Package: dataquieR
Title: Data Quality in Epidemiological Research
Version: 2.8.9
Authors@R: 
    c(person(given = "University Medicine Greifswald",
             role = "cph"),
      person(given = "Elisa",
             family = "Kasbohm",
             role = "aut",
             comment = c(ORCID = "0000-0001-5261-538X"),
             email = "elisa.kasbohm@uni-greifswald.de"),
      person(given = "Elena",
             family = "Salogni",
             role = "aut",
             comment = c(ORCID = "0009-0007-3767-7145"),
             email = "elena.salogni@uni-greifswald.de"),       
      person(given = "Joany",
             family = "Marino",
             role = "aut",
             comment = c(ORCID = "0000-0002-4657-3758"),
             email = "joany.marino@uni-greifswald.de"),
      person(given = "Adrian",
             family = "Richter",
             role = "aut",
             comment = c(ORCID = "0000-0002-3372-2021"),
             email = "richtera@uni-greifswald.de"),
      person(given = "Carsten Oliver",
             family = "Schmidt",
             role = "aut",
             comment = c(ORCID = "0000-0001-5266-9396"),
             email = "carsten.schmidt@uni-greifswald.de"),
      person(given = "Stephan",
             family = "Struckmann",
             role = c("aut", "cre"),
             comment = c(ORCID = "0000-0002-8565-7962"),
             email = "stephan.struckmann@uni-greifswald.de"),
      person(given = "German Research Foundation (DFG SCHM 2744/3-1, SCHM 2744/9-1, SCHM 2744/3-4)",
             role = "fnd"),
      person(given = "National Research Data Infrastructure for Personal Health Data: (NFDI 13/1)",
             role = "fnd"),
      person(given = "European Union’s Horizon 2020 programme (euCanSHare, grant agreement No. 825903)",
             role = "fnd"))
Description: Data quality assessments guided by a
    'data quality framework introduced by Schmidt and colleagues, 2021' 
    <doi:10.1186/s12874-021-01252-7> target the
    data quality dimensions integrity, completeness, consistency, and
    accuracy. The scope of applicable functions rests on the
    availability of extensive metadata which can be provided in
    spreadsheet tables. Either standardized (e.g. as 'html5' reports) or
    individually tailored reports can be generated. For an introduction
    into the specification of corresponding metadata, please refer to the
    'package website'
    <https://dataquality.qihs.uni-greifswald.de/VIN_Annotation_of_Metadata.html>.
License: BSD_2_clause + file LICENSE
URL: https://dataquality.qihs.uni-greifswald.de/
BugReports: https://gitlab.com/libreumg/dataquier/-/work_items/issues
Depends: R (>= 3.6.0)
Imports: dplyr (>= 1.0.2), emmeans, ggplot2 (>= 3.5.0), lme4,
        lubridate, MASS, parallelMap, patchwork (>= 1.3.0), R.devices,
        rlang, robustbase, qmrparser, utils, rio, readr, scales, withr,
        lifecycle, units, methods, hms
Suggests: S7 (>= 0.2.1), cowplot, grid, openxlsx2, grDevices, jsonlite,
        cli, whoami, DT (>= 0.23), htmltools, knitr, markdown,
        parallel, parallelly, rmarkdown, rstudioapi, testthat (>=
        3.1.9), tibble, vdiffr, pkgload, Rdpack, callr, colorspace,
        plotly (>= 4.11.0), htmlwidgets, future, processx, R6, shiny,
        xml2, mgcv, rvest, textutils, dbx, grImport2, rsvg, stringdist,
        rankICC, nnet, ordinal, storr, reticulate, stringi, lobstr,
        visNetwork
VignetteBuilder: knitr
Encoding: UTF-8
KeepSource: FALSE
Language: en-US
Config/testthat/parallel: true
Config/testthat/edition: 3
Config/testthat/start-first: dq_report_by_sm, dq_report2,
        dq_report_by_arguments, dq_report_by_pipesymbol_list,
        dq_report_by_s, dq_report_by_m, util_handle_complex_data_types,
        int_encoding_errors, plots, acc_loess, com_item_missingness,
        dq_report_by_na, dq_report_by_directories,
        con_limit_deviations, con_contradictions_redcap,
        com_segment_missingness, util_correct_variable_use
BuildManual: TRUE
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-11 16:23:28 UTC; struckmanns
Author: University Medicine Greifswald [cph],
  Elisa Kasbohm [aut] (ORCID: <https://orcid.org/0000-0001-5261-538X>),
  Elena Salogni [aut] (ORCID: <https://orcid.org/0009-0007-3767-7145>),
  Joany Marino [aut] (ORCID: <https://orcid.org/0000-0002-4657-3758>),
  Adrian Richter [aut] (ORCID: <https://orcid.org/0000-0002-3372-2021>),
  Carsten Oliver Schmidt [aut] (ORCID:
    <https://orcid.org/0000-0001-5266-9396>),
  Stephan Struckmann [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8565-7962>),
  German Research Foundation (DFG SCHM 2744/3-1, SCHM 2744/9-1, SCHM
    2744/3-4) [fnd],
  National Research Data Infrastructure for Personal Health Data: (NFDI
    13/1) [fnd],
  European Union’s Horizon 2020 programme (euCanSHare, grant agreement
    No. 825903) [fnd]
Maintainer: Stephan Struckmann <stephan.struckmann@uni-greifswald.de>
Repository: CRAN
Date/Publication: 2026-05-11 19:30:02 UTC
Built: R 4.6.0; ; 2026-05-11 22:49:20 UTC; unix

Package: dataquieR
Title: Data Quality in Epidemiological Research
Version: 2.8.9
Authors@R: 
    c(person(given = "University Medicine Greifswald",
             role = "cph"),
      person(given = "Elisa",
             family = "Kasbohm",
             role = "aut",
             comment = c(ORCID = "0000-0001-5261-538X"),
             email = "elisa.kasbohm@uni-greifswald.de"),
      person(given = "Elena",
             family = "Salogni",
             role = "aut",
             comment = c(ORCID = "0009-0007-3767-7145"),
             email = "elena.salogni@uni-greifswald.de"),       
      person(given = "Joany",
             family = "Marino",
             role = "aut",
             comment = c(ORCID = "0000-0002-4657-3758"),
             email = "joany.marino@uni-greifswald.de"),
      person(given = "Adrian",
             family = "Richter",
             role = "aut",
             comment = c(ORCID = "0000-0002-3372-2021"),
             email = "richtera@uni-greifswald.de"),
      person(given = "Carsten Oliver",
             family = "Schmidt",
             role = "aut",
             comment = c(ORCID = "0000-0001-5266-9396"),
             email = "carsten.schmidt@uni-greifswald.de"),
      person(given = "Stephan",
             family = "Struckmann",
             role = c("aut", "cre"),
             comment = c(ORCID = "0000-0002-8565-7962"),
             email = "stephan.struckmann@uni-greifswald.de"),
      person(given = "German Research Foundation (DFG SCHM 2744/3-1, SCHM 2744/9-1, SCHM 2744/3-4)",
             role = "fnd"),
      person(given = "National Research Data Infrastructure for Personal Health Data: (NFDI 13/1)",
             role = "fnd"),
      person(given = "European Union’s Horizon 2020 programme (euCanSHare, grant agreement No. 825903)",
             role = "fnd"))
Description: Data quality assessments guided by a
    'data quality framework introduced by Schmidt and colleagues, 2021' 
    <doi:10.1186/s12874-021-01252-7> target the
    data quality dimensions integrity, completeness, consistency, and
    accuracy. The scope of applicable functions rests on the
    availability of extensive metadata which can be provided in
    spreadsheet tables. Either standardized (e.g. as 'html5' reports) or
    individually tailored reports can be generated. For an introduction
    into the specification of corresponding metadata, please refer to the
    'package website'
    <https://dataquality.qihs.uni-greifswald.de/VIN_Annotation_of_Metadata.html>.
License: BSD_2_clause + file LICENSE
URL: https://dataquality.qihs.uni-greifswald.de/
BugReports: https://gitlab.com/libreumg/dataquier/-/work_items/issues
Depends: R (>= 3.6.0)
Imports: dplyr (>= 1.0.2), emmeans, ggplot2 (>= 3.5.0), lme4,
        lubridate, MASS, parallelMap, patchwork (>= 1.3.0), R.devices,
        rlang, robustbase, qmrparser, utils, rio, readr, scales, withr,
        lifecycle, units, methods, hms
Suggests: S7 (>= 0.2.1), cowplot, grid, openxlsx2, grDevices, jsonlite,
        cli, whoami, DT (>= 0.23), htmltools, knitr, markdown,
        parallel, parallelly, rmarkdown, rstudioapi, testthat (>=
        3.1.9), tibble, vdiffr, pkgload, Rdpack, callr, colorspace,
        plotly (>= 4.11.0), htmlwidgets, future, processx, R6, shiny,
        xml2, mgcv, rvest, textutils, dbx, grImport2, rsvg, stringdist,
        rankICC, nnet, ordinal, storr, reticulate, stringi, lobstr,
        visNetwork
VignetteBuilder: knitr
Encoding: UTF-8
KeepSource: FALSE
Language: en-US
Config/testthat/parallel: true
Config/testthat/edition: 3
Config/testthat/start-first: dq_report_by_sm, dq_report2,
        dq_report_by_arguments, dq_report_by_pipesymbol_list,
        dq_report_by_s, dq_report_by_m, util_handle_complex_data_types,
        int_encoding_errors, plots, acc_loess, com_item_missingness,
        dq_report_by_na, dq_report_by_directories,
        con_limit_deviations, con_contradictions_redcap,
        com_segment_missingness, util_correct_variable_use
BuildManual: TRUE
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-11 16:23:28 UTC; struckmanns
Author: University Medicine Greifswald [cph],
  Elisa Kasbohm [aut] (ORCID: <https://orcid.org/0000-0001-5261-538X>),
  Elena Salogni [aut] (ORCID: <https://orcid.org/0009-0007-3767-7145>),
  Joany Marino [aut] (ORCID: <https://orcid.org/0000-0002-4657-3758>),
  Adrian Richter [aut] (ORCID: <https://orcid.org/0000-0002-3372-2021>),
  Carsten Oliver Schmidt [aut] (ORCID:
    <https://orcid.org/0000-0001-5266-9396>),
  Stephan Struckmann [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8565-7962>),
  German Research Foundation (DFG SCHM 2744/3-1, SCHM 2744/9-1, SCHM
    2744/3-4) [fnd],
  National Research Data Infrastructure for Personal Health Data: (NFDI
    13/1) [fnd],
  European Union’s Horizon 2020 programme (euCanSHare, grant agreement
    No. 825903) [fnd]
Maintainer: Stephan Struckmann <stephan.struckmann@uni-greifswald.de>
Repository: CRAN
Date/Publication: 2026-05-11 19:30:02 UTC
Built: R 4.6.0; ; 2026-05-11 20:36:11 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$",dataquieR_result)
S3method("$",dataquieR_resultset2)
S3method("$",dq_lazy_ggplot)
S3method("$",dq_lazy_ggplot_s7)
S3method("$<-",dataquieR_resultset2)
S3method("%%",numeric_with_unit)
S3method("%/%",numeric_with_unit)
S3method("&",dq_lazy_ggplot)
S3method("*",dq_lazy_ggplot)
S3method("*",numeric_with_unit)
S3method("+",dq_lazy_ggplot)
S3method("+",numeric_with_unit)
S3method("-",dq_lazy_ggplot)
S3method("-",numeric_with_unit)
S3method("/",dq_lazy_ggplot)
S3method("/",numeric_with_unit)
S3method("[",dataquieR_result)
S3method("[",dataquieR_resultset2)
S3method("[<-",dataquieR_resultset2)
S3method("[[",dataquieR_result)
S3method("[[",dataquieR_resultset2)
S3method("[[",dq_lazy_ggplot)
S3method("[[",dq_lazy_ggplot_s7)
S3method("[[<-",dataquieR_resultset2)
S3method("^",numeric_with_unit)
S3method("names<-",dataquieR_translated)
S3method("|",dq_lazy_ggplot)
S3method(as.character,dataquieR_translated)
S3method(as.character,interval)
S3method(as.data.frame,dataquieR_resultset)
S3method(as.list,dataquieR_resultset)
S3method(as.list,dataquieR_resultset2)
S3method(dim,dataquieR_resultset2)
S3method(dimnames,dataquieR_resultset2)
S3method(droplevels,ReportSummaryTable)
S3method(ggplot_build,dq_lazy_ggplot)
S3method(ggplot_build,dq_lazy_ggplot_s7)
S3method(grid::grid.draw,svg_plot_proxy)
S3method(grid::grid.draw,util_pairs_ggplot_panels)
S3method(knitr::knit_print,ReportSummaryTable)
S3method(plot,dataquieR_summary)
S3method(prep_extract_summary,dataquieR_result)
S3method(prep_extract_summary,dataquieR_resultset2)
S3method(print,DataSlot)
S3method(print,ReportSummaryTable)
S3method(print,Slot)
S3method(print,StudyDataSlot)
S3method(print,TableSlot)
S3method(print,dataquieR_result)
S3method(print,dataquieR_resultset)
S3method(print,dataquieR_resultset2)
S3method(print,dataquieR_summary)
S3method(print,dataquieR_translated)
S3method(print,dq_lazy_ggplot)
S3method(print,dq_lazy_ggplot_s7)
S3method(print,interval)
S3method(print,list)
S3method(print,master_result)
S3method(print,numeric_with_unit)
S3method(print,svg_plot_proxy)
S3method(print,util_pairs_ggplot_panels)
S3method(rbind,ReportSummaryTable)
S3method(resnames,dataquieR_resultset2)
S3method(summary,dataquieR_resultset)
S3method(summary,dataquieR_resultset2)
S3method(util_undisclose,Slot)
S3method(util_undisclose,data.frame)
S3method(util_undisclose,dataquieR_result)
S3method(util_undisclose,dataquieR_resultset2)
S3method(util_undisclose,default)
S3method(util_undisclose,dq_lazy_ggplot)
S3method(util_undisclose,dq_lazy_ggplot_s7)
S3method(util_undisclose,gg)
S3method(util_undisclose,ggmatrix_fn_with_params)
S3method(util_undisclose,ggmatrix_plot_obj)
S3method(util_undisclose,ggplot_built)
S3method(util_undisclose,list)
S3method(util_undisclose,svg_plot_proxy)
S3method(util_undisclose,util_pairs_ggplot_panels)
export(.get_internal_api)
export(API_VERSION)
export(ASSOCIATION_DIRECTION)
export(ASSOCIATION_FORM)
export(ASSOCIATION_METRIC)
export(ASSOCIATION_RANGE)
export(CHECK_ID)
export(CHECK_LABEL)
export(CODE_CLASS)
export(CODE_CLASSES)
export(CODE_INTERPRET)
export(CODE_LABEL)
export(CODE_LIST_TABLE)
export(CODE_ORDER)
export(CODE_VALUE)
export(COMPUTATION_RULE)
export(CONTRADICTION_TERM)
export(CONTRADICTION_TYPE)
export(DATA_PREPARATION)
export(DATA_TYPES)
export(DATA_TYPES_OF_R_TYPE)
export(DF_CODE)
export(DF_ELEMENT_COUNT)
export(DF_ID_REF_TABLE)
export(DF_ID_VARS)
export(DF_NAME)
export(DF_RECORD_CHECK)
export(DF_RECORD_COUNT)
export(DF_UNIQUE_ID)
export(DF_UNIQUE_ROWS)
export(DISTRIBUTIONS)
export(GOLDSTANDARD)
export(MAHALANOBIS_THRESHOLD)
export(MISSING_CODE_RULES)
export(MULTIVARIATE_OUTLIER_CHECK)
export(MULTIVARIATE_OUTLIER_CHECKTYPE)
export(REL_VAL)
export(RULE)
export(SCALE_ACRONYM)
export(SCALE_LEVELS)
export(SCALE_NAME)
export(SEGMENT_ID_REF_TABLE)
export(SEGMENT_ID_TABLE)
export(SEGMENT_ID_VARS)
export(SEGMENT_MISS)
export(SEGMENT_PART_VARS)
export(SEGMENT_RECORD_CHECK)
export(SEGMENT_RECORD_COUNT)
export(SEGMENT_UNIQUE_ID)
export(SEGMENT_UNIQUE_ROWS)
export(SPLIT_CHAR)
export(VARATT_REQUIRE_LEVELS)
export(VARIABLE_LIST)
export(VARIABLE_LIST_ORDER)
export(VARIABLE_ROLES)
export(WELL_KNOWN_META_VARIABLE_NAMES)
export(acc_cat_distributions)
export(acc_distributions)
export(acc_distributions_ecdf)
export(acc_distributions_loc)
export(acc_distributions_only)
export(acc_distributions_prop)
export(acc_end_digits)
export(acc_loess)
export(acc_mahalanobis)
export(acc_mahalanobis_ratio)
export(acc_margins)
export(acc_multivariate_outlier)
export(acc_robust_univariate_outlier)
export(acc_shape_or_scale)
export(acc_univariate_outlier)
export(acc_varcomp)
export(com_item_missingness)
export(com_qualified_item_missingness)
export(com_qualified_segment_missingness)
export(com_segment_missingness)
export(com_unit_missingness)
export(con_contradictions)
export(con_contradictions_redcap)
export(con_inadmissible_categorical)
export(con_inadmissible_vocabulary)
export(con_limit_deviations)
export(contradiction_functions)
export(contradiction_functions_descriptions)
export(dataquieR.applicability_problem)
export(dataquieR.intrinsic_applicability_problem)
export(dataquieR.study_data_cache_metrics_env_default)
export(des_scatterplot_matrix)
export(des_summary)
export(des_summary_categorical)
export(des_summary_continuous)
export(dq_report)
export(dq_report2)
export(dq_report_by)
export(ggplotGrob.dq_lazy_ggplot)
export(ggplotGrob.dq_lazy_ggplot_s7)
export(ggplotly.dq_lazy_ggplot)
export(ggplotly.dq_lazy_ggplot_s7)
export(html_dependency_jspdf)
export(int_all_datastructure_dataframe)
export(int_all_datastructure_segment)
export(int_datatype_matrix)
export(int_duplicate_content)
export(int_duplicate_ids)
export(int_encoding_errors)
export(int_part_vars_structure)
export(int_sts_element_dataframe)
export(int_sts_element_segment)
export(int_unexp_elements)
export(int_unexp_records_dataframe)
export(int_unexp_records_segment)
export(int_unexp_records_set)
export(nres)
export(pipeline_recursive_result)
export(pipeline_vectorized)
export(plotly_build.dq_lazy_ggplot)
export(plotly_build.dq_lazy_ggplot_s7)
export(prep_acc_distributions_with_ecdf)
export(prep_add_cause_label_df)
export(prep_add_computed_variables)
export(prep_add_data_frames)
export(prep_add_missing_codes)
export(prep_add_to_meta)
export(prep_apply_coding)
export(prep_check_for_dataquieR_updates)
export(prep_check_meta_data_dataframe)
export(prep_check_meta_data_segment)
export(prep_check_meta_names)
export(prep_clean_labels)
export(prep_combine_report_summaries)
export(prep_compare_meta_with_study)
export(prep_create_meta)
export(prep_create_meta_data_file)
export(prep_create_storr_factory)
export(prep_datatype_from_data)
export(prep_deparse_assignments)
export(prep_deregister_progress_hook)
export(prep_dq_data_type_of)
export(prep_expand_codes)
export(prep_extract_cause_label_df)
export(prep_extract_classes_by_functions)
export(prep_extract_summary)
export(prep_fix_meta_id_dups)
export(prep_get_data_frame)
export(prep_get_labels)
export(prep_get_study_data_segment)
export(prep_get_user_name)
export(prep_get_variant)
export(prep_guess_encoding)
export(prep_is_translated)
export(prep_link_escape)
export(prep_list_dataframes)
export(prep_list_voc)
export(prep_load_folder_with_metadata)
export(prep_load_report)
export(prep_load_report_from_backend)
export(prep_load_workbook_like_file)
export(prep_map_labels)
export(prep_merge_study_data)
export(prep_meta_data_v1_to_item_level_meta_data)
export(prep_min_obs_level)
export(prep_open_in_excel)
export(prep_pmap)
export(prep_prepare_dataframes)
export(prep_purge_data_frame_cache)
export(prep_realize_ggplot)
export(prep_register_progress_hook)
export(prep_remove_from_cache)
export(prep_render_pie_chart_from_summaryclasses_ggplot2)
export(prep_render_pie_chart_from_summaryclasses_plotly)
export(prep_robust_guess_data_type)
export(prep_save_report)
export(prep_scalelevel_from_data_and_metadata)
export(prep_set_backend)
export(prep_study2meta)
export(prep_summary_to_classes)
export(prep_title_escape)
export(prep_undisclose)
export(prep_unsplit_val_tabs)
export(prep_valuelabels_from_data)
export(pro_applicability_matrix)
export(resnames)
export(to_basic.GeomPointrangeRobust)
import(patchwork)
import(qmrparser)
importFrom(MASS,fitdistr)
importFrom(dplyr,"%>%")
importFrom(dplyr,recode)
importFrom(ggplot2,aes)
importFrom(ggplot2,annotate)
importFrom(ggplot2,coord_cartesian)
importFrom(ggplot2,coord_flip)
importFrom(ggplot2,discrete_scale)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_text)
importFrom(ggplot2,expand_limits)
importFrom(ggplot2,expansion)
importFrom(ggplot2,facet_grid)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_bar)
importFrom(ggplot2,geom_boxplot)
importFrom(ggplot2,geom_col)
importFrom(ggplot2,geom_count)
importFrom(ggplot2,geom_density)
importFrom(ggplot2,geom_errorbar)
importFrom(ggplot2,geom_histogram)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_jitter)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_path)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_pointrange)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_tile)
importFrom(ggplot2,geom_violin)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggplot_build)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,labs)
importFrom(ggplot2,position_jitter)
importFrom(ggplot2,scale_alpha)
importFrom(ggplot2,scale_alpha_manual)
importFrom(ggplot2,scale_color_gradientn)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,scale_colour_manual)
importFrom(ggplot2,scale_fill_gradientn)
importFrom(ggplot2,scale_fill_manual)
importFrom(ggplot2,scale_linetype_manual)
importFrom(ggplot2,scale_size_continuous)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_x_datetime)
importFrom(ggplot2,scale_x_discrete)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,scale_y_discrete)
importFrom(ggplot2,sec_axis)
importFrom(ggplot2,stat_ecdf)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,vars)
importFrom(ggplot2,waiver)
importFrom(ggplot2,xlab)
importFrom(ggplot2,xlim)
importFrom(ggplot2,ylab)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRamp)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,gray.colors)
importFrom(grDevices,hcl.colors)
importFrom(grDevices,rgb)
importFrom(graphics,barplot)
importFrom(graphics,hist)
importFrom(methods,new)
importFrom(rlang,":=")
importFrom(rlang,.data)
importFrom(rlang,caller_fn)
importFrom(stats,IQR)
importFrom(stats,aggregate)
importFrom(stats,alias)
importFrom(stats,as.formula)
importFrom(stats,binomial)
importFrom(stats,complete.cases)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,dgamma)
importFrom(stats,dnorm)
importFrom(stats,dunif)
importFrom(stats,glm)
importFrom(stats,lm)
importFrom(stats,loess)
importFrom(stats,lowess)
importFrom(stats,mad)
importFrom(stats,mahalanobis)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,p.adjust)
importFrom(stats,pgamma)
importFrom(stats,pnorm)
importFrom(stats,poisson)
importFrom(stats,predict)
importFrom(stats,punif)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,reorder)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,var)
importFrom(utils,adist)
importFrom(utils,browseURL)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,object.size)
importFrom(utils,osVersion)
importFrom(utils,packageName)
importFrom(utils,packageVersion)
importFrom(utils,tail)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# `dataquieR`

<!-- badges: start -->

[![minimal R
version](https://img.shields.io/badge/R%3E%3D-3.6.0-6666ff.svg)](https://cran.r-project.org/)
[![Pipeline
Status](https://gitlab.com/libreumg/dataquier/badges/master/pipeline.svg?ignore_skipped=true)](https://libreumg.gitlab.io/dataquier/)
[![Coverage](https://codecov.io/gl/libreumg/dataquier/branch/master/graph/badge.svg?token=79TK6GQTMG)](https://app.codecov.io/gl/libreumg/dataquier)
[![CRAN-Version](https://www.r-pkg.org/badges/version/dataquieR)](https://cran.r-project.org/package=dataquieR)
![Latest
Release](https://gitlab.com/libreumg/dataquier/-/badges/release.svg)
[![DOI](https://img.shields.io/badge/DOI-10.32614%2FCRAN.package.dataquieR-00be00.svg)](https://doi.org/10.32614/CRAN.package.dataquieR)
[![CRAN-Downloads](https://cranlogs.r-pkg.org/badges/dataquieR)](https://www.r-pkg.org/pkg/dataquieR)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![`Lifecycle`](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![license](https://img.shields.io/badge/license-BSD_2_clause%20+%20file%20LICENSE-00be00.svg)](https://opensource.org/license/bsd-2-clause)
[![DOI](https://joss.theoj.org/papers/10.21105/joss.03093/status.svg)](https://doi.org/10.21105/joss.03093)
[![DOI](https://joss.theoj.org/papers/10.21105/joss.06581/status.svg)](https://doi.org/10.21105/joss.06581)

<!-- badges: end -->

The goal of `dataquieR` is to provide functions for assessing data
quality issues in studies, that can be used alone or in a data quality
pipeline. `dataquieR` also implements one generic pipeline producing
`htmltools` based HTML5 reports.

See also

[`https://dataquality.qihs.uni-greifswald.de`](https://dataquality.qihs.uni-greifswald.de)

------------------------------------------------------------------------

## Installation

You can install the released version of `dataquieR` from
[CRAN](https://CRAN.R-project.org/package=dataquieR) with:

``` r
install.packages("dataquieR")
```

The suggested packages can be directly installed by:

``` r
install.packages("dataquieR", dependencies = TRUE)
```

The developer version from
[`GitLab.com`](https://gitlab.com/libreumg/dataquier) can be installed
using:

``` r
if (!requireNamespace("remotes")) {
  install.packages("remotes")
}
remotes::install_gitlab("libreumg/dataquier")
```

For examples and additional documentation, please refer to our
[website](https://dataquality.qihs.uni-greifswald.de).

## dataquieR usage questionnaire

To help us improve `dataquieR`, we invite you to provide your feedback
by completing this short survey ([English](https://3x7.de/kd4l6) or
[German](https://3x7.de/4f884) version).

## Suggested packages

`dataquieR` reports can now use
[`plotly`](https://cran.r-project.org/package=plotly) if installed. That
means that, in the final report, you can zoom in the figures and get
information by hovering on the points, etc. To install `plotly` type:

``` r
install.packages("plotly")
```

To install all suggested packages, run:

``` r
prep_check_for_dataquieR_updates()
```

This command can also check for new beta releases of `dataquieR` from
our own server, so not from `CRAN`:

``` r
prep_check_for_dataquieR_updates(beta = TRUE)
```

***Hint*** If you are running `dataquieR` in an un-trusted setting,
namely, inside a server application, please consider disabling the
import of R-serialization files to prevent users from importing `RData`
(or `RDS` or even `R`) files, that trigger code execution on your
machine, see, e.g., [Ivan Krylov’s
blog](https://aitap.github.io/2024/05/02/unserialize.html) for the
reason:

``` r
# prevent rio from reading potentially code-containing files
options(rio.import.trust = FALSE)
```

If you do so, the example data won’t be loaded any more.

If you are using a version \>= 2.0.0 of `rio`, this will be the default,
so for running our examples, then, you’ll have to trust our files by
using e.g.
`withr::with_options(list(rio.import.trust = FALSE), prep_get_data_frame("study_data"))`
for loading our example study data into the data-frame cache, initially
and trusting our files loaded from

- <https://dataquality.qihs.uni-greifswald.de/extdata/study_data.RData>
- <https://dataquality.qihs.uni-greifswald.de/extdata/meta_data.RData>
- <https://dataquality.qihs.uni-greifswald.de/extdata/ship_meta.RDS>
- <https://dataquality.qihs.uni-greifswald.de/extdata/ship_subset1.RDS>
- <https://dataquality.qihs.uni-greifswald.de/extdata/ship_subset2.RDS>
- <https://dataquality.qihs.uni-greifswald.de/extdata/ship_subset3.RDS>
- <https://dataquality.qihs.uni-greifswald.de/extdata/ship.RDS>

## References

- [Software Paper](https://doi.org/10.21105/joss.06581) [![JOSS
  Article](https://joss.theoj.org/papers/10.21105/joss.06581/status.svg)](https://doi.org/10.21105/joss.06581)
- [Software Paper](https://doi.org/10.21105/joss.03093) [![JOSS
  Article](https://joss.theoj.org/papers/10.21105/joss.03093/status.svg)](https://doi.org/10.21105/joss.03093)
- [Data Quality Concept
  Paper](https://doi.org/10.1186/s12874-021-01252-7)
- [Data Quality Concept and Software Web
  Site](https://dataquality.qihs.uni-greifswald.de)

## Funding – see also [here](https://dataquality.qihs.uni-greifswald.de/Contact.html)

- German Research Foundation (`https://www.dfg.de/`) (DFG:
  `SCHM 2744/3–1` – initial concept and dataquieR development,
  `SCHM 2744/9-1` – `NFDI` Task Force `COVID-19` use case application;
  `SCHM 2744/3-4` – concept extensions, ongoing )

- [European Union’s Horizon 2020 research and innovation
  program](https://research-and-innovation.ec.europa.eu/funding/funding-opportunities/funding-programmes-and-open-calls/horizon-2020_en):
  [euCanSHare, grant agreement No. 825903](http://www.eucanshare.eu/) –
  [dataquieR](https://cran.r-project.org/package=dataquieR) refinements
  and implementations in the
  [Square2](https://pubmed.ncbi.nlm.nih.gov/28423853/) web application.

- [National Research Data Infrastructure for Personal Health
  Data](https://www.nfdi4health.de/en/): `NFDI 13/1` – extension based
  on revised metadata concept, ongoing.

- German National Cohort (NAKO Gesundheitsstudie) NAKO
  (`https://nako.de/`): `BMBF` (`https://www.bmbf.de/`): `01ER1301A` and
  `01ER1801A`

This dataset lists all the Media Types (MIME types), Media Subtypes, and their file extensions.

## Data

The details of the Media Types and Media Subtypes are taken from the [official registry of Media Types](http://www.iana.org/assignments/media-types/media-types.xhtml) maintained by IANA. The extension details are taken   [the website](http://svn.apache.org/viewvc/httpd/httpd/branches/2.2.x/docs/conf/mime.types?view=annotate) of the Apache Software Foundation.

## Preparation

The Type, Subtype, and Template name were copied from IANA's websites into a Google Sheets document. The link to the Template was generated in a fourth column in the same sheet by concatenating the Template name with a reference to the Template folder on IANA's website.

The extensions were copied from Apache's website into a separate sheet in the same Google Sheets document. The data was cleaned to place the extensions on their own in a single column without the Type and Subtype.

The extensions were finally added to the original sheet using VLOOKUP.


## License

These data are made available under the Public Domain Dedication and License v1.0 whose full text can be found at: http://www.opendatacommons.org/licenses/pddl/1.0/

#' Abbreviate snake_case function names to shortened `CamelCase`
#'
#' @param x a vector of indicator function names
#'
#' @return abbreviations
#'
#' @seealso [base::abbreviate]
#' @family process_functions
#' @concept reporting
#' @noRd
util_abbreviate <- function(x) {
  util_expect_scalar(x, allow_more_than_one = TRUE, allow_null = TRUE,
                     allow_na = TRUE, check_type = is.character)
  r <- vapply(strsplit(x, "_", fixed = TRUE), function(xx) {
    prefix <- head(xx, 1)
    known_prefix <- c("acc", "com", "con", "int", "des")
    if (any(prefix %in% known_prefix)) {
      prefix[prefix %in% known_prefix] <- c(
        "acc" = "a",
        "com" = "m",
        "con" = "c",
        "int" = "i",
        "des" = "d"
      )[[prefix[prefix %in% known_prefix]]]
    }
    suffix <- gsub("^(.)(..).*$", "\\U\\1\\E\\2", tail(xx, -1), perl = TRUE)
    paste0(prefix, paste(suffix, collapse = ""), collapse = "")
  },  FUN.VALUE = character(1))
  newly_ambig <- duplicated(x) != duplicated(r)
  while (any(newly_ambig)) {
    r[newly_ambig] <- paste0(r[newly_ambig], "\u00b0")
    newly_ambig <- duplicated(x) != duplicated(r)
  }
  r
}

#' Check for duplicated content
#'
#' @description
#' This function tests for duplicates entries in the data set. It is possible to
#' check duplicated entries by study segments or to consider only selected
#' segments.
#'
#' @param level [character] a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").
#' @param identifier_name_list [vector] the vector that contains the name of the identifier to be used in the assessment. For the study level, corresponds to the names of the different data frames. For the segment level, indicates the name of the segments.
#' @param study_data [data.frame] the data frame that contains the measurements, mandatory.
#' @param meta_data [data.frame] the data frame that contains metadata attributes of the study data, mandatory.
#' @param id_vars_list [list] the list containing the identifier variables names to be used in the assessment.
#' @param unique_rows [vector] named. for each segment, either true/false or `no_id` to exclude ID variables from check
#'
#' @inheritParams .template_function_indicator
#'
#' @return a [list] with
#'   - `SegmentData`: data frame with the results of the quality check for duplicated entries
#'   - `SegmentTable`: data frame with selected duplicated entries check results, used for the data quality report.
#'   - `Other`: vector with row indices of duplicated entries, if any, otherwise NULL.
#'
#' @family integrity_indicator_functions
#' @concept integrity_indicator
#' @keywords internal
util_int_duplicate_content_segment <- function(level = c("segment"),
                                               identifier_name_list,
                                               id_vars_list,
                                               unique_rows,
                                               study_data,
                                               meta_data, # item_level,
                                               meta_data_segment = "segment_level",
                                               segment_level) {
  label_col <- VAR_NAMES
  util_ck_arg_aliases()

  # Segment level check ----

  # Check arguments ----

  level <- util_match_arg(level)

  if (missing(identifier_name_list) &&
      missing(id_vars_list) &&
      missing(meta_data_segment) &&
      formals()$meta_data_segment %in% prep_list_dataframes()) {
    meta_data_segment <- force(meta_data_segment)
  }

  if (missing(identifier_name_list) &&
      missing(id_vars_list) &&
      !missing(meta_data_segment)) {
    meta_data_segment <- prep_check_meta_data_segment(meta_data_segment)
    meta_data_segment <-
      meta_data_segment[
        !util_empty(meta_data_segment[[SEGMENT_UNIQUE_ROWS]]) & (
        trimws(tolower(meta_data_segment[[SEGMENT_UNIQUE_ROWS]])) ==
          "no_id" |
        !util_is_na_0_empty_or_false(
          meta_data_segment[[SEGMENT_UNIQUE_ROWS]])), , drop = FALSE]
    # TODO: if nothing left
    identifier_name_list <- meta_data_segment[[STUDY_SEGMENT]];
    id_vars_list <- lapply(setNames(meta_data_segment[[SEGMENT_ID_VARS]], # TODO: use the constants everywhere: meta_data_segment[[SEGMENT_ID_VARS]], not ...$SEGMENT_ID_VARS
                                    nm = meta_data_segment[[STUDY_SEGMENT]]),
                           util_parse_assignments,
                           multi_variate_text = TRUE
    )
    id_vars_list <- lapply(id_vars_list, unlist, recursive = TRUE)
    unique_rows <- setNames(meta_data_segment[[SEGMENT_UNIQUE_ROWS]],
                            nm = meta_data_segment[[STUDY_SEGMENT]])
  } else if (!missing(meta_data_segment)) {
    util_error(c("I have %s and one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "id_vars_list"
                 )),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "id_vars_list"
                 )))
  } else if (missing(meta_data_segment) && (
    missing(identifier_name_list) ||
    missing(id_vars_list)
  )) {
    util_error(c("I don't have %s and also miss at least",
                 "one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "id_vars_list"
                 )),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "id_vars_list"
                 )))
  }

  util_expect_scalar(identifier_name_list,
                     allow_null = TRUE,
                     allow_more_than_one = TRUE,
                     check_type = is.character)

  if (missing(unique_rows)) {
    unique_rows <- setNames(rep(NA_character_, length(identifier_name_list)),
                            identifier_name_list);
  }
  util_expect_scalar(unique_rows, allow_more_than_one = TRUE,
                     check_type =
                       function(x) {
                         all(util_empty(x) | tolower(trimws(x)) %in%
                               c("f", "t", "true", "false", "no_id"))
                       })
  unique_rows <- tolower(trimws(unique_rows))
  unique_rows[util_empty(unique_rows)] <- "false"

  # map metadata to study data
  prep_prepare_dataframes(.allow_empty = TRUE)

  # check that specified segments are included in the metadata
  old_identifier_name_list <- identifier_name_list
  identifier_name_list <- intersect(identifier_name_list,
                                 meta_data[[STUDY_SEGMENT]])

  if (length(old_identifier_name_list) > length(identifier_name_list)) {
    util_message(
      "The segments in the %s do not match the segments in %s, considering only the intersection",
      dQuote("meta_data"),
      dQuote("meta_data_segment"),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = FALSE
    )
  }

  # Check for duplicated content  ----

  result <- lapply(setNames(nm = identifier_name_list), function(current_segment) {
    vars_in_current_segment <-
      util_get_vars_in_segment(segment = current_segment,
                               meta_data = meta_data,
                               label_col = VAR_NAMES)

    if (trimws(tolower(unique_rows[[current_segment]])) == "no_id") {
      vars_in_current_segment <- setdiff(vars_in_current_segment,
                                         id_vars_list[[current_segment]])
    }

    ds1_seg <- ds1[, intersect(colnames(ds1), vars_in_current_segment), FALSE]

    ds1_seg <- util_remove_empty_rows(ds1_seg)

    n_uniq <- nrow(unique(ds1_seg))

    res_tmp <- data.frame(
      check.names = FALSE,
      "Check" = "Duplicate records",
      "Segment" = current_segment,
      "Any duplicates" = ifelse(n_uniq < nrow(ds1_seg), TRUE, FALSE),
      "Number of duplicates" = nrow(ds1_seg) - n_uniq,
      "Percentage of duplicates" =
        round(100 * (nrow(ds1_seg) - n_uniq) / nrow(ds1_seg), 3),
      "GRADING" = ifelse(n_uniq < nrow(ds1_seg), 1, 0),
      stringsAsFactors = FALSE
    )

    if (res_tmp[[3]]) { # only if there are any duplicated observations
      vec_dup <- which(duplicated(ds1_seg)) # FIXME: the positions in the reduced data frame are mostly useless
      vec_dup <- NULL # TODO: Remove after fixing
    } else {
      vec_dup <- NULL
    }

    return(list(res_tmp, vec_dup))
  })

  res_df <- do.call(rbind.data.frame, lapply(result, `[[`, 1))
  duplicates_df <- do.call(cbind.data.frame, lapply(result, `[[`, 2))

  res_pipeline <- data.frame(
    "Segment" = res_df$Segment,
    "NUM_int_sts_dupl_content" = res_df$`Number of duplicates`, # TODO: Nowhere use numbers, always names, and return ordere here to use standardized names. labels for the not-pipelin-res can be read from an external file, late
    "PCT_int_sts_dupl_content" = res_df$`Percentage of duplicates`,
    "GRADING" = res_df$GRADING,
    stringsAsFactors = FALSE
  )

  return(list(
    SegmentData = res_df,
    SegmentTable = res_pipeline,
    Other = duplicates_df
  ))


}

#' Utility function to prepare the metadata for location checks
#'
#' @param resp_vars [variable list] the names of the measurement variables
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param report_problems [enum] Should missing metadata information be reported
#'                               as error, warning or message?
#'
#' @return a [list] with the location metric (mean or median) and expected
#'         range for the location check
#'
#' @family lookup_functions
#' @concept metadata_management
#' @noRd
util_prep_location_check <- function(resp_vars,
                                     meta_data,
                                     report_problems = c("error", "warning",
                                                         "message"),
                                     label_col = VAR_NAMES) {
  resp_vars <- prep_get_labels(resp_vars = resp_vars,
                               item_level = meta_data,
                               label_col = label_col,
                               force_label_col = "TRUE")
  report_problems <- match.arg(report_problems)
  rep_fun <- switch(report_problems,
                    error = util_error,
                    warning = util_warning,
                    message = util_message)
  loc_metric <- setNames(util_find_var_by_meta(resp_vars = resp_vars,
                                               target = "LOCATION_METRIC",
                                               meta_data = meta_data,
                                               allowed_sources = label_col),
                         nm = resp_vars)
  loc_metric <- trimws(tolower(loc_metric))

  if (!all(loc_metric[!util_empty(loc_metric)] %in% c("median", "mean"))) {
    rep_fun("Location checks can only be performed for mean or median values.",
            applicability_problem = TRUE,
            intrinsic_applicability_problem = TRUE)
    loc_metric[which(!(loc_metric %in% c("median", "mean")))] <- NA
  }

  loc_range <- setNames(util_find_var_by_meta(resp_vars = resp_vars,
                                              target = "LOCATION_RANGE",
                                              meta_data = meta_data,
                                              allowed_sources = label_col),
                        nm = resp_vars)
  loc_range <- lapply(loc_range, util_parse_interval)

  rvs_with_metric <- names(loc_metric)[which(!is.na(loc_metric))]
  rvs_with_range <- names(loc_range)[which(!is.na(loc_range))]
  rvs_with_both <- intersect(rvs_with_metric, rvs_with_range)
  rvs_with_none <- setdiff(resp_vars, unique(c(rvs_with_metric,
                                               rvs_with_range)))

  if (!all.equal(rvs_with_metric, rvs_with_range)) {
    rep_fun(paste0(
      "For ",
      paste(setdiff(union(rvs_with_metric, rvs_with_range),
                    rvs_with_both),
            collapse = ", "),
      ", the metadata for the expected location is incomplete. Both ",
      "LOCATION_RANGE and LOCATION_METRIC are required."),
      applicability_problem = TRUE)
    loc_metric[setdiff(resp_vars, rvs_with_both)] <- NA
    loc_range[setdiff(resp_vars, rvs_with_both)] <- NA
  }

  if (length(rvs_with_none) > 0) {
    rep_fun(paste0(
      "For ", paste(rvs_with_none, collapse = ", "),
      ", the metadata for a location check is missing (LOCATION_RANGE and ",
      "LOCATION_METRIC)."),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE)
  }

  loc_meta <- list("Metric" = loc_metric,
                   "Range" = loc_range)
  return(loc_meta)
}

.util_condition_once_seen <- new.env(parent = emptyenv())

util_condition_once_seen <- function(id) {
  exists(id, envir = .util_condition_once_seen, inherits = FALSE)
}

util_condition_once_mark_seen <- function(id) {
  assign(id, TRUE, envir = .util_condition_once_seen)
  invisible(TRUE)
}

util_clean_condition_once_cache <- function() {
  rm(list = ls(.util_condition_once_seen),
     envir = .util_condition_once_seen)
  invisible(NULL)
}

#' Produce a condition function
#'
#' @param .condition_type [character] the type of the conditions being created
#'   and signaled by the function, "error", "warning", or "message"
#'
#' @family condition_functions
#' @concept process
#' @noRd
util_condition_constructor_factory <- function(
                       .condition_type = c("error", "warning", "message")) {

  .condition_type <- match.arg(.condition_type)

  .signal_fkt <- switch (.condition_type,
                         error = stop,
                         warning = warning,
                         message = message)

  .cond_constructor <-
    switch (.condition_type,
            error = rlang::error_cnd,
            warning = rlang::warning_cnd,
            message = rlang::message_cnd)

  .caller_control_att <- paste0("dataquieR.",
                                toupper(.condition_type),
                                "S_WITH_CALLER")
  .caller_control_att_default <- get(paste0(.caller_control_att, "_default"))

  function(m, ..., applicability_problem = NA,
           intrinsic_applicability_problem = NA,
           integrity_indicator = "none", level = 0, immediate, title = "",
           additional_classes = c(), varname = NULL, once_id) {
    invis <- FALSE
    if (identical(Sys.getenv("TESTTHAT"), "true")) { # TODO: use the ensure_suggested/is_testing pattern, but do this efficiently, then
      if(!isTRUE(getOption("dataquieR.testthat_expect_message_active", NULL))) {
        invis <- TRUE
      }
    }
    if (identical(getOption("dataquieR.debug", dataquieR.debug_default),
                  TRUE)) {
      browser() # intended use of browser() -- dont modify this line
    }
    if (missing(immediate)) {
      immediate <- FALSE
    }
    m_args <- eval(quote(force(list(...))))
    if (!missing(once_id)) {
      util_expect_scalar(once_id, check_type = is.character)
    }
    # shows some false positive note on possible misplaced
    # ...
    # m_args <- lapply(
    #   rlang::call_args(rlang::call_match(dots_expand = FALSE))[["..."]], eval,
    #   envir = parent.frame())
    util_expect_scalar(integrity_indicator, allow_na = TRUE,
                       check_type = is.character)
    if (!(integrity_indicator %in% c(na.omit(subset(util_get_concept_info("dqi"),
                                                    get("Dimension") == "Integrity",
                                                    select = "abbreviation",
                                                    drop = TRUE)), "none"))) {
      util_error(
        "Internal error: %s is not a supported %s. Did you update %s?",
        dQuote(integrity_indicator),
        sQuote("integrity_indicator"),
        sQuote("dqi.rds"))
    }
    if (integrity_indicator == "none") {
      integrity_indicator <- NA_character_
    }
    util_stop_if_not(length(applicability_problem) == 1 &&
                       is.logical(applicability_problem))
    util_stop_if_not(length(intrinsic_applicability_problem) == 1 &&
                       is.logical(intrinsic_applicability_problem))

    start_from_call <- util_find_first_externally_called_functions_in_stacktrace()
    start_from_call <- length(sys.calls()) - start_from_call # refers to reverted sys.calls, so mirror the number
    caller. <- sys.call(1)
    calling <- character(0)
    if (FALSE) { # (!is.na(start_from_call)) {
      try(silent = TRUE, {
        caller. <- sys.call(start_from_call)
        calling <- util_deparse1(sys.call(start_from_call + 1))
        calling <- paste("when calling", calling)
      })
    } else {
      start_from_call <- 1
    }

    # https://stat.ethz.ch/pipermail/r-help/2011-November/295273.html
    str <- vapply(FUN.VALUE = character(1),
                  rev(sys.calls()), function(sc)
                    paste0(deparse(sc, nlines = 2),
                           collapse = "\n"))[
                             -seq_len(start_from_call)]
    # if (!!length(str)) {
    #   stacktrace <- (paste0(paste0("> ", str),
    #                         collapse = "\n"))
    # } else {
      stacktrace <- character(0)
    # }

    if (identical(as.logical(getOption("dataquieR.CONDITIONS_WITH_STACKTRACE", dataquieR.CONDITIONS_WITH_STACKTRACE_default)), FALSE)) {
      stacktrace <- ""
      if ((exists(".called_in_pipeline") && .called_in_pipeline) ||
          .condition_type != "error") {
        calling <- character(0)
      }
    } else {
    }
    if (identical(as.logical(getOption(.caller_control_att, .caller_control_att_default)), FALSE)) {
      caller. <- NULL
    }
    if (inherits(m, "try-error")) {
      m <- attr(m, "condition")
    }
    if (inherits(m, "condition")) {
      .m <- m
      m <- paste0(title, conditionMessage(m))
      if (m == "") {
        if (inherits(.m, "error")) {
          m <- "Error"
        } else if (inherits(.m, "warning")) {
          m <- "Warning"
        } else if (inherits(.m, "message")) {
          m <- "Message"
        } else {
          m <- "Condition -- should not be displayed, sorry. Please report."
        }
      }

      if (isTRUE(getOption("dataquieR.traceback",
                         dataquieR.traceback_default))) {
        tc <- rlang::trace_back(bottom = 2)
      } else {
        tc <- NULL
      }
      ec <-
        .cond_constructor(message = paste(c(m, calling, stacktrace),
                                          collapse = "\n"),
                          trace = tc,
                          use_cli_format = (!exists(".called_in_pipeline") ||
                                              !.called_in_pipeline),
                          call = caller.)

    } else {
      mm <- paste0(m,
                   collapse =
                     " ")
      if (nchar(mm) > 8192) {
        mm <- substr(mm, 1, 8192)
        mm <- sub("(%[^%]$)", "\\1", mm, perl = TRUE)
      }
      if (isTRUE(getOption("dataquieR.traceback",
                           dataquieR.traceback_default))) {
        tc <- rlang::trace_back(bottom = 2)
      } else {
        tc <- NULL
      }
      ec <-
        .cond_constructor(trace = tc,
                          use_cli_format = (!exists(".called_in_pipeline") ||
                                              !.called_in_pipeline),
                          message = paste0(c(do.call("sprintf", c(
          list(fmt = paste0(title, mm)),
          m_args)),
                                   calling, stacktrace), collapse = "\n"),
                    call = caller.)
    }
    attr(ec, "applicability_problem") <- applicability_problem
    attr(ec, "intrinsic_applicability_problem") <- intrinsic_applicability_problem
    attr(ec, "integrity_indicator") <- integrity_indicator
    attr(ec, "varname") <- varname
    dq_err_classes <- character(0)
    if (isTRUE(intrinsic_applicability_problem)) {
      dq_err_classes <- unique(c(
        dq_err_classes,
        dataquieR.intrinsic_applicability_problem
      ))
    }
    if (isTRUE(applicability_problem)) {
      dq_err_classes <- unique(c(
        dq_err_classes,
        dataquieR.applicability_problem
      ))
    }
    class(ec) <- unique(c(additional_classes, dq_err_classes, class(ec)))
    if (!isTRUE(.dq2_globs$.called_in_pipeline) && !missing(once_id)) {
      if (util_condition_once_seen(once_id)) {
        return(invisible(ec))
      }
      util_condition_once_mark_seen(once_id)
    }
    if (level >= getOption("dataquieR.CONDITIONS_LEVEL_TRHESHOLD",
                  dataquieR.CONDITIONS_LEVEL_TRHESHOLD_default) ||
        inherits(ec, "error")) {
      # .signal_fkt(ec)
      if (immediate && inherits(ec, "warning")) {
        cat("In",
            as.character(conditionCall(ec)),
            ":\n",
            conditionMessage(ec),
            "\n",
            file = stderr()) # rlang currently only calls warning
      }
      if (inherits(ec, "error")) {
        rlang::cnd_signal(ec)
      } else {
        if (inherits(ec, "warning")) { # TODO: add something like rlang::warn(.frequency = "once")
          rlang::cnd_signal(ec)
        } else {
          x <- capture.output(rlang::cnd_signal(ec), file = NULL, type = "message")
          if (!invis && length(x)) {
            cat(sep = "\n", x, file = stderr())
            if (!endsWith(x[length(x)], "\n")) {
              cat("\n", file = stderr())
            }
          }
        }
      }
    }
    invisible(ec)
  }
}

#' Pre-load a folder with named (usually more than) one table(s)
#'
#' The original purpose of this function is to load metadata, not study data.
#' If you want to load study data, you should keep them in a different folder,
#' then you can call this function once for the metadata and once for the study
#' data but this time setting `keep_types = TRUE` to avoid all data being read
#' as [character()].
#'
#' Note, that once loaded to the data frame cache, a file won't be read again,
#' except you call [prep_purge_data_frame_cache()] or
#' [prep_remove_from_cache()]. That is, if you call this function first, and
#' [prep_get_data_frame()] later, of if `dataquieR` wants to read a file, e.g.,
#' for [dq_report2()], the file will come from the cache in the way it was
#' initially read in (`keep_types` may thus be used inadequately).
#'
#' By default, this function works not recursively, but you can tweak that by
#' passing `...`-arguments passed through to the initially running
#' [list.files()] function.
#'
#' These can thereafter be referred to by their names only. Such files are,
#' e.g., spreadsheet-workbooks or `RData`-files.
#'
#' Note, that this function in contrast to [prep_get_data_frame] does neither
#' support selecting specific sheets/columns from a file.
#'
#' @param folder the folder name to load.
#' @param keep_types [logical] keep types as possibly defined in the file.
#'                             set `TRUE` for study data.
#' @param append [logical] if a data frame already exists in the cache
#'                         (by name), extend the existing one
#' @param ... arguments passed to [list.files()]
#'
#' @return `invisible(the cache environment)`
#' @export
#' @seealso [prep_add_data_frames]
#' @seealso [prep_get_data_frame]
#' @family data-frame-cache
prep_load_folder_with_metadata <- function(folder,
                                         keep_types = FALSE,
                                         append = FALSE,
                                         ...) {
  util_expect_scalar(append, check_type = is.logical)
  util_expect_scalar(folder, check_type = is.character)
  util_stop_if_not(
    "full.names not supported by prep_load_folder_with_metadata" =
                     (!"full.names" %in% rlang::call_args_names(
                       rlang::current_call())))
# IDEA: If folder is actually a zip file with zip extensions, uncompress first
# IDEA: If an html file is loaded, extract tables from it
# IDEA: Parse data:-URLs
# IDEA: Support iframes
  if (startsWith(folder, "https://") ||
      startsWith(folder, "http://")) {
    util_ensure_suggested(
      "rvest",
      goal =
        "read data from the internet using prep_load_folder_with_metadata()")
    fp <- tempfile()
    util_stop_if_not(!file.exists(fp))
    dir.create(fp)
    on.exit(try(unlink(fp, force = TRUE, recursive = TRUE, expand = FALSE),
                silent = TRUE))

    file_new <- file.path(fp, "index.html")
    try(utils::download.file(folder, destfile = file_new,
                             quiet = TRUE, mode = "wb"), silent = TRUE)
    if (file.exists(file_new)) {
      fl <- try(rvest::read_html(file_new), silent = TRUE)
      if (inherits(fl, "try-error")) {
        util_error("Could not read index from %s (%s): %s",
                   dQuote(folder),
                   dQuote(file_new),
                   conditionMessage(attr(fl, "condition")))
      }
      links <- rvest::html_nodes(fl, "a")
      all_refs <- rvest::html_attr(links, 'href')
      all_refs[!startsWith(tolower(all_refs), "http://") &
                 !startsWith(tolower(all_refs), "https://")] <-
        paste0(folder,
               "/", all_refs[!startsWith(tolower(all_refs), "http://") &
                           !startsWith(tolower(all_refs), "https://")])
      all_refs <- trimws(all_refs)
      lapply(
        all_refs,
        function(ref) {
          #ref_dec <- utils::URLdecode(ref)
          rf <- gsub("^.*\\/", "", ref, perl = TRUE)
          rf <- gsub("\\?.*$", "", rf)
          rf <- gsub("#.*$", "", rf)
          #rf_path <- gsub(print(rf), "", ref)
          ext <- ""
          ext <- try(util_fetch_ext(ref), silent = TRUE)
          # do not ignore content-disposition headers sent by the server (if they propose a file name)
          if (length(ext) != 1 ||
              !is.character(ext)) {
            msg <- "unknown reason"
            if (inherits(ext, "try-error")) {
              msg <- conditionMessage(attr(ext, "condition"))
            } else if (inherits(ext, "condition")) {
              msg <- conditionMessage(ext)
            }
            util_warning("Could not determine the file type of %s: %s",
                         dQuote(ref),
                         sQuote(msg))
            ext <- ""
          } else {
            if (!is.null(attr(ext, "file-name"))) {
              if (length(attr(ext, "file-name")) == 1 &&
                  !is.na(attr(ext, "file-name")))
                rf <- attr(ext, "file-name")
            }
            ext <- paste0(".", ext)
          }

          if (!endsWith(rf, ext)) {
            rf <- paste0(rf, ext)
          }

          try(utils::download.file(ref,
                                   destfile = file.path(fp, rf),
                                   quiet = TRUE, mode = "wb"),
              silent = TRUE)
        }
      )
      unlink(file_new, force = TRUE)
      folder <- fp
    }
  }

  util_stop_if_not(`Folder not found` = dir.exists(folder))
  util_stop_if_not(`Access denied` = file.access(folder) == 0)

  fls <- list.files(folder,
                    full.names = TRUE,
                    all.files = TRUE,
                    no.. = TRUE,
                    include.dirs = FALSE,
                    ...)
  if (any(startsWith(basename(fls), "~$") &
          (endsWith(basename(fls), ".xlsx") |
           endsWith(basename(fls), ".xls")
          ))) {
    util_warning(
      c("Found files that look like Excel",
        "working copies https://superuser.com/a/901749",
        "Do you have Excel open? This may cause warnings about",
        "files that cannot be opened/are locked. If not open, maybe, Excel",
        "had crashed before. You should close Excel, if it is not running,",
        "open it, it may address this. If nothing helps, consider moving",
        "all files whose names start with ~$ and end with .xls or xlsx away.")
    )
  }
  if (any(startsWith(basename(fls), ".~lock.") &
          endsWith(basename(fls), "#") # libreoffice
          )) {
    util_warning(
      c("Found files that look like LibreOffice/OpenOffice/StarOffice",
        "working copies https://superuser.com/a/901749",
        "Do you have one of these apps open? This may cause warnings about",
        "files that cannot be opened/are locked. If not open, maybe, one of",
        "these apps had crashed before. You should close all such apps.",
        "If they don't run, open them, they may address this. ",
        "If nothing helps, consider moving",
        "all files whose names start with .~lock. and end with # away.")
    )
  }
  lapply(fls, function(fn) {
    if (inherits(suppressWarnings(try(prep_load_workbook_like_file(fn,
                                                                   keep_types = keep_types,
                                                                   append = append),
                                      silent = TRUE)),
        "try-error")) {
      if (inherits(suppressWarnings(try(prep_get_data_frame(fn,
                                                            keep_types = keep_types),
                                        silent = TRUE)),
          "try-error")) {
        util_warning("Could not load %s, ignoring...", dQuote(fn))
      }
    }
  })

  invisible(.dataframe_environment())
}

test_that("prep_load_folder_with_metadata works", {
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("rvest")
  skip_on_cran()
  prep_purge_data_frame_cache()
  expect_equal(length(prep_list_dataframes()), 0)
  suppressWarnings(prep_load_folder_with_metadata("https://dataquality.qihs.uni-greifswald.de/extdata/fortests"))
  expect_gt(length(prep_list_dataframes()), 10)
})

#' Generate a HTML-based report from a [dataquieR] report
#'
#' @param x [dataquieR report v2][dq_report2].
#' @param dir [character] directory to store the rendered report's files,
#'   a temporary one,
#'   if omitted. Directory will be created, if missing
#' @param view [logical] display the report
#' @param disable_plotly [logical] do not use `plotly`, even if installed
#' @param block_load_factor [numeric] see [dataquieR.print_block_load_factor]
#' @param advanced_options [list] options to set during report computation,
#'                                see [options()]
#' @param ... additional arguments:
#' @param dashboard [logical] dashboard mode: `TRUE`: create a dashboard only,
#'                            `FALSE`: don't create a dashboard at all,
#'                            `NA` or missing: create a "normal" report with
#'                            a dashboard included.
#' @param cores [integer] number of cpu cores to use or a named list with
#'                        arguments for [parallelMap::parallelStart] or NULL,
#'                        if parallel has already been started by the caller.
#'                        Can also be a cluster.
#' @param force_overwrite [logical] force to overwrite `dir`, even if it exists
#'
#' @return file names of the generated report's HTML files
#' @export
#' @importFrom utils browseURL
print.dataquieR_resultset2 <- function(
    x,
    dir, # TODO: consolidate dir or output_dir see dq_report2 and dq_report_by
    view = TRUE,
    disable_plotly = FALSE,
    block_load_factor = getOption("dataquieR.print_block_load_factor",
                                  dataquieR.print_block_load_factor_default),
    advanced_options =  list(),
    dashboard = NA,
    force_overwrite = FALSE,
    ...,
    cores = list(mode = "socket",
                 logging = FALSE,
                 cpus = util_detect_cores(),
                 load.balancing = TRUE)) {

  util_ensure_suggested("parallel")


  util_expect_scalar(block_load_factor,
                     check_type =
                       util_is_numeric_in(min = 1, whole_num = FALSE),
                     convert_if_possible = as.numeric,
                     error_message =
                       sprintf(
                         "%s should be a number above 0",
                         sQuote("block_load_factor")))

  if (suppressWarnings(util_ensure_suggested("testthat", err = FALSE))) {
    if (testthat::is_testing()) {
      if (missing(cores)) {
        cores <- 2
      }
      if (!rlang::is_scalar_integerish(cores) ||
          cores > 2L) {
        if (!is.null(cores))
          util_warning(
            "Internal problem: %s should be an integer below %s in the context of %s",
            sQuote("cores"), dQuote("3"), sQuote("testthat"))
          cores <- 2
      }
    }
  }

  if (!is.null(cores)) {
    if (inherits(cores, "cluster")) {
      old_def_cl <- parallel::getDefaultCluster()
      parallel::setDefaultCluster(cores)
      on.exit(parallel::setDefaultCluster(old_def_cl), add = TRUE)
      old_o_def_cl <- options(
        parallelMap.cpus = length(parallel::getDefaultCluster()),
        parallelMap.load.balancing = TRUE,
        parallelMap.mode = "socket"
      )
      on.exit(options(old_o_def_cl), add = TRUE)
    } else if (inherits(cores, "list")) {
      suppressMessages(do.call(parallelMap::parallelStart, cores))
      on.exit({Sys.sleep(2);suppressMessages(parallelMap::parallelStop());}, add = TRUE) # whyever, rstudio needs these two seconds, it hangs, otherwise.
      on.exit({suppressMessages(gc(full = TRUE));}, add = TRUE)
    } else {
      suppressMessages(parallelMap::parallelStart("socket", cpus = cores,
                                                  logging = FALSE,
                                                  load.balancing = TRUE))
      on.exit({Sys.sleep(2);suppressMessages(parallelMap::parallelStop());}, add = TRUE) # whyever, rstudio needs these two seconds, it hangs, otherwise.
      on.exit(suppressMessages(gc(full = TRUE)), add = TRUE)
    }
    cores <- NULL
  }
  util_stop_if_not(is.list(advanced_options))
  old_O <- options(
    c(
      list(
        dataquieR.CONDITIONS_WITH_STACKTRACE = FALSE,
        dataquieR.ERRORS_WITH_CALLER = FALSE,
        dataquieR.MESSAGES_WITH_CALLER = FALSE,
        dataquieR.WARNINGS_WITH_CALLER = FALSE,
        DT.warn.size = FALSE
      ),
      advanced_options
    )
  )
  .old_.dq2_globs_.called_in_pipeline <- .dq2_globs$.called_in_pipeline
  .dq2_globs$.called_in_pipeline <- TRUE
  on.exit({
    .dq2_globs$.called_in_pipeline <- .old_.dq2_globs_.called_in_pipeline
    options(old_O)
  }, add = TRUE)

  opts <- c(as.list(environment()), list(...))
  opts$x <- NULL
  for (arg_name in setdiff(names(formals()), "...")) {
    missing_in_parent <- eval(call("missing", as.symbol(arg_name)))
    if (missing_in_parent) {
      opts[[arg_name]] <- NULL
    }
  }
  util_ensure_suggested(pkg = c("htmltools",
                                "DT", "rmarkdown",
                                "markdown"),
                        goal = "generate plain HTML-reports.")

  if (nrow(x) * ncol(x) * nres(x) == 0) {
    all_errors <- lapply(x, function(res) {
      e <- attr(res, "error")
      if (length(e)) {
        util_stop_if_not(length(e) == 1)
        e <- e[[1]]
        if (!is.null(attr(e, "applicability_problem")) &&
            is.logical(attr(e, "applicability_problem")) &&
            identical(attr(e, "applicability_problem"), TRUE)) {
          "Applicability problem(s)"
        } else {
          paste(conditionMessage(e), collapse = "\n") # nocov
        }
      } else {
        NULL
      }
    })
    all_errors <- all_errors[!vapply(all_errors, is.null, FUN.VALUE =
                                       logical(1))]
    all_errors <- unlist(all_errors, recursive = TRUE)
    all_errors <- gsub("\nwhen calling .*$", "", all_errors)
    all_errors <- unique(all_errors)
    if (length(all_errors) > 0)
      reason <- paste0(", possible reasons are: \n", paste("-", all_errors, collapse = "\n"))
    else
      reason <- "."
    util_error("Report is empty, no results at all%s",
               reason)
  }

  dir <- NULL

  template <- "default"

  if ("dir" %in% names(opts)) {
    dir <- opts[["dir"]]
    if (!is.character(dir) || length(dir) != 1) {
      util_error("dir must be a character(1)")
      dir <- NULL
    }
  }
  if ("template" %in% names(opts)) {
    template <- opts[["template"]]
    if (!is.character(template) || length(template) != 1)
      util_error("template must be a character(1)")
  }
  if ("view" %in% names(opts)) {
    view <- opts[["view"]]
    if (!is.logical(view) || length(view) != 1 || is.na(view))
      util_error("view must be a logical(1)")
  }
  if ("by_report" %in% names(opts)) {
    by_report <- opts[["by_report"]]
    if (!is.logical(by_report) || length(by_report) != 1 || is.na(by_report))
      util_error("by_report must be a logical(1)")
  } else {
    by_report <- FALSE
  }


  util_setup_rstudio_job("Rendering dq_report2 to HTML...",
                         n = 2 * nrow(x) * ncol(x))
  start_time <- Sys.time()

  if (length(dir) == 0) {
    dir <- tempfile()
    util_message("No output directory given (with dir=), setting it to %s",
                 dQuote(dir))
  }

  content_dir <- util_normalize_path(dir)
  content_file <- file.path(content_dir, "index.html")

  progress_msg("Page generation", sprintf("Writing to %s", dQuote(content_file)))

  dir <- file.path(dir, ".report")

  util_expect_scalar(force_overwrite, check_type = is.logical,
                     error_message = sprintf("%s must be a logical value",
                                             sQuote("force_overwrite")))

  if (!getOption("dataquieR.resume_print", dataquieR.resume_print_default)) {
    util_overwrite_if_requested(dir, force_overwrite)
  } else {
    if (!dir.create(dir, recursive = TRUE, showWarnings = FALSE)) {
      util_error("Could not create %s", dQuote(dir))
    }
  }

  dir <- util_normalize_path(dir)
  old_dir <- setwd(dir)
  on.exit({
    setwd(old_dir)
  }, add = TRUE)


  old_opt <- options(DT.warn.size=FALSE)
  on.exit({
    options(old_opt)
  }, add = TRUE)


  report <- x # the report object is defined here to be available
  rm(x)

  # rep_id <- attr(report, "properties")$rep_id
  rep_id <- util_make_report_id() # use a different ID for each rendering

  packageName <- utils::packageName()

  if (!inherits(attr(report, "min_render_version"), "numeric_version") ||
      attr(report, "min_render_version") > as.numeric_version("1.0.0")
  ) {
    util_warning(
      c("This report may be not rendered correctly by",
        "this version of %s (i.e., %s). I'll try, but consider",
        "upgradig %s, please:\n install.packages(\"%s\")"),
      packageName,
      as.character(packageVersion(packageName)),
      packageName,
      packageName
    )
  }

  if (!is.null(attr(report, "translation_version"))) {
    if (translation_version != attr(report, "translation_version")) {
      util_warning(
        c("This report may be not rendered correctly by",
          "this version of %s. The translations may have changed.",
          "Version in report: %s, version in installed %s: %s"),
        sQuote(packageName),
        sQuote(attr(report, "translation_version")),
        sQuote(paste(packageName, as.character(packageVersion(packageName)))),
        sQuote(translation_version)
      )
    }
  }

  util_message("Compiling HTML report, please wait...")


  template_file <- system.file("templates", template, "report.html", package =
                                 packageName)

  if (template_file == "") {
    if (file.exists(template)) {
      template_file <- template
    } else {
      util_error("Template %s not found", dQuote(template))
    }
  }

  title <- attr(report, "title")
  if (is.null(title)) {
    title <- "Data Quality Report"
  }

  .hi <- .hp <- .hm <- NULL
  content_file <- NULL
  if (!missing(dir)) {
    content_file <- file.path(dirname(dir), "index.html")
    list2env(util_init_html_progress(output_dir = content_dir,
                            content_file = content_file,
                            title = title,
                            view = view,
                            rep_id = rep_id), envir = environment())
  }

  if (is.na(dashboard) || dashboard) { # DASHBOARD
    progress_msg("Computing dashboard", "Computing dashboard...start")
    my_dashboard <- util_setup_dashboard(report, make_links = is.na(dashboard))
    progress_msg("Computing dashboard...", "Computing dashboard...done")
  } else {
    my_dashboard <- NULL
  }

  logo <- "logo.png"
  # <link rel="icon" type="image/png" href="{{ logo }}">
  if (!is.na(dashboard) && dashboard) {
    progress_msg("Writing dashboard...", "Start")
    my_dashboard <- htmltools::tagList(
      htmltools::tags$head(
        htmltools::tags$title(title),
        htmltools::htmlTemplate(
          text_ = '<link rel="icon" type="image/png" href="{{ logo }}">',
          logo = logo,
          document_ = FALSE)
      ),
      htmltools::HTML("<!-- done -->"),
      my_dashboard
    )
    # important: no progress update after this line should replace the index.html any more, so deregister
    prep_deregister_progress_hook(.hi, verbose = FALSE)
    prep_deregister_progress_hook(.hp, verbose = FALSE)
    prep_deregister_progress_hook(.hm, verbose = FALSE)
    progress_msg("Writing dashboard...", "Done")
    htmltools::save_html(my_dashboard,
                         libdir = file.path(content_dir, "lib"),
                         file = content_file)
    util_write_renderinfo_js_json(
      content_dir,
      rep_id,
      start_time = start_time,
      end_time = Sys.time()
    )
    fnms <- setNames(list(content_file), nm = paste0("dash-", title))
    if (util_really_rstudio()) {
      rstudioapi::executeCommand("activateConsole")
    }

  } else {

    progress_msg("Creating HTML files in memory...",
                 "Creating HTML files in memory...start")
    pages <- util_generate_pages_from_report(report, template,
                                             disable_plotly = disable_plotly,
                                             progress = progress,
                                             progress_msg = progress_msg,
                                             block_load_factor =
                                               block_load_factor,
                                             dir = dir,
                                             my_dashboard = my_dashboard)
    progress_msg("Creating HTML files in memory...",
                 "Creating HTML files in memory...finalize")

    all_ids <- util_extract_all_ids(pages)
    js_escape_string <- function(x) {
      x <- gsub("\\\\", "\\\\\\\\", x)  # Backslash first
      x <- gsub("\"", "\\\\\"", x)      # "
      x <- gsub("\n", "\\\\n", x)
      x <- gsub("\r", "\\\\r", x)
      x <- gsub("\t", "\\\\t", x)
      x
    }
    cat(sep = "",
        "window.all_ids = {\"all_ids\": ",
        paste0("[",
               paste0('"', js_escape_string(all_ids), '"', collapse = ", "),
               "]"),
        "}",
        file = file.path(dir, "anchor_list.js")
    )
    saveRDS(object = all_ids, file = file.path(dir, "anchor_list.RDS"))

    loading <-
      htmltools::includeHTML(
        system.file("loading.html",
                    package =
                      packageName))

    jqui <- rmarkdown::html_dependency_jqueryui()
    jqui$stylesheet <- "jquery-ui.min.css"

    deps_prepro <- util_copy_all_deps(dir = dir,
                                      pages,
                                      rmarkdown::html_dependency_jquery(),
                                      jqui,
                                      html_dependency_clipboard(),
                                      html_dependency_tippy(),
                                      rmarkdown::html_dependency_font_awesome(),
                                      html_dependency_dataquieR(),
                                      html_dependency_jspdf()
    )

    progress_msg("Creating HTML files in memory...",
                 "Creating HTML files in memory...done")

    progress_msg("Writing HTML files from memory...",
                 "Writing HTML files from memory...start")

    fnms <- lapply(
      setNames(nm = seq_along(pages)),
      util_create_page_file,
      progress_msg = progress_msg,
      progress = progress,
      pages = pages,
      rendered_pages = deps_prepro$rendered_pages,
      dir = dir,
      template_file = template_file,
      report = report,
      logo = logo,
      loading = loading,
      packageName = packageName,
      deps = deps_prepro$deps,
      title = title,
      by_report = by_report
    )

    util_setup_rstudio_job("Page generation: Finalization",
                           length(pages))

    progress_msg("Writing HTML files from memory...",
                 "Writing HTML files from memory...thumbnails/iframes")

    util_write_iframe_results(pages,
                              progress_msg = progress_msg,
                              progress = progress,
                              block_load_factor = block_load_factor,
                              template_file = system.file("templates",
                                                          template,
                                                          "iframe.html",
                                                          package =
                                                            packageName),
                              dir = dir)

    progress_msg("Creating HTML files in memory...",
                 "Writing HTML files from memory...done")

    end_time <- Sys.time()

    util_hide_file_windows(dir)

    progress_msg(sprintf("Wrote %d files", length(fnms)),
                 sprintf("Wrote %s", util_pretty_vector_string(fnms, n_max = 20)))

    # important: no progress update after this line should replace the index.html any more, so deregister
    prep_deregister_progress_hook(.hi, verbose = FALSE)
    prep_deregister_progress_hook(.hp, verbose = FALSE)
    prep_deregister_progress_hook(.hm, verbose = FALSE)

    util_write_index_html(
      file.path(content_dir, "index.html"),
      util_index_redirect_lines(
        title = title,
        target_rel = paste0(".report/", basename(fnms[[1]])),
        delay_ms = 300L
      )
    )

    util_write_renderinfo_js_json(
      content_dir,
      rep_id,
      start_time = start_time,
      end_time = end_time
    )

    if (util_really_rstudio()) {
      rstudioapi::executeCommand("activateConsole")
    }
  }

  if (view) {
    invisible(fnms)
  } else {
    fnms
  }
}

#' print a list of `dataquieR_result` objects
#'
#' @param x [list()] only, if all elements inherit from [dataquieR_result],
#'          this implementation runs
#' @param ... passed to other implementations
#'
#' @return undefined
#' @export
print.list <- function(x, ...) {
  util_stop_if_not(is.list(x))
  if (length(x) == 0) {
    return(NextMethod())
  }
  if (isTRUE(getOption('knitr.in.progress'))) {
    return(NextMethod())
  }
  entities <- try(unique(
    vapply(x,
           function(dqr) attr(attr(dqr, "call"), "entity"),
           FUN.VALUE = character(1))), silent = TRUE)
  if (length(entities) == 1 &&
      !util_is_try_error(entities) &&
      length(x) > 1) {
    util_ensure_suggested("htmltools")
    f <- withr::local_tempdir(.local_envir = rlang::global_env())
    withr::local_dir(f)
    jqui <- rmarkdown::html_dependency_jqueryui()
    jqui$stylesheet <- "jquery-ui.min.css"
    entity <- attr(attr(x[[1]], "call"), "entity")
    outputs <- lapply(x, function(dqr) {
      if (is.null(dqr) || inherits(dqr, "dataquieR_NULL")) {
        return(htmltools::tagList())
      }
      cn <- attr(dqr, "cn")
      nm <- paste0(cn, ".", entity)
      if (is.null(cn)) {
        title <- ""
        cn <- "Data Quality"
        nm <- ""
      } else {
        title <-
          util_alias2caption(cn,
                             long = TRUE)
      }
      section <- htmltools::tagList(html_dependency_dataquieR(iframe = FALSE),
                                    html_dependency_jspdf(),
                                    jqui,
                                    htmltools::h2(title),
                                    util_pretty_print(dqr = dqr,
                                                      nm = nm,
                                                      is_single_var = TRUE,
                                                      use_plot_ly = (util_ensure_suggested("plotly",
                                                                                           goal = "plot interactive figures",
                                                                                           err = FALSE)),
                                                      dir = f))
    })
    outputs <- c(list(htmltools::h1(entity)),
                 outputs)
    doc <- do.call(htmltools::tagList, outputs)
    htmltools::save_html(doc, "index.html")
    # FIXME: knitinprogress
    if (!("view" %in% names(list(...))) || (!identical(list(...)[["view"]],
                                                       FALSE))) {
      viewer <- getOption("viewer", utils::browseURL)
      viewer("index.html")
    }
    invisible(NULL)
  } else if (identical(
    try(all(vapply(x, function(y) inherits(y, "dataquieR_result") &&
                   inherits(y, "master_result"), FUN.VALUE = logical(1)),
            na.rm = TRUE), silent = TRUE), TRUE)) {
    ERRs <- vapply(x, FUN.VALUE = logical(1), FUN = function(y) {
      length(attr(y, "error")) > 0
    })
    if (any(ERRs)) {
      for (e in x[ERRs]) {
        for (cnd in attr(e, "error"))
          warning(cnd)
      }
    }
    x <- x[!ERRs]
    all_slots <- unique(unlist(lapply(x, names)))
    all_slots <- all_slots[vapply(all_slots,
                                  function(slot)
                                    all(vapply(x, function(dqr) slot %in%
                                                 names(dqr),
                                               FUN.VALUE = logical(1)),
                                        na.rm = TRUE),
                                  FUN.VALUE = logical(1))]
    names(x) <- paste0(rep("r.", length(x)), seq_along(x))
    nm1 <- unlist(lapply(x, attr, "function_name"))
    nm2 <- unlist(lapply(x, function(.dqr) {
      attr(attr(.dqr, "call"), "entity_name")
    }))
    if (!!length(nm1) &&
        length(nm1) == length(nm2)) {
      names(x) <- paste0(nm1, ".", nm2)
    }
    if (length(all_slots) > 0) {
      x <- lapply(setNames(nm = all_slots), function(slot) {
        cr <- util_combine_res(lapply(x, `[`, slot))
        if (length(cr) == 1) {
          r <- cr[[1]][[slot]]
          rownames(r) <- NULL
        } else {
          old_r <- cr
          r <- lapply(cr, `[[`, slot)
          if (slot == "SummaryPlot") {
            if (util_ensure_suggested("plotly",
                                      goal = "plot interactive figures",
                                      err = FALSE)) {
              as_plotlys <- unique(lapply(old_r, attr, "as_plotly"))
              as_plotly <- NULL
              if (length(as_plotlys) == 1) {
                if (!is.null(as_plotlys[[1]]) &&
                    exists(as_plotlys[[1]], mode = "function"))
                  as_plotly <- get(as_plotlys[[1]], mode = "function")
              }
              if (is.function(as_plotly)) {
                my_plots <- lapply(old_r, as_plotly)
              } else {
                my_plots <- lapply(r, util_plot_figure_plotly) # IDEA: Do we need sizing_hints, here?
              }
              my_plots <- mapply(SIMPLIFY = FALSE,
                                 p = my_plots,
                                 nm = names(my_plots), function(p, nm) {
                                   title <- paste0("", nm)
                                   title <- util_sub_string_right_from_.(nm)
                                   if (inherits(p, "plotly")) {
                                     # p <- plotly::layout(p, width = "30%", height = "30%", margin = list(b = 120))
                                     p <- plotly::config(p, responsive = TRUE, displaylogo = FALSE)
                                   }
                                   react_size <- "
                  window.dq_rs2_script_init = false;
                  window.dq_rs2_script_init_fkt = function() {
                    if (window.dq_rs2_script_init) {
                      return;
                    }
                    window.dq_rs2_script_init = true;
                    var pys = $(document).find('.js-plotly-plot')
                    for (var i = 0; i < pys.length; i++) {
                      var py = pys[i]
                      Plotly.relayout(py, {
                                    width: py.getBoundingClientRect().width,
                                    height: py.getBoundingClientRect().height
                                  })
                    }
                  }
                  $(window.dq_rs2_script_init_fkt)
                "
                                   htmltools::div(style = "float:left;width:30%;height:33%;overflow-y:hidden;",
                                                  htmltools::h2(style = "position:relative;top:0;height:0;margin:0;padding:0;z-index:9;", title),
                                                  p,
                                                  htmltools::tags$script(type = "text/javascript",
                                                                         htmltools::HTML(react_size)))
                                 })

              my_plots <- htmltools::tagList(c(my_plots,
                                               list(htmltools::br(style =
                                                                    "clear: both;"))))
              r <- my_plots
            } else {
              r <- patchwork::wrap_plots(r, ncol = 2)
            }
          }
          class(r) <- union("dataquieR_result", class(r))
        }
        r
      })
    } else {
      return(NextMethod())
    }
    class(x) <- c("master_result")
    print.master_result(x)
  } else {
    NextMethod()
  }
}

#' Find all tags with `iframe`-attributes
#'
#' @param x `html` object to search
#' @param key key to find
#'
#' @returns character vector with all found keys
#' @noRd
util_collect_tag_attrs <- function(x, key = "html_file") {
  found <- list()
  found_env <- environment()

  recurse <- function(tag) {
    if (inherits(tag, "shiny.tag.list") || inherits(tag, "list")) {
      lapply(tag, recurse)
    } else if (inherits(tag, "shiny.tag")) {
      if (!is.null(attr(tag, key))) {
        found_env$found[[length(found_env$found) + 1]] <- tag
      }
      if (!is.null(tag$children)) {
        recurse(tag$children)
      }
    }
  }

  recurse(x)
  found
}

#' Save ggplot with guaranteed inner panel size (no down-scaling)
#'
#' Saves a plot ensuring that the inner panel area is at least the requested
#' size in inches. Axes and text are scaled accordingly. If necessary, the
#' overall image is enlarged to preserve the original aspect ratio and avoid
#' shrinking. Supports complex plots with faceting or patchwork layouts.
#'
#' @param plot A ggplot object or patchwork object
#' @param filename Path to the output file
#' @param inner_width Target width (in inches) of the panel area
#' @param inner_height Target height (in inches) of the panel area
#' @param dpi Output resolution (default: 300)
#' @param base_text_factor Font scaling factor (default: 11)
#'
#' @return Nothing. Saves the file and returns invisibly.
#' @noRd
util_save_with_inner_size <- function(plot, filename,
                                      inner_width, inner_height,
                                      dpi = 300,
                                      base_text_factor = 11) {


  is_pairs_plot <-
    identical(attr(plot, "sizing_hints")$figure_type_id, "pairs_plot")

  min_text_pt <- 6.5
  max_text_pt <- 12
  scale_factor <- mean(c(inner_width / 6, inner_height / 4))
  base_size <- base_text_factor * scale_factor
  base_size <- max(min_text_pt, min(base_size, max_text_pt))

  if (!is_pairs_plot)
    plot <- plot + util_theme_with_scaling_limits(base_size = base_size)

  pw_widths <- NULL
  pw_heights <- NULL
  if (inherits(plot, "patchwork")) {
    pw_widths <- try(plot$patches$widths, silent = TRUE)
    if (util_is_try_error(pw_widths)) {
      pw_widths <- NULL
    }
    pw_heights <- try(plot$patches$heights, silent = TRUE)
    if (util_is_try_error(pw_heights)) {
      pw_heights <- NULL
    }
  }

  if (!is_pairs_plot) plot <- util_gg_stagger_x_axis_labels(
                                                plot,
                                                inner_width = inner_width,
                                                inner_height = inner_height,
                                                patchwork_widths = pw_widths,
                                                patchwork_heights = pw_heights,
                                                theme_base_size = base_size
                                              )

  as_plot_gtable <- function(plot) {
    if (inherits(plot, "patchwork")) {
      util_ensure_suggested("patchwork", err = TRUE)
      return(patchwork::patchworkGrob(plot))
    }
    ggplot2::ggplotGrob(plot)
  }

  sum_width_in <- function(x) {
    if (!length(x)) return(0)
    sum(grid::convertWidth(x, "in", valueOnly = TRUE))
  }

  sum_height_in <- function(x) {
    if (!length(x)) return(0)
    sum(grid::convertHeight(x, "in", valueOnly = TRUE))
  }

  find_panel_tracks <- function(gt) {
    nm <- gt$layout$name

    idx_panel <- grep("^panel($|[-;])", nm)
    idx_panel_area <- which(nm == "panel-area")

    idx <- sort(unique(c(idx_panel, idx_panel_area)))

    if (!length(idx)) {
      idx <- grep("panel", nm, fixed = TRUE)
    }

    if (!length(idx)) {
      return(list(rows = integer(0), cols = integer(0)))
    }

    rows <- unique(unlist(Map(seq.int, gt$layout$t[idx], gt$layout$b[idx])))
    cols <- unique(unlist(Map(seq.int, gt$layout$l[idx], gt$layout$r[idx])))

    list(rows = sort(rows), cols = sort(cols))
  }

  measure_non_panel_overhead_in <- function(gt) {
    tracks <- find_panel_tracks(gt)

    n_rows <- length(gt$heights)
    n_cols <- length(gt$widths)

    panel_rows <- tracks$rows
    panel_cols <- tracks$cols

    non_panel_rows <- setdiff(seq_len(n_rows), panel_rows)
    non_panel_cols <- setdiff(seq_len(n_cols), panel_cols)

    list(
      overhead_width = sum_width_in(gt$widths[non_panel_cols]),
      overhead_height = sum_height_in(gt$heights[non_panel_rows]),
      panel_rows = panel_rows,
      panel_cols = panel_cols
    )
  }

  g <- try(as_plot_gtable(plot), silent = TRUE)
  if (util_is_try_error(g)) {
    g <- NULL
  }

  if (is.null(g)) {
    overhead_width <- 0
    overhead_height <- 0
  } else {
    dims <- measure_non_panel_overhead_in(g)
    overhead_width <- dims$overhead_width
    overhead_height <- dims$overhead_height
  }

  pad_left <- 0.03
  pad_right <- 0.03
  pad_bottom <- 0.04
  pad_top <- 0.08

  total_width <- inner_width + overhead_width + pad_left + pad_right
  total_height <- inner_height + overhead_height + pad_top + pad_bottom

  e <- try(ggplot2::ggsave(
    filename = filename,
    plot = plot,
    width = total_width,
    height = total_height,
    dpi = dpi,
    units = "in",
    limitsize = FALSE
  ), silent = TRUE)

  if (util_is_try_error(e)) {
    msg <- conditionMessage(attr(e, "condition"))
    msg_stripped <- msg
    if (suppressWarnings(util_ensure_suggested("cli", err = FALSE))) {
      msg_stripped <- cli::ansi_strip(msg_stripped)
    }

    e <- try(ggplot2::ggsave(
      filename = filename,
      plot = util_ggplot_text(msg_stripped),
      width = total_width,
      height = total_height,
      dpi = dpi,
      units = "in",
      limitsize = FALSE
    ), silent = TRUE)

    util_warning("Could not write %s: %s",
                 dQuote(filename),
                 sQuote(msg))
  }

  suppressMessages(gc(verbose = FALSE))
  invisible(NULL)
}

#' Internal theme with scalable font size (not exported)
#'
#' Creates a `ggplot2::theme()` object where text elements are scaled
#' proportionally to a given base font size. Font sizes for axis text, axis
#' titles, legend text, and plot titles are explicitly set, not using
#' `rel()`. The legend key size is scaled in relation to the axis text size.
#'
#' @param base_size Base font size in pt (e.g., 12)
#' @param rel_axis Relative size for axis and legend text (default: 0.65)
#' @param rel_title Relative size for axis titles and plot titles (default: 0.75)
#'
#' @return A ggplot2 theme object with consistently scaled text
#' @noRd
util_theme_with_scaling_limits <- function(base_size,
                                           rel_axis = 0.72,
                                           rel_title = 0.9,
                                           rel_subtitle = 0.8) {
  ggplot2::theme(
    text            = ggplot2::element_text(size = base_size),
    axis.text       = ggplot2::element_text(size = base_size * rel_axis),
    axis.text.x     = ggplot2::element_text(size = base_size * rel_axis),
    axis.text.y     = ggplot2::element_text(size = base_size * rel_axis),
    axis.title      = ggplot2::element_text(size = base_size * rel_title),
    axis.title.x    = ggplot2::element_text(size = base_size * rel_title),
    axis.title.y    = ggplot2::element_text(size = base_size * rel_title),
    legend.text     = ggplot2::element_text(size = base_size * rel_axis),
    legend.title    = ggplot2::element_text(size = base_size * rel_title),
    legend.key.size = ggplot2::unit(rel_axis, "lines"),
    plot.title      = ggplot2::element_text(size = base_size * rel_title),
    plot.subtitle   = ggplot2::element_text(size = base_size * rel_subtitle)
  )
}

#' Write `iframe` HTML and thumbnails (parallel-aware)
#'
#' works serially, if no cluster runs. in parallel mode, it works
#' block-wise to show a progress bar.
#'
#' @param pages           list of shiny/HTML tags
#' @param progress_msg    `function(msg_short, msg_long)`
#' @param progress        `function(percent)`
#' @param block_load_factor [numeric] see
#'   [dataquieR.print_block_load_factor]
#' @param template_file file to use as a template for the HTML output
#' @param dir [character] directory to store the rendered report's files
#'
#' @return invisible(NULL)
#' @noRd
util_write_iframe_results <- function(
    pages,
    progress_msg,
    progress,
    block_load_factor = getOption(
      "dataquieR.print_block_load_factor",
      dataquieR.print_block_load_factor_default),
    template_file,
    dir
) {

  util_expect_scalar(
    block_load_factor,
    check_type         = util_is_numeric_in(min = 1, whole_num = FALSE),
    convert_if_possible = as.numeric,
    error_message       = sprintf("%s should be a number above 0",
                                  sQuote("block_load_factor"))
  )

  ## --- preparation ----------------------------------------------------------
  all_tags <- unlist(lapply(pages, util_collect_tag_attrs), recursive = FALSE)
  # process each output file only once
  tasks <- Filter(function(tag) !is.null(attr(tag, "html_file")), all_tags)
  outfiles <- vapply(tasks, attr, FUN.VALUE = character(1), "html_file")
  html_inners <- lapply(tasks, attr, "html_inner")

  rp <- util_copy_all_deps(dir = dir,
                           pages = html_inners)

  tasks <- Map(function(task, html_inner) {
    html_inner[["dependencies"]] <- NULL
    attr(task, "html_inner") <- html_inner
    task
  }, tasks, rp$rendered_pages)

  tasks <- tasks[!duplicated(outfiles)]

  total <- length(tasks)
  if (total == 0L) return(invisible(NULL))

  ## --- block-wise processing -------------------------------------------------
  cl         <- parallel::getDefaultCluster()
  if (!util_check_shared_filesystem(cl = cl)$shared) {
    util_message(
      c("File system not shared amoungst workers (e.g., through NFS),",
        "falling back to serial writing"))
    cl <- NULL
  }
  chunk_size <- max(1, ceiling(block_load_factor * length(cl)))
  blocks     <- split(tasks, ceiling(seq_along(tasks) / chunk_size))
  done       <- 0L

  for (chunk in blocks) {
    if (!is.null(cl)) {
      parallel::parLapplyLB(cl, chunk, util_write_one,
                            template_file = template_file, deps = rp$deps)
    } else {
      lapply(chunk, util_write_one, template_file = template_file,
             deps = rp$deps)
    }
    done <- done + length(chunk)
    progress(100 * done / total)
    progress_msg("Writing HTML/thumbnails",
                 sprintf("Progress: %d of %d", done, total))
  }

  invisible(NULL)
}

#' For a shiny tag, write all attached "FIG"-files
#'
#' @param tag the shiny tag
#' @param template_file the report template file to use
#' @param deps html-dependencies as html-tags
#'
#' @returns `invisible(NULL)`
#' @noRd
util_write_one <- function(tag, template_file, deps) {
  outfile        <- attr(tag, "html_file")
  thumb          <- attr(tag, "thumbnail_path")
  ggthumb        <- util_decompress(attr(tag, "ggthumb"))
  html_inner     <- attr(tag, "html_inner")
  thumbnail_args <- attr(tag, "thumbnail_args")

  # HTML
  logo <- "logo.png"
  withCallingHandlers({
    html_report <- htmltools::htmlTemplate(template_file,
                                           document_ = TRUE,
                                           logo = logo,
                                           deps = deps,
                                           content = html_inner,
                                           title = basename(outfile))
  },
  warning = function(cond) { # suppress a waning caused by ggplotly for barplots
    if (startsWith(conditionMessage(cond),
                   "'bar' objects don't have these attributes: 'mode'") ||
        startsWith(conditionMessage(cond),
                   "'box' objects don't have these attributes: 'mode'")) {
      invokeRestart("muffleWarning")
    }
  })


  f <- file(outfile, open = "w", encoding = "utf-8")
  on.exit(close(f))
  withCallingHandlers(
    cat(as.character(html_report), file = f),
    warning = function(w) {
      if (startsWith(conditionMessage(w),
                     "'bar' objects don't have these attributes: 'mode'") ||
          startsWith(conditionMessage(w),
                     "'box' objects don't have these attributes: 'mode'"))
        invokeRestart("muffleWarning")
    })

  # thumbnail (optional)
  figure_type_id <- thumbnail_args$figure_type_id

  if (length(figure_type_id) != 1 ||
      !is.character(figure_type_id)) {
    # FIXME: How can that be?!
    figure_type_id <- "dot_mat"
  }

  if (isTRUE(thumbnail_args$rotated)) {
    figure_type_id <-
      paste0(figure_type_id, "_rotated")
  }

  if (!is.null(ggthumb) && !file.exists(thumb) &&
      !is.null(thumbnail_args)) {
    suppressWarnings(suppressMessages(
      util_save_with_inner_size(
        plot           = ggthumb,
        filename       = thumb,
        inner_width    = thumbnail_args$width,
        inner_height   = thumbnail_args$height,
        dpi            = thumbnail_args$dpi,
        base_text_factor = switch(
          figure_type_id,
          dot_mat = 11,
          40)
      )))
  }
  invisible(NULL)
}

#' Check whether all workers see the same file-system
#'
#' @param dir [character] directory assumed to be shared **and writable**
#'                        by master and workers. Defaults to the current
#'                        working directory (`"."`).
#' @param cl `cluster` an existing `parallel` cluster **or** `NULL`. If
#'                     `NULL`, try `parallel::getDefaultCluster()` and
#'                     assume sequential mode if that fails.
#' @param cleanup [logical] remove the test-file once the check is finished
#'                          (default `TRUE`).
#' @param tmp_prefix [character] prefix for the temporary file name
#'                               (default `"fs_test_"`).
#'
#' @return *list* with three components
#'   * `shared`  `TRUE` if, and only if **all** workers confirmed the file,
#'   * `exists`  logical vector per worker,
#'   * `tmpFile` absolute path of the test-file (removed when `cleanup`).
#'
#' @family util_functions
#' @concept process
#' @noRd
util_check_shared_filesystem <- function(dir = ".",
                                         cl = NULL,
                                         cleanup = TRUE,
                                         tmp_prefix = "fs_test_") {

  ## ensure that package 'parallel' is available -----------------------------
  util_ensure_suggested("parallel")

  ## create a temporary test file on the master node -------------------------
  dir <- normalizePath(dir, mustWork = TRUE)
  testfile <- tempfile(tmp_prefix, tmpdir = dir)

  if (!file.create(testfile)) {
    util_error("Cannot create file %s", dQuote(testfile))
  }

  if (cleanup) {
    on.exit(unlink(testfile, force = TRUE), add = TRUE)
  }

  ## retrieve a running cluster, if not passed in ----------------------------
  if (is.null(cl)) {
    cl <- tryCatch(parallel::getDefaultCluster(),
                   error = function(e) NULL)
  }

  ## if no cluster is available, assume sequential execution -----------------
  if (is.null(cl)) {
    return(list(
      shared  = TRUE,
      exists  = TRUE,
      tmpFile = testfile
    ))
  }
  ## query each worker for file existence ------------------------------------
  exists_vec <- tryCatch({
    unlist(parallel::clusterCall(cl, file.exists, testfile),
           use.names = FALSE)
  }, error = function(e) {
    util_warning("clusterCall failed, sequential fallback: %s",
                 dQuote(conditionMessage(e)))
    TRUE
  })

  names(exists_vec) <- paste0("worker", seq_along(exists_vec))

  list(
    shared  = all(exists_vec),
    exists  = exists_vec,
    tmpFile = testfile
  )
}

# writes in subfolder .report and also renderinfo.json
util_write_renderinfo_js_json <- function(output_dir,
                                          rep_id = NULL,
                                          start_time = NULL,
                                          end_time = NULL) {
  output_dir <- file.path(output_dir, ".report")

  if (!dir.exists(output_dir)) {
    if (!dir.create(output_dir)) {
      util_error("Could not create %s -- exists as file?", output_dir)
    }
  }

  if (is.null(rep_id))
    rep_id <- util_make_report_id()

  if (is.null(start_time) ||
      is.null(end_time) ||
      any(is.na(start_time)) ||
      any(is.na(end_time)) ||
      length(start_time) != 1 ||
      length(end_time) != 1) {
    t <- "null"
  } else {
    t <- paste0("\"",
               as.character(round(
                 difftime(end_time, start_time, units = "mins"), 2)),
          " min", "\"")
  }

  fn <- file.path(output_dir, "renderinfo.json")
  tmp <- paste0(fn, ".tmp")

  txt <- paste0(
    "{\"renderingTime\": ",
    t,
    ", ",
    "\"reportId\": \"",
    rep_id,
    "\"}\n"
  )

  # write to temp file first (same directory => atomic rename on POSIX)
  cat(txt, file = tmp, sep = "")

  ok <- file.rename(tmp, fn)
  if (!isTRUE(ok)) {
    # Windows can fail if target exists / is locked -> best-effort replace
    unlink(fn, force = TRUE)
    ok2 <- file.rename(tmp, fn)
    if (!isTRUE(ok2)) {
      # last resort: copy then remove temp
      file.copy(tmp, fn, overwrite = TRUE)
      unlink(tmp, force = TRUE)
    }
  }

  fn <- file.path(output_dir, "renderinfo.js")
  tmp <- paste0(fn, ".tmp")

  js_txt <- paste0("window.renderingData = ", txt, ";")

  # write to temp file first (same directory => atomic rename on POSIX)
  cat(js_txt, file = tmp, sep = "")

  ok <- file.rename(tmp, fn)
  if (!isTRUE(ok)) {
    # Windows can fail if target exists / is locked -> best-effort replace
    unlink(fn, force = TRUE)
    ok2 <- file.rename(tmp, fn)
    if (!isTRUE(ok2)) {
      # last resort: copy then remove temp
      file.copy(tmp, fn, overwrite = TRUE)
      unlink(tmp, force = TRUE)
    }
  }

  invisible(rep_id)
}

util_write_index_html <- function(path, lines, tmp_name = ".index.html") {
  util_stop_if_not(length(path) == 1L)
  util_stop_if_not(is.character(lines))

  dir <- dirname(path)
  tmp <- file.path(dir, tmp_name)

  con <- file(tmp, open = "wb")
  on.exit({
    if (exists("con", inherits = FALSE) && inherits(try(close(con), silent = TRUE), "try-error")) {
      NULL
    }
  }, add = TRUE)

  writeLines(lines, con = con, sep = "\n", useBytes = TRUE)
  flush(con)
  close(con)
  rm(con)

  if (.Platform$OS.type != "windows") {
    ok <- file.rename(tmp, path)
    if (!ok) {
      if (file.exists(tmp)) {
        unlink(tmp, force = TRUE)
      }
      util_error("Failed to atomically replace: %s -> %s",
                 dQuote(tmp), dQuote(path))
    }
    return(invisible(path))
  }

  # Windows fallback: keep 'path' always present (not atomic, but no gap)
  ok <- file.copy(tmp, path, overwrite = TRUE)
  if (!ok) {
    if (file.exists(tmp)) {
      unlink(tmp, force = TRUE)
    }
    util_error("Failed to copy temp index into place: %s -> %s",
               dQuote(tmp), dQuote(path))
  }

  if (file.exists(tmp)) {
    for (i in seq_len(10L)) {
      if (unlink(tmp, force = TRUE) == 0L || !file.exists(tmp)) {
        break
      }
      Sys.sleep(0.05)
    }
  }

  invisible(path)
}

.outer_by_env <- new.env(parent = emptyenv())

util_index_loading_lines <- function(title,
                                     message = "Preparing report...",
                                     detail = NULL,
                                     reload_ms = 1200L,
                                     n = Inf,
                                     percent = 0,
                                     logo_rel = "logo.png") {

  title_escaped <- htmltools::htmlEscape(title)

  detail_html <- ""
  if (!!length(detail) && nzchar(detail)) {
    detail_html <- paste0(
      "<p class=\"detail\">",
      htmltools::htmlEscape(detail),
      "</p>"
    )
  }

  reload_s <- max(1, as.integer(round(as.integer(reload_ms) / 1000)))

  show_progress <- is.numeric(percent) && length(percent) == 1 &&
    is.finite(percent) && percent >= 0 && percent <= 100

  indicator_html <- if (show_progress) {
    paste0(
      "<div class=\"progress\" aria-hidden=\"true\">",
      "<div class=\"progress-bar\" style=\"width:", round(percent), "%\"></div>",
      "</div>",
      "<div class=\"progress-label\">", round(percent), "%</div>"
    )
  } else {
    "<div class=\"spinner\" aria-hidden=\"true\"></div>"
  }

  outer_html <- ""
  if (is.list(.outer_by_env$outer_by) &&
      length(.outer_by_env$outer_by) &&
      !is.null(.outer_by_env$outer_by$i) &&
      !is.null(.outer_by_env$outer_by$n)) {

    outer_i <- suppressWarnings(as.integer(.outer_by_env$outer_by$i))
    outer_n <- suppressWarnings(as.integer(.outer_by_env$outer_by$n))
    outer_msg <- .outer_by_env$outer_by$msg

    valid_outer <- length(outer_i) == 1 &&
      length(outer_n) == 1 &&
      !is.na(outer_i) &&
      !is.na(outer_n) &&
      outer_n > 0 &&
      outer_i >= 1 &&
      outer_i <= outer_n

    if (valid_outer) {
      outer_percent <- ((outer_i - 1) / outer_n) * 100
      outer_percent <- max(0, min(100, outer_percent))

      outer_msg_html <- ""
      if (!!length(outer_msg) && nzchar(outer_msg)) {
        outer_msg_html <- paste0(
          "<div class=\"outer-msg\">",
          htmltools::htmlEscape(outer_msg),
          "</div>"
        )
      }

      outer_html <- paste0(
        "<div class=\"outer-block\">",
        outer_msg_html,
        "<div class=\"progress outer-progress\" aria-hidden=\"true\">",
        "<div class=\"progress-bar\" style=\"width:", round(outer_percent), "%\"></div>",
        "</div>",
        "<div class=\"progress-label\">",
        "Overall: ", round(outer_percent), "% &mdash; step ",
        outer_i, " / ", outer_n,
        "</div>",
        "</div>"
      )
    } else {
      outer_parts <- character(0)

      if (!!length(outer_msg) && nzchar(outer_msg)) {
        outer_parts <- c(
          outer_parts,
          paste0(
            "<div class=\"outer-msg\">",
            htmltools::htmlEscape(outer_msg),
            "</div>"
          )
        )
      }

      if (length(outer_i) == 1 && !is.na(outer_i)) {
        if (length(outer_n) == 1 && !is.na(outer_n) && outer_n > 0) {
          outer_parts <- c(
            outer_parts,
            paste0(
              "<div class=\"progress-label\">",
              "Overall: step ", outer_i, " / ", outer_n,
              "</div>"
            )
          )
        } else {
          outer_parts <- c(
            outer_parts,
            paste0(
              "<div class=\"progress-label\">",
              "Overall: step ", outer_i,
              "</div>"
            )
          )
        }
      } else if (length(outer_n) == 1 && !is.na(outer_n) && outer_n > 0) {
        outer_parts <- c(
          outer_parts,
          paste0(
            "<div class=\"progress-label\">",
            "Overall: ", outer_n, " steps",
            "</div>"
          )
        )
      }

      outer_parts <- c(
        outer_parts,
        "<div class=\"spinner outer-spinner\" aria-hidden=\"true\"></div>"
      )

      outer_html <- paste0(
        "<div class=\"outer-block\">",
        paste(outer_parts, collapse = ""),
        "</div>"
      )
    }
  }

  # --- favicon (logo + small badge) ---
  icon_lines <- character(0)
  if (!is.null(logo_rel) && nzchar(logo_rel)) {
    badge_text <- if (show_progress) paste0(round(percent), "%") else "..."
    # Minimal SVG favicon with embedded logo + badge (bottom-right)
    svg <- paste0(
      "<svg xmlns='http://www.w3.org/2000/svg' width='64' height='64' viewBox='0 0 64 64'>",
      "<image href='", htmltools::htmlEscape(logo_rel), "' x='0' y='0' width='64' height='64'/>",
      "<circle cx='50' cy='50' r='13' fill='white' opacity='0.92'/>",
      "<circle cx='50' cy='50' r='12' fill='rgba(0,0,0,0.55)'/>",
      "<text x='50' y='54' text-anchor='middle' font-family='system-ui,-apple-system,sans-serif' ",
      "font-size='14' fill='white'>", htmltools::htmlEscape(badge_text), "</text>",
      "</svg>"
    )
    svg_enc <- utils::URLencode(svg, reserved = TRUE)
    icon_lines <- c(
      paste0("<link rel=\"icon\" type=\"image/svg+xml\" href=\"data:image/svg+xml,", svg_enc, "\">"),
      paste0("<link rel=\"apple-touch-icon\" href=\"", htmltools::htmlEscape(logo_rel), "\">")
    )
  }

  c(
    "<!DOCTYPE html>",
    "<html lang=\"en\">",
    "<head>",
    "<meta charset=\"UTF-8\">",
    "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">",
    "<meta http-equiv=\"Cache-Control\" content=\"no-store, no-cache, must-revalidate\">",
    "<meta http-equiv=\"Pragma\" content=\"no-cache\">",
    "<meta http-equiv=\"Expires\" content=\"0\">",
    icon_lines,
    # paste0("<meta http-equiv=\"refresh\" content=\"", reload_s, "\">"),
    "<script>",
    "(function(){",
    "  'use strict';",
    "  var ms = ", as.integer(reload_ms), ";",
    "  setTimeout(function(){",
    "    try { location.reload(); } catch(e) {}",
    "  }, Math.max(250, ms));",
    "})();",
    "</script>",
    "<title>...", title_escaped, "... Progressing</title>",
    "<style>",
    "body{font-family:system-ui,-apple-system,sans-serif;display:flex;justify-content:center;align-items:center;height:100vh;margin:0;background:#f8f9fa;color:#333;}",
    ".box{text-align:center;max-width:42rem;width:100%;padding:1.25rem;}",
    ".spinner{width:36px;height:36px;border:4px solid rgba(0,0,0,.15);border-top-color:rgba(0,0,0,.55);border-radius:50%;animation:spin 1s linear infinite;margin:0 auto .9rem;}",
    "@keyframes spin{to{transform:rotate(360deg);}}",
    ".progress{width:220px;height:10px;background:rgba(0,0,0,.15);border-radius:6px;margin:0 auto .5rem;overflow:hidden;}",
    ".progress-bar{height:100%;background:rgba(0,0,0,.55);transition:width .4s ease;}",
    ".progress-label{font-size:.85rem;opacity:.75;margin-bottom:.4rem;}",
    ".detail{opacity:.75;font-size:.95rem;}",
    ".outer-block{position:fixed;top:0;left:0;width:100%;padding-top:1rem;display:flex;flex-direction:column;align-items:center;background:#f8f9fa;z-index:1000;}",
    ".outer-progress{width:80%;max-width:42rem;}",
    ".outer-msg{font-size:.95rem;font-weight:600;margin-bottom:.35rem;}",
    ".outer-spinner{width:28px;height:28px;margin:0 auto .5rem;}",
    "</style>",
    "</head>",
    "<body>",
    "<div class=\"box\">",
    outer_html,
    indicator_html,
    "<h2>", htmltools::htmlEscape(message), "</h2>",
    detail_html,
    "</div>",
    "</body>",
    "</html>"
  )
}

util_index_redirect_lines <- function(title,
                                      target_rel,
                                      delay_ms = 0L,
                                      logo_rel = "logo.png") {

  title_escaped <- htmltools::htmlEscape(title)
  delay_ms <- max(0L, as.integer(delay_ms))
  target_js <- paste0('"', (gsub("\\\\", "\\\\\\\\", gsub("\"", "\\\\\"", target_rel))), '"')

  icon_lines <- character(0)
  if (!is.null(logo_rel) && nzchar(logo_rel)) {
    icon_lines <- c(
      paste0("<link rel=\"icon\" href=\"", htmltools::htmlEscape(logo_rel), "\">"),
      paste0("<link rel=\"apple-touch-icon\" href=\"", htmltools::htmlEscape(logo_rel), "\">")
    )
  }

  c(
    "<!DOCTYPE html>",
    "<html lang=\"en\">",
    "<!-- done -->",
    "<head>",
    "<meta charset=\"UTF-8\">",
    "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">",
    "<meta http-equiv=\"Cache-Control\" content=\"no-store, no-cache, must-revalidate\">",
    "<meta http-equiv=\"Pragma\" content=\"no-cache\">",
    "<meta http-equiv=\"Expires\" content=\"0\">",
    icon_lines,
    "<script>",
    "(function(){",
    "  'use strict';",
    paste0("  var target = ", target_js, ";"),
    "  var ms = ", delay_ms, ";",
    "  function go(){",
    "    try { location.replace(target); } catch(e) { try { location.href = target; } catch(e2) {} }",
    "  }",
    "  if (ms <= 0) go(); else setTimeout(go, ms);",
    "})();",
    "</script>",
    "<title>", title_escaped, "</title>",
    "</head>",
    "<body></body>",
    "</html>"
  )
}

util_index_error_lines <- function(title,
                                   message = "An error occurred",
                                   detail = NULL,
                                   logo_rel = "logo.png",
                                   reload_ms = 5000L) {

  title_escaped <- htmltools::htmlEscape(title)

  detail_html <- ""
  if (!is.null(detail) && nzchar(detail)) {
    detail_html <- paste0(
      "<pre class=\"detail\">",
      htmltools::htmlEscape(detail),
      "</pre>"
    )
  }

  # --- favicon (logo + red badge) ---
  icon_lines <- character(0)
  if (!is.null(logo_rel) && nzchar(logo_rel)) {
    svg <- paste(
      "<svg xmlns='http://www.w3.org/2000/svg' width='64' height='64' viewBox='0 0 64 64'>",
      "<image href='", htmltools::htmlEscape(logo_rel),
      "' x='0' y='0' width='64' height='64'/>",
      "<circle cx='50' cy='50' r='13' fill='white' opacity='0.95'/>",
      "<circle cx='50' cy='50' r='12' fill='#b00020'/>",
      "<text x='50' y='55' text-anchor='middle' font-family='system-ui,-apple-system,sans-serif' ",
      "font-size='20' fill='white'>!</text>",
      "</svg>",
      sep = "\n",
      collapse = "\n"
    )
    svg_enc <- utils::URLencode(svg, reserved = TRUE)
    icon_lines <- c(
      paste0("<link rel=\"icon\" type=\"image/svg+xml\" href=\"data:image/svg+xml,", svg_enc, "\">"),
      paste0("<link rel=\"apple-touch-icon\" href=\"", htmltools::htmlEscape(logo_rel), "\">")
    )
  }

  reload_ms <- as.integer(reload_ms)
  if (!is.finite(reload_ms) || reload_ms < 0L) reload_ms <- 5000L

  # --- JS: poll renderinfo.js; if it appears (or changes), soft-reload the page ---
  refresh_js <- paste(
    "(function(){",
    "  var delay=", reload_ms, ";",
    "  var baseUrl = (function(){",
    "    var u = window.location.href;",
    "    var i = u.indexOf('#'); if (i>=0) u = u.slice(0,i);",
    "    return u;",
    "  })();",
    "  function addBuster(u){",
    "    u = u.replace(/([?&])__dq_refresh=\\d+(&)?/g, function(m, sep, tail){",
    "      return sep === '?' ? (tail ? '?' : '') : (tail ? sep : '');",
    "    });",
    "    u = u.replace(/[?&]$/, '');",
    "    return u + (u.indexOf('?')>=0?'&':'?') + '__dq_refresh=' + Date.now();",
    "  }",
    "  function softReload(){",
    "    try { window.location.replace(addBuster(baseUrl)); } catch(e) {}",
    "  }",
    "",
    "  // Compute sibling URL: same dir as current document, file name '.report/renderinfo.js'",
    "  function renderinfoUrl(){",
    "    try {",
    "      var u = new URL(baseUrl);",
    "      // Replace last path segment with .report/renderinfo.js",
    "      u.pathname = u.pathname.replace(/\\/[^\\/]*$/, '/.report/renderinfo.js');",
    "      return u.toString();",
    "    } catch(e) {",
    "      // Fallback for older URL handling: manual",
    "      return baseUrl.replace(/\\/[^\\/]*$/, '/.report/renderinfo.js');",
    "    }",
    "  }",
    "",
    "  var lastToken = null;",
    "  var inFlight = false;",
    "  function check(){",
    "    if (inFlight) return;",
    "    inFlight = true;",
    "",
    "    // IMPORTANT: renderinfo.js should define window.renderingData = {...}",
    "    // When it does not exist -> that indicates an error; do NOT reload blindly.",
    "    var s = document.createElement('script');",
    "    s.async = true;",
    "    s.src = addBuster(renderinfoUrl());",
    "",
    "    s.onload = function(){",
    "      try {",
    "        var d = window.renderingData || null;",
    "        // If renderinfo.js loaded but didn't set data, treat as 'not ready' and keep polling.",
    "        if (!d) return;",
    "        // Build a token that changes when the report changes.",
    "        var token = String(d.reportId || '') + '|' + String(d.renderingTime || '');",
    "        if (lastToken === null) {",
    "          lastToken = token;",
    "          // renderinfo.js now exists => reload once to leave error page",
    "          softReload();",
    "          return;",
    "        }",
    "        if (token !== lastToken) {",
    "          lastToken = token;",
    "          softReload();",
    "          return;",
    "        }",
    "      } finally {",
    "        // cleanup",
    "        try { s.remove(); } catch(e) {}",
    "      }",
    "    };",
    "",
    "    s.onerror = function(){",
    "      // renderinfo.js missing -> error state; keep polling but do not reload",
    "      try { s.remove(); } catch(e) {}",
    "    };",
    "",
    "    // Append and always release the inFlight flag at end of this tick",
    "    (document.head || document.documentElement).appendChild(s);",
    "    window.setTimeout(function(){ inFlight = false; }, 0);",
    "  }",
    "",
    "  check();",
    "  window.setInterval(check, Math.max(1000, delay));",
    "})();",
    sep = "\n",
    collapse = "\n"
  )

  c(
    "<!DOCTYPE html>",
    "<html lang=\"en\">",
    "<head>",
    "<meta charset=\"UTF-8\">",
    "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">",
    "<meta http-equiv=\"Cache-Control\" content=\"no-store, no-cache, must-revalidate\">",
    "<meta http-equiv=\"Pragma\" content=\"no-cache\">",
    "<meta http-equiv=\"Expires\" content=\"0\">",
    icon_lines,
    "<title>\u2716 ", title_escaped, "</title>",
    "<style>",
    "body{font-family:system-ui,-apple-system,sans-serif;display:flex;justify-content:center;",
    "align-items:center;height:100vh;margin:0;background:#fff5f6;color:#2a0006;}",
    ".box{text-align:center;max-width:48rem;padding:1.5rem;}",
    ".icon{font-size:3rem;color:#b00020;margin-bottom:.6rem;}",
    "h2{color:#b00020;margin:.3rem 0 .6rem 0;}",
    ".detail{margin-top:.8rem;padding:.8rem;background:#fff;",
    "border:1px solid #f1b0b7;border-radius:6px;",
    "text-align:left;white-space:pre-wrap;max-height:40vh;overflow:auto;}",
    "</style>",
    "<script>", refresh_js, "</script>",
    "</head>",
    "<body>",
    "<div class=\"box\">",
    "<div class=\"icon\">\u26A0</div>",
    "<h2>", htmltools::htmlEscape(message), "</h2>",
    detail_html,
    "</div>",
    "</body>",
    "</html>"
  )
}

util_gg_stagger_x_axis_labels <- function(
    p,
    inner_width = NULL,
    inner_height = NULL,
    angle = NULL,
    auto_dodge = TRUE,
    max_dodge = 3L,
    min_labels_for_change = 28L,
    min_size = 4.2,
    title_scale = 0.90,
    subtitle_scale = 0.82,
    y_text_scale = 0.92,
    patchwork_widths = NULL,
    patchwork_heights = NULL,
    theme_base_size = NULL,
    verbose = FALSE) {

  ok <-
    (is.null(inner_width) || (length(inner_width) == 1L && is.numeric(inner_width) &&
                                is.finite(inner_width) && inner_width > 0)) &&
    (is.null(inner_height) || (length(inner_height) == 1L && is.numeric(inner_height) &&
                                 is.finite(inner_height) && inner_height > 0)) &&
    length(max_dodge) == 1L &&
    is.numeric(max_dodge) &&
    max_dodge >= 1 &&
    length(min_labels_for_change) == 1L &&
    is.numeric(min_labels_for_change) &&
    min_labels_for_change >= 1 &&
    length(min_size) == 1L &&
    is.numeric(min_size) &&
    min_size > 0 &&
    length(auto_dodge) == 1L &&
    is.logical(auto_dodge) &&
    length(verbose) == 1L &&
    is.logical(verbose) &&
    length(title_scale) == 1L &&
    is.numeric(title_scale) &&
    is.finite(title_scale) &&
    title_scale > 0 &&
    length(subtitle_scale) == 1L &&
    is.numeric(subtitle_scale) &&
    is.finite(subtitle_scale) &&
    subtitle_scale > 0 &&
    length(y_text_scale) == 1L &&
    is.numeric(y_text_scale) &&
    is.finite(y_text_scale) &&
    y_text_scale > 0 &&
    (is.null(angle) || (length(angle) == 1L && is.numeric(angle))) &&
    (is.null(patchwork_widths) ||
       (is.numeric(patchwork_widths) && length(patchwork_widths) >= 1L &&
          all(is.finite(patchwork_widths)) && all(patchwork_widths > 0))) &&
    (is.null(patchwork_heights) ||
       (is.numeric(patchwork_heights) && length(patchwork_heights) >= 1L &&
          all(is.finite(patchwork_heights)) && all(patchwork_heights > 0))) &&
    (is.null(theme_base_size) ||
       (length(theme_base_size) == 1L && is.numeric(theme_base_size) &&
          is.finite(theme_base_size) && theme_base_size > 0))

  if (!ok) {
    return(p)
  }

  `%||%` <- function(x, y) if (is.null(x)) y else x

  is_patchwork_obj <- function(x) {
    inherits(x, "patchwork") ||
      (!is.null(x$patches) &&
         is.list(x$patches) &&
         !is.null(x$patches$plots) &&
         is.list(x$patches$plots))
  }

  normalize_weights <- function(w, n, default = 1) {
    if (is.null(w) || !length(w)) {
      w <- rep(default, n)
    } else {
      if (length(w) < n) {
        w <- c(w, rep(tail(w, 1L), n - length(w)))
      } else if (length(w) > n) {
        w <- w[seq_len(n)]
      }
    }
    s <- sum(w)
    if (!is.finite(s) || s <= 0) {
      return(rep(1 / n, n))
    }
    w / s
  }

  get_theme_obj <- function(g) {
    th <- try(ggplot2::theme_get() + g$theme, silent = TRUE)
    if (util_is_try_error(th) || is.null(th)) {
      return(NULL)
    }
    th
  }

  get_local_theme_size <- function(g) {
    th_local <- g$theme
    if (is.null(th_local)) {
      return(NULL)
    }

    sz <- th_local$axis.text.x$size %||%
      th_local$axis.text$size %||%
      th_local$text$size

    if (is.null(sz) || !is.numeric(sz) || length(sz) != 1L || !is.finite(sz)) {
      return(NULL)
    }

    sz
  }

  get_theme_text_size <- function(g) {
    th <- get_theme_obj(g)
    if (is.null(th)) {
      return(11)
    }

    sz <- th$axis.text.x$size %||%
      th$axis.text$size %||%
      th$text$size %||%
      11

    if (!is.numeric(sz) || length(sz) != 1L || !is.finite(sz)) {
      return(11)
    }

    sz
  }

  get_theme_angle <- function(g) {
    th <- get_theme_obj(g)
    if (is.null(th)) {
      return(NULL)
    }

    ang <- th$axis.text.x$angle
    if (is.null(ang) || !is.numeric(ang) || length(ang) != 1L || !is.finite(ang)) {
      return(NULL)
    }

    ang
  }

  get_x_labels_info <- function(g) {
    b <- try(ggplot2::ggplot_build(g), silent = TRUE)
    if (util_is_try_error(b)) {
      return(NULL)
    }

    sc <- try(b$plot$scales$get_scales("x"), silent = TRUE)
    if (util_is_try_error(sc) || is.null(sc)) {
      return(NULL)
    }

    is_discrete_x <- inherits(sc, "ScaleDiscrete") ||
      inherits(sc, "ScaleDiscretePosition")

    if (!isTRUE(is_discrete_x)) {
      return(NULL)
    }

    pp <- try(b$layout$panel_params[[1L]], silent = TRUE)
    if (util_is_try_error(pp) || is.null(pp)) {
      return(NULL)
    }

    labels <- try(pp$x$get_labels(), silent = TRUE)
    if (util_is_try_error(labels) || is.null(labels)) {
      return(NULL)
    }

    labels <- as.character(labels)
    labels <- labels[nzchar(labels)]

    if (!length(labels)) {
      return(NULL)
    }

    list(
      labels = labels,
      n = length(labels),
      mean_nchar = mean(nchar(labels), na.rm = TRUE),
      max_nchar = max(nchar(labels), na.rm = TRUE)
    )
  }

  choose_target_size <- function(base_size, info, inner_width = NULL) {
    n_lab <- info$n
    mean_nchar <- info$mean_nchar
    max_nchar <- info$max_nchar

    target <- base_size

    if (n_lab <= 12L) {
      return(target)
    }

    if (n_lab <= 20L) {
      if (mean_nchar >= 10) {
        target <- min(target, base_size * 0.96)
      }
      return(max(min_size, target))
    }

    if (n_lab < min_labels_for_change) {
      if (!is.null(inner_width) && is.finite(inner_width) && inner_width > 0) {
        labels_per_in <- n_lab / inner_width
        if (labels_per_in >= 7) {
          target <- min(target, base_size * 0.92)
        }
      }
      return(max(min_size, target))
    }

    if (n_lab >= 28)  target <- min(target, 7.2)
    if (n_lab >= 40)  target <- min(target, 6.6)
    if (n_lab >= 60)  target <- min(target, 5.9)
    if (n_lab >= 90)  target <- min(target, 5.2)
    if (n_lab >= 130) target <- min(target, 4.7)
    if (n_lab >= 180) target <- min(target, 4.2)

    if (mean_nchar >= 8)  target <- target - 0.15
    if (mean_nchar >= 12) target <- target - 0.25
    if (max_nchar >= 20)  target <- target - 0.15

    if (!is.null(inner_width) && is.finite(inner_width) && inner_width > 0) {
      labels_per_in <- n_lab / inner_width

      if (labels_per_in >= 8)  target <- min(target, 6.3)
      if (labels_per_in >= 11) target <- min(target, 5.6)
      if (labels_per_in >= 14) target <- min(target, 5.0)
      if (labels_per_in >= 18) target <- min(target, 4.5)
    }

    max(min_size, target)
  }

  choose_n_dodge <- function(info, angle_eff, inner_width = NULL) {
    if (!isTRUE(auto_dodge)) {
      return(1L)
    }

    n_lab <- info$n
    mean_nchar <- info$mean_nchar

    if (n_lab <= 35L) {
      return(1L)
    }

    if (!is.null(angle_eff) && abs(angle_eff) >= 80) {
      return(1L)
    }

    n_dodge <- 1L

    if (n_lab >= 70 || mean_nchar >= 16) {
      n_dodge <- 2L
    }
    if (n_lab >= 140 || mean_nchar >= 24) {
      n_dodge <- 3L
    }

    if (!is.null(inner_width) && is.finite(inner_width) && inner_width > 0) {
      labels_per_in <- n_lab / inner_width
      if (labels_per_in >= 14) n_dodge <- max(n_dodge, 2L)
      if (labels_per_in >= 22) n_dodge <- max(n_dodge, 3L)
    }

    min(as.integer(max_dodge), n_dodge)
  }

  scale_patchwork_annotation_theme <- function(x, base_size_local) {
    if (!is_patchwork_obj(x)) {
      return(x)
    }

    ann_theme <- NULL
    if (!is.null(x$patches$annotation) &&
        !is.null(x$patches$annotation$theme)) {
      ann_theme <- x$patches$annotation$theme
    }

    if (is.null(ann_theme)) {
      ann_theme <- ggplot2::theme()
    }

    x$patches$annotation$theme <- ann_theme + ggplot2::theme(
      plot.title = ggplot2::element_text(
        size = max(min_size, base_size_local * 0.66)
      ),
      plot.subtitle = ggplot2::element_text(
        size = max(min_size, base_size_local * 0.58)
      ),
      plot.caption = ggplot2::element_text(
        size = max(min_size, base_size_local * 0.54)
      )
    )

    x
  }

  apply_one <- function(g, inner_width_local = inner_width, inner_height_local = inner_height) {
    if (!inherits(g, "ggplot") || is_patchwork_obj(g)) {
      return(g)
    }

    info <- get_x_labels_info(g)
    if (is.null(info)) {
      return(g)
    }

    base_size <- theme_base_size %||% get_local_theme_size(g) %||% get_theme_text_size(g)
    angle_eff <- angle %||% get_theme_angle(g)

    if (info$n <= 12L) {
      return(g)
    }

    n_dodge <- choose_n_dodge(info, angle_eff, inner_width = inner_width_local)
    target_size <- choose_target_size(base_size, info, inner_width = inner_width_local)

    if (isTRUE(verbose)) {
      msg <- sprintf(
        paste0(
          "util_gg_stagger_x_axis_labels(): n_labels=%d, mean_nchar=%.1f, ",
          "inner_width=%s, size %.1f -> %.1f, n.dodge=%d"
        ),
        info$n,
        info$mean_nchar,
        if (is.null(inner_width_local)) "NULL" else format(inner_width_local),
        base_size,
        target_size,
        n_dodge
      )
      message(msg)
    }

    th <- get_theme_obj(g)

    axis_text_x_old <- NULL
    axis_text_y_old <- NULL
    plot_title_old <- NULL
    plot_subtitle_old <- NULL
    text_old <- NULL

    if (!is.null(th)) {
      axis_text_x_old <- th$axis.text.x
      axis_text_y_old <- th$axis.text.y
      plot_title_old <- th$plot.title
      plot_subtitle_old <- th$plot.subtitle
      text_old <- th$text
    }

    inherit_el <- function(el, default = ggplot2::element_text()) {
      if (is.null(el)) default else el
    }

    axis_text_x_old <- inherit_el(axis_text_x_old)
    axis_text_y_old <- inherit_el(axis_text_y_old)
    plot_title_old <- inherit_el(plot_title_old)
    plot_subtitle_old <- inherit_el(plot_subtitle_old)
    text_old <- inherit_el(text_old)

    scaled_base_size <- base_size

    out <- g + ggplot2::theme(
      text = ggplot2::element_text(
        family = text_old$family,
        face = text_old$face,
        colour = text_old$colour,
        size = scaled_base_size,
        hjust = text_old$hjust,
        vjust = text_old$vjust,
        angle = text_old$angle,
        lineheight = text_old$lineheight,
        margin = text_old$margin,
        debug = text_old$debug
      ),
      axis.text.x = ggplot2::element_text(
        family = axis_text_x_old$family,
        face = axis_text_x_old$face,
        colour = axis_text_x_old$colour,
        size = target_size,
        hjust = axis_text_x_old$hjust %||% 1,
        vjust = axis_text_x_old$vjust %||% 0.5,
        angle = angle_eff %||% axis_text_x_old$angle,
        lineheight = axis_text_x_old$lineheight,
        margin = axis_text_x_old$margin,
        debug = axis_text_x_old$debug
      ),
      axis.text.y = ggplot2::element_text(
        family = axis_text_y_old$family,
        face = axis_text_y_old$face,
        colour = axis_text_y_old$colour,
        size = max(min_size, min(scaled_base_size * y_text_scale,
                                 max(min_size, target_size * y_text_scale))),
        hjust = axis_text_y_old$hjust,
        vjust = axis_text_y_old$vjust,
        angle = axis_text_y_old$angle,
        lineheight = axis_text_y_old$lineheight,
        margin = axis_text_y_old$margin,
        debug = axis_text_y_old$debug
      ),
      plot.title = ggplot2::element_text(
        family = plot_title_old$family,
        face = plot_title_old$face,
        colour = plot_title_old$colour,
        size = max(min_size, scaled_base_size * title_scale),
        hjust = plot_title_old$hjust,
        vjust = plot_title_old$vjust,
        angle = plot_title_old$angle,
        lineheight = plot_title_old$lineheight,
        margin = plot_title_old$margin,
        debug = plot_title_old$debug
      ),
      plot.subtitle = ggplot2::element_text(
        family = plot_subtitle_old$family,
        face = plot_subtitle_old$face,
        colour = plot_subtitle_old$colour,
        size = max(min_size, scaled_base_size * subtitle_scale),
        hjust = plot_subtitle_old$hjust,
        vjust = plot_subtitle_old$vjust,
        angle = plot_subtitle_old$angle,
        lineheight = plot_subtitle_old$lineheight,
        margin = plot_subtitle_old$margin,
        debug = plot_subtitle_old$debug
      )
    )

    if (n_dodge > 1L) {
      guide_args <- list(n.dodge = n_dodge)
      if (!is.null(angle_eff)) {
        guide_args$angle <- angle_eff
      }

      out <- out + ggplot2::guides(
        x = do.call(ggplot2::guide_axis, guide_args)
      )
    }

    out
  }

  recurse <- function(x,
                      inner_width_local = inner_width,
                      inner_height_local = inner_height,
                      widths_local = patchwork_widths,
                      heights_local = patchwork_heights) {

    if (inherits(x, "ggplot") && !is_patchwork_obj(x)) {
      return(apply_one(
        x,
        inner_width_local = inner_width_local,
        inner_height_local = inner_height_local
      ))
    }

    if (is_patchwork_obj(x)) {
      plots <- x$patches$plots
      if (is.null(plots) || !length(plots)) {
        return(scale_patchwork_annotation_theme(
          x,
          base_size_local = theme_base_size %||% 11
        ))
      }

      n <- length(plots)
      w_prop <- normalize_weights(widths_local, n)
      h_prop <- normalize_weights(heights_local, n)

      for (i in seq_along(plots)) {
        child_width <- inner_width_local
        child_height <- inner_height_local

        if (!is.null(inner_width_local)) {
          child_width <- inner_width_local * w_prop[i]
        }
        if (!is.null(inner_height_local)) {
          child_height <- inner_height_local * h_prop[i]
        }

        plots[[i]] <- recurse(
          plots[[i]],
          inner_width_local = child_width,
          inner_height_local = child_height,
          widths_local = NULL,
          heights_local = NULL
        )
      }

      x$patches$plots <- plots
      x <- scale_patchwork_annotation_theme(
        x,
        base_size_local = theme_base_size %||% 11
      )

      return(x)
    }

    x
  }

  out <- tryCatch(
    recurse(p),
    error = function(e) {
      util_warning(
        "util_gg_stagger_x_axis_labels() failed; returning input unchanged."
      )
      p
    }
  )

  if (is.null(out)) {
    return(p)
  }

  out
}


# IDEA: For PDF output, support https://cran.r-project.org/web/packages/xmpdf/index.html

test_that("print.list() falls back via NextMethod() for empty lists", {
  # print.default(list()) returns the object invisibly
  expect_invisible(print(list()))
})

test_that("print.list() falls back via NextMethod() during knitr", {
  withr::local_options(knitr.in.progress = TRUE)

  # If this branch is taken, the heavy code must NOT run; we guard it by mocking
  testthat::local_mocked_bindings(
    util_is_try_error = function(...) stop("should not be called"),
    util_ensure_suggested = function(...) stop("should not be called"),
    .package = "dataquieR"
  )

  expect_invisible(print(list(a = 1)))
})

test_that("print.list() renders single-entity multi-result HTML and respects view=FALSE", {
  # Two dummy "results" that share one entity
  mk_dqr <- function(cn) {
    dqr <- structure(list(), class = "dataquieR_result")
    attr(dqr, "call") <- structure(list(), entity = "AGE_GROUP_0", entity_name = "v00103")
    attr(dqr, "cn") <- cn
    dqr
  }

  x <- list(mk_dqr("A"), mk_dqr(NULL))  # include cn NULL to cover title/cn/nm fallback

  # Keep this branch self-contained and non-interactive.
  testthat::local_mocked_bindings(
    util_ensure_suggested = function(pkg, ...) {
      # Pretend suggests are available so the code continues.
      TRUE
    },
    html_dependency_dataquieR = function(...) {
      # Return a minimal dependency-like object (or NULL is fine for tagList)
      NULL
    },
    html_dependency_jspdf = function(...) NULL,
    util_alias2caption = function(cn, long = TRUE, ...) paste0("CAPTION-", cn),
    util_pretty_print = function(dqr, nm, is_single_var, use_plot_ly, dir, ...) {
      # Minimal HTML output
      htmltools::div(class = "pp", paste("nm:", nm))
    },
    .package = "dataquieR"
  )

  # Avoid launching a browser / viewer
  expect_null(print(x, view = FALSE))
})

test_that("print.list() master_result: warns for stored errors, then falls back via NextMethod()", {
  mk_master <- function(entity, with_err = FALSE) {
    y <- structure(list(), class = c("dataquieR_result", "master_result"))
    attr(y, "call") <- structure(list(), entity = entity, entity_name = "N")
    attr(y, "function_name") <- "acc_dummy"
    attr(y, "error") <- if (with_err) list(simpleWarning("stored warning from result")) else list()
    y
  }

  x <- list(
    mk_master("E1", TRUE),
    mk_master("E2", FALSE)
  )

  testthat::local_mocked_bindings(
    util_stop_if_not = function(...) invisible(TRUE),
    util_combine_res = function(...) stop("should not be called"),
    # IMPORTANT: fallback printing may call this via print.dataquieR_result()
    print.master_result = function(...) invisible(NULL),
    .package = "dataquieR"
  )

  # Reaches warning loop, then NextMethod() triggers default printing.
  expect_warning(
    print(x),
    "stored warning from result"
  )
})

test_that("util_correct_variable_use works", {
  skip_on_cran() # slow and implicitly tested by other tests
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  acc_test <- function(resp_variable, aux_variable, time_variable,
                       co_variables, group_vars, study_data, meta_data,
                       label_col) {
    prep_prepare_dataframes()
    util_correct_variable_use(resp_variable)
    util_correct_variable_use("resp_variable")
    util_correct_variable_use("aux_variable")
    util_correct_variable_use("time_variable")
    if (missing(co_variables)) {
      co_variables <- NA
      warning("No co_variables were defined")
    }
    util_correct_variable_use("co_variables", allow_na = TRUE,
                              allow_more_than_one = TRUE)
    util_correct_variable_use2(group_vars)
    co_variables <- na.omit(co_variables)
  }
  environment(acc_test) <- asNamespace("dataquieR")
  acc_test2 <- function(resp_variable, aux_variable, time_variable,
                       co_variables, group_vars, study_data, meta_data,
                       label_col) {
    prep_prepare_dataframes()
    expect_error(util_correct_variable_use(c(a, b)), regexp =
                   paste("argument arg_name must be of length 1, wrong use",
                         "of util_correct_variable_use."))
    expect_error(util_correct_variable_use(character(0)), regexp =
                   paste("argument arg_name must be of length 1, wrong use",
                         "of util_correct_variable_use."))

    expect_error(util_correct_variable_use(resp_variable, role = "invalid"),
                 regexp =
                   paste("Unknown variable-argument role: invalid for",
                         "argument resp_variable, wrong use",
                         "of util_correct_variable_use."))

    expect_error(util_correct_variable_use(invalid),
                 regexp =
                   paste("Unknown function argument invalid checked,",
                         "wrong use",
                         "of util_correct_variable_use."))
    resp_variable <- "v00000"
    label_col <- VAR_NAMES
    expect_silent(util_correct_variable_use(resp_variable))
    label_col <- "XXX"
    expect_silent(util_correct_variable_use(resp_variable))
    rm(label_col)
    expect_silent(util_correct_variable_use(resp_variable))
    delayedAssign("label_col", stop())
    expect_silent(util_correct_variable_use(resp_variable))
    label_col <- VAR_NAMES
  }
  environment(acc_test2) <- asNamespace("dataquieR")
  acc_test3 <- function(resp_variable, aux_variable, time_variable,
                        co_variables, group_vars, study_data, meta_data) {
    prep_prepare_dataframes()
    expect_error(util_correct_variable_use(c(a, b)), regexp =
                   paste("argument arg_name must be of length 1, wrong use",
                         "of util_correct_variable_use."))
    expect_error(util_correct_variable_use(character(0)), regexp =
                   paste("argument arg_name must be of length 1, wrong use",
                         "of util_correct_variable_use."))

    expect_error(util_correct_variable_use(resp_variable, role = "invalid"),
                 regexp =
                   paste("Unknown variable-argument role: invalid for",
                         "argument resp_variable, wrong use",
                         "of util_correct_variable_use."))

    expect_error(util_correct_variable_use(invalid),
                 regexp =
                   paste("Unknown function argument invalid checked,",
                         "wrong use",
                         "of util_correct_variable_use."))
    resp_variable <- "v00000"
    expect_silent(util_correct_variable_use(resp_variable))
  }
  environment(acc_test3) <- asNamespace("dataquieR")
  acc_test_type <-
    function(resp_variable, aux_variable, time_variable, co_variables,
                            group_vars, study_data, meta_data, need_type,
             label_col) {
    prep_prepare_dataframes()
    util_correct_variable_use(resp_variable, need_type = need_type,
      allow_more_than_one = TRUE)
  }
  environment(acc_test_type) <- asNamespace("dataquieR")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]
  expect_warning(
    expect_error(
      acc_test(study_data = study_data, meta_data = meta_data),
      regexp = "Argument resp_variable is NULL"
    ),
    regexp = paste(
      "Missing argument .+resp_variable.+ without default value.",
      "Setting to NULL. As a dataquieR developer"
    )
  )
  expect_error(
    acc_test("v00001"),
    regexp = "Need study data as a data frame"
  )
  expect_warning(
    expect_error(
      acc_test("v00001", study_data = study_data, meta_data = meta_data),
      regexp = paste(
        "Argument aux_variable is NULL"
      ),
      perl = TRUE
    ),
    regexp = paste(
      "Missing argument .+aux_variable.+ without default value. Setting",
      "to NULL. As a dataquieR developer,"
    ),
    perl = TRUE
  )
  expect_warning(
    expect_error(
      acc_test("v00001", "v00001", study_data = study_data,
               meta_data = meta_data),
      regexp = paste(
        "Argument time_variable is NULL"
      ),
      perl = TRUE
    ),
    regexp = paste(
      "Missing argument .+time_variable.+ without default value.",
      "Setting to NULL. As a dataquieR developer,"
    ),
    perl = TRUE
  )

  expect_error(
    acc_test("v00001", "v00001", "v00001", study_data = study_data,
     meta_data = meta_data, co_variables = 1:10),
    regexp = paste(
      "Need character variable names in argument co_variables"
    ),
    perl = TRUE
  )

  expect_warning(
    expect_error(
      acc_test("v00001", "v00001", "v00001", "v00001", study_data = study_data,
        meta_data = meta_data),
      regexp = paste(
        "Argument group_vars is NULL"
      ),
      perl = TRUE
    ),
    regexp = paste(
      "Missing argument .+group_vars.+ without default value.",
      "Setting to NULL. As a dataquieR developer,"
    ),
    perl = TRUE
  )

  expect_error(
    acc_test("v00001", "v00001", "v00001", "v00001",
      study_data = study_data, meta_data = meta_data,
      group_vars = c()
    ),
  regexp = paste(
    "Argument group_vars is NULL"
  ),
  perl = TRUE
  )

  expect_silent(
    acc_test("v00001", "v00001", "v00001", "v00001",
      study_data = study_data, meta_data = meta_data,
      group_vars = c("v00001")
    )
  )

  expect_error(
    acc_test("v00001", "v00001", "v00001", "v00001",
      study_data = study_data, meta_data = meta_data,
      group_vars = c("v00001", "v00002")
    ),
    regexp = paste(
      "Variable .+v00002.+ \\(group_vars\\) has NA observations,",
      "which is not allowed"
    ),
    perl = TRUE
  )

  expect_silent(
    acc_test("v00001", "v00001", "v00001", "v00001",
      study_data = study_data, meta_data = meta_data,
      group_vars = c("v00001", "v00000")
    )
  )

  expect_silent(
    acc_test_type(c("v00000", "v00001"),
      study_data = study_data,
      meta_data = meta_data, need_type = "integer|string"
    )
  )

  expect_error(
    acc_test_type(c("v00000", "v00001"),
      study_data = study_data,
      meta_data = meta_data, need_type = "integer|float"
    ),
    regexp = paste(
      "Argument .+resp_variable.+: Variable .+v00001.+ \\(string\\) does",
      "not have an allowed type \\(integer\\|float\\)"
    ),
    perl = TRUE
  )

  expect_error(
    acc_test_type(c("v00000", "v00001"),
      study_data = study_data,
      meta_data = meta_data, need_type = "integer|xxx"
    ),
    regexp = paste(
      "Internal error: .+resp_variable.+s .+need_type.+ contains invalid type",
      "names .+xxx.+allowed are .+integer.+,",
      ".+string.+, .+float.+, .+datetime.+.",
      "As a dataquieR developer, you should fix your call of",
      ".+acc_test_type+."
    ),
    perl = TRUE
  )

  expect_error(
    acc_test_type(c("v00000", "v00001"),
      study_data = study_data,
      meta_data = meta_data, need_type = "string"
    ),
    regexp = paste(
      "Argument .+resp_variable.+: Variable .+v00000.+ \\(integer\\) does",
      "not have an allowed type \\(string\\)"
    ),
    perl = TRUE
  )

  expect_error(
    acc_test_type(c("v00000", "v00001"),
      study_data = study_data,
      meta_data = meta_data, need_type = "!integer"
    ),
    regexp = paste(
      "Argument .+resp_variable.+: Variable .+v00000.+ \\(integer\\)",
      "has a disallowed type \\(.+integer.+\\)"
    ),
    perl = TRUE
  )

  expect_silent(
    acc_test_type(c("v00000", "v00001"),
      study_data = study_data,
      meta_data = meta_data, need_type = "!datetime"
    )
  )

  acc_test2("v00001", "v00001", "v00001", "v00001",
            study_data = study_data, meta_data = meta_data,
            group_vars = c("v00001", "v00000")
  )
  acc_test3("v00001", "v00001", "v00001", "v00001",
            study_data = study_data, meta_data = meta_data,
            group_vars = c("v00001", "v00000")
  )

  acc_test4 <- function(resp_variable, aux_variable, time_variable,
                        co_variables, group_vars, study_data, meta_data,
                        label_col) {
    prep_prepare_dataframes()
    util_correct_variable_use(resp_variable)
  }
  environment(acc_test4) <- asNamespace("dataquieR")
  delayedAssign("resp_variable", stop("Error"))

  suppressWarnings(expect_warning(expect_warning(
    expect_error(
      acc_test4(resp_variable = resp_variable,
              "v00001", "v00001", "v00001",
              study_data = study_data, meta_data = meta_data,
              group_vars = c("v00001", "v00000")),
      regexp = "Argument resp_variable is NULL"
    ),
    regexp = paste("Could not get value of argument",
                   "resp_variable for unexpected reasons.",
                   "Setting to NULL."
    ),
    perl = TRUE
  ), regexp = paste("Error")))

  resp_variable <- "v00001"

  acc_test5 <- function(resp_variable, aux_variable, time_variable,
                        co_variables, group_vars, study_data, meta_data,
                        label_col) {
    util_correct_variable_use(resp_variable)
  }
  environment(acc_test5) <- asNamespace("dataquieR")
  expect_error(
    acc_test5(resp_variable = "v00001", meta_data = meta_data),
    regexp = paste("Did not find merged study data and metadata ds1.",
                   "Wrong use of util_correct_variable_use?")
  )

  acc_test6 <- function(resp_variable, aux_variable, time_variable,
                        co_variables, group_vars, study_data, meta_data,
                        label_col, cmd, ...) {
    prep_prepare_dataframes()
    cmd()
    util_correct_variable_use(resp_variable, ...)
  }
  environment(acc_test6) <- asNamespace("dataquieR")
  expect_error(
    acc_test6(resp_variable = "v00001", study_data = study_data,
              meta_data = meta_data,
              cmd = function() { rm(meta_data, envir = parent.frame())}),
    regexp = paste("Did not find metadata.",
                   "Wrong use of util_correct_variable_use?")
  )
  expect_error(
    acc_test6(resp_variable = "v00001", study_data = study_data,
              meta_data = meta_data,
              cmd = function() { assign("meta_data", 42, parent.frame()) }),
    regexp = paste("meta_data does not provide a metadata data frame.",
                   "Wrong use of util_correct_variable_use?")
  )

  expect_error(
    acc_test6(resp_variable = "v00001", study_data = study_data,
              meta_data = meta_data,
              cmd = function() { assign("ds1", 42, parent.frame()) }),
    regexp = paste("ds1 does not provide merged study data and metadata.",
                   "Wrong use of util_correct_variable_use?")
  )

  expect_error(
    acc_test6(resp_variable = character(0), study_data = study_data,
              meta_data = meta_data,
              cmd = function() { }, allow_more_than_one = TRUE),
    regexp = paste("Need at least one element in argument resp_variable, got 0")
  )

  expect_error(
    acc_test6(resp_variable = letters, study_data = study_data,
              meta_data = meta_data,
              cmd = function() { }),
    regexp = paste("Need exactly one element in argument resp_variable, got",
                   "26: .a, b, c, d,.+x, y, z."),
    perl = TRUE
  )

  md0 <- meta_data
  md0$DATA_TYPE <- NA
  expect_warning(
    acc_test6(resp_variable = "v00001", study_data = study_data,
              meta_data = md0,
              cmd = function() { }, need_type = "string"),
    regexp = paste("predicted the.+DATA_TYPE"),
    perl = TRUE
  )

  expect_error(
    acc_test6(resp_variable = "v00000", study_data = study_data,
              meta_data = meta_data,
              cmd = function() { }, need_type = "string",
              allow_more_than_one = TRUE),
    regexp = paste("Argument .+resp_variable.+: Variable .+v00000.+ .integer.",
                   "does not have an allowed type .string."),
    perl = TRUE
  )

})

test_that("util_warning works", {
  skip_on_cran()
  
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  expect_warning(util_warning("The one and everything is %d (%s).", 42,
                              "Douglas Adams"),
                 regexp = ".*The one and everything is 42 .Douglas Adams.*",
                 perl = TRUE)

  testenv <- new.env(parent = baseenv())

  testenv$g <- function(...) {
    util_warning(...)
  }
  environment(testenv$g) <- asNamespace("dataquieR")

  f <- function(...) {
    g(...)
  }
  environment(f) <- testenv

  h <- function(...) {
    do.call(g, list(...))
  }
  environment(h) <- testenv

  expect_warning(f("The one and everything is %d (%s).", 42,
                    "Douglas Adams"),
                 regexp =
                   paste(
                     ".*The one and everything is",
                     "42 \\(Douglas Adams\\).*"
                   ),
                 perl = TRUE)
  expect_warning(h("The one and everything is %d (%s).", 42,
                   "Douglas Adams"),
                 regexp =
                   paste(
                     ".*The one and everything is",
                     "42 \\(Douglas Adams\\).*"
                   ),
                 perl = TRUE)
})
test_that("util_warning works", {
  skip_on_cran()
  
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  expect_warning(util_warning("The one and everything is %d (%s).", 42,
                              "Douglas Adams"),
                 regexp = "The one and everything is 42 (Douglas Adams).",
                 fixed = TRUE)

  g <- function(...) {
    util_warning(...)
  }
  environment(g) <- asNamespace("dataquieR")

  testenv <- new.env(parent = baseenv())
  testenv$gg <- force(g)

  f <- function(...) {
    gg(...)
  }
  environment(f) <- testenv

  h <- function(...) {
    do.call(gg, list(...))
  }
  environment(h) <- testenv

  expect_warning(f("The one and everything is %d (%s).", 42,
          "Douglas Adams"),
          regexp =
            paste(
              ".*The one and everything is",
              "42 \\(Douglas Adams\\).*"
            ),
          perl = TRUE)
  expect_warning(h("The one and everything is %d (%s).", 42,
         "Douglas Adams"),
         regexp =
           paste(
             ".*The one and everything is",
             "42 \\(Douglas Adams\\).*"
           ),
         ,
       perl = TRUE)

  x <- function(m) {
    warning(m)
  }
  w <- function(w) {
    util_warning(w)
    invokeRestart("muffleWarning")
  }
  expect_warning(
    withCallingHandlers(x(""), warning = w),
    regexp = "Warning",
    fixed = TRUE
  )
  expect_warning(
    withCallingHandlers(x("CAVE CANEM"), warning = w),
    regexp = "CAVE CANEM",
    fixed = TRUE
  )

})

#' return a flip term for `ggplot2` plots, if desired.
#'
#' @param w width of the plot to determine its aspect ratio
#' @param h height of the plot to determine its aspect ratio
#' @param p the `ggplot2` object, so far. If `w` or `h` are missing, `p` is used
#'        for an estimate on `w` and `h`, if both axes are discrete.
#' @param ref_env environment of the actual entry function, so that the correct
#'                formals can be detected.
#' @param ... additional arguments for `coord_flip` or `coord_cartesian`
#'
#' @return `coord_flip` or `coord_cartesian`
#' @importFrom ggplot2 coord_flip coord_cartesian
#'
#' @family process_functions
#' @concept process
#' @noRd
util_coord_flip <- function(w, h, p, ref_env, ...) {
  ## NEU: ggf. lazy-Plot einmal realisieren, bevor wir layer_scales() etc. nutzen
  if (!missing(p) && inherits(p, "dq_lazy_ggplot")) {
    p <- prep_realize_ggplot(p)
  }

  n <- 1
  if (!missing(ref_env)) {
    while (n < 4000 && !identical(rlang::caller_env(n = n), ref_env)) {
      n <- n + 1
    }
  }
  if (n >= 4000) {
    util_error(c("Internal error, sorry. As a dataquieR developer: There is",
                 "some call to util_coord_flip with a ref_env outside the",
                 "call trace, maybe inside a util_create_lean_ggplot()"))
  }
  rm(ref_env) # this reference should go to save resources
  calling_fn_name <- rlang::call_name(rlang::caller_call(n = n))
  if (!"flip_mode" %in% names(formals(calling_fn_name, envir =
                                      rlang::caller_env(n = n)))) {
    util_error("%s can only be called from a function with the %s argument",
               sQuote(rlang::caller_call()[[1]]),
               sQuote("flip_mode"))
  }
  .default <- formals(calling_fn_name, envir =
                        rlang::caller_env(n = n))$flip_mode
  .default <- eval(.default, envir = rlang::caller_env(n = n))
  if (eval(call("missing", as.symbol("flip_mode")), envir =
           rlang::caller_env(n = n))) {
    flip_mode <- getOption("dataquieR.flip_mode", dataquieR.flip_mode_default)
  } else {
    flip_mode <- get("flip_mode", parent.frame())
  }
  flip_mode <- util_match_arg(flip_mode, c("default", "flip", "noflip", "auto"))
  if (flip_mode == "default") {
    flip_mode <- .default
  }

  if ((flip_mode == "auto") && (missing(w) || missing(h))) {

    if (!missing(p)) {
      if (inherits(ggplot2::layer_scales(p)$x, "ScaleDiscrete") &&
          inherits(ggplot2::layer_scales(p)$y, "ScaleDiscrete")) {
        w <- length(ggplot2::layer_scales(p)$x$get_limits())
        h <- length(ggplot2::layer_scales(p)$y$get_limits())
      }
    }

    flip_mode <- .default
    if (flip_mode == "auto") {
      flip_mode <- dataquieR.flip_mode_default
    }
    if (flip_mode == "auto") {
      flip_mode <- "noflip"
    }
  }

  args <- list(...)

  if (flip_mode == "noflip") {
    return(util_create_lean_ggplot(
      do.call(coord_cartesian, args),
      args = args,
      .lazy = FALSE
    ))
  } else if (flip_mode == "flip") {
    return(util_create_lean_ggplot(
      do.call(coord_flip, args),
      args = args,
      .lazy = FALSE
    ))
  } else if (flip_mode == "auto") {
    if (w > h) {
      return(util_create_lean_ggplot(
        do.call(coord_flip, args),
        args = args,
        .lazy = FALSE
      ))
    } else {
      return(util_create_lean_ggplot(
        do.call(coord_cartesian, args),
        args = args,
        .lazy = FALSE
      ))
    }
  } else {
    util_error("Internal error 234243xx2423 -- Should never happen.")
  }
}



#' Lazily add a `coord` to a (possibly lazy) ggplot
#'
#' This helper attaches a coordinate system (e.g. \code{coord_flip()},
#' \code{coord_cartesian()}) to a plot. If \code{p} is a
#' \code{dq_lazy_ggplot}, a new lazy wrapper is created whose
#' expression materializes \code{p} and adds \code{coord}. For plain
#' \code{ggplot} objects, the function simply returns \code{p + coord}.
#'
#' @param p A \code{ggplot} or \code{dq_lazy_ggplot} object.
#' @param coord A ggplot2 coordinate object (e.g. from
#'   \code{ggplot2::coord_flip()}, \code{ggplot2::coord_cartesian()}).
#'
#' @return A \code{ggplot} or \code{dq_lazy_ggplot} object, depending
#'   on \code{p} and the global \code{dataquieR.lazy_plots} option.
#'
#' @noRd
util_lazy_add_coord <- function(p, coord) { # nolint
  ## "altes" Verhalten: für echte ggplots einfach p + coord
  if (!inherits(p, "dq_lazy_ggplot")) {
    return(p + coord)
  }

  ## globale Lazy-Option wie bei %lean+% auswerten
  lazy <- as.logical(getOption("dataquieR.lazy_plots",
                               dataquieR.lazy_plots_default))
  if (length(lazy) != 1 || is.na(lazy)) {
    util_warning(c(
      "Cannot use option dataquieR.lazy_plots %s as a logical value",
      "using %s"
    ),
    dQuote(paste(getOption("dataquieR.lazy_plots",
                           dataquieR.lazy_plots_default),
                 collapse = ",")),
    dQuote(dataquieR.lazy_plots_default))
    lazy <- as.logical(dataquieR.lazy_plots_default)
  }

  ## neue Lazy-Hülle: beim Realisieren wird p materialisiert und coord angehängt
  util_create_lean_ggplot(
    {
      prep_realize_ggplot(p) + coord
    },
    p     = p,
    coord = coord,
    .lazy = lazy
  )
}

#' Convert `x` to valid missing codes
#'
#' @param x [character] a vector of values
#'
#' @return converted `x`
#' @family robustness_functions
#' @noRd
 util_as_valid_missing_codes <- function(x) {
   # TODO: Use it everywhere, where we still have literally defined conversions, currently.
   r <- lapply(x, as.character)
  which_numeric <-  suppressWarnings(is.na(r) == is.na(as.numeric(r)))
  which_datetime <-  suppressWarnings(is.na(r) ==
                                       is.na(lapply(r, util_parse_date)))
  which_time <-  suppressWarnings(is.na(r) ==
                                        is.na(lapply(r, util_parse_time)))
  which_numeric <- which_numeric & !which_datetime & !which_time
  r[which_numeric] <- as.numeric(r[which_numeric])
  r[which_datetime] <- lapply(x[which_datetime], util_parse_date)
  r[which_time] <- lapply(x[which_time], util_parse_time)
  r[!which_numeric & !which_datetime & !which_time] <- NA_character_
  r
}

#' print implementation for the class `interval`
#'
#' such objects, for now, only occur in `RECCap` rules, so this function
#' is meant for internal use, mostly -- for now.
#'
#' @param x `interval` objects to print
#' @param ... not used yet
#'
#' @seealso base::print
#'
#' @return the printed object
#' @export
print.interval <- function(x, ...) {
  util_stop_if_not(inherits(x, "interval"))
  if (x$inc_l) { cat("[") } else { cat("(") }
  if (inherits(x$low, "POSIXct")) {
    cat(as.character(x$low, usetz = FALSE))
  } else {
    cat(as.character(x$low))
  }
  cat(";")
  if (inherits(x$upp, "POSIXct")) {
    cat(as.character(x$upp, usetz = FALSE))
  } else {
    cat(as.character(x$upp))
  }
  if (x$inc_u) { cat("]") } else { cat(")") }
  invisible(x)
}

#' `as.character` implementation for the class `interval`
#'
#' such objects, for now, only occur in `RECCap` rules, so this function
#' is meant for internal use, mostly -- for now.
#'
#' @param x `interval` objects to convert
#' @param ... not used yet
#'
#' @seealso base::as.character
#'
#' @return interval as character
#' @export
as.character.interval <- function(x, ...) {
  r <- util_suppress_output(capture.output(print.interval(x, ...)))
  r
}

#' HTML dependency for `jsPDF`
#'
#' Provides `jsPDF` for use in `Shiny` or `RMarkdown` via `htmltools`.
#'
#' @return An [htmltools::htmlDependency()] object
#' @export
html_dependency_jspdf <- function() {
  if (suppressWarnings(!util_ensure_suggested("visNetwork","PDF Downloads from Plotly Plots",
                        err = FALSE))) {
    return(htmltools::htmlDependency(
      name = "dummy",
      version = "0.0.1",
      src = ""
    ))
  }
  htmltools::htmlDependency(
    name = "jspdf",
    version = "1.3.2",
    src = c(file=system.file("htmlwidgets/lib/export/jsPDF", package="visNetwork")),
    script = "jspdf.debug.js"
  )
  # htmltools::htmlDependency(
  #   name = "jspdf",
  #   version = "2.5.1",
  #   src = system.file("jsPDF", package = "dataquieR"),
  #   script = "jspdf.umd.min.js"
  # )
}

#' Find all columns in item-level-metadata, that refer to some other variable
#'
#' @param meta_data [data.frame] the metadata
#'
#' @return [character] all column names referring to variables from item-level
#'                     metadata
#'
#' @family lookup_functions
#' @concept metadata_management
#' @noRd

util_variable_references <- function(meta_data = "item_level") {
  util_expect_data_frame(meta_data)
  colnames(meta_data)[startsWith(colnames(meta_data), "KEY_") |
                      startsWith(colnames(meta_data), "GROUP_VAR_") |
                      colnames(meta_data) == "TIME_VAR" |
                      colnames(meta_data) == "PART_VAR" |
                      colnames(meta_data) == "CO_VARS"]
}

util_fix_columns_in_dashboard_for_overview <- function(DashboardT, image_dir) {
  util_ensure_suggested("jsonlite",
                        goal = "dashboard views in overall-overviews",
                        err = TRUE)

  if (nrow(DashboardT) == 0) return(DashboardT)

  DashboardT <-
    util_extract_datauri_pngs(DashboardT, image_dir = image_dir)

  if (nrow(DashboardT) == 0) return(DashboardT)

  orig_cn <- colnames(DashboardT)

  trnsl <- util_translate(c("href", "value", "popup_href", "title"),
                          as_this_translation = colnames(DashboardT))

  make_vnlb_link <- function(to_show) {
    if (any(.cempt <- is.na(DashboardT[[to_show]])))
      DashboardT[[to_show]][.cempt] <- ""

    vapply(apply(DashboardT, 1, function(x) {
      res <- x
      try({
        cnt <- x[[to_show]]
        res <- htmltools::a(href = x[[trnsl[["href"]]]],
                            title = htmltools::HTML(as.character(
                              x[[trnsl[["value"]]]])),
                            onclick = htmltools::htmlTemplate(text_ = "(function(e) {
                                                        e.preventDefault();
                                                        showDataquieRResult({{url}}, {{link_url}}, {{title}});
                                                      })(event);",
                                                              url =
                                                                jsonlite::toJSON(paste0(x[[trnsl[["popup_href"]]]]), auto_unbox = TRUE),
                                                              link_url = jsonlite::toJSON(paste0(x[[trnsl[["href"]]]]), auto_unbox = TRUE),
                                                              title = jsonlite::toJSON(x[[trnsl[["title"]]]], auto_unbox = TRUE)
                            ),
                            htmltools::HTML(as.character(cnt)))
      })
      res
    }), as.character, FUN.VALUE = character(1))
  }

  nm <- util_translate(VAR_NAMES, as_this_translation = colnames(DashboardT))
  lb <- util_translate(LABEL, as_this_translation = colnames(DashboardT))
  fig <- util_translate("Figure", as_this_translation = colnames(DashboardT))
  gra <- util_translate("Graph", as_this_translation = colnames(DashboardT))
  fqvn <- util_translate("fq_VARNAME", as_this_translation = colnames(DashboardT))

  DashboardT[[lb]] <- make_vnlb_link(lb)
  DashboardT[[nm]] <- make_vnlb_link(fqvn)
  DashboardT[[fig]] <- make_vnlb_link(fig)
  DashboardT[[gra]] <- make_vnlb_link(gra)

  # ok <- head(levels(DashboardT$Classification), 1)
  #
  # if (length(ok) > 0) {
  #   rows_to_take <- DashboardT$Classification > ok
  # } else {
    rows_to_take <- rep(TRUE, nrow(DashboardT))
  # }

  rows_to_take[is.na(rows_to_take)] <- FALSE

  # columns_to_shorten <-
  #   vapply(DashboardT,
  #          function(x)
  #            suppressWarnings(max(nchar(as.character(x)), na.rm = TRUE)),
  #          FUN.VALUE = numeric(1)) > 2000
  columns_to_shorten <- rep(FALSE, ncol(DashboardT))


  if (any(columns_to_shorten)) {
    util_warning("Removed long columns from overall dashboard") # TODO: Better warning message
    DashboardT[, columns_to_shorten] <-
      NA_character_
  }


  DashboardT <-
    DashboardT[rows_to_take, , FALSE]

  DashboardT$title <- NULL
  DashboardT$href <- NULL
  DashboardT$popup_href <- NULL

  util_translated_colnames(DashboardT) <-
    util_translate(util_translate(colnames(DashboardT),
                                  as_this_translation = orig_cn,
                                  reverse = TRUE),
                   as_this_translation = orig_cn) # TODO: avoid reverse

  DashboardT
}

util_extract_datauri_pngs <- function(DashboardT, image_dir) {
  util_stop_if_not(is.data.frame(DashboardT))
  util_stop_if_not(length(image_dir) == 1L, is.character(image_dir), nzchar(image_dir))

  util_ensure_suggested(
    "jsonlite",
    goal = "dashboard views in overall-overviews",
    err = TRUE
  )

  image_dir_fs <- path.expand(image_dir)

  if (!dir.exists(image_dir_fs)) {
    dir.create(image_dir_fs, recursive = TRUE, showWarnings = FALSE)
  }

  image_dir_href <- basename(normalizePath(image_dir_fs, winslash = "/", mustWork = FALSE))

  normalize_data_uri <- function(uri) {
    paste0(
      "data:image/png;base64,",
      gsub(
        "\\s+",
        "",
        sub("^data:image/png;base64,", "", uri, perl = TRUE),
        perl = TRUE
      )
    )
  }

  replace_src_attr <- function(attr_text, new_path) {
    quote_chr <- sub(
      '^src\\s*=\\s*(["\\\']).*$',
      "\\1",
      attr_text,
      perl = TRUE
    )
    paste0("src=", quote_chr, new_path, quote_chr)
  }

  process_cell <- function(x, dedup_env, file_map_env, next_id_env) {
    if (is.na(x) || !grepl("data:image/png;base64", x, fixed = TRUE)) {
      return(x)
    }

    m <- gregexpr(
      'src\\s*=\\s*["\\\']data:image/png;base64,[A-Za-z0-9+/=[:space:]]+["\\\']',
      x,
      perl = TRUE
    )[[1L]]

    if (identical(m, -1L)) {
      return(x)
    }

    attrs <- regmatches(x, list(m))[[1L]]
    new_attrs <- attrs

    for (i in seq_along(attrs)) {
      attr_text <- attrs[[i]]

      uri <- sub(
        '^src\\s*=\\s*["\\\'](data:image/png;base64,[A-Za-z0-9+/=[:space:]]+)["\\\']$',
        "\\1",
        attr_text,
        perl = TRUE
      )

      uri_norm <- normalize_data_uri(uri)
      payload <- sub("^data:image/png;base64,", "", uri_norm, perl = TRUE)
      dedup_key <- paste0("k_", rlang::hash(payload))

      if (exists(dedup_key, envir = dedup_env, inherits = FALSE)) {
        rel_file <- get(dedup_key, envir = dedup_env, inherits = FALSE)
      } else {
        img_raw <- jsonlite::base64_dec(payload)

        rel_file <- sprintf("img_%04d.png", get("next_id", envir = next_id_env))
        assign("next_id", get("next_id", envir = next_id_env) + 1L, envir = next_id_env)

        writeBin(img_raw, file.path(image_dir_fs, rel_file))
        assign(dedup_key, rel_file, envir = dedup_env)

        file_map <- get("file_map", envir = file_map_env)
        file_map[[length(file_map) + 1L]] <- list(
          file = file.path(image_dir_fs, rel_file),
          uri = uri_norm
        )
        assign("file_map", file_map, envir = file_map_env)
      }

      new_attrs[[i]] <- replace_src_attr(
        attr_text,
        file.path(image_dir_href, rel_file)
      )
    }

    regmatches(x, list(m)) <- list(new_attrs)
    x
  }

  out <- DashboardT
  dedup_env <- new.env(parent = emptyenv())
  file_map_env <- new.env(parent = emptyenv())
  next_id_env <- new.env(parent = emptyenv())

  assign("file_map", list(), envir = file_map_env)
  assign("next_id", 1L, envir = next_id_env)

  hits <- matrix(
    FALSE,
    nrow = nrow(out),
    ncol = ncol(out),
    dimnames = list(rownames(out), names(out))
  )

  for (j in seq_along(out)) {
    if (is.factor(out[[j]])) {
      out[[j]] <- as.character(out[[j]])
    }

    if (!is.character(out[[j]])) {
      next
    }

    col_hits <- !is.na(out[[j]]) & grepl("data:image/png;base64", out[[j]], fixed = TRUE)
    hits[, j] <- col_hits

    if (any(col_hits)) {
      out[[j]][col_hits] <- vapply(
        out[[j]][col_hits],
        process_cell,
        character(1L),
        dedup_env = dedup_env,
        file_map_env = file_map_env,
        next_id_env = next_id_env
      )
    }
  }

  file_map_list <- get("file_map", envir = file_map_env)
  file_map <- if (length(file_map_list)) {
    do.call(
      rbind,
      lapply(file_map_list, function(x) {
        data.frame(
          file = x$file,
          uri = x$uri,
          stringsAsFactors = FALSE
        )
      })
    )
  } else {
    data.frame(
      file = character(),
      uri = character(),
      stringsAsFactors = FALSE
    )
  }

  out
}

# This file creates an environment with functions, that
# employ meta_data found in this environment (must be placed there by their
# caller, before)
# used by [util_generate_calls_for_function]

# CAVE: FUNCTIONS MUST NOT START WITH A . IF USED FOR POPULATION!!

#' `.meta_data_env` -- an environment for easy metadata access
#'
#' used by the dq_report2-pipeline
#' @seealso `meta_data_env_id_vars()` `meta_data_env_co_vars()`
#'          `meta_data_env_time_vars()` `meta_data_env_group_vars()`
#' @name meta_data_env
#' @keywords internal
.meta_data_env <- new.env(parent = environment())

#' Extract id variables for a given item or variable group
#' @param entity vector of item- or variable group identifiers
#' @details
#' In the environment, `target_meta_data` should be set either to
#' `item_level` or to `cross-item_level`.
#' @return a vector with id-variables for each entity-entry, having the
#'         `explode` attribute set to `FALSE`
#' @seealso [meta_data_env]
#' @name meta_data_env_id_vars
#' @noRd
.meta_data_env$id_vars <- function(entity) {
  util_expect_scalar(entity, check_type = is.character)
  if (target_meta_data == "cross-item_level") {
    entity <- names(util_parse_assignments(entity))
  } else if (target_meta_data == "item_level") {
    entity <- entity
  } else {
    util_error()
  }
  segments <- unique(util_find_var_by_meta(entity,
                                           target = STUDY_SEGMENT,
                                           meta_data = meta_data))
  r <- unique(meta_data_segment[meta_data_segment[[STUDY_SEGMENT]] %in%
                                  segments,
                                SEGMENT_ID_VARS])
  r <- names(util_parse_assignments(r))
  r <- util_find_var_by_meta( r,
                              meta_data,
                              label_col = label_col,
                              target = label_col,
                              ifnotfound = r)
  if (length(r) > 0)
    attr(r, "explode") <- FALSE
  r
}


#' Extract `MULTIVARIATE_OUTLIER_CHECK` for variable group
#' @param entity vector of item- or variable group identifiers
#' @details
#' In the environment, `target_meta_data` should be set either to
#' `item_level` or to `cross-item_level`.
#' @return a vector with flags for each entity-entry, having the
#'         `explode` attribute set to `FALSE`
#' @seealso [meta_data_env]
#' @name meta_data_env_criteria
#' @noRd
.meta_data_env$multivariate_outlier_check <- function(entity) {
  util_expect_scalar(entity, check_type = is.character)
  if (target_meta_data == "cross-item_level") {
    if (!CONTRADICTION_TERM %in% colnames(meta_data_cross_item)) {
      meta_data_cross_item[[CONTRADICTION_TERM]] <- NA_character_
    }
    r <- util_empty(
      meta_data_cross_item[!util_empty(meta_data_cross_item[[CHECK_ID]]) &
                             meta_data_cross_item[[CHECK_ID]] ==
                             entity, CONTRADICTION_TERM])
    if (MULTIVARIATE_OUTLIER_CHECK %in% colnames(meta_data_cross_item)) {
      .r <- tolower(trimws(
        meta_data_cross_item[!util_empty(meta_data_cross_item[[CHECK_ID]]) &
                               meta_data_cross_item[[CHECK_ID]] ==
                               entity, MULTIVARIATE_OUTLIER_CHECK]))
      if (.r %in% c("true", "1", "t", "+")) {
        r <- TRUE
      } else if (.r %in% c("false", "0", "f", "-")) {
        r <- FALSE
      } else {
        if (!util_empty(.r)) {
          util_warning("Found invalid entry in %s, treated as missing",
                       sQuote(MULTIVARIATE_OUTLIER_CHECK),
                       applicability_problem = TRUE)
        }
        mvolc <- getOption("dataquieR.MULTIVARIATE_OUTLIER_CHECK",
                      dataquieR.MULTIVARIATE_OUTLIER_CHECK_default)
        util_expect_scalar(mvolc)
        mvolc <- tolower(trimws(as.character(mvolc)))
        if (mvolc == "true") {
          r <- TRUE
        } else if (mvolc == "false") {
          r <- FALSE
        } else {
          if (mvolc != "auto") {
            util_warning("Found invalid entry in option %s, treated as %s",
                         sQuote("dataquieR.MULTIVARIATE_OUTLIER_CHECK"),
                         dQuote(dataquieR.MULTIVARIATE_OUTLIER_CHECK_default))
          }
        }
      }
    }
  } else {
    util_error()
  }
  if (length(r) > 0)
    attr(r, "explode") <- FALSE
  r
}

#' Extract `MAHALANOBIS_THRESHOLD` for variable group
#' @param entity vector of item- or variable group identifiers
#' @details
#' In the environment, `target_meta_data` should be set either to
#' `item_level` or to `cross-item_level`.
#' @return a vector with thresholds for each entity-entry, having the
#'         `explode` attribute set to `FALSE`
#' @seealso [meta_data_env]
#' @name meta_data_env_criteria
#' @noRd
.meta_data_env$mahalanobis_threshold <- function(entity) {
  util_expect_scalar(entity, check_type = is.character)
  if (target_meta_data == "cross-item_level") {
    .r <- getOption("dataquieR.MAHALANOBIS_THRESHOLD",
              dataquieR.MAHALANOBIS_THRESHOLD_default)
    if (MAHALANOBIS_THRESHOLD %in% colnames(meta_data_cross_item)) {
      .r <-
        meta_data_cross_item[!util_empty(meta_data_cross_item[[CHECK_ID]]) &
                               meta_data_cross_item[[CHECK_ID]] ==
                               entity, MAHALANOBIS_THRESHOLD]
      if (tolower(trimws(.r)) %in% c("true", "1", "t", "+")) {
        .r <- dataquieR.MAHALANOBIS_THRESHOLD_default
      }
    } else {
      return("") # do not run the check by default
    }
    r <- suppressWarnings(as.numeric(.r))
    if (!util_empty(r) && (
        r < 0 ||
        r > 1
    )) {
      r <- dataquieR.MAHALANOBIS_THRESHOLD_default
      util_warning("Found invalid entry in %s, used %s",
                   sQuote("MAHALANOBIS_THRESHOLD"),
                   dQuote(r),
                   applicability_problem = TRUE)
    }
  } else {
    util_error()
  }
  if (length(r) > 0)
    attr(r, "explode") <- FALSE
  r
}

#' Extract selected outlier criteria for a given item or variable group
#' @param entity vector of item- or variable group identifiers
#' @details
#' In the environment, `target_meta_data` should be set either to
#' `item_level` or to `cross-item_level`.
#' @return a vector with id-variables for each entity-entry, having the
#'         `explode` attribute set to `FALSE`
#' @seealso [meta_data_env]
#' @name meta_data_env_criteria
#' @noRd
.meta_data_env$criteria <- function(entity) {
  util_expect_scalar(entity, check_type = is.character)
  if (target_meta_data == "cross-item_level") {
    r <- tolower(trimws(unlist(util_parse_assignments(
      meta_data_cross_item[!util_empty(meta_data_cross_item[[CHECK_ID]]) &
                             meta_data_cross_item[[CHECK_ID]] ==
                           entity, MULTIVARIATE_OUTLIER_CHECKTYPE]))))

  } else if (target_meta_data == "item_level") {
    r <- tolower(trimws(unlist(util_parse_assignments(util_find_var_by_meta(
      entity,
                                meta_data,
                                label_col = label_col,
                                target = UNIVARIATE_OUTLIER_CHECKTYPE,
                                ifnotfound = NA_character_)))))
  } else {
    util_error()
  }
  if (length(r) > 0)
    attr(r, "explode") <- FALSE
  r
}

#' Extract outlier rules-number-threshold for a given item or variable group
#' @param entity vector of item- or variable group identifiers
#' @details
#' In the environment, `target_meta_data` should be set either to
#' `item_level` or to `cross-item_level`.
#' @return a vector with id-variables for each entity-entry, having the
#'         `explode` attribute set to `FALSE`
#' @seealso [meta_data_env]
#' @name meta_data_env_n_rules
#' @noRd
.meta_data_env$n_rules <- function(entity) {
  util_expect_scalar(entity, check_type = is.character)
  if (target_meta_data == "cross-item_level") {
    r <- unname(unlist(util_parse_assignments(
      meta_data_cross_item[!util_empty(meta_data_cross_item[[CHECK_ID]]) &
                              meta_data_cross_item[[CHECK_ID]] ==
                             entity, N_RULES])))

  } else if (target_meta_data == "item_level") {
    r <- unname(unlist(util_find_var_by_meta(entity,
                                      meta_data = meta_data,
                                      label_col = label_col,
                                      target = N_RULES,
                                      ifnotfound = NA_integer_)))
  } else {
    util_error()
  }
  if (identical(r, "NA")) {
    r <- NA_character_
  }
  r1 <- suppressWarnings(as.integer(r))
  if (any(is.na(r) != is.na(r1))) {
    util_warning("For %s, %s must be an integer number, it is %s",
                 dQuote(entity), sQuote(N_RULES), dQuote(r),
                 applicability_problem = TRUE)
  }
  r <- r1
  if (length(r) > 0)
    attr(r, "explode") <- FALSE
  r
}

# TODO: Add the following
#use_value_labels,  # TODO: make use_value_labels accessible from metadata
# replace_missing_by c(NA, LABEL, INTERPRET)
# use_replace_limits


#' Extract co-variables for a given item
#' @param entity vector of item-identifiers
#' @return a vector with co-variables for each entity-entry, having the
#'         `explode` attribute set to `FALSE`
#' @seealso [meta_data_env]
#' @name meta_data_env_co_vars
#' @keywords internal
.meta_data_env$co_vars <- function(resp_vars) {
  util_expect_scalar(resp_vars, check_type = is.character)
  r <- lapply(intersect(colnames(meta_data), CO_VARS), function(gv) {
    r <- util_map_labels(resp_vars, meta_data, from = label_col, to = gv,
                         ifnotfound = NA_character_)
    if (all(is.na(r))) {
      return(NA_character_)
    }
    r <- names(util_parse_assignments(r))
    util_find_var_by_meta(r,
                          meta_data,
                          label_col = label_col,
                          target = label_col,
                          ifnotfound = r)

  })
  r <- r[!is.na(r)]
  r <- unlist(r)
  if (length(r) > 0)
    attr(r, "explode") <- FALSE
  r
}

#' Extract measurement time variable for a given item
#' @param entity vector of item-identifiers
#' @return a vector with time-variables (usually one per item) for each
#'         entity-entry, having the `explode` attribute set to `TRUE`
#' @seealso [meta_data_env]
#' @name meta_data_env_time_vars
#' @noRd
.meta_data_env$time_vars <- function(resp_vars) {
  util_expect_scalar(resp_vars, check_type = is.character)
  r <- vapply(FUN.VALUE = character(1),
              intersect(colnames(meta_data), TIME_VAR), function(gv) {
                r <- util_map_labels(resp_vars, meta_data, from = label_col, to = gv,
                                     ifnotfound = NA_character_)
                util_find_var_by_meta(r,
                                      meta_data,
                                      label_col = label_col,
                                      target = label_col,
                                      ifnotfound = r)
              })
  if (length(r) > 0)
    attr(r, "explode") <- TRUE
  r
}

# TODO: write an internal documentation
.meta_data_env$resp_vars <- function(resp_vars, f = fkt) { # FIXME: Make multivariate
  if (length(resp_vars) == 0) return(resp_vars)
  util_stop_if_not(length(resp_vars) == 1)
  if (COMPUTED_VARIABLE_ROLE %in% colnames(meta_data)) {
    cvr <- meta_data[meta_data[[label_col]] == resp_vars,
                      COMPUTED_VARIABLE_ROLE, TRUE]
    if (length(cvr) == 1 && !util_empty(cvr)) {
      applicable_functions <- unlist(util_parse_assignments(util_get_concept_info("ssi",
                                                   get("SSI_METRICS") == cvr,
                                                   "functions",
                                                   drop = TRUE)))
      applicable_functions <- unique(gsub("\\..*$", "", applicable_functions))
      if (fkt %in% applicable_functions) {
        return(resp_vars)
      } else {
        return(character(0))
      }
    } else if (!apply_to_nssi) {
      return(character(0))
    }
  } else if (!apply_to_nssi) {
    return(character(0))
  }
  only_roles <- util_get_concept_info("implementations", get("function_R")
                                   == f, "only_roles")[["only_roles"]]
  if (length(only_roles) == 1 && !util_empty(only_roles)) {
    only_roles <- util_parse_assignments(only_roles)
  } else {
    if (length(only_roles) != 1) {
      util_warning(c("Internal warning, sorry; please report: not exactly",
                     "one entry for %s inside DQ_OBS"),
                   sQuote(f))
    }
    if (startsWith(f, "int_") ||
        startsWith(f, "des_")) {
      ## default for integrity and descripors
      only_roles <- unname(vapply(VARIABLE_ROLES,  # for all variables
                                  identity,
                                  FUN.VALUE = character(1)))
    } else { # default for neither int_ nor des_
      only_roles <- c(VARIABLE_ROLES$PRIMARY, VARIABLE_ROLES$SECONDARY)
    }
  }
  if (meta_data[meta_data[[label_col]] == resp_vars, VARIABLE_ROLE, TRUE]
      %in% only_roles) {
    resp_vars
  } else {
    character(0)
  }
}

#' Extract group variables for a given item
#' @param entity vector of item-identifiers
#' @return a vector with possible group-variables (can be more than  one per
#'         item) for each entity-entry, having the `explode` attribute
#'         set to `TRUE`
#' @name meta_data_env_group_vars
#' @seealso [meta_data_env]
#' @noRd
.meta_data_env$group_vars <- function(resp_vars) {
  util_expect_scalar(resp_vars, check_type = is.character)
  r <- vapply(FUN.VALUE = character(1),
              colnames(meta_data)[startsWith(colnames(meta_data),
                                             "GROUP_VAR_")], function(gv) {
                                               r <- util_map_labels(resp_vars, meta_data, from = label_col, to = gv,
                                                                    ifnotfound = NA_character_)
                                               util_find_var_by_meta(r,
                                                                     meta_data,
                                                                     label_col = label_col,
                                                                     target = label_col,
                                                                     ifnotfound = r)

                                             })
  r <- r[!util_empty(r)]
  r <- c(NA_character_, r)
  if (length(r) > 0)
    attr(r, "explode") <- TRUE
  r
}

.dqenv <- environment()
util_meta_data_env <- local({
  # make all the functions in the environment enclosed by this environment, too,
  # so that they can look up this environment for metadata
  fix_fkts <- function(e = .meta_data_env) {
    for (f in ls(e)) {
      if (is.function(e[[f]])) {
        environment(e[[f]]) <- e
      }
    }
    e
  }
  .dqenv$.meta_data_env <- fix_fkts(.meta_data_env)
  return(function(meta_data_v2,
                  label_col = LABEL,
                  item_level = "item_level",
                  meta_data_cross = "cross-item_level",
                  meta_data_segment = "segment_level",
                  meta_data_dataframe = "dataframe_level",
                  study_data) {
    .dfre <- new.env(parent = emptyenv())
    list2env(as.list(.dataframe_environment()), .dfre)
    with_dataframe_environment(
      env = .dfre,
      {
      util_maybe_load_meta_data_v2()
      clon <- rlang::env_clone(.meta_data_env)
      clon$.dfre <- .dfre
      clon$label_col <- label_col
      clon$meta_data <- data.frame(VAR_NAMES = character(0),
                                 LABEL = character(0),
                                 LONG_LABEL = character(0))
      clon$meta_data[[label_col]] <- character(0)
      try(
        clon$meta_data <- util_expect_data_frame(item_level,
                                               dont_assign = TRUE),
        silent = TRUE
      )
      if (!missing(study_data)) {
        clon$study_data <- util_expect_data_frame(study_data,
                                                dont_assign = TRUE,
                                                keep_types = TRUE)
      }
      clon$meta_data_segment <- data.frame(STUDY_SEGMENT = character(0))
      try(
        clon$meta_data_segment <- util_expect_data_frame(meta_data_segment,
                                               dont_assign = TRUE),
        silent = TRUE
      )
      clon$meta_data_cross_item <- data.frame(CHECK_ID = character(0),
                                            CHECK_LABEL = character(0))
      try(
        clon$meta_data_cross_item <- util_expect_data_frame(meta_data_cross,
                                               dont_assign = TRUE),
        silent = TRUE
      )
      clon$meta_data_dataframe <- data.frame(DF_NAME = character(0))
      try(
        clon$meta_data_dataframe <- util_expect_data_frame(meta_data_dataframe,
                                               dont_assign = TRUE),
        silent = TRUE
      )
      clon$provisionize_call <- function(cal, internal = FALSE, env) { # TODO: never pass the call, always find it on the stack to make this less quirks.
        util_expect_scalar(internal, check_type = is.logical)
        if (!internal) {
          cal <- substitute(cal)
        }
        util_stop_if_not(is.call(cal))
        util_stop_if_not(rlang::call_ns(cal) == "dataquieR")
        fkt <- rlang::call_name(cal)
        if (!internal && !(fkt %in% getNamespaceExports("dataquieR"))) {
          util_error("%s is not an exported function of %s",
                     sQuote(fkt),
                     sQuote(packageName()))
        }
        fn <- try(get(fkt, envir = asNamespace("dataquieR")), silent = TRUE)
        if (internal && util_is_try_error(fn) &&
            (identical(conditionMessage(attr(fn, "condition")),
                      "object 'FUN' not found") ||
             identical(conditionMessage(attr(fn, "condition")),
                       "invalid first argument"))) {
          fn <- try(rlang::caller_fn(3), silent = TRUE)
        }
        util_stop_if_not(is.function(fn))
        cal <- rlang::call_match(cal, fn) # normalize call, so that all arguments are nambed, if possible
        to_fill <-
          setdiff(intersect(names(formals(fn)), ls(parent.env(environment()))),
                  rlang::call_args_names(cal))
        if ("item_level" %in% names(cal) && !identical(cal$item_level,
                                                       rlang::missing_arg())) {
          to_fill <- setdiff(to_fill, "meta_data")
        }
        target_meta_data <- NULL
        if ("resp_vars" %in% names(formals(fn))) {
          if ("variable_group" %in% names(formals(fn))) {
            util_error(c("Internal error, sorry, please report:",
                         "%s must either work on item- or on cross-item-",
                         "level, but not both -- it has the formal arguments",
                         "%s as well as %s, which cannot be."),
                       sQuote(rlang::call_name(cal)),
                       sQuote("resp_vars"),
                       sQuote("variable_group")
                      )
          }
          target_meta_data <- "item_level"
          if ("resp_vars" %in% names(cal)) {
            if (!missing(env) && is.environment(env)) {
              cal$resp_vars <- eval(as.symbol("resp_vars"), envir = env)
            }
            entity <- util_find_var_by_meta(cal$resp_vars,
                                            meta_data = meta_data,
                                            label_col = label_col,
                                            target = label_col)
            if (any(is.na(entity)) && !internal) {
              util_warning(
                c("Could not find the following variables in %s: %s,",
                  "ignoring them."),
                sQuote("meta_data"),
                util_pretty_vector_string(cal$resp_vars[is.na(entity)]))
              entity <- entity[!is.na(entity)]
            }
          } else {
            entity <- formals(fn)[["resp_vars"]]
          }
        } else if ("variable_group" %in% names(formals(fn))) {
          if ("resp_vars" %in% names(formals(fn))) {
            util_error(c("Internal error, sorry, please report:",
                         "%s must either work on item- or on cross-item-",
                         "level, but not both -- it has the formal arguments",
                         "%s as well as %s, which cannot be."),
                       sQuote(rlang::call_name(cal)),
                       sQuote("resp_vars"),
                       sQuote("variable_group")
            )
          }
          target_meta_data <- "cross-item_level"
          if ("variable_group" %in% names(cal)) {
            if (!missing(env) && is.environment(env)) {
              cal$variable_group <- eval(as.symbol("variable_group"),
                                         envir = env)
            }
            entity <- cal$variable_group
          } else{
            entity <- formals(fn)[["variable_group"]]
          }
        }
        assign("target_meta_data", target_meta_data,
               envir = parent.env(environment()))
        withr::defer({
          rm("target_meta_data", envir = parent.env(environment()))
        })
        assign("fkt", fkt,
               envir = parent.env(environment()))
        withr::defer({
          rm("fkt", envir = parent.env(environment()))
        })
        for (arg in to_fill) {
          if (is.function(parent.env(environment())[[arg]]) &&
              !util_is_try_error(try(!missing(entity), silent = TRUE)) &&
              !missing(entity)) {
            .pv_args <- list(entity)
            .vals <- try(do.call(parent.env(environment())[[arg]], .pv_args), silent = TRUE)
            if (util_is_try_error(.vals)) {
              .vals <- NULL
            } else if (identical(attr(.vals, "explode"), TRUE)) {
              if (sum(!is.na(.vals)) == 1) {
                .vals <- .vals[!is.na(.vals)][[1]]
              } else {
                .vals <- .vals[!is.na(.vals)]
                if (length(.vals) == 0) {
                  rsnble <- "<none>"
                } else {
                  rsnble <- sQuote(paste0(.vals, " (", names(.vals), ")",
                                          collapse = ", "))
                }
                util_error(c("For %s, you need to specify %s explicitly,",
                             "according to your metadata, the following",
                             "values may be reasonable: %s"),
                           dQuote(entity),
                           sQuote(arg),
                           rsnble,
                           applicability_problem = TRUE,
                           intrinsic_applicability_problem = FALSE)
              }
            }
            if (!is.null(cal[[arg]]) || !is.null(.vals)) {
              cal[[arg]] <-
                .vals
            }
          } else {
            .vals <- parent.env(environment())[[arg]]
            if (!is.null(cal[[arg]]) || (!is.null(.vals) &&
                                         !is.function(.vals))) {
              cal[[arg]] <-
                .vals
            }
          }
        }
        cal
      }

      clon$call <- function(cal) {
        cl <- sys.call()
        cl[[1]] <- as.symbol("provisionize_call") # a bit hacky, and using debug output, it should not work this way, but it does?!
        with_dataframe_environment(env = .dfre,
                                   eval.parent(eval(cl)))
      }

      # Alternative solution using kind of a functinal this but with
      # cirucular references:
      # clon$call <- function(cal) {
      #   cal <- substitute(cal)
      #   prov_cal <-
      #     clon$provisionize_call(cal, internal = TRUE)
      #   with_dataframe_environment(env = .dfre,
      #                              eval.parent(prov_cal))
      #
      # }
      #
      # # The following code needs to be cleaned to save memory, but according to
      # # Gemini and ChatGPT does R a "Mark-and-Sweep" and will remove this, as
      # # soon as no reference to clon exists, any more. 2025-06-06 Str /
      # clon$clon <- clon
      # ##

      r <- fix_fkts(clon)
      r
    })
  })
})

#' Utility function for smoothes and plots adjusted longitudinal measurements
#'
#' @description
#' The following R implementation executes calculations for quality indicator
#' "Unexpected location" (see [here](
#' https://dataquality.qihs.uni-greifswald.de/PDQC_DQ_3_2_1_3.html
#' ). Local regression (LOESS) is a versatile statistical method to explore an
#' averaged course of time series
#' measurements (Cleveland, Devlin, and Grosse 1988). In context of
#' epidemiological data, repeated measurements using the same measurement
#' device or by the same examiner can be considered a time series. LOESS allows
#' to explore changes in these measurements over time.
#'
#' [Descriptor]
#'
#' @param resp_vars [variable] the name of the continuous (or binary)
#'                             measurement variable
#' @param group_vars [variable] the name of the observer, device or reader
#'                             variable
#' @param time_vars [variable] the name of the variable giving the time
#'                             of measurement
#' @param co_vars [variable list] a vector of co-variables for adjustment, for
#'                             example age and sex. Can be NULL (default) for no
#'                             adjustment.
#' @param min_obs_in_subgroup [integer] (optional argument) If `group_vars` is
#'                             specified, this argument can be used to specify
#'                             the minimum number of observations required for
#'                             each of the subgroups. Subgroups with fewer
#'                             observations are excluded. The default number
#'                             is `30`.
#' @param label_col [variable attribute] the name of the column in the metadata
#'                             with labels of variables
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param meta_data [data.frame] the data frame that contains metadata
#'                             attributes of study data
#' @param resolution [integer] the maximum number of time points used for
#'                             plotting the trend lines
#' @param comparison_lines [list] type and style of lines with which trend
#'                             lines are to be compared. Can be mean +/- 0.5
#'                             standard deviation (the factor can be specified
#'                             differently in `sd_factor`) or quartiles
#'                             (Q1, Q2, and Q3). Arguments `color` and
#'                             `linetype` are passed to [ggplot2::geom_line()].
#' @param mark_time_points [logical] mark time points with observations
#'                             (caution, there may be many marks)
#' @param plot_observations [logical] show observations as scatter plot in the
#'                             background. If there are `co_vars` specified,
#'                             the values of the observations in the plot will
#'                             also be adjusted for the specified covariables.
#' @param plot_format [enum] AUTO | COMBINED | FACETS | BOTH. Return the plot
#'                             as one combined plot for all groups or as
#'                             facet plots (one figure per group). `BOTH` will
#'                             return both variants, `AUTO` will decide based
#'                             on the number of observers.
#' @param n_group_max [integer] maximum number of categories to be displayed
#'                  individually for the grouping variable (`group_vars`,
#'                  devices / examiners)
#' @param enable_GAM [logical] Can LOESS computations be replaced by general
#'                  additive models to reduce memory consumption  for large
#'                  datasets?
#' @param exclude_constant_subgroups [logical] Should subgroups with constant
#'                  values be excluded?
#' @param min_bandwidth [numeric] lower limit for the LOESS bandwidth, should be
#'                  greater than 0 and less than or equal to 1. In general,
#'                  increasing the bandwidth leads to a smoother trend line.
#'
#' @return a [list] with:
#'   - `SummaryPlotList`: list with two plots if `plot_format = "BOTH"`,
#'   otherwise one of the two figures described below:
#'     - `Loess_fits_facets`: The plot contains LOESS-smoothed curves
#'       for each level of the `group_vars` in a separate panel. Added trend
#'       lines represent mean and standard deviation or quartiles (specified
#'       in `comparison_lines`) for moving windows over the whole data.
#'     - `Loess_fits_combined`: This plot combines all curves into one
#'       panel. Given a low number of levels in the `group_vars`, this plot
#'       eases comparisons. However, if the number increases this plot may
#'       be too crowded and unclear.
#'
#' @details
#'
#' If `mark_time_points` or `plot_observations` is selected, but would result in
#' plotting more than 400 points, only a sample of the data will be displayed.
#'
#' Limitations
#'
#' The application of LOESS requires model fitting, i.e. the smoothness
#' of a model is subject to a smoothing parameter (span).
#' Particularly in the presence of interval-based missing data, high
#' variability of measurements combined with a low number of
#' observations in one level of the `group_vars` may distort the fit.
#' Since our approach handles data without knowledge
#' of such underlying characteristics, finding the best fit is complicated if
#' computational costs should be minimal. The default of
#' LOESS in R uses a span of 0.75, which provides in most cases reasonable fits.
#' The function `util_acc_loess_continuous` adapts the span for
#' each level of the `group_vars`
#' (with at least as many observations as specified in `min_obs_in_subgroup`
#' and with at least three time points) based on the respective
#' number of observations.
#' LOESS consumes a lot of memory for larger datasets.
#' That is why `util_acc_loess_continuous`
#' switches to a generalized additive model with integrated smoothness
#' estimation (`gam` by `mgcv`) if there are 1000 observations or more for
#' at least one level of the `group_vars` (similar to `geom_smooth`
#' from `ggplot2`).
#'
#' @importFrom ggplot2 ggplot aes scale_color_manual xlab ylab geom_point geom_line facet_wrap theme_minimal ggtitle theme element_blank expand_limits
#' @importFrom stats as.formula lm loess predict na.omit glm binomial poisson sd cov var runif
#'
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_loess.html
#' )
#' @noRd
util_acc_loess_continuous <- function(
    resp_vars,
    label_col = NULL,
    study_data,
    item_level = "item_level",
    group_vars = NULL,
    time_vars,
    co_vars = NULL,
    min_obs_in_subgroup =
      getOption("dataquieR.acc_loess.min_obs_in_subgroup",
                dataquieR.acc_loess.min_obs_in_subgroup_default),
    resolution = 80,
    comparison_lines = list(type = c("mean/sd", "quartiles"),
                            color = "grey30",
                            linetype = 2,
                            sd_factor = 0.5),
    mark_time_points = getOption("dataquieR.acc_loess.mark_time_points",
                                 dataquieR.acc_loess.mark_time_points_default),
    plot_observations = getOption("dataquieR.acc_loess.plot_observations",
                                  dataquieR.acc_loess.plot_observations_default),
    plot_format =  getOption("dataquieR.acc_loess.plot_format",
                             dataquieR.acc_loess.plot_format_default),
    meta_data = item_level,
    n_group_max = getOption("dataquieR.max_group_var_levels_in_plot",
                            dataquieR.max_group_var_levels_in_plot_default),
    enable_GAM = getOption("dataquieR.GAM_for_LOESS",
                           dataquieR.GAM_for_LOESS_default),
    exclude_constant_subgroups =
      getOption("dataquieR.acc_loess.exclude_constant_subgroups",
                dataquieR.acc_loess.exclude_constant_subgroups_default),
    min_bandwidth = getOption("dataquieR.acc_loess.min_bw",
                              dataquieR.acc_loess.min_bw_default)) {
  # preps ----------------------------------------------------------------------
  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata = TRUE)

  # correct variable use?
  # (checked before, but included here to catch implementation errors)
  util_correct_variable_use("resp_vars",
                            need_type = "!string",
                            need_scale = "interval | ratio",
                            allow_all_obs_na = FALSE)
  util_correct_variable_use("group_vars",
                            need_scale = "nominal | ordinal",
                            allow_all_obs_na = TRUE,
                            allow_na = TRUE,
                            allow_null = TRUE)
  util_correct_variable_use("time_vars",
                            need_type = DATA_TYPES$DATETIME,
                            need_scale = "interval | ratio",
                            allow_all_obs_na = FALSE,
                            min_distinct_values = 3)
  util_correct_variable_use("co_vars",
                            overwrite = TRUE,
                            remove_not_found = TRUE,
                            allow_more_than_one = TRUE,
                            allow_all_obs_na = FALSE,
                            allow_na = TRUE,
                            allow_null = TRUE)

  # support time course plots without (sub-)groups
  if (is.null(group_vars) || all(util_empty(group_vars))) {
    # create a dummy grouping variable that is not yet contained in ds1
    group_vars <- "dummy_group"
    while (group_vars %in% colnames(ds1)) {
      group_vars <- paste0("dummy_group",
                           ceiling(runif(n = 1, min = 1, max = ncol(ds1) * 2)),
                           sep = "_")
    }
    ds1[[group_vars]] <- 1
    plot_title <- paste("Time course plot for", resp_vars)
    # The dummy variable should not be mentioned in the title of the plot.
  } else {
    plot_title <- paste("Effects of", group_vars, "in", resp_vars)
  }

  if (is.null(co_vars)) {
    co_vars <- character(0)
  }
  co_vars <- na.omit(co_vars)

  n_lev_gv_total <- length(na.omit(unique(ds1[[group_vars]])))

  # check that other arguments are specified correctly
  if (!is.list(comparison_lines) ||
      !all(names(comparison_lines) %in%
           c("type", "color", "linetype", "sd_factor")) ||
      ("type" %in% names(comparison_lines) &&
       !all(comparison_lines$type %in% c("mean/sd", "quartiles"))) ||
      ("type" %in% names(comparison_lines) &&
       "mean/sd" %in% comparison_lines$type  &&
       "sd_factor" %in% names(comparison_lines) &&
       (!is.numeric(comparison_lines$sd_factor) ||
        !is.finite(comparison_lines$sd_factor) ||
        length(comparison_lines$sd_factor) != 1
        )
      )
  ) {
    util_error(c(
      "%s needs to be a list of arguments as specified in the documentation."),
      dQuote("comparison_lines"),
      applicability_problem = TRUE)
  }
  if ("type" %in% names(comparison_lines)) {
    if (length(comparison_lines$type) > 1) {
      comparison_lines$type <- comparison_lines$type[1]
    }
    lines_to_add <- comparison_lines$type
  } else {
    lines_to_add <- "mean/sd"
  }
  sd_fac <- NULL
  if (lines_to_add == "mean/sd" & "sd_factor" %in% names(comparison_lines)) {
    sd_fac <- comparison_lines$sd_factor
  } else {
    sd_fac <- 0.5
  }
  lines_arg <- list(color = "grey30", linetype = 2)
  if ("color" %in% names(comparison_lines)) {
    lines_arg$color <- comparison_lines$color
  }
  if ("linetype" %in% names(comparison_lines)) {
    lines_arg$linetype <- comparison_lines$linetype
  }

  util_expect_scalar(mark_time_points,
                     check_type = is.logical)

  util_expect_scalar(plot_observations,
                     check_type = is.logical)

  # omit missing values and unnecessary variables
  n_prior <- nrow(ds1)
  ds1 <- ds1[, c(resp_vars, time_vars, group_vars, co_vars)]
  if (grepl("dummy_group", group_vars)) {
    # Only mention the 'dummy_group' in the message if it contributes any
    # missing values, otherwise do not mention it.
    if (any(is.na(ds1[complete.cases(ds1[, c(time_vars, co_vars)]),
                      group_vars]))) {
      msg_part1 <- paste0(c(group_vars, co_vars), collapse = ", ")
    } else {
      msg_part1 <- paste0(co_vars, collapse = ", ")
    }
  } else {
    msg_part1 <- paste0(c(group_vars, co_vars), collapse = ", ")
  }
  ds1 <- ds1[complete.cases(ds1[, c(time_vars, group_vars, co_vars)]), ]
  n_post <- nrow(ds1)
  msg <- NULL
  if (n_post < n_prior) {
    msg <- paste0(
      "Due to missing values in ",
      ifelse(nchar(msg_part1) > 0, paste0(msg_part1, " or "), ""),
      time_vars, ", N = ", n_prior - n_post,
      " observations were excluded. "
    )
  }
  n_prior <- n_post
  ds1 <- ds1[complete.cases(ds1), ]
  n_post <- nrow(ds1)
  if (n_post < n_prior) {
    msg <- paste0(
      msg, "Due to missing values in ", resp_vars, ", N = ",
      n_prior - n_post, " observations were excluded",
      ifelse(nchar(msg) > 0, " additionally.", "."))
  }
  if (!is.null(msg) && nchar(msg) > 0) {
    util_message(trimws(msg),
                 applicability_problem = FALSE)
  }

  # convert group_vars to factor
  if (!is.factor(ds1[[group_vars]])) {
    ds1[[group_vars]] <- factor(ds1[[group_vars]])
  }

  # TODO: use util_check_group_levels
  # too few observations per level?
  # check which groups do not have enough observations or time points
  tab_groups <- table(ds1[[group_vars]])
  groups_below_min_obs <- names(tab_groups)[tab_groups < min_obs_in_subgroup]
  tab_groups_tp <- vapply(levels(ds1[[group_vars]]), FUN.VALUE = numeric(1),
                          FUN = function(gr) {
                            length(unique(ds1[[time_vars]][
                              ds1[[group_vars]] == gr]))
                          })
  groups_with_few_tp <- names(tab_groups_tp)[tab_groups_tp < 3]
  if (length(groups_below_min_obs) > 0 | length(groups_with_few_tp) > 0) {
    to_excl <- unique(c(groups_below_min_obs, groups_with_few_tp))
    util_message(paste("Levels of the group_var with too few observations",
                       "were discarded",
                       paste0("(level",
                              ifelse(length(to_excl) > 1, "s ", " "),
                              paste(to_excl, collapse = ", "),
                              ").")
    ),
    applicability_problem = FALSE)
    # exclude levels with few observations or time points
    ds1 <- subset(ds1,
                  ds1[[group_vars]] %in%
                    setdiff(levels(ds1[[group_vars]]), to_excl))
    # drop unused levels
    if (!is.factor(ds1[[group_vars]])) {
      ds1[[group_vars]] <- factor(ds1[[group_vars]])
    } else {
      ds1[[group_vars]] <-
        droplevels(ds1[[group_vars]])
    }
  }

  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.",
               applicability_problem = TRUE)
  }

  if (exclude_constant_subgroups) {
    lvl_to_exclude <- levels(ds1[[group_vars]])[
      vapply(levels(ds1[[group_vars]]), FUN.VALUE = logical(1), function(gr) {
        vals <- ds1[[resp_vars]][which(as.character(ds1[[group_vars]]) == gr)]
        var(vals) == 0 || length(vals) < 2
      })
    ]
    if (length(lvl_to_exclude) > 0) {
      util_message(paste("Levels of the group_var with constant values",
                         "were discarded",
                         paste0("(level",
                                ifelse(length(lvl_to_exclude) > 1, "s ", " "),
                                paste(lvl_to_exclude, collapse = ", "),
                                ").")
      ),
      applicability_problem = FALSE)
      ds1 <- subset(ds1,
                    ds1[[group_vars]] %in%
                      setdiff(levels(ds1[[group_vars]]), lvl_to_exclude))
      # drop unused levels
      if (!is.factor(ds1[[group_vars]])) {
        ds1[[group_vars]] <- factor(ds1[[group_vars]])
      } else {
        ds1[[group_vars]] <-
          droplevels(ds1[[group_vars]])
      }
    }
  }

  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.",
               applicability_problem = TRUE)
  }

  # collapse 'rare' groups to reduce the number of levels, if needed
  tab_groups <- table(ds1[[group_vars]])
  have_other <- FALSE
  if (length(tab_groups) > n_group_max) {
    tab_groups <- tab_groups[order(tab_groups, decreasing = TRUE)]
    keep_gr <- names(tab_groups)[1:n_group_max]
    levels(ds1[[group_vars]])[which(!levels(ds1[[group_vars]])
                                    %in% keep_gr)] <- "other"
    tab_groups[["other"]] <- sum(ds1[[group_vars]] == "other")
    have_other <- TRUE
    # new category 'other' should always be the last one
    lvl_gr <-
      c(levels(ds1[[group_vars]])[which(levels(ds1[[group_vars]])
                                        %in% keep_gr)],
        "other")
    ds1[[group_vars]] <- as.character(ds1[[group_vars]])
    ds1[[group_vars]] <- factor(ds1[[group_vars]], levels = lvl_gr)
  }

  if (length(levels(ds1[[group_vars]])) < 2) {
    plot_format <- "COMBINED"
  }

  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.",
               applicability_problem = FALSE)
  }

  # order data by time and groups
  # (for plotting and for the moving window calculations)
  ds1 <- ds1[order(ds1[[time_vars]], ds1[[group_vars]]), ]
  # reduce time points according to the resolution, if needed
  tp_seq <- unique(ds1[[time_vars]])
  tp_round_seq <- util_optimize_sequence_across_time_var(
    time_var_data = tp_seq,
    n_points = resolution)
  ds1[["ROUND_TIME"]] <- suppressWarnings(util_parse_date(
      lubridate::round_date(ds1[[time_vars]], unit = tp_round_seq)))

  # store a numeric version of the original time variable for later calculations
  ds1$time_vars_num <- suppressWarnings(as.numeric(ds1[[time_vars]]))

  # Modelling group-wise trends ------------------------------------------------
  # adjust response for covariables (if any) using a linear model
  if (length(co_vars) > 0) {
    fmla <- as.formula(paste0(paste0(util_bQuote(resp_vars), "~"),
                              paste0(
                                paste0(util_bQuote(co_vars), collapse = " + "),
                                " + ",
                                util_bQuote(group_vars)
                              )))
    lmfit1 <- lm(fmla, data = ds1)
    group_marg <- data.frame(
      emmeans::emmeans(lmfit1, group_vars, type = "response"),
      check.names = FALSE)
    # store residuals (i.e., discard effects from covariables)
    # These values will be used for LOESS fits. In this way, we fit LOESS after
    # adjusting the response for the covariables.
    ds1$Residuals <-
      # estimated mean for each level of the grouping variable
      group_marg$emmean[match(ds1[[group_vars]], group_marg[, group_vars])] +
      # residuals: original value of the response variable - fitted value
      lmfit1$residuals
    rm(lmfit1) # Memory consumption
  } else {
    ds1$Residuals <- ds1[[resp_vars]]
  }

  # calculate LOESS smoothing parameter based on the number of observations
  bw_loess <- min(1, round(100/nrow(ds1), 2)) # upper limit: 1
  bw_loess <- max(min_bandwidth, bw_loess) # lower limit as specified

  # fit LOESS/GAM for each group separately
  grouped_ds1 <- split(ds1, ds1[[group_vars]])
  processed_grouped_ds1 <- lapply(grouped_ds1, function(data_i) {
    if (var(data_i[["Residuals"]]) == 0) { # constant for this subgroup
      df_i <- unique(data_i[, c("Residuals", "ROUND_TIME")])
      fit_vals <- df_i[["Residuals"]]
      data_i_seq <- df_i[["ROUND_TIME"]]
    } else if (max(tab_groups) > 1000 &&
        util_ensure_suggested("mgcv",
                              "use GAM from mgcv instead of loess for lower memory consumption",
                              err = FALSE) &&
        enable_GAM) {
      # If there are too many observations, switch to GAM instead of LOESS
      # because of memory consumption (if available).
      fit_i <- mgcv::gam(Residuals ~ s(time_vars_num, bs = "cs"),
                         method = "REML",
                         data = data_i)
      # To plot the trend line at the time points in `tp_round_seq`
      # (restricted to those values that lie within the observed time period
      # for this group), we need fitted values at these time points.
      data_i_seq <- tp_round_seq[
        (which(tp_round_seq == min(data_i[["ROUND_TIME"]]))):
          (which(tp_round_seq == max(data_i[["ROUND_TIME"]])))]
      data_i_seq_num <- suppressWarnings(as.numeric(data_i_seq))
      data_i_seq_num <- as.data.frame(data_i_seq_num)
      colnames(data_i_seq_num) <- "time_vars_num"

      fit_vals <- mgcv::predict.gam(fit_i, data_i_seq_num)
    } else { # LOWESS
      # fit LOWESS for data_i
      fit_i <- suppressWarnings(
        lowess(x = data_i[["ROUND_TIME"]],
               y = data_i[["Residuals"]],
               f = bw_loess))
      fit_i_df <- unique(as.data.frame(fit_i))
      fit_vals <- fit_i_df$y
      data_i_seq <- util_parse_date(fit_i_df$x)
    }

    pred_df <- data.frame(TIME = data_i_seq,
                          FITTED_VALUE = fit_vals,
                          GROUP = rep(data_i[[group_vars]][1],
                                      length(data_i_seq)))
    return(res_round_tp = pred_df[which(!is.na(pred_df$FITTED_VALUE)), ])
  })
  # https://stackoverflow.com/a/39838759
  fit_groups <- dplyr::bind_rows(processed_grouped_ds1)
  # Memory consumption
  rm(grouped_ds1)

  # Calculate comparison lines -------------------------------------------------
  # We will compute either mean and SD or quartiles using a moving window
  # approach on the original, complete dataset (if not adjusted for covariates)
  # or on the residuals after adjusting for covariates.
  if (length(tp_seq) >= 7) {
    mov_win_width <- round(0.3 * length(tp_seq))
    # split window into two parts to align it approximately at the middle
    # If the number is not even, we will have a smaller first part and a larger
    # second part (by one), otherwise both parts will be equal.
    part1 <- floor(0.5 * mov_win_width)
    part2 <- mov_win_width - part1
  } else { # We have to ensure that part1 contains at least one observation.
    part1 <- floor(0.5 * length(tp_seq))
    # note: length(tp_seq) >= 3 (ensured by checks during data preparation)
    part2 <- length(tp_seq) - part1
  }
  # To reduce computing time, we will only calculate values that are required
  # for the plot.
  mov_win_res <- lapply(tp_round_seq, function(tp) {
    if (tp %in% tp_seq) {
      i <- which(tp_seq == tp)
      res_i <- util_for_moving_window(tp_seq = tp_seq,
                                      ds1_resp_var = ds1[["Residuals"]],
                                      ds1_time_var = ds1[[time_vars]],
                                      i = i,
                                      part1 = part1,
                                      part2 = part2,
                                      mode = lines_to_add,
                                      sd_fac = sd_fac)
    } else {
      # We calculate the two moving windows below and above `tp` and interpolate
      # the required value.
      i1 <- which.max(tp_seq[which(tp_seq < tp)])
      i1 <- i1 + which(tp_seq < tp)[1] - 1
      i2 <- which.min(tp_seq[which(tp_seq > tp)])
      i2 <- i2 + which(tp_seq > tp)[1] - 1
      res_i <- util_for_moving_window(tp_seq = tp_seq,
                                      ds1_resp_var = ds1[["Residuals"]],
                                      ds1_time_var = ds1[[time_vars]],
                                      i = c(i1, i2),
                                      part1 = part1,
                                      part2 = part2,
                                      mode = lines_to_add,
                                      sd_fac = sd_fac)
      if (length(res_i) == 2) {
        data_i <- as.data.frame(do.call(rbind, res_i))
        if (all(is.na(data_i))) {
          res_i <- res_i[[1]]
        } else {
          data_i$TIME <- c(tp_seq[i1], tp_seq[i2])
          # interpolate value for `tp` using the two neighboring time points
          res_i <- vapply(c("low", "mid", "high"), FUN.VALUE = numeric(1),
                          FUN = function(cc) {
                            local_lm <- lm(as.formula(paste(cc, "~ TIME")),
                                           data = data_i)
                            return(suppressWarnings(
                              predict(local_lm,
                                      data.frame("TIME" = tp))))
                          })
        }
      }
    }
    return(res_i)
  })
  global_trends <- as.data.frame(do.call(rbind, mov_win_res))
  # fill in values at the beginning and at the end (`NA` because
  # there were not enough data for a full window)
  compl <- complete.cases(global_trends)
  if (any(compl)) {
    if (!compl[1]) {
      global_trends[1:(which(compl)[1] - 1), ] <- global_trends[which(compl)[1], ]
    }
    if (!compl[nrow(global_trends)]) {
      global_trends[(rev(which(compl))[1] + 1):nrow(global_trends), ] <-
        global_trends[rev(which(compl))[1], ]
    }
  }
  global_trends$TIME <- tp_round_seq

  # Plotting ------------------------------------------------------------------
  if (length(co_vars) > 0) {
    if (length(co_vars) < 10) {
      subtitle <- sprintf("adjusted for %s", paste0(co_vars, collapse = ", "))
    } else {
      subtitle <- sprintf("adjusted for %d variables", length(co_vars))
    }
  } else {
    subtitle <- ""
  }

  {
    n_lev_gv_reduced <-
      length(unique(fit_groups$GROUP))
    if (n_lev_gv_reduced < n_lev_gv_total) {
      appendix <- sprintf("%d of %d levels shown",
                          n_lev_gv_reduced - have_other,
                          n_lev_gv_total)
    } else {
      appendix <- sprintf("all %d levels shown",
                          n_lev_gv_reduced - have_other)
    }
    if (nzchar(subtitle)) {
      subtitle <- paste(subtitle,
                        appendix,
                        sep = " - ")
    } else {
      subtitle <- appendix
    }
  }

  if (any(compl)) {
    if (lines_to_add == "mean/sd") {
      lines_info <- paste("Trend lines shown for comparison indicate mean \u00B1",
                          sd_fac, "SD.")
    } else {
      lines_info <- paste("Trend lines shown for comparison indicate",
                          "quartiles Q1, Q2, Q3.")
    }
  } else {
    lines_info <- ""
  }

  # If observations should be included in the plot, we have to ensure that not
  # too many points are displayed.
  points_shown_max <- 400
  if (mark_time_points | plot_observations) {
    sel_obs <- list(
      # Dashed marks for observed time points on fitted trend line:
      "dp_facets" = 1:nrow(ds1), # for facets
      "dp_comb" = 1:nrow(ds1), # for combined plot
      # Scatter plot of observed values:
      "obs_facets" = 1:nrow(ds1), # for facets
      "obs_comb" = 1:nrow(ds1)) # for combined plot
    # If there are too many points to plot, we will only plot a sample of
    # them.
    if (nrow(ds1) > points_shown_max) {
      # Cut x axis (time) into five segments for (stratified) sampling.
      ds1_time_var_cut <- cut(as.numeric(ds1[[time_vars]]),
                              breaks = 5, labels = FALSE)
      ds1_tab <- table(ds1[[group_vars]], ds1_time_var_cut)
      ds1_tab_adj <- floor(ds1_tab / nrow(ds1) * points_shown_max)
      # store row indices for each group and time segment to ease data wrangling
      gr_ind <- setNames(nm = levels(ds1[[group_vars]]),
                         lapply(levels(ds1[[group_vars]]), function(gr) {
                           which(ds1[[group_vars]] == gr)
                         }))
      tp_ind <- setNames(nm = 1:max(ds1_time_var_cut),
                         lapply(1:max(ds1_time_var_cut), function(tp) {
                           which(ds1_time_var_cut == tp)
                         }))
      withr::local_seed(400) # TODO: Make the 400 option()able
      sel_obs$obs_facets <- sort(unlist(
        lapply(levels(ds1[[group_vars]]), function(gr) {
          if (sum(ds1_tab[gr, ]) > points_shown_max) {
            ds1_tab_gr_adj <- floor(ds1_tab[gr, ] / sum(ds1_tab[gr, ]) *
                                      points_shown_max)
            unlist(lapply(names(ds1_tab_gr_adj), function(tp) {
              gr_tp_ind <- intersect(gr_ind[[gr]], tp_ind[[tp]])
              if (length(gr_tp_ind) == 0 | ds1_tab_gr_adj[tp] == 0) {
                out <- integer(0)
              } else if (var(ds1[gr_tp_ind, "Residuals"]) == 0) {
                out <- sample(gr_tp_ind, size = ds1_tab_gr_adj[tp])
              } else {
                # keep 'extreme' values to show the true range
                gr_tp_max <- gr_tp_ind[which.max(ds1[gr_tp_ind, "Residuals"])]
                gr_tp_min <- gr_tp_ind[which.min(ds1[gr_tp_ind, "Residuals"])]
                if (ds1_tab_gr_adj[tp] <= 2) {
                  out <- sample(c(gr_tp_max, gr_tp_min),
                                size = ds1_tab_gr_adj[tp])
                } else {
                  out <- c(gr_tp_max, gr_tp_min,
                           sample(setdiff(gr_tp_ind, c(gr_tp_max, gr_tp_min)),
                                  size = ds1_tab_gr_adj[tp] - 2))
                }
              }
              return(out)
            }))
          } else {
            return(gr_ind[[gr]])
          }
        })))
      sel_obs$obs_comb <- sort(unlist(
        lapply(levels(ds1[[group_vars]]), function(gr) {
          unlist(
            lapply(colnames(ds1_tab_adj), function(tp) {
              gr_tp_ind <- intersect(gr_ind[[gr]], tp_ind[[tp]])
              if (length(gr_tp_ind) == 0 | ds1_tab_adj[gr, tp] == 0) {
                out <- integer(0)
              } else if (var(ds1[gr_tp_ind, "Residuals"]) == 0) {
                out <- sample(gr_tp_ind, size = ds1_tab_adj[gr, tp])
              } else {
                # keep 'extreme' values to show the true range
                gr_tp_max <- gr_tp_ind[which.max(ds1[gr_tp_ind, "Residuals"])]
                gr_tp_min <- gr_tp_ind[which.min(ds1[gr_tp_ind, "Residuals"])]
                if (ds1_tab_adj[gr, tp] <= 2) {
                  out <- sample(c(gr_tp_max, gr_tp_min),
                                size = ds1_tab_adj[gr, tp])
                } else {
                  out <- c(gr_tp_max, gr_tp_min,
                           sample(setdiff(gr_tp_ind, c(gr_tp_max, gr_tp_min)),
                                  size = ds1_tab_adj[gr, tp] - 2))
                }
              }
              return(out)
            }))
        })))
      # When we mark observed time points on the trend line, we can
      # omit 'duplicated' marks from observations at the same time point for
      # the same group.
      sel_obs$dp_facets <- sel_obs$obs_facets[
        which(!duplicated(ds1[sel_obs$obs_facets, c(group_vars, time_vars)]))]
      sel_obs$dp_comb <- sel_obs$obs_comb[
        which(!duplicated(ds1[sel_obs$obs_comb, c(group_vars, time_vars)]))]
    }
  }

  geom_dp_facets <- NULL
  geom_dp_comb <- NULL
  if (mark_time_points) {
    sel_obs_dp <- sort(unique(c(sel_obs$dp_facets, sel_obs$dp_comb)))
    mark_obs <- ds1[sel_obs_dp, c(time_vars, group_vars)]
    colnames(mark_obs) <- c("TIME", "GROUP")
    mark_obs$facets <- as.numeric(sel_obs_dp %in% sel_obs$dp_facets)
    mark_obs$comb <- as.numeric(sel_obs_dp %in% sel_obs$dp_comb)
    # If we want to show marks for the actual observed time points
    # (not rounded time points) on top of the trend line, we have to estimate
    # the position of the marks on the trend line. Otherwise, marks and trend
    # line will not match if the resolution is reduced considerably in
    # comparison to the actual resolution of the data.
    pred_pos_mark <-
      vapply(1:nrow(mark_obs), FUN.VALUE = numeric(1),
             function(i) {
               tp <- mark_obs[["TIME"]][i]
               gr <- mark_obs[["GROUP"]][i]
               pred_gr <- processed_grouped_ds1[[gr]]
               if (tp %in% pred_gr$TIME) {
                 res_i <- pred_gr$FITTED_VALUE[which(pred_gr$TIME == tp)]
               } else {
                 tp_ind1 <- which.max(pred_gr$TIME[which(pred_gr$TIME < tp)])
                 tp_ind1 <- tp_ind1 + which(pred_gr$TIME < tp)[1] - 1
                 tp_ind2 <- which.min(pred_gr$TIME[which(pred_gr$TIME > tp)])
                 tp_ind2 <- tp_ind2 + which(pred_gr$TIME > tp)[1] - 1
                 if (length(tp_ind1) == 0 & length(tp_ind2) == 0) {
                   res_i <- NA # should not happen
                 } else if (length(tp_ind1) == 0 | length(tp_ind2) == 0) {
                   res_i <- pred_gr$FITTED_VALUE[c(tp_ind1, tp_ind2)]
                 } else {
                   local_lm <- lm(FITTED_VALUE ~ TIME,
                                  data = pred_gr[tp_ind1:tp_ind2, ])
                   res_i <- suppressWarnings(predict(
                     local_lm, data.frame("TIME" = tp)))
                 }
               }
               return(res_i)
             })
    mark_obs$FITTED_VALUE <- pred_pos_mark
    geom_dp_facets <- geom_point(
      shape = "|",
      data = mark_obs[mark_obs$facets == 1, c("TIME", "GROUP", "FITTED_VALUE")]
    )
    geom_dp_comb <- geom_point(
      shape = "|",
      data = mark_obs[mark_obs$comb == 1, c("TIME", "GROUP", "FITTED_VALUE")]
    )
  }

  geom_obs_facets <- NULL
  geom_obs_comb <- NULL
  if (plot_observations) {
    scatter_obs_facets <- ds1[sel_obs$obs_facets,
                              c(time_vars, "Residuals", group_vars)]
    colnames(scatter_obs_facets) <- c("TIME", "FITTED_VALUE", "GROUP")
    geom_obs_facets <- geom_point(data = scatter_obs_facets,
                                  alpha = 0.2)
    scatter_obs_comb <- ds1[sel_obs$obs_comb,
                            c(time_vars, "Residuals", group_vars)]
    colnames(scatter_obs_comb) <- c("TIME", "FITTED_VALUE", "GROUP")
    geom_obs_comb <- geom_point(data = scatter_obs_comb,
                                alpha = 0.2)
  }

  levs <- as.character(levels(ds1[[group_vars]]))

  g_labs <- paste(
    levs,
    sprintf("(n=%d)", unname(tab_groups[levs]))
  )

  if (length(levels(ds1[[group_vars]])) <= 8) {
    hex_code <- c(
      "#56B4E9", "#E69F00",  "#009E73",
      "#F0E442", "#0072B2", "#D55E00", "#CC79A7", "#8C510A"
    )
    hex_code <- head(hex_code, length(levs))
    names(hex_code) <- levs
  } else {
    hex_code <- NULL
  }

  y_min <- mean(ds1$Residuals) - sd(ds1$Residuals)
  y_max <- mean(ds1$Residuals) + sd(ds1$Residuals)

  # Facet-Grids for categorical variable (observer/device)
  p1 <- util_create_lean_ggplot(
    ggplot(fit_groups,
           aes(x = .data$TIME,
               y = .data$FITTED_VALUE,
               color = .data$GROUP)) + {
                 if (!is.null(hex_code)) {
                   scale_color_manual(
                     values = hex_code,
                     breaks = levs,
                     labels = g_labs
                   ) # https://chatgpt.com/c/69e0ecc5-10a8-8397-9f51-0d365cfa1876
                 } else {
                   ggplot2::scale_color_discrete(
                     breaks = levs,
                     labels = g_labs
                   )
                 }
               } +
      xlab(lines_info) +
      ylab("") +
      geom_dp_facets +
      geom_obs_facets +
      geom_line() +
      facet_wrap(~ .data$GROUP, ncol = 2) + #TODO: What about this ~?
      expand_limits(y = c(y_min, y_max)) +
      theme_minimal() +
      ggtitle(plot_title, subtitle) +
      theme(legend.title = element_blank()),
    fit_groups = fit_groups,
    hex_code = hex_code,
    levs = levs,
    g_labs = g_labs,
    lines_info = lines_info,
    geom_dp_facets = geom_dp_facets,
    geom_obs_facets = geom_obs_facets,
    y_min = y_min,
    y_max = y_max,
    plot_title = plot_title,
    subtitle = subtitle)

  if (any(compl)) {
    p1 <- util_create_lean_ggplot(
      p1 +
        geom_line(data = global_trends, aes(y = .data$low, group = NA),
                  color = lines_arg$color, linetype = lines_arg$linetype) +
        geom_line(data = global_trends, aes(y = .data$mid, group = NA),
                  color = lines_arg$color, linetype = 1) +
        geom_line(data = global_trends, aes(y = .data$high, group = NA),
                  color = lines_arg$color, linetype = lines_arg$linetype),
      p1 = p1,
      global_trends = global_trends,
      lines_arg = lines_arg)
  }

  # combined plot
  p2 <- util_create_lean_ggplot(ggplot(fit_groups,
                                       aes(x = .data$TIME,
                                           y = .data$FITTED_VALUE,
                                           group = .data$GROUP,
                                           color = .data$GROUP)) + {
                                             if (!is.null(hex_code)) {
                                               scale_color_manual(
                                                 values = hex_code,
                                                 breaks = levs,
                                                 labels = g_labs
                                               ) # https://chatgpt.com/c/69e0ecc5-10a8-8397-9f51-0d365cfa1876
                                             } else {
                                               ggplot2::scale_color_discrete(
                                                 breaks = levs,
                                                 labels = g_labs
                                               )
                                             }
                                           } +
                                  xlab(lines_info) +
                                  ylab("") +
                                  geom_dp_comb +
                                  geom_obs_comb +
                                  geom_line() +
                                  expand_limits(y = c(y_min, y_max)) +
                                  theme_minimal() +
                                  ggtitle(plot_title, subtitle),
                                fit_groups = fit_groups,
                                hex_code = hex_code,
                                levs = levs,
                                g_labs = g_labs,
                                lines_info = lines_info,
                                geom_dp_comb = geom_dp_comb,
                                geom_obs_comb = geom_obs_comb,
                                y_min = y_min,
                                y_max = y_max,
                                plot_title = plot_title,
                                subtitle = subtitle)

  if (any(compl)) {
    p2 <- util_create_lean_ggplot(
      p2 +
        geom_line(data = global_trends, aes(y = .data$low, group = NA),
                  color = lines_arg$color, linetype = lines_arg$linetype) +
        geom_line(data = global_trends, aes(y = .data$mid, group = NA),
                  color = lines_arg$color, linetype = 1) +
        geom_line(data = global_trends, aes(y = .data$high, group = NA),
                  color = lines_arg$color, linetype = lines_arg$linetype),
      p2 = p2,
      global_trends = global_trends,
      lines_arg = lines_arg)
  }

  if (length(levels(ds1[[group_vars]])) > 1) {
    p2 <- util_create_lean_ggplot(
      p2 +
        theme(legend.title = element_blank()),
      p2 = p2)

  } else {
    p2 <-  util_create_lean_ggplot(
      p2 +
        theme(legend.position = "none"),
      p2 = p2)
  }

  p1 <- util_set_size(p1,
                      width_em = 45,
                      height_em = length(levels(ds1[[group_vars]])) * 15 / 2)
  p2 <- util_set_size(p2, 30, 15)

  pl <- list(
    Loess_fits_facets = p1,
    Loess_fits_combined = p2
  )

  if (length(plot_format) != 1 || !is.character(plot_format)) {
    plot_format <- "NOT character(1) STRING AT ALL"
  }

  #Add attribute with size hints to the combined plot
  if (!is.null(pl[["Loess_fits_combined"]])) {
    .plot1 <- pl[["Loess_fits_combined"]]
    obj1 <- util_create_lean_ggplot(ggplot2::ggplot_build(.plot1),
                                    .plot1= .plot1)
    min_point_line <- min(util_rbind(data_frames_list = obj1$data)$y, na.rm = TRUE)
    max_point_line <- max(util_rbind(data_frames_list = obj1$data)$y, na.rm = TRUE)
    n_groups <- length(unique(util_rbind(data_frames_list = obj1$data)$group))
    min_time <- min(util_rbind(data_frames_list = obj1$data)$x, na.rm = TRUE)
    max_time <- max(util_rbind(data_frames_list = obj1$data)$x, na.rm = TRUE)
    rm(obj1)
  }

  if (plot_format == "BOTH") {
    return(list(SummaryPlotList = pl))
  } else if (plot_format == "COMBINED") {

    return(util_attach_attr(list(SummaryPlotList = setNames(pl["Loess_fits_combined"],
                                           nm = resp_vars)),
                            sizing_hints = list(
                              figure_type_id = "dot_loess",
                              range = max_point_line - min_point_line,
                              no_char_y = max(nchar(c(round(max_point_line, digits = 2),
                                                      round(min_point_line, digits = 2)))),
                              n_groups = n_groups
                            )))
  } else if (plot_format == "FACETS") {
    return(list(SummaryPlotList = setNames(pl["Loess_fits_facets"],
                                           nm = resp_vars)))
  } else if (plot_format != "AUTO") {
    util_message("Unknown %s: %s -- will switch to default value AUTO.",
                 dQuote("plot_format"), dQuote(plot_format),
                 applicability_problem = TRUE)
  }
#  if (length(levels(ds1[[group_vars]])) < 15) {
    selection <- "Loess_fits_combined"
#  } else {
#    selection <- "Loess_fits_facets"
#  }

  pl <- pl[selection]

  return(util_attach_attr(list(SummaryPlotList = setNames(pl, nm = resp_vars)),
                          sizing_hints = list(
                            figure_type_id = "dot_loess",
                            range = max_point_line - min_point_line,
                            no_char_y = max(nchar(c(round(max_point_line, digits = 2),
                                                    round(min_point_line, digits = 2)))),
                            n_groups = n_groups
                          )))
}


util_for_moving_window <- function(tp_seq, ds1_resp_var, ds1_time_var,
                                   i, part1, part2,
                                   mode, sd_fac) {
  if (is.null(i)) {
    return(c("low" = NA,
             "mid" = NA,
             "high" = NA))
  }
  if (length(i) > 1) {
    res_out <- lapply(i, util_for_moving_window,
                      tp_seq = tp_seq, ds1_resp_var = ds1_resp_var,
                      ds1_time_var = ds1_time_var, part1 = part1, part2 = part2,
                      mode = mode, sd_fac = sd_fac)
    return(res_out)
  }
  if (i - part1 < 0 |
      i + part2 > length(tp_seq)) {
    # The moving window overlaps the beginning or end of the sequence of
    # observed time points.
    return(c("low" = NA,
             "mid" = NA,
             "high" = NA))
  } else {
    tp1 <- tp_seq[i - part1 + 1]
    # i is contained in the first part
    tp2 <- tp_seq[i + part2]
    data_i <- ds1_resp_var[ds1_time_var >= tp1 &
                             ds1_time_var <= tp2]
    if (mode == "mean/sd") {
      center <- mean(data_i)
      sd_emp <- sd(data_i)
      upper <- center + sd_fac * sd_emp
      lower <- center - sd_fac * sd_emp
    } else {
      qq <- quantile(data_i,
                     probs = c(0.25, 0.5, 0.75),
                     names = FALSE)
      lower <- qq[1]
      center <- qq[2]
      upper <- qq[3]
    }
    return(c("low" = lower,
             "mid" = center,
             "high" = upper))
  }
}

#' Return hex code colors from color names or `STATAReporter` syntax
#'
#' @param colors the colors, e.g.,"255 0 0" or "red" or "#ff0000"
#'
#' @return character vector with colors using `HTML` hexadecimal encoding, e..g,
#'         "#ff0000" for "red"
#' @noRd
util_col2rgb <- function(colors) {
  with_space <- grepl(" ", colors, fixed = TRUE)
  res <- colors
  rgbmatrix <- try(col2rgb(colors[!with_space], alpha = TRUE), silent = TRUE)
  if (inherits(rgbmatrix, "try-error")) {
    is_color <- vapply(FUN.VALUE = logical(1), colors[!with_space],
                          function(cl) {
                                  !inherits(try(col2rgb(cl, alpha = TRUE),
                                                silent = TRUE), "try-error")
                          })
    util_warning("No known colors: %s, replaced by black.",
                 util_pretty_vector_string(colors[!with_space][!is_color]),
                 applicability_problem = TRUE)
    colors[!with_space][!is_color] <- "black"
    rgbmatrix <- try(col2rgb(colors[!with_space], alpha = TRUE), silent = TRUE)
    util_stop_if_not(
      "Internal error 0x34556345345. Should not happen after cleaning above." =
        !inherits(rgbmatrix, "try-error"))
  }
  res[!with_space] <- apply(rgbmatrix, 2, function(cl) {
    paste0("#", paste0(vapply(cl, sprintf, fmt = "%02x",
                              FUN.VALUE = character(1)), collapse = ""))
  })
  ws <- suppressWarnings(lapply(strsplit(res[with_space],
                                                      "\\s"), as.numeric))
  invalid <- !vapply(ws, length, FUN.VALUE = integer(1)) %in% c(3, 4) |
      vapply(ws, function(triple) any(is.na(triple)),
                 FUN.VALUE = logical(1))
  if (any(invalid)) {
    util_warning("No known colors: %s, replaced by black.",
                 util_pretty_vector_string(colors[with_space][invalid]),
                 applicability_problem = TRUE)
    ws[invalid] <- list(c(0, 0, 0))
  }
  ws <- lapply(ws, sprintf, fmt = "%02x")
  ws <- vapply(ws, paste0, collapse = "",
                            FUN.VALUE = character(1))
  ws <- paste0("#", ws)
  res[with_space] <- ws
  unlist(res)
}

locale_available <- function(locale, category = "LC_TIME") {
  old <- Sys.getlocale(category)
  on.exit(suppressWarnings(Sys.setlocale(category, old)), add = TRUE)

  res <- suppressWarnings(
    tryCatch(
      Sys.setlocale(category, locale),
      warning = function(e) "",
      error   = function(e) ""
    )
  )

  !identical(res, "") && !is.na(res)
}

tz_available <- function(tz) {
  tz %in% OlsonNames()
}

has_valid_system_tz <- function() {
  tz <- attr(Sys.time(), "tzone")
  !is.null(tz) && nzchar(tz) && all(tz %in% OlsonNames())
}

require_english_locale_and_berlin_tz <- function(
    locale_candidates = c("en_US.UTF-8", "English.UTF-8"),
    tz = "Europe/Berlin",
    skip_on_cran = TRUE
) {
  if (isTRUE(skip_on_cran)) {
    testthat::skip_on_cran()
  }

  english_locale <- NULL
  for (loc in locale_candidates) {
    if (locale_available(loc, "LC_TIME")) {
      english_locale <- loc
      break
    }
  }

  if (is.null(english_locale)) {
    testthat::skip(
      sprintf(
        "No English UTF-8 LC_TIME locale available (tried: %s)",
        paste(locale_candidates, collapse = ", ")
      )
    )
  }

  if (!tz_available(tz)) {
    testthat::skip(
      sprintf("Timezone '%s' not available on this platform", tz)
    )
  }

  caller <- parent.frame()
  withr::local_locale(c(LC_TIME = english_locale), .local_envir = caller)
  withr::local_timezone(tz, .local_envir = caller)

  invisible(list(locale = english_locale, tz = tz))
}

test_that("distribution plot works", { # acc_distributions.R ----
  skip_on_cran()
  skip_if_not_installed("lobstr")
  skip_if_not_installed("vdiffr")

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)

  ({
    r <-
      acc_distributions(
        resp_vars = "DBP_0",
        study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
        meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
        label_col = LABEL)

    expect_doppelganger2("acc_distributions gr dbp0", r$SummaryPlotList$DBP_0)
    expect_lt(lobstr::obj_size(r$SummaryPlotList$DBP_0), 15 * 1024 * 1024)

    r <-
      acc_distributions(
        resp_vars = "EDUCATION_0",
        study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
        meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
        label_col = LABEL)

    expect_doppelganger2("acc_distributions def edu0", r$SummaryPlotList$EDUCATION_0)
    expect_lt(lobstr::obj_size(r$SummaryPlotList$EDUCATION_0), 15 * 1024 * 1024)

  })
})

test_that("loess plot works", { # acc_loess.R ----
  # testthat::local_reproducible_output()
  # skip_on_travis() # vdiffr fails
  skip_on_cran()
  skip_if_not_installed("lobstr")
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  skip_if_not_installed("vdiffr")
  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)

  require_english_locale_and_berlin_tz()

  ({
    time_vars <- prep_map_labels("DBP_0", meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|item_level", from = LABEL,
                                 to = TIME_VAR)
    group_vars <- prep_map_labels("DBP_0", meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|item_level", from = LABEL,
                                  to = GROUP_VAR_OBSERVER)
    r <-
      acc_loess(  resp_vars = "DBP_0",
                  study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                  meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
                  group_vars = group_vars,
                  time_vars = time_vars,
                  label_col = LABEL,
                  co_vars = "AGE_0")

    expect_doppelganger2("loess def dbp0", r$SummaryPlotList$DBP_0)
    expect_lt(lobstr::obj_size(r$SummaryPlotList$DBP_0), 15 * 1024 * 1024)

    r <- acc_loess(resp_vars = "EDUCATION_0",
                   study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                   meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
                   time_vars = time_vars,
                   group_vars = group_vars,
                   label_col = LABEL,
                   co_vars = "AGE_0")

    expect_doppelganger2("loess def edu0", r$SummaryPlotList$EDUCATION_0)
    expect_lt(lobstr::obj_size(r$SummaryPlotList$EDUCATION_0), 15 * 1024 * 1024)

    r <-
      acc_loess(  resp_vars = "DBP_0",
                  study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                  meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
                  group_vars = group_vars,
                  time_vars = time_vars,
                  plot_format = "FACETS",
                  label_col = LABEL,
                  co_vars = "AGE_0")

    expect_doppelganger2("loess fac dbp0", r$SummaryPlotList$DBP_0)
    expect_lt(lobstr::obj_size(r$SummaryPlotList$DBP_0), 15 * 1024 * 1024)

    r <-
      acc_loess(  resp_vars = "EDUCATION_0",
                  study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                  meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
                  time_vars = time_vars,
                  group_vars = group_vars,
                  plot_format = "FACETS",
                  label_col = LABEL,
                  co_vars = "AGE_0")

    expect_doppelganger2("loess fac edu0", r$SummaryPlotList$EDUCATION_0)
    expect_lt(lobstr::obj_size(r$SummaryPlotList$EDUCATION_0), 15 * 1024 * 1024)

  })
})

test_that("margins plot works", { # acc_margins.R -----
  skip_on_cran()
  skip_if_not_installed("lobstr")
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("vdiffr")

  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)
  ({
    group_vars <- prep_map_labels("DBP_0",
                                  meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|item_level",
                                  from = LABEL,
                                  to = GROUP_VAR_OBSERVER)
    r <-
      acc_margins(resp_vars = "DBP_0",
                  study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                  meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
                  group_vars = group_vars,
                  label_col = LABEL,
                  co_vars = "AGE_0",
                  sort_group_var_levels = FALSE)

    expect_doppelganger2("margins dbp0", r$SummaryPlot)
    r1 <- r

    prep_purge_data_frame_cache()
    skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
    prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
    md <- prep_get_data_frame("item_level")

    # tweak metadata to enable the test
    md$SCALE_LEVEL[md$LABEL == "EDUCATION_0"] <- SCALE_LEVELS$INTERVAL

    r <-
      acc_margins(resp_vars = "EDUCATION_0",
                  study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                  meta_data = md,
                  group_vars = group_vars,
                  label_col = LABEL,
                  co_vars = "AGE_0")

    expect_doppelganger2("margins edu0", r$SummaryPlot)
    skip_on_covr()
    expect_lt(lobstr::obj_size(r1$SummaryPlot), 15 * 1024 * 1024)
    expect_lt(lobstr::obj_size(r$SummaryPlot), 15 * 1024 * 1024)
  })
})

test_that("multivariate outlier plot works", { # acc_multivariate_outlier.R ----
  skip_on_cran()

  skip_if_not_installed("vdiffr")
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)

  withr::with_seed(32245253, {
    r <-
      acc_multivariate_outlier(
        variable_group = c("DBP_0", "SBP_0", "AGE_0"),
        label_col = LABEL,
        study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
        meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
        scale = FALSE)

    expect_doppelganger2("acc_multivariate_outlier test 1",
                                r$SummaryPlot)
    # skip_if(TRUE || R.version$major == "4" && R.version$minor == "5.1", message = "Strange size differences in lobstr on test containers with many, but not all R versions for this plot. No idea, but likely related with the FIXME at the end of this file.") # FIXME
    skip_if_not_installed("lobstr")
    skip_on_covr()
    expect_lt(lobstr::obj_size(r$SummaryPlot), 15 * 1024 * 1024)

  })
})


test_that("shape or scale plot works", { # acc_shape_or_scale.R -----
  skip_on_cran()
  skip_if_not_installed("lobstr")

  skip_if_not_installed("vdiffr")
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)
  ({
    r <-
      acc_shape_or_scale(
        resp_vars = "DBP_0",
        study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
        meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
        label_col = LABEL)

    expect_doppelganger2("shape or scale dbp0", r$SummaryPlot)
    expect_lt(lobstr::obj_size(r$SummaryPlot), 15 * 1024 * 1024)

  })
})

test_that("univariate outlier plot works", { # acc_univariate_outlier.R ----
  skip_on_cran()
  skip_if_not_installed("lobstr")
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  skip_if_not_installed("vdiffr")
  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)
  withr::with_seed(32245253, {
    r <-
      acc_univariate_outlier(
        resp_vars = c("DBP_0", "DEV_NO_0"),
        label_col = LABEL,
        study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
        meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")

    expect_doppelganger2("acc_univariate_outlier.R DBP_0",
                                r$SummaryPlotList$DBP_0)
    expect_lt(lobstr::obj_size(r$SummaryPlotList$DBP_0), 15 * 1024 * 1024)

    # Argument “resp_vars”: Variable 'DEV_NO_0' (nominal) does not have an
    # allowed scale level (interval | ratio)
    # In “resp_vars”, variables “DEV_NO_0” were excluded.
    expect_null(r$SummaryPlotList$DEV_NO_0)

  })
})

test_that("old contradiction plots work", { # con_contradictions.R ----
  skip_on_cran()
  skip_if_not_installed("lobstr")
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("vdiffr")

  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)

  ({
    check_table <- read.csv(
        "https://dataquality.qihs.uni-greifswald.de/extdata/contradiction_checks.csv",
      header = TRUE, sep = "#"
    )
    check_table[1, "tag"] <- "Logical"
    check_table[1, "Label"] <- "Becomes younger"
    check_table[2, "tag"] <- "Empirical"
    check_table[2, "Label"] <- "sex transformation"
    check_table[3, "tag"] <- "Empirical"
    check_table[3, "Label"] <- "looses academic degree"
    check_table[4, "tag"] <- "Logical"
    check_table[4, "Label"] <- "vegetarian eats meat"
    check_table[5, "tag"] <- "Logical"
    check_table[5, "Label"] <- "vegan eats meat"
    check_table[6, "tag"] <- "Empirical"
    check_table[6, "Label"] <- "non-veg* eats meat"
    check_table[7, "tag"] <- "Empirical"
    check_table[7, "Label"] <- "Non-smoker buys cigarettes"
    check_table[8, "tag"] <- "Empirical"
    check_table[8, "Label"] <- "Smoker always scrounges"
    check_table[9, "tag"] <- "Logical"
    check_table[9, "Label"] <- "Cuff didn't fit arm"
    check_table[10, "tag"] <- "Empirical"
    check_table[10, "Label"] <- "Very mature pregnant woman"
    check_table[1, "tag"] <- "Logical, Age-Related"
    check_table[10, "tag"] <- "Empirical, Age-Related"
    label_col <- "LABEL"
    threshold_value <- 1
    study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
    meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
    meta_data <- prep_scalelevel_from_data_and_metadata(meta_data = meta_data,
                                                        study_data = study_data)
    meta_data[startsWith(meta_data[[LABEL]], "EDUCATION_"), SCALE_LEVEL] <-
      SCALE_LEVELS$ORDINAL
    r <-
      con_contradictions(
        study_data = study_data, meta_data = meta_data, label_col =
          label_col,
        threshold_value = threshold_value, check_table = check_table,
        summarize_categories = TRUE
      )
    expect_doppelganger2("con_contradictions by tag",
                                r$SummaryPlot)
    expect_lt(lobstr::obj_size(r$SummaryPlot), 15 * 1024 * 1024)

    expect_doppelganger2("con_contradictions logical age checks, only",
                                r$`Logical, Age-Related`$SummaryPlot)
    expect_lt(lobstr::obj_size(r$`Logical, Age-Related`$SummaryPlot),
              15 * 1024 * 1024)

    expect_doppelganger2("con_contradictions all checks",
                                r$all_checks$SummaryPlot)
    expect_lt(lobstr::obj_size(r$all_checks$SummaryPlot),
              15 * 1024 * 1024)


  })
})

test_that("redcap based contradiction plots work", { # con_contradictions_redcap.R ----
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("lobstr")
  skip_if_not_installed("vdiffr")

  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)

  ({
    label_col <- "LABEL"
    threshold_value <- 1
    r <- con_contradictions_redcap(
      study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
      meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData", label_col = label_col,
      threshold_value = threshold_value, meta_data_cross_item =
        "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|cross-item_level",
      summarize_categories = TRUE
    )

    expect_doppelganger2("con_contradictions rc by tag",
                                r$SummaryPlot)
    expect_lt(lobstr::obj_size(r$SummaryPlot), 15 * 1024 * 1024)

    expect_doppelganger2("con_contradictions rc logical checks, only",
                                r$Other$LOGICAL$SummaryPlot)
    expect_lt(lobstr::obj_size(r$Other$LOGICAL$SummaryPlot), 15 * 1024 * 1024)

    expect_doppelganger2("con_contradictions rc all checks",
                                r$Other$all_checks$SummaryPlot)
    expect_lt(lobstr::obj_size(r$Other$all_checks$SummaryPlot),
              15 * 1024 * 1024)

  })
})

test_that("limit deviation plots work", { # con_limit_deviations.R ----
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("lobstr")
  skip_if_not_installed("vdiffr")

  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)
  set.seed(2012) # randomly scattered points should stay in their position for testing

  ({
    label_col <- "LABEL"
    threshold_value <- 1
    meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
    meta_data[meta_data$LABEL == "QUEST_DT_0", "HARD_LIMITS"] <-
      "[2018-01-01 00:00:00 CET; 2022-12-12 23:59:59 CET)"

    r1 <- con_limit_deviations( # all limits exist
      resp_vars = c("SBP_0", "ITEM_1_0", "QUEST_DT_0"),
      study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", meta_data = meta_data,
      label_col = label_col,
      limits = "HARD"
    )

    meta_data[meta_data$LABEL == "SBP_0", "HARD_LIMITS"] <-
      "[80; )"
    meta_data[meta_data$LABEL == "ITEM_1_0", "HARD_LIMITS"] <-
      "[0; )"
    meta_data[meta_data$LABEL == "QUEST_DT_0", "HARD_LIMITS"] <-
      "[2018-01-01 00:00:00 CET; )"

    r2 <- con_limit_deviations( # lower limits exist
      resp_vars = c("SBP_0", "ITEM_1_0", "QUEST_DT_0"),
      study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", meta_data = meta_data,
      label_col = label_col,
      limits = "HARD"
    )

    meta_data[meta_data$LABEL == "SBP_0", "HARD_LIMITS"] <-
      "(; 180]"
    meta_data[meta_data$LABEL == "ITEM_1_0", "HARD_LIMITS"] <-
      "(; 10]"
    meta_data[meta_data$LABEL == "QUEST_DT_0", "HARD_LIMITS"] <-
      "(; 2022-12-12 23:59:59 CET]"

    r3 <- con_limit_deviations( # upper limits exist
      resp_vars = c("SBP_0", "ITEM_1_0", "QUEST_DT_0"),
      study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
      meta_data = meta_data,
      label_col = label_col,
      limits = "HARD"
    )

    # meta_data[meta_data$LABEL == "SBP_0", "HARD_LIMITS"] <-
    #   NA_character_
    # meta_data[meta_data$LABEL == "ITEM_1_0", "HARD_LIMITS"] <-
    #   NA_character_
    # meta_data[meta_data$LABEL == "QUEST_DT_0", "HARD_LIMITS"] <-
    #   NA_character_
    #
    # r4 <- con_limit_deviations( # no limits exist
    #   resp_vars = c("SBP_0", "ITEM_1_0", "QUEST_DT_0"),
    #   study_data = "study_data", meta_data = meta_data,
    #   label_col = label_col,
    #   limits = "HARD"
    # )

    # > 20 < 20 obs, w/ and w/o limits (upper/lower), date vars .//. others

    expect_doppelganger2("con_limit_deviations all sbp_0",
                                r1$SummaryPlotList$SBP_0)
    expect_lt(lobstr::obj_size(r1$SummaryPlotList$SBP_0), 15 * 1024 * 1024)

    expect_doppelganger2("con_limit_deviations all item",
                                r1$SummaryPlotList$ITEM_1_0)
    expect_lt(lobstr::obj_size(r1$SummaryPlotList$ITEM_1_0), 15 * 1024 * 1024)

    expect_doppelganger2("con_limit_deviations all quest_dt",
                                r1$SummaryPlotList$QUEST_DT_0)
    expect_lt(lobstr::obj_size(r1$SummaryPlotList$QUEST_DT_0), 15 * 1024 * 1024)

    expect_doppelganger2("con_limit_deviations low sbp_0",
                                r2$SummaryPlotList$SBP_0)
    expect_lt(lobstr::obj_size(r2$SummaryPlotList$SBP_0), 15 * 1024 * 1024)

    expect_doppelganger2("con_limit_deviations low item",
                                r2$SummaryPlotList$ITEM_1_0)
    expect_lt(lobstr::obj_size(r2$SummaryPlotList$ITEM_1_0), 15 * 1024 * 1024)

    expect_doppelganger2("con_limit_deviations low quest_dt",
                                r2$SummaryPlotList$QUEST_DT_0)
    expect_lt(lobstr::obj_size(r2$SummaryPlotList$QUEST_DT_0), 15 * 1024 * 1024)

    expect_doppelganger2("con_limit_deviations upp sbp_0",
                                r3$SummaryPlotList$SBP_0)
    expect_lt(lobstr::obj_size(r3$SummaryPlotList$SBP_0), 15 * 1024 * 1024)

    expect_doppelganger2("con_limit_deviations upp item",
                                r3$SummaryPlotList$ITEM_1_0)
    expect_lt(lobstr::obj_size(r3$SummaryPlotList$ITEM_1_0), 15 * 1024 * 1024)

    expect_doppelganger2("con_limit_deviations upp quest_dt",
                                r3$SummaryPlotList$QUEST_DT_0)
    expect_lt(lobstr::obj_size(r3$SummaryPlotList$QUEST_DT_0), 15 * 1024 * 1024)


  })
})

test_that("data type matrix and print.ReportSummaryTable plots work", {
  # int_datatype_matrix.R and print.ReportSummaryTable.R ----
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("lobstr")
  skip_if_not_installed("vdiffr")

  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)

  ({
    label_col <- "LABEL"

    r <- int_datatype_matrix(
      resp_vars = c("SBP_0", "ITEM_1_0", "QUEST_DT_0"),
      study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
      meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
      label_col = label_col,
      split_segments = TRUE
    )

    expect_doppelganger2("int_datatype_matrix",
                                r$SummaryPlot)
    expect_lt(lobstr::obj_size(r$SummaryPlot), 15 * 1024 * 1024)

    expect_doppelganger2("int_datatype_matrix segment v50000",
                                r$DataTypePlotList$PART_QUESTIONNAIRE)
    expect_lt(lobstr::obj_size(r$DataTypePlotList$PART_QUESTIONNAIRE),
              15 * 1024 * 1024)

    expect_doppelganger2("int_datatype_matrix ReportSummaryTable",
                                print(r$ReportSummaryTable,
                                      view = FALSE,
                                      dt = FALSE))
    expect_lt(lobstr::obj_size(print(r$ReportSummaryTable,
                                     view = FALSE,
                                     dt = FALSE)),
              15 * 1024 * 1024)

    e <- structure(list(Variables = "CENTER_0", N = 2940L), # continuous vs not continuous; fully empty
              row.names = c(NA,  -1L),
              class = c("ReportSummaryTable", "data.frame"))

    expect_doppelganger2("Empty ReportSummaryTable",
                                print(e,
                                      view = FALSE,
                                      dt = FALSE))
    expect_lt(lobstr::obj_size(print(e,
                                     view = FALSE,
                                     dt = FALSE)),
              15 * 1024 * 1024)

    cont <- data.frame(Variables = letters[1:10L], N = 2940L,
                       a = sin(1:10L),
                       check.names = FALSE,
                       stringsAsFactors = FALSE)

    class(cont) <- c("ReportSummaryTable", "data.frame")

    expect_doppelganger2("ReportSummaryTable cont",
                                print(cont,
                                      view = FALSE,
                                      dt = FALSE))
    expect_lt(lobstr::obj_size(print(cont,
                                     view = FALSE,
                                     dt = FALSE)),
              15 * 1024 * 1024)
  })
})

test_that("pro-applicability matrix plots work", { # pro_applicability_matrix.R ----
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("lobstr")
  skip_if_not_installed("vdiffr")

  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)

  ({

    r <- pro_applicability_matrix(
      study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
      meta_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
      split_segments = FALSE,
      label_col = LABEL,
      meta_data_segment = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|segment_level",
      meta_data_dataframe = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|dataframe_level")

    expect_doppelganger2("pro_applicability_matrix",
                                r$ApplicabilityPlot)
    expect_lt(lobstr::obj_size(r$ApplicabilityPlot), 15 * 1024 * 1024)

    expect_doppelganger2("pro_applicability_matrix segment v50000",
                                r$ApplicabilityPlotList$PART_QUESTIONNAIRE)
    expect_lt(lobstr::obj_size(r$ApplicabilityPlotList$PART_QUESTIONNAIRE),
                               15 * 1024 * 1024)

    expect_doppelganger2("pro_applicability_matrix ReportSummaryTable",
                                print(r$ReportSummaryTable,
                                      view = FALSE,
                                      dt = FALSE))
    expect_lt(lobstr::obj_size(print(r$ReportSummaryTable,
                                     view = FALSE,
                                     dt = FALSE)), 15 * 1024 * 1024)

  })
})

test_that("heatmap with 1 threshold plots work", { # util_heatmap_1th.R ----
  skip_on_cran()
  skip_if_not_installed("lobstr")
  skip_if_not_installed("vdiffr")


  withr::local_options(dataquieR.CONDITIONS_LEVEL_TRHESHOLD = Inf)

  ({

    x <- iris
    x$a <- as.integer(x$Species)
    x$Species2 <- rev(x$Species)
    x$b <- as.integer(x$Species2)
    x$c <- x$a + x$b
    x$str <- round(x$Sepal.Width, 0)

    p1 <- util_heatmap_1th(df = x,
                           cat_vars = c("Species", "Species2"),
                           values = "c",
                           threshold = 0,
                           invert = FALSE)

    p2 <- util_heatmap_1th(df = x,
                           cat_vars = c("Species", "Species2"),
                           strata = "str",
                           values = "c",
                           threshold = 0,
                           invert = FALSE)

    p3 <- util_heatmap_1th(df = x,
                           cat_vars = "Species",
                           values = "a",
                           threshold = 0,
                           invert = FALSE)

    expect_doppelganger2("heatmap with 2 cat-vars but w/o strata vars",
                                p1)
    # FIXME: Enable after https://github.com/r-lib/lobstr/issues/48
    # expect_lt(lobstr::obj_size(p1), 15 * 1024 * 1024)

    expect_doppelganger2("heatmap with 2 cat-vars but w/ strata vars",
                                p2)
    # FIXME: Enable after https://github.com/r-lib/lobstr/issues/48
    # expect_lt(lobstr::obj_size(p2), 15 * 1024 * 1024)

    expect_doppelganger2("heatmap with 1 cat-var",
                                p3)
    # FIXME: Enable after https://github.com/r-lib/lobstr/issues/48
    # expect_lt(lobstr::obj_size(p3), 15 * 1024 * 1024)
  })

  # FIXME: @EK: Please add more plot tests for all flavors of plots, namely, distribution, margins, maybe loess (if some path through the function is missing)
})

test_that("con_contradictions_redcap works", {
  skip_on_cran() # slow, redcap parser is tested anyway, errors in plots obvious

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]
  meta_data_cross_item <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|cross-item_level")
  # ignore here 'special' data preparation steps, will be checked separately
  meta_data_cross_item$DATA_PREPARATION <- ""
  label_col <- "LABEL"
  threshold_value <- 1
  #expect_message2(
    expect_silent({
      default <- con_contradictions_redcap(
        study_data = study_data, meta_data = meta_data, label_col = label_col,
        threshold_value = threshold_value, meta_data_cross_item = meta_data_cross_item
      )
      off <- con_contradictions_redcap(
        study_data = study_data, meta_data = meta_data, label_col = label_col,
        threshold_value = threshold_value, meta_data_cross_item = meta_data_cross_item,
        summarize_categories = FALSE
      )
      on <- con_contradictions_redcap(
        study_data = study_data, meta_data = meta_data, label_col = label_col,
        threshold_value = threshold_value, meta_data_cross_item = meta_data_cross_item,
        summarize_categories = TRUE
      )
    })
  #)
  expect_equal(off$FlaggedStudyData, on$Other$all_checks$FlaggedStudyData)
  expect_equal(off$VariableGroupTable, on$Other$all_checks$VariableGroupTable)

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("summary contra_r pl1 ok",
                              on$Other$all_checks$SummaryPlot)
  expect_doppelganger2("summary contra_r pl2 ok",
                              default$SummaryPlot)
  expect_doppelganger2("summary contra_r pl3 ok",
                              off$SummaryPlot)
  expect_doppelganger2("one cat contra_r pl4 ok",
                              on$Other$EMPIRICAL$SummaryPlot)
})

test_that("con_contradictions_redcap works with tiny inputs", {
  skip_on_cran() # slow, redcap parser is tested anyway, errors in plots obvious

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  # catch if some objects will be reduced to scalars or vectors instead of dataframes or matrices
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]
  threshold_value <- 1

  meta_data_cross_item <- data.frame("CONTRADICTION_TERM" = "[AGE_0] < 18",
                            "CHECK_LABEL" = "Age hard limits check for testing",
                            "CONTRADICTION_TYPE" = "LOGICAL")
  meta_data_cross_item2 <- data.frame("CONTRADICTION_TERM" = "[v00003] < 18",
                             "CHECK_LABEL" = "Age hard limits check for testing",
                             "CONTRADICTION_TYPE" = "LOGICAL")
  meta_data_cross_item3 <- data.frame("CONTRADICTION_TERM" = c("[v00003] < 18", "[v00003] > 130"),
                             "CHECK_LABEL" = c("Age hard limits check for testing", "Age hard limits check for testing no. 2"),
                             "CONTRADICTION_TYPE" = c("LOGICAL", "EMPIRICAL"))
  tiny_sd <- study_data[, which(colnames(study_data) == meta_data$VAR_NAMES[which(meta_data$LABEL == "AGE_0")]), drop = FALSE]
  tiny_md <- meta_data[which(meta_data$LABEL == "AGE_0"), , drop = FALSE]
  tiny_md[[JUMP_LIST]][util_empty(tiny_md[[JUMP_LIST]])] <- SPLIT_CHAR
  tiny_md[[MISSING_LIST]][util_empty(tiny_md[[MISSING_LIST]])] <- SPLIT_CHAR
  #expect_message2(
  expect_message2({
    check1 <- con_contradictions_redcap( # only one contradiction check -> nrow(meta_data_cross_item) is 1
      study_data = tiny_sd, meta_data = tiny_md, label_col = "LABEL", # using VAR_NAMES and LABELs => two "needles" to generate the variable list
      threshold_value = threshold_value, meta_data_cross_item = meta_data_cross_item)
    check2 <- con_contradictions_redcap(
      study_data = tiny_sd, meta_data = tiny_md, label_col = "VAR_NAMES", # only one "needle" to generate the variable list
      threshold_value = threshold_value, meta_data_cross_item = meta_data_cross_item2) # only one contradiction check -> nrow(meta_data_cross_item2) is 1
    check3 <- con_contradictions_redcap(
      study_data = tiny_sd, meta_data = tiny_md, label_col = "VAR_NAMES", # only one "needle" to generate the variable list
      threshold_value = threshold_value, meta_data_cross_item = meta_data_cross_item3) # nrow(meta_data_cross_item3) is 2
  })
  #)
})

test_that("con_contradictions_redcap uses DATA_PREPARATION correctly", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]
  meta_data_cross_item <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|cross-item_level")

  # If nothing is specified in DATA_PREPARATION, missing value labels and values
  # outside hard limits should be replaced by NA.
  mdci <- meta_data_cross_item[1, ]
  res1 <- con_contradictions_redcap(study_data = study_data,
                                    meta_data = meta_data,
                                    label_col = LABEL,
                                    threshold_value = 1,
                                    meta_data_cross_item = mdci)
  mdci$DATA_PREPARATION <- "LIMITS | MISSING_NA"
  res2 <- con_contradictions_redcap(study_data = study_data,
                                    meta_data = meta_data,
                                    label_col = LABEL,
                                    threshold_value = 1,
                                    meta_data_cross_item = mdci)
  expect_equal(res1$FlaggedStudyData, res2$FlaggedStudyData)

  # check option MISSING_INTERPRET
  mdci$CHECK_LABEL <- "Inconsistent reason for missingness in number of children"
  mdci$CONTRADICTION_TERM <- "[N_BIRTH_0] > 0 and ([N_CHILD_0] in set('NE', 'P', 'NC'))"
  mdci$CONTRADICTION_TYPE <- "EMPIRICAL"
  mdci$DATA_PREPARATION <- "MISSING_INTERPRET"
  # We would expect these observations to be picked up:
  miss_tab <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|missing_table")
  sel_miss_codes <- miss_tab$CODE_VALUE[which(miss_tab$CODE_INTERPRET %in%
                                                c("NE", "NC", "P"))]
  check_res <- sum(table(study_data[which(study_data$v00021 %in% sel_miss_codes
                                          & study_data$v00027 > 0
                                          & study_data$v00027 < 8000),
                                    c("v00021", "v00027")]))
  # NE = 99981, NC = 99983, P = 99988
  #expect_equal(res3$VariableGroupTable$NUM_con_con, check_res)
  # but numerical variables are not yet supported here
  expect_warning({
    res3 <- con_contradictions_redcap(study_data = study_data,
                                    meta_data = meta_data,
                                    label_col = LABEL,
                                    threshold_value = 1,
                                    meta_data_cross_item = mdci)},
    regexp = "not yet supported for numerical variables")

  # check that the function catches problems with variables on smoking
  # (replacing missing value codes by NA not specified, but should be done;
  # inadmissible categorical values on top;
  # 91 observations match the contradiction rule)
  mdci <- meta_data_cross_item[grepl("Non-smokers inconsistency",
                                     meta_data_cross_item$CHECK_LABEL), ,
                               drop = FALSE]
  expect_message2({
    res4 <- con_contradictions_redcap(study_data = study_data,
                                    meta_data = meta_data,
                                    label_col = LABEL,
                                    threshold_value = 1,
                                    meta_data_cross_item = mdci)},
    regexp = sprintf("(%s|%s)",
      "replace the missing codes by NA, too",
      "Number of levels in variable greater than in character string"))
  expect_equal(res4$VariableGroupTable$NUM_con_con, 91)
})

#Temporarily skipped due to ship moving to website
test_that("no regression, rule errors should not be missed", {
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  sd1 <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS", keep_types = TRUE)
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx")
  md1 <- prep_get_data_frame("item_level")
  checks <- prep_get_data_frame("cross-item_level")
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = sd1,
                                           meta_data = md1)
  md1[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      md1[[VAR_NAMES]]
    ]

  checks[3,3] <- "[BODY_HEIGHT_0] < [OBS_SOMA_0]"
  checks[4,3] <- "[BODY_HEIGHT_0] < [EXAM_DT_0]"
  checks[4,3] <- "[BODY_HEIGHT] < 1.2"
  checks[5,3] <- "[SEX] < 3"
  checks <- checks[-c(1:2), ]
  checks <- checks[-c(4:10), ]

  suppressMessages(suppressWarnings(expect_warning(
    AnyContradictions <- con_contradictions_redcap(study_data = sd1,
                                                   meta_data       = md1,
                                                   label_col       = "LABEL",
                                                   meta_data_cross_item = checks,
                                                   threshold_value = 1),
    regexp = "object.+SEX.+not found"
  )))

  expect_equal(AnyContradictions$VariableGroupTable$NUM_con_con,
               c(2152, NA_real_, NA_real_))
  # the first test is by default comparing lexicographically, since obs_soma is a factor and LABEL is default for DATA_PREPARATION
  # the other two tests always fail because of missing variables
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  sd1 <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS", keep_types = TRUE)
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx")
  md1 <- prep_get_data_frame("item_level")
  checks <- prep_get_data_frame("cross-item_level")

  suppressWarnings(suppressMessages(
    AnyContradictions <- con_contradictions_redcap(study_data = sd1,
                                                   meta_data       = md1,
                                                   label_col       = "LABEL",
                                                   meta_data_cross_item = checks,
                                                   threshold_value = 1)))

  expect_equal(AnyContradictions$VariableGroupTable$NUM_con_con, c(35,
                                                             0,
                                                             0,
                                                             63,
                                                             12,
                                                             0,
                                                             0))
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  sd1 <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS", keep_types = TRUE)
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx")
  md1 <- prep_get_data_frame("item_level")
  checks <- prep_get_data_frame("cross-item_level")

  checks[1,3] <- "[sbp] < [dbp]" # variables not in dataset/metadata
  checks[2,3] <- "sbp2 < dbp2"   # omit brackets
  checks[6,3] <-
    "[diab_known] = \"yes\" NOT [diab_age] > 0" # uses variable names instead of labels
  checks <- checks[-c(3:5), ]
  checks <- checks[-c(4:9), ]


  suppressMessages(suppressWarnings(expect_warning({
    AnyContradictions <- con_contradictions_redcap(study_data = sd1,
                                                   meta_data       = md1,
                                                   label_col       = "LABEL",
                                                   meta_data_cross_item = checks,
                                                   threshold_value = 1)
    }, regexp = "Parser error"
    )))

  expect_true(all(is.na(AnyContradictions$VariableGroupTable$NUM_con_con)))

})

test_that("posix offset is 1970-01-01 01:00:00 CET)", {
  skip_on_cran()

  require_english_locale_and_berlin_tz()
  expect_equal(as.POSIXct(as.numeric(as.POSIXct("1975-12-16 03:46:00 CET")),
                          origin = min(as.POSIXct(Sys.Date()), 0)),
               as.POSIXct("1975-12-16 03:46:00 CET"))
})
test_that("util_interpret_limits works", {

  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta <- prep_create_meta(
    VAR_NAMES = 1:26,
    DATA_TYPE = c(rep(DATA_TYPES$INTEGER, 13), rep(DATA_TYPES$FLOAT, 9),
                  rep(DATA_TYPES$DATETIME, 4)),
    LABEL = LETTERS,
    MISSING_LIST = ""
  )
  expect_error(util_interpret_limits(meta), regexp =
                 "No column containing the term LIMIT")
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]] <- NA
  expect_message2(util_interpret_limits(meta), regexp =
                   "HARD_LIMITS has no defined intervals and is omitted.")
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]] <- "xx"
  expect_warning(util_interpret_limits(meta), regexp =
                   "Found invalid limits for .HARD_LIMITS.: .* will ignore these",
                 perl = TRUE)
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]] <- NA
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][1] <- "[0; 10)"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][2] <- "[0;Inf)"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][3] <- "(0; 10)"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][4] <- "[0; 10]"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][5] <- "[-Inf; 0]"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][6] <- "(-Inf; Inf]"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][7] <- "(0.1; 13.324]"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][8] <- "(0.1; 13324.0]"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][23] <- "(2020-01-01 00:00:00 CET; Inf]"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][24] <- "(+Inf; 2019-09-09 00:00:00 CET]"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][25] <- "(2020-01-01 00:00:00 CET; ]"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][26] <- "(+Inf; ]"
  m2 <- util_interpret_limits(meta)
  a <- m2[, c("HARD_LIMIT_LOW", "HARD_LIMIT_UP", "INCL_HARD_LIMIT_LOW",
              "INCL_HARD_LIMIT_UP")]
  b <- dplyr::tribble(
    ~HARD_LIMIT_LOW, ~HARD_LIMIT_UP, ~INCL_HARD_LIMIT_LOW, ~INCL_HARD_LIMIT_UP,
    0, 10, TRUE, FALSE,
    0, Inf, TRUE, FALSE,
    0, 10, FALSE, FALSE,
    0, 10, TRUE, TRUE,
    -Inf, 0, TRUE, TRUE,
    -Inf, Inf, FALSE, TRUE,
    0.1, 13.324, FALSE, TRUE,
    0.1, 13324, FALSE, TRUE,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    NA, NA, NA, NA,
    1577833200, Inf, FALSE, TRUE,
    Inf, 1567980000, FALSE, TRUE,
    1577833200, NA, FALSE, TRUE,
    Inf, NA, FALSE, TRUE,
  )
  expect_equal(a, b, tolerance = 1e-3, ignore_attr = TRUE)

  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]] <- NA
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][1] <- "[0; 0-0)"
  meta[[WELL_KNOWN_META_VARIABLE_NAMES$HARD_LIMITS]][2] <- "[3-3;Inf)"
  wm <- conditionMessage(capture_warning(m3 <- util_interpret_limits(meta)))

  expect_match(wm,
                 regexp =
                   paste0("Damaged (lower|upper)",
                          ".+HARD_LIMITS.+: .+(3-3|0-0).+ in .+[12].+"),
                 perl = TRUE
  )
})

test_that("con_contradictions works", {
  skip_on_cran() # slow and deprecated, use redcap rules, now
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  require_english_locale_and_berlin_tz()
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  meta_data <- prep_scalelevel_from_data_and_metadata(meta_data = meta_data,
                                                      study_data = study_data)
  meta_data[startsWith(meta_data[[LABEL]], "EDUCATION_"), SCALE_LEVEL] <-
    SCALE_LEVELS$ORDINAL
  check_table <- read.csv(
    "https://dataquality.qihs.uni-greifswald.de/extdata/contradiction_checks.csv",
    header = TRUE, sep = "#"
  )
  check_table[1, "tag"] <- "Logical"
  check_table[1, "Label"] <- "Becomes younger"
  check_table[2, "tag"] <- "Empirical"
  check_table[2, "Label"] <- "sex transformation"
  check_table[3, "tag"] <- "Empirical"
  check_table[3, "Label"] <- "looses academic degree"
  check_table[4, "tag"] <- "Logical"
  check_table[4, "Label"] <- "vegetarian eats meat"
  check_table[5, "tag"] <- "Logical"
  check_table[5, "Label"] <- "vegan eats meat"
  check_table[6, "tag"] <- "Empirical"
  check_table[6, "Label"] <- "non-veg* eats meat"
  check_table[7, "tag"] <- "Empirical"
  check_table[7, "Label"] <- "Non-smoker buys cigarettes"
  check_table[8, "tag"] <- "Empirical"
  check_table[8, "Label"] <- "Smoker always scrounges"
  check_table[9, "tag"] <- "Logical"
  check_table[9, "Label"] <- "Cuff didn't fit arm"
  check_table[10, "tag"] <- "Empirical"
  check_table[10, "Label"] <- "Very mature pregnant woman"
  label_col <- LABEL
  threshold_value <- 1
  check_table[1, "tag"] <- "Logical, Age-Related"
  check_table[10, "tag"] <- "Empirical, Age-Related"
  suppressMessages(suppressWarnings({
      default <- con_contradictions(
        study_data = study_data, meta_data = meta_data, label_col = label_col,
        threshold_value = threshold_value, check_table = check_table
      )
      off <- con_contradictions(
        study_data = study_data, meta_data = meta_data, label_col = label_col,
        threshold_value = threshold_value, check_table = check_table,
        summarize_categories = FALSE
      )
      on <- con_contradictions(
        study_data = study_data, meta_data = meta_data, label_col = label_col,
        threshold_value = threshold_value, check_table = check_table,
        summarize_categories = TRUE
      )
    }
  ))
  # expect_equal(off, default) because of plots not always true,
  # e.g.:   Component "SummaryPlot": Component "layers": Component 1:
  # Component 11: Component 1: target is not list-like
  expect_equal(off$FlaggedStudyData, on$all_checks$FlaggedStudyData)
  expect_equal(off$SummaryTable, on$all_checks$SummaryTable)
  expect_equal(off$SummaryData, on$all_checks$SummaryData)

  expect_lt(
    abs(suppressWarnings(sum(as.numeric(as.matrix(default$SummaryData)),
                             na.rm = TRUE)) - 12052.56), 10)

  expect_lt(
    abs(suppressWarnings(sum(as.numeric(as.matrix(on$Empirical$SummaryTable)),
                             na.rm = TRUE)) -  5474.32), 10)

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("summary contradictio1 plot ok",
                              on$all_checks$SummaryPlot)
  expect_doppelganger2("summary contradictio2 plot ok",
                              default$SummaryPlot)
  expect_doppelganger2("summary contradictio3 plot ok",
                              off$SummaryPlot)

  expect_doppelganger2("one cat contradiction plot ok",
                              on$Empirical$SummaryPlot)
})

test_that("util_acc_loess_bin works", {
  skip_on_cran() # slow

  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  # numeric response, contrast 'intermediate' and 'low'/'high' values,
  # without co_vars
  sd0 <- study_data
  sd0$v00003[1:10] <- NA
  sd0$v00002[11:20] <- NA
  expect_message2(
    res1 <-
      util_acc_loess_bin(resp_vars = "CRP_0", study_data = sd0,
                    meta_data = meta_data, group_vars = "DEV_NO_0",
                    time_vars = "LAB_DT_0", co_vars = NULL,
                    label_col = LABEL)
    ,
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )

  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$CRP_0)), "try-error"))

  # numeric response, contrast 'intermediate' and 'low'/'high' values,
  # with co_vars
  expect_message2(
    res2 <-
      util_acc_loess_bin(resp_vars = "CRP_0", study_data = sd0,
                    meta_data = meta_data, group_vars = "DEV_NO_0",
                    time_vars = "LAB_DT_0", co_vars = c("AGE_0", "SEX_0"),
                    label_col = LABEL)
    ,
    regexp = "Due to missing values in DEV_NO_0, AGE_0, SEX_0 or LAB_DT_0, N = 327 observations were excluded. Due to missing values in CRP_0, N = 130 observations were excluded"
  )

  expect_false(
    inherits(try(ggplot_build(res2$SummaryPlotList$CRP_0)), "try-error"))

  # nominal response, recoding to binary should be done automatically
  # no group_var
  expect_message2(
    res3 <-
      util_acc_loess_bin(resp_vars = "CENTER_0", study_data = study_data,
                    meta_data = meta_data, time_vars = "EXAM_DT_0",
                    label_col = LABEL) # plot is not helpful
  )

  expect_false(
    inherits(try(ggplot_build(res3$SummaryPlotList$CENTER_0)), "try-error"))

  # binary response, with group_var
  expect_message2(
    res4 <-
      util_acc_loess_bin(resp_vars = "ASTHMA_0", group_vars = "CENTER_0",
                    time_vars = "EXAM_DT_0",
                    study_data = study_data, meta_data = meta_data,
                    label_col = LABEL)
  )

  expect_false(
    inherits(try(ggplot_build(res4$SummaryPlotList$ASTHMA_0)), "try-error"))
})

test_that("acc_loess works without label_col and catches wrong inputs", {
  skip_on_cran() # slow

  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  skip_on_cran() # slow test
  skip_if_translated()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_warning(
    expect_error({
      res1 <-
        acc_loess(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data)
      },
      regexp = paste(".*Argument time_vars is NULL"),
      perl = TRUE
    ),
    regexp = paste("Missing argument .+time_vars.+ without default value.",
              "Setting to NULL. As a dataquieR developer"),
    perl = TRUE
  )

  expect_warning(
    expect_error({
      res1 <-
        acc_loess(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data, group_vars = "v00016")
    },
    regexp = paste(".*Argument time_vars is NULL"),
    perl = TRUE
    ),
    regexp = paste("Missing argument .+time_vars.+ without default value.",
                   "Setting to NULL. As a dataquieR developer"),
    perl = TRUE
  )

  sd1 <- study_data
  sd1[["v00017"]][1:1000] <- NA
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00014", study_data = sd1, # continuous variable
                meta_data = meta_data, group_vars = "v00016",
                time_vars = "v00017")
    ,
    regexp = "Due to missing values in v00016 or v00017, N = 1243 observations were excluded. Due to missing values in v00014, N = 82 observations were excluded"
  )

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00008", # categorical variable
                study_data = study_data,
                meta_data = meta_data,
                group_vars = "v00011",
                time_vars = "v00013")
    ,
    regexp = sprintf("(%s|%s|%s)",
                     paste("Due to missing values in v00011 or v00013,",
                           "N = 218 observations were excluded. Due to missing",
                           "values in v00008, N = 270 observations were",
                           "excluded additionally."),
                     "Recoded 1 variable.s.",
                     paste("Levels of the group_var with too few observations",
                           "were discarded .level USR_599..")
    )
  )

  sd1 <- study_data
  sd1[["v00017"]] <- as.character(sd1[["v00017"]])
  sd1[["v00017"]][1:1000] <- "2001-02-29"
  sd1[["v00013"]] <- as.character(sd1[["v00013"]])
  sd1[["v00013"]][1:1000] <- "2001-02-29"
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00014", study_data = sd1,
                meta_data = meta_data, group_vars = "v00016",
                time_vars = "v00017"), # ===> "LAB_DT_0"
    regexp = sprintf("(%s|%s)",
                     paste("Data type transformation of .v00017. introduced",
                           "1000 additional missing values"),
                     paste("Due to missing values in v00016 or v00017,",
                           "N = 1243 observations were excluded. Due to",
                           "missing values in v00014, N = 82 observations",
                           "were excluded")),
    perl = TRUE
  )
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00008", study_data = sd1,
                meta_data = meta_data, group_vars = "v00011",
                time_vars = "v00013")
    ,
    regexp = sprintf("(%s|%s|%s|%s)",
                     paste("Data type transformation of .v00013. introduced",
                           "1000 additional missing values"),
                     paste("Due to missing values in v00011 or v00013,",
                           "N = 1150 observations were excluded. Due to missing",
                           "values in v00008, N = 211 observations were",
                           "excluded additionally."),
                     "Recoded 1 variable.s.",
                     paste("Levels of the group_var with too few observations",
                           "were discarded .level USR_599..")
    )
  )

  expect_error(
    res1 <-
      acc_loess(resp_vars = "v00014", study_data = study_data,
                meta_data = meta_data, group_vars = "v00016",
                min_obs_in_subgroup = 1:2,
                time_vars = "v00017") # ===> "LAB_DT_0"
    ,
    regexp =
      paste("Need exactly one element in argument min_obs_in_subgroup, got 2"),
    perl = TRUE
  )

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00014", study_data = study_data,
                meta_data = meta_data, group_vars = "v00016",
                resolution = Inf,
                min_obs_in_subgroup = 30,
                time_vars = "v00017") # ===> "LAB_DT_0"
    ,
    regexp = sprintf("(%s|%s)",
                     paste("Argument resolution is not specified correctly and is set to 80 instead"),
                     paste("Due to missing values in v00016 or v00017, N = 308 observations were excluded. Due to missing values in v00014, N = 131 observations were excluded")),
    perl = TRUE
  )

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00014", study_data = study_data,
                meta_data = meta_data, group_vars = "v00016",
                resolution = "12",
                min_obs_in_subgroup = 30,
                time_vars = "v00017") # ===> "LAB_DT_0"
    ,
    regexp = "Due to missing values in v00016 or v00017, N = 308 observations were excluded. Due to missing values in v00014, N = 131 observations were excluded"
  )

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00014", study_data = study_data,
                meta_data = meta_data, group_vars = "v00016",
                resolution = complex(imaginary = 12),
                min_obs_in_subgroup = 30,
                time_vars = "v00017") # ===> "LAB_DT_0"
    ,
    regexp = sprintf("(%s|%s)",
                     "Argument resolution is not specified correctly and is set to 80 instead",
                     "Due to missing values in v00016 or v00017, N = 308 observations were excluded. Due to missing values in v00014, N = 131 observations were excluded")
  )

  suppressWarnings(
    expect_error(
      res1 <-
        acc_loess(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data, group_vars = "v00016",
                  mark_time_points = "x",
                  min_obs_in_subgroup = 30,
                  time_vars = "v00017") # ===> "LAB_DT_0"
      ,
      regexp =
        paste("Argument mark_time_points must be logical"),
      perl = TRUE
    )
  )

  suppressWarnings(
    expect_error(
      res1 <-
        acc_loess(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data, group_vars = "v00016",
                  mark_time_points = TRUE,
                  comparison_lines = 42,
                  min_obs_in_subgroup = 30,
                  time_vars = "v00017") # ===> "LAB_DT_0"
      ,
      regexp =
        paste(".+comparison_lines.+ needs to be a list of arguments",
              "as specified in the documentation."),
      perl = TRUE
    )
  )

  sd0 <- study_data
  sd0$v00017 <- "XXXX"
  suppressMessages(
    expect_error(
      res1 <-
        acc_loess(resp_vars = "v00014", study_data = sd0,
                  meta_data = meta_data, group_vars = "v00016",
                  mark_time_points = TRUE,
                  min_obs_in_subgroup = 30,
                  time_vars = "v00017") # ===> "LAB_DT_0"
      ,
      regexp = "Variable .v00017.+time_vars. has only NA observations",
      perl = TRUE
    )
  )

  suppressWarnings(
    expect_error(
      res1 <-
        acc_loess(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data, group_vars = "v00016",
                  mark_time_points = "TRUE",
                  min_obs_in_subgroup = 30,
                  time_vars = "v00017") # ===> "LAB_DT_0"
      , regexp = "Argument mark_time_points must be logical"
    )
  )

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00014", study_data = study_data,
                meta_data = meta_data, group_vars = "v00016",
                min_obs_in_subgroup = NA,
                time_vars = "v00017") # ===> "LAB_DT_0"
  )

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00014", study_data = study_data,
                meta_data = meta_data, group_vars = "v00016",
                min_obs_in_subgroup = "x",
                time_vars = "v00017") # ===> "LAB_DT_0"
    ,
    regexp = sprintf("(%s|%s)",
                     "Argument resolution is not specified correctly and is set to 80 instead",
                     "Due to missing values in v00016 or v00017, N = 308 observations were excluded. Due to missing values in v00014, N = 131 observations were excluded")
  )

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00014", study_data = study_data,
                meta_data = meta_data, group_vars = "v00016",
                time_vars = "v00017", plot_format = "BOTH") # ===> "LAB_DT_0"
    ,
    regexp = "Due to missing values in v00016 or v00017, N = 308 observations were excluded. Due to missing values in v00014, N = 131 observations were excluded"
  )

  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    2
  )
  expect_lt(
    suppressWarnings(abs(mean(as.numeric(
      as.matrix(res1$SummaryPlotList$Loess_fits_combined$data)),
      na.rm = TRUE) - 21.82126)), 50
  )
})

test_that("acc_loess works with label_col", {
  skip_on_cran() # slow

  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_warning(
    expect_error({
      res1 <-
        acc_loess(resp_vars = "CRP_0", study_data = study_data,
                  meta_data = meta_data,
                  label_col = LABEL)
    },
    regexp = paste(".*Argument time_vars is NULL"),
    perl = TRUE
    ),
    regexp = paste("Missing argument .+time_vars.+ without default value.",
                   "Setting to NULL. As a dataquieR developer"),
    perl = TRUE
  )

  expect_warning(
    expect_error({
      res1 <-
        acc_loess(resp_vars = "CRP_0", study_data = study_data,
                  meta_data = meta_data, group_vars = "DEV_NO_0",
                  label_col = LABEL)
    },
    regexp = paste(".*Argument time_vars is NULL"),
    perl = TRUE
    ),
    regexp = paste("Missing argument .+time_vars.+ without default value.",
                   "Setting to NULL. As a dataquieR developer"),
    perl = TRUE
  )

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "BOTH")
    ,
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )

  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    2
  )
  skip_on_cran()
  expect_lt(
    suppressWarnings(abs(mean(as.numeric(
      as.matrix(res1$SummaryPlotList$Loess_fits_combined$data)),
      na.rm = TRUE) - 21.82126)), 50
  )

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "VO2_CAPCAT_0", # categorical variable
                study_data = study_data,
                meta_data = meta_data,
                group_vars = "USR_VO2_0",
                time_vars = "EXAM_DT_0",
                label_col = LABEL)
    ,
    regexp = sprintf("(%s|%s|%s)",
                     paste("Due to missing values in USR_VO2_0 or EXAM_DT_0,",
                           "N = 218 observations were excluded. Due to missing",
                           "values in VO2_CAPCAT_0, N = 270 observations were",
                           "excluded additionally."),
                     "Recoded 1 variable.s.",
                     paste("Levels of the group_var with too few observations",
                           "were discarded .level USR_599..")
    )
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$VO2_CAPCAT_0)), "try-error"))

  sd0 <- study_data
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_211")] <- 0 # one subgroup has constant values
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "ASTHMA_0", # categorical variable
                study_data = sd0,
                meta_data = meta_data,
                group_vars = "USR_VO2_0",
                time_vars = "EXAM_DT_0",
                label_col = LABEL)
    ,
    regexp = sprintf("(%s|%s|%s)",
                     paste("Due to missing values in USR_VO2_0 or EXAM_DT_0,",
                           "N = 218 observations were excluded. Due to missing",
                           "values in ASTHMA_0, N = 259 observations were",
                           "excluded additionally."),
                     "Recoded 1 variable.s.",
                     paste("Levels of the group_var with too few observations",
                           "were discarded .level USR_599..")
    )
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$ASTHMA_0)), "try-error"))

  md0 <- meta_data
  md0[[SCALE_LEVEL]][md0[[LABEL]] == "VO2_CAPCAT_0"] <- SCALE_LEVELS$ORDINAL
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "VO2_CAPCAT_0", # categorical variable
                study_data = study_data,
                meta_data = md0,
                group_vars = "USR_VO2_0",
                time_vars = "EXAM_DT_0",
                label_col = LABEL)
    ,
    regexp = sprintf("(%s|%s|%s)",
                     paste("Due to missing values in USR_VO2_0 or EXAM_DT_0,",
                           "N = 218 observations were excluded. Due to missing",
                           "values in VO2_CAPCAT_0, N = 270 observations were",
                           "excluded additionally."),
                     "Recoded 1 variable.s.",
                     paste("Levels of the group_var with too few observations",
                           "were discarded .level USR_599..")
    )
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$VO2_CAPCAT_0)), "try-error"))

  sd0 <- study_data
  sd0[["v00014"]][which(sd0[["v00016"]] == 4)] <- 2.5 # one subgroup has constant values
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0",
                study_data = sd0,
                meta_data = meta_data,
                group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL)
    ,
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 121 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$CRP_0)), "try-error"))
})

test_that("acc_loess output matches", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "BOTH")
    ,
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    2
  )
  skip_on_cran()
  # TODO: skip_if_not(capabilities()["long.double"])
  # skip_on_travis() # vdiffr fails
  skip_if_not_installed("vdiffr")
  expect_doppelganger2("loess facets plot for CRP_0 ok",
                              res1$SummaryPlotList$Loess_fits_facets)
  expect_doppelganger2("loess combined plot for CRP_0 ok",
                              res1$SummaryPlotList$Loess_fits_combined)

  set.seed(32)
  nt <- 1000 # = nrow(study_data)/3
  p1 <- 0.3 # baseline probability
  p2 <- 0.7 # probability for the deviating examiner (at the end of the observation period or for a subset of observations)
  sd0 <- study_data
  sd0[["v00011"]] <- rep(c("USR_321", "USR_590", "USR_213"), 1000)
  set.seed(32)
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_321")] <-
    rbinom(n = nt, size = 1, prob = p1) # only noise
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_590")] <-
    rbinom(n = nt, size = 1,
           prob = (p2 - p1)/(nt - 1) * 1:nt + (p1 * nt - p2)/(nt - 1))
  # increase over time
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_213")] <-
    c(rbinom(n = floor(1/3 * nt), size = 1, prob = p1),
      rbinom(n = floor(1/3 * nt), size = 1, prob = p2),
      rbinom(n = nt - 2 * floor(1/3 * nt), size = 1, prob = p1))
  # switch between p1 and p2
  md0 <- meta_data
  md0[["RECODE_CASES"]] <- ""
  md0[["RECODE_CONTROL"]] <- ""
  md0[["RECODE_CASES"]][md0$VAR_NAMES == "v00007"] <- "yes"
  md0[["RECODE_CONTROL"]][md0$VAR_NAMES == "v00007"] <- "no"

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "ASTHMA_0", # categorical variable
                study_data = sd0,
                meta_data = md0,
                group_vars = "USR_VO2_0",
                time_vars = "EXAM_DT_0",
                label_col = LABEL)
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$ASTHMA_0)), "try-error"))
})

test_that("acc_loess min_obs_in_subgroups with label_col", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    expect_error({
      res1 <-
        acc_loess(resp_vars = "CRP_0", study_data = study_data,
                  meta_data = meta_data, group_vars = "DEV_NO_0",
                  time_vars = "LAB_DT_0",
                  label_col = LABEL, min_obs_in_subgroup = 999)
    },
    regexp = paste("No data left after data preparation."),
    perl = TRUE
    ),
    regexp =
      sprintf(
        "(%s|%s)",
        "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded.",
        paste("Levels of the group_var with too few observations",
              "were discarded .levels 1, 2, 3, 4, 5.+")
      ),
    perl = TRUE
  )
})

test_that("acc_loess with co-vars output matches", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  sd0 <- study_data
  sd0$v00003[1:10] <- NA
  sd0$v00002[11:20] <- NA
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = sd0,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0", co_vars = c("AGE_0", "SEX_0"),
                label_col = LABEL)
    ,
    regexp = "Due to missing values in DEV_NO_0, AGE_0, SEX_0 or LAB_DT_0, N = 327 observations were excluded. Due to missing values in CRP_0, N = 130 observations were excluded"
  )

  sd0 <- study_data
  sd0$v00014 <- as.factor(sd0$v00014)
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = sd0,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0", co_vars = c("AGE_0", "SEX_0"),
                label_col = LABEL, plot_format = "BOTH")
    ,
    regexp = "Due to missing values in DEV_NO_0, AGE_0, SEX_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )

  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    2
  )
  skip_on_cran()
  # TODO: skip_if_not(capabilities()["long.double"])
  # skip_on_travis() # vdiffr fails
  skip_if_not_installed("vdiffr")
  expect_doppelganger2("loess facets plot for CRP_0 with Covars ok",
                              res1$SummaryPlotList$Loess_fits_facets)
  expect_doppelganger2("loess combined plot for CRP_0 with Covars ok",
                              res1$SummaryPlotList$Loess_fits_combined)

  nt <- 1000 # = nrow(study_data)/3
  p1 <- 0.3 # baseline probability
  p2 <- 0.7 # probability for the deviating examiner (at the end of the observation period or for a subset of observations)
  p3 <- 0.5 # deviation due to the co_var (categorical)
  sd0 <- study_data
  sd0[["v00011"]] <- rep(c("USR_321", "USR_590", "USR_213"), 1000)
  sd0[["v00002"]] <- rep(c(0,1), 1500)
  set.seed(32)
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_321" & sd0[["v00002"]] == 0)] <-
    rbinom(n = nt/2, size = 1, prob = p1) # only noise
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_321" & sd0[["v00002"]] == 1)] <-
    rbinom(n = nt/2, size = 1, prob = p1 * p3) # only noise, with co-var effect
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_590")] <-
    rbinom(n = nt, size = 1,
           prob = ((p2 - p1)/(nt - 1) * 1:nt + (p1 * nt - p2)/(nt - 1)) *
             (1 - sd0[["v00002"]][which(sd0[["v00011"]] == "USR_590")] * (1 - p3)))
  # increase over time
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_213" & sd0[["v00002"]] == 0)] <-
    c(rbinom(n = floor(1/3 * nt/2), size = 1, prob = p1),
      rbinom(n = floor(1/3 * nt/2), size = 1, prob = p2),
      rbinom(n = nt/2 - 2 * floor(1/3 * nt/2), size = 1, prob = p1))
  # switch between p1 and p2
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_213" & sd0[["v00002"]] == 1)] <-
    c(rbinom(n = floor(1/3 * nt/2), size = 1, prob = p1 * p3),
      rbinom(n = floor(1/3 * nt/2), size = 1, prob = p2 * p3),
      rbinom(n = nt/2 - 2 * floor(1/3 * nt/2), size = 1, prob = p1 * p3))
  md0 <- meta_data
  md0[["RECODE_CASES"]] <- ""
  md0[["RECODE_CONTROL"]] <- ""
  md0[["RECODE_CASES"]][md0$VAR_NAMES == "v00007"] <- "yes"
  md0[["RECODE_CONTROL"]][md0$VAR_NAMES == "v00007"] <- "no"

#   plot_sex0 <- util_plot_categorical_vars(
#     resp_vars = "ASTHMA_0", time_vars = "EXAM_DT_0", group_vars = "USR_VO2_0",
#     study_data = sd0[which(sd0[["v00002"]] == 0), ], meta_data = md0)
#   plot_sex1 <- util_plot_categorical_vars(
#     resp_vars = "ASTHMA_0", time_vars = "EXAM_DT_0", group_vars = "USR_VO2_0",
#     study_data = sd0[which(sd0[["v00002"]] == 1), ], meta_data = md0)

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "ASTHMA_0", # categorical variable
                study_data = sd0,
                meta_data = md0,
                group_vars = "USR_VO2_0",
                time_vars = "EXAM_DT_0",
                co_vars = "SEX_0",
                label_col = LABEL)
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$ASTHMA_0)), "try-error"))
})

test_that("acc_loess works for all time ranges", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  skip_on_cran() # slow test
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  sd0 <- study_data
  v <- subset(meta_data, LABEL == "LAB_DT_0", VAR_NAMES, TRUE)
  sd0[[v]] <- min(sd0[[v]], na.rm = TRUE)
  expect_error(
    suppressWarnings(res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = sd0,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL)),
    regexp = paste("Variable .+LAB_DT_0.+time_vars.+has fewer distinct values",
                   "than required")
  )

  sd0 <- study_data
  sd0[["v00017"]] <- sample(sd0[["v00017"]][c(100, 700, 1500)],
                            size = nrow(sd0), replace = TRUE) #LAB_DT_0
  expect_silent(
    suppressMessages(res1 <-
                       acc_loess(resp_vars = "CRP_0", study_data = sd0,
                                 meta_data = meta_data, group_vars = "DEV_NO_0",
                                 time_vars = "LAB_DT_0",
                                 label_col = LABEL,
                                 comparison_lines = list(type = "quartiles",
                                                         color = "red",
                                                         linetype = 3)))
  )

  sd0 <- study_data
  # FIXME: The old adjust_type function does not interpret 19 as 0019 but as 2019. same hint, there. 1969 is the first year with 19, there.
  sd0[["v00017"]][1:3] <- sd0[["v00017"]][1:3] - 2000 * 60 * 60 * 24 * 365
  # extreme gap, unnoticed typo in year (inspired from real-world data case)
  md0 <- meta_data
  md0[["HARD_LIMITS"]][md0[["VAR_NAMES"]] == "v00017"] <- NA
  sd0[["v00013"]][1:3] <- sd0[["v00013"]][1:3] - 2000 * 60 * 60 * 24 * 365
  md0[["HARD_LIMITS"]][md0[["VAR_NAMES"]] == "v00013"] <- NA
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = sd0,
                meta_data = md0, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL,
                comparison_lines = list(type = "quartiles",
                                        color = "red",
                                        linetype = 3)),
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$CRP_0)), "try-error"))

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00008", # categorical variable
                study_data = sd0, meta_data = md0,
                group_vars = "v00011",
                time_vars = "v00013")
    ,
    regexp = sprintf("(%s|%s|%s)",
                     paste("Due to missing values in v00011 or v00013,",
                           "N = 218 observations were excluded. Due to missing",
                           "values in v00008, N = 270 observations were",
                           "excluded additionally."),
                     "Recoded 1 variable.s.",
                     paste("Levels of the group_var with too few observations",
                           "were discarded .level USR_599..")
    )
  )
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$v00008)), "try-error"))


  sd0 <- study_data
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = sd0,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                resolution = 0.1,
                label_col = LABEL),
    regexp = sprintf(
      "(%s|%s)",
      "Argument resolution is not specified correctly and is set to 80 instead",
      "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
    )
  )

  expect_silent(
    suppressMessages(res1 <-
                       acc_loess(resp_vars = "CRP_0", study_data = sd0,
                                 meta_data = meta_data, group_vars = "DEV_NO_0",
                                 time_vars = "LAB_DT_0",
                                 resolution = 10,
                                 label_col = LABEL))
  )

  expect_silent(
    suppressMessages(res1 <-
                       acc_loess(resp_vars = "CRP_0", study_data = sd0,
                                 meta_data = meta_data, group_vars = "DEV_NO_0",
                                 time_vars = "LAB_DT_0",
                                 resolution = 10000,
                                 label_col = LABEL))
  )

  set.seed(1024)
  sd0 <- data.frame("v00014" = c(rnorm(4000, mean = 10, sd = 1),
                                 rnorm(4000, mean = 20, sd = 10),
                                 rnorm(12, mean = 5, sd = 1),
                                 rnorm(50, mean = 5, sd = 1)),
                    "v00016" = c(rep(1, 4000),
                                 rep(2, 4000),
                                 rep(3, 12),
                                 rep(4, 50)),
                    "v00017" = c(sample(study_data$v00017,
                                        size = 8012, replace = TRUE),
                                 rep(min(study_data$v00017, na.rm = TRUE), 25),
                                 rep(max(study_data$v00017, na.rm = TRUE), 25)))
  md0 <- meta_data[which(meta_data$VAR_NAMES %in% colnames(sd0)), ]
  expect_message2(
    acc_test <-
      acc_loess(resp_vars = "v00014", study_data = sd0,
                meta_data = md0, group_vars = "v00016",
                time_vars = "v00017",
                plot_observations = TRUE),
    regexp = sprintf(
      "(%s|%s|%s)",
      paste("Not all entries in .KEY_STUDY_SEGMENT. are found in .VAR_NAMES.+",
            "Could convert 0 of these 3 to match .VAR_NAMES.+standard.+This",
            "may be caused by providing subsets of .meta_data.+"),
      "Due to missing values in",
      paste("Levels of the group_var with too few observations were",
            "discarded .levels 3, 4.+")
    ),
    perl = TRUE
  )
  expect_true("SummaryPlotList" %in% names(acc_test))
  expect_lt(
    suppressWarnings(abs(mean(as.numeric(
      as.matrix(acc_test$SummaryPlotList$v00014$data)),
      na.rm = TRUE) - mean(sd0$v00014))), 50
  )

  sd0 <- study_data
  g <- subset(meta_data, LABEL == "DEV_NO_0", VAR_NAMES, TRUE)
  set.seed(42)
  sd0[[g]][!is.na(sd0[[g]])] <-
    sample(x = 11, size = sum(!is.na(sd0[[g]])),
           replace = TRUE) # for >= 11 groups,
                           # R standard colors are used.
  expect_message2(
    res0 <-
      acc_loess(resp_vars = "CRP_0", study_data = sd0,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "BOTH")
  )

  sd0[[g]][!is.na(sd0[[g]])] <-
    sample(x = 8, size = sum(!is.na(sd0[[g]])),
           replace = TRUE) # for <= 8 groups,
                           # dataquieR standard colors are used.
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = sd0,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "BOTH")
  )

  g <- ggplot2::ggplot_build(res1$SummaryPlotList$Loess_fits_combined)
  got1 <- sort(unique(g$data[[1]][["colour"]]))
  hex_code <- sort(c( # the dataquieR colors
    "#E69F00", "#56B4E9", "#009E73",
    "#F0E442", "#0072B2", "#D55E00", "#CC79A7", "#8C510A"
  ))
  expect_equal(got1, hex_code)

  g <- ggplot2::ggplot_build(res0$SummaryPlotList$Loess_fits_combined)
  got0 <- sort(unique(g$data[[1]][["colour"]]))
  skip_if_not_installed("scales")
  ggs_default <- sort(scales::hue_pal()(11))

  expect_equal(got0, ggs_default)
})

test_that("acc_loess output matches plot_format=auto", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "AUTO")
    ,
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    1
  )
  sd1 <- study_data
  set.seed(42)
  sd1$v00016 <- sample(1:20, size = nrow(sd1), replace = TRUE)
  expect_message2(
    res2 <-
      acc_loess(resp_vars = "CRP_0", study_data = sd1,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "AUTO")
    ,
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 60 observations were excluded. Due to missing values in CRP_0, N = 241 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res2))
  expect_equal(
    length(res2$SummaryPlotList),
    1
  )
  skip_on_cran()
  # skip_on_travis() # vdiffr fails
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("loess plot for CRP_0 AUTO1 ok",
                              res1$SummaryPlotList$CRP_0)
  expect_doppelganger2("loess plot for CRP_0 AUTO2 ok",
                              res2$SummaryPlotList$CRP_0)
})

test_that("acc_loess output matches plot_format=combined", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "COMBINED")
    ,
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    1
  )
  skip_on_cran()
  # skip_on_travis() # vdiffr fails
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("loess combined plot for CRP_0 COMBINED ok",
                              res1$SummaryPlotList$CRP_0)
})

test_that("acc_loess output matches plot_format=facets", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "FACETS")
    ,
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    1
  )
  skip_on_cran()
  # skip_on_travis() # vdiffr fails
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("loess facets plot for CRP_0 FACETS ok",
                              res1$SummaryPlotList$CRP_0)
})

test_that("acc_loess output matches plot_format=both", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "BOTH")
    ,
    regexp = "Due to missing values in DEV_NO_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    2
  )
  skip_on_cran()
  # skip_on_travis() # vdiffr fails
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("loess facets plot for CRP_0 BOTH ok",
                              res1$SummaryPlotList$Loess_fits_facets)
  expect_doppelganger2("loess combined plot for CRP_0 BOTH ok",
                              res1$SummaryPlotList$Loess_fits_combined)
})

test_that("acc_loess output matches plot_format=invalid1", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_error(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = "invalid")
    ,
    regexp = "Argument plot_format must match the predicate"
  )
})

test_that("acc_loess output matches plot_format=invalid2", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_error(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL, plot_format = 1:10)
    ,
    regexp = "Need exactly one element in argument plot_format, got 10"
  )
})

test_that("acc_loess works without a grouping variable", {
  skip_on_cran() # slow

  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  skip_if_translated()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00014", # continuous variable
                study_data = study_data,
                meta_data = meta_data,
                time_vars = "v00017",
                label_col = LABEL)
    ,
    regexp = "Due to missing values in LAB_DT_0, N = 60 observations were excluded. Due to missing values in CRP_0, N = 241 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$CRP_0)), "try-error"))

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00008", # categorical variable
                study_data = study_data,
                meta_data = meta_data,
                time_vars = "v00013",
                label_col = LABEL)
    ,
    regexp = sprintf("(%s|%s)",
                     paste("Due to missing values in EXAM_DT_0,",
                           "N = 60 observations were excluded. Due to missing",
                           "values in VO2_CAPCAT_0, N = 345 observations were",
                           "excluded additionally."),
                     "Recoded 1 variable.s."
    )
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$VO2_CAPCAT_0)), "try-error"))
})

test_that("optional features for acc_loess work as expected", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0", co_vars = c("AGE_0", "SEX_0"),
                label_col = LABEL, mark_time_points = TRUE,
                plot_observations = TRUE, min_bandwidth = 0.5)
    ,
    regexp = "Due to missing values in DEV_NO_0, AGE_0, SEX_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$CRP_0)), "try-error"))

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0", co_vars = c("AGE_0", "SEX_0"),
                label_col = LABEL, enable_GAM = TRUE)
    ,
    regexp = "Due to missing values in DEV_NO_0, AGE_0, SEX_0 or LAB_DT_0, N = 308 observations were excluded. Due to missing values in CRP_0, N = 131 observations were excluded"
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$CRP_0)), "try-error"))

  expect_message2(
    res1 <-
      acc_loess(resp_vars = "v00008", # categorical variable
                study_data = study_data,
                meta_data = meta_data,
                time_vars = "v00013",
                label_col = LABEL, enable_GAM = TRUE)
    ,
    regexp = sprintf("(%s|%s)",
                     paste("Due to missing values in EXAM_DT_0,",
                           "N = 60 observations were excluded. Due to missing",
                           "values in VO2_CAPCAT_0, N = 345 observations were",
                           "excluded additionally."),
                     "Recoded 1 variable.s."
    )
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$VO2_CAPCAT_0)), "try-error"))

  sd0 <- study_data
  sd0[["v00007"]][which(sd0[["v00011"]] == "USR_211")] <- 0 # one subgroup has constant values
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "ASTHMA_0", # categorical variable
                study_data = sd0,
                meta_data = meta_data,
                group_vars = "USR_VO2_0",
                time_vars = "EXAM_DT_0",
                label_col = LABEL,
                exclude_constant_subgroups = TRUE,
                min_bandwidth = 0.5,
                min_proportion = 0.05)
    ,
    regexp = sprintf("(%s|%s|%s|%s)",
                     paste("Due to missing values in USR_VO2_0 or EXAM_DT_0,",
                           "N = 218 observations were excluded. Due to missing",
                           "values in ASTHMA_0, N = 259 observations were",
                           "excluded additionally."),
                     "Recoded 1 variable.s.",
                     paste("Levels of the group_var with too few observations",
                           "were discarded .level USR_599.."),
                     paste("Levels of the group_var with constant values",
                           "were discarded .level USR_211..")
    )
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$ASTHMA_0)), "try-error"))

  sd0 <- study_data
  sd0[["v00014"]][which(sd0[["v00016"]] == 4)] <- 2.5 # one subgroup has constant values
  expect_message2(
    res1 <-
      acc_loess(resp_vars = "CRP_0",
                study_data = sd0,
                meta_data = meta_data,
                group_vars = "DEV_NO_0",
                time_vars = "LAB_DT_0",
                label_col = LABEL,
                exclude_constant_subgroups = TRUE)
    ,
    regexp = sprintf("(%s|%s)",
                     paste("Due to missing values in DEV_NO_0 or LAB_DT_0,",
                           "N = 308 observations were excluded.",
                           "Due to missing values in CRP_0, N = 121",
                           "observations were excluded additionally"),
                     paste("Levels of the group_var with constant values",
                           "were discarded .level 4..")
    )
  )
  expect_true("SummaryPlotList" %in% names(res1))
  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlotList$CRP_0)), "try-error"))
})

test_that("print.dataquieR_resultset2 works", {
  skip_on_cran() # slow
  skip_if_not_installed("stringdist")
  prep_purge_data_frame_cache()
  study_data <- head(
    prep_get_data_frame(
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE),
    50)

  r <- dq_report2(study_data = study_data, resp_vars = "SBP_0",
                            meta_data_v2 =
               "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            cores = NULL,
                            dimensions = "Integrity")

  expect_warning(
    print(r, cores = 3, view = FALSE, dir = tempfile()),
    regexp = "Internal problem.+should be an integer below.+in the context.+"
  )

  expect_warning(print(r, cores = list(mode = "socket",
                        logging = FALSE,
                        cpus = 2,
                        load.balancing = TRUE),
        view = FALSE, dir = tempfile()),
        regexp = "Internal problem.+should be an integer below.+in the context.+"
  )

  skip_if_not_installed("parallel")
  cl <- parallel::makePSOCKcluster(1)
  withr::defer(parallel::stopCluster(cl))
  expect_warning(print(r, cores = cl, view = FALSE, dir = tempfile()),
                 regexp = "Internal problem.+should be an integer below.+in the context.+"
  )

  r <- dq_report2(study_data = study_data, resp_vars = "SBP_0",
                  meta_data_v2 =
                    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                  filter_indicator_functions = "acc_margins",
                  cores = NULL,
                  dimensions = "Integrity")
  expect_error({
    print(r)
  }, regexp = ".*results at all\\.")

  r <- dq_report2(study_data = study_data, resp_vars = "SBP_0",
                  meta_data_v2 =
                    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                  filter_indicator_functions = "con_inadmissible_categorical",
                  cores = NULL,
                  dimensions = "Consistency")
  expect_error({
    print(r)
  }, regexp = "Report is empty, no results at all.*Applicability")

})

test_that("class ReportSummaryTable", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  sd1 <- study_data
  md1 <- meta_data
  code_labels <- prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|missing_table")

  md1 <- prep_add_cause_label_df(md1, code_labels)

  item_miss <- com_item_missingness(study_data      = sd1,
                                    meta_data       = md1,
                                    label_col       = 'LABEL',
                                    show_causes     = TRUE,
                                    include_sysmiss = TRUE,
                                    threshold_value = 80,
                                    suppressWarnings = TRUE,
                                    drop_levels = TRUE,
                                    assume_consistent_codes = TRUE,
                                    expand_codes = TRUE
  )
  apm <- pro_applicability_matrix(study_data, meta_data)

  item_missc <- com_item_missingness(study_data      = sd1,
                                     resp_vars       = c("CENTER_0"),
                                     meta_data       = md1,
                                     label_col       = 'LABEL',
                                     show_causes     = TRUE,
                                     include_sysmiss = TRUE,
                                     threshold_value = 80,
                                     suppressWarnings = TRUE,
                                     drop_levels = !TRUE,
                                     assume_consistent_codes = TRUE,
                                     expand_codes = TRUE
  )

  item_misss <- com_item_missingness(study_data      = sd1,
                                     resp_vars       = c("SEX_0"),
                                     meta_data       = md1,
                                     label_col       = 'LABEL',
                                     show_causes     = TRUE,
                                     include_sysmiss = TRUE,
                                     threshold_value = 80,
                                     suppressWarnings = TRUE,
                                     drop_levels = !TRUE,
                                     assume_consistent_codes = TRUE,
                                     expand_codes = TRUE
  )

  item_missa <- com_item_missingness(study_data      = sd1,
                                     resp_vars       = c("AGE_0"),
                                     meta_data       = md1,
                                     label_col       = 'LABEL',
                                     show_causes     = TRUE,
                                     include_sysmiss = TRUE,
                                     threshold_value = 80,
                                     suppressWarnings = TRUE,
                                     drop_levels = !TRUE,
                                     assume_consistent_codes = TRUE,
                                     expand_codes = TRUE
  )

  item_missb <- com_item_missingness(study_data      = sd1,
                                     resp_vars       = c("SBP_0"),
                                     meta_data       = md1,
                                     label_col       = 'LABEL',
                                     show_causes     = TRUE,
                                     include_sysmiss = TRUE,
                                     threshold_value = 80,
                                     suppressWarnings = TRUE,
                                     drop_levels = !TRUE,
                                     assume_consistent_codes = TRUE,
                                     expand_codes = TRUE
  )

  item_miss_combined <- rbind(item_missa$ReportSummaryTable,
                              item_missb$ReportSummaryTable,
                              item_missc$ReportSummaryTable,
                              item_misss$ReportSummaryTable)

  expect_equal(nrow(item_miss_combined), 4)
  expect_equal(nrow(item_miss$ReportSummaryTable), 53)
  expect_s3_class(item_miss_combined, "ReportSummaryTable")
  expect_s3_class(apm$ReportSummaryTable, "ReportSummaryTable")

  item_miss_emtpy <- com_item_missingness(
     study_data      = sd1,
     resp_vars       = c("CENTER_0"),
     meta_data       = md1,
     label_col       = 'LABEL',
     show_causes     = TRUE,
     include_sysmiss = FALSE,
     threshold_value = 80,
     suppressWarnings = TRUE,
     drop_levels = TRUE,
     assume_consistent_codes = TRUE,
     expand_codes = TRUE
  )


  skip_on_cran()

  fkt2 <- function(x, size_min, size_max) {
    old <- options("viewer")
    td <- tempfile()
    stopifnot(dir.create(td))
    on.exit({
      options(old)
      unlink(td, TRUE, TRUE)
    })
    v <- function(x, ...) {
    }
    options(viewer = v)
    w <- print.ReportSummaryTable(x = x, dt = TRUE)
    expect_gt(object.size(w), size_min)
    expect_lt(object.size(w), size_max)
  }
  expect_warning(fkt2(item_miss_emtpy$ReportSummaryTable,
       size_max = 42000, size_min = 0), regexp = "Empty result")
  skip_if_not_installed("DT")
  fkt2(item_miss_combined, size_min = 50000,
       size_max = 80000)
  fkt2(item_missa$ReportSummaryTable, size_min = 40000,
       size_max = 60000)
  fkt2(item_missb$ReportSummaryTable, size_min = 50000,
       size_max = 70000)
  fkt2(item_missc$ReportSummaryTable, size_min = 40000,
       size_max = 60000)
  fkt2(item_misss$ReportSummaryTable, size_min = 40000,
       size_max = 60000)
  fkt2(item_miss$ReportSummaryTable, size_min = 80000,
       size_max = 200000)
  fkt2(apm$ReportSummaryTable, size_min = 50000,
       size_max = 200000)

  g1 <- R.devices::suppressGraphics(print(apm$ReportSummaryTable,
                                          dt = FALSE))
  g2 <- R.devices::suppressGraphics(print(item_miss$ReportSummaryTable,
                                          dt = FALSE))
  g3 <- R.devices::suppressGraphics(print(item_missa$ReportSummaryTable,
                                          dt = FALSE))
  expect_warning(R.devices::suppressGraphics(g4 <- print(item_miss_emtpy$ReportSummaryTable,
                                          dt = FALSE)), regexp = "Empty result")

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("app-ex-repsumtab", g1)
  expect_doppelganger2("im-ex1-repsumtab", g2)
  expect_doppelganger2("im-ex2-repsumtab", g3)
  expect_warning(expect_doppelganger2("im-empty-repsumtab", g4),
    regexp = "Empty result")

})

test_that("print.interval works", {
  skip_on_cran()
  require_english_locale_and_berlin_tz()
  expect_output(
    print(util_parse_interval("(1;)")),
    "(1;Inf)",
    fixed = TRUE
  )
  expect_output(
    print(util_parse_interval("[1;)")),
    "[1;Inf)",
    fixed = TRUE
  )
  expect_output(
    print(util_parse_interval("[1;]")),
    "[1;Inf]",
    fixed = TRUE
  )
  expect_output(
    print(util_parse_interval("(2001-01-01;)")),
    "(2001-01-01;Inf)",
    fixed = TRUE
  )
  expect_output(
    print(util_parse_interval("[2001-01-01;)")),
    "[2001-01-01;Inf)",
    fixed = TRUE
  )
  expect_output(
    print(util_parse_interval("[2001-01-01;]")),
    "[2001-01-01;Inf]",
    fixed = TRUE
  )
})

test_that("print.ReportSummaryTable works when called within the pipeline with duplicated labels for the plot, which are not the primary labels for the report", {
  skip_on_cran() # slow test
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("stringdist")
  skip_if_not_installed("DT")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  md0 <- meta_data
  md0$STUDY_SEGMENT <- NULL
  md0$LABEL[26:27] <- "EDUCATION"
  suppressWarningsMatching(
    dq <- dq_report2(study_data = study_data, item_level = md0, cores = NULL,
                     label_col = "LONG_LABEL",
                     filter_indicator_functions = "com_item_miss"),
    "Unique labels are required")
  # check that the VAR_NAMES attribute exists and is correct
  exp_vn <- prep_map_labels(x = c("EDUCATION_0", "EDUCATION_1"),
                            item_level = md0,
                            to = VAR_NAMES, from = LONG_LABEL)
  expect_equal(
    attr(dq$com_item_missingness.EDUCATION_0$ReportSummaryTable, "VAR_NAMES"),
    exp_vn[1]
  )
  # class(dq$com_item_missingness.EDUCATION_0$ReportSummaryTable) # is a ReportSummaryTable
  # class(dq["EDUCATION_0", "com_item_missingness", "ReportSummaryTable"]) # is a dataquieR_result
  # class(dq["EDUCATION_0", "com_item_missingness", "ReportSummaryTable"][[1]]) # is a ReportSummaryTable
  expect_equal(
    attr(dq["EDUCATION_0", "com_item_missingness", "ReportSummaryTable"][[1]], "VAR_NAMES"),
    exp_vn[1]
  )
  expect_equal(
    attr(dq[c("EDUCATION_0", "EDUCATION_1"), "com_item_missingness", "ReportSummaryTable"][[1]], "VAR_NAMES"),
    exp_vn
  )
  # check that the ReportSummaryTable can be printed
  expect_false(
    inherits(try(print(dq["EDUCATION_0", "com_item_missingness", "ReportSummaryTable"])), "try-error")
  )
  expect_false(
    inherits(try(dq[, "com_item_missingness", "ReportSummaryTable"]), "try-error")
  )
  suppressWarningsMatching({
    expect_false(
      inherits(try(print(dq[, "com_item_missingness", "ReportSummaryTable"])),
               "try-error")
    )
  }, regexps = ".*not being part of one of the segments.*")

  dq_rst <- dq[c("EDUCATION_0", "EDUCATION_1"), "com_item_missingness", "ReportSummaryTable"]
  expect_false(
    inherits(try(print(dq_rst)), "try-error")
  )
  expect_false(
    inherits(try(print(dq_rst$ReportSummaryTable)), "try-error")
  )
  expect_false(
    inherits(try(print(dq_rst[[1]])), "try-error")
  )
  # check that the expected axis labels appear in the figure
#  exp_lab <- exp_vn
#  names(exp_lab) <- rep("EDUCATION", 2)
  exp_lab <- c("v00018: EDUCATION", "v01018: EDUCATION")
#  exp_lab <- paste(names(exp_lab), paste0("(", exp_lab, ")"))
  dq_rst_plot <- ggplot_build(print(dq_rst$ReportSummaryTable))
  expect_true(
    identical(dq_rst_plot$layout$panel_params[[1]]$x$get_labels(),
      exp_lab) ||
    identical(dq_rst_plot$layout$panel_params[[1]]$y$get_labels(),
      exp_lab)
  )
})

test_that("util_adjust_data_type2==util_adjust_data_type2", {
  skip_on_cran() # depends on locales

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  ds1_old <- withr::with_options(
    list(dataquieR.old_type_adjust = "TRUE"),
    prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data)
  )
  ds1_new <- withr::with_options(
    list(dataquieR.old_type_adjust = "FALSE"),
    prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data)
  )

  expect_equal(ds1_new, ds1_old)

})

test_that("util_adjust_data_type2==util_adjust_data_type2_FCT", {
  skip_on_cran() # depends on locales

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # testthat::local_reproducible_output()
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  study_data$v00014 <- as.factor(study_data$v00014)

  ds1_old <- withr::with_options(
    list(dataquieR.old_type_adjust = "TRUE"),
    prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data)
  )
  ds1_new <- withr::with_options(
    list(dataquieR.old_type_adjust = "FALSE"),
    prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data)
  )

  expect_equal(ds1_new, ds1_old)

})

test_that("con_limit_deviations works", {
  skip_on_cran() # slow, errors obvious

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  require_english_locale_and_berlin_tz()
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]

  expect_message2(
    MyValueLimits <- con_limit_deviations(resp_vars  = c("AGE_0", "SBP_0",
                                                         "DBP_0", "SEX_0",
                                                         "QUEST_DT_0",
                                                         "EDUCATION_1",
                                                         "SMOKE_SHOP_0"),
                                          label_col  = "LABEL",
                                          study_data = study_data,
                                          meta_data  = meta_data),
    regexp = "No limits specified for SEX_0")

   expect_equal(
     sum(unname(rowSums(MyValueLimits$SummaryTable[#Sum all values outside HARD_LIMITS in SummaryTable
       order(MyValueLimits$SummaryTable$Variables),
       c("NUM_con_rvv_inum", "NUM_con_rvv_itdat")], na.rm = TRUE))),
     sum(as.numeric( #Sum all values outside HARD_LIMITS in SummaryData
       unname(
         vapply(
           X = unlist(
             as.vector(
               MyValueLimits$SummaryData[MyValueLimits$SummaryData$Limits == "HARD_LIMITS",
                                         c("Below limits N (%)",
                                           "Above limits N (%)")])),
           FUN = function(x) {
             gsub("\\s*\\([^\\)]+\\)", "", x)
             },
           FUN.VALUE = character(1)))))
   )
})

test_that("con_limit_deviations and timevars with < 20 integer sec-values ok", {
  skip_on_cran() # slow, errors obvious

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  vn <- subset(meta_data, LABEL == "QUEST_DT_0", VAR_NAMES, TRUE)

  sd0 <- study_data[
    study_data[[vn]] >= as.POSIXct("2018-05-01") &
      study_data[[vn]] < as.POSIXct("2018-05-15"), , FALSE]

  sd0[[vn]] <- round.POSIXt(sd0[[vn]], "days")

  expect_lt(length(unique(as.character(sd0[[vn]]))), 20)

  set.seed(1843) # randomly scattered points should stay in their position for testing
  MyValueLimits <- con_limit_deviations(resp_vars  = "QUEST_DT_0",
                                        label_col  = "LABEL",
                                        study_data = sd0,
                                        meta_data  = meta_data,
                                        limits     = "HARD_LIMITS")


  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  # TODO: skip_if_not(capabilities()["long.double"])
  suppressWarnings(
    expect_doppelganger2(
      "con_limit_deviations QUEST_DT_0",
                                MyValueLimits$SummaryPlotList$QUEST_DT_0)
  )

})

test_that("con_limit_deviations works w/o resp_vars", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  require_english_locale_and_berlin_tz()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  resp_vars <- c("AGE_0", "SBP_0",
                 "DBP_0", "SEX_0",
                 "QUEST_DT_0",
                 "EDUCATION_1",
                 "SMOKE_SHOP_0")

  resp_vars_names <- prep_map_labels(resp_vars,
                                     meta_data,
                                     from = LABEL,
                                     to = VAR_NAMES)

  sd0 <- study_data[, resp_vars_names, FALSE]
  md0 <- meta_data[meta_data$VAR_NAMES %in% resp_vars_names, , FALSE]
  md0$JUMP_LIST <- SPLIT_CHAR # signals unambiguously no codes intentionally


  expect_message2(
    MyValueLimits <- con_limit_deviations(label_col  = "LABEL",
                                          study_data = sd0,
                                          meta_data  = md0,
                                          limits     = "HARD_LIMITS",
                                          return_flagged_study_data = TRUE),
    regexp = sprintf("(%s|%s)",
                     paste("All variables for which limits are specified",
                           "in the metadata are used."),
                     paste("Not all entries in.+KEY_STUDY_SEGMENT.+are",
                           "found in.+VAR_NAMES.+")
    ),
    perl = TRUE
  )

  expect_equal(sum(rowSums(
    MyValueLimits$SummaryTable[,
      grepl("^FLG\\_.*$", colnames(MyValueLimits$SummaryTable))],
    na.rm = TRUE) > 0), 4)
  expect_true(all(MyValueLimits$FlaggedStudyData$QUEST_DT_0[which(as.logical(
    MyValueLimits$FlaggedStudyData$QUEST_DT_0_below_HARD))] <
      as.POSIXct("2018-01-01")))
  expect_true(all(MyValueLimits$FlaggedStudyData$QUEST_DT_0[which(!as.logical(
    MyValueLimits$FlaggedStudyData$QUEST_DT_0_below_HARD))] >=
      as.POSIXct("2018-01-01")))
})

test_that("con_limit_deviations handles errors", {
  skip_on_cran() # slow, errors obvious
  require_english_locale_and_berlin_tz()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  resp_vars <- c("AGE_0", "SBP_0",
                 "DBP_0", "SEX_0",
                 "QUEST_DT_0",
                 "EDUCATION_1",
                 "SMOKE_SHOP_0")

  resp_vars_names <- prep_map_labels(resp_vars,
                                     meta_data,
                                     from = LABEL,
                                     to = VAR_NAMES)

  sd0 <- study_data[, resp_vars_names, FALSE]
  md0 <- meta_data[meta_data$VAR_NAMES %in% resp_vars_names, , FALSE]
  md0$JUMP_LIST <- SPLIT_CHAR # signals unambiguously no codes intentionally
  # no limits:
  md0$HARD_LIMITS <- NA
  md0$DETECTION_LIMITS <- NA
  md0$SOFT_LIMITS <- NA

  expect_message2(expect_error(
    MyValueLimits <- con_limit_deviations(resp_vars = resp_vars,
                                          label_col  = "LABEL",
                                          study_data = sd0,
                                          meta_data  = md0),
    regexp = paste("The limit deviation check cannot be performed",
                   "without a metadata column specifying suitable limits")
  ),
  regexp = "Not all entries in .+KEY_STUDY_SEGMENT.+0.+7.*")

  resp_vars <- meta_data$LABEL
  resp_vars_names <- prep_map_labels(resp_vars,
                                     meta_data,
                                     from = LABEL,
                                     to = VAR_NAMES)

  sd0 <- study_data[, resp_vars_names, FALSE]
  md0 <- meta_data[meta_data$VAR_NAMES %in% resp_vars_names, , FALSE]
  md0$JUMP_LIST <- SPLIT_CHAR # signals unambiguously no codes intentionally
  md0$HARD_LIMITS[md0$DATA_TYPE == DATA_TYPES$STRING] <-
    "[1; 5]" # Some unsuitable limits

  expect_error(
      MyValueLimits <- suppressWarnings(
        con_limit_deviations(resp_vars = md0$LABEL, # missing, only resp_vars with suitable data type are picked by the function now.
                             label_col  = "LABEL",
                             study_data = sd0,
                             meta_data  = md0)),
      regexp = "Variable .+PSEUDO_ID.+ .+string.+ does not have an allowed type .+integer.+float.+datetime.+",
    perl = TRUE
  )

  sd0 <- study_data
  sd0[[prep_map_labels("BSG_0", meta_data, VAR_NAMES, LABEL)]] <- NA
  expect_error(
    MyValueLimits  <- con_limit_deviations(resp_vars  = "BSG_0",
                                         label_col  = "LABEL",
                                         study_data = sd0,
                                         meta_data  = meta_data,
                                         limits     = "HARD_LIMITS"),
    regexp = "Variable .+BSG_0.+ .+resp_vars.+ has only NA observations"
  )
})

test_that("con_limit_deviations and values < 0", {
  skip_on_cran() # slow, errors obvious

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  require_english_locale_and_berlin_tz()

  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  sd0 <- study_data
  x <- sd0[[prep_map_labels("EDUCATION_0", meta_data, VAR_NAMES, LABEL)]]

  x[!is.na(x) & x < 10] <- 0 - x[!is.na(x) & x < 10]

  sd0[[prep_map_labels("EDUCATION_0", meta_data, VAR_NAMES, LABEL)]] <- x

  md0 <- meta_data
  md0[md0$LABEL == "EDUCATION_0", HARD_LIMITS] <-
    "[-6;0]"
  md0[md0$LABEL == "EDUCATION_0", SCALE_LEVEL] <- SCALE_LEVELS$INTERVAL
  set.seed(2012) # randomly scattered points should stay in their position for testing
  MyValueLimits <- con_limit_deviations(resp_vars  = "EDUCATION_0",
                                        label_col  = "LABEL",
                                        study_data = sd0,
                                        meta_data  = md0,
                                        limits     = "HARD_LIMITS")

  expect_type(MyValueLimits, "list")

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  suppressWarnings(
    expect_doppelganger2(
      "con_limit_deviations for reverted",
      MyValueLimits$SummaryPlotList$EDUCATION_0)
  )

})

test_that("con_limit_deviations and values < 0 with max -1", {
  skip_on_cran() # slow, errors obvious
  require_english_locale_and_berlin_tz()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  sd0 <- study_data
  x <- sd0[[prep_map_labels("EDUCATION_0", meta_data, VAR_NAMES, LABEL)]]

  x[!is.na(x) & x < 10] <- -1 - x[!is.na(x) & x < 10]

  sd0[[prep_map_labels("EDUCATION_0", meta_data, VAR_NAMES, LABEL)]] <- x

  md0 <- meta_data
  md0[md0$LABEL == "EDUCATION_0", HARD_LIMITS] <-
    "[-7;-1]"
  md0[md0$LABEL == "EDUCATION_0", SCALE_LEVEL] <- SCALE_LEVELS$INTERVAL

  set.seed(2024) # randomly scattered points should stay in their position for testing
  MyValueLimits <- con_limit_deviations(resp_vars  = "EDUCATION_0",
                                        label_col  = "LABEL",
                                        study_data = sd0,
                                        meta_data  = md0,
                                        limits     = "HARD_LIMITS")

  expect_type(MyValueLimits, "list")

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  suppressWarnings(
    expect_doppelganger2(
      "con_limit_deviations reverse w/ -1 ==> 0 max",
      MyValueLimits$SummaryPlotList$EDUCATION_0)
  )

})

test_that("con_limit_deviations with no lower limit", {
  skip_on_cran() # slow, errors obvious
  require_english_locale_and_berlin_tz()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  sd0 <- study_data
  x <- sd0[[prep_map_labels("EDUCATION_0", meta_data, VAR_NAMES, LABEL)]]

  x[2:5] <- -2

  sd0[[prep_map_labels("EDUCATION_0", meta_data, VAR_NAMES, LABEL)]] <- x
  md0 <- meta_data
  md0[md0$LABEL == "EDUCATION_0", SCALE_LEVEL] <- SCALE_LEVELS$INTERVAL
  expect_silent(
    MyValueLimits1 <- con_limit_deviations(resp_vars  = "EDUCATION_0",
                                          label_col  = "LABEL",
                                          study_data = sd0,
                                          meta_data  = md0,
                                          limits     = "HARD_LIMITS")
  )

  md0[md0$LABEL == "EDUCATION_0", HARD_LIMITS] <- "[;6]"

  MyValueLimits2 <- con_limit_deviations(resp_vars  = "EDUCATION_0",
                                         label_col  = "LABEL",
                                         study_data = sd0,
                                         meta_data  = md0,
                                         limits     = "HARD_LIMITS")

  expect_equal(MyValueLimits1$SummaryTable$NUM_con_rvv_inum, 4,
               ignore_attr = TRUE)
  expect_equal(MyValueLimits2$SummaryTable$NUM_con_rvv_inum, 0,
               ignore_attr = TRUE)

  expect_equal(MyValueLimits1$SummaryData$Number[
    which(MyValueLimits1$SummaryData$Section == "above")], MyValueLimits2$SummaryData$Number[
      which(MyValueLimits2$SummaryData$Section == "above")])
})

test_that("con_limit_deviations with constant data", {
  skip_on_cran() # slow, errors obvious
  require_english_locale_and_berlin_tz()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  sd0 <- study_data
  sd0[[prep_map_labels("CRP_0", meta_data, VAR_NAMES, LABEL)]] <- 1.1

  md0 <- meta_data

  set.seed(12) # randomly scattered points should stay in their position for testing
  expect_silent(
    MyValueLimits  <- con_limit_deviations(resp_vars  = "CRP_0",
                                         label_col  = "LABEL",
                                         study_data = sd0,
                                         meta_data  = md0,
                                         limits     = "HARD_LIMITS"))

  expect_type(MyValueLimits, "list")

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  suppressWarnings(
    expect_doppelganger2(
      "con_limit_deviations 4 hists 1 val",
      MyValueLimits$SummaryPlotList$CRP_0)
  )

  md0[which(md0$LABEL == "CRP_0"), HARD_LIMITS] <- NA
  md0[which(md0$LABEL == "CRP_0"), DETECTION_LIMITS] <- NA
  md0[which(md0$LABEL == "CRP_0"), SOFT_LIMITS] <- NA
  md0[["NEW_LIMITS_ONE"]] <- NA
  md0[which(md0$LABEL == "CRP_0"), "NEW_LIMITS_ONE"] <- "(-Inf; 1.1)"
  md0[["NEW_LIMITS_TWO"]] <- NA
  md0[which(md0$LABEL == "CRP_0"), "NEW_LIMITS_TWO"] <- "(1.1; Inf)"

  expect_silent(
    MyValueLimits  <- con_limit_deviations(resp_vars  = "CRP_0",
                                           label_col  = "LABEL",
                                           study_data = sd0,
                                           meta_data  = md0,
                                           limits     = c("NEW_LIMITS_ONE", "NEW_LIMITS_TWO")))
})

test_that("con_limit_deviations does not crash with missing codes", {
  skip_on_cran() # slow, errors obvious
  require_english_locale_and_berlin_tz()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  sd0 <- study_data
  x <- sd0[[prep_map_labels("SBP_0", meta_data, VAR_NAMES, LABEL)]]

  x[[5]] <- 99999 # add a "forgotten missing code" to the data, which causes
  # too many breaks, that should be removed to avoid a crash

  sd0[[prep_map_labels("SBP_0", meta_data, VAR_NAMES, LABEL)]] <- x
  md0 <- meta_data

  expect_silent(
    MyValueLimits  <- con_limit_deviations(resp_vars  = "SBP_0",
                                          label_col  = "LABEL",
                                          study_data = sd0,
                                          meta_data  = md0,
                                          limits     = "HARD_LIMITS")
  )

  expect_type(MyValueLimits, "list")
  expect_equal(as.numeric(
    gsub("\\s*\\([^\\)]+\\)",
         "",
         unlist(MyValueLimits$SummaryData[MyValueLimits$SummaryData$Limits ==
                                            "HARD_LIMITS",
                                          "Above limits N (%)" ]))),
    1)
  expect_equal(as.numeric(
    gsub("\\s*\\([^\\)]+\\)",
         "",
         unlist(MyValueLimits$SummaryData[MyValueLimits$SummaryData$Limits ==
                                            "HARD_LIMITS", "Below limits N (%)" ]))),
    0)

  sd0 <- study_data
  x <- sd0[[prep_map_labels("SBP_0", meta_data, VAR_NAMES, LABEL)]]

  x[[5]] <- 100000 # add a less obvious "forgotten missing code" to the data,
  # which could cause too many breaks, that should be caught to avoid a crash

  sd0[[prep_map_labels("SBP_0", meta_data, VAR_NAMES, LABEL)]] <- x

  set.seed(20) # randomly scattered points should stay in their position for testing
  expect_silent(
    MyValueLimits  <- con_limit_deviations(resp_vars  = "SBP_0",
                                           label_col  = "LABEL",
                                           study_data = sd0,
                                           meta_data  = md0,
                                           limits     = "HARD_LIMITS")
  )

  expect_type(MyValueLimits, "list")
  expect_equal(
    as.numeric(gsub("\\s*\\([^\\)]+\\)",
                    "",
                    unlist(MyValueLimits$SummaryData[MyValueLimits$SummaryData$Limits ==
                                                       "HARD_LIMITS",
                                                     "Above limits N (%)" ]))),
    1)
  expect_equal(
    as.numeric(gsub("\\s*\\([^\\)]+\\)",
                    "",
                    unlist(MyValueLimits$SummaryData[MyValueLimits$SummaryData$Limits ==
                                                       "HARD_LIMITS",
                                                     "Below limits N (%)" ]))),
    0)
  expect_lt(nrow(MyValueLimits$SummaryPlotList[[1]]$data), 1000)

  skip_on_cran()
  # TODO: skip_if_not(capabilities()["long.double"])
  skip_if_not_installed("vdiffr")
  suppressWarnings(
    expect_doppelganger2(
      "con_limit_deviations histgrms + misscds",
      MyValueLimits$SummaryPlotList$SBP_0)
  )
})

test_that("con_limit_deviations does not crash with strong outliers", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # issue 116
  require_english_locale_and_berlin_tz()

  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  md0 <- meta_data
  md0[["HARD_LIMITS"]][which(md0$LABEL == "BSG_0")] <- "[0;5000]"
  sd0 <- study_data
  x <- sd0[[prep_map_labels("BSG_0", meta_data, VAR_NAMES, LABEL)]]
  x[c(2,8,92)] <- c(26027, 28097, 4031012019)
  sd0[[prep_map_labels("BSG_0", meta_data, VAR_NAMES, LABEL)]] <- x
  expect_silent(
    MyValueLimits  <- con_limit_deviations(resp_vars  = "BSG_0",
                                         label_col  = "LABEL",
                                         study_data = sd0,
                                         meta_data  = md0,
                                         limits     = "HARD_LIMITS")
  )

  expect_type(MyValueLimits, "list")
  expect_equal(MyValueLimits$SummaryTable$NUM_con_rvv_inum, 3,
               ignore_attr = TRUE)
})

test_that("con_limits_deviations does not crash with NAs in datetime vars", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # issue 106
  require_english_locale_and_berlin_tz()

  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  #sd0 <- study_data
  #x <- sd0[[prep_map_labels("QUEST_DT_0", meta_data, VAR_NAMES, LABEL)]]
  #any(is.na(x))
  #length(which(is.na(x)))
  # Variable QUEST_DT_0 already contains NA values
  expect_silent(
    MyValueLimits  <- con_limit_deviations(resp_vars  = "QUEST_DT_0",
                                         label_col  = "LABEL",
                                         study_data = study_data,
                                         meta_data  = meta_data,
                                         limits     = "HARD_LIMITS")
  )

  expect_type(MyValueLimits, "list")
  expect_equal(MyValueLimits$SummaryTable$NUM_con_rvv_itdat, 9,
               ignore_attr = TRUE)
})

test_that("con_limit_deviations works with no values within limits", {
  skip_on_cran() # slow, errors obvious
  require_english_locale_and_berlin_tz()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  md0 <- meta_data
  md0[["HARD_LIMITS"]][which(md0$LABEL == "BSG_0")] <- "[51.5;51.8]"
  sd0 <- study_data

  set.seed(24) # randomly scattered points should stay in their position for testing
  expect_silent(
    MyValueLimits  <- con_limit_deviations(resp_vars  = "BSG_0",
                                         label_col  = "LABEL",
                                         study_data = sd0,
                                         meta_data  = md0,
                                         limits     = "HARD_LIMITS")
  )

  expect_type(MyValueLimits, "list")
  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  suppressWarnings(
    expect_doppelganger2(
      "con_limit_deviations_nothing_within",
      MyValueLimits$SummaryPlotList$BSG_0)
  )
})

test_that("con_limit_deviations works with wrong datetime limits", {
  skip_on_cran() # slow, errors obvious
  require_english_locale_and_berlin_tz()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  md0 <- meta_data
  md0[["HARD_LIMITS"]][which(md0$LABEL == "QUEST_DT_0")] <- "[51.5;51.8]"
  sd0 <- study_data

  expect_warning(
      MyValueLimits <- con_limit_deviations(resp_vars  = "QUEST_DT_0",
                                            label_col  = "LABEL",
                                            study_data = sd0,
                                            meta_data  = md0,
                                            limits     = "HARD_LIMITS"),
  regexp = "Invalid limits detected")

  md0[["HARD_LIMITS"]][which(md0$LABEL == "QUEST_DT_0")] <-
    "[2018-10-01; 2017-10-01]"
  suppressWarnings(suppressMessages(expect_warning(expect_error(
    MyValueLimits <- con_limit_deviations(resp_vars  = "QUEST_DT_0",
                                          label_col  = "LABEL",
                                          study_data = sd0,
                                          meta_data  = md0,
                                          limits     = "HARD_LIMITS")))))

  md0[["HARD_LIMITS"]][which(md0$LABEL == "QUEST_DT_0")] <-
    "[0; test]"
  suppressWarnings(suppressMessages(expect_warning(expect_error(
    MyValueLimits <- con_limit_deviations(resp_vars  = "QUEST_DT_0",
                                          label_col  = "LABEL",
                                          study_data = sd0,
                                          meta_data  = md0,
                                          limits     = "HARD_LIMITS")))))
})

test_that("con_limit_deviations chooses suitable breaks for integer variables", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]

  md0 <- meta_data[meta_data[[LABEL]] %in% c("PSEUDO_ID", "AGE_0"), ]
  md0$KEY_STUDY_SEGMENT <- ""
  md0[["HARD_LIMITS"]][which(md0$LABEL == "AGE_0")] <- "[1; 110]"

  set.seed(2204)
  sd0 <- data.frame("v00001" = 1:(6*10^4 + 40),
                    "v00003" = c(
                      rep(0, 40),
                      round(abs(rnorm(mean = 53, sd = 12, n = 6*10^4)))))
  suppressMessages(suppressWarnings(
    cc_int <- con_limit_deviations(resp_vars = "AGE_0", study_data = sd0,
                                   meta_data = md0)
  ))
  plotdata <- ggplot2::ggplot_build(cc_int$SummaryPlotList[[1]])$data
  plotdata_bind <- util_rbind(data_frames_list = plotdata)
  plotdata_bind <- plotdata_bind[, c("xmin", "xmax")]
  plotdata_bind <- plotdata_bind[complete.cases(plotdata_bind), ]
  bin_widths <- plotdata_bind$xmax - plotdata_bind$xmin
  expect_true(all(util_is_integer(unique(round(bin_widths, 6)))))
})

test_that("con_limit_deviations complex limits alone", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  cil <- data.frame(VARIABLE_LIST = "SBP_0",
                                 HARD_LIMITS = paste0('[SEX_0] == "" or ([SEX_0] =="males" or ([SEX_0] =="females" and [SBP_0] > 100))'))

    MyValueLimits <- con_limit_deviations(resp_vars = "SBP_0",
                                          study_data = study_data,
                                          meta_data  = meta_data,
                                          meta_data_cross_item = cil)


    expect_equal(
      MyValueLimits$SummaryData$`All outside limits N (%)`, "2 (0.07)",
      ignore_attr = TRUE)

})


test_that("con_limit_deviations complex limits alone", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  cil <- data.frame(VARIABLE_LIST = "DBP_0",
                    HARD_LIMITS = paste0('[DBP_0] > 55'))

  MyValueLimits <- con_limit_deviations(resp_vars = "DBP_0",
                                        label_col = LABEL,
                                        study_data = study_data,
                                        meta_data  = meta_data,
                                        meta_data_cross_item = cil)


  expect_equal(
    MyValueLimits$SummaryData$`All outside limits N (%)`, "3 (0.1)",
    ignore_attr = TRUE)

})

test_that("Time-only variables limits", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                                    keep_types = TRUE)

  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  meta_data <- prep_get_data_frame("item_level")

  set.seed(12345)
  day_course <- unlist(lapply(
    0:23,
    function(h) {
      list(
        hms::hms(hours = h,
                 minutes = 3),
        hms::hms(hours = h,
                 minutes = 13),
        hms::hms(hours = h,
                 minutes = 24),
        hms::hms(hours = h,
                 minutes = 30),
        hms::hms(hours = h,
                 minutes = 50)
      )
    }
  ), recursive = FALSE)
  probs <-
    rep(c(.7, .2, .05, .04, 0.01), 24)
  times <- sample(x = day_course,
                  prob = probs,
                  size = nrow(study_data),
                  replace = TRUE)
  times[sample(seq_along(times),
               size = length(times) %/% 100,
               replace = TRUE)] <-
    list(hms::hms(seconds = 0, minutes = 0, hours = 0))
  study_data$v02000 <- times
  meta_data <- util_rbind(
    meta_data,
    data.frame(
      stringsAsFactors = FALSE,
      VAR_NAMES = "v02000",
      LABEL = "ADMIS_TM_0",
      DATA_TYPE = DATA_TYPES$TIME,
      SCALE_LEVEL = SCALE_LEVELS$INTERVAL,
      VALUE_LABELS = NA_character_,
      STANDARDIZED_VOCABULARY_TABLE = NA_character_,
      MISSING_LIST_TABLE = NA_character_,
      HARD_LIMITS = "[09:00:00;18:00:00]",
      DETECTION_LIMITS = NA_character_,
      SOFT_LIMITS = NA_character_,
      DISTRIBUTION = NA_character_,
      DECIMALS = NA_character_,
      DATA_ENTRY_TYPE = NA_character_,
      GROUP_VAR_OBSERVER = NA_character_,
      GROUP_VAR_DEVICE = NA_character_,
      TIME_VAR = NA_character_,
      STUDY_SEGMENT = "STUDY",
      PART_VAR = "PART_STUDY",
      VARIABLE_ROLE = VARIABLE_ROLES$PRIMARY,
      VARIABLE_ORDER = "54",
      LONG_LABEL = "Admission time",
      ELEMENT_HOMOGENITY_CHECKTYPE = NA_character_,
      UNIVARIATE_OUTLIER_CHECKTYPE = NA_character_,
      N_RULES = "4",
      LOCATION_METRIC = NA_character_,
      LOCATION_RANGE = NA_character_,
      PROPORTION_RANGE = NA_character_,
      REPEATED_MEASURES_VARS = NA_character_,
      REPEATED_MEASURES_GOLDSTANDARD = NA_character_,
      CO_VARS = NA_character_,
      MISSING_LIST = "00:00:00 = not available"
    )
  )

  r <- con_limit_deviations(
    "v02000", study_data = study_data, meta_data =
      meta_data, label_col = LABEL,
    meta_data_v2 =
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  suppressWarnings(
    expect_doppelganger2(
      "con_limit_deviations time only vars",
      r$SummaryPlotList$ADMIS_TM_0)
  )


  # with_testthat(con_limit_deviations("v02000", study_data = study_data, meta_data = meta_data, label_col = LABEL, meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx"))
  # with_testthat(con_limit_deviations("v00042", study_data = study_data, meta_data = meta_data, label_col = LABEL, meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx"))
})

test_that("util_plot_categorical_vars works", {
  skip_on_cran() # slow
  skip_if_translated()

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  require_english_locale_and_berlin_tz()

  expect_message2(
    t1 <- util_plot_categorical_vars(resp_vars = "v00103",
                                 group_vars = "v00012",
                                 time_vars = "v00013",
                                 study_data = study_data, meta_data = meta_data,
                                 n_cat_max = 10,
                                 n_group_max = 5,
                                 n_data_min = 50),
    "Missing some or all entries"
  )

  expect_message2(
    t2 <- util_plot_categorical_vars(resp_vars = "v00103",
                               time_vars = "v00013",
                               study_data = study_data, meta_data = meta_data,
                               n_cat_max = 10,
                               n_data_min = 50),
    "Missing some or all entries"
  )

  expect_message2(
    t3 <- util_plot_categorical_vars(resp_vars = "v00103",
                                 group_vars = "v00012",
                                 study_data = study_data, meta_data = meta_data,
                                 n_cat_max = 10,
                                 n_group_max = 5),
    "Missing some or all entries"
  )

  expect_message2(
    t4 <- util_plot_categorical_vars(resp_vars = "v00103",
                                 study_data = study_data, meta_data = meta_data,
                                 n_cat_max = 10),
    "Missing some or all entries"
  )
})

test_that("util_plot_categorical_vars works for nominal and ordinal variables", {
  skip_on_cran() # slow
  skip_if_translated()

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  require_english_locale_and_berlin_tz()

  meta_data[[SCALE_LEVEL]][meta_data[["VAR_NAMES"]] == "v00008"] <-
    SCALE_LEVELS$NOMINAL
  expect_message2(
    t5 <- util_plot_categorical_vars(resp_vars = "v00008",
                                     group_vars = "v00011",
                                     study_data = study_data, meta_data = meta_data,
                                     n_cat_max = 10,
                                     n_group_max = 5),
    "Missing some or all entries"
  )

  meta_data[[SCALE_LEVEL]][meta_data[["VAR_NAMES"]] == "v00008"] <-
    SCALE_LEVELS$ORDINAL
  expect_message2(
    t6 <- util_plot_categorical_vars(resp_vars = "v00008",
                                     group_vars = "v00011",
                                     study_data = study_data, meta_data = meta_data,
                                     n_cat_max = 10,
                                     n_group_max = 5),
    "Missing some or all entries"
  )

})

#' Wrapper function to check for segment data structure
#'
#' @description
#' This function tests for unexpected elements and records, as well as duplicated identifiers and content.
#' The unexpected element record check can be conducted by providing the number of expected records or
#' an additional table with the expected records.
#' It is possible to conduct the checks by study segments or to consider only selected
#' segments.
#'
#' [Indicator]
#'
#' @inheritParams .template_function_indicator
#'
#' @param meta_data_segment [data.frame] the data frame that contains the metadata for the segment level, mandatory
#' @param segment_level [data.frame] alias for `meta_data_segment`
#'
#' @return a [list] with
#'   - `SegmentTable`: data frame with selected check results, used for the data quality report.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' out_segment <- int_all_datastructure_segment(
#'   meta_data_segment = "meta_data_segment",
#'   study_data = "ship",
#'   meta_data = "ship_meta"
#' )
#'
#' study_data <- cars
#' meta_data <- dataquieR::prep_create_meta(VAR_NAMES = c("speedx", "distx"),
#'   DATA_TYPE = c("integer", "integer"), MISSING_LIST = "|", JUMP_LIST = "|",
#'   STUDY_SEGMENT = c("Intro", "Ex"))
#'
#' out_segment <- int_all_datastructure_segment(
#'   meta_data_segment = "meta_data_segment",
#'   study_data = study_data,
#'   meta_data = meta_data
#' )
#' }
int_all_datastructure_segment <- function(study_data,
                                          label_col,
                                          item_level = "item_level",
                                          meta_data = item_level,
                                          meta_data_v2,
                                          segment_level,
                                          meta_data_segment = "segment_level") {

  # Preps and checks ----
  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()

  # map metadata to study data
  prep_prepare_dataframes(.allow_empty = TRUE)
  if (!(STUDY_SEGMENT %in% colnames(meta_data))) {
    meta_data[[STUDY_SEGMENT]] <- "ALL" # TODO: Warn
  }
  meta_data_segment <- prep_check_meta_data_segment(meta_data_segment)

  id_vars_list <- lapply(setNames(meta_data_segment[[SEGMENT_ID_VARS]], # TODO: use the constants everywhere: meta_data_segment[[SEGMENT_ID_VARS]], not ...$SEGMENT_ID_VARS
                                  nm = meta_data_segment[[STUDY_SEGMENT]]),
                         util_parse_assignments,
                         multi_variate_text = TRUE
  )
  id_vars_list_vector <- lapply(id_vars_list, unlist, recursive = TRUE)

  id_vars_list_vector <- lapply(id_vars_list_vector,
                                util_map_labels,
                                meta_data = meta_data,
                                to = label_col)

  # 1. Unexpected data record count ----

  # subset metadata with entries
  meta_data_record_count_0 <-
    meta_data_segment[!util_empty(meta_data_segment[[SEGMENT_RECORD_COUNT]]), ,
                      drop = FALSE]

  unexp_records_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    unexp_records_out <- withr::with_options(list(
      dataquieR.testdebug = TRUE), int_unexp_records_segment(
        study_segment = meta_data_record_count_0[[STUDY_SEGMENT]],
        data_record_count = meta_data_record_count_0[[SEGMENT_RECORD_COUNT]],
        study_data = study_data, meta_data = meta_data, label_col = label_col))
  })


  # 2. Unexpected data record set ----
  # subset metadata with entries
  meta_data_record_set_1 <-
    meta_data_segment[!util_empty(meta_data_segment[[SEGMENT_RECORD_CHECK]]), ,
                      drop = FALSE
    ]

  unexp_records_id_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    unexp_records_id_out <- util_int_unexp_records_set_segment(
      id_vars_list =
        id_vars_list_vector[meta_data_record_set_1[[STUDY_SEGMENT]]],
      identifier_name_list = meta_data_record_set_1[[STUDY_SEGMENT]],
      valid_id_table_list = meta_data_record_set_1[[SEGMENT_ID_REF_TABLE]],
      meta_data_record_check_list =
        meta_data_record_set_1[[SEGMENT_RECORD_CHECK]],
      study_data = study_data,
      label_col = label_col,
      meta_data = meta_data)
  })


  # 3. Duplicates: ids ----
  meta_data_dup_ids_1 <-
    meta_data_segment[!util_empty(meta_data_segment[[SEGMENT_ID_VARS]]), ,
                      drop = FALSE
    ]

  duplicate_ids_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {

    duplicate_ids_out <- withr::with_options(list(
      dataquieR.testdebug = TRUE), int_duplicate_ids(
        level = "segment",
        id_vars_list = id_vars_list_vector[meta_data_dup_ids_1[[STUDY_SEGMENT]]],
        study_segment = meta_data_dup_ids_1[[STUDY_SEGMENT]],
        repetitions = meta_data_dup_ids_1[[SEGMENT_UNIQUE_ID]],
        study_data = study_data,
        meta_data = meta_data,
        label_col = label_col
      ))})


  # 4. Duplicates: content ----
  meta_data_dup_rows_1 <-
    meta_data_segment[!util_empty(meta_data_segment[[SEGMENT_UNIQUE_ROWS]]), ,
                      drop = FALSE
    ]

  meta_data_dup_rows_1 <-
    meta_data_dup_rows_1[
      trimws(tolower(meta_data_dup_rows_1[[SEGMENT_UNIQUE_ROWS]])) ==
        "no_id" |
        !util_is_na_0_empty_or_false(
          meta_data_dup_rows_1[[SEGMENT_UNIQUE_ROWS]]), ,
      drop = FALSE
    ]

  duplicate_rows_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    duplicate_rows_out <- withr::with_options(list(
      dataquieR.testdebug = TRUE), int_duplicate_content(
        level = "segment",
        identifier_name_list = meta_data_dup_rows_1[[STUDY_SEGMENT]],
        study_data = study_data,
        meta_data = meta_data,
        label_col = label_col,
        id_vars_list = id_vars_list_vector[meta_data_dup_rows_1[[STUDY_SEGMENT]]],
        unique_rows = setNames(meta_data_dup_rows_1[[SEGMENT_UNIQUE_ROWS]],
                               nm = meta_data_dup_rows_1[[STUDY_SEGMENT]])
      ))
  })

  # X. Unexpected data element set ----

  out_int_sts_element <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    out_int_sts_element <-
      withr::with_options(list(
        dataquieR.testdebug = TRUE), int_sts_element_segment(study_data = study_data,
                                                             label_col = label_col,
                                                             meta_data = meta_data)$SegmentTable)

    out_int_sts_element$GRADING <-
      ifelse(out_int_sts_element$NUM_int_sts_element == 0, 0, 1)
    rownames(out_int_sts_element) <- NULL
  })

  # Output ----
  result <- list(
    int_sts_countre = unexp_records_out$SegmentTable,
    int_sts_setrc = unexp_records_id_out$SegmentTable,
    int_sts_dupl_ids = duplicate_ids_out$SegmentTable,
    int_sts_dupl_content = duplicate_rows_out$SegmentTable,
    int_sts_element = out_int_sts_element
  )

  result <- result[vapply(result, FUN.VALUE = logical(1), # TODO: Why remove empty seg-dfs, and why not in int_all_datastructure_dataframe?!
                          FUN = function(df) {
                            !!prod(dim(df))
                          })]

  out_int_sts_elementData <- out_int_sts_element
  # (both already keyed by "Segment"; no DF_NAME swap needed here)

  # Translate indicator metric-like columns via helper; keep unknowns (e.g., "Segment") unchanged
  cn <- colnames(out_int_sts_elementData)
  if (length(cn) > 0) {
    colnames(out_int_sts_elementData) <- util_translate_indicator_metrics(
      cn,
      short = FALSE,
      long  = TRUE,
      ignore_unknown = TRUE
    )
  }

  resultData <- list(
    int_sts_countre = unexp_records_out$SegmentData,
    int_sts_setrc = unexp_records_id_out$SegmentData,
    int_sts_dupl_ids = duplicate_ids_out$SegmentData,
    int_sts_dupl_content = duplicate_rows_out$SegmentData,
    int_sts_element = out_int_sts_elementData
  )

  for (n in names(resultData)) {
    rownames(resultData[[n]]) <- NULL
  }

  dqi <- util_get_concept_info("dqi")
  dqi <- dqi[!util_empty(dqi$abbreviation) & !util_empty(dqi$Name), , FALSE]

  names(resultData) <-
    util_recode(
      names(resultData),
      dqi,
      "abbreviation",
      "Name",
      names(resultData)
    )

  SegmentTable <- util_merge_data_frame_list(result, "Segment") # TODO: why Segment and not STUDY_SEGMENT?!
  cn <- colnames(SegmentTable)
  cn[startsWith(cn, "GRADING.")] <- gsub("^GRADING\\.", "GRADING_",
                                         cn[startsWith(cn, "GRADING.")])
  colnames(SegmentTable) <- cn

  # SegmntData <- util_merge_data_frame_list(resultData, "Segment")

  SegmentData1 <- util_make_data_slot_from_table_slot(SegmentTable)
  if ("resp_vars" %in% colnames(SegmentTable)) {
    SegmentData1$`Unexp. Variables` <- SegmentTable$resp_vars
  }

  # Create the df
  SegmentData <- data.frame(Segment = SegmentData1$Segment)
  # Only if content is present, merge columns
  if(!is.null(SegmentData1$`Unexpected data record count (Number)`)){
    SegmentData$`Unexpected data record count N (%)` <- util_paste0_with_na(SegmentData1$`Unexpected data record count (Number)`, " (",
                                                                            SegmentData1$`Unexpected data record count (Percentage (0 to 100))`, ")")
  }
  SegmentData$`Unexpected data record count (Grading)`<- SegmentData1$`Unexpected data record count (Grading)`

  if(!is.null(SegmentData1$`Unexpected data record set (Number)`)){
    SegmentData$`Unexpected data record set N (%)` <- util_paste0_with_na(SegmentData1$`Unexpected data record set (Number)`, " (",
                                                                          SegmentData1$`Unexpected data record set (Percentage (0 to 100))`, ")")
  }
  SegmentData$`Unexpected data record set (Grading)`<- SegmentData1$`Unexpected data record set (Grading)`

  if(!is.null(SegmentData1$`Duplicates (Number)`)){
    SegmentData$`Duplicates N (%)` <- util_paste0_with_na(SegmentData1$`Duplicates (Number)`, " (",
                                                          SegmentData1$`Duplicates (Percentage (0 to 100))`, ")")
  }
  SegmentData$`Duplicates (Grading)`<- SegmentData1$`Duplicates (Grading)`

  if(!is.null(SegmentData1$`Unexp. Variables`)){
    SegmentData$`Unexp. Variables`<- SegmentData1$`Unexp. Variables`
  }

  if (!is.null(SegmentData1$`Unexpected data element set (Number)`)){
    SegmentData$`Unexpected data element set N (%)` <- util_paste0_with_na(SegmentData1$`Unexpected data element set (Number)`, " (",
                                                                           SegmentData1$`Unexpected data element set (Percentage (0 to 100))`, ")")
  }
  SegmentData$`Unexpected data element set (Grading)`<- SegmentData1$`Unexpected data element set (Grading)`

  rm(SegmentData1)

  #Add new attribute to the columns of SummaryData to define the datatype of each column
  strings_col <- c("Segment",
               "Unexpected data record count N (%)",
               "Unexpected data record set N (%)",
               "Duplicates N (%)",
               "Unexp. Variables",
               "Unexpected data element set N (%)")
  strings_col <- intersect(strings_col, colnames(SegmentData))

  integers_col <- c("Unexpected data record count (Grading)",
                "Unexpected data record set (Grading)",
                "Duplicates (Grading)",
                "Unexpected data element set (Grading)")
  integers_col<- intersect(integers_col, colnames(SegmentData))

  # Assign attributes in bulk
  if (length(strings_col) > 0 ) {
    SegmentData[strings_col] <- lapply(SegmentData[strings_col],
                                   function(x) {
                                     attr(x, DATA_TYPE) <- DATA_TYPES$STRING; x
                                   })
  }
  if (length(integers_col) > 0) {
    SegmentData[integers_col] <- lapply(SegmentData[integers_col],
                                    function(x) {
                                      attr(x, DATA_TYPE) <- DATA_TYPES$INTEGER; x
                                    })
  }

  return(list(
    SegmentTable = SegmentTable,
    SegmentData = SegmentData,
    SegmentDataList = resultData
  ))
}

#' Align column name case to meta_data VAR_NAMES
#'
#' Case-insensitively matches study-data column names to VAR_NAMES from
#' the meta_data and adjusts the capitalization of the column names so that
#' they match the corresponding VAR_NAMES.
#'
#' If there are case-insensitive duplicates in VAR_NAMES that differ only in
#' capitalization (e.g., "AGE" and "age"), the function stops with
#' util_error(..., applicability_error = TRUE).
#'
#' @param .colnames character vector of column names to be aligned.
#' @param .var_names character vector of VAR_NAMES from meta_data.
#'
#' @return character vector of the same length as .colnames, with
#'   capitalization aligned to .var_names where possible.
#' @noRd
util_align_colnames_case <- function(.colnames, .var_names) {
  .colnames  <- as.character(.colnames)
  .var_names <- as.character(.var_names)

  # check for case-insensitive duplicates in .var_names
  lv <- tolower(.var_names)

  if (any(duplicated(lv))) {
    # find all lower-case keys that appear multiple times
    dup_keys <- unique(lv[duplicated(lv)])

    # keep only those where the originals differ in case / spelling
    bad_keys <- dup_keys[
      vapply(
        dup_keys,
        function(k) {
          length(unique(.var_names[lv == k])) > 1L
        },
        logical(1L)
      )
    ]

    if (length(bad_keys) > 0L) {
      util_error(
        c("for case-insensitive study-data-column-names, I need",
        "case-insenstively distinct %s in the meta_data, but I have %s.",
        "either fix your meta_data or set the option %s to its default %s"),
        sQuote(VAR_NAMES),
        util_pretty_vector_string(bad_keys),
        sQuote("dataquieR.study_data_colnames_case_sensitive"),
        sQuote(dataquieR.study_data_colnames_case_sensitive_default),
        applicability_error = TRUE
      )
    }
  }

  # map .colnames (case-insensitive) to .var_names
  idx <- match(tolower(.colnames), lv)

  res <- .colnames
  hit <- !is.na(idx)
  res[hit] <- .var_names[idx[hit]]

  res
}

#' Check data for observer levels
#'
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param group_vars [variable] the name of the observer, device or reader
#'                              variable
#' @param min_obs_in_subgroup [integer] from=0. optional argument if
#'                                  `group_vars` are used. This argument
#'                                  specifies the minimum number of observations
#'                                  that is required to include a subgroup
#'                                   (level) of the group variable named by
#'                                    `group_vars` in the analysis. Subgroups
#'                                  with fewer observations are excluded.
#' @param max_obs_in_subgroup [integer] from=0. optional argument if
#'                                  `group_vars` are used. This argument
#'                                  specifies the maximum number of observations
#'                                  that is required to include a subgroup
#'                                   (level) of the group variable named by
#'                                    `group_vars` in the analysis. Subgroups
#'                                  with more observations are excluded.
#' @param min_subgroups [integer] from=0. optional argument if a "group_var" is
#'                                        used. This argument specifies the
#'                                        minimum no. of subgroups (levels)
#'                                        included "group_var". If the variable
#'                                        defined in "group_var" has fewer
#'                                        subgroups it is split for analysis.
#' @param max_subgroups [integer] from=0. optional argument if a "group_var" is
#'                                        used. This argument specifies the
#'                                        maximum no. of subgroups (levels)
#'                                        included "group_var". If the variable
#'                                        defined in "group_var" has more
#'                                        subgroups it is split for analysis.
#'
#' @return modified study data frame
#' @examples
#' \dontrun{
#' study_data <- prep_get_data_frame("study_data")
#' meta_data <- prep_get_data_frame("meta_data")
#' prep_prepare_dataframes(.label_col = LABEL)
#' util_check_group_levels(ds1, "CENTER_0")
#' dim(util_check_group_levels(ds1, "USR_BP_0", min_obs_in_subgroup = 400))
#' }
#'
#'
#' @seealso [prep_min_obs_level]
#' @family data_management
#' @concept robustness
#' @noRd
util_check_group_levels <- function(study_data,
                                    group_vars,
                                    min_obs_in_subgroup = -Inf,
                                    max_obs_in_subgroup = +Inf,
                                    min_subgroups = -Inf,
                                    max_subgroups = +Inf) {
  # util_correct_variable_use(group_vars) cannot work, here, should be called in the calling function, anyway
  util_expect_scalar(min_obs_in_subgroup, check_type = is.numeric)
  util_expect_scalar(max_obs_in_subgroup, check_type = is.numeric)
  util_expect_scalar(min_subgroups, check_type = is.numeric)
  util_expect_scalar(max_subgroups, check_type = is.numeric)
  util_stop_if_not(max_obs_in_subgroup >= min_obs_in_subgroup)
  util_stop_if_not(max_subgroups >= min_subgroups)
  l <- split(study_data, as.factor(study_data[, group_vars]))
  obs_per_subgroup <- vapply(l, nrow, FUN.VALUE = integer(1))
  n_subgroups <- length(l)
  if (is.finite(min_obs_in_subgroup)) {
    which_too_few <- which(obs_per_subgroup < min_obs_in_subgroup)
  } else {
    which_too_few <- integer(0)
  }
  if (is.finite(max_obs_in_subgroup)) {
    which_too_many <- which(obs_per_subgroup > max_obs_in_subgroup)
  } else {
    which_too_many <- integer(0)
  }
  which_valid <- setdiff(
    seq_along(obs_per_subgroup),
    union(which_too_few, which_too_many)
  )
  n_subgroups <- length(which_valid)

  modified_study_data <- do.call(rbind.data.frame, l[which_valid])
  if (length(which_valid) < length(l)) {
    util_message(
      c("Discarding %d observations (%d from %d levels of %s)",
        "because of too few/many observations",
        "per group"),
      nrow(study_data) - nrow(modified_study_data),
      length(l) - length(which_valid),
      length(l),
      dQuote(group_vars)
    )
  }

  if (is.finite(min_subgroups) && n_subgroups < min_subgroups) {
    util_error("Too few subgroups (%d < %d)", n_subgroups, min_subgroups)
  }
  if (is.finite(max_subgroups) && n_subgroups > max_subgroups) {
    attr(modified_study_data, "TOO_MANY") <-  TRUE
    return(modified_study_data) # TODO: SPLIT artificially
  }
  attr(modified_study_data, "TOO_MANY") <-  FALSE
  attributes(modified_study_data)[.ds1_attribute_names] <-
    attributes(study_data)[.ds1_attribute_names]
  modified_study_data
}

test_that("util_check_group_levels works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  f <- function(study_data, meta_data = "item_level", label_col) {
    ds1 <- prep_prepare_dataframes(
      .study_data = study_data,
      .meta_data = meta_data,
      .label_col = LABEL,
      .internal = TRUE)

    dim(util_check_group_levels(ds1, "CENTER_0"))

    expect_equal(
      nrow(util_check_group_levels(ds1, "USR_BP_0", min_obs_in_subgroup = 400)),
      448)
  }
  environment(f) <- environment(acc_margins)
  expect_message2(f(
    study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
    label_col = LABEL
  ))
})

#' Check, if `x` contains valid missing codes
#'
#' @param x a vector of missing codes
#'
#' @return a missing code not in `x`
#' @family metadata_management
#' @noRd
util_find_free_missing_code <- function(x) {
  x <- lapply(util_as_valid_missing_codes(x), as.character)
  if (all(suppressWarnings(is.na(x) ==
                                is.na(lapply(x, util_parse_date))))) {
    return(as.character(
      suppressWarnings(max(util_parse_date(x), na.rm = TRUE)) +
        lubridate::days(1)))
  } else if (all(suppressWarnings(is.na(x) ==
                                   is.na(lapply(x, util_parse_time))))) {
    return(util_as_character(
      hms::as_hms(
      suppressWarnings(max(util_parse_time(x), na.rm = TRUE)) +
        hms::hms(hours = 1))))
  } else if (all(suppressWarnings(is.na(x) == is.na(as.numeric(x))))) {
    return(as.character(suppressWarnings(max(as.numeric(x), na.rm = TRUE)) + 1))
  } else { # fallback to datetime
    return(suppressWarnings(do.call("max", c(lapply(x, util_parse_date),
                                list(na.rm = TRUE)))) +
      lubridate::days(1))
  }
}

test_that("util_find_free_missing_code works", {
  skip_on_cran()
  # Numeric codes
  x <- c(777,888,999)
  expected_next_code <-  1000
  expect_equal(util_find_free_missing_code(x), as.character(expected_next_code))

  # All values cannot be converted
  x <- c("abc", "def", NA, "ghi")
  expect_error(util_find_free_missing_code(x))

  # TODO: date-time values
})

#' Generate a RMarkdown-based report from a [dataquieR] report
#'
#'
#' @param ... deprecated
#'
#' @return deprecated
#' @export
#' @importFrom utils browseURL
print.dataquieR_resultset <- function(...) {
  lifecycle::deprecate_stop("2.1.0", what = "dq_report()", with = "dq_report2()")
}



# rstudioapi::jobRunScript()

#' Get Access to Utility Functions
#'
#' `r lifecycle::badge("experimental")`
#'
#' @param fkt function name
#' @param version version number to get
#'
#' @return an API object
#' @export
#'
#' @keywords internal
.get_internal_api <- function(fkt, version = API_VERSION, or_newer = TRUE) {
  # TODO: more life-cycle stuff, also for deprecation
  version <- try(as.package_version(version), silent = TRUE)
  if (util_is_try_error(version)) {
    util_error("You need to pass a version number in argument %s",
               sQuote("version"))
  }
  util_expect_scalar(or_newer, check_type = is.logical)
  if (!or_newer) {
    util_stop_if_not(version == API_VERSION)
  } else {
    util_stop_if_not(version <= API_VERSION)
  }
  f <- substitute(fkt)
  if (is.symbol(f)) {
    fkt <- as.character(f)
  }
  util_stop_if_not(is.character(fkt))
  util_stop_if_not(exists(fkt, mode = "function"))
  if (missing(version)) {
    util_warning(c("Hint: As a developer, you should not omit",
                   "the API version, when you access the internal API of %s,",
                   "write %s"),
                  sQuote(utils::packageName()),
                  sQuote(paste0(".get_internal_api(", fkt, ", version=\"",
                                API_VERSION, "\")"))
                 )
  }
  get(fkt, mode = "function")
}


#' Version of the API
#'
#' @seealso `.get_internal_api()`
#'
#' @export
#' @keywords internal
API_VERSION <- as.package_version("0.0.2")

test_that(".get_internal_api works", {
  skip_on_cran() # internal use, only
  expect_true(".get_internal_api" %in%
    getNamespaceExports("dataquieR"))
  x <- .get_internal_api("util_html_table", version = "0.0.1", TRUE)
  expect_identical(x, util_html_table)
  expect_error(.get_internal_api("util_html_table", version = "0.0.1", FALSE))
  expect_error(.get_internal_api("util_html_table", version = "10", FALSE))
  expect_error(.get_internal_api("util_html_table", version = "10", TRUE))
})

#' Data type conversion
#'
#' Utility function to convert a study variable to match the data type given in
#' the metadata, if possible.
#'
#' @param x the value
#' @param type expected data type
#'
#' @return the transformed values (if possible)
#'
#' @noRd
util_data_type_conversion <- function(x, type) {
  if (is.factor(x)) {
    if (isTRUE(as.logical(getOption("dataquieR.old_factor_handling",
                                    dataquieR.old_factor_handling_default)))) {
      # This is for backwards compatibility, but it may be more user friendly to
      # omit this in both adjust_data_type functions
      # noop
    } else {
      x <- as.character(x)
    }
  }
  # look-up list about how to convert
  converts <- setNames(
    list(
      function(x) {
        if (identical(attr(type, "orig_type"), "logical")) {
          as.integer(as.logical(x))
        } else {
          floor(as.numeric(x))
        }
      },
      # as.integer may fail for too large integer numbers
      as.numeric,
      as.character,
      util_parse_date,
      util_as_time_only
    ),
    nm = c(
      DATA_TYPES$INTEGER,
      DATA_TYPES$FLOAT,
      DATA_TYPES$STRING,
      DATA_TYPES$DATETIME,
      DATA_TYPES$TIME
    )
  )

  # function to convert to the matching data type
  .as <- try(converts[[type]], silent = TRUE)
  if (length(.as) != 1 || inherits(.as, "try-error")) {
    util_error("%s is not a known data type.", dQuote(type),
               applicability_problem = TRUE)
  }
  # perform data type conversion
  x2 <- suppressWarnings(do.call(.as, list(x)))

  return(x2)
}

#' Verify the data type of a value
#'
#' Function to verify the data type of a value.
#'
#' @param x the value
#' @param type expected data type
#' @param check_convertible [logical] also try, if a conversion to the
#'                                    declared data type would work.
#' @param threshold_value [numeric] from=0 to=100. percentage of failing
#'                                  conversions allowed.
#' @param return_percentages [logical] return the percentage of
#'                                mismatches.
#' @param check_conversion_stable [logical] do not distinguish convertible
#'                                          from convertible, but with issues
#' @param robust_na [logical] treat white-space-only-values as `NA`
#' @param vname [character] variable name being checked -- for error messages.
#'
#' @return if `return_percentages`: if not `check_convertible`, the percentage
#'                                of mismatches instead of logical value,
#'                                if `check_convertible`, return a named
#'                                vector with the percentages of all cases
#'                                (names of the vector are
#'                                `match`, `convertible_mismatch_stable`,
#'                                `convertible_mismatch_unstable`,
#'                                `nonconvertible_mismatch`)
#'         if not `return_percentages`: if `check_convertible` is `FALSE`,
#'         [logical] whether `x` is of the expected type
#'         if `check_convertible` is `TRUE`
#'         [integer] with the states `0, 1, 2, 3`: 0 = Mismatch, not convertible
#'                                                 1 = Match
#'                                                 2 = Mismatch, but convertible
#'                                                 3 = Mismatch, convertible,
#'                                                     but with issues (e.g.,
#'                                                     loss of decimal places)
#' @importFrom stats setNames
#'
#' @family data_management
#' @concept data_management
#' @noRd
util_check_data_type <- function(x, type, check_convertible = FALSE,
                                 threshold_value = 0, return_percentages =
                                   FALSE, check_conversion_stable = FALSE,
                                 robust_na = FALSE,
                                 vname = "data") {
  # FIXME: SLOW!!
  hash_id <-
    rlang::hash(list(x,
                type,
                check_convertible,
                check_conversion_stable,
                threshold_value,
                robust_na,
                return_percentages))
  # HINT: if dep on dataframe env: as.list(.dataframe_environment()), also remind
  # global options(). Also remind possible side effects
  if (exists(hash_id, .cache[[".cache"]])) {
    return(get(hash_id, .cache[[".cache"]]))
  }
  if (is.list(x) &&
      any(not_1 <- vapply(x, length, FUN.VALUE = integer(1)) != 1)) {
    varname <- NULL
    if (!missing(vname)) {
      varname <- vname
      vname <- dQuote(vname)
    }
    util_message(
      "Found %d inconsistent data values in %s (length <> 1). Removed.",
      sum(not_1),
      vname,
      integrity_indicator = "int_vfe_inhom",
      varname = varname
      )
    x[not_1] <- NA
  }
  if (robust_na) {
    empty <- util_empty
  } else {
    empty <- is.na
  }
  checks <- setNames(
    list(
      util_is_integer,
      is.numeric,
      is.character,
      lubridate::is.timepoint,
      util_is_time_only
    ),
    nm = c(
      DATA_TYPES$INTEGER,
      DATA_TYPES$FLOAT,
      DATA_TYPES$STRING,
      DATA_TYPES$DATETIME,
      DATA_TYPES$TIME
    )
  )

  # function to check whether the data type matches
  .is <- try(checks[[type]], silent = TRUE)
  if (length(.is) != 1 || inherits(.is, "try-error")) {
    util_error("%s is not a known data type.", dQuote(type),
               applicability_problem = TRUE)
  }
  # function to check whether the data type matches (disregarding `NA`s)
  .is_or_na <- function(...) all(empty(...) | .is(...))
  mismatches <- !vapply(x, .is_or_na, FUN.VALUE = logical(1))

  if (length(mismatches) < 1) {
    pct_mismatches <- 0
  } else {
    pct_mismatches <- sum(mismatches) / length(mismatches) * 100
    attr(pct_mismatches, "which") <- mismatches
  }


  if (check_convertible) {
    # perform data type conversion
    x2 <- util_data_type_conversion(x, type)

    if (check_conversion_stable) {
      stable <- util_conversion_stable(x, type)
    } else {
      stable <- rep(TRUE, length(x))
    }

    convertible_mismatch_all <-
      mismatches &
      vapply(x, empty, FUN.VALUE = logical(1)) ==
        vapply(x2, empty, FUN.VALUE = logical(1))

    convertible_mismatch_unstable <- convertible_mismatch_all & !stable
    convertible_mismatch_stable <- convertible_mismatch_all & stable


    pct_convertible_mismatch_stable <-
      sum(convertible_mismatch_stable) /
      length(convertible_mismatch_stable) * 100
    attr(pct_convertible_mismatch_stable, "which") <-
      convertible_mismatch_stable

    pct_convertible_mismatch_unstable <-
      sum(convertible_mismatch_unstable) /
      length(convertible_mismatch_unstable) * 100
    attr(pct_convertible_mismatch_unstable, "which") <-
      convertible_mismatch_unstable

    nonconvertible_mismatch <-
      vapply(x, empty, FUN.VALUE = logical(1)) !=
      vapply(x2, empty, FUN.VALUE = logical(1))
    if (length(nonconvertible_mismatch) < 1) {
      pct_nonconvertible_mismatch <- 0
    } else {
      pct_nonconvertible_mismatch <-
        sum(nonconvertible_mismatch) /
        length(nonconvertible_mismatch) * 100
    }
    attr(pct_nonconvertible_mismatch, "which") <-
      nonconvertible_mismatch

    if (return_percentages) {
      result <- c(
        match = 100 - pct_mismatches,
        convertible_mismatch_stable = pct_convertible_mismatch_stable,
        convertible_mismatch_unstable = pct_convertible_mismatch_unstable,
        nonconvertible_mismatch = pct_nonconvertible_mismatch
      )
      attr(result, "which") <-
        list(
          match = attr(pct_mismatches, "which"),
          convertible_mismatch_stable = attr(pct_convertible_mismatch_stable, "which"),
          convertible_mismatch_unstable = attr(pct_convertible_mismatch_unstable, "which"),
          nonconvertible_mismatch = attr(pct_nonconvertible_mismatch, "which")
        )
    } else {
      if (pct_nonconvertible_mismatch > threshold_value)
        { # NB: Order does matter
        # 0 = Mismatch, not convertible
        result <- 0L
      } else if (pct_mismatches <= threshold_value) {
        # 1 = Match
        result <- 1L
      } else if (pct_nonconvertible_mismatch <= threshold_value) {
        # 2 = Mismatch, but convertible
        result <- 2L
      } else {
        util_error(
          "Internal error in util_check_data_type, sorry, and please report")
      }
    }
  } else {
    if (return_percentages) {
      result <- pct_mismatches
    } else {
      result <- pct_mismatches <= threshold_value
    }
  }

  assign(hash_id, value = result, envir = .cache[[".cache"]])

  return(result)
}
.cache <- new.env(parent = emptyenv()) # HINT: This is exported by dq_reoprt2, but not yet for parallel rendering, which is complicated an not fully working anyways.
util_reset_cache <- function() {
  assign(x = ".cache", value = new.env(parent = emptyenv()), envir = .cache)
}
util_reset_cache()

test_that("util_check_data_type works", {
  skip_on_cran()
  expect_true(util_check_data_type(42, DATA_TYPES$INTEGER))
  expect_true(util_check_data_type(1:42, DATA_TYPES$INTEGER))
  expect_true(util_check_data_type(NA, DATA_TYPES$INTEGER))
  expect_false(util_check_data_type(c(TRUE, NA), DATA_TYPES$INTEGER))
  expect_true(util_check_data_type(c("text", NA), DATA_TYPES$STRING))
  expect_false(util_check_data_type(c(42.5, as.integer(12)),
                                    DATA_TYPES$INTEGER))
  expect_true(util_check_data_type(c(42.5, as.integer(12)), DATA_TYPES$FLOAT))
  expect_true(util_check_data_type(c(Sys.time(), as.integer(12)),
                                   DATA_TYPES$DATETIME))
  expect_false(util_check_data_type(c(as.integer(Sys.time()), as.integer(12)),
                                    DATA_TYPES$DATETIME))
  expect_false(util_check_data_type("2020-11-11 17:10:07 CET",
                                    DATA_TYPES$DATETIME))
  expect_true(util_check_data_type(as.POSIXct("2020-11-11 17:10:07 CET"),
                                   DATA_TYPES$DATETIME))
  expect_error(util_check_data_type(as.POSIXct("2020-11-11 17:10:07 CET"),
                                   "xyz"),
               regexp = ".+xyz.+ is not a known data type.")

  # 0 = Mismatch, not convertible
  # 1 = Match
  # 2 = Mismatch, but convertible

  expect_true(all(2 == util_check_data_type("2020-11-11 17:10:07 CET",
                                    DATA_TYPES$DATETIME,
                                    check_convertible = TRUE)))
  expect_true(all(0 == util_check_data_type("not a date",
                                            DATA_TYPES$DATETIME,
                                            check_convertible = TRUE)))

  expect_equal(util_check_data_type(42, DATA_TYPES$INTEGER,
                                   return_percentages = TRUE),
               util_attach_attr(0, which = FALSE))
  expect_equal(util_check_data_type(42.5, DATA_TYPES$INTEGER,
                                   return_percentages = TRUE),
               util_attach_attr(100, which = TRUE))
  expect_equal(util_check_data_type(c("42", "x"), DATA_TYPES$INTEGER,
                                    return_percentages = TRUE),
               util_attach_attr(100, which = c(`42` = TRUE, x = TRUE)))
  expect_equal(util_check_data_type(c("42", "x"), DATA_TYPES$INTEGER,
                                    check_convertible = TRUE,
                                    return_percentages = FALSE), 0)
  expect_equal(util_check_data_type(c(42, 2), DATA_TYPES$INTEGER,
                                    check_convertible = TRUE,
                                    return_percentages = FALSE), 1)
  expect_equal(util_check_data_type(c("42", "2"), DATA_TYPES$INTEGER,
                                    check_convertible = TRUE,
                                    return_percentages = FALSE), 2)
  expect_equal(util_check_data_type(c(42, 42.6, 42L), DATA_TYPES$INTEGER,
                                    check_convertible = TRUE,
                                    check_conversion_stable = TRUE,
                                    return_percentages = TRUE),
                   util_attach_attr(c(match = 200/3,
                     convertible_mismatch_stable = 0,
                     convertible_mismatch_unstable = 100/3,
                     nonconvertible_mismatch = 0), which = list(
                       match = c(FALSE, TRUE, FALSE),
                       convertible_mismatch_stable = c(FALSE, FALSE, FALSE),
                       convertible_mismatch_unstable = c(FALSE, TRUE, FALSE),
                       nonconvertible_mismatch = c(FALSE, FALSE, FALSE)
                     )))

})

#' Format numeric values with adaptive rounding
#'
#' Round numbers with a reasonable number of decimal places depending on magnitude.
#' Uses fixed notation for typical ranges and switches to scientific notation
#' for very small or large values.
#'
#' @param x a numeric vector to be rounded
#' @param digits a numeric value indicating the number of significant digits
#'
#' @return [character] formatted values
#'
#' @family data_management
#' @concept data_management
#' @noRd
util_round_to_decimal_places <- function (x, digits = 3) {

  util_expect_scalar(arg_name = x, allow_more_than_one = TRUE,
                     allow_null = TRUE, allow_na = TRUE,
                     check_type = is.numeric)

  util_expect_scalar(arg_name = digits, allow_more_than_one = FALSE,
                     allow_null = FALSE, allow_na = FALSE,
                     check_type = is.numeric)

  if (length(x) == 0) return(character(0))

  y_abs <- abs(x)
  finite <- is.finite(y_abs) & y_abs > 0

  use_sci <- FALSE
  if (any(finite)) {
    use_sci <- suppressWarnings(
      min(y_abs[finite]) < 0.001 ||
        max(y_abs[finite]) > 9999.999
    )
  }

  if (use_sci) {
    # minimal change: keep formatC but ensure consistent output
    y <- formatC(x, format = "e", digits = digits)
  } else {
    # NEW: use signif instead of round for better generic behaviour
    s <- signif(x, digits = digits)

    # limit decimal places to avoid long tails (git-friendly small change)
    s <- round(s, digits)

    # avoid scientific notation + trim zeros
    y <- format(s, trim = TRUE, scientific = FALSE)
  }

  return(y)
}
# TODO: Discuss
# guess_digits <- function(x, digs = getOption("digits", 4)) {
#   max(nchar(zapsmall(abs(x - floor(x)), digits = digs)) - 2, 0)
# }
#
# round(vector, xtable::xdigits(vector))
# or format() -- this is, what R uses
# format(3.66666654, digits = 2)

test_that("acc_margins works without label_col", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  expect_error({
    res1 <-
      acc_margins(resp_vars = "v00001", study_data = study_data,
                  meta_data = meta_data)
  },
  regexp = paste(".*Argument .+resp_vars.+: Variable .+v00001.+ \\(na\\)",
                 "has a disallowed scale level \\(.+na.+\\)"),
  perl = TRUE
  )

  expect_error( # float in group_vars
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "CRP_0",
                label_col = LABEL),
    regexp = paste("Argument .+group_vars.+: Variable .+CRP_0.+",
                   "\\(float\\) has a disallowed type \\(.+float.+\\)"),
    perl = TRUE
  )

  expect_error( # wrong group_vars
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "not_avail",
                label_col = LABEL),
    regexp = paste("Variable .+not_avail.+ \\(group_vars\\) not found in",
                   "metadata. Did you mean .+ASTHMA_0.+\\?"),
    perl = TRUE
  )

  expect_error( # no group_vars
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = NULL,
                label_col = LABEL),
    regexp = paste("Argument group_vars is NULL"),
    perl = TRUE
  )

  expect_error( # many group_vars
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = c("DEV_NO_0", "USR_BP_0"),
                label_col = LABEL),
    regexp = paste("Need exactly one element in argument group_vars,",
                   "got 2: .DEV_NO_0, USR_BP_0."),
    perl = TRUE
  )

  expect_message2(
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "SEX_0",
                label_col = LABEL),
    regexp = paste("Due to missing values in SEX_0, N = 60 observations",
                   "were excluded. Due to missing values in CRP_0, N = 241",
                   "observations were excluded additionally."),
    perl = TRUE
  )

  expect_message2( # float
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                label_col = LABEL)
  )

  expect_message2( # float util_margins_lm
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                label_col = LABEL, threshold_type = "none")
  )

  expect_message2( # float
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                threshold_value = data.frame(l = letters, L = LETTERS),
                label_col = LABEL),
    regexp = "threshold_value is not numeric.1.: .+,",
    perl = TRUE,
    all = FALSE
  )
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  md <- prep_get_data_frame("item_level")

  # test function for different scale levels, with and without sorting
  # 1. nominal (dichotomized or not)
  md$SCALE_LEVEL[md$LABEL == "MEAT_CONS_0"] <- SCALE_LEVELS$NOMINAL
  expect_message2( #util_margins_bin
    res2 <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = study_data,
                        meta_data = md, group_vars = "CENTER_0",
                        label_col = LABEL, dichotomize_categorical_resp = TRUE)
  )
  expect_false(
    inherits(try(ggplot_build(res2$SummaryPlot)), "try-error"))
  expect_message2(  #util_margins_bin
    res2_new <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = study_data,
                        meta_data = md, group_vars = "CENTER_0",
                        label_col = LABEL, dichotomize_categorical_resp = TRUE,
                        threshold_type = "none")
  )
  expect_message2(
    res2b <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = study_data,
                         meta_data = md, group_vars = "CENTER_0",
                         label_col = LABEL, dichotomize_categorical_resp = TRUE,
                         sort_group_var_levels = FALSE)
  )
  expect_false(
    inherits(try(ggplot_build(res2b$SummaryPlot)), "try-error"))

  #util_margins_bin binary
  expect_message2(
    res2b_bin <- acc_margins(resp_vars = "PREGNANT_0", study_data = study_data,
                         meta_data = md, group_vars = "USR_SOCDEM_0",
                         label_col = LABEL)
  )

  # should also work if some levels of the grouping variable did detect only
  # a single category
  sd1 <- study_data
  sd1[which(sd1[, md$VAR_NAMES[md$LABEL == "CENTER_0"]] == 1),
      md$VAR_NAMES[md$LABEL == "MEAT_CONS_0"]] <- 0
  sd1[which(sd1[, md$VAR_NAMES[md$LABEL == "CENTER_0"]] == 4),
      md$VAR_NAMES[md$LABEL == "MEAT_CONS_0"]] <- 1
  expect_message2( #util_margins_bin
    res2c <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = sd1,
                         meta_data = md, group_vars = "CENTER_0",
                         label_col = LABEL, dichotomize_categorical_resp = TRUE)
  )
  expect_false(
    inherits(try(ggplot_build(res2c$SummaryPlot)), "try-error"))

  # variant if the response variable is not being dichotomized
  res2d <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = study_data,  #util_margins_nom
                       meta_data = md, group_vars = "CENTER_0",
                       label_col = LABEL, dichotomize_categorical_resp = FALSE)
  expect_false(
    inherits(try(ggplot_build(res2d$SummaryPlot)), "try-error"))
  res2e <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = study_data,
                       meta_data = md, group_vars = "CENTER_0",
                       label_col = LABEL, dichotomize_categorical_resp = FALSE,
                       sort_group_var_levels = FALSE)
  expect_false(
    inherits(try(ggplot_build(res2e$SummaryPlot)), "try-error"))

  # 2. ordinal (linear model or ordinal regression)
  md$SCALE_LEVEL[md$LABEL == "MEAT_CONS_0"] <- SCALE_LEVELS$ORDINAL
  expect_message2( #util_margins_lm
    res3 <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = study_data,
                        meta_data = md, group_vars = "CENTER_0",
                        label_col = LABEL, sort_group_var_levels = TRUE,
                        include_numbers_in_figures = TRUE)
  )
  expect_false(
    inherits(try(ggplot_build(res3$SummaryPlot)), "try-error"))
  expect_message2( #util_margins_lm
    res3b <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = study_data,
                         meta_data = md, group_vars = "CENTER_0",
                         label_col = LABEL, sort_group_var_levels = FALSE,
                         include_numbers_in_figures = FALSE,
                         n_violin_max = 3)
  )
  expect_false(
    inherits(try(ggplot_build(res3b$SummaryPlot)), "try-error"))

  skip_if_not_installed("ordinal")

  expect_message2( #util_margins_ord
    res3c <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = study_data,
                         meta_data = md, group_vars = "CENTER_0",
                         label_col = LABEL, cut_off_linear_model_for_ord = NULL)
  )
  expect_false(
    inherits(try(ggplot_build(res3c$SummaryPlot)), "try-error"))
  expect_message2( #util_margins_ord
    res3d <- acc_margins(resp_vars = "MEAT_CONS_0", study_data = study_data,
                         meta_data = md, group_vars = "CENTER_0",
                         label_col = LABEL, cut_off_linear_model_for_ord = NULL,
                         sort_group_var_levels = FALSE)
  )
  expect_false(
    inherits(try(ggplot_build(res3d$SummaryPlot)), "try-error"))

  # 3. Poisson model
  md$SCALE_LEVEL[md$LABEL == "N_INJURIES_0"] <- SCALE_LEVELS$INTERVAL
  expect_message2( #util_margins_poi
    res4 <- acc_margins(resp_vars = "N_INJURIES_0", study_data = study_data,
                        meta_data = md, group_vars = "USR_SOCDEM_0",
                        label_col = LABEL, sort_group_var_levels = TRUE,
                        include_numbers_in_figures = FALSE)
  )
  expect_false(
    inherits(try(ggplot_build(res4$SummaryPlot)), "try-error"))
  expect_message2( #util_margins_poi
    res4b <- acc_margins(resp_vars = "N_INJURIES_0", study_data = study_data,
                         meta_data = md, group_vars = "USR_SOCDEM_0",
                         label_col = LABEL, sort_group_var_levels = FALSE,
                         include_numbers_in_figures = TRUE)
  )
  expect_false(
    inherits(try(ggplot_build(res4b$SummaryPlot)), "try-error"))

  expect_message2( #util_margins_poi
    res4b_new <- acc_margins(resp_vars = "N_INJURIES_0", study_data = study_data,
                         meta_data = md, group_vars = "USR_SOCDEM_0",
                         label_col = LABEL, sort_group_var_levels = FALSE,
                         include_numbers_in_figures = TRUE,
                         threshold_type = "none")
  )

  expect_error(
    res1 <-
      acc_margins(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data, group_vars = "v00001"),
    regexp = "Argument .group_vars.+Variable .v00001.+na. does not have an allowed scale level",
    perl = TRUE
  )

  expect_message2(
    res1 <-
      acc_margins(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data, group_vars = "v00016",
                  min_obs_in_subgroup = 2, sort_group_var_levels = FALSE,
                  n_violin_max = 3),
    regexp =
      paste("min_obs_in_subgroup is not specified correctly",
            "and is set to 5 instead.")
  )

  expect_message2(
      res1 <-
        acc_margins(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data, group_vars = "v00016"),
    regexp =
      sprintf(
        "(%s|%s)",
        paste("Due to missing values in v00016, N = 308",
              "observations were excluded. Due to missing values in",
              "v00014, N = 131 observations were excluded additionally."),
        paste("No or many threshold type specified and set to empirical.")
      ),
    perl = TRUE
  )

  expect_true(all(
    c("ResultData",
      "SummaryTable",
      "SummaryPlot"
      ) %in% names(res1)))

  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$ResultData)),
      na.rm = TRUE) - 2590.933)), 0.1
  )

  expect_equal(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE))), 0
  )

  expect_identical(colnames(res1$SummaryTable), c("Variables"))
})

test_that("acc_margins works with label_col", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_error({
    res1 <-
      acc_margins(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data)
  },
  regexp = paste("Argument group_vars is NULL"),
  perl = TRUE
)

suppressWarnings(expect_error(
  res1 <-
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
              meta_data = meta_data, group_vars = "DEV_NO_0",
              label_col = LABEL, threshold_type = "nonex",
              threshold_value = 1),
  regexp = ".+arg.+ should be one of .+empirical.+, .+user.+, .+none.+"
))

suppressWarnings(expect_error(
  res1 <-
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                label_col = LABEL, threshold_type = list()),
  regexp = ".+arg.+ must be NULL or a character vector",
  perl = TRUE
))

expect_message2(
  res1 <-
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                label_col = LABEL, threshold_type = "none",
                threshold_value = 1),
  regexp =
    sprintf(
      "(Due|%s|%s|%s|%ss|%s)",
      paste(
       "No or many minimum observation.+"),
      paste("No co_vars specified"),
      paste("Due to missing values.+308",
            "observations were excluded."),
      paste("Due to missing values.+131",
            "observations were excluded additionally."),
      paste("threshold_value is not numeric.+it to default value 1.")
    ),
  perl = TRUE
)

expect_message2(
  res1 <-
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                label_col = LABEL, threshold_type = "none"),
  regexp =
    sprintf(
      "(%s|%s|%s)",
      paste(
       "No or many minimum observation count was specified and is set to n=5."),
      paste("No co_vars specified"),
      paste("Due to missing values in DEV_NO_0, N = 308",
            "observations were excluded. Due to missing values in",
            "CRP_0, N = 131 observations were excluded additionally")
    ),
  perl = TRUE
)

expect_message2(
  res1 <-
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                label_col = LABEL, threshold_type = "user"),
  regexp =
    sprintf(
      "(%s|%s|%s|%s)",
      paste(
       "No or many minimum observation count was specified and is set to n=5."),
      paste("No co_vars specified"),
      paste("Due to missing values in DEV_NO_0, N = 308",
            "observations were excluded. Due to missing values in",
            "CRP_0, N = 131 observations were excluded additionally."),
      paste("Threshold was set to user but no value for the unit of",
            "measurements was defined.")),
  perl = TRUE
)

expect_message2(
  res1 <-
    acc_margins(resp_vars = "CRP_0", study_data = study_data,
                meta_data = meta_data, group_vars = "DEV_NO_0",
                label_col = LABEL, threshold_type = "empirical"),
  regexp =
    sprintf(
      "(%s|%s)",
      paste(
       "No or many minimum observation count was specified and is set to n=5."),
      paste("Due to missing values in DEV_NO_0, N = 308",
            "observations were excluded. Due to missing values in",
            "CRP_0, N = 131 observations were excluded additionally.")
    ),
  perl = TRUE
)

expect_message2(
    res1 <-
      acc_margins(resp_vars = "CRP_0", study_data = study_data,
                  meta_data = meta_data, group_vars = "DEV_NO_0",
                  label_col = LABEL, sort_group_var_levels = FALSE),
    regexp = paste("Due to missing values .+", "observations were excluded."),
    perl = TRUE
  )

  expect_true(all(
    c("ResultData",
      "SummaryTable",
      "SummaryPlot"
    ) %in% names(res1)))

  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$ResultData)),
      na.rm = TRUE) - 2590.933)), 0.1
  )

  expect_equal(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE))), 0
  )

  expect_identical(colnames(res1$SummaryTable), c("Variables"))

  skip_on_cran()
  skip_if_not_installed("vdiffr")

  # TODO: As soon as vdiffr supports the variant argument for snapshots -- https://github.com/r-lib/vdiffr/issues/125, use it everywhere.
  # r_minor_version <- function() paste0("R", getRversion()[, 1:2])
  # ggplot_version <- function() packageVersion("ggplot2")

  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("margins plot for CRP_0 ok",
                              res1$SummaryPlot)
})

test_that("acc_margins works with co_vars", {
  skip_on_cran() # slow

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  # 1. linear regression
  # example 1: obvious differences between examiners
  set.seed(1352)
  sd1 <- study_data
  sd1[["v00012"]] <- rep(c("USR_301", "USR_243", "USR_121"), 1000)
  sd1[["v00004"]][which(sd1[["v00012"]] == "USR_301")] <- 10
  sd1[["v00004"]][which(sd1[["v00012"]] == "USR_243")] <- 0
  sd1[["v00004"]][which(sd1[["v00012"]] == "USR_121")] <- -10
  sd1[["v00004"]] <- round(sd1[["v00004"]] + 70 +
                             5 * sd1[["v00002"]] +
                             0.5 * sd1[["v00003"]] + rnorm(n = nrow(sd1)))
  md1 <- meta_data
  md1[["HARD_LIMITS"]][md1[[VAR_NAMES]] == "v00004"] <- ""

  suppressMessages(
    res1_lm_unadj <- acc_margins(resp_vars = "SBP_0", # v00004
                                 group_vars = "USR_BP_0", # v00012
                                 co_vars = NULL,
                                 study_data = sd1,
                                 meta_data = md1,
                                 label_col = LABEL,
                                 sort_group_var_levels = FALSE)
  )
  # compare estimates from the model with empirical mean values
  mu_emp <- tapply(sd1[["v00004"]], sd1[["v00012"]], mean, na.rm = TRUE)
  mu_emp <- mu_emp[match(res1_lm_unadj$ResultData$USR_BP_0, names(mu_emp))]
  expect_lte(max(abs(as.numeric(res1_lm_unadj$ResultData$margins) -
                       as.numeric(util_round_to_decimal_places(as.vector(mu_emp))))),
             0.01)
  suppressMessages(
    res1_lm <- acc_margins(resp_vars = "SBP_0",
                           group_vars = "USR_BP_0",
                           co_vars = c("SEX_0", "AGE_0"),
                           study_data = sd1,
                           meta_data = md1,
                           label_col = LABEL,
                           sort_group_var_levels = FALSE)
  )
  # marginal means should be almost the same in both approaches
  expect_lte(
    max(abs(
      as.numeric(res1_lm_unadj$ResultData$margins) -
        as.numeric(res1_lm$ResultData$margins)
      )), 2)
  # standard errors should be much smaller when adjusting for the covariates in
  # this example
  expect_true(all(
    as.numeric(res1_lm_unadj$ResultData$SE) >
      as.numeric(res1_lm$ResultData$SE)
  ))

  # example 2: no differences between examiners after adjusting for covariates
  set.seed(1352)
  sd1 <- study_data
  age_cut3 <- as.numeric(cut(sd1[["v00003"]], breaks = 3))
  age_cut3[which(is.na(age_cut3))] <- 1
  sd1[["v00012"]] <- c("USR_301", "USR_243", "USR_121")[age_cut3]
  sd1[["v00002"]][which(sd1[["v00012"]] == "USR_301")] <- 0
  sd1[["v00002"]][which(sd1[["v00012"]] == "USR_243")] <- 1
  sd1[["v00002"]][which(sd1[["v00012"]] == "USR_121")] <-
    sample(c(0,1), replace = TRUE,
           size = length(which(sd1[["v00012"]] == "USR_121")))
  sd1[["v00004"]] <- round(70 + 20 * sd1[["v00002"]] + 0.2 * sd1[["v00003"]] +
                             rnorm(sd = 3, n = nrow(sd1)))
  md1 <- meta_data
  md1[["HARD_LIMITS"]][md1[[VAR_NAMES]] == "v00004"] <- ""
  suppressMessages(
    res2_lm_unadj <- acc_margins(resp_vars = "SBP_0", # v00004
                                 group_vars = "USR_BP_0", # v00012
                                 co_vars = NULL,
                                 study_data = sd1,
                                 meta_data = md1,
                                 label_col = LABEL,
                                 sort_group_var_levels = FALSE)
  )
  res2u_mar <- as.numeric(res2_lm_unadj$ResultData$margins)
  # compare estimates from the model with empirical mean values
  mu_emp <- tapply(sd1[["v00004"]], sd1[["v00012"]], mean, na.rm = TRUE)
  mu_emp <- mu_emp[match(res2_lm_unadj$ResultData$USR_BP_0, names(mu_emp))]

  expect_lte(max(abs(res2u_mar -
                       as.numeric(util_round_to_decimal_places(
                         as.vector(mu_emp))))), 0.01)
  # the estimated marginal means should deviate considerably
  expect_gte(max(res2u_mar) - min(res2u_mar), 15)
  suppressMessages(
    res2_lm <- acc_margins(resp_vars = "SBP_0",
                           group_vars = "USR_BP_0",
                           co_vars = c("SEX_0", "AGE_0"),
                           study_data = sd1,
                           meta_data = md1,
                           label_col = LABEL,
                           sort_group_var_levels = FALSE)
  )
  res2a_mar <- as.numeric(res2_lm$ResultData$margins)
  # the estimated marginal means should deviate only minimally
  expect_lte(max(res2a_mar) - min(res2a_mar), 2)
  # there should be levels of the grouping variable that deviate considerably
  # between both approaches, and one level of the grouping variable that
  # deviates only slightly
  expect_gte(max(abs(res2u_mar - res2a_mar)), 10)
  expect_lte(min(abs(res2u_mar - res2a_mar)), 2)

  # 2. binary regression
  # no differences between examiners after adjusting for covariates
  set.seed(603)
  sd1 <- study_data
  age_cut3 <- as.numeric(cut(sd1[["v00003"]], breaks = 3))
  age_cut3[which(is.na(age_cut3))] <- 1
  sd1[["v00011"]] <- c("USR_321", "USR_590", "USR_213")[age_cut3]
  sd1[["v00002"]][which(sd1[["v00011"]] == "USR_321")] <- 0
  sd1[["v00002"]][which(sd1[["v00011"]] == "USR_590")] <- 1
  sd1[["v00002"]][which(sd1[["v00011"]] == "USR_213")] <-
    sample(c(0,1), replace = TRUE,
           size = length(which(sd1[["v00011"]] == "USR_213")))
  mlin <- log(0.3) + 0.25 * sd1[["v00002"]] +
    2 * (sd1[["v00003"]] - mean(sd1[["v00003"]], na.rm = TRUE)) /
    sd(sd1[["v00003"]], na.rm = TRUE)
  mlin[which(is.na(mlin))] <- mean(mlin, na.rm = TRUE)
  sd1[["v00007"]] <- rbinom(n = nrow(sd1), size = 1, prob = 1/(1 + exp(-mlin)))
  md1 <- meta_data
  md1[["RECODE_CASES"]] <- ""
  md1[["RECODE_CONTROL"]] <- ""
  md1[["RECODE_CASES"]][md1$VAR_NAMES == "v00007"] <- "yes"
  md1[["RECODE_CONTROL"]][md1$VAR_NAMES == "v00007"] <- "no"

  suppressMessages(
    res3_bin_unadj <- acc_margins(resp_vars = "ASTHMA_0",
                                  study_data = sd1,
                                  meta_data = md1,
                                  group_vars = "USR_VO2_0",
                                  co_vars = NULL,
                                  label_col = LABEL)
  )
  res3u_mar <- as.numeric(res3_bin_unadj$ResultData$margins)
  # compare estimates from the model with empirical mean values
  mu_emp <- tapply(sd1[["v00007"]], sd1[["v00011"]], mean, na.rm = TRUE)
  mu_emp <- mu_emp[match(res3_bin_unadj$ResultData$USR_VO2_0, names(mu_emp))]

  expect_lte(max(abs(res3u_mar -
                       as.numeric(
                         util_round_to_decimal_places(as.vector(mu_emp))))),
             0.01)
  # the estimated marginal means should deviate considerably
  expect_gte(max(res3u_mar) - min(res3u_mar), 0.5)
  suppressMessages(
    res3_bin <- acc_margins(resp_vars = "ASTHMA_0",
                            study_data = sd1,
                            meta_data = md1,
                            group_vars = "USR_VO2_0",
                            co_vars = c("SEX_0", "AGE_0"),
                            label_col = LABEL)
  )
  res3a_mar <- as.numeric(res3_bin$ResultData$margins)
  # the estimated marginal means should deviate only minimally
  expect_lte(max(res3a_mar) - min(res3a_mar), 0.2)
  # there should be levels of the grouping variable that deviate considerably
  # between both approaches, and one level of the grouping variable that
  # deviates only slightly
  expect_gte(max(abs(res3u_mar - res3a_mar)), 0.5)
  expect_lte(min(abs(res3u_mar - res3a_mar)), 0.2)

  # 3. Poisson regression
  # example 1: obvious differences between examiners
  set.seed(1646)
  sd1 <- study_data
  parm_age <- (sd1[["v00003"]] - min(sd1[["v00003"]], na.rm = TRUE)) /
    (max(sd1[["v00003"]], na.rm = TRUE) - min(sd1[["v00003"]], na.rm = TRUE))
  parm_sex <- 0.4 * sd1[["v00002"]]
  sd1[["v00032"]] <- rep(c("USR_321", "USR_247", "USR_520"), 1000)
  sd1[["v00026"]][which(sd1[["v00032"]] == "USR_321")] <- 1
  sd1[["v00026"]][which(sd1[["v00032"]] == "USR_247")] <- 0.5
  sd1[["v00026"]][which(sd1[["v00032"]] == "USR_520")] <- 0
  ll1 <- exp(sd1[["v00026"]] + parm_age + parm_sex)
  ll1[which(is.na(ll1))] <- 1
  sd1[["v00026"]] <- rpois(lambda = ll1, n = nrow(sd1))
  # scale down to trigger Poisson regression (v2.5.0)
  if (max(sd1[["v00026"]], na.rm = TRUE) >= 20) {
    sd1[["v00026"]] <- round(sd1[["v00026"]]/max(sd1[["v00026"]]) * 19)
  }
  md1 <- meta_data
  md1[["HARD_LIMITS"]][md1[[VAR_NAMES]] == "v00026"] <- ""
  md1$SCALE_LEVEL[md1$LABEL == "N_INJURIES_0"] <- SCALE_LEVELS$INTERVAL

  suppressMessages(
    res5_poi_unadj <- acc_margins(resp_vars = "N_INJURIES_0",
                                  group_vars = "USR_SOCDEM_0",
                                  co_vars = NULL,
                                  study_data = sd1,
                                  meta_data = md1,
                                  label_col = LABEL,
                                  sort_group_var_levels = FALSE)
  )
  # compare estimates from the model with empirical mean values
  mu_emp <- tapply(sd1[["v00026"]], sd1[["v00032"]], mean, na.rm = TRUE)
  mu_emp <- mu_emp[match(res5_poi_unadj$ResultData$USR_SOCDEM_0, names(mu_emp))]
  expect_lte(max(abs(as.numeric(res5_poi_unadj$ResultData$margins) -
                  as.numeric(util_round_to_decimal_places(as.vector(mu_emp))))),
             0.01)
  suppressMessages(
    res5_poi <- acc_margins(resp_vars = "N_INJURIES_0",
                            group_vars = "USR_SOCDEM_0",
                            co_vars = c("SEX_0", "AGE_0"),
                            study_data = sd1,
                            meta_data = md1,
                            label_col = LABEL,
                            sort_group_var_levels = FALSE)
  )

  # marginal means should be almost the same in both approaches
  expect_lte(
    max(abs(
      as.numeric(res5_poi_unadj$ResultData$margins) -
        as.numeric(res5_poi$ResultData$margins)
    )), 2)

  # example 2: no differences between examiners after adjusting for covariates
  set.seed(2128)
  sd1 <- study_data
  age_cut3 <- as.numeric(cut(sd1[["v00003"]], breaks = 3))
  age_cut3[which(is.na(age_cut3))] <- 1
  sd1[["v00032"]] <- c("USR_321", "USR_247", "USR_520")[age_cut3]
  sd1[["v00002"]][which(sd1[["v00032"]] == "USR_321")] <- 0
  sd1[["v00002"]][which(sd1[["v00032"]] == "USR_520")] <- 1
  sd1[["v00002"]][which(sd1[["v00032"]] == "USR_247")] <-
    sample(c(0,1), replace = TRUE,
           size = length(which(sd1[["v00032"]] == "USR_247")))
  ll1 <- exp(parm_age + 0.5 * sd1[["v00002"]])
  ll1[which(is.na(ll1))] <- 1
  sd1[["v00026"]] <- rpois(lambda = ll1, n = nrow(sd1))
  md1 <- meta_data
  md1[["HARD_LIMITS"]][md1[[VAR_NAMES]] == "v00026"] <- ""
  md1$SCALE_LEVEL[md1$LABEL == "N_INJURIES_0"] <- SCALE_LEVELS$INTERVAL

  suppressMessages(
    res6_poi_unadj <- acc_margins(resp_vars = "N_INJURIES_0",
                                  group_vars = "USR_SOCDEM_0",
                                  co_vars = NULL,
                                  study_data = sd1,
                                  meta_data = md1,
                                  label_col = LABEL,
                                  sort_group_var_levels = FALSE)
  )
  # compare estimates from the model with empirical mean values
  mu_emp <- tapply(sd1[["v00026"]], sd1[["v00032"]], mean, na.rm = TRUE)
  mu_emp <- mu_emp[match(res6_poi_unadj$ResultData$USR_SOCDEM_0, names(mu_emp))]
  expect_lte(max(abs(as.numeric(res6_poi_unadj$ResultData$margins) -
                  as.numeric(util_round_to_decimal_places(as.vector(mu_emp))))),
             0.01)
  suppressMessages(
    res6_poi <- acc_margins(resp_vars = "N_INJURIES_0",
                            group_vars = "USR_SOCDEM_0",
                            co_vars = c("SEX_0", "AGE_0"),
                            study_data = sd1,
                            meta_data = md1,
                            label_col = LABEL,
                            sort_group_var_levels = FALSE)
  )
  res6u_mar <- as.numeric(res6_poi_unadj$ResultData$margins)
  res6a_mar <- as.numeric(res6_poi$ResultData$margins)
  # the estimated marginal means in the adjusted model should deviate
  # less than in the unadjusted model
  expect_gte(var(res6u_mar), var(res6a_mar))
  # the estimated marginal means in the adjusted model should deviate
  # only minimally
  expect_lte(max(res6a_mar) - min(res6a_mar), 0.5)
  # there should be levels of the grouping variable that deviate considerably
  # between both approaches, and one level of the grouping variable that
  # deviates only slightly
  expect_gte(max(abs(res6u_mar - res6a_mar)), 1)
  expect_lte(min(abs(res6u_mar - res6a_mar)), 0.5)

  # 4. ordinal regression
  # TODO
  md1 <- meta_data
  md1$SCALE_LEVEL[md1$LABEL == "MEAT_CONS_0"] <- SCALE_LEVELS$ORDINAL
  skip_if_not_installed("ordinal")
  expect_error(
    res8_ord <- acc_margins(resp_vars = "MEAT_CONS_0",
                            group_vars = "CENTER_0",
                            co_vars = "AGE_0",
                            study_data = study_data,
                            meta_data = md1,
                            label_col = LABEL,
                            cut_off_linear_model_for_ord = NULL,
                            sort_group_var_levels = FALSE),
    regexp = paste("Covariate argument for ordinal regression",
                   "not yet supported.")
  )

  # 5. multinomial regression
  set.seed(940)
  sd1 <- study_data
  age_cut3 <- as.numeric(cut(sd1[["v00003"]], breaks = 3))
  age_cut3[which(is.na(age_cut3))] <- 1
  sd1[["v00000"]] <- c(1:3)[age_cut3]
  sd1[["v00002"]][which(sd1[["v00002"]] == 1)] <-
    sample(c(0,1), replace = TRUE,
           size = length(which(sd1[["v00002"]] == 1))) #0
  sd1[["v00002"]][which(sd1[["v00002"]] == 2)] <-
    sample(c(0,1), replace = TRUE,
           size = length(which(sd1[["v00002"]] == 2)))#1
  sd1[["v00002"]][which(sd1[["v00002"]] == 3)] <-
    sample(c(0,1), replace = TRUE,
           size = length(which(sd1[["v00002"]] == 3)))
  xx <- #0.25 * sd1[["v00002"]] +
    #0.75 *
    (rank(sd1[["v00003"]], na.last = FALSE) - 1)/nrow(sd1)
  xx[which(is.na(xx))] <- mean(xx, na.rm = TRUE)
  xx <- (xx - min(xx))/(max(xx) - min(xx))
  sd1[["v00023"]] <- round(xx * 4)
  md1 <- meta_data
  md1[["HARD_LIMITS"]][md1[[VAR_NAMES]] == "v00023"] <- ""
  md1$SCALE_LEVEL[md1$LABEL == "MEAT_CONS_0"] <- SCALE_LEVELS$NOMINAL

  res10_cat_unadj <- acc_margins(resp_vars = "MEAT_CONS_0",
                                 group_vars = "CENTER_0",
                                 study_data = sd1,
                                 meta_data = md1,
                                 label_col = LABEL,
                                 dichotomize_categorical_resp = FALSE)

  expect_error(
    res10_cat <- acc_margins(resp_vars = "MEAT_CONS_0",
                             group_vars = "CENTER_0",
                             co_vars = c("AGE_0", "SEX_0"),
                             study_data = sd1,
                             meta_data = md1,
                             label_col = LABEL,
                             dichotomize_categorical_resp = FALSE),
    regexp = paste("Adjusting for covariates is currently",
                   "only supported for factor variables",
                   "with 2 or more levels")
  )
  # TODO
  res10_cat <- acc_margins(resp_vars = "MEAT_CONS_0",
                                 group_vars = "CENTER_0",
                           co_vars = "AGE_GROUP_0",
                                 study_data = sd1,
                                 meta_data = md1,
                                 label_col = LABEL,
                                 dichotomize_categorical_resp = FALSE)

})

#' Map based on largest common prefix
#'
#' @param needle [character]`(1)` item to search
#' @param haystack [character] items to find the entry sharing the largest
#'                             prefix with `needle`
#' @param split_char [character]`(1)` to split entries to atomic words (like
#' letters, if "" or `snake_elements`, if "_")
#' @param remove_var_suffix [logical]`(1)` remove potential suffix
#'                                         after the first dot `.`, before
#'                                         finding needle in haystack.
#'
#' @return [character]`(1)` with the fitting function name or [NA_character_]
#' @examples
#' \dontrun{ # internal function
#' util_map_by_largest_prefix(
#'   "acc_distributions_loc_ecdf_observer_time",
#'   names(dataquieR:::.manual$titles)
#' )
#' util_map_by_largest_prefix(
#'   "acc_distributions_loc_observer_time",
#'   names(dataquieR:::.manual$titles)
#' )
#' util_map_by_largest_prefix(
#'   "acc_distributions_loc_ecdf",
#'   names(dataquieR:::.manual$titles)
#' )
#' util_map_by_largest_prefix(
#'   "acc_distributions_loc",
#'   names(dataquieR:::.manual$titles)
#' )
#' }
#'
#' @family mapping
#' @concept concept
#' @noRd
util_map_by_largest_prefix <- function(
    needle,
    haystack,
    split_char = "_",
    remove_var_suffix = TRUE
    ) {
  util_expect_scalar(needle, check_type = is.character)
  util_expect_scalar(remove_var_suffix, check_type = is.logical)
  util_expect_scalar(haystack,
                     check_type = is.character,
                     allow_more_than_one = TRUE)
  if (is.null(names(haystack))) {
    haystack <- setNames(nm = haystack)
  }
  if (remove_var_suffix) {
    if (grepl(".", needle, fixed = TRUE)) {
      needle <- head(strsplit(needle, ".", fixed = TRUE)[[1]], 1)
    }
  }
  needle0 <- needle
  needle <- strsplit(needle, split_char, fixed = TRUE)[[1]]
  haystack <- strsplit(haystack, split_char, fixed = TRUE)
  lneedle <- length(needle)

  found <- FALSE
  i <- length(needle)
  while (!found) {
    needle <- needle[seq_len(i)]
    found <- any(vapply(haystack, FUN.VALUE = logical(1), function(straw) {
      length(straw) == i && all(straw == needle)
    }))
    i <- i - 1
    if (i < 0) {
      return(NA_character_)
    }
  }
  paste0(needle, collapse = split_char)
}

test_that("util_map_by_largest_prefix works", {
  skip_on_cran()

  expect_equal(util_map_by_largest_prefix(
    "acc_distributions_loc_ecdf_observer_time",
    names(dataquieR:::.manual$titles)
  ), "acc_distributions_loc")

  expect_equal(
    util_map_by_largest_prefix(
      "acc_distributions_loc_observer_time",
      names(dataquieR:::.manual$titles)
    ),
    "acc_distributions_loc"
  )

  expect_equal(
    util_map_by_largest_prefix(
      "acc_distributions_loc_ecdf",
      names(dataquieR:::.manual$titles)
    ),
    "acc_distributions_loc"
  )

  expect_equal(
    util_map_by_largest_prefix(
      "acc_distribution",
      names(dataquieR:::.manual$titles)
    ),
    NA_character_
  )

  expect_equal(
    util_map_by_largest_prefix(
      "acc_distributions_loc",
      names(dataquieR:::.manual$titles)
    ),
    "acc_distributions_loc"
  )

  expect_equal(
    util_map_by_largest_prefix(
      "acc_distributions_loessf",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributions", "xxx")),
    "acc_distributions"
  )

  expect_equal(
    util_map_by_largest_prefix(
      "acc_distributions_loessf",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributsons", "xxx")),
    NA_character_
  )

  expect_equal(
    util_map_by_largest_prefix(
      "acc_distributions_loess",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributions", "xxx", "acc_distributions_loess")
    ),
    "acc_distributions_loess"
  )

  expect_equal(
    util_map_by_largest_prefix(
      "acc_distributions_loessf",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributions", "xxx", "acc_distributions_loess")
    ),
    "acc_distributions"
  )

  expect_equal(
    util_map_by_largest_prefix(
      "acc_distributions_loess",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributions", "xxx", "acc_distributions_loessf")
    ),
    "acc_distributions"
  )

  expect_equal(
    util_map_by_largest_prefix(
      split_char = "",
      "acc_distributions_loessf",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributions", "xxx")),
    "acc_distributions"
  )


  expect_equal(
    util_map_by_largest_prefix(
      split_char = "",
      "acc_distributions_loess",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributions", "xxx", "acc_distributions_loess")
    ),
    "acc_distributions_loess"
  )

  expect_equal(
    util_map_by_largest_prefix(
      split_char = "",
      "acc_distributions_loessf",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributions", "xxx", "acc_distributions_loess")
    ),
    "acc_distributions_loess"
  )

  expect_equal(
    util_map_by_largest_prefix(
      split_char = "",
      "acc_distributions_loess",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributions", "xxx", "acc_distributions_loessf")
    ),
    "acc_distributions"
  )

  expect_equal(
    util_map_by_largest_prefix(
      split_char = "",
      "acc_distributions_loess",
      c("aaaa", "acc_distributions_loc_ecdf", "acc_distributions_loc",
        "acc_distributionsR", "xxx", "acc_distributions_loessf")
    ),
    NA_character_
  )


})

#' The jack of all trades device for tables
#'
#' @param ... passed to `DT::datatable`
#' @param tb the table as [data.frame]
#' @param columnDefs column specifications for the `datatables` JavaScript
#'                   object
#' @param autoWidth passed to the `datatables` JavaScript library
#' @param hideCols columns to hide (by name)
#' @param rowCallback passed to the `datatables` JavaScript library
#'                    (with default)
#' @param copy_row_names_to_column add a column 0 with `rownames`
#' @param tb_rownames number of columns from the left considered as row-names
#' @param meta_data the data dictionary for labels and similar stuff
#' @param options individually overwrites defaults in `options`
#'                passed to `DT::datatable`
#' @param link_variables considering row names being variables, convert
#'                      row names to links to the variable specific reports
#' @param rotate_headers rotate headers by 90 degrees
#' @param colnames column names for the table (defaults to `colnames(tb)`)
#' @param is_matrix_table create a heat map like table without padding
#' @param filter passed to `DT::datatable`
#' @param colnames_aliases2acronyms abbreviate column names considering being
#'                                  analysis matrix columns by their acronyms
#'                                  defined in square.
#' @param label_col label col used for mapping labels in case of
#'                  `link_variables` is used (that argument set to `TRUE` and
#'                  `Variables` or `VAR_NAMES` in `meta_data`)
#' @param fillContainer see `DT::datatable`
#' @param dl_fn file name for downloaded table -- see
#'      [https://datatables.net/reference/button/excel](https://datatables.net/reference/button/excel)
#' @param cols_are_indicatormetrics [logical] cannot be `TRUE`,
#'        `colnames_aliases2acronyms` is `TRUE`. `cols_are_indicatormetrics`
#'        controls, if the columns are really function calls or, if
#'        `cols_are_indicatormetrics` has been set to `TRUE`, the columns are
#'        indicator metrics.
#' @param rotate_for_one_row [logical] rotate one-row-tables
#' @param descs [character] descriptions of the columns for the hover-box shown
#'                          for the column names, if not missing, this overrides
#'                          the existing description stuff from known column
#'                          names. If you have an attribute "description" of the `tb`, then it
#'                          overwrites everything and appears as hover text
#' @param title [character] title for download formats, see
#'        [https://datatables.net/extensions/buttons/examples/html5/titleMessage.html](https://datatables.net/extensions/buttons/examples/html5/titleMessage.html)
#' @param messageTop  [character] subtitle for download formats, see
#'        [https://datatables.net/extensions/buttons/examples/html5/titleMessage.html](https://datatables.net/extensions/buttons/examples/html5/titleMessage.html)
#' @param messageBottom  [character] footer for download formats, see
#'        [https://datatables.net/extensions/buttons/examples/html5/titleMessage.html](https://datatables.net/extensions/buttons/examples/html5/titleMessage.html)
#' @param col_tags [list] if not `NULL`, a named `list()`, names are names used
#'                        to name a newly created  column-group hide/show
#'                        button, elements are column names belonging to each
#'                        column groups as defined by `colnames`
#' @param searchBuilder [logical] if `TRUE`, display a `searchBuilder`-Button.
#' @param init_search [list] object to initialize `searchBuilder`, see [`datatables.net`](https://datatables.net/reference/type/SearchBuilder.Criteria)
#' @param initial_col_tag [character] `col_tags` entry to activate initially
#' @param additional_init_args [list] if not missing or `NULL`, arguments passed to `JavaScript`, if `searchBuilder == TRUE`
#' @param additional_columnDefs [list] additional `columnDefs`, can be missing or `NULL`
#' @param df_escape [logical] apply `util_df_escape()` to the table
#'
#' @return the table to be added to an `rmd`/´`html` file as
#'         [htmlwidgets::htmlwidgets]
#' @seealso `util_formattable()`
#'
#' @family summary_functions
#' @concept html
#' @noRd
#'
util_html_table <- function(tb,
                  filter = "top",
                  columnDefs = NULL,
                  autoWidth = FALSE,
                  hideCols = character(0),
                  rowCallback = DT::JS("function(r,d) {$(r).attr('height', '2em')}"),
                  copy_row_names_to_column = !is.null(tb) &&
                    length(rownames(tb)) == nrow(tb) &&
                    !is.integer(attr(tb, "row.names")) && !all(seq_len(nrow(tb))
                                                               == rownames(tb)),
                  link_variables = TRUE,
                  tb_rownames = FALSE,
                  meta_data,
                  rotate_headers = FALSE,
                  fillContainer = TRUE,
                  ...,
                  colnames,
                  descs,
                  options = list(),
                  is_matrix_table = FALSE,
                  colnames_aliases2acronyms = is_matrix_table &&
                    !cols_are_indicatormetrics,
                  cols_are_indicatormetrics = FALSE,
                  label_col = LABEL,
                  dl_fn = "*",
                  rotate_for_one_row = FALSE,
                  title = dl_fn,
                  messageTop = NULL,
                  messageBottom = NULL,
                  col_tags = NULL,
                  searchBuilder = FALSE,
                  initial_col_tag,
                  init_search,
                  additional_init_args,
                  additional_columnDefs,
                  kv_table = !is.null(tb) && isTRUE(attr(tb, "kv_table")),
                  df_escape = FALSE
                  ) {
# caveat: the fixed columns filter may not work.
# TODO: add comments (descs) to header cells in excel export
# TODO: add figures to exports

  if (is.null(tb) || nrow(tb) == 0 || ncol(tb) == 0) {
    return()
  }

  force(copy_row_names_to_column)
  force(kv_table)

  util_ensure_suggested("htmltools", "Generating nice tables")
  util_ensure_suggested("DT", "Generating nice tables")
  util_ensure_suggested("htmlwidgets", "Generating nice tables")

  util_expect_scalar(rotate_for_one_row, check_type = is.logical)
  util_expect_scalar(cols_are_indicatormetrics, check_type = is.logical)
  util_expect_scalar(colnames_aliases2acronyms, check_type = is.logical)
  util_expect_scalar(is_matrix_table, check_type = is.logical)
  util_expect_scalar(df_escape, check_type = is.logical)

  util_stop_if_not(!(cols_are_indicatormetrics && colnames_aliases2acronyms))

  #  if we are currently being knit get our metadata from knitr
  if (missing(meta_data)) {
    if (isTRUE(getOption('knitr.in.progress')) &&
        exists("report", envir = knitr::knit_global())) {
      # https://stackoverflow.com/a/33121933
      meta_data <- knitr::knit_global()[["report"]][["meta_data"]]
    }
  }

  plain_label <- NULL
  try(plain_label <- attr(tb[["Variables"]], "plain_label"), silent = TRUE)

  tb <- util_html_table_prepare_data(tb
                                     , rotate_for_one_row
                                     , copy_row_names_to_column
                                     , df_escape = df_escape)

  if (link_variables && any(c("Variables", VAR_NAMES) %in%
                            base::colnames(tb)) && !missing(meta_data)) {
    result <- util_html_table_link_variables(tb = tb, descs = descs, meta_data = meta_data, label_col = label_col)
    tb <- result[["tb"]]
    if (!is.null(result[["descs"]])){
      descs <- result[["descs"]]
    }
  }

  # rotate ----
  if (rotate_for_one_row && nrow(tb) == 1) {
    tb <- util_table_rotator(tb)
    # is_html_escaped <- (identical(attr(tb, "is_html_escaped"), TRUE))
    # kv_table <- TRUE
    # tb <- data.frame(Name = base::colnames(tb),
    #                  Value = unlist(tb[1, , TRUE], recursive = FALSE,
    #                                 use.names = FALSE),
    #                  stringsAsFactors = FALSE)
    # attr(tb, "is_html_escaped") <- is_html_escaped
  }

  # column attributes ----
  if (missing(colnames)) {
    colnames <- base::colnames(tb)
  }

  if (is.null(columnDefs)) {
    columnDefs <-
      list(
        list(width = '5em',
             targets = (setdiff(seq_len(ncol(tb)), which(colnames(tb) %in% hideCols))) - 1
        ),
        list(visible = FALSE,
             searchable = FALSE,
             targets = which(colnames(tb) %in% hideCols) - 1
        )
      )
  }

  columnDefs <- c(
    columnDefs,
    list(
      list(
        targets = (which(colnames(tb) %in% c(VAR_NAMES, "Variables", "Labels", LABEL))) - 1,
        className = "dt-nowrap max_60vw"
      ),
      list(
        targets = (setdiff(seq_len(ncol(tb)), which(colnames(tb) %in% hideCols))) - 1,
        render = DT::JS("sort_vert_dt")
      )
    )
  )

  fnames <-
    vapply(colnames,
           util_map_by_largest_prefix,
           haystack = names(.manual$titles),
           FUN.VALUE = character(1))

  ftitles <-
    vapply(fnames, function(fn) {
      r <- util_alias2caption(fn, long = TRUE) # .manual$titles[[fn]];
      if (length(r) != 1) r <- NA_character_;
      r
    }, FUN.VALUE = character(1))

  ftitles[is.na(ftitles)] <- colnames[is.na(ftitles)]
  fnames[is.na(fnames)] <- colnames[is.na(fnames)]

  ftitles <- vapply(lapply(ftitles, htmltools::h3),
                      as.character, FUN.VALUE = character(1))

  coltitles <-
    vapply(colnames, function(cn) {
      r <- util_alias2caption(cn, long = TRUE) # .manual$titles[[fn]];
      if (length(r) != 1) r <- NA_character_;
      r
    }, FUN.VALUE = character(1))


  if (colnames_aliases2acronyms) {
    suffixes <-
      mapply(SIMPLIFY = FALSE, cn = colnames, fn = fnames,
             FUN = function(cn, fn) {
        if (startsWith(cn, fn)) {
          substr(cn, nchar(fn) + 1 + 1, nchar(cn)) # name + "_" (first +1), start is the next character (second +1)
        } else {
          cn
        }
      })
    acronyms <-
      util_map_labels(fnames,
                      util_get_concept_info("implementations"),
                      to = "dq_report2_short_title",
                      from = "function_R",
                      ifnotfound = util_abbreviate(fnames))
    suffixes <- gsub("_", " ", vapply(suffixes, as.character,
                                      FUN.VALUE = character(1)))
    suffixes[!util_empty(suffixes)] <- paste0(":",
                                             abbreviate(
                                               suffixes[!util_empty(suffixes)],
                                               minlength = 3
                                              )
                                            )
    acronyms <- paste0(acronyms, suffixes)
    names(acronyms) <- colnames
  } else if (cols_are_indicatormetrics) {
    # acronyms <- colnames
    acronyms <- util_translate_indicator_metrics(colnames,
                                                 short = TRUE,
                                                 long = FALSE)
  } else {
    acronyms <- colnames
    names(acronyms) <- colnames
  }

  if (cols_are_indicatormetrics) {
    fdescs <- rep("TODO", length(colnames))
  } else {
    fdescs <- vapply(fnames, FUN.VALUE = character(1), util_function_description)
  }

  if (missing(descs)) {
    if(!is.null(attr(tb, "description"))){
      descs <- attr(tb, "description")
      descs <- descs[colnames]
      descs[is.na(descs)] <- ""
      names(descs) <- colnames
    } else if (cols_are_indicatormetrics) {
      descs <- rep("TODO", length(colnames))
    } else {
      descs <- vapply(colnames, FUN.VALUE = character(1), util_col_description)
    }
  } else {
    util_expect_scalar(descs, allow_more_than_one = TRUE, check_type =
                         is.character)
    util_stop_if_not(
      `Need one description per column` =
        length(descs) == length(colnames))
  }

  acs <- acronyms[colnames]

  descs <- htmltools::htmlEscape(descs, attribute = TRUE)
  fdescs <- htmltools::htmlEscape(fdescs, attribute = TRUE)
  acs <- htmltools::htmlEscape(acs, attribute = TRUE)

  if (all(is.na(acs)) && all(colnames == "")) {
    acs <- colnames
  }

  cssClass <- if (rotate_headers) c("vertDT") else "myDT"

  if (rotate_headers && !searchBuilder) {
    cn <- paste0("<div class=\"colheader\" colname=\"",
               coltitles,
               "\" title=\"",
               paste(coltitles, descs, sep = "<br />\n\n"),
               "\">",
               vapply(
                 strsplit(
                   acs,
                   "", fixed = TRUE),
                 function(letters) {
                   paste0("<span>",
                          paste0(letters,
                                 collapse = ""),
                          "</span>")
                 },
                 FUN.VALUE = character(1)),
               "</div>")
   } else if (!searchBuilder) {
    cn <- paste0("<span colname=\"",
                   coltitles,
                   "\" title=\"",
                   paste(coltitles, descs, sep = "<br />\n\n"),
                   "\">",
                   acs,
                   "</span>")

  } else {
    cn <- paste(coltitles)
  }

  # these quotes are needed to prevent Excel from guessing
  # a title like "2020-01-01" may be a date to be displayed
  # as its integer representation
  # TODO: Make messsageTop == null work in JS (ie in ~/git/gitlab/QualityIndicatorFunctions/QualityIndicatorFunctions/inst/report-dt-style/report_dt.js)

  quote_ifnotnull <- function(v, default = v) {
    if(!is.null(v)) {
      return(dQuote(v))
    } else {
      return(default)
    }
   }

  title <- quote_ifnotnull(title)
  messageTop <- quote_ifnotnull(messageTop, "-")
  messageBottom <- quote_ifnotnull(messageBottom)

  messageTop = DT::JS(sprintf('function() { return setMsgTop.call(this, "%s"); }',
                              gsub('"', '\\"', messageTop, fixed = TRUE)))

  col_tags_filter_buttons <- list()
  if (!is.null(col_tags)) {
    for (ct in names(col_tags)) {
      css_class <- attr(col_tags[[ct]], "cssClass")
      in_group <- colnames %in% col_tags[[ct]]
      col_tags_filter_buttons <- c(col_tags_filter_buttons, list(list(
        extend = 'colvisGroup',
        className = paste('buttons-colvisGroup', css_class),
        text = ct,
        show = which(in_group) - 1,
        hide = which(!in_group) - 1,
        name = ct
      )))
    }
  }

  if (!missing(additional_columnDefs) && !is.null(additional_columnDefs) &&
      is.list(additional_columnDefs)) {
    columnDefs <- c(columnDefs, additional_columnDefs)
  }

  if (kv_table && ncol(tb) == 2) { # key-value-table, align to the center
    columnDefs <-
      c(list(list(className = 'dt-right', targets=0),
           list(className = 'dt-left', targets=1)),
      columnDefs)
  } else {
    columnDefs <- c(
      list(list(className = "dt-left", targets = "_all")),  # sane default
      util_get_datatables_alignments(tb),
      columnDefs
    )
  }

  .options <- list(
    dom = "Bt",
    buttons = c(
      list(
        list(
          extend = 'colvis',
          collectionLayout = 'fixed columns',
          postfixButtons = list( 'colvisRestore'
                               , '<input class="search_curr_colvis_input" onkeyup="search_curr_colvis()"></input>'
                               , '<button onclick="search_curr_colvis()">Toggle</button>'),
          columnText = DT::JS("function (dt, idx, title) {
              return dt.column(idx).header().textContent;
            }")
      )),
      col_tags_filter_buttons,
      util_make_export_buttons(dl_fn, title, messageBottom, messageTop)
      ),
    # https://github.com/rstudio/DT/issues/29
    columnDefs = columnDefs,
    autoWidth = autoWidth,
    rowCallback = rowCallback,
    autoFill = TRUE,
    scrollX = TRUE,
    scrollY = "55vh",
    scrollCollapse = TRUE,
    paging = FALSE,
    responsive = TRUE,
    fixedColumns = list(leftColumns = 1 + tb_rownames) # https://stackoverflow.com/a/51623663
  )

  is <- aia <- ict <- "null"

  if (searchBuilder) {
    result <- util_configure_search_builder(.options, init_search, additional_init_args)
    .options <- result[[".options"]]
    is <- result[["init_search"]]
    aia <- result[["additional_init_args"]]
  }

  if (!missing(initial_col_tag)) {
    ict <- sprintf('"%s"', initial_col_tag)
  }

  .options[names(options)] <- options

  if (is_matrix_table) {
    cssClass <- paste(cssClass, "matrixTable", collapse = " ")
  }

  ids <- as.character(floor(runif(2, min = 0, max = .Machine$integer.max)))
  pref <- vapply(nchar(.Machine$integer.max) - nchar(ids),
         function(n) { paste0(rep("0", n), collapse = "") },
         FUN.VALUE = character(1))
  my_id <- paste0("id", paste0(pref, ids), collapse = "")

  if (prep_is_translated(colnames(tb))) {
    colnames(tb) <- as.character(colnames(tb))
  }

  # TODO: Add links to table columns/rows not only the cells.
  dtable <- DT::datatable(tb,
                          elementId = my_id,
                          callback =
                            DT::JS(sprintf("return(dataquieRdtCallback({ 'initialColTag': %s, 'initSearch': %s, 'additional_init_args': %s, 'table': table }));", ict, is, aia)),
                          escape = FALSE,
                          class = cssClass,
                          filter = filter,
                          rownames = tb_rownames,
                          extensions = c("FixedColumns", # https://stackoverflow.com/a/51623663
                                         "Buttons",
                                         "SearchBuilder"),
                          options = .options,
                          colnames = cn,
                          fillContainer = fillContainer,
                          ...)

  if (is_matrix_table) {
    dtable <- htmlwidgets::onRender(dtable,
       "function(el, x) {
          if (!window.jQuery) return;
          var $ = window.jQuery;

          function findWrapper() {
            // el is usually inside the wrapper; if not, fall back to closest wrapper in parents
            var $el = $(el);
            var $w = $el.closest('.dataTables_wrapper');
            return $w.length ? $w : $el.parent().closest('.dataTables_wrapper');
          }

          function syncIn($root) {
            // copy computed bg from pre -> td, then clear pre bg
            $root.find('tbody td > pre').each(function () {
              var pre = this;
              var td = pre.parentNode;
              if (!td) return;

              var bg = window.getComputedStyle(pre).backgroundColor;
              if (!bg || bg === 'transparent' || bg === 'rgba(0, 0, 0, 0)') return;

              td.style.backgroundColor = bg;
              pre.style.backgroundColor = 'transparent';
            });
          }

          function syncAll() {
            var $w = findWrapper();
            if ($w && $w.length) {
              // all tables within this widget's wrapper (main + clones)
              syncIn($w);
            } else {
              // fallback: only within el
              syncIn($(el));
            }
          }

          // Try immediately (may be too early, but cheap)
          syncAll();

          // Wait until the actual DataTable is initialized, then hook events
          var tries = 0;
          (function hook() {
            tries++;

            var $w = findWrapper();
            var $main = $w && $w.length ? $w.find('table.dataTable').first() : $(el).find('table.dataTable').first();

            if ($main.length && $.fn.dataTable && $.fn.dataTable.isDataTable($main[0])) {
              // Sync now and on redraw-ish events
              syncAll();
              $main.on('draw.dt column-visibility.dt column-reorder.dt responsive-resize.dt', syncAll);
              return;
            }

            if (tries < 100) setTimeout(hook, 50); // ~5s max
          })();
       }")
  }

  js <- sprintf("if (!window.dtConfig) {
                  window.dtConfig = {}
                }
                window.dtConfig['%s'] = %s;", my_id, aia)

  dtable <- htmlwidgets::prependContent( dtable,
                                         htmltools::tags$script(DT::JS(js)) )

  jqui <- rmarkdown::html_dependency_jqueryui()
  jqui$stylesheet <- "jquery-ui.min.css"

  dtable$dependencies <- c(
    dtable$dependencies,
    list(
      rmarkdown::html_dependency_jquery(),
      jqui,
      html_dependency_report_dt() # always, since sort_vert is indep. from vert heads
    )
  )

  # dtable <- htmlwidgets::onRender(dtable, paste0("function(el, x) { utilHtmlTableInit(el, x, ", ifelse(isTRUE(fillContainer), "true", "false"), ",) }"))

  htmltools::tagList(htmltools::div(class = "table_top_spacer"),
                     htmltools::div(class = "table_result", dtable),
                     htmltools::br(style = "clear: both"))
}

#' Configure DT options for searchBuilder use
#'
#' @param .options  DT options
#' @param init_search util_html_table()
#' @param additional_init_args see util_html_table()
#'
#' @return a list containing each of the args altered to fit searchBuilder use
#' @keywords internal
#' @noRd
util_configure_search_builder <- function(.options, init_search, additional_init_args){
  .options[["searchBuilder"]] <- TRUE
  .options[["search"]] <- list(return = TRUE)
  .options[["dom"]] <- paste0("Q", .options[["dom"]])

  jsl <- util_ensure_suggested("jsonlite",
                              goal = "search filter init in tables",
                              err = FALSE)

  is <- aia <- "null"

  # return early without init search and additional args if we
  # are unable to generate the JSON
  if (!jsl) {
    return(list(.options = .options, init_search = is, additional_init_args = aia) )
  }

  if (!rlang::is_missing(init_search)) {
    is <- jsonlite::toJSON(init_search, auto_unbox = TRUE)
  }
  if (!rlang::is_missing(additional_init_args)) {
    aia <- jsonlite::toJSON(additional_init_args, auto_unbox = TRUE)
  }

  list(.options = .options, init_search = is, additional_init_args = aia)
}

#' Makes links from the variables in the table
#'
#' @param tb  the table
#' @param descs  descriptions
#' @param meta_data  see util_html_table()
#' @param label_col  see util_html_table()
#'
#' @return a list containing the table and, if applicable, descs
#' @keywords internal
#' @noRd
util_html_table_link_variables <- function(tb, descs, meta_data, label_col) {
  is_html_escaped <- (identical(attr(tb, "is_html_escaped"), TRUE))
  description <- attr(tb, "description")
  plain_label <- attr(tb[["Variables"]], "plain_label")

  Variables <- intersect(base::colnames(tb), c("Variables", VAR_NAMES))
  Variables <- head(Variables, 1)

  pretty_lb <- prep_get_labels( # TODO: use also translations for other labels
    resp_vars = tb[[Variables]],
    item_level = meta_data,
    label_col = label_col,
    label_class = "LONG",
    resp_vars_are_var_names_only = Variables == VAR_NAMES,
    resp_vars_match_label_col_only = Variables == "Variables"
  )

  from <- if (is.null(plain_label)) tb[[Variables]] else plain_label

  if (Variables == VAR_NAMES &&
      all(c(VAR_NAMES, label_col) %in% names(meta_data))) {
    tb[[Variables]] <- util_map_labels(
      from,
      meta_data = meta_data,
      to = label_col,
      from = VAR_NAMES,
      ifnotfound = from)
  }

  map_label_col_to_if_exists <- function (to, default) {
    if (to %in% names(meta_data)) {
      util_map_labels(
        from
      , meta_data = meta_data
      , to = to
      , from = label_col
      , ifnotfound = from
      )
    } else {
      default
    }
  }

  vn <- map_label_col_to_if_exists(VAR_NAMES, default = from)
  lb <- map_label_col_to_if_exists(LABEL, default = from)
  llb <- map_label_col_to_if_exists(LONG_LABEL, default = from)

  if (Variables == VAR_NAMES) {
    href <- tb[[Variables]]
    data <- vn
    hover_title <- pretty_lb
  } else {
    href <- from
    data <- pretty_lb
    hover_title <- vn
  }

  .filter <- data

  href <- paste0("VAR_",
                 prep_link_escape(href, html = TRUE),
                 ".html#",
                 htmltools::urlEncodePath(prep_link_escape(as.character(href))))

  # if we have a plain_label attribute, we should consider
  # the entries in the Variables column HTML
  if (!is.null(plain_label)) {
    data <- lapply(tb[["Variables"]], htmltools::HTML)
    hover_title <- prep_title_escape(plain_label, html = TRUE)
  } else {
    data <- prep_title_escape(data, html = TRUE)
    hover_title <- prep_title_escape(hover_title, html = TRUE)
  }

  onclick <- sprintf( 'if ( window.hasOwnProperty("dq_report2")
                            && window.dq_report2
                            && window.location != "%s") {
                          if (knowID("%s")) {
                              if (window.__dqPersistPopupHistory) window.__dqPersistPopupHistory();
                              window.location = "%s";
                          } else {
                              window.alert("No result available");
                          }
                        }'
                    , href
                    , href
                    , href )

  mk_hover_text <- function(hover_title, vn, lb, llb) {
    .m <- NULL
    try({
      .m <- meta_data[meta_data[[VAR_NAMES]] == vn, , FALSE]
      .m <- data.frame(fix.empty.names = FALSE,
                      check.names = FALSE,
                      ` ` = unlist(.m[1, , TRUE]))
      .m <- .m[!is.na(.m[[1]]), , FALSE]
      .m <- head(.m, 5)
      .m <- withr::with_options(list(knitr.kable.NA = ''),
                                     htmltools::HTML(
                                       knitr::kable(.m,
                                                    "html")))
    }, silent = TRUE)
    htmltools::div(
      style = htmltools::css(
        overflow_y = "scroll",
        max_height = "80vh"),
      htmltools::h5(hover_title),
      htmltools::tags$ul(
        htmltools::tags$li(vn),
        htmltools::tags$li(lb),
        htmltools::tags$li(llb)
      ),
      .m
    )
  }

  # prep_get_labels() or l18n
  hover_text <- mapply(
    hover_title = hover_title,
    vn = vn,
    lb = lb,
    llb = llb,
    SIMPLIFY = FALSE,
    FUN = mk_hover_text)

  links <- mapply(
    href = sprintf('javascript:console.log("%s")', href),
    onclick = onclick,
    title = hover_text,
    filter = .filter,
    data,
    SIMPLIFY = FALSE,
    FUN = htmltools::a
  )


  tb[[Variables]] <- (vapply(links,
          FUN = as.character,
          FUN.VALUE = character(1)))

  var_col <- which(base::colnames(tb) == "Variables")
  if (length(var_col) == 1) {
    ..left <- seq_len(var_col - 1)
    ..right <- var_col + seq_len(ncol(tb) - var_col)
    tb <- cbind(tb[..left], Variables = vn,
                Labels = tb[["Variables"]],
                tb[..right])

    if (!rlang::is_missing(descs)) {
      if (is.null(names(descs))) {
        descs <- c(descs[..left],
                   "Variables",
                   "Labels",
                   descs[..right])
      } else {
        descs <- c(descs[..left],
                   Variables = "Variables",
                   Labels = "Labels",
                   descs[..right])
      }
    }
  }

  if(rlang::is_missing(descs)) descs <- NULL

  attr(tb, "is_html_escaped") <- is_html_escaped
  attr(tb, "description") <- description
  attr(tb[["Variables"]], "plain_label") <- plain_label

  list(tb = tb, descs = descs)
}

#' Prepares a table for use in util_html_table()
#'
#' @param tb   the table
#' @param rotate_for_one_row   see util_html_table()
#' @param copy_row_names_to_column  see util_html_table()
#' @param df_escape see util_html_table()
#'
#' @return the table
#' @keywords internal
#' @noRd
util_html_table_prepare_data <- function(tb,
                                         rotate_for_one_row,
                                         copy_row_names_to_column,
                                         df_escape) {

  class(tb) <- setdiff(class(tb), "dataquieR_result")

  is_html_escaped <- (identical(attr(tb, "is_html_escaped"), TRUE))
  description <- attr(tb, "description")
  plain_label <- NULL
  try(plain_label <- attr(tb[["Variables"]], "plain_label"), silent = TRUE)

  if (is.matrix(tb)) {
    tb <- as.data.frame.matrix(tb, stringsAsFactors = FALSE)
  }

  # convert logical columns to character strings
  # Find all columns featuring any sort of logical also "TRUE" or so, and convert it to logicals

  ts <- c("t", "true", "1", "+")
  fs <- c("f", "false", "0", "-")
  nas <- c("", "na", NA_character_)
  lgs <- unique(c(ts, fs, nas))

  is_log_col <- function(cl) {
    if (is.logical(cl)) return(TRUE)
    ..dt <- attr(cl, DATA_TYPE)
    if (!is.null(..dt) && !any(is.na(..dt))) {
      if (trimws(tolower(..dt)) %in% c("logical", "bool", "boolean")) {
        return(TRUE)
      } else {
        return(FALSE)
      }
    }
    all(
      trimws(tolower(as.character(cl))) %in% lgs
    )
  }

  as_log_col <- function(cl) {
    if (!is.logical(cl)) {
      clc <- trimws(tolower(as.character(cl)))
      r <- logical(length(cl))
      r[clc %in% ts] <- TRUE
      r[clc %in% fs] <- FALSE
      r[clc %in% nas] <- NA # not really needed
      cl <- r
    }
    cl
  }

  log_cols <- vapply(tb, is_log_col, FUN.VALUE = logical(1))

  tb[, log_cols] <-
    vapply(tb[, log_cols, drop = FALSE],
           as_log_col, FUN.VALUE = logical(nrow(tb)))

  # round all numerical columns and convert to strings
  dtypes <- prep_datatype_from_data(tb, guess_character = TRUE)
  dtypes2 <- lapply(tb, attr, DATA_TYPE)
  if (any(vapply(dtypes2, length, FUN.VALUE = integer(1)) == 1)) {
    for (cl in names(dtypes2)) {
      dt2 <- dtypes2[[cl]]
      if (!is.null(dt2) && !any(is.na(dt2))) {
        dtypes[[cl]] <- dt2
      }
    }
  }
  fltcols <- !log_cols & dtypes %in% c(DATA_TYPES$FLOAT)
  flts_still_char <-
    vapply(tb[, fltcols, drop = FALSE], function(x) util_round_to_decimal_places(as.numeric(x)),
                 FUN.VALUE = character(nrow(tb)))
  tb[, fltcols] <-
    as.numeric(ifelse(trimws(flts_still_char) == "NA", NA, trimws(flts_still_char)))
  intcols <- !log_cols & dtypes %in% c(DATA_TYPES$INTEGER)
  tb[, intcols] <- vapply(tb[, intcols, drop = FALSE], function(x) {
    if (all(suppressWarnings(!is.na(x) & !is.na(as.logical(x)) & is.na(as.integer(x))))) {
      x <- suppressWarnings(as.logical(x))
    }
    as.integer(x)
  }, FUN.VALUE = integer(nrow(tb)))

  txtcols <- !log_cols & dtypes %in% c(DATA_TYPES$STRING)
  if ("Variables" %in% names(txtcols)) {
    txtcols[["Variables"]] <- FALSE
  }
  if (VAR_NAMES %in% names(txtcols)) {
    txtcols[[VAR_NAMES]] <- FALSE
  }
  tb[, txtcols] <- lapply(tb[, txtcols, drop = FALSE], function(cl) {
    if (is.list(cl) && all(vapply(cl, length, FUN.VALUE = integer(1)) == 1L) &&
        all(vapply(lapply(cl, class), length, FUN.VALUE = integer(1)) == 1L) &&
        all(vapply(cl, is.character, FUN.VALUE = logical(1)))) {
      cl <- unlist(cl)
    }
    if (!is.list(cl) && !is.factor(cl) && length(unique(cl)) <= 10) {
      if (!any(grepl("<\\s*[a-zA-Z][^>]*>", cl, perl = TRUE))) {
        try(cl <- as.factor(cl), silent = TRUE)
      }
    }
    cl
  })

  if (copy_row_names_to_column) {
    tb <- cbind.data.frame(data.frame(Variables = rownames(tb),
                                      stringsAsFactors = FALSE), tb)
    rownames(tb) <- NULL
  }

  attr(tb, "is_html_escaped") <- is_html_escaped
  attr(tb, "description") <- description
  attr(tb[["Variables"]], "plain_label") <- plain_label

  if (df_escape) {
    tb <- util_df_escape(tb)
  }

  tb
}


#' Creates extra button definitions for datatables
#'
#' @param dl_fn  download file name
#' @param title  title of all buttons
#' @param messageBottom  bottom message
#' @param messageTop  top message
#'
#' @return list of button definitions for the DT lib
#' @keywords internal
#' @noRd
util_make_export_buttons <- function(dl_fn, title, messageBottom, messageTop) {
  exportHeader <- DT::JS("function (data, columnIdx) {
                           if ($(data).length > 0) {
                             return $(data).attr(\"colname\");
                           } else {
                             return data;
                           }
                         }")

  mkFormatBody <- function(extend, extraJS = "") {
    DT::JS(paste0("function (data, row, column, node) {
                      r = data + \"\";
                      if (r.match(/\"data:image\\//g)) {
                        r = \"not supported in ", extend, "\"
                      }
                      if (r.match(/<table *>/g)) {
                        r = \"not supported in ", extend, "\"
                      }
                      r = r.replaceAll(/\\<br *\\/?\\>/g,\"  \");",
                      extraJS,"
                      return r;
                    }"))
  }

  extraJS <- "r = $('<div />').append(r).text().trim();
              if (isNumeric(r.replaceAll(/\\s/g, ''))) {
                r = r.replaceAll(/\\s/g, '');
              }"

  mkExportOpts <- function ( extend
                           , extraJS = ""
                           , orthogonal = "filter"
                           , rows = DT::JS("rowFilter")
                           , columns = ":visible"
                           , formatBody = mkFormatBody(extend, extraJS)
                           , formatHeader = exportHeader
                           , extraFormat = list()
                           , ... ) {
    c(list(
      orthogonal = orthogonal,
      rows = rows,
      columns = columns,
      format = c(list(
          body = force(formatBody)
        , header = force(formatHeader)
      ), extraFormat)
    ), list(...))
  }

  mkButton <- function (extend
                      , .title = title
                      , .messageTop = messageTop
                      , .messageBottom = messageBottom
                      , exportOptions = mkExportOpts(extend)
                      , ...
                      ) {
      c(list(
          extend = extend
        , title = .title
        , messageTop = .messageTop
        , messageBottom = .messageBottom
        , exportOptions = exportOptions
        ), list(...)
      )
  }

  list(
    mkButton( extend = "copy")
  , mkButton( extend = "excel"
            , filename = dl_fn
            , customize = DT::JS('customize_excel')
            , exportOptions = mkExportOpts( extend = "excel"
                                          , orthogonal = "display"
                                          , extraJS = extraJS )
            , autoFilter = TRUE )
  , mkButton( extend = "csv"
            , filename = dl_fn )
  , mkButton( extend = "pdf"
            , filename = dl_fn
            , orientation = "landscape"
            , customize = DT::JS('customize_pdf')
            , exportOptions = mkExportOpts( extend = "pdf"
                                          , orthogonal = "display"
                                          , extraJS = extraJS )
            )
  , mkButton( extend = "print"
            , message = dl_fn
            , autoPrint = TRUE
            , exportOptions =
                mkExportOpts( stripHtml = FALSE
                            , orthogonal = "display"
                            , formatBody =
                                DT::JS("function(d,x,c,n){ return d + \"\"; }") )
            )
  )

}


util_dt_normalize_cell_text <- function(x, keep_first_number = TRUE) {
  x <- trimws(gsub("&nbsp;", " ",
                   gsub("(?is)<[^>]+>", "",
                        gsub("(?is)<br\\s*/?>", "\n", as.character(x), perl = TRUE),
                        perl = TRUE),
                   fixed = TRUE))

  # Treat trailing percent as numeric-ish (e.g., "30%" -> "30")
  x <- sub("\\s*%\\s*$", "", x)

  if (!keep_first_number) return(x)

  num_pat <- "[-+]?((\\d{1,3}([ ,]\\d{3})+)|(\\d+))(\\.\\d+)?([eE][-+]?\\d+)?|[-+]?\\.\\d+([eE][-+]?\\d+)?"
  m <- regexpr(num_pat, x, perl = TRUE)
  trimws(ifelse(m > 0, regmatches(x, m), x))
}

util_get_datatables_alignments <- function(
    tb,
    prefer_right_for_datetime = TRUE,
    right_types = c(DATA_TYPES$INTEGER, DATA_TYPES$FLOAT),
    center_types = character(0),
    ...
) {
  stopifnot(is.data.frame(tb))

  Map(function(col, j) {
    align <- if (is.integer(col) || is.numeric(col)) "dt-right"
    else if (inherits(col, c("POSIXct", "POSIXt", "Date")))
      if (prefer_right_for_datetime) "dt-right" else "dt-center"
    else if (is.logical(col)) "dt-center"
    else {
      x <- util_dt_normalize_cell_text(col, TRUE)
      x <- x[!is.na(x) & x != ""]
      dtype <- if (!length(x)) DATA_TYPES$INTEGER else prep_robust_guess_data_type(x, ...)
      if (dtype %in% right_types) "dt-right"
      else if (dtype %in% center_types) "dt-center"
      else if (dtype %in% c(DATA_TYPES$DATETIME, DATA_TYPES$TIME))
        if (prefer_right_for_datetime) "dt-right" else "dt-center"
      else "dt-left"
    }
    list(className = align, targets = j - 1L)
  }, unname(tb), seq_along(tb))
}

#' Find a free variable name in the metadata
#'
#' looks up the column `target` in `meta_data` and searches for an unused entry
#' starting with `prefix`. Also excludes `also_not` from the possible results.
#'
#' @param meta_data item-level metadata.
#' @param prefix prefix for the returned variable name
#' @param target target column in metadata for uniqueness
#' @param also_not other possible ambiguous names to exclude
#'
#' @returns an unused unique name
#' @noRd
util_free_varname <- function(meta_data, prefix, target, also_not) {
  res <- prefix
  i <- 0
  while (res %in% c(meta_data[[target]], also_not)) {
    i <- i + 1
    res <- paste0(prefix, "_", i)
  }
  res
}

#' HTML Dependency for `tippy`
#'
#' @return the dependency
html_dependency_tippy <- function() {
  htmltools::htmlDependency(
    name = "tippy",
    version = "6.7.3",
    src = system.file("tippy", package = "dataquieR"),
    script = c("core.js", "tippy.js")
  )
}

cat1 <- as.character(util_as_cat(1))
cat2 <- as.character(util_as_cat(2))
cat3 <- as.character(util_as_cat(3))
cat4 <- as.character(util_as_cat(4))
cat5 <- as.character(util_as_cat(5))


util_funins <- function(f, expr, after = 1) { # https://stackoverflow.com/a/38733185
  expr <- substitute(expr)
  rlang::fn_body(f)<-as.call(append(as.list(rlang::fn_body(f)), expr, after=after))
  f
}

util_funwrap <- function(f, fn) {
  rlang::fn_body(f)<-
    call("{",
         call(fn, rlang::fn_body(f))
    )
  f
}

util_first_arg_study_data_or_resp_vars <- function() {
  cl <- rlang::caller_call()
  clan <- rlang::call_args_names(cl)
  cf <- rlang::caller_fn()
  formals_names <- names(formals(cf))
  if (length(formals_names) > 1 &&
      formals_names[[1]] %in% c("resp_vars", "variable_group", "label_col") &&
      "study_data" %in% formals_names  &&
      length(clan) >= 1
  ) {
    study_data_formal_pos <- which(formals_names == "study_data")
    first_arg_nm <- formals_names[[1]]
    if (clan[[1]] == "" &&
        !first_arg_nm %in% clan &&
        !"study_data" %in% clan) {
      missing_first_arg <- eval(call("missing",
                                     as.symbol(first_arg_nm)),
                                envir = rlang::caller_env())
      if (!missing_first_arg) {
        first_arg_val <- get(first_arg_nm, rlang::caller_env())
        if (is.data.frame(first_arg_val) ||
            isTRUE(try(first_arg_val %in% prep_list_dataframes(),
                       silent = TRUE)) ||
            is.data.frame(try(prep_get_data_frame(first_arg_val,
                                                  column_names_only = TRUE,
                                                  keep_types = TRUE),
                              silent = TRUE))) {
          args <- rlang::call_args(cl)[-1]

          # args <- c(setNames(list(formals(cf)[[1]]),
          #                    nm = first_arg_nm),
          #           rlang::call_args(cl)[-1])
          args[["study_data"]] <- first_arg_val
          repl_call <-
            do.call(call, c(list(rlang::call_name(cl)), args), quote = TRUE)
          return(repl_call)
        }
      }
    }
  }
  return(NULL)
}

util_maybe_eval_to_dataquieR_result <- function(my_call) {
  # detect recursion
  recursive_call <-
    sum(as.character(rlang::call_name(sys.call())) == vapply(sys.calls(),
                                                             function(x) {
                                                               r <-
                                                                 rlang::call_name(
                                                                   x
                                                                 )
                                                               if (is.null(r))
                                                                 r <-
                                                                   NA_character_
                                                               return(r)
                                                             },
                                                             FUN.VALUE =
                                                               character(1))) > 1
  if (is.na(recursive_call)) {
    recursive_call <- FALSE
  }
  my_call <- substitute(my_call)
  if (.called_in_pipeline || # TODO: in pipielines, the conditions are already collected, this may be change, later:
      getOption("dataquieR.testdebug", dataquieR.testdebug_default) ||
      getOption("dataquieR.dontwrapresults", dataquieR.dontwrapresults_default) ||
      identical(Sys.getenv("TESTTHAT"), "true") ||
      recursive_call) {  # don't do, if in testthat runs, so that expect_condition still works.
    eval.parent(my_call)
  } else {
    call_to_attach <- as.symbol("<call not found>")
    if (util_is_try_error(try(call_to_attach <- rlang::call_match(rlang::caller_call(),
                                                                  rlang::caller_fn()), silent = TRUE))) {
      util_warning(c("Internal Warning, sorry. Please report: This will",
                     "show a deprecation warning, but it should anyway",
                     "not happen."))
      try(call_to_attach <- rlang::call_match(my_call, rlang::call_fn(my_call)),
          silent = TRUE)
    }
    fn <- rlang::call_name(rlang::caller_call())
    .v <- try(eval(as.symbol("resp_vars"), envir = rlang::caller_env()),
              silent = TRUE)
    if (!util_is_try_error(.v) && !is.null(.v)) {
      nm <- paste0(.v, collapse = SPLIT_CHAR)
    } else {
      nm <- "[ALL]"
    }
    nm <- paste0(fn, ".", nm)
    r <- util_eval_to_dataquieR_result(my_call,
                                       nm = nm,
                                       function_name = fn,
                                       filter_result_slots = c(),
                                       env = parent.frame(),
                                       checkpoint_resumed = FALSE,
                                       called_in_pipeline = FALSE) # dont handle stuff twice
    if (!util_is_try_error(.v))
      attr(call_to_attach, "entity_name") <- .v
    .v <- try(eval(as.symbol("label_col"), envir = rlang::caller_env()),
              silent = TRUE)
    if (!util_is_try_error(.v))
      attr(call_to_attach, "label_col") <- .v
    attr(r, "call") <- call_to_attach
    attr(r, "function_name") <- fn
    r
  }
}

preps <- quote({
  # maybe fix order ####
  ....alt_call_res <- NULL
  ..cl <- util_first_arg_study_data_or_resp_vars()
  if (is.call(..cl)) {
    # util_warning("%s", util_pretty_vector_string(ls(rlang::caller_env(3))))
    # util_message("Not calling %s, calling %s, instead...",
    #              sQuote(util_deparse1(rlang::caller_call(0))),
    #              sQuote(util_deparse1(..cl)))
    ....alt_call_res <- force(eval(..cl, rlang::caller_env(3)))
    return(....alt_call_res)
  } else {
    rm(..cl)
  }
  #######
  if (!.called_in_pipeline && # TODO: in pipielines, the conditions are already collected, this may be change, later:
      !getOption("dataquieR.testdebug", dataquieR.testdebug_default) &&
      !identical(Sys.getenv("TESTTHAT"), "true")) {  # don't do, if in testthat runs, so that expect_condition still works.
    my_call <- rlang::caller_call(2)
    if ("meta_data_v2" %in% names(formals()) &&
        !is.null(my_call$meta_data_v2)) {
      util_maybe_load_meta_data_v2()
      meta_data_v2 <- rlang::missing_arg()
      my_call$meta_data_v2 <- NULL
    }
    # IDEA: Called recursively?
    # search for study-data:
    env_args <- formals(util_meta_data_env)
    ..what_i_have <- intersect(names(env_args), ls())
    missing_env_args <- vapply(..what_i_have, FUN.VALUE = logical(1),
                               function(object) {
                                 .m <- call("missing", object)
                                 r <- try(eval(.m, envir = rlang::caller_env(2)), silent = TRUE)
                                 return(identical(r, TRUE))
                               })
    ..what_i_have <- ..what_i_have[!missing_env_args]
    env_args[..what_i_have] <- mget(..what_i_have,
                                    inherits = FALSE)
    my_call <- rlang::call_match(my_call, rlang::current_fn())
    if (is.null(my_call$study_data) && "study_data" %in% names(formals())) {
      if (FALSE) { # FIXME ES: check the dataframe level metadata
        util_message("Using %s from your dataframe level metadata",
                     dQuote("study_data"))
        env_args <- list(study_data = .sd)
      } else if ("study_data" %in% prep_list_dataframes()) {
        util_message("Using %s from the dataframe cache.", dQuote("study_data"))
        .sd <- prep_get_data_frame("study_data")
        env_args <- list(study_data = .sd)
      } else if (exists("study_data", envir = rlang::caller_env(3))) {
        util_message("Using %s from your calling environment",
                     sQuote("study_data"))
        .sd <- get("study_data", envir = rlang::caller_env(3))
        env_args <- list(study_data = .sd)
      } else if (exists("study_data", envir = rlang::global_env())) {
        util_message("Using %s from your glabal environment",
                     sQuote("study_data"))
        .sd <- get("study_data", envir = rlang::global_env())
        env_args <- list(study_data = .sd)
      } else {
        util_warning(
          c("likely, the call will fail, because I could not find",
            "some %s, anywhere. Please pass them in your call or",
            "put them in your environment or data frame cache with",
            "the name %s"),
          sQuote("study_data"),
          dQuote("study_data"))
      }
    }
    .mde <- do.call(util_meta_data_env, env_args, quote = FALSE)
    #
    my_call_orig <- my_call
    my_call <- try(.mde$provisionize_call(my_call, internal = TRUE, # TODO: also handle other condtions
                                          env = environment()), silent = TRUE)
    if (util_is_try_error(my_call)) {
      if (.called_in_pipeline || # TODO: in pipielines, the conditions are already collected, this may be change, later:
          getOption("dataquieR.testdebug", dataquieR.testdebug_default) ||
          getOption("dataquieR.dontwrapresults", dataquieR.dontwrapresults_default) ||
          identical(Sys.getenv("TESTTHAT"), "true")) {  # don't do, if in testthat runs, so that expect_condition still works.
        util_error(my_call)
      } else {
        function_name <- "dataquieR"
        nm <- paste0(function_name, ".?????")
        try({
          function_name <- rlang::call_name(attr(my_call, "condition")$trace[[1]][[1]])
          fkt <- get(function_name,
                     mode = "function")
          rvs <- rlang::call_match(attr(my_call, "condition")$trace[[1]][[1]],
                                   fkt)[["resp_vars"]]
          if (length(rvs) > 1) rvs <- "[ALL]"
          if (length(rvs) < 1) rvs <- "?"
          gvs <- rlang::call_match(attr(my_call, "condition")$trace[[1]][[1]],
                                   fkt)[["group_vars"]] # TODO: this is not the name of the meta col but the content, but better than nothing
          if (length(gvs) == 1)
            nm <- paste0(function_name, "_", tolower(gvs), ".", rvs)
          else
            nm <- paste0(function_name, ".", rvs)

        },
        silent = TRUE
        )
        .r <- util_eval_to_dataquieR_result(
          util_error(my_call),
          function_name = function_name,
          nm = nm
        )
        class(.r) <- unique(c(class(.r), "master_result"))
        ....alt_call_res <- .r
        return(....alt_call_res)
      }
    }

    for(n in names(my_call)) {
      if (n %in% names(formals())) {
        if (!is.language(my_call[[n]])) {
          .v <- try(my_call[[n]], silent = TRUE)
        } else {
          .v <- try(eval(as.symbol(n)), silent = TRUE)
        }
        if (util_is_try_error(.v) || is.null(.v)) {
          .v2 <- eval.parent(my_call[[n]], n = 2)
          if (!util_is_try_error(.v2)) {
            .v <- .v2
          }
        }
        assign(n, .v)
      }
    }
  }
})

# TODO: Write a test for the following
util_decorator <- function(x,
                           f_nm = as.character(substitute(x))) {
  f <- util_funwrap(x, "util_maybe_eval_to_dataquieR_result")

  # reverse order (appends code)
  f <- util_funins(f,
                   if (!is.null(....alt_call_res)) {
                     return(....alt_call_res)
                   }
  )
  # f <- util_funins(f,
  # )
  f <- util_funins(f, eval(preps))
  # /reverse order (appends code)
  f
}

for (fkt in ls(pattern = "^(int|com|con|acc|des)_")) {
  assign(fkt, util_decorator(get(fkt), fkt))
}

try({
  if (suppressWarnings(util_ensure_suggested("pkgload", err = FALSE,
                                             goal =
                                             "provide help on report sections during development"))) {
    dev_package <- pkgload::is_dev_package(utils::packageName())
  } else {
    dev_package <- FALSE
  }

  if (dev_package) {
    roxygenise_call <- 0
    called_from_rogygen <- FALSE
    while (!is.null(cl <- rlang::caller_call(roxygenise_call)) &&
           !called_from_rogygen) {
      ns <- rlang::call_ns(cl)
      nm <- rlang::call_name(cl)
      if (is.null(ns)) ns <- "" else ns <- paste0(ns, "::")
      if (paste0(ns, nm) == "roxygen2::roxygenise") {
        called_from_rogygen <- TRUE
      } else {
        roxygenise_call <- roxygenise_call + 1
      }
    }
    if (called_from_rogygen) withr::defer(local({
      use_cli_format <-
        suppressWarnings(util_ensure_suggested("cli", err = FALSE,
                                               goal =
                                                 "nicer messages"))
      if (use_cli_format) {
        cmd <- sprintf("{.run [%s](%s)}", "dataquieR:::util_load_manual(TRUE)",
                       "dataquieR:::util_load_manual(TRUE)")
      } else {
        cmd <- "dataquieR:::util_load_manual(TRUE)"
      }
      need_rebuild <- FALSE
      pp <- dynGet("base_path", ifnotfound=getwd(), inherits=TRUE)
      withr::local_dir(pp)
      max_man <- max(vapply(list.files("man/", full.names = TRUE, all.files = TRUE,
                                       pattern = "*.Rd", ignore.case = TRUE), file.mtime,
                            FUN.VALUE = file.mtime("/")), na.rm = TRUE)
      rd_files <- sort(dir("man", pattern = "\\.[Rr]d$", full.names = TRUE))
      md5 <- tools::md5sum(rd_files)
      man_hash <- rlang::hash(unname(paste(names(md5), md5, sep = "\n")))
      old_man_hash <- ""
      try(
        old_man_hash <- rio::import("inst/manual.RData",
                                    which = "man_hash",
                                    trust = TRUE),
        silent = TRUE
      )
      if (old_man_hash == "" || old_man_hash != man_hash) {
        .update <- file.mtime("inst/manual.RData") < max_man
        .update <- .update || is.na(.update)
        if (.update) {
          rlang::inform(
            use_cli_format = use_cli_format,
            c(i = cli::format_inline(sprintf(paste("inst/manual.RData may be out of date, as a dataquieR developer,",
                                                   "please, consider updating it with %s -- doing this, now"),
                                             cmd))))
          need_rebuild <- TRUE
        }
      }
      old_man_hash <- ""
      try(
        old_man_hash <- rio::import("inst/indicator_or_descriptor.RData",
                                    which = "man_hash",
                                    trust = TRUE),
        silent = TRUE
      )
      if (old_man_hash == "" || old_man_hash != man_hash) {
        .update <- file.mtime("inst/indicator_or_descriptor.RData") < max_man
        .update <- .update || is.na(.update)
        if (.update) {
          rlang::inform(
            use_cli_format = use_cli_format,
            c(i = cli::format_inline(sprintf(paste("inst/indicator_or_descriptor.RData may be out of date, as a",
                                                   " dataquieR developer, please, consider updating it with",
                                                   "%s -- doing this, now"), cmd))))
          need_rebuild <- TRUE
        }
      }
      if (need_rebuild) {
        if (suppressWarnings(util_ensure_suggested("devtools", err = FALSE,
                                                   goal =
                                                   "provide help on report sections during development")) &&
            suppressWarnings(util_ensure_suggested("callr", err = FALSE,
                                                   goal =
                                                   "provide help on report sections during development"))) {

          recursive_call <- Sys.getenv("dataquieR_child", "") == "TRUE"
          if (!recursive_call) callr::r(env =
                                          c(dataquieR_child = "TRUE"),
                                        function(man_hash, pp) {
                                          withr::local_dir(pp)
                                          devtools::document()
                                          devtools::load_all()
                                          util_fix_backticks()
                                          util_load_manual(TRUE, man_hash = man_hash)
                                        }, args = list(man_hash = man_hash, pp = pp))
        }
      } else {
        util_fix_backticks()
      }
    }), envir = rlang::caller_env(roxygenise_call), priority = "last")
  }
}, silent = !TRUE)

.git_hash <- try(suppressWarnings(suppressMessages(
  system("git rev-parse HEAD",
         intern = TRUE,ignore.stderr = TRUE
  )
)), silent = TRUE)

if (util_is_try_error(.git_hash) ||
    !is.character(.git_hash) ||
    length(.git_hash) != 1 ||
    !grepl("^[a-z0-9]{40}$", .git_hash)) {
  .git_hash <- ""
}

util_dataquieR_version <- function() {
  gh <- .git_hash
  if (nzchar(gh)) {
    gh <- paste0(" (", gh, ")")
  }
  sprintf("%s%s",
          as.character(packageVersion(
            utils::packageName())),
          gh)
}

SSI_functions <-
  unlist(unique(lapply(strsplit(unlist(lapply(util_get_concept_info("ssi")$functions,
                                              util_parse_assignments, multi_variate_text = TRUE)), split = ".",
                                fixed = TRUE), `[[`, 1)))

# Hint: Debug with warpper turned off: withr::with_options(list(warn = 2, dataquieR.dontwrapresults = TRUE), com_item_missingness(study_data, resp_vars = "SBP_0", item_level = md0, suppressWarnings = TRUE))

# IDEA: util_vectorize for a more tailored version of:
# print(lapply(c("SBP_0", "DBP_0"), acc_margins))
# print(lapply(c("SBP_0", "DBP_0"), com_item_missingness))
# print(Vectorize(com_item_missingness, SIMPLIFY = FALSE)(c("SBP_0", "DBP_0")))
# print(Vectorize(acc_varcomp, SIMPLIFY = FALSE)(c("SBP_0", "DBP_0")))
# print(Vectorize(acc_margins, SIMPLIFY = FALSE)(c("SBP_0", "DBP_0")))
# print(lapply(c("SBP_0", "DBP_0"), \(x) acc_margins(resp_vars = x)))

#' Check for duplicated IDs
#'
#' @description
#' This function tests for duplicates entries in identifiers. It is possible to
#' check duplicated identifiers by study segments or to consider only selected
#' segments.
#'
#' @param id_vars_list [list] id variable names for each segment or data frame
#' @param study_data [data.frame] the data frame that contains the measurements, mandatory.
#' @param meta_data [data.frame] the data frame that contains metadata attributes of the study data, mandatory.
#' @param level [character] a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").
#' @param repetitions [vector] an integer vector indicating the number of allowed repetitions in the id_vars. Currently, no repetitions are supported.
#' @param study_segment [vector] the segments or data frame names being assessed
#'
#' @inheritParams .template_function_indicator
#'
#' @return a [list] with
#'   - `SegmentData`: data frame with the results of the quality check for duplicated identifiers
#'   - `SegmentTable`: data frame with selected duplicated identifiers check results, used for the data quality report.
#'   - `Other`: named list with inner lists of unique cases containing each the
#'              row indices of duplicated identifiers separated by "|" , if any.
#'              outer names are names of the segments.
#'              Use [prep_get_study_data_segment()] to get the data frame the
#'              indices refer to.
#'
#' @family integrity_indicator_functions
#' @concept integrity_indicator
#' @keywords internal
util_int_duplicate_ids_segment <- function(level = c("segment"),
                                           id_vars_list,
                                           study_segment,
                                           repetitions,
                                           study_data,
                                           meta_data,
                                           meta_data_segment = "segment_level",
                                           segment_level
                                           ) {

  util_ck_arg_aliases()

  level <- util_match_arg(level)

  # Segment level ----

  if (missing(id_vars_list) &&
      missing(study_segment) &&
      missing(repetitions) &&
      missing(meta_data_segment) &&
      formals()$meta_data_segment %in% prep_list_dataframes()) {
    meta_data_segment <- force(meta_data_segment)
  }

  if (missing(id_vars_list) &&
      missing(study_segment) &&
      missing(repetitions) &&
      !missing(meta_data_segment)) {
    meta_data_segment <- prep_check_meta_data_segment(meta_data_segment)

    meta_data_segment <-
      meta_data_segment[!util_empty(meta_data_segment[[SEGMENT_ID_VARS]]), ,
                        drop = FALSE
      ]
    meta_data_segment <-
      meta_data_segment[!util_empty(meta_data_segment[[SEGMENT_UNIQUE_ID]]), ,
                        drop = FALSE
      ]
    # TODO: if nothing left
    id_vars_list <- lapply(setNames(meta_data_segment[[SEGMENT_ID_VARS]],
                                    nm = meta_data_segment[[STUDY_SEGMENT]]),
                           util_parse_assignments,
                           multi_variate_text = TRUE );
    id_vars_list <- lapply(id_vars_list, unlist, recursive = TRUE)
    id_vars_list <- lapply(setNames(meta_data_segment[[SEGMENT_ID_VARS]],
                                    nm = meta_data_segment[[STUDY_SEGMENT]]),
                           util_parse_assignments,
                           multi_variate_text = TRUE
    )
    id_vars_list <- lapply(id_vars_list, unlist, recursive = TRUE)

    id_vars_list <- lapply(id_vars_list,
                           util_map_labels,
                           meta_data = meta_data,
                           to = label_col)

    study_segment <- meta_data_segment[[STUDY_SEGMENT]];
    repetitions <- meta_data_segment[[SEGMENT_UNIQUE_ID]]
  } else if (!missing(meta_data_segment)) {
    util_error(c("I have %s and one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "study_segment",
                   "repetitions",
                   "meta_data_segment"
                 )),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "study_segment",
                   "repetitions",
                   "meta_data_segment"
                 )))
  } else if (missing(meta_data_segment) && (
    missing(id_vars_list) ||
    missing(study_segment) ||
    missing(repetitions)
  )) {
    util_error(c("I don't have %s and also miss at least",
                 "one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "study_segment",
                   "repetitions"
                 )),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "study_segment",
                   "repetitions"
                 )))
  }


  # Checks ----

  # map metadata to study data
  prep_prepare_dataframes(.allow_empty = TRUE)

  # ID variable(s) must be specified
  util_stop_if_not(is.list(id_vars_list))
  util_stop_if_not(all(vapply(id_vars_list, is.character,
                              FUN.VALUE = logical(1))))

  util_expect_scalar(study_segment,
                     allow_null = TRUE,
                     allow_more_than_one = TRUE,
                     check_type = is.character)

  util_stop_if_not(length(study_segment) == length(id_vars_list))
  util_stop_if_not(length(study_segment) == length(repetitions))
  names(repetitions) <- study_segment

  util_ensure_in(names(id_vars_list), study_segment)
  util_ensure_in(study_segment, names(id_vars_list))

  level <- util_match_arg(level)

  # Check segments ----
  # check that specified segments are included in the metadata

  old_study_segment <- study_segment
  study_segment <- intersect(study_segment, meta_data[[STUDY_SEGMENT]])

  if (length(old_study_segment) > length(study_segment)) {
    util_message(
      "The study_segment in the %s do not match the study_segment in %s, considering only the intersection",
      dQuote("meta_data"),
      dQuote("meta_data_segment"),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = FALSE
    )
  }

  label_col <- attr(ds1, "label_col")

  # Check for duplicated record IDs  ----
  result <- lapply(setNames(nm = study_segment), function(current_segment) {

    reps <- repetitions[[current_segment]]

    id_vars <- id_vars_list[[current_segment]]
    id_vars <- util_find_var_by_meta(resp_vars = id_vars,
                                     meta_data = meta_data,
                                     label_col = label_col,
                                     target = label_col,
                                     ifnotfound = id_vars)
    id_vars <-
      util_ensure_in(id_vars,
                     colnames(ds1),
                     err_msg =
                       c(sprintf("ID variables in current segment %s",
                                 dQuote(current_segment)),
                         ": Missing %s from the study data,",
                         "did you mean %s? I'll remove the missing entries"),
                     applicability_problem = TRUE)


    if (length(id_vars) == 0) {
      util_warning(
        "The %s (%s) in the %s are not included in the %s",
        "ID variables",
        util_pretty_vector_string(id_vars_list[[current_segment]]),
        "metadata",
        "study data",
        applicability_problem = TRUE
      )
      return(data.frame(
        check.names = FALSE,
        "Check" = "IDs",
        "Segment" = current_segment,
        "Any duplicates" = NA,
        "Number of duplicates" = NA_real_,
        "Percentage of duplicates" = NA_real_,
        "GRADING" = NA_real_,
        stringsAsFactors = FALSE
      ))
    }

    vars_in_current_segment <-
      util_get_vars_in_segment(
        segment = current_segment,
        meta_data = meta_data,
        label_col = label_col
      )


    vars_in_current_segment <-
      util_ensure_in(vars_in_current_segment,
                     colnames(ds1),
                     err_msg =
                       c(sprintf("Study variables from current segment %s",
                                 dQuote(current_segment)),
                         ": Missing %s from the study data,",
                         "did you mean %s? I'll remove the missing entries"),
                     applicability_problem = TRUE)


    ds1_vars_in_current_segment <- ds1[, c(id_vars, vars_in_current_segment)]

    ds_reduced <- util_remove_empty_rows(
      ds1_vars_in_current_segment,
      id_vars = id_vars)


    dup_info <- util_find_duplicated_rows(
      study_data = ds_reduced,
      id_vars = id_vars,
      repeptitions = reps)

    n_uniq <- nrow(ds_reduced) - sum(dup_info$unexp_reps)

    n_total <-
      nrow(ds_reduced)

    res_tmp <- data.frame(
      check.names = FALSE,
      "Check" = "IDs",
      "Segment" = current_segment,
      "Any duplicates" = !!nrow(dup_info),
      "Number of duplicates" = n_total - n_uniq,
      "Percentage of duplicates" =
        round(100 * (n_total - n_uniq) /
                n_total, 3),
      "GRADING" = as.integer(n_uniq < n_total),
      "ID Vars" = prep_deparse_assignments(id_vars, mode = "string_codes"),
      stringsAsFactors = FALSE
    )

    if (res_tmp[[3]]) { # only if there are any duplicated observations
      vec_dup <- dup_info$which
    } else {
      vec_dup <- NULL
    }

    return(list(res_tmp, vec_dup))
  })

  res_df <- do.call(rbind.data.frame, lapply(result, `[[`, 1))

  duplicates_df <- lapply(result, `[[`, 2)

  res_pipeline <- data.frame(
    "Segment" = res_df$Segment,
    "NUM_int_sts_dupl_ids" = res_df$`Number of duplicates`,
    "PCT_int_sts_dupl_ids" = res_df$`Percentage of duplicates`,
    "GRADING" = res_df$GRADING,
    stringsAsFactors = FALSE
  )

  return(list(
    SegmentData = res_df,
    SegmentTable = res_pipeline,
    Other = duplicates_df
  ))

}

#' Create a `plotly` pie chart
#'
#' @param data data as returned by `prep_summary_to_classes` but
#'             summarized by one column (currently, we support
#'             `indicator_metric`, `call_names`,
#'              `STUDY_SEGMENT`, and `VAR_NAMES`)
#' @param meta_data [meta_data]
#'
#' @return a `htmltools` compatible object
#'
#' @family summary_functions
#' @export
prep_render_pie_chart_from_summaryclasses_plotly <- function(data, # FIXME: If use_plotly is FALSE?
                                                             meta_data = "item_level") {
  vars_to_include <- attr(data, "vars_to_include")
  ssi <- (identical(vars_to_include, "ssi"))
  # FIXME: Amend prep_render_pie_chart_from_summaryclasses_ggplot2 to handle this also.
  te <- topenv(parent.frame(1)) # see https://stackoverflow.com/a/27870803
  if (!(isNamespace(te) && getNamespaceName(te) == "dataquieR")) {
    lifecycle::deprecate_soft("2.1.0.9007",
                              "prep_render_pie_chart_from_summaryclasses_plotly()",
                              "plot.dataquieR_summary()")
  }

  if (!util_ensure_suggested(c("plotly", "htmltools"), err = FALSE)) {
    return(NULL)
  }

  if (nrow(data) == 0) {
    return(htmltools::browsable(htmltools::HTML("")))
  }

  grouped_by <- setdiff(colnames(data), c("class", "value", "percent", "note"))

  util_stop_if_not(length(grouped_by) == 1)

  groups <- unique(data[[grouped_by]])

  util_stop_if_not(length(groups) > 0)

  if (length(groups) > 1) {
    all_pys <- lapply(setNames(nm = groups), function(g) {
      prep_render_pie_chart_from_summaryclasses_plotly(
        data[data[[grouped_by]] == g, , FALSE], meta_data = meta_data)
    })
    # pys <-
    #   do.call(plotly::subplot, lapply(all_pys, plotly::subplot))
    # py <- plotly::layout(pys,
    #                # title = list(text =
    #                #                res$SummaryPlot$patches$annotation$title),
    #                margin = 0.01)
    ncols <- min(2, ceiling(sqrt(length(all_pys))))
    nrows <- ceiling(length(all_pys) / ncols)

    pys_matrix <- htmltools::tags$table(lapply(
      seq_len(nrows),
      function(rw) {
        if ((rw - 1) * ncols + 1 <= length(all_pys)) { # current row needed?
          cur_tr_as_list <- lapply(seq_len(ncols), function(rw, cl) {
            wch <- (rw - 1) * ncols + cl
            if (wch <= length(all_pys)) {
              o <- htmltools::tags$td(all_pys[wch])
            } else {
              o <- htmltools::tags$td()
            }
            o
          }, rw = rw)
          do.call(htmltools::tags$tr, cur_tr_as_list)
        } else { # should never be reached
          NULL
        }
      }
    ))

    py <- pys_matrix

    py <- htmltools::tags$table(py)
    py <- htmltools::browsable(py)
    return(py)
  }

  py_colors <- util_get_colors()
  py_colors["NA"] <- "lightgrey"
  labs <- util_get_labels_grading_class()
  labs["NA"] <- "Not classified"

  # wrap the text of the labs if too long. Max no. characters = 20
  no_char_labs<- vapply(labs, FUN = function(x){
    no_char <- nchar(x)
  }, FUN.VALUE = numeric(1))

  if (any(no_char_labs > 10)) {
    labs<- vapply(labs, FUN = function(x){
      #remove white spaces at beginning or end
      x <- trimws(x)

      #If labs >20 characters, cut it at 20
      if (nchar(x) > 20) {
        x <- substr(x, start = 1, stop = 20)
      }

      sst <- strsplit(x, '')[[1]]  #from https://stackoverflow.com/questions/11619616/how-to-split-a-string-into-substrings-of-a-given-length
      m <- matrix('', nrow=10,
                  ncol=(length(sst)+10-1)%/%10)
      m[seq_along(sst)] <- sst
      x <- apply(m, 2, paste, collapse='')
      rm(sst, m)
      #remove white spaces at beginning or end
      x <- trimws(x)
      x <- paste0(x, collapse = "<br>")
    }, FUN.VALUE = character(1))
  }


  if (all(is.na(data$class))) {
    return(htmltools::browsable(htmltools::HTML("")))
  }

  if (is.factor(data$class)) {
    data$class <- as.integer(gsub("^cat", "", data$class))
  }
  data$class <- factor(data$class,
                       levels = names(py_colors),
                       ordered = TRUE
  )

  data <- data[order(data$class, -data$value, decreasing = TRUE), , FALSE]
  hoverinfo <- "label+percent+value"

  if ("note" %in% colnames(data)) {
    hoverinfo <- paste0(hoverinfo, "+text")
  } else {
    data$note <- ""
  }

  if (length(groups) > 1) { # IDEA: sunburst would maybe work here, if we have a hiearchy in groups, currently, we do not have that.
    # Sunburst:
    # py <- plotly::plot_ly(
    #   ids = c("Eve", "Cain", "Seth", "Enos", "Noam", "Abel", "Awan", "Enoch", "Azura"),
    #   labels = paste(c("Eve", "Cain", "Seth", "Enos", "Noam", "Abel", "Awan", "Enoch", "Azura"), 42),
    #   parents = c("", "Eve", "Eve", "Seth", "Seth", "Eve", "Eve", "Awan", "Eve"),
    #   values = c(10, 14, 12, 10, 2, 6, 6, 4, 4),
    #   type = 'sunburst'
    # )
  } else {
    rotation_value <- 90
    #order data frame "data"
    data <- data[order(data$class),]


    py <- plotly::add_pie(plotly::plot_ly(data,
                                          height = 400,
                                          width = 400),
                          sort = FALSE,
                          direction = "clockwise",
                          rotation = rotation_value,
                          labels = labs[
                            paste(data$class)],
                          values = data$value,
                          hovertext = data$note,
                          #                name = "Segment",
                          #                hoverinfo = "label+percent+name",
                          hoverinfo = hoverinfo,
                          textinfo = 'label+percent+value',
                          showlegend = FALSE,
                          marker = list(
                            colors = py_colors[paste(data$class)]))
  }

  if (identical(grouped_by, "indicator_metric")) { # TODO: Hiearchical structure support
    title <- util_translate_indicator_metrics(groups, short = FALSE,
                                              long = FALSE)
    subtitle <- "percentage of QA classes"
  } else if (identical(grouped_by, "call_names")) { # TODO: Hiearchical structure support
    fnms <- util_cll_nm2fkt_nm(groups)
    if (nchar(fnms) < nchar(groups)) {
      suff <- gsub("^.*_", ": ", groups)
      substr(suff, 3, 3) <- toupper(substr(suff, 3, 3))
    } else {
      suff <- ""
    }
    title <- paste0(util_map_labels(fnms,
                                    util_get_concept_info("implementations"),
                                    to = "dq_report2_short_title",
                                    from = "function_R",
                                    ifnotfound = NA_character_), suff)
    subtitle <- "percentage of QA classes"
  } else if (identical(grouped_by, as.character(STUDY_SEGMENT))) { # TODO: Hiearchical structure support
    title <- groups
    subtitle <- "percentage of QA classes"
  } else if (identical(grouped_by, as.character(VAR_NAMES))) {
    util_expect_data_frame(meta_data)
    title <- prep_get_labels(groups,
                             meta_data = meta_data,
                             label_class = "LONG")
    subtitle <- "percentage of QA classes"
  } else if (identical(grouped_by, as.character("function_name"))) {
    title <- vapply(groups, util_alias2caption, long = TRUE,
                    FUN.VALUE = character(1))
    subtitle <- groups
  } else {
    title <- groups
    subtitle <- "percentage of QA classes"
    #subtitle <- NULL
    #    util_error("Unkown grouping by %s", sQuote(grouped_by))
  }

  if (!ssi) # FIXME: also for ssi
    subtitle <- paste(subtitle, sprintf(" -- %d of %d %s classified",
                                        sum(data$value, na.rm = TRUE),
                                        nrow(meta_data),
                                        ifelse(ssi, "scales", "variables")
    )) # TODO: Maybe, we should not compute this here, but earlier.


  #Define the space on top among the title and the plot conditionally
  angles <- data$value / sum(data$value) * 360
  cs1 <- cumsum(data$value / sum(data$value) * 360) #final limits in case started on top
  cs1 <- cs1 + rotation_value # final limits in case started at rotation_value

  lower_bound <- 345
  upper_bound <- 360
  values_in_range1 <- (cs1 >= lower_bound & cs1 <= upper_bound)
  count_in_range1 <- sum(values_in_range1)

  lower_bound <- 360
  upper_bound <- 375
  values_in_range2 <- (cs1 >= lower_bound & cs1 <= upper_bound)
  count_in_range2 <- sum(values_in_range2)


  #for bottom margin
  lower_bound <- 150
  upper_bound <- 180
  values_in_range3 <- (cs1 >= lower_bound & cs1 <= upper_bound)
  count_in_range3 <- sum(values_in_range3)

  lower_bound <- 180
  upper_bound <- 210
  values_in_range3 <- (cs1 >= lower_bound & cs1 <= upper_bound)
  count_in_range4 <- sum(values_in_range3)


  #Define the values for white spaces around the plot
  value_bottom_conditional <- 50
  value_on_top_conditional <- 50

  if(all(angles >= 60)) {
    #All the text is inside the plot, no need for extra space at top or bottom
    value_on_top_conditional <- 0
    value_bottom_conditional <- 0
  } else {
    if (count_in_range1 >= 2 || count_in_range2 >= 2) {
      value_on_top_conditional <- 170
    } else {
      value_on_top_conditional <- 60
    }

    if (count_in_range3 >= 2 || count_in_range4 >= 2) {
      value_bottom_conditional <- 170
    } else {
      value_bottom_conditional <- 60
    }
  }

  py <- plotly::layout(py,
                       title =
                         list(
                           text =
                             as.character(htmltools::tagList(
                               title,
                               htmltools::tags$sup(subtitle)
                             )),
                           y = 0.95,
                           yref = "container"),
                       autosize = FALSE,  #maybe can cause trouble
                       margin = list(
                         t = value_on_top_conditional,
                         r = 100,
                         l = 100,
                         b = value_bottom_conditional)
  )

  py <- plotly::config(py, displaylogo = FALSE)

  py <- htmltools::span(
    title = groups,
    `data-tippy-always-on` = "true",
    py
  )

  py <- htmltools::browsable(py)
  return(py)


#  py <- plotly::layout(py,
                       #title =
                       #  list(
                       #    text =
                       #      as.character(htmltools::tagList(
                       #        title,
                       #        htmltools::tags$sup(subtitle)
                        #     )),
                        #   y = 0.95,
                        #   yref = "container"),
#                       autosize = FALSE,  #maybe can cause trouble
#                       margin = list(
#                         t = value_on_top_conditional,
#                         r = 100,
#                         l = 100,
#                         b = value_bottom_conditional)
#  )

#  py <-
#    htmltools::div(htmltools::h4(title,
#                                 style = "font-family: sans-serif; margin-bottom: 0; margin-top: 2;"),
#                   htmltools::h5(subtitle,
#                                 style = "font-family: sans-serif; margin-top: 0; margin-bottom: 0; "),
#                   htmltools::div(py,
#                                  style = "align-self: center;margin-top: 0; "),
#                   style = "text-align: center; display: flex; flex-direction: column; align-item: center;",
#                   title = groups,
#                   `data-tippy-always-on` = "true")

#  py <- htmltools::browsable(py)

#  return(py)
}

#' Fix results from merge
#'
#' this function handles the result of [merge()]-calls, if `no.dups = TRUE` and
#' `suffixes = c("", "")`
#'
#' @param dfr data frame to fix
#' @param stop_if_incompatible [logical] stop if data frame can not be fixed
#'
#' @family data_management
#' @concept summary
#' @noRd
util_fix_merge_dups <- function(dfr,
                                stop_if_incompatible = TRUE) {
  if (nrow(dfr) == 0) {
    return(dfr)
  }
  dupcols <- unique(colnames(dfr)[duplicated(colnames(dfr))])
  for (dc in dupcols) {
    cur <- which(colnames(dfr) == dc)
    fix_nas <- dfr[, cur, drop = FALSE]
    xx <- apply(simplify = FALSE, fix_nas, 1, function(rw) {
      urw <- unique(rw)
      urw <- urw[!is.na(urw)]
      if (length(urw) > 1) {
        util_warning(c("Internal error: could not fix merge result, results are",
                       "incompatible. Sorry, please report"))
        return(rw)
      }
      if (length(urw) == 0) {
        # no action needed, all NA
        return(rw)
      }
      val <- urw[[1]]
      rw[] <- val
      return(rw)
    })
    fix_nas[] <- do.call(rbind, xx)
    # all(fix_nas[[1]] == fix_nas[[2]])
    fix_nas <- t(unique(t(fix_nas)))
    if(stop_if_incompatible) {
      util_stop_if_not(
        `Internal error, sorry, please report: fix_merge_dups failed` =
          ncol(fix_nas) == 1)
    } else {
      # if we have incompatible cases, try to fix them (!stop_if_incompatible)
      if (ncol(fix_nas) > 1) {
        colnames(dfr)[tail(cur, -1)] <-
          paste0(colnames(dfr)[tail(cur, -1)], ".", seq_len(length(cur)-1))  #TODO: check if there is not already a column with that name in all data, not just dfr
      }
    }

    # actually fix it!
    dfr[, head(cur, 1)] <- as.vector(fix_nas)
    dfr[, tail(cur, -1)] <- NULL
  }
  dfr
}

#' Guess a metadata data frame from study data.
#'
#' Guess a minimum metadata data frame from study data. Minimum required variable
#' attributes are:
#'
#' ```{r}
#' dataquieR:::util_get_var_att_names_of_level(VARATT_REQUIRE_LEVELS$REQUIRED)
#' ````
#'
#' The function also tries to detect missing codes.
#'
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param level [enum] levels to provide (see also [VARATT_REQUIRE_LEVELS])
#' @param cumulative [logical] include attributes of all levels up to level
#' @param convert_factors [logical] convert factor columns to coded integers.
#'                                  if selected, then also the study data will
#'                                  be updated and returned.
#' @param guess_missing_codes [logical] try to guess missing codes from the data
#' @param guess_character [logical] guess a data type for character columns
#'                                  based on the values
#' @return a meta_data data frame or a list with study data and metadata, if
#'         `convert_factors == TRUE`.
#' @export
#' @importFrom stats setNames
#' @examples
#' \dontrun{
#' dataquieR::prep_study2meta(Orange, convert_factors = FALSE)
#' }
prep_study2meta <- function(study_data, level = c(
                              VARATT_REQUIRE_LEVELS$REQUIRED,
                              VARATT_REQUIRE_LEVELS$RECOMMENDED
                            ),
                            cumulative = TRUE,
                            convert_factors = FALSE,
                            guess_missing_codes =
                              getOption("dataquieR.guess_missing_codes",
                                        dataquieR.guess_missing_codes_default),
                            guess_character =
                              getOption("dataquieR.guess_character",
                                        default =
                                          dataquieR.guess_character_default
                            )) {
  util_expect_scalar(guess_character, check_type = is.logical)
  withr::local_options(list(dataquieR.fix_column_type_on_read = TRUE))
  with_dataframe_environment(
    util_expect_data_frame(study_data, keep_types = TRUE) # prep_robust_guess_data_type
  )
  if (missing(study_data) || !is.data.frame(study_data)) {
    util_error("Need study data as a data frame")
  }

  if (length(convert_factors) != 1 || !is.logical(convert_factors)) {
    util_error("Argument %s must be logical(1)", dQuote("convert_factors"))
  }

  util_expect_scalar(guess_missing_codes, check_type = is.logical,
                     error_message = sprintf("%s needs to be one logical value",
                                             sQuote("guess_missing_codes")))

  study_data <- util_cast_off(study_data, "study_data")

  util_get_var_att_names_of_level(VARATT_REQUIRE_LEVELS$REQUIRED)

  var_names <- colnames(study_data)
  var_labels <- var_names # TODO: maybe do something nicer here, e.g., SNAKE_CASE -> Snake Case, or so.

  if (length(var_names) == 0) {
    util_error("No study variables found -- cannot proceed.")
  }

  datatypes <- prep_datatype_from_data(resp_vars = var_names, study_data =
                                         study_data, guess_character =
                                         guess_character)

  missing_list <-
    mapply(function(x, dt, cn) {
      if (dt %in% c(DATA_TYPES$INTEGER, DATA_TYPES$FLOAT)) {
        mcs <-
          unique(sort(suppressWarnings(as.numeric(x[
            util_looks_like_missing(as.numeric(x))]))))
        r <- paste(mcs, collapse = SPLIT_CHAR)
        if (length(mcs) > 0)
          util_message(
            "For %s, maybe, the following values are missing codes: %s",
            sQuote(cn),
            util_pretty_vector_string(mcs, n_max = 5)
          )
      } else {
        r <- ""
      }
      if (length(r) == 1 && util_empty(r))
        r <- SPLIT_CHAR # missing on purpose
      r
    }, study_data, datatypes, colnames(study_data))

  if (!guess_missing_codes) {
    missing_list <- SPLIT_CHAR
  }

  if (convert_factors) { # TODO: work with VALUE_LABEL_TABLE
    valuelabels <- prep_valuelabels_from_data(resp_vars = var_names,
                                              study_data = study_data)
  } else {
    valuelabels <- list()
    valuelabels[[VALUE_LABELS]] <- vapply(
      var_names,
      FUN.VALUE = character(1),
      function(v) {
        if (is.factor(study_data[[v]])) {
          lvs <- levels(study_data[[v]])
          if (is.ordered(study_data[[v]])) {
            split_char <- "<"
          } else {
            split_char <- SPLIT_CHAR
          }
          paste(lvs, collapse = sprintf(" %s ", split_char))
          # lvs[as.integer(study_data[[v]])]
        } else {
          NA_character_
        }
      }
    )
  }

  res <- prep_create_meta(
    VAR_NAMES = var_names,
    LABEL = var_labels,
    LONG_LABEL = var_labels,
    DATA_TYPE = datatypes,
    VALUE_LABELS = valuelabels[[VALUE_LABELS]],
    MISSING_LIST = missing_list,
    HARD_LIMITS = NA_character_,
    JUMP_LIST = NA_character_
  )

  generated_atts <- util_get_var_att_names_of_level(level,
                                                    cumulative = cumulative)

  if (SCALE_LEVEL %in% generated_atts) {
    .md <- res
    .md[[JUMP_LIST]] <- SPLIT_CHAR
    with_scale_level <-
      prep_scalelevel_from_data_and_metadata(#resp_vars = var_names,
                                             study_data = study_data,
                                             meta_data = .md,
                                             label_col = VAR_NAMES)

      res[[SCALE_LEVEL]] <- setNames(
        with_scale_level[[SCALE_LEVEL]], nm = with_scale_level[[VAR_NAMES]])[
          res[[VAR_NAMES]]]
  }

  missing_atts <- setdiff(generated_atts, colnames(res))
  if (length(missing_atts) > 0) {
    empty_cols <-
      do.call(data.frame, c(
        list(stringsAsFactors = FALSE),
        lapply(setNames(nm = missing_atts), function(x) {
          rep(NA_character_, nrow(res))
        })
      ))

    res <- cbind.data.frame(res, empty_cols)
  }

  if (MISSING_LIST_TABLE %in% generated_atts)
    generated_atts <- union(generated_atts, c(MISSING_LIST, JUMP_LIST))

  missing_atts <- setdiff(generated_atts, colnames(res))

  res <- res[, intersect(generated_atts, colnames(res))]

  if (length(missing_atts)) {
    util_error(
      c("Internal error. The function prep_study2meta should return a minimum",
      "metadata data frame, but the attributes %s are missing."),
      paste0(dQuote(
        missing_atts
      ), collapse = ", ")
    )
  }

  if (convert_factors) {
    # TODO: work with VALUE_LABEL_TABLE
    valuelabels <- prep_valuelabels_from_data(resp_vars = var_names,
                                              study_data = study_data)
    # data type should then be integer for codes
    datatypes[names(which(vapply(study_data, is.factor,
                                 FUN.VALUE = logical(1))))] <-
      DATA_TYPES$INTEGER
    res$DATA_TYPE <- unname(datatypes)
    res <- list(
      MetaData = res,
      ModifiedStudyData = valuelabels[["ModifiedStudyData"]]
    )
  }

  res
}

test_that("prep_apply_coding works", {
  skip_on_cran()
  study_data <- iris
  x <- prep_study2meta(study_data = study_data,
                       convert_factors = TRUE)
  x$MetaData[[STUDY_SEGMENT]] <- "SEGMENT"
  study_data$Species <- as.character(study_data$Species)
  expect_warning(
    expect_equal(
      prep_apply_coding(study_data = study_data,
                        item_level = x$MetaData)$Species,
      x$ModifiedStudyData$Species
    ),
    regexp = "Metadata does not provide a filled column called .+JUMP_LIST"
  )
})

test_that("prep_study2meta works", {
  skip_on_cran() # slow, rarely used, since w/o metdata, reports are not really informative
  
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  testthat::local_edition(3)
  expect_error(
    prep_study2meta(cars, convert_factors = 1:3),
    regexp = "Argument.+convert_factors.+must be logical.1."
  )

  expect_snapshot_value(style = "deparse",
                        prep_study2meta(iris, convert_factors = TRUE,
                                        guess_missing_codes = TRUE))

  meta_data <- prep_study2meta(cars, cumulative = FALSE,
                               guess_missing_codes = TRUE)
  reference <- data.frame(VAR_NAMES = colnames(cars),
                          LABEL = colnames(cars),
                          DATA_TYPE = "integer",
                          SCALE_LEVEL = c(SCALE_LEVELS$ORDINAL, # this is not correctly guessed
                                          SCALE_LEVELS$RATIO),
                          UNIT = NA_character_,
                          VALUE_LABELS = NA_character_,
                          MISSING_LIST_TABLE = NA_character_,
                          HARD_LIMITS = NA_character_,
                          GROUP_VAR_OBSERVER = NA_character_,
                          GROUP_VAR_DEVICE = NA_character_,
                          TIME_VAR = NA_character_,
                          PART_VAR = NA_character_,
                          STUDY_SEGMENT = NA_character_,
                          LOCATION_RANGE = NA_character_,
                          LOCATION_METRIC = NA_character_,
                          PROPORTION_RANGE = NA_character_,
                          UNIVARIATE_OUTLIER_CHECKTYPE = NA_character_,
                          N_RULES = NA_character_,
                          CO_VARS = NA_character_,
                          LONG_LABEL = colnames(cars),
                          MISSING_LIST = SPLIT_CHAR,
                          JUMP_LIST = NA_character_,
                          row.names = colnames(cars))
  reference <- reference[, intersect(colnames(meta_data), colnames(reference))]

  expect_equal(meta_data, reference)

  expect_error(prep_study2meta(42, guess_missing_codes = TRUE),
               regexp = ".+study_data.+ is not a data frame.")
  expect_error(prep_study2meta(NULL, guess_missing_codes = TRUE),
               regexp = ".+study_data.+ is not a data frame.")
  expect_error(prep_study2meta(),
               regexp = "Missing .+study_data.+")
  expect_error(prep_study2meta(cars[, c()], guess_missing_codes = TRUE),
               regexp = "No study variables found -- cannot proceed.")
})

test_that("prep_study2meta handles tibbles correctly", {
  skip_on_cran() # slow, rarely used, since w/o metdata, reports are not really informative
  skip_if_not_installed("tibble") # should never be skipped, since dplyr is a
                                  # dependency of dataquieR and dplr depends on
                                  # tibble
  meta_data <- prep_study2meta(tibble::as_tibble(cars), cumulative = FALSE,
                               guess_missing_codes = TRUE)
  expect_equal(meta_data,
               data.frame(VAR_NAMES = colnames(cars),
                          LABEL = colnames(cars),
                          DATA_TYPE = "integer",
                          SCALE_LEVEL = c(SCALE_LEVELS$ORDINAL, # this is not correctly guessed
                                          SCALE_LEVELS$RATIO),
                          UNIT = NA_character_,
                          VALUE_LABELS = NA_character_,
                          MISSING_LIST_TABLE = NA_character_,
                          HARD_LIMITS = NA_character_,
                          CO_VARS = NA_character_,
                          GROUP_VAR_OBSERVER = NA_character_,
                          GROUP_VAR_DEVICE = NA_character_,
                          TIME_VAR = NA_character_,
                          PART_VAR = NA_character_,
                          STUDY_SEGMENT = NA_character_,
                          LONG_LABEL = colnames(cars),
                          LOCATION_RANGE = NA_character_,
                          LOCATION_METRIC = NA_character_,
                          PROPORTION_RANGE = NA_character_,
                          UNIVARIATE_OUTLIER_CHECKTYPE = NA_character_,
                          N_RULES = NA_character_,
                          MISSING_LIST = SPLIT_CHAR,
                          JUMP_LIST = NA_character_,
                          stringsAsFactors = FALSE,
                          row.names = colnames(cars))
  )
    suppressWarnings(expect_warning(
      meta_data <-
        with_mocked_bindings(.package = "base",
                           requireNamespace = function(package, ...,
                                                       quietly = FALSE) {
        return(FALSE)
      },
        prep_study2meta(tibble::as_tibble(cars),
                        guess_missing_codes = TRUE)
      ),
    regexp =
      sprintf("%s",
              paste(".+study_data.+ looks like a tibble. However, the package",
                    ".+tibble.+ seems not to be available,")),
    perl = TRUE
  ))
  expect_equal(meta_data,
               data.frame(VAR_NAMES = colnames(cars),
                          LABEL = colnames(cars),
                          DATA_TYPE = "integer",
                          SCALE_LEVEL = c(SCALE_LEVELS$ORDINAL, # this is not correctly guessed
                                          SCALE_LEVELS$RATIO),
                          UNIT = NA_character_,
                          VALUE_LABELS = NA_character_,
                          MISSING_LIST_TABLE = NA_character_,
                          HARD_LIMITS = NA_character_,
                          CO_VARS = NA_character_,
                          GROUP_VAR_OBSERVER = NA_character_,
                          GROUP_VAR_DEVICE = NA_character_,
                          TIME_VAR = NA_character_,
                          PART_VAR = NA_character_,
                          STUDY_SEGMENT = NA_character_,
                          LONG_LABEL = colnames(cars),
                          LOCATION_RANGE = NA_character_,
                          LOCATION_METRIC = NA_character_,
                          PROPORTION_RANGE = NA_character_,
                          UNIVARIATE_OUTLIER_CHECKTYPE = NA_character_,
                          N_RULES = NA_character_,
                          MISSING_LIST = SPLIT_CHAR,
                          JUMP_LIST = NA_character_,
                          stringsAsFactors = FALSE,
                          row.names = colnames(cars))
  )
})

test_that("prep_study2meta works on study_data", {
  skip_on_cran() # slow, rarely used, since w/o metdata, reports are not really informative
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  study_data <- study_data[, order(colnames(study_data)), FALSE]
  meta_data <- meta_data[order(meta_data$VAR_NAMES), , FALSE]
  meta_data[meta_data$LABEL %in% c("SBP_0", "DBP_0", "ARM_CIRC_0", "BSG_0"),
            DATA_TYPE] <- # values are not really float
    DATA_TYPES$INTEGER
  guessed_meta_data <- prep_study2meta(study_data, guess_missing_codes = TRUE)
  expect_equal(
    guessed_meta_data,
    data.frame(
      VAR_NAMES = meta_data$VAR_NAMES,
      LABEL = meta_data$VAR_NAMES,
      DATA_TYPE = meta_data$DATA_TYPE,
      SCALE_LEVEL = c(
        "nominal",
        "na",
        "nominal",
        "ratio",
        "ratio",
        "ratio",
        "ratio",
        "nominal",
        "nominal",
        "ratio",
        "nominal",
        "nominal",
        "nominal",
        "interval",
        "ratio",
        "ratio",
        "nominal",
        "interval",
        "nominal",
        "nominal",
        "nominal",
        "ordinal",
        "nominal",
        "nominal",
        "nominal",
        "nominal",
        "ordinal",
        "ordinal",
        "nominal",
        "nominal",
        "nominal",
        "ordinal",
        "nominal",
        "interval",
        "ordinal",
        "ordinal",
        "ordinal",
        "ordinal",
        "ordinal",
        "ordinal",
        "ordinal",
        "ordinal",
        "interval",
        "nominal",
        "nominal",
        "nominal",
        "ratio",
        "nominal",
        "nominal",
        "nominal",
        "nominal",
        "nominal",
        "nominal"
      ),
      UNIT = NA_character_,
      VALUE_LABELS = NA_character_,
      MISSING_LIST_TABLE = NA_character_,
      HARD_LIMITS = NA_character_,
      CO_VARS = NA_character_,
      GROUP_VAR_OBSERVER = NA_character_,
      GROUP_VAR_DEVICE = NA_character_,
      TIME_VAR = NA_character_,
      PART_VAR = NA_character_,
      STUDY_SEGMENT = NA_character_,
      LONG_LABEL = meta_data$VAR_NAMES,
      LOCATION_RANGE = NA_character_,
      LOCATION_METRIC = NA_character_,
      PROPORTION_RANGE = NA_character_,
      UNIVARIATE_OUTLIER_CHECKTYPE = NA_character_,
      N_RULES = NA_character_,
      MISSING_LIST = c(
        SPLIT_CHAR,
        SPLIT_CHAR,
        SPLIT_CHAR,
        SPLIT_CHAR,
        "99980|99981|99982|99983|99984|99985|99986|99987|99988|99989|99990|99991|99992|99993|99994|99995",
        "99980|99981|99982|99983|99984|99985|99986|99987|99988|99989|99990|99991|99992|99993|99994|99995",
        "99980|99983|99987|99988|99989|99990|99991|99992|99993|99994|99995",
        "99980|99988|99989|99991|99993|99994|99995",
        SPLIT_CHAR,
        "99980|99981|99982|99983|99984|99985|99986|99988|99989|99990|99991|99992|99993|99994|99995",
        "99980|99987",
        SPLIT_CHAR,
        SPLIT_CHAR,
        SPLIT_CHAR,
        "99980|99981|99982|99983|99984|99985|99986|99988|99989|99990|99991|99992|99994|99995",
        "99980|99981|99982|99983|99984|99985|99986|99988|99989|99990|99991|99992|99994|99995",
        SPLIT_CHAR,
        SPLIT_CHAR,
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "88880|99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "88880|99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        SPLIT_CHAR,
        SPLIT_CHAR,
        "99980|99983|99988|99989|99991|99993|99994|99995",
        "99980|99983|99988|99989|99991|99993|99994|99995",
        "99980|99983|99988|99989|99991|99993|99994|99995",
        "99980|99983|99988|99989|99991|99993|99994|99995",
        "99980|99983|99988|99989|99991|99993|99994|99995",
        "99980|99983|99988|99989|99991|99993|99994|99995",
        "99980|99983|99988|99989|99991|99993|99994|99995",
        "99980|99983|99988|99989|99991|99993|99994|99995",
        SPLIT_CHAR,
        SPLIT_CHAR,
        "99980|99981|99982|99983|99984|99985|99986|99987|99988|99989|99990|99991|99992|99993|99994|99995",
        SPLIT_CHAR,
        SPLIT_CHAR,
        "99980|99983|99988|99989|99990|99991|99993|99994|99995",
        SPLIT_CHAR,
        SPLIT_CHAR,
        SPLIT_CHAR,
        SPLIT_CHAR,
        SPLIT_CHAR
      ),
      JUMP_LIST = NA_character_,
      stringsAsFactors = FALSE,
      row.names = colnames(study_data)
    )
  )
})

#' Prepare a label as part of a link for `RMD` files
#'
#' @param s the label
#' @param html prepare the label for direct `HTML` output instead of `RMD`
#'
#' @return the escaped label
#' @export
#'
prep_link_escape <- function(s, html = FALSE) { # TODO: Maybe use URLencode(), but take care: also file names are entered, here
  r <- gsub("*", "X", s, fixed = TRUE)
  r <- gsub("\u00b2", "", r, fixed = TRUE)
  r <- gsub("\u00b3", "", r, fixed = TRUE)
  r <- gsub("\u00B0", "", r, fixed = TRUE)
  r <- gsub("%", "", r, fixed = TRUE)
  r <- gsub("/", "", r, fixed = TRUE)
  r <- gsub("\\", "", r, fixed = TRUE)
  r <- gsub("'", "", r, fixed = TRUE)
  r <- gsub(":", "", r, fixed = TRUE)
  r <- gsub("\u20ac", "", r, fixed = TRUE)
  r <- gsub("`+", "", r, perl = TRUE)
  r <- gsub("\\$+", "", r, perl = TRUE)
  r <- gsub("^#+", "", r, perl = TRUE)
  r <- gsub("@", "", r, fixed = TRUE)
  r <- gsub("<", "", r, fixed = TRUE)
  r <- gsub(">", "", r, fixed = TRUE)
  r <- gsub("&", "", r, fixed = TRUE)
  r <- gsub("\"", "", r, fixed = TRUE)
  r <- gsub("_", "", r, fixed = TRUE)
  r <- gsub("~", "", r, fixed = TRUE)
  r <- gsub("\u00b5", "", r, fixed = TRUE)
  r <- gsub(";", "", r, fixed = TRUE)
  r <- gsub("\u00b0", "", r, fixed = TRUE)
  r <- gsub("?", "Q", r, fixed = TRUE)
  r <- gsub("#", "", r, fixed = TRUE)
  r <- gsub(")", "", r, fixed = TRUE)
  r <- gsub("(", "", r, fixed = TRUE)
  r <- gsub("|", "\uFF5C", r, fixed = TRUE)
  # r <- gsub("[\"/ <>&;°@µ~]", "", r, perl = TRUE)
  gsub("\\s+", "", r, perl = TRUE)
}

#' @title Check disjunctness of *_vars argument sets
#' @description
#' Checks that vectors supplied via arguments ending in `_vars` are pairwise
#' disjoint. If called with `...`, only the supplied arguments are checked
#' (legacy behavior). If called without `...`, the caller function is
#' introspected: all its `*_vars` arguments are resolved to their effective
#' values (passed, modified in the caller, or defaults) and checked.
#'
#' @param ... Optional named vectors to check. If provided, only these are
#'   considered.
#'
#' @return `invisible(NULL)`. Raises an error via `util_error()` on overlap.
#'
#' @noRd
#' @family utilities
util_disjunct_var_sets <- function(...) {
  dots <- rlang::dots_list(...)
  if (length(dots) > 0L) {
    arg_list <- dots
    arg_names <- names(arg_list)
    if (is.null(arg_names) || any(arg_names == "")) {
      arg_names <- vapply(
        substitute(list(...))[-1],
        rlang::as_label,
        character(1)
      )
    }
    names(arg_list) <- arg_names
  } else {
    caller_fn   <- rlang::caller_fn()
    caller_env  <- rlang::caller_env()
    caller_call <- rlang::caller_call()
    if (is.null(caller_fn) || is.null(caller_call)) return(invisible(NULL))

    formals_list <- formals(caller_fn)
    var_args <- grep("_vars$", names(formals_list), value = TRUE)
    if (length(var_args) == 0L) return(invisible(NULL))

    matched_call <- rlang::call_match(caller_call, caller_fn)
    matched_args <- rlang::call_args(matched_call)

    resolve_one <- function(nm) {
      if (rlang::env_has(caller_env, nm, inherit = TRUE)) {
        return(get(nm, envir = caller_env, inherits = TRUE))
      }
      if (nm %in% names(matched_args)) {
        return(rlang::eval_bare(matched_args[[nm]], env = caller_env))
      }
      def <- formals_list[[nm]]
      if (!rlang::is_missing(def)) {
        return(rlang::eval_bare(def, env = caller_env))
      }
      NULL
    }

    vals <- lapply(var_args, resolve_one)
    names(vals) <- var_args
    keep <- !vapply(vals, function(x) is.null(x) || length(x) == 0L,
                    logical(1))
    if (!any(keep)) return(invisible(NULL))
    arg_list <- vals[keep]
  }

  arg_names <- names(arg_list)
  pairs <- utils::combn(seq_along(arg_list), 2L, simplify = FALSE)

  overlaps <- lapply(pairs, function(ix) {
    ov <- intersect(arg_list[[ix[1]]], arg_list[[ix[2]]])
    if (length(ov) > 0L) {
      list(args = c(arg_names[[ix[1]]], arg_names[[ix[2]]]),
           values = ov)
    } else {
      NULL
    }
  })
  overlaps <- overlaps[!vapply(overlaps, is.null, logical(1))]

  if (length(overlaps) > 0L) {
    problematic_args <- unique(unlist(lapply(overlaps, `[[`, "args")))
    problematic_vals <- unique(unlist(lapply(overlaps, `[[`, "values")))
    util_error(
      c("Overlap in the arguments %s. %s cannot be parts of more than one ",
        "argument."),
      util_pretty_vector_string(problematic_args),
      util_pretty_vector_string(problematic_vals, n_max = 5),
      applicability_problem = TRUE
    )
  }

  invisible(NULL)
}

test_that("util_disjunct_var_sets works for disjoint sets", {
  expect_silent(util_disjunct_var_sets(
    resp_vars = c("a", "b"),
    co_vars = c("c", "d"),
    group_vars = c("e")
  ))
})

test_that("util_disjunct_var_sets throws error on overlap", {
  expect_error(
    util_disjunct_var_sets(
      resp_vars = c("a", "b"),
      co_vars = c("b", "c"),
      group_vars = c("e")
    ),
    regexp = "Overlap in the arguments .*b.*"
  )
})

test_that("util_disjunct_var_sets catches multiple overlaps", {
  expect_error(
    util_disjunct_var_sets(
      resp_vars = c("a", "b", "x"),
      co_vars = c("x", "y", "z"),
      group_vars = c("b", "w")
    ),
    regexp = "Overlap in the arguments .*resp_vars.*co_vars.*group_vars.*"
  )
})

test_that("no-dots: caller *_vars disjoint", {
  f <- function(resp_vars, co_vars) util_disjunct_var_sets()
  expect_silent(f(c("a", "b"), c("c", "d")))
})

test_that("no-dots: caller *_vars overlap", {
  f <- function(resp_vars, co_vars) util_disjunct_var_sets()
  expect_error(f(c("a", "b"), c("b", "c")),
               regexp = "Overlap in the arguments")
})

test_that("defaults are used when not passed", {
  f_ok <- function(resp_vars = c("a"), co_vars = c("b"))
    util_disjunct_var_sets()
  f_bad <- function(resp_vars = c("a"), co_vars = c("a"))
    util_disjunct_var_sets()
  expect_silent(f_ok())
  expect_error(f_bad(), regexp = "Overlap in the arguments .*a.*")
})

test_that("NULL/empty defaults => nothing to check", {
  f1 <- function(resp_vars = NULL, co_vars = NULL) util_disjunct_var_sets()
  f2 <- function(resp_vars = character(), co_vars = character())
    util_disjunct_var_sets()
  expect_silent(f1())
  expect_silent(f2())
})

test_that("caller has no *_vars formals", {
  f <- function(x, y) util_disjunct_var_sets()
  expect_silent(f(1, 2))
})

test_that("caller body overrides defaults (env wins)", {
  f_ok <- function(resp_vars = c("a"), co_vars = c("a")) {
    resp_vars <- "a"; co_vars <- "b"
    util_disjunct_var_sets()
  }
  f_bad <- function(resp_vars = c("a"), co_vars = c("b")) {
    co_vars <- "a"
    util_disjunct_var_sets()
  }
  expect_silent(f_ok())
  expect_error(f_bad(), regexp = "Overlap in the arguments .*a.*")
})

test_that("direct call without args is silent", {
  expect_silent(util_disjunct_var_sets())
})

#' Verify, that argument is a data frame
#'
#' stops with an error, if not. will add the columns, and return the resulting
#' extended data frame, and also updating the original data frame in the
#' calling environment, if #' `x` is empty (data frames easily break to
#' 0-columns in R, if they have not rows, e.g. using some `split`/`rbind`
#' pattern)
#'
#' @param x an object that is verified to be a `data.frame`.
#' @param col_names column names x must contain or named list of predicates to
#'                  check the columns (e.g.,
#'                  list(AGE=is.numeric, SEX=is.character))
#' @param convert_if_possible if given, for each column, a lambda can be given
#'                            similar to `col_names` check functions. This
#'                            lambda would be used to try a conversion. If
#'                            a conversion fails (returns `NA`, where the
#'                            input was not `util_empty'), an error
#'                            is still thrown, the data is converted, otherwise
#' @param dont_assign set `TRUE` to keep `x` in the caller environment untouched
#' @param keep_types [logical] keep types as possibly defined in a file, if the
#'                             data frame is loaded from one. set `TRUE` for
#'                             study data.
#' @param custom_errors [list] with error messages, specifically per column.
#'                      names of the list are column names, values are messages
#'                      ([character]).
#'
#' @return `invisible` data frame
#'
#' @family io
#' @concept data_frame
#' @noRd

util_expect_data_frame <- function(x, col_names, convert_if_possible,
                                   custom_errors,
                                   dont_assign, keep_types = FALSE,
                                   keep_encoding_errors = FALSE) {
  # TODO: Custom error message
  if (missing(dont_assign)) dont_assign <- FALSE
  if (!missing(keep_encoding_errors)) {
    util_expect_scalar(keep_encoding_errors,
                       check_type = is.logical,
                       error_message =
                         sprintf("%s needs to be to be a logical value.",
                                 sQuote("keep_encoding_errors")))
  }
  if (!missing(keep_types)) {
    util_expect_scalar(keep_types,
                       check_type = is.logical,
                       error_message =
                         sprintf("%s needs to be to be a logical value.",
                                 sQuote("keep_types")))
  }
  util_expect_scalar(dont_assign, check_type = is.logical, error_message =
                       sprintf("%s needs to be to be a logical value.",
                               sQuote("dont_assign")))
  symb <- substitute(x)
  mis_it <- try(eval.parent(call("missing", symb)), silent = TRUE)
  default <- try(get(symb, parent.frame()), silent = TRUE)
  if (inherits(default, "try-error") && identical(mis_it, TRUE)) {
    util_error("Missing %s",
               sQuote(util_deparse1(symb)))
  }
  if (missing(col_names)) {
    col_names <- character(0)
  }
  if (is.list(col_names)) {
    lambdas <- col_names
    if (!length(names(lambdas))) {
      util_error("%s needs to be a named list, if not a character vector",
                 sQuote("col_names"))
    }
    col_names <- names(lambdas)
  } else {
    util_expect_scalar(col_names,
                       allow_null = TRUE,
                       allow_more_than_one = TRUE)
    lambdas <- lapply(setNames(nm = col_names),
                      function(x) { function(y) TRUE })
  }
  if (missing(convert_if_possible)) {
    convert_if_possible <- NULL
  } else if (is.list(convert_if_possible)) {
    if (!length(names(convert_if_possible))) {
      util_error("%s needs to be a named list, if a list",
                 sQuote("convert_if_possible"))
    } else {
      util_ensure_in(names(convert_if_possible),
                     col_names,
                     err_msg = "",
                     error = FALSE)
    }
  } else if (length(convert_if_possible)) {
    util_error("%s needs to be a named list or NULL",
               sQuote("convert_if_possible"))
  }
  if (all(is.character(x)) && length(x) == 1) {
    # Handle data frame as char
    x <- prep_get_data_frame(x, keep_types = keep_types)
  }
  x <- util_cast_off(x, as.character(symb), .dont_cast_off_cols = TRUE)

  if (!is.data.frame(x)) {
    util_error("%s is not a data frame.", dQuote(symb))
  }

  if (!keep_encoding_errors) {
    x[] <- util_fix_encoding_cols(x)
  }


  if (any(!col_names %in% colnames(x))) {
    if (nrow(x) == 0) {
      cols <- union(
        colnames(x),
        col_names)
      x <- as.data.frame(
        matrix(nrow = 0, ncol = length(cols),
               dimnames = list(character(0), cols)),
        stringsAsFactors = FALSE, make.names = FALSE)
      if (!dont_assign)
        assign(as.character(symb), x, envir = parent.frame())
    } else {
      util_error("Missing columns %s from %s.",
                 paste(dQuote(col_names[!col_names %in% colnames(x)]),
                       collapse = ", "),
                 dQuote(symb))
    }
  }
  detect_mismatch <-function() {
    !vapply(names(lambdas[
      vapply(lambdas, is.function, FUN.VALUE = logical(1))]), function(cl) {
        r <- lambdas[[cl]](x[[cl]])
        if (length(r) == nrow(x) && is.logical(r)) {
          r <- all(r, na.rm = TRUE)
        }
        if (length(r) != 1 || is.na(r) || !is.logical(r)) {
          util_error(c("Internal error in util_expect_data_frame:",
                       "lambda %s did not return %s or %s"),
                     sQuote(cl), dQuote(TRUE), dQuote(FALSE),
                     applicability_problem = TRUE)
        }
        r
      }, FUN.VALUE = logical(1))
  }
  mismatch <- detect_mismatch()
  if (any(mismatch)) {
    if (is.list(convert_if_possible)) {
      for (cn in names(mismatch)) {
        convert <- convert_if_possible[[cn]]
        if (is.function(convert)) {
          x_cn <- convert(x[[cn]])
          if (!all(is.na(x_cn) == is.na(x[[cn]]))) {
            util_warning(
              "In %s, could not convert the whole column %s to match data type",
              dQuote(symb), dQuote(cn))
          } else {
            x[[cn]] <- x_cn
          }
        }
      }
    }
    mismatch <- detect_mismatch()
  }
  if (any(mismatch)) {
    if (!missing(custom_errors)) {
      util_stop_if_not("argument custom_errors needs to be a named list" =
                         is.list(custom_errors))
      cols_with_trouble <- names(which(mismatch))
      if (all(cols_with_trouble %in% names(custom_errors))) { # TODO: default should fill gaps in custom_errors
        util_error(
          applicability_problem = TRUE,
          paste0(custom_errors[cols_with_trouble], collapse = "\n"))
        }
    }
    util_error(
      applicability_problem = TRUE,
      "Data in columns in %s do not match: %s.\n%s",
      sQuote(symb),
      paste(dQuote(names(mismatch[mismatch])), collapse = ", "),
      do.call(paste, append(dQuote(vapply(lambdas[mismatch], util_deparse1,
                            FUN.VALUE = character(1))), list(collapse = ", ")))
    )
  }
  if (!dont_assign) {
    assign(as.character(symb), x, envir = parent.frame())
  }
  invisible(x)
}

#' Prepare and verify study data with metadata
#'
#' @description
#' This function ensures, that a data frame `ds1` with suitable variable
#' names study_data and meta_data exist as base [data.frame]s.
#'
#' @details
#'
#' This function defines `ds1` and modifies `study_data` and `meta_data` in the
#' environment of its caller (see [eval.parent]). It also defines or modifies
#' the object `label_col` in the calling environment. Almost all functions
#' exported by `dataquieR` call this function initially, so that aspects common
#' to all functions live here, e.g. testing, if an argument `meta_data` has been
#' given and features really a [data.frame]. It verifies the existence of
#' required metadata attributes ([VARATT_REQUIRE_LEVELS]). It can also replace
#' missing codes by `NA`s, and calls [prep_study2meta] to generate a minimum
#' set of metadata from the study data on the fly (should be amended, so
#' on-the-fly-calling is not recommended for an instructive use of `dataquieR`).
#'
#' The function also detects `tibbles`, which are then converted to base-R
#' [data.frame]s, which are expected by `dataquieR`.
#'
#' If `.internal` is `TRUE`, differently from the other utility function that
#' work in their caller's environment, this function modifies objects in the
#' calling function's environment. It defines a new object `ds1`,
#' it modifies `study_data` and/or `meta_data`
#' and `label_col`.
#'
#' @param .study_data if provided, use this data set as study_data
#' @param .meta_data if provided, use this data set as meta_data
#' @param .label_col if provided, use this as label_col
#' @param .replace_hard_limits replace `HARD_LIMIT` violations by `NA`,
#'                             defaults to `FALSE`.
#' @param .replace_missings replace missing codes, defaults to `TRUE`
#' @param .sm_code missing code for `NAs`, if they have been
#'                 re-coded by `util_combine_missing_lists`
#' @param .allow_empty allow `ds1` to be empty, i.e., 0 rows and/or 0 columns
#' @param .adjust_data_type ensure that the data type of variables in the study
#'                data corresponds to their data type specified in the metadata
#' @param .amend_scale_level ensure that `SCALE_LEVEL` is available in the
#'                           item-level `meta_data`. internally used to prevent
#'                           recursion, if called from
#'                           [prep_scalelevel_from_data_and_metadata()].
#' @param .internal [logical] internally called, modify caller's environment.
#' @param .apply_factor_metadata  [logical] convert categorical variables to
#'                                          labeled factors.
#' @param .apply_factor_metadata_inadm  [logical] convert categorical variables
#'                                          to labeled factors keeping
#'                                          inadmissible values. Implies, that
#'                                          .apply_factor_metadata will be set
#'                                          to `TRUE`, too.
#'
#' @seealso acc_margins
#'
#' @return `ds1` the study data with mapped column names, `invisible()`, if
#'         not `.internal`
#'
#' @examples
#' \dontrun{
#' acc_test1 <- function(resp_variable, aux_variable,
#'                       time_variable, co_variables,
#'                       group_vars, study_data, meta_data) {
#'   prep_prepare_dataframes()
#'   invisible(ds1)
#' }
#' acc_test2 <- function(resp_variable, aux_variable,
#'                       time_variable, co_variables,
#'                       group_vars, study_data, meta_data, label_col) {
#'   ds1 <- prep_prepare_dataframes(study_data, meta_data)
#'   invisible(ds1)
#' }
#' environment(acc_test1) <- asNamespace("dataquieR")
#' # perform this inside the package (not needed for functions that have been
#' # integrated with the package already)
#'
#' environment(acc_test2) <- asNamespace("dataquieR")
#' # perform this inside the package (not needed for functions that have been
#' # integrated with the package already)
#' acc_test3 <- function(resp_variable, aux_variable, time_variable,
#'                       co_variables, group_vars, study_data, meta_data,
#'                       label_col) {
#'   prep_prepare_dataframes()
#'   invisible(ds1)
#' }
#' acc_test4 <- function(resp_variable, aux_variable, time_variable,
#'                       co_variables, group_vars, study_data, meta_data,
#'                       label_col) {
#'   ds1 <- prep_prepare_dataframes(study_data, meta_data)
#'   invisible(ds1)
#' }
#' environment(acc_test3) <- asNamespace("dataquieR")
#' # perform this inside the package (not needed for functions that have been
#' # integrated with the package already)
#'
#' environment(acc_test4) <- asNamespace("dataquieR")
#' # perform this inside the package (not needed for functions that have been
#' # integrated with the package already)
#' meta_data <- prep_get_data_frame("meta_data")
#' study_data <- prep_get_data_frame("study_data")
#' try(acc_test1())
#' try(acc_test2())
#' acc_test1(study_data = study_data)
#' try(acc_test1(meta_data = meta_data))
#' try(acc_test2(study_data = 12, meta_data = meta_data))
#' print(head(acc_test1(study_data = study_data, meta_data = meta_data)))
#' print(head(acc_test2(study_data = study_data, meta_data = meta_data)))
#' print(head(acc_test3(study_data = study_data, meta_data = meta_data)))
#' print(head(acc_test3(study_data = study_data, meta_data = meta_data,
#'   label_col = LABEL)))
#' print(head(acc_test4(study_data = study_data, meta_data = meta_data)))
#' print(head(acc_test4(study_data = study_data, meta_data = meta_data,
#'   label_col = LABEL)))
#' try(acc_test2(study_data = NULL, meta_data = meta_data))
#' }
#'
#' @export
#'
#' @importFrom rlang caller_fn
#' @importFrom utils object.size
prep_prepare_dataframes <- function(.study_data, .meta_data, .label_col,
                                    .replace_hard_limits,
                                    .replace_missings, .sm_code = NULL,
                                    .allow_empty = FALSE,
                                    .adjust_data_type = TRUE,
                                    .amend_scale_level = TRUE,
                                    .apply_factor_metadata = FALSE,
                                    .apply_factor_metadata_inadm = FALSE,
                                    .internal =
                                      rlang::env_inherits(
                                        rlang::caller_env(),
                                        parent.env(environment()))) {

  case_insens <- util_is_na_0_empty_or_false(
    getOption("dataquieR.study_data_colnames_case_sensitive",
              dataquieR.study_data_colnames_case_sensitive_default))

#  dimension <- substr(rlang::call_name(rlang::caller_call()), 1, 3)
  util_expect_scalar(.sm_code,
                     check_type = util_all_is_integer,
                     allow_null = TRUE)

  util_expect_scalar(.apply_factor_metadata, check_type = is.logical)
  util_expect_scalar(.apply_factor_metadata_inadm, check_type = is.logical)
  if (.apply_factor_metadata_inadm)
    .apply_factor_metadata <- TRUE

  if (missing(.replace_hard_limits)) .replace_hard_limits <- FALSE
  util_expect_scalar(.replace_hard_limits, check_type = is.logical)

  util_expect_scalar(.allow_empty, check_type = is.logical)
  if (!missing(.replace_missings) && (length(.replace_missings) != 1 ||
                                      !is.logical(.replace_missings) ||
                                      is.na(.replace_missings))) {
    util_error(
      c("Internal error, sorry, please report: .replace_missings needs to",
        "be 1 logical value."))
  }
  if (missing(.replace_missings)) .replace_missings <- TRUE


  callfn <- caller_fn(1)

  caller_defaults <- suppressWarnings(formals(callfn))
  caller_formals <- names(caller_defaults)
  caller_has_default <- suppressWarnings({
    !vapply(formals(callfn), identical, rlang::missing_arg(),
            FUN.VALUE = logical(1))
  })

  missing_in_parent <- suppressWarnings({
    vapply(names(formals(callfn)), function(x) eval(call("missing",
                                                         as.symbol(x)),
                                                    envir = parent.frame(3)),
           FUN.VALUE = logical(1))
  })

  if (missing(.label_col)) {
    if ("label_col" %in% caller_formals) {
      if (!eval.parent(substitute({
        missing(label_col)
      }))) {
        quoted_label_col <- eval.parent(substitute(substitute(label_col)))
        .label_col <- try(eval(quoted_label_col, envir = parent.frame()),
                          silent = TRUE)
        if (inherits(.label_col, "try-error")) {
          .label_col <- try(eval(quoted_label_col, envir =
                                   dataquieR::WELL_KNOWN_META_VARIABLE_NAMES),
                            silent = TRUE)
        }
        if (inherits(.label_col, "try-error")) {
          util_error("Cannot resolve %s", dQuote(paste0("label_col", " = ",
                                                        quoted_label_col)),
                     applicability_problem = TRUE)
        }
      } else if (caller_has_default[["label_col"]]) {
        .label_col <- eval(caller_defaults[["label_col"]],
                           envir = asNamespace("dataquieR"), enclos =
                             parent.frame(2))
      } else {
        .label_col <- VAR_NAMES
      }
    } else if (exists("label_col", parent.frame())) {
      .label_col <- get("label_col", parent.frame())
    } else {
      .label_col <- VAR_NAMES
    }
  }

  # if no study_data have been provided -> error
  if (missing(.study_data)) {
    if ("study_data" %in% caller_formals) {
      if (!eval.parent(substitute({
        missing(study_data)
      }))) {
        if (exists("study_data", parent.frame())) { # TODO: Test this and the same for metadata
          .study_data <- try(eval(quote(study_data), parent.frame()),
                             silent = TRUE)
          if (inherits(.study_data, "try-error")) {
            cnd <- attr(.study_data, "condition")
            cnd$call <- sys.call(1)
            util_error(cnd)
          }
        } else {
          util_error("object %s not found", dQuote("study_data"))
        }
      } else if (caller_has_default[["study_data"]]) {
        .study_data <- try(eval.parent(caller_defaults[["study_data"]], n = 2),
                           silent = TRUE)
        if (inherits(.study_data, "try-error")) {
          cnd <- attr(.study_data, "condition")
          cnd$call <- sys.call(1)
          util_error(cnd)
        }
      } else {
        .study_data <- NULL
      }
    } else if (exists("study_data", parent.frame())) {
      .study_data <- get("study_data", parent.frame())
    } else {
      .study_data <- NULL
    }
  }

  e <- new.env(parent = environment())
  e$study_data <- .study_data
  .study_data <-
    eval(
      quote(try(util_expect_data_frame(study_data, keep_types = TRUE),
                silent = TRUE)),
      e
    )
  if (inherits(.study_data, "try-error")) {
    util_error("Need study data as a data frame: %s",
               conditionMessage(attr(.study_data, "condition")))
  }
  .study_data <- util_normalize_time_only_columns(.study_data)

  if (.internal &&
      !exists("already_ppdf", parent.frame()) &&
      all(c("item_level", "meta_data") %in% caller_formals) &&
      !isTRUE(missing_in_parent["item_level"]) &&
      !isTRUE(missing_in_parent["meta_data"]) &&
      !identical(dynGet("item_level"), dynGet("meta_data"))
      ) {
    # IDEA: if (rlang::hash())
    util_error(c("You cannot provide both, %s as well as %s",
                 "these arguments are synonyms and must be",
                 "used mutually exclusively"),
               sQuote("item_level"),
               sQuote("meta_data"))
    # see prep_get_labels
  }

  if (missing(.meta_data)) {
    if ("meta_data" %in% caller_formals &&
       (!isTRUE(missing_in_parent["meta_data"]) ||
                  caller_has_default[["meta_data"]])) {
      if (isTRUE(missing_in_parent["meta_data"])) {
        .meta_data <- try(eval(caller_defaults[["meta_data"]], parent.frame(2)),
                          silent = TRUE)
        if (util_is_try_error(.meta_data)) {
          .meta_data <- try(eval(caller_defaults[["meta_data"]], parent.frame(1)),
                            silent = TRUE)
        }
      } else {
        .meta_data <- try(eval(quote(meta_data), parent.frame(1)),
                          silent = TRUE)
      }
      if (inherits(.meta_data, "try-error")) {
        cnd <- attr(.meta_data, "condition")
        cnd$call <- sys.call(1)
        util_error(cnd)
      }
    }
  }


  if (missing(.meta_data)) {
    if ("item_level" %in% caller_formals &&
        (!isTRUE(missing_in_parent["item_level"]) ||
         caller_has_default[["item_level"]])) {
      if (isTRUE(missing_in_parent["item_level"])) {
        .meta_data <- try(eval(caller_defaults[["item_level"]], parent.frame(2)),
                          silent = TRUE)
        if (util_is_try_error(.meta_data)) {
          .meta_data <- try(eval(caller_defaults[["item_level"]], parent.frame(1)),
                            silent = TRUE)
        }
      } else {
        .meta_data <- try(eval(quote(item_level), parent.frame(1)),
                          silent = TRUE)
      }
      if (inherits(.meta_data, "try-error")) {
        cnd <- attr(.meta_data, "condition")
        cnd$call <- sys.call(1)
        util_error(cnd)
      }
    }
  }
  if (missing(.meta_data)) {
    if (exists("item_level", envir = parent.frame()) &&
        !("item_level" %in% caller_formals)) {
      .meta_data <- get("item_level", envir = parent.frame())
    } else if (exists("meta_data", envir = parent.frame()) &&
               !("meta_data" %in% caller_formals)) {
      .meta_data <- get("meta_data", envir = parent.frame())
    }
  }
  if (missing(.meta_data)) {
    if ("item_level" %in% prep_list_dataframes()) {
      .meta_data <- util_expect_data_frame("item_level")
    }
  }
  if (!missing(.meta_data) && !is.data.frame(.meta_data)) {
    if (util_is_try_error(try(util_expect_data_frame(.meta_data, keep_types = FALSE),
        silent = TRUE))) {
      .meta_data <- rlang::missing_arg()
      # util_error("Need metadata as a data frame", applicability_problem = TRUE)
    }
  }
  if (missing(.meta_data)) {
    w <- paste("Missing %s, try to guess a preliminary one from the data",
           "using %s. Please consider amending this minimum guess manually.")
    if (requireNamespace("cli", quietly = TRUE)) {
      w <- cli::bg_red(cli::col_br_yellow(w))
    }

    util_warning(
      w,
      dQuote("meta_data"),
      dQuote("prep_study2meta"),
      applicability_problem = TRUE, immediate = TRUE
    )
    .meta_data <- prep_study2meta(.study_data, level =
                                    VARATT_REQUIRE_LEVELS$REQUIRED) # recommended would include part vars, which causes many warnings
  } else if (!is.data.frame(.meta_data)) {
    util_error("Need metadata as a data frame", applicability_problem = TRUE)
  }

  .meta_data <- util_normalize_clt(meta_data = .meta_data)
  # prep_add_data_frames(item_level = .meta_data)

  # if no meta_data have been provided -> error
  e <- new.env(parent = environment())
  e$meta_data <- .meta_data
  .meta_data <-
    eval(
      quote(try(util_expect_data_frame(meta_data), silent = TRUE)),
      e
    )

  if (inherits(.meta_data, "try-error")) {
    util_error("Need metadata as a data frame: %s",
               conditionMessage(attr(.meta_data, "condition")))
  }

  util_expect_data_frame(.meta_data)

  if (!prod(dim(.meta_data))) {
    util_warning(
      c("Missing %s, try to guess a preliminary one from the data using %s.",
        "Please consider amending this minimum guess manually."),
      dQuote("meta_data"),
      dQuote("prep_prepare_dataframes"),
      applicability_problem = TRUE
    )
    .meta_data <- prep_study2meta(.study_data, level =
                                    VARATT_REQUIRE_LEVELS$REQUIRED)
  }

  if (.internal) {
    assign("already_ppdf", TRUE, envir = parent.frame())
  }

  if (is.null(.label_col)) {
    .label_col <- VAR_NAMES
  }

  study_data <- .study_data
  meta_data <- .meta_data
  label_col <- .label_col

  if (any(is.na(colnames(study_data)))) {
    util_error("%s must not feature columns with columns names being %s",
               sQuote("study_data"),
               sQuote("NA"), applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  if (case_insens) {
    colnames(study_data) <-
      util_align_colnames_case(.colnames = colnames(study_data),
                               .var_names = meta_data[[VAR_NAMES]])
  }

  # Exchanged to "label_col"
  if (!exists("label_col")) {
    label_col <- VAR_NAMES
  }

  try(if (missing(label_col)) {
    label_col <- VAR_NAMES
  }, silent = TRUE)

  util_expect_scalar(label_col, check_type = is.character,
                     error_message =
                       sprintf("%s needs to be of type character",
                               sQuote("label_col")))

  # TODO: include a check that VAR_NAMES/label_col exists, and that they are unique and not too long
  # intermediate solution - fill empty labels, since we will map to metadata label columns, and otherwise empty fields will cause errors
  for (lcol in unique(c(label_col, LABEL, LONG_LABEL))) {
    if (is.data.frame(meta_data) && lcol %in% colnames(meta_data)) {
      meta_data[[lcol]][which(util_empty(meta_data[[lcol]]))] <-
        meta_data[[VAR_NAMES]][which(util_empty(meta_data[[lcol]]))]
    }
  }

  meta_data <- prep_meta_data_v1_to_item_level_meta_data(meta_data =
                                                           meta_data,
                                                         verbose = FALSE,
                                                         label_col = label_col)

  if (!(DATA_TYPE %in% colnames(meta_data)) ||
      any(util_empty(meta_data[[DATA_TYPE]])) ||
      !all(meta_data[[DATA_TYPE]] %in% DATA_TYPES)) {
    if (!(DATA_TYPE %in% colnames(meta_data))) {
      meta_data[[DATA_TYPE]] <- NA_character_
    }

    meta_data <- .util_fix_data_types(meta_data, study_data)
  }

  if (!.called_in_pipeline) meta_data <- util_validate_known_meta(meta_data)

  study_data_hash <- rlang::hash(.study_data)
  meta_data_hash <- rlang::hash(.meta_data)
  .tabs <- sort(unique(unlist(meta_data[,
                                       endsWith(
                                         colnames(
                                           meta_data),
                                         "_TABLE")])))
  .tabs <- .tabs[!is.na(.tabs)]
  meta_tabs_hash <- rlang::hash(lapply(.tabs,
                                       function(tb) {
                                         try(prep_get_data_frame(tb),
                                             silent = TRUE)
                                       }))
  a <- formals(prep_prepare_dataframes)[.to_combine]
  a2 <- rlang::call_args(sys.call())
  a2 <- a2[intersect(names(a2), .to_combine)]
  a[names(a2)] <- a2
  missing_from_a <- vapply(a, rlang::is_missing, FUN.VALUE = logical(1))
  a[missing_from_a] <- mget(names(which(missing_from_a)))
  a <- rlang::hash(a)
  key <-
    paste0(a,
           "@", study_data_hash,
           "@", meta_data_hash,
           "@", meta_tabs_hash,
           "@", label_col,
           "@", getOption("dataquieR.old_type_adjust",
                          dataquieR.old_type_adjust_default))

  if (key %in% names(.study_data_cache)) {
    if (getOption("dataquieR.study_data_cache_metrics",
                  dataquieR.study_data_cache_metrics_default)) {
      e <- getOption("dataquieR.study_data_cache_metrics_env",
                     dataquieR.study_data_cache_metrics_env_default)
      if (!is.environment(e) ||
          rlang::env_is_locked(e) ||
          any(rlang::env_binding_are_locked(e, intersect(
            names(e),
            c("usage"))))) {
        util_warning(c("in `option()` %s, %s expects an unlocked",
                       "environment with an unlocked binding in %s.
                       Use %s, instead."),
                     sQuote("dataquieR.study_data_cache_metrics_env"),
                     sQuote(packageName),
                     sQuote("usage"),
                     sQuote('dataquieR.study_data_cache_metrics_env_default'))
        e <- dataquieR.study_data_cache_metrics_env_default
        e$usage <- list()
      }
      if (is.null(e$usage)) e$usage <- list()
      if (is.null(e$usage[[key]])) e$usage[[key]] <- 0
      e$usage[[key]] <- e$usage[[key]] + 1
    }
    ds1 <- .study_data_cache[[key]]
    study_data <- attr(ds1, "study_data") # IDEA, but not robust enough: use a hash as name and keep the dataframe in the cache only once
  } else {
    ds1 <- NULL
  }

  # If study_data exist and metadata were already mapped, then we can return ds1
  # directly, but only if all requested modifications are already considered
  # (if possible).
  if (isTRUE(attr(study_data, "MAPPED", exact = TRUE))) { # here, we known, study_data was a ds1, so it features a study_data already with original study data
    ds1 <- study_data # study_data is actually a ds1 already, that may be used
    study_data <- attr(ds1, "study_data") # get the original study_data w/o any attributes (XXX)
    # labels can be modified easily
    if (attr(ds1, "label_col") != label_col) {
      colnames(ds1) <-
        util_map_labels(colnames(ds1),
                        meta_data = meta_data,
                        from = attr(ds1, "label_col"),
                        to = label_col)
      attr(ds1, "label_col") <- label_col
    }
    ds1_ready <- TRUE

    # find out, if ds1 really can be re-used

    # replacement of missing value codes can not be undone
    if (isTRUE(attr(ds1, "Codes_to_NA")) && !.replace_missings) {
      ds1_ready <- FALSE
    } else if (isTRUE(attr(ds1, "Codes_to_NA")) != .replace_missings) {
      ds1_ready <- FALSE
    }
    # replacement of hard limits can not be undone
    if (isTRUE(attr(ds1, "HL_viol_to_NA")) &&
        !.replace_hard_limits) {
      ds1_ready <- FALSE
    } else if (isTRUE(attr(ds1, "HL_viol_to_NA")) != .replace_hard_limits) {
      ds1_ready <- FALSE
    }
    # replacement of hard limits can not be undone, also, it may prevent hard limit replacement, e.g.
    if (isTRUE(attr(ds1, "apply_fact_md"))) {
      ds1_ready <- FALSE
    }
    if (isTRUE(attr(ds1, "apply_fact_md_inadm"))) {
      ds1_ready <- FALSE
    }
    # data type correction can not be undone
    if (isTRUE(attr(ds1, "Data_type_matches")) &&
        !.adjust_data_type) {
      ds1_ready <- FALSE
    } else if (isTRUE(attr(ds1, "Data_type_matches")) != .adjust_data_type) {
      ds1_ready <- FALSE
    }
    check_sl <-
      (.amend_scale_level && SCALE_LEVEL %in% colnames(meta_data) &&
         !any(util_empty(meta_data[[SCALE_LEVEL]][meta_data$VAR_NAMES %in%
                                                    colnames(study_data)]))) ||
      !.amend_scale_level

    ds1_ready <- ds1_ready &&
      check_sl

    if (ds1_ready) { # ds1 can be re-used
      if (.internal) {
        assign("study_data", study_data, parent.frame())
        assign("meta_data", meta_data, parent.frame())
        assign("ds1", ds1, parent.frame())
        assign("label_col", label_col, parent.frame())
        return(invisible(ds1))
      } else {
        return(ds1)
      }
    } else { # need to create a valid ds1, now
      ds1 <- NULL
    }
  }

  # study_data does not feature ds1- attributes any more and ds1 needs to be built.

  if (.replace_missings) {
    # Are missing codes replaced?
    if (!isTRUE(attr(study_data, "Codes_to_NA"))) { # always true, here, because study_data is now not mapped
      study_data <-
        util_replace_codes_by_NA(
          study_data = study_data, meta_data = meta_data,
          split_char = SPLIT_CHAR, sm_code = .sm_code)
    }
  }

  if (!"VAR_NAMES" %in% colnames(meta_data)) {
    # Should get caught before by 'util_validate_known_meta'.
    util_error("'VAR_NAMES' not found in metadata [%s]",
               paste0(colnames(meta_data), collapse = ", "),
               applicability_problem = TRUE)
  }

  if (!isTRUE(attr(study_data, "MAPPED", exact = TRUE))) { # always true, here, because study_data is now not mapped
    study_data <- study_data[, order(colnames(study_data)), FALSE]
    meta_data <- meta_data[order(meta_data[[VAR_NAMES]]), , FALSE]
  }

  relevant_vars_for_warnings <- NULL

  # adjust data types, if enabled
  util_stop_if_not(
    `Pipeline should never request study data w/ unchanged datatypes, sorry, internal error, please report` =
                     !(!.adjust_data_type && .called_in_pipeline))
  if (.adjust_data_type || .apply_factor_metadata) {
    relevant_vars_for_warnings <- lapply(
      setNames(nm = names(.meta_data_env)[endsWith(names(.meta_data_env), "_vars")]), # all arguments populated by the pipeline with some variable references
      util_find_indicator_function_in_callers
    )
    if (is.null(relevant_vars_for_warnings$resp_vars)) { # original call was not for item level, so no filtering of problems by related items
      relevant_vars_for_warnings <- NULL
    }
    relevant_vars_for_warnings <- unlist(relevant_vars_for_warnings,
                                         recursive = TRUE)

    # Important: Remove all non-character matches (maybe, our call-stack kept the metadata-env-look-up-functions, whyever)
    relevant_vars_for_warnings <-
      relevant_vars_for_warnings[vapply(relevant_vars_for_warnings,
                                         is.character, FUN.VALUE = logical(1))]

    relevant_vars_for_warnings <-
      relevant_vars_for_warnings[!util_empty(relevant_vars_for_warnings)]

    try(relevant_vars_for_warnings <- # TODO: sometimes, relevant... does not have variables from label_col, if correct variable use used find var by names and mapped them to varnames already # this corresponds with TODO POSSIBLE_VARS in util_correct_variable_use
          util_find_var_by_meta(relevant_vars_for_warnings,
                                meta_data = meta_data,
                                label_col = label_col,
                                target = VAR_NAMES,
                                allowed_sources =
                                  c(VAR_NAMES, LABEL, LONG_LABEL, label_col),
                                ifnotfound = relevant_vars_for_warnings),
        silent = TRUE)

  }
  if (.adjust_data_type && !.called_in_pipeline) {
    study_data <- util_adjust_data_type(
      study_data = study_data,
      meta_data = meta_data,
      relevant_vars_for_warnings = relevant_vars_for_warnings)
  }

  if (.amend_scale_level && (!(SCALE_LEVEL %in% colnames(meta_data)) ||
                             any(util_empty(
                               meta_data[[SCALE_LEVEL]][meta_data$VAR_NAMES %in%
                                                        colnames(study_data)])))) {
    util_message(c(
      "Missing some or all entries in %s column in item-level %s. Predicting",
                   "it from the data -- please verify these predictions, they",
                   "may be wrong and lead to functions claiming not to be",
                   "reasonably applicable to a variable."),
                   sQuote(SCALE_LEVEL), "meta_data",
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = FALSE)

    meta_data <- prep_scalelevel_from_data_and_metadata(
      study_data = study_data,
      meta_data = meta_data,
      label_col = label_col
    )
  }

  # create ds1 -----------------------------------------------------------------
  if (is.null(ds1) || !isTRUE(attr(ds1, "MAPPED", exact = TRUE))) {
    ds1 <- util_map_all(label_col = label_col, study_data = study_data,
                        meta_data = meta_data)$df
  }
  if (!(.allow_empty) && ncol(ds1) * nrow(ds1) == 0) {
    util_error(
      "No data left. Aborting, since mapping of %s on %s was not possible",
      sQuote("meta_data"),
      sQuote("study_data"),
      applicability_problem = FALSE)
  }

  my_atts <- attributes(ds1)[.ds1_attribute_names]
  ds1 <- as.data.frame(ds1)
  attributes(ds1)[.ds1_attribute_names] <- my_atts

  meta_data <- as.data.frame(meta_data)

  .all <- ncol(ds1)
  my_atts <- attributes(ds1)[.ds1_attribute_names]
  ds1 <- ds1[, colnames(ds1) %in% meta_data[[label_col]], FALSE]
  attributes(ds1)[.ds1_attribute_names] <- my_atts
  .mapped <- ncol(ds1)
  if (.all > .mapped) { # nocov start
    # Should be dead code, because util_map_all is called above, which
    # performs an analogous check and clears out the unannotated variables
    # from the study data.
    util_warning("Lost %d variables, that I could not map using %s",
                 .all - .mapped, dQuote(label_col),
                 applicability_problem = TRUE)
  } # nocov end
  if (!.allow_empty) {
    # this would delete the metadata, if the mapping has failed totally
    meta_data <- meta_data[meta_data[[label_col]] %in% colnames(ds1), , FALSE]
  }

  if (STUDY_SEGMENT %in% colnames(meta_data) &&
      any(util_empty(meta_data[[STUDY_SEGMENT]]))) {
    dummy_name <- "SEGMENT"
    i <- 1
    while (dummy_name %in% meta_data[[STUDY_SEGMENT]]) {
      dummy_name <- sprintf("SEGMENT %d", i)
      i <- i + 1
    }
    util_message(c(
      "Some %s are NA. Will assign those to an artificial",
      "segment %s"), sQuote(STUDY_SEGMENT), dQuote(dummy_name),
      applicability_problem = TRUE
    )
    meta_data[[STUDY_SEGMENT]][util_empty(meta_data[[STUDY_SEGMENT]])] <-
      dummy_name
  }

  if (VARIABLE_ORDER %in% colnames(meta_data)) {
    meta_data[] <- meta_data[order(meta_data[[VARIABLE_ORDER]]), , drop = FALSE]
    vars <- meta_data[, label_col, drop = TRUE]
    vars <- vars[vars %in% colnames(ds1)]
    my_atts <- attributes(ds1)[.ds1_attribute_names]
    ds1 <- ds1[, vars, drop = FALSE]
    attributes(ds1)[.ds1_attribute_names] <- my_atts
  }

  if (!isTRUE(attr(study_data, "MAPPED", exact = TRUE)) &&  # !mapped is always true, here, because study_data is now not mapped
      PART_VAR %in% colnames(meta_data)) local({
    .kssvs <- intersect(colnames(study_data), meta_data[, PART_VAR])
    for (rv in .kssvs) {
      vals <- util_replace_codes_by_NA(
        study_data = study_data[, rv, drop = FALSE],
        meta_data = meta_data, sm_code = .sm_code)[[rv]]
      vals <- vals[!is.na(vals)]
      if (is.character(vals)) vals <- vals[trimws(vals) != ""]
      vals <- suppressWarnings(as.numeric(vals))
      if (!all(vals %in% c(0:1, NA), na.rm = TRUE)) {
        util_warning(c( # TODO: Maybe too frequently shown? see relevant_vars_for_warnings above
          "Found entries different from TRUE/FALSE/1/0 and <empty>, segment",
          "participation is expected, if values different from 0 are found.",
          "For the segment indicator variable %s, a table of inadmissible",
          "values: %s"),
          dQuote(rv),
          paste(capture.output(print(table(setdiff(vals, c(0:1, NA))))),
                collapse = "\n"))
      }
    }
  })

  attr(study_data, "Codes_to_NA") <- .replace_missings
  attr(ds1, "Codes_to_NA") <- .replace_missings

  attr(ds1, "MAPPED") <- TRUE
  attr(ds1, "label_col") <- label_col
  attr(ds1, "Data_type_matches") <- .adjust_data_type

  if (.replace_hard_limits) {
    # Are hard limit violations replaced?
    if (!isTRUE(attr(ds1, "HL_viol_to_NA"))) {
      ds1 <-
        util_replace_hard_limit_violations(
          study_data = ds1, meta_data = meta_data, label_col = label_col)
    }
  }

  attr(ds1, "HL_viol_to_NA") <- .replace_hard_limits

  if (.apply_factor_metadata) {
    relevant_vars_for_warnings_lb <-
      util_map_labels(relevant_vars_for_warnings, meta_data = meta_data,
                      from = VAR_NAMES, to = label_col,
                      relevant_vars_for_warnings)

    # convert columns to factors?
    if (!isTRUE(attr(ds1, "apply_fact_md"))) { # !apply_fact_md is always true, here, because study_data is now not mapped
      which_cols <- intersect(meta_data[[attr(ds1, "label_col")]],
                              colnames(ds1))
      ds1[, which_cols] <- lapply(which_cols,
                                  function(cl) { # IDEA: What to do with missing labels for numerical variables?
                                    # prep_load_workbook_like_file("meta_data_v2")
                                    # (prep_prepare_dataframes(.study_data = "study_data", .meta_data = il, .label_col = LABEL, .apply_factor_metadata = TRUE))
                                    if (!(
                                      SCALE_LEVEL %in% colnames(meta_data))) {
                                      util_error(c("%s was called with %s = %s",
                                                   "and %s = %s",
                                                   "but %s does not provide a",
                                                   "column %s -- please provide",
                                                   "this column or call with ",
                                                   "%s = %s"),
                                                 sQuote(rlang::call_name(sys.call(1))),
                                                 sQuote(".apply_factor_metadata"),
                                                 dQuote(.apply_factor_metadata),
                                                 sQuote(".amend_scale_level"),
                                                 dQuote(.amend_scale_level),
                                                 "meta_data",
                                                 sQuote(SCALE_LEVEL),
                                                 sQuote(".amend_scale_level"),
                                                 dQuote(TRUE),
                                                 applicability_problem = TRUE)
                                    }
                                    sl <- meta_data[meta_data[[attr(ds1, "label_col")]] == cl,
                                                    SCALE_LEVEL, drop = TRUE]
                                    if (sl %in% c(
                                      SCALE_LEVELS$NOMINAL, SCALE_LEVELS$ORDINAL
                                    )) {
                                      util_stop_if_not(
                                        `Internal error, sorry, please report: unexp. VALUE_LABELS` =
                                          util_empty(meta_data[
                                            meta_data[[attr(ds1, "label_col")]] == cl,
                                            VALUE_LABELS, drop = TRUE]))
                                      vlt <- try(prep_get_data_frame(
                                        meta_data[meta_data[[attr(ds1, "label_col")]] == cl,
                                                  VALUE_LABEL_TABLE, drop = TRUE]), silent = TRUE)
                                      levels <- NULL
                                      labels <- NULL
                                      if (!util_is_try_error(vlt)) {
                                        if (CODE_VALUE %in% colnames(vlt)) {
                                          levels <- vlt[[CODE_VALUE]]
                                        }
                                        if (CODE_LABEL %in% colnames(vlt)) {
                                          labels <- vlt[[CODE_LABEL]]
                                        }
                                        if (is.null(levels) && !is.null(labels)) {
                                          util_error(applicability_problem = TRUE,
                                                     "Have only code labels, but not code levels for %s",
                                                     dQuote(cl)
                                          )
                                        } else if (is.null(labels)) {
                                          labels <- levels
                                        }
                                      }
                                      ordered <- sl == SCALE_LEVELS$ORDINAL
                                      if (is.null(levels)) {
                                        levels <- unique(sort(ds1[[cl]]))
                                      }
                                      if (is.null(labels)) {
                                        labels <- levels
                                      }
                                      if (!.replace_missings) {
                                        .m <- util_get_code_list(x = cl,
                                                           code_name = MISSING_LIST,
                                                           mdf = meta_data,
                                                           label_col = label_col,
                                                           warning_if_no_list = FALSE,
                                                           warning_if_unsuitable_list = FALSE)
                                        .j <- util_get_code_list(x = cl,
                                                           code_name = JUMP_LIST,
                                                           mdf = meta_data,
                                                           label_col = label_col,
                                                           warning_if_no_list = FALSE,
                                                           warning_if_unsuitable_list = FALSE)
                                        defined <- levels %in% c(unname(.m),
                                                                 unname(.j))
                                        levels <- levels[!defined]
                                        labels <- labels[!defined]

                                        levels <- c(levels, unname(.j))
                                        labels <- c(labels, names(.j))
                                        levels <- c(levels, unname(.m))
                                        labels <- c(labels, names(.m))
                                      }
                                      if (.apply_factor_metadata_inadm) {
                                        empir <- unique(ds1[[cl]])
                                        empir <- empir[!is.na(empir)]
                                        empir <- setdiff(empir, levels)
                                        levels <- c(levels, empir)
                                        labels <- c(labels, empir)
                                      }
                                      .fct <- do.call(factor, list(
                                        x = ds1[[cl]],
                                        levels = levels,
                                        labels = labels,
                                        ordered = ordered
                                      ), quote = TRUE);
                                      if
                                        (any(is.na(.fct) != is.na(ds1[[cl]]))) {
                                        hint <- sprintf(paste(
                                          c("Found the following",
                                            "inadmissible categorical",
                                            "values in %s: %s -- I've",
                                            "removed them.")),
                                          sQuote(cl),
                                          util_pretty_vector_string(
                                            unique(ds1[[cl]][is.na(.fct) !=
                                                               is.na(ds1[[cl]])])
                                          ))
                                      } else {
                                        hint <- character(0)
                                      }
                                      .fct <-
                                        util_attach_attr(
                                          .fct,
                                          hint = hint)
                                    } else {
                                      .fct <- ds1[[cl]]
                                    }
                                    return(.fct)
                                  }
      )
    }
    # message for hints in columns - currently only inadm.cat.values.
    for (cl in intersect(relevant_vars_for_warnings_lb,
                         colnames(ds1))) {
      if (length(attr(ds1[[cl]], "hint")) > 0 &&
          !all(is.na(attr(ds1[[cl]], "hint")))) {
        util_message(attr(ds1[[cl]], "hint"))
      }
    }

    ##
  }

  attr(ds1, "apply_fact_md") <- .apply_factor_metadata
  attr(ds1, "apply_fact_md_inadm") <- .apply_factor_metadata_inadm

  study_data <- util_cast_off(study_data, "study_data", TRUE)
  meta_data <- util_cast_off(meta_data, "meta_data", TRUE)
  ds1 <- util_cast_off(ds1, "ds1", TRUE)
  attr(ds1, "study_data") <- study_data

  if (.internal) {
    assign("study_data", study_data, parent.frame())
    assign("meta_data", meta_data, parent.frame())
    assign("ds1", ds1, parent.frame())
    assign("label_col", label_col, parent.frame())
  }

  dataquieR.study_data_cache_max <-
    getOption("dataquieR.study_data_cache_max",
              dataquieR.study_data_cache_max_default)

  if (is.data.frame(ds1) &&
      !(key %in% names(.study_data_cache))) {
      if (sum(object.size(ds1),
              vapply(.study_data_cache, object.size, FUN.VALUE = numeric(1)),
              na.rm = TRUE) <= dataquieR.study_data_cache_max) {
        .study_data_cache[[key]] <- util_attach_attr(ds1, call = sys.call())
      } else if (getOption("dataquieR.study_data_cache_metrics",
                           dataquieR.study_data_cache_metrics_default)) {
        class(dataquieR.study_data_cache_max) <- "object_size"
        rlang::inform(
          sprintf(
            paste(
              "Maximum size for study_data cache (%s) reached, not caching.",
              "Adjust the %s to control"),
            format(dataquieR.study_data_cache_max, units = "auto"),
            sQuote("option(dataquieR.study_data_cache_max = )")
          ),
          .frequency = "once", .frequency_id = paste(key)
        )
      }
  }

  if (.internal) {
    invisible(ds1)
  } else {
    ds1
  }
}

util_maybe_load_meta_data_v2 <- function() {
  if ("meta_data_v2" %in% rlang::fn_fmls_names(fn = rlang::caller_fn(1))) {
    eval.parent(substitute({
      if (!missing(meta_data_v2)) {
        util_message("Have %s set, so I'll remove all loaded data frames",
                     sQuote("meta_data_v2"))
        prep_purge_data_frame_cache()
        prep_load_workbook_like_file(meta_data_v2)
        if (!exists("item_level", .dataframe_environment())) {
          w <- paste("Did not find any sheet named %s in %s, is this",
                     "really dataquieR version 2 metadata?")
          if (requireNamespace("cli", quietly = TRUE)) {
            w <- cli::bg_red(cli::col_br_yellow(w))
          }
          util_warning(w, dQuote("item_level"), dQuote(meta_data_v2),
                       immediate = TRUE)
        }
      }
    }))
  } else {
    util_error(c("Internal error: Need a %s formal, sorry. Please report.",
                 "As a dataquieR developer: You need to declare a formal",
                 "argument %s, if you want to call %s."),
               sQuote("meta_data_v2"),
               sQuote("meta_data_v2"),
               sQuote("util_maybe_load_meta_data_v2"))
  }
}

# fixes aliases for all metadata levels except item level for the data frame names as well as for arguemnts of exported functions
util_ck_arg_aliases <- function() {
  arg_maps_to = c( # this function does not handle item_level/meta_data, because this pair is addressed by prep_prepare_dataframes()
    segment_level = "meta_data_segment",
    cross_item_level = "meta_data_cross_item",
    `cross-item_level` = "meta_data_cross_item",
    dataframe_level = "meta_data_dataframe",
    item_computation_level =
      "meta_data_item_computation"
  )

  # handle metadata argument aliases --------
  caller_formal_names <- rlang::fn_fmls_names(fn = rlang::caller_fn(1))
  caller_formals <- rlang::fn_fmls(fn = rlang::caller_fn(1))
  missing_in_parent <- suppressWarnings({
    vapply(caller_formal_names, function(x) eval(call("missing",
                                                         as.symbol(x)),
                                                    envir = parent.frame(3)),
           FUN.VALUE = logical(1))
  })

  for (arg in intersect(names(arg_maps_to),
                        caller_formal_names)) {
    list_of_synonyms <- names(arg_maps_to)[arg_maps_to == arg_maps_to[[arg]]]
    for (syn in setdiff(
      union(list_of_synonyms, arg_maps_to[[arg]]), arg)) {
      if (syn %in% caller_formal_names) { # have both
        if (!missing_in_parent[[syn]] &&
            !missing_in_parent[[arg]] &&
            !identical(dynGet(syn), dynGet(arg))) {
          # IDEA: if (rlang::hash())
          util_error(c("You cannot provide both, %s as well as %s",
                       "these arguments are synonyms and must be",
                       "used mutually exclusively"),
                     sQuote(syn),
                     sQuote(arg))
        }
      }
    }
    v <- rlang::missing_arg()
    if (!missing_in_parent[[arg]]) {
      v <- eval.parent(as.symbol(arg))
    } else for (syn in setdiff(
      union(list_of_synonyms, arg_maps_to[[arg]]), arg)) {
      if (!missing_in_parent[[syn]]) {
        v <- eval.parent(as.symbol(syn))
      }
    }
    if (!missing(v)) {
      assign(arg_maps_to[[arg]], v, envir = parent.frame())
      for (syn in list_of_synonyms) {
        assign(syn, v, envir = parent.frame())
      }
    }
  }

  # handle data frame alias names --------
  # order does matter, because if the above stops,
  # the data frame cache stays untouched
  df_names <- prep_list_dataframes()
  cils2rn <- grep("(^|\\|)\\s*cross_item_level$", perl =  TRUE, value = TRUE,
                  df_names)
  for (cil2rn in cils2rn) { # rename dataframes named cross_item_level
    new_nm <- sub("cross_item_level$", "cross-item_level", cil2rn)
    if (!new_nm %in% df_names) {
      prep_add_data_frames(
        data_frame_list =
          setNames(list(
            prep_get_data_frame(cil2rn)
          ), nm = new_nm)
      )
    }
  }

}

# used, in case ds1 is copied or similar to keep the attributes in place
.ds1_attribute_names <- c("Codes_to_NA", "MAPPED", "label_col", "HL_viol_to_NA",
                          "Data_type_matches", "apply_fact_md",
                          "apply_fact_md_inadm", "study_data",
                          "normalized", "version")

.util_fix_data_types <- function(meta_data, study_data) {
  if (missing(study_data))
    study_data <- data.frame()
  which_wrong <-
    util_empty(meta_data[[DATA_TYPE]]) |
    !(meta_data[[DATA_TYPE]] %in% DATA_TYPES)

  wrong_names <- meta_data[which_wrong, VAR_NAMES, drop = TRUE]

  wrong_names <- intersect(colnames(study_data), wrong_names)

  datatypes <- prep_datatype_from_data(resp_vars = wrong_names,
                                       study_data = study_data)

  if (length(datatypes) != 0) {
    util_warning(
      c("For the variables %s, I have no valid %s in the %s. I've predicted",
        "the %s from the %s yielding %s."),
      util_pretty_vector_string(wrong_names),
      sQuote(DATA_TYPE),
      sQuote("meta_data"),
      sQuote(DATA_TYPE),
      sQuote("study_data"),
      dQuote(prep_deparse_assignments(names(datatypes), datatypes,
                                      mode = "string_codes")),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = FALSE
    )
  }
  meta_data[which_wrong, DATA_TYPE] <-
    datatypes[meta_data[which_wrong, VAR_NAMES]]

  meta_data
}

test_that("prep_min_obs_level works", {
  skip_on_cran() # function not used, yet
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  label_col <- LABEL
  prep_prepare_dataframes(.internal = TRUE)
  expect_message2(
    x <- prep_min_obs_level(ds1,
                            group_vars = "USR_BP_0",
                            min_obs_in_subgroup = 50),
    regexp =
      paste("The following levels: .+USR_559.+ have < 50",
            "observations and are disregarded"),
    perl = TRUE,
  )
  expect_equal(nrow(study_data) - nrow(x), 29)


  expect_error(
    x <- prep_min_obs_level(ds1,
                            group_vars = character(0),
                            min_obs_in_subgroup = 50
    ),
    regexp =
      paste(".+group_vars.+ is required to name exactly one variable."),
    perl = TRUE
  )

  expect_error(
    x <- prep_min_obs_level(ds1,
                            group_vars = NULL,
                            min_obs_in_subgroup = 50
                            ),
    regexp =
      paste(".+group_vars.+ is required to be a character.* argument."),
    perl = TRUE
  )

  expect_message2(
    expect_error(
      x <- prep_min_obs_level(ds1,
                              group_vars = letters,
                              min_obs_in_subgroup = 50
      ),
      regexp =
        paste(".+group_vars.+ = .+a.+ is not a variable."),
      perl = TRUE
    ),
    regexp = sprintf("(%s)",
                     paste("Subsets based only on one variable possible.")),
    perl = TRUE
  )

  expect_message2(
    x <- prep_min_obs_level(ds1,
                            group_vars = "USR_BP_0",
                            min_obs_in_subgroup = NA),
    regexp = sprintf("(%s|%s)",
      paste("The following levels: .+USR_559.+ have < 30",
            "observations and are disregarded"),
      paste("argument .+min_obs_in_subgroup.+ was missing,",
            "not of length 1 or NA, setting to its default, 30")),
    perl = TRUE
  )
  expect_equal(nrow(study_data) - nrow(x), 29)

  ds1. <- ds1
  ds1.$USR_BP_0 <- NA
  expect_message2(
    expect_error(
      x <- prep_min_obs_level(ds1.,
                              group_vars = "USR_BP_0",
                              min_obs_in_subgroup = NA),
      regexp = paste("For .+group_vars.+ = .+USR_BP_0.+,",
                     "observations cannot be counted."),
      perl = TRUE
    ),
    regexp = sprintf("(%s|%s)",
                     paste("The following levels: .+USR_559.+ have < 30",
                           "observations and are disregarded"),
                     paste("argument .+min_obs_in_subgroup.+ was missing,",
                           "not of length 1 or NA, setting to its default, 30")),
    perl = TRUE
  )

  ds1. <- subset(ds1, USR_BP_0 != "USR_559")
  expect_silent(
    x <- prep_min_obs_level(ds1.,
                            group_vars = "USR_BP_0",
                            min_obs_in_subgroup = 50)
  )
  expect_equal(nrow(ds1.) - nrow(x), 0)

})

test_that("util_heatmap_1th works", {
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_on_cran() # will be remoed, currently used only by segment missingness which is also tested
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  label_col <- LABEL
  prep_prepare_dataframes(.internal = TRUE)
  hm1 <- util_heatmap_1th(
    df = ds1,
    cat_vars = c("CENTER_0", "USR_BP_0"),
    values = "SBP_0",
    threshold = 100,
    invert = FALSE
  )
  expect_lt(
    abs(suppressWarnings(
      sum(as.numeric(as.matrix(hm1$SummaryPlot$data)), na.rm = TRUE)
    ) - 1134545),
    0.8
  )
  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  hm2 <- util_heatmap_1th(
    df = ds1,
    cat_vars = c("USR_BP_0"),
    values = "SBP_0",
    threshold = 100,
    invert = TRUE,
    strata = "CENTER_0"
  )
  hm3 <- util_heatmap_1th(
    df = ds1,
    cat_vars = c("USR_BP_0"),
    values = "SBP_0",
    threshold = 100,
    invert = TRUE,
    strata = "CENTER_0",
    right_intv = TRUE
  )
  suppressWarnings({
    expect_doppelganger2("util_heatmap_1th_1",
                                hm1$SummaryPlot)
    expect_doppelganger2("util_heatmap_1th_2",
                                hm2$SummaryPlot)
    expect_doppelganger2("util_heatmap_1th_3",
                                hm3$SummaryPlot)
  })
})

test_that("util_dichotomize works", {
  skip_on_cran() # TODO: activate?
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data[meta_data[[LABEL]] == "EATING_PREFS_0", RECODE_CONTROL] <-
    "0" # 0 = "eat meat" as reference
  meta_data[meta_data[[LABEL]] == "MEAT_CONS_0", RECODE_CASES] <-
    "1| 2|3 | 4 " # eats meat more frequently than 0="never"
  meta_data[meta_data[[LABEL]] == "SBP_0", RECODE_CASES] <-
    "(-Inf;100] |[140;Inf)"
  meta_data[meta_data[[LABEL]] == "SBP_0", RECODE_CONTROL] <-
    "[120;130)"
  m_study_data <- util_replace_codes_by_NA(study_data, meta_data)
  d_study_data <- util_dichotomize(m_study_data, meta_data)
  x <- d_study_data[,
                    prep_map_labels(c("EATING_PREFS_0", "MEAT_CONS_0"),
                                    from = LABEL, to = VAR_NAMES, meta_data = meta_data)]
  expect_equal(
    d_study_data[["v00022"]],
    ifelse(is.na(m_study_data[["v00022"]]),
           NA,
           ifelse(m_study_data[["v00022"]] %in% 0,
                  0, # eat meat
                  1  # veg*
           )
    )
  )
  expect_equal(
       d_study_data[["v00023"]],
       ifelse(is.na(m_study_data[["v00023"]]),
               NA,
         ifelse(m_study_data[["v00023"]] %in% 1:4,
                 1, # eats meat on some days
                 0  # never eats meat
               )
             )
     )
  col_int <- NA
  col_int[which(m_study_data[["v00004"]] <= 100)] <- 1
  col_int[which(m_study_data[["v00004"]] >= 140)] <- 1
  col_int[which(m_study_data[["v00004"]] >= 120 &
                  m_study_data[["v00004"]] < 130)] <- 0
  expect_equal(d_study_data[["v00004"]],
               col_int)
})

test_that("util_dichotomize is robust", {
  skip_on_cran() # TODO: activate?
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  # with wrong input for the recoding, the function returns the original data
  meta_data[meta_data[[LABEL]] == "EATING_PREFS_0", RECODE_CONTROL] <-
    "| |"
  meta_data[meta_data[[LABEL]] == "MEAT_CONS_0", RECODE_CASES] <-
    "1| 2|3 | 4 " # eats meat more than 0="never"
  meta_data[meta_data[[LABEL]] == "MEAT_CONS_0", RECODE_CONTROL] <-
    "1 "
  meta_data[meta_data[[LABEL]] == "SBP_0", RECODE_CASES] <-
    "(-Inf;112] |[125;Inf)"
  meta_data[meta_data[[LABEL]] == "SBP_0", RECODE_CONTROL] <-
    "[110;130)"
  m_study_data <- util_replace_codes_by_NA(study_data, meta_data)
  d_study_data <- suppressWarnings(util_dichotomize(m_study_data, meta_data))
  x <- d_study_data[,
                    prep_map_labels(c("EATING_PREFS_0", "MEAT_CONS_0"),
                                    from = LABEL, to = VAR_NAMES, meta_data = meta_data)]
  expect_equal(d_study_data[["v00022"]],
               m_study_data[["v00022"]])
  expect_equal(d_study_data[["v00023"]],
               m_study_data[["v00023"]])
  expect_equal(d_study_data[["v00004"]],
               m_study_data[["v00004"]])

  # ensure that the function can also be applied to ds1
  ds1 <- prep_prepare_dataframes(.study_data = study_data,
                                 .meta_data = meta_data, .label_col = LABEL,
                                 .apply_factor_metadata = TRUE)
  meta_data[meta_data[[LABEL]] == "SBP_0", RECODE_CONTROL] <- "(112;125)"
  meta_data[meta_data[[LABEL]] == "SBP_0", RECODE_CASES] <- ""
  meta_data[meta_data[[LABEL]] == "MEAT_CONS_0", RECODE_CASES] <- "1| 2|3 | 4" # factor from integer variable with integer codes
  meta_data[meta_data[[LABEL]] == "MEAT_CONS_0", RECODE_CONTROL] <- ""
  meta_data[meta_data[[LABEL]] == "EATING_PREFS_0", RECODE_CASES] <- "vegetarian | vegan" # factor from integer variable with recoding specified using value labels
  meta_data[meta_data[[LABEL]] == "EATING_PREFS_0", RECODE_CONTROL] <- ""
  meta_data[meta_data[[LABEL]] == "USR_SOCDEM_0", RECODE_CONTROL] <- "USR_321|USR_247|USR_520" # factor from string variable with recoding specified using value labels
  meta_data[meta_data[[LABEL]] == "USR_SOCDEM_0", RECODE_CASES] <- ""
  d_ds1 <- suppressWarnings(util_dichotomize(ds1, meta_data, label_col = LABEL))
  expect_equal(d_ds1[["MEAT_CONS_0"]],
               ifelse(is.na(m_study_data[["v00023"]]),
                      NA,
                      ifelse(m_study_data[["v00023"]] %in% 1:4,
                             1, # eats meat on some days
                             0  # never eats meat
                      )
               ))
  expect_equal(d_ds1[["EATING_PREFS_0"]],
               ifelse(is.na(m_study_data[["v00022"]]),
                      NA,
                      ifelse(m_study_data[["v00022"]] %in% 0,
                             0, # eat meat
                             1  # veg*
                      )
               ))
  expect_equal(d_ds1[["USR_SOCDEM_0"]],
               ifelse(is.na(ds1[["USR_SOCDEM_0"]]),
                      NA,
                      ifelse(ds1[["USR_SOCDEM_0"]] %in% c("USR_321", "USR_247", "USR_520"),
                             0,
                             1
                      )
               ))
  col_int <- NA
  col_int[which(m_study_data[["v00004"]] <= 112)] <- 1
  col_int[which(m_study_data[["v00004"]] >= 125)] <- 1
  col_int[which(m_study_data[["v00004"]] > 112 &
                  m_study_data[["v00004"]] < 125)] <- 0
  expect_equal(d_ds1[["SBP_0"]],
               col_int)
})

test_that("prep_prepare_dataframes works", {
  skip_on_cran()
  
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE,
                       dataquieR.guess_missing_codes = TRUE
                       )
  acc_test1 <- function(resp_variable, aux_variable,
                        time_variable, co_variables,
                        group_vars, study_data, meta_data) {
    prep_prepare_dataframes()
    invisible(ds1)
  }
  acc_test2 <- function(resp_variable, aux_variable,
                        time_variable, co_variables,
                        group_vars, study_data, meta_data, label_col) {
    ds1 <- prep_prepare_dataframes(study_data, meta_data)
    invisible(ds1)
  }
  environment(acc_test1) <- asNamespace("dataquieR")
  environment(acc_test2) <- asNamespace("dataquieR")
  acc_test3 <- function(resp_variable, aux_variable, time_variable,
                        co_variables, group_vars, study_data, meta_data,
                        label_col) {
    prep_prepare_dataframes()
    invisible(ds1)
  }
  acc_test4 <- function(resp_variable, aux_variable, time_variable,
                        co_variables, group_vars, study_data, meta_data,
                        label_col) {
    ds1 <- prep_prepare_dataframes(study_data, meta_data)
    invisible(ds1)
  }
  environment(acc_test3) <- asNamespace("dataquieR")
  environment(acc_test4) <- asNamespace("dataquieR")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]

  expect_error(
    acc_test1(),
    regexp = "Need study data as a data frame"
  )
  expect_error(
    acc_test2(),
    regexp = "Need study data as a data frame"
  )
  expect_warning(
    acc_test1(study_data = study_data),
    regexp =
      sprintf(
        "(?ms).*(%s|%s|%s|%s).*",
        paste("Missing .+meta_data.+, try to guess a preliminary one from the",
              "data using .+prep_study2meta.+. Please consider amending",
              "this minimum guess manually."),
        paste("Metadata does not provide a filled column called",
              ".+MISSING_LIST.+ for replacing codes with NAs."),
        paste("Metadata does not provide a filled column called .+JUMP_LIST.+",
              "for replacing codes with NAs."),
        paste("Some code labels or -values",
              "are missing from .+meta_data.+.")
      ),
    perl = TRUE
  )
  expect_error(
    acc_test1(meta_data = meta_data),
    regexp = "Need study data as a data frame"
  )
  expect_error(
    acc_test2(study_data = 12, meta_data = meta_data),
    regexp = "Need study data as a data frame"
  )
  expect_equal(colnames(acc_test1(study_data = study_data,
                                  meta_data = meta_data)),
               c(
                 "v00000",
                 "v00001",
                 "v00002",
                 "v00003",
                 "v00103",
                 "v01003",
                 "v01002",
                 "v10000",
                 "v00004",
                 "v00005",
                 "v00006",
                 "v00007",
                 "v00008",
                 "v00009",
                 "v00109",
                 "v00010",
                 "v00011",
                 "v00012",
                 "v00013",
                 "v20000",
                 "v00014",
                 "v00015",
                 "v00016",
                 "v00017",
                 "v30000",
                 "v00018",
                 "v01018",
                 "v00019",
                 "v00020",
                 "v00021",
                 "v00022",
                 "v00023",
                 "v00024",
                 "v00025",
                 "v00026",
                 "v00027",
                 "v00028",
                 "v00029",
                 "v00030",
                 "v00031",
                 "v00032",
                 "v00033",
                 "v40000",
                 "v00034",
                 "v00035",
                 "v00036",
                 "v00037",
                 "v00038",
                 "v00039",
                 "v00040",
                 "v00041",
                 "v00042",
                 "v50000"
               )
  )
  expect_equal(colnames(acc_test2(study_data = study_data,
                                  meta_data = meta_data)),
               c(
                 "v00000",
                 "v00001",
                 "v00002",
                 "v00003",
                 "v00103",
                 "v01003",
                 "v01002",
                 "v10000",
                 "v00004",
                 "v00005",
                 "v00006",
                 "v00007",
                 "v00008",
                 "v00009",
                 "v00109",
                 "v00010",
                 "v00011",
                 "v00012",
                 "v00013",
                 "v20000",
                 "v00014",
                 "v00015",
                 "v00016",
                 "v00017",
                 "v30000",
                 "v00018",
                 "v01018",
                 "v00019",
                 "v00020",
                 "v00021",
                 "v00022",
                 "v00023",
                 "v00024",
                 "v00025",
                 "v00026",
                 "v00027",
                 "v00028",
                 "v00029",
                 "v00030",
                 "v00031",
                 "v00032",
                 "v00033",
                 "v40000",
                 "v00034",
                 "v00035",
                 "v00036",
                 "v00037",
                 "v00038",
                 "v00039",
                 "v00040",
                 "v00041",
                 "v00042",
                 "v50000"
               )
  )
  expect_equal(colnames(acc_test3(study_data = study_data,
                                  meta_data = meta_data)),
               c(
                 "v00000",
                 "v00001",
                 "v00002",
                 "v00003",
                 "v00103",
                 "v01003",
                 "v01002",
                 "v10000",
                 "v00004",
                 "v00005",
                 "v00006",
                 "v00007",
                 "v00008",
                 "v00009",
                 "v00109",
                 "v00010",
                 "v00011",
                 "v00012",
                 "v00013",
                 "v20000",
                 "v00014",
                 "v00015",
                 "v00016",
                 "v00017",
                 "v30000",
                 "v00018",
                 "v01018",
                 "v00019",
                 "v00020",
                 "v00021",
                 "v00022",
                 "v00023",
                 "v00024",
                 "v00025",
                 "v00026",
                 "v00027",
                 "v00028",
                 "v00029",
                 "v00030",
                 "v00031",
                 "v00032",
                 "v00033",
                 "v40000",
                 "v00034",
                 "v00035",
                 "v00036",
                 "v00037",
                 "v00038",
                 "v00039",
                 "v00040",
                 "v00041",
                 "v00042",
                 "v50000"
               )
  )
  expect_equal(colnames(acc_test4(study_data = study_data,
                                  meta_data = meta_data)),
               c(
                 "v00000",
                 "v00001",
                 "v00002",
                 "v00003",
                 "v00103",
                 "v01003",
                 "v01002",
                 "v10000",
                 "v00004",
                 "v00005",
                 "v00006",
                 "v00007",
                 "v00008",
                 "v00009",
                 "v00109",
                 "v00010",
                 "v00011",
                 "v00012",
                 "v00013",
                 "v20000",
                 "v00014",
                 "v00015",
                 "v00016",
                 "v00017",
                 "v30000",
                 "v00018",
                 "v01018",
                 "v00019",
                 "v00020",
                 "v00021",
                 "v00022",
                 "v00023",
                 "v00024",
                 "v00025",
                 "v00026",
                 "v00027",
                 "v00028",
                 "v00029",
                 "v00030",
                 "v00031",
                 "v00032",
                 "v00033",
                 "v40000",
                 "v00034",
                 "v00035",
                 "v00036",
                 "v00037",
                 "v00038",
                 "v00039",
                 "v00040",
                 "v00041",
                 "v00042",
                 "v50000"
               )
  )

  expect_equal(
    colnames(
      acc_test3(
        study_data = study_data,
        meta_data = meta_data,
        label_col = "LABEL"
      )
    ),
    c(
      "CENTER_0",
      "PSEUDO_ID",
      "SEX_0",
      "AGE_0",
      "AGE_GROUP_0",
      "AGE_1",
      "SEX_1",
      "PART_STUDY",
      "SBP_0",
      "DBP_0",
      "GLOBAL_HEALTH_VAS_0",
      "ASTHMA_0",
      "VO2_CAPCAT_0",
      "ARM_CIRC_0",
      "ARM_CIRC_DISC_0",
      "ARM_CUFF_0",
      "USR_VO2_0",
      "USR_BP_0",
      "EXAM_DT_0",
      "PART_PHYS_EXAM",
      "CRP_0",
      "BSG_0",
      "DEV_NO_0",
      "LAB_DT_0",
      "PART_LAB",
      "EDUCATION_0",
      "EDUCATION_1",
      "FAM_STAT_0",
      "MARRIED_0",
      "N_CHILD_0",
      "EATING_PREFS_0",
      "MEAT_CONS_0",
      "SMOKING_0",
      "SMOKE_SHOP_0",
      "N_INJURIES_0",
      "N_BIRTH_0",
      "INCOME_GROUP_0",
      "PREGNANT_0",
      "MEDICATION_0",
      "N_ATC_CODES_0",
      "USR_SOCDEM_0",
      "INT_DT_0",
      "PART_INTERVIEW",
      "ITEM_1_0",
      "ITEM_2_0",
      "ITEM_3_0",
      "ITEM_4_0",
      "ITEM_5_0",
      "ITEM_6_0",
      "ITEM_7_0",
      "ITEM_8_0",
      "QUEST_DT_0",
      "PART_QUESTIONNAIRE"
    )
  )

  expect_equal(
    colnames(
      acc_test3(
        study_data = study_data,
        meta_data = meta_data,
        label_col = LABEL
      )
    ),
    c(
      "CENTER_0",
      "PSEUDO_ID",
      "SEX_0",
      "AGE_0",
      "AGE_GROUP_0",
      "AGE_1",
      "SEX_1",
      "PART_STUDY",
      "SBP_0",
      "DBP_0",
      "GLOBAL_HEALTH_VAS_0",
      "ASTHMA_0",
      "VO2_CAPCAT_0",
      "ARM_CIRC_0",
      "ARM_CIRC_DISC_0",
      "ARM_CUFF_0",
      "USR_VO2_0",
      "USR_BP_0",
      "EXAM_DT_0",
      "PART_PHYS_EXAM",
      "CRP_0",
      "BSG_0",
      "DEV_NO_0",
      "LAB_DT_0",
      "PART_LAB",
      "EDUCATION_0",
      "EDUCATION_1",
      "FAM_STAT_0",
      "MARRIED_0",
      "N_CHILD_0",
      "EATING_PREFS_0",
      "MEAT_CONS_0",
      "SMOKING_0",
      "SMOKE_SHOP_0",
      "N_INJURIES_0",
      "N_BIRTH_0",
      "INCOME_GROUP_0",
      "PREGNANT_0",
      "MEDICATION_0",
      "N_ATC_CODES_0",
      "USR_SOCDEM_0",
      "INT_DT_0",
      "PART_INTERVIEW",
      "ITEM_1_0",
      "ITEM_2_0",
      "ITEM_3_0",
      "ITEM_4_0",
      "ITEM_5_0",
      "ITEM_6_0",
      "ITEM_7_0",
      "ITEM_8_0",
      "QUEST_DT_0",
      "PART_QUESTIONNAIRE"
    )
  )

  expect_equal(
    colnames(
      acc_test4(
        study_data = study_data,
        meta_data = meta_data,
        label_col = LABEL
      )
    ),
    c(
      "CENTER_0",
      "PSEUDO_ID",
      "SEX_0",
      "AGE_0",
      "AGE_GROUP_0",
      "AGE_1",
      "SEX_1",
      "PART_STUDY",
      "SBP_0",
      "DBP_0",
      "GLOBAL_HEALTH_VAS_0",
      "ASTHMA_0",
      "VO2_CAPCAT_0",
      "ARM_CIRC_0",
      "ARM_CIRC_DISC_0",
      "ARM_CUFF_0",
      "USR_VO2_0",
      "USR_BP_0",
      "EXAM_DT_0",
      "PART_PHYS_EXAM",
      "CRP_0",
      "BSG_0",
      "DEV_NO_0",
      "LAB_DT_0",
      "PART_LAB",
      "EDUCATION_0",
      "EDUCATION_1",
      "FAM_STAT_0",
      "MARRIED_0",
      "N_CHILD_0",
      "EATING_PREFS_0",
      "MEAT_CONS_0",
      "SMOKING_0",
      "SMOKE_SHOP_0",
      "N_INJURIES_0",
      "N_BIRTH_0",
      "INCOME_GROUP_0",
      "PREGNANT_0",
      "MEDICATION_0",
      "N_ATC_CODES_0",
      "USR_SOCDEM_0",
      "INT_DT_0",
      "PART_INTERVIEW",
      "ITEM_1_0",
      "ITEM_2_0",
      "ITEM_3_0",
      "ITEM_4_0",
      "ITEM_5_0",
      "ITEM_6_0",
      "ITEM_7_0",
      "ITEM_8_0",
      "QUEST_DT_0",
      "PART_QUESTIONNAIRE"
    )
  )
  expect_error(
    acc_test2(study_data = NULL, meta_data = meta_data),
    regexp = "Need study data as a data frame"
  )
  expect_error(
    prep_prepare_dataframes(.replace_missings = cars),
    regexp = paste(".replace_missings needs to be 1 logical value."),
    fixed = TRUE
  )

  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]
  err_test1 <- function(resp_variable, aux_variable,
                        time_variable, co_variables,
                        group_vars, study_data, meta_data, label_col) {
    prep_prepare_dataframes()
    invisible(ds1)
  }
  environment(err_test1) <- asNamespace("dataquieR")

  if (exists("testxx")) rm(testxx)
  expect_error(err_test1(study_data = cars, label_col = testxx),
               regexp = "Cannot resolve .+label_col = testxx.+",
               perl = TRUE)


  sd99 <- study_data
  err_test2 <- function(resp_variable, aux_variable,
                        time_variable, co_variables,
                        group_vars, study_data = sd99,
                        meta_data, label_col) {
    prep_prepare_dataframes()
    invisible(ds1)
  }
  environment(err_test2) <- asNamespace("dataquieR")

  sd99 <- 42

  expect_error(err_test2(meta_data = meta_data),
               regexp = "Need study data as a data frame")

  err_test3 <- function(resp_variable, aux_variable,
                        time_variable, co_variables,
                        group_vars,
                        meta_data, label_col) {
    prep_prepare_dataframes()
    invisible(ds1)
  }
  environment(err_test3) <- asNamespace("dataquieR")
  if (exists("study_data")) rm(study_data)
  expect_error(err_test3(meta_data = meta_data),
               regexp = "Need study data as a data frame")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  md99 <- meta_data
  err_test4 <- function(resp_variable, aux_variable,
                        time_variable, co_variables,
                        group_vars, study_data,
                        meta_data = md99, label_col) {
    prep_prepare_dataframes()
    invisible(ds1)
  }
  environment(err_test4) <- asNamespace("dataquieR")

  expect_silent(err_test4(study_data = study_data))
  md99 <- 42
  expect_warning(err_test4(study_data = study_data),
               regexp = "guess a preliminary")
  err_test5 <- function(resp_variable, aux_variable,
                        time_variable, co_variables,
                        group_vars, study_data,
                        meta_data, label_col) {
    prep_prepare_dataframes()
    invisible(ds1)
  }
  environment(err_test5) <- asNamespace("dataquieR")
  expect_warning(err_test5(study_data = study_data),
                 regexp = sprintf("(%s|%s|%s)",
                                  paste("Missing .+meta_data.+, try to guess a",
                                        "preliminary one from the data using",
                                        ".+prep_study2meta.+. Please",
                                        "consider amending this minimum guess",
                                        "manually."),
                                  paste("Metadata does not provide a filled",
                                        "column called .+JUMP_LIST.+ for",
                                        "replacing codes with NAs."),
                                  paste("Some code labels or -values",
                                        "are missing from .+meta_data.+.")
                 ),
                 perl = TRUE)


  study_data <- tibble::as_tibble(study_data)
  meta_data <- tibble::as_tibble(meta_data)
  acc_test5 <- function(resp_variable, aux_variable, time_variable,
                        co_variables, group_vars, study_data, meta_data,
                        label_col) {
    expect_true(tibble::is_tibble(study_data))
    expect_true(tibble::is_tibble(meta_data))
    prep_prepare_dataframes()
    expect_false(tibble::is_tibble(study_data))
    expect_false(tibble::is_tibble(meta_data))
    expect_false(tibble::is_tibble(ds1))
    invisible(ds1)
  }
  environment(acc_test5) <- asNamespace("dataquieR")

  x <- acc_test5(study_data = study_data,
                 meta_data = meta_data, label_col = LABEL)
  y <- acc_test5(study_data = tibble::as_tibble(x),
                 meta_data = meta_data, label_col = LABEL)
  expect_identical(x, y) # because this has already been mapped.
  md99 <- meta_data
  md99$VAR_NAMES <- NULL
  expect_error(
    y <- acc_test5(study_data = study_data,
                   meta_data = md99, label_col = LABEL),
    regexp = paste(
      "Missing columns.+VAR_NAMES.+from .+meta_data.*"),
    perl = TRUE
  )
  sd99 <- study_data
  md99 <- meta_data
  colnames(sd99) <- sprintf("VAR_%06d", seq_len(ncol(sd99)))
  withr::with_options(list(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact"),
                      expect_conditions(
      expect_error(
        acc_test3(study_data = sd99, meta_data = md99),
        regexp = "No data left. Aborting"
      ),
      regexps = c(
                 paste("Lost 100. of the study data because of",
                       "missing.not assignable metadata"),
                 paste("Lost 100. of the metadata because of",
                       "missing.not assignable study.data"),
                 paste("Did not find any metadata for the following",
                       "variables from the study data. .+VAR_"),
                 paste("Found metadata for the following variables not",
                       "found in the study data: .+v00000.+, .+v00001.+,",
                       ".+v00002.+, .+v00003.+, .+v00004.+")
      ),
      classes = c("warning", "warning", "message", "message")
  ))
})

#' Utility function observations in subgroups
#'
#' This function uses `!is.na` to count the number of non-missing observations in subgroups of
#' the data (list) and in a set of user defined response variables. In some applications it is
#' required that the number of observations per e.g. factor level is higher than a user-defined
#' minimum number.
#'
#' @param x data frame
#' @param rvs variable names
#'
#' @return matrix of flags
#'
#' @seealso [prep_min_obs_level]
#' @seealso `util_check_group_levels()`
#'
#' @family robustness_functions
#' @concept data_management
#' @noRd
util_observations_in_subgroups <- function(x, rvs) {
  if (dim(x)[1] == 0) {
    x[1, ] <- t(rep(NA, length(x)))
  }

  y <- !apply(x[, rvs, drop = FALSE], 2, is.na)


  if (!is.matrix(y)) {
    y <- as.matrix(t(y))
  }

  colnames(y) <- paste0("OBSERVATIONS_OUT_", rvs)
  x <- cbind(x, y)
}

test_that("util_observations_in_subgroups works", {
  skip_on_cran()
  dat <- cars[1:10, , FALSE]
  dat[1:3, 1] <- NA
  dat[2:6, 2] <- NA
  r <- util_observations_in_subgroups(dat, c("speed", "dist"))
  expect_equal(r[, "OBSERVATIONS_OUT_speed", TRUE], c(rep(FALSE, 3),
                                                      rep(TRUE, 7)))
  expect_equal(r[, "OBSERVATIONS_OUT_dist", TRUE], c(TRUE,
                                                     rep(FALSE, 5),
                                                     rep(TRUE, 4)))


  dat <- cars[FALSE, "speed", !TRUE]
  r <- util_observations_in_subgroups(dat, c("speed"))
  expect_equal(r[, "OBSERVATIONS_OUT_speed", TRUE], FALSE)

})

#' Get the dimensions of a `dq_report2` result
#'
#' @param x a `dataquieR_resultset2` result
#'
#' @return dimensions
#' @export
dim.dataquieR_resultset2 <- function(x) {
  stopifnot(inherits(x, "dataquieR_resultset2"))
  c(length(rownames(x)), length(colnames(x)), length(resnames(x)))
}

#' Find first externally called function in the stack trace
#'
#' intended use: error messages for the user
#'
#' @param sfs reverse [sys.frames] to search in
#' @param cls reverse [sys.calls] to search in
#'
#' @return reverse [sys.frames] index of first non-dataquieR function in
#'         this stack
#'
#' @family condition_functions
#' @noRd
util_find_first_externally_called_functions_in_stacktrace <-
  function(sfs = rev(sys.frames()), cls = rev(sys.calls())) {
  first <- util_find_external_functions_in_stacktrace(sfs, cls)
  if (length(first) == 0) {
    NA_integer_
  } else {
    head(first, 1) - 2
  }
}

test_that("util_find_first_externally_called_functions_in_stacktrace works", {
  skip_on_cran()
  test1 <- function() {
    util_find_first_externally_called_functions_in_stacktrace()
  }
  environment(test1) <- asNamespace("dataquieR")
  test2 <- function() {
    test1 <- function() {
      util_find_first_externally_called_functions_in_stacktrace()
    }
    environment(test1) <- asNamespace("dataquieR")
    test1()
  }
  expect(
    ok = test2() == 1,
    info = "before shifting test2 to dataquieR, result is 1",
    failure_message = "external call detection wrong?"
  )
  environment(test2) <- asNamespace("dataquieR")
  expect(
    ok = test2() == 2,
    info = "after shifting test2 to dataquieR, result should now be 2",
    failure_message = "shifted, but external call detection still the old?"
  )
  expect_equal(
    util_find_first_externally_called_functions_in_stacktrace(
      sfs = NULL,
      cls = NULL),
    NA_integer_
  )
})

#' Plot a `dataquieR` summary
#'
#' @param x the `dataquieR` summary, see [summary()] and [dq_report2()]
#' @param y not yet used
#' @param ... not yet used
#' @param filter if given, this filters the summary, e.g.,
#'         `filter = call_names == "com_qualified_item_missingness"`
#' @param dont_plot suppress the actual plotting, just return a printable
#'                   object derived from `x`
#' @param stratify_by column to stratify the summary, may be one string.
#' @param disable_plotly [logical] do not use `plotly`, even if installed
#' @param vars_to_include [character] study | `ssi`. variables are study
#'                                    variables or computed social science
#'                                    indicator variables.
#' @param hierarchy not yet defined, but if an argument is given, a
#'                  sunburst chart is displayed, currently, only `DQ_OBS`
#'                  can be used a the hierarchy.
#' @param folder_of_report a named vector with the location of variable and
#'                         `call_names`
#' @param var_uniquenames a data frame with the original variable names and the
#'                        unique names in case of reports created with dq_report_by
#'                        containing the same variable in several reports
#'                        (e.g., creation of reports by sex)
#'
#' @return invisible html object
#' @export
#'
plot.dataquieR_summary <- function(x, y, ..., filter, dont_plot = FALSE,
                                   stratify_by,
                                   vars_to_include = "study",
                                   disable_plotly = FALSE,
                                   hierarchy,
                                   folder_of_report = NULL,
                                   var_uniquenames = NULL) {
  if (!disable_plotly) {
    util_ensure_suggested(pkg = c("plotly"),
                          goal = "generate interactive HTML-summaries.")
  }

  util_stop_if_not("y is not used for plotting summaries" = missing(y))
  util_ensure_suggested(pkg = c("htmltools",
                                "DT", "rmarkdown",
                                "markdown"),
                        goal = "generate plain HTML-summaries.")

  x <- util_reclassify_dataquieR_summary(x)

  repsum <- x

  indicator_metric <- NULL
  function_name <- NULL
  util_expect_scalar(dont_plot, check_type = is.logical)
  if (missing(stratify_by)) {
    stratify_by <- character(0)
  } else {
    util_expect_scalar(stratify_by, check_type = is.character)
  }
  this <- attr(repsum, "this")

  suitable_vars_sum <- util_filter_repsum(this$result,
                                          vars_to_include,
                                          this$meta_data,
                                          this$rownames_of_report,
                                          this$label_col)
  rownames_of_report <- attr(suitable_vars_sum, "rownames_of_report")

  if (!disable_plotly && !missing(hierarchy)) {
    # TODO: Support other hierarchies
    # TODO: Support filter / stratify_by like in the other plot versions.
    rs <- repsum
    attr(rs, "this") <- rlang::env_clone(this)
    attr(rs, "this")$result <- suitable_vars_sum
    if (is.null(attr(rs,"this")$result) ||
        is.null(attr(rs,"this")$result) ||
        !prod(dim(attr(rs,"this")$result))) {
      return(htmltools::HTML(""))
    }
    return(util_render_sunburst_from_summary_classes(rs,
                                                     folder_of_report =
                                                       folder_of_report,
                                                     var_uniquenames =
                                                       var_uniquenames,
                                                     ...))
  }

  # suppressMessages(this$result %>%
  #                    dplyr::filter(!startsWith(as.character(indicator_metric), "CAT_")) %>%
  #                    dplyr::filter(!startsWith(as.character(indicator_metric), "MSG_")) %>%
  #                    dplyr::filter(!is.na(class)) %>%
  #                    dplyr::group_by(function_name, class) %>%
  #                    dplyr::summarise(value = length(function_name)) ->
  #                    summ_per_fkt_fig)

  suppressMessages(suitable_vars_sum %>%
                     dplyr::filter(!startsWith(as.character(indicator_metric), "CAT_")) %>%
                     dplyr::filter(!startsWith(as.character(indicator_metric), "MSG_")) ->
                     all_per_variable_all_issue_classes_except_errors)

  if (!missing(filter)) {
    cl <-
      rlang::call2(dplyr::filter,
                   .data = all_per_variable_all_issue_classes_except_errors,
                   substitute(filter))
    all_per_variable_all_issue_classes_except_errors <- eval(cl,
                                                             envir = parent.frame())
  }

  if (nrow(all_per_variable_all_issue_classes_except_errors) == 0) {
    return(htmltools::HTML(""))
  }

  if (!all(stratify_by %in%
           colnames(all_per_variable_all_issue_classes_except_errors))) {
    util_error("Cannot stratify summary by %s, I don't know, what %s are.",
               util_pretty_vector_string(setdiff(stratify_by,
                  colnames(all_per_variable_all_issue_classes_except_errors))),
               util_pretty_vector_string(setdiff(stratify_by,
                colnames(all_per_variable_all_issue_classes_except_errors))))
  }

  all_per_variable_all_issue_classes_except_errors %>%
    dplyr::filter(!is.na(value)) %>%
    # dplyr::filter(!is.na(class)) %>%
    # dplyr::group_by(VAR_NAMES, !!stratify_by) %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(c(VAR_NAMES, stratify_by)))) %>%
    dplyr::summarise(class =
                       suppressWarnings(
                         util_as_cat(max(util_as_cat(class), na.rm = TRUE)))) ->
    worst_per_variable
  plot_tab <- worst_per_variable

  plot_tab %>%
    # dplyr::group_by(class) %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(c("class", stratify_by)))) %>%
    dplyr::summarise(value = length(VAR_NAMES),
                     note = util_pretty_vector_string(
                       n_max = 3,
                       suppressWarnings(prep_get_labels(VAR_NAMES,
                                                        max_len = 15,
                                                        label_class = "SHORT",
                                                        meta_data =
                                                          this$meta_data)))) ->
    sum_plot_tab

  if (length(stratify_by) == 0) {
    sum_plot_tab %>%
      dplyr::mutate(X =
                      sprintf(
                        "%d %s: %d classified by indicators",
                        ifelse(vars_to_include == "study", length(rownames_of_report), all_per_variable_all_issue_classes_except_errors %>%
                                 dplyr::filter(!is.na(value)) %>% dplyr::select(VAR_NAMES) %>% unique %>% nrow),
                        ifelse(vars_to_include == "study", "variables", "scales"),
                        ifelse(vars_to_include == "study", length(unique(plot_tab$VAR_NAMES)), all_per_variable_all_issue_classes_except_errors %>%
                             dplyr::filter(!is.na(value) & !is.na(class)) %>% dplyr::select(VAR_NAMES) %>% unique %>% nrow)
                      ),
                    class = as.integer(class)) ->
      sum_plot_tab
  }
  # summ_per_fkt_fig %>%
  #   dplyr::mutate(dimension = c(int = "Integrity",
  #                               acc = "Accuracy",
  #                               con = "Consistency",
  #                               com = "Completeness")[
  #                                 substr(function_name, 1, 3)]) %>%
  #   dplyr::group_by(dimension, class) %>%
  #   dplyr::summarise(value = sum(value)) -> summ_per_dim_fig

  #  sum_worst_per_variable$note <-

  attr(sum_plot_tab, "vars_to_include") <- vars_to_include
#  if (length(unique(sum_plot_tab$X)) > 0) {
  if (!disable_plotly) {
    summaryplots <-
      prep_render_pie_chart_from_summaryclasses_plotly(
        sum_plot_tab,
        meta_data = this$meta_data
      )
  } else {
    summaryplots <-
      prep_render_pie_chart_from_summaryclasses_ggplot2(
        sum_plot_tab,
        meta_data = this$meta_data
      )
  }
  # } else {
  #   summaryplots <- htmltools::tags$p(sprintf(
  #     "None of the %d variables was classified", length(rownames_of_report)))
  # }

  if (!inherits(summaryplots, "htmlwidget") &&
      !inherits(summaryplots, "html") &&
      !inherits(summaryplots, "shiny.tag") &&
      !inherits(summaryplots, "shiny.tag.list")) {
    if (!all(vapply(summaryplots, inherits, "htmlwidget",
                    FUN.VALUE = logical(1)) |
             vapply(summaryplots, inherits, "html",
                    FUN.VALUE = logical(1)) |
             vapply(summaryplots, inherits, "shiny.tag",
                    FUN.VALUE = logical(1)) |
             vapply(summaryplots, inherits, "shiny.tag.list",
                    FUN.VALUE = logical(1)))) {
      util_error(c("Internal error: Not all summaryplots are html htmlwidgets",
                   "or shiny.tags / shiny.tag.lists. Sorry, and please report",
                   "this bug. Thank you."))
    }
    summaryplots <- do.call(htmltools::tagList, summaryplots)
  }

  r <- htmltools::browsable(summaryplots)

  if (!dont_plot) {
    print(r)
  } else {
   invisible(r)
  }
}

#' Classify metrics from a report summary table
#'
#' @param report_summary [list()] as returned by
#'   [prep_extract_summary()]
#'
#' @return [data.frame] classes for the report summary table, long format
#' @family summary_functions
#' @export
prep_summary_to_classes <- function(report_summary) {
  te <- topenv(parent.frame(1)) # see https://stackoverflow.com/a/27870803
  if (!(isNamespace(te) && getNamespaceName(te) == "dataquieR")) {
    lifecycle::deprecate_soft("2.1.0.9007",
                              "prep_summary_to_classes()")
  }

  if (!inherits(report_summary, "dq_report2_summary") &&
      all(c("Data", "Table", "meta_data") %in% names(report_summary))){
    util_error(c("%s must be returned by %s or this function",
                 "i.e. feature %s, %s and %s"),
               sQuote("report_summary"),
               sQuote("prep_extract_summary"),
               sQuote("Data"), sQuote("Table"), sQuote("meta_data"))
  }

  report_summary_table <- report_summary$Table
  meta_data <- report_summary$meta_data

  metrices_maybe_with_call_name <-
    setNames(nm = setdiff(colnames(report_summary_table),
                          c(STUDY_SEGMENT, VAR_NAMES)))

  metrices <- gsub("^.*\\.", "", metrices_maybe_with_call_name)

  have_at_least_some_call_names <-
    any(grepl(".", metrices_maybe_with_call_name, fixed = TRUE), na.rm = TRUE)

  all_thresholds <- lapply(setNames(nm = metrices), function(metric) {
    r <- util_get_thresholds(indicator_metric = metric, meta_data = meta_data)
    lapply(r, lapply, util_parse_interval) # TODO: Maybe NA?
  })

  all <- lapply(metrices_maybe_with_call_name,
                function(metric_maybe_with_call_name) {
                  metric <- metrices[[metric_maybe_with_call_name]]
                  thresholds <- all_thresholds[[metric]]
                  col <- setNames(as.numeric(report_summary_table[[metric_maybe_with_call_name]]),
                                  nm = report_summary_table[[VAR_NAMES]])

                  classes <- lapply(setNames(nm = report_summary_table[[VAR_NAMES]]),
                                    FUN = function(vn) {
                                      setNames(mapply(
                                        SIMPLIFY = FALSE,
                                        USE.NAMES = FALSE,
                                        clss = names(thresholds[[vn]]),
                                        int = thresholds[[vn]],
                                        FUN = function(clss, int) {
                                          if (is.na(col[vn]) || !inherits(int, "interval"))
                                            return(NA)
                                          redcap_env$`in`(col[vn], int)
                                        }), nm = names(thresholds[[vn]]))
                                    })

                  values <- report_summary$Data[[metric_maybe_with_call_name]]
                  values_raw <- report_summary$Table[[metric_maybe_with_call_name]]

                  classes <- lapply(classes, vapply, FUN.VALUE = logical(1), identity)
                  classes <- lapply(classes, which)
                  classes_nrs <- lapply(classes, unname)
                  classes_nrs <- unlist(classes_nrs)
                  classes <- lapply(classes, names)
                  missing <- vapply(classes, length, FUN.VALUE = integer(1)) ==
                    0
                  classes[missing] <- NA
                  classes <- unlist(classes) # TODO: split it up here, return this first, then run the aggregation on it later
                  data.frame(class = classes,
                             VAR_NAMES = names(classes),
                             indicator_metric = rep(metric_maybe_with_call_name, length(classes)),
                             value = values,
                             values_raw = values_raw,
                             n_classes =
                               vapply(FUN.VALUE = integer(1),
                                      names(classes),
                                      function(vn) {
                                        r <- suppressWarnings(max(as.integer(names(thresholds[[vn]])),
                                            na.rm = TRUE))
                                        if (!is.finite(r)) {
                                          NA_integer_
                                        } else {
                                          r
                                        }
                                      })
                  )
                })
  all <- util_rbind(data_frames_list = all)
  rownames(all) <- NULL
  if (!prod(dim(all))) {
    all <- report_summary_table[, c(VAR_NAMES, STUDY_SEGMENT)]
  } else {
    all <- merge(all, report_summary_table[, c(VAR_NAMES, STUDY_SEGMENT)], by = VAR_NAMES, all.x = TRUE)
    if (have_at_least_some_call_names) {
      splitted <- all$indicator_metric # not yet splitted
      splitted[util_empty(splitted)] <- ""
      splitted <- strsplit(splitted, ".", fixed = TRUE)
      tuple_sizes <- vapply(splitted, length, FUN.VALUE = integer(1)) # should be always 1 or 2 according to function name and result column conventions
      valid_splitted <- tuple_sizes %in% 1:2
      if (!all(valid_splitted)) {
        util_warning(c("For some reason, the result names in this summary",
                       "feature columns with more than one . or empty names,",
                       "I'll ignore these. Internal error, please report"))
      }
      call_names <- rep("", nrow(all))
      indicator_metric <- rep("", nrow(all))
      call_names[tuple_sizes == 2] <-
        vapply(splitted[tuple_sizes == 2], `[[`, 1, FUN.VALUE = character(1))
      indicator_metric[tuple_sizes == 2] <-
        vapply(splitted[tuple_sizes == 2], `[[`, 2, FUN.VALUE = character(1))
      indicator_metric[tuple_sizes == 1] <-
        vapply(splitted[tuple_sizes == 1], `[[`, 1, FUN.VALUE = character(1))
      all$indicator_metric <- indicator_metric
      all$call_names <- call_names
    } else {
      all$call_names <- ""
    }
  }
  class(all) <- c("dq_report2_summaryclasses", "data.frame")
  return(all)
}


util_metrics_to_classes <- function(rs_table_long, meta_data, entity = "ITEM") {
  used_metrics <- setNames(nm = unique(rs_table_long[["indicator_metric"]]))

  all_thresholds <- lapply(used_metrics,
                           function(metric) {
    r <- util_get_thresholds(indicator_metric = metric, meta_data = meta_data)
    lapply(r, lapply, util_parse_interval)
  })

  all <- lapply(used_metrics,
                function(metric) {
                  curr_metric_rows <- rs_table_long$indicator_metric ==
                    metric
                  if (startsWith(metric, "MSG_")) {
                    res <- rs_table_long[curr_metric_rows, , FALSE]
                    res$class <- res$values_raw
                    return(res)
                  }
                  rs_table_long[curr_metric_rows, "values_raw"][rs_table_long[
                    curr_metric_rows, "values_raw"] == "TRUE"] <- 1
                  rs_table_long[curr_metric_rows, "values_raw"][rs_table_long[
                    curr_metric_rows, "values_raw"] == "FALSE"] <- 0
                  col <- suppressWarnings(as.numeric(
                    rs_table_long[curr_metric_rows, "values_raw"]))
                  if (entity == "ITEM") {
                    names(col) <- paste0(rs_table_long[curr_metric_rows, "call_names"],
                                         ".",
                                         rs_table_long[curr_metric_rows, VAR_NAMES])
                  } else { # TODO: Support other entities better by metadata model
                    names(col) <- paste0(rs_table_long[curr_metric_rows, "call_names"],
                                         seq_len(length(col))) # FIXME: so far, use first variables metric, which is not correct
                  }
                  thresholds <- all_thresholds[[metric]]

                  classes <- lapply(setNames(nm = names(col)),
                                    FUN = function(cn.vn) {
                                      if (entity == "ITEM") {
                                        vn <- util_sub_string_right_from_.(cn.vn)
                                      } else {
                                        vn <- names(thresholds)[[1]] # FIXME: so far, use first variables metric, which is not correct
                                      }
                                      cl_res <- setNames(mapply(
                                        SIMPLIFY = FALSE,
                                        USE.NAMES = FALSE,
                                        clss = names(thresholds[[vn]]),
                                        int = thresholds[[vn]],
                                        FUN = function(clss, int) {
                                          if (is.na(col[cn.vn]) || !inherits(int, "interval"))
                                            return(NA)
                                          valraw <- col[cn.vn]
                                          redcap_env$`in`(valraw, int)
                                        }), nm = names(thresholds[[vn]]))
                                      # if (metric == "FLG_acc_ud_prop" &&
                                      #     cn.vn == "acc_distributions_prop.v01002") browser()
                                      return(cl_res)
                                    })
# if (metric == "FLG_acc_ud_prop") browser()
                  classes <- lapply(classes, vapply, FUN.VALUE = logical(1),
                                    identity)
                  classes <- lapply(classes, which)
                  if (
                    any(vapply(classes, length, FUN.VALUE = integer(1))) > 1) {
                    util_warning(
                      c("Detected invalid grading rule with",
                        "overlapping intervals. Choosing the worst class."),
                      applicability_problem = TRUE
                      )
                    classes <- lapply(classes, tail, 1)
                  }
                  classes_nrs <- lapply(classes, unname)
                  classes_nrs <- unlist(classes_nrs)
                  classes <- lapply(classes, names)
                  missing <- vapply(classes, length, FUN.VALUE = integer(1)) ==
                    0
                  classes[missing] <- NA
                  classes <- unlist(classes) # TODO: split it up here, return this first, then run the aggregation on it later
                  res <- rs_table_long[curr_metric_rows, , FALSE]
                  res$class <- as.character(util_as_integer_cat(
                    util_as_cat(classes)))
                  res$n_classes <- vapply(FUN.VALUE = integer(1),
                                          names(classes),
                                          function(vn) {
                                            r <- suppressWarnings(max(as.integer(names(thresholds[[vn]])),
                                                                      na.rm = TRUE))
                                            if (!is.finite(r)) {
                                              NA_integer_
                                            } else {
                                              r
                                            }
                                          })
                  res
                })
  all <- util_rbind(data_frames_list = all)
  rownames(all) <- NULL
  # fill up all missing metrics, if at least once, a function returned them
  metrics_known_to_be_in_functions <- tapply(rs_table_long,
                                             rs_table_long$function_name,
                                             function(x) {
                                               r <- unique(x$indicator_metric)
                                               r[!startsWith(r, "CAT_") &
                                                   ! startsWith(r, "MSG_")]
                                              },
                                             simplify = FALSE)

  what_i_need <- unique(all[, setdiff(colnames(all),
                                 c("indicator_metric", "n_classes",
                                   "class", "value", "values_raw"))])

  metrics_known_to_be_in_functions2 <- util_rbind(data_frames_list =
               lapply(names(metrics_known_to_be_in_functions),
                      function(function_name) {
    cbind.data.frame(
      function_name = rep(function_name,
                          length(metrics_known_to_be_in_functions[[function_name]])),
      indicator_metric =
        metrics_known_to_be_in_functions[[function_name]],
      stringsAsFactors = FALSE)
                        }
    ))

  if (!prod(dim(metrics_known_to_be_in_functions2))) {
    what_i_really_need <-
      what_i_need
  } else {
    what_i_really_need <-
      merge(what_i_need, metrics_known_to_be_in_functions2, by = "function_name")
  }

  to_add <- dplyr::anti_join(
    what_i_really_need,
    all[, colnames(what_i_really_need), drop = FALSE],
    by = colnames(what_i_really_need)
  )

  if (nrow(to_add) > 0) {
    to_add[, setdiff(colnames(all),
                     colnames(what_i_really_need))] <- rep(NA, nrow(to_add))

    to_add <- to_add[, colnames(all), drop = FALSE]

    all <- util_rbind(all, to_add)
  }

  class(all) <- c("dq_report2_summaryclasses", "data.frame")
  return(all)
}

test_that("prep_render_pie_chart_from_summaryclasses_ggplot2 works", {
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_on_cran()
  skip_if_not_installed("stringdist")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  sd0 <- study_data[, 1:5]
  sd0$v00012 <- study_data$v00012
  md0 <- subset(meta_data, VAR_NAMES %in% colnames(sd0))
  md0$PART_VAR <- NULL

  report <- dq_report2(sd0, md0,
                       resp_vars = c("v00000", "v00001", "v00002",
                                     "v00003", "v00004", "v00012"),
                       filter_indicator_functions =
                         c("^com_item_missingness$",
                           "^acc_varcomp$"),
                       filter_result_slots =
                         c("^SummaryTable$"),
                       cores = NULL,
                       dimensions = # for speed, omit Accuracy
                         c("Integrity",
                           "Completeness",
                           "Consistency",
                           "Accuracy"))

  suppressWarnings({
    summary <- prep_extract_summary(report)
    classes <- prep_summary_to_classes(summary)
  })

  plot_tab <- classes %>%
    dplyr::filter(!is.na(value)) %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(c(VAR_NAMES, "call_names")))) %>%
    dplyr::summarise(class =
                       suppressWarnings(
                         util_as_cat(max(util_as_cat(class), na.rm = TRUE))))

  suppressWarnings(sum_plot_tab <- plot_tab %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(c("class", "call_names")))) %>%
    dplyr::summarise(value = length(VAR_NAMES),
                     note = util_pretty_vector_string(
                       n_max = 3,
                       suppressWarnings(prep_get_labels(VAR_NAMES,
                                                        max_len = 15,
                                                        label_class = "SHORT",
                                                        meta_data =
                                                          this$meta_data)))))



  suppressWarnings(
    r <- prep_render_pie_chart_from_summaryclasses_ggplot2(sum_plot_tab,
                                                      meta_data = md0))
  r <- as.character(r)
  # expect_snapshot_value(r, style = "deparse") snapshots do not work, here. the plots differ (like the label is on the test platform on top locally at the bottom of the big pie-piece)
  # do a very unspecific verification:

  expect_gt(nchar(r), 20000)

})

#' `Roxygen`-Template for indicator functions
#'
#' @param resp_vars [variable] the names of the measurement variables, if
#'                             missing or `NULL`, all variables will be checked
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#' @param meta_data_dataframe [data.frame] the data frame that contains the
#'                                          metadata for the data frame level
#' @param meta_data_segment [data.frame] -- optional: Segment level metadata
#' @param dataframe_level [data.frame] alias for `meta_data_dataframe`
#' @param segment_level [data.frame] alias for `meta_data_segment`
#' @return `invisible(NULL)`
#' @keywords internal
.template_function_indicator <-
  function(resp_vars, study_data, label_col, item_level, meta_data,
           meta_data_v2,
           meta_data_dataframe, meta_data_segment, dataframe_level,
           segment_level) { # nocov start
    util_error("nothing, just a template for Roxygen")
    invisible(NULL)
} # nocov end

#' an environment with functions available for `REDcap` rules
#'
#' @return environment
#'
#' @family redcap
#' @concept metadata_management
#' @noRd
util_get_redcap_rule_env <- function() {
  redcap_env
}

#' Utility function to parse dates/date-times
#'
#' @param dt a date vector, usually as a [character()], anything should work
#' @param tryFormats see [as.POSIXct()]
#' @param optional see [as.POSIXct()]
#' @param tz see [as.POSIXct()]
#'
#' @return the parsed `dt`
#'
#' @family parser_functions
#' @concept metadata_management
#' @noRd
util_parse_date <- function(dt, tryFormats, optional = TRUE, tz) {
  # this function is needed, here, so not in an own file
  if (is.numeric(dt)) {
    dt <- as.POSIXct(dt, origin = lubridate::origin, optional = optional)
  }
  if (missing(tryFormats)) {
    if (is.character(dt)) {
      r <- suppressWarnings(lubridate::as_datetime(dt))
      not_conv <- !is.na(dt) & is.na(r)
      if (any(not_conv)) {
        r[not_conv] <- suppressWarnings(as.POSIXct(dt[not_conv],
                                                   optional = optional,
                                                   tryFormats =
                                                     c("%Y-%m-%dT%H:%M:%OS",
                                                       "%Y/%m/%dT%H:%M:%OS",
                                                       "%Y-%m-%dT%H:%M",
                                                       "%Y/%m/%dT%H:%M",
                                                       "%Y-%m-%d %H:%M:%OS",
                                                       "%Y/%m/%d %H:%M:%OS",
                                                       "%Y-%m-%d %H:%M",
                                                       "%Y/%m/%d %H:%M",
                                                       "%Y-%m-%d",
                                                       "%Y/%m/%d")))
      }
      # r <- readr::parse_datetime(dt, %Z) too strict regarding timezones
    } else {
      r <- suppressWarnings(lubridate::as_datetime(dt))
    }
    if (!optional &&
        !all(is.na(r) == is.na(dt))) {
      util_error("Invalid date/time: %s",
                 util_pretty_vector_string(
                   n_max = 5,
                   dt[is.na(r) != is.na(dt)]
                 ))
    }
  } else {
    r <- suppressWarnings(as.POSIXct(dt,
                                     tryFormats = tryFormats,
                                     optional = optional))
  }
  if (inherits(r, "POSIXct")) {
    r <- as.POSIXct(round(r, units = "secs"))
  }
  if (missing(tz)) {
    r <- lubridate::with_tz(lubridate::force_tz(r))
  } else {
    r <- lubridate::with_tz(lubridate::force_tz(r, tzone = tz))
  }
  r
}

#' Utility function to parse times
#'
#' @param to a time-only vector, usually as a [character()], anything should
#'           work
#' @param tryFormats Character vector of time formats tried in order,
#'                   similar to [as.POSIXct()]. Defaults to "%H:%M:%S",
#'                   "%H:%M", "%H%M%S", "%H%M".
#' @param optional If FALSE, stop on un-parseable entries; if TRUE, set to NA.
#' @param tz Optional timezone; if missing, behaves like [lubridate::force_tz()].
#'
#' @return An [hms::hms] vector.
#'
#' @noRd
#' @family parser_functions
#' @concept metadata_management
util_parse_time <- function(
    to,
    tryFormats = c("%H:%M:%S", "%H:%M", "%H%M%S", "%H%M"),
    optional = TRUE,
    tz
) {
  # --------------------------------------------------------------------------
  # Early exit for already correct classes
  # --------------------------------------------------------------------------
  if (inherits(to, "hms")) return(to)
  if (inherits(to, "difftime")) return(hms::as_hms(to))
  if (inherits(to, "POSIXt"))  return(hms::as_hms(to))

  # --------------------------------------------------------------------------
  # Handle numeric input
  # --------------------------------------------------------------------------
  if (is.numeric(to)) {
    if (all(is.finite(to)) && all(to >= 0 & to < 1, na.rm = TRUE)) {
      # Excel-style fraction of a day
      return(hms::as_hms(to * 24 * 3600))
    }

    # Try HHMMSS / HHMM forms if integers
    to_chr <- sprintf("%06d", as.integer(to))
    suppressWarnings({
      trial <- tryCatch(
        as.difftime(to_chr, format = "%H%M%S", units = "secs"),
        error = function(e) NULL
      )
    })
    if (!is.null(trial) && any(!is.na(trial))) {
      return(hms::as_hms(trial))
    }

    # Fallback: interpret as seconds since epoch
    if (missing(tz)) {
      to <- as.POSIXct(to, origin = lubridate::origin)
      to <- lubridate::force_tz(to, tzone = lubridate::tz(to))
    } else {
      to <- as.POSIXct(to, origin = lubridate::origin, tz = tz)
      to <- lubridate::force_tz(to, tzone = tz)
    }
    return(hms::as_hms(to))
  }

  if (is.list(to)) {
    return(lapply(to, util_parse_time,
                  tryFormats = tryFormats,
                  optional = optional,
                  tz = tz))
  }

  # --------------------------------------------------------------------------
  # Character input normalization
  # --------------------------------------------------------------------------
  if (is.factor(to)) to <- as.character(to)
  to_chr <- as.character(to)
  to_chr[trimws(to_chr) == ""] <- NA_character_

  # Detect and reject date-time patterns
  has_date_pattern <- grepl("\\d{4}[-/]\\d{2}[-/]\\d{2}", to_chr, ignore.case = TRUE)
  if (any(has_date_pattern, na.rm = TRUE)) {
    bad <- unique(to_chr[has_date_pattern])
    util_warning(
      "Detected full date-time values; these are not pure times. Examples: %s",
      paste(utils::head(bad, 3L), collapse = ", ")
    )
    to_chr[has_date_pattern] <- NA_character_
  }

  n <- length(to_chr)
  parsed <- as.difftime(rep(NA_real_, n), units = "secs")

  # --------------------------------------------------------------------------
  # Try parsing with multiple formats
  # --------------------------------------------------------------------------
  for (fmt in tryFormats) {
    idx <- which(is.na(parsed))
    if (length(idx) == 0L) break
    suppressWarnings({
      trial <- tryCatch(
        as.difftime(to_chr[idx], format = fmt, units = "secs"),
        error = function(e) rep(as.difftime(NA_real_, units = "secs"), length(idx))
      )
    })
    ok <- !is.na(trial)
    if (any(ok)) parsed[idx[ok]] <- trial[ok]
  }

  # --------------------------------------------------------------------------
  # Warn or error for failed parsing
  # --------------------------------------------------------------------------
  if (anyNA(parsed)) {
    bad <- unique(to_chr[is.na(parsed) & !is.na(to_chr)])
    if (length(bad)) {
      msg <- sprintf(
        "Failed to parse %d/%d values. Examples: %s",
        sum(is.na(parsed)), n,
        paste(utils::head(bad, 3L), collapse = ", ")
      )
      if (!optional) stop(msg, call. = FALSE)
      else util_warning("%s", msg)
    }
  }

  # --------------------------------------------------------------------------
  # Return final hms vector
  # --------------------------------------------------------------------------
  hms::as_hms(parsed)
}


redcap_env <- (function() {
  penv <- new.env(parent = emptyenv())
# TODO: strcat?

  ### generic handling of redcap NAs ####
  # see gen_op below

  R_na2redcap_na <- function(x) {
    if (!is.vector(x)) {
      util_error("no complex arguemnts supported, esp. no %s",
                 dQuote(class(x)),
                 applicability_problem = TRUE)
    }
    if (is.factor(x) && !("" %in% levels(x))) {
      levels(x) <- c(levels(x), "")
    }
    if (!prep_dq_data_type_of(x) %in% c(DATA_TYPES$DATETIME,
                                        DATA_TYPES$TIME
                                        )) { # TODO datetime handling may be read as character? see also tag: 1893839
      if (any(is.na(x))) x[is.na(x)] <- ""
    }
    x
  }

  redcap_na2R_na <- function(x) {
    orig_dim <- dim(x)
    orig_dimnames <- dimnames(x)
    if (!is.atomic(x)) {
      util_error("no complex arguemnts supported, esp. no %s",
                 dQuote(class(x)),
                 applicability_problem = TRUE)
    }
    if (is.factor(x) && !("" %in% levels(x))) {
      levels(x) <- c(levels(x), "")
    }
    # if (prep_dq_data_type_of(x) == DATA_TYPES$DATETIME) {
    #   res <- x
    #   # resrna <- is.na(res)
    #   # if (identical(op, .Primitive("!="))) { # TODO: generalize
    #   #   if (any(resrna)) {
    #   #     res[resrna] <- TRUE
    #   #     res[resrna & is.na(x) & is.na(y)] <- FALSE
    #   #   }
    #   # } else if (identical(op, .Primitive("=="))) {
    #   #   if (any(resrna)) {
    #   #     res[resrna] <- FALSE
    #   #     res[resrna & is.na(x) & is.na(y)] <- TRUE
    #   #   }
    #   # } else {
    #   #   if (any(resrna)) res[resrna] <- ""
    #   # }
    # } else {
    #   xrcna <- is.na(x) | x == ""
    #   if (any(xrcna)) x[is.na(x)] <- ""
    #
    # }
    if (!prep_dq_data_type_of(x) %in% c(DATA_TYPES$DATETIME,
                                        DATA_TYPES$TIME)) { # TODO datetime handling may be read as character? see also tag: 1893839
      x_na <- is.na(x)
      if (any(x_na)) {
        if (is.factor(x)) {
          x[x_na] <- ""
        } else {
          x <- as.character(x)
          x[x_na] <- ""
        }
      }
    }
    if (!is.null(orig_dim)) {
      dim(x) <- orig_dim
      dimnames(x) <- orig_dimnames
    }
    x
  }

  decorate_fkt_redcap_na <- function(fkt) {
    function(...) {
      args <- list(...)
      args <- lapply(args, R_na2redcap_na)
      r <- do.call(fkt, args)
      redcap_na2R_na(r)
    }
  }

  gen_op <- function(op) {
    # support REDcap NAs in operators; see decorate_fkt_redcap_na
    function(x, y) {
#      util_warning("Used expected thing")
#      util_warning("op = %s", util_deparse1(op))
      # if (identical(op, `==`)) browser()
      if (identical(op, `&`) ||
          identical(op, `|`)
          ) {
        x <- suppressWarnings(as.logical(x))
        y <- suppressWarnings(as.logical(y))
      }
      if (is.factor(x) && !("" %in% levels(x))) {
        levels(x) <- c(levels(x), "")
      }
      if (is.factor(y) && !("" %in% levels(y))) {
        levels(y) <- c(levels(y), "")
      }
#      util_warning("dt(x) = %s", prep_dq_data_type_of(x))
#      util_warning("dt(y) = %s", prep_dq_data_type_of(y))
      if (prep_dq_data_type_of(x) %in% c(DATA_TYPES$DATETIME,
                                         DATA_TYPES$TIME) &&
          prep_dq_data_type_of(y) == DATA_TYPES$STRING) {
        yrcna <- y == ""
        if (any(yrcna)) y[[yrcna]] <- NA
      }
      if (prep_dq_data_type_of(x) == DATA_TYPES$STRING &&
          prep_dq_data_type_of(y) %in% c(DATA_TYPES$DATETIME,
                                         DATA_TYPES$TIME)) {
        xrcna <- x == ""
        if (any(xrcna)) x[[x == ""]] <- NA
      }
      if (prep_dq_data_type_of(x) == DATA_TYPES$DATETIME ||
          prep_dq_data_type_of(y) == DATA_TYPES$DATETIME) {
        xx <- try(util_parse_date(x, optional = FALSE), silent = TRUE)
        yy <- try(util_parse_date(y, optional = FALSE), silent = TRUE)
        if (util_is_try_error(xx)) {
          util_error("Rule evaluation: Could not interpret %s as %s",
                     util_pretty_vector_string(x), sQuote(DATA_TYPES$DATETIME))
        }
        if (util_is_try_error(yy)) {
          util_error("Rule evaluation: Could not interpret %s as %s",
                     util_pretty_vector_string(y), sQuote(DATA_TYPES$DATETIME))
        }
        res <- op(xx, yy)
        resrna <- is.na(res)
        if (identical(op, .Primitive("!="))) {
          if (any(resrna)) {
            res[resrna] <- TRUE
            res[resrna & is.na(x) & is.na(y)] <- FALSE
          }
        } else if (identical(op, .Primitive("=="))) {
          if (any(resrna)) {
            res[resrna] <- FALSE
            res[resrna & is.na(x) & is.na(y)] <- TRUE
          }
        } else {
          if (!inherits(res, "difftime") && any(resrna)) res[resrna] <- ""
        }
      } else if (prep_dq_data_type_of(x) == DATA_TYPES$TIME ||
                 prep_dq_data_type_of(y) == DATA_TYPES$TIME) {
        xx <- try(util_parse_time(x, optional = FALSE), silent = TRUE)
        yy <- try(util_parse_time(y, optional = FALSE), silent = TRUE)
        if (util_is_try_error(xx)) {
          util_error("Rule evaluation: Could not interpret %s as %s",
                     util_pretty_vector_string(x), sQuote(DATA_TYPES$TIME))
        }
        if (util_is_try_error(yy)) {
          util_error("Rule evaluation: Could not interpret %s as %s",
                     util_pretty_vector_string(y), sQuote(DATA_TYPES$TIME))
        }
        res <- op(xx, yy)
        resrna <- is.na(res)
        if (identical(op, .Primitive("!="))) {
          if (any(resrna)) {
            res[resrna] <- TRUE
            res[resrna & is.na(x) & is.na(y)] <- FALSE
          }
        } else if (identical(op, .Primitive("=="))) {
          if (any(resrna)) {
            res[resrna] <- FALSE
            res[resrna & is.na(x) & is.na(y)] <- TRUE
          }
        } else {
          if (any(resrna)) res[resrna] <- ""
        }
      } else {
        xrcna <- util_empty(x)
        yrcna <- util_empty(y)
        if (is.logical(x) && is.logical(y)) {
          res <- op(x, y)
          if (any(is.na(res)))
            res[is.na(res)] <- ""
        } else if (is.numeric(x) && is.numeric(y)) {
          res <- op(x, y)
          if (any(is.na(res)))
            res[is.na(res)] <- ""
        } else if (prep_dq_data_type_of(x) == DATA_TYPES$DATETIME &&
                   prep_dq_data_type_of(y) == DATA_TYPES$DATETIME) {
          xx <- try(util_parse_date(x, optional = FALSE), silent = TRUE)
          yy <- try(util_parse_date(y, optional = FALSE), silent = TRUE)
          if (util_is_try_error(xx)) {
            util_error("Rule evaluation: Could not interpret %s as %s",
                       util_pretty_vector_string(x), sQuote(DATA_TYPES$DATETIME))
          }
          if (util_is_try_error(yy)) {
            util_error("Rule evaluation: Could not interpret %s as %s",
                       util_pretty_vector_string(y), sQuote(DATA_TYPES$DATETIME))
          }
          res <- op(xx, yy)
          if (any(is.na(res)))
            res[is.na(res)] <- ""
        } else if (prep_dq_data_type_of(x) == DATA_TYPES$TIME &&
                   prep_dq_data_type_of(y) == DATA_TYPES$TIME) {
          xx <- try(util_parse_time(x, optional = FALSE), silent = TRUE)
          yy <- try(util_parse_time(y, optional = FALSE), silent = TRUE)
          if (util_is_try_error(xx)) {
            util_error("Rule evaluation: Could not interpret %s as %s",
                       util_pretty_vector_string(x), sQuote(DATA_TYPES$TIME))
          }
          if (util_is_try_error(yy)) {
            util_error("Rule evaluation: Could not interpret %s as %s",
                       util_pretty_vector_string(y), sQuote(DATA_TYPES$TIME))
          }
          res <- op(xx, yy)
          if (any(is.na(res)))
            res[is.na(res)] <- ""
        } else {
          if (any(xrcna)) x[is.na(x)] <- ""
          if (any(yrcna)) y[is.na(y)] <- ""
          res <- try(op(x, y), silent = TRUE)
          if (inherits(res, "try-error") ||
              (is.factor(x) && !is.factor(y)) ||
              (!is.factor(x) && is.factor(y))) {
            .x <- trimws(as.character(x))
            .x[is.na(x)] <- ""
            .y <- trimws(as.character(y))
            .y[is.na(y)] <- ""
            res <- try(op(as.character(.x), as.character(.y)), silent = TRUE)
            if (inherits(res, "try-error")) {
              res <- try(op(readr::parse_guess(.x), readr::parse_guess(.y)),
                         silent = TRUE)
            }
          }
          res
        }
      }
      if (inherits(res, "difftime")) {
        #TODO: check redCap NA handling
        res <- as.numeric(res, units = "secs")
        res[is.na(res)] <- ""
      }
      res
    }
  }

  ### opreators/functions ####

  # avoid operator precedence by R
  penv[["or"]] <- gen_op(base::`|`)
  attr(penv[["or"]], "prio") <- -2
  penv[["and"]] <- gen_op(base::`&`)
  attr(penv[["and"]], "prio") <- -1

  penv[["("]] <- base::`(`


  penv[["**"]] <- gen_op(base::`^`) #  must not be in prio=3 if prio 3 has no order
  attr(penv[["**"]], "order") <- -1
  attr(penv[["**"]], "prio") <- 3

  penv[["^"]] <- gen_op(base::`^`)
  attr(penv[["^"]], "prio") <- 3

  penv[["+"]] <- gen_op(base::`+`)
  attr(penv[["+"]], "prio") <- 1
  penv[["-"]] <- gen_op(base::`-`)
  attr(penv[["-"]], "prio") <- 1
  penv[["*"]] <- gen_op(base::`*`)
  attr(penv[["*"]], "prio") <- 2
  penv[["/"]] <- gen_op(base::`/`)
  attr(penv[["/"]], "prio") <- 2

  penv[["prod"]] <- decorate_fkt_redcap_na(base::`prod`)
  penv[["sum"]] <- decorate_fkt_redcap_na(base::`sum`)

  make_num <- function(number) {
    orig_dim <- dim(number)
    orig_dimnames <- dimnames(number)
    was_posixct <- inherits(number, "POSIXct")
    tzone <- attr(number, "tzone")
    number[trimws(number) == ""] <- NA
    numnumber <- as.numeric(number)
    if (suppressWarnings((!all(is.na(numnumber) == is.na(number)))))
      util_error("%s not numeric",
                 dQuote(util_deparse1(substitute(number))))
    if (was_posixct) {
      numnumber <- as.POSIXct(numnumber, origin = "1970-01-01", tz = tzone)
    }
    if (!is.null(orig_dim)) {
      dim(numnumber) <- orig_dim
      dimnames(numnumber) <- orig_dimnames
    }
    numnumber
  }

  get_input_as_num_matrix <- function(...) {
    args <- list(...)

    # Check if all inputs are POSIXct
    all_posixct <- all(vapply(args, inherits, logical(1), "POSIXct"))

    # Harmonize time zones if all are POSIXct
    if (all_posixct) {
      tzones <- vapply(args, function(x) attr(x, "tzone") %||% "", character(1))
      tzones <- tzones[tzones != ""]
      if (length(unique(tzones)) > 1) {
        warning("Different time zones found; coercing all to first non-empty one.")
      }
      common_tzone <- if (length(tzones) > 0) tzones[[1]] else "UTC"

      # Convert all elements to common time zone
      args <- lapply(args, function(x) as.POSIXct(x, tz = common_tzone))
    }

    # Manually cbind, set class later
    x <- do.call(cbind, args)

    # Restore POSIXct class if applicable
    if (all_posixct) {
      class(x) <- c("POSIXct", "POSIXt", "matrix")
      attr(x, "tzone") <- common_tzone
    }

    # Convert to numeric (or retain POSIXct structure)
    x <- make_num(x)

    x
  }

  # All functions from ####
  # https://docs.google.com/document/d/1l3nGBgqqPKi5PtMe75g7q0dny8QzGMd_/edit?tab=t.0

  penv[["round"]] <- function(number, decimal_places = 0) {
    numnumber <- make_num(number)
    decimal_places <- make_num(decimal_places)
    redcap_na2R_na(round(numnumber, decimal_places))
  }
  attr(penv[["round"]], "order") <- 99

  penv[["roundup"]] <- function(number, decimal_places = 0) {
    number <- make_num(number)
    decimal_places <- make_num(decimal_places)
    redcap_na2R_na(ceiling(number * (10^(decimal_places))) / (10^(decimal_places)))
  }

  penv[["rounddown"]] <- function(number, decimal_places = 0) {
    number <- make_num(number)
    decimal_places <- make_num(decimal_places)
    redcap_na2R_na(floor(number * (10^(decimal_places))) / (10^(decimal_places)))
  }

  penv[["sqrt"]] <- function(number) {
    number <- make_num(number)
    redcap_na2R_na(sqrt(number))
  }

  penv[["abs"]] <- function(number) {
    number <- make_num(number)
    redcap_na2R_na(abs(number))
  }

  penv[["min"]] <- function(...) {
    x <- make_num(c(...))
    redcap_na2R_na(min(x, na.rm = TRUE))
  }

  penv[["max"]] <- function(...) {
    x <- make_num(c(...))
    redcap_na2R_na(max(x, na.rm = TRUE))
  }

  penv[["mean"]] <- function(...) {
    x <- make_num(c(...))
    redcap_na2R_na(mean(x, na.rm = TRUE))
  }

  penv[["rowMin"]] <- function(...) {
    x <- get_input_as_num_matrix(...)
    x <- suppressWarnings(apply(x, 1, min, na.rm = TRUE))
    if (inherits(x, "numeric") && all(vapply(list(...), inherits, logical(1), "POSIXct"))) {
      x <- as.POSIXct(x, origin = "1970-01-01", tz = attr(list(...)[[1]], "tzone"))
    }
    redcap_na2R_na(x)
  }

  penv[["rowMax"]] <- function(...) {
    x <- get_input_as_num_matrix(...)
    x <- suppressWarnings(apply(x, 1, max, na.rm = TRUE))
    if (inherits(x, "numeric") && all(vapply(list(...), inherits, logical(1), "POSIXct"))) {
      x <- as.POSIXct(x, origin = "1970-01-01", tz = attr(list(...)[[1]], "tzone"))
    }
    redcap_na2R_na(x)
  }

  penv[["rowMean"]] <- function(...) {
    x <- get_input_as_num_matrix(...)
    x <- apply(x, 1, mean, na.rm = TRUE)
    if (inherits(x, "numeric") && all(vapply(list(...), inherits, logical(1), "POSIXct"))) {
      x <- as.POSIXct(x, origin = "1970-01-01", tz = attr(list(...)[[1]], "tzone"))
    }
    redcap_na2R_na(x)
  }

  penv[["median"]] <- function(...) {
    x <- make_num(c(...))
    redcap_na2R_na(median(x, na.rm = TRUE))
  }

  penv[["stdev"]] <- function(...) {
    x <- make_num(c(...))
    redcap_na2R_na(sd(x, na.rm = TRUE))
  }

  penv[["util_parse_date"]] <- util_parse_date
  attr(penv[["util_parse_date"]], "internal") <- TRUE

  penv[["util_parse_time"]] <- util_parse_time
  attr(penv[["util_parse_time"]], "internal") <- TRUE

  successive_dates <- function(..., strictly = TRUE) {
    util_expect_scalar(strictly, check_type = is.logical)
    if (length(unique(vapply(list(...), length, FUN.VALUE = integer(1))))
        != 1) {
      util_warning(
        c("%s was called with vectors of differing lengths. This should",
          "never happen."),
          sQuote("successive_dates")
      )
      return(FALSE)
    }
    all_dates <- list(...)
    FUN <- function(...) {
      dates <- list(...)
      if (length(dates) < 2) {
        return(TRUE)
      }
      is_timepoint <- vapply(dates, lubridate::is.timepoint, FUN.VALUE =
                               logical(1))
      if (!all(is_timepoint)) {
        dates[!is_timepoint] <- lapply(dates[!is_timepoint],
                                       function(x) {
                                         if (util_empty(x)) {
                                           NULL
                                         } else
                                         try(
                                           util_parse_date(x,
                                                      tryFormats =
                                                    c("%Y-%m-%d %H:%M:%OS",
                                                          "%Y-%m-%d"),
                                                    optional = FALSE),
                                           silent = TRUE
                                          )
                                        })
        dates <- dates[!vapply(dates, is.null, FUN.VALUE = logical(1))]
        is_timepoint <- vapply(dates, lubridate::is.timepoint, FUN.VALUE =
                                 logical(1))
        if (!all(is_timepoint)) {
          util_warning(c("Found non-dates in %s: %s, so,",
                         "dates are not successive."),
                       sQuote("successive_dates"),
                       util_pretty_vector_string(dates[!is_timepoint]))
          return(NA)
        }
      }
      LHS_indices_to_check <- (head(seq_along(dates), -1))
      if (strictly) {
        .cmp_fun <- function(lhs_index) {
          dates[[lhs_index]] < dates[[lhs_index + 1]]
        }
      } else {
        .cmp_fun <- function(lhs_index) {
          dates[[lhs_index]] <= dates[[lhs_index + 1]]
        }
      }
      vapply(LHS_indices_to_check, FUN.VALUE = logical(1),
             FUN = .cmp_fun)
    }
    vapply(do.call(what = mapply,
            args =
              c(
                all_dates,
                list(
                  FUN = FUN,
                  SIMPLIFY = FALSE,
                  USE.NAMES = FALSE,
                  MoreArgs = list()
                )
              )
            ), all, na.rm = TRUE, FUN.VALUE = logical(1))
  }

  penv[["successive_dates"]] <- function(...) successive_dates(...,
                                                               strictly = FALSE)
  penv[["strictly_successive_dates"]] <- function(...)
    successive_dates(...,  strictly = TRUE)

  penv[["maxLongStr"]] <- function(...) {
    # maximum longstring
    all_data <- cbind(...) # TODO: Do we need anything like get_input_as_num_matrix??
    apply(all_data, 1, function(rw) { max(rle(rw)$length) })
  }

  penv[["NCOL"]] <- function(...) {
    length(list(...))
  }

  penv[["NROW"]] <- function(...) {
    if (length(list(...)) < 1) {
      return(0)
    } else {
      if (1 != length(unique(vapply(list(...), length, FUN.VALUE = integer(1))))) {
        util_warning(
          c("NROW was used for a set of variables with different",
            "observation counts (including NAs)"),
          applicability_problem = TRUE)
      }
      return(length(list(...)[[1]]))
    }
  }

  penv[["perc_miss_in_row"]] <- function(...) {
    all_data <- cbind(...)
    apply(all_data, 1, function(rw) { 100 * sum(is.na(rw)) / length(rw) })
  }


  penv[["CASES_NA_BELOW"]] <- function(p, ...) {
    if (inherits(p, "MISSING")) {
      p <- 100
    }
    util_expect_scalar(p, check_type =
                         util_is_numeric_in(min = 0,
                                            max = 100),
                       error_message =
"Percentage of variables need not to be NA must be a number between 0 and 100")
    # filter rows with PCT_NA>x
    all_data <- apply(cbind(...), 1:2, as.numeric) # FIXME: Use get_input_as_num_matrix
    i <- 100 * rowSums(is.na(all_data)) / ncol(all_data) <= p
    if (sum(i) < nrow(all_data)) {
      all_data[!i, ] <- NA
      util_message("Deleting %d records because of < %d%% variables filled",
                   nrow(all_data) - sum(i),
                   p)
    }
    all_data # [i, , drop = FALSE] cannot delete records totally, later, they won't match the dataset any more.
  }

  penv[["IRV"]] <- function(...) { # TODO: maybe, first, run some convert_factors -- data_preparation -- should be addressed by LABEL in DATA_PREPARATION, but untested, so far!?
    # Intra-individual Response variability
    all_data <- apply(cbind(...), 1:2, as.numeric) # FIXME: Use get_input_as_num_matrix
    apply(all_data, 1, sd, na.rm = TRUE)
  }

  penv[["QCHISQ"]] <- function(p, df) {
    stats::qchisq(p = p, df = df)
  }

  penv[["mahalanobis_threshold"]] <- function(x) {
    if (missing(x) || length(x) == 0) {
      return(NA_real_)
    }
    util_expect_scalar(x)
    by_opt <- getOption("dataquieR.MAHALANOBIS_THRESHOLD",
                    dataquieR.MAHALANOBIS_THRESHOLD_default)
    if (tolower(trimws(by_opt)) %in% c("true", "1", "t", "+")) {
      by_opt <- dataquieR.MAHALANOBIS_THRESHOLD_default
    }
    if (tolower(trimws(x)) %in% c("true", "1", "t", "+")) {
      .r <- by_opt
    } else {
      .r <- suppressWarnings(as.numeric(x))
    }
    if (length(.r) != 1) {
      .r <- NA_real_
    }
    .r
  }

  penv[["MAHALANOBIS"]] <- function(...) {

    x <- get_input_as_num_matrix(...)

    keep_rows <- rowSums(is.na(x)) == 0

    x_clean <- x[keep_rows, , drop = FALSE]

    # Calculation of the Mahalanobis distance
    mahalanobis(x, colMeans(x_clean), cov(x_clean))
  }

  # TODO: penv[["group_var"]], ...
  penv[["measurement_time"]] <- function(...) { # FIXME: Add-On-System for such additional functions. How to handle SSI entries, if such a function is missing?
      which_vars <-
      as.character(rlang::call_args(sys.call(length(sys.calls()))))
    meta_data <-
      get("[meta_data]", parent.frame())
    label_col <-
      get("[label_col]", parent.frame())
    if (TIME_VAR %in% colnames(meta_data)) {
      time_vars <-
        util_find_var_by_meta(which_vars,
                              meta_data = meta_data,
                              label_col = label_col,
                              target = TIME_VAR)
    } else {
      util_error("Don't have any time varialbes in the metadata",
                 applicability_problem = TRUE)
    }
    time_vars <-
      util_find_var_by_meta(time_vars,
                            meta_data = meta_data,
                            label_col = label_col,
                            target = label_col)
    res <- lapply(time_vars, get, envir = parent.frame())
    res <- suppressWarnings(lapply(res, util_parse_date))
    m <- do.call(cbind, res)
    class(m) <- c("POSIXct", "POSIXt")
    m
  }
  penv[["measurement_end_time"]] <- function(...) {
    which_vars <-
      as.character(rlang::call_args(sys.call(length(sys.calls()))))
    meta_data <-
      get("[meta_data]", parent.frame())
    label_col <-
      get("[label_col]", parent.frame())
    if (TIME_VAR_END %in% colnames(meta_data)) {
      time_vars <-
        util_find_var_by_meta(which_vars,
                              meta_data = meta_data,
                              label_col = label_col,
                              target = TIME_VAR_END)
    } else if (TIME_VAR %in% colnames(meta_data)) {
      time_vars <-
        util_find_var_by_meta(which_vars,
                              meta_data = meta_data,
                              label_col = label_col,
                              target = TIME_VAR)
    } else {
      util_error("Don't have any time varialbes in the metadata",
                 applicability_problem = TRUE)
    }
    time_vars <-
      util_find_var_by_meta(time_vars,
                            meta_data = meta_data,
                            label_col = label_col,
                            target = label_col)
    res <- lapply(time_vars, get, envir = parent.frame())
    res <- suppressWarnings(lapply(res, util_parse_date))
    m <- do.call(cbind, res)
    class(m) <- c("POSIXct", "POSIXt")
    m
  }

  penv[["datediff"]] <- function(date1, # TODO: if used with vectors of posixct, this crashes: util_eval_rule(util_parse_redcap_rule('datediff(set("2020-11-01", "", "2020-11-10"), "2020-01-01", "y", "Y-M-D", true)'), ds1 = ds1, meta_data = meta_data, use_value_labels = FALSE)
                                 date2,
                                 units = "",
                                 date_format = "Y-M-D",
                                 Return_Signed_Value = FALSE) {

    util_expect_scalar(arg_name = units,
                       allow_na = FALSE, allow_null = FALSE,
                       check_type = function(x) x %in% c("y", "M", "d",
                                           "h", "m", "s", ""))

    util_expect_scalar(arg_name = date_format,
                       allow_na = FALSE, allow_null = FALSE,
                       check_type = function(x) x %in% c("Y-M-D",
                                                         "M-D-Y",
                                                         "D-M-Y"))

    try_formats <- list(
      `Y-M-D` = list("%Y-%m-%d %H:%M:%OS", "%Y-%m-%d"),
      `M-D-Y` = list("%m-%d-%Y %H:%M:%OS", "%m-%d-%Y"),
      `D-M-Y` = list("%d-%m-%Y %H:%M:%OS", "%d-%m-%Y")
    )[[date_format]]

    util_expect_scalar(arg_name = Return_Signed_Value,
                       allow_na = FALSE, allow_null = FALSE,
                       check_type = is.logical)

    util_expect_scalar(arg_name = date1,
                       allow_na = TRUE,
                       allow_null = FALSE,
                       allow_more_than_one = TRUE,
                       check_type = function(x) {
                         prep_dq_data_type_of(x) %in%
                           c(DATA_TYPES$DATETIME,
                             DATA_TYPES$STRING)
                       })

    util_expect_scalar(arg_name = date2,
                       allow_na = TRUE,
                       allow_null = FALSE,
                       allow_more_than_one = TRUE,
                       check_type = function(x) {
                         prep_dq_data_type_of(x) %in%
                           c(DATA_TYPES$DATETIME,
                             DATA_TYPES$STRING)
                       })

    if (prep_dq_data_type_of(date1) == DATA_TYPES$STRING) {
      try({
        date1 <- util_parse_date(date1, tryFormats = unlist(try_formats),
                                 optional = FALSE)
      }, silent = TRUE)
      if (prep_dq_data_type_of(date1) != DATA_TYPES$DATETIME) {
        date1 <- NA
      }
    }
    if (prep_dq_data_type_of(date2) == DATA_TYPES$STRING) {
      try({
        date2 <- util_parse_date(date2, tryFormats = unlist(try_formats),
                                 optional = FALSE)
      }, silent = TRUE)
      if (prep_dq_data_type_of(date2) != DATA_TYPES$DATETIME) {
        date2 <- NA
      }
    }

    if (units == "") units <- "a"

    u <-
      c("a" = "auto",
        "s" = "secs",
        "m" = "mins",
        "h" = "hours",
        "d" = "days",
        "M" = "months",
        "y" = "years")[[units]]

    if (units %in% c("M", "y")) {
      u <- "days"
    }

    res <- as.numeric(difftime(date1, date2,
                               units = u))

    if (units == "M") {
      res <- res / 30.44
    } else if (units == "y") {
      res <- res / 365.2425
    }

    if (!Return_Signed_Value) {
      res <- abs(res)
    }

    redcap_na2R_na(res)
  }

  # but use a new rule for this, it collides with REDcapexpression, use sth. like concat(keyword("if"),  charParser("("), logical_expression(), ...))
  # if ([xx] < 99, [xx], "NaN")
  # penv[["="]] <- base::`==`
  # penv[["=="]] <- base::`==`
  # penv[["!="]] <- base::`!=`
  # penv[["<>"]] <- base::`!=`
  # penv[[">="]] <- base::`>=`
  # penv[["<="]] <- base::`<=`
  # penv[["<"]] <- base::`<`
  # penv[[">"]] <- base::`>`
  # penv[["if"]] <- base::`ifelse`

  # Note: For each name, ensure, that no op function with the same prefix is in the same prio class
  # So, attr(., "order") is needed. Default order will be 0
  penv[["="]] <- gen_op(`==`)
  attr(penv[["="]], "prio") <- 0
  penv[["=="]] <- gen_op(`==`) #  must not be in prio=0 if prio 0 has no order
  attr(penv[["=="]], "order") <- -1
  attr(penv[["=="]], "prio") <- 0
  penv[["!="]] <- gen_op(`!=`)
  attr(penv[["!="]], "prio") <- 0
  penv[["<>"]] <- gen_op(`!=`) #  must not be in prio=0 if prio 0 has no order
  attr(penv[["<>"]], "order") <- -1
  attr(penv[["<>"]], "prio") <- 0
  penv[["<"]] <- gen_op(`<`)
  attr(penv[["<"]], "prio") <- 0
  penv[[">"]] <- gen_op(`>`)
  attr(penv[[">"]], "prio") <- 0
  penv[["<="]] <- gen_op(`<=`)
  attr(penv[["<="]], "prio") <- 0
  attr(penv[["<="]], "order") <- -1
  penv[[">="]] <- gen_op(`>=`)
  attr(penv[[">="]], "prio") <- 0
  attr(penv[[">="]], "order") <- -1

  penv[["set"]] <- function(...) {
    base::unique(base::c(...))
  }

  penv[["interval"]] <- function(inc_l, low, upp, inc_u) {
    util_expect_scalar(inc_l, check_type = base::is.logical)
    util_expect_scalar(inc_u, check_type = base::is.logical)
    if (inherits(low, "POSIXct")) {
      util_expect_scalar(low, check_type = function(x) {
        inherits(low, "POSIXct")
      })
    } else if (inherits(low, "hms")) {
      util_expect_scalar(low, check_type = function(x) {
        inherits(low, "hms")
      })
    } else {
      util_expect_scalar(low, check_type = base::is.numeric)
    }
    if (inherits(upp, "POSIXct")) {
      util_expect_scalar(upp, check_type = function(x) {
        inherits(upp, "POSIXct")
      })
    } else if (inherits(low, "hms")) {
      util_expect_scalar(low, check_type = function(x) {
        inherits(low, "hms")
      })
    } else  {
      util_expect_scalar(upp, check_type = base::is.numeric)
    }
    i <- list(inc_l = inc_l, low = low, upp = upp, inc_u = inc_u)
    class(i) <- "interval"
    i
  }
  attr(penv[["interval"]], "order") <- -1 # before in

  my_in <- gen_op(`%in%`)

  penv[["in"]] <- function(x, table) {
    if (inherits(table, "interval")) {
      r <- rep(TRUE, length(x))
      inc_l <- table$inc_l
      if (is.infinite(table$low) && table$low < 0) { # -Inf
        inc_l <- TRUE # despite the math rules, we have to "include" Inf, so
        # that -Inf in (-Inf; 0) works
      }
      inc_u <- table$inc_u
      if (is.infinite(table$upp) && table$upp > 0) { # +Inf
        inc_u <- TRUE # despite the math rules, we have to "include" Inf, so
        # that Inf in (Inf; Inf) works
      }
      if (inc_l) {
        r <- r & (x >= table$low)
      } else {
        r <- r & (x > table$low)
      }
      if (inc_u) {
        r <- r & (x <= table$upp)
      } else {
        r <- r & (x < table$upp)
      }
      r
    } else {
      my_in(x, table)
    }
  }
  attr(penv[["in"]], "prio") <- -1

  penv[["not in"]] <- function(x, table) {
    !penv[["in"]](x, table)
  }
  attr(penv[["not in"]], "prio") <- -1
  attr(penv[["not in"]], "order") <- -1

  penv[["if"]] <- decorate_fkt_redcap_na(function(test, yes, no) {
    ifelse(test, yes, no)
  })

  penv[["not"]] <- decorate_fkt_redcap_na(function(x0) {
    x <- suppressWarnings(as.logical(x0))
    which_not <- util_empty(x) != util_empty(x0)
    if (any(which_not)) {
      util_warning("Not a logical value: %s. %s returns %s",
                   util_pretty_vector_string(paste(x0[which_not]), n_max = 4),
                   sQuote("not"),
                   util_pretty_vector_string(paste(x[which_not])), n_max = 4)
      NA
    } else {
      !x
    }
  })

  penv[["pi"]] <- pi
  penv[["e"]] <- exp(1)
  penv[["pearson"]] <- function(...) {
    x <- get_input_as_num_matrix(...)
    x <- cor(x, use = "complete.obs", method = "pearson")
    x
  }
  penv[["spearman"]] <- function(...) {
    x <- get_input_as_num_matrix(...)
    x <- cor(x, use = "complete.obs", method = "spearman")
    x
  }
  penv[["kendall"]] <- function(...) {
    x <- get_input_as_num_matrix(...)
    x <- cor(x, use = "complete.obs", method = "kendall")
    x
  }

  penv[["."]] <- structure(NA, class = "MISSING")

  `in` <- function(...) {
    `%in%`(...)
  }

#  penv[["id"]] <- function(x) x

  makeActiveBinding("today", function() { util_parse_date(Sys.Date()) }, penv)

  penv
})()

test_that("util_parse_redcap_rule works", {
  skip_on_cran()
  skip_if_not_installed(c("withr"))
  skip_if_not_installed(c("qmrparser"))
  skip_if_not_installed(c("callr"))
  # TODO: also consider the results of the function, not only, that it does not report syntax errors
  debug <- 1
  suppressMessages({
    expect_message2(util_parse_redcap_rule("(12)", debug = debug))
    expect_message2(util_parse_redcap_rule("12", debug = debug))
    expect_warning(util_parse_redcap_rule("xxxsin(12)", debug = debug))
    expect_message2(util_parse_redcap_rule("sum(12)", debug = debug))
    expect_message2(util_parse_redcap_rule("(12 + 1)", debug = debug))
    expect_message2(util_parse_redcap_rule("(12 + 1 * 2)", debug = debug))
    expect_message2(util_parse_redcap_rule("(12 + 1) * 2", debug = debug))
    expect_message2(util_parse_redcap_rule("12 + 1 * 2", debug = debug))
    expect_message2(util_parse_redcap_rule("12 + (1 * 2)", debug = debug))
    expect_message2(util_parse_redcap_rule("(12 + (1 * 2))", debug = debug))
    expect_message2(util_parse_redcap_rule("(12 + (1 * 2)) > 1", debug = debug))
    expect_message2(util_parse_redcap_rule("((12 + (1 * 2)) > 1)", debug = debug))
    expect_message2(util_parse_redcap_rule("12 + (1 * 2) > 1", debug = debug))
    expect_message2(util_parse_redcap_rule("(12 + (1 * 2) > 1)", debug = debug))
    expect_silent(util_parse_redcap_rule("12 + (1 * 2) > 1"))
    expect_silent(util_parse_redcap_rule("12 + (1 * 2) > 1 and true"))
    expect_silent(util_parse_redcap_rule("12 + (1 * 2) > 1 and [speed] > 1"))
    expect_silent(util_parse_redcap_rule("(12 + (1 * 2) > 1) and ([speed] > 1)"))
    expect_silent(util_parse_redcap_rule("(1 * 2) > 1 and true"))
    expect_message2(util_parse_redcap_rule("(1 * 2) > 1 and 2 > 1", debug = debug))
    expect_silent(util_parse_redcap_rule(rule = '[speed] > 5 and [dist] > 42 or 1 = "2"'))
    expect_silent(util_parse_redcap_rule(rule = '[speed] > 5 or [dist] > 42 and 1 = "2"'))
    expect_error(util_parse_redcap_rule(rule = '[speed] > 5', entry_pred = "non existing"))
    expect_silent(util_parse_redcap_rule(rule = '([speed] > 5)'))
    expect_silent(util_parse_redcap_rule(rule = '[speed]', entry_pred = "term_expression"))
    expect_silent(util_parse_redcap_rule(rule = '[speed] > 5 and [dist] > 42'))
    expect_silent(util_parse_redcap_rule(rule = '[speed] > 5'))
    expect_silent(util_parse_redcap_rule(rule = '[speed] > 5', entry_pred = "term_expression"))
    expect_silent(util_parse_redcap_rule(rule = '[speed] > 5 or 2 > 1', entry_pred = "term_expression"))
    expect_silent(util_parse_redcap_rule(rule = '[speed] > 5 and 12 > 5'))
    expect_silent(util_parse_redcap_rule(rule = '[speed]'))
    expect_silent(util_parse_redcap_rule(rule = '12'))
    expect_silent(util_parse_redcap_rule(rule = '([speed] > 5) and (1 = 2)'))
    expect_silent(util_parse_redcap_rule(rule = '5*5'))
    expect_silent (util_parse_redcap_rule(rule = 'prod(5*5)'))
    expect_silent(util_parse_redcap_rule(rule = 'prod(5*5) < 1'))
    expect_silent(util_parse_redcap_rule(rule = 'prod(5*5) < 1 and true'))
    expect_silent(util_parse_redcap_rule(rule = 'prod(5*5) < 1 and true or false'))
    expect_silent(util_parse_redcap_rule(rule = 'prod(5*5) < 1 and (true or false)'))
    expect_silent(util_parse_redcap_rule(rule = '(prod(5*5) < 1 and (true) or false)'))
    expect_silent(util_parse_redcap_rule(rule = '(prod(5*5) < 1 and true) or false'))
    expect_silent(util_parse_redcap_rule(rule = '(prod(5*5) < (1) and true) or false'))
    expect_silent(util_parse_redcap_rule(rule = '1 + 3 * 3'))
    expect_silent(util_parse_redcap_rule(rule = '1 + 3 * 3 and false'))
    expect_silent(util_parse_redcap_rule("[a] = 12 or [b] = 13"))
    expect_silent(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] <> "" and datediff([con_consentdt],[sda_osd1dt],"d",true) < 0'))
    expect_silent(cars[eval(util_parse_redcap_rule(rule = '[speed] > 5 and [dist] > 42 or 1 = "2"'), cars, util_get_redcap_rule_env()), ])
    expect_silent(util_parse_redcap_rule('datediff([con_consentdt],[sda_osd1dt],"d",true) < 0'))
    expect_silent(util_parse_redcap_rule('datediff([con_consentdt],[sda_osd1dt],"d",true)'))
    expect_message2(util_parse_redcap_rule('datediff([con_consentdt],[sda_osd1dt],"d",true)', debug = debug))
    expect_message2(util_parse_redcap_rule('datediff([con_consentdt],[sda_osd1dt],"d",true)', debug = debug, entry_pred = "function_expression"))
    expect_message2(util_parse_redcap_rule('[con_consentdt],[sda_osd1dt],"d",true', debug = debug, entry_pred = "arg_part"))
    expect_message2(util_parse_redcap_rule('[con_consentdt]', debug = debug, entry_pred = "term_expression"))
    expect_message2(util_parse_redcap_rule('[con_consentdt]', debug = debug, entry_pred = "arg"))
    expect_message2(util_parse_redcap_rule('[con_consentdt]', debug = debug, entry_pred = "symbol_expression"))
    expect_error(eval(util_parse_redcap_rule('datediff([con_consentdt],[sda_osd1dt],"d",true)', debug = debug, entry_pred = "function_expression"), cars, util_get_redcap_rule_env()))
    expect_error(eval(util_parse_redcap_rule('datediff([con_consentdt],[sda_osd1dt],"d",true)', entry_pred = "function_expression"), cars, util_get_redcap_rule_env()))
    expect_error(suppressWarnings(suppressMessages(eval(util_parse_redcap_rule('datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true)', entry_pred = "function_expression"), cars, util_get_redcap_rule_env()))))
    x <- data.frame(con_consentdt = c(as.POSIXct("2020-01-01"), as.POSIXct("2020-10-20")), sda_osd1dt = c(as.POSIXct("2020-01-20"), as.POSIXct("2020-10-01")))
    expect_silent(eval(util_parse_redcap_rule('datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true)', entry_pred = "function_expression"), x, util_get_redcap_rule_env()))
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] <> "" and datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true)'), x, util_get_redcap_rule_env()))
    x <- cars
    expect_error(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] <> "" and datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true)'), x, util_get_redcap_rule_env()))
    x <- data.frame(con_consentdt = c(as.POSIXct("2020-01-01"), as.POSIXct("2020-10-20")), sda_osd1dt = c(as.POSIXct("2020-01-20"), as.POSIXct("2020-10-01")))
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] <> "" and datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true) < 10'), x, util_get_redcap_rule_env()))
    expect_silent(eval(util_parse_redcap_rule('datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true) < 10'), x, util_get_redcap_rule_env()))
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] <> ""'), x, util_get_redcap_rule_env()))
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] == ""'), x, util_get_redcap_rule_env()))
    x$sda_osd1dt <- NA
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] == ""'), x, util_get_redcap_rule_env()))
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] == "" and [sda_osd1dt] == ""'), x, util_get_redcap_rule_env()))
    x$sda_osd1dt[1] <- as.POSIXct("2020-01-01")
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] == "" and [sda_osd1dt] == ""'), x, util_get_redcap_rule_env()))
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] == "" or [sda_osd1dt] == ""'), x, util_get_redcap_rule_env()))
    expect_silent(expect_error(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] <> "" and datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true) < 10'), x, util_get_redcap_rule_env())))
    x <- data.frame(con_consentdt = c(as.POSIXct("2020-01-01"), as.POSIXct("2020-10-20")), sda_osd1dt = c(as.POSIXct("2020-01-20"), as.POSIXct("2020-10-01")))
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] <> "today" and datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true) < 10'), x, util_get_redcap_rule_env()))
    x$sda_osd1dt[1] <- Sys.Date()
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] <> "today" and datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true) < 10'), x, util_get_redcap_rule_env()))
    expect_silent(eval(util_parse_redcap_rule('[con_consentdt] <> "" and [sda_osd1dt] = "today" and datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true) < 10'), x, util_get_redcap_rule_env()))
    expect_silent(lapply(lapply(setNames(nm = c("(12) > 0 and [speed] > 12", "4+4*5+6")), util_parse_redcap_rule, debug = 0), eval, cars, util_get_redcap_rule_env()))
    expect_equal(
      util_eval_rule(util_parse_redcap_rule("12 * ([speed] + not(not(true or false)))", debug = 0), ds1 = cars, use_value_labels = FALSE),
      with(cars, 12 * (speed + !(!(TRUE || FALSE))))
      )
    expect_equal(
      util_eval_rule(util_parse_redcap_rule("if([speed] > 5, 1, 0)", debug = 0), ds1 = cars, use_value_labels = FALSE),
      with(cars, ifelse(speed > 5, 1, 0))
    )

    x <- tibble::tribble( ~ con_consentdt, ~ cont_inddt, ~ sda_osd1dt,
      as.POSIXct("2020-01-01"), as.POSIXct("2020-01-20"), as.POSIXct("2020-01-20"),
      as.POSIXct("2020-10-20"), as.POSIXct("2020-10-20 15:00:00"), as.POSIXct("2020-10-20 17:00:00")
    )

    expect_equal(
      util_eval_rule(util_parse_redcap_rule("successive_dates([con_consentdt], [cont_inddt], [sda_osd1dt])", debug = 0),
                     ds1 = x, use_value_labels = FALSE),
      with(x, con_consentdt <= cont_inddt & cont_inddt <= sda_osd1dt)
    )

    expect_equal(
      util_eval_rule(util_parse_redcap_rule("strictly_successive_dates([con_consentdt], [cont_inddt], [sda_osd1dt])", debug = 0),
                     ds1 = x, use_value_labels = FALSE),
      with(x, con_consentdt < cont_inddt & cont_inddt < sda_osd1dt)
    )
  })
})

#' Utility function verifying syntax of known metadata columns
#'
#' This function goes through metadata columns, `dataquieR` supports
#' and verifies for these, that they follow its metadata conventions.
#'
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#'
#' @return [data.frame] possibly modified `meta_data`, `invisible()`
#'
#' @family metadata_management
#' @concept robustness
#' @noRd
util_validate_known_meta <- function(meta_data) {
  # TODO: use redcap parser instead, include tests for LOCATION columns and PROPORTIONS columns
  if (!VAR_NAMES %in% colnames(meta_data)) { # avoid errors in checks, if
    # not a complete metadata frame is being checked.
    meta_data[[VAR_NAMES]] <- paste0("v", seq_len(nrow(meta_data)))
  }
  if (any(util_empty(meta_data[[VAR_NAMES]]))) {
    util_warning("Found variables w/o %s. Added dummy names", sQuote(VAR_NAMES),
                 applicability_problem = TRUE)
    meta_data[[VAR_NAMES]][util_empty(meta_data[[VAR_NAMES]])] <-
      paste0("v", seq_len(nrow(meta_data)))[util_empty(meta_data[[VAR_NAMES]])]
  }
  if (any(duplicated(meta_data))) {
    util_error(
      "Found duplicated rows in item-level metadata. Please fix.",
      applicability_problem = TRUE)
    # sorrry, cannot fix on the fly, because the context, where this is called, is too heterogeneous.
#    meta_data <- meta_data[!duplicated(meta_data), , FALSE]
  }
  for (no_dups in
        c(VAR_NAMES,
          grep("^LABEL(_.+)?$", colnames(meta_data)),
          grep("^LONG_LABEL(_.+)?$", colnames(meta_data)))
       ) {
    if (any(duplicated(meta_data[[no_dups]]))) {
      util_error("Found duplicated %s in item-level metadata: %s. Please fix.",
                 sQuote(no_dups),
                 util_pretty_vector_string(
                   sort(unique(meta_data[[no_dups]][
                     duplicated(meta_data[[no_dups]])]))
                 ),
                 applicability_problem = TRUE)
    }
  }
  util_validate_missing_lists(meta_data)
  env <- new.env(environment())
  env$error <- character(0)
  if (any(grepl("^(HARD|SOFT|DETECTION)_LIMITS$", colnames(meta_data)))) {
    suppressMessages(suppressWarnings(
      l <- withCallingHandlers(
        util_interpret_limits(mdata = meta_data),
        warning = function(w) env$error <- c(env$error, conditionMessage(w)),
        error = function(e) env$error <- c(env$error, conditionMessage(e))
      )
    ))
    newlimcols <- grep("INCL_([^_]*)_LIMIT_(UP|LOW)", colnames(l), value = TRUE)
    newlimcols <- gsub("^INCL_", "", newlimcols)
    newlimcols <- unique(gsub("_LIMIT_(LOW|UP)$", "", newlimcols))
    for (limc in newlimcols) {
      wrong_interval_order <- l[, sprintf("%s_LIMIT_LOW", limc), FALSE] >
        l[, sprintf("%s_LIMIT_UP", limc), FALSE]
      wrong_interval_order[is.na(wrong_interval_order)] <- FALSE
      if (any(wrong_interval_order)) {
        env$error <- c(env$error,
                   sprintf("Found %s with lower limit > upper limit: %s",
                           limc,
                           unique(unlist(l[wrong_interval_order,
                                    sprintf("%s_LIMITS", limc),
                                    TRUE]))))
      }
    }
    env$error <- unique(env$error)
    if (length(env$error) > 0) {
      util_warning("Invalid limits detected: %s",
                   paste0(env$error, collapse = "\n"),
                   applicability_problem = TRUE)
    }
  }

  env$error <- character(0)

  md9 <- meta_data
  if (!VALUE_LABEL_TABLE %in% colnames(md9)) {
    md9[[VALUE_LABEL_TABLE]] <- NA_character_
  }
  if (!STANDARDIZED_VOCABULARY_TABLE %in% colnames(md9)) {
    md9[[STANDARDIZED_VOCABULARY_TABLE]] <- NA_character_
  }
  if (!SCALE_LEVEL %in% colnames(md9)) {
    md9[[SCALE_LEVEL]] <- NA_character_
  }
  invisible(mapply(
    vlt = md9[[VALUE_LABEL_TABLE]],
    svt = md9[[STANDARDIZED_VOCABULARY_TABLE]],
    sl = md9[[SCALE_LEVEL]],
    vn = md9[[VAR_NAMES]],
    FUN = function(vlt, svt, sl, vn) {
      if (!util_empty(vlt)) {
        df <- try(prep_get_data_frame(vlt))
        if (util_is_try_error(df)) {
          env$error <- c(env$error,
                         sprintf(
                           "Cannot load %s %s for %s: %s",
                           sQuote(VALUE_LABEL_TABLE),
                           dQuote(vlt),
                           dQuote(vn),
                           util_condition_from_try_error(df)))
        } else {
          if (!CODE_VALUE %in% colnames(df)) {
            env$error <- c(env$error,
                           sprintf(
                             "%s %s is missing column %s for %s",
                             sQuote(VALUE_LABEL_TABLE),
                             dQuote(vlt),
                             sQuote(CODE_VALUE),
                             dQuote(vn)))
          } else if (length(unique(df[[CODE_VALUE]])) == 0) {
            env$error <- c(env$error,
                           sprintf(
                             "%s %s is empty for %s",
                             sQuote(VALUE_LABEL_TABLE),
                             dQuote(vlt),
                             dQuote(vn)))
          } else if (length(unique(df[[CODE_VALUE]])) == 1) {
            env$error <- c(env$error,
                           sprintf(
                             "%s %s is suspicious for %s: only one value",
                             sQuote(VALUE_LABEL_TABLE),
                             dQuote(vlt),
                             dQuote(vn)))
          }
          if (CODE_ORDER %in% colnames(df) &&
              !is.na(sl) &&
              sl != SCALE_LEVELS$ORDINAL) {
            env$error <- c(env$error,
                           sprintf(
                             "Ordinal variable with unordered %s: %s",
                             sQuote(VALUE_LABEL_TABLE),
                             dQuote(vn)))
          }
          if (!CODE_ORDER %in% colnames(df) &&
              !is.na(sl) &&
              sl != SCALE_LEVELS$NOMINAL) {
            env$error <- c(env$error,
                           sprintf(
                             "Nominal variable with ordered %s: %s",
                             sQuote(VALUE_LABEL_TABLE),
                             dQuote(vn)))
          }
        }
      }
    }
  ))

  # TODO: Check STANDARDIZED_VOCABULARY_TABLE, too
  if (VALUE_LABELS %in% colnames(meta_data)) {
    invisible(mapply(vl = meta_data[[VALUE_LABELS]],
           vn = meta_data[[VAR_NAMES]],
      FUN = function(vl, vn) {
        if (!util_empty(vl)) {
          if (!grepl(SPLIT_CHAR, vl, fixed = TRUE) &&
              !grepl("<", vl, fixed = TRUE)) {
            env$error <- c(env$error,
                           sprintf(
           "Suspicious value labels (only 1 level) detected: %s (%s)",
                             dQuote(vl), dQuote(vn)))
          }
        }
      }
    ))
    # TODO: do the check for unsuitable missing code type from util_get_code_list (util_get_code_list) here, instead, and summarize the warnings.
    # TODO: Check whether VALUE_LABELS and entries in the study data match (e.g.,
    # SEX_0 coded as string ("female", "male"), but VALUE_LABELS are
    # 0 = "female" | 1 = "male", or the other way around - SEX_0 coded as
    # integer, but VALUE_LABELS give only "female", "male"
    if (SCALE_LEVEL %in% colnames(meta_data)) {
      invalid_scale_levels <-
        !util_empty(meta_data[[SCALE_LEVEL]]) &
        !(meta_data[[SCALE_LEVEL]] %in% unlist(SCALE_LEVELS))
      if (any(invalid_scale_levels)) {
        util_warning("Found invalid scale levels: %s",
          util_pretty_vector_string(
            paste0(meta_data[invalid_scale_levels, VAR_NAMES, drop = TRUE],
                   ": ",
                  meta_data[invalid_scale_levels, SCALE_LEVEL, drop = TRUE]))
        )
      }
      value_list_scale_level_missmatch <-
        meta_data[[SCALE_LEVEL]] == SCALE_LEVELS$NOMINAL
      value_list_scale_level_missmatch[
        vapply(
          FUN.VALUE = character(1),
          util_parse_assignments(
            meta_data[[VALUE_LABELS]],
            multi_variate_text = TRUE,
            split_char = c(SPLIT_CHAR, "<")), attr, "split_char") == SPLIT_CHAR
      ] <- FALSE
      if (any(value_list_scale_level_missmatch)) {
        env$error <- c(env$error,
                       sprintf(
              "Value labels with %s found for variable(s) %s with %s = %s",
                        dQuote("<"),
                        paste(dQuote(meta_data[value_list_scale_level_missmatch,
                                         VAR_NAMES, drop = TRUE]),
                              collapse = ", "),
                        sQuote(SCALE_LEVEL),
                        dQuote(SCALE_LEVELS$NOMINAL)))
      }
    }
    env$error <- unique(env$error)
    if (length(env$error) > 0)
      util_warning(paste(env$error, collapse = "\n"),
                   applicability_problem = TRUE)
  }
  if (any(c(MISSING_LIST, JUMP_LIST) %in% colnames(meta_data))) {
    env$error <- character(0)
    for (l in intersect(c(MISSING_LIST, JUMP_LIST), colnames(meta_data))) {
      vl <- meta_data[[l]]
      notnumdat <- !grepl(sprintf(
        "^[%s0-9\\s\\-\\.\\: ]*$",
        SPLIT_CHAR), perl = TRUE, vl)
      not_assign <- vapply(vl, FUN.VALUE = logical(1),
                           function(x) {
                             if (is.na(x)) return(FALSE);
                             v <- names(util_parse_assignments(x))
                             any(is.na(v) != suppressWarnings(
                                 is.na(as.numeric(v))) &
                                   is.na(v) != suppressWarnings(
                                     is.na(util_parse_date(v))) &
                                   is.na(v) != suppressWarnings(
                                     is.na(util_parse_time(v)))
                                   )
                           })
      if (any(!util_empty(vl) & notnumdat & not_assign)) {
        env$error <- c(env$error,
                   sprintf(
                     "Suspicious %s: not numeric/date/time/assignment: %s",
                     sQuote(l),
                     paste(dQuote(unique(sort(
                       as.character(vl[!util_empty(vl) & notnumdat & not_assign])))),
                       collapse = ", ")))
      } else {
        vlc <- as.character(vl)
        if (any(is.na(vlc) != is.na(vl))) { # nocov start
          util_warning(c("This should never be displayed,",
                         "so you've found a bug in dataquieR.",
                         "Please report it as bug in",
                         "%s to %s"),
                       dQuote("util_validate_known_meta"),
                       dQuote(utils::packageDescription("dataquieR")[["BugReports"]]),
                       applicability_problem = FALSE)
          env$error <- c(env$error,
            sprintf("Found missing or jump code not being a character: %s",
                       dQuote(
                         deparse(head(vl[is.na(vlc) != is.na(vl)], 1),
                                 nlines = 1)))
          )
        } # nocov end
        vl <- vlc
        lists <- strsplit(vl,
                 split = sprintf("\\s*[%s]\\s*", SPLIT_CHAR),
                 perl = TRUE,
                )
        dups <-
          vapply(lists, FUN.VALUE = logical(1), function(x) any(duplicated(x)))
        if (any(dups)) {
          dup_lists <- unique(unlist(
            lapply(lists[dups], paste, collapse = sprintf(" %s ", SPLIT_CHAR))))
          util_warning(
            "Duplicates in %s: %s. Maybe another missing code is not listed?",
                       dQuote(l),
                       paste0(sQuote(dup_lists), collapse = ", "),
            applicability_problem = TRUE)
        }
      }
    }
    env$error <- unique(env$error)
    if (length(env$error) > 0)
      util_warning(env$error, applicability_problem = TRUE)
  }

  invisible(meta_data)
}

#' Abbreviate a vector of strings
#'
#' @param initial [character] vector with stuff to abbreviate
#' @param max_value_label_len [integer] maximum length (may not strictly
#'                                      be met, if not possible keeping a maybe
#'                                      detected uniqueness of `initial`)
#'
#' @return [character] uniquely abbreviated `initial`
#' @family string_functions
#' @concept string
#' @noRd
util_abbreviate_unique <- function(initial, max_value_label_len) {
  # TODO: use code from util_ensure_label for abbreviations
  no_dups <- !any(duplicated(initial))
  abb <- substr(initial, 1, max_value_label_len)
  if (no_dups) while (any(dups <- duplicated(abb))) {
    abb[dups] <- substr(initial[dups], 1, max(1, max_value_label_len -
                          nchar(sum(dups))))
    abb[dups] <- paste0(abb[dups], seq_len(sum(dups)))
  }
  abb
}

test_that("util_validate_known_meta works", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # TODO: Also test retunred value
  local({
    meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
    md1 <- meta_data
    expect_silent(util_validate_known_meta(meta_data = md1))
    md1$VAR_NAMES <- NULL
    expect_silent(util_validate_known_meta(meta_data = md1))
    md1 <- meta_data
    md1$HARD_LIMITS[[10]] <- "[9; 1]"
    expect_warning(util_validate_known_meta(meta_data = md1),
                   regexp = sprintf("(%s|%s)",
                     paste("Invalid limits detected: Found HARD with",
                           "lower limit > upper limit: \\[9; 1\\]"),
          paste("Some code labels or -values are missing from .+meta_data.+")),
                   perl = TRUE)
    md1 <- meta_data
    md1$VALUE_LABELS[[53]] <- "0 = no"
    expect_warning(util_validate_known_meta(meta_data = md1),
                   regexp = sprintf("(%s|%s)",
                                    paste("Suspicious value labels",
                                          ".only 1 level. detected:",
                                          ".+0 = no.+"),
                                    paste("Some code labels or -values are",
                                          "missing from .+meta_data.+")
                                    ),
                   perl = TRUE)
    md1 <- meta_data
    md1$MISSING_LIST[[44]] <- "x| 9"

    expect_warning(expect_warning(
        util_validate_known_meta(meta_data = md1),
       regexp = paste("Some missing codes are not numeric"),
       perl = TRUE),
     regexp = paste("Suspicious .+MISSING_LIST.+:",
                    "not numeric/date/time/assignment"),
     perl = TRUE)

    md1 <- meta_data
    md1$MISSING_LIST[[44]] <- list(1:10)
    expect_warning(expect_warning(
      util_validate_known_meta(meta_data = md1),
      regexp = paste("Some missing codes are not numeric"),
      perl = TRUE),
      regexp = paste("Suspicious .+MISSING_LIST.+:",
                     "not numeric/date/time/assignment"),
      perl = TRUE)

    md1 <- meta_data
    md1$MISSING_LIST[[44]] <- paste(md1$MISSING_LIST[[44]],
                                    md1$MISSING_LIST[[44]],
                                    collapse = SPLIT_CHAR)
    expect_warning(expect_warning(
      util_validate_known_meta(meta_data = md1),
      regexp = paste("Some missing codes are not numeric"),
      perl = TRUE),
      regexp = paste("Duplicates in"),
      perl = TRUE)
  })
})

#' Compares study data data types with the ones expected according to the
#' metadata
#'
#' Utility function to compare data type of study data with those defined
#' in metadata
#'
#' @param sdf the [data.frame] of study data
#' @param mdf the [data.frame] of associated static metadata
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param check_convertible [logical] also try, if a conversion to the
#'                                    declared data type would work.
#' @param check_conversion_stable [logical] do not distinguish convertible
#'                                          from convertible, but with issues
#' @param threshold_value [numeric] from=0 to=100. percentage failing
#'                                  conversions allowed if `check_convertible`
#'                                  is `TRUE`.
#' @param return_percentages [logical] return the percentage of
#'                                mismatches.
#'
#' @return for `return_percentages == FALSE`: if `check_convertible` is `FALSE`,
#'         a binary vector `(0, 1)` if data type applies,
#'         if `check_convertible` is `TRUE``
#'         a vector with the states `0, 1, 2, 3`: 0 = Mismatch, not convertible
#'                                                1 = Match
#'                                                2 = Mismatch, but convertible
#'                                                3 = Mismatch, convertible, but
#'                                                    with issues (e.g., loss
#'                                                    of decimal places)
#'         for `return_percentages == TRUE`: a data frame with percentages of
#'         non-matching datatypes according, each column is a variable, the
#'         rows follow the vectors returned by `util_check_data_type`.
#' @importFrom stats setNames
#'
#' @seealso [prep_dq_data_type_of]
#' @seealso [prep_datatype_from_data]
#'
#' @family data_management
#' @concept robustness
#' @noRd
#'
util_compare_meta_with_study <- function(sdf, mdf, label_col,
                                         check_convertible = FALSE,
                                         threshold_value = 0,
                                         return_percentages = FALSE,
                                         check_conversion_stable = FALSE
                                         ) {
  if (any(trimws(colnames(sdf)) == "")) {
    sdf[, trimws(colnames(sdf)) == ""] <- paste0("v",
                                                 seq_len(
                                                   sum(trimws(colnames(sdf)) ==
                                                         "")))
    util_warning("Found columns w/o names in %s, using dummy names",
               dQuote(label_col),
               applicability_problem = TRUE)
  }
  if (any(trimws(mdf[[label_col]]) == "")) {
    mdf[trimws(mdf[[label_col]]) == "", label_col] <-
      paste0("v", seq_len(sum(trimws(mdf[[label_col]]) == "")))
    util_warning("Found empty labels in %s, using dummy labels",
               sQuote("study_data"), applicability_problem = TRUE)
  }
  missing_in_study_data <- !(mdf[[label_col]] %in% colnames(sdf))
  sdf <- sdf[, as.character(mdf[[label_col]][!missing_in_study_data]),
             drop = FALSE]
  if (return_percentages) {
    if (!check_conversion_stable) {
      util_error("%s = %s is only allowed, if %s = %s",
                  sQuote("check_conversion_stable"),
                  sQuote("FALSE"),
                  sQuote("return_percentages"),
                  sQuote("FALSE")
                 )
    }
    if (check_convertible)
      res_template <- numeric(4)
    else
      res_template <- numeric(1)
  } else {
    res_template <- integer(1)
  }
  sdf[util_empty(sdf)] <- NA # enable robust_na = FALSE option to accelerate -- util_empty keeps structure, important for nrow or ncol < 2
  is_data_type <- lapply(setNames(nm = colnames(sdf)),
                         function(x, check_convertible, threshold_value,
                                  return_percentages, check_conversion_stable) {
    ..vct <- sdf[, x, drop = TRUE]
    attr(..vct, "..cn") <- x
    res <- util_check_data_type(..vct,
                                mdf$DATA_TYPE[mdf[[label_col]] == x],
                                check_convertible = check_convertible,
                                robust_na = FALSE,
                                threshold_value = threshold_value,
                                return_percentages = return_percentages,
                                check_conversion_stable =
                                  check_conversion_stable,
                                vname = x)
    bitsToInt<-function(x) { # https://stackoverflow.com/a/25411493
      packBits(rev(c(rep(FALSE, 32-length(x)%%32), as.logical(x))), "integer")
    }
    which_vec <- NULL
    try({
      which_vec <- # loosing this on vapply. also: this is too slow.
        apply(as.matrix(do.call(cbind.data.frame, attr(res, "which"))) * 1, 1,
              bitsToInt)
    }, silent = TRUE)

    if (return_percentages) {
      util_attach_attr(setNames(as.numeric(res), nm = names(res)),
                       which_vec = which_vec)
    } else {
      util_attach_attr(as.integer(res), which_vec = which_vec)
    }
  },
  check_convertible = check_convertible,
  threshold_value = threshold_value,
  return_percentages = return_percentages,
  check_conversion_stable = check_conversion_stable)
  which_vec <- NULL
  if (length(res_template) > 1) {
    which_vec <- lapply(is_data_type, attr, "which_vec")
  }
  is_data_type <- vapply(is_data_type, identity, FUN.VALUE = res_template)
  if (length(res_template) > 1) {
    is_data_type <- as.data.frame(is_data_type)
  }
  attr(is_data_type, "which_vec") <- which_vec
  return(is_data_type)
}

test_that("util_compare_meta_with_study works", {#####
  skip_on_cran()
  study_data <- cars
  meta_data <- prep_create_meta(
    VAR_NAMES = c("speed", "dist"),
    LABEL = c("Speed", "Stopping distance"),
    DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER),
    MISSING_LIST = ""
  )

  expect_equal(util_compare_meta_with_study(study_data, meta_data,
                                            label_col = VAR_NAMES),
               c(speed = 1, dist = 1))

  study_data$speed <- study_data$speed + rnorm(nrow(study_data))
  expect_equal(util_compare_meta_with_study(study_data, meta_data,
                                            label_col = VAR_NAMES),
               c(speed = 0, dist = 1))

  study_data <- cars
  meta_data <- prep_create_meta(
    VAR_NAMES = c("speed", "dist"),
    LABEL = c("Speed", "Stopping distance"),
    DATA_TYPE = c(DATA_TYPES$FLOAT, DATA_TYPES$FLOAT),
    MISSING_LIST = ""
  )
  expect_equal(util_compare_meta_with_study(study_data, meta_data,
                                            label_col = VAR_NAMES),
               c(speed = 1, dist = 1))
  study_data$speed <- study_data$speed + rnorm(nrow(study_data))
  expect_equal(util_compare_meta_with_study(study_data, meta_data,
                                            label_col = VAR_NAMES),
               c(speed = 1, dist = 1))

  # note, that other data types are tested with the function
  # util_check_data_type, that util_compare_meta_with_study bases on

})

#' @name util_deparse1
#'
#' @title Expression De-Parsing
#'
#' @description
#'
#' Turn unevaluated expressions into character strings.
#'
#' @details
#'
#' This is a simple utility function for R < 4.0.0 to ensure a string
#' result (character vector of length one),
#' typically used in name construction, as `util_deparse1(substitute(.))`.
#'
#' This avoids a dependency on `backports` and on R >= 4.0.0.
#'
#' @param expr any R expression.
#' @param collapse a string, passed to `paste()`
#' @param width.cutoff integer in \[20, 500\] determining the
#'                     cutoff (in bytes) at which line-breaking is tried.
#' @param ... further arguments passed to `deparse()`.
#'
#' @return the deparsed expression
#'
#' @family condition_functions
#' @concept process
#' @noRd

NULL # Documentation is assigned to null (@name in line 1)

if (exists("deparse1", envir = baseenv())) {
  util_deparse1 <- base::deparse1
} else {
  util_deparse1 <-
    function (expr,
              collapse = " ",
              width.cutoff = 500L,
              ...) {
      paste(
        deparse(
          expr = expr,
          width.cutoff = width.cutoff,
          ...),
        collapse = collapse
      )
    }
}

#' Extract all properties of a `ReportSummaryTable`
#'
#' @param x `ReportSummaryTable` object
#'
#' @return [list] with all properties
#'
#' @noRd
util_init_respum_tab <- function(x) {
  my_cols <- c("#7f0000", "#b30000", "#d7301f", "#ef6548", "#fc8d59",
               "#fdbb84", "#fdd49e", "#fee8c8", "#2166AC")

  higher_means <- attr(x, "higher_means")
  if (is.null(higher_means)) higher_means <- "worse"
  continuous <- attr(x, "continuous")
  if (is.null(continuous)) continuous <- TRUE
  colcode <- attr(x, "colcode")
  if (is.null(colcode)) {
    continuous <- TRUE
  }
  level_names <- attr(x, "level_names")

  relative <- attr(x, "relative")
  if (is.null(relative)) relative <- continuous
  colscale <- attr(x, "colscale")
  if (is.null(colscale)) {
    colscale <- my_cols
  }

  if (!is.null(attr(x, "flip_mode"))) {
    flip_mode <- attr(x, "flip_mode")
  }

  if (higher_means != "worse")
    colscale <- rev(colscale)

  as.list(environment())
}

#' RStudio crashes on parallel calls in some versions on Darwin based operating
#' systems with R 4
#'
#' @return invisible null
#'
#' @family robustness_functions
#' @concept system
#' @noRd
util_fix_rstudio_bugs <- function() { # nocov start
  if (!util_really_rstudio()) return(invisible(NULL))
  # https://github.com/rstudio/rstudio/issues/6692
  if (Sys.getenv("RSTUDIO") == "1" && !nzchar(Sys.getenv("RSTUDIO_TERM")) &&
    Sys.info()["sysname"] == "Darwin" && getRversion() >= "4.0.0") {
    if (requireNamespace("rstudioapi", quietly = TRUE)) {
      if (rstudioapi::isAvailable()) {
        rstudio_too_old <- rstudioapi::versionInfo()$version < "1.3.1056"
      } else {
        rstudio_too_old <- FALSE
      }
    } else {
      util_user_hint(sprintf(
        paste0("Without the package %s, I cannot decide, if your RStudio is",
               "at least at version 1.3.1056, so I'll activate a work-around",
               "for a known parallel-bug fixed in newer RStudios."),
        dQuote("rstudioapi")
      ))
      rstudio_too_old <- TRUE
    }
    if (rstudio_too_old) {
      util_user_hint(paste0("Enabling workaround for an RStudio bug:",
                             "https://github.com/rstudio/rstudio/issues/6692"))
      if (exists("setDefaultClusterOptions", asNamespace("parallel"),
                 mode = "function")) {
        get("setDefaultClusterOptions", asNamespace("parallel"),
            mode = "function")(setup_strategy = "sequential")
      }
    }
  }
  invisible(NULL)
} # nocov end

#' Compute Indicators for Qualified Item Missingness
#'
#' [Indicator]
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable list] the name of the measurement variables
#' @param expected_observations [enum] HIERARCHY | ALL | SEGMENT. Report the
#'                                     number of observations expected using
#'                                     the old `PART_VAR` concept. See
#'                                     [com_item_missingness] for an
#'                                     explanation.
#'
#' @return A [list] with:
#'   - `SummaryTable`: [data.frame] containing data quality checks for
#'                     "Non-response rate" (`PCT_com_qum_nonresp`) and
#'                     "Refusal rate" (`PCT_com_qum_refusal`) for each response
#'                     variable in `resp_vars`.
#'   - `SummaryData`: a [data.frame] containing data quality checks for
#'                    “Non-response rate” and "Refusal rate"
#'                    for a report
#'
#' @export
com_qualified_item_missingness <-    function(resp_vars,
                                              study_data,
                                              label_col = NULL,
                                              item_level = "item_level",
                                              expected_observations = c("HIERARCHY",
                                                                        "ALL",
                                                                        "SEGMENT"),
                                              meta_data = item_level,
                                              meta_data_v2,
                                              meta_data_segment,
                                              segment_level) {
  # TODO: Allow also here dates values as missing codes
  # TODO (for the example): Add also missing table assignments to the synthetic cases with PART_VAR working following the old implementation

  # Preparation of the input ----

  util_maybe_load_meta_data_v2()

  util_ck_arg_aliases()

  prep_prepare_dataframes(.replace_missings = FALSE)

  expected_observations <- util_match_arg(expected_observations)

  meta_data_segment <- try(prep_check_meta_data_segment(meta_data_segment),
      silent = TRUE)

  if (missing(resp_vars) || length(resp_vars) == 0) {
    resp_vars <- meta_data[[label_col]]
  }

  util_correct_variable_use(resp_vars,
                            allow_more_than_one = TRUE,
                            allow_all_obs_na = TRUE,
                            allow_any_obs_na = TRUE)

  # allowed AAPOR-States
  AAPOR_STATES <- c("I", "P", "PL", "R", "BO", "NC", "O", "UH", "UO", "NE")

  # try to get segment part var names
  seg_pv_vn <- character(0)
  try({
    if (is.data.frame(meta_data_segment) && SEGMENT_PART_VARS %in%
        colnames(meta_data_segment))
      seg_pv_vn <- util_find_var_by_meta(meta_data_segment[[SEGMENT_PART_VARS]])
  }, silent = TRUE)

  # Loop over all resp_vars ----

  aapors <-
    lapply(setNames(nm = resp_vars),
           function(rv) { # rv is the currently examined study variable's name
             rv_vn <- meta_data[meta_data[[label_col]] == rv, # find the variable name
                                VAR_NAMES]
             if (rv %in% colnames(ds1)) { # check, if the status variable really exists in the study data
               .cur_miss <- meta_data[meta_data[[label_col]] == rv, # find the missing code table assigned to the current study variable
                                      MISSING_LIST_TABLE]
               if (length(.cur_miss) == 0) {
                 .cur_miss <- NA_character_ # nocov: Cannot be reached, because if rv is not in the metadata, we fail already in util_correct_variable_use, otherwise, we may have NA in MISSING_LIST_TABLE, but not length(.) == 0
               }
               cur_miss <- try(util_expect_data_frame(x = .cur_miss, # this is up to here the name of the missing table, not the table itself
                                                      dont_assign = TRUE, # don't replace the value of .cur_miss by the data frame, just return it
                                                      col_names = # check/ensure, content of the data frame matches expectations
                                                        list(
                                                          CODE_VALUE = util_is_valid_missing_codes,
                                                          CODE_INTERPRET = function(x) {
                                                            is.character(x) & x %in% AAPOR_STATES
                                                          }
                                                        ),
                                                      custom_errors = list(
                                                        CODE_VALUE = "CODE_VALUE must be numeric or DATETIME",
                                                        CODE_INTERPRET = paste("CODE_INTERPRET must be one of",
                                                                               util_pretty_vector_string(AAPOR_STATES))
                                                      ),
                                                      convert_if_possible = list( # try to convert values to match expectations, if produces NAs, an error is thrown. Thereafter, checks from col_names above will re-run.
                                                        CODE_VALUE = util_as_valid_missing_codes,
                                                        CODE_INTERPRET = function(x) {
                                                          trimws(toupper(x))
                                                        }
                                                      )), silent = TRUE)
               if (inherits(cur_miss, "try-error")) { # some problems with fetching and checking/fixing the data frame by util_expect_data_frame
                 # nocov start: Cannot be reached, because most potential problems will have occurred already in prep_prepare_data_frames > prep_meta_data_v1_to_item_level_meta_data
                 if (!is.na(.cur_miss)) util_warning(
                   "Could not load missing-match-table %s for response variable %s: %s",
                   dQuote(.cur_miss),
                   dQuote(rv),
                   conditionMessage(attr(cur_miss, "condition")),
                   applicability_problem = TRUE) else # nocov end
                     util_warning(
                       "No missing-match-table for response variable %s",
                       dQuote(rv),
                       applicability_problem = TRUE)
                 return(NULL)
               } else { # if a missing table could be properly fetched and is consistent

                 cur_miss <- # remove rows from the missing table, that are explicitly not assigned to the rv
                   util_filter_missing_list_table_for_rv(cur_miss, rv, rv_vn)

                 # really compute stuff ----
                 code_for_P <- head(cur_miss[cur_miss[[CODE_INTERPRET]] == "I",
                                             CODE_VALUE, TRUE], 1) # get the first code for I to replace the real measurements by this code
                 if (length(code_for_P) == 0) {
                   code_for_P <- util_find_free_missing_code(cur_miss$CODE_VALUE)
                   P_df <- data.frame(
                     CODE_VALUE = code_for_P,
                     CODE_LABEL = "Participation",
                     CODE_INTERPRET = "I",
                     stringsAsFactors = FALSE
                   )
                   cur_miss <- util_rbind(cur_miss,
                                          P_df)
                   code_for_P <- head(cur_miss[cur_miss[[CODE_INTERPRET]] == "I",
                                               CODE_VALUE, TRUE], 1) # get the first code for I to replace the real measurements by this code
                 }
                 if (length(code_for_P) != 1) {
                   # nocov start: Cannot be reached, because if there would've been no I code, we would have one after the last block. If we had > 1, we would have chosen the first one already, wlog
                   util_error(
                     applicability_problem = TRUE,
                     c("Could not find any participation related code in %s %s",
                       "for qualified missingness of item %s"),
                     sQuote(MISSING_LIST_TABLE),
                     dQuote(.cur_miss),
                     dQuote(rv)
                   )
                   # nocov end
                 }

                 rx <- ds1[[rv]]

                 rx[!is.na(rx) & !(rx %in% cur_miss[cur_miss[[CODE_INTERPRET]] != "I",
                                                    CODE_VALUE, TRUE])] <- code_for_P

                 rx <- suppressWarnings(as.numeric(rx)) # TODO: Handle dates, handle errors

                 r <- util_table_of_vct(factor( # counts the number of "CODE_INTERPRED"/AAPOR letters for the observations, by converting the value to a factor and tabulating it
                   rx, # the (not-yet) AAPOR codes
                   levels = cur_miss[[CODE_VALUE]], # TODO: Streamline with prep_extract_cause_label_df and prep_add_cause_label_df # CODES used in the rv are on the missing-code table in the column CODE_VALUE
                   labels = cur_miss[[CODE_INTERPRET]]))

                 r <- as.list(setNames(r$Freq, nm = r$Var1)) # convert the data frame with columns Var1 and Freq to a named list, names are the AAPOR codes, values are the frequencies

                 r$N <- sum(!util_empty(ds1[[rv]]))
                 r$N2 <- util_count_expected_observations(rv,
                                                          study_data = ds1,
                                                          meta_data = meta_data,
                                                          label_col = label_col,
                                                          expected_observations =
                                                            expected_observations)

                 if (length(r$P) == 1 && !is.na(r$P) && r$P != 0) {
                   # find out, if the variable is in segment part vars
                   if (is.data.frame(meta_data_segment) &&
                       rv_vn %in% seg_pv_vn) {
                      # this is still ok, it's a part var
                   } else {
                     util_warning(
                       c("%0.2f%% of the values in %s have a missing code for",
                         "%s assigned, which is impossible on item level. Treating",
                         "these cases as %s"),
                       round(r$P / r$N * 100, 2),
                       sQuote(rv),
                       sQuote("P"),
                       sQuote("UO"),
                       applicability_problem = TRUE)
                     if (length(r$UO) == 1 && !is.na(r$UO) && r$UO != 0) {
                       r$UO <- r$UO + r$P
                     } else {
                       r$UO <- r$P
                     }
                     r$P <- 0
                   }
                 }


                 # Generate replacement list
                 rclean <- r

                 # Enter 0 for NULL in replacement list to always enable computations
                 for (name in AAPOR_STATES) {
                   if (is.null(rclean[[name]])) {
                     rclean[[name]] <- 0
                   }
                 }

                 # Compute response, denial, ... rates ----

                 # DQI-2002    Based on participation only
                 # NRR1 for all who did not participate
                 # Minimal information: Study, Segment: (I | P ) -> NOTE NE not used in denominator
                 # RR1<-(I+P)/((I+P+PL) + (R+BO+NC+O) + (UH+UO)
                 if(!is.null(r$I) | !is.null(r$P) ) {
                   r$RR1 <- (rclean$I + rclean$P) / ((rclean$I+rclean$P+rclean$PL) + (rclean$R+rclean$BO+rclean$NC+rclean$O) + (rclean$UH+rclean$UO))
                   r$NRR1 <- 1 - r$RR1
                   r$PCT_com_qum_nonresp <- 100 * r$NRR1
                 } else {
                   # nocov start
                   # we will always have at least I == 0, because this is ensured above.
                   util_warning(
                     "Preconditions to generate Nonresponse Rate 1 not given for study variable %s",
                     dQuote(rv),
                     applicability_problem = TRUE)
                   r$NRR1 <- NA_real_
                   r$RR1 <- NA_real_
                   r$PCT_com_qum_nonresp <- NA_real_
                   # nocov end
                 }

                 # NRR2 based on participation and scheduled dates
                 # Minimal information: Study, Segment: (I | P | PL ) -> NOTE NE not used in denominator
                 # RR2<-(I+P+PL)/((I+P+PL) + (R+BO+NC+O) + (UH+UO)
                 if(!is.null(r$I) | !is.null(r$P)  | !is.null(r$PL) ) {
                   r$RR2 <- (rclean$I + rclean$P + rclean$PL) / ((rclean$I+rclean$P+rclean$PL) + (rclean$R+rclean$BO+rclean$NC+rclean$O) + (rclean$UH+rclean$UO))
                   r$NRR2 <- 1 - r$RR2
                 } else {
                   # nocov start
                   # we will always have at least I == 0, because this is ensured above.
                   util_warning(
                     "Preconditions to generate Nonresponse Rate 1 not given for variable %s",
                     dQuote(rv),
                     applicability_problem = TRUE)
                   r$NRR2 <- NA_real_
                   r$RR2 <- NA_real_
                   # nocov end
                 }


                 # DQI-2003    Refusal rate
                 # NRR1 for all who refused at any stage
                 # Minimal information: Study, Segment: R | BO
                 # REF1<-(R+BO)/((I+P+PL) + (R+BO+NC+O) + (UH+UO)
                 if(!is.null(r$R) | !is.null(r$BO) ) {
                   r$REF1 <- (rclean$R + rclean$BO) / ((rclean$I+rclean$P+rclean$PL) + (rclean$R+rclean$BO+rclean$NC+rclean$O) + (rclean$UH+rclean$UO))
                   r$PCT_com_qum_refusal <- 100 * r$REF1
                 } else {
                   util_warning(
                     "Preconditions to generate Nonresponse Rate 1 not given for variable %s", # TODO: Write nicer warnings, here.
                     dQuote(rv),
                     applicability_problem = TRUE)
                   r$REF1 <- NA_real_
                   r$PCT_com_qum_refusal <- NA_real_
                 }

                 r <- cbind( # create a new data frame with a column Variables and all the rates and values for the current rv 1 row, only.
                   data.frame(Variables = rv, stringsAsFactors = FALSE, check.names = FALSE, row.names = NULL),
                   data.frame(r, stringsAsFactors = FALSE, check.names = FALSE, row.names = NULL)
                 )
               }
             } else { # study data lacks the status variable for the  current variable, omit this segment
               # nocov start
               # unreachabel, becasue if rv would not be in colnames(ds1), util_correct_variable_use would have stopped, already
               util_warning("Missing variable %s from %s",
                            dQuote(rv),
                            sQuote("study_data"),
                            applicability_problem = TRUE,
                            intrinsic_applicability_problem = TRUE)
               return(NULL)
               # nocov end
             }
           })

  SummaryTable <- util_rbind(data_frames_list = aapors) # Combine all the 1-row-tables to a large rate-table
  rownames(SummaryTable) <- NULL

  # --- Start refactored part: use helpers for selection + translation ---
  # Select only indicator metric columns from SummaryTable
  indicator_cols <- util_extract_indicator_metrics(SummaryTable)

  # Assemble SummaryData = Variables + indicator columns
  SummaryData <- cbind(
    data.frame(Variables = SummaryTable[["Variables"]], stringsAsFactors = FALSE, check.names = FALSE, row.names = NULL),
    indicator_cols
  )

  # Format percentage columns BEFORE renaming (we rely on original "PCT_" prefixes)
  pct_cols <- colnames(SummaryData)[startsWith(colnames(SummaryData), "PCT_")]
  if (length(pct_cols)) {
    SummaryData[, pct_cols] <- lapply(SummaryData[, pct_cols, drop = FALSE],
                                      function(cl) paste0(round(cl, 2), "%"))
  }

  # Translate column names using the canonical helper; keep unknown names as-is
  colnames(SummaryData) <- util_translate_indicator_metrics(
    colnames(SummaryData),
    short = FALSE,
    long  = TRUE,
    ignore_unknown = TRUE
  )
  # --- End refactored part ---

  # SummaryTable$GRADING <- SummaryTable[["PCT_com_qum_nonresp"]] > 10 # only to avoid empty output in issue matrix, can be removed, now.

  text_to_display <- util_get_hovertext("[com_qualified_item_missingness_hover]")
  attr(SummaryData, "description") <- text_to_display

  return(list( # return the results
    SummaryTable = SummaryTable,
    SummaryData = SummaryData
  ))
}

test_that("com_qualified_item_missingness works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  r1 <- com_qualified_item_missingness(resp_vars = "SBP_0",
                                 study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                                 label_col = LABEL)
  expect_identical(dim(r1$SummaryTable), c(1L, 15L))
  suppressWarnings(expect_warning(
    r2 <- com_qualified_item_missingness(study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                                   label_col = LABEL),
    regexp = "No missing-match-table"
  ))
  expect_identical(dim(r2$SummaryTable), c(36L, 15L))

  mt <- prep_get_data_frame("missing_table")
  mt <- mt[!(mt$CODE_INTERPRET %in% c("I", "P", "PL")), , FALSE] # for qual.segment.missingness, this would trigger warnings
  prep_add_data_frames(missing_table = mt)
  r3 <- com_qualified_item_missingness(resp_vars = "SBP_0",
                                       study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                                       label_col = LABEL)

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  mt <- prep_get_data_frame("missing_table")
  mt <- mt[!(mt$CODE_INTERPRET %in% c("R", "BO")), , FALSE]
  prep_add_data_frames(missing_table = mt)
  expect_warning(
    r4 <- com_qualified_item_missingness(resp_vars = "SBP_0",
                                         study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                                         label_col = LABEL),
    regexp = ".*Nonresponse Rate 1.*"
  )

})

# devtools::test_coverage_active_file()

#' View a file in most suitable viewer
#'
#' @param file the file to view
#'
#' @return
#' `invisible(file)`
#'
#' @family system_functions
#' @concept reporting
#' @noRd
util_view_file <- function(file) {
  viewer <- getOption("viewer")
  if (is.null(viewer)) { # nocov start
    if (!identical(Sys.getenv("TESTTHAT"), "true")) {
      # Viewer stuff cannot be tested w/o a GUI
      if (util_really_rstudio() &&
          requireNamespace("rstudioapi", quietly = TRUE) &&
          (exists("viewer", asNamespace("rstudioapi"), mode = "function")) &&
          (exists("isAvailable", asNamespace("rstudioapi"), mode = "function")) &&
          rstudioapi::isAvailable()) {
        rstudioapi::viewer(file)
      } else {
        browseURL(paste0("file://", normalizePath(file)))
      }
    }
  } else { # nocov end
    viewer(file)
  }
  invisible(file)
}

util_works_in_rs_viewer <- function(fn) {
  if (!util_really_rstudio()) return(FALSE)
  is_existing_path_under_tempdir <- function(fn) {
    stopifnot(is.character(fn), length(fn) == 1L, nzchar(fn))
    if (!file.exists(fn)) return(FALSE)

    td <- normalizePath(tempdir(), winslash = "/", mustWork = TRUE)
    fn_real <- normalizePath(fn, winslash = "/", mustWork = TRUE)

    td_prefix <- paste0(td, "/")
    identical(fn_real, td) || startsWith(fn_real, td_prefix)
  }
  is_existing_path_under_tempdir(fn)
}

#' @noRd
util_parallel_classic <- function(all_calls,
                                  worker,
                                  n_nodes,
                                  progress,
                                  debug_parallel,
                                  my_storr_object) {
  outer_env <- parent.frame()
  # was length(parallel::getDefaultCluster()), but parallelMap doesn't use
  # defaultcluster any more!!
  tasks_per_node <- ceiling(length(all_calls) / n_nodes)
  indices <- 1:length(all_calls)
  length(indices) <- n_nodes * tasks_per_node # make equal length
  task_matrix <- matrix(indices, ncol = n_nodes, nrow = tasks_per_node,
                        byrow = TRUE)

  function_names <- lapply(all_calls, rlang::call_name)

  if (getOption("parallelMap.mode") %in% c("BatchJobs", "batchtools")) {
    task_matrix <- matrix(seq_along(all_calls), nrow = 1) # for batch jobs, don't split
  }

  r <- unlist(lapply(
    1:nrow(task_matrix),
    function(row) {
      slices <- task_matrix[row, ]
      slices <- slices[!is.na(slices)]
      if (length(all_calls))
        progress(100 * row / nrow(task_matrix))
      if (!dynGet(".is_testing", ifnotfound = FALSE)) {
        util_message(
          sprintf("%s [%s] %d of %d, %s", Sys.time(), "INFO",
                  row, nrow(task_matrix), "DQ")
        ) # TODO: Use RStudio job if available
      }
      if (debug_parallel) {
        # TODO: log something about the current chunk
      }
      acs <- all_calls[slices]
      R.devices::suppressGraphics(
        # don't use any auto graphics device (needed for certain
        # parallelization methods)
        parallelMap::parallelMap(impute.error = identity,
                                    setNames(nm = names(acs), seq_along(acs)),
                                    function_name = function_names[slices],
                                    fun = function(cl, acs, ...) {
                                      worker(acs[[cl]], ...)
                                    },
                                    more.args = list(
                                      env = outer_env,
                                      my_storr_object = my_storr_object,
                                      acs = acs
                                    ),
                                 simplify = FALSE,
                                 use.names = TRUE,
                                    nm = names(acs))
      )
    }
  ), recursive = FALSE)

  r
}

#' Instantiate a new metadata file
#'
#' @param file_name [character] file path to write to
#' @param study_data [data.frame] optional, study data to guess metadata from
#' @param open [logical] open the file after creation
#' @param overwrite [logical] overwrite `file`, if exists
#'
#' @return `invisible(NULL)`
#' @export
prep_create_meta_data_file <- function(file_name,
                                       study_data,
                                       open = TRUE,
                                       overwrite = FALSE) {
  util_expect_scalar(file_name,
                     check_type = is.character,
                     error_message =
                       sprintf("%s needs to be a character string.",
                               sQuote("file_name")))
  util_expect_scalar(open,
                     check_type = is.logical,
                     error_message =
                       sprintf("%s needs to be logical",
                               sQuote("open")))

  util_expect_scalar(overwrite,
                     check_type = is.logical,
                     error_message =
                       sprintf("%s needs to be logical",
                               sQuote("overwrite")))

  if (!missing(study_data)) {
    util_expect_data_frame(study_data, keep_types = TRUE)
  }

  if (!overwrite && file.exists(file_name)) {
    cll <- sys.call()
    if (identical(as.character(cll[[1]]), "prep_create_meta_data_file")) {
      # not called as dataquieR::prep_create_meta_data_file, but
      # as prep_create_meta_data_file()
      prefix <- paste0(packageName(), "::")
    } else {
      prefix <- ""
    }
    cmd <- paste0(deparse(nlines = -1,
                   rlang::call_modify(cll,
                                      overwrite = TRUE)),
                  collapse = "\n")
    if (suppressWarnings(util_ensure_suggested("cli", err = FALSE))) {
      util_error(
        cli::cli_text(
          sprintf(
"%s already exists, will not overwrite. Call \n{.run [%s](%s%s)}\n to overwrite.",
    dQuote(file_name), cmd, prefix, cmd))
      )
    } else {
      util_error(
        "%s already exists, will not overwrite. Call \n%s\n to overwrite.",
        dQuote(file_name),
        cmd
      )
    }
  }
  mdl <- util_rio_import_list("https://dataquality.qihs.uni-greifswald.de/extdata/meta_data_v2.xlsx",
                              keep_types = FALSE)
  old_mlts <- unique(mdl[["item_level"]][[MISSING_LIST_TABLE]])
  old_id_rft <- unique(c(mdl$segment_level$SEGMENT_ID_REF_TABLE,
                         mdl$dataframe_level$DF_ID_REF_TABLE))
  if (!missing(study_data)) {
    old_cache <- prep_list_dataframes()
    mdl[["item_level"]] <-
      prep_study2meta(study_data = study_data)

    mdl[["item_level"]] <-
      prep_meta_data_v1_to_item_level_meta_data(
        meta_data = mdl[["item_level"]],
        verbose = FALSE)

    mdl[["item_level"]][[MISSING_LIST]] <- NULL
    mdl[["item_level"]][[JUMP_LIST]] <- NULL

    mlts <- unique(mdl[["item_level"]][[MISSING_LIST_TABLE]])

    for (mlt in mlts) {
      if (!is.na(mlt)) try(mdl[[mlt]] <- prep_get_data_frame(mlt),
                           silent = TRUE)
    }
    new_cache <- prep_list_dataframes()
    rm(list = setdiff(new_cache, old_cache), envir = .dataframe_environment())
  }
  mdl[old_mlts] <- NULL
  mdl[old_id_rft] <- NULL
  for (sheet in
       setdiff(grep("_level$", value = TRUE, names(mdl)), c("item_level"))) {
    mdl[[sheet]] <- mdl[[sheet]][FALSE, , FALSE]
  }

  e <- try(rio::export(mdl, file = file_name),
           silent = TRUE)
  if (inherits(e, "try-error")) {
    util_error("Could not write %s: %s",
               dQuote(file_name),
               conditionMessage(attr(e, "condition")))
  }
  if (open) {
    browseURL(file_name)
  }
  invisible(NULL)
}

test_that("multiplication works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  d <- withr::local_tempdir()
  withr::local_dir(d)
  prep_create_meta_data_file("Test.xlsx",
                             cars,
                             open = FALSE,
                             overwrite = TRUE)
  expect_gt(file.size("Test.xlsx"), 1000)
  md0 <- prep_get_data_frame("Test.xlsx|item_level")
  expect_snapshot(md0)
})

#' Pre-load a file with named (usually more than) one table(s)
#'
#' These can thereafter be referred to by their names only. Such files are,
#' e.g., spreadsheet-workbooks or `RData`-files.
#'
#' Note, that this function in contrast to [prep_get_data_frame] does neither
#' support selecting specific sheets/columns from a file.
#'
#' @param file the file name to load.
#' @param keep_types [logical] keep types as possibly defined in the file.
#'                             set `TRUE` for study data.
#' @param append [logical] if a data frame already exists in the cache
#'                         (by name), extend the existing one
#'
#' @return `invisible(the cache environment)`
#' @export
#' @seealso [prep_add_data_frames]
#' @seealso [prep_get_data_frame]
#' @family data-frame-cache
prep_load_workbook_like_file <- function(file,
                                         keep_types = FALSE,
                                         append = FALSE) {
  util_expect_scalar(append, check_type = is.logical)
  if (!missing(keep_types)) {
    util_expect_scalar(keep_types,
                       check_type = is.logical,
                       error_message =
                         sprintf("%s needs to be to be a logical value.",
                                 sQuote("keep_types")))
  }

  util_expect_scalar(file, check_type = is.character,
                     allow_more_than_one = TRUE)

  if (length(file) > 1 && (append || all(vapply(file, function(fn) {
      (tolower(tools::file_ext(trimws(fn))) %in% c("odm", "xml") &&
             file.exists(fn) && !dir.exists(fn) &&
             util_is_odm_xml(fn))
    }, FUN.VALUE = logical(1))))) {
    if (!append) {
    }
    for (fn in file) {
      prep_load_workbook_like_file(file = fn,
                                   keep_types = keep_types,
                                   append = TRUE)
    }
    return(invisible(.dataframe_environment()))
  }

  if (length(file) > 1) {
    util_error(
      c("Only for ODM files, we assume append = TRUE,",
        "if more than one is read at the same time"))
  }

  if (file %in% c("meta_data_v2",
                  "ship_meta_dataframe",
                  "ship_meta_v2")) {
    util_stop_if_not(
      `no shortcuts in tests` =
        !identical(Sys.getenv("TESTTHAT"), "true"))
    file <- sprintf(
      "https://dataquality.qihs.uni-greifswald.de/extdata/%s.xlsx",
      file)
  }

  if (startsWith(file, "https://") ||
      startsWith(file, "http://")) {
    fp <- tempfile()
    util_stop_if_not(!file.exists(fp))
    dir.create(fp)
    on.exit(try(unlink(fp, force = TRUE, recursive = TRUE, expand = FALSE),
                silent = TRUE))
    file_dec <- utils::URLdecode(file)
    fn <- gsub("^.*\\/", "", file_dec, perl = TRUE)
    fn <- gsub("\\?.*$", "", fn)
    fn <- gsub("#.*$", "", fn)

    ext <- ""
    ext <- try(util_fetch_ext(file), silent = TRUE)
    # do not ignore content-disposition headers sent by the server (if they propose a file name)
    if (length(ext) != 1 ||
        !is.character(ext)) {
      msg <- "unknown reason"
      if (inherits(ext, "try-error")) {
        msg <- conditionMessage(attr(ext, "condition"))
      } else if (inherits(ext, "condition")) {
        msg <- conditionMessage(ext)
      }
      util_warning("Could not determine the file type of %s: %s",
                   dQuote(file),
                   sQuote(msg))
      ext <- ""
    } else {
      if (!is.null(attr(ext, "file-name"))) {
        if (length(attr(ext, "file-name")) == 1 &&
            !is.na(attr(ext, "file-name")))
          fn <- attr(ext, "file-name")
      }
      ext <- paste0(".", ext)
    }

    if (!endsWith(fn, ext)) {
      fn <- paste0(fn, ext)
    }

    file_new <- file.path(fp, fn)

    try(utils::download.file(file, destfile = file_new, quiet = TRUE, mode = "wb"), silent = TRUE)
    if (file.exists(file_new)) {
      file <- file_new
    }
  }

  fn <- file

  r <- util_rio_import_list(fn, keep_types = keep_types)

  if (inherits(r, "try-error") || !is.list(r) ||
      any(vapply(r, is.data.frame, FUN.VALUE = logical(1)))) {
    fn0 <- system.file("extdata", paste0(fn, ".RDS"),
                         package = "dataquieR")
    if (file.exists(fn0)) {
      fn <- fn0
    } else {
      fn0 <- system.file("extdata", paste0(fn, ".RData"),
                           package = "dataquieR")
      if (file.exists(fn0)) {
        fn <- fn0
      } else {
        fn0 <- system.file("extdata", paste0(fn, ".xlsx"),
                             package = "dataquieR")
        if (file.exists(fn0)) {
          fn <- fn0
        }
      }
    }
    r <- util_rio_import_list(fn, keep_types = keep_types)
  }

  if (inherits(r, "try-error") || !is.list(r) ||
      any(
        !vapply(r, is.null, FUN.VALUE = logical(1))
        &
        !vapply(r, is.data.frame, FUN.VALUE = logical(1)))) {
    if (inherits(r, "try-error")) {
      error <- gsub("%", "%%", fixed = TRUE,
                          conditionMessage(attr(r, "condition")))
    } else if (file.access(fn, 4)) { # file.access has inverse result logics
      error <- "Access to denied"
    } else if (!file.exists(fn)) {
      error <- "File not found"
    } else if (dir.exists(fn)) {
      error <- "It is a folder"
    } else if (any(
      !vapply(r, is.null, FUN.VALUE = logical(1))
      &
      !vapply(r, is.data.frame, FUN.VALUE = logical(1)))) {
      error <-
        "File does not feature a workbook-like structure, i.e., a named list of data frames"
    } else {
      error <- "Unkonwn reasons."
    }
    util_error("Could not load/find file %s: %s", dQuote(file),
               error)
  }
  data_frame_list <- r

# addition to load a csv file
  if (is.null(names(data_frame_list)) && length(data_frame_list) == 1) {
    names(data_frame_list) <- paste0(basename(fn))
  }

  prep_add_data_frames(data_frame_list = data_frame_list, append = append)
  data_frame_list2 <- data_frame_list
  names(data_frame_list2) <- paste0(basename(fn),
                                    SPLIT_CHAR,
                                    names(data_frame_list2))
  r <- prep_add_data_frames(data_frame_list = data_frame_list2, append = append)
  trunc_name <- basename(fn)
  if (all(grepl("\\.(RDS|RData|xlsx)$", trunc_name, perl = TRUE))) {
    trunc_name <- gsub("\\.(RDS|RData|xlsx)$", "", trunc_name, perl = TRUE)
    data_frame_list3 <- data_frame_list
    names(data_frame_list3) <- paste0(trunc_name,
                                      SPLIT_CHAR,
                                      names(data_frame_list3))
    r <- prep_add_data_frames(data_frame_list = data_frame_list3,
                              append = append)
  }
  invisible(r)
}

test_that("test-int_sts_element_dataframe works", {
  skip_on_cran() # online, fragile
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx")
  sd0 <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS", keep_types = TRUE)
  md0 <- prep_get_data_frame("item_level")
  dl0 <- prep_get_data_frame("dataframe_level")
  dl0$DF_NAME <- "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS"
  md0 <- md0[!(md0$VAR_NAMES %in% c("sex")), , FALSE] # remove a variable
  ised <- int_sts_element_dataframe(item_level = md0, meta_data_dataframe = dl0)
  expect_snapshot_value(ised, style = "deparse")
})

# tests/testthat/test-util_generate_pages_from_report-coverage.R

.get_mini_report <- function() {
  env <- testthat::test_env()

  if (!is.null(env$.mini_report)) {
    return(env$.mini_report)
  }

  skip_on_cran() # slow, errors unlikely
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  dataquieR::prep_load_workbook_like_file(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx"
  )

  withr::defer(
    dataquieR::prep_purge_data_frame_cache(),
    envir = env
  )

  env$.mini_report <- dataquieR::dq_report2(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
    resp_vars = c("SEX_0", "AGE_0"),
    dimensions = c("int"),
    label_col = LABEL,
    cores = NULL
  )

  env$.mini_report
}

.call_pages <- function(report, ...) {
  dataquieR:::util_generate_pages_from_report( # nolint
    report = report,
    template = "default",
    disable_plotly = TRUE,

    # required + avoid recursive defaults
    progress = function(pct) invisible(NULL),
    progress_msg = function(...) invisible(NULL),

    block_load_factor = 1,
    dir = tempdir(),
    my_dashboard = NULL,

    ...
  )
}

test_that("util_generate_pages_from_report(): renders pages for a mini report", {
  skip_if_not_installed("stringdist")
  report <- .get_mini_report()

  pages <- .call_pages(report)
  expect_true(is.list(pages))
  expect_true(length(pages) >= 1)

  html <- paste(unlist(pages, recursive = TRUE), collapse = "\n")
  expect_true(nchar(html) > 0)
})

test_that("util_generate_pages_from_report(): label modifications section (if present) is rendered", {
  skip_if_not_installed("stringdist")
  report <- .get_mini_report()

  pages <- .call_pages(report)
  html <- paste(unlist(pages, recursive = TRUE), collapse = "\n")

  txt <- attr(report, "label_modification_text")
  if (!is.null(txt) && length(txt) == 1L && nchar(txt) > 0) {
    expect_match(html, "Label modifications", fixed = TRUE)
  } else {
    expect_true(TRUE) # branch executed, but nothing to assert
  }
})

test_that("util_generate_pages_from_report(): integrity issues before pipeline (if present) are rendered", {
  skip_if_not_installed("stringdist")
  report <- .get_mini_report()

  pages <- .call_pages(report)
  html <- paste(unlist(pages, recursive = TRUE), collapse = "\n")

  issues <- attr(report, "integrity_issues_before_pipeline")
  if (!is.null(issues) && length(issues) > 0) {
    expect_match(html, "Integrity", fixed = FALSE)
  } else {
    expect_true(TRUE)
  }
})

test_that("util_generate_pages_from_report(): falls back when properties are missing", {
  skip_if_not_installed("stringdist")
  report <- .get_mini_report()

  # modify a copy to avoid polluting the cached object
  report2 <- report
  attr(report2, "properties") <- NULL

  pages <- .call_pages(report2)
  html <- paste(unlist(pages, recursive = TRUE), collapse = "\n")

  expect_match(html, "No report properties found", fixed = TRUE)
})

test_that("util_generate_pages_from_report(): errors if parallelMap.mode is not local", {
  skip_if_not_installed("stringdist")
  report <- .get_mini_report()
  withr::local_options(list(parallelMap.mode = "snow"))

  expect_error(
    .call_pages(report),
    "parallel rendering of reports is not supported",
    fixed = TRUE
  )
})

test_that("dq_report_by works with pipe list in argum", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_on_cran() # slow test
  target <- withr::local_tempdir("testdqareportby_pipe")

  study_data <- head(
    prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE),
    10)

  study_data <- study_data[setdiff(names(study_data), c("v00035", "v00036",
                                                        "v00037","v00038",
                                                        "v00039","v00040",
                                                        "v00041"))]

  prep_load_workbook_like_file( "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  item_level <- prep_get_data_frame("item_level")
  item_level <- item_level[setdiff(names(item_level), c("HARD_LIMITS",
                                                        "DETECTION_LIMITS",
                                                        "SOFT_LIMITS",
                                                        "DISTRIBUTION",
                                                        "DECIMALS",
                                                        "DATA_ENTRY_TYPE" , "GROUP_VAR_DEVICE" ,
                                                        "GROUP_VAR_OBSERVER" ,
                                                        "ELEMENT_HOMOGENITY_CHECKTYPE",
                                                        "UNIVARIATE_OUTLIER_CHECKTYPE",
                                                        "N_RULES",
                                                        "LOCATION_METRIC",
                                                        "LOCATION_RANGE",
                                                        "PROPORTION_RANGE",
                                                        "REPEATED_MEASURES_VARS" ,
                                                        "REPEATED_MEASURES_GOLDSTANDARD",
                                                        "CO_VARS" ))]

  item_level<- item_level[!item_level$VAR_NAMES %in% c("v00035", "v00036",
                                                       "v00037","v00038",
                                                       "v00039","v00040",
                                                       "v00041"), ]


  expect_message2(dq_report_by(study_data = study_data,
                              dimensions = "int",
                              cores = NULL,
                              segment_column = STUDY_SEGMENT,
                              segment_select ="STUDY | LAB",
                              item_level = item_level,
                              subgroup = "[v00003]<48",
                              strata_column = "SEX_0",
                              strata_select ="0|1",
                              #meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "m1"),
                              also_print = TRUE))

  prep_load_workbook_like_file( "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  study_data[is.na(study_data[, "v00008"]), "v00008"] <- "B"
  expect_warning(dq_report_by(study_data = study_data,
                              dimensions = "int",
                              cores = NULL,
                              segment_column = STUDY_SEGMENT,
                              segment_exclude ="LAB | INTERVIEW | QUESTIONNAIRE | PHYS_EXAM",
                              item_level = item_level,
                              strata_column = "VO2_CAPCAT_0",
                              strata_exclude ="B |C| D ",
                              selection_type = "value",
                              #meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "m2"),
                              also_print = TRUE),
                 regexp = "The stratum.*strata.*D.*not present in the study data")

  prep_load_workbook_like_file( "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  expect_message2(dq_report_by(study_data = c("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS",
                                             "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
                              meta_data_v2 ="https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "m5"),
                              cores = NULL,
                              also_print=TRUE,
                              segment_column = STUDY_SEGMENT,
                              segment_exclude = c("PHYS_EXAM", "LAB",
                                                  "INTERVIEW", "QUESTIONNAIRE"),
                              strata_column = "v00002",
                              strata_select = "females",
                              dimensions = "Integrity",
                              selection_type = "v_label",
                              subgroup = "[v00003]<40",
                              id_vars = "v00001|v00000"))

})

test_that("test_create_descriptive_summ", {

  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  target <- withr::local_tempdir("testdessummary")

  sd1 <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE),
              20)

  expect_message2(
    des_summary(study_data = sd1,
                meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx"))

  sd1 <- sd1[, 4:7]
  desc1 <- des_summary(study_data = sd1)
  expect_equal(sum(as.numeric(desc1$SummaryData$Mean)),
                297.9)

  desc2 <- des_summary(resp_vars = c("v00003", "v00004"),
                   study_data = sd1)
  expect_equal(sum(as.numeric(desc2$SummaryData$Mean)),
                171.4)
  expect_equal(sum(as.numeric(desc2$SummaryData$SD)),
               10.93)
  expect_equal(sum(as.numeric(desc2$SummaryData$CV)),
               13.27)
  expect_equal(sum(as.numeric(desc2$SummaryData$Kurtosis)),
               -2.5740)
  expect_equal(sum(as.numeric(desc2$SummaryData$Median)),
               171)

})

test_that("test_create_descriptive_summ_to", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                                    keep_types = TRUE)

  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  meta_data <- prep_get_data_frame("item_level")

  set.seed(12345)
  day_course <- unlist(lapply(
    0:23,
    function(h) {
      list(
        hms::hms(hours = h,
                 minutes = 3),
        hms::hms(hours = h,
                 minutes = 13),
        hms::hms(hours = h,
                 minutes = 24),
        hms::hms(hours = h,
                 minutes = 30),
        hms::hms(hours = h,
                 minutes = 50)
      )
    }
  ), recursive = FALSE)
  probs <-
    rep(c(.7, .2, .05, .04, 0.01), 24)
  times <- sample(x = day_course,
                  prob = probs,
                  size = nrow(study_data),
                  replace = TRUE)
  study_data$v02000 <- times
  meta_data <- util_rbind(
    meta_data,
    data.frame(
      stringsAsFactors = FALSE,
      VAR_NAMES = "v02000",
      LABEL = "ADMIS_TM_0",
      DATA_TYPE = DATA_TYPES$TIME,
      SCALE_LEVEL = SCALE_LEVELS$INTERVAL,
      VALUE_LABELS = NA_character_,
      STANDARDIZED_VOCABULARY_TABLE = NA_character_,
      MISSING_LIST_TABLE = NA_character_,
      HARD_LIMITS = "[09:00:00;18:00:00]",
      DETECTION_LIMITS = NA_character_,
      SOFT_LIMITS = NA_character_,
      DISTRIBUTION = NA_character_,
      DECIMALS = NA_character_,
      DATA_ENTRY_TYPE = NA_character_,
      GROUP_VAR_OBSERVER = NA_character_,
      GROUP_VAR_DEVICE = NA_character_,
      TIME_VAR = NA_character_,
      STUDY_SEGMENT = "STUDY",
      PART_VAR = "PART_STUDY",
      VARIABLE_ROLE = "intro",
      VARIABLE_ORDER = "54",
      LONG_LABEL = "Admission time",
      ELEMENT_HOMOGENITY_CHECKTYPE = NA_character_,
      UNIVARIATE_OUTLIER_CHECKTYPE = NA_character_,
      N_RULES = "4",
      LOCATION_METRIC = NA_character_,
      LOCATION_RANGE = NA_character_,
      PROPORTION_RANGE = NA_character_,
      REPEATED_MEASURES_VARS = NA_character_,
      REPEATED_MEASURES_GOLDSTANDARD = NA_character_,
      CO_VARS = NA_character_
    )
  )

  r <-
    des_summary(
      resp_vars = "ADMIS_TM_0",
      study_data = study_data,
      label_col = LABEL,
      meta_data = meta_data)

})

test_that("dq_report2 works", {
  skip_if_not_installed("DT")
  skip_if_not_installed("stringdist")
  skip_if_not_installed("markdown")
  skip_on_cran() # slow, parallel, ...

  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  study_data <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), 100)
  meta_data <- prep_get_data_frame("item_level")

  mlt <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx| missing_table")

  prep_purge_data_frame_cache()

  prep_add_data_frames(`missing_table` = mlt)

  invisible(testthat::capture_output_lines(gc(full = TRUE, verbose = FALSE)))

  sd0 <- study_data[, 1:5]
  sd0$v00012 <- study_data$v00012
  md0 <- subset(meta_data, VAR_NAMES %in% colnames(sd0))
  md0$PART_VAR <- NULL

  # md0$MISSING_LIST_TABLE <- NULL

  # don't include huge reports as RData in the package
  # Suppress warnings since we do not test dq_report2
  # here in the first place
  report <- dq_report2(sd0, md0,
                       resp_vars = c("v00000", "v00001", "v00002",
                                     "v00003", "v00004", "v00012"),
                       filter_indicator_functions =
                         c("^com_item_missingness$",
                           "^acc_varcomp$"),
                       filter_result_slots =
                         c("^SummaryTable$"),
                       cores = NULL,
                       dimensions = # for speed, omit Accuracy
                         c("Integrity",
                           "Completeness",
                           "Consistency",
                           "Accuracy"))

  sts <- report[, "com_item_missingness", "SummaryTable"]

  expect_equal(sum(sts$SummaryTable$GRADING), 1)

  expect_silent(invisible(summary(report)))

  r <- report
  r$acc_varcomp_observer.SBP_0$SummaryTable <- NULL
  expect_silent(summary(r))

  r <- report
  r$acc_varcomp_observer.SBP_0$SummaryTable <-
    r$acc_varcomp_observer.SBP_0$SummaryTable[FALSE, , FALSE]
  expect_silent(summary(r))

  expect_error(
    report <-
      suppressWarnings(suppressMessages(dq_report2(sd0, md0,
                                 cores = NULL,
                                 dimensions = 42))
      ),
    regexp =
      sprintf("The argument %s must be character or NULL",
              sQuote("dimensions")),
    perl = TRUE
  )


  expect_warning(
    report <-
      (dq_report2(sd0, md0,
                  resp_vars = c("v00000", "v00001", "v00002",
                                "v00003", "v00004", "v00012"),
                  filter_indicator_functions =
                    c("^com_item_missingness$",
                      "^acc_varcomp$"),
                  filter_result_slots =
                    c("^SummaryTable$"),
                  cores = NULL,
                   dimensions = c("invalid"),
      )),
    regexp =
      paste("(?ms)Missing",
            ".+invalid.+from.+Completeness.+Consis.+Accuracy.+Integrity.+did",
            "you mean.+Integrity.+"),
    perl = TRUE
  )

  # md0$MISSING_LIST_TABLE <- NULL

  expect_silent(
    report <-
      suppressMessages(dq_report2(sd0, md0,
                                  resp_vars = c("v00000", "v00001", "v00002",
                                                "v00003", "v00004", "v00012"),
                                  filter_indicator_functions =
                                    c("^com_item_missingness$",
                                      "^acc_varcomp$"),
                                  filter_result_slots =
                                    c("^SummaryTable$"),
                                  cores = NULL,
                                 strata_attribute = "GROUP_VAR_XXX",
                                 dimensions = c("Completeness"),
      )
      )
  )

  report <- suppressWarnings(dq_report2(sd0, md0,
                                       cores = NULL,
                                       label_col = LABEL,
                                       dimensions = # for speed, omit Accuracy
                                         c("Completeness",
                                           "Consistency",
                                           "Accuracy"),
                                       resp_vars = c("SBP_0", "SEX_0"),
                                       filter_indicator_functions =
                                         c("^com_item_missingness$",
                                           "^acc_distributions_loc$",
                                           "^acc_margins$"),
                                       filter_result_slots =
                                         c("^SummaryTable$"),
                                       specific_args = list(
                                         acc_margins =
                                           list(min_obs_in_subgroup = 40),
                                         acc_distributions_loc =
                                           list(flip_mode = "flip"),
                                         com_item_missingness = list(
                                           label_col = LONG_LABEL
                                         ))
  ))

  expect_equal(
    attr(report$acc_margins_observer.SBP_0, "call")[["min_obs_in_subgroup"]],
    40)

  expect_equal(
    attr(report$acc_distributions_loc.SBP_0, "call")[["flip_mode"]],
    "flip")

  expect_null(
    attr(report$acc_distributions_loc_ecdf_observer.SBP_0,
         "call")[["flip_mode"]])

  expect_equal(
    attr(report$com_item_missingness.SBP_0, "call")[["label_col"]],
    LABEL) # this should not be overwritable

  report <- suppressWarnings(dq_report2(sd0, md0,
                                       resp_vars = c("SBP_0", "SEX_0"),
                                       filter_indicator_functions =
                                         c("^acc_distributions_loc_ecdf$",
                                           "^acc_distributions_loc$"),
                                       cores = NULL,
                                       flip_mode = "flip",
                                       dimensions = # for speed, omit Accuracy
                                         c("Completeness",
                                           "Consistency",
                                           "Accuracy"))
  )

  expect_equal(attr(report$acc_distributions_loc.SBP_0, "call")$resp_vars,
               "SBP_0") # resp_vars cannot be overwritten, but SEX_0 is not a pimary output

  expect_equal(attr(report$acc_distributions_loc.SBP_0, "call")$flip_mode,
               "flip")

  md1 <- md0
  md1$LABEL <- c("CENTER_0",
                 "",
                 "CENTER_0 DUPLICATE", # will become a duplicated label
                 "CENTER_0", # direct duplication of the first label
                 "Have you been physically vigorously active in the past 12 hours ('physically vigorously active' means at least 30 minutes of jogging or fast cycling, digging up your garden, carrying heavy objects weighing more than 10 kg for a long time, or similar physical activities)?", # very long label
                 "Hybpvaitp1hpvamal3mojofcduygchowmt1kfaltospa") # matches the very long label after abbreviation
  md1$VAR_NAMES[2] <- "yOvCzPY60JRjmrYb16Tsd6qMymal4B5Skw9rZ5PHSCtaBqOVglAKcguPkQhakampFJcC8xqLbZJs7kZUdKH804pbOmM5ORPVabrkEkVkiWbakWiixZ99NRYF6BP8SRxzNYY2tED7DjmhMUwk0t674RjH828jq9zoTJgDxYP6nEdHBxhmXJh0ClCPjGsi1q" # very long variable name that should get caught and not be used as label as it is
  colnames(sd0)[2] <- md1$VAR_NAMES[2]

  suppressWarningsMatching(
    expect_warning(expect_warning(
      report <- dq_report2(sd0, md1,
                           label_col = LABEL,
                           cores = NULL,
                           dimensions =
                             c("Integrity",
                               "Consistency")),
      regexp = ".*Labels are required to create a report.*"),
      regexp = ".*Unique labels are required to create a report.*"),
    c("Some variables have labels with more than 60 characters in .+LABEL.+",
      "Unique labels are required",
      ".*duplicated in the metadata and cannot be used as label.*")
  )
  suppressWarningsMatching(
    expect_warning(
      report <- dq_report2(sd0, md1,
                           label_col = VAR_NAMES,
                           cores = NULL,
                           dimensions =
                             c("Integrity",
                               "Consistency")),
      regexp = ".*This will cause suboptimal outputs and possibly.*"
      ),
    c("Some variables have no label in .+LABEL", "Unique labels are required",
      "more than 60 characters")
  )

  md1$VAR_NAMES[2] <- "" # this will be considered different
  colnames(sd0)[2] <- "" # because both are missing, and NA maybe unequal
                         # from NA

  suppressWarningsMatching(
    expect_warning(expect_warning(expect_warning(
            report <- dq_report2(sd0, md1,
                                 label_col = LABEL,
                                 cores = NULL,
                                 dimensions =
                                   c("Integrity",
                                     "Consistency")),
            regexp = "Need.+VAR_NAMES.+in.*meta_data",
            perl = TRUE
          ),
          regexp = "Some variables have duplicated labels in .+LABEL.+",
          perl = TRUE
        ),
      regexp =
        "Some variables have labels with more than 60 characters in .+LABEL.+",
      perl = TRUE
    ),
    c("(Some variables have duplicated labels in .+LABEL.+|Need.+VAR_NAMES.+in.*meta_data|Some variables have labels with more than 60 characters in .+LABEL.+|Lost 16.7% of the study data because of missing/not assignable metadata|.+dummy names|Some variables have no label in .+LABEL)",
      ".*duplicated in the metadata and cannot be used as label.*")
  )

  suppressWarningsMatching(
      report <- dq_report2(sd0, md1,
                           label_col = VAR_NAMES,
                           cores = NULL,
                           dimensions =
                             c("Integrity",
                               "Consistency")),
    "(Unique labels are required|Some variables have labels with more than 60 characters in .+LABEL.+|Lost 16.7% of the study data because of missing/not assignable metadata|Need.+VAR_NAMES.+discard.+|.+dummy names|Some variables have no label in .+LABEL)"
  )

  #warning message that there is no metadata
  prep_purge_data_frame_cache()
  study_data <- head(
    prep_get_data_frame(
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
    50)

  expect_warning(dq_report2(study_data = study_data,
                            cores = NULL,
                 dimensions = "Integrity"),
                 regexp = "NO ITEM LEVEL METADATA.+PLEASE CONSIDER PASSING.+")


  md1 <-
    prep_get_data_frame(
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx| item_level")
  md1 <- md1[, !colnames(md1) %in% c("VARIABLE_ROLE", "MISSING_LIST_TABLE")]


  #message about no assigned variable roles
  suppressWarningsMatching(dq_report2(study_data = study_data,
                            meta_data = md1,
                            cores = NULL,
                            dimensions = "Integrity"),
                           "Metadata does not provide.+for replacing codes with NAs.")

  md1 <-
    prep_get_data_frame(
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx| item_level")
  md1 <- md1[, !colnames(md1) %in% "MISSING_LIST_TABLE"]
  md1[1, VARIABLE_ROLE] <- "non-existing_role"


  # message about non existing role (ignore warnings about missing JUMP_LIST and
  # MISSING_LIST)
  suppressWarningsMatching(dq_report2(study_data = study_data,
                           meta_data = md1,
                           cores = NULL,
                           dimensions = "Integrity"),
                 "Metadata does not provide.+for replacing codes with NAs.")


  #message of variable not found in the metadata
  md1<- md1[, colnames(md1) %in% c("VAR_NAMES", "LABEL", "DATA_TYPE",
                                   "SCALE_LEVEL", "VALUE_LABELS")]
  md1 <- rbind(md1, data.frame(VAR_NAMES = "no_real_var_name",
                               LABEL = "no_var_name",
                               DATA_TYPE = "integer",
                               SCALE_LEVEL = "nominal",
                               VALUE_LABELS = NA_character_))

  expect_message2(dq_report2(study_data = study_data,
                            meta_data = md1,
                            cores = NULL,
                            dimensions = "Integrity"))

})

test_that("no issues with parallel running reports", {
  skip_if_not_installed("DT")
  skip_if_not_installed("stringdist")
  skip_if_not_installed("markdown")
  skip_on_cran() # needs maybe too many cores
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  study_data <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), 100)
  meta_data <- prep_get_data_frame("item_level")

  mlt <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx| missing_table")

  prep_purge_data_frame_cache()

  prep_add_data_frames(`missing_table` = mlt)

  invisible(testthat::capture_output_lines(gc(full = TRUE, verbose = FALSE)))

  sd0 <- study_data[, 1:5]
  sd0$v00012 <- study_data$v00012
  md0 <- subset(meta_data, VAR_NAMES %in% colnames(sd0))
  md0$PART_VAR <- NULL

  r <- without_testthat(
    withr::with_options( # otherwise, cores = 2 is disallowed
      list(dataquieR.tmp_no_load_all = TRUE, # even in testthat do not run load_all in this specific test, it would hide possible problems in the default code branch not being executed under load_all in util_evalute_calls
           dataquieR.dt_adjust = FALSE), { # speed up
    dq_report2(sd0, md0,
                    resp_vars = c("v00000", "v00001", "v00002",
                                  "v00003", "v00004", "v00012"),
                    filter_indicator_functions =
                      c("^com_item_missingness$",
                        "^acc_varcomp$"),
                    filter_result_slots =
                      c("^SummaryTable$"),
                    cores = 2,
                    mode = "queue",
                    dimensions = # for speed, omit Accuracy
                      c("Integrity",
                        "Completeness",
                        "Consistency",
                        "Accuracy"))
      })
  )

  res_is_null <- vapply(r,
         inherits, "dataquieR_NULL",
         FUN.VALUE = logical(1))

  res_is_iap <-
    any(vapply(r,
           function(rres) {
             if (length(attr(rres, "error")) > 1) {
               fail("We should not have > 1 error per result")
             }
             if (length(attr(rres, "error")) == 0) {
               FALSE
             } else {
               vapply(attr(rres, "error"),
                      inherits, dataquieR.intrinsic_applicability_problem,
                      FUN.VALUE = logical(1))
             }
           }, FUN.VALUE = logical(1)))

  res_is_ap <-
    any(vapply(r,
               function(rres) {
                 if (length(attr(rres, "error")) > 1) {
                   fail("We should not have > 1 error per result")
                 }
                 if (length(attr(rres, "error")) == 0) {
                   FALSE
                 } else {
                   vapply(attr(rres, "error"),
                          inherits, dataquieR.applicability_problem,
                          FUN.VALUE = logical(1))
                 }
               }, FUN.VALUE = logical(1)))

  expect_all_false(res_is_null & !res_is_ap & !res_is_iap) # should not have unknown errors in the report

})

# sd0 <- study_data
# md0 <- meta_data
# md0 <- md0[which(md0[[LABEL]] %in%
#                    c("CENTER_0", "CRP_0", "BSG_0", "PSEUD0_ID", "DEV_NO_0", "PART_STUDY", "PART_LAB")), ]
# md0[[LABEL]] <- ""
# # create a grouping variable with large integer numbers and without specifying value labels
# sd0$v11111 <- sd0$v00000 * 1000
# md0[["KEY_DEVICE"]][md0[["VAR_NAMES"]] == "v00015"] <- "v11111"
# md0 <- rbind(md0, c("v11111", "", "integer", rep(NA_character_, ncol(md0) - 3)))
# sd0 <- sd0[which(colnames(sd0) %in% md0[[VAR_NAMES]])]
# test_rep <- dq_report2(study_data = sd0, meta_data = md0, dimensions = NULL)

test_that("Time-only variables general", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("stringdist")

  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                                    keep_types = TRUE)

  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  meta_data <- prep_get_data_frame("item_level")

  meta_data <- util_rbind(
    meta_data,
    data.frame(
      stringsAsFactors = FALSE,
      VAR_NAMES = "v02000",
      LABEL = "ADMIS_TM_0",
      DATA_TYPE = DATA_TYPES$TIME,
      SCALE_LEVEL = SCALE_LEVELS$INTERVAL,
      VALUE_LABELS = NA_character_,
      STANDARDIZED_VOCABULARY_TABLE = NA_character_,
      MISSING_LIST_TABLE = NA_character_,
      HARD_LIMITS = "[09:00:00;18:00:00]",
      DETECTION_LIMITS = NA_character_,
      SOFT_LIMITS = NA_character_,
      DISTRIBUTION = NA_character_,
      DECIMALS = NA_character_,
      DATA_ENTRY_TYPE = NA_character_,
      GROUP_VAR_OBSERVER = NA_character_,
      GROUP_VAR_DEVICE = NA_character_,
      TIME_VAR = NA_character_,
      STUDY_SEGMENT = "STUDY",
      PART_VAR = "PART_STUDY",
      VARIABLE_ROLE = VARIABLE_ROLES$PRIMARY,
      VARIABLE_ORDER = "54",
      LONG_LABEL = "Admission time",
      ELEMENT_HOMOGENITY_CHECKTYPE = NA_character_,
      UNIVARIATE_OUTLIER_CHECKTYPE = NA_character_,
      N_RULES = "4",
      LOCATION_METRIC = NA_character_,
      LOCATION_RANGE = NA_character_,
      PROPORTION_RANGE = NA_character_,
      REPEATED_MEASURES_VARS = NA_character_,
      REPEATED_MEASURES_GOLDSTANDARD = NA_character_,
      CO_VARS = NA_character_,
      MISSING_LIST = "00:00:00 = not available"
    )
  )

  compute_me <- function() {
    set.seed(12345)
    r <- dq_report2(
      resp_vars = "ADMIS_TM_0",
      study_data = study_data, label_col = LABEL,meta_data = meta_data,
      filter_indicator_functions =
        c("^com_item_missingness$",
          "^con_limit_deviations$"),
      filter_result_slots =
        c("^SummaryTable$"),
      cores = NULL,
      meta_data_v2 =
        "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
    r
  }

  set.seed(12345)
  day_course <- unlist(lapply(
    0:23,
    function(h) {
      list(
        hms::hms(hours = h,
                 minutes = 3),
        hms::hms(hours = h,
                 minutes = 13),
        hms::hms(hours = h,
                 minutes = 24),
        hms::hms(hours = h,
                 minutes = 30),
        hms::hms(hours = h,
                 minutes = 50)
      )
    }
  ), recursive = FALSE)
  probs <-
    rep(c(.7, .2, .05, .04, 0.01), 24)
  times <- sample(x = day_course,
                  prob = probs,
                  size = nrow(study_data),
                  replace = TRUE)
  times[sample(seq_along(times),
               size = length(times) %/% 100,
               replace = !TRUE)] <-
    list(hms::hms(seconds = 0, minutes = 0, hours = 0))

  study_data$v02000 <- times

  r <- compute_me()

  expect_true(is.list(attr(r, "integrity_issues_before_pipeline")) &&
                length(attr(r, "integrity_issues_before_pipeline")) == 0)

  expect_length(r, 2)

  expect_equal(r$com_item_missingness.ADMIS_TM_0$SummaryTable$`Missing codes N`,
               30)
  expect_equal(r$com_item_missingness.ADMIS_TM_0$SummaryTable$`Sysmiss N`, 0)
  expect_equal(r$con_limit_deviations.ADMIS_TM_0$SummaryTable$NUM_con_rvv_inum,
               1864,
               ignore_attr = TRUE)

  times <- sample(x = day_course,
                  prob = probs,
                  size = nrow(study_data),
                  replace = TRUE)
  # introduce some missing
  times[sample(seq_along(times),
               size = length(times) %/% 100,
               replace = !TRUE)] <-
    hms::hms(seconds = 0, minutes = 0, hours = 0)

  study_data$v02000 <- times

  r <- compute_me()

  expect_true(is.list(attr(r, "integrity_issues_before_pipeline")) &&
                length(attr(r, "integrity_issues_before_pipeline")) == 1)

  expect_length(r, 2)

  expect_equal(r$com_item_missingness.ADMIS_TM_0$SummaryTable$`Missing codes N`,
               30)
  expect_equal(r$com_item_missingness.ADMIS_TM_0$SummaryTable$`Sysmiss N`, 0)
  expect_equal(r$con_limit_deviations.ADMIS_TM_0$SummaryTable$NUM_con_rvv_inum,
               1876,
               ignore_attr = TRUE)


  times[c(4, 6)] <-
    cars
  times[[17]] <- iris
  study_data$v02000 <- times

  r <- compute_me()

  expect_true(is.list(attr(r, "integrity_issues_before_pipeline")) &&
                length(attr(r, "integrity_issues_before_pipeline")) == 2)

  expect_length(r, 2)

  expect_equal(r$com_item_missingness.ADMIS_TM_0$SummaryTable$`Missing codes N`,
               30)
  expect_equal(r$com_item_missingness.ADMIS_TM_0$SummaryTable$`Sysmiss N`, 3)
  expect_equal(r$con_limit_deviations.ADMIS_TM_0$SummaryTable$NUM_con_rvv_inum,
               1875,
               ignore_attr = TRUE)

  skip_if_not_installed("chron")

  probs <-
    rep(c(.7, .2, .05, .04, 0.01), 24)
  times <- sample(x = day_course,
                  prob = probs,
                  size = nrow(study_data),
                  replace = TRUE)
  times[sample(seq_along(times),
               size = length(times) %/% 100,
               replace = !TRUE)] <-
    list(hms::hms(seconds = 0, minutes = 0, hours = 0))

  times <-
    get("as.times", envir = asNamespace("chron"))(util_as_character(times))

  study_data$v02000 <- times

  r <- compute_me()

  expect_true(is.list(attr(r, "integrity_issues_before_pipeline")) &&
                length(attr(r, "integrity_issues_before_pipeline")) == 0)

  expect_length(r, 2)

  expect_equal(r$com_item_missingness.ADMIS_TM_0$SummaryTable$`Missing codes N`,
               30)
  expect_equal(r$com_item_missingness.ADMIS_TM_0$SummaryTable$`Sysmiss N`, 0)
  expect_equal(r$con_limit_deviations.ADMIS_TM_0$SummaryTable$NUM_con_rvv_inum,
               1876,
               ignore_attr = TRUE)


})

test_that("complex_data_types", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                                    keep_types = TRUE)

  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  meta_data <- prep_get_data_frame("item_level")

  meta_data[[DATA_TYPE]][meta_data[[LABEL]] == "N_CHILD_0"] <- "cOuNt"
  meta_data[[SCALE_LEVEL]][meta_data[[LABEL]] == "N_CHILD_0"] <- ""

  # dq_report2 ----

  r <- dq_report2(
    resp_vars = "N_CHILD_0",
    study_data = study_data, label_col = LABEL, meta_data = meta_data,
    dimensions = NULL,
    cores = NULL,
    meta_data_v2 =
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  repsum <- summary(r)
  expect_snapshot(
    attr(r, "meta_data")[attr(r, "meta_data")[[LABEL]] == "N_CHILD_0", , FALSE]
  )

  expect_snapshot(repsum)

  # Call one function, only ----
  r <- acc_distributions(
    resp_vars = "N_CHILD_0",
    study_data = study_data, label_col = LABEL, meta_data = meta_data,
    meta_data_v2 =
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  skip_if_not_installed("vdiffr")
  expect_doppelganger2("EXTENDED_DATA_TYPE cnt dist",
                       r$SummaryPlotList$N_CHILD_0)

  # use pre-filled columns, such as SCALE_LEVEL, EXTENDED_DATA_TYPE, ... ----

  meta_data[[HARD_LIMITS]][meta_data[[LABEL]] == "N_CHILD_0"] <-
    "[0; 7]"
  r <- acc_distributions(
    resp_vars = "N_CHILD_0",
    study_data = study_data, label_col = LABEL, meta_data = meta_data,
    meta_data_v2 =
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  skip_if_not_installed("vdiffr")
  expect_doppelganger2("EXTENDED_DATA_TYPE cnt dist2",
                       r$SummaryPlotList$N_CHILD_0)

})

test_that("util_meta_data_env works", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  b <- util_meta_data_env(meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx",
                          study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS")
  xx <- b$call(acc_margins(resp_vars = "sbp1",
                     group_vars = "DEV_BP_0"))
  prep_purge_data_frame_cache()
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx")
  yy <- acc_margins(resp_vars = "sbp1",
                    group_vars = "DEV_BP_0",
                    study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS",
                    meta_data = "item_level",
                    co_vars = c("SEX_0", "AGE_0"),
                    label_col = LABEL)
  expect_equal(xx$ResultData, yy$ResultData, ignore_attr = TRUE)
  expect_equal(xx$SummaryPlot, yy$SummaryPlot, ignore_attr = TRUE)
  expect_equal(xx$SummaryTable[, -1, FALSE],
               yy$SummaryTable[, -1, FALSE],
               ignore_attr = TRUE)
})

test_that("prep_compare_meta_with_study works", {
  skip_on_cran() # online, fragile
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  sd0 <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  my_warnings <- capture_warnings({
    res <- prep_compare_meta_with_study(sd0)
  })
  expect_snapshot_value(res, style = "deparse")
  expect_snapshot_value(my_warnings, style = "deparse")
})

test_that("util_varcomp_robust works", {
  skip_if_not_installed("rankICC")
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  r <- util_varcomp_robust(resp_vars = "SBP_0",
                           group_vars = "USR_BP_0",
                           study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                           meta_data = "item_level")
  expect_snapshot(r)
})

test_that("com_qualified_segment_missingness no meta", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  expect_warning(
    expect_warning(
      expect_warning(
        expect_warning(
          expect_warning(
            res <- com_qualified_segment_missingness(
              study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
              meta_data = "item_level",
              meta_data_segment = "segment_level"),
            regexp = "Missing or doubled.+SEGMENT_PART.+segm.+for.+"
          ),
          regexp = "Missing or doubled.+SEGMENT_PART.+segm.+for.+"
        ),
        regexp = "Missing or doubled.+SEGMENT_PART.+segm.+for.+"
      ),
      regexp = "Missing or doubled.+SEGMENT_PART.+segm.+for.+"
    ),
    regexp = "Missing or doubled.+SEGMENT_PART.+segm.+for.+"
  )
  expect_type(res, "list")
  expect_s3_class(res$SegmentTable, "data.frame")
  expect_s3_class(res$SegmentData, "data.frame")
  expect_length(res$SegmentData, 0)
  expect_length(res$SegmentTable, 0)
})

test_that("com_qualified_segment_missingness with meta", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx")
  res <- com_qualified_segment_missingness(
    study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS",
    meta_data = "item_level",
    meta_data_segment = "segment_level")
  expect_type(res, "list")
  expect_s3_class(res$SegmentTable, "data.frame")
  expect_s3_class(res$SegmentData, "data.frame")
  expect_equal(nrow(res$SegmentData), 4)
  expect_equal(nrow(res$SegmentTable), 4)
  expect_true(any(res$SegmentTable[, 2:ncol(res$SegmentTable)] > 0))
  expect_true(any(res$SegmentData[, 2:ncol(res$SegmentData)] > 0))
})

test_that("nres works", {
  skip_on_cran() # slow, errors unlikely
  skip_if_not_installed("stringdist")

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  withr::defer(prep_purge_data_frame_cache())
  # item_level <- prep_get_data_frame("item_level")
  # item_level$MISSING_LIST_TABLE <- NULL
  # prep_add_data_frames(item_level)
  report <-
    dq_report2("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", dimensions = c("int"), label_col = "LABEL",
               cores = NULL, filter_result_slots = NULL);
  expect_equal(nres(report), 13)
})

test_that("util_parallel_futures works", {
  # testthat::skip_if(identical(Sys.getenv("R_COVR"), "true"),
  #                   message = "Crashes, if instrumented")
  skip_on_cran() # slow, parallel, ...
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("future")
  skip_if_not_installed("stringdist")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  study_data <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), 100)
  meta_data <- prep_get_data_frame("item_level")

  mlt <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx| missing_table")

  prep_purge_data_frame_cache()

  prep_add_data_frames(`missing_table` = mlt)

  invisible(testthat::capture_output_lines(gc(full = TRUE, verbose = FALSE)))

  sd0 <- study_data[, 1:5]
  sd0$v00012 <- study_data$v00012
  md0 <- subset(meta_data, VAR_NAMES %in% colnames(sd0))
  md0$PART_VAR <- NULL

  # md0$MISSING_LIST_TABLE <- NULL

  # don't include huge reports as RData in the package
  # Suppress warnings since we do not test dq_report2
  # here in the first place
  expect_warning({
    expect_warning({
      report <- dq_report2(sd0, md0,
                           resp_vars = c("v00000", "v00001", "v00002",
                                         "v00003", "v00004", "v00012"),
                           filter_indicator_functions =
                             c("^com_item_missingness$",
                               "^acc_varcomp$"),
                           filter_result_slots =
                             c("^SummaryTable$"),
                           cores = list(mode = "socket", cores = 1),
                           mode = "futures",
                           dimensions = # for speed, omit Accuracy
                             c("Integrity",
                               "Completeness",
                               "Consistency",
                               "Accuracy"))
    }, regexp =
      ".*context*",
    perl = TRUE)}, regexp =
      ".*context*",
    perl = TRUE)

  if (nres(report) == 0) {
    if (identical(Sys.getenv("CI_PROJECT_ID"), "10015470")) { # only in our gitlab CI pipeline
      p <- file.path(path.expand("~"), "addtional_output")
      if (!dir.exists(p)) {
        dir.create(p, recursive = TRUE)
      }
      if (dir.exists(p)) {
        save(report, file = file.path(p, "util_parallel_futures_report.RData"))
      }
    }
  }

  expect_equal(dim(report), c(6, 3, 1))
  expect_snapshot(summary(report))

})

test_that("util_has_no_group_vars works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  expect_equal(util_has_no_group_vars("SBP_0"), FALSE)
  expect_equal(util_has_no_group_vars("DBP_0"), FALSE)
  expect_error(util_has_no_group_vars(NA))
  expect_equal(util_has_no_group_vars("SEX_0"), TRUE)
})

test_that("acc_distributions_ecdf works", {
  skip_on_cran() # slow and tested elsewhere
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("colorspace")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx|item_level")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  t1 <- acc_distributions_ecdf(group_vars = "obs_soma",
                                 study_data = study_data,
                                 meta_data = meta_data)

  expect_equal(length(names(t1$SummaryPlotList)), 13)

  expect_false(
    inherits(try(ggplot_build(t1$SummaryPlot$exdate)), "try-error"))
  expect_false(
    inherits(try(ggplot_build(t1$SummaryPlot$cholesterol)), "try-error"))
})

test_that("prep_undisclose works", {
  skip_on_cran() # slow, parallel, ...
  skip_if_not_installed("stringdist")

  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  study_data <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), 100)
  meta_data <- prep_get_data_frame("item_level")

  mlt <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx| missing_table")

  prep_purge_data_frame_cache()

  prep_add_data_frames(`missing_table` = mlt)

  invisible(testthat::capture_output_lines(gc(full = TRUE, verbose = FALSE)))

  sd0 <- study_data[, 1:5]
  sd0$v00012 <- study_data$v00012
  md0 <- subset(meta_data, VAR_NAMES %in% colnames(sd0))
  md0$PART_VAR <- NULL

  # md0$MISSING_LIST_TABLE <- NULL

  # don't include huge reports as RData in the package
  # Suppress warnings since we do not test dq_report2
  # here in the first place
  report <- dq_report2(sd0, md0,
                       resp_vars = c("v00000", "v00001", "v00002",
                                     "v00003", "v00004", "v00012"),
                       filter_indicator_functions =
                         c("distrib",
                           "^acc_univariate_outlier$"),
                       cores = NULL,
                       dimensions = # for speed, omit Accuracy
                         c("Integrity",
                           "Completeness",
                           "Consistency",
                           "Accuracy"))

  expect_equal(nrow(report$acc_univariate_outlier.SBP_0$SummaryPlotList$SBP_0$data), 88)
  report_x <- prep_undisclose(report)
  expect_equal(length(report_x$acc_univariate_outlier.SBP_0$SummaryPlotList$SBP_0$data), 0)
})

test_that("prep_save_report correctly saves a report with correct parameters", {
  skip_on_cran() # slow, parallel, ...
  skip_if_not_installed("stringdist")

  # Define a temporary file path for the test
  tmp_file <- tempfile(fileext = ".gz")

  # Not a dq_report2 object
  df <- data.frame(a =  1:5, b = letters[1:5])
  expect_error(prep_save_report(df, tmp_file))
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # Create example report
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <- prep_get_data_frame("item_level")

  sd0 <- study_data[, 1:5]
  sd0$v00012 <- study_data$v00012
  md0 <- subset(meta_data, VAR_NAMES %in% colnames(sd0))
  md0$PART_VAR <- NULL

  # don't include huge reports as RData in the package
  # Suppress warnings since we do not test dq_report2
  # here in the first place
  report <- dq_report2(sd0, md0,
                       resp_vars = c("v00000", "v00001", "v00002",
                                     "v00003", "v00004", "v00012"),
                       filter_indicator_functions =
                         c("^com_item_missingness$",
                           "^acc_varcomp$"),
                       filter_result_slots =
                         c("^SummaryTable$"),
                       cores = NULL,
                       dimensions = # for speed, omit Accuracy
                         c("Integrity",
                           "Completeness",
                           "Consistency",
                           "Accuracy"))

  prep_save_report(report, tmp_file)

  # Check that the file was created
  expect_true(file.exists(tmp_file))

  # Check the file size
  expect_gt(file.info(tmp_file)$size,  0)

  # Check for invalid compression_level
  expect_error(prep_save_report(report, tmp_file, compression_level =  10))

  # Clean up
  on.exit(unlink(tmp_file))
})

test_that("dq_report_by works with content s", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
    skip_on_cran() # slow test
    # TODO: test with all sorts of _by calls and on windows
    target <- withr::local_tempdir("testdqareportby")

    sd0 <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), 100)

    expect_message2(dq_report_by(sd0,
                 dimensions = "int",
                 cores = NULL,
                 strata_column = "SEX_0",
                 strata_select = "0", selection_type = "value",
                 meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                 output_dir = !!file.path(target, "s"),
                 also_print = TRUE))

    # TODO: check target path for all expected artifacts, namely the dq2 files, the .output_dirs containing the HTML reports and the summary RDS files

})


test_that("dq_report_by works with complex limits", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_on_cran() # slow test
  # TODO: test with all sorts of _by calls and on windows
  target <- withr::local_tempdir("testdqareportby")

  sd0 <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), 50)

  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  md <- prep_get_data_frame("item_level")
  cil <- prep_get_data_frame("cross-item_level")
  cil <- rbind(cil, c("SBP_0", rep(NA, 12) ))
  cil$HARD_LIMITS <- c(rep(NA, 12), "[SBP_0] > 80 and [SBP_0] < 180")

  expect_message2(dq_report_by(sd0,
                              dimensions = "con",
                              cores = NULL,
                              segment_select = "PHYS_EXAM",
                              item_level = md,
                              meta_data_cross_item = cil,
                              output_dir = !!file.path(target, "c"),
                              also_print = TRUE))

})

test_that("acc_cat_distributions works", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  r <- acc_cat_distributions(study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                             resp_vars = "EDUCATION_1",
                             group_vars = "USR_SOCDEM_0",
                             label_col = LABEL)
  skip_if_not_installed("vdiffr")
  skip_on_cran()
  expect_doppelganger2("acc_cat_distributions EDUCATION_1",
                       r$SummaryPlot)

})

test_that("int_encoding_errors works", {
  skip_on_cran() # depends on OS code page support
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_purge_data_frame_cache()

  sd <-
    prep_get_data_frame(
  "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"
  )

  sd$v00001[[1]] <- "ä"
  skip_if(gsub("[^a-z0-9]", "", tolower(Encoding(sd$v00001[[1]]))) != "utf8",
          "Platform seems not to use UTF-8")

  # Introduce an encoding error
  Encoding(sd$v00001[[1]]) <- "latin1"

  expect_warning(
    x <- int_encoding_errors(study_data = sd),
    regexp = "neither"

  )

  expect_equal(sum(x$SummaryTable$NUM_int_uenc), 1)
  expect_equal(which(x$SummaryTable$NUM_int_uenc > 0), 1)

  prep_purge_data_frame_cache()
  prep_load_workbook_like_file(
"https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  dl <- prep_get_data_frame("dataframe_level")

  dl[[ENCODING]] <- c("utf-8")
  dl[[DF_CODE]] <- "sd"
  prep_add_data_frames(dataframe_level = dl)

  il <- prep_get_data_frame("item_level")

  il[[DATAFRAMES]] <- "sd"
  il[[ENCODING]] <- "utf-8"
  prep_add_data_frames(item_level = il)

  x <- int_encoding_errors(study_data = sd)
  expect_equal(sum(x$SummaryTable$NUM_int_uenc), 1)
  expect_equal(which(x$SummaryTable$NUM_int_uenc > 0), 1)

  prep_purge_data_frame_cache()
  prep_load_workbook_like_file(
"https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  dl <- prep_get_data_frame("dataframe_level")

  dl[[ENCODING]] <- c("utf-8")
  dl[[DF_CODE]] <- "sd"
  prep_add_data_frames(dataframe_level = dl)

  il <- prep_get_data_frame("item_level")

  il[[DATAFRAMES]] <- "sd"
  prep_add_data_frames(item_level = il)

  x <- int_encoding_errors(study_data = sd)
  expect_equal(sum(x$SummaryTable$NUM_int_uenc), 1)
  expect_equal(which(x$SummaryTable$NUM_int_uenc > 0), 1)

  prep_purge_data_frame_cache()
  prep_load_workbook_like_file(
"https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  il <- prep_get_data_frame("item_level")

  il[[ENCODING]] <- "utf-8"
  prep_add_data_frames(item_level = il)

  x <- int_encoding_errors(study_data = sd)
  expect_equal(sum(x$SummaryTable$NUM_int_uenc), 1)
  expect_equal(which(x$SummaryTable$NUM_int_uenc > 0), 1)

})

test_that("reports with encoding errors work", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  skip_on_cran() # depends on OS code page support
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_purge_data_frame_cache()

  sd <-
    prep_get_data_frame(
  "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"
    )

  sd <- head(sd, 10)

  sd$v00012[[1]] <- "ä"
  skip_if(gsub("[^a-z0-9]", "", tolower(Encoding(sd$v00012[[1]]))) != "utf8",
          "Platform seems not to use UTF-8")

  # Introduce an encoding error
  Encoding(sd$v00012[[1]]) <- "latin1"

  r <- dq_report2(sd,
                  resp_vars = c("SBP_0", "ITEM_1_0"),
                  dimensions = NULL,
                  cores = NULL,
                  meta_data_v2 =
"https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  td <- withr::local_tempdir("testdqareportby")
  print(r, view = FALSE, dir = td)

  sz <- sum(file.size(list.files(td,
                                 recursive = TRUE,
                                 full.names = TRUE,
                                 all.files = TRUE)))
  expect_gt(sz, 13000000)

})

#' Replace limit violations (HARD_LIMITS) by NAs
#'
#' @param study_data [study_data]
#' @param meta_data [study_data]
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#'
#' @return modified `study_data`
#'
#' @family data_management
#' @concept metadata_management
#' @noRd
 util_replace_hard_limit_violations <- function(study_data,
                                                meta_data,
                                                label_col) {

   if (!isTRUE(attr(study_data, "Codes_to_NA"))) {
     util_error(
       c("Missing codes have to have been replaced already by NA, before %s",
         "violations can be replaced by NA. This is an internal error.",
         "As a dataquieR developer, please ensure, that you only pass",
         "%s that have the %s attribute (set, e.g., by %s or %s) as %s to %s."),
       sQuote(HARD_LIMITS),
       sQuote("study_data"),
       dQuote("Codes_to_NA"),
       sQuote("prep_prepare_dataframes"),
       sQuote("util_replace_codes_by_NA"),
       sQuote("study_data"),
       sQuote("util_replace_hard_limit_violations")
       )
   }
   if (!isTRUE(attr(study_data, "MAPPED", exact = TRUE))) {
     util_error(
       c("%s must have been mapped with %s, before %s",
         "violations can be replaced by NA. This is an internal error.",
         "As a dataquieR developer, please ensure, that you only pass",
         "%s that have the %s attribute (set, e.g., by %s) as %s to %s."),
       sQuote("study_data"),
       sQuote("meta_data"),
       sQuote(HARD_LIMITS),
       sQuote("study_data"),
       dQuote("MAPPED"),
       sQuote("prep_prepare_dataframes"),
       sQuote("study_data"),
       sQuote("util_replace_hard_limit_violations")
     )
   }

   if (!HARD_LIMITS %in% names(meta_data)) {
     util_message("Cannot replace %s violations, because %s do not provide %s.",
                  sQuote(HARD_LIMITS),
                  dQuote("meta_data"),
                  sQuote(HARD_LIMITS),
                  applicability_problem = TRUE)
     attr(study_data, "HL_viol_to_NA") <- TRUE
     return(study_data)
   } else {
     hl <- setNames(meta_data[[HARD_LIMITS]], nm = meta_data[[label_col]])
     hl[util_empty(hl)] <- NA
   }

   i <- lapply(hl, util_parse_interval)

   if (length(meta_data[[DATA_TYPE]]) == 0) {
     meta_data[[DATA_TYPE]] <- "n/a in metadata"
   }

   study_data[] <- mapply(study_data, i, names(i), meta_data[[DATA_TYPE]],
                          SIMPLIFY = FALSE,
                          FUN = function(col, int, nm, dt) {
     if (inherits(int, "interval") && (prep_dq_data_type_of(col) %in%
                                        c(DATA_TYPES$INTEGER,
                                          DATA_TYPES$FLOAT,
                                          DATA_TYPES$DATETIME,
                                          DATA_TYPES$TIME))) {
       outside <- !redcap_env$`in`(col, int)
       outside[is.na(outside)] <- FALSE
       col[outside] <- NA
       col
     } else {
       if (inherits(int, "interval")) {
         util_warning(c("For %s, I have %s but the column is of type %s",
                        "(metadata say %s)"),
                      dQuote(nm), sQuote(HARD_LIMITS),
                      sQuote(prep_dq_data_type_of(col)),
                      sQuote(dt), applicability_problem = TRUE)
       }
       col
     }
   })

   attr(study_data, "HL_viol_to_NA") <- TRUE
   study_data
 }

#' HTML Dependency for report headers in `DT::datatable`
#'
#' @return the dependency
html_dependency_report_dt <- function() {
  htmltools::htmlDependency(
    name = "report-dt-style"
    ,version = "0.0.1"
    ,src = c(file = system.file("report-dt-style", package = "dataquieR"))
    ,stylesheet = "report-dt-style.css"
    ,script = "report_dt.js"
  )
}

#' Save a `dq_report2`
#'
#' @param report [dataquieR_resultset2] the report
#' @param file [character] the file name to write to
#' @param compression_level [integer]  from=0 to=9. Compression level.
#'        9 is very slow.
#'
#' @return `invisible(NULL)`
#' @export
prep_save_report <- function(report, file, compression_level = 3) {
  util_expect_scalar(file, check_type = is.character)
  if (!inherits(report, "dataquieR_resultset2")) {
    util_error("Can only save %s reports",
               sQuote("dq_report2"))
  }
  util_expect_scalar(compression_level,
                     check_type = util_is_integer)
  if (compression_level < 0 || compression_level > 9) {
    util_error("%s msut be an integer between 0 and 9, not %d",
               sQuote("compression_level"),
               compression_level)
  }
  zz <- gzcon(file(file, "wb"), level = compression_level)
  serialize(object = report, connection = zz, version = 3)
  close(zz)
  invisible(NULL)
}

#' Load a `dq_report2`
#'
#' @param file [character] the file name to load from
#'
#' @return [dataquieR_resultset2] the report
#' @export
prep_load_report <- function(file) {
  util_expect_scalar(file, check_type = is.character)
  util_stop_if_not(file.exists(file))
  zz <- gzcon(file(file, "rb"))
  report <- unserialize(zz)
  close(zz)
  report
}

#' Insert missing codes for `NA`s based on rules
#'
#' @param resp_vars [variable list] the name of the measurement variables to be
#'   modified, all from `rules`, if omitted
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param rules [data.frame] with the columns:
#'   - `resp_vars` or `VAR_NAMES`:
#'                  Variable, whose `NA`-values should be replaced by jump codes
#'   - `CODE_CLASS`: Either `MISSING` or `JUMP`: Is the currently described case
#'                   an expected missing value (`JUMP`) or not (`MISSING`)
#'   - `CODE_VALUE`: The jump code or missing code
#'   - `CODE_LABEL`: A label describing the reason for the missing value
#'   - `RULE`: A rule in `REDcap` style (see, e.g.,
#'   [`REDcap` help](https://help.redcap.ualberta.ca/help-and-faq/project-best-practices/data-quality/example-data-quality-rules),
#'   [`REDcap` how-to](https://docs.google.com/document/d/1l3nGBgqqPKi5PtMe75g7q0dny8QzGMd_/edit?tab=t.0)), and
#'   [`REDcap` branching logic](https://www.iths.org/wp-content/uploads/REDCap-Branching-Logic-2017-202.pdf)
#'   that describes cases for the missing
#'   - `DATA_PREPARATION`: optional. if available, and `use_value_labels`` is
#'                                   either missing or `NA`, this columns
#'                                   controls the rule handling.
#' @param overwrite [logical] Also insert missing codes, if the values are not
#'                            `NA`
#' @param use_value_labels [logical] In rules for factors, use the value labels,
#'                                   not the codes. Defaults to `TRUE`, if any
#'                                   `VALUE_LABELS` are given in the metadata,
#'                                   and no `DATA_PREPARATION` exists for the
#'                                   `rules`. `NA` means to use
#'                                   `DATA_PREPARATION`, if available.
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return a `list` with the entries:
#'   - `ModifiedStudyData`: Study data with `NA`s replaced by the `CODE_VALUE`
#'   - `ModifiedMetaData`: Metadata having the new codes amended in the columns
#'                         `JUMP_LIST` or `MISSING_LIST`, respectively
#' @export
#'
prep_add_missing_codes <- function(resp_vars,
                                   study_data,
                                   meta_data_v2,
                                   item_level = "item_level",
                                   label_col,
                                   rules,
                                   use_value_labels = NA,
                                   overwrite = FALSE,
                                   meta_data = item_level) {
  util_maybe_load_meta_data_v2()
  util_expect_scalar(overwrite, check_type = is.logical)

  prep_prepare_dataframes(.replace_missings = FALSE,
                          .replace_hard_limits = FALSE,
                          .adjust_data_type = FALSE,
                          .amend_scale_level = FALSE)

  util_expect_scalar(use_value_labels, check_type = is.logical, allow_na = TRUE)

  util_expect_data_frame(rules, c(CODE_CLASS,
                                  CODE_LABEL, CODE_VALUE, RULE))

  if (all(c("resp_vars", VAR_NAMES) %in% colnames(rules))) {
    util_error(c("Have %s as well as %s in %s. This is not supported,",
                 "give only one of these columns, please."),
               dQuote("resp_vars"),
               dQuote(VAR_NAMES),
               sQuote("rules"),
               applicability_problem = TRUE)
  }

  colnames(rules)[colnames(rules) == "resp_vars"] <-
    VAR_NAMES

  util_expect_data_frame(rules, c(VAR_NAMES, CODE_CLASS,
                                  CODE_LABEL, CODE_VALUE, RULE))

  rules <- util_explode_var_names_lists(
    var_names_list =
      util_parse_assignments(rules[[VAR_NAMES]], multi_variate_text = TRUE),
    dframe = rules,
    col_name = VAR_NAMES)

  rules[[VAR_NAMES]] <- util_find_var_by_meta(
    rules[[VAR_NAMES]],
    meta_data = meta_data,
    label_col = label_col,
    target = label_col,
    ifnotfound = rules[[VAR_NAMES]]
  )

  rules_var_in_meta <- rules[[VAR_NAMES]] %in% meta_data[[label_col]]
  if (!all(rules_var_in_meta)) {
    util_message(
      c("Found the following %s/%s in %s,",
        "which are not in the %s: %s.",
        "Ignoring these rules."),
      sQuote(VAR_NAMES),
      sQuote("resp_vars"),
      sQuote("rules"),
      sQuote("meta_data"),
      util_pretty_vector_string(rules[[VAR_NAMES]][!rules_var_in_meta])
    )
    rules <- rules[rules_var_in_meta, , FALSE]
  }


  util_expect_data_frame(rules, list(VAR_NAMES = function(x) {
                                        all(x %in% colnames(ds1))
                                      },
                                     CODE_CLASS = function(x) {
                                       all(x %in% c("MISSING", "JUMP"))
                                     },
                                     CODE_LABEL = is.character,
                                     CODE_VALUE = util_is_valid_missing_codes,
                                     RULE = is.character))

  if (missing(resp_vars) || identical(resp_vars, NA)) {
    resp_vars <- rules[[VAR_NAMES]]
  }

  util_correct_variable_use(resp_vars, allow_more_than_one = TRUE)

  rules <- rules[rules[[VAR_NAMES]] %in% resp_vars, , FALSE]

  rules[[CODE_VALUE]] <- util_as_valid_missing_codes(rules[[CODE_VALUE]])

  compiled_rules <- lapply(setNames(nm = rules[[RULE]]), util_parse_redcap_rule)

  if (!DATA_PREPARATION %in% colnames(rules)) { # TODO: Also suppport replace_missings_by and replace_limits by arguments
    if (is.na(use_value_labels)) {
      if (
        (VALUE_LABELS %in% colnames(meta_data) &&
           any(!util_empty(meta_data[[VALUE_LABELS]]))) ||
        (VALUE_LABEL_TABLE %in% colnames(meta_data) &&
           any(!util_empty(meta_data[[VALUE_LABEL_TABLE]]))) ||
        (STANDARDIZED_VOCABULARY_TABLE %in% colnames(meta_data) &&
           any(!util_empty(meta_data[[STANDARDIZED_VOCABULARY_TABLE]]))))
        rules[[DATA_PREPARATION]] <- "LABEL"
      else
        rules[[DATA_PREPARATION]] <- ""
    } else {
      if (use_value_labels) {
        rules[[DATA_PREPARATION]] <- "LABEL"
      } else {
        rules[[DATA_PREPARATION]] <- ""
      }
    }
  }

  rule_match <- mapply(SIMPLIFY = FALSE,
                       rule = compiled_rules,
                       data_preparation = rules[[DATA_PREPARATION]],
                       FUN = function(rule, data_preparation) {
    # TODO: implement handling of DATA_PREPARATION fully in util_eval_rule, instead
    to_apply <- util_parse_assignments(data_preparation)
    replace_missing_by <- ""
    if (sum("MISSING_INTERPRET" %in% to_apply,
            "MISSING_LABEL" %in% to_apply,
            "MISSING_NA" %in% to_apply) > 1) {
      util_warning("Invalid %s in computation rules. Falling back to %s",
                   sQuote(DATA_PREPARATION),
                   dQuote("MISSING_NA"))
      to_apply <- to_apply[!startsWith(to_apply, "MISSING_")]
      to_apply <- c(to_apply, "MISSING_NA")
    }
    if ("MISSING_NA" %in% to_apply) {
      replace_missing_by <- "NA"
    }
    if ("MISSING_INTERPRET" %in% to_apply) {
      replace_missing_by <- "INTERPRET"
    }
    if ("MISSING_LABEL" %in% to_apply) {
      replace_missing_by <- "LABEL"
    }

    util_eval_rule(rule = rule,
                   ds1 = ds1,
                   meta_data = meta_data,
                   replace_missing_by = replace_missing_by,
                   use_value_labels = ("LABEL" %in% to_apply),
                   replace_limits = ("LIMITS" %in% to_apply))
  })

  na <- lapply(setNames(rules[[VAR_NAMES]], nm = rules[[RULE]]), function(rv) {
    r <- util_empty(ds1[[rv]])
    if (overwrite) {
      r[] <- TRUE
    }
    r
  })

  util_stop_if_not(length(rule_match) == length(na))

  add_miss <- lapply(setNames(nm = rules[[RULE]]),
                     function(rl) {
                       util_stop_if_not(length(na[[rl]]) ==
                                          length(rule_match[[rl]]))
                       rm <- as.logical(rule_match[[rl]])
                       rm[is.na(rm)] <- FALSE
                       na[[rl]] & rm
                      })

  workload <- as.data.frame(add_miss,
                            stringsAsFactors = FALSE,
                            check.names = FALSE)

  target_variable <- function(nm) {
    rules[rules[[RULE]] == nm, VAR_NAMES, TRUE]
  }

  target_code <- function(nm) {
    rules[rules[[RULE]] == nm, CODE_VALUE, TRUE]
  }

  for (rulenm in rules[[RULE]]) {
    tv <- target_variable(rulenm)
    for (tv0 in tv) {
      ds1[[tv0]][workload[[rulenm]]] <-
        rep(target_code(rulenm), sum(workload[[rulenm]]))
    }
  }

  cause_label_df_list <- prep_extract_cause_label_df(meta_data = meta_data,
                                                     label_col = label_col)

  meta_data <- cause_label_df_list$meta_data
  cause_label_df <- cause_label_df_list$cause_label_df

  colnames(cause_label_df)[colnames(cause_label_df) == "resp_vars"] <-
    VAR_NAMES

  cause_label_df <- rbind(cause_label_df,
                          rules[, colnames(cause_label_df)])

  if (!CODE_LABEL %in% colnames(cause_label_df)) {
    cause_label_df[[CODE_LABEL]] <-
      paste(cause_label_df[[CODE_CLASS]], cause_label_df[[CODE_VALUE]])
  }

  colnames(cause_label_df)[colnames(cause_label_df) == VAR_NAMES] <-
    "resp_vars"

  meta_data <-
    prep_add_cause_label_df(meta_data = meta_data,
                            cause_label_df = cause_label_df,
                            label_col = label_col)

  ModifiedStudyData <- ds1

  ModifiedStudyData[] <- lapply(ModifiedStudyData, unlist)

  colnames(ModifiedStudyData) <-
    prep_map_labels( colnames(ds1),
                     to = VAR_NAMES,
                     from = label_col,
                     meta_data = meta_data)

  ModifiedMetaData <- meta_data

  return(list(ModifiedStudyData = ModifiedStudyData,
              ModifiedMetaData = ModifiedMetaData))

}

#' Explode list-columns of variable names to row-wise representation
#'
#' Expands a data frame by repeating each row according to the length of the
#' corresponding element in `var_names_list`. Each element of the list must be a
#' character vector. The resulting data frame contains one row per entry of all
#' character vectors, with the original columns preserved and an additional
#' column holding the exploded values.
#'
#' @param var_names_list A list of character vectors. Must have the same length
#'   as the number of rows in `dframe`.
#' @param dframe A `data.frame` whose rows are to be repeated.
#' @param col_name A single character string giving the name of the new column
#'   that will contain the unlisted values from `var_names_list`. Defaults to
#'   `"var_names"`.
#'
#' @returns A `data.frame` with repeated rows from `dframe` and one additional
#'   column named according to `col_name`, containing the values from
#'   `var_names_list`. Rows corresponding to empty elements (`character(0)`) in
#'   `var_names_list` are omitted.
#' @noRd
util_explode_var_names_lists <- function(var_names_list, dframe,
                                         col_name = "var_names") {

  util_stop_if_not(
    "'var_names_list' must be a list" = is.list(var_names_list)
  )

  util_stop_if_not(
    "'dframe' must be a data.frame" = is.data.frame(dframe)
  )

  util_stop_if_not(
    "'col_name' must be a single character value" =
      is.character(col_name) && length(col_name) == 1L && !is.na(col_name)
  )

  util_stop_if_not(
    "length(var_names_list) must match nrow(dframe)" =
      length(var_names_list) == nrow(dframe)
  )

  lens <- lengths(var_names_list)

  idx <- rep(seq_len(nrow(dframe)), lens)

  out <- dframe[idx, , drop = FALSE]

  out[[col_name]] <- unlist(var_names_list, use.names = FALSE)

  row.names(out) <- NULL

  out
}

test_that("prep_add_missing_codes works", {
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  rules <- tibble::tribble(
    ~resp_vars, ~CODE_CLASS, ~CODE_LABEL, ~CODE_VALUE, ~RULE,
    "PREGNANT_0", "JUMP", "No pregnancies in males", "9999", '[SEX_0]=1',
   )
   r <- prep_add_missing_codes(NA, study_data = study_data, meta_data =
   meta_data,
     label_col = "LABEL", rules = rules, use_value_labels = FALSE)
   subset(r$ModifiedMetaData, LABEL == "PREGNANT_0", JUMP_LIST)
   subset(meta_data, LABEL == "PREGNANT_0", JUMP_LIST)
   vn <- subset(r$ModifiedMetaData, LABEL == "PREGNANT_0", VAR_NAMES)[[1]]
   expect_snapshot(table(study_data[[vn]], useNA = "always"))
   expect_snapshot(table(r$ModifiedStudyData[[vn]], useNA = "always"))
   r <- prep_add_missing_codes(NA, study_data = study_data, meta_data =
     meta_data, label_col = "LABEL", rules = rules, use_value_labels = FALSE,
     overwrite = TRUE)
   expect_snapshot(table(study_data[[vn]], useNA = "always"))
   expect_snapshot(table(r$ModifiedStudyData[[vn]], useNA = "always"))

  rules <- tibble::tribble(
    ~resp_vars, ~CODE_CLASS, ~CODE_LABEL, ~CODE_VALUE, ~RULE,
    "PREGNANT_0", "JUMP", "No pregnancies in males", "9999", '[SEX_0]="males"',
   )
   r <- prep_add_missing_codes(NA, study_data = study_data,
     meta_data = meta_data,
     label_col = "LABEL", rules = rules, use_value_labels = TRUE,
     overwrite = FALSE)
   expect_snapshot(table(study_data[[vn]], useNA = "always"))
   expect_snapshot(table(r$ModifiedStudyData[[vn]], useNA = "always"))

  rules <- tibble::tribble(
    ~resp_vars, ~CODE_CLASS, ~CODE_LABEL, ~CODE_VALUE, ~RULE,
    "PREGNANT_0", "JUMP", "No pregs in males", "9999", '[v00002]="males"',
   )
   r <- prep_add_missing_codes(NA, study_data = study_data,
     meta_data = meta_data,
     label_col = "LABEL", rules = rules, use_value_labels = TRUE,
     overwrite = FALSE)
   expect_snapshot(table(study_data[[vn]], useNA = "always"))
   expect_snapshot(table(r$ModifiedStudyData[[vn]], useNA = "always"))
   # devtools::load_all(".")

  study_data$v00002 <- ifelse(study_data$v00002 == "0", "females", "males")
  meta_data[meta_data$LABEL == "SEX_0", "VALUE_LABELS"] <- "females|males"
  rules <- tibble::tribble(
    ~resp_vars, ~CODE_CLASS, ~CODE_LABEL, ~CODE_VALUE, ~RULE,
    "PREGNANT_0", "JUMP", "No pregnancies in males", "9999", '[v00002]="males"',
  )
  r <- prep_add_missing_codes(NA, study_data = study_data, meta_data =
                              meta_data,
                              label_col = "LABEL", rules = rules,
                              use_value_labels = TRUE, overwrite = FALSE)
  expect_snapshot(table(study_data[[vn]], useNA = "always"))
  expect_snapshot(table(r$ModifiedStudyData[[vn]], useNA = "always"))
})

#' Utility function to create histograms
#'
#' A helper function for simple histograms.
#'
#' @param plot_data a `data.frame` without missing values
#' @param num_var column name of the numerical or datetime variable
#'                in `plot_data` (if omitted, the first column is assumed to
#'                contain this variable)
#' @param fill_var column name of the categorical variable in `plot_data` which
#'                 will be used for coloring stacked histograms
#' @param facet_var column name of the categorical variable in `plot_data`
#'                  which will be used to create facets
#' @param nbins_max the maximum number of bins for the histogram (see
#'                  `util_optimize_histogram_bins`)
#' @param colors vector of colors, or a single color
#' @param is_datetime if `TRUE`, the x-axis will be adapted for the
#'              datetime format
#' @param is_time if `TRUE`, the x-axis will be adapted for the
#'              time-only format
#'
#' @return a histogram
#'
#' @importFrom ggplot2 ggplot aes theme_minimal xlab ylab geom_col scale_x_continuous scale_x_datetime expansion facet_grid theme element_text
#' @importFrom grDevices colorRampPalette
#'
#' @noRd
util_histogram <- function(plot_data,
                           num_var = colnames(plot_data)[1],
                           fill_var = NULL,
                           facet_var = NULL,
                           nbins_max = 100,
                           colors = "#2166AC",
                           is_datetime = FALSE,
                           is_time = FALSE) { # FIXME: EK: see https://chatgpt.com/s/t_69d517c5f1488191a4afb6fe4763b8cc for time-only-variables at least, the binning can look like overlapping bins triggering warnings like "`position_stack()` requires non-overlapping x intervals." at rendering time. I'm suppressing them now, but fixing would be preferred. # xxxx remove after removing xxxxx from acc_distributions and addressing this issuse correctly, here.
  # compute bin breaks
  # if the plot is faceted, optimize bin breaks for the most frequent category
  bb_opt_sel <- seq_len(nrow(plot_data))
  if (!is.null(facet_var)) {
    # Which category occurs most frequently?
    tab_cat <- table(plot_data[, facet_var])
    most_freq <- names(tab_cat)[which.max(tab_cat)]
    bb_opt_sel <- which(plot_data[, facet_var] == most_freq)
    # We have to ensure that the list of bin breaks includes also the overall
    # minimum and maximum value. Otherwise the faceted plots will be limited to
    # the range of values from the most frequent category.
    bb_opt_sel <- unique(c(bb_opt_sel,
                           which.min(plot_data[, num_var]),
                           which.max(plot_data[, num_var])))
  }

  bin_breaks <- suppressMessages(util_optimize_histogram_bins(
    x = plot_data[bb_opt_sel, num_var],
    nbins_max = nbins_max
  ))
  breaks_x <- bin_breaks[[1]]

  # compute bin heights, prepare data for plotting
  if (!is.null(facet_var) | !is.null(fill_var)) { # histogram with groups
    split_var <- c(facet_var, fill_var)
    split_data <- split(plot_data, plot_data[, split_var])
    split_data <- split_data[vapply(split_data, nrow, integer(1)) > 0]
    plot_data2 <- lapply(seq_along(split_data), function(ll) {
      h1 <- hist(split_data[[ll]][, num_var], plot = FALSE, breaks = breaks_x)
      return(data.frame(histogram_x = h1$mids,
                        histogram_y = h1$counts,
                        split_data[[ll]][1, split_var, drop = FALSE],
                        row.names = NULL))
    })
    plot_data2 <- do.call(rbind, plot_data2)
    if (is_datetime) {
      plot_data2[["histogram_x"]] <- util_parse_date(plot_data2[["histogram_x"]])
    } else if (is_time) {
      plot_data2[["histogram_x"]] <- util_parse_time(plot_data2[["histogram_x"]])
    }
  } else { # plain histogram
    bin_heights <- hist(as.numeric(plot_data[, num_var]),
                        plot = FALSE,
                        breaks = breaks_x)
    if (is_datetime) {
      plot_data2 <- data.frame(histogram_x = util_parse_date(bin_heights$mids),
                               histogram_y = bin_heights$counts)
    } else if (is_time) {
      plot_data2 <- data.frame(histogram_x = util_parse_time(bin_heights$mids),
                               histogram_y = bin_heights$counts)
    } else {
      plot_data2 <- data.frame(histogram_x = bin_heights$mids,
                               histogram_y = bin_heights$counts)
    }
  }
  width_col <- as.numeric(breaks_x)
  width_col <- width_col[2] - width_col[1]

  # create histogram
  if (!is.null(fill_var)) { # histogram with color-coded groups
    if (length(colors) < length(levels(plot_data[, fill_var]))) {
      if (length(colors) == 1) {
        colors <- c("gray90", colors, "gray20")
      }
      colors <- colorRampPalette(colors)(length(levels(plot_data[, fill_var])))
    }
    his <- util_create_lean_ggplot(
      ggplot(data = plot_data2, aes(x = .data[["histogram_x"]],
                                    y = .data[["histogram_y"]],
                                    fill = .data[[fill_var]])) +
        theme_minimal() +
        xlab("") +
        ylab("") +
        geom_col(width = width_col) +
        scale_fill_manual(values = colors, name = ""),
      plot_data2 = plot_data2,
      fill_var = fill_var,
      width_col = width_col,
      colors = colors
    )
  } else { # 'plain' histogram
    his <- util_create_lean_ggplot(
      ggplot(data = plot_data2, aes(x = .data[["histogram_x"]],
                                    y = .data[["histogram_y"]])) +
        theme_minimal() +
        xlab("") +
        ylab("") +
        geom_col(width = width_col,
                 fill = colors[1],
                 color = colors[1]),
      plot_data2 = plot_data2,
      width_col = width_col,
      colors = colors
    )
  }

  if (!is_datetime && !is_time) {
    his <- his + util_create_lean_ggplot(
      scale_x_continuous(expand = expansion(mult = 0.1)))
  } else if (is_datetime && !is_time) {
    his <- his + util_create_lean_ggplot(
      scale_x_datetime(expand = expansion(mult = 0.1)))
  } else {
    # should be time
    his <- his + util_create_lean_ggplot(
      ggplot2::scale_x_time(expand = expansion(mult = 0.1)))
  }

  if (!is.null(facet_var)) {
    his <- his + util_create_lean_ggplot(
      facet_grid(.data[[facet_var]] ~ ., scales = "free_y"),
      facet_var = facet_var) +
        theme(strip.text = element_text(size = 14))
  }

  return(his)
}

#' Simple subsampling for plotting
#'
#' @param data [data.frame] containing the variables to be used for selection.
#' @param x [character] naming the primary variable. In univariate mode,
#'          this is the variable used for histogram-oriented selection. In
#'          bivariate mode, this is the predictor variable used in the
#'          scatterplot-oriented selection and in the linear regression `y ~ x`.
#' @param y [character] optional. Names the second variable.
#'                      If `NULL`, the function runs in univariate mode.
#'                      If provided, the function runs in bivariate mode.
#' @param nmax [integer] Target maximum number of observations to include in
#'                       the selected subset.
#'                       If the number of eligible observations is less than or
#'                       equal to `nmax`, all eligible observations are
#'                       returned.
#' @param pinc [numeric] Proportion controlling the grid-based coverage
#'                       component:
#'                         - In the univariate case, `ceiling(nmax * pinc)`
#'                           bins are created over the range of the unique
#'                           observed values of x.
#'                         - In the bivariate case, `ceiling(nmax * pinc)`
#'                           approximately determines the number of 2D grid
#'                           cells used for support coverage of the `(x, y)`
#'                           distribution.
#'                       Typical values are small, for example `0.02` to `0.10`.
#' @param resids [numeric] Proportion controlling the residual-based inclusion
#'                         component in the bivariate case. The number of
#'                         additional observations selected on the basis of
#'                         large absolute studentized residuals is approximately
#'                         `ceiling(nmax * resids)`.
#'                         This argument is ignored in the univariate case.
#' @param random [logical] flag controlling whether only simple random sampling
#'                         is used:
#'                           - `FALSE`: use the structured subsampling algorithm
#'                           - `TRUE`: use only simple random sampling of size
#'                                     `nmax`

#' @param case_id [character] Optional. Names a variable in `data`
#'                            containing unique case identifiers.
#'                            If `NULL`, row numbers of `data` are used as
#'                            case identifiers.
#' @param seed [integer] Optional. Used to set the random seed before selection.
#'                       This allows reproducible random subsampling.
#' @description
#'
#' # Univariate mode:
#' - create one-row-per-unique-x dataset
#' - create ceiling(nmax * pinc) bins over full x range
#' - select the maximum x from each occupied bin
#' - also include the minimum x
#' - draw ceiling(nmax * (1 - pinc)) random cases
#'   from cases not already selected by the grid step
#'
#' # Bivariate mode:
#' - create one-row-per-unique-(x,y) dataset
#' - create a 2D grid with about ceiling(nmax * pinc) cells
#' - select one case from each occupied cell
#' - also include min/max x and min/max y
#' - compute y ~ x on all complete cases
#' - include the top ceiling(nmax * resids) absolute
#'   studentized residuals
#' - draw ceiling(nmax * (1 - pinc - resids)) random cases
#'   from cases not already selected
#'
#' @returns
#'   vector of case identifiers to include
#'
#' @noRd
util_subsample_cases <- function(
    data,
    x,
    y = NULL,
    nmax = 5000L,
    pinc = 0.05,
    resids = 0.01,
    random = FALSE,
    case_id = NULL,
    seed = NULL) {

  # ---------- checks ----------
  util_expect_data_frame(data)
  ds1 <- data

  meta_data <- prep_study2meta(ds1, level = VARATT_REQUIRE_LEVELS$REQUIRED,
                               convert_factors = FALSE,
                               cumulative = TRUE,
                               guess_missing_codes = FALSE,
                               guess_character = FALSE)

  util_correct_variable_use(x)
  if (!missing(y)) {
    util_correct_variable_use(y)
  }
  util_expect_scalar(nmax,
                     check_type = util_is_numeric_in(min = 1,
                                                     whole_num = TRUE,
                                                     finite = TRUE),
                     error_message = "'nmax' must be a positive integer >= 1.")

  util_expect_scalar(pinc,
                     check_type = util_is_numeric_in(min = 0,
                                                     max = 1,
                                                     whole_num = FALSE,
                                                     finite = TRUE),
                     error_message =
                       "'pinc' must be a single number in [0, 1].")

  util_expect_scalar(resids,
                     check_type = util_is_numeric_in(min = 0,
                                                     max = 1,
                                                     whole_num = FALSE,
                                                     finite = TRUE),
                     error_message =
                       "'resids' must be a single number in [0, 1].")

  util_expect_scalar(random, check_type = is.logical)

  if (!missing(case_id)) {
    util_correct_variable_use(case_id)
  }

  nmax <- as.integer(round(nmax))

  if (!is.null(seed)) withr::local_seed(seed)

  ids_all <- if (is.null(case_id)) seq_len(nrow(data)) else data[[case_id]]

  simple_sample_ids <- function(ids, size) {
    if (size <= 0L || length(ids) == 0L) return(ids[FALSE])
    if (length(ids) <= size) return(ids)
    sample(ids, size = size, replace = FALSE)
  }

  # ---------- eligible data ----------
  needed_vars <- c(x, y)
  needed_vars <- needed_vars[!is.null(needed_vars)]

  cc <- stats::complete.cases(data[, needed_vars, drop = FALSE])
  d <- data[cc, , drop = FALSE]
  d$.case_id <- ids_all[cc]

  nobs <- nrow(d)

  if (nobs == 0L) return(ids_all[FALSE])
  if (nobs <= nmax) return(d$.case_id)

  # ---------- random-only mode ----------
  if (random) {
    return(simple_sample_ids(d$.case_id, nmax))
  }

  # ============================================================
  # A. UNIVARIATE
  # ============================================================
  if (is.null(y)) {
    xvals <- d[[x]]

    # one row per unique x
    ord_unique <- !duplicated(xvals)
    d_unique <- d[ord_unique, , drop = FALSE]
    x_unique <- d_unique[[x]]

    # number of bins
    B <- max(1L, as.integer(ceiling(nmax * pinc)))

    # grid selection
    if (length(x_unique) == 1L || max(x_unique, na.rm = TRUE) ==
        min(x_unique, na.rm = TRUE)) {
      i_grid_case <- d_unique$.case_id[1L]
    } else {
      breaks <- seq(min(x_unique, na.rm = TRUE),
                    max(x_unique, na.rm = TRUE),
                    length.out = B + 1L)

      bin_id <- cut(x_unique,
                    breaks = breaks,
                    include.lowest = TRUE,
                    right = TRUE,
                    labels = FALSE)

      split_idx <- split(seq_along(x_unique), bin_id, drop = TRUE)

      max_idx_local <- vapply(split_idx, function(idx) {
        idx[which.max(x_unique[idx])]
      }, integer(1L))

      i_grid_case <- d_unique$.case_id[max_idx_local]
    }

    # make sure minimum x is included
    min_case <- d_unique$.case_id[which.min(x_unique)[1L]]
    i_case <- unique(c(i_grid_case, min_case))

    # random fill
    nrand <- as.integer(ceiling(nmax * (1 - pinc)))
    remaining <- d[!(d$.case_id %in% i_case), , drop = FALSE]
    srand <- simple_sample_ids(remaining$.case_id, nrand)

    return(unique(c(i_case, srand)))
  }

  # ============================================================
  # B. BIVARIATE
  # ============================================================
  xvals <- d[[x]]
  yvals <- d[[y]]

  # one row per unique (x, y)
  ord_unique <- !duplicated(d[, c(x, y), drop = FALSE])
  d_unique <- d[ord_unique, , drop = FALSE]
  x_unique <- d_unique[[x]]
  y_unique <- d_unique[[y]]

  # target number of grid cells
  B <- max(1L, as.integer(ceiling(nmax * pinc)))

  # choose a roughly square grid
  nx <- max(1L, floor(sqrt(B)))
  ny <- max(1L, ceiling(B / nx))

  # grid selection
  if ((length(x_unique) == 1L || max(x_unique, na.rm = TRUE) == min(x_unique, na.rm = TRUE)) &&
      (length(y_unique) == 1L || max(y_unique, na.rm = TRUE) == min(y_unique, na.rm = TRUE))) {

    i_grid_case <- d_unique$.case_id[1L]

  } else if (max(x_unique, na.rm = TRUE) == min(x_unique, na.rm = TRUE)) {

    # x constant -> bin only on y
    breaks_y <- seq(min(y_unique, na.rm = TRUE),
                    max(y_unique, na.rm = TRUE),
                    length.out = ny + 1L)

    by <- cut(y_unique,
              breaks = breaks_y,
              include.lowest = TRUE,
              right = TRUE,
              labels = FALSE)

    split_idx <- split(seq_along(y_unique), by, drop = TRUE)
    rep_idx <- vapply(split_idx, function(idx) idx[1L], integer(1L))
    i_grid_case <- d_unique$.case_id[rep_idx]

  } else if (max(y_unique, na.rm = TRUE) == min(y_unique, na.rm = TRUE)) {

    # y constant -> bin only on x
    breaks_x <- seq(min(x_unique, na.rm = TRUE),
                    max(x_unique, na.rm = TRUE),
                    length.out = nx + 1L)

    bx <- cut(x_unique,
              breaks = breaks_x,
              include.lowest = TRUE,
              right = TRUE,
              labels = FALSE)

    split_idx <- split(seq_along(x_unique), bx, drop = TRUE)
    rep_idx <- vapply(split_idx, function(idx) idx[1L], integer(1L))
    i_grid_case <- d_unique$.case_id[rep_idx]

  } else {

    breaks_x <- seq(min(x_unique, na.rm = TRUE),
                    max(x_unique, na.rm = TRUE),
                    length.out = nx + 1L)
    breaks_y <- seq(min(y_unique, na.rm = TRUE),
                    max(y_unique, na.rm = TRUE),
                    length.out = ny + 1L)

    bx <- cut(x_unique,
              breaks = breaks_x,
              include.lowest = TRUE,
              right = TRUE,
              labels = FALSE)
    by <- cut(y_unique,
              breaks = breaks_y,
              include.lowest = TRUE,
              right = TRUE,
              labels = FALSE)

    cell_id <- paste(bx, by, sep = "_")
    split_idx <- split(seq_along(x_unique), cell_id, drop = TRUE)
    rep_idx <- vapply(split_idx, function(idx) idx[1L], integer(1L))
    i_grid_case <- d_unique$.case_id[rep_idx]
  }

  # make sure min/max x and min/max y are included
  boundary_cases <- c(
    d_unique$.case_id[which.min(x_unique)[1L]],
    d_unique$.case_id[which.max(x_unique)[1L]],
    d_unique$.case_id[which.min(y_unique)[1L]],
    d_unique$.case_id[which.max(y_unique)[1L]]
  )

  i_case <- unique(c(i_grid_case, boundary_cases))

  # residual-extreme cases from y ~ x
  ir_case <- d$.case_id[FALSE]

  regression_feasible <-
    resids > 0 &&
    nrow(d) >= 3L &&
    is.finite(stats::var(d[[x]], na.rm = TRUE)) &&
    stats::var(d[[x]], na.rm = TRUE) > 0

  if (regression_feasible) {
    fit <- try(stats::lm(stats::as.formula(paste(y, "~", x)), data = d), silent = TRUE)

    if (!util_is_try_error(fit)) {
      rstud <- try(abs(stats::rstudent(fit)), silent = TRUE)

      if (!util_is_try_error(rstud) &&
          is.numeric(rstud) &&
          length(rstud) == nrow(d) &&
          any(is.finite(rstud))) {

        rstud[!is.finite(rstud)] <- -Inf
        nres <- as.integer(ceiling(nmax * resids))

        if (nres > 0L) {
          ord <- order(rstud, decreasing = TRUE)
          keep_n <- min(nres, sum(is.finite(rstud)))
          keep_local <- ord[seq_len(keep_n)]
          ir_case <- d$.case_id[keep_local]
        }
      }
    }
  }

  i_case <- unique(c(i_case, ir_case))

  # random fill
  nrand <- as.integer(ceiling(nmax * (1 - pinc - resids)))
  if (nrand < 0L) nrand <- 0L

  remaining <- d[!(d$.case_id %in% i_case), , drop = FALSE]
  srand <- simple_sample_ids(remaining$.case_id, nrand)

  unique(c(i_case, srand))
}

#' Attach attributes to an object and return it
#'
#' @param x the object
#' @param ... named arguments, each becomes an attributes
#'
#' @return `x`, having the desired attributes attached
#'
#' @family process_functions
#' @concept process
#' @noRd
util_attach_attr <- function(x, ...) {
  atts <- list(...)
  attributes(x)[names(atts)] <- atts
  x
}

test_that("literal bracketed variable names are kept", {
  valid_names <- c("AGE_0", "SEX_0")

  expect_equal(
    util_expand_pattern_rules("[AGE_0] > 55", valid_names),
    util_attach_attr("[AGE_0] > 55", mismatches = list())
  )
})

test_that("wildcards expand matching variable names", {
  valid_names <- c("SEX_0", "PREGNANT_0", "PREGNANT_1", "AGE_0")

  expect_equal(
    util_expand_pattern_rules("[PREGNANT_*]", valid_names),
    util_attach_attr(c("[PREGNANT_0]", "[PREGNANT_1]"), mismatches = list())
  )
})

test_that("digit wildcard expands one digit only", {
  valid_names <- c("PREGNANT_0", "PREGNANT_09", "PREGNANT_x")

  expect_equal(
    util_expand_pattern_rules("[PREGNANT_#]", valid_names),
    util_attach_attr("[PREGNANT_0]", mismatches = list())
  )
})

test_that("two digit wildcard expands exactly two digits", {
  valid_names <- c("PREGNANT_0", "PREGNANT_09", "PREGNANT_99")

  expect_equal(
    util_expand_pattern_rules("[PREGNANT_##]", valid_names),
    util_attach_attr(c("[PREGNANT_09]", "[PREGNANT_99]"), mismatches = list())
  )
})

test_that("one or two digit wildcard expands one or two digits", {
  valid_names <- c("PREGNANT_0", "PREGNANT_09", "PREGNANT_123")

  expect_equal(
    util_expand_pattern_rules("[PREGNANT_#?]", valid_names),
    util_attach_attr(c("[PREGNANT_0]", "[PREGNANT_09]"), mismatches = list())
  )
})

test_that("captures are reused across variables", {
  valid_names <- c(
    "PREGNANT_0", "PREGNANT_1",
    "AGE_0", "AGE_1", "AGE_2"
  )

  expect_equal(
    util_expand_pattern_rules(
      "[PREGNANT_{W:#}] = \"yes\" and [AGE_{W}] > 55",
      valid_names
    ),
    util_attach_attr(c(
      "[PREGNANT_0] = \"yes\" and [AGE_0] > 55",
      "[PREGNANT_1] = \"yes\" and [AGE_1] > 55"
    ), mismatches = list())
  )
})

test_that("captures do not create mixed combinations", {
  valid_names <- c("PREGNANT_0", "PREGNANT_1", "AGE_0")

  expect_equal(
    util_expand_pattern_rules(
      "[PREGNANT_{W:#}] = \"yes\" and [AGE_{W}] > 55",
      valid_names
    ),
    util_attach_attr("[PREGNANT_0] = \"yes\" and [AGE_0] > 55", mismatches = list())
  )
})

test_that("capture arithmetic with minus offset works", {
  valid_names <- c("AGE_0", "AGE_1", "AGE_2")

  expect_equal(
    util_expand_pattern_rules("[AGE_{CE:#}] < [AGE_{CE-1}]", valid_names),
    util_attach_attr(c("[AGE_1] < [AGE_0]", "[AGE_2] < [AGE_1]"), mismatches = list())
  )
})

test_that("capture arithmetic with plus offset works", {
  valid_names <- c("AGE_0", "AGE_1", "AGE_2")

  expect_equal(
    util_expand_pattern_rules("[AGE_{CE:#}] < [AGE_{CE+1}]", valid_names),
    util_attach_attr(c("[AGE_0] < [AGE_1]", "[AGE_1] < [AGE_2]"), mismatches = list())
  )
})

test_that("invalid arithmetic results are dropped", {
  valid_names <- c("AGE_0", "AGE_1")

  expect_equal(
    util_expand_pattern_rules("[AGE_{CE:#}] < [AGE_{CE-1}]", valid_names),
    util_attach_attr("[AGE_1] < [AGE_0]", mismatches = list())
  )
})

test_that("arithmetic on non-integer captures drops generated rule", {
  valid_names <- c("AGE_A", "AGE_B")

  expect_equal(
    util_expand_pattern_rules("[AGE_{CE:?}] < [AGE_{CE-1}]", valid_names),
    util_attach_attr(character(0), mismatches = list())
  )
})

test_that("escaped mini-language characters are treated literally", {
  valid_names <- c("Pregnant? 0", "Pregnant* 0", "AGE_0")

  expect_equal(
    util_expand_pattern_rules(
      "[Pregnant\\? {W:#}] and [AGE_{W}]",
      valid_names
    ),
    util_attach_attr("[Pregnant? 0] and [AGE_0]", mismatches = list())
  )
})

test_that("variable labels with spaces and punctuation work", {
  valid_names <- c("Age at exam 0", "Pregnant? 0")

  expect_equal(
    util_expand_pattern_rules(
      "[Pregnant\\? {W:#}] and [Age at exam {W}]",
      valid_names
    ),
    util_attach_attr("[Pregnant? 0] and [Age at exam 0]", mismatches = list())
  )
})

test_that("valid names must not contain square brackets", {
  expect_warning(
    util_expand_pattern_rules("[AGE_0]", c("AGE_0", "AGE[1]")),
    "valid names must not contain"
  )
})

test_that("undefined capture references fail clearly", {
  expect_error(
    util_expand_pattern_rules("[AGE_{W}]", "AGE_0"),
    "used before being defined"
  )
})

test_that("unmatched rules return character zero", {
  expect_equal(
    util_expand_pattern_rules("[DOES_NOT_EXIST_*]", "AGE_0"),
    util_attach_attr(character(0), mismatches = list("DOES_NOT_EXIST_*"))
  )
})

test_that("expanded cross-item pattern rules get unique IDs and labels", {
  meta_data <- data.frame(
    VAR_NAMES = c("sbp_0", "dbp_0", "age_0", "age_1", "age_2"),
    LABEL = c("sbp_0", "dbp_0", "age_0", "age_1", "age_2"),
    LONG_LABEL = c("sbp_0", "dbp_0", "age_0", "age_1", "age_2"),
    ORIGINAL_VAR_NAMES = c("sbp_0", "dbp_0", "age_0", "age_1", "age_2"),
    ORIGINAL_LABEL = c("sbp_0", "dbp_0", "age_0", "age_1", "age_2"),
    stringsAsFactors = FALSE
  )
  meta_data_cross_item <- data.frame(
    CHECK_ID = c("bloodpressure", "age"),
    CHECK_LABEL = c("bloodpressure", "age"),
    CONTRADICTION_TERM = c(
      "[sbp_0] > [dbp_0]",
      "[age_{W:#}] < [age_{W-1}]"
    ),
    stringsAsFactors = FALSE
  )

  out <- util_normalize_cross_item(
    meta_data = meta_data,
    meta_data_cross_item = meta_data_cross_item
  )

  expect_equal(out[[CHECK_ID]], c("bloodpressure", "age_1", "age_2"))
  expect_equal(out[[CHECK_LABEL]], c("bloodpressure", "age #1", "age #2"))
  expect_equal(
    out[[CONTRADICTION_TERM]],
    c("[sbp_0] > [dbp_0]", "[age_1] < [age_0]", "[age_2] < [age_1]")
  )
})

#' Bind data frames row-based
#'
#' if not all data frames share  all columns, missing columns will be filled with
#' `NA`s.
#'
#' @param ... [data.frame] none more more data frames
#' @param data_frames_list [list] optional, a list of data frames
#'
#' @return [data.frame] all data frames appended
#'
#' @examples
#' \dontrun{
#' util_rbind(head(cars), head(iris))
#' util_rbind(head(cars), tail(cars))
#' util_rbind(head(cars)[, "dist", FALSE], tail(cars)[, "speed", FALSE])
#' }
#'
#' @family data_management
#' @concept process
#' @noRd
util_rbind <- function(..., data_frames_list = list()) {
  data_frames_list <- c(list(...), data_frames_list)
  data_frames_list <- data_frames_list[!vapply(data_frames_list, is.null,
                                              FUN.VALUE = logical(1))]
  data_frames_list <-
    lapply(data_frames_list, util_expect_data_frame, dont_assign = TRUE)

  all_cn <- lapply(data_frames_list, colnames)

  tnsltd <- FALSE

  if (length(all_cn) > 0) {
    tnsltd <- inherits(all_cn[[1]], "dataquieR_translated")
  }

  if (tnsltd) {
    util_stop_if_not(
      "Internal error, sorry. Please report. Combining incompatibly translated column names (1)" =
        all(vapply(all_cn, inherits, "dataquieR_translated", FUN.VALUE = logical(1))))

    ns <- unique(vapply(all_cn, attr, "ns", FUN.VALUE = character(1)))
    lang <- unique(vapply(all_cn, attr, "lang", FUN.VALUE = character(1)))

    util_stop_if_not(
      "Internal error, sorry. Please report. Combining incompatibly translated column names (2)" =
        length(ns) == 1 &&
        length(lang) == 1)
  }


  all_cols <- unique(unlist(lapply(data_frames_list, colnames),
                            recursive = TRUE))

  data_frames_list <- lapply(data_frames_list, function(dfr) {
    if (nrow(dfr) > 0) {
      for (cl in setdiff(all_cols, colnames(dfr))) {
        dfr[[cl]] <- NA
      }
      dfr
    } else {
      NULL
    }
  })

  all_data_types <-
    vapply(setNames(nm = all_cols), function(cl) {
      dt <- unique(lapply(lapply(data_frames_list, `[[`, cl),
             attr,
             WELL_KNOWN_META_VARIABLE_NAMES$DATA_TYPE))
      if (length(dt) == 1) {
        r <- dt[[1]]
      } else {
        # util_warning("Internal error, sorry, please report: data types of a combined output table incompatible")
        r <- NA_character_
      }
      if (is.null(r)) {
        r <- NA_character_
      }
      r
    }, FUN.VALUE = character(1)) # This is used also from onLoad, so maybe, DATA_TYPE has not been defined, yet

  r <- do.call(rbind.data.frame, data_frames_list)

  if (length(all_data_types) == ncol(r)) {
    for (cl in colnames(r)) {
      attr(r[[cl]], WELL_KNOWN_META_VARIABLE_NAMES$DATA_TYPE) <-
        all_data_types[[cl]]
    }
  }

  if (tnsltd) {
    util_translated_colnames(r) <-
      util_attach_attr(
        colnames(r),
        class = "dataquieR_translated",
        names = unname(unlist(unique(lapply(all_cn, function(x) setNames(nm = as.character(x), ifelse(names(x) == "", as.character(x), names(x)))[colnames(r)])))),
        ns = ns,
        lang = lang
      )
  }

  r
}

#' Map a vector of values based on an assignment table
#'
#' @param values [vector] the vector
#' @param mapping_table [data.frame] a table with the mapping table
#' @param from [character] the name of the column with the "old values"
#' @param to [character] the name of the column with the "new values"
#' @param default [character] either one character or on character per value,
#'                            being used, if an entry from `values` is not
#'                            in the `from` column in `mapping_table
#'
#' @return the mapped values
#'
#' @seealso [dplyr::recode]
#'
#' @importFrom dplyr recode
#'
#' @family mapping
#' @concept process
#' @noRd
util_recode <- function(values, mapping_table, from, to, default = NULL) {
  util_expect_scalar(values,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     allow_na = TRUE)
  util_expect_scalar(from, check_type = is.character)
  util_expect_scalar(to, check_type = is.character)
  if (length(default) == 1) {
    util_expect_scalar(default,
                       allow_na = TRUE, check_type = is.character)
  } else {
    util_expect_scalar(default,
                       allow_na = TRUE,
                       allow_null = TRUE,
                       allow_more_than_one = TRUE,
                       min_length = length(values),
                       max_length = length(values))
  }
  util_expect_data_frame(mapping_table, c(from, to))
  mapping_list <- setNames(mapping_table[[to]], nm = mapping_table[[from]])
  mapping_list <- c(list(.x = values),
                    mapping_list,
                    list(.default = default))
  do.call(recode, mapping_list)
}

#' Tabulate a vector
#'
#' does the same as `as.data.frame(table(x))` but
#' guarantees a data frame with two columns is returned
#'
#' @param Var1 vector to ta tabulate
#'
#' @return a data frame with columns `Var1` and `Freq`
#'
#' @family data_management
#' @concept robustness
#' @noRd
util_table_of_vct <- function(Var1) {
  util_expect_scalar(Var1,
                     allow_more_than_one = TRUE,
                     allow_na = TRUE,
                     allow_null = TRUE)
  r <- as.data.frame(table(Var1))
  if (!identical(colnames(r), c("Var1", "Freq"))) {
    r <- data.frame(Var1 = integer(0), Freq = integer(0))
  }
  r
}

#' Maps label column metadata on study data variable names
#'
#' Maps a certain label column from the metadata
#' to the study data frame.
#'
#' @param label_col the variable of the metadata that contains the variable
#'                  names of the study data
#' @param study_data the name of the data frame that contains the measurements
#' @param meta_data the name of the data frame that contains metadata attributes
#'                  of study data
#'
#' @return [list] with slot `df` with a study data frame with mapped column
#'                names
#'
#' @importFrom utils adist
#'
#' @family mapping
#' @concept metadata_management
#' @noRd
util_map_all <- function(label_col = VAR_NAMES,
                         study_data,
                         meta_data) {

  ################# checks ##################

  if (!nrow(meta_data)) {
    util_error("%s is empty", sQuote("meta_data"))
  }

  if (length(label_col) != 1 || !is.character(label_col)) {
    util_error(
      c("label_col must be exactly 1 metadata attribute,",
        "neither a vector nor NULL."), applicability_problem = TRUE)
  }

  if (!label_col %in% colnames(meta_data)) {
    fuzzy_match <- vapply(
      label_col,
      function(v) {
        colnames(meta_data)[which.min(adist(trimws(v),
                                            trimws(colnames(meta_data)),
                                            ignore.case = TRUE, fixed = TRUE))]
      },
      ""
    )
    util_error("label_col %s not found in metadata. Did you mean %s?",
               dQuote(label_col), dQuote(fuzzy_match),
               applicability_problem = TRUE)
  }

  if (!VAR_NAMES %in% colnames(meta_data)) {
    util_error("VAR_NAMES not found in metadata.",
               applicability_problem = TRUE)
  }

  if (any(duplicated(meta_data[[VAR_NAMES]]), na.rm = TRUE)) {
    util_error(
      c("The following variable names are duplicated in the metadata",
        "and cannot be used as label therefore: %s"),
      paste0(collapse = ", ",
             dQuote(unique(meta_data[[VAR_NAMES]][(
               duplicated(meta_data[[VAR_NAMES]]))]))),
      applicability_problem = TRUE
    )
  }

  if (any(duplicated(meta_data[[label_col]]), na.rm = TRUE)) {
    util_error(
      c("The following %s are duplicated in the metadata",
        "and cannot be used as label therefore: %s"),
      sQuote(label_col),
      paste0(collapse = ", ",
             dQuote(unique(meta_data[[label_col]][(
               duplicated(meta_data[[label_col]]))]))),
      applicability_problem = TRUE
    )
  }

  if (any(is.na(meta_data[[VAR_NAMES]]))) {
    util_error(
      c("For the following variables, some variable",
        "names are missing in the metadata: %s"),
      paste0("Variable No. #", which(is.na(meta_data[[VAR_NAMES]])),
             collapse = ", "), applicability_problem = TRUE
    )
  }

  if (any(is.na(meta_data[[label_col]]))) {
    util_error(
      c("For the following variables, some %s are missing in the metadata and",
        "cannot be used as label therefore: %s"),
      sQuote(label_col),
      paste0("Variable No. #", which(is.na(meta_data[[label_col]])),
             collapse = ", "),
      applicability_problem = TRUE
    )
  }

  ################# mapping ##################

  # select only relevant variables from study_data
  counter <- sum(!(colnames(study_data) %in% meta_data[[VAR_NAMES]]))
  if (ncol(study_data) > 0) {
    lost <- counter / ncol(study_data)
  } else {
    if (counter > 0) {
      lost <- Inf
    } else {
      lost <- 0
    }
  }

  if (!(exists(".called_in_pipeline") && .called_in_pipeline) &&
      lost > 0 && !getOption("dataquieR.ELEMENT_MISSMATCH_CHECKTYPE",
                             dataquieR.ELEMENT_MISSMATCH_CHECKTYPE_default) %in%
      c("none", "subset_m")) {
    util_warning(
      "Lost %g%% of the study data because of missing/not assignable metadata",
      round(lost * 100, 1),
      applicability_problem = TRUE, integrity_indicator = "int_sts_element")
    vars <- colnames(study_data)[!(colnames(study_data) %in%
                                     meta_data[[VAR_NAMES]])]
    util_message(
      paste("Did not find any metadata for the following",
             "variables from the study data: %s"),
      paste0(dQuote(vars),
             collapse = ", "), integrity_indicator = "int_sts_element",
      varname = vars
    )
  }

  counter <- sum(!(meta_data[[VAR_NAMES]] %in% colnames(study_data)))
  if (nrow(meta_data) > 0) {
    unlost <- counter / nrow(meta_data)
  } else {
    if (counter > 0) {
      unlost <- Inf
    } else {
      unlost <- 0
    }
  }

  if (!(exists(".called_in_pipeline") && .called_in_pipeline) &&
      unlost > 0 && !getOption("dataquieR.ELEMENT_MISSMATCH_CHECKTYPE",
                               dataquieR.ELEMENT_MISSMATCH_CHECKTYPE_default) %in%
      c("none", "subset_u")) {
    util_warning(
      "Lost %g%% of the metadata because of missing/not assignable study data",
      round(unlost * 100, 1),
      applicability_problem = TRUE, integrity_indicator = "int_sts_element",
      intrinsic_applicability_problem = TRUE)
    vars <- meta_data[[VAR_NAMES]][!(meta_data[[VAR_NAMES]] %in%
                                       colnames(study_data))]
    util_message(
      paste("Found metadata for the following variables",
            "not found in the study data: %s"),
      paste0(dQuote(vars),
             collapse = ", "), integrity_indicator = "int_sts_element",
      varname = vars
    )
  }

  ds1 <- study_data[, colnames(study_data) %in% meta_data[[VAR_NAMES]],
                    drop = FALSE]

  # select variable name from meta_data
  cn <- util_map_labels(colnames(ds1), meta_data, label_col, ifnotfound = "")

  invalid <- trimws(cn) == ""
  invalid[is.na(invalid)] <- TRUE

  dups <- duplicated(cn)
  dups[is.na(dups)] <- TRUE

  e <- character(0)
  if (any(invalid)) {
    e <- c(e, sprintf(
      "Mapping of metadata on study data yielded invalid variable labels: %s",
      paste0(dQuote(cn[invalid]), collapse = ", ")
    ))
  }

  if (any(dups)) {
    e <- c(e, sprintf(
      "Mapping of metadata on study data yielded duplicated variable labels: %s",
      paste0(dQuote(cn[dups]), collapse = ", ")
    ))
  }

  if (length(e) > 0) {
    util_error(paste0(e, collapse = "\n"), applicability_problem = TRUE)
  }

  colnames(ds1) <- cn

  return(list(df = ds1))
}

test_that("util_map_all works", {
  skip_on_cran()
  
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  md <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = DATA_TYPES$FLOAT,
    LABEL = LETTERS,
    MISSING_LIST = "",
    nums = seq_along(letters)
  )
  sd <- as.data.frame(lapply(setNames(nm = letters), paste, 1:10),
                      stringsAsFactors = FALSE)
  mapped <- util_map_all(
    label_col = LABEL,
    study_data = sd,
    meta_data = md
  )

  expect_equal(colnames(mapped$df), LETTERS)

  mapped <- util_map_all(
    label_col = "nums",
    study_data = sd,
    meta_data = md
  )

  expect_error(util_map_all(
      label_col = c("nums", "kkk"),
      study_data = sd,
      meta_data = md
    ),
    regexp =
      paste("label_col must be exactly 1 metadata attribute,",
            "neither a vector nor NULL.")
  )

  expect_error(util_map_all(
    label_col = c(),
    study_data = sd,
    meta_data = md
  ),
  regexp =
    paste("label_col must be exactly 1 metadata attribute,",
          "neither a vector nor NULL.")
  )

  expect_error(util_map_all(
    label_col = "NO NO NO NO NO",
    study_data = sd,
    meta_data = md
  ), regexp = paste("label_col .+NO NO NO NO NO.+ not found in metadata.",
                    "Did you mean .+VAR_NAMES.+"))

  expect_error(util_map_all(
    label_col = "speed",
    study_data = sd,
    meta_data = cars
  ), regexp = paste(".*VAR_NAMES not found in metadata."))

  expect_error(util_map_all(
    label_col = "DATA_TYPE",
    study_data = sd,
    meta_data = md
  ), regexp =
    paste("The following .+DATA_TYPE.+ are duplicated in the",
          "metadata and cannot be used as label therefore: .+float.+"))

  mdx <- md
  mdx$VAR_NAMES[[5]] <- mdx$VAR_NAMES[[1]]
  expect_error(util_map_all(
    label_col = LABEL,
    study_data = sd,
    meta_data = mdx
  ), regexp =
    paste("The following variable names are duplicated in the",
          "metadata and cannot be used as label therefore: .+a.+"))

  mdx <- md
  mdx$VAR_NAMES[[5]] <- NA
  expect_error(util_map_all(
    label_col = LABEL,
    study_data = sd,
    meta_data = mdx
  ), regexp = paste("For the following variables, some variable names are",
                    "missing in the metadata: Variable No. #5"))

  mdx <- md
  mdx$LABEL[[5]] <- NA
  expect_error(util_map_all(
    label_col = LABEL,
    study_data = sd,
    meta_data = mdx
  ), regexp = paste("For the following variables, some .+LABEL.+ are missing",
                    "in the metadata and cannot be used as label therefore:",
                    "Variable No. #5"))

  mdx <- md
  mdx$VAR_NAMES[[1]] <- "xxx"
  mdx$VAR_NAMES[[4]] <- "yyy"

  withr::with_options(list(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact"),
                      expect_warning(
    expect_warning(
      expect_message2(
        expect_message2(
          invisible(capture.output(util_map_all(
            label_col = LABEL,
            study_data = sd,
            meta_data = mdx
          ))),
          regexp = paste("Did not find any metadata for the following",
                          "variables from the study data: .+a.+, .+d.+"),
        perl = TRUE),
      regexp = paste("Found metadata for the following variables not",
                     "found in the study data: .+xxx.+, .+yyy.+"),
      perl = TRUE),
      regexp = paste("Lost 7.7% of the study data because",
                     "of missing/not assignable metadata"),
      perl = TRUE
    ),
    regexp = paste("Lost 7.7% of the metadata because of",
                   "missing/not assignable study data"),
    perl = TRUE
  ))

  mdx <- md
  mdx$LABEL[[5]] <- "    "
  expect_error(util_map_all(
    label_col = LABEL,
    study_data = sd,
    meta_data = mdx
  ), regexp = paste("Mapping of metadata on study data yielded invalid",
                    "variable labels:"))

  expect_equal(colnames(mapped$df), as.character(seq_along(letters)))

})

#' Check, if `x` is a try-error
#'
#' @param x object to test
#'
#' @return [logical()] if it is a try-error
#' @noRd
util_is_try_error <- function(x) {
  inherits(x, "try-error")
}

#' Extract condition from try error
#'
#' @param x the try-error object
#'
#' @return [condition] of the try-error
#' @noRd
util_condition_from_try_error <- function(x) {
  if (!util_is_try_error(x)) {
    util_error("Internal error %s: Not a try-error. Please report, sorry.",
               dQuote(class(x)))
  } else {
    attr(x, "condition")
  }
}

test_that("util_condition_from_try_error works", {
  x <- util_condition_from_try_error(try(stop("Test"), silent = TRUE))
  expect_s3_class(x, "simpleError")
  expect_equal(conditionMessage(x), "Test")
})

#' Names of a `dataquieR` report object (v2.0)
#'
#' @param x the result object
#'
#' @return the names
#'
#' @method dimnames dataquieR_resultset2
#' @export
dimnames.dataquieR_resultset2 <- function(x) {
  matrix_list <- attr(x, "matrix_list")
  row_indices <- attr(matrix_list, "row_indices")
  col_indices <- attr(matrix_list, "col_indices")
  list(names(sort(row_indices)), # TODO: sort also in Square2's access function
       names(sort(col_indices)), # TODO: sort also in Square2's access function
       resnames(x)
       )
}

#' Print a `dataquieR` summary
#'
#' @param x the `dataquieR` summary, see [summary()] and [dq_report2()]
#' @param ... not yet used
#' @param dont_print suppress the actual printing, just return a printable
#'                   object derived from `x`
#' @param grouped_by define the columns of the resulting matrix. It can be either
#' "call_names", one column per function,  or "indicator_metric", one column per indicator
#' or both c("call_names", "indicator_metric"). The last combination is the default
#' @param folder_of_report a named vector with the location of variable and
#'                         `call_names`
#' @param vars_to_include `"study"`, `"ssi"` or `c("study", "ssi")`. variables
#'        to include
#' @return invisible html object
#' @export
#'
print.dataquieR_summary <- function(x, ..., grouped_by =
                                      c("call_names",
                                        "indicator_metric"),
                                    dont_print = FALSE,
                                    folder_of_report = NULL,
                                    vars_to_include = c("study")) {
  util_ensure_suggested(pkg = c("htmltools",
                                "DT", "rmarkdown",
                                "markdown"),
                        goal = "generate plain HTML-summaries.")
  util_expect_scalar(vars_to_include,
                     allow_more_than_one = TRUE,
                     min_length = 1,
                     max_length = 2,
                     check_type = is.character)
  util_match_arg(vars_to_include, c("study", "ssi"), several_ok = TRUE)

  #check and normalize grouped_by
  util_expect_scalar(grouped_by, allow_more_than_one = TRUE,
                     min_length = 1, check_type = is.character,
                     error_message = "not allowed value for grouped_by")
  grouped_by <- util_match_arg(grouped_by, several_ok = TRUE)
  grouped_by <- unique(grouped_by)

  x <- util_reclassify_dataquieR_summary(x)

  y <- util_render_table_dataquieR_summary(x, grouped_by = grouped_by,
                                           folder_of_report = folder_of_report,
                                           vars_to_include = vars_to_include)

  if (!dont_print) {
    print(y)
  } else {
    invisible(y)
  }
}

#' Set up an RStudio job
#'
#' Also defines a progress function and a `progress_msg` function in
#' the caller's environment.
#'
#' In `RStudio` its job system will be used, for `shiny::withProgress`
#' based calls, this will require min and max being set to 0 and 1 (defaults).
#' If `cli` is available, it will be used, in all other cases, just `message`s
#' will be created.
#'
#' @param job_name a name for the job
#' @param n number of steps (needed, if, e.g., `progressr` should be used)
#'
#' @return
#' list: the `progress` function and the `progress_msg` function
#'
#' @examples
#' \dontrun{
#'   test <- function() {
#'     util_setup_rstudio_job("xx")
#'     Sys.sleep(5)
#'     progress(50)
#'     progress_msg("halfway through")
#'     Sys.sleep(5)
#'     progress(100)
#'     Sys.sleep(1)
#'   }
#'   test()
#' }
#'
#' @family process_functions
#' @concept reporting
#' @noRd
util_setup_rstudio_job <- function(job_name = "Job", n) {

  if (missing(n)) {
    util_error(
      c("Internal error, sorry, please report: arguemnt %s is mandatory",
        "for %s"),
      sQuote("n"),
      sQuote("util_setup_rstudio_job"))
  }

  # Find context ----

  progress_init_fkt <-
    getOption("dataquieR.progress_init_fkt", dataquieR.progress_init_fkt_default)

  progress_fkt <-
    getOption("dataquieR.progress_fkt", dataquieR.progress_fkt_default)

  progress_msg_fkt <-
    getOption("dataquieR.progress_msg_fkt", dataquieR.progress_msg_fkt_default)

  is_shiny <- suppressWarnings(util_ensure_suggested("shiny", err = FALSE)) &&
    (!is.null(shiny::getDefaultReactiveDomain()))

  is_rstudio <-
    !is_shiny &&
    suppressWarnings(util_ensure_suggested("rstudioapi", err = FALSE)) &&
    rstudioapi::isAvailable()

  is_rstudio <- is_rstudio && util_really_rstudio()

  is_rstudio <- is_rstudio && !is.function(progress_msg_fkt)
  is_rstudio <- is_rstudio && !is.function(progress_fkt)

  is_cli <-
    !is_shiny &&
    !is_rstudio &&
    suppressWarnings(util_ensure_suggested("cli", err = FALSE))

  is_cli <- is_cli && !is.function(progress_msg_fkt)
  is_cli <- is_cli && !is.function(progress_fkt)

  p <- parent.frame()

  # RStudio: remove existing job, if a function calls us repeatedly ----
  if (is_rstudio && exists("rstudiojob", envir = p) &&
      !is.null(get("rstudiojob", envir = p))) {
    try(rstudioapi::jobRemove(get("rstudiojob", envir = p)), silent = TRUE)
    rstudiojob <- NULL
    assign("rstudiojob", rstudiojob, envir = p)
  }
  rstudiojob <- NULL

  # RStudio: Create new Job ----
  if (is_rstudio &&
      !is.function(progress_fkt) &&
      !is.function(progress_msg_fkt) &&
      !is.function(progress_init_fkt)
      ) {
    try({
      rstudiojob <- rstudioapi::jobAdd(job_name,
                                      progressUnits = 100L)
    }, silent = TRUE)
  }
  assign("rstudiojob", rstudiojob, envir = p) # always

  if (is_cli &&
      !is.function(progress_fkt) &&
      !is.function(progress_msg_fkt) &&
      !is.function(progress_init_fkt)) {
    cli::cli_progress_bar("Task...", total = 100, .envir = p)
  }

  # RStudio: Hook to remove job, if our caller is left ----
  do.call("on.exit", # TODO: see and maybe use withr::defer_parent?
          list(quote({
            Sys.sleep(1) # some strange concurrency problem with RStudio
            if (exists(".is_testing", inherits = FALSE)) {
              rm(".is_testing", inherits = FALSE)
            }
            if (!is.null(rstudiojob)) { # nocov start
              try({
                rstudioapi::jobRemove(rstudiojob)
                rstudiojob <- NULL
              }, silent = TRUE)
            } # nocov end
          }), add = TRUE),
          envir = p)

  assign(".is_testing",
    suppressWarnings(util_ensure_suggested("testthat", err = FALSE)) &&
    testthat::is_testing(),
    envir = p)

  # Define default progress functions ----
  progress <- function(percent, is_rstudio, is_shiny, is_cli, e) {
    util_call_progress_hooks("progress", percent = percent)
    if (isTRUE(e$.is_testing)) {
      return()
    }
    if (length(percent) != 1)
      return()
    if (is.na(percent))
      return()
    if (!is.numeric(percent))
      return()
    if (percent < 0)
      return()
    if (percent > 100)
      return()
    if (is_rstudio && !is.null(rstudiojob)) { # nocov start
      try({
        rstudioapi::jobSetProgress(rstudiojob, percent)
      }, silent = TRUE)
    } else if (is_shiny) {
      shiny::setProgress(value = percent / 100)
    } else if (is_cli) { # nocov end
      try(cli::cli_progress_update(set = percent, .envir = e), silent = TRUE)
    } else { # nocov start
      util_message("|%s>", paste(rep("#", percent), collapse = ""))
    } # nocov end
  }
  if (is.function(progress_fkt)) if
      (length(intersect(names(formals(progress_fkt)),
                names(formals(progress)))) ==
      length(union(names(formals(progress_fkt)),
                       names(formals(progress))))) {
    progress <- progress_fkt
  } else if (is.function(progress_fkt)) {
    util_error(
      c("option %s must refer to a compatible function",
        "in doubt, unset this option()"),
       dQuote("dataquieR.progress_fkt")
    )
  }
  formals(progress)$is_rstudio <- force(is_rstudio)
  formals(progress)$is_shiny <- force(is_shiny)
  formals(progress)$is_cli <- force(is_cli)
  formals(progress)$e <- p
  environment(progress) <- p
  assign("progress", progress, envir = p)

  progress_msg <- function(status, msg, is_rstudio, is_shiny, e) {
    util_call_progress_hooks("msg", status = status, msg = msg)
    if (isTRUE(e$.is_testing)) {
      return()
    }
    if (missing(msg) && !missing(status)) {
      msg <- status
      status <- ""
    }
    if (is_rstudio && !is.null(rstudiojob)) { # nocov start
      try({
        rstudioapi::jobSetStatus(rstudiojob, status)
        rstudioapi::jobAddOutput(rstudiojob, paste0(msg, "\n"))
      }, silent = TRUE)
    } else if (is_shiny) {
      shiny::setProgress(message = msg)
    } else if (is_cli) { # nocov end
      try(cli::cli_progress_update(status = msg, .envir = e), silent = TRUE)
    } else { # nocov start
      util_message("|%s ###", msg)
    } # nocov end
  }
  if (is.function(progress_msg_fkt)) if
      (length(intersect(names(formals(progress_msg_fkt)),
                       names(formals(progress_msg)))) ==
      length(union(names(formals(progress_msg_fkt)),
                   names(formals(progress_msg))))) {
    progress_msg <- progress_msg_fkt
  } else if (is.function(progress_msg_fkt)) {
    util_error(
      c("option %s must refer to a compatible function",
        "in doubt, unset this option()"),
      dQuote("dataquieR.progress_msg_fkt")
    )
  }
  formals(progress_msg)$is_rstudio <- force(is_rstudio)
  formals(progress_msg)$is_shiny <- force(is_shiny)
  formals(progress_msg)$is_cli <- force(is_cli)
  formals(progress_msg)$e <- p
  if (rlang::is_missing(formals(progress_msg)$msg))
    formals(progress_msg)$msg <- ""

  if (is.function(progress_init_fkt)) {
    if (!identical(names(formals(progress_init_fkt)), "n")) {
      util_error(
        c("option %s must refer to a compatible function",
          "in doubt, unset this option()"),
        dQuote("dataquieR.progress_init_fkt")
      )
    }
    progress_init_fkt(n = n)
    util_call_progress_hooks("init", n = n)
  }

  environment(progress_msg) <- p
  assign("progress_msg", progress_msg, envir = p)

  progress_msg(job_name)

  invisible(list(progress = progress,
                 progress_msg = progress_msg))
}

.progress_hooks <- new.env(parent = emptyenv())

#' Register a hook function for progresses in computation/rendering
#'
#' The order hooks are called is not defined.
#'
#' @param type [character] what event
#' @param hook [function] hook function
#'
#' @returns [character] a handle for de-registering, `invisible`
#' @export
prep_register_progress_hook <- function(type = c("progress", "init", "msg"),
                                        hook) {
  type <- util_match_arg(type)
  util_expect_scalar(hook,
                     check_type = is.function,
                     allow_na = TRUE,
                     error_message = sprintf("%s must be a function",
                                             sQuote("hook")))
  if (type == "init" && !"n" %in% names(formals(hook))) {
    util_error("A progress hook of type %s must feature a formal argument %s",
               sQuote(hook), sQuote("n"))
  }
  if (type == "progress" && !"percent" %in% names(formals(hook))) {
    util_error("A progress hook of type %s must feature a formal argument %s",
               sQuote(hook), sQuote("percent"))
  }
  if (type == "msg" && !all(c("status", "msg") %in% names(formals(hook)))) {
    util_error(c("A progress hook of type %s must feature formal arguments %s",
                 "and %s"),
               sQuote(hook), sQuote("percent"), sQuote("msg"))
  }
  handle <- rlang::hash(list(type, hook))
  if (!exists(type, .progress_hooks, mode = "list")) {
    .progress_hooks[[type]] <- list()
  }
  if (!is.null(.progress_hooks[[type]][[handle]])) {
    util_message("This hook has already been registered, won't register twice")
  } else {
    .progress_hooks[[type]][[handle]] <- hook
  }
  return(invisible(handle))
}

#' De-register a hook function for progresses in computation/rendering
#'
#' @param handle [character] the handle
#' @param verbose [logical] message, if `handle` has currently no registration
#'
#' @returns [logical] `invisible(TRUE)` on success
#' @export
prep_deregister_progress_hook <- function(handle, verbose = TRUE) {
  util_expect_scalar(verbose,
                     check_type = is.logical,
                     error_message = sprintf("%s must be a logical",
                                             sQuote("verbose")))
  util_expect_scalar(handle,
                     check_type = is.character,
                     error_message = sprintf("%s must be a character",
                                             sQuote("handle")))
  found <- FALSE
  for (type in eval(formals(prep_register_progress_hook)[["type"]],
                    envir = baseenv())) {
    if (!exists(type, .progress_hooks, mode = "list")) {
      .progress_hooks[[type]] <- list()
    }
    if (!is.null(.progress_hooks[[type]][[handle]])) {
      .progress_hooks[[type]][[handle]] <- NULL
      found <- TRUE
    }
  }
  if (!found && verbose) {
    util_message(c("This hook has not been registered or already",
                   "de-registered, cannot de-register"))
  }
  return(invisible(found))
}

util_call_progress_hooks <- function(type = c("progress", "init", "msg"), ...) {
  type <- util_match_arg(type)
  if (!exists(type, .progress_hooks, mode = "list")) {
    .progress_hooks[[type]] <- list()
  }
  for (handle in names(.progress_hooks[[type]])) {
    if (is.function(.progress_hooks[[type]][[handle]])) {
      worked <-
        try(
          util_call_with_only_existing_formals(
            .progress_hooks[[type]][[handle]], ...),
          silent = TRUE)
      if (util_is_try_error(worked)) {
        util_message("Could not call a hook function: %s\n%s",
                     head(.progress_hooks[[type]][[handle]]),
                     conditionMessage(attr(worked, "condition")))
      }
    } else {
      util_message(
        c("Could not call a hook function, which should never",
          "have been registered: %s. Internal error, sorry. Please report."),
        dQuote(util_deparse1(.progress_hooks[[type]][[handle]]))
      )
    }
  }
}

#' Detect cores
#'
#' See `parallel::detectCores` for further details.
#'
#' @return number of available CPU cores.
#'
#' @family system_functions
#' @concept process
#' @noRd
util_detect_cores <- function() {
  if (requireNamespace("parallelly", quietly = TRUE)) {
    return(parallelly::availableCores())
  # } else if (requireNamespace("rJava", quietly = TRUE)) {
  #   rJava::.jinit()
  #   rt <- rJava::.jcall("java/lang/Runtime", "Ljava/lang/Runtime;",
  #                       method = "getRuntime")
  #   cpus <- rJava::.jcall(rt, "I", "availableProcessors")
  #  return(cpus)
  } else if (requireNamespace("parallel", quietly = TRUE)) {
    r <- parallel::detectCores()
    if (length(r) != 1 || is.na(r) || !util_is_integer(r) || r < 1) {
      r <- 1
    }
    return(r)
  } else{
    util_warning(c("None of the suggested packages %s are found,",
                   "autodetection of CPU cores disabled --",
                   "using default of 1 core only."),
                 util_pretty_vector_string(c(
                   "parallel",
                   "parallelly")))#,
                   #"rJava")))
    return(1)
  }
}

test_that("util_detect_cores works", {
  skip_on_cran() # not really useful, yet
  skip_if_not_installed("parallelly")
  expect_equal(util_detect_cores(), parallelly::availableCores())
  expect_warning(expect_equal(with_mocked_bindings(
    .package = "base",
    requireNamespace = function(...) {
    return(FALSE)
  },
  util_detect_cores()
  ), 1),
  regexp = "None of the suggested packages.*are found")
})

#' check, if a scalar/vector function argument matches expectations
#'
#' @param arg_name the argument
#' @param allow_more_than_one allow vectors
#' @param allow_null allow NULL
#' @param allow_na allow `NAs`
#' @param min_length minimum length of the argument's value
#' @param max_length maximum length of the argument's value
#' @param check_type a predicate function, that must return `TRUE` on the
#'                   argument's value.
#' @param convert_if_possible if given, a lambda can be given
#'                            similar to `check_type` This
#'                            lambda would be used to try a conversion. If
#'                            a conversion fails (returns `NA`, where the
#'                            input was not `util_empty'), an error
#'                            is still thrown, the data is converted, otherwise
#' @param dont_assign set `TRUE` to keep `x` in the caller environment untouched
#' @param conversion_may_replace_NA if set to `TRUE`, we can define a function
#'                    in `convert_if_possible` that replaces `NA` values without
#'                    causing a warning, but this option is set to `FALSE` by
#'                    default to catch possible conversion problems (use it with
#'                    caution).
#' @param error_message if `check_type()` returned `FALSE`, show this instead of
#'                      a default error message.
#'
#' @return the value of arg_name -- but this is updated in the calling
#'         frame anyway.
#'
#' @examples
#' \dontrun{
#' f <- function(x) {
#'   util_expect_scalar(x, check_type = is.integer)
#' }
#' f(42L)
#' try(f(42))
#' g <- function(x) {
#'   util_expect_scalar(x, check_type = is.integer, convert_if_possible =
#'           as.integer)
#' }
#' g(42L)
#' g(42)
#' }
#'
#' @family robustness_functions
#' @concept robustness
#' @noRd
util_expect_scalar <- function(arg_name,
                               allow_more_than_one = FALSE,
                               allow_null = FALSE,
                               allow_na = FALSE,
                               min_length = -Inf,
                               max_length = Inf,
                               check_type,
                               convert_if_possible,
                               conversion_may_replace_NA = FALSE,
                               dont_assign = FALSE,
                               error_message) {
  if (!missing(error_message)) {
    util_expect_scalar(arg_name = error_message,
                       # error_message = Do not use error_message here to
                       #                 prevent infinite recursion!!
                       check_type = is.character)
  }

  if (missing(convert_if_possible)) {
    convert_if_possible <- NULL
  } else if (!is.function(convert_if_possible)) {
    util_error("%s needs to be a function",
               sQuote("convert_if_possible"))
  }
  if (missing(dont_assign)) dont_assign <- FALSE
  if (length(dont_assign) != 1 || is.na(dont_assign) ||
      !is.logical(dont_assign)) {
    util_error("%s needs to be TRUE or FALSE", sQuote("dont_assign"))
  }
  # interpret the arg_name argument
  try({
    arg_name <- as.character(substitute(arg_name))
  }, # if called with a symbol, get that symbol as a character,
  silent = TRUE # if called with a character, do nothing
  )
  if (!all(is.character(arg_name))) { # nocov start
    # if we did not get a character up to here, there is an error in
    # calling this util_expect_scalar
    # this should never happen anyway, so cannot test this
    util_error(c(
      "argument arg_name must be either one character or one symbol,",
      "wrong use of util_expect_scalar?"))
  } # nocov end
  if (length(arg_name) != 1) {
    # if we have 0 or more than one argument name, this is also an
    # error in calling util_expect_scalar
    util_error(c(
      "argument arg_name must be of length 1,",
      "wrong use of util_expect_scalar?"))
  }

  p <- parent.frame(1) # access the caller's environment

  if (!exists(arg_name, envir = p)) {
    # check, if the function argument exists in the calling function.
    util_error(
      c("Unknown function argument %s checked,",
        "wrong use of util_expect_scalar?"),
      arg_name)
  }

  missing_in_parent <- eval.parent(call("missing", as.symbol(arg_name)))

  arg_value <- try(get(arg_name, envir = p), silent = TRUE)
  # try to get the value of the callers argument called `arg_name`.
  if (inherits(arg_value, "try-error")) {
    if (missing_in_parent) {
      util_warning(
        c("Missing argument %s without default value. Setting to NULL. As",
          "a dataquieR developer, please add a default value for %s to",
          "remove this warning."),
        dQuote(arg_name), dQuote(arg_name),
        applicability_problem = TRUE)
    } else {
      util_warning(
        c("Could not get value of argument %s for unexpected reasons. Setting",
          "to NULL."), arg_name, applicability_problem = TRUE)
      util_warning(arg_value, applicability_problem = TRUE)
    }
    arg_value <- NULL
  }

  if (!allow_null && is.null(arg_value)) {
    # if we need the argument in arg_name, but the user provided NULL
    util_error("Argument %s is NULL", arg_name,
               applicability_problem = TRUE) # this is an error
  }

  if (allow_more_than_one) {
    # if we allow more than one arg_value in the argument arg_name
    if (!allow_null && length(arg_value) == 0) {
      # and we do need at least one arg_value name here, but the user did not
      # provide any
      util_error("Need at least one element in argument %s, got 0", arg_name,
                 applicability_problem = TRUE)
      # this is an error
    }

    if (length(dim(arg_value))) {
      util_error(
        "Need a vector in argument %s, got an object with %d dimensions",
        arg_name,
        length(dim(arg_value)),
        applicability_problem = TRUE)
    }

  } else {# if we expect one value in the argument arg_name at most
    if ((!allow_null && length(arg_value) != 1) ||
        (allow_null && length(arg_value) > 1)) {
      # but the user gave more than one or none although
      # allow_null prohibits this
      util_error("Need exactly one element in argument %s, got %d: [%s]",
                 arg_name, length(arg_value), paste0(arg_value, collapse = ", "),
                 applicability_problem = TRUE)
      # this is an error
    }
  }


  if (length(min_length) != 1) {
    if (!all(is.numeric(min_length))) {
      util_error(c("Need numeric min_length names in argument %s",
                   "wrong use of util_expect_scalar?"),
                 arg_name,
                 applicability_problem = FALSE)
    }
  }

  if (length(max_length) != 1) {
    if (!all(is.numeric(max_length))) {
      util_error(c("Need numeric max_length names in argument %s",
                   "wrong use of util_expect_scalar?"),
                 arg_name,
                 applicability_problem = FALSE)
    }
  }

  if (length(arg_value) < min_length || length(arg_value) > max_length) {
    if (!(allow_null && is.null(arg_value))) {
      util_error("Argument %s must have a length in [%s:%s]",
                 arg_name,
                 as.character(min_length),
                 as.character(max_length),
                 applicability_problem = TRUE)
    }
  }


  my_check_type <- function(arg_value) {
    type_match <- try(check_type(arg_value))
    if (length(type_match) != 1 ||
        !is.logical((type_match)) ||
        is.na(type_match)) {
      util_error(c("Need a lambda function as a predicate as %s for the",
                   "argument %s, but the function returned %s",
                   "wrong use of util_expect_scalar?"),
                 dQuote("check_type"),
                 arg_name,
                 sQuote(paste(deparse(type_match, nlines = 2),
                              collapse = "\n")),
                 applicability_problem = FALSE)
    }
    type_match
  }

  if (length(arg_value) > 0 && (!missing(check_type))) {
    if (!is.function(check_type)) {
      util_error(c("Need a lambda function as a predicate as %s for the",
                   "argument %s",
                   "wrong use of util_expect_scalar?"),
                 dQuote("check_type"),
                 arg_name,
                 applicability_problem = FALSE)
    }
    type_match <- my_check_type(arg_value)
    if (!type_match && !is.null(convert_if_possible)) {
      x_arg <- convert_if_possible(arg_value)
      if (!conversion_may_replace_NA) {
        if (!all(is.na(x_arg) == is.na(arg_value))) {
          util_warning(
            "In %s, could not convert the whole vector to match data type",
            dQuote(arg_name))
        } else {
          arg_value <- x_arg
        }
        type_match <- my_check_type(arg_value)
      } else {
        if (!all(which(is.na(x_arg)) %in% which(is.na(arg_value)))) {
          util_warning(
            "In %s, conversion introduced NAs",
            dQuote(arg_name))
        } else {
          arg_value <- x_arg
        }
        type_match <- my_check_type(arg_value)
      }
    }
    if (!type_match) {
      if (!missing(error_message)) {
        util_error(error_message,
                   applicability_problem = TRUE)
      } else {
        if (!is.null(error_msg <- attr(check_type, "error_msg"))) {
          msg <- error_msg
        } else if (identical(check_type, is.character)) {
          msg <- "be characters"
        } else if (identical(check_type, is.integer)) {
          msg <- "be integer numbers"
        } else if (identical(check_type, is.double)) {
          msg <- "be floating point"
        } else if (identical(check_type, is.numeric)) {
          msg <- "be numeric"
        } else if (identical(check_type, is.logical)) {
          msg <- "be logical"
        } else {
          msg <- sprintf("match the predicate %s",
                  dQuote(paste(head(deparse(check_type)), collapse = " ")))
        }
        util_error("Argument %s must %s",
                   arg_name,
                   msg,
                   applicability_problem = TRUE)
      }
    }
  }

  if (!allow_na && any(is.na(arg_value))) {
    util_error("Argument %s must not contain NAs",
               arg_name,
               applicability_problem = TRUE)
  }

  if (!dont_assign)
    assign(arg_name, arg_value, envir = p) # re-assign the possibly modified
  # argument value in the caller's environment

  arg_value
}

#' Import a data frame
#'
#' see [`rio::import`], but with argument `keep_types` and modified error
#' handling.
#'
#' @param fn the file name to load.
#' @param keep_types [logical] keep types as possibly defined in the file.
#'                             set `TRUE` for study data.
#' @param ... additional arguments for [rio::import]
#'
#' @return [data.frame] as in [rio::import]
#' @noRd
util_rio_import <- function(fn, keep_types, ...) {
  r <- .util_rio_import_lambda(fn, keep_types, lambda = rio::import, ...)
  .util_warn_on_possible_study_data_read_with_keep_types_false_or_vv(
    r,
    fn = fn,
    keep_types = keep_types)
  r
}

#' Import list of data frames
#'
#' see [`rio::import_list`], but with argument `keep_types` and modified error
#' handling.
#'
#' @param fn the file name to load.
#' @param keep_types [logical] keep types as possibly defined in the file.
#'                             set `TRUE` for study data.
#' @param ... additional arguments for [rio::import_list]
#'
#' @return [list] as in [rio::import_list]
#' @noRd
#'
util_rio_import_list <- function(fn, keep_types, ...) {
  if (tolower(tools::file_ext(trimws(fn))) %in% c("odm", "xml") &&
      file.exists(fn) && !dir.exists(fn) &&
      util_is_odm_xml(fn)) {
    if (!odm_installed) {
      util_error(c("%s looks like an ODM file. %s cannot read these without",
                     "the help of dataquieR2odm, which you need to install",
                     "separately, see %s"),
                   dQuote(fn),
                   sQuote(utils::packageName()),
                   sQuote(
                   'https://dataquality.qihs.uni-greifswald.de/DownloadR.html'))
    }
    my_read_list <- try(util_odm2dataquieR(fn), silent = TRUE)
    if (!util_is_try_error(my_read_list) && is.list(my_read_list)) {
      if (all(c("item_level", CODE_LIST_TABLE) %in% names(my_read_list))) {
        return(my_read_list[c("item_level", CODE_LIST_TABLE)])
      }
    }
  }
  r <- .util_rio_import_lambda(fn, keep_types, lambda = rio::import_list, ...)
  .util_warn_on_possible_study_data_read_with_keep_types_false_or_vv(
    r,
    fn = fn,
    keep_types = keep_types)
  r
}

.util_warn_on_possible_study_data_read_with_keep_types_false_or_vv <-
  function(x, fn, keep_types, sheet = "", segments = character(0)) {
  segments <- unique(c(segments, unlist(eapply(.dataframe_environment(),
                                               function(df0) {
                                                 if (!is.data.frame(df0)) return(character(0))
                                                 df0[, intersect(colnames(df0), c(KEY_STUDY_SEGMENT, STUDY_SEGMENT))]
                                               }))))
  other_allowed_names <- sort(c(
    "indicator_metric",
    "Description",
    paste0("dqi_cat_", 1:20),
    "category",
    "label",
    "color",
    "ID",
    "Metadata sheet",
    "Column name",
    "Definition",
    "Use in dataquieR reporting"
  ))
  if (keep_types) { # expecting study data
    if (!is.data.frame(x)) {
      segments <- unique(c(segments, unlist(lapply(x, function(df0) {
        if (!is.data.frame(df0)) return(character(0))
        df0[, intersect(colnames(df0), c(KEY_STUDY_SEGMENT, STUDY_SEGMENT))]
      }))))
      segments <- segments[!util_empty(segments)]
      mapply(SIMPLIFY = FALSE,
             x = x,
             sheet = names(x),
             FUN =
            .util_warn_on_possible_study_data_read_with_keep_types_false_or_vv,
             MoreArgs = list(
               fn = fn,
               keep_types = TRUE,
               segments = segments
            )
      )
    } else {
      ratio <- sum(vapply(colnames(x), # of known metadata column names
                   function(cn) {
                     exists(cn,
                            envir = getNamespace("dataquieR"),
                            inherits = FALSE) || cn %in% c(other_allowed_names,
                                                           segments)
                   }, FUN.VALUE = logical(1))) / ncol(x)
      if (ratio >= .8) { # >= 80% well known metadata names
        util_message(c("%s: Found many (%g %%) column names that look like",
                       "meta data but %s was set to %s -- did you read",
                       "the file keeping column types on purpose, this",
                       "is not suitable for metadata. Colunm names: %s"),
                     dQuote(fn),
                     100 * ratio, sQuote("keep_types"), dQuote(keep_types),
                     util_pretty_vector_string(colnames(x)))
      }
    }
  } else { # expecting metadata
    if (!is.data.frame(x)) {
      segments <- unique(c(segments, unlist(lapply(x, function(df0) {
        if (!is.data.frame(df0)) return(character(0))
        df0[, intersect(colnames(df0), c(KEY_STUDY_SEGMENT, STUDY_SEGMENT))]
      }))))
      segments <- segments[!util_empty(segments)]
      mapply(SIMPLIFY = FALSE,
             x = x,
             sheet = names(x),
             FUN =
            .util_warn_on_possible_study_data_read_with_keep_types_false_or_vv,
             MoreArgs = list(
               fn = fn,
               keep_types = FALSE,
               segments = segments
             )
      )
    } else {
      ratio <- sum(vapply(colnames(x), # of known metadata column names
                   function(cn) {
                     exists(cn,
                            envir = getNamespace("dataquieR"),
                            inherits = FALSE) || cn %in% c(other_allowed_names,
                                                           segments)
                   }, FUN.VALUE = logical(1))) / ncol(x)
      if (ratio < .05) { # < 5% well known metadata names
        util_message(c("%s: Found only few (%g %%) column names that look like",
                       "meta data but %s was set to %s -- did you read",
                       "the file as text-only on purpose, this",
                       "is not suitable for study data. Colunm names: %s"),
                     dQuote(fn),
                     100 * ratio, sQuote("keep_types"), dQuote(keep_types),
                     util_pretty_vector_string(colnames(x)))
      }
    }
  }
}

.util_rio_import_lambda <- function(fn, keep_types, lambda, ...) { # TODO: Detect invalid utf-8 codes and remove such. use readr::guess_encoding()??
  withr::local_options(list("datatable.na.strings" = c("NA",
                                                       "-",
                                                       "",
                                                       "na",
                                                       "n.a.",
                                                       "n/a")))
  is_list <- (identical(lambda, rio::import_list))
  if (!identical(lambda, rio::import) && !identical(lambda, rio::import_list)) {
    util_error(
      "Internal error, sorry, please report: invalid lambda in %s",
      sQuote("util_rio_import_lambda"))
  }
  util_expect_scalar(keep_types,
                     check_type = is.logical,
                     error_message =
                       sprintf("%s needs to be to be a logical value.",
                               sQuote("keep_types")))
  # prep_guess_encoding(fn)
  file_type_needs_type_adjust <-
    getOption("dataquieR.fix_column_type_on_read",
              dataquieR.fix_column_type_on_read_default) &&
      (all(grepl(".", fn, fixed = TRUE)) &&
         rio::get_info(fn)$input %in% .formats_with_weird_types)
  need_type_adjust <- FALSE
  # prep_robust_guess_data_type(c("2021-02-29"))
  # lapply(util_rio_import("/Library/Frameworks/R.framework/Versions/4.1/Resources/library/rgdal/gdal/gcs.csv", keep_types = FALSE), prep_robust_guess_data_type)
#
#   study_data <- util_adjust_data_type(
#     study_data = study_data,
#     meta_data = meta_data,
#     relevant_vars_for_warnings = relevant_vars_for_warnings)
  # rio::get_info("/Library/Frameworks/R.framework/Versions/4.1/Resources/library/rgdal/gdal/gcs.csv")$input
  # study_data <- util_adjust_data_type(
  #   study_data = study_data,
  #   meta_data = meta_data,
  #   relevant_vars_for_warnings = relevant_vars_for_warnings)
  if (endsWith(tolower(fn), ".ods")) {
    col_types <- NA
  } else if (!endsWith(tolower(fn), ".xls") &&
             !endsWith(tolower(fn), ".xlsx")) {
    col_types <- "character"
  } else {
    col_types <- "text"
  }
  need_read_as_text <-
    (!keep_types || file_type_needs_type_adjust)
  warns1 <- new.env(parent = emptyenv())
  warns1$warns <- list()
  warns2 <- new.env(parent = emptyenv())
  warns2$warns <- list()
  r <- suppressWarnings(suppressMessages(
    util_suppress_output(try(lambda(fn, ...), silent = TRUE), warns = warns1)))
  if (need_read_as_text && !util_is_try_error(r)) {
    .rio_res_is_not_null <- function(rx, .is_list) {
      if (.is_list) {
        any(vapply(rx, .rio_res_is_not_null, .is_list = FALSE,
                   FUN.VALUE = logical(1)))
      } else {
        !!prod(dim(rx))
      }
    }
    r2 <- util_suppress_output(suppressMessages(suppressWarnings(
      try(lambda(fn, ..., col_types = col_types,
                    colClasses = "character"), silent = TRUE))), warns = warns2)
    if (!util_is_try_error(r2) && .rio_res_is_not_null(r2,
                                                       .is_list = is_list)) {
      r <- r2
      need_type_adjust <- keep_types
    } else if (.rio_res_is_not_null(r, .is_list = is_list)) {
      util_warning("Could not read %s as text-only data frame",
                   dQuote(fn),
                   applicability_problem = TRUE)
    }
  }
  if (all(grepl(".", fn, fixed = TRUE)) &&
    rio::get_info(fn)$input %in% c("csv", "tsv")) { # data.table::fread not returning NA for "", they are only used, if not in quotes in the csv/tsv
    r[] <- lapply(r, function(x) {
      x[x %in% getOption("datatable.na.strings")] <- NA
      x
    })
  }
  if (need_type_adjust) {
    e <- environment()
    trouble_cols <- character(0)
    .adjust_type <- function(dfr, sheet) {
      dfr[] <-
        lapply(setNames(nm = colnames(dfr)), function(nm, sheet) {
          x <- dfr[[nm]]
          dt0 <- try(prep_robust_guess_data_type(x, it = 200),
                     silent = TRUE)
          if (!util_is_try_error(dt0) && !is.na(dt0)) {
            x <- util_data_type_conversion(x, dt0)
          } else {
            if (!missing(sheet)) {
              nm <- paste(sheet, SPLIT_CHAR, nm)
            }
            e$trouble_cols <- c(e$trouble_cols, nm)
          }
          return(x)
        }, sheet = sheet)
      dfr
    }
    if (is_list) {
      r <- mapply(SIMPLIFY = FALSE, dfr = r, sheet = names(r),
                  FUN = .adjust_type)
    } else {
      r <- .adjust_type(r)
    }
    if (length(e$trouble_cols) > 0) {
      util_warning(
        "Could not adjust the data type for the following columns: %s",
        util_pretty_vector_string(e$trouble_cols, n_max = 5))
    }
  }
  is_excel <- endsWith(tolower(fn), ".xls") || endsWith(tolower(fn), ".xlsx")
  if (isTRUE(keep_types) && is_excel) {
    if (is_list) {
      r <- lapply(r, .util_convert_time_only_cols_df)
    } else {
      r <- .util_convert_time_only_cols_df(r)
    }
  }
  all_warns <- c(warns1$warns,
                 warns2$warns)
  if (length(all_warns) > 0) {
    all_warns <- lapply(all_warns, conditionMessage)
    util_warning("Could not correctly read %s: %s",
                 fn, util_pretty_vector_string(all_warns, n_max = 4))
  }
  r
}

.formats_with_weird_types <- local({
  suppressMessages(all_formats <- gsub("^.import.rio_", "", as.character(utils::.S3methods(".import", envir = asNamespace("rio")))))
  mime_ext <- strsplit(.mime_chache$tb$Extensions, "\\s+", perl = TRUE)
  formats_with_mime_types <- lapply(setNames(nm = all_formats), function(fmt) {
    r <-
      .mime_chache$tb[(vapply(mime_ext, `%in%`, x = fmt,
                              FUN.VALUE = logical(1))),
                      c("Type", "Subtype"), drop = TRUE]
    cbind(list(fmt = rep(fmt, nrow(r))), r)
  })
  formats_with_mime_types <- formats_with_mime_types[0 < vapply(
    formats_with_mime_types,
    nrow,
    FUN.VALUE = integer(1)
  )]
  formats_with_mime_types <- util_rbind(data_frames_list =
                                          formats_with_mime_types[0 < vapply(
                                            formats_with_mime_types,
                                            nrow,
                                            FUN.VALUE = integer(1)
                                          )])
  formats_with_weird_types <-
    formats_with_mime_types[formats_with_mime_types$Type == "text" |
                              formats_with_mime_types$fmt %in%  # the latter feature typed cells, not columns
                              c("ods", "xls", "xlsx"),
                            "fmt", drop = TRUE];
  formats_with_weird_types
})

# Detect whether a POSIXct vector is an Excel "time-only" column
.util_is_excel_time_only <- function(x) {
  if (!inherits(x, "POSIXct")) return(FALSE)
  if (all(is.na(x))) return(FALSE)
  # Excel time-only values are datetimes anchored at an origin date
  # (Windows origin 1899-12-30/31; macOS origin 1904-01-01)
  ux <- unique(as.Date(x, tz = "UTC"))
  origins <- as.Date(c("1899-12-30", "1899-12-31", "1904-01-01"))
  # treat as time-only if ALL non-NA dates are one of the known origins
  all(ux %in% origins)
}

# Convert POSIXct Excel time-only to hms (or difftime fallback)
.util_posixct_to_hms <- function(x) {
  secs <- suppressWarnings(as.numeric(x)) %% (24 * 3600)
  secs[is.na(x)] <- NA_real_
  if (requireNamespace("hms", quietly = TRUE)) {
    return(hms::as_hms(secs))
  } else {
    # fallback to difftime if hms isn't available
    return(structure(secs, class = "difftime", units = "secs"))
  }
}

# Convert all time-only columns in a data.frame
.util_convert_time_only_cols_df <- function(dfr) {
  if (!is.data.frame(dfr) || !ncol(dfr)) return(dfr)
  for (cn in names(dfr)) {
    x <- dfr[[cn]]
    if (.util_is_excel_time_only(x)) {
      dfr[[cn]] <- .util_posixct_to_hms(x)
    }
  }
  dfr
}


# FIXME: If I read the files with robust data types, int_datatype_somehing cannot run, so find a way to read it as string, first.

#' Checks user-defined contradictions in study data
#'
#' @description
#' This approach considers a contradiction if impossible combinations of data
#' are observed in one participant. For example, if age of a participant is
#' recorded repeatedly the value of age is (unfortunately) not able to decline.
#' Most cases of contradictions rest on comparison of two variables.
#'
#' Important to note, each value that is used for comparison may represent a
#' possible characteristic but the combination of these two values is considered
#' to be impossible. The approach does not consider implausible or inadmissible
#' values.
#'
#' [Descriptor]
#'
#' @details
#' ### Algorithm of this implementation:
#'
#'  - Select all variables in the data with defined contradiction rules (static
#'    metadata column CONTRADICTIONS)
#'  - Remove missing codes from the study data (if defined in the metadata)
#'  - Remove measurements deviating from limits defined in the metadata
#'  - Assign label to levels of categorical variables (if applicable)
#'  - Apply contradiction checks on predefined sets of variables
#'  - Identification of measurements fulfilling contradiction rules. Therefore
#'    two output data frames are generated:
#'    - on the level of observation to flag each contradictory value
#'      combination, and
#'    - a summary table for each contradiction check.
#'  - A summary plot illustrating the number of contradictions is generated.
#'
#' List function.
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable list] the name of the measurement variables
#' @param threshold_value [numeric] from=0 to=100. a numerical value
#'                                                 ranging from 0-100
#' @param check_table [data.frame] contradiction rules table.  Table defining
#'                                 contradictions. See details for
#'                                 its required structure.
#' @param summarize_categories [logical] Needs a column 'tag' in the
#'                             `check_table`.
#'                             If set, a summary output is generated for the
#'                             defined categories plus one plot per
#'                             category.
#' @return
#' If `summarize_categories` is `FALSE`:
#' A [list] with:
#'   - `FlaggedStudyData`: The first output of the contradiction function is a
#'                         data frame of similar dimension regarding the number
#'                         of observations in the study data. In addition, for
#'                         each applied check on the variables an additional
#'                         column is added which flags observations with a
#'                         contradiction given the applied check.
#'   - `SummaryTable`: The second output summarizes this information into one
#'                     data frame. This output can be used to provide an
#'                     executive overview on the amount of contradictions. This
#'                     output is meant for automatic digestion within pipelines.
#'   - `SummaryData`: The third output is the same as `SummaryTable` but for
#'                    human readers.
#'   - `SummaryPlot`: The fourth output visualizes summarized information
#'     of `SummaryData`.
#'
#' if `summarize_categories` is `TRUE`, other objects are returned:
#' one per category named by that category (e.g. "Empirical") containing a
#' result for contradictions within that category only. Additionally, in the
#' slot `all_checks` a result as it would have been returned with
#' `summarize_categories` set to `FALSE`. Finally, a slot `SummaryData` is
#' returned containing sums per Category and an according [ggplot2::ggplot] in
#' `SummaryPlot`.
#'
#' @export
#'
#' @importFrom ggplot2 ggplot geom_bar scale_fill_manual theme_minimal scale_y_continuous geom_hline coord_flip theme aes geom_text xlab scale_x_continuous sec_axis
#' @importFrom stats setNames reorder
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions.html
#' )
con_contradictions <- function(resp_vars = NULL,
                               study_data,
                               label_col,
                               item_level = "item_level",
                               threshold_value,
                               check_table,
                               summarize_categories = FALSE,
                               # flip_mode = "flip" # TODO: Fix noflip graph
                               meta_data = item_level,
                               meta_data_v2) {
  # Preps ----------------------------------------------------------------------
  util_maybe_load_meta_data_v2()
  # labels used instead of variable names?
  if (!(missing(label_col)) && label_col != VAR_NAMES) {
    util_message(
      sprintf(paste("Labels of variables from %s will be used.",
                    "In this case columns A and B in check_tables must",
                    "refer to labels.", collapse = " "),
      dQuote(label_col))
    )
  } else {
    util_message(paste("Variable names will be used. In this case columns A",
                  "and B in check_tables must refer to variable names."))
  }

  # map meta to study
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata = TRUE)

  util_correct_variable_use("resp_vars",
    allow_more_than_one = TRUE,
    allow_null = TRUE,
    allow_any_obs_na = TRUE
  )

  util_expect_data_frame(check_table, c("ID", "Function_name", "A", "A_levels",
                                        "A_value", "B", "B_levels",  "B_value",
                                        "Label"))

  # table of specified contradictions
  if (missing(check_table) || !is.data.frame(check_table)) {
    util_error(
      c("Missing check_table --",
        "cannot apply contradictions checks w/o contradiction rules"),
      applicability_problem = TRUE)
  }

  if (missing(threshold_value)) {
    threshold_value <- 0
    if (!.called_in_pipeline)
      util_message("No %s has been set, will use default %d",
                 dQuote("threshold_value"), threshold_value,
                 applicability_problem = TRUE)
  }

  ct <- check_table
  ct$Label <- as.character(ct$Label)

  # colors:
  cols <- c("0" = "#2166AC", "1" = "#B2182B")

  # columns:
  expected_cols <- c(
    "ID",
    "Label",
    "Function_name",
    "A",
    "B",
    "A_value",
    "B_value",
    "A_levels",
    "B_levels"
  )

  missing_cols <- !(expected_cols %in% colnames(ct))

  if (any(missing_cols)) {
    util_error(
      "Missing the following columns in the check_table: %s",
      dQuote(expected_cols[missing_cols]),
      applicability_problem = TRUE
    )
  }

  if (summarize_categories) { # TODO: Optimize graphics as in the redcap version of this function.
    # if we want to summarize contradictions per category
    if (!("tag" %in% colnames(ct))) {
      util_error(c(
        "Cannot summerize categories of contradictions,",
        "because these are not defined in the check_table as column 'tag'."),
        applicability_problem = TRUE)
    }
    splitted_tags <- lapply(strsplit(ct$tag, SPLIT_CHAR, fixed = TRUE), trimws)
    tags <- sort(unique(unlist(splitted_tags)))
    tags <- setNames(nm = tags)
    tags_ext <- tags
    tags_ext[["all_checks"]] <- NA
    result <- lapply(tags_ext, function(atag) {
      # generate one output per category (stratified)
      if (is.na(atag)) {
        new_ct <- ct[, -which(colnames(ct) == "tag"), drop = FALSE]
      } else {
        contains_tag <- function(x, tg) {
          any(x == tg, na.rm = TRUE)
        }
        rows_matching_tag <- vapply(splitted_tags, contains_tag, tg = atag,
                                    logical(1))
        new_ct <- ct[rows_matching_tag, -which(colnames(ct) == "tag"),
                     drop = FALSE]
      }
      con_contradictions(
        resp_vars = resp_vars, study_data = study_data,
        meta_data = meta_data, label_col = label_col,
        threshold_value = threshold_value, check_table = new_ct,
        summarize_categories = FALSE
        # , flip_mode = flip_mode TODO
      )
    })
    rx <- lapply(tags_ext, function(atag) {
      # and summarize the contradictions per category/tag
      if (is.na(atag)) {
        sum(rowSums(result[["all_checks"]]$FlaggedStudyData[, -1, drop = FALSE],
                    na.rm = TRUE) > 0) /
          nrow(result[["all_checks"]]$FlaggedStudyData) * 100
      } else {
        sum(rowSums(result[[atag]]$FlaggedStudyData[, -1, drop = FALSE],
                    na.rm = TRUE) > 0) /
          nrow(result[[atag]]$FlaggedStudyData) * 100
      }
    })
    rx <- data.frame(
      category = names(rx),
      percent = unlist(rx),
      GRADING = ordered(ifelse(unlist(rx) > threshold_value, 1, 0))
    )
    result$SummaryData <- rx
    # Plot for summarized contradiction checks -----------------------------------------------------
    p <- util_create_lean_ggplot(ggplot(rx, aes(x = seq_len(nrow(rx)), y = percent,
                        fill = as.ordered(GRADING))) +
      geom_bar(stat = "identity") +
      scale_fill_manual(values = cols, name = " ", guide = "none") +
      theme_minimal() +
      scale_y_continuous(name = "(%)",
                         limits = (c(0, max(1.2 * max(rx$percent),
                                            threshold_value))),
                         expand = expansion(mult = c(0,0.05))) +
      scale_x_continuous(breaks = seq_len(nrow(rx)),
                         sec.axis = sec_axis(~., # TDOO: checl ~
                                             breaks = seq_len(nrow(rx)),
                                             labels = rx$category),
                         trans = "reverse") +
      xlab("Category of applied contradiction checks") +
      geom_hline(yintercept = threshold_value, color = "red", linetype = 2) +
      geom_text(label = paste0(" ", round(rx$percent, 2), "%"),
                hjust = 0, vjust = 0.5) +
      coord_flip() + # TODO
      theme(axis.text.y.right = element_text(size = 14),
            axis.text.y.left = element_blank()),
      rx = rx,
      cols = cols,
      threshold_value = threshold_value
      )

    # p <- p + util_coord_flip(p = p) # TODO: estimate w and h, if p is not using discrete axes util_lazy_add_coord(p, fli)

    # https://stackoverflow.com/a/51795017
#    bp <- ggplot_build(p)
#    w <- 2 * length(bp$layout$panel_params[[1]]$x$get_labels())
#    if (w == 0) {
 #     w <- 10
#    }
#    w <- w + 2 +
#      max(nchar(bp$layout$panel_params[[1]]$y$get_labels()),
#          na.rm = TRUE)
#    h <- 2 * length(bp$layout$panel_params[[1]]$y$get_labels())
 #   if (h == 0) {
#      h <- 10
#    }
 #   h <- h + 15

#    p <- util_set_size(p, width_em = w, height_em = h)

    result$SummaryPlot <- p

    return(result)
  } else {
    ct$A_levels <- as.character(ct$A_levels)
    ct$B_levels <- as.character(ct$B_levels)

    # check and prep metadata
    if (!(CONTRADICTIONS %in% colnames(meta_data))) {
      util_error(
        c("Missing column %s in metadata cannot apply",
          "contradictions checks w/o contradiction rules"),
        dQuote(CONTRADICTIONS),
        applicability_problem = TRUE
      )
    }

    meta_data[["CONTRADICTIONS"]] <-
      as.character(meta_data[["CONTRADICTIONS"]])

    # no variables defined?
    if (length(resp_vars) == 0) {
      if (all(is.na(meta_data[[CONTRADICTIONS]]))) {
        util_error(paste0("No Variables with defined CONTRADICTIONS."),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      } else {
        util_message(paste0(
          "All variables with CONTRADICTIONS in the metadata are used."),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE)
        resp_vars <- meta_data[[label_col]][!(is.na(meta_data[[CONTRADICTIONS]]))]
        resp_vars <- intersect(resp_vars, colnames(ds1))
      }
    } else {
      # contradictions defined at all?
      if (all(is.na(meta_data[[CONTRADICTIONS]][meta_data[[label_col]] %in%
                                                resp_vars]))) {
        util_error(paste0("No Variables with defined CONTRADICTIONS."),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      }
      # no contradictions for some variables?
      rvs_with_contr <- meta_data[[label_col]][!(is.na(meta_data[[CONTRADICTIONS]])) &
                                       meta_data[[label_col]] %in% resp_vars]
      if (length(rvs_with_contr) < length(resp_vars)) {
        util_message(paste0("The variables ", resp_vars[!(resp_vars %in% rvs_with_contr)],
                            " have no defined CONTRADICTIONS.",
                            collapse = ", "),
                     applicability_problem = TRUE)
      }
      resp_vars <- rvs_with_contr
    }

    # select contradiction checks
    # get checks from metadata
    cl <- meta_data[[CONTRADICTIONS]][!(is.na(meta_data$CONTRADICTIONS))]

    # is list ?
    cl <- unlist(cl)

    # select unique checks
    cl <- unique(as.numeric(unlist(strsplit(as.character(cl), SPLIT_CHAR,
                                            fixed = TRUE))))

    cl <- intersect(cl, ct$ID)

    cl <- cl[order(cl)]

    summary_df1 <- data.frame(Obs = 1:nrow(ds1))

    summary_df2 <- data.frame(
      Check_type = rep(NA, length(cl)),
      Check_ID = rep(NA, length(cl)),
      Study_variables = rep(NA, length(cl)),
      A_levels = rep(NA, length(cl)),
      B_levels = rep(NA, length(cl)),
      N = rep(NA, length(cl)),
      Percent = rep(NA, length(cl)),
      Grading = rep(NA, length(cl)),
      Label = rep(NA, length(cl))
    )

    for (i in seq_along(cl)) {
      prior_names <- names(summary_df1)

      # prepare columns and name of respective check
      summary_df1[i + 1] <- NA
      names(summary_df1) <- c(prior_names, paste0("grading_", cl[i]))

      # which check function is to be applied
      check <- paste(ct$Function_name[ct$ID == cl[i]])

      # parse levels
      a_lev <- gsub("'", "", ct$A_levels[ct$ID == cl[i]])
      a_lev <- unlist(strsplit(a_lev, SPLIT_CHAR, fixed = TRUE))
      a_lev <- trimws(a_lev)

      b_lev <- gsub("'", "", ct$B_levels[ct$ID == cl[i]])
      b_lev <- unlist(strsplit(b_lev, SPLIT_CHAR, fixed = TRUE))
      b_lev <- trimws(b_lev)
      # apply check
      summary_df1[i + 1] <-
        contradiction_functions[[check]](study_data = ds1,
        A = paste(ct$A[ct$ID == cl[i]]),
        A_levels = a_lev,
        A_value = ct$A_value[ct$ID == cl[i]],
        B = paste(ct$B[ct$ID == cl[i]]),
        B_levels = b_lev,
        B_value = ct$B_value[ct$ID == cl[i]]
      )

      # summarize checks
      summary_df2[i, 1] <- cl[i]
      summary_df2[i, 2] <- check
      summary_df2[i, 3] <- paste0(
        "A is: ", ct$A[ct$ID == cl[i]], "; ",
        "B is: ", ct$B[ct$ID == cl[i]]
      )
      summary_df2[i, 4] <- paste(a_lev, collapse = SPLIT_CHAR)
      summary_df2[i, 5] <- paste(b_lev, collapse = SPLIT_CHAR)
      summary_df2[i, 6] <- sum(summary_df1[, i + 1], na.rm = TRUE)
      summary_df2[i, 7] <- sum(summary_df1[, i + 1], na.rm = TRUE) /
        dim(ds1)[1] * 100
      summary_df2[i, 8] <- ifelse(summary_df2[i, 7] > threshold_value, 1, 0)
      summary_df2[i, 9] <- ct$Label[ct$ID == cl[i]]
    }

    summary_df2$Percent <- round(summary_df2$Percent, digits = 2)

    names(summary_df2) <- c(
      "Check ID", "Check type", "Variables A and B", "A Levels",
      "B Levels", "Contradictions (N)", "Contradictions (%)",
      "GRADING", "Label"
    )

    summary_df2$GRADING <- ordered(summary_df2$GRADING) # FIXME: Update GRADING -> Standard Name

    x <- util_as_numeric(reorder(summary_df2[, 1], -summary_df2[, 1]))
    lbs <- as.character(reorder(summary_df2[, 9], -summary_df2[, 1]))
    # plot summary_df2
    p <- util_create_lean_ggplot(ggplot(summary_df2, aes(x = x, y = .data[["Contradictions (%)"]],
                                 fill =
                                    as.ordered(GRADING))) +
      geom_bar(stat = "identity") +
      geom_text(
        label = paste0(" ", round(summary_df2[, 7], digits = 2), "%"),
        hjust = 0, vjust = 0.5
      ) +
      scale_fill_manual(values = cols, name = " ", guide = "none") +
      theme_minimal() +
      xlab("IDs of applied checks") +
      scale_y_continuous(name = "(%)",
                         limits = (c(0, max(1.2 * max(summary_df2[, 7]),
                                            threshold_value)))) +
      scale_x_continuous(breaks = x, sec.axis =
                           sec_axis(~., breaks = x, labels = lbs)) + # TODO: checl ~ ??
      geom_hline(yintercept = threshold_value, color = "red", linetype = 2) +
      coord_flip() + # TODO
      theme(text = element_text(size = 20)),
      summary_df2 = summary_df2,
      threshold_value = threshold_value,
      x = x,
      lbs = lbs,
      cols = cols)

    # p <- p + util_coord_flip(p = p) # TODO: estimate w and h, if p is not using discrete axes util_lazy_add_coord(p, fli)

    # create SummaryTable object
    st1 <- summary_df2
    st1$`Variables A and B` <- gsub("A is: ", "", st1$`Variables A and B`)
    st1$`Variables A and B` <- gsub("B is: ", "", st1$`Variables A and B`)
    st1$Variables <- unlist(lapply(st1$`Variables A and B`,
                                   function(x) unlist(strsplit(x, ";",
                                                               fixed =
                                                                 TRUE))[1]))
    st1$`Reference variable` <- unlist(lapply(st1$`Variables A and B`,
                                              function(x) unlist(
                                                strsplit(x, ";", fixed =
                                                           TRUE))[2]))
    st1$`Variables A and B` <- NULL
    st1 <- st1[, c(9, 10, 1:8)]
    #st1 <- dplyr::rename(st1, c("GRADING" = "Grading"))

#    suppressWarnings({
      # suppress wrong warnings: https://github.com/tidyverse/ggplot2/pull/4439/commits
      # find out size of the plot https://stackoverflow.com/a/51795017
#      bp <- ggplot_build(p)
#      w <- 2 * length(bp$layout$panel_params[[1]]$x$get_labels())
#      if (w == 0) {
#        w <- 10
 #     }
 #     w <- w + 2 +
 #       max(nchar(bp$layout$panel_params[[1]]$y$get_labels()),
 #           na.rm = TRUE)
 #     w <- w +
 #       max(nchar(bp$layout$panel_params[[1]]$y.sec$get_labels()),
 #           na.rm = TRUE)
 #     h <- 2 * length(bp$layout$panel_params[[1]]$y$get_labels())
#      if (h == 0) {
#        h <- 10
#      }
 #     h <- h + 15
#      p <- util_set_size(p, width_em = w, height_em = h)
#    })

    # Output
    return(list(
      FlaggedStudyData = summary_df1,
      SummaryTable = st1, # TODO: VariableGroupTable
      SummaryData = summary_df2,
      SummaryPlot = p
    ))
  }

  # Never called, just for documentation.
  return(list( # nocov start
    FlaggedStudyData = summary_df1,
    SummaryTable = st1,
    SummaryData = summary_df2,
    SummaryPlot = p
  )) # nocov end
}

#' Counts all individuals with no measurements at all
#'
#' @description
#' This implementation examines a crude version of unit missingness or
#' unit-nonresponse (Kalton and Kasprzyk 1986), i.e. if all measurement
#' variables in the study data are missing for an observation it has unit
#' missingness.
#'
#' The function can be applied on stratified data. In this case strata_vars must
#' be specified.
#'
#' [Descriptor]
#'
#' @details
#'
#' This implementations calculates a crude rate of unit-missingness. This type
#' of missingness may have several causes and is an important research outcome.
#' For example, unit-nonresponse may be selective regarding the targeted study
#' population or technical reasons such as record-linkage may cause
#' unit-missingness.
#'
#' It has to be discriminated form segment and item missingness, since different
#' causes and mechanisms may be the reason for unit-missingness.
#'
#' ### Hint
#' This function does not support a `resp_vars` argument but `id_vars`, which
#' have a roughly inverse logic behind: id_vars with values do not prevent a row
#' from being considered missing, because an ID is the only hint for a unit that
#' elsewise would not occur in the data at all.
#'
#' List function.
#'
#' @inheritParams .template_function_indicator
#'
#' @param id_vars [variable list] optional, a (vectorized) call of ID-variables
#'                                that should not be
#'                                considered in the calculation of unit-
#'                                missingness
#' @param strata_vars [variable] optional, a string or integer variable used for
#'                               stratification
#'
#' @return A list with:
#'   - `FlaggedStudyData`: [data.frame] with id-only-rows flagged in a column
#'                         `Unit_missing`
#'   - `SummaryData`: [data.frame] with numbers and percentages of unit
#'                                missingness
#'
#' @export
#'
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_com_impl_unit_missingness.html
#' )
com_unit_missingness <- function(id_vars = NULL,
                                 strata_vars = NULL,
                                 label_col,
                                 study_data,
                                 item_level = "item_level",
                                 meta_data = item_level,
                                 meta_data_v2) { # TODO: Discuss, if this function can be fully removed?

  util_maybe_load_meta_data_v2()
  # map study and metadata
  prep_prepare_dataframes()
  ds1_labelled <- prep_prepare_dataframes(.apply_factor_metadata_inadm = TRUE,
                                          .internal = FALSE)

  part_vars <- meta_data[[PART_VAR]]

  # correct variable usage
  util_correct_variable_use("id_vars",
    allow_more_than_one = TRUE,
    allow_null          = TRUE
  )

  util_correct_variable_use("strata_vars",
    allow_null = TRUE,
    need_type = "!float"
  )

  if (is.null(id_vars)) {
    util_message(
      c("No ID-variables specified, all variables are",
        "considered to be measurements."),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE
      )
  }

  # initialize result dataframe
  sumdf1 <- ds1

  # compute unit missingness for individuals having
  # NA in all columns (except ID-vars)
  if (!(is.null(id_vars)) || !(is.null(strata_vars))) {
    leave_out <- c()
    if (!(is.null(id_vars))) {
      util_correct_variable_use("id_vars",
        allow_na = TRUE, allow_more_than_one = TRUE,
        allow_null = TRUE, allow_all_obs_na = TRUE, allow_any_obs_na = TRUE
      )
      leave_out <- union(leave_out, id_vars)
    }
    if (!(is.null(strata_vars))) {
      util_correct_variable_use("strata_vars",
        allow_na = FALSE, allow_more_than_one = FALSE,
        allow_null = TRUE, allow_all_obs_na = FALSE, allow_any_obs_na = TRUE
      )
      leave_out <- union(leave_out, strata_vars)
    }
    sumdf1$Unit_missing <- as.integer(apply(ds1[, -which(names(ds1) %in%
                                                           leave_out)], 1,
                                            function(x) all(is.na(x))))
  } else {
    sumdf1$Unit_missing <- as.integer(apply(ds1, 1, function(x) all(is.na(x))))
  }

  UMR <- data.frame(
    "N" = sum(sumdf1$Unit_missing, na.rm = TRUE),
    "%" = round(sum(sumdf1$Unit_missing, na.rm = TRUE) / dim(sumdf1)[1] * 100,
                digits = 2)
  )

  # summarize for strata_vars
  if (!(is.null(strata_vars))) {
    sumdf1[, setdiff(strata_vars, part_vars)] <-
      ds1_labelled[, setdiff(strata_vars, part_vars), FALSE]

    sumdf2 <- as.data.frame.matrix(table(sumdf1[[strata_vars]],
                                         sumdf1$Unit_missing))
    if (!any(sumdf1$Unit_missing, na.rm = TRUE)) {
      sumdf2$N_UNIT_MISSINGS <- 0
    }
    colnames(sumdf2) <- c("N_OBS", "N_UNIT_MISSINGS")
    sumdf2[[strata_vars]] <- rownames(sumdf2)
    rownames(sumdf2) <- NULL
    sumdf2 <- sumdf2[, c(strata_vars, c("N_OBS", "N_UNIT_MISSINGS"))]
    sumdf2$"N_UNIT_MISSINGS_(%)" <- round(sumdf2$N_UNIT_MISSINGS /
                                            sumdf2$N_OBS * 100, digits = 2)
  }

  if (!(is.null(strata_vars))) {
    return(list(FlaggedStudyData = sumdf1, SummaryData = sumdf2))
  } else {
    return(list(FlaggedStudyData = sumdf1, SummaryData = UMR))
  }
}

test_that("com_unit_missingness works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  r <- com_unit_missingness(study_data, item_level = meta_data,
                            label_col = LABEL,
                       id_vars = "PSEUDO_ID", strata_vars = "CENTER_0")
  expect_equal(
    length(intersect(
      names(r),
      c("FlaggedStudyData", "SummaryData")
    )), length(union(
      names(r),
      c("FlaggedStudyData", "SummaryData")
    ))
  )
  expect_equal(r$SummaryData,
               structure(
                 list(
                   CENTER_0 = c("Berlin",
                                "Hamburg",
                                "Leipzig",
                                "Cologne",
                                "Munich"),
                   N_OBS = c(617L, 581L, 593L, 564L, 585L),
                   N_UNIT_MISSINGS = c(15L,  11L, 9L, 13L, 12L),
                   "N_UNIT_MISSINGS_(%)" = c(2.43, 1.89, 1.52,  2.3, 2.05)
                 ),
                 row.names = c(NA,-5L),
                 class = "data.frame"
               ))
  expect_equal(sum(r$FlaggedStudyData$Unit_missing == 1), 60)
  expect_equal(unique(r$FlaggedStudyData$Unit_missing), 0:1)
})

#' Can we really be sure to run `RStudio`
#'
#' `Jetbrain's` `Idea` and so on fake to be `RStudio` by having `RStudio` in
#' `.Platform$GUI`.
#'
#' @return `TRUE`, if really sure to be `RStudio`, `FALSE`, otherwise.
#'
#' @noRd
util_really_rstudio <- function() {

  is_shiny <- suppressWarnings(util_ensure_suggested("shiny", err = FALSE)) &&
    (!is.null(shiny::getDefaultReactiveDomain()))

  is_rstudio <-
    !is_shiny &&
    suppressWarnings(util_ensure_suggested("rstudioapi", err = FALSE)) &&
    rstudioapi::isAvailable()

  is_not_jetbrains <- identical(Sys.getenv("RSTUDIO"), "1") &&
    !exists(".jetbrains", globalenv(), mode = "environment")

  is_rstudio && is_not_jetbrains

}

#' @seealso [print.ReportSummaryTable]
#' @family figure_functions
#' @concept figure
#' @noRd

util_plotly_font_size <- function(n_labels, space = 200,
                                  max_font_size = 16, min_font_size = 12) {
  if (n_labels > 15 || min_font_size > max_font_size || n_labels == 0 ||
      n_labels >= space || space == 0) {
    return(10)
  } else if (n_labels > 0 && n_labels < 16){
    return(12)
  } else {
    res <- min_font_size + (log(space/n_labels) - log(space/15)) *
      (max_font_size - min_font_size) / (log(space) - log(space/15))
    return(round(res))
  }
}

#' Delete rows from summary table for `SSI` or non-`SSI` variables
#'
#' @inheritParams .template_function_indicator
#' @param repsumtab [data.frame] the report summary table
#' @param vars_to_include `"study"`, `"ssi"` or `c("study", "ssi")`. variables to include
#'
#' @returns [data.frame] the filtered `repsumtab` with attribute
#'   `rownames_of_report`, also filtered
#' @keywords internal
util_filter_repsum <- function(repsumtab, vars_to_include, meta_data,
                               rownames_of_report, label_col) {
  util_expect_scalar(vars_to_include,
                     allow_more_than_one = TRUE,
                     min_length = 1,
                     max_length = 2,
                     check_type = is.character)
  util_match_arg(vars_to_include, c("study", "ssi"), several_ok = TRUE)

  ssi_vars <-
    meta_data[[VAR_NAMES]][!util_empty(meta_data[[COMPUTED_VARIABLE_ROLE]])]
  ssi_vars_lb <-
    meta_data[[label_col]][!util_empty(meta_data[[COMPUTED_VARIABLE_ROLE]])]

  if (!"ssi" %in% vars_to_include) {
    repsumtab <-
      repsumtab[!repsumtab$VAR_NAMES %in% ssi_vars, , FALSE]
    rownames_of_report <- rownames_of_report[!(rownames_of_report %in%
                                                 ssi_vars_lb)]
  }
  if (!"study" %in% vars_to_include) {
    repsumtab <-
      repsumtab[repsumtab$VAR_NAMES %in% ssi_vars, , FALSE]
    rownames_of_report <- rownames_of_report[(rownames_of_report %in%
                                                ssi_vars_lb)]
  }

  util_attach_attr(repsumtab, rownames_of_report = rownames_of_report)
}

#' Create a data frame containing all the results from summaries of reports
#'
#' @param to_combine [vector] a list containing the summaries of reports
#'                            obtained with `summary(report)`
#' @param type [character] if `type` is `unique_vars` it means that the variable
#'                        names are unique and there is not need to add a prefix
#'                        to the variables and labels to specify the
#'                        report of origin.
#'                        If `type` is `repeated_vars` a prefix will be used to
#'                        specify the report of origin of each variable
#'
#' @return a summary of summaries of `dataquieR` reports
#'
#' @noRd
util_combine_list_report_summaries <- function(to_combine,
                                               type = c("unique_vars",
                                                        "repeated_vars")) {

  # check if all the elements in to_combine are of class "dataquieR_summary"
  for (x in seq_along(to_combine)) {
    util_stop_if_not(inherits(to_combine[[x]], "dataquieR_summary"))
  }

  # Recursive function
  if (length(to_combine) == 0) {
    # create an empty dataframe
    result2 <- data.frame()
    # use var_names as labels
    label_col <- VAR_NAMES
    # create a paragraph stating that there is no result
    res <- "Empty summary"
    class(res) <- "dataquieR_summary"

    this <- new.env(parent = emptyenv())

    for (prop in c("rowmaxes", # maximum category for each row in the summary table
                   # and all the ellipsis_arguments from
                   # util_get_html_cell_for_result, above:
                   "labels_of_var_names_in_report",
                   "alias_names",
                   "stopped_functions",
                   "colors",
                   "labels",
                   "order_of",
                   "filter_of",
                   "ordered_call_names",
                   "ordered_var_names"
    )) {
      this[[prop]] <- NULL
    }
    this[["result"]] <- data.frame()
    this[["title"]] <- ""
    this[["subtitle"]] <- ""
    this[["meta_data"]] <- NULL
    this[["label_col"]] <- VAR_NAMES
    this[["rownames_of_report"]] <- character(0)
    this[["colnames_of_report"]] <- character(0)

    return(util_attach_attr(res,
                            repsum_wide = util_attach_attr(result2,
                                                           label_col = label_col),
                            rule_digest = rlang::hash(list(util_get_rule_sets(),
                                                           util_get_ruleset_formats())),
                            this = this))
  } else if (length(to_combine) == 1) {
    return(to_combine[[1]])
  } else if (length(to_combine) == 2) {
    this1 <- attr(to_combine[[1]], "this")
    this2 <- attr(to_combine[[2]], "this")

    ## Import results
    result_1 <- this1$result
    if (!prod(dim(result_1)) ||
        all(startsWith(as.character(result_1$indicator_metric), "CAT_") |
            startsWith(as.character(result_1$indicator_metric), "MSG_"))) {
      # create an empty data frame
      result_1 <- data.frame()
    }
    result_2 <- this2$result
    if (!prod(dim(result_2)) ||
        all(startsWith(as.character(result_1$indicator_metric), "CAT_") |
            startsWith(as.character(result_1$indicator_metric), "MSG_"))) {
      # create an empty data frame
      result_2 <- data.frame()
    }

    ## import meta_data
    md1 <- this1$meta_data
    md2 <- this2$meta_data

    ## check labels_of_var_names_in_report
    if (length(this1$labels_of_var_names_in_report) == 0) {  #TODO: is this possible?
      this1$labels_of_var_names_in_report <- vector("character")
    }
    if (length(this2$labels_of_var_names_in_report) == 0) {
      this2$labels_of_var_names_in_report <- vector("character")
    }


    #In case the variables are unique the type unique_vars is used
    if (type == "unique_vars") {
      #add a column origin in both result data frames (works also in case empty)
      result_1$origin <- rep(paste0(this1$title), nrow(result_1))
      result_2$origin <- rep(paste0(this2$title), nrow(result_2))
      result <- dplyr::bind_rows(result_1, result_2)
      rm(result_1, result_2)

      ## labels_of_var_names_in_report
      labels_of_var_names_in_report <- c(this1$labels_of_var_names_in_report,
                                         this2$labels_of_var_names_in_report)
      labels_of_var_names_in_report <-  #Remove duplicates
        labels_of_var_names_in_report[match(unique(
          names(labels_of_var_names_in_report)),
          names(labels_of_var_names_in_report))]
      #TODO: meta_data can not be missing right??
      ## meta_data
      if (!"new_label" %in% names(md1)) { #in case we are merging the first 2 summaries
        md1$origin <- paste0(this1$title)
        #new column for label_col with the origin followed by the original label_col
        lab1 <- md1[this1$label_col]
        names(lab1) <- "new_label"
        md1 <- cbind(md1, lab1)
      }
      md2$origin <- paste0(this2$title)
      #new column for label_col with the origin followed by the label_col
      lab1 <- md2[this2$label_col]
      names(lab1) <- "new_label"
      md2 <- cbind(md2, lab1)
      rm(lab1)
      #merge the metadata
      meta_data <- merge(md1, md2, all = TRUE) #FIXME: remove duplicates also if summaries came from 2 different metadata files

      #In case of presence of duplicated rows (excluding origin and VARIABLE_ROLE)
      meta_data <- meta_data[!duplicated(meta_data[,
                                                   setdiff(colnames(meta_data),
                                                           c("origin",
                                                             VARIABLE_ROLE)),
                                                   drop = FALSE]), ,
                             drop = FALSE]

      attr(meta_data, "version") <- 2
      attr(meta_data, "normalized") <- TRUE


      ## ordered_var_names
      ordered_var_names <- c(this1$ordered_var_names, this2$ordered_var_names)
      ordered_var_names <-
        ordered_var_names[match(unique(names(ordered_var_names)),
                                names(ordered_var_names))]

      ## rowmaxes - tibble
      rowmaxes <- dplyr::bind_rows(this1$rowmaxes, this2$rowmaxes)

      ## rownames_of_report
      rownames_of_report <- unique(c(this1$rownames_of_report,
                                     this2$rownames_of_report))

    } else if (type == "repeated_vars") { #in case vars may be not unique
      #Report name 1
      if (!"..Origin" %in% names(result_1)) {
        used_data_file <- this1$used_data_file
        #leave only file name and extension, if the vector contains a path
        if (grepl(.Platform$file.sep, used_data_file, fixed = TRUE)) {
          used_data_file <- gsub(".+\\/(.+\\..+$)", "\\1", used_data_file)
        }
        stratum <- this1$stratum
        segm <- this1$segment
        if (segm == "all_variables") {
          #in case there is only strata_column and not segment_column
          origin_name1 <- paste0(used_data_file, "-", stratum)
        } else {
          origin_name1 <- paste0(used_data_file, "-", stratum, "-", segm)
        }
        rm(used_data_file, stratum, segm)
      } else {
        origin_name1 <- unique(this1$result$..Origin)
      }

      #Report name 2
      used_data_file <- this2$used_data_file
      #leave only file name and extension, if the vector contains a path
      if (grepl(.Platform$file.sep, used_data_file, fixed = TRUE)) {
        used_data_file <- gsub(".+\\/(.+\\..+$)", "\\1", used_data_file)
      }
      stratum <- this2$stratum
      segm <- this2$segment
      if (segm == "all_variables") {
        origin_name2 <- paste0(used_data_file, "-", stratum)
      } else {
        origin_name2 <- paste0(used_data_file, "-", stratum, "-", segm)
      }
      rm(used_data_file, stratum)
      gc()

      # check that names are not the same
      orig1 <- as.vector(origin_name1)
      orig2 <- as.vector(origin_name2)
      if (any(orig1 %in% orig2)) {
        util_error("Two reports can not have the same title %s",
                   dQuote(origin_name2))
      }

      #Create a data frame to map var_names and labels to new with prefix
      if (!"..Origin" %in% names(result_1)) {
        df_map1 <- this1$meta_data
        df_map1 <- df_map1[, names(df_map1) %in%
                             c("VAR_NAMES", this1$label_col), drop = FALSE]

        df_map1$new_VAR_NAMES <- paste0(origin_name1, "-",
                                        df_map1$VAR_NAMES)
        df_map1$new_label_col <- paste0(origin_name1, "-",
                                        df_map1[, names(df_map1) %in%
                                                  this1$label_col])
      }

      df_map2 <- this2$meta_data
      df_map2 <- df_map2[, names(df_map2) %in% c("VAR_NAMES", this2$label_col), drop = FALSE]
      df_map2$new_VAR_NAMES <- paste0(origin_name2, "-", df_map2$VAR_NAMES)
      df_map2$new_label_col <- paste0(origin_name2, "-",
                                      df_map2[, names(df_map2) %in%
                                                this2$label_col])

      #add a column origin in both result data frames (works also in case empty)
      if (!"..Origin" %in% names(result_1)) {
        #in case we are merging the first 2 summaries
        result_1$..Origin <- rep(origin_name1, nrow(result_1))
        result_1$Original_var_name <- result_1$VAR_NAMES
        result_1 <- dplyr::mutate(result_1,
                                  VAR_NAMES = paste0(result_1$..Origin,
                                                     "-", result_1$VAR_NAMES))
      }
      #the second summary is always a new one in all loops
      result_2$..Origin <- rep(origin_name2, nrow(result_2))
      result_2$Original_var_name <- result_2$VAR_NAMES
      result_2 <- dplyr::mutate(result_2,
                                VAR_NAMES = paste0(result_2$..Origin,
                                                   "-", result_2$VAR_NAMES))
      result <- rbind(result_1, result_2)

      ## meta_data
      if (!"..Origin" %in% names(md1)) {
        #in case we are merging the first 2 summaries
        md1$..Origin <- rep(origin_name1, nrow(md1))
        # long label is used for the list of variables in the final table, so if
        # present I add the prefix of ..Origin
        if ("LONG_LABEL" %in% names(md1)) {
          md1 <- dplyr::mutate(md1, LONG_LABEL = paste(md1$..Origin, LONG_LABEL))
        }
      }
      md2$..Origin <- origin_name2
      #long label is used for the list of variables in the final table, so if
      # present I add the prefix of origin
      if ("LONG_LABEL" %in% names(md2)) {
        md2 <- dplyr::mutate(md2, LONG_LABEL = paste(md2$..Origin, LONG_LABEL))
      }

      if (!"new_label" %in% names(md1)) {
        #in case we are merging the first 2 summaries
        #new column for label_col with the ..Origin followed by the label_col
        lab1 <- md1[this1$label_col]
        names(lab1) <- "new_label"
        md1 <- cbind(md1, lab1)
        md1 <- dplyr::mutate(md1, new_label = paste(md1$..Origin, "-",
                                                    md1$new_label))

        ## rowmaxes - replace VAR_NAMES with VAR_NAMES with prefix Report title
        rowmaxes1 <- this1$rowmaxes
        temp1 <- util_map_labels(this1$rowmaxes$VAR_NAMES,
                                 meta_data = df_map1,
                                 to = "new_VAR_NAMES",
                                 from = VAR_NAMES)
        names(temp1) <- NULL
        rowmaxes1$VAR_NAMES <- temp1

        ## rownames_of_report
        rownames_of_report1 <- util_map_labels(this1$rownames_of_report,
                                               meta_data = df_map1,
                                               to = "new_label_col",
                                               from = this1$label_col)

        # add the report title in the VAR_NAMES and LABELS of the
        ## ordered_var_names
        labs_new <- util_map_labels(this1$ordered_var_names,
                                    meta_data = df_map1,
                                    to = "new_label_col",
                                    from = VAR_NAMES)

        # replace var_names with new_var_names
        new_var_names1 <- util_map_labels(this1$ordered_var_names,
                                          meta_data = df_map1,
                                          to = "new_VAR_NAMES",
                                          from = "VAR_NAMES")
        ord1 <- setNames(object = new_var_names1, nm =  labs_new)

        ## labels_of_var_names_in_report
        # replace labels with new_labels
        labs_new <- util_map_labels(this1$labels_of_var_names_in_report,
                                    meta_data = df_map1,
                                    to = "new_label_col",
                                    from = this1$label_col)

        # replace var_names with new_var_names
        new_var_names1 <- util_map_labels(
          names(this1$labels_of_var_names_in_report),
          meta_data = df_map1,
          to = "new_VAR_NAMES",
          from = "VAR_NAMES")
        tb_for_names1 <- setNames(object = labs_new, nm =  new_var_names1)

        # now modify VAR_NAMES also in md1
        md1$Original_var_name <- md1$VAR_NAMES
        md1 <- dplyr::mutate(md1,
                             VAR_NAMES = paste0(md1$..Origin,
                                                "-", md1$VAR_NAMES))
        rm(lab1, labs_new, new_var_names1)
      }

      #new column for label_col with the origin followed by the label_col
      lab1 <- md2[this2$label_col]
      names(lab1) <- "new_label"
      md2 <- cbind(md2, lab1)
      md2 <- dplyr::mutate(md2, new_label = paste(md2$..Origin, "-",
                                                  md2$new_label))

      ## ordered_var_names
      # add the report title in the VAR_NAMES and LABELS of the ordered_var_names
      labs_new <- util_map_labels(this2$ordered_var_names,
                                  meta_data = df_map2,
                                  to = "new_VAR_NAMES",
                                  from = VAR_NAMES)

      # replace var_names with new_var_names
      new_var_names2 <- util_map_labels(this2$ordered_var_names,
                                        meta_data = df_map2,
                                        to = "new_VAR_NAMES",
                                        from = "VAR_NAMES")
      ord2 <- setNames(object = new_var_names2, nm =  labs_new)

      if (!exists("ord1")) {
        ord1 <- this1$ordered_var_names
      }
      ordered_var_names <- c(ord1, ord2)
      ordered_var_names <-
        ordered_var_names[match(unique(names(ordered_var_names)),
                                names(ordered_var_names))]
      rm(lab1, labs_new, new_var_names2, ord1, ord2)

      ## labels_of_var_names_in_report
      # replace labels with new_labels
      labs_new <- util_map_labels(this2$labels_of_var_names_in_report,
                                  meta_data = df_map2,
                                  to = "new_label_col",
                                  from = this2$label_col)

      # replace var_names with new_var_names
      new_var_names2 <- util_map_labels(
        names(this2$labels_of_var_names_in_report),
        meta_data = df_map2,
        to = "new_VAR_NAMES",
        from = "VAR_NAMES")
      tb_for_names2 <- setNames(object = labs_new, nm =  new_var_names2)
      if (!exists("tb_for_names1")) {
        tb_for_names1 <- this1$labels_of_var_names_in_report
      }

      #merge the two vectors
      labels_of_var_names_in_report <- c(tb_for_names1, tb_for_names2)
      #Remove duplicates
      labels_of_var_names_in_report <-
        labels_of_var_names_in_report[match(
          unique(names(labels_of_var_names_in_report)),
          names(labels_of_var_names_in_report))]
      rm(labs_new, new_var_names2, tb_for_names1, tb_for_names2)

      #rowmaxes - tibble
      if (!exists("rowmaxes1")) {
        #From the second round on the name contains already the prefix
        rowmaxes1 <- this1$rowmaxes
      }
      rowmaxes2 <- this2$rowmaxes
      temp_name <- util_map_labels(this2$rowmaxes$VAR_NAMES,
                                   meta_data = df_map2,
                                   to = "new_VAR_NAMES",
                                   from = VAR_NAMES)
      names(temp_name) <- NULL
      rowmaxes2$VAR_NAMES <- temp_name
      rm(temp_name)
      rowmaxes <- rbind(rowmaxes1, rowmaxes2)

      #rownames_of_report
      if (!exists("rownames_of_report1")) {
        #From the second round on the name contains already the prefix
        rownames_of_report1 <- this1$rownames_of_report
      }
      rownames_of_report2 <- util_map_labels(this2$rownames_of_report,
                                             meta_data = md2,
                                             to = "new_label",
                                             from = this2$label_col)
      rownames_of_report <- unique(c(rownames_of_report1, rownames_of_report2))

      # now modify VAR_NAMES also in md2
      md2$Original_var_name <- md2$VAR_NAMES
      md2 <- dplyr::mutate(md2, VAR_NAMES = paste0(md2$..Origin, "-",
                                                   md2$VAR_NAMES))
      #merge the metadata
      meta_data <- merge(md1, md2, all = TRUE)
      attr(meta_data, "version") <- 2
      attr(meta_data, "normalized") <- TRUE
      rm(md1, md2)
    }

    ## alias_names
    alias_names <- c(this1$alias_names, this2$alias_names)
    alias_names <- alias_names[match(unique(names(alias_names)),
                                     names(alias_names))]

    ## ordered_call_names
    ordered_call_names <- unique(c(this1$ordered_call_names,
                                   this2$ordered_call_names))

    ## colnames_of_report
    colnames_of_report <- unique(c(this1$colnames_of_report,
                                   this2$colnames_of_report))

    #stopped_functions
    stopped_functions <- c(this1$stopped_functions, this2$stopped_functions)
    stopped_functions <-
      stopped_functions[match(unique(names(stopped_functions)),
                              names(stopped_functions))]

    #subtitle
    comp1 <- all.equal(this1$subtitle, this2$subtitle)
    if (length(comp1) == 1 &&
        comp1 == TRUE) {
      #TODO: check if it did not become too long
      subtitle <- this1$subtitle
    } else {
      subtitle <- paste0(this1$subtitle, " - ", this2$subtitle)
      attr(subtitle, "default") <- TRUE
    }

    #title
    comp1 <- all.equal(this1$title, this2$title)
    if (length(comp1) == 1 &&
        comp1 == TRUE) {
      #TODO: check if it did not become too long
      title <- this1$title
    } else {
      title <- paste0(this1$title, " - ", this2$title)
      attr(title, "default") <- TRUE
    }
    rm(comp1)

    l1 <- this1$label_col
    l2 <- this2$label_col
    while (!is.null(attr(l1, "orig_label_col"))) {
      l1 <- attr(l1, "orig_label_col")
    }
    while (!is.null(attr(l1, "orig_label_col"))) {
      l2 <- attr(l2, "orig_label_col")
    }

    if (l1 != l2) {
      util_error(
        "Cannot combine summaries if the reports used a different label_col")
    }

    #label_col
    label_col <- util_attach_attr("new_label", orig_label_col = this1$label_col)
    attr(label_col, "var_att_required") <- "recommended"

    #filter_of
    ## Combine lists by element names. It works also if all list are identical
    l_all <- list(this1$filter_of, this2$filter_of)
    keys <- unique(unlist(lapply(l_all, names)))
    list_all <- as.data.frame(do.call(mapply, c(FUN = c,
                                                lapply(l_all, `[`, keys))))
    filter_of <- lapply(
      names(list_all),
      FUN = function(x) {
        paste0(unique(list_all[x]), collapse = "-")
      }
    )
    filter_of <- setNames(filter_of, names(list_all))

    ## labels
    #labels are exactly the same as filter_of
    labels <- filter_of

    rm(l_all, keys, list_all)

    #colors
    colors <- this1$colors
    comp1 <- all.equal(this1$colors, this2$colors)
    if (length(comp1) > 1 || comp1 != TRUE) {
      util_warning("Colors of first report in the list will be used")
    }

    #order_of is always the same 5 categories and NAs
    order_of <- this1$order_of

    summary_of_summaries <- NA
    class(summary_of_summaries) <- c("dataquieR_summary")

    this <- new.env(parent = emptyenv())
    for (prop in c(
      "result",
      "rowmaxes",
      "labels_of_var_names_in_report",
      "alias_names",
      "stopped_functions",
      "colors",
      "labels",
      "order_of",
      "filter_of",
      "ordered_call_names",
      "ordered_var_names",
      "rownames_of_report"
    )) {
      this[[prop]] <- get(prop)
    }

    # needed for the summary downlaod button
    this[["title"]] <- title
    this[["subtitle"]] <- subtitle
    # needed to pretty print
    this[["meta_data"]] <- meta_data
    # "label_col", # needed to pretty print
    this[["label_col"]] <- label_col
    attr(summary_of_summaries, "this") <- this

    return(summary_of_summaries)
  } else {
    hd <- head(to_combine, 2)
    tl <- tail(to_combine, -2)
    hd_repl <- util_combine_list_report_summaries(hd, type = type)
    new_to_combine <- c(list(hd_repl), tl)
    res1 <- util_combine_list_report_summaries(new_to_combine, type = type)
    return(res1)
  }

}

#' Check for duplicated content
#'
#' @description
#' This function tests for duplicates entries in the data set. It is possible to
#' check duplicated entries by study segments or to consider only selected
#' segments.
#'
#' @inheritParams .template_function_indicator
#' @param level [character] a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").
#' @param identifier_name_list [vector] the vector that contains the name of the identifier to be used in the assessment. For the study level, corresponds to the names of the different data frames. For the segment level, indicates the name of the segments.
#' @param id_vars_list [list] the list containing the identifier variables names to be used in the assessment.
#' @param unique_rows [vector] named. for each data frame, either true/false or `no_id` to exclude ID variables from check
#' @param ... Not used.
#'
#' @return a [list] with
#'   - `SegmentData`: data frame with the results of the quality check for duplicated entries
#'   - `SegmentTable`: data frame with selected duplicated entries check results, used for the data quality report.
#'   - `Other`: vector with row indices of duplicated entries, if any, otherwise NULL.
#'
#' @family integrity_indicator_functions
#' @concept integrity_indicator
#' @keywords internal
util_int_duplicate_content_dataframe <- function(level = c("dataframe"),
                                                 identifier_name_list,
                                                 id_vars_list,
                                                 unique_rows,
                                                 meta_data_dataframe = "dataframe_level",
                                                 ...,
                                                 dataframe_level) {
  # Dataframe level check ----

  #meta_data_dup_rows_1[[DF_NAME]]

  # Check arguments ----

  util_ck_arg_aliases()

  level <- util_match_arg(level)


  if (missing(identifier_name_list) &&
      missing(id_vars_list) &&
      missing(meta_data_dataframe) &&
      formals()$meta_data_dataframe %in% prep_list_dataframes()) {
    meta_data_dataframe <- force(meta_data_dataframe)
  }

  if (missing(identifier_name_list) &&
      missing(id_vars_list) &&
      !missing(meta_data_dataframe)) {
    meta_data_dataframe <- prep_check_meta_data_dataframe(meta_data_dataframe)
    meta_data_dataframe <- meta_data_dataframe[
      vapply(meta_data_dataframe[[DF_NAME]],
             function(x) { !util_is_try_error(try(prep_get_data_frame(data_frame_name = x,
                                                                      keep_types = TRUE), silent = TRUE)) },
             FUN.VALUE = logical(1))
      , , drop = FALSE]
    meta_data_dataframe <-
      meta_data_dataframe[
        !util_empty(meta_data_dataframe[[DF_UNIQUE_ROWS]]) & (
          trimws(tolower(meta_data_dataframe[[DF_UNIQUE_ROWS]])) ==
            "no_id" |
          !util_is_na_0_empty_or_false(meta_data_dataframe[[DF_UNIQUE_ROWS]])
        ), ]
    identifier_name_list <- meta_data_dataframe[[DF_NAME]];
    id_vars_list <- lapply(setNames(meta_data_dataframe[[DF_ID_VARS]],
                                    nm = meta_data_dataframe[[DF_NAME]]),
                           util_parse_assignments,
                           multi_variate_text = TRUE )

    id_vars_list <- lapply(id_vars_list, unlist, recursive = TRUE)
    unique_rows <- setNames(meta_data_dataframe[[DF_UNIQUE_ROWS]],
                            nm = meta_data_dataframe[[DF_NAME]])
  } else if (!missing(meta_data_dataframe)) {
    util_error(c("I have %s and one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "id_vars_list"
                 )),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "id_vars_list"
                 )))
  } else if (missing(meta_data_dataframe) && (
    missing(identifier_name_list) ||
    missing(id_vars_list)
  )) {
    util_error(c("I don't have %s and also miss at least",
                 "one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "id_vars_list"
                 )),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "id_vars_list"
                 )))
  }


  util_expect_scalar(identifier_name_list,
                     allow_null = TRUE,
                     allow_more_than_one = TRUE,
                     check_type = is.character)

  if (missing(unique_rows)) {
    unique_rows <- setNames(rep(NA_character_, length(identifier_name_list)),
                            identifier_name_list);
  }
  util_expect_scalar(unique_rows, allow_more_than_one = TRUE,
                     check_type =
                       function(x) {
                         all(util_empty(x) | tolower(trimws(x)) %in%
                               c("f", "t", "true", "false", "no_id"))
                       })
  unique_rows <- tolower(trimws(unique_rows))
  unique_rows[util_empty(unique_rows)] <- "false"

  # Check for duplicated content  ----

  result <- lapply(setNames(nm = identifier_name_list), function(current_df) {

    if (trimws(tolower(unique_rows[[current_df]])) == "no_id") {
      idvl <- id_vars_list[[current_df]]
    } else {
      idvl <- character(0)
    }

    # Convert data from list to data frame
    data_current_df <- util_expect_data_frame(current_df, dont_assign = TRUE)

    data_current_df <-
      data_current_df[, setdiff(colnames(data_current_df),
                                idvl), FALSE]


    n_uniq <- nrow(unique(data_current_df))

    res_tmp <- data.frame(
      check.names = FALSE,
      "Check" = "Duplicates",
      "Data frame" = current_df,
      "Any duplicates" = ifelse(n_uniq < nrow(data_current_df), TRUE, FALSE),
      "Number of duplicates" = nrow(data_current_df) - n_uniq,
      "Percentage of duplicates" =
        round(100 * (nrow(data_current_df) - n_uniq) / nrow(data_current_df), 3),
      "GRADING" = ifelse(n_uniq < nrow(data_current_df), 1, 0),
      stringsAsFactors = FALSE
    )

    if (res_tmp[[3]]) { # only if there are any duplicated observations
      vec_dup <- which(duplicated(data_current_df)) # FIXME: Fix, this will omit the first elements, see id check for more hints, also on allowed repeats.
      vec_dup <- NULL # TODO: remove after fix
    } else {
      vec_dup <- NULL
    }

    return(list(res_tmp, vec_dup))
  })

  res_df <- do.call(rbind.data.frame, lapply(result, `[[`, 1))
  duplicates_df <- do.call(cbind.data.frame, lapply(result, `[[`, 2))

  res_pipeline <- data.frame(
    "Level" = "Dataframe",
    "DF_NAME" = res_df[["Data frame"]],
    "NUM_int_sts_dupl_content" = res_df[["Number of duplicates"]],
    "PCT_int_sts_dupl_content" = res_df[["Percentage of duplicates"]],
    "GRADING" = res_df[["GRADING"]],
    stringsAsFactors = FALSE
  )

  return(list(
    DataframeData = res_df,
    DataframeTable = res_pipeline,
    Other = duplicates_df
  ))

}

#' Warn about a problem in `varname`, if `x` has no natural order
#'
#' Also warns, if R does not have a comparison operator for `x`.
#'
#' @param x [vector] of data
#' @param varname [character] len=1. Variable name for warning messages
#'
#' @return invisible(NULL)
#'
#' @family robustness_functions
#' @concept robustness
#' @noRd
util_warn_unordered <- function(x, varname) {
  if (missing(varname)) varname <- as.character(substitute(x))
  if (sum(dim(x) != 1) > 1) {
    util_error(
      "util_warn_unordered only works on effectively one-dimensional input")
  }
  warn <- FALSE
  if (is.factor(x)) {
    if (!is.ordered(x)) {
      warn <- TRUE
    }
  } else if (!is.numeric(x)) {
    if (!identical(prep_dq_data_type_of(x),
        tolower(DATA_TYPES$DATETIME))) {
      warn <- TRUE
    }
  }
  if (warn) {
    util_warning(
      c(
        "Don't know, how to compare values of %s (%s) -- comparisons",
        "may still be possbile, but they could be meaningless."
      ),
      dQuote(varname),
      paste0(sQuote(unique(c(class(x), typeof(x)))), collapse = ", "),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE
    )
  }
  invisible(NULL)
}

test_that("util_warn_unordered works", {
  skip_on_cran()
  skip_if_translated()
  expect_warning(util_warn_unordered("Hallo", "x"),
                 perl = TRUE,
                 regexp =
                   "Don't know, how to compare values of .+x.+ (.+character.+)")
  expect_silent(util_warn_unordered(17, "xyz"))
  expect_warning(util_warn_unordered("17", "x"),
                 perl = TRUE,
                 regexp =
                   "Don't know, how to compare values of .+x.+ (.+character.+)")
  expect_warning(util_warn_unordered(complex(1, 2), "x"),
                 perl = TRUE,
                 regexp =
                   "Don't know, how to compare values of .+x.+ (.+complex.+)")
  expect_warning(util_warn_unordered(complex(1, 2), "xya"),
                 perl = TRUE,
                 regexp =
                   "Don't know, how to compare values of .+xya.+ (.+complex.+)")
  expect_warning(util_warn_unordered(complex(1, 2), "xyz"),
                 perl = TRUE,
                 regexp =
                   "Don't know, how to compare values of .+xyz.+ (.+complex.+)")
  expect_error(util_warn_unordered(cars, "xyz"),
               regexp = "only works on effectively one-dimensional input")
  expect_warning(util_warn_unordered(c, "xyz"),
                 perl = TRUE,
                 regexp =
                   "Don't know, how to compare values of .+xyz.+ (.+builtin+)")
  expect_warning(util_warn_unordered(c),
                 perl = TRUE,
                 regexp =
                   "Don't know, how to compare values of .+c.+ (.+builtin+)")
  expect_error(util_warn_unordered(matrix(1:9, nrow = 3)),
               regexp = "only works on effectively one-dimensional input")
  expect_silent(util_warn_unordered(matrix(1:9, nrow = 1)))
  expect_error(util_warn_unordered(),
               regexp = "argument .+x.+ is missing, with no default")
})

#' @noRd
util_queue_cluster_setup <- function(
                                n_nodes,
                                progress,
                                debug_parallel,
                                my_storr_object) {
  util_ensure_suggested(c("R6", "processx", "callr"))
  self <- NULL # https://github.com/r-lib/R6/issues/230#issuecomment-862462217
  private <- NULL # https://github.com/r-lib/R6/issues/230#issuecomment-862462217
  task_q <- R6::R6Class( # from the callr package by Csárdi and Chang
    "task_q",
    public = list(
      initialize = function(concurrency = n_nodes) {
        private$start_workers(concurrency)
        invisible(self)
      },
      list_tasks = function() private$tasks,
      get_num_waiting = function()
        sum(!private$tasks$idle & private$tasks$state == "waiting"),
      get_num_running = function()
        sum(!private$tasks$idle & private$tasks$state == "running"),
      get_num_done = function() sum(private$tasks$state == "done"),
      is_idle = function() sum(!private$tasks$idle) == 0,

      push = function(fun, args = list(), id = NULL) {
        if (is.null(id)) id <- private$get_next_id()
        if (id %in% private$tasks$id) stop("Duplicate task id")
        before <- which(private$tasks$idle)[1]
        private$tasks <- tibble::add_row(private$tasks, .before = before,
                                         id = id, idle = FALSE, state = "waiting", fun = list(fun),
                                         args = list(args), worker = list(NULL), result = list(NULL))
        private$schedule()
        invisible(id)
      },

#      push2 = function(expr, id = NULL, work_fkt) {
#        self$push(id = id,
#          function(work_fkt, expr, id) {
#            r <- list(
#              try(work_fkt(expr, environment()))
#            )
#            names(r) <- id
#            r
#            },
#          list(id = id,
#               expr = expr,
#               work_fkt = work_fkt))
#      },

      compute_report = function(all_calls, worker, step) {
        for (.rno in seq_len(1 + (length(all_calls) %/% step))) { # at least 1
          rno <- (step * (.rno - 1)) + seq_len(step)
          rno <- rno[rno <= length(all_calls)]
          calls <- all_calls[rno]
          nms <- names(calls)
          id <- paste0(nms, collapse = "; ")
          self$push(id = id,
                    function(work_fkt, expr, id, nms, my_storr_object) {
                      r <- mapply(SIMPLIFY = FALSE,
                                  e = setNames(expr, nm = nms),
                                  nm = nms,
                                  FUN =
                        function(e, nm) {
                          try(work_fkt(e, environment(), nm = nm,
                                       my_storr_object = my_storr_object))
                      })
                      r
                    },
                    list(id = id,
                         expr = calls,
                         work_fkt = worker,
                         nms = nms,
                         my_storr_object = my_storr_object))
        }
        r <- self$results()
        r <- r[names(all_calls)]
        return(r)
      },

#      compute_report_old = function(all_calls, worker) {
#        for (rno in seq_len(length(all_calls))) {
#          call <- all_calls[[rno]]
#          nm <- names(all_calls)[[rno]]
#          self$push2(call, id = nm, work_fkt = worker)
#        }
#        self$results()[names(all_calls)]
#      },

      poll = function(timeout = 0) {
        limit <- Sys.time() + timeout
        as_ms <- function(x) if (x == Inf) -1L else as.integer(x)
        repeat{
          topoll <- which(private$tasks$state == "running")
          conns <- lapply(
            private$tasks$worker[topoll],
            function(x) x$get_poll_connection())
          pr <- processx::poll(conns, as_ms(timeout))
          private$tasks$state[topoll][pr == "ready"] <- "ready"
          private$schedule()
          ret <- private$tasks$id[private$tasks$state == "done"]
          if (is.finite(timeout)) timeout <- limit - Sys.time()
          if (length(ret) || timeout < 0) break;
        }
        ret
      },

      pop = function(timeout = 0) {
        if (is.na(done <- self$poll(timeout)[1])) return(NULL)
        row <- match(done, private$tasks$id)
        result <- private$tasks$result[[row]]
        private$tasks <- private$tasks[-row, ]
        c(result, list(task_id = done))
      },

      workerEval = function(fkt, args) {
        private$wEval(fkt, args)
      },

      export = function(...) {
        for (nm in c(...)) {
          # cat("exporting", nm, "\n")
          private$wEval(function(x, y) {
            e <- rlang::global_env()
            assign(x = x, value = y, envir = e) }, list(
              nm,
              get(nm, parent.frame())
            ))
        }
      },

      results = function() {
        r <- list()
        n <- sum(!private$tasks$idle)
        while (!self$is_idle()) {
          task_result <- self$pop(Inf)
          progress(100 * (n-sum(!private$tasks$idle)) / n)
          # if (5 > length(task_result$result)) browser()
          r <- c(r, task_result$result)
        #   if (!self$is_idle())
        #     Sys.sleep(10)
        }
        r
      }

    ),

    private = list(
      tasks = NULL,
      next_id = 1L,
      get_next_id = function() {
        id <- private$next_id
        private$next_id <- id + 1L
        paste0(".", id)
      },

      wEval = function(fkt, args = list()) {
        stopifnot(all(private$tasks$state == "ready"))
        lapply(private$tasks$worker, function(w) {
          while (w$get_state() == "starting") {
            Sys.sleep(1)
            w$read()
          }
          w$run(fkt, args)
        })
      },

      start_workers = function(concurrency) {
        private$tasks <- tibble::tibble(
          id = character(), idle = logical(),
          state = c("waiting", "running", "ready", "done")[NULL],
          fun = list(), args = list(), worker = list(), result = list())
        for (i in seq_len(concurrency)) {

          rs <- callr::r_session$new(wait = FALSE)
          # https://github.com/r-lib/callr/issues/90#issuecomment-444278278
          rs$poll_process(1000)
          rs$read()

          private$tasks <- tibble::add_row(private$tasks,
                                           id = paste0(".idle-", i), idle = TRUE, state = "ready", # state = "running",
                                           fun = list(NULL), args = list(NULL), worker = list(rs),
                                           result = list(NULL))
        }
      },

      schedule = function() {
        ready <- which(private$tasks$state == "ready")
        if (!length(ready)) return()
        rss <- private$tasks$worker[ready]

        private$tasks$result[ready] <- lapply(rss, function(x) x$read())
        private$tasks$worker[ready] <- replicate(length(ready), NULL)
        private$tasks$state[ready] <-
          ifelse(private$tasks$idle[ready], "waiting", "done")

        waiting <- which(private$tasks$state == "waiting")[1:length(ready)]
        private$tasks$worker[waiting] <- rss
        private$tasks$state[waiting] <-
          ifelse(private$tasks$idle[waiting], "ready", "running")
        lapply(waiting, function(i) {
          if (! private$tasks$idle[i]) {
            private$tasks$worker[[i]]$call(private$tasks$fun[[i]],
                                           private$tasks$args[[i]])
          }
        })
      }
    )
  )
  task_q$new()
}

test_that("dq_report_by works with input and output dir", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_on_cran()
  target <- withr::local_tempdir("testdqareportby_v2")
  target <- gsub('\\\\','/', target)
  sd1 <- head(
    prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE),
    n = 10)

  library(rio)
  export(sd1, file.path(target, "data_origin.RData") )


  try(utils::download.file(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
    destfile = file.path(target, "meta_data_v2.xlsx"),
    quiet = TRUE, mode = "wb"), silent = TRUE)

  target1 <- file.path(target, "test01")

  expect_message2(dq_report_by(study_data = "data_origin.RData",
                              input_dir = target,
                              dimensions = "int",
                              cores = NULL,
                              strata_column = "v00002",
                              strata_select = "0",
                              meta_data_v2 = "meta_data_v2.xlsx",
                              output_dir = target1,
                              also_print = TRUE,
                              advanced_options = list(rio.import.trust = TRUE)))


  expect_error(dq_report_by(study_data = "data_origin.RData",
                              input_dir = target,
                              dimensions = "int",
                              cores = NULL,
                              strata_column = "v00002",
                              strata_select = "0",
                              meta_data_v2 = "meta_data_v2.xlsx",
                              output_dir = target1,
                              also_print = TRUE,
                            advanced_options = list(rio.import.trust = TRUE)),
               regexp = ".+already exists. cannot use this as an output folder")


  target_v3 <- withr::local_tempdir("testdqareportby_v3")
  target_v3 <- gsub('\\\\','/', target_v3)

  try(utils::download.file(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
    destfile = file.path(target_v3, "meta_data_v2.xlsx"),
    quiet = TRUE, mode = "wb"), silent = TRUE)

  sd1 <- head(
    prep_get_data_frame(
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
    n = 10)
  export(sd1, file.path(target_v3, "data_origin.RData"))

  sd2 <- tail(
    prep_get_data_frame(
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
    n = 10)
  export(sd2, file.path(target_v3, "data_origin_2.RData") )

  target2 <- file.path(target_v3, "test02")

  expect_warning(dq_report_by(study_data = c("data_origin.RData",
                                             "data_origin_2.RData"),
                              input_dir = target_v3,
                              dimensions = "int",
                              cores = NULL,
                              strata_column = "v00002",
                              strata_select = "0",
                              meta_data_v2 = "meta_data_v2.xlsx",
                              output_dir = target2,
                              also_print = TRUE,
                              advanced_options = list(rio.import.trust = TRUE)),
                 regexp = "Because no id.+created duplicated rows.")

})

test_that("util_queue_cluster_setup works", {
  skip_on_cran() # runs parallel
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  q <- util_queue_cluster_setup(1, function(...) {}, FALSE, NULL)
  r1 <- q$compute_report(all_calls = list(a = call("ls"),
                                          b = call("search"),
                                          c = call("class", as.symbol("dt0"))),
                         worker = function(exp, env, nm, my_storr_object) {
                           eval(exp, env)
                         },
                         step = 1)
  dt0 <- cars
  q$export("dt0")
  q$workerEval(function(lib) {
    require(lib, character.only = TRUE)
  }, list(lib = "ggplot2"))
  r2 <- q$compute_report(all_calls = list(a = call("ls"),
                                         b = call("search"),
                                         c = call("class", as.symbol("dt0"))),
                        worker = function(exp, env, nm, my_storr_object) {
                          eval(exp, env)
                        },
                        step = 1)
  q <- NULL
  gc()
  expect_equal(setdiff(r2$b, r1$b), "package:ggplot2")
  expect_s3_class(r1$c, "try-error")
  expect_equal(conditionMessage(attr(r1$c, "condition")),
               "object 'dt0' not found")
  expect_equal(r2$c, "data.frame")
  expect_equal(r1$a, r2$a)
  expect_equal(r1$a, c("e", "nm"))
})

#' Integer breaks for `ggplot2`
#'
#' creates integer-only breaks
#'
#' @param x the values
#' @param n integer giving the desired number of intervals. Non-integer values are rounded down.
#'
#' @return breaks suitable for `scale_*_continuous`' `breaks` argument
#'
#' @author [Sarah](https://stackoverflow.com/users/2738526/sarah)
#'
#' @seealso [StackOverflow](https://stackoverflow.com/a/57086284)
#'
#' @noRd
#'
#' @examples
#' \dontrun{
#' big_numbers1 <- data.frame(x = 1:5, y = c(0:1, 0, 1, 0))
#' big_numbers2 <- data.frame(x = 1:5, y = c(0:1, 0, 1, 0) + 1000000)
#'
#' big_numbers_plot1 <- ggplot(big_numbers1, aes(x = x, y = y)) +
#'   geom_point()
#'
#' big_numbers_plot2 <- ggplot(big_numbers2, aes(x = x, y = y)) +
#'   geom_point()
#'
#' big_numbers_plot1 + scale_y_continuous()
#' big_numbers_plot1 + scale_y_continuous(breaks = util_int_breaks_rounded)
#'
#' big_numbers_plot2 + scale_y_continuous()
#' big_numbers_plot2 + scale_y_continuous(breaks = util_int_breaks_rounded)
#' }
#'
util_int_breaks_rounded <- function(x, n = 5) {
  unique(pretty(x, n)[round(pretty(x, n),1) %% 1 == 0])
}

#' Encoding Errors
#'
#' @description
#' Detects errors in the character encoding of string variables
#'
#' [Indicator]
#'
#' @details
#' Strings are stored based on
#' [code tables](https://en.wikipedia.org/wiki/Character_encoding), nowadays,
#' typically as
#' [UTF-8](https://en.wikipedia.org/wiki/UTF-8). However, other code systems
#' are still in use, so, sometimes, strings from different systems are mixed
#' in the data. This indicator checks for such problems and returns the count
#' of entries per variable, that do not match the reference coding system, which
#' is estimated from the study data (addition of metadata field is planned).
#'
#' If not specified in the metadata (columns `ENCODING` in item- or data-frame-
#' level, the encoding is guessed from the data). Otherwise, it may be any
#' supported encoding as returned by `iconvlist()`.
#'
#' @inheritParams .template_function_indicator
#' @param ref_encs reference encodings (names are `resp_vars`)
#'
#' @return a list with:
#'   - `SummaryTable`: [data.frame] with information on such problems
#'   - `SummaryData`: [data.frame] human readable version of `SummaryTable`
#'   - `FlaggedStudyData`: [data.frame] tells for each entry in study data if
#'                         its encoding is OK. has the same dimensions as
#'                         `study_data`

#' @export
int_encoding_errors <- function(resp_vars = NULL,
                                study_data,
                                label_col,
                                meta_data_dataframe =
                                  "dataframe_level",
                                item_level = "item_level",
                                ref_encs,
                                meta_data = item_level,
                                meta_data_v2,
                                dataframe_level) {


  util_maybe_load_meta_data_v2()
  util_first_arg_study_data_or_resp_vars()
  util_ck_arg_aliases()

  meta_data_dataframe <-
    try(prep_check_meta_data_dataframe(meta_data_dataframe), silent = TRUE)

  if (util_is_try_error(meta_data_dataframe)) {
    meta_data_dataframe <- data.frame()
  }

  util_expect_data_frame(study_data, keep_types =  TRUE,
                         keep_encoding_errors = TRUE)

  if (missing(label_col)) {
    label_col <- VAR_NAMES
  }

  if (!missing(ref_encs)) {
    util_expect_scalar(ref_encs, allow_more_than_one = TRUE)
  }

  meta_data <- try(util_expect_data_frame(meta_data), silent = TRUE)

  if (util_is_try_error(meta_data) || nrow(meta_data) == 0) {
    meta_data <- data.frame(
      VAR_NAMES = colnames(study_data),
      DATA_TYPE = prep_datatype_from_data(resp_vars = colnames(study_data),
                                          study_data =
                                            study_data)
    )
    meta_data[[label_col]] <- meta_data[[VAR_NAMES]]
  }

  util_expect_data_frame(meta_data, col_names = c(
    VAR_NAMES,
    label_col
  ))

  if (missing(resp_vars) || is.null(resp_vars)) {
    resp_vars <- colnames(study_data)
  }

  util_expect_scalar(resp_vars,
                     allow_more_than_one = TRUE,
                     error_message =
                       "Need either no resp_vars or valid variable names")

  rv <- util_find_var_by_meta(resp_vars,
                              meta_data = meta_data,
                              target = VAR_NAMES,
                              ifnotfound = NA,
                              label_col = label_col);
  rv <- rv[!is.na(rv)]
  study_data <- study_data[, rv, FALSE]
  resp_vars <- prep_get_labels(rv,
                               label_col = label_col,
                               item_level = meta_data,
                               label_class = "LONG",
                               resp_vars_are_var_names_only = TRUE)
  ds1_label_col <- attr(resp_vars, "label_col")
  colnames(study_data) <- resp_vars
  ds1 <- study_data

  # correct variable use?
  util_correct_variable_use(resp_vars,
                            allow_more_than_one = TRUE,
                            allow_null          = TRUE,
                            allow_any_obs_na    = TRUE # not using need_type allow running w/o metadata
  )
  resp_vars_character <- names(which(vapply(study_data,
                                is.character,
                                FUN.VALUE = logical(1))))

  if (!any(resp_vars %in% resp_vars_character)) {
    util_error(c("No %s are of type character, cannot work on",
                   "the non-character variables."),
                 sQuote("resp_vars"),
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
  } else if (!all(resp_vars %in% resp_vars_character)) {
    util_message(c("Not all %s are of type character, leave out",
                   "the non-character variables."),
                 sQuote("resp_vars"),
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
  }
  resp_vars <- intersect(resp_vars,
            resp_vars_character)
  if (missing(ref_encs)) {
    # ref_encs <- vapply(res, attr, "ref_enc", FUN.VALUE = character(1))
    ref_encs <- util_get_encoding(study_data = ds1,
                                  label_col = ds1_label_col,
                                  meta_data = meta_data,
                                  meta_data_dataframe = meta_data_dataframe)
  }
  res <- util_verify_encoding(study_data[,
                                         resp_vars,
                                         FALSE], ref_encs = ref_encs)
  cols_enc <- attr(res, "cols_enc")
  enc_guessed <- vapply(setNames(nm = names(cols_enc)), function(nm) {
    x <- cols_enc[[nm]]
    ref_enc <- "---~unkonwn~----"
    try(
      ref_enc <- ref_encs[[nm]],
      silent = TRUE
    )
    x <- x[order(x[["confidence"]], decreasing = TRUE), , FALSE]
    x[x[["encoding"]] == ref_enc, "encoding"] <- paste("(*)", ref_enc)
    prep_deparse_assignments(x[["encoding"]],
                             x[["confidence"]],
                             mode = "string_codes")
  }, FUN.VALUE = character(1))
  num_prolems_per_var <-
    vapply(res, length, FUN.VALUE = integer(1))
  num_prolems_per_var[setdiff(resp_vars, names(num_prolems_per_var))] <- 0
  num_prolems_per_var <- num_prolems_per_var[resp_vars]
  pct_prolems_per_var <- 100 * num_prolems_per_var / nrow(study_data) # TODO: or remove NAs for denominator?
  SummaryTable <- data.frame(check.names = FALSE,
    Variables = resp_vars, # resp_vars_lcol[resp_vars],
    `Defined Encoding` = ref_encs[resp_vars],
    `Guessed Encoding` = enc_guessed[resp_vars],
    PCT_int_uenc = pct_prolems_per_var,
    NUM_int_uenc = num_prolems_per_var
  )
  # TODO: Fix this after we have an indicator for this in DQ_OBS
  SummaryData <- SummaryTable[, 1:3]
  SummaryData$`Mismatches` <- paste0(
    num_prolems_per_var, " (",
    util_round_to_decimal_places(pct_prolems_per_var), "%)"
  )
  FlaggedStudyData <- study_data
  FlaggedStudyData[] <- FALSE
  FlaggedStudyData[] <- lapply(names(FlaggedStudyData), function(cn) {
    r <- FlaggedStudyData[[cn]]
    if (cn %in% names(res)) {
      r[as.integer(res[[cn]])] <- TRUE
    }
    r
  })

  resp_vars_lcol <- prep_map_labels(SummaryData$Variables,
                                    ifnotfound = SummaryData$Variables,
                                    item_level = meta_data,
                                    to = label_col,
                                    from = ds1_label_col)

  SummaryData$Variables <- util_attach_attr(SummaryData$Variables,
                                            plain_label =
                                              resp_vars_lcol[resp_vars])

  rownames(SummaryTable) <- NULL # html table would display them
  rownames(SummaryData) <- NULL
  rownames(FlaggedStudyData) <- NULL

  list(
    SummaryTable = SummaryTable,
    SummaryData = SummaryData,
    FlaggedStudyData = FlaggedStudyData
  )
}

# Encoding(nako_WS_fil$ws_note) <- "UTF-8"
# nako_WS_fil$ws_note<- iconv(nako_WS_fil$ws_note, "UTF-8", "UTF-8", sub = '')

#' Utility function to parse assignments
#'
#' This function parses labels & level assignments in the format
#' `1 = male | 2 = female`. The function also handles `m = male | f = female`,
#' but this would not match the metadata concept. The split-character can
#' be given, if not the default from [SPLIT_CHAR] is to be used, but this
#' would also violate the metadata concept.
#'
#' @param text Text to be parsed
#' @param split_char Character separating assignments, may be a vector, then
#'                   all will be tried and the the most likely matching one will
#'                   be returned as attribute `split_char` of the result.
#' @param multi_variate_text don't paste text but parse element-wise
#' @param split_on_any_split_char split on any split `split_char`, if > 1 given.
#'
#' @return the parsed assignments as a named list
#'
#' @importFrom stats setNames
#'
#' @examples
#' \dontrun{
#' md <- prep_get_data_frame("meta_data")
#' vl <- md$VALUE_LABELS
#' vl[[50]] <- "low<medium < high"
#' a <- util_parse_assignments(vl, split_char = c(SPLIT_CHAR, "<"),
#'   multi_variate_text = TRUE)
#' b <- util_parse_assignments(vl, split_char = c(SPLIT_CHAR, "<"),
#'   split_on_any_split_char = TRUE, multi_variate_text = TRUE)
#' is_ordered <- vapply(a, attr, "split_char", FUN.VALUE = character(1)) == "<"
#' md$VALUE_LABELS[[50]] <- "low<medium < high"
#' md$VALUE_LABELS[[51]] <- "1 = low< 2=medium < 3=high"
#' md$VALUE_LABELS[[49]] <- "2 = medium< 1=low < 3=high" # counter intuitive
#' with_sl <- prep_scalelevel_from_data_and_metadata(study_data = "study_data",
#'   meta_data = md)
#' View(with_sl[, union(SCALE_LEVEL, colnames(with_sl))])
#' }
#'
#' @family parser_functions
#' @concept metadata_management
#' @noRd
util_parse_assignments <- function(text, split_char = SPLIT_CHAR,
                                   multi_variate_text = FALSE, # Dont change default here, many calls of this function rely on a non-list-result
                                   split_on_any_split_char = FALSE
                                   ) {
  if (!multi_variate_text) {
    if (all(util_empty(text))) {
      text <- NA_character_
    } else {
      if (is.list(text)) {
        text <- unlist(text, recursive = TRUE, use.names = FALSE)
      }
      text <- paste0(text, collapse = "\n")
    }
  }
  use_regexp <- FALSE
  if (length(split_char) > 1) {
    if (!split_on_any_split_char) {
      r <- lapply(X = setNames(nm = split_char),
                  FUN = util_parse_assignments,
                  text = text,
                  multi_variate_text = TRUE)
      rl <- lapply(r, vapply, length, FUN.VALUE = integer(1))
      rl <- vapply(rl, identity, FUN.VALUE = integer(length(text)))
      if (is.null(dim(rl))) {
        rl <- matrix(rl,
                     ncol = length(split_char),
                     dimnames = list(as.character(seq_len(length(text))),
                                     names(rl)))
      }
      rlwm <- apply(rl, 1, which.max, simplify = FALSE)
      atts <- vapply(rlwm, names,
                     FUN.VALUE = character(1))
      res <- mapply(SIMPLIFY = FALSE,
                    wm = rlwm, i = seq_along(rlwm), function(wm, i) {
        util_attach_attr(r[[wm]][[i]], split_char = atts[[i]])
      })
      return(res)
    } else {
      use_regexp <- TRUE
    }
  }
  res <- lapply(text, function(x) {
    if (is.list(x)) {
      x <- paste(unlist(x, recursive = TRUE, use.names = FALSE),
                 collapse = "\n")
    }
    if (use_regexp) {
      split <- sprintf("[%s]", paste0(gsub("[", "\\[", fixed = TRUE,
                                           gsub("]", "\\]", fixed = TRUE,
                                                split_char)), collapse = ""))
      if (all(util_empty(gsub(split, "", perl = TRUE, x)))) {
        return(setNames(list(), nm = character(0)))
      }
      assignments <- base::strsplit(x = as.character(x),
                                    split = split, perl = TRUE)[[1]]
    } else {
      if (all(util_empty(gsub(split_char, "", fixed = TRUE, x)))) {
        return(setNames(list(), nm = character(0)))
      }
      assignments <- base::strsplit(x = as.character(x),
                                    split = split_char, fixed = TRUE)[[1]]
    }

    keys <- trimws(gsub(pattern = "(?ms)\\s*=\\s*.*$", replacement = "",
                        x = assignments, perl = TRUE))
    values <- trimws(gsub(pattern = "(?ms)^.*?\\s*=\\s*", replacement = "\\1",
                          x = assignments, perl = TRUE))
    as.list(setNames(values, keys))
  })
  res <- lapply(res,
                function(r) {
                  if (!length(r)) {
                    return(r)
                  } else if (
                    all(startsWith(as.character(unlist(r)), "'") &
                        endsWith(as.character(unlist(r)), "'"))) {
                    r <- substr(r, 2, nchar(r))
                    r <- substr(r, 1, nchar(r) - 1)
                  } else                   if (
                    all(startsWith(as.character(unlist(r)), '"') &
                        endsWith(as.character(unlist(r)), '"'))) {
                    r <- substr(r, 2, nchar(r))
                    r <- substr(r, 1, nchar(r) - 1)
                  }
                  r
                })
  if (getOption("dataquieR.VALUE_LABELS_htmlescaped",
                dataquieR.VALUE_LABELS_htmlescaped_default)) {
    util_ensure_suggested(
      "textutils",
      "use the option(dataquieR.VALUE_LABELS_htmlescaped = TRUE) ")
    res[] <- lapply(res, lapply, textutils::HTMLdecode)
  }
  if (!multi_variate_text) {
    if (length(res) != 1) { # nocov start
      util_error(c("univariate use of util_parse_assignments returned %d",
                   "results. Sorry, this should not happen, internal error."),
                 length(res))
    } # nocov end
    return(res[[1]])
  } else {
    return(res)
  }
}

test_that("util_parse_assignments works", {
  skip_on_cran()
  expected <- list("1" = "married", "2" = "single", "3" = "divorced",
                   "4" = "widowed")

  util_parse_assignments(
    "1 = married| \n 2 \t = \r  single|   3 =divorced|4=widowed")

  expect_equal(
    util_parse_assignments(
      "1 = married| \n 2 \t = \r  single|   3 =divorced|4=widowed"),
    expected
  )
})

test_that(paste("util_parse_assignments equivalent with an old independent",
                "development from con inadmissible categories"), {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")

  got <- lapply(lapply(setNames(meta_data[[VALUE_LABELS]], nm =
                                  meta_data$VAR_NAMES),
                util_parse_assignments), unlist) # get for all item

  got <- mapply(lapply(got, names), got, FUN = setNames) # swap names and values

  label_col <- LABEL
  rvs <- meta_data[[LABEL]]

  expected <-
    lapply(
      lapply(
        setNames(meta_data[[VALUE_LABELS]][meta_data[[label_col]] %in% rvs],
                 nm = meta_data[[VAR_NAMES]][meta_data[[label_col]] %in% rvs]),
        function(x) {
          lapply(
            trimws(unlist(strsplit(x, SPLIT_CHAR, fixed = TRUE))),
            function(x) {
              setNames(lapply(
                x,
                function(x) trimws(unlist(strsplit(x, "=", fixed = TRUE))[1])
              ), nm = lapply(
                x,
                function(x) {
                  y <- unlist(strsplit(x, "=", fixed = TRUE))
                  if (length(y) > 1)
                    trimws(paste(y[-1], collapse = "="))
                  else
                    trimws(paste(y, collapse = "=")) # in case of no name, use
                                                     # the value as the name
                }
              ))
            }
          )
        }
      ),
      unlist
    )

  expected[vapply(expected, identical, c("NA" = NA_character_),
                  FUN.VALUE = logical(1))] <- list(NULL)

  expect_equal(got, expected)

})

#' Select really [numeric] variables
#'
#' Reduce `resp_vars` to those, which are either `float` or `integer` without
#' [VALUE_LABELS], i.e. likely [numeric] but not a [factor]
#'
#' @param resp_vars [variable list] len=1-2. the name of the continuous
#'                                           measurement variable
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param warn [logical] warn about removed variable names
#' @param stop [logical] stop on no matching `resp_var`
#'
#' @return [character] vector of matching `resp_vars`.
#'
#' @importFrom stats setNames
#'
#' @family metadata_management
#' @concept metadata_management
#' @noRd
util_no_value_labels <- function(resp_vars, meta_data, label_col, warn = TRUE,
                                 stop = TRUE) {
  util_stop_if_not(ncol(meta_data) > 1)
  matching_vars <- vapply(FUN.VALUE = logical(1), X = setNames(nm = resp_vars),
                          FUN = function(rv) {
    row <- meta_data[meta_data[[label_col]] == rv, , TRUE]
    vl <- row[[VALUE_LABELS]]
    vlt <- row[[VALUE_LABEL_TABLE]]
    svt <- row[[STANDARDIZED_VOCABULARY_TABLE]]
    dt <- row[[DATA_TYPE]]
    if (dt == DATA_TYPES$FLOAT) {
      return(TRUE)
    } else if (dt == DATA_TYPES$INTEGER) {
      return(
        (length(vl) == 0 || is.na(vl) || trimws(vl) == "") &&
        (length(vlt) == 0 || is.na(vlt) || trimws(vlt) == "") && # TODO: check, if exists
        (length(svt) == 0 || is.na(svt) || trimws(svt) == "") # TODO: check, if exists
      )
    } else {
      return(FALSE)
    }
  })
  if (stop && all(!matching_vars)) {
    util_error(c(
      "None of the variables %s is float or integer",
      "without VALUE_LABELS; aborting."),
      paste0(dQuote(resp_vars), collapse = ", "),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE)
  } else if (warn && any(!matching_vars)) {
    util_message(
      c("The variables %s are neither float nor integer",
        "without VALUE_LABELS. Ignoring those"),
      paste0(dQuote(resp_vars[!matching_vars]), collapse = ", "),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE
    )
  }
  resp_vars[matching_vars]
}

test_that("util_no_value_labels works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  local({
    meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
    expect_identical(util_no_value_labels("CENTER_0", meta_data = meta_data,
                                          label_col = LABEL, stop = FALSE,
                                          warn = FALSE), character(0))
    expect_message2(util_no_value_labels("CENTER_0", meta_data = meta_data,
                                        label_col = LABEL, stop = FALSE,
                                        warn = TRUE),
      regexp = paste("The variables .+CENTER_0.+ are neither float",
                     "nor integer without VALUE_LABELS. Ignoring those"),
      perl = TRUE
    )
    expect_error(util_no_value_labels("CENTER_0", meta_data = meta_data,
                                      label_col = LABEL, stop = TRUE,
                                      warn = FALSE),
      regexp = paste("None of the variables .+CENTER_0.+ is float or",
                     "integer without VALUE_LABELS; aborting."),
      class = "error"
    )
    expect_identical(
      sort(util_no_value_labels(resp_vars = meta_data[[LABEL]],
                                meta_data = meta_data, label_col = LABEL,
                                stop = FALSE, warn = FALSE)),
      c(
        "AGE_0", "AGE_1", "ARM_CIRC_0", "BSG_0", "CRP_0", "DBP_0", "DEV_NO_0",
        "GLOBAL_HEALTH_VAS_0", "ITEM_1_0", "ITEM_2_0", "ITEM_3_0",
        "ITEM_4_0", "ITEM_5_0", "ITEM_6_0", "ITEM_7_0", "ITEM_8_0",
        "N_ATC_CODES_0", "N_BIRTH_0", "N_CHILD_0", "N_INJURIES_0", "SBP_0"
      )
    )
  })
})

#' Check for integer values
#'
#' This function checks if a variable is integer.
#'
#' @param x the object to test
#' @param tol precision of the detection. Values deviating more than `tol` from
#'            their closest integer value will not be deemed integer.
#'
#' @return `TRUE` or `FALSE`
#' @seealso is.integer
#'
#' Copied from the documentation of \link[base]{is.integer}
#'
#' \link[base]{is.integer} detects, if the storage mode of an R-object is
#' integer. Usually, users want to know, if the values are integer. As suggested
#' by \link[base]{is.integer}'s documentation, `is.wholenumber` does so.
#'
#'
#' @family robustness_functions
#' @concept data_management
#' @noRd
util_is_integer <- function(x, tol = .Machine$double.eps^0.5) {
  if (is.numeric(x)) {
    r <- abs(x - round(x)) < tol & !is.nan(x)
    # & x <= .Machine$integer.max & this would return, whether x can be stored as an integer.
    #   x >= - .Machine$integer.max
  } else {
    r <- rep(FALSE, length(x))
  }
  r[is.na(r)] <- TRUE # NA is not not an integer
  r
}

#' Sort a vector by order given in some other vector
#'
#' @param x the vector
#' @param order the "order vector
#' @param ... additional arguments passed to `sort`
#'
#' @examples
#' \dontrun{
#' util_sort_by_order(c("a", "b", "a", "c", "d"), letters)
#' }
#'
#' @family summary_functions
#' @concept reporting
#' @noRd

util_sort_by_order <- function(x, order, ...) {
  x <- ordered(x, order)
  as.character(sort(x, ...))
}

#' Get the order of a vector with general order given in some other vector
#'
#' @param x the vector
#' @param order the "order vector
#' @param ... additional arguments passed to `order`
#'
#' @examples
#' \dontrun{
#' util_order_by_order(c("a", "b", "a", "c", "d"), letters)
#' }
#' @family reporting_functions
#' @concept summary
#' @noRd
util_order_by_order <- function(x, order, ...) {
  x <- ordered(x, order)
  order(x, ...)
}

#' Get sub-string left from first `.`
#'
#' @param x the string with a least one `.`
#'
#' @examples
#' \dontrun{
#' util_sub_string_left_from_.(c("a.b", "asdf.xyz", "asdf.jkl.zuio"))
#' }
#'
#' @family string_functions
#' @concept reporting
#' @noRd
util_sub_string_left_from_. <- function(x) {
  pos <- regexpr(".", x, fixed = TRUE)

  util_stop_if_not(
    "there must be at least one dot to separate the col. name from the row name (wolog this is the first dot)" =
      all(pos > 0)
  )

  substring(x, 1L, pos - 1L)
}
#' Get sub-string right from first `.`
#'
#' @param x the string with a least one `.`
#'
#' @examples
#' \dontrun{
#' util_sub_string_right_from_.(c("a.b", "asdf.xyz"))
#' util_sub_string_right_from_.(c("a.b", "asdf.xy.z"))
#' util_sub_string_right_from_.(c("ab", "asdxy.z"))
#' util_sub_string_right_from_.(c("a.b", "asdxy.z."))
#' util_sub_string_right_from_.(c("a.b", "asdxy.z.."))
#' }
#'
#' @family string_functions
#' @concept reporting
#' @noRd
util_sub_string_right_from_. <- function(x) {
  pos <- regexpr(".", x, fixed = TRUE)

  util_stop_if_not(
    "there must be at least one dot to separate the col. name from the row name (wolog this is the first dot)" =
      all(pos > 0)
  )

  substring(x, pos + 1L)
}

#' Collect all errors, warnings, or messages so that they are combined for a combined result
#'
#' @noRd

util_collapse_msgs <- function(class, all_of_f) { # class is either error, warning or message
  # extract and create a list of all the messages by class
  allmsgsofclass <- lapply(all_of_f, attr, class)
  # TODO: This is maybe not very general, but the function is used for result attached messages, not for the general overview
  allmsgsofclass <- lapply(allmsgsofclass, function(msgofclass) {
    msgofclass[!vapply( # do not show conditions about calls, that would not be possible at all (e.g., loess for only categorical scales)
      lapply(msgofclass, attr, "intrinsic_applicability_problem"),
      identical,
      TRUE,
      FUN.VALUE = logical(1)
    )]
  })
  # /TODO
  msgs <- lapply(allmsgsofclass, vapply, conditionMessage, FUN.VALUE = character(1)); # extract and create a list of all the messages by class
  # the messages are grouped to avoid repetitions, so the messages are amended
  nms <- gsub("^.*?\\.", "", names(msgs), perl = TRUE)
  # remove variable names
  dfr <- do.call(rbind.data.frame, mapply(SIMPLIFY = FALSE, msgs, nms, FUN = function(msg_lst, varname) {
    msg_lst <- gsub(sprintf("(\\W|^)\\Q%s\\E(\\W|$)", varname), "<VARIABLE>", msg_lst, perl = TRUE)
    if (length(msg_lst) > 0) {
      r <- data.frame(varname = varname, message = msg_lst)
    } else {
      r <- data.frame(varname = character(0), message = character(0))
    }
    r
  }))
  # remove numbers
  dfr$message <- gsub("(\\W|^)\\d+(\\W|$)", " <NUMBER> ", dfr$message, perl = TRUE)
  dfr$message <- trimws(gsub(" +", " ", dfr$message, perl = TRUE))
  if (!prod(dim(dfr))) {
    return(character(0))
  }
  # group messages
  dfr <- dplyr::summarize(dplyr::group_by(dfr, message), varname =
                            paste(get("varname"), collapse = ", "))
  msgs <-
    mapply(dfr$varname, dfr$message,
           FUN = function(name, msg) {
             nm <- dQuote(strsplit(name, ", ", fixed = TRUE)[[1]])
             if (length(nm) > 5) { # truncate variable names for the current warning message group
               nm <- c(head(nm, 4), "...")
             } else if (length(nm) == 1 && nm == dQuote("[ALL]")) {
               nm <- "all variables"
             }
             paste0("For ", paste(nm, collapse = ", "), ": ", msg)
           })
  unname(msgs)
}

#' Create a `ggplot2` pie chart
#'
#' needs `htmltools`
#'
#' @param data data as returned by `prep_summary_to_classes` but
#'             summarized by one column (currently, we support
#'             `indicator_metric`, `STUDY_SEGMENT`, and `VAR_NAMES`)
#' @param meta_data [meta_data]
#'
#' @return a `htmltools` compatible object or `NULL`, if package is missing
#'
#' @family summary_functions
#' @export
prep_render_pie_chart_from_summaryclasses_ggplot2 <- function(data,
                                                      meta_data = "item_level") {
  vars_to_include <- attr(data, "vars_to_include")
  ssi <- (identical(vars_to_include, "ssi"))
  te <- topenv(parent.frame(1)) # see https://stackoverflow.com/a/27870803
  if (!(isNamespace(te) && getNamespaceName(te) == "dataquieR")) {
    lifecycle::deprecate_soft("2.1.0.9007",
                              "prep_render_pie_chart_from_summaryclasses_ggplot2()",
                              "plot.dataquieR_summary()")
  }
  if (!util_ensure_suggested(c("htmltools"), err = FALSE)) {
    return(NULL)
  }

  if (nrow(data) == 0) {
    return(htmltools::browsable(htmltools::HTML("")))
  }

  grouped_by <- setdiff(colnames(data), c("class", "value", "percent", "note"))

  util_stop_if_not(length(grouped_by) == 1)

  groups <- unique(data[[grouped_by]])

  util_stop_if_not(length(groups) > 0)

  if (length(groups) > 1) {
    all_pys <- lapply(setNames(nm = groups), function(g) {
      prep_render_pie_chart_from_summaryclasses_ggplot2(
        data[data[[grouped_by]] == g, , FALSE], meta_data = meta_data)
    })
    ncols <- min(2, ceiling(sqrt(length(all_pys))))
    nrows <- ceiling(length(all_pys) / ncols)

    pys_matrix <- htmltools::tags$table(lapply(
      seq_len(nrows),
      function(rw) {
        if ((rw - 1) * ncols + 1 <= length(all_pys)) { # current row needed?
          cur_tr_as_list <- lapply(seq_len(ncols), function(rw, cl) {
            wch <- (rw - 1) * ncols + cl
            if (wch <= length(all_pys)) {
              o <- htmltools::tags$td(all_pys[wch])
            } else {
              o <- htmltools::tags$td()
            }
            o
          }, rw = rw)
          do.call(htmltools::tags$tr, cur_tr_as_list)
        } else { # should never be reached
          NULL
        }
      }
    ))

    py <- pys_matrix

    py <- htmltools::tags$table(py)
    py <- htmltools::browsable(py)
    return(py)
  }

  gg_colors <- util_get_colors()
  names(gg_colors) <- util_get_labels_grading_class()[names(gg_colors)]

  if (all(is.na(data$class))) {
    return(htmltools::browsable(htmltools::HTML("")))
  }

  if (is.factor(data$class)) {
    data$class <- as.integer(gsub("^cat", "", data$class))
  }
  data$class <- factor(data$class,
                       levels = names(util_get_labels_grading_class()),
                       labels = util_get_labels_grading_class(),
                       ordered = TRUE
  )

  data <- data[order(data$class, -data$value, decreasing = TRUE), , FALSE]


  p <- ggplot(data, aes(x = 1,
                        y = value,
                        fill = class)) +
    geom_bar(stat = "identity", width = 1) +
    geom_text(
      aes(x = 1.1,
          label = scales::percent(value / sum(value))),
      position = ggplot2::position_stack(vjust = 0.5),
      check_overlap = TRUE
    ) +
    ggplot2::coord_polar("y", start = 0, clip = "off") +
    scale_fill_manual(values = gg_colors) +
    ggplot2::theme_minimal() +
    ggplot2::theme(axis.line = ggplot2::element_blank(),
                   axis.text.x = ggplot2::element_blank(),
                   axis.text.y = ggplot2::element_blank(),
                   axis.ticks = ggplot2::element_blank(),
                   axis.title.x = ggplot2::element_blank(),
                   axis.title.y = ggplot2::element_blank(),
                   # legend.position = "none",
                   legend.title = ggplot2::element_blank(),
                   panel.background = ggplot2::element_blank(),
                   panel.border = ggplot2::element_blank(),
                   panel.grid.major = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank(),
                   plot.background = ggplot2::element_blank())

  if (identical(grouped_by, "indicator_metric")) { # TODO: Hiearchical structure support
    title <- util_translate_indicator_metrics(groups, short = FALSE,
                                              long = FALSE)
    subtitle <- "percentage of QA classes"
  } else if (identical(grouped_by, "call_names")) { # TODO: Hiearchical structure support
    fnms <- util_cll_nm2fkt_nm(groups)
    if (nchar(fnms) < nchar(groups)) {
      suff <- gsub("^.*_", ": ", groups)
      substr(suff, 3, 3) <- toupper(substr(suff, 3, 3))
    } else {
      suff <- ""
    }
    title <- paste0(util_map_labels(fnms,
                                    util_get_concept_info("implementations"),
                                    to = "dq_report2_short_title",
                                    from = "function_R",
                                    ifnotfound = NA_character_), suff)
    subtitle <- "percentage of QA classes"
  } else if (identical(grouped_by, as.character(STUDY_SEGMENT))) { # TODO: Hiearchical structure support
    title <- groups
    subtitle <- "percentage of QA classes"
  } else if (identical(grouped_by, as.character(VAR_NAMES))) {
    util_expect_data_frame(meta_data)
    title <- prep_get_labels(groups,
                             meta_data = meta_data,
                             label_class = "LONG")
    subtitle <- "percentage of QA classes"
  } else if (identical(grouped_by, as.character("function_name"))) {
    title <- vapply(groups, util_alias2caption, long = TRUE,
                    FUN.VALUE = character(1))
    subtitle <- groups
  } else {
    title <- groups
    subtitle <- "percentage of QA classes"
    #subtitle <- NULL
    #    util_error("Unkown grouping by %s", sQuote(grouped_by))
  }

  if (!ssi) # FIXME: also for ssi
    subtitle <- paste(subtitle, sprintf(" -- %d of %d %s classified",
                                        sum(data$value, na.rm = TRUE),
                                        nrow(meta_data),
                                        ifelse(ssi, "scales", "variables")
  )) # TODO: Maybe, we should not compute this here, but earlier.

  p <- p + ggtitle(
    title, subtitle
  )

  htmltools::plotTag(width = 400, height = 400,
    p, alt = "A summary of data quality assessments"
  )

}

util_gg_get <- function(x, name, default = NULL) {
  if (is.null(x) || is.null(name)) {
    return(default)
  }

  out <- default

  # --- ggproto objects: don't second-guess their internals -------------------
  if (inherits(x, "ggproto")) {
    # For ggproto, $ is the canonical accessor for fields
    val <- tryCatch(x[[name]], error = function(e) NULL)
    if (!is.null(val)) {
      out <- val
    } else if (exists(name, envir = x, inherits = FALSE)) {
      out <- get(name, envir = x, inherits = FALSE)
    } else {
      return(default)
    }

    # --- environments (non-ggproto) -------------------------------------------
  } else if (is.environment(x)) {
    if (exists(name, envir = x, inherits = FALSE)) {
      out <- get(name, envir = x, inherits = FALSE)
    } else {
      return(default)
    }

    # --- named lists -----------------------------------------------------------
  } else if (is.list(x) && !is.null(names(x)) && name %in% names(x)) {
    out <- x[[name]]

    # --- attributes fallback ---------------------------------------------------
  } else {
    att <- attr(x, name, exact = TRUE)
    if (!is.null(att)) {
      out <- att
    } else {
      return(default)
    }
  }

  # For some known list-y fields, normalize environments to lists
  if (is.environment(out) &&
      !inherits(x, "ggproto") &&
      name %in% c("data", "layers", "panel_scales_x")) {
    out <- as.list(out)
  }

  out
}

#' Convert a study variable to a [factor]
#'
#' @param resp_vars [variable list] the name of the measurement variables
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param assume_consistent_codes [logical] assume, that missing codes are
#'                                          consistent for all variables
#' @param have_cause_label_df [logical] is a missing-code table available
#' @param code_name [character] all lists from the [meta_data] to use for the
#'                              coding.
#' @param include_sysmiss [logical] add also a factor level for data values
#'                                  that were `NA` in the original study data
#'                                  (system missingness).
#'
#' @return `study_data` converted to factors using the coding provided in
#'         `code_name`
#'
#' @family data_management
#' @concept metadata_management
#' @noRd
util_study_var2factor <- function(resp_vars = NULL, study_data,
                                  meta_data = "item_level", # TODO: make a prep, see util_assign_levlabs, make it work also for VALUE_LABELS, not only for missing codes, optionally. Default should be FALSE or all existing calls need to be edited to turn off this feature. Maybe, the prep is just a copy of the util with modified defaults, as in  progress <- other_function; formals(progress)$is_rstudio <- force(is_rstudio)
                                  label_col = LABEL,        # TODO: This is now available in prep_prepare_dataframes(.apply_factor_metadata = ) or prep_prepare_dataframes(.apply_factor_metadata_inadm = )
                                  assume_consistent_codes = TRUE,
                                  have_cause_label_df = FALSE,
                                  code_name = c(JUMP_LIST, MISSING_LIST),
                                  include_sysmiss = TRUE) {
  util_expect_scalar(assume_consistent_codes, check_type = is.logical)
  util_expect_scalar(include_sysmiss, check_type = is.logical)
  util_expect_scalar(code_name, allow_more_than_one = TRUE, allow_null = TRUE,
                     check_type = is.character)
  util_expect_data_frame(meta_data, code_name)
  .meta_data <- meta_data # FIXME: Save metadata here, but we should not rely on VALUE_LABELS, any more
  prep_prepare_dataframes(.replace_missings = FALSE)
  meta_data <- .meta_data
  util_correct_variable_use2(resp_vars,
                             allow_more_than_one = TRUE,
                             allow_na = TRUE,
                             allow_null = TRUE,
                             allow_all_obs_na = TRUE,
                             allow_any_obs_na = TRUE)
  if (!length(resp_vars)) {
    resp_vars <- colnames(ds1)
  }
  r <- lapply(setNames(nm = resp_vars), function(rv) {
    l <- util_get_combined_code_lists(rv,
                                      mdf = meta_data,
                                      code_name = code_name,
                                      label_col = label_col,
                                      warning_if_no_list = FALSE,
                                      assume_consistent_codes = assume_consistent_codes,
                                      have_cause_label_df = have_cause_label_df)
    orig_r <- ds1[[rv]]
    r <- ds1[[rv]]
    if (lubridate::is.timepoint(r)) {
      r <- as.character(r)
      r <- gsub(" 00:00:00$", "", r) # only whole days are allowed missing codes
    }
    r <- suppressWarnings(
      factor(r,
             labels = names(l),
             levels = l))
    if (include_sysmiss) {
      levels(r) <- c(levels(r), .SM_LAB)
      r[is.na(orig_r)] <- .SM_LAB
    }
    r
  })
  ds2 <- ds1[, resp_vars, drop = FALSE]
  ds2[] <- r

  return(ds2)

}

#' Open a data frame in Excel
#'
#' @details
#' if the file cannot be read on function exit, NULL will be returned
#'
#'
#' @param dfr the data frame
#'
#' @export
#'
#' @return potentially modified data frame after dialog was closed
prep_open_in_excel <- function(dfr) { # nocov start
  nms <- as.character(substitute(dfr))
  if (length(nms) != 1)
    nms <- "dataframe"
  fn <- paste0(nms, ".xlsx")
  if (is.character(dfr)) {
    fn <- paste0(dfr, ".xlsx")
  }
  util_expect_data_frame(dfr)
  withr::with_tempdir({
    if (util_ensure_suggested("openxlsx2", "Freeze the header row",
                              err = FALSE)) {
      openxlsx2::wb_workbook()$
        add_worksheet()$
        add_data(x = dfr)$
        add_filter(sheet = 1, rows = 1, seq_len(ncol(dfr)))$
        save(fn)
    } else {
      rio::export(dfr, format = "xlsx", file = fn)
    }
    fn <- normalizePath(fn)
    browseURL(fn)
    rstudioapi::showDialog("Excel", "Close, when you have finished using Excel")
    res <- NULL
    try({
      res <- rio::import(fn)
    })
  })
  invisible(res)
} # nocov end

#' Generate a tag for a specific result
#'
#' for `dq_report2`
#'
#' @param varname variable to create an anchor for
#' @param callname function call to create an anchor for
#' @param order_context anchor created on variable overview or indicator
#'                      overview page
#' @param name replaces `varname` and `callname`, must contain the `.`
#'             separator, then
#'
#' @return the `htmltools` tag
#'
#' @family reporting_functions
#' @concept process
#' @noRd
util_generate_anchor_tag <- function(
    varname,
    callname,
    order_context = c("variable",
                      "indicator"),
    name) {
  util_ensure_suggested("htmltools")
  if (!missing(name)) {
    util_stop_if_not(missing(varname) && missing(callname))
    varname <- sub("^.*?\\.", "", name)
    callname <- sub("^([^\\.]*).*$", "\\1", name)
    if (varname == callname && callname == name) {
      if (order_context == "variable") {
        varname <- ""
      } else {
        callname <- ""
      }
    }
    if (varname == "[ALL]") varname <- ""
    if (callname == "[ALL]") callname <- ""
  }
  dot <- ""
  if (nzchar(varname)) {
    varname <- prep_link_escape(varname, html = TRUE)
    if (nzchar(callname)) {
      dot <- "."
    }
  }
  if (util_match_arg(order_context) == "variable") {
    id <- sprintf("%s%s%s", varname, dot, callname)
  } else {
    id <- sprintf("%s%s%s", callname, dot, varname)
  }
  htmltools::a(id = id)
}

#' Generate a link to a specific result
#'
#' for `dq_report2`
#'
#' @param varname variable to create a link to
#' @param callname function call to create a link to
#' @param order_context link created to variable overview or indicator
#'                      overview page
#' @param name replaces `varname` and `callname`, must contain the `.`
#'             separator, then
#' @param title optional, replaces auto-generated link title
#'
#' @return the `htmltools` tag
#'
#' @family reporting_functions
#' @concept process
#' @noRd
util_generate_anchor_link <- function(
    varname,
    callname,
    order_context = c("variable",
                      "indicator"),
    name,
    title) {
  util_ensure_suggested("htmltools")
  if (!missing(name)) {
    util_stop_if_not(missing(varname) && missing(callname))
    varname <- sub("^.*?\\.", "", name)
    callname <- sub("^([^\\.]*).*$", "\\1", name)
    if (varname == callname && callname == name) {
      if (order_context == "variable") {
        util_error("For links to an variable page, I need to know the variable")
      } else {
        util_error(
          "For links to an indicator page, I need to know the indicator")
      }
    }
    if (varname == "[ALL]") varname <- ""
    if (callname == "[ALL]") callname <- ""
  }
  orig_varname <- varname
  dot <- ""
  if (nzchar(varname)) {
    varname <- prep_link_escape(varname, html = TRUE)
    if (nzchar(as.character(callname))) {
      dot <- "."
    }
  }
  if (util_match_arg(order_context) == "variable") {
    if (!missing(title)) {
      .title <- title
    } else {
      .title <- prep_title_escape(util_alias2caption(callname, long = TRUE),
                                  html = TRUE)
    }
    href <- sprintf("VAR_%s.html#%s%s%s", varname, varname, dot, callname)
  } else {
    .title <- prep_title_escape(orig_varname, html = TRUE)
    if (startsWith(callname, "acc_")) {
      fil <- paste0("dim_", "acc_", callname, ".html")
      # the "dim_" prefix is required because otherwise windows will ignore a file called con.html confusing it with a special device con:
    } else {
      dim <- sub("_.*$", "", callname)
      fil <- paste0("dim_", dim, ".html")
    }
    href <- sprintf("%s#%s%s%s", fil, callname, dot, varname)
  }
  if (!missing(title)) {
    .title <- title
  }
  if (!nzchar(.title) && any(grepl(".[ALL]", fixed = TRUE, name))) {
    .title <- gsub(".[ALL]", "", fixed = TRUE, name)
  }
  htmltools::a(href = href, .title)
}

#' Recycle vectors to a common size
#'
#' Recycles multiple vectors to a common target size, similar in spirit to
#' `vctrs::vec_recycle_common()`, but limited to plain vector-like inputs.
#'
#' Recycling is only allowed if each input has length `1` or if its length
#' divides the target size without remainder.
#'
#' @param ... Vectors to recycle.
#' @param .size Optional target size. If omitted, the maximum input length is
#'   used.
#'
#' @return A list containing the recycled inputs in the original order.
#'
#' @noRd
util_recycle <- function(..., .size = NULL) {
  xs <- list(...)

  if (!length(xs)) {
    return(list())
  }

  is_vector_like <- vapply(
    xs,
    function(x) is.atomic(x) || is.list(x) || is.expression(x),
    logical(1L)
  )

  if (!all(is_vector_like)) {
    bad <- which(!is_vector_like)[1L]
    util_error(
      "Argument %s is of unsupported class %s",
      bad,
      sQuote(paste(class(xs[[bad]]), collapse = ", "))
    )
  }

  lens <- lengths(xs)

  if (is.null(.size)) {
    .size <- max(lens)
  } else {
    if (!is.numeric(.size) || length(.size) != 1L || is.na(.size) ||
        .size < 0 || .size != as.integer(.size)) {
      util_error(
        "Argument %s must be a single non-negative integer",
        sQuote(".size")
      )
    }
    .size <- as.integer(.size)
  }

  bad_len <- which(lens != 1L & lens != 0L & (.size %% lens != 0L))
  if (length(bad_len)) {
    i <- bad_len[1L]
    util_error(
      "Can't recycle argument %s of length %s to size %s",
      i,
      lens[[i]],
      .size
    )
  }

  lapply(xs, function(x) rep(x, length.out = .size))
}

test_that("util_recycle() returns empty list for no inputs", {
  expect_identical(util_recycle(), list())
})

test_that("util_recycle() leaves equally sized vectors unchanged", {
  x <- 1:3
  y <- c("a", "b", "c")

  res <- util_recycle(x, y)

  expect_length(res, 2L)
  expect_identical(res[[1L]], x)
  expect_identical(res[[2L]], y)
})

test_that("util_recycle() recycles length-1 vectors to common size", {
  x <- 1
  y <- 1:4

  res <- util_recycle(x, y)

  expect_length(res, 2L)
  expect_identical(res[[1L]], c(1, 1, 1, 1))
  expect_identical(res[[2L]], y)
})

test_that("util_recycle() recycles shorter vectors to maximum size", {
  x <- 1:2
  y <- 1:4

  res <- util_recycle(x, y)

  expect_length(res, 2L)
  expect_identical(res[[1L]], c(1L, 2L, 1L, 2L))
  expect_identical(res[[2L]], y)
})

test_that("util_recycle() respects explicit .size", {
  x <- 1:2
  y <- 9

  res <- util_recycle(x, y, .size = 4L)

  expect_length(res, 2L)
  expect_identical(res[[1L]], c(1L, 2L, 1L, 2L))
  expect_identical(res[[2L]], c(9, 9, 9, 9))
})

test_that("util_recycle() supports zero-length vectors if target size is zero", {
  x <- integer()
  y <- character()

  res <- util_recycle(x, y)

  expect_length(res, 2L)
  expect_identical(res[[1L]], integer())
  expect_identical(res[[2L]], character())
})

test_that("util_recycle() errors for incompatible lengths", {
  expect_error(
    util_recycle(1:2, 1:3),
    "Can't recycle"
  )
})

test_that("util_recycle() errors for incompatible explicit .size", {
  expect_error(
    util_recycle(1:2, .size = 3L),
    "Can't recycle"
  )
})

test_that("util_recycle() errors for invalid .size", {
  expect_error(
    util_recycle(1:2, .size = NA_integer_),
    ".size"
  )
  expect_error(
    util_recycle(1:2, .size = -1L),
    ".size"
  )
  expect_error(
    util_recycle(1:2, .size = 1.5),
    ".size"
  )
  expect_error(
    util_recycle(1:2, .size = c(2L, 3L)),
    ".size"
  )
})

test_that("util_recycle() errors for unsupported inputs", {
  expect_error(
    util_recycle(environment()),
    "unsupported class"
  )
})

#' Combine two report summaries
#'
#' @param ... objects returned by [prep_extract_summary]
#' @param summaries_list if given, [list] of objects returned by
#'                       [prep_extract_summary]
#' @param amend_segment_names [logical] use names of the `summaries_list` and
#'                            argument names as segment prefixes
#'
#' @return combined summaries
#' @family summary_functions
#' @export
prep_combine_report_summaries <- function(...,
                                          summaries_list,
                                          amend_segment_names = FALSE) {
  te <- topenv(parent.frame(1)) # see https://stackoverflow.com/a/27870803
  if (!(isNamespace(te) && getNamespaceName(te) == "dataquieR")) {
    lifecycle::deprecate_soft("2.1.0.9007",
                              "prep_combine_report_summaries()")
  }
  if (missing(summaries_list)) summaries_list <- list()
  if (!is.list(summaries_list)) {
    util_error("%s needs to be a list, if passed", sQuote("summaries_list"))
  }
  summaries_list <- c(list(...), summaries_list)
  util_expect_scalar(amend_segment_names, check_type = is.logical)

  if (amend_segment_names && any(util_empty(names(summaries_list)))) {
    util_error(
      "All summaries must be named arguments/list elements, if %s is %s",
      sQuote("amend_segment_names"),
      dQuote("TRUE"))
  }

  if (!all(vapply(summaries_list, function(summary) {
    inherits(summary, "dq_report2_summary") &&
      all(c("Data", "Table", "meta_data") %in% names(summary))
  }, FUN.VALUE = logical(1)))) {
    util_error(c("All summaries passed must be returned by %s or this function",
                 "i.e. feature %s, %s and %s"),
               sQuote("prep_extract_summary"),
               sQuote("Data"), sQuote("Table"), sQuote("meta_data"))
  }

  has_study_segment <- function(x) STUDY_SEGMENT %in% colnames(x)
  if (!all(vapply(FUN.VALUE = logical(1),
         lapply(summaries_list, `[[`, "Data"),
         has_study_segment
  ))) {
    util_error(
      c("All summaries passed must be returned by %s or this function,",
        "i.e., feature %s entries with %s columns"),
      sQuote("prep_extract_summary"),
      sQuote("Data"),
      sQuote(STUDY_SEGMENT))
  }

  if (!all(vapply(FUN.VALUE = logical(1),
                  lapply(summaries_list, `[[`, "Table"),
                  has_study_segment
  ))) {
    util_error(
      c("All summaries passed must be returned by %s or this function,",
        "i.e., feature %s entries with %s columns"),
      sQuote("prep_extract_summary"),
      sQuote("Table"),
      sQuote(STUDY_SEGMENT))
  }

  if (amend_segment_names) {
    summaries_list <- mapply(SIMPLIFY = FALSE,
                             summary = summaries_list,
                             name = names(summaries_list),
                             function(summary, name) {
                               summary$Data[[STUDY_SEGMENT]] <-
                                 paste(name, summary$Data[[STUDY_SEGMENT]],
                                       sep = ": ")
                               summary$Table[[STUDY_SEGMENT]] <-
                                 paste(name, summary$Table[[STUDY_SEGMENT]],
                                       sep = ": ")
                               if (!STUDY_SEGMENT %in%
                                   colnames(summary$meta_data)) {
                                 summary$meta_data[[STUDY_SEGMENT]] <- "Study"
                               }
                               summary$meta_data[[STUDY_SEGMENT]] <-
                                 paste(name, summary$meta_data[[STUDY_SEGMENT]],
                                       sep = ": ")
                               summary
                             })
  }

  summaries_list <- lapply(summaries_list, function(summary) {
    rownames(summary$Data) <- NULL
    rownames(summary$Table) <- NULL
    summary
  })

  combined <- list(
    Data =
      util_rbind(data_frames_list = lapply(summaries_list, `[[`, "Data")),
    Table =
      util_rbind(data_frames_list = lapply(summaries_list, `[[`, "Table")),
    meta_data =
      util_rbind(data_frames_list = lapply(summaries_list, `[[`, "meta_data"))
  )

  # VAR_NAMES may be ambiguous
  if (any(vapply(lapply(lapply(combined, `[[`, VAR_NAMES), duplicated),
         any,
         FUN.VALUE = logical(1)))) {
    util_warning( # FIXME: This warning is shown, even, if it is possible
      # FIXME: Allow to combine such reports, if the results are distinct,
      # still, i.e., one report with ICC, the other with item
      # missingness
      # by definiion, a message, but the effects are too big, so here a warning
      c("Some of the summaries comprise overlapping variables, will",
        "pick the first summary for each"),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = FALSE)
    combined <-
      lapply(combined, function(x) {
        x <- x[!duplicated(x[[VAR_NAMES]]), , FALSE]
        x
      })
  }

  class(combined) <- "dq_report2_summary"

  combined
}

test_that("prep_combine_report_summaries works", {
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_on_cran()
  skip_if_not_installed("stringdist")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  sd0 <- study_data[, 1:6]
  sd0$v00012 <- study_data$v00012
  md0 <- subset(meta_data, VAR_NAMES %in% colnames(sd0))
  md0$PART_VAR <- NULL

  report <- dq_report2(sd0, md0,
                       resp_vars = c("v00000", "v00001", "v00002",
                                     "v00003", "v00004", "v00012"),
                       filter_indicator_functions =
                         c("^com_item_missingness$",
                           "^acc_varcomp$"),
                       filter_result_slots =
                         c("^SummaryTable$"),
                       cores = NULL,
                       dimensions = # for speed, omit Accuracy
                         c("Integrity",
                           "Completeness",
                           "Consistency",
                           "Accuracy"))

  report2 <- dq_report2(sd0, md0,
                       resp_vars = c("v00000", "v00001", "v00002",
                                     "v00003", "v00005", "v00012"),
                       filter_indicator_functions =
                         c("^com_item_missingness$",
                           "^acc_varcomp$"),
                       filter_result_slots =
                         c("^SummaryTable$"),
                       cores = NULL,
                       dimensions = # for speed, omit Accuracy
                         c("Integrity",
                           "Completeness",
                           "Consistency",
                           "Accuracy"))

  suppressWarnings({
    summary <- prep_extract_summary(report)
    summary2 <- prep_extract_summary(report2)
  })

  # ignore deprecation warning and known bug of a superfluous warning
  suppressWarnings(comb_sum <- prep_combine_report_summaries(summary, summary2))

  expect_snapshot(comb_sum)
})

#' Test, if values of x are empty, i.e. NA or whitespace characters
#' @param x the vector to test
#' @return a logical vector, same length as x; TRUE, if resp. element in x is
#'         "empty"
#'
#' @family robustness_functions
#' @concept process
#' @noRd
util_empty <- function(x) {
  xx <- "not_empty"
  if (util_is_try_error(try(xx <- trimws(x), silent = TRUE))) {
    try(xx <- vapply(x,
                     function(y) {
                       z <- try(r <- trimws(y), silent = TRUE)
                       if (util_is_try_error(z)) {
                         r <- y
                       }
                       r
                      },
                     FUN.VALUE = character(1)), silent = TRUE)
  }
  unname(is.na(x) | xx == "")
}

test_that("util_empty works", {
  skip_on_cran()
  expect_identical(util_empty(c()), logical(0))
  expect_identical(util_empty(c()), logical(0))
  expect_identical(util_empty(c("", "k")), c(TRUE, FALSE))
  expect_identical(util_empty(c("", "k", NA)), c(TRUE, FALSE, TRUE))
  expect_identical(util_empty(NA), TRUE)
  expect_identical(util_empty(Inf), FALSE)
  expect_identical(util_empty(NaN), TRUE)
  expect_identical(util_empty(0), FALSE)
})

#' Call a function with `...`, dropping arguments not present in formals
#'
#' Helper to forward `...` to a target function while silently removing
#' arguments that are not part of the target's formals (unless the target
#' itself has `...`). The call is evaluated in the caller environment so that
#' functions depending on the calling frame (e.g. `ls()`) behave as expected.
#'
#' @param f A function (or something coercible via [rlang::as_function()]).
#' @param ... Arguments to forward.
#'
#' @return The return value of `f(...)`.
#'
#' @noRd
util_call_with_only_existing_formals <- function(f, ...) {
  f <- rlang::as_function(f)

  dots <- rlang::dots_list(...)
  fm   <- rlang::fn_fmls(f)

  if (!("..." %in% names(fm))) {
    dots <- dots[intersect(names(dots), names(fm))]
  }

  rlang::eval_bare(
    rlang::call2(f, !!!dots),
    env = rlang::caller_env()
  )
}

#' Evaluate a parsed redcap rule for given study data
#'
#' also allows to use `VAR_NAMES` in the rules,
#' if other labels have been selected
#'
#' @param rule the redcap rule (parsed, already)
#' @param ds1 the study data as prepared by `prep_prepare_dataframes`
#' @param meta_data the metadata
#' @param use_value_labels map columns with `VALUE_LABELS` as factor variables
#' @param replace_limits [logical] replace hard limit violations by `NA`, implies
#'                                 `replace_missing_by = NA`
#' @param replace_missing_by [enum] LABEL | INTERPRET | NA . Missing codes should
#'                                  be replaced by the missing labels, the
#'                                  `AAPOR` codes from the missing table or
#'                                  by `NA`. Can also be an empty string to keep
#'                                  the codes.
#'
#' @return the result of the parsed rule
#'
#' @family redcap
#' @concept process
#' @noRd
util_eval_rule <- function(rule, ds1, meta_data = "item_level",
                           use_value_labels,
                           replace_missing_by = "NA",
                           replace_limits = TRUE) { # IDEA: Use the new .apply_factor_metadata options, somehow
  if (replace_limits && replace_missing_by != "NA") {
    util_message(
      c("Cannot replace hard limits, if missing codes are not deleted. I will",
        "therefore replace the missing codes by NA, too."),
      applicability_problem = TRUE)
    replace_missing_by <- "NA"
  }
  use_missing_labels <- FALSE
  if (replace_missing_by %in% c("LABEL", "INTERPRET")) {
    use_missing_labels <- TRUE
    use_value_labels <- TRUE
  }
  if (missing(meta_data)) { # TODO: Have a default, now?!
    use_value_labels <- FALSE
    meta_data <- data.frame()
  }
  util_expect_data_frame(meta_data)

  if (replace_missing_by %in% c("LABEL", "INTERPRET")) {
    # missing label and AAPOR codes currently only supported for
    # categorical variables:
    vars_data_type <- setNames(nm = all.vars(rule),
                               util_find_var_by_meta(
                                 resp_vars = all.vars(rule),
                                 target = "DATA_TYPE",
                                 meta_data = meta_data))
    if (!VALUE_LABELS %in% colnames(meta_data)) {
      meta_data[[VALUE_LABELS]] <- NA_character_
    }
    if (!VALUE_LABEL_TABLE %in% colnames(meta_data)) {
      meta_data[[VALUE_LABEL_TABLE]] <- NA_character_
    }

    # TODO: Support:
    # if (!STANDARDIZED_VOCABULARY_TABLE %in% colnames(meta_data)) {
    #   meta_data[[STANDARDIZED_VOCABULARY_TABLE]] <- NA_character_
    # }
    vars_value_labels <- setNames(nm = all.vars(rule),
                                  util_find_var_by_meta(
                                    resp_vars = all.vars(rule),
                                    target = VALUE_LABELS,
                                    meta_data = meta_data))
    vars_value_label_tables <- setNames(nm = all.vars(rule),
                                  util_find_var_by_meta(
                                    resp_vars = all.vars(rule),
                                    target = VALUE_LABEL_TABLE,
                                    meta_data = meta_data))
    not_yet_supported <- intersect(names(vars_data_type)[which(vars_data_type %in% c("float", "integer"))],
                                   union(
                                     names(vars_value_labels)[which(util_empty(vars_value_labels))], # TODO: STANDARDIZED_VOCABULARY_TABLE
                                     names(vars_value_label_tables)[which(util_empty(vars_value_label_tables))]
                                   ))
    if (length(not_yet_supported) > 0) {
      util_error(paste0("Replacement of missing value codes by labels or ",
                        "AAPOR codes is not yet supported for numerical ",
                        "variables (",
                        paste0(not_yet_supported, collapse = ", "),
                        ")."),
                        applicability_problem = TRUE)
    }
  }
  if (!missing(ds1)) {
    problem <-
      isTRUE(attr(ds1, ("Codes_to_NA"))) &&
         replace_missing_by != "NA" # then, an unsuitable ds1 has been passed
    util_stop_if_not(!problem) # an unsuitable ds1 has been passed
    if (!!prod(dim(meta_data))) {
      ds1 <- prep_prepare_dataframes(.replace_hard_limits = replace_limits,
                              .replace_missings = (replace_missing_by == "NA"),
                              .study_data = ds1,
                              .label_col = attr(ds1, "label_col"))
    }
  }
  if (missing(use_value_labels)) {
    use_value_labels <- VALUE_LABELS %in% colnames(meta_data) &&
      any(!util_empty(meta_data[[VALUE_LABELS]]))
    use_value_labels <- use_value_labels || (
      VALUE_LABEL_TABLE %in% colnames(meta_data) &&
      any(!util_empty(meta_data[[VALUE_LABEL_TABLE]])))
    # TODO: STANDARDIZED_VOCABULARY_TABLE
  }
  if (missing(ds1)) {
    ds1 <- parent.frame()
  }
  if (is.environment(ds1)) {
    ds1 <- as.list(ds1)
  }
  if (!is.list(ds1)) {
    util_error("%s must be an environment or a list (or a dataframe)",
               dQuote("ds1"))
  }
  redcap_rule_env <- util_get_redcap_rule_env()
  label_col <- attr(ds1, "label_col")
  if (use_value_labels || use_missing_labels) {
    util_stop_if_not(attr(ds1, "MAPPED"))
    util_stop_if_not(!util_empty(label_col))
    if (!(VALUE_LABEL_TABLE %in% colnames(meta_data))) {
      meta_data[[VALUE_LABEL_TABLE]] <- NA_character_
    }
    if (!(VALUE_LABELS %in% colnames(meta_data))) {
      meta_data[[VALUE_LABELS]] <- NA_character_
    }
    cols_with_valuelabels <- meta_data[!util_empty(meta_data[[VALUE_LABELS]]) |
                                    !util_empty(meta_data[[VALUE_LABEL_TABLE]]), # TODO: address VALUE_LABELS = c("m|f", "") ...
                                       label_col, drop = TRUE] # TODO: STANDARDIZED_VOCABULARY_TABLE
    ds1_with_labels <- ds1
    if (!MISSING_LIST_TABLE %in% colnames(meta_data)) {
      meta_data[[MISSING_LIST_TABLE]] <- NA_character_
    }
    ds1_with_labels[, cols_with_valuelabels] <- lapply(cols_with_valuelabels,
      function(cn) {
        if (replace_missing_by == "INTERPRET") {
          mltab_name <- meta_data[
            meta_data[[label_col]] == cn,
            MISSING_LIST_TABLE,
            drop = TRUE]
          if (!is.na(mltab_name))
            mltab <-
              try(prep_get_data_frame(mltab_name), silent = TRUE)
          else
            mltab <- data.frame(
              CODE_INTERPRET = character(0),
              CODE_VALUE = character(0)
            )
          if (inherits(mltab, "try-error") ||
              !(CODE_INTERPRET %in% colnames(mltab)) ||
              !(CODE_VALUE %in% colnames(mltab))) {
            util_warning(
              c("For %s, I have no %s assignments from %s, so I will",
                "not replace missing codes."),
              dQuote(cn),
              sQuote(CODE_INTERPRET),
              sQuote(CODE_VALUE)
            )
            CODE_INTERPRET <- NULL;
          } else {
            mltab <-
              util_filter_missing_list_table_for_rv(mltab,
                                                    cn,
                                                    prep_map_labels( # TODO: Do the mapping outside the apply function loop
                                                      cn,
                                                      meta_data = meta_data,
                                                      to = VAR_NAMES,
                                                      from = label_col,
                                                      ifnotfound = cn,
                                                      warn_ambiguous = FALSE
                                                    ))
            CODE_INTERPRET <-
              setNames(mltab[[CODE_VALUE]], nm = mltab[[CODE_INTERPRET]])
          }
        }

        labs <- # TODO: discard, we have now VALUE_LABEL_TABLE, mostly always
          meta_data[
            meta_data[[label_col]] == cn,
            VALUE_LABELS,
            drop = TRUE]

        util_stop_if_not(
`Internal error, sorry, please report. VALUE_LABELS still in util_eval_rule()` =
          length(labs) == 0 ||
            all(util_empty(labs))
        )

        labtabs <- meta_data[
          meta_data[[label_col]] == cn,
          VALUE_LABEL_TABLE,
          drop = TRUE]

        labs <- vapply(labtabs, # FIXME: This is a slow and ugly work around.
                       function(vlt) {
                         vltdf <- try(prep_get_data_frame(vlt))
                         if (is.data.frame(vltdf)) {
                           if (CODE_VALUE %in% colnames(vltdf)) {
                             if (!CODE_LABEL %in% colnames(vltdf)) {
                               vltdf[[CODE_LABEL]] <- vltdf[[CODE_VALUE]]
                             }
                             return(prep_deparse_assignments(
                               mode = "string_codes",
                               codes = vltdf[[CODE_VALUE]],
                               labels = vltdf[[CODE_LABEL]]))
                           }
                         }
                         return(SPLIT_CHAR)
                       }, FUN.VALUE = character(1))


        if (!use_value_labels) {
          labs <- paste(unique(ds1[[cn]]), # TODO: This can definitely be bad in terms of performance
                        collapse = sprintf(" %s ", SPLIT_CHAR))
        }
        # if (all(util_empty(labs))) {
        #   labs <- paste(unique(ds1[[cn]]), # TODO: This can definitely be bad in terms of performance
        #                 collapse = sprintf(" %s ", SPLIT_CHAR))
        # }
        labs <- as.character(labs)
        ml <- util_parse_assignments(multi_variate_text = TRUE,
          meta_data[
            meta_data[[label_col]] == cn,
            MISSING_LIST,
            drop = TRUE])[[1]]
        ml <- ml[!util_empty(gsub(SPLIT_CHAR, "", ml))]
        mcodes <- suppressWarnings(trimws(names(ml)))
        jl <- util_parse_assignments(multi_variate_text = TRUE,
          meta_data[
            meta_data[[label_col]] == cn,
            JUMP_LIST,
            drop = TRUE])[[1]]
        jl <- jl[!util_empty(gsub(SPLIT_CHAR, "", jl))]
        jcodes <- suppressWarnings(trimws(names(jl)))

        if (replace_missing_by == "") { # keep the codes
          jl <-
            prep_deparse_assignments(codes = jcodes,
                                     labels = jcodes)
          ml <-
            prep_deparse_assignments(codes = mcodes,
                                     labels = mcodes)
        } else if (replace_missing_by == "LABEL") { # use the missing/jump labels
          jl <-
            prep_deparse_assignments(codes = jcodes,
                                     labels = jl)
          ml <-
            prep_deparse_assignments(codes = mcodes,
                                     labels = ml)
        } else if (replace_missing_by == "INTERPRET") { # use AAPOR codes
          if (!is.null(CODE_INTERPRET)) {
            jl <- ""
            ml <-
              prep_deparse_assignments(codes = CODE_INTERPRET,
                                       labels = names(CODE_INTERPRET))
          } else { # keep the codes
            util_warning("I do not have a %s with %s to use them for rules",
                         sQuote(MISSING_LIST_TABLE),
                         sQuote(CODE_INTERPRET))
            jl <-
              prep_deparse_assignments(codes = jcodes,
                                       labels = jcodes)
            ml <-
              prep_deparse_assignments(codes = mcodes,
                                       labels = mcodes)
          }
        } else {
          # "NA" is already handled before by prep_prepare_dataframes with .replace_missings
          jl <- ""
          ml <- ""
        }

        jl <- jl[!is.na(jl)]
        ml <- ml[!is.na(ml)]

        if (!any(grepl(SPLIT_CHAR, labs, fixed = TRUE)) &&
            (any(grepl("<", labs, fixed = TRUE)))) {
          labs <- gsub("<", SPLIT_CHAR, labs, fixed = TRUE)
        }

        labs <- paste(c(labs, jl, ml), collapse = sprintf(" %s ", SPLIT_CHAR))

        col_as_factor <-
          util_assign_levlabs(variable = ds1[[cn]],
                            string_of_levlabs = labs,
                            splitchar = SPLIT_CHAR,
                            assignchar = "=",
                            variable_name = cn,
                            warn_if_inadmissible = FALSE) # inadmissible categorical values are checked in an indicator, already
        if (!is.factor(col_as_factor)) {
          col_as_factor <- as.factor(col_as_factor)
        }
        col_as_factor
      }
    )
    ds2 <- ds1_with_labels
  } else {
    ds2 <- ds1
  }
  ds2_0 <- ds2
  if (length(label_col) == 1 && label_col != VAR_NAMES) {
    ds2_varnames <- ds2_0
    colnames(ds2_varnames) <-
      util_map_labels(colnames(ds2_varnames),
                      meta_data = meta_data,
                      to = VAR_NAMES,
                      from = label_col)
    ds2 <- cbind(ds2, ds2_varnames)
  }
  if (length(label_col) == 1 && label_col != LABEL) {
    ds2_varnames <- ds2_0
    colnames(ds2_varnames) <-
      util_map_labels(colnames(ds2_varnames),
                      meta_data = meta_data,
                      to = LABEL,
                      from = label_col)
    ds2 <- cbind(ds2, ds2_varnames)
  }
  if (LONG_LABEL %in% colnames(meta_data) &&
      length(label_col) == 1 && label_col != LONG_LABEL) {
    ds2_varnames <- ds2_0
    colnames(ds2_varnames) <-
      util_map_labels(colnames(ds2_varnames),
                      meta_data = meta_data,
                      to = LONG_LABEL,
                      from = label_col)
    ds2 <- cbind(ds2, ds2_varnames)
  }
  if ("ORIGINAL_VAR_NAMES" %in% colnames(meta_data)) {
    ds2_varnames <- ds2_0
    colnames(ds2_varnames) <-
      util_map_labels(colnames(ds2_varnames),
                      meta_data = meta_data,
                      to = "ORIGINAL_VAR_NAMES",
                      from = label_col)
    ds2 <- cbind(ds2, ds2_varnames)
  }
  if ("ORIGINAL_LABEL" %in% colnames(meta_data)) {
    ds2_varnames <- ds2_0
    colnames(ds2_varnames) <-
      util_map_labels(colnames(ds2_varnames),
                      meta_data = meta_data,
                      to = "ORIGINAL_LABEL",
                      from = label_col)
    ds2 <- cbind(ds2, ds2_varnames)
  }
  # TODO: replace missing value codes only for
  # intersect(colnames(ds2), all.vars(rule))
  # to reduce the number of warning messages (i.e., do not throw a warning for
  # variables which are not used in the rule)
  enclos <- new.env(parent = redcap_rule_env)
  enclos[["[meta_data]"]] <- meta_data # something, redcap cannot refert to, ensured by its strange name
  enclos[["[label_col]"]] <- label_col # something, redcap cannot refert to, ensured by its strange name
  # add metadata here
  res <- eval(expr = rule,
              envir = ds2,
              enclos = enclos)
  if (is.character(res)) {
    res <- readr::parse_guess(res)
  }
  res
}

#' @title The `dataquieR` package about Data Quality in
#'        Epidemiological Research
#' @name dataquieR
#' @aliases dataquieR-package dataquieR-options
#' @docType package
#' @description
#' For a quick start please read [dq_report2] and maybe the vignettes
#' or the package's [website](https://dataquality.qihs.uni-greifswald.de/).
#'
#' # Options
#'
#' This package features the following `options()`:
#'
#' `r paste0(' - [', grep("^dataquieR\\.", names(options()), value = TRUE), ']', collapse = "\n")`
#'
#' @family options
#' @references
#' - [JOSS 2024](https://joss.theoj.org/papers/10.21105/joss.06581)
#' - [JOSS 2021](https://joss.theoj.org/papers/10.21105/joss.03093)
#' - [DQ_OBS data quality framework](https://www.researchsquare.com/article/rs-119457/v1)
#'
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

#' Utility function to calculate Mahalanobis distances for a group of variables
#'
#' @param x [data.frame] containing the variables to use to
#'                        calculate the Mahalanobis distance
#' @param rv a vector containing the names of the variables
#' @param check_label_rv [character] the label of the check list used to define
#'                                    the group of variables in the cross-item metadata
#'                                    or "variable_group" if the list of variables
#'                                    is provided as an argument
#'
#' @return a list with:
#'   - `x_with_MD`: [data.frame] containing the original variables with the
#'                                addition of a new column containing the
#'                                Mahalanobis distance (only for complete cases)
#'   - `df`: [numeric] the degree of freedom (no.variables)
#'
#' @family util_functions
#' @concept outlier
#' @keywords internal
#'
#' @noRd
util_generate_mahalanobis_dist <- function(x,
                                           rv,
                                           check_label_rv){
  x$row_numbers <- c(1:nrow(x))
  n_prior <- dim(x)[1]
  ds1completecases <- x[rowSums(is.na(x[, rv, drop = FALSE])) == 0, ,
                                drop = FALSE]


  n_post <- dim(ds1completecases)[1]

  if (n_post == 0) {
    util_error("No observational unit with complete cases for all variables %s. Aborting.",
               paste0(sQuote(rv), collapse = ", "),
               applicability_problem = FALSE)
  }

  if (n_post < n_prior) {
    util_message(paste0(
      "Due to missing values",
      " N=", n_prior - n_post,
      " observational units were excluded."
    ), applicability_problem = FALSE)
  }

  # Mahalanobis ----------------------------------------------------------------
  # no. variables used to calculate the MD
  degree_freedom <- ncol(ds1completecases[, rv]) #NOTE: this is the no. columns or no. variables

  # Estimate covariance of response variables
  Sx <- cov(ds1completecases[, rv, drop = FALSE])

  # Calculate squared Mahalanobis distance
  ds1completecases[[paste0("MD_", check_label_rv)]] <-
    mahalanobis(ds1completecases[, rv],
                colMeans(ds1completecases[, rv, drop = FALSE]),
                Sx)

  ds1completecases <- merge(x,
            ds1completecases,
            by = intersect(colnames(x), colnames(ds1completecases)),
            all.x = TRUE)

  ds1completecases <- ds1completecases[, names(ds1completecases) != "row_numbers"]


  return(list(x_with_MD = ds1completecases,
              df = degree_freedom))

}

#' Extract named capture groups from String
#'
#' created with support from `ChatGPT` in April 2025
#'
#' @param pattern [character] regular expression with named capture groups,
#'                            e.g. `""^\\[(?<row>\\d+), +(?<col>\\d+)\\]:.*""`
#' @param text_vector [character] haystack
#'
#' @returns [data.frame] with matches
#'
#' @seealso [grep()]
#'
#' @noRd
util_extract_named_groups <- function(pattern, text_vector) {
  m <- regexpr(pattern, text_vector, perl = TRUE)

  starts <- attr(m, "capture.start")
  lengths <- attr(m, "capture.length")
  names <- attr(m, "capture.names")

  if (is.null(starts)) {
    # Kein Match oder keine Gruppen
    return(data.frame())
  }

  # Sicherstellen, dass es sich um Matrizen handelt
  if (is.null(dim(starts))) {
    starts <- matrix(starts, nrow = 1)
    lengths <- matrix(lengths, nrow = 1)
  }

  # Ergebnisliste: je Element ein benannter Vektor der Gruppenwerte
  result <- lapply(seq_along(text_vector), function(i) {
    s_row <- starts[i, ]
    l_row <- lengths[i, ]
    grp_vals <- mapply(function(start, len) {
      if (start != -1) substr(text_vector[i], start, start + len - 1) else NA_character_
    }, s_row, l_row, SIMPLIFY = TRUE)
    setNames(as.list(grp_vals), names)
  })

  # In Data Frame umwandeln
  result_df <- do.call(rbind.data.frame, result)
  rownames(result_df) <- NULL
  return(result_df)}

#' Get description for an indicator function
#'
#' @param fname the function name
#'
#' @return the description
#'
#' @noRd
util_function_description <- function(fname) {
  # def <-
  #   util_map_labels(fname,
  #                   util_get_concept_info("implementations"),
  #                   to = "Definition",
  #                   from = "function_R",
  #                   ifnotfound = "")
  # if (gsub("[^a-z]", "", tolower(trimws(def))) == "na") {
  #   def <- ""
  # }
  # expl <-
  #   util_map_labels(fname,
  #                   util_get_concept_info("implementations"),
  #                   to = "Explanation",
  #                   from = "function_R",
  #                   ifnotfound = "")
  # if (gsub("[^a-z]", "", tolower(trimws(expl))) == "na") {
  #   expl <- ""
  # }
  # rel <-
  #   util_map_labels(fname,
  #                   util_get_concept_info("implementations"),
  #                   to = "Relevance",
  #                   from = "function_R",
  #                   ifnotfound = "")
  # if (gsub("[^a-z]", "", tolower(trimws(rel))) == "na") {
  #   rel <- ""
  # }
  # guidance <-
  #   util_map_labels(fname,
  #                   util_get_concept_info("implementations"),
  #                   to = "Guidance",
  #                   from = "function_R",
  #                   ifnotfound = "")
  # if (gsub("[^a-z]", "", tolower(trimws(guidance))) == "na") {
  #   guidance <- ""
  # }
  int <-
    util_map_labels(fname,
                    util_get_concept_info("implementations"),
                    to = "Interpretation",
                    from = "function_R",
                    ifnotfound = "")
  if (is.na(int) || gsub("[^a-z]", "", tolower(trimws(int))) == "na") {
    int <- ""
  }
  desc <- paste(int, sep = "\n")
  if (util_empty(desc)) {
    desc <-
      .manual$descriptions;
    if (length(desc) != 1) desc <- sprintf(
      "<i>No description found for <tt>%s</tt></i>",
      dQuote(fname));
  } else {
    if (suppressWarnings(util_ensure_suggested("markdown", err = FALSE))) {
      int <-
        markdown::markdownToHTML(int, fragment.only = TRUE)
    }
  }
  desc
}

# options("testthat.default_reporter" = testthat::RStudioReporter)

library(dataquieR)

.on_cran <- function() {
  env <- Sys.getenv("NOT_CRAN")
  if (identical(env, "")) {
    !interactive()
  } else {
    !isTRUE(as.logical(env))
  }
}

if (.on_cran()) {
  Sys.setenv(TESTTHAT_CPUS = "1")
}

testthat::test_check("dataquieR")

#' Produce an error message with a useful short stack trace. Then it stops the
#' execution.
#'
#' @param m error message or a [condition]
#' @param ... arguments for [sprintf] on m, if m is a character
#' @param applicability_problem [logical] `TRUE`, if an applicability issue,
#'                                        that is, the information for
#'                                        computation is missing
#'                                        (that is, an error that indicates
#'                                        missing metadata) or an error because
#'                                        the requirements of
#'                                        the stopped function were not met,
#'                                        e.g., a `barplot` was called for
#'                                        metric data. We can have logical
#'                                        or empirical applicability problems.
#'                                        empirical is the default, if the
#'                                        argument
#'                                        `intrinsic_applicability_problem` is
#'                                        left unset or set to `FALSE`.
#' @param intrinsic_applicability_problem [logical] `TRUE`, if this is a
#'                                        logical applicability issue, that is,
#'                                        the computation makes no sense
#'                                        (for example, an error of unsuitable
#'                                        `resp_vars`). Intrinsic/logical
#'                                        applicability problems are also
#'                                        applicability problems. Non-logical
#'                                        applicability problems are called
#'                                        empirical applicability problems.
#' @param integrity_indicator [character] the message is an integrity problem,
#'                                        here is the indicator abbreviation..
#' @param level [integer] level of the error message (defaults to 0). Higher
#'                        levels are more severe.
#' @param once_id [character] If specified, suppress further conditions
#'                 with the same ID during this R session.
#' @param immediate [logical] not used.
#' @param additional_classes [character] additional classes the
#'                                       thrown condition object should inherit
#'                                       from, first.
#'
#' @return nothing, its purpose is to stop.
#'
#' @family condition_functions
#' @concept process
#' @noRd
util_error <- util_condition_constructor_factory("error")

#' Checks user-defined contradictions in study data
#'
#' @description
#' This approach considers a contradiction if impossible combinations of data
#' are observed in one participant. For example, if age of a participant is
#' recorded repeatedly the value of age is (unfortunately) not able to decline.
#' Most cases of contradictions rest on comparison of two variables.
#'
#' Important to note, each value that is used for comparison may represent a
#' possible characteristic but the combination of these two values is considered
#' to be impossible. The approach does not consider implausible or inadmissible
#' values.
#'
#' [Indicator]
#'
#' @details
#' ### Algorithm of this implementation:
#'
#'  - Remove missing codes from the study data (if defined in the metadata)
#'  - Remove measurements deviating from limits defined in the metadata
#'  - Assign label to levels of categorical variables (if applicable)
#'  - Apply contradiction checks (given as `REDCap`-like rules in a separate
#'    metadata table)
#'  - Identification of measurements fulfilling contradiction rules. Therefore
#'    two output data frames are generated:
#'    - on the level of observation to flag each contradictory value
#'      combination, and
#'    - a summary table for each contradiction check.
#'  - A summary plot illustrating the number of contradictions is generated.
#'
#' List function.
#'
#' @inheritParams .template_function_indicator
#'
#' @param threshold_value [numeric] from=0 to=100. a numerical value
#'                                                 ranging from 0-100
#' @param meta_data_cross_item [data.frame] contradiction rules table.  Table
#'                                 defining contradictions. See
#'                                 [online documentation](https://dataquality.qihs.uni-greifswald.de/VIN_Cross_Item_Level_Metadata.html)
#'                                 for its required structure.
#' @param summarize_categories [logical] Needs a column `CONTRADICTION_TYPE` in
#'                             the `meta_data_cross_item`.
#'                             If set, a summary output is generated for the
#'                             defined categories plus one plot per
#'                             category. TODO: Not yet controllable by metadata.
#' @param use_value_labels [logical] Deprecated in favor of [DATA_PREPARATION].
#'                             If set to `TRUE`, labels can be used in the
#'                             `REDCap` syntax to specify contraction checks for
#'                             categorical variables. If set to `FALSE`,
#'                             contractions have to be specified using the coded
#'                             values. In case that this argument is not set in
#'                             the function call, it will be set to `TRUE` if
#'                             the metadata contains a column `VALUE_LABELS`
#'                             which is not empty.
#' @param cross_item_level [data.frame] alias for `meta_data_cross_item`
#' @param `cross-item_level` [data.frame] alias for `meta_data_cross_item`
#'
#' @return
#' If `summarize_categories` is `FALSE`:
#' A [list] with:
#'   - `FlaggedStudyData`: The first output of the contradiction function is a
#'                         data frame of similar dimension regarding the number
#'                         of observations in the study data. In addition, for
#'                         each applied check on the variables an additional
#'                         column is added which flags observations with a
#'                         contradiction given the applied check.
#'   - `VariableGroupData`: The second output summarizes this information
#'                     into one
#'                     data frame. This output can be used to provide an
#'                     executive overview on the amount of contradictions.
#'   - `VariableGroupTable`: A subset of `VariableGroupData` used within the
#'                           pipeline.
#'   - `SummaryPlot`: The third output visualizes summarized information
#'                    of `SummaryData`.
#'
#' If `summarize_categories` is `TRUE`, other objects are returned:
#' A list with one element `Other`, a list with the following entries:
#' One per category named by that category (e.g. "Empirical") containing a
#' result for contradiction checks within that category only. Additionally, in the
#' slot `all_checks`, a result as it would have been returned with
#' `summarize_categories` set to `FALSE`. Finally, in
#' the top-level list, a slot `SummaryData` is
#' returned containing sums per Category and an according [ggplot2::ggplot] in
#' `SummaryPlot`.
#'
#' @export
#'
#' @importFrom ggplot2 ggplot geom_bar scale_fill_manual theme_minimal scale_y_continuous geom_hline coord_flip theme aes geom_text xlab scale_x_continuous sec_axis
#' @importFrom stats setNames
#' @seealso
#' [Online Documentation for the function](
#' https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions_redcap.html
#' )
#' [meta_data_cross]
#' [Online Documentation for the required cross-item-level metadata](
#' https://dataquality.qihs.uni-greifswald.de/Cross_Item_Level_Metadata.html
#' )
con_contradictions_redcap <- function(study_data,
                                      item_level = "item_level",
                                      label_col, threshold_value,
                                      meta_data_cross_item = "cross-item_level",
                                      use_value_labels,
                                      summarize_categories = FALSE,
                                      # flip_mode = "flip", # TODO: Fix noflip graph
                                      meta_data = item_level,
                                      cross_item_level,
                                      `cross-item_level`,
                                      meta_data_v2
) {

  # preps ----------------------------------------------------------------------
  util_maybe_load_meta_data_v2()

  util_ck_arg_aliases()

  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = FALSE,
                          .replace_missings = FALSE) # replacements are performed later

  if (!missing(use_value_labels)) {
    lifecycle::deprecate_stop(when = "2.1.0",
                              what = "con_contradictions_redcap(use_value_labels)",
                              details =
                                "Please use DATA_PREPARATION in meta_data_cross_item now."
    )
  }

  # table of specified contradictions
  util_expect_data_frame(meta_data_cross_item, list(
    CONTRADICTION_TERM = is.character
  ))
  if (!CHECK_LABEL %in% colnames(meta_data_cross_item)) {
    if (nrow(meta_data_cross_item) > 0) {
      meta_data_cross_item[[CHECK_LABEL]] <-
        paste0("Check #", seq_len(nrow(meta_data_cross_item)))
    } else {
      meta_data_cross_item[[CHECK_LABEL]] <- character(0)
    }

  }
  util_expect_data_frame(meta_data_cross_item, list(
    CONTRADICTION_TERM = is.character,
    CHECK_LABEL = is.character
  ))

  meta_data_cross_item <-
    meta_data_cross_item[!util_empty(
      meta_data_cross_item[[CONTRADICTION_TERM]]), , FALSE]

  meta_data_cross_item <- util_normalize_cross_item(
    meta_data = meta_data,
    meta_data_cross_item = meta_data_cross_item,
    label_col = label_col
  )

  # There might be rows without contradiction rules (NAs), which should be removed first.
  if (any(is.na(meta_data_cross_item[[CONTRADICTION_TERM]]))) {
    meta_data_cross_item <- meta_data_cross_item[-which(is.na(meta_data_cross_item[[CONTRADICTION_TERM]])), ]
  }

  if (missing(threshold_value)) {
    threshold_value <- NA_real_
    if (!.called_in_pipeline)
      util_message("No %s has been set, will use default %d",
                   dQuote("threshold_value"), threshold_value,
                   applicability_problem = TRUE)
  } else {
    util_expect_scalar(threshold_value, allow_na = TRUE,
                       check_type = function(x) {
                         if (is.na(x)) {
                           return(TRUE)
                         }
                         is.numeric(x) && !is.na(x) && x >= 0 && x <= 100
                       },
                       error_message = sprintf("%s must be a number between %d and %d",
                                               sQuote("threshold_value"),
                                               0,
                                               100)
    )
    threshold_value <- as.numeric(threshold_value)
  }

  util_expect_scalar(summarize_categories, check_type = is.logical)

  # parse redcap rules to obtain interpretable contradiction checks
  compiled_rules <- lapply(setNames(nm = meta_data_cross_item[[CONTRADICTION_TERM]]),
                           util_parse_redcap_rule)

  # colors
  cols <- c("0" = "#2166AC", "1" = "#B2182B")

  # summarize contradictions per category given in CONTRADICTION_TYPE -------------------------------------
  if (summarize_categories) {
    if (!(CONTRADICTION_TYPE %in% colnames(meta_data_cross_item))) {
      util_error(c(
        "Cannot summerize categories of contradictions,",
        "because these are not defined in the meta_data_cross_item",
        "as column %s."),
        sQuote(CONTRADICTION_TYPE),
        applicability_problem = TRUE)
    }

    split_tags <- lapply(strsplit(meta_data_cross_item[[CONTRADICTION_TYPE]], SPLIT_CHAR, fixed = TRUE), trimws)
    tags <- sort(unique(unlist(split_tags)))
    tags <- setNames(nm = tags)
    tags_ext <- tags
    tags_ext[["all_checks"]] <- NA

    result <- lapply(tags_ext, function(atag) {
      # generate one output per category (stratified)
      if (is.na(atag)) {
        new_ct <- meta_data_cross_item[, , #-which(colnames(meta_data_cross_item) == CONTRADICTION_TYPE),
                                       drop = FALSE]
      } else {
        contains_tag <- function(x, tg) {
          any(x == tg, na.rm = TRUE)
        }
        rows_matching_tag <- vapply(split_tags, contains_tag, tg = atag,
                                    logical(1))
        new_ct <- meta_data_cross_item[rows_matching_tag, , #-which(colnames(meta_data_cross_item) == CONTRADICTION_TYPE),
                                       drop = FALSE]
      }
      # recursive call of the function only for the contradiction checks of the currently selected category in "atag"
      r <- try(con_contradictions_redcap(
        study_data = study_data,
        meta_data = meta_data, label_col = label_col,
        threshold_value = threshold_value, meta_data_cross_item = new_ct,
        summarize_categories = FALSE
        # , flip_mode = flip_mode TODO
      ), silent = TRUE)
      if (inherits(r, "try-error")) {
        list(FlaggedStudyData = data.frame())
      } else {
        r
      }
    })

    # summarize the outputs of the recursive calls
    rx <- lapply(tags_ext, function(atag) {
      if (is.na(atag)) {
        round(sum(rowSums(result[["all_checks"]]$FlaggedStudyData[, -1, drop = FALSE],
                          na.rm = TRUE) > 0) /
                nrow(result[["all_checks"]]$FlaggedStudyData) * 100, digits = 2)
      } else {
        round(sum(rowSums(result[[atag]]$FlaggedStudyData[, -1, drop = FALSE],
                          na.rm = TRUE) > 0) /
                nrow(result[[atag]]$FlaggedStudyData) * 100, digits = 2)
      }
    })
    rx_num <- lapply(tags_ext, function(atag) {
      if (is.na(atag)) {
        sum(rowSums(result[["all_checks"]]$FlaggedStudyData[, -1, drop = FALSE],
                    na.rm = TRUE) > 0)
      } else {
        sum(rowSums(result[[atag]]$FlaggedStudyData[, -1, drop = FALSE],
                    na.rm = TRUE) > 0)
      }
    })
    rx <- data.frame(
      CONTRADICTION_TYPE = names(rx),
      PCT_con_con = unlist(rx),
      NUM_con_con = unlist(rx_num),
      GRADING = ordered(ifelse(unlist(rx) > threshold_value, 1, 0))
    )
    if ("LOGICAL" %in% rx[[CONTRADICTION_TYPE]]) {
      rx$PCT_con_con_contc <- rep(NA_real_, nrow(rx))
      rx$PCT_con_con_contc[
        rx[[CONTRADICTION_TYPE]] %in% c("LOGICAL")
      ] <- rx$PCT_con_con[rx[[CONTRADICTION_TYPE]]
                          %in% c("LOGICAL")]
      rx$NUM_con_con_contc <- rep(NA_integer_, nrow(rx))
      rx$NUM_con_con_contc[
        rx[[CONTRADICTION_TYPE]] %in% c("LOGICAL")
      ] <- rx$PCT_con_con[rx[[CONTRADICTION_TYPE]]
                          %in% c("LOGICAL")]
    }
    if ("EMPIRICAL" %in% rx[[CONTRADICTION_TYPE]]) {
      rx$PCT_con_con_contu <- rep(NA_real_, nrow(rx))
      rx$PCT_con_con_contu[
        rx[[CONTRADICTION_TYPE]] %in% c("EMPIRICAL")
      ] <- rx$PCT_con_con[rx[[CONTRADICTION_TYPE]]
                          %in% c("EMPIRICAL")]
      rx$NUM_con_con_contu <- rep(NA_integer_, nrow(rx))
      rx$NUM_con_con_contu[
        rx[[CONTRADICTION_TYPE]] %in% c("EMPIRICAL")
      ] <- rx$PCT_con_con[rx[[CONTRADICTION_TYPE]]
                          %in% c("EMPIRICAL")]
    }

    result$OtherTable <- rx
    # Create Data Slot
    result$OtherData <- rx
    result$OtherData$PCT_con_con_contc <- NULL
    result$OtherData$PCT_con_con_contu <- NULL
    result$OtherData$GRADING <- NULL
    result$OtherData <- util_make_data_slot_from_table_slot(result$OtherData)

    e <- new.env(parent = environment(con_contradictions_redcap))
    e$rx <- rx
    e$meta_data <- meta_data

    cls_rx <- rlang::new_quosure(quote(
      util_make_cls_binding(rx, meta_data = meta_data)), e)

    # Plot for summarized contradiction checks -----------------------------------------------------
    # TODO: work on the changes here and below with position = "top"
    p <- util_create_lean_ggplot(
      ggplot(rx, aes(x = seq_along(CONTRADICTION_TYPE), y = PCT_con_con,
                        fill = (if (!is.na(threshold_value)) {
                          as.ordered(GRADING)
                        } else {
                          !!cls_rx
                        }))) +
      geom_bar(stat = "identity") +
      theme_minimal() +
      scale_y_continuous(name = "(%)",
                         limits = (c(0, max(1.2 * suppressWarnings(
                           max(rx$PCT_con_con)),
                           threshold_value)))#,
                         #expand = expansion(add = 0.5, mult = c(0, 0.1))
                         ) +
      scale_x_continuous(breaks = seq_len(nrow(rx)),
                         labels = rx[[CONTRADICTION_TYPE]]#,
                         #position = "top",
                         #trans = "reverse"
                         ) +
      # xlab("Category of applied contradiction checks") +
      xlab("") +
      (if (!is.na(threshold_value)) {
        scale_fill_manual(values = cols, name = " ", guide = "none")
      } else {
        util_scale_fill_dataquieR()
      }) +
      (if (!is.na(threshold_value)) {
        geom_hline(yintercept = threshold_value,
                   color = "red", linetype = 2)
      }) +
      geom_text(label = util_paste0_with_na(" ", rx$PCT_con_con, "%"),
                hjust = 0, vjust = 0.5, size = 3.5) +
      coord_flip() + # TODO
      theme(axis.text.x = element_text(size = 10),
            axis.text.y.right = element_text(size = 10),
            axis.text.y.left = element_text(size = 10), #element_blank(),
            legend.title = element_blank()),
      rx = rx,
      cls_rx = cls_rx,
      threshold_value = threshold_value,
      cols = cols
    )

    # p <- p + util_coord_flip(p = p) # TODO: estimate w and h, if p is not using discrete axes util_lazy_add_coord(p, fli)

    # https://stackoverflow.com/a/51795017
    bp <- ggplot_build(p)
    w <- 2 * length(bp$layout$panel_params[[1]]$x$get_labels())
    if (w == 0) {
      w <- 10
    }
    w <- w + 2 +
      max(nchar(bp$layout$panel_params[[1]]$y$get_labels()),
          na.rm = TRUE)
    h <- 2 * length(bp$layout$panel_params[[1]]$y$get_labels())
    if (h == 0) {
      h <- 10
    }
    h <- h + 15

    p <- util_set_size(p, width_em = w, height_em = h)

    result$SummaryPlot <- p

    to_Other <- setdiff(names(result), c("SummaryData",
                                         "OtherTable",
                                         "OtherData",
                                         "SummaryPlot"))

    Other <- result[to_Other]

    result[to_Other] <- NULL

    result$Other <- Other


    return(util_attach_attr(
      result,
      as_plotly = "util_as_plotly_con_contradictions_redcap",
      sizing_hints = list(
        figure_type_id = "bar_chart",
        rotated = TRUE,
        number_of_bars = nrow(p$data),
        range = max(p$data$PCT_con_con) - min(p$data$PCT_con_con)
      )
    ))

  } else {
    # run contradiction checks without summarizing -------------------------------------------------------
    # apply contradiction checks -------------------------------------------------------------------------
    rule_match <- mapply(
      SIMPLIFY = FALSE,
      rule = compiled_rules,
      prep = util_parse_assignments(
        meta_data_cross_item[[DATA_PREPARATION]], multi_variate_text = TRUE),
      FUN = function(rule, prep) {
        prep <- as.character(names(prep))
        use_value_labels <- ("LABEL"  # meta_data_cross_item has been normalized, already
                             %in% prep)
        if ("MISSING_NA" %in% prep) {
          replace_missing_by <- "NA"
        } else if ("MISSING_LABEL" %in% prep) {
          replace_missing_by <- "LABEL"
        } else if ("MISSING_INTERPRET" %in% prep) {
          replace_missing_by <- "INTERPRET"
        } else {
          replace_missing_by <- ""
        }
        replace_limits <- ("LIMITS"
                           %in% prep)
        if (is.list(rule) && !length(rule) && is.null(attr(rule, "class"))) {
          r <- try(util_error("Parser error"), silent = TRUE)
        } else {
          r <- try(util_eval_rule(rule = rule,
                                  ds1 = ds1,
                                  meta_data = meta_data,
                                  use_value_labels = use_value_labels,
                                  replace_missing_by = replace_missing_by,
                                  replace_limits = replace_limits
          ),
          silent = TRUE)
        }
        if (inherits(r, "try-error")) {
          rule_src <- attr(rule, "src")
          if (length(rule_src) == 0) {
            rule_src <- util_deparse1(rule)
          }
          util_warning("Could not evaluate rule %s: %s",
                       dQuote(rule_src),
                       conditionMessage(attr(r, "condition")))
          r <- "error"
        }
        r
      })

    rule_errors <- vapply(rule_match, identical, "error",
                          FUN.VALUE = logical(1))
    rule_match <- lapply(rule_match, as.logical)

    list_element_length <- vapply(rule_match, length, FUN.VALUE = integer(1))
    if (any(list_element_length == 1)) {
      # not all columns of same length, fix this for as.data.frame
      rule_match[list_element_length == 1] <- lapply(
        rule_match[list_element_length == 1],
        function(to_recycle) {
          rep(to_recycle, nrow(ds1))
        })
    }

    if (length(unique(vapply(rule_match, length, FUN.VALUE = integer(1)))) > 1) {
      util_error(c("Internal error: unexpected inhomogeneous length of rules result.",
                   "This is an internal error, please excuse and contact the dataquieR developers."))
    }

    if (length(unique(vapply(rule_match, length, FUN.VALUE = integer(1)))) == 0) {
      summary_df1 <- data.frame(Obs = seq_len(nrow(ds1)))
    } else {
      summary_df1 <- cbind(data.frame(Obs = seq_len(nrow(ds1))),
                           as.data.frame(rule_match))
    }

    colnames(summary_df1)[-1] <- paste0("flag_con",
                                        formatC(seq_len(nrow(meta_data_cross_item)),
                                                width = nchar(nrow(meta_data_cross_item)),
                                                format = "d",
                                                flag = "0"))

    summary_df2 <- meta_data_cross_item

    summary_df2$NUM_con_con <- as.numeric(lapply(rule_match, sum, na.rm = TRUE))
    summary_df2$NUM_con_con[rule_errors] <- rep(NA_integer_, sum(rule_errors))

    summary_df2$PCT_con_con <- round(summary_df2$NUM_con_con / nrow(ds1) * 100,
                                     digits = 2)


    if (CONTRADICTION_TYPE %in% colnames(summary_df2)) {
      summary_df2[["CONTRADICTION_TYPE"]] <-
        trimws(toupper(summary_df2[["CONTRADICTION_TYPE"]]))
      # logical
      summary_df2$NUM_con_con_contc <- rep(NA_integer_, nrow(summary_df2))
      summary_df2$NUM_con_con_contc[
        summary_df2[["CONTRADICTION_TYPE"]] %in% c("LOGICAL")
      ] <- summary_df2$NUM_con_con[summary_df2[["CONTRADICTION_TYPE"]]
                                   %in% c("LOGICAL")]
      summary_df2$PCT_con_con_contc <- rep(NA_real_, nrow(summary_df2))
      summary_df2$PCT_con_con_contc[
        summary_df2[["CONTRADICTION_TYPE"]] %in% c("LOGICAL")
      ] <- summary_df2$PCT_con_con[summary_df2[["CONTRADICTION_TYPE"]]
                                   %in% c("LOGICAL")]


      summary_df2$NUM_con_con_contu <- rep(NA_integer_, nrow(summary_df2))
      summary_df2$NUM_con_con_contu[
        summary_df2[["CONTRADICTION_TYPE"]] %in% c("EMPIRICAL")
      ] <- summary_df2$NUM_con_con[summary_df2[["CONTRADICTION_TYPE"]]
                                   %in% c("EMPIRICAL")]
      summary_df2$PCT_con_con_contu <- rep(NA_real_, nrow(summary_df2))
      summary_df2$PCT_con_con_contu[
        summary_df2[[CONTRADICTION_TYPE]] %in% c("EMPIRICAL")
      ] <- summary_df2$PCT_con_con[summary_df2[[CONTRADICTION_TYPE]]
                                   %in% c("EMPIRICAL")]

    } else {
      summary_df2[["CONTRADICTION_TYPE"]] <- rep(NA_character_, nrow(summary_df2))
    }

    summary_df2$GRADING <- ifelse(summary_df2$PCT_con_con > threshold_value,
                                  1, 0)

    summary_df2 <- summary_df2[order(summary_df2[["PCT_con_con"]],
                                     decreasing = TRUE), ]
    summary_df2 <- summary_df2[order(summary_df2[[CONTRADICTION_TYPE]],
                                     decreasing = TRUE), ]

    summary_df2_for_return <- summary_df2

    summary_df2 <- summary_df2[rev(seq_len(nrow(summary_df2))), ]


    e <- new.env(parent = environment(con_contradictions_redcap))
    e$summary_df2 <- summary_df2
    e$meta_data <- meta_data

    cls_summary_df2 <- rlang::new_quosure(quote(
      util_make_cls_binding(summary_df2, meta_data = meta_data)), e)

    ctype_pal <- setNames( # does not really depend on grading formats
      scales::pal_hue(h = c(0,360))(n =
                                      length(unique(summary_df2[[CONTRADICTION_TYPE]]
                                      ))
      ), nm = unique(summary_df2[[CONTRADICTION_TYPE]]))
    ctype_pal[["LOGICAL"]] <- getOption("dataquieR.col_con_con_logical",
                                        dataquieR.col_con_con_logical_default)
    ctype_pal[["EMPIRICAL"]] <- getOption("dataquieR.col_con_con_empirical",
                                        dataquieR.col_con_con_empirical_default)
    # Plot for all contradiction checks --------------------------------------------------------
    p <- util_create_lean_ggplot(
      ggplot(summary_df2, aes(x = seq_along(CHECK_ID),
                                 y = PCT_con_con,
                                 color = CONTRADICTION_TYPE,
                                 fill = (if (!is.na(threshold_value)) {
                                   as.ordered(GRADING)
                                 } else {
                                   !!cls_summary_df2
                                 }))) +
      geom_bar(stat = "identity") +
      theme_minimal() +
      # xlab("Applied contradiction checks") +
      xlab("") +
      scale_y_continuous(name = "(%)",
                         limits = c(0, max(1.2 * suppressWarnings(
                           max(summary_df2$PCT_con_con)),
                           threshold_value)),
                         expand = expansion(
                          # add = 0.5,
                           mult = c(0, 0.1))
                         ) +
      scale_x_continuous(breaks = seq_len(nrow(summary_df2)),
                         labels = summary_df2$CHECK_LABEL#,
                       #  position = "top"
                         #,
                         #trans = "reverse"
                         ) +
      ggplot2::scale_color_discrete(type = ctype_pal) +
      (if (!is.na(threshold_value)) {
        scale_fill_manual(values = cols, name = " ", guide = "none")
      } else {
        util_scale_fill_dataquieR()
      }) +
      (if (!is.na(threshold_value)) {
        geom_hline(yintercept = threshold_value,
                   color = "red", linetype = 2)
      }) +
      geom_text(label = util_paste0_with_na(" ", summary_df2$PCT_con_con, "%"),
                hjust = 0, vjust = 0.5, size = 3.5) +
      coord_flip() + # TODO
      theme(axis.text.x = element_text(size = 10),
            axis.text.y.right = element_text(size = 10),
            axis.text.y.left = element_text(size = 10), #element_blank(),
            legend.title = element_blank()),
      summary_df2 = summary_df2,
      cls_summary_df2 = cls_summary_df2,
      ctype_pal = ctype_pal,
      threshold_value = threshold_value,
      cols = cols
    )

    if (!prod(dim(summary_df2))) {
      util_error("No contradiction check defined",
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    }

    # create Data Slot
    summary_df3 <- summary_df2_for_return[, intersect(c(VARIABLE_LIST, CHECK_LABEL,
                                             "NUM_con_con",
                                             "PCT_con_con",
                                             CONTRADICTION_TYPE,
                                             "GRADING"), colnames(summary_df2_for_return))]
    summary_df3 <- util_make_data_slot_from_table_slot(summary_df3)

    #add sizing information
    obj1 <- ggplot2::ggplot_build(p)
    number_of_bars <- nrow(util_gg_get(obj1, "data")[[1]])
    range <- max(c(
      util_rbind(data_frames_list = util_gg_get(obj1, "data"))$ymax,
      util_rbind(data_frames_list = util_gg_get(obj1, "data"))$yintercept
    ), na.rm = TRUE) -
      min(c(
        util_rbind(data_frames_list = util_gg_get(obj1, "data"))$ymax,
        util_rbind(data_frames_list = util_gg_get(obj1, "data"))$yintercept
      ), na.rm = TRUE)

    no_char_vars = max(c(0, nchar(summary_df2$CHECK_LABEL)))
    no_char_numbers = max(c(0, nchar(round(util_gg_get(obj1, "data")[[1]]$ymax,digits = 0)),
                          na.rm = TRUE))

    summary_df2$VARIABLE_LIST_ORDER <- NULL # used only internally for
                                            # computed variables.
    summary_df2_for_return$VARIABLE_LIST_ORDER <-
                                       NULL # used only internally for
                                            # computed variables.
    # Output
    return(util_attach_attr(list(
      FlaggedStudyData = summary_df1,
      VariableGroupTable = summary_df2_for_return,
      VariableGroupData = summary_df3,
      SummaryPlot = p
    ),
    as_plotly = "util_as_plotly_con_contradictions_redcap",
    sizing_hints = list(
      figure_type_id = "bar_chart",
      rotated = TRUE,
      number_of_bars = number_of_bars,
      range = range,
      no_char_vars = no_char_vars,
      no_char_numbers = no_char_numbers
    )
    ))
  }

  # Never called, just for documentation.
  return(list( # nocov start
    FlaggedStudyData = summary_df1,
    VariableGroupTable = summary_df2,
    VariableGroupData = summary_df3,
    SummaryPlot = p
  )) # nocov end
}

util_reshape_for_cls_binding <- local({
  cache_env <- new.env(parent = emptyenv())
  function(rx, idvars) {
    key <- rlang::hash(list(rx, idvars))
    if (exists(key, envir = cache_env, inherits = FALSE)) {
      return(get(key, envir = cache_env, inherits = FALSE))
    }
    reshaped <-     stats::reshape(data = rx,
                                           timevar = "indicator_metric",
                           idvar = idvars,
                                           times = c("PCT_con_con",
                                                     "PCT_con_con_contc",
                                                     "PCT_con_con_contu",
                                                     "NUM_con_con",
                                                     "NUM_con_con_contc",
                                                     "NUM_con_con_contu"),
                                           varying = list(c("PCT_con_con",
                                                            "PCT_con_con_contc",
                                                            "PCT_con_con_contu",
                                                            "NUM_con_con",
                                                            "NUM_con_con_contc",
                                                            "NUM_con_con_contu")),
                                           v.names = "values_raw",
                                           direction = "long")
    assign(key, reshaped, envir = cache_env)
    reshaped
  }
})

util_make_cls_binding <- local({
  gg_cache_env <- new.env(parent = emptyenv())
  function(rx, meta_data) {

  rs <- util_get_rule_sets()
  rf <- util_get_ruleset_formats()

  key <- paste0("KEY_", rlang::hash(list(rs, rf, rx, meta_data)))
  if (exists(key, envir = gg_cache_env)) {
    return((get(key, envir = gg_cache_env)))
  }

  # FIXME: Do not expect all metrics
  grading_labs <- util_get_labels_grading_class()
  grading_colors <- util_get_colors()
  grading_colors["NA"] <- "#888888"
  grading_rules <- rs[["0"]]

  if (is.data.frame(grading_rules) &&
      is.character(grading_labs) &&
      length(grading_labs) > 0) {
    rx$.orig_order <- seq_len(nrow(rx))

    idvars <- intersect(c(VARIABLE_LIST, CHECK_LABEL,
                          CHECK_ID, ".orig_order"),
                        colnames(rx))
    summ <- util_reshape_for_cls_binding(rx, idvars)
    summ$call_names <- paste0("con_contradictions_redcap") # TODO: after exgtension of metadat model, handle a grading rule column in cross-item paste0("con_contradictions_redcap.", summ$CHECK_ID)
    summ$function_name <- paste0("con_contradictions_redcap")
    summ <- util_metrics_to_classes(summ, meta_data, entity = "CROSS_ITEM")
    cls <- summ[, c(idvars, "class"), drop = FALSE]
    cls <- unsplit(lapply(
      split(cls, cls[, idvars]), FUN =
        function(x) {
          if (any(!is.na(as.numeric(x$class)))) {
            x$class <- max(as.numeric(x$class), na.rm = TRUE)
          } else {
            x$class <- rep(NA_integer_, nrow(x))
          }
          x
        }), cls[, idvars])
    cls <- unique(cls)
    rx <- merge(rx, cls, by = idvars, sort = FALSE)
    # rx <- unique(rx) this changes the order
    rx$class <- grading_colors[paste(rx$class)]
    # rx$class <- factor(rx$class,
    #                    ordered = TRUE,
    #                    levels = unname(grading_colors),
    #                    labels = grading_labs)
    rx <- rx[order(rx[[".orig_order"]]), , FALSE]
    res <- rx$class
  } else {
    res <- NA
  }
  assign(key, res, envir = gg_cache_env)
  return(res)
}})

util_scale_fill_dataquieR <- function(...) {
  r <- ggplot2::scale_fill_identity(
    name = " ",
    labels = setNames(
      util_get_labels_grading_class(),
      util_get_colors()
    ),
    guide = ggplot2::guide_legend()
  )
  r$get_labels <- function(x, ..., self) {
    p <- setNames(nm = util_get_colors(),
                  util_get_labels_grading_class())
    p[x]
  }
  r
}

#' @family plotly_shims
#' @concept plotly_shims
#' @noRd
util_as_plotly_con_contradictions_redcap <- function(res, ...) {
  if (!util_ensure_suggested("plotly", err = FALSE)) {
    return(htmltools::HTML("No Plotly"))
  }
  # Maybe, we have Other, but certainly, we have SummaryPlot
  # fix the Legend
  col_map <- setNames(nm = util_get_colors(), util_get_labels_grading_class())
  col_map["#888888"] <- "NA"
  # hline, no legnedn from .. or guide hidden --> no legend needed, legacy mode
  # with threshold
  all_geoms <- lapply(util_gg_get(res$SummaryPlot, "layers"), `[[`, "geom")
  vlines <- vapply(all_geoms, inherits, "GeomVline",
                   FUN.VALUE = logical(1))
  hlines <- vapply(all_geoms, inherits, "GeomHline",
                   FUN.VALUE = logical(1))
  legacy_mode <- !!sum(vlines, hlines)
  py <- util_plot_figure_plotly(res[["SummaryPlot"]], attr(res, "sizing_hints"))
  if (legacy_mode) { # no legend in gg, legend static, here
    if (any(vlines)) {
      all_geoms[vlines][[1]]$parameters
      thr <- util_gg_get(res$SummaryPlot, "layers")[vlines][[1]]$data$xintercept
    } else {
      all_geoms[hlines][[1]]$parameters
      thr <- util_gg_get(res$SummaryPlot, "layers")[hlines][[1]]$data$yintercept
    }
    col_map_yn <- c("1" = sprintf("> %.4g%%", thr),
                    "0" = sprintf("\u2264 %.4g%%", thr))
    for (i in seq_along(py$x$data)) {
      py$x$data[[i]]$name <- lapply(
        strsplit(gsub("^\\((.*)\\)$", "\\1", py$x$data[[i]]$name), split = ","),
        function(scales) {
          scales[scales %in% names(col_map_yn)] <-
            col_map_yn[scales[scales %in% names(col_map_yn)]]
          paste0("(", paste0(scales, collapse = ","), ")")
        }
      )
    }
  } else {
    for (i in seq_along(py$x$data)) {
      py$x$data[[i]]$name <- lapply(
        strsplit(gsub("^\\((.*)\\)$", "\\1", py$x$data[[i]]$name), split = ","),
        function(scales) {
          scales[scales %in% names(col_map)] <-
            col_map[scales[scales %in% names(col_map)]]
          paste0("", paste0(scales, collapse = ", "), "")
        }
      )
    }
  }
  py
}

#' Convert a pipeline result data frame to named encapsulated lists
#'
#' @description
#'
#' Deprecated
#'
#' @param ... Deprecated
#'
#' @return Deprecated
#'
#' @export
#' @importFrom stats setNames
pipeline_recursive_result <- function(...) { # nocov start
  lifecycle::deprecate_stop("2.1.0", what = "dq_report()", with = "dq_report2()")
} # nocov end

#' Convert a [character()] to a `numeric_with_unit`
#'
#' @param x [character()] to convert/parse
#'
#' @returns `numeric_with_unit`
#' @noRd
util_as_numeric_with_unit <- function(x) {
  n <- trimws(gsub("^([\\d\\.\\-eE\\+, ]+).*$", "\\1", x, perl = TRUE))
  u <- trimws(gsub("^[\\d\\.\\-eE\\+, ]+", "", x, perl = TRUE))
  ok <- try({
    n <- as.numeric(n)
    bu <- util_unit2baseunit(u)
    prefix <- sub(paste0(bu, "$"), "", u)
    if (nzchar(prefix)) {
      fct <- UNIT_PREFIX_FACTORS[prefix]
    } else {
      fct <- 1
    }
    value <- fct * n
    unit <- bu
  }, silent = TRUE)
  if (!util_is_try_error(ok)) {
    return(structure(
      as.numeric(value),
      unit = unit,
      class = "numeric_with_unit"
    ))
  } else {
    util_warning("Could not parse number with unit: %s",
                 dQuote(x))
    return(value)
  }
}

#' Print a number with unit
#'
#' @param x number with unit
#' @param ... not used
#'
#' @returns `invisible(x)`
#' @export
print.numeric_with_unit <- function(x, ...) {
  util_stop_if_not("Intenal error, sorry, please report: Inadmissible call of print.numeric_with_unit" =
                   inherits(x, "numeric_with_unit"))
  cat(format(x))
  if (!!length(attr(x, "unit")) &&
    nzchar(attr(x, "unit"))) cat(" [", attr(x, "unit"), "]", sep = "")
  cat("\n")
  invisible(x)
}

#' Operator caring for units
#'
#' @param e1 first argument
#' @param e2 second argument
#'
#' @returns result
#' @keywords internal
util_op_numeric_with_unit <- function(e1, e2) { # TODO: Vectorize, nicer error messages, better support for 1 + 2%, better maybe some implementations of +.character and similar, either concatenating or summing up controlled by some option()
  # TODO: < > ==
  # assumes, e1 and e2 have been created using util_unit2baseunit(), so
  # unit prefixes have been normalized, already
  # preps
  count_def <- attr(UNIT_IS_COUNT, "def")
  count_def[""] <- 1
  # 1st try to make e1, e2 numeric only, if these have UNIT_IS_COUNT units or are not numeric_with_unit at all.
  is1 <- inherits(e1, "numeric_with_unit")
  is2 <- inherits(e2, "numeric_with_unit")
  if (is1) {
    u1 <- attr(e1, "unit")
    if (!length(u1)) {
      u1 <- ""
    }
    if (!!length(u1) && !is.na(count_def[u1])) {
      e1 <- as.numeric(count_def[u1] * as.numeric(e1))
      u1 <- NULL
    }
  }
  if (is2) {
    u2 <- attr(e2, "unit")
    if (!length(u2)) {
      u2 <- ""
    }
    if (!!length(u2) && !is.na(count_def[u2])) {
      e2 <- as.numeric(count_def[u2] * as.numeric(e2))
      u2 <- NULL
    }
  }

  # unitless stuff may now be plain numbers
  is1 <- inherits(e1, "numeric_with_unit")
  is2 <- inherits(e2, "numeric_with_unit")

  result_u <- NULL

  if (is1 || is2) {
    op <- rlang::call_name(sys.call())
    my_op <- sub("\\.numeric_with_unit", "", op)
    u1 <- attr(e1, "unit")
    if (!length(u1)) {
      u1 <- ""
    }
    u2 <- attr(e2, "unit")
    if (!length(u2)) {
      u2 <- ""
    }
    if (u1 == "") { is1 <- FALSE }
    if (u2 == "") { is2 <- FALSE }
    if (is1 && is2) { # both have a unit
      if (my_op %in% c("*", "/", "%%", "%/%")) {
        if (u1 != u2) {
          util_error("Cannot calclate %s %s %s",
                     u1, my_op, u2)
        }
        if (my_op %in% c("/", "%/%")) {
          result_u <- NULL
        } else if (my_op %in% c("%%")) {
          result_u <- u1
        } else if (my_op %in% c("*")) {
          util_warning("derived units not yet fully supported")
          result_u <- sprintf("%s%s%s", u1, my_op, u2)
        }
      } else if (my_op %in% c("+", "-")) {
        if (u1 != u2) {
          util_error("Cannot calclate %s %s %s",
                     u1, my_op, u2)
        }
        result_u <- u1
      } else if (my_op %in% c("^")) {
        util_error("Cannot calclate %s %s %s",
                   u1, my_op, u2)
      }
    } else { # only one has a real unit
      if (my_op %in% c("*", "/", "%%", "%/%")) {
        # return(NextMethod())
        result_u <- paste0(u1, u2)
      } else if (my_op %in% c("^")) {
        if (u2 == "") {
          util_warning("derived units not yet fully supported")
          result_u <- sprintf("%s^%d", u1, e2)
        } else {
          util_error("Cannot calclate %s %s %s",
                     u1, my_op, u2)
        }
      } else if (my_op %in% c("+", "-")) {
        if (u1 != u2) {
          util_error("Cannot calclate %s %s %s",
                     u1, my_op, u2)
        }
      }
    }
  }
  r <- NextMethod()
  if (is.null(result_u)) {
    r <- as.numeric(r)
  } else {
    attr(r, "unit") <- result_u
  }
  return(r)
}

#' @inherit util_op_numeric_with_unit
#' @export
`*.numeric_with_unit` <- util_op_numeric_with_unit

#' @inherit util_op_numeric_with_unit
#' @export
`+.numeric_with_unit` <- util_op_numeric_with_unit

#' @inherit util_op_numeric_with_unit
#' @export
`-.numeric_with_unit` <- util_op_numeric_with_unit

#' @inherit util_op_numeric_with_unit
#' @export
`/.numeric_with_unit` <- util_op_numeric_with_unit

#' @inherit util_op_numeric_with_unit
#' @export
`%%.numeric_with_unit` <- util_op_numeric_with_unit

#' @inherit util_op_numeric_with_unit
#' @export
`%/%.numeric_with_unit` <- util_op_numeric_with_unit

#' @inherit util_op_numeric_with_unit
#' @export
`^.numeric_with_unit` <- util_op_numeric_with_unit

#' Create an overview of the reports created with `dq_report_by`
#'
#' writes to the files
#' `index.html`, `dashboard.html`, `tables.html` in `output_dir`.
#'
#' @param output_dir [character] the directory in which all reports are searched
#'                               and the overview is saved
#'
#' @return `invisible(NULL)`
#' @noRd
util_create_report_by_overview <- function(output_dir) {

  util_ensure_suggested("jsonlite",
                        goal = "overall-overviews",
                        err = TRUE)

  # prepare the path of the output_dir with the / at the end
  out_dir <- output_dir
  if (!endsWith(out_dir, .Platform$file.sep)) {
    out_dir <- paste0(out_dir, .Platform$file.sep)
  }

  strata_column <- NULL
  rm("strata_column")
  segment_column <- NULL
  rm("segment_column")
  strata_column_label <- NULL
  rm("strata_column_label")
  subgroup <- NULL
  rm("subgroup")
  mod_label <- NULL
  rm("mod_label")
  title <- NULL
  rm("title")
  disable_plotly <- NULL
  rm("disable_plotly")
  rep_id <- NULL
  rm("rep_id")
  start_time <- NULL
  rm("start_time")
  subtitle <- NULL
  rm("subtitle")
  author <- NULL
  rm("author")
  user_info <- NULL
  rm("user_info")
  by_call <- NULL
  rm("by_call")

  list2env(readRDS(file.path(out_dir, "report_by_meta.RDS")),
           envir = environment())
  # this contains as list of
  # @param strata_column [character] name of a study variable to stratify the
  #                                    report by. It can be null
  # @param segment_column [character] name of a metadata attribute
  #                                    usable to split the report in
  #                                    sections of variables. It can be null
  # @param strata_column_label [character] the label of the variable used as
  #                                           strata_column
  # @param subgroup [character] optional, to define subgroups of cases
  # @param mod_label [list] `util_ensure_label()` info
  # @param title [character] a title for the overview `HTML`-file
  # @param disable_plotly [logical] do not use `plotly`, even if installed
  # @param rep_id [character] unique ID for the report to detect changes
  # @param start_time [as.POSIXct] time point when this computation has started

  if (nzchar(trimws(subtitle))) {
    subtitle <- paste0(subtitle, ": ")
  }

  util_expect_scalar(title, check_type = is.character)

  packageName <- utils::packageName()

  doc_title <- title

  # create a summary of summaries
  # import all the R objects containing a report summary and create a list
  sum_names <- list.files(path = output_dir, pattern = "^report_summary_")
  sum_names2 <- vapply(sum_names, function(x)
    paste0(out_dir, x), FUN.VALUE = character(1))
  list_summaries2 <- lapply(sum_names2, FUN = readRDS)

  # create a toc
  toc <- lapply(sum_names, function(x) {
    path_name <- paste0(out_dir, x)
    if (!"report.html" %in%
        list.files(path = file.path(sub("/$", "", out_dir),
                                    sub("^report_summary", "report",
                                        sub(".RDS$", "", x)),
                                    ".report"))) {
      return(NULL)
    }
    temp_file <- readRDS(path_name)
    this1 <- attr(temp_file, "this")
    sdn <- this1$sdn
    path_name <- file.path(output_dir, gsub("[^a-zA-Z0-9_\\.]", "",
                                            sprintf("report_%s", sdn)))
    location <- gsub("[^a-zA-Z0-9_\\.]", "",
                     sprintf("report_%s", sdn))
    toc <- list()
    toc[[this1$stratum]] <- list()
    toc[[this1$stratum]][[this1$segment]] <- location
    return(toc)
  })
  toc <- toc[!vapply(toc, is.null, FUN.VALUE = logical(1))]

  # create a vector containing all variables and their location in folder
  vars_reportlocation <- lapply(sum_names, function(x) {
    name_folder <- gsub("(report_summary_)(.*)\\.RDS", "\\2", x)
    name_folder <- paste0("report_", name_folder)
    path_name <- paste0(out_dir, x)
    temp_file <- readRDS(path_name)
    this1 <- attr(temp_file, "this")
    used_data_file <- this1$used_data_file
    # leave only file name and extension, if the vector contains a path
    if (grepl(.Platform$file.sep, used_data_file, fixed = TRUE)) {
      used_data_file <- gsub(".+\\/(.+\\..+$)", "\\1", used_data_file)
    }
    stratum <- this1$stratum
    segm <- this1$segment
    # the complete name changes depending of the split selected
    if (is.null(strata_column)) {
      origin_name <- paste0(used_data_file, "-", segm)
    } else if (!is.null(strata_column) && is.null(segment_column)) {
      origin_name <- paste0(used_data_file, "-", stratum)
    } else if (!is.null(strata_column) && !is.null(segment_column)) {
      origin_name <- paste0(used_data_file, "-", stratum, "-", segm)
    }

    temp_res <- this1$result[, colnames(this1$result) %in%
                               c(VAR_NAMES, "call_names"), drop = FALSE]

    if (this1$label_col != VAR_NAMES) {
      temp_varnames_colnames <- this1$meta_data[, colnames(this1$meta_data) %in%
                                                  c(VAR_NAMES, this1$label_col)]
      temp_res$VAR_NAMES <- util_map_labels(temp_res$VAR_NAMES,
                                            meta_data = temp_varnames_colnames,
                                            from = VAR_NAMES,
                                            to = this1$label_col)
    }

    temp_res$md_names <- paste0(temp_res$VAR_NAMES,
                                rep(".", nrow(temp_res)),
                                temp_res$call_names)
    if (nrow(temp_res) == 0) {
      return(setNames(list(), nm = character(0)))
    }
    md_names <- temp_res$md_names
    md_names <- paste0(origin_name, "-", md_names)
    if (is.null(strata_column) || (!is.null(strata_column) &&
                                   length(list_summaries2) == 1)) {
      origin_name <- paste0(origin_name, "-")
      md_names <- sub(origin_name, "", md_names, fixed = TRUE)
    }

    md_names <- unique(md_names)
    label_col_and_folder <- list(setNames(rep(name_folder,
                                              times = length(md_names)),
                                          nm = md_names))
    return(label_col_and_folder)
  })

  vars_reportlocation <- unlist(vars_reportlocation)

  # In case of creation of report by strata (more than one strata) the VAR_NAMES
  # changes so a conversion dataframe is needed to create the links to the folders
  if (!is.null(strata_column) && length(list_summaries2) > 1) {
    # create a vector with the variable_names and the unique variable names
    # created when there is a strata selection
    vars_originalnames <- mapply(x = sum_names,
                                 SIMPLIFY = FALSE,
                                 FUN = function(x) {
                                   name_folder <- gsub("(report_summary_)(.*)\\.RDS",
                                                       "\\2", x)
                                   name_folder <- paste0("report_", name_folder)
                                   path_name <- paste0(out_dir, x)
                                   temp_file <- readRDS(path_name)
                                   this1 <- attr(temp_file, "this")
                                   used_data_file <- this1$used_data_file
                                   if (grepl(.Platform$file.sep,
                                             used_data_file,
                                             fixed = TRUE)) {
                                     used_data_file <- gsub(".+\\/(.+\\..+$)",
                                                            "\\1",
                                                            used_data_file)
                                   }
                                   stratum <- this1$stratum
                                   segm <- this1$segment
                                   # the complete name changes depending of the split selected
                                   if (is.null(strata_column)) {
                                     origin_name <- paste0(used_data_file,
                                                           "-",
                                                           segm)
                                   } else if (!is.null(strata_column) &&
                                              is.null(segment_column)) {
                                     origin_name <- paste0(used_data_file,
                                                           "-",
                                                           stratum)
                                   } else if (!is.null(strata_column) &&
                                              !is.null(segment_column)) {
                                     origin_name <- paste0(used_data_file,
                                                           "-",
                                                           stratum,
                                                           "-",
                                                           segm)
                                   }

                                   temp_res <- this1$result[, colnames(this1$result) %in%
                                                              c(VAR_NAMES),
                                                            drop = FALSE]
                                   temp_res <- temp_res[!duplicated(temp_res),
                                                        ,
                                                        drop = FALSE]
                                   temp_res$LABEL <- rep(NA, nrow(temp_res))
                                   if (this1$label_col != VAR_NAMES) {
                                     temp_varnames_colnames <-
                                       this1$meta_data[, colnames(this1$meta_data) %in%
                                                         c(VAR_NAMES,
                                                           this1$label_col)]
                                     temp_res$LABEL <- util_map_labels(
                                       temp_res$VAR_NAMES,
                                       meta_data = temp_varnames_colnames,
                                       from = VAR_NAMES,
                                       to = this1$label_col
                                     )
                                   }
                                   temp_res$md_names <- rep(NA, nrow(temp_res))
                                   md_names <- temp_res$VAR_NAMES
                                   md_names <- paste0(origin_name, "-", md_names)
                                   if (is.null(strata_column) ||
                                       (!is.null(strata_column) &&
                                        length(list_summaries2) == 1)) {
                                     origin_name <- paste0(origin_name, "-")
                                     md_names <- sub(origin_name,
                                                     "",
                                                     md_names,
                                                     fixed = TRUE)
                                   }
                                   if (nrow(temp_res) > 0) {
                                     temp_res$md_names <- md_names
                                   }

                                   temp_res$new_name_with_label <- paste0(temp_res$LABEL)
                                   new_name_with_label <- temp_res$new_name_with_label
                                   new_name_with_label <- paste0(origin_name,
                                                                 "-",
                                                                 new_name_with_label)
                                   if (is.null(strata_column) ||
                                       (!is.null(strata_column) &&
                                        length(list_summaries2) == 1)) {
                                     origin_name <- paste0(origin_name, "-")
                                     new_name_with_label <- sub(origin_name,
                                                                "",
                                                                new_name_with_label,
                                                                fixed = TRUE)
                                   }
                                   if (nrow(temp_res) > 0) {
                                     temp_res$new_name_with_label <- new_name_with_label
                                   }

                                   temp_res$name_matching_result2_Variables <-
                                     paste0(temp_res$LABEL)
                                   name_matching_result2_Variables <-
                                     temp_res$name_matching_result2_Variables
                                   name_matching_result2_Variables <-
                                     paste0(origin_name,
                                            " - ",
                                            name_matching_result2_Variables)
                                   if (is.null(strata_column) ||
                                       (!is.null(strata_column) &&
                                        length(list_summaries2) == 1)) {
                                     origin_name <- paste0(origin_name, "-")
                                     name_matching_result2_Variables <-
                                       sub(origin_name,
                                           "",
                                           name_matching_result2_Variables,
                                           fixed = TRUE)
                                   }
                                   if (nrow(temp_res) > 0) {
                                     temp_res$name_matching_result2_Variables <-
                                       name_matching_result2_Variables
                                   }
                                   label_col_and_folder <-
                                     data.frame(
                                       new_names_with_varnames = temp_res$md_names,
                                       new_names_with_label = temp_res$new_name_with_label,
                                       result2_Variables_match =
                                         temp_res$name_matching_result2_Variables,
                                       var_names = temp_res$VAR_NAMES,
                                       name_label = temp_res$LABEL
                                     )
                                   return(label_col_and_folder)
                                 })
  } else {
    vars_originalnames <- NULL
  }

  # create a summary of summaries to be included in the overview page
  if (is.null(strata_column)) {
    summary_all <- util_combine_list_report_summaries(list_summaries2,
                                                      type = "unique_vars")
  } else {
    summary_all <- util_combine_list_report_summaries(list_summaries2,
                                                      type = "repeated_vars")
  }

  # render the table and the pie chart of the overview summary
  TBSummaries <-
    util_render_table_dataquieR_summary(summary_all,
                                        folder_of_report = vars_reportlocation,
                                        var_uniquenames = vars_originalnames)
  PlotSummaries <- plot(summary_all,
                        dont_plot = TRUE,
                        disable_plotly = disable_plotly,
                        folder_of_report = vars_reportlocation,
                        var_uniquenames = vars_originalnames)

  dashboard_names <- list.files(path = output_dir,
                                pattern = "^report_dashboard_",
                                full.names = TRUE)

  all_dashboards <- lapply(setNames(nm = dashboard_names), FUN = function(f) {
    db <- readRDS(f)
    if (is.data.frame(db) && nrow(db) > 0) {
      if (is.null(strata_column)) {
        db$fq_VARNAME <- db$..VAR_NAMES
      } else {
        db$fq_VARNAME <- paste0(attr(db, "name_of_study_data"),
                                "-",
                                ifelse(!is.null(strata_column),
                                       paste0(attr(db, "level_name"), "-"),
                                       ""
                                ),
                                ifelse(!is.null(segment_column),
                                       paste0(
                                         util_with_orig_names(db)[[STUDY_SEGMENT]],
                                              "-"),
                                       ""
                                ),
                                db$..VAR_NAMES)
      }
    }
    db
  })
  all_dashboards_df <- util_rbind(data_frames_list = all_dashboards)
  rownames(all_dashboards_df) <- NULL
  full_sum <- util_add_links_to_summary_table(
    attr(summary_all, "this")$result,
    attr(summary_all, "this"),
    folder_of_report = vars_reportlocation)

  orig_cn <- colnames(all_dashboards_df)
  if (!!prod(dim(all_dashboards_df))) {
    to_rm <- util_untranslated_colnames(all_dashboards_df) %in%
      c("title", "href", "popup_href")
  } else {
    to_rm <- rep(FALSE, length(orig_cn))
  }

  ns <- attr(orig_cn, "ns")
  lang <- attr(orig_cn, "lang")
  if (!!length(orig_cn)) {
    class <- attr(orig_cn, "class")
  } else {
    class <- "dataquieR_translated"
  }
  nms <- attr(orig_cn, "names")

  all_dashboards_df[to_rm] <- NULL
  orig_cn <- orig_cn[!to_rm]
  nms <- nms[!to_rm]

  attr(orig_cn, "lang") <- lang
  attr(orig_cn, "ns") <- ns
  attr(orig_cn, "names") <- nms
  attr(orig_cn, "class") <- class

  util_translated_colnames(all_dashboards_df) <-
    orig_cn

  if (!!prod(dim(all_dashboards_df)) && !!prod(dim(full_sum))) {
    DashboardT <-
      merge(x = all_dashboards_df,
            y = full_sum[, -which(colnames(full_sum) == "value"), FALSE],
            by.x = util_translate(c("fq_VARNAME",
                                    "call_names",
                                    "indicator_metric"),
                                  as_this_translation = colnames(all_dashboards_df)),
            by.y = c("VAR_NAMES", "call_names", "indicator_metric"),
            all = FALSE)


    util_translated_colnames(DashboardT) <-
      util_translate(util_translate(colnames(DashboardT),
                                    as_this_translation = colnames(all_dashboards_df),
                                    reverse = TRUE),
                     as_this_translation = colnames(all_dashboards_df)) # TODO: avoid reverse

    DashboardT <- util_fix_columns_in_dashboard_for_overview(
      DashboardT,
      image_dir = file.path(output_dir, "dashboard_images")
    )

    Dashboard <- util_dashboard_table2widget(
      DashboardT,
      unique(lapply(lapply(list_summaries2, attr, "this"), `[[`, "label_col"))[[1]]
    )
  } else {
    Dashboard <- htmltools::HTML("")
  }

  if (!disable_plotly) {
    SunburstSummaries <- plot(summary_all,
                              dont_plot = TRUE,
                              disable_plotly = disable_plotly,
                              hierarchy = TRUE,
                              folder_of_report = vars_reportlocation,
                              var_uniquenames = vars_originalnames)
  } else {
    SunburstSummaries <- htmltools::HTML("")
  }

  # Create an all_ids file
  all_ids <- lapply(sum_names, function(x) {
    name_folder <- gsub("(report_summary_)(.*)\\.RDS", "\\2", x)

    path_all_ids_file <- file.path(out_dir,
                                   paste0("report_", name_folder),
                                   ".report",
                                   "anchor_list.RDS")
    if (!file.exists(path_all_ids_file)) {
      util_warning(
        "Cannot read %s for %s -- Internal error, sorry. Please report",
        dQuote(path_all_ids_file),
        dQuote(x)
      )
      return(character(0))
    }
    temp_file <- readRDS(path_all_ids_file)
    temp_file <- file.path(fsep = "/",
                           paste0("report_", name_folder),
                           ".report",
                           temp_file)
    return(temp_file)
  })

  # create an overall anchor list with all created results, useful not to
  # create links to non-existing results
  all_ids <- c(unlist(all_ids), "index.html", "dashboard.html", "tables.html")

  cat(
    sep = "",
    "window.all_ids = {\"all_ids\": ",
    paste0("[",
           paste0('"', all_ids, '"', collapse = ", "),
           "]"),
    "}",
    file = file.path(out_dir, "anchor_list.js")
  )

  # creates the html overview page that links all sub-reports created
  if (!is.null(strata_column) && !is.null(segment_column)) {
    level_seg <- names(unlist(toc))
    get_level <- function(str) {
      sub("\\..*", "", str)
    }
    sdlevel <- vapply(level_seg, get_level, FUN.VALUE = character(1))
    get_segm <- function(str) {
      sub(".*\\.", "", str)
    }
    segs <- vapply(level_seg, get_segm, FUN.VALUE = character(1))
    sdlevel <- substr(sdlevel, nchar(strata_column_label) + 2, 10000)
    title <- paste0("Report: ", segment_column, " = ", segs, ", ",
                    strata_column_label, " = ", sdlevel)
  } else if (is.null(strata_column) && !is.null(segment_column)) {
    level_seg <- names(unlist(toc))
    get_level <- function(str) {
      sub("\\..*", "", str)
    }
    sdlevel <- vapply(level_seg, get_level, FUN.VALUE = character(1))
    get_segm <- function(str) {
      sub(".*\\.", "", str)
    }
    segs <- vapply(level_seg, get_segm, FUN.VALUE = character(1))
    title <- paste0("Report: ", segment_column, " = ", segs, ".", sdlevel)
  } else if (!is.null(strata_column) && is.null(segment_column)) {
    sdlevel <- names(unlist(toc))
    sdlevel <- substr(sdlevel, nchar(strata_column_label) + 2, 10000)
    title <- paste("Report: ", strata_column_label, " = ", sdlevel)
  } else if (is.null(strata_column) && is.null(segment_column) &&
             !is.null(subgroup)) {
    if (is.character(title)) {
      title <- paste0(title, ". Subgroup: ", subgroup)
    } else {
      title <- paste0("Report on subgroup: ", subgroup)
    }
  } else {
    title <- paste0("Report on all variables and observations")
  }

  # create the link for the created reports
  if (length(toc) == 0) {
    href <- character(0)
  } else {
    href <- paste0(unlist(toc), "/.report/report.html")
  }
  util_ensure_suggested("htmltools", "generate reports TOC")

  has_legend <- exists("..INFO_SD_NAME_FOR_REPORT", .dataframe_environment())

  overview_css_dep <- htmltools::htmlDependency(
    name = "dataquieR.dq_report_by_overview",
    version = "0.0.1",
    src = system.file("menu", package = "dataquieR"),
    stylesheet = c("dataquieR_overview.css")
  )


  build_report_list <- function(href, title) {
    htmltools::tags$div(
      class = "dq-overview-card dq-overview-card-links",
      htmltools::tags$div(
        class = "dq-overview-card-title",
        "List of created reports"
      ),
      htmltools::tags$ul(
        class = "dq-overview-report-list",
        lapply(seq_along(href), function(i) {
          htmltools::tags$li(
            class = "dq-overview-report-item",
            htmltools::tags$a(
              class = "dq-overview-report-link",
              href = href[[i]],
              title[[i]]
            )
          )
        })
      )
    )
  }

  build_legend_block <- function() {
    info_sd <- as.list(prep_get_data_frame("..INFO_SD_NAME_FOR_REPORT"))

    htmltools::tags$div(
      class = "dq-overview-card dq-overview-card-legend",
      htmltools::tags$div(
        class = "dq-overview-card-title",
        "Legend"
      ),
      htmltools::tags$div(
        class = "dq-overview-legend-grid",
        lapply(names(info_sd), function(x) {
          htmltools::tags$div(
            class = "dq-overview-legend-item",
            htmltools::tags$div(
              class = "dq-overview-legend-key",
              x
            ),
            htmltools::tags$div(
              class = "dq-overview-legend-value",
              as.character(info_sd[[x]])
            )
          )
        })
      )
    )
  }

  build_tab_nav <- function(active_file) {
    tabs <- list(
      list(file = "index.html", label = "Sunburst"),
      list(file = "dashboard.html", label = "Dashboard"),
      list(file = "tables.html", label = "Summary Table")
    )

    htmltools::tags$div(
      class = "dq-overview-tabs",
      lapply(tabs, function(tab) {
        classes <- "dq-overview-tab"
        if (identical(tab$file, active_file)) {
          classes <- paste(classes, "active")
        }
        htmltools::tags$a(
          class = classes,
          href = tab$file,
          `data-no-existance-check` =
            jsonlite::toJSON(TRUE, auto_unbox = TRUE),
          tab$label
        )
      })
    )
  }

  build_footer <- function(start_time, end_time) {
    p <- NULL
    if (is.list(user_info)) {
      # TODO: jsTree
      p <- user_info
      p[vapply(p, is.call, FUN.VALUE = logical(1))] <-
        vapply(lapply(p[vapply(p, is.call, FUN.VALUE = logical(1))], deparse),
               paste, collapse = "\n", FUN.VALUE = character(1))
      p[vapply(p, inherits, "POSIXt", FUN.VALUE = logical(1))] <-
        vapply(p[vapply(p, inherits, "POSIXt", FUN.VALUE = logical(1))],
               paste, collapse=" ", FUN.VALUE = character(1))
      p <- p[vapply(p, is.vector, FUN.VALUE = logical(1))]
      p <- p[vapply(p, length, FUN.VALUE = integer(1)) == 1]
      p <- data.frame(`  ` = names(p),
                      ` ` = unlist(unname(p)),
                      stringsAsFactors = FALSE,
                      check.names = FALSE)
    }


    htmltools::tagList(
      util_html_table(
        title = prep_title_escape(doc_title, TRUE),
        rotate_for_one_row = TRUE,
        util_rbind(data.frame(
          check.names = FALSE,
          fix.empty.names = FALSE,
          `Overall computation time` =
            as.character(htmltools::span(`data-content` = "renderingTime")),
          `dataquieR version` = util_dataquieR_version(),
          `Timestamp` = format(start_time),
          `Author` = author,
          `Call` =
            as.character(htmltools::pre(paste0(deparse(by_call),
                                               collapse = "\n")))
        ), p)
      )
    )
  }

  build_section_card <- function(title_text, content) {
    htmltools::tags$div(
      class = "dq-overview-section-card",
      htmltools::tags$div(
        class = "dq-overview-section-heading",
        title_text
      ),
      htmltools::tags$div(
        class = "dq-overview-section-body",
        content
      )
    )
  }

  build_overview_toc <- function(active_file,
                                 include_plot_summaries = FALSE,
                                 include_sunburst_summaries = FALSE,
                                 include_dashboard = FALSE,
                                 include_tb_summaries = FALSE,
                                 end_time) {

    summary_sections <- list()

    if (include_plot_summaries) {
      summary_sections <- c(
        summary_sections,
        list(build_section_card("Plots", PlotSummaries))
      )
    }
    if (include_sunburst_summaries) {
      summary_sections <- c(
        summary_sections,
        list(build_section_card("Sunburst", SunburstSummaries))
      )
    }
    if (include_dashboard) {
      summary_sections <- c(
        summary_sections,
        list(build_section_card("Dashboard", Dashboard))
      )
    }
    if (include_tb_summaries) {
      summary_sections <- c(
        summary_sections,
        list(build_section_card("Tables", TBSummaries))
      )
    }

    header_subtitle <- paste0(
      util_by_header_from_args(by_call),
#      "\nInteractive overview of created reports",
      if (is.character(doc_title) && length(doc_title) == 1L &&
          nzchar(doc_title)) {
        paste0(" \u2014 ", doc_title)
      } else {
        ""
      }
    )

    do.call(
      htmltools::tagList,
      list(
        htmltools::tags$script(src = ".report/renderinfo.js",
                               type = "text/javascript"),
        htmltools::div(class = "navbar"),
        htmltools::tags$script(
          type = "text/javascript",
          "window.dq_report_by_overview = true"
        ),
        htmltools::tags$script(
          type = "text/javascript",
          src = "anchor_list.js"
        ),
        htmltools::tags$div(
          class = "dq-overview-app",
          htmltools::tags$div(
            class = "dq-overview-shell",
            htmltools::tags$div(
              class = "dq-overview-topbar",
              htmltools::tags$div(
                class = "dq-overview-eyebrow",
                "dataquieR overview"
              ),
              htmltools::tags$h1(
                class = "dq-overview-title",
                doc_title
              ),
              htmltools::tags$p(
                class = "dq-overview-subtitle",
                paste0(author, ": ", header_subtitle,
                       " (", format(start_time), ")")
              ),
              build_tab_nav(active_file)
            ),
            htmltools::tags$div(
              class = "dq-overview-main",
              summary_sections,
              build_report_list(href = href, title = title),
              if (has_legend) build_legend_block()
            ),
            htmltools::tags$div(
              class = "dq-overview-footer",
              build_footer(start_time = start_time,
                           end_time = end_time)
            )
          )
        )
      )
    )
  }

  if (nchar(mod_label$label_modification_text) > 0) {
    notes_labels <- htmltools::div(
      htmltools::h3("Label modifications"),
      util_html_table(util_df_escape(mod_label$label_modification_table),
                      dl_fn = "Label_modifications")
    )
  } else {
    notes_labels <- htmltools::div()
  }

  logo_rel <- "logo.png"

  file.copy(system.file("logos",
                        "dataquieR_48x48.png",
                        package = packageName),
            file.path(output_dir, logo_rel))

  save_overview_html <- function(toc, file_name) {
    htmltools::save_html(
      htmltools::tagList(
        overview_css_dep,
        rmarkdown::html_dependency_jquery(),
        html_dependency_dataquieR(iframe = FALSE),
        htmltools::tags$head(
          htmltools::tags$script(
            type = "text/javascript",
            "window.dq_report_by_overview = true"
          ),
          htmltools::tags$title(doc_title),
          htmltools::tags$link(
            rel = "icon",
            type = "image/png",
            href = logo_rel
          ),
          htmltools::tags$meta(name = "viewport",
                               content = "width=device-width, initial-scale=1")
        ),
        htmltools::HTML("<!-- done -->"),
        toc,
        htmltools::tags$div(
          class = "dq-overview-shell dq-overview-notes",
          notes_labels
        )
      ),
      file = file.path(output_dir, file_name)
    )
  }

  end_time <- Sys.time()

  toc_index <- build_overview_toc(
    active_file = "index.html",
    include_plot_summaries = FALSE,
    include_sunburst_summaries = TRUE,
    include_dashboard = FALSE,
    include_tb_summaries = FALSE,
    end_time = end_time
  )

  toc_dashboard <- build_overview_toc(
    active_file = "dashboard.html",
    include_plot_summaries = TRUE,
    include_sunburst_summaries = FALSE,
    include_dashboard = TRUE,
    include_tb_summaries = FALSE,
    end_time = end_time
  )

  toc_tables <- build_overview_toc(
    active_file = "tables.html",
    include_plot_summaries = TRUE,
    include_sunburst_summaries = FALSE,
    include_dashboard = FALSE,
    include_tb_summaries = TRUE,
    end_time = end_time
  )

  save_overview_html(toc_index, "index.html")
  save_overview_html(toc_dashboard, "dashboard.html")
  save_overview_html(toc_tables, "tables.html")

  util_write_renderinfo_js_json(
    output_dir,
    rep_id,
    start_time = start_time,
    end_time = end_time
  )

  invisible(NULL)
}

#' Create a Short Subtitle From a dq_report_by Call
#'
#' @param by_call a call to `dq_report_by()` (possibly unevaluated)
#'
#' @return character(1)
#' @noRd
util_by_header_from_args <- function(by_call) {
  stopifnot(!missing(by_call))

  bcl <- rlang::call_match(
    call = by_call,
    fn = dq_report_by
  )

  args <- rlang::call_args(bcl)

  deparse1_safe <- function(x) {
    if (is.null(x)) return(NULL)
    paste(deparse(x), collapse = "")
  }

  clean_expr <- function(x) {
    if (is.null(x)) return(NULL)
    gsub('^"|"$', "", x)
  }

  as_label <- function(x) {
    if (is.null(x)) {
      return(NULL)
    }

    if (is.character(x)) {
      return(x)
    }

    if (is.symbol(x)) {
      return(as.character(x))
    }

    if (is.call(x) && identical(x[[1]], as.name("::")) && length(x) == 3) {
      return(as.character(x[[3]]))
    }

    clean_expr(deparse1_safe(x))
  }

  normalize_chr <- function(x) {
    if (is.null(x)) {
      return(NULL)
    }

    if (is.character(x)) {
      return(x)
    }

    if (is.symbol(x)) {
      return(as.character(x))
    }

    if (is.call(x) && identical(x[[1]], as.name("c"))) {
      vals <- unlist(lapply(as.list(x)[-1], function(xx) {
        if (is.character(xx)) {
          xx
        } else if (is.symbol(xx)) {
          as.character(xx)
        } else if (is.call(xx) &&
                   identical(xx[[1]], as.name("::")) &&
                   length(xx) == 3) {
          as.character(xx[[3]])
        } else {
          clean_expr(deparse1_safe(xx))
        }
      }), use.names = FALSE)
      return(vals)
    }

    clean_expr(deparse1_safe(x))
  }

  collapse_labels <- function(x) {
    if (!length(x)) {
      return("")
    }
    if (length(x) == 1) {
      return(x)
    }
    if (length(x) == 2) {
      return(paste(x, collapse = " & "))
    }

    paste0(
      paste(x[-length(x)], collapse = ", "),
      ", & ",
      x[length(x)]
    )
  }

  strata <- if ("strata_column" %in% names(args)) {
    as_label(args$strata_column)
  } else {
    NULL
  }

  segment <- if ("segment_column" %in% names(args)) {
    if (is.null(args$segment_column)) NULL else as_label(args$segment_column)
  } else {
    NULL
  }

  dimensions <- if ("dimensions" %in% names(args)) {
    normalize_chr(args$dimensions)
  } else {
    NULL
  }

  default_dims <- c("des", "int", "com", "con")
  full_dims <- c("des", "int", "com", "con", "acc")
  optional_dims <- c("com", "con", "acc")

  if (is.null(dimensions)) {
    dimensions <- default_dims
  }

  dimensions <- unique(tolower(dimensions))

  if (identical(dimensions, "all")) {
    dimensions <- full_dims
  }

  dim_label <- function(x) {
    if (x %in% names(dims)) {
      unname(dims[[x]])
    } else {
      x
    }
  }

  parts <- character()

  if (!is.null(strata)) {
    parts <- c(parts, sprintf("Stratified by %s", strata))
  }

  if (!is.null(segment)) {
    parts <- c(
      parts,
      sprintf(
        "Split by Segments as Defined in %s in the Metadata",
        segment
      )
    )
  }

  present_opt <- intersect(optional_dims, dimensions)
  added_opt <- setdiff(present_opt, intersect(optional_dims, default_dims))
  missing_opt <- setdiff(optional_dims, present_opt)

  if (length(added_opt)) {
    parts <- c(
      parts,
      sprintf(
        "Including %s Checks",
        collapse_labels(vapply(added_opt, dim_label, character(1)))
      )
    )
  }

  if (length(missing_opt) && !identical(sort(missing_opt), "acc")) {
    parts <- c(
      parts,
      sprintf(
        "Without %s Checks",
        collapse_labels(vapply(missing_opt, dim_label, character(1)))
      )
    )
  }

  if (!length(parts)) {
    return("Data Quality Report Bundle")
  }

  paste("Data Quality Report Bundle", paste(parts, collapse = " and "))
}

#' Combine data frames by merging
#'
#' This is an extension of `merge` working for a list of data frames.
#'
#' @param data_frames [list] of [data.frame]s
#' @param id_vars [character] the variable(s) to merge the data frames by.
#'                            each of them must exist in all data frames.
#'
#' @return [data.frame] combination of data frames
#'
#'
#' @seealso [prep_merge_study_data]
#'
#' @family data_management
#' @concept data_management
#' @noRd
util_merge_data_frame_list <- function(data_frames, id_vars) {
  util_expect_scalar(id_vars, allow_more_than_one = TRUE)
  util_stop_if_not(is.list(data_frames))
  data_frames <- data_frames[vapply(data_frames, is.data.frame,
                                    FUN.VALUE = logical(1))]
  invisible(lapply(names(data_frames),
                   function(nm) {
                     assign(nm,  # ensure a speaking error message
                            data_frames[[nm]])
                     eval(call( # ensure a speaking error message
                       "util_expect_data_frame",
                       as.symbol(nm),
                       id_vars
                     ))
                   })
            )
  all_col_names <- unlist(lapply(data_frames, colnames), recursive = TRUE)
  ambiguous_column_names <- setdiff(
    unique(all_col_names[duplicated(all_col_names)]), id_vars)
  data_frames <- lapply(setNames(nm = names(data_frames)), function(nm) {
    dfr <- data_frames[[nm]]
    acn <- intersect(ambiguous_column_names, colnames(dfr))
    dfr[, paste0(acn, ".", nm)] <-
      dfr[, acn, drop = FALSE]
    dfr[, acn] <- NULL
    dfr
  })
  res <- Reduce(function(x, y) {
    rownames(x) <- rownames(y) <- NULL
    in_both <- setdiff(intersect(colnames(x), colnames(y)), id_vars)
    res <- merge(x,
                 y,
                 by = id_vars,
                 all = TRUE)
    res
  }, data_frames)
  for (acn in ambiguous_column_names) {
    all_cns <- intersect(colnames(res), paste0(acn, ".", names(data_frames)))

    all_equal <- function(cols, .res = res) {
      if (length(cols) < 2) {
        TRUE
      } else {
        .res[[cols[[2]]]][is.na(.res[[cols[[2]]]])] <- # na.rm doesn't work, if recursion cuts off head and tail had an NA but later some value different from the old head
          .res[[cols[[1]]]][is.na(.res[[cols[[2]]]])]
        all(.res[, cols[[1]]] == .res[, cols[[2]]], na.rm = TRUE) &&
          Recall(tail(cols, -1), .res = .res)
      }
    }
    if (all_equal(all_cns)) {
      res[[acn]] <- unlist(
        lapply(apply(res[, all_cns], 1, unique), setdiff, NA),
        recursive = FALSE)
      res[, all_cns] <- NULL
    }
  }
  res
}

#' Get the Table with Known Vocabularies
#'
#' @param .data_frame_list [environment] cache for loaded data frames
#'
#' @return [data.frame] the (combined) table with known vocabularies
#' @noRd
util_get_voc_tab <- function(.data_frame_list = .dataframe_environment()) {
  voc_tab <- util_get_concept_info("voc")
  if ("<>" %in% names(.data_frame_list)) { # also labels?
    my_voc <- .dataframe_environment()[["<>"]]
    if (!is.data.frame(my_voc) || !all(c("voc", "url") %in% colnames(my_voc))) {
      util_warning(
        c("Invalid <>-Bookmark-Table, need two columns %s in",
          "a data frame. Ignoring the user-provided table %s"),
        util_pretty_vector_string(c("voc", "url")),
        dQuote("<>"),
        applicability_problem = TRUE,
        intrinsic_applicability_problem = FALSE
      )
    } else {
      voc_tab <- voc_tab[!(voc_tab$voc %in% my_voc$voc), , FALSE]
      voc_tab <- util_rbind(my_voc, voc_tab)
    }
  }
  voc_tab
}

#' Create a caption from an alias name of a `dq_report2` result
#'
#' @param alias alias name
#' @param long return result based on `menu_title_report`,
#'             `matrix_column_title_report` otherwise
#'
#' @return caption
#'
#' @seealso `util_html_table`
#' @family reporting_functions
#' @concept html
#' @noRd
util_alias2caption <- function(alias, long = FALSE) {

  util_expect_scalar(long, check_type = is.logical)

  if (length(alias) != 1 ||
      !is.character(alias) ||
      is.na(alias)) {
    return("Unkown Alias")
  }

  fname <- util_map_by_largest_prefix(
    alias,
    haystack = names(.manual$titles))

  if (is.na(fname)) {
    fname <- alias
  }

  if (long) {
    ftitle <-
      util_map_labels(fname,
                      util_get_concept_info("implementations"),
                      to = "menu_title_report",
                      from = "function_R",
                      ifnotfound = NA_character_)
    if (util_empty(ftitle)) {
      ftitle <-
        util_map_labels(fname,
                        util_get_concept_info("implementations"),
                        to = "Implementationform",
                        from = "function_R",
                        ifnotfound = NA_character_)
    }
  } else {
    ftitle <-
      util_map_labels(fname,
                      util_get_concept_info("implementations"),
                      to = "matrix_column_title_report",
                      from = "function_R",
                      ifnotfound = NA_character_)
    if (util_empty(ftitle)) {
      ftitle <-
        util_map_labels(fname,
                        util_get_concept_info("implementations"),
                        to = "dq_report2_short_title",
                        from = "function_R",
                        ifnotfound = NA_character_)
    }
  }

  if (is.na(ftitle)) {
    ftitle <-
      r <- .manual$titles[[fname]];
    if (length(r) != 1 || is.na(r)) ftitle <- r <- alias;
  }

  if (startsWith(alias, fname)) {
    suffix <- substr(alias, nchar(fname) + 1 + 1, nchar(alias)) # name + "_" (first +1), start is the next character (second +1)
  } else {
    suffix <- alias
  }

  # acronyms <-
  #   util_map_labels(fnames,
  #                   util_get_concept_info("implementations"),
  #                   to = "dq_report2_short_title",
  #                   from = "function_R",
  #                   ifnotfound = util_abbreviate(fnames))

  suffix <- gsub("_", " ", suffix)
  if (!util_empty(suffix))
    suffix <- paste0(" ", suffix)

  r  <- paste0(ftitle, suffix)

  names(r) <- alias

  r

}

#' Get machine variant for snapshot tests
#'
#' @return [character] the variant
#'
#' @keywords internal
#' @export
prep_get_variant <- function() {
  v <- paste0("R", getRversion()[, 1:2])
#  if (v == "R4.4") { # current developer version
#    return(NULL)
#  } else {
    return(v)
#  }
}

#' utility function to test for applicability of marginal means plots
#'
#' Test for applicability of detection limits checks
#'
#' @param x [data.frame] metadata
#' @param dta [logical] vector, 1 = matching data type,
#'                            0 = non-matching data type
#'
#' @return [factor] 0-3 for each variable in metadata
#' - 0 data type mismatch and not applicable
#' - 1 data type mismatches but applicable
#' - 2 data type matches but not applicable
#' - 3 data type matches and applicable
#' - 4 not applicable because of not suitable data type
#'
#' @seealso [pro_applicability_matrix]
#' @noRd

util_app_mar <- function(x, dta) {
  c1 <- rowSums(!is.na(x[, grep("^GROUP_VAR_", colnames(x),
         perl = TRUE, value = TRUE), drop = FALSE])) > 0
  aa <- paste0(dta, as.integer(c1))
  score <- as.numeric(recode(as.factor(aa), "00" = 0, "01" = 1, "10" = 2,
                             "11" = 3))
  score <- ifelse(x[["DATA_TYPE"]] %in% c("integer", "float"), score, 4)
  score <- as.factor(score)
  return(score)
}

# This file creates an environment with functions, that are used in the
# function `des_summary`

#' `util_des_functions_env` -- an environment for `des_summary` functions
#' @name util_des_functions_env
#' @noRd
util_des_functions_env <- new.env(parent = emptyenv())

#' Compute the median for categorical variables
#'
#' @name util_compute_median_cat
#' @param x a vector
#'
#' @seealso [`StackOverflow`](https://stackoverflow.com/a/7925162)
#' @author Richie Cotton
#'
#' @return median for categorical variables
#' @noRd
util_des_functions_env$util_compute_median_cat <- function(x) {
  levs <- levels(x)
  m <- median(as.integer(x), na.rm = TRUE)
  if(floor(m) != m)
  {
    util_message("Median is between two values; using the first one")
    m <- floor(m)
  }
  ordered(m, labels = levs, levels = seq_along(levs))
}


#' Compute mode for categorical variables
#'
#' @name util_compute_mode_cat
#' @param x a vector
#'
#' @return the first 3 mode and the number of the remaining ones
#' @noRd
util_des_functions_env$util_compute_mode_cat <- function(x) {
  x <- x[!is.na(x)]
  mode_value <-  unique(x)[tabulate(match(x, unique(x))) ==
                             max(tabulate(match(x, unique(x))))]
  if (length(mode_value) > 3) {
    l1 <- length(mode_value) - 3
    mode_value <- paste0(paste(c(head(mode_value, 3)), collapse = " "),
                         " and other ",
                         l1, " categories")
  } else if (length(mode_value) > 1 && length(mode_value) < 4) {
    mode_value <- paste(mode_value,collapse = " ")
  }
  mode_value <- format(mode_value)
  return(mode_value)
}


#' Compute mode for continuous variables (excluding datetime)
#'
#' @name util_compute_mode_contin
#' @param x a numeric vector
#'
#' @return the mode
#' @noRd
util_des_functions_env$util_compute_mode_contin<- function(x) {
  x <- x[!is.na(x)]
  mode_value <-  unique(x)[tabulate(match(x, unique(x))) ==
                             max(tabulate(match(x, unique(x))))]
  if(length(mode_value)==1) {
    r <- format(mode_value)
  } else  if(length(mode_value)==2) {
    mode_value <- paste(mode_value,collapse = " ")
    r <- format(mode_value)
  } else if (length(mode_value)>2) {
    n_mode <- length(mode_value)
    mode_value <- mode_value[1:2]
    n_mode_removed <- n_mode - length(mode_value)
    mode_value <- paste(mode_value,collapse = " ")
    r <- paste0(format(mode_value), " and other ",
                n_mode_removed ," values")
  }
  return(r)
}


#' Compute mode for datetime variables
#'
#' @name util_compute_mode_datetime
#' @param x a vector containing date time values
#'
#' @return the first 2 mode as date time and the number of the remaining ones
#' @noRd
util_des_functions_env$util_compute_mode_datetime <- function(x) {
  x <- x[!is.na(x)]
  mode_value <-  unique(x)[tabulate(match(x, unique(x))) ==
                             max(tabulate(match(x, unique(x))))]
  if(length(mode_value)==1) {
    r <- format(mode_value, usetz = TRUE)
  } else if(length(mode_value)==2) {
    mode_value <- format(mode_value, usetz = TRUE)
    mode_value <- paste(mode_value,collapse = " ")
    r <- format(mode_value, usetz = TRUE)
  } else if (length(mode_value)>2) {
    mode_value <- format(mode_value, usetz = TRUE)
    n_mode <- length(mode_value)
    mode_value<- mode_value[1]
    n_mode_removed <- n_mode - length(mode_value)
    mode_value <- paste(mode_value,collapse = " ")
    r <- paste0(format(mode_value, usetz = TRUE), " and other ",
                n_mode_removed ," dates")
  }
  return(r)
}


#' Compute the interquartile range (IQR), the first quartile (Q1) and the
#' third quartile (Q3) for ordinal variables
#'
#' @name util_compute_IQR_ord
#' @param x a vector
#'
#' @return a [list] with:
#'   - `IQR`: a string containing the range
#'   - `IQR_q1_q3`: a string containing Q1 and Q3
#'
#'
#'  a list containing the  and a string containing Q1 and Q3
#' @noRd
util_des_functions_env$util_compute_IQR_ord <- function(x) {
  #Only for ORDINAL vars
  q <- quantile(x, na.rm = TRUE, names = TRUE, type = 1)
  names(q) <- paste0("Q", seq_len(length(q)) - 1)
  q <- q[c("Q1", "Q3")]
  r <- prep_deparse_assignments(labels = as.character(q),
                                codes = names(q),
                                mode = "string_codes")
  iqr <- IQR(na.rm = TRUE, x = x, type = 1)
  r <- paste0(format(iqr), " (", r, ")")
  return(list("IQR" = iqr,
              "IQR_q1_q3" = r))
}


#' Compute the interquartile range (IQR), the first quartile (Q1) and the
#' third quartile (Q3) for numerical variables
#'
#' @name util_compute_IQR_contin
#' @param x a vector
#'
#' @return a [list] with:
#'   - `IQR`: a numeric value indicating the interquartile range
#'   - `IQR_q1_q3`: a string containing Q1 and Q3
#' @noRd
util_des_functions_env$util_compute_IQR_contin <- function (x) {
  q <- quantile(x, na.rm = TRUE, names = FALSE)
  names(q) <- paste0("Q", seq_len(length(q)) - 1)
  q <- q[c("Q1", "Q3")]
  r <- prep_deparse_assignments(labels = as.character(q),
                                codes = names(q),
                                mode = "string_codes")
  iqr <- IQR(na.rm = TRUE, x = x, type = 7)
  r<- paste0(format(iqr),
             " (", r, ")")
  rm(q)
  return(list("IQR" = iqr,
              "IQR_q1_q3" = r))
}


#' Compute the interquartile range (IQR), the first quartile (Q1) and
#' the third quartile (Q3) for datetime variables
#'
#' @name util_compute_IQR_datetime
#' @param x a vector
#'
#' @return a [list] with:
#'   - `IQR`: the interquartile range in seconds
#'   - `IQR_q1_q3`: a string containing Q1 and Q3
#' @noRd
util_des_functions_env$util_compute_IQR_datetime <- function(x) {
  q <- quantile(x, na.rm = TRUE, names = FALSE)
  names(q) <- paste0("Q", seq_len(length(q)) - 1)
  q <- q[c("Q1", "Q3")]
  r <- prep_deparse_assignments(labels = format(q, usetz = TRUE),
                                codes = names(q),
                                mode = "string_codes")
  iqr <- IQR(na.rm = TRUE, x = x,
             type = 7)
  #iqr1 <- format(as.difftime(iqr, units = "secs"), usetz = TRUE)


  iqr1 <- as.difftime(iqr, units = "secs")

#total_seconds <- lubridate::as.duration(
#  as.numeric(iqr1, units = "secs"))
#iqr1 <- util_des_functions_env$util_format_duration_human(start = total_seconds)

  iqr1 <-
    util_des_functions_env$util_compute_difftime_auto(iqr1)
  iqr1 <- format(iqr1, usetz = TRUE)

  r<- paste0(iqr1, " (", r, ")")
  rm(q)
  return(list("IQR" = iqr1,
              "IQR_q1_q3" = r))
}


#' Compute the Skewness
#'
#' @name util_compute_skewness
#' @param x a numeric vector
#'
#' @return the Skewness
#' @noRd
util_des_functions_env$util_compute_skewness <- function(x) {
  n <- length(x)
  mean_x <- mean(x, na.rm = TRUE)
  sd_x <- sd(x, na.rm = TRUE)
  skewness <- (sum((x - mean_x)^3, na.rm = TRUE) / (n * sd_x^3))
  return(skewness)
}


#' Compute SE.Skewness
#'
#' @name util_compute_SE_skewness
#' @param x a numeric vector
#'
#' @return the standard error of skewness
#' @noRd
util_des_functions_env$util_compute_SE_skewness <- function(x) {
  n <- length(x)
  se_skewness <- sqrt((6 * n * (n - 1)) /
                        ((n - 2) * (n + 1) * (n + 3)))
  return(se_skewness)
}


#' Compute Kurtosis
#'
#' @name util_compute_kurtosis
#' @param x a numeric vector
#'
#' @return the Kurtosis
#' @noRd
util_des_functions_env$util_compute_kurtosis <- function(x) {
  n <- length(x)
  mean_x <- mean(x, na.rm = TRUE)
  sd_x <- sd(x, na.rm = TRUE)
  kurtosis <- sum((x - mean_x)^4, na.rm = TRUE) / (n * sd_x^4) - 3
  return(kurtosis)
}


#' Compute the frequency of each category
#'
#' @name util_compute_frequency_table
#' @param x a vector
#'
#' @return a string of the categories and their frequency, separated by a pipe symbol
#' @noRd
util_des_functions_env$util_compute_frequency_table <- function(x) {
  # create a dataframe from the vector calculating the frequency of each value
  if(nrow(table(x)) > 1) {
    cont_tab <- as.data.frame(sort(table(x),
                                   decreasing = TRUE,
                                   na.last = FALSE))

  } else {
    cont_tab <- as.data.frame(table(x), na.last = FALSE)
  }

  cont_tab[[1]] <- sQuote(cont_tab[[1]]) # to prevent values from being treated as numeric
  # rename the table
  colnames(cont_tab) <- c("Value", "Freq")
  r <- prep_deparse_assignments(labels = cont_tab$Value,
                                cont_tab$Freq,
                                mode = "string_codes")
  r
}


#' Compute a distribution plot for categorical variables
#'
#' @name util_compute_graph_cat
#' @param x a vector
#'
#' @return  the plot in html code
#' @noRd
util_des_functions_env$util_compute_graph_cat <- function(x) {
  mrg <- 2
  r <- htmltools::plotTag({
    withr::local_par(list(
      mar = rep(mrg, 4),
      oma = rep(0, 4)
    ))
    barplot(table(x), main = NULL,
            xlab = NULL,
            ylab = NULL)},
    width = 250,
    height = 200,
    alt = paste("Histogram"))
  r <- as.character(r)
  r <- paste0('<div style="min-width: 300px">', r, "</div>")
  r
}


#' Compute a distribution plot for datetime variables
#'
#' @name util_compute_graph_datetime
#' @param x a vector
#'
#' @return  the plot in html code
#' @noRd
util_des_functions_env$util_compute_graph_datetime <- function(x) {
  mrg <- 2

  is_time_only <- inherits(x, "hms")
  x0 <- x[!is.na(x)]

  if (length(x0) == 0L) {
    return('<div style="min-width: 300px"><em>no data</em></div>')
  }

  if (is_time_only) {
    # ---- hms: in Sekunden seit Mitternacht, Achse selbst zeichnen ----
    sec <- as.numeric(x0) %% 86400

    br_raw <- suppressMessages(util_optimize_histogram_bins(x = sec,
                                                            nbins_max = 100))
    br <- unique(sort(unlist(br_raw)))
    if (length(br) < 2L) br <- pretty(sec, n = 10)

    # hübsche Ticks (in Sekunden), später als hms formatieren
    at <- pretty(range(br, na.rm = TRUE))

    r <- htmltools::plotTag({
      withr::local_par(list(mar = rep(mrg, 4), oma = rep(0, 4)))

      hist(sec,
           main = NULL, xlab = NULL, ylab = NULL,
           breaks = br, freq = TRUE, xaxt = "n")

      axis(1, at = at, labels = format(hms::as_hms(at)))
    },
    width = 250, height = 200, alt = "Histogram")

    r <- as.character(r)
    paste0('<div style="min-width: 300px">', r, "</div>")

  } else {
    # ---- POSIXct: echte Zeitachse, axis.POSIXct nutzen ----
    if (!inherits(x0, "POSIXt")) {
      util_error(
        "x must be POSIXct/POSIXlt or hms. Internal error, sorry, please report.")
    }

    breaks <- unlist(suppressMessages(util_optimize_histogram_bins(
      x = x,
      nbins_max = 100
    )))
    breaks <- util_parse_date(breaks)
    breaks <- unique(sort(breaks))
    breaks <- scales::pretty_breaks()(breaks)
    r <- htmltools::plotTag({
      withr::local_par(list(
        mar = rep(mrg, 4),
        oma = rep(0, 4)
      ))

      hist(x, main = NULL,
           xlab = NULL,
           ylab = NULL,
           breaks = breaks,
           freq = TRUE)},
      width = 250,
      height = 200,
      alt = paste("Histogram")) #, labs[[rv]]))
    r <- as.character(r)
    paste0('<div style="min-width: 300px">', r, "</div>")
  }
}



#' Compute a distribution plot for continuous variables (excluding datetime)
#'
#' @name util_compute_graph_cont
#' @param x a vector
#'
#' @return  the plot in html code
#' @noRd
util_des_functions_env$util_compute_graph_cont <- function(x) {
  mrg <- 2
  r <- htmltools::plotTag({
    withr::local_par(list(
      mar = rep(mrg, 4),
      oma = rep(0, 4)
    ))
    hist(x, main = NULL,
         xlab = NULL,
         ylab = NULL)},
    width = 250,
    height = 200,
    alt = paste("Histogram")) # of", labs[[rv]]))
  r <- as.character(r)
  r <- paste0('<div style="min-width: 300px">', r, "</div>")
  r

}


#' Move the first row of a data frame to its column names
#'
#' @name util_first_row_to_colnames
#' @param dfr [data.frame]
#'
#' @return [data.frame] with first row as column names
#' @noRd
util_des_functions_env$util_first_row_to_colnames <- function(dfr) {
  colnames(dfr) <- dfr[1, , TRUE]
  dfr <- tail(dfr, -1)
  dfr
}


#' Compute a plot containing the frequency table for categorical variables
#'
#' @name util_compute_graph_frequency_table
#' @param x a vector
#'
#' @return  the plot
#' @noRd
util_des_functions_env$util_compute_graph_frequency_table <- function(x) {
  # create a data frame from the vector calculating the frequency of each value
  if(nrow(table(x)) > 1) {
    cont_tab <- as.data.frame(sort(table(x),
                                   decreasing = TRUE,
                                   na.last = FALSE))

  } else {
    cont_tab <- as.data.frame(table(x), na.last = FALSE)
  }
  # to prevent values from being treated as numeric
  cont_tab[[1]] <- sQuote(cont_tab[[1]])
  # rename the table
  colnames(cont_tab) <- c("Value", "Freq")

  # in case of tables with more than 5 rows, sum last values (less frequent)
  if (nrow(cont_tab) > 5) {
    foot <- tail(cont_tab, -5)
    foot <- c("Others", sum(foot[[2]], na.rm = TRUE))
    cont_tab <- rbind(head(cont_tab, 5), foot)
  }
  # save the formatted table in r
  r <- paste0(util_formattable(util_des_functions_env$util_first_row_to_colnames(
    as.data.frame(t(cont_tab))),
    min_color = c(235, 235, 235),
    max_color = c(20, 20, 235)))
  rm(cont_tab)
  return(r)
}


#' A function to reduce a vector to the only one value that is not NA
#'
#' @name util_combine_cols_content
#' @param rw a vector, e.g., the row of a data frame with only one value and all other NAs
#' @param subject what the data content refers to
#'
#' @return  a value
#' @noRd
util_des_functions_env$util_combine_cols_content <- function(rw, subject) {
  to_keep <- !is.na(rw)
  if (sum(to_keep) > 1) {
    util_warning(c("It should not be possible for",
                   "%s to have more than one",
                   "result per variable -- internal error, please report."),
                 subject)
    NA
  } else {
    result <- rw[to_keep]
    names(result) <- NULL
    if(length(result)==0) {
      result <-  ""
    }
    result
  }
}

#' A function to format from secs to appropriate readable time
#'
#' @name util_compute_difftime_auto
#' @param x a vector
#'
#' @return  a string
#' @noRd
#
util_des_functions_env$util_compute_difftime_auto <- function(time_diff) {
  # check if it is an acceptable object
  if (!inherits(time_diff, "difftime")) {
    util_error(
      "Internal error, sorry. Please report.Input must be a 'difftime' object.")
  }
  # Convert to seconds
  total_seconds <- as.numeric(time_diff, units = "secs")

  # Define thresholds for switching units
  seconds_in_minute <- 60
  seconds_in_hour <- 60 * seconds_in_minute
  seconds_in_day <- 24 * seconds_in_hour
  seconds_in_year <- 365.25 * seconds_in_day
  seconds_in_month <- seconds_in_year / 12

  if (abs(total_seconds) < seconds_in_hour) {
    # If less than an hour, show in minutes (or seconds if very small)
    if (abs(total_seconds) < seconds_in_minute * 2) { # If less than 2 minutes
      return(paste(round(total_seconds, 2), "seconds"))
    } else {
      minutes <- as.numeric(time_diff, units = "mins")
      return(paste(round(minutes, 2), "minutes"))
    }
  } else if (abs(total_seconds) < seconds_in_day * 2) { # If less than 2 days
    hours <- as.numeric(time_diff, units = "hours")
    return(paste(round(hours, 2), "hours"))
  } else if (abs(total_seconds) < seconds_in_month * 2) { # If less than 2 months
    days <- as.numeric(time_diff, units = "days")
    return(paste(round(days, 2), "days"))
  } else if (abs(total_seconds) < seconds_in_year * 2) { # If less than 2 years
    # For months, lubridate's period is better for accuracy if specific dates are involved.
    # But for general display, we can approximate:
    months <- total_seconds / seconds_in_month
    return(paste(round(months, 2), "months"))
  } else {
    years <- total_seconds / seconds_in_year
    return(paste(round(years, 2), "years"))
  }
}


#' A function to format from two dates interval to appropriate readable time
#'
#' @name util_format_duration_human
#' @param x a vector
#'
#' @return  a string
#' @noRd
util_des_functions_env$util_format_duration_human <-
  function(start, end = NULL,
           units = list(
             year    = c("year", "years"),
             month   = c("month", "months"),
             week    = c("week", "weeks"),
             day     = c("day", "days"),
             hour    = c("hour", "hours"),
             minute  = c("minute", "minutes"),
             second  = c("second", "seconds"),
             and     = "and"
           )) {

    # Eingabeinterpretation
    if (!is.null(end)) {
      if (inherits(start, "hms")) {
        intv <- lubridate::interval(start, end, tzone = "UTC")
      } else {
        intv <- lubridate::interval(start, end)
      }
      p <- lubridate::as.period(intv)
    } else {
      if (inherits(start, "difftime")) {
        p <- lubridate::as.period(as.duration(start))
      } else if (inherits(start, "Duration")) {
        p <- lubridate::as.period(start)
      } else if (inherits(start, "Period")) {
        p <- start
      } else {
        util_error(c(
          "Internal error, sorry. Please report: Invalid input type. Provide",
          "two time points or a difftime/duration/period object."))
      }
    }

    # Zerlegung
    years   <- lubridate::year(p)
    months  <- lubridate::month(p)
    days    <- lubridate::day(p)
    hours   <- lubridate::hour(p)
    minutes <- lubridate::minute(p)
    seconds <- lubridate::second(p)
    #round seconds when there are more than 2 decimals
    if (seconds != 0) {
       seconds <- round(seconds, digits = 2)
    }


    # Wochen extrahieren
    weeks <- days %/% 7
    days  <- days %% 7

    # Pluralisierung
    pluralize <- function(value, labels) {
      if (value == 1) paste(value, labels[1])
      else if (value > 0) paste(value, labels[2])
      else NULL
    }

    # Liste bauen
    parts <- c(
      pluralize(years,   units$year),
      pluralize(months,  units$month),
      pluralize(weeks,   units$week),
      pluralize(days,    units$day),
      pluralize(hours,   units$hour),
      pluralize(minutes, units$minute),
      pluralize(seconds, units$second)
    )

    # Ausgabeformat
    if (length(parts) > 1) {
#      paste(paste(parts[-length(parts)], collapse = ", "), units$and, parts[length(parts)])
      paste(paste(parts, collapse = ", "))
    } else if (length(parts) == 1) {
      parts
    } else {
      paste("0", units$second[2])
    }
  }

util_order_of_indicator_metrics <- function(indicator_metrics) {

  abbreviationMetrics <- util_get_concept_info("abbreviationMetrics")
  abbreviationMetrics$order <-
    rank(abbreviationMetrics$order)
  dqi <- util_get_concept_info("dqi")
  dqi$order_nr <-
    rank(dqi$order_nr)
  order_value <-
    vapply(indicator_metrics, FUN.VALUE = integer(1), FUN = function(x) {
      util_stop_if_not(length(x) == 1)
      nm <- strsplit(x, "_", fixed = TRUE)[[1]]
      if (length(nm) >= 2) {

        m <- head(subset(abbreviationMetrics, get("Abbreviation") == nm[[1]],
                         "order", drop = TRUE), 1)

        d <- head(subset(dqi, get("abbreviation") == paste(tail(nm, -1),
                                                           collapse = "_"),
                         "order_nr", drop = TRUE), 1)

        if (length(m) == length(d) && length(d) == 1 &&
            !util_empty(m) && !util_empty(d)) {
          as.integer(d * 100 + m)
        } else {
          NA_integer_
        }
      } else {
        NA_integer_
      }
    })

  order_value[is.na(order_value)] <-
    99999999

  rank(order_value)

}

#' Return all variables in the segment `segment`
#'
#' @param segment [character] the segment as specified in `STUDY_SEGMENT`
#' @param meta_data [data.frame] the metadata
#' @param label_col [character] the metadata attribute used for naming the
#'                              variables
#'
#' @return vector of variable names
#'
#' @family metadata_management
#' @concept metadata_management
#' @noRd


util_get_vars_in_segment <- function(segment, meta_data = "item_level",
                                     label_col = LABEL) {
  util_expect_scalar(segment, check_type = is.character)
  util_expect_data_frame(meta_data, list(
    STUDY_SEGMENT = is.character,
    VAR_NAMES = is.character
  ))
  if (!(LABEL %in% colnames(meta_data))) {
    meta_data[[LABEL]] <- meta_data[[VAR_NAMES]]
  }
  def <- list(
    STUDY_SEGMENT = is.character,
    VAR_NAMES = is.character,
    LABEL = is.character
  )
  def[[label_col]] <- is.character
  util_expect_data_frame(meta_data, def)
  kss <- meta_data[[STUDY_SEGMENT]]
  # kss_lc <- prep_map_labels(kss,
  #                           to = LABEL,
  #                           meta_data = meta_data,
  #                           ifnotfound = kss)
  # kss_lb <- prep_map_labels(kss,
  #                 to = label_col,
  #                 meta_data = meta_data,
  #                 ifnotfound = kss)
  r <-
    unique(c(
      meta_data[kss == segment, label_col, drop = TRUE]#,
#      meta_data[kss_lc == segment, label_col, drop = TRUE],
#      meta_data[kss_lb == segment, label_col, drop = TRUE]
    ))
  sort(r)
}

#' Utility function to parse intervals
#'
#' @param int an interval as string, e.g., "[0;Inf)"
#'
#' @return the parsed interval with elements `inc_l` (Is the lower limit
#' included?), `low` (the value of the lower limit), `inc_u` (Is the upper
#' limit included?), `upp` (the value of the upper limit)
#'
#' @family parser_functions
#' @concept metadata_management
#' @noRd

util_parse_interval <- function(int) {
  if (util_empty(int)) {
    return(NA)
  }
  if (exists(int, .interval_cache)) {
    return(get(int, .interval_cache))
  }
  r <- util_eval_rule(util_parse_redcap_rule(int,
                                             entry_pred = "interval",
                                             must_eof = TRUE))
  if (!inherits(r, "interval") || r$low > r$upp) {
    r <- NA
  }
  assign(int, r, .interval_cache)
  r
}

.interval_cache <- new.env(parent = emptyenv())

#' Normalize and check cross-item-level metadata
#'
#' @param meta_data [meta_data]
#' @param meta_data_cross_item [cross-item-level metadata][meta_data_cross()]
#' @param label_col [character] label column to use for variable naming
#'
#' @return normalized and checked [cross-item-level metadata][meta_data_cross()]
#'
#' @family meta_data_cross
#' @seealso [meta_data_cross()]
#'
#' @noRd
util_normalize_cross_item <-
  function(
    meta_data = "item_level",
    meta_data_cross_item = "cross-item_level",
    label_col = LABEL) {
    # table of specified contradictions
    util_expect_data_frame(meta_data, list(
      VAR_NAMES = is.character
    ))
    util_expect_data_frame(meta_data_cross_item)
    if (!nrow(meta_data_cross_item)) {
      return(data.frame(
        VARIABLE_LIST = character(0),
        CHECK_LABEL = character(0),
        CONTRADICTION_TERM = character(0),
        CONTRADICTION_TYPE = character(0),
        MULTIVARIATE_OUTLIER_CHECKTYPE = character(0),
        N_RULES = integer(0),
        ASSOCIATION_RANGE = character(0),
        ASSOCIATION_METRIC = character(0),
        ASSOCIATION_DIRECTION = character(0),
        ASSOCIATION_FORM = character(0),
        REL_VAL = character(0),
        GOLDSTANDARD = character(0),
        CHECK_ID = character(0),
        DATA_PREPARATION = character(0)
      ))
    }

    if (!CONTRADICTION_TERM %in% colnames(meta_data_cross_item)) {
      meta_data_cross_item[[CONTRADICTION_TERM]] <- rep(NA_character_,
                                                    nrow(meta_data_cross_item))
    }

    # ensure unique identifiers for cross-item checks --------------------------
    if (!CHECK_ID %in% colnames(meta_data_cross_item)) { # as early as possible
      meta_data_cross_item[[CHECK_ID]] <- seq_len(nrow(meta_data_cross_item))
    }
    if (any(duplicated(meta_data_cross_item[[CHECK_ID]]))) {
      util_message(c("cross-item-level metadata: %s must be unique.",
                     "All IDs will be replaced"),
                   dQuote(CHECK_ID),
                   applicability_problem = TRUE)
      meta_data_cross_item[[CHECK_ID]] <- seq_len(nrow(meta_data_cross_item))
    }
    meta_data_cross_item[[CHECK_ID]] <- trimws(meta_data_cross_item[[CHECK_ID]])


    # Find all valid variable names:
    valid_name_cols <- eval(formals(util_find_var_by_meta)$allowed_sources,
                            envir = environment())
    valid_name_cols <- c(valid_name_cols,
                         VAR_NAMES,
                         label_col,
                         LABEL,
                         LONG_LABEL,
                         "ORIGINAL_VAR_NAMES",
                         "ORIGINAL_LABEL")
    lng <- getOption("dataquieR.lang", dataquieR.lang_default)
    if (nzchar(lng)) {
      valid_name_cols <- c(valid_name_cols,
                           paste0(valid_name_cols, "_", lng))
    }
    valid_name_cols <- intersect(unique(trimws(valid_name_cols)),
                                 colnames(meta_data))

    valid_names <- sort(unique(as.vector(unlist(meta_data[, valid_name_cols]))))

    # Expand CONTRADICTION_TERM using util_expand_pattern_rules
    # do before using the term for filling in variable_list
    expanded_contradiction_terms <-
      lapply(meta_data_cross_item[[CONTRADICTION_TERM]], function(rl) {
        r <- util_expand_pattern_rules(rl, valid_names = valid_names)
        if (length(r) == 0)
          r <- rl
        r
      })
    expanded_lengths <- lengths(expanded_contradiction_terms)
    expanded_indices <- sequence(expanded_lengths)
    expanded_source_rows <- rep(seq_len(nrow(meta_data_cross_item)),
                                expanded_lengths)

    meta_data_cross_item <-
      util_explode_var_names_lists(
        expanded_contradiction_terms,
        dframe = meta_data_cross_item,
        col_name = CONTRADICTION_TERM
      )

    expanded_clones <- expanded_lengths[expanded_source_rows] > 1L
    if (any(expanded_clones)) {
      original_check_ids <- meta_data_cross_item[[CHECK_ID]]
      meta_data_cross_item[[CHECK_ID]][expanded_clones] <- paste0(
        original_check_ids[expanded_clones],
        "_",
        expanded_indices[expanded_clones]
      )
      meta_data_cross_item[[CHECK_ID]] <- make.unique(
        meta_data_cross_item[[CHECK_ID]],
        sep = "_"
      )

      if (CHECK_LABEL %in% colnames(meta_data_cross_item)) {
        base_labels <- meta_data_cross_item[[CHECK_LABEL]]
        base_labels[util_empty(base_labels)] <- original_check_ids[
          util_empty(base_labels)]
        meta_data_cross_item[[CHECK_LABEL]][expanded_clones] <- paste0(
          base_labels[expanded_clones],
          " #",
          expanded_indices[expanded_clones]
        )
      }
    }

    if (VARIABLE_LIST %in% colnames(meta_data_cross_item)) {
      # Expand patterns in list column
      meta_data_cross_item[[VARIABLE_LIST]] <-
        unname(vapply(mapply(vl = meta_data_cross_item[[VARIABLE_LIST]],
                             id = meta_data_cross_item[[CHECK_ID]],
                             SIMPLIFY = FALSE,
                      function(vl, id) {
                        r <- util_expand_pattern_rules(
                          paste0("[",
                                 unlist(util_parse_assignments(
                                   vl,
                                   multi_variate_text = TRUE)),
                                 "]"),
                          valid_names = valid_names)
                        if (length(not_found <- attr(r, "mismatches")) > 0) {
                          util_warning(c("Entry VARIABLE_LIST for check %s",
                                         "maybe not specified correctly: %s"),
                                       paste(id),
                                       util_pretty_vector_string(
                                         as.vector(not_found)),
                                       applicability_problem = TRUE)

                        }
                        r <- gsub("^\\[(.*)\\]$", "\\1", trimws(r))
                        r <- paste(r, collapse = SPLIT_CHAR)
                        if (nzchar(r)) {
                          r
                        } else {
                          NA_character_
                        }
                      }), FUN = identity, FUN.VALUE = character(1)))
    }

    if (VARIABLE_LIST %in% colnames(meta_data_cross_item)) {
      meta_data_cross_item[[VARIABLE_LIST_ORDER]] <-
        meta_data_cross_item[[VARIABLE_LIST]]
    }

    # fill empty entries in VARIABLE_LIST and check for possible mismatches
    # between VARIABLE_LIST and CONTRADICTION_TERM -----------------------------
    if (!VARIABLE_LIST %in% colnames(meta_data_cross_item) ||
        any(util_empty(meta_data_cross_item[[VARIABLE_LIST]])) ||
        any(!util_empty(meta_data_cross_item[[CONTRADICTION_TERM]]))) {
      # generate VARIABLE_LIST entries from CONTRADICTION_TERM entries
      # TODO: Support "[" and "]" in variable labels
      needles_var_names <- unique(c(meta_data[[VAR_NAMES]],
                                    meta_data[[label_col]],
                                    meta_data[[LABEL]],
                                    meta_data[[LONG_LABEL]],
                                    meta_data[["ORIGINAL_VAR_NAMES"]],
                                    meta_data[["ORIGINAL_LABEL"]]))
      needles <- paste0("[",
                        needles_var_names,
                        "]")
      x <- vapply(setNames(needles, nm = needles_var_names),
                  grepl,
                  setNames(nm = meta_data_cross_item[[CONTRADICTION_TERM]]),
                  fixed = TRUE,
                  FUN.VALUE = logical(length = nrow(meta_data_cross_item))
      )
      if (is.vector(x)) {
        x <- as.matrix(t(x))
      }
      # variablelist_from_term <- lapply(lapply(lapply(apply(x, 1, which, simplify = FALSE), # apply supports simplify from R 4.1.0
      #                                      names), sort), unique)                # so use the following, less intuitive code line
      variablelist_from_term <- unname(lapply(as.data.frame(t(x)),
                                    function(xx) unique(sort(colnames(x)[xx]))))
      variablelist_from_term <-
        lapply(variablelist_from_term, paste0, collapse = sprintf(" %s ", SPLIT_CHAR))

      vl_from_term_empty <-
        vapply(variablelist_from_term, util_empty, FUN.VALUE = logical(1))

      if (!VARIABLE_LIST %in% colnames(meta_data_cross_item)) {
        meta_data_cross_item[[VARIABLE_LIST]] <- NA
      }

      # Empty VARIABLE_LIST entries can be replaced with the generated lists.
      vl_empty <- util_empty(meta_data_cross_item[[VARIABLE_LIST]])
      meta_data_cross_item[[VARIABLE_LIST]][vl_empty] <- variablelist_from_term[vl_empty]
      # For existing VARIABLE_LIST entries, we have to check whether they match
      # with the generated lists from CONTRADICTION_TERM. This will be done
      # below, after mapping the lists to `label_col`, because then both
      # lists can be compared more easily. We store the generated entries that
      # have to be checked in a temporary column.
      if (!all(seq_len(nrow(meta_data_cross_item)) %in% vl_empty)) {
        meta_data_cross_item[["VARIABLE_LIST_FROM_TERM"]] <- NA
        meta_data_cross_item[[
          "VARIABLE_LIST_FROM_TERM"]][!vl_empty & !vl_from_term_empty] <-
          variablelist_from_term[!vl_empty & !vl_from_term_empty]
      }
    }

    # parse variable lists to `label_col` to ensure uniform naming
    variablelist <- lapply(
        util_parse_assignments(meta_data_cross_item[[VARIABLE_LIST]],
                               multi_variate_text = TRUE
                               ), names)
    variablelist <- lapply(variablelist, function(x) util_find_var_by_meta(
      resp_vars = x,
      meta_data = meta_data,
      label_col = label_col,
      target = label_col,
      ifnotfound = NA_character_))
    # compare with lists from CONTRADICTION_TERM, if necessary
    if ("VARIABLE_LIST_FROM_TERM" %in% colnames(meta_data_cross_item)) {
      vl_from_term_filled <-
        !util_empty(meta_data_cross_item[["VARIABLE_LIST_FROM_TERM"]])
      variablelist_from_term <- lapply(util_parse_assignments(
        meta_data_cross_item[["VARIABLE_LIST_FROM_TERM"]][vl_from_term_filled],
        multi_variate_text = TRUE), names)
      variablelist_from_term <- lapply(variablelist_from_term, function(x) util_find_var_by_meta(
        resp_vars = x,
        meta_data = meta_data,
        label_col = label_col,
        target = label_col,
        ifnotfound = NA_character_))
      variablelist[vl_from_term_filled] <- mapply(
             vl = variablelist[vl_from_term_filled],
             vl_from_term = variablelist_from_term,
             id = meta_data_cross_item[[CHECK_ID]][vl_from_term_filled],
             SIMPLIFY = FALSE,
             FUN = function(vl, vl_from_term, id) { # TODO: Still needed after util_expand_pattern_rules
               not_in_vl <- vl_from_term[!(vl_from_term %in% vl)]
               not_in_vl_from_term <- vl[!(vl %in% vl_from_term)]
               if (length(not_in_vl) > 0 || length(not_in_vl_from_term) > 0) {
                 msg <- paste("Entry VARIABLE_LIST for check",
                              id,
                              "was not specified correctly and adapted:")
                 if (length(not_in_vl_from_term) > 0) {
                   msg <- paste(msg, "Variable(s)",
                                paste(not_in_vl_from_term, collapse = ", "),
                                ifelse(length(not_in_vl_from_term) > 1, "were", "was"),
                                "not used by the", sQuote(CONTRADICTION_TERM))
                 }
                 if (length(not_in_vl) > 0) {
                   if (length(not_in_vl_from_term) > 0) {
                     msg <- paste0(msg, ". Also,")
                   }
                   msg <- paste(msg, "Variable(s)",
                                paste(not_in_vl, collapse = ", "),
                                ifelse(length(not_in_vl) > 1, "were", "was"),
                                "missing from", sQuote(VARIABLE_LIST))
                 }
                 util_warning(msg, applicability_problem = FALSE)
                 # replace variable list with the generated one
                 return(vl_from_term)
               }
               return(vl)
             })
      # delete the temporary column
      meta_data_cross_item <-
        meta_data_cross_item[, -which(colnames(meta_data_cross_item) ==
                                        "VARIABLE_LIST_FROM_TERM")]
    }
    meta_data_cross_item[[VARIABLE_LIST]] <-
      lapply(variablelist, paste, collapse = sprintf(" %s ", SPLIT_CHAR))

    # ensure unique labels for the cross-item checks ---------------------------
    if (!CHECK_LABEL %in% colnames(meta_data_cross_item)) {
      meta_data_cross_item[[CHECK_LABEL]] <- meta_data_cross_item[[CHECK_ID]]
    }
    meta_data_cross_item[[CHECK_LABEL]][
      is.na(meta_data_cross_item[[CHECK_LABEL]])] <- paste0(
        "Check #",
        seq_along(meta_data_cross_item[[CHECK_LABEL]])[
          is.na(meta_data_cross_item[[CHECK_LABEL]])])
    if (any(duplicated(meta_data_cross_item[[CHECK_LABEL]]))) {
      util_message(c("Check labels cannot have duplicates in",
                   "cross-item_level metadata. I'll fix that"),
                   applicability_problem = TRUE)
      while(any(duplicated(meta_data_cross_item[[CHECK_LABEL]]))) {
        meta_data_cross_item[[CHECK_LABEL]][
          duplicated(meta_data_cross_item[[CHECK_LABEL]])] <- paste0(
            "Check #",
            seq_along(meta_data_cross_item[[CHECK_LABEL]])[
              duplicated(meta_data_cross_item[[CHECK_LABEL]])])

      }
    }

    # check DATA_PREPARATION ---------------------------------------------------
    set <- character(0)
    if ((VALUE_LABELS %in% colnames(meta_data) &&
         !all(util_empty(meta_data[[VALUE_LABELS]]))) ||
        (VALUE_LABEL_TABLE %in% colnames(meta_data) &&
         !all(util_empty(meta_data[[VALUE_LABEL_TABLE]]))) # ||
        # (STANDARDIZED_VOCABULARY_TABLE %in% colnames(meta_data) &&
        #  !all(util_empty(meta_data[[STANDARDIZED_VOCABULARY_TABLE]])))
        # FIXME: Enable for this case, too
    ) {
      set <- c(set, "LABEL")
    }

    set <- c(set, "MISSING_NA") # MISSING_LABEL MISSING_INTERPRET

    set <- c(set, "LIMITS")

    default_data_preparation <- paste(set, collapse = sprintf(" %s ",
                                                              SPLIT_CHAR))

    if (!DATA_PREPARATION %in% colnames(meta_data_cross_item)) {
      meta_data_cross_item[[DATA_PREPARATION]] <-
        rep(default_data_preparation, nrow(meta_data_cross_item))
    } else {
      meta_data_cross_item[[DATA_PREPARATION]][
        util_empty(meta_data_cross_item[[DATA_PREPARATION]])] <-
        rep(default_data_preparation,
            sum(util_empty(meta_data_cross_item[[DATA_PREPARATION]])))
    }

    meta_data_cross_item[[DATA_PREPARATION]] <- trimws(
      toupper(meta_data_cross_item[[DATA_PREPARATION]]))

    meta_data_cross_item[[DATA_PREPARATION]] <-
      gsub(
        "\\s+",
        " ",
        perl = TRUE,
        meta_data_cross_item[[DATA_PREPARATION]])

    unknown_tags <- lapply(lapply(
      util_parse_assignments(meta_data_cross_item[[DATA_PREPARATION]],
                             multi_variate_text = TRUE), names),
      FUN = function(x) { x[!(x %in%
                                c("LABEL",
                                  "MISSING_NA",
                                  "MISSING_LABEL",
                                  "MISSING_INTERPRET",
                                  "LIMITS" ))] } )

    meta_data_cross_item[[DATA_PREPARATION]] <- lapply(lapply(
      util_parse_assignments(meta_data_cross_item[[DATA_PREPARATION]],
                             multi_variate_text = TRUE), names),
      FUN = function(x) {
          tags <-
            x[(x %in%
                 c("LABEL",
                   "MISSING_NA",
                   "MISSING_LABEL",
                   "MISSING_INTERPRET",
                   "LIMITS" ))]
          if (any(duplicated(tags))) {
            util_message(
              c("Found duplicated %s tags in cross-item_level",
                "metadata cells. I'll ignore the duplicates."),
              sQuote(DATA_PREPARATION),
              applicability_problem = TRUE
            )
            tags <- unique(tags)
          }
          paste(tags, collapse = sprintf(" %s ", SPLIT_CHAR))
        }
    )

    meta_data_cross_item[[DATA_PREPARATION]][
      util_empty(meta_data_cross_item[[DATA_PREPARATION]])] <-
      sprintf(" %s ", SPLIT_CHAR)

    dupl_tags <- vapply(lapply(
      util_parse_assignments(meta_data_cross_item[[DATA_PREPARATION]],
                                 multi_variate_text = TRUE), names),
      FUN.VALUE = logical(1),
      FUN = function(x) { sum(startsWith(x,
          "MISSING_")) > 1 } )

    meta_data_cross_item$VARIABLE_LIST <-
      vapply(meta_data_cross_item$VARIABLE_LIST, paste, collapse = sprintf(" %s ", SPLIT_CHAR), FUN.VALUE = character(1))

    meta_data_cross_item$DATA_PREPARATION <-
      vapply(meta_data_cross_item$DATA_PREPARATION, paste, collapse = sprintf(" %s ", SPLIT_CHAR), FUN.VALUE = character(1))

    mapply(unknown_tags,
           paste0(
             "#",
             meta_data_cross_item[[CHECK_ID]],
             ": ",
             meta_data_cross_item[[CHECK_LABEL]],
             " (",
             meta_data_cross_item[[VARIABLE_LIST]],
             ")"
           ),
           SIMPLIFY = FALSE,
           FUN = function(tags, name) {
             if (length(tags) > 0)
               util_message(
                 c("In cross-item_level metadata, I found unknown %s tags: %s",
                   "Will ignore such tags."),
                 sQuote(DATA_PREPARATION),
                 util_pretty_vector_string(tags),
                 applicability_problem = TRUE)
           })

    if (any(dupl_tags)) {
      util_message(
        c("Found the following contradicting %s in %s:\n\n",
          "%s\n\nThese will be replaced by the default (%s",
          "default depends on %s),",
          "which may cause",
          "non-functional rules."),
        sQuote(DATA_PREPARATION),
        "cross-item_level metadata",
        paste0(
          "\t#",
          meta_data_cross_item[dupl_tags, CHECK_ID],
          ": ",
          meta_data_cross_item[dupl_tags, CHECK_LABEL],
          " (",
          meta_data_cross_item[dupl_tags, VARIABLE_LIST],
          "): ",
          meta_data_cross_item[dupl_tags, DATA_PREPARATION],
          collapse = "\n"
        ),
        default_data_preparation,
        "item_level metadata",
        applicability_problem = TRUE
      )
      meta_data_cross_item[dupl_tags, DATA_PREPARATION] <-
        default_data_preparation
    }

    # Check *_LIMITS
    for (lim in c(HARD_LIMITS, SOFT_LIMITS, DETECTION_LIMITS)) {
      if (lim %in% colnames(meta_data_cross_item)) {
        meta_data_cross_item[[lim]] <- vapply(
          meta_data_cross_item[[lim]],
          FUN.VALUE = character(1),
          FUN = function(rl) {
            rr <- withCallingHandlers(
              if (!util_empty(rl)) {
                if (length(util_parse_redcap_rule(rl)) > 0) {
                  rl
                } else {
                  NA_character_
                }
              } else {
                NA_character_
              },
              warning = function(cnd) {
                rl <- NA_character_
                m <- conditionMessage(cnd)
                util_warning("Cannot parse %s in cross item level: %s -- %s",
                             dQuote(lim),
                             sQuote(m),
                             dQuote(rl),
                             applicability_problem = TRUE
                             )
                invokeRestart("muffleWarning")
              }
            )
            rr
          }
        )
      }
    }

    # TODO: check the other columns
    meta_data_cross_item
}

#' Get data frame for a given segment
#'
#' @inheritParams .template_function_indicator
#' @param segment [character] name of the segment to return data for
#'
#' @return [data.frame] the data for the segment
#' @export
prep_get_study_data_segment <- function(
    segment,
    study_data,
    item_level = "item_level",
    meta_data = item_level,
    meta_data_v2,
    segment_level,
    meta_data_segment = "segment_level") {
  # Preps and checks ----
  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()
  util_expect_scalar(segment, check_type = is.character,
                     error_message =
                       sprintf("Need a segment name in %s", sQuote("segment")))
  meta_data_segment <- try(prep_check_meta_data_segment(meta_data_segment),
                           silent = TRUE)
  if (util_is_try_error(meta_data_segment)) {
    util_warning("%s metadata missing/corrupted: %s",
                 sQuote("segment_level"),
                 conditionMessage(attr(meta_data_segment, "condition",
                                       exact = TRUE)))
    meta_data_segment <- data.frame()
  }
  # map metadata to study data
  prep_prepare_dataframes(.allow_empty =  TRUE,
                          .adjust_data_type = FALSE,
                          .replace_hard_limits = FALSE,
                          .amend_scale_level = FALSE,
                          .replace_missings = FALSE,
                          .label_col = VAR_NAMES)
  if (!(STUDY_SEGMENT %in% colnames(meta_data))) {
    meta_data[[STUDY_SEGMENT]] <- "ALL" # TODO: Warn
  }
  util_ensure_in(applicability_problem = TRUE, error = TRUE,
    segment, meta_data[[STUDY_SEGMENT]],
    sprintf("Did not find segment %s in item_level metadata %s",
            dQuote(segment), sQuote(STUDY_SEGMENT)
    )
  )

  id_vars <- meta_data_segment[meta_data_segment[[STUDY_SEGMENT]] == segment,
                               SEGMENT_ID_VARS,
                               drop = TRUE]
  if (length(id_vars) != 1) {
    util_warning("No ID-vars in segment level metadata for segment %s found",
                 sQuote(segment), applicability_problem = TRUE)
    id_vars <- character(0)
  } else {
    id_vars <-
      util_parse_assignments(id_vars)[[1]]
    id_vars <- util_find_var_by_meta(id_vars,
                                     meta_data = meta_data,
                                     label_col = label_col,
                                     target = VAR_NAMES,
                                     ifnotfound = id_vars)
    id_vars <- intersect(colnames(ds1), id_vars)
  }

  util_remove_empty_rows(
    id_vars = id_vars,
    ds1[, c(id_vars, util_get_vars_in_segment(segment,
                             meta_data = meta_data,
                             label_col = label_col)), FALSE]
  )
}

test_that("prep_get_study_data_segment works", {
  skip_on_cran() # needs online access
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  sds <-
    prep_get_study_data_segment("LAB", "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                                meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  expect_snapshot_value(sds, style = "json2")
})

#' Check, if `x` contains valid missing codes
#'
#' @param x a vector of values
#'
#' @return `TRUE` or `FALSE`
#' @family robustness_functions
#' @noRd
util_is_valid_missing_codes <- function(x) {
  # TODO: Use it everywhere, where we still have literally defined checks, currently.
  if (suppressWarnings(all(na.rm = TRUE,
          is.na(x) == is.na(as.numeric(x))))) {
    return(TRUE)
  }
  if (all(na.rm = TRUE, is.na(x) == suppressWarnings(
    is.na(util_parse_date(x))))) {
    return(TRUE)
  }
  if (all(na.rm = TRUE, is.na(x) == suppressWarnings(
    is.na(util_parse_time(x))))) {
    return(TRUE)
  }
  return(FALSE)
}

test_that("util_is_valid_missing_codes works", {
  skip_on_cran()
  # Only NA values
  expect_true(util_is_valid_missing_codes(c(NA, NA, NA)))

  # Mixed NA and numeric values
  expect_true(util_is_valid_missing_codes(c(NA,  1,  2)))

  # Character values that are numeric strings
  expect_false(util_is_valid_missing_codes(c("NA", "1", "2")))

  # Character values that cannot be coerced to numeric
  expect_false(util_is_valid_missing_codes(c("NA", "abc", "def")))

  # Date strings that can be converted to datetime
  expect_true(util_is_valid_missing_codes(c("2024-01-01  12:00:00", "2024-01-02  13:00:00")))

  # Date strings that cannot be converted to datetime without warnings
  expect_false(util_is_valid_missing_codes(c("not a date", "2024-13-01  12:00:00")))
})

#' All indicator functions of `dataquieR`
#' @return [character] names of all indicator functions
#'
#' @noRd
util_all_ind_functions <- function() {
  all_inds <-
    lapply(paste0(names(dims), '_'),
           startsWith,
           x = getNamespaceExports(utils::packageName()))
  all_inds <- Reduce(`|`, all_inds)
  getNamespaceExports(utils::packageName())[all_inds]
  # TOOD: Use also, e.g., in util_generate_calls instead of manual names
}

test_that("indic-fkt-args-order", {
  skip_on_cran() # internal conventions

  ind_fkts <- util_all_ind_functions()

  .desired_order <-
    c(
      "resp_vars",
      "group_vars",
      "time_vars",
      "co_vars",
      "study_data",
      "label_col",
      "item_level",
      "meta_data",
      "meta_data_v2"#,
      # ...
    )

  # skip_if_not_installed("devtools")
  # devtools::document()
  # util_load_manual(TRUE)
  mismatch <- !vapply(FUN.VALUE = logical(1),
                     setNames(nm = ind_fkts), function(fkt) {
    desired_order <- .desired_order
    fmls <- names(formals(fkt))
    if ("resp_vars" %in% fmls &&
        which(fmls == "resp_vars") != 1) { # resp_vars can only once in the argument list
      util_message("%s should be the 1st arg in %s",
                   dQuote("resp_vars"),
                   dQuote(fkt))
      return(TRUE)
    } else if (!("resp_vars" %in% fmls)) {
      # if the function does not work on item_level,
      # the _vars arguments are less important
      desired_order <- desired_order[!endsWith(desired_order,
                                               "_vars")]
      desired_order <- setdiff(desired_order, "label_col")
    }
    arg_order <-
      match(fmls, desired_order)
    arg_order <-
      arg_order[!is.na(arg_order)]
    # order of arguments should follow the order in desired_order
    diff_arg_order <- diff(rank(arg_order))
    all_args_in_order <- all(diff_arg_order == 1, na.rm = TRUE)
    if (!all_args_in_order) {
      util_message("args not in order: %s ; Function: %s",
                   prep_deparse_assignments(match(fmls, desired_order),
                                            fmls),
                   dQuote(fkt))
    }
    all_args_in_order
  })
  functions_with_wrong_arg_order <-
    names(which(mismatch))
  expect_equal(functions_with_wrong_arg_order, character(0))
})

test_that("indic-fkt-args-miss", {
  skip_on_cran() # internal conventions

  ind_fkts <- util_all_ind_functions()

  all_or_none <- list(
    c("meta_data", "item_level"),
    c("meta_data_segment", "segment_level"),
    c("meta_data_dataframe", "dataframe_level"),
    c("meta_data_cross_item", "cross-item_level","cross_item_level")
    # TODO: computed items
  )

  names(all_or_none) <- vapply(all_or_none, `[`, 1, FUN.VALUE = character(1))

  mismatch <- vapply(FUN.VALUE = logical(1),
                      setNames(nm = ind_fkts), function(fkt) {
                        fmls <- names(formals(fkt))
                        any(vapply(all_or_none, function(aon) {
                          any(aon %in% fmls) && !all(aon %in% fmls)
                        }, FUN.VALUE = logical(1)))
                      })
  functions_incomplete_arg_list <-
    names(which(mismatch))
  expect_equal(functions_incomplete_arg_list, character(0))
})

test_that("indic-fkt-args-have-mdv2", {
  skip_on_cran() # internal conventions

  ind_fkts <- util_all_ind_functions()

  mismatch <- vapply(FUN.VALUE = logical(1),
                     setNames(nm = ind_fkts), function(fkt) {
                       fmls <- names(formals(fkt))
                       !"meta_data_v2" %in% fmls
                     })
  functions_no_meta_data_v2 <-
    names(which(mismatch))
  expect_equal(functions_no_meta_data_v2, character(0))

  mismatch <- vapply(FUN.VALUE = logical(1),
                     setNames(nm = ind_fkts), function(fkt) {
                       fmls <- names(formals(fkt))
                       if ("meta_data_v2" %in% fmls) {
                         bd <- paste0(as.character(body(fkt)),
                                      collapse = "\n")
                         !grepl("(^|\n)\\s*util_maybe_load_meta_data_v2\\(",
                                bd,
                                perl = TRUE)
                       } else {
                         FALSE
                       }
                     })
  functions_w_meta_data_v2_wo_util_maybe_load_meta_data_v2 <-
    names(which(mismatch))
  expect_equal(functions_w_meta_data_v2_wo_util_maybe_load_meta_data_v2,
               character(0))

})

#' Try hard, to map a variable
#'
#' does not warn on ambiguities nor if not found (but in the latter case,
#' it returns `ifnotfound`)
#'
#' @param resp_vars variables to map from
#' @param meta_data metadata
#' @param allowed_sources allowed names to map from (as metadata columns)
#' @param target metadata attribute to map to
#' @param label_col label-col to map from, if not `allowed_sources` should be
#'                  entirely passed
#' @param ifnotfound [list] A list of values to be used if the item is not
#'                          found: it will be coerced to a list if necessary.
#'
#' @return vector of mapped target names of resp_vars
#'
#' @family metadata_management
#' @concept metadata_management
#' @noRd
util_find_var_by_meta <- function(resp_vars,
                                  meta_data = "item_level",
                                  label_col = LABEL,
                                  allowed_sources = c(VAR_NAMES,
                                                      label_col,
                                                      LABEL,
                                                      LONG_LABEL,
                                                      "ORIGINAL_VAR_NAMES",
                                                      "ORIGINAL_LABEL"),
                                  target = VAR_NAMES,
                                  ifnotfound = NA_character_) {
  util_expect_scalar(resp_vars, allow_more_than_one = TRUE,
                     allow_na = TRUE, allow_null = TRUE,
                     check_type = is.character)
  util_expect_scalar(ifnotfound, allow_more_than_one = TRUE, allow_na = TRUE,
                     allow_null = TRUE)
  if (length(ifnotfound) == 0) {
    ifnotfound <- NA_character_
  }
  if (!(length(ifnotfound) %in% c(1, length(resp_vars))))
    util_error("%s must have either length 1 or the same length as %s",
               sQuote("ifnotfound"), sQuote("resp_vars"))
  if (length(ifnotfound) == 1) ifnotfound <- rep(ifnotfound, length(resp_vars))
  if (all(is.na(resp_vars))) {
    return(ifnotfound)
  }
  util_expect_data_frame(meta_data)
  if (!any(allowed_sources %in% colnames(meta_data))) {
    util_error(c("Did",
                 "not find any mappable column in item-level metadata."))
  }

  # This is not needed, since labels have not significantly changed
  # meta_data <- prep_meta_data_v1_to_item_level_meta_data(meta_data = meta_data,
  #                                                        verbose = FALSE,
  #                                                        label_col = label_col)
  allowed_sources <- unique(allowed_sources)
  allowed_sources <- intersect(allowed_sources, colnames(meta_data))
  target <- util_ensure_in(target, colnames(meta_data))
  if ((length(target) != 1) || (length(allowed_sources) < 1)) {
    map_res <- lapply(setNames(nm = allowed_sources), function(oc) {
      rep(NA_character_, length(resp_vars))
    })
  } else {
    map_res <- lapply(setNames(nm = allowed_sources), function(oc) {
      unname(
        util_map_labels(
          resp_vars,
          warn_ambiguous = FALSE,
          meta_data = meta_data,
          to = target,
          from = oc,
          ifnotfound = NA_character_
        )
      )
    })
  }
  map_res <- as.data.frame(map_res)
  # check whether the user-specified variable could not be found in any of the other columns:
  map_res_NA <- vapply(
    seq_len(nrow(map_res)), # TODO: Use seq_len in all places, where this mapping is used (likely, util_correct_variable_use and prep_meta_data_v1_to_item_level_meta_data)
    FUN.VALUE = logical(1),
    FUN = function(i) {
      all(is.na(map_res[i, ]))
    }
  )
  r <- vapply(
    seq_len(nrow(map_res)),
    FUN.VALUE = character(1),
    FUN = function(i) {
      unique(na.omit(as.character(map_res[i, ])))[1]
      # We select here the first element, such that in case of ambiguities
      # VAR_NAME is preferred over LABEL and LONG_LABEL, and LABEL over LONG_LABEL.
      # If there are no ambiguities, this works as well.
    }
  )
  not_found <- !util_empty(resp_vars) & is.na(r)
  r[not_found] <- ifnotfound[not_found]
  r
}

#' Fix metadata duplicates
#'
#' if [VAR_NAMES] have duplicates, maybe, it's because of ID-vars assigned
#' to different study segments multiple times (they should be in one "intro"-
#' segment, only), which is not the intended use of `STUDY_SEGMENT`.
#' Naturally, they will be part of more than one data-frame, so
#' this would also qualify for a dump duplicate, only, which can safely be
#' removed. Only ID-vars are by default assumed to have such duplicates in item
#' level metadata allowed.
#'
#' @inheritParams .template_function_indicator
#'
#' @returns [meta_data]
#' @export
#' @examples
#' \dontrun{
#' il <- prep_get_data_frame("item_level")
#' il <- rbind(il, il)
#' il$STUDY_SEGMENT[2] <- "X"
#' il2 <- prep_fix_meta_id_dups(meta_data_v2 = "meta_data_v2", item_level = il)
#' il2$STUDY_SEGMENT
#' il$STUDY_SEGMENT[3] <- "X"
#' il3 <- prep_fix_meta_id_dups(meta_data_v2 = "meta_data_v2", item_level = il)
#' il3$STUDY_SEGMENT
#' }
prep_fix_meta_id_dups <- function(meta_data_segment = "segment_level",
                                  meta_data_dataframe =
                                    "dataframe_level",
                                  item_level = "item_level",
                                  meta_data = item_level,
                                  meta_data_v2,
                                  segment_level,
                                  dataframe_level) {
  # Preps and checks ----
  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()
  util_expect_data_frame(meta_data)
  meta_data <- meta_data[!duplicated(meta_data), , FALSE]
  prep_check_meta_names(meta_data = meta_data, level = REQUIRED)
  duplicated_vn <- duplicated(meta_data[[VAR_NAMES]])
  if (!any(duplicated_vn)) return(meta_data);

  meta_data_dataframe <- prep_check_meta_data_dataframe(meta_data_dataframe)
  meta_data_segment <- prep_check_meta_data_segment(meta_data_segment)
  if (!DF_ID_VARS %in% names(meta_data_dataframe))
    meta_data_dataframe[[DF_ID_VARS]] <- NA_character_
  if (!SEGMENT_ID_VARS %in% names(meta_data_segment))
    meta_data_segment[[SEGMENT_ID_VARS]] <- NA_character_
  all_id_vars <-
    trimws(unique(unlist(util_parse_assignments(multi_variate_text = TRUE,
                                                      c(
    as.vector(unlist(meta_data_segment[, SEGMENT_ID_VARS, drop = TRUE])),
    as.vector(unlist(meta_data_dataframe[, DF_ID_VARS, drop = TRUE]))
  )))))
  all_id_vars <- all_id_vars[!is.na(all_id_vars)]

  all_id_vars <- util_find_var_by_meta(all_id_vars,
                                       meta_data = meta_data)

  duplicated_il_wo <- duplicated(meta_data[, setdiff(colnames(meta_data),
                                                     c(STUDY_SEGMENT,
                                                       DATAFRAMES))])
  if (any(duplicated_vn & !duplicated_il_wo)) {
    util_error( # FIXME: With dq_report_by and fixing colnames(of_affected_study_data), too.
      "Fixing of different variables using the same name not yet supported",
      applicability_problem = TRUE)
  } else {
    dup_names <- as.vector(meta_data[duplicated_vn, VAR_NAMES, TRUE])
    if (!all(dup_names %in% all_id_vars)) {
      util_warning(
        c("Some none-id-vars are duplicated but differ only in %s or %s",
          "will fix that, but you may have very inconsistent metadata"),
        sQuote(STUDY_SEGMENT),
        sQuote(DATAFRAMES), applicability_problem = TRUE)
    } else {
      util_message("Overwriting inconsistent parts of metadata...", # TODO: Better message
                   applicability_problem = TRUE)
    }
    meta_data[meta_data[[VAR_NAMES]] %in% dup_names,
              intersect(colnames(meta_data),
                                       c(DATAFRAMES,
                                         STUDY_SEGMENT))] <-
      "**INTRO**" ## TODO: Better name
    r <- meta_data[!duplicated_vn, , FALSE]
    return(r)
  }
}

#' Utility function ensuring valid labels and variable names
#'
#' Valid labels should not be empty, be unique and do not exceed a certain
#' length.
#'
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#'
#' @return a list containing the study data, possibly with adapted column names,
#'         the metadata, possibly with adapted labels, and a string and a
#'         table informing about the changes
#'
#' @noRd
util_ensure_label <- function(meta_data, label_col) { # TODO: Add somehow also to util_validate_known_meta, after VAR_NAMES are not changed any more
  util_stop_if_not(label_col %in% colnames(meta_data))
  if (grepl("^LONG_", label_col)) {
    max_label_len <- MAX_LONG_LABEL_LEN
  } else {
    max_label_len <- MAX_LABEL_LEN
  }
  res <- .util_ensure_label(meta_data = meta_data,
                            label_col = label_col,
                            max_label_len = max_label_len)
  for (lc in setdiff(unique(intersect(colnames(meta_data), c(
    LABEL,
    LONG_LABEL,
    grep("^LABEL_", colnames(meta_data), value = TRUE),
    grep("^LONG_LABEL_", colnames(meta_data), value = TRUE)
    ))), label_col)) {
    if (grepl("^LONG_", lc)) {
      max_label_len <- MAX_LONG_LABEL_LEN
    } else {
      max_label_len <- MAX_LABEL_LEN
    }
    cur_res <- .util_ensure_label(meta_data = meta_data,
                       label_col = lc,
                       max_label_len = max_label_len)
    if (lc == VAR_NAMES) {
      lc <- cur_res$label_col
      if (label_col == VAR_NAMES) {
        res$label_col <- cur_res$label_col
      }
    }
    res$meta_data[[lc]] <-
      cur_res$meta_data[[lc]]
    if (!grepl(cur_res$label_modification_text,
               res$label_modification_text,
               fixed = TRUE)) {
      res$label_modification_text = paste0(
        res$label_modification_text,
        cur_res$label_modification_text,
        collapse = "\n"
      )
      res$label_modification_table <-
        rbind(res$label_modification_table,
              cur_res$label_modification_table)
    }
  }
  res
}

#' Utility function ensuring valid labels and variable names -- internal version
#'
#' really shortens only label-col
#'
#' @param max_label_len [integer] maximum length for the labels, defaults to
#'                                `r MAX_LABEL_LEN`.
#' @return a list containing the study data, possibly with adapted column names,
#'         the metadata, possibly with adapted labels, and a string and a
#'         table informing about the changes
#'
#' @noRd
#' @inheritParams util_ensure_label
.util_ensure_label <- function(meta_data, label_col,
                              max_label_len = MAX_LABEL_LEN) { # TODO: Add somehow also to util_validate_known_meta, after VAR_NAMES are not changed any more
  util_expect_scalar(label_col, check_type = is.character)
  util_expect_data_frame(meta_data, col_names = c(label_col, VAR_NAMES))

  if (any(util_empty(meta_data[[VAR_NAMES]]))) {
    util_warning(c("Need %s in %s, for all variables, some are missing.",
                 "I'll discard all %s without %s."),
               sQuote(VAR_NAMES),
               sQuote("meta_data"),
               sQuote("meta_data"),
               sQuote(VAR_NAMES),
               applicability_problem = TRUE)
    meta_data <- meta_data[!util_empty(meta_data[[VAR_NAMES]]), , FALSE]
  }
  original <- list(
    meta_data = meta_data,
    label_col = label_col
  )
  if (label_col == VAR_NAMES) {
    i <- 0
    new_label_col <- label_col
    while (new_label_col %in% colnames(meta_data)) {
        i <- i + 1
        new_label_col <- paste0(label_col, "_", i)
    }
    meta_data[[new_label_col]] <-
      meta_data[[label_col]]
    label_col <- new_label_col
  }
  N_label <- length(original$meta_data[[original$label_col]])
  modified_label <- original$meta_data[[original$label_col]]
  label_modification_text <- NULL
  label_modification_table <- data.frame(
    "Label-Column" = rep(original$label_col, N_label),
    "Original label" = original$meta_data[[original$label_col]],
    "Modified label" = vector(mode = "character", length = N_label),
    "Reason" = vector(mode = "character", length = N_label),
    check.names = FALSE
  )

  # check for missing labels ---------------------------------------------------
  if (any(ind_no_label <- util_empty(modified_label))) {
    if (sum(ind_no_label) > 0) {
      util_warning(c("Some variables have no label in %s in %s.",
                     "Labels are required to create a report,",
                     "that's why missing labels will be replaced",
                     "provisionally. Please add missing labels in",
                     "your metadata."),
                   dQuote(label_col),
                   sQuote("meta_data"),
                   applicability_problem = TRUE
      )
      new_label <- meta_data[[VAR_NAMES]][ind_no_label]
    } else {
      new_label <- character(0)
    }

    # check for possible duplication
    while (any(new_label %in% modified_label)) {
      # estimate the required number of digits:
      nn <- ceiling(log10(N_label))
      new_label <- paste("NO LABEL", sprintf(
        paste0("%0", nn + 2, "d"),
        sample(1:10^(nn + 1), # sample a random number
               # the set from which we sample is larger than the number of
               # variables, so even if there are already similar labels in
               # the study data, there are most likely enough unassigned
               # numbers left
               size = sum(ind_no_label),
               replace = FALSE)))
    }

    # apply new labels
    if (sum(ind_no_label) > 0) {
      modified_label[ind_no_label] <- new_label
      label_modification_text <-
        paste0(paste(sQuote(new_label), collapse = ", "),
               ifelse(sum(ind_no_label) > 1,
                      paste(" were introduced for",
                            sum(ind_no_label),
                            "missing labels."),
                      " was introduced for one missing label."))
      label_modification_table$Reason[ind_no_label] <- "No label specified."
    }
  }

  # check for duplicated labels ------------------------------------------------

  if (any(duplicated(modified_label))) {
    dupl_lab <- unique(modified_label[which(duplicated(modified_label))])
    if (length(dupl_lab) > 0) {
      util_warning(c("Some variables have duplicated labels in %s in %s.",
                     "Unique labels are required to create a report,",
                     "that's why duplicated labels will be replaced",
                     "provisionally. Please modify the labels in",
                     "your metadata or select a suitable column."),
                   dQuote(label_col),
                   sQuote("meta_data"),
                   applicability_problem = TRUE
      )
    }
    mod_lab_before <- modified_label

    dps <- util_duplicated_inclding_first(modified_label)
    if (any(dps)) {
      # original$meta_data[[original$label_col]]
      modified_label[dps] <- paste0(original$meta_data[[VAR_NAMES]][dps],
                                    ": ",
                                    modified_label[dps])
    }

    dupl_lab <- unique(modified_label[which(duplicated(modified_label))])

    for (ll in dupl_lab) {
      ii <- which(modified_label == ll)
      ndupl_ll <- length(ii)
      if (ndupl_ll > 2) {
        # label is duplicated more than once, so we need to add numbers
        label_suffix <- c("", # the first occurrence does not get a suffix
                          paste("DUPLICATE", sprintf(
                            paste0("%0",
                                   # estimate the required number of digits:
                                   ceiling(log10(ndupl_ll - 1)),
                                   "d"),
                            seq_len(ndupl_ll - 1))))
      } else {
        label_suffix <- c("", "DUPLICATE")
      }
      new_label <- trimws(paste(ll, label_suffix))
      # check for possible duplication
      while (any(new_label[-1] %in% modified_label)) {
        jj <- which(new_label %in% modified_label)[-1]
        # estimate the required number of digits:
        nn <- ceiling(log10(N_label))
        new_label[jj] <- paste(new_label[jj], sprintf(
          paste0("%0", nn + 2, "d"),
          sample(1:10^(nn + 1), # sample a random number
                 # the set from which we sample is larger than the number of
                 # variables, so even if there are already similar labels in
                 # the study data, there are most likely enough unassigned
                 # numbers left
                 size = length(jj),
                 replace = FALSE)))
      }
      modified_label[ii] <- new_label
    }

    ind_dupl <- which(mod_lab_before != modified_label)
    if (length(ind_dupl) > 0) {
      label_modification_text <- c(
        label_modification_text,
        paste0(
          paste(sQuote(modified_label[ind_dupl]), collapse = ", "),
          ifelse(length(ind_dupl) > 1,
                 " were introduced for duplicated labels.",
                 " was introduced for one duplicated label.")
        ))
      label_modification_table$Reason[ind_dupl] <- trimws(
        paste(label_modification_table$Reason[ind_dupl], "Duplicated label."))
    }
  }

  # abbreviate() warns on non-ASCII characters, by default.
  my_abbreviate <- function(names.arg, method = "both.sides", ...) {  # FIXME: this qualifies for an own function for abbreviating, because we need it for value labels, too
    x <- names.arg
    x_words <- strsplit(x, "\\s+")
    uc_first_words <-
      lapply(x_words, function(x) {
        substr(x, 1, 1) <- toupper(substr(x, 1, 1));
        x
      })
    uc_first_words <-
      vapply(uc_first_words, paste, collapse = " ", FUN.VALUE = character(1))
    suppressWarnings(abbreviate(names.arg = uc_first_words, method = method,
                                ...))
  }

  # abbreviate long labels -----------------------------------------------------
  if (any(nchar(modified_label) > max_label_len)) {
    ind_long_label <- which(nchar(modified_label) > max_label_len)
    if (length(ind_long_label) > 0) {
      util_warning(c("Some variables have labels with more than %d characters",
                     "in %s in %s.",
                     "This will cause suboptimal outputs and possibly also",
                     "failures when rendering the report,",
                     "due to issues with the maximum length of file names",
                     "in your operating system or file system.",
                     "This will be fixed provisionally. Please shorten",
                     "your labels or choose another label column."),
                   max_label_len,
                   dQuote(label_col),
                   sQuote("meta_data"),
                   applicability_problem = TRUE,
                   additional_classes = LONG_LABEL_EXCEPTION
      )
    }

    # remove punctuation from labels, if possible, before abbreviating (gives
    # more readable abbreviations)
    new_label <- unname(my_abbreviate(gsub("[[:punct:]]+", replacement = " ",
                                        modified_label[ind_long_label]),
                                   minlength = max_label_len))
    if (any(util_empty(new_label))) {
      new_label <- unname(my_abbreviate(modified_label[ind_long_label],
                                     minlength = max_label_len))
    }
    if (any(nchar(new_label) > max_label_len)) {
      new_label <- substr(new_label, 1, max_label_len)
    }
    label_modification_table$Reason[ind_long_label] <- trimws(
      paste(label_modification_table$Reason[ind_long_label],
            "The label is too long.")
    )
    # check for possible duplication between abbreviated labels and the
    # other, unchanged labels
    temp_label <- modified_label
    temp_label[ind_long_label] <- new_label
    guard <- 0
    while (any(duplicated(temp_label))) {
      # add the duplicated labels to the selection of labels that should
      # be adapted by 'abbreviate'
      ind_long_label <- c(ind_long_label,
                          which(modified_label %in% new_label))
      ind_long_label <- sort(unique(ind_long_label))
      # rerun the 'abbreviate' function to get shortened, unique labels
      new_label <- trimws(unname(my_abbreviate(gsub("[[:punct:]]+", replacement = " ",
                                          modified_label[ind_long_label]),
                                     minlength = max_label_len)))
      if (any(util_empty(new_label))) {
        new_label <- trimws(unname(my_abbreviate(modified_label[ind_long_label],
                                       minlength = max_label_len)))
      }
      if (any(duplicated(new_label))) {
        new_label <- trimws(unname(my_abbreviate(modified_label[ind_long_label],
                                                 minlength = max_label_len)))
      }
      if (any(duplicated(new_label))) {
        util_message(c("Could not shorten labels to at max. %d",
                       "characters preserving uniqueness. Try to",
                       "allow more characters setting the %s option()",
                       "or amend your labels"),
                     max_label_len, sQuote("dataquieR.MAX_LABEL_LEN"),
                     applicability_problem = TRUE)
        new_label <- make.unique(new_label) # may yield too long labels, but here, there is not better choice
      }
      temp_label[ind_long_label] <- new_label
      guard <- guard + 1
      if (guard > 1000) {
        util_error(c("Internal error: Could not shorten labels to at max. %d",
                     "characters preserving uniqueness. Try to",
                     "allow more characters setting the %s option()",
                     "or amend your labels. I gave up after %d tries (!!).",
                     "This should not happen, sorry, please report."),
                   max_label_len, sQuote("dataquieR.MAX_LABEL_LEN"), 1000,
                   applicability_problem = TRUE)
      }
    }

    modified_label <- temp_label

    if (any(util_empty(label_modification_table$Reason[ind_long_label]))) {
      ind_long_label2 <- intersect(ind_long_label,
                                   which(util_empty(
                                     label_modification_table$Reason)))
      label_modification_table$Reason[ind_long_label2] <- trimws(
        paste(label_modification_table$Reason[ind_long_label2],
              "The label was identical to an abbreviated label.")
      )
    }

    if (any(ind_no_label)) {
      # If we store the original labels or variable names (only needed if they
      # are abbreviated), then we have to replace empty entries, otherwise we
      # will run into errors later with 'util_find_var_by_meta'
      # running 'util_map_labels' for 'ORIGINAL_LABEL'.
      original$meta_data[[original$label_col]][ind_no_label] <- "(missing)"
    }
    meta_data[["ORIGINAL_LABEL"]] <- original$meta_data[[original$label_col]]

    if (length(ind_long_label) > 0) {
      label_modification_text <- c(
        label_modification_text,
        paste0(paste(sQuote(modified_label[ind_long_label]), collapse = ", "),
               ifelse(length(ind_long_label) > 1,
                      paste0(" were introduced as abbreviated labels for ",
                             paste(sQuote(original$meta_data[[original$label_col]][ind_long_label]),
                                   collapse = ", "), "."),
                      paste0(
                        " was introduced as an abbreviation for the label ",
                             sQuote(original$meta_data[[original$label_col]][ind_long_label]), ".")
               )
        )
      )
    }
  }

  # apply modified labels now that all checks are done -------------------------

  if (is.data.frame(meta_data)) {
    meta_data[[label_col]] <- modified_label
  }

  label_modification_table[, "Modified label"] <- modified_label
  # only show entries with modifications
  label_modification_table <- label_modification_table[which(
    label_modification_table[, "Original label"] !=
      label_modification_table[, "Modified label"] |
      !util_empty(label_modification_table[, "Reason"])
  ), ]

  return(list(meta_data = meta_data,
              label_modification_text = trimws(paste(label_modification_text,
                                                       collapse = " ")),
              label_modification_table = label_modification_table,
              label_col = label_col
              ))
}

#' Utility function to compute and optimize bin breaks for histograms
#'
#' @param x a vector of data values (numeric or datetime)
#' @param interval_freedman_diaconis range of values which should be included to
#'               calculate the Freedman-Diaconis bandwidth (e.g., for
#'               `con_limit_deviations` only values within limits) in interval
#'               notation (e.g., `[0;100]`)
#' @param nbins_max the maximum number of bins for the histogram. Strong
#'                  outliers can cause too many narrow bins, which might be
#'                  even to narrow to be plotted. This also results in large
#'                  files and rendering problems. So it is sensible to limit
#'                  the number of bins. The function will produce a message if
#'                  it reduces the number of bins in such a case. Reasons
#'                  could be unspecified missing value codes, or minimum or
#'                  maximum values far away from most of the data values, a few
#'                  number of unique values, or (for `con_limit_deviations`)
#'                  no or few values within limits.
#' @param cuts a vector of values at which breaks between bins should occur
#'
#' @return a list with bin breaks, if needed separated for each segment
#'         of the plot
#'
#' @family figure_functions
#' @concept figure
#' @noRd
util_optimize_histogram_bins <- function(x, interval_freedman_diaconis = NULL,
                                         nbins_max = 100, cuts = NULL) {
  # check and prepare input ----------------------------------------------------
  if (inherits(x, "POSIXlt")) x <- util_parse_date(x)
  util_expect_scalar(x, allow_more_than_one = TRUE, allow_na = TRUE,
                     check_type = function(x) {
    is.numeric(x) || "POSIXct" %in% class(x) || inherits(x, "hms")
  })
  if (any(is.infinite(x))) {
    x[is.infinite(x)] <- NA
  }
  if (length(x) == 0 || all(util_empty(x))) {
    util_error("Nothing to plot!", applicability_problem = TRUE)
  }

  x <- x[!is.na(x)]
  min_x <- min(x)
  max_x <- max(x)
  if (inherits(x, "hms")) {
    min_x <- hms::as_hms(min_x)
    max_x <- hms::as_hms(max_x)
  }

  if (inherits(cuts, "POSIXlt")) cuts <- util_parse_date(cuts)
  util_expect_scalar(cuts, allow_null = TRUE, allow_more_than_one = TRUE,
                     check_type = function(x) {
                       is.numeric(x) || "POSIXct" %in% class(x)
                     })

  util_expect_scalar(interval_freedman_diaconis,
                     allow_null = TRUE, allow_more_than_one = TRUE)
  if (is.null(interval_freedman_diaconis)) {
    #interval_freedman_diaconis <- util_parse_interval(
    #  paste0("[", min_x, ";", max_x, "]"))
    interval_freedman_diaconis <-
      redcap_env$interval(inc_l = TRUE, inc_u = TRUE,
                          low = min_x, upp = max_x)
  } else {
    if (!("interval" %in% class(interval_freedman_diaconis))) {  #TOFIX: if the interval is 110; Inf it stops with an error
      interval_freedman_diaconis <-
        util_parse_interval(interval_freedman_diaconis)
    }
    if (!("interval" %in% class(interval_freedman_diaconis))) {
      util_error("The interval is incorrect.")
    }
    cuts <- c(cuts, interval_freedman_diaconis$low,
              interval_freedman_diaconis$upp)
  }

  min_x <- as.numeric(min_x)
  max_x <- as.numeric(max_x)
  cuts <- as.numeric(cuts)
  cuts <- cuts[which(!is.na(cuts) & !is.infinite(cuts))]
  cuts <- sort(unique(cuts))
  if (length(cuts) > 1) {
    # omit 'cuts' at (practically) the same value (not caught by 'unique')
    cuts <- cuts[cuts[-1] - cuts[-length(cuts)] > .Machine$double.eps]
  }
  if (length(cuts) > 0) {
    if (cuts[1] - min_x >= .Machine$double.eps) {
      cuts <- c(-Inf, cuts)
    }
    if (max_x - cuts[length(cuts)] >= .Machine$double.eps) {
      cuts <- c(cuts, Inf)
    }
  } # Note: 'cuts' can still be null

  min_plot <- min(c(min_x, cuts[!is.infinite(cuts)]))
  max_plot <- max(c(max_x, cuts[!is.infinite(cuts)]))

  util_expect_scalar(nbins_max,
                     check_type = util_is_numeric_in(min = 1, max = 500,
                                                     whole_num = TRUE))

  # step 1: Freedman-Diaconis bandwidth ----------------------------------------
  x_within <- x[redcap_env$`in`(x, interval_freedman_diaconis)]
  # If there are no values within the specified interval, we discard the
  # interval and consider all values instead.
  if (length(x_within) == 0) {
    x_within <- x
    min_within <- min_x
    max_within <- max_x
  } else {
    min_within <- min(x_within)
    if (!is.infinite(interval_freedman_diaconis$low)) {
      min_within <- min(min_within, interval_freedman_diaconis$low)
    }
    min_within <- as.numeric(min_within)
    max_within <- max(x_within)
    if (!is.infinite(interval_freedman_diaconis$upp)) {
      max_within <- max(max_within, interval_freedman_diaconis$upp)
    }
    max_within <- as.numeric(max_within)
  }
  # - calculate span
  dif <- max_within - min_within
  # - calculate bandwidth according to Freedman-Diaconis:
  #   (2 * IQR(data) / length(data)^(1/3))
  iqr_bw <- IQR(x_within)
  if (is.na(iqr_bw)) iqr_bw <- 0
  n_bw <- length(x_within)
  if (n_bw == 0) n_bw <- 1
  bw <- 2 * iqr_bw / n_bw^(1 / 3)
  if (bw == 0) bw <- 1

  # step 2: adjust bandwidth ---------------------------------------------------
  # - calculate the number of bins between the lower and upper limit of the
  #   interval (or between the minimum and maximum value, see above), based on
  #   the calculated bandwidth
  # - round the number of bins within limits to an integer value to obtain bin
  #   breaks at the limits
  # - get the new bandwidth for the rounded number of bins within limits
  nbins_within <- round(dif / bw)
  if (nbins_within == 0) nbins_within <- 1
  byX <- dif / nbins_within

  # - calculate a lower limit for the bandwidth based on the maximum number of
  #   bins
  bw_min <- (max_x - min_x) / nbins_max
  # - for constant variables: dif = 0 => byX = 0 and also bw_min = 0
  # - in this case, we have to provide a good default value for bw_min
  # - for this, we can consider the distances between segments (if any), or
  #   use the 'pretty' function
  if (length(cuts) > 1) {
    # calculate span for each segment
    dif_seq <- cuts[-1] - cuts[-(length(cuts))]
  } else {
    dif_seq <- max_plot - min_plot
  }
  if (bw_min == 0) {
    # consider distance between cuts
    min_dist <- suppressWarnings(
      min(dif_seq[which(!is.infinite(dif_seq) & dif_seq > 0)]))
    # consider proposal from 'pretty' function
    pp <- pretty(c(min_plot, max_plot), min.n = 1)
    bw_min <- min(c(min_dist, pp[2] - pp[1]))
  }
  # - check whether the bandwidth `byX` is not below the minimum bandwidth.
  #   If it is, use 'floor' with the minimum bandwidth to calculate and round
  #   the number of bins within limits.
  #   This approach will fail in keeping the bandwidth at or above 'bw_min',
  #   if 'dif' < 'bw_min' (i.e., the minimum possible bin width is larger than
  #   the range within limits, or, respectively, larger than the range from the
  #   minimum value to the maximum value). In this case, we have to switch to
  #   different bin sizes for the segments and will allow to have a smaller bin
  #   within limits, if needed. For constant variables, the single bar of the
  #   histogram will have width 'bw_min'.
  if (byX < bw_min) {
    nbins_within <- floor(dif / bw_min)
    if (nbins_within == 0) {
      byX <- bw_min
    } else {
      byX <- dif / nbins_within
    }
  }
  # - if all values are integer values, then we have to ensure that the
  #   bandwidth `byX` is an integer >= 1
  if (all(util_is_integer(x))) {
    byX <- max(round(byX), 1)
  }
  # - calculate the number of bins for each segment
  nbins_seq <- vapply(dif_seq, FUN.VALUE = numeric(1), function(dd) {
    ceiling(dd / byX)
  })
  # Note: Due to the forced breaks at the limits, we could have now a few bins
  # more than 'nbins_max'.

  # The forced breaks can change the bandwidth in each segment. In case there
  # is a segment with an open upper/lower limits, we could use the average of
  # the bandwidths from the fixed segments for it, to improve the overall look
  # of the segmented histogram.
  if (length(cuts) > 1) {
    byX_mean <- dif_seq / nbins_seq
    byX_mean <- byX_mean[is.finite(byX_mean)]
    if (length(byX_mean) > 0) {
      byX_mean <- sum(byX_mean) / length(byX_mean)
      if (byX_mean > bw_min) {
        byX <- byX_mean
      }
    }
    if (is.infinite(nbins_seq[1])) {
      nbins_seq[1] <- ceiling(abs(cuts[2] - min_plot) / byX)
    }
    if (is.infinite(nbins_seq[length(nbins_seq)])) {
      nbins_seq[length(nbins_seq)] <-
        ceiling(abs(max_plot - cuts[length(cuts) - 1]) / byX)
    }
  }

  # step 3: get the position of bin breaks -------------------------------------
  if (length(cuts) <= 1) {
    all_breaks <- list(pretty(c(min_plot, max_plot),
                              n = max(nbins_seq, 1), min.n = 1))
    if ("POSIXct" %in% class(x)) {
      all_breaks[[1]] <- util_parse_date(all_breaks[[1]])
    }
  } else {
    all_breaks <- vector(mode = "list", length = length(nbins_seq))
    for (i in seq_along(nbins_seq)) {
      if (is.infinite(cuts[i])) {
        # We ensured by construction that cuts[i] and cuts[i+1] can not both
        # be infinite.
        all_breaks[[i]] <- seq(from = cuts[i+1] - nbins_seq[i] * byX,
                               to = cuts[i+1],
                               length.out = nbins_seq[i] + 1)
      } else if (is.infinite(cuts[i+1])) {
        all_breaks[[i]] <- seq(from = cuts[i],
                               to = cuts[i] + nbins_seq[i] * byX,
                               length.out = nbins_seq[i] + 1)
      } else if (nbins_seq[i] <= 1) {
        all_breaks[[i]] <- c(cuts[i], cuts[i+1])
      } else {
        all_breaks[[i]] <- seq(from = cuts[i],
                               to = cuts[i+1],
                               length.out = nbins_seq[i] + 1)
      }
      if ("POSIXct" %in% class(x)) {
        all_breaks[[i]] <- util_parse_date(all_breaks[[i]])
      }
    }
  }

  return(breaks = all_breaks)
}

#' Remove a specified element from the data frame cache
#'
#' @param object_to_remove [character] name of the object to be removed
#' as character string (quoted), or character vector containing the names of the
#' objects to remove from the cache
#'
#' @return nothing
#'
#' @family data-frame-cache
#' @export
#'
#' @examples
#' \dontrun{
#' prep_load_workbook_like_file("meta_data_v2") #load metadata in the cache
#' ls(.dataframe_environment()) #get the list of dataframes in the cache
#'
#' #remove cross-item_level from the cache
#' prep_remove_from_cache("cross-item_level")
#'
#' #remove dataframe_level and expected_id from the cache
#' prep_remove_from_cache(c("dataframe_level", "expected_id"))
#'
#' #remove missing_table and segment_level from the cache
#' x<- c("missing_table", "segment_level")
#' prep_remove_from_cache(x)
#' }
#'
prep_remove_from_cache <- function(object_to_remove) {
  #if the argument is one object, e.g. "meta_data"
  if(length(object_to_remove) == 1) {
    # check if the object is present in the cache and remove it
    if(object_to_remove %in% ls(.dataframe_environment())){
    rm(list=capture.output(cat(object_to_remove)),
       envir = .dataframe_environment())
    util_message(paste0(object_to_remove, " have been removed from cache"))
    } else { #provide a message is the object is not present in the cache
    util_message(paste0(object_to_remove, " not present in the cache"))
    }
  } else if (length(object_to_remove) > 1) {  #if the argument is a vector
    # check if all objects in the vector are present in the cache and remove them
    if(all(object_to_remove %in% ls(.dataframe_environment()))){
      rm(list=paste(object_to_remove), envir = .dataframe_environment())
      util_message(sprintf("%s have been removed from cache",
                           paste0(object_to_remove, collapse = " & ")))
    } else { # provide a message if not all objects in the vector are present in
      # the cache
      util_message(sprintf("Not all the objects (%s) are present in the cache",
                           paste0(object_to_remove, collapse = ", ")))
    }
  }
}

#' Plot a `ggplot2` figure without `plotly`
#'
#' @param x [ggplot2::ggplot2] object
#' @param sizing_hints `object` additional metadata about the natural figure size
#'
#' @return `htmltools` compatible object
#'
#' @noRd
util_plot_figure_no_plotly <- function(x, sizing_hints = NULL) {
  if (capabilities("cairo") && suppressWarnings(util_ensure_suggested("grDevices",
                                                     "render vector graphics",
                                                     err = FALSE))) {
    # x <- htmltools::plotTag(x, "", suppressSize = "xy", attribs =
    #           list(class = "dataquieRfigure"),
    #           device = grDevices::svg, mimeType = "image/svg+xml",
    #           # width = 800, height = 600,
    #           pixelratio = 1/72
    # )
    # # preserveAspectRatio="none"
    # x$attribs$src
    w <- 800
    h <- 600
    if (!is.null(sizing_hints)) {
      sizing_hints <- util_finalize_sizing_hints(sizing_hints = sizing_hints)
      if (!is.null(sizing_hints$w_in_cm) &&
          !is.null(sizing_hints$h_in_cm)) {
        w <- sizing_hints$w_in_cm / 2.54 * 72
        h <- sizing_hints$h_in_cm / 2.54 * 72
      }
    }
    x <- util_plot_svg_to_uri(x, w = w, h = h)
  } else {
    x <- htmltools::plotTag(x, "", suppressSize = "xy", attribs =
                              list(class = "dataquieRfigure"))#, width = 640)
  }
  x
}

test_that("util_plot_figure_no_plotly works", {
  skip_on_cran()
  # Create a test plot
  test_plot <- plot(cars)
  result <- util_plot_figure_no_plotly(test_plot)

  # Check that the result is an HTML plot tag
  expect_s3_class(result, "shiny.tag")

})

#' Normalize time-only columns to `hms`
#'
#' @description
#' Convert columns that contain *time-only* class values to [`hms::hms()`].
#' Supported time-only classes are:
#' - `chron::times` (fraction of day),
#' - `data.table::ITime` (integer seconds since midnight), and
#' - `calcal::time_of_day` (hour/minute/second record).
#'
#' Columns that are already `hms` are left unchanged. Date-time (`POSIXt`)
#' columns and plain text/numeric columns are not modified.
#'
#' List-columns are handled robustly: if they contain any time-only objects,
#' each element is converted to `hms`; unconvertible scalars become `NA_hms`
#' to avoid downstream failures. Inhomogeneities are reported via
#' `util_message()` / `util_warning()` using integrity indicator
#' `"int_vfe_inhom"`, analogous to existing handlers.
#'
#' @param study_data [data.frame]
#'   Input data frame whose time-only columns should be normalized.
#'
#' @return
#' A modified copy of `study_data` where detected time-only columns are of
#' class `hms`.
#'
#' @examples
#' \dontrun{
#'   d <- data.frame(
#'     t_num   = c(32, 47),
#'     t_chron = chron::times(c("09:10:00", "10:20:30")),
#'     t_text  = c("09:10:00", "10:20:30"),
#'     t_posix = as.POSIXct(c("2020-01-01 09:10:00", "2020-01-01 10:20:30")),
#'     stringsAsFactors = FALSE
#'   )
#'   # Optional if packages are available:
#'   # d$t_itime  <- data.table::as.ITime(c("09:10:00", "10:20:30"))
#'   # d$t_calcal <- calcal::time_of_day(
#'   #   hour = c(9, 10), minute = c(10, 20), second = c(0, 30)
#'   # )
#'   d2 <- util_normalize_time_only_columns(d)
#'   str(d2$t_chron)  # hms
#'   str(d2$t_text)   # character (unchanged)
#'   str(d2$t_posix)  # POSIXct (unchanged)
#' }
#'
#' @noRd
util_normalize_time_only_columns <- function(study_data) {
  util_expect_data_frame(study_data)

  # ---- helpers -------------------------------------------------------------

  is_scalar_na <- function(x) {
    r <- FALSE
    try(r <- length(x) == 1 && is.na(x), silent = TRUE)
    r
  }

  as_hms_na <- function() hms::as_hms(NA_character_)

  # Class predicates
  inherits_times <- function(x) inherits(x, "times")           # chron::times
  inherits_itime <- function(x) inherits(x, "ITime")           # data.table
  inherits_hms   <- function(x) inherits(x, "hms")
  inherits_posix <- function(x) inherits(x, "POSIXt")
  inherits_calcal_tod <- function(x) inherits(x, "time_of_day")  # calcal

  # Convert a single element (possibly from a list-column) to hms
  elem_to_hms <- function(x) {
    if (inherits_hms(x) && length(x) == 1) return(x)
    if (is_scalar_na(x)) return(as_hms_na())

    # chron::times -> fraction of day * 86400
    if (inherits_times(x)) {
      sec <- try(as.numeric(x) * 86400, silent = TRUE)
      if (util_is_try_error(sec) || length(sec) != 1 || is.na(sec)) {
        return(as_hms_na())
      }
      return(hms::hms(sec))
    }

    # data.table::ITime -> integer seconds since midnight
    if (inherits_itime(x)) {
      sec <- try(as.integer(x), silent = TRUE)
      if (util_is_try_error(sec) || length(sec) != 1 || is.na(sec)) {
        return(as_hms_na())
      }
      return(hms::hms(sec))
    }

    # calcal::time_of_day -> record (expect scalar in list-columns)
    if (inherits_calcal_tod(x)) {
      if (length(x) != 1) return(as_hms_na())
      hh <- try(as.numeric(unclass(x)$hour),   silent = TRUE)
      mm <- try(as.numeric(unclass(x)$minute), silent = TRUE)
      ss <- try(as.numeric(unclass(x)$second), silent = TRUE)
      ok <- !(util_is_try_error(hh) | util_is_try_error(mm) |
                util_is_try_error(ss)) &&
        is.finite(hh) && is.finite(mm) && is.finite(ss)
      if (!ok) return(as_hms_na())
      return(hms::hms(h = hh, m = mm, s = ss))
    }

    # Explicitly exclude POSIXt (date-time) from "time-only"
    if (inherits_posix(x)) return(as_hms_na())

    # Conservative extra: numeric seconds or difftime scalars are acceptable
    if (length(x) == 1 && (is.numeric(x) || inherits(x, "difftime"))) {
      r <- try(hms::as_hms(x), silent = TRUE)
      if (!util_is_try_error(r)) return(r)
    }

    # Everything else -> NA to keep downstream code stable
    as_hms_na()
  }

  # Decide if a whole column counts as "time-only" we should normalize
  is_timeonly_column <- function(x) {
    if (inherits_hms(x))        return(FALSE)
    if (inherits_posix(x))      return(FALSE)
    if (inherits_times(x))      return(TRUE)
    if (inherits_itime(x))      return(TRUE)
    if (inherits_calcal_tod(x)) return(TRUE)
    if (is.character(x))        return(FALSE)
    if (!is.list(x))            return(FALSE)
    any(vapply(
      x,
      function(e)
        inherits_times(e) || inherits_itime(e) || inherits_calcal_tod(e),
      FUN.VALUE = logical(1)
    ))
  }

  emit_log <- function(varname, count_non_hms, lost, is_hms_before) {
    in_pipe <- FALSE
    tmp <- try(.called_in_pipeline2(), silent = TRUE)
    if (!util_is_try_error(tmp)) in_pipe <- isTRUE(tmp)

    vn <- dQuote(if (length(varname)) varname else "data")
    label_time <- if (!is.null(DATA_TYPES$TIME)) sQuote(DATA_TYPES$TIME)
    else sQuote("time")

    if (!in_pipe) {
      .cnd <- if (lost > 0) util_warning else util_message
      .cnd(
        c("Found %d non-hms %s values in %s -- %d of which",
          "could not be casted to hms"),
        count_non_hms, label_time, vn, lost,
        integrity_indicator = "int_vfe_inhom", varname = varname
      )
    } else {
      util_message(
        c("Found %d non-hms %s values in %s"),
        sum(!is_hms_before), label_time, vn,
        integrity_indicator = "int_vfe_inhom", varname = varname
      )
    }
  }

  # ---- main ----------------------------------------------------------------

  for (j in seq_along(study_data)) {
    vname <- names(study_data)[j]
    x <- study_data[[j]]

    if (!is_timeonly_column(x)) next

    # Atomic chron::times vector
    if (inherits_times(x)) {
      sec <- try(as.numeric(x) * 86400, silent = TRUE)
      if (!util_is_try_error(sec)) {
        study_data[[j]] <- hms::hms(sec)
      }
      next
    }

    # Atomic data.table::ITime vector
    if (inherits_itime(x)) {
      sec <- try(as.integer(x), silent = TRUE)
      if (!util_is_try_error(sec)) {
        study_data[[j]] <- hms::hms(sec)
      }
      next
    }

    # Atomic calcal::time_of_day vector (vctrs record)
    if (inherits_calcal_tod(x)) {
      hh <- suppressWarnings(as.numeric(unclass(x)$hour))
      mm <- suppressWarnings(as.numeric(unclass(x)$minute))
      ss <- suppressWarnings(as.numeric(unclass(x)$second))
      ok <- is.finite(hh) & is.finite(mm) & is.finite(ss)
      seconds <- hh * 3600 + mm * 60 + ss
      seconds[!ok] <- NA_real_
      study_data[[j]] <- hms::hms(seconds)
      next
    }

    # List-column containing time-only objects (possibly mixed)
    if (is.list(x)) {
      is_hms_before <- vapply(x, inherits_hms, FUN.VALUE = logical(1))
      is_na_before  <- lapply(x, is_scalar_na)

      x2 <- lapply(x, elem_to_hms)

      is_na_after <- lapply(x2, is_scalar_na)
      lost <- sum(!vapply(
        mapply(is_na_before, is_na_after, FUN = identical, SIMPLIFY = FALSE),
        FUN.VALUE = logical(1), FUN = identity
      ))

      count_non_hms <- sum(!is_hms_before)
      if (count_non_hms > 0) {
        emit_log(vname, count_non_hms, lost, is_hms_before)
      }

      len1 <- vapply(x2, length, integer(1)) == 1L
      if (all(len1)) {
        vals <- suppressWarnings(vapply(
          x2,
          function(e) if (inherits(e, "hms") && length(e) == 1) as.numeric(e)
          else NA_real_,
          numeric(1)
        ))
        study_data[[j]] <- hms::hms(vals)
      } else {
        flat <- try(
          hms::as_hms(unlist(x2, recursive = FALSE, use.names = FALSE)),
          silent = TRUE
        )
        if (!util_is_try_error(flat)) {
          study_data[[j]] <- flat
        } else {
          study_data[[j]] <- x2
        }
      }
      next
    }
  }

  study_data
}

#' Convert missing codes in metadata format v1.0 and a missing-cause-table
#' to v2.0 missing list / jump list assignments
#'
#' The function has to working modes. If `replace_meta_data` is `TRUE`, by
#' default, if `cause_label_df` contains a column
#' named `resp_vars`, then the missing/jump codes in
#' `meta_data[, c(MISSING_CODES, JUMP_CODES)]` will be overwritten, otherwise,
#' it will be labeled using the `cause_label_df`.
#'
#' If a column `resp_vars` exists, then rows with a value in `resp_vars` will
#' only be used for the corresponding variable.
#'
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param cause_label_df [data.frame] missing code table. If missing codes have
#'                                    labels the respective data frame can be
#'                                    specified here, see [cause_label_df]
#' @param assume_consistent_codes [logical] if TRUE and no labels are given and
#'                                          the same missing/jump code is used
#'                                          for more than one variable, the
#'                                          labels assigned for this code will
#'                                          be the same for all variables.
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param replace_meta_data [logical] if `TRUE`, ignore existing missing codes
#'                                    and jump codes and replace them with
#'                                    data from the `cause_label_df`. Otherwise,
#'                                    copy the labels from `cause_label_df` to
#'                                    the existing code columns.
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#' @return [data.frame] updated metadata including all the code labels in
#'                      missing/jump lists
#' @seealso [prep_extract_cause_label_df]
#'
#' @export
#'
prep_add_cause_label_df <- function(item_level = "item_level",
                                    cause_label_df,
                                    label_col = VAR_NAMES,
                                    assume_consistent_codes = TRUE,
                                    replace_meta_data =
                                      ("resp_vars" %in%
                                         colnames(cause_label_df)),
                                    meta_data = item_level,
                                    meta_data_v2) {

  util_maybe_load_meta_data_v2()

  util_expect_data_frame(meta_data, c(DATA_TYPE))
  util_expect_data_frame(cause_label_df, c(CODE_VALUE, CODE_LABEL))
  if (CODE_CLASS %in% colnames(cause_label_df)) {
    cause_label_df[[CODE_CLASS]][is.na(cause_label_df[[CODE_CLASS]])] <-
      ""
  }
  if (length(label_col) != 1 ||
      is.na(label_col) ||
      !is.character(label_col)) {
    util_error("Need one character value in %s",
               dQuote("label_col"))
  }
  if (!label_col %in% colnames(meta_data)) {
    util_error("No column %s in %s", dQuote(label_col), dQuote("meta_data"))
  }
  if (length(assume_consistent_codes) != 1 ||
      is.na(assume_consistent_codes) ||
      !is.logical(assume_consistent_codes)) {
    util_error("Need one logical value in %s",
               dQuote("assume_consistent_codes"))
  }
  if (length(replace_meta_data) != 1 ||
      is.na(replace_meta_data) ||
      !is.logical(replace_meta_data)) {
    util_error("Need one logical value in %s", dQuote("replace_meta_data"))
  }

  have_resp_vars_col <- ("resp_vars" %in%
                           colnames(cause_label_df))

  amend_cl <- function(rv, lst, tp) {
    lst_nm <- sprintf("%s_LIST", lst)
    if (lst_nm %in% colnames(meta_data)) {
      cl <- meta_data[meta_data[[label_col]] == rv, lst_nm, TRUE]
    } else {
      cl <- NA_character_
    }
    if (any(is.na(cl)) && !replace_meta_data) {
      NA_character_
    } else {
      if (CODE_CLASS %in% colnames(cause_label_df)) {
        cause_label_df <- cause_label_df[
                                 cause_label_df[[CODE_CLASS]] == lst, , FALSE]
      } else if (replace_meta_data && lst != "MISSING") {
        cause_label_df <- subset(cause_label_df, FALSE)
      }
      if (have_resp_vars_col) {
        cause_label_df <- subset(cause_label_df,
                                 is.na(get("resp_vars")) |
                                 trimws(get("resp_vars")) == "" |
                                 get("resp_vars") == rv)
      }
      if (replace_meta_data) {
        cls <- cause_label_df$CODE_VALUE
      } else {
        if (tp %in% c(DATA_TYPES$DATETIME)) {
          cls <- util_parse_date(names(util_parse_assignments(cl)))
        } else if (tp %in% c(DATA_TYPES$TIME)) {
          cls <- util_parse_time(names(util_parse_assignments(cl)))
        } else {
          cls <- as.integer(names(util_parse_assignments(cl)))
        }
      }
      cldf <- cause_label_df[
                     as.character(cause_label_df[[CODE_VALUE]]) %in% as.character(cls),
                     , FALSE]
      miss <- cls[!(as.character(cls) %in%
                      as.character(cause_label_df[[CODE_VALUE]]))]
      if (length(miss) > 0) {
        if (assume_consistent_codes) {
          cldf1 <- data.frame(CODE_VALUE = miss,
                              CODE_LABEL = paste(lst, miss),
                              stringsAsFactors = FALSE)
        } else {
          cldf1 <- data.frame(CODE_VALUE = miss,
                              CODE_LABEL = paste(lst, rv, miss),
                              stringsAsFactors = FALSE)
        }
        cldf <- util_rbind(cldf, cldf1)
      }
      prep_deparse_assignments(codes = cldf$CODE_VALUE,
                               labels = cldf$CODE_LABEL)
    }
  }

  meta_data$MISSING_LIST <-
    vapply(mapply(
      rv = meta_data[[label_col]],
      tp = meta_data[[DATA_TYPE]],
      FUN = amend_cl,
      MoreArgs = list(lst = "MISSING")),
      as.character,
      FUN.VALUE = character(1))

  meta_data$JUMP_LIST <-
    vapply(mapply(
      rv = meta_data[[label_col]],
      tp = meta_data[[DATA_TYPE]],
      FUN = amend_cl,
      MoreArgs = list(lst = "JUMP")),
      as.character,
      FUN.VALUE = character(1))

  meta_data
}

#' Parallel or fallback to serial `lapply`
#'
#' Wraps `parallel::parLapplyLB`, but if no cluster is available,
#' falls back to `lapply()`. The cluster is automatically detected using
#' `parallel::getDefaultCluster()`.
#'
#' @param cl cluster Optional. A cluster object as returned by `makeCluster()`.
#'                     If `NULL`, attempts to use `parallel::getDefaultCluster()`.
#'                     If that is also `NULL`, falls back to serial `lapply()`.
#' @param X [list] The list of elements to iterate over.
#' @param fun [function] The function to apply to each element.
#' @param ... Further arguments passed to `fun`.
#' @param chunk.size [integer] Optional chunk size passed to `parLapplyLB`.
#'
#' @return A list of results, like `lapply()` or `parLapplyLB()` would return.
#' @noRd
util_par_lapply_lb <- function(cl = parallel::getDefaultCluster(),
                               X, fun, ..., chunk.size = NULL) {
  if (is.null(cl)) {
    # No cluster: fallback to serial execution
    return(lapply(X, fun, ...))
  } else {
    # Cluster execution
    if (is.null(chunk.size)) {
      return(parallel::parLapplyLB(cl = cl, X = X, fun = fun, ...))
    } else {
      return(parallel::parLapplyLB(cl = cl, X = X, fun = fun,
                                   chunk.size = chunk.size, ...))
    }
  }
}

.on_cran <- function() {
  env <- Sys.getenv("NOT_CRAN")
  if (identical(env, "")) {
    !interactive()
  } else {
    !isTRUE(as.logical(env))
  }
}

if (.on_cran()) {
  .expect_doppelganger2 <- function(...) {
    testthat::skip_on_cran()
  }
} else {
  if (requireNamespace("vdiffr", quietly = TRUE) &&
      packageVersion("vdiffr") >= "1.0.8" &&
      "variant" %in% names(formals(vdiffr::expect_doppelganger))) {
    .expect_doppelganger2 <- vdiffr::expect_doppelganger
    formals(.expect_doppelganger2)$variant <- prep_get_variant()
  } else { # here, the variant formal was missing.
    .expect_doppelganger2 <- function (title, fig, path = deprecated(), ...,
                                      writer = write_svg,
                                      cran = FALSE, variant =
                                        prep_get_variant())
    { # see vdiffr::expect_doppelganger, ugly patch to avoid https://github.com/r-lib/vdiffr/issues/125
      # this file is a patched version from vdiffr (MIT License, original authors: see packageDescription("vdiffr"))
      testthat::local_edition(3)
      fig_name <- str_standardise(title)
      file <- paste0(fig_name, ".svg")
      testthat::announce_snapshot_file(name = file)
      testcase <- make_testcase_file(fig_name)
      writer(fig, testcase, title)
      if (!missing(...)) {
        lifecycle::deprecate_soft("1.0.0", "vdiffr::expect_doppelganger(... = )",
        )
      }
      if (lifecycle::is_present(path)) {
        lifecycle::deprecate_soft("1.0.0", "vdiffr::expect_doppelganger(path = )",
        )
      }
      if (is_graphics_engine_stale()) {
        testthat::skip(paste_line("The R graphics engine is too old.",
                                  "Please update to R 4.1.0 and regenerate the vdiffr snapshots."))
      }
      withCallingHandlers(testthat::expect_snapshot_file(variant = variant,
                                                         testcase, name = file, cran = cran, compare = testthat::compare_file_text),
                          expectation_failure = function(cnd) {
                            if (is_snapshot_stale(title, testcase)) {
                              testthat::skip(paste_line("SVG snapshot generated under a different vdiffr version.",
                                                        i = "Please update your snapshots."))
                            }
                            if (!is_null(snapshotter <- get_snapshotter())) {
                              path_old <- snapshot_path(snapshotter, file)
                              path_new <- snapshot_path(snapshotter, paste0(fig_name,
                                                                            ".new.svg"))
                              if (all(file.exists(path_old, path_new))) {
                                push_log(fig_name, path_old, path_new)
                              }
                            }
                          })
    }
    if (requireNamespace("vdiffr", quietly = TRUE)) {
      environment(expect_doppelganger2) <- asNamespace("vdiffr")
    }
  }
}
expect_doppelganger2 <- function(title, fig, ...) {
  if (util_is_gg(fig))
    fig <- prep_realize_ggplot(fig)
  .expect_doppelganger2(title, fig, ...)
}

test_that("acc_multivariate_outlier works with 3 args", {
  skip_on_cran() # slow and tested elsewhere
  skip_if_translated()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_error(
    res1 <-
      acc_multivariate_outlier(
        study_data = study_data, meta_data = meta_data, scale = FALSE),
    regexp =
      "Argument variable_group is NULL",
    perl = TRUE
  )

  expect_error(
    res1 <-
      acc_multivariate_outlier(
        variable_group = "v00014",
        study_data = study_data, meta_data = meta_data, scale = FALSE),
    regexp =
      "Need at least two variables for multivariate outliers.",
    perl = TRUE
  )

  expect_message2(
    res1 <-
      acc_multivariate_outlier(variable_group = c("v00014", "v00006"),
                         study_data = study_data, meta_data = meta_data,
                         scale = FALSE),
    regexp =
      sprintf(
        "(%s|%s)",
        paste("As no ID-var has been specified the rownumbers will be used."),
        paste("Due to missing values in v00014, v00006 or dq_id N=602",
              "observations were excluded.")
      ),
    perl = TRUE
  )

  expect_true(all(c("FlaggedStudyData",
                    "SummaryTable",
                    "SummaryPlot") %in% names(res1)))
  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$FlaggedStudyData)),
      na.rm = TRUE) - 4492191)), 0.5
  )
  expect_equal(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE))), 182+3.13
  )
})

test_that("acc_multivariate_outlier works with label_col", {
  skip_on_cran() # slow and tested elsewhere
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_error(
    res1 <-
      acc_multivariate_outlier(
        label_col = LABEL,
        study_data = study_data, meta_data = meta_data, scale = FALSE),
    regexp =
      "Argument variable_group is NULL",
    perl = TRUE
  )

  expect_error(
    res1 <-
      acc_multivariate_outlier(
        label_col = LABEL,
        variable_group = "CRP_0",
        study_data = study_data, meta_data = meta_data, scale = FALSE),
    regexp =
      "Need at least two variables for multivariate outliers.",
    perl = TRUE
  )

  expect_message2(
    res1 <-
      acc_multivariate_outlier(variable_group = c("CRP_0", "GLOBAL_HEALTH_VAS_0"),
                               label_col = LABEL,
                               study_data = study_data, meta_data = meta_data,
                               scale = FALSE),
    regexp =
      sprintf(
        "(%s|%s)",
        paste("As no ID-var has been specified the rownumbers will be used."),
        paste("Due to missing values in CRP_0, GLOBAL_HEALTH_VAS_0 or",
              "dq_id N=602 observations were excluded.")
      ),
    perl = TRUE
  )

  expect_true(all(c("FlaggedStudyData",
                    "SummaryTable",
                    "SummaryPlot") %in% names(res1)))
  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$FlaggedStudyData)),
      na.rm = TRUE) - 4492191)), 0.5
  )
  expect_equal(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE))), 182+3.13
  )

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2(
    "acc_mv_outlierCRP0GLOBHEAVA0",
                              res1$SummaryPlot)
})

test_that("acc_multivariate_outlier works with min-max-scaling", {
  skip_on_cran() # slow and tested elsewhere
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_multivariate_outlier(variable_group = c("CRP_0", "GLOBAL_HEALTH_VAS_0"),
                               label_col = LABEL,
                               study_data = study_data, meta_data = meta_data,
                               scale = TRUE),
    regexp =
      sprintf(
        "(%s|%s)",
        paste("As no ID-var has been specified the rownumbers will be used."),
        paste("Due to missing values in CRP_0, GLOBAL_HEALTH_VAS_0 or",
              "dq_id N=602 observations were excluded.")
      ),
    perl = TRUE
  )

  expect_true(all(c("FlaggedStudyData",
                    "SummaryTable",
                    "SummaryPlot") %in% names(res1)))

  expect_false(
    inherits(try(ggplot_build(res1$SummaryPlot)), "try-error"))
})

# TODO:  a test for invalid values in the criteria argument.

test_that("acc_shape_or_scale works with 3 args", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]


  expect_message2(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = meta_data,
                         dist_col = DISTRIBUTION,
                         guess = c(TRUE, FALSE, TRUE)),
    regexp = sprintf(
      "(%s|%s)",
      paste("guess should be a scalar logical value.",
            "Have more than one value, use the first one only"),
      paste("Due to missing values in v00014 301 observations were deleted.")
    ),
    perl = TRUE
  )

  expect_message2(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = meta_data,
                         dist_col = DISTRIBUTION,
                         end_digits = c(TRUE, FALSE, TRUE)),
    regexp = sprintf(
      "(%s|%s)",
      paste("end_digits should be a scalar logical value.",
            "Have more than one value, use the first one only"),
      paste("Due to missing values in v00014 301 observations were deleted.")
    ),
    perl = TRUE
  )

  expect_error(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = meta_data,
                         dist_col = DISTRIBUTION,
                         par1 = "xxx"),
    regexp = "par1 should be a numeric value"
  )

  expect_error(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = meta_data,
                         dist_col = DISTRIBUTION,
                         par1 = 0,
                         par2 = "xxx",
                         guess = FALSE),
    regexp = "par2 should be a numeric value"
  )

  md1 <- meta_data
  md1[md1$VAR_NAMES == "v00014", DISTRIBUTION] <- NA
  suppressMessages(suppressWarnings(expect_error(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = md1,
                         dist_col = DISTRIBUTION,
                         par1 = 0,
                         par2 = 1,
                         guess = FALSE),
    regexp = "No distribution specified for v00014 in DISTRIBUTION"
  )))

  md1 <- meta_data
  md1[md1$VAR_NAMES == "v00014", DISTRIBUTION] <- "dirichlet"
  suppressWarnings(expect_error(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = md1,
                         dist_col = DISTRIBUTION,
                         par1 = 0,
                         par2 = 1,
                         guess = FALSE),
    regexp = "This distribution .+dirichlet.+ is not supported yet..."
  ))

  expect_message2(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00006", # uniform and integer
                         study_data = study_data,
                         meta_data = meta_data,
                         dist_col = DISTRIBUTION,
                         par1 = 1,
                         par2 = 9,
                         guess = FALSE),
    regexp = "Due to missing values in v00006 382 observations were deleted."
  )
  expect_equal(sum(1 == res1$ResultData$GRADING), 8)
  expect_error(
    suppressWarnings(
      res1 <-
        acc_shape_or_scale(resp_vars = "v00014",
                           study_data = study_data,
                           meta_data = meta_data,
                           dist_col = DISTRIBUTION,
                           par1 = Inf,
                           par2 = 1,
                           guess = FALSE)
    ),
    regexp = paste(
      "Since .+guess.+ is not true finite numerical",
      "parameters must be prespecified"),
    perl = TRUE
  )

expect_error(
  res1 <-
    acc_shape_or_scale(resp_vars = "v00001",
                       study_data = study_data,
                       meta_data = meta_data,
                       dist_col = DISTRIBUTION,
                       par1 = 0,
                       par2 = 1,
                       guess = FALSE),
  regexp = paste("Argument .resp_vars.+ Variable .v00001.+string. does",
                 "not have an allowed type"),
  perl = TRUE
)
expect_message2(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = meta_data,
                         dist_col = DISTRIBUTION,
                         par1 = 0:10,
                         par2 = 1:11),
    regexp = sprintf(
      "(%s|%s|%s|%s)",
      paste("Since parameters were specified: .+guess.+ is set to false"),
      paste("par1 should be a scalar numeric value. Have more than one value,",
            "use the first one only"),
      paste("par2 should be a scalar numeric value. Have more than one value,",
            "use the first one only"),
      paste("Due to missing values in v00014 301 observations were deleted.")
    ),
    perl = TRUE
  )
expect_message2(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = meta_data,
                         dist_col = DISTRIBUTION,
                         par1 = 0,
                         par2 = 1),
    regexp = sprintf(
      "(%s|%s)",
      paste("Since parameters were specified: .+guess.+ is set to false"),
      paste("Due to missing values in v00014 301 observations were deleted.")
    ),
    perl = TRUE
  )
expect_message2(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = meta_data,
                         dist_col = DISTRIBUTION,
                         guess = c(NA, TRUE)),
    regexp = sprintf("(%s|%s)",
                     paste("Have more than one value for guess,",
                           "use the first one only"),
                     paste("Due to missing values in v00014 301",
                           "observations were deleted.")),
    perl = TRUE
  )
  md1 <- meta_data[, setdiff(colnames(meta_data), DISTRIBUTION)]
  expect_error(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data,
                         meta_data = md1,
                         dist_col = DISTRIBUTION),
    regexp = "Did not find variable attribute DISTRIBUTION in the meta_data"
  )
  expect_warning(
    expect_error(
      res1 <-
        acc_shape_or_scale(study_data = study_data, meta_data = meta_data),
      regexp =
        "Argument resp_vars is NULL",
      perl = TRUE
    ),
    regexp =
      sprintf(
        "(%s)",
        paste("Missing argument .+resp_vars.+ without default value.",
              "Setting to NULL. As a dataquieR developer,")
      ),
    perl = TRUE
  )

  expect_message2(
    res1 <-
      acc_shape_or_scale(resp_vars = "v00014",
                         study_data = study_data, meta_data = meta_data),
    regexp =
      sprintf(
        "(%s|%s)",
        paste("A column of the metaddata specifying the distributions has",
              "not been specified. Trying the default .+DISTRIBUTION.+."),
        paste("Due to missing values in v00014 301 observations were deleted.")
      ),
    perl = TRUE
  )

  expect_true(all(c("ResultData",
                  "SummaryPlot",
                  "SummaryTable") %in% names(res1)))
  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$ResultData)),
      na.rm = TRUE) - 5484.87)), 0.5
  )
  expect_equal(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE))), 1
  )
})

test_that("acc_shape_or_scale works with label_col", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]

    expect_warning(
    expect_error(
      res1 <-
        acc_shape_or_scale(study_data = study_data, meta_data = meta_data,
                           label_col = LABEL),
      regexp =
        "Argument resp_vars is NULL",
      perl = TRUE
    ),
    regexp =
      sprintf(
        "(%s)",
        paste("Missing argument .+resp_vars.+ without default value.",
              "Setting to NULL. As a dataquieR developer,")
      ),
    perl = TRUE
  )

  expect_message2(
    res1 <-
      acc_shape_or_scale(resp_vars = "CRP_0",
                         study_data = study_data, meta_data = meta_data,
                         label_col = LABEL),
    regexp =
      sprintf(
        "(%s|%s)",
        paste("A column of the metaddata specifying the distributions has",
              "not been specified. Trying the default .+DISTRIBUTION.+."),
        paste("Due to missing values in CRP_0 301 observations were deleted.")
      ),
    perl = TRUE
  )

  expect_true(all(c("ResultData",
                    "SummaryPlot",
                    "SummaryTable") %in% names(res1)))
  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$ResultData)),
      na.rm = TRUE) - 5484.87)), 0.5
  )
  expect_equal(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE))), 1
  )
  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("shape_or_scale plot for CRP_0 ok",
                              res1$SummaryPlot)
})

test_that("acc_end_digits works with 2 args", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_error(
    res1 <-
      acc_end_digits(study_data = study_data, meta_data = meta_data),
    regexp = "Argument resp_vars is NULL",
    perl = TRUE
  )
  expect_message2(
    res1 <-
      acc_end_digits(resp_vars = "v00014", study_data = study_data,
                     meta_data = meta_data),
    regexp =
      sprintf(
        "(%s)",
        paste("Due to missing values in v00014 301",
              "observations were deleted.")
      ),
    perl = TRUE
  )

  expect_true(all(c("SummaryTable", "SummaryPlot") %in% names(res1)))
  expect_equal(
    nrow(res1$SummaryTable),
    1
  )
  expect_equal(
    ncol(res1$SummaryTable),
    2
  )
  expect_equal(
    suppressWarnings(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE)), 1
  )
})

test_that("acc_end_digits works with label_col", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_end_digits(resp_vars = "CRP_0", study_data = study_data,
                     meta_data = meta_data, label_col = LABEL),
    regexp =
      sprintf(
        "(%s)",
        paste("Due to missing values in CRP_0",
              "301 observations were deleted.")
      ),
    perl = TRUE
  )
  expect_true(all(c("SummaryTable", "SummaryPlot") %in% names(res1)))
  expect_equal(
    nrow(res1$SummaryTable),
    1
  )
  expect_equal(
    ncol(res1$SummaryTable),
    2
  )
  expect_equal(
    suppressWarnings(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE)), 1
  )
})

test_that("acc_end_digits works image check", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # TODO: skip_if_not(capabilities()["long.double"])
  skip_if_not_installed("vdiffr")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  expect_message2(
    res1 <-
      acc_end_digits(resp_vars = "CRP_0",
                     study_data = study_data, meta_data = meta_data,
                     label_col = LABEL),
    regexp =
      sprintf(
        "(%s)",
        paste("Due to missing values in CRP_0",
              "301 observations were deleted.")
      ),
    perl = TRUE
  )
  skip_on_cran()
  expect_doppelganger2("enddigits plot for CRP_0 ok",
                              res1$SummaryPlot)
})

test_that("acc_end_digits is robust", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  sd0 <- study_data
  sd0$v00014[[2]] <- Inf
  expect_error(
    res1 <-
      acc_end_digits(resp_vars = "CRP_0",
                     study_data = sd0, meta_data = meta_data,
                     label_col = LABEL),
    regexp =
      sprintf(
        "(%s)",
        paste("Values in .+resp_vars.+ must not contain infinite data")
      ),
    perl = TRUE
  )
  sd0 <- study_data
  md0 <- meta_data
  md0[[DECIMALS]] <- NULL
  expect_error(
    res1 <-
      acc_end_digits(resp_vars = "CRP_0",
                     study_data = sd0, meta_data = md0,
                     label_col = LABEL),
    regexp =
      sprintf(
        "(%s)",
        paste("The number of digits following the",
              "decimal point must be prespecified.")
      ),
    perl = TRUE
  )
  sd0 <- study_data
  sd0$v00010[[2]] <- sd0$v00010[[2]] + 0.5
  md0 <- meta_data
  expect_error(
    res1 <-
      acc_end_digits(resp_vars = "ARM_CUFF_0",
                     study_data = sd0, meta_data = md0,
                     label_col = LABEL),
    regexp =
      paste("Argument .+resp_vars.+: Variable .+ARM_CUFF_0.+ .nominal. does",
            "not have an allowed scale level"),
    perl = TRUE
  )

  sd0 <- study_data
  sd0$v00014 <- c(rep(1:9, 3000/9), 1:3)
  sd0$v00014 <- sd0$v00014 / 1000
  sd0$CRP_0_x_last <- sd0$v00014
  sd0$CRP_0_x_last1 <- sd0$v00014
  md0 <- prep_add_to_meta(VAR_NAMES = "CRP_0_x_last",
                          DATA_TYPE = DATA_TYPES$STRING,
                          LABEL = "xxx",
                          VALUE_LABELS = NA,
                          meta_data)
  md0 <- prep_add_to_meta(VAR_NAMES = "CRP_0_x_last1",
                          DATA_TYPE = DATA_TYPES$STRING,
                          LABEL = "yyy",
                          VALUE_LABELS = NA,
                          md0)
  res1 <-
    acc_end_digits(resp_vars = "CRP_0",
                   study_data = sd0, meta_data = md0,
                   label_col = LABEL)

  sd0 <- study_data
  sd0$v00014 <- c(rep(1:9, 3000/9), 1:3)
  sd0$v00014 <- sd0$v00014 / 1000
  sd0$CRP_0_x_last <- sd0$v00014
  sd0$CRP_0_x_last1 <- sd0$v00014
  md0 <- prep_add_to_meta(VAR_NAMES = "CRP_0_x_last",
                          DATA_TYPE = DATA_TYPES$STRING,
                          LABEL = "xxx",
                          VALUE_LABELS = NA,
                          meta_data)
  md0 <- prep_add_to_meta(VAR_NAMES = "CRP_0_x_last1",
                          DATA_TYPE = DATA_TYPES$STRING,
                          LABEL = "yyy",
                          VALUE_LABELS = NA,
                          md0)
  md0$DISTRIBUTION <- NULL
  res1 <-
    acc_end_digits(resp_vars = "CRP_0",
                   study_data = sd0, meta_data = md0,
                   label_col = LABEL)

})

test_that("com_segment_missingness works", {
  skip_on_cran()
  
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = FALSE,
                   dataquieR.ERRORS_WITH_CALLER = FALSE,
                   dataquieR.WARNINGS_WITH_CALLER = FALSE,
                   dataquieR.MESSAGES_WITH_CALLER = FALSE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]
  expect_message2(
    r <- com_segment_missingness(study_data, meta_data, label_col = LABEL,
                               threshold_value = NA, color_gradient_direction = "above",
                               exclude_roles = VARIABLE_ROLES$PROCESS),
    regexp = sprintf("%s|%s",
                     paste("Study variables: .+ARM_CUFF_0.+,",
                           ".+USR_VO2_0.+, .+USR_BP_0.+,",
                           ".+EXAM_DT_0.+, .+DEV_NO_0.+,",
                           ".+LAB_DT_0.+, .+USR_SOCDEM_0.+,",
                           ".+INT_DT_0.+, .+QUEST_DT_0.+",
                           "are not considered due to their",
                           "VARIABLE_ROLE."),
                     paste("threshold_value should be a single number between",
                           "0 and 100. Invalid value specified,",
                           "setting to 10%."))
  )

  expect_message2(
    r <- com_segment_missingness(study_data, meta_data, label_col = LABEL,
                                 threshold_value = NA, color_gradient_direction = "above"),
    regexp = sprintf("%s|%s|%s",
                     paste("Study variables: .+ARM_CUFF_0.+,",
                           ".+USR_VO2_0.+, .+USR_BP_0.+,",
                           ".+EXAM_DT_0.+, .+DEV_NO_0.+,",
                           ".+LAB_DT_0.+, .+USR_SOCDEM_0.+,",
                           ".+INT_DT_0.+, .+QUEST_DT_0.+",
                           "are not considered due to their",
                           "VARIABLE_ROLE."),
                     paste("threshold_value should be a single number between",
                           "0 and 100. Invalid value specified,",
                           "setting to 10%."),
                     paste("Formal exclude_roles is used with default:",
                           "all process variables are not included here."))
  )

  expect_message2(
    r <- com_segment_missingness(study_data, meta_data,
                                 threshold_value = NA, color_gradient_direction = "above"),
    regexp = sprintf("%s|%s|%s",
                     paste("Study variables: .+v00010.+,",
                           ".+v00011.+, .+v00012.+,",
                           ".+v00013.+, .+v00016.+,",
                           ".+v00017.+, .+v00032.+,",
                           ".+v00033.+, .+v00042.+",
                           "are not considered due to their",
                           "VARIABLE_ROLE."),
                     paste("threshold_value should be a single number between",
                           "0 and 100. Invalid value specified,",
                           "setting to 10%."),
                     paste("Formal exclude_roles is used with default:",
                           "all process variables are not included here."))
  )

  expect_message2(
    r <- com_segment_missingness(study_data, meta_data, label_col = LABEL,
                                 threshold_value = NA, color_gradient_direction = "above",
                                 strata_vars = "CENTER_0"),
    regexp = sprintf("%s|%s|%s",
                     paste("Study variables: .+ARM_CUFF_0.+,",
                           ".+USR_VO2_0.+, .+USR_BP_0.+,",
                           ".+EXAM_DT_0.+, .+DEV_NO_0.+,",
                           ".+LAB_DT_0.+, .+USR_SOCDEM_0.+,",
                           ".+INT_DT_0.+, .+QUEST_DT_0.+",
                           "are not considered due to their",
                           "VARIABLE_ROLE."),
                     paste("threshold_value should be a single number between",
                           "0 and 100. Invalid value specified,",
                           "setting to 10%."),
                     paste("Formal exclude_roles is used with default:",
                           "all process variables are not included here."))
  )

  meta_data2 <- meta_data
  meta_data2$KEY_STUDY_SEGMENT <- NULL
  meta_data2$STUDY_SEGMENT <- NULL
  expect_error(suppressWarnings(suppressMessages(
    r <- com_segment_missingness(study_data, meta_data2,
                                 threshold_value = 10, color_gradient_direction = "above",
                                 exclude_roles = c(VARIABLE_ROLES$PROCESS,
                                                   "invalid"))
    )),
    regexp = paste(".*Metadata do not contain",
                   "the column STUDY_SEGMENT"),
    perl = TRUE
  )

  meta_data2 <- meta_data
  meta_data2$LONG_LABEL <- NA
  expect_warning(
    r <- com_segment_missingness(study_data, meta_data2,
                                 threshold_value = 10, color_gradient_direction = "above",
                                 exclude_roles = c(VARIABLE_ROLES$PROCESS,
                                                   "invalid")),
    regexp = sprintf("%s|%s",
                     paste("Specified VARIABLE_ROLE.s.:",
                           ".+invalid.+ was not found in metadata, only:",
                           ".+process.+ is used."),
                     paste("Study variables: .+v00010.+, .+v00011.+,",
                           ".+v00012.+, .+v00013.+, .+v00016.+, .+v00017.+,",
                           ".+v00032.+, .+v00033.+, .+v00042.+ are not",
                           "considered due to their VARIABLE_ROLE.")),
    perl = TRUE
  )

  expect_warning(
    r <- com_segment_missingness(study_data, meta_data,
                                 threshold_value = 10, color_gradient_direction = "above",
                                 exclude_roles = c(VARIABLE_ROLES$PROCESS,
                                                   "invalid")),
    regexp = sprintf("%s|%s",
                     paste("Specified VARIABLE_ROLE.s.:",
                           ".+invalid.+ was not found in metadata, only:",
                           ".+process.+ is used."),
                     paste("Study variables: .+v00010.+, .+v00011.+,",
                           ".+v00012.+, .+v00013.+, .+v00016.+, .+v00017.+,",
                           ".+v00032.+, .+v00033.+, .+v00042.+ are not",
                           "considered due to their VARIABLE_ROLE.")),
    perl = TRUE
  )

  expect_warning(
    r <- com_segment_missingness(study_data, meta_data, label_col = LABEL,
                                 threshold_value = 10, color_gradient_direction = "above",
                                 exclude_roles = c(VARIABLE_ROLES$PROCESS,
                                                   "invalid")),
    regexp = sprintf("%s|%s",
                     paste("Specified VARIABLE_ROLE.s.:",
                           ".+invalid.+ was not found in metadata, only:",
                           ".+process.+ is used."),
                     paste("Study variables: .+ARM_CUFF_0.+,",
                           ".+USR_VO2_0.+, .+USR_BP_0.+,",
                           ".+EXAM_DT_0.+, .+DEV_NO_0.+,",
                           ".+LAB_DT_0.+, .+USR_SOCDEM_0.+,",
                           ".+INT_DT_0.+, .+QUEST_DT_0.+",
                           "are not considered due to their",
                           "VARIABLE_ROLE.")),
    perl = TRUE
  )

  expect_error(suppressWarnings(
    r <- com_segment_missingness(study_data, meta_data, label_col = LABEL,
                                 threshold_value = 10, color_gradient_direction = "invalid",
                                 exclude_roles = VARIABLE_ROLES$PROCESS)
    ),
    regexp = paste("Parameter .+color_gradient_direction.+ should be either .+above.+ or",
                   ".+below.+, but not .+invalid.+."),
    perl = TRUE
  )

  expect_error(suppressMessages(
    r <- com_segment_missingness(study_data, meta_data, label_col = LABEL,
                                 threshold_value = 10, color_gradient_direction = 1:2,
                                 exclude_roles = VARIABLE_ROLES$PROCESS)
    ),
    regexp = paste("Parameter .+color_gradient_direction.+ should be of length",
                   "1, but not 2."),
    perl = TRUE
  )

  expect_message2(
    r <- com_segment_missingness(study_data, meta_data, label_col = LABEL,
                                 threshold_value = 10, color_gradient_direction = "above",
                                 exclude_roles = VARIABLE_ROLES$PROCESS),
    regexp = paste("Study variables: .+ARM_CUFF_0.+,",
                   ".+USR_VO2_0.+, .+USR_BP_0.+,",
                   ".+EXAM_DT_0.+, .+DEV_NO_0.+,",
                   ".+LAB_DT_0.+, .+USR_SOCDEM_0.+,",
                   ".+INT_DT_0.+, .+QUEST_DT_0.+",
                   "are not considered due to their",
                   "VARIABLE_ROLE.")
  )
  expect_message2(
    r <- com_segment_missingness(study_data, meta_data, label_col = LABEL,
                                 threshold_value = 10, color_gradient_direction = "below",
                                 exclude_roles = VARIABLE_ROLES$PROCESS),
    regexp = paste("Study variables: .+ARM_CUFF_0.+,",
                   ".+USR_VO2_0.+, .+USR_BP_0.+,",
                   ".+EXAM_DT_0.+, .+DEV_NO_0.+,",
                   ".+LAB_DT_0.+, .+USR_SOCDEM_0.+,",
                   ".+INT_DT_0.+, .+QUEST_DT_0.+",
                   "are not considered due to their",
                   "VARIABLE_ROLE.")
  )
  expect_equal(
    length(intersect(
      names(r),
      c("ResultData", "SummaryPlot", "ReportSummaryTable")
    )), length(union(
      names(r),
      c("ResultData", "SummaryPlot", "ReportSummaryTable")
    ))
  )
  expect_true(abs(suppressWarnings(sum(as.numeric(as.matrix(
    r$ResultData)), na.rm = TRUE)) - 15288.63) < 2)

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("segment missingness plot ok",
                              r$SummaryPlot)
})

test_that("com_segment_missingness works w/g (group|strata)_vars", {
  skip_on_cran() # slow and not frequently used
  
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = FALSE,
                   dataquieR.ERRORS_WITH_CALLER = FALSE,
                   dataquieR.WARNINGS_WITH_CALLER = FALSE,
                   dataquieR.MESSAGES_WITH_CALLER = FALSE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]
  expect_message2({
    r1 <- com_segment_missingness(study_data, meta_data, strata_vars = "CENTER_0",
                                  threshold_value = 5,
                                  color_gradient_direction = "above",
                                  exclude_roles = VARIABLE_ROLES$PROCESS)
    r2 <- com_segment_missingness(study_data, meta_data, strata_vars = "CENTER_0",
                                  group_vars = "SEX_0",
                                  threshold_value = 5,
                                  color_gradient_direction = "above",
                                  exclude_roles = VARIABLE_ROLES$PROCESS)
    r3 <- com_segment_missingness(study_data, meta_data, group_vars = "SEX_0",
                                  threshold_value = 5,
                                  color_gradient_direction = "above",
                                  exclude_roles = VARIABLE_ROLES$PROCESS)
    },
    regexp = "Study variables: .+ are not considered due to their VARIABLE_ROLE.",
    perl = TRUE
  )
  testthat::local_edition(3)
  expect_snapshot_value(style = "deparse",
                        r1$ResultData)
  expect_snapshot_value(style = "deparse",
                        r2$ResultData)
  expect_snapshot_value(style = "deparse",
                        r3$ResultData)
  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("segment missingness plot r1 ok",
                              r1$SummaryPlot)
  expect_doppelganger2("segment missingness plot r2 ok",
                              r2$SummaryPlot)
  expect_doppelganger2("segment missingness plot r3 ok",
                              r3$SummaryPlot)
})

test_that("com_item_missingness works", {
  skip_on_cran() # slow and errors will be obvious.

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  md0 <- meta_data
  md0$STUDY_SEGMENT <- NULL
  expect_message2(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = md0,
                         suppressWarnings = TRUE)
  ))

  expect_equal(com_item_missingness(resp_vars = "v00001",
                                    study_data = study_data,
                                    item_level = md0,
                                    suppressWarnings = TRUE,
                                    threshold_value = 100,
                                    include_sysmiss = FALSE,
                                    drop_levels = TRUE,
                                    assume_consistent_codes = TRUE,
                                    expand_codes = TRUE,
                                    show_causes = TRUE,
                                    expected_observations =
                                      "ALL")$SummaryTable$GRADING,
               0)

  skip_on_cran() # too many tests make things slow too.

  expect_error(invisible(
    com_item_missingness(study_data, meta_data,
                         suppressWarnings = TRUE,
                         include_sysmiss = 1:10)
  ), regexp = "Need exactly one element in argument include_sysmiss")

  w <- capture_warnings(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data,
                         suppressWarnings = TRUE,
                         include_sysmiss = TRUE)
  ))
  w <- gsub("(\n|^|\r)+>.*$", "", w)
  expect_false(any(grepl("include_sysmiss", w)))

  w <- capture_warnings(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data,
                         suppressWarnings = TRUE,
                         include_sysmiss = FALSE)
  ))
  w <- gsub("(\n|^|\r)+>.*$", "", w)
  expect_false(any(grepl("include_sysmiss", w)))

  expect_message2(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data,
                         suppressWarnings = TRUE)
  ), regexp =
    paste("The mandatory argument threshold_value was",
          "not defined and is set to the default of 90%."))

  expect_error(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data, suppressWarnings = NA)
  ), regexp = "Argument suppressWarnings must not contain NAs")

  expect_error(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data, suppressWarnings = "xx")
  ), regexp =
    paste("Argument suppressWarnings must be logical"))

  expect_error(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data, suppressWarnings = 1:2)
  ), regexp = "Need exactly one element in argument suppressWarnings, got 2")

  w <- capture_warnings(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data, suppressWarnings = TRUE)
  ))
  w <- gsub("(\n|^|\r)+>.*$", "", w)
  expect_false(any(grepl("Setting suppressWarnings to its default FALSE", w
        )))

  expect_error(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data, suppressWarnings = c(TRUE,
                                                                   FALSE))
  ), regexp = "Need exactly one element in argument suppressWarnings, got 2")


  w <- capture_warnings(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data, suppressWarnings = c(TRUE))
  ))
  w <- gsub("(\n|^|\r)+>.*$", "", w)
  expect_false(any(grepl("Setting suppressWarnings to its default FALSE",
                         w)))


  expect_error(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data, suppressWarnings = TRUE,
                         cause_label_df = 42, threshold_value = .9)
  ), regexp =
    paste(".+cause_label_df.+ is not a data frame."))

  expect_message2(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data, suppressWarnings = TRUE,
                         threshold_value = "XX")
  ), regexp = paste(
    "Could not convert threshold_value .+XX.+ to a number.",
    "Set to default value 90%."), perl = TRUE)

  cause_label_df <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|missing_table")

  cause_label_df$CODE_VALUE[16] <-
    cause_label_df$CODE_VALUE[15]
  suppressMessages(suppressWarnings(expect_warning(invisible(
    com_item_missingness(study_data = study_data,
                         item_level = meta_data, suppressWarnings = FALSE,
                         cause_label_df = cause_label_df)
  ), regexp =
    "code.* with more than one meaning")))

  suppressMessages(expect_conditions({
    i1 <- com_item_missingness(study_data = study_data,
                               item_level = meta_data,)
                               },
                       regexps =
                         c(
                           paste("There are \\d+ meassurements",
                                 "of .+",
                                 "for participants not being",
                                 "part of one of the segments .+"),
                           paste("There are \\d+ meassurements",
                                 "of .+",
                                 "for participants not being",
                                 "part of one of the segments .+"),
                           paste("There are \\d+ meassurements",
                                 "of .+",
                                 "for participants not being",
                                 "part of one of the segments .+"),
                           paste("There are \\d+ meassurements",
                                       "of .+",
                                       "for participants not being",
                                       "part of one of the segments .+")
                         ),
                  classes = c("warning")))
  expect_lt(suppressWarnings(
    abs(
      sum(as.numeric(as.matrix(i1$SummaryTable)), na.rm = TRUE) -
        479149.9)), 50)

  code_labels <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx|missing_table")
  suppressMessages(expect_conditions({
    i2 <- com_item_missingness(study_data = study_data,
                               item_level = meta_data,
                                            cause_label_df = code_labels)
    },
                       regexps =
                         c(paste(".*more than one meaning.*"),
                           paste("Combining .+ and assignments in .+ in",
                                 ".+ is discouraged. This may cause",
                                 "errors."),
                           paste("There are \\d+ meassurements",
                                 "of .+",
                                 "for participants not being",
                                 "part of one of the segments .+"),
                           paste("There are \\d+ meassurements",
                                 "of .+",
                                 "for participants not being",
                                 "part of one of the segments .+"),
                           paste("There are \\d+ meassurements",
                                 "of .+",
                                 "for participants not being",
                                 "part of one of the segments .+"),
                           paste("There are \\d+ meassurements",
                                 "of .+",
                                 "for participants not being",
                                 "part of one of the segments .+"),
                           paste("The argument .+cause_label_df.+ has been",
                                 "deprecated. It",
                                 "will be in a future",
                                 "version be removed.")
                         ),
                  classes = c("warning")))
  expect_lt(suppressWarnings(
    abs(
      sum(as.numeric(as.matrix(i2$SummaryTable)), na.rm = TRUE) -
        479149.9)), 50)

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("item missingness no labels",
                              i1$SummaryPlot)
  expect_doppelganger2("item missingness w labels",
                              i2$SummaryPlot)
})

test_that("int_datatype_matrix works", {
  skip_on_cran() # slow and not so complicated. also, errors will be obvious.
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]

  expect_error({
    appmatrix <- int_datatype_matrix(study_data = study_data,
                                     meta_data = meta_data,
                                     label_col = LABEL,
                                     max_vars_per_plot = c())
    },
    regexp = "max_vars_per_plot must be one strictly positive"
  )

  expect_error(
    int_datatype_matrix(resp_vars = "SEX_0",
                        study_data = study_data,
                        meta_data = meta_data[,
                                              setdiff(colnames(meta_data),
                                                      "DATA_TYPE")],
                        label_col = LABEL),
    regexp =
      paste("Missing columns .+DATA_TYPE.+ from .+meta_data.+.")
  )

  md0 <- meta_data
  md0$DATA_TYPE[2] <- NA

  expect_warning(
    int_datatype_matrix(resp_vars = "SEX_0",
                        study_data = study_data,
                        meta_data = md0,
                        label_col = LABEL),
    regexp =
      paste("I've predicted the.+DATA_TYPE.+")
  )

  md0 <- meta_data
  md0$DATA_TYPE[2] <- "MY_TYPE"

  expect_silent(
    expect_warning(
      int_datatype_matrix(resp_vars = "SEX_0",
                          study_data = study_data,
                          meta_data = md0,
                          label_col = LABEL),
      regexp =
        "yielding.+v00001 = string"
    )
  )

  expect_warning(
    appmatrix <- int_datatype_matrix(
                                   study_data = study_data,
                                   meta_data = meta_data[,
                                                         setdiff(
                                                           colnames(meta_data),
                                                           c(STUDY_SEGMENT,
                                                             KEY_STUDY_SEGMENT)
                                                         ),
                                                         drop = FALSE],
                                   label_col = LABEL,
                                   split_segments = TRUE),
    regexp =  paste("Stratification for STUDY_SEGMENT is not possible",
                    "due to missing metadata.",
                    "Will split arbitrarily avoiding too large figures")
  )

  expect_equal(length(appmatrix$DataTypePlotList), 3)

  appmatrix <- int_datatype_matrix(resp_vars = "SEX_0",
                                   study_data = study_data,
                                   meta_data = meta_data,
                                   label_col = LABEL)

  expect_true(all(appmatrix$ReportSummaryTable$MATCH == 0))
  expect_equal(nrow(appmatrix$ReportSummaryTable), 1)
  expect_gte(ncol(appmatrix$ReportSummaryTable), 4)
  expect_identical(sort(gsub("[^a-zA-Z0-9]", "",
                             tolower(colnames(appmatrix$ReportSummaryTable)))),
                   sort(c("convertiblemismatchstable",
                          "convertiblemismatchunstable", "nonconvertiblemismatch",
                          "n", "variables" )))

  appmatrix <- int_datatype_matrix(study_data = study_data,
                                   meta_data = meta_data,
                                   label_col = LABEL)

  expect_equal(nrow(appmatrix$ReportSummaryTable), 53)
  expect_gte(ncol(appmatrix$ReportSummaryTable), 4)
  expect_true(
    all(appmatrix$ReportSummaryTable$`convertible mismatch, unstable` == 0))
  expect_true(
    all(appmatrix$ReportSummaryTable$`convertible mismatch, stable` == 0))
  expect_true(all(appmatrix$ReportSummaryTable$`nonconvertible mismatch` == 0))

  skip_on_cran()
  # TODO: skip_if_not(capabilities()["long.double"])
  skip_if_not_installed("vdiffr")

  expect_doppelganger2("integrity datatype",
                              appmatrix$SummaryPlot)

  for (n in names(appmatrix$DataTypePlotList)) {
    expect_doppelganger2(sprintf("intDt%s", n),
                                appmatrix$DataTypePlotList[[n]])
  }

})

test_that("pro_applicability_matrix works", {
  skip_on_cran() # deprecated
  
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]

  for (max_vars_per_plot in list(
    1:10, -1, -Inf, NA, NaN, complex(real = 1), "A", letters
  )) {
    expect_error(
      appmatrix <- pro_applicability_matrix(study_data = study_data,
                                            meta_data = meta_data,
                                            label_col = LABEL,
                                            max_vars_per_plot =
                                              max_vars_per_plot),
      regexp =
        paste("max_vars_per_plot must be one strictly positive",
              "non-complex integer value, may be Inf."),
      perl = TRUE)
  }

  for (max_vars_per_plot in list(
    20, Inf
  )) {
    expect_silent(
      appmatrix <- pro_applicability_matrix(study_data = study_data,
                                            meta_data = meta_data,
                                            label_col = LABEL,
                                            max_vars_per_plot =
                                              max_vars_per_plot))
  }

  md0 <- meta_data
  md0$DATA_TYPE <- NULL
  expect_error(
    appmatrix <- pro_applicability_matrix(study_data = study_data,
                                          meta_data = md0,
                                          label_col = LABEL,
                                          max_vars_per_plot =
                                            max_vars_per_plot),
    regexp =
      paste("Missing columns .+DATA_TYPE.+ from .+meta_data.+"),
    perl = TRUE)

  md0 <- meta_data
  md0$DATA_TYPE[[2]] <- NA
  expect_warning(
    appmatrix <- pro_applicability_matrix(study_data = study_data,
                                          meta_data = md0,
                                          label_col = LABEL,
                                          max_vars_per_plot =
                                            max_vars_per_plot),
    regexp =
      paste("yielding .+v00001 = string.+"),
    perl = TRUE)

  md0 <- meta_data
  md0$DATA_TYPE[[2]] <- "Ordinal"
  expect_warning(
      appmatrix <- pro_applicability_matrix(study_data = study_data,
                                            meta_data = md0,
                                            label_col = LABEL,
                                            max_vars_per_plot =
                                              max_vars_per_plot),
      regexp =
        paste("yielding .+v00001 = string.+"),
    perl = TRUE)

  expect_silent(
    appmatrix <- pro_applicability_matrix(study_data = study_data,
                                          meta_data = meta_data,
                                          label_col = LABEL,
                                          split_segments = TRUE)
  )

  md0 <- meta_data
  if (KEY_STUDY_SEGMENT %in% names(md0))
    md0[[KEY_STUDY_SEGMENT]][[2]] <- NA
  if (STUDY_SEGMENT %in% names(md0))
    md0[[STUDY_SEGMENT]][[2]] <- NA
  expect_message2(
    appmatrix <- pro_applicability_matrix(study_data = study_data,
                                          meta_data = md0,
                                          label_col = LABEL,
                                          split_segments = TRUE),
    regexp =
      paste("Some .+STUDY_SEGMENT.+ are NA.",
            "Will assign those to an artificial segment .+SEGMENT.+"),
    perl = TRUE
  )

  expect_message2(
    appmatrix <- pro_applicability_matrix(study_data = study_data,
                                          meta_data = meta_data,
                                          label_col = LABEL,
                                          split_segments = TRUE,
                                          max_vars_per_plot = 2),
    regexp =
      paste(".*Will split segemnt",
            ".+ arbitrarily avoiding too large figures"),
    perl = TRUE
  )


  md0 <- meta_data
  md0$KEY_STUDY_SEGMENT <- NULL
  md0$STUDY_SEGMENT <- NULL
  expect_message2(
    appmatrix <- pro_applicability_matrix(study_data = study_data,
                                          meta_data = md0,
                                          label_col = LABEL,
                                          split_segments = TRUE),
    regexp = paste("Stratification for STUDY_SEGMENT is not",
                   "possible due to missing metadata. Will split arbitrarily",
                   "avoiding too large figures"),
    perl = TRUE
  )

  appmatrix <- pro_applicability_matrix(study_data = study_data,
                                        meta_data = meta_data,
                                        label_col = LABEL)

  expect_length(appmatrix$ApplicabilityPlotList, 5)
  expect_lt(abs(suppressWarnings(sum(na.rm = TRUE,
    as.numeric(as.matrix(appmatrix$SummaryTable))
  )) - 3149), 5)

  skip_on_cran()
  skip_if_not_installed("vdiffr")
  # TODO: skip_if_not(capabilities()["long.double"])
  expect_doppelganger2("appmatrix plot ok",
                              appmatrix$ApplicabilityPlot)
  expect_doppelganger2("appmatrix segment v10000",
                              appmatrix$ApplicabilityPlotList$v10000)

})

.dq_lazy_state <- new.env(parent = emptyenv())

.dq_lazy_state$s7_ready <- FALSE
.dq_lazy_state$s7_class <- NULL

dq_lazy_unwrap <- function(x) {
  if (!inherits(x, "S7_object")) return(x)
  payload <- tryCatch(x@payload, error = function(e) NULL)
  if (!is.null(payload) && inherits(payload, "dq_lazy_ggplot")) return(payload)
  x
}

#' Support function to stop, if an optional package is not installed
#'
#' This function stops, if a package is not installed but needed for using an
#' optional feature of `dataquieR`.
#'
#' @param pkg needed package
#' @param err [logical] Should the function throw an error (default) or a
#'            warning?
#' @param goal feature description for error message.
#' @param and_import import the listed function to the caller's environment
#' @return `TRUE` if all packages in `pkg` are available, `FALSE` if at least
#'         one of the packages is missing.
#' @examples
#' \dontrun{ # internal use, only
#' f <- function() {
#'   util_ensure_suggested <- get("util_ensure_suggested",
#'     asNamespace("dataquieR"))
#'   util_ensure_suggested("ggplot2", "Test",
#'       and_import = "(ggplot|geom_.*|aes)")
#'   print(ggplot(cars, aes(x = speed)) + geom_histogram())
#' }
#' f()
#' }
#'
#' @family robustness_functions
#' @concept process
#' @noRd
util_ensure_suggested <- function(pkg, goal =
                                    ifelse(
                                      is.null(
                                        rlang::caller_call()),
                                      "work",
                                      paste("call", sQuote(rlang::call_name(
                                                      rlang::caller_call())))),
                                  err = TRUE, and_import = c()) {
  util_expect_scalar(err, check_type = is.logical)
  missingp <- !vapply(pkg,
                      FUN.VALUE = logical(1),
                      requireNamespace,
                      quietly = TRUE)
  if (err && any(missingp)) {
    rlang::check_installed(
      call = rlang::caller_env(),
      pkg = pkg,
      reason = paste("to", goal)
    )
    missingp <- !vapply(pkg,
                        FUN.VALUE = logical(1),
                        requireNamespace,
                        quietly = TRUE)
  }
  if (any(missingp)) {
    if (err) {
      lambda <- util_error
    } else {
      lambda <- util_warning
    }
    if (rlang::is_installed("cli")) {
      install_all <-
        cli::format_inline(
          sprintf("Call {.run [%s](%s)} to install all suggested packages",
                  "prep_check_for_dataquieR_updates()",
                  "dataquieR::prep_check_for_dataquieR_updates()")
        )
    } else {
      install_all <-
        sprintf("Call %s to install all suggested packages",
                dQuote("prep_check_for_dataquieR_updates()"))
    }
    lambda(c("Missing the package(s) %s to %s.",
             "Install with install.packages(%s).",
             install_all),
           paste0(dQuote(pkg[missingp]), collapse = ", "),
           goal,
           deparse(pkg[missingp]))
    return(FALSE)
  } else {
    if (length(and_import) > 0) {
      caller_env <- parent.frame()
      for (nm in ls(
        pattern = and_import, name = asNamespace(pkg))) {
        if (nm %in% getNamespaceExports(pkg)) {
          assign(nm, get(nm, asNamespace(pkg)), caller_env)
        }
      }
    }
    return(TRUE)
  }
}

.util_optional_packages <- new.env(parent = emptyenv())

util_have_suggested <- function(pkg, goal =
                                  ifelse(
                                    is.null(
                                      rlang::caller_call()),
                                    "work",
                                    paste("call", sQuote(rlang::call_name(
                                      rlang::caller_call()))))) {
  if (!exists(pkg, envir = .util_optional_packages, inherits = FALSE)) {
    assign(pkg,
           suppressWarnings(suppressMessages(
             util_ensure_suggested(pkg, goal = goal, err = FALSE)
           )),
           envir = .util_optional_packages)
  }

  get(pkg, envir = .util_optional_packages, inherits = FALSE)
}

test_that("util_ensure_suggested works", {
  skip_on_cran()
  ip <- installed.packages()
  existing <- "ggplot2"
  if (!(existing %in% ip[, "Package"])) {
    fail(paste("Package", dQuote(existing),
               "is a dependency as of writing this.",
               "Therefore, it is assumed to be an installed package here.",
               "However, it is missing."))
  }
  found <- FALSE
  i <- 1
  while (!found) { # look for an uninstalled package name
    unexist  <- sprintf("%s_%d", existing, i)
    i <- i + 1
    found <- !(unexist %in% ip[, "Package"])
  }
  expect_silent(util_ensure_suggested(existing, "test the function"))
  expect_error(util_ensure_suggested(unexist, "test the function"),
              regexp = "The package.+is required to\\s+test\\s+the\\s+function",
               perl = TRUE)
})

#' @inherit resnames
#'
#' @export
# importFrom dataquieR resnames
resnames.dataquieR_resultset2 <- function(x) {
  rsn <- attr(x, "resnames")
  if (!is.null(rsn)) {
    return(rsn)
  } else {
    return(sort(unique(unname(unlist(lapply(x, names))))))
  }
}

#' Generate function calls for a given indicator function
#'
#' new reporting pipeline v2.0
#'
#' @param fkt the indicator function's name
#' @param meta_data the item level metadata data frame
#' @param label_col the label column
#' @param meta_data_segment segment level metadata
#' @param meta_data_dataframe data frame level metadata
#' @param meta_data_cross_item cross-item level metadata
#' @param specific_args argument overrides for specific functions
#' @param arg_overrides general argument overrides
#' @param resp_vars variables to be respected
#' @param ssi_functions [vector] functions for `SSI-computation`
#' @param non_ssi_functions  functions computation excluding `SSI-computation`
#'
#' @return function calls for the given function
#'
#' @family reporting_functions
#' @concept process
#' @noRd
util_generate_calls_for_function <-
  function(fkt,
           meta_data,
           label_col,
           meta_data_segment,
           meta_data_dataframe,
           meta_data_cross_item,
           specific_args,
           arg_overrides,
           resp_vars,
           ssi_functions,
           non_ssi_functions) { # TODO: Document
  .meta_data_env$fkt <- fkt
  .meta_data_env$meta_data <- meta_data
  .meta_data_env$label_col <- label_col
  .meta_data_env$meta_data_segment <- meta_data_segment
  .meta_data_env$meta_data_dataframe <- meta_data_dataframe
  .meta_data_env$meta_data_cross_item <- meta_data_cross_item
  .meta_data_env$apply_to_ssi <- (fkt %in% ssi_functions)
  .meta_data_env$apply_to_nssi <- (fkt %in% non_ssi_functions)
  on.exit({
    .meta_data_env$fkt <- NULL
    .meta_data_env$meta_data <- NULL
    .meta_data_env$label_col <- NULL
    .meta_data_env$meta_data_segment <- NULL
    .meta_data_env$meta_data_dataframe <- NULL
    .meta_data_env$meta_data_cross_item <- NULL
    .meta_data_env$target_meta_data <- NULL
    .meta_data_env$apply_to_ssi <- NULL
    .meta_data_env$apply_to_nssi <- NULL
  })
  .to_fill <- formals(fkt)
  to_fill <- list()
  to_fill[intersect(names(.to_fill), names(arg_overrides))] <-
    arg_overrides[intersect(names(.to_fill), names(arg_overrides))]
  to_fill[intersect(names(.to_fill), names(specific_args[[fkt]]))] <-
    specific_args[[fkt]][
      intersect(names(.to_fill), names(specific_args[[fkt]]))]
  if ("study_data" %in% names(.to_fill)) {
    to_fill[["study_data"]] <- quote(study_data)
  }
  if ("meta_data_cross_item" %in% names(.to_fill)) {
    to_fill[["meta_data_cross_item"]] <- quote(meta_data_cross_item)
  }
  if ("meta_data_dataframe" %in% names(.to_fill)) {
    to_fill[["meta_data_dataframe"]] <- quote(meta_data_dataframe)
  }
  if ("meta_data_segment" %in% names(.to_fill)) {
    to_fill[["meta_data_segment"]] <- quote(meta_data_segment)
  }
  if ("meta_data" %in% names(.to_fill)) {
    to_fill[["meta_data"]] <- quote(meta_data)
  }
  if ("label_col" %in% names(.to_fill)) {
    to_fill[["label_col"]] <- label_col
  }
  if ("resp_vars" %in% names(.to_fill)) {
    if ("variable_group" %in% names(.to_fill)) {
      util_error("")
    }
    .meta_data_env$target_meta_data <- "item_level"
    lapply(setNames(nm = resp_vars), function(rv) {
      fillers <- names(.meta_data_env)
      fillers <- fillers[vapply(FUN.VALUE = logical(1),
                                fillers,
                                function(f) {
                                  is.function(.meta_data_env[[f]])
                                })]
      can_fill <- intersect(names(.to_fill), fillers)
      to_fill[can_fill] <- # TODO: Find a solution to deliver NULL as missing.
        lapply(setNames(nm = can_fill), function(filler) {
          .meta_data_env[[filler]](rv)
        })
      # to_fill$resp_vars <- rv
      to_fill
    })
  } else if ("variable_group" %in% names(.to_fill)) {
    if ("resp_vars" %in% names(.to_fill)) {
      util_error("")
    }
    .meta_data_env$target_meta_data <- "cross-item_level"
    ck_id <- meta_data_cross_item[[CHECK_ID]]
    ck_id[util_empty(ck_id)] <- NA_character_
    nm <- meta_data_cross_item[[CHECK_LABEL]]
    nm[util_empty(nm)] <- NA_character_
    if (any(is.na(nm))) {
      util_message("Removing rows from %s because %s is missing.",
                   sQuote("cross-item_level"),
                   sQuote(CHECK_LABEL),
                   applicability_problem = TRUE)
    }
    ck_id <- setNames(ck_id[!is.na(nm) & ! is.na(ck_id)],
                   nm = nm[!is.na(nm) & ! is.na(ck_id)])
    # TODO: only select ck_id that incorporate variables that match var_list
    lapply(ck_id, function(ci) {
      if (!util_empty(ci)) {
        on.exit(.meta_data_env$meta_data_cross_item <- meta_data_cross_item)
        .meta_data_env$meta_data_cross_item <- meta_data_cross_item[
          (is.na(meta_data_cross_item[[CHECK_ID]]) & is.na(ci)) |
            (meta_data_cross_item[[CHECK_ID]] == ci), , drop = FALSE]
        fillers <- names(.meta_data_env)
        fillers <- fillers[vapply(FUN.VALUE = logical(1),
                                  fillers,
                                  function(f) {
                                    is.function(.meta_data_env[[f]])
                                  })]
        can_fill <- intersect(names(.to_fill), fillers)
        to_fill[can_fill] <- # TODO: Find a solution to deliver NULL as missing.
          lapply(setNames(nm = can_fill), function(filler) {
            .meta_data_env[[filler]](ci)
          })

        vg <- .meta_data_env$meta_data_cross_item[[VARIABLE_LIST]]

        variable_group <- names(util_parse_assignments(vg))
        variable_group <- util_find_var_by_meta(variable_group, ## TODO: Still needed? Has normalized always been called before??
                                                meta_data,
                                                label_col = label_col,
                                                target = label_col,
                                                ifnotfound = variable_group)
        to_fill$variable_group <- variable_group
      }
      to_fill
    })
  } else {
    # util_error("") not an error, may be a function w/o entity form the int dim
    list(`[ALL]` = to_fill)
  }
}

#' Validate code lists for missing and/or jump codes
#'
#' will warn/stop on problems
#'
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param cause_label_df [data.frame] missing code table. If missing codes have
#'                                    labels the respective data frame can be
#'                                    specified here, see [cause_label_df]
#' @param assume_consistent_codes [logical] if TRUE and no labels are given and
#'                                          the same missing/jump code is used
#'                                          for more than one variable, the
#'                                          labels assigned for this code will
#'                                          be the same for all variables.
#' @param expand_codes [logical] if TRUE, code labels are copied from other
#'                               variables, if the code is the same and the
#'                               label is set somewhere
#' @param suppressWarnings [logical] warn about consistency issues with missing
#'                                   and jump lists
#'
#'
#' @return [list] with entries:
#'   - `cause_label_df` updated data frame with labels for missing codes
#'
#' @family metadata_management
#' @concept robustness
#' @noRd


util_validate_missing_lists <-
  function(meta_data, cause_label_df,
           assume_consistent_codes = FALSE,
           expand_codes = FALSE,
           suppressWarnings = FALSE,
           label_col) {
    if (missing(label_col)) {
      label_col <- WELL_KNOWN_META_VARIABLE_NAMES$VAR_NAMES
    }
    if (missing(meta_data) ||
        !is.data.frame(meta_data) ||
        !(all(c(MISSING_LIST, label_col) %in% colnames(meta_data)))) {
      util_error("Need at least %s with columns %s and %s",
                 dQuote("meta_data"),
                 dQuote(MISSING_LIST),
                 dQuote(label_col),
                 applicability_problem = TRUE)
    }

    if (!(JUMP_LIST %in% colnames(meta_data))) {
      meta_data[[JUMP_LIST]] <- NA_character_
    }

    cause_label_df_given <- !missing(cause_label_df)

    if (!cause_label_df_given) {
      cause_label_df <- data.frame(
        CODE_VALUE = NA,
        CODE_LABEL = NA,
        stringsAsFactors = FALSE
      )[FALSE, , FALSE]
    }

    not_empty <- function(x) {
      !is.na(x) & trimws(gsub(SPLIT_CHAR, "", x, fixed = TRUE)) != ""
    }

    is_assignment <- function(x) {
      grepl("=", x, fixed = TRUE)
    }

    if (!suppressWarnings && cause_label_df_given &&
        (
          any(is_assignment(meta_data$JUMP_LIST)) ||
          any(is_assignment(meta_data$MISSING_LIST))
        )) {
      util_warning(c("Combining %s and assignments in %s and/or %s",
                     "in %s is discouraged. This may cause errors."),
                   dQuote("cause_label_df"),
                   dQuote(MISSING_LIST),
                   dQuote(JUMP_LIST),
                   dQuote("meta_data"),
                   applicability_problem = TRUE)
    }

    if (!suppressWarnings &&
        (
          any(is_assignment(meta_data$JUMP_LIST)) ||
          any(is_assignment(meta_data$MISSING_LIST))
        ) &&
        (
          any(not_empty(meta_data$JUMP_LIST) &
             !is_assignment(meta_data$JUMP_LIST)) ||
         any(not_empty(meta_data$MISSING_LIST) &
             !is_assignment(meta_data$MISSING_LIST))
        )
      ) {
      util_warning(c("%s and assignment notation in %s or %s in %s have been",
                     "combined. This is discouraged",
                     "and may cause unexpected results. Please use %s to add",
                     "all entries from %s to %s"),
                   dQuote("cause_label_df"),
                   dQuote(MISSING_LIST),
                   dQuote(JUMP_LIST),
                   dQuote("meta_data"),
                   dQuote("prep_add_cause_label_df"),
                   dQuote("cause_label_df"),
                   dQuote("meta_data"),
                   applicability_problem = TRUE)
    }

    if (!is.data.frame(cause_label_df) ||
        !(all(c(CODE_VALUE, CODE_LABEL) %in% colnames(cause_label_df)))) {
      util_message(
        c("Need columns %s in %s, which must be a data frame if given.",
          "Will ignore this argument"),
        paste(dQuote(c(CODE_VALUE, CODE_LABEL)), collapse = ", "),
        dQuote("cause_label_df"),
        applicability_problem = TRUE)
      cause_label_df <- data.frame(
        CODE_VALUE = NA,
        CODE_LABEL = NA,
        stringsAsFactors = FALSE
      )[FALSE, , FALSE]
    }

    if (CODE_CLASS %in% colnames(cause_label_df)) {
      if (!suppressWarnings &&
          !all(cause_label_df$CODE_CLASS %in% c("MISSING", "JUMP", NA))) {
        util_error("Only %s and %s are valid values for the column %s in %s.",
                   dQuote("MISSING"),
                   dQuote("JUMP"),
                   dQuote(CODE_CLASS),
                   dQuote("cause_label_df"),
                   applicability_problem = TRUE)
      }
    }

    if (!suppressWarnings &&
       (any(is.na(cause_label_df$CODE_VALUE)) ||
        any(is.na(cause_label_df$CODE_LABEL)))) {
      util_warning("Some code labels or -values are missing from %s.",
                   dQuote("cause_label_df"),
                   applicability_problem = TRUE)
    }

    parsed_lists_md <-  unlist(
      util_parse_assignments(c(meta_data$JUMP_LIST,
                               meta_data$MISSING_LIST),
                               multi_variate_text = TRUE),
      recursive = TRUE)
    parsed_lists_cldf <- setNames(cause_label_df$CODE_LABEL,
                     nm = cause_label_df$CODE_VALUE)
    parsed_lists <- c(parsed_lists_md,
                      parsed_lists_cldf)
    parsed_lists <- parsed_lists[!is.na(parsed_lists)]

    not_numbers <-
      (
         is.na(suppressWarnings(as.numeric(names(parsed_lists)))) &
         suppressWarnings(vapply(lapply(
           names(parsed_lists), util_parse_date), is.na,
                FUN.VALUE = logical(1))) &
           suppressWarnings(vapply(lapply(
             names(parsed_lists), util_parse_time), is.na,
             FUN.VALUE = logical(1)))
      ) !=
      (is.na(names(parsed_lists)) |
      trimws(names(parsed_lists)) == "" |
      trimws(names(parsed_lists)) == "NA")

    if (!suppressWarnings && any(not_numbers)) {
      util_warning(c("Some missing codes are not numeric or date/time. This is",
                     "not supported yet: %s"),
                   paste(dQuote(names(parsed_lists[not_numbers])),
                         collapse = ", "),
                   applicability_problem = TRUE)
    }

    if (cause_label_df_given) {
      not_in_cldf <- sort(unique(names(parsed_lists_md)[!(names(parsed_lists_md) %in% c("NA", cause_label_df$CODE_VALUE))]))
      if (!suppressWarnings && length(not_in_cldf) > 0) {
        util_warning("Found jump/missing codes in %s not mentioned in %s: %s",
                     dQuote("meta_data"),
                     dQuote("cause_label_df"),
                     paste(sQuote(not_in_cldf), collapse = ", "))
      }

      not_in_md <- unique(sort(names(parsed_lists_cldf)[!(names(parsed_lists_cldf) %in% c("NA", names(parsed_lists_md)))]))
      if (!suppressWarnings && length(not_in_md) > 0) {
        util_warning("Found jump/missing codes in %s not mentioned in %s: %s",
                     dQuote("cause_label_df"),
                     dQuote("meta_data"),
                     paste(sQuote(not_in_md), collapse = ", "))
      }
    }

    parsed_lists <- parsed_lists[!not_numbers &
                                   trimws(names(parsed_lists)) != ""]

    # not_named <- parsed_lists[
    #   as.numeric(names(parsed_lists)) == parsed_lists]
    #
    # if (!suppressWarnings &&
    #     length(not_named)) {
    #   util_warning("Some code labels or -values are not named: %s.",
    #                paste(sQuote(unique(not_named)),
    #                  collapse = ", "),
    #                applicability_problem = TRUE)
    # }

    if (!"resp_vars" %in% colnames(cause_label_df)) {
      nm <- meta_data[, label_col, FALSE]
      colnames(nm) <- "resp_vars"
      cause_label_df <- merge(cause_label_df,
                              nm,
                              by = c(),
                              all = TRUE)
      if (colnames(cause_label_df)[ncol(cause_label_df)] == "y[FALSE, ]") {
        colnames(cause_label_df)[ncol(cause_label_df)] <- "resp_vars"
      }
    }

    if (!CODE_CLASS %in% colnames(cause_label_df)) {
      cause_label_df$CODE_CLASS <- rep("MISSING", nrow(cause_label_df))
    }

    cldf2 <- prep_extract_cause_label_df(meta_data = meta_data,
                                         label_col = label_col)$cause_label_df

    if (!!prod(dim(cause_label_df)) &&
        !!prod(dim(cldf2)))
    cause_label_df <- util_rbind(
      cause_label_df,
      cldf2
    ) else if (!prod(dim(cause_label_df)) &&
               !!prod(dim(cldf2))) {
      cause_label_df <- cldf2
    }

    cause_label_df <- cause_label_df[!duplicated(cause_label_df), , FALSE]

    vars_with_dups <-
      vapply(split(cause_label_df,
                   cause_label_df$resp_vars),
             function(x) {
               x <- x[!duplicated(x), , FALSE]
               y <- x$CODE_VALUE
               anyDuplicated(y) > 0
             },
             FUN.VALUE = logical(1))

    if (!suppressWarnings && any(vars_with_dups)) {
      util_warning(
        "Found at least one missing code with more than one meaning for %s.",
        paste(dQuote(names(vars_with_dups[vars_with_dups])), collapse = ", "),
        applicability_problem = TRUE
        )
    }

    if (assume_consistent_codes) {
      cld <-
        cause_label_df[, c(CODE_VALUE, CODE_LABEL, CODE_CLASS), FALSE]
      cld <- cld[!duplicated(cld), , FALSE]
      d <- duplicated(cld$CODE_VALUE)
      if (any(d)) {
        dd <- unique(cld[d, CODE_VALUE, TRUE])
        if (!suppressWarnings) {
          util_warning(
            "Found missing code(s) with more than one meaning:\n%s",
            paste(capture.output(
              print(cld[cld$CODE_VALUE %in% dd, , FALSE])),
              collapse = "\n"),
            applicability_problem = TRUE
          )
        }
      }
    }

    if (assume_consistent_codes) {
      cause_label_df$AUTO <- cause_label_df$CODE_LABEL ==
        paste(cause_label_df$CODE_CLASS, cause_label_df$CODE_VALUE)
    } else {
      cause_label_df$AUTO <- cause_label_df$CODE_LABEL ==
        paste(cause_label_df$CODE_CLASS, cause_label_df$resp_vars, cause_label_df$CODE_VALUE)
    }

    if (expand_codes) {
      s_cause_label_df <- split(cause_label_df,
            list(cause_label_df$CODE_VALUE)) # , cause_label_df$CODE_CLASS
      warn_expand <- function(cldf) {
        my_labels <- cldf[!cldf$AUTO, CODE_LABEL, TRUE]
        my_labels <- my_labels[!is.na(my_labels)]
        if (!suppressWarnings && length(unique(my_labels)) == 1) {
          util_message("Would use label %s for all values coded with %s",
                       dQuote(unique(my_labels)),
                       dQuote(unique(cldf$CODE_VALUE)),
                       applicability_problem = TRUE)
        }
      }
      lapply(s_cause_label_df, warn_expand)
    }
    invisible(list(cause_label_df = cause_label_df))
}

#' Detect Expected Observations
#'
#' For each participant, check, if an observation was expected, given the
#' `PART_VARS` from item-level metadata
#'
#' [Descriptor]
#'
#' @param study_data [study_data] must have all relevant `PART_VARS` to avoid
#'                                false-positives on `PART_VARS` missing from
#'                                `study_data`
#' @param item_level [meta_data] must be complete to avoid false positives on
#'                              non-existing `PART_VARS`
#' @param meta_data [data.frame] old name for `item_level`
#' @param label_col [character] mapping attribute `colnames(study_data)` vs.
#'                              `meta_data[label_col]`
#' @param expected_observations [enum] HIERARCHY | SEGMENT. How should
#'                                     `PART_VARS` be handled:
#'                                     - `SEGMENT`: if `PART_VAR` is 1, an
#'                                       observation is expected
#'                                     - `HIERARCHY`: the default, if the
#'                                       `PART_VAR` is 1 for this variable and
#'                                       also for all `PART_VARS` of `PART_VARS`
#'                                       up in the hierarchy, an observation is
#'                                       expected.
#' @param disclose_problem_paprt_var_data [logical] show the problematic data
#'                                        (`PART_VAR` only)
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return empty list, so far -- the function only warns.
#' @export
int_part_vars_structure <- # TODO: Support segment level metadata links to SEGMENT_PART_VARS
  function(label_col,
           study_data,
           item_level = "item_level",
           expected_observations = c("HIERARCHY",
                                     "SEGMENT"),
           disclose_problem_paprt_var_data = FALSE,
           meta_data = item_level,
           meta_data_v2) {
  util_maybe_load_meta_data_v2()
  util_expect_scalar(disclose_problem_paprt_var_data, check_type = is.logical)

  util_expect_scalar(expected_observations, allow_more_than_one = TRUE)
  expected_observations <- util_match_arg(expected_observations)
  util_expect_scalar(expected_observations)

  prep_prepare_dataframes()

  if (!PART_VAR %in% colnames(meta_data)) {
    meta_data[[PART_VAR]] <- NA_character_
  }

  pv <- meta_data[, PART_VAR]
  pv <- pv[!util_empty(pv)]
  if (length(pv) < nrow(meta_data)) {
    util_warning("Found %d variables w/o %s in %s",
                 nrow(meta_data) - length(pv),
                 sQuote("PART_VARS"),
                 sQuote("meta_data"))
  }

  if (!all(pv %in%
           c(meta_data[[label_col]],
             meta_data[[VAR_NAMES]],
             meta_data[[LABEL]]
           ))) {
    ms <- !(pv %in%
      c(meta_data[[label_col]],
        meta_data[[VAR_NAMES]],
        meta_data[[LABEL]]
      ))

    util_warning("Missing %d %s from the %s: %s",
                 sum(ms),
                 sQuote("PART_VARS"),
                 sQuote("meta_data"),
                 paste0(dQuote(pv[ms]), collapse = ", "))
  }

  if (!all(pv %in%
           c(colnames(ds1),
             meta_data[meta_data[[label_col]] %in% colnames(ds1),
                       VAR_NAMES, drop = TRUE],
             meta_data[meta_data[[label_col]] %in% colnames(ds1),
                       LABEL, drop = TRUE]
             )
           )) {
    ms <- !(pv %in%
              c(colnames(ds1),
                meta_data[meta_data[[label_col]] %in% colnames(ds1),
                          VAR_NAMES, drop = TRUE],
                meta_data[meta_data[[label_col]] %in% colnames(ds1),
                          LABEL, drop = TRUE]
              ))

    util_warning("Missing %d %s from the %s: %s",
                 sum(ms),
                 sQuote("PART_VARS"),
                 sQuote("study_data"),
                 paste0(dQuote(pv[ms]), collapse = ", "))
  }

  represening_rvs <- sort(unique(util_map_labels(pv,
                                     meta_data,
                                     to = label_col,
                                     from = PART_VAR)))


  r <- lapply(represening_rvs, function(rv) {

    all_need_to_be_1 <- # and the order is from root to leaf in the PART_VAR
                        # hierarchy
      util_all_intro_vars_for_rv(rv, ds1, meta_data, label_col,
                                 expected_observations = expected_observations)

    missing_vars <- setdiff(all_need_to_be_1, colnames(ds1))
    if (any(missing_vars)) {
      util_warning(c("Missing %s from %s, I fill it with NA.",
                     "This may cause inconsistencies, if below in the",
                     "hierarchy, something is expected"),
                   paste0(dQuote(missing_vars), collapse = ", "),
                   sQuote("meta_data"))
      ds1[, missing_vars] <- NA
    }
    ds1 <- ds1[, all_need_to_be_1, drop = FALSE]

## FIXME TOO SLOW
    # all part_bits must start with 1*(0|NA)* ("" contained implicitly.)
    # part_bits <- apply(ds1, 1, paste0, collapse = "")
    # structure_damaged <- !grepl("1*(0|NA)*", part_bits, perl = TRUE)
    # part_bits <- apply(ds1, 1, function(part_bits_vector) {
    #   # but handle NA, "" better and return always 1 character per var
    #   part_bits_vector[util_is_na_0_empty_or_false(part_bits_vector)] <- 0
    #   paste0(part_bits_vector, collapse = "")
    # })
    # structure_damaged <- !grepl("1*0*", part_bits, perl = TRUE)


    test_part_bits <- function(part_bits_vector) { # TODO: Not here, write an Integrity function
      # tst_vld <- c(1, 1, 1, 0, 0, NA, 0)
      # tst_ivl <- c(1, 1, 1, 0, 0, 1, 0)
      # part_bits_vector <- tst_vld
      # test_part_bits(part_bits_vector)
      # part_bits_vector <- tst_ivl
      # test_part_bits(part_bits_vector)
      part_bits_vector <- !util_is_na_0_empty_or_false(part_bits_vector)
      seqs <- rle(part_bits_vector)
      ( ((any(head(seqs$values, 1) == FALSE) && (length(seqs$values)) > 1)) ||
          any(tail(seqs$values, -1))
      )
    }
    structure_damaged <- apply(ds1, 1, test_part_bits)


    if (any(structure_damaged)) { # TODO: test this
      case_data <- case_data <- ds1[structure_damaged, , FALSE]
      case_data <- paste0(capture.output(print(case_data)), collapse = "\n")
      case_data <- paste0(":\n", case_data, "\n")
      if (!disclose_problem_paprt_var_data) {
        case_data <- ""
      }
      util_warning(c("In %s, I found %d inconsistencies in %s hiearchy, an",
                     "observation of a sub-segment cannot be expected, if the",
                     "segment where the sub-segment belongs to is not.",
                     "Found such cases, e.g., in lines %s%s of the %s%s"),
                   dQuote(util_map_labels(rv, meta_data, from = label_col,
                                          to = STUDY_SEGMENT)),
                   sum(structure_damaged),
                   dQuote(PART_VAR),
                   paste(head(which(structure_damaged), 5), collapse = ", "),
                   (if (sum(structure_damaged, na.rm = TRUE) > 5) ", ..."
                    else ""),
                   sQuote("study_data"),
                   case_data)
    }
  })
  list()
}

#' String check for results/combined results
#'
#' detect, if `x` starts with `<prefix>.` or equals `<prefix>`,
#' if results have been combined
#'
#' @param x [character] haystack
#' @param prefix [character] needle
#' @param sep [character] separation string
#'
#' @return [logical] if entries in x start with prefix-DOT/equal to prefix
#'
#' @noRd
util_startsWith_prefix._or_equals_prefix <- function(x, prefix, sep = ".") {
  startsWith(x, paste0(prefix, sep)) || x == prefix
}

test_that("util_startsWith_prefix._or_equals_prefix works", {
  skip_on_cran()

  # String that should start with the prefix
  expect_true(util_startsWith_prefix._or_equals_prefix("foo.baz", "foo"))

  # String that matches identically
  expect_true(util_startsWith_prefix._or_equals_prefix("foo", "foo"))

  # Strings that should not start with the prefix
  expect_false(util_startsWith_prefix._or_equals_prefix("baz.foo", "foo"))
  expect_false(util_startsWith_prefix._or_equals_prefix("quux", "foo"))
  expect_false(util_startsWith_prefix._or_equals_prefix("", "foo"))

  # Edge case
  expect_error(util_startsWith_prefix._or_equals_prefix(NA, "foo"))
})

#' Estimates variance components
#'
#' @description
#' Variance based models and intraclass correlations (ICC) are approaches to
#' examine the impact of so-called process variables on the measurements. This
#' implementation is model-based.
#'
#' **NB:** The term ICC is frequently used to describe the agreement between
#' different observers, examiners or even devices. In respective settings a good
#' agreement is pursued. ICC-values can vary between `[-1;1]` and an ICC close
#' to 1 is desired (Koo and Li 2016, Müller and Büttner 1994).
#'
#' However, in multi-level analysis the ICC is interpreted differently. Please
#' see Snijders et al. (Sniders and Bosker 1999). In this context the proportion
#' of variance explained by respective group levels indicate an influence of (at
#' least one) level of the respective group_vars. An ICC close to 0 is desired.
#'
#' @details
#' # ALGORITHM OF THIS IMPLEMENTATION:
#'
#' - This implementation is yet restricted to data of type float.
#' - Missing codes are removed from resp_vars (if defined in the metadata)
#' - Deviations from limits, as defined in the metadata, are removed
#' - A linear mixed-effects model is estimated for resp_vars using co_vars and
#'   group_vars for adjustment.
#' - An output data frame is generated for group_vars indicating the ICC.
#'
#' @importFrom stats as.formula na.omit median
#'
#' @param resp_vars [variable list] the names of the continuous measurement
#'                                  variables
#' @param group_vars [variable list] the names of the resp. observer, device or
#'                                   reader variables
#' @param co_vars [variable list] a vector of covariables, e.g. age and sex for
#'                                adjustment
#' @param min_obs_in_subgroup [integer] from=0. optional argument if a
#'                                   "group_var" is used. This argument
#'                                   specifies the minimum no. of observations
#'                                   that is required to include a subgroup
#'                                   (level) of the "group_var" in the analysis.
#'                                   Subgroups with fewer observations are
#'                                   excluded. The default is 30.
#' @param min_subgroups [integer] from=0. optional argument if a "group_var" is
#'                                        used. This argument specifies the
#'                                        minimum no. of subgroups (levels)
#'                                        included "group_var". If the variable
#'                                        defined in "group_var" has fewer
#'                                        subgroups it is not used for analysis.
#'                                        The default is 5.
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return a list with:
#'   - `SummaryTable`: data frame with ICCs per `rvs`
#'   - `SummaryData`: data frame with ICCs per `rvs`
#'   - `ScalarValue_max_icc`: maximum variance contribution value by group_vars
#'   - `ScalarValue_argmax_icc`: variable with maximum variance contribution by
#'                               group_vars
#'
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_varcomp.html
#' )
#'
#' @noRd
#'
util_acc_varcomp <-
  function(resp_vars = NULL,
           label_col = NULL,
           study_data,
           item_level = "item_level",
           group_vars,
           co_vars = NULL,
           min_obs_in_subgroup = 30,
           min_subgroups = 5,
           meta_data = item_level,
           meta_data_v2) { # TODO: does this function remove observers w/o observations?

  # preps ----------------------------------------------------------------------
  util_maybe_load_meta_data_v2()

  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = TRUE)

  .min_obs_in_subgroup <- suppressWarnings(as.integer(min_obs_in_subgroup))
  if (is.na(.min_obs_in_subgroup)) {
    util_message(c(
    "Could not convert min_obs_in_subgroup %s to a number.",
    "Set to standard value."
    ),
     dQuote(as.character(min_obs_in_subgroup)),
    applicability_problem = TRUE
   )
    min_obs_in_subgroup <- 30
  } else {
    min_obs_in_subgroup <- .min_obs_in_subgroup
  }

  .min_subgroups <- suppressWarnings(as.integer(min_subgroups))
  if (is.na(.min_subgroups)) {
    util_message(
      "Could not convert min_subgroups %s to a number. Set to standard value.",
      dQuote(as.character(min_subgroups)),
      applicability_problem = TRUE
    )
    min_subgroups <- 5
  } else {
    min_subgroups <- .min_subgroups
  }

  # correct variable use?
  util_correct_variable_use("resp_vars",
    allow_more_than_one = TRUE,
    allow_null = TRUE,
    need_type = "float",
    need_scale = "interval|ratio"
  )

  util_correct_variable_use("group_vars",
    allow_null = TRUE,
    allow_any_obs_na = TRUE,
    allow_more_than_one = !TRUE, # TODO: Remove the whole functionality.
    need_type = "!float",
    need_scale = "nominal|ordinal"
  )

  util_correct_variable_use("co_vars",
                            overwrite = TRUE,
                            remove_not_found = TRUE,
                            allow_na = TRUE,
                            allow_more_than_one = TRUE,
                            allow_null = TRUE
                            )

  rvs <- resp_vars # TODO: Fix rvs <-> resp_vars

  if (length(rvs) == 0) {
    rvs <- colnames(ds1[, vapply(ds1, is.numeric, FUN.VALUE = logical(1))])
    rvs <- setdiff(rvs, group_vars)
    rvs <- setdiff(rvs, co_vars)
    if (length(group_vars) != 1)
      util_error(
        "Need exactly 1 group_vars, if all applicable resp_vars are used",
        applicability_problem = TRUE)
  }

  if (length(group_vars) == 1 && length(rvs) > 1) {
    group_vars <- rep(group_vars, length(rvs))
    if (!.called_in_pipeline)
      util_message(sprintf("using the same group var %s for all resp_vars",
                    dQuote(group_vars[[1]])
                    ))
  }

  if (length(group_vars) != length(rvs)) {
    util_error(
      c("acc_varcomp expects one group_var per resp_var. Here,",
        "it has been called with %d resp_vars but %d group_vars"),
      length(rvs), length(group_vars),
      applicability_problem = TRUE)
  }

  ds1[, unique(group_vars)] <- lapply(ds1[, unique(group_vars), drop = FALSE],
                                      factor)

  # (3) if no covariables are defined for adjustment only the intercept is
  #     modelled
  if (is.null(co_vars) || length(co_vars) == 0) {
    co_vars <- "1"
    cv <- character(0)
  } else {
    cv <- co_vars
    co_vars <- util_bQuote(co_vars)
  }

  # missing checks
  # -> encoded missings or jumps
  # -> resp_variables
  # -> variable types (factor(), ...)


  icc_output <- apply(cbind(rvs, group_vars), 1, function(row) {
    rv <- row[[1]]
    group_var <- row[[2]]

    check_df <- util_table_of_vct(ds1[[group_var]])

    # too few observations in >1 level of the random effect
    critical_levels <- levels(check_df$Var1)[check_df$Freq <
                                               min_obs_in_subgroup]
    if (length(critical_levels) > 0) {
      util_message("Levels %s were excluded due to fewer than %d observations.",
                   paste0(c(vapply(head(critical_levels, 10), dQuote, ""),
                            if (length(critical_levels) <= 10)
                              character(0) else "..."),
                          collapse = ", "),
                   min_obs_in_subgroup,
                   applicability_problem = FALSE)
      # exclude levels with too few observations
      ds1 <- ds1[!(ds1[[group_var]] %in% critical_levels), ]
      # dropping unused levels
      ds1[[group_var]] <- factor(ds1[[group_var]])
    }

    check_df <-
      util_table_of_vct(ds1[[group_var]]) # number of observers may have changed

    # fewer than min_subgroups levels of random effect
    if (length(check_df[, 1]) < min_subgroups) {
      util_message("%d < %d levels in %s. Will not compute ICCs for %s.",
                   length(check_df[, 1]),
                   min_subgroups,
                   dQuote(group_var),
                   dQuote(rv),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      return(data.frame(
        Variables = NA,
        Object = NA,
        Model.Call = NA,
        ICC = NA,
        Class.Number = NA,
        Mean.Class.Size = NA,
        Median.Class.Size = NA,
        Min.Class.Size = NA,
        Max.Class.Size = NA,
        convergence.problem = FALSE
      )[FALSE, , FALSE])
    }

    ds1[, rv] <- lapply(ds1[, rv, drop = FALSE], util_as_numeric)

    # prepare data frame for output
    res_df <- data.frame(
      Variables = NA,
      Object = NA,
      Model.Call = NA,
      ICC = NA,
      Class.Number = nrow(check_df),
      Mean.Class.Size = mean(check_df$Freq),
      Median.Class.Size = median(check_df$Freq),
      Min.Class.Size = min(check_df$Freq),
      Max.Class.Size = max(check_df$Freq),
      convergence.problem = FALSE
    )

    # fill output df
    res_df[1, 1] <- rv
    res_df[1, 2] <- group_var
    # create model formula from function arguments
    fmla <- as.formula(paste0(paste0(util_bQuote(rv), "~"),
                              paste0(c(co_vars, paste0("(1|", util_bQuote(group_var), ")")),
                                     collapse = "+")))
    # save formula for results
    res_df[1, 3] <- Reduce(paste, deparse(fmla))

    # remove misisng in used variables
    subdf <- na.omit(ds1[, c(rv, intersect(cv, colnames(ds1)),
                             group_var), drop = FALSE])
    # drop unused levels (https://stackoverflow.com/a/1197154)
    subdf[] <- lapply(subdf, function(x) if (is.factor(x)) factor(x) else x)

    env <- environment()

    convergence.problem <- FALSE

    suppressMessages(withCallingHandlers(
      # fit mixed effects model with formula
      fit_lmer <- try(lme4::lmer(fmla, data = subdf, REML = TRUE)),
      message = function(m) {
        if (inherits(m, "simpleMessage") && any(
            grepl("singular",
             trimws(conditionMessage(m))))) {
          env$convergence.problem <- TRUE
        } else {
          util_message(m) # nocov
          # This is to handle other possible problems with the model
          # Since it is for unexpected problems, I cannot write a test
        }
      },
      warning = function(w) {
        # nocov start
        # on travis, the try above prevents a stop on convergence problems
        # for some unclear reasons. therefore, on travis,
        # the following code cannot be reached. However, locally, this
        # is possible.
        haystack <- trimws(paste0(conditionMessage(w), collapse = "\n"))
        if (inherits(w, "simpleWarning") && any(vapply(
            c("unable to evaluate scaled gradient",
              "Problem with Hessian check",
              "Model is nearly unidentifiable: very large eigenvalue",
              "convergence code [0-9\\-].+ from nloptwrap",
              "Model failed to converge with"
              ), grepl, haystack, perl = TRUE, FUN.VALUE = logical(1)))) {
          env$convergence.problem <- TRUE
          invokeRestart("muffleWarning")
        } else {
          warning(w)
          # This is to handle other possible problems with the model
          # Since it is for unexpected problems, I cannot write a test
        }
        # nocov end
      }
    ))

    res_df[1, "convergence.problem"] <- convergence.problem

    if (inherits(fit_lmer, "try-error")) {
      res_df[1, "convergence.problem"] <- TRUE
    } else {
      # extract variance of random effects
      v_tab <- as.data.frame(lme4::VarCorr(fit_lmer))
      res_df$Class.Number[[1]] <- lme4::ngrps(fit_lmer)
      # calculate icc
      res_df[1, 4] <- round(v_tab$vcov[1] / (v_tab$vcov[1] + v_tab$vcov[2]), 3)
    }

    class_sizes <- util_table_of_vct(subdf[, group_var])
    res_df <- within(res_df, {
      Mean.Class.Size[[1]] <- mean(class_sizes$Freq)
      Median.Class.Size[[1]] <- median(class_sizes$Freq)
      Min.Class.Size[[1]] <- min(class_sizes$Freq)
      Max.Class.Size[[1]] <- max(class_sizes$Freq)
    })

    # output
    return(res_df)
  })

  icc_output <- as.data.frame(do.call(dplyr::bind_rows, icc_output),
                              stringsAsFactors = FALSE)
  rownames(icc_output) <- NULL

  max_icc <- suppressWarnings(max(icc_output$ICC, na.rm = TRUE))
  argmax_icc <- icc_output$Variables[which.max(icc_output$ICC)]

  # format
  rownames(icc_output) <- NULL
  SummaryTable <- icc_output
  SummaryData <- icc_output
  names(SummaryTable)[names(SummaryTable) == "ICC"] <- "ICC_acc_ud_loc"

  # fix rounding decimal places and columns present
  SummaryData <- SummaryData[, setdiff(colnames(SummaryData), c("Model.Call"))]
  SummaryData$ICC <- util_round_to_decimal_places(SummaryData$ICC)
  SummaryData$Mean.Class.Size <-
    util_round_to_decimal_places(SummaryData$Mean.Class.Size)
  SummaryData$Median.Class.Size <-
    util_round_to_decimal_places(SummaryData$Median.Class.Size, 1)


  return(list(SummaryTable = SummaryTable,
              SummaryData = SummaryData,
              ScalarValue_max_icc = max_icc,
              ScalarValue_argmax_icc = argmax_icc))
}

# dataquieR_resultset2 <- function(...) {
#   this <- list(...)
#   # dataquieR_resultset_verify2(this) # TODO: Implement me!!
#   class(this) <- dataquieR_resultset_class2
#   this
# }
dataquieR_resultset_class2 <- "dataquieR_resultset2"

#' Get namespace for attributes
#'
#' @param my_storr_object the `storr` object
#'
#' @return the namespace name
#' @noRd
util_get_storr_att_namespace <- function(my_storr_object) {
  r <- my_storr_object$default_namespace
  r <- paste0(r, ".attributes")
  r
}

#' Get namespace for computation status
#'
#' @param my_storr_object the `storr` object
#'
#' @return the namespace name
#' @noRd
util_get_storr_stat_namespace <- function(my_storr_object) {
  r <- my_storr_object$default_namespace
  r <- paste0(r, ".status")
  r
}

#' Get namespace specifically for summary attributes for speed-up
#'
#' @param my_storr_object the `storr` object
#'
#' @return the namespace name
#' @noRd
util_get_storr_summ_namespace <- function(my_storr_object) {
  r <- my_storr_object$default_namespace
  r <- paste0(r, ".summary")
  r
}

#' Get a single result from a `dataquieR 2` report
#'
#' @param x the report
#' @param el the index
#'
#' @return the `dataquieR` result object
#' @export
`[[.dataquieR_resultset2` <- function(x, el) {
  my_storr_object <- util_get_storr_object_from_report(x)

  if (is.null(my_storr_object)) {
    # default
    all_calls <- attr(x, "all_calls")

    r <- NextMethod()

    if (is.raw(r)) {
      r <- util_decompress(r)
    }

    r_names <- names(x)

    if (is.numeric(el) && suppressWarnings(as.integer(el) == as.numeric(el)) &&
        el > 0 && el <= length(r_names)) {
      slot <- r_names[[el]]
    } else if (is.character(el)) {
      slot <- el
    } else {
      r <- NULL
    }
    if (slot %in% names(all_calls))
      r <- util_fix_dataquieR_result(r = r,
                                     slot = slot,
                                     cl = all_calls[[slot]])
    if (is.raw(r)) {
      if (isTRUE(attr(x, "raw"))) { # attr raw is set by .access_dq_rs2, if only storr part of this is needed
        return(r)
      }
      r <- util_decompress(r)
    }

    return(r)
  }

  stopifnot(inherits(my_storr_object, "storr"))

  r_names <- my_storr_object$get("names", namespace =
                                   util_get_storr_att_namespace(my_storr_object))

  if (is.numeric(el) && suppressWarnings(as.integer(el) == as.numeric(el)) &&
      el > 0 && el <= length(r_names)) {
    slot <- r_names[[el]]
    r <- NULL
    if (my_storr_object$exists(slot)) {
      r <- my_storr_object$get(slot)
      if (is.raw(r)) {
        if (isTRUE(attr(x, "raw"))) { # attr raw is set by .access_dq_rs2, if only storr part of this is needed
          return(r)
        }
        r <- util_decompress(r)
      }
    }
    if (!inherits(r, "dataquieR_result")) {
      all_calls <- attr(x, "all_calls")

      r <- util_fix_dataquieR_result(r = r,
                                     slot = slot,
                                     cl = all_calls[[slot]])
      if (is.raw(r)) {
        if (isTRUE(attr(x, "raw"))) { # attr raw is set by .access_dq_rs2, if only storr part of this is needed
          return(r)
        }
        r <- util_decompress(r)
      }

    }
    return(r)
  } else if (el %in% r_names) {
    slot <- el
    r <- NULL
    if (my_storr_object$exists(slot)) {
      r <- my_storr_object$get(slot)
      if (is.raw(r)) {
        if (isTRUE(attr(x, "raw"))) { # attr raw is set by .access_dq_rs2, if only storr part of this is needed
          return(r)
        }
        r <- util_decompress(r)
      }
    }
    if (!inherits(r, "dataquieR_result")) {
      all_calls <- attr(x, "all_calls")

      r <- util_fix_dataquieR_result(r = r,
                                     slot = slot,
                                     cl = all_calls[[slot]])

      if (is.raw(r)) {
        if (isTRUE(attr(x, "raw"))) { # attr raw is set by .access_dq_rs2, if only storr part of this is needed
          return(r)
        }
        r <- util_decompress(r)
      }
    }
    return(r)
  } else {
    return(NULL)
  }
}

# # IDEA: attr, attr<-, attrbiutes, ..., cave wrt my_storr_object, which needs to be found by the primitive, always, postponed, since obviously impossible
# Solution, so far: keep the attributes in all backends (this will not be kept in sync automatically, but it should still work)
# does not really work, we need to handle attributes separately
# trace ("attr", edit =  function (...)
# {
#   args <- list(...)
#   x <- NULL
#   if (length(args) > 0) {
#     x <- args[[1]]
#   }
#   if (inherits(x, "dataquieR_resultset2")) {
#     return(42)
#   }
#   .prim <- .Primitive("attr")
#   .prim(...)
# }

# attr.dataquieR_resultset2 <- function(x, which, exact = FALSE) {
#
#     y <- unclass(x)
#     my_storr_object <- attr(y, "my_storr_object")
#
#     if (is.null(my_storr_object)) {
#       # default
#       return(NextMethod())
#     }
#
#     stopifnot(inherits(my_storr_object, "storr"))
#
#     my_storr_object$get(which, namespace = util_get_storr_att_namespace(my_storr_object))
# }


#' Set a single result from a `dataquieR 2` report
#'
#' @param x the report
#' @param el the index
#' @param value the single result
#'
#' @return the `dataquieR` result object
#' @export
`[[<-.dataquieR_resultset2` <- function(x, el, value) { # TODO: verify class of value

  if (is.null(value)) {
    value <- list()
    class(value) <- "dataquieR_NULL"
  }

  if (is.null(x[[el]])) {
    util_error("Extending reports is not supported.")
  }

  my_storr_object <- util_get_storr_object_from_report(x)

  if (is.null(my_storr_object)) {
    # default
    return(NextMethod())
  }

  stopifnot(inherits(my_storr_object, "storr"))

  r_names <- my_storr_object$get("names", namespace =
                                   util_get_storr_att_namespace(my_storr_object))

  if (is.numeric(el) && suppressWarnings(as.integer(el) == as.numeric(el)) &&
      el > 0 && el <= length(r_names)) {
    my_storr_object$set(r_names[[el]], value = value)
    return(x)
  } else if (el %in% r_names) {
    my_storr_object$set(el, value = value)
    return(x)
  } else {
    util_error("element not found, extending reports not yet supported, sorry")
  }
}

#' Write to a report
#'
#' Overwriting of elements only list-wise supported
#'
#' @param x a `dataquieR_resultset2
#' @param ... if this contains only one entry and this entry is not named
#'            or its name is `els`, then, the report will be accessed in
#'            list mode.
#' @param value new value to write
#'
#' @return nothing, stops
#' @export
`[<-.dataquieR_resultset2` <- function(x, ..., value) {
  if (nargs() > 3) {
    util_error("You cannot write subsets of a dataquieR report, yet.")
  }
  if (nargs() == 3 && !identical(rlang::missing_arg(),
                                 rlang::call_args(sys.call())[[2]])) {
    args <- list(...)
  } else {
    args <- list()
  }
  if (identical(names(args), "els") ||
      (length(args) == 1 && is.null(names(args))) ||
      (length(args) == 0)) {
    # list mode
    return(`.access_dq_rs2<-`(x, seq_along(names(x)), value))
  } else {
    util_error("You cannot write subsets of a dataquieR report, yet.")
  }
}

#' Access single results from a [dataquieR_resultset2] report
#' @aliases cash-.dataquieR_resultset2
#' @inherit [[.dataquieR_resultset2
#' @export
`$.dataquieR_resultset2` <- `[[.dataquieR_resultset2`

#' Write single results from a [dataquieR_resultset2] report
#' @aliases cash-set-.dataquieR_resultset2.Rd
#' @inherit [[<-.dataquieR_resultset2
#' @export
`$<-.dataquieR_resultset2` <- `[[<-.dataquieR_resultset2`

#' inefficient way to convert a report to a list. try [prep_set_backend()]
#'
#' @param x [dataquieR_resultset2]
#' @param ... no used
#'
#' @return [list]
#' @export
as.list.dataquieR_resultset2 <- function(x, ...) {
  my_storr_object <- util_get_storr_object_from_report(x)

  if (is.null(my_storr_object)) {
    # default
    return(NextMethod())
  } else {
    if (!getOption("dataquieR.convert_to_list_for_lapply",
                  dataquieR.convert_to_list_for_lapply_default) &&
        identical(rlang::env_parent(rlang::caller_env()),
                  asNamespace("base")) && # called from base, likely *apply
        rlang::call_name(rlang::caller_call()) %in% c("lapply",
                                                      "vapply",
                                                      "sapply")
        ) { # no conversion needed, interface already compatible with list()
      return(x)
    }
    util_warning("as.list is inefficient for dataquieR_resultset2 objects",
            immediate = TRUE)
    # print(rlang::trace_back())
    lapply(x, identity)
  }
}

#' Class [`r dataquieR_resultset_class2`].
#'
#' @seealso [dq_report2]
#' @aliases .dataquieR_resultset2
#' @importFrom methods new
dataquieR_resultset2 <- methods::setClass("dataquieR_resultset2")

#' Change the back-end of a report
#'
#' with this function, you can move a report from/to a `storr` storage.
#'
#' @param r [dataquieR_resultset2] the report
#' @param storr_factory `storr` the `storr` storage or `NULL`, to move
#'                                the report fully back into the RAM.
#' @param amend [logical] if there is already data in.`storr_factory`,
#'                        use it anyways -- unsupported, so far!
#'
#' @return [dataquieR_resultset2] but now with the desired back-end
#' @export
prep_set_backend <- function(r, storr_factory = NULL, amend = FALSE) {
  util_expect_scalar(amend, check_type = is.logical)
  stopifnot(inherits(r, "dataquieR_resultset2"))
  if (is.null(storr_factory)) {
    my_storr_object <- util_get_storr_object_from_report(r)
    stopifnot(inherits(my_storr_object, "storr"))
    atts_r_nm <- my_storr_object$list(namespace =
                                        util_get_storr_att_namespace(my_storr_object))
    atts_r <- setNames(my_storr_object$mget(atts_r_nm,
                                            namespace =
                                              util_get_storr_att_namespace(my_storr_object)),
                       atts_r_nm
    )
    r_nm <- atts_r[["names"]] # my_storr_object$list()
    r <- setNames(my_storr_object$mget(r_nm), r_nm);
    attr(r, "my_storr_object") <- NULL; # my_storr_object # never ever restore this
    attributes(r) <- atts_r
  } else {

    my_storr_object <- util_storr_object(storr_factory)

    stopifnot(inherits(my_storr_object, "storr"))

    if (!is.null(my_storr_object)) {

      if (!is.null(my_storr_object) && (
        length(my_storr_object$list()) > 0 ||
        length(my_storr_object$list(
          util_get_storr_att_namespace(my_storr_object))) > 0 ||
        length(my_storr_object$list(
          util_get_storr_summ_namespace(my_storr_object))) > 0
      )) {
        if (amend) {
          util_message(c("Your storr-object is not empty, but %s was set %s,",
                         "so I'll amend the storage object. This is unsupported,",
                         "yet, so expect strange behavior."),
                       dQuote("amend"), sQuote(TRUE))
        } else {
          util_error(c("Your storr-object is not empty, and %s was set %s,",
                       "so I won't amend the storage object, which would",
                       "still be unsupported, so could cause strange behavior.",
                       "We strongly recommend to use clear storr objects (or",
                       "at least the default namespace (%s in your case)",
                       "and its sister namespaces (the default namespace suffixed",
                       "with %s and %s, should be empty. In case of %s, just",
                       "delete the folder that backs the storr."),
                     dQuote("amend"),
                     sQuote(FALSE),
                     sQuote(my_storr_object$default_namespace),
                     sQuote(".attributes"),
                     sQuote(".summary"),
                     sQuote("driver_rds")
          )
        }
      }


      my_storr_object <- util_fix_storr_object(my_storr_object)
      atts_r <- attributes(r)
      atts_r[["my_storr_object"]] <- NULL # dont save this ever
      my_storr_object$mset(key = names(atts_r), value = atts_r, namespace =
                             util_get_storr_att_namespace(my_storr_object))

      my_storr_object$mset(key = names(r), value = r)

      my_storr_object$mset(
        key = names(r),
        value = lapply(r, attr, "r_summary"),
        namespace = util_get_storr_summ_namespace(my_storr_object))


      r[] <- lapply(r, function(x) NA)
      attr(r, "my_storr_object") <- my_storr_object
      class(r) <- "dataquieR_resultset2"
    }
  }
  r
}

#' Load a report from a back-end
#'
#' @param namespace the namespace to read the report's results from
#' @param storr_factory a function returning a `storr` object holding the report
#' @param db_dir [character] path to the directory for the back-end, if
#'                           a `storr_rds` or `storr_torr` is used.
#'
#' @return [dataquieR_resultset2] the report
#' @export
#' @seealso [prep_create_storr_factory()]
#' @examples
#' \dontrun{
#' r <- dataquieR::dq_report2("study_data", meta_data_v2 = "meta_data_v2",
#'                            dimensions = NULL)
#' storr_factory <- prep_create_storr_factory()
#' r_storr <- prep_set_backend(r, storr_factory)
#' r_restorr <- prep_set_backend(r_storr, NULL)
#' r_loaded <- prep_load_report_from_backend(storr_factory)
#' }
prep_load_report_from_backend <- function(
    namespace = "objects",
    db_dir,
    storr_factory = prep_create_storr_factory(namespace = namespace,
                                               db_dir = db_dir)) {

  my_storr_object <- util_storr_object(storr_factory)

  if (is.null(my_storr_object)) {
    util_error("You did not pass a valid storr factory in the argument %s",
               sQuote("storr_factory"))
  }

  atts_r_nm <- my_storr_object$list(namespace =
                                      util_get_storr_att_namespace(my_storr_object))
  atts_r <- setNames(my_storr_object$mget(atts_r_nm, namespace =
                                            util_get_storr_att_namespace(my_storr_object)),
                     atts_r_nm
  )
  r_nm <- atts_r[["names"]] # my_storr_object$list()
  r <- lapply(r_nm, function(x) NULL)
  attributes(r) <- atts_r
  attr(r, "my_storr_object") <- my_storr_object
  class(r) <- "dataquieR_resultset2"
  r
}

#' Get a subset of a `dataquieR` `dq_report2` report
#'
#' @param x the report
#'
#' @param row the variable names, must be unique
#' @param col the function-call-names, must be unique
#' @param res the result slot, must be unique
#' @param drop drop, if length is 1
#' @param els used, if in list-mode with named argument
#' @param as_raw retrieve the result maybe as compressed `raw` `util_compress()`
#'               serialized object
#'
#' @return a list with results, depending on `drop` and the number of results,
#'         the list may contain all requested results in sub-lists. The order
#'         of the results follows the order of the row/column/result-names given
#'
#' @export
`[.dataquieR_resultset2` <- function(x, row, col, res, drop = FALSE,
                                     els = row, as_raw = FALSE) {
  util_stop_if_not(inherits(x, "dataquieR_resultset2"))

  if (identical(rlang::call_args_names(sys.call()), c("", "")) ||
      identical(rlang::call_args_names(sys.call()), c("", "els"))) {
    # list mode
    return(.access_dq_rs2(x, els, as_raw = as_raw))
  }

  cn <- attr(x, "cn")
  rn <- attr(x, "rn")

  if (missing(col)) {
    col_matches <- rep(TRUE, length(attr(x, "names")))
  } else {
    if (!is.vector(col)) {
      util_error("column coordinate %s is not a vector/scalar", sQuote(util_deparse1(
        substitute(col))))
    }
    util_stop_if_not(!any(duplicated(col)))
    col_matches <- cn %in% col
  }

  if (missing(row)) {
    row_matches <- rep(TRUE, length(attr(x, "names")))
  } else {
    util_stop_if_not(!any(duplicated(row)))
    # row_matches <- rn %in% c(row, "[ALL]") # TODO: Ensure, that only one match is found (int_ should return [all] but nothing else, com_, eg, vv)
    row_matches <- rn %in% row
  }

  matches <- row_matches & col_matches

  r <- .access_dq_rs2(x, matches, as_raw = as_raw)
  rcn <- cn[matches]
  rrn <- rn[matches]
  if (!missing(col)) {
    first_order <- ordered(rcn, col)
  } else {
    first_order <- seq_along(rcn)
  }
  if (!missing(row)) {
    second_order <- ordered(rrn, row)
  } else {
    second_order <- seq_along(rrn)
  }
  r <- r[order(first_order, second_order)]

  if (!missing(res) && length(res)) {
    util_stop_if_not(!any(duplicated(res)))
    errors <- lapply(r, attr, "error")
    errors <- vapply(errors, length, FUN.VALUE = integer(1)) > 0
    r[!errors] <- lapply(r[!errors], `[`, res, drop = drop)
    # if (drop) r[!errors] <- lapply(r[!errors], `[[`, 1)
    # if (drop && length(r) == 1) { r <- r[[1]] }
    want_combine <- TRUE
  } else {
    want_combine <- FALSE
  }

  if (!as_raw && want_combine) {
    r <- r[!vapply(r, function(rs) {
      all(vapply(rs, is.null, FUN.VALUE = logical(1)))
    }, FUN.VALUE = logical(1))]
    if (length(r) > 0 &&
        !all(vapply(r, FUN.VALUE = logical(1),
                    function(x) {
                      all(vapply(x,
                                 function(y) {
                                   length(y) == 0
                                 }, FUN.VALUE = logical(1)))
                    }))) {
      if (!util_is_try_error(try(r <- util_combine_res(r), silent = TRUE))) {
        if (length(r) > 1) {
          for (i in seq_along(r)) {
            class(r[[i]]) <- unique(c("master_result",
                                      "dataquieR_result",
                                      class(r[[i]])))
          }
          # class(r) <- unique(c("dataquieR_result", class(r)))
        } else {
          r <- util_dataquieR_result(r[[1]])
        }
      }
    } else {
      class(r) <- union(c("dataquieR_NULL", "dataquieR_result"), class(r))
    }
  }

  if (length(r) == 1 && drop) {
    r <- r[[1]]
  }

  return(r)

}

#' Access elements from a `dataquieR_resultset2`
#'
#' does so, but similar to `[` for lists.
#'
#' @param x the `dataquieR_resultset2`
#' @param els the selector (character, number or logical)
#' @param as_raw retrieve the result maybe as compressed `raw` `util_compress()`
#'               serialized object
#'
#' @return the sub-list of `x`
#' @noRd
.access_dq_rs2 <- function(x, els, as_raw = FALSE) {
  my_storr_object <- util_get_storr_object_from_report(x)

  # if (is.null(my_storr_object)) {}
  # if (is.raw(r)) {
  #   r <- util_decompress(r)
  # }

  if (!as_raw) {
    .classit <- identity
  } else {
    if (is.null(my_storr_object)) {
      .classit <- unclass # THIS WONT WORK, IF STORR_FACTORY HAS BEEN SET
    } else {
      .classit <- function(x) {
        # [[ behaves differently if attr raw is set
        attr(x, "raw") <- TRUE
        x
      }
    }
  }
  if (is.character(els)) {
    lapply(setNames(nm = els), function(el) .classit(x)[[el]])
  } else if (is.numeric(els)) {
    lapply(setNames(nm = names(x)[els]), function(el) .classit(x)[[el]])
  } else if (is.logical(els)) {
    Recall(x, which(els), as_raw = as_raw)
  } else {
    util_error(
      c("Access to report can use numbers, logical vectors or names as index,",
        "but not %s"), util_pretty_vector_string(class(els)))
  }
}

#' Write elements from a `dataquieR_resultset2`
#'
#' does so, but similar to `[` for lists.
#'
#' @param x the `dataquieR_resultset2`
#' @param els the selector (character, number or logical)
#' @param value `dataquieR_result` to write
#'
#' @return the modified `x`
#' @noRd
`.access_dq_rs2<-` <- function(x, els, value) {
  if (is.logical(els)) {
    els <- which(els)
  }
  if (is.character(els)) {
    values <- rep(value, length.out = length(els))
    for (el in seq_along(els)) {
      x[[els[[el]]]] <- values[[el]]
    }
  } else if (is.numeric(els)) {
    values <- rep(value, length.out = length(els))
    for (el in seq_along(els)) {
      x[[els[[el]]]] <- values[[el]]
    }
  } else {
    util_error(
      c("Access to report can use numbers, logical vectors or names as index,",
        "but not %s"), util_pretty_vector_string(class(els)))
  }
  return(x)
}

#' Ensure `dataquieR` result of class `dataquieR_result`
#'
#' @param r `object` hopefully of class `dataquieR_result`
#' @param slot [character] name of the result in the report
#' @param cl [call] call that should have produced the result
#'
#' @returns `r`, if valid, otherwise an error result
#' @noRd
util_fix_dataquieR_result <- function(r, slot, cl) {
  if (!inherits(r, "dataquieR_result")) {
    r <- util_eval_to_dataquieR_result(init = TRUE,
                                       quote({util_error(paste("No result available for unkown reasons",
                                                               "(out of memory? try to reduce the number",
                                                               "of parallel running jobs using the",
                                                               "`cores` argument)"))}),
                                       nm = slot,
                                       function_name = rlang::call_name(cl),
                                       my_call = cl)
  }
  return(r)
}

test_that("dataquieR_resultset2 class", {
  skip_on_cran() # slow, parallel, ...
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("storr")

  # db_dir <- "/tmp/asdfasdf"
  # unlink(db_dir, recursive = T, force = T)
  db_dir <-  withr::local_tempdir()
  db_dir2 <-  withr::local_tempdir()

  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")

  study_data <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), 100)
  meta_data <- prep_get_data_frame("item_level")

  mlt <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx| missing_table")

  prep_purge_data_frame_cache()

  prep_add_data_frames(`missing_table` = mlt)

  invisible(testthat::capture_output_lines(gc(full = TRUE, verbose = FALSE)))

  sd0 <- study_data[, 1:5]
  sd0$v00012 <- study_data$v00012
  md0 <- subset(meta_data, VAR_NAMES %in% colnames(sd0))
  md0$PART_VAR <- NULL
  md1 <- md0
  md1$LABEL <- c("CENTER_0",
                 "",
                 "CENTER_0 DUPLICATE", # will become a duplicated label
                 "CENTER_0", # direct duplication of the first label
                 "Have you been physically vigorously active in the past 12 hours ('physically vigorously active' means at least 30 minutes of jogging or fast cycling, digging up your garden, carrying heavy objects weighing more than 10 kg for a long time, or similar physical activities)?", # very long label
                 "Hybpvaitp1hpvamal3mojofcduygchowmt1kfaltospa") # matches the very long label after abbreviation
  md1$VAR_NAMES[2] <- "yOvCzPY60JRjmrYb16Tsd6qMymal4B5Skw9rZ5PHSCtaBqOVglAKcguPkQhakampFJcC8xqLbZJs7kZUdKH804pbOmM5ORPVabrkEkVkiWbakWiixZ99NRYF6BP8SRxzNYY2tED7DjmhMUwk0t674RjH828jq9zoTJgDxYP6nEdHBxhmXJh0ClCPjGsi1q" # very long variable name that should get caught and not be used as label as it is
  colnames(sd0)[2] <- md1$VAR_NAMES[2]

  sf <- prep_create_storr_factory(namespace = "Test",
                            db_dir = db_dir)

  suppressWarningsMatching(
    report <- dq_report2(sd0, md1,
                         storr_factory =
                           sf,
                         label_col = VAR_NAMES,
                         cores = NULL,
                         dimensions =
                           c("Integrity",
                             "Consistency")),
    "(Some variables have labels with more than 60 characters in.+|Unique labels are required|Lost 16.7% of the study data because of missing/not assignable metadata|Need.+VAR_NAMES.+discard.+|.+dummy names|Some variables have no label in .+LABEL)"
  )
  report0 <- report

  testidx <- head(which(names(report) == "des_summary_categorical.v00000"), 1)
  expect_length(testidx, 1)

  x0 <- force(report0[[testidx]]$SummaryTable)
  expect_s3_class(report0[["con_limit_deviations.v00004"]], "master_result")
  expect_null(report0[["con_limit_deviations.v0000xx4"]])


  suppressWarningsMatching({
    report[[1]] <- report[[1]]
    report[["con_limit_deviations.v00004"]] <-
      report[["con_limit_deviations.v00004"]]

    expect_error(
      report[["con_limit_deviations.v0000xx4"]] <- NULL,
      regexp = "Extending reports is not supported")
  }, regexps = ".*inefficient.*")

  x1 <- force(report[[testidx]]$SummaryTable)
  # save(x0, x1, report, file = "/tmp/xx.RData")

  expect_s3_class(report[["con_limit_deviations.v00004"]], "master_result")
  expect_null(report[["con_limit_deviations.v0000xx4"]])

  report0 <- prep_set_backend(report, NULL)
  report2 <- prep_set_backend(report0,
                              prep_create_storr_factory(namespace = "Test",
                                                        db_dir = db_dir2))
  report3 <- prep_load_report_from_backend(
      namespace = "Test",
      db_dir = db_dir2
    )

  expect_equal(
    report0$`int_all_datastructure_dataframe.[ALL]`,
    report$`int_all_datastructure_dataframe.[ALL]`,
    ignore_attr = "system.time"
  )

  expect_equal(
    report2$`int_all_datastructure_dataframe.[ALL]`,
    report$`int_all_datastructure_dataframe.[ALL]`,
    ignore_attr = "system.time"
  )

  expect_equal(
    report3$`int_all_datastructure_dataframe.[ALL]`,
    report$`int_all_datastructure_dataframe.[ALL]`,
    ignore_attr = "system.time"
  )

  # testthat::skip_if(identical(Sys.getenv("R_COVR"), "true"),
  #                   message = "Does not work, if instrumented")


  expect_s3_class(x0, "TableSlot")
  expect_s3_class(x1, "TableSlot")

})

test_that("storr helper functions", {
  skip_if_not_installed("storr")
  sf1 <- prep_create_storr_factory()
  so1 <- sf1()
  sf2 <- prep_create_storr_factory(namespace = "Test")
  so2 <- sf2()
  expect_equal(util_get_storr_att_namespace(so1), "objects.attributes")
  expect_equal(util_get_storr_att_namespace(so2), "Test.attributes")
  expect_equal(util_get_storr_summ_namespace(so1), "objects.summary")
  expect_equal(util_get_storr_summ_namespace(so2), "Test.summary")
})

#' Convert `VALUE_LABELS` to separate tables
#'
#' @inheritParams .template_function_indicator
#' @param max_value_label_len [integer] maximum length for value labels
#'
#' @return [data.frame] metadata with `VALUE_LABEL_TABLE` instead of
#'   `VALUE_LABELS` (or none of these, if absent)
#'
#' @noRd
#' @examples
#' \dontrun{
#' prep_purge_data_frame_cache()
#' prep_load_workbook_like_file("meta_data_v2")
#' util_normalize_value_labels()
#' prep_add_data_frames(test_labs =
#'   tibble::tribble(~ CODE_VALUE, ~ CODE_LABEL, 17L, "Test", 19L, "Test",
#'     17L, "TestX"))
#' il <- prep_get_data_frame("item_level")
#' if (!VALUE_LABEL_TABLE %in% colnames(il)) {
#'   il$VALUE_LABEL_TABLE <- NA_character_
#' }
#' il$VALUE_LABEL_TABLE[[1]] <- "test_labs"
#' il$VALUE_LABELS[[1]] <- "17 = TestY"
#' prep_add_data_frames(item_level = il)
#' util_normalize_value_labels()
#' }
#'
util_normalize_value_labels <- function(meta_data = "item_level",
                                        max_value_label_len =
                                          getOption(
                                            "dataquieR.MAX_VALUE_LABEL_LEN",
                                      dataquieR.MAX_VALUE_LABEL_LEN_default)) {

  util_expect_data_frame(meta_data)

  if (!any(c(VALUE_LABELS, VALUE_LABEL_TABLE) %in% colnames(meta_data))) {
    return(meta_data)
  }

  util_expect_scalar(max_value_label_len,
                     check_type = is.numeric,
                     error_message = sprintf("%s must be a numeric value",
                                             sQuote("dataquieR.MAX_VALUE_LABEL_LEN")))

  if (VALUE_LABELS %in% colnames(meta_data)) {
    if (VALUE_LABEL_TABLE %in% colnames(meta_data) &&
        any(!util_empty(meta_data[!util_empty(meta_data[[VALUE_LABELS]]),
                                  VALUE_LABEL_TABLE]))) {
      util_warning(
        "Cannot mix %s and %s, trying to fix, but expect inconsistencies.",
        sQuote(VALUE_LABELS),
        sQuote(VALUE_LABEL_TABLE), applicability_problem = TRUE)

      # load all data frames
      invisible(lapply(
        meta_data[!util_empty(meta_data[[VALUE_LABEL_TABLE]]),
                  VALUE_LABEL_TABLE, drop = TRUE], prep_get_data_frame))
    }

    md <- meta_data
    md <- md[!duplicated(md[[VALUE_LABELS]]), , FALSE]

    def_cats <- setNames(lapply(util_parse_assignments(md[[VALUE_LABELS]],
                                       split_on_any_split_char = FALSE,
                                       split_char = c(SPLIT_CHAR, '<'),
                                       multi_variate_text = TRUE),
                                function(x) {
      if (length(x) > 0) {
        r <- data.frame(CODE_VALUE = names(x),
                   CODE_LABEL = unlist(unname(x), recursive = FALSE),
                   row.names = NULL,
                   check.rows = FALSE,
                   check.names = FALSE,
                   fix.empty.names = FALSE,
                   stringsAsFactors = FALSE)
      } else {
        r <- data.frame(CODE_VALUE = character(0),
                   CODE_LABEL = character(0),
                   row.names = NULL,
                   check.rows = FALSE,
                   check.names = FALSE,
                   fix.empty.names = FALSE,
                   stringsAsFactors = FALSE)
      }
      if (attr(x, "split_char") == '<') {
        r[[CODE_ORDER]] <- rank(r[[CODE_VALUE]], ties.method = "random")
      }
      return(r)
    }), nm = md[[VALUE_LABELS]])

    names(def_cats) <- .util_generate_value_label_table_name(names(def_cats))

    if (any(names(def_cats) %in% prep_list_dataframes())) {
      already <-
        names(def_cats)[names(def_cats) %in% prep_list_dataframes()]
      for (vlt in already) {
        def_cats[[vlt]] <-
          util_combine_value_label_tables(prep_get_data_frame(vlt),
                                          def_cats[[vlt]])
      }
    }

    prep_add_data_frames(data_frame_list = def_cats)
  }

  if (!(VALUE_LABELS %in% colnames(meta_data))) {
    meta_data[[VALUE_LABELS]] <- NA_character_
  }

  if (!(VALUE_LABEL_TABLE %in% colnames(meta_data))) {
    meta_data[[VALUE_LABEL_TABLE]] <- NA_character_
  }

  # fix records with VALUE_LABELS and VALUE_LABEL_TABLE
  both <-
    !util_empty(meta_data[[VALUE_LABELS]]) &
      !util_empty(meta_data[[VALUE_LABEL_TABLE]])

  for (i in which(both)) {
    orig_nm <- meta_data[[VALUE_LABEL_TABLE]][[i]]
    cmbd_nm <- paste0(orig_nm, "_",
                       VALUE_LABELS)

    vlnm <-
      .util_generate_value_label_table_name(meta_data[[VALUE_LABELS]][[i]])

    vldf <- def_cats[[vlnm]]

    if ((cmbd_nm %in% prep_list_dataframes())) {

      exdf <- prep_get_data_frame(cmbd_nm)

      testdf <- merge(vldf, exdf, by = intersect(c(CODE_VALUE, CODE_ORDER), # FIXME: CODE_ORDER: honor its values, not only its existence
                                                 intersect(colnames(vldf),
                                                           colnames(exdf))),
                      all = TRUE)

      if (nrow(testdf) != nrow(exdf) || nrow(testdf) != nrow(vldf)) {
        # util_warning(
        #   "Will overwrite the following data frame: %s",
        #   dQuote(cmbd_nm), applicability_problem = TRUE)
      }
    }

    meta_data[[VALUE_LABELS]][[i]] <- NA
    meta_data[[VALUE_LABEL_TABLE]][[i]] <- cmbd_nm

    prep_add_data_frames(data_frame_list = setNames(
      list(x = util_combine_value_label_tables(
        prep_get_data_frame(orig_nm),
        vldf)),
      nm = cmbd_nm
    ))

  }

  meta_data[!util_empty(meta_data[[VALUE_LABELS]]),
            VALUE_LABEL_TABLE] <-
    .util_generate_value_label_table_name(
      meta_data[!util_empty(meta_data[[VALUE_LABELS]]), VALUE_LABELS])

  meta_data[[VALUE_LABELS]] <- NULL

  is_too_long <- function(x) {
    nchar(x) > max_value_label_len
  }
  to_fix <- unique(meta_data[[VALUE_LABEL_TABLE]]) # TODO: Standardized voc;
  to_fix <- to_fix[!is.na(to_fix)]
  lapply(to_fix, function(vlt) {
    cur_df <- try(prep_get_data_frame(vlt), silent = TRUE)
    if (is.data.frame(cur_df)) {
      if (!(CODE_LABEL %in% colnames(cur_df)) &&
          CODE_VALUE %in% colnames(cur_df) &&
          any(is_too_long(cur_df[[CODE_VALUE]]))) {
        cur_df[[CODE_LABEL]] <- cur_df[[CODE_VALUE]]
      }
      if (CODE_LABEL %in% colnames(cur_df) &&
          any(is_too_long(cur_df[[CODE_LABEL]]))) {
        cur_df[is_too_long(cur_df[[CODE_LABEL]]), CODE_LABEL] <-
          util_abbreviate_unique(
            cur_df[is_too_long(cur_df[[CODE_LABEL]]), CODE_LABEL],
            max_value_label_len = max_value_label_len)
      }
      prep_add_data_frames(data_frame_list =
                             setNames(list(cur_df), nm = vlt))
      # } else if (util_is_try_error(cur_df)) { this will be shown anyways
      #   util_warning("Could not find value label table %s",
      #                applicability_problem = TRUE)
    }
  })

  meta_data

}

#' Combine two value lists
#'
#' @param vlt1 [value_label_table]
#' @param vlt2 [value_label_table]
#'
#' @return [value_label_table]
#' @noRd
#' @examples
#' \dontrun{
#' util_combine_value_label_tables(
#'   tibble::tribble(~ CODE_VALUE, ~ CODE_LABEL, 17L, "Test", 19L, "Test", 17L, "TestX"),
#'   tibble::tribble(~ CODE_VALUE, ~ CODE_LABEL, 17L, "Test", 19L, "Test", 17L, "TestX"))
#' }
util_combine_value_label_tables <- function(vlt1, vlt2) {
  vlt <- unique(util_rbind(vlt1, vlt2))
  if (any(duplicated(vlt[[CODE_VALUE]]))) {
    # util_warning(c("Duplicated %ss detected: %s. Fixing",
    #                "by merging the labels"),
    #              sQuote(CODE_VALUE),
    #              util_pretty_vector_string(
    #                unique(vlt[[CODE_VALUE]][duplicated(vlt[[CODE_VALUE]])])
    #              ),
    #              applicability_problem = TRUE)
    dups <- split(vlt, vlt[[CODE_VALUE]])
    dups <- lapply(dups, function(dup) {
      dup[[CODE_LABEL]] <-
        common_label <- paste(dup[[CODE_LABEL]],
                              collapse = sprintf(" %s ", SPLIT_CHAR))
      dup
    })
    vlt <- unique(util_rbind(data_frames_list = dups))
  }
  rownames(vlt) <- NULL
  vlt
}

.util_generate_value_label_table_name <- function(vl) {

  res <- util_parse_assignments(
    vl,
    split_on_any_split_char = FALSE,
    split_char = c(SPLIT_CHAR, '<'),
    multi_variate_text = TRUE)

  first_labs <- vapply(res, function(res_i) {
    suff <- ifelse(length(res_i) > 2, "...", "")
    res_i <- paste0(substr(head(res_i, 2), 1, 8), collapse = " ")
    paste0(res_i, suff)
  }, FUN.VALUE = character(1))

  hashes <- vapply(res, function(res_i) {
    hash <- substr(rlang::hash(res_i), 1, 5) # TODO: don't ignore hash collisions
  }, FUN.VALUE = character(1))

  res <- paste0("LABS_", first_labs, "::", hashes)
  return(res)
}

# TODO: Also work with only one sheet featuring columns CODE_LABEL, CODE_VALUE, CODE_ORDER and VALUE_LABEL_TABLE

#' Utility function to create a plot similar to the margins plots for nominal variables
#'
#' This function is still under development. It uses the `nnet` package to
#' compute multinomial logistic regression models.
#'
#' @param resp_vars  [variable] the name of the nominal measurement variable
#' @param group_vars [variable] the name of the observer, device or
#'                              reader variable
#' @param co_vars [variable list] a vector of covariables, e.g. age and sex for
#'                              adjustment
#' @param min_obs_in_subgroup [integer] from=0. This optional argument specifies
#'                       the minimum number of observations that is required to
#'                       include a subgroup (level) of the `group_var` in the
#'                       analysis.
#' @param min_obs_in_cat [integer] This optional argument specifies the minimum
#'                       number of observations that is required to include
#'                       a category (level) of the outcome (`resp_vars`) in
#'                       the analysis.
#' @param ds1 [data.frame] the data frame that contains the measurements, after
#'                       replacing missing value codes by `NA`, excluding
#'                       inadmissible values and transforming categorical
#'                       variables to factors.
#' @param label_col [variable attribute] the name of the column in the metadata
#'                       with labels of variables
#' @param adjusted_hint [character] hint, if adjusted for `co_vars`
#' @param title [character] title for the plot
#' @param sort_group_var_levels [logical] Should the levels of the grouping
#'                        variable be sorted descending by the number of
#'                        observations (in the figure)?
#'
#' @return A table and a matching plot.
#'
#' @importFrom ggplot2 ggplot geom_point geom_pointrange theme_minimal theme xlab ylab ggtitle scale_colour_manual aes facet_grid
#' @importFrom stats p.adjust
#'
#' @noRd
util_margins_nom <- function(resp_vars = NULL, group_vars = NULL, co_vars = NULL,
                             min_obs_in_subgroup = 5, min_obs_in_cat = 5,
                             ds1, label_col,
                             adjusted_hint = "",
                             title = "",
                             sort_group_var_levels =
                               getOption("dataquieR.acc_margins_sort",
                                         dataquieR.acc_margins_sort_default)) {
  # preps and checks -----------------------------------------------------------
  util_ensure_suggested("nnet")
  # ensure that the response variable is not constant
  var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])
  if (var_prop$NDistinct < 2) {
    util_error("The response variable is constant (after data preparation).",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }
  # ensure that there are enough observations for the model
  count_nom <- util_table_of_vct(ds1[[resp_vars]])
  if (any(count_nom[, 2] < min_obs_in_cat)) {
    util_error("Not enough observations per category (after data preparation).")
  }
  count_nom <- util_table_of_vct(ds1[[group_vars]])
  if (any(count_nom[, 2] < min_obs_in_subgroup)) {
    util_error("Not enough observations per group (after data preparation).")
  }
  # ensure that the model supports the specified covariates
  if (length(co_vars) > 0) {
    covar_prop <- util_dist_selection(ds1[, co_vars, drop = FALSE])
    covar_fact <- setNames(nm = co_vars,
                           vapply(ds1[, co_vars, drop = FALSE], FUN.VALUE = logical(1),
                           FUN = is.factor))
    covar_excl <- union(which(covar_prop$NCategory <= 2),
                        which(!covar_fact))
    if (length(covar_excl) > 0) { # TODO: work in progress!
      util_error(paste("Adjusting for covariates is currently",
                       "only supported for factor variables with",
                       "2 or more levels"))
    }
  }

  # modelling ------------------------------------------------------------------
  # if no co_vars are defined for adjustment only the intercept is modelled
  if (length(co_vars) == 0) {
    co_vars <- "1"
    co_vars_bQ <- co_vars
  } else {
    co_vars_bQ <- util_bQuote(co_vars)
  }

  # build model formula
  fmla <- as.formula(paste0(
    paste0(util_bQuote(resp_vars), "~"),
    paste0(
      paste0(co_vars_bQ, collapse = " + "),
      " + ",
      util_bQuote(group_vars)
    )
  ))

  # multinomial logistic regression
  invisible(capture.output({
    model <- suppressMessages(nnet::multinom(fmla, data = ds1))
  }))
  nom_res_fmla <- as.formula(paste0(
    "~ ",
    util_bQuote(group_vars),
    " | ",
    util_bQuote(resp_vars)
  ))
  nom_int_fmla <- as.formula(paste0(
    "~ 1 | ",
    util_bQuote(resp_vars)
  ))

  res_df <- data.frame(emmeans::emmeans(model, nom_res_fmla),
                       check.names = FALSE)

  summary_ds <- as.data.frame(
    dplyr::summarize(
      dplyr::group_by_at(ds1[, c(resp_vars, group_vars), drop = FALSE],
                         unname(c(group_vars, resp_vars)), .drop = FALSE),
      sample_size = dplyr::n()))

  res_df <- merge(res_df, summary_ds, by = c(group_vars, resp_vars),
                  all.x = TRUE)
  res_df <- dplyr::rename(res_df, c("margins" = "prob", "LCL" = "lower.CL",
                                    "UCL" = "upper.CL"))

  # adjusted overall mean
  omv <- data.frame(emmeans::emmeans(model, nom_int_fmla))
  omv <- dplyr::rename(omv, c("margins" = "prob", "LCL" = "lower.CL",
                              "UCL" = "upper.CL"))

  # detect deviations ----------------------------------------------------------
  # proposal: use Wald tests (here z-tests, two-sided) to detect
  # outlying group var levels
  nom_test <- merge(res_df, omv[, c(resp_vars, "margins")],
                    by = resp_vars, all = TRUE, suffixes = c("", ".overall"))
  test_df <- nom_test[, c(resp_vars, group_vars)]
  test_df$z <- (nom_test$margins - nom_test$margins.overall)/nom_test$SE
  test_df$p <- (1 - pnorm(abs(test_df$z), 0, 1)) * 2
  # adjust p values for multiple testing
  test_df$GRADING <- as.numeric(p.adjust(test_df$p, method = "fdr") < 0.05)
  res_df <- merge(res_df, test_df[, c(resp_vars, group_vars, "GRADING")],
                  by = c(resp_vars, group_vars), all = TRUE)

  # figure ---------------------------------------------------------------------
  omv[, group_vars] <- "overall"
  nom_plot_data <- merge(res_df, omv,
                         by = c(resp_vars, group_vars, "margins", "SE", "df",
                                "LCL", "UCL"),
                         all = TRUE)
  nom_plot_data$GRADING[which(is.na(nom_plot_data$GRADING))] <- "-1"

  warn_code <- c("1" = "#B2182B", "0" = "#2166AC", "-1" = "grey20")

  if (sort_group_var_levels) {
    # order levels of the grouping variable by number of observations
    tbl_gr <- util_table_of_vct(ds1[[group_vars]]) %>%
      dplyr::arrange(dplyr::desc(.data$Var1)) %>%
      dplyr::arrange(.data$Freq)
    nom_plot_data[[group_vars]] <- factor(nom_plot_data[[group_vars]],
                                          levels = c(as.character(tbl_gr$Var1), "overall"))
    res_df <- res_df %>%
      dplyr::arrange(.data[[resp_vars]],
                     match(.data[[group_vars]], rev(as.character(tbl_gr$Var1))))
  } else { # original order in plot, but from top of the y-axis to the bottom
    nom_plot_data[[group_vars]] <- factor(nom_plot_data[[group_vars]],
                                          levels = c(rev(levels(ds1[[group_vars]])), "overall"))
  }

  nom_plot_data[["facet"]] <- nom_plot_data[[resp_vars]] # make facet findable to prevent util_compress_ggplots_in_res from deleting this column

  res_plot <- util_create_lean_ggplot(ggplot(nom_plot_data,
                                             aes(x = .data[["margins"]],
                                                 y = .data[[group_vars]],
                                                 col = .data[["GRADING"]])) +
                                        util_geom_pointrange_robust(aes(xmin = .data[["LCL"]],
                                                            xmax = .data[["UCL"]])) +
                                        geom_point() +
                                        scale_color_manual(values = warn_code,
                                                           guide = "none") +
                                        facet_grid(.data[["facet"]] ~ .) +
                                        theme_minimal() +
                                        xlab("probability") +
                                        ylab("") +
                                        theme(strip.text = element_text(size = 14)) +
                                        ggtitle(label = title,
                                                subtitle = adjusted_hint),
                                      nom_plot_data = nom_plot_data,
                                      group_vars = group_vars,
                                      warn_code = warn_code,
                                      title = title,
                                      adjusted_hint = adjusted_hint)

  return(list("plot_data" = res_df, "plot" = res_plot))
}

#' @family missing_functions
#' @concept missing
#' @noRd
util_seg_table <- function(ds2, study_data, meta_data, expected_observations =
                         c("HIERARCHY",
                           "ALL",
                           "SEGMENT")) {
  make_df <- function(x, ...) {
    r <- data.frame(CODES = as.character(rownames(x)), Freq = x,
                    row.names = NULL)
    r
  }
  lapply(lapply(lapply(setNames(nm = colnames(ds2)), function(x) {
    y <- ds2[util_observation_expected(x, study_data, meta_data, VAR_NAMES,
                                       expected_observations =
                                         expected_observations), x,
             drop = TRUE]
    table(y)
  }), as.matrix), make_df)
}

util_readr_cols_from_metadata <- function(meta_data) {
  util_expect_data_frame(meta_data, c(VAR_NAMES, DATA_TYPE))
  r <- setNames(readr_coltypes[trimws(tolower(meta_data$DATA_TYPE))],
           nm = meta_data$VAR_NAMES)
  unkown_dt <- vapply(r, is.null, FUN.VALUE = logical(1))
  if (any(unkown_dt)) {
    util_warning(
      c("For the following variables, the %s is not one of %s: %s\n",
        "--> falling back to %s"),
      sQuote(DATA_TYPE),
      util_pretty_vector_string(DATA_TYPES),
      prep_deparse_assignments(names(which(unkown_dt)),
                               meta_data$DATA_TYPE[unkown_dt],
                               mode = "string_codes"),
      dQuote(DATA_TYPES$STRING),
      applicability_problem = TRUE
    )
    r[unkown_dt] <- lapply(r[unkown_dt], function(x) readr::col_character())
  }
  r
}

readr_coltypes <-
  list(DATA_TYPES$INTEGER, readr::col_double(),
       DATA_TYPES$STRING, readr::col_character(),
       DATA_TYPES$FLOAT, readr::col_double(),
       DATA_TYPES$DATETIME, readr::col_datetime(),
       DATA_TYPES$TIME, readr::col_time())

readr_coltypes <-
  setNames(readr_coltypes[2*seq_len(length(readr_coltypes)/2)],
           nm = readr_coltypes[2*seq_len(length(readr_coltypes)/2) - 1])

util_adjust_data_type2 <- function(study_data,
                                   meta_data,
                                   relevant_vars_for_warnings = character(0),
                                   language_code =
                                     getOption("dataquieR.locale",
                                               dataquieR.locale_default)) {
  util_expect_data_frame(study_data, keep_types = TRUE)
  if (length(relevant_vars_for_warnings) == 0) {
    relevant_vars_for_warnings <- colnames(study_data)
  }
  if (isTRUE(attr(study_data, "Data_type_matches"))) {
    return(study_data)
  }
  if (.called_in_pipeline && rlang::is_integerish(dynGet("cores")) &&
      as.integer(dynGet("cores")) > 1 &&
      suppressWarnings(util_ensure_suggested("parallel", err = FALSE))) {
    mycl <- parallel::makePSOCKcluster(as.integer(dynGet("cores")))
    parallel::clusterCall(mycl, library, "dataquieR", character.only = TRUE)
    parallel::clusterCall(mycl, loadNamespace, "hms")
    on.exit(parallel::stopCluster(mycl))
  } else {
    mycl <- NULL
  }
  for (cn in colnames(study_data)) {
    if (!is.null(study_data[[cn]])) {
      attr(study_data[[cn]], "..cn") <-
        cn
    }
  }
  if (isTRUE(as.logical(getOption("dataquieR.old_factor_handling",
                dataquieR.old_factor_handling_default)))) {
    # This is for backwards compatibility, but it may be more user friendly to
    # omit this in both adjust_data_type functions
    study_data[, vapply(study_data, is.factor, FUN.VALUE = logical(1))] <-
      lapply(study_data[,
                        vapply(study_data, is.factor, FUN.VALUE = logical(1)),
                        drop = FALSE], as.integer)
  } else {
    study_data[, vapply(study_data, is.factor, FUN.VALUE = logical(1))] <-
      util_par_lapply_lb(cl = mycl, study_data[,
                        vapply(study_data, is.factor, FUN.VALUE = logical(1)),
                        drop = FALSE], function(x) util_as_character(x))
  }
  study_data[] <- util_par_lapply_lb(cl = mycl,
                                     study_data, util_as_character)
  to_warn <- character(0)
  my_env <- environment()
  old_names <- names(study_data)
  no_name <- !nzchar(names(study_data))
  if (any(no_name)) {
    # find unused prefix
    prefix <- "#"
    while (any(startsWith(names(study_data), prefix))) {
      prefix <- paste0("#", prefix)
    }
    names(study_data)[no_name] <-
      paste0(prefix, seq_len(sum(no_name)))
  }
  e <- l10n_info()$codeset
  if (is.null(e))
    e <- "UTF-8" # no reliable way of detecting the current encoding
  locale <- readr::locale(
    date_names = language_code,
#    date_format = , # keep the default
#    time_format = , # keep the default
    decimal_mark = Sys.localeconv()[["decimal_point"]],
    grouping_mark = Sys.localeconv()[["thousands_sep"]],
    tz = Sys.timezone(),
    encoding = e,
    asciify = FALSE)
  locale$tz <- Sys.timezone()
  col_types <- util_readr_cols_from_metadata(meta_data) # also checks for VAR_NAMES and DATA_TYPE
  num <- intersect(colnames(study_data),
                     meta_data[meta_data[[DATA_TYPE]] %in% c(
                       DATA_TYPES$FLOAT,
                       DATA_TYPES$INTEGER),
                               VAR_NAMES])
  intg <- intersect(colnames(study_data),
                            meta_data[meta_data[[DATA_TYPE]] ==
                                        DATA_TYPES$INTEGER,
                              VAR_NAMES])

  datim <- intersect(colnames(study_data),
                     meta_data[meta_data[[DATA_TYPE]] == DATA_TYPES$DATETIME,
                               VAR_NAMES])
  if (length(intg) > 0) {
    study_data[, intg] <- util_par_lapply_lb(cl = mycl,
      study_data[, intg, drop = FALSE],
      function(cl) {
        r <- cl
        r <- trimws(tolower(r))
        r[r %in% c("t", "true")] <- "1"
        r[r %in% c("f", "false")] <- "1"
        r
      }
    )
  }
  if (length(datim) > 0) {
    study_data[, datim] <- util_par_lapply_lb(cl = mycl,
      study_data[, datim, drop = FALSE],
      function(cl) {
        r <- cl
        r <- # prepend 000 to one-digit years
          gsub("^(\\d)\\-(\\d\\d)\\-(\\d\\d)(\\W.*|)$", "000\\1-\\2-\\3\\4",
               r)
        r <- # prepend 00 to two-digit years
          gsub("^(\\d\\d)\\-(\\d\\d)\\-(\\d\\d)(\\W.*|)$", "00\\1-\\2-\\3\\4",
             r)
        r <- # prepend 0 to three-digit years
          gsub("^(\\d\\d\\d)\\-(\\d\\d)\\-(\\d\\d)(\\W.*|)$", "0\\1-\\2-\\3\\4",
               r)
        r
      }
    )
  }

  if (isTRUE(as.logical(getOption("dataquieR.type_adjust_parallel",
                       dataquieR.type_adjust_parallel_default)))) {
    processed_list <- util_par_lapply_lb(
      cl = mycl,
      X  = seq_along(study_data),
      fun = function(i, study_data, col_types, locale, relevant_vars_for_warnings) {
        e <- environment()
        e$local_warns <- character()

        col_name <- names(study_data)[i]
        col_vec  <- study_data[[i]]

        df_one <- data.frame(col_vec, stringsAsFactors = FALSE)
        names(df_one) <- col_name

        if (!col_name %in% names(col_types)) {
          col_types[[col_name]] <- readr::col_guess()
        }
        converted_df <- withCallingHandlers(
          readr::type_convert(
            df_one,
            col_types     = col_types[col_name],
            trim_ws       = TRUE,
            guess_integer = FALSE,
            locale        = locale,
            na            = c("", "NA", "Inf", "+Inf", "-Inf", "NaN")
          ),
          warning = function(w) {
            # cat(conditionMessage(w), file = "/tmp/asdf", append = TRUE)
            # cat("\n", file = "/tmp/asdf", append = TRUE)
            cm <- util_extract_named_groups(
              "^\\[(?<row>\\d+), +(?<col>\\d+)\\]:.*",
              conditionMessage(w)
            )
            if (nrow(cm) > 0) {
              e$local_warns <- unique(c(e$local_warns, col_name))
              invokeRestart("muffleWarning")
            }
          }
        )

        list(
          data  = converted_df[[col_name]],
          warns = e$local_warns
        )
      },
      study_data                   = study_data,
      col_types                    = col_types,
      locale                       = locale,
      relevant_vars_for_warnings   = relevant_vars_for_warnings
    )

    data_list   <- lapply(processed_list, `[[`, "data")
    study_data0 <- as.data.frame(data_list, stringsAsFactors = FALSE)
    names(study_data0) <- names(study_data)

    warn_lists <- lapply(processed_list, `[[`, "warns")
    to_warn    <- unique(unlist(warn_lists))
  } else {
    study_data0 <- withCallingHandlers(readr::type_convert(
      # somehow, tehre is a wrning for TIME, currently (in parallel but hopefully, also here)
      # r <- dq_report2(study_data, meta_data = meta_data,
      #    meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
      #    and use meta_data and study_data according to test-con_limit_devaitions with time-var
      study_data,
      col_types = col_types,
      trim_ws = TRUE,
      guess_integer = FALSE,
      locale = locale,
      na = c("", "NA", "Inf", "+Inf", "-Inf", "NaN")
    ), warning = function(c) {
      cm <- util_extract_named_groups(
        "^\\[(?<row>\\d+), +(?<col>\\d+)\\]:.*",
        conditionMessage(c))
      if (nrow(cm) > 0) {
        my_env$to_warn <- c(
          my_env$to_warn,
          intersect(relevant_vars_for_warnings,
                    colnames(study_data)[as.integer(cm[, "col"])])
        )
        invokeRestart("muffleWarning")
      }
    })
  }

  if (length(num) > 0) {
    study_data0[, num] <- mapply(
      SIMPLIFY = FALSE,
      cl = study_data0[, num, drop = FALSE],
      orig = study_data[, num, drop = FALSE],
      FUN = function(cl, orig) {
        if (any(is.na(cl))) {
          cl[is.na(cl) & orig %in% c("Inf", "+Inf", "-Inf", "NaN")] <-
            as.vector(as.numeric(
              orig[is.na(cl) & orig %in% c("Inf", "+Inf", "-Inf", "NaN")]))
        }
        cl
      }
    )
  }
  study_data <- study_data0
  rm(study_data0)
  names(study_data) <- old_names
  dt_cols <- vapply(study_data, inherits, "POSIXt", FUN.VALUE = logical(1))
  study_data[, dt_cols] <- util_par_lapply_lb(cl = mycl, study_data[, dt_cols, drop = FALSE],
                                  function(cl) {
                                    cl <- as.POSIXct(round(as.POSIXct(cl), units = "secs"))
                                    cl <- lubridate::with_tz(lubridate::force_tz(cl))
                                  })
  study_data[, intg] <- util_par_lapply_lb(cl = mycl, study_data[, intg, drop = FALSE],
                               function(cl) { # TRUE -> 1, ...
                                 floor(cl)
                               })
  if (length(to_warn) > 0) {
    tb <- table(to_warn)
    invisible(apply(as.data.frame(tb), 1, function(rv_Freq) {
      rv <- rv_Freq["to_warn"]
      Freq <- rv_Freq["Freq"] # is this number also correct or lists?
      util_message(
        paste("Data type transformation of", dQuote(rv), "introduced",
              Freq,
              "additional missing values."),
        applicability_problem = TRUE)
    }))
  }
  attr(study_data, "Data_type_matches") <- TRUE
  study_data
}

# m <- prep_get_data_frame("meta_data")
# s <- prep_get_data_frame("study_data", keep_types = TRUE)
# s <- rbind(s, s, s, s, s, s)
# s <- rbind(s, s, s, s, s, s)
#
#
# microbenchmark::microbenchmark(
#   util_adjust_data_type(s, meta_data = m),
#   util_adjust_data_type2(s, meta_data = m)
# )
#
# Unit: seconds
# expr      min       lq     mean   median       uq       max neval
# util_adjust_data_type(s, meta_data = m) 3.910896 4.021739 4.163222 4.046629 4.070163 15.708896   100
# util_adjust_data_type2(s, meta_data = m) 2.797905 2.935524 3.025309 2.994379 3.035081  4.840821   100
# >

#' Combine results for Single Variables
#'
#' to, e.g., a data frame with one row per variable or a similar heat-map,
#' see [print.ReportSummaryTable()].
#'
#' @param all_of_f all results of a function
#'
#' @return row-bound combined results
#'
#' @noRd
util_combine_res <- function(all_of_f) {
  # getting the call names for all_of_f (subset of the report)
  cn <- unique(unlist(lapply(all_of_f, attr, "cn")))
  if (length(cn) != 1) {
    cn <- ""
  }
  #remove the class dataquieR_result
  all_of_f <- lapply(all_of_f, function(x) {
    class(x) <- setdiff(class(x), "dataquieR_result")
    x
  })

  # combine results for the indicator functions related overview----
  # extracting call names (all until first dot)
  nms <- sub("^([^\\.]*).*$", "\\1", names(all_of_f))
  # combine only if results come from the same call
  util_stop_if_not(length(unique(nms)) == 1)
  cll <- nms[[1]]
  # get the function name
  fkt <- util_map_by_largest_prefix(
    cll,
    haystack = util_all_ind_functions())
  # try to combine results, mostly relevant for indicator function outputs (where single variable is false)
  # check the results for the existence of certain output types
  # each result is a logical vector
  PLOTs <- !vapply(lapply(all_of_f, `[[`, "SummaryPlot"), is.null,
                   FUN.VALUE = logical(1))
  PLOT_LISTSs <- !vapply(lapply(all_of_f, `[[`, "SummaryPlotList"), is.null,
                         FUN.VALUE = logical(1))

  STs <- !vapply(lapply(all_of_f, `[[`, "SummaryTable"), is.null,
                 FUN.VALUE = logical(1)) # TODO: Do the same for segment and data frame level output
  SDs <- !vapply(lapply(all_of_f, `[[`, "SummaryData"), is.null,
                 FUN.VALUE = logical(1)) # TODO: Do the same for segment and data frame level output
  RDs <- !vapply(lapply(all_of_f, `[[`, "ResultData"), is.null,
                 FUN.VALUE = logical(1)) # TODO: Do the same for segment and data frame level output
  RSTs <- !vapply(lapply(all_of_f, `[[`, "ReportSummaryTable"), is.null,
                  FUN.VALUE = logical(1))
  NULLs <- vapply(all_of_f, inherits, "dataquieR_NULL",
                  FUN.VALUE = logical(1))

  # TODO: Add "VariableGroupTable", "VariableGroupData",

  ERRORs <- util_collapse_msgs("error", all_of_f)
  WARNINGs <- util_collapse_msgs("warning", all_of_f)
  MESSAGEs <- util_collapse_msgs("message", all_of_f)

  if ((fkt %in% c(
    "con_limit_deviations",  # check if we are working with a limits function
    "con_hard_limits",
    "con_soft_limits",
    "con_detection_limits"
    )) && (any(PLOT_LISTSs) || any(PLOTs))) {
    return(all_of_f) # use the limits plots, if available, not the ReportSummaryTable if we are working with a limits function
    # check if we have to combine some single variable results to a multivariable result
  } else if ((!any(RSTs)) && (any(PLOT_LISTSs) || any(PLOTs))) {
    return(all_of_f) # use the plots, if available
    # otherwise, use the ReportSummaryTable, SummaryData, or SummaryTable, in that order of preference
    # and combine the results (rbind) using util_combine_res
  } else if (any(RSTs)) {
    RESs <- RSTs
    slot <- "ReportSummaryTable"
  } else if (any(RDs)) {
    RESs <- RDs
    slot <- "ResultData"
  } else if (any(SDs)) {
    RESs <- SDs
    slot <- "SummaryData"
  } else if (any(STs)) {
    RESs <- STs
    slot <- "SummaryTable"
  } else { # TODO: Add "VariableGroupTable", "VariableGroupData",
    return(all_of_f)
  }

  util_stop_if_not(all(RESs | NULLs)) # NULLs contains all results that are NULL, check that all results can be combined, except missing results

  # extract all call attributes to combine them
  clls <- lapply(all_of_f[RESs & !NULLs], attr, "call")
  clls <- lapply(clls, deparse)
  clls <- vapply(clls, paste0, collapse = "\n",
                 FUN.VALUE = character(1))
  clls <- paste0(clls, "$", slot, collapse = ", \n\t")
  clls <- paste0("rbind(\n\t", clls, "\n)")

  # for ReportSummaryTables: extract and combine VAR_NAMES attributes
  vns <- NULL
  if (any(RSTs)) {
    vns <- unlist(lapply(unname(lapply(all_of_f[RESs & !NULLs], `[[`, slot)), attr, "VAR_NAMES"))
  }

  # copy hover text for table headers
  description <- lapply(lapply(all_of_f[RESs & !NULLs], `[[`, slot), attr, "description")
  description <- unique(description)


  # rescue plain label attributes if assigned for single result tables
  plain_label_atts <- lapply(lapply(lapply(all_of_f[RESs & !NULLs],
                                           `[[`,
                                           slot), `[[`, "Variables"), attr, "plain_label")
  #plain label should be only 1 or no plain_label(error), be characters, not NA
  if (all(vapply(plain_label_atts, length, FUN.VALUE = integer(1)) %in% 0:1) &&
      all(vapply(plain_label_atts, is.character, FUN.VALUE = logical(1)) == TRUE) &&
      !any(vapply(plain_label_atts, identical, NA_character_, FUN.VALUE = logical(1)))) {
    plain_label_atts <- unname(unlist(plain_label_atts))
  } else {
    if (!is.null(plain_label_atts) &&
        !all(vapply(plain_label_atts, length, FUN.VALUE = integer(1)) == 0)) {
      util_error(
        "Internal error, sorry, please report: invalid plain label atts")
    }
    plain_label_atts <- NULL
  }

  # combine results (ReportSummaryTable and tables)
  # select results according to the logical vectors, extract the corresponding slots, and then bind by row
  # then write the combined result to all_of_f keeping its original structure (a list of encapsulated lists)
  all_of_f <- list(setNames(list(do.call(util_rbind, lapply(all_of_f[RESs & !NULLs],
                                                       `[[`,
                                                       slot))), nm = slot))
  if (!is.null(all_of_f[[1]][[slot]][["Variables"]])) {
    attr(all_of_f[[1]][[slot]][["Variables"]], "plain_label") <-
      plain_label_atts
  }
  attr(all_of_f[[1]], "call") <- clls
  if (!is.null(vns)) { # attach VAR_NAMES for ReportSummaryTables
    attr(all_of_f[[1]][[slot]], "VAR_NAMES") <- vns
  }

  # add attribute "description" for hover text
  if (length(description)>1) {
    util_warning(c("Internal error: sorry please report to the developers",
                   " - util_combine_res incompatible results"))
  } else if (length(description)==1) {
    attr(all_of_f[[1]][[slot]], "description") <- description[[1]]
  }


  # reattach the error/message/warning attributes but now using the combined version (in ERRORs/WARNINGs,...)
  if (any(trimws(ERRORs) != ""))
    attr(all_of_f[[1]], "error") <- list(simpleError(paste(ERRORs, collapse = "\n")))
  else
    attr(all_of_f[[1]], "error") <- list()
  if (any(trimws(WARNINGs) != ""))
    attr(all_of_f[[1]], "warning") <- list(simpleWarning(paste(WARNINGs, collapse = "\n")))
  else
    attr(all_of_f[[1]], "warning") <- list()
  if (any(trimws(MESSAGEs) != ""))
    attr(all_of_f[[1]], "message") <- list(simpleMessage(paste(MESSAGEs, collapse = "\n")))
  else
    attr(all_of_f[[1]], "message") <- list()
  names(all_of_f) <- cll

  attr(all_of_f[[1]], "cn") <- cn

  return(all_of_f)
}

#' Summarize a [dataquieR] report
#'
#' @param ... Deprecated
#'
#' @description
#' Deprecated
#' @return Deprecated
#' @export
#' @importFrom stats setNames
#' @importFrom dplyr %>%
summary.dataquieR_resultset <- function(...) {
  lifecycle::deprecate_stop("2.1.0", what = "dq_report()", with = "dq_report2()")
}

#' Get encoding from metadata or guess it from data
#'
#' @inheritParams .template_function_indicator
#'
#' @return named vector of valid encoding strings matching `resp_vars`
#' @noRd
util_get_encoding <- function(resp_vars = colnames(study_data),
                              study_data,
                              label_col,
                              meta_data,
                              meta_data_dataframe) {
  result <- setNames(rep(NA_character_, length(resp_vars)),
                     nm = resp_vars)
  # from item-level
  if (ENCODING %in% colnames(meta_data)) {
    meta_data[[ENCODING]] <- trimws(meta_data[[ENCODING]])
    meta_data[[ENCODING]][meta_data[[ENCODING]] == ""] <- NA_character_
    rvs <- resp_vars[is.na(result[resp_vars])]
    result[rvs] <-
      setNames(meta_data[[ENCODING]], nm = meta_data[[label_col]])[rvs]
    result[rvs] <- toupper(result[rvs])
    unkown <- setdiff(result, .encs_avail)
    if (length(unkown) > 0) {
      util_warning("Found unkown encodings in metadata: %s",
                   util_pretty_vector_string(unkown))
      result[(!(result[rvs] %in% .encs_avail))] <- NA_character_
    }
  }
  # from data-frame-level
  if (DATAFRAMES %in% colnames(meta_data) &&
      ENCODING %in% colnames(meta_data_dataframe) &&
      DF_CODE %in% colnames(meta_data_dataframe)) {
    meta_data_dataframe[[ENCODING]] <- trimws(meta_data_dataframe[[ENCODING]])
    meta_data_dataframe[[ENCODING]][meta_data_dataframe[[ENCODING]] == ""] <- NA_character_
    rvs <- resp_vars[is.na(result[resp_vars])]
    dfs <- setNames(meta_data[[DATAFRAMES]], nm = meta_data[[label_col]])
    result[rvs] <-
      vapply(rvs, function(rv) {
        dfcodes <- util_parse_assignments(dfs[[rv]])
        enc <- meta_data_dataframe[meta_data_dataframe[[DF_CODE]] %in%
                                         dfcodes, ENCODING]
        enc <- unique(enc)
        enc <- enc[!util_empty(enc)]

        if (length(enc) > 1) {
          util_warning(c("Found more than one encodings for variable %s",
                         "from dataframe level metadata, namely %s,",
                         "because the variable exists in the dataframes %s.",
                         "This is not yet supported. I'll use the first, %s."),
                       sQuote(rv),
                       util_pretty_vector_string(enc),
                       util_pretty_vector_string(dfcodes),
                       dQuote(head(enc, 1)),
                       applicability_problem = TRUE,
                       intrinsic_applicability_problem = TRUE)
          enc <- head(enc, 1)
        } else if (length(enc) < 1) {
          enc <- NA_character_
        }
        enc
      }, FUN.VALUE = character(1))
    result[rvs] <- toupper(result[rvs])
    unkown <- setdiff(result, .encs_avail)
    if (length(unkown) > 0) {
      util_warning("Found unkown encodings in metadata: %s",
                   util_pretty_vector_string(unkown))
      result[(!(result[rvs] %in% .encs_avail))] <- NA_character_
    }
  }
  # from guessing
  rvs <- resp_vars[is.na(result[resp_vars])]
  if (length(rvs) > 0) {
    util_warning(c("Encoding information neither in item-",
                 "nor dataframe-level-metadata found for %s, guessing from",
                 "data."),
                 util_pretty_vector_string(rvs), applicability_problem = TRUE)
  }
  result[rvs] <- vapply(setNames(nm = rvs),
         function(rv) {
           prep_guess_encoding(x = study_data[[rv]])$encoding[[1]]
          }, FUN.VALUE = character(1))
  result[rvs] <- toupper(result[rvs])
  result[(!(result[rvs] %in% .encs_avail))] <- NA_character_
  result[resp_vars]
}
.encs_avail <- toupper(c(iconvlist(), NA_character_, "unknown"))

#' Utility function Tukey outlier rule
#'
#' This function calculates outliers according to the rule of Tukey.
#'
#' @param x [numeric] data to check for outliers
#'
#' @return binary vector
#'
#' @importFrom stats quantile IQR
#'
#' @family outlier_functions
#' @concept outlier
#' @noRd

util_tukey <- function(x) {
  xq1 <- as.numeric(quantile(x, na.rm = TRUE)[2])
  xiqr <- IQR(x, na.rm = TRUE)
  xq3 <- as.numeric(quantile(x, na.rm = TRUE)[4])
  lth <- xq1 - 1.5 * xiqr
  uth <- xq3 + 1.5 * xiqr
  xbin <- ifelse(x < lth | x > uth, 1, 0)
  return(xbin)
}

#' Utility function to replace missing codes by `NA`s
#'
#' Substitute all missing codes in a [data.frame] by `NA`.
#'
#' @param study_data Study data including jump/missing codes as specified in the
#'                   code conventions
#' @param meta_data Metadata as specified in the code conventions
#' @param split_char Character separating missing codes
#' @param sm_code missing code for `NAs`, if they have been
#'                re-coded by `util_combine_missing_lists`
#'
#' Codes are expected to be numeric.
#'
#' @importFrom stats setNames
#' @return a list with a modified data frame and some counts
#'
#' @family missing_functions
#' @concept metadata_management
#' @noRd
util_replace_codes_by_NA <- function(study_data, meta_data = "item_level",
                                     split_char = SPLIT_CHAR,
                                     sm_code = NULL) {
  util_expect_scalar(sm_code, # TODO: This needs an update, too, but sm_code is only used, if the un-exported function util_combine_missing_lists would have been applied before
                     check_type = util_all_is_integer,
                     allow_null = TRUE)
  sdf <- study_data
  mdf <- meta_data
  # data_records is edited but not explicitly copied, so we can expect the
  # MAPPED-attribute to be preserved.
  util_expect_data_frame(sdf)
  util_expect_data_frame(mdf)

  label_col <- VAR_NAMES
  if (!is.null(attr(sdf, "label_col"))) {
    label_col <- attr(sdf, "label_col")
  }

  for (code_name in c(MISSING_LIST, JUMP_LIST)) {
    if (!(code_name %in% names(mdf)) || all(is.na(mdf[[code_name]]) |
                                            trimws(mdf[[code_name]]) == "")) {
      util_warning(
        c("Metadata does not provide a filled column",
          "called %s for replacing codes with NAs."),
        dQuote(code_name),
        applicability_problem = TRUE)
    }
  }

  # apply functions on studydata and create list
  miss <- lapply(setNames(nm = names(sdf)), util_get_code_list,
                 MISSING_LIST, split_char,
    mdf = mdf, label_col = label_col, warning_if_no_list = FALSE,
    warning_if_unsuitable_list = FALSE
  )
  jump <- lapply(setNames(nm = names(sdf)), util_get_code_list,
                 JUMP_LIST, split_char,
    mdf = mdf, label_col = label_col, warning_if_no_list = FALSE,
    warning_if_unsuitable_list = FALSE
  )

  # miss_n <- list()
  # jump_n <- list()
  #
  # miss_n <- lapply(
  #   mapply(`%in%`, sdf, miss, SIMPLIFY = FALSE, USE.NAMES = TRUE), sum,
  #   na.rm = TRUE)
  # jump_n <- lapply(
  #   mapply(`%in%`, sdf, jump, SIMPLIFY = FALSE, USE.NAMES = TRUE), sum,
  #   na.rm = TRUE)

  replace <- function(x, l, sm_code) {
    # if (lubridate::is.timepoint(x) && length(sm_code) && is.numeric(sm_code))
    if ((length(l) == 0 && length(sm_code == 0)) || !length(x)) {
      return(x)
    }
    if (lubridate::is.timepoint(x)) {
      if (!lubridate::is.timepoint(l)) {
        return(x)
      }
      x <- as.POSIXct(round(as.POSIXct(x), units = "secs"))
      x <- lubridate::with_tz(lubridate::force_tz(x))
      l <- as.POSIXct(round(as.POSIXct(l), units = "secs"))
      l <- lubridate::with_tz(lubridate::force_tz(l))
    } else if (lubridate::is.timepoint(l)) {
      return(x)
    }
    if (!is.null(sm_code)) {
      l <- unique(c(l, sm_code))
    }
    x[x %in% l] <- NA # TODO: do this a bit more stable wrt numeric issues
    x
  }
  environment(replace) <- parent.env(environment())

  sdf[] <- mapply(replace, x = sdf, l = miss,
                  MoreArgs = list(sm_code = sm_code),
                  SIMPLIFY = FALSE)
  sdf[] <- mapply(replace, x = sdf, l = jump,
                  MoreArgs = list(sm_code = sm_code),
                  SIMPLIFY = FALSE)

  attr(sdf, "Codes_to_NA") <- TRUE
  attr(sdf, "MAPPED") <- attr(study_data, "MAPPED")

  return(sdf)
}

test_that("util_replace_codes_by_na works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  local({
    meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
    study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
    sd0 <- study_data[1:30, 30:35, FALSE]
    md0 <- meta_data[meta_data$VAR_NAMES %in% colnames(sd0), , FALSE]
    expected <- structure(list(v00024 = c(
        0, 1, 0, NA, NA,
        1, 0, 0, 1, 0, NA, 1, 1, 0, 0, 1, NA, 0, 1, 0, NA, 0, 0, 0, 1,
        0, NA, 1, 1, 0
      ), v00025 = c(
        NA, 1, NA, NA, NA, NA, NA, NA, 4,
        NA, NA, 3, 0, NA, NA, 0, NA, NA, NA, NA, NA, NA, NA, NA, 1, NA,
        NA, 3, 3, NA
      ), v00026 = c(
        3, 7, 4, 6, NA, 11, 2, 7, 3, 8, NA,
        6, 7, 7, 4, 2, NA, NA, 3, 5, NA, 3, NA, 7, 8, 11, NA, 4, 7, 8
      ), v00027 = c(
        NA, 4, 3, NA, NA, NA, 4, NA, 3, 6, NA, NA, 2, 4,
        4, 2, NA, NA, NA, NA, NA, 2, NA, 2, NA, NA, NA, 4, NA, 2
      ), v00028 = c(
        1,
        2, 0, 2, NA, 2, 1, 3, 1, 1, NA, 2, 0, 2, 0, NA, NA, 1, 4, 1,
        NA, 1, 3, 5, 3, 3, NA, NA, 3, NA
      ), v00029 = c(
        0, 0, NA, 0, NA,
        NA, NA, NA, 0, 0, NA, NA, 0, 0, 1, 0, NA, NA, NA, NA, NA, 0,
        NA, 0, 1, NA, NA, 0, NA, 0
      )), row.names = c(NA, 30L), class = "data.frame", Codes_to_NA = TRUE)
    got <- dataquieR:::util_replace_codes_by_NA(
      study_data = sd0, meta_data = md0)
    expect_equal(got, expected)
  })
})

#' Create a predicate function to check for certain numeric properties
#'
#' useful, e.g., for `util_expect_data_frame()` and `util_expect_scalar()`. The
#' generated function returns on `TRUE` or `FALSE`, even if called with a
#' vector.
#'
#' @param min if given, minimum for numeric values
#' @param max if given, maximum for numeric values
#' @param whole_num if TRUE, expect a whole number
#' @param set if given, a set, the value must be in (see `util_match_arg()`)
#' @param finite Are `Inf` and `-Inf` invalid values? (FALSE by default)
#'
#' @return a function that checks an `x` for the properties.
#' @examples
#' \dontrun{
#' util_is_numeric_in(min = 0)(42)
#' util_is_numeric_in(min = 43)(42)
#' util_is_numeric_in(max = 3)(42)
#' util_is_numeric_in(whole_num = TRUE)(42)
#' util_is_numeric_in(whole_num = TRUE)(42.1)
#' util_is_numeric_in(set = c(1, 3, 5))(1)
#' util_is_numeric_in(set = c(1, 3, 5))(2)
#' }
#'
#' @family robustness_functions
#' @concept data_management
#' @noRd
util_is_numeric_in <- function(min = -Inf,
                               max = +Inf,
                               whole_num = FALSE,
                               finite = FALSE,
                               set = NULL) {
  error_msg <- "be numeric"
  if (finite) {
    error_msg <- paste(error_msg, "and finite")
  }
  if (whole_num) {
    error_msg <- paste(error_msg, "and a whole number")
  }
  if (length(set)) {
    error_msg <- paste(error_msg, sprintf("and in %s",
                                         util_pretty_vector_string(set)))
  }
  if (is.finite(min)) {
    error_msg <- paste(error_msg, sprintf("and larger/equal than %s",
                                         sQuote(min)))
  }
  if (is.finite(max)) {
    error_msg <- paste(error_msg, sprintf("and smaller/equal than %s",
                                         sQuote(max)))
  }
  util_attach_attr(function(x) {
    if (!is.numeric(x)) {
      return(FALSE)
    }
    if (finite && any(!is.finite(x))) {
      return(FALSE)
    }
    if (whole_num && any(!util_is_integer(x))) {
      return(FALSE)
    }
    if (any(x > max | x < min)) {
      return(FALSE)
    }
    if (!is.null(set)) {
      if (any(!(x %in% set))) {
        return(FALSE)
      }
    }
    TRUE
  }, error_msg = error_msg)
}

#' Replacement for `htmltools::plotTag`
#'
#' the function is specifically designed for fully scalable `SVG` figures.
#'
#' @param expr plot expression
#' @param w width
#' @param h height
#'
#' `w` and `h` are mostly used for the relation of fixed text sizes to
#' the figure size.
#'
#' @return `htmltools` compatible object
#' @noRd
util_plot_svg_to_uri <- function(expr, w = 800, h = 600) {
  tmpfil <- NULL
  withr::with_tempfile("tmpfil", fileext = ".svg", {
    htmltools::capturePlot(expr = {
      rlang::eval_tidy(expr)
    },
    filename = tmpfil,
    device = grDevices::svg,
    width = w / 72, height = h / 72,
#    pointsize = 12
#    res = 1/72
    )
    .svg <- readLines(tmpfil)
    .svg <- gsub("<svg ", '<svg preserveAspectRatio="none" ', .svg, fixed = TRUE)
    writeLines(.svg, tmpfil)
    b64 <- R.devices::asDataURI(tmpfil,
                         mime = "image/svg+xml")
    htmltools::browsable(htmltools::tags$img(src = b64,
                                             class = "dataquieRfigure"))
  })
}

#' similar to match.arg
#'
#' will only warn and return a cleaned `x`.
#'
#' @param x [character] vector of needles
#' @param set [character] vector representing the haystack
#' @param err_msg [character] optional error message. Use %s twice, once for
#'                            the missing elements and once for proposals
#' @param error [logical] if `TRUE`, the execution will stop with an error,
#'                        if not all `x` are elements of `set`, otherwise,
#'                        it will throw a warning and "clean" the vector `x`
#'                        from unexpected elements.
#' @param applicability_problem [logical] error indicates unsuitable resp_vars
#'
#' @return [character] invisible(intersect(x, set))
#'
#' @family robustness_functions
#' @concept robustness
#' @noRd
util_ensure_in <- function(x, set, err_msg, error = FALSE, applicability_problem = NA) {
  util_expect_scalar(error, check_type = is.logical)
  if (missing(err_msg)) {
    err_msg <- sprintf("Missing %%s from %s, did you mean %%s?", sQuote(
      util_deparse1(substitute(set))))
  }
  mis <- !(x %in% set)
  if (sum(mis) > 0) {

    prop <-
      vapply(x[mis],
           FUN.VALUE = character(1),
           function(v) {
              set[which.min(adist(trimws(v),
                                  trimws(set),
                                  ignore.case = TRUE,
                                  fixed = TRUE))]
           }
      )

    ifelse(error, util_error, util_warning)(
      err_msg,
      paste0(dQuote(x[mis]), collapse = ", "),
      paste0(dQuote(prop), collapse = ", "),
      applicability_problem = applicability_problem
    )
    x[mis]
  }
  invisible(intersect(x, set))
}

#' Combine `ReportSummaryTable` outputs
#'
#' @param ... `ReportSummaryTable` objects to combine.
#'
#' @description
#' Using this `rbind` implementation, you can combine different
#' heatmap-like results of the class `ReportSummaryTable`.
#'
#' @seealso [base::rbind.data.frame]
#'
#' @export
#'
rbind.ReportSummaryTable <- function(...) {
  a <- list(...)
  if (!all(vapply(a, is.data.frame, FUN.VALUE = logical(1)))) {
    util_error("Can only bind ReportSummaryTables")
  }
  if (!all(vapply(a, inherits, what = "ReportSummaryTable",
                  FUN.VALUE = logical(1)))) {
    util_error("Can only bind ReportSummaryTables")
  }
  a <- a[!!vapply(a, nrow, FUN.VALUE = integer(1))]
  a <- a[!!vapply(a, ncol, FUN.VALUE = integer(1))]
  if (length(a) <= 2) {
    if (length(a) == 0) {
      x <- rbind.data.frame(
        deparse.level = 1,
        make.row.names = TRUE,
        stringsAsFactors = FALSE,
        factor.exclude = TRUE
      )
      class(x) <- union("ReportSummaryTable", class(x))
      return(x)
    } else if (length(a) == 1) {
      x <- a[[1]]
      y <- data.frame(Variables = character(0), N = character(0))
      class(y) <- union("ReportSummaryTable", class(y))
    } else if (length(a) == 2) {
      x <- a[[1]]
      y <- a[[2]]
    }

    if (nrow(x) == 0) {
      return(y)
    }

    if (nrow(y) == 0) {
      return(x)
    }

    cols <- union(colnames(x), colnames(y))
    x[setdiff(cols, colnames(x))] <- numeric(nrow(x))
    y[setdiff(cols, colnames(y))] <- numeric(nrow(y))
    r <- rbind.data.frame(x[, cols, drop = FALSE], y[, cols, drop = FALSE],
                          deparse.level = 1,
                          make.row.names = TRUE,
                          stringsAsFactors = FALSE,
                          factor.exclude = TRUE
    )
    class(r) <- union("ReportSummaryTable", class(r))
    attr(r, "higher_means") <- attr(x, "higher_means")
    attr(r, "flip_mode") <- attr(x, "flip_mode")
    attr(r, "continuous") <- attr(x, "continuous")
    attr(r, "colscale") <- attr(x, "colscale")
    attr(r, "colcode") <- attr(x, "colcode")
    attr(r, "level_names") <- attr(x, "level_names")
    attr(r, "relative") <- attr(x, "relative")
    attr(r, "VAR_NAMES") <- c(attr(x, "VAR_NAMES"), attr(y, "VAR_NAMES"))
    r
  } else { # recursive call to bind ReportSummaryTables for more than two variables
    x <-
      do.call(Recall,
            c(list(
              a[[1]],
              a[[2]]
            ),
            list(
              # deparse.level = 1,
              # make.row.names = TRUE,
              # stringsAsFactors = FALSE,
              # factor.exclude = TRUE
            ))
    )
    y <- a[3:length(a)]
    do.call(Recall,
            c(list(
              x
            ),
            y,
            list(
                # deparse.level = 1,
                # make.row.names = TRUE,
                # stringsAsFactors = FALSE,
                # factor.exclude = TRUE
              ))
    )
  }
}

test_that("rbind.ReportSummaryTable works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  md0 <- meta_data

  expect_true(all(dim(rbind.ReportSummaryTable()) == 0))

  empty <- data.frame()
  expect_error(rbind.ReportSummaryTable(empty),
               regexp = "Can only bind ReportSummaryTables")
  class(empty) <- union("ReportSummaryTable", class(empty))

  expect_true(all(dim(rbind.ReportSummaryTable(empty)) == 0))

  im1 <- suppressMessages(suppressWarnings(com_item_missingness(
    resp_vars = "SBP_0",
                                               study_data, md0,
                              label_col = LABEL,
                              suppressWarnings = TRUE,
                              threshold_value = 100,
                              include_sysmiss = FALSE,
                              show_causes = TRUE,
                              drop_levels = TRUE,
                              assume_consistent_codes = TRUE,
                              expand_codes = TRUE)))

  expect_equal(
    rbind.ReportSummaryTable(im1$ReportSummaryTable),
    structure(
      list(
        "MISSING 99980" = 1L,
        "MISSING 99981" = 5L,
        "MISSING 99982" = 3L,
        "MISSING 99983" = 4L,
        "MISSING 99984" = 4L,
        "MISSING 99985" = 3L,
        "MISSING 99986" = 4L,
        "MISSING 99987" = 6L,
        "MISSING 99988" = 4L,
        "MISSING 99989" = 4L,
        "MISSING 99990" = 82L,
        "MISSING 99991" = 6L,
        "MISSING 99992" = 2L,
        "MISSING 99993" = 2L,
        "MISSING 99994" = 9L,
        "MISSING 99995" = 1L,
        Variables = structure("SBP_0", label_col =
                                "LABEL"),
        N = 2940L
      ),
      row.names = c(NA,-1L),
      class = c("ReportSummaryTable", "data.frame"),
      VAR_NAMES = setNames("v00004", "SBP_0")
    )
  )

  im2 <- suppressWarnings(suppressMessages(com_item_missingness(resp_vars = "DBP_0", study_data, md0,
                              label_col = LABEL,
                              suppressWarnings = TRUE,
                              threshold_value = 100,
                              include_sysmiss = FALSE,
                              show_causes = TRUE,
                              drop_levels = TRUE,
                              assume_consistent_codes = TRUE,
                              expand_codes = TRUE)))

  expect_equal(
    rbind.ReportSummaryTable(im2$ReportSummaryTable),
    structure(
      list(
        "MISSING 99980" = 3L,
        "MISSING 99981" = 3L,
        "MISSING 99982" = 4L,
        "MISSING 99983" = 6L,
        "MISSING 99984" = 7L,
        "MISSING 99985" = 12L,
        "MISSING 99986" = 5L,
        "MISSING 99987" = 3L,
        "MISSING 99988" = 4L,
        "MISSING 99989" = 5L,
        "MISSING 99990" = 85L,
        "MISSING 99991" = 10L,
        "MISSING 99992" = 5L,
        "MISSING 99993" = 4L,
        "MISSING 99994" = 3L,
        "MISSING 99995" = 4L,
        Variables = structure("DBP_0", label_col =
                                "LABEL"),
        N = 2940L
      ),
      row.names = c(NA,-1L),
      class = c("ReportSummaryTable", "data.frame"),
      VAR_NAMES = setNames("v00005", "DBP_0")
    )
  )

  im3 <- suppressWarnings(suppressMessages(
    com_item_missingness(resp_vars = c("AGE_0", "AGE_1"), study_data, md0,
                              label_col = LABEL,
                              suppressWarnings = TRUE,
                              threshold_value = 100,
                              include_sysmiss = FALSE,
                              show_causes = TRUE,
                              drop_levels = TRUE,
                              assume_consistent_codes = TRUE,
                              expand_codes = TRUE)))
  expect_equal(
    rbind.ReportSummaryTable(im3$ReportSummaryTable),
    structure(
      list(
        Variables = structure(c("AGE_0", "AGE_1"), label_col =
                                "LABEL"),
        N = c(2940L,  2940L)
      ),
      row.names = c(NA,-2L),
      class = c("ReportSummaryTable",  "data.frame"),
      VAR_NAMES = setNames(c("v00003", "v01003"), c("AGE_0", "AGE_1"))
    )
  )

  expect_equal(rbind.ReportSummaryTable(im2$ReportSummaryTable,
                                        im3$ReportSummaryTable,
                                        im1$ReportSummaryTable),
               structure(
                 list(
                   "MISSING 99980" = c(3, 0, 0, 1),
                   "MISSING 99981" = c(3,  0, 0, 5),
                   "MISSING 99982" = c(4, 0, 0, 3),
                   "MISSING 99983" = c(6,  0, 0, 4),
                   "MISSING 99984" = c(7, 0, 0, 4),
                   "MISSING 99985" = c(12,  0, 0, 3),
                   "MISSING 99986" = c(5, 0, 0, 4),
                   "MISSING 99987" = c(3,  0, 0, 6),
                   "MISSING 99988" = c(4, 0, 0, 4),
                   "MISSING 99989" = c(5,  0, 0, 4),
                   "MISSING 99990" = c(85, 0, 0, 82),
                   "MISSING 99991" = c(10,  0, 0, 6),
                   "MISSING 99992" = c(5, 0, 0, 2),
                   "MISSING 99993" = c(4,  0, 0, 2),
                   "MISSING 99994" = c(3, 0, 0, 9),
                   "MISSING 99995" = c(4,  0, 0, 1),
                   Variables = structure(c("DBP_0", "AGE_0", "AGE_1", "SBP_0"),
                                         label_col = "LABEL"),
                   N = c(2940L, 2940L, 2940L, 2940L)
                 ),
                 row.names = c(NA,-4L),
                 class = c("ReportSummaryTable", "data.frame"),
                 VAR_NAMES = setNames(c("v00005", "v00003", "v01003", "v00004"),
                                      c("DBP_0", "AGE_0", "AGE_1", "SBP_0"))
               )
  )

})

util_fetch_ext <- function(url) {
  #ask the server for headers
  headers <- try(base::curlGetHeaders(url), silent = TRUE)
  if (inherits(headers, "try-error")) {
    return(attr(headers, "condition"))
  } else {
    # TODO: content-disposition
    #search for the content type
    mt <- grep("^content-type:", headers, value = TRUE, ignore.case = TRUE)
    if (length(mt) != 1) {
      return(errorCondition("No content-type header found"))
    }
  }
  #remove the content_type at the beginning of mime
  mt <- trimws(gsub("^content-type:", "", trimws(mt), ignore.case = TRUE))
  #delete everything after the ; including the ; (extra information)
  mt <- gsub(";.*$", "", mt)
  #look to the corresponding file extension from a table delivered with the
  # package (if more than 1 extension, takes the first)
  r <- head(unlist(strsplit(
    subset(.mime_chache$tb, get("Media Type") == mt, "Extensions", drop = TRUE),
    "\\s",
  )), 1)
  #get the file name from the header content-disposition
  file_name <- grep("^content-disposition", headers, value = TRUE,
                    ignore.case = TRUE)
  if (length(file_name) == 1 && is.character(file_name)) {
    # see https://httpwg.org/specs/rfc6266.html#n-grammar
    # and https://www.rfc-editor.org/rfc/rfc5987.html#section-3.2
    if (grepl("filename*=", tolower(file_name), fixed = TRUE)) {
      file_name <- utils::URLdecode(
        gsub(".*filename\\*\\s*=\\s*.*[^']*'[^']*'([^;]*).*$", "\\1",
             file_name, ignore.case = TRUE)) # TODO:  we ignore charset and language
      attr(r, "file-name") <- file_name
    } else if (grepl("filename=", tolower(file_name), fixed = TRUE)) {
      file_name <-
        gsub("^.*filename\\s*=\\s*\"(.*)\".*$", "\\1", file_name,
             ignore.case = TRUE)
      attr(r, "file-name") <- file_name
    }
  }
  r
}

.mime_chache <- new.env(parent = emptyenv())

local({
  f <- system.file("mimetypes", "media-types.RDS", package = "dataquieR")
  if (0 != file.access(f, 4)) {
    util_error("Cannot read file %s. Internal error.", dQuote(f))
  }
  tb <- readRDS(f)
  assign("tb", tb, envir = .mime_chache)
})

#' Render a table summarizing dataquieR results
#'
#' @param x a report summary (`summary(r)`)
#' @param grouped_by define the columns of the resulting matrix. It can be either
#'                  "call_names", one column per function, or "indicator_metric",
#'                  one column per indicator or both
#'                  `c("call_names", "indicator_metric")`.
#'                  The last combination is the default
#' @param folder_of_report a named vector with the location of variable and call_names
#' @param var_uniquenames a data frame with the original variable names and the
#'                        unique names in case of reports created with dq_report_by
#'                        containing the same variable in several reports
#'                        (e.g., creation of reports by sex)
#' @inheritParams util_filter_repsum
#' @return something, `htmltools` can render
#'
#' @noRd
#'
util_render_table_dataquieR_summary <- function(x,
                                                grouped_by =
                                                  c("call_names",
                                                    "indicator_metric"),
                                                folder_of_report = NULL,
                                                var_uniquenames = NULL,
                                                vars_to_include = c("study")) { # FIXME: -> util_*, remove other table printer functions, make prep_extract_summary util_* and, maybe, remove ggplot2 alternative pre for pie charts
  # make check happy:
  ordered_call_names <- NULL
  ordered_var_names <- NULL
  html <- NULL

  #order grouped_by if more than one is present
  if (length(grouped_by) > 1) {
    grouped_by <- sort(grouped_by)
  }

  # start
  util_stop_if_not(inherits(x, "dataquieR_summary"))
  # this represent the current object
  this <- attr(x, "this")
  # define an environment to be attached
  withr::with_environment(this, {
    for (n in names(this)) {
      assign(n, get(n, envir = this), envir = environment())
    }
    result <- util_filter_repsum(result, vars_to_include, meta_data,
                                 rownames_of_report,
                                 label_col)
    rownames_of_report <- attr(result, "rownames_of_report")
    labels <- this$labels #obviously not attached by with_environment
    # extract the colors from the grading_format file
    gg_colors <- util_get_colors()
    # rename the categories to have "Ok" or "Critical" instead of cat1 or cat5
    names(gg_colors) <- util_get_labels_grading_class()[names(gg_colors)]

    # For robustness: in case none of the assessments work, create an empty result
    # if result is empty or contains only errors
    if (!prod(dim(result)) || all(
      startsWith(as.character(result$indicator_metric), "CAT_") |
      startsWith(as.character(result$indicator_metric), "MSG_"))) {
      # create a paragraph stating that there is no result
      res <- (htmltools::browsable(htmltools::tagList(htmltools::p(
        "Empty summary"))))
      # create an empty dataframe
      result2 <- data.frame()
      # use var_names as labels
      label_col <- VAR_NAMES
    } else {  # in the case of actual metrics and not only error messages
      #===== Table and html text preparation ===
      result_full <- result

      #create a named vector with names = call_names and values = function_name
      function_name_mapping <- setNames(result_full$function_name,
                                        nm = result_full$call_names)

      # create the function name that corresponds to the actual call_names,
      # e.g., with con_hard_limits instead of con_limit_deviations
      result_full$function_name2 <-
        apply(result_full[, "call_names",
                          drop = FALSE],
              1, FUN = function(x) {
                util_map_by_largest_prefix(x,
                                           haystack = util_all_ind_functions())
              })

      # add a column with suffixes
      result_full <-
        dplyr::mutate(result_full,
                      suffixes = ifelse(
                        startsWith(result_full[["call_names"]],
                                   result_full[["function_name2"]]),
                        substr(result_full[["call_names"]],
                               nchar(result_full[["function_name2"]]) +
                                 1 + 1,
                               nchar(result_full[["call_names"]])),
                        # name + "_" (first +1), start is the next character (second +1)
                        result_full[["call_names"]]
                      ))

      # create a new column with indicator_metric and suffixes
      # if present (excluding MSG and cAT indicators)
      result_full <-
        dplyr::mutate(result_full,
                      indicator_metric_suff =
                        ifelse(startsWith(as.character(result_full$indicator_metric),
                                          "MSG_") |
                                 startsWith(as.character(result_full$indicator_metric),
                                            "CAT_"),
                               result_full[["indicator_metric"]],
                               ifelse(!util_empty(result_full[["suffixes"]]),
                                      paste0(result_full[["indicator_metric"]],
                                             "_",
                                             result_full[["suffixes"]]),
                                      result_full[["indicator_metric"]])))

      # Create 3 different objects containing errors, category of errors, and
      # the actual results with the indicator metrics

      # Start selecting everything that starts with MSG (Error message)
      result_err_msg <-
        result_full[startsWith(as.character(result_full$indicator_metric),
                               "MSG_"),
                    c(VAR_NAMES, "call_names",
                      "indicator_metric_suff",
                      "class"),
                    drop = FALSE]
      # modify the class to be "err_msg"
      result_err_msg <- dplyr::rename(result_err_msg, err_msg = "class")

      # Select everything starting with CAT (Error category)
      result_err_cat <-
        result_full[startsWith(as.character(result_full$indicator_metric),
                               "CAT_"),
                    c(VAR_NAMES, "call_names",
                      "indicator_metric_suff",
                      "class"),
                    drop = FALSE]
      # Rename the class for CAT errors as "err_cat"
      result_err_cat <- dplyr::rename(result_err_cat, err_cat = "class")

      # Remove all error messages (CAT and MSG) and obtain an object with actual
      # results (metrics) only
      result_full <-
        result_full[!startsWith(as.character(result_full$indicator_metric_suff),
                                "CAT_"), , drop = FALSE]
      result_full <-
        result_full[!startsWith(as.character(result_full$indicator_metric_suff),
                                "MSG_"), , drop = FALSE]
      #=====  Add or modify columns in the 3 objects before merging them ======

      #===== prepare the object with the results =======

      # Add a column with the numeric value of class e.g., 1 to 5
      result_full$class_num <- result_full$class #save it first as numeric

      # Convert column "class" to factor (e.g., Critical, Ok, NA)
      class_categ <- util_get_labels_grading_class()
      result_full$class <- factor(result_full$class,
                                  # take the levels of the grading and the labels
                                  # from the grading_ruleset
                                  levels = names(class_categ),
                                  labels = class_categ,
                                  ordered = TRUE)
      rm(class_categ)

      #===== prepare the object with the error messages =======

      # Remove empty rows and leave only rows with actual error messages
      result_err_msg <- result_err_msg[!util_empty(result_err_msg$err_msg), ,
                                       drop = FALSE]

      #remove duplicated rows in case present
      result_err_msg<- dplyr::distinct(result_err_msg)

      # in case there are duplicated rows due to the message
      # "No results computed"
      # separate duplicates in a group and then remove the
      # "No results computed" duplicates before to add them back
      values_with_dupl <- subset(
        result_err_msg,
        stats::ave(seq_len(nrow(result_err_msg)),
                   interaction(result_err_msg[, c("indicator_metric_suff",
                                                  "VAR_NAMES", "call_names")]),
                   FUN = length
        ) > 1
      )

      result_err_msg_without_duplicate <-
        dplyr::anti_join(result_err_msg,
                         values_with_dupl,
                         by = c("indicator_metric_suff",
                                "VAR_NAMES",
                                "call_names",
                                "err_msg"))
      values_with_dupl <-
        values_with_dupl[order(values_with_dupl[,"indicator_metric_suff"],
                               values_with_dupl[,"VAR_NAMES"],
                               values_with_dupl[,"call_names"],
                               values_with_dupl[, "err_msg"]) ,]
      values_with_dupl <- values_with_dupl[!grepl("No results computed",
                                                  values_with_dupl$err_msg),]
      result_err_msg <- rbind(result_err_msg_without_duplicate,
                              values_with_dupl)

      # from long to wide format
      # result_err_msg <- tidyr::spread(result_err_msg, indicator_metric, err_msg)
      result_err_msg <- stats::reshape(result_err_msg,
                                       idvar = c("VAR_NAMES", "call_names"),
                                       timevar = "indicator_metric_suff",
                                       direction = "wide")

      names(result_err_msg)[startsWith(names(result_err_msg), "err_msg.")] <-
        substr(names(result_err_msg)[startsWith(names(result_err_msg),
                                                "err_msg.")],
               nchar("err_msg.") + 1,
               nchar(names(result_err_msg)[startsWith(names(result_err_msg),
                                                      "err_msg.")]))

      # Robustness: in case of empty cells replace them with NAs
      result_err_msg[util_empty(result_err_msg)] <- NA

      # add tag <li> to each error message (if the column exists in the report)
      if("MSG_anamat" %in% colnames(result_err_msg)) {
        result_err_msg <- dplyr::mutate(result_err_msg,
                                        MSG_anamat =
                                          ifelse(!util_empty(get("MSG_anamat")),
                                                 paste0("<li>",
                                                        get("MSG_anamat"),
                                                        "</li>"),
                                                 NA))
      }

      if("MSG_applicability" %in% colnames(result_err_msg)) {
        result_err_msg <-
          dplyr::mutate(result_err_msg,
                        MSG_applicability =
                          ifelse(!util_empty(get("MSG_applicability")),
                                 paste0("<li>",
                                        get("MSG_applicability"),
                                        "</li>"), NA))
      }

      if("MSG_error" %in% colnames(result_err_msg)) {
        result_err_msg <- dplyr::mutate(result_err_msg,
                                        MSG_error =
                                          ifelse(!util_empty(get("MSG_error")),
                                                 paste0("<li>",
                                                        get("MSG_error"),
                                                        "</li>"), NA))
      }

      # create a column with all error messages (if there are error messages)
      # FIXED: Here, error messages are copied to all variables.
      # We need to group by VAR_NAMES and call_names
      ky_cols <- c("VAR_NAMES", "call_names")
      dt_cols <- setdiff(colnames(result_err_msg), ky_cols)
      result_err_msg <- tapply(simplify = FALSE,
                               result_err_msg, as.list(result_err_msg[, ky_cols]),
                               function(x) {
                                 res <- unique(x[, ky_cols, FALSE])
                                 if (nrow(x) == 0) {
                                   return("")
                                 }
                                 paste(unique(
                                   na.omit(unlist(x[, dt_cols], recursive = TRUE))),
                                   collapse = "\n")
                               })
      result_err_msg <- array2DF(result_err_msg,
                                 responseName = "All_err_msg",
                                 simplify = FALSE,
                                 allowLong = TRUE)

      #===== prepare the object with the categories of errors =======

      # Remove empty rows and leave only rows with actual error messages
      result_err_cat <- result_err_cat[!util_empty(result_err_cat$err_cat), ,
                                       drop = FALSE]
      #remove duplicated rows in case present
      result_err_cat <- dplyr::distinct(result_err_cat)

      # in case there are duplicated rows separate duplicates in
      # a group and then order by decreasing values of err_cat
      # then keep only worst (higher values) of err_cat
      values_with_dupl <- subset(
        result_err_cat,
        stats::ave(seq_len(nrow(result_err_cat)),
                   interaction(result_err_cat[, c("indicator_metric_suff",
                                                  "VAR_NAMES", "call_names")]),
                   FUN = length) > 1)

      result_err_cat_without_duplicate <-
        dplyr::anti_join(result_err_cat,
                         values_with_dupl,
                         by = c("indicator_metric_suff",
                                "VAR_NAMES",
                                "call_names",
                                "err_cat"))
      values_with_dupl$err_cat <- as.numeric(values_with_dupl$err_cat)
      # order by decreasing values of err_cat
      values_with_dupl <-
        values_with_dupl[order(values_with_dupl[,"indicator_metric_suff"],
                               values_with_dupl[,"VAR_NAMES"],
                               values_with_dupl[,"call_names"],
                               -values_with_dupl[, "err_cat"]) ,]
      values_with_dupl <-    #keep first row of duplicated values
        values_with_dupl[!duplicated(values_with_dupl[,
                                                      c("indicator_metric_suff",
                                                        "VAR_NAMES",
                                                        "call_names")]),]
      result_err_cat <- rbind(result_err_cat_without_duplicate, values_with_dupl)

      #from long to wide format
      #result_err_cat <- tidyr::spread(result_err_cat, indicator_metric, err_cat)
      result_err_cat <- stats::reshape(result_err_cat,
                                       idvar = c("VAR_NAMES", "call_names"),
                                       timevar = "indicator_metric_suff",
                                       direction = "wide")

      names(result_err_cat)[startsWith(names(result_err_cat), "err_cat.")] <-
        substr(names(result_err_cat)[startsWith(names(result_err_cat),
                                                "err_cat.")],
               nchar("err_cat.") + 1,
               nchar(names(result_err_cat)[startsWith(names(result_err_cat),
                                                      "err_cat.")]))

      # Add column other_error_stop5 - logic T/F; if T unexpected error has occurred
      result_err_cat$other_error_stop5 <- result_err_cat$CAT_error == 5

      # Add column is_indicator - logic T/F; if F is not an indicator
      result_err_cat$is_indicator <-
        result_err_cat$CAT_indicator_or_descriptor == 1

      #===== Merge the 3 object in one data frame =======

      # Merge results and errors in one object (errors become 4 extra columns)
      if (!VAR_NAMES %in% colnames(result_err_msg))
        result_err_msg[[VAR_NAMES]] <- rep(NA_character_, nrow(result_err_msg))
      result_full <- merge(
        result_full,
        result_err_msg,
        by = c(VAR_NAMES, "call_names"), all = TRUE)

      if (!VAR_NAMES %in% colnames(result_err_cat))
        result_err_cat[[VAR_NAMES]] <- rep(NA_character_, nrow(result_err_cat))
      result_full <- merge(
        result_full,
        result_err_cat,
        by = c(VAR_NAMES, "call_names"), all = TRUE)


      #===== Modify the obtained dataframe =======

      # Sort result_full by indicator_metrics
      result_full <- result_full[order(result_full[["indicator_metric"]]), ,
                                 drop = FALSE]

      # For robustness: names in labels and colors should match, otherwise
      # there is inconsistency in the grading_ruleset.
      # Check if the names match
      if (!identical(names(labels), names(colors)) ||
          !all(paste0("cat", seq_len(5)) %in% names(colors))) {
        util_warning(
          c("Could not find enough categories in custom format file,",
            "falling back to default"), applicability_problem = TRUE,
          intrinsic_applicability_problem = FALSE)
        old_o <- options(
          dataquieR.grading_formats =
            system.file("grading_formats.xlsx",
                        package = "dataquieR"),
          dataquieR.grading_rulesets =
            system.file("grading_rulesets.xlsx",
                        package = "dataquieR"))
        on.exit(options(old_o))
        labels <- util_get_labels_grading_class()
        colors <- util_get_colors()
        names(labels) <- paste0("cat", names(labels))
        names(colors) <- paste0("cat", names(colors))
      }

      # Add a column T/F to check if the indicator metric has a value
      result_full$have_value <- !util_empty(result_full$indicator_metric)

      # Add a column label
      result_full$LABEL <-
        this$labels_of_var_names_in_report[result_full$VAR_NAMES]

      # Add a column islab - with the category (e.g., "Ok") or "Not classified"
      result_full <-
        result_full %>% dplyr::mutate(islab = ifelse(!util_empty(class),
                                                     as.character(class),
                                                     "Not classified"))

      # Add a logic vector checking if the result is reasonably possible
      if("CAT_anamat" %in% colnames(result_full)) {
        result_full$not_applicable <- result_full[["CAT_anamat"]] != 1
      }

      # Add a logic vector checking if there are all metadata to compute the
      # indicator, if the value is not 1 something is missing
      if("CAT_applicability" %in% colnames(result_full)) {
        result_full$meta_data_sufficient <-
          result_full[["CAT_applicability"]] == 1
      }
      # Add a column with the categories as. cat1, cat2 and so on
      result_full$class_cat <- util_as_cat(result_full$class_num)

      # Add a column "class_color" with the corresponding color of the categories and NAs
      result_full$class_color <-  NA
      result_full$class_color[is.na(result_full$class_color) &
                                !is.na(result_full$class_cat) &
                                result_full$is_indicator] <-
        colors[paste(result_full$class_cat[is.na(result_full$class_color) &
                                             !is.na(result_full$class_cat) &
                                             result_full$is_indicator])]

      if("not_applicable" %in% colnames(result_full)){
        result_full$class_color[is.na(result_full$class_color) &
                                  result_full$not_applicable &
                                  result_full$is_indicator] <-  "#ffffff"
      }


      if("meta_data_sufficient" %in% colnames(result_full)){
        result_full$class_color[is.na(result_full$class_color) &
                                  !result_full$meta_data_sufficient &
                                  result_full$is_indicator] <- "#444444"
      }

      result_full$class_color[is.na(result_full$class_color) &
                                (result_full$other_error_stop5 |
                                   result_full$is_indicator)] <-  "#aaaaaa"

      result_full$class_color[is.na(result_full$class_color)] <-  "#ffffff"

      # Add a column to determine the brightness of the text
      # (black/white depending on the background)
      result_full$fg_color <- vapply(result_full$class_color,
                                     FUN = util_get_fg_color,
                                     FUN.VALUE = character(1))

      # create a column title_text - containing the title for the hover text
      result_full$title <-
        vapply(unique(result_full$call_names),
               util_alias2caption,
               long = FALSE,
               FUN.VALUE= character(1))[result_full$call_names]

      result_full <- result_full %>%
        dplyr::mutate(title_text = paste0(get("title"), " of ", get("LABEL")))

      # create the text of the hover box. A tag for header of level 4 with nice caption
      # and a tag for classification (e.g. Critical)
      result_full$hover_text <- paste0("<h4>", result_full[["title_text"]],
                                       "</h4> <h5>", result_full[["islab"]],
                                       "</h5>")
      result_full <- result_full[, -which(names(result_full) == c("title_text")),
                                 drop = FALSE]


      #create an alternative title_text and hover_text in case only indicator metric is selected
      result_full$title_ind <- vapply(unique(result_full$indicator_metric),
                                      util_translate_indicator_metrics,
                                      FUN.VALUE=
                                        character(1))[result_full$indicator_metric]

      result_full <- result_full %>% dplyr::mutate(title_text_ind =
                                                     paste0(get("title_ind"),
                                                            " of ",
                                                            get("LABEL")))

      # create the text of the hover box. A tag for header of level 4 with nice caption
      # and a tag for classification (e.g. Critical)
      result_full$hover_text_ind <-
        paste0("<h4>", result_full[["title_text_ind"]],
               "</h4> <h5>", result_full[["islab"]],
               "</h5>")

      result_full <- result_full[, -which(names(result_full) ==
                                            c("title_text_ind")),
                                 drop = FALSE]

      # create a readable text for the indicator metric result
      # example: Unexpected location (Percentage (0 to 100)) = 0.00%: Ok
      result_full <-
        dplyr::mutate(result_full,
                      speaking_name_indic_metric = ifelse(
                        !util_empty(get("indicator_metric")),
                        paste0(util_translate_indicator_metrics(
                          result_full[["indicator_metric"]])),
                        NA))

      result_full <-
        dplyr::mutate(result_full,
                      speaking_name_indic_metric = ifelse(
                        !util_empty(result_full[["suffixes"]]),
                        paste0(result_full[["speaking_name_indic_metric"]],
                               " ", result_full[["suffixes"]]),
                        result_full[["speaking_name_indic_metric"]]
                      ))

      result_full <-
        dplyr::mutate(result_full,
                      values_readable = ifelse(
                        !util_empty(result_full[["value"]]),
                        paste0(result_full[["speaking_name_indic_metric"]],
                               " = ", result_full[["value"]],
                               ifelse(
                                 !is.na(result_full[["class"]]),
                                 paste0(": ",
                                        result_full[["class"]]),
                                 ""),
                               " - ", result_full[["title"]]),
                        "No results available"
                      ))


      result_full <- result_full[, -which(names(result_full) ==
                                            c("speaking_name_indic_metric")),
                                 drop = FALSE]

      # define which should be in bold (there is a metric and it has a class)
      result_full <- dplyr::mutate(result_full,
                                   values_readable =
                                     ifelse(get("have_value") &
                                              !util_empty(get("class")),
                                            paste("<strong>",
                                                  get("values_readable"),
                                                  "</strong>"),
                                            get("values_readable")))

       # create a link

      result_full <-
        util_add_links_to_summary_table(result_full,
                                        this,
                                        folder_of_report = ) # TODO: Let util_add_links_to_summary_table handle the folder-of-report and remove the respective code from R/util_render_table_dataquieR_summary.R


      rf <- result_full[, c(LABEL, "call_names",
                            "value", "fg_color", "href", "popup_href", "title", "title_ind"),
                        drop=FALSE]
      rf$orig_label_col <- attr(result_full, "orig_label_col")

      result_full$ln <-
        apply(rf, 1,
              FUN = function(x) {
                if (!is.null(folder_of_report)) { # in case it is not a summary of a single report
                  search_name <- gsub("\\s+", "", paste0(x[["orig_label_col"]], ".", x[["call_names"]]))
                  name_of_r <- unique(folder_of_report[names(folder_of_report)[
                    names(folder_of_report) %in% search_name]])
                  if (!length(name_of_r)) {
                    name_of_r <- ""
                    prefix <- ""
                  } else {
                    prefix <-
                      paste0(name_of_r, "/.report/")
                    name_of_r <- paste0(name_of_r, " -- ")
                  }
                } else { # in case it is a summary of a single report
                  name_of_r <- ""
                  prefix <- ""
                }
                dlg_cap <- paste0(name_of_r, paste0(x[["title"]], x[["title_ind"]]))
                link <- htmltools::a(href = paste0(prefix, x[["href"]]),
                                     title = htmltools::HTML(as.character(
                                       x[["value"]])),
                                       onclick = htmltools::htmlTemplate(text_ = "(function(e) {
                                                        e.preventDefault();
                                                        // showDataquieRResult({{url}}, {{link_url}}, {{title}}); this is triggered by the pre onclick handler, already
                                                      })(event);",
                                                   url =
                                                     jsonlite::toJSON(paste0(prefix, x[["popup_href"]]), auto_unbox = TRUE),
                                                   link_url = jsonlite::toJSON(paste0(prefix, x[["href"]]), auto_unbox = TRUE),
                                                   title = jsonlite::toJSON(dlg_cap, auto_unbox = TRUE)
                                       ),
                                     htmltools::HTML(as.character(x[["value"]])))

                link$attribs$style <- c(link$attribs$style,
                                        "text-decoration:none;display:block;")
                link$attribs$style <- paste0(link$attribs$style,
                                             sprintf("color:%s;",
                                                     x[["fg_color"]]))
                as.character(link)
              })

      result_full <- dplyr::mutate(result_full, ln =
                                     ifelse(util_empty(get("value")),
                                            gsub("(<a.*?[^/>]*>)[^/<]*(.*)",
                                                 "\\1&nbsp;\\2",
                                                 x = get("ln"),
                                                 perl = TRUE),
                                            get("ln")))


      #extract href to a columns and remove href as an attribute in the link
      result_full$href <- gsub('<a.*? *href *= *"([^"]*)".*?>.*',
                               "\\1",
                               x = result_full$ln, perl = TRUE)
      result_full$ln <- gsub('(<a.*?) *href *= *"[^"]*"(.*?>)',
                             "\\1\\2",
                             x = result_full$ln, perl = TRUE)
      result_full$ln <- gsub('(<a.*?) *onclick *= *"[^"]*"(.*?>)',
                             "\\1\\2",
                             x = result_full$ln, perl = TRUE)

      # for all rows is_indicator == FALSE append something to All_err_msg
      result_full <-
        dplyr::mutate(result_full,
                      All_err_msg = ifelse(!get("is_indicator"),
                                           paste0(get("All_err_msg"),
                                                  " ",
                                                  '<span class="dataquieR-error-message"> ',
                                                  get("function_name"),
                                                  " is a descriptor, only.",
                                                  "</span>"),
                                           get("All_err_msg")))


      # create hover addition containing error messages
      result_full <- dplyr::mutate(result_full,
                                   hover = ifelse(!util_empty(get("All_err_msg")),
                                                  paste0("<br>","Messages: ",
                                                         "<br>",
                                                         "\n<ul>\n",
                                                         get("All_err_msg"),
                                                         "\n</ul>\n"),
                                                  paste0("")))


      # initialize the text to display inside the cell - it should be never observed
      result_full$celltext <- paste(result_full$VAR_NAMES,
                                    result_full$call_names, collapse = "/")
      result_full$o <- NA  # Initialize the order
      result_full$f <- htmltools::HTML("&nbsp;") #initialize the filters


      # define order and filter
      result_full$o <- this$order_of[paste(result_full$class_cat)]
      result_full$f <- this$filter_of[paste(result_full$class_cat)]

      #Complete text for the hover box
      result_full$text_complete_hover <- paste0(result_full$hover_text, "<br/>",
                                                result_full$values_readable,
                                                "<br/>",
                                                result_full$hover)
      result_full$text_complete_hover <-
        htmltools::htmlEscape(result_full$text_complete_hover,
                              attribute = TRUE)
      if(is.null(folder_of_report)) {
        #in case of a summary of a single report
        # create the complete href sentence to add after in the html text
        result_full$href_sentence <-
          sprintf('"if (window.hasOwnProperty(&quot;dq_report2&quot;) &amp;&amp; window.dq_report2 &amp;&amp; window.location != &quot;%s&quot;) { if (knowID(&quot;%s&quot;)) { window.location = &quot;%s&quot; } else { window.alert(&quot;No result available&quot;) } }"',
                  result_full$href,
                  result_full$href, result_full$href)

      } else {
        #in case of a summary of a group of reports
        # create the complete href sentence to add after in the html text
        result_full$href_sentence <-
          sprintf('"if (window.hasOwnProperty(&quot;dq_report_by_overview&quot;) &amp;&amp; window.dq_report_by_overview &amp;&amp; window.location != &quot;%s&quot;) { if (knowID(&quot;%s&quot;)) { window.location = &quot;%s&quot; } else { window.alert(&quot;No result available&quot;) } }"',
                  result_full$href,
                  result_full$href, result_full$href)

      }
      rf <- result_full[, c(LABEL, "href_sentence", "class_color",
                            "o", "f",
                            "text_complete_hover", "ln",
                            "call_names",
                            "value", "fg_color", "href", "popup_href", "title", "title_ind"),
                        drop=FALSE]
      rf$orig_label_col <- attr(result_full, "orig_label_col")
      # create a column with the final html text
      try(result_full$html <-
            apply(rf, 1,
                  FUN = function(x) {
                    ln <- x[["ln"]]
                    filter <- x[["f"]]
                    sort_value <- x[["o"]]
                    if (!is.null(folder_of_report)) { # in case it is not a summary of a single report
                      search_name <- gsub("\\s+", "", paste0(x[["orig_label_col"]], ".", x[["call_names"]]))
                      name_of_r <- unique(folder_of_report[names(folder_of_report)[
                        names(folder_of_report) %in% search_name]])
                      if (!length(name_of_r)) {
                        name_of_r <- ""
                        prefix <- ""
                      } else {
                        prefix <-
                          paste0(name_of_r, "/.report/")
                        name_of_r <- paste0(name_of_r, " -- ")
                      }
                    } else { # in case it is a summary of a single report
                      name_of_r <- ""
                      prefix <- ""
                    }
                    dlg_cap <- paste0(name_of_r, paste0(x[["title"]], x[["title_ind"]]))

                    if (!util_empty(x[["value"]])) { # NOT no results available
                      onclick <- paste0("onclick=\"",
                      htmltools::htmlEscape(htmltools::htmlTemplate(text_ = "(function(e) {
                                                        e.preventDefault();
                                                        showDataquieRResult({{url}}, {{link_url}}, {{title}});
                                                      })(event);",
                                                                    url =
                                                                      jsonlite::toJSON(paste0(prefix, x[["popup_href"]]), auto_unbox = TRUE),
                                                                    link_url = jsonlite::toJSON(x[["href"]], auto_unbox = TRUE), # this has a prefix here, already.
                                                                    # TODO: href shouldn't have a prefix, here
                                                                    title = jsonlite::toJSON(dlg_cap, auto_unbox = TRUE)
                      ), attribute = TRUE),
                      "\"")
                      cursor_style <- "cursor: pointer;"
                    } else {
                      onclick <- ""
                      cursor_style <- ""
                    }
                    #  x <- htmltools::htmlEscape(x, attribute = TRUE)
                    text_for_html <- paste0("<pre ",
                                            onclick,
                                            " style=",
                                            sprintf('"height: 100%%; min-height: 2em; margin: 0em; padding: 0em; background: %s; %s text-align: center;"',
                                                    x[["class_color"]], cursor_style),
                                            " sort=\"",
                                            sort_value, "\"",
                                            " filter=\"",
                                            filter,"\"",
                                            " title=\"",
                                            x[["text_complete_hover"]],"\"> ",
                                            ln, "\n</pre>")
                    as.character(text_for_html)
                  }))

      #####==== In case the argument grouped_by is selected ========
      # if not selected the default is to have both grouped_by =  c("call_names", "indicator_metric")
      if(identical(sort(grouped_by), c("call_names", "indicator_metric"))){

        result <- result_full[, -which(names(result_full) %in%
                                         c("value_readable",
                                           "hover_text",
                                           "ln", "href",
                                           "hover",
                                           "text_complete_hover")),
                              drop = FALSE]
        rm(result_full)
      } else if (grouped_by == "call_names" | grouped_by == "indicator_metric") {

        result_full <- result_full[,c("VAR_NAMES","call_names",
                                      "class", "indicator_metric",
                                      "indicator_metric_suff", "suffixes",
                                      "value", "values_raw",
                                      "n_classes", "STUDY_SEGMENT",
                                      "function_name", "class_num",
                                      "All_err_msg", "LABEL",
                                      "class_cat", "class_color",
                                      "values_readable",
                                      "hover_text", "ln", "hover_text_ind",
                                      "href_sentence", "celltext", "o", "f"),
                                   drop = FALSE]

        # ==== Separate the case of call_names and indicator_metric ===

        if(grouped_by == "call_names") {
          #####==== In case the argument grouped_by = "call_names" is selected===

          # decreasing order of VAR_NAMES, call_names and class_num
          result_full <- result_full[order(result_full$VAR_NAMES,
                                           result_full$call_names,
                                           result_full$class_num,
                                           decreasing = TRUE), , drop = FALSE]

          #create a code with VAR_NAMES and call_names
          result_full$code <- paste(result_full$VAR_NAMES,
                                    result_full$call_names)
        } else if (grouped_by == "indicator_metric") {
          #####==== In case the argument grouped_by = "indicator_metric" is selected====

          # decreasing order of VAR_NAMES, call_names and class_num
          result_full <- result_full[order(result_full$VAR_NAMES,
                                           result_full$indicator_metric_suff,
                                           result_full$class_num,
                                           decreasing = TRUE), , drop = FALSE]

          #create a code with VAR_NAMES and indicator_metric
          result_full$code <- paste(result_full$VAR_NAMES,
                                    result_full$indicator_metric_suff)

        }

        #####==== The following part is the same for the 2 cases
        # grouped_by = "indicator_metric" or
        # grouped_by = "call_names"  =========

        # return only first occurrence in the compared vectors, to get only the first row with
        # the worst value
        result_call_name <- result_full[match(unique(result_full$code),
                                              result_full$code), ,drop = FALSE]

        # return rows of result_full that do not have a match in result_call_name
        # If NULL = all variables in common across the 2 dataframe will be used
        result_full <- suppressMessages(dplyr::anti_join(result_full,
                                                         result_call_name,
                                                         by = NULL))

        # create a temporary object tO remove strong tag from values_readable
        List_other_indicators <- result_full[ , c("VAR_NAMES",
                                                  "call_names",
                                                  "code",
                                                  "indicator_metric",
                                                  "indicator_metric_suff",
                                                  "suffixes",
                                                  "class",
                                                  "values_readable"),
                                              drop = FALSE]
        # TO remove strong tag : remove all rows without a result (class = NA)
        List_other_indicators <- dplyr::filter(List_other_indicators,
                                               get("values_readable") !=
                                                 "No results available")

        # TO remove strong tag: keep only a part of the string
        # First create a temporary data frame with the code and the worst class
        class_worst_result <- result_call_name[, c("code", "class"),
                                               drop = FALSE]
        class_worst_result <-
          class_worst_result[!util_empty(class_worst_result$class), ,
                             drop = FALSE]
        colnames(class_worst_result)[colnames(class_worst_result) == "class"] =
          "class_worst_result"
        # Then merge data frames to add a column containing the worst class in the
        # list of other indicators
        List_other_indicators <- merge(List_other_indicators,
                                       class_worst_result,
                                       by = "code",
                                       all.x = TRUE)
        rm(class_worst_result)

        # Finally remove the strong tag only if the class is different from
        # the worst class
        List_other_indicators <-
          dplyr::mutate(List_other_indicators,
                        values_readable =
                          ifelse(class ==
                                   get("class_worst_result") |
                                   is.na(get("class_worst_result")),
                                 get("values_readable"),
                                 gsub('<strong> *(.*)<.*',
                                      "\\1",
                                      x = List_other_indicators$values_readable,
                                      perl = TRUE)))

        # create a row with the content of multiple rows creating a bullet list -
        # if there are other rows
        # first add the tag li to each row in values_readable
        if(nrow(List_other_indicators) != 0) {
          List_other_indicators$values_readable <-
            paste0("<li>",
                   List_other_indicators$values_readable,
                   "</li>")
          # Then i merged the rows by code (VAR_NAMES call_names)
          List_other_indicators <- List_other_indicators %>%
            dplyr::group_by(get("code")) %>%
            dplyr::summarise(values_readable = paste(get("values_readable"),
                                                     collapse = " "))
          # renamed to have the first column "code" and not get("code")
          colnames(List_other_indicators) <- c("code", "values_readable")


          #create a bullet points list
          #      List_other_indicators$values_readable <- paste0("\n<ul>\n",
          #                                                     List_other_indicators$values_readable,
          #                                                     "\n</ul>\n")
          #      # I add now the list of other value readable to be add under the indicator
          # that is in bold and define the color of the cell
          result_call_name <- merge(result_call_name, List_other_indicators,
                                    by = "code", all.x = TRUE)


          # create a column with all value_readable
          result_call_name$values_readable_complete <-
            apply(result_call_name[, names(result_call_name) %in%
                                     c("values_readable.x",
                                       "values_readable.y")],
                  1, function(x) paste(na.omit(x), collapse = "\n"))
          # remove non useful columns
          result_call_name <-
            result_call_name[, -which(names(result_call_name) %in%
                                        c("values_readable.x",
                                          "values_readable.y")),
                             drop = FALSE]
          rm(List_other_indicators)
        }

        #Now create the new complete list of error messages
        List_other_errors <- result_full[ ,
                                          c("VAR_NAMES",
                                            "call_names",
                                            "code",
                                            "indicator_metric",
                                            "indicator_metric_suff",
                                            "suffixes",
                                            "class",
                                            "All_err_msg"), drop = FALSE]
        # remove rows without an error message
        List_other_errors <-
          List_other_errors[!is.na(List_other_errors$All_err_msg), ,
                            drop = FALSE]
        # keep only non-repeated error messages for VAR_NAMES and call_names

        List_other_errors <- List_other_errors[, c("code", "All_err_msg"),
                                               drop = FALSE]
        List_other_errors <- unique(List_other_errors[ , c("code",
                                                           "All_err_msg"),
                                                       drop = FALSE])

        # remove error messages matching the error message of the indicator with the
        # worst result
        # create a temporary dataframe
        errors_in_result_call_name <-
          result_call_name[ , c("code", "All_err_msg"), drop = FALSE]

        # return rows of List_other_errors that do not have a match in errors_in_result_call_name
        List_other_errors <-
          suppressMessages(dplyr::anti_join(List_other_errors,
                                            errors_in_result_call_name,
                                            by = NULL))
        rm(errors_in_result_call_name, result_full)

        # Then merge the rows by code (VAR_NAMES call_names)
        List_other_errors <- List_other_errors %>%
          dplyr::group_by(get("code")) %>%
          dplyr::summarise(All_err_msg = paste(get("All_err_msg"),
                                               collapse = " "))
        # renamed to have the first column "code" and not get("code")
        colnames(List_other_errors)<- c("code", "All_err_msg")

        #Then add the error messages to the dataframe
        result_call_name <- merge(result_call_name, List_other_errors,
                                  by = "code", all.x = TRUE)

        # create a column with all error messages
        result_call_name$All_err_msg_complete <-
          apply(result_call_name[, names(result_call_name) %in%
                                   c("All_err_msg.x", "All_err_msg.y")],
                1, function(x) paste(na.omit(x),
                                     collapse = "\n"))

        # create hover containing the bullet point list of error messages and the word
        # message is there is messages in All_err_msg_complete
        result_call_name <-
          dplyr::mutate(result_call_name,
                        hover =
                          ifelse(!util_empty(get("All_err_msg_complete")),
                                 paste0("<br>","Messages: ", "<br>",
                                        "\n<ul>\n",
                                        get("All_err_msg_complete"),
                                        "\n</ul>\n"),
                                 paste0("")))
        # remove non useful columns
        result_call_name <-
          result_call_name[,-which(names(result_call_name) %in%
                                     c("All_err_msg.x",
                                       "All_err_msg.y")),
                           drop = FALSE]
        rm(List_other_errors)



        # == modify the title of the hover text in case only indicator metric is selected

        if("call_names" %in% grouped_by) {
          # complete text for the hover box
          result_call_name$text_complete_hover <-
            paste0(result_call_name$hover_text, "<br/>",
                   result_call_name$values_readable, "<br/>",
                   result_call_name$hover)
          result_call_name$text_complete_hover <-
            htmltools::htmlEscape(result_call_name$text_complete_hover,
                                  attribute = TRUE)
        }  else if (grouped_by == "indicator_metric" ) {

          # complete text for the hover box
          result_call_name$text_complete_hover <-
            paste0(result_call_name$hover_text_ind, "<br/>",
                   result_call_name$values_readable, "<br/>",
                   result_call_name$hover)
          result_call_name$text_complete_hover <-
            htmltools::htmlEscape(result_call_name$text_complete_hover,
                                  attribute = TRUE)
        }

        # create the new html column
        try(result_call_name$html <-
              apply(
                result_call_name[, c("href_sentence",
                                     "class_color",
                                     "o",
                                     "f",
                                     "text_complete_hover",
                                     "ln"), drop =
                                   FALSE],
                1,
                FUN = function(x) {
                  ln <- x[["ln"]]
                  filter <- x[["f"]]
                  sort_value <- x[["o"]]
                  #  x <- htmltools::htmlEscape(x, attribute = TRUE)
                  text_for_html <-
                    paste0(
                      "<pre ",
#                      "onclick=",
#                      x[["href_sentence"]],
                      " style=",
                      sprintf(
                        '"height: 100%%; min-height: 2em; margin: 0em; padding: 0em; background: %s; text-align: center;"', # cursor: pointer;
                        x[["class_color"]]
                      ),
                      " sort=\"",
                      sort_value,
                      "\"",
                      " filter=\"",
                      filter,
                      "\"",
                      " title=\"",
                      x[["text_complete_hover"]],
                      "\"> ",
                      ln,
                      "\n</pre>"
                    )
                  as.character(text_for_html)
                }
              ))
        result<- result_call_name
      }

      # remove cases without indicator metric
      if("indicator_metric"%in% grouped_by){
        result<- result[!util_empty(result$indicator_metric), , drop = FALSE]
      }

      #rename value as category and html as value
      result1 <- dplyr::rename(result, category = value, value = html)

      # ==== create column of the final table in the case call_names is in the argument
      if("call_names" %in% grouped_by) {

        if (length(grouped_by) > 1) {
          result2 <- result1[, c("VAR_NAMES",
                                 "call_names",
                                 "indicator_metric",
                                 "value"),
                             drop = FALSE]
          result2$cl <- paste0(result2$call_names, "_",
                               result2$indicator_metric)

          result2 <- result2[, c("VAR_NAMES","cl", "value"), drop = FALSE]

          #remove duplicated rows in case present
          result2 <-  dplyr::distinct(result2)

          result2 <- stats::reshape(result2,
                                    idvar = c("VAR_NAMES"),
                                    timevar = "cl",
                                    direction = "wide")

          names(result2)[startsWith(names(result2), "value.")] <-
            substr(names(result2)[startsWith(names(result2),
                                             "value.")],
                   nchar("value.") + 1,
                   nchar(names(result2)[startsWith(names(result2),
                                                   "value.")]))
        } else {
          result2 <- result1[, c("VAR_NAMES",
                                 "call_names",
                                 "indicator_metric",
                                 "value"),
                             drop = FALSE]
          result2 <- dplyr::select(result2, "VAR_NAMES",
                                   paste0(grouped_by),
                                   "value")
          names(result2) <- c("VAR_NAMES", "cl", "value")

          #remove dupliates in case present
          result2 <- dplyr::distinct(result2)

          result2 <- stats::reshape(result2,
                                    idvar = c("VAR_NAMES"),
                                    timevar = "cl",
                                    direction = "wide")

          names(result2)[startsWith(names(result2), "value.")] <-
            substr(names(result2)[startsWith(names(result2),
                                             "value.")],
                   nchar("value.") + 1,
                   nchar(names(result2)[startsWith(names(result2),
                                                   "value.")]))

        }

        # create a dataframe with unique call_names in the first column and
        # indicator_metrics in the second column
        mapping_of_cns <-
          expand.grid(call_names = unique(result$call_names),
                      indicator_metric = unique(na.omit(result$indicator_metric)),
                      stringsAsFactors = FALSE)


        # append a new column with function names
        mapping_of_cns$fname <-
          function_name_mapping[mapping_of_cns[["call_names"]]];
        # create a new column called colname containing the combination
        # selected for grouped-by. Example:
        # if only indicator_metric was selected, the column will contains only indicator_metrics
        if (all(c("call_names", "indicator_metric") %in% grouped_by)) {
          mapping_of_cns$colname <- paste0(mapping_of_cns[["call_names"]], '_',
                                           mapping_of_cns[["indicator_metric"]])
        } else if (all(c("call_names") %in% grouped_by)) {
          mapping_of_cns$colname <- paste0(mapping_of_cns[["call_names"]])
          #     } else if(all(c("indicator_metric") %in% grouped_by)) {
          #       mapping_of_cns$colname <- paste0(mapping_of_cns[["indicator_metric"]])
        } else {
          util_error("this group by is not yet supported")
        }
        # create a list of suffixes for each call_name and function name
        # example for acc_loess_observer, the suffix is observer

        mapping_of_cns$fname2 <-
          apply(mapping_of_cns[, "call_names", drop = FALSE],
                1, FUN = function(x) {
                  util_map_by_largest_prefix(x,
                                             haystack =
                                               util_all_ind_functions())
                })

        suffixes <- mapply(SIMPLIFY = FALSE,
                           cn = mapping_of_cns[["call_names"]],
                           fn = mapping_of_cns$fname2,
                           FUN = function(cn, fn) {
                             if (startsWith(cn, fn)) {
                               substr(cn, nchar(fn) + 1 + 1, nchar(cn))
                               # name + "_" (first +1), start is the next character (second +1)
                             } else {
                               cn
                             }
                           })

        # get a table with for each function, the short-name, implementations and so on
        acronyms_tab <- util_get_concept_info("implementations")

        # === Define the title of the matrix columns =====
        # First select the short name of the function from "implementations" in
        # dq control file
        mapping_of_cns$acronyms <-
          util_map_labels(mapping_of_cns$fname2,
                          acronyms_tab,
                          # to = "dq_report2_short_title",
                          to = "matrix_column_title_report",
                          from = "function_R",
                          ifnotfound = util_abbreviate(mapping_of_cns$fname))
        # Prepare the suffixes
        suffixes <- gsub("_", " ", vapply(suffixes, as.character,
                                          FUN.VALUE = character(1)))
        suffixes[!util_empty(suffixes)] <-
          paste0(":",
                 abbreviate(
                   suffixes[!util_empty(suffixes)],
                   minlength = 3
                 ))

        # Prepare the indicator names
        mapping_of_cns$imtitle <-
          util_translate_indicator_metrics(mapping_of_cns[["indicator_metric"]],
                                           long = FALSE,
                                           short = FALSE)
        mapping_of_cns$imtitle[util_empty(mapping_of_cns$imtitle)] <-
          mapping_of_cns[["indicator_metric"]][util_empty(mapping_of_cns$imtitle)]

        mapping_of_cns$imtitle_short <-
          util_translate_indicator_metrics(mapping_of_cns[["indicator_metric"]],
                                           long = FALSE, short = TRUE)
        mapping_of_cns$imtitle_short[util_empty(mapping_of_cns$imtitle_short)] <-
          mapping_of_cns[["indicator_metric"]][util_empty(mapping_of_cns$imtitle_short)]

        # Create the actual column names for the matrix in column "coltitle"
        if (all(c("call_names", "indicator_metric") %in% grouped_by)) {
          mapping_of_cns$coltitle <- paste0(mapping_of_cns$acronyms, suffixes,
                                            ", ", mapping_of_cns$imtitle_short)
        } else if (all(c("call_names") %in% grouped_by)) {
          mapping_of_cns$coltitle <- paste0(mapping_of_cns$acronyms, suffixes)
          #      } else if(all(c("indicator_metric") %in% grouped_by)) {
          #        mapping_of_cns$coltitle <- paste0(mapping_of_cns$imtitle, suffixes)
        } else {
          util_error("this group by is not yet supported")
        }

        # Definition of the description of the hover text of column titles in the final matrix
        #      if("call_names" %in% grouped_by){
        # in all other cases of grouped_by
        # create description of the hover text of column titles in the final matrix
        mapping_of_cns$coldesc <- vapply(mapping_of_cns$fname,
                                         FUN.VALUE = character(1),
                                         util_col_description)

        # order the column based on the object ordered_call_names
        # First create an order based on call names e.g., 3
        mapping_of_cns$order_call <-
          setNames(seq(length(this$ordered_call_names)),
                   nm = this$ordered_call_names)[
                     mapping_of_cns[["call_names"]]]
        # FIXME: If called with grouped_by = c("call_names"), the columns imtitle and imtitle_short are NA at their second occurrence
        # Then create an order based on indicator metric e.g., 9
        mapping_of_cns$order_slot <- util_order_of_indicator_metrics(mapping_of_cns[["indicator_metric"]]) # was rank(mapping_of_cns[["indicator_metric"]])

        # Then combine the 2 values e.g., 3009
        mapping_of_cns$order <-
          (mapping_of_cns$order_call * 10 **
             ceiling(log(max(mapping_of_cns$order_slot, na.rm = TRUE),
                         base = 10))) + mapping_of_cns$order_slot

        # create an object with the order
        order_of_cols <- c(
          VAR_NAMES = 0,
          setNames(mapping_of_cns$order, nm = mapping_of_cns$colname)
        )

        # check if the columns of result2 match the values in order_of_cols
        util_stop_if_not(all(colnames(result2) %in% names(order_of_cols)))

        # reorder the column of the result based on the order of the object just created
        result2 <-
          result2[, order(order_of_cols[colnames(result2)]), FALSE]


        # create a list containing for each combination of the selected grouped_by
        # the correspondent description
        descs <- setNames(mapping_of_cns$coldesc,
                          nm = mapping_of_cns$colname)[colnames(result2)]
        descs[util_empty(descs)] <- ""

        # rename the columns of result2 with the user friendly name
        colnames_tech <- colnames(result2)
        colnames(result2)[colnames(result2) %in% mapping_of_cns$colname] <-
          setNames(mapping_of_cns$coltitle,
                   nm = mapping_of_cns$colname)[
                     colnames(result2)[colnames(result2) %in%
                                         mapping_of_cns$colname]]

        # creation of the final matrix in case indicator_metric is selected ===
      }    else if (grouped_by == "indicator_metric" ) {
        result2 <- result1[, c("VAR_NAMES",
                               "indicator_metric_suff",
                               "value"), drop = FALSE]

        #remove dupliates in case present
        result2 <- dplyr::distinct(result2)

        result2  <- stats::reshape(result2,
                                   idvar = c("VAR_NAMES"),
                                   timevar = "indicator_metric_suff",
                                   direction = "wide")

        names(result2)[startsWith(names(result2), "value.")] <-
          substr(names(result2)[startsWith(names(result2),
                                           "value.")],
                 nchar("value.") + 1,
                 nchar(names(result2)[startsWith(names(result2),
                                                 "value.")]))

        # create a dataframe with unique call_names in the first column and
        # indicator_metrics in the second column
        mapping_of_cns <-
          expand.grid(call_names = unique(result$call_names),
                      indicator_metric = unique(result$indicator_metric),
                      stringsAsFactors = FALSE)


        # add a new column with the function name
        mapping_of_cns$function_name2 <-
          apply(mapping_of_cns[, "call_names",
                               drop = FALSE],
                1, FUN = function(x) {
                  util_map_by_largest_prefix(x,
                                             haystack =
                                               util_all_ind_functions())
                })

        # add a column with suffixes
        mapping_of_cns <-
          dplyr::mutate(mapping_of_cns,
                        suffixes = ifelse(
                          startsWith(mapping_of_cns[["call_names"]],
                                     mapping_of_cns[["function_name2"]]),
                          substr(mapping_of_cns[["call_names"]],
                                 nchar(mapping_of_cns[["function_name2"]]) +
                                   1 + 1,
                                 nchar(mapping_of_cns[["call_names"]])),
                          # name + "_" (first +1), start is the next character (second +1)
                          mapping_of_cns[["call_names"]]
                        ))

        # create a new column with indicator_metric and suffixes if present
        mapping_of_cns <-
          dplyr::mutate(mapping_of_cns,
                        indicator_metric_suff =
                          ifelse(!util_empty(get("suffixes")),
                                 paste0(get("indicator_metric"),"_",
                                        get("suffixes")),
                                 get("indicator_metric")))

        # append a new column with function names
        mapping_of_cns$fname <-
          function_name_mapping[mapping_of_cns[["call_names"]]];
        # create a new column called colname containing the combination
        # selected for grouped-by. Example:
        # if only indicator_metric was selected, the column will contains only indicator_metrics
        mapping_of_cns$colname <-
          paste0(mapping_of_cns[["indicator_metric"]])
        mapping_of_cns$colname_suff <-
          paste0(mapping_of_cns[["indicator_metric_suff"]])

        # Prepare the indicator names
        mapping_of_cns$imtitle <-
          util_translate_indicator_metrics(mapping_of_cns[["indicator_metric"]],
                                           long = FALSE, short = FALSE)
        mapping_of_cns$imtitle[util_empty(mapping_of_cns$imtitle)] <-
          mapping_of_cns[["indicator_metric"]][util_empty(mapping_of_cns$imtitle)]

         # Create the actual column names for the matrix in column "coltitle"
        mapping_of_cns$coltitle <- paste0(mapping_of_cns$imtitle, " ",
                                          mapping_of_cns$suffixes)

        # Definition of the description of the hover text of column titles in the final matrix
        # remove prefix from indicator name
        mapping_of_cns$indicators_no_prefix <-
          gsub(".*?_(.*)","\\1",
               vapply(mapping_of_cns$colname,
                      as.character,
                      FUN.VALUE = character(1)))
        # import the dqi informations
        dqi_info <- util_get_concept_info("dqi") # order_nr. AbbreviationMetrics$order
        # use the Definition from dqi to create the description of the hover text
        mapping_of_cns$coldesc <-
          util_map_labels(mapping_of_cns$indicators_no_prefix,
                          dqi_info,
                          to = "Definition",
                          from = "abbreviation")


        # order the columns
        # create an order based on indicator metric e.g., 9
        mapping_of_cns$order <- util_order_of_indicator_metrics(mapping_of_cns[["indicator_metric_suff"]]) # was rank(mapping_of_cns[["indicator_metric_suff"]])

        # create an object with the order
        order_of_cols <- c(
          VAR_NAMES = 0,
          setNames(mapping_of_cns$order, nm = mapping_of_cns$colname_suff)
        )

        util_stop_if_not(all(colnames(result2) %in% names(order_of_cols)))

        # reorder the column of the result based on the order of the object just created
        result2 <-
          result2[, order(order_of_cols[colnames(result2)]), FALSE]

        # create a list containing for each combination of the selected grouped_by
        # the correspondent description
        descs <- setNames(mapping_of_cns$coldesc,
                          nm = mapping_of_cns$colname_suff)[colnames(result2)]
        descs[util_empty(descs)] <- ""

        # rename the columns of result2 with the user friendly name
        colnames_tech <- colnames(result2)
        colnames(result2)[colnames(result2) %in% mapping_of_cns$colname_suff] <-
          setNames(mapping_of_cns$coltitle,
                   nm = mapping_of_cns$colname_suff)[
                     colnames(result2)[colnames(result2) %in%
                                         mapping_of_cns$colname_suff]]

      }

      # replace var_names with labels in the first column of the matrix
      if (VAR_NAMES %in% colnames(result2)) {
        Variables <- util_map_labels(result2[[VAR_NAMES]],
                                     meta_data = meta_data,
                                     to = this$label_col,
                                     from = VAR_NAMES)
        # label_col <- attr(Variables, "label_col")
        result2[[VAR_NAMES]] <- Variables
        colnames(result2)[colnames(result2) == VAR_NAMES] <- "Variables"
        label_col <- this$label_col
      } else {
        label_col <- VAR_NAMES
      }

      # In case of summary for dq_report_by
      # Create the href for each cell "Total"
      if (!is.null(folder_of_report) ) {

        # In case there is more than one strata, and the name of the variables
        # are modified because repeated in several folders
        if (!is.null(var_uniquenames)) {
          # Prepare vars_folder_of_report: a named vector containing the new name
          # of the variable (e.g., "sd1-SEX_0_1-ITEM_5_0"). The names of this
          # vector contains the name of the folder of the report
          vars_folder_of_report <- setNames(names(folder_of_report),
                                            folder_of_report)
          vars_folder_of_report <- gsub("^(.*)\\.(.*)" , "\\1",
                                        vars_folder_of_report)
          vars_folder_of_report <-
            vars_folder_of_report[!duplicated(vars_folder_of_report)]
          #Prepare a data frame with the different names of the same variable
          var_uniquenames <- do.call("rbind", var_uniquenames)
          rownames(var_uniquenames) <- NULL

          #Prepare a named vector of the variables (e.g., SEX_0) and the folder
          name_and_folder <- vars_folder_of_report
          simple_var_names <-
            var_uniquenames[var_uniquenames$new_names_with_label %in%
                              name_and_folder, , drop = FALSE]
          name_and_folder <- setNames(simple_var_names$name_label,
                                     nm = names(name_and_folder))

          # ix the names of this vector so that include the complete path to a
          # variable inside the folder
          names(vars_folder_of_report) <-
            mapply(x = prep_link_escape(name_and_folder),
                   y = names(vars_folder_of_report),
                   SIMPLIFY = FALSE,
                   FUN = function(x, y){
                     z <- paste0(y, "/.report/VAR_", x, ".html#", x)
                     return(z)
                   })

          #rename content of vars_folder_of_report, so that now contain
          # the new name e.g.,  "sd1-SEX_0_0-v00000" instead of
          # previous "sd1-SEX_0_0-CENTER_0" (to match content of this$rowmaxes)
          var_uniquenames_fil <-
            var_uniquenames[var_uniquenames$new_names_with_label %in%
                              vars_folder_of_report, , drop = FALSE]
          vars_folder_of_report <-
            setNames(var_uniquenames_fil$new_names_with_varnames,
                     nm = names(vars_folder_of_report))

          # make unique the names a created named vector containing
          # this$rowmaxes$cell_text. The names comes now from the
          # column this$rowmaxes$VAR_NAMES, because
          # the names of this$rowmaxes$cell_text contain repeated variable names
          new_rowmaxes_cell_text <- this$rowmaxes$cell_text
          names(new_rowmaxes_cell_text) <- this$rowmaxes$VAR_NAMES

          # replace the href in rowmaxes used for links from var_names and total cells
          this$rowmaxes$cell_text <-
            mapply(needle = setNames(nm = names(new_rowmaxes_cell_text)),
                   MoreArgs = list(
                     haystack = new_rowmaxes_cell_text,
                     corresponding_value_for_needle = vars_folder_of_report),
                   SIMPLIFY = FALSE,
                   FUN = function(needle, corresponding_value_for_needle,
                                  haystack){
                     haystack <- haystack[[as.character(needle)]]
                     haystack <-
                       gsub('(?=VAR_)(.*)(?="\\sstyle)',
                            names(corresponding_value_for_needle[
                              corresponding_value_for_needle %in% needle]),
                            haystack,
                            perl = TRUE)
                     return(haystack)
                   })
        } else {
          #Create a named vector vars_folder_of_report containing the variable
          # name (e.g., "SEX_0"). The names are the folder of the variable
          vars_folder_of_report <- setNames(names(folder_of_report),
                                            folder_of_report)
          vars_folder_of_report <- gsub("^(.*)\\.(.*)" , "\\1",
                                        vars_folder_of_report)
          vars_folder_of_report <-
            vars_folder_of_report[!duplicated(vars_folder_of_report)]

          names(vars_folder_of_report) <-
            mapply(variable_nm = prep_link_escape(vars_folder_of_report),
                   folder_for_href = names(vars_folder_of_report),
                   SIMPLIFY = FALSE,
                   FUN = function(variable_nm, folder_for_href){
                     z <- paste0(folder_for_href, "/.report/VAR_",
                                 variable_nm, ".html#", variable_nm)
                     return(z)
                   })

          # replace the href in rowmaxes used for links from var_names and Total cells
          this$rowmaxes$cell_text <-
            mapply(x = setNames(nm = names(this$rowmaxes$cell_text)),
                   MoreArgs = list(
                     z = this$rowmaxes$cell_text,
                     y = vars_folder_of_report),
                   SIMPLIFY = FALSE,
                   FUN = function(x, y, z){
                     z <- z[[as.character(x)]]

                     z <- gsub('(?=VAR_)(.*)(?="\\sstyle)',
                               names(y[y %in% x]),
                               z,
                               perl = TRUE)
                     return(z)
                   })
        }
      }

      this$rowmaxes$VAR_NAMES_mapped <- prep_map_labels(this$rowmaxes$VAR_NAMES,
                                                        meta_data = meta_data,
                                                        to = label_col)

      ordered_var_names_mapped <- prep_map_labels(ordered_var_names,
                                                  meta_data = meta_data,
                                                  to = label_col)

      result2 <- result2[util_order_by_order(result2$Variables,
                                             ordered_var_names_mapped),
                         , drop = FALSE]

      result2$Total <- setNames(this$rowmaxes$cell_text,
                                nm = this$rowmaxes$VAR_NAMES_mapped)[result2$Variables]

      descs <- c(descs, "Total DQ for Items")


      colnames_speaking_from_tech <- setNames(colnames(result2),
                                              nm = colnames_tech)

      #In case of summary for dq_report_by
      if (!is.null(folder_of_report)) {
        # if more than 1 strata is present
        if (!is.null(var_uniquenames)) {
          # create a named vector with as variable name the new name that
          # will appear on the table (it has extra spaces)
          # (e.g.,  "sd1-SEX_0_1 - QUEST_DT_0" instead of
          # "sd1-SEX_0_1-QUEST_DT_0") and as name
          # the folder containing the variable
          new_var_name_in_table <- vars_folder_of_report
          simple_var_names <-
            var_uniquenames[var_uniquenames$new_names_with_varnames %in%
                              new_var_name_in_table, , drop = FALSE]

          new_var_name_in_table <-
            setNames(simple_var_names$result2_Variables_match,
                     nm = names(new_var_name_in_table))

          # create a new named vector href with the correct folder (as name) and
          # the new variable name to read in the table (with extra white spaces)
          href <-
            setNames(names(new_var_name_in_table[new_var_name_in_table %in%
                                                   result2$Variables]),
                     nm = new_var_name_in_table[new_var_name_in_table %in%
                                                  result2$Variables])
          filter <- setNames(result2$Variables, nm = result2$Variables)
          data <- setNames(result2$Variables, nm = result2$Variables)

        } else {
          # define the href, a named vector containing the folder of the variable.
          # The names contain the variable name (e.g., "SEX_0")
          href <-
            setNames(names(vars_folder_of_report[vars_folder_of_report %in%
                                                   result2$Variables]),
                     nm = vars_folder_of_report[vars_folder_of_report %in%
                                                  result2$Variables])
          filter <- setNames(result2$Variables, nm = result2$Variables)
          data <- setNames(result2$Variables, nm = result2$Variables)
        }

        result2$Variables <-
          mapply(x = setNames(nm = result2$Variables),
                 MoreArgs = (list(
                   href = #sprintf('javascript:console.log("%s")', href),
                     href,
                   filter = filter,
                   data =  data
                 )), SIMPLIFY = FALSE,
                 FUN = function(x, href, filter, data){
                   text_for_html <-
                     paste0("<pre ",
                            'onclick=\"\"',
                            ' style=',
                            '"height: 100%; min-height: 2em; font-size: 16px; font-family: serif; font-style: normal; margin: 0em; color: #0066CC; padding: 0em; background: #ffffff; cursor: pointer; text-align: right;"',
                            " filter=\"", filter[x],"\">\n",
                            "<a href=\"", href[names(href) %in% x], '"',
                            'style=\"text-decoration:none;display:block;',
                            "\"> ",
                            data[x], "</a>\n</pre>")
                   as.character(text_for_html)
                   return(text_for_html)
                 })

        tb <-
          util_suppress_warnings(util_html_table(
            result2,
            filter = "top", options = list(scrollCollapse = TRUE,
                                           scrollY = "75vh"),
            is_matrix_table = TRUE, rotate_headers =
              (ncol(result2) > 5),
            link_variables = FALSE,
            cols_are_indicatormetrics = !TRUE,
            colnames_aliases2acronyms = FALSE,
            descs = descs,
            meta_data = meta_data,
            label_col = label_col,
            dl_fn = prep_link_escape(paste0("Summary-",
                                            this[["title"]],
                                            "-",
                                            this[["subtitle"]]),
                                     html = TRUE),
            title = this[["title"]],
            messageTop = this[["subtitle"]],
            messageBottom = sprintf("generated by %s %s",
                                    paste(packageName()),
                                    paste(packageVersion(
                                      packageName()))
            ),
            col_tags = list(`All columns` = colnames(result2),
                            `Integrity` = c("Variables",
                                            colnames_speaking_from_tech[grep(
                                              "^int_", value = TRUE,
                                              colnames_tech)], "Total"),
                            `Completeness` = c("Variables",
                                               colnames_speaking_from_tech[grep(
                                                 "^com_", value = TRUE,
                                                 colnames_tech)], "Total"),
                            `Consistency` = c("Variables",
                                              colnames_speaking_from_tech[grep(
                                                "^con_", value = TRUE,
                                                colnames_tech)], "Total"),
                            `Accuracy` = c("Variables",
                                           colnames_speaking_from_tech[grep(
                                             "^acc_", value = TRUE,
                                             colnames_tech)], "Total")
            )
          ), classes = LONG_LABEL_EXCEPTION)
      } else {
        tb <- util_suppress_warnings(
          util_html_table(
            result2,
            filter = "top", options = list(scrollCollapse = TRUE,
                                           scrollY = "75vh"),
            is_matrix_table = TRUE, rotate_headers =
              (ncol(result2) > 5),
            cols_are_indicatormetrics = !TRUE,
            colnames_aliases2acronyms = FALSE,
            descs = descs,
            meta_data = meta_data,
            label_col = label_col,
            dl_fn = prep_link_escape(paste0("Summary-",
                                            this[["title"]],
                                            "-",
                                            this[["subtitle"]]),
                                     html = TRUE),
            title = this[["title"]],
            messageTop = this[["subtitle"]],
            messageBottom = sprintf("generated by %s %s",
                                    paste(packageName()),
                                    paste(packageVersion(
                                      packageName()))
            ),
            col_tags = list(`All columns` = colnames(result2),
                            `Integrity` = c("Variables",
                                            colnames_speaking_from_tech[grep(
                                              "^int_", value = TRUE,
                                              colnames_tech)], "Total"),
                            `Completeness` = c("Variables",
                                               colnames_speaking_from_tech[grep(
                                                 "^com_", value = TRUE,
                                                 colnames_tech)], "Total"),
                            `Consistency` = c("Variables",
                                              colnames_speaking_from_tech[grep(
                                                "^con_", value = TRUE,
                                                colnames_tech)], "Total"),
                            `Accuracy` = c("Variables",
                                           colnames_speaking_from_tech[grep(
                                             "^acc_", value = TRUE,
                                             colnames_tech)], "Total")
            )
          ), classes = LONG_LABEL_EXCEPTION)
      }

      jqui <- rmarkdown::html_dependency_jqueryui()
      jqui$stylesheet <- "jquery-ui.min.css"

      res <- htmltools::browsable(htmltools::tagList(
        rmarkdown::html_dependency_jquery(),
        jqui,
        htmltools::htmlDependency(
          name = "tippy",
          version = "6.7.3",
          src = system.file("tippy", package = "dataquieR"),
          script = c("core.js", "tippy.js")
        ),
        htmltools::htmlDependency(
          name = "clipboard",
          version = "2.0.11",
          src = system.file("clipboard", package = "dataquieR"),
          script = c("clipboard.min.js")
        ),
        html_dependency_dataquieR(),
        tb,
        htmltools::tags$script(paste0('$(function() {',
                                      'if (!window.hasOwnProperty("dq_report2") || !window.dq_report2) {',
                                      'if (!window.hasOwnProperty("dq_report_by_overview") || !window.dq_report_by_overview) {',
                                      ' $("a").attr("href", "javascript:alert(\\"links work in dq_report2 reports, only.\\")")',
                                      '} else {' ,
                                      ' ',
                                      '}',
                                      '}} )'))
      ))
    }

    util_attach_attr(res,
                     repsum_wide = util_attach_attr(result2,
                                                    label_col = label_col),
                     this = this)
  })}

#' Function that calculated height and width values for `script_iframe`
#'
#' @param sizing_hints [list] containing information for setting
#'                           the size of the `iframe`
#'
#' @return a list with figure_type_id, w, and h; sizes are as `CSS`, existing
#'         elements are kept, `w_in_cm` and `h_in_cm` are estimates for the
#'         size in centimeters on a typical computer display (in 2024)
#' @noRd
util_finalize_sizing_hints <- function(sizing_hints) {
  original_sizing_hint <- sizing_hints
  if (!is.list(original_sizing_hint)) {
    original_sizing_hint <- list()
  }
  sizing_hints <- try(.util_finalize_sizing_hints(sizing_hints = sizing_hints))
  if (util_is_try_error(sizing_hints)) {
    sizing_hints <- list()
  }
  .w <- 14
  .uw <- "cm"
  .h <- 10
  .uh <- "cm"

  if (is.null(sizing_hints$w)) {
    sizing_hints$w <- sprintf("%s%s", as.character(.w), .uw)
  }

  if (is.null(sizing_hints$h)) {
    sizing_hints$h <- sprintf("%s%s", as.character(.h), .uh)
  }

  normalize_unit <- function(css_value) {
    ..v <- css_value
    is_sreen_rel <- FALSE
    if (all(grepl("^\\s*calc\\s*\\(.+\\)\\s*;?\\s*$", ..v))) {
      ..v <- gsub("^\\s*calc\\s*\\(", "", ..v)
      ..v<- gsub("\\)\\s*;?\\s*$", "", ..v)
      # use typical screen size
      ..v_old <- ..v
      ..v <- gsub("var\\(\\s*--screen-width\\s*\\)", "1920", ..v)
      ..v <- gsub("var\\(\\s*--screen-height\\s*\\)", "1080", ..v)
      if (..v != ..v_old) {
        is_sreen_rel <- TRUE
      }
    }
    ..v <- gsub(";\\s*$", "", ..v)
    .units <- c("cm", "mm", "in", "px", "pt", "pc", "em", "ex", "ch",
                "rem", "vw", "vh", "vmin", "vmax", "%")
    for (.unit in .units) {
      # 1px = 1/96th of 1in
      pattern <- paste0("((?:\\d*\\.)?\\d+)(", .unit, ")")
      ready <- FALSE
      while (!ready) {
        matches <- gregexpr(pattern, ..v)
        ready <- length(matches) == 0 || matches[[1]] == -1
        if (!ready) {
          e <- matches[[1]]
          new_val <- substr(..v, e, e+attr(e, "match.length")-1)
          match <- regexec(pattern, new_val)[[1]]
          from_number <- match[2]
          to_number <- from_number + attr(match, "match.length")[2]
          from_unit <- match[3]
          to_unit <- from_unit + attr(match, "match.length")[3]
          number <- as.numeric(substr(
            new_val, from_number, to_number - 1))
          unit <- .unit
          if (unit == "px") { # assume pixels according to https://www.w3schools.com/cssref/css_units.php
            if (!is_sreen_rel)
              util_message(c("Internal error, sorry: dataquieR",
                             "developers should never specify sizes in",
                             "pixels, because the size depends on the",
                             "output resolution, then. Please report."))
            unit <- "in"
            number <- number / 96
          } else if (unit == "pt") {
            unit <- "in"
            number <- number / 72
          } else if (unit == "pc") {
            unit <- "in"
            number <- number * 12 / 72
          } else if (unit == "em") { # relative units from here
            unit <- "in"
            number <- number * 12 / 72 # assuming a font size of 12pt
          } else if (unit == "ex") {
            unit <- "in"
            number <- number * 12 / 72 # assuming a font size of 12pt
          } else if (unit == "ch") {
            unit <- "in"
            number <- number * 12 / 72 # assuming a font size of 12pt
          } else if (unit == "rem") {
            unit <- "in"
            number <- number * 12 / 72 # assuming a font size of 12pt
          } else if (unit == "vw") {
            unit <- "in"
            number <- number * 1920 / 100
          } else if (unit == "vh") {
            unit <- "in"
            number <- number * 1080 / 100
          } else if (unit == "vmin") {
            unit <- "in"
            number <- number * 1080 / 100
          } else if (unit == "vmax") {
            unit <- "in"
            number <- number * 1920 / 100
          } else if (unit == "%") {
            unit <- "in"
            number <- number * 1920 / 100
          }

          number <- units::set_units(number, unit, mode = "standard")
          cm <- "dummy" # avoid notes about cm
          units(number) <- units::make_units(cm)
          ..v <- paste0(
            substring(..v, 1, e-1),
            units::drop_units(number),
            substring(..v, e + attr(e, "match.length"), 99999)
          )
        }
      }
    }
    if (length(..v) != 1 ||
        !grepl("^[0-9\\.\\*\\+\\-\\/ ]+$", ..v, perl = TRUE)) {
      util_error(
        c(
          "Internal error in util_finalize_sizing_hints, parsing final sizing",
          "hints, sorry. Please report: ",
          "Could not parse unit: %s"
        ),
        dQuote(css_value)
      )
    } else {
      ..v <- try(eval(parse(text = ..v), envir = baseenv()),
                 silent = TRUE)
      if (util_is_try_error(..v)) {
        util_error(
          c(
            "Internal error in util_finalize_sizing_hints, parsing final sizing",
            "hints, sorry. Please report: ",
            "Could not compute unit: %s"
          ),
          dQuote(css_value)
        )

      }
    }
    ..v
  }
  # "calc(var(--screen-height) * 0.65 * 1px);"
  # css does not allow space between number and unit; allowed units according to https://www.w3schools.com/cssref/css_units.php
  # cm, mm, in, px, pt, pc // relative: em, ex, ch, rem, vw, vh, vmin, vmax, %
  # relative, unsupported by the units package: px, pt, pc, em, ex, ch, rem, vw, vh vmin, vmax, %

  sizing_hints$w_in_cm <- normalize_unit(sizing_hints$w)
  sizing_hints$h_in_cm <- normalize_unit(sizing_hints$h)

  new_hints <- setdiff(
    names(original_sizing_hint),
    names(sizing_hints))
  sizing_hints[new_hints] <-
    original_sizing_hint[new_hints]

  sizing_hints
}

.util_finalize_sizing_hints<- function(sizing_hints) {

  ##CASE 1: bar_chart
  if (!is.null(sizing_hints$figure_type_id) &&
      sizing_hints$figure_type_id == "bar_chart") {

    #In case of missing hints, give an error
    if (is.null(sizing_hints$number_of_bars) ||
        is.null(sizing_hints$range)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }

    # BAR CHART CASE1: it is rotated (bars are horizontal)
    if (!is.null(sizing_hints$rotated) &&
        sizing_hints$rotated == TRUE) {
      if (!is.null(sizing_hints$number_of_bars) &&
          sizing_hints$number_of_bars < 15) {
        h <- 300
      } else if (!is.null(sizing_hints$number_of_bars) &&
                 sizing_hints$number_of_bars >= 15 &&
                 sizing_hints$number_of_bars < 101) {
        h <-  sizing_hints$number_of_bars * 22
      } else {
        h <- "calc(var(--screen-height) * 0.5 * 1px);"
      }
      w <- 600

      # final height in cm
      if(is.numeric(h)) {
        h <- round(h * 0.0264583333, digits = 4)
        h <- paste0(h, "cm;")
      }

      if (length(sizing_hints$no_char_numbers) != 1 ||
          is.na(sizing_hints$no_char_numbers) ||
          sizing_hints$no_char_numbers < 4) {
        sizing_hints$no_char_numbers <- 4
      }
      if (length(sizing_hints$no_char_vars) != 1 ||
          is.na(sizing_hints$no_char_vars) ||
          sizing_hints$no_char_vars < 4) {
        sizing_hints$no_char_vars <- 4
      }

      #final width in cm
      if(is.numeric(w)) {
        w <- w + 16 * sizing_hints$no_char_vars
        if(w > 1200) {
          w <- "calc(var(--screen-width) * 0.9 * 1px);"
        } else {
          w <- round(w * 0.0264583333, digits = 4)
          w <- paste0(w, "cm;")
        }
      }
    } else {
      # BAR CHART CASE2: it is NOT rotated (bars are vertical)
      if (!is.null(sizing_hints$number_of_bars) &&
          sizing_hints$number_of_bars < 16) {
        w <- 500
      } else if (!is.null(sizing_hints$number_of_bars) &&
                 sizing_hints$number_of_bars > 15 &&
                 sizing_hints$number_of_bars < 31) {
        w <- 800
      } else {
        w <- "calc(var(--screen-width) * 0.6* 1px);"
      }
      if (sizing_hints$range < 301 ||
          sizing_hints$number_of_bars < 21) {
        h <- 350
      } else if (sizing_hints$range > 300 &&
                 sizing_hints$range < 501) {
        h <- 450
      } else {
        h <- "calc(var(--screen-height) * 0.5 * 1px);"
      }

      if(length(sizing_hints$no_char_x) != 1 || sizing_hints$no_char_x < 4) {
        sizing_hints$no_char_x <- 4
      }
      if(length(sizing_hints$no_char_y) != 1 || sizing_hints$no_char_y < 4) {
        sizing_hints$no_char_y <- 4
      }
      # final height in cm
      if(is.numeric(h)) {
        h <- h + 16 * sizing_hints$no_char_y
        h <- round(h * 0.0264583333, digits = 4)
        h <- paste0(h, "cm;")
      }
      # final width in cm
      if(is.numeric(w)) {
        w <- w + 16 * sizing_hints$no_char_x
        w <- round(w * 0.0264583333, digits = 4)
        w <- paste0(w, "cm;")
      }
    }
    ##CASE 2: dot_mat
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "dot_mat") {
    #In case of missing hints, give an error
    if (is.null(sizing_hints$number_of_vars) ||
        is.null(sizing_hints$number_of_cat)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }
    # DOT_MAT CASE1: it is rotated
    # (vars on the y axis, and categories of
    # missing or levels of groups are on the x axis)
    if (!is.null(sizing_hints$rotated) &&
        sizing_hints$rotated == TRUE) {
      if (sizing_hints$number_of_vars < 21) {
        h <- 250
      } else {
        h <- sizing_hints$number_of_vars * 12
      }

      if (sizing_hints$number_of_cat < 7) {
        w <- 300
      } else if (sizing_hints$number_of_cat > 6 &&
                 sizing_hints$number_of_cat < 16) {
        w <-  sizing_hints$number_of_cat * 45
      } else {
        w <- sizing_hints$number_of_cat * 35
      }

      if(sizing_hints$no_char_vars < 8) {
        sizing_hints$no_char_vars <- 8
      }
      if(sizing_hints$no_char_cat < 4) {
        sizing_hints$no_char_cat <- 4
      }
      # final height in cm
      if(is.numeric(h)) {
        h <- h + 16 * sizing_hints$no_char_cat
        h <- round(h * 0.0264583333, digits = 4)
        h <- paste0(h, "cm;")
      }
      # final width in cm
      if(is.numeric(w)) {
        w <- w + 16 * sizing_hints$no_char_vars
        w <- round(w * 0.0264583333, digits = 4)
        w <- paste0(w, "cm;")
      }
    } else {
      # DOT_MAT CASE2: it is NOT rotated
      # (vars on the x axis, and categories of
      # missing or levels of groups are on the y axis)
      if (sizing_hints$number_of_cat < 4) {
        h <- 150
      } else if (sizing_hints$number_of_cat >3 &&
                 sizing_hints$number_of_cat < 15) {
        h <- 250
      } else {
        h <- 100 + sizing_hints$number_of_cat * 16
      }
      if (sizing_hints$number_of_vars < 7) {
        w <- 300
      } else if (6 < sizing_hints$number_of_vars &&
                 sizing_hints$number_of_vars < 15) {
        w <-  sizing_hints$number_of_vars * 48
      } else {
        w <- sizing_hints$number_of_vars * 35
      }

      if(sizing_hints$no_char_vars < 4) {
        sizing_hints$no_char_vars <- 4
      }
      if(sizing_hints$no_char_cat < 4) {
        sizing_hints$no_char_cat <- 4
      }
      # final height in cm
      if(is.numeric(h)) {
        h <- h + 16 * sizing_hints$no_char_vars
        h <- round(h * 0.0264583333, digits = 4)
        h <- paste0(h, "cm;")
      }
      # final width in cm
      if(is.numeric(w)) {
        w <- w + 16 * sizing_hints$no_char_cat
        w <- round(w * 0.0264583333, digits = 4)
        w <- paste0(w, "cm;")
      }
    }
    ##CASE 2b: dot_mat DATATYPE matrix with fix size
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "dot_mat_fix") {
    #In case of missing hints, give an error
    if (is.null(sizing_hints$number_of_vars) ||
        is.null(sizing_hints$number_of_cat)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }
    # DOT_MAT can not be rotated
    # DOT_MAT CASE2: it is NOT rotated
    # (vars on the x axis, and categories of
    # missing or levels of groups are on the y axis)
    h <- 25
    if (sizing_hints$number_of_vars < 7) {
      w <- 300
    } else if (6 < sizing_hints$number_of_vars &&
               sizing_hints$number_of_vars < 15) {
      w <-  sizing_hints$number_of_vars * 48
    } else {
      w <- sizing_hints$number_of_vars * 35
    }

    if(sizing_hints$no_char_vars < 4) {
      sizing_hints$no_char_vars <- 4
    }
    if(sizing_hints$no_char_cat < 4) {
      sizing_hints$no_char_cat <- 4
    }
    # final height in cm
    if(is.numeric(h)) {
      h <- h + 16 * sizing_hints$no_char_vars
      h <- round(h * 0.0264583333, digits = 4)
      h <- paste0(h, "cm;")
    }
    # final width in cm
    if(is.numeric(w)) {
      w <- w + 16 * sizing_hints$no_char_cat
      w <- round(w * 0.0264583333, digits = 4)
      w <- paste0(w, "cm;")
    }
    ##CASE 3: scatt_plot, only used now for univariate outliers
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "scatt_plot") {
    #In case of missing hints, give an error
    if (is.null(sizing_hints$range) ||
        is.null(sizing_hints$number_of_vars)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }
    if (sizing_hints$range < 20) {
      h <- 200
    } else if (sizing_hints$range > 19 &&
               sizing_hints$range < 101) {
      h <- 350
    } else if (sizing_hints$range > 100 &&
               sizing_hints$range < 251) {
      h <- 400
    } else {
      h <- 600
    }
    if (sizing_hints$number_of_vars == 1) {
      w <- 750
    } else {
      # should be never reached as univariate outliers always have 1 var
      w <- "calc(var(--screen-width) * 0.5 * 1px);"
    }

    if(sizing_hints$no_char_y < 4) {
      sizing_hints$no_char_y <- 4
    }
    # final height in cm
    if(is.numeric(h)) {
      h <- round(h * 0.0264583333, digits = 4)
      h <- paste0(h, "cm;")
    }
    # final width in cm
    if(is.numeric(w)) {
      w <- w + 16 * sizing_hints$no_char_y
      w <- round(w * 0.0264583333, digits = 4)
      w <- paste0(w, "cm;")
    }
    ## CASE 3 bis: mahalanobis_plot
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "mahalanobis_plot") {
    #In case of missing hints, give an error
    if (is.null(sizing_hints$range_x) ||
        is.null(sizing_hints$range_y)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }
    if (sizing_hints$range_y < 20) {
      h <- 200
    } else if (sizing_hints$range_y > 19 &&
               sizing_hints$range_y < 101) {
      h <- 350
    } else if (sizing_hints$range_y > 100 &&
               sizing_hints$range_y < 251) {
      h <- 400
    } else {
      h <- 600
    }
    if (sizing_hints$range_x < 201) {
      w <- 400
    } else if (sizing_hints$range_y > 200 &&
               sizing_hints$range_y < 401) {
      w <- 600
    } else {
      w <- "calc(var(--screen-width) * 0.5 * 1px);"
    }

    # final height in cm
    if(is.numeric(h)) {
      h <- round(h * 0.0264583333, digits = 4)
      h <- paste0(h, "cm;")
    }
    # final width in cm
    if(is.numeric(w)) {
      w <- round(w * 0.0264583333, digits = 4)
      w <- paste0(w, "cm;")
      ##CASE 4: bar_limit
    }
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "bar_limit") {
    #In case of missing hints, give an error
    if (is.null(sizing_hints$range) ||
        is.null(sizing_hints$no_bars_in_all_w)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }
    # here rotated would be always FALSE, as the plots are always with vertical
    # bars
    if (sizing_hints$no_bars_in_all_w < 31) {
      w <- 500
    } else if (sizing_hints$no_bars_in_all_w > 30 &&
               sizing_hints$no_bars_in_all_w < 41) {
      w <- sizing_hints$no_bars_in_all_w * 20
    } else {
      w <- "calc(var(--screen-width) * 0.6 * 1px);"
    }
    if (sizing_hints$range < 301 ||
        sizing_hints$no_bars_in_all_w < 21) {
      h <- 350
    } else if (sizing_hints$range > 300 &&
               sizing_hints$range < 501) {
      h <- 450
    } else {
      h <- "calc(var(--screen-height) * 0.5 * 1px);"
    }

    if(sizing_hints$no_char_y < 4) {
      sizing_hints$no_char_y <- 4
    }
    # final height in cm
    if(is.numeric(h)) {
      h <- round(h * 0.0264583333, digits = 4)
      h <- paste0(h, "cm;")
    }
    # final width in cm
    if(is.numeric(w)) {
      w <- w + 16 * sizing_hints$no_char_y
      w <- round(w * 0.0264583333, digits = 4)
      w <- paste0(w, "cm;")
    }
    ##CASE 5: dot_loess
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "dot_loess") {
    #In case of missing hints, give an error
    if (is.null(sizing_hints$n_groups)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }
    # always with time on the x axis
    if (sizing_hints$n_groups < 21) {
      h <- 400
      w <- "calc(var(--screen-width) * 0.5 * 1px);"
    } else if (sizing_hints$n_groups > 20) {
      h <- "calc(var(--screen-height) * 0.65 * 1px);"
      w <- "calc(var(--screen-width) * 0.9 * 1px);"
    }

    if(sizing_hints$no_char_y < 4) {
      sizing_hints$no_char_y <- 4
    }
    # final height in cm
    if(is.numeric(h)) {
      h <- round(h * 0.0264583333, digits = 4)
      h <- paste0(h, "cm;")
    }
    # final width in cm
    if(is.numeric(w)) {
      w <- w + 16 * sizing_hints$no_char_y
      w <- round(w * 0.0264583333, digits = 4)
      w <- paste0(w, "cm;")
    }
    ##CASE 6: marg_plot
    # FIXME: @Elena, please handle the case, where the plot does not feature an overall distribution plot: (length(res$SummaryPlot) == 1)
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "marg_plot") {
    #In case of missing hints, give an error
    if (is.null(sizing_hints$n_groups) ||
        is.null(sizing_hints$type_plot)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }
    # there are 3 options: count_plot that always vary from 0 to 1,
    # violin_plot, and rotated_plot that include 2 different plot with the axis inverted
    if (sizing_hints$type_plot == "count_plot") {
      # MARG_PLOT CASE1: COUNT_PLOT
      h <- 400
      if (sizing_hints$n_groups < 11) {
        w <- 400
      } else if (sizing_hints$n_groups >10 &&
                 sizing_hints$n_groups < 21) {
        w <-  600
      } else {
        w <- sizing_hints$n_groups * 30
      }
      if(sizing_hints$no_char_y < 4) {
        sizing_hints$no_char_y <- 4
      }
      # final height in cm
      if(is.numeric(h)) {
        h <- h + 16 * sizing_hints$no_char_x
        h <- round(h * 0.0264583333, digits = 4)
        h <- paste0(h, "cm;")
      }
      # final width in cm
      if(is.numeric(w)) {
        w <- w + 16 * sizing_hints$no_char_y
        w <- round(w * 0.0264583333, digits = 4)
        w <- paste0(w, "cm;")
      }
    } else if (sizing_hints$type_plot == "violin_plot") {
      # MARG_PLOT CASE2: VIOLIN_PLOT
      h <- 400
      if (sizing_hints$n_groups < 7) {
        w <- 600
      } else if (sizing_hints$n_groups > 6 &&
                 sizing_hints$n_groups < 21) {
        w <- 800
      } else {
        w = sizing_hints$n_groups * 40
      }

      if(length(sizing_hints$no_char_y) != 1 || sizing_hints$no_char_y < 4) {
        sizing_hints$no_char_y <- 4
      }
      # final height in cm
      if(is.numeric(h)) {
        h <- h + 16 * sizing_hints$no_char_x
        h <- round(h * 0.0264583333, digits = 4)
        h <- paste0(h, "cm;")
      }
      # final width in cm
      if(is.numeric(w)) {
        w <- w + 16 * sizing_hints$no_char_y
        w <- round(w * 0.0264583333, digits = 4)
        w <- paste0(w, "cm;")
      }
    } else if (sizing_hints$type_plot == "rotated_plot") {
      # MARG_PLOT CASE3: ROTATED_PLOT
      w <- 500
      if (sizing_hints$n_groups < 11) {
        h <- 400
      } else if (sizing_hints$n_groups >10 &&
                 sizing_hints$n_groups < 21) {
        h <- 600
      } else {
        h <- "calc(var(--screen-height) * 0.5 * 1px);"
      }

      if(sizing_hints$no_char_y < 4) {
        sizing_hints$no_char_y <- 4
      }
      # final height in cm
      if(is.numeric(h)) {
        h <- h + 16 * sizing_hints$no_char_y
        h <- round(h * 0.0264583333, digits = 4)
        h <- paste0(h, "cm;")
      }
      # final width in cm
      if(is.numeric(w)) {
        w <- w + 16 * sizing_hints$no_char_x
        w <- round(w * 0.0264583333, digits = 4)
        w <- paste0(w, "cm;")
      }
    }
    ##CASE 7: pairs_plot in des_scatterplot_matrix
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "pairs_plot") {
    #In case of missing hints, give an error
    if (is.null(sizing_hints$number_of_vars)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }

    if (sizing_hints$number_of_vars < 6) {
      h <- 200 * sizing_hints$number_of_vars
      w <- 300 * sizing_hints$number_of_vars
    } else {
      h <- "calc(var(--screen-height) * 0.65 * 1px);"
      w <- "calc(var(--screen-width) * 0.9 * 1px);"
    }

    # final height in cm
    if(is.numeric(h)) {
      h <- round(h * 0.0264583333, digits = 4)
      h <- paste0(h, "cm;")
    }
    # final width in cm
    if(is.numeric(w)) {
      w <- round(w * 0.0264583333, digits = 4)
      w <- paste0(w, "cm;")
    }
    ##CASE 8: connected points with lines in multivariate outliers
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "multivar_plot") {
    #In case of missing hints, give an error
    if (is.null(sizing_hints$number_of_vars) ||
        is.null(sizing_hints$range)) {
      util_error(c("Internal error, sorry, please report:",
                   "Missing sizing hint values for %s"),
                 dQuote(sizing_hints$figure_type_id))
    }

    if (sizing_hints$number_of_vars < 7) {
      w <- 100 + 200 * sizing_hints$number_of_vars
    } else {
      w <- "calc(var(--screen-width) * 0.9 * 1px);"
    }

    if (sizing_hints$range < 101) {
      h <- 400
    } else if (sizing_hints$range > 100 &&
               sizing_hints$range < 401) {
      h <- 600
    } else  {
      h <- "calc(var(--screen-height) * 0.65 * 1px);"
    }

    if(sizing_hints$no_char_y < 4) {
      sizing_hints$no_char_y <- 4
    }

    # final height in cm
    if(is.numeric(h)) {
      h <- round(h * 0.0264583333, digits = 4)
      h <- paste0(h, "cm;")
    }
    # final width in cm
    if(is.numeric(w)) {
      w <- w + 16 * sizing_hints$no_char_y
      w <- round(w * 0.0264583333, digits = 4)
      w <- paste0(w, "cm;")
    }
    ##CASE 999: test only
  } else if (!is.null(sizing_hints$figure_type_id) &&
             sizing_hints$figure_type_id == "pass_through") {
    w <- sizing_hints$w
    h <- sizing_hints$h
  } else {
    # none of the plots before
    w <- NULL
    h <- NULL
  }
  return(list(figure_type_id = sizing_hints$figure_type_id,
              w = w,
              h = h))
}

#' Ensure, sizing hint sticks at the `dqr`, only
#'
#' @param dqr a `dataquieR` result
#' @param x a plot object
#'
#' @return a list with `dqr` and `x`, but fixed
#' @noRd
util_fix_sizing_hints <- function(dqr, x) {
  if (!is.null(attr(x, "sizing_hints"))) {
    if (!is.null(attr(dqr, "sizing_hints"))) {
      if (!identical(attr(x, "sizing_hints"),
                     attr(dqr, "sizing_hints"))) # FIXME: This is not numerically stable
        util_error(c("Internal error: two different",
                     "sizing_hints should not be,",
                     "sorry. Please report this error."))
    }
    attr(dqr, "sizing_hints") <- attr(x, "sizing_hints")
    attr(x, "sizing_hints") <- NULL
  }
  return(list(dqr = dqr, x = x))
}

test_that("util_finalize_sizing_hints works", {
  skip_on_cran()
  skip_if_not_installed("jsonlite")

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "scatt_plot",
      "range": 63,
      "number_of_vars": 1,
      "no_char_y":3
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 21.5, tolerance = 0.1)
  expect_equal(sh$h_in_cm, 9.3, tolerance = 0.1)

  expect_true(any(
    grepl("atomic vectors",
          capture.output(sh <- util_finalize_sizing_hints("x"),
                         type = "message")
    ))
  )

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "10px"
    }')

  expect_message2(sh <- util_finalize_sizing_hints(sh),
                         regexp = "sizes in pixels")

  expect_equal(sh$w_in_cm, 0.26, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "10pt"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 0.35, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "10pc"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 4.2, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "1em"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 0.4, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "1ex"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 0.4, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "1ch"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 0.4, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "1rem"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 0.4, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "1vw"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 48.8, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "1vh"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 27.4, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "1vmin"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 27.4, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "1vmax"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 48.8, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id": "pass_through",
      "w": "1%"
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 48.8, tolerance = 0.1)

  sh <- jsonlite::parse_json(
    '{"figure_type_id":"dot_loess",
      "range":28.2342,
      "no_char_y":6,
      "n_groups":14
    }')

  sh <- util_finalize_sizing_hints(sh)

  expect_equal(sh$w_in_cm, 25.4, tolerance = 0.1)

})

#' An exception class assigned for exceptions caused by long
#' variable labels
#' @noRd
LONG_LABEL_EXCEPTION <- "LONG_LABEL_EXCEPTION"
# FIXME: Use some exception class everywhere with some hierarchical concept for these classes

#' An exception class assigned for exceptions caused by trying to apply
#' a non-applicable indicator function, which is not caused by deficient
#' metadata
#'
#' Also amending meta data could not make the function running, e.g., a test
#' for numbers applied to a character.
#'
#' @export
dataquieR.intrinsic_applicability_problem <-
  "dataquieR.intrinsic_applicability_problem"

#' An exception class assigned for exceptions caused by trying to apply
#' a non-applicable indicator function
#'
#' Amending metadata could make the function running, e.g., a test
#' for missingness without any declared missing codes
#'
#' @export
dataquieR.applicability_problem <-
  "dataquieR.applicability_problem"

#' Detect Expected Observations
#'
#' For each participant, check, if an observation was expected, given the
#' `PART_VARS` from item-level metadata
#'
#' @param rv [character] the response variable, for that a value may be expected
#' @param study_data [study_data]
#' @param meta_data [meta_data]
#' @param label_col [character] mapping attribute `colnames(study_data)` vs.
#'                              `meta_data[label_col]`
#' @param expected_observations [enum] HIERARCHY | ALL | SEGMENT. How should
#'                                     `PART_VARS` be handled:
#'                                     - `ALL`: Ignore, all observations are
#'                                       expected
#'                                     - `SEGMENT`: if `PART_VAR` is 1, an
#'                                       observation is expected
#'                                     - `HIERARCHY`: the default, if the
#'                                       `PART_VAR` is 1 for this variable and
#'                                       also for all `PART_VARS` of `PART_VARS`
#'                                       up in the hierarchy, an observation is
#'                                       expected.
#'
#' @return a vector with `TRUE` or `FALSE` for each row of `study_data`, if for
#'         `study_data[rv]` a value is expected.
#'
#' @family missing_functions
#' @concept missing
#' @noRd
util_observation_expected <- # TODO: Support segment level metadata links to SEGMENT_PART_VARS
  function(rv, study_data, meta_data, label_col = LABEL,
           expected_observations =
             c("HIERARCHY",
               "ALL",
               "SEGMENT")) { # TODO: make prep

    util_expect_scalar(expected_observations, allow_more_than_one = TRUE)
    expected_observations <- match.arg(expected_observations)
    util_expect_scalar(expected_observations)

    all_need_to_be_1 <- # and the order is from root to leaf in the PART_VAR
                        # hierarchy
      util_all_intro_vars_for_rv(rv, study_data, meta_data, label_col,
                                 expected_observations = expected_observations)

    sd <- study_data
    missing_vars <- setdiff(all_need_to_be_1, colnames(study_data))
    if (any(missing_vars)) {
      util_warning(c("Missing %s from %s, I fill it with NA.",
                     "This may cause inconsistencies, if below in the",
                     "hierarchy, something is expected"),
                   paste0(dQuote(missing_vars), collapse = ", "),
                   sQuote("meta_data"))
      sd[, missing_vars] <- NA
    }
    sd <- sd[, all_need_to_be_1, drop = FALSE]

    sd[] <- !sapply(sd, util_is_na_0_empty_or_false)
    return(
      rowSums(sd, na.rm = TRUE) ==
        length(all_need_to_be_1)
    )
  }

#' @noRd
util_parallel_futures <- function(all_calls,
                                  worker,
                                  n_nodes,
                                  progress,
                                  debug_parallel,
                                  my_storr_object) {
  util_ensure_suggested("future")
  outer_env <- parent.frame()
  if (!is.null(parallel::getDefaultCluster())) {
    oplan <- future::plan(list(future::tweak(future::cluster,
                                             persistent = TRUE,
                             workers = parallel::getDefaultCluster()),
                       future::multisession))
    on.exit(future::plan(oplan), add = TRUE)
  }
    # don't use any auto graphics device (needed for certain
    # parallelization methods)
  rp <- lapply(setNames(seq_along(all_calls), nm = names(all_calls)),
           function(i) {
             progress(100 * i/length(all_calls))

             future::future(seed = TRUE, # IDEA: Use future_promise: https://rstudio.github.io/promises/articles/future_promise.html
               {
                 worker(all_calls[[i]], env = outer_env, nm =
                          names(all_calls)[[i]],
                        function_name = rlang::call_name(all_calls[[i]]),
                        my_storr_object = my_storr_object)
               }
              )

             #   label = names(all_calls)[[i]],
             #   expr = {
             #       R.devices::suppressGraphics(
             #           worker(all_calls[[i]], env = environment()))
             #      }
             # )
           }
  )
  r <- future::value(rp)
  # repeat {
  #   rsvld <- vapply(rp, resolved, FUN.VALUE = logical(1))
  #   if (all(rsvld)) break;
  #   progress(100 * sum(rsvld) / length(f))
  # }
  # later::run_now()
  # repeat {
  #   rsvld <- vapply(rp, resolved, FUN.VALUE = logical(1))
  #   if (all(rsvld)) break;
  #   progress(100 * sum(rsvld) / length(f))
  # }
  # as.list(r)
  r
}

#' Verify and normalize metadata on data frame level
#'
#' if possible, mismatching data types are converted (`"true"` becomes `TRUE`)
#'
#' missing columns are added, filled with `NA`, if this is valid, i.e., n.a.
#' for `DF_NAME` as the key column
#'
#' @param meta_data_dataframe [data.frame] data frame or path/url of a metadata
#'                                       sheet for the data frame level
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#' @param dataframe_level [data.frame] alias for `meta_data_dataframe`
#' @return standardized metadata sheet as data frame
#' @export
#'
#' @examples
#' \dontrun{
#' mds <- prep_check_meta_data_dataframe("ship_meta_dataframe|dataframe_level") # also converts
#' print(mds)
#' prep_check_meta_data_dataframe(mds)
#' mds1 <- mds
#' mds1$DF_RECORD_COUNT <- NULL
#' print(prep_check_meta_data_dataframe(mds1)) # fixes the missing column by NAs
#' mds1 <- mds
#' mds1$DF_UNIQUE_ROWS[[2]] <- "xxx" # not convertible
#' # print(prep_check_meta_data_dataframe(mds1)) # fail
#' mds1 <- mds
#' mds1$DF_UNIQUE_ID[[2]] <- 12
#' # print(prep_check_meta_data_dataframe(mds1)) # fail
#' }
prep_check_meta_data_dataframe <- function(meta_data_dataframe =
                                             "dataframe_level",
                                           meta_data_v2,
                                           dataframe_level) {
  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()
  util_expect_data_frame(meta_data_dataframe)

  if (!(DF_ELEMENT_COUNT %in% colnames(meta_data_dataframe))) {
    meta_data_dataframe[[DF_ELEMENT_COUNT]] <- rep(NA_integer_, nrow(meta_data_dataframe))
  }
  if (!(DF_RECORD_COUNT %in% colnames(meta_data_dataframe))) {
    meta_data_dataframe[[DF_RECORD_COUNT]] <- rep(NA_integer_, nrow(meta_data_dataframe))
  }
  if (!(DF_ID_REF_TABLE %in% colnames(meta_data_dataframe))) {
    meta_data_dataframe[[DF_ID_REF_TABLE]] <- rep(NA_character_, nrow(meta_data_dataframe))
  }
  if (!(DF_RECORD_CHECK %in% colnames(meta_data_dataframe))) {
    meta_data_dataframe[[DF_RECORD_CHECK]] <- rep(NA_character_, nrow(meta_data_dataframe))
  }
  if (!(DF_ID_VARS %in% colnames(meta_data_dataframe))) {
    meta_data_dataframe[[DF_ID_VARS]] <- rep(NA_character_, nrow(meta_data_dataframe))
  }
  if (!(DF_UNIQUE_ID %in% colnames(meta_data_dataframe))) {
    meta_data_dataframe[[DF_UNIQUE_ID]] <- rep(NA, nrow(meta_data_dataframe))
  }
  if (!(DF_UNIQUE_ROWS %in% colnames(meta_data_dataframe))) {
    meta_data_dataframe[[DF_UNIQUE_ROWS]] <- rep(NA, nrow(meta_data_dataframe))
  }
  if (!(DF_CODE %in% colnames(meta_data_dataframe))) {
    meta_data_dataframe[[DF_CODE]] <- rep(NA, nrow(meta_data_dataframe))
  }

  r <- util_expect_data_frame(
    custom_errors = list(
      DF_ID_VARS = "Column DF_ID_VARS must be character",
      DF_NAME = "Column DF_NAME must be character",
      DF_CODE = "Column DF_CODE must be character",
      DF_RECORD_COUNT = "Column DF_RECORD_COUNT must be integer",
      DF_ELEMENT_COUNT = "Column DF_ELEMENT_COUNT must be integer",
      DF_ID_REF_TABLE = "Column DF_ID_REF_TABLE must be character",
      DF_RECORD_CHECK = "Column DF_RECORD_CHECK must be subset, superset or exact",
      DF_UNIQUE_ROWS = "Column DF_UNIQUE_ROWS must be true, false, or no_id",
      DF_UNIQUE_ID = "Column DF_UNIQUE_ID must be integer"
    ),
    meta_data_dataframe,
    list(
      DF_ID_VARS = is.character, # TODO: Write utility functions for converting data types with proper warnings
      DF_NAME = is.character,
      DF_CODE = is.character,
      DF_RECORD_COUNT = util_all_is_integer,
      DF_ELEMENT_COUNT = util_all_is_integer,
      DF_ID_REF_TABLE = is.character,
      DF_RECORD_CHECK = function(x) {
        all(util_empty(x) | x %in% c("superset", "subset", "exact"))
      },
      DF_UNIQUE_ROWS = function(x) {
        all(util_empty(x) | tolower(trimws(x)) %in%
              c("f", "t", "true", "false", "no_id"))
      },
      DF_UNIQUE_ID = util_all_is_integer
    ), # TODO: We have DF_ID_REF_TABLE but DF_ID_TABLE. Could you add a todo for this?
    list(
      DF_ID_VARS = as.character,
      DF_NAME = as.character,
      DF_CODE = as.character,
      DF_RECORD_COUNT = as.integer,
      DF_ELEMENT_COUNT = as.integer,
      DF_ID_REF_TABLE = as.character,
      DF_RECORD_CHECK = function(x) {
        r <-
          factor(tolower(trimws(as.character(x))),
                 levels = c(c("superset", "subset", "exact")))
        levels(r)[as.numeric(r)]
      },
      DF_UNIQUE_ROWS = as.character,
      DF_UNIQUE_ID = as.integer
    )
  )

  no_key <- util_empty(r[[DF_NAME]])
  sum_no_key <- sum(no_key)

  if (sum_no_key) {
    util_message("Removing %d rows from %s, because %s is empty.",
                 sum_no_key,
                 dQuote("meta_data_dataframe"),
                 sQuote(DF_NAME),
                 applicability_problem = TRUE
                 )
  }

  r <- r[!no_key, , drop = FALSE]

  # Check each DF_CODE and DF_NAME not used more than once
  if (any(duplicated(r[[DF_NAME]], incomparables = NA))) {
    util_error("Found duplicated dataframes in dataframe level metadata: %s",
               util_pretty_vector_string(
                 sQuote(r[[DF_NAME]][
                   duplicated(r[[DF_NAME]], incomparables = NA)])
               ),
               applicability_problem = TRUE)
  }

  if (any(duplicated(r[[DF_CODE]], incomparables = NA))) {
    util_error(
      "Found duplicated dataframe codes in dataframe level metadata: %s",
      util_pretty_vector_string(
        sQuote(r[[DF_CODE]][
          duplicated(r[[DF_CODE]], incomparables = NA)])
      ),
      applicability_problem = TRUE)
  }

  r
}

# util_expand_pattern_rules.R
# This file contains a helper to expand pattern-based variable references
# inside cross-item rules. Only content inside square brackets [ ... ] is
# interpreted as a pattern. Everything else is left unchanged.

util_expand_pattern_rules <- function(pattern_rules, valid_names) {

  e <- environment()

  mismatches <- list()

  # Basic input checks
  util_stop_if_not(is.character(pattern_rules), is.character(valid_names))
  if (any(inv <- grepl("[][]", valid_names))) {
    util_warning("valid names must not contain '[' or ']'.",
                 applicability_problem = TRUE)
    valid_names <- valid_names[!inv]
  }

  # Escape regex meta characters to ensure user input cannot alter regex logic
  .rx_escape_chr <- function(x) {
    gsub("([][{}()+*?^$|\\\\.,])", "\\\\\\1", x, perl = TRUE)
  }

  # Read escaped character (e.g. \*, \?, \#)
  .read_escaped <- function(x, i) {
    n <- nchar(x)
    if (i < n && substr(x, i, i) == "\\") {
      list(value = substr(x, i + 1L, i + 1L), next_i = i + 2L)
    } else {
      list(value = substr(x, i, i), next_i = i + 1L)
    }
  }

  # Find closing brace for capture definitions { ... }
  .find_unescaped <- function(x, char, start) {
    n <- nchar(x)
    i <- start
    while (i <= n) {
      if (substr(x, i, i) == "\\") {
        i <- i + 2L
      } else if (substr(x, i, i) == char) {
        return(i)
      } else {
        i <- i + 1L
      }
    }
    NA_integer_
  }

  # Resolve capture references such as {W}, {W-1}, or {W+1}.
  # Arithmetic is only attempted for integer capture values.
  .resolve_capture_reference <- function(cap_expr, captures) {
    m <- regexec("^([A-Za-z][A-Za-z0-9_]*)([+-][0-9]+)?$", cap_expr, perl = TRUE)
    mm <- regmatches(cap_expr, m)[[1L]]

    if (!length(mm)) {
      util_error(sprintf("Invalid capture reference {%s}.", cap_expr), call. = FALSE)
    }

    ref_name <- mm[[2L]]
    offset <- if (length(mm) >= 3L && nzchar(mm[[3L]])) {
      as.integer(mm[[3L]])
    } else {
      0L
    }

    if (is.null(captures[[ref_name]])) {
      util_error("Capture {%s} used before being defined.", ref_name)
    }

    ref_value <- captures[[ref_name]]

    if (offset != 0L) {
      if (!grepl("^-?[0-9]+$", ref_value)) {
        return(NULL)
      }
      ref_value <- as.character(as.integer(ref_value) + offset)
    }

    ref_value
  }

  # Translate wildcard mini-language into regex
  .compile_fragment <- function(x) {
    rx <- character()
    i <- 1L
    n <- nchar(x)

    while (i <= n) {
      ch <- substr(x, i, i)
      ch2 <- if (i < n) substr(x, i, i + 1L) else ""

      if (ch == "\\") {
        esc <- .read_escaped(x, i)
        rx <- c(rx, .rx_escape_chr(esc$value))
        i <- esc$next_i
      } else if (ch2 == "#?") {
        rx <- c(rx, "[0-9]{1,2}")
        i <- i + 2L
      } else if (ch2 == "##") {
        rx <- c(rx, "[0-9]{2}")
        i <- i + 2L
      } else if (ch == "#") {
        rx <- c(rx, "[0-9]")
        i <- i + 1L
      } else if (ch == "*") {
        rx <- c(rx, ".*")
        i <- i + 1L
      } else if (ch == "?") {
        rx <- c(rx, ".")
        i <- i + 1L
      } else {
        rx <- c(rx, .rx_escape_chr(ch))
        i <- i + 1L
      }
    }

    paste0(rx, collapse = "")
  }

  # Compile full variable pattern with optional capture groups
  .compile_var_pattern <- function(pattern, captures = list()) {
    rx <- character()
    capture_names <- character()
    pattern_invalid <- FALSE
    i <- 1L
    n <- nchar(pattern)

    while (i <= n) {
      ch <- substr(pattern, i, i)

      if (ch == "\\") {
        esc <- .read_escaped(pattern, i)
        rx <- c(rx, .rx_escape_chr(esc$value))
        i <- esc$next_i
      } else if (ch == "{") {
        close <- .find_unescaped(pattern, "}", i + 1L)
        if (is.na(close)) util_error("Unclosed capture in pattern")

        inner <- substr(pattern, i + 1L, close - 1L)
        def <- strsplit(inner, ":", fixed = TRUE)[[1L]]
        cap_name <- def[[1L]]

        if (length(def) == 1L) {
          ref_value <- .resolve_capture_reference(cap_name, captures)

          if (is.null(ref_value)) {
            pattern_invalid <- TRUE
            rx <- c(rx, "(?!)")
          } else {
            rx <- c(rx, .rx_escape_chr(ref_value))
          }
        } else if (!is.null(captures[[cap_name]])) {
          rx <- c(rx, .rx_escape_chr(captures[[cap_name]]))
        } else {
          rx <- c(rx, "(", .compile_fragment(def[[2L]]), ")")
          capture_names <- c(capture_names, cap_name)
        }

        i <- close + 1L
      } else {
        next_special <- regexpr("[\\\\{]", substr(pattern, i, n), perl = TRUE)[[1L]]
        end <- if (next_special < 0L) n else i + next_special - 2L
        rx <- c(rx, .compile_fragment(substr(pattern, i, end)))
        i <- end + 1L
      }
    }

    list(
      regex = paste0("^", paste0(rx, collapse = ""), "$"),
      capture_names = capture_names,
      pattern_invalid = pattern_invalid
    )
  }

  # Expand a single token against valid names
  .expand_token <- function(token, captures = list()) {
    compiled <- .compile_var_pattern(token, captures)

    hits <- valid_names[grepl(compiled$regex, valid_names, perl = TRUE)]

    if (!length(hits)) {
      return(structure(list(), pattern_invalid = compiled$pattern_invalid))
    }

    lapply(hits, function(nm) {
      new_captures <- captures

      if (length(compiled$capture_names)) {
        m <- regexec(compiled$regex, nm, perl = TRUE)
        vals <- regmatches(nm, m)[[1L]][-1L]

        for (i in seq_along(compiled$capture_names)) {
          new_captures[[compiled$capture_names[[i]]]] <- vals[[i]]
        }
      }

      list(name = nm, captures = new_captures)
    })
  }

  # Expand full rule
  .extract_bracket_tokens <- function(rule) {
    m <- gregexpr("\\[[^][]+\\]", rule, perl = TRUE)
    tokens <- regmatches(rule, m)[[1L]]
    if (!length(tokens)) return(character(0))
    sub("^\\[", "", sub("\\]$", "", tokens))
  }

  .expand_rule <- function(rule) {
    tokens <- .extract_bracket_tokens(rule)
    states <- list(list(text = rule, captures = list()))

    for (token in tokens) {
      new_states <- list()
      token_had_hit <- FALSE

      for (state in states) {
        hits <- .expand_token(token, state$captures)

        if (length(hits) || isTRUE(attr(hits, "pattern_invalid"))) {
          token_had_hit <- TRUE
        }

        for (hit in hits) {
          new_states[[length(new_states) + 1L]] <- list(
            text = sub(
              paste0("[", token, "]"),
              paste0("[", hit$name, "]"),
              state$text,
              fixed = TRUE
            ),
            captures = hit$captures
          )
        }
      }

      if (!token_had_hit) {
        assign("mismatches",
               c(e$mismatches, token),
               envir = e)
      }

      states <- new_states
      if (!length(states)) break
    }

    unique(vapply(states, `[[`, character(1), "text"))
  }

  res <- force(unlist(lapply(pattern_rules, .expand_rule), use.names = FALSE))

  util_attach_attr(
    res,
    mismatches = unique(mismatches)
  )
}

#' Import vector of hover text for tables in the report
#'
#' @param x name of the tables. They are `meta_data`,
#'            `meta_data_segment`, `meta_data_dataframe`, `meta_data_cross_item`,
#'            `meta_data_item_computation`, `com_item_missingness`,
#'            `int_datatype_matrix`, `con_inadmissible_categorical`,
#'            `rulesetformat`, `gradingruleset`
#' @return named vector containing the hover text from the file `metadata-hovertext.rds`
#'          in the inst folder. Names correspond to column names in the metadata
#'          tables
#'
#' @family html
#' @concept reporting
#' @noRd
util_get_hovertext <- function(x) {
  f <- system.file(paste0("hovertext", ".rds"), package = "dataquieR")
  if (exists(f, .concept_chache, mode = "list")) {
    list_of_hover_metadata <- get(f, .concept_chache, mode = "list")
  } else {
    if (0 != file.access(f, 4)) {
      util_error("Cannot read file %s. Internal error.",
                 dQuote("metadata-hovertext"))
    }
    list_of_hover_metadata <- readRDS(f)
    assign(f, list_of_hover_metadata, .concept_chache)
  }
  list_of_hover <- list_of_hover_metadata[[x]]
  list_of_hover
}

#' Extract all ids from a list of `htmltools` objects
#'
#' @param pages the list of objects
#'
#' @return a character vector with valid targets
#' @family html
#' @noRd
util_extract_all_ids <- function(pages) {
  all_ids <- lapply(lapply(pages, .util_extract_all_ids_from_htmltools), unlist)
  all_ids <- lapply(all_ids, unname)
  all_ids <- all_ids[!vapply(all_ids, is.null, FUN.VALUE = logical(1))]
  all_ids <- unlist(
    lapply(names(all_ids), function(fn) { paste0(fn, "#", all_ids[[fn]]) } ))
  all_ids <- c(gsub("#.*$", "", all_ids), all_ids)
  all_ids <- unique(c("report.html", all_ids))
  all_ids
}
.util_extract_all_ids_from_htmltools <- function(html_stuff) {
  r <- list()
  if ("children" %in% names(html_stuff)) {
    r <- c(r, .util_extract_all_ids_from_htmltools(html_stuff$children))
    html_stuff$children <- NULL
  }
  if ("attribs" %in% names(html_stuff)) {
    if ("id" %in% names(html_stuff$attribs)) {
      r <- c(r, html_stuff$attribs$id)
    }
    html_stuff$attribs <- NULL
  }
  if (is.list(html_stuff) || length(html_stuff) > 1) {
    c(r, lapply(html_stuff, .util_extract_all_ids_from_htmltools))
  } else {
    r
  }
}

#' Return the category for a result
#'
#' messages do not cause any category, warnings are `cat3`, errors are `cat5`
#'
#' @param result a `dataquieR_resultset2` result
#' @param aspect an aspect/problem category of results
#'               (error, applicability error)
#' @param ... not used
#'
#' @return a category, see `util_as_cat()`
#'
#' @family summary_functions
#' @noRd
util_get_category_for_result <- function(result, aspect =
                                        c("applicability", "error",
                                          "anamat", "indicator_or_descriptor"),
                                      ...) {

  aspect <- util_match_arg(aspect, several_ok = FALSE)

  ##### preps -----
  util_ensure_suggested("htmltools", "Generating nice tables")

  ### indicator_or_descriptor -----

  if (aspect %in% c("indicator_or_descriptor")) {
    function_name <- attr(result, "function_name")
    if (is.null(function_name)) { # if we do not have a result, we cannot tell
      return(util_as_cat(NA))
    }
    if (get(function_name, .indicator_or_descriptor)) {
      return(cat1) # It's a data quality indicator
    } else {
      return(cat3) # It's a descriptor
    }
    # # TODO: Fix "error" must be added in squarereportrenderer
    # expected_a_result <-
    #   (util_get_category_for_result(
    #     result, "applicability") %in% c(cat1, cat2, cat3))
    # # TODO: Fix error must be added in squarereportrenderer
    # expected_a_result <-
    #   expected_a_result &&
    #   !is.na(util_get_category_for_result(result, "anamat"))
    # if (length(result) == 0 || inherits(result, "dataquieR_NULL")) {
    #   if (expected_a_result)
    #     return(cat4) # result missing for unknown reasons
    #   else
    #     return(cat3) # result missing, but because of n.a. (intrinsic or due to missing metadata)
    # }
    #
    # if (is.data.frame(result$SummaryTable)) { # TODO: This is not enough, now we have not only GRADING and we have SegmentTable, ... -- better mark indicator functions as such.
    #   if (!is.null(result$SummaryTable$GRADING)) {
    #     if (all(is.na(result$SummaryTable$GRADING))) {
    #       return(cat4) # grading column exists but is empty
    #     } else {
    #       return(cat1) # It's a data quality indicator
    #     }
    #   } else {
    #     # no grading column provided
    #     return(cat3) # It's a descriptor
    #   }
    #   util_error("this should be unreachable. Internal error, please repor a bug, sorry.")
    # }
    # return(cat3) # It's a descriptor
  } else { ##### error or applicability or anamat -----
    if (length(attr(result, "error")) > 0) { # some error occurred ----
      util_stop_if_not(length(attr(result, "error")) == 1)
      cnd <- attr(result, "error")[[1]]
      applicability_problem <- attr(cnd, "applicability_problem")
      if (is.null(applicability_problem) || is.na(applicability_problem)) {
        applicability_problem <- FALSE
      }
      intrinsic_applicability_problem <- attr(cnd,
                                              "intrinsic_applicability_problem")
      if (is.null(intrinsic_applicability_problem) ||
          is.na(intrinsic_applicability_problem)) {
        intrinsic_applicability_problem <- FALSE
      }
      if (aspect == "anamat" && applicability_problem &&
          intrinsic_applicability_problem) {
        return(cat5)
      } else if (aspect == "anamat" && applicability_problem &&
          !intrinsic_applicability_problem) {
        return(cat1)
      } else if (aspect == "anamat" && !applicability_problem) {
        return(cat1)
      } else if (aspect == "applicability" && applicability_problem &&
                 !intrinsic_applicability_problem) {
        return(cat5) # applicability error was asked and occurred
      } else if (aspect == "error"  && !applicability_problem) {
        return(cat5) # other error was asked and occurred
      }
      return(cat3) # error of the other class (not asked now) occurred
    }
    # If we have warnings, we cannot return, we have to go over all warnings and return cat3, if we find a warning of the aspect's class
    res <- cat1
    if (length(attr(result, "warning")) > 0) { # some warning occurred ----
      for (w in attr(result, "warning")) {
        applicability_problem <- attr(w, "applicability_problem")
        if (is.null(applicability_problem) || is.na(applicability_problem)) {
          applicability_problem <- FALSE
        }
        intrinsic_applicability_problem <- attr(w,
                                                "intrinsic_applicability_problem")
        if (is.null(intrinsic_applicability_problem) ||
            is.na(intrinsic_applicability_problem)) {
          intrinsic_applicability_problem <- FALSE
        }
        if (aspect == "anamat" && applicability_problem &&
            intrinsic_applicability_problem) {
          # if (res != cat3)
          #   res <- washedoutgreen # maybe no example possible?
        } else if (aspect == "anamat" && applicability_problem &&
                   !intrinsic_applicability_problem) {
          # if (res != cat3)
          #   res <- cat1
        } else if (aspect == "anamat" && !applicability_problem) {
          # if (res != cat3)
          #   res <- cat1
        } else if (aspect == "applicability") {
          if (applicability_problem && !intrinsic_applicability_problem) {
            # res <- cat3
            return(cat3)
          }
        } else if (aspect == "error") {
          if (!applicability_problem) {
            # res <- cat3
            return(cat3)
          }
        } else util_error("internal error in get color for result: %s",
                          dQuote(aspect))
      }
    }
    return(cat1)
  }
}

#' Extract summary from data quality results
#'
#' Generic function, currently supports [dq_report2] and [dataquieR_result]
#'
#' @param r [dq_report2] or [dataquieR_result] object
#' @param ... further arguments, maybe needed for some implementations
#'
#' @export
#' @return [list] with two slots `Data` and `Table` with [data.frame]s
#'                featuring all metrics columns
#'                from the report or result in `x`,
#'                the [STUDY_SEGMENT] and the [VAR_NAMES].
#'                In case of `Data`, the columns are formatted nicely but still
#'                with the standardized column names -- use
#'                `util_translate_indicator_metrics()` to rename them nicely. In
#'                case of `Table`, just as they are.
#' @family summary_functions
prep_extract_summary <- function(r,
                                 ...) {
  te <- topenv(parent.frame(1)) # see https://stackoverflow.com/a/27870803
  if (!(isNamespace(te) && getNamespaceName(te) == "dataquieR")) {
    lifecycle::deprecate_soft("2.1.0.9007",
                              "prep_extract_summary()",
                              "summary()")
  }
  UseMethod("prep_extract_summary")
}

#' HTML Dependency for `dataquieR`
#'
#' generate all dependencies used in static `dataquieR` reports
#'
#' @param iframe [logical]`(1)` if `TRUE`, create the dependency used in figure
#' `iframes`.
#'
#' @return the dependency
html_dependency_dataquieR <- function(iframe = FALSE) {
  util_expect_scalar(iframe, check_type = is.logical)
  if (iframe) {
    htmltools:: htmlDependency("menu", "1.0.1",
                               src = system.file("menu", package =
                                                   "dataquieR"),
                               stylesheet = c("style.css", "style_iframe.css"),
                               script = c("script.js", "script2.js",
                                          "script_iframe.js"))
  } else {
    htmltools:: htmlDependency("menu", "1.0.1",
                               src = system.file("menu", package =
                                                   "dataquieR"),
                               stylesheet = c("style.css", "style_toplevel.css"),
                               script = c("script.js", "script2.js",
                                          "script_toplevel.js"))
  }
}

#' Get the thresholds for grading
#'
#' @param indicator_metric which indicator metric to be classified
#' @param meta_data the item level metadata
#'
#' @return named list (names are `VAR_NAMES`, values are named vectors of
#'         intervals, names in the vectors are the category numbers)
#' @family summary_functions
#' @noRd
util_get_thresholds <- function(indicator_metric, meta_data) {

  vars <- meta_data[[VAR_NAMES]]

  if (!(GRADING_RULESET %in% colnames(meta_data))) {
    meta_data[[GRADING_RULESET]] <- 0
  }
  meta_data[
    util_empty(meta_data[[GRADING_RULESET]]),
    GRADING_RULESET] <-
    0

  rsts <- util_get_rule_sets()
  availab_rulesets <- meta_data[[GRADING_RULESET]] %in% names(rsts)
  if (any(!availab_rulesets)) {##### TODO: issue #723 https://gitlab.com/libreumg/dataquier/-/work_items/723
    util_warning(c("The following rulesets are used by item-level metadata",
                   "but missing from the ruleset file: %s. Falling back to",
                   "default (0) for them"),
                 util_pretty_vector_string(
                   unique(meta_data[[GRADING_RULESET]][!availab_rulesets])))
    meta_data[[GRADING_RULESET]][!availab_rulesets] <- "0"
  }
  trs <- lapply(setNames(nm = vars), function(vn) {

    rs <-
      as.character(
        meta_data[meta_data[[VAR_NAMES]] == vn, GRADING_RULESET, TRUE]
      )
    rst <- rsts[[rs]][rsts[[rs]]$indicator_metric == indicator_metric, , FALSE]
    if (nrow(rst) > 1) {
      util_message("More than one ruleset for %s / %s, using the first one",
                   dQuote(vn),
                   dQuote(indicator_metric),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = FALSE)
      rst <- head(rst, 1)
    } else if (nrow(rst) == 0) {
      return(setNames(nm = character(0)))
    }
    if(is.null(rst[["dqi_catnum"]])) { #TODO: make it robust
      cats <- seq_len(5)
    } else {
      cats <- seq_len(as.integer(rst[["dqi_catnum"]]))
    }

    cat_names <- as.character(cats)
    breaks <- paste0("dqi_cat_", cats)
    breaks <- unlist(rst[, breaks])
    setNames(breaks, nm = cat_names)
  })

  trs
}

#' Checks the validity of metadata w.r.t. the provided column names
#'
#' @description
#' This function verifies, if a data frame complies to metadata conventions and
#' provides a given richness of meta information as specified by `level`.
#'
#' @details
#'
#' Note, that only the given level is checked despite, levels are somehow
#' hierarchical.
#'
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param level [enum] level of requirement (see also [VARATT_REQUIRE_LEVELS]).
#'                     set to `NULL` to deactivate the check of richness.
#' @param character.only [logical] a logical indicating whether level can be
#'                                 assumed to be character strings.
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return a logical with:
#'   - invisible(TRUE). In case of problems with the metadata, a condition is
#'      raised (`stop()`).
#' @export
#'
#' @examples
#' \dontrun{
#' prep_check_meta_names(data.frame(VAR_NAMES = 1, DATA_TYPE = 2,
#'                       MISSING_LIST = 3))
#'
#' prep_check_meta_names(
#'   data.frame(
#'     VAR_NAMES = 1, DATA_TYPE = 2, MISSING_LIST = 3,
#'     LABEL = "LABEL", VALUE_LABELS = "VALUE_LABELS",
#'     JUMP_LIST = "JUMP_LIST", HARD_LIMITS = "HARD_LIMITS",
#'     GROUP_VAR_OBSERVER = "GROUP_VAR_OBSERVER",
#'     GROUP_VAR_DEVICE = "GROUP_VAR_DEVICE",
#'     TIME_VAR = "TIME_VAR",
#'     PART_VAR = "PART_VAR",
#'     STUDY_SEGMENT = "STUDY_SEGMENT",
#'     LOCATION_RANGE = "LOCATION_RANGE",
#'     LOCATION_METRIC = "LOCATION_METRIC",
#'     PROPORTION_RANGE = "PROPORTION_RANGE",
#'     MISSING_LIST_TABLE = "MISSING_LIST_TABLE",
#'     CO_VARS = "CO_VARS",
#'     LONG_LABEL = "LONG_LABEL"
#'   ),
#'   RECOMMENDED
#' )
#'
#' prep_check_meta_names(
#'   data.frame(
#'     VAR_NAMES = 1, DATA_TYPE = 2, MISSING_LIST = 3,
#'     LABEL = "LABEL", VALUE_LABELS = "VALUE_LABELS",
#'     JUMP_LIST = "JUMP_LIST", HARD_LIMITS = "HARD_LIMITS",
#'     GROUP_VAR_OBSERVER = "GROUP_VAR_OBSERVER",
#'     GROUP_VAR_DEVICE = "GROUP_VAR_DEVICE",
#'     TIME_VAR = "TIME_VAR",
#'     PART_VAR = "PART_VAR",
#'     STUDY_SEGMENT = "STUDY_SEGMENT",
#'     LOCATION_RANGE = "LOCATION_RANGE",
#'     LOCATION_METRIC = "LOCATION_METRIC",
#'     PROPORTION_RANGE = "PROPORTION_RANGE",
#'     DETECTION_LIMITS = "DETECTION_LIMITS", SOFT_LIMITS = "SOFT_LIMITS",
#'     CONTRADICTIONS = "CONTRADICTIONS", DISTRIBUTION = "DISTRIBUTION",
#'     DECIMALS = "DECIMALS", VARIABLE_ROLE = "VARIABLE_ROLE",
#'     DATA_ENTRY_TYPE = "DATA_ENTRY_TYPE",
#'     CO_VARS = "CO_VARS",
#'     END_DIGIT_CHECK = "END_DIGIT_CHECK",
#'     VARIABLE_ORDER = "VARIABLE_ORDER", LONG_LABEL =
#'       "LONG_LABEL", recode = "recode",
#'       MISSING_LIST_TABLE = "MISSING_LIST_TABLE"
#'   ),
#'   OPTIONAL
#' )
#'
#' # Next one will fail
#' try(
#'   prep_check_meta_names(data.frame(VAR_NAMES = 1, DATA_TYPE = 2,
#'     MISSING_LIST = 3), TECHNICAL)
#' )
#' }
prep_check_meta_names <- function(
    item_level = "item_level", level,
                                  character.only = FALSE,
                                  meta_data = item_level,
    meta_data_v2) {
  util_maybe_load_meta_data_v2()
  util_expect_data_frame(meta_data)
  if (missing(level)) {
    level <- VARATT_REQUIRE_LEVELS$REQUIRED
  } else {
    if (character.only) {
      level <- as.character(level)
    } else {
      lv <- substitute(expr = level)
      if (length(lv) == 3 &&
          as.character(lv[[1]]) == "$" &&
          as.character(lv[[2]]) =="VARATT_REQUIRE_LEVELS")
        util_error(
          c("%s is called either with %s",
            "set to %s or with names from %s without quotes."),
          sQuote("prep_check_meta_names"),
          sQuote("character.only"),
          dQuote(TRUE),
          dQuote('VARATT_REQUIRE_LEVELS')
        )
      level <- as.character(lv)
    }
    if (length(level) > 0) {
      level <- try(match.arg(level, choices = c(names(VARATT_REQUIRE_LEVELS),
                                                unlist(VARATT_REQUIRE_LEVELS)),
                             several.ok = FALSE), silent = TRUE)
      if (inherits(level, "try-error")) {
        util_error(
          "Error regarding argument %s: %s",
          dQuote("level"),
          conditionMessage(attr(level, "condition"))
        )
      }
      if (level %in% names(VARATT_REQUIRE_LEVELS)) {
        level <- VARATT_REQUIRE_LEVELS[[level]]
      }
    }
  }
  env <- new.env(environment())
  env$res <- try(stop("Internal error"), silent = TRUE)
  try(
    withCallingHandlers({
        if (!is.data.frame(meta_data)) {
          util_error("metadata is not a data frame at all")
        }
        required_atts <- util_get_var_att_names_of_level(level)
        missing_atts <- setdiff(required_atts, colnames(meta_data))
        if (WELL_KNOWN_META_VARIABLE_NAMES$MISSING_LIST_TABLE %in% # no mlt
            missing_atts && WELL_KNOWN_META_VARIABLE_NAMES$MISSING_LIST %in%
              colnames(meta_data)) { # no problem, if missing list is there
            missing_atts <- setdiff(missing_atts,
                                    WELL_KNOWN_META_VARIABLE_NAMES$
                                      MISSING_LIST_TABLE)
        }
        if (length(missing_atts) > 0) {
          util_error(
            c("Missing %s: Not all variable attributes of requirement",
              "level %s (%s) are available in the metadata (%s)."),
            paste0(dQuote(missing_atts), collapse = ", "),
            dQuote(level),
            paste0(dQuote(required_atts), collapse = ", "),
            paste0(dQuote(colnames(meta_data)), collapse = ", "),
            applicability_problem = TRUE
          )
        }
        if (!all(colnames(meta_data) %in% required_atts)) {
          suspicious_names <-
            colnames(meta_data)[!(colnames(meta_data) %in% required_atts)]

          normLev.fnc <- function(a, b, ...) {
            # https://stackoverflow.com/q/10140923
            drop(adist(a, b, ...) / nchar(attr(adist(a, b, counts = TRUE, ...),
                                               "trafos")))
          }

          fuzzy_match <- vapply(
            # and try to guess, what the user wanted to put there
            suspicious_names,
            function(v) {
              distances <- normLev.fnc(trimws(v), trimws(required_atts),
                                       ignore.case = TRUE, fixed = TRUE)
              if (any(distances < 0.3)) {
                required_atts[which.min(distances)]
              } else {
                NA_character_
              }
            },
            ""
          )
          fuzzy_match <- fuzzy_match[!is.na(fuzzy_match)]
          if (length(fuzzy_match) > 0) {
            util_warning(c("Found the following addtional metadata columns,",
                           "which look like typos of defined names: %s"),
                         paste(dQuote(names(fuzzy_match)), dQuote(fuzzy_match),
                               sep = " -> ", collapse = ", "),
                         applicability_problem = TRUE)
          }
        }
        env$res <- TRUE
      },
      error = function(cond) {
        env$res <- cond
      }
    ),
    silent = TRUE
  )
  if (inherits(env$res, "condition")) {
    env$res[["call"]] <- sys.call()
    util_error(env$res)
  }
  invisible(env$res)
}

test_that("prep_check_meta_names works", {
  skip_on_cran()
  
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                  dataquieR.ERRORS_WITH_CALLER = TRUE,
                  dataquieR.WARNINGS_WITH_CALLER = TRUE,
                  dataquieR.MESSAGES_WITH_CALLER = TRUE)
 expect_silent(prep_check_meta_names(data.frame(), level = NULL))

 expect_silent( prep_check_meta_names(data.frame(VAR_NAMES = 1, DATA_TYPE = 2,
                        MISSING_LIST = 3)))

  expect_silent(prep_check_meta_names(
    data.frame(
      VAR_NAMES = 1, DATA_TYPE = 2, MISSING_LIST = 3,
      SCALE_LEVEL = NA_character_,
      UNIT = NA_character_,
      LABEL = "LABEL", VALUE_LABELS = "VALUE_LABELS",
      JUMP_LIST = "JUMP_LIST", HARD_LIMITS = "HARD_LIMITS",
      GROUP_VAR_OBSERVER = "GROUP_VAR_OBSERVER",
      GROUP_VAR_DEVICE = "GROUP_VAR_DEVICE",
      TIME_VAR = "TIME_VAR",
      STUDY_SEGMENT = "STUDY_SEGMENT",
      PART_VAR = "PART_VAR",
      LOCATION_RANGE = "LOCATION_RANGE",
      LOCATION_METRIC = "LOCATION_METRIC",
      PROPORTION_RANGE = "PROPORTION_RANGE",
      MISSING_LIST_TABLE = "MISSING_LIST_TABLE",
      LONG_LABEL = "LONG_LABEL",
      CO_VARS = "CO_VARS",
      UNIVARIATE_OUTLIER_CHECKTYPE = "UNIVARIATE_OUTLIER_CHECKTYPE",
      N_RULES = "N_RULES"
    ),
    RECOMMENDED
  ))

  expect_silent(prep_check_meta_names(
    data.frame(
      VAR_NAMES = 1, DATA_TYPE = 2, MISSING_LIST = 3,
      SCALE_LEVEL = NA_character_,
      UNIT = NA_character_,
      LABEL = "LABEL", VALUE_LABELS = "VALUE_LABELS",
      VALUE_LABEL_TABLE = "VALUE_LABEL_TABLE",
      STANDARDIZED_VOCABULARY_TABLE = "STANDARDIZED_VOCABULARY_TABLE",
      JUMP_LIST = "JUMP_LIST", HARD_LIMITS = "HARD_LIMITS",
      GROUP_VAR_OBSERVER = "GROUP_VAR_OBSERVER",
      GROUP_VAR_DEVICE = "GROUP_VAR_DEVICE",
      TIME_VAR = "TIME_VAR",
      STUDY_SEGMENT = "STUDY_SEGMENT",
      PART_VAR = "PART_VAR",
      DETECTION_LIMITS = "DETECTION_LIMITS", SOFT_LIMITS = "SOFT_LIMITS",
      CONTRADICTIONS = "CONTRADICTIONS", DISTRIBUTION = "DISTRIBUTION",
      DECIMALS = "DECIMALS", VARIABLE_ROLE = "VARIABLE_ROLE",
      DATA_ENTRY_TYPE = "DATA_ENTRY_TYPE",
      END_DIGIT_CHECK = "END_DIGIT_CHECK",
      VARIABLE_ORDER = "VARIABLE_ORDER", LONG_LABEL =
        "LONG_LABEL", recode = "recode",
      LOCATION_RANGE = "LOCATION_RANGE",
      LOCATION_METRIC = "LOCATION_METRIC",
      PROPORTION_RANGE = "PROPORTION_RANGE",
      MISSING_LIST_TABLE = "MISSING_LIST_TABLE",
      CO_VARS = "CO_VARS",
      GRADING_RULESET = "GRADING_RULESET",
      RECODE_CASES = "RECODE_CASES",
      RECODE_CONTROL = "RECODE_CONTROL",
      DATAFRAMES = "DATAFRAMES",
      ENCODING = "ENCODING",
      UNIVARIATE_OUTLIER_CHECKTYPE = "UNIVARIATE_OUTLIER_CHECKTYPE",
      N_RULES = "N_RULES",
      EXTENDED_DATA_TYPE = "EXTENDED_DATA_TYPE",
      TIME_VAR_END = "TIME_VAR_END"
    ),
    OPTIONAL
  ))

  expect_error(
    prep_check_meta_names(data.frame(VAR_NAMES = 1, DATA_TYPE = 2,
    MISSING_LIST = 3), TECHNICAL),
    regexp = "Not all variable attributes of requirement level .+technical.+",
    perl = TRUE
  )

  expect_error(
    prep_check_meta_names(data.frame(VAR_NAMES = 1, DATA_TYPE = 2,
                                     MISSING_LIST = 3), "TECHNICAL",
                          character.only = TRUE),
    regexp = "Not all variable attributes of requirement level .+technical.+",
    perl = TRUE
  )

  expect_error(
    prep_check_meta_names(data.frame(VAR_NAMES = 1, DATA_TYPE = 2,
                                     MISSING_LIST = 3), "TEXCHNICAL",
                          character.only = TRUE),
    regexp = paste("Error regarding argument .+level.+: .+arg.+ should be",
                   "one of .+COMPATIBILITY.+, .+REQUIRED.+, .+RECOMMENDED.+,",
                   ".+OPTIONAL.+, .+TECHNICAL.+"),
    perl = TRUE
  )

  expect_error(
    prep_check_meta_names(meta_data = list(VAR_NAMES = 1, DATA_TYPE = 2,
                                           MISSING_LIST = 3), "TECHNICAL",
                          character.only = TRUE),
    regexp = paste(".+meta_data.+",
                   "is not a data frame"),
    perl = TRUE
  )

  expect_warning(
    prep_check_meta_names(
      data.frame(
        VAR_NAMES = 1, DATA_TYPE = 2, MISSING_LIST = 3,
        LABEL = "LABEL", VALUE_LABELS = "VALUE_LABELS",
        JUMP_LIST = "JUMP_LIST", HARD_LIMITS = "HARD_LIMITS",
        GROUP_VAR_OBSERVER = "GROUP_VAR_OBSERVER",
        GROUP_VAR_DEVICE = "GROUP_VAR_DEVICE",
        TIME_VAR = "TIME_VAR",
        STUDY_SEGMENT = "STUDY_SEGMENT",
        PART_VAR = "PART_VAR",
        DETECTION_LIMITS = "DETECTION_LIMITS", SOFT_LIMITS = "SOFT_LIMITS",
        CONTRADICTIONS = "CONTRADICTIONS", DISTRIBUTION = "DISTRIBUTION",
        DECIMALS = "DECIMALS", VARIABLE_ROLE = "VARIABLE_ROLE",
        DATA_ENTRY_TYPE = "DATA_ENTRY_TYPE",
        END_DIGIT_CHECK = "END_DIGIT_CHECK",
        VARIABLE_ORDER = "VARIABLE_ORDER", LONG_LABEL =
          "LONG_LABEL", recode = "recode", VARNAMES = 19
      ),
      REQUIRED
    ),
    regexp = paste(
      "Found the following addtional metadata columns,",
      "which look like typos of defined names: .+VARNAMES.+ -> .+VAR_NAMES.+"
    ),
    perl  = TRUE
  )
})

#' Create a dynamic dimension related page for the report
#'
#' @param report [dataquieR_resultset2] a `dq_report2` report
#' @param use_plot_ly [logical] use `plotly`, if available.
#' @param template [character] template to use for the `dq_report2` report.
#' @param block_load_factor [numeric] multiply size of parallel compute blocks
#'                                    by this factor.
#' @param repsum the `dataquieR` summary, see [summary()] and [dq_report2()]
#' @param dir [character] output directory for potential `iframes`.
#'
#' @return list of arguments for `append_single_page()` defined locally in
#'   `util_generate_pages_from_report()`.
#'
#' @keywords internal
util_html_for_dims <- function(report, use_plot_ly, template,
                               block_load_factor, repsum, dir) {
  util_setup_rstudio_job("Page generation: Indicator View",
                         n = ncol(report))

  meta_data <- attr(report, "meta_data")
  label_col <- attr(report, "label_col")

  util_stop_if_not(inherits(report, "dataquieR_resultset2"))

  # find which dimensions are included in the report
  dims_in_rep <-
    unique(vapply(strsplit(colnames(report), "_"), `[[`, 1,
                  FUN.VALUE = character(1)))

  dims_in_rep <- util_sort_by_order(dims_in_rep, names(dims))

  # match the functions to the dimensions
  clls_in_rp <- lapply(setNames(nm = dims_in_rep),
                       function(prefix)
                         colnames(report)[startsWith(colnames(report),
                                                     prefix)])


  clls_in_rep_2 <-
    factor(vapply(strsplit(colnames(report), '_', fixed = TRUE), `[[`, 1,
                  FUN.VALUE = character(1)), names(dims), ordered = TRUE)
  sorted_clls <- colnames(report)[order(order(clls_in_rep_2))]

  i <- 0
  n <- ncol(report)

  # already done, before in util_generate_pages_from_report
  # suppressWarnings(
  #   parallelMap::parallelExport("report", "use_plot_ly", "template"))
  # parallelMap::parallelLibrary(utils::packageName())
  cores <- util_get_cores_safe()
  if (!is.numeric(cores) || !util_is_integer(cores) || !is.finite(cores)) {
    cores <- 1
  }
  block_size <- block_load_factor * cores
  all_pages <- list()
  nblocks <- ceiling(ncol(report) / block_size)
  for (cur_block in seq_len(nblocks) - 1) { # create the single variable pages
    block_indices <- seq(1 + cur_block * block_size,
                         min(cur_block * block_size + block_size,
                             ncol(report)))
    clls_in_chunk <- colnames(report)[block_indices]
    clls_in_chunk <- clls_in_chunk[!is.na(clls_in_chunk)]
    progress(i/n * 100)

    progress_msg("Page generation", sprintf("Single Indicators %s",
                                            paste(sQuote(clls_in_chunk),
                                                  collapse = ", ")))
    results <- lapply(setNames(nm = clls_in_chunk), function(nm) {
      report[, nm, as_raw = TRUE]
    })

    combined_list <- Map(function(results, clls) list(results = results, clls = clls), results = results, clls = clls_in_chunk)

    chunk_of_pages <- util_par_lapply_lb(
#      chunk.size = cores,
      X = combined_list,
      fun = util_isolate_function(function(cml, ...) {
        cml$results[vapply(cml$results, is.raw, FUN.VALUE = logical(1))] <-
          lapply(cml$results[vapply(cml$results, is.raw, FUN.VALUE = logical(1))], util_decompress) # TODO: avoid re-compression
          #filter the results to keep only non-ssi variables
          cml$results <- # remove results for SSI part of the report
           cml$results[vapply(lapply(cml$results, attr, COMPUTED_VARIABLE_ROLE),
                     is.null, FUN.VALUE = logical(1))]

         util_html_for_dim(results = cml$results,
                                      cll = cml$clls, ...)
      }),
      repsum = repsum,
      function_alias_map =
        attributes(attributes(report)$matrix_list)$function_alias_map,
      meta_data = meta_data,
      label_col = label_col,
      use_plot_ly = use_plot_ly,
      dir = dir,
      template = template,
      wd = getwd()
    )
    all_pages <- c(all_pages, chunk_of_pages)
    i <- i + length(clls_in_chunk)
    progress(i/n * 100)
  }
  all_pages <- all_pages[vapply(all_pages, length,
                                          FUN.VALUE = integer(1)) > 0]
  all_pages
}

util_html_for_dim <- function(results, cll, repsum, function_alias_map,
                              meta_data, label_col, use_plot_ly, dir,
                              template, wd) {
  if (length(results) == 0) {
    return(NULL)
  }
  setwd(wd)

  cur_dm <- sub("_.*$", "", cll)

  if (cur_dm %in% c("acc")) {  # acc is in separate files
    fname <- paste0("dim_", cur_dm, "_", cll, ".html") # add the function name for accuracy
  } else {
    fname <- paste0("dim_", cur_dm, ".html") # define the file name, one file per dimension if not in accuracy
  }

  .fn <- paste0("dim_", cur_dm, "_", cll, ".RDS") # add the function name for accuracy

  if (getOption("dataquieR.resume_print", dataquieR.resume_print_default)) {
    # TODO: [html_files_should_also_work_alone] Check, if we have the html file, already, see util_create_page_file
    # do not re-compute, then
  }

  if (getOption("dataquieR.resume_print", dataquieR.resume_print_default)
      && file.exists(.fn)) {
    cached <- try(readRDS(.fn), silent = TRUE)
    if (!util_is_try_error(cached)) {
      return(cached)
    }
  }

  call_names <- NULL

  drop_down <- dims[[cur_dm]] # fetches the name of the content for the first-level drop-down menu
  fkt <- util_cll_nm2fkt_nm(cll, function_alias_map = function_alias_map)
  sumplot <- plot(repsum, filter = call_names == cll, dont_plot = TRUE,
                  stratify_by = "indicator_metric",
                  disable_plotly = !use_plot_ly)

  is_single_var <- FALSE
  # indicator overview page
  # take all results for the current function
  all_of_f <- util_combine_res(results)

  # try to also combine preferred_summary_slots
  resn <- intersect(resnames.dataquieR_resultset2(results),
                    preferred_summary_slots)
  resn <- setdiff(resn, resnames.dataquieR_resultset2(all_of_f))
  if (!any(resnames.dataquieR_resultset2(all_of_f) %in%
           preferred_summary_slots)) {
    # we have already a summary slot displayed --
    # TODO: Find a way to define, if we want to see two tables, really
    if (length(resn) > 0) { # we have additional summaries to append
      resn <- head(resn, 1) # take the top-prio slot from the remaining ones
      all_of_f1 <- util_combine_res(lapply(results, `[`, resn))
      all_of_f[[cll]][[resn]] <- all_of_f1[[cll]][[resn]]
    }
  }
  # create a page by iterating over all_of_f entries and their respective names
  # for each element the util_pretty_print function is called, converting each dataquieR result to an htmltools compatible object
  # the do.call(htmltools::tagList) converts the output as an htmltools compatible tag list ("as.tagList.list")
  # we have to unname the list of tags to prevent taglists from generating named html attributes
  page <- do.call(htmltools::tagList, unname(mapply(dqr = all_of_f,
                                                    nm = names(all_of_f),
                                                    FUN =
                                                      util_pretty_print,
                                                    MoreArgs =
                                                      list(
                                                        meta_data =
                                                          meta_data,
                                                        label_col =
                                                          label_col,
                                                        use_plot_ly =
                                                          use_plot_ly,
                                                        is_single_var =
                                                          FALSE,
                                                        dir = dir),
                                                    SIMPLIFY = FALSE
  )))
  # check that the length of the page is positive and that it does not only contain null
  if ((length(page) && !all(vapply(page, is.null, FUN.VALUE = logical(1))))) {
    if ("link" %in% unique(unlist(lapply(page, names)))) { # if we have page navigation links
      # extract link information for the page navigation menu (created by util_pretty_print)
      all_links <- lapply(page, `[[`, "link")
      # add links to the page navigation menu
      all_links <- all_links[vapply(all_links, length, # remove empty links
                                    FUN.VALUE = integer(1)) != 0]
      all_links <- lapply(all_links, htmltools::tags$li) # add li to each tag to create the items
      all_links <- do.call(htmltools::tagList, all_links) # convert the list to a tag list, which is htmltools compatible
      page_menu <- util_float_index_menu(object = all_links) # create the menu
      # remove the original links (as created by util_pretty_print in the wrong places)
      page <- lapply(page, function(p) {
        p$link <- NULL
        p
      })
      # call template
      page <- htmltools::htmlTemplate(
        system.file("templates", template, "single_indicator_with_menu.html",
                    package = utils::packageName()),
        page_menu = page_menu,
        cll = cll,
        fkt = fkt,
        title = util_alias2caption(cll, long = TRUE),
        description = htmltools::HTML(util_function_description(fkt)),
        page = page,
        online_ref = util_online_ref(fkt),
        items_indicators = util_get_concept_links(fkt),
        sumplot = sumplot
      )
      # the "dim_" prefix is required because otherwise windows will ignore a file called con.html confusing it with a special device con:
    } else { # for pages without navigation menu, not used
      page <- htmltools::htmlTemplate( #
        system.file("templates", template, "single_indicator.html",
                    package = utils::packageName()),
        cll = cll,
        fkt = fkt,
        title = util_alias2caption(cll, long = TRUE),
        description = util_function_description(fkt),
        page = page,
        online_ref = util_online_ref(fkt),
        sumplot = sumplot
      )
    }
    #Check implementations and if there is a menu_location_report info, use it to distribute results
    impls <- util_get_concept_info("implementations")
    if(fkt %in% impls$function_R &&
      impls[impls$function_R == fkt, "menu_location_report", drop = TRUE] != "") {
      drop_down <- impls[impls$function_R == fkt, "menu_location_report", drop = TRUE]
    }

    title <- util_alias2caption(cll, long = TRUE);
    title <- util_attach_attr(title,
                     alternative_names =
                       as.vector(unlist(c(cll,
                                        util_alias2caption(cll,
                                                           long = FALSE)))))

    r <- list( drop_down,
               title,
               fname,
               page)
    if (getOption("dataquieR.resume_print", dataquieR.resume_print_default)) {
      saveRDS(r, file = .fn, compress = "xz")
    }
    r
  }

}

#' Check for package updates
#'
#' @param beta [logical] check for beta version too
#' @param deps [logical] check for missing (optional) dependencies
#' @param ask [logical] ask for updates
#'
#' @return `invisible(NULL)`
#' @export
prep_check_for_dataquieR_updates <- function(beta = FALSE,
                                             deps = TRUE,
                                             ask = interactive()) { # nocov start
  util_expect_scalar(beta, check_type = is.logical)
  util_expect_scalar(deps, check_type = is.logical)
  util_expect_scalar(ask, check_type = is.logical)
  pn <- packageName()
  util_message("Looking for updates of %s. Currently installed: %s",
               sQuote(pn),
               sQuote(packageVersion(pn)))
  if (beta) {
    r <- list(
      QIHS = "https://packages.qihs.uni-greifswald.de//repository/ship-snapshot-r/"
    )
  } else {
    r <- list()
  }
  rlang::local_options(repos = c(
        r,
        getOption("repos")
    )
  )
  if (beta) {
    utils::update.packages(oldPkgs = pn, ask = ask, dependencies = deps,
                           type = "source")
  } else {
    utils::update.packages(oldPkgs = pn, ask = ask, dependencies = deps)
  }
  all_deps <- unique(trimws(unlist(strsplit(split = ",", fixed = TRUE, x =
                          unname(unlist(utils::packageDescription("dataquieR",
    fields = c("Depends", "Imports", "LinkingTo", "Suggests", "Enhances"))))))))
  all_deps <- gsub("\\s+", " ", all_deps)
  all_deps <- all_deps[!util_empty(all_deps)]
  all_deps <- all_deps[!startsWith(all_deps, "R ")]
  if (ask) {
    rlang::check_installed(all_deps,
                           sprintf("as (soft) dependencies of %s",
                                   dQuote(packageName())))
  } else {
    to_install <- !vapply(all_deps, rlang::is_installed,
                          FUN.VALUE=logical(1))
    if (rlang::is_installed("pak")) {
      cat("Installing using pak: ", dQuote(names(to_install[to_install])), "\n")
      pkg_install <- rlang::env_get(rlang::ns_env("pak"), "pkg_install")
      pkg_install(names(to_install[to_install]), ask = FALSE)
    } else {
      cat("Installing ", dQuote(names(to_install[to_install])), "\n")
      utils::install.packages(names(to_install[to_install]))
    }
  }
  invisible(NULL)
} # nocov end

#' Support function to identify the levels of a process variable with minimum
#' number of observations
#'
#' @description
#' utility function to subset data based on minimum number of observation per
#' level
#'
#' @details
#' This functions removes observations having fewer than `min_obs_in_subgroup`
#' distinct values in a group variable, e.g. blood pressure measurements
#' performed by an examiner having fewer than e.g. 50 measurements done. It
#' displays a warning, if samples/rows are removed and returns the modified
#' study data frame.
#'
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param group_vars [variable list] the name grouping variable
#' @param min_obs_in_subgroup [integer] optional argument if a "group_var" is
#'                                      used. This argument specifies the
#'                                      minimum no. of observations that is
#'                                      required to include a subgroup (level)
#'                                      of the "group_var" in the analysis.
#'                                      Subgroups with less observations are
#'                                      excluded. The default is 30.
#'
#' @return a data frame with:
#'   - a subsample of original data
#' @export
prep_min_obs_level <- function(study_data, group_vars, min_obs_in_subgroup) {
  util_expect_data_frame(study_data, keep_types = TRUE)
  if (missing(group_vars) || !is.character(group_vars)) {
    util_error("%s is required to be a character(1) argument.",
               dQuote("group_vars"), applicability_problem = TRUE)
  }

  if (length(group_vars) > 1) {
    util_message("Subsets based only on one variable possible.",
                 applicability_problem = TRUE)
    group_vars <- group_vars[1]
  }

  if (length(group_vars) != 1) {
    util_error("%s is required to name exactly one variable.",
                 dQuote("group_vars"), applicability_problem = TRUE)
  }

  if (!(group_vars %in% colnames(study_data))) {
    util_error("%s is not a variable.",
               paste(sQuote("group_vars"), "=", dQuote(group_vars)),
               applicability_problem = TRUE)
  }

  if (missing(min_obs_in_subgroup) || length(min_obs_in_subgroup) != 1 ||
      is.na(min_obs_in_subgroup)) {
    if (!.called_in_pipeline || !missing(min_obs_in_subgroup))
      util_message(
        c("argument %s was missing, not of length 1 or NA, setting to its",
          "default, 30"),
        dQuote("min_obs_in_subgroup"),
        applicability_problem = TRUE)
    min_obs_in_subgroup <- 30
  }

  X <- util_table_of_vct(study_data[[group_vars]])

  if (suppressWarnings(!is.finite(min(X[["Freq"]])))) {
    util_error("For %s, observations cannot be counted.",
               paste(sQuote("group_vars"), "=", dQuote(group_vars)))
  }

  # too few observations in >1 level of group_vars
  if (min(X[, 2]) < min_obs_in_subgroup) {
    critical_levels <- levels(X$Var1)[X$Freq < min_obs_in_subgroup]
    util_message(
      "The following levels: %s have < %d observations and are disregarded",
      paste0(dQuote(critical_levels), collapse = ", "),
      min_obs_in_subgroup,
      applicability_problem = FALSE
    )
    subsdf <- study_data[!(study_data[[group_vars]] %in% critical_levels), ]
  } else {
    subsdf <- study_data
  }


  return(subsdf)
}

#' Utility function parallel version of `purrr::pmap`
#'
#' Parallel version of `purrr::pmap`.
#'
#'
#' @param .l [data.frame] with one call per line and one function argument
#'                        per column
#' @param .f [`function`] to call with the arguments from `.l`
#' @param ... additional, static arguments for calling `.f`
#'
#' @param cores number of cpu cores to use or a (named) list with arguments for
#'              [parallelMap::parallelStart] or NULL, if parallel has already
#'              been started by the caller.
#' @param use_cache [logical] set to FALSE to omit re-using already distributed
#'                            study- and metadata on a parallel cluster
#'
#' @seealso `purrr::pmap`
#' @seealso [Stack Overflow post](https://stackoverflow.com/a/47575143)
#'
#' @author [Aurèle](https://stackoverflow.com/users/6197649)
#' @author S Struckmann
#'
#' @return [list] of results of the function calls
#'
#' @family process_functions
#' @concept reporting
#' @noRd
util_par_pmap <- function(.l, .f, ...,
                          cores = list(mode = "socket",
                                       cpus = util_detect_cores(),
                                       logging = FALSE,
                                       load.balancing = TRUE),
                          use_cache = FALSE) {
  if (!is.null(cores)) {
    if (inherits(cores, "list")) {
      suppressMessages(do.call(parallelMap::parallelStart, cores))
    } else {
      suppressMessages(parallelMap::parallelStart("socket", cpus = cores,
                                                  logging = FALSE,
                                                  load.balancing = TRUE))
    }
    on.exit({Sys.sleep(2);suppressMessages(parallelMap::parallelStop());}, add = TRUE) # whyever, rstudio needs these two seconds, it hangs, otherwise.
  } # TODO: else it fails with Error in `assign(n, get(n, envir = env), envir = ee)`: cannot add bindings to a locked environment -- guess, the same problem I had in util_evaluate_calls see ''if (parallelMap::parallelGetOptions()$settings$mode == "local") {'' therein
  more_args <- list(...)
  if ("meta_data" %in% names(more_args)) {
    meta_data <- more_args[["meta_data"]]
    if (use_cache &&
        !all(unlist(parallelMap::parallelMap(exists, "meta_data")))) {
      suppressWarnings(parallelMap::parallelExport("meta_data"))
      more_args[["meta_data"]] <- NULL
    }
  }
  if ("study_data" %in% names(more_args)) {
    study_data <- more_args[["study_data"]]
    if (use_cache &&
        !all(unlist(parallelMap::parallelMap(exists, "study_data")))) {
      suppressWarnings(parallelMap::parallelExport("study_data"))
      more_args[["study_data"]] <- NULL
    }
  }
  do.call(
    parallelMap::parallelMap,
    c(.l, list(
      fun = .f, more.args = more_args, simplify = FALSE, use.names = FALSE,
      show.info = FALSE, impute.error = identity
    ))
  )
  #  do.call(
  #    parallel::mcmapply,
  #    c(.l, list(FUN = .f, MoreArgs = list(...), SIMPLIFY = FALSE,
  #         mc.cores = mc.cores))
  #  )
}

test_that("util_par_pmap works", {
  skip_on_cran() # slow and meant for parallel processing
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  plan <- dplyr::tribble(
    ~ x, ~ y,
    rnorm(1:10000), rnorm(1:10000),
    rnorm(1:10000), rnorm(1:10000),
    rnorm(1:10000), rnorm(1:10000),
    rnorm(1:10000), rnorm(1:10000),
    rnorm(1:10000), rnorm(1:10000),
    rnorm(1:10000), rnorm(1:10000),
    rnorm(1:10000), rnorm(1:10000),
    rnorm(1:10000), rnorm(1:10000),
    rnorm(1:10000), rnorm(1:10000),
    rnorm(1:10000), rnorm(1:10000)
  )
  r <- unlist(dataquieR:::util_par_pmap(plan, cor, cores = 1),
              recursive = FALSE)
  expect_vector(r, numeric(0), 10)
  expect_equal(ignore_attr = TRUE, mapply(plan$x, plan$y, FUN = cor),
                    r)
  myCor <- function(..., meta_data, study_data) {
    testthat::expect_true(exists("meta_data"))
    testthat::expect_true(exists("study_data"))
    cor(...)
  }
  s <- unlist(dataquieR:::util_par_pmap(.l = plan,
                                        .f =  myCor,
                                        meta_data = cars,
                                        study_data = cars,
                                        cores = list(mode = "socket",
                                                     cpus = 1,
                                                     logging = FALSE,
                                                     load.balancing = TRUE),
                                        use_cache = TRUE),
              recursive = FALSE)
  expect_equal(s, r)
})

#' Utility function to compute model-based ICC depending on the (statistical) data type
#'
#' @description
#' This function is still under construction. It is designed to run for any
#' statistical data type as follows:
#'   - Variables with only two distinct values will be modeled by mixed effects
#'     logistic regression.
#'   - Nominal variables will be transformed to binary variables. This can be
#'     user-specified using the metadata columns `RECODE_CASES` and/or
#'     `RECODE_CONTROL`. Otherwise, the most frequent category will be assigned
#'     to cases and the remaining categories to control. As for other binary
#'     variables, the ICC will be computed using a mixed effects logistic
#'     regression.
#'   - Ordinal variables will be analyzed by linear mixed effects models, if
#'     every level of the variable has at least as many observations as
#'     specified in the argument `cut_off_linear_model_for_ord`. Otherwise, the
#'     data will be modeled by a mixed effects ordered regression, if the
#'     package `ordinal` is available.
#'   - Metric variables with integer values are analyzed by linear mixed
#'     effects models.
#'   - For variables with data type `float`, the existing implementation
#'     `acc_varcomp` is called, which also uses linear mixed effects models.
#'
#' [Indicator]
#'
#' @export
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars  [variable] the name of the measurement variable
#' @param group_vars [variable] the name of the examiner, device or
#'                              reader variable
#' @param co_vars    [variable list] a vector of covariables, e.g. age and sex,
#'                              for adjustment
#' @param min_obs_in_subgroup [integer] from=0. This optional argument specifies
#'                              the minimum number of observations that is
#'                              required to include a subgroup (level) of the
#'                              `group_var` in the analysis. Subgroups with less
#'                              observations are excluded.
#' @param min_subgroups [integer] from=0. This optional argument specifies
#'                              the minimum number of subgroups (level) of the
#'                              `group_var` that is required to run the
#'                              analysis. If there are less subgroups, the
#'                              analysis is not conducted.
#' @param cut_off_linear_model_for_ord [integer] from=0. This optional argument
#'                              specifies the minimum number of observations for
#'                              individual levels of an ordinal outcome
#'                              (`resp_var`) that is required to run a linear
#'                              mixed effects model instead of a mixed effects
#'                              ordered regression (i.e., a cut-off value above
#'                              which linear models are considered a good
#'                              approximation). The argument can be set to
#'                              `NULL` if ordered regression models are
#'                              preferred for ordinal data in any case.
#' @param threshold_value Deprecated.
#' @return The function returns two data frames, 'SummaryTable' and
#'         'SummaryData', that differ only in the names of the columns.
#'
#' @details
#' Not yet described
#'
acc_varcomp <- function(resp_vars = NULL,
                        group_vars = NULL,
                        co_vars = NULL,
                        study_data,
                        label_col,
                        item_level = "item_level",
                        min_obs_in_subgroup = 10,
                        min_subgroups = 5,
                        cut_off_linear_model_for_ord = 10,
                        threshold_value = lifecycle::deprecated(),
                        meta_data = item_level,
                        meta_data_v2) {
  # preps and checks -----------------------------------------------------------

  if (lifecycle::is_present(threshold_value)) {

    # Signal the deprecation to the user
    lifecycle::deprecate_stop("2.5.0",
                              'dataquieR::acc_varcomp(threshold_value)',
                              details =
                                c("use the grading rules instaed, see",
                                  "GRADING_RULESET and online documentation"))

  }

  util_maybe_load_meta_data_v2()

  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata = TRUE)

  util_correct_variable_use("resp_vars",
                            allow_all_obs_na = FALSE,
                            need_scale = "!na")

  util_correct_variable_use("group_vars",
                            allow_na = TRUE,
                            allow_all_obs_na = TRUE,
                            need_type = "!float",
                            need_scale = "nominal | ordinal")

  if (length(unique(setdiff(group_vars, NA_character_))) != 1) {
    util_error("Missing entries not allowed in argument %s",
               "group_vars",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = .called_in_pipeline) #TODO: clean up, find a way not to call this function from pipeline at all if there is no group_vr
  }
 #util_correct_variable_use("group_vars",
 #                           allow_all_obs_na = FALSE,
 #                           need_type = "!float",
 #                           need_scale = "nominal | ordinal")

  util_correct_variable_use("co_vars",
                            overwrite = TRUE,
                            remove_not_found = TRUE,
                            allow_na = TRUE,
                            allow_more_than_one = TRUE,
                            allow_null = TRUE)
  co_vars <- co_vars[which(!util_empty(co_vars))]
  if (is.null(co_vars)) {
    co_vars <- character(0)
  }

  util_expect_scalar(min_obs_in_subgroup,
                     error_message = "min_obs_in_subgroup needs to be integer > 0",
                     check_type = util_is_numeric_in(min = 0,
                                                     whole_num = TRUE,
                                                     finite = TRUE))
  util_expect_scalar(min_subgroups,
                     error_message = "min_subgroups needs to be integer > 0",
                     check_type = util_is_numeric_in(min = 0,
                                                     whole_num = TRUE,
                                                     finite = TRUE))

  util_expect_scalar(cut_off_linear_model_for_ord,
                     error_message =
                       "cut_off_linear_model_for_ord needs to be integer > 0",
                     allow_null = TRUE,
                     check_type = util_is_numeric_in(min = 0,
                                                     whole_num = TRUE,
                                                     finite = TRUE))

  ds1 <- ds1[, c(resp_vars, group_vars, co_vars), drop = FALSE]
  ds1 <- ds1[complete.cases(ds1), , drop = FALSE]

  # ensure that the grouping variable has the specified minimum number of levels
  check_df <- util_table_of_vct(ds1[[group_vars]])
  if (length(check_df[, 1]) < min_subgroups) {
    util_error("%d < %d levels in %s. Will not compute ICCs for %s.",
               length(check_df[, 1]),
               min_subgroups,
               dQuote(group_vars),
               dQuote(resp_vars),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  # ensure that included levels of the grouping variable have the specified
  # minimum number of observations
  critical_levels <- levels(check_df$Var1)[check_df$Freq <
                                             min_obs_in_subgroup]
  if (length(critical_levels) > 0) {
    util_message("Levels %s were excluded due to less than %d observations.",
                 paste0(c(vapply(head(critical_levels, 10), dQuote, ""),
                          if (length(critical_levels) <= 10)
                            character(0) else "..."),
                        collapse = ", "),
                 min_obs_in_subgroup,
                 applicability_problem = FALSE)
    # exclude levels with too few observations
    ds1 <- ds1[!(ds1[[group_vars]] %in% critical_levels), ]
    levels(ds1[[group_vars]])[
      which(levels(ds1[[group_vars]]) %in% critical_levels)] <- NA
  }
  ds1 <- ds1[complete.cases(ds1), , drop = FALSE]
  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.")
  }

  rvs_bin_note <- NULL

  # check first if there is a user-specified recoding
  if (("RECODE_CASES" %in% colnames(meta_data) &&
       !util_empty(meta_data[meta_data[[label_col]] == resp_vars, RECODE_CASES])) ||
      ("RECODE_CONTROL" %in% colnames(meta_data) &&
       !util_empty(meta_data[meta_data[[label_col]] == resp_vars, RECODE_CONTROL]))) {
    rvs_bin <- util_dichotomize(
      study_data = ds1[, resp_vars, drop = FALSE],
      meta_data = meta_data,
      label_col = label_col)
    rvs_bin_note <- attr(rvs_bin, "Dichotomization")[[resp_vars]]
    ds1[[resp_vars]] <- unlist(rvs_bin)
    ds1 <- ds1[complete.cases(ds1), , drop = FALSE]
  }

  # ensure that the response variable is not constant
  var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])
  if (var_prop$NDistinct < 2) {
    util_error("The response variable is constant after data preparation.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  # compute model-based ICC depending on the statistical data type -------------
  # prepare model formula
  if (length(co_vars) == 0) {
    .co_vars <- "1"
  } else {
    .co_vars <- util_bQuote(co_vars)
  }
  fmla <- as.formula(paste0(
    util_bQuote(resp_vars), " ~ ",
    paste0(c(.co_vars, paste0("(1|", util_bQuote(group_vars), ")")),
           collapse = "+")))

  var_scale <- meta_data[[SCALE_LEVEL]][meta_data[[label_col]] == resp_vars]
  var_dtype <- meta_data[[DATA_TYPE]][meta_data[[label_col]] == resp_vars]


  ## if loop
  if (var_prop$NDistinct == 1) {
    util_error("The response variable is constant.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  } else if (var_prop$NDistinct == 2 && (
    var_scale %in% c("ordinal", "ratio", "interval"))) { ## binary response ----------------------------
    check_df <- util_table_of_vct(ds1[[resp_vars]])
    if (any(check_df$Freq < min_obs_in_subgroup)) {
      util_error("Too few observations with different values in %s.",
                 dQuote(resp_vars),
                 applicability_problem = TRUE)
    }
    # code the most frequent value as 1, the other as 0
    check_df <- check_df[order(check_df$Freq, decreasing = TRUE), ]
    ds1[[resp_vars]] <- ifelse(ds1[[resp_vars]] == check_df$Var1[1], 1, 0)
    rvs_bin_note <- paste(
      paste("Cases (1):", check_df[1, 1]),
      paste("Control (0):", check_df[2, 1]),
      sep = ". ")
    # mixed effects logistic regression
    fit1 <- try(lme4::glmer(formula = fmla, data = ds1,
                            family = binomial,
                            control = lme4::glmerControl(optimizer = "bobyqa"),
                            nAGQ = 10))
    icc <- NA
    if (!inherits(fit1, "try-error")) {
      # extract variance of random effects
      v_tab <- as.data.frame(lme4::VarCorr(fit1))
      # calculate icc
      icc <- round(v_tab$vcov / (v_tab$vcov + pi^2/3), 3)
    }
    check_gr <- util_table_of_vct(ds1[[group_vars]])
    res_icc <- data.frame(
      Variables = resp_vars,  #TODO: add the information somewhere else
      #   Category = check_df$Var1[1],
      Object = group_vars,
      Model.Call = Reduce(paste, deparse(fmla)),
      ICC = icc,
      Class.Number = nrow(check_gr),
      Mean.Class.Size = mean(check_gr$Freq),
      Median.Class.Size = median(check_gr$Freq),
      Min.Class.Size = min(check_gr$Freq),
      Max.Class.Size = max(check_gr$Freq),
      convergence.problem = inherits(fit1, "try-error")
    )
  } else if (var_scale == SCALE_LEVELS$NOMINAL) { ## nominal response ----------
    count_nom <- util_table_of_vct(ds1[[resp_vars]])
    count_nom <- count_nom[order(count_nom[, 2], decreasing = TRUE), ]
    min_obs_in_cat <- 2 * length(levels(ds1[[group_vars]]))
    count_nom$below_thresh <- count_nom[, 2] < min_obs_in_cat
    # Nominal response variables will be transformed to binary variables.
    if (!(RECODE_CASES %in% colnames(meta_data))) {
      meta_data[[RECODE_CASES]] <- ""
    }
    if (!(RECODE_CONTROL %in% colnames(meta_data))) {
      meta_data[[RECODE_CONTROL]] <- ""
    }
    if (util_empty(meta_data[[RECODE_CASES]][meta_data[[label_col]] == resp_vars]) &
        util_empty(meta_data[[RECODE_CONTROL]][meta_data[[label_col]] == resp_vars])) {
      # If the recoding is not defined in the metadata, dataquieR will use the
      # most frequent category as 'cases', the remaining categories as
      # 'control'. If there were too few observations in the most frequent
      # category for the 'cases', we will include further categories until we
      # reach the lower limit specified by 'min_obs_in_cat' (but we stop it
      # before all categories are being combined into one group).
      ind_cases <- 1
      n_cases <- count_nom[1, 2]
      while (n_cases < min_obs_in_cat & max(ind_cases) < nrow(count_nom) - 1) {
        ind_cases <- seq_len(max(ind_cases) + 1)
        n_cases <- sum(count_nom[ind_cases, 2])
      }
      meta_data[[RECODE_CASES]][meta_data[[label_col]] == resp_vars] <-
        paste(count_nom[ind_cases, 1], collapse = " | ")
    }
    rvs_bin <- util_dichotomize(
      study_data = ds1[, resp_vars, drop = FALSE],
      meta_data = meta_data,
      label_col = label_col)
    rvs_bin_note <- attr(rvs_bin, "Dichotomization")[[resp_vars]]
    ds1[[resp_vars]] <- unlist(rvs_bin)
    ds1 <- ds1[complete.cases(ds1), ]
    # ensure that the response variable is not constant
    var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])
    if (var_prop$NDistinct < 2) {
      util_error("The response variable is constant (after data preparation).",
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    }
    # ensure that the response variable is binary
    if (var_prop$NDistinct != 2) {
      util_error("The response variable is not binary.",
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    }
    # ensure that there are enough observations for the model
    # count_bin <- util_table_of_vct(ds1[[resp_vars]])
    # if (any(count_bin[, 2] < min_obs_in_cat)) {
    #   util_error("Not enough observations per category (after data preparation).")
    # }
    count_bin <- util_table_of_vct(ds1[[group_vars]])
    if (any(count_bin[, 2] < min_obs_in_subgroup)) {
      util_error("Not enough observations per group (after data preparation).")
    }
    fit1 <- try(lme4::glmer(formula = fmla, data = ds1,
                            family = "binomial",
                            control = lme4::glmerControl(optimizer = "bobyqa"),
                            nAGQ = 10))
    icc <- NA
    if (!inherits(fit1, "try-error")) {
      # extract variance of random effects
      v_tab <- as.data.frame(lme4::VarCorr(fit1))
      # calculate icc
      icc <- round(v_tab$vcov / (v_tab$vcov + pi^2/3), 3)
    }
    res_icc <- data.frame(
      Variables = resp_vars,
      Object = group_vars,
      Model.Call = Reduce(paste, deparse(fmla)),
      ICC = icc,
      Class.Number = nrow(count_bin),
      Mean.Class.Size = mean(count_bin$Freq),
      Median.Class.Size = median(count_bin$Freq),
      Min.Class.Size = min(count_bin$Freq),
      Max.Class.Size = max(count_bin$Freq),
      convergence.problem = inherits(fit1, "try-error")
    )
  } else if (var_scale == SCALE_LEVELS$ORDINAL && var_prop$NDistinct > 2) { ## ordinal response ----------
    # Ordinal response variables will either be analyzed as integer responses
    # or using the 'ordinal' package.
    count_ord <- util_table_of_vct(ds1[[resp_vars]])
    if (!is.null(cut_off_linear_model_for_ord) &&
        all(count_ord[, 2] >= cut_off_linear_model_for_ord)) {
      ds1[[resp_vars]] <- as.numeric(ds1[[resp_vars]]) - 1
      fit1 <- try(lme4::lmer(formula = fmla, data = ds1, REML = TRUE))
      icc <- NA
      if (!inherits(fit1, "try-error")) {
        # extract variance
        v_tab <- as.data.frame(lme4::VarCorr(fit1, comp = "Variance"))
        v_tab <- v_tab$vcov
        # calculate icc
        icc <- round(v_tab[1] / (v_tab[1] + v_tab[2]), 3)
      }
      check_gr <- util_table_of_vct(ds1[[group_vars]])
      res_icc <- data.frame(
        Variables = resp_vars,
        Object = group_vars,
        Model.Call = Reduce(paste, deparse(fmla)),
        ICC = icc,
        Class.Number = nrow(check_gr),
        Mean.Class.Size = mean(check_gr$Freq),
        Median.Class.Size = median(check_gr$Freq),
        Min.Class.Size = min(check_gr$Freq),
        Max.Class.Size = max(check_gr$Freq),
        convergence.problem = inherits(fit1, "try-error")
      )
    } else {
      util_ensure_suggested("ordinal")
      fit1 <- try(ordinal::clmm(formula = fmla, data = ds1,
                                Hess = TRUE, nAGQ = 10))
      icc <- NA
      if (!inherits(fit1, "try-error")) {
        # extract variance of random effects (intercept)
        re_var <- as.numeric(lme4::VarCorr(fit1, comp = "Variance"))
        # calculate icc
        icc <- round(re_var / (re_var + pi^2/3), 3)
      }
      check_gr <- util_table_of_vct(ds1[[group_vars]])
      res_icc <- data.frame(
        Variables = resp_vars,
        Object = group_vars,
        Model.Call = Reduce(paste, deparse(fmla)),
        ICC = icc,
        Class.Number = nrow(check_gr),
        Mean.Class.Size = mean(check_gr$Freq),
        Median.Class.Size = median(check_gr$Freq),
        Min.Class.Size = min(check_gr$Freq),
        Max.Class.Size = max(check_gr$Freq),
        convergence.problem = inherits(fit1, "try-error")
      )
    }
  } else if (var_scale == SCALE_LEVELS$RATIO && var_prop$NDistinct > 2) {
    if( var_dtype == DATA_TYPES$INTEGER) { #TODO: extract code from here
      ## integer values, variable on interval or ratio scale ---------------------
      # linear model with random effects
      fit1 <- try(lme4::lmer(formula = fmla, data = ds1, REML = TRUE))
      # other options:
      # - generalized linear model with random effects using the poisson
      # distribution
      # fit2 <- try(lme4::glmer(formula = fmla, data = ds1, family = "poisson",
      #                         nAGQ = 10))
      # - generalized linear model with random effects using the
      # negative binomial distribution
      # fit3 <- try(lme4::glmer.nb(formula = fmla, data = ds1, nAGQ = 10))
      icc <- NA
      if (!inherits(fit1, "try-error")) {
        # extract variance
        v_tab <- as.data.frame(lme4::VarCorr(fit1, comp = "Variance"))
        v_tab <- v_tab$vcov
        # calculate icc
        icc <- round(v_tab[1] / (v_tab[1] + v_tab[2]), 3)
      }
      check_gr <- util_table_of_vct(ds1[[group_vars]])
      res_icc <- data.frame(
        Variables = resp_vars,
        Object = group_vars,
        Model.Call = Reduce(paste, deparse(fmla)),
        ICC = icc,
        Class.Number = nrow(check_gr),
        Mean.Class.Size = mean(check_gr$Freq),
        Median.Class.Size = median(check_gr$Freq),
        Min.Class.Size = min(check_gr$Freq),
        Max.Class.Size = max(check_gr$Freq),
        convergence.problem = inherits(fit1, "try-error")
      )
    } else if(var_dtype == DATA_TYPES$FLOAT){
      if (length(co_vars) == 1 && co_vars == "1") {
        co_vars <- NULL
      }
      res_icc <- util_acc_varcomp(resp_vars = resp_vars, group_vars = group_vars,
                                  co_vars = co_vars,
                                  min_obs_in_subgroup = min_obs_in_subgroup,
                                  min_subgroups = min_subgroups,
                                  label_col = label_col,
                                  study_data = study_data, meta_data = meta_data)
    } else {
      util_error("No method implemented for scale level ratio and data types string or datetime, sorry.",
                 applicability_problem = TRUE)
    }
  } else if (var_scale == SCALE_LEVELS$INTERVAL && var_prop$NDistinct > 2){
    if (var_dtype == DATA_TYPES$INTEGER) {
      ## integer values, variable on interval or ratio scale ---------------------
      # linear model with random effects
      fit1 <- try(lme4::lmer(formula = fmla, data = ds1, REML = TRUE))
      # other options:
      # - generalized linear model with random effects using the poisson
      # distribution
      # fit2 <- try(lme4::glmer(formula = fmla, data = ds1, family = "poisson",
      #                         nAGQ = 10))
      # - generalized linear model with random effects using the
      # negative binomial distribution
      # fit3 <- try(lme4::glmer.nb(formula = fmla, data = ds1, nAGQ = 10))
      icc <- NA
      if (!inherits(fit1, "try-error")) {
        # extract variance
        v_tab <- as.data.frame(lme4::VarCorr(fit1, comp = "Variance"))
        v_tab <- v_tab$vcov
        # calculate icc
        icc <- round(v_tab[1] / (v_tab[1] + v_tab[2]), 3)
      }
      check_gr <- util_table_of_vct(ds1[[group_vars]])
      res_icc <- data.frame(
        Variables = resp_vars,
        Object = group_vars,
        Model.Call = Reduce(paste, deparse(fmla)),
        ICC = icc,
        Class.Number = nrow(check_gr),
        Mean.Class.Size = mean(check_gr$Freq),
        Median.Class.Size = median(check_gr$Freq),
        Min.Class.Size = min(check_gr$Freq),
        Max.Class.Size = max(check_gr$Freq),
        convergence.problem = inherits(fit1, "try-error")
      )
    } else if(var_dtype == DATA_TYPES$FLOAT){
      if (length(co_vars) == 1 && co_vars == "1") {
        co_vars <- NULL
      }
      res_icc <- util_acc_varcomp(resp_vars = resp_vars, group_vars = group_vars,
                                  co_vars = co_vars,
                                  min_obs_in_subgroup = min_obs_in_subgroup,
                                  min_subgroups = min_subgroups,
                                  label_col = label_col,
                                  study_data = study_data, meta_data = meta_data)
    } else {
      util_error("No method implemented for scale level interval and data types string or datetime, sorry.",
                 applicability_problem = TRUE)
    }
  }

  if (is.data.frame(res_icc)) {
    res_icc$ICC <- util_round_to_decimal_places(res_icc$ICC)
    res_icc$Mean.Class.Size <-
      util_round_to_decimal_places(res_icc$Mean.Class.Size)
    res_icc$Median.Class.Size <-
      util_round_to_decimal_places(res_icc$Median.Class.Size, 1)

    sumtab <- res_icc
    names(sumtab)[names(sumtab) == "ICC"] <- "ICC_acc_ud_loc"

    SummaryData <- res_icc

    SummaryData <- SummaryData[, setdiff(colnames(SummaryData), c("Model.Call"))]

    # TODO: How to return rvs_bin_note?
    return(list("SummaryTable" = sumtab,
                "SummaryData" = SummaryData))
  } else { # output created by acc_varcomp
    return(res_icc)
  }
}

test_that("varcomp works", {
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  # nominal outcome
  # meta_data[[SCALE_LEVEL]][meta_data[[LABEL]] == "EATING_PREFS_0"]
  res1 <- acc_varcomp(
    resp_vars = "EATING_PREFS_0",
    group_vars = "CENTER_0",
    study_data = study_data, meta_data = meta_data, label_col = LABEL)

  expect_false(inherits(res1, "try-error"))

  # ordinal outcome
  meta_data[[SCALE_LEVEL]][meta_data[[LABEL]] == "MEAT_CONS_0"] <- SCALE_LEVELS$ORDINAL
  res2 <- acc_varcomp(
    resp_vars = "MEAT_CONS_0",
    group_vars = "USR_SOCDEM_0",
    study_data = study_data, meta_data = meta_data, label_col = LABEL)

  expect_false(inherits(res2, "try-error"))

  skip_if_not_installed("ordinal")

  # integer
  # meta_data[[DATA_TYPE]][meta_data[[LABEL]] == "N_INJURIES_0"]
  res3 <- acc_varcomp(
    resp_vars = "N_INJURIES_0",
    group_vars = "USR_SOCDEM_0",
    study_data = study_data, meta_data = meta_data, label_col = LABEL)

  expect_false(inherits(res3, "try-error"))

  # float
  res4 <- acc_varcomp(
    resp_vars = "CRP_0",
    group_vars = "CENTER_0",
    study_data = study_data, meta_data = meta_data, label_col = LABEL)

  expect_false(inherits(res4, "try-error"))
})

test_that("acc_varcomp works without label_col", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  expect_error({
    res1 <-
      acc_varcomp(resp_vars = "v00014", study_data = study_data,
                  meta_data = meta_data)
  }, regexp =
    paste("Argument group_vars is NULL"),
    perl = TRUE
  )

  expect_error(
    res1 <-
      acc_varcomp(resp_vars = c("CRP_0", "SBP_0"), study_data = study_data,
                  meta_data = meta_data, group_vars = c("DEV_NO_0", "USR_BP_0"),
                  label_col = LABEL),
    regexp =
      sprintf(
        "(%s)",
  paste("Need exactly one element in argument resp_vars, got 2: .CRP_0, SBP_0.")
      ),
    perl = TRUE
  )

  suppressMessages(
    res1 <-
      acc_varcomp(resp_vars = c("DBP_0"), study_data = study_data,
                  meta_data = meta_data, group_vars = c("USR_BP_0"),
                  label_col = LABEL))

  expect_true(all(
    c("SummaryTable",
      "ScalarValue_max_icc",
      "ScalarValue_argmax_icc"
    ) %in% names(res1)))

  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE) - 783.112)), 0.1
  )

  expect_true(res1$ScalarValue_max_icc == 0.112)
  expect_true(res1$ScalarValue_argmax_icc == "DBP_0")

})

test_that("acc_varcomp works with label_col", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  expect_error({
      res1 <-
        acc_varcomp(resp_vars = "SBP_0", study_data = study_data,
                    label_col = LABEL,
                    meta_data = meta_data)
    },
    regexp = paste("Argument group_vars is NULL")
  )

  res1 <-
    acc_varcomp(resp_vars = "v00014", study_data = study_data,
                meta_data = meta_data, group_vars = "v00016")

  expect_true(all(
    c("SummaryTable",
      "ScalarValue_max_icc",
      "ScalarValue_argmax_icc"
    ) %in% names(res1)))

  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE) - 2082.221)), 0.1
  )

  expect_equal(res1$ScalarValue_max_icc, 0.021)
  expect_equal(res1$ScalarValue_argmax_icc, "v00014")


})

test_that("acc_varcomp works illegal min_obs_in_subgroup/min_subgroups", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  suppressMessages(expect_error(
    res1 <-
      acc_varcomp(resp_vars = c("DBP_0"), study_data = study_data,
                  meta_data = meta_data, group_vars = c("USR_BP_0"),
                  label_col = LABEL, min_obs_in_subgroup = "k",
                  min_subgroups = "x"),
    regexp =
      sprintf(
        "(%s|%s|%s)",
        paste("min_obs_in_subgroup needs to be integer > 0"),
        paste("Could not convert min_subgroups .+x.+ to a number.",
              "Set to standard value."),
        paste("Levels .+USR_559.+ were excluded",
              "due to fewer than 30 observations.")
      )
  ))

  suppressMessages(
    res1 <-
      acc_varcomp(resp_vars = c("DBP_0"), study_data = study_data,
                  meta_data = meta_data, group_vars = c("USR_BP_0"),
                  label_col = LABEL)
  )

  expect_true(all(
    c("SummaryTable",
      "ScalarValue_max_icc",
      "ScalarValue_argmax_icc"
    ) %in% names(res1)))

  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE) - 783.112)), 0.1
  )

  expect_true(res1$ScalarValue_max_icc == 0.112)
  expect_true(res1$ScalarValue_argmax_icc == "DBP_0")

})

test_that("acc_varcomp works without resp_vars", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  expect_error(
    res1 <-
      acc_varcomp(study_data = study_data,
                  meta_data = meta_data,
                  group_vars = c("v00016", "v00012")),
    regexp =
      paste("Argument resp_vars is NULL")

  )

  res1 <-
    suppressMessages(acc_varcomp(study_data = study_data,
                meta_data = meta_data,
                resp_vars = "v00004",
                group_vars = "v00016"))

  expect_true(all(
    c("SummaryTable",
      "ScalarValue_max_icc",
      "ScalarValue_argmax_icc"
    ) %in% names(res1)))

  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE) - 1932.009)), 0.1
  )

  expect_equal(res1$ScalarValue_max_icc, 0.008)
  expect_equal(res1$ScalarValue_argmax_icc, "v00004")


})

test_that("acc_varcomp stops on too few subgroups", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  expect_error(
    res1 <-
      acc_varcomp(resp_vars = "v00014",
                  study_data = study_data,
                  meta_data = meta_data,
                  group_vars = "v00016",
                  min_subgroups = 50),
    regexp = "5 . 50 levels in .+v00016.+ Will not compute ICCs for .+v00014.+."

  )

  # expect_length(res1$ScalarValue_argmax_icc, 0)
  # expect_equal(res1$ScalarValue_max_icc, -Inf)
  # expect_equal(nrow(res1$SummaryTable), 0)
  # expect_equal(ncol(res1$SummaryTable), 11)

})

#' Wrapper function to check for studies data structure
#'
#' @description
#' This function tests for unexpected elements and records, as well as duplicated identifiers and content.
#' The unexpected element record check can be conducted by providing the number of expected records or
#' an additional table with the expected records.
#' It is possible to conduct the checks by study segments or to consider only selected
#' segments.
#'
#' [Indicator]
#'
#' @inheritParams .template_function_indicator
#'
#' @return a [list] with
#'   - `DataframeTable`: data frame with selected check results, used for the data quality report.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' out_dataframe <- int_all_datastructure_dataframe(
#'   meta_data_dataframe = "meta_data_dataframe",
#'   meta_data = "ship_meta"
#' )
#' md0 <- prep_get_data_frame("ship_meta")
#' md0
#' md0$VAR_NAMES
#' md0$VAR_NAMES[[1]] <- "Id" # is this missmatch reported -- is the data frame
#'                            # also reported, if nothing is wrong with it
#' out_dataframe <- int_all_datastructure_dataframe(
#'   meta_data_dataframe = "meta_data_dataframe",
#'   meta_data = md0
#' )
#'
#' # This is the "normal" procedure for inside pipeline
#' # but outside this function  checktype is exact by default
#' options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "subset_u")
#' lapply(setNames(nm = prep_get_data_frame("meta_data_dataframe")$DF_NAME),
#'   int_sts_element_dataframe, meta_data = md0)
#' md0$VAR_NAMES[[1]] <-
#'   "id" # is this missmatch reported -- is the data frame also reported,
#'        # if nothing is wrong with it
#' lapply(setNames(nm = prep_get_data_frame("meta_data_dataframe")$DF_NAME),
#'   int_sts_element_dataframe, meta_data = md0)
#' options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact")
#' }
#'
int_all_datastructure_dataframe <- function(meta_data_dataframe =
                                              "dataframe_level",
                                            item_level = "item_level",
                                            meta_data = item_level,
                                            meta_data_v2,
                                            dataframe_level) {

  # Preps and checks ----

  util_maybe_load_meta_data_v2()

  util_ck_arg_aliases()

  meta_data_dataframe <- prep_check_meta_data_dataframe(meta_data_dataframe)
  util_expect_data_frame(meta_data)
  prep_check_meta_names(meta_data = meta_data,
                        level = REQUIRED)

  study_data_list <- lapply(
    setNames(nm = meta_data_dataframe[[DF_NAME]]),
    function(dfn) {
      r <- NULL
      try(r <- prep_get_data_frame(dfn), silent = TRUE)
      if (!is.data.frame(r)) {
        util_warning("Could not load/find data frame %s. Trying it without.",
                     dQuote(dfn))
      }
      r
    }
  )

  study_data_list <-
    study_data_list[!vapply(
      study_data_list, is.null, FUN.VALUE = logical(1))]

  df_name_ok <- meta_data_dataframe[[DF_NAME]] %in% names(study_data_list)

  if (!all(df_name_ok)) {
    util_warning("Losing %d data frame(s), because they could not be loaded",
                 sum(!df_name_ok)
    )
  }

  meta_data_dataframe <- meta_data_dataframe[df_name_ok, , FALSE]

  id_vars_list <- lapply(setNames(meta_data_dataframe[[DF_ID_VARS]],
                                  nm = meta_data_dataframe[[DF_NAME]]),
                         util_parse_assignments,
                         multi_variate_text = TRUE )

  id_vars_list_vector <- lapply(id_vars_list, unlist, recursive = TRUE)


  # 1. Unexpected data element count ----

  # subset metadata with entries
  meta_data_element_count_1 <-
    meta_data_dataframe[
      !util_empty(meta_data_dataframe[[DF_ELEMENT_COUNT]]), ,
      drop = FALSE
    ]

  unexp_element_count_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    unexp_element_count_out <- withr::with_options(list(
      dataquieR.testdebug = TRUE),
      int_unexp_elements(
        identifier_name_list = meta_data_element_count_1[[DF_NAME]],
        data_element_count = meta_data_element_count_1[[DF_ELEMENT_COUNT]]
      ))
  })

  # 2. Unexpected data element set ----

  unexp_element_set_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    # subset_m cases are not reported here, since they are not usually a problem, if variable-level metadata is a large file for many data frames with diverse variable sets in, each.

    meta_data[[STUDY_SEGMENT]] <- NULL
    unexp_element_set_out <-withr::with_options(list(
      dataquieR.testdebug = TRUE), int_sts_element_dataframe(item_level = meta_data,
                                                             meta_data_dataframe =
                                                               meta_data_dataframe
      )$DataframeTable)

    unexp_element_set_out$GRADING <-
      ifelse(unexp_element_set_out$NUM_int_sts_element == 0, 0, 1)
  })

  # 3. Unexpected data record count ----

  # subset metadata with entries
  meta_data_record_count_1 <-
    meta_data_dataframe[
      !util_empty(meta_data_dataframe[[DF_RECORD_COUNT]]), ,
      drop = FALSE
    ]

  unexp_records_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    unexp_records_out <- withr::with_options(list(
      dataquieR.testdebug = TRUE), int_unexp_records_dataframe(
        identifier_name_list = meta_data_record_count_1[[DF_NAME]],
        data_record_count = meta_data_record_count_1[[DF_RECORD_COUNT]]
      ))
  })

  # 4. Unexpected data record set ----

  # subset metadata with entries
  meta_data_record_set_1 <-
    meta_data_dataframe[
      !util_empty(meta_data_dataframe[[DF_RECORD_CHECK]]), ,
      drop = FALSE
    ]

  #  meta_data_record_set_1

  unexp_records_id_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    unexp_records_id_out <- util_int_unexp_records_set_dataframe(
      id_vars_list = id_vars_list_vector[meta_data_record_set_1[[DF_NAME]]],
      identifier_name_list = meta_data_record_set_1[[DF_NAME]],
      valid_id_table_list = meta_data_record_set_1[[DF_ID_REF_TABLE]],
      meta_data_record_check_list = meta_data_record_set_1[[DF_RECORD_CHECK]]
    )
  })


  # 5. Duplicates: ids ----

  meta_data_dup_ids_1 <-
    meta_data_dataframe[
      !util_empty(meta_data_dataframe[[DF_UNIQUE_ID]]) &
        !util_empty(meta_data_dataframe[[DF_ID_VARS]]), ,
      drop = FALSE
    ]

  duplicate_ids_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    duplicate_ids_out <- withr::with_options(list(
      dataquieR.testdebug = TRUE), int_duplicate_ids(
        level = "dataframe",
        identifier_name_list = meta_data_dup_ids_1[[DF_NAME]],
        id_vars_list = id_vars_list_vector[meta_data_dup_ids_1[[DF_NAME]]],
        repetitions = meta_data_dup_ids_1[[DF_UNIQUE_ID]]))
  })

  # 6. Duplicates: content ----

  meta_data_dup_rows_1 <-
    meta_data_dataframe[
      !util_empty(meta_data_dataframe[[DF_UNIQUE_ROWS]]), ,
      drop = FALSE
    ]

  meta_data_dup_rows_1 <-
    meta_data_dup_rows_1[
      trimws(tolower(meta_data_dup_rows_1[[DF_UNIQUE_ROWS]])) ==
        "no_id" |
        !util_is_na_0_empty_or_false(meta_data_dup_rows_1[[DF_UNIQUE_ROWS]]), ]

  duplicates_rows_out <- NULL # TODO: capture errors form the next all and put them to the matrices
  try(silent = TRUE, {
    duplicates_rows_out <- withr::with_options(list(
      dataquieR.testdebug = TRUE), int_duplicate_content(
        level = "dataframe",
        identifier_name_list = meta_data_dup_rows_1[[DF_NAME]],
        id_vars_list = id_vars_list_vector[meta_data_dup_rows_1[[DF_NAME]]],
        unique_rows = setNames(meta_data_dup_rows_1[[DF_UNIQUE_ROWS]],
                               nm = meta_data_dup_rows_1[[DF_NAME]])
      ))
  })

  # Output ----

  result <- list(
    int_sts_countel = unexp_element_count_out$DataframeTable,
    int_sts_element = unexp_element_set_out,
    int_sts_countre = unexp_records_out$DataframeTable,
    int_sts_setrc = unexp_records_id_out$DataframeTable,
    int_sts_dupl_ids = duplicate_ids_out$DataframeTable,
    int_sts_dupl_row = duplicates_rows_out$DataframeTable
  )

  unexp_element_set_outData <- unexp_element_set_out
  unexp_element_set_outData$`Data frame` <-
    unexp_element_set_outData[[DF_NAME]]
  unexp_element_set_outData[[DF_NAME]] <- NULL

  # Translate indicator metric columns via helper (leave unknowns untouched)
  cn <- colnames(unexp_element_set_outData)
  if (length(cn) > 0) {
    colnames(unexp_element_set_outData) <- util_translate_indicator_metrics(
      cn,
      short = FALSE,
      long  = TRUE,
      ignore_unknown = TRUE
    )
  }

  resultData <- list(
    int_sts_countel = unexp_element_count_out$DataframeData,
    int_sts_element = unexp_element_set_outData,
    int_sts_countre = unexp_records_out$DataframeData,
    int_sts_setrc = unexp_records_id_out$DataframeData,
    int_sts_dupl_ids = duplicate_ids_out$DataframeData,
    int_sts_dupl_row = duplicates_rows_out$DataframeData
  )

  for (n in names(resultData)) {
    rownames(resultData[[n]]) <- NULL
  }

  dqi <- util_get_concept_info("dqi")
  dqi <- dqi[!util_empty(dqi$abbreviation) & !util_empty(dqi$Name), , FALSE]

  names(resultData) <-
    util_recode(
      names(resultData),
      dqi,
      "abbreviation",
      "Name",
      names(resultData)
    )

  DataframeTable <- util_merge_data_frame_list(result, "DF_NAME")
  cn <- colnames(DataframeTable)
  if (length(cn) > 0) {
    cn[startsWith(cn, "GRADING.")] <- gsub("^GRADING\\.", "GRADING_",
                                           cn[startsWith(cn, "GRADING.")])
    colnames(DataframeTable) <- cn
  }
  # DataframeData <- util_merge_data_frame_list(resultData, "Data frame")

  DataframeData_1 <- util_make_data_slot_from_table_slot(DataframeTable)
  if ("resp_vars" %in% colnames(DataframeTable)) {
    DataframeData_1$`Unexp. Variables` <- DataframeTable$resp_vars
  }

  DataframeData <- data.frame(Dataframe = DataframeTable[[DF_NAME]])
  # Replace separate columns with just one column containing N and %
  # Only if content is present, merge columns
  if(!is.null(DataframeData_1$`Unexpected data element count (Number)`)){
    DataframeData$`Unexpected data element count N (%)` <- util_paste0_with_na(DataframeData_1$`Unexpected data element count (Number)`, " (",
                                                                               DataframeData_1$`Unexpected data element count (Percentage (0 to 100))`, ")")
    DataframeData$`Unexpected data element count (Grading)`<- DataframeData_1$`Unexpected data element count (Grading)`
  }

  if(!is.null(DataframeData_1$`Unexp. Variables`)){
    DataframeData$`Unexp. Variables`<- DataframeData_1$`Unexp. Variables`
  }

  if(!is.null(DataframeData_1$`Unexpected data element set (Number)`)){
    DataframeData$`Unexpected data element set N (%)` <- util_paste0_with_na(DataframeData_1$`Unexpected data element set (Number)`, " (",
                                                                             DataframeData_1$`Unexpected data element set (Percentage (0 to 100))`, ")")
    DataframeData$`Unexpected data element set (Grading)`<- DataframeData_1$`Unexpected data element set (Grading)`
  }

  if(!is.null(DataframeData_1$`Unexpected data record count (Number)`)){
    DataframeData$`Unexpected data record count N (%)` <- util_paste0_with_na(DataframeData_1$`Unexpected data record count (Number)`, " (",
                                                                              DataframeData_1$`Unexpected data record count (Percentage (0 to 100))`, ")")
    DataframeData$`Unexpected data record count (Grading)`<- DataframeData_1$`Unexpected data record count (Grading)`
  }

  if(!is.null(DataframeData_1$`Unexpected data record set (Number)`)){
    DataframeData$`Unexpected data record set N (%)` <- util_paste0_with_na(DataframeData_1$`Unexpected data record set (Number)`, " (",
                                                                            DataframeData_1$`Unexpected data record set (Percentage (0 to 100))`, ")")
    DataframeData$`Unexpected data record set (Grading)`<- DataframeData_1$`Unexpected data record set (Grading)`
  }

  if(!is.null(DataframeData_1$`Duplicates (Number)`)){
    DataframeData$`Duplicates N (%)`<- util_paste0_with_na(DataframeData_1$`Duplicates (Number)`," (",
                                                           DataframeData_1$`Duplicates (Percentage (0 to 100))`, ")")
    DataframeData$`Duplicates (Grading)`<- DataframeData_1$`Duplicates (Grading)`
  }

  rm(DataframeData_1)


  return(list(
    DataframeTable = DataframeTable,
    DataframeData = DataframeData,
    DataframeDataList = resultData
  ))
}
# FIXME: In INT, we see errors for partiaql reports, because study_data was split by segments, so variables are missing -- maybe meta_data was split but not study  data?

#' Guess the data type of a vector
#'
#' @param x a vector with characters
#' @param k [numeric] sample size, if less than `floor(length(x) / (it/20)))`,
#'                    minimum sample size is 1.
#' @param it [integer] number of iterations when taking samples
#'
#' @details
#' # Algorithm
#' This function takes `x` and tries to guess the data type of random subsets of
#' this vector using [readr::guess_parser()]. The RNG is initialized with a
#' constant, so the function stays deterministic. It does such sub-sample based
#' checks `it` times, the majority of the detected datatype determines the
#' guessed data type.
#'
#' @return a guess of the data type of `x`. An attribute `orig_type` is also
#' attached to give the more detailed guess returned by [readr::guess_parser()].
#' @export
#' @seealso [prep_datatype_from_data()]
prep_robust_guess_data_type <- function(x, k = 50, it = 200) {
  withr::with_seed(4242, {
    util_expect_scalar(x,
                       allow_more_than_one = TRUE,
                       allow_na = TRUE,
                       check_type = is.character,
                       error_message = "Need a character vector in argument x")
    if (all(is.na(x))) {
      util_error("argument x cannot be NA only")
    }
    x <- trimws(x)
    x <- x[!is.na(x) & x != ""]
    if (length(x) == 0) {
      return(DATA_TYPES$INTEGER)
    }
    smpl_size <- max(min(k, floor(length(x) / (it/20))), 1)
    gp <- vapply(seq_len(it), function(i) {
      sub_x <- sample(x,
                      size =
                        smpl_size, replace = FALSE)
      readr::guess_parser(sub_x, guess_integer = TRUE)
    }, FUN.VALUE = character(1))
    parsers <- as.matrix(table(gp))[, 1]
    util_attach_attr(unname(c(
      date = DATA_TYPES$DATETIME,
      time = DATA_TYPES$TIME,
      datetime = DATA_TYPES$DATETIME,
      integer = DATA_TYPES$INTEGER,
      double = DATA_TYPES$FLOAT,
      character = DATA_TYPES$STRING,
      logical = DATA_TYPES$INTEGER,
      number = DATA_TYPES$FLOAT
    )[names(which.max(parsers))]),
      orig_type = names(which.max(parsers))
    )
  })
}

#' Guess encoding of text or text files
#'
#' @param x [character] string to guess encoding for
#' @param file [character] file to guess encoding for
#'
#' @return encoding
#' @export
prep_guess_encoding <- function(x, file) {
  if ((missing(file) && missing(x)) ||
      (!missing(file) && !missing(x))) {
    util_error("Can neither have both nor none of the arguments x and file")
  }
  if (!missing(file)) {
    readr::guess_encoding(file, n_max = -1, threshold = 0)
  } else {
    util_expect_scalar(x, allow_more_than_one = TRUE, check_type =
                         is.character, convert_if_possible = as.character,
                       allow_na = TRUE)
    x <- x[validEnc(x)]
    x <- x[!is.na(x)]
    guesses <- table(unlist(lapply(x, Encoding)))
    guess1 <- names(which.max(guesses))
    if (length(x) > 0 && is.null(guess1) || identical(guess1, "unknown")) {
      xx <- lapply(x, charToRaw)
      guess1 <- readr::guess_encoding(xx,
                                      n_max = -1,
                                      threshold = 0)
    } else {
      if (length(guesses) == 0) {
        guesses <- data.frame(encoding = "unknown",
                              confidence = 1)
      }
      guess1 <- setNames(data.frame(guesses), nm = c("encoding", "confidence"))
      guess1$confidence <- guess1$confidence / sum(guess1$confidence,
                                                   na.rm = TRUE)
    }
    guess1$encoding <- as.character(guess1$encoding)
    guess1$confidence <- as.numeric(guess1$confidence)
    return(guess1)
    # if (length(x) == 0) x <- ""
    # y <- paste0(x, collapse = "\n")
    # if (length(y) == 1 && !any(grepl("\n", y, fixed = TRUE))) {
    #   y <- paste0(y, "\n")
    # }
    # readr::guess_encoding(y, n_max = -1) # ensure, nrow > 0
  }
}

#' Verify encoding
#' @param dt0 [data.frame] data to verify
#' @param ref_encs [character] names are column names of `dt0`, values their
#'                             expected encoding, can be missing.
#'
#' @noRd
#' @examples
#' \dontrun{
#'   dt0 <-
#'     prep_get_data_frame(
#'     file.path("~",
#'       "rsync", "nako_mrt_qs$", "exporte", "NAKO_Datensatz_bereinigte_Daten",
#'       "NatCoEdc_Export", "export_mannheim_30.csv"))
#'   util_verify_encoding(dt0)
#'   dt0$mrt_note[[1]] <- iconv("Härbärt", "UTF-8", "cp1252")
#'   util_verify_encoding(dt0)
#'   dt0$mrt_note[[15]] <- iconv("Härbärt", "UTF-8", "cp1252")
#'   util_verify_encoding(dt0)
#'   dt0$mrt_note[[1]] <- "Härbärt"
#'   util_verify_encoding(dt0)
#'   dt0$mrt_note[[17]] <- iconv("Härbärt", "UTF-8", "latin3")
#'   util_verify_encoding(dt0)
#' }
util_verify_encoding <- function(dt0, ref_encs) {
  find_cols_with_clear_encoding <- function(x) {
    gs <- prep_guess_encoding(x);
    util_attach_attr((all(validEnc(x))) &&
      any(c("UTF-8", "ASCII") %in% gs$encoding) &&
      !any(gs$confidence < 1), enc = gs)
  }
  x_matches_ref_enc <- function(x, ref_enc) {
    if (ref_enc == "unknown") {
      return(TRUE)
    };
    y <- iconv(x, ref_enc, "UTF-8", sub = '');
    r <- (is.na(x) & is.na(y)) |
      (!is.na(x) & !is.na(y) & (x == y));
    r
  }
  cols_enc <- lapply(dt0, find_cols_with_clear_encoding)
  cols_with_unclear_encoding <- setNames(nm = names(which(!vapply(
                                           cols_enc,
                                           identity,
                                           FUN.VALUE = logical(1)),
                                   useNames = TRUE)))
  not_missing_refencs <- !missing(ref_encs)
  res <- lapply(
    setNames(nm = cols_with_unclear_encoding),
    function(col_unclear) {
      if (not_missing_refencs) {
        ref_enc <- ref_encs[[col_unclear]]
      } else {
        ref_enc <-
          prep_guess_encoding(dt0[[col_unclear]])$encoding[[1]]
      }
      res <- which(!vapply(dt0[[col_unclear]],
                    x_matches_ref_enc,
                    ref_enc = ref_enc,
                    FUN.VALUE = logical(1)))
      act_enc <- lapply(setNames(dt0[res, col_unclear],
                                 nm = res), prep_guess_encoding)
      util_attach_attr(res,
                       ref_enc = ref_enc,
                       act_enc = act_enc)
    })
  res[vapply(res, length, FUN.VALUE = integer(1)) == 0] <- NULL
  util_attach_attr(res, cols_enc = lapply(cols_enc, attr, "enc"))
}

#' utility function applicability of distribution function's shape or scale
#' check
#'
#' Test for applicability of checks for deviation form expected
#' probability distribution shapes/scales
#'
#' @param x [data.frame] metadata
#' @param dta [logical] vector, 1=matching data type,
#'                  0 = non-matching data type
#'
#' @return [factor] 0-3 for each variable in metadata
#' - 0 data type mismatch and not applicable
#' - 1 data type mismatches but applicable
#' - 2 data type matches but not applicable
#' - 3 data type matches and applicable
#' - 4 not applicable because of not suitable data type
#'
#' @seealso [pro_applicability_matrix]
#' @noRd

util_app_sos <- function(x, dta) {
  if ("DISTRIBUTION" %in% names(x)) {
    c1 <- ifelse(is.na(x["DISTRIBUTION"]), 0, 1)
  } else {
    c1 <- rep(0, times = dim(x)[1])
  }

  aa <- paste0(dta, c1)
  score <- as.numeric(recode(as.factor(aa), "00" = 0, "01" = 1,
                             "10" = 2, "11" = 3))
  score <- ifelse(x[["DATA_TYPE"]] %in% c("float", "integer"), score, 4)
  score <- as.factor(score)
  return(score)
}

test_that("util_app_sos works", {
  skip_on_cran() # deprecated

  # Create a test metadata set
  md <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = c(rep(DATA_TYPES$FLOAT, 13), rep(DATA_TYPES$INTEGER, 10),
                  DATA_TYPES$STRING, DATA_TYPES$DATETIME, DATA_TYPES$STRING),
    MISSING_LIST = "",
    GROUP_VAR_OBSERVER = "a",
    DISTRIBUTION = c(rep(DISTRIBUTIONS$NORMAL, 10),
                     DISTRIBUTIONS$GAMMA,
                     DISTRIBUTIONS$UNIFORM,
                     rep(NA, 14)),
    DETECTION_LIMITS =
      c(rep("[0;9)", 10), rep(NA, 6), rep("", 10))
  )

  result1 <- util_app_sos(md, as.factor(rep(1, nrow(md))))

  # Check that the result is a factor
  expect_s3_class(result1, "factor")

  # check result
  expect_equal(result1,
               as.factor(c(rep(3, 12), rep(2, 11), rep(4, 3))))

  md0 <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = c(rep(DATA_TYPES$FLOAT, 13), rep(DATA_TYPES$INTEGER, 10),
                  DATA_TYPES$STRING, DATA_TYPES$DATETIME, DATA_TYPES$STRING),
    MISSING_LIST = "",
    GROUP_VAR_OBSERVER = "a",
    DETECTION_LIMITS =
      c(rep("[0;9)", 10), rep(NA, 6), rep("", 10))
  )

  result2 <- util_app_sos(md0, as.factor(rep(1, nrow(md))))

  # check result
  expect_equal(result2,
               as.factor(c(rep(2, 23), rep(4, 3))))
})

#' Display a sunburst chart
#'
#' @param repsum a report summary
#' @param remove_lines display lines?
#' @param ex exponent to steepen color gradient
#' @param folder_of_report a named vector with the location of variable and
#'                         `call_names`
#' @param var_uniquenames a data frame with the original variable names and the
#'                        unique names in case of reports created with dq_report_by
#'                        containing the same variable in several reports
#'                        (e.g., creation of reports by sex)
#'
#' @returns a `plotly` object
#' @noRd
util_render_sunburst_from_summary_classes <- function(repsum,
                                                      remove_lines,
                                                      ex = 5,
                                                      folder_of_report = NULL,
                                                      var_uniquenames = NULL) {

  util_ensure_suggested(pkg = "plotly", goal = "generate sunburst plot")
  util_ensure_suggested(pkg = "htmlwidgets", goal = "generate sunburst plot")
  util_ensure_suggested(pkg = "rmarkdown", goal = "generate sunburst plot")

  # NSE variables (for R CMD check): avoid "no visible binding" notes
  Dimension <- Domain <- Name <- Parent_Element_ID <- IndicatorID <-
    abbreviation <- Level <- Indicator <- class <- LABEL_use <- max.class <-
    severity_num <- mean_class_within_name <- mean_class_within_dom <-
    mean_class_within_dim <- color <- id <- parent <- path <- L1 <-
    L2 <- L3 <- L4 <- severity_mix <- n <- font_size <- id_match <-
    children <- child_vals <- n_leaves_dim <- n_leaves_dom <-
    href <- popup_href <- title <- n_leaves_name <- NULL


  this <- attr(repsum, "this")

  if (missing(remove_lines)) {
    remove_lines <- length(unique(this$colnames_of_report)) *
      length(unique(this$rownames_of_report)) > 1000 # TODO: find empirically a reasonable default.
  }

  repsum2 <- util_dashboard_table(repsum, folder_of_report = folder_of_report)
  is_by <- !is.null(folder_of_report)

  repsum2$LABEL_use <-
    prep_get_labels(repsum2$VAR_NAMES,item_level = this$meta_data,
                    label_col = this$label_col, max_len = 40, label_class = "SHORT",
                    resp_vars_are_var_names_only = TRUE, resp_vars_match_label_col_only = FALSE)

  dqi <- util_get_concept_info("dqi")

  dqi$Indicator = dqi$Name
  dqi$Indicator[dqi$Level != 3] <- NA

  cols_dqi <- c("Dimension","Domain","Name","Parent_Element_ID","IndicatorID","abbreviation","Level","Indicator")

  dqi <-
    dqi %>%
    dplyr::select(dplyr::all_of(cols_dqi)) %>%
    dplyr::distinct(dplyr::across(dplyr::all_of(cols_dqi)), .keep_all = FALSE)


  repsum2$indicator_metric <- sub("^[^_]+_", "", repsum2$indicator_metric)

  df <- dplyr::left_join(dqi, repsum2, by = c("abbreviation" = "indicator_metric"))

  sunburst_df <-
    df %>%
    dplyr::select("Parent_Element_ID",
                  "IndicatorID" ,
                  "LABEL_use" ,
                  "Dimension",
                  "Domain",
                  "Name",
                  "class",
                  "Level",
                  "href",
                  "popup_href",
                  "title")

  dqi_3 <-
    dqi %>%
    dplyr::filter(get("Level") == 3) %>%
    dplyr::select("Dimension",
                  "Domain",
                  "Name",
                  "Level") %>%
    dplyr::mutate("LABEL_use" = NA,
                  "max.class" = NA)

  cols_sunburst_df <- c("Dimension", "Domain", "Name", "LABEL_use", "Level", "href", "popup_href", "title")

  sunburst_df <-
    sunburst_df %>%
    dplyr::group_by(dplyr::across(dplyr::all_of(cols_sunburst_df))) %>%
    dplyr::mutate("Level" = dplyr::case_when(
      !is.na(get("LABEL_use")) ~ 4,
      TRUE ~ get("Level")
    )) %>%
    dplyr::summarize("max.class" = suppressWarnings(max(get("class"), na.rm = TRUE))) %>%
    dplyr::select("Dimension", "Domain", "Name", "LABEL_use", "Level", "max.class", "href", "popup_href", "title") %>%
    dplyr::filter(get("Level") != 3)

  sunburst_df <- dplyr::bind_rows(dqi_3, sunburst_df)

  # Define a severity → numeric scale
  sunburst_df$severity_num <- as.numeric(sunburst_df$max.class) # now, we have numbers according to the level order in the ordered class factor

  # Level 3 = Mean der Leaves
  sunburst_df <- sunburst_df %>%
    dplyr::group_by(Dimension, Domain, Name) %>%
    dplyr::mutate(
      mean_class_within_name = mean(severity_num[Level == 4], na.rm = TRUE),
      severity_num = ifelse(Level == 3, mean_class_within_name, severity_num)
    ) %>%
    dplyr::ungroup()

  # Level 2
  sunburst_df <- sunburst_df %>%
    dplyr::group_by(Dimension, Domain) %>%
    dplyr::mutate(
      mean_class_within_dom = mean(severity_num[Level == 4], na.rm = TRUE),
      severity_num = ifelse(Level == 2, mean_class_within_dom, severity_num)
    ) %>%
    dplyr::ungroup()

  # Level 1
  sunburst_df <- sunburst_df %>%
    dplyr::group_by(Dimension) %>%
    dplyr::mutate(
      mean_class_within_dim = mean(severity_num[Level == 4], na.rm = TRUE),
      severity_num = ifelse(Level == 1, mean_class_within_dim, severity_num)
    ) %>%
    dplyr::ungroup()

  sunburst_df$LABEL_use <-
    vapply(sunburst_df$LABEL_use, function(x) {
      if (is.null(x)) NA_character_ else as.character(x)
    }, FUN.VALUE = character(1))

  sunburst_df <-
    sunburst_df %>%
    dplyr::mutate("LABEL_use" = dplyr::case_when(
      get("Level") == 1 ~ get("Name"),
      get("Level") == 2 ~ get("Name"),
      get("Level") == 3 ~ get("Name"),
      TRUE ~ get("LABEL_use")
    ))

  # Create an ID and parent relationship for sunburst
  sunburst_full <- sunburst_df %>%
    dplyr::filter(!is.na(get("severity_num"))) %>%
    dplyr::mutate(
      "id" = dplyr::case_when( # ids must be allowed anchor names in a URL
        get("Level") == 1 ~ get("Dimension"),
        get("Level") == 2 ~ paste(get("Dimension"), get("Domain"), sep = " - "),
        get("Level") == 3 ~ paste(get("Dimension"), get("Domain"), get("Name"), sep = " - "),
        get("Level") == 4 ~ paste(get("Dimension"), get("Domain"), get("Name"), get("LABEL_use"), sep = " - ")
      ),
      "parent" = dplyr::case_when(
        get("Level") == 1 ~ "",
        get("Level") == 2 ~ get("Dimension"),
        get("Level") == 3 ~ paste(get("Dimension"), get("Domain"), sep = " - "),
        get("Level") == 4 ~ paste(get("Dimension"), get("Domain"), get("Name"), sep = " - ")

      )
    )  %>%
    # ---- leaf counts (descendant leaves); leaves are Level==4 ----
  dplyr::left_join(
    dplyr::filter(., get("Level") == 4) %>%
      dplyr::count(Dimension, name = "n_leaves_dim"),
    by = "Dimension"
  ) %>%
    dplyr::left_join(
      dplyr::filter(., get("Level") == 4) %>%
        dplyr::count(Dimension, Domain, name = "n_leaves_dom"),
      by = c("Dimension", "Domain")
    ) %>%
    dplyr::left_join(
      dplyr::filter(., get("Level") == 4) %>%
        dplyr::count(Dimension, Domain, Name, name = "n_leaves_name"),
      by = c("Dimension", "Domain", "Name")
    ) %>%
    dplyr::mutate(
      n_leaves = dplyr::case_when(
        get("Level") == 1 ~ as.integer(n_leaves_dim),
        get("Level") == 2 ~ as.integer(n_leaves_dom),
        get("Level") == 3 ~ as.integer(n_leaves_name),
        get("Level") == 4 ~ 1L,
        TRUE ~ NA_integer_
      )
    ) %>%
    dplyr::select(-n_leaves_dim, -n_leaves_dom, -n_leaves_name)

  line_cfg <- if (remove_lines) {
    list(width = 0, color = 'rgba(0,0,0,0)')
  } else {
    list(width = 1, color = 'white') # Standard: white seperator lines
  }

  min_font <- 12
  max_font <- 30

  sunburst_full <- sunburst_full %>%
    dplyr::mutate(
      # scale based on severity like in a word-cloud
      font_size = min_font + (severity_num - 1) / 4 * (max_font - min_font)
    )

  sunburst_full$font_size[is.na(sunburst_full$font_size)] <- min_font

  sunburst_full <- sunburst_full %>%
    dplyr::mutate(
      customdata = Map(
        function(href, popup_href, title, severity_num) {
          list(
            href = href,
            popup_href = popup_href,
            title = title,
            severity_num = severity_num
          )
        },
        href,
        popup_href,
        title,
        severity_num
      )
    )

  # ---------------------------
  # Plot
  # ---------------------------
  py <- plotly::plot_ly(
    data = sunburst_full,
    ids = ~ id,
    labels = ~ LABEL_use,
    parents = ~ parent,
    values = ~ severity_num * n_leaves,
    customdata = ~ customdata,
    type = 'sunburst',
    branchvalues = 'total',
    textinfo = "label",
    textfont = list(size = ~ font_size),
    marker = list(
      colors = ~ 1 - (( 1 - ((severity_num - 1)/4)) ^ ex),
      line = line_cfg,
      # colorscale = "Bluered",
      # colorscale = "RdBu",
      # colorscale = list(
      #   list(1.0, "#f2f2f2"),
      #   list(2.0, "#d9d9d9"),
      #   list(3.0, "#f4c7c3"),
      #   list(4.0, "#e57368"),
      #   list(5.0, "#c43c39")
      # ),
      colorscale = lapply(mapply(SIMPLIFY = FALSE, nm = 1.0 * (1:5-1)/4, cl = util_get_colors(), list), unname),
      cmin = 0,
      cmax = 1,
      showscale = FALSE
    )
  )

  py <- htmlwidgets::onRender(
    py,
    "sunburst_on_render",
    data = list(ex_init = ex)
  )

  py$height <- NULL
  py$width  <- NULL
  # py <- plotly::config(py, responsive = TRUE)

  py <- plotly::layout(py,
                       autosize = is_by
  )

  py <- plotly::config(py, displaylogo = FALSE)

  py <- util_decorate_plotly(py)

  py <- util_plotly_add_modebar_buttons(py,
                   list(
                     list(
                       name = "Back to root",
                       icon = htmlwidgets::JS("Plotly.Icons.home"),
                       click = htmlwidgets::JS("Plotly_dq_go_root")
                      )
                   )
  )

  py <- util_plotly_remove_modebar_button_added(py, "Restore initial Size")

  py <- util_plotly_modebar_right(py, 50)

  py <- plotly::layout(py,
                       margin = list(r = 40))

  py$sizingPolicy$defaultHeight <- "100%"

  if (.called_in_pipeline) {
    dummy <- htmltools::tagList()
  } else {
    dummy <-
      htmltools::tagList(htmltools::div(class = "navbar"),
                         htmltools::div(class = "content"),
                         htmltools::tags$script(type = "text/javascript",
                                                "window.dq_report2 = true")
      )
  }

  if (is_by) {
    py <- htmltools::tagList(dummy,
                             htmltools::div(
                               style="width:100vw;height:100vh",
                               `data-tippy-always-on` = "true",
                               py
                             )
    )
  } else {
    py <- htmltools::tagList(dummy,
                             htmltools::div(
                               style="width:100vw;height:95vh;overflow:hidden;",
                               `data-tippy-always-on` = "true",
                               py
                             )
    )
  }


  if (!.called_in_pipeline) {
    py <- htmltools::tagList(rmarkdown::html_dependency_jquery(),
                           html_dependency_clipboard(),
                           html_dependency_tippy(),
                           html_dependency_dataquieR(iframe = FALSE),
                           py)
  }

  py <- htmltools::browsable(py)

  py

}

#' Force `Plotly` `modebar` to the top-right (`htmlwidgets`)
#'
#' Some CSS setups or late resizes can make `Plotly` place the `modebar` (buttons)
#' on the left. This helper pins it to the right via a widget-local JS hook.
#'
#' @param p A `plotly` `htmlwidget`.
#' @param right_px Inset from the right edge (CSS px).

#' @return The modified widget.
#' @noRd
util_plotly_modebar_right <- function(p, right_px = 16) {
  stopifnot(is.numeric(right_px), length(right_px) == 1L, is.finite(right_px))

  htmlwidgets::onRender(
    p,
    sprintf("
function(el, x) {
  function pin() {
    var mbs = el.querySelectorAll('.modebar, .modebar-container');
    for (var i = 0; i < mbs.length; i++) {
      mbs[i].style.left  = 'auto';
      mbs[i].style.right = '%spx';
    }
  }

  pin();

  if (typeof Plotly !== 'undefined' && el && typeof el.on === 'function') {
    el.on('plotly_afterplot', pin);
    el.on('plotly_relayout',  pin);
  }

  requestAnimationFrame(pin);
  setTimeout(pin, 0);
  setTimeout(pin, 50);
}
", as.integer(round(right_px)))
  )
}

util_plotly_remove_modebar_button_added <- function(p, name) {
  cfg <- p$x$config
  mba <- cfg$modeBarButtonsToAdd
  if (is.null(mba) || !length(mba)) return(p)

  cfg$modeBarButtonsToAdd <- Filter(
    function(btn) {
      # keep unknown shapes by default
      if (is.list(btn)) {
        # custom buttons are lists with $name
        return(is.null(btn$name) || !identical(btn$name, name))
      }
      # atomic entries (sometimes characters) -> drop only if equal
      if (is.atomic(btn) && length(btn) == 1L) {
        return(!identical(as.character(btn), name))
      }
      TRUE
    },
    mba
  )

  p$x$config <- cfg
  p
}
# Nur AGE_0 wird organge?

#' Support function to create [data.frame]s of metadata
#'
#' @description
#' Create a metadata data frame and map names.
#' Generally, this function only creates a [data.frame], but using
#' this constructor instead of calling
#' `data.frame(..., stringsAsFactors = FALSE)`, it becomes possible, to adapt
#' the metadata [data.frame] in later developments, e.g. if we decide to use
#' classes for the metadata, or if certain standard names of variable attributes
#' change. Also, a validity check is possible to implement here.
#'
#' @details
#' For now, this calls [data.frame], but it already renames variable attributes,
#' if they have a different name assigned in [WELL_KNOWN_META_VARIABLE_NAMES],
#' e.g. `WELL_KNOWN_META_VARIABLE_NAMES$RECODE` maps to `recode` in lower case.
#'
#' NB: `dataquieR` exports all names from WELL_KNOWN_META_VARIABLE_NAME as
#' symbols, so `RECODE` also contains `"recode"`.
#'
#' @param ... named column vectors, names will be mapped using
#'            [WELL_KNOWN_META_VARIABLE_NAMES],
#'            if included in [WELL_KNOWN_META_VARIABLE_NAMES]
#'            can also be a data frame, then its column names will be mapped
#'            using [WELL_KNOWN_META_VARIABLE_NAMES]
#' @param stringsAsFactors [logical] if the argument is a list of vectors, a
#'                         data frame will be
#'                         created. In this case, `stringsAsFactors` controls,
#'                         whether characters will be auto-converted to Factors,
#'                         which defaults here always to false independent from
#'                         the [default.stringsAsFactors].
#' @param level [enum] level of requirement (see also [VARATT_REQUIRE_LEVELS])
#'                     set to `NULL`, if not a complete metadata frame is
#'                     created.
#' @param character.only [logical] a logical indicating whether level can be
#'                                 assumed to be character strings.
#'
#' @seealso [WELL_KNOWN_META_VARIABLE_NAMES]
#' @return a data frame with:
#'   - metadata attribute names mapped and
#'   - metadata checked using [prep_check_meta_names] and do some more
#'     verification about conventions, such as check for valid intervals
#'     in limits)
#' @export
#' @importFrom stats setNames
prep_create_meta <- function(..., stringsAsFactors = FALSE,
                             level, character.only = FALSE) {
  if (length(list(...)) == 1 && is.data.frame(list(...)[[1]])) {
    metas <- list(...)[[1]]
    cn <- colnames(metas)
  } else {
    args <- list(...)
    if (any(duplicated(names(args)))) {
      util_error("Found duplicated meta columns: %s",
                 paste0(dQuote(unique(names(args)[duplicated(names(args))])),
                        collapse = ", "),
                 applicability_problem = TRUE)
    }

    args_lens <- vapply(args, length, integer(1))

    null_args <- args_lens == 0

    if (any(null_args)) {
      util_warning(c("The following variable attributes are",
                   "NULL, will ignore these: %s"),
                   paste0(dQuote(names(null_args[null_args])),
                          collapse = ", "),
                   applicability_problem = TRUE)
      args <- args[!null_args]
      args_lens <- vapply(args, length, integer(1))
    }

    max_len <- max(args_lens)

    if (any(max_len %% args_lens != 0)) {
      util_error(c("The given variable attributes have different lengths and",
                   "cannot be equalized by repeating the shorter ones to",
                   "create a data frame with as %d rows (the longest",
                   "variable attribute vector provided). R would say:",
                   "arguments imply differing number of rows: %s"),
                   max_len,
                   paste0(sort(unique(args_lens)), collapse = ", "))
    }

    metas <- do.call(data.frame, c(args, stringsAsFactors = stringsAsFactors))
    cn <- names(args)
  }
  if (ncol(metas) > 0) {
    cn <-  mget(
      x = cn,
      envir = as.environment(WELL_KNOWN_META_VARIABLE_NAMES),
      mode = "character",
      ifnotfound = setNames(nm = colnames(metas)),
      inherits = FALSE
    )
    colnames(metas) <- cn
  }
  if (missing(level)) {
    level <- VARATT_REQUIRE_LEVELS$REQUIRED
  } else {
    if (character.only) {
      level <- as.character(level)
    } else {
      level <- as.character(substitute(expr = level))
    }
    if (length(level) > 0) {
      level <- try(match.arg(level, choices = c(names(VARATT_REQUIRE_LEVELS),
                                                unlist(VARATT_REQUIRE_LEVELS)),
                             several.ok = FALSE), silent = TRUE)
      if (inherits(level, "try-error")) {
        util_error(
          "Error regarding argument %s: %s",
          dQuote("level"),
          conditionMessage(attr(level, "condition")),
          applicability_problem = TRUE
        )
      }
    }
  }
  # level <- VARATT_REQUIRE_LEVELS[[level]]

  prep_check_meta_names(item_level = metas,
                        level = level, character.only = TRUE)
  if (any(c(JUMP_LIST, MISSING_LIST, MISSING_LIST_TABLE) %in%
      util_get_var_att_names_of_level(level))) {
    metas <- util_validate_known_meta(metas) # TODO: MISSING_LIST_TABLE
  }
  return(metas)
}

test_that("util_map_labels works", {
  skip_on_cran()
  meta_data <- prep_create_meta(
    VAR_NAMES = c("ID", "SEX", "AGE", "DOE"),
    LABEL = c("Pseudo-ID", "Gender", "Age", "Examination Date"),
    MISSING_LIST = "",
    DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER, DATA_TYPES$INTEGER,
                  DATA_TYPES$DATETIME)
  )
  expect_equal(util_map_labels(c("AGE", "DOE"), meta_data),
               c(AGE = "Age", DOE = "Examination Date"))
  expect_error(util_map_labels(c("NOT_AVAIL", "AGE"), meta_data),
               regexp = "value for .+NOT_AVAIL.+ not found", perl = TRUE)
  expect_equal(util_map_labels(c("AGE", "NOT_AVAIL"), meta_data,
                               ifnotfound = NA_character_),
               c(AGE = "Age", NOT_AVAIL = NA_character_))
  expect_equal(util_map_labels(c("AGE", "NOT_AVAIL"), meta_data,
                               ifnotfound = 42),
               c(AGE = "Age", NOT_AVAIL = 42))
})

test_that("util_get_code_list works", {
  skip_on_cran()
  
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  mdf <- prep_create_meta(
    VAR_NAMES = c("age", "sex"),
    DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER),
    MISSING_LIST = c(NA, "-1 = -|-2 = -|-5 = -"),
    JUMP_LIST = c("999 = -", "")
  )
  expect_warning(expect_equal(
    util_get_code_list(c("age"), "MISSING_LIST",
      split_char = SPLIT_CHAR,
      mdf = mdf,
      label_col = VAR_NAMES,
      warning_if_no_list = TRUE
    ),
    numeric(0)
  ),
  perl = TRUE,
  regexp = paste0("Could not find .MISSING_LIST. for",
                  " .age. in the meta_data for replacing codes with NAs.")
  )
  expect_equal(
    util_get_code_list(c("sex"), "MISSING_LIST",
      split_char = SPLIT_CHAR,
      mdf = mdf,
      label_col = VAR_NAMES,
      warning_if_no_list = TRUE
    ),
    c(`-` = -1,`-` =  -2, `-` = -5)
  )
  expect_equal(
    util_get_code_list(c("sex"), "JUMP_LIST",
      split_char = SPLIT_CHAR,
      mdf = mdf,
      label_col = VAR_NAMES,
      warning_if_no_list = TRUE
    ),
    setNames(numeric(0), character(0))
  )
  expect_equal(
    util_get_code_list(c("age"), "JUMP_LIST",
      split_char = SPLIT_CHAR,
      mdf = mdf,
      label_col = VAR_NAMES,
      warning_if_no_list = TRUE
    ),
    c(`-` = 999)
  )
  expect_warning(
    util_get_code_list(c("age"), "XJUMP_LIST",
                       split_char = SPLIT_CHAR,
                       mdf = mdf,
                       label_col = VAR_NAMES,
                       warning_if_no_list = TRUE
    ),
    regexp = paste(
      "Metadata does not provide a column called .*XJUMP_LIST.*",
      "for replacing codes with NAs."
    ),
    perl = TRUE
  )
  expect_warning(
    util_get_code_list(c("age"), "JUMP_LIST",
                       split_char = SPLIT_CHAR,
                       mdf = mdf,
                       label_col = "xx",
                       warning_if_no_list = TRUE
    ),
    regexp = paste(
      "Metadata does not provide a column called .+xx.+",
      "for the labels."
    ),
    perl = TRUE
  )
  # util_get_code_list(c("sex", "age"), "JUMP_LIST",
  #                    split_char = SPLIT_CHAR,
  #                    mdf = mdf,
  #                    label_col = VAR_NAMES,
  #                    warning_if_no_list = TRUE)
})

test_that("util_get_code_list warns about non-numeric codes", {
  skip_on_cran()
  
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  mdf <- prep_create_meta(
    VAR_NAMES = c("age", "sex"),
    DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER),
    MISSING_LIST = c(NA, NA),
    JUMP_LIST = c("999=-", "")
  )
  mdf$MISSING_LIST[2] <- "-1|-2|-5|x"
  expect_warning(expect_equal(
    util_get_code_list(c("sex"), "MISSING_LIST",
                       split_char = SPLIT_CHAR,
                       mdf = mdf,
                       label_col = VAR_NAMES,
                       warning_if_no_list = TRUE
    ),
    c(`-1` = -1, `-2` = -2, `-5` = -5)
  ),
  perl = TRUE,
  regexp = paste("Some codes ..MISSING_LIST.. were",
                  "not numeric/assignment for .sex.: .x., these will be ignored")
  )
})

test_that("util_get_code_list recognizes no code on purpose", {
  skip_on_cran()
  mdf <- prep_create_meta(
    VAR_NAMES = c("age", "sex"),
    DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER),
    MISSING_LIST = c(SPLIT_CHAR, SPLIT_CHAR),
    JUMP_LIST = c("999 = ", "")
  )
  expect_silent(expect_equal(
    util_get_code_list(c("sex"), "MISSING_LIST",
                       split_char = SPLIT_CHAR,
                       mdf = mdf,
                       label_col = VAR_NAMES,
                       warning_if_no_list = TRUE
    ),
    setNames(numeric(0), character(0))
  ))
})

test_that("util_app_iac works", {
  skip_on_cran() # deprecated
  md <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = c(rep(DATA_TYPES$FLOAT, 13), rep(DATA_TYPES$INTEGER, 10),
                  DATA_TYPES$STRING, DATA_TYPES$DATETIME, DATA_TYPES$STRING),
    MISSING_LIST = "",
    DETECTION_LIMITS =
      c(rep("[0;9)", 10), rep(NA, 6), rep("", 10))
  )
  expect_equal(util_app_iac(md, as.factor(rep(1, nrow(md)))),
               as.factor(c(rep(4, 13), rep(2, 11), 4, 2)))
  md <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = c(rep(DATA_TYPES$FLOAT, 13), rep(DATA_TYPES$INTEGER, 10),
                  DATA_TYPES$STRING, DATA_TYPES$DATETIME, DATA_TYPES$STRING),
    MISSING_LIST = "",
    VALUE_LABELS =
      head(c(rep(c(NA_character_, "", "12 = x | 14 = z", "|"),
                 ceiling(26 / 4))), 26),
    DETECTION_LIMITS =
      c(rep("[0;9)", 10), rep(NA, 6), rep("", 10))
  )
  testthat::local_edition(3)
  expect_snapshot_value(style = "deparse",
    util_app_iac(md, as.factor(c(rep(rep(1, nrow(md)), 52),
    rep(rep(0, nrow(md)), 52)))))
})

test_that("util_app_sm works", {
  skip_on_cran() # deprecated
  md <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = DATA_TYPES$FLOAT,
    MISSING_LIST = "",
    KEY_STUDY_SEGMENT =
      c(rep("a", 10), rep(NA, 6), rep("b", 10))
  )
  suppressMessages(md <- prep_meta_data_v1_to_item_level_meta_data(md))
  expect_equal(util_app_sm(md, as.factor(rep(1, nrow(md)))),
               as.factor(c(rep(3, 10), rep(2, 6), rep(3, 10))))
})

test_that("util_app_hl works", {
  skip_on_cran()
  # 0. Non-Matching data types
  #  - return 1: data type mismatches but applicable
  #  - return 4: not applicable because of not suitable data type

  # Create a test data set
  study_data <- cars
  meta_data <- prep_create_meta(
    VAR_NAMES = c("speed", "dist"),
    LABEL = c("Speed", "Stopping distance"),
    DATA_TYPE = c(DATA_TYPES$FLOAT, DATA_TYPES$STRING),
    MISSING_LIST = "",
    HARD_LIMITS = c("[5.5;10.25]", "")
  )

  result0 <- util_app_hl(meta_data, dta = c(0, 0))

  # Check that the result is a factor
  expect_s3_class(result0, "factor")

  # Check that the result is as expected
  expect_equal(result0, as.factor(c(1, 4)))

  # 1. Matching data types but no hard limits defined
  #  - return 2: data type matches but not applicable

  # Create a test data set
  meta_data <- prep_create_meta(
    VAR_NAMES = c("speed", "dist"),
    LABEL = c("Speed", "Stopping distance"),
    DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER),
    MISSING_LIST = ""
  )

  result1 <- util_app_hl(meta_data, dta = c(1,1))

  # Check that the result is a factor
  expect_s3_class(result1, "factor")

  # Check that the result is as expected
  expect_equal(result1, as.factor(c(2, 2)))

  # 2. Matching data types and hard limits defined
  #  - return 3: data type matches and applicable

  # Create a test data set
  meta_data <- prep_create_meta(
    VAR_NAMES = c("speed", "dist"),
    LABEL = c("Speed", "Stopping distance"),
    DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER),
    MISSING_LIST = "",
    HARD_LIMITS = c("[5;25]", "[1;100]")
  )

  result2 <- util_app_hl(meta_data, dta = c(1,1))

  # Check that the result is as expected
  expect_equal(result2, as.factor(c(3, 3)))

  # 3. Non-Matching data types and hard limits defined
  #  - return 3: data type matches and applicable
  #  - return 4: not applicable because of not suitable data type

  # Create a test data set
  meta_data <- prep_create_meta(
    VAR_NAMES = c("speed", "dist"),
    LABEL = c("Speed", "Stopping distance"),
    DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$STRING),
    MISSING_LIST = "",
    HARD_LIMITS = c("[5;25]", "[1;100]")
  )

  result3 <- util_app_hl(meta_data, dta = c(1,1))

  # Check that the result is as expected
  expect_equal(result3, as.factor(c(3, 4)))

})

test_that("util_app_dl works", {
  skip_on_cran() # deprecated

  # Create a test metadata set
  md <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = c(rep(DATA_TYPES$FLOAT, 13), rep(DATA_TYPES$INTEGER, 10),
                  DATA_TYPES$STRING, DATA_TYPES$DATETIME, DATA_TYPES$STRING),
    MISSING_LIST = "",
    DETECTION_LIMITS =
      c(rep("[0;9)", 10), rep(NA, 6), rep("", 10))
  )

  md_no_names <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = c(rep(DATA_TYPES$FLOAT, 13), rep(DATA_TYPES$INTEGER, 10),
                  DATA_TYPES$STRING, DATA_TYPES$DATETIME, DATA_TYPES$STRING),
    MISSING_LIST = ""
  )

  result1 <- util_app_dl(md, as.factor(rep(1, nrow(md))))
  result2 <- util_app_dl(md_no_names, as.factor(rep(1, nrow(md_no_names))))

  # Check that the result is a factor
  expect_s3_class(result1, "factor")

  # Check that the result is as expected
  expect_equal(result1,
               as.factor(c(rep(3, 10), rep(2, 6), rep(3, 7), rep(4, 3))))

  expect_equal(result2,
               as.factor(c(rep(2, 23), rep(4, 3))))
})

test_that("util_app_im works", {
  skip_on_cran() # deprecated
  expect_warning(
    md <- prep_create_meta(
      VAR_NAMES = letters,
      DATA_TYPE = DATA_TYPES$FLOAT,
      MISSING_LIST =
        c(rep("1 = NA | 9 = -", 10), rep(NA, 3), rep("1 = NA | 9 = -", 13)),
      JUMP_LIST =
        c(rep("1 = Not planned | 9 = exp. missing", 12), rep(NA, 3), "",
          rep("1 = Not planned | 9 = exp. missing", 10))
    ),
    regexp = paste("Found at least one missing code with more than one meaning",
                   "for .a., .b., .c., .d., .e., .f., .g., .h., .i., .j.,",
                   ".q., .r., .s., .t., .u., .v., .w., .x., .y., .z..")
  )
  expect_equal(util_app_im(md, as.factor(rep(1, nrow(md)))),
               as.factor(c(rep(3, 12), rep(2, 1), rep(3, 13))))
  md$JUMP_LIST <- NULL
  md$MISSING_LIST <- NULL
  expect_equal(util_app_im(md, as.factor(rep(1, nrow(md)))),
               as.factor(rep(2, 26)))
})

test_that("util_app_mol works", {
  skip_on_cran() # deprecated
  md <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = c(rep(DATA_TYPES$FLOAT, 13), rep(DATA_TYPES$INTEGER, 10),
                  DATA_TYPES$STRING, DATA_TYPES$DATETIME, DATA_TYPES$STRING),
    MISSING_LIST = "",
    GROUP_VAR_OBSERVER = "a",
    DETECTION_LIMITS =
      c(rep("[0;9)", 10), rep(NA, 6), rep("", 10))
  )
  expect_equal(util_app_mol(md, as.factor(rep(1, nrow(md)))),
               as.factor(c(rep(3, 23), rep(4, 3))))
})

test_that("prep_create_meta works", {
  skip_on_cran()
  l <- VARATT_REQUIRE_LEVELS$REQUIRED
  expect_warning(
    meta_data1 <- prep_create_meta(
      VAR_NAMES = letters,
      DATA_TYPE = DATA_TYPES$INTEGER,
      LABEL = LETTERS,
      MISSING_LIST = "999 = a|998 = b",
      XYZ = character(0),
      level = l,
      character.only = TRUE
    ),
    regexp = sprintf("(%s)",
      "The following variable attributes are NULL, will ignore these: .+XYZ.+"
    ),
    perl = TRUE
  )

  meta_data <- prep_create_meta(
    VAR_NAMES = letters,
    DATA_TYPE = DATA_TYPES$INTEGER,
    LABEL = LETTERS,
    MISSING_LIST = "999 = a|998 = b"
  )
  expc <- data.frame(
    stringsAsFactors = FALSE,
    VAR_NAMES = letters,
    DATA_TYPE = DATA_TYPES$INTEGER,
    LABEL = LETTERS,
    MISSING_LIST = "999 = a|998 = b"
  )
  expect_identical(meta_data1, expc)
  expect_identical(meta_data, expc)
  meta_data2 <- prep_create_meta(meta_data)
  expect_identical(meta_data2, expc)
  expect_error(
    meta_data <- prep_create_meta(
      VAR_NAMES = letters,
      DATA_TYPE = DATA_TYPES$INTEGER,
      LABEL = LETTERS,
      MISSING_LIST = "999|998",
      MISSING_LIST = "999|998"
    ),
    regexp = "Found duplicated meta columns: .+MISSING_LIST.+",
    perl = TRUE
  )
  expect_error(
    meta_data <- prep_create_meta(
      VAR_NAMES = letters,
      DATA_TYPE = DATA_TYPES$INTEGER,
      LABEL = LETTERS,
      XYZ = 1:3,
      MISSING_LIST = "999|998"
    ),
    regexp = paste("The given variable attributes have different lengths and",
                   "cannot be equalized by repeating the shorter ones to",
                   "create a data frame with as 26 rows .the longest variable",
                   "attribute vector provided.. R would say:",
                   "arguments imply differing number of rows: 1, 3, 26"),
    perl = TRUE
  )
})

test_that("prep_map_labels works", {
  skip_on_cran()
  
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                   dataquieR.ERRORS_WITH_CALLER = TRUE,
                   dataquieR.WARNINGS_WITH_CALLER = TRUE,
                   dataquieR.MESSAGES_WITH_CALLER = TRUE)
  meta_data <- prep_create_meta(
    VAR_NAMES = c("ID", "SEX", "AGE", "DOE"),
    LABEL = c("Pseudo-ID", "Gender", "Age", "Examination Date"),
    DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER, DATA_TYPES$INTEGER,
                  DATA_TYPES$DATETIME),
    MISSING_LIST = ""
  )
  expect_equal(prep_map_labels(c("AGE", "DOE"), meta_data),
                   c(AGE = "Age", DOE = "Examination Date"))
  expect_error(prep_map_labels(c("NOT_AVAIL", "AGE"), meta_data),
              regexp = "value for .+NOT_AVAIL.+ not found", perl = TRUE)
  expect_equal(prep_map_labels(c("AGE", "NOT_AVAIL"), meta_data,
                               ifnotfound = NA_character_),
               c(AGE = "Age", NOT_AVAIL = NA_character_))
  expect_equal(prep_map_labels(c("AGE", "NOT_AVAIL"), meta_data,
                               ifnotfound = 42),
               c(AGE = "Age", NOT_AVAIL = 42))
  expect_error(prep_map_labels(meta_data = 42),
               regexp =
                 paste(".+meta_data.+ is not a data frame")
  )
  expect_error(prep_map_labels(x = function(){},
                               meta_data = meta_data),
               regexp =
                 paste("all variable source names must be characters")
  )
  expect_error(prep_map_labels(x = meta_data$VAR_NAMES,
                               meta_data = meta_data,
                               to = 42),
               regexp =
                 paste("Need exactly one existing variable",
                       "attribute name to map variables to")
  )
  expect_error(prep_map_labels(x = meta_data$VAR_NAMES,
                               meta_data = meta_data,
                               to = letters[1:5]),
               regexp =
                 paste("Need exactly one existing variable",
                       "attribute name to map variables to")
  )
  expect_error(prep_map_labels(x = meta_data$VAR_NAMES,
                               meta_data = meta_data,
                               to = "vegetable soup"),
               regexp =
                 paste("Need exactly one existing variable",
                       "attribute name to map variables to")
  )

  expect_error(prep_map_labels(x = meta_data$VAR_NAMES,
                               meta_data = meta_data,
                               from = 42),
               regexp =
                 paste("Need exactly one variable",
                       "attribute name to use as variable name on mapping")
  )
  expect_error(prep_map_labels(x = meta_data$VAR_NAMES,
                               meta_data = meta_data,
                               from = letters[1:5]),
               regexp =
                 paste("Need exactly one variable",
                       "attribute name to use as variable name on mapping")
  )
  expect_error(prep_map_labels(x = meta_data$VAR_NAMES,
                               meta_data = meta_data,
                               from = "vegetable soup"),
               regexp =
                 paste("Need exactly one variable",
                       "attribute name to use as variable name on mapping")
  )


})

#' Return messages/warnings/notes/error messages for a result
#'
#' @param result a `dataquieR_resultset2` result
#' @param aspect an aspect/problem category of results
#' @param collapse either a lambda function or a separator for combining
#'                 multiple messages for the same result
#' @param ... not used
#'
#' @return hover texts for results with data quality issues,
#'         run-time errors, warnings or notes (aka messages)
#'
#' @family summary_functions
#' @concept reporting
#' @noRd
util_get_message_for_result <- function(result,
                                   aspect = c("applicability", "error",
                                              "anamat", "indicator_or_descriptor"),
                                   collapse = "\n<br />\n", ...) {
  # check if the aspect is an allowed name (robustness)
  aspect <- util_match_arg(aspect, several_ok = FALSE)

  if(aspect == "indicator_or_descriptor"){
    return("")
  }

  if (!inherits(result, "dataquieR_result")) {
    return("No results computed")
  }

  if (!(aspect %in% c("applicability", "anamat"))) {
    expected_a_result <-
      (util_get_category_for_result(result, "applicability") %in% c(cat1, cat2, cat3))
    # TODO: Fix error must be added in squarereportrenderer
    expected_a_result <-
      expected_a_result && !is.na(util_get_category_for_result(result, "anamat"))

  } else {
    expected_a_result <- NA
  }
  msgs <- character(0)
  if (length(attr(result, "message")) > 0) {
    for (w in attr(result, "message")) {
      applicability_problem <- attr(w, "applicability_problem")
      if (is.null(applicability_problem) || is.na(applicability_problem)) {
        applicability_problem <- FALSE
      }
      intrinsic_applicability_problem <- attr(w,
                                              "intrinsic_applicability_problem")
      if (is.null(intrinsic_applicability_problem) ||
          is.na(intrinsic_applicability_problem)) {
        intrinsic_applicability_problem <- FALSE
      }
      if (aspect %in% c("applicability", "anamat")) {
        if (aspect == "applicability" &&
            applicability_problem && !intrinsic_applicability_problem) {
          msgs <- c(msgs, paste("<span class=\"dataquieR-message-message\">",
                                gsub("\n>.*$", "", #gsub("^.*?: ", "",
                                                  conditionMessage(w)),
                                "</span>"))
        } else if ((aspect == "anamat") &&
                    applicability_problem && intrinsic_applicability_problem) {
          msgs <- c(msgs, paste("<span class=\"dataquieR-message-message\">",
                                gsub("\n>.*$", "", #gsub("^.*?: ", "",
                                                  conditionMessage(w)),
                                "</span>"))
        }
      } else {
        if (!applicability_problem && !intrinsic_applicability_problem) {
          msgs <- c(msgs, paste("<span class=\"dataquieR-message-message\">",
                                gsub("\n>.*$", "", #gsub("^.*?: ", "",
                                                  conditionMessage(w)),
                                "</span>"))
        }
      }
    }
  }
  if (length(attr(result, "warning")) > 0) {
    for (w in attr(result, "warning")) {
      applicability_problem <- attr(w, "applicability_problem")
      if (is.null(applicability_problem) || is.na(applicability_problem)) {
        applicability_problem <- FALSE
      }
      intrinsic_applicability_problem <- attr(w,
                                              "intrinsic_applicability_problem")
      if (is.null(intrinsic_applicability_problem) ||
          is.na(intrinsic_applicability_problem)) {
        intrinsic_applicability_problem <- FALSE
      }
      if (aspect == "applicability") {
        if (applicability_problem && !intrinsic_applicability_problem) {
          msgs <- c(msgs, paste("<span class=\"dataquieR-warning-message\">",
                                gsub("\n>.*$", "", #gsub("^.*?: ", "",
                                     conditionMessage(w)),
                                "</span>"))
        }
      } else if (aspect == "anamat") {
        if (applicability_problem && intrinsic_applicability_problem) {
          msgs <- c(msgs, paste("<span class=\"dataquieR-warning-message\">",
                                gsub("\n>.*$", "", #gsub("^.*?: ", "",
                                     conditionMessage(w)),
                                "</span>"))
        }
      } else {
        if (!applicability_problem && !intrinsic_applicability_problem) {
          msgs <- c(msgs, paste("<span class=\"dataquieR-warning-message\">",
                                gsub("\n>.*$", "", #gsub("^.*?: ", "",
                                     conditionMessage(w)),
                                "</span>"))
        }
      }
    }
  }
  if (length(attr(result, "error")) > 0) {
    for (w in attr(result, "error")) {
      applicability_problem <- attr(w, "applicability_problem")
      if (is.null(applicability_problem) || is.na(applicability_problem)) {
        applicability_problem <- FALSE
      }
      intrinsic_applicability_problem <- attr(w,
                                              "intrinsic_applicability_problem")
      if (is.null(intrinsic_applicability_problem) ||
          is.na(intrinsic_applicability_problem)) {
        intrinsic_applicability_problem <- FALSE
      }
      if (aspect %in% c("applicability", "anamat")) {
        if (aspect == "applicability" &&
            applicability_problem && !intrinsic_applicability_problem) {
          msgs <- c(msgs, paste("<span class=\"dataquieR-error-message\">",
                                gsub("\n>.*$", "", #gsub("^.*?: ", "",
                                                  conditionMessage(w)),
                    "</span>"))
        } else if ((aspect == "anamat") &&
                    applicability_problem && intrinsic_applicability_problem) {
          msgs <- c(msgs, paste("<span class=\"dataquieR-error-message\">",
                                gsub("\n>.*$", "", #gsub("^.*?: ", "",
                                                  conditionMessage(w)),
                    "</span>"))
        }
      } else {
        if (!applicability_problem) {
          msgs <- c(msgs, paste("<span class=\"dataquieR-error-message\">",
                                gsub("\n>.*$", "", #gsub("^.*?: ", "",
                                                  conditionMessage(w)),
                    "</span>"))
        }
      }
    }
  }

  msgs <- unique(msgs)

  if (is.function(collapse)) {
    collapse(rev(msgs))
  } else {
    paste0(rev(msgs), collapse = collapse)
  }
}

#' Convert item-level metadata from v1.0 to v2.0
#'
#' This function is idempotent..
#'
#' The `options("dataquieR.force_item_specific_missing_codes")` (default
#' `FALSE`) tells the system, to always fill in `res_vars` columns to the
#' `MISSING_LIST_TABLE`, even, if the column already exists, but is empty.
#'
#' @param item_level [data.frame] the old item-level-metadata
#' @param meta_data [data.frame] old name for `item_level`
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param verbose [logical] display all estimated decisions, defaults to `TRUE`,
#'                          except if called in a [dq_report2] pipeline.
#' @param cause_label_df [data.frame] missing code table, see [cause_label_df].
#'                                    Optional. If this argument is given,
#'                                    you can add missing code tables.
#'
#' @return [data.frame] the updated metadata
#' @export
prep_meta_data_v1_to_item_level_meta_data <- function(item_level = "item_level",
                                                      verbose = TRUE,
                                                      label_col = LABEL,
                                                      cause_label_df,
                                                      meta_data = item_level) {

  # TODO: Rename, the function normalizes item_level metdata, not only in v1.0.
  # TOOD: Amend documentation: If you call it with v2 metadata, you can specify MISSING_LIST_TABLE; if so, then this table is searched using prep_get_dataframe, so it must have been loaded before,
  #        e.g. sing prep_load_workbook_like_file("inst/extdata/meta_data_extended.xlsx") or prep_add_data_frames(...). As a result, JUMP_LIST, MISSING_LIST and MISSING_LIST_TABLE will be consistent.
  #       JUMP_LIST and MISSING_LIST are used by all old functions to find information about jump codes and missing codes.
  #       MISSING_LIST_TABLE is compatible with the data frame passed as `cause_label_df` to `com_item_missingness`, so it works also with `prep_add_cause_label_df` and can be extracted form the JUMP_LIST and MISSING_LIST (without the v2 mapping columns for qualified mssingness) using `prep_extract_cause_label_df`
  # TODO: Re-Check, if besides prep_prepare_dataframes, all other a places, where meta_data is used, call this function

  if (missing(verbose) && .called_in_pipeline) {
    verbose <- FALSE
  }

  if (identical(attr(meta_data, "version"), 2)) {
    return(.util_internal_normalize_meta_data(meta_data, label_col))
  }

  util_expect_data_frame(meta_data,
                         col_names = list(VAR_NAMES = is.character),
                         convert_if_possible = list(VAR_NAMES = as.character))
  util_expect_scalar(verbose, check_type = is.logical)

  if (!LABEL %in% colnames(meta_data)) {
    meta_data[[LABEL]] <- meta_data[[VAR_NAMES]]
  }

  if (!label_col %in% colnames(meta_data)) {
    util_warning("Missing %s from %s.",
                 dQuote(label_col),
                 sQuote("meta_data"))
    meta_data[[label_col]] <- NA_character_
  }

  # End digits requested? ----
  if (DATA_ENTRY_TYPE %in% colnames(meta_data) &&
      END_DIGIT_CHECK %in% colnames(meta_data)) {
    util_warning("Have %s and %s in %s, will prefer %s.",
                 sQuote(END_DIGIT_CHECK),
                 sQuote(DATA_ENTRY_TYPE),
                 sQuote("meta_data"),
                 sQuote(END_DIGIT_CHECK))
    meta_data[[DATA_ENTRY_TYPE]] <- NULL
  }

  if (!END_DIGIT_CHECK %in% colnames(meta_data) &&
      DATA_ENTRY_TYPE %in% colnames(meta_data)) {
    meta_data[[END_DIGIT_CHECK]] <-
      !tolower(trimws(meta_data[[DATA_ENTRY_TYPE]])) %in% c("auto",
                                                            "automatic",
                                                            "machine",
                                                            "computational",
                                                            "electronic") &
      !util_is_na_0_empty_or_false(meta_data[[DATA_ENTRY_TYPE]])
    meta_data[[DATA_ENTRY_TYPE]] <- NULL
  }

  if (!END_DIGIT_CHECK %in% colnames(meta_data)) { # defaults to FALSE
    meta_data[[END_DIGIT_CHECK]] <- rep(FALSE, nrow(meta_data))
  }

  # KEY_STUDY_SEGMENT -> STUDY_SEGMENT and PART_VAR ----
  if (KEY_STUDY_SEGMENT %in% colnames(meta_data)) {
    if (!all(meta_data[[KEY_STUDY_SEGMENT]] %in% c(NA_character_,
                                                   meta_data[[VAR_NAMES]]))) {
      if (all(meta_data[[KEY_STUDY_SEGMENT]] %in% c(NA_character_,
                                                    meta_data[[label_col]]))) {
        util_message(
          c("All entries in %s are %s. Could convert them to",
            "%s."),
          dQuote(KEY_STUDY_SEGMENT),
          dQuote(label_col),
          dQuote(VAR_NAMES),
          applicability_problem = TRUE
        )
        meta_data[[KEY_STUDY_SEGMENT]] <-
          util_map_labels(meta_data[[KEY_STUDY_SEGMENT]],
                          meta_data,
                          to = VAR_NAMES,
                          from = label_col,
                          ifnotfound = meta_data[[KEY_STUDY_SEGMENT]])
      } else if (all(meta_data[[KEY_STUDY_SEGMENT]] %in% c(NA_character_,
                                                           meta_data[[LABEL]]
                                                           ))) {
        util_message(
          c("All entries in %s are %s. Could convert them to",
            "%s."),
          dQuote(KEY_STUDY_SEGMENT),
          dQuote(LABEL),
          dQuote(VAR_NAMES),
          applicability_problem = TRUE
        )
        meta_data[[KEY_STUDY_SEGMENT]] <-
          util_map_labels(meta_data[[KEY_STUDY_SEGMENT]],
                          meta_data,
                          to = VAR_NAMES,
                          from = LABEL,
                          ifnotfound = meta_data[[KEY_STUDY_SEGMENT]])
      } else {
        not_matched <- !(meta_data[[KEY_STUDY_SEGMENT]] %in%
                           c(NA_character_, meta_data[[VAR_NAMES]]))
        meta_data[[KEY_STUDY_SEGMENT]][not_matched] <-
          util_map_labels(meta_data[[KEY_STUDY_SEGMENT]][not_matched],
                          meta_data,
                          to = VAR_NAMES,
                          from = label_col,
                          ifnotfound = meta_data[[KEY_STUDY_SEGMENT]][not_matched])
        meta_data[[KEY_STUDY_SEGMENT]][not_matched] <-
          util_map_labels(meta_data[[KEY_STUDY_SEGMENT]][not_matched],
                          meta_data,
                          to = VAR_NAMES,
                          from = LABEL,
                          ifnotfound =
                            meta_data[[KEY_STUDY_SEGMENT]][not_matched])
        not_matched <- !(meta_data[[KEY_STUDY_SEGMENT]] %in%
                           c(NA_character_, meta_data[[VAR_NAMES]]))
        util_message(
          c("Not all entries in %s are found in %s. Could convert %d of",
            "these %d to match %s. This may be caused by providing",
            "subsets of %s."),
          dQuote(KEY_STUDY_SEGMENT),
          dQuote(VAR_NAMES),
          length(not_matched) - sum(not_matched, na.rm = TRUE),
          length(not_matched),
          dQuote(VAR_NAMES),
          dQuote("meta_data"),
          applicability_problem = TRUE
        )
      }
    }
  }

  if (KEY_STUDY_SEGMENT %in% colnames(meta_data)) {
    if (all(meta_data[[KEY_STUDY_SEGMENT]] %in%
            c(meta_data[[VAR_NAMES]], NA_character_))) {
      # KEY_STUDY_SEGMENT is PART_VAR, new STUDY_SEGMENT is the LABEL of the
      # referred variable
      if (verbose)
        util_message("%s is %s, new %s is the %s of the referred variable",
                     sQuote(KEY_STUDY_SEGMENT),
                     sQuote(PART_VAR),
                     sQuote(STUDY_SEGMENT),
                     sQuote(LABEL))
      meta_data[[PART_VAR]] <- meta_data[[KEY_STUDY_SEGMENT]]
      meta_data[[STUDY_SEGMENT]] <-
        util_map_labels(meta_data[[KEY_STUDY_SEGMENT]],
                        meta_data = meta_data,
                        warn_ambiguous = FALSE,
                        ifnotfound = meta_data[[KEY_STUDY_SEGMENT]])
      meta_data[[KEY_STUDY_SEGMENT]] <- NULL
    } else if (!any(meta_data[[KEY_STUDY_SEGMENT]] %in%
                    c(meta_data[[VAR_NAMES]]))) { # NA_character_ is not allowed here, because NA should not cause doubts, if keystudysegment contains varnames, NA is not a varname
      # KEY_STUDY_SEGMENT is STUDY_SEGMENT, PART_VAR is empty
      if (verbose)
        util_message("%s is %s, %s is empty",
                     sQuote(KEY_STUDY_SEGMENT),
                     sQuote(STUDY_SEGMENT),
                     sQuote(PART_VAR))
      meta_data[[STUDY_SEGMENT]] <-
        meta_data[[KEY_STUDY_SEGMENT]]
      meta_data[[KEY_STUDY_SEGMENT]] <- NULL
    } else {
      util_error(c("Some, but not all entries in %s found in %s. So, I cannot",
                   "decide, if %s contains links to %s or labels for %s only."),
                 sQuote(KEY_STUDY_SEGMENT), sQuote(VAR_NAMES),
                 sQuote(KEY_STUDY_SEGMENT), sQuote(PART_VAR),
                 sQuote(STUDY_SEGMENT))
    }
  }

  if (PART_VAR %in% colnames(meta_data)) {
    if (!all(meta_data[[PART_VAR]] %in% c(NA_character_,
                                          meta_data[[VAR_NAMES]]))) {
      if (all(meta_data[[PART_VAR]] %in% c(NA_character_,
                                           meta_data[[label_col]]))) {
        # util_message(
        #   c("All entries in %s (old: %s) are %s. Could convert them to",
        #     "standard, i.e., %s."),
        #   dQuote(PART_VAR),
        #   dQuote(KEY_STUDY_SEGMENT),
        #   dQuote(label_col),
        #   dQuote(VAR_NAMES),
        #   applicability_problem = TRUE
        # )
        meta_data[[PART_VAR]] <-
          util_map_labels(meta_data[[PART_VAR]],
                          meta_data,
                          to = VAR_NAMES,
                          from = label_col,
                          ifnotfound = meta_data[[PART_VAR]])
      } else if (all(meta_data[[PART_VAR]] %in% c(NA_character_,
                                                  meta_data[[LABEL]]))) {
        # TODO: Be robust, if LABEL, LONG_LABEL, VAR_NAMES and label_col are mixed
        # util_message(
        #   c("All entries in %s (old: %s) are %s. Could convert them to",
        #     "standard, i.e., %s."),
        #   dQuote(PART_VAR),
        #   dQuote(KEY_STUDY_SEGMENT),
        #   dQuote(LABEL),
        #   dQuote(VAR_NAMES),
        #   applicability_problem = TRUE
        # )
        meta_data[[PART_VAR]] <-
          util_map_labels(meta_data[[PART_VAR]],
                          meta_data,
                          to = VAR_NAMES,
                          from = LABEL,
                          ifnotfound = meta_data[[PART_VAR]])
      } else {
        not_matched <- !(meta_data[[PART_VAR]] %in% c(NA_character_,
                                                      meta_data[[VAR_NAMES]]))
        meta_data[[PART_VAR]][not_matched] <-
          util_map_labels(meta_data[[PART_VAR]][not_matched],
                          meta_data,
                          to = VAR_NAMES,
                          from = label_col,
                          ifnotfound = meta_data[[PART_VAR]][not_matched])
        meta_data[[PART_VAR]][not_matched] <-
          util_map_labels(meta_data[[PART_VAR]][not_matched],
                          meta_data,
                          to = VAR_NAMES,
                          from = LABEL,
                          ifnotfound = meta_data[[PART_VAR]][not_matched])
        not_matched <- !(meta_data[[PART_VAR]] %in%
                           c(NA_character_, meta_data[[VAR_NAMES]]))
        util_message(
          c("Not all entries in %s (old: %s) are found in %s.",
            "Could convert %d of",
            "these %d to match %s. This may be caused by providing",
            "subsets of %s."),
          dQuote(PART_VAR),
          dQuote(KEY_STUDY_SEGMENT),
          dQuote(VAR_NAMES),
          length(not_matched) - sum(not_matched, na.rm = TRUE),
          length(not_matched),
          dQuote(VAR_NAMES),
          dQuote("meta_data"),
          applicability_problem = TRUE
        )
      }
    }
  }

  # KEY_DATETIME -> TIME_VAR ----
  if (KEY_DATETIME %in% colnames(meta_data)) {
    meta_data[[TIME_VAR]] <- meta_data[[KEY_DATETIME]]
    meta_data[[KEY_DATETIME]] <- NULL
  }

  # GROUP_VAR_OBSERVER, GROUP_VAR_DEVICE -- KEY_* -> GROUP_VAR_* ----
  colnames(meta_data) <- gsub("KEY_(.*)",
                             "GROUP_VAR_\\1",
                             perl = TRUE,
                             colnames(meta_data))

  if (!missing(cause_label_df)) {
    util_expect_data_frame(cause_label_df,
                           c(CODE_VALUE, CODE_LABEL))
    meta_data[] <-
      prep_add_cause_label_df(meta_data = meta_data,
                              cause_label_df = cause_label_df,
                              label_col = VAR_NAMES)
  }

  attr(meta_data, "version") <- 2

  return(.util_internal_normalize_meta_data(meta_data, label_col = label_col))
}


#' Make normalizations of v2.0 item_level metadata.
#'
#' Requires referred missing-tables being available by
#' `prep_get_data_frame`.
#'
#' @inheritParams prep_meta_data_v1_to_item_level_meta_data
#' @noRd
.util_internal_normalize_meta_data <- function(meta_data = "item_level",
                                               label_col = LABEL,
                                               verbose = TRUE) {
  if (!identical(attr(meta_data, "version"), 2)) {
    util_error(c(
      "Internal error: .util_internal_normalize_meta_data called on meta_data",
      "below v2.0"))
  }

  meta_data <- util_handle_complex_data_types(meta_data)

  if (VALUE_LABELS %in% names(meta_data)) {
    # this should be checked outside  (!identical(attr(meta_data,
    # "normalized"), TRUE)), because the user could have modified processed
    # metadata adding a column VALUE_LABELS in between, so the data claims to
    # be normalized, but it isn't any more. However, the existence of the
    # column VALUE_LABELS will always tell us, if we have something to
    # do, here. Maybe, we need to use check-sums for such attributes that
    # express the grade of normalization of an object?
    meta_data <- util_normalize_value_labels(meta_data = meta_data)
  }

  already_warned <- c()
  already_warned_env <- environment()

  if (!identical(attr(meta_data, "normalized"), TRUE)) {

    for (lb in intersect(c(LABEL, LONG_LABEL, label_col, grep("^LONG_LABEL_",
                                                    names(meta_data),
                                                    value = TRUE,
                                                    perl = TRUE),
                 grep("^LABEL_",
                      names(meta_data),
                      value = TRUE,
                      perl = TRUE)), colnames(meta_data))) {
      .mis <- util_empty(meta_data[[lb]])
      rep <- character(nrow(meta_data))
      if (LABEL %in% names(meta_data)) rep <- meta_data[[LABEL]]
      .nolb <- util_empty(rep)

      if (any(.nolb)) {
        util_warning(c("Some variables have no label in %s in %s.",
                       "Labels are required to create a report,",
                       "that's why missing labels will be replaced",
                       "provisionally. Please add missing labels in",
                       "your metadata."),
                     dQuote(LABEL),
                     sQuote("meta_data"),
                     applicability_problem = TRUE
        )
      }

      rep[.nolb] <- meta_data[[VAR_NAMES]][.nolb]
      if (any(.mis)) {
        meta_data[[lb]][.mis] <- rep[.mis]
      }
    }

    if (!N_RULES %in% names(meta_data)) {
      meta_data[[N_RULES]] <- NA_integer_
    }

    if (!UNIVARIATE_OUTLIER_CHECKTYPE %in% names(meta_data)) {
      meta_data[[UNIVARIATE_OUTLIER_CHECKTYPE]] <- NA_character_
    }

    if (is.null(meta_data[[MISSING_LIST_TABLE]])) {
      meta_data[[MISSING_LIST_TABLE]] <- rep(NA_character_, nrow(meta_data))
    }
    no_mlt <- util_empty(meta_data[[MISSING_LIST_TABLE]])
    has_both_mlt_and_ml_jl <- !no_mlt &
      (!util_empty(meta_data[[JUMP_LIST]]) |
         !util_empty(meta_data[[MISSING_LIST]]))
    if (any(has_both_mlt_and_ml_jl)) { # have to combine lists
      util_warning(c("Should not have %s and %s / %s for the same variable.",
                     "I'll try to combine them. This will take some time, so",
                     "you should fix this in your %s to avoid waiting."),
                   dQuote(MISSING_LIST_TABLE),
                   dQuote(MISSING_LIST),
                   dQuote(JUMP_LIST),
                   sQuote("meta_data"),
                   applicability_problem = TRUE)
      for (r in which(has_both_mlt_and_ml_jl)) {
        tb <- data.frame(CODE_VALUE = character(0),
                         CODE_LABEL = character(0),
                         CODE_INTERPRET = character(0))
        try({
          tb <- prep_get_data_frame(meta_data[r, MISSING_LIST_TABLE])
          if (!CODE_VALUE %in% colnames(tb)) {
            tb[[CODE_VALUE]] <- NA
          }
          tb <- tb[!util_empty(tb[[CODE_VALUE]]), ,
                                 FALSE]
          if (CODE_INTERPRET %in% colnames(tb) &&
              !CODE_CLASS %in% colnames(tb)) {
            tb[[CODE_CLASS]] <- ifelse(tb[[CODE_INTERPRET]] %in%
                                           c("NE"), "JUMP", "MISSING") # TODO: Verify, if, e.g., PL is also a JUMP
          }
        })
        legacy_tb <- data.frame(CODE_VALUE = character(0),
                                CODE_LABEL = character(0))
        try({
          legacy_tb <- prep_extract_cause_label_df(
            meta_data[r, , drop = FALSE], label_col = label_col)$cause_label_df
          util_stop_if_not(length(unique(legacy_tb$resp_vars)) < 2)
          legacy_tb$resp_vars <- NULL # is always only for the item in row r
          if (!CODE_VALUE %in% colnames(legacy_tb)) {
            legacy_tb[[CODE_VALUE]] <- NA
          }
          legacy_tb <- legacy_tb[!util_empty(legacy_tb[[CODE_VALUE]]), ,
                                 FALSE]
        })
        cld_name <- paste0(meta_data[r, MISSING_LIST_TABLE], '_',
                           prep_link_escape(meta_data[r, label_col],
                                            html = TRUE))
        if (exists( # find an available name for a new data frame
          cld_name,
          envir = .dataframe_environment())) {
          i <- 0
          while (exists( # find an available name for a new data frame
            paste0(cld_name, "_", i),
            envir = .dataframe_environment())) i <- i + 1
          cld_name <- paste0(cld_name, "_", i)
        }
        meta_data[r, MISSING_LIST] <- NA
        meta_data[r, JUMP_LIST] <- NA
        meta_data[r, MISSING_LIST_TABLE] <- NA
        combined_table <- merge(legacy_tb, tb, by = CODE_VALUE, all = TRUE)
        if (all(c("CODE_LABEL.x", "CODE_LABEL.y") %in%
                colnames(combined_table))) {
          combined_table[util_empty(combined_table$CODE_LABEL.x) |
                           combined_table$CODE_LABEL.x ==
            paste("MISSING", combined_table$CODE_VALUE), "CODE_LABEL.x"] <-
            combined_table[util_empty(combined_table$CODE_LABEL.x) |
                             combined_table$CODE_LABEL.x ==
                             paste("MISSING", combined_table$CODE_VALUE),
                           "CODE_LABEL.y"]

          combined_table[util_empty(combined_table$CODE_LABEL.x) |
                           combined_table$CODE_LABEL.x ==
            paste("JUMP", combined_table$CODE_VALUE), "CODE_LABEL.x"] <-
            combined_table[util_empty(combined_table$CODE_LABEL.x) |
                             combined_table$CODE_LABEL.x ==
                             paste("JUMP", combined_table$CODE_VALUE),
                           "CODE_LABEL.y"]

          combined_table[util_empty(combined_table$CODE_LABEL.y) |
                           combined_table$CODE_LABEL.y ==
            paste("MISSING", combined_table$CODE_VALUE), "CODE_LABEL.y"] <-
            combined_table[util_empty(combined_table$CODE_LABEL.y) |
                             combined_table$CODE_LABEL.y ==
                             paste("MISSING", combined_table$CODE_VALUE),
                           "CODE_LABEL.x"]

          combined_table[util_empty(combined_table$CODE_LABEL.y) |
                           combined_table$CODE_LABEL.y ==
            paste("JUMP", combined_table$CODE_VALUE), "CODE_LABEL.y"] <-
            combined_table[util_empty(combined_table$CODE_LABEL.y) |
                             combined_table$CODE_LABEL.y ==
                             paste("JUMP", combined_table$CODE_VALUE),
                           "CODE_LABEL.x"]
          combined_table$CODE_LABEL <- combined_table$CODE_LABEL.x
          combined_table$CODE_LABEL.x <- NULL
          combined_table$CODE_LABEL.y <- NULL
        }
        if (all(c("CODE_CLASS.x", "CODE_CLASS.y") %in%
                colnames(combined_table))) {
          combined_table[util_empty(combined_table$CODE_CLASS.x),
                         "CODE_CLASS.x"] <-
            combined_table[util_empty(combined_table$CODE_CLASS.x),
                           "CODE_CLASS.y"]
          combined_table[util_empty(combined_table$CODE_CLASS.y),
                         "CODE_CLASS.y"] <-
            combined_table[util_empty(combined_table$CODE_CLASS.y),
                           "CODE_CLASS.x"]
          if (any(
            (!is.na(combined_table$CODE_CLASS.y) &
             combined_table$CODE_CLASS.x != combined_table$CODE_CLASS.y))) {
            util_warning(
              c("Code classes in old and new missing code settings",
                "don't match for %s, missing and jump codes overlap.",
                "Will use the new %s: %s as the reference."),
              dQuote(meta_data[r, label_col]),
              sQuote(MISSING_LIST_TABLE),
              dQuote(cld_name),
              applicability_problem = TRUE)
          }
          combined_table$CODE_CLASS <- combined_table$CODE_CLASS.y
          combined_table$CODE_CLASS.x <- NULL
          combined_table$CODE_CLASS.y <- NULL
        }
        if (prod(dim(combined_table))) {
          l <- list()
          l[[cld_name]] <- combined_table
          prep_add_data_frames(data_frame_list = l)
          # assign the new cause label df as MISSING_LIST_TABLE
          meta_data[r, MISSING_LIST_TABLE] <- cld_name
        }
      }
    }
    if (any(no_mlt)) {
      # If no missing list table, yet, for a variable
      # generate one from JUMP_LIST/MISSING_LIST
      cld <- prep_extract_cause_label_df(
        meta_data = meta_data[no_mlt, , FALSE],
        label_col = VAR_NAMES)
      meta_data[no_mlt, colnames(cld$meta_data)] <- cld$meta_data
      i <- 0
      while (exists( # find an available name for a new data frame
        paste0(MISSING_LIST_TABLE, "_", i),
        envir = .dataframe_environment())) i <- i + 1
      cld_name <- paste0(MISSING_LIST_TABLE, "_", i)

      # store in the internal data frame cache
      l <- list()

      if (prod(dim(cld$cause_label_df))) {
        l[[cld_name]] <- cld$cause_label_df
        prep_add_data_frames(data_frame_list = l)

        # assign the new cause label df as MISSING_LIST_TABLE
        meta_data[no_mlt, MISSING_LIST_TABLE] <- cld_name
      }
    }

    mlts <- mapply(
      SIMPLIFY = FALSE,
      setNames(meta_data[[MISSING_LIST_TABLE]], nm = meta_data[[VAR_NAMES]]),
      meta_data[[VAR_NAMES]],
      FUN = function(cld, vn) {
        if (all(util_empty(cld))) {
          return(NULL)
        }
        r <- data.frame()
        try(silent = TRUE,
          withCallingHandlers(
            r <- prep_get_data_frame(cld),
          error = function(cnd) {
            if (!isTRUE(already_warned_env$already_warned[cld])) {
              already_warned_env$already_warned[cld] <- TRUE
              my_cnd <- warningCondition(
                sprintf(
                  "Could not load table %s: %s",
                  dQuote(cld),
                  sQuote(sub("\nwhen calling .*$", "", conditionMessage(cnd)))
                )
              )
              util_warning(my_cnd, immediate = TRUE) # TODO: Check for sheet names very close to standard names, warn in such cases.
            }
          })
        )
        if (!prod(dim(r))) return(data.frame())
        if (!("resp_vars" %in% colnames(r))) {
          r$resp_vars <- vn
        } else {
          have_vn <- !util_empty(r$resp_vars)
          mapped_vn <- r$resp_vars[have_vn]
          other_col <- intersect(c(VAR_NAMES, LABEL, LONG_LABEL),
                                 colnames(meta_data)) # TODO: For all similar mapping functions (util_find_var_by_meta and util_correct_variable_use)
          map_res <- lapply(setNames(nm = other_col), function(oc) {
            unname(
              util_map_labels(
                mapped_vn,
                warn_ambiguous = FALSE,
                meta_data = meta_data,
                to = VAR_NAMES,
                from = oc,
                ifnotfound = mapped_vn
              )
            )
          })
          map_res <- as.data.frame(t(as.data.frame(map_res)))
          mapped_vn <-
            vapply(
              FUN.VALUE = character(1),
              map_res,
              FUN = function(mr) {
                mr <- mr[!is.na(mr)]
                mr <- mr[[1]]
                as.character(mr)
                # unique(na.omit(as.character(map_res[i, ])))[1]
                # We select here the first element, such that in case of ambiguities
                # VAR_NAME is preferred over LABEL and LONG_LABEL, and LABEL over LONG_LABEL.
                # If there are no ambiguities, this works as well.
              }
            )
          r$resp_vars[have_vn] <- mapped_vn
          if (isTRUE(getOption("dataquieR.force_item_specific_missing_codes",
                               dataquieR.force_item_specific_missing_codes_default))) {
            r$resp_vars[!have_vn] <- vn
          }
          r
        }
        r[util_empty(r$resp_vars) | r$resp_vars == vn, , FALSE]
      }
    )

    mlts <- lapply(mlts, function(mlt) { # avoid coerction of missing code values
      if (CODE_VALUE %in% colnames(mlt))
        mlt[[CODE_VALUE]] <- as.character(mlt[[CODE_VALUE]])
      mlt
    })

    mlts <- util_rbind(data_frames_list = mlts)

    # Ensure, that we now only have variable specific cause_label_df rows
    if (prod(dim(mlts))) {
      util_stop_if_not(!is.null(mlts$resp_vars))
      util_stop_if_not(all(!util_empty(mlts$resp_vars)))

      # ensure, that all missing codes for an item are in MISSING/JUMP_LIST
      new <- prep_add_cause_label_df(meta_data = meta_data,
                                     cause_label_df = mlts,
                                     label_col = VAR_NAMES)

      new_cols <- setdiff(colnames(new), colnames(meta_data))
      if (length(new_cols)) {
        meta_data[, new_cols] <- NA
      }

      lost_cols <- setdiff(colnames(meta_data), colnames(new))
      if (length(lost_cols)) {
        new[, lost_cols] <- NA
      }

      util_stop_if_not(length(intersect(colnames(meta_data), colnames(new))) ==
                         length(union(colnames(meta_data), colnames(new))))

      if (!all(colnames(meta_data) == colnames(new))) {
        new <- new[, colnames(meta_data)]
      }

      meta_data[] <- new
    }
    attr(meta_data, "normalized") <- TRUE
  }
  meta_data
}

# TODO: add some metadata management functios (write to XLSX, edit, ...)

#' utility function for the outliers rule of Hubert and Vandervieren 2008
#'
#' function to calculate outliers according to the rule of Huber et al. This
#' function requires the package `robustbase`
#'
#' @param x [numeric] data to check for outliers
#'
#' @return binary vector
#'
#' @family outlier_functions
#' @concept outlier
#' @noRd

util_hubert <- function(x) {
  lth <- robustbase::adjboxStats(x, doScale = FALSE)$stats[1]
  uth <- robustbase::adjboxStats(x, doScale = FALSE)$stats[5]
  xbin <- ifelse(x < lth | x > uth, 1, 0)
  return(xbin)
}

util_plotly_text <- function(text) {
  util_ensure_suggested("plotly")
  util_expect_scalar(text, check_type = is.character)
  text <- paste(text, collapse = "\n");
  if (suppressWarnings(util_ensure_suggested("cli", err = FALSE)))
    text <- cli::ansi_strip(text)

  plotly::plot_ly(
    type = "scatter",
    mode = "text",
    x = 0.5,
    y = 0.5,
    text = text,
    textposition = "middle center"
  ) %>%
    plotly::layout(
      xaxis = list(visible = FALSE, range = c(0, 1)),
      yaxis = list(visible = FALSE, range = c(0, 1)),
      showlegend = FALSE
    )
}

#' Utility function to create a margins plot for binary variables
#'
#' @param resp_vars  [variable] the name of the binary measurement variable
#' @param group_vars [variable] the name of the observer, device or
#'                              reader variable
#' @param co_vars [variable list] a vector of covariables, e.g. age and sex for
#'                              adjustment
#' @param min_obs_in_subgroup [integer] from=0. This optional argument specifies
#'                       the minimum number of observations that is required to
#'                       include a subgroup (level) of the `group_var` in the
#'                       analysis.
#' @param min_obs_in_cat [integer] This optional argument specifies the minimum
#'                       number of observations that is required to include
#'                       a category (level) of the outcome (`resp_vars`) in
#'                       the analysis.
#' @param ds1 [data.frame] the data frame that contains the measurements, after
#'                       replacing missing value codes by `NA`, excluding
#'                       inadmissible values and transforming categorical
#'                       variables to factors.
#' @param label_col [variable attribute] the name of the column in the metadata
#'                       with labels of variables
#' @param threshold_type [enum] empirical | user | none. See `acc_margins`.
#' @param threshold_value [numeric] see `acc_margins`
#' @param caption [string] a caption for the plot (optional, typically used to
#'                       report the coding of cases and control group)
#' @param adjusted_hint [character] hint, if adjusted for `co_vars`
#' @param title [character] title for the plot
#' @param sort_group_var_levels [logical] Should the levels of the grouping
#'                        variable be sorted descending by the number of
#'                        observations (in the figure)?
#' @param include_numbers_in_figures [logical] Should the figure report the
#'                        number of observations for each level of the grouping
#'                        variable?
#' @param no_overall_in_bin [logical] Suppress overall distribution in 'margins'
#'                                    figures for binary outcomes
#' @param no_geom_count_in_bin [logical] Suppress counts 'margins'
#'                                       figures for binary outcomes, so they
#'.                                      are not always including 0 and 1.
#'
#' @return A table and a matching plot.
#'
#' @importFrom ggplot2 ggplot geom_violin geom_boxplot geom_pointrange geom_text element_blank element_text geom_density coord_flip annotate ggplot_build theme_minimal labs theme scale_colour_manual geom_count aes geom_vline geom_hline
#' @import patchwork
#'
#' @noRd
util_margins_bin <- function(resp_vars = NULL, group_vars = NULL, co_vars = NULL,
                             threshold_type = NULL, threshold_value,
                             min_obs_in_subgroup = 5, min_obs_in_cat = 5,
                             caption = NULL, ds1, label_col,
                             adjusted_hint = "",
                             title = "",
                             sort_group_var_levels =
                               getOption("dataquieR.acc_margins_sort",
                                         dataquieR.acc_margins_sort_default),
                             include_numbers_in_figures =
                               getOption("dataquieR.acc_margins_num",
                                         dataquieR.acc_margins_num_default),
                             no_overall_in_bin =
                               getOption("dataquieR.no_overall_in_bin",
                                         dataquieR.no_overall_in_bin_default),
                             no_geom_count_in_bin =
                               getOption("dataquieR.no_geom_count_in_bin",
                                      dataquieR.no_geom_count_in_bin_default)) {
  # preps and checks -----------------------------------------------------------
  # to avoid "no visible binding for global variable ‘sample_size’"

  if (no_geom_count_in_bin &&  # so it does not scale to 0/1
      !no_overall_in_bin) { # but we want to have the overall distribution
    util_message("Cannot have %s = %s and %s = %s, setting %s to %s",
                 sQuote("no_geom_count_in_bin"),
                 dQuote(TRUE),
                 sQuote("no_overall_in_bin"),
                 dQuote(FALSE),
                 sQuote("no_overall_in_bin"),
                 dQuote(TRUE)
    )
    no_overall_in_bin <- TRUE
  }

  sample_size <- NULL
  # ensure that the response variable is not constant
  var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])
  if (var_prop$NDistinct < 2) {
    util_error("The response variable is constant (after data preparation).",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }
  # ensure that the response variable is binary
  if (var_prop$NDistinct != 2) {
    util_error("The response variable is not binary.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }
  # ensure that there are enough observations for the model
  count_bin <- util_table_of_vct(ds1[[resp_vars]])
  if (any(count_bin[, 2] < min_obs_in_cat)) {
    util_error("Not enough observations per category (after data preparation).")
  }
  count_bin <- util_table_of_vct(ds1[[group_vars]])
  if (any(count_bin[, 2] < min_obs_in_subgroup)) {
    util_error("Not enough observations per group (after data preparation).")
  }

  # modelling ------------------------------------------------------------------
  # if no co_vars are defined for adjustment only the intercept is modelled
  if (length(co_vars) == 0) {
    co_vars <- "1"
    co_vars_bQ <- co_vars
  } else {
    co_vars_bQ <- util_bQuote(co_vars)
  }

  # build model formula
  fmla <- as.formula(paste0(
    paste0(util_bQuote(resp_vars), "~"),
    paste0(
      paste0(co_vars_bQ, collapse = " + "),
      " + ",
      util_bQuote(group_vars)
    )
  ))

  suppressWarnings({
    model <- glm(fmla, data = ds1, family = binomial(link = "logit"))
  })
  res_df <- data.frame(emmeans::emmeans(model, group_vars, type = "response"),
                       check.names = FALSE)

  # adjust for covariates, if needed
  ds1$resp_var_adj <-
    # estimated mean for each level of the grouping variable
    res_df$prob[match(ds1[[group_vars]], res_df[, group_vars])] +
    # residuals: original value of the response variable - fitted value
    ds1[[resp_vars]] - model$fitted.values

  summary_ds <- as.data.frame(
    dplyr::summarize(
      dplyr::group_by_at(ds1[, c(resp_vars, group_vars), drop = FALSE],
                         unname(group_vars)),
      sample_size = dplyr::n()))

  res_df <- merge(res_df, summary_ds, by = group_vars, all.x = TRUE)
  res_df <- dplyr::rename(res_df, c("margins" = "prob", "LCL" = "asymp.LCL",
                                    "UCL" = "asymp.UCL"))

  # adjusted overall mean
  omv <- data.frame(emmeans::emmeans(model, "1", type = "response"))
  omv <- dplyr::rename(omv, c("margins" = "prob", "LCL" = "asymp.LCL",
                              "UCL" = "asymp.UCL"))
  res_df$overall <- omv$margins # TODO: never used?

  # thresholds -----------------------------------------------------------------
  if (threshold_type %in% c("empirical", "none")) {
    th <- mean(ds1[["resp_var_adj"]]) # TODO: use estimate in 'omv'?
    th <- th * (1 - th) # TODO: Is variance here a good estimator? Maybe align
    # with measurements of deviation for nominal variables?
    parn <- c(
      paste("-", threshold_value, "TH", sep = ""), "Prob.",
      paste("+", threshold_value, "TH", sep = "")
    )
    pars <- as.vector(c(
      omv$margins - threshold_value * th, omv$margins,
      omv$margins + threshold_value * th
    ))

    res_df$threshold <- threshold_value
    res_df$GRADING <- ifelse(res_df$margins < pars[1] |
                               res_df$margins > pars[3], 1, 0)
  } else if (threshold_type == "user") {
    th <- threshold_value
    parn <- c("", paste0("Prob.=", threshold_value, sep = ""), "")
    pars <- as.vector(c(th, th, th))

    res_df$threshold <- th
    res_df$GRADING <- mapply(
      function(th, l, u) {
        ifelse(th >= l & th <= u, 0, 1)
      },
      res_df$threshold, res_df$LCL, res_df$UCL
    )
  }

  # figure ---------------------------------------------------------------------
  # drop confidence intervals for group_var levels that have zero variance to
  # ensure the readability of the figures
  res_df_plot <- res_df[, c(group_vars, "margins", "LCL", "UCL",
                            "GRADING", "sample_size")]
  gr_var <- tapply(ds1[["resp_var_adj"]], ds1[[group_vars]], var)
  gr_zero <- gr_var < sqrt(.Machine$double.eps)
  if (any(gr_zero)) {
    gr_ci_excl <- names(gr_var)[gr_zero]
    res_df_plot$LCL[as.character(res_df_plot[, group_vars]) %in% gr_ci_excl] <- NA
    res_df_plot$UCL[as.character(res_df_plot[, group_vars]) %in% gr_ci_excl] <- NA
  }

  # use offset for annotation depending on variable scale
  offs <- ifelse(th <= 50, th / 10, th / 20)

  warn_code <- c("1" = "#B2182B", "0" = "#2166AC")

  if (sort_group_var_levels) {
    # order levels of the grouping variable by number of observations
    tbl_gr <- util_table_of_vct(ds1[[group_vars]]) %>%
      dplyr::arrange(dplyr::desc(.data$Freq))
    ds1[[group_vars]] <- factor(ds1[[group_vars]],
                                levels = as.character(tbl_gr$Var1))
    res_df <- res_df[match(levels(ds1[[group_vars]]),
                           as.character(res_df[, group_vars])), ]
  }

  # Plot 1: hybrid density/boxplot graph
  .ds00 <- ds1[, c("resp_var_adj", group_vars), drop = FALSE]
  if (all(!is.finite(.ds00$resp_var_adj))) {
    util_error("No data left.")
  }

  if (no_geom_count_in_bin) { # FIXME: Elisa had some corner cases, where this did not match the overall distribution any more.
    y_lims <- c(
      max(0, min(res_df_plot$LCL, na.rm = TRUE)),
      min(1, max(res_df_plot$UCL, na.rm = TRUE))
    )
    ### extract 0/1-values and group
    # tmp_count_df <- data.frame(
    #   .grp = .ds00[[group_vars]],
    #   .y = round(.ds00[["resp_var_adj"]])
    # )
    #
    # ## only 0/1
    # tmp_count_df <- tmp_count_df[!is.na(tmp_count_df$.grp) &
    #                                !is.na(tmp_count_df$.y) &
    #                                tmp_count_df$.y %in% c(0, 1), , drop = FALSE]
    #
    # ## frequencies
    # tab <- table(tmp_count_df$.grp, tmp_count_df$.y)
    #
    # ## ensure coluns 0 and 1
    # tab_df <- data.frame(
    #   .grp = rownames(tab),
    #   n0 = if ("0" %in% colnames(tab)) tab[, "0"] else 0,
    #   n1 = if ("1" %in% colnames(tab)) tab[, "1"] else 0,
    #   row.names = NULL,
    #   check.names = FALSE
    # )
    #
    # ## add to res_df_plot
    # label_df <- merge(
    #   data.frame(.grp = res_df_plot[[group_vars]], res_df_plot, check.names = FALSE),
    #   tab_df,
    #   by = ".grp",
    #   all.x = TRUE,
    #   sort = FALSE
    # )
    #
    # ## NAs -> 0
    # label_df$n0[is.na(label_df$n0)] <- 0
    # label_df$n1[is.na(label_df$n1)] <- 0
    #
    # ## create labels
    # label_df$label <- paste0(label_df$n0, "+", label_df$n1)
    gcnt <- NULL
    # summary_ds$sample_size <- paste0(summary_ds$sample_size,
    #                                  "=\n",
    #                                 label_df$label)
  } else {
    y_lims <- c(0, 1)
    gcnt <- geom_count(aes(alpha = 0.9), color = "gray")
  }

  p1 <- util_create_lean_ggplot(ggplot(data = .ds00,
                                       aes(x = .data[[group_vars]],
                                           y = round(.data[["resp_var_adj"]]))) +
                                  gcnt +
                                  util_geom_pointrange_robust(data = res_df_plot, aes(
                                    x = .data[[group_vars]],
                                    y = margins,
                                    ymin = LCL,
                                    ymax = UCL,
                                    color = as.factor(GRADING)),#, n = sample_size
                                    shape = 18,
                                    linewidth = 1,
                                    inherit.aes = FALSE,
                                    size = .5) +
                                  theme_minimal() +
                                  labs(x = "", y = "") +
                                  theme(
                                    legend.position = "none",
                                    legend.title = element_blank(),
                                    text = element_text(size = 16),
                                    axis.text.x = element_text(angle = 90,
                                                               vjust = 0.5,
                                                               hjust = 1),
                                    plot.margin = ggplot2::unit(c(2, 0, 2, 0),
                                                                "mm")) +
                                  scale_colour_manual(values = warn_code) +
                                  ggplot2::expand_limits(y = y_lims),
                                .ds00 = .ds00,
                                group_vars = group_vars,
                                res_df_plot = res_df_plot,
                                warn_code = warn_code,
                                y_lims = y_lims,
                                gcnt = gcnt)

  p1 <- util_create_lean_ggplot(p1 +
                                  ggplot2::coord_cartesian(ylim = y_lims,
                                                            clip = "off"),
                                p1 = p1,
                                y_lims = y_lims)

    if (include_numbers_in_figures) {
      p1 <- util_create_lean_ggplot(p1 +
                                      geom_text(data = summary_ds,
                                                aes(x = .data[[group_vars]],
                                                    y = Inf,
                                                    label = sample_size),
                                                inherit.aes = FALSE,
                                                hjust = 0.5,
                                                vjust = -0.2,
                                                angle = 90) +
                                      annotate("text",
                                               x = 0.5,
                                               y = Inf,
                                               label = "N",
                                               vjust = -0.2) +
                                      theme(plot.margin = ggplot2::unit(c(8, 0, 2, 0),
                                                                        "mm")),
                                    p1 = p1,
                                    summary_ds = summary_ds,
                                    group_vars = group_vars,
                                    y_lims = y_lims,
                                    sample_size = sample_size)
    }

  if (threshold_type != "none") {
    p1 <- util_create_lean_ggplot(p1 +
                                    geom_hline(yintercept = pars[2],
                                               color = "red") +
                                    geom_hline(yintercept = pars[-2],
                                               color = "red",
                                               linetype = 2),
                                  p1 = p1,
                                  pars = pars)

  } else {
    p1 <- util_create_lean_ggplot(p1 +
                                    geom_hline(yintercept = pars[2],
                                               color = "red"),
                                  p1 = p1,
                                  pars = pars)
  }

  # Plot 2: overall distributional plot flipped on y-axis of plot 1
  .ds01 <- ds1[, "resp_var_adj", drop = FALSE]
  get_y_scale <-
    util_create_lean_ggplot(ggplot(.ds01,
                                   aes(x = round(.data[["resp_var_adj"]]))) +
                              geom_density(alpha = 0.35),
                            .ds01 = .ds01)

  build <- ggplot2::ggplot_build(get_y_scale)
  data1 <- util_gg_get(build, "data")[[1]]
  yvals <- util_gg_get(data1, "y")

  aty <- mean(range(yvals))


  p2 <- util_create_lean_ggplot(ggplot(.ds01,
                                       aes(round(.data[["resp_var_adj"]]))) +
                                  geom_density(alpha = 0.35) +
                                  coord_flip() + # util_lazy_add_coord(p, fli)
                                  theme_minimal() +
                                  labs(x = NULL, y = NULL) +
                                  ggplot2::xlim(c(min(min(round(.ds01)), pars),
                                                  max(max(round(.ds01)) + 0.3,
                                                      pars + offs))) +
                                  theme(axis.text.y = element_blank(),
                                        axis.text.x = element_blank(),
                                        text = element_text(size = 16),
                                        plot.margin =
                                          ggplot2::unit(c(0, 2, 0, 2), "mm")) +
                                  ggplot2::expand_limits(x = y_lims),
                                .ds01 = .ds01,
                                pars = pars,
                                offs = offs,
                                y_lims = y_lims)

  if (threshold_type != "none") {
    p2 <- util_create_lean_ggplot(p2 +
                                    annotate(geom = "text",
                                             x = pars + offs,
                                             y = aty,
                                             label = parn) +
                                    geom_vline(xintercept = pars[2],
                                               color = "red") +
                                    geom_vline(xintercept = pars[-2],
                                               color = "red",
                                               linetype = 2),
                                  p2 = p2,
                                  pars = pars,
                                  offs = offs,
                                  aty = aty,
                                  parn = parn)

  } else {
    p2 <- util_create_lean_ggplot(
      p2 +
        annotate(geom = "text",
                 x = pars + offs,
                 y = aty,
                 label = c("", parn[2], "")) +
        geom_vline(xintercept = pars[2], color = "red"),
      p2 = p2,
      pars = pars,
      offs = offs,
      aty = aty,
      parn = parn)
  }

  if (is.null(caption)) {
    caption <- ""
  }

  # combine plots and add the title
  if (no_overall_in_bin) {
    p2 <- NULL
    my_plot_layout <- plot_layout(nrow = 1)
  } else {
    my_plot_layout <- plot_layout(nrow = 1,
                                  widths = c(5, 1))
  }

  res_plot <- # TODO: For all patchwork-calls, add information to reproduce the layout in plot.ly
    util_create_lean_ggplot(
      p1 +
        p2 +
        my_plot_layout +
        plot_annotation(title = title,
                        subtitle = adjusted_hint,
                        caption = caption),
      p1 = p1,
      p2 = p2,
      title = title,
      adjusted_hint = adjusted_hint,
      caption = caption,
      my_plot_layout = my_plot_layout)

  # output ---------------------------------------------------------------------
  return(list("plot_data" = res_df, "plot" = res_plot))
}

#' return all matches of an expression
#'
#' @param data a character vector
#' @param pattern a character string containing a regular expression
#'
#' @return A list with matching elements or NULL (in case on non-matching elements)
#' @author Josh O'Brien
#' @seealso [Stack Overflow](https://stackoverflow.com/a/10215257)
#' @examples
#' \dontrun{ # not exported, so not tested
#' dat0 <- list("a sentence with citation (Ref. 12), (Ref. 13), and then (Ref. 14)",
#'   "another sentence without reference")
#' pat <- "Ref. (\\d+)"
#' util_extract_matches(dat0, pat)
#' }
#'
#' @family process_functions
#' @concept string
#' @noRd
util_extract_matches <- function(data, pattern) {
  util_expect_scalar(pattern)
  util_expect_scalar(data,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     allow_na = TRUE)
  .util_extract_matches <- function(dt) {
    start <-  gregexpr(pattern, dt)[[1]]
    stop  <-  start + attr(start, "match.length") - 1
    if(-1 %in% start) {
      NULL
    } else {
      mapply(substr, start, stop, MoreArgs = list(x = dt))
    }
  }
  lapply(data, .util_extract_matches)
}

util_try_with_trace <- function(expr, silent = FALSE) {
  force(silent)
  tryCatch(
    expr,
    error = function(cnd) {
      # Attach an rlang backtrace to ANY error (base or rlang)
      cnd <- rlang::cnd_entrace(cnd)

      msg <- conditionMessage(cnd)

      if (!silent) {
        # Print a compact message; keep the heavy trace in the attribute
        message("Error: ", msg)
      }

      # Return a classic try-error *character* vector with a "condition" attribute
      structure(
        msg,
        class = "try-error",
        condition = cnd
      )
    }
  )
}

statichtmlWidget <- function(html, width = NULL, height = NULL,
                             elementId = NULL, sizingPolicy,
                             js = list()) {
  util_ensure_suggested("htmlwidgets", "Render results in RMarkdown")
  if (missing(sizingPolicy)) sizingPolicy <- htmlwidgets::sizingPolicy(fill = TRUE)
  util_ensure_suggested("htmltools", "Render results in RMarkdown")
  #o <- htmltools::knit_print.shiny.tag(html)

  #deps <- c(attr(o, "knit_meta"))
  #  deps <- c(list(rmarkdown::html_dependency_jquery()), attr(o, "knit_meta")) # order matters, jquery before potential datatable.js
  #  deps <- c(attr(o, "knit_meta"), list(rmarkdown::html_dependency_jquery())) # order matters, jquery after potential plotly

  o <- htmltools::as.tags(html)

  deps <- htmltools::findDependencies(o)
  deps <- c(list(rmarkdown::html_dependency_jquery()), deps) # order matters, jquery before potential datatable.js
  # forward options using x
  deps <- deps[!duplicated(
    vapply(deps, "[[", "name", FUN.VALUE = character(1)))]
  x = list(
    html = as.character(htmltools::div(class = "htmlwidget_container",
                                       o))#htmltools::HTML(o)))
    , js = htmlwidgets::JS(js)

  )

  # create widget
  w <- htmlwidgets::createWidget(
    name = 'statichtmlWidget',
    dependencies = deps,
    x,
    width = width,
    height = height,
    package = 'dataquieR',
    elementId = elementId,
    sizingPolicy = sizingPolicy
  )

  #if (any(grepl("plotly", o))) browser()

  w
}

test_that("statichtmlWidget works", {
  skip_if_not_installed("htmlwidgets")

  shw <- statichtmlWidget(htmltools::HTML("<div>Hello!!</div>"),
                          width = 100,
                          height = 20,
                          elementId = "asdf",
                          sizingPolicy =
                            htmlwidgets::sizingPolicy(),
                          js = "alert('Hello!')")
  tmpdir <- withr::local_tempdir()
  tmpfil <- file.path(tmpdir, "index.html")
  htmltools::save_html(shw, tmpfil)
  expect_true(file.exists(tmpfil))
  expect_gt(file.size(tmpfil), 620)
  expect_true(dir.exists(file.path(tmpdir, "lib")))
  expect_true(any(grepl('id="asdf"',
        fixed = TRUE,
        x = readLines(tmpfil))))
  expect_true(any(grepl("<div>Hello!!<\\/div>",
                        fixed = TRUE,
                        x = readLines(tmpfil))))
  expect_true(any(grepl("alert('Hello!')",
                        fixed = TRUE,
                        x = readLines(tmpfil))))

})

#' Get data types from data
#'
#' @param resp_vars [variable] names of the variables to fetch the data type
#'                             from the data
#' @param study_data [data.frame] the data frame that contains the measurements
#'                                Hint: Only data frames supported, no URL
#'                                or file names.
#' @param .dont_cast_off_cols [logical] internal use, only
#' @param guess_character [logical] guess a data type for character columns
#'                                  based on the values
#'
#' @return vector of data types
#' @export
#' @importFrom stats setNames
#' @seealso [prep_robust_guess_data_type()]
#'
#' @examples
#' \dontrun{
#' dataquieR::prep_datatype_from_data(cars)
#' }
prep_datatype_from_data <-
  function(resp_vars = colnames(study_data), study_data,
           .dont_cast_off_cols = FALSE,
           guess_character = getOption("dataquieR.guess_character",
                                       default =
                                         dataquieR.guess_character_default
           )) {
  util_expect_scalar(guess_character, check_type = is.logical)
  if (!missing(resp_vars) && is.data.frame(resp_vars) && missing(study_data)) {
    study_data <- resp_vars
    resp_vars <- colnames(study_data)
  } else if (missing(study_data) || !is.data.frame(study_data)) {
    util_error("Need study data as a data frame in the argument %s",
               dQuote("study_data"))
  }

  study_data <- util_cast_off(study_data, "study_data", .dont_cast_off_cols)

  if (ncol(study_data) == 0) {
    return(character(0))
  }

  if (!is.character(resp_vars)) {
    util_error(
      "%s should be missing or give variable names referring the study_data.",
      dQuote("resp_vars"), applicability_problem = TRUE)
  }

  if (!(all(resp_vars %in% colnames(study_data)))) {
    util_message(c(
      "The following %s are missing from the %s.",
      "Won't return a type for them: %s"),
      dQuote("resp_vars"),
      dQuote("study_data"),
      sQuote(resp_vars[!(resp_vars %in% colnames(study_data))]),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE
    )
  }

  types <- vapply(setNames(nm = resp_vars), FUN.VALUE = character(1),
                  function(variable) {
    if (variable %in% colnames(study_data)) {
      r <- prep_dq_data_type_of(study_data[[variable]],
                                guess_character = guess_character)
      if (length(r) == 0) {
        r <- DATA_TYPES$STRING
      }
      r
    } else {
      NA_character_
    }
  })
  types
}

test_that("prep_datatype_from_data works", {
  skip_on_cran()
  expect_equal(prep_datatype_from_data(cars),
               c(speed = "integer", dist = "integer"))
  expect_equal(prep_datatype_from_data("speed", cars),
               c(speed = "integer"))
  expect_message2(
    expect_equal(prep_datatype_from_data("speedx", cars),
               c(speedx = NA_character_)),
    regexp = paste("The following .+resp_vars.+ are missing from",
                   "the .+study_data.+.",
                   "Won't return a type for them: .+speedx.+"),
    perl = TRUE
  )
  expect_error(prep_datatype_from_data("xx"),
               regexp = paste("Need study data as a data",
                              "frame in the argument .+study_data.+"),
               perl = TRUE)
  expect_error(prep_datatype_from_data(),
               regexp = paste("Need study data as a data",
                              "frame in the argument .+study_data.+"),
               perl = TRUE)
  expect_error(
    prep_datatype_from_data(42, cars),
    regexp =
      paste(".+resp_vars.+ should be missing",
        "or give variable names referring the study_data."
      ),
    perl = TRUE)

  xx <- cars
  class(xx$speed) <- "unknown_class"
  expect_equal(prep_datatype_from_data(xx),
               c(speed = "string", dist = "integer"))

})

test_that("prep_datatype_from_data works with tibble", {
  skip_on_cran()
  skip_if_not_installed("tibble")

  expect_equal(prep_datatype_from_data(tibble::as_tibble(cars)),
               c(speed = "integer", dist = "integer"))
})

#' Split table with mixed code/missing lists to single tables
#'
#' resulting tables are populated to the data frame cache.
#'
#' @param val_tab [data.frame] tables in one long data frame.
#'
#' @return `invisible(NULL)`
#'
#' @noRd
util_split_val_tab <- function(val_tab = CODE_LIST_TABLE) {
  util_expect_data_frame(val_tab, c(CODE_VALUE, CODE_CLASS))

  if (all(c(MISSING_LIST_TABLE, VALUE_LABEL_TABLE) %in% colnames(val_tab))) {
    if (length(setdiff(intersect(val_tab[[MISSING_LIST_TABLE]],
                  val_tab[[VALUE_LABEL_TABLE]]), NA)) > 0) {
      util_warning(c("Found the some table names",
                     "assigned as value and as well as as missing list table"),
                   applicability_problem = TRUE
                   )
    }
  }

  if (CODE_LIST_TABLE %in% colnames(val_tab)) {
    if (any(c(MISSING_LIST_TABLE, VALUE_LABEL_TABLE) %in% colnames(val_tab))) {
      util_warning("Have %s as well as %s or %s in %s, ignoring %s.",
                   sQuote(CODE_LIST_TABLE),
                   sQuote(MISSING_LIST_TABLE),
                   sQuote(VALUE_LABEL_TABLE),
                   dQuote(val_tab),
                   sQuote(CODE_LIST_TABLE),
                   applicability_problem = TRUE)
      val_tab[[CODE_LIST_TABLE]] <- NULL
    } else {
      vtn <- trimws(val_tab[[CODE_LIST_TABLE]])
      vtn <- vtn[!is.na(vtn)]
      val_tab[[MISSING_LIST_TABLE]] <- ""
      val_tab[[VALUE_LABEL_TABLE]] <- ""
      for (tbn in vtn) {
        which_mlt <- trimws(val_tab[[CODE_LIST_TABLE]]) == tbn &
          trimws(val_tab[[CODE_CLASS]]) %in% c("MISSING", "JUMP")
        if (any(which_mlt)) {
          val_tab[[MISSING_LIST_TABLE]][which_mlt] <- tbn
        }
        which_vlt <- trimws(val_tab[[CODE_LIST_TABLE]]) == tbn &
          trimws(val_tab[[CODE_CLASS]]) %in% c("VALUE")
        if (any(which_vlt)) {
          val_tab[[VALUE_LABEL_TABLE]][which_vlt] <- tbn
        }
      }
      val_tab[[CODE_LIST_TABLE]] <- NULL
    }
  }

  if (MISSING_LIST_TABLE %in% colnames(val_tab)) {
    mlt <- val_tab
    mlt[[MISSING_LIST_TABLE]][util_empty(mlt[[MISSING_LIST_TABLE]])] <-
      NA_character_
    mlts <- split(mlt, mlt[[MISSING_LIST_TABLE]])

    mlts <- mlts[vapply(mlts, FUN.VALUE = logical(1),
                        function(mlt) {
                          any(mlt[[CODE_CLASS]] %in% c(CODE_CLASSES$MISSING,
                                                       CODE_CLASSES$JUMP))
                        })]
    mlts <- lapply(mlts, `[[<-`, MISSING_LIST_TABLE, value = NULL)
    mlts <- lapply(mlts, `[[<-`, VALUE_LABEL_TABLE, value = NULL)
    mlts <- lapply(mlts, `[[<-`, CODE_ORDER, value = NULL) # TODO: Support code order also for missing tables
  }

  if (VALUE_LABEL_TABLE %in% colnames(val_tab)) {
    vlt <- val_tab
    vlt[[VALUE_LABEL_TABLE]][util_empty(vlt[[VALUE_LABEL_TABLE]])] <-
      NA_character_
    vlts <- split(vlt, vlt[[VALUE_LABEL_TABLE]])

    vlts <- vlts[vapply(vlts, FUN.VALUE = logical(1),
                        function(vlt) {
                          any(vlt[[CODE_CLASS]] %in% c(CODE_CLASSES$VALUE))
                        })]
    vlts <- lapply(vlts, `[[<-`, MISSING_LIST_TABLE, value = NULL)
    vlts <- lapply(vlts, `[[<-`, VALUE_LABEL_TABLE, value = NULL)
    vlts <- lapply(vlts, `[[<-`, CODE_CLASS, value = NULL)
    vlts <- lapply(vlts, `[[<-`, CODE_INTERPRET, value = NULL)
    vlts <- lapply(vlts, function(vlt) {
      if (CODE_ORDER %in% colnames(vlt)) {
        if (all(util_empty(vlt[[CODE_ORDER]]))) {
          vlt[[CODE_ORDER]] <- NULL
        }
      }
      vlt
    })
  }

  mixed_tab_names <- intersect(names(mlts), names(vlts))
  if (length(mixed_tab_names) > 0) {
    # have tables that have the same name but feature different content,
    # missing codes as well as value codes. need to disentangle these
    util_warning(
      c("Found code-list-tables that feature missing codes",
        "as well as value codes. This is not in line with dataquieR's",
        "metadata model, you should disentangle such tables. I'll add",
        "mixed tables, but this may cause problems."),
        applicability_problem = TRUE)
    # XXX
    vlts[mixed_tab_names] <- # overwrite the vlts with the combined version
      mapply(SIMPLIFY = FALSE,
           vlt = vlts[mixed_tab_names],
           mlt = mlts[mixed_tab_names],
           FUN = function(mlt, vlt) {
             util_rbind(mlt, vlt)
           })
  }

  prep_add_data_frames(data_frame_list = mlts)
  # write vlts after mlts, so the updated combined tables from XXX will
  # prevail
  prep_add_data_frames(data_frame_list = vlts)


  invisible(NULL)
}

#' Combine all missing and value lists to one big table
#'
#' @param meta_data [data.frame] item level meta data to be used, defaults to
#'   `"item_level"`
#' @param val_tab [character] name of the table being created: This table will
#'   be added to the data frame cache (or overwritten). If `NULL`, the table
#'   will only be returned
#'
#' @return [data.frame] the combined table
#'
#' @export
prep_unsplit_val_tabs <- function(meta_data = "item_level",
                                  val_tab = NULL) {
  util_expect_data_frame(meta_data)
  meta_data <- prep_meta_data_v1_to_item_level_meta_data(
    util_normalize_value_labels(meta_data))

  if (!VALUE_LABEL_TABLE %in% colnames(meta_data)) {
    meta_data[[VALUE_LABEL_TABLE]] <- ""
  }

  if (!MISSING_LIST_TABLE %in% colnames(meta_data)) {
    meta_data[[MISSING_LIST_TABLE]] <- ""
  }

  if (!is.null(val_tab) && val_tab %in% prep_list_dataframes()) {
    util_warning("Have already a table named %s, will overwrite this table.",
                 dQuote(val_tab))
  }

  clt <- c(meta_data[[VALUE_LABEL_TABLE]],
           meta_data[[MISSING_LIST_TABLE]])

  clt <- intersect(prep_list_dataframes(), clt)

  clts <- lapply(clt, prep_get_data_frame)

  tps <- lapply(clt, function(cltnm) {
    if (cltnm %in% meta_data[[VALUE_LABEL_TABLE]]) {
      return(VALUE_LABEL_TABLE)
    } else {
      return(MISSING_LIST_TABLE)
    }
  })

  clts <- mapply(tb = clts, nm = clt, tp = tps, FUN = function(tb, nm, tp) {
    tb[[tp]] <- nm
    if (tp == VALUE_LABEL_TABLE) {
      tb[[CODE_CLASS]] <- CODE_CLASSES$VALUE
    }
    tb
  })

  cltb <- util_rbind(data_frames_list = clts)

  if (!is.null(val_tab)) {

    data_frame_list <- list()

    data_frame_list[[val_tab]] <- cltb

    prep_add_data_frames(data_frame_list = data_frame_list)

  }

  invisible(util_attach_attr(cltb, meta_data = meta_data))
}

util_handle_val_tab <- function() {
  if (CODE_LIST_TABLE %in% prep_list_dataframes()) {
    util_split_val_tab(CODE_LIST_TABLE)
    rm(list = CODE_LIST_TABLE, envir = .dataframe_environment()) # TODO: Use resp. function from Elena, once available here from the dq_report_by-branch.
  }
}

test_that("util_split_val_tab works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_purge_data_frame_cache()

  test <- tibble::tribble(
    ~ CODE_VALUE, ~ CODE_LABEL, ~ CODE_CLASS, ~ CODE_INTERPRET, ~ CODE_ORDER,
    ~ VALUE_LABEL_TABLE, ~ MISSING_LIST_TABLE,
    1234, "Test1", "MISSING", "P", 1, "vtab1", "mtab1",
    2234, "Test2", "JUMP", "P", 1, "vtab1", "mtab1",
    3234, "Test3", "VALUE", NA, 1, "vtab1", "vtab1",
    4234, "Test4", "MISSING", "P", 1, "vtab1", "mtab2",
    5234, "Test5", "MISSING", "P", 1, "vtab1", "mtab2",
    3234, "Test999", "VALUE", NA, 1, "vtab2", NA
  )
  expect_warning(util_split_val_tab(test))
  expect_equal(sort(prep_list_dataframes()),
               c("mtab1", "mtab2", "vtab1", "vtab2"))
  prep_purge_data_frame_cache()

  test <- tibble::tribble(
    ~ CODE_VALUE, ~ CODE_LABEL, ~ CODE_CLASS, ~ CODE_INTERPRET, ~ CODE_ORDER,
    ~ VALUE_LABEL_TABLE, ~ MISSING_LIST_TABLE,
    1234, "Test1", "MISSING", "P", 1, "vtab1", "mtab1",
    2234, "Test2", "JUMP", "P", 1, "vtab1", "mtab1",
    3234, "Test3", "VALUE", NA, 1, "vtab1", "",
    4234, "Test4", "MISSING", "P", 1, "vtab1", "mtab2",
    5234, "Test5", "MISSING", "P", 1, "vtab1", "mtab2",
    3234, "Test999", "VALUE", NA, 1, "vtab2", NA
  )
  util_split_val_tab(test)
  expect_equal(sort(prep_list_dataframes()),
               c("mtab1", "mtab2", "vtab1", "vtab2"))
  prep_purge_data_frame_cache()

  test <- tibble::tribble(
    ~ CODE_VALUE, ~ CODE_LABEL, ~ CODE_CLASS, ~ CODE_INTERPRET, ~ CODE_ORDER,
    ~ CODE_LIST_TABLE,
    1234, "Test1", "MISSING", "P", 1, "tab1",
    2234, "Test2", "JUMP", "P", 1, "tab1",
    3234, "Test3", "VALUE", NA, 1, "tab1",
    4234, "Test4", "MISSING", "P", 1, "tab1",
    5234, "Test5", "MISSING", "P", 1, "tab1",
    3234, "Test999", "VALUE", NA, 1, "tab2",
    3235, "Test99x9", "VALUE", NA, 2, "tab2",
    3236, "Test999a", "MISSING", NA, 1, "tab4",
    3237, "Test99x9b", "JUMP", NA, 2, "tab4",
    0, "no", "VALUE", NA, 1, "tab9",
    1, "yes", "VALUE", NA, 2, "tab9",
    3234, "MISSING", "MISSING", "P", 1, "tab9"
  )
  expect_warning(util_split_val_tab(test))
  expect_equal(prep_list_dataframes(), c("tab1", "tab2", "tab4", "tab9"))

  # prep_purge_data_frame_cache()
  # prep_load_workbook_like_file("meta_data_v2")
  # ilmd <- prep_get_data_frame("item_level")
  # ilmd <- util_normalize_value_labels(ilmd)
  # ilmd$VALUE_LABEL_TABLE[8] <- "tab9"
  # ilmd$MISSING_LIST_TABLE[8] <- "tab9"
  # ilmd$VARIABLE_ROLE[8] <- VARIABLE_ROLES$PRIMARY
  # prep_add_data_frames(item_level = ilmd)
  # prep_get_labels(ilmd$VAR_NAMES[[8]])
  # # r <- dq_report2("study_data", dimensions = NULL)
})

test_that("prep_unsplit_val_tabs works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_purge_data_frame_cache()
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  prep_unsplit_val_tabs(val_tab = CODE_LIST_TABLE)
  # prep_open_in_excel(CODE_LIST_TABLE)
  clt <- prep_get_data_frame(CODE_LIST_TABLE)
  # View(clt)
  expect_snapshot(clt)

  prep_purge_data_frame_cache()
  util_split_val_tab(clt)
  expect_snapshot(prep_list_dataframes())
})

#' Verify the class `ReportSummaryTable`
#'
#' @param tb [data.frame] object to be a `ReportSummaryTable`
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data. Used to translate
#'                               variable names, if given.
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @return [data.frame] maybe fixed `ReportSummaryTable`
#' @noRd
util_validate_report_summary_table <- function(tb, meta_data, label_col) {
  util_expect_data_frame(tb, c("Variables", "N"))
  # test, all columns except Variables numeric
  util_stop_if_not(
  `ReportSummaryTable: need to have numeric columns only (except "Variables")` =
                     all(vapply(tb[, setdiff(colnames(tb), c("Variables"))],
         is.numeric, FUN.VALUE = logical(1))))
  if (!is.factor(tb$Variables)) {
    tb$Variables <- as.character(tb$Variables)
  }
  util_stop_if_not(
    `ReportSummaryTables: The column Variables needs to be w/o NAs` =
      all(!is.na(tb$Variables)))

  tb_var_names <- NULL
  if (!is.null(attr(tb, "VAR_NAMES"))) {
    tb_var_names <- attr(tb, "VAR_NAMES")
    if (length(tb_var_names) != nrow(tb)) tb_var_names <- NULL
  }
  if (!missing(meta_data) && !missing(label_col)) {
    if (!(VARIABLE_ORDER %in% colnames(meta_data))) {
      meta_data[[VARIABLE_ORDER]] <-
        seq_len(nrow(meta_data))
    }
    # varorder <- prep_get_labels(
    #   resp_vars = tb$Variables,
    #   item_level = meta_data,
    #   label_col = VARIABLE_ORDER,
    #   resp_vars_match_label_col_only = TRUE,
    #   label_class = "SHORT",
    #   force_label_col = "TRUE"
    # )
    if (is.null(tb_var_names)) {
      tb_var_names <- try(prep_map_labels(
        as.character(tb$Variables),
        item_level = meta_data,
        to = VAR_NAMES,
        from = label_col), silent = TRUE)
      if (inherits(tb_var_names, "try-error")) tb_var_names <- NULL
    }
    if (is.factor(tb$Variables)) {
      levels(tb$Variables) <-
        prep_get_labels(
          resp_vars = levels(tb$Variables),
          item_level = meta_data,
          label_col = label_col,
          resp_vars_match_label_col_only = TRUE,
          label_class = "SHORT"
        )
    } else {
      tb$Variables <- prep_get_labels(
        resp_vars = tb$Variables,
        item_level = meta_data,
        label_col = label_col,
        resp_vars_match_label_col_only = TRUE,
        label_class = "SHORT"
      )
    }
    # tb[] <- tb[order(varorder), , FALSE]
  } else if (!missing(meta_data) || !missing(label_col)) {
    util_error(
      "ReportSummaryTable: Need to have either %s and %s or neither of them",
      sQuote("meta_data"), sQuote("label_col"))
  }
  # 'Variables' should be unique
  if (any(duplicated(tb$Variables))) {
    nm_dup <- tb$Variables[duplicated(tb$Variables)]
    ind_dup <- which(tb$Variables %in% nm_dup)
    if (!is.null(tb_var_names)) {
      tb$Variables[ind_dup] <-
        paste0(tb$Variables[ind_dup], " (", tb_var_names[ind_dup], ")")
    } else {
      tb[ind_dup, ] <- tb[ind_dup, ] %>%
        dplyr::group_by(Variables) %>%
        dplyr::mutate(Variables =
                        paste(Variables, paste0("(", dplyr::row_number(), ")"))) %>%
        as.data.frame()
    }
  }
  # If the first try fails, fix duplicates here strictly:
  if (any(duplicated(tb$Variables))) {
    nm_dup <- tb$Variables[duplicated(tb$Variables)]
    ind_dup <- which(tb$Variables %in% nm_dup)
    tb$Variables[ind_dup] <- paste(tb$Variables[ind_dup], seq_along(ind_dup))
  }
  util_stop_if_not(
    `ReportSummaryTables: The column Variables needs to be w/o duplicates` =
      !any(duplicated(tb$Variables)))
  # TODO: Also verify attributes
  class(tb) <-
    union("ReportSummaryTable", class(tb))
  if (is.null(attr(tb, "VAR_NAMES")) && !is.null(tb_var_names)) {
    attr(tb, "VAR_NAMES") <- tb_var_names
  }
  tb
}

#' Generate a full DQ report, v2
#'
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param cores [integer] number of cpu cores to use or a named list with
#'                        arguments for [parallelMap::parallelStart] or NULL,
#'                        if parallel has already been started by the caller.
#'                        Can also be a cluster.
#' @param debug_parallel [logical] print blocks currently evaluated in parallel
#' @param meta_data_segment [data.frame] -- optional: Segment level metadata
#' @param meta_data_dataframe [data.frame] -- optional: Data frame level
#'                                                                 metadata
#' @param all_calls [list] a list of calls
#' @param meta_data_cross_item [data.frame] -- optional: cross-item level
#'                                             metadata
#' @param resp_vars [variable list] the name of the measurement variables
#'                                  for the report.
#' @param filter_result_slots [character] regular expressions, only
#'                                               if an indicator function's
#'                                               result's name
#'                                               matches one of these, it'll
#'                                               be used for the report. If
#'                                               of length zero, no filtering
#'                                               is performed.
#' @param checkpoint_resumed [logical] if using a `storr_factory` and the back-
#'                                     end there is already filled
#'                                     compute all missing result and add them
#'                                     to the back-end.
#' @param content_file [character] internal use-- only for progress messages
#' @inheritParams dq_report2
#'
#' @return a [dataquieR_resultset2]. Can be printed creating a RMarkdown-report.
#'
#' @family reporting_functions
#' @concept process
#' @noRd
util_evaluate_calls <-
  function(all_calls,
           study_data,
           meta_data,
           label_col,
           meta_data_segment,
           meta_data_dataframe,
           meta_data_cross_item,
           resp_vars,
           filter_result_slots,
           cores,
           debug_parallel,
           mode = c("default", "futures", "queue", "parallel"),
           mode_args,
           my_storr_object = NULL,
           checkpoint_resumed,
           dt_adjust = as.logical(getOption("dataquieR.dt_adjust",
                                               dataquieR.dt_adjust_default)),
           content_file) {

    conds <- NULL # integrity issues outside the pipeline are collected here

    my_storr_object <- util_fix_storr_object(my_storr_object)

    function_names <- vapply(lapply(all_calls, `[[`, 1), as.character,
                             FUN.VALUE = character(1))

    if (!is.null(my_storr_object) && checkpoint_resumed) {
      all_calls_cp <- all_calls[setdiff(names(all_calls),
                                        my_storr_object$list(namespace =
                                                               util_get_storr_stat_namespace(my_storr_object)))]
    } else {
      all_calls_cp <- all_calls
    }

    # if (!is.null(my_storr_object)) {
    #   # if (inherits(my_storr_object$driver, "driver_thor")) {
    #   #   util_error( # IDEA: if we would write whenever we collect results and increase progress, we would not write concurrently to the storr.
    #   #     c("CAVEAT: LMDB databases (as used by the package thor) are",
    #   #       "not supported because of known issues with file-locking"))
    #   # }
    #   my_storr_object$set(NO_SHARED_STORR, TRUE) # will be deleted during computation
    #   # https://stackoverflow.com/a/74509506 -- be as robust as possible for local database backends
    #   invisible(lapply(names(all_calls_cp), function(slot) {
    #     invisible(util_eval_to_dataquieR_result(init = TRUE,
    #       quote({util_error(paste("No result available for unkown reasons",
    #                               "(out of memory? try to reduce the number",
    #                               "of parallel running jobs using the",
    #                               "`cores` argument)"))}),
    #       filter_result_slots = ".*", nm = slot,
    #       my_storr_object = my_storr_object,
    #       function_name = function_names[[slot]],
    #       my_call = all_calls_cp[[slot]]))
    #   }))
    #   my_storr_object$flush_cache()
    #
    # }

    if (length(mode_args) > 0) {
      if (!is.list(mode_args) || is.null(names(mode_args)) ||
          any(util_empty(names(mode_args)))) {
        util_message("%s needs to be a named list",
                     sQuote("mode_args"))
        mode_args <- list()
      }
    } else {
      mode_args <- list()
    }


    mode <- util_match_arg(mode)

    invisible(force(parallelMap::parallelGetOptions()$settings$mode)) # ensure, that this package is loaded here: it sets onLoad some options that
                                                                      # we want to overwrite below. doing so w/o having th package loaded will
                                                                      # overwrite the options, if parallelMap is acutally loaded after the option
                                                                      # setting code below.

    # maybe also https://cran.r-project.org/web/packages/parabar/readme/README.html
    # TODO: add all the objects, that Square2 writes for being full compatible
    r <- list()
    util_ensure_suggested("parallel")
    util_setup_rstudio_job(
      "Computing dq_report2, parallel computation",
      n = length(all_calls))

    progress_msg("Cluster setup", "initializing parallel mode, if applicable")
    if (!missing(content_file))
      progress_msg("Cluster setup",
                   sprintf("content_file = %s", dQuote(content_file)))

    if (mode == "queue") {
      if ((eval.parent(call("missing", as.symbol("cores"))) &&
          identical(cores,
                    eval.parent(formals(rlang::caller_fn())$cores))) ||
          missing(cores)) {
        cores <- util_detect_cores()
      }
      if (length(cores) != 1 ||
          !util_is_integer(cores) ||
          is.na(cores) ||
          cores > util_detect_cores()) {
        cores <- util_detect_cores()
        util_message(c("For mode %s, %s can only be an integer(1) <= %d,",
                       "it to its maximum"),
                     dQuote(mode),
                     sQuote("cores"),
                     cores)
      }
      q <- util_queue_cluster_setup(n_nodes = cores,
                                    progress = progress,
                                    debug_parallel = debug_parallel,
                                    my_storr_object = my_storr_object)
    } else {
      q <- NULL
    }

    if (length(all_calls_cp) > 0) {
      if (is.null(q)) {
        if (!is.null(cores)) {
          if (inherits(cores, "cluster")) {
            old_def_cl <- parallel::getDefaultCluster()
            parallel::setDefaultCluster(cores)
            on.exit(parallel::setDefaultCluster(old_def_cl), add = TRUE)
            old_o_def_cl <- options(
              parallelMap.cpus = length(parallel::getDefaultCluster()),
              parallelMap.load.balancing = TRUE,
              parallelMap.mode = "socket"
            )
            on.exit(options(old_o_def_cl), add = TRUE)
          } else if (inherits(cores, "list")) {
            suppressMessages(do.call(parallelMap::parallelStart, cores))
            on.exit({Sys.sleep(2);suppressMessages(parallelMap::parallelStop());}, add = TRUE) # whyever, rstudio needs these two seconds, it hangs, otherwise.
          } else {
            suppressMessages(parallelMap::parallelStart("socket", cpus = cores,
                                                        logging = FALSE,
                                                        load.balancing = TRUE))
            on.exit({Sys.sleep(2);suppressMessages(parallelMap::parallelStop());}, add = TRUE) # whyever, rstudio needs these two seconds, it hangs, otherwise.
          }
          cores <- NULL
        } else if
        (getOption("parallelMap.mode") != "BatchJobs" &&
         getOption("parallelMap.mode") != "batchtools" &&
         !is.null(parallel::getDefaultCluster())) {
          old_o_def_cl <- options(
            parallelMap.cpus = length(parallel::getDefaultCluster()),
            parallelMap.load.balancing = TRUE,
            parallelMap.mode = "socket"
          )
          on.exit(options(old_o_def_cl), add = TRUE)
        }

        oldO <- options(parallelMap.show.info = FALSE)
        on.exit(options(oldO), add = TRUE)

        parlib <- function(lib) {
          suppressMessages(suppressPackageStartupMessages(
            parallelMap::parallelLibrary(lib, show.info = FALSE)))
        }
        parloadNS <- function(lib) {
          .exp <- substitute({
            suppressMessages(suppressPackageStartupMessages(
              loadNamespace(lib)))
            invisible(NULL)
          })
          parexp(".exp")
          par_eval_q(eval(.exp))
        }
        parexp <- parallelMap::parallelExport
        par_eval_q <- function(expr) {
          if (is.null(parallel::getDefaultCluster())) {
            eval(expr)
          } else {
            do.call(
              parallel::clusterEvalQ,
              list(cl = NULL, substitute(expr))
            )
          }
        }

      } else {
        parloadNS <- function(lib) {
          q$workerEval(function(lib) {
            suppressMessages(suppressPackageStartupMessages(
              loadNamespace(lib)))
          }, list(lib = lib))
        }
        parlib <- function(lib) {
          q$workerEval(function(lib) {
            suppressMessages(suppressPackageStartupMessages(
              require(lib, quietly = TRUE, character.only = TRUE)))
          }, list(lib = lib))
        }
        parexp <- function(...) {
          q$export(...)
        }
        par_eval_q <- function(expr) { # may not work as expected
          q$workerEval(function(expr) {
            eval(expr)
            }, list(expr = substitute(expr)))
        }
      }

      progress_msg("Cluster setup: initializing parallel mode, if applicable", "loading library")

      if (suppressWarnings(util_ensure_suggested("pkgload", err = FALSE,
                                                 goal =
                                                 "not really needed"))) {
        dev_package <- pkgload::is_dev_package(utils::packageName())
      } else {
        dev_package <- FALSE
      }

      if (dev_package && !is.null(parallel::getDefaultCluster()) &&
          !isTRUE(getOption("dataquieR.tmp_no_load_all"))) {
        .d <- system.file(package = utils::packageName())
        .exp <- substitute({
          pkgload::load_all(path = .d)
          invisible(NULL)
        })
        parexp(".exp")
        par_eval_q(eval(.exp))
      } else {
        suppressWarnings(suppressMessages(try({
          parlib(utils::packageName())
        }, silent = TRUE)))
      }
      parloadNS("hms")

      ..e <- environment()
      conds <- list()

      suppressWarnings(suppressMessages(withCallingHandlers({

        if (dt_adjust) {
          progress_msg("Cluster setup: initializing parallel mode, if applicable",
                       "consolidating data types 1...")

          ## Adjust data type and compute int_data_type_matrix before

          withr::with_options(list(dataquieR.testdebug = TRUE), # TODO: Find an internal solution to suppress convenience argument fillers and printers
                              int_datatype_matrix_res <- int_datatype_matrix(
                                study_data = study_data,
                                meta_data = meta_data,
                                label_col = label_col
                              )
          )
        } else {
          withr::with_options(list(dataquieR.testdebug = TRUE), # TODO: Find an internal solution to suppress convenience argument fillers and printers
            int_datatype_matrix_res <- int_datatype_matrix(
              study_data = study_data[1, , FALSE],
              meta_data = meta_data,
              label_col = label_col
            )
          )
        }

      },
      condition = function(cnd) {
        ..e$conds <- c(..e$conds, list(cnd))
      })))

      conds <- unique(conds[vapply(lapply(conds, attr, "integrity_indicator"),
                                   length, FUN.VALUE = integer(1)) == 1]) # IDEA: Don't discard "other" conditions, except such with non-matching var_name-attribute, see also my_conds below.


      old_.dq2_globs.called_in_pipeline <- .dq2_globs$.called_in_pipeline
      .dq2_globs$.called_in_pipeline <- TRUE
      on.exit(.dq2_globs$.called_in_pipeline <-
                old_.dq2_globs.called_in_pipeline, add = TRUE)

      if (dt_adjust) {
        progress_msg("Cluster setup: initializing parallel mode, if applicable",
                     "consolidating data types 2...")
        study_data <- util_adjust_data_type(
          study_data = study_data,
          meta_data = meta_data,
          relevant_vars_for_warnings = NULL)
      }

      progress_msg("Cluster setup: initializing parallel mode, if applicable", "exporting data")

      suppressWarnings(parexp("study_data", "meta_data", "label_col", "meta_data_segment", "meta_data_dataframe", "meta_data_cross_item",
                              "my_storr_object"))
      .options <- options() #options to be copied to the children (child process)
      .options <- .options[startsWith(names(.options), "dataquieR.")] #only dataquieR options selected

      progress_msg("Cluster setup: initializing parallel mode, if applicable", "exporting options")

      suppressWarnings(parexp(".options"))

      progress_msg("Cluster setup: initializing parallel mode, if applicable", "exporting data frame cache")

      dataframes_list <- as.list(.dataframe_environment())
      suppressWarnings(parexp("dataframes_list"))

      progress_msg("Cluster setup: initializing parallel mode, if applicable", "exporting other caches")

      if (getOption("dataquieR.precomputeStudyData",
                    default = dataquieR.precomputeStudyData_default)) {
        cache_as_list <- as.list(.cache[[".cache"]])
        study_data_cache <- as.list(.study_data_cache)
      } else {
        cache_as_list <- list()
        study_data_cache <- list()
      }

      suppressWarnings(parexp("cache_as_list"))
      suppressWarnings(parexp("study_data_cache"))

      progress_msg("Cluster setup: initializing parallel mode, if applicable", "finalizing setup of compute nodes")

      if (!is.null(q) || !is.null(parallel::getDefaultCluster())) {
        par_eval_q({
          ..glbs <- get(".dq2_globs", envir = asNamespace("dataquieR"))
          ..glbs$.called_in_pipeline <- TRUE
        })
        par_eval_q(options(.options))
        par_eval_q(dataquieR::prep_add_data_frames(
          data_frame_list = dataframes_list))
        par_eval_q({
          assign(
            x = ".cache",
            envir = get(".cache", envir = asNamespace("dataquieR")),
            value = as.environment(cache_as_list))
          list2env(study_data_cache, get(".study_data_cache",
                                         envir = asNamespace("dataquieR")))
        })
        u8 <- par_eval_q(l10n_info()[["UTF-8"]])
        u8 <- vapply(u8, identity, FUN.VALUE = logical(1))
        if (!all(u8)) {
          util_warning(c("%d of the %d cluster nodes do not support",
                         "UTF-8, this may cause trouble with the encoding.",
                         "For Windows nodes, you should use R > 4.2.0",
                         "on all nodes. Also, all nodes should use a UTF-8",
                         "character set by default (see Sys.setlocale())"),
                       sum(!u8), length(u8))
        }
      } else {
        dataquieR::prep_add_data_frames(
          data_frame_list = dataframes_list)
        assign(
          x = ".cache",
          envir = get(".cache", envir = asNamespace("dataquieR")),
          value = as.environment(cache_as_list))
      }
      # parallelMap::parallelLapply(1:10, function(x) {dataquieR::prep_list_dataframes()})

      # this exports the static data to the cluster, this is always done, even if
      # already avail. functions don't touch the exported data.
      currentCpus <- parallelMap::parallelGetOptions()$settings$cpus

      worker <- util_eval_to_dataquieR_result
      formals(worker)$filter_result_slots <- filter_result_slots
      force(formals(worker)$filter_result_slots)
      formals(worker)$checkpoint_resumed <- checkpoint_resumed
      force(formals(worker)$checkpoint_resumed)

      if (parallelMap::parallelGetOptions()$settings$mode == "local") {
        environment(worker) <- new.env(parent = asNamespace(utils::packageName()))
      } else if (parallelMap::parallelGetOptions()$settings$mode == "multicore") {
        environment(worker) <- environment()
      } else {
        environment(worker) <- asNamespace(utils::packageName())
      }

      progress_msg("Computation", "computing report")

      n_nodes <- max(1, as.integer(currentCpus[[1]]), na.rm = TRUE)

      if (mode == "futures") { # have/use futures
        r <- util_parallel_futures(
          all_calls = all_calls_cp,
          n_nodes = n_nodes,
          progress = progress,
          worker = worker,
          debug_parallel = debug_parallel,
          my_storr_object = my_storr_object
        )
      } else if (mode == "queue") {
        step <- 15
        if ("step" %in% names(mode_args)) {
          step <- mode_args[["step"]]
          if (length(step) != 1 || !is.numeric(step) ||
              !is.vector(step) || !is.finite(step) || !util_is_integer(step) ||
            step <= 0 || step > 10000) {
            util_message(
              c("%s needs to be a positive scalar integer value <= %d, falling",
                "back to default %d"),
              dQuote("step"),
              10000,
              15)
            step <- 15
          }
        }
        r <- q$compute_report(all_calls = all_calls_cp,
                              worker = worker,
                              step = step)
      } else {
        r <- util_parallel_classic(
          all_calls = all_calls_cp,
          n_nodes = n_nodes,
          progress = progress,
          worker = worker,
          debug_parallel = debug_parallel,
          my_storr_object = my_storr_object
        )
      }

      if (!dynGet(".is_testing", ifnotfound = FALSE)) {
        util_message(
          sprintf("%s [%s], %s", Sys.time(), "INFO",
                  "DQ -- done")
        ) # TODO: Use RStudio job if available
      }
    }

    r_with_already_computed_r <- setNames(r[names(all_calls)], names(all_calls))
    r_with_already_computed_r[setdiff(names(all_calls), names(r))] <- NA

    r <- r_with_already_computed_r

    my_storr_object <- util_fix_storr_object(my_storr_object)

    if (!is.null(my_storr_object) && length(unclass(r)) > 0 &&
        length(all_calls_cp) > 0 &&
        my_storr_object$exists(NO_SHARED_STORR) &&
        identical(my_storr_object$get(NO_SHARED_STORR), TRUE)
        ) {
      my_storr_object$del(NO_SHARED_STORR)
      util_error("Your storr backend must be shared amongst the compute nodes.")
    }

    progress_msg("Computation", "finalizing report")

    # old <- class(r)
    attr(r, "all_calls") <- all_calls

    class(r) <- union(dataquieR_resultset_class2,
                      "square_results")
    attr(r, "my_storr_object") <- my_storr_object
    # class(r) <- old need?
    if (!is.null(my_storr_object) &&
        inherits(my_storr_object, "storr") &&
        !util_is_try_error(try(my_storr_object$list(), silent = TRUE))) {
      # to make summary work
      atts_r <- attributes(r)
      atts_r[["my_storr_object"]] <- NULL # dont save this ever
      my_storr_object$mset(key = names(atts_r), value = atts_r, namespace =
                             util_get_storr_att_namespace(my_storr_object))
    }

    if (!is.null(my_storr_object)) {
      my_storr_object$flush_cache()
      # missing_objects <- setdiff(names(r), my_storr_object$list())
      # invisible(lapply(missing_objects, function(mo) {
      #   invisible(util_eval_to_dataquieR_result(
      #       quote({util_error(paste("No result available for unkown reasons",
      #                        "(out of memory? try to reduce the number",
      #                        "of parallel running jobs using the",
      #                        "`cores` argument)"))}),
      #       filter_result_slots = ".*", nm = mo,
      #       my_storr_object = my_storr_object,
      #       function_name = function_names[[mo]],
      #       my_call = all_calls[[mo]], checkpoint_resumed = FALSE))
      # }))
    } else {
      # r[] <- mapply(function_names, all_calls, r, attr(all_calls, "cn"),
      #               i = seq_len(length(function_names)),
      #               i_n = length(function_names),
      #               SIMPLIFY = FALSE,
      #               FUN = function(nm, cl, r, cn, i, i_n) {
      #   # util_message("Fixing %s -- %d of %d", sQuote(nm), i, i_n) # FIXME: SLOW
      #   # added to fix bug in unexpected crashing single jobs.
      #   # if (is.null(r)) {
      #   #    r <-
      #   #      util_eval_to_dataquieR_result(
      #   #        quote({util_error(paste("No result available for unkown reasons",
      #   #                         "(out of memory? try to reduce the number",
      #   #                         "of parallel running jobs using the",
      #   #                         "`cores` argument)"))}),
      #   #        filter_result_slots = ".*", nm = nm,
      #   #        function_name = nm, checkpoint_resumed = FALSE,
      #   #        my_call = cl)
      #   # }
      #
      #   # now done directly, when result is computed:
      #   # attr(r, "function_name") <- nm --> util_eval_to_dataquieR_result
      #   # attr(r, "cn") <- cn
      #   # attr(r, "call") <- cl
      #   ########################################
      #   r
      # })
    }

    # do after itdm, see below:
    # if (!is.null(my_storr_object)) {
    #   namespace = util_get_storr_summ_namespace(my_storr_object)
    #   all_sums <- my_storr_object$mget(my_storr_object$list(
    #     namespace = namespace),
    #     namespace = namespace)
    # } else {
    #   all_sums <- lapply(r, attr, "r_summary")
    # }
    #
    # # c("SummaryData", "SummaryTable", "ReportSummaryTable"), # this is always included (integrity), but later
    # rsn <- sort(unique(unname(unlist(lapply(all_sums, attr, "resnames")))))
    # attr(r, "resnames") <- rsn

    if (!is.null(my_storr_object) &&
        inherits(my_storr_object, "storr") &&
        !util_is_try_error(try(my_storr_object$list(), silent = TRUE))) {
      # to make summary work
      atts_r <- attributes(r)
      atts_r[["my_storr_object"]] <- NULL # dont save this ever
      my_storr_object$mset(key = names(atts_r), value = atts_r, namespace =
                             util_get_storr_att_namespace(my_storr_object))
    }


    if (any(is.na(names(r)))) {

      nms <- paste(
        vapply(r, attr, "cn", FUN.VALUE = character(1)),
        vapply(lapply(r, attr, "call"), function(cl) {
          attr(cl, "entity_name")
        }, FUN.VALUE = character(1)), sep = ".")

      names(r)[is.na(names(r))] <-
        nms[is.na(names(r))]
      util_stop_if_not(
        `Internal error, sorry, please report: report name inconsistency` =
          all(nms == names(r))
      )

    }

    # overwrite the result for int_datatype_matrix_res
#    if (!is.null(r)) { # TODO: could that happen?

    # add left-out integrity results to the report -----
    if (length(r)) {
      idtm_variable_labels <-
        vapply(lapply(.access_dq_rs2(r, startsWith(names(r), "int_datatype_matrix.")),
                      attr, "call"), attr, "entity_name",
               FUN.VALUE = character(1))

      int_datatype_matrix. <-
        lapply(idtm_variable_labels, function(lab) {
          res <- r[[paste0("int_datatype_matrix.", lab)]]
          attr(res, "error") <- NULL
          attr(res, "warning") <- NULL
          attr(res, "message") <- NULL
          if (!dt_adjust) {
            attr(res, "error") <- list(attr(try(util_error(
              c("data type check was disabled",
                "(argument dt_adjust or option dataquieR.dt_adjust)"),
              applicability_problem = TRUE),
              silent = TRUE), "condition"))
            attr(res, "error")[[1]]$trace <- NULL
            res$SummaryTable <- NULL
            res$SummaryData <- NULL
            res$ReportSummaryTable <- NULL
            class(res) <- union("dataquieR_NULL", class(res))
          } else {
            res$SummaryTable <-
              int_datatype_matrix_res$SummaryTable[
                int_datatype_matrix_res$SummaryTable$Variables == lab,
                , FALSE
              ]

            res$SummaryData <-
              int_datatype_matrix_res$SummaryData[
                int_datatype_matrix_res$SummaryData$Variables == lab,
                , FALSE
              ]

            res$ReportSummaryTable <- NULL
            res$ReportSummaryTable <-
              int_datatype_matrix_res$ReportSummaryTable[
                int_datatype_matrix_res$ReportSummaryTable$Variables == lab,
                , FALSE
              ]
          }
          my_conds <- conds[vapply(conds, function(cnd) {
            (identical(attr(cnd, "varname"), lab)) ||
              (is.null(cnd))
          }, FUN.VALUE = logical(1))]

          if (length(my_conds) > 0) {
            for (cnd in my_conds) {
              if (inherits(cnd, "error")) {
                attr(res, "error") <- c(attr(res, "error"),
                                        list(cnd))
              }
              if (inherits(cnd, "warning")) {
                attr(res, "warning") <- c(attr(res, "warning"),
                                          list(cnd))
              }
              if (inherits(cnd, "message")) {
                attr(res, "message") <- c(attr(res, "message"),
                                          list(cnd))
              }
            }
          }

          res
        })
      #    }

      int_datatype_matrix. <- mapply(SIMPLIFY = FALSE,
                                     r = int_datatype_matrix.,
                                     nm = names(int_datatype_matrix.),
                                     function(r, nm) {
                                       s <- prep_extract_summary(r)
                                       r_summary1 <-
                                         suppressWarnings(
                                           prep_summary_to_classes(s)) # FIXME: split this function, we do not want classes, here

                                       CAT_ <-
                                         vapply(setNames(nm = c("applicability", "error", "anamat", "indicator_or_descriptor")), function(aspect) {
                                           as.character(as.numeric(util_as_cat(util_get_category_for_result(r, aspect = aspect))))
                                         }, FUN.VALUE = character(1))

                                       MSG_ <-
                                         vapply(setNames(nm = c("applicability", "error", "anamat", "indicator_or_descriptor")), function(aspect) {
                                           util_get_message_for_result(r, aspect = aspect)
                                         }, FUN.VALUE = character(1))

                                       names(CAT_) <- paste0("CAT_", names(CAT_))
                                       names(MSG_) <- paste0("MSG_", names(MSG_))

                                       # rownames(CAT_) <- NULL
                                       # rownames(MSG_) <- NULL
                                       # rownames(r) <- NULL

                                       # if (is.data.frame(r_summary) && nrow(r_summary) > 0) {
                                       #   r_summary <- cbind.data.frame(r_summary, t(CAT_), t(MSG_))
                                       # } else {
                                       r_summary <- cbind.data.frame(t(CAT_), t(MSG_))
                                       # }
                                       my_call <- attr(r, "call")
                                       r_summary <- data.frame(
                                         VAR_NAMES = unname(attr(my_call, VAR_NAMES)),
                                         STUDY_SEGMENT = unname(attr(my_call, STUDY_SEGMENT)),
                                         call_names = unname(attr(r, "cn")),
                                         value = as.character(r_summary),
                                         values_raw = as.character(r_summary),
                                         function_name = "int_datatype_matrix",
                                         indicator_metric = names(r_summary)
                                       )
                                       r_summary <- util_rbind(r_summary, r_summary1)

                                       r_summary$function_name <- "int_datatype_matrix"

                                       attr(r_summary, "resnames") <- names(r)
                                       attr(r, "r_summary") <- r_summary
                                       if (!is.null(my_storr_object) &&
                                           inherits(my_storr_object, "storr") &&
                                           !util_is_try_error(try(my_storr_object$list(), silent = TRUE))) {
                                         my_storr_object$set(key = nm, value = r)
                                         my_storr_object$set(
                                           key = nm,
                                           value = r_summary,
                                           namespace = util_get_storr_summ_namespace(my_storr_object))
                                         my_storr_object$flush_cache()
                                         try(my_storr_object$driver$disconnect(), silent = TRUE)
                                         try(my_storr_object$driver$env$sync(force = TRUE), silent = TRUE)

                                         r <- NA
                                       } else if (!is.null(my_storr_object)) {
                                         r <- "Invalid storr object"
                                       }
                                       r
                                     })

      if (is.null(my_storr_object)) { # only the RAM based version needs this line, otherwise, we overwrite the results also in the back-end with NAs, but we won't write the r_summary, then
        .access_dq_rs2(r, startsWith(names(r), "int_datatype_matrix.")) <-
          int_datatype_matrix. # TODO: make the access function aware of potential summary attributes and write in such cases also the summary to the back-end to make the condtions here unneeded
      }

      # do after itdm, see above
      if (!is.null(my_storr_object)) {
        namespace = util_get_storr_summ_namespace(my_storr_object)
        all_sums <- my_storr_object$mget(my_storr_object$list(
          namespace = namespace),
          namespace = namespace)
      } else {
        all_sums <- lapply(r, attr, "r_summary")
      }

      rsn <- sort(unique(unname(unlist(lapply(all_sums, attr, "resnames")))))
      attr(r, "resnames") <- rsn


      # overwrite the result for int_datatype_matrix_res / DONE
    }

    # make report compatible with old Square2 reports -----

    function_names <- setNames(nm = function_names)

    aliases <- gsub("\\..*$", "", names(r))

    matrix_list <- lapply(setNames(nm = setdiff(aliases,
                                                attr(all_calls,
                                                     "multivariatcol"))),
                          function(alias) {
                            lapply(setNames(nm = resp_vars),
                                   function(vn) {
                                     if (is.call(
                                       all_calls[[paste0(alias, ".", vn)]])) {
                                       rlang::call_args(
                                         all_calls[[paste0(alias, ".", vn)]]
                                       )
                                     } else {
                                       setNames(list(), nm = character(0))
                                     }
                                   })
                          })

    function_names <- function_names[!duplicated(aliases)]
    aliases <- aliases[!duplicated(aliases)]

    descriptions = .manual$descriptions[function_names]
    if (length(descriptions) == 0) {
      descriptions <- rep(NA_character_, length(function_names))
    }
    descriptions[vapply(descriptions, is.null, FUN.VALUE = logical(1))] <-
      NA_character_
    descriptions = unlist(descriptions, recursive = FALSE)

    function_alias_map <- data.frame(
      fk_report = rep(NA_integer_, length(aliases)),
      alias = aliases,
      acronym = util_abbreviate(aliases),
      description = descriptions,
      fk_function = rep(NA_integer_, length(aliases)),
      # TODO: check if .manual actually exists
      function_description = descriptions,
      name = function_names,
      stringsAsFactors = FALSE)

    attr(matrix_list, "function_alias_map") <- function_alias_map

    fn <- unique(function_names)

    function2category <- setNames(c(
      des = "Descriptors",
      int = "Integrity",
      com = "Completeness",
      con = "Consistency",
      acc = "Accuracy"
    )[gsub(
      "_.*$",
      "",
      fn)],
    nm = fn)

    attr(matrix_list, "function2category") <- function2category

    dim_ranks <- setNames(c(
      des = 0,
      int = 1,
      com = 2,
      con = 3,
      acc = 4
    )[gsub(
      "_.*$",
      "",
      aliases)],
    nm = aliases)


    drr <- rank(dim_ranks)
    fnr <- rank(function_names)
    alr <- rank(aliases)

    base <- max(c(drr, fnr, alr, 10), na.rm = TRUE)

    attr(matrix_list, "col_indices") <- setNames(
      (drr * base * base + fnr * base + alr) * 10,
      nm = aliases)

    vo <- as.numeric(meta_data[[VARIABLE_ORDER]])
    if (0 == length(vo)) {
      vo <- seq_len(nrow(meta_data)) * 10
    }
    offset <- max(vo, na.rm = TRUE) + 1
    if (is.infinite(offset))
      offset <- 1
    vo[util_empty(vo)] <- offset + seq_len(sum(util_empty(vo)))
    attr(matrix_list, "row_indices") <- setNames(vo,
                                                 nm = meta_data[[label_col]])
    if (!length(r)) {
      r <- list()
    }
    attr(r, "matrix_list") <- matrix_list
    attr(r, "label_col") <- label_col
    attr(r, "meta_data") <- meta_data # one discapency from SQ2 reults, here: We have *always* v2.0 metadata, here. Will not yet write a back-converter, as long as this is not really needed.

    attr(r, "meta_data_segment") <- meta_data_segment
    attr(r, "meta_data_dataframe") <- meta_data_dataframe
    attr(r, "meta_data_cross_item") <- meta_data_cross_item

    # Include all tables referred to by the standard metadata ----
    meta_data_frames <-
      grep("^meta_data", names(attributes(r)), value = TRUE)

    refs <- unname(unlist(lapply(meta_data_frames,
                       function(mdf_name) {
                         mdf <- attr(r, mdf_name)
                         cls <- grep("_TABLE$", colnames(mdf), value = TRUE)
                         if (length(cls)) {
                           unlist(mdf[, cls], recursive = TRUE)
                         } else {
                           character(0)
                         }
                       }), recursive = FALSE))

    # FIXME: ensure, that no study data are exposed, here. Flag all meta data frames in the cache as exportable.

    refs <- refs[!util_empty(refs)]
    refs <- gsub(sprintf("\\s*\\%s\\s*", SPLIT_CHAR),
                 SPLIT_CHAR, refs)
    refs <- gsub(sprintf("(\\%s.*?)\\%s.*$", SPLIT_CHAR, SPLIT_CHAR),
                 "\\1", refs)
    refs <- unique(sort(refs))

    refs <- lapply(setNames(nm = refs), function(dfn) {
      r <- data.frame(`NA` = paste(dQuote(dfn), "is not available."),
                      check.names = FALSE)
      if (!getOption("dataquieR.non_disclosure",
                     dataquieR.non_disclosure_default)) { # this is anyway removed below, if the non-disclosure option is TRUE, but we do not needto add it first.
        try(
          r <- prep_get_data_frame(dfn),
          silent = TRUE
        )
      }
      r
    })

    attr(r, "referred_tables") <- refs

    # add information on the dimension names ----

    attr(r, "study_data_dimnames") <- dimnames(study_data)
    attr(r, "cn") <- attr(all_calls, "cn")
    attr(r, "rn") <- attr(all_calls, "rn")
    attr(r, "integrity_issues_before_pipeline") <- conds

    attr(r, "dt_adjust") <- dt_adjust

    # class is to be compatible with Square2 -----

    class(r) <- union(dataquieR_resultset_class2,
                      "square_results")

    if (getOption("dataquieR.non_disclosure",
                  dataquieR.non_disclosure_default)) {
      progress_msg("Computation", "undisclosing report")
      r <- util_undisclose(r)
    }

    progress_msg("Computation", "finished")

    # Return report ----
    r
}

NO_SHARED_STORR <- "NO_SHARED_STORR"

#' Add labels to `ggplot`
#'
#' EXPERIMENTAL
#'
#' @param ... EXPERIMENTAL
#' @param meta_data the metadata
#' @param label_col the label columns
#'
#' @return a modified `ggplot`
#'
#' @concept process
#' @noRd
util_gg_var_label <- function(...,
                              meta_data = get("meta_data", parent.frame()),
                              label_col = get("label_col", parent.frame())) { # nocov start

  lp <- ggplot2::last_plot()

  # ggplot_build kann (je nach Version) S7-Objekte liefern, daher später util_gg_get
  p <- ggplot2::ggplot_build(lp)

  l <- list(...)

  # ----- Labels aus dem Plot holen -------------------------------------------
  labels_obj <- util_gg_get(lp, "labels")

  if (!is.null(labels_obj)) {
    yy <- unlist(labels_obj[trimws(labels_obj) != ""],
                 recursive = FALSE)
    xx <- prep_map_labels(
      yy,
      meta_data   = meta_data,
      to          = label_col,
      ifnotfound  = NA_character_
    )
    xx <- xx[!is.na(xx)]
    zz <- setNames(xx[yy], nm = names(yy))
    l[names(zz)] <- zz
  }

  # ----- Mappings aus dem ggplot_build-Objekt holen --------------------------
  plot_obj <- util_gg_get(p, "plot")
  mapping  <- if (!is.null(plot_obj)) util_gg_get(plot_obj, "mapping") else NULL

  if (!is.null(mapping)) {
    for (n in names(mapping)) {
      r <- try({
        rlang::quo_name(mapping[[n]])
      }, silent = TRUE)

      if (!inherits(r, "try-error") &&
          length(r) == 1 &&
          is.character(r) &&
          r != "<unknown>") {
        l[n] <- r
      }
    }
  }

  # ----- Labels mappen und labs() anwenden -----------------------------------
  l <- l[!is.na(l)]

  l[] <- prep_map_labels(
    unlist(l, recursive = FALSE),
    meta_data   = meta_data,
    to          = label_col,
    ifnotfound  = setNames(nm = unlist(l, recursive = FALSE))
  )

  do.call(ggplot2::labs, l)
} # nocov end

user_hints <- new.env(parent = emptyenv())
user_hints$l <- list()
#' Save a hint to the user during package load
#'
#' @param x [character] the hint
#'
#' @return `invisible(NULL)`
#'
#' @family system_functions
#' @concept process
#' @noRd
util_user_hint <- function(x) {
  util_expect_scalar(x, check_type = is.character)
  user_hints$l <-
    c(user_hints$l, x)
}

test_that("util_user_hint works", {
  skip_on_cran()
  test_text <- sprintf(
    paste0("Without the package %s, I cannot decide, if your RStudio is",
           "at least at version 1.3.1056, so I'll activate a work-around",
           "for a known parallel-bug fixed in newer RStudios."),
    dQuote("rstudioapi")
  )

  result <- util_user_hint(test_text)

  # Check that the result is silent
  expect_silent(result)

  # Check that the result is a list
  expect_type(result, "list")
})

#' HTML Dependency for vertical headers in `DT::datatable`
#'
#' @return the dependency
html_dependency_vert_dt <- function() {
  htmltools::htmlDependency(
    name = "vertical-dt-style"
    ,version = "0.0.1"
    ,src = c(file = system.file("vertical-dt-style", package = "dataquieR"))
    ,stylesheet = "vertical-dt-style.css"
    ,script = "sort_heatmap_dt.js"
  )
}

#' Fetch a label for a variable based on its purpose
#'
#' @param resp_vars [variable list] the variable names to fetch for
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param max_len [integer] the maximum label length to return, if not possible
#'                          w/o causing ambiguous labels, the labels may still
#'                          be longer. For `label_class == "LONG"`, it defaults
#'                          to `200`, while for `label_class == "SHORT"` to `30`
#' @param label_class [enum] SHORT | LONG. which sort of label
#'                           according to the metadata model should be returned
#' @param label_lang [character] optional language suffix, if available in
#'                               the metadata. Can be controlled by the option
#'                               `dataquieR.lang`.
#' @param resp_vars_are_var_names_only [logical] If `TRUE`, do not use
#'                                               other labels than [VAR_NAMES]
#'                                               for finding `resp_vars` in
#'                                               `meta_data`
#' @param resp_vars_match_label_col_only [logical] If `TRUE`, do not use
#'                                               other labels than those,
#'                                               referred by `label_col`
#'                                               for finding `resp_vars` in
#'                                               `meta_data`
#'
#' @return [character] suitable labels for each `resp_vars`, names of this
#' vector are [VAR_NAMES]
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#' @param force_label_col [enum] auto | FALSE | TRUE. if `TRUE`, always use
#'                               labels according `label_col`, `FALSE` means
#'                               use labels matching best the function's
#'                               requirements, `auto` means `FALSE`, if in a
#'                               `dq_report()` and `TRUE`, otherwise.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' prep_load_workbook_like_file("meta_data_v2")
#' prep_get_labels("SEX_0", label_class = "SHORT", max_len = 2)
#' }
prep_get_labels <- function(
    resp_vars,
    item_level = "item_level",
    label_col,
    max_len,
    label_class = c("SHORT", "LONG"),
    label_lang = getOption("dataquieR.lang", dataquieR.lang_default),
    resp_vars_are_var_names_only = FALSE,
    resp_vars_match_label_col_only = FALSE,
    meta_data = item_level,
    meta_data_v2,
    force_label_col = getOption("dataquieR.force_label_col",
                                dataquieR.force_label_col_default)) {

  if (!missing(item_level) && !missing(meta_data) &&
      !identical(item_level, meta_data)) {
    util_error(c("You cannot provide both, %s as well as %s",
                 "these arguments are synonyms and must be",
                 "used mutually exclusively"),
               sQuote("item_level"),
               sQuote("meta_data"))
    # see prep_prepare_dataframes
  }

  util_maybe_load_meta_data_v2()

  util_expect_scalar(force_label_col,
                     check_type = is.character,
                     error_message =
                       "force_label_col must be 1 character value")

  util_match_arg(force_label_col, c(
    "auto", "FALSE", "TRUE"
  ))

  label_class <- util_match_arg(label_class)
  util_expect_scalar(label_class)
  if (missing(max_len)) {
    if (label_class == "SHORT") {
      max_len <- MAX_LABEL_LEN
    } else {
      max_len <- MAX_LONG_LABEL_LEN
    }
  }

  util_expect_scalar(max_len,
                     check_type = util_is_numeric_in(min = 1,
                                                     whole_num = TRUE,
                                                     finite = TRUE))
  if (!missing(label_col)) {
    util_expect_scalar(label_col,
                       check_type = is.character,
                       error_message =
                         sprintf("%s needs to be a character string.",
                                 sQuote("label_col")))
    no_lc <- FALSE
  } else {
    no_lc <- TRUE
    label_col <- VAR_NAMES
  }
  util_expect_scalar(label_lang,
                     check_type = is.character,
                     error_message =
                       sprintf("%s needs to be a character string.",
                               sQuote("label_lang")))
  util_expect_scalar(resp_vars_are_var_names_only,
                     check_type = is.logical,
                     error_message =
                       sprintf("%s needs to be a logical value.",
                               sQuote("resp_vars_are_var_names_only")))
  util_expect_scalar(resp_vars_match_label_col_only,
                     check_type = is.logical,
                     error_message =
                       sprintf("%s needs to be a logical value.",
                               sQuote("resp_vars_match_label_col_only")))
  if (resp_vars_match_label_col_only &&
      resp_vars_are_var_names_only) {
    util_error("Call with either %s or %s set TRUE, not both",
               sQuote("resp_vars_match_label_col_only"),
               sQuote("resp_vars_match_label_col_only"))
  }
  util_expect_data_frame(
    meta_data,
    col_names = VAR_NAMES)

  # this function does not need missing stuff and this would slow down
  # prep_meta_data_v1_to_item_level_meta_data
  meta_data[[MISSING_LIST]] <- NULL
  meta_data[[MISSING_LIST_TABLE]] <- NULL
  meta_data[[JUMP_LIST]] <- NULL
  meta_data[[VALUE_LABELS]] <- NULL
  meta_data <- prep_meta_data_v1_to_item_level_meta_data(meta_data =
                                                           meta_data,
                                                         verbose = FALSE,
                                                         label_col = label_col)

  if (missing(resp_vars)) {
    resp_vars <- meta_data[[VAR_NAMES]]
    resp_vars_are_var_names_only <- TRUE
    resp_vars_match_label_col_only <- FALSE
  }

  util_expect_scalar(resp_vars,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     allow_na = TRUE)

  if (!no_lc && !label_col %in% colnames(meta_data)) {
    util_error("%s = %s, if specified, needs to exist in %s.",
               sQuote("label_col"),
               dQuote(label_col),
               sQuote("meta_data"))
  }

  if (trimws(toupper(force_label_col)) == "AUTO") {
    force_label_col <- !no_lc && !.called_in_pipeline
  } else {
    force_label_col <- trimws(toupper(force_label_col)) == "TRUE"
  }

  if (!resp_vars_are_var_names_only &&
      !resp_vars_match_label_col_only) {
    vns <- suppressMessages(util_find_var_by_meta(
      label_col = label_col,
      resp_vars = resp_vars,
      ifnotfound = resp_vars,
      meta_data = meta_data,
      target = VAR_NAMES
    ))
  } else {
    if (resp_vars_match_label_col_only) {
      vns <- suppressMessages(prep_map_labels(
        x = resp_vars,
        meta_data = meta_data,
        from = label_col,
        to = VAR_NAMES,
        ifnotfound = resp_vars,
        warn_ambiguous = FALSE
      ))

    } else {
      vns <- suppressMessages(prep_map_labels(
        x = resp_vars,
        meta_data = meta_data,
        to = VAR_NAMES,
        ifnotfound = resp_vars,
        warn_ambiguous = FALSE
      ))
    }
  }

  if (force_label_col) {
    to <- label_col
  } else {
    if (label_class == "SHORT") {
      to <- c(LABEL, LONG_LABEL, VAR_NAMES)
    } else if (label_class == "LONG") {
      to <- c(LONG_LABEL, LABEL, VAR_NAMES)
    } else {
      # internal error
      util_error("Please excuse and inform us about Internal error 439285")
    }
  }

  to <- c(paste(to, label_lang, sep = "_"), to)  #see what happens with label_lang

  to <- c(to, gsub("_.+?$", "", setdiff(to, c(VAR_NAMES, LONG_LABEL))))

  to <- head(
    intersect(
      to,
      colnames(meta_data)),
    1)

  if (length(to) == 0) {
    return(setNames(nm = resp_vars))
  }

  uel <- .util_ensure_label(meta_data[meta_data[[VAR_NAMES]] %in% vns,
                                     c(VAR_NAMES, to),  FALSE],
                    label_col = to,
                    max_label_len = max_len)

  res <- prep_map_labels( # do not attach "to" as the used label col if the follwoing fails or vn was already not a.
    x = vns,
    meta_data = uel$meta_data,
    to = uel$label_col,
    ifnotfound = vns,
    warn_ambiguous = FALSE
  )

  res[util_empty(res)] <- resp_vars[util_empty(res)]

  util_attach_attr(res, # see above
                   label_col = to)

}

#' Get Function called for a Call Name
#'
#' get aliases from report attributes and then replace them by the actual
#' function name
#'
#' @param cll_names [character] then systematic function call name to fetch its
#'                             function name
#' @param report [dataquieR_resultset2] the report
#' @param function_alias_map [data.frame] the `function_alias_map` of a report,
#'                           alternative foe passing the report
#'
#' @return [character] the function name
#'
#' @noRd
util_cll_nm2fkt_nm <- function(cll_names, report, function_alias_map) {
  if (missing(report) && missing(function_alias_map)) {
    vapply(cll_names, function(cll_name) {
      fns <- util_all_ind_functions()
      fkt_name <- unique(names(which(vapply(fns, function(fn) {
        util_startsWith_prefix._or_equals_prefix(
          x = cll_name,
          prefix = fn,
          sep = "_")
      }, FUN.VALUE = logical(1)))))
      if (length(fkt_name) != 1) fkt_name <- cll_name
      # post-process function names
      if (fkt_name %in% c("con_hard_limits",
                          "con_soft_limits",
                          "con_detection_limits")) {
        fkt_name <- "con_limit_deviations"
      } else if (fkt_name %in% c("acc_robust_univariate_outlier")) {
        fkt_name <- "acc_univariate_outlier"
      }

      return(fkt_name)
    }, FUN.VALUE = character(1))
  } else {
    if (missing(function_alias_map)) {
      function_alias_map <-
        attributes(attributes(report)$matrix_list)$function_alias_map
    }
    vapply(cll_names,
           function_alias_map = function_alias_map,
           function(cll_name, function_alias_map) {
      fkt_name <- cll_name
      f <- cll_name
      f <- try(subset(
        function_alias_map,
        alias == cll_name, "name", drop = TRUE),
        silent = TRUE)
      if (is.vector(f) && is.character(f) && (length(f) == 1)) {
        fkt_name <- f
      }

      # post-process function names
      if (fkt_name %in% c("con_hard_limits",
                          "con_soft_limits",
                          "con_detection_limits")) {
        fkt_name <- "con_limit_deviations"
      } else if (fkt_name %in% c("acc_robust_univariate_outlier")) {
        fkt_name <- "acc_univariate_outlier"
      }

      return(fkt_name)
    }, FUN.VALUE = character(1))
  }
}

#' Clear data frame cache
#'
#' @return nothing
#'
#' @family data-frame-cache
#' @export
prep_purge_data_frame_cache <- function() {
  util_prune_odm_cache()
  util_clean_condition_once_cache()
  rm(list = ls(.dataframe_environment(), all.names = TRUE),
     envir = .dataframe_environment())
  util_message("All cached dataframes have been purged.")
}

test_that("mahalanobis works", {
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  df1 <- data.frame(
    ID = c(letters[1:26], "aa", "bb", "cc", "dd"),
    Q1 = c(5, 4, 1, 2, 2, 99999, 2,rep(1, 10), 2, 2, 2, 2, 3, 2, 1, 1, 2, 2, 2, 2, 1),
    Q2 = c(2, 5, rep(1, 4), 2, rep(1, 13), 2, 2, 1, 1, 1, 1, 1, 1, 2, 1),
    Q3 = c(4, 5, rep(1, 20),3, 1, 1, 1, 1, 1, 1, 1 ),
    Q4 = c(NA, 1, 4, 4, 5, 3, 4, 5, 3, 3, 5, 5, 4, 4, 4, 3, 4, 3, 4, 5, 5, 4, 4, 3, 4, 4, 4, 4, 5, 4),
    Q5 = c(88888, 1, 3, 4, 3, 3, 4, 3, 4, 3, 4, 3, 3, 3, 3, 2, 3, 3, 3, 3, 3, 4, 4, 3, 3, 3, 4, 3, 4, 3),
    Q6 = c(88888, 1, 5, 5, 5, 4, 5, 4, 5, 3, 4, 5, 4, 4,rep(5, 5), 4, 4, 5, 5, 4, 4, 5, 5, 4, 5, 5),
    Q7 = c(NA, 2, 3, 3, 2, rep(3, 4), 4, 3, 3, 4, 3, 3, 4, 3, 2, 3, 3, 4, 3, 4, 3, 3, 4, 3, 3, 3, 4),
    Q8 = c(NA, 1, rep(4,5), 5, 4, 4, 3, 4, 3, 3, 4, 4, 5, 5, 4, 4, 3, 3, 4, 4, 4, 4, 5, 5, 4, 4),
    Q9 = c(NA, 1, 4, 3, 3, 4, 3, 3, 2, 3, 3, 2, rep(3, 4), 4, 3, 3, 4, 3, 3, 3, 4, 3, 4, 4, 3, 3, 3),
    Q10 = c(1, 1, 2, 1, rep(2, 4), 1, 2, 2, 2, 3, 3, rep(2, 4), 3, rep(2, 4),3, 2, 2, 3, 2, 1, 2),
    stringsAsFactors = FALSE
  )

  #PREPARE_METADATA
  # Define the common value labels to avoid repetitive typing
  likert_labels <-
    "1 = not apply | 2 = rather not apply | 3 = intermediate | 4 = rather apply | 5 = apply completely"

  # Create the item_level data frame
  item_level <- data.frame(
    VAR_NAMES = c("ID", paste0("Q", 1:10), "T1", "T2"),
    LABEL = c(
      "pseudo id", "shy and reserved", "trust people", "tasks throughly",
      "relaxed", "active imagination", "sociable", "find faults in others",
      "lazy", "nervous", "artistic interest", "begin", "end"
    ),
    DATA_TYPE = c("string", rep("integer", 10), "datetime", "datetime"),
    SCALE_LEVEL = c("na", rep("ordinal", 10), "interval", "interval"),
    TIME_VAR = c(NA, rep("T1", 10), NA, NA),
    TIME_VAR_END = c(NA, rep("T2", 10), NA, NA),
    MISSING_LIST_TABLE = c(NA, rep("missing_table", 10), "missing_table", "missing_table"),
    VALUE_LABELS = c(
      NA,
      rep(likert_labels, 10),
      NA,
      NA
    ),
    STUDY_SEGMENT = c(NA, rep("Questionnaire", 12)),
    VARIABLE_ROLE = c("intro", rep("primary", 10), "process", "process"),
    stringsAsFactors = FALSE
  )

  #add it to the cache
  prep_purge_data_frame_cache()
  prep_add_data_frames(item_level)

  # Create the missing_table data frame
  missing_table <- data.frame(
    CODE_VALUE = c(99999, 99998, 99997, 88888),
    CODE_LABEL = c(
      "Missing - other reason",
      "Missing - exclusion criteria",
      "Missing - refusal",
      "JUMP 88880"
    ),
    CODE_CLASS = c("MISSING", "MISSING", "MISSING", "JUMP"),
    stringsAsFactors = FALSE
  )

  # add in the cache
  prep_add_data_frames(missing_table)


  # Create the cross_item_level data frame
  cross_item_level <- data.frame(
    VARIABLE_LIST = c(
      "Q1 | Q2 | Q3 | Q4 | Q5 | Q6",
      "Q7 | Q8 | Q9 | Q10",
      "Q1 | Q2 | Q3 | Q4 | Q5 | Q6 | Q7 | Q8 | Q9 | Q10"
    ),
    CHECK_LABEL = c(
      "First part",
      "Second part",
      "all questionnaire"
    ),
    CONTRADICTION_TERM = c(NA, NA, NA),
    CONTRADICTION_TYPE = c(NA, NA, NA),
    MULTIVARIATE_OUTLIER_CHECKTYPE = c(NA, NA, NA),
    N_RULES = c(NA, NA, NA),
    MAXIMUM_LONG_STRING = c("[;2)", "[;]", "[;3]"),
    IRV = c(NA, "[-1; 1]", NA),
    IRV_ARGS = c(NA, "20", NA),
    stringsAsFactors = FALSE
  )

  ## add in the cache
  prep_add_data_frames(cross_item_level)


  mahal_res <- acc_mahalanobis(variable_group = c("Q1", "Q2", "Q3", "Q4", "Q5", "Q6", "Q7", "Q8", "Q9", "Q10"),
                       study_data = df1,
                       meta_data = "item_level")


  expect_equal(mahal_res$SummaryTable$NUM_ssc_mah, 2)
  expect_equal(mahal_res$SummaryTable$observational_units_removed, 2)
  expect_equal(mahal_res$SummaryData$`MD_outliers (%)`, 7.14)
})

test_that("prep_get_data_frame works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_on_cran()
  prep_purge_data_frame_cache()
  expect_equal(length(prep_list_dataframes()), 0)
  expect_error(prep_get_data_frame("", .data_frame_list = cars),
               "must be an environment")
  expect_error(prep_get_data_frame("ship"),
               "shortcuts in test")
  expect_error(prep_get_data_frame("ship_subset1"),
               "shortcuts in test")
  expect_error(prep_get_data_frame("ship_meta_v2"),
               "shortcuts in test")
  expect_error(prep_get_data_frame("ship_meta_dataframe"),
               "shortcuts in test")
  expect_error(prep_get_data_frame("ship_meta"),
               "shortcuts in test")
  expect_error(prep_get_data_frame("study_data|study_data"),
               "shortcuts in test")
  expect_error(prep_get_data_frame("meta_data_v2"),
               "shortcuts in test")
  expect_error(prep_get_data_frame("meta_data"),
               "shortcuts in test")


})

#' Verify an object of class `r dataquieR_resultset_class`
#'
#' Deprecated
#'
#' @param ... Deprecated
#'
#' @return Deprecated
#'
dataquieR_resultset_verify <- function(...) {
  lifecycle::deprecate_stop("2.1.0", what = "dq_report()", with = "dq_report2()")
}

#' All available probability distributions for
#' [acc_shape_or_scale]
#'
#'  - `uniform` For uniform distribution
#'  - `normal` For Gaussian distribution
#'  - `gamma` For a gamma distribution
#'
#' @export
DISTRIBUTIONS <- list(
  UNIFORM = "uniform",
  NORMAL = "normal",
  GAMMA = "gamma"
)

#' @title Names of DQ dimensions
#' @name dimensions
#' @description
#'
#' a vector of data quality dimensions. The supported
#' dimensions are Completeness, Consistency and Accuracy.
#'
#' @seealso [Data Quality Concept](
#'   https://dataquality.qihs.uni-greifswald.de/DQconceptNew.html)
#'
#' @return Only a definition, not a function, so no return value
#'
dimensions = c("Completeness", "Consistency", "Accuracy")

#' Data Types
#'
#' ## Data Types of Study Data
#' In the metadata, the following entries are allowed for the
#' [variable attribute] [DATA_TYPE]:
#'
#'  - `integer` for integer numbers
#'  - `string` for text/string/character data
#'  - `float` for decimal/floating point numbers
#'  - `datetime` for timepoints
#'  - `time` for time of day
#'
#' ## Data Types of Function Arguments
#' As function arguments, [dataquieR] uses additional type specifications:
#'
#' - `numeric` is a numerical value ([float] or [integer]), but it is not an
#'   allowed `DATA_TYPE` in the metadata. However, some functions may accept
#'   `float` or `integer` for specific function arguments. This is, where we
#'   use the term `numeric`.
#' - `enum` allows one element out of a set of allowed options similar to
#'   [match.arg]
#' - `set` allows a subset out of a set of allowed options similar to
#'   [match.arg] with `several.ok = TRUE`.
#' - `variable` Function arguments of this type expect a character scalar that
#'              specifies one variable using the variable identifier given in
#'              the metadata attribute `VAR_NAMES` or, if `label_col` is set,
#'              given in the metadata attribute given in that argument.
#'              Labels can easily be translated using [prep_map_labels]
#' - `variable list` Function arguments of this type expect a character vector
#'                   that specifies variables using the variable identifiers
#'                   given in the metadata attribute `VAR_NAMES` or,
#'                   if `label_col` is set, given in the metadata attribute
#'                   given in that argument. Labels can easily be translated
#'                   using [prep_map_labels]
#'
#' @seealso [integer] [string]
#' @aliases integer string float datetime time numeric enum FLOAT INTEGER STRING DATETIME TIME variable set
#' @rawRd \alias{variable list}
#'
#' @export
DATA_TYPES <- list(
  INTEGER = "integer",
  STRING = "string",
  FLOAT = "float",
  DATETIME = "datetime",
  TIME = "time"
)

#' Scale Levels
#'
#' ## Scale Levels of Study Data according to `Stevens's` Typology
#' In the metadata, the following entries are allowed for the
#' [variable attribute] [SCALE_LEVEL]:
#'
#'  - `nominal` for categorical variables
#'  - `ordinal` for ordinal variables (i.e., comparison of values is possible)
#'  - `interval` for interval scales, i.e., distances are meaningful
#'  - `ratio` for ratio scales, i.e., ratios are meaningful
#'  - `na` for variables, that contain e.g. unstructured texts, `json`,
#'          `xml`, ... to distinguish them from variables, that still need to
#'          have the `SCALE_LEVEL` estimated by
#'          `prep_scalelevel_from_data_and_metadata()`
#'
#' ## Examples
#'
#' - sex, eye color -- `nominal`
#' - income group, education level -- `ordinal`
#' - temperature in degree Celsius -- `interval`
#' - body weight, temperature in Kelvin -- `ratio`
#'
#' @seealso [Wikipedia](https://en.wikipedia.org/wiki/Level_of_measurement)
#' @export
SCALE_LEVELS <-
  list(
    NOMINAL = "nominal",
    ORDINAL = "ordinal",
    INTERVAL = "interval",
    RATIO = "ratio",
    'NA' = "na"
  ) # TODO: validate item level metadata `SCALE_LEVEL` entries somewhere.

#' All available data types, mapped from their respective
#' R types
#' @seealso [`prep_dq_data_type_of`]
#' @export
DATA_TYPES_OF_R_TYPE <- list(
  # order important for prep_dq_data_type_of --
  # start with the most restrictive types
  logical = DATA_TYPES$INTEGER,
  integer = DATA_TYPES$INTEGER,
  ordered = DATA_TYPES$STRING,
  factor = DATA_TYPES$STRING,
  POSIXct = DATA_TYPES$DATETIME,
  POSIXlt = DATA_TYPES$DATETIME,
  POSIXt = DATA_TYPES$DATETIME,
  Date = DATA_TYPES$DATETIME,
  dates = DATA_TYPES$DATETIME,
  times = DATA_TYPES$TIME,
  hms = DATA_TYPES$TIME,
  times = DATA_TYPES$DATETIME,
  double = DATA_TYPES$FLOAT,
  numeric = DATA_TYPES$FLOAT,
  character = DATA_TYPES$STRING,
  ITime = DATA_TYPES$TIME,
  time_of_day = DATA_TYPES$TIME
)

#' Variable roles can be one of the following:
#'
#'   - `intro` a variable holding consent-data
#'   - `primary` a primary outcome variable
#'   - `secondary` a secondary outcome variable
#'   - `process` a variable describing the measurement process
#'   - `suppress` a variable added on the fly computing sub-reports, i.e., by
#'                [dq_report_by] to have all referred variables available,
#'                even if they are not part of the currently processed segment.
#'                But they will only be fully assessed in their real segment's
#'                report.
#'
#' @rawRd \alias{variable roles}
#' @export
VARIABLE_ROLES <- list(
  INTRO = "intro",
  PRIMARY = "primary",
  SECONDARY = "secondary",
  PROCESS = "process",
  SUPPRESS = "suppress"
)

#' @name MAXIMUM_LONG_STRING
#' @title Cross-item level metadata attribute name
#' @description
#'
#' TODO
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @family SSI
NULL

#' @name MAHALANOBIS_RATIO
#' @title Cross-item level metadata attribute name
#' @description
#'
#' TODO
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @family SSI
NULL

#' @name IRV
#' @title Cross-item level metadata attribute name
#' @description
#'
#' TODO
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @family SSI
NULL

#' @name TOTRESPT
#' @title Cross-item level metadata attribute name
#' @description
#'
#' TODO
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @family SSI
NULL

#' @name RELCOMPL_SPEED
#' @title Cross-item level metadata attribute name
#' @description
#'
#' TODO
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @family SSI
NULL

#' @name RESPT_PER_ITEM
#' @title Cross-item level metadata attribute name
#' @description
#'
#' TODO
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @family SSI
NULL

#' @name RELCOMPL_SPEED
#' @title Cross-item level metadata attribute name
#' @description
#'
#' TODO
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @family SSI
NULL

#' @name MISS_RESP
#' @title Cross-item level metadata attribute name
#' @description
#'
#' TODO
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @family SSI
NULL


#' @name COMPUTED_VARIABLE_ROLES
#' @title `SSI` related Cross-item level metadata attribute names
#' Computed Variable roles can be one of the following:
#' @description
#'
#'   - `MAXIMUM_LONG_STRING` Social Science: Computed Indicator Variable,
#'                           maximum long string
#'   - `IRV` Social Science: Computed Indicator Variable, `IRV`
#'   - `TOTRESPT` Social Science: Computed Indicator Variable, `TOTRESPT`
#'   - `RESPT_PER_ITEM` Social Science: Computed Indicator Variable, `RESPT_PER_ITEM`
#'   - `RELCOMPL_SPEED` Social Science: Computed Indicator Variable, `RELCOMPL_SPEED`
#'   - `MISS_RESP` Social Science: Computed Indicator Variable, `MISS_RESP`
#'   - `NA` Social Science: Computed Indicator Variable -- N/A
#'
#' @seealso [VARIABLE_ROLES]
#' @family meta_data_cross
#' @family SSI
NULL

#' Character used  by default as a separator in metadata such as
#'           missing codes
#'
#' This 1 character is according to our metadata concept `r dQuote(SPLIT_CHAR)`.
#' @export
SPLIT_CHAR <- "|"

# constants for the names of defined metadata attributes / meta reference
# attributes
REQUIREMENT_ATT <- "var_att_required"

#' Requirement levels of certain metadata columns
#'
#' These levels are cumulatively used by the function [prep_create_meta] and
#' related in the argument `level` therein.
#'
#' currently available:
#'
#' ```{r echo=FALSE,results='asis'}
#' cat(
#'   paste(" -", sQuote(unlist(names(VARATT_REQUIRE_LEVELS))), "=",
#'               dQuote(unlist(VARATT_REQUIRE_LEVELS)), collapse = "\n")
#' )
#' ```
#'
#' @export
#' @aliases COMPATIBILITY REQUIRED RECOMMENDED OPTIONAL TECHNICAL UNKNOWN
VARATT_REQUIRE_LEVELS <- list(
  COMPATIBILITY = "compatibility",
  REQUIRED = "required",
  RECOMMENDED = "recommended",
  OPTIONAL = "optional",
  TECHNICAL = "technical"
)

VARATT_REQUIRE_LEVELS_ORDER <- c(
  # for cumulative access from util_get_var_att_names_of_level
  VARATT_REQUIRE_LEVELS$REQUIRED,
  VARATT_REQUIRE_LEVELS$RECOMMENDED,
  VARATT_REQUIRE_LEVELS$OPTIONAL,
  VARATT_REQUIRE_LEVELS$COMPATIBILITY,
  VARATT_REQUIRE_LEVELS$TECHNICAL
)

#' Well-known metadata column names, names of metadata columns
#'
#' names of the variable attributes in the metadata frame holding
#' the names of the respective observers, devices, lower limits for plausible
#' values, upper limits for plausible values, lower limits for allowed values,
#' upper limits for allowed values, the variable name (column name, e.g.
#' v0020349) used in the study data,  the variable name used for processing
#' (readable name, e.g. RR_DIAST_1) and in parameters of the QA-Functions, the
#' variable label, variable long label, variable short label, variable data
#' type (see also [DATA_TYPES]), re-code for definition of lists of event
#' categories, missing lists and jump lists as CSV strings. For valid units
#' see [UNITS].
#'
#' all entries of this list will be mapped to the package's exported NAMESPACE
#' environment directly, i.e. they are available directly by their names too:
#'
#' `r paste0(' - [', names(WELL_KNOWN_META_VARIABLE_NAMES), ']', collapse = "\n")`
#'
#' @rawRd \alias{variable attribute}
#'
#' @eval paste("@aliases", paste(names(WELL_KNOWN_META_VARIABLE_NAMES), collapse = " "))
#'
#' @seealso [meta_data_segment] for `STUDY_SEGMENT`
#' @seealso [online](https://dataquality.qihs.uni-greifswald.de/VIN_Item_Level_Metadata.html)
#' @family UNITS
#'
#' @examples
#' print(WELL_KNOWN_META_VARIABLE_NAMES$VAR_NAMES)
#' # print(VAR_NAMES) # should usually also work
#' @export
WELL_KNOWN_META_VARIABLE_NAMES <- list(
  VAR_NAMES = structure("VAR_NAMES", var_att_required =
                          VARATT_REQUIRE_LEVELS$REQUIRED),
  LABEL = structure("LABEL", var_att_required =
                      VARATT_REQUIRE_LEVELS$RECOMMENDED),
  DATA_TYPE = structure("DATA_TYPE", var_att_required =
                          VARATT_REQUIRE_LEVELS$REQUIRED),
  SCALE_LEVEL = structure("SCALE_LEVEL", var_att_required =
                            VARATT_REQUIRE_LEVELS$RECOMMENDED),
  UNIT = structure("UNIT", var_att_required =
                            VARATT_REQUIRE_LEVELS$RECOMMENDED),
  VALUE_LABELS = structure("VALUE_LABELS", var_att_required =
                             VARATT_REQUIRE_LEVELS$RECOMMENDED),
  VALUE_LABEL_TABLE = structure("VALUE_LABEL_TABLE", var_att_required =
                             VARATT_REQUIRE_LEVELS$OPTIONAL),
  MISSING_LIST = structure("MISSING_LIST", var_att_required =
                             VARATT_REQUIRE_LEVELS$COMPATIBILITY),
  JUMP_LIST = structure("JUMP_LIST", var_att_required =
                          VARATT_REQUIRE_LEVELS$COMPATIBILITY),
  MISSING_LIST_TABLE = structure("MISSING_LIST_TABLE", var_att_required =
                                   VARATT_REQUIRE_LEVELS$REQUIRED),
  HARD_LIMITS = structure("HARD_LIMITS", var_att_required =
                            VARATT_REQUIRE_LEVELS$RECOMMENDED),
  DETECTION_LIMITS = structure("DETECTION_LIMITS", var_att_required =
                                 VARATT_REQUIRE_LEVELS$OPTIONAL),
  SOFT_LIMITS = structure("SOFT_LIMITS", var_att_required =
                            VARATT_REQUIRE_LEVELS$OPTIONAL),
  CONTRADICTIONS = structure("CONTRADICTIONS", var_att_required =
                               VARATT_REQUIRE_LEVELS$OPTIONAL),
  DISTRIBUTION = structure("DISTRIBUTION", var_att_required =
                             VARATT_REQUIRE_LEVELS$OPTIONAL),
  DECIMALS = structure("DECIMALS", var_att_required =
                         VARATT_REQUIRE_LEVELS$OPTIONAL),
  DATA_ENTRY_TYPE = structure("DATA_ENTRY_TYPE", var_att_required =
                                VARATT_REQUIRE_LEVELS$OPTIONAL),
  END_DIGIT_CHECK = structure("END_DIGIT_CHECK", var_att_required =
                                VARATT_REQUIRE_LEVELS$OPTIONAL),
  CO_VARS = structure("CO_VARS", var_att_required =
                        VARATT_REQUIRE_LEVELS$RECOMMENDED),
  GROUP_VAR_OBSERVER = structure("GROUP_VAR_OBSERVER", var_att_required =
                             VARATT_REQUIRE_LEVELS$RECOMMENDED),
  GROUP_VAR_DEVICE = structure("GROUP_VAR_DEVICE", var_att_required =
                             VARATT_REQUIRE_LEVELS$RECOMMENDED),
  KEY_OBSERVER = structure("KEY_OBSERVER", var_att_required =
                             VARATT_REQUIRE_LEVELS$COMPATIBILITY),
  KEY_DEVICE = structure("KEY_DEVICE", var_att_required =
                           VARATT_REQUIRE_LEVELS$COMPATIBILITY),
  TIME_VAR = structure("TIME_VAR", var_att_required =
                             VARATT_REQUIRE_LEVELS$RECOMMENDED),
  TIME_VAR_END = structure("TIME_VAR_END", var_att_required =
                         VARATT_REQUIRE_LEVELS$OPTIONAL),
  KEY_DATETIME = structure("KEY_DATETIME", var_att_required =
                             VARATT_REQUIRE_LEVELS$COMPATIBILITY),
  PART_VAR = structure("PART_VAR", var_att_required =
                         VARATT_REQUIRE_LEVELS$RECOMMENDED),
  STUDY_SEGMENT = structure("STUDY_SEGMENT", var_att_required =
                              VARATT_REQUIRE_LEVELS$RECOMMENDED),
  KEY_STUDY_SEGMENT = structure("KEY_STUDY_SEGMENT", var_att_required =
                                  VARATT_REQUIRE_LEVELS$COMPATIBILITY),
  VARIABLE_ROLE = structure("VARIABLE_ROLE", var_att_required =
                              VARATT_REQUIRE_LEVELS$OPTIONAL),
  VARIABLE_ORDER = structure("VARIABLE_ORDER", var_att_required =
                               VARATT_REQUIRE_LEVELS$OPTIONAL),
  LONG_LABEL = structure("LONG_LABEL", var_att_required =
                           VARATT_REQUIRE_LEVELS$RECOMMENDED),
  SOFT_LIMIT_LOW = structure("SOFT_LIMIT_LOW", var_att_required =
                               VARATT_REQUIRE_LEVELS$TECHNICAL),
  SOFT_LIMIT_UP = structure("SOFT_LIMIT_UP", var_att_required =
                              VARATT_REQUIRE_LEVELS$TECHNICAL),
  HARD_LIMIT_LOW = structure("HARD_LIMIT_LOW", var_att_required =
                               VARATT_REQUIRE_LEVELS$TECHNICAL),
  HARD_LIMIT_UP = structure("HARD_LIMIT_UP", var_att_required =
                              VARATT_REQUIRE_LEVELS$TECHNICAL),
  DETECTION_LIMIT_LOW = structure("DETECTION_LIMIT_LOW", var_att_required =
                                    VARATT_REQUIRE_LEVELS$TECHNICAL),
  DETECTION_LIMIT_UP = structure("DETECTION_LIMIT_UP", var_att_required =
                                   VARATT_REQUIRE_LEVELS$TECHNICAL),
  INCL_SOFT_LIMIT_LOW = structure("INCL_SOFT_LIMIT_LOW", var_att_required =
                                    VARATT_REQUIRE_LEVELS$TECHNICAL),
  INCL_SOFT_LIMIT_UP = structure("INCL_SOFT_LIMIT_UP", var_att_required =
                                   VARATT_REQUIRE_LEVELS$TECHNICAL),
  INCL_HARD_LIMIT_LOW = structure("INCL_HARD_LIMIT_LOW", var_att_required =
                                    VARATT_REQUIRE_LEVELS$TECHNICAL),
  INCL_HARD_LIMIT_UP = structure("INCL_HARD_LIMIT_UP", var_att_required =
                                   VARATT_REQUIRE_LEVELS$TECHNICAL),
  LOCATION_RANGE = structure("LOCATION_RANGE", var_att_required =
                               VARATT_REQUIRE_LEVELS$RECOMMENDED),
  LOCATION_METRIC = structure("LOCATION_METRIC", var_att_required =
                                VARATT_REQUIRE_LEVELS$RECOMMENDED),
  PROPORTION_RANGE = structure("PROPORTION_RANGE", var_att_required =
                                 VARATT_REQUIRE_LEVELS$RECOMMENDED),
  LOCATION_LIMIT_LOW = structure("LOCATION_LIMIT_LOW", var_att_required =
                                   VARATT_REQUIRE_LEVELS$TECHNICAL),
  LOCATION_LIMIT_UP = structure("LOCATION_LIMIT_UP", var_att_required =
                                  VARATT_REQUIRE_LEVELS$TECHNICAL),
  INCL_LOCATION_LIMIT_LOW = structure("INCL_LOCATION_LIMIT_LOW", var_att_required =
                                        VARATT_REQUIRE_LEVELS$TECHNICAL),
  INCL_LOCATION_LIMIT_UP = structure("INCL_LOCATION_LIMIT_UP", var_att_required =
                                       VARATT_REQUIRE_LEVELS$TECHNICAL),
  PROPORTION_LIMIT_LOW = structure("PROPORTION_LIMIT_LOW", var_att_required =
                                     VARATT_REQUIRE_LEVELS$TECHNICAL),
  PROPORTION_LIMIT_UP = structure("PROPORTION_LIMIT_UP", var_att_required =
                                    VARATT_REQUIRE_LEVELS$TECHNICAL),
  INCL_PROPORTION_LIMIT_LOW = structure("INCL_PROPORTION_LIMIT_LOW", var_att_required =
                                          VARATT_REQUIRE_LEVELS$TECHNICAL),
  INCL_PROPORTION_LIMIT_UP = structure("INCL_PROPORTION_LIMIT_UP", var_att_required =
                                         VARATT_REQUIRE_LEVELS$TECHNICAL),
  #  VARSHORTLABEL = structure("varshortlabel", var_att_required =
  #                                     VARATT_REQUIRE_LEVELS$OPTIONAL),
  RECODE_CASES = structure("RECODE_CASES", var_att_required =
                       VARATT_REQUIRE_LEVELS$OPTIONAL),
  RECODE_CONTROL = structure("RECODE_CONTROL", var_att_required =
                             VARATT_REQUIRE_LEVELS$OPTIONAL),
  GRADING_RULESET = structure("GRADING_RULESET", var_att_required =
                                VARATT_REQUIRE_LEVELS$OPTIONAL),
  STANDARDIZED_VOCABULARY_TABLE = structure("STANDARDIZED_VOCABULARY_TABLE",
                                            var_att_required =
                                              VARATT_REQUIRE_LEVELS$OPTIONAL),
  DATAFRAMES = structure("DATAFRAMES",
                         var_att_required =
                           VARATT_REQUIRE_LEVELS$OPTIONAL),
  ENCODING = structure("ENCODING",
                         var_att_required =
                           VARATT_REQUIRE_LEVELS$OPTIONAL),
  UNIVARIATE_OUTLIER_CHECKTYPE = structure("UNIVARIATE_OUTLIER_CHECKTYPE",
                                           var_att_required =
                                             VARATT_REQUIRE_LEVELS$RECOMMENDED),
  N_RULES = structure("N_RULES",
                       var_att_required =
                         VARATT_REQUIRE_LEVELS$RECOMMENDED),
  EXTENDED_DATA_TYPE = structure("EXTENDED_DATA_TYPE", var_att_required =
                                   VARATT_REQUIRE_LEVELS$OPTIONAL),
  COMPUTED_VARIABLE_ROLE = structure("COMPUTED_VARIABLE_ROLE",
                                     var_att_required =
                                       VARATT_REQUIRE_LEVELS$TECHNICAL
                                     )
)

.onAttach <- function(...) { # nocov start
  # if (packageVersion("ggplot2") >= as.package_version("4.0.0")) {
  #   packageStartupMessage(paste(
  #     "ggplot2 >= 4.0.0 uses S7 classes, which can cause performance issues.",
  #     "We are currently trying to address these, in doubt, try to downgrade",
  #     "ggplot2, if you consider trouble after updating ggplot2"))
  # }

  if (length(user_hints$l) > 0) {
    packageStartupMessage(paste(user_hints$l, collapse = "\n"))
  }
}
# nocov end

.onLoad <- function(...) { # nocov start

  if ((suppressWarnings(util_ensure_suggested("jsonlite", err = FALSE)))) {
    methods::setClass("dataquieR_translated")
    methods::setMethod(methods::getGeneric("asJSON", package = "jsonlite"),
              "dataquieR_translated",
              asJSON.dataquieR_translated)
  }


  delayedAssign("odm_installed",
         suppressWarnings(
           util_ensure_suggested("dataquieR2odm",
                                 goal = "Read ODM files",
                                 err = FALSE)),
         eval.env = asNamespace("dataquieR"),
         asNamespace("dataquieR"))

  to_set <-
    grep("^dataquieR.*_default$", ls(asNamespace("dataquieR")), value = TRUE)

  for (o in to_set) {
    nm <- sub("_default$", "", o)
    if (is.null(getOption(nm))) {
      do.call("options", setNames(list(get(o, asNamespace("dataquieR"))),
                                  nm = nm))
    }
  }

  ## cowplot integration (only if installed)
  if ((suppressWarnings(util_ensure_suggested("cowplot", err = FALSE)))) {
    util_s3_register("cowplot::as_grob", "dq_lazy_ggplot_s7", as_grob.dq_lazy_ggplot_s7)
    util_s3_register("cowplot::as_grob", "dq_lazy_ggplot",    as_grob.dq_lazy_ggplot)
  }

  ## plotly integration (only if installed)

  if ((suppressWarnings(util_ensure_suggested("plotly", err = FALSE)))) {
    util_s3_register("plotly::ggplotly",     "dq_lazy_ggplot", ggplotly.dq_lazy_ggplot)
    util_s3_register("plotly::plotly_build", "dq_lazy_ggplot", plotly_build.dq_lazy_ggplot)
    util_s3_register("plotly::ggplotly",     "dq_lazy_ggplot_s7", ggplotly.dq_lazy_ggplot_s7)
    util_s3_register("plotly::plotly_build", "dq_lazy_ggplot_s7", plotly_build.dq_lazy_ggplot_s7)
    util_s3_register("plotly::to_basic", "GeomPointrangeRobust", to_basic.GeomPointrangeRobust)
  }

  if (packageVersion("ggplot2") >= as.package_version("3.5.2")) {
    f <- get("is_ggplot",
             envir = environment(ggplot2::ggplot))
  } else {
    f <- get("is.ggplot",
             envir = environment(ggplot2::ggplot))

  }
  assign("util_is_gg_plot",
         f,
         environment(util_is_gg_plot))

  if (getRversion() >= "2.15.1") {
    utils::globalVariables(
      c(
        names(WELL_KNOWN_META_VARIABLE_NAMES),
        "....alt_call_res",
        "colcode",
        "colscale",
        "continuous",
        "level_names",
        "APP_SCORE",
        "FITTED_VALUE",
        "GRADING",
        "IMPLEMENTATION",
        "INTERVALS",
        "LCL",
        "LOWER_CL",
        "PROB",
        "TIME",
        "UCL",
        "UPPER_CL",
        "VARIABLES",
        "PCT_con_con",
        "Variables",
        "category",
        "lwl",
        "margins",
        "percent",
        "upl",
        "x",
        "y",
        "z2",
        "ds1",
        "meta_data",
        "variable",
        "value",
        "Rules",
        "progress",
        "progress_msg",
        "REQUIRED", # a var att requirement level, using NSE
        "label_col", # generated by prep_prepare_dataframes like ds1
        "tr", # htmltools withTags
        "td", # htmltools withTags
        "th", # htmltools withTags
        "table", # htmltools withTags
        "." # dplyr
      )
    )
  }

  if (requireNamespace("pkgload", quietly = TRUE)) {
    is_dev_package <- pkgload::is_dev_package
  } else {
    is_dev_package <- function(...) { return(FALSE) }
  }

  for (name in names(WELL_KNOWN_META_VARIABLE_NAMES)) {
    if (exists(name, asNamespace("dataquieR")) &&
        !is_dev_package("dataquieR")) {
      util_warning("Variable %s is in dataquieR too!", name,
                   applicability_problem = FALSE)
    }
    assign(name, WELL_KNOWN_META_VARIABLE_NAMES[[name]],
           asNamespace("dataquieR"))
  }
  if (!is_dev_package("dataquieR")) {
    namespaceExport(asNamespace("dataquieR"),
                    names(WELL_KNOWN_META_VARIABLE_NAMES))
  }

  if (file.exists(system.file("ssi.rds", package = "dataquieR"))) {
    local({
      .ssi <- readRDS(system.file("ssi.rds", package = "dataquieR"))
      COMPUTED_VARIABLE_ROLES <-
        as.list(setNames(nm = .ssi$SSI_METRICS))
      COMPUTED_VARIABLE_ROLES$`NA` <- "na"
      assign("COMPUTED_VARIABLE_ROLES", COMPUTED_VARIABLE_ROLES,
             asNamespace("dataquieR"))
      for (name in names(COMPUTED_VARIABLE_ROLES)) {
        if (exists(name, asNamespace("dataquieR")) &&
            !is_dev_package("dataquieR")) {
          util_warning("Variable %s is in dataquieR too!", name,
                       applicability_problem = FALSE)
        }
        assign(name, COMPUTED_VARIABLE_ROLES[[name]],
               asNamespace("dataquieR"))
      }
      if (!is_dev_package("dataquieR")) {
        namespaceExport(asNamespace("dataquieR"),
                        "COMPUTED_VARIABLE_ROLES")
        namespaceExport(asNamespace("dataquieR"),
                        setdiff(names(COMPUTED_VARIABLE_ROLES), "NA"))
      }
    })
  }

  if (is_dev_package("dataquieR") &&
      exists("util_load_helpers") &&
      is.function(get("util_load_helpers"))) {
    get("util_load_helpers")()
  }

  makeActiveBinding(".manual", function() {
    if (!length(ls(..manual))) {
      util_load_manual()
    }
    if (!length(..manual$titles)) {
      wrnPat <- paste("Did not find the reference manual for %s.",
                      "This can cause all-white summmaries")
      if (suppressWarnings(util_ensure_suggested("cli", err = FALSE))) {
        wrnPat <- cli::cli_alert_warning(
          cli::bg_black(cli::col_yellow(wrnPat)))
      }
      rlang::warn(sprintf(
        wrnPat,
        dQuote(rlang::ns_env_name())),
        .frequency = "regularly", .frequency_id = "dataquieRManual")
    }
    return(..manual)
  }, environment(util_load_manual))

  makeActiveBinding(".indicator_or_descriptor", function() {
    if (!length(ls(..indicator_or_descriptor))) {
      util_load_manual()
    }
    return(..indicator_or_descriptor)
  }, environment(util_load_manual))

  makeActiveBinding(".called_in_pipeline", function() {
    .dq2_globs$.called_in_pipeline
  }, environment(.onLoad))

  makeActiveBinding("MAX_LONG_LABEL_LEN", function() {
    min(.MAX_LABEL_LEN, getOption("dataquieR.MAX_LONG_LABEL_LEN",
                                  dataquieR.MAX_LONG_LABEL_LEN_default),
        na.rm = TRUE)
  }, env = environment(.onLoad))

  makeActiveBinding("MAX_LABEL_LEN", function() {
    min(.MAX_LABEL_LEN, getOption("dataquieR.MAX_LABEL_LEN",
                                  dataquieR.MAX_LABEL_LEN_default),
        na.rm = TRUE)
  }, env = environment(.onLoad))

  util_fix_rstudio_bugs()

  if (!l10n_info()[["UTF-8"]]) {
    update_recommended <-
      (.Platform$OS.type == "windows" &&
       (R.version$major < 4 || (R.version$major >= 4 && R.version$minor < 2)))
    util_user_hint(sprintf(
      paste("Your R session is not using a UTF-8 character set,",
      "be prepared to encoding problems. Also use %s. %s"),
      "Sys.setlocale() to select a UTF-8 locale in your .Rprofile",
      ifelse(update_recommended,
        paste(
          "\nOn Windows, you should consider updating R to a version >= 4.2.0,",
            "which supports UTF-8."),
          "")
    ))
  }

  unit_env <- new.env(parent = emptyenv())

  makeActiveBinding("UNITS", function() {
    if (!exists("valud", envir = unit_env)) {
      if (!requireNamespace("xml2", quietly = TRUE)) { # for units
        return(character(0))
      }
      valud <- suppressMessages(units::valid_udunits())
      assign("valud", valud, envir = unit_env)
    }
    valud <- get("valud", envir = unit_env)
    u <- valud[["symbol"]]
    u <- u[!util_empty(u)]
    u
  }, env = asNamespace(packageName()))

  makeActiveBinding("UNIT_IS_COUNT", function() {
    if (!exists("valud", envir = unit_env)) {
      if (!requireNamespace("xml2", quietly = TRUE)) { # for units
        return(character(0))
      }
      valud <- suppressMessages(units::valid_udunits())
      assign("valud", valud, envir = unit_env)
    }
    valud <- get("valud", envir = unit_env)
    valud <- valud[!util_empty(valud[["symbol"]]), , FALSE]
    u <- valud[is.na(valud$def) ==
             is.na(suppressWarnings(as.numeric(valud$def))), "symbol", TRUE]
    u <- u[!util_empty(u)]
    u <- setNames(nm = u)
    attr(u, "def") <-
      suppressWarnings(setNames(nm = valud$symbol, as.numeric(valud$def))[u])
    u
  }, env = asNamespace(packageName()))

  makeActiveBinding("UNIT_SOURCES", function() {
    if (!exists("valud", envir = unit_env)) {
      if (!requireNamespace("xml2", quietly = TRUE)) { # for units
        return(character(0))
      }
      valud <- suppressMessages(units::valid_udunits())
      assign("valud", valud, envir = unit_env)
    }
    valud <- get("valud", envir = unit_env)
    valud <- valud[!util_empty(valud[["symbol"]]), , FALSE]
    u <- setNames(nm = valud[["symbol"]], valud[["source_xml"]])
    u
  }, env = asNamespace(packageName()))

  makeActiveBinding("UNIT_PREFIXES", function() {
    if (!exists("p", envir = unit_env)) {
      if (!requireNamespace("xml2", quietly = TRUE)) { # for units
        return(character(0))
      }
      p <- suppressMessages(unique(
        unname(unlist(units::valid_udunits_prefixes()[, 1:3]))))
      p <- p[!util_empty(p)]
      assign("p", trimws(unlist(strsplit(p, ",", fixed = TRUE))), envir =
               unit_env)
    }
    get("p", envir = unit_env)
  }, env = asNamespace(packageName()))

  makeActiveBinding("UNIT_PREFIX_FACTORS", function() {
    if (!exists("pf", envir = unit_env)) {
      if (!requireNamespace("xml2", quietly = TRUE)) { # for units
        return(character(0))
      }
      ..pf <- suppressMessages(units::valid_udunits_prefixes())
      pf <- suppressMessages(unique(
        unname(unlist(..pf[, 1:3]))))
      pf <- pf[!util_empty(pf)]
      pf <- vapply(trimws(unlist(strsplit(pf, ",", fixed = TRUE))),
                          function(pfx) {
                            r <- ..pf[..pf[[1]] == pfx |
                                   ..pf[[2]] == pfx |
                                   ..pf[[3]] == pfx, "value", drop = TRUE]
                            if (length(r) == 0) {
                              r <- 0
                            }
                            r
                          },
                          FUN.VALUE = numeric(1))
      pf <- pf[pf != 0]
      dps <- duplicated(names(pf))
      if (any(dps)) {
        util_warning(
          c("Removing duplicated unit prefixes %s from udunits. Please",
            "verify, if your installation of the package units returns valid",
            "prefixes."),
          util_pretty_vector_string(sort(unique(names(pf)[dps])))
        )
        pf <- pf[!dps]
      }
      assign("pf",
             pf,
             envir =
               unit_env)
    }
    get("pf", envir = unit_env)
  }, env = asNamespace(packageName()))

  namespaceExport(asNamespace("dataquieR"),
                  c("UNITS", "UNIT_PREFIXES", "UNIT_SOURCES", "UNIT_IS_COUNT"))

  suppressMessages({
    # methods::setMethod('vapply', methods::signature(X="dataquieR_resultset2"),
    #                    function(X, FUN, FUN.VALUE, ..., USE.NAMES = TRUE) {
    #                      vapply(setNames(nm = names(X)), function(nm) {
    #                        FUN(X[[nm]], ...)
    #                      }, FUN.VALUE = FUN.VALUE, USE.NAMES = USE.NAMES)
    #                    })
    #
    # methods::setMethod('lapply', methods::signature(X="dataquieR_resultset2"),
    #                    function(X, FUN, ...) {
    #                      lapply(setNames(nm = names(X)), function(nm) {
    #                        FUN(X[[nm]], ...)
    #                      })
    #                    })
    #
    # methods::setMethod('sapply', methods::signature(X="dataquieR_resultset2"),
    #                    function(X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE) {
    #                      sapply(setNames(nm = names(X)), function(nm) {
    #                        FUN(X[[nm]], ...)
    #                      }, simplify = simplify, USE.NAMES = USE.NAMES)
    #                    })

    # methods::setMethod('vapply', methods::signature(X="dataquieR_resultset2"),
    #                    function(X, FUN, FUN.VALUE, ..., USE.NAMES = TRUE) {
    #                      vapply(setNames(nm = names(X)), function(nm) {
    #                        FUN(X[[nm]], ...)
    #                      }, FUN.VALUE = FUN.VALUE, USE.NAMES = USE.NAMES)
    #                    }, where = globalenv())
    #
    # methods::setMethod('lapply', methods::signature(X="dataquieR_resultset2"),
    #                    function(X, FUN, ...) {
    #                      lapply(setNames(nm = names(X)), function(nm) {
    #                        FUN(X[[nm]], ...)
    #                      })
    #                    }, where = globalenv())
    #
    # methods::setMethod('sapply', methods::signature(X="dataquieR_resultset2"),
    #                    function(X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE) {
    #                      sapply(setNames(nm = names(X)), function(nm) {
    #                        FUN(X[[nm]], ...)
    #                      }, simplify = simplify, USE.NAMES = USE.NAMES)
    #                    }, where = globalenv())


    if (!inherits(try(as.environment("tools:rstudio"),
                      silent = TRUE), "try-error") &&
        exists(".rs.getNames", as.environment("tools:rstudio"))) {
      # no need to load the full object just to fetch the names.
      .dataquieR_orig_util_rs_get_names <- # FIXME: don't, if already done
        get(".rs.getNames", as.environment("tools:rstudio"))
      if (!is.null(attr(.dataquieR_orig_util_rs_get_names,
                        ".dataquieR_orig_util_rs_get_names"))) {
        .dataquieR_orig_util_rs_get_names <-
          attr(.dataquieR_orig_util_rs_get_names,
               ".dataquieR_orig_util_rs_get_names")
      }
      util_rs_get_names <- function(object) {
        if (inherits(object, "dataquieR_resultset2")) {
          r <- names(object)
          if (!is.null(r))
            attr(r, "meta") <- rep("", length(r))
          return(r)
        } else {
          eval.parent(body(.dataquieR_orig_util_rs_get_names))
        }
      }
      environment(util_rs_get_names) <-
        environment(.dataquieR_orig_util_rs_get_names)
      attr(.dataquieR_orig_util_rs_get_names,
           ".dataquieR_orig_util_rs_get_names") <-
        .dataquieR_orig_util_rs_get_names
      attr(util_rs_get_names, ".dataquieR_orig_util_rs_get_names") <-
        .dataquieR_orig_util_rs_get_names
      assign(".dataquieR_orig_util_rs_get_names",
             .dataquieR_orig_util_rs_get_names,
             environment(util_rs_get_names))
      assign(".rs.getNames", util_rs_get_names, as.environment("tools:rstudio"))
    }
  })

  assign("MEM_COMPRESS_CAPABILITIES",
         names(which(.util_mem_compress_capabilities())),
         environment(.util_mem_compress_capabilities))

  if (requireNamespace("S7", quietly = TRUE)) {
    # register all S7 methods for the package (recommended by S7)
    S7::methods_register()

    # ensure your dq_lazy S7 class + operator methods exist in THIS session
    suppressMessages(dq_lazy_register_s7())
  }


}
# nocov end

MEM_COMPRESS_CAPABILITIES <- "none"

# name of the additional system missingness column in com_item_missingness
.SM_LAB <- "NAs (System missings)"

#' Data frame with labels for missing- and jump-codes
#' #' Metadata about value and missing codes
#'
#'

#' @name value/missing-lists
#' @aliases cause_label_df missing_matchtable CODE_VALUE CODE_LABEL CODE_CLASS CODE_INTERPRET value_label_table
#' @description
#' [data.frame] with the following columns:
#'   - `CODE_VALUE`: [numeric] | [DATETIME] Missing or categorical code
#'                                (the number or date representing a
#'                                missing/category)
#'   - `CODE_LABEL`: [character] a label for the missing code or category
#'   - `CODE_CLASS`: [enum] JUMP | MISSING. For missing lists: Class of the
#'                             missing code.
#'   - `CODE_INTERPRET` [enum] I | P | PL | R | BO | NC | O | UH | UO | NE.
#'                             For missing lists: Class of the missing code
#'                             according to
#'       [`AAPOR`](https://aapor.org/standards-and-ethics/standard-definitions/).
#'   - `resp_vars`: [character] For missing lists: optional, if a missing code
#'                              is specific for some
#'                              variables, it is listed for each such variable
#'                              with one entry in `resp_vars`, If `NA`, the
#'                              code is assumed shared among all variables.
#'                              For v1.0 metadata, you need to refer to
#'                              `VAR_NAMES` here.
#'
#' @seealso [Online](https://dataquality.qihs.uni-greifswald.de/VIN_Item_Level_Metadata.html#MISSING_LIST_TABLE)
#' @seealso [com_item_missingness()]
#' @seealso [com_segment_missingness()]
#' @seealso [com_qualified_item_missingness()]
#' @seealso [com_qualified_segment_missingness()]
#' @seealso [con_inadmissible_categorical()]
#' @seealso [con_inadmissible_vocabulary()]
#' @seealso [MISSING_LIST_TABLE]
#' @seealso [VALUE_LABEL_TABLE]
#' @seealso [STANDARDIZED_VOCABULARY_TABLE]
#' @seealso [cause_label_df]
NULL

#' Data frame with contradiction rules
#' @name check_table
#' @seealso [meta_data_cross]
#' @description
#' Two versions exist, the newer one is used by [con_contradictions_redcap] and
#' is described [here](https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions_redcap.html#Example_output).,
#' the older one used by [con_contradictions] is described
#' [here](https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions.html#Example_output).
NULL

#' Data frame with metadata about the study data on variable level
#'
#' @name meta_data
#' @description
#'
#' Variable level metadata.
#'
#' @seealso [further details on variable level metadata.](https://dataquality.qihs.uni-greifswald.de/VIN_Annotation_of_Metadata.html)
#' @seealso [meta_data_segment]
#' @seealso [meta_data_dataframe]
#'
NULL

#' Data frame with the study data whose quality is being assessed
#' @name study_data
#' @description
#' Study data is expected in wide format. If should contain all variables
#' for all segments in one large table, even, if some variables are not
#' measured for all observational utils (study participants).
NULL

#' Well known columns on the `meta_data_segment` sheet
#' @name meta_data_segment
#' @description
#' Metadata describing study segments, e.g., a full questionnaire, not its
#' items.
NULL

#' Default Name of the Table featuring Code Lists
#'
#' @export
CODE_LIST_TABLE = "CODE_LIST_TABLE"

# this has already been defined
# @inherit meta_data_segment
# @export
# STUDY_SEGMENT <- "STUDY_SEGMENT"

#' Segment level metadata attribute name
#'
#' Number of expected data records in each segment.
#' [numeric]. Check only conducted if number entered
#'
#' @seealso [meta_data_segment]
#'
#' @export
SEGMENT_RECORD_COUNT <- "SEGMENT_RECORD_COUNT"

#' Segment level metadata attribute name
#'
#' The name of the data frame containing the reference IDs to be compared
#' with the IDs in the targeted segment.
#'
#' @seealso [meta_data_segment]
#'
#' @export
SEGMENT_ID_REF_TABLE <- "SEGMENT_ID_REF_TABLE"

#' Deprecated segment level metadata attribute name
#'
#' The name of the data frame containing the reference IDs to be compared
#' with the IDs in the targeted segment.
#'
#' Please use [SEGMENT_ID_REF_TABLE]
#'
#' @export
SEGMENT_ID_TABLE <- SEGMENT_ID_REF_TABLE

#' Segment level metadata attribute name
#'
#' The type of check to be conducted when comparing the reference ID
#' table with the IDs in a segment.
#'
#' @seealso [meta_data_segment]
#'
#' @export
SEGMENT_RECORD_CHECK <- "SEGMENT_RECORD_CHECK"

#' Segment level metadata attribute name
#'
#' All variables that are to be used as one single ID
#' variable (combined key) in a segment.
#'
#' @seealso [meta_data_segment]
#'
#' @export
SEGMENT_ID_VARS <- "SEGMENT_ID_VARS"

#' Segment level metadata attribute name
#'
#' @seealso [DF_UNIQUE_ID]
#'
#' @seealso [meta_data_segment]
#'
#' @export
SEGMENT_UNIQUE_ID <- "SEGMENT_UNIQUE_ID"

#' Segment level metadata attribute name
#'
#' Specifies whether identical data is permitted across rows in a
#' segment (excluding ID variables)
#'
#' @seealso [meta_data_segment]
#'
#' @export
SEGMENT_UNIQUE_ROWS <- "SEGMENT_UNIQUE_ROWS"

#' Segment level metadata attribute name
#'
#' The name of the segment participation status variable
#'
#' @seealso [meta_data_segment]
#'
#' @export
SEGMENT_PART_VARS <- "SEGMENT_PART_VARS"

#' Segment level metadata attribute name
#'
#' `true` or `false` to suppress crude segment missingness output
#' (`Completeness/Misg. Segments` in the report). Defaults to compute
#' the output, if more than one segment is available in the item-level
#' metadata.
#'
#' @seealso [meta_data_segment]
#'
#' @export
SEGMENT_MISS <- "SEGMENT_MISS"

#' Well known columns on the `meta_data_dataframe` sheet
#' @name meta_data_dataframe
#' @description
#' Metadata describing data delivered on one data frame/table sheet,
#' e.g., a full questionnaire, not its items.
NULL

#' Data frame level metadata attribute name
#'
#' Name of the data frame
#'
#' @seealso [meta_data_dataframe]
#'
#' @export
DF_NAME <- "DF_NAME"

#' Data frame level metadata attribute name
#'
#' Name of the data frame
#'
#' @seealso [meta_data_dataframe]
#'
#' @export
DF_CODE <- "DF_CODE"

#' Data frame level metadata attribute name
#'
#' Number of expected data elements in a data frame.
#' [numeric]. Check only conducted if number entered
#'
#' @seealso [meta_data_dataframe]
#'
#' @export
DF_ELEMENT_COUNT <- "DF_ELEMENT_COUNT"

#' Data frame level metadata attribute name
#'
#' Number of expected data records in a data frame.
#' [numeric]. Check only conducted if number entered
#'
#' @seealso [meta_data_dataframe]
#'
#' @export
DF_RECORD_COUNT <- "DF_RECORD_COUNT"

#' Data frame level metadata attribute name
#'
#' The name of the data frame containing the reference IDs to be
#' compared with the IDs in the study data set.
#'
#' @seealso [meta_data_dataframe]
#'
#' @export
DF_ID_REF_TABLE <- "DF_ID_REF_TABLE"

#' Data frame level metadata attribute name
#'
#' The type of check to be conducted when comparing the reference
#' ID table with the IDs delivered in the study data files.
#'
#' @seealso [meta_data_dataframe]
#'
#' @export
DF_RECORD_CHECK <- "DF_RECORD_CHECK"

#' Data frame level metadata attribute name
#'
#' Defines expectancies on the uniqueness of the IDs
#' across the rows of a data frame, or the number of times some ID can be repeated.
#'
#' @seealso [meta_data_dataframe]
#'
#' @export
DF_UNIQUE_ID <- "DF_UNIQUE_ID"

#' Data frame level metadata attribute name
#'
#' All variables that are to be used as one single ID
#' variable (combined key) in a data frame.
#'
#' @seealso [meta_data_dataframe]
#'
#' @export
DF_ID_VARS <- "DF_ID_VARS"

#' Data frame level metadata attribute name
#'
#' Specifies whether identical data is permitted across rows in a
#' data frame (excluding ID variables)
#'
#' @seealso [meta_data_dataframe]
#'
#' @export
DF_UNIQUE_ROWS <- "DF_UNIQUE_ROWS"

#' Well known columns on the `item_computation_level` sheet
#' @name meta_data_computation
#' @family meta_data_cross
#' @description
#' Computation rules
#' TODO
#'
NULL

#' Well known columns on the `cross-item_level` sheet
#' @name meta_data_cross
#' @seealso [check_table]
#' @seealso [Online Documentation](https://dataquality.qihs.uni-greifswald.de/VIN_Cross_Item_Level_Metadata.html)
#' @family meta_data_cross
#' @description
#' Metadata describing groups of variables, e.g., for their multivariate
#' distribution or for defining contradiction rules.
#'
NULL

#' Valid unit symbols according to [units::valid_udunits()]
#'
#' like m, g, N, ...
#'
#' @name UNITS
#' @family UNITS
NULL

#' Is a unit a count according to [units::valid_udunits()]
#'
#' see column `def`, therein
#'
#' like `%`, `ppt`, `ppm`
#'
#' @name UNIT_IS_COUNT
#' @family UNITS
NULL

#' Maturity stage of a unit according to [units::valid_udunits()]
#'
#' see column `source_xml` therein, i.e., base, derived, accepted, or common
#'
#' @name UNIT_SOURCES
#' @family UNITS
NULL

#' Factors related to unit prefixes [units::valid_udunits_prefixes()]
#'
#' named `numeric` vector
#'
#' translates k, m, M, c, ... to 1000, 0.001, ...
#'
#' @name UNIT_PREFIX_FACTORS
#' @family UNITS
NULL

#' Valid unit prefixes according to [units::valid_udunits_prefixes()]
#'
#' like k, m, M, c, ...
#'
#' @name UNIT_PREFIXES
#' @family UNITS
NULL

#' Cross-item level metadata attribute name
#'
#' Specifies the unique labels for cross-item level metadata records
#'
#' if missing, `dataquieR` will create such labels
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
CHECK_LABEL <- "CHECK_LABEL"

#' Cross-item level metadata attribute name
#' TODO
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @export
SCALE_NAME <- "SCALE_NAME"

#' Cross-item level metadata attribute name
#' TODO
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @export
SCALE_ACRONYM <- "SCALE_ACRONYM"

#' Cross-item level metadata attribute name
#' TODO
#' internal use, only
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @export
VARIABLE_LIST_ORDER <- "VARIABLE_LIST_ORDER"

#' Cross-item level metadata attribute name
#'
#' Specifies the unique IDs for cross-item level metadata records
#'
#' if missing, `dataquieR` will create such IDs
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
CHECK_ID <- "CHECK_ID"

#' Cross-item level metadata attribute name
#'
#' Specifies a group of variables for multivariate analyses. Separated
#' by `r SPLIT_CHAR`, please use variable names from [VAR_NAMES] or
#' a label as specified in `label_col`, usually [LABEL] or [LONG_LABEL].
#'
#' if missing, `dataquieR` will create such IDs from [CONTRADICTION_TERM],
#' if specified.
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
VARIABLE_LIST <- "VARIABLE_LIST"

#' Cross-item level metadata attribute name
#'
#' Note: in some `prep_`-functions, this field is named `RULE`
#'
#' Specifies a contradiction rule. Use `REDCap` like syntax, see
#' [online vignette](https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions_redcap.html)
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#' @aliases RULE
#'
#' @export
CONTRADICTION_TERM <- "CONTRADICTION_TERM"

#' Cross-item level metadata attribute name
#'
#' Specifies the type of a contradiction. According to the data quality
#' concept, there are logical and empirical contradictions, see
#' [online vignette](https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions_redcap.html)
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
CONTRADICTION_TYPE <- "CONTRADICTION_TYPE"

#' Cross-item level metadata attribute name
#'
#' Select, which outlier criteria to compute, see [acc_multivariate_outlier].
#'
#' You can leave the cell empty, then, all checks will apply. If you enter
#' a set of methods, the maximum for [N_RULES] changes. See also
#' [`UNIVARIATE_OUTLIER_CHECKTYPE`].
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
MULTIVARIATE_OUTLIER_CHECKTYPE <- "MULTIVARIATE_OUTLIER_CHECKTYPE"

#' Cross-item level metadata attribute name
#'
#' @seealso [meta_data_computation]
#' @family meta_data_computation
#'
#' @export
COMPUTATION_RULE = "COMPUTATION_RULE"

#' Cross-item level metadata attribute name
#'
#' Select, whether to compute [acc_multivariate_outlier].
#'
#' You can leave the cell empty, then the depends on the setting of the
#' `option` [dataquieR.MULTIVARIATE_OUTLIER_CHECK]. If this column is missing,
#' all this is the same as having all cells empty and
#' `dataquieR.MULTIVARIATE_OUTLIER_CHECK` set to `"auto"`.
#'
#' See also [`MULTIVARIATE_OUTLIER_CHECKTYPE`].
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
MULTIVARIATE_OUTLIER_CHECK <- "MULTIVARIATE_OUTLIER_CHECK"


#' Cross-item level metadata attribute name
#'
#' Select, whether to compute [acc_mahalanobis].
#'
#' You can leave the cell empty, then the depends on the setting of the
#' `option` [dataquieR.MULTIVARIATE_OUTLIER_CHECK]. If this column is missing,
#' all this is the same as having all cells empty and
#' `dataquieR.MULTIVARIATE_OUTLIER_CHECK` set to `"auto"`.
#'
#' See also [`MULTIVARIATE_OUTLIER_CHECKTYPE`].
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
MAHALANOBIS_THRESHOLD <- "MAHALANOBIS_THRESHOLD"

#' Cross-item level metadata attribute name
#'
#' Specifies the allowable range of an association. The inclusion of the
#' endpoints follows standard mathematical notation using round brackets
#' for open intervals and square brackets for closed intervals.
#' Values must be separated by a semicolon.
#'
#' @family meta_data_cross
#' @seealso [meta_data_cross]
#'
#' @export
ASSOCIATION_RANGE <- "ASSOCIATION_RANGE"

#' Cross-item level metadata attribute name
#'
#' The metric underlying the association in [ASSOCIATION_RANGE]. The input is
#' a string that specifies the analysis algorithm to be used.
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
ASSOCIATION_METRIC <- "ASSOCIATION_METRIC"

#' Cross-item level metadata attribute name
#'
#' The allowable direction of an association. The input is a string that can be
#' either "positive" or "negative".
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
ASSOCIATION_DIRECTION <- "ASSOCIATION_DIRECTION"

#' Cross-item level metadata attribute name
#'
#' The allowable form of association. The string specifies the form based on a
#' selected list.
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
ASSOCIATION_FORM <- "ASSOCIATION_FORM"

#' Cross-item level metadata attribute name
#'
#' Specifies the type of reliability or validity analysis. The string specifies
#' the analysis algorithm to be used, and can be either "inter-class" or
#' "intra-class".
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
REL_VAL <- "REL_VAL"

#' Cross-item level metadata attribute name
#'
#' Defines the measurement variable to be used as a known gold standard. Only
#' one variable can be defined as the gold standard.
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
GOLDSTANDARD <- "GOLDSTANDARD"

#' Cross-item level metadata attribute name
#'
#' For contradiction rules, the required pre-processing steps that can be given.
#' Note: `MISSING_LABEL`, `MISSING_INTERPRET` may not work for non-factor
#' variables
#'
#' LABEL LIMITS MISSING_NA MISSING_LABEL MISSING_INTERPRET
#'
#' @seealso [meta_data_cross]
#' @family meta_data_cross
#'
#' @export
DATA_PREPARATION <- "DATA_PREPARATION"

#' @export
CODE_VALUE <- "CODE_VALUE"

#' @export
RULE <- "RULE"

#' @export
CODE_LABEL <- "CODE_LABEL"

#' @export
CODE_INTERPRET <- "CODE_INTERPRET"

#' @export
CODE_CLASS <- "CODE_CLASS"

# TODO Only existence of CODE_ORDER is checked, order not yet used
#' Only existence is checked, order not yet used
#' @export
CODE_ORDER <- "CODE_ORDER"

#' types of value codes
#' @export
CODE_CLASSES <- list(MISSING = "MISSING",
                  JUMP = "JUMP",
                  VALUE = "VALUE")

#' Name of the sheet with rules to introduce missing codes in the pipeline
#' @export
MISSING_CODE_RULES = "MISSING_CODE_RULES"

#' Dimension Titles for Prefixes
#'
#' order does matter, because it defines the order in the `dq_report2`.
#'
#' @seealso `util_html_for_var()`
#' @seealso `util_html_for_dims()`
dims <- c(
  des = "Descriptive statistics",
  int = "Integrity",
  com = "Completeness",
  con = "Consistency",
  acc = "Accuracy"
)

# Use cli, if available.
rlang::local_use_cli()

#' Metadata sheet name containing VALUE_LABEL_TABLES
#' This metadata sheet can contain both value labels of several
#' VALUE_LABEL_TABLE and also Missing and JUMP tables
#' @export
CODE_LIST_TABLE <- "CODE_LIST_TABLE"

# for global options for dataquieR not exposed to the user like options()
dataquieR.properties <- new.env(parent = emptyenv())

.set_properties <- function(p) {
  list2env(p, dataquieR.properties)
}

with_pipeline <- withr::with_(new = FALSE,
  function(x) {
    res <- force(.dq2_globs$.called_in_pipeline)
    if (missing(x)) {
      .dq2_globs$.called_in_pipeline <- TRUE
    } else {
      .dq2_globs$.called_in_pipeline <- x
    }
    res
  })

without_pipeline <- withr::with_(new = FALSE,
  function(x) {
    res <- force(.dq2_globs$.called_in_pipeline)
    if (missing(x)) {
      .dq2_globs$.called_in_pipeline <- FALSE
    } else {
      .dq2_globs$.called_in_pipeline <- x
    }
    res
  })

.called_in_pipeline2 <- function() {
  n <- 1L
  repeat {
    cl <- rlang::caller_call(n)
    if (is.null(cl)) return(FALSE)  # End of call stack reached

    if (identical(rlang::call_name(cl), "dq_report2")) {
      # Try to resolve the actual function and its environment
      fn <- rlang::caller_fn(n)
      if (is.null(fn)) return(TRUE)  # If we can't resolve it, assume match
      # is good

      fn_env_top <- base::topenv(rlang::fn_env(fn))
      # Accept both Namespace (normal case) and Package environment
      # (e.g. devtools::load_all)
      is_from_pkg <- identical(fn_env_top, rlang::ns_env("dataquieR")) ||
        identical(rlang::env_name(fn_env_top), "package:dataquieR")
      return(is_from_pkg)
    }

    n <- n + 1L  # Check next caller up the stack
  }
}

dq_lazy_register_s7 <- function() {
  if (.dq_lazy_state$s7_ready) return(invisible(TRUE))
  if (!requireNamespace("S7", quietly = TRUE)) return(invisible(FALSE))

  # Create an S7 wrapper class around your existing S3 lazy object
  cls <- S7::new_class(
    "dq_lazy_ggplot_s7",
    properties = list(payload = S7::class_any)
  )
  .dq_lazy_state$s7_class <- cls

  # Define S7 operator methods for the wrapper.
  # NOTE: S7’s method-registration API is `S7::method()` in recent versions.
  # We guard it so you get a clear error if the API name differs.
  if (!exists("method", envir = asNamespace("S7"), inherits = FALSE)) {
    stop("S7 is installed but S7::method() was not found; adjust registration for your S7 version.")
  }

  base_ops <- get("base_ops", envir = asNamespace("S7"))

  S7::method(base_ops[["|"]], signature = list(cls, S7::class_any)) <- function(e1, e2) {
    p1 <- prep_realize_ggplot(dq_lazy_unwrap(e1))
    e2 <- dq_lazy_unwrap(e2)
    if (inherits(e2, "dq_lazy_ggplot")) e2 <- prep_realize_ggplot(e2)
    f <- get("|.ggplot", envir = asNamespace("patchwork"))
    f(p1, e2)
  }

  S7::method(base_ops[["/"]], signature = list(cls, S7::class_any)) <- function(e1, e2) {
    p1 <- prep_realize_ggplot(dq_lazy_unwrap(e1))
    e2 <- dq_lazy_unwrap(e2)
    if (inherits(e2, "dq_lazy_ggplot")) e2 <- prep_realize_ggplot(e2)
    f <- get("/.ggplot", envir = asNamespace("patchwork"))
    f(p1, e2)
  }

  S7::method(base_ops[["+"]], signature = list(cls, S7::class_any)) <- function(e1, e2) {
    p1 <- prep_realize_ggplot(dq_lazy_unwrap(e1))
    e2 <- dq_lazy_unwrap(e2)
    if (inherits(e2, "dq_lazy_ggplot")) e2 <- prep_realize_ggplot(e2)
    p1 + e2
  }

  S7::method(base_ops[["-"]], signature = list(cls, S7::class_any)) <- function(e1, e2) {
    p1 <- prep_realize_ggplot(dq_lazy_unwrap(e1))
    e2 <- dq_lazy_unwrap(e2)
    if (inherits(e2, "dq_lazy_ggplot")) e2 <- prep_realize_ggplot(e2)
    p1 - e2
  }

  S7::method(base_ops[["&"]], signature = list(cls, S7::class_any)) <- function(e1, e2) {
    p1 <- prep_realize_ggplot(dq_lazy_unwrap(e1))
    e2 <- dq_lazy_unwrap(e2)
    if (inherits(e2, "dq_lazy_ggplot")) e2 <- prep_realize_ggplot(e2)
    p1 & e2
  }

  S7::method(base_ops[["*"]], signature = list(cls, S7::class_any)) <- function(e1, e2) {
    p1 <- prep_realize_ggplot(dq_lazy_unwrap(e1))
    e2 <- dq_lazy_unwrap(e2)
    if (inherits(e2, "dq_lazy_ggplot")) e2 <- prep_realize_ggplot(e2)
    p1 * e2
  }

  .dq_lazy_state$s7_ready <- TRUE
  invisible(TRUE)
}

test_that(".onLoad works", {
  skip_if_not_installed("pkgload")
  skip_on_cran()
  try(.onLoad(), silent = TRUE)
  expect_equal(
    intersect(
      names(WELL_KNOWN_META_VARIABLE_NAMES),
      names(pkgload::pkg_env("dataquieR"))
    ),
    names(WELL_KNOWN_META_VARIABLE_NAMES)
  )
})

test_that(".set_properties works", {
  p <- as.list(dataquieR.properties)
  withr::defer(list2env(p, dataquieR.properties))
  .set_properties(list(a = "b"))
  expect_identical(dataquieR.properties$a, "b")
})

test_that("with_pipeline works", {
  expect_false(.dq2_globs$.called_in_pipeline)
  with_pipeline(expect_true(.dq2_globs$.called_in_pipeline))
})

#' convert a value to character
#'
#' @param x the value
#' @param error [logical] if `TRUE`, an error is thrown, a warning otherwise
#'                        in case of a conversion error
#' @param error_msg error message to be  displayed, if conversion was not
#'                  possible
#' @param ... additional arguments passed to `util_error` or `util_warning`
#'            respectively in case of an error, and if an `error_msg` has
#'            been passed
#'
#' @return as.character(x)
#'
#' @family robustness_functions
#' @concept robustness
#' @noRd
util_ensure_character <- function(x, error = FALSE, error_msg, ...) {
  if (missing(error_msg)) {
    error_msg <-
      "conversion of %s to a character was not possible for all of its values"
    arg_name <- deparse1(substitute(x))
    error_args <- list(error_msg, sQuote(arg_name))
  } else {
    util_expect_scalar(error_msg,
                       check_type = is.character,
                       allow_more_than_one = TRUE)
    error_args <- list(error_msg, ...)
  }
  util_expect_scalar(error, check_type = is.logical)
  x1 <- suppressWarnings(as.character(x))
  if (!all(is.na(x) == is.na(x1))) {
    do.call(ifelse(error, util_error, util_warning), error_args)
  }
  x1
}

test_that("util_ensure_character works", {
  skip_on_cran()
  # Check that the result is a character
  expect_type(util_ensure_character(666), "character")

  # Check that the result is as expected
  expect_equal(util_ensure_character(c(666, "abc", pi)), c("666", "abc", "3.14159265358979"))

  # Check error messages
  ID <- c("111", "222", "333", "444")
  Age <- c(23, 41, NA, "7a")
  df <- data.frame(ID, Age)

  expect_error(util_ensure_character(df, TRUE), ".*was not possible for all of its values.*")
  expect_error(util_ensure_character(df, TRUE, "My error message"), "My error message")
  })

#' Check for unexpected data record count within segments
#' @description
#' This function contrasts the expected record number in each study segment in
#' the metadata with the actual record number in each segment data frame.
#'
#' [Indicator]
#'
#' @inheritParams .template_function_indicator
#'
#' @param data_record_count [integer]  an integer vector with the number of expected data records, mandatory.
#' @param study_segment [character] a character vector indicating the name of each study data frame, mandatory.
#'
#' @return a [list] with
#'   - `SegmentData`: data frame with the results of the quality check for unexpected data elements
#'   - `SegmentTable`: data frame with selected unexpected data elements check results, used for the data quality report.
#'
#' @details
#' The current implementation does not take into account jump or missing codes, the function is rather based on checking whether NAs are present in the study data
#'
#' @export
int_unexp_records_segment <- function(study_segment,
                                      study_data,
                                      label_col,
                                      item_level = "item_level",
                                      data_record_count, # TODO: DONT PASS 2 VECTORS FOR ASSINGMENTS
                                      meta_data = item_level,
                                      meta_data_segment = "segment_level",
                                      meta_data_v2,
                                      segment_level
                                      ) {

  # Preps and checks ----
  util_maybe_load_meta_data_v2()

  util_ck_arg_aliases()

  if (missing(study_segment) &&
      missing(data_record_count) &&
      missing(meta_data_segment) &&
      formals()$meta_data_segment %in% prep_list_dataframes()) {
    meta_data_segment <- force(meta_data_segment)
  }

  if (missing(study_segment) &&
      missing(data_record_count) &&
      !missing(meta_data_segment)) {
    meta_data_segment <- prep_check_meta_data_segment(meta_data_segment)
    meta_data_segment <- meta_data_segment[
      !util_empty(meta_data_segment[[SEGMENT_RECORD_COUNT]])
      , , drop = FALSE]
    # TODO: if nothing left
    study_segment <- meta_data_segment[[STUDY_SEGMENT]];
    data_record_count <- meta_data_segment[[SEGMENT_RECORD_COUNT]]
  } else if (!missing(meta_data_segment)) {
    util_error(c("I have %s and one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("study_segment",
                   "data_record_count"
                 )),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("study_segment",
                   "data_record_count"
                 )))
  } else if (missing(meta_data_segment) && (
    missing(study_segment) ||
    missing(data_record_count)
  )) {
    util_error(c("I don't have %s and also miss at least",
                 "one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("study_segment",
                   "data_record_count"
                 )),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("study_segment",
                   "data_record_count"
                 )))
  }
  prep_prepare_dataframes(.allow_empty = TRUE)

  # meta_data$STUDY_SEGMENT <-
  #   util_map_labels(meta_data$STUDY_SEGMENT,
  #                   meta_data = meta_data,
  #                   to = label_col,
  #                   ifnotfound = meta_data$STUDY_SEGMENT)

  # Check arguments ----

  util_expect_scalar(study_segment,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)

  util_expect_scalar(data_record_count,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = util_all_is_integer)

  util_stop_if_not(length(data_record_count) ==
                     length(study_segment))

  # check that specified segments are included in the metadata
  old_segments <- study_segment
  segments <- intersect(study_segment, meta_data$STUDY_SEGMENT)

  if (length(old_segments) > length(segments)) {
    util_message(
      c("The segments in the %s do not match the segments in %s,",
        "considering only the intersection"),
      dQuote("meta_data"),
      dQuote("meta_data_segment"),
      applicability_problem = TRUE
    )
  }

  # Check for unexpected records  ----
  names(data_record_count) <- segments

  result <- lapply(setNames(nm = segments), function(current_segment) {

    vars_in_current_segment <-
      util_get_vars_in_segment(current_segment, meta_data = meta_data,
                               label_col = label_col)

    vars_in_current_segment <- intersect(colnames(ds1),
                                         vars_in_current_segment)

    data_records_0 <- util_remove_empty_rows(ds1[, c(vars_in_current_segment)])
    data_records_1 <- subset(data_records_0,
      rowSums(is.na(data_records_0)) != length(vars_in_current_segment),
      drop = FALSE
    )

    # TODO: use `util_observation_expected`
    # The user could have more control to specify which subset of jump codes should be used
    # data_records_1 <- subset(rowSums(data_records_0 %in% participation_jump_codes) == length(vars_in_current_segment))

    # Select segment variables from data
    data_records_cnt <- nrow(data_records_1)
    metadata_records_cnt <- data_record_count[[current_segment]]

    res_tmp <- data.frame(
      check.names = FALSE,
      "Segment" = current_segment,
      "Check" = "Records",
      "Unexpected records" = !(data_records_cnt == metadata_records_cnt),
      "Number of records in data" = data_records_cnt,
      "Number of records in metadata" = metadata_records_cnt,
      "Number of mismatches" =
        abs(round(data_records_cnt - metadata_records_cnt, 3)),
      "Percentage of mismatches" =
        abs(round(100 * (data_records_cnt - metadata_records_cnt) /
                    metadata_records_cnt, 3)),
      "GRADING" = ifelse(data_records_cnt == metadata_records_cnt, 0, 1),
      stringsAsFactors = FALSE
    )
    rownames(res_tmp) <- NULL
    return(res_tmp)
  })

  res_df <- do.call(rbind.data.frame, result)

  rownames(res_df) <- NULL

  res_pipeline <- data.frame(
    "Segment" = res_df$Segment,
    "NUM_int_sts_countre" = res_df$`Number of mismatches`,
    "PCT_int_sts_countre" = res_df$`Percentage of mismatches`,
    "GRADING" = res_df$GRADING,
    stringsAsFactors = FALSE
  )

  return(list(
    SegmentData = res_df,
    SegmentTable = res_pipeline
  ))
}

# nocov start
#' Robust pointrange geom that turns non-finite range endpoints into NA
#'
#' This preserves normal ggplot2 aes inheritance because the cleaning happens
#' after mappings have been evaluated.
#'
#' @inheritParams ggplot2::geom_pointrange
#' @noRd
util_geom_pointrange_robust <- function(mapping = NULL, data = NULL,
                                        stat = "identity",
                                        position = "identity",
                                        ...,
                                        na.rm = TRUE,
                                        show.legend = NA,
                                        inherit.aes = TRUE) {
  if (!identical(na.rm, TRUE)) {
    util_error("Cannot call util_geom_pointrange_robust with na.rm = FALSE")
  }

  ggplot2::layer(
    data = data,
    mapping = mapping,
    stat = stat,
    geom = GeomPointrangeRobust,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = rlang::list2(
      na.rm = na.rm,
      ...
    )
  )
}

#' @noRd
util_geom_pointrange_robust_setup_data <- function(data, params) {
  data <- ggplot2::GeomPointrange$setup_data(data, params)

  range_aes <- base::intersect(
    c("xmin", "xmax", "ymin", "ymax"),
    base::names(data)
  )

  for (nm in range_aes) {
    x <- data[[nm]]
    if (base::is.numeric(x)) {
      x[!base::is.finite(x)] <- NA_real_
      data[[nm]] <- x
    }
  }

  data
}

environment(util_geom_pointrange_robust_setup_data) <- baseenv()

GeomPointrangeRobust <- local({
  ee <- new.env(parent = baseenv())
  ee$Parent <- ggplot2::GeomPointrange
  ee$setup_fun <- util_geom_pointrange_robust_setup_data

  evalq(
    ggplot2::ggproto(
      "GeomPointrangeRobust",
      Parent,
      extra_params = base::c(Parent$extra_params, "na.rm"),
      setup_data = setup_fun
    ),
    envir = ee
  )
})
#' Internally used point-range
#'
#' @param data the data returned by `ggplot2::ggplot_build()`
#' @param prestats_data the data before statistics are computed.
#' @param layout the panel layout.
#' @param params parameters for the geom, statistic, and 'constant' aesthetics
#' @param p a ggplot2 object (the conversion may depend on scales, for instance).
#' @param ... currently ignored
#'
#' @export
to_basic.GeomPointrangeRobust <- function(data, prestats_data, layout, params,
                                          p, ...) {
  for (nm in c("xmin", "xmax", "ymin", "ymax")) {
    if (nm %in% names(data) && is.numeric(data[[nm]])) {
      x <- data[[nm]]
      x[!is.finite(x)] <- NA_real_
      data[[nm]] <- x
    }
  }

  data$width <- 0

  list(
    util_prefix_class(data, "GeomErrorbar"),
    util_prefix_class(data, "GeomPoint")
  )
}

#' @noRd
util_prefix_class <- function(x, cls) {
  class(x) <- c(cls, setdiff(class(x), cls))
  x
}
# nocov end

#' Get description for a call
#'
#' @param cn the call name
#'
#' @return the description
#'
#' @noRd
util_col_description <- function(cn) {

  fname <- util_map_by_largest_prefix(
    cn,
    haystack = names(.manual$titles))

  if (is.na(fname)) {
    fname <- cn
  }

  return(util_function_description(fname))

}

#' Isolate enclosure of a function to avoid exporting too much for parallel
#'
#' @param FUN [function()] the function to isolate
#' @param vars [character()] object names from `envir` to have in the
#'                           isolated version of `FUN`
#' @param envir [environment()] environment to copy the objects listed in
#'                              `vars` from
#'
#' @returns [function()] the isolated function
#' @noRd
util_isolate_function <- function(FUN, vars = character(),
                                  envir = parent.frame()) {
  # get objects to keep
  obj_list <- mget(vars, envir = envir, inherits = TRUE)

  # build clean environment
  fn_env <- list2env(obj_list, parent = environment(util_isolate_function))

  # update closure
  environment(FUN) <- fn_env

  return(FUN)
}

#' Amend missing item-level metadata, only
#'
#' @inheritParams .template_function_indicator
#' @inheritParams prep_study2meta
#' @returns [data.frame] fixed `meta_data`
#' @noRd
util_amend_missing_metadata <- function(study_data,
                                        meta_data,
                                        level = c(
                                          VARATT_REQUIRE_LEVELS$REQUIRED,
                                          VARATT_REQUIRE_LEVELS$RECOMMENDED
                                        ),
                                        cumulative = TRUE,
                                        guess_missing_codes =
                                          getOption("dataquieR.guess_missing_codes",
                                                    dataquieR.guess_missing_codes_default)) {
  to_amend <- !(colnames(study_data) %in% meta_data[[VAR_NAMES]])
  if (any(to_amend)) {
    util_message(
      "Missing %s from %s, amending guessed %s for these items/variables...",
      util_pretty_vector_string(colnames(study_data)[to_amend]),
      sQuote("meta_data"),
      "item-level metadata",
      applicability_problem = TRUE,
      intrinsic_applicability_problem = FALSE
    )
    am_md <- prep_study2meta(
      study_data =
        study_data[, colnames(study_data)[to_amend], FALSE],
      level = level,
      cumulative = cumulative,
      convert_factors = FALSE,
      guess_missing_codes = guess_missing_codes
    )
    meta_data <- util_rbind(meta_data, am_md)
  }
  return(meta_data)
}

#' Utility function to prepare the metadata for proportion checks
#'
#' @param resp_vars [variable list] the names of the measurement variables
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param ds1 [data.frame] the data frame that contains the measurements
#'                               (hint: missing value codes should be excluded,
#'                               so the function should be called with `ds1`, if
#'                               available)
#' @param report_problems [enum] Should missing metadata information be reported
#'                               as error, warning or message?
#'
#' @inheritParams .template_function_indicator
#'
#' @return a [list] with the expected range for the proportion check
#'
#' @family lookup_functions
#' @concept metadata_management
#' @noRd

util_prep_proportion_check <- function(resp_vars,
                                       meta_data,
                                       ds1,
                                       report_problems = c("error", "warning",
                                                           "message"),
                                       label_col = attr(ds1, "label_col")) {
  resp_vars <- prep_get_labels(resp_vars = resp_vars,
                               item_level = meta_data,
                               label_col = label_col,
                               force_label_col = "TRUE")
  if (!isTRUE(attr(ds1, "apply_fact_md_inadm"))) {
    util_error("Internal error, sorry, please report: proprotion check called with bare study data or without inadmissible values.") # TODO: I think we should also allow to have inadmissible values already removed because this utility function is called in the accuracy section.
  }
  report_problems <- match.arg(report_problems)
  prop_range <- setNames(util_find_var_by_meta(resp_vars = resp_vars,
                                               target = PROPORTION_RANGE,
                                               meta_data = meta_data,
                                               allowed_sources = label_col),
                         nm = resp_vars)
  if (VALUE_LABELS %in% colnames(meta_data)) {
    util_error("Internal error, sorry, please report: VALUE_LABELS in util_prep_proportion_check.")
  }
  # meta_data <- util_normalize_value_labels(meta_data) # TODO: STANDARDIZED_VOCABULARY_TABLE
  if (VALUE_LABEL_TABLE %in% colnames(meta_data)) {
    val_label_tables <- setNames(util_find_var_by_meta(resp_vars = resp_vars,
                                                 target = VALUE_LABEL_TABLE,
                                                 meta_data = meta_data,
                                                 allowed_sources = label_col),
                           nm = resp_vars)
  } else {
    val_label_tables <- setNames(rep(NA_character_, length(resp_vars)),
                               nm = resp_vars)
  }


  prop_range_int <-
    lapply(setNames(nm = resp_vars), function(rv) {
      rv_is_factor <- is.factor(ds1[[rv]])
      if (rv_is_factor) {
        vv_codes <- levels(ds1[[rv]])
      } else {
        vv_codes <- unique(ds1[[rv]])
        vv_codes <- sort(vv_codes[!is.na(vv_codes)])
      }
      VL <- NULL
      if (rv_is_factor) {
        vlt <- val_label_tables[[rv]]
        if (!is.null(vlt) && !util_empty(vlt)) {
          vlt_rv <- prep_get_data_frame(vlt)
          VL <- setNames(vlt_rv[[CODE_VALUE]], nm = vlt_rv[[CODE_LABEL]])
        }
      }
      if (grepl(SPLIT_CHAR, prop_range[[rv]], fixed = TRUE) |
          grepl(" in ", prop_range[[rv]])) {
        # ranges given per category
        input <- gsub(" in ", " = ", prop_range[[rv]])
        int_list <- util_parse_assignments(input)
        if (rv_is_factor & !is.null(VL) & all(names(int_list) %in% VL)) {
          # If the levels are given as integers, we need to transform them to
          # value labels.
          names(int_list) <- unlist(lapply(names(int_list), function(cc) {
            names(VL)[VL == as.character(cc)]
          }))
        }
        range_per_cat <- lapply(setNames(nm = vv_codes), function(cc) {
          if (cc %in% names(int_list)) {
            util_parse_interval(int_list[[as.character(cc)]])
          } else {
            NA
          }
        })
      } else { # one range given
        int <- util_parse_interval(prop_range[[rv]])
        range_per_cat <- lapply(setNames(nm = vv_codes), function(cc) { int })
      }
      if (all(is.na(range_per_cat))) {
        util_warning(paste0(
          "For ", rv, ", the given PROPORTION_RANGE could not be ",
          "interpreted as an interval."),
          applicability_problem = TRUE)
       # range_per_cat <- NA
      }
      range_per_cat
    })

  rvs_with_prop <- names(prop_range_int)[
    vapply(prop_range_int,
           FUN = function(x) {
             if (is.list(x)) {
               !all(is.na(x))
             } else {
               !is.na(x)
             }},
           FUN.VALUE = logical(1))]

  rep_fun <- switch(report_problems,
         error = util_error,
         warning = util_warning,
         message = util_message)

  if (length(resp_vars) > length(rvs_with_prop)) {
    rep_fun(paste0(
      "For ", paste(setdiff(resp_vars, rvs_with_prop), collapse = ", "),
      ", the metadata for a proportion check is missing (PROPORTION_RANGE)."),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE)
  }

  return(list("Range" = prop_range_int))
}

#' Paste strings but keep NA
#'
#' @param ... other arguments passed to `paste`
#'
#' @return [character] pasted strings
#'
#' @noRd
util_paste_with_na <- function(...) {
  res <- paste(...)
  res[as.logical(rowSums(do.call(data.frame, lapply(list(...), is.na))))] <-
    NA_character_
  res
}

#' Paste strings but keep NA (`paste0`)
#'
#' @param ... other arguments passed to `paste0`
#'
#' @return [character] pasted strings
#'
#' @noRd
util_paste0_with_na <- function(...) {
  res <- paste0(...)
  res[as.logical(rowSums(do.call(data.frame, lapply(list(...), is.na))))] <-
    NA_character_
  res
}

#' Support function to scan variable labels for applicability
#'
#' @description
#'
#' Adjust labels in meta_data to be valid variable names in formulas for
#' diverse r functions, such as [`glm`] or [`lme4::lmer`].
#'
#' @details
#'
#' Hint: The following is still true, but the functions should be capable
#' of doing potentially needed fixes on-the-fly automatically, so likely
#' you will not need this function any more.
#'
#' Currently, labels as given by `label_col` arguments in the most functions
#' are directly used in formula, so that they become natural part of the
#' outputs, but different models expect differently strict syntax for such
#' formulas, especially for valid variable names. `prep_clean_labels` removes
#' all potentially inadmissible characters from variable names (no guarantee,
#' that some exotic model still rejects the names, but minimizing the number
#' of exotic characters). However, variable names are modified, may become
#' unreadable or indistinguishable from other variable names. For the latter
#' case, a `stop` call is possible, controlled by the `no_dups` argument.
#'
#' A warning is emitted, if modifications were necessary.
#'
#' @param label_col [character] label attribute to adjust or character vector to
#'                              adjust, depending on `meta_data` argument is
#'                              given or missing.
#' @param item_level [data.frame] metadata data frame: If `label_col` is a label
#'                               attribute to adjust, this is the metadata
#'                               table to process on. If missing, `label_col`
#'                               must be a character vector with values to
#'                               adjust.
#' @param meta_data [data.frame] old name for `item_level`
#' @param no_dups   [logical] disallow duplicates in input or output vectors of
#'                            the function, then, prep_clean_labels would call
#'                            `stop()` on duplicated labels.
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return a data.frame with:
#'  - if `meta_data` is set, a list with:
#'    - modified `meta_data[, label_col]` column
#'  - if `meta_data` is not set, adjusted labels that then were directly given
#'    in label_col
#'
#' @export
#' @importFrom stats setNames
#'
#' @examples
#' \dontrun{
#' meta_data1 <- data.frame(
#'   LABEL =
#'     c(
#'       "syst. Blood pressure (mmHg) 1",
#'       "1st heart frequency in MHz",
#'       "body surface (\\u33A1)"
#'     )
#' )
#' print(meta_data1)
#' print(prep_clean_labels(meta_data1$LABEL))
#' meta_data1 <- prep_clean_labels("LABEL", meta_data1)
#' print(meta_data1)
#' }
prep_clean_labels <- function(label_col, item_level = "item_level",
                              no_dups = FALSE, meta_data = item_level,
                              meta_data_v2) {
  util_maybe_load_meta_data_v2()
  if (missing(label_col)) {
    util_error("Need at least on paramter")
  }
  if (is.factor(label_col)) {
    label_col <- as.character(label_col)
  }
  if (!(missing(meta_data) && is.character(meta_data))) {
    util_expect_data_frame(meta_data)
  }
  if (!(missing(item_level) && is.character(item_level))) {
    util_expect_data_frame(item_level)
  }
  util_stop_if_not((is.character(label_col) && missing(meta_data) &&
                      missing(item_level)) ||
              (is.data.frame(meta_data) || is.data.frame(item_level)))
  if (!missing(meta_data)) {
    util_expect_data_frame(meta_data, c(label_col))
    orig_col <- meta_data[[label_col]]
  } else {
    orig_col <- label_col
  }
  if (no_dups && any(duplicated(orig_col))) {
    util_error("Have duplicates in desired variable labels",
               applicability_problem = TRUE)
  }
  adjusted_col <- gsub("[^a-zA-Z0-9_]+", "_", orig_col)
  adjusted_col <- gsub("^[^a-zA-Z]+", "", adjusted_col)
  if (no_dups && any(duplicated(adjusted_col))) {
    util_error(c(
      "Have duplicates in desired variable labels after",
      "adjusting them to be valid variable names"),
      applicability_problem = TRUE)
  }
  if (any(na.omit(orig_col != adjusted_col))) {
    if (!missing(meta_data)) {
      util_message("Adjusted labels in %s to be valid variable names.",
                   dQuote(label_col), applicability_problem = TRUE)
    } else {
      util_message("Adjusted labels to be valid variable names.",
                   applicability_problem = TRUE)
    }
  }
  if (!missing(meta_data)) {
    meta_data[[label_col]] <- adjusted_col
  } else {
    meta_data <- adjusted_col
  }
  meta_data
}

test_that("prep_clean_labels works", {
  skip_on_cran()
  meta_data1 <- data.frame(
    LABEL =
      c(
        "syst. Blood pressure (mmHg) 1",
        "1st heart frequency in MHz",
        "body surface (\\u33A1)"
      )
  )
  expect_message2(
    expect_equal(prep_clean_labels(meta_data1$LABEL),
                 c("syst_Blood_pressure_mmHg_1", "st_heart_frequency_in_MHz",
                   "body_surface_u33A1_")),
    regexp = "Adjusted labels to be valid variable names."
  )
  expect_message2(
    expect_equal(prep_clean_labels("LABEL", meta_data1),
               structure(list(LABEL = c("syst_Blood_pressure_mmHg_1",
                                        "st_heart_frequency_in_MHz",
                                        "body_surface_u33A1_")),
                         row.names = c(NA, -3L), class = "data.frame")),
    regexp = "Adjusted labels in .{1,4}LABEL.{1,4} to be valid variable names."
  )
  meta_data2 <- data.frame(
    LABEL =
      c(
        "syst. Blood pressure (mmHg) 1",
        "syst. Blood pressure  mmHg! 1",
        "body surface (\\u33A1)"
      )
  )
  expect_error(print(prep_clean_labels(meta_data2$LABEL, no_dups = TRUE)),
                   regexp = "Have duplicates in desired variable labels"
  )
  expect_message2(
    expect_equal(prep_clean_labels(meta_data2$LABEL, no_dups = FALSE),
               c("syst_Blood_pressure_mmHg_1", "syst_Blood_pressure_mmHg_1",
                 "body_surface_u33A1_")),
    regexp = "Adjusted labels to be valid variable names."
  )
})

#' Add computation rules from cross-item-level-indicator-columns
#'
#' @param meta_data_item_computation existing computation rules
#' @param meta_data_cross_item cross-item metadata
#' @param meta_data main metadata
#' @param label_col column to store label copies
#'
#' @noRd
#'
#' @returns list with updated `meta_data_item_computation` and `meta_data`
#' @examples
#' \dontrun{
#' dq_report2("study_data", meta_data_v2 = "~/tmp/mls.xlsx")
#' }
util_add_computed_internals <- function(meta_data_item_computation,
                                        meta_data_cross_item,
                                        meta_data,
                                        label_col) {
  if (missing(meta_data_cross_item)) {
    meta_data_cross_item <- NULL
  }
  suppressWarnings(suppressMessages(
    meta_data_cross_item <- util_normalize_cross_item(
      meta_data = meta_data,
      meta_data_cross_item = meta_data_cross_item,
      label_col = label_col
    )
  ))

  meta_data_item_computation_internal <- data.frame(
    VAR_NAMES = character(0),
    COMPUTATION_RULE = character(0),
    DATA_PREPARATION = character(0),
    CHECK_ID = character(0)
  )

  meta_data_internal <- data.frame(
    VAR_NAMES = character(0),
    DATA_TYPE = character(0)
  )

  COLUMNS_WITH_SSI <- setdiff(names(COMPUTED_VARIABLE_ROLES), "NA")

  for (ssi_col in COLUMNS_WITH_SSI) {
    if (!ssi_col %in% colnames(meta_data_cross_item)) next

    rows <- which(!util_empty(meta_data_cross_item[[ssi_col]]))
    if (length(rows) == 0) next

    rule_info <- util_get_concept_info("ssi", get("SSI_METRICS") == ssi_col)
    rule_template <- rule_info$rule_template[1]
    data_preparation <- rule_info$data_preparation[1]
    extra_col <- rule_info$extra_arguments[1]
    human_prefix_from_result_caption <- rule_info$result_caption[1]
    human_prefix_from_menu_label <- rule_info$menu_label[1]
    scale_level <- SCALE_LEVELS[[rule_info$scale_level[1]]]
    data_type <- DATA_TYPES[[rule_info$data_type[1]]]

    e <- new.env(parent = emptyenv())
    e$given <- character(0)
    e$given_lab <- character(0)
    e$given_llab <- character(0)
    e$given_seg <- character(0)

    tech_prefix <- gsub("[^A-Z0-9_]", "", toupper(ssi_col))

    to_add <- lapply(rows, function(rw, e) {
      parsed <- suppressWarnings(
        tryCatch(
          util_parse_interval(meta_data_cross_item[rw, ssi_col]),
          warning = function(w) {
            util_warning(
              "Invalid interval in column %s at row %d: %s - %s.",
              ssi_col,
              rw,
              meta_data_cross_item[rw, ssi_col],
              conditionMessage(w),
              applications_problem = TRUE
            )
            NA
          }
        )
      )
      if (!inherits(parsed, "interval")) {
        return(NULL)
      }

      check_id <- meta_data_cross_item[rw, CHECK_ID]
      hard_limits <- meta_data_cross_item[rw, ssi_col]
      check_label <- meta_data_cross_item[rw, CHECK_LABEL]
      scale_name <- meta_data_cross_item[rw, SCALE_NAME]
      scale_acronym <- meta_data_cross_item[rw, SCALE_ACRONYM]
      computed_variable_role <- COMPUTED_VARIABLE_ROLES[[ssi_col]]
      if (!length(computed_variable_role)) {
        util_error(
          c("Internal error, sorry, please report: Detected unkown",
            "computed varialbe role: %s"),
          ssi_col
        )
      }

      study_segment <- util_free_varname(
        meta_data,
        paste0("COMPUTED_", tech_prefix),
        target = STUDY_SEGMENT,
        also_not = e$given_seg
      )
      e$given_seg <- c(e$given_seg, study_segment)

      prefix <- tech_prefix
      if (!!length(scale_name) && !util_empty(scale_name)) {
        prefix <- paste0(prefix, "_", scale_name)
      } else if (!!length(check_label) && !util_empty(check_label)) {
        prefix <- paste0(prefix, "_", check_label)
      } else if (!!length(scale_acronym) && !util_empty(scale_acronym)) {
        prefix <- paste0(prefix, "_", scale_acronym)
      }
      var_names <- util_free_varname(
        meta_data, prefix, target = VAR_NAMES, also_not = e$given
      )
      e$given <- c(e$given, var_names)

      prefix <- paste0(human_prefix_from_menu_label, ": ")
      if (!!length(scale_name) && !util_empty(scale_name)) {
        prefix <- paste0(prefix, scale_name)
        if (!!length(check_label) && !util_empty(check_label)) {
          prefix <- paste0(prefix, " (", check_label, ")")
        }
      } else if (!!length(check_label) && !util_empty(check_label)) {
        prefix <- paste0(prefix, check_label)
      }
      label <- util_free_varname(
        meta_data, prefix, target = LABEL, also_not = e$given_lab
      )
      e$given_lab <- c(e$given_lab, var_names)

      long_prefix <- human_prefix_from_result_caption
      if (!!length(scale_name) && !util_empty(scale_name)) {
        long_prefix <- paste0(long_prefix, " for ", scale_name)
      }
      if (!!length(scale_acronym) && !util_empty(scale_acronym) &&
          !!length(check_label) && !util_empty(check_label)) {
        long_prefix <- paste0(long_prefix, " (", scale_acronym, " -- ",
                              check_label, ")")
      } else if (!!length(check_label) && !util_empty(check_label)) {
        long_prefix <- paste0(long_prefix, " (", check_label, ")")
      } else if (!!length(scale_acronym) && !util_empty(scale_acronym)) {
        long_prefix <- paste0(long_prefix, " (", scale_acronym, ")")
      }
      long_label <- util_free_varname(
        meta_data, long_prefix, target = LONG_LABEL, also_not = e$given_llab
      )
      e$given_llab <- c(e$given_llab, var_names)

      var_expr <- paste(
        collapse = ", ",
        paste0("[", util_parse_assignments(
          meta_data_cross_item[rw, VARIABLE_LIST_ORDER]), "]")
      )

      rule_filled <- rule_template
      rule_filled <- gsub("!!!VAR_NAMES", var_expr, rule_filled, fixed = TRUE)

      if (!is.na(extra_col)) {
        extra_args <- NA
        if (extra_col %in% colnames(meta_data_cross_item))
        if (util_empty(meta_data_cross_item[rw, extra_col])) {
          extra_args <- "[.]"
        } else {
          extra_args <-
            util_parse_assignments(meta_data_cross_item[rw, extra_col])
        }
        rule_filled <- gsub(paste0("!!!", extra_col), extra_args,
                            rule_filled, fixed = TRUE)
      }

      computation_rule <- rule_filled
      variable_role <- VARIABLE_ROLES$PRIMARY

      list(
        check_id = check_id,
        var_names = var_names,
        computation_rule = computation_rule,
        data_preparation = data_preparation,
        scale_level = scale_level,
        data_type = data_type,
        variable_role = variable_role,
        computed_variable_role = computed_variable_role,
        hard_limits = hard_limits,
        label = label,
        long_label = long_label,
        study_segment = study_segment
      )
    }, e = e)

    to_add <- Filter(Negate(is.null), to_add)

    .meta_data_item_computation_internal <- util_rbind(
      data_frames_list = lapply(lapply(to_add, `[`, tolower(
        c(VAR_NAMES, COMPUTATION_RULE, DATA_PREPARATION, CHECK_ID))),
        as.data.frame)
    )
    colnames(.meta_data_item_computation_internal) <-
      toupper(colnames(.meta_data_item_computation_internal))
    meta_data_item_computation_internal <- util_rbind(
      meta_data_item_computation_internal,
      .meta_data_item_computation_internal
    )

    .meta_data_internal <- util_rbind(
      data_frames_list = lapply(lapply(to_add, function(x) {
        row <- as.data.frame(x[tolower(c(
          VAR_NAMES,
          SCALE_LEVEL,
          DATA_TYPE,
          VARIABLE_ROLE,
          COMPUTED_VARIABLE_ROLE,
          HARD_LIMITS,
          LABEL,
          LONG_LABEL,
          STUDY_SEGMENT,
          CHECK_ID
        ))])
        names(row) <- toupper(names(row))
        if (!is.null(label_col) &&
            !(toupper(label_col) %in% c(LABEL, LONG_LABEL, VAR_NAMES))) {
          row[[label_col]] <- x$label
        }
        row
      }), identity)
    )
    colnames(.meta_data_internal) <- toupper(colnames(.meta_data_internal))
    meta_data_internal <- util_rbind(meta_data_internal, .meta_data_internal)
  }

  list(
    meta_data_item_computation = util_rbind(
      meta_data_item_computation,
      meta_data_item_computation_internal
    ),
    meta_data = util_rbind(
      meta_data,
      meta_data_internal
    )
  )
}

#' Verify and normalize metadata on segment level
#'
#' if possible, mismatching data types are converted (`"true"` becomes `TRUE`)
#'
#' missing columns are added, filled with `NA`, if this is valid, i.e., n.a.
#' for `STUDY_SEGMENT` as the key column
#'
#' @param meta_data_segment [data.frame] data frame or path/url of a metadata
#'                                       sheet for the segment level
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#' @param segment_level [data.frame] alias for `meta_data_segment`
#' @return standardized metadata sheet as data frame
#' @export
#'
#' @examples
#' \dontrun{
#' mds <- prep_check_meta_data_segment("ship_meta_v2|segment_level") # also converts
#' print(mds)
#' prep_check_meta_data_segment(mds)
#' mds1 <- mds
#' mds1$SEGMENT_RECORD_COUNT <- NULL
#' print(prep_check_meta_data_segment(mds1)) # fixes the missing column by NAs
#' mds1 <- mds
#' mds1$SEGMENT_UNIQUE_ROWS[[2]] <- "xxx" # not convertible
#' # print(prep_check_meta_data_segment(mds1)) # fail
#' }
prep_check_meta_data_segment <- function(meta_data_segment = "segment_level",
                                         meta_data_v2,
                                         segment_level
                                         ) {

  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()
  util_expect_data_frame(meta_data_segment)

  if (!(SEGMENT_RECORD_COUNT %in% colnames(meta_data_segment))) {
    meta_data_segment$SEGMENT_RECORD_COUNT <- rep(NA_integer_,
                                                  nrow(meta_data_segment))
  }
  if (!(SEGMENT_ID_REF_TABLE %in% colnames(meta_data_segment))) {
    if ("SEGMENT_ID_TABLE" %in% colnames(meta_data_segment)) {
      util_message("Did not find %s in metadata, but %s, renaming it...",
                   dQuote("SEGMENT_ID_REF_TABLE"),
                   dQuote("SEGMENT_ID_TABLE"),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = FALSE)
      colnames(meta_data_segment)[colnames(meta_data_segment) ==
                                    "SEGMENT_ID_TABLE"] <-
        SEGMENT_ID_REF_TABLE
    } else {
      meta_data_segment$SEGMENT_ID_REF_TABLE <- rep(NA_character_,
                                                    nrow(meta_data_segment))
    }
  }
  if (!(SEGMENT_RECORD_CHECK %in% colnames(meta_data_segment))) {
    meta_data_segment$SEGMENT_RECORD_CHECK <- rep(NA_character_,
                                                  nrow(meta_data_segment))
  }
  if (!(SEGMENT_ID_VARS %in% colnames(meta_data_segment))) {
    meta_data_segment$SEGMENT_ID_VARS <- rep(NA_character_,
                                             nrow(meta_data_segment))
  }
  if (!(SEGMENT_PART_VARS %in% colnames(meta_data_segment))) {
    meta_data_segment$SEGMENT_PART_VARS <- rep(NA_character_,
                                               nrow(meta_data_segment))
  }
  if (!(SEGMENT_UNIQUE_ROWS %in% colnames(meta_data_segment))) {
    meta_data_segment$SEGMENT_UNIQUE_ROWS <- rep(NA,
                                                 nrow(meta_data_segment))
  }
  if (!(SEGMENT_UNIQUE_ID %in% colnames(meta_data_segment))) {
    meta_data_segment[[SEGMENT_UNIQUE_ID]] <- rep(1,
                                                  nrow(meta_data_segment))
  }

  r <- util_expect_data_frame(
    custom_errors = list(
      SEGMENT_ID_VARS = "Column DF_ID_VARS must be character",
      SEGMENT_PART_VARS = "Column SEGMENT_PART_VARS must be character",
      STUDY_SEGMENT = "Column STUDY_SEGMENT must be character",
      SEGMENT_RECORD_COUNT = "Column SEGMENT_RECORD_COUNT must be integer",
      SEGMENT_ID_REF_TABLE = "Column SEGMENT_ID_REF_TABLE must be character",
      SEGMENT_RECORD_CHECK = "Column SEGMENT_RECORD_CHECK must be subset, superset or exact",
      SEGMENT_UNIQUE_ROWS = "Column SEGMENT_UNIQUE_ROWS must be true, false, or no_id",
      SEGMENT_UNIQUE_ID = "Column SEGMENT_UNIQUE_ID must be integer"
    ),
    meta_data_segment,
    list(
      SEGMENT_ID_VARS = is.character,
      SEGMENT_PART_VARS = is.character,
      STUDY_SEGMENT = is.character,
      SEGMENT_RECORD_COUNT = util_all_is_integer,
      SEGMENT_ID_REF_TABLE = is.character,
      SEGMENT_RECORD_CHECK = function(x) {
        all(util_empty(x) | x %in% c("superset", "subset", "exact"))
      },
      SEGMENT_UNIQUE_ROWS = function(x) {
        all(util_empty(x) | tolower(trimws(x)) %in%
              c("f", "t", "true", "false", "no_id"))
      },
      SEGMENT_UNIQUE_ID = util_all_is_integer
    ),
    list(
      SEGMENT_ID_VARS = as.character, # TODO: Write utility functions for converting data types with proper warnings
      SEGMENT_PART_VARS = as.character,
      STUDY_SEGMENT = as.character,
      SEGMENT_RECORD_COUNT = as.integer,
      SEGMENT_ID_REF_TABLE = as.character,
      SEGMENT_RECORD_CHECK = function(x) {
        r <-
          factor(tolower(trimws(as.character(x))),
                 levels = c("superset", "subset", "exact"))
        levels(r)[as.numeric(r)]
      },
      SEGMENT_UNIQUE_ROWS = as.character,
      SEGMENT_UNIQUE_ID = as.integer
    )
  )

  no_key <- util_empty(r[[STUDY_SEGMENT]])
  sum_no_key <- sum(no_key)

  if (sum_no_key) {
    util_message("Removing %d rows from %s, because %s is empty.",
                 sum_no_key,
                 dQuote("meta_data_segment"),
                 sQuote(STUDY_SEGMENT),
                 applicability_problem = TRUE
    )
  }

  r <- r[!no_key, , drop = FALSE]

  r
}

#' Internal function only existing for technical reasons, planned to be removed in
#' future releases
#'
#' `r lifecycle::badge("experimental")`
#'
#' @description Please use instead the function [acc_mahalanobis()]
#'
#' [Indicator]
#'
#' @details
#' # ALGORITHM OF THIS IMPLEMENTATION:
#' - Implementation is restricted to variables of type integer
#' - Remove missing codes from the study data (if defined in the metadata)
#' - The covariance matrix is estimated for all variables of `resp_vars`
#' - The `Mahalanobis` distance of each observation is calculated
#'   \eqn{MD^2_i  = (x_i - \mu)^T \Sigma^{-1} (x_i -  \mu)}
#' - The default to consider a value an `outlier` is to use the 0.975 quantile
#'    of a theoretical chi-square distribution with degrees of freedom
#'    equals to the number of variables used to calculate the
#'    `Mahalanobis` distance (`Mayrhofer and Filzmoser`, 2023)
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable] the names of the computed variable
#'                                    containing `Mahalanobis` distance ratio
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param label_col [variable attribute] the name of the column in the
#'                                       metadata containing the labels of
#'                                       the variables
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param meta_data_v2 [character] path or file name of the workbook like
#'                                 metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`
#' @param meta_data_cross_item [data.frame] -- Cross-item level metadata
#' @param cross_item_level [data.frame] alias for `meta_data_cross_item`
#' @param `cross-item_level` [data.frame] alias for `meta_data_cross_item`
#'
#'
#' @return a list with:
#'   - `SummaryData`: [data.frame] underlying the plot with user friendly caption
#'   - `SummaryTable`: [data.frame] underlying the plot
#'   - `SummaryPlot`: [ggplot2::ggplot2] Q-Q plot of squared `Mahalanobis`
#'                                        distances vs. a theoretical
#'                                        chi-squared distribution showing `outliers`.
#'   - `FlaggedStudyData` [data.frame] contains the original data frame of the
#'                                      variables used to calculate
#'                                      the squared `Mahalanobis` distances
#'                                      with an additional column indicating if
#'                                      for a group of variables if the
#'                                      observational unit is a
#'                                      multivariate `outlier`.
#'
#' @export
#'
#' @importFrom ggplot2 ggplot aes geom_path  scale_color_manual geom_point discrete_scale theme_minimal scale_alpha_manual
#'
#' @importFrom stats mahalanobis
#' @importFrom rlang .data
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_multivariate_outlier.html
#' )
acc_mahalanobis_ratio <- function(resp_vars = NULL,
                                  study_data,
                                  label_col = VAR_NAMES,
                                  item_level = "item_level",
                                  meta_data = item_level,
                                  meta_data_v2,
                                  meta_data_cross_item = "cross-item_level",
                                  cross_item_level,
                                  `cross-item_level`) {
  # preps -----------------------------------------------
  util_maybe_load_meta_data_v2()
  # Load cross-item_level metadata and normalize it ----
  # check if there is a cross item metadata and if it is a data frame
  # in case is not present, create an empty data frame for cross item metadata
  try(util_expect_data_frame(meta_data_cross_item), silent = TRUE)
  if (!is.data.frame(meta_data_cross_item)) {
    util_message(sprintf(
      "No cross-item level metadata %s found",
      sQuote(meta_data_cross_item)))
    meta_data_cross_item <- data.frame(VARIABLE_LIST = character(0),
                                       CHECK_LABEL = character(0))
  }

  # First normalize input for meta_data_cross_item from the user
  meta_data_cross_item <- util_normalize_cross_item(
    meta_data = meta_data,
    meta_data_cross_item = meta_data_cross_item,
    label_col = label_col
  )

  #Check label_col
  if (missing(label_col)) {
    orig_label_col <- rlang::missing_arg()
  } else {
    orig_label_col <- force(label_col)
  }

  label_col <- attr(prep_get_labels("",
                                    item_level = meta_data,
                                    label_class = "SHORT",
                                    label_col = label_col),
                    "label_col")

  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = FALSE) #Otherwise it already removes outliers (ratio values > 1)!

  #check for variable role in the metadata for the resp_var
  util_correct_variable_use(resp_vars,
                            need_type = DATA_TYPES$FLOAT,
                            need_scale = SCALE_LEVELS$RATIO,
                            need_computed_role = COMPUTED_VARIABLE_ROLES$MAHALANOBIS_RATIO
  )
  #Check resp_vars
  util_expect_scalar(
    arg_name = resp_vars,
    allow_more_than_one = FALSE,
    check_type = is.character
  )


  # select current variable from data ------------------------------------
  #Select CHECK_ID of the current variables group
  current_check_id <- util_map_labels(resp_vars, meta_data = meta_data, to = CHECK_ID, from = label_col)
  all_checkID_with_vars <- setNames(meta_data_cross_item$VARIABLE_LIST,
                                    nm = meta_data_cross_item$CHECK_ID)
  intermediate2 <- lapply(util_parse_assignments(all_checkID_with_vars,
                                                 multi_variate_text = TRUE),
                          lapply,
                          prep_get_labels,
                          label_col = VAR_NAMES,
                          force_label_col = "TRUE",
                          item_level = meta_data)
  intermediate3 <- lapply(intermediate2, unique)
  no_vars_per_check_ID <- vapply(lapply(intermediate3,
                                        function(vl) intersect(unlist(vl), meta_data$VAR_NAMES)), length,
                                 FUN.VALUE = integer(1))
 # vars_per_check_ID  <- lapply(intermediate2, names)

  current_no_vars <- as.numeric(no_vars_per_check_ID[current_check_id])
  current_vars <- all_checkID_with_vars[[current_check_id]]
  rm(all_checkID_with_vars, intermediate2, intermediate3, no_vars_per_check_ID)


  FlaggedStudyData <- ds1
  current_MD_ratio <- ds1[[resp_vars]]
  current_MD_ratio <-  current_MD_ratio[order(current_MD_ratio)]

  #current_MD_ratio <- current_MD_ratio[order(current_MD_ratio)]


  #Define the threshold ----
  threshold_defined <-meta_data_cross_item[meta_data_cross_item$CHECK_ID == current_check_id,
                         MAHALANOBIS_THRESHOLD]
  if (tolower(trimws(threshold_defined)) %in% c("true", "1", "t", "+")) {
    mahalanobis_threshold <- dataquieR.MAHALANOBIS_THRESHOLD_default
  } else {
    mahalanobis_threshold <- as.numeric(threshold_defined)
    util_expect_scalar(mahalanobis_threshold)
  }
  rm(threshold_defined)


  MD_outliers_threshold <- stats::qchisq(mahalanobis_threshold,
                                         df = current_no_vars)




  n_prior <- length(current_MD_ratio)
  current_MD_ratio_fin <- current_MD_ratio[!is.na(current_MD_ratio)]
  n_post <- length(current_MD_ratio_fin)

  if (n_post == 0) {
    util_error("No samples with complete cases in the variables %s. Aborting.",
               paste0(sQuote(vars), collapse = ", "),
               applicability_problem = FALSE)
  }

  if (n_post < n_prior) {
    util_message(paste0(
      "Due to missing values",
      " N=", n_prior - n_post,
      " observational units were excluded."
    ), applicability_problem = FALSE)
  }


  #Create plot
  res_MD <- util_create_mahalanobis_ggplot(MD_ratio = current_MD_ratio_fin,
                                           mahalanobis_threshold = mahalanobis_threshold,
                                           df = current_no_vars)
  p1 <- res_MD$plot_MD

  # new complete data with the column indicating the outliers
  FlaggedStudyData$MD_outliers <- NA
  FlaggedStudyData$MD_outliers <- ifelse(FlaggedStudyData[resp_vars] > 1, 1, 0)
  n_non_ol <- sum(FlaggedStudyData$MD_outliers == 0,na.rm = TRUE)
  n_devs <- sum(FlaggedStudyData$MD_outliers == 1,na.rm = TRUE)



  # create summary table
  st1 <- data.frame(Variables = resp_vars)
  st1$"MD_outliers (N)" <- n_devs
  st1$"MD_outliers (%)" <- round(n_devs/n_post*100,
                                 digits = 2)
  st1$"N" <- n_post
  st1$"observational_units_removed" <- n_prior - n_post
  st1$"mahalanobis_threshold" <- mahalanobis_threshold

  SummaryData <- st1
  SummaryTable <- st1
  names(SummaryTable)[names(SummaryTable) == "MD_outliers (N)"] <- "NUM_ssc_mah" # add indicator to DQ_OBS to make grading rules work, here.
  names(SummaryTable)[names(SummaryTable) == "MD_outliers (%)"] <- "PCT_ssc_mah"
  SummaryTable <- SummaryTable[, c("Variables", "NUM_ssc_mah", "PCT_ssc_mah")]



  #Add new attribute to the columns of SummaryData to define the datatype of each column
  attr(SummaryData$Variables, DATA_TYPE) <- DATA_TYPES$STRING
  attr(SummaryData$`MD_outliers (N)`, DATA_TYPE) <- DATA_TYPES$INTEGER
  attr(SummaryData$`MD_outliers (%)`, DATA_TYPE) <- DATA_TYPES$FLOAT
  attr(SummaryData$N, DATA_TYPE) <- DATA_TYPES$INTEGER
  attr(SummaryData$observational_units_removed, DATA_TYPE) <- DATA_TYPES$INTEGER
  attr(SummaryData$mahalanobis_threshold, DATA_TYPE) <- DATA_TYPES$FLOAT


  return(list(FlaggedStudyData = FlaggedStudyData,
              SummaryTable = SummaryTable,
              SummaryData = SummaryData,
              SummaryPlot = p1
         ))
}

#' Utility function to create a plot similar to the margins plots for ordinal variables
#'
#' This function is still under development. It uses the `ordinal` package to
#' compute ordered regression models.
#'
#' @param resp_vars  [variable] the name of the ordinal measurement variable
#' @param group_vars [variable] the name of the observer, device or
#'                              reader variable
#' @param co_vars [variable list] a vector of covariables, e.g. age and sex for
#'                              adjustment
#' @param min_obs_in_subgroup [integer] from=0. This optional argument specifies
#'                       the minimum number of observations that is required to
#'                       include a subgroup (level) of the `group_var` in the
#'                       analysis.
#' @param min_subgroups [integer] from=3. The model provided by the `ordinal`
#'                       package requires at least three different subgroups
#'                       (levels) of the `group_var`. Users might want to
#'                       increase this threshold to obtain results only for
#'                       variables with a sufficient number of `group_var`
#'                       levels (observers, devices, etc.).
#' @param ds1 [data.frame] the data frame that contains the measurements, after
#'                       replacing missing value codes by `NA`, excluding
#'                       inadmissible values and transforming categorical
#'                       variables to factors.
#' @param label_col [variable attribute] the name of the column in the metadata
#'                       with labels of variables
#' @param adjusted_hint [character] hint, if adjusted for `co_vars`
#' @param title [character] title for the plot
#' @param sort_group_var_levels [logical] Should the levels of the grouping
#'                        variable be sorted descending by the number of
#'                        observations (in the figure)?
#'
#' @return A table and a matching plot.
#'
#' @importFrom ggplot2 ggplot geom_point geom_pointrange theme_minimal theme labs xlab ylab ggtitle scale_colour_manual aes geom_vline
#' @importFrom stats qnorm p.adjust
#'
#' @noRd
util_margins_ord <- function(resp_vars = NULL, group_vars = NULL, co_vars = NULL,
                             min_obs_in_subgroup = 5, min_subgroups = 5,
                             ds1, label_col,
                             adjusted_hint = "",
                             title = "",
                             sort_group_var_levels =
                               getOption("dataquieR.acc_margins_sort",
                                         dataquieR.acc_margins_sort_default)) {
  # preps and checks -----------------------------------------------------------
  util_ensure_suggested("ordinal")
  # prep_prepare_dataframes(.replace_hard_limits = TRUE,
  #                         .apply_factor_metadata = TRUE)
  #
  # util_correct_variable_use("resp_vars",
  #                           allow_all_obs_na = FALSE,
  #                           need_scale = "ordinal")
  # util_correct_variable_use("group_vars",
  #                           allow_all_obs_na = FALSE,
  #                           need_type = "!float",
  #                           need_scale = "nominal | ordinal")
  # util_correct_variable_use("co_vars",
  # overwrite = TRUE,
  # remove_not_found = TRUE,
  #                           allow_na = TRUE,
  #                           allow_more_than_one = TRUE,
  #                           allow_null = TRUE)
  # co_vars <- co_vars[which(!util_empty(co_vars))]
#
#   util_expect_scalar(min_obs_in_subgroup,
#                      check_type = util_is_numeric_in(min = 0,
#                                                      whole_num = TRUE,
#                                                      finite = TRUE))
  util_expect_scalar(min_subgroups,
                     check_type = util_is_numeric_in(min = 3,
  # 'ordinal' package requires at least three levels for the random-effect terms
  # (otherwise, they could be included as fixed effects?)
  # https://stat.ethz.ch/pipermail/r-help/2014-June/375429.html
                                                     whole_num = TRUE,
                                                     finite = TRUE))

  # ds1 <- ds1[, c(resp_vars, group_vars, co_vars), drop = FALSE]
  # ds1 <- ds1[complete.cases(ds1), , drop = FALSE]

  # ensure that included levels of the grouping variable have the specified
  # minimum number of observations
  # check_df <- util_table_of_vct(ds1[[group_vars]])
  # critical_levels <- levels(check_df$Var1)[check_df$Freq <
  #                                            min_obs_in_subgroup]
  # if (length(critical_levels) > 0) {
  #   util_message("Levels %s were excluded due to less than %d observations.",
  #                paste0(c(vapply(head(critical_levels, 10), dQuote, ""),
  #                         if (length(critical_levels) <= 10)
  #                           character(0) else "..."),
  #                       collapse = ", "),
  #                min_obs_in_subgroup,
  #                applicability_problem = FALSE)
  #   # exclude levels with too few observations
  #   ds1 <- ds1[!(ds1[[group_vars]] %in% critical_levels), ]
  #   levels(ds1[[group_vars]])[
  #     which(levels(ds1[[group_vars]]) %in% critical_levels)] <- NA
  # }
  # ds1 <- ds1[complete.cases(ds1), , drop = FALSE]
  # if (nrow(ds1) == 0) {
  #   util_error("No data left after data preparation.")
  # }

  # TODO: work in progress!
  if (length(co_vars) > 0 && all(!util_empty(co_vars))) {
    util_error(paste("Covariate argument for ordinal regression",
                       "not yet supported."))
  }

  # ensure that the grouping variable has the specified minimum number of levels
  check_df <- util_table_of_vct(ds1[[group_vars]])
  if (length(check_df[, 1]) < min_subgroups) {
    util_error("%d < %d levels in %s.",
               length(check_df[, 1]),
               min_subgroups,
               dQuote(group_vars),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  # ensure that the response variable is not constant
  var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])
  if (var_prop$NDistinct < 2) {
    util_error("The response variable is constant after data preparation.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  # modelling ------------------------------------------------------------------
  # prepare model formula
  if (length(co_vars) == 0) {
    co_vars <- "1"
  } else {
    co_vars <- util_bQuote(co_vars)
  }
  fmla <- as.formula(paste0(
    util_bQuote(resp_vars), " ~ ",
    paste0(c(co_vars, paste0("(1|", util_bQuote(group_vars), ")")),
           collapse = "+")))

  fit1 <- try(ordinal::clmm(formula = fmla, data = ds1,
                            Hess = TRUE, nAGQ = 10))
  if (inherits(fit1, "try-error")) {
    fit1 <- try(ordinal::clmm(formula = fmla, data = ds1,
                              Hess = TRUE))
  }
  if (inherits(fit1, "try-error")) {
    util_error("Model could not be fit automatically.")
  }

  if (sort_group_var_levels) {
    # order levels of the grouping variable by number of observations
    tbl_gr <- check_df %>%
      dplyr::arrange(dplyr::desc(.data$Var1)) %>%
      dplyr::arrange(.data$Freq)
    ds1[[group_vars]] <- factor(ds1[[group_vars]],
                                levels = as.character(tbl_gr$Var1))
  } else { # original order in plot, but from top of the y-axis to the bottom
    ds1[[group_vars]] <- factor(ds1[[group_vars]],
                                levels = rev(levels(ds1[[group_vars]])))
  }

  # calculation of random effects confidence intervals (using conditional modes)
  # taken from the vignette:
  # 'A Tutorial on fitting Cumulative Link Mixed Models with clmm2 from
  # the ordinal Package'
  # https://cran.r-project.org/web/packages/ordinal/vignettes/clmm2_tutorial.pdf
  plot_data <- data.frame(
    group = factor(levels(ds1[[group_vars]]), levels = levels(ds1[[group_vars]])),
    ran_ef = fit1$ranef,
    lower = fit1$ranef - qnorm(0.975) * sqrt(fit1$condVar),
    upper = fit1$ranef + qnorm(0.975) * sqrt(fit1$condVar))
  colnames(check_df) <- c("group", "sample_size")
  plot_data <- merge(plot_data, check_df)
  if (sort_group_var_levels) {
    plot_data <- plot_data %>%
      dplyr::arrange(dplyr::desc(.data[["sample_size"]]))
  }
  # Wald tests (here z-tests, two-sided, H0: mu = 0) to detect outlying levels
  # of the grouping variable
  plot_data$z <- fit1$ranef/sqrt(fit1$condVar)
  plot_data$p <- (1 - pnorm(abs(plot_data$z), 0, 1)) * 2
  # adjust p values for multiple testing
  plot_data$GRADING <- as.character(as.numeric(
    p.adjust(plot_data$p, method = "fdr") < 0.05))

  # figure ---------------------------------------------------------------------
  warn_code <- c("1" = "#B2182B", "0" = "#2166AC")

  p1 <- util_create_lean_ggplot(
    ggplot(plot_data,
           aes(x = .data[["ran_ef"]],
               y = .data[["group"]],
               col = .data[["GRADING"]])) +
      # +/- 1 SD of random effects (group_var) around 0
      geom_vline(xintercept = -sqrt(as.numeric(ordinal::VarCorr(fit1)[[group_vars]])),
                 linetype = 2,
                 col = "gray60") +
      geom_vline(xintercept = sqrt(as.numeric(ordinal::VarCorr(fit1)[[group_vars]])),
                 linetype = 2,
                 col = "gray60") +
      util_geom_pointrange_robust(aes(xmin = .data[["lower"]],
                          xmax = .data[["upper"]])) +
      geom_point() +
      scale_color_manual(values = warn_code,
                         guide = "none") +
      xlab("conditional mode") +
      ylab("") +
      theme_minimal() +
      ggtitle(label = title,
              subtitle = adjusted_hint) +
      labs(caption = "Dashed lines mark the interval 0 \u00B1 SD."),
    plot_data = plot_data,
    fit1 = fit1,
    group_vars = group_vars,
    warn_code = warn_code,
    title = title,
    adjusted_hint = adjusted_hint)

  # alternative code, for testing and further development ----------------------
  # # the emmeans approach works if we include the grouping variable as
  # # fixed effect:
  # fmla <- as.formula(paste0(
  #   util_bQuote(resp_vars), " ~ ",
  #   paste0(c(co_vars, util_bQuote(group_vars)),
  #          collapse = "+")))
  # fit1 <- try(ordinal::clm(formula = fmla, data = ds1,
  #                           Hess = TRUE, nAGQ = 10))
  # plot(emmeans::emmeans(fit1, group_vars))
  #
  # # or if we include them as scale effect:
  # # https://cran.r-project.org/web/packages/emmeans/vignettes/sophisticated.html#ordinal
  # fmla <- as.formula(paste0(
  #   util_bQuote(resp_vars), " ~ ",
  #   paste0(co_vars, collapse = "+")))
  #
  # fit1 <- try(ordinal::clm(
  #   formula = fmla,
  #   scale = as.formula(paste0("~", util_bQuote(group_vars))),
  #   data = ds1,
  #   Hess = TRUE, nAGQ = 10))
  #
  # summary(emmeans::ref_grid(fit1, mode = "scale"), type = "response")
  # # the first entry gives the reference level, the others are being compared
  # # to this level
  # # see https://cran.r-project.org/web/packages/emmeans/vignettes/sophisticated.html#ordinal
  # plot(summary(emmeans::ref_grid(fit1, mode = "scale"), type = "prob"))

  plot_data <- dplyr::rename(plot_data, c("margins" = "ran_ef", "LCL" = "lower",
                                          "UCL" = "upper"))
  plot_data$z <- NULL
  plot_data$p <- NULL

  return(list("plot" = p1, "plot_data" = plot_data))
}

#' Check referred variables
#'
#' This function operates in the environment of its caller
#' (using [eval.parent], similar to [Function like C-Preprocessor-Macros](
#' https://www.programiz.com/c-programming/c-preprocessor-macros#example-define)
#' ).
#' Different from the other utility function that work
#' in the caller's environment ([prep_prepare_dataframes]), It has no side
#' effects except that the argument
#' of the calling function specified in `arg_name` is normalized (set to its
#' default or a general default if missing, variable names being all white
#' space replaced by NAs).
#' It expects two objects in the caller's environment: `ds1` and `meta_data`.
#' `meta_data` is the metadata data frame and `ds1` is produced by a preceding
#' call of [prep_prepare_dataframes] using `meta_data` and `study_data`.
#' So this function can only be used after calling the function
#' [prep_prepare_dataframes]
#'
#' `util_correct_variable_use` and `util_correct_variable_use2` differ only in
#' the default of the argument `role`.
#'
#' `util_correct_variable_use` and `util_correct_variable_use2` put strong
#' effort on producing compressible
#' error messages to the caller's caller (who is typically an end user of
#' a `dataquieR` function).
#'
#' The function ensures, that a specified argument of its caller that refers
#' variable names
#' (one or more as character vector) matches some expectations.
#'
#' This function accesses the caller's environment!
#'
#' @param arg_name [character] Name of a function argument of the caller
#'                             of `util_correct_variable_use`
#' @param allow_na [logical] default = FALSE. allow NAs in the variable names
#'                                            argument given in `arg_name`
#' @param allow_more_than_one [logical] default = FALSE. allow more than one
#'                                                  variable names in `arg_name`
#' @param allow_null [logical] default = FALSE. allow an empty variable name
#'                                             vector in the argument `arg_name`
#' @param allow_all_obs_na [logical] default = TRUE. check observations for not
#'                                                   being all `NA`
#' @param allow_any_obs_na [logical] default = TRUE. check observations for
#'                                               being complete without any `NA`
#' @param min_distinct_values [integer] Minimum number of distinct observed
#'                              values of a study variable
#' @param need_type [character] if not `NA`, variables must be of data type
#'                                       `need_type` according to the metadata,
#'                                       can be a pipe (`|`) separated list of
#'                                       allowed data types. Use `!` to exclude
#'                                       a type. See [DATA_TYPES] for the
#'                                       predefined variable types of the
#'                                       `dataquieR` concept.
#' @param need_scale [character] if not `NA`, variables must be of scale level
#'                                       `need_scale` according to the metadata,
#'                                       can be a pipe (`|`) separated list of
#'                                     allowed scale levels. Use `!` to exclude
#'                                       a level. See [SCALE_LEVELS] for the
#'                                       predefined scale levels of the
#'                                       `dataquieR` concept.
#' @param need_role [character] see `need_scale`, but looks for `VARIABLE_ROLE`
#' @param need_computed_role [character] see `need_scale`, but looks for
#'                                       `COMPUTED_VARIABLE_ROLE`
#' @param role [character] variable-argument role. Set different defaults for
#'                         all `allow`-arguments and `need_type` of
#'                         this `util_correct_variable_use.`. If given, it
#'                         defines the intended use
#'                         of the verified argument. For typical arguments and
#'                         typical use cases, roles
#'                         are predefined in [.variable_arg_roles].
#'                         The role's defaults can be overwritten by the
#'                         arguments. If `role` is "" (default),
#'                         the standards are `allow_na = FALSE`,
#'                         `allow_more_than_one = FALSE`, `allow_null = FALSE`,
#'                         `allow_all_obs_na = TRUE`, `allow_any_obs_na = TRUE`,
#'                         and `need_type = NA`.
#'                         Use `util_correct_variable_use2` for using the
#'                         `arg_name` as default for `role`.
#'                         See [.variable_arg_roles] for currently available
#'                         variable-argument roles.
#' @param overwrite [logical] overwrite vector of variable names
#'                         to match the labels given in `label_col`.
#' @param do_not_stop [logical] do not throw an error, if one of the variables
#'                         violates `allow_all_obs_na`, `allow_any_obs_na` or
#'                         `min_distinct_values`. Instead, the variable will be
#'                         removed from `arg_name` in the parent environment
#'                         with a warning. This is helpful for functions which
#'                         work with multiple variables.
#' @param remove_not_found [logical]
#'                          if variables are missing from the sutdy data, ignore
#'                          them
#'
#' @seealso [.variable_arg_roles]
#' @importFrom stats na.omit
#'
#' @family robustness_functions
#' @concept robustness
#' @noRd
util_correct_variable_use <- function(arg_name,
                                      allow_na,
                                      allow_more_than_one,
                                      allow_null,
                                      allow_all_obs_na,
                                      allow_any_obs_na,
                                      min_distinct_values,
                                      need_type,
                                      need_scale,
                                      need_role,
                                      need_computed_role,
                                      role = "",
                                      overwrite = TRUE,
                                      do_not_stop = FALSE,
                                      remove_not_found = FALSE) {
  re_check <- TRUE
  while (re_check) {
    re_check <- FALSE
    # interpret the arg_name argument
    try({
      arg_name <- as.character(substitute(arg_name))
    }, # if called with a symbol, get that symbol as a character,
    silent = TRUE # if called with a character, do nothing
    )
    if (!all(is.character(arg_name))) { # nocov start
      # if we did not get a character up to here, there is an error in
      # calling this util_correct_variable_use
      # this should never happen anyway, so cannot test this
      util_error(c(
        "argument arg_name must be either one character or one symbol,",
        "wrong use of util_correct_variable_use?"))
    } # nocov end
    if (length(arg_name) != 1) {
      # if we have 0 or more than one argument name, this is also an
      # error in calling util_correct_variable_use
      util_error(c(
        "argument arg_name must be of length 1,",
        "wrong use of util_correct_variable_use?"))
    }

    # verify the "roles" argument
    if (role %in% .variable_arg_roles$name) {
      # if `role` is a known variable-argument role
      if (missing(allow_na)) { # if missing the allow_na argument
        allow_na <- subset(.variable_arg_roles, get("name") == role,
                           "allow_na", drop = TRUE)
      } # fetch it from the role's defaults
      if (missing(allow_more_than_one)) {
        # if missing the allow_more_than_one argument
        allow_more_than_one <- subset(.variable_arg_roles, get("name") == role,
                                      "allow_more_than_one", drop = TRUE)
      } # fetch it from the role's defaults
      if (missing(allow_null)) { # if missing the allow_null argument
        allow_null <- subset(.variable_arg_roles, get("name") == role,
                             "allow_null", drop = TRUE)
      } # fetch it from the role's defaults
      if (missing(allow_all_obs_na)) { # if missing the allow_all_obs_na argument
        allow_all_obs_na <- subset(.variable_arg_roles, get("name") == role,
                                   "allow_all_obs_na", drop = TRUE)
      } # fetch it from the role's defaults
      if (missing(allow_any_obs_na)) { # if missing the allow_any_obs_na argument
        allow_any_obs_na <- subset(.variable_arg_roles, get("name") == role,
                                   "allow_any_obs_na", drop = TRUE)
      } # fetch it from the role's defaults
      if (missing(min_distinct_values)) {
        # if missing the min_distinct_values argument
        min_distinct_values <- subset(.variable_arg_roles, get("name") == role,
                                      "min_distinct_values", drop = TRUE)
      } # fetch it from the role's defaults
      if (missing(need_type)) { # if missing the need_type argument
        need_type <- subset(.variable_arg_roles, get("name") == role,
                            "need_type", drop = TRUE)
      } # fetch it from the role's defaults
      if (missing(need_scale)) { # if missing the need_scale argument
        need_scale <- subset(.variable_arg_roles, get("name") == role,
                             "need_scale", drop = TRUE)
      } # fetch it from the role's defaults
      if (missing(need_role)) { # if missing the need_role argument
        need_role <- subset(.variable_arg_roles, get("name") == role,
                             "need_role", drop = TRUE)
      } # fetch it from the role's defaults
      if (missing(need_computed_role)) { # if missing the need_computed_role argument
        need_computed_role <- subset(.variable_arg_roles, get("name") == role,
                             "need_computed_role", drop = TRUE)
      } # fetch it from the role's defaults
    } else { # no known variable-argument in argument `role`
      if (role != "") { # if role is not empty
        util_error(
          c("Unknown variable-argument role: %s for argument %s,",
            "wrong use of util_correct_variable_use?"),
          role, arg_name
        ) # this is again a wrong use of this function.
      }
      if (missing(allow_na)) {
        # set the default for allow_na if it has not been given
        allow_na <- FALSE
      }
      if (missing(allow_more_than_one)) {
        # set the default for allow_more_than_one if it has not been given
        allow_more_than_one <- FALSE
      }
      if (missing(allow_null)) {
        # set the default for allow_null if it has not been given
        allow_null <- FALSE
      }
      if (missing(allow_all_obs_na)) {
        # set the default for allow_all_obs_na if it has not been given
        allow_all_obs_na <- TRUE
      }
      if (missing(allow_any_obs_na)) {
        # set the default for allow_any_obs_na if it has not been given
        allow_any_obs_na <- TRUE
      }
      if (missing(min_distinct_values)) {
        # set the default for min_distinct_values if it has not been given
        min_distinct_values <- 0
      }
      if (missing(need_type)) {
        # set the default for need_type if it has not been given
        need_type <- NA
      }
      if (missing(need_scale)) {
        # set the default for need_scale if it has not been given
        need_scale <- NA
      }
      if (missing(need_role)) {
        # set the default for need_role if it has not been given
        need_role <- NA
      }
      if (missing(need_computed_role)) {
        # set the default for need_computed_role if it has not been given
        need_computed_role <- NA
      }
    }

    p <- parent.frame(1) # access the caller's environment
    if (!exists(arg_name, envir = p)) {
      # check, if the function argument exists in the calling function.
      util_error(
        c("Unknown function argument %s checked,",
          "wrong use of util_correct_variable_use?"),
        arg_name)
    }

    # if we have a label col, use it
    if (exists("label_col", envir = p)) {
      # calling function has a label_col argument
      label_col <- try(get("label_col", envir = p), silent = TRUE)
      # try to get its value
      if (inherits(label_col, "try-error")) {
        # no value avail, i.e. no default and not given
        label_col <- VAR_NAMES # use VAR_NAMES as default
      }
    } else {
      label_col <- VAR_NAMES
      # calling function does not support label_col, so use VAR_NAMES as default
    }

    missing_in_parent <- eval.parent(call("missing", as.symbol(arg_name)))

    variable <- try(get(arg_name, envir = p), silent = TRUE)
    # try to get the value of the callers argument called `arg_name`.
    if (inherits(variable, "try-error")) {
      if (missing_in_parent) {
        util_warning(
          c("Missing argument %s without default value. Setting to NULL. As",
            "a dataquieR developer, please add a default value for %s to",
            "remove this warning."),
          dQuote(arg_name), dQuote(arg_name),
          applicability_problem = TRUE)
      } else {
        util_warning(
          c("Could not get value of argument %s for unexpected reasons. Setting",
            "to NULL."), arg_name, applicability_problem = TRUE)
        util_warning(variable, applicability_problem = TRUE)
      }
      variable <- NULL
    }

    if (!exists("ds1", envir = p)) {
      # if the calling function does not have a ds1, this is a wrong use of
      # util_correct_variable_use
      util_error(c(
        "Did not find merged study data and metadata ds1.",
        "Wrong use of util_correct_variable_use?"))
    }

    if (!exists("meta_data", envir = p)) {
      # if the calling function does not have a meta_data, this is a wrong use of
      # util_correct_variable_use
      util_error(
        "Did not find metadata. Wrong use of util_correct_variable_use?")
    }

    ds1 <- try(get("ds1", envir = p), silent = TRUE)
    # try to fetch the ds1 object from the callers environment
    if (!is.data.frame(ds1)) { # it this is not a data frame, this is again a
      # wrong use of util_correct_variable_use
      util_error(
        c("ds1 does not provide merged study data and metadata.",
          "Wrong use of util_correct_variable_use?"))
    }

    meta_data <- try(get("meta_data", envir = p), silent = TRUE)
    # try to fetch the meta_data object from the callers environment
    if (!is.data.frame(meta_data)) {
      # it this is not a data frame, this is again a wrong use of
      # util_correct_variable_use
      util_error(
        c("meta_data does not provide a metadata data frame.",
          "Wrong use of util_correct_variable_use?"))
    }

    if (!allow_null && is.null(variable)) {
      # if we need the argument in arg_name, but the user provided NULL
      util_error("Argument %s is NULL", arg_name,
                 applicability_problem = TRUE) # this is an error
    }

    if (allow_more_than_one) {
      # if we allow more than one variable in the argument arg_name
      if (!allow_null && length(variable) == 0) {
        # and we do need at least one variable name here, but the user did not
        # provide any
        util_error("Need at least one element in argument %s, got 0", arg_name,
                   applicability_problem = TRUE)
        # this is an error
      }
    } else { # if we expect one value in the argument arg_name at most
      if ((!allow_null && length(variable) != 1) ||
          (allow_null && length(variable) > 1)) {
        # but the user gave more than one or none although
        # allow_null prohibits this
        util_error("Need exactly one element in argument %s, got %d: [%s]",
                   arg_name, length(variable), paste0(variable, collapse = ", "),
                   applicability_problem = TRUE)
        # this is an error
      }
    }
    if (length(variable) > 0) { # if we have at least one variable name
      if (!all(is.character(variable) |
               is.na(variable))) { # and one name is not a character string
        util_error("Need character variable names in argument %s", arg_name,
                   applicability_problem = TRUE)
        # this is (at least up to now, may allow symbols here similar
        # to subset or with) an error
      }
      variable[trimws(variable) == ""] <- NA # replace all variable names that are
      # empty characters by NA.
    }
    assign(arg_name, variable, envir = p) # re-assign the possibly modified
    # argument value in the caller's environment
    if (!allow_na && any(is.na(variable))) {
      # if we have NAs in the list of variable names in `arg_name`
      # and this is not allowed
      util_error("Missing entries not allowed in argument %s", arg_name,
                 applicability_problem = TRUE)
      # then this is an error.
    }

    # Check whether variables can be found in the data (using all possible label columns).
    possible_vars <- unique(c(meta_data[[VAR_NAMES]], # TODO this corresponds with TODO POSSIBLE_VARS in prep_prepare_dataframes
                              meta_data[[LABEL]],
                              meta_data[[LONG_LABEL]],
                              meta_data[[label_col]]))
    if (!all(na.omit(variable) %in% possible_vars)) {
      # if we have variable names in the caller's variable argument `arg_name`
      # missing in the data
      non_matching_vars <-
        na.omit(variable)[!(na.omit(variable) %in% possible_vars)] # find them

      only_in_sd <-
        try({
          n <- colnames(get("study_data", envir = p))
          n[na.omit(match(tolower(non_matching_vars), tolower(n)))]
        }, silent = TRUE)

      if (util_is_try_error(only_in_sd)) {
        only_in_sd <- character(0)
      }

      if (length(only_in_sd) > 0) {
        hint_maybe_miss_in_meta <- sprintf(
        paste("\nHINT: I've found the following variables in %s not covered",
          "by %s: %s, so, maybe, your metadata is incomplete?"),
          "study data",
          "metadata",
          util_pretty_vector_string(only_in_sd)
        )
      } else {
        hint_maybe_miss_in_meta <- ""
      }

      fuzzy_match <- vapply( # and try to guess what the user wanted to put there
        non_matching_vars,
        function(v) {
          unique(possible_vars)[which.min(adist(
            trimws(v),
            trimws(unique(possible_vars)),
            ignore.case = TRUE,
            fixed = TRUE
          ))]
        }, "")
      if (remove_not_found) {
        cnd_f <- util_message
      } else {
        cnd_f <- util_error
      }
      cnd_f(
        "Variable '%s' (%s) not found in metadata. Did you mean '%s'?%s",
        # then emit an informative error message.
        paste0(non_matching_vars, collapse = ", "),
        arg_name,
        paste0(fuzzy_match, collapse = ", "),
        hint_maybe_miss_in_meta,
        applicability_problem = TRUE,
        intrinsic_applicability_problem = TRUE
      )
      variable <- setdiff(non_matching_vars, non_matching_vars)
      re_check <- TRUE
      if (overwrite) {
        # Pass the 'variable' vector to the environment to use it within the calling function.
        assign(arg_name, variable, envir = parent.frame())
      }
    } else {
      # If users mix VAR_NAMES and LABELs (and maybe also LONG_LABELs), we need to
      # map the variable names to the column names of ds1.
      if (!all(na.omit(variable) %in% colnames(ds1))) {
        non_matching_ind <- intersect(which(!(variable %in% colnames(ds1))),
                                      which(!is.na(variable)))
        non_matching_vars <- variable[non_matching_ind]
        other_col <- setdiff(c(VAR_NAMES, LABEL, LONG_LABEL), label_col)
        other_col <- intersect(other_col, colnames(meta_data))
        map_res <- lapply(setNames(nm = other_col), function(oc) {
          unname(
            util_map_labels(
              non_matching_vars,
              warn_ambiguous = TRUE,
              meta_data = meta_data,
              to = label_col,
              from = oc,
              ifnotfound = NA_character_
            )
          )
        })
        map_res <- as.data.frame(map_res)
        # check whether the user-specified variable could not be found in any of the other columns:
        map_res_NA <- vapply(
          1:nrow(map_res),
          FUN.VALUE = logical(1),
          FUN = function(i) {
            all(is.na(map_res[i, ]))
          }
        )
        if (any(map_res_NA)) { # may be dead code?
          util_error("Variable '%s' (%s) not found in the metadata.",
                     paste0(non_matching_vars[map_res_NA],
                            collapse = ", "),
                     arg_name,
                     applicability_problem = TRUE)
        }
        # check whether the user-specified variable was found in different label columns and maps to different variables (= ambiguity)
        map_res_ambiguous <- vapply(
          1:nrow(map_res),
          FUN.VALUE = logical(1),
          FUN = function(i) {
            length(unique(na.omit(as.character(map_res[i, ])))) > 1
          }
        )
        if (any(map_res_ambiguous)) {
          util_warning("Variable '%s' (%s) is ambiguous in the metadata.",
                       paste0(non_matching_vars[map_res_ambiguous],
                              collapse = ", "),
                       arg_name,
                       applicability_problem = TRUE)
        }

        variable[non_matching_ind] <- vapply(
          1:nrow(map_res),
          FUN.VALUE = character(1),
          FUN = function(i) {
            unique(na.omit(as.character(map_res[i, ])))[1]
            # We select here the first element, such that in case of ambiguities
            # VAR_NAME is preferred over LABEL and LONG_LABEL, and LABEL over LONG_LABEL.
            # If there are no ambiguities, this works as well.
          }
        )
        # The 'variable' vector contains now all variable names according to label_col.
        if (overwrite) {
          # Pass the 'variable' vector to the environment to use it within the calling function.
          assign(arg_name, variable, envir = parent.frame())
        }
      }
    }

    record_problem <- ifelse(do_not_stop, util_message, util_error)
    applicability_problem <- ifelse(do_not_stop, FALSE, TRUE)
    intrinsic_applicability_problem <- ifelse(do_not_stop, FALSE, TRUE)
    data_problem_container <- new.env(parent = emptyenv())
    for (v in variable) { # now, check all variable values from the data
      if (!allow_all_obs_na && all(is.na(ds1[[v]]))) {
        if (is.na(v)) {
          # if all observations are NA and this is prohibited
          record_problem("There are NAs in the variable names passed in %s",
                         arg_name,
                         applicability_problem = applicability_problem,
                         intrinsic_applicability_problem =
                           intrinsic_applicability_problem)
        } else{
          # if all observations are NA and this is prohibited
          record_problem("Variable '%s' (%s) has only NA observations",
                         na.omit(v)[na.omit(v) %in% colnames(ds1)], arg_name,
                         applicability_problem = applicability_problem,
                         intrinsic_applicability_problem =
                           intrinsic_applicability_problem)
        }
        data_problem_container[[v]] <- TRUE
      }
      # names(empty_containter) gibt die geflaggten Variablen
      if (!allow_any_obs_na && any(is.na(ds1[[v]]))) {
        # if no NAs are allowed at all, but we have some
        record_problem("Variable '%s' (%s) has NA observations, which is not allowed",
                       na.omit(v)[na.omit(v) %in% colnames(ds1)], arg_name,
                       applicability_problem = applicability_problem,
                       intrinsic_applicability_problem =
                         intrinsic_applicability_problem)
        data_problem_container[[v]] <- TRUE
      }

      if (min_distinct_values > 0) {
        uniq_rv <- unique(ds1[[v]])
        uniq_rv <- uniq_rv[!util_empty(uniq_rv)]
        n_distinct_values <- length(uniq_rv)
        if (n_distinct_values < min_distinct_values) {
          # if we have fewer distinct values than required
          record_problem(c("Variable '%s' (%s) has fewer distinct values than required",
                           "for the argument %s of %s"),
                         na.omit(v)[na.omit(v) %in% colnames(ds1)], arg_name,
                         sQuote(arg_name),
                         sQuote(rlang::caller_call()[[1]]),
                         applicability_problem = applicability_problem,
                         intrinsic_applicability_problem =
                           intrinsic_applicability_problem)
          data_problem_container[[v]] <- TRUE
        }
      }
    }

    if (length(names(data_problem_container)) > 0) {
      variable <- setdiff(variable, names(data_problem_container))
      if (length(variable) == 0) {
        util_error(
          "In %s, none of the specified variables matches the requirements.",
          dQuote(arg_name),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
      } else {
        util_warning(
          c("In %s, variables %s were excluded."),
          dQuote(arg_name),
          paste(dQuote(names(data_problem_container)), collapse = ", "),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
        assign(arg_name, variable, envir = p)
      }
    }

    data_type_problem_container <- new.env(parent = emptyenv())

    # data type given in metadata? (similar to util_get_meta_from_var ...)
    if (DATA_TYPE %in% colnames(meta_data)) {
      types <- meta_data[meta_data[[label_col]] %in% variable, DATA_TYPE]
      # try to fetch the types from the metadata.
      names(types) <- meta_data[meta_data[[label_col]] %in% variable, label_col]
      types <- types[!is.na(types) & !is.na(names(types))]
    } else {
      types <- NULL
      names(types) <- NULL
    }

    data_scale_problem_container <- new.env(parent = emptyenv())

    # data type given in metadata? (similar to util_get_meta_from_var ...)
    if (SCALE_LEVEL %in% colnames(meta_data)) {
      scales <- meta_data[meta_data[[label_col]] %in% variable, SCALE_LEVEL]
      # try to fetch the scales from the metadata.
      names(scales) <- meta_data[meta_data[[label_col]] %in% variable, label_col]
      scales <- scales[!is.na(scales) & !is.na(names(scales))]
    } else {
      scales <- NULL
      names(scales) <- NULL
    }

    data_role_problem_container <- new.env(parent = emptyenv())

    # variable role given in metadata? (similar to util_get_meta_from_var ...)
    if (VARIABLE_ROLE %in% colnames(meta_data)) {
      roles <- meta_data[meta_data[[label_col]] %in% variable, VARIABLE_ROLE]
      # try to fetch the roles from the metadata.
      names(roles) <- meta_data[meta_data[[label_col]] %in% variable, label_col]
      roles <- roles[!is.na(roles) & !is.na(names(roles))]
    } else {
      roles <- NULL
      names(roles) <- NULL
    }

    data_computed_role_problem_container <- new.env(parent = emptyenv())

    # computed variable role given in metadata? (similar to util_get_meta_from_var ...)
    if (!COMPUTED_VARIABLE_ROLE %in% colnames(meta_data)) {
      meta_data[[COMPUTED_VARIABLE_ROLE]] <- rep("na", nrow(meta_data))
    }
    meta_data[[COMPUTED_VARIABLE_ROLE]][
      is.na(meta_data[[COMPUTED_VARIABLE_ROLE]])] <- "na"
    computed_roles <- meta_data[meta_data[[label_col]] %in% variable,
                                COMPUTED_VARIABLE_ROLE]
    # try to fetch the computed roles from the metadata.
    names(computed_roles) <-
      meta_data[meta_data[[label_col]] %in% variable, label_col]
    computed_roles <-
      computed_roles[!is.na(computed_roles) & !is.na(names(computed_roles))]

    if (!all(is.na(need_type)) && length(types) <
        length(unique(variable[!is.na(variable)]))) {
      util_warning(
        c("In %s, variables with types matching %s should be specified, but not",
          "all variables have a type assigned in the metadata.",
          "I have %d variables but only %d types."),
        dQuote(arg_name),
        dQuote(need_type),
        length(unique(variable[!is.na(variable)])),
        length(types),
        applicability_problem = TRUE
      )
    }

    for (..vname in names(types)) {
      # for all variables that have a data type assigned in the metadata
      type <- types[[..vname]]
      # fetch the type of the currently checked variable
      if (!is.null(type) && !is.na(type) && !is.na(need_type)) {
        # if we have a needed type and a variable type for the currently checked
        # variable,
        need_type. <- tolower(trimws(unlist(strsplit(need_type, "|",
                                                     fixed = TRUE),
                                            recursive = TRUE)))
        # parse the type (split at pipes |)
        not_type <-
          tolower(trimws(substring(need_type.[startsWith(need_type., "!")], 2)))
        # and parse the type for negation (e.g. *not* INTEGER)
        well_type <-
          tolower(trimws(need_type.[!startsWith(need_type., "!")]))
        # and parse the type for no negation resp. (e.g. INTEGER)

        if ((length(well_type) > 0) &&
            (!all(well_type %in%
                  tolower(trimws(names(DATA_TYPES)))))) {
          util_error(c(
            "Internal error:",
            "%s's %s contains invalid type names %s (allowed are %s).",
            "As a dataquieR developer, you should fix your call of %s."
          ),
          sQuote(arg_name),
          sQuote("need_type"),
          paste(dQuote(
            well_type[!(well_type %in%
                          tolower(trimws(names(DATA_TYPES))))]
          ), collapse = ", "),
          paste(dQuote(tolower(trimws(names(DATA_TYPES)))), collapse = ", "),
          sQuote(rlang::caller_call()[[1]])
          )
        }

        if ((length(not_type) > 0) &&
            (!all(not_type %in%
                  tolower(trimws(names(DATA_TYPES)))))) {
          util_error(c(
            "Internal error:",
            "%s's %s contains invalid !type names %s (allowed are %s).",
            "As a dataquieR developer, you should fix your call of %s."
          ),
          sQuote(arg_name),
          sQuote("need_type"),
          paste(dQuote(
            not_type[!(not_type %in%
                         tolower(trimws(names(DATA_TYPES))))]
          ), collapse = ", "),
          paste(dQuote(tolower(trimws(names(DATA_TYPES)))), collapse = ", "),
          sQuote(rlang::caller_call()[[1]])
          )
        }

        if ((length(well_type) > 0) && (!tolower(trimws(type)) %in% well_type)) {
          # now check for the allowed types
          record_problem(
            "Argument %s: Variable '%s' (%s) does not have an allowed type (%s)",
            dQuote(arg_name), ..vname, tolower(trimws(type)), need_type,
            applicability_problem = TRUE,
            intrinsic_applicability_problem = TRUE)
          data_type_problem_container[[..vname]] <- TRUE
        }
        if ((length(not_type) > 0) && (tolower(trimws(type)) %in% not_type)) {
          # and for the disallowed types
          record_problem(
            "Argument %s: Variable '%s' (%s) has a disallowed type (%s)",
            dQuote(arg_name), ..vname, tolower(trimws(type)),
            util_pretty_vector_string(not_type),
            applicability_problem = TRUE,
            intrinsic_applicability_problem = TRUE)
          data_type_problem_container[[..vname]] <- TRUE
        }
        # Check, if variable is really of declared data type.
        sd_type <- prep_datatype_from_data(..vname, ds1)
        if (!is.na(sd_type) && sd_type != type) {
          if (sd_type != DATA_TYPES$INTEGER || type != DATA_TYPES$FLOAT) {
            util_message(
              c("Argument %s: Variable '%s' (%s) does not have matching",
                "data type in the study data (%s)"),
              dQuote(arg_name), ..vname, tolower(trimws(type)),
              prep_dq_data_type_of(ds1[, ..vname, TRUE]),
              applicability_problem = TRUE)
          }
        }
      } else { # nocov start
        # if the type is NA, it is not part of the types vector
        # so this is dead code, likely (see
        # "types <- types[!is.na(types) & !is.na(names(types))]" above).
        if (!is.na(need_type)) {
          util_warning(
            c("Argument %s: No data type found for '%s' in the",
              "metadata, cannot check type"), dQuote(arg_name), variable,
            applicability_problem = TRUE)
        }
      } # nocov end
    }
    if (length(names(data_type_problem_container)) > 0) {
      variable <- setdiff(variable, names(data_type_problem_container))
      if (length(variable) == 0) {
        util_error(
          "In %s, none of the specified variables matches the requirements.",
          dQuote(arg_name),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
      } else {
        util_warning(
          c("In %s, variables %s were excluded."),
          dQuote(arg_name),
          paste(dQuote(names(data_type_problem_container)), collapse = ", "),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
        assign(arg_name, variable, envir = p)
      }
    }
    # TODO EK: Should we support hierarchy of sclae levels here, i.e.,
    # util_correct_variable_use(resp_vars, need_scale = SCALE_LEVELS$ORDINAL)
    # would pass, if SBP_0 is ratio so even "better" than ordinal? But maybe,
    # we want to explicitly break the hierarchy for certain descriptors, e.g., run
    # a function only for ordinal variables but not for ratio ones.
    if (!all(is.na(need_scale)) && length(scales) <
        length(unique(variable[!is.na(variable)]))) {
      util_warning(
        c("In %s, variables with scale levels matching %s should be specified, but not",
          "all variables have a scale level assigned in the metadata.",
          "I have %d variables but only %d scale levels"),
        dQuote(arg_name),
        dQuote(need_scale),
        length(unique(variable[!is.na(variable)])),
        length(scales),
        applicability_problem = TRUE
      )
    }

    for (..vname in names(scales)) {
      # for all variables that have a scale level assigned in the metadata
      scale <- scales[[..vname]]
      # fetch the scale level of the currently checked variable
      if (!is.null(scale) && !is.na(scale) && !is.na(need_scale)) {
        # if we have a needed scale level and a variable scale leve for the
        # currently checked variable,
        need_scale. <- tolower(trimws(unlist(strsplit(need_scale, "|",
                                                      fixed = TRUE),
                                             recursive = TRUE)))
        # parse the scale level (split at pipes |)
        not_scale <-
          tolower(trimws(substring(need_scale.[startsWith(need_scale., "!")], 2)))
        # and parse the scale level for negation (e.g. *not* NOMINAL)
        well_scale <-
          tolower(trimws(need_scale.[!startsWith(need_scale., "!")]))
        # and parse the scale level for no negation resp. (e.g. NOMINAL)

        if ((length(well_scale) > 0) &&
            (!all(well_scale %in%
                  tolower(trimws(names(SCALE_LEVELS)))))) {
          util_error(c( # TODO EK This should not happen outside pipelines, then it should be a warning only.
            "Internal error:",
            "%s's %s contains invalid scale-level names %s (allowed are %s).",
            "As a dataquieR developer, you should fix your call of %s."
          ),
          sQuote(arg_name),
          sQuote("need_scale"),
          paste(dQuote(
            well_scale[!(well_scale %in%
                           tolower(trimws(names(SCALE_LEVELS))))]
          ), collapse = ", "),
          paste(dQuote(tolower(trimws(names(SCALE_LEVELS)))), collapse = ", "),
          sQuote(rlang::caller_call()[[1]])
          )
        }

        if ((length(not_scale) > 0) &&
            (!all(not_scale %in%
                  tolower(trimws(names(SCALE_LEVELS)))))) {
          util_error(c(
            "Internal error:",
            "%s's %s contains invalid !scale-level names %s (allowed are %s).",
            "As a dataquieR developer, you should fix your call of %s."
          ),
          sQuote(arg_name),
          sQuote("need_scale"),
          paste(dQuote(
            not_scale[!(not_scale %in%
                          tolower(trimws(names(SCALE_LEVELS))))]
          ), collapse = ", "),
          paste(dQuote(tolower(trimws(names(SCALE_LEVELS)))), collapse = ", "),
          sQuote(rlang::caller_call()[[1]])
          )
        }

        if ((length(well_scale) > 0) && (!tolower(trimws(scale)) %in% well_scale)) {
          # now check for the allowed scale levels
          record_problem(
            "Argument %s: Variable '%s' (%s) does not have an allowed scale level (%s)",
            dQuote(arg_name), ..vname, tolower(trimws(scale)), need_scale,
            applicability_problem = TRUE,
            intrinsic_applicability_problem = TRUE)
          data_scale_problem_container[[..vname]] <- TRUE
        }
        if ((length(not_scale) > 0) && (tolower(trimws(scale)) %in% not_scale)) {
          # and for the disallowed scale levels
          record_problem(
            "Argument %s: Variable '%s' (%s) has a disallowed scale level (%s)",
            dQuote(arg_name), ..vname, tolower(trimws(scale)),
            util_pretty_vector_string(not_scale),
            applicability_problem = TRUE,
            intrinsic_applicability_problem = TRUE)
          data_scale_problem_container[[..vname]] <- TRUE
        }
        # TODO: Implement this check also for scale levels
        # # Check, if variable is really of declared scale level.
        # sd_scale <- prep_scale_level_from_data(..vname, ds1)
        # if (!is.na(sd_scale) && sd_scale != scale) {
        #   util_message(
        #     c("Argument %s: Variable '%s' (%s) does not have matching",
        #       "scale level in the study data (%s)"),
        #     dQuote(arg_name), ..vname, tolower(trimws(scale)),
        #     prep_scale_level_from_data(ds1[, ..vname, TRUE]),
        #     applicability_problem = TRUE)
        # }
      } else { # nocov start
        # if the scale is NA, it is not part of the scales vector
        # so this is dead code, likely (see
        # "types <- types[!is.na(types) & !is.na(names(types))]" above).
        if (!is.na(need_scale)) {
          util_warning(
            c("Argument %s: No scale level found for '%s' in the",
              "metadata, cannot check scale level"), dQuote(arg_name), variable,
            applicability_problem = TRUE)
        }
      } # nocov end
    }
    if (length(names(data_scale_problem_container)) > 0) {
      variable <- setdiff(variable, names(data_scale_problem_container))
      if (length(variable) == 0) {
        util_error(
          "In %s, none of the specified variables matches the requirements.",
          dQuote(arg_name),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
      } else {
        util_warning(
          c("In %s, variables %s were excluded."),
          dQuote(arg_name),
          paste(dQuote(names(data_scale_problem_container)), collapse = ", "),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
        assign(arg_name, variable, envir = p)
      }
    }

    if (!all(is.na(need_role)) && length(roles) <
        length(unique(variable[!is.na(variable)]))) {
      util_warning(
        c("In %s, variables with roles matching %s should be specified, but not",
          "all variables have a role assigned in the metadata.",
          "I have %d variables but only %d roles"),
        dQuote(arg_name),
        dQuote(need_role),
        length(unique(variable[!is.na(variable)])),
        length(roles),
        applicability_problem = TRUE
      )
    }

    for (..vname in names(roles)) {
      # for all variables that have a variable role assigned in the metadata
      role <- roles[[..vname]]
      # fetch the variable role of the currently checked variable
      if (!is.null(role) && !is.na(role) && !is.na(need_role)) {
        # if we have a needed variable role and a variable role for the
        # currently checked variable,
        need_role. <- tolower(trimws(unlist(strsplit(need_role, "|",
                                                     fixed = TRUE),
                                            recursive = TRUE)))
        # parse the variable role (split at pipes |)
        not_role <-
          tolower(trimws(substring(need_role.[startsWith(need_role., "!")], 2)))
        # and parse the variable role for negation (e.g. *not* CONTEXT)
        well_role <-
          tolower(trimws(need_role.[!startsWith(need_role., "!")]))
        # and parse the variable role for no negation resp. (e.g. ANALYSIS)

        if ((length(well_role) > 0) &&
            (!all(well_role %in%
                  tolower(trimws(names(VARIABLE_ROLES)))))) {
          util_error(c(
            "Internal error:",
            "%s's %s contains invalid variable role names %s (allowed are %s).",
            "As a dataquieR developer, you should fix your call of %s."
          ),
          sQuote(arg_name),
          sQuote("need_role"),
          paste(dQuote(
            well_role[!(well_role %in%
                          tolower(trimws(names(VARIABLE_ROLES))))]
          ), collapse = ", "),
          paste(dQuote(tolower(trimws(names(VARIABLE_ROLES)))), collapse = ", "),
          sQuote(rlang::caller_call()[[1]])
          )
        }

        if ((length(not_role) > 0) &&
            (!all(not_role %in%
                  tolower(trimws(names(VARIABLE_ROLES)))))) {
          util_error(c(
            "Internal error:",
            "%s's %s contains invalid !variable role names %s (allowed are %s).",
            "As a dataquieR developer, you should fix your call of %s."
          ),
          sQuote(arg_name),
          sQuote("need_role"),
          paste(dQuote(
            not_role[!(not_role %in%
                         tolower(trimws(names(VARIABLE_ROLES))))]
          ), collapse = ", "),
          paste(dQuote(tolower(trimws(names(VARIABLE_ROLES)))), collapse = ", "),
          sQuote(rlang::caller_call()[[1]])
          )
        }

        if ((length(well_role) > 0) && (!tolower(trimws(role)) %in% well_role)) {
          # now check for the allowed variable roles
          record_problem(
            "Argument %s: Variable '%s' (%s) does not have an allowed variable role (%s)",
            dQuote(arg_name), ..vname, tolower(trimws(role)), need_role,
            applicability_problem = TRUE,
            intrinsic_applicability_problem = TRUE)
          data_role_problem_container[[..vname]] <- TRUE
        }
        if ((length(not_role) > 0) && (tolower(trimws(role)) %in% not_role)) {
          # and for the disallowed variable roles
          record_problem(
            "Argument %s: Variable '%s' (%s) has a disallowed variable role (%s)",
            dQuote(arg_name), ..vname, tolower(trimws(role)),
            util_pretty_vector_string(not_role),
            applicability_problem = TRUE,
            intrinsic_applicability_problem = TRUE)
          data_role_problem_container[[..vname]] <- TRUE
        }
      } else { # nocov start
        if (!is.na(need_role)) {
          util_warning(
            c("Argument %s: No variable role found for '%s' in the",
              "metadata, cannot check variable role"), dQuote(arg_name), variable,
            applicability_problem = TRUE)
        }
      } # nocov end
    }
    if (length(names(data_role_problem_container)) > 0) {
      variable <- setdiff(variable, names(data_role_problem_container))
      if (length(variable) == 0) {
        util_error(
          "In %s, none of the specified variables matches the requirements.",
          dQuote(arg_name),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
      } else {
        util_warning(
          c("In %s, variables %s were excluded."),
          dQuote(arg_name),
          paste(dQuote(names(data_role_problem_container)), collapse = ", "),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
        assign(arg_name, variable, envir = p)
      }
    }

    if (!all(is.na(need_computed_role)) && length(computed_roles) <
        length(unique(variable[!is.na(variable)]))) {
      util_warning(
        c("In %s, variables with computed roles matching %s should be specified, but not",
          "all variables have a computed role assigned in the metadata.",
          "I have %d variables but only %d computed roles"),
        dQuote(arg_name),
        dQuote(need_computed_role),
        length(unique(variable[!is.na(variable)])),
        length(computed_roles),
        applicability_problem = TRUE
      )
    }

    for (..vname in names(computed_roles)) {
      # for all variables that have a computed variable role assigned in the metadata
      computed_role <- computed_roles[[..vname]]
      # fetch the computed variable role of the currently checked variable
      if (!is.null(computed_role) && !is.na(computed_role) &&
          !is.na(need_computed_role)) {
        # if we have a needed computed variable role and a computed variable
        # role for the currently checked variable,
        need_computed_role. <-
          tolower(trimws(unlist(strsplit(need_computed_role, "|",
                                         fixed = TRUE),
                                recursive = TRUE)))
        # parse the computed variable role (split at pipes |)
        not_computed_role <-
          tolower(trimws(substring(
            need_computed_role.[startsWith(need_computed_role., "!")], 2)))
        # and parse the computed variable role for negation
        well_computed_role <-
          tolower(trimws(
            need_computed_role.[!startsWith(need_computed_role., "!")]))
        # and parse the computed variable role for no negation

        if ((length(well_computed_role) > 0) &&
            (!all(well_computed_role %in%
                  tolower(trimws(names(COMPUTED_VARIABLE_ROLES)))))) {
          util_error(c(
            "Internal error:",
            "%s's %s contains invalid computed variable role names %s (allowed are %s).",
            "As a dataquieR developer, you should fix your call of %s."
          ),
          sQuote(arg_name),
          sQuote("need_computed_role"),
          paste(dQuote(
            well_computed_role[!(well_computed_role %in%
                                   tolower(trimws(names(COMPUTED_VARIABLE_ROLES))))]
          ), collapse = ", "),
          paste(dQuote(tolower(trimws(names(COMPUTED_VARIABLE_ROLES)))), collapse = ", "),
          sQuote(rlang::caller_call()[[1]])
          )
        }

        if ((length(not_computed_role) > 0) &&
            (!all(not_computed_role %in%
                  tolower(trimws(names(COMPUTED_VARIABLE_ROLES)))))) {
          util_error(c(
            "Internal error:",
            "%s's %s contains invalid !computed variable role names %s (allowed are %s).",
            "As a dataquieR developer, you should fix your call of %s."
          ),
          sQuote(arg_name),
          sQuote("need_computed_role"),
          paste(dQuote(
            not_computed_role[!(not_computed_role %in%
                                  tolower(trimws(names(COMPUTED_VARIABLE_ROLES))))]
          ), collapse = ", "),
          paste(dQuote(tolower(trimws(names(COMPUTED_VARIABLE_ROLES)))), collapse = ", "),
          sQuote(rlang::caller_call()[[1]])
          )
        }

        if ((length(well_computed_role) > 0) &&
            (!tolower(trimws(computed_role)) %in% well_computed_role)) {
          # now check for the allowed computed variable roles
          record_problem(
            "Argument %s: Variable '%s' (%s) does not have an allowed computed variable role (%s)",
            dQuote(arg_name), ..vname, tolower(trimws(computed_role)),
            need_computed_role,
            applicability_problem = TRUE,
            intrinsic_applicability_problem = TRUE)
          data_computed_role_problem_container[[..vname]] <- TRUE
        }
        if ((length(not_computed_role) > 0) &&
            (tolower(trimws(computed_role)) %in% not_computed_role)) {
          # and for the disallowed computed variable roles
          record_problem(
            "Argument %s: Variable '%s' (%s) has a disallowed computed variable role (%s)",
            dQuote(arg_name), ..vname, tolower(trimws(computed_role)),
            util_pretty_vector_string(not_computed_role),
            applicability_problem = TRUE,
            intrinsic_applicability_problem = TRUE)
          data_computed_role_problem_container[[..vname]] <- TRUE
        }
      } else { # nocov start
        if (!is.na(need_computed_role)) {
          util_warning(
            c("Argument %s: No computed variable role found for '%s' in the",
              "metadata, cannot check computed variable role"),
            dQuote(arg_name), variable,
            applicability_problem = TRUE)
        }
      } # nocov end
    }
    if (length(names(data_computed_role_problem_container)) > 0) {
      variable <- setdiff(variable, names(data_computed_role_problem_container))
      if (length(variable) == 0) {
        util_error(
          "In %s, none of the specified variables matches the requirements.",
          dQuote(arg_name),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
      } else {
        util_warning(
          c("In %s, variables %s were excluded."),
          dQuote(arg_name),
          paste(dQuote(names(data_computed_role_problem_container)),
                collapse = ", "),
          applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE
        )
        assign(arg_name, variable, envir = p)
      }
    }
  }
}

#' Variable-argument roles
#'
#' A Variable-argument role is the intended use of an argument of a indicator
#' function -- an argument that refers variables.
#' In general for the table .variable_arg_roles, the suffix _var means one
#' variable allowed,
#' while _vars means more than one. The default  sets of arguments
#' for `util_correct_variable_use`/`util_correct_variable_use2` are defined
#' from the point of usage, e.g. if it could be, that NAs are in
#' the list of variable names, the function should be able to remove certain
#' response variables
#' from the output and not disallow them by setting `allow_na` to `FALSE`.
#'
#' @seealso `util_correct_variable_use()`
#' @seealso `util_correct_variable_use2()`
#'
#' @keywords internal
.variable_arg_roles <-
  dplyr::tribble(
    ~name, ~allow_na, ~allow_more_than_one, ~allow_null, ~allow_all_obs_na,
    ~allow_any_obs_na, ~min_distinct_values, ~need_type, ~need_scale,
    ~need_role, ~need_computed_role,
    "resp_var", FALSE, FALSE, FALSE, FALSE, TRUE, 0, NA, NA, NA, NA,
    "response_var", FALSE, FALSE, FALSE, FALSE, TRUE, 0, NA, NA, NA, NA,
    "resp_vars", FALSE, TRUE, FALSE, FALSE, TRUE, 0, NA, NA, NA, NA,
    "response_vars", FALSE, TRUE, FALSE, FALSE, TRUE, 0, NA, NA, NA, NA,
    "control_vars", TRUE, TRUE, TRUE, FALSE, TRUE, 0, NA, NA, NA, NA,
    "co_vars", TRUE, TRUE, TRUE, FALSE, TRUE, 0, NA, NA, NA, NA,
    "cluster_vars", FALSE, TRUE, FALSE, FALSE, FALSE, 0, NA, NA, NA, NA,
    "group_vars", FALSE, TRUE, FALSE, FALSE, FALSE, 0, NA, NA, NA, NA,
    "level_vars", FALSE, TRUE, FALSE, FALSE, FALSE, 0, NA, NA, NA, NA,
    "cluster_var", FALSE, FALSE, TRUE, FALSE, TRUE, 0, NA, NA, NA, NA,
    "level_var", FALSE, FALSE, TRUE, FALSE, TRUE, 0, NA, NA, NA, NA,
    "strata_vars", FALSE, FALSE, TRUE, FALSE, TRUE, 0, NA, NA, NA, NA,
    "time_vars", FALSE, FALSE, FALSE, FALSE, FALSE, 0, NA, NA, NA, NA,
    "id_vars", FALSE, FALSE, FALSE, FALSE, FALSE, 0, NA, NA, NA, NA
  )
# "process_vars",
# "key_segment_vars"

#' @importFrom stats na.omit
#' @rdname util_correct_variable_use
#' @family robustness_functions
#' @concept robustness
#' @noRd
util_correct_variable_use2 <- util_correct_variable_use
formals(util_correct_variable_use2)[["role"]] <- quote(arg_name)

#' Evaluate an expression and create a `dataquieR_result` object from
#' it's evaluated value
#'
#' if an error occurs, the function will return a corresponding object
#' representing that error. all conditions will be recorded and replayed,
#' whenever the result is printed by [print.dataquieR_result].
#'
#' @param expression the expression
#' @param env the environment to evaluate the expression in
#' @param filter_result_slots [character] regular expressions, only
#'                                               if an indicator function's
#'                                               result's name
#'                                               matches one of these, it'll
#'                                               be used for the report. If
#'                                               of length zero, no filtering
#'                                               is performed.
#' @param nm [character] name for the computed result
#' @param function_name [character] name of the function to be executed
#' @param my_storr_object a `storr` object to store the result in
#' @param my_call the call being executed (equivalent to `expression`)
#' @param init [logical] is this an initial call to compute dummy results?
#' @param called_in_pipeline [logical] if the evaluation should be considered
#'                                     as part of a pipeline.
#' @param checkpoint_resumed [logical] if `my_storr_object` and `nm` are set,
#'                                     check, if the result already exists and
#'                                     do not re-compute in such cases.
#'
#' @return a `dataquieR_result` object
#'
#' @family reporting_functions
#' @concept process
#' @noRd
util_eval_to_dataquieR_result <- function(expression, env = parent.frame(),
                                          filter_result_slots, nm,
                                          function_name,
                                          my_call = expression,
                                          my_storr_object = NULL,
                                          init = FALSE,
                                          called_in_pipeline = TRUE,
                                          checkpoint_resumed = FALSE) {
  withr::local_options(list(rlang_trace_top_env = rlang::current_env()))
  if (!isTRUE(getOption("dataquieR.traceback",
                        dataquieR.traceback_default))) {
    withr::local_options(list(rlang_backtrace_on_error = "none"))
  }
  if (missing(function_name)) {
    error <- try(
      function_name <- rlang::call_name(expression),
      silent = TRUE)
    if (util_is_try_error(error)) {
      function_name <- "unkown function"
    }
  }

  use_storr <- FALSE

  storr_err <-
    try(my_storr_object <- util_fix_storr_object(my_storr_object),
        silent = TRUE)

  if (util_is_try_error(storr_err)) {
    return(attr(storr_err, "condition"))
  }
  if (!is.null(my_storr_object) &&
      inherits(my_storr_object, "storr") &&
      !util_is_try_error(try(my_storr_object$list(), silent = TRUE))) {
    use_storr <- TRUE
    if (!init) {
      my_storr_object$del(NO_SHARED_STORR)
    }
  }
  .old_.dq2_globs_.called_in_pipeline <- .dq2_globs$.called_in_pipeline
  .dq2_globs$.called_in_pipeline <- called_in_pipeline
  on.exit({
    .dq2_globs$.called_in_pipeline <- .old_.dq2_globs_.called_in_pipeline
  })
  errors <- list()
  warnings <- list()
  messages <- list()
  e <- environment()
  collect_condition <- function(cnd) {
    if (!isTRUE(getOption("dataquieR.traceback",
                          dataquieR.traceback_default))) {
      attr(cnd, "trace") <- NULL
      cnd$trace <- NULL
    }
    if (inherits(cnd, "error")) {
      e$errors <- c(e$errors, list(cnd))
    } else if (inherits(cnd, "warning")) {
      e$warnings <- c(e$warnings, list(cnd))
    } else if (inherits(cnd, "message")) {
      e$messages <- c(e$messages, list(cnd))
    } else {
      util_error("")
    }
  }
  #r_summary <- data.frame()

  need_compute <- TRUE

  if (checkpoint_resumed &&
      called_in_pipeline &&
      use_storr &&
      my_storr_object$exists(nm, namespace =
                             util_get_storr_stat_namespace(my_storr_object))) {
    # r <- my_storr_object$get(nm)
    # if (inherits(r, "dataquieR_result")) { # FIXME: Write some Ack to the storr instead of this
    #   if (inherits(r, "dataquieR_NULL")) {
    #     err <- attr(r, "error")
    #     if (length(err) > 0) {
    #       util_stop_if_not(
    #         "Internal error, sorry. Please report: Got more than one error in one dataquieR_result." =
    #           length(err) != 1
    #       )
    #       intrinsic_applicability_problem <-
    #         attr(err[[1]], "intrinsic_applicability_problem")
    #       applicability_problem <-
    #         attr(err[[1]], "applicability_problem")
    #       need_compute <-
    #         !intrinsic_applicability_problem &&
    #         !applicability_problem # TODO: Maybe, if metadata are allowed to change in between, need_compute could be TRUE for applicability problems
    #     }
    #   } else {
        need_compute <- FALSE
      # }
    # }
  }
  if (!need_compute) {
    r <- NA
  } else {
    add_attribs <- new.env(parent = emptyenv())
    r_summary1 <- data.frame()
    r <- list()
    class(r) <- union("empty", class(r))
    tm <- system.time(
    suppressWarnings(suppressMessages(try(withCallingHandlers(
      {
        r <-  eval(expression, envir = env)
        if (length(r)) {
          if (length(filter_result_slots)) {
            r <- util_filter_names_by_regexps(r,
                                              filter_result_slots)
          }
          if (.called_in_pipeline) {
            r <- util_compress_ggplots_in_res(r)
          }
        }
      },
      error = collect_condition,
      warning = collect_condition,
      message = collect_condition
    ), silent = TRUE))))
    if (length(r) == 0) {
      r <- list()
      class(r) <- union("dataquieR_NULL", class(r))
    } else if ("SummaryTable" %in% names(r)) {
      meta_data <- try(eval(quote(meta_data), envir = env), silent = TRUE)
      label_col <- try(eval(quote(label_col), envir = env), silent = TRUE)
      if (is.data.frame(meta_data) &&
          is.character(label_col) &&
          length(label_col) == 1) {
      r$SummaryTable <- util_map_to_other_metrics(
                                          r$SummaryTable,
                                          eval(quote(meta_data),
                                               envir = env),
                                          get("label_col",
                                              envir = env,
                                              inherits = TRUE),
                                          add_attribs = add_attribs)

      }
    }
    for (a in names(add_attribs)) {
      attr(r, a) <- add_attribs[[a]]
    }
    attr(r, "error") <- errors
    attr(r, "warning") <- warnings
    attr(r, "message") <- messages
    attr(r, "system.time") <- tm

    attr(r, "function_name") <- function_name #### from util_evalute_calls
    attr(r, "cn") <- util_sub_string_left_from_.(nm)
    attr(r, "call") <- my_call

    class(r) <- union("dataquieR_result", class(r))
    res_check <- try(r <- util_dataquieR_result(r), silent = TRUE)
    if (util_is_try_error(res_check)) {
      attr(r, "error") <- util_attach_attr(list(attr(res_check, "condition")),
                                          class = "dataquieR_invalid_result_error"
      )
    }

    if (called_in_pipeline) {

      if (length(r) > 0) {
        try({
          # r_summary <- util_extract_indicator_metrics(r$SummaryTable)
          s <- prep_extract_summary(r)
          r_summary1 <- suppressWarnings(prep_summary_to_classes(s)) # FIXME: split this function, we do not want classes, here
          #      r_summary <- r_summary0$Table

        }, silent = TRUE)
      }

      CAT_ <-
        vapply(setNames(nm = c("applicability", "error", "anamat", "indicator_or_descriptor")), function(aspect) {
          as.character(as.numeric(util_as_cat(util_get_category_for_result(r, aspect = aspect))))
        }, FUN.VALUE = character(1))

      MSG_ <-
        vapply(setNames(nm = c("applicability", "error", "anamat", "indicator_or_descriptor")), function(aspect) {
          util_get_message_for_result(r, aspect = aspect)
        }, FUN.VALUE = character(1))

      names(CAT_) <- paste0("CAT_", names(CAT_))
      names(MSG_) <- paste0("MSG_", names(MSG_))

      # rownames(CAT_) <- NULL
      # rownames(MSG_) <- NULL
      # rownames(r) <- NULL

      # if (is.data.frame(r_summary) && nrow(r_summary) > 0) {
      #   r_summary <- cbind.data.frame(r_summary, t(CAT_), t(MSG_))
      # } else {
      r_summary <- cbind.data.frame(t(CAT_), t(MSG_))
      # }

      r_summary <- data.frame(
        VAR_NAMES = unname(attr(my_call, VAR_NAMES)),
        STUDY_SEGMENT = unname(attr(my_call, STUDY_SEGMENT)),
        call_names = unname(attr(r, "cn")),
        value = as.character(r_summary),
        values_raw = as.character(r_summary),
        function_name = function_name,
        indicator_metric = names(r_summary)
      )
      r_summary <- util_rbind(r_summary, r_summary1)
      r_summary$function_name <- function_name

      attr(r_summary, "resnames") <- names(r)

      attr(r, "r_summary") <- r_summary

    }

  #  save(my_storr_object, r, file = "/tmp/debug")

    if (!is.null(my_storr_object) &&
        inherits(my_storr_object, "storr") &&
        !util_is_try_error(try(my_storr_object$list(), silent = TRUE))) { # TODO: During tests, check for RDS files in the back-end larger than some threshold
      # util_message("Before: %s", strsplit(as.character(system("df -h /", intern = TRUE)), "\\s+")[[2]][4])
      # util_message("OS1: %s", capture.output(print(object.size(r), units = "MB", digits = 2)))
      # util_message("OS2: %s", capture.output(print(object.size(r_summary), units = "MB", digits = 2)))
      my_storr_object$set(
        key = nm,
        value = r_summary,
        namespace = util_get_storr_summ_namespace(my_storr_object))
      if (!init && (!inherits(r, "dataquieR_NULL") ||
          length(attr(r, "error")) == 0 ||
          isTRUE(attr(attr(r, "error")[[1]], "intrinsic_applicability_problem")))) {
        if (!init) {
          my_storr_object$set(key = nm, value = TRUE, namespace =
                                util_get_storr_stat_namespace(my_storr_object))
        }
        if (called_in_pipeline)
          r <- util_compress(r)
        my_storr_object$set(key = nm, value = r)
      }
      my_storr_object$flush_cache() # TODO: Needed?
      # util_message("After: %s", strsplit(as.character(system("df -h /", intern = TRUE)), "\\s+")[[2]][4])
      # if (util_as_numeric_with_unit(paste0(strsplit(as.character(system("df -H /", intern = TRUE)), "\\s+")[[2]][4], "b")) < util_as_numeric_with_unit("10Gb"))
      #   util_error("Stopped by too low disk space")
      # # if (util_get_avail_ram() < 1024^2 * 2) { # less than 2 MByte free
      # #   util_error("Stopped by too low RAM")
      # # }
      # util_message("Free: %s", capture.output(print(util_get_avail_ram(), units = "GiB")))
      r <- NA
    } else if (!is.null(my_storr_object)) {
      r <- "Invalid storr object"
    } else {
      if (called_in_pipeline)
        r <- util_compress(r)
    }

    try(my_storr_object$driver$disconnect(), silent = TRUE)

    try(my_storr_object$driver$env$sync(force = TRUE), silent = TRUE)
  }

  invisible(gc(verbose = FALSE, full = FALSE))
  r
}

.dq2_globs <- new.env(parent = emptyenv())
.dq2_globs$.called_in_pipeline <- FALSE

#' Adjust the data types of study data, if needed
#'
#' @param study_data [data.frame] the study data
#' @param meta_data [meta_data] `VAR_NAMES` relevant for warnings about
#'                              conversion errors
#' @param relevant_vars_for_warnings [character]
#'
#' @return [data.frame] modified study data
#' @noRd
util_adjust_data_type <- function(study_data,
                                  meta_data, # TODO: Use maybe readr::type_convert
                                  relevant_vars_for_warnings = character(0)) {
  # FIXME: The old adjust_type function does not interpret 19 as 0019 but as 2019. same hint in the test for test-acc_loess.R -- 1969 is the first year with 19, there.
  if (!isTRUE(as.logical(getOption("dataquieR.old_type_adjust",
                dataquieR.old_type_adjust_default)))) {
    cl <- sys.call()
    cl[[1]] <-
      rlang::call2(":::", rlang::sym("dataquieR"),
                   rlang::sym("util_adjust_data_type2"))
    return(eval(cl,
                envir = parent.frame(),
                enclos = parent.frame()))
  }
  # prevent repeated data type checks if adjusted previously
  if (!isTRUE(attr(study_data, "Data_type_matches"))) {
    check_type <- util_compare_meta_with_study(
      sdf = study_data,
      mdf = meta_data,
      label_col = VAR_NAMES,
      check_convertible = TRUE, # slow, but now cached
      check_conversion_stable = FALSE,
      threshold_value = 0)

    if (any(check_type != 1)) {
      vars_to_transform <- names(check_type)[check_type != 1]
      expected_type <- setNames(util_map_labels(
        x = vars_to_transform,
        to = DATA_TYPE,
        from = VAR_NAMES,
        ifnotfound =
          DATA_TYPES$STRING,
        meta_data = meta_data,
        warn_ambiguous = FALSE),
        nm = vars_to_transform)

      transf_sd <- lapply(
        setNames(nm = vars_to_transform),
        function(rv) {
          rv2 <- util_data_type_conversion(study_data[[rv]],
                                           expected_type[[rv]])
          if (!identical(which(is.na(study_data[[rv]])),
                         which(is.na(rv2)))) {
            if (length(relevant_vars_for_warnings) == 0 || # did not find relevant vars vector, show warning in any case
                rv %in% relevant_vars_for_warnings) {
              util_message(
                paste("Data type transformation of", dQuote(rv), "introduced",
                      length(which(is.na(rv2))) -
                        length(which(is.na(study_data[[rv]]))),
                      "additional missing values."),
                applicability_problem = TRUE)
            }
          }
          return(rv2)
        })
      study_data[, vars_to_transform] <- transf_sd
    }
    logcs <- names(which(vapply(study_data, is.logical, FUN.VALUE = logical(1)))) # these should be NA only.
    if (identical(Sys.getenv("TESTTHAT"), "true")) {
      util_stop_if_not(
        "all logical columns should be NA only." =
          all(is.na(unlist(as.vector(study_data[, logcs])))))
    }
    if (length(logcs) > 0) {
      expected_type <- setNames(util_map_labels( # convert them to the desired NA type -- can happen, if data type was compatible, so no conversion was done
        x = logcs,
        to = DATA_TYPE,
        from = VAR_NAMES,
        ifnotfound =
          DATA_TYPES$FLOAT,
        meta_data = meta_data,
        warn_ambiguous = FALSE),
        nm = logcs)
      lgcs_which_are_flts <- names(which(expected_type == DATA_TYPES$FLOAT))
      lgcs_which_are_intgs <- names(which(expected_type == DATA_TYPES$INTEGER))
      study_data[, lgcs_which_are_intgs] <-
        lapply(study_data[, lgcs_which_are_intgs, drop = FALSE],
               as.integer)
      study_data[, lgcs_which_are_flts] <-
        lapply(study_data[, lgcs_which_are_flts, drop = FALSE],
               as.double)
    }
    attr(study_data, "Data_type_matches") <- TRUE
  }
  study_data
}

#' Generate a report summary table
#'
#' @param object a square result set
#' @param aspect an aspect/problem category of results
#' @param FUN function to apply to the cells of the result table
#' @param ... not used
#' @param collapse passed to `FUN`
#'
#' @return a summary of a `dataquieR` report
#' @export
#' @examples
#' \dontrun{
#'   util_html_table(summary(report),
#'        filter = "top", options = list(scrollCollapse = TRUE, scrollY = "75vh"),
#'        is_matrix_table = TRUE, rotate_headers = TRUE
#'   )
#' }
summary.dataquieR_resultset2 <- function(object, aspect = c("applicability", "error", "anamat", "indicator_or_descriptor"),
                                   FUN,
                                   collapse = "\n<br />\n",
                                   ...) {
  my_storr_object <- util_get_storr_object_from_report(object)
  # FIXME: Go only once ofer all results
  # TODO: depending on the back-end, do not access the single results but load from the summary namespace
  if (!is.null(attr(object, "repsum")) &&
      identical(rlang::call_args_names(rlang::call_match()), "object") &&
      identical(attr(attr(object, "repsum"), "rule_digest"),
                rlang::hash(list(util_get_rule_sets(),
                                 util_get_ruleset_formats())))
      ) {
    return(attr(object, "repsum"))
  }
  if (missing(FUN)) {# TODO: I need this long-format overall dq summary data frame also elsewhere
    util_stop_if_not("aspect is only supported for specific FUN values" =
                       missing(aspect))

    labels_of_var_names_in_report <- setNames(
      attr(object, "meta_data")[[attr(object, "label_col")]],
      nm = attr(object, "meta_data")[[VAR_NAMES]]
    )

    var_names_of_labels_in_report <- setNames(
      nm = attr(object, "meta_data")[[attr(object, "label_col")]],
      attr(object, "meta_data")[[VAR_NAMES]]
    )

    labels <- util_get_labels_grading_class()
    colors <- util_get_colors()
    names(labels) <- paste0("cat", names(labels))
    names(colors) <- paste0("cat", names(colors))
    if (!identical(names(labels), names(colors)) ||
        !all(paste0("cat", seq_len(5)) %in% names(colors))) {
      util_warning(
        c("Could not find enough categories in custom format file,",
          "falling back to default"), applicability_problem = TRUE,
        intrinsic_applicability_problem = FALSE)
      old_o <- options(
        dataquieR.grading_formats =
          system.file("grading_formats.xlsx",
                      package = "dataquieR"),
        dataquieR.grading_rulesets =
          system.file("grading_rulesets.xlsx",
                      package = "dataquieR"))
      on.exit(options(old_o))
      labels <- util_get_labels_grading_class()
      colors <- util_get_colors()
      names(labels) <- paste0("cat", names(labels))
      names(colors) <- paste0("cat", names(colors))
    }

    labels <- as.list(labels)
    colors <- as.list(colors)

    labels[["catNA"]] <- htmltools::HTML("\u00A0") # this is nbsp, but it reads " ", even in the HTML source code
    labels[["NA"]] <- htmltools::HTML("\u00A0")
    colors[["catNA"]] <- "#ffffff00" # transparent / white
    colors[["NA"]] <- "#ffffff00" # transparent / white

    icons <- lapply(labels, function(x) htmltools::HTML("\u00A0")) # not available, yet

    order_of <- suppressWarnings(setNames(as.integer(gsub("^cat", "",
                              util_as_cat(names(colors)))), nm = names(colors)))

    filter_of <- labels

    if (!is.null(my_storr_object)) {
      namespace = util_get_storr_summ_namespace(my_storr_object)
      all_sums <- my_storr_object$mget(my_storr_object$list(
        namespace = namespace),
                          namespace = namespace)
    } else {
      all_sums <- lapply(object, attr, "r_summary")
    }

    result <- util_rbind(data_frames_list = all_sums)

    if (any(is.na(result$n_classes))) {
      result[is.na(result$n_classes), "n_classes"] <- 5
    }

    ordered_call_names <- colnames(object)
    ordered_var_names <- var_names_of_labels_in_report[rownames(object)]


    # spreaded_result <- lapply(
    #   setNames(nm = intersect(ordered_call_names, result$call_names)),
    #   function(cn) vapply(
    #     FUN.VALUE = character(1),
    #     setNames(nm = intersect(ordered_var_names, result$VAR_NAMES)),
    #     util_get_html_cell_for_result,
    #     cn = cn,
    #     result = result,
    #     icons = icons,
    #     labels = labels,
    #     colors = colors,
    #     order_of = order_of,
    #     filter_of = filter_of,
    #     labels_of_var_names_in_report = labels_of_var_names_in_report,
    #     var_names_of_labels_in_report = var_names_of_labels_in_report
    #   )
    # )
    # spreaded_result_df <- do.call(data.frame, spreaded_result)
    # Variables <-
    #   labels_of_var_names_in_report[rownames(spreaded_result_df)]
    # spreaded_result_df <- cbind(
    #   data.frame(Variables = Variables), spreaded_result_df)
    # rownames(spreaded_result_df) <- NULL

    indicator_metric <- NULL

    if (!prod(dim(result))) {
      cls <-
        c("VAR_NAMES", "class", "indicator_metric", "value", "values_raw",
          "n_classes", "STUDY_SEGMENT", "call_names", "function_name")
      result <-
        as.data.frame(setNames(rep(list(character(0)), length(cls)), nm = cls))
    }

    if (!"class" %in% colnames(result)) {
      result$class <- rep(NA, nrow(result))
    }

#    result$class <- 1 # fixme: this needs to be replaced by the interval code R/prep_summary_to_classes.R
#    result$class[is.na(result$class)] <- "NA"
    result$class <- util_as_cat(result$class)

    result <- result[!is.na(result[[VAR_NAMES]]), , FALSE] # remove all non-variable-related stuff, not yet supported, here.

    # compute stopped_functions
    stopped_functions <- vapply(object,
                                inherits, what = "dataquieR_NULL",
                                FUN.VALUE = logical(1) )





    # spreaded_result_df$Total <- rowmaxes$cell_text[
    #   spreaded_result_df$Variables]
    #
    # class(spreaded_result_df) <- c("dataquieR_summary", "data.frame")

    spreaded_result_df <- NA
    class(spreaded_result_df) <- c("dataquieR_summary")


    matrix_list <- attr(object, "matrix_list")
    function_alias_map <- attr(matrix_list, "function_alias_map")
    function_alias_map <- unique(function_alias_map[ , c("name", "alias"), drop = FALSE])
    alias_names <- setNames(function_alias_map$name, nm = function_alias_map$alias)

    # FIXME: avoid to do this already once, before
    result <- util_metrics_to_classes(result,
                                      attr(object, "meta_data")) # re-classify

    rowmaxes <- util_compute_rowmaxes(result = result,
                                      labels = labels,
                                      colors = colors,
                                      order_of = order_of,
                                      filter_of = filter_of,
                                      labels_of_var_names_in_report =
                                        labels_of_var_names_in_report)


    this <- new.env(parent = emptyenv())

    for (prop in c("result", # long format of the summary, column indicator_metric contains CAT_ and MSG_, which are special process classes
                   "rowmaxes", # maximum category for each row in the summary table
                   # and all the ellipsis_arguments from
                   # util_get_html_cell_for_result, above:
                   "labels_of_var_names_in_report",
                   "alias_names",
                   "stopped_functions",
                   "colors",
                   "labels",
                   "order_of",
                   "filter_of",
                   "ordered_call_names",
                   "ordered_var_names"
    )) {
      this[[prop]] <- get(prop)
    }

    # needed for the summary downlaod button
    this[["title"]] <- attr(object, "title")
    this[["subtitle"]] <- attr(object, "subtitle")

    # needed to pretty print
    this[["meta_data"]] <- attr(object, "meta_data")

    # "label_col", # needed to pretty print
    this[["label_col"]] <- attr(object, "label_col")

    # needed for pretty plot
    this[["rownames_of_report"]] <- rownames(object)
    this[["colnames_of_report"]] <- colnames(object)

    attr(spreaded_result_df, "this") <- this
    attr(spreaded_result_df, "rule_digest") <-
              rlang::hash(list(util_get_rule_sets(),
                               util_get_ruleset_formats()))
    spreaded_result_df

  } else {
    f <- substitute(FUN)
    FUN <- force(eval(f, enclos = parent.frame(), envir = environment()))
    util_ensure_suggested("htmltools", "Generating nice tables")
    util_stop_if_not(inherits(object, "dataquieR_resultset2"))
    aspect <- util_match_arg(aspect, several_ok = FALSE)

    rn_obj <- rownames(object)
    cn_obj <- colnames(object)

    rn_to_use <- vapply(rn_obj, function(rn) {
      any(vapply(object[rn, , drop = TRUE],
                 inherits,
                 "dataquieR_result",
                 FUN.VALUE = logical(1)))
    }, FUN.VALUE = logical(1))

    cn_to_use <-
      vapply(cn_obj, function(cn) {
        # TODO: check entity attribute instead and omit outputs with entity != item
        x <- object[, cn, drop = FALSE]
        length(x) > 0 &&
          any(!endsWith(names(x), ".[ALL]")) &&
          all(vapply(names(x), function(listname) {
            any(endsWith(listname, paste0(".", rownames(object))))
          }, FUN.VALUE = logical(1)))
      }, FUN.VALUE = logical(1))

    rn_obj <- rn_obj[rn_to_use]
    cn_obj <- cn_obj[cn_to_use]

    if (!any(rn_to_use) || !any(cn_to_use)) {
      return(
        matrix(nrow = 0, ncol = 2,
               dimnames = list(c(), c(VAR_NAMES, STUDY_SEGMENT)))
      )
    }

    do.call(rbind, lapply(setNames(nm = rn_obj),
                          FUN = function(rn) {
                            vapply(setNames(nm = cn_obj),
                                   FUN.VALUE = character(1),
                                   FUN = function(cn, aspect, collapse) {
                                     r <- FUN(object[rn, cn, drop = TRUE],
                                              aspect = aspect,
                                              collapse = collapse,
                                              rn = rn,
                                              cn = cn)
                                     # if (inherits(r, "try-error")) {
                                     #   debug(FUN)
                                     #   r <- FUN(object[rn, cn, drop = TRUE],
                                     #            aspect = aspect,
                                     #            collapse = collapse,
                                     #            rn = rn,
                                     #            cn = cn)
                                     # }
                                     # if (!is.character(r) || length(r) != 1)
                                     r
                                   },
                                   aspect = aspect,
                                   collapse = collapse)
                          }
    )) # TOOD: "Any-Issue" Column
  }
}


util_reclassify_dataquieR_summary <- function(x) {
  util_stop_if_not(inherits(x, "dataquieR_summary"))

  if (identical(attr(x, "rule_digest"),
            rlang::hash(list(util_get_rule_sets(),
                             util_get_ruleset_formats()))))
    return(x)

  rs_table_long <- attr(x, "this")$result
  meta_data <- attr(x, "this")$meta_data

  attr(x, "this")$result <-
    util_metrics_to_classes(rs_table_long, meta_data)

  attr(x, "this")$rowmaxes <- util_compute_rowmaxes(
    result = attr(x, "this")$result,
    labels = attr(x, "this")$labels,
    colors = attr(x, "this")$colors,
    order_of = attr(x, "this")$order_of,
    filter_of = attr(x, "this")$filter_of,
    labels_of_var_names_in_report =
      attr(x, "this")$labels_of_var_names_in_report)

  attr(x, "rule_digest") <-
    rlang::hash(list(util_get_rule_sets(),
                     util_get_ruleset_formats()))

  x
}

util_compute_rowmaxes <- function(result,
                                  labels,
                                  colors,
                                  order_of,
                                  filter_of,
                                  labels_of_var_names_in_report) {
  indicator_metric <- NULL # make R CMD check happy
  if (!VAR_NAMES %in% colnames(result)) {
    result[[VAR_NAMES]] <- rep(NA_character_, nrow(result))
  }
  if (!"class" %in% colnames(result)) {
    result[["class"]] <- rep(NA_character_, nrow(result))
  }
  result %>% dplyr::filter(!startsWith(as.character(indicator_metric), "CAT_") &
                             !startsWith(as.character(indicator_metric), "MSG_")) %>%
    dplyr::group_by(VAR_NAMES) %>%
    dplyr::summarise(rowmax = suppressWarnings(######
                                               max(util_as_cat(class), na.rm = TRUE))) ->
    rowmaxes

  rowmax <- paste(rowmaxes$rowmax) # also make NA -> "NA"

  rowmaxes$label <- vapply(labels[rowmax], identity,
                           FUN.VALUE = character(1))

  rowmaxes$color <- vapply(colors[rowmax], identity,
                           FUN.VALUE = character(1))

  rowmaxes$order <- vapply(order_of[rowmax], identity,
                           FUN.VALUE = integer(1))

  rowmaxes$filter <- vapply(filter_of[rowmax], identity,
                            FUN.VALUE = character(1))

  get_cell_text <- function(lb, cl, o, f, vn) {
    # util_message("lb = %s, cl = %s, o = %s, f = %s, vn = %s",
    #              sQuote(lb), sQuote(cl), sQuote(o), sQuote(f), sQuote(vn))
    link <- util_generate_anchor_link(labels_of_var_names_in_report[[vn]],
                                      "",
                                      title =
                                        htmltools::HTML(as.character(lb)))
    link$attribs$style <- c(link$attribs$style,
                            "text-decoration:none;display:block;")
    fg_color <- util_get_fg_color(cl)
    link$attribs$style <- c(link$attribs$style, sprintf("color:%s;", fg_color))

    paste0(
      htmltools::tagList(
        # htmltools::p(cn),
        # htmltools::p(vn)
        htmltools::pre(
          onclick = "",
          style = sprintf("height: 100%%; min-height: 2em; margin: 0em; padding: 0em; background: %s; cursor: pointer; text-align: center;", cl),
          sort = o,
          filter = f,
          title = "",
          link)
      ), collapse = "\n")
  }

  rowmaxes$cell_text <- setNames(vapply(mapply(
    SIMPLIFY = FALSE,
    FUN = get_cell_text,
    lb = rowmaxes$label,
    cl = rowmaxes$color,
    o = rowmaxes$order,
    f = rowmaxes$filter,
    vn = rowmaxes$VAR_NAMES
  ),
  FUN = identity,
  FUN.VALUE = character(1)),
  nm = labels_of_var_names_in_report[rowmaxes$VAR_NAMES])

  return(rowmaxes)

}

#' Get variable attributes of a certain provision level
#'
#' This function returns all variable attribute names of a certain
#' metadata provision level or of more than one level.
#'
#' @param level level(s) of requirement
#' @param cumulative include all names from more basic levels
#'
#' @return all matching variable attribute names
#'
#' @family metadata_management
#' @concept metadata_management
#' @noRd


util_get_var_att_names_of_level <- function(level, cumulative = TRUE) {
  if (length(level) > 0) {
    level <- match.arg(level, choices = unlist(VARATT_REQUIRE_LEVELS),
                       several.ok = TRUE)
  } else {
    return(character(0))
  }
  if (cumulative) {
    max_level <- max(which(VARATT_REQUIRE_LEVELS_ORDER %in% level))
    level <- head(VARATT_REQUIRE_LEVELS_ORDER, max_level)
  }
  r <- vapply(WELL_KNOWN_META_VARIABLE_NAMES, function(att_name) {
    if (attr(att_name, REQUIREMENT_ATT) %in% level) {
      att_name
    } else {
      NA_character_
    }
  }, character(1))
  r[!is.na(r)]
}

#' remove all records, that have at least one `NA` in any of the given variables
#'
#' @param study_data the study data frame
#' @param vars the variables being checked for `NA`s
#'
#' @return modified study_data data frame
#'
#' @examples
#' \dontrun{
#' dta <- iris
#' dim(util_remove_na_records(dta))
#' dta$Species[4:6] <- NA
#' dim(util_remove_na_records(dta))
#' dim(util_remove_na_records(dta, c("Sepal.Length", "Petal.Length")))
#' }
#'
#'
#' @family data_management
#' @concept missing
#' @noRd
util_remove_na_records <- function(study_data, vars = colnames(study_data)) {
  obs_wo_na <- rowSums(is.na(study_data[, vars, FALSE])) == 0
  if (sum(!obs_wo_na) > 0) {
    util_message(
      c("Removing %d observations because of NAs in some of",
        "the following columns: %s"),
      sum(!obs_wo_na),
      paste0(dQuote(vars), collapse = ", ")
    )
  }
  r <- study_data[obs_wo_na, , FALSE]
  attributes(r)[.ds1_attribute_names] <-
    attributes(study_data)[.ds1_attribute_names]
  r
}

test_that("util_remove_na_records works", {
  skip_on_cran()

  result <- util_remove_na_records(cars)

  # Check that the result has the correct structure
  expect_s3_class(result, "data.frame")

  # Check that the result has the correct values when there are no NAs
  expect_identical(result, cars)

  # Check that the result shows a message when NAs where removed
  # Set a seed for reproducibility
  set.seed(123)

  # Generate random indices to replace with NA
  indices_to_replace1 <- sample(1:length(cars$dist), size = 10)
  indices_to_replace2 <- sample(1:length(cars$dist), size = 10)

  # Replace values at selected indices with NA
  cars_na <- cars
  cars_na$dist[indices_to_replace1] <- NA
  cars_na$speed[indices_to_replace2] <- NA

  # TODO: improve the message

  # Find how many NAs are left
  # nas <- sum(rowSums(is.na(cars_na[, colnames(cars_na), FALSE])) > 0)

  expect_message2(result_remove_na <- util_remove_na_records(cars_na),
                 regexp = ". observations because of NAs in some of the following columns.",
                 perl = TRUE)

})

#' Remove unused levels from `ReportSummaryTable`
#'
#' @param x `ReportSummaryTable` an object from which to drop unused factor
#'          levels.
#' @param ... no used.
#'
#' @return `ReportSummaryTable` with all (`NA` or `0`)-columns removed
#' @export
droplevels.ReportSummaryTable <- function(x, ...) {
  util_stop_if_not(inherits(x, "ReportSummaryTable"))
  cols_to_change <- setdiff(colnames(x), c("Variables", "N"))
  we_can_be_dropped <- vapply(x[, cols_to_change], function(cl) {
    all(is.na(cl) | 0 == cl)
  }, FUN.VALUE = logical(1))
  x[, names(which(we_can_be_dropped))] <- NULL
  x
}

util_is_time_only <- function(x, na_ok = TRUE) { # TODO: See also in redcap-rules-env
  # Already "pure time" classes
  if (inherits(x, "hms")) return(TRUE)
  if (inherits(x, "times")) return(TRUE)                  # chron::times

  # Character strings that parse cleanly as time-of-day
  if (is.character(x)) {
    ok <- !is.na(suppressWarnings(readr::parse_time(x)))
    return(if (na_ok) all(ok | is.na(x)) else all(ok))
  }

  # Numeric that clearly represents time-of-day (seconds or Excel fractions)
  if (is.numeric(x)) {
    # Treat values in [0, 1) as Excel day fractions (time of day),
    # otherwise treat as seconds modulo 24h
    v <- x[!is.na(x)]
    looks_like_fraction <- length(v) && all(v >= 0 & v < 1)
    if (looks_like_fraction) return(TRUE)
    # Seconds in a reasonable range? accept as time of day too.
    if (length(v) && all(v >= 0 & v < 86400*2)) return(TRUE)  # allow 0..<2 days
  }

  if (is.list(x)) {
    return(all(vapply(x, util_is_time_only, na_ok = na_ok,
                      FUN.VALUE = logical(1))))
  }

  # POSIXt: has a date, but can be reduced to time-of-day
  inherits(x, "POSIXt") && !is.null(x)
}

# ------------- Coercion to hms ----------------------------------------------

util_as_time_only <- function(x, tz = "UTC") { # TODO: See util_parse_time
  # Normalize any “time-ish” input to hms::hms (seconds since midnight).
  sec_per_day <- 86400

  if (inherits(x, "hms")) return(x)

  if (inherits(x, "times")) {              # chron::times -> hms
    return(hms::as_hms(as.numeric(x) * sec_per_day))
  }

  if (inherits(x, "POSIXt")) {
    # Extract time-of-day in given tz
    tt <- as.POSIXlt(x, tz = tz)
    secs <- tt$hour * 3600 + tt$min * 60 + floor(tt$sec)
    return(hms::as_hms(secs %% sec_per_day))
  }

  if (inherits(x, "difftime")) {
    return(hms::as_hms(as.numeric(x, units = "secs") %% sec_per_day))
  }

  if (is.character(x)) {
    # Robust parser from readr -> hms
    return(suppressWarnings(readr::parse_time(x)))
  }

  if (is.factor(x)) {
    return(hms::as_hms(as.character(x)))
  }

  if (is.numeric(x)) {
    # Two conventions:
    # - [0,1) -> Excel fraction of a day
    # - else  -> seconds since midnight (or arbitrary seconds -> modulo 24h)
    v <- x
    is_frac <- !is.na(v) & v >= 0 & v < 1
    secs <- ifelse(is_frac, v * sec_per_day, v)
    return(hms::as_hms(secs %% sec_per_day))
  }

  if (is.list(x)) {
    return(lapply(x, function(x) {
      r <- NA
      try(
        r <- util_as_time_only(x, tz = tz),
        silent = TRUE
      )
      return(r)
    }))
  }

  # Logical, lists, etc. -> try character fallback
  hms::as_hms(as.character(x))
}

#' Get colors for each russet `DQ` category
#'
#' @return named vector of colors, names are categories (e.g, "1" to "5")
#'         values are colors as `HTML` `RGB` hexadecimal strings
#' @family summary_functions
#' @concept process
#' @noRd
util_get_colors <- function() {

  rsts <- util_get_ruleset_formats()

  cat_names <- as.character(rsts$category)
  if (!"color" %in% colnames(rsts)) {
    rsts[["color"]] <-
      rep("128 128 128", nrow(rsts))
  }
  colors <- rsts[["color"]]
  colors <- util_col2rgb(colors)
  setNames(colors, nm = cat_names)
}

#' Expand code labels across variables
#'
#' Code labels are copied from other
#' variables, if the code is the same and the
#' label is set only for some variables
#'
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param suppressWarnings [logical] show warnings, if labels are expanded
#' @param mix_jumps_and_missings [logical] ignore the class of the codes for
#'                               label expansion, i.e., use missing code labels
#'                               as jump code labels, if the values are the
#'                               same.
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return [data.frame] an updated metadata data frame.
#' @export
#'
#' @examples
#' \dontrun{
#' meta_data <- prep_get_data_frame("meta_data")
#' meta_data$JUMP_LIST[meta_data$VAR_NAMES == "v00003"] <- "99980 = NOOP"
#' md <- prep_expand_codes(meta_data)
#' md$JUMP_LIST
#' md$MISSING_LIST
#' md <- prep_expand_codes(meta_data, mix_jumps_and_missings = TRUE)
#' md$JUMP_LIST
#' md$MISSING_LIST
#' meta_data <- prep_get_data_frame("meta_data")
#' meta_data$MISSING_LIST[meta_data$VAR_NAMES == "v00003"] <- "99980 = NOOP"
#' md <- prep_expand_codes(meta_data)
#' md$JUMP_LIST
#' md$MISSING_LIST
#' }
prep_expand_codes <- function(item_level = "item_level",
                              suppressWarnings = FALSE,
                              mix_jumps_and_missings = FALSE,
                              meta_data_v2,
                              meta_data = item_level) {
  util_maybe_load_meta_data_v2()
  util_expect_scalar(suppressWarnings, check_type = is.logical)
  util_expect_scalar(mix_jumps_and_missings, check_type = is.logical)
  util_expect_data_frame(meta_data, MISSING_LIST)

  Xcause_label_df <- prep_extract_cause_label_df(meta_data = meta_data,
                                                label_col = VAR_NAMES)

  cause_label_df <- Xcause_label_df$cause_label_df


  cause_label_df$AUTO <- cause_label_df$CODE_LABEL ==
    paste(cause_label_df$CODE_CLASS, cause_label_df$CODE_VALUE)


  if (mix_jumps_and_missings) {
    s_cause_label_df <- split(cause_label_df,
                              list(cause_label_df$CODE_VALUE))
  } else {
    s_cause_label_df <- split(cause_label_df,
                              list(cause_label_df$CODE_VALUE,
                                   cause_label_df$CODE_CLASS))
  }

  expand <- function(cldf) {
    have_auto <- any(cldf$AUTO)
    my_labels <- cldf[!cldf$AUTO, CODE_LABEL, TRUE]
    my_labels <- trimws(my_labels[!is.na(my_labels)])
    if (length(unique(my_labels)) == 1) {
      if (!suppressWarnings && have_auto) {
        util_message("Expand label %s for all values coded with %s",
                     dQuote(unique(my_labels)),
                     dQuote(unique(cldf$CODE_VALUE)),
                     applicability_problem = TRUE)
      }
      cldf$CODE_LABEL <- rep(unique(my_labels), nrow(cldf))
    }
    cldf
  }
  cause_label_df <-
    do.call(rbind.data.frame, lapply(s_cause_label_df, expand))

  meta_data <- prep_add_cause_label_df(Xcause_label_df$meta_data,
                                       cause_label_df,
                                       assume_consistent_codes = TRUE,
                                       replace_meta_data = FALSE)
  meta_data
}

#' Suppress warnings conditionally
#'
#' @param expr [expression] to evaluate
#' @param classes [character] classes of warning-conditions to suppress
#'
#' @returns the result of `expr`
#'
#' @family condition_functions
#' @concept process
#' @noRd
util_suppress_warnings <- function(expr, classes = "warning") {
  withCallingHandlers(
    expr = expr,
    warning = function(cnd) {
      if (inherits(cnd, classes)) {
        tryInvokeRestart("muffleWarning")
      }
    }
  )
}

#' Merge a list of study data frames to one (sparse) study data frame
#'
#' @param study_data_list [list] the list
#'
#' @return [data.frame] [study_data]
#' @export
prep_merge_study_data <- function(study_data_list) {
  # TODO: Implement me!
  # Reduce(merge, study_data_list) # similar to this,
  # see sq2psqlrepository:::getStudyData.sq2psqlRepository
}

#' Calculate and plot `Mahalanobis` distances
#'
#' @description
#' A standard tool to calculate `Mahalanobis` distance.
#' In this approach the squared `Mahalanobis` distance is calculated for ordinal
#' variables (treated as continuous) to identify inattentive responses.
#' It calculates the distance for each observational unit from the sample mean.
#' The greater the distance, the atypical the responses.
#'
#' [Indicator]
#'
#' @details
#' # ALGORITHM OF THIS IMPLEMENTATION:
#' - Implementation is restricted to variables of type integer
#' - Remove missing codes from the study data (if defined in the metadata)
#' - The covariance matrix is estimated for all variables from `variable_group`
#' - The `Mahalanobis` distance of each observation is calculated
#'   \eqn{MD^2_i  = (x_i - \mu)^T \Sigma^{-1} (x_i -  \mu)}
#' - The default to consider a value an `outlier` is to use the 0.975 quantile
#'    of a theoretical chi-square distribution with degrees of freedom
#'    equals to the number of variables used to calculate the
#'    `Mahalanobis` distance (`Mayrhofer and Filzmoser`, 2023)
#'
#' @inheritParams .template_function_indicator
#'
#' @param variable_group [variable list] the names of the variables used to
#'                                        calculate the `Mahalanobis` distance
#' @param study_data [data.frame] the data frame that contains the measurements
#'
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param meta_data_cross_item [data.frame] -- Cross-item level metadata
#' @param label_col [variable attribute] the name of the column in the
#'                                       metadata containing the labels of
#'                                       the variables
#' @param cross_item_level [data.frame] alias for `meta_data_cross_item`
#' @param `cross-item_level` [data.frame] alias for `meta_data_cross_item`
#' @param meta_data_v2 [character] path or file name of the workbook like
#'                                 metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`
#' @param mahalanobis_threshold [numeric] the confidence level to use to define
#'                                        `outliers`, if not stated it is by default
#'                                        0.975.
#'
#' @return a list with:
#'   - `SummaryTable`: [data.frame] underlying the plot
#'   - `SummaryData`: [data.frame] underlying the plot with speaking column labels
#'   - `SummaryPlot`: [ggplot2::ggplot2] Q-Q plot of squared `Mahalanobis`
#'                                        distances vs. a theoretical
#'                                        chi-squared distribution showing `outliers`.
#'   - `FlaggedStudyData`: [data.frame] contains the original data frame of the
#'                                      variables used to calculate
#'                                      the squared `Mahalanobis` distances
#'                                      with the additional column,
#'                                      containing the squared
#'                                       `Mahalanobis` distance, and a column
#'                                       called `MD_outliers`, that contains
#'                                       1 if the observational unit is considered
#'                                       a multivariate `outlier`.
#'
#' @export
#' @importFrom ggplot2 ggplot aes geom_path  scale_color_manual geom_point discrete_scale theme_minimal scale_alpha_manual
#'
#' @importFrom stats mahalanobis
#' @importFrom rlang .data
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_multivariate_outlier.html
#' )
acc_mahalanobis <- function(variable_group = NULL,
                            study_data,
                            item_level = "item_level",
                            meta_data = item_level,
                            meta_data_cross_item = "cross-item_level",
                            label_col = VAR_NAMES,
                            meta_data_v2,
                            cross_item_level,
                            `cross-item_level`,
                            mahalanobis_threshold =
                              suppressWarnings(
                                as.numeric(
                                  getOption("dataquieR.MAHALANOBIS_THRESHOLD",
                                            dataquieR.MAHALANOBIS_THRESHOLD_default)))) {
  #Preps------
  if(.called_in_pipeline) {
    util_error(m = "This function is not meant to run in the pipeline",
               intrinsic_applicability_problem = TRUE
    )
  }

  util_maybe_load_meta_data_v2()

  if (missing(label_col)) {
    orig_label_col <- rlang::missing_arg()
  } else {
    orig_label_col <- force(label_col)
  }

  label_col <- attr(prep_get_labels("",
                                    item_level = meta_data,
                                    label_class = "SHORT",
                                    label_col = label_col),
                    "label_col")
  # Load cross-item_level metadata and normalize it ----
  # check if there is a cross item metadata and if it is a data frame
  # in case is not present, create an empty data frame for cross item metadata
  try(util_expect_data_frame(meta_data_cross_item), silent = TRUE)
  if (!is.data.frame(meta_data_cross_item)) {
    util_message(sprintf(
      "No cross-item level metadata %s found",
      dQuote(meta_data_cross_item)))
    meta_data_cross_item <- data.frame(VARIABLE_LIST = character(0),
                                       CHECK_LABEL = character(0))
  }

  # First normalize input for meta_data_cross_item from the user
  meta_data_cross_item <- util_normalize_cross_item(
    meta_data = meta_data,
    meta_data_cross_item = meta_data_cross_item,
    label_col = label_col
  )
  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = TRUE)


  #Define variable groups
  vars <- NULL
  if(!is.null(variable_group)) {
    util_correct_variable_use("variable_group",
                              allow_more_than_one = TRUE,
                              allow_any_obs_na = TRUE,
                              need_type = "integer | float",
                              need_scale = "interval | ratio | ordinal" #ES: should I leave only ordinal?
    )

    if (length(variable_group) == 1) {
      util_error("Need at least two variables for Mahalanobis distance.",
                 applicability_problem = TRUE)
    }

    vars <- setNames(list(variable_group[!is.na(variable_group)]), "variable_group")



    if (length(mahalanobis_threshold) != 1 ||
        !is.numeric(mahalanobis_threshold) ||
        !is.finite(mahalanobis_threshold)) {
      util_message(
        c("The mahalanobis_threshold argument is not correct",
          "default (%g) is used."),
        dataquieR.MAHALANOBIS_THRESHOLD_default,
        applicability_problem = TRUE)
      mahalanobis_threshold <- dataquieR.MAHALANOBIS_THRESHOLD_default
    }

    mahalanobis_threshold <- setNames(list(mahalanobis_threshold), "variable_group")
  } else {

    given_mahal_cols <- intersect(c("MAHALANOBIS_THRESHOLD", "MAHALANOBIS_RATIO"),
                                  colnames(meta_data_cross_item))
    if (!all(util_empty(as.vector(meta_data_cross_item[, given_mahal_cols,
                                                       FALSE])))) {
      #reduce the cross-item_level content to only the rows that contains mahal. info
      cur_cross <-
        meta_data_cross_item[!util_empty(meta_data_cross_item[[MAHALANOBIS_THRESHOLD]]),
                             , FALSE]
      if (nrow(cur_cross) == 0) {
        vars <- NULL
      } else {
        #Select only the column of interest
        cur_cross <- cur_cross[, intersect(c(VARIABLE_LIST,
                                             CHECK_ID,
                                             CHECK_LABEL,
                                             DATA_PREPARATION,
                                             MAHALANOBIS_THRESHOLD,
                                             MAHALANOBIS_RATIO),
                                           colnames(cur_cross)
        ),
        FALSE]

        cur_cross$MAHALANOBIS_THRESHOLD <- ifelse(
          tolower(as.character(cur_cross$MAHALANOBIS_THRESHOLD)) %in%
            c("true", "1", "t", "+") |
            is.na(suppressWarnings(as.numeric(
              as.character(cur_cross$MAHALANOBIS_THRESHOLD)))),
          dataquieR.MAHALANOBIS_THRESHOLD_default,
          as.numeric(as.character(cur_cross$MAHALANOBIS_THRESHOLD))
        )


        vars <- setNames(lapply(cur_cross$VARIABLE_LIST, function(a) {
          unname(unlist(util_parse_assignments(a)))
        }), cur_cross$CHECK_LABEL)

        mahalanobis_threshold <- as.list(setNames(
          as.numeric(cur_cross$MAHALANOBIS_THRESHOLD),
          cur_cross$CHECK_LABEL
        ))
      }
    }
  }

  #vars is a list
  if (is.null(vars)) {
    util_error(util_error("No variables provided to calculate Mahalanobis distance",
                          applicability_problem = TRUE))
  }

  plot_list <- lapply(setNames(nm = names(vars), vars), function(rv) {
    #Filter to only keep the columns in the variable_group
    ds1_group <- ds1[, rv, drop = FALSE]
    check_label_rv <- names(vars)[sapply(vars, function(x) identical(x, rv))]
    MD_res <- util_generate_mahalanobis_dist(ds1_group,
                                             rv,
                                             check_label_rv)

    ds1_group <- MD_res$x_with_MD
    degree_freedom <- MD_res$df
    rm(MD_res)
    ds1plot <- ds1_group[rowSums(is.na(ds1_group[, rv, drop = FALSE])) == 0, ,
                         drop = FALSE]

    # Threshold to identify multivariate outliers (argument mahalanobis_threshold)
    # is by default 0.975 (Mayrhofer and Filzmoser, 2023) or set by the user
    MD_outliers_threshold <- unname(stats::qchisq(mahalanobis_threshold[[check_label_rv]],
                                                  df = degree_freedom))

    #TODO: Check! This differs from careless function mahad that calculates MD even if there are missing values.

    #create a column with outliers
    ds1plot[[paste0("MD_outliers_", check_label_rv)]] <- NA
    ds1plot[[paste0("MD_outliers_", check_label_rv)]] <-
      ifelse(ds1plot[[paste0("MD_", check_label_rv)]] > MD_outliers_threshold, 1, 0)
    n_non_ol <- sum(ds1plot[[paste0("MD_outliers_", check_label_rv)]] == 0)
    n_devs <- sum(ds1plot[[paste0("MD_outliers_", check_label_rv)]] == 1)


    # Q-Q plot ----
    p_df <- length(ds1plot[[paste0("MD_", check_label_rv)]])
    probabilities <- stats::ppoints(p_df)

    theoretical_quantiles <- stats::qchisq(probabilities,
                                           df = degree_freedom) #fixed, df is not the no. rows, but it is the no. columns
    ds1plot <- ds1plot[order(ds1plot[[paste0("MD_", check_label_rv)]]), ]
    ds1plot$MD_ratio <- ds1plot[[paste0("MD_", check_label_rv)]]/MD_outliers_threshold


    #Utility function to create the plot
    res_MD <- util_create_mahalanobis_ggplot(MD_ratio = ds1plot$MD_ratio,
                                             mahalanobis_threshold = mahalanobis_threshold[[check_label_rv]],
                                             df = degree_freedom)
    p1 <- res_MD$plot_MD

    return(p1)


  })
  ################

  sumdat <- do.call(rbind.data.frame,
                    lapply(setNames(nm = names(vars), vars), function(rv) {
                      #Filter to only keep the columns in the variable_group
                      ds1_group <- ds1[, rv, drop = FALSE]
                      check_label_rv <- names(vars)[sapply(vars,
                                                           function(x) identical(x, rv))]
                      MD_res <- util_generate_mahalanobis_dist(ds1_group,
                                                               rv,
                                                               check_label_rv)
                      ds1_group <- MD_res$x_with_MD
                      degree_freedom <- MD_res$df
                      rm(MD_res)


                      # Threshold to identify multivariate outliers (argument mahalanobis_threshold)
                      # is by default 0.975 (Mayrhofer and Filzmoser, 2023) or set by the user
                      MD_outliers_threshold <-
                        unname(stats::qchisq(mahalanobis_threshold[[check_label_rv]],
                                             df = degree_freedom))

                      #TODO: Check! This differs from careless function mahad that calculates MD even if there are missing values.
                      #create a column with outliers
                      ds1_group[[paste0("MD_outliers_", check_label_rv)]] <- NA
                      ds1_group[[paste0("MD_outliers_", check_label_rv)]] <-
                        ifelse(ds1_group[[paste0("MD_", check_label_rv)]] > MD_outliers_threshold, 1, 0)
                      n_non_ol <- sum(ds1_group[[paste0("MD_outliers_",
                                                        check_label_rv)]] == 0,
                                      na.rm = TRUE)
                      n_devs <- sum(ds1_group[[paste0("MD_outliers_",
                                                      check_label_rv)]] == 1,
                                    na.rm = TRUE)
                      nas_obs_units <- sum(rowSums(is.na(ds1_group)) > 0)
                      nrows_completecases <- nrow(ds1_group)-nas_obs_units
                      # create summary table
                      st1 <- data.frame(Variables = check_label_rv)
                      st1$"MD_outliers (N)" <- n_devs
                      st1$"MD_outliers (%)" <- round(n_devs/nrows_completecases*100,
                                                     digits = 2)
                      st1$"N" <- nrows_completecases
                      st1$"observational_units_removed" <- nrow(ds1_group) -
                        nrows_completecases
                      st1$"mahalanobis_threshold" <-
                        mahalanobis_threshold[[check_label_rv]]
                      SummaryData <- st1
                    }))

  SummaryTable <- sumdat
  names(SummaryTable)[names(SummaryTable) == "MD_outliers (N)"] <- "NUM_ssc_mah"
  names(SummaryTable)[names(SummaryTable) == "MD_outliers (%)"] <- "PCT_ssc_mah"


  FlaggedStudyData <- lapply(setNames(nm = names(vars), vars), function(rv) {
    #Filter to only keep the columns in the variable_group
    ds1_group <- ds1[, rv, drop = FALSE]
    check_label_rv <- names(vars)[sapply(vars, function(x) identical(x, rv))]
    MD_res <- util_generate_mahalanobis_dist(ds1_group,
                                             rv,
                                             check_label_rv)
    ds1_group <- MD_res$x_with_MD
    degree_freedom <- MD_res$df
    rm(MD_res)

    # Threshold to identify multivariate outliers (argument mahalanobis_threshold)
    # is by default 0.975 (Mayrhofer and Filzmoser, 2023) or set by the user
    MD_outliers_threshold <- unname(stats::qchisq(mahalanobis_threshold[[check_label_rv]],
                                                  df = degree_freedom))
    #create a column with outliers
    ds1_group[[paste0("MD_outliers_", check_label_rv)]] <- NA
    ds1_group[[paste0("MD_outliers_", check_label_rv)]] <-
      ifelse(ds1_group[[paste0("MD_", check_label_rv)]] > MD_outliers_threshold, 1, 0)
    ds1_group$row_n <- c(1:nrow(ds1_group))

    return(ds1_group)
  })

  if (length(FlaggedStudyData) == 1) {
    print("The list has exactly one element.")
    FlaggedStudyData_all <- FlaggedStudyData[[1]]
    FlaggedStudyData_all <-
      FlaggedStudyData_all[, names(FlaggedStudyData_all) != "row_n"]

  } else {
    row_counts <- sapply(FlaggedStudyData, nrow)
    if (length(unique(row_counts)) != 1) {
      util_error("Internal error, sorry: The original data frame should have the same number of rows. Please report")
    }
    FlaggedStudyData_all <-  Reduce(function(x, y) {
      extra_cols <- c("row_n", setdiff(colnames(y), colnames(x)))
      merge(x, y[, extra_cols], by = "row_n", all = TRUE)
    }, FlaggedStudyData)


    FlaggedStudyData_all <-
      FlaggedStudyData_all[, names(FlaggedStudyData_all) != "row_n"]

  }

  return(list(SummaryTable = SummaryTable,
              SummaryData = sumdat,
              SummaryPlotList = plot_list,
              FlaggedStudyData = FlaggedStudyData_all
  ))
}

#' Find externally called function in the stack trace
#'
#' intended use: error messages for the user
#'
#' @param sfs reverse [sys.frames] to search in
#' @param cls reverse [sys.calls] to search in
#'
#' @return vector of [logical]s stating for each index, if it had been called
#'          externally
#'
#' @family condition_functions
#' @concept process
#' @noRd
util_find_external_functions_in_stacktrace <-
  function(sfs = rev(sys.frames()),
           cls = rev(sys.calls())) {

  safe_parent_env <- function(...) {
    try(silent = TRUE,
        parent.env(...))
  }

  is_me <-
    vapply(sfs, rlang::env_inherits, ancestor = parent.env(environment()),
           FUN.VALUE = logical(1))

  # frame_parents <- lapply(sfs, safe_parent_env)
  # frame_grand_parents <- lapply(frame_parents, safe_parent_env)
  # frame_great_grand_parents <- lapply(frame_grand_parents, safe_parent_env)
  # is_me <- vapply(frame_parents, identical, parent.env(environment()),
  #                 FUN.VALUE = logical(1))
  # is_me <- is_me | vapply(frame_grand_parents, identical,
  #                         parent.env(environment()), FUN.VALUE = logical(1))
  # is_me <- is_me | vapply(frame_great_grand_parents, identical,
  #                         parent.env(environment()),
  #                         FUN.VALUE = logical(1))
  is_base <- vapply(sfs, identical, asNamespace("base"),
                    FUN.VALUE =
                      logical(1)) # base never calls me, but by using do.call
  if (requireNamespace("parallel", quietly = TRUE))
    is_parallel <- vapply(sfs, rlang::env_inherits,
                          asNamespace("parallel"),
                      FUN.VALUE =
                        logical(1)) # base never calls me, but by using do.call
  else
    is_parallel <- is_base
  is_exception_handler_or_lambda <-
    withCallingHandlers(
    vapply(cls, function(cll) {
        if (length(cll) > 0) {
            if (!is.symbol(cll[[1]]))
              return(TRUE)
            if (cll[[1]] == as.symbol("withCallingHandlers"))
              return(TRUE)
            if (cll[[1]] == as.symbol("doTryCatch"))
              return(TRUE)
            if (cll[[1]] == as.symbol("tryCatch"))
              return(TRUE)
            if (cll[[1]] == as.symbol("tryCatchOne"))
              return(TRUE)
            if (cll[[1]] == as.symbol("tryCatchList"))
              return(TRUE)
            if (cll[[1]] == as.symbol("try"))
              return(TRUE)
            if (cll[[1]] == as.symbol("do.call"))
              return(TRUE)
      }
      return(FALSE)
    }, FUN.VALUE = logical(1)),
  error = browser
  )
  which(!is_me & !is_base & !is_parallel & !is_exception_handler_or_lambda)

}

test_that("util_find_external_functions_in_stacktrace works", {
  skip_on_cran()
  test1 <- function() {
    util_find_external_functions_in_stacktrace()
  }
  environment(test1) <- asNamespace("dataquieR")
  test2 <- function() {
    test1 <- function() {
      util_find_external_functions_in_stacktrace()
    }
    environment(test1) <- asNamespace("dataquieR")
    test1()
  }
  a <- test2()
  expect_equal(
    a[1:1], 3
  )
  environment(test2) <- asNamespace("dataquieR")
  b <- test2()
  expect_equal(
    b, a[-1]
  )
})

test_that("util_find_external_functions_in_stacktrace works from emptyenv", {
  skip_on_cran()
  test1 <- function() {
    util_find_external_functions_in_stacktrace()
  }
  environment(test1) <- asNamespace("dataquieR")
  test2 <- function() {
    test1 <- function() {
      util_find_external_functions_in_stacktrace()
    }
    environment(test1) <- asNamespace("dataquieR")
    test1()
  }
  test3 <- function() {
    test1 <- function() {
      util_find_external_functions_in_stacktrace()
    }
    environment(test1) <- asNamespace("dataquieR")
    test1()
  }
  a <- test2()
  expect_equal(
    a[1:1], 3
  )
  environment(test2) <- asNamespace("dataquieR")
  b <- test2()
  expect_equal(
    b, a[-1]
  )
  test3()
  # Regression test for "the empty environment has no parent"
  test4 <- function() {
    testtest <- function() {
      util_find_external_functions_in_stacktrace()
    }
    environment(testtest) <- asNamespace("dataquieR")
    testtest()
  }
  atestenv <- new.env(parent = emptyenv())
  atestenv$`{` <- `{`
  atestenv$`<-` <- `<-`
  atestenv$`function` <- `function`
  atestenv$asNamespace <- asNamespace
  atestenv$`environment<-` <- `environment<-`
  environment(test4) <- atestenv
  test4()
})

#' Read additional concept tables
#'
#' @param filename RDS-file name without extension to read from
#' @param ... passed to [subset]
#'
#' @return a data frame
#'
#' @family concept_functions
#' @concept reporting
#' @noRd

util_get_concept_info <- function(filename, ...) {
  if (exists(filename, .concept_chache, mode = "list")) {
    dfr <- get(filename, .concept_chache, mode = "list")
  } else {
    f <- system.file(paste0(filename, ".rds"), package = "dataquieR")
    if (0 != file.access(f, 4)) {
      util_error("Cannot read file %s. Internal error.", dQuote(filename))
    }
    dfr <- readRDS(f)
    assign(filename, dfr, .concept_chache)
  }
  cl <- sys.call()
  cl[[1]] <- as.symbol("subset")
  cl[[2]] <- NULL
  cl$x <- dfr
  r <- try(eval(cl,
       envir = parent.frame(),
       enclos = environment()), silent = TRUE)
  if (util_is_try_error(r)) {
    util_warning(r)
    r <- NULL
  }
  r
}
.concept_chache <- new.env(parent = emptyenv())

# Title
#
# @returns
# @seealso https://stackoverflow.com/a/76664004
# @author Alexander Christensen
# util_get_avail_ram <- function() {
#   # Get operating system
#   OS <- tolower(Sys.info()["sysname"])
#
#   # Branch based on OS
#   if (OS == "windows") {
#     # Windows
#
#     # System information
#     system_info <- system("systeminfo", intern = TRUE)
#
#     # Get available memory
#     value <- system_info[grep("Available Physical Memory", system_info)]
#
#     # Remove extraneous information
#     value <- gsub("Available Physical Memory: ", "", value)
#     value <- gsub("\\,", "", value)
#
#     # Convert to bytes
#     value_split <- unlist(strsplit(value, split = " "))
#
#     # Check for second value
#     bytes <- as.numeric(value_split[1]) * switch(value_split[2],
#                                                  "KB" = 1e03,
#                                                  "MB" = 1e06,
#                                                  "GB" = 1e09)
#
#   } else if (OS == "linux") {
#     # Linux
#
#     # Split system information
#     info_split <- strsplit(system("free", intern = TRUE), split = " ")
#
#     # Remove "Mem:" and "Swap:"
#     info_split <- lapply(info_split, function(x) {
#       gsub("Mem:", "", x)
#     })
#     info_split <- lapply(info_split, function(x) {
#       gsub("Swap:", "", x)
#     })
#
#     # Get actual values
#     info_split <- lapply(info_split, function(x) {
#       x[x != ""]
#     })
#
#     # Bind values
#     info_split <- do.call(rbind, info_split[1:2])
#
#     # Get free values
#     bytes <- as.numeric(info_split[2, info_split[1, ] == "free"])
#
#   } else {
#     # Mac
#
#     # System information
#     system_info <- system("top -l 1 -s 0 | grep PhysMem", intern = TRUE)
#
#     # Get everything after comma
#     unused <- gsub(" .*,", "", system_info)
#
#     # Get values only
#     value <- gsub("PhysMem: ", "", unused)
#     value <- gsub(" unused.", "", value)
#
#     # Check for bytes
#     if (grepl("M", value)) {
#       bytes <- as.numeric(gsub("M", "", value)) * 1e06
#     } else if (grepl("G", value)) {
#       bytes <- as.numeric(gsub("G", "", value)) * 1e09
#     } else if (grepl("K", value)) {
#       bytes <- as.numeric(gsub("K", "", value)) * 1e03
#     }
#
#   }
#
#   class(bytes) <- "object_size"
#
#   # Return bytes
#   return(bytes)
#
# }

#' Extract columns of a `SummaryTable` (or Segment, ...)
#'
#' @param Table [data.frame], a table
#'
#' @family summary_functions
#' @noRd
#'
#' @return [data.frame] columns with indicator metrics from `Table`
util_extract_indicator_metrics <- function(Table) {
  # Table <- data.frame(PCT_com_qum_nonresp = 1:10, xxxxx = letters[1:10])
  abbreviationMetrics <- util_get_concept_info("abbreviationMetrics")
  dqi <- util_get_concept_info("dqi")
  cols_for_output <-
    vapply(colnames(Table), FUN.VALUE = character(1), FUN = function(x) {
      util_stop_if_not(length(x) == 1)
      nm <- strsplit(x, "_", fixed = TRUE)[[1]]
      if (length(nm) >= 2) {
        m <- head(subset(abbreviationMetrics, get("Abbreviation") == nm[[1]],
                         "Metrics", drop = TRUE), 1)
        d <- head(subset(dqi, get("abbreviation") == paste(tail(nm, -1),
                                                           collapse = "_"),
                         "Name", drop = TRUE), 1)
        if (length(m) == length(d) && length(d) == 1 &&
            !util_empty(m) && !util_empty(d)) {
          sprintf("%s (%s)", d, m)
        } else {
          NA_character_
        }
      } else {
        NA_character_
      }
    })
  Table[names(cols_for_output[!is.na(cols_for_output)])]
}

#' Get the dataquieR `DATA_TYPE` of `x`
#'
#' @param x object to define the dataquieR data type of
#' @param guess_character [logical] guess a data type for character columns
#'                                  based on the values
#'
#' @return the dataquieR data type as listed in `DATA_TYPES`
#' @seealso [`DATA_TYPES_OF_R_TYPE`]
#' @export
prep_dq_data_type_of <- function(x, guess_character =
                                    getOption("dataquieR.guess_character",
                                              default =
                                              dataquieR.guess_character_default
                                            )) {
  util_expect_scalar(guess_character, check_type = is.logical)
  if (is.list(x) && all(vapply(x, inherits, "hms", FUN.VALUE = logical(1)))) {
    return(tolower(DATA_TYPES_OF_R_TYPE$hms))
  }
  if (is.list(x) && all(vapply(x, inherits, "times", FUN.VALUE = logical(1)))) {
    return(tolower(DATA_TYPES_OF_R_TYPE$times))
  }
  r <- head(unique(intersect(names(DATA_TYPES_OF_R_TYPE), class(x))), 1)

  if (length(r) != 1) {
    NULL
  } else {
    if (r %in% c("numeric", "float", "double")) {
      if (all(util_is_integer(x))) {
        r <- "integer"
      }
    } else if (guess_character && r == "character") {
      r <- switch (readr::guess_parser(x, guess_integer = TRUE),
        integer = "integer",
        character = "character",
        double = "double",
        number = "numeric",
        time = "hms",
        date = "POSIXct",
        datetime = "POSIXct",
        logical = "logical",
        "character" # fallback
      )
      # readr::guess_parser(as.character(1:10), guess_integer = TRUE) # integer
      # readr::guess_parser("asdf", guess_integer = TRUE) # character
      # readr::guess_parser("1.4", guess_integer = TRUE) # double, number
      # readr::guess_parser("12:00:00", guess_integer = TRUE) # time
      # readr::guess_parser("2007-08-07", guess_integer = TRUE) # date
      # readr::guess_parser("2007-08-07 01:01:01", guess_integer = TRUE) # datetime
      # readr::guess_parser("true", guess_integer = TRUE) # logical
      # factor or guess are obviously never returned.
    }

    tolower(DATA_TYPES_OF_R_TYPE[[r]])
  }
}

#' Print a [dataquieR] result returned by [dq_report2]
#' @aliases dataquieR_result
#' @param x [list] a dataquieR result from [dq_report2] or
#'                 `util_eval_to_dataquieR_result`
#' @param ... passed to print. Additionally, the argument `slot` may be passed
#'            to print only specific sub-results.
#' @seealso `util_pretty_print()`
#' @return see print
#' @export
print.dataquieR_result <- function(x, ...) {
  old_opts <- options(
    dataquieR.CONDITIONS_WITH_STACKTRACE = FALSE,
    dataquieR.ERRORS_WITH_CALLER = FALSE,
    dataquieR.WARNINGS_WITH_CALLER = FALSE,
    dataquieR.MESSAGES_WITH_CALLER = FALSE)
  on.exit(old_opts)
  if (length(attr(x, "message")) > 0) {
    for (m in attr(x, "message")) {
      util_message(m)
    }
  }
  if (length(attr(x, "warning")) > 0) {
    for (w in attr(x, "warning")) {
      util_warning(w)
    }
  }
  error_shown <- FALSE
  if (length(attr(x, "error")) > 0) {
    e <- attr(x, "error")[[1]]
    try(util_error(e))
    error_shown <- TRUE
  }
  attr(x, "message") <- NULL
  attr(x, "warning") <- NULL
  attr(x, "error") <- NULL
  if (inherits(x, "empty")) {
    return()
  }
  class(x) <- setdiff(class(x), c("dataquieR_result", "square_result_list"))
  if (inherits(x, "dataquieR_NULL")) {
    x <- NULL
  }
  opts <- list(...)
  if ("slot" %in% names(opts)) {
    if (opts$slot %in% names(x)) {
      if (!("view" %in% names(list(...))) || (!identical(list(...)[["view"]],
                                                      FALSE)))
        print(x[[opts$slot]])
      else
        invisible(x[[opts$slot]])
    } else {
      if (!error_shown) util_error("Cannot find %s in result", opts$slot)
    }
  } else {
    # TODO: switch based on context (knitting, interactive, ...)
    if (error_shown && is.null(x)) {
      return()
    }
    if (!("view" %in% names(list(...))) || (!identical(list(...)[["view"]],
                                                       FALSE))) {
      print(x, ...) # NextMethod()
    } else {
      invisible(x)
    }
  }
}

ALLOWED_DATAQUIER_RESULT_NAMES <- character(0)

(function() {
  prefixes <-
    c("Result", "Dataframe", "Segment", "Summary", "VariableGroup")
  pre_prefixes <-
    c("Modified", "Flagged", "")
  suffixes <-
    c("Table", "Data")
  singleton <-
    c("ReportSummaryTable",
      "SummaryPlot",
      "SummaryPlotList",
      "PlotlyPlot",
      "DataTypePlotList",
      "DataframeDataList",
      "SegmentDataList",
      "VariableGroupPlotList",
      "ModifiedStudyData",
      "FlaggedStudyData",
      "OtherData",
      "OtherTable",
      "Other")
  all_names <- expand.grid(
    pre_prefixes, prefixes, suffixes
  )
  assign("ALLOWED_DATAQUIER_RESULT_NAMES",
         c(apply(all_names, 1, paste0, collapse = ""), singleton),
         parent.frame())
})()

util_is_gg <- function(x) {
  # Handle S7-wrapped dq_lazy_ggplot
  if (dq_lazy_register_s7() &&
      inherits(x, "S7_object") &&
      !is.null(.dq_lazy_state$s7_class) &&
      S7::S7_inherits(x, .dq_lazy_state$s7_class)) {
    return(TRUE)
  }
  return(inherits(x, "gg") || inherits(x, "dq_lazy_ggplot") ||
           inherits(x, "") ||
           inherits(x, "util_pairs_ggplot_panels") ||
           inherits(x, "svg_plot_proxy") ||
           util_is_gg_plot(x))
}

util_is_gg_plot <- function(x) {
  stop(
    paste("Internal error, sorry. Please report! Will be availble later. As",
          "a dataquieR developer: util_is_gg_plot cannot be used during package",
          "load.")
  )
}

# r <- dq_report2("ship", meta_data_v2 = "ship_meta_v2", dimensions = NULL); r2 <- dq_report2("study_data", meta_data_v2 = "meta_data_v2", dimensions = NULL);
# find internal errors: which(vapply(lapply(r2, attr, "error"), inherits, "dataquieR_invalid_result_error", FUN.VALUE = logical(1)))
# see https://nextcloud.uni-greifswald.de/index.php/apps/onlyoffice/23259435?filePath=%2F5230.group_qig%2FTools%2FdataquieR%2FdataquieR%20revised%20metadata%2Fdq_control%20revised%20v3-9_E8.xlsx
util_dataquieR_result <- function(r) {
  if (inherits(r, "dataquieR_NULL") ||
      length(attr(r, "error")) == 1) {
    return(r)
  }
  util_stop_if_not(is.list(r))
  # if (any(vapply(r, is.null, FUN.VALUE = logical(1)))) browser() -- this somehow happened?!
  # r <- r[!vapply(r, is.null, FUN.VALUE = logical(1))]
  # if (length(r) == 0) {
  #   class(r) <- union(class(r), "dataquieR_NULL")
  #   return(r)
  # }
  util_stop_if_not(all(trimws(names(r)) != ""))
  which_not <- names(r)[
    !startsWith(names(r), "ScalarValue_") &
    !names(r) %in% ALLOWED_DATAQUIER_RESULT_NAMES]
  if (length(which_not) > 0) {
    util_error(c("Internal error, sorry. Found an unexpected",
                 "result %s, please report."),
               util_pretty_vector_string(which_not))
  }
  .util_is_data_frame_or_length0 <- function(x) {
    length(x) == 0 || is.data.frame(x)
  }
  util_stop_if_not(all(vapply(r[endsWith(names(r), "Table")],
                              .util_is_data_frame_or_length0, FUN.VALUE = logical(1))))
  TableSlots <- endsWith(names(r), "Table")  &
    vapply(r, .util_is_data_frame_or_length0, FUN.VALUE = logical(1))
  if (any(TableSlots)) {
    for (TableSlot in names(r)[TableSlots]) {
      class(r[[TableSlot]]) <- union("TableSlot", class(r[[TableSlot]]))
      if (inherits(r[[TableSlot]], "ReportSummaryTable")) {
        class(r[[TableSlot]]) <- union("ReportSummaryTable",
                                       class(r[[TableSlot]]))
      }
    }
  }
  DataSlots <- endsWith(names(r), "Data") &
    ! endsWith(names(r), "StudyData") &
    vapply(r, .util_is_data_frame_or_length0, FUN.VALUE = logical(1))
  if (any(DataSlots)) {
    for (DataSlot in names(r)[DataSlots]) {
      class(r[[DataSlot]]) <- union("DataSlot", class(r[[DataSlot]]))
    }
  }
  StudyDataSlots <- endsWith(names(r), "StudyData") &
    vapply(r, .util_is_data_frame_or_length0, FUN.VALUE = logical(1))
  if (any(StudyDataSlots)) {
    for (StudyDataSlot in names(r)[StudyDataSlots]) {
      class(r[[StudyDataSlot]]) <- union("StudyDataSlot",
                                         class(r[[StudyDataSlot]]))
    }
  }
  if ("Other" %in% names(r)) {
    class(r[["Other"]]) <- union("Other", class(r[["Other"]]))
  }
  if ("ReportSummaryTable" %in% names(r)) {
    util_stop_if_not(inherits(r$ReportSummaryTable, "ReportSummaryTable"))
  }
  if ("PlotlyPlot" %in% names(r)) {
    util_stop_if_not(inherits(r$PlotlyPlot, "plotly"))
  }
  if ("SummaryPlot" %in% names(r)) {
    util_stop_if_not(util_is_gg(r$SummaryPlot))
  }
  if ("SummaryPlotList" %in% names(r)) {
    util_stop_if_not(is.list(r$SummaryPlotList))
    util_stop_if_not(all(vapply(r$SummaryPlotList,
                                util_is_gg,
                                FUN.VALUE = logical(1))))
  }
  if ("DataTypePlotList" %in% names(r)) {
    util_stop_if_not(is.list(r$DataTypePlotList))
    util_stop_if_not(all(vapply(r$DataTypePlotList,
                                util_is_gg,
                                FUN.VALUE = logical(1))))
  }
  if ("DataframeDataList" %in% names(r)) {
    util_stop_if_not(is.list(r$DataframeDataList))
  }
  if ("SegmentDataList" %in% names(r)) {
    util_stop_if_not(is.list(r$SegmentDataList))
  }
  if ("VariableGroupPlotList" %in% names(r)) {
    util_stop_if_not(is.list(r$VariableGroupPlotList))
    util_stop_if_not(all(vapply(r$VariableGroupPlotList,
                                util_is_gg,
                                FUN.VALUE = logical(1))))
  }
  if ("ResultTable" %in% names(r)) {
    util_stop_if_not(.util_is_data_frame_or_length0(r$ResultTable))
    util_stop_if_not(length(r$ResultTable) == 0 ||
                       ncol(r$ResultTable) == 0 ||
                       "ResultName" %in% colnames(r$ResultTable))
  }
  if ("ResultData" %in% names(r)) {
    util_stop_if_not(.util_is_data_frame_or_length0(r$ResultData))
  }
  if ("OtherData" %in% names(r)) {
    util_stop_if_not(.util_is_data_frame_or_length0(r$OtherTable))
  }
  if ("OtherTable" %in% names(r)) {
    util_stop_if_not(.util_is_data_frame_or_length0(r$OtherTable))
  }
  if ("DataframeTable" %in% names(r)) {
    util_stop_if_not(.util_is_data_frame_or_length0(r$DataframeTable))
    util_stop_if_not(length(r$DataframeTable) == 0 ||
                       ncol(r$DataframeTable) == 0 ||
                       "DF_NAME" %in% colnames(r$DataframeTable))
  }
  if ("SegmentTable" %in% names(r)) {
    util_stop_if_not(.util_is_data_frame_or_length0(r$SegmentTable))
    util_stop_if_not(length(r$SegmentTable) == 0 ||
                       ncol(r$SegmentTable) == 0 ||
                       "Segment" %in% colnames(r$SegmentTable))
  }
  if ("SummaryTable" %in% names(r)) {
    util_stop_if_not(.util_is_data_frame_or_length0(r$SummaryTable))
    util_stop_if_not(length(r$SummaryTable) == 0 ||
                       ncol(r$SummaryTable) == 0 ||
                       "Variables" %in% colnames(r$SummaryTable))
  }
  if ("ReportSummaryTable" %in% names(r)) {
    util_stop_if_not(.util_is_data_frame_or_length0(r$ReportSummaryTable))
    util_stop_if_not(length(r$ReportSummaryTable) == 0 ||
                       ncol(r$ReportSummaryTable) == 0 ||
                       "Variables" %in% colnames(r$ReportSummaryTable))
    util_stop_if_not(length(r$ReportSummaryTable) == 0 ||
                       ncol(r$ReportSummaryTable) == 0 ||
                       "N" %in% colnames(r$ReportSummaryTable))
  }
  if ("VariableGroupTable" %in% names(r)) {
    util_stop_if_not(.util_is_data_frame_or_length0(r$VariableGroupTable))
    util_stop_if_not(length(r$VariableGroupTable) == 0 ||
                       ncol(r$VariableGroupTable) == 0 ||
                       "VARIABLE_LIST" %in% colnames(r$VariableGroupTable))
  }
  class(r) <- union(c("dataquieR_result", "master_result"), class(r))
  r
}

#' Print a `StudyDataSlot` object
#'
#' @param x the object
#' @param ... not used
#'
#' @return see print
#' @export
print.StudyDataSlot <- function(x, ...) {
  util_ensure_suggested("tibble")
  r <- tibble::as_tibble(x)
  if (!("view" %in% names(list(...))) || (!identical(list(...)[["view"]],
                                                  FALSE)))
    print(r)
  else
    invisible(r)
}

#' Print a `DataSlot` object
#'
#' @param x the object
#' @param ... not used
#'
#' @return see print
#' @export
print.DataSlot <- function(x, ...) {
  util_ensure_suggested("htmltools")
  r <- util_html_table(x)
  if (!is.null(r)) r <- htmltools::browsable(r)
  if (!("view" %in% names(list(...))) || (!identical(list(...)[["view"]],
                                                  FALSE))) {
    if (isTRUE(getOption('knitr.in.progress'))) {
      util_ensure_suggested("knitr", "knit-print")
      class(x) <- setdiff(class(x), "DataSlot")
      knitr::knit_print(x)
    } else {
      print(r)
    }
  } else
    invisible(r)
}

#' Print a `TableSlot` object
#'
#' @param x the object
#' @param ... not used
#'
#' @return see print
#' @export
print.TableSlot <- function(x, ...) {
  util_ensure_suggested("htmltools")
  r <- util_make_data_slot_from_table_slot(x)
  r <- util_html_table(r)
  if (!is.null(r)) r <- htmltools::browsable(r)
  if (!("view" %in% names(list(...))) || (!identical(list(...)[["view"]],
                                                     FALSE)))
    print(r)
  else
    invisible(r)
}

#' Print a `master_result` object
#'
#' @param x the object
#' @param template the template for the `iframes`, not used, so far.
#' @param ... not used
#'
#' @return `invisible(NULL)`
#' @export
print.master_result <- function(x, template = "default", ...) {
  template <- "default"
  util_ensure_suggested("htmltools")
  if (isTRUE(getOption('knitr.in.progress'))) {
    f <- withr::local_tempdir(.local_envir = knitr::knit_global())
  } else {
    f <- withr::local_tempdir(.local_envir = rlang::global_env())
  }
  withr::local_dir(f)
  jqui <- rmarkdown::html_dependency_jqueryui()
  jqui$stylesheet <- "jquery-ui.min.css"
  if (!is.null(attr(x, "function_name")) &&
      attr(x, "function_name") %in%
      c("con_limit_deviations",  # check if we are working with a limits function
        "con_hard_limits",
        "con_soft_limits",
        "con_detection_limits")) {
    # FIXME: Remove special treatment of con_limit_deviations in favor of suitable result slots in DQ_OBS
    x$ReportSummaryTable <- NULL
  }
  cnt <- util_pretty_print(dqr = x, nm = attr(x, "cn"),
                           is_single_var = FALSE,
                           use_plot_ly = util_ensure_suggested("plotly", "plot interactive figures", err = FALSE),
                           dir = f,
                           ...)
  util_write_iframe_results(pages = cnt,
                            progress_msg = function(...) {},
                            progress = function(...) {},
                            template_file = system.file("templates",
                                                        template,
                                                        "iframe.html",
                                                        package =
                                                          packageName()),
                            dir = f)
  doc <- htmltools::tagList(rmarkdown::html_dependency_jquery(),
                            html_dependency_tippy(),
                            html_dependency_clipboard(),
                            html_dependency_dataquieR(iframe = FALSE),
                            html_dependency_jspdf(),
                            jqui,
                            htmltools::div(class = "navbar"),
                     cnt,
                     htmltools::tags$script('window.dataquieR_single_result = true ; $(function(){$("body").css("overflow", ""); $(".navbar").hide(); $(".default-target").height("1em");})'),
                     # htmltools::tags$script('    setTimeout(function() {
                     # debugger
                     #                        window.dispatchEvent(new Event("resize")) }, 500)')
                     )

  deps_prepro <- util_copy_all_deps(dir = f,
                                    doc,
                                    rmarkdown::html_dependency_jquery(),
                                    jqui,
                                    html_dependency_clipboard(),
                                    html_dependency_tippy(),
                                    rmarkdown::html_dependency_font_awesome(),
                                    html_dependency_dataquieR(),
                                    html_dependency_jspdf()
  )

  html_result <- htmltools::htmlTemplate(system.file("templates",
                                                     template,
                                                     "report.html",
                                                     package =
                                                       packageName()),
                                         document_ = TRUE,
                                         by_report = FALSE,
                                         spage = doc,
                                         logo = NULL,
                                         menu = NULL,
                                         loading = NULL,
                                         deps = deps_prepro$deps,
                                         title = attr(x, "nm"),
                                         backlink = NULL,
                                         header = NULL)

  f <- file(description = "index.html", open = "w", encoding = "utf-8")
  on.exit(close(f))

  cat(as.character(html_result), file = f)

  if (!("view" %in% names(list(...))) || (!identical(list(...)[["view"]],
                                                  FALSE))) {
    if (isTRUE(getOption('knitr.in.progress'))) {
      util_ensure_suggested("htmlwidgets", "Render results in RMarkdown")
      #sp <- htmlwidgets::sizingPolicy(fill = TRUE)

      util_ensure_suggested("knitr", "knit-print")
      if (knitr::is_latex_output()) { # TODO: LaTeX
        util_warning(c("%s in R markdown not yet supported by %s for printing",
                       "full results"),
                      sQuote(knitr::pandoc_to()), sQuote(packageName()))
        return("")
      } else if (knitr::is_html_output()) {
        return(knitr::knit_print(
          statichtmlWidget(doc,
                           js =
                             '$(function(){$("body").css("overflow", ""); })')))
      } else { # TODO: Word, ...?
        # knitr::opts_chunk$get("height")
        util_warning(c("%s in R markdown not yet supported by %s for printing",
                       "full results"),
                     sQuote(knitr::pandoc_to()), sQuote(packageName()))
        return("")
      }
    } else {
      viewer <- getOption("viewer", utils::browseURL)
      viewer("index.html")
    }
  }
  invisible(NULL)
}


# #' exportS3Method knitr::knit_print
# knit_print.Slot <- function(x, ...) {
#   util_ensure_suggested("knitr", "Call knit_print")
#   knitr::asis_output("***TEST***")
# }

#' Print a `Slot` object
#'
#' displays all warnings and stuff. then it prints `x`.
#'
#' @param x the object
#' @param ... not used
#'
#' @return calls the next print method
#' @export
print.Slot <- function(x, ...) {
  if (any(inherits(x, c("Other", "ReportSummaryTable")))) {
    return(NextMethod())
  }
  if (length(attr(x, "message")) > 0) {
    for (m in attr(x, "message")) {
      util_message(m)
    }
  }
  if (length(attr(x, "warning")) > 0) {
    for (w in attr(x, "warning")) {
      util_warning(w)
    }
  }
  error_shown <- FALSE
  if (length(attr(x, "error")) > 0) {
    e <- attr(x, "error")[[1]]
    try(util_error(e))
    error_shown <- TRUE
  }
  attr(x, "message") <- NULL
  attr(x, "warning") <- NULL
  attr(x, "error") <- NULL
  withr::with_pdf(NULL,
                  o <- capture.output(rr <- withVisible(NextMethod())))
  r <- rr$value
  v <- rr$visible
  if (!is.null(r))
    class(r) <- setdiff(class(r), "Slot")
  if ((v || !("view" %in% names(list(...))) || (!identical(list(...)[["view"]],
                                                     FALSE)))) {
    print(r)
    # if (any(nzchar(o))) {
    #   cat(o, sep = "\n")
    # }
  } else {
    invisible(r)
  }
}

#' Filter collection based on its `names()` using regular expressions
#'
#' @param collection a named collection (list, vector, ...)
#' @param regexps [character] a vector of regular expressions
#' @param negate [logical] Invert the meaning of the regular expressions, so
#'                         keep only entries, that do not match any of the
#'                         `regexps`.
#'
#' @return `collection` reduced to entries, that's names match at least any
#'         expression from `regexps`
#' @examples
#' \dontrun{ # internal function
#' util_filter_names_by_regexps(iris, c("epa", "eta"))
#' }
#'
#' @family string_functions
#' @concept process
#' @noRd
util_filter_names_by_regexps <- function(collection,
                                         regexps,
                                         negate = FALSE) {

  if (is.null(names(collection))) {
    util_error("Need names in argument %s", sQuote("collection"))
  }

  util_expect_scalar(regexps,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)

  util_expect_scalar(negate, check_type = is.logical)

  if (length(regexps) == 0 || length(collection) == 0) {
    return(collection)
  }

  matches_mat <-
    matrix(vapply(
      regexps,
      grepl,
      names(collection),
      perl = TRUE,
      FUN.VALUE =
        logical(length(collection))
    ), nrow = length(collection), ncol = length(regexps))

  if (is.null(dim(matches_mat))) { # TODO: is this ever executed?
    util_stop_if_not(length(regexps) == 1 && length(collection) == 1)
    if (length(regexps) > 1) { # TODO: ever executed?
      # only one long row of match?(T/F)-entries
      #             RE1 RE2 RE3 RE4 RE5
      # acc_margins T   F   F   T   T
      if (negate) {
        if (all(!as.vector(matches_mat), na.rm = TRUE)) {
          r <- collection
          nm <- names(collection)
        } else {
          r <- character(0)
          nm <- character(0)
        }
      } else {
        if (any(as.vector(matches_mat), na.rm = TRUE)) {
          r <- collection
          nm <- names(collection)
        } else {
          r <- character(0)
          nm <- character(0)
        }
      }
    } else {
      # only one long column of match?(T/F)-entries
      #               RE1
      # acc_margins   T
      # com_item_miss F
      # con_limit_dev F
      if (negate) {
        matches_mat <- ! matches_mat
      }
      r <- collection[matches_mat]
      nm <- names(collection[matches_mat])
    }
  } else {
    if (negate) {
      r <- collection[rowSums(matches_mat) == 0]
      nm <- names(collection[rowSums(matches_mat) == 0])
    } else {
      r <- collection[rowSums(matches_mat) > 0]
      nm <- names(collection[rowSums(matches_mat) > 0])
    }
  }

  mostattributes(r) <- attributes(collection)
  names(r) <- nm

  r
}

test_that("util_filter_names_regexps works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  # selecting functions I want to keep with regexp----
  expect_equal(util_filter_names_by_regexps(
    setNames(nm = c("acc_margins", "com_item_miss", "con_limit_dev")),
    c("^acc_", "^con_")),
    setNames(nm = c("acc_margins", "con_limit_dev" )))

  expect_equal(util_filter_names_by_regexps(
    setNames(nm = c("acc_margins", "com_item_miss", "con_limit_dev")),
    c("^acc_")),
    setNames(nm = c("acc_margins")))

  expect_equal(util_filter_names_by_regexps(
    setNames(nm = c("acc_margins", "com_item_miss", "con_limit_dev")),
    c()),
    setNames(nm = c("acc_margins", "com_item_miss", "con_limit_dev")))

  expect_error(util_filter_names_by_regexps(
    setNames(nm = c()), c("^acc_", "^con_")),
    regexp = paste(".*Need names.*"),
    perl = TRUE)

  expect_identical(util_filter_names_by_regexps(
    setNames(nm = character(0)),
    c("^acc_", "^con_")),
    setNames(character(0), character(0)))

  expect_equal(util_filter_names_by_regexps(
    setNames(nm = c("acc_margins")),
    c("^acc_", "^con_")),
    setNames(nm = c("acc_margins")))

  expect_identical(util_filter_names_by_regexps(
    setNames(nm = c("acc_margins")),
    c("^xacc_", "^con_")),
    setNames(character(0), character(0)))

  # excluding functions I do not want to keep with regexp----
  expect_equal(util_filter_names_by_regexps(
    setNames(nm = c("acc_margins", "com_item_miss", "con_limit_dev")),
    c("^acc_", "^con_"),
    negate = TRUE),
    setNames(nm = c("com_item_miss")))

  expect_equal(util_filter_names_by_regexps(
    setNames(nm = c("acc_margins", "com_item_miss", "con_limit_dev")),
    c("^acc_"),
    negate = T),
    setNames(nm = c("com_item_miss", "con_limit_dev")))

  expect_equal(util_filter_names_by_regexps(
    setNames(nm = c("acc_margins", "com_item_miss", "con_limit_dev")),
    c(),
    negate = T),
    setNames(nm = c("acc_margins", "com_item_miss", "con_limit_dev")))

  expect_error(util_filter_names_by_regexps(
    setNames(nm = c()),
    c("^acc_", "^con_"),
    negate = TRUE),
    regexp = paste(".*Need names.*"),
    perl = TRUE)

  expect_identical(util_filter_names_by_regexps(
    setNames(nm = character(0)),
    c("^acc_", "^con_"),
    negate = TRUE),
    setNames(character(0), character(0)))

  expect_equal(util_filter_names_by_regexps(
    setNames(nm = c("acc_margins")),
    c("^acc_", "^con_"),
    negate = T),
    setNames(character(0), character(0)))


})

#' All valid `voc:` vocabularies
#'
#' @return [character()] all `voc:` suffixes allowed for
#'                       [prep_get_data_frame()].
#' @export
#'
#' @examples
#' \dontrun{
#' prep_list_dataframes()
#' prep_list_voc()
#' prep_get_data_frame("<ICD10>")
#' my_voc <-
#'   tibble::tribble(
#'     ~ voc, ~ url,
#'     "test", "data:datasets|iris|Species+Sepal.Length")
#' prep_add_data_frames(`<>` = my_voc)
#' prep_list_dataframes()
#' prep_list_voc()
#' prep_get_data_frame("<test>")
#' prep_get_data_frame("<ICD10>")
#' my_voc <-
#'   tibble::tribble(
#'     ~ voc, ~ url,
#'     "ICD10", "data:datasets|iris|Species+Sepal.Length")
#' prep_add_data_frames(`<>` = my_voc)
#' prep_list_dataframes()
#' prep_list_voc()
#' prep_get_data_frame("<ICD10>")
#' }
#'
prep_list_voc <- function() { # TODO: Also the custom tables
  voc_tab <- util_get_voc_tab()
  # table lookup to modify fn to package: or data:
  try(subset(voc_tab, select = "voc", drop = TRUE))
}

#' escape `"`
#'
#' @param s haystack
#'
#' @return `s` with `"` replaced by `&quot;`
#'
#' @concept reporting
#' @noRd

util_html_attr_quote_escape <- function(s) {
  gsub('"', "&quot;", s, fixed = TRUE)
}

test_that("util_html_attr_quote_escape escapes quotes properly", {
  skip_on_cran()
  input_string <- 'Hello "World"!'
  expected_output <- 'Hello &quot;World&quot;!'
  actual_output <- util_html_attr_quote_escape(input_string)
  expect_equal(actual_output, expected_output)
})

#' Calculate and plot Mahalanobis distances
#'
#' @description
#' A standard tool to detect multivariate outliers is the Mahalanobis distance.
#' This approach is very helpful for the interpretation of the plausibility of a
#' measurement given the value of another.
#' In this approach the Mahalanobis distance is used as a univariate measure
#' itself. We apply the same rules for the identification of outliers as in
#' univariate outliers:
#' - the classical approach from Tukey: \eqn{1.5 * IQR} from the
#'   1st (\eqn{Q_{25}}) or 3rd (\eqn{Q_{75}}) quartile.
#' - the 3SD approach, i.e. any measurement of the Mahalanobis
#'   distance not in the interval of \eqn{\bar{x} \pm 3*\sigma} is considered an
#'   outlier.
#' - the approach from Hubert for skewed distributions which is embedded in the
#'   R package \pkg{robustbase}
#' - a completely heuristic approach named \eqn{\sigma}-gap.
#'
#' For further details, please see the vignette for univariate outlier.
#'
#' [Indicator]
#'
#' @details
#' # ALGORITHM OF THIS IMPLEMENTATION:
#' - Implementation is restricted to variables of type float
#' - Remove missing codes from the study data (if defined in the metadata)
#' - The covariance matrix is estimated for all variables from `variable_group`
#' - The Mahalanobis distance of each observation is calculated
#'   \eqn{MD^2_i  = (x_i - \mu)^T \Sigma^{-1} (x_i -  \mu)}
#' - The four rules mentioned above are applied on this distance for
#'   each observation in the study data
#' - An output data frame is generated that flags each outlier
#' - A parallel coordinate plot indicates respective outliers
#'
#' List function.
#'
#' @inheritParams .template_function_indicator
#'
#' @param variable_group [variable list] the names of the continuous
#'                                       measurement variables building
#'                                       a group, for that multivariate outliers
#'                                       make sense.
#' @param id_vars [variable] optional, an ID variable of
#'                        the study data. If not specified row numbers are used.
#' @param n_rules [numeric] from=1 to=4. the no. of rules that must be violated
#'                                       to classify as outlier
#' @param max_non_outliers_plot [integer] from=0. Maximum number of non-outlier
#'                                                points to be plot. If more
#'                                                points exist, a subsample will
#'                                                be plotted only. Note, that
#'                                                sampling is not deterministic.
#' @param criteria [set] tukey | 3SD | hubert | sigmagap. a vector with
#'                       methods to be used for detecting outliers.
#' @param scale [logical] Should min-max-scaling be applied per variable?
#' @param multivariate_outlier_check [logical] really check, pipeline use,
#'                                              only.
#'
#' @return a list with:
#'   - `SummaryTable`: [data.frame] underlying the plot
#'   - `SummaryPlot`: [ggplot2::ggplot2] outlier plot
#'   - `FlaggedStudyData` [data.frame] contains the original data frame with
#'                                     the additional columns `tukey`,
#'                                     `3SD`,
#'                                     `hubert`, and `sigmagap`. Every
#'                                     observation
#'                                     is coded 0 if no outlier was detected in
#'                                     the respective column and 1 if an
#'                                     outlier was detected. This can be used
#'                                     to exclude observations with outliers.
#'
#' @export
#' @importFrom ggplot2 ggplot aes geom_path  scale_color_manual geom_point discrete_scale theme_minimal scale_alpha_manual
#'
#' @importFrom stats mahalanobis
#' @importFrom rlang .data
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_multivariate_outlier.html
#' )
acc_multivariate_outlier <- function(variable_group = NULL,
                                     id_vars = NULL,
                                     label_col = VAR_NAMES,
                                     study_data,
                                     item_level = "item_level",
                                     n_rules = 4,
                                     max_non_outliers_plot = 10000,
                                     criteria = c("tukey", "3sd",
                                                  "hubert", "sigmagap"),
                                     meta_data = item_level,
                                     meta_data_v2,
                                     scale = getOption("dataquieR.acc_multivariate_outlier.scale",
                                                       dataquieR.acc_multivariate_outlier.scale_default),
                                     multivariate_outlier_check = TRUE) { # TODO: see univ.out to select criteria to use

  # preps ----------------------------------------------------------------------
   util_expect_scalar(multivariate_outlier_check,
                     check_type = is.logical)
  if (.called_in_pipeline && !multivariate_outlier_check) {
    util_error("No multivariate outliers requested",
               intrinsic_applicability_problem = TRUE)
  }

  util_maybe_load_meta_data_v2()

  if (missing(label_col)) {
    orig_label_col <- rlang::missing_arg()
  } else {
    orig_label_col <- force(label_col)
  }

  label_col <- attr(prep_get_labels("",
                                    item_level = meta_data,
                                    label_class = "SHORT",
                                    label_col = label_col),
                    "label_col")

  #campatibility with previous name (sixsigma)
  #replace "sixsigma" (if any attributed in criteria) with 3SD
  old_name<- "sixsigma"
  if(any(trimws(tolower(criteria)) %in% old_name)== TRUE){
    criteria[criteria=="sixsigma"] <- "3sd"
  }

  #all lowercase
  criteria <- trimws(tolower(criteria))


  util_expect_scalar(criteria,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)


  if (length(unique(criteria)) < 1 ||
      length(unique(criteria)) >
      length(eval(formals(acc_multivariate_outlier)$criteria)) ||
      !all(criteria %in% eval(formals(acc_multivariate_outlier)$criteria))) {
    if (!.called_in_pipeline)
      util_message(c("The formal criteria must have > 0 and < %d entries.",
                   "Allowed values are %s.",
                   "I was called with %s, falling back to default %s."),
                 length(eval(formals(acc_multivariate_outlier)$criteria)),
                 paste(dQuote(eval(formals(acc_multivariate_outlier)$criteria)),
                       collapse = ", "),
                 paste(dQuote(unique(criteria)), collapse = ", "),
                 paste(dQuote(eval(formals(acc_multivariate_outlier)$criteria)),
                       collapse = ", "), applicability_problem = TRUE)
    criteria <- eval(formals(acc_multivariate_outlier)$criteria)
  }

  if (length(n_rules) != 1 || !is.numeric(n_rules) ||
      !all(util_is_integer(n_rules)) ||
      !(n_rules %in% seq_len(length(unique(criteria))))) {
    if ((!.called_in_pipeline))
      util_message(
      "The formal n_rules is not an integer between 1 and %d, default (%d) is used.",
      length(unique(criteria)),
      min(eval(formals(acc_multivariate_outlier)$n_rules),
          length(unique(criteria))),
      applicability_problem = TRUE)
    n_rules <- min(eval(formals(acc_multivariate_outlier)$n_rules),
                   length(unique(criteria)))

  }
    # rules in 1:4?
  if (n_rules != 4) {
    if (!(n_rules %in% 1:4)) {
      util_message(
        "The formal n_rules is not an integer of 1 to 4, default (4) is used. ",
        applicability_problem = TRUE)
      n_rules <- 4
    }
  }

  if (length(max_non_outliers_plot) != 1 ||
      !is.numeric(max_non_outliers_plot) ||
      !all(util_is_integer(max_non_outliers_plot)) ||
      (max_non_outliers_plot < 0)) {
    util_message(
      c("The formal max_non_outliers_plot is not an integer >= 0,",
        "default (%d) is used."),
      formals(acc_multivariate_outlier)$max_non_outliers_plot,
      applicability_problem = TRUE)
    max_non_outliers_plot <-
      formals(acc_multivariate_outlier)$max_non_outliers_plot
  }

  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = TRUE)

#  needed_scale <- paste0("interval", SPLIT_CHAR, "ratio")
#  if (COMPUTED_VARIABLE_ROLE %in% colnames(meta_data)) {
 #   if (#one of the variables inside variable_group   ??
 #     %in% meta_data$COMPUTED_VARIABLE_ROLE) {
 #     needed_scale <- paste0(needed_scale, SPLIT_CHAR, "ordinal")
 #   }
#  }


  util_correct_variable_use("variable_group",
    allow_more_than_one = TRUE,
    allow_any_obs_na = TRUE,
    need_type = "integer | float",
    need_scale = "interval | ratio"
  )




  if (length(variable_group) == 1) {
    util_error("Need at least two variables for multivariate outliers.",
               applicability_problem = TRUE)
  }

  # check correct use of ID-variable if specified
  util_correct_variable_use("id_vars",
    allow_any_obs_na = TRUE,
    allow_null = TRUE,
    need_type = "string | integer",
    need_scale = "nominal | ordinal | na"
  )

  # no use of id_vars?
  if (!length(id_vars)) {
    ds1$dq_id <- rownames(ds1)
    id_vars <- "dq_id"
    util_message("As no ID-var has been specified the rownumbers will be used.",
                 applicability_problem = TRUE)
  }

  if (length(id_vars) > 1) {
    id_vars <- id_vars[1]
  }

  # missing data in any of the variables?
  vars <- c(variable_group, id_vars)
  vars <- vars[!is.na(vars)]

  n_prior <- dim(ds1)[1]
  ds1plot <- ds1[rowSums(is.na(ds1[, vars, drop = FALSE])) == 0, ,
    drop = FALSE
  ]
  n_post <- dim(ds1plot)[1]

  if (n_post == 0) {
    util_error("No samples with without missing values in one of %s. Aborting.",
               paste0(dQuote(vars), collapse = ", "),
               applicability_problem = FALSE)
  }

  if (n_post < n_prior) {
    util_message(paste0(
      "Due to missing values in ", paste0(variable_group, collapse = ", "),
      " or ", id_vars, " N=", n_prior - n_post,
      " observations were excluded."
    ), applicability_problem = FALSE)
  }

  variable_group <-
    util_no_value_labels(
      resp_vars = variable_group,
      meta_data = meta_data,
      label_col = label_col,
      warn = TRUE,
      stop = TRUE
    )

  if (length(variable_group) < 2) {
    util_error(
      "Fewer than two variables left, no multivariate analysis possible.",
      applicability_problem = TRUE)
  }

  # Mahalanobis ----------------------------------------------------------------

  # Estimate covariance of response variables
  Sx <- cov(ds1plot[, variable_group, drop = FALSE])

  # Calculation of the Mahalanobis distance
  ds1plot$MD <- mahalanobis(ds1plot[, variable_group], colMeans(ds1plot[, variable_group,
                                                             drop = FALSE]), Sx)

  # outlier identification
  # Initialize with NA
  ds1plot$tukey <- NA
  ds1plot$threeSD <- NA
  ds1plot$hubert <- NA
  ds1plot$sigmagap <- NA
  # View(ds1)
  # apply outlier functions to plot-df
  # after export/final built  correct the call of the utility functions
  ds1plot$tukey <- util_tukey(ds1plot$MD)
  ds1plot$threeSD <- util_3SD(ds1plot$MD)
  ds1plot$hubert <- util_hubert(ds1plot$MD)
  ds1plot$sigmagap <- util_sigmagap(ds1plot$MD)

  #Fix the problem with name 3SD starting with a number replacing it with threeSD---
  orig_name<- "3sd"
  if(any(criteria %in% orig_name)== TRUE){
    criteria[criteria=="3sd"] <- "threeSD"
  }

  # calculate summary of all outlier definitions
  ds1plot$Rules <-
    apply(ds1plot[, criteria, drop = FALSE], 1, sum)

  n_non_ol <- sum(ds1plot$Rules == 0)

  ### remove non-outliers, if too many
  if (max_non_outliers_plot < n_non_ol) {

    dsi_non_ol <- ds1plot[ds1plot$Rules == 0, , FALSE]
    dsi_ol <- ds1plot[ds1plot$Rules > 0, , FALSE]

    subsel_non_ol <- sample(seq_len(nrow(dsi_non_ol)),
                            size =
                              min(max_non_outliers_plot, nrow(dsi_non_ol)))

    ds1plot <- rbind.data.frame(dsi_non_ol[subsel_non_ol, , FALSE], dsi_ol)

    util_message(
      c("For %s, %d from %d non-outlier data values were",
        "sampled to avoid large plots."),
      sQuote(sprintf("acc_multivariate_outlier(%s)",
                     paste0(dQuote(variable_group), collapse = ", "))),
      max_non_outliers_plot,
      n_non_ol,
      applicability_problem = FALSE
    )
  }
  ###

  # create summary table
  # outlier deviating according to all rules?
  n_devs <- ifelse(ds1plot$Rules >= n_rules, 1, 0)
  grading <- max(n_devs, na.rm = TRUE)

  st1 <- data.frame(Variables = paste0(variable_group, collapse = " | "))
  st1$"Tukey (N)" <- sum(ds1plot$tukey)
  st1$"3SD (N)" <-  sum(ds1plot$threeSD)
  st1$"Hubert (N)" <- sum(ds1plot$hubert)
  st1$"Sigma-gap (N)" <- sum(ds1plot$sigmagap)
  st1$"Outliers (N)" <- sum(ds1plot$Rules >= n_rules)
  st1$"Outliers (%)" <- round(sum(ds1plot$Rules >= n_rules)/length(ds1plot$MD)*100, digits = 2)
  st1$GRADING <- grading

  SummaryTable <- st1
  names(SummaryTable)[names(SummaryTable) == "Outliers (N)"] <- 'NUM_acc_ud_outlm'
  names(SummaryTable)[names(SummaryTable) == 'Outliers (%)'] <- 'PCT_acc_ud_outlm'

  # reshape wide to long
#  ds2plot <-
#    melt(ds1plot[, c(variable_group, id_vars, "Rules")], measure.vars = variable_group, id.vars =
#           c(id_vars, "Rules"))
  ds2plot <-
    stats::reshape(data = ds1plot[, c(variable_group, id_vars, "Rules")],
                   idvar =  c(id_vars, "Rules"),
                   varying =
                     colnames(ds1plot[,
                                      c(variable_group,
                                        id_vars,
                                        "Rules")])[!(colnames(ds1plot[,
                                                                      c(variable_group,
                                                                        id_vars, "Rules")]) %in%
                                                       c(id_vars, "Rules"))],
                   v.names = "value",
                   times =
                     colnames(ds1plot[, c(variable_group,
                                          id_vars,
                                          "Rules")])[!(colnames(ds1plot[,
                                                                        c(variable_group,
                                                                          id_vars,
                                                                          "Rules")]) %in%
                                                         c(id_vars, "Rules"))] ,
                               direction = "long")
  lev <- variable_group
  ds2plot$time <- factor(ds2plot$time, levels = lev, ordered = FALSE)
  rownames(ds2plot) <- NULL
  names(ds2plot)[names(ds2plot) == "time"] <- "variable"
  ds2plot$orig_value <- ds2plot$value

  # apply min-max-scaling if selected
  if (scale) {
    ds2plot <- ds2plot %>%
      dplyr::group_by(variable) %>%
      dplyr::mutate(value = (value - min(value)) / (max(value) - min(value))) %>%
      dplyr::ungroup()
  }

  # use named color vector
  disc_cols <- c("#2166AC", "#fdd49e", "#fc8d59", "#d7301f", "#7f0000")
  names(disc_cols) <- c(0:4)

  # use ID/Rules as factor
  ds2plot[[id_vars]] <- factor(ds2plot[[id_vars]])
  ds2plot$Rules <- factor(ds2plot$Rules, ordered = TRUE)


  # transparency
  .a <- c(0.2, 0.7, 0.8, 0.9, 1)
  .a <- .a[seq_along(levels(ds2plot$Rules))]
  names(.a) <- levels(ds2plot$Rules)

  # size
  .s <- c(0.05, 0.2, 0.3, 0.4, 0.5)
  .s <- .s[seq_along(levels(ds2plot$Rules))]
  names(.s) <- levels(ds2plot$Rules)

  # PLOT
  p <- util_create_lean_ggplot(
    ggplot(ds2plot, aes(x = factor(variable, levels = variable_group),
                        y = value, colour = Rules,
                        group = .data[[id_vars]],
                        label = .data$orig_value)) +
      geom_path(aes(alpha = Rules,
                    linewidth = Rules),
                position = "identity") +
      scale_color_manual(values = disc_cols) +
      geom_point(aes(alpha = Rules)) +
      scale_alpha_manual(values = .a) +
      discrete_scale("linewidth", # deprecated since ggplot 3.6.0: "outlier_rules_scale",
                     palette = function(n) {
                       c(0.05, 0.2, 0.3, 0.4, 0.5)
                     }) +
      xlab("") + ylab("") +
      theme_minimal(),
    ds2plot = ds2plot,
    variable_group = variable_group,
    id_vars = id_vars,
    disc_cols = disc_cols,
    .a = .a)

  if (scale) {
    p <- p + util_create_lean_ggplot(
                                   scale_y_continuous(breaks = c(0, 1),
                                                      labels = c("min", "max")))
  }

  # Information for sizing
  number_of_vars <- length(unique(ds2plot$variable))
  min_value <- min(ds2plot$value, na.rm = TRUE)
  max_value <- max(ds2plot$value, na.rm = TRUE)
  range <- max_value - min_value
  no_char_y <- nchar(max_value)

  #Sizing information - a default
  attr(p, "sizing_hints") <- list(
    figure_type_id = "multivar_plot",
    number_of_vars = number_of_vars,
    range = range,
    no_char_y = no_char_y
  )
  rm(min_value, max_value, range, number_of_vars, no_char_y)


  SummaryData <- st1

  # # fix Variables column in SummaryData
  # vs <- lapply(util_parse_assignments(SummaryData$Variables,
  #                               multi_variate_text = TRUE),
  #        prep_get_labels,
  #        item_level = meta_data,
  #        label_col = label_col,
  #        label_class = "SHORT",
  #        resp_vars_match_label_col_only = TRUE
  # )
  # SummaryData$Variables <-
  #   lapply(vs, paste0, collapse = sprintf(" %s ", SPLIT_CHAR))

  #############################################

  if (!missing(orig_label_col)) { # map back SummaryTable to requested label_col and tables to suitable table labels
    for (slot in c("SummaryTable", "SummaryData")) {
      tb <- get(slot)
      vs <- lapply(util_parse_assignments(tb$Variables,
                                          multi_variate_text = TRUE),
                   function(x) {
                     prep_map_labels(
                       x = x,
                       item_level = meta_data,
                       from = label_col,
                       to  = orig_label_col,
                       ifnotfound = x,
                       warn_ambiguous = FALSE
                     )
                   }
      )
      tb$Variables <-
        lapply(vs, paste0, collapse = sprintf(" %s ", SPLIT_CHAR))

      tb[["Variables"]] <-
        prep_map_labels(x = tb[["Variables"]],
                        item_level = meta_data,
                        to  = orig_label_col,
                        from = label_col,
                        ifnotfound = tb[["Variables"]],
                        warn_ambiguous = FALSE)
      assign(slot, tb)
    }
  }

  ##########################################
  return(list(FlaggedStudyData = ds1plot,
              SummaryTable = SummaryTable,  # TODO: VariableGroupTable, maybe other functions, too?
              SummaryData = SummaryData,  # TODO: VariableGroupTable, maybe other functions, too?
              SummaryPlot =
                util_set_size(p,
                              width_em = 25 +
                                1.2 * length(variable_group),
                              height_em = 25
                              )))
}

# This file creates an environment with functions, that generate an
# html menu
# used by [print.dataquieR_resultset2]

#' `.menu_env` -- an environment for HTML menu creation
#'
#' used by the dq_report2-pipeline
#' @name menu_env
#' @noRd
.menu_env <- new.env(parent = environment())

#' Create a single menu entry
#'
#' @param title of the entry
#' @param id linked `href`, defaults to modified title. can be a word, then
#'        a single-page-link with an anchor tag is created.
#' @param ... additional arguments for the menu link
#' @return html-a-tag object
#' @name menu_env_menu_entry
#' @keywords internal
.menu_env$menu_entry <- function(title,
           id = title,
           ...) {
  if (!grepl("#", id, fixed = TRUE) &&
      !startsWith(id, "http://") &&
      !startsWith(id, "https://") &&
      !startsWith(id, "file:") &&
      !startsWith(id, "ftp://") &&
      !startsWith(id, "ftps://") &&
      !startsWith(id, "sftp://") &&
      !startsWith(id, "ssh:") &&
      !startsWith(id, "gopher:") &&
      !startsWith(id, "scp:") &&
      !startsWith(id, "rsync:")) {
    hash_if_needed <- "#"
    id <- prep_link_escape(id,
                           html = TRUE)
  } else {
    hash_if_needed <- ""
    link <- sub("#.*$", "", id)
    hash <- sub("^.*?#", "", id)
    hash <- htmltools::urlEncodePath(as.character(hash))
    id <- paste0(link, "#", hash)
  }
  if (!is.null(attr(title, "alternative_names")) &&
      suppressWarnings(util_ensure_suggested("jsonlite",
                                             goal = "Alias names in menu for search",
                                             err = FALSE))) {
    alternative_names <-
      jsonlite::toJSON(attr(title, "alternative_names"), auto_unbox = TRUE)
  } else {
    alternative_names <- NULL
  }

  htmltools::a(href=sprintf(
    "%s%s",
    hash_if_needed,
    id),
    `data-alternative-names` = alternative_names,
                 title, # htmltools::htmlEscape(title),
                 ...)
}

#' Creates a drop-down menu
#'
#' @param title name of the entry in the main menu
#' @param menu_description description, displayed, if the main
#'                         menu entry itself is clicked
#' @param ... the sub-menu-entries
#' @param id id for the entry, defaults to modified title
#'
#' @return html div object
#' @name menu_env_drop_down
#' @keywords internal
.menu_env$drop_down <- function(title,
                                menu_description,
                      ..., id = prep_link_escape(title)) {
  htmltools::div(class = "dropdown",
                 id = id,
                 onclick = sprintf("showDescription('%s', '%s'); event.stopPropagation()",
                                   id,
                                   htmltools::htmlEscape(menu_description,
                                                         TRUE)),
                 htmltools::tags$button(class="dropbtn",
                                        htmltools::tagList(
                                          htmltools::tags$p(htmltools::htmlEscape(title)),
                                          htmltools::tags$i(
                                            class="fa fa-caret-down"))),
                 htmltools::div(class="dropdown-content",
                                # https://stackoverflow.com/a/33225276
                                ...
                 )
  )
}

#' Generate the menu for a report
#'
#' @param pages encapsulated `list` with report pages as `tagList` objects,
#'              its names are the desired file names
#'
#' @return the html-`taglist` for the menu
#' @name menu_env-menu
#' @keywords internal
.menu_env$menu <- function(pages) {
  entry_env <- environment()
  entries_of_dd <- lapply(names(pages), function(fn) {
    dd <- lapply(names(pages[[fn]]), function(sp) {
      r <- attr(pages[[fn]][[sp]], "dropdown")
      if (length(r) != 1) {
        r <- "Dropdown"
        attr(entry_env$pages[[fn]][[sp]], "dropdown") <- r
      }
      r
    })
    dd <- unique(dd)
    lapply(setNames(nm = dd), function(ddn) {
      eoddn <- lapply(names(pages[[fn]]), function(sp) {
        if (attr(pages[[fn]][[sp]], "dropdown") == ddn) {
          util_attach_attr(sp,
                           fn = fn,
                           alternative_names =
                             attr(pages[[fn]][[sp]]$attribs$id, "alternative_names")
                           )
        } else {
          NULL
        }
      })
      eoddn
    })
  })
  all_dd <- unique(unlist(lapply(entries_of_dd, names)))

  menu_from_pages <- lapply(all_dd, function(ddn) {
    util_ensure_suggested("markdown")
    ddmen <- lapply(unlist(lapply(entries_of_dd, `[[`, ddn), recursive = FALSE),
                    function(me) {
                      if (is.null(me)) {
                        NULL
                      } else {
                        do.call("call", args = c(list("menu_entry",
                                                      me,
                                                      sprintf("%s#%s", attr(me, "fn"), me))),
                                quote = TRUE)
                      }
    })
    concept_info <- subset(util_get_concept_info("dqi"),
           get("Dimension") == ddn & get("Level") == 1 &
             get("dataquierR pipeline include") == 1,
           select = c("Definition", "Explanation", "Guidance", "abbreviation",
                      "IndicatorID"),
           drop = FALSE)
    if (nrow(concept_info) == 1) { # https://dataquality.qihs.uni-greifswald.de/PDQC_DQ_1_0_0_0.html
      if (!util_empty(concept_info$IndicatorID))
        href <- sprintf(
          "https://dataquality.qihs.uni-greifswald.de/PDQC_%s.html",
          concept_info$IndicatorID)
      else
        href = NULL
      menu_description <-
        htmltools::tagList(
          htmltools::h2(ddn),
          htmltools::tags$p(
            htmltools::a(
              href=
                href,
              target="_blank",
              title="Online reference",
              ddn
            ),
            ifelse(util_empty(concept_info$abbreviation), "", sprintf(
              " -- related indicator function names are prefixed with %s",
              dQuote(concept_info$abbreviation)))
          ),
          htmltools::h3("Definition"),
          htmltools::tags$p(
            htmltools::HTML(markdown::markdownToHTML(
              text = concept_info$Definition, fragment.only = TRUE))
          ),
          htmltools::h3("Explanation"),
          htmltools::tags$p(
            htmltools::HTML(markdown::markdownToHTML(
              text = concept_info$Explanation, fragment.only = TRUE))
          ),
          htmltools::h3("Guidance"),
          htmltools::tags$p(
            htmltools::HTML(markdown::markdownToHTML(
              text = concept_info$Guidance, fragment.only = TRUE))
          )
        )
    } else {
      menu_description <-
        htmltools::tagList(
          htmltools::h2(ddn),
          htmltools::tags$p(
            sprintf("No unique description for %s in DQ concept", dQuote(ddn)),
          )
        )
    }
    do.call("call", args = c(list("drop_down",
                                  ddn,
                                  menu_description =
                                    as.character(menu_description)),
                             ddmen), quote = TRUE)
  })
  do.call(htmltools::tagList,
          lapply(unlist(menu_from_pages), eval, envir = entry_env),
          quote = FALSE)
}

# make all the functions in the environment enclosed by this environment, too,
# so that they can look up this environment for metadata
for (f in ls(.menu_env)) {
  if (is.function(.menu_env[[f]])) {
    environment(.menu_env[[f]]) <- .menu_env
  }
}

# Makes the manual accessible at run-time
#' Holds parts of the manual at run-time
#' @noRd
..manual <- new.env(parent = emptyenv())

# Makes Indicator .//. Descriptor information
# from the manual accessible at run-time
#' Holds Indicator .// Descriptor assignments from the manual at run-time
#' @noRd
..indicator_or_descriptor <- new.env(parent = emptyenv())

#' @name Indicator
#' @title Indicator Function
#' @description
#' A function that returns some value that correlates with the magnitude of
#' a certain class of data quality problems. Typically, in `dataquieR`, such
#' functions return a `SummaryTable` that features columns with names, that
#' start with a short abbreviation that describes the specific semantics of
#' the value (e.g., `PCT` for a percentage or `COR` for a correlation) and
#' the public name of the indicator according to the data quality concept
#' `DQ_OBS`, e.g., `com_qum_nonresp` for item-non-response-rate. A name could
#' therefore be `PCT_com_qum_nonresp`.
#'
#' The object `Indicator` only contains the name used internally to tag
#' such functions.
#'
#' @docType data
#' @keywords internal
#' @seealso [Descriptor]
Indicator <- "Indicator"

#' @name Descriptor
#' @title Descriptor Function
#' @description
#' A function that returns some figure or table to assess data quality, but it
#' does not return a value correlating with the magnitude of a data quality
#' problem. It's the opposite of an [Indicator].
#'
#' The object `Descriptor` only contains the name used internally to tag
#' such functions.
#'
#' @docType data
#' @keywords internal
#' @seealso [Indicator]
Descriptor <- "Descriptor"

util_fix_backticks <- function(path = "man") {
  rd_files <- list.files(path, "\\.Rd$", full.names = TRUE)

  for (f in rd_files) {
    x <- readLines(f, warn = FALSE, encoding = "UTF-8")

    x <- gsub(
      "\\\\item\\{(?:\\\\\\\\)?`?cross-item_level`?(?:\\\\\\\\)?\\}",
      "\\\\item{`cross-item_level`}",
      x,
      perl = TRUE
    )

    writeLines(x, f, useBytes = TRUE)
  }

  invisible(rd_files)
}

#' being called by the active binding function for .manual
#'
#' @param rebuild rebuild the cache
#' @param target file for `..manual`
#' @param target2 file for `..indicator_or_descriptor`
#' @param man_hash internal use: hash-sum over the manual to prevent rebuild
#'                 if not changed.
#'
#' @family reporting_functions
#' @concept system
#' @noRd
util_load_manual <- function(rebuild = FALSE,
                             target = "inst/manual.RData",
                             target2 = "inst/indicator_or_descriptor.RData",
                             man_hash = "") {
  if (!rebuild) {
    .manual_file <-
      system.file("manual.RData", package = "dataquieR")
    .indicator_or_descriptor_file <-
      system.file("indicator_or_descriptor.RData", package = "dataquieR")
    if (file.exists(.manual_file) &&
        file.exists(.indicator_or_descriptor_file)) {
      if ((!util_is_try_error(try(load(.manual_file, envir = ..manual),
                                 silent = TRUE))) &&
           !util_is_try_error(try(load(.indicator_or_descriptor_file, envir =
                                       ..indicator_or_descriptor),
                               silent = TRUE)))

        return()
    }
  }
  is_dev_package <- function() FALSE
  if (suppressWarnings(util_ensure_suggested("pkgload", err = FALSE,
                                             goal =
                                             "provide help on report sections during development",
                                             and_import = "is_dev_package"))) {
    dev_package <- is_dev_package(utils::packageName())
  } else {
    dev_package <- FALSE
  }
  ind_fkts <- setNames(nm =
             ls(envir = parent.env(environment()),
                pattern = "^(des|int|com|con|acc)_"))
  if (FALSE && dev_package) {
    ..manual$rd_objects <- lapply(ind_fkts,
                                 function(fn) eval(call("?", as.symbol(fn))))
  } else {
    if (suppressWarnings(util_ensure_suggested("Rdpack", err = FALSE,
                                               goal =
                                               "provide help on report sections",
    ))) {
      ..manual$rd_objects <- lapply(ind_fkts,
                                   function(f) {
                                     o <- Rdpack::Rdo_fetch(f,
                                                          utils::packageName(),
                                                            installed = FALSE)
                                     list(#path = attr(o, "Rdfile"),
                                          doc = o)
                                   })
    } else {
      ..manual$rd_objects <- lapply(ind_fkts,
                                   function(f) list(
                                     # path = "",
                                     doc = NULL))
    }
  }


  # docs <- lapply(..manual$rd_objects, function(rdo) {
  #   if (!!length(rdo) &&
  #       "path" %in% names(rdo) &&
  #       length(rdo[["path"]]) == 1 &&
  #       is.character(rdo[["path"]]) &&
  #       file.exists(rdo[["path"]])) {
  #       doc <- tools::parse_Rd(rdo[["path"]])
  #     list(
  #       doc = doc,
  #       rd_as_latex = rd_as_latex
  #     )
  #   } else {
  #     list(
  #       doc = NULL,
  #       rd_as_latex = NULL
  #     )
  #   }
  # })
  docs <- ..manual$rd_objects

  for (function_name in names(docs)) {
    assign(function_name,
           NA,
           envir = ..indicator_or_descriptor)
    doc <- docs[[function_name]]$doc

    is_descriptor <- FALSE
    is_indicator <- FALSE

    if (!is.null(doc)) {
      find_rd_tag_link <- function(el, parent = NULL) {
        if (!is.null(attr(el, "Rd_tag"))) {
          if (identical(attr(el, "Rd_tag"), "TEXT") &&
              identical(attr(parent, "Rd_tag"), "\\link") &&
              length(el) == 1 &&
              !is.na(el) &&
              is.character(el)) {
            if (el %in% c("Indicator", "Descriptor")) {
              return(el)
            }
          }
        }
        if (!identical(el, parent)) {
          return(lapply(el, find_rd_tag_link, parent = el))
        } else {
          return(NULL)
        }
      }
      ind_or_desc <- unlist(find_rd_tag_link(doc))
      is_indicator <- "Indicator" %in% ind_or_desc
      is_descriptor <- "Descriptor" %in% ind_or_desc
    }

    if (is_indicator && is_descriptor) {
      util_warning(
        c("Internal error, sorry. Please report: For function %s,",
          "the information, if it is an",
          "Indicator or a Descriptor is set to both in the",
          "Roxygen2 comments. Will classify it as a %s, for now."),
        sQuote(function_name),
        sQuote(Descriptor)
      )
      assign(function_name,
             FALSE,
             envir = ..indicator_or_descriptor)
    } else if (!is_indicator && !is_descriptor) {
      if (length(doc) > 0) {
        util_warning(
          c("Internal error, sorry. Please report: For function %s,",
            "the information, if it is an",
            "Indicator or a Descriptor is not correctly available in the",
            "Roxygen2 comments. Will classify it as a %s, for now."),
          sQuote(function_name),
          sQuote(Descriptor)
        )
      }
      assign(function_name,
             FALSE,
             envir = ..indicator_or_descriptor)
    } else {
      assign(function_name,
             is_indicator,
             envir = ..indicator_or_descriptor)
    }
  }

  ..manual$descriptions <- lapply(setNames(nm = names(docs)),
                                  function(function_name) {
    doc <- docs[[function_name]]$doc
    if (!is.null(doc)) {
      descr <- paste(unlist(doc[lapply(doc, attr, "Rd_tag") == "\\description"]),
                     collapse = "")
      descr <- trimws(gsub("\\s+", " ", descr))
    } else {
      descr <- ""
    }
    descr
  })

  ..manual$titles <- lapply(setNames(nm = names(docs)),
                            function(function_name) {
      doc <- docs[[function_name]]$doc
      if (!is.null(doc)) {
        title <- paste(unlist(doc[lapply(doc, attr, "Rd_tag") == "\\title"]),
                       collapse = "")
        title <- trimws(gsub("\\s+", " ", title))
      } else {
        title <- ""
      }
      title
  })
  if (rebuild) {
    ..manual$rd_objects <- NULL
    ..manual$man_hash <- man_hash
    save(list = ls(envir = ..manual, all.names = TRUE), envir = ..manual, file = target,
         compress = "xz")
    ..indicator_or_descriptor$man_hash <- man_hash
    save(list = ls(envir = ..indicator_or_descriptor, all.names = TRUE),
         envir = ..indicator_or_descriptor, file = target2,
         compress = "xz")
  }
}

test_that("reflection works", {
  skip_on_cran() # snot really a problem, if not working. also obvious, if buggy. slow.
  util_load_manual()
  expect("titles" %in% names(..manual),
         failure_message = "load manual did not read titles")
  expect_true(
    any(
      vapply(
        mget(setdiff(ls(.indicator_or_descriptor), "man_hash"),
             envir = .indicator_or_descriptor), identity,
        FUN.VALUE = logical(1))), info = "Any Indicator in Package")
})

#' utility function for the applicability of outlier detection
#'
#' Test for applicability of univariate outlier detection
#'
#' @param x [data.frame] metadata
#' @param dta [logical] vector, 1=matching data type,
#'                  0 = non-matching data type
#'
#' @return [factor] 0-3 for each variable in metadata
#' - 0 data type mismatch and not applicable
#' - 1 data type mismatches but applicable
#' - 2 data type matches but not applicable
#' - 3 data type matches and applicable
#' - 4 not applicable because of not suitable data type
#'
#' @seealso [pro_applicability_matrix]
#' @noRd

util_app_ol <- function(x, dta) {
  aa <- paste0(dta, 1)
  score <- as.numeric(recode(as.factor(aa), "00" = 0, "01" = 1,
                             "10" = 2, "11" = 3))
  if (is.null(x[[VALUE_LABELS]])) {
    x[[VALUE_LABELS]] <- NA
  }
  if (is.null(x[[VALUE_LABEL_TABLE]])) {
    x[[VALUE_LABEL_TABLE]] <- NA
  }
  if (is.null(x[[STANDARDIZED_VOCABULARY_TABLE]])) {
    x[[STANDARDIZED_VOCABULARY_TABLE]] <- NA
  }
  score <- ifelse(x[["DATA_TYPE"]] == DATA_TYPES$FLOAT |
                    (x[["DATA_TYPE"]] == DATA_TYPES$INTEGER &
                       (is.null(x[[VALUE_LABELS]]) | util_empty(x[[VALUE_LABELS]])) &
                       (is.null(x[[VALUE_LABEL_TABLE]]) | util_empty(x[[VALUE_LABEL_TABLE]])) &
                       (is.null(x[[STANDARDIZED_VOCABULARY_TABLE]]) |
                          util_empty(x[[STANDARDIZED_VOCABULARY_TABLE]]))),
                  score, 4)
  score <- as.factor(score)
  return(score)
}

#' utility function to assign labels to levels
#'
#' function to assign labels to levels of a variable
#'
#' DEPRECATED from v2.5.0
#'
#' @param variable [vector] vector with values of a study variable
#' @param string_of_levlabs [character] len=1. value labels,
#'                                             e.g. `1 = no | 2 = yes`
#' @param splitchar [character] len=1. splitting character(s) in
#'                                 `string_of_levlabs`, usually [SPLIT_CHAR]
#' @param assignchar [character] len=1. assignment operator character(s) in
#'                                  `string_of_levlabs`, usually ` = ` or `: `
#' @param ordered the function converts `variable` to a [factor], by default to
#'                an [ordered] [factor] assuming LHS of assignments being
#'                meaningful numbers, e.g. `1 = low | 2 = medium | 3 = high`.
#'                If no special order is given, set `ordered` to `FALSE`,
#'                e.g. for `1 = male | 2 = female` or
#'                `1 = low | 2 = high | 3 = medium`.
#' @param variable_name [character] the name of the variable being converted
#'                                  for warning messages
#' @param warn_if_inadmissible [logical] warn on [con_inadmissible_categorical]
#'                                       values
#'
#' @return a factor with labels assigned to categorical variables
#'         (if available)
#'
#' @family data_management
#' @concept data_management
#' @noRd
util_assign_levlabs <- function(variable, string_of_levlabs, splitchar,
                                assignchar, ordered = TRUE, variable_name = "",
                                warn_if_inadmissible = TRUE) {
  # FIXME: deprecate
  # lifecycle::deprecate_soft("2.5.0",
  #                           what = "util_assign_levlabs()",
  #                           with = "prep_prepare_dataframes()")
  # TODO: handle VALUE_LABELS w/o codes, e.g. male | female
  util_expect_scalar(variable_name, check_type = is.character)
  if (!util_empty(variable_name)) {
    variable_name <- sprintf(" for variable %s", dQuote(variable_name))
  }
  util_expect_scalar(string_of_levlabs, check_type = is.character)
  util_expect_scalar(splitchar, check_type = is.character)
  util_expect_scalar(assignchar, check_type = is.character)
  util_expect_scalar(ordered, check_type = is.logical)
  util_expect_scalar(variable, # not nec. numeric
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     allow_na = TRUE)

  #TODO: util_deparse_assignments
  #TODO: rewrite, user should not need ot lookup metadata, this could be done here
  #TODO: make available as a prep funciton
  # split one string into multiple
  levlab_list <- strsplit(string_of_levlabs, split = splitchar, fixed = TRUE)
  # remove lreading/trailing blanks
  levlab_list <- trimws(unlist(levlab_list), which = "both")

  levlab_list <- levlab_list[!util_empty(levlab_list)]

  # get levels
  .levs <- unlist(lapply(levlab_list, function(x)
    unlist(strsplit(x, "=", fixed = TRUE))[1]))
  .levs <- trimws(.levs, which = "both")
  # get labels
  labs <- unlist(lapply(levlab_list, function(x)
    unlist(strsplit(x, "=", fixed = TRUE))[2]))
  labs <- trimws(labs, which = "both")

  # levels to numeric if applicable
  if (suppressWarnings(all(!is.na(as.numeric(.levs))))) {
    levs <- as.numeric(.levs)
  } else {
    levs <- .levs
  }

  # warnings:
  if (length(levs) != length(labs)) {
    # Dead code?
    util_warning("Number of levels does not match number of labels%s.", # nocov
                 variable_name, applicability_problem = TRUE) # nocov
  }

  if (warn_if_inadmissible && length(levs) <
      length(unique(variable[!(is.na(variable))]))) {
    util_warning(
      "Number of levels in variable greater than in character string%s.",
      variable_name, applicability_problem = TRUE)
  }

  if (warn_if_inadmissible && any(is.na(labs))) {
    util_warning("No labels assigned for some levels, use levels as labels%s",
                 variable_name, applicability_problem = TRUE)
    labs[is.na(labs)] <- levs[is.na(labs)]
  }

  # assign levels and labels to vector
  if ((length(levs) == length(labs))) {
    # & (length(levs) >=
    #                                     length(
    #                                       unique(
    #                                         variable[!(is.na(variable))])))) {
    if (warn_if_inadmissible & !all(variable %in% c(NA, levs))) {
      util_warning(
        c("Inadmissible categorical values found, use levels as labels%s"),
                   variable_name, applicability_problem = TRUE)
      inadm <- unique(variable[!(variable %in% c(NA, levs))])
      levs <- c(levs, inadm)
      labs <- c(labs, inadm)
    }
    variable <- factor(variable, levels = levs,
                       labels = labs, ordered = ordered)
  }

  return(variable)
}

test_that("util_assign_levlabs works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  vname <- "v50000"
     labelled_var <-
       dataquieR:::util_assign_levlabs(
         variable = study_data[[vname]],
         string_of_levlabs = subset(meta_data,
                                    VAR_NAMES == vname,
                                    VALUE_LABELS,
                                    TRUE),
         splitchar = SPLIT_CHAR,
         assignchar = " = "
       )

     expect_identical(
       table(labelled_var, useNA = "always"),
       structure(c(76L, 2864L, 60L), .Dim = 3L,
         .Dimnames = list(labelled_var = c("no",  "yes", NA)),
         class = "table")
     )
     expect_true(labelled_var[2021] < labelled_var[1])
     expect_true(is.ordered(labelled_var))

     labelled_var1 <-
       dataquieR:::util_assign_levlabs(
         ordered = FALSE,
         variable = study_data[[vname]],
         string_of_levlabs = subset(meta_data,
                                    VAR_NAMES == vname,
                                    VALUE_LABELS,
                                    TRUE),
         splitchar = SPLIT_CHAR,
         assignchar = " = "
       )

     expect_identical(
       table(labelled_var1, useNA = "always"),
       structure(c(76L, 2864L, 60L), .Dim = 3L,
                 .Dimnames = list(labelled_var1 = c("no",  "yes", NA)),
                 class = "table")
     )
     expect_false(is.ordered(labelled_var1))

     variable <- study_data[[vname]]
     string_of_levlabs <- subset(meta_data,
            VAR_NAMES == vname,
            VALUE_LABELS,
            TRUE)
     labelled_var <-
             dataquieR:::util_assign_levlabs(
                     variable = variable,
                     string_of_levlabs = string_of_levlabs,
                     splitchar = SPLIT_CHAR,
                     assignchar = " = "
             )
     non_num_labs <- as.character(labelled_var)
     string_of_levlabs_now <- "no = no | yes = yes"
     labelled_var2 <- util_assign_levlabs(
             variable = non_num_labs,
             string_of_levlabs = string_of_levlabs_now,
             splitchar = SPLIT_CHAR,
             assignchar = " = "
     )
     expect_equal(labelled_var2, labelled_var)
     string_of_levlabs_now <- "no = no | yes"
     expect_warning(invisible(
             util_assign_levlabs(
                     variable = non_num_labs,
                     string_of_levlabs = string_of_levlabs_now,
                     splitchar = SPLIT_CHAR,
                     assignchar = " = "
             )),
             regexp =
                     "use levels as labels",
             fixed = TRUE
     )
     string_of_levlabs_now <- "no = no = no"
     expect_warning(expect_warning(invisible(
             util_assign_levlabs(
                     variable = non_num_labs,
                     string_of_levlabs = string_of_levlabs_now,
                     splitchar = SPLIT_CHAR,
                     assignchar = " = "
             )),
             regexp = sprintf("(%s|%s)",
          "Number of levels in variable greater than in character string.",
          "use levels as labels"
             ),
             perl = TRUE
     ),
     "Inadmissible categorical values found, use levels as labels",
     perl = TRUE)
})

#' Compute something comparable from an `ordered`
#'
#' interpolates categories of an ordinal variable
#'
#' @param codes [numeric()] `n` values
#' @param maxlevel_old [integer()] number of categories of `codes`
#' @param maxlevel_new [integer()] number of categories for output
#'
#' @return [integer()] `n` values in `{1, ..., maxlevel_new}`
#' @noRd
util_standardize_ordinal_codes <- function(codes, maxlevel_old, maxlevel_new) {
  util_expect_scalar(maxlevel_old, check_type = util_is_numeric_in(min = 2,
                                                                   whole_num = TRUE,
                                                                   finite = TRUE))
  util_expect_scalar(maxlevel_new, check_type = util_is_numeric_in(min = 2,
                                                                   whole_num = TRUE,
                                                                   finite = TRUE))
  util_expect_scalar(codes,
                     allow_more_than_one = TRUE,
                     check_type = util_is_numeric_in(min = 1,
                                                     max = maxlevel_old,
                                                     whole_num = TRUE,
                                                     finite = TRUE))
  as.integer(round((codes - 1) * ((maxlevel_new - 1) / (maxlevel_old - 1))) + 1)
}

test_that("util_standardize_ordinal_codes works", {
  skip_on_cran()
  # Increasing ordinal codes
  codes <- c(1,  2,  3,  4,  5)
  maxlevel_old <-  5
  maxlevel_new <-  10
  expected_result <- c(1,  3,  5,  8,  10)
  expect_equal(util_standardize_ordinal_codes(codes, maxlevel_old, maxlevel_new), expected_result)

  # Decreasing ordinal codes
  codes <- c(5,  4,  3,  2,  1)
  expected_result <- c(10,  8,  5,  3,  1)
  expect_equal(util_standardize_ordinal_codes(codes, maxlevel_old, maxlevel_new), expected_result)

  # Equal ordinal codes
  codes <- rep(3,  5)
  expected_result <- rep(5,  5)
  expect_equal(util_standardize_ordinal_codes(codes, maxlevel_old, maxlevel_new), expected_result)

  # Invalid input
  expect_error(util_standardize_ordinal_codes(c(1,  2, "three"), maxlevel_old, maxlevel_new))
  expect_error(util_standardize_ordinal_codes(codes,  0, maxlevel_new))
  expect_error(util_standardize_ordinal_codes(codes, -1, maxlevel_new))
})

#' Read data from files/URLs
#'
#' data_frame_name can be a file path or an URL you can append a pipe and a
#' sheet name for Excel files or object name e.g. for `RData` files. Numbers
#' may also work. All file formats supported by your `rio` installation will
#' work.
#'
#' The data frames will be cached automatically, you can define an alternative
#' environment for this using the argument `.data_frame_list`, and you can purge
#' the cache using [prep_purge_data_frame_cache].
#'
#' Use [prep_add_data_frames] to manually add data frames to the
#' cache, e.g., if you have loaded them from more complex sources, before.
#'
#' @param data_frame_name [character] name of the data frame to read, see
#'   details
#' @param .data_frame_list [environment] cache for loaded data frames
#' @param keep_types [logical] keep types as possibly defined in a file, if the
#'                             data frame is loaded from one. set `TRUE` for
#'                             study data.
#' @param column_names_only [logical] if TRUE imports only headers (column names)
#'                                    of the data frame and no content
#'                                    (an empty data frame)
#'
#' @return [data.frame] a data frame
#' @export
#' @seealso [prep_add_data_frames]
#' @seealso [prep_load_workbook_like_file]
#' @family data-frame-cache
#'
#' @examples
#' \dontrun{
#' bl <- as.factor(prep_get_data_frame(
#'   paste0("https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus",
#'     "/Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=",
#'     "publicationFile|COVID_Todesfälle_BL|Bundesland"))[[1]])
#'
#' n <- as.numeric(prep_get_data_frame(paste0(
#'   "https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus/",
#'   "Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=",
#'   "publicationFile|COVID_Todesfälle_BL|Anzahl verstorbene",
#'   " COVID-19 Fälle"))[[1]])
#' plot(bl, n)
#' # Working names would be to date (2022-10-21), e.g.:
#' #
#' # https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus/ \\
#' #    Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=publicationFile
#' # https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus/  \\
#' #    Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=publicationFile|2
#' # https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus/ \\
#' #    Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=publicationFile|name
#' # study_data
#' # ship
#' # meta_data
#' # ship_meta
#' #
#' prep_get_data_frame("meta_data | meta_data")
#' }
prep_get_data_frame <- function(data_frame_name,
                                .data_frame_list = .dataframe_environment(),
                                keep_types = FALSE,
                                column_names_only = FALSE) {
  # IDEA: also allow to address rows or cells from dataframes using the syntax in data_frame_name
  if (!missing(keep_types)) {
    util_expect_scalar(keep_types,
                       check_type = is.logical,
                       error_message =
                         sprintf("%s needs to be to be a logical value.",
                                 sQuote("keep_types")))
  }

  util_expect_scalar(column_names_only,
                     check_type = is.logical,
                     error_message =
                       sprintf("%s needs to be to be a logical value.",
                               sQuote("column_names_only")))

  data_frame_name <- gsub("\\s*\\|\\s*", "|", data_frame_name)

  util_expect_scalar(data_frame_name, check_type = is.character)
  if (!is.environment(.data_frame_list)) {
    util_error("%s must be an environment, if specified",
               sQuote(".data_frame_list"))
  }

  # Special case of ship.RDS not present in the package anymore
  if (grepl(sprintf("^ship(\\s*\\%s.*)?$", SPLIT_CHAR), data_frame_name) ) {
    util_stop_if_not(
      `no shortcuts in tests` =
        !identical(Sys.getenv("TESTTHAT"), "true"))

    # nocov start
    # this cannot be checked, because we disallow shortcuts in tests, see above.
    data_frame_name <-
      sub(sprintf("^ship(\\s*\\%s.*)?$", SPLIT_CHAR),
          "https://dataquality.qihs.uni-greifswald.de/extdata/ship.RDS\\1",
          data_frame_name)
    # nocov end
  }

  # Special case of ship.RDS not present in the package anymore
  if (grepl(sprintf("^ship_subset[1-3](\\s*\\%s.*)?$", SPLIT_CHAR),
            data_frame_name) ) {
    util_stop_if_not(
      `no shortcuts in tests` =
        !identical(Sys.getenv("TESTTHAT"), "true"))

    # nocov start
    # this cannot be checked, because we disallow shortcuts in tests, see above.
    data_frame_name <-
      sub(sprintf("^ship_subset([1-3])(\\s*\\%s.*)?$", SPLIT_CHAR),
    "https://dataquality.qihs.uni-greifswald.de/extdata/ship_subset\\1.RDS\\2",
          data_frame_name)
    # nocov end
  }

  # Special case speed-up
  if ("grading_rulesets" == data_frame_name &&
      !exists(data_frame_name,
                 envir = .dataframe_environment(),
                 mode = "list")) {
    data_frame_name <- system.file("grading_rulesets.xlsx",
                                   package = "dataquieR")
  }
  # Special case speed-up
  if ("grading_formats" == data_frame_name &&
      !exists(data_frame_name,
              envir = .dataframe_environment(),
              mode = "list")) {
    data_frame_name <- system.file("grading_formats.xlsx",
                                   package = "dataquieR")
  }


  # Special case of ship.RDS not present in the package anymore
  if (grepl(sprintf("^ship_meta_v2(\\s*\\%s.*)?$", SPLIT_CHAR), data_frame_name) ) {
    util_stop_if_not(
      `no shortcuts in tests` =
        !identical(Sys.getenv("TESTTHAT"), "true"))

    # nocov start
    # this cannot be checked, because we disallow shortcuts in tests, see above.
    data_frame_name <-
      sub(sprintf("^ship_meta_v2(\\s*\\%s.*)?$", SPLIT_CHAR),
      "https://dataquality.qihs.uni-greifswald.de/extdata/ship_meta_v2.xlsx\\1",
          data_frame_name)
    # nocov end
  }

  # Special case of ship.RDS not present in the package anymore
  if (grepl(sprintf("^ship_meta_dataframe(\\s*\\%s.*)?$", SPLIT_CHAR), data_frame_name) ) {
    util_stop_if_not(
      `no shortcuts in tests` =
        !identical(Sys.getenv("TESTTHAT"), "true"))

    # nocov start
    # this cannot be checked, because we disallow shortcuts in tests, see above.
    data_frame_name <-
      sub(sprintf("^ship_meta_dataframe(\\s*\\%s.*)?$", SPLIT_CHAR),
"https://dataquality.qihs.uni-greifswald.de/extdata/ship_meta_dataframe.xlsx\\1",
          data_frame_name)
    # nocov end
  }

  # Special case of ship.RDS not present in the package anymore
  if (grepl(sprintf("^ship_meta(\\s*\\%s.*)?$", SPLIT_CHAR), data_frame_name) ) {
    util_stop_if_not(
      `no shortcuts in tests` =
        !identical(Sys.getenv("TESTTHAT"), "true"))

    # nocov start
    # this cannot be checked, because we disallow shortcuts in tests, see above.
    data_frame_name <-
      sub(sprintf("^ship_meta(\\s*\\%s.*)?$", SPLIT_CHAR),
          "https://dataquality.qihs.uni-greifswald.de/extdata/ship_meta.RDS\\1",
          data_frame_name)
    # nocov end
  }

  # Special case of study_data.RData not present in the package anymore
  if (grepl(sprintf("^study_data(\\s*\\%s.*)?$", SPLIT_CHAR),
            data_frame_name) && !exists(data_frame_name,
                                        envir = .dataframe_environment(),
                                        mode = "list")) {
    util_stop_if_not(
      `no shortcuts in tests` =
        !identical(Sys.getenv("TESTTHAT"), "true"))

    # nocov start
    # this cannot be checked, because we disallow shortcuts in tests, see above.
    data_frame_name <-
      sub(sprintf("^study_data(\\s*\\%s.*)?$", SPLIT_CHAR),
      "https://dataquality.qihs.uni-greifswald.de/extdata/study_data.RData\\1",
          data_frame_name)
    # nocov end
  }

  # Special case of study_data.RData not present in the package anymore
  if (grepl(sprintf("^meta_data_v2(\\s*\\%s.*)?$", SPLIT_CHAR),
            data_frame_name) ) {
    util_stop_if_not(
      `no shortcuts in tests` =
        !identical(Sys.getenv("TESTTHAT"), "true"))

    # nocov start
    # this cannot be checked, because we disallow shortcuts in tests, see above.
    data_frame_name <-
      sub(sprintf("^meta_data_v2(\\s*\\%s.*)?$", SPLIT_CHAR),
      "https://dataquality.qihs.uni-greifswald.de/extdata/meta_data_v2.xlsx\\1",
          data_frame_name)
    # nocov end
  }

  # Special case of study_data.RData not present in the package anymore
  if (grepl(sprintf("^meta_data(\\s*\\%s.*)?$", SPLIT_CHAR),
            data_frame_name) ) {
    util_stop_if_not(
      `no shortcuts in tests` =
        !identical(Sys.getenv("TESTTHAT"), "true"))

    # nocov start
    # this cannot be checked, because we disallow shortcuts in tests, see above.
    data_frame_name <-
      sub(sprintf("^meta_data(\\s*\\%s.*)?$", SPLIT_CHAR),
        "https://dataquality.qihs.uni-greifswald.de/extdata/meta_data.RData\\1",
          data_frame_name)
    # nocov end
  }

  if (all(startsWith(
    data_frame_name,
    "https://pfau.qihs.uni-greifswald.de/dfg_website_rendered/extdata/fortests/")
    | startsWith(
    data_frame_name,
    "https://dataquality.qihs.uni-greifswald.de/extdata/"))) {
    # trust our own example files
    withr::local_options(list(rio.import.trust = TRUE))
  }

  if (!column_names_only && nzchar(data_frame_name) && exists(data_frame_name,
                                   envir = .data_frame_list, mode = "list")) {
    r <- get(data_frame_name, envir = .data_frame_list, mode = "list")
    if (is.data.frame(r)) {
      return(r)
    }
  }

  # util_ensure_suggested(c("rio"), "Load data from files")

  fn <- data_frame_name
  which <- NULL
  col <- NULL

  if (any(grepl(SPLIT_CHAR, data_frame_name, fixed = TRUE))) {
    splitted <- trimws(strsplit(data_frame_name, SPLIT_CHAR, fixed = TRUE)[[1]])

    if (length(splitted) > 2) { # 1-column data frame requested
      col <- tail(splitted, 1)
      splitted <- head(splitted, -1)
    }

    which <- tail(splitted, 1)
    suppressWarnings(
      if (is.finite(as.integer(which)) && as.integer(which) < 500) {
        which <- as.integer(which)
      }
    )
    fn <- paste0(head(splitted, -1), collapse = SPLIT_CHAR)
  }

  r <- NULL

  if (startsWith(fn, "<") &&
      endsWith(fn, ">")) {
    fn <- paste0("voc:", substr(fn, 2, nchar(fn) - 1))
  }
  if (startsWith(fn, "voc:")) {
    .voc <- sub("^voc:", "", fn)
    voc_tab <- util_get_voc_tab()
    # table lookup to modify fn to package: or data:
    recall_arg <-
      try(subset(voc_tab, get("voc") == .voc, "url", drop = TRUE))
    if (!inherits(recall_arg, "try-error") && length(recall_arg) == 1) {
      return(Recall(recall_arg))
    }
    # data("icd_meta_codes", envir = e)
  }
  # auto request to install missing packages
  if (startsWith(fn, "dbx:")) {# FIXME: Test this
    try({
      util_ensure_suggested("dbx", goal = "accessing databases")
      dbxurl <- sub("^dbx:", "", fn)
      connect_args <- list(url = dbxurl)
      opt_args <- getOption("dataquieR")
      if (is.list(opt_args)) {
        connect_args <-
          connect_args[!(names(connect_args) %in% names(opt_args))]
        connect_args <- c(connect_args, opt_args)
      }
      con <- do.call(dbx::dbxConnect, connect_args)
      withr::with_db_connection(con, { # TODO: Where conditions, also for the other url types maye with @
        if (is.null(which)) {
          util_error("Need a table name for database access %s.",
                     dQuote(fn),
                     applicability_problem = TRUE)
        }
        which <- gsub("[^a-zA-Z0-9_\\.\\[\\]]", "", which)
        if (is.null(col)) {
          col <- "*"
        } else {
          col <- gsub("[^a-zA-Z0-9_\\.\\[\\]]", "", col)
        }
        r <- dbx::dbxSelect(con,  #TODO: consider the argument column_names_only also in this case
                            sprintf("SELECT %s FROM %s",
                                    col,
                                    which))
      })
      if (!is.data.frame(r)) {
        util_error("Could not load %s",
                   dQuote(fn),
                   applicability_problem = TRUE)
      }
    })
  } else if (startsWith(fn, "extdata:")) {
    p <- sub("^extdata:", "", fn)
    p <- strsplit(p, "/", fixed = TRUE)[[1]]
    package <- p[[1]]
    util_ensure_suggested(package, goal = "load data from it")
    p[[1]] <- "extdata"
    fn <- do.call(system.file, c(as.list(p), list(package = package)))
  } else if (startsWith(fn, "package:")) {
    p <- sub("^package:", "", fn)
    p <- strsplit(p, "/", fixed = TRUE)[[1]]
    package <- p[[1]]
    util_ensure_suggested(package, goal = "load data from it")
    p <- p[-1]
    fn <- do.call(system.file, c(as.list(p), list(package = package)))
  } else if (startsWith(fn, "data:")) try({
    p <- sub("^data:", "", fn)
    p <- strsplit(p, "/", fixed = TRUE)[[1]]
    package <- p[[1]]
    util_ensure_suggested(package, goal = "load data from it")
    p <- p[-1]
    if (length(p) > 0 || length(which) != 1) {
      util_error("Cannot load %s", dQuote(fn), applicability_problem = TRUE)
    }
    # also allow missing package, then use search() or installed.packages()
    .e <- new.env(parent = emptyenv())
    utils::data(list = which,
         package = package,
         envir = .e)
    r <- .e[[which]]  #TODO: consider the new argument column_names_only also here
    if (!is.data.frame(r)) {
      util_error("Could not load %s",
                 dQuote(fn),
                 applicability_problem = TRUE)
    }
  })

  # create a conditional named list based on argument column_names_only
  if (!is.data.frame(r)) {
    arguments_for_rio_import <- list(fn = fn,
                                     keep_types = keep_types)
    if (column_names_only) {
      to_add_list <- list(n_max = 0, nrows = 0)
      arguments_for_rio_import <- c(arguments_for_rio_import, to_add_list)
    } else {
      if (!is.null(which)) {
        to_add_list <- list(which = which)
        arguments_for_rio_import <- c(arguments_for_rio_import, to_add_list)
      }
    }
    r <- do.call(util_rio_import, arguments_for_rio_import)
  }

  if (inherits(r, "try-error") || !is.data.frame(r)) {
    fn0 <- system.file("extdata", paste0(fn, ".RDS"), package = "dataquieR")
    if (file.exists(fn0)) {
      fn <- fn0
    } else {
      fn0 <- system.file("extdata", paste0(fn, ".RData"), package = "dataquieR")
      if (file.exists(fn0)) {
        fn <- fn0
      } else {
        fn0 <- system.file("extdata", paste0(fn, ".xlsx"),
                           package = "dataquieR")
        if (file.exists(fn0)) {
          fn <- fn0
        }
      }
    }


    # create a conditional named list based on argument column_names_only
    arguments_for_rio_import <- list(fn = fn,
                                     keep_types = keep_types)
    if (column_names_only) {
      to_add_list <- list(n_max = 0, nrows = 0)
      arguments_for_rio_import <- c(arguments_for_rio_import, to_add_list)
    } else {
      if (!is.null(which)) {
        to_add_list <- list(which = which)
        arguments_for_rio_import <- c(arguments_for_rio_import, to_add_list)
      }
    }

    r <- do.call(util_rio_import, arguments_for_rio_import)


    if (inherits(r, "try-error")) {
      if (conditionMessage(attr(r, "condition")) ==
          "No such file" && length(prep_list_dataframes()) == 0) {
        hint <- sprintf("\nDid you forget to call %s, %s or %s?",
                        sQuote("prep_load_workbook_like_file"),
                        sQuote("prep_add_data_frames"),
                        sQuote("prep_load_folder_with_metadata")
        )
      } else {
        hint <- ""
      }
      util_error("Cannot read file %s using %s: %s%s",
                 dQuote(fn),
                 sQuote("rio"),
                 conditionMessage(attr(r, "condition")),
                 hint)
    }

    if (!is.data.frame(r)) {
      util_error("File %s did not contain a table (data frame) according to %s",
                 dQuote(fn),
                 sQuote("rio"))
    }


  }

  if (!is.null(col)) {
    if (grepl("+", col, fixed = TRUE)) {
      col <- strsplit(col, "+", fixed = TRUE)[[1]]
    }
    if (!(all(col %in% colnames(r)))) {
      util_error("%s does not contain all of the columns %s on/in %s",
                 dQuote(fn),
                 dQuote(util_pretty_vector_string(col)),
                 dQuote(which))
    }
    r <- r[, col, drop = FALSE]
  }

  # delete invalid character codes
  r[] <- util_fix_encoding_cols(r)

  if (!column_names_only) {
    assign(data_frame_name, r, envir = .data_frame_list)
  }
  return(r)
}

.global.dataframe_environment <- new.env(parent = emptyenv())

util_set_dataframe_environment <- function(env = NULL) {
  if (is.null(env)) {
    assign(x = "dataframe_environment",
           value = .global.dataframe_environment,
           envir = dataquieR.properties)
  } else {
    util_stop_if_not(is.environment(env))
    assign(x = "dataframe_environment",
           value = env,
           envir = dataquieR.properties)
  }
}

.dataframe_environment <- function() {
  return(get(x = "dataframe_environment",
             envir = dataquieR.properties,
             mode = "environment"))
}

util_set_dataframe_environment(NULL)

with_dataframe_environment <- function(expr,
                                       env = new.env(parent = emptyenv())) {
  old_env <- .dataframe_environment()
  withr::defer({
    util_set_dataframe_environment(old_env)
  })
  util_set_dataframe_environment(env)
  force(rlang::eval_tidy(expr, env = parent.frame()))
}

test_that("dq_report_by works with content m", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
    skip_on_cran() # slow test
    # TODO: test with all sorts of _by calls and on windows
    target <- withr::local_tempdir("testdqareportby")

    study_data <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), 100)


    #Added a new grading ruleset
    default_grading_ruleset <-
      prep_get_data_frame(system.file("grading_rulesets.xlsx",
                                      package = "dataquieR"))

    default_grading_ruleset[default_grading_ruleset$indicator_metric ==
                              "PCT_con_rvv_icat",] <- c("0",
                                                        "PCT_con_rvv_icat",
                                                        NA,
                                                        NA,
                                                        NA,
                                                        NA,
                                                        "[0; 0]",
                                                        "(0; 100]")

    prep_add_data_frames(grading_rulesets = default_grading_ruleset)

    expect_message2(dq_report_by(study_data = study_data,
                 dimensions = "int",
                 cores = NULL,
                 segment_column = STUDY_SEGMENT,
                 segment_select = c("STUDY", "LAB"),
                 meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                 output_dir = file.path(target, "m"),
                 also_print = TRUE))

    # TODO: check target path for all expected artifacts, namely the dq2 files, the .output_dirs containing the HTML reports and the summary RDS files

})

test_that("util_ds1_eval_env works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  md <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  sd <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  md <- md[md$VAR_NAMES %in% colnames(sd), , FALSE]
  sd <- sd[, intersect(md$VAR_NAMES, colnames(sd)), FALSE]
  md$xx <- abbreviate(md$LABEL)
  e <- util_ds1_eval_env(sd, md,
                         label_col = "xx")
  expect_equal(
    sort(ls(e)),
    sort(unique(c(
      md$VAR_NAMES,
      md$LABEL,
      md$LONG_LABEL,
      md$xx)))
  )
})

test_that("dq_report_by error_arguments_test", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_on_cran() # slow test
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  target <- withr::local_tempdir("testdqareportby")

  sd0 <- head(prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
    20)
  sd0 <- sd0[, head(colnames(sd0), 10)]

  md0 <- prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx | item_level")

  md0 <- md0[md0$VAR_NAMES %in% colnames(sd0), ]
  # fewer segments to speed up things.
  # can be solved in a more elegant way after
  # https://gitlab.com/libreumg/dataquier/-/issues/481

  md0$STUDY_SEGMENT <- "STUDY"


  expect_warning(expect_warning(dq_report_by(study_data = sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            subgroup = "[v00003] == 106",
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm"),
                            also_print = TRUE),
                 regexp = "No data available to create the report.+"),
                 regexp = ".*impossible condition.+")

  expect_warning(dq_report_by(study_data = sd0,
                              meta_data = md0,
                              dimensions = "int",
                              cores = NULL,
                              segment_column = STUDY_SEGMENT,
                              subgroup = "[v00003] > 1",
                              meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "sm1"),
                              also_print = TRUE),
               regexp = "The number of cases did not change after applying the subgroup.+")

  expect_error(dq_report_by(sd0,
                              meta_data = md0,
                              dimensions = "int",
                              cores = NULL,
                              segment_column = STUDY_SEGMENT,
                              strata_select = "2",
                              meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "sm2"),
                              also_print = TRUE),
               "strata_column is needed for selecting the strata", fixed = TRUE)


  expect_message2(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            strata_column = "v00002",
                            strata_select = "^ma",
                            selection_type = "regex",
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm3"),
                            also_print = TRUE))

  #Label_col and strata not matching
  expect_error(dq_report_by(sd0,
                              meta_data = md0,
                              dimensions = "int",
                              cores = NULL,
                              segment_column = STUDY_SEGMENT,
                              label_col = VAR_NAMES,
                              strata_column = "SEX_0",
                              meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "sm4"),
                              also_print = TRUE),
               regexp = "The strata_column provided does not.+item_level_metadata")

  #no existing strata matches the regex
  expect_error(dq_report_by(sd0,
                              meta_data = md0,
                              dimensions = "int",
                              cores = NULL,
                              segment_column = STUDY_SEGMENT,
                              strata_column = "v00002",
                              strata_select = "^hello",
                              selection_type = "regex",
                              meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "sm5"),
                              also_print = TRUE),
               regexp = "No value or value label.+strata.+select")

  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            strata_exclude = "1",
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm6"),
                            also_print = TRUE),
               "strata_column is needed for excluding the strata", fixed = TRUE)

  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            selection_type = "value",
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm7"),
                            also_print = TRUE),
               paste0("selection_type can only be specified if strata_select",
                      " or strata_exclude are present"), fixed = TRUE)

  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            strata_column = "SEX_0",
                            strata_exclude = "1",
                            selection_type = "values",
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm8"),
                            also_print = TRUE),
               paste0('Internal error: The selection_type can only',
                      ' be "value", "v_label", or "regex"'), fixed = TRUE)

  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            subgroup = c("[AGE_0]>30", "[SBP_0]>100"),
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm9"),
                            also_print = TRUE),
               paste0("Need exactly one element in argument subgroup, ",
                      "got 2: [[AGE_0]>30, [SBP_0]>100]"),
               fixed = TRUE)

  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            input_dir = "test/error",
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm10"),
                            also_print = TRUE),
               regexp = ".+test/error.+does not exist. Provide an.+input_dir.+containing the study data")

  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            id_vars = "non-existing",
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm11"),
                            also_print = TRUE),
               regexp = "The id_vars.+non-existing.+is not present in the item_level metadata")


  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm12"),
                            strata_column = "SEX_0",
                            strata_exclude = "female", selection_type = "v_label",
                            also_print = TRUE),
               regexp = "Internal error: No value label.+strata_exclude")

  expect_message2(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm13"),
                            strata_column = "SEX_0",
                            strata_exclude = "females", selection_type = "v_label",
                            also_print = TRUE))


  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm14"),
                            strata_column = "SEX_0",
                            strata_exclude = "^femm", selection_type = "regex",
                            also_print = TRUE),
               regexp = "Internal error: No value or value label.+strata_exclude")


  expect_warning(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm15"),
                            strata_column = "SEX_0",
                            strata_exclude = "^fem",
                            also_print = TRUE),
                 regexp = "The argument selection_type not provided.+expressions")


  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm16"),
                            strata_column = "SEX_0", selection_type = "value",
                            strata_exclude = c("0", "1"),
                            also_print = TRUE),
               regexp = "Internal error: No strata remain.+ones")


  expect_error(dq_report_by(sd0,
                            meta_data = md0,
                            dimensions = "int",
                            label_col = "VAR_NAMES",
                            cores = NULL,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm17"),
                            strata_column = "SEX_0",
                            also_print = TRUE),
               regexp = "The strata_column provided does not correpond.+item_level_metadata")

  #remove columns VAR_NAMES from metadata
  md1 <-  md0[ , !(names(md0) == VAR_NAMES)]
  expect_error(dq_report_by(sd0,
                            meta_data = md1,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm18"),
                            also_print = TRUE),
               regexp = "Did not find the mandatory column.+VAR_NAMES.+in the.+meta_data.+")

  #remove columns STUDY_SEGMENT from metadata
  md2 <-  md0[ , !(names(md0) == STUDY_SEGMENT)]
  expect_error(dq_report_by(sd0,
                            meta_data = md2,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = STUDY_SEGMENT,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm19"),
                            also_print = TRUE),
               regexp = "No metadata attribute.+STUDY_SEGMENT.+found for segmenting DQ report.")

  expect_error(dq_report_by(sd0,
                            meta_data = md2,
                            dimensions = "int",
                            cores = NULL,
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm20"),
                            also_print = TRUE),
               regexp = "No information for split provided.+segment_column = NULL.+")


  expect_error(dq_report_by(sd0,
                            meta_data = md2,
                            dimensions = "int",
                            cores = NULL,
                            strata_column = "sex_two",
                            meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                            output_dir = !!file.path(target, "sm21"),
                            also_print = TRUE),
               regexp = "The strata_column provided does not correpond.+item_level_metadata")

  expect_message2(dq_report_by(study_data = c("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS",
                                             "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
                              meta_data_v2 ="https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "comb1"),
                              cores = NULL,
                              also_print=TRUE,
                              segment_column = STUDY_SEGMENT,
                              segment_exclude = c("PHYS_EXAM", "LAB",
                                                  "INTERVIEW", "QUESTIONNAIRE"),
                              strata_column = "v00002",
                              strata_select = "females",
                              dimensions = "Integrity",
                              selection_type = "v_label",
                              subgroup = "[v00003]<40",
                              id_vars = "v00001"))


  sd1 <- head(prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
    20)

  expect_message2(dq_report_by(resp_vars = c("v00000", "v00001", "v00002",
                                            "v00003", "v00004", "v00005",
                                            "v00006", "v00018"),
                              study_data = sd1,
                              cores = NULL,
                              meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "comb2"),
                              also_print=TRUE,
                              segment_column = STUDY_SEGMENT,
                              segment_select = "^STU",
                              dimensions = "Integrity"))


  sd3 <- head(prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
    20)
  sd3 <- sd3[, head(colnames(sd3), 10)]

  md3 <- prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx | item_level")

  md3 <- md3[md3$VAR_NAMES %in% colnames(sd3), ]

  md3$STUDY_SEGMENT <- "STUDY"
  md3$DATAFRAMES <- "my_sd"

  df3 <- prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx | dataframe_level")

  df3$DF_ID_REF_TABLE <- NA
  df3$DF_RECORD_CHECK <- NA
  df3$DF_UNIQUE_ID <- NA
  df3$DF_CODE <- "my_sd"

  segl3 <- prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx | segment_level")

  segl3$SEGMENT_ID_TABLE <- NA
  segl3$SEGMENT_ID_VARS <- NA
  segl3$SEGMENT_RECORD_CHECK<- NA
  segl3$SEGMENT_UNIQUE_ROWS <- NA
  md3$MISSING_LIST_TABLE <- NA

  expect_message2(dq_report_by(sd3,
                              meta_data = md3,
                              meta_data_dataframe = df3,
                              meta_data_segment = segl3,
                              dimensions = "int",
                              cores = NULL,
                              segment_column = STUDY_SEGMENT,
                              segment_select = "^ST",
                              output_dir = !!file.path(target, "comb3"),
                              also_print = TRUE))



df3$DF_NAME <- "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"
expect_message2(dq_report_by(meta_data = md3,
                            meta_data_dataframe = df3,
                            cores = NULL,
                            segment_select = "STUDY",
                            output_dir = !!file.path(target, "comb4"),
                            also_print = TRUE ))
df3 <- rbind(df3,
             data.frame(DF_ELEMENT_COUNT = NA,
                        DF_RECORD_COUNT = NA,
                        DF_ID_REF_TABLE = NA,
                        DF_RECORD_CHECK = NA,
                        DF_UNIQUE_ID= NA,
                        DF_ID_VARS = "id",
                        DF_UNIQUE_ROWS = NA,
                        DF_CODE = "ship",
                        DF_NAME = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS"))


expect_message2(dq_report_by(meta_data = md3,
                            meta_data_dataframe = df3,
                            cores = NULL,
                            segment_select = "STUDY",
                            output_dir = !!file.path(target, "comb5"),
                            also_print = TRUE ))

md3$REPORT_NAME <- c("A", "B")
expect_message2(dq_report_by(sd3,
                            meta_data = md3,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = "REPORT_NAME",
                            segment_select = "A",
                            output_dir = !!file.path(target, "comb6"),
                            also_print = TRUE))

expect_error(dq_report_by(sd3,
                            meta_data = md3,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = "REPORTT_NAME",
                            output_dir = !!file.path(target, "comb7"),
                            also_print = TRUE),
             regexp = "No metadata.+for segmenting DQ report.")

expect_error(dq_report_by(sd3,
                          meta_data = md3,
                          dimensions = "int",
                          cores = NULL,
                          strata_column = "non-existent_strata",
                          output_dir = !!file.path(target, "comb8"),
                          also_print = TRUE),
             regexp = "The strata_column provided.+in the item_level_metadata")


md3 <- md3[, !colnames(md3) %in% "DATAFRAMES"]
df3 <- df3[, !colnames(df3) %in% "DF_CODE"]
expect_warning(dq_report_by(meta_data = md3,
                            meta_data_dataframe = df3,
                            cores = NULL,
                            segment_select = "STUDY",
                            output_dir = !!file.path(target, "comb9"),
                            also_print = TRUE ),
               regexp = "Because no id variable is available.+duplicated rows.")

#Without LABEL and VARIABLE_ROLE in metadata
sd1 <- head(prep_get_data_frame(
  "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
  20)
sd1 <- sd1[, head(colnames(sd1), 10)]

md1 <- prep_get_data_frame(
  "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx | item_level")

md1 <- md1[md1$VAR_NAMES %in% colnames(sd1), ]
md1 <- md1[, !colnames(md1) %in% c("LABEL", "VARIABLE_ROLE")]
md1$MISSING_LIST_TABLE <- NA

expect_message2(dq_report_by(study_data = sd1,
                            meta_data = md1,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = NULL,
                            output_dir = !!file.path(target, "test1"),
                            also_print = TRUE))

#only segment_select, but from STUDY_SEGMENT
expect_message2(dq_report_by(study_data = sd1,
                            meta_data = md1,
                            dimensions = "int",
                            cores = NULL,
                            segment_select = "PHYS_EXAM",
                            output_dir = !!file.path(target, "test2"),
                            also_print = TRUE))

#only segment_select but not present in STUDY_SEGMENT
expect_error(dq_report_by(study_data = sd1,
                          meta_data = md1,
                          dimensions = "int",
                          cores = NULL,
                          segment_select = "THIS_EXAM",
                          output_dir = !!file.path(target, "test3"),
                          also_print = TRUE),
             regexp = "segment_select values are.+not assigned.")


#only segment_exclude, but from STUDY_SEGMENT
expect_message2(dq_report_by(study_data = sd1,
                            meta_data = md1,
                            dimensions = "int",
                            cores = NULL,
                            segment_exclude = "PHYS_EXAM",
                            output_dir = !!file.path(target, "test4"),
                            also_print = TRUE))

#remove all segments with segment_exclude
expect_error(dq_report_by(study_data = sd1,
                            meta_data = md1,
                            dimensions = "int",
                            cores = NULL,
                            segment_column = "STUDY_SEGMENT",
                            segment_exclude = c("PHYS_EXAM", "STUDY"),
                            output_dir = !!file.path(target, "test5"),
                            also_print = TRUE),
             regexp = "No segment.+column level left after removing unwanted segments.+")

#only segment_exclude but not present in STUDY_SEGMENT
expect_error(dq_report_by(study_data = sd1,
                          meta_data = md1,
                          dimensions = "int",
                          cores = NULL,
                          segment_exclude = "THIS_EXAM",
                          output_dir = !!file.path(target, "test6"),
                          also_print = TRUE),
             regexp = "segment_exclude values.+not assigned.")


#no STUDY_SEGMENT and no segment_select provided
#Without LABEL and VARIABLE_ROLE in metadata
md2 <- prep_get_data_frame(
  "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx | item_level")

md2 <- md2[md2$VAR_NAMES %in% colnames(sd1), ]
md2 <- md2[, !colnames(md2) %in% "STUDY_SEGMENT"]
md2$MISSING_LIST_TABLE <- NA

expect_error(dq_report_by(study_data = sd1,
                          meta_data = md2,
                          dimensions = "int",
                          cores = NULL,
                          output_dir = !!file.path(target, "test7"),
                          also_print = TRUE),
             regexp = "No information for split provided..+segment_column = NULL.+")

})

test_that("con_inadmissible_categorical works", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]
  {
    IAVCatAll <- con_inadmissible_categorical(study_data = study_data,
                                              meta_data  = meta_data,
                                              label_col  = "LABEL")
  } %>% expect_message2(
    regexp = paste("All variables with VALUE_LABELS.+",
                           "in the metadata are used."),
    perl = TRUE
  ) %>% expect_message2(
    regexp = paste("The following variable.s.: .+_IAV.+flag.s.",
                           "inadmissible values."),
    perl = TRUE
  )

  expect_equal(sum(IAVCatAll$SummaryTable$GRADING), 5)
  expect_equal(sum(IAVCatAll$FlaggedStudyData$EDUCATION_1_IAV), 3)

  expect_silent(suppressWarnings({
    IAVCatAll <- con_inadmissible_categorical(study_data = study_data,
                                              meta_data  = meta_data,
                                              label_col  = "LABEL",
                                              resp_vars =
                                                c("MARRIED_0",
                                                  "SMOKING_0",
                                                  "PREGNANT_0"))

    IAVCatAll <- con_inadmissible_categorical(study_data = study_data,
                                              meta_data  = meta_data,
                                              label_col  = "LABEL",
                                              resp_vars = c("MARRIED_0"))
  }))

})

test_that("prep_add_to_meta works", {
  skip_on_cran()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  md <- prep_add_to_meta(VAR_NAMES = c("X", "Y"),
                         LABEL = c("x", "y"),
                         DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$FLOAT),
                         VALUE_LABELS = c("1 = female | 2 = male", NA),
                         LONG_LABEL = c("Ix", "Ypsilon"),
                         test = 3:4,
                         meta_data = meta_data
                        )
  expect_false("test" %in% colnames(md))
  expect_equal(nrow(md), 55)
  new <- md[md$VAR_NAMES %in% c("X", "Y"),
            c("VAR_NAMES", "LABEL", "DATA_TYPE", "VALUE_LABELS", "LONG_LABEL")]
  expect_identical(new,
                   structure(list(VAR_NAMES = c("X", "Y"), LABEL = c("x", "y"),
                                  DATA_TYPE = c("integer", "float"),
                                  VALUE_LABELS = c("1 = female | 2 = male", NA),
                                  LONG_LABEL = c("Ix", "Ypsilon")),
                             row.names = 54:55, class = "data.frame"))
})

test_that("int_unexp_elements works", {
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_on_cran()
  skip("TODO") # TODO
  sd1 <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS", keep_types = TRUE)
  md1 <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta.RDS")

  expect_error(
    int_unexp_elements(
      segments = c("INTRO", "INTERVIEW"),
      data_element_count = c(4, 11),
      meta_data = md1)
  )

  out <- int_unexp_elements(
    segments = c("INTRO", "INTERVIEW"),
    data_element_count = c(4, 11),
    study_data = sd1,
    meta_data = md1)

  expect_equal(out$SegmentData$Number.of.mistmatches[[1]], 0)

  testthat::local_edition(3)
  expect_snapshot_value(style = "deparse",

  )

})

test_that("dq_report_by works with content sm", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_on_cran() # slow test
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # TODO: test with all sorts of _by calls and on windows
  target <- withr::local_tempdir("testdqareportby")

  sd0 <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), 20)
  md0 <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx | item_level")

  # fewer segments to speed up things.
  # can be solved in a more elegant way after
  # https://gitlab.com/libreumg/dataquier/-/issues/481

  md0$STUDY_SEGMENT <- "STUDY"

  dq_report_by( sd0,
                meta_data = md0,
                dimensions = "int",
                cores = NULL,
                segment_column = STUDY_SEGMENT,
                strata_column = "SEX_0",
                meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                output_dir = file.path(target, "sm"),
                also_print = TRUE)

  rm(md0)

  sd0[2, "v00002"] <- 2
  expect_warning(dq_report_by(study_data = sd0,
                              meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              output_dir = !!file.path(target, "sm2"),
                              cores = NULL,
                              also_print=TRUE,
                              strata_column ="v00002",
                              segment_column = "STUDY_SEGMENT",
                              strata_select = "1",
                              segment_select = "STUDY",
                              dimensions = "Integrity"),
                 "The stratum/strata.+not present in the metadata")

  # TODO: check target path for all expected artifacts, namely the dq2 files, the .output_dirs containing the HTML reports and the summary RDS files

})

test_that("acc_univariate_outlier works without label_col", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]

  sd1 <- study_data # no outliers at all
  set.seed(234325)
  sd1$v00014 <- rnorm(nrow(sd1), sd = 0.01)
  sd1 <- sd1[sd1$v00014 < 0.02 & sd1$v00014 > -0.02, ]
  expect_message2(
    no_ol_det <- acc_univariate_outlier(resp_vars = "v00014", study_data = sd1,
                                        meta_data = meta_data, max_non_outliers_plot = 0),
    regexp =
      paste("For .+v00014.+, 0 from 1466 non-outlier data values",
            "were sampled to avoid large plots."))

  expect_match(
    ggplot2::ggplot_build(no_ol_det$SummaryPlotList$v00014)$data[[1]]$label,
    "No outliers detected for .+v00014.+",
    perl = TRUE
  )

  expect_message2(invisible(
    acc_univariate_outlier(resp_vars = "v00014", study_data = study_data,
                           meta_data = meta_data, n_rules = cars)),
    regexp =
      "The formal n_rules is not an integer between 1 and 4, default .4. is used.")

  expect_message2(invisible(
    acc_univariate_outlier(resp_vars = "v00014", study_data = study_data,
                           meta_data = meta_data,
                           max_non_outliers_plot = 100)),
    regexp =
      paste("For .+v00014.+, 100 from 2633 non-outlier data values",
            "were sampled to avoid large plots."))

  expect_message2(invisible(
    acc_univariate_outlier(resp_vars = "v00014", study_data = study_data,
                           meta_data = meta_data,
                           max_non_outliers_plot = -42)),
    regexp =
      paste("The formal max_non_outliers_plot is not an",
            "integer >= 0, default .10000. is used."))

  sd1 <- study_data[, c("v00103", "v00001", "v00013")]
  md1 <- subset(meta_data, VAR_NAMES %in% colnames(sd1))
  md1$JUMP_LIST <- "|"
  md1$MISSING_LIST <- "|"

  expect_message2(
    expect_error(invisible(
      acc_univariate_outlier(study_data = sd1,
                             meta_data = md1)),
      regexp =
        paste("No variables with suitable data type defined.")),
    regexp = "The following variables: were selected.",
    all = !TRUE,
    fixed = TRUE
  )

  md0 <- meta_data
  md0$DATA_TYPE <- NULL
  expect_error(invisible(
    acc_univariate_outlier(study_data = study_data,
                           meta_data = md0, n_rules = 4)),
    regexp = "Missing columns .+DATA_TYPE.+ from .+meta_data.+"
  )

  md0 <- meta_data
  md0$DATA_TYPE[[1]] <- NA
  expect_message2(
    expect_warning(
      invisible(
        acc_univariate_outlier(study_data = study_data,
                               meta_data = md0, n_rules = 4))
      ,
      regexp = sprintf("(%s)",
                       paste(
                         "For the variables.+?v00000.+?I have no valid.+?DATA_TYPE.+?in the",
                         ".+?meta_data.+?I.+?ve predicted the.+?DATA_TYPE.+?from",
                         "the.+?study_data.+?yielding.+?v00000 = integer.+?"
                       )
      ),
      perl = TRUE
    ),
    perl = TRUE,
    regexp = sprintf("(?ms).*(%s|%s|%s).*", paste("The following variables: v00000, v00002, v00003,",
                                                  "v01003, v01002, v10000, v00004, v00005, v00006, v00007, v00009, v00109,",
                                                  "v00010, v20000, v00014, v00015, v00016, v30000, v00018, v01018, v00019,",
                                                  "v00020, v00021, v00022, v00023, v00024, v00025, v00026, v00027, v00028,",
                                                  "v00029, v00030, v00031, v40000, v00034, v00035, v00036, v00037, v00038,",
                                                  "v00039, v00040, v00041, v50000.*"),
                     paste("Variables:.+?v00000.+?,.+?v00002.+?,.+?v00003.+?,.+?v01003.+?,.+?v01002.+?,.+?v10000.+?,.+?v00004.+?,.+?v00005.+?,.+?v00007.+?,.+?v00009.+?,.+?v00109.+?,.+?v00010.+?,.+?v20000.+?,.+?v00015.+?,.+?v00016.+?,.+?v30000.+?,.+?v00018.+?,.+?v01018.+?,.+?v00019.+?,.+?v00020.+?,.+?v00021.+?,.+?v00022.+?,.+?v00023.+?,.+?v00024.+?,.+?v00025.+?,.+?v00026.+?,.+?v00027.+?,.+?v00028.+?,.+?v00029.+?,.+?v00030.+?,.+?v00031.+?,.+?v40000.+?,.+?v00034.+?,.+?v00035.+?,.+?v00036.+?,.+?v00037.+?,.+?v00038.+?,.+?v00039.+?,.+?v00040.+?,.+?v00041.+?,.+?v50000.+?show integer values only, but will be nonetheless considered."),
                     paste("The variables.+?v00000.+?,.+?v00002.+?,.+?v01002.+?,.+?v10000.+?,.+?v00007.+?,.+?v00109.+?,.+?v00010.+?,.+?v20000.+?,.+?v30000.+?,.+?v00018.+?,.+?v01018.+?,.+?v00019.+?,.+?v00020.+?,.+?v00022.+?,.+?v00023.+?,.+?v00024.+?,.+?v00025.+?,.+?v00028.+?,.+?v00029.+?,.+?v00030.+?,.+?v40000.+?,.+?v50000.+? are neither float nor integer without VALUE_LABELS. Ignoring those.*")
    )
  )

  res1 <-
    acc_univariate_outlier(resp_vars = "v00014", study_data = study_data,
                           meta_data = meta_data)

  expect_true(all(
    c("SummaryTable",
      "SummaryPlotList"
    ) %in% names(res1)))

  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE) -2699 - 112.45)), 0.1
  )

  expect_equal(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryPlotList$v00014$data)),
      na.rm = TRUE) - 8005.172)), 0
  )

  expect_identical(colnames(res1$SummaryTable),
                   c("Variables", "Mean", "No.records", "SD", "Median", "Skewness",
                     "Tukey (N)",  "3SD (N)", "Hubert (N)",
                     "Sigma-gap (N)", "NUM_acc_ud_outlu",  "Outliers, low (N)",
                     "Outliers, high (N)", "GRADING", "PCT_acc_ud_outlu"))
})

test_that("acc_univariate_outlier works with label_col", {
  skip_on_cran() # slow, errors obvious
  skip_if_translated()
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data2 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)
  meta_data[[SCALE_LEVEL]] <-
    setNames(meta_data2[[SCALE_LEVEL]], nm = meta_data2[[VAR_NAMES]])[
      meta_data[[VAR_NAMES]]
    ]

  md0 <- meta_data
  md0$DATA_TYPE <- NA
  {
    res1 <-
      acc_univariate_outlier(resp_vars = c("USR_SOCDEM_0", "CRP_0"),
                             study_data = study_data,
                             meta_data = md0, label_col = LABEL)
  } %>%
    expect_message2("Argument.+resp_vars.+Variable.+USR_SOCDEM_0.+string.+does not have an allowed type.+integer.+float.+", perl = TRUE) %>%
    expect_message2("Argument.+resp_vars.+Variable.+USR_SOCDEM_0.+nominal.+does not have an allowed scale level.+interval.+ratio.+", perl = TRUE) %>%
    expect_warning("I've predicted the.+DATA_TYPE.+from the.+study_data", perl = TRUE) %>%
    expect_warning("resp_vars.+USR_SOCDEM_0.+were excluded", perl = TRUE) %>%
    expect_warning("resp_vars.+USR_SOCDEM_0.+were excluded", perl = TRUE)

  {
    res1 <-
      acc_univariate_outlier(resp_vars = c("USR_SOCDEM_0", "CRP_0"),
                             study_data = study_data,
                             meta_data = md0, label_col = LABEL,
                             exclude_roles = "XXX")
    } %>%
    expect_message2("Argument.+resp_vars.+Variable.+USR_SOCDEM_0.+string.+does not have an allowed type.+integer.+float.+", perl = TRUE) %>%
    expect_message2("Argument.+resp_vars.+Variable.+USR_SOCDEM_0.+nominal.+does not have an allowed scale level.+interval.+ratio.+", perl = TRUE) %>%
    expect_warning("Specified VARIABLE_ROLE not in meta_data. No exclusion applied.") %>%
    expect_warning(".*predicted.+DATA_TYPE.*", perl = TRUE) %>%
    expect_warning("resp_vars.+USR_SOCDEM_0.+were excluded", perl = TRUE) %>%
    expect_warning("resp_vars.+USR_SOCDEM_0.+were excluded", perl = TRUE)


  {
    res1 <-
      acc_univariate_outlier(
        resp_vars = c("USR_SOCDEM_0", "CRP_0",
                      "CENTER_0", "SEX_0", "AGE_0"),
        study_data = study_data,
        meta_data = md0,
        label_col = LABEL,
        exclude_roles = c("intro", "process")
      )
  } %>%
    expect_warning(regexp =
                     paste("variables.+SEX_0.+USR_SOCDEM_0.+CENTER_0.+were",
                           "excluded."), perl = TRUE) %>%
    expect_warning(paste("variables.+USR_SOCDEM_0.+were excluded."),
                   perl = TRUE) %>%
    expect_warning(paste("For the variables",
                         ".*predicted.+DATA_TYPE.*from the .study_data."),
                   perl = TRUE) %>%
    expect_message2(paste(
      "Argument.+resp_vars.+Variable.+USR_SOCDEM_0.+string.+does",
      "not have an allowed type.+integer.+float.+"
    ), perl = TRUE) %>%
    expect_message2(paste(
      "Argument.+resp_vars.+Variable.+USR_SOCDEM_0.+nominal.+does not have an allowed scale level.+interval.+ratio.+"
    ), perl = TRUE) %>%
    expect_message2(paste(
      "Argument.+resp_vars.+Variable.+SEX_0.+nominal.+does not have an allowed scale level.+interval.+ratio.+"
    ), perl = TRUE) %>%
    expect_message2(paste("Argument.+resp_vars.+Variable.+CENTER_0.+nominal.+does not have an allowed scale level.+interval.+ratio.+"
    ), perl = TRUE) %>%
    expect_message2(paste("Study variables.+AGE_0.+have",
                         "been excluded."), perl = TRUE)

  sd0 <- study_data
  sd0$v00014 <- NA_integer_
  expect_message2(
    #    expect_warning(
    expect_error(
      res1 <-
        acc_univariate_outlier(resp_vars = c("CRP_0", "AGE_0"),
                               study_data = sd0,
                               meta_data = meta_data, label_col = LABEL,
                               exclude_roles = c("intro", "process")),
      regexp = "No data left, aborting",
      perl = TRUE
    ),
    #   regexp = "In .resp_vars.+variables .CRP_0. were excluded",
    #   perl = TRUE
    # ),
    regexp = sprintf("(%s|%s|%s)",
                     paste(
                       "Argument.+resp_vars.+Variable.+CRP_0.+nominal.+does not have an allowed scale level.+interval.+ratio.+"
                     ),
                     paste("Study variables.+AGE_0.+have",
                           "been excluded."),
                     paste("Variables: .+CRP_0.+show integer values only,",
                           "but will be nonetheless considered.")
    ),
    perl = TRUE
  )


  sd0 <- study_data
  sd0$v00014 <- NA
  expect_message2( # no data left v14(crp0) wird rausgworfen, weil in diesem test NA gesetzt. komsich: je nach option ist sie dann noch in den resp_vars oder nicht.
    expect_error(
      res1 <-
        acc_univariate_outlier(resp_vars = c("CRP_0", "AGE_0"),
                               study_data = sd0,
                               meta_data = meta_data, label_col = LABEL,
                               exclude_roles = c("intro", "process")),
      regexp = "No data left, aborting.",
      perl = TRUE
    ),
    regexp = sprintf("(%s|%s)",
                     paste("Variables .+CRP_0.+ are not of type float or",
                           "integer and will be removed from univariate",
                           "outlier analysis."),
                     paste("Study variables:",
                           ".+AGE_0.+ have been excluded.")
    ),
    perl = TRUE
  )

  res1 <-
    acc_univariate_outlier(resp_vars = "CRP_0", study_data = study_data,
                           meta_data = meta_data, label_col = LABEL,
                           n_rules = 1)
  expect_identical(res1$SummaryTable$GRADING, 1)

  res1 <-
    acc_univariate_outlier(resp_vars = "CRP_0", study_data = study_data,
                           meta_data = meta_data, label_col = LABEL)

  expect_true(all(
    c("SummaryTable",
      "SummaryPlotList"
    ) %in% names(res1)))

  expect_lt(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryTable)),
      na.rm = TRUE) -2699 - 112.45)), 0.1
  )

  expect_equal(
    suppressWarnings(abs(sum(as.numeric(
      as.matrix(res1$SummaryPlotList$CRP_0$data)),
      na.rm = TRUE) - 8005.172)), 0
  )

  expect_identical(colnames(res1$SummaryTable),
                   c("Variables", "Mean", "No.records", "SD", "Median", "Skewness",
                     "Tukey (N)",  "3SD (N)", "Hubert (N)",
                     "Sigma-gap (N)", "NUM_acc_ud_outlu",  "Outliers, low (N)",
                     "Outliers, high (N)", "GRADING", "PCT_acc_ud_outlu"))

})

test_that("dq_report_by error_arguments_test", {
  skip_on_cran() # slow test
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")
  skip_if_not_installed("storr")

  target <- withr::local_tempdir("testdqareporstorr")
  storr_dir <- withr::local_tempdir("testdqareporstorr_dir")

  sd0 <- head(prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
    20)
  sd0 <- sd0[, head(colnames(sd0), 10)]

  md0 <- prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx | item_level")

  md0 <- md0[md0$VAR_NAMES %in% colnames(sd0), ]
  # fewer segments to speed up things.
  # can be solved in a more elegant way after
  # https://gitlab.com/libreumg/dataquier/-/issues/481

  md0$STUDY_SEGMENT <- "STUDY"


  expect_message2(dq_report_by(sd0,
               meta_data = md0,
               dimensions = "int",
               strata_column = "SEX_0",
               cores = NULL,
               storr_factory = prep_create_storr_factory(storr_dir, "TestReport"),
               meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
               output_dir = file.path(target, "sm5"),
               also_print = TRUE))

  expect_true(
    dir.exists(file.path(target, "sm5", "report_sd0_SEX_0_0_all_variables"))
  )
  expect_true(
    dir.exists(file.path(target, "sm5", "report_sd0_SEX_0_1_all_variables"))
  )

})

test_that("dq_report_by works with subgroup", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_on_cran() # slow test
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  # TODO: test with all sorts of _by calls and on windows
  target <- withr::local_tempdir("testdqareportby")

  sd1 <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE),
              20)

  expect_message2(dq_report_by(meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              study_data = sd1,
                              output_dir = !!file.path(target, "sb"),
                              also_print = TRUE,
                              cores = NULL,
                              label_col = "LABEL",
                              segment_column = "STUDY_SEGMENT",
                              segment_select = c("LAB"),
                              strata_column = "SEX_0",
                              strata_exclude = "0",
                              selection_type = "value",
                              subgroup = "[v00003] == 56",
                              dimensions = "Integrity"))

  expect_message2(dq_report_by(meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              study_data = sd1,
                              also_print = FALSE,
                              cores = NULL,
                              label_col = "LABEL",
                              segment_column = "STUDY_SEGMENT",
                              segment_select = c("LAB"),
                              strata_column = "SEX_0",
                              strata_exclude = "0",
                              selection_type = "value",
                              subgroup = "[AGE_0] == 56",
                              dimensions = "Integrity"))

  expect_error(dq_report_by(meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                              study_data = sd1,
                              also_print = FALSE,
                              cores = NULL,
                              label_col = "LABEL",
                              segment_column = "STUDY_SEGMENT",
                              segment_select = c("LAB"),
                              strata_column = "SEX_0",
                              strata_exclude = "0",
                              selection_type = "value",
                              subgroup = "[AGGGE_0] == 56",
                              dimensions = "Integrity"),
               regexp = "The subgroup rule.+not acceptable.+AGGGE_0")
})

test_that("dq_report_by error_respvars", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_on_cran() # slow test
  # TODO: test with all sorts of _by calls and on windows
  target <- withr::local_tempdir("testdqareportbyrespv")


  study_data <-
    head(prep_get_data_frame(
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"),
      20)



  expect_message2(dq_report_by(
    resp_vars = c("CENTER_0", "PSEUDO_ID",
                  "AGE_0", "SEX_0",
                  "ARM_CUFF_0",
                  "SBP_0",
                  "DBP_0", "CRP_0", "BSG_0" ),
    study_data = study_data,
    meta_data_v2 =
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
    cores = NULL,
    also_print = FALSE,
    dimensions = "Integrity",
    segment_column = "STUDY_SEGMENT"))

  dq_report_by(
    resp_vars = c("v00000", "v00001", "v00002", "v00003",
                  "v00010",
                  "v00004", "v00005",
                  "v00020", "v00021"),
    cores = NULL,
    study_data = study_data,
    meta_data_v2 =
      "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
    output_dir = file.path(target, "resp2"),
    also_print = TRUE,
    dimensions = NULL,
    segment_column = "STUDY_SEGMENT")

})

test_that("prep_acc_distributions_with_ecdf works", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  skip_if_not_installed("colorspace")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  expect_silent(
    r1 <- prep_acc_distributions_with_ecdf(resp_vars = "SBP_0",
                                           group_vars = "USR_BP_0",
                                           study_data = study_data,
                                           meta_data = meta_data)
  )

  expect_false(
    inherits(try(ggplot_build(r1$SummaryPlot)), "try-error"))

})

test_that("test_render_report_summary_1", {

  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
  skip_on_cran() # slow test
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  target <- withr::local_tempdir("testrendersummary")


  sd0 <- head(prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE), n = 20)
  sd0 <- sd0[, 1:15]

  md0 <- prep_get_data_frame(
    "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx | item_level")

  md0 <- md0[md0$VAR_NAMES %in% colnames(sd0), ]

  r1<- suppressWarnings(dq_report2(study_data = sd0,
                  cores = NULL,
                  meta_data = md0,
                  dimensions = c("Completeness")))

  sumrep <- summary(r1)

  skip_if_not_installed("htmltools")
  skip_if_not_installed("DT")
  skip_if_not_installed("rmarkdown")
  skip_if_not_installed("markdown")

  calnames <- print(sumrep, grouped_by = "call_names")
  indic <- print(sumrep, grouped_by = "indicator_metric")

  expect_equal(names(calnames[[6]][[2]][[3]][[1]][[1]]$data)[1:9],
               c("Variables",
                 "Labels",
                 "Descr stats-Cat",
                 "Descr stats-Cont",
                 "Data type error",
                 "Invalid encoding",
                 "Miss values-Item",
                 "Resp-rates-Item",
                 "Total" ))

  expect_equal(trimws(names(indic[[6]][[2]][[3]][[1]][[1]]$data)[1:8]),
               c("Variables",
                 "Labels",
                 "Data type mismatch (%)",
                 "Data type mismatch (cat.)",
                 "Inadmissible data format (%)",
                 "Inadmissible data format (N)",
                 "Missing values (%)",
                 "Total"))

})

#' removes empty rows from `x`
#'
#' @param x [data.frame] a data frame to be cleaned
#' @param id_vars [character] column names, that will be treated as empty
#'
#' @return [data.frame] reduced `x`
#'
#' @family missing_functions
#' @concept data_management
#' @noRd
util_remove_empty_rows <- function(x, id_vars = character(0)) {
  util_expect_scalar(id_vars,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)
  util_expect_data_frame(x, id_vars)
  x[rowSums(!util_empty(as.matrix(x[, setdiff(colnames(x), id_vars)]))) != 0, ,
    FALSE]
}

test_that("util_remove_empty_rows works", {
  skip_on_cran()

  result <- util_remove_empty_rows(cars)

  # Check that the result has the correct structure
  expect_s3_class(result, "data.frame")

  # Check that the result has the correct values when there are no empty rows
  expect_identical(result, cars)

  # Check that the result has the correct number of rows when there are empty rows
  # Set a seed for reproducibility
  set.seed(123)

  # Generate random indices to replace with empty string
  n_empty_strings <- 10
  indices_to_replace1 <- sample(1:length(cars$dist), size = n_empty_strings)

  # Replace values at selected indices with empty string
  cars_na <- cars
  cars_na$dist[indices_to_replace1] <- ""
  cars_na$speed[indices_to_replace1] <- ""

  result_na <- util_remove_empty_rows(cars_na)

  expect_equal(nrow(result_na), nrow(cars)-n_empty_strings)

})

#' Smoothes and plots adjusted longitudinal measurements and longitudinal trends
#' from logistic regression models
#'
#' @description
#' The following R implementation executes calculations for quality indicator
#' "Unexpected location" (see [here](
#' https://dataquality.qihs.uni-greifswald.de/PDQC_DQ_3_2_1_3.html
#' ). Local regression (LOESS) is a versatile statistical method to explore an
#' averaged course of time series
#' measurements (Cleveland, Devlin, and Grosse 1988). In context of
#' epidemiological data, repeated measurements using the same measurement
#' device or by the same examiner can be considered a time series. LOESS allows
#' to explore changes in these measurements over time.
#'
#' [Descriptor]
#'
#' @export
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable] the name of the continuous measurement variable
#' @param group_vars [variable] the name of the observer, device or reader
#'                             variable
#' @param time_vars [variable] the name of the variable giving the time
#'                             of measurement
#' @param co_vars [variable list] a vector of covariables for adjustment, for
#'                             example age and sex. Can be NULL (default) for no
#'                             adjustment.
#' @param min_obs_in_subgroup [integer] (optional argument) If `group_vars` is
#'                             specified, this argument can be used to specify
#'                             the minimum number of observations required for
#'                             each of the subgroups. Subgroups with fewer
#'                             observations are excluded. The default number
#'                             is `30`.
#' @param resolution [numeric] the maximum number of time points used for
#'                             plotting the trend lines
#' @param comparison_lines [list] type and style of lines with which trend
#'                             lines are to be compared. Can be mean +/- 0.5
#'                             standard deviation (the factor can be specified
#'                             differently in `sd_factor`) or quartiles
#'                             (Q1, Q2, and Q3). Arguments `color` and
#'                             `linetype` are passed to [ggplot2::geom_line()].
#' @param mark_time_points [logical] mark time points with observations
#'                             (caution, there may be many marks)
#' @param plot_observations [logical] show observations as scatter plot in the
#'                             background. If there are `co_vars` specified,
#'                             the values of the observations in the plot will
#'                             also be adjusted for the specified covariables.
#' @param plot_format [enum] AUTO | COMBINED | FACETS | BOTH. Return the plot
#'                             as one combined plot for all groups or as
#'                             facet plots (one figure per group). `BOTH` will
#'                             return both variants, `AUTO` will decide based
#'                             on the number of observers.
#' @param n_group_max [integer] maximum number of categories to be displayed
#'                  individually for the grouping variable (`group_vars`,
#'                  devices / examiners)
#' @param enable_GAM [logical] Can LOESS computations be replaced by general
#'                  additive models to reduce memory consumption  for large
#'                  datasets?
#' @param exclude_constant_subgroups [logical] Should subgroups with constant
#'                  values be excluded?
#' @param min_bandwidth [numeric] lower limit for the LOESS bandwidth, should be
#'                  greater than 0 and less than or equal to 1. In general,
#'                  increasing the bandwidth leads to a smoother trend line.
#' @param min_proportion [numeric] lower limit for the proportion of the smaller
#'                  group (cases or controls) for creating a LOESS figure,
#'                  should be greater than 0 and less than 0.4.
#'
#' @return a [list] with:
#'   - `SummaryPlotList`: list with two plots if `plot_format = "BOTH"`,
#'   otherwise one of the two figures described below:
#'     - `Loess_fits_facets`: The plot contains LOESS-smoothed curves
#'       for each level of the `group_vars` in a separate panel. Added trend
#'       lines represent mean and standard deviation or quartiles (specified
#'       in `comparison_lines`) for moving windows over the whole data.
#'     - `Loess_fits_combined`: This plot combines all curves into one
#'       panel. Given a low number of levels in the `group_vars`, this plot
#'       eases comparisons. However, if the number increases this plot may
#'       be too crowded and unclear.
#'
#' @details
#'
#' If `mark_time_points` or `plot_observations` is selected, but would result in
#' plotting more than 400 points, only a sample of the data will be displayed.
#'
#' Limitations
#'
#' The application of LOESS requires model fitting, i.e. the smoothness
#' of a model is subject to a smoothing parameter (span).
#' Particularly in the presence of interval-based missing data, high
#' variability of measurements combined with a low number of
#' observations in one level of the `group_vars` may distort the fit.
#' Since our approach handles data without knowledge
#' of such underlying characteristics, finding the best fit is complicated if
#' computational costs should be minimal. The default of
#' LOESS in R uses a span of 0.75, which provides in most cases reasonable fits.
#' The function `acc_loess` adapts the span for each level of the `group_vars`
#' (with at least as many observations as specified in `min_obs_in_subgroup`
#' and with at least three time points) based on the respective
#' number of observations.
#' LOESS consumes a lot of memory for larger datasets. That is why `acc_loess`
#' switches to a generalized additive model with integrated smoothness
#' estimation (`gam` by `mgcv`) if there are 1000 observations or more for
#' at least one level of the `group_vars` (similar to `geom_smooth`
#' from `ggplot2`).
#'
#' @importFrom ggplot2 ggplot aes scale_color_manual xlab ylab geom_point geom_line facet_wrap theme_minimal ggtitle theme element_blank
#' @importFrom stats as.formula lm loess lowess predict na.omit glm binomial poisson sd cov var runif
#'
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_loess.html
#' )
acc_loess <- function(resp_vars,
                      group_vars = NULL,
                      time_vars,
                      co_vars = NULL,
                      study_data,
                      label_col = VAR_NAMES,
                      item_level = "item_level",
                      min_obs_in_subgroup =
                        getOption("dataquieR.acc_loess.min_obs_in_subgroup",
                               dataquieR.acc_loess.min_obs_in_subgroup_default),
                      resolution = 80,
                      comparison_lines = list(type = c("mean/sd", "quartiles"),
                                              color = "grey30",
                                              linetype = 2,
                                              sd_factor = 0.5),
                      mark_time_points =
                        getOption("dataquieR.acc_loess.mark_time_points",
                                  dataquieR.acc_loess.mark_time_points_default),
                      plot_observations =
                        getOption("dataquieR.acc_loess.plot_observations",
                                  dataquieR.acc_loess.plot_observations_default),
                      plot_format =
                        getOption("dataquieR.acc_loess.plot_format",
                                  dataquieR.acc_loess.plot_format_default),
                      meta_data = item_level,
                      meta_data_v2,
                      n_group_max =
                        getOption("dataquieR.max_group_var_levels_in_plot",
                                  dataquieR.max_group_var_levels_in_plot_default),
                      enable_GAM = getOption("dataquieR.GAM_for_LOESS",
                                             dataquieR.GAM_for_LOESS_default),
                      exclude_constant_subgroups =
                        getOption("dataquieR.acc_loess.exclude_constant_subgroups",
                                  dataquieR.acc_loess.exclude_constant_subgroups_default),
                      min_bandwidth =
                        getOption("dataquieR.acc_loess.min_bw",
                                  dataquieR.acc_loess.min_bw_default),
                      min_proportion =
                        getOption("dataquieR.acc_loess.min_proportion",
                                  dataquieR.acc_loess.min_proportion_default)) {
  # preps ----------------------------------------------------------------------
  util_maybe_load_meta_data_v2()

  label_col <- attr(prep_get_labels("",
                                    item_level = meta_data,
                                    label_class = "LONG",
                                    label_col = label_col),
                    "label_col")

  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata = TRUE)

  util_correct_variable_use("resp_vars",
                            need_scale = "!na",
                            allow_all_obs_na = FALSE)
  util_correct_variable_use("group_vars",
                            need_scale = "nominal | ordinal",
                            allow_all_obs_na = TRUE,
                            allow_na = TRUE,
                            allow_null = TRUE)
  util_correct_variable_use("time_vars",
                            need_type = DATA_TYPES$DATETIME,
                            need_scale = "interval | ratio",
                            allow_all_obs_na = FALSE,
                            min_distinct_values = 3)
  util_correct_variable_use("co_vars",
                            overwrite = TRUE,
                            remove_not_found = TRUE,
                            allow_more_than_one = TRUE,
                            allow_all_obs_na = FALSE,
                            allow_na = TRUE,
                            allow_null = TRUE)

  # check that other arguments are specified correctly
  util_expect_scalar(
    min_obs_in_subgroup,
    check_type = util_is_numeric_in(min = 1, whole_num = TRUE, finite = TRUE),
    convert_if_possible = function(x) {
      x1 <- suppressWarnings(as.integer(x))
      if (is.na(x1) ||
          !util_is_numeric_in(min = 1, whole_num = TRUE,
                              finite = TRUE)(x1)) {
        x1 <- 30L
        util_message(paste(
          "Argument min_obs_in_subgroup is not specified",
          "correctly and is set to 30 instead."),
          applicability_problem = TRUE)
      }
      x1
    },
    conversion_may_replace_NA = TRUE)
  util_expect_scalar(
    resolution,
    check_type = util_is_numeric_in(min = 3, whole_num = TRUE, finite = TRUE),
    convert_if_possible = function(x) {
      x1 <- suppressWarnings(as.integer(x))
      if (is.na(x1) ||
          !util_is_numeric_in(min = 3, whole_num = TRUE,
                              finite = TRUE)(x1)) {
        x1 <- 80L
        util_message(
          paste("Argument resolution is not specified",
                "correctly and is set to 80 instead."),
          applicability_problem = TRUE)
      }
      x1
    },
    conversion_may_replace_NA = TRUE)
  util_expect_scalar(n_group_max,
                     check_type = util_is_numeric_in(min = 1, whole_num = TRUE))
  util_expect_scalar(enable_GAM, check_type = is.logical)
  util_expect_scalar(exclude_constant_subgroups, check_type = is.logical)
  util_expect_scalar(plot_format,
                     check_type = function(x) {
                       is.character(x) &&
                         any(grepl(x,
                                   c("AUTO", "COMBINED", "FACETS", "BOTH",
                                     "auto", "combined", "facets", "both" ),
                                   fixed = TRUE)) })
  util_expect_scalar(
    min_bandwidth,
    check_type = function(x) {
      util_is_numeric_in(min = 0, max = 1)(x) && x != 0 })
  util_expect_scalar(min_proportion,
                     check_type = util_is_numeric_in(min = 0, max = 0.4))

  # check data properties to choose a suitable method
  ds1 <- ds1[, c(resp_vars, time_vars, group_vars, co_vars)]
  ds1 <- ds1[complete.cases(ds1), ]
  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.",
               applicability_problem = TRUE)
  }
  var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])
  scl <- meta_data[[SCALE_LEVEL]][meta_data[[label_col]] == resp_vars]

  # generate a LOESS plot using a suitable method ------------------------------
  if (var_prop$NDistinct > 9 &&
      scl %in% c(SCALE_LEVELS$RATIO, SCALE_LEVELS$INTERVAL)) {
    SummaryPlotList_from_util <-
      util_acc_loess_continuous(resp_vars = resp_vars,
                                group_vars = group_vars,
                                time_vars = time_vars,
                                co_vars = co_vars,
                                study_data = study_data,
                                meta_data = meta_data,
                                label_col = label_col,
                                min_obs_in_subgroup = min_obs_in_subgroup,
                                resolution = resolution,
                                comparison_lines = comparison_lines,
                                mark_time_points = mark_time_points,
                                plot_observations = plot_observations,
                                plot_format = plot_format,
                                n_group_max = n_group_max,
                                enable_GAM = enable_GAM,
                                exclude_constant_subgroups =
                                  exclude_constant_subgroups,
                                min_bandwidth = min_bandwidth)
  } else if ((var_prop$NDistinct < 10 &&
              scl %in% c(SCALE_LEVELS$RATIO, SCALE_LEVELS$INTERVAL)) ||
             scl %in% c(SCALE_LEVELS$NOMINAL, SCALE_LEVELS$ORDINAL)) {
    SummaryPlotList_from_util <-
      util_acc_loess_bin(resp_vars = resp_vars,
                         group_vars = group_vars,
                         time_vars = time_vars,
                         co_vars = co_vars,
                         study_data = study_data,
                         meta_data = meta_data,
                         label_col = label_col,
                         min_obs_in_subgroup = min_obs_in_subgroup,
                         resolution = resolution,
                         plot_format = plot_format,
                         n_group_max = n_group_max,
                         enable_GAM = enable_GAM,
                         exclude_constant_subgroups =
                           exclude_constant_subgroups,
                         min_bandwidth = min_bandwidth,
                         min_proportion = min_proportion)
  } else {
    util_error("Variable '%s' has a disallowed scale level (%s)",
               dQuote(resp_vars),
               dQuote(tolower(trimws(scl))),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }
  return(util_attach_attr(SummaryPlotList_from_util,
                          as_plotly = "util_as_plotly_acc_loess"))
}

#' @family plotly_shims
#' @concept plotly_shims
#' @noRd
util_as_plotly_acc_loess <- function(res, ...) {
  util_ensure_suggested("plotly")
  if ("SummaryPlotList" %in% names(res)) {
    SummaryPlot <- util_remove_dataquieR_result_class(res$SummaryPlotList[[1]])
  } else {
    SummaryPlot <- util_remove_dataquieR_result_class(res$SummaryPlot)
  }
  build_obj <- ggplot2::ggplot_build(SummaryPlot)
  plot_obj <- util_gg_get(build_obj, "plot")
  labels_obj <- if (!is.null(plot_obj)) util_gg_get(plot_obj, "labels") else NULL
  title <- if (!is.null(labels_obj)) util_gg_get(labels_obj, "title") else NULL
  subtitle <- if (!is.null(labels_obj)) util_gg_get(labels_obj, "subtitle") else NULL
  sc <- plot_obj$scales$get_scales("colour")
  breaks <- sc$get_breaks()
  labels <- setNames(sc$get_labels(breaks), nm = breaks)
  py <- util_ggplotly(SummaryPlot, ...)
  for (i in seq_along(labels)) {
    if (py$x$data[[i]]$name %in% names(labels))
    py$x$data[[i]]$name <-
      labels[py$x$data[[i]]$name]
  }
  if (!is.null(title) && is.null(subtitle)) {
    py <- plotly::layout(py, title = title)
  } else if (!is.null(title) && !is.null(subtitle)) {
    # if (packageVersion("plotly") > "4.10.4") { # should be supported, soon
    #   py <- plotly::layout(py, title = list(text = title,
    #                                       subtitle = list(text = subtitle)))
    # } else {
    py <- plotly::layout(
      py,
      margin = list(b = 100),
      title = list(
        text = paste0(
          "<sub>",
          title,
          "\u00a0\u2014\u00a0",
          subtitle,
          "</sub>"
        )
      )
    )
  }
  py
}

#' Remove tables referred to by metadata and use `SVG` for most figures
#'
#' @param x an object to un-disclose
#' @param ... further arguments, used for pointing to the `dataquieR_result`
#'            object, if called recursively
#'
#' @return undisclosed object
#' @noRd
util_undisclose <- function(x, ...) {
  UseMethod("util_undisclose")
}

#' @export
util_undisclose.dq_lazy_ggplot <- function(x, ...) {
  util_undisclose(prep_realize_ggplot(x), ...)
}

#' @export
util_undisclose.default <- function(x, ...) {
  if (is.atomic(x)) {
    return(x)
  }
  util_error("Internal error: object of class %s in report.",
             util_pretty_vector_string(quote = sQuote, class(x)))
}

#' @export
util_undisclose.dataquieR_resultset2 <- function(x, ...) {
  if ("cores" %in% names(list(...))) {
    cores <- list(...)$cores
  }
  if ((("cores" %in% names(list(...))) ||
      .called_in_pipeline) && rlang::is_integerish(dynGet("cores")) &&
      as.integer(dynGet("cores")) > 1 &&
      suppressWarnings(util_ensure_suggested("parallel", err = FALSE))) {
    mycl <- parallel::makePSOCKcluster(as.integer(dynGet("cores")))
    parallel::clusterCall(mycl, library, "dataquieR", character.only = TRUE)
    parallel::clusterCall(mycl, loadNamespace, "hms")
    on.exit(parallel::stopCluster(mycl))
  } else {
    mycl <- parallel::getDefaultCluster()
  }

  my_tabs <- lapply(setNames(nm = names(attr(x, "referred_tables"))),
                   function(dfn) {
                      data.frame(`NA` = paste(dQuote(dfn), "is not available."),
                                              check.names = FALSE)
                  })

  attr(x, "referred_tables")[] <- my_tabs

  x[] <- util_par_lapply_lb(cl = mycl,
                            x, util_undisclose, ...)
  return(x)
}

#' @export
util_undisclose.dataquieR_result <- function(x, ...) {
  if (length(setdiff(class(x),
                     c("dataquieR_result", "list", "dataquieR_NULL",
                       "master_result", "Slot"))) > 0) {
    return(NextMethod())
  }
  dataquieR_result <- x
  if ("PlotlyPlot" %in% names(x)) {
    # class plotly
    if (any(endsWith(setdiff(names(x), "PlotlyPlot"), "Plot")) ||
        any(endsWith(setdiff(names(x), "PlotlyPlot"), "PlotList"))) {
      x$PlotlyPlot <- NULL
    } else {

      # ensure, sizing hint sticks at the dqr, only
      fixed <- util_fix_sizing_hints(dqr = dataquieR_result, x = x$PlotlyPlot)

      x$SummaryPlot <- try(util_plotly2svg_object(x$PlotlyPlot,
                                                  sizing_hints =
                                                    attr(fixed$dqr,
                                                         "sizing_hints")),
                            silent = TRUE)
      if (util_is_try_error(x$SummaryPlot)) {
        util_warning(
          c("Could not convert a plotly to an SVG or PNG for",
            "undisclosing data. Will delete an output slot. Maybe, a",
            "suggested package is missing: %s"), sQuote(conditionMessage(
              attr(x$SummaryPlot, "condition")
            )))
        x$SummaryPlot <- NULL
      }
      x$PlotlyPlot <- NULL
    }
  }
  x[] <- lapply(x, util_undisclose, dataquieR_result = dataquieR_result, ...)
  return(x)
}

#' @export
util_undisclose.list <- function(x, ...) {
  x[] <- lapply(x, util_undisclose, ...)
  return(x)
}

#' @export
util_undisclose.Slot <- function(x, ...) {
  if (length(setdiff(class(x),
                     c("dataquieR_result", "list", "dataquieR_NULL",
                       "master_result", "Slot"))) > 0) {
    return(NextMethod())
  }
  x[] <- lapply(x, util_undisclose, ...)

  return(x)
}

#' @export
util_undisclose.gg <- function(x, ...) {
  dataquieR_result <- list(...)[["dataquieR_result"]]
  if (util_is_svg_object(x)) {
    return(x)
  }
  fixed <- util_fix_sizing_hints(dqr = dataquieR_result, x = x)
  return(suppressWarnings(util_plot2svg_object(x, sizing_hints =
                                                 attr(fixed$dqr,
                                                      "sizing_hints"))))
}

#' @export
util_undisclose.util_pairs_ggplot_panels <- util_undisclose.gg

#' @export
util_undisclose.svg_plot_proxy <- util_undisclose.gg

#' @export
util_undisclose.ggmatrix_plot_obj <- util_undisclose.gg

#' @export
util_undisclose.ggmatrix_fn_with_params <- util_undisclose.gg

#' @export
util_undisclose.ggplot_built <- util_undisclose.gg

#' @export
util_undisclose.data.frame <- function(x, ...) {
  return(x)
}

#' Remove data disclosing details
#'
#' new function: no warranty, so far.
#'
#' @param x an object to un-disclose, a
#' @param cores can be an integer with a number of cores to use. if not
#'              specified, the function uses the default cluster, if available
#'              and falls back to serial un-disclosing, otherwise.
#'
#' @return undisclosed object
#' @export
prep_undisclose <- function(x, cores) {
  if (!(inherits(x, "dataquieR_resultset2") ||
        inherits(x, "dataquieR_result"))) {
    util_error("%s works for results or reports, only",
               sQuote("prep_undisclose")
               )
  }
  util_message("%s comes without any warranty, so far",
               sQuote("prep_undisclose"));
  if (missing(cores)) {
    return(suppressMessages(util_undisclose(x)))
  } else {
    return(suppressMessages(util_undisclose(x, cores = cores)))
  }

}

#' Fetch a missing code list from the metadata
#'
#' get missing codes from metadata (e.g. [MISSING_LIST] or [JUMP_LIST])
#'
#' @param x [variable] the name of the variable to retrieve code lists for.
#'                     only one variable at a time is supported, *not*
#'                     vectorized!!
#' @param code_name [variable attribute] [JUMP_LIST] or [MISSING_LIST]:
#'                                       Which codes to retrieve.
#' @param split_char [character] len = 1. Character(s) used to separate
#'                               different codes in the metadata, usually `|`,
#'                               as in `99999|99998|99997`.
#' @param mdf [data.frame] the data frame that contains metadata
#'                         attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param warning_if_no_list [logical] len = 1. If `TRUE`, a warning is
#'                                              displayed, if no missing
#'                                              codes are available for a
#'                                              variable.
#' @param warning_if_unsuitable_list [logical] len = 1. If `TRUE`, a warning is
#'                                              displayed, if missing
#'                                              codes do not match with a
#'                                              variable' data type.
#'
#' @return [numeric] vector of missing codes.
#'
#' @family missing_functions
#' @concept metadata_management
#' @noRd
util_get_code_list <- function(x, code_name, split_char = SPLIT_CHAR, mdf,
                               label_col = VAR_NAMES,
                               warning_if_no_list = TRUE,
                               warning_if_unsuitable_list = TRUE) {
  if (!(label_col %in% names(mdf))) {
    if (warning_if_no_list) {
      util_warning(
        "Metadata does not provide a column called %s for the labels.",
        dQuote(label_col),
        applicability_problem = TRUE)
    }
    return(numeric(0))
  }
  if (!(code_name %in% names(mdf))) {
    if (warning_if_no_list) {
      util_warning(
        c("Metadata does not provide a column called %s",
          "for replacing codes with NAs."),
        dQuote(code_name),
        applicability_problem = TRUE)
    }
    return(numeric(0))
  }

  c_list <- mdf[[code_name]][mdf[[label_col]] == x]

  if (length(c_list) != 1 || is.na(c_list)) {
    if (warning_if_no_list) {
      util_warning(
        c("Could not find %s for %s in the meta_data",
          "for replacing codes with NAs."),
        dQuote(code_name),
        dQuote(x),
        applicability_problem = TRUE)
    }
    return(numeric(0))
  }

  cl <- c_list[[1]] # ?

  if (length(cl) != 1) { # nocov start
    util_error(c("Internal error: Have more than one codelist for a",
                 "variable in the metadata -- this should not happen."))
  }  # nocov end

  # res <- unlist(strsplit(cl, split_char, fixed = TRUE))
  if (code_name != VALUE_LABELS) {
    r <- util_parse_assignments(cl, split_char = SPLIT_CHAR)
  } else {
    # FIXME: "Deprecated, should never be used"
    r <- util_parse_assignments(cl,
                                split_on_any_split_char = TRUE,
                                split_char = c(SPLIT_CHAR, '<'),)
  }
  if (length(r) == 0) {
    r <- setNames(character(0), character(0))
  } else {
    r <- setNames(unlist(r, recursive = FALSE), nm = names(r))
  }
  res <- names(r)
  res[util_empty(res)] <- NA
  if (DATA_TYPE %in% colnames(mdf) &&
      any(!is.na(mdf[[DATA_TYPE]])) &&
      all(na.rm = TRUE, mdf[[DATA_TYPE]][
        !is.na(mdf[[label_col]]) & mdf[[label_col]] == x] ==
      DATA_TYPES$DATETIME)) {
    dt_res <- suppressWarnings(util_parse_date(res))
    if (sum(is.na(res)) < sum(is.na(dt_res))) {
      if (warning_if_unsuitable_list) util_warning(
"Some codes (%s) were not datetime/assignment for %s: %s, these will be ignored",
        dQuote(code_name),
        dQuote(x),
        paste(sQuote(res[is.na(dt_res) != is.na(res)]),
              collapse = ", "),
        applicability_problem = TRUE)
    }
    names(dt_res) <- r
    r <- dt_res[!is.na(dt_res)]
  } else if (DATA_TYPE %in% colnames(mdf) &&
              any(!is.na(mdf[[DATA_TYPE]])) &&
              all(na.rm = TRUE, mdf[[DATA_TYPE]][
                !is.na(mdf[[label_col]]) & mdf[[label_col]] == x] ==
                DATA_TYPES$TIME)) {
    to_res <- suppressWarnings(util_parse_time(res))
    if (sum(is.na(res)) < sum(is.na(to_res))) {
      if (warning_if_unsuitable_list) util_warning(
        "Some codes (%s) were not time/assignment for %s: %s, these will be ignored",
        dQuote(code_name),
        dQuote(x),
        paste(sQuote(res[is.na(to_res) != is.na(res)]),
              collapse = ", "),
        applicability_problem = TRUE)
    }
    names(to_res) <- r
    r <- to_res[!is.na(to_res)]
  } else {
    numeric_res <- suppressWarnings(as.numeric(res))
    if (sum(is.na(res)) < sum(is.na(numeric_res))) {
      if (warning_if_unsuitable_list) util_warning(
"Some codes (%s) were not numeric/assignment for %s: %s, these will be ignored",
        dQuote(code_name),
        dQuote(x),
        paste(sQuote(res[is.na(numeric_res) != is.na(res)]),
              collapse = ", "),
        applicability_problem = TRUE)
    }
    names(numeric_res) <- r
    r <- numeric_res[!is.na(numeric_res)]
  }
  r
}

#' For a group of variables (original) the function provides all original plus
#' referred variables in the metadata and a new item_level metadata
#' including information on the original variables and the referred variables
#'
#' @param meta_data [data.frame] old name for `item_level`
#' @param meta_data_segment [data.frame] -- optional: Segment level metadata
#' @param meta_data_dataframe [data.frame] -- optional if `study_data` is present:
#'                                                     Data frame level metadata
#' @param meta_data_cross_item [data.frame] -- optional: Cross-item level
#'                                                                 metadata
#' @param meta_data_item_computation [data.frame] -- optional: Computed items
#'                                                                     metadata
#' @param resp_vars [variable list] the name of the original variables.
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param id_vars [variable] a vector containing the name/s of the variables
#'                            containing ids
#' @param vars_in_subgroup [variable] a vector containing the name/s of the
#'                                    variable/s mentioned inside the
#'                                    subgroup rule
#' @param strata_column [variable] name of a study variable used to stratify the
#'                                    report by and to add as referred variable
#'
#'
#' @return a named list containing the referred variables and
#' a new item_level metadata including information on the original variables
#' and the referred variables
#'
#' @noRd

util_referred_vars <- function(resp_vars,
                               id_vars = character(0),
                               vars_in_subgroup = character(0),
                               meta_data,
                               label_col,
                               meta_data_segment = NULL,
                               meta_data_dataframe = NULL,
                               meta_data_cross_item = NULL,
                               meta_data_item_computation = NULL,
                               strata_column = NULL) {
  #Preparation-----
    key_cols <- util_variable_references(meta_data)
  #checks ----
    #TODO: add checks

   # util_expect_data_frame(meta_data,
    #col_names = list(DATAFRAMES = is.character))

  #-----
  vars <- resp_vars
  repeat {
    # referred variables from key_cols----
    referred_vars <-
      unique(unlist(meta_data[meta_data[[VAR_NAMES]] %in%
                                vars, key_cols], recursive = TRUE))
    # referred variables from cross-item level metadata----
    referred_vars <-
      c(referred_vars, unlist(
        util_parse_assignments(meta_data_cross_item[[VARIABLE_LIST]],
                               multi_variate_text = TRUE)
      ))


    # referred variables from cross-item level metadata from *_LIMITS columns----
    # Define a vector of the LIMITS column names to iterate over
    limit_cols <- c("HARD_LIMITS", "SOFT_LIMITS", "DETECTION_LIMITS")

    # if at least one limit column is non-empty
    if (any(sapply(limit_cols, function(col)
      any(!util_empty(meta_data_cross_item[[col]]))))) {

      # Pre-calculate needles_var_names and needles once
      needles_var_names <- unique(c(meta_data[[VAR_NAMES]],
                                    meta_data[[label_col]],
                                    meta_data[[LABEL]],
                                    meta_data[[LONG_LABEL]],
                                    meta_data[["ORIGINAL_VAR_NAMES"]],
                                    meta_data[["ORIGINAL_LABEL"]]))
      needles <- paste0("[", needles_var_names, "]")

      results_list <- lapply(
        limit_cols,
        function(col) {
          if (any(!util_empty(meta_data_cross_item[[col]]))) {
            x <- vapply(setNames(needles, nm = needles_var_names),
                        grepl,
                        setNames(nm = meta_data_cross_item[[col]]),
                        fixed = TRUE,
                        FUN.VALUE = logical(length =
                                              nrow(meta_data_cross_item))
            )
            # Ensure x is a matrix
            if (is.vector(x)) {
              x <- as.matrix(t(x))
            }
            # Extract variable names, ensure unique and sorted per row
            variablelist_raw <-
              unname(lapply(as.data.frame(t(x)),
                            function(xx) unique(sort(colnames(x)[xx]))))
            # Collapse the list of variable names into a single string
            variablelist <-
              lapply(variablelist_raw,
                     paste0,
                     collapse = sprintf(" %s ", SPLIT_CHAR))
            return(variablelist)
          }
          return(NULL) # Return NULL if the limit column is empty
        }
      )
      variable_list_limits <- unique(unname(unlist(results_list)))
      #remove empty results
      variable_list_limits <- variable_list_limits[variable_list_limits != ""]
      # separate variables as single ones
      variable_list_limits <- lapply(X = variable_list_limits,
                                     FUN = util_parse_assignments)
      # get a vector of variables from all limits
      variable_list_limits <- unique(unname(unlist(variable_list_limits)))


      #merge the variablea to the referred_vars
      referred_vars <-
        unique(c(referred_vars, variable_list_limits))

    }

    # referred variables from meta_data_item_computation metadata----
    referred_vars <-
      c(referred_vars, unlist(
        util_parse_assignments(meta_data_item_computation[[VARIABLE_LIST]],
                               multi_variate_text = TRUE)
      ))

    # looks to all possible labels and
    # return var_names
    referred_vars <-
      unique(util_find_var_by_meta(unique(
        unlist(
          util_parse_assignments(referred_vars,
                                 multi_variate_text = TRUE)
        )
      ), meta_data = meta_data))
    #remove na's
    referred_vars <-
      referred_vars[!is.na(referred_vars)]

    if (all(referred_vars %in% vars)) {
      #if all variables are already present
      # than it stops
      break
    } else {
      # otherwise add the new variables and
      # restart the loop
      vars <-
        union(vars, referred_vars)
    }
  }

  resp_vars_plus_referred <- vars
  rm(vars)

  # Add the var of strata_column if present
  if (!is.null(strata_column)) {
    var_split <- strata_column
    names(var_split) <- NULL
    resp_vars_plus_referred <- unique(c(resp_vars_plus_referred, var_split))
    rm(var_split)
  }
  # Add the var of id_vars if present
  if (length(id_vars) != 0) {
    resp_vars_plus_referred <- unique(c(resp_vars_plus_referred, id_vars))
  }
  # Add the var of vars_in_subgroup if present
  if (length(vars_in_subgroup) != 0) {
    resp_vars_plus_referred <- unique(c(resp_vars_plus_referred,
                                        vars_in_subgroup))
  }
  # Create a temporary subset of item_level metadata with only
  # resp_vars and referred vars
  temp_subset <-
    meta_data[meta_data$VAR_NAMES %in% resp_vars_plus_referred,
              , drop = FALSE]
  not_resp_vars <- setdiff(resp_vars_plus_referred, resp_vars)

  # Give a warning if not all resp_vars are present in the item_level metadata
  if(length(setdiff(resp_vars, temp_subset$VAR_NAMES)) > 0 )  {
    util_warning(
      "Variables in resp_vars that are not present in the item_level metadata: %s",
      setdiff(resp_vars, temp_subset$VAR_NAMES),
      applicability_problem = TRUE
    )
  }

  # Add all ID vars from the dataframe_level metadata----
  # obtain the code of the data frames
  if(!is.null(meta_data_dataframe) ||
     nrow(meta_data_dataframe) > 0) {
    if(DATAFRAMES %in% colnames(meta_data) &&
       DF_CODE %in% colnames(meta_data_dataframe)) {
      itemlevel_subset_df <-
        temp_subset[, c(VAR_NAMES, DATAFRAMES), drop = FALSE]

      dfr_code <- unique(unname(unlist(
        util_parse_assignments(
          itemlevel_subset_df$DATAFRAMES,
          split_char = SPLIT_CHAR,
          multi_variate_text = TRUE) )))

      #obtain the id from the dataframe interested by the resp-vars and referred vars
      id_vars_referred <- meta_data_dataframe[meta_data_dataframe[[DF_CODE]] %in%
                                                dfr_code, DF_ID_VARS, drop = TRUE]
      rm(dfr_code)
    } else {
      #obtain the id from the dataframe interested by the resp-vars and referred vars
      id_vars_referred <- meta_data_dataframe[, DF_ID_VARS, drop = TRUE]
    }

  }  else {
    id_vars_referred <- character(0)
  }


if(!is.null(meta_data_segment)) {
    #Obtain the ID from segment level if present
    list_seg <-
      unique(temp_subset[, STUDY_SEGMENT, drop = TRUE])
    #obtain the id from the segment interested by the resp-vars and referred vars
    id_vars_seg_referred <- meta_data_segment[meta_data_segment[[STUDY_SEGMENT]] %in%
                                                list_seg, SEGMENT_ID_VARS, drop = TRUE]
    #Put together all the id vars
    id_vars_referred <- unique(c(id_vars_referred,id_vars_seg_referred))
    rm(list_seg, id_vars_seg_referred)

  }

   # reduce to vector without | in case multiple vars stated in one cell
  id_vars_referred <-
    unlist(sapply(id_vars_referred,
                  function(x) {
                    util_parse_assignments(x, split_char = c(SPLIT_CHAR),
                                           multi_variate_text = TRUE,
                                           split_on_any_split_char = TRUE)
                  })) #it also get rid of eventual NA in the id_vars_referred
  names(id_vars_referred) <- NULL
  id_vars_referred <- unique(id_vars_referred)

  #Create a new complete list of variables with resp_vars, referred_vars and id vars
  resp_vars_complete <- union(resp_vars_plus_referred, id_vars_referred)
  rm(resp_vars_plus_referred)



  #obtain the metadata for the vars----
  # Set the VARIABLE_ROLE of referred_vars as suppress
  vars_referred <- union(id_vars_referred, not_resp_vars)
  vars_referred <- setdiff(vars_referred, resp_vars)
  rm(id_vars_referred, not_resp_vars)


  md_complete <-
    meta_data[meta_data$VAR_NAMES %in% resp_vars_complete,
              , drop = FALSE]

  md_complete <- md_complete %>%
    dplyr::mutate(VARIABLE_ROLE = ifelse(VAR_NAMES %in% vars_referred,
                                         "suppress",
                                         VARIABLE_ROLE))


  # Create the new item level metadata with only rows of resp vars and referred+id
  rm(temp_subset, vars_referred, key_cols)

 return(list(vars_complete = resp_vars_complete,
             md_complete = md_complete))

}

#' Utility function single quote string
#'
#' This function generates usual single-quotes for each element of the character
#' vector.
#'
#' @param string Character vector
#'
#' @return quoted string
#'
#' @family string_functions
#' @concept process
#' @noRd
util_set_sQuoteString <- function(string) {
  old <- options(useFancyQuotes = FALSE)
  on.exit(options(old))
  sQuote(string)
}

#' utility function for the applicability of categorical admissibility
#'
#' Test for applicability of categorical admissibility
#'
#' @param x [data.frame] metadata
#' @param dta [logical] vector, 1=matching data type,
#'                  0 = non-matching data type
#'
#' @return [factor] 0-3 for each variable in metadata
#' - 0 data type mismatch and not applicable
#' - 1 data type mismatches but applicable
#' - 2 data type matches but not applicable
#' - 3 data type matches and applicable
#' - 4 not applicable because of not suitable data type
#'
#' @seealso [pro_applicability_matrix]
#' @noRd

util_app_iac <- function(x, dta) {
  if (VALUE_LABELS %in% names(x)) {
    c1 <- ifelse(!is.na(x[[VALUE_LABELS]]), 1, 0)
  } else {
    c1 <- rep(0, times = dim(x)[1])
  }

  if (VALUE_LABEL_TABLE %in% names(x)) {
    c1 <- c1 | ifelse(!is.na(x[[VALUE_LABEL_TABLE]]), 1, 0)
  }

  if (STANDARDIZED_VOCABULARY_TABLE %in% names(x)) {
    c1 <- c1 | ifelse(!is.na(x[[STANDARDIZED_VOCABULARY_TABLE]]), 1, 0)
  }

  if (DATA_TYPE %in% names(x)) {
    c2 <- ifelse(!is.na(x[[DATA_TYPE]]) &
                   (x[[DATA_TYPE]] %in% c(DATA_TYPES$INTEGER,
                                          DATA_TYPES$STRING))
                 , 1, 0)
  } else {
    c2 <- rep(0, times = dim(x)[1])
  }


  aa <- paste0(dta, as.integer(c1))
  score <- as.numeric(recode(as.factor(aa), "00" = 0, "01" = 1,
                             "10" = 2, "11" = 3))
  score[c2 == 0] <- 4
  score <- as.factor(score)
  return(score)
}

#' Prepare a vector four output
#'
#' @param v the vector
#' @param quote function, used for quoting -- `sQuote` or `dQuote`
#' @param n_max maximum number of elements of `v` to display, if not missing.
#'
#' @return the "pretty" collapsed vector as a string.
#'
#' @family string_functions
#' @concept process
#' @noRd
util_pretty_vector_string <- function(v, quote = dQuote, n_max = length(v)) {
  util_expect_scalar(
    v,
    allow_more_than_one = TRUE,
    allow_null = TRUE,
    allow_na = TRUE)
  if (length(v) == 0) {
    return("")
  }
  util_expect_scalar(n_max, check_type = util_is_numeric_in(min = 1,
                                                            finite = TRUE,
                                                            whole_num = TRUE))
  util_stop_if_not(is.function(quote))
  util_stop_if_not(length(formals(quote)) > 0)
  if (!length(v)) {
    return(quote(""))
  } else {
    if (any(is.na(v))) {
      v[is.na(v)] <- ""
    }
    if (length(v) > n_max) {
      v <- head(quote(v), n_max)
      v <- c(v, "...")
    } else {
      v <- quote(v)
    }
    return(paste0(v, collapse = ", "))
  }
}

#' Summarizes missingness for individuals in specific segments
#'
#' @description
#' ### This implementation can be applied in two use cases:
#'
#'  1. participation in study segments is not recorded by respective variables,
#'     e.g. a participant's refusal to attend a specific examination is not
#'     recorded.
#'  2. participation in study segments is recorded by respective
#'     variables.
#'
#' Use case *(1)* will be common in smaller studies. For the calculation of
#' segment missingness it is assumed that study variables are nested in
#' respective segments. This structure must be specified in the static metadata.
#' The R-function identifies all variables within each segment and returns TRUE
#' if all variables within a segment are missing, otherwise FALSE.
#'
#' Use case *(2)* assumes a more complex structure of study data and metadata.
#' The study data comprise so-called intro-variables (either TRUE/FALSE or codes
#' for non-participation). The column `PART_VAR` in the metadata is
#' filled by variable-IDs indicating for each variable the respective
#' intro-variable. This structure has the benefit that subsequent calculation of
#' item missingness obtains correct denominators for the calculation of
#' missingness rates.
#'
#' [Descriptor]
#'
#' @details
#' ### Implementation and use of thresholds
#' This implementation uses one threshold to discriminate critical from
#' non-critical values. If direction is above than all values below the
#' threshold_value are normal (displayed in dark blue in the plot and flagged
#' with GRADING = 0 in the dataframe). All values above the threshold_value are
#' considered critical. The more they deviate from the threshold the displayed
#' color shifts to dark red. All critical values are highlighted with GRADING =
#' 1 in the summary data frame. By default, highest values are always shown in
#' dark red irrespective of the absolute deviation.
#'
#' If direction is below than all values above the threshold_value are normal
#' (displayed in dark blue, GRADING = 0).
#'
#' ### Hint
#' This function does not support a `resp_vars` argument but `exclude_roles` to
#' specify variables not relevant for detecting a missing segment.
#'
#' List function.
#'
#' @inheritParams .template_function_indicator
#'
#' @param group_vars [variable] the name of a variable used for grouping,
#'                              defaults to *NULL* for not grouping output
#' @param strata_vars [variable] the name of a variable used for stratification,
#'                               defaults to NULL for not grouping output
#' @param meta_data_segment [data.frame] Segment level metadata. Optional.
#' @param threshold_value [numeric] from=0 to=100. a numerical value ranging
#'                                                 from 0-100
#' @param direction [enum] low | high. "high" or "low", i.e. are deviations
#'                                     above/below the threshold critical. This argument is deprecated and replaced by *color_gradient_direction*.
#' @param color_gradient_direction [enum] above | below. "above" or "below", i.e. are deviations
#'                                     above or below the threshold critical? (default: above)
#' @param exclude_roles [variable roles] a character (vector) of variable roles
#'                                       not included
#'
#' @param expected_observations [enum] HIERARCHY | ALL | SEGMENT. If ALL, all
#'                                     observations are expected to comprise
#'                                     all study segments. If SEGMENT, the
#'                                     `PART_VAR` is expected to point
#'                                     to a variable with values of 0 and 1,
#'                                     indicating whether the variable was
#'                                     expected to be observed for each data
#'                                     row. If HIERARCHY, this is also
#'                                     checked recursively, so, if a variable
#'                                     points to such a participation variable,
#'                                     and that other variable does has also
#'                                     a `PART_VAR` entry pointing
#'                                     to a variable, the observation of the
#'                                     initial variable is only
#'                                     expected, if both segment variables are
#'                                     1.
#' @param segment_level [data.frame] alias for `meta_data_segment`
#'
#' @return a list with:
#'   - `ResultData`: data frame about segment missingness
#'   - `SummaryPlot`: ggplot2 heatmap plot: a heatmap-like graphic that
#'                    highlights critical values depending on the respective
#'                    threshold_value and direction.
#'   - `ReportSummaryTable`: data frame underlying `SummaryPlot`
#'
#' @export
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_com_impl_segment_missingness.html
#' )
com_segment_missingness <- function(study_data,
                                    item_level = "item_level",
                                    strata_vars = NULL,
                                    group_vars = NULL,
                                    label_col,
                                    threshold_value,
                                    direction,
                                    color_gradient_direction,
                                    expected_observations = c("HIERARCHY",
                                                              "ALL",
                                                              "SEGMENT"),
                                    exclude_roles =
                                      c(VARIABLE_ROLES$PROCESS),
                                    meta_data = item_level,
                                    meta_data_v2,
                                    segment_level,
                                    meta_data_segment) {

  #########
  # STOPS #
  #########

  util_maybe_load_meta_data_v2()

  util_ck_arg_aliases()

  if (!missing(meta_data_segment)) {
    meta_data_segment <- util_expect_data_frame(meta_data_segment)
  } else {
    meta_data_segment <- data.frame()
  }

  expected_observations_missing <- missing(expected_observations)
  util_expect_scalar(expected_observations, allow_more_than_one = TRUE)
  expected_observations <- match.arg(expected_observations)
  util_expect_scalar(expected_observations)

  int_part_vars_structure(study_data = study_data,
                          meta_data = meta_data,
                          label_col = label_col,
                          expected_observations = expected_observations,
                          disclose_problem_paprt_var_data = FALSE)

  if (missing(threshold_value) ||
      length(threshold_value) != 1 ||
      !is.numeric(threshold_value)) {
    if (!missing(threshold_value) || !.called_in_pipeline) util_message(
      c("threshold_value should be a single number between 0 and 100.",
      "Invalid value specified, setting to 10%%."),
      applicability_problem = TRUE)
    threshold_value <- 10
  }

  if (missing(color_gradient_direction)) {
    if (!.called_in_pipeline) util_message(c(
      "No specification of color gradient direction found.",
      "The function interprets values above the threshold as violations."),
      applicability_problem = TRUE)
    color_gradient_direction <- "above"
  }

  if (length(color_gradient_direction) != 1) {
    util_error(
      "Parameter %s, if not missing, should be of length 1, but not %d.",
      dQuote("color_gradient_direction"), length(color_gradient_direction),
      applicability_problem = TRUE)
  }

  if (!all(color_gradient_direction %in% c("above", "below"))) {
    util_error(
      "Parameter %s should be either %s or %s, but not %s.",
      dQuote("color_gradient_direction"),
      dQuote("above"), dQuote("below"), dQuote(color_gradient_direction),
      applicability_problem = TRUE
    )
  }

  if (!(missing(direction))) {
    if (direction %in% c("high", "low")) {
      if ((direction == "low" && color_gradient_direction == "above") ||
          (direction == "high" && color_gradient_direction == "below")) {
        util_error(
          "Conflicting options for %s and %s (%s is deprecated).",
          dQuote("color_gradient_direction"),
          dQuote("direction"), dQuote("direction"),
          applicability_problem = TRUE
        )
      } else {
        util_warning("%s is deprecated.", dQuote("direction"),
                     applicability_problem = FALSE)
      }
    } else {
      util_warning("%s is deprecated.", dQuote("direction"),
                   applicability_problem = FALSE)
    }
  }


  ####################
  # PREPS AND CHECKS #
  ####################

  # map meta to study
  prep_prepare_dataframes()
  ds1_labelled <- prep_prepare_dataframes(.apply_factor_metadata_inadm = TRUE,
                                          .internal = FALSE)

  if (!expected_observations_missing && expected_observations != "ALL" &&
      !(PART_VAR %in% colnames(meta_data))) {
    util_message(c("For %s = %s, a column %s is needed in %s. Falling",
                   "back to %s = %s."),
                 sQuote("expected_observations"),
                 dQuote(expected_observations),
                 dQuote(PART_VAR),
                 sQuote("meta_data"),
                 sQuote("expected_observations"),
                 dQuote("ALL"),
                 applicability_problem = TRUE
    )
    expected_observations <- "ALL"
  }

  # correct variable usage
  util_correct_variable_use("strata_vars",
    allow_null = TRUE,
    need_type = "!float"
  )

  util_correct_variable_use("group_vars",
    allow_null = TRUE,
    need_type = "!float"
  )


  # exclude_roles = c("process", "not")

  # should some variables not be considered?
  if (VARIABLE_ROLE %in% names(meta_data)) {

    # a: not all roles specified found in metadata
    if (!(all(exclude_roles %in% meta_data[[VARIABLE_ROLE]]))) {
      if (any(exclude_roles %in% meta_data[[VARIABLE_ROLE]])) {
        util_warning(paste0(
          "Specified VARIABLE_ROLE(s): '",
          exclude_roles[!(exclude_roles %in% meta_data[[VARIABLE_ROLE]])],
          "' was not found in metadata, only: '",
          exclude_roles[exclude_roles %in% meta_data[[VARIABLE_ROLE]]],
          "' is used."
        ), applicability_problem = TRUE)

        exclude_roles <- exclude_roles[exclude_roles %in%
                                         meta_data[[VARIABLE_ROLE]]]

        which_vars_not <-
          meta_data[[label_col]][meta_data[[VARIABLE_ROLE]] %in%
                                   c(exclude_roles, VARIABLE_ROLES$SUPPRESS)]
        if (missing(label_col)) {
          which_vars_not <-
            meta_data[[VAR_NAMES]][meta_data[[VARIABLE_ROLE]] %in%
                                     c(exclude_roles, VARIABLE_ROLES$SUPPRESS)]
        }
        which_vars_not <- setdiff(which_vars_not, strata_vars)
        which_vars_not <- setdiff(which_vars_not, group_vars)
        if (length(intersect(names(ds1), which_vars_not)) > 0) {
          util_message(paste0(
            "Study variables: ",
            paste(dQuote(intersect(names(ds1), which_vars_not)),
                  collapse = ", "),
            " are not considered due to their VARIABLE_ROLE."
          ), applicability_problem = TRUE,
          intrinsic_applicability_problem = TRUE)
        }
        ds1 <- ds1[, !(names(ds1) %in% which_vars_not)]
        ds1_labelled <- ds1_labelled[, !(names(ds1_labelled) %in%
                                           which_vars_not)]
      } else {
        exclude_roles <- FALSE
        util_warning(
          c("Specified VARIABLE_ROLE(s) were not found in metadata.",
            "All variables are included here."),
          applicability_problem = TRUE)
      }


      # b: all roles are found in metadata
    } else {
      if (missing(exclude_roles)) {
        if (!.called_in_pipeline) util_message(
          c("Formal exclude_roles is used with default: all process variables",
            "are not included here."), applicability_problem = TRUE)
      }

      which_vars_not <- meta_data[[label_col]][meta_data[[VARIABLE_ROLE]] %in%
                                                 c(exclude_roles,
                                                   VARIABLE_ROLES$SUPPRESS)]
      if (missing(label_col)) {
        which_vars_not <- meta_data[[VAR_NAMES]][meta_data[[VARIABLE_ROLE]] %in%
                                                   c(exclude_roles,
                                                     VARIABLE_ROLES$SUPPRESS)]
      }
      which_vars_not <- setdiff(which_vars_not, strata_vars)
      which_vars_not <- setdiff(which_vars_not, group_vars)
      if (length(intersect(names(ds1), which_vars_not)) > 0) {
        util_message(paste0(
          "Study variables: ", paste(dQuote(intersect(names(ds1),
                                                      which_vars_not)),
                                     collapse = ", "),
          " are not considered due to their VARIABLE_ROLE."
        ),
        applicability_problem = TRUE,
        intrinsic_applicability_problem = TRUE)
      }
      ds1 <- ds1[, !(names(ds1) %in% which_vars_not)]
      ds1_labelled <- ds1_labelled[, !(names(ds1_labelled) %in%
                                         which_vars_not)]

    }
  } else {
    # since there are no roles defined exclusion is set to false
    exclude_roles <- FALSE
    util_message(
      c("VARIABLE_ROLE has not been defined in the metadata,",
        "therefore all variables within segments are used."),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE) # TODO: PART_VAR or STUDY_SEGMENT
  }

  # Which segments?
  if (!(STUDY_SEGMENT %in% names(meta_data))) {
    util_error("Metadata do not contain the column STUDY_SEGMENT",
               applicability_problem = TRUE)
  }

  meta_data[[STUDY_SEGMENT]][is.na(meta_data[[STUDY_SEGMENT]])] <- ""

  if (VARIABLE_ROLE %in% colnames(meta_data)) {
    seg_names <- meta_data[
      util_empty(meta_data[[VARIABLE_ROLE]]) |
        meta_data[[VARIABLE_ROLE]] !=
        VARIABLE_ROLES$SUPPRESS, # omit segments added on-the-fly as dependencies
      STUDY_SEGMENT,
      drop = TRUE] # meta_data[[LONG_LABEL]][meta_data$VAR_NAMES %in% part_vars]
  } else {
    seg_names <- meta_data[
      , STUDY_SEGMENT,
      drop = TRUE] # meta_data[[LONG_LABEL]][meta_data$VAR_NAMES %in% part_vars]
  }

  if (!(PART_VAR %in% names(meta_data))) {
    util_warning("Metadata do not contain the column PART_VAR",
                 applicability_problem = TRUE)
    pv <- seg_names
    pv[!startsWith(pv, "PART_")] <-
      paste0("PART_", pv[!startsWith(pv, "PART_")])
    while (any(pv %in% c(meta_data[[VAR_NAMES]],
                         meta_data[[LABEL]],
                         meta_data[[label_col]])
    )) {
      pv[pv %in% c(meta_data[[VAR_NAMES]],
                   meta_data[[LABEL]],
                   meta_data[[label_col]])] <-
        paste0("_", pv[pv %in% c(meta_data[[VAR_NAMES]],
                            meta_data[[LABEL]],
                            meta_data[[label_col]])], "_")
    }
    meta_data[[PART_VAR]] <- pv
  }

  part_vars <- meta_data[[PART_VAR]]
  names(part_vars) <- seg_names

  keep <- !is.na(part_vars) & !is.na(seg_names)

  part_vars <- part_vars[keep]

  part_vars <- part_vars[!duplicated(part_vars)]

  sn <- seg_names[!is.na(seg_names)]
  sn <- setNames(nm = unique(sn))
  .sn <- sn

  # do not compute segment plots, if not explicitly requested, if only one
  # segment in data
  keep_in_segmiss <- character(0)
  if (all(c(STUDY_SEGMENT, SEGMENT_MISS) %in% colnames(meta_data_segment))) {
    nas <- util_empty(meta_data_segment[[SEGMENT_MISS]])
    not_in_output <-
      util_is_na_0_empty_or_false(meta_data_segment[[SEGMENT_MISS]]) &
      !nas
    must_in_output <-
      !util_is_na_0_empty_or_false(meta_data_segment[[SEGMENT_MISS]])
    auto <- meta_data_segment[nas, STUDY_SEGMENT, drop = TRUE]
    remove <- meta_data_segment[not_in_output, STUDY_SEGMENT, drop = TRUE]
    keep_in_segmiss <- meta_data_segment[must_in_output,
                                         STUDY_SEGMENT, drop = TRUE]
    sn <- union(keep_in_segmiss, setdiff(sn, remove))
  }

  sn <- intersect(.sn, sn)

  sn <- sn[!is.na(sn)]
  sn <- setNames(nm = unique(sn))

  if (length(sn) < 2 && length(keep_in_segmiss) == 0) {
    util_error("No segment missingness plot for fewer than two segments",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  # determine which vars per segment
  var_sets <- lapply(sn, util_get_vars_in_segment,
               meta_data = meta_data,
               label_col = label_col)

  # remove all part_vars
  if (TRUE) {
    remove_part_vars <- function(vars, meta_data, label_col) {
      v <- util_map_labels(vars, meta_data, to = VAR_NAMES, from = label_col)
      vars[!(v %in% part_vars)]
    }
    var_sets <- lapply( var_sets, remove_part_vars,
                        meta_data = meta_data,
                        label_col = label_col)
  }

  # remove variables excluded by roles
  if (!isFALSE(exclude_roles)) {
    remove_roles <- function(vars, meta_data, label_col) {
      roles <- util_map_labels(vars,
                               meta_data = meta_data,
                               to = VARIABLE_ROLE,
                               from = label_col)
      vars[!(roles %in% exclude_roles)]
    }
    var_sets <- lapply( var_sets, remove_roles,
                        meta_data = meta_data,
                        label_col = label_col)
  }

  var_sets <- var_sets[vapply(var_sets, length, FUN.VALUE = integer(1)) > 0]

  # Which groups?
  if (length(group_vars) > 0) {
    is_grouped <- TRUE

    ds1[, setdiff(group_vars, part_vars)] <-
      ds1_labelled[, setdiff(group_vars, part_vars), FALSE]
    gr <- unique(ds1[[group_vars]][!is.na(ds1[[group_vars]])])
    gr <- gr[order(gr)]
    # covariables for plot
    cvs <- c(group_vars, "Examinations")

    # missings in grouping variable?
    ds1_labelled <- ds1_labelled[!is.na(ds1[[group_vars]]), ]
    ds1 <- ds1[!is.na(ds1[[group_vars]]), ]
  } else {
    is_grouped <- FALSE
    gr <- 1
    group_vars <- "Group"
    ds1$Group <- 1
    cvs <- "Examinations"
  }

  if (length(strata_vars) > 0) {
    # No. of strata levels and labels
    if (dim(ds1)[1] != dim(ds1[!is.na(ds1[[strata_vars]]), ])[1]) {
      ds1_labelled <- ds1_labelled[!is.na(ds1[[strata_vars]]), ]
      ds1 <- ds1[!is.na(ds1[[strata_vars]]), ]
      util_message(paste0("Some observations in ", strata_vars,
                          " are NA and were removed."),
                   applicability_problem = FALSE)
    }

    ds1[, setdiff(strata_vars, part_vars)] <-
      ds1_labelled[, setdiff(strata_vars, part_vars), FALSE]

    strata <- unique(ds1[[strata_vars]])[!is.na(unique(ds1[[strata_vars]]))]
    strata <- strata[order(strata)]
    # covariables for plot
    cvs <- c(strata_vars, group_vars, "Examinations")
  }

  # create result dataframe by factor combinations
  if (length(strata_vars) == 0) {
    res_df <- expand.grid(Group = gr, Examinations = names(var_sets))
    colnames(res_df) <- c(group_vars, "Examinations")
  } else {
    res_df <- expand.grid(
      Strata = strata, Group = gr, Examinations = names(var_sets),
      stringsAsFactors = TRUE
    )
    colnames(res_df) <- c(strata_vars, group_vars, "Examinations")
    res_df <- res_df[order(res_df[[strata_vars]], res_df[[group_vars]]), ]
  }

  ################
  # CALCULATIONS #
  ################

  myfun <- function(x) {
    all(is.na(x))
  }
  Ns <- c()
  Ms <- c()

  if (length(strata_vars) == 0) {
    for (j in seq_along(var_sets)) {
      for (i in seq_along(gr)) {
        one_var_from_seg_w_r_o_g <- var_sets[[j]][[1]] # empty var set is impossible, since segments are read from item-level-metadata here. so, the first one could represent for expected observations the whole segment.
        check_df <- ds1[
          util_observation_expected(rv = one_var_from_seg_w_r_o_g,
                                    study_data = ds1,
                                    meta_data = meta_data,
                                    label_col = label_col,
                                    expected_observations =
                                      expected_observations) &
            ds1[[group_vars]] == gr[i],
          c(as.character(unlist(var_sets[j]))),
          drop = FALSE]
        Ns <- c(Ns, nrow(check_df))
        Ms <- c(Ms, sum(apply(check_df, 1, myfun)))
      }
    }
  } else {
    for (i in seq_along(strata)) {
      for (j in seq_along(gr)) {
        for (k in seq_along(var_sets)) {
          one_var_from_seg_w_r_o_g <- var_sets[[k]][[1]] # empty var set is impossible, since segments are read from item-level-metadata here. so, the first one could represent for expected observations the whole segment.
          check_df <-
            ds1[util_observation_expected(rv = one_var_from_seg_w_r_o_g,
                                      study_data = ds1,
                                      meta_data = meta_data,
                                      label_col = label_col,
                                      expected_observations =
                                        expected_observations) &
            ds1[[strata_vars]] == strata[i] & ds1[[group_vars]] == gr[j],
          c(as.character(unlist(var_sets[k])))
          ]
          Ns <- c(Ns, nrow(check_df))
          Ms <- c(Ms, sum(apply(check_df, 1, myfun)))
        }
      }
    }
  }

  res_df$"No. Participants" <- Ns # TODO: Write an indicator for segments expected to be empty according to some intro variable
  res_df$"No. missing segments" <- Ms
  res_df$"(%) of missing segments" <- round(res_df$`No. missing segments` /
                                              res_df$`No. Participants` *
                                              100, digits = 2)
  res_df$"(%) of missing segments" <-
    as.numeric(res_df$"(%) of missing segments")
  res_df$threshold <- threshold_value
  res_df$direction <- color_gradient_direction

  res_df$"(%) of missing segments"[
    !is.finite(res_df$"(%) of missing segments")] <- NA_real_ # to avoid Inf %

  if (color_gradient_direction == "above") {
    res_df$GRADING <- ifelse(res_df$"(%) of missing segments" >
                               threshold_value, 1, 0) # FIXME: Update GRADING -> Standard Name
  } else {
    res_df$GRADING <- ifelse(res_df$"(%) of missing segments" <
                               threshold_value, 1, 0) # FIXME: Update GRADING -> Standard Name
  }

  # plot
  inversion <- ifelse(color_gradient_direction == "below", 1, 0)

  # order result data frame by grouping variable
  if (length(strata_vars) == 0) {
    # order result data frame by grouping variable
    res_df <- res_df[order(res_df[[group_vars]]), ]
    if (!is_grouped) {
      repsumtab <- res_df[, c("Examinations",
                              "No. missing segments",
                              "No. Participants")]
      colnames(repsumtab)[c(1,3)] <- c("Variables", "N")
      attr(repsumtab, "higher_means") <- ifelse(
        color_gradient_direction == "above",
        "worse",
        "")
      repsumtab <- util_validate_report_summary_table(repsumtab,
                                                  meta_data = meta_data,
                                                  label_col = label_col)
      p <- print(util_attach_attr(repsumtab, relative = FALSE),
            view = FALSE)

    } else { # TODO: Implemen t something useful in print.ReportSummaryTable
      p <- util_heatmap_1th(
       df = res_df, cat_vars = cvs, values = "(%) of missing segments",
       right_intv = TRUE, threshold = threshold_value,
       invert = inversion
      )$SummaryPlot
      repsumtab <- NULL
    }
  } else { # TODO: Implemen t something useful in print.ReportSummaryTable
    # order result data frame by grouping variable
    res_df <- res_df[order(res_df[[strata_vars]], res_df[[group_vars]]), ]
    # repsumtab <- res_df[, c("Examinations",
    #                         "No. of missing segments",
    #                         "No. of Participants")]
    # colnames(repsumtab)[c(1,3)] <- c("Variables", "N")
    # attr(repsumtab, "higher_means") <- ifelse(
    #   color_gradient_direction == "above",
    #   "worse",
    #   "")
    # class(repsumtab) <- union("ReportSummaryTable", class(repsumtab))
    # p <- print(repsumtab, view = FALSE))
    p <- util_heatmap_1th( # TODO: Implemen t something useful in print.ReportSummaryTable
      df = res_df, cat_vars = cvs[-1], values = "(%) of missing segments",
      right_intv = TRUE, threshold = threshold_value,
      invert = inversion, strata = strata_vars
    )$SummaryPlot
    repsumtab <- NULL
  }
#
#   suppressWarnings({
#     # suppress wrong warnings: https://github.com/tidyverse/ggplot2/pull/4439/commits
#     # find out size of the plot https://stackoverflow.com/a/51795017
#     bp <- ggplot_build(p)
#     w <- 2 * length(bp$layout$panel_params[[1]]$x$get_labels())
#     if (w == 0) {
#       w <- 10
#     }
#     w <- w + 2 +
#       max(nchar(bp$layout$panel_params[[1]]$y$get_labels()),
#           na.rm = TRUE)
#     h <- 2 * length(bp$layout$panel_params[[1]]$y$get_labels())
#     if (h == 0) {
#       h <- 10
#     }
#     h <- h + 15
#
#     p <- util_set_size(p, width_em = w, height_em = h)
#   })

  text_to_display <- util_get_hovertext("[com_segment_missingness_hover]")
  attr(res_df, "description") <- text_to_display

  rst_atts <- util_init_respum_tab(repsumtab)

  if (rst_atts$relative) {
    v <- repsumtab$`No. of missing segments` / repsumtab$N  * 100
  } else {
    v <- repsumtab$`No. of missing segments`
  }
#  min_bar_height <- suppressWarnings(min(v, na.rm = TRUE))
#  max_bar_height <- suppressWarnings(max(v, na.rm = TRUE))



#  if(min_bar_height == Inf ||
 #    min_bar_height == -Inf ||
#     max_bar_height == Inf ||
#     max_bar_height == -Inf ) {
#    range <- 400 #an intermediate size plot in height
#  } else {
#    range <- max_bar_height - min_bar_height
#  }


  res_df <- res_df[, !names(res_df) %in% c("GRADING")]


  #Add new attribute to the columns of ResultData to define the datatype of each column
  attr(res_df[[paste(colnames(res_df)[1])]], DATA_TYPE) <- DATA_TYPES$STRING #The column name depends on the presence of the argument group_vars, if not present is Group otherwise is the name of the variable in group_vars
  attr(res_df$Examinations, DATA_TYPE) <- DATA_TYPES$STRING
  attr(res_df$`No. Participants`, DATA_TYPE) <- DATA_TYPES$INTEGER
  attr(res_df$`No. missing segments`, DATA_TYPE) <- DATA_TYPES$INTEGER
  attr(res_df$`(%) of missing segments`, DATA_TYPE) <- DATA_TYPES$FLOAT
  attr(res_df$threshold, DATA_TYPE) <- DATA_TYPES$INTEGER
  attr(res_df$direction, DATA_TYPE) <- DATA_TYPES$STRING


  return(util_attach_attr(list(ResultData = res_df,
              ReportSummaryTable = repsumtab,
              SummaryPlot = p)#,
             # sizing_hints = list(
            #    figure_type_id = "bar_chart",
            #    rotated = FALSE,
            #    number_of_bars = length(unique(repsumtab$Variables)),
            #    range = range
            #  )
  ))
}

expect_conditions <- function(expr, regexps, classes) {
  cl <- substitute(expr)
  .res <- regexps
  .res[] <- ""
  classes <- paste0(.res, classes) # recycle
  for (ire in seq_along(regexps)) {
    re <- regexps[[ire]]
    cls <- classes[[ire]]
    cl <- do.call(rlang::call2,
                       list(.fn = paste0("expect_", cls),
                            object = quote(cl),
                            regexp = re))
  }
  eval(cl, envir = parent.frame())
}

suppressWarningsMatching <- function(expr, regexps) {
  withCallingHandlers(expr,
                      warning = function(cnd) {
                        msg <- conditionMessage(cnd)
                        if (any(vapply(regexps, function(x)
                          all(grepl(x, paste0(msg, collapse = "\n"))),
                          FUN.VALUE = logical(1)))) {
                          invokeRestart("muffleWarning")
                        }
                      })
}

#' @seealso `util_study_var2factor()`
#' @family missing_functions
#' @concept metadata_management
#' @noRd
util_get_combined_code_lists <- function(x,
                                         code_name,
                                         split_char = SPLIT_CHAR,
                                         mdf,
                                         label_col = VAR_NAMES,
                                         warning_if_no_list = TRUE,
                                         assume_consistent_codes = TRUE,
                                         have_cause_label_df = FALSE) {
  l <- lapply(code_name, function(cn) {
    lb <- sub("_.*?$", "", cn, perl = TRUE)
    .l <-
      util_get_code_list(x, cn, split_char, mdf,
                         label_col,
                         warning_if_no_list)
    if (!have_cause_label_df ||
        (!(lb %in% c("JUMP", "MISSING")))) {
      if (any(names(.l) == as.character(.l))) {
        if (assume_consistent_codes) {
          names(.l)[names(.l) == as.character(.l)] <-
            paste(lb, .l[names(.l) == as.character(.l)])
        } else {
          names(.l)[names(.l) == as.character(.l)] <-
            paste(lb, x, .l[names(.l) == as.character(.l)])
        }
      }
    }
    setNames(as.character(.l), nm = names(.l))
  })
  unlist(l, recursive = FALSE)
}

#' Utility function to characterize study variables
#'
#' This function summarizes some properties of measurement variables.
#'
#' @param study_data study data, pre-processed with `prep_prepare_dataframes`
#'                   to replace missing value codes by `NA`
#' @param val_lab    deprecated
#'
#' @return data frame with one row for each variable in the study data and the
#'         following columns:
#' `Variables` contains the names of the variables
#' `IsInteger` contains a check whether the variable contains integer values
#'             only (variables coded as factor will be converted to integers)
#' `IsMultCat` contains a check for variables with integer or string values
#'             whether there are more than two categories
#' `NCategory` contains the number of distinct values for variables with
#'             values coded as integers or strings (excluding `NA` and
#'             empty entries)
#' `AnyNegative` contains a check whether the variable contains any negative
#'             values
#' `NDistinct` contains the number of distinct values
#' `PropZeroes` reports the proportion of zeroes
#'
#'
#' @family metadata_management
#' @concept process
#' @noRd
util_dist_selection <- function(study_data, val_lab = lifecycle::deprecated()) {
  if (lifecycle::is_present(val_lab)) {
    # Signal the deprecation to the user
    lifecycle::deprecate_warn(
      "2.5.0",
      "dataquieR::util_dist_selection(val_lab = )")
  }
  # TODO EK: Is this function maybe deprecated?
  # TODO: discuss function name (does not select a distance), could be named 'util_vars_properties' instead
  .x <- as.data.frame(study_data)
  .r <- data.frame(
    Variables   = colnames(.x),
    IsInteger   = c(rep(NA, length(colnames(.x)))),
    IsMultCat   = c(rep(NA, length(colnames(.x)))),
    NCategory   = c(rep(NA, length(colnames(.x)))),
    AnyNegative = c(rep(NA, length(colnames(.x)))),
    NDistinct   = c(rep(NA, length(colnames(.x)))),
    PropZeroes  = c(rep(NA, length(colnames(.x))))#,
    # HasValueLabels = c(rep(NA, length(colnames(.x))))
  )

  # convert variables coded as factors to integers
  .x[, vapply(FUN.VALUE = logical(1), .x, is.factor)] <-
    vapply(FUN.VALUE = integer(nrow(.x)),
           .x[, vapply(FUN.VALUE = logical(1), .x, is.factor), drop = FALSE],
           as.integer)

  # identify integer values (includes both positive and negative values)
  .r$IsInteger <- vapply(FUN.VALUE = logical(1), .x,
                         function(.y)
                           all(util_is_integer(.y), na.rm = TRUE))

  # identify negative values
  .r$AnyNegative <- vapply(FUN.VALUE = logical(1), .x,
                           function(.y)
                             any(.y < 0, na.rm = TRUE))

  # count the number of distinct values, excluding `NA`s and empty fields
  .r$NDistinct <- vapply(FUN.VALUE = numeric(1), .x,
                         function(.y)
                           length(unique(.y[which(!util_empty(.y))])))

  # calculate the proportion of zeroes
  .r$PropZeroes <- vapply(FUN.VALUE = numeric(1), .x,
                          function(.y)
                            length(which(.y == 0)) /
                            length(.y[which(!util_empty(.y))]))

  is_char <- vapply(FUN.VALUE = logical(1), .x,
                    function(.y)
                      is.character(.y))
  # number of categories (can be coded as integer or as string)
  # ind_get_cat <- which(.r$IsInteger | is_char | .r$HasValueLabels)
  ind_get_cat <- which(.r$IsInteger | is_char)
  .r$IsMultCat[ind_get_cat] <- FALSE
  if (length(ind_get_cat) > 0) {
    .r$NCategory[ind_get_cat] <-
      vapply(FUN.VALUE = numeric(1), .x[, ind_get_cat, drop = FALSE],
             function(.y)
               length(unique(.y[which(!util_empty(.y))])))
    # more than two categories (can be coded as integer or as string)
    .r$IsMultCat[ind_get_cat] <- ifelse(.r$NCategory[ind_get_cat] > 2,
                                        TRUE, FALSE)
  }

  ### Output
  return(Results = .r)
}

test_that("util_dist_selection works", {
  skip_on_cran()
  study_data <- data.frame(
    integer = c(1, 2, 3, 4),
    factor = factor(c(1, 2, 2, 3),
      levels = 1:3,
      labels = c("male", "female", "undefined")
    ),
    ordered = ordered(c(3, 2, 1, 3),
      levels = 3:1,
      labels = c("low", "medium", "high")
    ),
    integer = as.integer(c(1, 2, 3, 4)),
    logical = c(TRUE, FALSE, TRUE, FALSE),
    float = 1:4 * pi
  )
  dist_info <- util_dist_selection(study_data)
  expect_equal(
    dist_info$IsInteger,
    c(rep(TRUE, 4), FALSE, FALSE)
  )
  expect_equal(
    dist_info$IsMultCat,
    c(rep(TRUE, 4), NA, NA)
  )
  expect_equal(
    dist_info$NCategory,
    c(4, 3, 3, 4, NA, NA)
  )
})

#' Make `it` scalable, if it is a figure (parallel-safe version)
#'
#' This version avoids writing to disk in parallel execution.
#' Instead, it attaches metadata for later serialization.
#'
#' @param it [htmltools::tag()] compatible object
#' @param dir [character] output directory for potential `iframe`.
#' @param nm [character] name for the `iframe`d file, if one is created
#' @param fkt [character] function name of the indicator function that created `it`.
#' @param sizing_hints `object` additional metadata about the natural figure size
#' @param ggthumb [ggplot2::ggplot()] optional preview plot
#'
#' @return [htmltools::tagList()] tagged object with metadata for later rendering
#' @noRd
util_iframe_it_if_needed <- function(it, dir, nm, fkt, sizing_hints, ggthumb) {
  functionName <- fkt

  if (!isTRUE(getOption('knitr.in.progress')) &&
      (inherits(it, "plotly") ||
       (inherits(it, "shiny.tag") && it$name == "img"))) {

    util_expect_scalar(dir, check_type = is.character)
    util_expect_scalar(nm, check_type = is.character)
    util_stop_if_not(dir.exists(dir))

    if (!nzchar(paste0("", nm))) {
      fig_framed_file <- tempfile(pattern = "FIG_", fileext = ".html", tmpdir = dir)
      fig_framed_file <- basename(fig_framed_file)
    } else {
      fig_framed_file <- paste0("FIG_", prep_link_escape(nm, html = TRUE), ".html")
    }

    full_fig_file <- file.path(dir, fig_framed_file)
    fig_framed_thumb_file <- gsub("\\.html$", ".png", fig_framed_file)
    thumb_file <- file.path(dir, fig_framed_thumb_file)

    sizing_hints <- util_finalize_sizing_hints(sizing_hints = sizing_hints)

    # original proportions
    aspect_ratio <- sizing_hints$w_in_cm / sizing_hints$h_in_cm
    if (!is.finite(aspect_ratio)) aspect_ratio <- 1
    min.width <- 320
    min.height <- 250
    min.width.from.height <- min.height * aspect_ratio
    min.height.from.width <- min.width / aspect_ratio
    if (min.width.from.height > min.width) {
      min.width <- min.width.from.height
    } else {
      min.height <- min.height.from.width
    }
    min.width <- paste0(min.width, "px");
    min.height <- paste0(min.height, "px");


    if (util_is_gg(ggthumb)) {
      optimized_args <- util_get_restricted_size_args_for_figure(
        MAX_SIZE = 5 * 1024 * 1024,
        max_w_in_cm = 53,
        max_h_in_cm = 30,
        sizing_hints = sizing_hints
      )

      sizing_hints$w_in_cm <- optimized_args$sizing_hints_updated$w_in_cm
      sizing_hints$h_in_cm <- optimized_args$sizing_hints_updated$h_in_cm
    } else {
      optimized_args <- NULL
    }

    if (suppressWarnings(util_ensure_suggested("jsonlite", goal = "generate figures in reports", err = FALSE))) {
      sizing_hint_script <- paste0(
        'window.sizing_hints = ',
        jsonlite::toJSON(sizing_hints, auto_unbox = FALSE)
      )
    } else {
      sizing_hint_script <- 'window.sizing_hints = "No figure size hints, need R package jsonlite for this."'
    }

    util_ensure_suggested("rmarkdown",
                          goal = "Creating the Print Dialog for Plotly")

    jqui <- rmarkdown::html_dependency_jqueryui()
    jqui$stylesheet <- "jquery-ui.min.css"

    html_inner <- htmltools::tagList(
      rmarkdown::html_dependency_jquery(),
      jqui,
      htmltools::tags$script(type = "text/javascript", sizing_hint_script),
      htmltools::tags$span(`data-nm` = nm, id = "nm"),
      htmltools::tags$span(`data-functionName` = functionName, id = "functionName"),
      html_dependency_dataquieR(iframe = TRUE),
      html_dependency_jspdf(),
      it
    )

    frameTag <- htmltools::tags$iframe(
      src = fig_framed_file,
      style = htmltools::css(border = "0", width = "100%", height = "calc(100% - 5px)")
    )

    if (util_is_gg(ggthumb)) {
      my_style <- "margin:0px;width:100%;height:100%;object-fit:scale-down;object-position: left top;cursor:zoom-in;display:block;"
      imgTag <- htmltools::tagList(htmltools::img(
        src = basename(thumb_file),
        style = my_style,
        onmouseup = "tglePyHandler()",
        `data-iframe` = frameTag
      ))
    } else {
      imgTag <- frameTag
    }

    final_div <- htmltools::div(
      class = "scaler",
      `data-initialw` = sizing_hints$w,
      `data-initialh` = sizing_hints$h,
      style = htmltools::css(
        width = sizing_hints$w,
        height = sizing_hints$h,
        min.width = min.width,
        min.height = min.height,
        resize = "both",
        overflow = "auto",
        border = "1px"
      ),
      imgTag
    )

    attr(final_div, "html_file") <- full_fig_file
    attr(final_div, "html_inner") <- html_inner
    # attr(final_div, "dependencies") <- htmltools::htmlDependencies(html_inner))
    if (util_is_gg(ggthumb)) {
      attr(final_div, "thumbnail_path") <- thumb_file
      attr(final_div, "ggthumb") <- util_compress(ggthumb)
      attr(final_div, "thumbnail_args") <- list(
        width = optimized_args$width,
        height = optimized_args$height,
        dpi = optimized_args$dpi,
        figure_type_id = sizing_hints$figure_type_id,
        rotated = sizing_hints$rotated
      )
    }

    return(final_div)
  }

  return(htmltools::div(it))
}

util_get_restricted_size_args_for_figure <- function(MAX_SIZE, max_w_in_cm, max_h_in_cm, sizing_hints) {
  # 32 bit color-space (4 bytes per pixel)
  BYTES_PER_PIXEL <- 4

  # original proportions
  original_w_cm <- sizing_hints$w_in_cm
  original_h_cm <- sizing_hints$h_in_cm
  aspect_ratio <- original_w_cm / original_h_cm

  # resize if too big keeping aspect ratio
  if (original_w_cm > max_w_in_cm || original_h_cm > max_h_in_cm) {
    scale_factor_w <- max_w_in_cm / original_w_cm
    scale_factor_h <- max_h_in_cm / original_h_cm

    # preserve aspect ratio
    scale_factor <- min(scale_factor_w, scale_factor_h)

    sizing_hints_updated_w_cm <- original_w_cm * scale_factor
    sizing_hints_updated_h_cm <- original_h_cm * scale_factor
  } else {
    sizing_hints_updated_w_cm <- original_w_cm
    sizing_hints_updated_h_cm <- original_h_cm
  }

  # convert to inches for ggsave()
  width_in_inches <- sizing_hints_updated_w_cm / 2.54
  height_in_inches <- sizing_hints_updated_h_cm / 2.54

  # compute number of pixels given the color space and the maximum file size
  max_pixels <- floor(MAX_SIZE / BYTES_PER_PIXEL)

  # compute max. dpi based on physical dimensions and max_pixels
  # we know: max_pixels = (width_in_inches * dpi) * (height_in_inches * dpi)
  # max_pixels = width_in_inches * height_in_inches * dpi^2
  # dpi^2 = max_pixels / (width_in_inches * height_in_inches)
  # dpi = sqrt(max_pixels / (width_in_inches * height_in_inches))

  # reboustness
  if (width_in_inches <= 0 || height_in_inches <= 0) {
    util_warning(c("Internal error, sorry. Please report: Invalid physical",
                   "dimensions (width or height is zero or negative).",
                   "Cannot calculate DPI."))
    return(list(sizing_hints_updated =
                  list(w_in_cm = sizing_hints_updated_w_cm,
                       h_in_cm = sizing_hints_updated_h_cm),
                dpi = 72,
                width = width_in_inches,
                height = height_in_inches))
  }

  max_dpi_from_size <- sqrt(max_pixels / (width_in_inches * height_in_inches))

  # rounding (should not be larger than max_dpi_from_size, so floor())
  # but limit resultion to at least 72 dpi for typical screens
  dpi <- max(72, floor(max_dpi_from_size))

  # Aktualisiere die sizing_hints_updated für die Rückgabe
  sizing_hints_updated <- list(w_in_cm = sizing_hints_updated_w_cm,
                               h_in_cm = sizing_hints_updated_h_cm)

  return(list(sizing_hints_updated = sizing_hints_updated,
              dpi = dpi,
              width = width_in_inches,
              height = height_in_inches))
}

#' Get the size of the currently set cluster
#'
#' or 1, if no cluster has been set
#'
#' @returns [integer] size of the default cluster
#' @noRd
util_get_cores_safe <- function() {
  cl <- parallel::getDefaultCluster()
  if (is.null(cl)) {
    return(1)
  } else {
    return(length(cl))
  }
}

#' Generate an execution/calling plan for computing a report from the metadata
#'
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param dimensions [dimensions] Vector of dimensions to address in the report.
#'                   Allowed values in the vector are Completeness, Consistency,
#'                   and Accuracy. The generated report will only cover the
#'                   listed data quality dimensions. Accuracy is computational
#'                   expensive, so this dimension is not enabled by default.
#'                   Completeness should be included, if Consistency is
#'                   included, and Consistency should be included, if Accuracy
#'                   is included to avoid misleading detections of e.g. missing
#'                   codes as outliers, please refer to the data quality concept
#'                   for more details. Integrity is always included.
#' @param meta_data_segment [data.frame] -- optional: Segment level metadata
#' @param meta_data_dataframe [data.frame] -- optional: Data frame level
#'                                                                 metadata
#' @param meta_data_cross_item [data.frame] -- optional: Cross-item level
#'                                                                 metadata
#' @param specific_args [list] named list of arguments specifically for one of
#'                             the called functions, the of the list elements
#'                             correspond to the indicator functions whose calls
#'                             should be modified. The elements are lists of
#'                             arguments.
#' @param arg_overrides [list] arguments to be passed to all called indicator
#'                             functions if applicable.
#' @param resp_vars variables to be respected, `NULL` means to use all.
#' @param filter_indicator_functions [character] regular expressions, only
#'                                               if an indicator function's name
#'                                               matches one of these, it'll
#'                                               be used for the report. If
#'                                               of length zero, no filtering
#'                                               is performed.
#' @param exclude_indicator_functions [character] regular expressions,
#'                                               if an indicator function's name
#'                                               matches one of these, it'll
#'                                               be excluded from the report. If
#'                                               of length zero, no filtering
#'                                               is performed.
#'
#' @return a list of calls
#'
#' @family reporting_functions
#' @concept process
#' @noRd
util_generate_calls <- function(dimensions,
                                meta_data,
                                label_col,
                                meta_data_segment,
                                meta_data_dataframe,
                                meta_data_cross_item,
                                specific_args,
                                arg_overrides,
                                resp_vars,
                                filter_indicator_functions,
                                exclude_indicator_functions) {
  .dimensions <- unique(c("Descriptors", "Integrity", dimensions))
  .dimensions <- substr(tolower(.dimensions), 1, 3)
  .dimensions[.dimensions == "int"] <- "int_all" # use the wrappers, only
  .dimensions <- paste(paste0(.dimensions, "_"), collapse = "|")
  .dimensions <- sprintf("^(%s)", .dimensions)
  exports <- getNamespaceExports(utils::packageName()) # TODO: Use util_all_ind_functions
  ind_functions <- grep(.dimensions, exports, value = TRUE)
  ind_functions <- c(ind_functions, "int_encoding_errors")
  ind_functions <- ind_functions[
    vapply(ind_functions,
           exists,
           envir = getNamespace(utils::packageName()),
           mode = "function", FUN.VALUE = logical(1))]
  ind_functions <- c(setdiff(ind_functions, c(
    "con_contradictions", # we use the new cross-item-level now
    "acc_robust_univariate_outlier", # is just a synonym for acc_univariate_outlier
    "acc_distributions", # there are specific functions for location and proportion checks
    "com_unit_missingness", # function does not really create something reasonable, currently
    "des_summary" # function just meant to be called directly by the user, more specific functions are available in the pipeline
  )), "int_datatype_matrix") # the data type matrix function does not start with int_all, but integrity should always run. However, it is still run before the pipeline, see util_evalute_calls and int_data_type_matrix

  ind_functions <- setNames(nm = ind_functions)

  ind_functions <- util_filter_names_by_regexps(ind_functions,
                                                filter_indicator_functions)

  ind_functions <- util_filter_names_by_regexps(ind_functions,
                                                exclude_indicator_functions,
                                                negate = TRUE)

  # handle SSI functions
  ssi_functions <- character(0)
  try({
    ssi_functions <- unlist(util_parse_assignments(util_get_concept_info("ssi",
                                                                          get("SSI_METRICS") %in% colnames(meta_data_cross_item),
                                                                          "functions",
                                                                          drop = TRUE)))
  }, silent = TRUE)
  ssi_functions <- setNames(nm = unique(gsub("\\..*$", "", ssi_functions)))
  # /SSI functions

  all_calls <- lapply(setNames(nm = union(ind_functions, ssi_functions)),
                      util_generate_calls_for_function,
                      meta_data = meta_data,
                      label_col = label_col,
                      meta_data_segment = meta_data_segment,
                      meta_data_dataframe = meta_data_dataframe,
                      meta_data_cross_item = meta_data_cross_item,
                      specific_args = specific_args,
                      arg_overrides = arg_overrides,
                      resp_vars = resp_vars,
                      ssi_functions = ssi_functions,
                      non_ssi_functions = ind_functions
  )

  # TODO: https://gitlab.com/libreumg/dataquier/-/issues/147, before

  explode_splinter <- function(entity_name, splinter, fkt) {
    to_explode <- vapply(lapply(splinter, attr, "explode"),
                         identical,
                         FUN.VALUE = logical(1),
                         y = TRUE)
    if (all(!to_explode)) {
      res <- do.call(what = call,
                     args = c(list(name = fkt), splinter),
                     quote = TRUE)
      attr(res, "entity_name") <- entity_name
      attr(res, VAR_NAMES) <- util_map_labels(
        x = entity_name,
        to = VAR_NAMES,
        from = label_col,
        ifnotfound = NA_character_,
        meta_data = meta_data
      )
      if (!(STUDY_SEGMENT %in% colnames(meta_data))) {
        meta_data[[STUDY_SEGMENT]] <- rep(NA_character_, nrow(meta_data))
      }
      attr(res, STUDY_SEGMENT) <- util_map_labels(
        x = entity_name,
        to = STUDY_SEGMENT,
        from = label_col,
        ifnotfound = NA_character_,
        meta_data = meta_data
      )
      if (!(GRADING_RULESET %in% colnames(meta_data))) {
        meta_data[[GRADING_RULESET]] <- rep(0, nrow(meta_data))
      }
      attr(res, GRADING_RULESET) <- util_map_labels(
        x = entity_name,
        to = GRADING_RULESET,
        from = label_col,
        ifnotfound = 0,
        meta_data = meta_data
      )
      attr(res, "label_col") <- label_col
      return(res)
    }
    my_splinter <- head(splinter[to_explode], 1)
    remainder <- tail(splinter[to_explode], -1)
    not_a_bomb <- splinter[!to_explode]
    lapply(my_splinter[[1]], function(my_splinter_part) {
      my_arg <- setNames(
        list(my_splinter_part),
        nm = names(my_splinter)
      )
      arg_list <- c(my_arg, remainder, not_a_bomb)
      explode_splinter(entity_name = entity_name,
                       splinter = arg_list,
                       fkt = fkt)
    })
  }

  explode <- function(fkt, bomb) {
    mapply(SIMPLIFY = FALSE,
           names(bomb),
           bomb,
           FUN = explode_splinter,
           MoreArgs = list(fkt = fkt))
  }

  all_calls <- unlist(mapply(SIMPLIFY = FALSE,
                             names(all_calls), all_calls, FUN = explode))


  has_resp_vars <- vapply(lapply(all_calls, names),
                          `%in%`, x = "resp_vars", FUN.VALUE = logical(1))
  suppress <- has_resp_vars &
    vapply(lapply(all_calls, `[[`, "resp_vars"), length,
                                             FUN.VALUE = integer(1)) == 0

  all_calls <- all_calls[!suppress]

  make_name <- function(nm, call) {
    entity_name <- attr(call, "entity_name")
    r <- gsub(".", "_", fixed = TRUE, sub( # the alias names must not contain dots ".", the first dot separates the call alias from the entity name (mostly VAR_NAMES)
      paste0("\\.\\Q",
             entity_name,
             "\\E"),
      "",
      nm,
      perl = TRUE))
    r <- sub("_TIME_VAR", "", r) # all with explode = TRUE
    r <- sub("_GROUP_VAR", "", r)
    r <- paste0(tolower(r), ".", entity_name)
    r
  }

  names(all_calls) <- unlist(mapply(SIMPLIFY = FALSE,
                                    names(all_calls), all_calls, FUN = make_name))

  util_stop_if_not(all(grepl(fixed = TRUE, ".", names(all_calls)))) # there must be at least one dot to separated the col. name from the row name (wolog this is the first dots)

  if (length(all_calls)) {
    rn <- util_sub_string_right_from_.(names(all_calls))
    cn <- util_sub_string_left_from_.(names(all_calls))
  } else {
    rn <- character(0)
    cn <- character(0)
    all_calls <- list()
  }

  attr(all_calls, "rn") <- rn
  attr(all_calls, "cn") <- cn


  multivariatecol <- names(all_calls)[vapply(lapply(all_calls,
                                                    `[[`, "resp_vars"), length,
                                             FUN.VALUE = integer(1)) != 1] # all calls that passed more ore fewer than one variable in resp_vars

  multivariatecol <- gsub("\\..*$", "", multivariatecol)

  attr(all_calls, "multivariatcol") <- multivariatecol # SQ2 compatibility, everything, that does not address item_level, i.e., that is not called once for each variable; CAVE: Strange name in SQ2 standard w/o "e"

  all_calls
}

#' Create a dashboard-table from a report summary
#'
#' @param repsum a report summary from `summary(report)`
#' @param folder_of_report a named vector with the location of variable and
#'                         `call_names`
#'
#' @family html
#' @concept process
#' @noRd
util_dashboard_table <- function(repsum, folder_of_report = NULL) {
  this <- attr(repsum, "this")
  bigTable <- this$result
  tb <- suppressWarnings(
    bigTable[!is.na(as.numeric(bigTable$values_raw)), , FALSE])
  label_col <- this$label_col
  meta_data <- this$meta_data

  rowmaxes <- this$rowmaxes
  var_classes <- setNames(util_as_integer_cat(
    rowmaxes$rowmax), nm = rowmaxes$VAR_NAMES)
  var_classes <- ordered(var_classes,
                         levels = names(util_get_labels_grading_class()),
                         labels = util_get_labels_grading_class())

  tb[[label_col]] <-
    prep_map_labels(tb$VAR_NAMES, meta_data = meta_data, to = label_col)

  indicator_metric <- NULL # to make R CMD check happy

  if (length(tb) > 0 && prod(dim(tb)) > 0) {
    tb <- subset(tb,
                 !startsWith(indicator_metric, "CAT_") &
                   !startsWith(indicator_metric, "MSG_")
    )
  }

  tb$var_class <- var_classes[tb$VAR_NAMES]

  cols <- util_get_colors()[tb$class]
  labs <- util_get_labels_grading_class()[tb$class]
  fg_cols <- util_get_fg_color(cols)

  tb$`Class` <- ifelse(is.na(labs), "", paste0( # TODO: Find out, if we can remove this safely, when we have factors in datatables always working
    labs
  ))

  tb$`Metric` <- util_translate_indicator_metrics(tb$indicator_metric)
  tb$`Call` <- vapply(tb$call, util_alias2caption, long = TRUE,
                      FUN.VALUE = character(1))
  tb$n_classes <- NULL

  meta_data$STUDY_SEGMENT <- NULL # delete, it is already in tb

  tb <- merge(tb, meta_data, by = intersect(intersect(colnames(tb),
                                                      colnames(meta_data)),
                                            c(VAR_NAMES, label_col)))

  tb$class <- ordered(tb$class,
                      levels = names(util_get_labels_grading_class()),
                      labels = util_get_labels_grading_class())

  my_order1 <- c(
    label_col,
    VAR_NAMES,
    STUDY_SEGMENT,
    "Call",
    "Metric",
    "value",
    "Class",
    "call_names",
    "values_raw",
    "function_name",
    "indicator_metric",
    "class",
    "var_class",
    LABEL,
    DATA_TYPE,
    SCALE_LEVEL
  )

  my_order2 <- c(
    STANDARDIZED_VOCABULARY_TABLE,
    MISSING_LIST_TABLE,
    HARD_LIMITS,
    DETECTION_LIMITS,
    SOFT_LIMITS,
    DISTRIBUTION,
    DECIMALS,
    GROUP_VAR_OBSERVER,
    GROUP_VAR_DEVICE,
    TIME_VAR,
    PART_VAR,
    VARIABLE_ROLE,
    VARIABLE_ORDER,
    LONG_LABEL,
    "ELEMENT_HOMOGENITY_CHECKTYPE",
    UNIVARIATE_OUTLIER_CHECKTYPE,
    N_RULES,
    LOCATION_METRIC,
    LOCATION_RANGE,
    PROPORTION_RANGE,
    "REPEATED_MEASURES_VARS",
    "REPEATED_MEASURES_GOLDSTANDARD",
    CO_VARS,
    END_DIGIT_CHECK,
    VALUE_LABEL_TABLE,
    MISSING_LIST,
    JUMP_LIST
  )

  my_order <- unique(c(intersect(my_order1, colnames(tb)),
                sort(setdiff(colnames(tb), union(my_order1, my_order2))),
                intersect(my_order2, colnames(tb))))

  tb <- tb[, my_order, FALSE]

  tb <- util_add_links_to_summary_table(tb,
                                        this,
                                        folder_of_report = folder_of_report)

  return(util_attach_attr(tb,
                          label_col = label_col))
}

util_add_links_to_summary_table <- function(tb, this, folder_of_report = NULL) {
  label_col <- this$label_col
  while (!is.null(attr(label_col, "orig_label_col"))) {
    label_col <- attr(label_col, "orig_label_col")
  }
  tb[[label_col]] <-
    prep_map_labels(tb$VAR_NAMES,
                    meta_data = this$meta_data,
                    to = label_col)
  tb[[this$label_col]] <-
    prep_map_labels(tb$VAR_NAMES,
                    meta_data = this$meta_data,
                    to = this$label_col)
  if (nrow(tb) == 0) {
    tb$href <- character(0)
    tb$popup_href <- character(0)
    tb$title <- character(0)
  } else {
    tb$href <-
      paste0("VAR_", prep_link_escape(tb[[label_col]],
                                      html = TRUE),
             ".html#",
             prep_link_escape(as.character(tb[[label_col]])),
             ".", tb[["call_names"]])

    # showDataquieRResult("VAR_AGE0.html#nm=acc_distributions_only.AGE_0", "VAR_AGE0.html#v00003: Age B/L", "Gemüsesuppe")
    tb$popup_href <-
      paste0("VAR_", prep_link_escape(tb[[label_col]],
                                      html = TRUE),
             ".html#nm=", tb$call_names, ".",
             as.character(tb[[label_col]]))

    tb$title <- paste0(tb[[this$label_col]],
                       ": ",
                       vapply(tb$call_names, util_alias2caption, long = TRUE, FUN.VALUE = character(1)),
                       ifelse(!is.na(tb$title_ind), paste0(" (",
                       tb$title_ind,
                       ")"), ""))
  }

  if (!is.null(folder_of_report)) { # TODO: Use utility function util_add_links_to_summary_table
    tb$href <-
      paste0(folder_of_report[gsub("\\s+", "", paste0(tb[[this$label_col]], ".", tb[["call_names"]]))],
             "/.report/",
             tb$href)
    tb$popup_href <-
      paste0(folder_of_report[gsub("\\s+", "", paste0(tb[[this$label_col]], ".", tb[["call_names"]]))],
             "/.report/",
             tb$popup_href)
  }

  util_attach_attr(tb, orig_label_col = tb[[this$label_col]])
}

util_get_concept_links <- function(fkt) {

  # generates pages with navigation menu
  # load concept to get current indicator links in reports

  concept <- util_get_concept_info("dqi")
  concept <- concept[c("PublicID", "Name", "function_R")]

  fkt2concept <- subset(
    concept,
    get("function_R") == fkt
  )

  # create link tags
  get_links <- function(indicator_id, indicator_name) {
    htmltools::tags$a(
      target = "_blank",
      href = paste(
        "https://dataquality.qihs.uni-greifswald.de/id/#",
        indicator_id,
        sep = ""
      ),
      indicator_name
    )
  }

  # create un-ordered item list for each indicator
  links <- mapply(
    get_links,
    indicator_id   = fkt2concept$PublicID,
    indicator_name = fkt2concept$Name,
    SIMPLIFY = FALSE
  )
  names(links) <- NULL

  items <- lapply(links, htmltools::tags$li)

  htmltools::tags$ul(items)
}

#' Compare observed versus expected distributions
#'
#' @description
#' This implementation contrasts the empirical distribution of a measurement
#' variables against assumed distributions. The approach is adapted from the
#' idea of rootograms (Tukey 1977) which is also applicable for count data
#' (Kleiber and Zeileis 2016).
#'
#' [Indicator]
#'
#' @details
#' # ALGORITHM OF THIS IMPLEMENTATION:
#'   - This implementation is restricted to data of type float or integer.
#'   - Missing codes are removed from resp_vars (if defined in the metadata)
#'   - The user must specify the column of the metadata containing probability
#'     distribution (currently only: normal, uniform, gamma)
#'  - Parameters of each distribution can be estimated from the data or are
#'    specified by the user
#'  - A histogram-like plot contrasts the empirical vs. the technical
#'    distribution
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable] the name of the continuous measurement variable
#' @param dist_col [variable attribute] the name of the variable attribute in
#'                                      meta_data that provides the expected
#'                                      distribution of a study variable
#' @param guess [logical] estimate parameters
#' @param par1 [numeric] first parameter of the distribution if applicable
#' @param par2 [numeric] second parameter of the distribution if applicable
#' @param end_digits [logical] internal use. check for end digits preferences
#' @inheritParams acc_distributions
#'
#' @return a list with:
#'   - `ResultData`: [data.frame] underlying the plot
#'   - `SummaryPlot`: [ggplot2::ggplot2] probability distribution plot
#'   - `SummaryTable`: [data.frame] with the columns `Variables` and `FLG_acc_ud_shape`
#'
#' @export
#' @importFrom MASS fitdistr
#' @importFrom ggplot2 ggplot theme_minimal geom_bar aes scale_fill_manual geom_line geom_hline xlab scale_color_manual geom_errorbar
#' @importFrom stats dunif dgamma dnorm punif pgamma pnorm median
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_shape_or_scale.html
#' )
acc_shape_or_scale <- function(resp_vars,
                               study_data,
                               label_col,
                               item_level = "item_level",
                               dist_col,
                               guess,
                               par1, par2,
                               end_digits,
                               flip_mode = "noflip",
                               meta_data = item_level,
                               meta_data_v2) {
  # TODO: remove dist_col, expect column "DISTRIBUTIONS"
  # preps ----------------------------------------------------------------------
  util_maybe_load_meta_data_v2()
  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .replace_missings = TRUE)

  # correct variable use?
  util_correct_variable_use("resp_vars",
                            need_type = "integer | float",
                            need_scale = "interval | ratio")

  if (missing(dist_col)) {
    dist_col <- DISTRIBUTION
    # util_message(
    #   c("A column of the metaddata specifying the distributions has",
    #   "not been specified. Trying the default %s."),
    #   dQuote(DISTRIBUTION), applicability_problem = TRUE)
  }

  if (!(dist_col %in% colnames(meta_data))) {
    util_error("Did not find variable attribute %s in the meta_data",
                       dist_col, applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  if (missing(guess) || is.na(guess[[1]])) {
    if (!missing(guess) && length(guess) > 1) {
      util_message("Have more than one value for guess, use the first one only",
                   applicability_problem = TRUE)
    }
    if (!missing(par1) && !missing(par2)) {
      guess <- FALSE
      util_message("Since parameters were specified: 'guess' is set to false",
                   applicability_problem = TRUE)
    } else {
      guess <- TRUE
    }
  }

  if (!(missing(end_digits)) && length(end_digits) > 1) {
    util_message(
      c("end_digits should be a scalar logical value. Have more than one",
        "value, use the first one only"),
      applicability_problem = TRUE)
    end_digits <- end_digits[[1]]
  }

  if (!missing(guess) && length(guess) > 1) {
    util_message(
      c("guess should be a scalar logical value.",
        "Have more than one value, use the first one only"),
      applicability_problem = TRUE)
    guess <- guess[[1]]
  }

  if (!missing(par1) && length(par1) > 1) {
    util_message(
      c("par1 should be a scalar numeric value. Have more than one value,",
        "use the first one only"),
      applicability_problem = TRUE)
    par1 <- par1[[1]]
  }

  if (!missing(par2) && length(par2) > 1) {
    util_message(
      c("par2 should be a scalar numeric value.",
        "Have more than one value, use the first one only"),
      applicability_problem = TRUE)
    par2 <- par2[[1]]
  }

  if (!missing(guess) && !all(is.logical(guess))) {
    guess <- suppressWarnings(as.logical(guess))
    if ((!all(is.logical(guess))) || any(is.na(guess))) {
      util_error("guess should be a logical value",
                 applicability_problem = TRUE)
    }
  }

  if (!missing(end_digits) && !all(is.logical(end_digits))) {
    end_digits <- suppressWarnings(as.logical(end_digits))
    if ((!all(is.logical(end_digits))) || any(is.na(end_digits))) {
      util_error("end_digits should be a logical value",
                 applicability_problem = TRUE)
    }
  }

  if (!missing(par1) && !all(is.numeric(par1))) {
    par1 <- suppressWarnings(as.numeric(par1))
    if ((!all(is.numeric(par1))) || any(is.na(par1))) {
      util_error("par1 should be a numeric value",
                 applicability_problem = TRUE)
    }
  }

  if (!missing(par2) && !all(is.numeric(par2))) {
    par2 <- suppressWarnings(as.numeric(par2))
    if ((!all(is.numeric(par2))) || any(is.na(par2))) {
      util_error("par2 should be a numeric value",
                 applicability_problem = TRUE)
    }
  }

  if (length(ds1[[resp_vars]]) == 0L || mode(ds1[[resp_vars]]) != "numeric") {
    util_error("resp_vars == '%s' must be a non-empty numeric variable",
               resp_vars, applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  # missings in resp_vars?
  n_prior <- dim(ds1)[1]
  ds1 <- ds1[!(is.na(ds1[[resp_vars]])), ]
  n_post <- dim(ds1)[1]

  if (n_post < n_prior) {
    util_message(paste0("Due to missing values in ", resp_vars, " ",
                        n_prior - n_post,
      " observations were deleted.",
      collapse = " "
    ), applicability_problem = FALSE)
  }

  if (guess == FALSE && (missing(par1) || missing(par2) ||
                         !is.finite(par1) || !is.finite(par2))) {
    util_error(c("Since 'guess' is not true finite numerical",
                 "parameters must be prespecified"),
               applicability_problem = TRUE)
  }

  if (missing(end_digits)) {
    end_digits <- FALSE
  }

  if (
    !is.character(meta_data[meta_data[[label_col]] == resp_vars, dist_col]) ||
    is.na(meta_data[meta_data[[label_col]] == resp_vars, dist_col]) ||
    trimws(meta_data[meta_data[[label_col]] == resp_vars, dist_col]) == "") {
    util_error(paste0("No distribution specified for ", resp_vars, " in ",
                      dist_col, collapse = ""),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  dist <- paste0(meta_data[meta_data[[label_col]] == resp_vars, dist_col])

  if (end_digits == TRUE) {
    dist <- "uniform"
  }

  if (guess && dist == "uniform") {
    par1 <- min(ds1[[resp_vars]], na.rm = TRUE)
    par2 <- max(ds1[[resp_vars]], na.rm = TRUE)
  }

  if (guess && dist %in% c("normal", "gamma")) {
    suppressWarnings(par1 <- fitdistr(ds1[[resp_vars]],
                                      densfun = dist)$estimate[1])
    suppressWarnings(par2 <- fitdistr(ds1[[resp_vars]],
                                      densfun = dist)$estimate[2])
  }

  # assign densities/probability arguments
  d.fun <- switch(dist, uniform = dunif, gamma = dgamma, normal = dnorm, NULL)
  p.fun <- switch(dist, uniform = punif, gamma = pgamma, normal = pnorm, NULL)

  if (is.null(d.fun)) {
    util_error(sprintf("This distribution '%s' is not supported yet...", dist),
               applicability_problem = TRUE)
  }

  # create histogram
  h1 <- graphics::hist(ds1[[resp_vars]], plot = FALSE)
  nobs <- sum(h1$counts)

  if (dist == "uniform" && (par1 < min(h1$breaks) | par2 > max(h1$breaks))) {
    breaks <- seq(min(par1, min(h1$breaks)), max(par2, max(h1$breaks)), by = 1)
    h1 <- graphics::hist(ds1[[resp_vars]], breaks = breaks, plot = FALSE)
    nobs <- sum(h1$counts)
  }

  if (identical(getOption("dataquieR.acc_shape_or_scale_ci",
                           dataquieR.acc_shape_or_scale_ci_default),
                 "binomial")) {
    ci <- util_binomial_ci
  } else if (identical(getOption("dataquieR.acc_shape_or_scale_ci",
                               dataquieR.acc_shape_or_scale_ci_default),
                     "multinomial")) {
    ci <- util_multinomial_ci
  } else {
    util_error("option %s should be either %s or %s",
               sQuote("dataquieR.acc_shape_or_scale_ci"),
               dQuote("multinomial"),
               dQuote("binomial"))
  }

  if (end_digits == FALSE) {
    if (all(util_is_integer(ds1[[resp_vars]])) &
        length(unique(ds1[[resp_vars]])) < 20 & dist == "uniform") {
      df1 <- data.frame(
        INTERVALS = factor(unique(ds1[[resp_vars]])[
          order(unique(ds1[[resp_vars]]))]),
        COUNT = as.vector(table(ds1[[resp_vars]])),
        PROB = as.vector(table(ds1[[resp_vars]])) / nobs,
        EXP_PROB = round(1 / length(unique(ds1[[resp_vars]])), digits = 2),
        EXP_COUNT = nobs * 1 / length(unique(ds1[[resp_vars]])),
        LOWER_CL = ci(as.vector(table(ds1[[resp_vars]])),
                                 alpha = 0.05 / (length(h1$mids) - 1))[, 1],
        UPPER_CL = ci(as.vector(table(ds1[[resp_vars]])),
                                 alpha = 0.05 / (length(h1$mids) - 1))[, 2]
      )
    } else {
      df1 <- data.frame(
        INTERVALS = h1$mids,
        COUNT = h1$counts,
        PROB = h1$counts / nobs,
        EXP_PROB = min(diff(h1$mids)) * d.fun(h1$mids, par1, par2),
        EXP_COUNT = nobs * diff(p.fun(h1$breaks, par1, par2)),
        LOWER_CL = ci(h1$counts,
                                 alpha = 0.05 / (length(h1$mids) - 1))[, 1],
        UPPER_CL = ci(h1$counts,
                                 alpha = 0.05 / (length(h1$mids) - 1))[, 2]
      )
    }
  } else {
    ds1[[resp_vars]] <- factor(ds1[[resp_vars]], levels = 0:9)
    df1 <- data.frame(
      INTERVALS = factor(0:9),
      COUNT = as.vector(table(ds1[[resp_vars]])),
      PROB = as.vector(table(ds1[[resp_vars]])) / nobs,
      EXP_PROB = 0.1,
      EXP_COUNT = nobs * 0.1,
      LOWER_CL = ci(as.vector(table(ds1[[resp_vars]])),
                               alpha = 0.05 / (length(h1$mids) - 1))[, 1],
      UPPER_CL = ci(as.vector(table(ds1[[resp_vars]])),
                               alpha = 0.05 / (length(h1$mids) - 1))[, 2]
    )
  }

  df1$GRADING <- factor(ifelse(df1$EXP_PROB < df1$UPPER_CL &
                                 df1$EXP_PROB > df1$LOWER_CL, 0, 1))

  df1$PROB <- round(df1$PROB, digits = 2)
  df1$EXP_PROB <- round(df1$EXP_PROB, digits = 2)
  df1$EXP_COUNT <- round(df1$EXP_COUNT, digits = 2)
  df1$LOWER_CL <- round(df1$LOWER_CL, digits = 2)
  df1$UPPER_CL <- round(df1$UPPER_CL, digits = 2)


  x2 <- seq(floor(min(h1$breaks)), ceiling(max(h1$breaks)),
            length.out = length(h1$breaks) * 100)
  y_line <- min(diff(h1$breaks)) * d.fun(x2, par1, par2)

  if (end_digits) {
    y_line <- rep(0.1, length = length(x2))
  }

  if (all(util_is_integer(ds1[[resp_vars]])) &
      length(unique(ds1[[resp_vars]])) < 20 & dist == "uniform") {
    y_line <- rep(round(1 / length(unique(ds1[[resp_vars]])), digits = 2),
                  length = length(x2))
  }

  df2 <- data.frame(
    x2 = x2,
    y_line = y_line
  )

  grading_cols <- c("#2166AC", "#B2182B")
  names(grading_cols) <- c("0", "1")

  p1 <- util_create_lean_ggplot(ggplot(df1, aes(x = INTERVALS, y = PROB)) +
    theme_minimal() +
    geom_bar(aes(fill = GRADING), stat = "identity") +
    scale_fill_manual(values = grading_cols, guide = "none") +
    geom_errorbar(aes(ymin = LOWER_CL, ymax = UPPER_CL), width = 0.1) + {
      if (!(end_digits) & dist != "uniform") {
        geom_line(data = df2, aes(x = x2, y = y_line, color = "#E69F00"),
                  linewidth = 1)
      }
    } + {
      if (!(end_digits) & dist == "uniform") {
        geom_hline(yintercept = unique(y_line), color = "#E69F00",
                   linewidth = 1)
      }
    } + {
      if (!(end_digits) & dist == "uniform") {
        xlab("Values")
      }
    } + {
      if (end_digits) geom_hline(yintercept = 0.1, color = "#E69F00",
                                 linewidth = 1)
    } + {
      if (end_digits) xlab("End digits")
    } + scale_color_manual(values = c("#E69F00"), guide = "none"),
    df1 = df1,
    grading_cols = grading_cols,
    end_digits = end_digits,
    x2 = x2,
    y_line = y_line,
    dist = dist,
    df2 = df2)

  p1 <- p1 + theme(legend.position = "none",
                   axis.text.y = element_text(size = 10),
                   axis.text.x = element_text(size = 10)) # also suppresses the legend in the ggplotly figure
  fli <- util_coord_flip(p = p1)
  p1 <- util_lazy_add_coord(p1, fli) # TODO: estimate w and h, since p is not using discrete axes
  is_flipped <- inherits(fli, "CoordFlip")
  p1 <- util_set_size(p1);
  #Put the old SummaryData table (that was not organized one row per variable)
  # inside the new result OtherTable that can be further processed
  OtherTable <- df1

  #OtherTable <- OtherTable[, !names(OtherTable) %in% c("GRADING")]

  #Create the SummaryTable
  SummaryTable <-  data.frame(Variables = resp_vars,
    FLG_acc_ud_shape = 1 - prod(1 - util_as_numeric(df1$GRADING))
  )


  no_intervals_flagged <- sum(as.numeric(as.character(df1$GRADING)))
  no_intervals_over_tot <- sum(as.numeric(as.character(df1$GRADING)))/nrow(df1)

  #SummaryData
#  SummaryData <-  data.frame(Variables = resp_vars,
#                             "Intervals_flagged" =
#                               paste0(no_intervals_flagged, " (",
#                                      no_intervals_over_tot,
#                                      "%)"))

  return(util_attach_attr(list(
    SummaryPlot = p1,
    ResultData  = OtherTable,
    SummaryTable = SummaryTable
  ), sizing_hints = list(
    figure_type_id = "bar_chart",
    rotated = is_flipped,
    number_of_bars = nrow(p1$data),
    range = max(p1$data$UPPER_CL) - min(p1$data$LOWER_CL),
    no_char_x = 4,
    no_char_y = 4
  )))
}

util_overwrite_if_requested <- function(dir, force_overwrite) {
  if (file.exists(dir)) {
    if (force_overwrite) {
      to_remove <-
        list.files(
          path = dir, all.files = TRUE,
          full.names = TRUE,
          recursive = TRUE,
          include.dirs = TRUE,
          no.. = TRUE
        )
      # Paranoia before deletion:
      to_remove_norm <- vapply(
        to_remove,
        normalizePath,
        FUN.VALUE = character(1),
        mustWork = FALSE
      )
      if (any(unlist(
        lapply(c(to_remove_norm), strsplit, .Platform$file.sep,
               fixed = TRUE)) == ".."))
        util_error("Refusing to unlink paths containing '..'")
      unlink(to_remove_norm,
             recursive = TRUE,
             force = TRUE,
             expand = FALSE)
      if (file.exists(dir) && !dir.exists(dir)) {
        util_error(c("%s already exists as a file, not a directory, cannot",
                     "use this as an output folder."),
                   dQuote(dir))
      }
    } else {
      util_error("%s already exists, cannot use this as an output folder.",
                 dQuote(dir))
    }
  }
  if (!dir.exists(dir))
    if (!dir.create(dir, recursive = TRUE, showWarnings = FALSE)) {
      util_error("Could not create %s", dQuote(dir))
  }
}

#' If on Windows, hide a file
#'
#' @param fn the file path + name
#'
#' @return `invisible(NULL)`
#'
#' @noRd
util_hide_file_windows <- function(fn) {
  if (.Platform$OS.type != "windows") return();
  fn <- normalizePath(fn)
  if (file.exists(fn)) {
    system(sprintf("attrib +h %s", shQuote(fn, type = "cmd")))
  }
  invisible(NULL)
}

#' Get duplicates including first instance
#'
#' @param x object to pass do [base::duplicated()] as `x`
#' @param ... passed to [base::duplicated()] as `...`
#'
#' @returns see [base::duplicated()]
#' @noRd
#' @seealso [`StackOverflow`](https://stackoverflow.com/a/7854620)
#' @author [`Joshua Ulrich`](https://stackoverflow.com/users/271616/joshua-ulrich)
util_duplicated_inclding_first <- function(x, ...) {
  if (is.atomic(x) || is.data.frame(x)) {
    return(duplicated(x, ...) | duplicated(x, fromLast=TRUE, ...))
  } else {
    util_error("Object of unsupported class %s passed",
               sQuote(paste(class(x), collapse = ", ")))
  }
}

#' Get formats for `DQ` categories
#'
#' @family summary_functions
#' @noRd
#'
#' @return [data.frame] columns: `categories` (e.g., "1" to "5"),
#'   `color` (e.g., "33 102 172", "67 147 195", "227 186 20", "214 96 77",
#'   178 23 43"), `label` (e.g., "OK", "unclear", "moderate", "important",
#'    "critical" )
util_get_ruleset_formats <- function() {

  rs <- util_get_rule_sets()

  if(!"dqi_catnum" %in% colnames(rs)) {
    max_cats <- 5
  } else {
    max_cats <- suppressWarnings(
      vapply(lapply(lapply(rs, `[[`, "dqi_catnum"), as.integer), max,
             na.rm = TRUE, FUN.VALUE = integer(1)))
  }

  shipped_ruleset_formats <- system.file("grading_formats.xlsx",
                                  package = "dataquieR")
  if (!nzchar(shipped_ruleset_formats) &&
      suppressWarnings(util_ensure_suggested("pkgload", err = FALSE)) &&
      pkgload::is_dev_package("dataquieR")) {
    if (util_is_try_error(try(silent = TRUE,
      shipped_ruleset_formats <- pkgload::package_file("inst",
                                                    "grading_formats.xlsx",
                                                    path = find.package(
                                                      "dataquieR"))))) {
      rlang::warn(sprintf(
        "Could not find package source, trying to use %s from installed package",
        sQuote("grading_formats.xlsx")
      ),
      .frequency_id =
        "pkgload_confusionfmts",
      .frequency = "once")
      shipped_ruleset_formats <-
        names(head(which(vapply(
          setNames(nm = file.path(.libPaths(), "dataquieR",
                                  "grading_formats.xlsx")), file.exists,
          FUN.VALUE = logical(1))), 1))
      if (length(shipped_ruleset_formats) != 1) {
        shipped_ruleset_formats <- ""
      }
    }
  }
  if (!nzchar(shipped_ruleset_formats)) {
    util_error(
      "Internal error with pkgload, please report, sorry: Could not find %s.",
      sQuote("shipped_ruleset_formats"))
  }
  shipped_ruleset_formats <- prep_get_data_frame(shipped_ruleset_formats)

  reftab <- try(
    prep_get_data_frame(getOption("dataquieR.grading_formats", dataquieR.grading_formats_default)),
    silent = TRUE
  ) # https://gitlab.com/libreumg/dataquier/-/issues/34#note_1597720844
  if (!inherits(reftab, "data.frame")) {
    if (getOption("dataquieR.grading_formats", dataquieR.grading_formats_default) != dataquieR.grading_formats_default) {
      util_message("Could not find ruleset formats %s, using the default formats.",
                   dQuote(getOption("dataquieR.grading_formats",
                                    dataquieR.grading_formats_default)),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = FALSE)
    }
    reftab <- shipped_ruleset_formats
  }

  util_expect_data_frame(reftab, col_names = c("category", "label", "color"))

  need <- seq_len(max_cats)
  have <- unique(reftab$category)

  if (!all(need %in% have)) {
    util_error(
      applicability_problem = TRUE,
      intrinsic_applicability_problem = FALSE,
      c("Did not find formats for all categories, need %s, found %s.",
        "%s may not match %s or one of these is not available."
      ),
      util_pretty_vector_string(need),
      util_pretty_vector_string(have),
      dQuote(getOption("dataquieR.grading_formats",
                       dataquieR.grading_formats_default)),
      dQuote(getOption("dataquieR.grading_rulesets",
                       dataquieR.grading_rulesets_default))
    )
  }

  reftab <- reftab[!util_empty(reftab$category), , FALSE]
  reftab[["category"]] <-
    suppressWarnings(as.integer(reftab[["category"]]))

  reftab
}

#' Prepare a label as part of a title text for `RMD` files
#'
#' @param s the label
#' @param html prepare the label for direct `HTML` output instead of `RMD`
#'
#' @return the escaped label
#' @export
#'
prep_title_escape <- function(s, html = FALSE) {
  # r <- gsub("&", "&amp;", s, fixed = TRUE)
  # r <- gsub("^#+", "", r, perl = TRUE)
  # r <- gsub("/", "&#47;", r, fixed = TRUE)
  # r <- gsub("@", "&#64;", r, fixed = TRUE)
  # r <- gsub("<", "&lt;", r, fixed = TRUE)
  # r <- gsub(">", "&gt;", r, fixed = TRUE)
  # r <- gsub("\"", "&#92;", r, fixed = TRUE)
  # r <- gsub("_", "&#95;", r, fixed = TRUE)
  # r <- gsub("~", "&#126;", r, fixed = TRUE)
  if (html) {
    r <- lapply(s, htmltools::pre)
    r <- vapply(s, as.character, FUN.VALUE = character(1))
  } else {
    r <- s
    r <- gsub("`", "", r, fixed = TRUE)
    r <- paste0("`", r, "`")
  }
  r
}

test_that("prep_title_escape works", {
  skip_on_cran()
  # html = TRUE
  expected_resut <- c('Hello' = "Hello", 'world!' = "world!")
  expect_equal(prep_title_escape(c("Hello", "world!"), html = TRUE), expected_resut)

  # html = FALSE
  expect_equal(prep_title_escape("Hello world!", html = FALSE), "`Hello world!`")
  expect_equal(prep_title_escape(c("String1", "String2"), html = FALSE), c("`String1`", "`String2`"))
  expect_equal(prep_title_escape("Test `String`", html = FALSE), "`Test String`")
})

# RECORD_MISSMATCH_CHECKTYPE

dataquieR.ELEMENT_MISSMATCH_CHECKTYPE_default <- "subset_u"

#' @name dataquieR.ELEMENT_MISSMATCH_CHECKTYPE
#' @title Metadata describes more than the current study data
#' @description
#' - `none`: no check will be provided about the match of variables and
#'          records available in the study data and described in the metadata
#' - `exact`: There must be a 1:1 match between the study data and metadata
#'            regarding data frames and segments variables and records
#' - `subset_u`: study data are a subset of metadata. All variables from the study
#'               data are expected to be present in the metadata, but one or
#'               more variables in the metadata are not expected to be
#'               present in the study data.
#'               In this case a variable present in
#'               the study data but not in the metadata would produce an issue.
#' - `subset_m`: metadata are a subset of study data. All variables in the metadata
#'               are expected to be present in the study data, but one or more
#'               variables in the study data are not expected to be
#'               present in the metadata.
#' @family options
#' @docType data
NULL

dataquieR.scale_level_heuristics_control_metriclevels_default <- 25
dataquieR.scale_level_heuristics_control_binaryrecodelimit_default <- 8

#' @name dataquieR.scale_level_heuristics_control_metriclevels
#' @title Number of levels to consider a variable metric in absence
#'        of [SCALE_LEVEL]
#' @description
#' If `SCALE_LEVEL` is not specified in the `meta_data`, it will be inferred
#' using a heuristic. This option defines, for numeric variables, the maximum
#' number of distinct data values for a variable to be considered categorical,
#' not ordinal.
#' @family options
#' @docType data
NULL

#' @name dataquieR.scale_level_heuristics_control_binaryrecodelimit
#' @title Number of levels to consider a variable ordinal in absence
#'        of [SCALE_LEVEL]
#' @description
#' If `SCALE_LEVEL` is not specified in the `meta_data`, it will be inferred
#' using a heuristic. This option defines, for numeric variables, the maximum
#' number of distinct data values for a variable to be considered ordinal.
#' @family options
#' @docType data
NULL

# TODO: Deprecate this block? until including dataquieR.CONDITIONS_WITH_STACKTRACE
dataquieR.ERRORS_WITH_CALLER_default <- TRUE
#' @name dataquieR.ERRORS_WITH_CALLER
#' @title Set caller for error conditions (to be deprecated)
#' @description
#' to be deprecated
#' @family options
#' @docType data
NULL

dataquieR.WARNINGS_WITH_CALLER_default <- TRUE
#' @name dataquieR.WARNINGS_WITH_CALLER
#' @title Set caller for warning conditions (to be deprecated)
#' @description
#' to be deprecated
#' @family options
#' @docType data
NULL

dataquieR.MESSAGES_WITH_CALLER_default <- FALSE
#' @name dataquieR.MESSAGES_WITH_CALLER
#' @title Set caller for message conditions (to be deprecated)
#' @description
#' to be deprecated
#' @family options
#' @docType data
NULL

dataquieR.CONDITIONS_WITH_STACKTRACE_default <- FALSE
#' @name dataquieR.CONDITIONS_WITH_STACKTRACE
#' @title Add stack-trace in condition messages (to be deprecated)
#' @description
#' to be deprecated
#' @family options
#' @docType data
NULL

dataquieR.CONDITIONS_LEVEL_TRHESHOLD_default <- 0
#' @name dataquieR.CONDITIONS_LEVEL_TRHESHOLD
#' @title Log Level
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.traceback_default <- FALSE
#' @name dataquieR.traceback
#' @title Include full trace-back in captured conditions
#' @description
#' Caveat: Needs really much memory
#' @family options
#' @docType data
NULL

dataquieR.acc_shape_or_scale_ci_default <- "multinomial"
#' @name dataquieR.acc_shape_or_scale_ci
#' @title Method for confidence intervals in `acc_shape_or_scale()` and `acc_end_digits()`
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.debug_default <- FALSE
#' @name dataquieR.debug
#' @title Call [browser()] on errors
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.flip_mode_default <- "default" # flip, noflip, auto
#' @name dataquieR.flip_mode
#' @title Flip-Mode to Use for figures
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.force_item_specific_missing_codes_default <- FALSE
#' @name dataquieR.force_item_specific_missing_codes
#' @title Converting [MISSING_LIST]/[JUMP_LIST] to a [MISSING_LIST_TABLE] create on list per item
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.progress_init_fkt_default <- NULL
#' @name progress_init_fkt
#' @title [function] to call on progress initialization
#' @description
#' has one argument, `n`, reporting the number of steps in the current
#' job. needed, e.g., by packages, such as `progressr`.
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.progress_fkt_default <- NULL
#' @name dataquieR.progress_fkt_default
#' @title [function] to call on progress increase
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.progress_msg_fkt_default <- NULL
#' @name dataquieR.progress_msg_fkt_default
#' @title [function] to call on progress message update
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.grading_rulesets_default <- "grading_rulesets"
#' @name dataquieR.grading_rulesets
#' @title Name of the [data.frame] featuring [GRADING_RULESET]
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.grading_formats_default <- "grading_formats"
#' @name dataquieR.grading_formats
#' @title Name of the [data.frame] featuring a format for grading-values
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.VALUE_LABELS_htmlescaped_default <- FALSE
#' @name dataquieR.VALUE_LABELS_htmlescaped
#' @title Assume, all [VALUE_LABELS] are [HTML escaped](https://www.w3.org/International/questions/qa-escapes)
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.acc_loess.min_obs_in_subgroup_default <- 30
#' @name dataquieR.acc_loess.min_obs_in_subgroup
#' @title Minimum observations for a level to be included by [acc_loess()]
#' @description
#' specify the minimum number of observations required for
#' each of the subgroups. Subgroups with fewer
#' observations are excluded. The default number
#' is `r dataquieR.acc_loess.min_obs_in_subgroup_default`.
#' @family options
#' @docType data
NULL

dataquieR.acc_loess.mark_time_points_default <- FALSE
#' @name dataquieR.acc_loess.mark_time_points
#' @title Display time-points in LOESS plots
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.acc_loess.plot_observations_default <- FALSE
#' @name dataquieR.acc_loess.plot_observations
#' @title Display observations in LOESS plots
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.acc_loess.plot_format_default <- "COMBINED"
#' @name dataquieR.acc_loess.plot_format
#' @title default for Plot-Format in [acc_loess()]
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.acc_multivariate_outlier.scale_default <- TRUE
#' @name dataquieR.acc_multivariate_outlier.scale
#' @title Apply min-max scaling in parallel coordinates figure to inspect multivariate outliers
#' @description
#' boolean, TRUE or FALSE
#' @family options
#' @docType data
NULL

dataquieR.non_disclosure_default <- FALSE
#' @name dataquieR.non_disclosure
#' @title Remove all observation-level-real-data from reports
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.lang_default <- ""
#' @name dataquieR.lang
#' @title Language-Suffix for metadata Label-Columns
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.force_label_col_default <- "auto"
#' @name dataquieR.force_label_col
#' @title Control, how the `label_col` argument is used.
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.testdebug_default <- FALSE
#' @name dataquieR.testdebug
#' @title Disable all interactively used metadata-based function argument provision
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.dontwrapresults_default <- FALSE
#' @name dataquieR.dontwrapresults
#' @title Disable automatic post-processing of `dataquieR` function results
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.guess_missing_codes_default <- FALSE
#' @name dataquieR.guess_missing_codes
#' @title Control, if `dataquieR` tries to guess missing-codes from the study data in absence of metadata
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.col_con_con_logical_default <- "#B35451"
#' @name dataquieR.col_con_con_logical
#' @title Color for logical contradictions
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.col_con_con_empirical_default <- "#B39651"
#' @name dataquieR.col_con_con_empirical
#' @title Color for empirical contradictions
#' @description
#' TODO
#' @family options
#' @docType data
NULL

dataquieR.fix_column_type_on_read_default <- FALSE
#' @name dataquieR.fix_column_type_on_read
#' @title Try to avoid fallback to string columns when reading files
#' @description
#' If a file does not feature column data types or features data types
#' cell-based, choose that type which matches the majority of the
#' sampled cells of a column for the column's data type.
#' @details
#' This may make you miss data type problems but it could fix them, so
#' [prep_get_data_frame()] works better.
#' @family options
#' @docType data
NULL

dataquieR.MAX_VALUE_LABEL_LEN_default <- 60
#' @name dataquieR.MAX_VALUE_LABEL_LEN
#' @title Maximum length for value labels
#' @description
#' value labels are restricted to this length
#' @family options
#' @docType data
NULL

# Maximum length for labels according to
# file names https://stackoverflow.com/a/265782/4242747
# and label handling in plot.ly
.MAX_LABEL_LEN <- 200L

dataquieR.MAX_LABEL_LEN_default <- 60
#' @name dataquieR.MAX_LABEL_LEN
#' @title Maximum length for variable labels [LABEL]
#' @description
#' All variable labels will be shortened to fit this maximum length.
#' Cannot be larger than 200 for technical reasons.
#' @family options
#' @docType data
NULL

dataquieR.MAX_LONG_LABEL_LEN_default <- 200
#' @name dataquieR.MAX_LONG_LABEL_LEN
#' @title Maximum length for long variable labels [LONG_LABEL]
#' @description
#' All long variable labels will be shortened to fit this maximum length.
#' Cannot be larger than 200 for technical reasons.
#' @family options
#' @docType data
NULL

dataquieR.max_group_var_levels_in_plot_default <- 20
#' @name dataquieR.max_group_var_levels_in_plot
#' @title Maximum number of levels of the grouping variable shown
#' individually in figures
#' @description
#' If there are more examiners or devices than can be shown individually,
#' they will be collapsed into "other".
#' @family options
#' @docType data
NULL

dataquieR.max_cat_resp_var_levels_in_plot_default <- 6
#' @name dataquieR.max_cat_resp_var_levels_in_plot
#' @title Maximum number of levels of the categorical response variable shown
#' individually in figures
#' @description
#' If there are more levels of a categorical response variable than can be
#' shown individually, they will be collapsed into "other".
#' @family options
#' @docType data
NULL

dataquieR.min_time_points_for_cat_resp_var_default <- 20
#' @name dataquieR.min_time_points_for_cat_resp_var
#' @title Minimum number of data points to create a time course plot for an
#' individual level of a categorical response variable
#' @description
#' If there are less observations for an individual level of a categorical
#' variable, it will not be shown in the time course plot.
#' @family options
#' @docType data
NULL

dataquieR.min_obs_per_group_var_in_plot_default <- 10
#' @name dataquieR.min_obs_per_group_var_in_plot
#' @title Minimum number of observations per grouping variable that is required
#' to include an individual level of the grouping variable in a figure
#' @description
#' Levels of the grouping variable with fewer observations than specified here
#' will be excluded from the figure.
#' @family options
#' @docType data
NULL

dataquieR.max_group_var_levels_with_violins_default <- 10
#' @name dataquieR.max_group_var_levels_with_violins
#' @title Maximum number of levels of the grouping variable shown
#' with individual histograms ('violins') in 'margins' figures
#' @description
#' If there are more examiners or devices, the figure will be reduced to
#' box-plots to save space.
#' @family options
#' @docType data
NULL

dataquieR.no_overall_in_bin_default <- TRUE
#' @name dataquieR.no_overall_in_bin
#' @title Suppress overall distribution in 'margins' figures for binary outcomes
#' @family options
#' @docType data
NULL

dataquieR.no_geom_count_in_bin_default <- TRUE
#' @name dataquieR.no_geom_count_in_bin
#' @title Suppress counts 'margins' figures for binary outcomes
#' @description
#' so they are not always including 0 and 1.
#' @family options
#' @docType data
NULL

dataquieR.MULTIVARIATE_OUTLIER_CHECK_default <- "auto"
#' @name dataquieR.MULTIVARIATE_OUTLIER_CHECK
#' @title Default availability of multivariate outlier checks in reports
#' @description
#' can be
#'   - `TRUE`: for `cross-item_level`-groups with `MULTIVARIATE_OUTLIER_CHECK`
#'             empty, do a multivariate outlier check
#'   - `FALSE`: for `cross-item_level`-groups with `MULTIVARIATE_OUTLIER_CHECK`
#'             empty, don't do a multivariate outlier check
#'   - `"auto"`: for `cross-item_level`-groups with `MULTIVARIATE_OUTLIER_CHECK`
#'             empty, do multivariate outlier checks, if there is no entry in
#'             the column [CONTRADICTION_TERM].
#' @family options
#' @docType data
NULL

dataquieR.MAHALANOBIS_THRESHOLD_default <- 0.975
#' @name dataquieR.MAHALANOBIS_THRESHOLD
#' @title Default availability of Mahalanobis based multivariate outlier checks in reports
#' @description
#' a number, see corresponding argument in [acc_mahalanobis()]
#' @family options
#' @docType data
NULL

dataquieR.acc_margins_sort_default <- TRUE
#' @name dataquieR.acc_margins_sort
#' @title Sort levels of the grouping variable in the 'margins' figures
#' @description
#' If this option is set to `TRUE`, the levels of the grouping variable in the
#' figure are sorted in descending order according to the number of
#' observations so that levels with more observations are easier to identify.
#' Otherwise, the original order of the levels is retained.
#' @family options
#' @docType data
NULL

dataquieR.acc_margins_num_default <- TRUE
#' @name dataquieR.acc_margins_num
#' @title Include number of observations for each level of the grouping variable
#' in the 'margins' figure
#' @description
#' If this option is set to `FALSE`, the figures created by `acc_margins` will
#' not include the number of observations for each level of the grouping
#' variable. This can be used to obtain clean static plots.
#' @family options
#' @docType data
NULL

dataquieR.des_summary_hard_lim_remove_default <- FALSE
#' @name dataquieR.des_summary_hard_lim_remove
#' @title Removal of hard limits from data before calculating
#'        descriptive statistics.
#' @description
#' can be
#'   - `TRUE`: values outside hard limits will be removed from the data
#'             before calculating descriptive statistics
#'   - `FALSE`: values outside hard limits will not be removed from the original
#'              data
#' @family options
#' @docType data
NULL

dataquieR.GAM_for_LOESS_default <- FALSE
#' @name dataquieR.GAM_for_LOESS
#' @title Enable to switch to a general additive model instead of LOESS
#' @description
#' If this option is set to `TRUE`, time course plots will use general additive
#' models (GAM) instead of LOESS when the number of observations exceeds a
#' specified threshold. LOESS computations for large datasets have a high
#' memory consumption.
#' @family options
#' @docType data
NULL

dataquieR.acc_loess.exclude_constant_subgroups_default <- FALSE
#' @name dataquieR.acc_loess.exclude_constant_subgroups
#' @title Exclude subgroups with constant values from LOESS figure
#' @description
#' If this option is set to `TRUE`, time course plots will only show subgroups
#' with more than one distinct value. This might improve the readability of
#' the figure.
#' @family options
#' @docType data
NULL

dataquieR.acc_loess.min_bw_default <- 0.2
#' @name dataquieR.acc_loess.min_bw
#' @title Lower limit for the LOESS bandwidth
#' @description
#' The value should be greater than 0 and less than or equal to 1. In general,
#' increasing the bandwidth leads to a smoother trend line.
#' @family options
#' @docType data
NULL

dataquieR.acc_loess.min_proportion_default <- 0.1
#' @name dataquieR.acc_loess.min_proportion
#' @title Lower limit for the proportion of cases or controls to create a
#' smoothed time trend figure
#' @description
#' The value should be greater than 0 and less than 0.4. If the proportion of
#' cases or controls is lower than the specified value, the LOESS figure will
#' not be created for the specified binary outcome.
#' @family options
#' @docType data
NULL

dataquieR.droplevels_ReportSummaryTable_default <- FALSE
#' @name dataquieR.droplevels_ReportSummaryTable
#' @title Show also unused levels in heatmaps
#' @description
#' if `TRUE`, levels not taken will not be displayed when printing/plotting
#' heatmap tables
#' @family options
#' @docType data
NULL

dataquieR.resume_checkpoint_default <- FALSE
#' @name dataquieR.resume_checkpoint
#' @title If result already exists in a `storr` back-end, re-use it
#' @description
#' if `TRUE`, computation won't be repeated, if a result already exist in the
#' output `storr`
#' @family options
#' @docType data
NULL

dataquieR.convert_to_list_for_lapply_default <- FALSE
#' @name dataquieR.convert_to_list_for_lapply
#' @title If report uses a `storr` back-end, do not convert to base-list
#' @description
#' if `TRUE` and a report uses a `storr`-back-end, convert it to a base list,
#' i.e., copy to the RAM, even if this would likely not be really needed for
#' apply-calls
#' @family options
#' @docType data
NULL

dataquieR.resume_print_default <- FALSE
#' @name dataquieR.resume_print
#' @title If output folder is not empty, try to resume stopped `print()`
#' @description
#' if `TRUE` and a report was already partially printed with also this option
#' `TRUE`, then, a second call to `print()` will resume the printing.
#' @family options
#' @docType data
NULL

dataquieR.study_data_cache_max_default <- Inf
#' @name dataquieR.study_data_cache_max
#' @title Maximum size of cache for curated study data
#' @description
#' `dataquieR` caches all used flavors of curated study data, e.g., having
#' missing codes replaced by `NA`s, having hard limits replaced by `NA`, ...
#' For larger sets of study data this can be very RAM consuming, so you can
#' control here the maximum size for this cache. Also, this cache is distributed
#' to all compute nodes in case of parallel computation, which may be very time-
#' consuming, and, on single-node-parallelization, also, it may be even more
#' RAM-consuming then.
#' @family options
#' @family study_data_cache
#' @docType data
NULL

dataquieR.precomputeStudyData_default <- FALSE
#' @name dataquieR.precomputeStudyData
#' @title Pre-compute different curation levels of study data
#' @description
#' as described in [dataquieR.study_data_cache_max], different flavors of
#' the study data are cached. With this option, you control, if before a report
#' is computed, a frequently needed bunch of such flavors are pre-computed and
#' distributed to the compute nodes. However, this may be time- and RAM-
#' consuming, so, you can turn the pre-computation off, which will still allow
#' the individual compute nodes to maintain such a cache but then growing on
#' demand on individual nodes separately, only. If
#' [dataquieR.study_data_cache_max] cannot handle all flavors, they may still be
#' pre-computed but immediately discarded.
#' @family options
#' @family study_data_cache
#' @docType data
NULL

dataquieR.study_data_cache_quick_fill_default <- TRUE
#' @name dataquieR.study_data_cache_quick_fill
#' @title Control the pre-computation of curation levels of study data
#' @description
#' as described in [`dataquieR.precomputeStudyData`], different flavors of
#' the study data are cached. With this option, you control, if before a report
#' is computed, only frequently needed bunch of such flavors are pre-computed,
#' or simply all possible flavors. Won't have any effect, if pre-computation
#' has been turned off.
#' @family options
#' @family study_data_cache
#' @docType data
NULL

dataquieR.study_data_cache_metrics_default <- FALSE
#' @name dataquieR.study_data_cache_metrics
#' @title Collect metrics on cache usage of study data cache
#' @description
#' if `TRUE`, collect metrics on the usage of the study data cache
#' described here: [dataquieR.study_data_cache_max]. Won't work, fully,
#' if running in parallel.
#' `r lifecycle::badge("experimental")`
#' @family options
#' @family study_data_cache
#' @docType data
NULL

#' @export
#' @title Default space for some metrics during report computation
#' @description
#' `r lifecycle::badge("experimental")`
#' @family study_data_cache
dataquieR.study_data_cache_metrics_env_default <- new.env(parent = emptyenv())

#' @name dataquieR.study_data_cache_metrics_env
#' @title [environment] for storing metrics on the study data cache
#' @description
#' this is the environment, where metrics will be stored, if
#' [dataquieR.study_data_cache_metrics]-`option()` has been set `TRUE`.
#' `r lifecycle::badge("experimental")`
#' @family options
#' @family study_data_cache
#' @docType data
NULL

dataquieR.ignore_empty_vars_default <- "auto"
#' @name dataquieR.ignore_empty_vars
#' @title [character] remove variables with only empty values
#' @description
#' remove variables with only empty values (`NA`, `".  "`,
#' `""` or similar) from reports. `auto` means, such variables are removed, if
#' we have more than 20% of the variables empty.
#' @family options
#' @family study_data_cache
#' @docType data
NULL # 20% are implemented at dq_report2, currently, search for ignore_empty_vars20 in this file.


if (rlang::is_installed("stringi")) {
  locale_string <- stringi::stri_locale_get()
} else {
  if (is.null(getOption("dataquieR.locale"))) {
    packageStartupMessage(
      paste("stringi is not installed to detect system default locale.",
            "Falling back to en_US. You can set the option dataquieR.locale to",
            "your desired ISO Code for the locale to avoid this check."))
  }
  locale_string <- "en_US"
}
language_code <- sub("_.*", "", locale_string)

dataquieR.locale_default <- language_code
#' @name dataquieR.locale
#' @title [character] default language for type conversion
#' @description
#' the language to use for type conversions (`en`, `de`, `fr`, `cn`, `ca`, ...)
#' only used by `util_adjust_data_type2()`, currently
#' @family options
#' @docType data
NULL # TODO: Overlaps in parts with dataquieR.lang

dataquieR.type_adjust_parallel_default <- "TRUE"
#' @name dataquieR.type_adjust_parallel
#' @title [character] try to do type adjustments in parallel
#' only, if [dq_report2()] was called with `cores = 2` or higher.
#' @family options
#' @docType data
NULL

dataquieR.old_type_adjust_default <- "FALSE"
#' @name dataquieR.old_type_adjust
#' @title [character] use the old type conversion code (slower)
#' @family options
#' @docType data
NULL
# For a benchmark, set the option to "TRUE", then run
# microbenchmark::microbenchmark(
#   util_adjust_data_type(s, meta_data = m),
#   util_adjust_data_type2(s, meta_data = m)
# )

dataquieR.old_factor_handling_default <- "FALSE"
#' @name dataquieR.old_factor_handling
#' @title [character] use the old handling of study data already featuring
#'                    factors
#' @description
#' if `study_data` comes as a data frame, it may already feature factors. if
#' a column has the `DATA_TYPE` `integer` in the meta data, the factor was
#' converted to integer using [as.integer()], which caused unexpected behavior.
#' if this option is set to `"FALSE"` (the new default), the conversion will now
#' try to apply `as.character(column_data)`, first.
#'
#' @family options
#' @docType data
NULL

dataquieR.print_block_load_factor_default <-
  Sys.getenv("DATAQUIER_PRINT_BLOCK_LOAD_FACTOR", unset = "1")
#' @name dataquieR.print_block_load_factor
#' @title [numeric]
#' @description
#' multiply size of parallel compute blocks
#' by this factor. the higher it is set,
#' the less smooth progress bar grows, but
#' setting it to a huge number can really
#' speed up the rendering process by
#' approx. 10%. Either set to 1 for full
#' progress control or large (e.g., 1000000)
#' for maximum speed.
#' @family options
#' @family study_data_cache
#' @docType data
NULL

dataquieR.guess_character_default <- FALSE
#' @name dataquieR.guess_character
#' @title For metadata guessing, try to guess [DATA_TYPE] from the data values
#' @description
#' By default, the [DATA_TYPE] is derived from the R data type of the study
#' data. However, when data are imported from plain text files, it can be more
#' appropriate to examine the actual values and infer the data type based on
#' their content. This option enables that behavior: set
#' `dataquieR.guess_character` to `TRUE` to infer data types from the observed
#' values rather than relying solely on the column’s `class` in the data frame.
#' @family options
#' @docType data
NULL

dataquieR.dt_adjust_default <- "TRUE"
#' @name dataquieR.dt_adjust
#' @title [character] Adjust data types according to metadata
#' @description
#' also reports inadmissible data types. can be turned off for performance
#' reasons, if the data source is already type-safe (e.g., a database)
#' ***use with care, may cause pipelines breaking (maybe only in the final
#' rendering step), if the data type is incorrectly set for some columns.***
#'
#' @family options
#' @docType data
NULL

dataquieR.study_data_colnames_case_sensitive_default <- "TRUE"
#' @name dataquieR.study_data_colnames_case_sensitive
#' @title [character] Are column names in study data considered
#'                    case-sensitive for mapping
#' @description
#' if `TRUE`, `colnames(study_data)` replaced by the capitalization used in the
#' metadata using a case-insensitive matching, first.
#'
#' @family options
#' @docType data
NULL

dataquieR.lazy_plots_default <- "TRUE"
#' @name dataquieR.lazy_plots
#' @title [character] plots realized lazy
#' @description
#' if `TRUE`, plots are not realized until needed in side reports to save
#' memory.
#'
#' @family options
#' @docType data
NULL

dataquieR.lazy_plots_cache_default <- "FALSE"
#' @name dataquieR.lazy_plots_cache
#' @title [character] cache realizations
#' @description
#' if `TRUE`, realized plots are cached, may need more memory.
#'
#' @family options
#' @docType data
NULL

dataquieR.lazy_plots_gg_compatibility_default <- "TRUE"
#' @name dataquieR.lazy_plots_gg_compatibility
#' @title [character] be as compatible with `ggplot2` objects as possible
#' @description
#' if `TRUE`, plot promises are blessed in an `S7` class so they behave almost
#' like "real" `ggplot2` objects, so you normally do not need to call
#' [prep_realize_ggplot()] on them. However, this comes with a small memory
#' overhead, so, you can disable this.
#'
#' @family options
#' @docType data
NULL

#' Utility function to dichotomize variables
#'
#' This function uses the metadata attributes `RECODE_CASES` and/or
#' `RECODE_CONTROL` to dichotomize the data. 'Cases' will be recoded to 1,
#' 'controls' to 0. The recoding can be specified by an interval (for metric
#' variables) or by a list of categories separated by the 'SPLIT_CHAR'. Recoding
#' will be used for data quality checks that include a regression model.
#'
#' @param study_data study data without jump/missing codes as specified in the
#'                   code conventions
#' @param meta_data metadata as specified in the code conventions
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#'
#'
#' @family data_management
#' @concept data_management
#' @noRd
util_dichotomize <- function(study_data,
                             meta_data,
                             label_col = VAR_NAMES) {

  # if (!isTRUE(attr(study_data, "Codes_to_NA"))) {
  #   util_error("Expect to have missings already replaced.")
  # }
  # would be a nice check, but the attribute gets lost when subsetting ds1,
  # e.g. to select certain columns and rows

  vrcnt <- 0
  list_out <- list()

  for (var in colnames(study_data)) {
    recode_cases <- meta_data[meta_data[[label_col]] == var, RECODE_CASES]
    recode_control <- meta_data[meta_data[[label_col]] == var, RECODE_CONTROL]
    case_vec <- NULL
    contr_vec <- NULL
    val_lab_applied <- is.factor(study_data[[var]])
    val_lab <- util_parse_assignments(
      meta_data[[VALUE_LABELS]][meta_data[[label_col]] == var])
    # TODO STS: What needs to be done in case we have a VALUE_LABEL_TABLE
    # instead of VALUE_LABELS here?

    if ((!is.null(recode_cases) && !util_empty(recode_cases)) ||
        (!is.null(recode_control) && !util_empty(recode_control))) {

      # step 1: preprocess metadata information
      if (!is.null(recode_cases) && !util_empty(recode_cases)) {
        case_vec <- trimws(strsplit(recode_cases, SPLIT_CHAR, fixed = TRUE)[[1]])
        case_vec <- case_vec[!util_empty(case_vec)]
        if (length(case_vec) > 0 & !val_lab_applied) {
          # The definition might use the interval notation:
          case_int <- suppressWarnings(lapply(case_vec, util_parse_interval))
          is_interval <- vapply(case_int, FUN.VALUE = logical(1),
                                function(int) { inherits(int, "interval") })
          case_int <- case_int[is_interval]
          if (length(case_int) > 0) {
            case_vec <- case_int
          }
        } else if (length(case_vec) > 0 & val_lab_applied) {
          if (all(case_vec %in% names(val_lab))) {
            # If the levels are given as integers, we need to transform them to
            # value labels.
            case_vec <- unlist(lapply(case_vec, function(cc) {
              val_lab[[as.character(cc)]]
            }))
          }
        }
      }
      if (!is.null(recode_control) && !util_empty(recode_control)) { # same as for cases above
        contr_vec <- trimws(strsplit(recode_control, SPLIT_CHAR, fixed = TRUE)[[1]])
        contr_vec <- contr_vec[!util_empty(contr_vec)]
        if (length(contr_vec) > 0 & !val_lab_applied) {
          contr_int <- suppressWarnings(lapply(contr_vec, util_parse_interval))
          is_interval <- vapply(contr_int, FUN.VALUE = logical(1),
                                function(int) { inherits(int, "interval") })
          contr_int <- contr_int[is_interval]
          if (length(contr_int) > 0) {
            contr_vec <- contr_int
          }
        } else if (length(contr_vec) > 0 & val_lab_applied) {
          if (all(contr_vec %in% names(val_lab))) {
            contr_vec <- unlist(lapply(contr_vec, function(cc) {
              val_lab[[as.character(cc)]]
            }))
          }
        }
      }

      # step 2: apply recoding
      if (length(case_vec) == 0 & length(contr_vec) == 0) {
        util_warning(paste("Variable", var, "could not be recoded."))
      } else {
        col <- study_data[[var]]
        na <- is.na(col)
        if (length(case_vec) > 0) {
          if (is.list(case_vec)) {
            case_ind <- apply(as.data.frame(
              lapply(case_vec, function(ii) { redcap_env$`in`(col, ii) })),
              MARGIN = 1, FUN = any)
          } else {
            case_ind <- col %in% case_vec
          }
          if (length(contr_vec) > 0) {
            if (is.list(contr_vec)) {
              contr_ind <- apply(as.data.frame(
                lapply(contr_vec, function(ii) { redcap_env$`in`(col, ii) })),
                MARGIN = 1, FUN = any)
            } else {
              contr_ind <- col %in% contr_vec
            }
          } else {
            contr_ind <- !case_ind
          }
        } else {
          if (is.list(contr_vec)) {
            contr_ind <- apply(as.data.frame(
              lapply(contr_vec, function(ii) { redcap_env$`in`(col, ii) })),
              MARGIN = 1, FUN = any)
          } else {
            contr_ind <- col %in% contr_vec
          }
          case_ind <- !contr_ind
        }

        if (length(c(which(contr_ind), which(case_ind))) == 0) {
          util_warning(paste("Variable", var, "could not be recoded."))
        } else if (length(intersect(which(contr_ind), which(case_ind))) > 0) {
          util_warning(paste("Variable", var, "could not be recoded.",
                             "Cases and controls overlap."))
        } else {
          vrcnt <- vrcnt + 1
          col <- NA
          col[case_ind] <- 1
          col[contr_ind] <- 0
          col[na] <- NA
          study_data[, var] <- col
          list_out[[var]] <- paste(
            paste("Cases (1):",
                  ifelse(length(case_vec) == 0,
                         "any other value",
                         paste(unlist(lapply(case_vec, function(cc) {
                           if (is.list(cc)) {
                             paste0(ifelse(cc$inc_l, "[", "("),
                                    cc$low, ";", cc$upp, ifelse(cc$inc_u, "]", ")"))
                             } else {
                               cc
                             }})), collapse = " | "))),
            paste("Control (0):",
                  ifelse(length(contr_vec) == 0,
                         "any other value",
                         paste(unlist(lapply(contr_vec, function(cc) {
                           if (is.list(cc)) {
                             paste0(ifelse(cc$inc_l, "[", "("),
                                    cc$low, ";", cc$upp, ifelse(cc$inc_u, "]", ")"))
                           } else {
                             cc
                           }})), collapse = " | "))),
            sep = ". ")
        }
      }
    }
  }

  if (vrcnt > 0) {
    util_message(sprintf("Recoded %d variable(s).", vrcnt))
  }

  attr(study_data, "Dichotomization") <- list_out
  return(util_attach_attr(study_data))
}

#' Find a foreground color for a background
#'
#' black or white
#'
#' @param cl colors
#'
#' @return black or white for each cl
#' @seealso [`stackoverflow.com`](https://stackoverflow.com/a/24810681)
#' @noRd
util_get_fg_color <- function(cl) {
  cl <- col2rgb(util_col2rgb(cl), alpha = TRUE)
  brightness <- cl["red", ] * 0.299 + cl["green", ] * 0.587 +
    cl["blue", ] * 0.114
  ifelse(brightness > 160, "#000000", "#ffffff")
}

#' Plot a `ggplot2` figure using `plotly`
#'
#' @param x [ggplot2::ggplot2] object
#' @param sizing_hints `object` additional metadata about the natural figure size
#'
#' @return `htmltools` compatible object
#'
#' @noRd
util_plot_figure_plotly <- function(x, sizing_hints = NULL) {
  withCallingHandlers({
    if (inherits(x, "dq_lazy_ggplot")) x <- prep_realize_ggplot(x)
    if (inherits(x, "ggplot") && inherits(x, "patchwork")) {
      x <- htmltools::plotTag(x, "", width = 640) # leave patchworks as is, so far: TODO!!
    } else if ((inherits(x, "ggplot") || inherits(x, "ggmatrix") ||
                inherits(x, "svg_plot_proxy") ||
                inherits(x, "util_pairs_ggplot_panels")) &&
               !util_is_svg_object(x)) {

      title    <- NULL
      subtitle <- NULL

      ## Titel und Untertitel S7-sicher aus ggplot_build(x) holen
      try({
        build_obj   <- ggplot2::ggplot_build(x)
        plot_obj    <- util_gg_get(build_obj, "plot")
        labels_obj  <- if (!is.null(plot_obj)) util_gg_get(plot_obj, "labels") else NULL
        title       <- if (!is.null(labels_obj)) util_gg_get(labels_obj, "title") else NULL
      }, silent = TRUE)

      try({
        build_obj   <- ggplot2::ggplot_build(x)
        plot_obj    <- util_gg_get(build_obj, "plot")
        labels_obj  <- if (!is.null(plot_obj)) util_gg_get(plot_obj, "labels") else NULL
        subtitle    <- if (!is.null(labels_obj)) util_gg_get(labels_obj, "subtitle") else NULL
      }, silent = TRUE)

      py <- attr(x, "py")
      if (is.null(py)) {
        py <- attr(x, "plotly")
      }
      if (is.null(py) || !inherits(py, "plotly")) {
        x <- util_remove_dataquieR_result_class(x)
        # x <- x + ggplot2::theme(aspect.ratio = NULL) # prevent a warning
        py <- suppressWarnings( # prevent a warning
          force(util_ggplotly(x)))#, height = 480, width = 1040)))
        # TODO: only, if attr(dqr, "dont_util_adjust_geom_text_for_plotly") is not TRUE
        py <- util_adjust_geom_text_for_plotly(py)
      }
      x <- plotly::layout(py,
                          autosize = !FALSE,
                          margin = list(autoexpand = !FALSE,
                                        r = 200,
                                        b = 100)
      )
      if (!is.null(title) && is.null(subtitle)) {
        x <- plotly::layout(x, title = title)
      } else if (!is.null(title) && !is.null(subtitle)) {
        # if (packageVersion("plotly") > "4.10.4") { # should be supported, soon
        #   x <- plotly::layout(x, title = list(text = title,
        #                                       subtitle = list(text = subtitle)))
        # } else {
        x <- plotly::layout(
          x,
          title = list(
            text = paste0(
              title,
              "<br />",
              "<sub>",
              subtitle,
              "</sub>"
            )
          )
        )
        # }
      }
      x <- plotly::config(x, # TODO: generalize and combine with util_pretty_print
                          responsive = TRUE,
                          autosizable = TRUE,
                          fillFrame = TRUE,
                          displaylogo = FALSE,
                          modeBarButtonsToRemove = list("toImage"),
                          modeBarButtonsToAdd =
                            list(
                              list(
                                name = "Save as image",
                                icon = htmlwidgets::JS("Plotly.Icons.camera"),
                                click = htmlwidgets::JS(paste(sep = "\n",
                                                              "function(gd) {",
                                                              "  plotlyDL(gd)",
                                                              "}"))
                              ),
                              list(
                                name = "Restore initial Size",
                                icon =
                                  htmlwidgets::JS("PlotlyIconshomeRED()"),
                                click =
                                  htmlwidgets::JS(
                                    paste(sep = "\n",
                                          "function() {",
                                          "   if (togglePlotlyWindowZoom instanceof Function) {",
                                          "     togglePlotlyWindowZoom()",
                                          "   }",
                                          "}"
                                    )
                                  )),
                              list(
                                name = "Print plot",
                                icon = htmlwidgets::JS("PlotlyIconsprinter()"),
                                click = htmlwidgets::JS(paste(sep = "\n",
                                                              "function(gd) {",
                                                              "  plotlyPrint(gd)",
                                                              "}"))
                              ),
                              list(
                                name = "Download as PDF",
                                icon = htmlwidgets::JS("PlotlyIconspdf()"),
                                click = htmlwidgets::JS(paste(sep = "\n",
                                                              "function(gd) {",
                                                              "  downloadPlotlyAsPDF(gd, {orientation: 'landscape', dpi: 300});",
                                                              "}"))
                              )))
      util_ensure_suggested("htmlwidgets",
                            goal = "Creating the Print Dialog for Plotly")
      x <- htmlwidgets::onRender(x, "
                          function(el, x) {
                            if (typeof injectPlotlyPrintDialog === 'function') {
                              injectPlotlyPrintDialog();
                            }
                          }
                        ")
    } else {
      x <- util_plot_figure_no_plotly(x = x, sizing_hints = sizing_hints)
    }
  }, # nocov start
  warning = function(cond) { # suppress a waning caused by ggplotly for barplots
    if (startsWith(conditionMessage(cond),
                   "'bar' objects don't have these attributes: 'mode'") ||
        startsWith(conditionMessage(cond),
                   "'box' objects don't have these attributes: 'mode'")) {
      invokeRestart("muffleWarning")
    }
    if (any(grepl("the mode", conditionMessage(cond)))) {
      invokeRestart("muffleWarning")
    }
  },
  message = function(cond) { # suppress eine Message von ggplotly
    if (startsWith(conditionMessage(cond),
                   "'bar' objects don't have these attributes: 'mode'") ||
        startsWith(conditionMessage(cond),
                   "'box' objects don't have these attributes: 'mode'")) {
      invokeRestart("muffleMessage")
    }
    if (any(grepl("the mode", conditionMessage(cond)))) {
      invokeRestart("muffleMessage")
    } # nocov end
  })
  x
}

#' Return the pre-computed `plotly` from a `dataquieR` result
#'
#' @param res the `dataquieR` result
#' @param ... not used
#'
#' @return a `plotly` object
#' @family plotly_shims
#' @concept plotly_shims
#' @noRd
util_as_plotly_from_res <- function(res, ...) {
  util_stop_if_not(
    "Internal error, sorry: res w/o PlotlyPlot in util_as_plotly_from_res(), please report" =
      "PlotlyPlot" %in% names(res))
  have_plot_ly <- util_ensure_suggested("plotly", "plot interactive figures", err = FALSE)
  util_stop_if_not("Internal error, sorry: plotly uninstalled in util_as_plotly_from_res(), please report" =
                     have_plot_ly)
  py <- res[["PlotlyPlot"]]
  util_stop_if_not("Internal error in util_as_plotly_from_res(), sorry: PlotlyPlot should be a plotly, please report" =
                     inherits(py, "plotly"))
  py
}

#' Wrapper around `plotly::ggplotly` for optional dependency and lazy plots
#'
#' @param p a \code{ggplot} or \code{dq_lazy_ggplot} object.
#' @param ... passed through to \code{plotly::ggplotly}.
#'
#' @return A \code{plotly} object.
#'
#' @noRd
util_ggplotly <- function(p , ...) {
  util_ensure_suggested("plotly")
  if (inherits(p, "dq_lazy_ggplot")) {
    p <- prep_realize_ggplot(p)
  }
  r <- try(plotly::ggplotly(p, ...))
  if (util_is_try_error(r)) {
    r <- util_plotly_text(conditionMessage(attr(r, "condition")))
  }
  # try to reduce size of the plotly object
  r$x$visdat   <- NULL
  r$x$attrs    <- NULL    # no manipulation with R needed any more.
  r$x$cur_data <- NULL
  r
}

#' Wrapper around plotly::plotly_build for optional dependency and lazy plots
#'
#' @noRd
util_plotly_build <- function(p, ...) {
  util_ensure_suggested("plotly")
  if (inherits(p, "dq_lazy_ggplot")) {
    p <- prep_realize_ggplot(p)
  }

  plotly::plotly_build(p, ...)
}

test_that("util_plot_figure_plotly works", {
  skip_if_not_installed("plotly")
  skip_on_cran()
  # Create test ggplot
  p1 <- ggplot(mtcars) +
    geom_point(aes(mpg, disp))

  result <- util_plot_figure_plotly(p1)

  # Check that the result is a plotly object
  expect_s3_class(result, "plotly")

  # Create a test patchwork
  p2 <- ggplot(mtcars) +
    geom_boxplot(aes(gear, disp, group = gear))
  p3 <- p1 + p2

  result_2 <- util_plot_figure_plotly(p3)

  # Check that the result is an HTML plot tag
  expect_s3_class(result_2, "shiny.tag")
})

#' Re-Code labels with their respective codes according to the `meta_data`
#'
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return [data.frame] modified study data with labels replaced by the codes
#' @export
prep_apply_coding <- function(study_data, meta_data_v2,
                              item_level = "item_level",
                              meta_data = item_level) {
  util_maybe_load_meta_data_v2()
  util_expect_data_frame(study_data, keep_types = TRUE)
  prep_prepare_dataframes(.adjust_data_type = FALSE)
  util_expect_data_frame(meta_data, c(VAR_NAMES, VALUE_LABEL_TABLE))
  resp_vars <-
    meta_data[!util_empty(meta_data[[VALUE_LABEL_TABLE]]), VAR_NAMES, drop = TRUE]
  value_label_tables <- prep_map_labels(resp_vars,
                                  VALUE_LABEL_TABLE,
                                  item_level = meta_data)
  sdrv <-
    mapply(resp_vars, value_label_tables, SIMPLIFY = FALSE, FUN =
             function(rv, vlt) {
      vltdf <- prep_get_data_frame(vlt)
      VL <- setNames(vltdf[[CODE_VALUE]], nm = vltdf[[CODE_LABEL]])
      dt <- study_data[, rv, drop = TRUE]
      r <- VL[as.character(dt)]
      r[is.na(r) & ! is.na(dt)] <- dt[is.na(r) & ! is.na(dt)]
      if (
        all(!is.na(suppressWarnings(as.integer(vltdf[[CODE_VALUE]]))))
        &&
        (all(vltdf[[CODE_VALUE]] ==
             suppressWarnings(as.integer(vltdf[[CODE_VALUE]]))))
      ) {
        r <- as.integer(r)
      }
      r
    })
  study_data[, resp_vars] <- sdrv
  study_data
}

#' Check for unexpected data record count at the data frame level
#'
#' @description
#' This function contrasts the expected record number in each study in
#' the metadata with the actual record number in each study data frame.
#'
#' [Indicator]
#'
#' @param data_record_count [integer]  an integer vector with the number of expected data records per study data frame, mandatory.
#' @param identifier_name_list [character] a character vector indicating the name of each study data frame, mandatory.
#'
#' @inheritParams .template_function_indicator
#'
#' @return a [list] with
#'   - `DataframeData`: data frame with the results of the quality check for unexpected data elements
#'   - `DataframeTable`: data frame with selected unexpected data elements check results, used for the data quality report.
#'
#' @export
int_unexp_records_dataframe <- function(identifier_name_list,
                              data_record_count,
                              meta_data_dataframe = "dataframe_level",
                              meta_data_v2,
                              dataframe_level) {

  # Checks arguments ----
  util_maybe_load_meta_data_v2()

  util_ck_arg_aliases()

  if (missing(identifier_name_list) &&
      missing(data_record_count) &&
      missing(meta_data_dataframe) &&
      formals()$meta_data_dataframe %in% prep_list_dataframes()) {
    meta_data_dataframe <- force(meta_data_dataframe)
  }

  if (missing(identifier_name_list) &&
      missing(data_record_count) &&
      !missing(meta_data_dataframe)) {
    meta_data_dataframe <- prep_check_meta_data_dataframe(meta_data_dataframe)
    meta_data_dataframe <- meta_data_dataframe[
      vapply(meta_data_dataframe[[DF_NAME]],
             function(x) { !util_is_try_error(try(prep_get_data_frame(data_frame_name = x,
                                                                      keep_types = TRUE), silent = TRUE)) },
             FUN.VALUE = logical(1))
      , , drop = FALSE]
    meta_data_dataframe <- meta_data_dataframe[
      !util_empty(meta_data_dataframe[[DF_RECORD_COUNT]])
      , , drop = FALSE]
    # TODO: if nothing left
    identifier_name_list <- meta_data_dataframe[[DF_NAME]];
    data_record_count <- meta_data_dataframe[[DF_RECORD_COUNT]]
  } else if (!missing(meta_data_dataframe)) {
    util_error(c("I have %s and one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "data_record_count"
                 )),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "data_record_count"
                 )))
  } else if (missing(meta_data_dataframe) && (
    missing(identifier_name_list) ||
    missing(data_record_count)
  )) {
    util_error(c("I don't have %s and also miss at least",
                 "one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "data_record_count"
                 )),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("identifier_name_list",
                   "data_record_count"
                 )))
  }

  util_expect_scalar(identifier_name_list,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)

  util_expect_scalar(data_record_count,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.numeric)

  util_stop_if_not(length(identifier_name_list) == length(data_record_count),
                   label =
                     sprintf("In %s, %s and %s should have the same length: %s",
                             dQuote("int_unexp_elements"),
                             dQuote("identifier_name_list"),
                             dQuote("data_record_count"),
                             "They represent a mapping."
                     ))

  # Check for unexpected records  ----

  names(data_record_count) <- identifier_name_list

  result <- lapply(setNames(nm = identifier_name_list), function(current_df) {

    # Convert data from list to data frame
    data_current_df <- util_expect_data_frame(current_df, dont_assign = TRUE)

    # Select variables from data and metadata
    data_records <- nrow(data_current_df)
    metadata_records <- data_record_count[[current_df]]

    res_tmp <- data.frame(
      check.names = FALSE,
      "Check" = "Records",
      "Data frame" = current_df,
      "Unexpected records" = !(data_records == metadata_records),
      "Number of records in data" = data_records,
      "Number of records in metadata" = metadata_records,
      "Number of mismatches" =
        abs(round(data_records - metadata_records, 3)),
      "Percentage of mismatches" =
        abs(round(100 * ( data_records - metadata_records ) / metadata_records, 3)),
      "GRADING" = ifelse(data_records == metadata_records, 0, 1),
      stringsAsFactors = FALSE
    )

    return(res_tmp)
  })

  res_df <- do.call(rbind.data.frame, result)

  res_pipeline <- data.frame( # TODO: make res_df from these names, not vice versa.
    "Level" = "Dataframe",
    "DF_NAME" = res_df[["Data frame"]],
    "NUM_int_sts_countre" = res_df[["Number of mismatches"]],
    "PCT_int_sts_countre" = res_df[["Percentage of mismatches"]],
    "GRADING" = res_df[["GRADING"]],
    stringsAsFactors = FALSE
  )

  return(list(
    DataframeData = res_df,
    DataframeTable = res_pipeline
  ))
}

#' Convert a category to an ordered factor (`1:5`)
#'
#' @param category vector with categories
#'
#' @return an ordered factor
#' @family summary_functions
#' @noRd
util_as_cat <- function(category) {
  if (is.character(category)) {
    category <- gsub("^cat", "", category)
  }
  category <- suppressWarnings(as.integer(category))
  range <- 1:max(5L, category, na.rm = TRUE)
  factor(
    category,
    levels = range,
    labels = paste0("cat", range),
    ordered = TRUE
  )
}

#' Convert a category to a number (`1:5`)
#'
#' @param category vector with categories
#'
#' @return an integer
#' @family summary_functions
#' @noRd
util_as_integer_cat <- function(category) {
  if (!is.ordered(category)) category <- util_as_cat(category)
  util_stop_if_not(all(startsWith(levels(category), "cat")))
  numcat <- substring(levels(category), 4)
  util_stop_if_not(all(numcat == as.integer(numcat)))
  levels(category) <- gsub("^cat", "", levels(category))
  util_as_numeric(category)
}

#' Convert a full `dataquieR` report to a `data.frame`
#'
#' @description
#'
#' Deprecated
#'
#' @param ... Deprecated
#' @param x Deprecated
#'
#' @return Deprecated
#' @export
as.data.frame.dataquieR_resultset <- function(x, ...) {
  lifecycle::deprecate_stop("2.1.0", what = "dq_report()", with = "dq_report2()")
}

#' Compute Descriptive Statistics - continuous variables
#'
#' @description
#'
#' generates a descriptive overview of continuous variables (ratio and interval) in `resp_vars`.
#'
#' [Descriptor]
#'
#' @details
#' TODO
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable] the name of the continuous measurement variable
#' @param hard_limits_removal [logical] if TRUE values outside hard limits are
#'                                      removed from the data before calculating
#'                                      descriptive statistics.
#'                                      The default is FALSE
#' @param ... arguments to be passed to all called indicator functions if
#'            applicable.
#'
#' @return a [list] with:
#'   - `SummaryTable`: [data.frame]
#'   - `SummaryData`: [data.frame]
#' @export
#'
#' @examples
#' \dontrun{
#' prep_load_workbook_like_file("meta_data_v2")
#' xx <- des_summary_continuous(study_data = "study_data", meta_data =
#'                               prep_get_data_frame("item_level"))
#' xx$SummaryData
#' }
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_des_impl_summary.html
#' )
#' @importFrom graphics barplot hist
#' @importFrom stats mad
#'
des_summary_continuous <- function(resp_vars = NULL, # IDEA: group_vars = NULL, co_vars = NULL,
                                   study_data,
                                   label_col,
                                   item_level = "item_level",
                                   meta_data = item_level,
                                   meta_data_v2,
                                   hard_limits_removal =
                                     getOption("dataquieR.des_summary_hard_lim_remove",
                                               dataquieR.des_summary_hard_lim_remove_default),
                                   ...) {
  # TODO: add figures also to Excel Exports and prevent column type guessing
  # by datatables export correctly

  ##### Preparation -----
  # Metadata and study data set up ----
  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()
  util_expect_data_frame(study_data)
  try(util_expect_data_frame(meta_data), silent = TRUE)
  item_level <- meta_data <-
    util_amend_missing_metadata(study_data = study_data,
                                meta_data = meta_data,
                                level = VARATT_REQUIRE_LEVELS$RECOMMENDED)
  if (util_is_try_error(
    erobj <- try(prep_prepare_dataframes(.replace_hard_limits = FALSE,
                                         .replace_missings = FALSE,
                                         .amend_scale_level = TRUE,
                                         .apply_factor_metadata = FALSE,
                                         .apply_factor_metadata_inadm = FALSE
    ), silent = TRUE))) {
    w <- "%s has problems: %s, estimating %s"
    if (suppressWarnings(util_ensure_suggested("cli", err = FALSE))) {
      w <- cli::bg_black(cli::col_yellow(w))
    }
    util_warning(w, sQuote("meta_data"),
                 dQuote(conditionMessage(util_condition_from_try_error(erobj))),
                 sQuote("meta_data"),
                 immediate = TRUE)
    meta_data <-
      prep_study2meta(study_data, level = VARATT_REQUIRE_LEVELS$RECOMMENDED)
    prep_prepare_dataframes(.meta_data = meta_data,
                            .replace_hard_limits = FALSE,
                            .replace_missings = FALSE,
                            .amend_scale_level = TRUE,
                            .apply_factor_metadata = FALSE,
                            .apply_factor_metadata_inadm = FALSE
    )
  }

  # Define resp_vars----
  if (length(resp_vars) == 0) {
    #in case of missing resp_vars use all variables in the study data
    suppress <- function(...) suppressWarnings(suppressMessages(...))
    resp_vars <- colnames(ds1)
  } else {
    #in case of resp_vars present, check them
    suppress <- eval
  }

  #Check and remove resp_vars not present in the study data
  suppress(util_correct_variable_use(resp_vars,
                                     allow_more_than_one = TRUE,
                                     do_not_stop = TRUE,
                                     need_scale = "interval | ratio"))
  # resp_vars <- intersect(colnames(ds1), resp_vars)

  # Select only resp_vars that are continuous
  df_continuous <- meta_data[meta_data[, label_col, drop = TRUE] %in%
                               resp_vars, ]


  study_data_new <- ds1[, colnames(ds1) %in% resp_vars, drop = FALSE]

  colnames(study_data_new) <-
    util_map_labels(colnames(study_data_new),
                    meta_data = df_continuous,
                    to = VAR_NAMES,
                    from = label_col)

  # Create the descriptive statistics
  result1 <- des_summary(study_data = study_data_new,
                         item_level = df_continuous,
                         label_col = label_col)

  # Select only interested columns for SummaryData
  names_vector <- c("Variables",
                    "Type",
                    "STUDY_SEGMENT",
                    "Mean",
                    "SD",
                    "Median",
                    "Mode",
                    "IQR (Quartiles)",
                    "MAD",
                    "Range (Min - Max)",
                    "CV",
                    "Skewness (SE)",
                    "Kurtosis",
                    "Valid",
                    "Missing",
                    "Graph")

  summary_data <-
    result1$SummaryData[ ,
                         colnames(result1$SummaryData) %in% names_vector,
                         drop = FALSE]
  summary_table <-
    result1$SummaryTable

  summary_data[is.na(summary_data)] <- ""

  return(list(SummaryData = util_attach_attr(summary_data,
                                             is_html_escaped = TRUE),
              SummaryTable = util_attach_attr(summary_table,
                                              is_html_escaped = TRUE)))
}

#' make an html dashboard for a report
#'
#' @param report [dataquieR report v2][dq_report2]
#' @param make_links [logical] add links to variables
#' @param return_table_only [logical] if `TRUE` returns the table only,
#'                                    otherwise and `htmlwidget`
#'
#' @return a `htmltools` compatible dashboard of report results
#'         or `NULL`, if package is missing
#' @noRd
util_setup_dashboard <- function(report,
                                 make_links = FALSE,
                                 return_table_only = FALSE) {
  if (!util_ensure_suggested(c("htmltools"), err = FALSE)) {
    return(NULL)
  }
  if (!util_ensure_suggested(c("jsonlite"), err = FALSE)) {
    return(NULL)
  }
  # IDEA: Group by Variables, if wanted:
  # https://datatables.net/extensions/rowgroup/examples/initialisation/customRow.html
  title <- attr(report, "title")
  if (is.null(title)) {
    title <- "Data Quality Report"
  }

  table <- util_dashboard_table(summary(report)) #TODO: test with filter to un-existing functions.

  #
  # return early if table is empty anyway
  #
  if (is.null(table) || nrow(table) == 0 || ncol(table) == 0) {
      return()
  }

  table[["..VAR_NAMES"]] <- table[[VAR_NAMES]] # these are the original VAR_NAMES needed for mapping, maybe later.

  margin <- 2


  #
  # If SummaryPlot or SummaryPlotList are there and ggplot objects,
  # add them for each result
  #
  if (attr(report, "label_col") ==
      attr(table, "label_col")) {
    plot_names <- c("SummaryPlot", "SummaryPlotList")
    table$Figure <- mapply(
      SIMPLIFY = FALSE,
      var_label = table[[attr(table, "label_col")]],
      call_name = table$call_names,
      function(var_label, call_name) {
        fn_result_name <- paste0(call_name, ".", var_label)
        if (fn_result_name %in% names(report) &&
          any(plot_names %in% names(report[[fn_result_name]]))) {

          result_plot_name <- head(intersect(plot_names
                            , names(report[[fn_result_name]]))
                        , 2)
          if (util_is_gg(report[[fn_result_name]][[result_plot_name]])) {
            result <- htmltools::plotTag(
              {
                withr::local_par(list(
                  mar = rep(margin, 4),
                  oma = rep(0, 4)
                ))
                suppressWarnings(suppressMessages(
                  print(report[[fn_result_name]][[result_plot_name]])
                ))
              },
              width = 250,
              height = 200,
              alt = paste("Figure")
            )
            as.character(result)
          } else {
            NA_character_
          }
        } else {
          NA_character_
        }
      }
    )
  }

  #
  # if a row has a figure and it's value is "T" or "F", replace the value with the figure
  #
  logicals_with_figure = !is.na(table$Figure) & table$value %in% c("T", "F")
  table$value[logicals_with_figure] <-table$Figure[logicals_with_figure]

  #
  # get continuous and categorical summaries
  #
  add_summary_graphs <- function(table, summary_name, graph_name) {
    if (summary_name %in% colnames(report)) {
      des <- report[, summary_name, "SummaryTable"]$SummaryTable
      des[[attr(report, "label_col")]] <-
        des$Variables
      des[[graph_name]] <- des$Graph
      des$Graph <- NULL
      if (!is.null(dim(des)) && nrow(des) > 0) {
        table <- suppressWarnings(merge(table, des, all.x = TRUE,
                                     by = attr(report, "label_col"),
                                     suffixes = c("", "")))
        table <- util_fix_merge_dups(table)
      }
    }
    table
  }

  table <- add_summary_graphs(table, "des_summary_continuous", "GraphCon")
  table <- add_summary_graphs(table, "des_summary_categorical", "GraphCat")

  no_lb <- is.na(table$Variable_names) # happens for variables w/o any descriptive statistics (SCALE_LEVEL is "na"), since the column comes from des_summary()

  table$Variable_names[no_lb] <-
      vapply(lapply(apply(table[no_lb, intersect(unique(rev(c(
      VAR_NAMES, LABEL, attr(report, "label_col"), LONG_LABEL
    ))), colnames(table)), FALSE], 1, unique, simplify = FALSE), unique),
    paste0,
    collapse = "<br />",
    FUN.VALUE = character(1))

  # GraphCon
  # GraphCat
  # Variable Label
  # Variable Names
  # Figure

  if (!"GraphCat" %in% colnames(table)) {
    table$GraphCat <- NA
  }
  if (!"GraphCon" %in% colnames(table)) {
    table$GraphCon <- NA
  }

  #
  #  if both GraphCon and GraphCat are there, GraphCon is default and GraphCat is fallback.
  #  otherwise, take what is there
  #
  if ("des_summary_continuous" %in% colnames(report) &&
      "des_summary_categorical" %in% colnames(report)  ) {
    table$Graph <- table$GraphCon
    table$GraphCon <- NULL
    table$Graph[is.na(table$Graph)] <- table$GraphCat[is.na(table$Graph)]
    table$GraphCat <- NULL
  } else if ("des_summary_continuous" %in% colnames(report)) {
    table$Graph <- table$GraphCon
    table$GraphCon <- NULL
  } else if ("des_summary_categorical" %in% colnames(report)) {
    table$Graph <- table$GraphCat
    table$GraphCat <- NULL
  }

  # put Graph to the left if it is there
  table <- table[, intersect(unique(c("Graph", colnames(table))), colnames(table)), FALSE]

  if (make_links) {
    table0 = table
    #
    # for each of label_col, "Graph" and VAR_NAMES that is in colnames, make a link from the content
    #
    for (cn in intersect(colnames(table), c(
      attr(report, "label_col"),
      "Graph",
      VAR_NAMES
    ))) {
      cnt <- table0[[cn]]
      if (any(.no_cnt <- is.na(cnt))) {
        cnt[.no_cnt] <- "&nbsp;"
      }
      lb <- table0[[attr(report, "label_col")]]
      cl <- table0[["call_names"]]
      href <- table[["href"]]
      popup_href <- table[["popup_href"]]
      title <- table[["title"]]
      table[[cn]] <- mapply(SIMPLIFY = FALSE,
                         cnt = cnt,
                         lb = lb,
                         cl = cl,
                         href = href,
                         popup_href = popup_href,
                         title = title,
                         function(cnt, lb, cl, href, popup_href, title) {
                          # if (is.na(cnt)) browser()
                           as.character(htmltools::a(htmltools::HTML(cnt),
                                                     href = href,
                                                     onclick = htmltools::htmlTemplate(text_ = "(function(e) {
                                                        e.preventDefault();
                                                        showDataquieRResult({{url}}, {{link_url}}, {{title}});
                                                      })(event);",
                                                      url =
                                                        jsonlite::toJSON(popup_href, auto_unbox = TRUE),
                                                      link_url = jsonlite::toJSON(href, auto_unbox = TRUE),
                                                      title = jsonlite::toJSON(title, auto_unbox = TRUE)
                                                      )))
                         })
    }
    table$href <- table$popup_href <- table$title <- NULL
    rm("table0")
  }

  # get indices of colnames and subtract 1 so they make sense for javascript
  var_class_col_js_idx <- which(colnames(table) == "var_class") - 1
  Class_col_js_idx <- which(colnames(table) == "Class") - 1
  class_raw_col_js_idx <- which(colnames(table) == "class") - 1
  name_col_js_idx <- which(colnames(table) == attr(report, "label_col")) - 1

  table$Class <- table$class # we can handle factors, now

  # translate all colnames
  util_translated_colnames(table) <- util_translate(colnames(table), ns = "dashboard_table")

  #
  # unlist all columns that are lists
  #
  list_columns <- vapply(table, is.list, FUN.VALUE = logical(1))

  if (any(list_columns)) {
    table[list_columns] <-
      lapply(table[list_columns], unlist )
  }

  attr(table, "indexes") <- list(
    var_class_col_js_idx = var_class_col_js_idx,
    Class_col_js_idx = Class_col_js_idx,
    class_raw_col_js_idx = class_raw_col_js_idx,
    name_col_js_idx = name_col_js_idx
  )

  #
  if (return_table_only) {
    return(table)
  }

  my_dashboard <- util_dashboard_table2widget(table, attr(report, "label_col"))

  my_dashboard
}

util_dashboard_table2widget <- function(table, label_col) {

  indexes <- attr(table, "indexes")
  var_class_col_js_idx <- indexes$var_class_col_js_idx
  Class_col_js_idx <- indexes$Class_col_js_idx
  class_raw_col_js_idx <- indexes$class_raw_col_js_idx
  name_col_js_idx <- indexes$name_col_js_idx

  #
  # make actual dashboard from the data
  #
  my_dashboard <-
    util_html_table(
      table,
      descs = setNames(rep("", ncol(table)),
                       colnames(table)),
      searchBuilder = TRUE, # -- var_class not ok and var_class no unclear
      col_tags = list(
        init = as.character(util_translate(unique(c(label_col,
                                       VAR_NAMES,
                                       "Metric",
                                       "value",
                                       "Graph",
                                       "Class",
                                       "Call",
                                       "var_class")), as_this_translation = colnames(table))),
        all = colnames(table)
      ),
      initial_col_tag = "init",

      # '{"criteria":[{"condition":"!=","data":"var_class","type":"string","value":["Ok"]},{"condition":"!null","data":"Class","type":"string","value":[]}],"logic":"AND"}'
      # xx <- jsonlite::parse_json(...)
      # cat(deparse(xx), sep = "\n")
      #
      # initial search criteria: var_class != grading_class[[1]] && Class != Null
      #
      init_search =
        list(criteria = list(list(condition = "!=",
                                  data = as.character(util_translate("var_class", as_this_translation = colnames(table))),
                                  type = "string",
                                  value = list(util_get_labels_grading_class()[["1"]])),
                             list(condition = "!null",
                                  data = as.character(util_translate("Class", as_this_translation = colnames(table))),
                                  type = "string", value = list())), logic = "AND"),
      #
      # arguments that are passed directly to javascript
      #
      additional_init_args = list(
        grading_cols = as.character(util_translate(unique(c("Class",
                                               "class",
                                               "var_class")), as_this_translation = colnames(table))),
        secondary_order = setNames(list(
          c(var_class_col_js_idx, name_col_js_idx, class_raw_col_js_idx)
        ), nm = as.character(util_translate(unique(c("var_class")), as_this_translation = colnames(table)))),
        grading_order = util_get_labels_grading_class(),
        grading_colors = util_get_colors(),
        fg_colors = util_get_fg_color(util_get_colors())
      ) , # , this is done in js in report_dt.js:sort_vert_dt() -- if (type == 'sort')
      # additional_columnDefs = list(list(
      #   orderData = c(var_class_col_js_idx, class_raw_col_js_idx),
      #   targets = Class_col_js_idx
      # ))
      additional_columnDefs = list(list(
        className = "dt-right",
        targets = which(colnames(table) %in% c("Value"), # column value is a mixed column, but it should nevertheless be right-aligned
                                                         ) - 1
      ))
    )
  htmltools::tagList(html_dependency_tippy(),
                     html_dependency_clipboard(),
                     my_dashboard)
}

#' Produce a condition message with a useful short stack trace.
#'
#' @inheritParams util_error
#'
#' @param m a message or a [condition]
#' @param ... arguments for [sprintf] on m, if m is a character
#' @param integrity_indicator [character] the message is an integrity problem,
#'                                        here is the indicator abbreviation..
#' @param level [integer] level of the message (defaults to 0). Higher
#'                        levels are more severe.
#' @param once_id [character] If specified, suppress further conditions
#'                 with the same ID during this R session.
#' @param immediate [logical] not used.
#'
#' @return [condition] the condition object, if the execution is not stopped
#'
#' @family condition_functions
#' @concept process
#' @noRd
util_message <- util_condition_constructor_factory("message")

#' Extract Parts of a `dataquieR` Result Object
#'
#' @param x the `dataquieR` result object
#'
#' @param ... arguments passed to the implementation for lists.
#'
#' @return the sub-list of the `dataquieR` result object with all messages
#'         still attached
#'
#' @seealso  [base::Extract]
#' @export
#'
#' @noRd
`[.dataquieR_result` <- function(x, ...) {
  r <- NextMethod()
  attr(r, "error") <- attr(x, "error")
  attr(r, "message") <- attr(x, "message")
  attr(r, "warning") <- attr(x, "warning")
  attr(r, "as_plotly") <- attr(x, "as_plotly")
  attr(r, "dont_util_adjust_geom_text_for_plotly") <- attr(x, "dont_util_adjust_geom_text_for_plotly")
  attr(r, "function_name") <- attr(x, "function_name")
  attr(r, "cn") <- attr(x, "cn")
  attr(r, "call") <- attr(x, "call")
  class(r) <- unique(c("dataquieR_result", class(r)))
  r
}

#' Extract Elements of a `dataquieR` Result Object
#'
#' @param x the `dataquieR` result object
#'
#' @param ... arguments passed to the implementation for lists.
#'
#' @return the element of the `dataquieR` result object with all messages
#'         still attached
#'
#' @seealso  [base::Extract]
#' @export
#'
#' @noRd
`[[.dataquieR_result` <- function(x, ...) {
  r <- NextMethod()
  if (is.null(r)) {
    r <- list()
    class(r) <- union("dataquieR_NULL", class(r))
  }
  if (!util_is_gg(x)) {
    attr(r, "error") <- attr(x, "error")
    attr(r, "message") <- attr(x, "message")
    attr(r, "warning") <- attr(x, "warning")
    # do not assign this class, here:
    # class(r) <- unique(c("dataquieR_result", class(r)))
    class(r) <- unique(c("Slot", class(r)))
  }
  r
}

#' Extract elements of a `dataquieR` Result Object
#'
#' @param x the `dataquieR` result object
#'
#' @param ... arguments passed to the implementation for lists.
#'
#' @return the element of the `dataquieR` result object with all messages
#'         still attached
#'
#' @seealso  [base::Extract]
#' @export
#'
#' @noRd
`$.dataquieR_result` <- `[[.dataquieR_result`

#' Creates a Link to our Website
#'
#' i.e., to a vignette on the website
#'
#' @param fkt_name [character] function name to generate a link for
#'
#' @return [character] the link
#'
#' @noRd
util_online_ref <- function(fkt_name) {

  # convert function names to website names
  fkt_name <- strsplit(fkt_name, "_", fixed = TRUE)[[1]]
  fkt_name_impl <-  c("VIN", head(fkt_name, 1), "impl", tail(fkt_name, -1))
  fkt_name <- c("VIN", fkt_name)
  fkt_name <- paste(fkt_name, collapse = "_")
  fkt_name_impl <- paste(fkt_name_impl, collapse = "_")

  fkt_name <-
    sprintf("https://dataquality.qihs.uni-greifswald.de/%s.html", fkt_name)

  fkt_name_impl <-
    sprintf("https://dataquality.qihs.uni-greifswald.de/%s.html", fkt_name_impl)


  if (identical(
    try(attr(
      base::curlGetHeaders(fkt_name), "status"), silent = TRUE), 200L)) {
    return(fkt_name)
  }

  return(fkt_name_impl)
}

#' Copy default dependencies to the report's lib directory
#'
#' @param dir report directory
#' @param pages all pages to write
#' @param ... additional `htmltools::htmlDependency` objects to be added to all
#'            pages, also
#'
#' @return pre-processed dependencies -- a list with `deps` and rendered `pages`
#'
#' @family reporting_functions
#' @concept process
#' @noRd
util_copy_all_deps <- function(dir, pages, ...) {
# FIXME: Crashes for no dependencies
  libdir <- file.path(dir, "lib")

  withCallingHandlers({
    rendered_pages <- lapply(pages, htmltools::renderTags)
  },
  warning = function(cond) { # suppress a waning caused by ggplotly for barplots
    if (startsWith(conditionMessage(cond),
                   "'bar' objects don't have these attributes: 'mode'") ||
        startsWith(conditionMessage(cond),
                   "'box' objects don't have these attributes: 'mode'")) {
      invokeRestart("muffleWarning")
    }
  })

  deps <- c(lapply(rendered_pages, `[[`, "dependencies"),
            GLOBAL_ = list(list(...)))

  names(deps)[util_empty(names(deps))] <- paste0("__NULL__",
                                                         seq_len(sum(util_empty(
                                                           names(deps)))),
                                                         "__NULL__")

  deps_cnt <- lapply(deps, length)

  pos_in_file <- lapply(deps_cnt, seq_len)

  deps <- unlist(deps, recursive = FALSE)

  deps_info <- data.frame(index = seq_len(length(deps)),
                          file_name =
                            unlist(lapply(names(deps_cnt), function(nm) rep(nm, deps_cnt[[nm]]))),
                          pos_in_file = unlist(pos_in_file))
  deps_info$version <- vapply(deps, `[[`, "version", FUN.VALUE = character(1))
  deps_info$name <- vapply(deps, `[[`, "name", FUN.VALUE = character(1))
  deps_info$take <- rep(FALSE, nrow(deps_info))

  # Omit older version, if libraries are duplicated
  deps_info <- split(deps_info, deps_info$name)

  deps_info <- lapply(deps_info, function(x) {
    r <- x[order(numeric_version(x$version),
                 decreasing = TRUE,
                 na.last = TRUE), , FALSE]
    r[1, "take"] <- TRUE
    r
  })

  deps_info <- do.call(rbind, deps_info)

  # overall order of dependencies

  n_deps <- length(unique(deps_info$name))

  # find dep, which is either not in a file or at first position

  order_of_deps <- unique(unname(unlist(lapply(setNames(nm = seq_len(n_deps)), function(pos) {
    lapply(setNames(nm = unique(deps_info$name)), function(nm) {
      r <- deps_info[deps_info$name == nm & deps_info$pos_in_file > pos, ,
                     FALSE]
      if (nrow(r)) {
        NULL
      } else {
        nm
      }
    })
  }))))


  index <- deps_info[deps_info$take, "index", TRUE]

  deps <- deps[index]

  names(deps) <- vapply(deps, `[[`, "name", FUN.VALUE = character(1))

  deps <- deps[order_of_deps]

  deps <-
    lapply(deps,
           htmltools::copyDependencyToDir,
           outputDir = libdir,
           mustWork = TRUE) # no external http-dependencies allowed (mustWork)

  deps <-
    lapply(deps,
           htmltools::makeDependencyRelative,
           basepath = dir,
           mustWork = TRUE) # no external http-dependencies allowed (mustWork)

  list(deps = htmltools::renderDependencies(deps, "file"), # TODO: What, if two versions of the same lib are used (in acc_acc_distributions_observer, we have an old jQuery injected by plotly)
       rendered_pages = rendered_pages)
}

#' Utility function to distribute points across a time variable
#'
#' @param time_var_data vector of the data points of the time variable
#' @param n_points      maximum number of points to distribute across the
#'                      time variable (minimum: 3)
#' @param prop_grid     proportion of points given in `n_points` that should be
#'                      distributed in an equally spaced grid across the time
#'                      variable (minimum: 0.1, maximum: 1). The remaining
#'                      proportion of points will be spaced according to the
#'                      distribution of the time variable's data points.
#'
#' @return a sequence of points in datetime format
#' @noRd
util_optimize_sequence_across_time_var <- function(time_var_data,
                                                   n_points,
                                                   prop_grid = 0.5) { # TODO EK: Please adjust for time-only
  util_expect_scalar(n_points,
                     check_type = util_is_numeric_in(min = 3,
                                                     whole_num = TRUE,
                                                     finite = TRUE))
  util_expect_scalar(prop_grid,
                     check_type = util_is_numeric_in(min = 0.1, max = 1))

  time_var_data <- na.omit(time_var_data)
  tp_seq <- unique(time_var_data)
  tp_seq <- sort(tp_seq)

  if (length(tp_seq) < 2) {
    util_error("Internal error in sequence optimization.") # should not occur
  }

  if (n_points < length(tp_seq)) { # reduce time points
    # distribute some points in a fixed grid
    n_grid <- max(floor(prop_grid * n_points), 3)
    # at least three points in a grid
    # compute number of seconds between two consecutive time points
    period <- (max(tp_seq) - min(tp_seq)) / (n_grid - 1)
    secs <- suppressWarnings(as.integer(as.double(period, units = "secs")))
    if (secs <= 0) {
      secs <- 1
    }
    # This defines a new sequence of equally spaced time points ranging from
    # the earliest observed time point to the last one.
    # This sequence is stored in `tp_round_seq`. The number of time points
    # equals the number given in `n_equ_spaced`.
    tp_round_seq <- suppressWarnings(util_parse_date(
      seq(from = min(tp_seq),
          to = max(tp_seq),
          by = secs)))
    tp_round_seq[length(tp_round_seq)] <- max(tp_seq) # can otherwise deviate
    # due to rounding/numerical errors
    # distribute remaining points (if any) according to the distribution of
    # the data
    if (n_points - n_grid > 0) {
      tp_tab <- util_table_of_vct(
        cut(time_var_data,
            breaks = c(tp_round_seq[1:(length(tp_round_seq) - 1)],
                       tp_round_seq[length(tp_round_seq)] + 1)))
      tp_tab$weight <- tp_tab$Freq/sum(tp_tab$Freq)
      tp_tab$n_add <- floor(tp_tab$weight * (n_points - n_grid))
      tp_round_seq <- c(do.call(c,
        lapply(seq_along(tp_round_seq[-1]), function(tp_i) {
          if (tp_tab$n_add[tp_i] > 0) {
            period <- (tp_round_seq[tp_i + 1] - tp_round_seq[tp_i]) /
              (tp_tab$n_add[tp_i] + 1)
            secs <- suppressWarnings(as.integer(
              as.double(period, units = "secs")))
            if (secs <= 0) {
              secs <- 1
            }
            suppressWarnings(util_parse_date(
              tp_round_seq[tp_i] + c(0, seq_len(tp_tab$n_add[tp_i])) * secs))
          } else {
            tp_round_seq[tp_i]
          }
        })
      ), max(tp_seq))
    }
  } else {
    tp_round_seq <- tp_seq
  }

  return(tp_round_seq)
}

#' Detects variable levels not specified in metadata
#'
#' @description
#' For each categorical variable, value lists should be defined in the metadata.
#'  This implementation will examine, if all observed levels in the study data
#'  are valid.
#'
#' [Indicator]
#'
#' @details
#' ### Algorithm of this implementation:
#'
#'  - Remove missing codes from the study data (if defined in the metadata)
#'  - Interpretation of variable specific VALUE_LABELS as supplied in the
#'    metadata.
#'  - Identification of measurements not corresponding to the expected
#'    categories. Therefore two output data frames are generated:
#'    - on the level of observation to flag each undefined category, and
#'    - a summary table for each variable.
#'  - Values not corresponding to defined categories are removed in a data frame
#'    of modified study data
#'
#' @export
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable list] the name of the measurement variables
#' @param threshold_value [numeric] from=0 to=100. a numerical value ranging
#'                                           from 0-100.
#'
#' @return a list with:
#' - `SummaryData`: data frame summarizing inadmissible categories with the
#' columns:
#'   - `Variables`: variable name/label
#'   - `OBSERVED_CATEGORIES`: the categories observed in the study data
#'   - `DEFINED_CATEGORIES`: the categories defined in the metadata
#'   - `NON_MATCHING`: the categories observed but not defined
#'   - `NON_MATCHING_N`: the number of observations with categories not defined
#'   - `NON_MATCHING_N_PER_CATEGORY`: the number of observations for each of the
#'                unexpected categories
#' - `SummaryTable`: data frame for the `dataquieR` pipeline reporting the number
#'                   and percentage of inadmissible categorical values
#' - `ModifiedStudyData`: study data having inadmissible categories removed
#' - `FlaggedStudyData`: study data having cases with inadmissible categories
#' flagged
#'
#' @importFrom utils capture.output
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_inadmissible_categorical.html
#' )
con_inadmissible_categorical <- function(resp_vars = NULL,
                                         study_data,
                                         label_col,
                                         item_level = "item_level",
                                         threshold_value = 0,
                                         meta_data = item_level,
                                         meta_data_v2) {

  # preps ----------------------------------------------------------------------
  util_maybe_load_meta_data_v2()

  # detect, if called as "con_inadmissible_vocabulary"
  voc_mode <-
    rlang::call_name(rlang::caller_call(4)) == "con_inadmissible_vocabulary"


  as_numeric_no_warn <- function(...) {
    suppressWarnings(as.numeric(...))
  }

  # map metadata to study data
  prep_prepare_dataframes()

  util_stop_if_not(
`Internal error, sorry, please report. VALUE_LABELS after prep_prepare_dataframes` =
       is.null(meta_data[[VALUE_LABELS]]))

  if (voc_mode) {
    meta_data[[VALUE_LABEL_TABLE]] <- NULL
    meta_data[[VALUE_LABEL_TABLE]] <- meta_data[[STANDARDIZED_VOCABULARY_TABLE]]
  }


  if (!VALUE_LABEL_TABLE %in% colnames(meta_data)) {
    if (voc_mode) {
      util_error(paste0("Function con_inadmissible_vocabulary requires",
                        "the metadata column STANDARDIZED_VOCABULARY_TABLE."),
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)

    } else {
      util_error(paste0("Function con_inadmissible_categorical requires",
                        "the metadata column VALUE_LABELS or VALUE_LABEL_TABLE."),
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)

    }
  }

  util_correct_variable_use("resp_vars",
                            allow_more_than_one = TRUE,
                            allow_null = TRUE,
                            allow_any_obs_na = TRUE,
                            need_type = "integer | string", # TODO: What about datetime as categorical variable, if there are, e.g., only three timepoints?
                            need_scale = "nominal | ordinal"
  )

  if (length(resp_vars) == 0) {
    # Select all variables with VALUE_LABEL_TABLE (if any), if resp_vars were not specified.
    if (all(util_empty(meta_data[[VALUE_LABEL_TABLE]]))) {
      if (voc_mode) {
        util_error(c("No variables with defined",
                     "STANDARDIZED_VOCABULARY_TABLE"),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      } else {
        util_error(c("No variables with defined",
                     "VALUE_LABELS or VALUE_LABEL_TABLE."),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      }
    } else {
      if (voc_mode) {
        util_message(c("All variables with STANDARDIZED_VOCABULARY_TABLE",
                       "in the metadata are used,",
                       "if SCALE_LEVEL is nominal or ordinal."),
                     applicability_problem = TRUE,
                     intrinsic_applicability_problem = TRUE)
      } else {
        util_message(c("All variables with VALUE_LABELS or",
                       "VALUE_LABEL_TABLE in the metadata are used,",
                       "if SCALE_LEVEL is nominal or ordinal."),
                     applicability_problem = TRUE,
                     intrinsic_applicability_problem = TRUE)
      }
      resp_vars <-
        intersect(meta_data[[label_col]][!(
          util_empty(meta_data[[VALUE_LABEL_TABLE]])) &
            meta_data[[SCALE_LEVEL]] %in% c(
              SCALE_LEVELS$NOMINAL, SCALE_LEVELS$ORDINAL)],
                  colnames(ds1))
    }
  } else {
    # Check for specified resp_vars whether VALUE_LABEL_TABLE have been defined at all.
    if (all(util_empty(
      meta_data[[VALUE_LABEL_TABLE]][meta_data[[label_col]] %in% resp_vars]))) {
      if (voc_mode) {
        util_error("No variables with defined STANDARDIZED_VOCABULARY_TABLE.",
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      } else {
        util_error("No variables with defined VALUE_LABELS or VALUE_LABEL_TABLE.",
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      }
    }
    # check whether VALUE_LABEL_TABLE are missing for some of the specified resp_vars
    rvs <- meta_data[[label_col]][!(
      util_empty(meta_data[[VALUE_LABEL_TABLE]])) &
                                    meta_data[[label_col]] %in% resp_vars]
    if (length(rvs) < length(resp_vars)) {
      if (voc_mode) {
        util_message(paste0("The variables ", resp_vars[!(resp_vars %in% rvs)],
                          " have no defined",
                          "STANDARDIZED_VOCABULARY_TABLE."),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      } else {
        util_message(paste0("The variables ", resp_vars[!(resp_vars %in% rvs)],
                            " have no defined",
                            "VALUE_LABELS or VALUE_LABEL_TABLE."),
                     applicability_problem = TRUE,
                     intrinsic_applicability_problem = TRUE)
      }
    }
    resp_vars <- rvs
  }

  util_correct_variable_use("resp_vars",
                            allow_more_than_one = TRUE,
                            allow_null = TRUE,
                            allow_any_obs_na = TRUE,
                            need_type = "integer | string", # TODO: What about datetime as categorical variable, if there are, e.g., only three timepoints?
                            need_scale = "nominal | ordinal"
  )

  if (length(resp_vars) == 0) {
    util_error("No categorical variables with value lists / vocabulary found.",
               applicability_problem = TRUE)
  }

  util_expect_scalar(threshold_value,
                     check_type = util_is_numeric_in(min = 0, max = 100))

  #util_expect_scalar(use_value_labels, check_type = is.logical)

  # Calculations ---------------------------------------------------------------

  sumdf1 <- data.frame(Variables = resp_vars)

  # Which categories were used in the study data?
  get_obs_cats <- function(x) {
    as.character(sort(unique(x[!is.na(x)])))
  }
  sumdf1$OBSERVED_CATEGORIES <- lapply(ds1[, resp_vars, drop = FALSE],
                                       get_obs_cats)

  voc_names <- setNames(
    util_map_labels(
      resp_vars,
      meta_data = meta_data,
      to = VALUE_LABEL_TABLE,
      from = label_col,
      ifnotfound = NA
    ),
    nm = resp_vars
  )

  # Which categories were defined in metadata?
  def_cats <- lapply(voc_names, function(vlt) {
        vl <- NA_character_
        if (!is.na(vlt)) {
          codes <- try(prep_get_data_frame(vlt), silent = TRUE)
          if (util_is_try_error(codes)) {
            util_warning(codes,
                         title = "Could not fetch value_labels/vocabulary %s, ignoring:\n",
                         dQuote(vlt))
          } else {
            # TODO: Handle | in [[1]] or [[2]]
            if (voc_mode && ncol(codes) == 1) {
              vl <- codes
              colnames(vl) <- CODE_VALUE
            } else if (voc_mode && ncol(codes) == 2) {
              vl <- codes
              colnames(vl) <- c(CODE_VALUE, CODE_LABEL)
            } else if (voc_mode) {
              url <-
                subset(util_get_voc_tab(), paste0("<", get("voc"), ">") == vlt,
                       "url", drop = TRUE)
              util_warning(c("For the standardized vocabulary table %s,",
                             "there are %d columns from the table %s defined.",
                             "There should be 1 or 2 (2nd one would be the",
                             "labels).",
                             "Ignoring this table."),
                           dQuote(vlt),
                           ncol(codes),
                           dQuote(url))
            } else {
              vl <- codes
            }
          }
        }

    if (is.data.frame(vl)) {
      if (CODE_VALUE %in% colnames(vl)) {
        if (CODE_LABEL %in% colnames(vl)) {
          return(setNames(as.character(vl[[CODE_VALUE]]), nm = vl[[CODE_LABEL]]))
        } else {
          return(as.character(vl[[CODE_VALUE]]))
        }
      } else {
        return(NA_character_)
      }
    } else {
      return(NA_character_)
    }
  })

  sumdf1$DEFINED_CATEGORIES <- def_cats


  # Which categories were used in the study data but not defined in the metadata?
  which_not <- lapply(
    seq_along(sumdf1$OBSERVED_CATEGORIES),
    function(x) !(sumdf1$OBSERVED_CATEGORIES[[x]] %in%
                    sumdf1$DEFINED_CATEGORIES[[x]])
  )
  sumdf1$NON_MATCHING <- lapply(
    seq_along(sumdf1$OBSERVED_CATEGORIES),
    function(x) paste0(sumdf1$OBSERVED_CATEGORIES[[x]][which_not[[x]]])
  )
  # If there are no unexpected categories, the entry should be NA (instead of "character(0)").
  sumdf1$NON_MATCHING <- lapply(
    sumdf1$NON_MATCHING,
    function(x) if (identical(x, character(0))) NA_character_ else x
  )

  # Count the number of non-matching observations in total and per unexpected category.
  sumdf1$NON_MATCHING_N <- NA
  sumdf1$NON_MATCHING_N_PER_CATEGORY <- NA
  # msdf = modified study data
  msdf <- ds1

  for (i in seq_len(nrow(sumdf1))) {
    if (!is.na(sumdf1$NON_MATCHING[i])) {
      if (all(!is.na(as_numeric_no_warn(unlist(sumdf1$NON_MATCHING[i]))))) {
        sumdf1$NON_MATCHING_N[i] <- sum(ds1[[paste(sumdf1$Variables[i])]] %in%
                                          as_numeric_no_warn(unlist(
                                            sumdf1$NON_MATCHING[i])),
          na.rm = TRUE
        )
        n_per_cat <- table(ds1[[sumdf1$Variables[i]]][which(
          ds1[[sumdf1$Variables[i]]] %in%
            as_numeric_no_warn(unlist(sumdf1$NON_MATCHING[i])))])
        n_per_cat <- n_per_cat[order(n_per_cat, decreasing = TRUE)]
        sumdf1$NON_MATCHING_N_PER_CATEGORY[i] <- paste(
          paste0(n_per_cat, " (", dQuote(names(n_per_cat)), ")"),
          collapse = ", ")
        msdf[[paste(sumdf1$Variables[i])]][
          msdf[[paste(sumdf1$Variables[i])]] %in%
            as_numeric_no_warn(unlist(sumdf1$NON_MATCHING[i]))] <- NA
        ds1[[paste0(sumdf1$Variables[i], "_IAV")]] <-
          ifelse(ds1[[paste(sumdf1$Variables[i])]] %in%
                   as_numeric_no_warn(unlist(sumdf1$NON_MATCHING[i])), 1, 0)
      } else {
        sumdf1$NON_MATCHING_N[i] <- sum(ds1[[paste(sumdf1$Variables[i])]] %in%
                                          unlist(sumdf1$NON_MATCHING[i]),
          na.rm = TRUE
        )
        n_per_cat <- table(ds1[[sumdf1$Variables[i]]][which(
          ds1[[sumdf1$Variables[i]]] %in% unlist(sumdf1$NON_MATCHING[i]))])
        sumdf1$NON_MATCHING_N_PER_CATEGORY[i] <- paste(
          paste0(n_per_cat, " (", dQuote(names(n_per_cat)), ")"),
          collapse = ", ")
        msdf[[paste(sumdf1$Variables[i])]][
          msdf[[paste(sumdf1$Variables[i])]] %in%
            as_numeric_no_warn(unlist(sumdf1$NON_MATCHING[i]))] <- NA
        ds1[[paste0(sumdf1$Variables[i], "_IAV")]] <-
          ifelse(ds1[[paste(sumdf1$Variables[i])]] %in%
                   as_numeric_no_warn(unlist(sumdf1$NON_MATCHING[i])), 1, 0)
      }
    }
  }

  # any IAVs?
  checkIAV <- names(ds1)[grep("_IAV", names(ds1))]
  if (length(checkIAV) > 0) {
    if (!.called_in_pipeline) util_message(paste0("The following variable(s): ",
                        paste0(checkIAV, collapse = ", "),
                        " flag(s) inadmissible values."),
                 applicability_problem = FALSE)
  }

  # attribute
  attr(msdf, "rmIAVcat") <- TRUE

  # if (use_value_labels) {
  #   sumdf1$DEFINED_CATEGORIES <- lapply(sumdf1[["DEFINED_CATEGORIES"]], function(x) {
  #     if (all(!is.na(as_numeric_no_warn(x)))) {
  #       x <- paste0(x, " (", dQuote(names(x)), ")")
  #     }
  #     x
  #   })
  # }

  # handle lists
  sumdf1[["OBSERVED_CATEGORIES"]] <-
    as.vector(unlist(lapply(sumdf1[["OBSERVED_CATEGORIES"]],
                            paste, collapse = ", ")))

  if (voc_mode) {
    sumdf1[["DEFINED_CATEGORIES"]] <- voc_names[sumdf1[["Variables"]]]
  } else {
    sumdf1[["DEFINED_CATEGORIES"]] <-
      as.vector(unlist(lapply(sumdf1[["DEFINED_CATEGORIES"]],
                              paste, collapse = ", ")))
  }

  sumdf1[["NON_MATCHING"]] <-
    as.vector(unlist(lapply(sumdf1[["NON_MATCHING"]],
                            paste, collapse = ", ")))
  sumdf1$NON_MATCHING <-
    ifelse(sumdf1$NON_MATCHING == "NA", "", sumdf1$NON_MATCHING)
  sumdf1$NON_MATCHING_N <-
    ifelse(is.na(sumdf1$NON_MATCHING_N), 0, sumdf1$NON_MATCHING_N)
  sumdf1$NON_MATCHING_N_PER_CATEGORY <-
    ifelse(is.na(sumdf1$NON_MATCHING_N_PER_CATEGORY), 0,
           sumdf1$NON_MATCHING_N_PER_CATEGORY)
  sumdf1[["GRADING"]] <-
    ifelse(sumdf1$NON_MATCHING_N > threshold_value / 100 * nrow(ds1), 1, 0)

  sumdf2 <- sumdf1[, c("Variables", "NON_MATCHING_N")]

  if (voc_mode) {
    colnames(sumdf2)[2] <- "NUM_con_rvv_icat" # FIXME: use correct indicator metrics
  } else {
    colnames(sumdf2)[2] <- "NUM_con_rvv_icat"
  }
  sumdf2$PCT_con_rvv_icat <- round(sumdf2$NUM_con_rvv_icat / nrow(ds1) * 100, 1)
  sumdf2$GRADING <- sumdf1$GRADING
  sumdf2$FLG_con_rvv_icat <- ifelse(sumdf2$GRADING == 1, TRUE, FALSE)


  #remove GRADING from sumdf1 that is going to be the SummaryData
  sumdf1$GRADING <- NULL

  # to add descriptions in the hover text of the headers of the table
  text_to_display <- util_get_hovertext("[con_inadmissible_categorical_hover]")

  attr(sumdf1, "description") <- text_to_display

  #Add new attribute to the columns of SummaryData to define the datatype of each column
  attr(sumdf1$Variables, DATA_TYPE) <- DATA_TYPES$STRING
  attr(sumdf1$OBSERVED_CATEGORIES, DATA_TYPE) <- DATA_TYPES$STRING
  attr(sumdf1$DEFINED_CATEGORIES, DATA_TYPE) <- DATA_TYPES$STRING
  attr(sumdf1$NON_MATCHING, DATA_TYPE) <- DATA_TYPES$STRING
  attr(sumdf1$NON_MATCHING_N, DATA_TYPE) <- DATA_TYPES$INTEGER
  attr(sumdf1$NON_MATCHING_N_PER_CATEGORY, DATA_TYPE) <- DATA_TYPES$STRING


   # TODO EK: Use util_pretty_vector_string for SummaryData to avoid very long strings
  # Examples:
  # util_pretty_vector_string(letters, n_max = 30)
  # util_pretty_vector_string(letters, n_max = 3)

  return(list(SummaryData = sumdf1,
              SummaryTable = sumdf2,
              ModifiedStudyData = msdf,
              FlaggedStudyData = ds1))
}

#' Detects variable levels not specified in standardized vocabulary
#'
#' @description
#' For each categorical variable, value lists should be defined in the metadata.
#'  This implementation will examine, if all observed levels in the study data
#'  are valid.
#'
#' [Indicator]
#'
#' @details
#' ### Algorithm of this implementation:
#'
#'  - Remove missing codes from the study data (if defined in the metadata)
#'  - Interpretation of variable specific VALUE_LABELS as supplied in the
#'    metadata.
#'  - Identification of measurements not corresponding to the expected
#'    categories. Therefore two output data frames are generated:
#'    - on the level of observation to flag each undefined category, and
#'    - a summary table for each variable.
#'  - Values not corresponding to defined categories are removed in a data frame
#'    of modified study data
#'
#' @export
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable list] the name of the measurement variables
#' @param threshold_value [numeric] from=0 to=100. a numerical value ranging
#'                                           from 0-100.
#'
#' @return a list with:
#' - `SummaryData`: data frame summarizing inadmissible categories with the
#' columns:
#'   - `Variables`: variable name/label
#'   - `OBSERVED_CATEGORIES`: the categories observed in the study data
#'   - `DEFINED_CATEGORIES`: the categories defined in the metadata
#'   - `NON_MATCHING`: the categories observed but not defined
#'   - `NON_MATCHING_N`: the number of observations with categories not defined
#'   - `NON_MATCHING_N_PER_CATEGORY`: the number of observations for each of the
#'                unexpected categories
#'   - `GRADING`: indicator TRUE/FALSE if inadmissible categorical values were
#'     observed (more than indicated by the `threshold_value`)
#' - `SummaryTable`: data frame for the `dataquieR` pipeline reporting the number
#'                   and percentage of inadmissible categorical values
#' - `ModifiedStudyData`: study data having inadmissible categories removed
#' - `FlaggedStudyData`: study data having cases with inadmissible categories
#' flagged
#'
#' @importFrom utils capture.output
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_inadmissible_categorical.html
#' )
#' @examples
#' \dontrun{
#' sdt <- data.frame(DIAG = c("B050", "B051", "B052", "B999"),
#'                   MED0 = c("S01XA28", "N07XX18", "ABC", NA), stringsAsFactors = FALSE)
#' mdt <- tibble::tribble(
#' ~ VAR_NAMES, ~ DATA_TYPE, ~ STANDARDIZED_VOCABULARY_TABLE, ~ SCALE_LEVEL, ~ LABEL,
#' "DIAG", "string", "<ICD10>", "nominal", "Diagnosis",
#' "MED0", "string", "<ATC>", "nominal", "Medication"
#' )
#' con_inadmissible_vocabulary(NULL, sdt, mdt, label_col = LABEL)
#' prep_load_workbook_like_file("meta_data_v2")
#' il <- prep_get_data_frame("item_level")
#' il$STANDARDIZED_VOCABULARY_TABLE[[11]] <- "<ICD10GM>"
#' il$DATA_TYPE[[11]] <- DATA_TYPES$INTEGER
#' il$SCALE_LEVEL[[11]] <- SCALE_LEVELS$NOMINAL
#' prep_add_data_frames(item_level = il)
#' r <- dq_report2("study_data", dimensions = "con")
#' r <- dq_report2("study_data", dimensions = "con",
#'      advanced_options = list(dataquieR.non_disclosure = TRUE))
#' r
#' }
con_inadmissible_vocabulary <- con_inadmissible_categorical

# TODO: allow gsub before checking, because of this:
# > head(prep_get_data_frame("<ICD10>")[[1]])
# [1] "A00"  "A00"  "A00"  "A000" "A001" "A009"
# > head(prep_get_data_frame("<ICD10GM>")[[1]])
# [1] "A00.0" "A00.1" "A00.9" "A01.-" "A01.0" "A01.1"

#' Test for likely misspelled data frame references
#'
#' checks, if some data frame names may have typos in their names
#'
#' @param name_of_study_data [character] names of study data, such are expected
#'
#' @return `invisible(NULL)`, messages / warns only.
#'
#' @noRd
util_verify_names <- function(name_of_study_data = character(0)) {
  x <- .util_verify_names(observed_names = prep_list_dataframes(),
                          name_of_study_data = name_of_study_data)
  if (length(x$warn) > 0) {
    util_warning("Found potentially misspelled tables: %s",
                 paste(dQuote(x$warn2), "=?=>", sQuote(x$warn),
                       collapse = ", "), immediate = TRUE)
  }
  if (length(x$dontknow > 0)) {
    util_message("Found unexpected tables: %s",
                 util_pretty_vector_string(x$dontknow))
  }
  return(invisible(NULL))
}

.util_verify_names <- function(standard_names = c("CODE_LIST_TABLE",
                                                  "MISSING_CODE_RULES",
                                                  "cross-item_level",
                                                  "dataframe_level",
                                                  "item_level",
                                                  "segment_level",
                                                  "item_computation_level",
                                                  "study_data",
                                                  "columns definition",
                                                  "<>",
                                                  name_of_study_data),
                              observed_names = prep_list_dataframes(),
                              name_of_study_data = character(0)) {


  observed_names <- setdiff(observed_names,
         c(
           "https://dataquality.qihs.uni-greifswald.de/extdata/study_data.RData",
           "https://dataquality.qihs.uni-greifswald.de/extdata/ship.RDS",
           "https://dataquality.qihs.uni-greifswald.de/extdata/meta_data.RData",
           "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData",
           "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData"
          )
  )
  observed_names <- observed_names[!startsWith(observed_names,
    system.file(package = utils::packageName()))]
  observed_names <-
    grep(
      "^LABS_.*::[a-f0-9]",
      observed_names,
      invert = TRUE,
      value = TRUE)

  # fill referred_names
  len_ref_nm <- -1
  referred_names <- c()
  #Loops runs while new referred names are found
  while (len_ref_nm < length(referred_names)) {
    len_ref_nm <- length(referred_names)
    #check for all observed names and return them
    referred_names_df <- lapply(observed_names, function(x) {
      r <- NULL; try(r <- prep_get_data_frame(x), silent = TRUE); r})
    # check if they are data frames and takes the columns ending with _table
    referred_names_cur <- lapply(referred_names_df, function(x) {
      if (is.data.frame(x) && ncol(x) > 0) {
        x[, endsWith(colnames(x), "_TABLE")]
      } else {
        character(0)
      }
    })
    #update referred_names
    referred_names <- unique(c(referred_names,
                               as.vector(unlist(referred_names_cur, recursive = TRUE))))
  }
  # Note: we started from observed_names but that may not be
  # the expected names, because the user can overwrite it

  #add all data frames from dataframe_level
  if ("dataframe_level" %in% prep_list_dataframes()) {
    referred_names <- c(referred_names,
                        prep_get_data_frame("dataframe_level")[[DF_NAME]])
  }

  #normalize all names
  referred_names <- referred_names[!is.na(referred_names)]
  referred_names <-  trimws(referred_names)
  referred_names <-  gsub(sprintf("\\s*([%s])\\s*", SPLIT_CHAR), "\\1",
                          referred_names)

  observed_names <- trimws(observed_names)
  observed_names <-  gsub(sprintf("\\s*([%s])\\s*", SPLIT_CHAR), "\\1",
                           observed_names)

  prefixes <- unique(trimws(
    gsub(sprintf("^(.*?)[%s].*", SPLIT_CHAR), "\\1",
         grep(SPLIT_CHAR, value = TRUE, fixed = TRUE, observed_names))))

  prefixes_to_be_processed <- grep(".",
                                   prefixes,
                                   value = TRUE,
                                   fixed = TRUE,
                                   invert = TRUE)
  prefixes_to_be_processed <-
    prefixes_to_be_processed[!grepl( ":",
                                     prefixes_to_be_processed,
                                     fixed = TRUE)]

  prefixes <- c(prefixes,
                paste0("https://dataquality.qihs.uni-greifswald.de/extdata/",
                       prefixes_to_be_processed,
                       ".xlsx"))


  referred_names <-  c(referred_names, vapply(prefixes, FUN = function(prefix) {
    paste0(prefix, SPLIT_CHAR, c(standard_names, referred_names)) },
    FUN.VALUE = character(length(c(standard_names, referred_names)))))



  #Add the referred names to standard (-> expected names)
  standard_names <- c(standard_names, referred_names)
  standard_names <- standard_names[!is.na(standard_names)]

  #try to compare expected names with names available, warn about data frames not expected
  d1 <- adist(toupper(standard_names), toupper(observed_names))
  d2 <- adist((standard_names), (observed_names))
  rownames(d1) <- rownames(d2) <- standard_names
  colnames(d1) <- colnames(d2) <- observed_names
  names(dimnames(d1)) <- names(dimnames(d2)) <- c("standard_names", "observed_names")
  dn <- t(t(d2) / nchar(rownames(d2))) # https://stackoverflow.com/a/20596490
  dn1 <- t(t(d1) / nchar(rownames(d1))) # https://stackoverflow.com/a/20596490
  # redo the matrices with stringdist if that is available
  if (util_ensure_suggested("stringdist", err = FALSE, goal = "guessing typos")) {
    dn <- stringdist::stringdistmatrix(standard_names,
                                       observed_names, method = "jw", p = 0.1)
    dn1 <- stringdist::stringdistmatrix(toupper(standard_names),
                                        toupper(observed_names),
                                        method = "jw", p = 0.1)
  }
  # compute the minimum distance to an expected name for each column (observed names)
  xx <- suppressWarnings(apply(dn, 2, min))

  #initialize a new matrix
  non_standard <- matrix()
  #find non-matching columns (observed names) and keep only them
  try(non_standard <- d2[, xx > 0, drop = FALSE], silent = TRUE)

  #find the closest expected names for this non-matching
  likely <- setNames(rownames(d1)[apply(d1, 2, which.min)], nm = colnames(d1))
  #compute likelihood
  suppressWarnings(llhd <- 1 - apply(dn1, 2, min))
  props <- setNames(llhd, nm = likely)
  #identify proposals that are not closer than 90%
  dontknow <- names(likely[llhd <= 0.9])
  if (!is.null(dontknow)) for (sn in c(standard_names, observed_names))
    dontknow <- dontknow[!endsWith(dontknow, paste0("|", sn))]
  list(case_insensitive = d1, case_sensitive = d2, mismatches = non_standard,
       likely = likely, llhd = llhd, warn = likely[llhd > 0.9 & llhd < 1.0],
       warn2 = names(likely)[llhd > 0.9 & llhd < 1.0],
       dontknow = dontknow)
}

#' @inherit prep_map_labels
#'
#' @family mapping
#' @concept metadata_management
#' @noRd
util_map_labels <- function(x, meta_data = "item_level",
                            to = LABEL, from = VAR_NAMES, ifnotfound,
                            warn_ambiguous = FALSE) {
  if (!missing(x) && length(x) == 0) {
    return(character(0))
  }
  util_expect_data_frame(meta_data, c(from, to))

  if (warn_ambiguous) {
    ambiguities <- vapply(as.character(x),
                          FUN.VALUE = logical(1),
                          FUN = function(xx) {
                            length(which(meta_data[[from]] == xx)) > 1
                          })
    if (any(ambiguities)) {
      util_warning(paste("There are several entries '%s' in the metadata column",
                         "%s. Mapping to metadata column %s is ambiguous."),
                   paste0(as.character(x)[ambiguities], collapse = ", "),
                   from,
                   to,
                   applicability_problem = TRUE)
    }
  }

  nm <- meta_data[[from]]
  isempty <- util_empty(nm)
  nempty <- sum(isempty)

  i <- 0
  replnm <- paste0("..", i + seq_len(nempty))
  while (length(intersect(replnm, union(nm, x))) > 0) {
    i <- i + 1
    replnm <- paste0("..", i + seq_len(nempty))
  }

  nm[isempty] <- replnm

  x[!nzchar(as.character(x))] <- NA_character_

  unlist(mget(as.character(x),
              as.environment(as.list(
                setNames(as.character(meta_data[[to]]),
                         nm = nm)
              )),
              ifnotfound = ifnotfound))
}

#' Return the logged-in User's Full Name
#'
#' If `whoami` is not installed, the user name from
#' `Sys.info()` is returned.
#'
#' Can be overridden by options or environment:
#'
#' `options(FULLNAME = "Stephan Struckmann")`
#'
#' `Sys.setenv(FULLNAME = "Stephan Struckmann")`
#'
#' @return [character] the user's name
#'
#' @export
prep_get_user_name <- function() {
  if (length(options("FULLNAME")) == 1 &&
      length(options("FULLNAME")[[1]]) == 1 &&
      options("FULLNAME")[[1]] != "") {
    options("FULLNAME")
  } else if (Sys.getenv("FULLNAME", "") != "") {
    Sys.getenv("FULLNAME", "")
  } else if (requireNamespace('whoami', quietly = TRUE)) {
    whoami::fullname(fallback = Sys.info()[['user']])
  } else {
    Sys.info()[['user']]
  }
}

util_map_to_other_metrics <- function(sumtab,
                                      meta_data,
                                      label_col,
                                      add_attribs) {

  if (.called_in_pipeline &&
      COMPUTED_VARIABLE_ROLE %in% colnames(meta_data)) {
    cvr <-
      util_map_labels(sumtab$Variables,
                      meta_data = meta_data,
                      to = COMPUTED_VARIABLE_ROLE,
                      from = label_col,
                      ifnotfound = NA,
                      warn_ambiguous = FALSE)

    if (is.environment(add_attribs) && length(cvr) == 1 && is.character(cvr) &&
        !all(util_empty(cvr))) {
      add_attribs[[COMPUTED_VARIABLE_ROLE]] <- cvr
    }

    if (nrow(sumtab) == 1 && length(cvr) == 1 && is.character(cvr) &&
        !all(util_empty(cvr))) {

      cvim <-
        util_get_concept_info("computed_vars_ind_mapping",
                              get("computed_role") == cvr,
                              c("needle", "replacement"), drop = FALSE)

      sumtab <- sumtab[, c("Variables",
                           grep(paste0(".*(",
                                       paste(cvim$needle, collapse = "|"),
                                       ").*"),
                                colnames(sumtab),
                                value = TRUE, perl = TRUE)), FALSE]
      for (i in seq_len(nrow(cvim))) {
        needle <- cvim$needle[i]
        replacement <- cvim$replacement[i]

        colnames(sumtab) <- gsub(
          paste0("^([A-Z1-9]+_)", needle, "(.*)$"),
          paste0("\\1", replacement, "\\2"),
          colnames(sumtab),
          perl = TRUE
        )
      }

    }
  }

  return(sumtab)
}

odm_installed <- FALSE

odm_cache <- new.env(parent = emptyenv())

util_prune_odm_cache <- function() {
  rm(envir = odm_cache, list = ls(odm_cache, all.names = TRUE))
}

util_odm2dataquieR <- function(...) {
  key <- rlang::hash(list(...))
  if (exists(key, odm_cache)) {
    return(get(key, odm_cache))
  }
  odm2dataquieR <- function(...) {
    util_message(
      "Optional feature requires 'dataquieR2odm'. Get it from %s",
      sQuote("https://dataquality.qihs.uni-greifswald.de/DownloadR.html"));
    invisible(NULL)
  }
  if (odm_installed) {
    suppressWarnings(
      util_ensure_suggested("dataquieR2odm",
                            goal = "Read ODM files",
                            err = FALSE,
                            and_import = "odm2dataquieR"))
  }
  r <- odm2dataquieR(...)
  assign(key, r, odm_cache)
  r
}

util_is_odm_xml <- function(path, bytes = 65536L) {
  con <- file(path, open = "rb")
  on.exit(close(con), add = TRUE)

  x <- readChar(con, nchars = bytes, useBytes = TRUE)

  # UTF-8 BOM removal
  x <- sub("^\ufeff", "", x, useBytes = TRUE)

  # root element muss ODM sein
  if (!grepl("<\\s*([[:alnum:]_\\-]+:)?ODM\\b", x, perl = TRUE))
    return(FALSE)

  # any ODM namespace version is ok
  grepl("xmlns(:[[:alnum:]_\\-]+)?\\s*=\\s*['\"]http://www\\.cdisc\\.org/ns/odm/",
        x, perl = TRUE)
}

#' Call (nearly) one "Accuracy" function with many parameterizations at once
#' automatically
#'
#' @param ... Deprecated
#'
#' @description
#'
#' Deprecated
#'
#'
#' @return Deprecated
#'
#' @export
#' @importFrom stats setNames
#'
pipeline_vectorized <- function(...) { # nocov start
  lifecycle::deprecate_stop("2.1.0", what = "pipeline_vectorized()",
                      with = "dq_report2()")
} # nocov end

util_handle_complex_data_types <- function(meta_data) {

  melt <- function(meta_data, rows, column, value, force = FALSE, tp) {
    util_stop_if_not(is.data.frame(meta_data))
    util_expect_scalar(value)
    util_expect_scalar(column, check_type = is.character)
    util_expect_scalar(rows,
                       allow_more_than_one = TRUE,
                       min_length = nrow(meta_data),
                       max_length = nrow(meta_data),
                       check_type = is.logical)
    if (force) util_stop_if_not(!missing(tp))
    if (nrow(meta_data) == 0) {
      return(meta_data)
    }
    if (!column %in% colnames(meta_data)) {
      meta_data[[column]] <- NA_character_
    }
    if (force) {
      if (column != DATA_TYPE &&
          any(rows & !util_empty(meta_data[[column]]) &
          tolower(trimws(meta_data[[column]])) !=
          tolower(trimws(value)), na.rm = TRUE)) {
        util_warning(
          c("Overwriting some entries in %s in",
            "the %s, because they have %s set to %s"),
          dQuote(column),
          sQuote("meta_data"),
          sQuote(DATA_TYPE),
          dQuote(tp)
        )
      }
    } else {
      rows <- rows & util_empty(meta_data[[column]])
    }
    meta_data[rows, column] <- as.character(value)
    meta_data
  }

  if (DATA_TYPE %in% colnames(meta_data)) {
    # Handle non-basic data types

    tp <- tolower(trimws(meta_data[[DATA_TYPE]]))

    count_rows <- tp == "count"
    if (any(count_rows, na.rm = TRUE)) {
      meta_data <- melt(meta_data,
                        count_rows,
                        DATA_TYPE,
                        DATA_TYPES$INTEGER, force = TRUE, tp = tp)
      meta_data <- melt(meta_data,
                        count_rows,
                        SCALE_LEVEL,
                        SCALE_LEVELS$RATIO)
      meta_data <- melt(meta_data,
                        count_rows,
                        HARD_LIMITS,
                        "[0; Inf)")
      meta_data <- melt(meta_data,
                        count_rows,
                        EXTENDED_DATA_TYPE,
                        "count", force = TRUE, tp = tp)
      # TODO Implement these distributions in distributional indicator functions:
      # meta_data[[DISTRIBUTION]][count_rows] <- "Poisson | Negative Binomial"
    }

    unkown <- !tp %in% DATA_TYPES
    if (any(unkown)) {
      meta_data <- .util_fix_data_types(meta_data)
    }
  }

  return(meta_data)
}

#' Utility function outliers according to the rule of Huber et al.
#'
#' This function calculates outliers according to the rule of Huber et al.
#'
#' @param x [numeric] data to check for outliers
#'
#'
#' @return binary vector
#'
#' @importFrom stats sd
#'
#' @family outlier_functions
#' @concept outlier
#' @noRd

util_sigmagap <- function(x) {

  # sd
  xsd <- sd(x, na.rm = TRUE)
  xmu <- mean(x, na.rm = TRUE)

  # dataframe of original values and their distances
  ints <- data.frame(RN = seq_along(x), VALUE = x)
  ints <- ints[order(ints$VALUE), ]
  ints$int <- c(0, diff(ints$VALUE))
  ints$sigmagap <- ifelse(ints$int > xsd, 1, 0)

  if (any(!is.na(ints$sigmagap)) && max(ints$sigmagap, na.rm = TRUE) == 1) {
    # if break is low
    if (max(ints$VALUE[which(ints$sigmagap == 1)], na.rm = TRUE) < xmu) {
      ints$sigmagap[1:min(which(ints$sigmagap == 1), na.rm = TRUE)] <- 1
    }
    if (min(ints$VALUE[which(ints$sigmagap == 1)], na.rm = TRUE) > xmu) {
      ints$sigmagap[min(which(ints$sigmagap == 1), na.rm = TRUE):length(x)] <-
        1
    }
  }

  # order to original seq of data
  ints <- ints[order(ints$RN), ]

  xbin <- ints$sigmagap

  return(xbin)
}

set_state_inspector(function() {
  # normal purpose of this function is to check for changes of the global state
  # here, we abuse this function to ensure, that before any test runs,
  # the caches are emptied. Maybe, we need to be more careful and edit
  # each and any test to explicitly run the functions below, if needed.
  suppressMessages(prep_purge_data_frame_cache())
  options(rio.import.trust = TRUE) # the tests run well-known files, only
  # Sys.unsetenv("NOT_CRAN")
  prep_list_dataframes()
  util_reset_cache()
  invisible(NULL) # returned objects are considered global state
})

#' Check applicability of DQ functions on study data
#'
#' @description
#' Checks applicability of DQ functions based on study data and metadata
#' characteristics
#'
#' @details
#' This is a preparatory support function that compares study data with
#' associated metadata. A prerequisite of this function is that the no. of
#' columns in the study data complies with the no. of rows in the metadata.
#'
#' For each existing R-implementation, the function searches for necessary
#' static metadata and returns a heatmap like matrix indicating the
#' applicability of each data quality implementation.
#'
#' In addition, the data type defined in the metadata is compared with the
#' observed data type in the study data.
#'
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param split_segments [logical] return one matrix per study segment
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param max_vars_per_plot [integer] from=0. The maximum number of variables
#'                                            per single plot.
#' @param meta_data_segment [data.frame] -- optional: Segment level metadata
#' @param meta_data_dataframe [data.frame] -- optional: Data frame level
#'                                                                 metadata
#' @param meta_data [data.frame] old name for `item_level`
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#' @param segment_level [data.frame] alias for `meta_data_segment`
#' @param dataframe_level [data.frame] alias for `meta_data_dataframe`
#'
#' @inheritParams acc_distributions
#'
#' @return a list with:
#'   - `SummaryTable`: data frame about the applicability of each indicator
#'                   function (each function in a column).
#'                   its [integer] values can be one of the following four
#'                   categories:
#'                               0. Non-matching datatype + Incomplete metadata,
#'                               1. Non-matching datatype + complete metadata,
#'                               2. Matching datatype + Incomplete metadata,
#'                               3. Matching datatype + complete metadata,
#'                               4. Not applicable according to data type
#'   - `ApplicabilityPlot`: [ggplot2::ggplot2] heatmap plot, graphical representation of
#'                       `SummaryTable`
#'   - `ApplicabilityPlotList`: [list] of plots per (maybe artificial) segment
#'   - `ReportSummaryTable`: data frame underlying `ApplicabilityPlot`
#' @export
#' @importFrom ggplot2 ggplot aes geom_tile scale_fill_manual facet_wrap theme_minimal scale_x_discrete xlab guides guide_legend theme element_text
pro_applicability_matrix <- function(study_data,
                                     item_level = "item_level",
                                     split_segments =
                                      FALSE,
                                     label_col,
                                     max_vars_per_plot = 20,
                                     meta_data_segment,
                                     meta_data_dataframe,
                                     flip_mode = "noflip",
                                     meta_data_v2,
                                     meta_data = item_level,
                                     segment_level,
                                     dataframe_level) { # TODO: add meta_data_cross_item
  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()
  if (!missing(meta_data_segment) && !is.null(meta_data_segment)) {
    meta_data_segment <-
      prep_check_meta_data_segment(meta_data_segment) # TODO: Which columns are mandatory
    # TODO: If an alternative argument dq_control has been set, this may point to either something composed (e.g., excel workbook or rdata file) featuring all other standard data frames as sheets, or to a sheet naming all other standard data frames
  } else {
    meta_data_segment <- NULL # as.list(formals((function(x){})))[[1]] # after overwriting x, x is not missing(x) any more
  }
  if (!missing(meta_data_dataframe) && !is.null(meta_data_dataframe)) {
    meta_data_dataframe <-
      prep_check_meta_data_dataframe(meta_data_dataframe) # TODO: Which columns are mandatory # TODO: study data may be merged here
  } else {
    meta_data_dataframe <- NULL
  }

  if (length(max_vars_per_plot) != 1 || !util_is_integer(max_vars_per_plot) ||
      is.na(max_vars_per_plot) || is.nan(max_vars_per_plot) ||
      max_vars_per_plot < 1) {
    util_error(c(
      "max_vars_per_plot must be one strictly positive non-complex integer",
      "value, may be Inf."
    ), applicability_problem = TRUE)
  }

  prep_prepare_dataframes(.replace_missings = FALSE)

  Variables <- names(ds1)
  # this matrix will be shown to users
  app_matrix <- data.frame(Variables = Variables)
  # this matrix can be used to trigger computations
  # tri_matrix <- data.frame(Variables = Variables)
  # defined for merge of strata (optional)
  by_y <- label_col

  # DATA_TYPE defined in metadata?
  if (!("DATA_TYPE" %in% names(meta_data))) {
    util_error(
      c("The attribute DATA_TYPE is not contained in the metadata but is",
        "required for this function."), applicability_problem = TRUE)
  }

  # variables with missing DATA_TYPE?
  if (any(is.na(meta_data$DATA_TYPE))) {
    whichnot <- as.character(meta_data[[label_col]][is.na(meta_data$DATA_TYPE)])
    util_error(paste0("The DATA_TYPE for variable(s) <<", whichnot,
                      ">> is not defined in the metadata."),
               applicability_problem = TRUE)
  }

  # check whether data types adhere to conventions
  if (!all(unique(meta_data[[DATA_TYPE]]) %in% DATA_TYPES)) {
    whichnot <- dplyr::setdiff(unique(meta_data$DATA_TYPE), DATA_TYPES)
    util_warning(paste0("The data type(s): <<", whichnot,
                        ">> is not eligible in the metadata concept."),
                 applicability_problem = TRUE)
    util_error("Please map data types to: %s.", paste0(dQuote(DATA_TYPES),
                                                       collapse = ", "),
               applicability_problem = TRUE)
  }

  # DATA TYPE CONSISTENCY ------------------------------------------------------
  # This step compares the datatype as represented in study data with expected
  # datatypes in metadata and creates a binary vector indicating whether data
  # type fits or not
  dt_appl <- as.numeric(util_compare_meta_with_study(sdf = ds1, mdf = meta_data,
                                                     label_col = label_col,
                                                     check_conversion_stable =
                                                       TRUE))
  # combine convertible w/ and convertible w/o issues
  dt_appl[!is.na(dt_appl) & dt_appl == 3] <- 2


  # HINT: For each added function that returns MoifiedStudyData,
  # check, if this should be used downstream the dq_report-pipeline. If not
  # it must be added to the default values of the dont_modify_study_data_by
  # argument of dq_report.

  # INTEGRITY ---------------------------------------------------------------
  # crude unit missingness always applicable
  app_matrix$"int_all_datastructure_dataframe" <-
    ifelse(is.null(meta_data_dataframe), 4, 3)
  app_matrix$"int_all_datastructure_dataframe" <-
    as.factor(app_matrix$"int_all_datastructure_dataframe")

  app_matrix$"int_all_datastructure_segment" <-
    ifelse(is.null(meta_data_segment), 4, 3)
  app_matrix$"int_all_datastructure_segment" <-
    as.factor(app_matrix$"int_all_datastructure_segment")

  app_matrix$"int_datatype_matrix" <- 3
  app_matrix$"int_datatype_matrix" <-
    as.factor(app_matrix$"int_datatype_matrix")

  # COMPLETENESS ---------------------------------------------------------------
  # crude unit missingness always applicable
  app_matrix$"com_unit_missingness" <- 3
  app_matrix$"com_unit_missingness" <-
    as.factor(app_matrix$"com_unit_missingness")
  app_matrix$"com_segment_missingness" <- util_app_sm(meta_data, dt_appl) # TODO: Rename all app functions to contain the full indicator name as for util_app_con_contradictions_redcap below
  app_matrix$"com_item_missingness" <- util_app_im(meta_data, dt_appl) # TODO: Add com_qualified_item_missingness and com_qualified_segment_missingness

  # CONSISTENCY ----------------------------------------------------------------
  app_matrix$"con_hard_limits" <- util_app_hl(meta_data, dt_appl)
  app_matrix$"con_soft_limits" <- util_app_sl(meta_data, dt_appl)
  app_matrix$"con_detection_limits" <- util_app_dl(meta_data, dt_appl)
  app_matrix$"con_inadmissible_categorical" <- util_app_iac(meta_data, dt_appl)
  app_matrix$"con_contradictions" <- util_app_cd(meta_data, dt_appl)
  app_matrix$"con_contradictions_redcap" <-
    util_app_con_contradictions_redcap(meta_data, dt_appl)

  # ACCURACY -------------------------------------------------------------------
  app_matrix$"acc_univariate_outlier" <- util_app_ol(meta_data, dt_appl)
  app_matrix$"acc_distributions" <- util_app_dc(meta_data, dt_appl)
  app_matrix$"acc_margins" <- util_app_mar(meta_data, dt_appl)
  app_matrix$"acc_varcomp" <- util_app_vc(meta_data, dt_appl)
  app_matrix$"acc_loess" <- util_app_loess(meta_data, dt_appl)
  app_matrix$"acc_shape_or_scale" <- util_app_sos(meta_data, dt_appl)
  app_matrix$"acc_end_digits" <- util_app_ed(meta_data, dt_appl)
  app_matrix$"acc_multivariate_outlier" <- util_app_mol(meta_data, dt_appl)

  # SHOULD STRATIFICATION OF SEGMENTS BE USED? ---------------------------------

  # is STUDY_SEGMENT in metadata and none of the entries has NA?
  strata_defined <- STUDY_SEGMENT %in% names(meta_data) &&
    !all(util_empty(meta_data[[STUDY_SEGMENT]]))

  if (!strata_defined && split_segments) {
    util_message(c(
      "Stratification for STUDY_SEGMENT is not possible due to missing",
      "metadata. Will split arbitrarily avoiding too large figures"
    ), applicability_problem = TRUE)
    nvars <- nrow(meta_data)
    meta_data$STUDY_SEGMENT <- paste0("Block #", ceiling(1:nvars /
                                                             max_vars_per_plot))
  } else if (strata_defined && split_segments) {
    if (any(is.na(meta_data$STUDY_SEGMENT))) { # HINT: This is now dead code
      util_message(c(
        "Some STUDY_SEGMENT are NA. Will assign those to an artificial",
        "segment %s"), dQuote("Other"),
        applicability_problem = TRUE
      )
      meta_data$STUDY_SEGMENT[is.na(meta_data$STUDY_SEGMENT)] <- "Other"
    }
    too_big_blocks <- table(meta_data$STUDY_SEGMENT) > max_vars_per_plot
    too_big_blocks <- names(too_big_blocks)[too_big_blocks]
    for (too_big_block in too_big_blocks) {
      util_message(
        "Will split segemnt %s arbitrarily avoiding too large figures",
        dQuote(too_big_block),
        applicability_problem = FALSE
      )
      nvars <- sum(meta_data$STUDY_SEGMENT == too_big_block, na.rm = TRUE)
      meta_data$STUDY_SEGMENT[
        meta_data$STUDY_SEGMENT == too_big_block] <-
        paste0(
          meta_data$STUDY_SEGMENT[meta_data$STUDY_SEGMENT ==
                                        too_big_block],
          "#",
          ceiling(1:nvars / max_vars_per_plot)
        )
    }
  }

  if (!(STUDY_SEGMENT %in% names(meta_data))) {
    meta_data[[STUDY_SEGMENT]] <- "Study"
  }

  # merge relation to segments to app_matrix
  app_matrix <- merge(app_matrix, meta_data[, intersect(c(VAR_NAMES, LABEL,
                                                          STUDY_SEGMENT,
                                                          label_col),
                                                        colnames(meta_data)),
                                            FALSE],
    by.x = "Variables",
    by.y = by_y
  )
  app_matrix <- app_matrix[, !(names(app_matrix) %in% c(VAR_NAMES, LABEL,
                                                        label_col))]

  # PREPARE PLOT ---------------------------------------------------------------
  # reorder according VARIABLE_ORDER (optional)
  if (VARIABLE_ORDER %in% colnames(meta_data)) {
    meta_data <- meta_data[order(meta_data$VARIABLE_ORDER), ]
    meta_data[[VARIABLE_ORDER]][is.na(meta_data[[VARIABLE_ORDER]])] <-
      max(meta_data[[VARIABLE_ORDER]], na.rm = TRUE)
    meta_data[[VARIABLE_ORDER]][is.na(meta_data[[VARIABLE_ORDER]])] <-
      1
    app_matrix <- app_matrix[na.omit(match(meta_data[[label_col]],
                                     app_matrix$Variables)), ]
  }

  # assign factor levels
  app_matrix$Variables <- factor(app_matrix$Variables, levels =
                                   app_matrix$Variables)


  # reshape wide to long
#  app_matrix_long <- melt(app_matrix, id.vars = c("Variables",
#                                                  STUDY_SEGMENT))
  app_matrix_long <-
    stats::reshape(data = app_matrix,
                   idvar = c("Variables", STUDY_SEGMENT),
                   varying = colnames(app_matrix)[2:(ncol(app_matrix)-1)],
                   v.names = "value",
                   times = colnames(app_matrix)[2:(ncol(app_matrix)-1)],
                   direction = "long")

  levs <- unique(app_matrix_long$time)
  #sort levels
  #levs <- sort(levs)
  levs_int <- levs[grep("int_", levs)]
  levs_com <- levs[grep("com_", levs)]
  levs_con <- levs[grep("con_", levs)]
  levs_acc <- levs[grep("acc_", levs)]

  #order the factor levels
  app_matrix_long$time <- factor(app_matrix_long$time,
                                  levels = c(levs_int,levs_com,
                                             levs_con, levs_acc))
  rownames(app_matrix_long) <- NULL
  colnames(app_matrix_long) <- c("VARIABLES", "STUDY_SEGMENT", "IMPLEMENTATION",
                                 "APP_SCORE")

  #Order rows by int, com, con, and acc
  app_matrix_long_int <- app_matrix_long[grep("int_",
                                             app_matrix_long$IMPLEMENTATION), ,
                                        drop = FALSE]
  app_matrix_long_int <-
    app_matrix_long_int[order(app_matrix_long_int$IMPLEMENTATION), , drop = FALSE]


  app_matrix_long_com <- app_matrix_long[grep("com_",
                                             app_matrix_long$IMPLEMENTATION), ,
                                        drop = FALSE]
  app_matrix_long_com <-
    app_matrix_long_com[order(app_matrix_long_com$IMPLEMENTATION), , drop = FALSE]

  app_matrix_long_con<- app_matrix_long[grep("con_",
                                             app_matrix_long$IMPLEMENTATION), ,
                                        drop = FALSE]
  app_matrix_long_con <-
    app_matrix_long_con[order(app_matrix_long_con$IMPLEMENTATION), , drop = FALSE]

  app_matrix_long_acc <- app_matrix_long[grep("acc_",
                                             app_matrix_long$IMPLEMENTATION), ,
                                        drop = FALSE]
  app_matrix_long_acc <-
    app_matrix_long_acc[order(app_matrix_long_acc$IMPLEMENTATION), , drop = FALSE]

  app_matrix_long<- rbind(app_matrix_long_int, app_matrix_long_com,
                          app_matrix_long_con, app_matrix_long_acc)

  lbs <- c(
    "Non-matching datatype + Incomplete metadata",
    "Non-matching datatype + complete metadata",
    "Matching datatype + Incomplete metadata",
    "Matching datatype + complete metadata",
    "Not applicable according to data type"
  )
  lvls <- c(
    0,
    1,
    2,
    3,
    4
  )
  # assign factor labels
  app_matrix_long$APP_SCORE <- factor(app_matrix_long$APP_SCORE,
    levels = lvls,
    labels = lbs
  )
  # PLOT -----------------------------------------------------------------------
  dred  <- "#B2182B"
  lred  <- "#EF6548"
  lblue <- "#92C5DE"
  dblue <- "#2166AC"
  grey  <- "#B0B0B0"

  colcode <- c(dred, lred, lblue, dblue, grey)
  names(colcode) <- lbs

  ratio <- dim(app_matrix)[1] / dim(app_matrix)[2]

  ref_env <- environment()

  plot_me <- function(m) {
    fli <- util_coord_flip(ref_env = ref_env)
    util_create_lean_ggplot(
      ggplot(m, aes(
        x = IMPLEMENTATION, y = VARIABLES,
        fill = APP_SCORE
      )) +
        geom_tile(colour = "white", linewidth = 0.8) + # https://github.com/tidyverse/ggplot2/issues/5051
        scale_fill_manual(values = colcode, name = " ") +
        {
          if (split_segments) facet_wrap(~STUDY_SEGMENT,
                                         scales = "free_y")# TODO: check ~
        } +
        theme_minimal() +
        scale_x_discrete(position = "top") +
        xlab("") +
        guides(fill = guide_legend(
          ncol = 1, nrow = length(colcode),
          byrow = TRUE
        )) +
        theme(
          legend.position = "bottom",
          axis.text.x = element_text(angle = 90, hjust = 0),
          axis.text.y = element_text(size = 10)#,
          #aspect.ratio = ratio
        ) + fli,
      m = m,
      colcode = colcode,
      split_segments = split_segments,
      fli = fli)
  }

  p <- plot_me(app_matrix_long)

  pl <- lapply(
    split(app_matrix_long, app_matrix_long$STUDY_SEGMENT),
    plot_me
  )


  ReportSummaryTable <- app_matrix
  ReportSummaryTable$STUDY_SEGMENT <- NULL

  ReportSummaryTable[2:ncol(ReportSummaryTable)] <-
    lapply(ReportSummaryTable[2:ncol(ReportSummaryTable)], util_as_numeric)

#  ReportSummaryTable[2:ncol(ReportSummaryTable)] <-
#    4 - ReportSummaryTable[2:ncol(ReportSummaryTable)]

  ReportSummaryTable$N <- 4

  attr(ReportSummaryTable, "higher_means") <- "better"
  attr(ReportSummaryTable, "continuous") <- FALSE
  attr(ReportSummaryTable, "colcode") <- colcode
  attr(ReportSummaryTable, "level_names") <-
    setNames(nm = 0:4, levels(app_matrix_long$APP_SCORE))


  ReportSummaryTable <- util_validate_report_summary_table(ReportSummaryTable,
                                                           meta_data = meta_data,
                                                           label_col = label_col)

  return(list(
    ApplicabilityPlot = p,
    ApplicabilityPlotList = pl,
    SummaryTable = app_matrix,
    ReportSummaryTable = ReportSummaryTable
  ))
}

#' Filter a `MISSING_LIST_TABLE` for rows matching the variable `rv`
#'
#' In `MISSING_LIST_TABLE`, a column `resp_vars` may be specified. If so,
#' and if, for a row, this column is not empty, then that row only affects the
#' one variable specified in that cell
#'
#' @param table [cause_label_df] a data frame with missing codes and
#'                               optionally `resp_vars`. It also comprises
#'                               labels and optionally an interpretation
#'                               column with `AAPOR` codes. Must already cover
#'                               the variable `rv`, i.e., item level metadata
#'                               is not checked to find the suitable missing
#'                               table for `rv`.
#'
#' @param rv [variable] the response variable to filter the missing list for
#'                      specified by a label.
#' @param rv2 [variable] the response variable to filter the missing list for
#'                       specified by a `VAR_NAMES`-name.
#'
#' @return [data.frame] the row-wise bound data frames as one data frame
#'
#' @family missing_functions
#' @concept metadata_management
#' @noRd
util_filter_missing_list_table_for_rv <- function(table, rv, rv2 = rv) {
  util_expect_scalar(rv, check_type = is.character)
  util_expect_scalar(rv2, check_type = is.character)
  util_expect_data_frame(table)
  if (!("resp_vars" %in% colnames(table))) { # no filter column, all rows match
    return(table)
  }
  unspecific <- table[util_empty(table[["resp_vars"]]), ,
                      FALSE] # rows w/o resp_vars restriction
  specific <- table[!util_empty(table[["resp_vars"]]) &
                      trimws(table[["resp_vars"]]) %in% c(rv, rv2), ,
                    FALSE]  # rows restricted by resp_vars on rv
  rbind(specific, unspecific) # combine both
}

#' Can a vector be converted to  a defined `DATA_TYPE`
#'
#' the function also checks, if the conversion is perfect, or if something
#' is lost (e.g., decimal places), or something is strange (like arbitrary
#' suffixes in a date, just note, that
#' `as.POSIXct("2020-01-01 12:00:00 CET asdf")` does not fail in `R`), but
#' `util_conversion_stable("2020-01-01 12:00:00 CET asdf", DATA_TYPES$DATETIME)`
#' will return `FALSE`.
#'
#' *HINT:*
#' `util_conversion_stable(.Machine$integer.max + 1, DATA_TYPES$INTEGER)` seems
#' to work correctly, although `is.integer(.Machine$integer.max + 1)`
#' returns `FALSE`.
#'
#' @param vector [vector] input vector,
#' @param data_type [enum] The type, to what the conversion should be tried.
#' @param return_percentages [logical] return the percentage of stable
#'                                     conversions or matches.
#'
#' @return [numeric] ratio of convertible entries in `vector`
#' @noRd
util_conversion_stable <- function(vector, data_type,
                                   return_percentages = FALSE) {
  util_expect_scalar(return_percentages, check_type = is.logical)
  util_expect_scalar(vector,
                     allow_more_than_one = TRUE,
                     allow_na = TRUE,
                     error_message =
                       sprintf(
                         "argument %s needs to be convertible to a string",
                                             sQuote("vector")),
                     check_type = is.character,
                     convert_if_possible = util_as_character)
  util_match_arg(data_type, DATA_TYPES)
  if (data_type == DATA_TYPES$INTEGER) {
    as_target <- util_data_type_conversion(vector, data_type)
    if (identical(attr(data_type, "orig_type"), "logical")) {
      as_num <- util_data_type_conversion(as.logical(vector), DATA_TYPES$FLOAT)
      res <-
        (
          (util_empty(as_target) & util_empty(vector)) |
            ((!util_empty(as_target) & !util_empty(vector) &
                as_target == as_num))
        )
    } else {
      as_num <- util_data_type_conversion(vector, DATA_TYPES$FLOAT)
      res <-
        (
          (util_empty(as_target) & util_empty(vector)) |
            ((!util_empty(as_target) & !util_empty(vector) &
                as_target == as_num))
        )
    }
  } else if (data_type == DATA_TYPES$FLOAT) { # scientific notation
    as_target <- util_data_type_conversion(vector, data_type)
    res <- util_empty(vector) == util_empty(as_target)
  } else if (data_type == DATA_TYPES$DATETIME) {
    # FIXME: integrate somehow my_parse_date to util_data_type_conversion -- also parse_time and so on
    # as_target <- util_data_type_conversion(string, data_type)
    as_target <- .my_parse_date(vector)
    res <- util_empty(vector) == util_empty(as_target)
  } else if (data_type == DATA_TYPES$TIME) {
    as_target <- util_as_time_only(vector)
    res <- util_empty(vector) == util_empty(as_target)
  } else if (data_type == DATA_TYPES$STRING) {
    res <- rep(TRUE, length(vector))
  }
  if (return_percentages) {
    res <- sum(res) / length(vector) * 100
  }
  res
}

# x <- c("2009-09-29x", "2012-11-29 CET", "2015-29-12", "2009-09-29x", "2009-09-29x", "2012-11-29 12:00:00 CET", "2015-29-12 13:00:00", "2009-09-29x")
# x2 <- as.character(as.POSIXct(rnorm(1000000, mean = 1000000000, sd = 100000000)))
.my_parse_date <- function(x) {
  x <- trimws(x, "right")
  # remove all OlsonNames, if endswith
  for (n in union(OlsonNames(), c("CEST", "CET"))) {
    subst <- !is.na(x) & endsWith(x, n)
    l <- nchar(n)
    x[subst] <-
      substr(x[subst], 1, nchar(x[subst]) - l)
  }
  # IDEA: can using this in favor of lubridate everywhere remove the dependency from lubridate: no, lubridate has also round_date and duration-related stuff?
  return(suppressWarnings(readr::parse_datetime(x,
                                                locale =
                                                  readr::locale(
                                                    tz =
                                                      Sys.timezone()))))
}

#' Return names of result slots (e.g., 3rd dimension of dataquieR results)
#'
#' @param x the objects
#'
#' @return character vector with names
#' @export
resnames <- function(x) {
  UseMethod("resnames")
}

################################################################################
#' Remove specific classes from a ggplot `plot_env` environment
#'
#' Useful to remove large objects before writing to disk with `qs` or `rds`.
#' Also deletes parent environment of the plot environment.
#' Also deletes unneeded variables
#'
#' @param r the object
#'
#' @seealso [HERE](https://github.com/tidyverse/ggplot2/issues/3619#issuecomment-628021555)
#'
#' @noRd
util_compress_ggplots_in_res <- function(r) {
  if (isTRUE(attr(r, "from_ReportSummaryTable"))) {
    return(NULL) # never store plots of reportsummarytables, because the original objects are already in the report
  }
  if (util_is_gg_plot(r)) {
    r$plot_env <- emptyenv()
    # https://stackoverflow.com/questions/75698707/how-to-extract-variable-names-from-aes-mapping-in-r/75699079#75699079
    #mv <- unique(unlist(lapply(r$mapping, all.vars))) # does not work for quosures .data[["variable_name"]] - gives '.data' instead of 'variable_name'
    mv <- unique(unlist(lapply(r$mapping, function(ll) {
      quo_ll_map <- rlang::quo_get_expr(ll)
      if (".data" %in% as.character(quo_ll_map)) {
        colnames(r$data)[which(colnames(r$data) %in% as.character(quo_ll_map))]
      } else {
        all.vars(ll)
      }
    })))
    mv <- unique(c(mv,
                   unlist(lapply(r$layers,
                                 function(lly)  {
                                   lapply(lly$mapping, function(ll) {
                                     quo_ll_map <- NULL
                                     try(quo_ll_map <- rlang::quo_get_expr(ll),
                                         silent = TRUE)
                                     if (".data" %in% as.character(quo_ll_map)) { #TODO: also handle the data in the geoms layers
                                       colnames(r$data)[which(colnames(r$data) %in%
                                                                as.character(quo_ll_map))]
                                     } else {
                                       all.vars(ll)
                                     }
                                   })
                                 }))))
    mv <- c(mv, "facet", "resp_var_adj") # keep column facet as used in util_margins_nom, column resp_var_adj is used in ..._lm, ..._bin, ..._poi

    if ("facet" %in% names(r)) {
      facet_v <- NULL
      try({
        facet_v <- r$facet$vars()
      }, silent = TRUE)
      if (is.character(facet_v) && length(facet_v) > 0) {
        mv <- c(mv, facet_v)
      }
    }

    if ("data" %in% names(r) && !is.null(dim(r$data))) {
      r$data <- r$data[, intersect(colnames(r$data), mv), drop = FALSE]
    }
  } else if (is.list(r)) {
    if (!inherits(r, "dq_lazy_ggplot")) {
      r[] <- lapply(r, util_compress_ggplots_in_res)
      r[vapply(r, is.null, logical(1))] <- NULL
    }
  }
  return(r)
}

#' Extract old function based summary from data quality results
#'
#' @param r [dq_report2]
#'
#' @return [data.frame] long format, compatible with [prep_summary_to_classes()]
#' @export
#'
#' @family summary_functions
prep_extract_classes_by_functions <- function(r) {
  te <- topenv(parent.frame(1)) # see https://stackoverflow.com/a/27870803
  if (!(isNamespace(te) && getNamespaceName(te) == "dataquieR")) {
    lifecycle::deprecate_soft("2.1.0.9007",
                              "prep_combine_report_summaries()")
  }

  res <-
    prep_summary_to_classes(
      prep_extract_summary(r))

  res$function_name <-
    vapply(FUN.VALUE = character(1),
           setNames(nm = res$call_names),
           util_cll_nm2fkt_nm,
           report = r)

  if (!"class" %in% colnames(res)) {
    res$class <- rep(util_as_cat(NA), nrow(res))
  }

  if (!"indicator_metric" %in% colnames(res)) {
    res$indicator_metric <- rep(NA_character_, nrow(res))
  }

  if (!"value" %in% colnames(res)) {
    res$value <- rep(NA_character_, nrow(res))
  }

  if (!"values_raw" %in% colnames(res)) {
    res$values_raw <- rep(NA_character_, nrow(res))
  }

  if (!"n_classes" %in% colnames(res)) {
    res$n_classes <- rep(NA_character_, nrow(res))
  }

  if (!"call_names" %in% colnames(res)) {
    res$call_names <- rep(NA_character_, nrow(res))
  }

  res[, c("VAR_NAMES", "class", "indicator_metric", "value", "values_raw",
          "n_classes",
          "STUDY_SEGMENT", "call_names", "function_name"), drop = FALSE]

}

#' Compute Indicators for Qualified Segment Missingness
#'
#' [Indicator]
#'
#' @inheritParams .template_function_indicator
#'
#' @param meta_data_segment [data.frame] Segment level metadata
#' @param expected_observations [enum] HIERARCHY | ALL | SEGMENT. Report the
#'                                     number of observations expected using
#'                                     the old `PART_VAR` concept. See
#'                                     [com_item_missingness] for an
#'                                     explanation.
#' @param segment_level [data.frame] alias for `meta_data_segment`
#'
#' @return A [list] with:
#'   - `SegmentTable`: [data.frame] containing data quality checks for
#'                     "Non-response rate" (`PCT_com_qum_nonresp`) and
#'                     "Refusal rate" (`PCT_com_qum_refusal`) for each segment.
#'   - `SegmentData`: a [data.frame] containing data quality checks for
#'                    "Unexpected location" and "Unexpected proportion" per
#'                     segment for a report
#'
#' @export
com_qualified_segment_missingness <- function(label_col = NULL,
                                              study_data,
                                              item_level = "item_level",
                                              expected_observations = c("HIERARCHY",
                                                                        "ALL",
                                                                        "SEGMENT"),
                                              meta_data = item_level,
                                              meta_data_v2,
                                              meta_data_segment,
                                              segment_level) {

  # Preparation of the input ----
  util_maybe_load_meta_data_v2()

  util_ck_arg_aliases()

  prep_prepare_dataframes(.replace_missings = FALSE)

  expected_observations <- util_match_arg(expected_observations)

  meta_data_segment <- prep_check_meta_data_segment(meta_data_segment)

  # allowed AAPOR-States
  AAPOR_STATES <- c("I", "P", "PL", "R", "BO", "NC", "O", "UH", "UO", "NE")

  # Loop over all segments defined on segment-level metadata ----

  aapors <-
    lapply(setNames(nm = meta_data_segment[[STUDY_SEGMENT]]),
           function(segment) { # segment is the currently examined segment's name
             cur_seg <- meta_data_segment[ # only the line(s) from metadata segment, that refer to the currently examined segment
               meta_data_segment[[STUDY_SEGMENT]] == segment, , FALSE]
             util_stop_if_not(nrow(cur_seg) == 1)
             .cur_state_var <- cur_seg[[SEGMENT_PART_VARS]] # participation status variable of the current segment
             if (length(.cur_state_var) != 1 || is.na(.cur_state_var)) {
               util_warning("Missing or doubled %s in %s for %s",
                            sQuote(SEGMENT_PART_VARS),
                            sQuote("meta_data_segment"),
                            dQuote(segment),
                            applicability_problem = TRUE)
               return(NULL)
             }
             cur_state_var_vn <- meta_data[meta_data[[label_col]] == .cur_state_var |
                                             meta_data[[VAR_NAMES]] == .cur_state_var,
                                           "VAR_NAMES"]
             cur_state_var <- meta_data[meta_data[[label_col]] == .cur_state_var |
                                          meta_data[[VAR_NAMES]] == .cur_state_var,
                                        label_col]
             if (length(cur_state_var) != 1) {
               util_warning("Missing or doubled %s in %s",
                            dQuote(.cur_state_var),
                            sQuote("meta_data"),
                            applicability_problem = TRUE)
               return(NULL)
             }
             if (cur_state_var %in% colnames(ds1)) {
               .cur_miss <- meta_data[meta_data[[label_col]] == cur_state_var,
                                      MISSING_LIST_TABLE]
               if (length(.cur_miss) == 0) {
                 .cur_miss <- NA_character_
               }
               cur_miss <- try(util_expect_data_frame(
                 x = .cur_miss,
                 dont_assign = TRUE,
                 col_names = list(
                   CODE_VALUE = util_is_valid_missing_codes,
                   CODE_INTERPRET = function(x) is.character(x) & x %in% AAPOR_STATES
                 ),
                 custom_errors = list(
                   CODE_VALUE = "CODE_VALUE must be numeric or DATETIME",
                   CODE_INTERPRET = paste("CODE_INTERPRET must be one of",
                                          util_pretty_vector_string(AAPOR_STATES))
                 ),
                 convert_if_possible = list(
                   CODE_VALUE = util_as_valid_missing_codes,
                   CODE_INTERPRET = function(x) trimws(toupper(x))
                 )
               ), silent = TRUE)
               if (inherits(cur_miss, "try-error")) {
                 if (!is.na(.cur_miss)) {
                   util_warning(
                     "Could not load missing-match-table %s for %s variable %s for segment %s: %s",
                     dQuote(.cur_miss),
                     sQuote(SEGMENT_PART_VARS),
                     dQuote(cur_state_var),
                     dQuote(segment),
                     conditionMessage(attr(cur_miss, "condition")),
                     applicability_problem = TRUE
                   )
                 } else {
                   util_warning(
                     "No missing-match-table for response variable %s for segment %s",
                     dQuote(cur_state_var),
                     dQuote(segment),
                     applicability_problem = TRUE
                   )
                 }
                 return(NULL)
               } else {
                 cur_miss <- util_filter_missing_list_table_for_rv(
                   cur_miss, cur_state_var, cur_state_var_vn
                 )

                 # really compute stuff ----
                 r <- util_table_of_vct(factor(
                   ds1[[cur_state_var]],
                   levels = cur_miss[[CODE_VALUE]],
                   labels = cur_miss[[CODE_INTERPRET]]
                 ))

                 r <- as.list(setNames(r$Freq, nm = r$Var1))

                 # replacement list
                 rclean <- r
                 for (name in AAPOR_STATES) {
                   if (is.null(rclean[[name]])) rclean[[name]] <- 0
                 }

                 # RR1 / NRR1
                 if (!is.null(r$I) | !is.null(r$P)) {
                   r$RR1 <- (rclean$I + rclean$P) /
                     ((rclean$I + rclean$P + rclean$PL) +
                        (rclean$R + rclean$BO + rclean$NC + rclean$O) +
                        (rclean$UH + rclean$UO))
                   r$NRR1 <- 1 - r$RR1
                   r$PCT_com_qum_nonresp <- 100 * r$NRR1
                 } else {
                   util_warning(
                     "Preconditions to generate Nonresponse Rate 1 not given for segment %s",
                     dQuote(segment),
                     applicability_problem = TRUE)
                   r$NRR1 <- NA_real_
                   r$RR1 <- NA_real_
                   r$PCT_com_qum_nonresp <- NA_real_
                 }

                 # RR2 / NRR2
                 if (!is.null(r$I) | !is.null(r$P) | !is.null(r$PL)) {
                   r$RR2 <- (rclean$I + rclean$P + rclean$PL) /
                     ((rclean$I + rclean$P + rclean$PL) +
                        (rclean$R + rclean$BO + rclean$NC + rclean$O) +
                        (rclean$UH + rclean$UO))
                   r$NRR2 <- 1 - r$RR2
                 } else {
                   util_warning(
                     "Preconditions to generate Nonresponse Rate 1 not given for segment %s",
                     dQuote(segment),
                     applicability_problem = TRUE)
                   r$NRR2 <- NA_real_
                   r$RR2 <- NA_real_
                 }

                 # REF1 / PCT_refusal
                 if (!is.null(r$R) | !is.null(r$BO)) {
                   r$REF1 <- (rclean$R + rclean$BO) /
                     ((rclean$I + rclean$P + rclean$PL) +
                        (rclean$R + rclean$BO + rclean$NC + rclean$O) +
                        (rclean$UH + rclean$UO))
                   r$PCT_com_qum_refusal <- 100 * r$REF1
                 } else {
                   util_warning(
                     "Preconditions to generate Nonresponse Rate 1 not given for segment %s",
                     dQuote(segment),
                     applicability_problem = TRUE)
                   r$REF1 <- NA_real_
                   r$PCT_com_qum_refusal <- NA_real_
                 }

                 r$N  <- sum(!util_empty(ds1[[cur_state_var]]))
                 r$N2 <- util_count_expected_observations(
                   cur_state_var,
                   study_data = study_data,
                   meta_data = meta_data,
                   label_col = label_col,
                   expected_observations = expected_observations
                 )

                 r <- cbind(
                   data.frame(Segment = segment, stringsAsFactors = FALSE, check.names = FALSE, row.names = NULL),
                   data.frame(r, stringsAsFactors = FALSE, check.names = FALSE, row.names = NULL)
                 )
               }
             } else {
               util_warning("Missing %s variable %s from %s for %s",
                            sQuote(SEGMENT_PART_VARS),
                            dQuote(cur_state_var),
                            sQuote("study_data"),
                            dQuote(segment),
                            applicability_problem = TRUE,
                            intrinsic_applicability_problem = TRUE)
               return(NULL)
             }
           })

  SegmentTable <- util_rbind(data_frames_list = aapors)
  rownames(SegmentTable) <- NULL

  # --- Start refactored part: use helpers for selection + translation ---
  # Select only indicator metric columns
  indicator_cols <- util_extract_indicator_metrics(SegmentTable)

  # Assemble SegmentData = Segment + indicator columns
  SegmentData <- cbind(
    data.frame(Segment = SegmentTable[["Segment"]], stringsAsFactors = FALSE, check.names = FALSE, row.names = NULL),
    indicator_cols
  )

  # Format percentage columns BEFORE renaming (based on original PCT_ prefixes)
  pct_cols <- colnames(SegmentData)[startsWith(colnames(SegmentData), "PCT_")]
  if (length(pct_cols)) {
    SegmentData[, pct_cols] <- lapply(
      SegmentData[, pct_cols, drop = FALSE],
      function(cl) paste0(round(cl, 2), "%")
    )
  }

  # Translate column headers via helper (keep unknowns such as "Segment" intact)
  colnames(SegmentData) <- util_translate_indicator_metrics(
    colnames(SegmentData),
    short = FALSE,
    long  = TRUE,
    ignore_unknown = TRUE
  )
  # --- End refactored part ---

  ## TODO: to discuss
  # if (!prod(dim(SegmentTable))) {
  #   util_error("Can not compute qualified segment missingness",
  #              applicability_problem = TRUE)
  # }

  text_to_display <- util_get_hovertext("[com_qualified_segment_missingness_hover]")
  attr(SegmentData, "description") <- text_to_display

  return(list(
    SegmentTable = SegmentTable,
    SegmentData  = SegmentData
  ))
}

#' Utility Function Heatmap with 1 Threshold
#'
#' Function to create heatmap-like plot given one threshold -- works for
#' percentages for now.
#'
#' @param df [data.frame] with data to display as a heatmap.
#' @param cat_vars [variable list] len=1-2. Variables to group by. Up to 2
#'                                          group levels supported.
#' @param values [variable] the name of the percentage variable
#' @param threshold [numeric] lowest acceptable value
#' @param right_intv [logical] len=1. If `FALSE` (default), intervals used
#'                                    to define color ranges in the heatmap
#'                                    are closed on the left side, if `TRUE` on
#'                                    the right side, respectively.
#' @param invert [logical] len=1. If `TRUE`, high values are better, warning
#'                                colors are used for low values. `FALSE` works
#'                                vice versa.
#' @param cols deprecated, ignored.
#' @param strata [variable] optional, the name of a variable
#'                          used for stratification
#' `inheritParams` `acc_distributions`
#' @return a [list] with:
#'   - `SummaryPlot`: [ggplot2::ggplot] object with the heatmap
#'
#' @importFrom ggplot2 ggplot aes geom_bar geom_text scale_fill_manual scale_y_discrete coord_flip facet_grid theme_minimal scale_x_discrete
#'
#'
#' @family figure_functions
#' @concept summary
#' @noRd
util_heatmap_1th <- function(df, cat_vars, values, threshold, right_intv,
                             invert, cols, strata
                             # , flip_mode # TODO: pass through from all usages of this function
                             ) {

  # STOPs
  if (!(length(cat_vars) %in% c(1, 2))) {
    util_error(paste0(
      "Argument cat_vars can have 1 or 2 elements. You specified ",
      length(cat_vars), "elements."), applicability_problem = TRUE)
  }

  if (!(is.numeric(df[[values]]))) {
    util_error(paste0("The variable you specified under ",
                      values, " must be numeric."),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  if (missing(threshold)) {
    util_error(paste0("No threshold has been specified"),
               applicability_problem = TRUE)
  }

  # Preps
  df$z2 <- df[[values]]

  if (missing(right_intv)) {
    right_intv <- FALSE
  }

  # maximum of data values
  if (invert == 0) {
    maxz <- max(df$z2, na.rm = TRUE)
  } else {
    maxz <- min(df$z2, na.rm = TRUE)
  }

  # interpret threshold as lowest acceptable value
  minz <- threshold

  warn_cols <- c(
    "#7f0000", "#b30000", "#d7301f", "#ef6548", "#fc8d59",
    "#fdbb84", "#fdd49e", "#fee8c8", "#2166AC"
  )

  range <- maxz - minz

  # values below thresholds == BLUE, above threshold a range to darkred
  steps <- range / (length(warn_cols) - 1)

  # create step-wise color-coding
  midrange <- c(threshold)
  nsteps <- (range / steps - 1)
  if (is.finite(nsteps)) {
    for (i in 1:nsteps) {
      midrange <- c(midrange, i * steps + threshold)
    }
  }

  # extract unique levels (if range is too small)
  midrange <- unique(round(midrange, 2))

  # Categorize values
  df[[values]] <- cut(df[[values]], breaks = c(-Inf, midrange, Inf),
                      right = right_intv)

  if (invert == 0) {
    disc_cols <- rev(warn_cols)
    names(disc_cols) <- levels(df[[values]])
  } else {
    disc_cols <- warn_cols
    names(disc_cols) <- levels(df[[values]])
  }

  # allocate factor variables to x and y axis according to number of levels
  if (length(cat_vars) == 2) {
    df[[cat_vars[1]]] <- factor(df[[cat_vars[1]]])
    df[[cat_vars[2]]] <- factor(df[[cat_vars[2]]])

    A <- length(levels(df[[cat_vars[1]]]))
    B <- length(levels(df[[cat_vars[2]]]))

    if (A < B) {
      df$x <- df[[cat_vars[1]]]
      df$y <- df[[cat_vars[2]]]
      namex <- cat_vars[1]
      namey <- cat_vars[2]
    } else {
      df$x <- df[[cat_vars[2]]]
      df$y <- df[[cat_vars[1]]]
      namex <- cat_vars[2]
      namey <- cat_vars[1]
    }

    if (!missing(strata)) {
      p <- ggplot(df, aes(x, y, fill = .data[[values]])) +
        facet_grid(.data[[strata]] ~ .) + # TODO: test ~
        geom_tile(colour = "white", linewidth = 0.8) + # https://github.com/tidyverse/ggplot2/issues/5051
        geom_text(label = paste0(round(df$z2, 2), " %")) +
        scale_fill_manual(values = disc_cols, name = " ") +
        theme_minimal() +
        theme(axis.text.x = element_text(angle = 30, hjust = 1)) +
        scale_x_discrete(name = namex) +
        scale_y_discrete(expand = c(0, 0), name = namey,
                         limits = rev(levels(df$y))) +
        xlab("Study segments")
    } else {
      p <- ggplot(df, aes(x, y, fill = .data[[values]])) +
        geom_tile(colour = "white", linewidth = 0.8) + # https://github.com/tidyverse/ggplot2/issues/5051
        geom_text(label = paste0(round(df$z2, 2), " %")) +
        scale_fill_manual(values = disc_cols, name = " ") +
        theme_minimal() +
        theme(axis.text.x = element_text(angle = 30, hjust = 1)) +
        scale_x_discrete(name = namex) +
        scale_y_discrete(expand = c(0, 0), name = namey,
                         limits = rev(levels(df$y))) +
        xlab("Study segments")
    }
  } else {
    namex <- cat_vars

    p <- ggplot(df, aes(.data[[cat_vars]], y = z2, fill = .data[[values]])) +
      geom_bar(stat = "identity", na.rm = TRUE) +
      geom_text(label = paste0(" ", round(df$z2, digits = 2), "%"),
                hjust = 0, vjust = 0.5) +
      scale_fill_manual(values = disc_cols, name = " ") +
      theme_minimal() +
      scale_x_discrete(name = namex) +
      scale_y_continuous(name = "(%)",
                         limits = c(0, 1.2 * max(df$z2))) +
      coord_flip() # TODO: use util_coord_flip util_lazy_add_coord(p, fli)
  }

  return(list(SummaryPlot = p))
}

#' Verify item-level metadata
#'
#' @description
#'
#' are the provided item-level `meta_data` plausible given `study_data`?
#'
#' @inheritParams .template_function_indicator
#'
#' @return an [invisible()] list with the entries.
#'   - `pred` [data.frame] metadata predicted from `study_data`,
#'     reduced to such metadata also available in the provided metadata
#'   - `prov` [data.frame] provided metadata,
#'     reduced to such metadata also available in the provided `study_data`
#'   - `ml_error` [character] `VAR_NAMES` of variables with potentially wrong
#'     `MISSING_LIST`
#'   - `sl_error` [character] `VAR_NAMES` of variables with potentially wrong
#'     `SCALE_LEVEL`
#'   - `dt_error` [character] `VAR_NAMES` of variables with potentially wrong
#'     `DATA_TYPE`
#' @export
prep_compare_meta_with_study <- function(study_data,
         label_col,
         item_level = "item_level",
         meta_data = item_level,
         meta_data_v2) {

  util_maybe_load_meta_data_v2()

  prep_prepare_dataframes(
    .replace_hard_limits = FALSE,
    .replace_missings = FALSE,
    .adjust_data_type = FALSE,
    .amend_scale_level = FALSE,
    .apply_factor_metadata = FALSE,
    )

  prov <- meta_data

  suppressMessages(suppressWarnings(
    pred <- prep_study2meta(study_data = study_data,
                            guess_missing_codes = TRUE)
  ))

  rownames(pred) <- pred$VAR_NAMES
  rownames(prov) <- prov$VAR_NAMES

  both <- intersect(rownames(pred), rownames(prov))

  pred <- pred[both, , FALSE]
  prov <- prov[both, , FALSE]

  dt_error <- both[pred$DATA_TYPE != prov$DATA_TYPE]
  sl_error <- both[pred$SCALE_LEVEL != prov$SCALE_LEVEL]
  ml_error <- both[pred$MISSING_LIST != prov$MISSING_LIST]

  if (length(dt_error) > 0) {
    util_warning(
      "In the item-level %s %s may have assigned the wrong %s given %s",
      sQuote("meta_data"),
      util_pretty_vector_string(dt_error),
      sQuote(DATA_TYPE),
      sQuote("study_data")
    )
  }
  if (length(sl_error) > 0) {
    util_warning(
      "In the item-level %s %s may have assigned the wrong %s given %s",
      sQuote("meta_data"),
      util_pretty_vector_string(sl_error),
      sQuote(SCALE_LEVEL),
      sQuote("study_data")
    )
  }
  if (length(ml_error) > 0) {
    util_warning(
      "In the item-level %s %s may have assigned the wrong %s given %s",
      sQuote("meta_data"),
      util_pretty_vector_string(ml_error),
      sQuote(MISSING_LIST),
      sQuote("study_data")
    )
  }
  return(invisible(list(
    pred = pred,
    prov = prov,
    ml_error = ml_error,
    sl_error = sl_error,
    dt_error = dt_error
  )))
}

#' Extract all missing/jump codes from metadata and export a
#' cause-label-data-frame
#'
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return [list] with the entries
#'   - `meta_data` [data.frame] a data frame that contains updated metadata --
#'                              you still need to add a column
#'                              [MISSING_LIST_TABLE] and add the
#'                              `cause_label_df` as such to the metadata
#'                              cache using [prep_add_data_frames()], manually.
#'   - `cause_label_df` [data.frame] missing code table. If missing codes have
#'                                    labels the respective data frame are
#'                                    specified here, see [cause_label_df].
#'
#' @export
#'
#' @seealso [prep_add_cause_label_df]
prep_extract_cause_label_df <- function(item_level = "item_level",
                                       label_col = VAR_NAMES,
                                       meta_data_v2,
                                       meta_data = item_level) {
  util_maybe_load_meta_data_v2()
  util_expect_scalar(label_col, check_type = is.character)
  util_expect_data_frame(meta_data, c(label_col, DATA_TYPE))
  if (JUMP_LIST %in% colnames(meta_data)) {
    jl <- util_parse_assignments(meta_data$JUMP_LIST,
                                 multi_variate_text = TRUE)
  } else {
    jl <- lapply(seq_len(nrow(meta_data)),
                 function(x) setNames(list(NA), NA_character_))
  }

  if (MISSING_LIST %in% colnames(meta_data)) {
    ml <- util_parse_assignments(meta_data$MISSING_LIST,
                               multi_variate_text = TRUE)
  } else {
    ml <- lapply(seq_len(nrow(meta_data)),
                 function(x) setNames(list(NA), NA_character_))
  }

  to_df <- function(l, cls) {
    cause_label_df <- mapply(l, meta_data[[label_col]],
                             meta_data[[DATA_TYPE]], FUN =
             function(x, v, tp) {
               if (!is.na(tp) && tp == DATA_TYPES$DATETIME) {
                 dfr <- data.frame(CODE_VALUE =
                                     as.character(
                                     suppressWarnings(util_parse_date(
                                       names(x)))),
                                   CODE_LABEL = unname(unlist(x)))
                 prfx <- as.character(dfr$CODE_LABEL) ==
                   as.character(dfr$CODE_VALUE)
                 prfx <- prfx[!is.na(prfx)]
                 if (any(prfx, na.rm = TRUE)) {
                   dfr[prfx, CODE_LABEL] <-
                     paste(cls, dfr[prfx, CODE_LABEL])
                 }
                 dfr$CODE_CLASS <- rep(cls, nrow(dfr))
                 dfr$resp_vars <- rep(v, nrow(dfr))
                 dfr[!is.na(dfr$CODE_VALUE), , FALSE]
               } else if (!is.na(tp) && tp == DATA_TYPES$TIME) {
                 dfr <- data.frame(CODE_VALUE =
                                     as.character(
                                       suppressWarnings(util_parse_time(
                                         names(x)))),
                                   CODE_LABEL = unname(unlist(x)))
                 prfx <- as.character(dfr$CODE_LABEL) ==
                   as.character(dfr$CODE_VALUE)
                 prfx <- prfx[!is.na(prfx)]
                 if (any(prfx, na.rm = TRUE)) {
                   dfr[prfx, CODE_LABEL] <-
                     paste(cls, dfr[prfx, CODE_LABEL])
                 }
                 dfr$CODE_CLASS <- rep(cls, nrow(dfr))
                 dfr$resp_vars <- rep(v, nrow(dfr))
                 dfr[!is.na(dfr$CODE_VALUE), , FALSE]
               } else {
                 dfr <- data.frame(CODE_VALUE =
                                     as.character(
                                     suppressWarnings(as.numeric(names(x)))),
                                   CODE_LABEL = unname(unlist(x)))
                 prfx <- dfr$CODE_LABEL == dfr$CODE_VALUE
                 prfx <- prfx[!is.na(prfx)]
                 if (any(prfx, na.rm = TRUE)) {
                   dfr[prfx, CODE_LABEL] <-
                     paste(cls, dfr[prfx, CODE_LABEL])
                 }
                 dfr$CODE_CLASS <- rep(cls, nrow(dfr))
                 dfr$resp_vars <- rep(v, nrow(dfr))
                 dfr[!is.na(dfr$CODE_VALUE), , FALSE]
               }
             }, SIMPLIFY = FALSE)
    do.call(rbind.data.frame, cause_label_df)
  }

  cause_label_df <- rbind.data.frame(to_df(jl, "JUMP"), to_df(ml, "MISSING"))

  dt <- !is.na(meta_data[[DATA_TYPE]]) &
    meta_data[[DATA_TYPE]] == DATA_TYPES$DATETIME

  to <- !is.na(meta_data[[DATA_TYPE]]) &
    meta_data[[DATA_TYPE]] == DATA_TYPES$TIME

  ml2 <- ml
  jl2 <- jl

  coll <- sprintf(" %s ", SPLIT_CHAR)
  coll_to <- function(x)
    paste(.util_format_hms(
      na.omit(suppressWarnings(util_parse_time(names(x))))),
          collapse = coll)
  coll_dt <- function(x)
    paste(na.omit(suppressWarnings(util_parse_date(names(x)))),
          collapse = coll)
  coll_ndt <- function(x)
    paste(na.omit(suppressWarnings(as.numeric(names(x)))),
          collapse = coll)

  ml_empty_on_purpose <- trimws(meta_data$MISSING_LIST) == SPLIT_CHAR
  ml2[dt] <- vapply(ml[dt], coll_dt, FUN.VALUE = character(1))
  ml2[to] <- vapply(ml[to], coll_to, FUN.VALUE = character(1))
  ml2[!dt & !to] <- vapply(ml[!dt & !to], coll_ndt, FUN.VALUE = character(1))
  ml2[trimws(ml2) == ""] <- NA
  meta_data$MISSING_LIST <- unlist(ml2, recursive = FALSE)

  jl_empty_on_purpose <- trimws(meta_data$JUMP_LIST) == SPLIT_CHAR
  jl2[dt] <- vapply(jl[dt], coll_dt, FUN.VALUE = character(1))
  jl2[to] <- vapply(jl[to], coll_to, FUN.VALUE = character(1))
  jl2[!dt & !to] <- vapply(jl[!dt & !to], coll_ndt, FUN.VALUE = character(1))
  jl2[trimws(jl2) == ""] <- NA
  meta_data$JUMP_LIST <- unlist(jl2, recursive = FALSE)

  meta_data$MISSING_LIST[is.na(meta_data$MISSING_LIST) & ml_empty_on_purpose] <-
    SPLIT_CHAR

  meta_data$JUMP_LIST[is.na(meta_data$JUMP_LIST) & jl_empty_on_purpose] <-
    SPLIT_CHAR


  return(list(
    meta_data = meta_data,
    cause_label_df = cause_label_df
  ))
}

#' Summarize missingness columnwise (in variable)
#'
#' @description
#' Item-Missingness (also referred to as item nonresponse (De Leeuw et al.
#' 2003)) describes the missingness of single values, e.g. blanks or empty data
#' cells in a data set. Item-Missingness occurs for example in case a respondent
#' does not provide information for a certain question, a question is overlooked
#' by accident, a programming failure occurs or a provided answer were missed
#' while entering the data.
#'
#' [Indicator]
#'
#' @details
#' # ALGORITHM OF THIS IMPLEMENTATION:
#'
#'  - Lists of missing codes and, if applicable, jump codes are selected from
#'    the metadata
#'  - The no. of system missings (NA) in each variable is calculated
#'  - The no. of used missing codes is calculated for each variable
#'  - The no. of used jump codes is calculated for each variable
#'  - Two result dataframes (1: on the level of observations, 2: a summary for
#'    each variable) are generated
#'  - *OPTIONAL:* if `show_causes` is selected, one summary plot for all
#'                `resp_vars` is provided
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable list] the name of the measurement variables
#' @param show_causes [logical] if TRUE, then the distribution of missing codes
#'                              is shown
#' @param cause_label_df [data.frame] missing code table. If missing codes have
#'                                    labels the respective data frame can be
#'                                    specified here or in the metadata as
#'                                    assignments, see [cause_label_df]
#' @param include_sysmiss [logical] Optional, if TRUE system missingness (NAs)
#'                                  is evaluated in the summary plot
#' @param threshold_value [numeric] from=0 to=100. a numerical value ranging
#'                                                 from 0-100
#' @param suppressWarnings [logical] warn about consistency issues with missing
#'                                   and jump lists
#' @param assume_consistent_codes [logical] if TRUE and no labels are given and
#'                                          the same missing/jump code is used
#'                                          for more than one variable, the
#'                                          labels assigned for this code are
#'                                          treated as being be the same for
#'                                          all variables.
#' @param expand_codes [logical] if TRUE, code labels are copied from other
#'                               variables, if the code is the same and the
#'                               label is set somewhere
#'@param drop_levels [logical] if TRUE, do not display unused missing codes in
#'                             the figure legend.
#'
#' @param expected_observations [enum] HIERARCHY | ALL | SEGMENT. If ALL, all
#'                                     observations are expected to comprise
#'                                     all study segments. If SEGMENT, the
#'                                     `PART_VAR` is expected to point
#'                                     to a variable with values of 0 and 1,
#'                                     indicating whether the variable was
#'                                     expected to be observed for each data
#'                                     row. If HIERARCHY, this is also
#'                                     checked recursively, so, if a variable
#'                                     points to such a participation variable,
#'                                     and that other variable does has also
#'                                     a `PART_VAR` entry pointing
#'                                     to a variable, the observation of the
#'                                     initial variable is only
#'                                     expected, if both segment variables are
#'                                     1.
#' @param pretty_print [logical] deprecated. If you want to have a human
#'                               readable output, use `SummaryData` instead
#'                               of `SummaryTable`
#'
#' @return a list with:
#'   - `SummaryTable`: data frame about item missingness per response variable
#'   - `SummaryData`: data frame about item missingness per response variable
#'                    formatted for user
#'   - `SummaryPlot`: ggplot2 heatmap plot, if show_causes was TRUE
#'   - `ReportSummaryTable`: data frame underlying `SummaryPlot`
#'
#' @export
#' @importFrom ggplot2 ggplot facet_wrap geom_bar theme_minimal theme annotate scale_fill_gradientn theme element_blank
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_com_impl_item_missingness.html
#' )
com_item_missingness  <- function(resp_vars = NULL,
                                  study_data,
                                  label_col,
                                  item_level = "item_level",
                                  show_causes = TRUE,
                                  cause_label_df,
                                  include_sysmiss = TRUE,
                                  threshold_value,
                                  suppressWarnings = FALSE,
                                  assume_consistent_codes = TRUE,
                                  expand_codes = assume_consistent_codes,
                                  drop_levels = FALSE,
                                  expected_observations = c("HIERARCHY",
                                                            "ALL",
                                                            "SEGMENT"),
                                  pretty_print = lifecycle::deprecated(),
                                  meta_data = item_level,
                                  meta_data_v2) {
  util_maybe_load_meta_data_v2()

  if (lifecycle::is_present(pretty_print)) {

    # Signal the deprecation to the user
    lifecycle::deprecate_stop("2.1.0",
                              "dataquieR::com_item_missingness(pretty_print = 'FALSE')",
                              "dataquieR::com_item_missingness()",
                              details = c("Replace as follows: dataquieR::com_item_missingness(pretty_print = FALSE)$SummaryTable --> dataquieR::com_item_missingness()$SummaryTable",
                                          "dataquieR::com_item_missingness(pretty_print = TRUE)$SummaryTable --> dataquieR::com_item_missingness()$SummaryData",
                                          "dataquieR::com_item_missingness()$SummaryTable --> dataquieR::com_item_missingness()$SummaryData"))

  }
  expected_observations_missing <- missing(expected_observations)
  util_expect_scalar(expected_observations, allow_more_than_one = TRUE)
  expected_observations <- match.arg(expected_observations)
  util_expect_scalar(expected_observations)

  util_expect_scalar(assume_consistent_codes, check_type = is.logical)
  util_expect_scalar(include_sysmiss, check_type = is.logical)
  util_expect_scalar(suppressWarnings, check_type = is.logical)
  util_expect_scalar(drop_levels, check_type = is.logical)

  if (is.character(study_data)) {
    study_data <- prep_get_data_frame(study_data)
  }

  study_data <- util_cast_off(study_data, "study_data", TRUE)
  .sd <- study_data
  prep_prepare_dataframes(.replace_missings = FALSE)
  study_data <- .sd

  if (missing(threshold_value)) {
    if (!.called_in_pipeline) util_message(
      c("The mandatory argument threshold_value was not",
        "defined and is set to the default of 90%%."),
      applicability_problem = TRUE)
    threshold_value <- 90
  }
  .threshold_value <- suppressWarnings(as.numeric(threshold_value))
  if (is.na(.threshold_value)) {
    util_message(
      c("Could not convert threshold_value %s to a number.",
        "Set to default value 90%%."),
      dQuote(as.character(threshold_value)),
      applicability_problem = TRUE
    )
    threshold_value <- 90
  } else {
    threshold_value <- .threshold_value
  }


  if (!missing(show_causes) &&
      length(show_causes) == 1 &&
      is.logical(show_causes) &&
      !show_causes) {
    util_warning(c("The argument %s has been deprecated. It will be ignored",
                   "and in a future version be removed."),
                 dQuote("show_causes"))
  }

  if (!missing(cause_label_df)) {
    util_expect_data_frame(cause_label_df, c(CODE_VALUE, CODE_LABEL))
    util_warning(c("The argument %s has been deprecated. It will be",
                   "in a future version be removed."),
                 dQuote("cause_label_df"))
    meta_data <-
      prep_add_cause_label_df(meta_data = meta_data,
                              cause_label_df = cause_label_df,
                              label_col = label_col,
                              assume_consistent_codes = assume_consistent_codes)
  }

  if (!suppressWarnings) {
    util_validate_missing_lists(meta_data = meta_data,
                                cause_label_df = cause_label_df,
                                assume_consistent_codes =
                                  assume_consistent_codes,
                                expand_codes = FALSE,
                                suppressWarnings = suppressWarnings,
                                label_col = label_col
    )
  }

  if (missing(resp_vars) || is.null(resp_vars)) {
    resp_vars <- meta_data[[label_col]]
    if (VARIABLE_ORDER %in% colnames(meta_data)) {
      try(
        resp_vars <- resp_vars[order(util_as_numeric(meta_data[[VARIABLE_ORDER]],
                                                     warn =
                                                       sprintf("%s is not fully numeric: %%s",
                                                               sQuote(VARIABLE_ORDER))))],
        silent = TRUE
      )
    }
  }

  util_correct_variable_use("resp_vars",
                            allow_more_than_one = TRUE,
                            allow_all_obs_na = TRUE,
                            allow_null = TRUE)

  if (!expected_observations_missing && expected_observations != "ALL" &&
      !(PART_VAR %in% colnames(meta_data))) {
    util_message(c("For %s = %s, a column %s is needed in %s. Falling",
                   "back to %s = %s."),
                 sQuote("expected_observations"),
                 dQuote(expected_observations),
                 dQuote(PART_VAR),
                 sQuote("meta_data"),
                 sQuote("expected_observations"),
                 dQuote("ALL"),
                 applicability_problem = TRUE
    )
    expected_observations <- "ALL"
  }

  if (expand_codes) {
    meta_data <- prep_expand_codes(meta_data,
                                   suppressWarnings,
                                   mix_jumps_and_missings = FALSE)
  }

  {
    r <- util_study_var2factor(study_data = study_data, meta_data = meta_data,
                               resp_vars = resp_vars, label_col = label_col,
                               assume_consistent_codes = assume_consistent_codes,
                               have_cause_label_df = !missing(cause_label_df),
                               code_name = MISSING_LIST,
                               include_sysmiss = FALSE)
    colnames(r) <-
      util_map_labels(colnames(r),
                      meta_data,
                      to = VAR_NAMES,
                      from = label_col)
    m <-
      vapply(lapply(util_seg_table(r, study_data,
                                   meta_data,
                                   expected_observations = expected_observations),
                    `[[`, "Freq"), sum, FUN.VALUE = integer(1))
  }
  {
    r <- util_study_var2factor(study_data = study_data, meta_data = meta_data,
                               resp_vars = resp_vars, label_col = label_col,
                               assume_consistent_codes = assume_consistent_codes,
                               have_cause_label_df = !missing(cause_label_df),
                               code_name = JUMP_LIST,
                               include_sysmiss = FALSE)
    colnames(r) <-
      util_map_labels(colnames(r),
                      meta_data,
                      to = VAR_NAMES,
                      from = label_col)
    j <-
      vapply(lapply(util_seg_table(r, study_data,
                                   meta_data,
                                   expected_observations = expected_observations),
                    `[[`, "Freq"), sum, FUN.VALUE = integer(1))
  }
  SysMiss <-
    vapply(prep_map_labels(resp_vars,
                           meta_data =  meta_data,
                           to = VAR_NAMES,
                           from = label_col),
           function(rv) {
             sum(is.na(study_data[[rv]][
               util_observation_expected(rv = rv,
                                         study_data = study_data,
                                         meta_data = meta_data,
                                         label_col = VAR_NAMES,
                                         expected_observations =
                                           expected_observations)]))
                    },
                    FUN.VALUE = integer(1))

  N_obs <-
    util_count_expected_observations(resp_vars,
                                     study_data = ds1,
                                     meta_data = meta_data,
                                     label_col = label_col,
                                     expected_observations = expected_observations)

  DV_N <- vapply(
    resp_vars,
    function(rv) {
      in_seg <-
        util_observation_expected(rv = rv,
                                  study_data = ds1,
                                  meta_data = meta_data,
                                  label_col = label_col,
                                  expected_observations = expected_observations)
      {
        vn <- util_map_labels(rv, meta_data,
                              VAR_NAMES, label_col)
        vals <- study_data[!in_seg, vn, drop = FALSE]
        vals <- util_replace_codes_by_NA(
          study_data = vals, meta_data = meta_data)[[vn]]
        have_data_not_expected <- !util_is_na_0_empty_or_false(vals)
        # also allow 0/FALSE, if no data is expected
        if (any(have_data_not_expected) && !suppressWarnings) {
          segvars <- util_all_intro_vars_for_rv(rv,
                                                ds1, meta_data,
                                                label_col,
                                                expected_observations =
                                                  expected_observations)
          util_warning(
            c("There are %d meassurements of %s for participants",
              "not being part of one of the segments %s"),
            sum(have_data_not_expected),
            dQuote(rv),
            paste(dQuote(util_map_labels(
              x = util_map_labels(segvars, meta_data, from = label_col,
                                  to = VAR_NAMES),
              meta_data = meta_data,
                                         from = PART_VAR, to = STUDY_SEGMENT)),
              paste0("(if ", dQuote(segvars), " \u2260 1)"),
                  collapse = ", "),
            applicability_problem = TRUE)
        }
      }
      sum(!is.na(study_data[in_seg, util_map_labels(rv, meta_data,
                                                    VAR_NAMES, label_col)]))
    }, FUN.VALUE = integer(1))

  N_meas <- DV_N - m -  j

  SummaryTable <- data.frame(check.names = FALSE,
                             Variables = resp_vars,
                             `Expected observations N` = N_obs,
                             `Sysmiss N` = SysMiss,
                             `Datavalues N` = DV_N,
                             `Missing codes N` = m,
                             `Jumps N` = j,
                             `Measurements N` = N_meas
  )

  rownames(SummaryTable) <- NULL

  SysMissP <- round(SysMiss / N_obs * 100, digits = 2)
  DV_N_P <- round(DV_N / N_obs * 100, digits = 2)
  m_P <- round(m / N_obs * 100, digits = 2)
  j_P <- round(j / N_obs * 100, digits = 2)
  N_meas_P <- round(N_meas / (N_obs - j) * 100, digits = 2)
  Missing_expected_obs<- SysMiss+j+m

  SummaryTable$Missing_expected_obs<- Missing_expected_obs
  Missing_expected_obs_P <- round((SysMiss+j+m)/N_obs*100, digits = 2)
  PCT_com_crm_mv <- round((SysMiss+j+m)/ nrow(ds1)* 100, digits = 2)
  SummaryTable$PCT_com_crm_mv<- PCT_com_crm_mv
  SummaryTable$GRADING <- ifelse(N_meas_P < threshold_value, 1, 0)

  #Create SummaryData
  SummaryData<-SummaryTable[, !names(SummaryTable) %in% c("GRADING")]

  names(SummaryData)[names(SummaryData) == "Observations N"] <-
    "Expected observations N"

  names(SummaryData)[names(SummaryData) == "Sysmiss N"] <-
    "Sysmiss N (%)"
  SummaryData$`Sysmiss N (%)`<- paste0(SysMiss, " (", SysMissP, ")")

  names(SummaryData)[names(SummaryData) == "Datavalues N"] <-
    "Datavalues N (%)"
  SummaryData$`Datavalues N (%)`<- paste0(DV_N, " (", DV_N_P, ")")

  names(SummaryData)[names(SummaryData) == "Missing codes N"] <-
    "Missing codes N (%)"
  SummaryData$`Missing codes N (%)`<- paste0(m, " (", m_P, ")")

  names(SummaryData)[names(SummaryData) == "Jumps N"] <- "Jumps N (%)"
  SummaryData$`Jumps N (%)`<- paste0(j, " (", j_P, ")")

  names(SummaryData)[names(SummaryData) == "Measurements N"] <-
    "Measurements N (%)"
  SummaryData$`Measurements N (%)`<- paste0(N_meas, " (", N_meas_P, ")")

  names(SummaryData)[names(SummaryData) == "Missing_expected_obs"] <-
    "Missing expected obs. N (%)"
  SummaryData$`Missing expected obs. N (%)`<- paste0(Missing_expected_obs,
                                                     " (",
                                                     Missing_expected_obs_P, ")")

  names(SummaryData)[names(SummaryData) == "PCT_com_crm_mv"] <-
    "Crude missingness N (%)"
  SummaryData$`Crude missingness N (%)`<- paste0(SysMiss+j+m, " (",
                                                 PCT_com_crm_mv, ")")


  # to add formulas in hover text:
  # create a named vector with names = column names of the SummaryData and
  # values = what you want to be displayed in the hover text
  text_to_display <- util_get_hovertext("[com_item_missingness_hover]")


  #  result_table$check_MC <- paste0(N_class_MC, " (", N_MC_avail, ")")
  #  result_table$check_JC <- paste0(N_class_JC, " (", N_JC_avail, ")")

  r <- util_study_var2factor(study_data = study_data, meta_data = meta_data,
                             resp_vars = resp_vars, label_col = label_col,
                             assume_consistent_codes = assume_consistent_codes,
                             have_cause_label_df = !missing(cause_label_df),
                             include_sysmiss = include_sysmiss)
  colnames(r) <-
    util_map_labels(colnames(r),
                    meta_data,
                    to = VAR_NAMES,
                    from = label_col)
  r <- util_seg_table(r, study_data, meta_data,
                      expected_observations = expected_observations)
  r <- r[vapply(r, ncol, FUN.VALUE = integer(1)) == 2]
  r <- lapply(names(r), function(nm) setNames(r[[nm]], c("CODES", nm)))
  mctab <- Reduce(function(dtf1, dtf2) merge(dtf1, dtf2, by = "CODES",
                                             all = TRUE), r)
  rownames(mctab) <- mctab$CODES
  mctab$CODES <- NULL
  mctab <- t(mctab)
  mctab[is.na(mctab)] <- 0
  mctab <- data.frame(mctab, check.names = FALSE)
  if (drop_levels) {
    mctab <- mctab[, colSums(mctab) != 0, drop = FALSE]
  }
  if (!ncol(mctab)) {
    mctab <- data.frame(check.names = FALSE,
                        Variables = resp_vars,
                        stringsAsFactors = FALSE)
  } else {
    mctab$Variables <- prep_map_labels(rownames(mctab),
                                       meta_data = meta_data,
                                       to = label_col,
                                       from = VAR_NAMES)
  }
  rownames(mctab) <- NULL
  mctab$N <-
    util_count_expected_observations(mctab$Variables,
                                     study_data = ds1,
                                     meta_data = meta_data,
                                     label_col = label_col,
                                     expected_observations = expected_observations)

  mctab <- util_validate_report_summary_table(mctab,
                                              meta_data = meta_data,
                                              label_col = label_col)

  attr(SummaryData, "description") <- text_to_display

  list(SummaryTable = SummaryTable,
       SummaryData = SummaryData,
       SummaryPlot = print(util_attach_attr(mctab, relative = FALSE),
                           view = FALSE),
       ReportSummaryTable = mctab)
}

#' utility function applicability of segment missingness
#'
#' Test for applicability of segment missingness
#'
#' @param x [data.frame] metadata
#' @param dta [logical] vector, 1=matching data type,
#'                  0 = non-matching data type
#'
#' @return [factor] 0-3 for each variable in metadata
#' - 0 data type mismatch and not applicable
#' - 1 data type mismatches but applicable
#' - 2 data type matches but not applicable
#' - 3 data type matches and applicable
#'
#' @seealso [pro_applicability_matrix]
#' @noRd

util_app_sm <- function(x, dta) {
  if (all(c(PART_VAR, STUDY_SEGMENT) %in% names(x))) {
    c1 <- ifelse(is.na(x[[STUDY_SEGMENT]]) |
                   is.na(x[[PART_VAR]]), 0, 1)
  } else {
    c1 <- rep(0, times = dim(x)[1])
  }

  aa <- paste0(dta, c1)
  score <- as.numeric(recode(as.factor(aa), "00" = 0, "01" = 1,
                             "10" = 2, "11" = 3))
  score <- as.factor(score)
  return(score)
}

#' Put in back-ticks
#'
#' also escape potential back-ticks in x
#'
#' @param x a string
#'
#' @return x in back-ticks
#'
#' @seealso `util_backtickQuote`
#' @family process_functions
#' @concept data_management
#' @noRd
util_bQuote <- function(x) {
  if (length(x) == 0)
    return(character(0))
  nna <- !is.na(x)
  x[nna] <- gsub("\\", "\\\\", x[nna], fixed = TRUE)
  x[nna] <- gsub("`", "\\`", x[nna], fixed = TRUE)
  x[nna] <- paste0("`", x[nna], "`")
  x
}

#' Rename columns of a `SummaryTable` (or Segment, ...) to look nice
#'
#' @param Table [data.frame], a table
#'
#' @return renamed table
#'
#' @family reporting_functions
#' @concept summary
#' @noRd
util_make_data_slot_from_table_slot <- function(Table) {

  # 1) Select indicator metric columns via helper
  indicator_cols <- util_extract_indicator_metrics(Table)

  # 2) Choose base/technical columns (if present) and build final column order
  base_candidates <- c("Variables", "Segment", "DF_NAME", "CHECK_LABEL",
                       "Dataframe", "CONTRADICTION_TYPE")
  base_keep <- intersect(base_candidates, colnames(Table))

  select_names <- c(base_keep, colnames(indicator_cols))
  # Ensure unique, in case of overlaps
  select_names <- unique(select_names)

  # If nothing to select, return an empty DataSlot with preserved classes
  Data <- Table[, select_names, drop = FALSE]

  # 3) Prepare percent-column mask based on ORIGINAL names (before renaming)
  pct_mask <- startsWith(select_names, "PCT_")

  # 4) Translate headers with the helper (keep unknowns untouched)
  translated <- util_translate_indicator_metrics(
    select_names,
    short = FALSE,
    long  = TRUE,
    ignore_unknown = TRUE
  )

  # 5) Apply manual recodes for known technical columns
  manual_map <- c(
    Variables           = "Variables",
    Segment             = "Segment",
    DF_NAME             = "Dataframe",
    CHECK_LABEL         = "Check",
    Dataframe           = "Dataframe",
    CONTRADICTION_TYPE  = "Contradiction Type"
  )
  # Overwrite translated labels where we have explicit recodes
  idx_manual <- match(names(manual_map), select_names, nomatch = 0L)
  if (any(idx_manual > 0L)) {
    translated[idx_manual[idx_manual > 0L]] <-
      unname(manual_map[names(manual_map)[idx_manual > 0L]])
  }

  # 6) Rename columns
  colnames(Data) <- translated

  # 7) Format percentage columns by ORIGINAL-name mask (same positions)
  if (any(pct_mask)) {
    Data[, pct_mask] <- lapply(Data[, pct_mask, drop = FALSE], function(cl) {
      if (length(cl) == 0 || all(util_empty(cl))) {
        NA_character_
      } else {
        util_paste0_with_na(round(cl, 2), "%")
      }
    })
  }

  # 8) Adjust class tags
  if (inherits(Data, "TableSlot")) {
    class(Data) <- setdiff(class(Data), "TableSlot")
  }
  if (!inherits(Data, "DataSlot")) {
    class(Data) <- union("DataSlot", class(Data))
  }

  Data
}

util_ggplot_text <- function(text,
                        fontsize = 12,
                        fontfamily = "",
                        hjust = 0.5,
                        vjust = 0.5) {
  text <- paste(text, collapse = "\n");
  if (suppressWarnings(util_ensure_suggested("cli", err = FALSE)))
    text <- cli::ansi_strip(text)

  ggplot() +
    ggplot2::theme_void() +
    annotate("text",
             x = 0.5, y = 0.5,
             label = text,
             size = fontsize / ggplot2::.pt,  # ggplot2 fontsize uses mm, convert from pt
             family = fontfamily,
             hjust = hjust,
             vjust = vjust
    ) +
    xlim(0, 1) + ggplot2::ylim(0, 1)
}

#' Check for one value only
#'
#' utility function to identify variables with one value only.
#'
#' @param x vector with values
#'
#' @return logical(1): TRUE, if -- except NA -- exactly only one value
#'                           is observed in `x`,
#'                     FALSE otherwise
#'
#' @family robustness_functions
#' @concept data_management
#' @noRd


util_check_one_unique_value <- function(x) {
  howmany <- length(unique(x[!is.na(x)]))
  return(as.logical(ifelse(howmany == 1, 1, 0)))
}

test_that("util_check_one_unique_value works", {
  skip_on_cran()
  expect_false(util_check_one_unique_value(1:2))
  expect_false(util_check_one_unique_value(c(1:2, NA)))
  expect_true(util_check_one_unique_value(c(1, NA)))
  expect_false(util_check_one_unique_value(c(NA, NA)))
})

#' Detect base unit from composite units
#'
#' @param unit [character] a unit
#' @param warn_ambiguities [logical] warn about all ambiguous units
#' @param unique [logical] choose the more `SI`-like unit in case of ambiguities
#'
#' @return [character] all possible or the preferable (unique set `TRUE`)
#'                     base units. Can be `character(0)`, if unit is invalid
#'                     or uniqueness was requested, but even using precedence
#'                     rules of `SI`-closeness do not help selecting the most
#'                     suitable unit.
#'
#' @noRd
#'
#' @examples
#' \dontrun{
#' util_unit2baseunit("%")
#' util_unit2baseunit("d%")
#'
#' # Invalid unit
#' util_unit2baseunit("aa%")
#' util_unit2baseunit("aa%", unique = FALSE)
#'
#' util_unit2baseunit("a%")
#'
#' # Invalid unit
#' util_unit2baseunit("e%")
#' util_unit2baseunit("e%", unique = FALSE)
#'
#' util_unit2baseunit("E%")
#' util_unit2baseunit("Eg")
#'
#' # Invalid unit
#' util_unit2baseunit("E")
#' util_unit2baseunit("E", unique = FALSE)
#'
#' util_unit2baseunit("EC")
#' util_unit2baseunit("EK")
#' util_unit2baseunit("µg")
#' util_unit2baseunit("mg")
#' util_unit2baseunit("°C")
#' util_unit2baseunit("k°C")
#' util_unit2baseunit("kK")
#' util_unit2baseunit("nK")
#'
#' # Ambiguous units, if used with unique = FALSE
#' util_unit2baseunit("kg")
#' util_unit2baseunit("cd")
#' util_unit2baseunit("Pa")
#' util_unit2baseunit("kat")
#' util_unit2baseunit("min")
#'
#' # atto atom units or astronomical units, both in state "accepted"
#' util_unit2baseunit("au")
#' util_unit2baseunit("au", unique = FALSE)
#'
#' # astronomical units or micro are, both in state "accepted"
#' util_unit2baseunit("ua")
#' util_unit2baseunit("ua", unique = FALSE)
#'
#' util_unit2baseunit("kt")
#'
#' # parts per trillion or pico US_liquid_pint, both in state "common",
#' # but in this case, plain count units will be preferred
#' util_unit2baseunit("ppt")
#' util_unit2baseunit("ppt", unique = FALSE)
#'
#' util_unit2baseunit("ft")
#' util_unit2baseunit("yd")
#' util_unit2baseunit("pt")
#'
#' # actually the same, but both only common, and to my knowledge not-so-common
#' # gram-force vs. kilogram-force (kilo pond)
#' util_unit2baseunit("kgf")
#' util_unit2baseunit("kgf", unique = FALSE)
#'
#' util_unit2baseunit("at")
#' util_unit2baseunit("ph")
#' util_unit2baseunit("nt")
#' }
util_unit2baseunit <- function(unit,
                               warn_ambiguities = !exists("warn_ambiguities",
                                                         .unit2baseunitenv),
                               unique = TRUE) {
  if (warn_ambiguities) {
    assign("warn_ambiguities", FALSE, .unit2baseunitenv)
    problems <- vapply(lapply(setNames(nm = UNITS), util_unit2baseunit,
                              warn_ambiguities = FALSE, unique = FALSE),
           length, FUN.VALUE = integer(1)) != 1
    if (any(problems)) {
      util_message(
        "Found ambiguous units (could also be prefix + some other unit): %s",
        paste0(dQuote(names(problems[problems])), collapse = ", "))
    }
  }
  # all_units <-
  #   apply(expand.grid(UNIT_PREFIXES, UNITS), 1, paste0, collapse = "")
  myUNIT_PREFIXES <- c(UNIT_PREFIXES, "")
  mp <- lapply(setNames(nm = UNITS),
         FUN = function(suf, pre) {
             paste0(pre, suf)
         },
         myUNIT_PREFIXES)
  base_unit <- vapply(mp, function(block) {
    any(block == unit)
  }, FUN.VALUE = logical(1))
  base_unit <- which(base_unit)
  r <- names(mp)[base_unit]
  allowed_set <- c("base", "derived", "accepted", "common")
  orig_r <- r
  while (unique && length(r) > 1 && length(allowed_set) > 0) {
    orig_r <- r
    r <- r[UNIT_SOURCES[r] %in% allowed_set]
    allowed_set <- head(allowed_set, -1)
  }
  if (unique && length(r) == 0 &&
      length(intersect(orig_r, UNIT_IS_COUNT)) == 1) {
    r <- intersect(orig_r, UNIT_IS_COUNT)
  }
  r
}
.unit2baseunitenv <- new.env(parent = emptyenv())

test_that("util_unit2baseunit works", {
  skip_on_cran() # OS dependent

  suppressMessages(util_unit2baseunit("%"))

  expect_equal(util_unit2baseunit("%"), "%")
  expect_equal(util_unit2baseunit("d%"), "%")

  # Invalid unit
  expect_length(util_unit2baseunit("aa%"), 0)
  expect_length(util_unit2baseunit("aa%", unique = FALSE), 0)

  expect_equal(util_unit2baseunit("a%"), "%")

  # Invalid unit
  expect_length(util_unit2baseunit("e%"), 0)
  expect_length(util_unit2baseunit("e%", unique = FALSE), 0)

  expect_equal(util_unit2baseunit("E%"), "%")
  expect_equal(util_unit2baseunit("Eg"), "g")

  # Invalid unit
  expect_length(util_unit2baseunit("E"), 0)
  expect_length(util_unit2baseunit("E", unique = FALSE), 0)

  expect_equal(util_unit2baseunit("EC"), "C")
  expect_equal(util_unit2baseunit("EK"), "K")
  expect_equal(util_unit2baseunit("µg"), "g")
  expect_equal(util_unit2baseunit("mg"), "g")
  expect_equal(util_unit2baseunit("°C"), "°C")
  expect_equal(util_unit2baseunit("k°C"), "°C")
  expect_equal(util_unit2baseunit("kK"), "K")
  expect_equal(util_unit2baseunit("nK"), "K")

  # Ambiguous units, if used with unique = FALSE
  expect_equal(util_unit2baseunit("kg"), "kg")
  expect_equal(util_unit2baseunit("cd"), "cd")
  expect_equal(util_unit2baseunit("Pa"), "Pa")
  expect_equal(util_unit2baseunit("kat"), "kat")
  expect_equal(util_unit2baseunit("min"), "min")

  # atto atom units or astronomical units, both in state "accepted"
  expect_length(util_unit2baseunit("au"), 0)
  expect_equal(util_unit2baseunit("au", unique = FALSE), c("u", "au"))

  # astronomical units or micro are, both in state "accepted"
  expect_length(util_unit2baseunit("ua"), 0)
  expect_equal(util_unit2baseunit("ua", unique = FALSE), c("ua", "a"))

  expect_equal(util_unit2baseunit("kt"), "t")

  # parts per trillion or pico US_liquid_pint, both in state "common",
  # but in this case, plain count units will be preferred
  expect_equal(util_unit2baseunit("ppt"), "ppt")
  expect_equal(util_unit2baseunit("ppt", unique = FALSE), c("ppt", "pt"))

  expect_equal(  util_unit2baseunit("ft"), "t")
  expect_equal(util_unit2baseunit("yd"), "d")
  expect_equal(util_unit2baseunit("pt"), "t")

  # actually the same, but both only common, and to my knowledge not-so-common
  # gram-force vs. kilogram-force (kilo pond)
  expect_length(util_unit2baseunit("kgf"), 0)
  expect_equal(util_unit2baseunit("kgf", unique = FALSE), c("kgf", "gf"))

  expect_equal(util_unit2baseunit("at"), "t")
  expect_equal(util_unit2baseunit("ph"), "h")
  expect_equal(util_unit2baseunit("nt"), "t")
})

#' Utility function to plot a combined figure for distribution checks
#'
#' @description
#' Data quality indicator checks "Unexpected location" with histograms and
#' plots of empirical cumulative distributions for the subgroups.
#'
#' @export
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable list] the name of the measurement variable
#' @param group_vars [variable list] the name of the observer, device or
#'                                   reader variable
#' @param n_group_max maximum number of categories to be displayed individually
#'                  for the grouping variable (`group_vars`, devices / examiners)
#' @param n_obs_per_group_min minimum number of data points per group to create
#'                  a graph for an individual category of the `group_vars` variable
#'
#' @return A `SummaryPlot`.
#'
#' @importFrom ggplot2 ggplot_build xlim
prep_acc_distributions_with_ecdf <- function(resp_vars = NULL,
                                             group_vars = NULL,
                                             study_data,
                                             label_col,
                                             item_level = "item_level",
                                             meta_data = item_level,
                                             meta_data_v2,
                                             n_group_max = getOption("dataquieR.max_group_var_levels_in_plot", dataquieR.max_group_var_levels_in_plot_default),
                                             n_obs_per_group_min = getOption("dataquieR.min_obs_per_group_var_in_plot", dataquieR.min_obs_per_group_var_in_plot_default)) {
  # preps ----------------------------------------------------------------------
  util_maybe_load_meta_data_v2()
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          #.apply_factor_metadata = TRUE, # can be omitted in favor of .apply_factor_metadata_inadm
                          .apply_factor_metadata_inadm = TRUE
  )

  util_correct_variable_use(resp_vars,
                            allow_any_obs_na = TRUE,
                            allow_all_obs_na = FALSE,
                            min_distinct_values = 10,
                            need_scale = "interval | ratio"
  )

  util_correct_variable_use(group_vars,
                            allow_more_than_one = FALSE,
                            allow_any_obs_na = TRUE,
                            allow_all_obs_na = FALSE,
                            min_distinct_values = 2,
                            need_type = "!float",
                            need_scale = "nominal | ordinal"
  )

  util_expect_scalar(n_group_max,
                     check_type = util_is_numeric_in(min = 2,
                                                     whole_num = TRUE,
                                                     finite = TRUE))

  util_expect_scalar(n_obs_per_group_min,
                     check_type = util_is_numeric_in(min = 0,
                                                     whole_num = TRUE,
                                                     finite = TRUE))

  part1 <- acc_distributions(resp_vars = resp_vars,
                             study_data = study_data,
                             meta_data = meta_data,
                             label_col = label_col,
                             flip_mode = "noflip")
  p1 <- part1$SummaryPlotList[[resp_vars]]
  built   <- ggplot2::ggplot_build(p1)
  layout  <- util_gg_get(built, "layout")
  scalesx <- util_gg_get(layout, "panel_scales_x")[[1]]
  rngobj  <- util_gg_get(scalesx, "range")
  xx      <- util_gg_get(rngobj, "range")
  xx_exp <- xx
  xx_exp[1] <- xx_exp[1] - 0.05 * (xx_exp[2] - xx_exp[1])
  xx_exp[2] <- xx_exp[2] + 0.05 * (xx_exp[2] - xx_exp[1])
  suppressMessages(
    p1 <- p1 %lean+% util_create_lean_ggplot(
      xlim(xx_exp),
      xx_exp = xx_exp
    )
  )
  part2 <- acc_distributions_ecdf(resp_vars = resp_vars, group_vars = group_vars,
                                  study_data = study_data,
                                  label_col = label_col,
                                  meta_data = meta_data,
                                  n_group_max = n_group_max,
                                  n_obs_per_group_min = n_obs_per_group_min)
  p2prep <- part2$SummaryPlotList[[resp_vars]]
  suppressMessages(
    p2 <- p2prep + util_create_lean_ggplot(
        xlim(xx_exp),
      xx_exp = xx_exp
    )
  )

  P <-
    p1 %lean+% p2 %lean+%
      plot_layout(ncol = 1) %lean+%
      plot_annotation(tag_levels = 'A')

  return(list(SummaryPlot = util_set_size(P)))
}

#' @family plotly_shims
#' @concept plotly_shims
#' @noRd
util_as_plotly_prep_acc_distributions_with_ecdf <- function(res, ...) {
  if (length(res$SummaryPlot) != 1) {
    nres <- length(res$SummaryPlot)
    return(util_ggplotly(util_create_lean_ggplot(ggplot2::ggplot() +
                                                      ggplot2::annotate("text", x = 0, y = 0,
                                                                        label =
                                                                          sprintf(paste("Internal error: I should  have exactly 1 result, if",
                                                                                        "calling plotly for a dq_report2 otuput. I have %d."),
                                                                                  nres)) +
                                                      theme(
                                                        axis.line = element_blank(),
                                                        axis.text.x = element_blank(),
                                                        axis.text.y = element_blank(),
                                                        axis.ticks = element_blank(),
                                                        axis.title.x = element_blank(),
                                                        axis.title.y = element_blank(),
                                                        legend.position = "none",
                                                        panel.background = element_blank(),
                                                        panel.border = element_blank(),
                                                        panel.grid.major = element_blank(),
                                                        panel.grid.minor = element_blank(),
                                                        plot.background = element_blank()
                                                      ),
                                                    nres = nres)))
  }
  res$SummaryPlot <-
    util_remove_dataquieR_result_class(res$SummaryPlot)
  # use res$SummaryPlot, not something from the enclosure
  # of the result, that may contain study data.
  util_ensure_suggested("plotly")

  if (inherits(res$SummaryPlot, "patchwork")) {
    py1 <- try(util_ggplotly(res$SummaryPlot[[1]],
                                ...), silent = TRUE)
    py2 <- try(util_ggplotly(res$SummaryPlot[[2]],
                                ...), silent = TRUE)
    util_stop_if_not(!inherits(py1, "try-error"))
    util_stop_if_not(!inherits(py2, "try-error"))
    # https://plotly.com/r/subplots/#subplots-with-shared-yaxes
    plotly::subplot(
      plotly::add_annotations( # https://stackoverflow.com/a/59191142
        py1,
        text = "A",
        x = 0,
        y = 1,
        yref = "paper",
        xref = "paper",
        xanchor = "left",
        yanchor = "top",
        yshift = 20,
        showarrow = FALSE,
        font = list(size = 15)
      ),
      plotly::add_annotations( # https://stackoverflow.com/a/59191142
        py2,
        text = "B",
        x = 0,
        y = 1,
        yref = "paper",
        xref = "paper",
        xanchor = "left",
        yanchor = "top",
        yshift = 20,
        showarrow = FALSE,
        font = list(size = 15)
      ),
      nrows = 2,
      shareX = TRUE)
  } else {
    py <- util_ggplotly(res$SummaryPlot, ...)
    plotly::layout(py, xaxis = list(tickangle = "auto"))
  }
}

#' Get value labels from data
#'
#' Detects factors and converts them to compatible metadata/study data.
#'
#' @param resp_vars [variable] names of the variables to fetch the value labels
#'                             from the data
#' @param study_data [data.frame] the data frame that contains the measurements
#'
#' @return a [list] with:
#'   - `VALUE_LABELS`: vector of value labels and modified study data
#'   - `ModifiedStudyData`: study data with factors as integers
#'
#' @export
#' @importFrom stats setNames
#'
#' @examples
#' \dontrun{
#' dataquieR::prep_datatype_from_data(iris)
#' }
prep_valuelabels_from_data <-
  function(resp_vars = colnames(study_data), study_data) {
  if (!missing(resp_vars) && is.data.frame(resp_vars) && missing(study_data)) {
    study_data <- resp_vars
    resp_vars <- colnames(study_data)
  } else if (missing(study_data) || !is.data.frame(study_data)) {
    util_error("Need study data as a data frame in the argument %s",
               dQuote("study_data"))
  }

  util_expect_data_frame(study_data, keep_types = TRUE)

  study_data <- util_cast_off(study_data, "study_data")

  if ((length(study_data) == 0) || !is.character(resp_vars)) {
    util_error(
      "%s should be missing or give variable names referring the study_data.",
      dQuote("resp_vars"), applicability_problem = TRUE)
  }

  if (!(all(resp_vars %in% colnames(study_data)))) {
    util_message(c(
      "The following %s are missing from the %s.",
      "Won't return a type for them: %s"),
      dQuote("resp_vars"),
      dQuote("study_data"),
      sQuote(resp_vars[!(resp_vars %in% colnames(study_data))]),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE
    )
  }

  factor_vars <-
    colnames(study_data)[vapply(study_data, is.factor, FUN.VALUE = logical(1))]

  factor_resp_vars <- intersect(factor_vars, resp_vars)

  valuelabels <- vapply(setNames(nm = resp_vars),
                        FUN.VALUE = character(1),
                  function(variable) {
      if (variable %in% factor_resp_vars) {
        fctr <- factor(levels = levels(study_data[[variable]]),
                       x = levels(study_data[[variable]]))
        if (is.ordered(study_data[[variable]])) {
          sc <- "<"
        } else {
          sc <- SPLIT_CHAR
        }
        paste(as.integer(fctr), "=", levels(fctr)[as.integer(fctr)],
              collapse = sprintf(" %s ", sc))
      } else {
        NA_character_
      }
  })

  modifiedstudydata <- study_data
  modifiedstudydata[, factor_resp_vars] <-
    lapply(study_data[, factor_resp_vars, FALSE], as.integer)

  return(list(
    VALUE_LABELS = valuelabels, # TODO: create VALUE_LABEL_TABLE, instead
    ModifiedStudyData = modifiedstudydata
  ))
}

#' Rotate 1-row data frames to key-value data frames
#'
#' if `nrow(tb) > 1`, `util_table_rotator` just returns `tb`.
#'
#' @param tb [data.frame] a data frame
#'
#' @return [data.frame] but transposed
#'
#' @noRd
util_table_rotator <- function(tb) {  #TODO: should take care of the description attribute
  util_expect_data_frame(tb)
  if (nrow(tb) == 1) {
    is_html_escaped <- attr(tb, "is_html_escaped")
    tb[] <- lapply(tb, FUN = as.character)
    util_attach_attr(data.frame(
      check.names = FALSE,
      fix.empty.names = FALSE,
      row.names = NULL,
      ` ` = colnames(tb),
      ` ` = unlist(tb[1, , drop = TRUE])
    ), is_html_escaped = is_html_escaped, kv_table = TRUE)
  } else {
    tb
  }
}

#' Plots simple HTML tables with background color scale
#'
#' @param tb [data.frame] the table as [data.frame] with mostly numbers
#' @param min_val [numeric] minimum value for the numbers in `tb`
#' @param max_val [numeric] maximum value for the numbers in `tb`
#' @param min_color [numeric] vector with the RGB color values for the
#'                            minimum color, values between 0 and 255
#' @param max_color [numeric] vector with the RGB color values for the
#'                            maximum color, values between 0 and 255
#' @param soften [function] to be applied to the relative values between 0
#'                          and 1 before mapping them to a color
#' @param style_header [character] to be applied to style the HTML header of
#'                                 the table
#' @param text_color_mode [enum] bw | gs. Should the text be displayed in black
#'                               and white or using a grey scale? In both cases,
#'                               the color will be adapted to the background.
#' @param hover_texts [data.frame] if not `NULL`, this data frame contains html
#'                                 code displayed when the user's mouse pointer
#'                                 moves inside corresponding cells from
#'                                 `tb`. Can contain `HTML` code.
#' @param escape_all_content [logical] if `TRUE`, treat `tb` and `hover_texts`
#'                                     using some `HTML` escaping function
#'
#' @seealso `util_html_table()`
#'
#' @return `htmltools` compatible object
#' @examples
#' \dontrun{
#'
#' tb <- as.data.frame(matrix(ncol = 5, nrow = 5))
#' tb[] <- sample(1:100, prod(dim(tb)), replace = TRUE)
#' tb[, 1] <- paste("case", 1:nrow(tb))
#' htmltools::browsable(util_formattable(tb))
#' htmltools::browsable(util_formattable(tb[, -1]))
#' }
#'
#' @noRd
util_formattable <- function(tb,
                             min_val = min(tb, na.rm = TRUE),
                             max_val = max(tb, na.rm = TRUE),
                             min_color = c(0, 0, 255),
                             max_color = c(255, 0, 0),
                             soften =
                               function(x) stats::plogis(x,
                                                  location = 0.5,
                                                  scale = 0.1),
                             style_header = "font-weight: bold;",
                             text_color_mode = c("bw", "gs"),
                             hover_texts = NULL,
                             escape_all_content = TRUE) {
  util_ensure_suggested("htmltools", "create colored tables")
  text_color_mode <- match.arg(text_color_mode)
  tb_val <- as.matrix(as.data.frame(suppressWarnings(lapply(tb, as.numeric))))
  if (missing(min_val)) {
    min_val <- min(tb_val, na.rm = TRUE)
  }
  if (missing(max_val)) {
    max_val <- max(tb_val, na.rm = TRUE)
  }
  rel_points <- ((tb_val - min_val) / abs(max_val - min_val))
  colors <- apply(soften(rel_points),
                  1:2,
                  function(x)
                    setNames((min_color + x * (max_color - min_color)) / 255,
                             c("red", "green", "blue")),
                  simplify = FALSE)
  htmltools::withTags(table(
    if (length(style_header) == ncol(tb)) {
      # style defined for each entry
      tr(lapply(seq_along(colnames(tb)), function(cc) {
        th(style = style_header[cc], colnames(tb)[cc])
      }))
    } else {
      if (length(style_header) > 1) {
        # style not well defined
        style_header <- style_header[1]
      }
      tr(lapply(colnames(tb), function(header) {
        th(style = style_header, header)
      }))
    },
    lapply(1:nrow(tb),
           function(rw)
             tr(
               lapply(1:ncol(tb),
                      function(cl) {
                        color <- try(do.call(rgb, as.list(colors[rw, cl][[1]])),
                                     silent = TRUE)
                        if (inherits(color, "try-error")) {
                          color <- "white"#"#bbbbbb"
                        }
                        if (text_color_mode == "gs") {
                          # invert
                          txtcolor <- 1 - colors[rw, cl][[1]]
                          if (suppressWarnings(max(abs(0.5 - txtcolor),
                                                   na.rm = TRUE) < 0.1)) {
                            txtcolor <- c(0, 0, 0)
                          }
                          #greyscale, https://www.baeldung.com/cs/convert-rgb-to-grayscale#3-luminosity-method
                          gs <-
                            0.3 * txtcolor[[1]] +
                            0.59 * txtcolor[[2]] +
                            0.11 * txtcolor[[3]]
                          txtcolor <- rep(gs, 3)
                        } else {
                          # text in black or white, depending on the brightness of the background, https://stackoverflow.com/questions/11867545/change-text-color-based-on-brightness-of-the-covered-background-area, https://www.w3.org/TR/AERT/#color-contrast
                          if (any(is.na(colors[rw, cl][[1]]))) {
                            txtcolor <- c(0, 0, 0)
                          } else {
                            brightness <- (299 * 255 * colors[rw, cl][[1]][1] +
                                             587 * 255 * colors[rw, cl][[1]][2] + 114 * 255 * colors[rw, cl][[1]][3]) / 1000
                            if (brightness > 125) {
                              txtcolor <- c(0, 0, 0)
                            } else {
                              txtcolor <- c(1, 1, 1)
                            }
                          }
                        }
                        txtcolor <- try(do.call(rgb, as.list(txtcolor)),
                          silent = TRUE)
                        if (inherits(txtcolor, "try-error")) {
                          txtcolor <- "#222222"
                        }
                        if (!is.null(hover_texts)) {
                          hover_text <- hover_texts[rw, cl]
                          if (escape_all_content) {
                            hover_text <- htmltools::htmlEscape(hover_text)
                          }
                        } else {
                          hover_text <- NULL
                        }
                        val <- tb[rw, cl]
                        if (!escape_all_content) {
                          val <- htmltools::HTML(val)
                        }
                        td(
                          style =
                            sprintf("background-color: %s; color: %s; text-align: right;",
                                    color, txtcolor),
                          title = hover_text,
                          val
                        )
                      })
             )
  )))
}

#' Add data frames to the pre-loaded / cache data frame environment
#'
#' These can be referred to by their names, then, wherever `dataquieR` expects
#' a [data.frame] -- just pass a character instead. If this character is not
#' found, `dataquieR` would additionally look for files with the name and for
#' `URLs`. You can also refer to specific sheets of a workbook or specific
#' object from an `RData` by appending a pipe symbol and its name. A second
#' pipe symbol allows to extract certain columns from such sheets (but
#' they will remain data frames).
#'
#' @param ... data frames, if passed with names, these will be the names
#'            of these tables in the data frame environment. If not, then the
#'            names in the calling environment will be used.
#' @param data_frame_list a named list with data frames. Also these will be
#'                        added and names will be handled as for the `...`
#'                        argument.
#' @param append [logical] if a data frame already exists in the cache
#'                         (by name), extend the existing one
#'
#' @return [data.frame] `invisible(the cache environment)`
#' @export
#' @seealso [prep_load_workbook_like_file]
#' @seealso [prep_get_data_frame]
#' @family data-frame-cache
prep_add_data_frames <- function(...,
                                 data_frame_list = list(),
                                 append = FALSE) {
  util_expect_scalar(append, check_type = is.logical)
  ellipse <- list(...)
  if (is.null(names(ellipse))) {
    names(ellipse) <- rep("", length(ellipse))
  }
  unnamed <- !nzchar(names(ellipse))
  symbols <- sys.call()[-1]
  if ("data_frame_list" %in% names(symbols)) {
    symbols <- as.character(
      symbols[-which(names(symbols) == "data_frame_list")])
  } else {
    symbols <- as.character(symbols)
  }
  names(ellipse)[unnamed] <- symbols[unnamed]
  if (!is.list(data_frame_list)) {
    util_error("%s needs to be a list, if given", dQuote(data_frame_list))
  }
  symbols <- as.character(substitute(data_frame_list))[-1]
  unnamed <- !nzchar(names(data_frame_list))
  names(data_frame_list)[unnamed] <- symbols[unnamed]
  data_frame_list <- c(ellipse, data_frame_list)
  named <- nzchar(names(data_frame_list))
  if (!all(named)) {
    util_error(
      c("Some unnamed data frames could not even automatically be named",
        "Have you passed an empty string?")
    )
  }
  characters <- vapply(data_frame_list, is.character, FUN.VALUE = logical(1)) &
    (1 == vapply(data_frame_list, length, FUN.VALUE = integer(1)))
  data_frame_list[characters] <- lapply(data_frame_list[characters],
                                        prep_get_data_frame)
  nulls <- vapply(data_frame_list, is.null, FUN.VALUE = logical(1))
  data_frame_list <- data_frame_list[!nulls]
  errors <- !vapply(data_frame_list, is.data.frame, FUN.VALUE = logical(1))

  data_frame_list[errors] <- lapply(data_frame_list[errors], function(x) {
    if (is.null(dim(x))) {
      try(stop("Dimension-less object is not a data frame"), silent = TRUE)
    } else {
      try(as.data.frame(x), silent = TRUE)
    }
  })

  errors <- !vapply(data_frame_list, is.data.frame, FUN.VALUE = logical(1))

  if (any(errors)) {
    util_error(
      "Not all entries are data frames or loadable file names/URLs: %s",
      paste0(dQuote(names(data_frame_list)[errors]), collapse = ", "))
  }
  for (n in names(data_frame_list)) {
    if (append && exists(n, envir = .dataframe_environment())) {
      existing <- get(n, envir = .dataframe_environment())
      if (is.data.frame(existing)) {
        data_frame_list[[n]] <- util_rbind(
          existing,
          data_frame_list[[n]]
        )
      }
    }
    assign(
      n,
      data_frame_list[[n]],
      envir = .dataframe_environment())
  }
  invisible(.dataframe_environment())
}

#' Compute Descriptive Statistics
#' @description
#'
#' generates a descriptive overview of the variables in `resp_vars`.
#'
#' [Descriptor]
#'
#' @details
#' TODO
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable] the name of the measurement variables
#' @param hard_limits_removal [logical] if TRUE values outside hard limits are
#'                                      removed from the data before calculating
#'                                      descriptive statistics.
#'                                      The default is FALSE
#' @param ... arguments to be passed to all called indicator functions if
#'            applicable.
#'
#' @return a [list] with:
#'   - `SummaryTable`: [data.frame]
#'   - `SummaryData`: [data.frame]
#' @export
#'
#' @examples
#' \dontrun{
#' xx <- des_summary(study_data = "study_data", meta_data_v2 = "meta_data_v2")
#' xx$SummaryData
#' }
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_des_summary.html
#' )
#' @importFrom graphics barplot hist
#' @importFrom stats mad
#'

des_summary <- function(resp_vars = NULL,
                        study_data,
                        label_col,
                        item_level = "item_level",
                        meta_data = item_level,
                        meta_data_v2,
                        hard_limits_removal =
                          getOption("dataquieR.des_summary_hard_lim_remove",
                                    dataquieR.des_summary_hard_lim_remove_default),
                        ...) {

  ##TODO: if a variable is present in the study data but not in the metadata is
  # not in the final descriptive statistics! It is already lost at line 37!

  # PREPARATION----
  # Metadata and study data set up ----
  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()
  util_expect_data_frame(study_data)
  try(util_expect_data_frame(meta_data), silent = TRUE)
  item_level <- meta_data <-
    util_amend_missing_metadata(study_data = study_data,
                                meta_data = meta_data,
                                level = VARATT_REQUIRE_LEVELS$RECOMMENDED)
  if (util_is_try_error(
    erobj <- try(prep_prepare_dataframes(.replace_hard_limits = hard_limits_removal,
                                         .apply_factor_metadata = TRUE),
                 silent = TRUE))) {
    w <- "%s has problems: %s, estimating %s"
    if (suppressWarnings(util_ensure_suggested("cli", err = FALSE))) {
      w <- cli::bg_black(cli::col_yellow(w))
    }
    util_warning(w, sQuote("meta_data"),
                 dQuote(conditionMessage(util_condition_from_try_error(erobj))),
                 sQuote("meta_data"),
                 immediate = TRUE)
    meta_data <-
      prep_study2meta(study_data, level = VARATT_REQUIRE_LEVELS$RECOMMENDED)
    prep_prepare_dataframes(.meta_data = meta_data,
                            .replace_hard_limits = hard_limits_removal,
                            .apply_factor_metadata = TRUE)
  }

  #ORDER of variables: reorder both metadata and study data with same order----
  if(!VARIABLE_ORDER %in% colnames(meta_data)){
    #VARIABLE_ORDER missing, add appearance order of variables in the study data
    mapping_order_of_vars <- setNames(seq_len(ncol(ds1)),
                                      nm = names(ds1))
  } else {
    #VARIABLE_ORDER present, so take that as order numbers
    meta_data$VARIABLE_ORDER <-
      suppressWarnings(as.numeric(meta_data$VARIABLE_ORDER))
    meta_data$VARIABLE_ORDER <- rank(meta_data$VARIABLE_ORDER,
                                     ties.method = "first",
                                     na.last = TRUE)
    mapping_order_of_vars <- setNames(meta_data$VARIABLE_ORDER,
                                      nm = meta_data[[label_col]])
  }
  # Order both study data and metadata variables
  ds1 <- ds1[, order(mapping_order_of_vars[colnames(ds1)]), drop = FALSE]
  meta_data <- meta_data[order(mapping_order_of_vars[meta_data[[label_col]]]), ,
                         drop = FALSE]

  # Define resp_vars and filter for them----
  if (length(resp_vars) == 0) {
    #in case of missing resp_vars use all variables in the study data
    suppress <- function(...) suppressWarnings(suppressMessages(...))
    resp_vars <- colnames(ds1)
  } else {
    # in case of resp_vars present, check them
    suppress <- eval
  }
  suppress(util_correct_variable_use(resp_vars,
                                     allow_more_than_one = TRUE,
                                     do_not_stop = TRUE,
                                     # I remove here the variables with
                                     # na in the scale level
                                     need_scale = "!na"))
  resp_vars <- intersect(colnames(ds1), resp_vars)


  # Select only resp_vars for both study data and metadata
  ds1_filtered <- ds1[, colnames(ds1) %in% resp_vars, drop = FALSE]
  md_filtered <- meta_data[meta_data[, label_col, drop = TRUE] %in%
                             resp_vars, ]
  md_filtered$nprog <- 1:nrow(md_filtered)

  #Define categorical variables----
  # nominal variables
  nominal_cols <- md_filtered$SCALE_LEVEL %in% c(SCALE_LEVELS$NOMINAL)
  my_nominal_ds <- ds1_filtered[, nominal_cols, drop = FALSE]
  my_nominal_meta <- md_filtered[nominal_cols, ,drop = FALSE]

  # ordinal variables
  ordinal_cols <- md_filtered$SCALE_LEVEL %in% c(SCALE_LEVELS$ORDINAL)
  my_ordinal_ds <- ds1_filtered[, ordinal_cols, drop = FALSE]
  my_ordinal_meta <- md_filtered[ordinal_cols, ,drop = FALSE]

  # nominal and ordinal variables
  categorical_cols <- md_filtered$SCALE_LEVEL %in% c(SCALE_LEVELS$NOMINAL,
                                                     SCALE_LEVELS$ORDINAL)
  my_categor_ds <- ds1_filtered[, categorical_cols, drop = FALSE]
  my_categor_meta <- md_filtered[categorical_cols, ,drop = FALSE]

  rm(nominal_cols, ordinal_cols, categorical_cols)


  #TODO: if we have a scale level ratio and interval, and datatype is not numeric is a problem
  # can it be remedied or not-- should be done is prep_prepare_dataframe??

  #Define continuous variables----
  ratio_interv_cols <- md_filtered$SCALE_LEVEL %in% c(SCALE_LEVELS$RATIO,
                                                      SCALE_LEVELS$INTERVAL)
  my_ratioint_ds <- ds1_filtered[, ratio_interv_cols, drop = FALSE]
  my_ratioint_meta <- md_filtered[ratio_interv_cols, ,drop = FALSE]

  ratioint_notime_cols <- md_filtered$SCALE_LEVEL %in%
    c(SCALE_LEVELS$RATIO,
      SCALE_LEVELS$INTERVAL) &
    !(md_filtered$DATA_TYPE %in% c(DATA_TYPES$DATETIME,
                                   DATA_TYPES$TIME))
  my_continuous_ds <- ds1_filtered[, ratioint_notime_cols, drop = FALSE]
  my_continuous_meta <- md_filtered[ratioint_notime_cols, ,drop = FALSE]


  datetime_cols <- md_filtered$SCALE_LEVEL %in%
    c(SCALE_LEVELS$INTERVAL) &
    md_filtered$DATA_TYPE %in% c(DATA_TYPES$DATETIME,
                                 DATA_TYPES$TIME)
  my_datetime_ds <- ds1_filtered[, datetime_cols, drop = FALSE]
  my_datetime_meta <- md_filtered[datetime_cols, ,drop = FALSE]


  #need also an extra subset for CV, Kurtosis, and Skewness
  num_cols <- md_filtered$SCALE_LEVEL %in%
    c(SCALE_LEVELS$RATIO, SCALE_LEVELS$INTERVAL) &
    md_filtered$DATA_TYPE %in%
    c(DATA_TYPES$INTEGER, DATA_TYPES$FLOAT)
  my_num_ds <- ds1_filtered[, num_cols, drop = FALSE]
  my_num_meta <- md_filtered[num_cols, ,drop = FALSE]

  rm(ratio_interv_cols, ratioint_notime_cols, datetime_cols, num_cols)

  # Create an overview table - Summary_table----
  #First all names/label columns----
  summary_table <- md_filtered[,
                               colnames(md_filtered) %in%
                                 label_col, drop = FALSE]

  # Create all_labels together and add to summary_data----
  summary_labels <- md_filtered[,
                                colnames(md_filtered) %in%
                                  c(VAR_NAMES,
                                    LABEL,
                                    label_col,
                                    LONG_LABEL,
                                    colnames(md_filtered)[
                                      startsWith(colnames(md_filtered),
                                                 paste0(LABEL, "_"))],
                                    colnames(md_filtered)[
                                      startsWith(colnames(md_filtered),
                                                 paste0(LONG_LABEL, "_"))]
                                  ), drop = FALSE]

  rev_summary_labels <- rev(summary_labels)
  rownames(rev_summary_labels) =  NULL
  names_tb <- names(rev_summary_labels)
  rev_summary_labels <- cbind(rev_summary_labels,
                              vapply(apply(rev_summary_labels, 1,
                                           unique,
                                           simplify = FALSE),
                                     paste,
                                     collapse = "<br />",
                                     FUN.VALUE = character(1)))
  names(rev_summary_labels) <- c(names_tb,"Variable_names")
  Variable_names <- setNames(nm = rev_summary_labels[[label_col]],
                             rev_summary_labels$Variable_names)
  rm(rev_summary_labels, names_tb, summary_labels)
  # add the new column to summary data, matching the label_col columns
  summary_table$Variable_names <- util_attach_attr(
    Variable_names[match(summary_table[[label_col]],
                         names(Variable_names))],
    plain_label = summary_table[[label_col]]
  )

  # Add data type, scale level, and study segment----
  summary_table <- cbind(summary_table,
                         md_filtered[,
                                     colnames(md_filtered) %in%
                                       c("SCALE_LEVEL",
                                         "DATA_TYPE",
                                         "STUDY_SEGMENT")])

  ### Mean ----
  # Mean ratio and interval variables - no datetime
  vec1 <- unlist(lapply(my_continuous_ds,
                        mean, na.rm = TRUE))
  summary_table$mean_contin <- vec1[match(summary_table[[label_col]],
                                          names(vec1))]
  rm(vec1)

  #create empty ones in case no continuous data are present and round
  if (!"mean_contin" %in% colnames(summary_table)) {
    summary_table$mean_contin <- NA
  } else {
    summary_table$mean_contin[!is.na(summary_table$mean_contin)] <-
      format(util_round_to_decimal_places(
        summary_table$mean_contin[!is.na(summary_table$mean_contin)]))
    summary_table$mean_contin <- trimws(summary_table$mean_contin)
  }

  # Mean datetime variables
  vec1 <- unlist(lapply(my_datetime_ds,
                        function(x) {
                          mean_value <- mean(x, na.rm = TRUE)
                          r <- format(mean_value, usetz = TRUE)
                          return(r)
                        }
  ))
  summary_table$mean_datetime <- vec1[match(summary_table[[label_col]],
                                            names(vec1))]
  rm(vec1)

  #create empty ones in case no continuous data are present
  if (!"mean_datetime" %in% colnames(summary_table)) {
    summary_table$mean_datetime <- NA
  }

  #Combine columns Mean
  existing_columns <- intersect(colnames(summary_table),
                                c("mean_contin", "mean_datetime"))
  if(length(existing_columns) > 0) {
    summary_table$Mean <-
      apply(summary_table[, existing_columns, drop = FALSE], 1,
            util_des_functions_env$util_combine_cols_content, subject = "Mean")
  } else {
    summary_table$Mean <- NA
  }

  # identify columns NOT to remove
  cols_to_keep <- setdiff(colnames(summary_table), existing_columns)
  # remove the extra columns the data frame
  summary_table <- summary_table[, cols_to_keep]
  rm(cols_to_keep, existing_columns)


  ### Median----
  # Median of ordinal variables
  vec1 <- unlist(lapply(my_ordinal_ds,
                        util_des_functions_env$util_compute_median_cat))
  summary_table$median_cat <- vec1[match(summary_table[[label_col]],
                                         names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"median_cat" %in% colnames(summary_table)) {
    summary_table$median_cat <- NA
  }

  # Median of ratio and interval - no datetime
  vec1 <- unlist(lapply(my_continuous_ds,
                        median, na.rm = TRUE))
  summary_table$median_contin <- vec1[match(summary_table[[label_col]],
                                            names(vec1))]
  rm(vec1)
  #create empty one in case not created and round
  if (!"median_contin" %in% colnames(summary_table)) {
    summary_table$median_contin <- NA
  } else{
    summary_table$median_contin[!is.na(summary_table$median_contin)] <-
      format(util_round_to_decimal_places(
        summary_table$median_contin[!is.na(summary_table$median_contin)]))
    summary_table$median_contin <- trimws(summary_table$median_contin)
  }

  # Median datetime
  vec1 <- unlist(lapply(my_datetime_ds,
                        function(x){
                          m <- median(x, na.rm = TRUE)
                          format(m, usetz = TRUE)
                        }
  ))
  summary_table$median_datetime <-
    vec1[match(summary_table[[label_col]],names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"median_datetime" %in% colnames(summary_table)) {
    summary_table$median_datetime <- NA
  }


  # Merge the 3 median values in one column
  existing_columns <- intersect(colnames(summary_table),
                                c("median_cat",
                                  "median_contin",
                                  "median_datetime"))
  if(length(existing_columns) > 0) {
    summary_table$Median <-
      apply(summary_table[, existing_columns, drop = FALSE], 1,
            util_des_functions_env$util_combine_cols_content,
            subject = "Median")
  } else {
    summary_table$Median <- NA
  }

  # identify columns NOT to remove
  cols_to_keep <- setdiff(colnames(summary_table), existing_columns)
  # remove the extra columns the data frame
  summary_table <- summary_table[, cols_to_keep]
  rm(cols_to_keep, existing_columns)



  ### Mode----
  # Mode of categorical variables
  vec1 <- unlist(lapply(my_categor_ds,
                        util_des_functions_env$util_compute_mode_cat))
  summary_table$mode_cat <- vec1[match(summary_table[[label_col]],
                                       names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"mode_cat" %in% colnames(summary_table)) {
    summary_table$mode_cat <- NA
  }

  # Mode of ratio and interval - no datetime
  vec1 <- unlist(lapply(my_continuous_ds,
                        util_des_functions_env$util_compute_mode_contin))
  summary_table$mode_contin <- vec1[match(summary_table[[label_col]],
                                          names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"mode_contin" %in% colnames(summary_table)) {
    summary_table$mode_contin <- NA
  }

  # Mode of datetime vars
  vec1 <- unlist(lapply(my_datetime_ds,
                        util_des_functions_env$util_compute_mode_datetime))
  summary_table$mode_datetime <- vec1[match(summary_table[[label_col]],
                                            names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"mode_datetime" %in% colnames(summary_table)) {
    summary_table$mode_datetime <- NA
  }

  #Combine columns Mode
  existing_columns <- intersect(colnames(summary_table),
                                c("mode_cat",
                                  "mode_contin",
                                  "mode_datetime"))
  if(length(existing_columns) > 0) {
    summary_table$Mode <-
      apply(summary_table[, existing_columns, drop = FALSE], 1,
            util_des_functions_env$util_combine_cols_content,
            subject = "Mode")
  } else {
    summary_table$Mode <- NA
  }

  # identify columns NOT to remove
  cols_to_keep <- setdiff(colnames(summary_table), existing_columns)
  # remove the extra columns the data frame
  summary_table <- summary_table[, cols_to_keep]
  rm(cols_to_keep, existing_columns)



  ### SD----
  # SD ratio and interval variables - no datetime
  vec1 <- unlist(lapply(my_continuous_ds,
                        sd, na.rm = TRUE))
  summary_table$SD_contin <- vec1[match(summary_table[[label_col]],
                                        names(vec1))]
  rm(vec1)
  #create empty one in case not created and round
  if (!"SD_contin" %in% colnames(summary_table)) {
    summary_table$SD_contin <- NA
  } else {
    summary_table$SD_contin[!is.na(summary_table$SD_contin)] <-
      format(util_round_to_decimal_places(
        summary_table$SD_contin[!is.na(summary_table$SD_contin)]))
    summary_table$SD_contin <- trimws(summary_table$SD_contin)
  }

  # SD datetime variables
  vec1 <-
    unlist(lapply(my_datetime_ds,
                  function(x) {
                    time_in_sec <- as.difftime(sd(x, na.rm = TRUE),
                                               units = 'secs')
                    #total_seconds <- lubridate::as.duration(
                    #  as.numeric(time_in_sec, units = "secs"))
                    #formatted_time <- util_des_functions_env$util_format_duration_human(start = total_seconds)
                    formatted_time <-
                      util_des_functions_env$util_compute_difftime_auto(time_in_sec)
                    format(formatted_time,
                           usetz = TRUE)
                  }))
  summary_table$SD_datetime <- vec1[match(summary_table[[label_col]],
                                          names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"SD_datetime" %in% colnames(summary_table)) {
    summary_table$SD_datetime <- NA
  }

  #Combine columns SD
  existing_columns <- intersect(colnames(summary_table),
                                c("SD_contin",
                                  "SD_datetime"))
  if(length(existing_columns) > 0) {
    summary_table$SD <-
      apply(summary_table[, existing_columns, drop = FALSE], 1,
            util_des_functions_env$util_combine_cols_content,
            subject = "SD")
  } else {
    summary_table$SD <- NA
  }

  # identify columns NOT to remove
  cols_to_keep <- setdiff(colnames(summary_table), existing_columns)
  # remove the extra columns the data frame
  summary_table <- summary_table[, cols_to_keep]
  rm(cols_to_keep, existing_columns)



  ### IQR----
  # IQR ordinal variables
  vec1 <- lapply(my_ordinal_ds,
                 util_des_functions_env$util_compute_IQR_ord)
  #actual range
  IQR <- vapply(vec1, `[[`, "IQR",
                FUN.VALUE = numeric(1))
  # range + Q1 and Q3
  IQR_q1_q3 <- vapply(vec1, `[[`, "IQR_q1_q3",
                      FUN.VALUE = character(1))
  rm(vec1)
  #Add both to summary_table
  summary_table$IQR_ord <- IQR[match(summary_table[[label_col]],
                                     names(IQR))]
  summary_table$IQR_ord_q1_q3 <- IQR_q1_q3[match(summary_table[[label_col]],
                                                 names(IQR_q1_q3))]
  #create empty one in case not created
  if (!"IQR_ord" %in% colnames(summary_table)) {
    summary_table$IQR_ord <- NA
  } else {
    summary_table$IQR_ord <- format(summary_table$IQR_ord)
  }
  if (!"IQR_ord_q1_q3" %in% colnames(summary_table)) {
    summary_table$IQR_ord_q1_q3 <- NA
  }

  # IQR ratio and interval - no datetime
  vec1 <- lapply(my_continuous_ds,
                 util_des_functions_env$util_compute_IQR_contin)
  #actual range
  IQR <- vapply(vec1, `[[`, "IQR", FUN.VALUE = numeric(1))
  # range + Q1 and Q3
  IQR_q1_q3 <- vapply(vec1, `[[`, "IQR_q1_q3", FUN.VALUE = character(1))
  rm(vec1)
  #Add both to summary_table
  summary_table$IQR_cont <- IQR[match(summary_table[[label_col]],
                                      names(IQR))]
  summary_table$IQR_cont_q1_q3 <- IQR_q1_q3[match(summary_table[[label_col]],
                                                  names(IQR_q1_q3))]
  #create empty one in case not created
  if (!"IQR_cont" %in% colnames(summary_table)) {
    summary_table$IQR_cont <- NA
  } else {
    summary_table$IQR_cont[!is.na(summary_table$IQR_cont)] <-
      format(util_round_to_decimal_places(
        summary_table$IQR_cont[!is.na(summary_table$IQR_cont)]))
  }
  if (!"IQR_cont_q1_q3" %in% colnames(summary_table)) {
    summary_table$IQR_cont_q1_q3 <- NA
  }

  # IQR datetime variables
  vec1 <- lapply(my_datetime_ds,
                 util_des_functions_env$util_compute_IQR_datetime)
  #actual range
  IQR <- vapply(vec1, `[[`, "IQR", FUN.VALUE = character(1))
  # range + Q1 and Q3
  IQR_q1_q3 <- vapply(vec1, `[[`, "IQR_q1_q3", FUN.VALUE = character(1))
  rm(vec1)
  #Add both to summary_table
  summary_table$IQR_datetime <- IQR[match(summary_table[[label_col]],
                                          names(IQR))]
  summary_table$IQR_datetime_q1_q3 <-
    IQR_q1_q3[match(summary_table[[label_col]],
                    names(IQR_q1_q3))]
  #create empty one in case not created
  if (!"IQR_datetime" %in% colnames(summary_table)) {
    summary_table$IQR_datetime <- NA
  }
  if (!"IQR_datetime_q1_q3" %in% colnames(summary_table)) {
    summary_table$IQR_datetime_q1_q3 <- NA
  }

  # Create a column `IQR (Quartiles)`with all IQR values
  # (IQR_ord_q1_q3, IQR_cont_q1_q3, IQR_datetime_q1_q3)
  #Combine columns SD
  existing_columns <- intersect(colnames(summary_table),
                                c("IQR_ord_q1_q3",
                                  "IQR_cont_q1_q3",
                                  "IQR_datetime_q1_q3"))
  if(length(existing_columns) > 0) {
    summary_table$`IQR (Quartiles)` <-
      apply(summary_table[, existing_columns, drop = FALSE], 1,
            util_des_functions_env$util_combine_cols_content,
            subject = "IQR (Quartiles)")
  } else {
    summary_table$`IQR (Quartiles)` <- NA
  }

  # identify columns NOT to remove
  existing_columns <- intersect(colnames(summary_table),
                                c("IQR_ord_q1_q3",
                                  "IQR_cont_q1_q3",
                                  "IQR_datetime_q1_q3",
                                  "IQR_ord",
                                  "IQR_cont",
                                  "IQR_datetime"))
  cols_to_keep <- setdiff(colnames(summary_table), existing_columns)
  # remove the extra columns the data frame
  summary_table <- summary_table[, cols_to_keep]
  rm(cols_to_keep, existing_columns)



  ### MAD----
  # MAD ratio and interval variables - no datetime
  vec1 <- unlist(lapply(my_continuous_ds,
                        mad, na.rm = TRUE))

  summary_table$MAD_contin <- vec1[match(summary_table[[label_col]],
                                         names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"MAD_contin" %in% colnames(summary_table)) {
    summary_table$MAD_contin <- NA
  } else {
    summary_table$MAD_contin[!is.na(summary_table$MAD_contin)] <-
      format(util_round_to_decimal_places(
        summary_table$MAD_contin[!is.na(summary_table$MAD_contin)]))
    summary_table$MAD_contin <- trimws(summary_table$MAD_contin)
  }

  # MAD datetime variables
  vec1 <-
    unlist(lapply(my_datetime_ds,
                  function(x) {
                    time_in_sec <- as.difftime(mad(x, na.rm = TRUE),
                                               units = 'secs')
                    # total_seconds <- lubridate::as.duration(
                    #   as.numeric(time_in_sec, units = "secs"))

                    #formatted_time <-
                    #  util_des_functions_env$util_format_duration_human(start = total_seconds)

                    formatted_time <-
                      util_des_functions_env$util_compute_difftime_auto(time_in_sec)
                    format(formatted_time,
                           usetz = TRUE)
                  }))
  summary_table$MAD_datetime <- vec1[match(summary_table[[label_col]],
                                           names(vec1))]
  rm(vec1)

  #create empty one in case not created
  if (!"MAD_datetime" %in% colnames(summary_table)) {
    summary_table$MAD_datetime <- NA
  }

  # merge MAD values
  existing_columns <- intersect(colnames(summary_table),
                                c("MAD_contin",
                                  "MAD_datetime"))
  if(length(existing_columns) > 0) {
    summary_table$MAD <-
      apply(summary_table[, existing_columns, drop = FALSE], 1,
            util_des_functions_env$util_combine_cols_content,
            subject = "MAD")
  } else {
    summary_table$MAD <- NA
  }

  # identify columns NOT to remove
  cols_to_keep <- setdiff(colnames(summary_table), existing_columns)
  # remove the extra columns the data frame
  summary_table <- summary_table[, cols_to_keep]
  rm(cols_to_keep, existing_columns)



  ### Range Min max----
  # Min max ordinal variables
  #min
  vec1 <- unlist(lapply(my_ordinal_ds,
                        min, na.rm = TRUE))
  if (!is.null(vec1)) {
    summary_table$min_ord <- vec1[match(summary_table[[label_col]],
                                        names(vec1))]
    rm(vec1)
  }
  #max
  vec1 <- unlist(lapply(my_ordinal_ds,
                        max, na.rm = TRUE))
  if (!is.null(vec1)) {
    summary_table$max_ord <- vec1[match(summary_table[[label_col]],
                                        names(vec1))]
    rm(vec1)
  }
  summary_table$min_max_ord <- NA
  result <- !is.na(summary_table$min_ord) & !is.na(summary_table$max_ord)
  # create the min_max_ord vars when results are present
  summary_table$min_max_ord[result] <-
    paste0("(",
           summary_table$min_ord[result], " - ",
           summary_table$max_ord[result], ")")

  #create empty one in case not created
  if (!"min_ord" %in% colnames(summary_table)) {
    summary_table$min_ord <- NA
  }
  if (!"max_ord" %in% colnames(summary_table)) {
    summary_table$max_ord <- NA
  }
  if (!"min_max_ord" %in% colnames(summary_table)) {
    summary_table$min_max_ord <- NA
  }


  # Min max ratio and interval variables - no datetime
  #min
  vec1 <- unlist(lapply(my_continuous_ds,
                        min, na.rm = TRUE))
  if (!is.null(vec1)) {
    summary_table$min_contin <- vec1[match(summary_table[[label_col]],
                                           names(vec1))]
    rm(vec1)
  }
  #max
  vec1 <- unlist(lapply(my_continuous_ds,
                        max, na.rm = TRUE))
  if (!is.null(vec1)) {
    summary_table$max_contin <- vec1[match(summary_table[[label_col]],
                                           names(vec1))]
    rm(vec1)
  }
  # string with min and max
  #create empty one in case not created
  if (!"min_contin" %in% colnames(summary_table)) {
    summary_table$min_contin <- NA
  }
  if (!"max_contin" %in% colnames(summary_table)) {
    summary_table$max_contin <- NA
  }
  summary_table$range_temp <- summary_table$max_contin-summary_table$min_contin
  result <- !is.na(summary_table$min_contin) & !is.na(summary_table$max_contin)
  summary_table$min_max_contin <- NA
  summary_table$min_max_contin[result] <-
    paste0( summary_table$range_temp[result],
            " (",
            summary_table$min_contin[result], " - ",
            summary_table$max_contin[result], ")")
  #create empty one in case not created
  if (!"min_max_contin" %in% colnames(summary_table)) {
    summary_table$min_max_contin <- NA
  }
  summary_table$range_temp <- NULL


  # Min max datetime variables
  # min
  vec1 <- unlist(lapply(my_datetime_ds,
                        function(x){
                          m <- min(x, na.rm = TRUE)
                          format(m, usetz = TRUE)
                        }))

  if (!is.null(vec1)) {
    summary_table$min_datetime <- vec1[match(summary_table[[label_col]],
                                             names(vec1))]
    rm(vec1)
  }
  #max
  vec1 <- unlist(lapply(my_datetime_ds,
                        function(x){
                          m <- max(x, na.rm = TRUE)
                          format(m, usetz = TRUE)
                        }
  ))
  if (!is.null(vec1)) {
    summary_table$max_datetime <- vec1[match(summary_table[[label_col]],
                                             names(vec1))]
    rm(vec1)
  }
  # string with min and max
  vec1 <-
    unlist(lapply(my_datetime_ds,
                  function(x){
                    max <- max(x, na.rm = TRUE)
                    min <- min(x, na.rm = TRUE)
                    if (inherits(x, "hms")) {
                      max <- hms::as_hms(max)
                      min <- hms::as_hms(min)
                    }
                    # iqr_interval <- lubridate::interval(min, max)
                    # Convert to a period human-readable
                    # formatted_time <- lubridate::as.period(iqr_interval)
                    # formatted_time <-gsub("(^| )0[ymdHMS]", "", formatted_time)
                    # print(formatted_time)
                    time_in_sec <- as.difftime(max - min,
                                               units = 'secs')
                    formatted_time <-
                      util_des_functions_env$util_format_duration_human(
                        start = min,
                        end = max)
                    paste0(formatted_time,
                           " (",
                           format(min, usetz = TRUE), " - ",
                           format(max, usetz = TRUE), ")")
                  }))
  if (!is.null(vec1)) {
    summary_table$min_max_datetime <- vec1[match(summary_table[[label_col]],
                                                 names(vec1))]
    rm(vec1)
  }
  #create empty one in case not created
  if (!"min_datetime" %in% colnames(summary_table)) {
    summary_table$min_datetime <- NA
  }
  if (!"max_datetime" %in% colnames(summary_table)) {
    summary_table$max_datetime <- NA
  }
  if (!"min_max_datetime" %in% colnames(summary_table)) {
    summary_table$min_max_datetime <- NA
  }
  #merge min-max values
  existing_columns <- intersect(colnames(summary_table),
                                c("min_max_ord",
                                  "min_max_contin",
                                  "min_max_datetime"))
  if(length(existing_columns) > 0) {
    summary_table$`Range (Min - Max)` <-
      apply(summary_table[, existing_columns, drop = FALSE], 1,
            util_des_functions_env$util_combine_cols_content,
            subject = "Range (Min - Max)")
  } else {
    summary_table$`Range (Min - Max)` <- NA
  }
  # identify columns NOT to remove
  existing_columns <- intersect(colnames(summary_table),
                                c("min_max_ord",
                                  "min_max_contin",
                                  "min_max_datetime",
                                  "min_ord",
                                  "max_ord",
                                  "min_contin",
                                  "max_contin",
                                  "min_datetime",
                                  "max_datetime"))
  cols_to_keep <- setdiff(colnames(summary_table), existing_columns)
  # remove the extra columns the data frame
  summary_table <- summary_table[, cols_to_keep]
  rm(cols_to_keep, existing_columns)



  ### CV ----
  # CV numeric variables - only continuous that are integer or float
  vec1 <- unlist(lapply(my_num_ds,
                        function(x){
                          sd(x, na.rm = TRUE) /
                            mean(x, na.rm = TRUE) * 100
                        }))
  summary_table$CV <- vec1[match(summary_table[[label_col]],
                                 names(vec1))]
  rm(vec1)

  #create empty one in case not created
  if (!"CV" %in% colnames(summary_table)) {
    summary_table$CV <- NA
  } else {
    summary_table$CV[!is.na(summary_table$CV)] <-
      format(util_round_to_decimal_places(
        summary_table$CV[!is.na(summary_table$CV)]))
    summary_table$CV <- trimws(summary_table$CV)
  }
  # replace NAs with ""
  no_result <- is.na(summary_table$CV)
  # replace NA with ""
  summary_table$CV[no_result] <- ""



  #### Skewness value ----
  # Skewness - only continuous that are integer or float
  vec1 <- unlist(lapply(my_num_ds,
                        util_des_functions_env$util_compute_skewness))
  summary_table$skewness_value <- vec1[match(summary_table[[label_col]],
                                             names(vec1))]
  rm(vec1)
  #round results
  if (!"skewness_value" %in% colnames(summary_table)) {
    summary_table$skewness_value <- ""
  } else {
    no_result <- is.na(summary_table$skewness_value)
    summary_table$skewness_value[!no_result] <-
      format(util_round_to_decimal_places(summary_table$skewness_value[!no_result]))
    summary_table$skewness_value <- trimws(summary_table$skewness_value)
    # replace NA with ""
    summary_table$skewness_value[no_result] <- ""
  }


  #### SE Skewness ----
  # SE Skewness  - only continuous that are integer or float
  vec1 <- unlist(lapply(my_num_ds,
                        util_des_functions_env$util_compute_SE_skewness))
  summary_table$SE_skewness <- vec1[match(summary_table[[label_col]],
                                          names(vec1))]
  rm(vec1)
  #round results
  if (!"SE_skewness" %in% colnames(summary_table)) {
    summary_table$SE_skewness <- ""
  } else {
    no_result <- is.na(summary_table$SE_skewness)
    summary_table$SE_skewness[!no_result] <-
      format(util_round_to_decimal_places(summary_table$SE_skewness[!no_result]))
    summary_table$SE_skewness <- trimws(summary_table$SE_skewness)
    # replace NA with ""
    summary_table$SE_skewness[no_result] <- ""
  }


  ### Skewness----
  # merge skewness and replace NA with ""
  if ("skewness_value" %in% colnames(summary_table) &&
      !all(summary_table$skewness_value == "")) {
    #merge skewness
    no_result <- is.na(summary_table$skewness_value) |  summary_table$skewness_value==""
    summary_table$Skewness[!no_result] <-
      paste0(summary_table$skewness_value[!no_result], " (",
             summary_table$SE_skewness[!no_result], ")")
    # replace NA with "" and deleted not anymore needed cols
    summary_table$Skewness[no_result] <- ""
  }

  if (!"Skewness" %in% colnames(summary_table)) {
    summary_table$Skewness <- NA
  }



  ### Kurtosis ----
  # Kurtosis  - only continuous that are integer or float
  vec1 <- unlist(lapply(my_num_ds,
                        util_des_functions_env$util_compute_kurtosis))
  summary_table$Kurtosis <- vec1[match(summary_table[[label_col]],
                                       names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"Kurtosis" %in% colnames(summary_table)) {
    summary_table$Kurtosis <- NA
  } else {
    summary_table$Kurtosis[!is.na(summary_table$Kurtosis)] <-
      format(util_round_to_decimal_places(
        summary_table$Kurtosis[!is.na(summary_table$Kurtosis)]))
    summary_table$Kurtosis <- trimws(summary_table$Kurtosis)
  }

  # replace Na with ""
  if ("Kurtosis" %in% colnames(summary_table) &&
      !all(is.na(summary_table$Kurtosis))) {
    no_result  <- is.na(summary_table$Kurtosis)
    summary_table$Kurtosis[no_result] <- ""
  }



  ### Frequency table - only categorical-----
  # Obtain a table for ordinal and nominal variables - my_categor_ds
  #### 1. No.unique_values(included NA) ----
  vec1 <- unlist(lapply(my_categor_ds,
                        function(x){
                          length(unique(x))
                        }
  ))
  summary_table$No_categories_incl_NA <-
    vec1[match(summary_table[[label_col]],
               names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"No_categories_incl_NA" %in% colnames(summary_table)) {
    summary_table$No_categories_incl_NA <- NA
  }


  #### 2. Levels with N----
  vec1 <- unlist(lapply(my_categor_ds,
                        util_des_functions_env$util_compute_frequency_table))
  summary_table$Level_freq <- vec1[match(summary_table[[label_col]],
                                         names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"Level_freq" %in% colnames(summary_table)) {
    summary_table$Level_freq <- NA
  }

  ### Valid----
  # Valid N
  vec1 <- unlist(lapply(ds1_filtered,
                        function(x) sum(is.finite(x))
  ))
  summary_table$valid_N <- vec1[match(summary_table[[label_col]],
                                      names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"valid_N" %in% colnames(summary_table)) {
    summary_table$valid_N <- NA
  }

  # Valid %
  vec1 <- unlist(lapply(ds1_filtered,
                        function(x) {
                          perc_result <- sum(is.finite(x)) / length(x) * 100
                          return(perc_result)
                        }
  ))
  summary_table$valid_PCT <- vec1[match(summary_table[[label_col]],
                                        names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"valid_PCT" %in% colnames(summary_table)) {
    summary_table$valid_PCT <- NA
  } else{
    summary_table$valid_PCT <-
      format(util_round_to_decimal_places(summary_table$valid_PCT,
                                          digits = 2))
    summary_table$valid_PCT <- trimws(summary_table$valid_PCT)
  }

  # merge valid N and %
  summary_table$Valid <-
    paste0(summary_table$valid_N, " (",
           summary_table$valid_PCT,
           "%)")

  ### Missing ----
  # Missing N
  vec1 <- unlist(lapply(ds1_filtered,
                        function(x) sum(util_empty(x))
  ))
  summary_table$missing_N <- vec1[match(summary_table[[label_col]],
                                        names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"missing_N" %in% colnames(summary_table)) {
    summary_table$missing_N <- NA
  }

  # Missing %
  vec1 <- unlist(lapply(ds1_filtered,
                        function(x) {
                          perc_result <- (sum(util_empty(x)) / length(x)) * 100
                          return(perc_result)
                        }
  ))
  summary_table$missing_PCT <- vec1[match(summary_table[[label_col]],
                                          names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"missing_PCT" %in% colnames(summary_table)) {
    summary_table$missing_PCT <- NA
  } else {
    summary_table$missing_PCT <- format(
      util_round_to_decimal_places(summary_table$missing_PCT, digits = 2))
    summary_table$missing_PCT <- trimws(summary_table$missing_PCT)
  }

  #merge Missing N and %
  summary_table$Missing <-
    paste0(summary_table$missing_N, " (",
           summary_table$missing_PCT ,
           "%)")

  ### Graph -----
  # Graph nominal and ordinal variables
  vec1 <- unlist(lapply(my_categor_ds,
                        util_des_functions_env$util_compute_graph_cat))

  summary_table$graph_cat <- vec1[match(summary_table[[label_col]],
                                        names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"graph_cat" %in% colnames(summary_table)) {
    summary_table$graph_cat <- NA
  }

  # Graph DATETIME variables
  vec1 <- unlist(lapply(my_datetime_ds,
                        util_des_functions_env$util_compute_graph_datetime))
  summary_table$graph_datetime <- vec1[match(summary_table[[label_col]],
                                             names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"graph_datetime" %in% colnames(summary_table)) {
    summary_table$graph_datetime <- NA
  }

  # Graph integer and float variables
  vec1 <- unlist(lapply(my_num_ds,
                        util_des_functions_env$util_compute_graph_cont))
  summary_table$graph_cont <- vec1[match(summary_table[[label_col]],
                                         names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"graph_cont" %in% colnames(summary_table)) {
    summary_table$graph_cont <- NA
  }

  #Combine graphs
  existing_columns <- intersect(colnames(summary_table),
                                c("graph_cat",
                                  "graph_datetime",
                                  "graph_cont"))
  if(length(existing_columns) > 0) {
    summary_table$Graph <-
      apply(summary_table[, existing_columns, drop = FALSE], 1,
            util_des_functions_env$util_combine_cols_content,
            subject = "Graph")
  } else {
    summary_table$Graph <- NA
  }

  # identify columns NOT to remove
  cols_to_keep <- setdiff(colnames(summary_table), existing_columns)
  # remove the extra columns the data frame
  summary_table <- summary_table[, cols_to_keep]
  rm(cols_to_keep, existing_columns)



  # Additional column to summary_table to make it compatible with previous version
  # Type----
  vec_Type <- setNames(nm = summary_table[[label_col]],
                       paste0(summary_table$SCALE_LEVEL,
                              ", ", summary_table$DATA_TYPE))
  summary_table$Type <- vec_Type[match(summary_table[[label_col]],
                                       names(vec_Type))]


  # Create a summary_data starting from summary_table ----
  summary_data <- summary_table

  ### Remove unused columns----
  summary_data <- summary_data[, !(colnames(summary_data) %in%
                                     c(SCALE_LEVEL,DATA_TYPE ))]

  ### Create the plot of the frequency table----
  # only categorical
  vec1 <-
    unlist(lapply(my_categor_ds,
                  util_des_functions_env$util_compute_graph_frequency_table))
  summary_data$"Frequency table" <- vec1[match(summary_data[[label_col]],
                                               names(vec1))]
  rm(vec1)
  #create empty one in case not created
  if (!"Frequency table" %in% colnames(summary_data)) {
    summary_data$`Frequency table` <- NA
  }

  # Clean summary data table----
  col_vector <- c("Variable_names",
                  "Type",
                  "STUDY_SEGMENT",
                  "Mean",
                  "SD",
                  "Median",
                  "Mode",
                  "IQR (Quartiles)",
                  "MAD",
                  "Range (Min - Max)",
                  "CV",
                  "Skewness",
                  "Kurtosis",
                  "No_categories_incl_NA",
                  "Frequency table",
                  "Valid",
                  "Missing",
                  "Graph")

  summary_data <- summary_data[, colnames(summary_data) %in%
                                 col_vector]

  valid_cols_in_vec <- intersect(col_vector, names(summary_data))
  summary_data <- summary_data[, valid_cols_in_vec, drop = FALSE]

  ### correct names and replace NA with ""
  # correct names summary_table
  names(summary_table)[names(summary_table) == label_col] <- "Variables"
  # correct names summary_data
  names(summary_data)[names(summary_data) == "Variable_names"] <- "Variables"
  names(summary_data)[names(summary_data) == "Skewness"] <- "Skewness (SE)"
  names(summary_data)[names(summary_data) == "No_categories_incl_NA"] <- "No. categories (incl.NAs)"
  # replace NA in summary_data with ""
  summary_data[is.na(summary_data)] <- ""



  return(list(SummaryData = util_attach_attr(summary_data,
                                             is_html_escaped = TRUE),
              SummaryTable = util_attach_attr(summary_table,
                                              is_html_escaped = TRUE)))
}

#' Support function to allocate labels to variables
#'
#' @description
#' Map variables to certain attributes, e.g. by default their labels.
#'
#' @details
#' This function basically calls `colnames(study_data) <- meta_data$LABEL`,
#' ensuring correct merging/joining of study data columns to the corresponding
#' metadata rows, even if the orders differ. If a variable/study_data-column
#' name is not found in `meta_data[[from]]` (default `from = VAR_NAMES`),
#' either stop is called or, if `ifnotfound` has been assigned a value, that
#' value is returned. See [`mget`], which is internally used by this function.
#'
#' The function not only maps to the `LABEL` column, but `to` can be any
#' metadata variable attribute, so the function can also be used, to get, e.g.
#' all `HARD_LIMITS` from the metadata.
#'
#' @param x [character] variable names, character vector, see parameter from
#' @param item_level [data.frame] metadata data frame, if, as a `dataquieR`
#'                               developer, you do not have
#'                               **item-level-metadata**, you should use
#'                               `util_map_labels()` instead to avoid consistency
#'                               checks on for item-level `meta_data`.
#' @param meta_data [data.frame] old name for `item_level`
#' @param to [character] variable attribute to map to
#' @param from [character] variable identifier to map from
#' @param ifnotfound [list] A list of values to be used if the item is not
#'                          found: it will be coerced to a list if necessary.
#' @param warn_ambiguous [logical] print a warning if mapping variables from
#'                            `from` to `to` produces ambiguous identifiers.
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return a character vector with:
#'   - mapped values
#' @export
#'
#' @examples
#' \dontrun{
#' meta_data <- prep_create_meta(
#'   VAR_NAMES = c("ID", "SEX", "AGE", "DOE"),
#'   LABEL = c("Pseudo-ID", "Gender", "Age", "Examination Date"),
#'   DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER, DATA_TYPES$INTEGER,
#'                  DATA_TYPES$DATETIME),
#'   MISSING_LIST = ""
#' )
#' stopifnot(all(prep_map_labels(c("AGE", "DOE"), meta_data) == c("Age",
#'                                                  "Examination Date")))
#' }
prep_map_labels <- function(x, item_level = "item_level",
                            to = LABEL, from = VAR_NAMES, ifnotfound,
                            warn_ambiguous = FALSE,
                            meta_data_v2,
                            meta_data = item_level) {
  if (!missing(x) && length(x) == 0) {
    return(character(0))
  }

  if (!missing(item_level) && !missing(meta_data) &&
      !identical(item_level, meta_data)) {
    util_error(c("You cannot provide both, %s as well as %s",
                 "these arguments are synonyms and must be",
                 "used mutually exclusively"),
               sQuote("item_level"),
               sQuote("meta_data"))
    # see prep_prepare_dataframes
  }

  util_maybe_load_meta_data_v2()
  util_expect_data_frame(meta_data)

  # This is not needed, since labels have not significantly changed
  # meta_data <- suppressMessages(
  #   prep_meta_data_v1_to_item_level_meta_data(
  #     meta_data = meta_data, verbose = TRUE, label_col = LABEL))
  if (!is.data.frame(meta_data)) {
    util_error(
      c("Need metadata as a data frame for mapping",
        "variables to their attributes"),
      applicability_problem = TRUE
    )
  }
  .x <- try(as.character(x), silent = TRUE)
  if (inherits(.x, "try-error") || any(is.na(x) != is.na(.x))) {
    util_error("all variable source names must be characters",
               applicability_problem = TRUE)
  }
  x <- .x
  if (!is.character(to) || length(to) != 1 || !(to %in% colnames(meta_data))) {
    util_error(
      "Need exactly one existing variable attribute name to map variables to",
      applicability_problem = TRUE)
  }
  if (!is.character(from) || length(from) != 1 ||
      !(from %in% colnames(meta_data))) {
    util_error(
      c("Need exactly one variable attribute name",
        "to use as variable name on mapping"), applicability_problem = TRUE)
  }
  util_expect_scalar(warn_ambiguous, check_type = is.logical)
  util_map_labels(x, meta_data, to, from, ifnotfound, warn_ambiguous)
}

#' Check for unexpected data record set
#'
#' @description
#' This function tests that the identifiers match a provided record set. It is possible to
#' check for unexpected data record sets by study segments or to consider only selected
#' segments.
#'
#' [Indicator]
#'
#' @param level [character] a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").
#' @param ... Depending on `level`, passed to either
#'            `util_int_unexp_records_set_segment` or
#'            `util_int_unexp_records_set_dataframe`
#'
#' @inheritParams .template_function_indicator
#'
#' @return a [list]. Depending on `level`, see
#'   `util_int_unexp_records_set_segment` or
#'   `util_int_unexp_records_set_dataframe` for a description of the outputs.
#'
#' @export
#'
int_unexp_records_set <- function(level = c("dataframe", "segment"),
                                  study_data,
                                  item_level = "item_level",
                                  label_col,
                                  meta_data = item_level,
                                  meta_data_v2,
                                  ...) {
  util_maybe_load_meta_data_v2()

  level <- util_match_arg(level)
  fname <- rlang::call_name(rlang::frame_call())
  fname <- paste("util", fname, level, sep = "_")
  miss_label_col <- missing(label_col)
  if (miss_label_col) {
    label_col <- NULL
  }
  if (missing(study_data)) {
    cl_l <- list(fname, level = level, item_level = item_level,
                 meta_data = meta_data,
                 label_col = label_col, ...)
  } else {
    cl_l <- list(fname, level = level, item_level = item_level,
                 meta_data = meta_data, study_data = study_data,
                 label_col = label_col, ...)
  }
  if (missing(item_level) && !missing(meta_data)) {
    cl_l$item_level <- NULL
  }
  if (miss_label_col) {
    cl_l$label_col <- NULL
  }
  cl_l <- cl_l[names(cl_l) %in% c("", names(formals(fname)))]
  cl2 <- do.call("call",
                 cl_l)
  eval(cl2)

}

test_that("int_unexp_records_set works", {
  skip_on_cran() # slow
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")

  r <- int_unexp_records_set(
    level = "segment",
    study_data = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS",
    meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship_meta_v2.xlsx",
    label_col = LABEL
  )
  expect_snapshot(r)
})

#' Count Expected Observations
#'
#' Count participants, if an observation was expected, given the
#' `PART_VARS` from item-level metadata
#'
#' @param resp_vars [character] the response variables, for that a value may be
#'                              expected
#' @param study_data [study_data]
#' @param meta_data [meta_data]
#' @param label_col [character] mapping attribute `colnames(study_data)` vs.
#'                              `meta_data[label_col]`
#' @param expected_observations [enum] HIERARCHY | ALL | SEGMENT. How should
#'                                     `PART_VARS` be handled:
#'                                     - `ALL`: Ignore, all observations are
#'                                       expected
#'                                     - `SEGMENT`: if `PART_VAR` is 1, an
#'                                       observation is expected
#'                                     - `HIERARCHY`: the default, if the
#'                                       `PART_VAR` is 1 for this variable and
#'                                       also for all `PART_VARS` of `PART_VARS`
#'                                       up in the hierarchy, an observation is
#'                                       expected.
#'
#' @return a vector with the number of expected observations for each
#'         `resp_vars`.
#'
#' @family missing_functions
#' @concept process
#' @noRd
util_count_expected_observations <- function(resp_vars, study_data, meta_data,
                                             label_col = LABEL,
                                             expected_observations =
                                               c("HIERARCHY",
                                                 "ALL",
                                                 "SEGMENT")) {
  vapply(resp_vars, function(rv) {
    sum(util_observation_expected(rv = rv,
                                  study_data = study_data,
                                  meta_data = meta_data,
                                  label_col = label_col,
                                  expected_observations =
                                    expected_observations),
        na.rm = TRUE)
  },
  FUN.VALUE = integer(1))
}

#' utility function for the applicability of contradiction checks
#'
#' Test for applicability of contradiction checks
#'
#' @param x [data.frame] metadata
#' @param dta [logical] vector, 1=matching data type,
#'                  0 = non-matching data type
#'
#' @return [factor] 0-3 for each variable in metadata
#' - 0 data type mismatch and not applicable
#' - 1 data type mismatches but applicable
#' - 2 data type matches but not applicable
#' - 3 data type matches and applicable
#'
#' @seealso [pro_applicability_matrix]
#' @noRd

util_app_con_contradictions_redcap <- function(x, dta) { # TODO: would need meta_data_cross_item
  c1 <- rep(1, times = dim(x)[1])

  aa <- paste0(dta, c1)
  score <- as.numeric(recode(as.factor(aa), "00" = 0, "01"
                             = 1, "10" = 2, "11" = 3))
  score <- as.factor(score)
  return(score)
}

#' Utility function to create bar plots
#'
#' A helper function for simple bar plots. The layout is intended for data with
#' positive numbers only (e.g., counts/frequencies).
#'
#' @param plot_data the data for the plot. It should consist of one column
#'                  specifying the categories, and a second column giving the
#'                  respective numbers / counts per category. It may contain
#'                  another column to specify the coloring of the bars
#'                  (`fill_var`).
#' @param cat_var column name of the categorical variable in `plot_data`
#' @param num_var column name of the numerical variable in `plot_data`
#' @param relative if `TRUE`, numbers will be interpreted as percentages
#'                 (values in `num_var` should lie within `[0,1]`)
#' @param show_numbers if `TRUE`, numbers will be displayed on top of the bars
#' @param fill_var column name of the variable in `plot_data` which will be used
#'                 to color the bars individually
#' @param colors vector of colors, or a single color
#' @param show_color_legend if `TRUE`, a legend for the colors will be displayed
#' @param flip if `TRUE`, bars will be oriented horizontally
#'
#' @return a bar plot
#'
#' @noRd
util_bar_plot <- function(plot_data, cat_var, num_var, relative = FALSE,
                          show_numbers = TRUE,
                          fill_var = NULL, colors = "#2166AC",
                          show_color_legend = FALSE,
                          flip = FALSE) {

  # base plot
  bar_plot <- util_create_lean_ggplot(
    ggplot(plot_data,
           aes(x = .data[[cat_var]],
               y = .data[[num_var]])) +
      theme_minimal() +
      xlab("") +
      ylab(""),
    plot_data = plot_data,
    cat_var = cat_var,
    num_var = num_var
  )

  # color setup, adding the bars
  if (!is.null(fill_var)) {
    if (is.numeric(plot_data[[fill_var]])) {
      if (relative) {
        scale_fill <- scale_fill_gradientn(colors = colors,
                                           labels = scales::percent,
                                           name = "")
      } else {
        scale_fill <- scale_fill_gradientn(colors = colors,
                                           name = "")
      }
    } else {
      scale_fill <- scale_fill_manual(values = colors, name = "")
    }
    bar_plot <- bar_plot %lean+%  util_create_lean_ggplot(
        geom_col(aes(fill = .data[[fill_var]]),
                 width = 0.8),
      fill_var = fill_var
    ) %lean+% scale_fill
  } else {
    bar_plot <- bar_plot %lean+% util_create_lean_ggplot(
        geom_col(fill = colors[1],
                 width = 0.8),
      colors = colors
    )
  }
  # y-axis setup, show numbers as percentages if needed
  if (relative) {
    scale_y <- util_create_lean_ggplot(scale_y_continuous(labels = scales::percent,
                                  expand = expansion(mult = c(0, 0.05))))
    num_labels <- paste0(round(plot_data[[num_var]] * 100, digits = 2), "%")
  } else {
    scale_y <- util_create_lean_ggplot(scale_y_continuous(expand = expansion(mult = c(0, 0.05))))
    num_labels <- as.character(plot_data[[num_var]])
  }
  bar_plot <- bar_plot %lean+% scale_y
  # some layout options must be adapted to the orientation of the plot:
  # - larger font size for category labels
  # - no grid lines parallel to the bars
  # - position of numbers above/below or next to the bars
  if (!flip) {
    scale_x <- util_create_lean_ggplot(scale_x_discrete(drop = FALSE,
                                expand = expansion(add = 0.5, mult = 0.1)))
    ly <- util_create_lean_ggplot(theme(axis.text.x = element_text(angle = 35, hjust = 1, size = 10),
                axis.text.y = element_text(size = 10),
                panel.grid.major.x = element_blank(),
                panel.grid.minor.x = element_blank()))
    if (show_color_legend) {
      gui <- guides(fill = guide_legend(position = "right"))
    } else {
      gui <- guides(fill = "none")
    }
    # set up label positions
    vert_just <- ifelse(plot_data[[num_var]] / max(plot_data[[num_var]]) < 0.4, # include potential threshold line here in max
                        -0.5, # above the bar
                        1.5) # within the bar
    horiz_just <- 0.5
    pos_labels <- ifelse(vert_just <= 0, "outside", "within")
  } else {
    bar_plot <- util_lazy_add_coord(bar_plot, coord_flip())
    scale_x <- util_create_lean_ggplot(scale_x_discrete(limits = rev,
                                drop = FALSE,
                                expand = expansion(add = 0.5, mult = 0.1)))
    ly <- util_create_lean_ggplot(theme(axis.text.y = element_text(size = 10),
                axis.text.x = element_text(size = 10),
            panel.grid.major.y = element_blank(),
            panel.grid.minor.y = element_blank()))
    if (show_color_legend) {
      gui <- util_create_lean_ggplot(guides(fill = guide_legend(position = "bottom", byrow = TRUE)))
    } else {
      gui <- util_create_lean_ggplot(guides(fill = "none"))
    }
    # set up label positions
    vert_just <- 0.5
    horiz_just <- ifelse(plot_data[[num_var]] / max(plot_data[[num_var]]) < 0.4, # include potential threshold line here in max
                         0, # on the right side next to the bar
                         1) # within the bar
    pos_labels <- ifelse(horiz_just == 0, "outside", "within")
    num_labels <- paste0(" ", num_labels, " ")
  }
  bar_plot <- bar_plot %lean+% scale_x %lean+% ly %lean+% gui

  # color setup for labels on/next to the bars
  if (show_numbers) {
    ld <- ggplot2::layer_data(bar_plot)
    col_bars <- ld$fill # hex code
    brightness <- apply(col2rgb(col_bars), 2, function(rgb_col) {
      (299 * rgb_col[1] +
         587 * rgb_col[2] +
         114 * rgb_col[3]) / 1000
    })
    # show label text in black or white, depending on the brightness of the background, https://stackoverflow.com/questions/11867545/change-text-color-based-on-brightness-of-the-covered-background-area, https://www.w3.org/TR/AERT/#color-contrast
    col_labels <- ifelse(pos_labels == "within" & brightness < 130,
                         "white", "black") # this is referred to by util_as_plotly_util_plot_categorical_vars
    bar_plot <- bar_plot %lean+% util_create_lean_ggplot(
                                          geom_text(label = num_labels,
                                                    vjust = vert_just,
                                                    hjust = horiz_just,
                                                    color = col_labels,
                                                    size = 3.5),
                                        num_labels = num_labels,
                                        vert_just = vert_just,
                                        horiz_just = horiz_just,
                                        col_labels = col_labels)

  }
  return(bar_plot)
}

#' Convert factors to label-corresponding numeric values
#'
#' Converts a vector factor aware of numeric values not being scrambled.
#'
#' @param v the vector
#' @param warn if not missing: character with error message stating conversion
#'   error
#'
#' @return the converted vector
#'
#' @family conversion_functions
#' @concept data_management
#' @noRd
util_as_numeric <- function(v, warn) {
  if (is.factor(v) && !is.ordered(v)) {
    r <- suppressWarnings(as.numeric(levels(v))[v])
  } else {
    r <- suppressWarnings(as.numeric(v))
  }
  conversion_errors <- (is.na(v) | "" == trimws(v)) != is.na(r)
  v_conversion_errors <- v[conversion_errors]
  if (length(v_conversion_errors) > 5) {
    v_conversion_errors <- head(v_conversion_errors, 5)
    dot <- ", ..."
  } else {
    dot <- ""
  }
  if (any(conversion_errors)) {
    if (is.factor(v)) {
      util_warning("Could not convert %s%s to numeric values",
                   paste0(dQuote(v_conversion_errors), collapse = ", "), dot,
                   applicability_problem = FALSE)
    }
    if (!missing(warn) && length(warn) == 1 && is.character(warn) &&
        !is.na(warn)) {
      util_warning(warn, paste0(dQuote(v_conversion_errors), collapse = ", "),
                   applicability_problem = FALSE)
    }
  }
  r
}

test_that("util_as_numeric works", {
  skip_on_cran()
  a <- dataquieR:::util_as_numeric(1:10,
                             warn = TRUE)
  b <- dataquieR:::util_as_numeric(as.character(1:10),
                              warn = TRUE)
  c <- dataquieR:::util_as_numeric(factor(1:10,
                                     levels = 10:1,
                                     labels = as.character(1:10)),
                              warn = TRUE)
  expect_warning(
       d <- dataquieR:::util_as_numeric(factor(1:10,
                                     levels = 10:1,
                                     labels =
                                       paste("Grade", as.character(1:10))),
                              warn = TRUE),
       regexp = paste(
         "Could not convert .+Grade 10.+, .+Grade 9.+, .+Grade 8.+,",
         ".+Grade 7.+, .+Grade 6.+, ... to numeric values"),
       perl = TRUE)

  e <- dataquieR:::util_as_numeric(ordered(1:10,
                                     levels = 10:1,
                                     labels = as.character(1:10)),
                              warn = TRUE)

  expect_equal(a, 1:10)
  expect_equal(b, 1:10)
  expect_equal(c, 10:1)
  expect_equal(d, rep(NA_real_, 10))
  expect_equal(e, 10:1)
})

#' Utility function for judging whether a character vector does not appear to
#' be a categorical variable
#'
#' The function considers the following properties:
#' - the maximum number of characters (to identify free text fields with
#' long entries),
#' - the relative frequency of punctuation and space characters per element (to
#' identify, e.g., JSON or XML elements, which are structured by those
#' characters),
#' - the relative frequency of elements (categorical variables would have
#' a low proportion of unique values in comparison to other variables).
#'
#' @param vec a character vector
#'
#' @return TRUE or FALSE
#'
#' @noRd
util_string_is_not_categorical <- function(vec) {
  vec <- vec[!util_empty(vec)]
  if (length(vec) == 0) {
    return(TRUE)
  }
  if (!is.character(vec)) {
    util_warning("Wrong use of util_string_is_not_categorical!")
    return(FALSE)
  }
  vec_uniq <- unique(vec)
  # free-text fields can contain very long strings
  long_elements <- max(nchar(vec_uniq,
                             type = "bytes", # must also work for wrong encoding
                             allowNA = TRUE,
                             keepNA = FALSE)) > 100
  # JSON, XML or similar elements contain more punctuation symbols and possibly
  # space characters than expected for categorical variables
  many_non_alphanum_char <- median(
    vapply(vec_uniq, FUN.VALUE = numeric(1),
         function(vv) {
           sum(
             grepl('[[:punct:]|[:space:]]',
                   unlist(strsplit(vv, split = ""))
                   )
             ) / nchar(vv,
                       type = "bytes", # must also work for wrong encoding
                       allowNA = TRUE,
                       keepNA = FALSE)
         })
  ) > 0.4
  # For categorical variables, we expect a low proportion of unique values. So
  # if there are only few duplicates, the vector is most likely not a
  # categorical variable.
  if (length(vec) >= 20) {
    few_duplicates <- (length(vec) - length(vec_uniq)) / length(vec) < 0.2
  } else {
    few_duplicates <- FALSE
  }
  return(long_elements | many_non_alphanum_char | few_duplicates)
}

#' Plot to un-disclosed `ggplot` object
#'
#' @param expr plot expression
#' @param w width in cm
#' @param h height in cm
#'
#' @return `ggplot` object, but rendered (no original data included)
#'
#' @noRd
util_plot2svg_object <- function(expr, w = 21.2, h = 15.9, sizing_hints) {
  util_ensure_suggested("grImport2")
  util_ensure_suggested("rsvg")
  orig_sizing_hints <- sizing_hints
  if (!is.null(sizing_hints)) {
    sizing_hints <- util_finalize_sizing_hints(sizing_hints = sizing_hints)
    if (!is.null(sizing_hints) &&
        !is.null(sizing_hints$h_in_cm) &&
        !is.null(sizing_hints$w_in_cm)) {
      w <- sizing_hints$w_in_cm
      h <- sizing_hints$h_in_cm
    }
  }
  tmpfil <- NULL
  withr::with_tempfile("tmpfil", fileext = ".svg", {
    htmltools::capturePlot(expr = {
      rlang::eval_tidy(expr)
    },
    filename = tmpfil,
    device = grDevices::svg,
    width = w / 2.54, height = h / 2.54,
    #    pointsize = 12
    #    res = 1/72
    )
    .svg <- readLines(tmpfil)
    .svg <- gsub("<svg ", '<svg preserveAspectRatio="none" ', .svg, fixed = TRUE)
    # writeLines(.svg, tmpfil)
    rsvg::rsvg_svg(charToRaw(paste0(.svg, collapse = "\n")), tmpfil)
    # res <- as.environment(list(
    #   x = magick::image_read_svg(tmpfil, width = w, height = h)
    # ))
    # class(res) <- "dataquieR_undisclosed_figure"
    raw <- grImport2::readPicture(tmpfil)
    res <- util_svg_plot_proxy(tmpfil)
    attr(res, "sizing_hints") <- orig_sizing_hints
    return(res)
  })
}

#' `Plotly` to un-disclosed `ggplot` object
#'
#' @param plotly the object
#' @param w width in cm
#' @param h height in cm
#'
#' @return `ggplot` object, but rendered (no original data included)
#'
#' @noRd
util_plotly2svg_object <- function(plotly, w = 21.2, h = 15.9, sizing_hints) { # FIXME: Also for thumbnails, if not ggplot exists.
  util_ensure_suggested("grImport2")
  util_ensure_suggested("rsvg")
  util_ensure_suggested("plotly")
  util_ensure_suggested("reticulate")
  orig_sizing_hints <- sizing_hints
  if (!is.null(sizing_hints)) {
    sizing_hints <- util_finalize_sizing_hints(sizing_hints = sizing_hints)
    if (!is.null(sizing_hints) &&
        !is.null(sizing_hints$h_in_cm) &&
        !is.null(sizing_hints$w_in_cm)) {
      w <- sizing_hints$w_in_cm
      h <- sizing_hints$h_in_cm
    }
  }
  # install.packages('reticulate')
  # reticulate::install_miniconda()
  # reticulate::conda_install('r-reticulate', 'python-kaleido')
  # reticulate::conda_install('r-reticulate', 'plotly', channel = 'plotly')
  # reticulate::use_miniconda('r-reticulate')

  tmpfil <- NULL
  withr::with_tempfile("tmpfil", fileext = ".svg", {

    fn <- try({
      plotly::save_image(p = plotly, file = tmpfil,
                         width = w / 2.54 * 96,
                         height = h / 2.54 * 96)
    }, silent = TRUE)

    if (util_is_try_error(fn)) {
      util_error(
        c("Could not use %s to convert a plotly to a static image:\n",
          "%s",
          "\nYou can try to fix that by setting up reticulate properly and",
          "setting everything up as described in %s. You can also file a bug",
          "report, because this shold be needed only as a fallback during",
          "the development of new indicator functions."
        ),
        sQuote("plotly::save_image()"),
        dQuote(conditionMessage(attr(fn, "condition"))),
        sQuote("? plotly::save_image")
      )
      # For STS: reticulate::use_python("/Users/struckmanns/Library/r-miniconda-arm64/envs/r-reticulate/bin/python", required = T)
    }

    .svg <- readLines(tmpfil, warn = FALSE)
    .svg <- gsub("<svg ", '<svg preserveAspectRatio="none" ', .svg, fixed = TRUE)
    # writeLines(.svg, tmpfil)
    rsvg::rsvg_svg(charToRaw(paste0(.svg, collapse = "\n")), tmpfil)
    res <- util_svg_plot_proxy(tmpfil)
    attr(res, "sizing_hints") <- orig_sizing_hints
    return(res)
  })
}

#' Detect un-disclosed `ggplot`
#'
#' @param x the object to check
#'
#' @return `TRUE` or `FALSE`
#' @noRd
util_is_svg_object <- function(x) {
  inherits(x, "svg_plot_proxy") || (
    util_is_gg_plot(x) &&
      all(vapply(lapply(x$layers, `[[`, "geom"),
                 inherits, "GeomDrawGrob", FUN.VALUE = logical(1)))
  )
}

util_svg_plot_proxy <- function(svg_file) {
  svg_raw <-
    charToRaw(paste(readLines(svg_file, warn = FALSE), collapse = "\n"))
  structure(list(svg = svg_raw),
            class = "svg_plot_proxy")
}

#' @exportS3Method grid::grid.draw
grid.draw.svg_plot_proxy <- function(x, ...) {
  util_ensure_suggested("grImport2")

  tmp_svg <- tempfile(fileext = ".svg")
  on.exit(unlink(tmp_svg), add = TRUE)
  writeBin(x$svg, tmp_svg)

  pic <- grImport2::readPicture(tmp_svg)
  grob <- grImport2::pictureGrob(pic, clip = "off", distort = TRUE)
  grid::grid.draw(grob)
}

#' @export
print.svg_plot_proxy <- function(x, ...) {
  grid::grid.newpage()
  grid::grid.draw(x)
  invisible(x)
}

#' Create a dynamic single variable page for the report
#'
#' @param results [list] a list of subsets of the report matching `cur_var`
#' @param cur_var [character] variable name for single variable pages
#' @param use_plot_ly [logical] use `plotly`, if available.
#' @param template [character] template to use for the `dq_report2` report.
#' @param note_meta [character] notes on the metadata for a single variable
#'                              (if needed)
#' @param rendered_repsum the `dataquieR` summary, see [summary()],
#'   [dq_report2()] and [print.dataquieR_summary()]
#' @param dir [character] output directory for potential `iframes`.
#'
#'
#' @return list of arguments for `append_single_page()` defined locally in
#'   `util_generate_pages_from_report()`.
#'
#' @keywords internal
util_html_for_var <- function(results, cur_var, use_plot_ly, template,
                              note_meta = c(), rendered_repsum, dir,
                              meta_data, label_col,
                              dims_in_rep,
                              clls_in_rep,
                              function_alias_map) {

  fname <-
    sprintf("VAR_%s.html", prep_link_escape(cur_var, #remove special characters from variable name
                                          html = TRUE))

  if (getOption("dataquieR.resume_print", dataquieR.resume_print_default)) {
    # TODO: [html_files_should_also_work_alone] Check, if we have the html file, already, see util_create_page_file
    # do not re-compute, then
  }
  .fn <- file.path(dir, sub("\\.html$", ".RDS", fname))
  if (getOption("dataquieR.resume_print", dataquieR.resume_print_default)
      && file.exists(.fn)) {
    cached <- try(readRDS(.fn), silent = TRUE)
    if (!util_is_try_error(cached)) {
      return(cached)
    }
  }

  util_stop_if_not(is.list(results))
  util_stop_if_not(!missing("cur_var") && length(cur_var) == 1 &&
                     !is.na(cur_var) &&
                      is.character(cur_var))


  rep_sum_el <- htmltools::HTML("")
  this <- attr(rendered_repsum, "this")
  repsum_wide <- attr(rendered_repsum, "repsum_wide")
  Variables <- util_map_labels(repsum_wide$Variables, meta_data, label_col,
                  attr(repsum_wide, "label_col"))
  if (any(Variables == cur_var)) {
    rs <- repsum_wide[Variables == cur_var,
                 setdiff(colnames(repsum_wide), "Variables"), FALSE]
    rs <- data.frame(x = colnames(rs), v = unname(unlist((rs[1, , TRUE]))))
    rs <- rbind.data.frame(rs[rs$x == "Total", , drop = FALSE],
                           rs[rs$x != "Total", , drop = FALSE])
    rs <-
      rs[!grepl("background: #ffffff", rs[, 2], fixed = TRUE), , drop = FALSE]
    rs <-
      rs[!is.na(rs[, 2]), , drop = FALSE]

    texts <- vapply(rs[, 1, drop = TRUE],
                    util_alias2caption, long = TRUE, FUN.VALUE = character(1))

    x <- rs[, 2, drop = TRUE]

    # 1) href="..."
    href1 <- sub('^.*href="([^"]+)".*$', "\\1", x)

    # nur echte Treffer behalten
    href1[href1 == x] <- NA_character_

    # 2) window.location = &quot;...&quot;
    href2 <- sub(
      '^.*window\\.location\\s*=\\s*&quot;([^&]+)&quot;.*$',
      "\\1",
      x
    )
    href2[href2 == x] <- NA_character_

    # 3) showDataquieRResult(&quot;...&quot;, &quot;TARGET&quot;, &quot;...&quot;)
    href3 <- sub(
      '^.*showDataquieRResult\\(&quot;[^&]*&quot;\\s*,\\s*&quot;([^&]+)&quot;.*$',
      "\\1",
      x
    )
    href3[href3 == x] <- NA_character_

    # Fallback-Kaskade pro Zeile
    href_final <- ifelse(
      !is.na(href1), href1,
      ifelse(!is.na(href2), href2, href3)
    )

    links <- vapply(
      mapply(
        href = href_final,
        text = texts,
        no_res = grepl("No results available", rs[, 2]),
        SIMPLIFY = FALSE,
        FUN = function(href, text, no_res) {
          if (any(grepl(".", href, fixed = TRUE)) && !no_res) {
            paste0("<a href=\"", href, "\">", text, "</a>")
          } else {
            paste0("<a style=\"color:black;cursor:default;text-decoration:none;\">", text, "</a>")
          }
        }
      ),
      FUN = paste0, collapse = "",
      FUN.VALUE = character(1)
    )

    rs[, 1] <- links

    # Remove onclick from a elements
    rs[, 2] <- gsub("(<a\\b[^>]*?)\\s+onclick\\s*=\\s*(['\"]).*?\\2", "\\1", rs[, 2], perl = TRUE, ignore.case = TRUE)
    rs <- rs[!is.na(href3), , FALSE]

    if (!!prod(dim(rs))) {
      rep_sum_el <- htmltools::browsable(util_html_table(rs, filter = "top", options = list(scrollCollapse = TRUE, scrollY = "75vh"),
                                           is_matrix_table = TRUE, rotate_headers = TRUE,
                                           meta_data = this$meta_data,
                                           label_col = this$label_col,
                                           dl_fn = "DQ Report Summary",
                                           link_variables = FALSE, colnames = c(" ", " ")))
    }
  }

  rep_sum_el <- htmltools::tagList(
    anchor = util_generate_anchor_tag(cur_var, "Summary", "variable"),
    link = util_generate_anchor_link(cur_var, "Summary", "variable"), # these links are moved later
    rep_sum_el
  )

  svp <- list()

  for (cur_dm in dims_in_rep) {
    # util_message("Dimension %s", cur_dm)
    drop_down <- dims[[cur_dm]] # fetches the name of the content for the first-level drop-down menu
    j <- 0
    m <- length(clls_in_rep[[cur_dm]])
    for (cll in clls_in_rep[[cur_dm]]) { # loop over the functions in each dimension
      fkt <- util_cll_nm2fkt_nm(cll, function_alias_map = function_alias_map)
      all_of_f <- results[paste0(cll, '.', cur_var)] # extract the single result
      all_of_f <- # remove non-existing results
        all_of_f[!vapply(all_of_f, is.null, FUN.VALUE = logical(1))]

     # use results as they are, not combined by util_combine_res. see else branch below for comments on mapply
      output <- do.call(htmltools::tagList, unname(mapply(dqr = all_of_f,
                                                          nm =
                                                            names(all_of_f),
                                                          FUN =
                                                            util_pretty_print,
                                                          MoreArgs =
                                                            list(
                                                              meta_data =
                                                                meta_data,
                                                              label_col =
                                                                label_col,
                                                              use_plot_ly =
                                                                use_plot_ly,
                                                              is_single_var =
                                                                TRUE,
                                                              dir = dir),
                                                          SIMPLIFY = FALSE)))
      # if (cll == "con_inadmissible_categorical" && cur_var == "SBP_0") browser()
      # if we have some result, then store it in the svp list for the current single variable page
      if (length(output) && !all(vapply(output, is.null, FUN.VALUE = logical(1)))) {
        # title <- .manual$titles[[fkt]]

        title <- util_alias2caption(cll, long = TRUE)
        description <- util_function_description(fkt)

        anchor <- output[[1]]$anchor # extract anchor tag and move it above the caption of the current result
        output[[1]]$anchor <- htmltools::span() # delete anchor tag from its orignal position

        svp[[cll]] <- htmltools::tagList(
          htmltools::span( # span is requird to make [[2]][[1]]$link on page navi menu creation below work
            anchor, # put the anchor tag on top
            htmltools::h4(style = "margin-top: 3em", # add function name as a ~subtitle, add space to prevent plot.lys to overlap
                          htmltools::a(
                            target = "_blank",
                            href = util_online_ref(fkt),
                            title)),
            htmltools::HTML("<!-- "),
            htmltools::h5(paste(unique(c(cll, fkt)),
                                collapse = ": ")),
            htmltools::HTML(" -->"),
            htmltools::div(class="infobutton", htmltools::tagList(
                           htmltools::HTML(description),
                           htmltools::h6("Concept relations:"),
                           htmltools::p(util_get_concept_links(fkt))
                           ))),
          output)
      } else {
        svp[[cll]] <- NULL
      }

    }

  }


  ..vn <- prep_map_labels(cur_var,
                          meta_data = meta_data,
                          to = VAR_NAMES,
                          from = label_col,
                          ifnotfound = "",
                          warn_ambiguous = FALSE)
  ..lbs <- prep_get_labels(cur_var,
                           meta_data = meta_data,
                           label_col = label_col,
                           resp_vars_match_label_col_only = TRUE,
                           label_class = "SHORT")
  if (startsWith(..lbs, ..vn)) {
    ..vn <- ""
  }
  title <- htmltools::tagList(
    htmltools::h1(prep_get_labels(cur_var,
                                  max_len = 9999999,
                                  meta_data = meta_data,
                                  label_col = label_col,
                                  resp_vars_match_label_col_only = TRUE,
                                  label_class = "LONG")),
     htmltools::h2(paste(
       ..vn,
      ..lbs
       ))
  )

  cur_md <- meta_data[meta_data[[label_col]] == cur_var,
                      , FALSE] # subset the metadata for the current variable to show it in the report

  { # link tables to their representation in this report-site
    cur_md[, !grepl("_TABLE$", colnames(cur_md))] <-
      util_df_escape(cur_md[, !grepl("_TABLE$", colnames(cur_md)),
                            drop = FALSE])

    for (cn in grep("_TABLE$", colnames(cur_md), value = TRUE)) {

      cur_md[!util_empty(cur_md[[cn]]), cn] <- vapply(FUN.VALUE = character(1),
                                                      cur_md[!util_empty(cur_md[[cn]]), cn],
                                                      FUN = function(tn) {
                                                        paste(as.character(htmltools::a(href = paste0(
                                                          prep_link_escape(tn, html = TRUE), ".html"), tn)),
                                                          collapse = "")
                                                      }
      )

    }

  }

  md_info <- htmltools::tagList(## link tables
    anchor = util_generate_anchor_tag(cur_var, "meta_data", "variable"),
    link = util_generate_anchor_link(cur_var, "meta_data", "variable"), # these links are moved later
    htmltools::h1("Item-level Metadata"),
    htmltools::div(note_meta),
    util_html_table(data.frame(Name = names(cur_md),
                               Value = unname(t(cur_md))),
                    meta_data = meta_data,
                    label_col = label_col,
                    copy_row_names_to_column = FALSE,
                    dl_fn = paste("Item-level_Metadata_", cur_var))
  )

  svp <- c(list(rep_sum_el = list("", list(rep_sum_el))), # add the rep_sum_el info to the svp keeping the same structure of svp
           svp,
           list(md_info = list("", list(md_info))) # add the metadata info to the svp keeping the same structure of svp
           )

  body <- do.call(htmltools::tagList, unname(svp))
  # move the links to the correct position to create the page navigation menu
  # this time variable-wise (see code above for comments)
  all_links <- lapply(body, function(x) x[[2]][[1]]$link)
  all_links <- all_links[vapply(all_links, length,
                                FUN.VALUE = integer(1)) != 0]
  all_links <- lapply(all_links, htmltools::tags$li)
  all_links <- do.call(htmltools::tagList, all_links)

  page_menu <- util_float_index_menu(object = all_links)
  body <- lapply(body, function(x) {
    x[[2]][[1]]$link <- NULL
    x
  })
  body <- htmltools::tagList(page_menu, body)

  add_anchor <- util_generate_anchor_tag(cur_var, "", "variable")

  is_var_page_js <- htmltools::tags$script(type = "text/javascript",
                                           htmltools::HTML("window.is_svp = true;"))

  page <- htmltools::tagList(is_var_page_js, add_anchor, title, body)

  title <- prep_get_labels(cur_var,
                           meta_data = meta_data,
                           label_col = label_col,
                           max_len = 9999999,
                           resp_vars_match_label_col_only = TRUE,
                           label_class = "LONG")
  if (startsWith(x = title, prefix = sprintf("%s:", names(title)))) {
    title <- sprintf("%s", title)
  } else {
    title <- sprintf("%s: %s", names(title), title)
  }


  r <- list(list("Single Variables",
                 util_attach_attr(title,
                                  alternative_names =
                                    as.vector(unlist(
                                      cur_md[, startsWith(colnames(cur_md), LABEL) |
                                               startsWith(colnames(cur_md), LONG_LABEL) |
                                               colnames(cur_md) == VAR_NAMES,
                                             drop = TRUE]))),
                 fname,
                 page))

  if (getOption("dataquieR.resume_print", dataquieR.resume_print_default)) {
    saveRDS(r, file = .fn, compress = "xz")
  }

  return(r)

}

#' Generate a stratified full DQ report
#'
#' @param resp_vars [variable] the names of the measurement variables, if
#'                             missing or `NULL`, all variables will be included
#' @param id_vars [variable] a vector containing the name/s of the variables
#'                            containing ids, to
#'                            be used to merge multiple data frames if provided
#'                            in `study_data` and to be add to referred vars
#' @param study_data [data.frame] the data frame that contains the measurements:
#'                                it can be an R object (e.g., `bia`), a
#'                                data frame (e.g., `"C:/Users/data/bia.dta"`),
#'                                a vector containing data frames files (e.g.,
#'                                c(`"C:/Users/data/bia.dta"`,
#'                                `C:/Users/data/biames.dta"`)), or it can be
#'                                left empty and the data frames are provided
#'                                in the data frame level metadata. If only the
#'                                file name without path is provided
#'                                (e.g., `"bia.dta"`), the file
#'                                name needs the extension and the path must be
#'                                provided in the argument `input_dir`. It can
#'                                also contain only the file name in case of
#'                                example data from the package `dataquieR`
#'                                (e.g., `"study_data"` or `"ship"`)
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param meta_data_segment [data.frame] -- optional: Segment level metadata
#' @param meta_data_dataframe [data.frame] -- optional if `study_data` is
#'                                            present: Data frame level metadata
#' @param meta_data_cross_item [data.frame] -- optional: Cross-item level
#'                                                                 metadata
#' @param meta_data_item_computation [data.frame] -- optional: Computed items
#'                                                                     metadata
#' @param ... arguments to be passed through to [dq_report] or [dq_report2]
#' @param label_col [variable attribute] the name of the column in the
#'                                       metadata containing the labels of
#'                                       the variables
#' @param meta_data_v2 [character] path or file name of the workbook like
#'                                 metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`
#' @param segment_column [variable attribute] name of a metadata attribute
#'                                             usable to split the report in
#'                                             sections of variables, e.g. all
#'                                             blood-pressure related variables.
#'                                             By default,
#'                                             reports are split by
#'                                             [STUDY_SEGMENT] if available and
#'                                             no segment_column nor
#'                                             strata_column or subgroup
#'                                             are defined.
#'                                             To create an un-split report
#'                                             please write explicitly
#'                                             the argument
#'                                             'segment_column = NULL'
#' @param strata_column [variable] name of a study variable to stratify the
#'                                    report by, e.g. the study centers.
#'                                    Both labels and `VAR_NAMES` are accepted.
#'                                    In case of NAs in the selected variable,
#'                                    a separate report containing the NAs
#'                                    subset will be created
#' @param strata_select [character] if given, the strata of strata_column
#'                                       are limited to the content of this
#'                                       vector. A character vector or a regular
#'                                       expression can be provided
#'                                       (e.g., "^a.*$"). This argument can not
#'                                       be used if no strata_column is
#'                                       provided
#' @param segment_select [character] if given, the levels of segment_column
#'                                       are limited to the content of this
#'                                       vector. A character vector or a
#'                                       regular expression (e.g., ".*_EXAM$")
#'                                       can be provided.
#'                                       This argument can not be used if no
#'                                       segment_column is provided.
#' @param input_dir [character] if given, the study data files that have
#'                              no path and that are not URL are searched in
#'                              this directory. Also `meta_data_v2` is searched
#'                              in this directory if no path is provided
#' @param advanced_options [list] options to set during report computation,
#'                                see [options()]
#' @param output_dir [character] if given, the output is not returned but saved
#'                               in this directory
#' @param missing_tables [character] the name of the data frame containing the
#'                                   missing codes, it can be a vector if more
#'                                   than one table is provided. Example:
#'                                   `c("missing_table1", "missing_table2")`
#' @param also_print [logical] if `output_dir` is not `NULL`, also create
#'                             `HTML` output for each report using
#'                             [print.dataquieR_resultset2()]
#'                               written to the path `output_dir`
#' @param disable_plotly [logical] do not use `plotly`, even if installed
#' @param selection_type [character] optional, can only be specified if a
#'                                              `strata_select` or
#'                                              `strata_exclude` is specified.
#'                                              If not present the
#'                                              function try to guess what the
#'                                              user typed as `strata_select` or
#'                                               `strata_exclude`.
#'                                              There are 3 options:
#'                                              `value` indicating that the
#'                                              stratum selected is a value and
#'                                              not a value_label.
#'                                              For example `"0"`;
#'                                              `v_label` indicating that the
#'                                              stratum specified is a label.
#'                                              For example `"male"`.
#'                                              `regex` indicating that the user
#'                                              specified strata using a regular
#'                                              expression. For example `"^Ber"`
#'                                              to select all strata starting
#'                                              with that letters
#' @param segment_exclude [character] optional, can only be specified if a
#'                                              `segment_column` is specified.
#'                                               The levels of `segment_column`
#'                                               will not include the content of
#'                                               this argument.
#'                                               A character vector or
#'                                               a regular  expression can be
#'                                               provided  (e.g., "^STU").
#' @param strata_exclude [character] optional, can only be specified if a
#'                                              `strata_column` is specified.
#'                                              The strata of `strata_column`
#'                                               will not include the content of
#'                                               this argument.
#'                                               A character vector or
#'                                               a regular  expression can be
#'                                               provided  (e.g., "^STU").
#' @param subgroup [character] optional, to define subgroups of cases. Rules are
#'                                      to be written as `REDCap` rules.
#'                                      Only VAR_NAMES are accepted in the rules.
#' @param cross_item_level [data.frame] alias for `meta_data_cross_item`
#' @param `cross-item_level` [data.frame] alias for `meta_data_cross_item`
#' @param segment_level [data.frame] alias for `meta_data_segment`
#' @param dataframe_level [data.frame] alias for `meta_data_dataframe`
#' @param item_computation_level [data.frame] alias for
#'                               `meta_data_item_computation`
#' @param storr_factory [function] `NULL`, or
#'                        a function returning a `storr` object as
#'                        back-end for the report's results. If used with
#'                        `cores > 1`, the storage must be accessible from all
#'                        cores and capable of concurrent writing according
#'                        to `storr`. Hint: `dataquieR` currently only supports
#'                        `storr::storr_rds()`, officially, while other back-
#'                        ends may nevertheless work, yet, they are not tested.
#' @param amend [logical] if there is already data in.`storr_factory`,
#'                        use it anyways -- unsupported, so far!
#' @param checkpoint_resumed [logical] if using a `storr_factory` and the back-
#'                                     end there is already filled, and if
#'                                     `amend` is missing or set to `TRUE`,
#'                                     compute all missing result and add them
#'                                     to the back-end.
#' @param view [logical] open the returned report
#'
#' @return A named [list] of named [list]s of [dq_report2] reports, returned
#'         invisibly unless `view = TRUE`. If `output_dir` is given, the result
#'         is still returned (invisibly), and optionally opened in a browser
#'         (`view = TRUE`, `also_print = TRUE`).
#'
#' @param force_overwrite [logical] force to overwrite `output_dir`, even if it
#'                                 exists
#' @param title [character] optional argument to specify the title for
#'                          the data quality report bundle
#' @param subtitle [character] optional argument to specify a subtitle for
#'                             the data quality report bundle
#' @param user_info [list] additional info stored with the report bundle,
#'                          e.g., comments, title, ...
#' @param author [character] author for the report bundle's documents.
#'
#' @seealso [dq_report]
#'
#' @export
#'
#' @examples
#' \dontrun{ # really long-running example.
#' prep_load_workbook_like_file("meta_data_v2")
#' rep <- dq_report_by("study_data", label_col =
#'   LABEL, strata_column = "CENTER_0")
#' rep <- dq_report_by("study_data",
#'   label_col = LABEL, strata_column = "CENTER_0",
#'   segment_column = NULL
#' )
#' unlink("/tmp/testRep/", force = TRUE, recursive = TRUE)
#' dq_report_by("study_data",
#'   label_col = LABEL, strata_column = "CENTER_0",
#'   segment_column = STUDY_SEGMENT, output_dir = "/tmp/testRep"
#' )
#' unlink("/tmp/testRep/", force = TRUE, recursive = TRUE)
#' dq_report_by("study_data",
#'   label_col = LABEL, strata_column = "CENTER_0",
#'   segment_column = NULL, output_dir = "/tmp/testRep"
#' )
#' dq_report_by("study_data",
#'   label_col = LABEL,
#'   segment_column = STUDY_SEGMENT, output_dir = "/tmp/testRep"
#' )
#' dq_report_by("study_data",
#'   label_col = LABEL,
#'   segment_column = STUDY_SEGMENT, output_dir = "/tmp/testRep",
#'   also_print = TRUE
#' )
#' dq_report_by(study_data = "study_data", meta_data_v2 = "meta_data_v2",
#'   advanced_options = list(dataquieR.study_data_cache_max = 0,
#'   dataquieR.study_data_cache_metrics = TRUE,
#'   dataquieR.study_data_cache_metrics_env = environment()),
#'   cores = NULL, dimensions = "int")
#' dq_report_by(study_data = "study_data", meta_data_v2 = "meta_data_v2",
#'   advanced_options = list(dataquieR.study_data_cache_max = 0),
#'   cores = NULL, dimensions = "int")
#' }
dq_report_by <- function(study_data,
                         item_level = "item_level",
                         meta_data_segment = "segment_level",
                         meta_data_dataframe = "dataframe_level",
                         meta_data_cross_item = "cross-item_level",
                         meta_data_item_computation = "item_computation_level",
                         missing_tables = NULL,
                         label_col,
                         meta_data_v2,
                         segment_column = NULL,
                         strata_column = NULL,
                         strata_select = NULL,
                         selection_type = NULL,
                         segment_select = NULL,
                         segment_exclude = NULL,
                         strata_exclude = NULL,
                         subgroup = NULL,
                         resp_vars = character(0),
                         id_vars = NULL,
                         advanced_options =  list(),
                         storr_factory = NULL,
                         amend = FALSE,
                         checkpoint_resumed =
                           getOption("dataquieR.resume_checkpoint",
                                     dataquieR.resume_checkpoint_default),
                         ...,
                         output_dir = NULL,
                         input_dir = NULL,
                         also_print = FALSE, # TODO: is this needed? or should this default to TRUE, whenever we have an output_dir?
                         force_overwrite = FALSE,
                         disable_plotly = FALSE,
                         view = TRUE,
                         meta_data = item_level,
                         cross_item_level,
                         `cross-item_level`,
                         segment_level,
                         dataframe_level,
                         item_computation_level,
                         author = prep_get_user_name(),
                         title = ifelse(is.null(output_dir),
                                        "Data quality report Bundle",
                                        paste0(basename(output_dir))),
                         subtitle = as.character(Sys.Date()),
                         user_info = NULL) {

  by_call <- rlang::caller_call(0)

  .outer_by_env$outer_by <- list(
    i = NA,
    n = NA,
    msg = "Preparing computation..."
  )

  on.exit({.outer_by_env$outer_by <- NULL}, add = TRUE)

  start_time <- Sys.time()

  rep_id <- util_make_report_id()

  if (!suppressWarnings(util_ensure_suggested("plotly",
                                              goal =
                                              "creating interactive figures",
                                              err = FALSE))) {
    if (!isTRUE(disable_plotly)) {
      util_message("Without the package plotly, you miss interactive figures.")
      disable_plotly <- TRUE
    }
  }

  dots <- rlang::dots_list(...)

  util_stop_if_not(is.list(advanced_options))
  util_expect_scalar(view, check_type = is.logical)

  old_O <- options(c(
    list(
      dataquieR.CONDITIONS_WITH_STACKTRACE = FALSE,
      dataquieR.ERRORS_WITH_CALLER = FALSE,
      dataquieR.MESSAGES_WITH_CALLER = FALSE,
      dataquieR.WARNINGS_WITH_CALLER = FALSE,
      DT.warn.size = FALSE
    ),
    advanced_options
  ))
  on.exit(options(old_O), add = TRUE)

  # store the call to use it later for the technical info in the reports
  call_report_by <- paste(deparse(sys.call()), collapse = "")

  # ?re-think the next line with
  # https://gitlab.com/libreumg/dataquier/-/issues/482
  # clear the study data cache
  # util_purge_study_data_cache()

  # Check the arguments (exception of segment_select and segment_exclude)----

  # check if both strata_column and strata_select are present
  if (!is.null(strata_select) && is.null(strata_column)) {
    util_error("strata_column is needed for selecting the strata")
  }

  # check if both strata_column and strata_exclude are present
  if (!is.null(strata_exclude) && is.null(strata_column)) {
    util_error("strata_column is needed for excluding the strata")
  }

  # check if both strata_select and strata_exclude are present
  # to use selection_type
  if (!is.null(selection_type) &&
      (is.null(strata_select) && is.null(strata_exclude))) {
    util_error(
      c(
        "selection_type can only be specified if",
        "strata_select or strata_exclude are present"
      )
    )
  }

  # check if the provided selection_type is acceptable
  if (!is.null(selection_type)) {
    util_stop_if_not(selection_type %in%
                       c("value", "v_label", "regex"),
                     label =
                       'The selection_type can only be "value", "v_label", or "regex"')
  }

  # check if subgroup rule is a string
  if (!is.null(subgroup)) {
    util_expect_scalar(
      arg_name = subgroup,
      allow_more_than_one = FALSE,
      check_type = is.character
    )
  }

  util_expect_scalar(force_overwrite, check_type = is.logical)

  ### check output directory only if also_print == TRUE
  # if users specify an output dir, check if it is a scalar, a character
  # string, check if also_print is a scalar and logical,
  # stop if the specified dir already exists
  # provide an error if the dir can not be created
  if((!missing (also_print) && also_print == TRUE) ) {
    if (!missing(output_dir)) {
      util_expect_scalar(output_dir, check_type = is.character)
      util_expect_scalar(also_print, check_type = is.logical)
      util_overwrite_if_requested(output_dir, force_overwrite)
      output_dir <- util_normalize_path(output_dir)
    }
  }

  .hi <- .hp <- .hm <- NULL
  content_file <- NULL
  if (!missing(output_dir) && also_print) {
    content_file <- file.path(output_dir, "index.html")
    list2env(util_init_html_progress(output_dir = output_dir,
                            content_file = content_file,
                            title = title,
                            view = view,
                            rep_id = rep_id), envir = environment())
  }

  ### check input directory
  # if users specify an input_dir, check if it is a scalar and a character
  # string, and stop if the specified dir does not exists
  # provide an error if the dir do not exists
  if (!missing(input_dir)) {
    util_expect_scalar(input_dir, check_type = is.character)
    if (!dir.exists(input_dir)) {
      util_error(
        "%s does not exist. Provide an %s containing the study data",
        dQuote(input_dir),
        sQuote("input_dir"))
    }
  }

  # check if resp_vars is a string vector
  if (length(resp_vars) > 0) {
    util_expect_scalar(
      arg_name = resp_vars,
      allow_more_than_one = TRUE,
      check_type = is.character
    )
  } else {
    resp_vars_complete <- character(0)
  }


  # Check and import item_level metadata ----
  ### load meta_data_v2 and item_level metadata
  # in case of presence of meta_data_v2 purge the cache and load it
  if (!missing(meta_data_v2)) {
    util_message("Have %s set, so I'll remove all loaded data frames",
                 sQuote("meta_data_v2"))


    #save the grading rule-sets and formats and re-add to cache after purge
    grading_rl <- NULL
    if (getOption("dataquieR.grading_rulesets",
                  dataquieR.grading_rulesets_default) %in%
        prep_list_dataframes()) {
      grading_rl <- prep_get_data_frame(getOption("dataquieR.grading_rulesets",
                                                  dataquieR.grading_rulesets_default))
    }
   format_rl <- NULL
    if (options("dataquieR.grading_formats" =
                dataquieR.grading_formats_default) %in%
        prep_list_dataframes()) {
      format_rl <- prep_get_data_frame(options("dataquieR.grading_formats" =
                                                 dataquieR.grading_formats_default))
    }
    prep_purge_data_frame_cache()
    if(!is.null(grading_rl)) {
      prep_add_data_frames(
        data_frame_list =
          setNames(list(grading_rl),
                   nm = getOption("dataquieR.grading_rulesets",
                                  dataquieR.grading_rulesets_default)))
    }
    if(!is.null(format_rl)) {
      prep_add_data_frames(
        data_frame_list =
          setNames(list(format_rl),
                   nm = options("dataquieR.grading_formats" =
                                  dataquieR.grading_formats_default)))
    }
    #try to import the metadata, and if not possible, try to add
    # the path if provided
    m <- try(prep_load_workbook_like_file(meta_data_v2), silent = TRUE)
    if (inherits(m, "try-error")) {
      # in case the name of the file is provided without a path and there is an
      # input_dir, the name is fixed adding the path at the beginning
      if (!is.null(input_dir)) {
        if (!grepl(.Platform$file.sep, meta_data_v2, fixed = TRUE)) {
          if (endsWith(input_dir, .Platform$file.sep)) {
            input_dir <- substr(input_dir, 1, nchar(input_dir) - 1)
          }
          meta_data_v2 <- file.path(input_dir, meta_data_v2)
        }
      }
      prep_load_workbook_like_file(meta_data_v2)
    }

    # check if item level metadata is in the cache and provide an error if not;
    # if it is not present predicts item-level from the data
    if (!is.data.frame(meta_data) &&
        (length(meta_data) != 1 || (!is.character(meta_data)) ||
         !exists(meta_data, .dataframe_environment()) )) {
      w <- paste(
        "Did not find any sheet named %s in %s, is this",
        "really dataquieR version 2 metadata?"
      )
      if (requireNamespace("cli", quietly = TRUE)) {
        w <- cli::bg_red(cli::col_br_yellow(w))
      }
      util_warning(w, dQuote(meta_data),
                   dQuote(meta_data_v2),
                   immediate = TRUE)
    }
  }

  # check if meta_data (item_level) is a data frame,
  # if not look for it in the cache and in the file system
  util_expect_data_frame(meta_data)

  util_ck_arg_aliases()

  name_sd <- character(0)

  # Getting name of file indicated by user to add it as argument in util_verify_names
  # (not including names from dataframe_level metadata)
  if (!missing(study_data)) {
    # 1. the study_data argument is specified by the user, and is only 1 data frame
    ## 1a. the data frame is an object in the environment
    if (is.data.frame(study_data)) {
       # get the name of the study data from the function call
      name_sd <- head(as.character(substitute(study_data)), 1)

    } else if (length(study_data) == 1 &&
              is.character(study_data)) {
      # 1b. the name of a data frame (not loaded yet) was indicated by the user
      # add the path before the name of the study data
      if (!is.null(input_dir)) {
        if (!grepl(.Platform$file.sep, study_data, fixed = TRUE)) {
          if (endsWith(input_dir, .Platform$file.sep)) {
            input_dir <- substr(input_dir, 1, nchar(input_dir) - 1)
          }
          name_sd <- file.path(input_dir, study_data)
        }
      } else {
        name_sd <- study_data
      }

    } else if (!is.data.frame(study_data) && length(study_data) > 1) {
      ## 1c. two or more study data indicated as vector by the user
      # if the study names are a vector of names without path and they are
      # not URL, add the input_dir before the names of each study data
      # create a list containing all names of the study data
      if (!is.null(input_dir)) {
        name_sd <- vapply(study_data, function(x) {
          res <- x
          if (!grepl(.Platform$file.sep, x, fixed = TRUE)) {
            if (endsWith(input_dir, .Platform$file.sep)) {
              input_dir <- substr(input_dir, 1, nchar(input_dir) - 1)
            }
            res <- file.path(input_dir, x)
          }
          return(res)
        }, FUN.VALUE = character(1))
        names(name_sd) <- NULL
      } else {
        name_sd <- study_data
      }
    }
  } else {
    name_sd <- character(0)
  }


  util_verify_names(name_of_study_data = name_sd)
  rm(name_sd)



  ##back-compatibility for column names in item_level_metadata
  my_args <- list(...)
  if ("cause_label_df" %in% names(my_args)) {
    cause_label_df <- my_args$cause_label_df
  } else {
    cause_label_df <- rlang::missing_arg()
  }

  # check if meta_data (item_level) contains the column VAR_NAMES
  suppressWarnings(util_ensure_in(
    VAR_NAMES,
    names(meta_data),
    error = TRUE,
    err_msg =
      sprintf("Did not find the mandatory column %%s in the %s.",
        sQuote("meta_data"))
  ))

  # define label_col as "LABEL" if they are not specified by users
  # if LABEL is missing then replace with VAR_NAMES
  if (missing(label_col) && (LABEL %in% names(meta_data))) {
    label_col <- LABEL
  } else if (missing(label_col) && !(LABEL %in% names(meta_data))) {
    label_col <- VAR_NAMES
  }

  # fix to rename columns from old metadata to new
  # (e.g., KEY_STUDY_SEGMENT > STUDY_SEGMENT)
  # even if cause_label_df is missing
  if (missing(cause_label_df)) {
    try(meta_data <- do.call(
      prep_meta_data_v1_to_item_level_meta_data,  #TODO: check very slow
      list(
        meta_data = meta_data,
        label_col = label_col,
        verbose = FALSE
      )
    ),
    silent = TRUE)
  } else {
    try(meta_data <- do.call(
      prep_meta_data_v1_to_item_level_meta_data,
      list(
        meta_data = meta_data,
        label_col = label_col,
        cause_label_df = cause_label_df,
        verbose = FALSE
      )
    ),
    silent = TRUE)
  }

  label_col_provided <- label_col
  mod_label <- util_ensure_label(meta_data = meta_data,  #here it creates VAR_NAMES_1
                                 label_col = label_col)
  if (!is.null(mod_label$label_modification_text)) {
    # There were changes in the metadata.
    meta_data <- mod_label$meta_data
    label_col <- mod_label$label_col
  }

  attr(meta_data, "normalized") <- TRUE
  attr(meta_data, "version") <- 2

  rm(my_args)

  # check if a column VARIABLE_ROLE is present in the metadata.
  # if not create one with always "primary" as role
  if (!(VARIABLE_ROLE %in% colnames(meta_data))) {
    util_message(
      "No %s assigned in item level metadata. Defaulting to %s.",
      sQuote(VARIABLE_ROLE),
      dQuote(VARIABLE_ROLES$PRIMARY),
      applicability_problem = TRUE)
    meta_data$VARIABLE_ROLE <- VARIABLE_ROLES$PRIMARY
  }

  util_stop_if_not(
    `Internal error, sorry: meta_data should be a data frame in dq_report_by. Please report` =
      is.data.frame(meta_data))

  # if DATAFRAMES is present in the item_level metadata
  # check if it is a character vector.
  if (DATAFRAMES %in% colnames(meta_data)) {
    util_expect_data_frame(meta_data,
                           col_names = list(DATAFRAMES = is.character))
  }

  # Check if id_vars is a string vector, and is present in item_level md----
  if (!is.null(id_vars)) {
    util_expect_scalar(
      arg_name = id_vars,
      allow_more_than_one = TRUE,
      check_type = is.character
    )

    #Check if id_vars is written with pipe, in case fix it
    if (length(id_vars) == 1) {
      id_vars <-
        unname(unlist(util_parse_assignments(id_vars,
                                             split_char = SPLIT_CHAR)))
    }
    # check if the id_vars is present in the item_level md
    id_vars <- vapply(id_vars, FUN = function(x){
      if (!x %in% meta_data$VAR_NAMES) {
        var_x <- x
        x <- try(util_map_labels(x,
                             meta_data = meta_data,
                             from = label_col,
                             to = VAR_NAMES), silent = TRUE)
        if (util_is_try_error(x)) {
          util_error(c("The id_vars %s is not present",
                       " in the item_level metadata"),
                     dQuote(var_x))
        } else {
          names(x) <- NULL
        }
        rm(var_x)
        return(x)
      } else {x}
    }, FUN.VALUE = character(1) )
    names(id_vars) <- NULL

  } else {
    #if not specified as argument by the user
    id_vars <- character(0)
  }


  # Check the arguments segment_select and segment_exclude----
  # check if both segment_column and segment_select are present
  if (!is.null(segment_select) && is.null(segment_column)) {

    if(STUDY_SEGMENT %in% colnames(meta_data)) {
      segment_column <- STUDY_SEGMENT
      possible_segments <- unique(meta_data$STUDY_SEGMENT)
      if (any(!segment_select %in% possible_segments, na.rm = TRUE)) {
        util_error(c("segment_select values are not present in the ",
                     "column 'STUDY_SEGMENT' that is assumed to be the ",
                     "segment_column when this is not assigned."))
      }
    } else {
      util_error(c("No segment_column provided and no STUDY_SEGMENT ",
                   "available in the metadata"))
    }
  }

  #check if both segment_column and segment_exclude are present
  if (!is.null(segment_exclude) && is.null(segment_column)) {
    if(STUDY_SEGMENT %in% colnames(meta_data)) {
      segment_column <- STUDY_SEGMENT
      possible_segments <- unique(meta_data$STUDY_SEGMENT)
      if(any(!segment_exclude %in% possible_segments, na.rm = TRUE)) {
        util_error(c("segment_exclude values are not present in the ",
                     "column 'STUDY_SEGMENT' that is assumed to be the ",
                     "segment_column when this is not assigned."))
      }
    } else {
      util_error(c("No segment_column provided and no STUDY_SEGMENT",
                   " available in the metadata"))
      }
  }

  # Load cross-item_level metadata and normalize it ----
  # check if there is a cross item metadata and if it is a data frame
  # in case is not present, create an empty data frame for cross item metadata
  try(util_expect_data_frame(meta_data_cross_item), silent = TRUE)
  if (!is.data.frame(meta_data_cross_item)) {
    util_message(sprintf(
      "No cross-item level metadata %s found",
      dQuote(meta_data_cross_item)))
    meta_data_cross_item <- data.frame(VARIABLE_LIST = character(0),
                                       CHECK_LABEL = character(0))
  }

  # First normalize input for meta_data_cross_item from the user
  meta_data_cross_item <- util_normalize_cross_item(
    meta_data = meta_data,
    meta_data_cross_item = meta_data_cross_item,
    label_col = label_col
  )

  # Load segment_level metadata ----
  try(util_expect_data_frame(meta_data_segment,
                             col_names = list(STUDY_SEGMENT = is.character)),
      silent = TRUE)
  # if not present, create an empty one
  if (!is.data.frame(meta_data_segment)) {
    util_message("No segment level metadata %s found",
                 dQuote(meta_data_segment))
    if(!is.null(meta_data$STUDY_SEGMENT)) {  #Add study segments if present
      meta_data_segment <- data.frame(STUDY_SEGMENT =
                                        unique(meta_data$STUDY_SEGMENT),
                                      SEGMENT_ID_VARS = NA_character_)
    } else {
      meta_data_segment <- data.frame(STUDY_SEGMENT =
                                        character(0),
                                      SEGMENT_ID_VARS = character(0))
    }
  }

  if(!SEGMENT_ID_VARS %in% colnames(meta_data_segment)){
    meta_data_segment$SEGMENT_ID_VARS <- NA_character_
  }


  # Load data frame_level metadata ----
  # check if there is a dataframe level metadata or provide message
  try(util_expect_data_frame(meta_data_dataframe), silent = TRUE)
  if (!is.data.frame(meta_data_dataframe)) {
    util_message(sprintf(
      "No dataframe level metadata %s found",
      dQuote(meta_data_dataframe)))
    meta_data_dataframe <- data.frame(DF_NAME = character(0),
                                      DF_CODE = character(0),
                                      DF_ID_VARS = character(0))
  } else if (is.data.frame(meta_data_dataframe) &&
             (!missing(study_data) && !is.data.frame(study_data) &&
              length(study_data) > 1)) {
    # two or more study data indicated as vector by the user
    util_message(c("When multiple data frames are provided for the",
                   " argument 'study data', ",
                   "the dataframe_level_metadata information will be ignored"))
    meta_data_dataframe <- data.frame(DF_NAME = character(0),
                                      DF_CODE = character(0),
                                      DF_ID_VARS = character(0))
  } else {
    # check presence of expected columns in the dataframe_level metadata
    util_expect_data_frame(meta_data_dataframe,
                           col_names = list(DF_NAME = is.character))

    if (DF_ID_VARS %in% colnames(meta_data_dataframe)) {
      util_expect_data_frame(meta_data_dataframe,
                             col_names = list(DF_ID_VARS = is.character))
    } else {
      meta_data_dataframe$DF_ID_VARS <- NA_character_
    }

    if (DF_CODE %in% colnames(meta_data_dataframe)) {
      util_expect_data_frame(meta_data_dataframe,
                             col_names = list(DF_CODE = is.character))
    } else {
      meta_data_dataframe$DF_CODE <- NA_character_
    }

    if(nrow(meta_data_dataframe) > 0) {
      # Add the path to the dataframe name if not already present,
      # if input_dir is provided
      if (!is.null(input_dir)) {
        meta_data_dataframe$DF_NAME <-
          vapply(
            meta_data_dataframe$DF_NAME,
            FUN = function(x) {
              if (!grepl(.Platform$file.sep, x, fixed = TRUE)) {
                if (endsWith(input_dir, .Platform$file.sep)) {
                  input_dir <- substr(input_dir, 1, nchar(input_dir) - 1)
                }
                x <- file.path(input_dir, x)
                return(x)
              } else {
                return(x)
              }
            },
            FUN.VALUE = character(1)
          )}
    }
  }

  # Load "meta_data_item_computation" metadata and complete VARIABLE_LIST ----
  # check if there is a computed_items metadata and if it is a data frame
  # in case is not present, create an empty data frame for computed_items
  try(util_expect_data_frame(meta_data_item_computation),
      silent = TRUE)
  if (!is.data.frame(meta_data_item_computation)) {
    util_message(sprintf(
      "No meta_data_item_computation %s found",
      dQuote(meta_data_item_computation)))
    meta_data_item_computation <- data.frame(VAR_NAMES = character(0),
                                             COMPUTATION_RULE = character(0))
  }

  if (VARIABLE_LIST %in% colnames(meta_data_cross_item)) {
    meta_data_cross_item[[VARIABLE_LIST_ORDER]] <-
      meta_data_cross_item[[VARIABLE_LIST]]
  }

  ### create VARIABLE_LIST entries from COMPUTATION_RULE entries
  needles_var_names <- unique(c(
    meta_data[[VAR_NAMES]],
    meta_data[[label_col]],
    meta_data[[LABEL]],
    meta_data[[LONG_LABEL]],
    meta_data[["ORIGINAL_VAR_NAMES"]],
    meta_data[["ORIGINAL_LABEL"]]
  ))
  needles <- paste0("[", needles_var_names, "]")
  x <- vapply(
    setNames(needles, nm = needles_var_names),
    grepl,
    setNames(nm = meta_data_item_computation[["COMPUTATION_RULE"]]),
    fixed = TRUE,
    FUN.VALUE = logical(length = nrow(meta_data_item_computation))
  )
  if (is.vector(x)) {
    x <- as.matrix(t(x))
  }
  variablelist <- unname(lapply(as.data.frame(t(x)), function(xx)
    unique(sort(colnames(
      x
    )[xx]))))
  variablelist <-
    lapply(variablelist, paste0, collapse = sprintf(" %s ", SPLIT_CHAR))

  if (!VARIABLE_LIST %in% colnames(meta_data_item_computation) &&
      nrow(meta_data_item_computation) > 0) {
    meta_data_item_computation[[VARIABLE_LIST]] <- NA_character_
  } else if (!VARIABLE_LIST %in% colnames(meta_data_item_computation) &&
             nrow(meta_data_item_computation) == 0) {
    meta_data_item_computation[[VARIABLE_LIST]] <- character(0)
  }
  vl_empty <- util_empty(meta_data_item_computation[[VARIABLE_LIST]])
  meta_data_item_computation[[VARIABLE_LIST]][vl_empty] <-
    variablelist[vl_empty]

  # define vars_in_subgroup----
  if(!is.null(subgroup)){
    needles_var_names <- unique(c(meta_data[[VAR_NAMES]],
                                  meta_data[[label_col]],
                                  meta_data[[LABEL]],
                                  meta_data[[LONG_LABEL]],
                                  meta_data[["ORIGINAL_VAR_NAMES"]],
                                  meta_data[["ORIGINAL_LABEL"]]))
    needles <- paste0("[", needles_var_names, "]")
    x <- vapply(setNames(needles, nm = needles_var_names),
                grepl,
                setNames(nm = subgroup),
                fixed = TRUE,
                FUN.VALUE = logical(length = length(subgroup)))
    if (is.vector(x)) {
      x <- as.matrix(t(x))
    }
    vars_in_subgroup <- unlist(unname(lapply(as.data.frame(t(x)), function(xx)
      unique(sort(colnames(x)[xx])))))
    rm(needles_var_names, needles, x)
  } else {
    vars_in_subgroup <- character(0)
  }

  vars_in_subgroup <- #if not in VAR_NAMES reduce it to VAR_NAMES
    vapply(vars_in_subgroup, FUN = function(x) {
      if(!x %in% meta_data$VAR_NAMES) {
        x <- util_map_labels(
          x,
          meta_data = meta_data,
          from = label_col,
          to = VAR_NAMES)
        names(x) <- NULL
        return(x)
      } else {
        x
      }
    }, FUN.VALUE = character(1))

  # Define needed objects for later use----
  # initialize split_segment to FALSE
  # it can be: FALSE if there is no splitting or TRUE otherwise
  split_segments <- FALSE


  # Prepare/filter item_level metadata when argument resp_vars is provided ----
  if(!missing(resp_vars)) {
    #if resp_vars is not VAR_NAMES but LABELS, first turn them to VAR_NAMES
    resp_vars <- vapply(resp_vars, FUN = function(x) {
      if(!x %in% meta_data$VAR_NAMES) {
        x <- util_map_labels(
          x,
          meta_data = meta_data,
          from = label_col,
          to = VAR_NAMES)
        names(x) <- NULL
        return(x)
      } else {
        x
      }
    }, FUN.VALUE = character(1))

    # 1st: extract all rules from cross-item containing a var from resp_vars
    #combine var_names with labels
    vars <- util_map_labels(resp_vars,
                            meta_data = meta_data,
                            to = label_col,
                            from = VAR_NAMES)
    # create a list with the vars from the column
    # variable_list in cross-item metadata
    rules_vars <-
      util_parse_assignments(meta_data_cross_item$VARIABLE_LIST,
                             multi_variate_text =
                               TRUE)

    # intersect vars in the rules with vars
    # in the segment to see which rules affect the
    # current segment and discard rules not affected by the resp_vars
    rules_to_use_cil <-
      vapply(lapply(rules_vars, intersect, vars),
             length,
             FUN.VALUE = integer(1)) > 0
    cil_in_resp_vars <- meta_data_cross_item[rules_to_use_cil, , FALSE]

    rm(rules_to_use_cil, rules_vars)


    # 2nd: extract all rules containing a resp_vars from item_computation_level
    # extracting the variable names from the column
    # VAR_NAMES in meta_data_item_computation
    comp_vars <- util_map_labels(
      meta_data_item_computation$VAR_NAMES,
      meta_data = meta_data,
      to = label_col,
      from = VAR_NAMES)
    # intersect vars in the rules with resp_vars
    computed_to_use <-
      vapply(lapply(comp_vars, intersect, vars),
             length,
             FUN.VALUE = integer(1)) > 0

    # discard rules not affected by resp_vars
    computed_in_resp_vars <- meta_data_item_computation[computed_to_use, ,
                                                        FALSE]

    rm(computed_to_use, comp_vars, vars)

    # Obtain all the original plus referred variables and
    # the filtered item_level metadata.
    # All the referred variables added has a variable role = 'suppress'
    overview_referred <- util_referred_vars(resp_vars = resp_vars,
                                            id_vars = id_vars,
                                            vars_in_subgroup = vars_in_subgroup,
                                            label_col = label_col,
                                            meta_data = meta_data,
                                            meta_data_segment =
                                              meta_data_segment,
                                            meta_data_dataframe =
                                              meta_data_dataframe,
                                            meta_data_cross_item =
                                              cil_in_resp_vars,
                                            meta_data_item_computation =
                                              computed_in_resp_vars,
                                            strata_column = strata_column)
    meta_data_cross_item <- cil_in_resp_vars
    meta_data_item_computation <- computed_in_resp_vars
    resp_vars_complete <- overview_referred$vars_complete
    meta_data <- overview_referred$md_complete
    attr(meta_data, "normalized") <- TRUE
    attr(meta_data, "version") <- 2

    rm(overview_referred)

    #Modify also meta_data_item_computation to remove non necessary rows
    meta_data_item_computation <-
      meta_data_item_computation[meta_data_item_computation$VAR_NAMES %in%
                                   resp_vars_complete, , drop = FALSE]
  }


  # Define the split ----
  # if nothing specified, it tries to separate data by STUDY SEGMENT
  # (unless there is a subgroup, or strata_column, or
  # specifically stated by the user "segment_column = NULL"
  # if there is "segment_column = NULL" it is nowhere in the if statement,
  # and it does not create any split
  if ((missing(segment_column) && is.null(segment_column)) &&
      is.null(strata_column) && is.null(subgroup)) {
    if (STUDY_SEGMENT %in% colnames(meta_data)) {
      segment_column <- STUDY_SEGMENT
    } else {
      util_error(
        paste0(
          "No information for split provided. ",
          "Missing both strata_column, segment_column, and subgroup.",
          " To have an unsplit report please set the ",
          "argument 'segment_column = NULL'"
        ),
        applicability_problem = TRUE
      )
    }
  } else if (is.null(segment_column) &&
             !is.null(strata_column)) {
    # if strata_column specified, check if the variable is in VAR_NAMES and
    # set a new object with the label if possible
    if (label_col_provided != VAR_NAMES && !is.null(strata_column)) {
      strata_column1 <- strata_column
      strata_column <- try(util_map_labels(strata_column,
                                           meta_data, VAR_NAMES,
                                           label_col),
                           silent = TRUE)
      # if strata_column is a VAR_NAME then there is an error instead of a
      # vector in strata_column, so replace with the original value
      if (!is.vector(strata_column)) {
        strata_column <- strata_column1
      }
      rm(strata_column1)
      if (!strata_column %in% meta_data[[VAR_NAMES]]) {
        util_error(
          paste0(
            "The strata_column provided does not correpond ",
            "to any variable in the item_level_metadata"
          ),
          applicability_problem = TRUE
        )
      }
      strata_column_label <- util_map_labels(strata_column,
                                             meta_data,
                                             label_col,
                                             VAR_NAMES)
    } else if (label_col_provided == VAR_NAMES && !is.null(strata_column)) {
      if (!strata_column %in% meta_data[VAR_NAMES]) {
        util_error(
          paste0(
            "The strata_column provided does not correpond ",
            "to any variable in the item_level_metadata"
          ),
          applicability_problem = TRUE
        )
      }
      strata_column_label <- util_map_labels(strata_column,
                                             meta_data, label_col,
                                             VAR_NAMES)
    }
    # if only segment_column is provided, check if the value corresponds to
    # a column in the item_level_metadata
  } else if (!is.null(segment_column) &&
             is.null(strata_column)) {
    if (!(segment_column %in% colnames(meta_data))) {
      util_error(
        "No metadata attribute %s found for segmenting DQ report.",
        dQuote(segment_column))
    }
    # if both segment_column and strata_column are provided
    # check if the segment_column provided corresponds to a column
    # in the item_level_metadata and
    # check if the variable of strata_column is in VAR_NAMES and
    # set a new object with the label if possible
  } else if (!is.null(segment_column) &&
             !is.null(strata_column)) {
    if (!(segment_column %in% colnames(meta_data))) {
      util_error(
        "No metadata attribute %s found for segmenting DQ report.",
        dQuote(segment_column))
    }
    if (label_col_provided != VAR_NAMES && !is.null(strata_column)) {
      strata_column1 <- strata_column
      strata_column <- try(util_map_labels(strata_column,
                                           meta_data, VAR_NAMES,
                                           label_col),
                           silent = TRUE)
      # if strata_column is a VAR_NAME then there is an error instead of a
      # vector in strata_column, so replace with the original value

      if (!is.vector(strata_column)) {
        strata_column <- strata_column1
      }
      rm(strata_column1)
      if (!strata_column %in% meta_data[, VAR_NAMES, drop = TRUE]) {
        util_error(
          paste0(
            "The strata_column provided does not correpond ",
            "to any variable in the item_level_metadata"
          ),
          applicability_problem = TRUE
        )
      }
      strata_column_label <- util_map_labels(strata_column,
                                             meta_data, label_col,
                                             VAR_NAMES)
    } else if (label_col_provided == VAR_NAMES && !is.null(strata_column)) {
      if (!strata_column %in% meta_data[, VAR_NAMES, drop = TRUE]) {
        util_error(
          paste0(
            "The strata_column provided does not correpond ",
            "to any variable in the item_level_metadata"
          ),
          applicability_problem = TRUE
        )
      }
      strata_column_label <- util_map_labels(strata_column,
                                             meta_data, label_col,
                                             VAR_NAMES)
    }
  }

  # Define the study data----
  #if the study_data argument is specified by the user
  if (!missing(study_data)) {
    ### case 1: only one study data indicated----
    #the data frame is an object in the environment
    if (is.data.frame(study_data)) {
      # if it is a data frame, look up the call of the function
      # to get the name of the study data from the call
      name_of_study_data <-
        head(as.character(substitute(study_data)), 1)
      # add the study data to the cache
      prep_add_data_frames(data_frame_list = setNames(list(study_data),
                                                      nm = name_of_study_data))
      # create also here for compatibility a list_sd_columns,
      # a list containing one element in this case,
      # containing the names of the study data variables
      dataframe_names <- name_of_study_data
      list_sd_columns <- colnames(study_data)
      list_sd_columns <- list(list_sd_columns)
      names(list_sd_columns) <- dataframe_names
    } else if (length(study_data) == 1 &&
               is.character(study_data)) {
      # if the user indicate a name of a data frame,
      # use that name (not loaded yet)

      # add the path before the name of the study data
      if (!is.null(input_dir)) {
        if (!grepl(.Platform$file.sep, study_data, fixed = TRUE)) {
          if (endsWith(input_dir, .Platform$file.sep)) {
            input_dir <- substr(input_dir, 1, nchar(input_dir) - 1)
          }
          study_data <- file.path(input_dir, study_data)
        }
      }
      name_of_study_data <- study_data

      # create also here for compatibility a list_sd_columns,
      # a list containing one element in this case,
      # containing the names of the study data variables
      dataframe_names <- name_of_study_data
      list_sd_columns <- colnames(prep_get_data_frame(dataframe_names))
      list_sd_columns <- list(list_sd_columns)
      names(list_sd_columns) <- dataframe_names
    } else if (!is.data.frame(study_data) && length(study_data) > 1) {
      ## case 2: two or more study data indicated as vector by the user----
      # if the study names are a vector of names without path and they are
      # not URL, add the input_dir before the names of each study data

      # create a list containing all names of the study data
      if (!is.null(input_dir)) {
        dataframe_names <- vapply(study_data, function(x) {
          res <- x
          if (!grepl(.Platform$file.sep, x, fixed = TRUE)) {
            if (endsWith(input_dir, .Platform$file.sep)) {
              input_dir <- substr(input_dir, 1, nchar(input_dir) - 1)
            }
            res <- file.path(input_dir, x)
          }
          return(res)
        }, FUN.VALUE = character(1))
        names(dataframe_names) <- NULL
      } else {
        dataframe_names<- study_data
      }



      # import the name of the variables in the study data
      list_sd_columns <- lapply(setNames(nm = dataframe_names), function (nm) {
        columns_df <-
          try(prep_get_data_frame(nm,
                                  column_names_only = TRUE,
                                  keep_types = TRUE),
              silent = TRUE)
        if (util_is_try_error(columns_df)) {
          columns_df <- NULL
        }
        columns_df
      })
      #import works also with URLs, in combination with n_max and nrows = 0
      list_sd_columns <- lapply(list_sd_columns, colnames)
      names(list_sd_columns) <- dataframe_names
    } else {
      util_error(c("The provided ",
                   "study_data argument of class %s is not supported"),
                 util_pretty_vector_string(class(study_data)),
                 applicability_problem = TRUE)
    }
  } else if (missing(study_data) &&
             is.data.frame(meta_data_dataframe)) {
    ### case 3: no study data name provided by the user, use meta_data_dataframe----
    # import the study_data names from meta_data_dataframe

     #Following 4 rowsNot needed, It is not possible to arrive here without a valid DF_NAME
     # check if the column DF_NAME is not empty
#    if (all(is.na(meta_data_dataframe$DF_NAME))) {
#      util_error("Column %s in dataframe_level metadata can not be empty.",
#                 sQuote("DF_NAME"))
#    }

    # vector of the names present in the dataframe_level metadata
    dataframe_names <- meta_data_dataframe[, DF_NAME, drop = TRUE]

    #if there are information about the DATAFRAMES and DF_CODE
    if ((DF_CODE) %in% colnames(meta_data_dataframe) &&
        DATAFRAMES %in% colnames(meta_data)) {
      # Create a vector with the name of the study_data as before with
      # the actual path but also with the relative DF_CODE
      study_data_withcode <- meta_data_dataframe[, c(DF_NAME, DF_CODE),
                                                 drop = FALSE]
      list_sd_columns <- NULL

    } else {
      #if there are NO information about the DATAFRAMES and DF_CODE
      # import only headers of study data files
      list_sd_columns <- lapply(setNames(nm = dataframe_names), function (nm) {
        columns_df <-
          try(prep_get_data_frame(nm,
                                  column_names_only = TRUE,
                                  keep_types = TRUE),
              silent = TRUE)
        if (util_is_try_error(columns_df)) {
          columns_df <- NULL
        }
        columns_df
      })
      #import works also with URLs in combination with n_max = 0 and nrows = 0
      list_sd_columns <- lapply(list_sd_columns, colnames)
      names(list_sd_columns) <- dataframe_names
    }
    # stop if no study data provided and no dataframe_level metadata available
    # or dataframe_level metadata is empty
  } else if (missing(study_data) &&
             (!is.data.frame(meta_data_dataframe) ||
              nrow(meta_data_dataframe) == 0 )) {
    util_error(
      c("Not possible to create reports as no study data and no",
        "dataframe level metadata %s with study",
        "data names are available"),
      dQuote(meta_data_dataframe))
  }

  # Define name_of_study_data to NULL if it does not exist yet----
  if (!exists("name_of_study_data")) {
    name_of_study_data <- NULL
  }



  # Prepare the metadata depending on the presence of a segment_column or not----
  ### 1st case: the segment_column is present----
  if (!is.null(segment_column)) {
    # if there are empty entries in the column defined for the split, set
    # a new non-used segment name (e.g., na1) and use it for empty rows
    split_segments <- TRUE
    .md <- meta_data[[segment_column]]
    i <- ""
    while (any(.md == paste0("na", i), na.rm = TRUE)) {
      if (i == "") {
        i <- 0
      }
      i <- i + 1
    }
    .md[util_empty(.md)] <- paste0("na", i)
    meta_data[[segment_column]] <- .md
    # define an object containing the segments to create
    segments <- unique(meta_data[[segment_column]])

    # backwards compatibility-replace var_names with label in segments to create
    if (label_col_provided != VAR_NAMES &&
        all(segments %in% meta_data[[VAR_NAMES]])) {
      segmentNames <- util_map_labels(segments, meta_data, label_col)
    } else {
      segmentNames <- segments
    }
    rm(segments)

    # Add selection of 1 or multiple levels of segment_column or select using a
    # regex
    if (!is.null(segment_select)) {
      #Fix segment_select in case is written as "A | B" instead of c("A", "B")
      if (length(segment_select) == 1) {
        segment_select <-  unname(unlist(util_parse_assignments(segment_select,
                                                                split_char = SPLIT_CHAR)))
      }
      # check if the typed levels exist in the segment_column possible levels
      # before to select it
      if (!any(segment_select %in% segmentNames)) {
        #if there is more than one element it means it is not a pattern, but
        # a list of names
        if (length(segment_select) > 1) {
          # stop if selection does not match any level
          util_error(
            "No segment_column level matches the provided names: %s",
            dQuote(segment_select) )
        } else {
          # if the argument is not present in the level names, check if it is
          # a pattern, and if it also does not match any level then provide a
          # warning and stop
          all_segmentNames <- segmentNames
          segmentNames <- segmentNames[grepl(segment_select, segmentNames)]
          if (length(segmentNames) == 0) {
            # stop if selection (name or pattern) does
            # not match any level
            util_error(
              c("No segment_column level matches the provided name or",
                "pattern: %s"),
              dQuote(segment_select))
            segmentNames <- all_segmentNames
          }
          rm(all_segmentNames)
        }
      } else {
        segmentNames <- segmentNames[segmentNames %in% segment_select]
      }
    }

    ### remove segments to exclude
    if (!is.null(segment_exclude)) {
      #Fix segment_exclude in case is written as "A | B" instead of c("A", "B")
      if (length(segment_exclude) == 1) {
        segment_exclude <-
          unname(unlist(util_parse_assignments(segment_exclude,
                                               split_char = SPLIT_CHAR)))
      }

      # check if the typed levels exist in the segment_column possible levels
      # before to select it
      if (!any(segment_exclude %in% segmentNames)) {
        #if there is more than one element it means it is not a pattern, but
        # a list of names
        if (length(segment_exclude) > 1) {
          # stop if selection does not match any level
          util_error(
            "No segment_column level matches the provided names to exclude: %s",
            dQuote(segment_exclude))
        } else {
          # if the argument is not present in the level names, check if it is
          # a pattern, and if it also does not match any level then provide a
          # warning and stop
          all_segmentNames <- segmentNames
          unwanted_seg <- segmentNames[grepl(segment_exclude, segmentNames)]
          segmentNames <- setdiff(segmentNames, unwanted_seg)
          if (length(segmentNames) == 0) {
            # stop if selection (name or pattern) has no segment left
            util_error(
              c("No segment_column level left after removing ",
                "unwanted segments: %s"),
              dQuote(segment_exclude))
          }
          rm(all_segmentNames)
        }
      } else {
        #exclude unwanted segments
        segmentNames <- segmentNames[!(segmentNames %in% segment_exclude)]
        if (length(segmentNames) == 0) {
          # stop if selection has no segment left
          util_error(
            c("No segment_column level left after removing ",
              "unwanted segments: %s"),
            dQuote(segment_exclude))
        }
      }
    }


    # Filter cross-item level for the rules that can contain variables
    # in the segment (cross-item_level already normalized)
    # extract all rules containing a variable from the current evaluated segment
    cil_in_segment <- lapply(setNames(segmentNames, nm = segmentNames),
                             function(segment) {
                               vars <-
                                 meta_data[meta_data[[segment_column]] ==
                                             segment, VAR_NAMES]
                               #replace var_names with labels
                               vars <- util_map_labels(vars,
                                                       meta_data = meta_data,
                                                       to = label_col,
                                                       from = VAR_NAMES)
                               # extracting the variable names from the column
                               # variable_list in cross-item metadata to have
                               # a vector of variable names for each rule in the
                               # list rules_vars
                               rules_vars <-
                                 util_parse_assignments(
                                   meta_data_cross_item$VARIABLE_LIST,
                                   multi_variate_text = TRUE)
                               # intersect vars in the rules with vars
                               # in the segment to discard rules not
                               # affecting the current segment
                               rules_to_use <-
                                 vapply(lapply(rules_vars, intersect, vars),
                                        length,
                                        FUN.VALUE = integer(1)) > 0
                               meta_data_cross_item[rules_to_use, , FALSE]
                             })

    # Filter computed items to extract all rules to compute
    # containing a VAR_NAMES from the current evaluated segment
    computed_in_segment <-
      lapply(setNames(segmentNames, nm = segmentNames),
             function(segment) {
               vars <-
                 meta_data[meta_data[[segment_column]] ==
                             segment, VAR_NAMES]
               #replace var_names with labels
               vars <- util_map_labels(vars,
                                       meta_data = meta_data,
                                       to = label_col,
                                       from = VAR_NAMES)
               # extracting the variable names from the column
               # VAR_NAMES in meta_data_item_computation
               comp_vars <- util_map_labels(
                 meta_data_item_computation$VAR_NAMES,
                 meta_data = meta_data,
                 to = label_col,
                 from = VAR_NAMES
               )
               # intersect vars in the rules with vars
               # in the segment to see which VAR_NAMES affect the
               # current segment
               computed_to_use <-
                 vapply(lapply(comp_vars, intersect, vars),
                        length,
                        FUN.VALUE = integer(1)) > 0
               # discard rules not affected by the
               # current segment
               meta_data_item_computation[computed_to_use, , FALSE]
             })

    # Select only the data frames interested by the current segment
    ### if there is a dataframe level metadata----
    if (is.data.frame(meta_data_dataframe) &&
        nrow(meta_data_dataframe) > 0) { #this is introduced in the case of
      # study_data argument filled and so the dataframe level will be ignored

      # check if the column DF_NAME is not empty
      if (all(is.na(meta_data_dataframe$DF_NAME))) {
        util_error("Column %s in dataframe_level metadata can not be empty.",
                   sQuote("DF_NAME"))
      }

      # extract the data frame per segment
      #### caseA: there are DF_CODE and DATAFRAMES: select the dataframes ----
      # based on the DF_CODE of the variables in the segment
      if ((DF_CODE) %in% colnames(meta_data_dataframe) &&
          DATAFRAMES %in% colnames(meta_data)) {
        dfr_in_segment <-
          lapply(setNames(segmentNames, nm = segmentNames), function(segment) {
            vars <-
              meta_data[meta_data[[segment_column]] ==
                          segment, c(VAR_NAMES, DATAFRAMES), drop = FALSE]

            dfr_CODE_from_item_level <- unique(unname(unlist(
              util_parse_assignments(
                vars$DATAFRAMES,
                split_char = SPLIT_CHAR,
                multi_variate_text = TRUE
              )
            )))

            dfr_to_use <-
              meta_data_dataframe[meta_data_dataframe[[DF_CODE]] %in%
                                    dfr_CODE_from_item_level, , FALSE]
          })
      } else {
        #### caseB: there are no DF_CODE and DATAFRAMES ----
        dfr_in_segment <-
          lapply(setNames(segmentNames, nm = segmentNames), function(segment) {
            vars <-
              meta_data[meta_data[[segment_column]] ==
                          segment, VAR_NAMES]
            dfr_to_use <-
              vapply(lapply(list_sd_columns, intersect, vars),
                     length,
                     FUN.VALUE = integer(1)) > 0
            # convert to data frame with row.names = file names
            # and column with TRUE or FALSE
            dfr_to_use <- do.call(rbind, lapply(dfr_to_use, as.data.frame))
            colnames(dfr_to_use) <- "intersection"
            dfr_to_use <-
              dfr_to_use[dfr_to_use$intersection == TRUE, , drop = FALSE]
            dfr_to_use <- row.names(dfr_to_use)
            #discard data frames not containing the vars of this segment)
            dfr_to_use <-
              dataframe_names[dataframe_names %in% dfr_to_use]

            meta_data_dataframe[meta_data_dataframe[[DF_NAME]] %in%
                                  dfr_to_use, , FALSE]
          })
      }

    } else if (!is.data.frame(meta_data_dataframe) ||
               (is.data.frame(meta_data_dataframe) &&
                nrow(meta_data_dataframe) == 0)) {
      ### in case there is no dataframe metadata or it is empty----
      if (length(study_data) == 1 || is.data.frame(study_data)) {
        meta_data_dataframe <- data.frame(DF_NAME =
                                            name_of_study_data)
        dfr_in_segment <-
          lapply(setNames(segmentNames, nm = segmentNames), function(segment) {
            data.frame(DF_NAME = name_of_study_data,
                       DF_CODE = NA_character_,
                       DF_ID_VARS = NA_character_)
          })
      } else if (!is.data.frame(study_data) &&
                 length(study_data) > 1) {
        meta_data_dataframe <- data.frame(DF_NAME = dataframe_names,
                                          DF_CODE = NA_character_,
                                          DF_ID_VARS = NA_character_)

        dfr_in_segment <-
          lapply(setNames(segmentNames, nm = segmentNames), function(segment) {
            vars <-
              meta_data[meta_data[[segment_column]] ==
                          segment, VAR_NAMES]
            dfr_to_use <-
              vapply(lapply(list_sd_columns, intersect, vars),
                     length,
                     FUN.VALUE = integer(1)) > 0
            #convert to data frame with row.names = file names and
            # column with TRUE or FALSE
            dfr_to_use <- do.call(rbind, lapply(dfr_to_use, as.data.frame))
            colnames(dfr_to_use) <- "intersection"
            dfr_to_use <-
              dfr_to_use[dfr_to_use$intersection == TRUE, , drop = FALSE]
            dfr_to_use <- row.names(dfr_to_use)
            #discard data frames not containing the vars of this segment
            dfr_to_use <-
              dataframe_names[dataframe_names %in% dfr_to_use]
            meta_data_dataframe <-
              meta_data_dataframe[meta_data_dataframe$DF_NAME %in% dfr_to_use,
                                  , drop = FALSE]
          })

      }
    }

    # Prepare segment level meta_data, selecting the row matching
    # the current segment
    # if the segment_column is study_segment, separate it by segment
    if (segment_column == STUDY_SEGMENT) {
      seg_in_segment <-
        lapply(setNames(segmentNames, nm = segmentNames), function(segment) {
          meta_data_segment[meta_data_segment[[STUDY_SEGMENT]] ==
                              segment, , FALSE]
        })
    } else {
      # in case the segment_column is not the segment, need to first
      # create a list with the corresponding segments for each segment_column
      if (!STUDY_SEGMENT %in% colnames(meta_data)) {
        #In case there is no STUDY_SEGMENT
        meta_data$STUDY_SEGMENT <- "all"
      }
      segments_list <- meta_data[, c(segment_column, STUDY_SEGMENT), #TODO: need back compatibility with KEY_STUDY_SEGMENT using prep_meta_data_v1_to_item_level_meta_data?
                                 drop = FALSE]
      segments_list <- unique(segments_list)
      segments_list <-
        lapply(setNames(segmentNames, nm = segmentNames), function(piece) {
          segments_list[segments_list[segment_column] ==
                          piece, STUDY_SEGMENT, TRUE]
        })
      seg_in_segment <-
        lapply(setNames(segmentNames, nm = segmentNames), function(x) {
          meta_data_segment[meta_data_segment[[STUDY_SEGMENT]] %in%
                              c(segments_list[[x]]), , FALSE]
        })
      # in case of segment_select not present in the data, there will be
      # elements $<NA> in the list. Remove them
      seg_in_segment <- seg_in_segment[!is.na(names(seg_in_segment))]
    }


    #For each segment define which variables to include (using a repeat-loop)
    if(!missing(resp_vars)) {
      # if there is the argument resp_vars
      vars_in_segment <-
        lapply(setNames(segmentNames, nm = segmentNames),
               function(segment) {

                 # First select the variables that are indicated
                 # as part of the segment in the item-level metadata
                 # (in this case the metadata was filtered and contains
                 # original and referred variables)
                 vars <-
                   meta_data[meta_data[[segment_column]] ==
                               segment, VAR_NAMES]
                 #include variables that are in resp_vars and in this segment
                 vars<- vars[vars %in% resp_vars]
                 overview_vars_md <-
                   util_referred_vars(resp_vars = vars,
                                      id_vars = id_vars,
                                      vars_in_subgroup = vars_in_subgroup,
                                      label_col = label_col,
                                      meta_data = meta_data,
                                      meta_data_segment =
                                        seg_in_segment[[segment]],
                                      meta_data_dataframe =
                                        dfr_in_segment[[segment]],
                                      meta_data_cross_item =
                                        cil_in_segment[[segment]],
                                      meta_data_item_computation =
                                        computed_in_segment[[segment]],
                                      strata_column = strata_column)

                 vars <- overview_vars_md$vars_complete
               })

      # Subset the item_level metadata per each segment
      md_in_segment <-
        lapply(setNames(segmentNames, nm = segmentNames),
               function(segment) {
                 # First select the variables that are indicated
                 # as part of the segment in the item-level metadata
                 # (in this case the metadata was filtered and contains
                 # original and referred variables
                 vars <-
                   meta_data[meta_data[[segment_column]] ==
                               segment, VAR_NAMES]
                 #include variables that are in resp_vars and in this segment
                 vars<- vars[vars %in% resp_vars]
                 overview_vars_md <-
                   util_referred_vars(resp_vars = vars,
                                      id_vars = id_vars,
                                      vars_in_subgroup = vars_in_subgroup,
                                      label_col = label_col,
                                      meta_data = meta_data,
                                      meta_data_segment =
                                        seg_in_segment[[segment]],
                                      meta_data_dataframe =
                                        dfr_in_segment[[segment]],
                                      meta_data_cross_item =
                                        cil_in_segment[[segment]],
                                      meta_data_item_computation =
                                        computed_in_segment[[segment]],
                                      strata_column = strata_column)

                 md_seg <- overview_vars_md$md_complete
                 attr(md_seg, "normalized") <- TRUE
                 attr(md_seg, "version") <- 2
                 return(md_seg)
               })





      # create resp_vars_in_segment
      resp_vars_in_segment <- lapply(setNames(segmentNames, nm = segmentNames),
                                     function(segment) {
                                       vars_in_segment <-
                                         vars_in_segment[[segment]][
                                           vars_in_segment[[segment]] %in%
                                                                      resp_vars]
                                     })
    } else {
      # if there is no argument resp_vars
      vars_in_segment <-
        lapply(setNames(segmentNames, nm = segmentNames),
               function(segment) {
                 # First select the variables that are indicated
                 # as part of the segment in the item-level metadata
                 vars <-
                   meta_data[meta_data[[segment_column]] ==
                               segment, VAR_NAMES]
                 overview_vars_md <-
                   util_referred_vars(resp_vars = vars,
                                      id_vars = id_vars,
                                      vars_in_subgroup = vars_in_subgroup,
                                      label_col = label_col,
                                      meta_data = meta_data,
                                      meta_data_segment =
                                        seg_in_segment[[segment]],
                                      meta_data_dataframe =
                                        dfr_in_segment[[segment]],
                                      meta_data_cross_item =
                                        cil_in_segment[[segment]],
                                      meta_data_item_computation =
                                        computed_in_segment[[segment]],
                                      strata_column = strata_column)
                 vars <- overview_vars_md$vars_complete
               })

      # Define the item_level metadata for each segment
      md_in_segment <-
        lapply(setNames(segmentNames, nm = segmentNames),
               function(segment) {
                 # First select the variables that are indicated
                 # as part of the segment in the item-level metadata
                 vars <-
                   meta_data[meta_data[[segment_column]] ==
                               segment, VAR_NAMES]
                 overview_vars_md <-
                   util_referred_vars(resp_vars = vars,
                                      id_vars = id_vars,
                                      vars_in_subgroup = vars_in_subgroup,
                                      label_col = label_col,
                                      meta_data = meta_data,
                                      meta_data_segment =
                                        seg_in_segment[[segment]],
                                      meta_data_dataframe =
                                        dfr_in_segment[[segment]],
                                      meta_data_cross_item =
                                        cil_in_segment[[segment]],
                                      meta_data_item_computation =
                                        computed_in_segment[[segment]],
                                      strata_column = strata_column)

                 md_seg <- overview_vars_md$md_complete
                 attr(md_seg, "normalized") <- TRUE
                 attr(md_seg, "version") <- 2
                 return(md_seg)
               })


      #create resp_vars_in_segment = character(0)
      resp_vars_in_segment <-
        lapply(setNames(segmentNames, nm = segmentNames),
               function(segment) {
                 x <- character(0)
               })
    }



  } else {
    #### 2nd case: the segment_column is not present (NULL),----
    # include all variables and all metadata in category all_variables
    # create a list anyways but only with 1 element for all cross-item metadata
    cil_in_segment <- list(all_variables = meta_data_cross_item)

    # create a list anyways but only with 1 element for all computed vars md
    computed_in_segment <- list(all_variables = meta_data_item_computation)

    # Prepare metadata at the segment level
    try(util_expect_data_frame(meta_data_segment,
                               col_names = list(STUDY_SEGMENT = is.character)),
        silent = TRUE)
    # if not present, create an empty one
    if (!is.data.frame(meta_data_segment)) {
      util_message("No segment level metadata %s found",
                   dQuote(meta_data_segment))
      meta_data_segment <- data.frame(STUDY_SEGMENT =
                                        unique(meta_data$STUDY_SEGMENT))
    }
    seg_in_segment <- list(all_variables = meta_data_segment)

    # prepare the dataframe_level metadata, put all in a list element
    # names "all_variables" if present, if not create an empty one
    if (is.data.frame(meta_data_dataframe)) {
      dfr_in_segment <- list(all_variables = meta_data_dataframe)
    } else {
      meta_data_dataframe <- data.frame(DF_NAME = dataframe_names,
                                        DF_CODE = NA_character_,
                                        DF_ID_VARS = NA_character_)
      dfr_in_segment <- list(all_variables = meta_data_dataframe)
    }

    # prepare the list of variables in a list containing all the variables
    # in the item_level metadata
    vars_in_segment <- list(all_variables = meta_data[[VAR_NAMES]])

    # create resp_vars_in_segment <- character(0)
    if(!missing(resp_vars)) {
      resp_vars_in_segment <- list(all_variables = resp_vars)
    } else {
      resp_vars_in_segment <- list(all_variables = character(0))
    }






    #Add all the meta_data (item_level)
    md_in_segment <- list(all_variables = meta_data)

    #Addition for progressing bar
    segmentNames <- "all_variables"
  }


  # if no study data was specified by the user, set it to NULL -----
  # this fixes an issue in the following overall loop when it is missing
  if (missing(study_data)) {
    study_data <- NULL
  }




  # Calculate no. strata for the job progress bar----
  if (is.null(strata_column)) {
    n_strata <- 1
  } else {
    # Value labels from VALUE_LABEL_TABLE and CODE_LIST_TABLE
    if (is.na(meta_data[meta_data[VAR_NAMES] == strata_column,
                        VALUE_LABELS, drop = TRUE]) ||
        is.null(meta_data[meta_data[VAR_NAMES] == strata_column,
                          VALUE_LABELS, drop = TRUE])) {
      value_label_table_name <-
        meta_data[meta_data[VAR_NAMES] == strata_column,
                  VALUE_LABEL_TABLE, drop = TRUE]

      try(util_expect_data_frame(value_label_table_name),
          silent = TRUE)
      if (!is.data.frame(value_label_table_name)) {
        try(util_expect_data_frame("CODE_LIST_TABLE"), silent = TRUE)
        if (!is.data.frame(CODE_LIST_TABLE)) {
          util_message(sprintf(
            "No value_label_table_name %s found",
            dQuote(value_label_table_name)))
          value_label_table_name <- data.frame(CODE_VALUE = character(0),
                                               CODE_LABEL = character(0))
        } else {
          #select in CODE_LIST_TABLE$VALUE_LABEL_TABLE only the
          # value_label_table_name
          value_label_table_name <-
            CODE_LIST_TABLE[CODE_LIST_TABLE$VALUE_LABEL_TABLE ==
                              value_label_table_name, , drop = FALSE]
        }
      }
      expected_strata <- setNames(value_label_table_name$CODE_LABEL,
                                  nm = value_label_table_name$CODE_VALUE)
    } else {
      # if value labels are in a column in item_level metadata
      expected_strata <- unlist(
        util_parse_assignments(
          meta_data[meta_data[VAR_NAMES] == strata_column, VALUE_LABELS,
                    drop = TRUE],
          split_char = SPLIT_CHAR,
          split_on_any_split_char = TRUE,
          multi_variate_text = TRUE
        )
      )
    }



    # in case there is only a data frame as study_data
    if (length(list_sd_columns) == 1) {
      # if the dataframe already exists in the current environment assign it
      # otherwise import it
      tab1 <- util_expect_data_frame(names(list_sd_columns), dont_assign = TRUE)
      original_strata <- unique(tab1[[strata_column]])
      # check if all expected strata are present in the study data
      if (length(setdiff(names(expected_strata), original_strata)) > 0 ) {
        util_warning(c("The stratum/strata %s is/are ",
                       "not present in the study data"),
                     dQuote(names(expected_strata)[!(names(expected_strata) %in%
                                                       original_strata)]))
      }
      # check if in the study data there are strata not expected
      #removed NA from check
      if (length(setdiff(original_strata[!is.na(original_strata)],
                         names(expected_strata)) > 0 )) {
        util_warning(
          "The stratum/strata %s is/are not present in the metadata",
          dQuote(original_strata[!(original_strata %in%
                                     names(expected_strata))]))
      }
      n_strata <- length(original_strata)
      # in case the there are more dataframes as study_data
    } else if (length(list_sd_columns) > 1) {
      # obtain the names of the needed study data and select
      # only dataframes containing variables of this segment
      tab_to_import <-
        vapply(lapply(list_sd_columns, intersect, strata_column),
               length,
               FUN.VALUE = integer(1)) > 0
      tab_to_import <- names(tab_to_import[tab_to_import %in% TRUE])

      if (length(tab_to_import) == 1) {
        tab1 <- util_expect_data_frame(tab_to_import, dont_assign = TRUE)
        original_strata <- unique(tab1[[strata_column]])
      } else if (length(tab_to_import) > 1) {
        tabs <- lapply(setNames(nm = tab_to_import),
                       util_expect_data_frame,
                       dont_assign = TRUE)
        tab1 <- Reduce(function(x, y) {
          merge(x, y, all = TRUE)
        }, tabs)
        original_strata <- unique(tab1[[strata_column]])
      }
      n_strata <- length(original_strata)
    } else if (is.null(list_sd_columns)) {
      # In case in which the DF_CODE and DATAFRAME are present
      # obtain the names of the needed study data and select
      # only dataframes containing variables of strata_column
      dtf_CODE_of_strata_col <- meta_data[meta_data[[VAR_NAMES]] %in%
                                            strata_column,
                                          c(VAR_NAMES, DATAFRAMES),
                                          drop = FALSE]


      dfr_CODE_of_strata_col <- unique(unname(unlist(
        util_parse_assignments(
          dtf_CODE_of_strata_col$DATAFRAMES,
          split_char = SPLIT_CHAR,
          multi_variate_text = TRUE
        )
      )))

      tab_to_import <-
        study_data_withcode[study_data_withcode[[DF_CODE]] %in%
                              dfr_CODE_of_strata_col, DF_NAME, drop = TRUE]


      if (length(tab_to_import) == 1) {
        tab1 <- util_expect_data_frame(tab_to_import, dont_assign = TRUE)
        original_strata <- unique(tab1[[strata_column]])
      } else if (length(tab_to_import) > 1) {
        tabs <- lapply(setNames(nm = tab_to_import),
                       util_expect_data_frame,
                       dont_assign = TRUE)
        tab1 <- Reduce(function(x, y) {
          merge(x, y, all = TRUE)
        }, tabs)
        original_strata <- unique(tab1[[strata_column]])
      }
      n_strata <- length(original_strata)
    }

    # Define the strata in case strata_select is used
    if (!is.null(strata_select)) {
      #Fix strata_select in case is written as "A | B" instead of c("A", "B")
      if (length(strata_select) == 1) {
        strata_select <-
          unname(unlist(util_parse_assignments(strata_select,
                                               split_char = SPLIT_CHAR)))
      }


      # if there is a strata_select argument
      if (!is.null(selection_type)) {
        if (selection_type == "value") {
          util_stop_if_not(any(strata_select %in% original_strata),
                           label =
                             paste0("No values in the variable correspond ",
                                    "to the strata_select"))
          original_strata <-
            original_strata[original_strata %in% strata_select]
          n_strata <- length(original_strata)

        } else if (selection_type == "v_label") {
          util_stop_if_not(any(strata_select %in% expected_strata), label =
                             paste0("No value label in the variable ",
                                    "correspond to the strata_select"))
          expected_strata <-
            expected_strata[expected_strata %in% strata_select]
          original_strata <-
            original_strata[original_strata %in% names(expected_strata)]
          n_strata <- length(original_strata)
        } else if (selection_type == "regex") {
          name_pattern_strata <-
            original_strata[grepl(strata_select, original_strata)]
          name_pattern_strata2 <- expected_strata[grepl(strata_select,
                                                        expected_strata)]

          name_pattern_strata <- unique(c(name_pattern_strata,
                                          names(name_pattern_strata2)))

          util_stop_if_not(length(name_pattern_strata) > 0, label =
                             paste0("No value or value label in the variable",
                                    " correspond to the strata_select"))
          n_strata <- length(name_pattern_strata)
        }
      } else {
        #if selection_type is null, try to guess the typed strata
        if (!any(strata_select %in% original_strata)) {
          #if the strata does not match the list from data
          if (!any(strata_select %in% expected_strata)) {
            #if the strata does not match the value list labels
            # if the argument is not present in the level names, check if it is
            # a pattern
            name_pattern_strata <-
              original_strata[grepl(strata_select, original_strata)]
            name_pattern_strata <- c(name_pattern_strata,
                                     expected_strata[grepl(strata_select,
                                                           expected_strata)])
            if (length(name_pattern_strata) == 0) {
              util_error("%s does not corresponds to any strata",
                         dQuote(strata_select))
            } else {
              n_strata <- length(name_pattern_strata)
            }
          } else {
            # if the strata matches a value label
            expected_strata <-
              expected_strata[expected_strata %in% strata_select]
            original_strata <-
              original_strata[original_strata %in% names(expected_strata)]
            n_strata <- length(original_strata)

          }
        } else {
          # if the strata matches a value of the variable in the data
          original_strata <-
            original_strata[original_strata %in% strata_select]
          n_strata <- length(original_strata)
        }
      }
    }

    ## Define the strata in case strata_exclude is used
    if (!is.null(strata_exclude)) {
      if (length(strata_exclude) == 1) {
        strata_exclude <-
          unname(unlist(util_parse_assignments(strata_exclude,
                                               split_char = SPLIT_CHAR)))
      }
      # if there is a strata_exclude argument
      if (!is.null(selection_type)) {
        if (selection_type == "value") {
          util_stop_if_not(any(strata_exclude %in% original_strata), label =
                             paste0("No values in the variable correspond",
                                    " to the strata_exclude"))
          original_strata <-
            original_strata[!(original_strata %in% strata_exclude)]
          n_strata <- length(original_strata)

        } else if (selection_type == "v_label") {
          util_stop_if_not(any(strata_exclude %in% expected_strata), label =
                             paste0("No value label in the variable ",
                                    "correspond to the strata_exclude"))
          expected_strata <-
            expected_strata[!(expected_strata %in% strata_exclude)]
          original_strata <-
            original_strata[original_strata %in% names(expected_strata)]
          n_strata <- length(original_strata)
        } else if (selection_type == "regex") {
          unwanted_seg_orig <-
            original_strata[grepl(strata_exclude, original_strata)]
          unwanted_seg_label <- expected_strata[grepl(strata_exclude,
                                                      expected_strata)]
          unwanted_seg_label <- original_strata[original_strata %in%
                                                  names(unwanted_seg_label)]
          unwanted_seg <- unique(c(unwanted_seg_orig, unwanted_seg_label))
          rm(unwanted_seg_orig, unwanted_seg_label)

          name_pattern_strata	<- setdiff(original_strata, unwanted_seg)
          util_stop_if_not(length(unwanted_seg) > 0, label =
                             paste0("No value or value label in the variable",
                                    " correspond to the strata_exclude"))
          n_strata <- length(name_pattern_strata)
        }

      } else {
        #if selection_type is null, try to guess the typed strata
        if (!any(strata_exclude %in% original_strata)) {
          #if the strata does not match the list from data
          if (!any(strata_exclude %in% expected_strata)) {
            #if the strata does not match the value list labels
            # if the argument is not present in the level names, check if it is
            # a pattern
            unwanted_seg_orig <-
              original_strata[grepl(strata_exclude, original_strata)]
            unwanted_seg_label <- expected_strata[grepl(strata_exclude,
                                                        expected_strata)]
            unwanted_seg_label <- original_strata[original_strata %in%
                                                    names(unwanted_seg_label)]
            unwanted_seg <- unique(c(unwanted_seg_orig, unwanted_seg_label))
            rm(unwanted_seg_orig, unwanted_seg_label)

            name_pattern_strata	<- setdiff(original_strata, unwanted_seg)
            if (length(unwanted_seg) == 0) {
              util_error("%s does not corresponds to any strata",
                         dQuote(strata_exclude))
            } else {
              n_strata <- length(name_pattern_strata)
            }
          } else {
            # if the strata matches a value label
            expected_strata <-
              expected_strata[!(expected_strata %in% strata_exclude)]
            original_strata <-
              original_strata[original_strata %in% names(expected_strata)]
            n_strata <- length(original_strata)

          }
        } else {
          # if the strata matches a value of the variable in the data
          original_strata <-
            original_strata[!(original_strata %in% strata_exclude)]
          n_strata <- length(original_strata)
        }
      }
    }
  }

  # Create a job to report progress and give it a name ----
  util_setup_rstudio_job("dq_report_by",
                         n = length(segmentNames) * n_strata)
  # create an extra environment for the progress
  p <- new.env(parent = emptyenv())
  p$i <- 0
  p$N <- length(segmentNames) * n_strata

  # OUTER list: split base on segment----
  # Return a list of lists of results. The outer list is for the split based on
  # segment_column e.g., study segments. The inner list is for the strata
  # (split based on the strata_column)
  overall_res <- mapply(
    #here starts the outer loop by segment_column (by segment or another column)
    vars_in_segment = vars_in_segment[order(names(vars_in_segment))],
    cur_seg = sort(names(vars_in_segment)),
    meta_data = md_in_segment[order(names(md_in_segment))],
    meta_data_dataframe = dfr_in_segment[order(names(dfr_in_segment))] ,
    seg_in_segment = seg_in_segment[order(names(seg_in_segment))] ,
    MoreArgs = list(
      list_sd_columns = list_sd_columns,
      name_of_study_data = name_of_study_data,
      call_report_by = call_report_by,
      subgroup = subgroup,
      resp_vars_in_segment = resp_vars_in_segment,
      id_vars = id_vars,
      vars_in_subgroup = vars_in_subgroup
    ),
    SIMPLIFY = FALSE,
    FUN = function(vars_in_segment,
                   cur_seg,
                   meta_data,
                   meta_data_dataframe,
                   list_sd_columns,
                   seg_in_segment,
                   name_of_study_data,
                   call_report_by,
                   subgroup,
                   resp_vars_in_segment,
                   id_vars,
                   vars_in_subgroup) {

      attr(meta_data, "normalized") <- TRUE
      attr(meta_data, "version") <- 2

      # filter resp_vars_in_segment for the right segment if present
      # in case there is only a data frame as study_data
      if (length(list_sd_columns) == 1) {
        dataframe_names_cur_seg <- names(list_sd_columns)
        # if the study_data = object
        if (is.data.frame(study_data)) {
          dataframes_cur_seg <- study_data
          name_files <-
            head(as.character(substitute(study_data)), 1)
        } else {
          # import the study data files matching the name
          dataframes_cur_seg <-
            util_expect_data_frame(names(list_sd_columns), dont_assign = TRUE)
          name_files <- paste(names(list_sd_columns))
        }

        # in case there is only one study data in the data frame level
        if (is.null(name_of_study_data)) {
          name_of_study_data <- name_files
        }
        # in case the there are more data frames as study_data
      } else if (length(list_sd_columns) > 1 ||
                 is.null(list_sd_columns)) {


        #if there is CODE_DF the list_sd_columns is set to NULL
        if (is.null(list_sd_columns)) {
          # obtain the names of the needed study data and select
          # only data frames containing variables of this segment
          vars_in_seg <-
            meta_data[meta_data[[VAR_NAMES]] %in% vars_in_segment,
                      c(VAR_NAMES, DATAFRAMES), drop = FALSE]

          dfr_CODE_of_vars_in_seg <- unique(unname(unlist(
            util_parse_assignments(
              vars_in_seg$DATAFRAMES,
              split_char = SPLIT_CHAR,
              multi_variate_text = TRUE
            )
          )))

          dataframe_names_cur_seg <-
            study_data_withcode[study_data_withcode[[DF_CODE]] %in%
                                  dfr_CODE_of_vars_in_seg, DF_NAME,
                                drop = TRUE]

          rm(vars_in_seg, dfr_CODE_of_vars_in_seg)
        } else {
          # obtain the names of the needed study data and select
          # only dataframes containing variables of this segment
          dataframe_names_cur_seg <-
            vapply(lapply(list_sd_columns, intersect, vars_in_segment),
                   length,
                   FUN.VALUE = integer(1)) > 0
          dataframe_names_cur_seg <-
            dataframe_names_cur_seg[dataframe_names_cur_seg %in% TRUE]
          dataframe_names_cur_seg <-
            names(dataframe_names_cur_seg)
        }


        # import the study data files matching the names
        # in the dataframe_level metadata
        dataframes_cur_seg <-
          lapply(setNames(nm = dataframe_names_cur_seg),
                 util_expect_data_frame,
                 dont_assign = TRUE)



        # create a name for the study data merging all the
        # data frames names, after ordering and making them unique
        dataframe_names_cur_seg <- sort(unique(dataframe_names_cur_seg))
        name_of_study_data <- paste(dataframe_names_cur_seg, collapse = ", ")
      }


      #THIS IS A REPETITION BUT IT IS NEEDED BECAUSE there may be other
      # referred variables
      # Look to the data frame interested by the current segment
      dfr_in_segment <-
        lapply(setNames(dataframe_names_cur_seg, nm = dataframe_names_cur_seg),
               function(xx) {
                 meta_data_dataframe[meta_data_dataframe[[DF_NAME]] == xx, ,
                                     FALSE]
               })
      dfr_in_segment <-
        list(dfr = meta_data_dataframe[meta_data_dataframe[[DF_NAME]] %in%
                                         dataframe_names_cur_seg, , FALSE])
      names(dfr_in_segment) <- cur_seg

      # Create sd_merge and filter for var in segment if needed----
      #if there is no need to merge files
      if (length(list_sd_columns) == 1) {
        sd_merged <- dataframes_cur_seg
        #Filter for vars_in_segment
        sd_merged <- sd_merged[ , colnames(sd_merged) %in% vars_in_segment,
                                drop = FALSE]


      } else if (length(list_sd_columns) > 1 ||
                 is.null(list_sd_columns)) {
        # there is need to merge files
        #if the files are specified from user (length(study_data) > 1) or
        # there is no argument study_data
        if (is.null(list_sd_columns)) {
          #if there is no argument study_data
          # get id vars of the dataframe level metadata for merging
          # (only from dataframe and NOT from segment)
          dfr_in_segment <- dfr_in_segment[[1]]
          if (all(is.na(dfr_in_segment$DF_ID_VARS))) {
            # stop if all column is empty
            util_warning(
              paste0("Column %s in dataframe_level ",
                     "metadata is empty."),
              sQuote("DF_ID_VARS"))
          }

          id_variable <- unique(dfr_in_segment[, DF_ID_VARS, drop = TRUE])

          # Add the argument id_vars (works even if null or character(0))
          id_variable <- unique(c(id_variable, id_vars))

          util_expect_scalar(
            id_variable,
            allow_more_than_one = TRUE,
            allow_null = TRUE,     #It can be an empty id
            allow_na = TRUE)

          id_variable <-
            unlist(sapply(id_variable,
                          function(x) {
                            util_parse_assignments(
                              x,
                              split_char =
                                c(SPLIT_CHAR),
                              multi_variate_text =
                                TRUE,
                              split_on_any_split_char =
                                TRUE
                            )
                          })) #it also get rid of eventual NA in the id_variable
          names(id_variable) <- NULL
          id_variable <- unique(id_variable)


          if((DF_CODE) %in% colnames(dfr_in_segment) &&
             DATAFRAMES %in% colnames(meta_data)) {
            # In case there is a DATAFRAMES and DF_CODE, filter the data frames
            # imported in dataframe_cur_seg,
            # so that the var_names comes from the right data frame

            # First create a list with data frame names and list of variables
            # from metadata
            # corresponding to that data frame
            # (needed variables are in vars_in_segment)
            df1 <-
              meta_data[meta_data[[VAR_NAMES]] %in%
                          vars_in_segment, c(VAR_NAMES, DATAFRAMES),
                        drop = FALSE]
            dtf_with_expected_vars_from_itemlv <- lapply(
              setNames(nm = dataframe_names_cur_seg),
              FUN = function(x) {
                code_df <-
                  study_data_withcode[study_data_withcode[[DF_NAME]] %in%
                                        x, DF_CODE, drop = TRUE]
                df2 <- df1[grepl(sprintf("\\b(%s)\\b", code_df),
                                 df1$DATAFRAMES), ]
                df2[[VAR_NAMES]]
              }
            )

            # Filter each data frame imported in dataframe_cur_seg so that the
            # variables are matching the relative data frame in DATAFRAMES
            # in item_level md.
            # This way if a there are variables with the same name
            # in different data frames,they will be imported only
            # from the indicated DATAFRAMES in item_level and
            # not all of them.

            # Add always the id_vars to any data frame
            # (inside list_vars_DATAFRAMES) more specifically only import
            # id variables indicate in DF_ID_VARS of data frames and
            # the argument id_vars

            # Replace the dataframe_cur_seg with the one filtered
            dataframes_cur_seg <- lapply(
              setNames(nm = names(dtf_with_expected_vars_from_itemlv)),
              FUN = function(x) {
                # get a list of vars for each data frame based on the
                # DATAFRAMES col in item_level md
                list_vars_DATAFRAMES <-
                  dtf_with_expected_vars_from_itemlv[[x]]
                # get a list of id_vars from data
                my_id_vars <- dfr_in_segment[dfr_in_segment$DF_NAME == x,
                                             DF_ID_VARS, drop = TRUE]
                util_expect_scalar(
                  my_id_vars,
                  allow_more_than_one = FALSE,
                  allow_null = TRUE,
                  allow_na = TRUE
                )
                my_id_vars <- unlist(
                  util_parse_assignments(
                    my_id_vars,
                    split_char =
                      c(SPLIT_CHAR),
                    multi_variate_text =
                      TRUE,
                    split_on_any_split_char =
                      TRUE
                  )
                )

                names(my_id_vars) <- NULL
                my_id_vars <- unique(my_id_vars)
                # if a id_vars is provided as argument, this is added to
                # all segments
                my_id_vars <- unique(c(my_id_vars, id_vars))

                #add the id_vars to the list
                list_vars_DATAFRAMES <- c(list_vars_DATAFRAMES, my_id_vars)

                #get a list of vars for each data frame in dataframes_cur_seg
                list_vars_curseg <- names(dataframes_cur_seg[[x]])
                dataframes_cur_seg[[x]] <-
                  dataframes_cur_seg[[x]][,
                                          colnames(dataframes_cur_seg[[x]]) %in%
                                            list_vars_DATAFRAMES, drop = FALSE]
                dataframes_cur_seg[[x]]
              }
            )
          }

        } else {
          #TODO: check I think this can never happen
          #if the files are specified from user (length(list_sd_columns) > 1)
          id_variable <- id_vars
        }

        # If a segment is empty, do not create a report
        if (length(dataframes_cur_seg) == 0 ) {
          util_warning(sprintf("No data available for the segment %s",
                               sQuote(cur_seg)))
          return(NULL)
        }


        #Before merging select only variables in the current segment
        dataframes_cur_seg <-
          mapply(dataframes_cur_seg,
                 MoreArgs = list(vars_in_segment =
                                   vars_in_segment),
                 SIMPLIFY = FALSE,
                 FUN = function(x,
                                vars_in_segment){
                   x <- x[, colnames(x) %in% vars_in_segment, drop = FALSE]
                   return(x)
                   })

        #Before merging obtain the list of all variables
        vars_before <- unique(unlist(lapply(names(dataframes_cur_seg),
                                            FUN = function(x) {
                                              colnames(dataframes_cur_seg[[x]])
                                            })))


        #if there is no id_vars
        if(is.null(id_variable) || length(id_variable) == 0 ) {
          util_warning(c("Because no id variable is available, merging ",
                         "data frames could have created duplicated rows."))
          # marge data frame pairwise, without duplicating columns in common
          sd_merged <- Reduce(function(x, y) {
            merge(x, y, all = TRUE)
          }, dataframes_cur_seg)
        } else {
          # merge data frame pairwise, without duplicating columns in common
          sd_merged <- Reduce(function(x, y) {
            partial_merge <-
              suppressWarnings(merge(x, y, all = TRUE,
                                     by = intersect(
                                       intersect(colnames(x), colnames(y)),
                                       id_variable),
                                     suffixes = c("", "")
              ))
            partial_merge <-
              suppressWarnings(util_fix_merge_dups(partial_merge,
                                                   FALSE))
            return(partial_merge)
          }, dataframes_cur_seg)
        }

        #Check what happens to the variables during the merge
        if (length(vars_before) > length(colnames(sd_merged))) {
          util_warning(
            "Lost %d variables due to mapping problems. %d variables left.",
            length(vars_before) - length(colnames(sd_merged)),
            setdiff(vars_before, colnames(sd_merged)),
            applicability_problem = TRUE
          )
        } else if (length(vars_before) < length(colnames(sd_merged))) {
          util_warning(
            sprintf("There are duplicated variables due to merging: %s",
                    dQuote(setdiff(colnames(sd_merged),
                                   vars_before))),
            applicability_problem = TRUE
          )
        }
      }

      # Filter sd_merged by subgroup keeping only needed records----
      if (!is.null(subgroup)) {
        nrow_df <- nrow(sd_merged)
        # parse redcap rules to obtain contradiction in an usable way
        rule <- util_parse_redcap_rule(subgroup)

        sd_merged <- try(sd_merged[util_eval_rule(rule, ds1 = sd_merged,
                                                  meta_data = meta_data), ],
                         silent = TRUE)
        if (inherits(sd_merged, "try-error")) {
          # if the subgroup selection did not work
          err <- conditionMessage(attr(sd_merged, "condition"))
          util_error("The subgroup rule %s was not acceptable: %s",
                     dQuote(subgroup), err)
        }
        if (nrow(sd_merged) == nrow_df) {
          util_warning(
            c("The number of cases did not change after applying the",
              "subgroup filter %s"),
            dQuote(subgroup))
        } else if (nrow(sd_merged) == 0 && nrow_df > 0) {
          util_warning(
            c(
              "After using subgroup rule %s, no dataset is left. You may",
              "have provided an impossible condition, e.g., filtering",
              "for a variable equal to %s, but the actual levels would be %s."
            ),
            dQuote(subgroup),
            dQuote("females"),
            dQuote("female")
          )
        }
        rm(nrow_df, rule)
      }



      # Define the strata_column ----
      # if no argument to split present,
      # create a list anyways with all_observations
      if (is.null(strata_column) ||
          !strata_column %in% colnames(sd_merged)) {
        .sd_list <- list(all_observations = sd_merged)
      } else if (!is.null(strata_column) &&
                 strata_column %in% colnames(sd_merged)) {
        # split the data
        .sd_list <- split(sd_merged, sd_merged[[strata_column]])


        # provide a warning message in case there are NAs in the study data in
        # the column selected as strata_column
        if (anyNA(sd_merged[[strata_column]])) {
          # util_warning(m = sprintf(c("There are %d missing values inside the",
          #                           " column selected as strata_column"),
          #                         sum(is.na(sd_merged[[strata_column]])) )
          # )
          new_group_NA <- sd_merged[is.na(sd_merged[[strata_column]]), ]
          list_new_group_NA <- setNames(list(new_group_NA), nm = "NAs_group")
          .sd_list <- c(.sd_list, list_new_group_NA)
          rm(new_group_NA, list_new_group_NA)
        }

        ## filtering for the study_data_strata (vector of names or regexp)
        if (!is.null(strata_select)) {
          # Value labels from VALUE_LABEL_TABLE and CODE_LIST_TABLE
          if (is.na(meta_data[meta_data[VAR_NAMES] == strata_column,
                              VALUE_LABELS, drop = TRUE]) ||
              is.null(meta_data[meta_data[VAR_NAMES] == strata_column,
                                VALUE_LABELS, drop = TRUE])) {
            value_label_table_name <- meta_data[meta_data[VAR_NAMES] ==
                                                  strata_column,
                                                VALUE_LABEL_TABLE, drop = TRUE]

            try(util_expect_data_frame(value_label_table_name),
                silent = TRUE)
            if (!is.data.frame(value_label_table_name)) {
              try(util_expect_data_frame("CODE_LIST_TABLE"),
                  silent = TRUE)
              if (!is.data.frame(CODE_LIST_TABLE)) {
                util_message(sprintf(
                  "No value_label_table_name %s found",
                  dQuote(value_label_table_name)
                ))
                value_label_table_name <- data.frame(CODE_VALUE = character(0),
                                                     CODE_LABEL = character(0))
              } else {
                #select in CODE_LIST_TABLE$VALUE_LABEL_TABLE only the value_label_table_name
                value_label_table_name <-
                  CODE_LIST_TABLE[CODE_LIST_TABLE$VALUE_LABEL_TABLE ==
                                    value_label_table_name, , drop = FALSE]
              }
            }
            expected_strata <- setNames(value_label_table_name$CODE_LABEL,
                                        nm = value_label_table_name$CODE_VALUE)
          } else {
            expected_strata <- unlist(
              util_parse_assignments(
                meta_data[meta_data[VAR_NAMES] == strata_column,
                          VALUE_LABELS, drop = TRUE],
                split_char = SPLIT_CHAR,
                split_on_any_split_char = TRUE,
                multi_variate_text = TRUE
              )
            )
          }


          ## definition of selection_type and strata_select
          if (!is.null(selection_type)) {
            if (selection_type == "value") {
              # check if the typed levels exist in the segment_column possible
              # levels before to select it
              if (!any(strata_select %in% names(.sd_list))) {
                # stop if selection does not match any strata
                util_error(
                  c("No strata_column stratum matches the",
                    "provided names: %s"),
                  dQuote(strata_select))
              } else {
                .sd_list <- .sd_list[strata_select]
                # in case empty, remove empty from list
                # and give a warning
                .sd_list <- .sd_list[!is.na(names(.sd_list))]
                if (length(strata_select) > length(.sd_list)) {
                  util_warning(c(
                    "One or more strata were not",
                    "present in the study_data"
                  ))
                }
              }
            } else if (selection_type == "v_label") {
              value_of_strata_select <-
                expected_strata[expected_strata %in% strata_select]
              value_of_strata_select <- names(value_of_strata_select)
              # check if the typed levels exist in the segment_column possible
              # levels before to select it
              if (length(value_of_strata_select) == 0 &&
                  !any(value_of_strata_select %in% names(.sd_list))) {
                # stop if selection does not match any strata
                util_error(
                  c("No strata_column stratum matches the",
                    "provided names: %s"),
                  dQuote(strata_select))
              } else {
                .sd_list <- .sd_list[value_of_strata_select]
                # in case a strata is not present, remove empty from list
                # and give a warning
                .sd_list <- .sd_list[!is.na(names(.sd_list))]
                if (length(value_of_strata_select) > length(.sd_list)) {
                  util_warning(c(
                    "One or more strata were not",
                    "present in the study_data"
                  ))
                }
              }
            } else if (selection_type == "regex") {
              all_labels_in_sd_list <-
                expected_strata[names(expected_strata) %in% names(.sd_list)]
              value_of_strata_select <-
                all_labels_in_sd_list[grepl(strata_select, all_labels_in_sd_list)]
              if (length(names(.sd_list)[grepl(strata_select,
                                               names(.sd_list))]) > 0) {
                #regex match at least one strata
                .sd_list <- .sd_list[grepl(strata_select, names(.sd_list))]
                .sd_list <- .sd_list[!is.na(names(.sd_list))]
              } else if (length(names(.sd_list)[names(.sd_list) %in%
                                                names(value_of_strata_select)]) > 0) {
                .sd_list <- .sd_list[names(.sd_list)[names(.sd_list) %in%
                                                       names(value_of_strata_select)]]
                .sd_list <- .sd_list[!is.na(names(.sd_list))]
              } else {
                util_error(
                  c("No strata_column stratum matches the",
                    "provided regular expression: %s"),
                  dQuote(strata_select))
              }
            }
          } else {
            # if  selection_type is null
            # give a warning and then try to find matches
            util_message(
              c(
                "The argument selection_type is not provided, ",
                "the function will try to find any matches with value, ",
                "value labels or possible regular expressions"
              )
            )
            #check if the strata_select matches a names(.sd_list)
            if (any(strata_select %in% names(.sd_list))) {
              .sd_list <- .sd_list[strata_select]
              # in case present, remove empty from list
              .sd_list <- .sd_list[!is.na(names(.sd_list))]
            } else if (any(strata_select %in% expected_strata)) {
              #in case the strata_select matches a label of names(.sd_list)
              labels_strata_select <- expected_strata[expected_strata %in%
                                                        strata_select]
              value_of_strata_select <- names(labels_strata_select)
              .sd_list <- .sd_list[value_of_strata_select]
              # in case present, remove empty from list
              .sd_list <- .sd_list[!is.na(names(.sd_list))]
            } else if (length(names(.sd_list)[grepl(strata_select,
                                                    names(.sd_list))]) > 0) {
              #regex matches a value
              .sd_list <- .sd_list[names(.sd_list)[grepl(strata_select,
                                                         names(.sd_list))]]
            } else {
              labels_strata_select <-
                expected_strata[grepl(strata_select, expected_strata)]

              if (length(labels_strata_select) > 0) {
                value_of_strata_select <- names(labels_strata_select)
                .sd_list <- .sd_list[value_of_strata_select]
                # in case present, remove empty from list
                .sd_list <- .sd_list[!is.na(names(.sd_list))]
              } else {
                #it does not maches any labels or anything else
                util_error(
                  c("No strata_column stratum matches the",
                    "provided names: %s" ),
                  dQuote(strata_select))
              }
            }
          }
        }

        ## remove the strata to exclude
        if (!is.null(strata_exclude)) {
          # Value labels from VALUE_LABEL_TABLE and CODE_LIST_TABLE
          if (is.na(meta_data[meta_data[VAR_NAMES] == strata_column,
                              VALUE_LABELS, drop = TRUE]) ||
              is.null(meta_data[meta_data[VAR_NAMES] == strata_column,
                                VALUE_LABELS, drop = TRUE])) {
            value_label_table_name <- meta_data[meta_data[VAR_NAMES] ==
                                                  strata_column,
                                                VALUE_LABEL_TABLE, drop = TRUE]

            try(util_expect_data_frame(value_label_table_name),
                silent = TRUE)
            if (!is.data.frame(value_label_table_name)) {
              try(util_expect_data_frame("CODE_LIST_TABLE"),
                  silent = TRUE)
              if (!is.data.frame(CODE_LIST_TABLE)) {
                util_message(sprintf(
                  "No value_label_table_name %s found",
                  dQuote(value_label_table_name)))
                value_label_table_name <- data.frame(CODE_VALUE = character(0),
                                                     CODE_LABEL = character(0))
              } else {
                #select in CODE_LIST_TABLE$VALUE_LABEL_TABLE only the value_label_table_name
                value_label_table_name <-
                  CODE_LIST_TABLE[CODE_LIST_TABLE$VALUE_LABEL_TABLE ==
                                    value_label_table_name, , drop = FALSE]
              }
            }
            expected_strata <- setNames(value_label_table_name$CODE_LABEL,
                                        nm = value_label_table_name$CODE_VALUE)
          } else {
            expected_strata <- unlist(
              util_parse_assignments(
                meta_data[meta_data[VAR_NAMES] == strata_column,
                          VALUE_LABELS, drop = TRUE],
                split_char = SPLIT_CHAR,
                split_on_any_split_char = TRUE,
                multi_variate_text = TRUE
              )
            )
          }

          if (!is.null(selection_type)) {
            if (selection_type == "value") {
              # check if the typed levels exist in the segment_column possible
              # levels before to select it
              if (!any(strata_exclude %in% names(.sd_list))) {
                # stop if selection does not match any strata
                util_error(
                  c("No strata_column stratum matches the",
                    "strata to exclude: %s"),
                  dQuote(strata_exclude))
              } else {
                #remove strata_exclude elements from list
                .sd_list <- .sd_list[names(.sd_list) %in% strata_exclude == FALSE]
                # stop if removing all strata
                util_stop_if_not(length(.sd_list) > 0,
                                 label = paste0("No strata remain after",
                                                " excluding selected ones"))
              }
            } else if (selection_type == "v_label") {
              value_of_strata_select <-
                expected_strata[expected_strata %in% strata_exclude]
              value_of_strata_select <- names(value_of_strata_select)
              # check if the typed levels exist in the segment_column possible
              # levels before to select it
              if (length(value_of_strata_select) == 0 &&
                  !any(value_of_strata_select %in% names(.sd_list))) {
                # stop if selection does not match any strata
                util_error(
                  c("No strata_column stratum matches the",
                    "strata_exclude: %s"),
                  dQuote(strata_exclude))
              } else {
                .sd_list <- .sd_list[names(.sd_list) %in%
                                       value_of_strata_select == FALSE]
                #stop if no strata remain
                if (length(.sd_list) == 0) {
                  util_error(
                    c("No strata_column stratum remains",
                      "after removing strata_exclude: %s"),
                    dQuote(strata_exclude))
                }
              }
            } else if (selection_type == "regex") {
              all_labels_in_sd_list <-
                expected_strata[names(expected_strata) %in% names(.sd_list)]
              value_of_strata_select <-
                all_labels_in_sd_list[grepl(strata_exclude,
                                            all_labels_in_sd_list)]
              if (length(names(.sd_list)[grepl(strata_exclude,
                                               names(.sd_list))]) > 0) {
                #regex match at least one strata - remove the strata
                .sd_list <- .sd_list[!grepl(strata_exclude, names(.sd_list))]
                #stop if no strata remain
                if (length(.sd_list) == 0) {
                  util_error(
                    c("No strata_column stratum remains",
                      "after removing strata_exclude: %s"),
                    dQuote(strata_exclude))
                }
              } else if (length(names(.sd_list)[names(.sd_list) %in%
                                                names(value_of_strata_select)]) > 0) {
                .sd_list <- .sd_list[names(.sd_list) %in%
                                       names(value_of_strata_select) == FALSE]
              } else {
                util_error(
                  c("No strata_column stratum matches the",
                    "provided regular expression: %s"),
                  dQuote(strata_exclude))
              }
            }
          } else {
            # if  selection_type is null
            # give a warning and then try to find matches
            util_warning(
              c(
                "The argument selection_type not provided, ",
                "the function will try to find any matches with value, ",
                "value labels or possible regular expressions"
              )
            )
            #check if the strata_exclude matches a names(.sd_list)
            if (any(strata_exclude %in% names(.sd_list))) {
              #remove strata_exclude elements from list
              .sd_list <- .sd_list[names(.sd_list) %in% strata_exclude == FALSE]
              # stop if removing all strata
              util_stop_if_not(length(.sd_list) > 0,
                               label = "No strata remain after excluding selected ones")
            } else if (any(strata_exclude %in% expected_strata)) {
              #in case the strata_exclude matches a label of names(.sd_list)
              value_of_strata_select <-
                expected_strata[expected_strata %in% strata_exclude]
              value_of_strata_select <- names(value_of_strata_select)
              # check if the typed levels exist in the segment_column possible
              # levels before to select it
              if (length(value_of_strata_select) == 0 &&
                  !any(value_of_strata_select %in% names(.sd_list))) {
                # stop if selection does not match any strata
                util_error(
                  c("No strata_column stratum matches the",
                    "strata_exclude: %s"),
                  dQuote(strata_exclude))
              } else {
                .sd_list <- .sd_list[names(.sd_list) %in%
                                       value_of_strata_select == FALSE]
                #stop if no strata remain
                if (length(.sd_list) == 0) {
                  util_error(
                    c("No strata_column stratum remains",
                      "after removing strata_exclude: %s"),
                    dQuote(strata_exclude))
                }
              }
            } else {
              #in case no matches found for value or v_label try regex
              all_labels_in_sd_list <-
                expected_strata[names(expected_strata) %in% names(.sd_list)]
              value_of_strata_select <-
                all_labels_in_sd_list[grepl(strata_exclude,
                                            all_labels_in_sd_list)]
              #if the regex matches a value
              if (length(names(.sd_list)[grepl(strata_exclude,
                                               names(.sd_list))]) > 0) {
                #regex match at least one strata - remove the strata
                .sd_list <- .sd_list[!grepl(strata_exclude, names(.sd_list))]
                #stop if no strata remain
                if (length(.sd_list) == 0) {
                  util_error(
                    c("No strata_column stratum remains",
                      "after removing strata_exclude: %s"),
                    dQuote(strata_exclude))
                }
                #if the regex matches a v_label
              } else if (length(names(.sd_list)[names(.sd_list) %in%
                                                names(value_of_strata_select)]) > 0) {
                .sd_list <- .sd_list[names(.sd_list) %in%
                                       names(value_of_strata_select) == FALSE]
              } else {
                #if the regex also has no matches
                util_error(
                  c("No strata_column stratum matches the",
                    "provided strata_exclude: %s"),
                  dQuote(strata_exclude))
              }
            }
          }
        }

        # convert names of the groups
        # to VAR_NAMES_actualGroupValues
        names(.sd_list) <- paste0(strata_column_label, "_", names(.sd_list))
      }

      # INNER loop starts here: split by strata_column ----
      mapply(
        sd = .sd_list,
        sdn = names(.sd_list),
        MoreArgs = list(
          md = meta_data,
          sd_merged = sd_merged,
          name_of_study_data = name_of_study_data,
          call_report_by = call_report_by,
          resp_vars_in_segment = resp_vars_in_segment
        ),
        SIMPLIFY = FALSE,
        FUN = function(sd,
                       sdn,
                       md,
                       sd_merged,
                       name_of_study_data,
                       call_report_by,
                       resp_vars_in_segment) {

          # import a list containing original study data (too long), and
          # new assigned names, in case it was created in a previous apply,
          # otherwise create an empty one
          if (exists("..INFO_SD_NAME_FOR_REPORT", .dataframe_environment())) {
            info_sd_name <- prep_get_data_frame("..INFO_SD_NAME_FOR_REPORT")
            info_sd_name <- as.list(info_sd_name)
          } else {
            info_sd_name <- list()
          }
          # Define a new name composed of study_data_current
          # stratum_current segment
          level_name <- sdn

          # message on progresses
          util_message(sprintf(
            "Segment %s, Stratum %s..",
            sQuote(cur_seg),
            sQuote(level_name)))
          progress_msg(sprintf(
            "Segment %s, Stratum %s...",
            sQuote(cur_seg),
            sQuote(level_name)))
          # update the progress bar
          p$i <- p$i + 1
          progress(100 * p$i / p$N)
          .outer_by_env$outer_by <- list(
            i = p$i,
            n = p$N,
            msg = sprintf(
              "Segment %s, Stratum %s...",
              sQuote(cur_seg),
              sQuote(level_name))
          )

          # add the study data name that are too long to a list
          # containing original study data names and
          # new assigned names
          if (!is.null(name_of_study_data) && (
            startsWith(name_of_study_data, "https://") ||
            startsWith(name_of_study_data, "http://") ||
            startsWith(name_of_study_data, "ftp://") ||
            startsWith(name_of_study_data, "ftps://") ||
            startsWith(name_of_study_data, "dbx://") ||
            nchar(name_of_study_data) > 20
          )) {
            # if not already present in the list,
            # add the name to a general list for the overview
            if (!name_of_study_data %in%
                unlist(info_sd_name, use.names = FALSE)) {
              info_sd_name <-
                c(info_sd_name, setNames(as.list(name_of_study_data),
                                         nm = paste0("SD", length(
                                           info_sd_name
                                         ) + 1)))
            }

            # add the name to a local list for the report
            info_sd_name_vector <- setNames(names(info_sd_name), info_sd_name)
            info_sd_name_vector <-
              info_sd_name_vector[names(info_sd_name_vector) ==
                                    name_of_study_data]
            # turn it to a list to be used as arg. in dq_report2
            info_sd_name_per_report <-
              setNames(as.list(names(info_sd_name_vector)), info_sd_name_vector)
          } else {
            info_sd_name_vector <- NULL
            info_sd_name_per_report <- NULL
          }
          #use abbreviation SD# instead of study data file names in folder names
          if (!is.null(info_sd_name_vector)) {
            name_of_study_data <- info_sd_name_vector[name_of_study_data]
          }

          # Define a new name composed of study_data_current
          # stratum_current segment
          sdn <- paste0(name_of_study_data, "_", sdn, "_", cur_seg)

          subtitle_report <- paste0("Study data: ", name_of_study_data)

          if (!is.null(subgroup)) {
            subtitle_report <- paste0(subtitle_report, " Subgroup: ", subgroup)
          }


          #     }
          # update the cache with the new list info_sd_name for the overview
          if (length(info_sd_name) != 0) {
            ..INFO_SD_NAME_FOR_REPORT <- data.frame(info_sd_name)
            prep_add_data_frames("..INFO_SD_NAME_FOR_REPORT" =
                                   ..INFO_SD_NAME_FOR_REPORT)
          }
          title_report <-
            paste0("Data quality report on ", level_name, ".", cur_seg)

          # Created to later clean the cache to improve speed using function
          # prep_remove_from_cache
          old_dataframes <- prep_list_dataframes()

          # add missing tables specified as arguments
          if (!is.null(missing_tables)) {
            suppressMessages(lapply(
              missing_tables,
              FUN = function(x) {
                assign(x, get(x), envir = .dataframe_environment())
                invisible(.dataframe_environment())
              }
            ))
          }

          #in case of study_data being used as argument indicating the name of
          # the data frame, saved it here to prevent its change in the cache
          # during running the function dq_report2, and restored after
          if ("study_data" %in% names(.dataframe_environment())) {
            saved_study_data <- prep_get_data_frame("study_data")
          } else {
            saved_study_data <- NULL
          }

          if (!is.null(storr_factory)) {
            storr_factory_clone <- rlang::duplicate(storr_factory)
            environment(storr_factory_clone) <-
              rlang::env_clone(environment(storr_factory_clone))
            ns <- get("namespace", environment(storr_factory))
            ns <- paste0(ns, "/", sdn)
            assign("namespace", ns, environment(storr_factory_clone))
          } else {
            storr_factory_clone <- NULL
          }


          # If a segment is empty, do not create a report
          if (nrow(sd) == 0 || is.null(sd)) {
            util_warning(sprintf("No data available to create the report %s",
                                 sQuote(title_report)))
            return(NULL)
          }

          allowed_args <- names(formals(dataquieR::dq_report2))

          clean_dots <- dots[names(dots) %in% allowed_args]

          ui <- info_sd_name_per_report

          if (!is.null(user_info) && is.list(user_info)) {
            ui <- user_info
            ui[names(info_sd_name_per_report)] <- info_sd_name_per_report
          }

          args <- c(
            list(
              study_data = sd,
              meta_data = md,
              resp_vars = resp_vars_in_segment[[cur_seg]],
              meta_data_cross_item = cil_in_segment[[cur_seg]],
              meta_data_segment = seg_in_segment[[cur_seg]],
              meta_data_dataframe = dfr_in_segment[[1]],
              meta_data_item_computation = computed_in_segment[[cur_seg]],
              label_col = label_col,
              split_segments = split_segments,
              user_info = ui,
              title = title_report,
              subtitle = subtitle_report,
              storr_factory = storr_factory_clone,
              amend = amend,
              checkpoint_resumed = checkpoint_resumed,
              name_of_study_data = name_of_study_data,
              author = author
            ),
            clean_dots
          )

          args$output_dir <- NULL

          # create the different sub-reports ----
          r <- try(
            do.call(dq_report2, args)
          )

          ###Check here if it is an error, put r <- NULL, empty report()
          attr(r, "label_modification_text") <- trimws(paste(
            attr(r, "label_modification_text"),
            mod_label$label_modification_text
          ))

          attr(r, "label_modification_table") <-
            rbind(attr(r, "label_modification_table"),
                  mod_label$label_modification_table)


          # Define subgroup for technical information
          if (!is.null(subgroup)) {
            Subgroup_info <- subgroup
          } else {
            Subgroup_info <- "Not specified"
          }

          # Add information in the "Technical information" part of the report
          attr(r, "properties") <- c(
            attr(r, "properties"),
            list(
              Study_data = name_of_study_data,
              Segment = cur_seg,
              Stratum = level_name,
              Subgroup = Subgroup_info
            )
          )

          rm(sd_merged, sd, md, Subgroup_info)
          #remove the files creating by the dq_report function from the cache
          new_dataframes <- prep_list_dataframes()
          names_to_remove <- setdiff(new_dataframes, old_dataframes)
          suppressMessages(prep_remove_from_cache(names_to_remove))
          # restore the study_data as mentioned before running the report
          suppressMessages(prep_remove_from_cache("study_data"))
          if (!is.null(saved_study_data)) {
            prep_add_data_frames(study_data = saved_study_data)
          }
          gc()

          # in case an output directory is defined, save reports and delete them ----
          if (length(output_dir) == 1 && dir.exists(output_dir)) {
            if (inherits(r, "try-error")) {
              #check if instead of the report there is an error
              util_warning(
                "Could not compute report for %s: %s.",
                dQuote(sdn),
                sQuote(conditionMessage(attr(
                  r, "condition")))
              )
            } else {
              # Save the report as R object
              s_res <-
                try(prep_save_report(r, file.path(output_dir, gsub(
                  "[^a-zA-Z0-9_\\.]",
                  "",
                  sprintf("report_%s.dq2", sdn)
                ))))
              if (inherits(s_res, "try-error")) {
                # if the report could not be saved
                # gives a warning
                util_warning(
                  "Could not save report for %s: %s.",
                  dQuote(sdn),
                  sQuote(conditionMessage(attr(
                    s_res, "condition")))
                )
              }
              # remove object to reduce memory use
              rm(s_res)

              # Save the summaries for a quick overview of all the reports ----
              obj <- summary(r)
              this <- attr(obj, "this")
              this$stratum <- level_name
              this$used_data_file <- name_of_study_data
              this$segment <- cur_seg #not for combining
              this$sdn <- sdn
              attr(obj, "this") <- this

              s_res <-
                try(saveRDS(obj, file.path(output_dir, gsub(
                  "[^a-zA-Z0-9_\\.]",
                  "",
                  sprintf("report_summary_%s.RDS", sdn)
                ))))
              if (inherits(s_res, "try-error")) {
                # if the summary could not be saved
                util_warning(
                  "Could not save report summary for %s: %s.",
                  dQuote(sdn),
                  sQuote(conditionMessage(attr(
                    s_res, "condition")))
                )
              }
              obj <- NULL
              try({
                obj  <- util_setup_dashboard(r,
                                             make_links = FALSE,
                                             return_table_only = TRUE)
                if (!is.null(obj)) {
                  attr(obj, "name_of_study_data") <- name_of_study_data
                  attr(obj, "level_name") <- level_name
                }
              })
              # paste0(name_of_study_data, dashboard$Variables, level_name)
              s_res <-
                try(saveRDS(obj, file.path(output_dir, gsub(
                  "[^a-zA-Z0-9_\\.]",
                  "",
                  sprintf("report_dashboard_%s.RDS", sdn)
                ))))
              if (inherits(s_res, "try-error")) {
                # if the summary could not be saved
                util_warning(
                  "Could not save report dashboard table for %s: %s.",
                  dQuote(sdn),
                  sQuote(conditionMessage(attr(
                    s_res, "condition")))
                )
              }
              # remove object to reduce memory use
              rm(s_res)
              .dir <- file.path(output_dir,
                                gsub("[^a-zA-Z0-9_\\.]", "",
                                     sprintf("report_%s", sdn)))
              dir.create(.dir,
                         showWarnings = FALSE,
                         recursive = TRUE)
              if (!dir.exists(.dir)) {
                # warning if the directory could not be created
                util_warning(
                  paste0("Could not create directory %s. ",
                         "No output for this segment"),
                  dQuote(.dir))
              }
              # prepare the folder for saving the rendered report html file
              if (also_print) {
                if (dir.exists(.dir)) {
                  # these two arguments may be passed to the print function
                  pass_args <- dots[names(dots) %in%
                                      c("cores", "block_load_factor")]
                  # save the rendered html files in the directory
                  p_res <- util_try_with_trace(rlang::eval_bare(
                    rlang::call2(print.dataquieR_resultset2,
                                 r,
                                 dir = .dir,
                                 view = FALSE,
                                 disable_plotly = disable_plotly,
                                 by_report = TRUE,
                                 !!!pass_args
                    )
                  ))
                  #creating the back link
                  if (inherits(p_res, "try-error")) {
                    if (isTRUE(getOption("dataquieR.traceback",
                                         dataquieR.traceback_default))) {
                      util_warning(util_condition_from_try_error(p_res))
                    }
                    # gives a warning if the report could not be rendered
                    util_warning(
                      "Could not create HTML report for %s: %s.",
                      dQuote(.dir),
                      sQuote(conditionMessage(
                        attr(p_res, "condition")))
                    )
                  }
                  rm(p_res)
                }
              }
            }
            # remove the reports to reduce memory use
            rm(r)
            gc()
            return(invisible(NULL))
          } else {
            #if no output directory is indicated, return the report
            return(r)
          }
          gc()
        }
      )
    }
  )

  if (!missing(output_dir) && dir.exists(output_dir)) {
    maybe <- function(x) {
      if (exists(x, envir = parent.frame())) {
        setNames(list(rlang::maybe_missing(get(x, envir = parent.frame()))),
                 nm = x)
      } else {
        list()
      }
    }
    saveRDS(
      c(
        maybe("strata_column"),
        maybe("segment_column"),
        maybe("strata_column_label"),
        maybe("subgroup"),
        maybe("mod_label"),
        maybe("disable_plotly"),
        maybe("title"),
        maybe("start_time"),
        maybe("rep_id"),
        maybe("subtitle"),
        maybe("author"),
        maybe("user_info"),
        maybe("by_call")
      ),
      file = file.path(output_dir, "report_by_meta.RDS")
    )
  }

  # create the html overview page that links all sub-reports created -----
  if (!missing(output_dir) && also_print) {
    util_create_report_by_overview(
      output_dir = output_dir
    )
  }
  prep_purge_data_frame_cache()

  if (util_really_rstudio()) {
    rstudioapi::executeCommand("activateConsole")
  }

  if (missing(output_dir)) {
    if (view) {
      return(overall_res)
    } else {
      return(invisible(overall_res))
    }
  } else {
    if (view && also_print) {
      # util_view_file(content_file) done by exit handler
      return(overall_res)
    } else {
      return(invisible(overall_res))
    }
  }

}

# return a list of .hi, .hp, and .hm, the handles for the
# progress hooks for init, progress and progress-messages
# so use it as follows:
# .hi <- .hp <- .hm <- NULL
# list2env(util_init_html_progress(...), envir = environent)
util_init_html_progress <- function(output_dir, content_file, title, view,
                                    rep_id, start_time = Sys.time()) {
  util_seed_last_error()
  force(start_time)
  packageName <- utils::packageName()
  hook_store <- new.env(parent = emptyenv())
  hook_store$n <- Inf
  hook_store$percent <- 0
  hook_store$status <- "Initializing"
  hook_store$msg <- ""

  hook <- function(n = hook_store$n,
                   percent = hook_store$percent,
                   status = hook_store$status,
                   msg = hook_store$msg) {
    if (!missing(n)) {
      hook_store$n <- n
    }
    if (!missing(percent)) {
      hook_store$percent <- percent
    }
    if (!missing(status)) {
      hook_store$status <- status
    }
    if (!missing(msg)) {
      hook_store$msg <- msg
    }
    util_write_index_html(
      content_file,
      util_index_loading_lines(
        title  = title,
        message = hook_store$status,
        detail  = hook_store$msg,
        reload_ms = 1200L,
        n = hook_store$n,
        percent = hook_store$percent
      )
    )
    util_write_renderinfo_js_json(output_dir = output_dir,
                                  rep_id = rep_id,
                                  start_time = start_time,
                                  end_time = Sys.time())
  }

  .hi <- prep_register_progress_hook(type = "init",
                                     hook)
  withr::defer_parent(prep_deregister_progress_hook(.hi, verbose = FALSE))
  .hp <- prep_register_progress_hook(type = "progress",
                                     hook)
  withr::defer_parent(prep_deregister_progress_hook(.hp, verbose = FALSE))
  .hm <- prep_register_progress_hook(type = "msg",
                                     hook)
  withr::defer_parent(prep_deregister_progress_hook(.hm, verbose = FALSE))

  if (!dir.exists(file.path(output_dir, ".report"))) {
    if (!dir.create(file.path(output_dir, ".report"))) {
      util_error("Could not create %s for HTML output",
                 dQuote(file.path(output_dir, ".report")))
    }
  }

  file.copy(system.file("logos",
                        "dataquieR_48x48.png",
                        package = packageName),
            file.path(output_dir, ".report", "logo.png"))

  if (view) {
    util_write_index_html(
      content_file,
      util_index_loading_lines(
        title  = title,
        message = hook_store$status,
        detail  = hook_store$msg,
        reload_ms = 1200L,
        n = hook_store$n,
        percent = hook_store$percent
      )
    )
    util_write_renderinfo_js_json(output_dir = output_dir,
                                  rep_id = rep_id,
                                  start_time = start_time,
                                  end_time = Sys.time())
    withr::defer_parent({
      cl <- ""
      cl <- suppressWarnings(try(readLines(content_file), silent = TRUE))
      if (!any(grepl(fixed = TRUE, "<!-- done -->", cl))) {
        util_write_index_html(
          content_file,
          util_index_error_lines(
            title = title,
            message = "Report was not created... it was cancelled or an error occurred."
          )
        )
        #print(dirname(content_file))
        # util_message("Could not find done in %s: %s", dQuote(content_file), util_pretty_vector_string(cl))
        unlink(file.path(dirname(content_file), "renderinfo.js")) # this is the marker for an error
      }
    })
    if (util_works_in_rs_viewer(content_file)) {
      # RStudio has a deadlock, if the autorefresh runs in its viewer, whatever, we have progressbars there, anyways.
      withr::defer_parent({
        cl <- ""
        cl <- suppressWarnings(try(readLines(content_file), silent = TRUE))
        if (!any(grepl(fixed = TRUE, "<!-- done -->", cl))) {
          if (!util_is_last_error_sentinel()) {
            p <- NULL
            try(p <- rlang::last_error(), silent = TRUE)
            rlang::abort("Cancelled.",
                         parent = p)
          } else {
            util_error("Cancelled.")
          }
        } else {
          util_view_file(content_file)
        }
      })
    } else {
      util_view_file(content_file)
    }
  }
  list(.hi = .hi,
       .hp = .hp,
       .hm = .hm)
}

util_seed_last_error <- function(context = "pipeline phase started") {
  cnd <- attr(try(
    rlang::abort(
      message = context,
      class = "my_last_error_sentinel"
    ), silent = TRUE
  ), "condition")
  rlang::entrace(cnd)
}

util_is_last_error_sentinel <- function() {
  err <- tryCatch(rlang::last_error(), error = function(e) NULL)
  inherits(err, "my_last_error_sentinel")
}

test_that("dq_report_by works with content na", {
  skip_if_not_installed("DT")
  skip_if_not_installed("markdown")
  skip_if_not_installed("stringdist")

  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  withr::local_options(dataquieR.CONDITIONS_WITH_STACKTRACE = TRUE,
                       dataquieR.ERRORS_WITH_CALLER = TRUE,
                       dataquieR.WARNINGS_WITH_CALLER = TRUE,
                       dataquieR.MESSAGES_WITH_CALLER = TRUE)
    skip_on_cran() # slow test
    # TODO: test with all sorts of _by calls and on windows
    target <- withr::local_tempdir("testdqareportby")

    expect_message2(dq_report_by("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData",
                 dimensions = "int",
                 cores = NULL,
                 segment_column = NULL,
                 meta_data_v2 = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx",
                 output_dir = !!file.path(target, "na"),
                 also_print = TRUE))

    # TODO: check target path for all expected artifacts, namely the dq2 files, the .output_dirs containing the HTML reports and the summary RDS files

})

test_that("default behavior returns list of resultsets", {
  skip_on_cran() # slow
  skip_if_not_installed("stringdist")
  study_data <- data.frame(id = 1:4,
                           CENTER = c("A", "A", "B", "B"),
                           val1 = c(1, NA, 3, 4),
                           val2 = c(5, 6, NA, 8)
  )
  meta_data <- data.frame(
    VAR_NAMES = c("val1", "val2", "CENTER"),
    LABEL = c("Variable 1", "Variable 2", "Center"),
    DATA_TYPE = c("integer", "integer", "string"),
    MISSING_LIST = c("", "", ""),
    VALUE_LABELS = c(NA, NA, "A|B"),
    stringsAsFactors = FALSE
  )

  rep <- dq_report_by(study_data, meta_data, label_col = "LABEL",
                      strata_column = "CENTER", segment_column = NULL, cores = NULL)
  expect_true(is.list(rep))
  expect_named(rep$all_variables, c("Center_A", "Center_B"))
  expect_true(all(sapply(rep$all_variables, inherits, "dataquieR_resultset2")))
})

test_that("segment_column = NULL returns flat structure", {
  skip_on_cran() # slow
  skip_if_not_installed("stringdist")
  study_data <- data.frame(id = 1:4,
                           CENTER = c("A", "A", "B", "B"),
                           val1 = c(1, NA, 3, 4),
                           val2 = c(5, 6, NA, 8)
  )
  meta_data <- data.frame(
    VAR_NAMES = c("val1", "val2", "CENTER"),
    LABEL = c("Variable 1", "Variable 2", "Center"),
    DATA_TYPE = c("integer", "integer", "string"),
    MISSING_LIST = c("", "", ""),
    VALUE_LABELS = c(NA, NA, "A|B"),
    stringsAsFactors = FALSE
  )
  rep2 <- dq_report_by(study_data, meta_data, label_col = "LABEL",
                       strata_column = "CENTER", segment_column = NULL, cores = NULL)
  expect_true(all(sapply(rep2, function(r) is.list(r) && is.null(r$segment))))
})

test_that("disable_plotly disables plotly in HTML output", {
  skip_on_cran() # slow

  skip_if_not_installed("DT")

  study_data <- data.frame(id = 1:4,
                           CENTER = c("A", "A", "B", "B"),
                           val1 = c(1, NA, 3, 4),
                           val2 = c(5, 6, NA, 8)
  )
  meta_data <- data.frame(
    VAR_NAMES = c("val1", "val2", "CENTER"),
    LABEL = c("Variable 1", "Variable 2", "Center"),
    DATA_TYPE = c("integer", "integer", "string"),
    MISSING_LIST = c("", "", ""),
    VALUE_LABELS = c(NA, NA, "A|B"),
    stringsAsFactors = FALSE
  )
  td <- tempfile()
  on.exit(unlink(td, recursive = TRUE), add = TRUE)
  result <- dq_report_by(study_data, meta_data, label_col = "LABEL",
                         strata_column = "CENTER",
                         output_dir = td, also_print = TRUE, view = FALSE,
                         disable_plotly = TRUE, segment_column = NULL,
                         cores = NULL)
  html_files <- list.files(td, pattern = "\\.html$", full.names = TRUE,
                           all.files = TRUE,
                           recursive = TRUE)
  expect_gt(length(html_files), 0)
  html_contents <- paste(unlist(lapply(html_files,
    function(f) readLines(f, warn = FALSE))), collapse = "\n")

  expect_false(grepl("plotly",
                     gsub("disable_plotly", "", html_contents, fixed = TRUE),
                     fixed = TRUE))
})

test_that("also_print writes html files and returns a result list", {
  skip_on_cran() # slow

  skip_if_not_installed("DT")
  study_data <- data.frame(id = 1:4,
                           CENTER = c("A", "A", "B", "B"),
                           val1 = c(1, NA, 3, 4),
                           val2 = c(5, 6, NA, 8)
  )
  meta_data <- data.frame(
    VAR_NAMES = c("val1", "val2", "CENTER"),
    LABEL = c("Variable 1", "Variable 2", "Center"),
    DATA_TYPE = c("integer", "integer", "string"),
    MISSING_LIST = c("", "", ""),
    VALUE_LABELS = c(NA, NA, "A|B"),
    stringsAsFactors = FALSE
  )
  td <- tempfile()
  on.exit(unlink(td, recursive = TRUE), add = TRUE)
  result <- dq_report_by(study_data, meta_data, label_col = "LABEL",
                         strata_column = "CENTER",
                         output_dir = td, also_print = TRUE, segment_column = NULL, cores = NULL)
  expect_type(result, "list")
  html_files <- list.files(td, pattern = "\\.html$")
  expect_gt(length(html_files), 0)
})

test_that("missing label_col is handled gracefully", {
  skip_on_cran() # slow

  skip_if_not_installed("stringdist")
  study_data <- data.frame(id = 1:4,
                           CENTER = c("A", "A", "B", "B"),
                           val1 = c(1, NA, 3, 4),
                           val2 = c(5, 6, NA, 8)
  )
  meta_data <- data.frame(
    VAR_NAMES = c("val1", "val2", "CENTER"),
    DATA_TYPE = c("integer", "integer", "string"),
    MISSING_LIST = c("", "", ""),
    VALUE_LABELS = c(NA, NA, "A|B"),
    stringsAsFactors = FALSE
  )

  result <- dq_report_by(
    study_data, meta_data,
    label_col = "LABEL",
    strata_column = "CENTER",
    segment_column = NULL,
    cores = NULL
  )

  expect_type(result, "list")
})

#' Utility function to interpret mathematical interval notation
#'
#' Utility function to split limit definitions into interpretable elements
#'
#' @param mdata [data.frame] the data frame that contains metadata
#'                               attributes of study data
#'
#' @importFrom utils head
#'
#' @return augments metadata by interpretable limit columns
#'
#' @seealso `util_validate_known_meta()`
#' @family parser_functions
#' @concept robustness
#' @noRd
util_interpret_limits <- function(mdata) { # TODO: Use the redcap parser, instead!!

  report_generation_time <- as.character(Sys.time())

  # grep all columns of data with the notation of LIMITS
  lv <- colnames(mdata[grep("LIMIT", colnames(mdata))])

  if (length(lv) == 0) {
    util_error("No column containing the term LIMIT.",
               applicability_problem = TRUE)
  }

  # all not empty?
  ne <- apply(mdata[, lv, drop = FALSE], 2, function(x) all(is.na(x)))

  if (any(ne)) {
    util_message(paste0("The column: ", lv[ne],
                        " has no defined intervals and is omitted."),
                 applicability_problem = TRUE)
  }

  # don't consider empty columns at all
  lv <- lv[!(ne)]

  if (length(lv) == 0) {
    return(mdata)
  }

  # prefix/type of LIMITS
  pv <- lapply(strsplit(lv, split = "_", fixed = TRUE), function(syllables) {
    paste0(syllables[!grepl("LIMIT", syllables)], collapse = "_")
  })

  # result
  mdata_ext <- mdata

  # the code below works only on non-empty elements of the column
  # the number of elements will most likely vary between columns
  for (i in seq_along(lv)) {
    valid1 <- grepl(
      "[\\[(]([0-9\\.Ee+\\-]*|[\\+\\-]?Inf);([0-9\\.Ee+\\-]*|[\\+\\-]?Inf)[\\])]",
      gsub(" ", "", mdata[[lv[i]]], fixed = TRUE), perl = TRUE)
    valid2 <- vapply(mdata[[lv[i]]], FUN.VALUE = logical(1),
                     function(x) {
                       if (is.na(x) || trimws(x) == "") {
                         return(FALSE)
                       }
                       x <- gsub("^[(\\[]+", "", x, perl = TRUE)
                       x <- gsub("[)\\]]+$", "", x, perl = TRUE)
                       x <- paste("", x, "")
                       xs <- trimws(strsplit(x, ";", fixed = TRUE)[[1]])
                       if (length(xs) != 2) {
                         return(FALSE)
                       }
                       xs1 <- xs[[1]]
                       xs2 <- xs[[2]]
                       if (xs1 %in% c("", "Inf", "+Inf", "-Inf", "today")) {
                         a <- TRUE
                       } else {
                         a <- !inherits(try(suppressWarnings(util_parse_date(xs1,
                                                                             optional = FALSE)),
                                            silent = TRUE),
                                        "try-error")
                         if (!a) {
                           # also accept time-only (hms) limits
                           a <- !inherits(try(suppressWarnings(util_parse_time(xs1,
                                                                               optional = FALSE)),
                                              silent = TRUE),
                                          "try-error")
                         }
                       }
                       if (xs2 %in% c("", "Inf", "+Inf", "-Inf", "today")) {
                         b <- TRUE
                       } else {
                         b <- !inherits(try(suppressWarnings(util_parse_date(xs2,
                                                                             optional = FALSE)),
                                            silent = TRUE),
                                        "try-error")
                         if (!b) {
                           # also accept time-only (hms) limits
                           b <- !inherits(try(suppressWarnings(util_parse_time(xs2,
                                                                               optional = FALSE)),
                                              silent = TRUE),
                                          "try-error")
                         }
                       }
                       a && b
                     })

    valid <- valid1 | valid2

    if (any(!valid & !util_empty(mdata[[lv[i]]]))) { # TODO: Would we allow [1; 12[ ??
      util_warning(
        "Found invalid limits for %s: %s%s -- will ignore these",
        sQuote(lv[i]),
        paste(head(dQuote(mdata[[lv[i]]][!valid & !is.na(mdata[[lv[i]]])]), 5),
              collapse = ", "),
        (ifelse(sum(!valid & !is.na(mdata[[lv[i]]])) > 5, ", ...", "")),
        applicability_problem = TRUE
      )
    }

    # select rows with entries in limits
    X <- mdata[valid & !(is.na(mdata[lv[i]])), c(VAR_NAMES, DATA_TYPE, lv[i])]

    # Split limits
    myfun1 <- function(x) {
      trimws(unlist(strsplit(as.character(x), split = ";")), "both")
    }

    # add to dataframe
    if (any(valid)) {
      X <- cbind.data.frame(X, (t(apply(as.data.frame(X[[lv[i]]]), 1, myfun1))))

      # extract values
      X_LOWER <- (gsub(X[, "1"],
                       replacement = "",
                       pattern = "[\\(|\\[]",
                       perl = TRUE
      ))
      X_UPPER <- (gsub(X[, "2"],
                       replacement = "",
                       pattern = "[\\)|\\]]",
                       perl = TRUE
      ))
      ## date
      date_vars <- X[[DATA_TYPE]] == DATA_TYPES$DATETIME
      date_vars[is.na(date_vars)] <- FALSE

      # --- detect time-only (hms) style limits (heuristic if no explicit TIME type) ---
      has_colon_lower <- grepl(":", X_LOWER, fixed = TRUE)
      has_colon_upper <- grepl(":", X_UPPER, fixed = TRUE)
      looks_like_date_lower <- grepl("\\b\\d{4}[-/]\\d{2}[-/]\\d{2}\\b", X_LOWER)
      looks_like_date_upper <- grepl("\\b\\d{4}[-/]\\d{2}[-/]\\d{2}\\b", X_UPPER)
      time_vars <- (!date_vars) & ((has_colon_lower | has_colon_upper) &
                                     !(looks_like_date_lower | looks_like_date_upper))
      time_vars[is.na(time_vars)] <- FALSE
      # If you have DATA_TYPES$TIME, you could also OR it in:
      # time_vars <- (X[[DATA_TYPE]] == DATA_TYPES$TIME) | time_vars

      IS_LOWER_INF <- grepl(perl = TRUE, ignore.case = TRUE,
                            "^\\s*[\\+\\-]?\\s*Inf\\s*$", X_LOWER[date_vars])
      IS_UPPER_INF <- grepl(perl = TRUE, ignore.case = TRUE,
                            "^\\s*[\\+\\-]?\\s*Inf\\s*$", X_UPPER[date_vars])
      VAL_LOWER_INF <- X_LOWER[date_vars][IS_LOWER_INF]
      VAL_UPPER_INF <- X_UPPER[date_vars][IS_UPPER_INF]

      IS_LOWER_NOW <- grepl(perl = TRUE, ignore.case = TRUE,
                            "^\\s*today\\s*$", X_LOWER[date_vars])
      IS_UPPER_NOW <- grepl(perl = TRUE, ignore.case = TRUE,
                            "^\\s*today\\s*$", X_UPPER[date_vars])

      X_LOWER[date_vars][IS_LOWER_NOW] <- report_generation_time
      X_UPPER[date_vars][IS_UPPER_NOW] <- report_generation_time
      X_LOWER[date_vars][IS_LOWER_INF] <- NA
      X_UPPER[date_vars][IS_UPPER_INF] <- NA
      X_LOWER[date_vars][trimws(X_LOWER[date_vars]) == ""] <- NA
      X_UPPER[date_vars][trimws(X_UPPER[date_vars]) == ""] <- NA

      # extract values
      X$LOWER[date_vars] <- suppressWarnings(
        as.numeric(util_parse_date(X_LOWER[date_vars], optional = TRUE)))
      X$UPPER[date_vars] <- suppressWarnings(
        as.numeric(util_parse_date(X_UPPER[date_vars], optional = TRUE)))

      X$LOWER[date_vars][IS_LOWER_INF] <- suppressWarnings(
        as.numeric(VAL_LOWER_INF))
      X$UPPER[date_vars][IS_UPPER_INF] <- suppressWarnings(
        as.numeric(VAL_UPPER_INF))

      # --- time-only (hms) limits: interpret as seconds since midnight ---
      if (any(time_vars)) {
        IS_LOWER_INF_T <- grepl(perl = TRUE, ignore.case = TRUE,
                                "^\\s*[\\+\\-]?\\s*Inf\\s*$", X_LOWER[time_vars])
        IS_UPPER_INF_T <- grepl(perl = TRUE, ignore.case = TRUE,
                                "^\\s*[\\+\\-]?\\s*Inf\\s*$", X_UPPER[time_vars])
        IS_LOWER_NOW_T <- grepl(perl = TRUE, ignore.case = TRUE,
                                "^\\s*today\\s*$", X_LOWER[time_vars])
        IS_UPPER_NOW_T <- grepl(perl = TRUE, ignore.case = TRUE,
                                "^\\s*today\\s*$", X_UPPER[time_vars])

        # normalize empty/Inf/today to NA for time-only
        X_LOWER[time_vars][trimws(X_LOWER[time_vars]) == ""] <- NA
        X_UPPER[time_vars][trimws(X_UPPER[time_vars]) == ""] <- NA
        X_LOWER[time_vars][IS_LOWER_INF_T | IS_LOWER_NOW_T] <- NA
        X_UPPER[time_vars][IS_UPPER_INF_T | IS_UPPER_NOW_T] <- NA

        lower_t <- suppressWarnings(util_parse_time(X_LOWER[time_vars], optional = TRUE))
        upper_t <- suppressWarnings(util_parse_time(X_UPPER[time_vars], optional = TRUE))

        X$LOWER[time_vars] <- suppressWarnings(as.numeric(lower_t))
        X$UPPER[time_vars] <- suppressWarnings(as.numeric(upper_t))
      }

      # numeric (everything else)
      X$LOWER[!date_vars & !time_vars] <- suppressWarnings(as.numeric(X_LOWER[!date_vars & !time_vars]))
      X$UPPER[!date_vars & !time_vars] <- suppressWarnings(as.numeric(X_UPPER[!date_vars & !time_vars]))

      damaged_lower_limit <- ((!is.na(X_LOWER) & X_LOWER != "") &
                                is.na(X$LOWER))
      if (any(damaged_lower_limit)) {
        field <- sQuote(lv[[i]])
        defects <- paste(head(dQuote(X_LOWER[damaged_lower_limit]), 5), "in",
                         head(sQuote(X$VAR_NAMES[damaged_lower_limit]), 5),
                         collapse = ", ")
        dots <- ifelse(sum(damaged_lower_limit) > 5, ", ...", "")
        util_warning(
          "Damaged lower %s: %s%s",
          field,
          defects,
          dots,
          applicability_problem = TRUE
        )
      }

      damaged_upper_limit <- ((!is.na(X_UPPER) & X_UPPER != "") &
                                is.na(X$UPPER))
      if (any(damaged_upper_limit)) {
        field <- sQuote(lv[[i]])
        defects <- paste(head(dQuote(X_UPPER[damaged_upper_limit]), 5), "in",
                         head(sQuote(X$VAR_NAMES[damaged_upper_limit]), 5),
                         collapse = ", ")
        dots <- ifelse(sum(damaged_upper_limit) > 5, ", ...", "")
        util_warning(
          "Damaged upper %s: %s%s",
          field,
          defects,
          dots,
          applicability_problem = TRUE
        )
      }

      damaged_upper_limit <- ((!is.na(X_UPPER) & X_UPPER != "") &
                                is.na(X$UPPER))

      # identify brace type
      p1 <- "("
      p2 <- ")"

      # interpret brace to logical
      X$INCL_LOWER <- ifelse(grepl(p1, X[, "1"], fixed = TRUE), FALSE, TRUE)
      X$INCL_UPPER <- ifelse(grepl(p2, X[, "2"], fixed = TRUE), FALSE, TRUE)

      X <- X[, c("VAR_NAMES", "LOWER", "UPPER", "INCL_LOWER", "INCL_UPPER")]

      colnames(X) <- c(
        "VAR_NAMES",
        paste0(pv[i], "_LIMIT_LOW"),
        paste0(pv[i], "_LIMIT_UP"),
        paste0("INCL_", paste0(pv[i], "_LIMIT_LOW")),
        paste0("INCL_", paste0(pv[i], "_LIMIT_UP"))
      )
      # --> for dates, use as.POSIXct(X$UPPER, origin = min(Sys.time(), 0))
      #     to translate back.

      mdata_ext <- merge(mdata_ext, X, by = "VAR_NAMES", all.x = TRUE)
    }
  }

  return(mdata_ext)
}

#' Get all `PART_VARS` for a response variable (from item-level metadata)
#'
#' @param rv [character] the response variable's name
#' @param study_data [study_data]
#' @param meta_data [meta_data]
#' @param label_col [character] the metadata attribute to map `meta_data` on
#'                              `study_data` based on `colnames(study_data)`
#' @param expected_observations [enum] HIERARCHY | ALL | SEGMENT. How should
#'                                     `PART_VARS` be handled:
#'                                     - `ALL`: Ignore, all observations are
#'                                       expected
#'                                     - `SEGMENT`: if `PART_VAR` is 1, an
#'                                       observation is expected
#'                                     - `HIERARCHY`: the default, if the
#'                                       `PART_VAR` is 1 for this variable and
#'                                       also for all `PART_VARS` of `PART_VARS`
#'                                       up in the hierarchy, an observation is
#'                                       expected.
#'
#' @return [character] all `PART_VARS` for `rv` from item level metadata.
#'    For `expected_observations = HIERARCHY`, the more general `PART_VARS`
#'    (i.e., up, in the hierarchy) are more left in the vector, e.g.:
#'    `PART_STUDY, PART_PHYSICAL_EXAMINATIONS, PART_BLOODPRESSURE`
#'
#' @family missing_functions
#' @concept missing
#' @noRd
util_all_intro_vars_for_rv <- function(rv, study_data, meta_data,
                                       label_col = LABEL,
                                       expected_observations =
                                         c("HIERARCHY",
                                           "ALL",
                                           "SEGMENT")) {
  r <- .util_all_intro_vars_for_rv(rv, study_data, meta_data, label_col,
                                   expected_observations)
  util_map_labels(r, meta_data, to = label_col, ifnotfound = r)

}

.util_all_intro_vars_for_rv <-
  function(rv, study_data, meta_data, label_col = LABEL,
           expected_observations =
             c("HIERARCHY",
               "ALL",
               "SEGMENT")) {

    meta_data <- prep_meta_data_v1_to_item_level_meta_data(meta_data,
                                                           verbose = FALSE,
                                                           label_col =
                                                             label_col)

    util_expect_scalar(expected_observations, allow_more_than_one = TRUE,
                       check_type = is.character)
    expected_observations <- match.arg(expected_observations)
    util_expect_scalar(expected_observations)

    if (expected_observations == "ALL" ||
        !(PART_VAR %in% names(meta_data))) {
      return(NULL)
    }
    if (label_col != VAR_NAMES) {
      rv <-
        prep_map_labels(
          rv,
          from = label_col,
          to = VAR_NAMES,
          meta_data =  meta_data
        )
    }
    kss <-
      prep_map_labels(
        rv,
        from = VAR_NAMES,
        to = PART_VAR,
        meta_data =  meta_data
      )
    rv <- unname(rv)
    kss <- unname(kss)
    if (identical(rv, kss)) {
      return(NULL)
    }
    if (expected_observations == "HIERARCHY" &&
        kss %in% meta_data$VAR_NAMES) {
      return(c(Recall(kss, study_data, meta_data, VAR_NAMES), kss))
    }
    if (expected_observations != "ALL") {
      var_kss <- kss %in% meta_data[[VAR_NAMES]]
      if (any(!var_kss)) {
        util_warning(c("For %s = %s, %s in %s must contain names of",
                       "study variables inidcating participation in a",
                       "study segment. %s is/are not in %s in %s."),
                     sQuote("expected_observations"),
                     dQuote(expected_observations),
                     dQuote(PART_VAR),
                     sQuote("meta_data"),
                     paste(dQuote(kss[!var_kss]), collapse = ", "),
                     dQuote(VAR_NAMES),
                     sQuote("meta_data")
                     )
      }
    }
    intersect(kss, meta_data[[VAR_NAMES]])
  }

#' Plots and checks for distributions for categorical variables
#'
#' @description
#' This function creates distribution plots for categorical variables.
#'
#' [Descriptor]
#'
#' @details
#' To complete
#'
#' @export
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable] the name of the measurement variable
#' @param group_vars [variable] the name of the observer, device or
#'                                   reader variable
#' @param n_cat_max maximum number of categories to be displayed individually
#'                  for the categorical variable (`resp_vars`)
#' @param n_group_max maximum number of categories to be displayed individually
#'                  for the grouping variable (`group_vars`, devices / examiners)
#' @param n_data_min minimum number of data points to create a time course plot
#'                   for an individual category of the `resp_vars` variable
#'
#' @return A [list] with:
#'   - `SummaryPlot`: [ggplot2::ggplot] for the response variable in
#'                    `resp_vars`.
#'
#' @seealso
#' [Online Documentation](https://dataquality.qihs.uni-greifswald.de/)
acc_cat_distributions <- function(resp_vars = NULL,
                                  group_vars = NULL,
                                  study_data,
                                  label_col,
                                  item_level = "item_level",
                                  meta_data = item_level,
                                  meta_data_v2,
                                  n_cat_max =
                                    getOption("dataquieR.max_cat_resp_var_levels_in_plot",
                                              dataquieR.max_cat_resp_var_levels_in_plot_default),
                                  n_group_max =
                                    getOption("dataquieR.max_group_var_levels_in_plot",
                                              dataquieR.max_group_var_levels_in_plot_default),
                                  n_data_min =
                                    getOption("dataquieR.min_time_points_for_cat_resp_var",
                                              dataquieR.min_time_points_for_cat_resp_var_default)) {
  # preps ----------------------------------------------------------------------
  # map metadata to study data
  util_maybe_load_meta_data_v2()
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata_inadm = TRUE)

  # If no response variable is defined, all suitable variables will be selected.
  if (length(resp_vars) == 0) {
    util_message(
      c("All variables defined to be nominal and ordinal in the metadata are used",
        "by acc_cat_distribution."),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE)
    resp_vars <- meta_data[[label_col]][meta_data[[SCALE_LEVEL]] %in%
                                          c(SCALE_LEVELS$NOMINAL,
                                            SCALE_LEVELS$ORDINAL)]
    resp_vars <- intersect(resp_vars, colnames(ds1))
    if (length(resp_vars) == 0) {
      util_error("No suitable variables were defined for acc_cat_distributions.",
                 applicability_problem = TRUE)  #TODO: maybe intrinsic
    }
  }

  # set up grouping variables, if needed
  if (all(is.na(group_vars)) ||
      length(group_vars) == 0) {
    group_vars <- NULL
  } else {
    util_correct_variable_use(group_vars,
                              allow_null = TRUE,
                              allow_more_than_one = FALSE,
                              allow_any_obs_na = TRUE,
                              allow_all_obs_na = FALSE,
                              need_type = "!float",
                              need_scale = "nominal | ordinal"
    )

    if (length(group_vars) > 0) {
      # The grouping variable(s) should not be included as response variable(s).
      if (any(group_vars %in% resp_vars)) {
        resp_vars <- resp_vars[-which(resp_vars %in% group_vars)]
        util_warning(paste("Removed grouping variable from response variables",
                           "for acc_cat_distributions."),
                     applicability_problem = TRUE)
      }
      if (length(resp_vars) == 0) {

        my_message <- util_error

        if (.called_in_pipeline) {
          my_message <- util_error
        } else {
          my_message <- util_warning
        }

        my_message("No variables left to analyse for acc_cat_distributions.",
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
        x <- util_create_lean_ggplot(
          ggplot() +
            annotate("text", x = 0, y = 0, label = "Empty result.") +
            theme(
              axis.line = element_blank(),
              axis.text.x = element_blank(),
              axis.text.y = element_blank(),
              axis.ticks = element_blank(),
              axis.title.x = element_blank(),
              axis.title.y = element_blank(),
              legend.position = "none",
              panel.background = element_blank(),
              panel.border = element_blank(),
              panel.grid.major = element_blank(),
              panel.grid.minor = element_blank(),
              plot.background = element_blank()
            )
        )
        x <- util_set_size(x)

        return(list(SummaryPlot = x))
      }
    }
  }

  util_correct_variable_use(resp_vars,
                            allow_more_than_one = FALSE,
                            allow_any_obs_na = TRUE,
                            allow_all_obs_na = FALSE,
                            min_distinct_values = 2,
                            need_scale = "nominal|ordinal"
  )

  # create the plot
  p <- util_plot_categorical_vars(resp_vars = resp_vars,
                                  group_vars = group_vars,
                                  study_data = study_data,
                                  meta_data = meta_data,
                                  n_cat_max = n_cat_max,
                                  n_group_max = n_group_max,
                                  n_data_min = n_data_min)

  rvlab <- prep_get_labels(resp_vars,
                           item_level = meta_data,
                           label_col = label_col,
                           resp_vars_match_label_col_only = TRUE,
                           label_class = "LONG")

  gvlab <- prep_get_labels(group_vars,
                           item_level = meta_data,
                           label_col = label_col,
                           resp_vars_match_label_col_only = TRUE,
                           label_class = "LONG")

  as_plotly <- attr(p, "as_plotly")

  # p <- p + ggplot2::ggtitle(rvlab,
  #                           sprintf("Effects of %s",
  #                                   gvlab)
  # )
  p <- p + util_create_lean_ggplot(
    ylab(gvlab), # + xlab(rvlab)
    gvlab = gvlab
  )
  #  p <- util_set_size(p)

  attr(p, "as_plotly") <- NULL
  dont_util_adjust_geom_text_for_plotly <-
    attr(p, "dont_util_adjust_geom_text_for_plotly")
  attr(p, "dont_util_adjust_geom_text_for_plotly") <- NULL

  return(util_attach_attr(list(SummaryPlot = p),
                          as_plotly = as_plotly,
                          dont_util_adjust_geom_text_for_plotly =
                            dont_util_adjust_geom_text_for_plotly))
}

#' `dataquieR` version of match.arg
#'
#' does not support partial matching, but will display the most likely
#' match as a warning/error.
#'
#' @param arg the argument
#' @param choices the choices
#' @param several_ok allow more than one entry in `arg`
#' @param error `stop()`, if `arg` is not in `choices`
#'              (warns and cleans `arg`, otherwise)
#'
#' @return "cleaned" `arg`
#'
#' @family robustness_functions
#' @concept string
#' @noRd
util_match_arg <- function(arg, choices, several_ok = FALSE, error = TRUE) {
#TODO: util_expect_scalar check_type = as.character
  if (missing(arg)) {
    util_error("%s needs the argument %s",
               sQuote("util_match_arg"),
               sQuote("arg"))
  }

  arg_name <- util_deparse1(substitute(arg))

  calling_fkt <- rlang::caller_fn()
  fkt_name <- "<unknown function>"
  try({
    fkt_name <- as.character(rlang::caller_call()[[1]])
  }, silent = TRUE)

  if (missing(choices)) {
    formal_args <-formals(calling_fkt)
    choices1 <- eval(
      formal_args[[arg_name]],
      envir = parent.frame()
    )
  } else {
    choices1 <- choices
  }

  choices1 <- util_ensure_character(
    choices1,
    error = TRUE,
    error_msg = c("For argument %s of function %s,",
                  "not all choices passed to %s could be interpreted as character."),
    sQuote(arg_name),
    sQuote(fkt_name),
    sQuote("util_match_arg")
  )

  choices1 <- unique(choices1)

  util_expect_scalar(
    choices1,
    allow_null = TRUE,
    allow_more_than_one = TRUE)

  if (length(choices1) < 1) {
    util_error(
      "the function %s does not provide any choice for its argument %s",
      sQuote(fkt_name), sQuote(arg_name))
  }

  if ((eval.parent(call("missing", arg_name)) ||
       !(arg_name %in% names(rlang::call_match(rlang::caller_call(),
                                               rlang::caller_fn())))) &&
      missing(choices) &&
      all(arg == choices1) &&
      length(arg) != 1 &&
      !several_ok) {
    arg <- NULL
  }

  arg1 <- util_ensure_character(
    arg,
    error = TRUE,
    error_msg = c("For argument %s of function %s,",
                  "not all choices passed to %s could be interpreted as character."),
    sQuote(arg_name),
    sQuote(fkt_name),
    sQuote("util_match_arg")
  )

  e <- new.env(parent = environment())

  assign(arg_name, arg1, e)

  cl <- call("util_expect_scalar", as.symbol(arg_name),
             allow_more_than_one = TRUE,
             allow_null = TRUE,
             allow_na = TRUE,
             check_type = is.character)

  eval(cl, envir = e)

  if (length(arg1) == 0) {
    arg1 <- choices1[[1]]
  }

  if (!several_ok && length(arg1) != 1) {
    util_error("the function %s needs exactly one entry in %s",
               sQuote(fkt_name), sQuote(arg_name))
  }

  util_ensure_in(
    arg1,
    choices1,
    err_msg =
      sprintf(
        paste("argument %s of function %s should be one of %s.",
              "%%s is not an allowed value, did you mean %%s?"),
        sQuote(arg_name),
        sQuote(fkt_name),
        util_pretty_vector_string(choices1)
    ),
    error = error)
}

#' Utility function for smoothed longitudinal trends from logistic regression models
#'
#' This function is under development. It computes a logistic regression for
#' binary variables and visualizes smoothed time trends of the residuals by
#' LOESS or GAM. The function can also be called for non-binary outcome
#' variables. These will be transformed to binary variables, either using
#' user-specified groups in the metadata columns `RECODE_CASES` and/or
#' `RECODE_CONTROL` (see `util_dichotomize`), or it will attempt to recode the
#' variables automatically. For nominal variables, it will consider the most
#' frequent category as 'cases' and every other category as 'control', if there
#' are more than two categories. Nominal variables with only two distinct values
#' will be transformed by assigning the less frequent category to 'cases' and
#' the more frequent category to 'control'. For variables of other statistical
#' data types, values inside the interquartile range are considered as
#' 'control', values outside this range as 'cases'. Variables with few
#' different values are transformed in a simplified way to obtain two groups.
#'
#' [Descriptor]
#'
#' @param resp_vars  [variable] the name of the (binary) measurement variable
#' @param group_vars [variable] the name of the observer, device or
#'                              reader variable
#' @param co_vars [variable list] a vector of co-variables, e.g. age and sex for
#'                              adjustment
#' @param time_vars  [variable] the name of the variable giving the time
#'                              of measurement
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param meta_data [data.frame] the data frame that contains metadata
#'                              attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                              with labels of variables
#' @param min_obs_in_subgroup [integer] from=0. This optional argument specifies
#'                       the minimum number of observations that is required to
#'                       include a subgroup (level) of the `group_var` in the
#'                       analysis. Subgroups with less observations are
#'                       excluded.
#' @param resolution [integer] the maximum number of time points used for
#'                             plotting the trend lines
#' @param plot_format [enum] AUTO | COMBINED | FACETS | BOTH. Return the plot
#'                             as one combined plot for all groups or as
#'                             facet plots (one figure per group). `BOTH` will
#'                             return both variants, `AUTO` will decide based
#'                             on the number of observers.
#' @param n_group_max [integer] maximum number of categories to be displayed
#'                  individually for the grouping variable (`group_vars`,
#'                  devices / examiners)
#' @param enable_GAM [logical] Can LOESS computations be replaced by general
#'                  additive models to reduce memory consumption  for large
#'                  datasets?
#' @param exclude_constant_subgroups [logical] Should subgroups with constant
#'                  values be excluded?
#' @param min_bandwidth [numeric] lower limit for the LOESS bandwidth, should be
#'                  greater than 0 and less than or equal to 1. In general,
#'                  increasing the bandwidth leads to a smoother trend line.
#' @param min_proportion [numeric] lower limit for the proportion of the smaller
#'                  group (cases or controls) for creating a LOESS figure,
#'                  should be greater than 0 and less than 0.4.
#'
#' @return a [list] with:
#'   - `SummaryPlotList`: a plot.
#'
#' @importFrom ggplot2 ggplot aes scale_color_manual xlab ylab geom_line facet_wrap theme_minimal ggtitle theme element_blank expand_limits
#' @importFrom stats as.formula lm loess predict na.omit glm binomial poisson sd cov var runif
#' @noRd
util_acc_loess_bin <- function(
    resp_vars,
    label_col = NULL,
    study_data,
    item_level = "item_level",
    group_vars = NULL,
    time_vars,
    co_vars = NULL,
    min_obs_in_subgroup =
      getOption("dataquieR.acc_loess.min_obs_in_subgroup",
                dataquieR.acc_loess.min_obs_in_subgroup_default),
    resolution = 80,
    plot_format = getOption("dataquieR.acc_loess.plot_format",
                            dataquieR.acc_loess.plot_format_default),
    meta_data = item_level,
    n_group_max = getOption("dataquieR.max_group_var_levels_in_plot",
                            dataquieR.max_group_var_levels_in_plot_default),
    enable_GAM = getOption("dataquieR.GAM_for_LOESS",
                           dataquieR.GAM_for_LOESS_default),
    exclude_constant_subgroups =
      getOption("dataquieR.acc_loess.exclude_constant_subgroups",
                dataquieR.acc_loess.exclude_constant_subgroups_default),
    min_bandwidth = getOption("dataquieR.acc_loess.min_bw",
                              dataquieR.acc_loess.min_bw_default),
    min_proportion = getOption("dataquieR.acc_loess.min_proportion",
                               dataquieR.acc_loess.min_proportion_default)) {
  # preps ----------------------------------------------------------------------
  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata = TRUE)

  # correct variable use?
  # (checked before, but included here to catch implementation errors)
  util_correct_variable_use("resp_vars",
                            need_scale = "!na",
                            allow_all_obs_na = FALSE)
  util_correct_variable_use("group_vars",
                            need_scale = "nominal | ordinal",
                            allow_all_obs_na = TRUE,
                            allow_na = TRUE,
                            allow_null = TRUE)
  util_correct_variable_use("time_vars",
                            need_type = DATA_TYPES$DATETIME,
                            need_scale = "interval | ratio",
                            allow_all_obs_na = FALSE,
                            min_distinct_values = 3)
  util_correct_variable_use("co_vars",
                            overwrite = TRUE,
                            remove_not_found = TRUE,
                            allow_more_than_one = TRUE,
                            allow_all_obs_na = FALSE,
                            allow_na = TRUE,
                            allow_null = TRUE)

  # support time course plots without (sub-)groups
  if (is.null(group_vars) || all(util_empty(group_vars))) {
    # create a dummy grouping variable that is not yet contained in ds1
    group_vars <- "dummy_group"
    while (group_vars %in% colnames(ds1)) {
      group_vars <- paste0("dummy_group",
                           ceiling(runif(n = 1, min = 1, max = ncol(ds1) * 2)),
                           sep = "_")
    }
    ds1[[group_vars]] <- 1
    plot_title <- paste("Time course plot for", resp_vars)
    # The dummy variable should not be mentioned in the title of the plot.
  } else {
    plot_title <- paste("Effects of", group_vars, "in", resp_vars)
  }

  if (is.null(co_vars)) {
    co_vars <- character(0)
  }
  co_vars <- na.omit(co_vars)

  n_lev_gv_total <- length(na.omit(unique(ds1[[group_vars]])))

  # omit missing values and unnecessary variables
  n_prior <- nrow(ds1)
  ds1 <- ds1[, c(resp_vars, time_vars, group_vars, co_vars)]
  if (grepl("dummy_group", group_vars)) {
    # Only mention the 'dummy_group' in the message if it contributes any
    # missing values, otherwise do not mention it.
    if (any(is.na(ds1[complete.cases(ds1[, c(time_vars, co_vars)]),
                      group_vars]))) {
      msg_part1 <- paste0(c(group_vars, co_vars), collapse = ", ")
    } else {
      msg_part1 <- paste0(co_vars, collapse = ", ")
    }
  } else {
    msg_part1 <- paste0(c(group_vars, co_vars), collapse = ", ")
  }
  ds1 <- ds1[complete.cases(ds1[, c(time_vars, group_vars, co_vars)]), ]
  n_post <- nrow(ds1)
  msg <- NULL
  if (n_post < n_prior) {
    msg <- paste0(
      "Due to missing values in ",
      ifelse(nchar(msg_part1) > 0, paste0(msg_part1, " or "), ""),
      time_vars, ", N = ", n_prior - n_post,
      " observations were excluded. "
    )
  }
  n_prior <- n_post
  ds1 <- ds1[complete.cases(ds1), ]
  n_post <- nrow(ds1)
  if (n_post < n_prior) {
    msg <- paste0(
      msg, "Due to missing values in ", resp_vars, ", N = ",
      n_prior - n_post, " observations were excluded",
      ifelse(nchar(msg) > 0, " additionally.", "."))
  }
  if (!is.null(msg) && nchar(msg) > 0) {
    util_message(trimws(msg),
                 applicability_problem = FALSE)
  }

  # dichotomize resp_vars
  if (!("RECODE_CASES" %in% colnames(meta_data))) {
    meta_data[[RECODE_CASES]] <- ""
  }
  if (!("RECODE_CONTROL" %in% colnames(meta_data))) {
    meta_data[[RECODE_CONTROL]] <- ""
  }
  var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])

  if (var_prop$NDistinct == 1) {
    util_error("The response variable is constant after data preparation.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  } else {
    if (util_empty(meta_data[[RECODE_CASES]][meta_data[[label_col]] == resp_vars]) &
        util_empty(meta_data[[RECODE_CONTROL]][meta_data[[label_col]] == resp_vars])) {
      # => the recoding is not defined in the metadata. dataquieR will try to
      # define a helpful categorization into two groups:
      if (meta_data[[SCALE_LEVEL]][meta_data[[label_col]] == resp_vars] ==
          SCALE_LEVELS$NOMINAL) {
        count_nom <- util_table_of_vct(ds1[[resp_vars]])

        if (var_prop$NDistinct == 2) {
          # less frequent category as 'cases'
          count_nom <- count_nom[which(count_nom[, 2] > 0), ]
          count_nom <- count_nom[order(count_nom[, 2], decreasing = FALSE), ]
          meta_data[[RECODE_CASES]][meta_data[[label_col]] == resp_vars] <-
            paste(count_nom[1, 1])
        } else {
          # more than two groups: dataquieR will use the most frequent category
          # as 'cases', the remaining categories as 'control'.
          count_nom <- count_nom[order(count_nom[, 2], decreasing = TRUE), ]
          meta_data[[RECODE_CASES]][meta_data[[label_col]] == resp_vars] <-
            paste(count_nom[1, 1])
        }
      } else {
        # Automatic recoding will be based on quartiles
        qu_rvs <- quantile(as.numeric(ds1[[resp_vars]]), type = 1)
        val_lab_applied <- is.factor(ds1[[resp_vars]])
        if (length(unique(qu_rvs)) < 3) {
          # the value that occurs predominantly will become the 'control'
          tab_qu <- util_table_of_vct(qu_rvs)
          tab_qu <- tab_qu[order(tab_qu[, 2], decreasing = TRUE), ]
          if (val_lab_applied) {
            meta_data[[RECODE_CONTROL]][meta_data[[label_col]] == resp_vars] <-
              as.character(tab_qu[1,1])
          } else {
            meta_data[[RECODE_CONTROL]][meta_data[[label_col]] == resp_vars] <-
              paste0("[", tab_qu[1,1], ";", tab_qu[1,1], "]")
          }
        } else if (var_prop$NDistinct < 10) {
          # If there are few values (more than two), consider the
          # upper quartile as 'cases'.
          if (val_lab_applied) {
            meta_data[[RECODE_CASES]][meta_data[[label_col]] == resp_vars] <-
              paste(levels(ds1[[resp_vars]])[qu_rvs[4]:qu_rvs[5]],
                    collapse = " | ")
          } else {
            meta_data[[RECODE_CASES]][meta_data[[label_col]] == resp_vars] <-
              paste0("[", qu_rvs[4], ";", qu_rvs[5], "]")
          }
        } else {
          # Values below Q1 and/or above Q3 as ('extreme') cases,
          # values between Q1 and Q3 (i.e., central 50% of values) as control
          if (val_lab_applied) {
            meta_data[[RECODE_CONTROL]][meta_data[[label_col]] == resp_vars] <-
              paste(levels(ds1[[resp_vars]])[qu_rvs[2]:qu_rvs[4]],
                    collapse = " | ")
          } else {
            meta_data[[RECODE_CONTROL]][meta_data[[label_col]] == resp_vars] <-
              paste0("[", qu_rvs[2], ";", qu_rvs[4], "]")
          }
        }
      }
    }
  }

  rvs_bin <- util_dichotomize(
    study_data = ds1[, resp_vars, drop = FALSE],
    meta_data = meta_data,
    label_col = label_col)
  rvs_bin_note <- attr(rvs_bin, "Dichotomization")[[resp_vars]]
  ds1[[resp_vars]] <- unlist(rvs_bin)
  ds1 <- ds1[complete.cases(ds1), ]

  # convert group_vars to factor (needed for example for the dummy group)
  if (!is.factor(ds1[[group_vars]])) {
    ds1[[group_vars]] <- factor(ds1[[group_vars]])
  }

  # TODO: use util_check_group_levels
  # too few observations per level?
  # check which groups do not have enough observations or time points
  tab_groups <- table(ds1[[group_vars]])
  groups_below_min_obs <- names(tab_groups)[tab_groups < min_obs_in_subgroup]
  tab_groups_tp <- vapply(levels(ds1[[group_vars]]), FUN.VALUE = numeric(1),
                          FUN = function(gr) {
                            length(unique(ds1[[time_vars]][
                              ds1[[group_vars]] == gr]))
                          })
  groups_with_few_tp <- names(tab_groups_tp)[tab_groups_tp < 3]
  if (length(groups_below_min_obs) > 0 | length(groups_with_few_tp) > 0) {
    to_excl <- unique(c(groups_below_min_obs, groups_with_few_tp))
    util_message(paste("Levels of the group_var with too few observations",
                       "were discarded",
                       paste0("(level",
                              ifelse(length(to_excl) > 1, "s ", " "),
                              paste(to_excl, collapse = ", "),
                              ").")
    ),
    applicability_problem = FALSE)
    # exclude levels with few observations or time points
    ds1 <- subset(ds1,
                  ds1[[group_vars]] %in%
                    setdiff(levels(ds1[[group_vars]]), to_excl))
    # drop unused levels
    if (!is.factor(ds1[[group_vars]])) {
      ds1[[group_vars]] <- factor(ds1[[group_vars]])
    } else {
      ds1[[group_vars]] <-
        droplevels(ds1[[group_vars]])
    }
  }

  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.",
               applicability_problem = TRUE)
  }

  if (exclude_constant_subgroups) {
    lvl_to_exclude <- levels(ds1[[group_vars]])[
      vapply(levels(ds1[[group_vars]]), FUN.VALUE = logical(1), function(gr) {
        check1 <- util_dist_selection(
          ds1[[resp_vars]][which(as.character(ds1[[group_vars]]) == gr)]
        )
        check1$NDistinct < 2 ||
          check1$PropZeroes > 0.95 ||
          check1$PropZeroes < 0.05
      })
    ]
    if (length(lvl_to_exclude) > 0) {
      util_message(paste("Levels of the group_var with constant values",
                         "were discarded",
                         paste0("(level",
                                ifelse(length(lvl_to_exclude) > 1, "s ", " "),
                                paste(lvl_to_exclude, collapse = ", "),
                                ").")
      ),
      applicability_problem = FALSE)
      ds1 <- subset(ds1,
                    ds1[[group_vars]] %in%
                      setdiff(levels(ds1[[group_vars]]), lvl_to_exclude))
      # drop unused levels
      if (!is.factor(ds1[[group_vars]])) {
        ds1[[group_vars]] <- factor(ds1[[group_vars]])
      } else {
        ds1[[group_vars]] <-
          droplevels(ds1[[group_vars]])
      }
    }
  }

  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.",
               applicability_problem = TRUE)
  }

  # collapse 'rare' groups to reduce the number of levels, if needed
  tab_groups <- table(ds1[[group_vars]])
  have_other <- FALSE
  if (length(tab_groups) > n_group_max) {
    tab_groups <- tab_groups[order(tab_groups, decreasing = TRUE)]
    keep_gr <- names(tab_groups)[1:n_group_max]
    levels(ds1[[group_vars]])[which(!levels(ds1[[group_vars]])
                                    %in% keep_gr)] <- "other"
    tab_groups[["other"]] <- sum(ds1[[group_vars]] == "other")
    # new category 'other' should always be the last one
    have_other <- TRUE
    lvl_gr <-
      c(levels(ds1[[group_vars]])[which(levels(ds1[[group_vars]])
                                        %in% keep_gr)],
        "other")
    ds1[[group_vars]] <- as.character(ds1[[group_vars]])
    ds1[[group_vars]] <- factor(ds1[[group_vars]], levels = lvl_gr)
  }

  if (length(levels(ds1[[group_vars]])) < 2) {
    plot_format <- "COMBINED"
  }

  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.",
               applicability_problem = FALSE)
  }

  var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])
  if (var_prop$NDistinct < 2) {
    util_error("The response variable is constant after data preparation.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }
  if (var_prop$NDistinct > 2) {
    util_error("The response variable is not binary after data preparation.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }
  if (var_prop$PropZeroes > 1 - min_proportion ||
      var_prop$PropZeroes < min_proportion) {
    util_error("The response variable contains too few cases/controls.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  # order data by time and groups
  # (for plotting and for the moving window calculations)
  ds1 <- ds1[order(ds1[[time_vars]], ds1[[group_vars]]), ]
  # reduce time points according to the resolution, if needed
  tp_seq <- unique(ds1[[time_vars]])
  tp_round_seq <- util_optimize_sequence_across_time_var(
    time_var_data = tp_seq,
    n_points = resolution)
  ds1[["ROUND_TIME"]] <- suppressWarnings(util_parse_date(
    lubridate::round_date(ds1[[time_vars]], unit = tp_round_seq)))

  # store a numeric version of the original time variable for later calculations
  ds1$time_vars_num <- suppressWarnings(as.numeric(ds1[[time_vars]]))

  # Modelling group-wise trends ------------------------------------------------
  # adjust response for covariables (if any)
  if (length(co_vars) > 0) {
    fmla <- as.formula(paste0(paste0(util_bQuote(resp_vars), "~"),
                              paste0(
                                paste0(util_bQuote(co_vars), collapse = " + "),
                                " + ",
                                util_bQuote(group_vars)
                              )))
    suppressWarnings({
      logfit1 <- glm(fmla, data = ds1, family = binomial(link = "logit"))
    })
    group_marg <- data.frame(
      emmeans::emmeans(logfit1, group_vars, type = "response"),
      check.names = FALSE)
    # store residuals
    # These values will be used for LOESS fits. In this way, we fit LOESS after
    # adjusting the response for the covariables.
    ds1$Residuals <-
      # estimated mean for each level of the grouping variable
      group_marg$prob[match(ds1[[group_vars]], group_marg[, group_vars])] +
      # residuals: original value of the response variable - fitted value
      ds1[[resp_vars]] - logfit1$fitted.values
    rm(logfit1) # Memory consumption
  } else {
    ds1$Residuals <- ds1[[resp_vars]]
  }

  # calculate LOESS smoothing parameter based on the number of observations
  bw_loess <- min(1, round(100/nrow(ds1), 2)) # upper limit: 1
  bw_loess <- max(min_bandwidth, bw_loess) # lower limit as specified

  # fit LOESS/GAM for each group separately
  grouped_ds1 <- split(ds1, ds1[[group_vars]])
  processed_grouped_ds1 <- lapply(grouped_ds1, function(data_i) {
    if (var(data_i[["Residuals"]]) == 0) { # constant for this subgroup
      df_i <- unique(data_i[, c("Residuals", "ROUND_TIME")])
      fit_vals <- df_i[["Residuals"]]
      data_i_seq <- df_i[["ROUND_TIME"]]
    } else if (max(tab_groups) > 1000 &&
        util_ensure_suggested("mgcv",
                              "use GAM from mgcv instead of loess for lower memory consumption",
                              err = FALSE) &&
        enable_GAM) {
      # If there are too many observations, switch to GAM instead of LOESS
      # because of memory consumption (if available).
      fit_i <- mgcv::gam(Residuals ~ s(time_vars_num, bs = "cs"),
                         method = "REML",
                         data = data_i)
      # To plot the trend line at the time points in `tp_round_seq`
      # (restricted to those values that lie within the observed time period
      # for this group), we need fitted values at these time points.
      data_i_seq <- tp_round_seq[
        (which(tp_round_seq == min(data_i[["ROUND_TIME"]]))):
          (which(tp_round_seq == max(data_i[["ROUND_TIME"]])))]
      data_i_seq_num <- suppressWarnings(as.numeric(data_i_seq))
      data_i_seq_num <- as.data.frame(data_i_seq_num)
      colnames(data_i_seq_num) <- "time_vars_num"

      fit_vals <- mgcv::predict.gam(fit_i, data_i_seq_num)
    } else { # LOWESS
      # fit LOWESS for data_i
      fit_i <- suppressWarnings(
        lowess(x = data_i[["ROUND_TIME"]],
               y = data_i[["Residuals"]],
               f = bw_loess,
               iter = 0)) # For binary variables, the `robustifying iterations`
      # can lead to an underestimation of the proportion of cases.
      fit_i_df <- unique(as.data.frame(fit_i))
      fit_vals <- fit_i_df$y
      val_bel_0 <- fit_vals < 0
      val_abv_1 <- fit_vals > 1
      if (any(c(val_bel_0, val_abv_1))) {
        fit_vals[val_bel_0] <- 0
        fit_vals[val_abv_1] <- 1
      }
      data_i_seq <- util_parse_date(fit_i_df$x)
    }

    pred_df <- data.frame(TIME = data_i_seq,
                          FITTED_VALUE = fit_vals,
                          GROUP = rep(data_i[[group_vars]][1],
                                      length(data_i_seq)))
    return(res_round_tp = pred_df[which(!is.na(pred_df$FITTED_VALUE)), ])
  })
  # https://stackoverflow.com/a/39838759
  fit_groups <- dplyr::bind_rows(processed_grouped_ds1)
  # Memory consumption
  rm(grouped_ds1)

  # Plotting ------------------------------------------------------------------
  if (length(co_vars) > 0) {
    if (length(co_vars) < 10) {
      subtitle <- sprintf("adjusted for %s", paste0(co_vars, collapse = ", "))
    } else {
      subtitle <- sprintf("adjusted for %d variables", length(co_vars))
    }
  } else {
    subtitle <- ""
  }

  {
    n_lev_gv_reduced <-
      length(unique(fit_groups$GROUP))
    if (n_lev_gv_reduced < n_lev_gv_total) {
      appendix <- sprintf("%d of %d levels shown",
                         n_lev_gv_reduced - have_other,
                         n_lev_gv_total)
    } else {
      appendix <- sprintf("all %d levels shown",
                         n_lev_gv_reduced - have_other)
    }
    if (nzchar(subtitle)) {
      subtitle <- paste(subtitle,
                        appendix,
                        sep = " - ")
    } else {
      subtitle <- appendix
    }
  }

  levs <- as.character(levels(ds1[[group_vars]]))

  g_labs <- paste(
    levs,
    sprintf("(n=%d)", unname(tab_groups[levs]))
  )

  if (length(levels(ds1[[group_vars]])) <= 8) {
    hex_code <- c(
      "#56B4E9", "#E69F00", "#009E73",
      "#F0E442", "#0072B2", "#D55E00", "#CC79A7", "#8C510A"
    )
    hex_code <- head(hex_code, length(levs))
    names(hex_code) <- levs
  } else {
    hex_code <- NULL
  }

  y_min <- max(0, mean(ds1$Residuals) - sd(ds1$Residuals))
  y_max <- min(1, mean(ds1$Residuals) + sd(ds1$Residuals))

  # Facet-Grids for categorical variable (observer/device)
  p1 <- util_create_lean_ggplot(ggplot(fit_groups,
                                       aes(x = .data$TIME,
                                           y = .data$FITTED_VALUE,
                                           color = .data$GROUP)) + {
                                             if (!is.null(hex_code)) {
                                               scale_color_manual(
                                                 values = hex_code,
                                                 breaks = levs,
                                                 labels = g_labs
                                               ) # https://chatgpt.com/c/69e0ecc5-10a8-8397-9f51-0d365cfa1876
                                             } else {
                                               ggplot2::scale_color_discrete(
                                                 breaks = levs,
                                                 labels = g_labs
                                               )
                                             }} +
                                  xlab(rvs_bin_note) +
                                  ylab("") +
                                  geom_line() +
                                  facet_wrap(~ .data$GROUP, ncol = 2) + #TODO: What about this ~?
                                  expand_limits(y = c(y_min, y_max)) +
                                  theme_minimal() +
                                  ggtitle(plot_title, subtitle) +
                                  theme(legend.title = element_blank()),
                                fit_groups = fit_groups,
                                hex_code = hex_code,
                                levs = levs,
                                g_labs = g_labs,
                                rvs_bin_note = rvs_bin_note,
                                y_max = y_max,
                                y_min = y_min,
                                plot_title = plot_title,
                                subtitle = subtitle)


  # combined plot
  p2 <- util_create_lean_ggplot(ggplot(fit_groups,
                                       aes(x = .data$TIME,
                                           y = .data$FITTED_VALUE,
                                           group = .data$GROUP,
                                           color = .data$GROUP)) + {
                                             if (!is.null(hex_code)) {
                                               scale_color_manual(
                                                 values = hex_code,
                                                 breaks = levs,
                                                 labels = g_labs
                                               ) # https://chatgpt.com/c/69e0ecc5-10a8-8397-9f51-0d365cfa1876
                                             } else {
                                               ggplot2::scale_color_discrete(
                                                 breaks = levs,
                                                 labels = g_labs
                                               )
                                             }} +
                                  xlab(rvs_bin_note) +
                                  ylab("") +
                                  geom_line() +
                                  expand_limits(y = c(y_min, y_max)) +
                                  theme_minimal() +
                                  ggtitle(plot_title, subtitle),
                                fit_groups = fit_groups,
                                hex_code = hex_code,
                                levs = levs,
                                g_labs = g_labs,
                                rvs_bin_note = rvs_bin_note,
                                y_min = y_min,
                                y_max = y_max,
                                plot_title = plot_title,
                                subtitle = subtitle)

  if (length(levels(ds1[[group_vars]])) > 1) {
    p2 <- util_create_lean_ggplot(p2 +
                                    theme(legend.title = element_blank()),
                                  p2 = p2)
  } else {
    p2 <- util_create_lean_ggplot(p2 +
                                    theme(legend.position = "none"),
                                  p2 = p2)
  }

  p1 <- util_set_size(p1,
                      width_em = 45,
                      height_em = length(levels(ds1[[group_vars]])) * 15 / 2)
  p2 <- util_set_size(p2, 30, 15)

  pl <- list(
    Loess_fits_facets = p1,
    Loess_fits_combined = p2
  )

  if (length(plot_format) != 1 || !is.character(plot_format)) {
    plot_format <- "NOT character(1) STRING AT ALL"
  }

  # Add attribute with size hints to the combined plot
  if (!is.null(pl[["Loess_fits_combined"]])) {
    .plot1 <- pl[["Loess_fits_combined"]]
    obj1 <- util_create_lean_ggplot(
      ggplot2::ggplot_build(.plot1),
      .plot1 = .plot1)
    min_point_line <- min(util_rbind(data_frames_list = obj1$data)$y, na.rm = TRUE)
    max_point_line <- max(util_rbind(data_frames_list = obj1$data)$y, na.rm = TRUE)
    n_groups <- length(unique(util_rbind(data_frames_list = obj1$data)$group))
    min_time <- min(util_rbind(data_frames_list = obj1$data)$x, na.rm = TRUE)
    max_time <- max(util_rbind(data_frames_list = obj1$data)$x, na.rm = TRUE)
    rm(obj1)
  }

  if (plot_format == "BOTH") {
    return(list(SummaryPlotList = pl))
  } else if (plot_format == "COMBINED") {

    return(util_attach_attr(
      list(SummaryPlotList = setNames(pl["Loess_fits_combined"],
                                      nm = resp_vars)),
      sizing_hints = list(
        figure_type_id = "dot_loess",
        range = max_point_line - min_point_line,
        no_char_y = max(nchar(c(round(max_point_line, digits = 2),
                                round(min_point_line, digits = 2)))),
        n_groups = n_groups
      )))
  } else if (plot_format == "FACETS") {
    return(list(SummaryPlotList = setNames(pl["Loess_fits_facets"],
                                           nm = resp_vars)))
  } else if (plot_format != "AUTO") {
    util_message("Unknown %s: %s -- will switch to default value AUTO.",
                 dQuote("plot_format"), dQuote(plot_format),
                 applicability_problem = TRUE)
  }
#  if (length(levels(ds1[[group_vars]])) < 15) {
    selection <- "Loess_fits_combined"
#  } else {
#    selection <- "Loess_fits_facets"
#  }
  pl <- pl[selection]
  names(pl) <- resp_vars

  return(util_attach_attr(
    list(SummaryPlotList = pl),
    sizing_hints = list(
      figure_type_id = "dot_loess",
      range = max_point_line - min_point_line,
      no_char_y = max(nchar(c(round(max_point_line, digits = 2),
                              round(min_point_line, digits = 2)))),
      n_groups = n_groups
    )))
  # TODO: This function can only be called with single resp_vars, not with a
  # vector of resp_vars. Should the output be renamed to 'SummaryPlot'?
}

#' Normalize paths, even if the tail does not exist yet
#'
#' Similar to [base::normalizePath()], but also resolves paths where a suffix
#' of the path does not yet exist. The longest existing prefix is normalized via
#' [base::normalizePath()], while the remaining non-existing tail is normalized
#' syntactically by resolving `.` and `..`.
#'
#' @param path A character vector of paths.
#' @param winslash The separator to use on Windows, passed through to
#'   [base::normalizePath()].
#' @param mustWork Logical scalar. If `TRUE`, an error is thrown when the final
#'   path does not exist. If `NA`, a warning is emitted. If `FALSE`, no error is
#'   thrown.
#'
#' @return A character vector of normalized paths.
#'
#' @noRd
util_normalize_path <- function(path, winslash = "/", mustWork = FALSE) {
  if (!is.character(path)) {
    util_error("`path` must be a character vector.")
  }
  if (length(winslash) != 1L || is.na(winslash)) {
    util_error("`winslash` must be a non-missing character scalar.")
  }
  if (length(mustWork) != 1L || !is.logical(mustWork)) {
    util_error("`mustWork` must be a logical scalar.")
  }

  is_abs_path <- function(x) {
    grepl("^(/|~|[A-Za-z]:[/\\\\]|[/\\\\]{2})", x)
  }

  path_dirname <- function(x) {
    x2 <- gsub("[/\\\\]+$", "", x)
    if (!nzchar(x2)) {
      return(x)
    }
    dirname(x2)
  }

  split_path_components <- function(x) {
    parts <- strsplit(x, "[/\\\\]+", perl = TRUE)[[1L]]
    parts[nzchar(parts)]
  }

  normalize_tail <- function(x) {
    parts <- split_path_components(x)
    if (!length(parts)) {
      return("")
    }

    stack <- character()

    for (part in parts) {
      if (identical(part, ".") || !nzchar(part)) {
        next
      }
      if (identical(part, "..")) {
        if (length(stack)) {
          stack <- stack[-length(stack)]
        }
        next
      }
      stack <- c(stack, part)
    }

    paste(stack, collapse = "/")
  }

  normalize_one <- function(p) {
    if (is.na(p)) {
      return(base::normalizePath(
        path = p,
        winslash = winslash,
        mustWork = mustWork
      ))
    }

    p_exp <- path.expand(p)
    p_abs <- if (is_abs_path(p_exp)) p_exp else file.path(getwd(), p_exp)

    p_abs <- gsub("[/\\\\]+", .Platform$file.sep, p_abs)

    if (file.exists(p_abs)) {
      return(base::normalizePath(
        path = p_abs,
        winslash = winslash,
        mustWork = mustWork
      ))
    }

    prefix <- p_abs
    repeat {
      parent <- path_dirname(prefix)
      if (file.exists(prefix) || identical(parent, prefix)) {
        break
      }
      prefix <- parent
    }

    prefix_norm <- base::normalizePath(
      path = prefix,
      winslash = winslash,
      mustWork = FALSE
    )

    prefix_cmp <- gsub("[/\\\\]+", "/", prefix)
    path_cmp <- gsub("[/\\\\]+", "/", p_abs)

    rest <- substr(path_cmp, nchar(prefix_cmp) + 1L, nchar(path_cmp))
    rest <- sub("^[/\\\\]+", "", rest)
    rest <- normalize_tail(rest)

    out <- if (nzchar(rest)) {
      file.path(prefix_norm, rest)
    } else {
      prefix_norm
    }

    if (.Platform$OS.type == "windows") {
      if (identical(winslash, "/")) {
        out <- gsub("\\\\", "/", out)
      } else if (identical(winslash, "\\")) {
        out <- gsub("/", "\\\\", out)
      }
    }

    exists_out <- file.exists(out)

    if (isTRUE(mustWork) && !exists_out) {
      util_error("path[%d]=\"%s\": No such file or directory", 1L, p)
    }
    if (is.na(mustWork) && !exists_out) {
      warning(
        sprintf("path[%d]=\"%s\": No such file or directory", 1L, p),
        call. = FALSE
      )
    }

    out
  }

  out <- vapply(path, normalize_one, character(1L), USE.NAMES = FALSE)

  names(out) <- names(path)
  out
}

#' Internal constructor for the internal class [`r dataquieR_resultset_class`].
#'
#' creates an object of the class [`r dataquieR_resultset_class`].
#'
#' @param ... properties stored in the object
#'
#' @details
#' The class features the following methods:
#'  `r paste0(" * [", methods(class = "dataquieR_resultset"), "]")`
#'
#' @return an object of the class [`r dataquieR_resultset_class`].
#' @seealso [dq_report]
dataquieR_resultset <- function(...) {
  this <- list(...)
  dataquieR_resultset_verify(this)
  class(this) <- dataquieR_resultset_class
  this
}
dataquieR_resultset_class <- "dataquieR_resultset"

#' Search for a formal in the stack trace
#'
#' Similar to [dynGet()], find a symbol in the closest data quality indicator
#' function and return its value. Can `stop()`, if symbol evaluation causes a
#' stop.
#'
#' @param symbol symbol to find
#'
#' @return value of the symbol, if available, `NULL` otherwise
#'
#' @family condition_functions
#' @noRd
util_find_indicator_function_in_callers <- function(symbol = "resp_vars") {
  n <- 1
  found <- FALSE
  try({
    while (!any( rlang::call_name(rlang::caller_call(n)) %in%
                 names(.indicator_or_descriptor))) {
    # while (!any( rlang::call_name(rlang::caller_call(n)) %in%
    #              names(.indicator_or_descriptor)) ||
    #        # find ony, if resp_vars are in the call, so not missing, but you still could write ...(resp_vars = ,)
    #        !any(symbol %in% rlang::call_args_names(rlang::caller_call(n)))) {
#      str(rlang::call_args(rlang::caller_call(n)))
      n <- n + 1
    }
    found <- TRUE
  }, silent = TRUE)
  if (found) {
    r <- withr::with_language("en", try(dynGet(symbol,
                inherits = TRUE,
                ifnotfound = NULL,
                minframe = n), silent = TRUE)) # n - 1 is relative to my caller, but for dynGet from here, it fits
    if (inherits(r, "try-error")) {
      cnd <- attr(r, "condition")
      if (conditionMessage(cnd) ==
          sprintf("argument \"%s\" is missing, with no default", symbol)) {
        return(NULL)
      } else {
        util_error(cnd)
      }
    }
    return(r)
  } else {
    return(NULL)
  }
}

#' HTML Dependency for report headers in `clipboard`
#'
#' @return the dependency
html_dependency_clipboard <- function() {
  htmltools::htmlDependency(
    name = "clipboard",
    version = "2.0.11",
    src = system.file("clipboard", package = "dataquieR"),
    script = c("clipboard.min.js")
  )
}

#' ECDF plots for distribution checks
#'
#' @description
#' Data quality indicator checks "Unexpected location" and "Unexpected
#' proportion" if a grouping variable is included: Plots of
#' empirical cumulative distributions for the subgroups.
#'
#' [Descriptor]
#'
#' @export
#'
#' @inheritParams .template_function_indicator
#' @param resp_vars [variable list] the names of the measurement variables
#' @param group_vars [variable list] the name of the observer, device or
#'                                   reader variable
#' @param n_group_max maximum number of categories to be displayed individually
#'                  for the grouping variable (`group_vars`, devices / examiners)
#' @param n_obs_per_group_min minimum number of data points per group to create
#'                  a graph for an individual category of the `group_vars` variable
#'
#' @return A [list] with:
#'   - `SummaryPlotList`: [list] of [ggplot2::ggplot]s for each response variable in
#'                    `resp_vars`.
#'
#' @importFrom ggplot2 ggplot aes stat_ecdf scale_x_continuous scale_x_datetime theme_minimal theme element_text element_blank labs
#' @importFrom rlang .data
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_distributions.html
#' )
acc_distributions_ecdf <- function(resp_vars = NULL,
                                   group_vars = NULL,
                                   study_data,
                                   label_col,
                                   item_level = "item_level",
                                   meta_data = item_level,
                                   meta_data_v2,
                                   n_group_max = getOption("dataquieR.max_group_var_levels_in_plot", dataquieR.max_group_var_levels_in_plot_default),
                                   n_obs_per_group_min = getOption("dataquieR.min_obs_per_group_var_in_plot", dataquieR.min_obs_per_group_var_in_plot_default)) {
  # preps ----------------------------------------------------------------------
  util_maybe_load_meta_data_v2()
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          #.apply_factor_metadata = TRUE, # can be omitted in favor of .apply_factor_metadata_inadm
                          .apply_factor_metadata_inadm = TRUE
  )

  # If no response variable is defined, all suitable variables will be selected.
  if (length(resp_vars) == 0) {
    util_message(
      c("All variables with interval or ratio scale according to the metadata",
        "are used by acc_distributions_ecdf."),
      applicability_problem = TRUE, intrinsic_applicability_problem = TRUE)
    resp_vars <- meta_data[[label_col]][meta_data$SCALE_LEVEL %in%
                                          c("interval", "ratio")]
    resp_vars <- intersect(resp_vars, colnames(ds1))
    if (length(resp_vars) == 0) {
      util_error("No suitable variables were defined for acc_distributions_ecdf.",
                 applicability_problem = TRUE)
    }
  }

  util_correct_variable_use(resp_vars,
                            allow_more_than_one = TRUE,
                            allow_any_obs_na = TRUE,
                            allow_all_obs_na = FALSE,
                            min_distinct_values = 10,
                            do_not_stop = ifelse(length(resp_vars) > 1,
                                                 TRUE, FALSE),
                            need_scale = "interval | ratio"
  )

  util_correct_variable_use(group_vars,
                            allow_more_than_one = FALSE,
                            allow_any_obs_na = TRUE,
                            allow_all_obs_na = FALSE,
                            min_distinct_values = 2,
                            need_type = "!float",
                            need_scale = "nominal | ordinal"
  )

  # The grouping variable should not be included as response variable.
  if (any(group_vars %in% resp_vars)) {
    resp_vars <- resp_vars[-which(resp_vars %in% group_vars)]
    util_warning(paste("Removed grouping variable from response variables",
                       "for acc_distributions_ecdf."),
                 applicability_problem = TRUE)
  }
  if (length(resp_vars) == 0) {
    util_error("No variables left to analyse for acc_distributions_ecdf.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  util_expect_scalar(n_group_max,
                     check_type = util_is_numeric_in(min = 2,
                                                     whole_num = TRUE,
                                                     finite = TRUE))

  util_expect_scalar(n_obs_per_group_min,
                     check_type = util_is_numeric_in(min = 0,
                                                     whole_num = TRUE,
                                                     finite = TRUE))

  # Which variables are of type 'datetime'?
  is_datetime_var <- vapply(resp_vars, function(rv) {
    meta_data[["DATA_TYPE"]][meta_data[[label_col]] == rv] ==
      DATA_TYPES$DATETIME
  }, FUN.VALUE = logical(1))

  # omit observations with missing grouping variable
  ds1 <- ds1[!(is.na(ds1[[group_vars]])), , drop = FALSE]

  ds1[[group_vars]] <- as.factor(ds1[[group_vars]]) # there is only one
  # grouping variable, we have ensured this by allow_more_than_one = FALSE

  tab_gr <- table(ds1[[group_vars]])

  # exclude levels of the grouping variable with too few observations
  if (any(tab_gr) < n_obs_per_group_min) {
    keep_gr <- names(tab_gr)[which(tab_gr >= n_obs_per_group_min)]
    levels(ds1[[group_vars]])[which(!levels(ds1[[group_vars]]) %in%
                                      keep_gr)] <- NA
    ds1 <- ds1[!(is.na(ds1[[group_vars]])), , drop = FALSE]
    tab_gr <- table(ds1[[group_vars]])
  }

  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.",
               applicability_problem = TRUE)
  }

  # collapse 'rare' groups to reduce the number of levels, if needed
  if (length(tab_gr) > n_group_max) {
    tab_gr <- tab_gr[order(tab_gr, decreasing = TRUE)]
    keep_gr <- names(tab_gr)[1:n_group_max]
    levels(ds1[[group_vars]])[which(!levels(ds1[[group_vars]])
                                    %in% keep_gr)] <- "other"
    # new category 'other' should always be the last one
    lvl_gr <-
      c(levels(ds1[[group_vars]])[which(levels(ds1[[group_vars]])
                                        %in% keep_gr)],
        "other")
    ds1[[group_vars]] <- as.character(ds1[[group_vars]])
    ds1[[group_vars]] <- factor(ds1[[group_vars]], levels = lvl_gr)
  }

  # find suitable labels
  lbg <- paste0(prep_get_labels(group_vars,
                                item_level = meta_data,
                                label_col = label_col,
                                resp_vars_match_label_col_only = TRUE,
                                label_class = "SHORT"))

  is_datetime_var <- vapply(resp_vars, function(rv) {
    meta_data[["DATA_TYPE"]][meta_data[[label_col]] == rv] ==
      DATA_TYPES$DATETIME
  }, FUN.VALUE = logical(1))

  # plot -----------------------------------------------------------------------
  ref_env <- environment()

  txtspec <- element_text(
    colour = "black", hjust = .5,
    vjust = .5, face = "plain", size = 10
  )

  plot_list <- lapply(setNames(nm = resp_vars), function(rv) {

    # find suitable labels
    lbr <- paste0(prep_get_labels(rv,
                                  item_level = meta_data,
                                  label_col = label_col,
                                  resp_vars_match_label_col_only = TRUE,
                                  label_class = "SHORT"))

    # omit NAs from data to prevent ggplot2 warning messages
    ds1 <- ds1[!(is.na(ds1[[rv]])), , drop = FALSE]

    .ds00 <- ds1[, c(rv, group_vars), drop = FALSE]

    pp <- util_create_lean_ggplot(
      ggplot(data = .ds00,
             aes(x = .data[[rv]], colour = .data[[group_vars]])) +
        stat_ecdf(geom = "step") +
        labs(x = "", y = paste0("ECDF: ", lbr, " (by ", lbg, ")")) +
        theme_minimal() +
        theme(
          title = txtspec,
          axis.text.x = txtspec,
          axis.text.y = txtspec,
          axis.title.x = txtspec,
          axis.title.y = txtspec,
          legend.title = element_blank()
        ),
      .ds00 = .ds00,
      rv = rv,
      group_vars = group_vars,
      lbr = lbr,
      lbg = lbg,
      txtspec = txtspec
    )

    if (!is_datetime_var[[rv]]) {
      pp <- pp + util_create_lean_ggplot(
        scale_x_continuous(expand = expansion(mult = 0.1))
      )
    } else {
      pp <- pp + util_create_lean_ggplot(
        scale_x_datetime(expand = expansion(mult = 0.1))
      )
    }

    if (util_ensure_suggested("colorspace",
                              "use the colorspace color scale",
                              err = FALSE)) {
      pp <- pp + util_create_lean_ggplot(
          colorspace::scale_color_discrete_sequential(palette = "Plasma",
                                                      na.value = "grey")
      )
    }

    util_set_size(pp)
  })

  return(list(SummaryPlotList = plot_list))
}

#' Heuristics to amend a SCALE_LEVEL column and a UNIT column in the metadata
#'
#' ...if missing
#'
#' @param resp_vars [variable list] deprecated, the function always addresses
#'                                  all variables.
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return [data.frame] modified metadata
#'
#' @export
#'
#' @examples
#' \dontrun{
#'   prep_load_workbook_like_file("meta_data_v2")
#'   prep_scalelevel_from_data_and_metadata(study_data = "study_data")
#' }
prep_scalelevel_from_data_and_metadata <- function(resp_vars =
                                                     lifecycle::deprecated(),
                                                   study_data,
                                                   item_level = "item_level",
                                                   label_col = LABEL,
                                                   meta_data = item_level,
                                                   meta_data_v2) {
  if (lifecycle::is_present(resp_vars)) {
    # Signal the deprecation to the user
    lifecycle::deprecate_warn(
      "2.5.0",
      "dataquieR::prep_scalelevel_from_data_and_metadata(resp_vars = )")
  }
  resp_vars <- NULL
  util_maybe_load_meta_data_v2()
  # TODO: shorten function name, e.g., 'prep_amend_scale_level'
  if (
    inherits(
      try(
        util_expect_data_frame(meta_data),
        silent = TRUE),
      "try-error")) {
    w <-
      "No item-level metadata provided at all. Will try to predict them."
    if (requireNamespace("cli", quietly = TRUE)) {
      w <- cli::bg_red(cli::col_br_yellow(w))
    }

    util_warning(w, immediate = TRUE)
    util_warning(w, immediate = TRUE)
    meta_data <- prep_study2meta(study_data)
  }

  prep_prepare_dataframes(
    .replace_hard_limits = TRUE,
    .amend_scale_level = FALSE, # prevent recursion
    .replace_missings = TRUE
  )

  util_correct_variable_use(resp_vars,
                            allow_more_than_one = TRUE,
                            allow_null = TRUE,
                            overwrite = TRUE,
                            remove_not_found = TRUE)

  if (is.null(resp_vars)) {
    resp_vars <- colnames(ds1)
  }

  md <- meta_data[meta_data[[label_col]] %in% resp_vars,
                  intersect(c(label_col, SCALE_LEVEL, DATA_TYPE, UNIT,
                              VALUE_LABELS, VALUE_LABEL_TABLE),
                            colnames(meta_data)), FALSE]
  distsel <- util_dist_selection(ds1)
  # TODO: What about missing codes like ".A" in metadata? They are considered
  # to be negative values, even in string columns! (".A" < 0 gives TRUE)
  # (problem with util_dist_selection, not actually here)

  meta_info <-
    merge(md, distsel, by.x = label_col, by.y = "Variables", all = TRUE)

  for (cl in c(SCALE_LEVEL, DATA_TYPE, UNIT, VALUE_LABELS, VALUE_LABEL_TABLE)) {
    if (!(cl %in% colnames(meta_info))) {
      meta_info[[cl]] <- NA_character_
    }
  }

  meta_info <- meta_info[,
                         intersect(
                           c(
                             label_col,
                             DATA_TYPE,
                             VALUE_LABELS,
                             VALUE_LABEL_TABLE,
                             "IsInteger",
                             "IsMultCat",
                             "NCategory",
                             "AnyNegative",
                             SCALE_LEVEL,
                             UNIT
                           ),
                           colnames(meta_info)
                         ), drop = FALSE]

  varrefs <- util_variable_references(meta_data = meta_data)

  # 1. Variables with value labels ---------------------------------------------
  # - the value labels indicate an order: ordinal scale
  # - the value labels are separated by `|`: nominal scale
  # assignments <- util_parse_assignments(meta_info[[VALUE_LABELS]],
  #                                       split_char = c(SPLIT_CHAR, "<"),
  #                                       multi_variate_text = TRUE)
  #
  assignments <- lapply(meta_info[[VALUE_LABEL_TABLE]],
                        function(vltnm) {
                          if (vltnm %in% prep_list_dataframes()) {
                            rdf <- prep_get_data_frame(vltnm)
                            if (!(CODE_VALUE %in% colnames(rdf))) {
                              util_warning(
                            "Missing at least a column %s in %s %s. Ignoring.",
                                sQuote(CODE_VALUE),
                                VALUE_LABEL_TABLE,
                                dQuote(vltnm),
                                applicability_problem = TRUE
                              )
                              r <- setNames(nm = character(0))
                              split_char <- SPLIT_CHAR
                              util_attach_attr(r, split_char = split_char)
                            }
                            if (!(CODE_LABEL %in% colnames(rdf))) {
                              rdf[[CODE_LABEL]] <- rdf[[CODE_VALUE]]
                            }
                            r <- setNames(rdf[[CODE_LABEL]],
                                          nm = rdf[[CODE_VALUE]])
                            split_char <- ifelse(
                              CODE_ORDER %in% colnames(rdf),
                              "<",
                              SPLIT_CHAR
                            )
                            util_attach_attr(r, split_char = split_char)
                          } else {
                            r <- setNames(nm = character(0))
                            split_char <- SPLIT_CHAR
                            util_attach_attr(r, split_char = split_char)
                          }
                        })

  meta_info$is_ordered <- vapply(assignments, attr, "split_char",
                                 FUN.VALUE = character(1)) == "<"

  if (any(meta_info$is_ordered)) { # check for counter-intuitive ordered lists
    # e.g., 2 = low < 1 = high
    # QUESTION TO STS: Should this check rather be part of prep_prepare_dataframes?
    for (a in assignments[meta_info$is_ordered]) {
      if (all(suppressWarnings(is.na(as.integer(names(a)))) ==
              is.na(names(a))) &&  # integer codes
          !all(diff(as.integer(names(a))) > 0)
      ) {
        util_message("Found counter-intuitive %s/%s: %s",
                     VALUE_LABELS,
                     VALUE_LABEL_TABLE,
                     prep_deparse_assignments(labels = vapply(a, identity,
                                                              FUN.VALUE = character(1)),
                                              codes = names(a),
                                              split_char = "<"))
      }
    }
  }

  meta_info[
    util_empty(meta_info$SCALE_LEVEL) &
      meta_info$DATA_TYPE %in% c(DATA_TYPES$STRING, DATA_TYPES$INTEGER) &
      (!util_empty(meta_info$VALUE_LABELS) | !util_empty(meta_info$VALUE_LABEL_TABLE)), SCALE_LEVEL] <-
    ifelse(meta_info$is_ordered[util_empty(meta_info$SCALE_LEVEL) &
                                  meta_info$DATA_TYPE %in%
                                    c(DATA_TYPES$STRING,
                                    DATA_TYPES$INTEGER) &
                                  (!util_empty(meta_info$VALUE_LABELS) | !util_empty(meta_info$VALUE_LABEL_TABLE))],
           SCALE_LEVELS$ORDINAL,
           SCALE_LEVELS$NOMINAL)

  # identical(attr(util_parse_assignments("a < 2", split_char = c(SPLIT_CHAR, "<")), "split_char"), "<")

  # FIXME: also consider STANDARDIZED_VOCABULARY (once it is implemented)

  # 2. Variables used as grouping variables: nominal scale ---------------------
  # for example examiners and devices
  all_group_vars <- varrefs[
    startsWith(varrefs, "GROUP_VAR_")]# TODO: What about the old schema using KEY_OBSERVER, etc.?

  if (length(all_group_vars) > 0) {
    used_as_group_var <- unique(sort(unlist(apply(meta_data[all_group_vars],
                                                  1:2,
                                                  util_find_var_by_meta,
                                                  meta_data = meta_data,
                                                  label_col = label_col,
                                                  target = label_col
    ), recursive = TRUE)))

    meta_info$used_as_group_var <- meta_info[[label_col]] %in% used_as_group_var

    meta_info[
      util_empty(meta_info$SCALE_LEVEL) &
        meta_info$used_as_group_var, SCALE_LEVEL] <-
      SCALE_LEVELS$NOMINAL
  }

  # 3. Variables used as time variables: interval scale ------------------------
  all_time_vars <- varrefs[
    startsWith(varrefs, "TIME_VAR")]

  if (length(all_time_vars) > 0) {
    used_as_time_var <- unique(sort(unlist(apply(meta_data[all_time_vars],
                                                 1:2,
                                                 util_find_var_by_meta,
                                                 meta_data = meta_data,
                                                 label_col = label_col,
                                                 target = label_col
    ), recursive = TRUE)))

    meta_info$used_as_time_var <- meta_info[[label_col]] %in% used_as_time_var

    meta_info[
      util_empty(meta_info$SCALE_LEVEL) &
        meta_info$used_as_time_var, SCALE_LEVEL] <-
      SCALE_LEVELS$INTERVAL
  }

  # 4. Variables of type 'datetime': interval scale ----------------------------
  meta_info[
    util_empty(meta_info$SCALE_LEVEL) &
      meta_info$DATA_TYPE %in% c(DATA_TYPES$DATETIME, DATA_TYPES$TIME),
    SCALE_LEVEL] <-
    SCALE_LEVELS$INTERVAL

  # 5. Variables of type 'string': nominal scale or 'na' -----------------------
  string_vars <- meta_info[
    util_empty(meta_info$SCALE_LEVEL) &
      meta_info$DATA_TYPE %in% c(DATA_TYPES$STRING),
    label_col]

  if (length(string_vars) > 0) {
    string_not_cat <- vapply(ds1[, string_vars, drop = FALSE],
                             FUN.VALUE = logical(1),
                             FUN = util_string_is_not_categorical)

    meta_info[meta_info[[label_col]] %in% string_vars, SCALE_LEVEL] <-
      ifelse(string_not_cat, SCALE_LEVELS$`NA`, SCALE_LEVELS$NOMINAL)
  }

  # 6. Classify remaining integer and float values -----------------------------
  binaryrecodelimit <-
    getOption("dataquieR.scale_level_heuristics_control_binaryrecodelimit",
            dataquieR.scale_level_heuristics_control_binaryrecodelimit_default)

  metriclevels <-
    getOption("dataquieR.scale_level_heuristics_control_metriclevels",
            dataquieR.scale_level_heuristics_control_metriclevels_default)

  if (metriclevels < binaryrecodelimit) {
    metriclevels <- ifelse(
      dataquieR.scale_level_heuristics_control_metriclevels_default <
        binaryrecodelimit,
      binaryrecodelimit + 10,
      dataquieR.scale_level_heuristics_control_metriclevels_default)

    util_warning(paste(
      "The threshold for metric variables was set too low and set to",
      metriclevels,
      "instead."),
      applicability_problem = TRUE)
  }

  # number of distinct values > 'metriclevels' (default: 25) and
  # minimum value >= 0: ratio scale
  meta_info[
    util_empty(meta_info$SCALE_LEVEL) &
    meta_info$DATA_TYPE %in% c(DATA_TYPES$INTEGER, DATA_TYPES$FLOAT) &
    ((meta_info$IsInteger & meta_info$NCategory > metriclevels) |
     !meta_info$IsInteger) &
    !meta_info$AnyNegative, SCALE_LEVEL] <-
    SCALE_LEVELS$RATIO

  # number of distinct values > 'metriclevels' (default: 25) and
  # minimum value < 0: interval scale
  meta_info[
    util_empty(meta_info$SCALE_LEVEL) &
      meta_info$DATA_TYPE %in% c(DATA_TYPES$INTEGER, DATA_TYPES$FLOAT) &
      ((meta_info$IsInteger & meta_info$NCategory > metriclevels) |
      !meta_info$IsInteger) &
      meta_info$AnyNegative, SCALE_LEVEL] <-
    SCALE_LEVELS$INTERVAL

  # number of distinct values <= 'metriclevels' (default: 25) and
  # number of distinct values > 'binaryrecodelimit' (default: 8): ordinal
  meta_info[
    util_empty(meta_info$SCALE_LEVEL) &
      meta_info$DATA_TYPE %in% c(DATA_TYPES$INTEGER, DATA_TYPES$FLOAT) &
      meta_info$IsInteger &
      meta_info$NCategory <= metriclevels &
      meta_info$NCategory > binaryrecodelimit, SCALE_LEVEL] <-
    SCALE_LEVELS$ORDINAL

  # number of distinct values <= 'binaryrecodelimit' (default: 8): nominal
  meta_info[
    util_empty(meta_info$SCALE_LEVEL) &
      meta_info$DATA_TYPE %in% c(DATA_TYPES$INTEGER, DATA_TYPES$FLOAT) &
      meta_info$IsInteger &
      meta_info$NCategory <= binaryrecodelimit, SCALE_LEVEL] <-
    SCALE_LEVELS$NOMINAL

### TODO: This should not be required here anymore, but let's test it first.
  # write also something for variables that do not fit the `Stevens's` Typology
  # e.g., notes, letters, free text, other structured content (xml, json, ...)
  meta_info[
    util_empty(meta_info$SCALE_LEVEL), SCALE_LEVEL] <-
      SCALE_LEVELS$`NA`

  meta_data[[SCALE_LEVEL]] <- setNames(meta_info[[SCALE_LEVEL]],
                                       nm = meta_info[[label_col]])[meta_data[[label_col]]]

  meta_data
}


# https://gitlab.com/libreumg/dataquier/-/issues/125
# https://gitlab.com/libreumg/dataquier/-/issues/148

# TODO: Create Test Issue: We should improve the handling of VALUE_LABLELS like male | female (i.e., w/o a coding, the actual values are in the data, which is not so uncommon) -> maybe, we should convert these to our standard case similarly to prep_apply_coding and prep_valuelabels_from_data

# TODO: Discuss: How to reduce the message about inconsistent "low < high" but "nominal" in the metadata
# TODO: Share Updated metadata with examples for all new features (< in value labels, value labels w/o number -- old feature actually, scale_level set vs. missing; add some lorem-ipsum column; add suitable variable role secondary for variables, which are now mostly missing)

# IDEA: Use a scale-free-network assumption as a foundation to expect exponentionally distributed values for key columns

# For testing:
#' prep_load_workbook_like_file("meta_data_v2")
#' View(prep_scalelevel_from_data_and_metadata(study_data = "study_data")[, union(SCALE_LEVEL, colnames(prep_scalelevel_from_data_and_metadata(study_data = "study_data")))])

#' Create an HTML file for the [dq_report2]
#'
#' @param page_nr the number of the page being created
#' @param pages list with all page-contents named by their desired file names
#' @param rendered_pages list with all rendered (`htmltools::renderTags`) page-contents named by their desired file names
#' @param template_file the report template file to use
#' @param report the output of [dq_report2]
#' @param packageName the name of the current package
#' @param dir target directory
#' @param logo logo `PNG` file
#' @param loading loading animation div
#' @param deps dependencies, as pre-processed by
#'             `htmltools::copyDependencyToDir` and
#'             `htmltools::renderDependencies`
#' @param progress_msg [closure] to call with progress information
#' @param progress [closure] to call with progress information
#' @param title [character] the web browser's window name
#' @param by_report [logical] this report html is part of a set of reports,
#'                            add a back-link
#'
#' @return `invisible(file_name)`
#'
#' @family reporting_functions
#' @concept process
#' @noRd
util_create_page_file <- function(page_nr,
                                  pages,
                                  rendered_pages,
                                  dir,
                                  template_file,
                                  report,
                                  logo,
                                  loading,
                                  packageName,
                                  deps,
                                  progress_msg,
                                  progress,
                                  title,
                                  by_report) {

  page <- names(pages)[page_nr] # the name of the page-file being created

  file_name <- file.path(dir, page)

  if (getOption("dataquieR.resume_print", dataquieR.resume_print_default) &&
       util_is_html_file_complete(file_name)) {
    progress(page_nr / length(pages) * 100)
    return(invisible(file_name))
  }

  # util_message("Writing %s...", dQuote(page))
  progress_msg("", sprintf("Writing %s...", dQuote(page)))

  util_stop_if_not(endsWith(page, ".html"))
  util_stop_if_not(!is.null(pages[[page]]))

  pg <- rendered_pages[[page]]

  pg[["dependencies"]] <- NULL

  if (by_report) {
    backlink <- htmltools::div(
      style = htmltools::css(
        font.size = "18px",
        font.weight = "bold",
        position = "fixed",
        right = "0",
        top = "125px"#,
        #border =  paste("6px", "solid", "#6495ED")
      ),
      htmltools::a(
        href = "#",
        onclick =
          'if (window.__dqPersistPopupHistory) window.__dqPersistPopupHistory();window.location.href = "../../index.html"',
        "Back to reports' overview"
      )
    )
  } else {
    backlink <- NULL
  }

  if (!is.null(attr(report, "title")) &&
      !isTRUE(attr(attr(report, "title"), "default"))) {
    header_text <- attr(report, "title")
    if (!is.null(attr(report, "subtitle")) &&
        !isTRUE(attr(attr(report, "subtitle"), "default"))) {
      header_text <- paste0(header_text, ": ", attr(report, "subtitle"))
    }
  } else {
    header_text <- NULL
  }

  if (!is.null(header_text)) {
    header <- htmltools::tagList(
      htmltools::p(class = "dq-title",
                   htmltools::tags$a(href = "report.html", header_text))
    )
  } else {
    header <- NULL
  }

  if (by_report) {
    by_report <- "true"
  } else {
    by_report <- "false"
  }

  html_report <- htmltools::htmlTemplate(template_file,
                                  by_report = by_report,
                                  document_ = TRUE,
                                  spage = pg,
                                  logo = logo,
                                  menu = .menu_env$menu(pages),
                                  loading = loading,
                                  deps = deps,
                                  title = title,
                                  backlink = backlink,
                                  header = header)

  #fix: sort reportsummarytable by first (sysmiss) and varaible column

  # https://atomiks.github.io/tippyjs/v6/all-props/

  # htmltools::save_html(html_report,
  #                      libdir = file.path(dir, "lib"),
  #                      file = file_name)

  f <- file(description = file_name , open = "w", encoding = "utf-8")
  on.exit(close(f))

  withCallingHandlers({
    cat(as.character(html_report), file = f)
    # TODO: [html_files_should_also_work_alone] ack and remove RDS -- then check, if html exists, really in util_html_for_*

  },
  warning = function(cond) { # suppress a waning caused by ggplotly for barplots
    if (startsWith(conditionMessage(cond),
                   "'bar' objects don't have these attributes: 'mode'") ||
        startsWith(conditionMessage(cond),
                   "'box' objects don't have these attributes: 'mode'")) {
      invokeRestart("muffleWarning")
    }
  })

  progress(page_nr / length(pages) * 100)

  invisible(file_name)
}

util_make_report_id <- function() {
  # Stable enough uniqueness: time + pid + random
  ts <- format(Sys.time(), "%Y%m%dT%H%M%OS6", tz = "UTC")
  pid <- Sys.getpid()
  rnd <- paste(sample(c(letters, LETTERS, 0:9), 16, replace = TRUE), collapse = "")
  paste0("dq-", ts, "-p", pid, "-", rnd)
}

#' Detect falsish values
#'
#' @param x a value/vector of values
#'
#' @return vector of logical values:
#'    `TRUE`, wherever x is somehow empty
#'
#' @family missing_functions
#' @concept robustness
#' @noRd
util_is_na_0_empty_or_false <- function(x) { # TODO: generic?
  cs <- paste0("util_is_na_0_empty_or_false.", rlang::hash(x))
  if (exists(cs, envir = .study_data_cache)) { # TODO: maybe RAM intensive?
    return(get(cs, envir = .study_data_cache))
  }
  if (inherits(x, "hms")) { # maybe more general: !is.vector(x)
    x <- util_as_character(x)
  }
  y <- x
  class(y) <- "logical"
  attributes(y) <- NULL
  y[] <- FALSE
  y[is.na(x)] <- TRUE
  y[(trimws(x) %in% c("false", "FALSE", "F", "f", "-", ""))] <- TRUE
  idx <- !suppressWarnings(as.numeric(x))
  idx[is.na(idx)] <- FALSE
  y[idx] <- TRUE
  y <- as.logical(y)
  assign(cs, y, envir = .study_data_cache)
  return(y)
}

#' Simultaneous confidence intervals for multinomial proportions
#'
#' Sison & Glaz style implementation with the same user-facing API as
#' MultinomialCI::multinomialCI(x, alpha, verbose = FALSE).
#'
#' This implementation follows the published Sison & Glaz / May & Johnson
#' formulas and is written independently in R. It is suitable as a permissively
#' licensed replacement. If you redistribute it, keep the attribution note below.
#'
#' Attribution note:
#'   The algorithmic structure and formulas used here are consistent with the
#'   Sison-Glaz method as described in:
#'   - Sison, C.P. and Glaz, J. (1995)
#'   - May, W.L. and Johnson, W.D. (1997, 2000)
#'   and with the BSD-3 licensed statsmodels implementation documentation.
#'
#' @param x Integer vector of counts.
#' @param alpha Significance level in `[0, 1]`.
#' @param verbose Logical.
#'
#' @return Numeric matrix with two columns: lower and upper.
#' @noRd
util_multinomial_ci <- function(x, alpha, verbose = FALSE) {
  .util_multinomial_ci_validate_input(x = x, alpha = alpha, verbose = verbose)

  x <- as.numeric(x)
  n <- sum(x)
  k <- length(x)

  if (n <= 0) {
    util_error("sum(x) must be positive.", call. = FALSE)
  }

  p_hat <- x / n

  # Keep the documented [0, 1] boundary behavior user-friendly.
  if (alpha <= 0) {
    out <- cbind(lower = rep(0, k), upper = rep(1, k))
    return(unname(out))
  }
  if (alpha >= 1) {
    out <- cbind(lower = p_hat, upper = p_hat)
    return(unname(out))
  }

  .poisson_interval <- function(interval, lambda) {
    lower <- interval[1]
    upper <- interval[2]

    prob <- stats::ppois(upper, lambda = lambda) -
      stats::ppois(lower - 1, lambda = lambda)

    if (lambda == 0 && is.nan(prob)) {
      return(as.numeric(lower - 1 < 0))
    }

    prob
  }

  .truncated_poisson_factorial_moment <- function(interval, r, lambda) {
    lower <- interval[1]
    upper <- interval[2]

    denom <- .poisson_interval(c(lower, upper), lambda)
    if (denom <= 0) {
      return(0)
    }

    lambda^r * (
      1 - (
        (.poisson_interval(c(upper - r + 1, upper), lambda) -
           .poisson_interval(c(lower - r, lower - 1), lambda)) / denom
      )
    )
  }

  .edgeworth <- function(intervals, counts) {
    mu_r <- lapply(
      1:4,
      function(r) {
        vapply(
          seq_along(counts),
          function(i) .truncated_poisson_factorial_moment(intervals[[i]], r, counts[i]),
          numeric(1)
        )
      }
    )

    mu_r1 <- mu_r[[1]]
    mu_r2 <- mu_r[[2]]
    mu_r3 <- mu_r[[3]]
    mu_r4 <- mu_r[[4]]

    mu  <- mu_r1
    mu2 <- mu_r2 + mu - mu^2
    mu3 <- mu_r3 + mu_r2 * (3 - 3 * mu) + mu - 3 * mu^2 + 2 * mu^3
    mu4 <- mu_r4 +
      mu_r3 * (6 - 4 * mu) +
      mu_r2 * (7 - 12 * mu + 6 * mu^2) +
      mu - 4 * mu^2 + 6 * mu^3 - 3 * mu^4

    sum_mu2 <- sum(mu2)
    if (sum_mu2 <= 0) {
      return(0)
    }

    g1 <- sum(mu3) / (sum_mu2^(3 / 2))
    g2 <- (sum(mu4) - 3 * sum(mu2^2)) / (sum_mu2^2)

    x_std <- (n - sum(mu)) / sqrt(sum_mu2)
    phi <- exp(-(x_std^2) / 2) / sqrt(2 * pi)

    H3 <- x_std^3 - 3 * x_std
    H4 <- x_std^4 - 6 * x_std^2 + 3
    H6 <- x_std^6 - 15 * x_std^4 + 45 * x_std^2 - 15

    f <- phi * (1 + g1 * H3 / 6 + g2 * H4 / 24 + g1^2 * H6 / 72)
    f / sqrt(sum_mu2)
  }

  .approximated_multinomial_interval <- function(intervals, counts, n) {
    probs <- vapply(
      seq_along(counts),
      function(i) .poisson_interval(intervals[[i]], counts[i]),
      numeric(1)
    )

    if (any(probs <= 0)) {
      return(0)
    }

    ew <- .edgeworth(intervals, counts)
    if (!is.finite(ew) || ew <= 0) {
      return(0)
    }

    exp(sum(log(probs)) + log(ew) - stats::dpois(n, lambda = n, log = TRUE))
  }

  .nu <- function(c_val, counts, n) {
    intervals <- lapply(
      counts,
      function(count_i) c(max(count_i - c_val, 0), min(count_i + c_val, n))
    )
    .approximated_multinomial_interval(intervals = intervals, counts = counts, n = n)
  }

  c_val <- 1
  nu_c <- .nu(c_val, x, n)
  nu_cp1 <- .nu(c_val + 1, x, n)

  while (!(nu_c <= (1 - alpha) && (1 - alpha) < nu_cp1)) {
    if (verbose) {
      util_message(paste0(
        "c = ", c_val,
        ", nu(c) = ", format(nu_c, digits = 8),
        ", nu(c+1) = ", format(nu_cp1, digits = 8)
      ))
    }

    if (c_val > n) {
      util_error("Could not find a value c satisfying nu(c) <= 1 - alpha < nu(c + 1).",
           call. = FALSE)
    }

    c_val <- c_val + 1
    nu_c <- nu_cp1
    nu_cp1 <- .nu(c_val + 1, x, n)
  }

  g <- (1 - alpha - nu_c) / (nu_cp1 - nu_c)

  if (verbose) {
    util_message(paste0(
      "Final: c = ", c_val,
      ", nu(c) = ", format(nu_c, digits = 10),
      ", nu(c+1) = ", format(nu_cp1, digits = 10),
      ", gamma = ", format(g, digits = 10)
    ))
  }

  lower <- pmax(p_hat - c_val / n, 0)
  upper <- pmin(p_hat + (c_val + 2 * g) / n, 1)

  out <- cbind(lower = lower, upper = upper)
  unname(out)
}

.util_multinomial_ci_validate_input <- function(x, alpha, verbose) {
  if (missing(x)) {
    util_error("Argument 'x' is missing.", call. = FALSE)
  }
  if (!is.numeric(x) || !is.vector(x) || length(x) < 2L) {
    util_error("'x' must be a numeric vector of length >= 2.", call. = FALSE)
  }
  if (any(!is.finite(x))) {
    util_error("'x' must contain only finite values.", call. = FALSE)
  }
  if (any(x < 0)) {
    util_error("'x' must contain non-negative counts.", call. = FALSE)
  }
  if (any(abs(x - round(x)) > sqrt(.Machine$double.eps))) {
    util_error("'x' must contain integer counts.", call. = FALSE)
  }

  if (missing(alpha) || !is.numeric(alpha) || length(alpha) != 1L ||
      !is.finite(alpha) || alpha < 0 || alpha > 1) {
    util_error("'alpha' must be a single real number in [0, 1].", call. = FALSE)
  }

  if (!is.logical(verbose) || length(verbose) != 1L || is.na(verbose)) {
    util_error("'verbose' must be TRUE or FALSE.", call. = FALSE)
  }
}

#' Simple binomial confidence intervals for `multinomial` counts
#'
#' Computes independent binomial confidence intervals for each
#' category proportion using the same interface as
#' `MultinomialCI::multinomialCI()`.
#'
#' @param x Integer vector of counts.
#' @param alpha Significance level in `[0, 1].`
#' @param verbose Logical (unused; kept for API compatibility).
#'
#' @return Numeric matrix with columns: lower and upper.
#' @noRd
util_binomial_ci <- function(x, alpha, verbose = FALSE) {

  .util_multinomial_ci_validate_input(
    x = x,
    alpha = alpha,
    verbose = verbose
  )

  x <- as.numeric(x)

  n <- sum(x)

  out <- t(vapply(
    x,
    function(xi) {

      ci <- stats::prop.test(
        x = xi,
        n = n,
        conf.level = 1 - alpha
      )$conf.int

      c(
        lower = ci[1],
        upper = ci[2]
      )
    },
    numeric(2)
  ))

  unname(out)
}

#' Utility function to create a margins plot from linear regression models
#'
#' @param resp_vars  [variable] the name of the measurement variable
#' @param group_vars [variable] the name of the observer, device or
#'                              reader variable
#' @param co_vars [variable list] a vector of covariables, e.g. age and sex for
#'                              adjustment
#' @param threshold_type [enum] empirical | user | none. See `acc_margins`.
#' @param threshold_value [numeric] see `acc_margins`
#' @param min_obs_in_subgroup [integer] from=0. This optional argument specifies
#'                       the minimum number of observations that is required to
#'                       include a subgroup (level) of the `group_var` in the
#'                       analysis.
#' @param ds1 [data.frame] the data frame that contains the measurements, after
#'                       replacing missing value codes by `NA`, excluding
#'                       inadmissible values and transforming categorical
#'                       variables to factors.
#' @param label_col [variable attribute] the name of the column in the metadata
#'                       with labels of variables
#' @param levels [levels()] of the original ordinal variable, if applicable.
#'                          Used for axis tick labels.
#' @param adjusted_hint [character] hint, if adjusted for `co_vars`
#' @param title [character] title for the plot
#' @param n_violin_max [integer] from=0. This optional argument specifies
#'                       the maximum number of levels of the `group_var` for
#'                       which violin plots will be shown in the figure.
#' @param sort_group_var_levels [logical] Should the levels of the grouping
#'                        variable be sorted descending by the number of
#'                        observations (in the figure)?
#' @param include_numbers_in_figures [logical] Should the figure report the
#'                        number of observations for each level of the grouping
#'                        variable?
#'
#' @return A table and a matching plot.
#'
#' @importFrom ggplot2 ggplot geom_violin geom_boxplot geom_pointrange geom_text element_blank element_text geom_density coord_flip annotate ggplot_build theme_minimal labs theme scale_colour_manual geom_count aes geom_vline geom_hline
#' @import patchwork
#'
#' @noRd
util_margins_lm <- function(resp_vars = NULL, group_vars = NULL, co_vars = NULL,
                            threshold_type = NULL, threshold_value,
                            min_obs_in_subgroup = 5, ds1, label_col,
                            levels = NULL,
                            adjusted_hint = "",
                            title = "",
                            n_violin_max = getOption("dataquieR.max_group_var_levels_with_violins", dataquieR.max_group_var_levels_with_violins_default),
                            sort_group_var_levels =
                              getOption("dataquieR.acc_margins_sort",
                                        dataquieR.acc_margins_sort_default),
                            include_numbers_in_figures =
                              getOption("dataquieR.acc_margins_num",
                                        dataquieR.acc_margins_num_default)) {
  # preps and checks -----------------------------------------------------------
  # to avoid "no visible binding for global variable ‘sample_size’"
  sample_size <- NULL
  # ensure that the response variable is not constant
  var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])
  if (var_prop$NDistinct < 2) {
    util_error("The response variable is constant (after data preparation).",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }
  # ensure that there are enough observations for the model
  countdf <- util_table_of_vct(ds1[[group_vars]])
  if (any(countdf[, 2] < min_obs_in_subgroup)) {
    util_error("Not enough observations per group (after data preparation).")
  }

  if (nrow(countdf) <= n_violin_max) {
    show_violins <- TRUE
  } else {
    show_violins <- FALSE
  }

  # modelling ------------------------------------------------------------------
  # if no co_vars are defined for adjustment only the intercept is modelled
  if (length(co_vars) == 0) {
    co_vars <- "1"
    co_vars_bQ <- co_vars
  } else {
    co_vars_bQ <- util_bQuote(co_vars)
  }

  # build model formula
  fmla <- as.formula(paste0(
    paste0(util_bQuote(resp_vars), "~"),
    paste0(
      paste0(co_vars_bQ, collapse = " + "),
      " + ",
      util_bQuote(group_vars)
    )
  ))

  model <- lm(fmla, data = ds1)
  res_df <- data.frame(emmeans::emmeans(model, group_vars, type = "response"),
                       check.names = FALSE)

  # adjust for covariates, if needed
  ds1$resp_var_adj <-
    # estimated mean for each level of the grouping variable
    res_df$emmean[match(ds1[[group_vars]], res_df[, group_vars])] +
    # residuals: original value of the response variable - fitted value
    model$residuals

  summary_ds <- as.data.frame(
    dplyr::summarize(
      dplyr::group_by_at(ds1[, c(resp_vars, group_vars), drop = FALSE],
                         unname(group_vars)),
      sample_size = dplyr::n()))

  res_df <- merge(res_df, summary_ds, by = group_vars, all.x = TRUE)
  res_df <- dplyr::rename(res_df, c("margins" = "emmean", "LCL" = "lower.CL",
                                    "UCL" = "upper.CL"))

  # adjusted overall mean
  omv <- data.frame(emmeans::emmeans(model, "1", type = "response"))
  omv <- dplyr::rename(omv, c("margins" = "emmean", "LCL" = "lower.CL",
                                  "UCL" = "upper.CL"))
  res_df$overall <- omv$margins # TODO: never used?

  # thresholds -----------------------------------------------------------------
  if (threshold_type %in% c("empirical", "none")) {
    th <- sd(ds1[["resp_var_adj"]])  # TODO: here we use an empirical estimate,
    # but in other margin functions we use estimates based on properties of the
    # assumed distribution
    parn <- c(
      paste("-", threshold_value, "TH", sep = ""), "Mean",
      paste("+", threshold_value, "TH", sep = "")
    )
    pars <- as.vector(c(
      omv$margins - threshold_value * th, omv$margins,
      omv$margins + threshold_value * th
    ))

    res_df$threshold <- threshold_value
    res_df$GRADING <- ifelse(res_df$margins < pars[1] |
                               res_df$margins > pars[3], 1, 0)
  } else if (threshold_type == "user") {
    th <- threshold_value
    parn <- c("", paste0("Mean=", threshold_value, sep = ""), "")
    pars <- as.vector(c(th, th, th))

    res_df$threshold <- th
    res_df$GRADING <- mapply(
      function(th, l, u) {
        ifelse(th >= l & th <= u, 0, 1)
      },
      res_df$threshold, res_df$LCL, res_df$UCL
    )
  }

  # figure ---------------------------------------------------------------------

  # use offset for annotation depending on variable scale
  offs <- ifelse(th <= 50, th / 10, th / 20)

  warn_code <- c("1" = "#B2182B", "0" = "#2166AC")

  if (sort_group_var_levels) {
    # order levels of the grouping variable by number of observations
    tbl_gr <- util_table_of_vct(ds1[[group_vars]]) %>%
      dplyr::arrange(dplyr::desc(.data$Freq))
    ds1[[group_vars]] <- factor(ds1[[group_vars]],
                                levels = as.character(tbl_gr$Var1))
    res_df <- res_df[match(levels(ds1[[group_vars]]),
                           as.character(res_df[, group_vars])), ]
  }

  # Plot 1: hybrid density/boxplot graph
  .ds00 <- ds1[, c("resp_var_adj", group_vars), drop = FALSE]
  if (all(!is.finite(.ds00$resp_var_adj))) {
    util_error("No data left.")
  }
  p1 <-
    util_create_lean_ggplot(ggplot(data = .ds00,
                                   aes(x = .data[[group_vars]],
                                       y = .data[["resp_var_adj"]])),
                            .ds00 = .ds00,
                            group_vars = group_vars)
  if (show_violins) {
    p1 <- util_create_lean_ggplot(p1 +
                                    geom_violin(alpha = 0.9,
                                                fill = "gray99") +
                                    ggplot2::stat_ydensity( # FIXME: ggplot2 4.0.0 now draws actual data quantiles (not “density-inferred” ones), so the lines may shift very slightly compared to pre-4.0.0 outputs, but the intent (e.g., median at 0.5) is the same. ggplot2.tidyverse.org   -- do we need to change some documentation?
                                      geom = "violin",
                                      alpha = 0.9,
                                      fill  = "gray99",
                                      quantiles = 0.5
                                    ),
                                  p1 = p1)
  }
  p1 <- util_create_lean_ggplot(p1 +
                                  geom_boxplot(width = 0.1,
                                               fill = "white",
                                               color = "gray",
                                               alpha = 0.5) +
                                  util_geom_pointrange_robust(
                                    data = res_df, aes(
                                      x = factor(.data[[group_vars]]),
                                      y = margins,
                                      ymin = LCL,
                                      ymax = UCL,
                                      color = as.factor(GRADING)#, n = sample_size
                                    ),
                                    shape = 18,
                                    linewidth = 1,
                                    inherit.aes = FALSE,
                                    size = .5) +
                                  theme_minimal() +
                                  labs(x = "", y = "") +
                                  theme(
                                    legend.position = "none",
                                    legend.title = element_blank(),
                                    text = element_text(size = 16),
                                    axis.text.x = element_text(angle = 90,
                                                               vjust = 0.5,
                                                               hjust = 1)
                                  ) +
                                  scale_colour_manual(values = warn_code),
                                p1 = p1,
                                res_df = res_df,
                                group_vars = group_vars,
                                warn_code = warn_code)

  y_lims <- range(c(.ds00[["resp_var_adj"]], pars, pars + offs),
                  finite = TRUE)

  p1 <- util_create_lean_ggplot(
    p1 +
      ggplot2::coord_cartesian(ylim = y_lims,
                                                            clip = "off"),
    p1 = p1,
    y_lims = y_lims)

  if (include_numbers_in_figures) {
    p1 <- util_create_lean_ggplot(p1 +
                                    geom_text(data = summary_ds,
                                              aes(x = .data[[group_vars]],
                                                  y = Inf,
                                                  label = sample_size),
                                              inherit.aes = FALSE,
                                              hjust = 0.5,
                                              vjust = -0.2,
                                              angle = 90) +
                                    annotate("text",
                                             x = 0.5,
                                             y = Inf,
                                             label = "N",
                                             vjust = -0.2) +
                                    theme(plot.margin = ggplot2::unit(c(8, 0, 2, 0),
                                                                      "mm")),
                                  p1 = p1,
                                  summary_ds = summary_ds,
                                  group_vars = group_vars,
                                  y_lims = y_lims
    )
  }

  if (!is.null(levels)) {
    try(p1 <-
          util_create_lean_ggplot(
            p1 +
              ggplot2::scale_y_continuous(breaks =
                                            as.integer(names(levels)),
                                          labels = levels),
            p1 = p1,
            levels = levels),
        silent = TRUE)
  }

  if (threshold_type != "none") {
    p1 <- util_create_lean_ggplot(p1 +
                                    geom_hline(yintercept = pars[2],
                                               color = "red") +
                                    geom_hline(yintercept = pars[-2],
                                               color = "red",
                                               linetype = 2),
                                  p1 = p1,
                                  pars = pars)
  } else {
    p1 <- util_create_lean_ggplot(p1 +
                                    geom_hline(yintercept = pars[2],
                                               color = "red"),
                                  p1 = p1,
                                  pars = pars)
  }

  # Plot 2: overall distributional plot flipped on y-axis of plot 1
  .ds03 <- ds1[, "resp_var_adj", drop = FALSE]
  get_y_scale <- util_create_lean_ggplot(
    ggplot(.ds03,
           aes(x = .data[["resp_var_adj"]])) +
      geom_density(alpha = 0.35),
    .ds03 = .ds03)

  build <- ggplot2::ggplot_build(get_y_scale)
  data1 <- util_gg_get(build, "data")[[1]]
  yvals <- util_gg_get(data1, "y")

  aty <- mean(range(yvals))

  p2 <- util_create_lean_ggplot(ggplot(.ds03,
                                       aes(.data[["resp_var_adj"]])) +
                                  geom_density(alpha = 0.35) +
                                  coord_flip() + # util_lazy_add_coord(p, fli)
                                  theme_minimal() +
                                  labs(x = NULL, y = NULL) +
                                  ggplot2::xlim(y_lims) +
                                  theme(axis.text.y = element_blank(),
                                        axis.text.x = element_blank(),
                                        text = element_text(size = 16)),
                                .ds03 = .ds03,
                                y_lims = y_lims)

  if (threshold_type != "none") {
    p2 <- util_create_lean_ggplot(
      p2 +
        annotate(geom = "text",
                 x = pars + offs,
                 y = aty,
                 label = parn) +
        geom_vline(xintercept = pars[2], color = "red") +
        geom_vline(xintercept = pars[-2], color = "red", linetype = 2),
      p2 = p2,
      pars = pars,
      offs = offs,
      aty = aty,
      parn = parn)
  } else {
    p2 <- util_create_lean_ggplot(
      p2 +
        annotate(geom = "text",
                 x = pars + offs,
                 y = aty,
                 label = c("", parn[2], "")) +
        geom_vline(xintercept = pars[2], color = "red"),
      p2 = p2,
      pars = pars,
      offs = offs,
      aty = aty,
      parn = parn)
  }
  # combine plots and add the title
  res_plot <- # TODO: For all patchwork-calls, add information to reproduce the layout in plot.ly
    util_create_lean_ggplot(
      p1 +
        p2 +
        plot_layout(nrow = 1,
                    widths = c(5, 1)) +
        plot_annotation(title = title,
                        subtitle = adjusted_hint),
      p1 = p1,
      p2 = p2,
      title = title,
      adjusted_hint = adjusted_hint)

  # output ---------------------------------------------------------------------
  return(list("plot_data" = res_df, "plot" = res_plot))
}

#' Attaches attributes about the recommended minimum
#' absolute sizes to the plot p
#'
#' @param p [ggplot2::ggplot] the plot
#' @param width_em [numeric] len=1. the minimum width hint in `em`
#' @param height_em [numeric] len=1. the minimum height in `em`
#'
#' @return p the modified plot
#'
#' @family reporting_functions
#' @concept figure
#' @noRd
util_set_size <- function(p, width_em = NA_integer_, height_em = NA_integer_) {
  if (!util_is_gg(p)) {
    util_error("Internal error, p should be a ggplot object, not: %s",
               util_pretty_vector_string(class(p)))
  }
  attr(p, "width_em") <- width_em
  attr(p, "height_em") <- height_em
  return(p)
}

test_that("util_set_size works", {
  skip_on_cran()

  expect_error(util_set_size(plot(mtcars$wt, mtcars$mpg),
                             width_em =  6, height_em =  4))

  # Example ggplot object
  p <- ggplot(mtcars, aes(wt, mpg)) + geom_point()

  # Set width and height using util_set_size
  p_resized <- util_set_size(p, width_em =  6, height_em =  4)

  # Check that the attributes have been set correctly
  expect_equal(attr(p_resized, "width_em"),  6)
  expect_equal(attr(p_resized, "height_em"),  4)
})

#' convenience function to abbreviate all(util_is_integer(...))
#'
#' @inheritParams util_is_integer
#'
#' @return `TRUE`, if all entries are integer-like, `FALSE` otherwise
#'
#' @seealso `util_is_integer`
#' @family process_functions
#' @concept process
#' @noRd
util_all_is_integer <- function(x) {
  all(util_is_integer(x))
}

#' Tail for R
#'
#' @param file_name [character] a file name
#' @param bytes [integer] number of bytes to return, careful with multi-character
#'.             encodings like `utf-8`
#'
#' @returns last `bytes` bytes from `file_name`
#' @noRd
util_tail_file <- function(file_name, bytes) {
  s <- file.size(file_name)
  offs <- max(0, s - bytes)
  con <- file(file_name, open = "rb")
  withr::defer(close(con))
  seek(con, offs, rw = "rb")
  rawToChar(readBin(con, "raw", bytes))
}

#' Verify, if an `HTML` file looks complete
#'
#' @param file_name [character] file name
#'
#' @returns `TRUE`, if the file features an `HTML`-end tag within its last 20
#'                  bytes.
#' @noRd
util_is_html_file_complete <- function(file_name) {
  r <- suppressWarnings(try(
    file.exists(file_name) &&
    file.access(file_name, 2) == 0 &&
    endsWith(tolower(trimws(util_tail_file(file_name, 20))),  "</html>"),
    silent = TRUE))
  r <- !util_is_try_error(r) && r
}

#' Compute Pairwise Correlations
#'
#' works on variable groups (`cross-item_level`), which are expected to show
#' a Pearson correlation
#'
#' [Descriptor] # TODO: This can be an indicator
#'
#' @inheritParams .template_function_indicator
#'
#' @param meta_data_cross_item [meta_data_cross]
#' @param cross_item_level [data.frame] alias for `meta_data_cross_item`
#' @param `cross-item_level` [data.frame] alias for `meta_data_cross_item`
#'
#' @return a `list` with the slots:
#'   - `SummaryPlotList`: for each variable group a [ggplot2::ggplot] object with
#'                        pairwise correlation plots
#'   - `SummaryData`: table with columns `VARIABLE_LIST`, `cors`,
#'                    `max_cor`, `min_cor`
#'   - `SummaryTable`: like `SummaryData`, but machine readable and with
#'                     stable column names.
#' @export
#'
#' @importFrom stats cor
#'
#' @examples
#' \dontrun{
#' devtools::load_all()
#' prep_load_workbook_like_file("meta_data_v2")
#' des_scatterplot_matrix("study_data")
#' }
des_scatterplot_matrix <- function(label_col, # FIXME: This needs a variable_group formal, too
                                   study_data,
                                   item_level = "item_level",
                                   meta_data_cross_item = "cross-item_level",
                                   meta_data = item_level,
                                   meta_data_v2,
                                   cross_item_level,
                                   `cross-item_level`) {
  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()
  prep_prepare_dataframes(.apply_factor_metadata = TRUE)

  meta_data_cross_item <- util_normalize_cross_item(
    meta_data = meta_data,
    meta_data_cross_item = meta_data_cross_item,
    label_col = label_col)
  VariableGroupPlotList <- lapply(which(
    useNames = TRUE,
    trimws(tolower(setNames(meta_data_cross_item$ASSOCIATION_METRIC,
                            nm = meta_data_cross_item$CHECK_LABEL))) %in%
      c("pearson", "spearman")), function(vg) {
        rvs <-
          vapply(
            util_parse_assignments(meta_data_cross_item[vg, VARIABLE_LIST]),
            identity, FUN.VALUE = character(1))
       rvs <- rvs[vapply(rvs, FUN = function(x) {
         !all(is.na(ds1[[x]]))
         }, FUN.VALUE = logical(1) )]
       if (length(rvs) == 0) {
         r_py <- htmltools::div(
           htmltools::h3("No data available to create the plot"))
         r <- util_create_lean_ggplot(ggplot() +
           annotate("text", x = 0, y = 0,
                    label = "No data available to create the plot") +
           theme(
             axis.line = element_blank(),
             axis.text.x = element_blank(),
             axis.text.y = element_blank(),
             axis.ticks = element_blank(),
             axis.title.x = element_blank(),
             axis.title.y = element_blank(),
             legend.position = "none",
             panel.background = element_blank(),
             panel.border = element_blank(),
             panel.grid.major = element_blank(),
             panel.grid.minor = element_blank(),
             plot.background = element_blank()
           ))
        number_of_vars <- 0
        attr(r, "plotly") <- r_py
       } else {
         columnLabels <- prep_get_labels(rvs,
                                         item_level = meta_data,
                                         label_col = label_col,
                                         resp_vars_match_label_col_only = TRUE,
                                         label_class = "SHORT")
         correlation_method <-
           trimws(tolower(meta_data_cross_item[vg, ASSOCIATION_METRIC]))
         .dt0 <- ds1[, rvs]
         cc <- complete.cases(.dt0)
         if (any(!cc, na.rm = TRUE)) {
           util_warning(
      "Found %d incomplete cases in %s. Discarding such rows for correlations.",
                        sum(!cc), util_pretty_vector_string(columnLabels))
         }
         .dt0 <- .dt0[cc, , FALSE]
         r <- util_create_lean_ggplot(util_pairs_ggplot(.dt0,
                              columnLabels = columnLabels,
                              correlation_method = correlation_method),
                              .dt0 = .dt0,
                              columnLabels = columnLabels,
                              correlation_method = correlation_method
         )
         r_py <- util_pairs_plotly(.dt0,
                                   columnLabels = columnLabels,
                                   correlation_method = correlation_method)
         attr(r, "plotly") <- r_py
         number_of_vars <- length(rvs)
       }


       #Sizing information - a default
       attr(r, "sizing_hints") <- list(
         figure_type_id = "pairs_plot",
         number_of_vars = number_of_vars
       )
       return(r)
      })

  names(VariableGroupPlotList) <-
    as.character(meta_data_cross_item[
      trimws(tolower(setNames(meta_data_cross_item$ASSOCIATION_METRIC,
                              nm = meta_data_cross_item$CHECK_LABEL))) %in%
        c("pearson", "spearman"), CHECK_LABEL])
  VariableGroupTable <-
    util_rbind(data_frames_list = lapply(which(
      useNames = TRUE,
      trimws(tolower(setNames(meta_data_cross_item$ASSOCIATION_METRIC,
                              nm = meta_data_cross_item$CHECK_LABEL))) %in%
        c("pearson", "spearman")), function(vg) {

          vl <- meta_data_cross_item[vg, VARIABLE_LIST]

          correlation_method <-
            trimws(tolower(meta_data_cross_item[vg, ASSOCIATION_METRIC]))

          cor_range <-
            util_parse_interval(
              trimws(tolower(meta_data_cross_item[vg, ASSOCIATION_RANGE])))

          rvs <-
            vapply(
              util_parse_assignments(vl),
              identity, FUN.VALUE = character(1))

          rvs <- rvs[vapply(rvs, FUN = function(x) {
            !all(is.na(ds1[[x]]))
          }, FUN.VALUE = logical(1) )]
          if (length(rvs) == 0) {
            cors <- matrix(ncol = 0, nrow = 0)
          } else {
            cors <- cor(ds1[, rvs, FALSE], use = "complete.obs",
                        method = correlation_method)
          }

          cors_combs <-
            t(apply(expand.grid(rvs, rvs), 1, sort))

          cors_combs <-
            cors_combs[!duplicated(cors_combs), , drop = FALSE]

          cors_combs <-
            cors_combs[
              apply(cors_combs, 1, function(r) length(unique(r)) > 1), ,
              drop = FALSE
            ]

          if (!!prod(dim(cors_combs))) {
            cors_of_combs <- setNames(apply(cors_combs, 1, function(r) {
              cors[r[[1]], r[[2]]]
            }), nm = apply(cors_combs, 1, function(r) {
              sprintf("cor(%s, %s)", r[[1]], r[[2]])
            }))
          } else {
            cors_of_combs <- setNames(numeric(0), nm = character(0))
          }

          data.frame(VARIABLE_LIST = vl,
                     pretty_cors = prep_deparse_assignments(
                                                     labels = format(util_round_to_decimal_places(cors_of_combs)),
                                                     codes = names(cors_of_combs),
                                                     mode = "string_codes"),
                     pretty_max_cor = format(max(cors_of_combs, na.rm = TRUE)),
                     cors = prep_deparse_assignments(labels = cors_of_combs,
                                                    codes = names(cors_of_combs),
                                                    mode = "string_codes"),
                     max_cor = max(cors_of_combs, na.rm = TRUE),
                     pretty_in_range = redcap_env$`in`(max(cors_of_combs, na.rm = TRUE),
                                                cor_range),
                     pretty_range = as.character(cor_range),
                     in_range = redcap_env$`in`(max(cors_of_combs, na.rm = TRUE),
                                                       cor_range)
          )
          #            # max(cor()))
        }))
  VariableGroupData <- VariableGroupTable[, c(VARIABLE_LIST, grep("^pretty_.*$",
                                                                  names(VariableGroupTable), value = TRUE))]
  colnames(VariableGroupData) <- gsub("^pretty_", "", colnames(VariableGroupData))
  util_attach_attr(list(
    VariableGroupPlotList = VariableGroupPlotList,
    VariableGroupTable = VariableGroupTable[, !startsWith(colnames(VariableGroupTable), "pretty_")],
    VariableGroupData = VariableGroupData
  ), "as_plotly" = "util_as_plotly_des_scatterplot_matrix")
}

util_as_plotly_des_scatterplot_matrix <- function(dqr) {
  util_ensure_suggested("plotly")
  if (!identical(length(dqr$VariableGroupPlotList), 1L)) {
    print("FIXME") # FIXME
    plotly::plot_ly() %>%
      plotly::add_text(x = 0.5, y = 0.5,
                       text = "> 1 group in scatter plot matrix is not yet supported",
                       textfont = list(size = 24, color = "red"),
                       showlegend = FALSE) %>%
      plotly::layout(
        xaxis = list(showticklabels = FALSE, zeroline = FALSE, showgrid = FALSE),
        yaxis = list(showticklabels = FALSE, zeroline = FALSE, showgrid = FALSE),
        margin = list(l = 0, r = 0, t = 0, b = 0)
      )
  } else {
    attr(dqr$VariableGroupPlotList[[1]], "plotly")
  }
}

#' Check for unexpected data record set
#'
#' @description
#' This function tests that the identifiers match a provided record set. It is possible to
#' check for unexpected data record sets by study segments or to consider only selected
#' segments.
#'
#' @param level [character] a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").
#' @param id_vars_list [list] the list containing the identifier variables names to be used in the assessment.
#' @param identifier_name_list [list] the list that contains the name of the identifier to be used in the assessment. For the study level, corresponds to the names of the different data frames. For the segment level, indicates the name of the segments.
#' @param valid_id_table_list [list] the reference list with the identifier variable values.
#' @param meta_data_record_check_list [character] a character vector indicating the type of check to conduct, either "subset" or "exact".
#' @param study_data [data.frame] the data frame that contains the measurements, mandatory.
#' @param meta_data [data.frame] the data frame that contains metadata attributes of the study data, mandatory.
#'
#' @inheritParams .template_function_indicator
#'
#' @return a [list] with
#'   - `SegmentData`: data frame with the results of the quality check for unexpected data elements
#'   - `SegmentTable`: data frame with selected unexpected data elements check results, used for the data quality report.
#'   - `UnexpectedRecords`: vector with row indices of duplicated records, if any, otherwise NULL.
#'
#' @family integrity_indicator_functions
#' @concept integrity_indicator
#' @keywords internal
util_int_unexp_records_set_segment <- function(level = c("segment"),
                                  id_vars_list,
                                  identifier_name_list,
                                  valid_id_table_list,
                                  meta_data_record_check_list,
                                  study_data,
                                  label_col,
                                  meta_data, item_level,
                                  meta_data_segment = "segment_level",
                                  segment_level) {

  util_ck_arg_aliases()

  # 1. Segment level check ----

  # Checks arguments ----
  level <- util_match_arg(level)

  if (missing(id_vars_list) &&
      missing(identifier_name_list) &&
      missing(valid_id_table_list) &&
      missing(meta_data_record_check_list) &&
      missing(meta_data_segment) &&
      formals()$meta_data_segment %in% prep_list_dataframes()) {
    meta_data_segment <- force(meta_data_segment)
  }

  if (missing(id_vars_list) &&
      missing(identifier_name_list) &&
      missing(valid_id_table_list) &&
      missing(meta_data_record_check_list) &&
      !missing(meta_data_segment)) {
    meta_data_segment <- prep_check_meta_data_segment(meta_data_segment)

    meta_data_segment <-
      meta_data_segment[!util_empty(
        meta_data_segment[[SEGMENT_RECORD_CHECK]]), , drop = FALSE]
    # TODO: if nothing left
    id_vars_list <- lapply(setNames(meta_data_segment[[SEGMENT_ID_VARS]],
                                    nm = meta_data_segment[[STUDY_SEGMENT]]),
                           util_parse_assignments,
                           multi_variate_text = TRUE );
    id_vars_list <- lapply(id_vars_list, unlist, recursive = TRUE)
    id_vars_list <- lapply(setNames(meta_data_segment[[SEGMENT_ID_VARS]], # TODO: use the constants everywhere: meta_data_segment[[SEGMENT_ID_VARS]], not ...$SEGMENT_ID_VARS
                                    nm = meta_data_segment[[STUDY_SEGMENT]]),
                           util_parse_assignments,
                           multi_variate_text = TRUE
    )
    id_vars_list <- lapply(id_vars_list, unlist, recursive = TRUE)

    id_vars_list <- lapply(id_vars_list,
                                  util_map_labels,
                                  meta_data = meta_data,
                                  to = label_col)

    identifier_name_list <- meta_data_segment[[STUDY_SEGMENT]];
    valid_id_table_list <- meta_data_segment[[SEGMENT_ID_REF_TABLE]];
    meta_data_record_check_list <- meta_data_segment[[SEGMENT_RECORD_CHECK]];
  } else if (!missing(meta_data_segment)) {
    util_error(c("I have %s and one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "valid_id_table_list",
                   "meta_data_record_check_list"
                 )),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "valid_id_table_list",
                   "meta_data_record_check_list"
                 )))
  } else if (missing(meta_data_segment) && (
    missing(id_vars_list) ||
    missing(identifier_name_list) ||
    missing(valid_id_table_list) ||
    missing(meta_data_record_check_list)
  )) {
    util_error(c("I don't have %s and also miss at least",
                 "one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "valid_id_table_list",
                   "meta_data_record_check_list"
                 )),
               sQuote("meta_data_segment"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "valid_id_table_list",
                   "meta_data_record_check_list"
                 )))
  }

  .nrw <- length(identifier_name_list)

  util_stop_if_not(length(id_vars_list) == .nrw)
  util_stop_if_not(length(identifier_name_list) == .nrw)
  util_stop_if_not(length(valid_id_table_list) == .nrw)
  util_stop_if_not(length(meta_data_record_check_list) == .nrw)

  names(id_vars_list) <- identifier_name_list
  names(meta_data_record_check_list) <- identifier_name_list
  names(valid_id_table_list) <- identifier_name_list

  prep_prepare_dataframes()

  # meta_data$STUDY_SEGMENT <-
  #   util_map_labels(meta_data$STUDY_SEGMENT,
  #                   to = label_col,
  #                   meta_data = meta_data,
  #                   ifnotfound = meta_data$STUDY_SEGMENT)


  # Checks segments ----

  segments <- identifier_name_list

  # TODO: check
  if (missing(segments) || is.null(segments)) {
    # assign all segments
    meta_data_segment_reduced_record <- data.frame()
    segments <- meta_data_segment_reduced_record$STUDY_SEGMENT
  }

  # check that specified segments are included in the metadata
  old_segments <- segments
  segments <- intersect(segments, meta_data$STUDY_SEGMENT)

  if (length(old_segments) > length(segments)) {
    util_warning(
      "The segments in the %s do not match the segments in %s, considering only the intersection",
      dQuote("meta_data"),
      dQuote("meta_data_segment"),
      applicability_problem = TRUE,
      intrinsic_applicability_problem = TRUE
    )
  }

  # Check for unexpected records  ----
  names(meta_data_record_check_list) <- segments
  names(valid_id_table_list) <- segments

  result <- lapply(setNames(nm = segments), function(current_segment) {
    valid_id_table <- valid_id_table_list[[current_segment]]
    util_expect_data_frame(valid_id_table)

    vars_in_current_segment <- meta_data[
      meta_data[[STUDY_SEGMENT]] == current_segment,
      label_col,
      drop = TRUE]

    id_vars <- id_vars_list[[current_segment]]
    id_vars <- id_vars[!util_empty(id_vars)]

    if (length(id_vars) == 0) {
      util_warning(
        "No %d defined in %d, skipping the check for unexpected record set",
        dQuote("SEGMENT_ID_VARS"),
        dQuote("meta_data_segment"),
        applicability_problem = TRUE,
        intrinsic_applicability_problem = TRUE
      )
      return(
        res_pipeline <- data.frame(
          "Level" = "Segment",
          "NUM_int_sts_setrc" = 0,
          "PCT_int_sts_setrc" = 0,
          "GRADING" = 0,
          stringsAsFactors = FALSE
        )[FALSE, , drop = FALSE]
      )
    }

    id_vars <-
      util_ensure_in(id_vars,
                     colnames(ds1),
                     err_msg =
                       c(sprintf("ID variables in current segment %s",
                                 dQuote(current_segment)),
                         ": Missing %s from the study data,",
                         "did you mean %s? I'll remove the missing entries"))

    vars_in_current_segment <-
      util_ensure_in(vars_in_current_segment,
                     colnames(ds1),
                     err_msg =
                       c(sprintf("Study variables from current segment %s",
                                 dQuote(current_segment)),
                         ": Missing %s from the study data,",
                         "did you mean %s? I'll remove the missing entries"))


    # Select ids from variables in each segment from data
    data_ids_1 <- ds1[, c(id_vars, vars_in_current_segment)]
    data_ids <- util_remove_empty_rows(data_ids_1, id_vars = id_vars)

    if (current_segment %in% colnames(valid_id_table)) {
      metadata_ids <- valid_id_table[, current_segment]
    } else if ("ID" %in% colnames(valid_id_table)) {
      metadata_ids <- valid_id_table[, "ID"]
    } else {
      metadata_ids <- character(0)
    }
    metadata_ids <- metadata_ids[!util_empty(metadata_ids)]

    # TODO: implement
    if (length(id_vars) > 1) {
      util_warning(c("Check for multiple IDs is not currently supported,",
                     "but you could assign value labels to the each variable"),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      return(
        res_pipeline <- data.frame(
          "Level" = "Segment",
          "NUM_int_sts_setrc" = 0,
          "PCT_int_sts_setrc" = 0,
          "GRADING" = 0,
          stringsAsFactors = FALSE
        )[FALSE, , drop = FALSE]
      )

      # TODO: supported by reference ids written as c("9871 | 4567", "id | exdate")
    }

    # Check membership of the id vectors
    unex_records_tmp <- data_ids[[id_vars]][!(data_ids[[id_vars]] %in%
                                                metadata_ids)]

    miss_records_tmp <- metadata_ids[!(metadata_ids %in%
                                                data_ids[[id_vars]])]

    # TODO: check definition of exact and subset
    if (all(data_ids[[id_vars]] %in% metadata_ids) &&
        all(metadata_ids %in% data_ids[[id_vars]])) {
      match_actual <- "exact"
    } else if (all(data_ids[[id_vars]] %in% metadata_ids)) {
      match_actual <- "subset"
    } else if (all(metadata_ids %in% data_ids[[id_vars]])) {
      match_actual <- "superset"
    } else {
      match_actual <- "mismatch"
    }
    match_expected <- meta_data_record_check_list[[current_segment]]

    res_tmp <- data.frame(
      check.names = FALSE,
      "Check" = "Record set",
      "Segment" = current_segment,
      "Unexpected records in set?" = !(length(unex_records_tmp) == 0),
      "Number of records in data" = length(data_ids[[id_vars]]),
      "Number of records in metadata" = length(metadata_ids),
      "Number of mismatches" = length(unex_records_tmp),
      "Percentage of mismatches" =
        abs(round(100 * length(unex_records_tmp) / (
          length(data_ids[[id_vars]]) +
                                                       length(
                                                         metadata_ids)), 3)),
      "Expected match type" = match_expected,
      "Actual match type" = match_actual,
      "GRADING" = ifelse((match_expected[1] == match_actual), 0, 1),
      stringsAsFactors = FALSE
    )

    # TODO: not working when multiple id vars are used
    if (any(!util_empty(unex_records_tmp))) {
      vec_unex_records <- unex_records_tmp[!util_empty(unex_records_tmp)]
    } else {
      vec_unex_records <- NULL
    }

    mism_lin <- as.character(which(!(data_ids[[id_vars]] %in%
                                       metadata_ids)))

    return(list(res_tmp, data.frame(Segment = rep(current_segment,
                                            length(vec_unex_records)),
                                    UnexpectedID = vec_unex_records,
                                    Line = mism_lin)))
  })

  res_df <- do.call(rbind.data.frame, lapply(result, `[[`, 1))
  unex_records_df <- dplyr::bind_rows(lapply(result, `[[`, 2))

  res_pipeline <- data.frame(
    "Segment" = res_df$Segment,
    "NUM_int_sts_setrc" = res_df$`Number of mismatches`,
    "PCT_int_sts_setrc" = res_df$`Percentage of mismatches`,
    "GRADING" = res_df$GRADING,
    stringsAsFactors = FALSE
  )

  return(list(
    SegmentData = res_df,
    SegmentTable = res_pipeline,
    Other = unex_records_df
  ))

}

#' Verify assumptions made by the code, that must be TRUE
#'
#' @seealso [`stopifnot`]
#'
#' @param ... see [`stopifnot`]
#' @param label [character] a label for the assumptions, can be missing
#' @param label_only [logical] if `TRUE` and `label` is given, the condition
#'                                will not be displayed, if `FALSE`
#'
#' @return `invisible(FALSE)`, if not stopped.
#'
#'
#' @family robustness_functions
#' @concept condition
#' @noRd
util_stop_if_not <- function(..., label, label_only) { # FIXME: Strange ... problems in some cases
  cc <- rlang::current_call()
  # rlang::call_name(cl)
  # setdiff(rlang::call_args_names(rlang::caller_call(n = 0)), setdiff(names(formals(rlang::call_name(rlang::current_call()))), "..."))
  my_own <-
    setdiff(names(formals(rlang::call_name(cc))), "...")
  zappings <- rep(list(rlang::zap()), length(my_own))
  names(zappings) <- my_own
  mod_args <- c(list(.call = cc), zappings)
  cl <- do.call(rlang::call_modify, mod_args, quote = TRUE)
  cl[[1]] <- rlang::sym("stopifnot") # overwrites also <ns> in <ns>::
  ok <- try(eval.parent(cl), silent = TRUE)
  if (missing(label)) {
    label <- ""
  }
  if (missing(label_only)) {
    label_only <- TRUE
  }
  util_expect_scalar(label, check_type = is.character)
  util_expect_scalar(label_only, check_type = is.logical)
  if (inherits(ok, "try-error")) {
    cm <- conditionMessage(attr(ok, "condition"))
    if (label_only && nzchar(label)) cm <- label
    if (!label_only && nzchar(label) && nzchar(cm))
      cm <- paste0(label, ": ", cm)
    util_error(
      paste("Internal error:", gsub("%", "%%", fixed = TRUE, cm)),
      applicability_problem = FALSE)
  }
  invisible(ok)
}

#' Utility function for 3SD deviations rule
#'
#' This function calculates outliers according to the rule of 3SD deviations.
#'
#' @param x [numeric] data to check for outliers
#'
#' @return binary vector
#'
#' @importFrom stats sd
#'
#' @family outlier_functions
#' @concept outlier
#' @noRd
util_3SD <- function(x) {
  xmu <- mean(x, na.rm = TRUE)
  xsd <- sd(x, na.rm = TRUE)
  lth <- xmu - 3 * xsd
  uth <- xmu + 3 * xsd
  xbin <- ifelse(x < lth | x > uth, 1, 0)
  return(xbin)
}

#' Utility function to check whether a variable has no grouping variable assigned
#'
#' @inheritParams .template_function_indicator
#' @param resp_vars [variable list] the name of a measurement variable
#'
#' @return boolean
#'
#' @noRd
util_has_no_group_vars <- function(resp_vars,
                                   label_col = LABEL,
                                   meta_data = "item_level") {
  util_expect_scalar(resp_vars)
  util_expect_data_frame(meta_data)
  columns <- grep("^GROUP_VAR_.*", colnames(meta_data), value = TRUE)
  resp_vars <- util_find_var_by_meta(resp_vars,
                                     meta_data = meta_data,
                                     label_col = label_col)
  if (is.na(resp_vars)) { # nocov start
    return(TRUE) # likely not reachable because of util_expet_scalar, above
  } # nocov end
  rowSums(!util_empty(meta_data[meta_data[["VAR_NAMES"]] == resp_vars,
                                columns, FALSE])) == 0
}

#' Generate a full DQ report
#'
#' Deprecated
#'
#' @param ... Deprecated
#' @return Deprecated
#' @export
#' @importFrom stats setNames
dq_report <- function(...) { # nocov start
  lifecycle::deprecate_stop("2.1.0", what = "dq_report()", with = "dq_report2()")
} # nocov end

#' Check for unexpected data record set
#'
#' @description
#' This function tests that the identifiers match a provided record set. It is possible to
#' check for unexpected data record sets by study segments or to consider only selected
#' segments.
#'
#' @inheritParams .template_function_indicator
#' @param level [character] a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").
#' @param id_vars_list [list] the list containing the identifier variables names to be used in the assessment.
#' @param identifier_name_list [list] the list that contains the name of the identifier to be used in the assessment. For the study level, corresponds to the names of the different data frames. For the segment level, indicates the name of the segments.
#' @param valid_id_table_list [list] the reference list with the identifier variable values.
#' @param meta_data_record_check_list [character] a character vector indicating the type of check to conduct, either "subset" or "exact".
#' @param ... not used
#'
#' @return a [list] with
#'   - `SegmentData`: data frame with the results of the quality check for unexpected data elements
#'   - `SegmentTable`: data frame with selected unexpected data elements check results, used for the data quality report.
#'   - `UnexpectedRecords`: vector with row indices of duplicated records, if any, otherwise NULL.
#'
#' @family integrity_indicator_functions
#' @concept integrity_indicator
#' @keywords internal
util_int_unexp_records_set_dataframe <- function(level = c("dataframe"),
                                  id_vars_list,
                                  identifier_name_list,
                                  valid_id_table_list,
                                  meta_data_record_check_list,
                                  meta_data_dataframe = "dataframe_level",
                                  ...,
                                  dataframe_level) {

  util_ck_arg_aliases()

  # 1. Dataframe level check ----

  # Checks/Prepare arguments ----
  level <- util_match_arg(level)

  if (missing(id_vars_list) &&
      missing(identifier_name_list) &&
      missing(valid_id_table_list) &&
      missing(meta_data_record_check_list) &&
      missing(meta_data_dataframe) &&
      formals()$meta_data_dataframe %in% prep_list_dataframes()) {
    meta_data_dataframe <- force(meta_data_dataframe)
  }

  if (missing(id_vars_list) &&
      missing(identifier_name_list) &&
      missing(valid_id_table_list) &&
      missing(meta_data_record_check_list) &&
      !missing(meta_data_dataframe)) {
      meta_data_dataframe <- prep_check_meta_data_dataframe(meta_data_dataframe)
      meta_data_dataframe <- meta_data_dataframe[
        vapply(meta_data_dataframe[[DF_NAME]],
               function(x) { !util_is_try_error(try(prep_get_data_frame(data_frame_name = x,
                                                                    keep_types = TRUE), silent = TRUE)) },
               FUN.VALUE = logical(1))
        , , drop = FALSE]
      meta_data_dataframe <- meta_data_dataframe[
        vapply(meta_data_dataframe[[DF_ID_REF_TABLE]],
               function(x) { !util_is_try_error(try(prep_get_data_frame(data_frame_name = x), silent = TRUE)) },
               FUN.VALUE = logical(1))
        , , drop = FALSE]
      # TODO: if nothing left
      id_vars_list <- lapply(setNames(meta_data_dataframe[[DF_ID_VARS]],
                                      nm = meta_data_dataframe[[DF_NAME]]),
                             util_parse_assignments,
                             multi_variate_text = TRUE );
      id_vars_list <- lapply(id_vars_list, unlist, recursive = TRUE)
      identifier_name_list <- meta_data_dataframe[[DF_NAME]];
      valid_id_table_list <- meta_data_dataframe[[DF_ID_REF_TABLE]];
      meta_data_record_check_list <- meta_data_dataframe[[DF_RECORD_CHECK]];
  } else if (!missing(meta_data_dataframe)) {
    util_error(c("I have %s and one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "valid_id_table_list",
                   "meta_data_record_check_list"
                   )),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "valid_id_table_list",
                   "meta_data_record_check_list"
                 )))
  } else if (missing(meta_data_dataframe) && (
    missing(id_vars_list) ||
    missing(identifier_name_list) ||
    missing(valid_id_table_list) ||
    missing(meta_data_record_check_list)
  )) {
    util_error(c("I don't have %s and also miss at least",
                 "one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "valid_id_table_list",
                   "meta_data_record_check_list"
                 )),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "valid_id_table_list",
                   "meta_data_record_check_list"
                 )))
  }
  .nrw <- length(identifier_name_list)

  util_stop_if_not(length(id_vars_list) == .nrw)
  util_stop_if_not(length(identifier_name_list) == .nrw)
  util_stop_if_not(length(valid_id_table_list) == .nrw)
  util_stop_if_not(length(meta_data_record_check_list) == .nrw)

  names(id_vars_list) <- identifier_name_list
  names(meta_data_record_check_list) <- identifier_name_list
  names(valid_id_table_list) <- identifier_name_list

  # Check for unexpected records  ----

  result <- lapply(setNames(nm = identifier_name_list), function(current_df) {

    # Convert data from list to data frame
    # data_current_df <- do.call(rbind.data.frame, study_data[current_df])
    data_current_df <- util_expect_data_frame(current_df, dont_assign = TRUE)

    valid_id_table <- valid_id_table_list[[current_df]]
    util_expect_data_frame(valid_id_table)

    id_vars <- id_vars_list[[current_df]] # TODO: id_vars may be unavailable.
    id_vars <- id_vars[!util_empty(id_vars)]

    if (length(id_vars) == 0) {
      util_warning(
        "No %s defined in %s, skipping the check for unexpected record set",
        dQuote("DF_ID_VARS"),
        dQuote("meta_data_studies"),
        applicability_problem = TRUE,
        intrinsic_applicability_problem = TRUE
      )
      return(
        res_pipeline <- data.frame(
          "Level" = "Data frame",
          "NUM_int_sts_setrc" = 0,
          "PCT_int_sts_setrc" = 0,
          "GRADING" = 0,
          stringsAsFactors = FALSE
        )[FALSE, , drop = FALSE]
      )
    }

    # Select ids from data
    data_ids <- util_remove_empty_rows(data_current_df, id_vars = id_vars)

    metadata_ids <- valid_id_table
    metadata_ids <- metadata_ids[!util_empty(metadata_ids)]

    if (length(id_vars) > 1) {
      util_warning(c("Check for multiple IDs is not currently supported,",
                     "but you could assign value labels to the each variable"),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      return(
        res_pipeline <- data.frame(
          "Level" = "Data frame",
          "NUM_int_sts_setrc" = 0,
          "PCT_int_sts_setrc" = 0,
          "GRADING" = 0,
          stringsAsFactors = FALSE
        )[FALSE, , drop = FALSE]
      )

      # IDEA: supported by reference ids written as c("9871 | 4567", "id | exdate")
    }

    # Check membership of the id vectors
    unex_records_tmp <- data_ids[[id_vars]][!(data_ids[[id_vars]] %in%
                                                metadata_ids)]
    miss_records_tmp <- metadata_ids[!(metadata_ids %in%
                                         data_ids[[id_vars]])]

    # TODO: check definition of exact and subset
    if (all(data_ids[[id_vars]] %in% metadata_ids) &&
        all(metadata_ids %in% data_ids[[id_vars]])) {
      match_actual <- "exact"
    } else if (all(data_ids[[id_vars]] %in% metadata_ids)) {
      match_actual <- "subset"
    } else if (all(metadata_ids %in% data_ids[[id_vars]])) {
      match_actual <- "superset"
    } else {
      match_actual <- "mismatch"
    }
    match_expected <- meta_data_record_check_list[[current_df]]

    res_tmp <- data.frame(
      check.names = FALSE,
      "Check" = "Record set",
      "Data frame" = current_df,
      "Unexpected records in set?" = !(length(unex_records_tmp) == 0),
      "Number of records in data" = length(data_ids[, id_vars]),
      "Number of records in metadata" = length(metadata_ids),
      "Number of mismatches" =
        length(unex_records_tmp),
      "Percentage of mismatches" =
        abs(round(100 * length(unex_records_tmp) / length(metadata_ids), 3)),
      "Expected match type" = match_expected,
      "Actual match type" = match_actual,
      "GRADING" = ifelse((match_expected == match_actual), 0, 1),
      stringsAsFactors = FALSE
    )

    if (any(!util_empty(unex_records_tmp))) {
      vec_unex_records <- unex_records_tmp[!util_empty(unex_records_tmp)]
    } else {
      vec_unex_records <- NULL
    }

    mism_lin <- as.character(which(!(data_ids[[id_vars]] %in%
                                   metadata_ids)))

    return(list(res_tmp, data.frame(Dataframe = rep(current_df,
                                            length(vec_unex_records)),
                                    UnexpectedID = vec_unex_records,
                                    Line = mism_lin)))
  })

  res_df <- do.call(rbind.data.frame, lapply(result, `[[`, 1))
  unex_records_df <- dplyr::bind_rows(lapply(result, `[[`, 2))

  res_pipeline <- data.frame(
    "Level" = "Dataframe",
    "DF_NAME" = res_df[, 2],
    "NUM_int_sts_setrc" = res_df[, 6],
    "PCT_int_sts_setrc" = res_df[, 7],
    "GRADING" = res_df$GRADING,
    stringsAsFactors = FALSE
  )

  return(list(
    DataframeData = res_df,
    DataframeTable = res_pipeline,
    Other = unex_records_df
  ))

}

#' Estimate marginal means, see [emmeans::emmeans]
#'
#' @description
#' This function examines the impact of so-called process variables on a
#' measurement variable. This implementation combines a descriptive and a
#' model-based approach. Process variables that can be considered in this
#' implementation must be categorical. It is currently not possible to
#' consider more than one process variable within one function call.
#' The measurement variable can be adjusted for (multiple) covariables, such as
#' age or sex, for example.
#'
#' Marginal means rests on model-based results, i.e. a significantly different
#' marginal mean depends on sample size. Particularly in large studies, small
#' and irrelevant differences may become significant. The contrary holds if
#' sample size is low.
#'
#' [Indicator]
#'
#' @details
#' Limitations
#'
#' Selecting the appropriate distribution is complex. Dozens of continuous,
#' discrete or mixed distributions are conceivable in the context of
#' epidemiological data. Their exact exploration is beyond the scope of this
#' data quality approach. The present function uses the help function
#' `util_dist_selection`, the assigned `SCALE_LEVEL` and the `DATA_TYPE`
#' to discriminate the following cases:
#' \itemize{
#'   \item continuous data
#'   \item binary data
#'   \item count data with <= 20 distinct values
#'   \item count data with > 20 distinct values (treated as continuous)
#'   \item nominal data
#'   \item ordinal data
#'  }
#' Continuous data and count data with more than 20 distinct values are analyzed
#' by linear models. Count data with up to 20 distinct values are modeled by a
#' Poisson regression. For binary data, the implementation uses logistic
#' regression.
#' Nominal response variables will either be transformed to binary variables or
#' analyzed by multinomial logistic regression models. The latter option is only
#' available if the argument `dichotomize_categorical_resp` is set to `FALSE`
#' and if the package `nnet` is installed. The transformation to a binary
#' variable can be user-specified using the metadata columns `RECODE_CASES`
#' and/or `RECODE_CONTROL`. Otherwise, the most frequent category will be
#' assigned to cases and the remaining categories to control.
#' For ordinal response variables, the argument `cut_off_linear_model_for_ord`
#' controls whether the data is analyzed in the same way as continuous data:
#' If every level of the variable has at least as many observations as specified
#' in the argument, the data will be analyzed by a linear model. Otherwise,
#' the data will be modeled by a ordered regression, if the package `ordinal`
#' is installed.
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars  [variable] the name of the measurement variable
#' @param group_vars [variable list] len=1-1. the name of the observer, device
#'                                   or reader variable
#' @param co_vars [variable list] a vector of covariables, e.g. age and sex for
#'                              adjustment
#' @param threshold_type [enum] empirical | user | none. In case `empirical` is
#'                       chosen, a multiplier of the scale measure is used.
#'                       In case of `user`, a value of the mean or probability
#'                       (binary data) has to be defined
#'                       see [`Implementation and use of thresholds`](https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_margins.html#Implementation_and_use_of_thresholds) in the  online documentation).
#'                       In case of `none`, no thresholds are displayed and no
#'                       flagging of unusual group levels is applied.
#' @param threshold_value [numeric] a multiplier or absolute value (see
#'                       [`Implementation and use of thresholds`](https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_margins.html#Implementation_and_use_of_thresholds) in the
#'                       online documentation).
#' @param min_obs_in_subgroup [integer] from=0. This optional argument specifies
#'                       the minimum number of observations that is required to
#'                       include a subgroup (level) of the `group_var` in the
#'                       analysis. Subgroups with less observations are
#'                       excluded.
#' @param min_obs_in_cat [integer] This optional argument specifies the minimum
#'                        number of observations that is required to include
#'                        a category (level) of the outcome (`resp_vars`) in
#'                        the analysis. Categories with less observations are
#'                        combined into one group. If the collapsed category
#'                        contains less observations than required, it will be
#'                        excluded from the analysis.
#' @param dichotomize_categorical_resp [logical] Should nominal response
#'                        variables always be transformed to binary variables?
#' @param cut_off_linear_model_for_ord [integer] from=0. This optional argument
#'                        specifies the minimum number of observations for
#'                        individual levels of an ordinal outcome (`resp_var`)
#'                        that is required to run a linear model instead of an
#'                        ordered regression (i.e., a cut-off value above which
#'                        linear models are considered a good approximation).
#'                        The argument can be set to `NULL` if ordered
#'                        regression models are preferred for ordinal data in
#'                        any case.
#' @param sort_group_var_levels [logical] Should the levels of the grouping
#'                        variable be sorted descending by the number of
#'                        observations? Note that ordinal grouping variables
#'                        will not be reordered.
#' @param include_numbers_in_figures [logical] Should the figure report the
#'                        number of observations for each level of the grouping
#'                        variable?
#' @param n_violin_max [integer] from=0. This optional argument specifies
#'                       the maximum number of levels of the `group_var` for
#'                       which violin plots will be shown in the figure.
#'
#' @param no_overall_in_bin [logical] Suppress overall distribution in 'margins'
#'                                    figures for binary outcomes
#' @param no_geom_count_in_bin [logical] Suppress counts 'margins'
#'                                       figures for binary outcomes, so they
#'.                                      are not always including 0 and 1.
#' @return a list with:
#'   - `SummaryTable`: [data.frame] underlying the plot
#'   - `ResultData`: [data.frame]
#'   - `SummaryPlot`: [ggplot2::ggplot()] margins plot
#'
#' @export
#'
#' @importFrom utils tail head
#'
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_margins.html
#' )
acc_margins <- function(resp_vars = NULL,
                        group_vars = NULL,
                        co_vars = NULL,
                        study_data,
                        label_col,
                        item_level = "item_level",
                        threshold_type = "empirical",
                        threshold_value,
                        min_obs_in_subgroup = 5,
                        min_obs_in_cat = 5,
                        dichotomize_categorical_resp = TRUE,
                        cut_off_linear_model_for_ord = 10,
                        meta_data = item_level,
                        meta_data_v2,
                        sort_group_var_levels =
                          getOption("dataquieR.acc_margins_sort",
                                    dataquieR.acc_margins_sort_default),
                        include_numbers_in_figures =
                          getOption("dataquieR.acc_margins_num",
                                    dataquieR.acc_margins_num_default),
                        n_violin_max =
                          getOption("dataquieR.max_group_var_levels_with_violins",
                                    dataquieR.max_group_var_levels_with_violins_default),
                        no_overall_in_bin =
                          getOption("dataquieR.no_overall_in_bin",
                                    dataquieR.no_overall_in_bin_default),
                        no_geom_count_in_bin =
                          getOption("dataquieR.no_geom_count_in_bin",
                                    dataquieR.no_geom_count_in_bin_default)) { # TODO: flip_mode =
  #prep for meta_data_v2
  util_maybe_load_meta_data_v2()

  # to avoid "no visible binding for global variable ‘sample_size’"
  sample_size <- NULL
  # preps ----------------------------------------------------------------------
  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata = TRUE)

  util_correct_variable_use("resp_vars",
                            need_scale = "!na",
                            min_distinct_values = 2)

  util_correct_variable_use("group_vars",
                            allow_any_obs_na = TRUE,
                            need_type = "!float",
                            need_scale = "nominal | ordinal",
                            min_distinct_values = 2)

  util_correct_variable_use("co_vars",
                            overwrite = TRUE,
                            remove_not_found = TRUE,
                            allow_more_than_one = TRUE,
                            allow_all_obs_na = FALSE,
                            allow_null = TRUE,
                            allow_na = TRUE,
                            allow_any_obs_na = TRUE)

  util_disjunct_var_sets(resp_vars, co_vars, group_vars)

  co_vars <- na.omit(co_vars)
  if (is.null(co_vars)) {
    co_vars <- character(0)
  }

  # replace dollar signs for emmeans
  dollar <- "\uFE69"
  colnames(ds1) <- gsub("$", dollar, fixed = TRUE, colnames(ds1))
  if (any(grepl('$', fixed = TRUE, c(resp_vars, group_vars, co_vars)))) {
    util_message(c("emmeans used by acc_margins does not support variable",
                   "names containing %s, replacing this symbol by an",
                   "equivalent unicode character %s"),
                 dQuote("$"),
                 dQuote(dollar), applicability_problem = TRUE,
                 intrinsic_applicability_problem = FALSE)
  }
  original_resp_vars <- resp_vars
  original_group_vars <- group_vars
  original_co_vars <- co_vars
  resp_vars <- gsub("$", dollar, fixed = TRUE, resp_vars)
  group_vars <- gsub("$", dollar, fixed = TRUE, group_vars)
  co_vars <- gsub("$", dollar, fixed = TRUE, co_vars)

  if (length(setdiff(co_vars, "1")) > 0) {
    lb <- prep_get_labels(original_co_vars,
                          item_level = meta_data,
                          label_col = label_col,
                          label_class = "LONG",
                          resp_vars_match_label_col_only = TRUE)
    if (length(lb) < 4) {
      adjusted_hint <- sprintf("adjusted for %s", paste0(lb, collapse = ", "))
    } else {
      adjusted_hint <- sprintf("adjusted for %d variables", length(lb))
    }
  } else {
    adjusted_hint <- ""
  }

  title <- paste(prep_get_labels(original_group_vars,
                                 item_level = meta_data,
                                 label_col = label_col,
                                 label_class = "LONG",
                                 resp_vars_match_label_col_only = TRUE),
                 "margins in",
                 prep_get_labels(original_resp_vars,
                                 item_level = meta_data,
                                 label_col = label_col,
                                 label_class = "LONG",
                                 resp_vars_match_label_col_only = TRUE))

  util_expect_scalar(min_obs_in_subgroup,
                     check_type = util_is_numeric_in(min = 5,
                                                     whole_num = TRUE,
                                                     finite = TRUE),
                     convert_if_possible = function(x) {
                       x1 <- suppressWarnings(as.integer(x))
                       if (is.na(x1) ||
                           !util_is_numeric_in(min = 5, whole_num = TRUE,
                                               finite = TRUE)(x1)) {
                         x1 <- as.integer(5)
                         util_message(
                           paste("min_obs_in_subgroup is not specified",
                                 "correctly and is set to 5 instead."),
                           applicability_problem = TRUE)
                       }
                       x1
                     })

  util_expect_scalar(min_obs_in_cat,
                     check_type = util_is_numeric_in(min = 1,
                                                     whole_num = TRUE,
                                                     finite = TRUE))

  util_expect_scalar(sort_group_var_levels, check_type = is.logical)
  if (meta_data[[SCALE_LEVEL]][
    meta_data[[label_col]] == original_group_vars] %in% SCALE_LEVELS$ORDINAL) {
    sort_group_var_levels <- FALSE
  }
  util_expect_scalar(include_numbers_in_figures, check_type = is.logical)
  util_expect_scalar(n_violin_max,
                     check_type = util_is_numeric_in(min = 0))

  util_expect_scalar(no_overall_in_bin, check_type = is.logical)
  util_expect_scalar(no_geom_count_in_bin, check_type = is.logical)

  # omit missing values and unnecessary variables
  n_prior <- nrow(ds1)
  ds1 <- ds1[, c(resp_vars, group_vars, co_vars), drop = FALSE]
  ds1 <- ds1[complete.cases(ds1[, c(group_vars, co_vars)]), ]
  n_post <- nrow(ds1)
  msg <- NULL
  if (n_post < n_prior) {
    msg <- paste0(
      "Due to missing values in ",
      ifelse(length(co_vars) > 0,
             paste(paste0(co_vars, collapse = ", "), "or "),
             ""),
      group_vars, ", N = ", n_prior - n_post,
      " observations were excluded. "
    )
  }
  n_prior <- n_post
  ds1 <- ds1[complete.cases(ds1), ]
  n_post <- nrow(ds1)
  if (n_post < n_prior) {
    msg <- paste0(
      msg, "Due to missing values in ", resp_vars, ", N = ",
      n_prior - n_post, " observations were excluded",
      ifelse(nchar(msg) > 0, " additionally.", "."))
  }

  if (length(msg) > 0 && nchar(msg) > 0) {
    util_message(trimws(msg),
                 applicability_problem = FALSE)
  }

  if (!(prod(dim(ds1)))) {
    util_error("No data left after data preparation.")
  }

  # ensure that included levels of the grouping variable have the specified
  # minimum number of observations
  check_df <- util_table_of_vct(ds1[[group_vars]])
  critical_levels <- levels(check_df$Var1)[check_df$Freq <
                                             min_obs_in_subgroup]
  if (length(critical_levels) > 0) {
    util_message("Levels %s were excluded due to less than %d observations.",
                 paste0(c(vapply(head(critical_levels, 10), dQuote, ""),
                          if (length(critical_levels) <= 10)
                            character(0) else "..."),
                        collapse = ", "),
                 min_obs_in_subgroup,
                 applicability_problem = FALSE)
    # exclude levels with too few observations
    ds1 <- ds1[!(ds1[[group_vars]] %in% critical_levels), ]
    levels(ds1[[group_vars]])[
      which(levels(ds1[[group_vars]]) %in% critical_levels)] <- NA
  }
  ds1 <- ds1[complete.cases(ds1), , drop = FALSE]
  if (nrow(ds1) == 0) {
    util_error("No data left after data preparation.")
  }

  if (!missing(threshold_value)) {
    if (is.vector(threshold_value)) {
      .threshold_value <- as.numeric(threshold_value)
    } else {
      .threshold_value <- NA
    }
    if (length(threshold_value) != 1 || is.na(.threshold_value)) {
      util_message(
        "threshold_value is not numeric(1): %s, setting it to default value 1.",
        dQuote(head(try(as.character(threshold_value)), 1)),
        applicability_problem = TRUE)
      threshold_value <- 1
    } else {
      threshold_value <-
        .threshold_value
    }
  }

  if (is.null(threshold_type) || (!is.list(threshold_type)
                                  && length(threshold_type) != 1)) {
    if (
      !is.null(threshold_type)
      ||
      !.called_in_pipeline
    ) util_message("No or many threshold type specified and set to empirical.",
                   applicability_problem = TRUE)
    threshold_type <- "empirical"
  }

  threshold_type <- match.arg(threshold_type, c("empirical", "user", "none"))

  # no relative distance (based on SD) to mean defined?
  if (threshold_type %in% c("empirical", "none") & missing(threshold_value)) {
    threshold_value <- 1
  }

  # threshold is user but no value defined -> switch to empirical
  if (threshold_type == "user" & missing(threshold_value)) {
    util_message(
      c(
        "Threshold was set to user but no value for the unit of measurements",
        "was defined.\n",
        "The function switches to one SD as default."),
      applicability_problem = TRUE)
    threshold_type == "empirical"
    threshold_value <- 1
  }

  # check first if there is a user-specified recoding
  if (("RECODE_CASES" %in% colnames(meta_data) &&
       !util_empty(meta_data[meta_data[[label_col]] == resp_vars, RECODE_CASES])) ||
      ("RECODE_CONTROL" %in% colnames(meta_data) &&
       !util_empty(meta_data[meta_data[[label_col]] == resp_vars, RECODE_CONTROL]))) {
    rvs_bin <- util_dichotomize(
      study_data = ds1[, resp_vars, drop = FALSE],
      meta_data = meta_data,
      label_col = label_col)
    rvs_bin_note <- attr(rvs_bin, "Dichotomization")[[resp_vars]]
    ds1[[resp_vars]] <- unlist(rvs_bin)
    ds1 <- ds1[complete.cases(ds1), , drop = FALSE]
  }

  var_prop <- util_dist_selection(ds1[, resp_vars, drop = FALSE])
  if (var_prop$NDistinct < 2) {
    util_error("The response variable is constant after data preparation.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  # call utility functions based on scale and data properties ------------------
  var_scale <- meta_data[[SCALE_LEVEL]][meta_data[[label_col]] == resp_vars]
  var_dtype <- meta_data[[DATA_TYPE]][meta_data[[label_col]] == resp_vars]

  ###1st CASE: there are only 2 distinct values
  if (var_prop$NDistinct == 2) {
    if (nrow(ds1) < 2 * min_obs_in_cat) {
      util_error("Not enough data (after data preparation).",
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    }
    # recode binary variable to 0/1, if needed
    if (!all(unique(ds1[[resp_vars]]) %in% c(0, 1))) {
      bin_codes <- names(sort(table(ds1[[resp_vars]])))
      mf <- tail(bin_codes, 1)
      lf <- head(bin_codes, 1)
      # https://stackoverflow.com/questions/12187187/
      #     how-to-retrieve-the-most-repeated-value-in-a-
      #                                         column-present-in-a-data-frame
      col <- NA
      col[ds1[[resp_vars]] == mf] <- 0
      col[ds1[[resp_vars]] == lf] <- 1
      ds1[[resp_vars]] <- col
      rvs_bin_note <- paste(
        paste("Cases (1):", lf),
        paste("Control (0):", mf),
        sep = ". ")
    } else {
      rvs_bin_note <- "Cases (1): 1. Control (0): 0"
    }
    # run margins function for binary response
    mar_out <- util_margins_bin(resp_vars = resp_vars,
                                group_vars = group_vars,
                                co_vars = co_vars,
                                threshold_type = threshold_type,
                                threshold_value = threshold_value,
                                min_obs_in_subgroup = min_obs_in_subgroup,
                                min_obs_in_cat = min_obs_in_cat,
                                caption = rvs_bin_note,
                                ds1 = ds1,
                                label_col = label_col,
                                adjusted_hint = adjusted_hint,
                                title = title,
                                sort_group_var_levels = sort_group_var_levels,
                                include_numbers_in_figures =
                                  include_numbers_in_figures,
                                no_overall_in_bin = no_overall_in_bin,
                                no_geom_count_in_bin = no_geom_count_in_bin)

    .plot1 <- mar_out$plot
    obj1 <- util_create_lean_ggplot(
      ggplot2::ggplot_build(.plot1),
      .plot1  = .plot1
    )
    obj1_data <- util_rbind(
      data_frames_list = util_gg_get(obj1, "data")
    )
    min_value <- min(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
    max_value <- max(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
    range_values <- max_value - min_value

    no_char_y <- nchar(range_values)
    rm(obj1, obj1_data)
    type_plot <- "count_plot"


    ###2nd CASE: NOMINAL (2 possible results)
  } else if (var_scale == SCALE_LEVELS$NOMINAL) {
    if (nrow(ds1) < 2 * min_obs_in_cat) {
      util_error("Not enough data (after data preparation).",
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    }
    # Nominal response variables will either be transformed to binary variables
    # or analyzed by multinomial logistic regression models.
    count_nom <- util_table_of_vct(ds1[[resp_vars]])
    count_nom <- count_nom[which(count_nom[, 2] > 0), ]
    count_nom <- count_nom[order(count_nom[, 2], decreasing = TRUE), ]
    count_nom$below_thresh <- count_nom[, 2] < min_obs_in_cat
    # catch cases were the nominal response variable has to be analyzed as a
    # binary variable:
    if (nrow(count_nom) == 2 | # i.e., there are only two categories
        length(which(count_nom$below_thresh)) >= nrow(count_nom) - 1 |
        # i.e., (almost) all categories have too few observations to be
        # analyzed individually
        (sum(count_nom[which(count_nom$below_thresh), 2]) < min_obs_in_cat &
         length(which(!count_nom$below_thresh)) == 2)
        # If we would collapse all rare categories, they would still have too
        # few observations to form a third category. Thus, we can only analyse
        # two categories individually.
    ) {
      dichotomize_categorical_resp <- TRUE
    }

    if (dichotomize_categorical_resp) {
      # Dichotomized nominal variables will be analyzed by logistic models.
      # The dichotomization can be user-defined in the metadata (RECODE_CASES
      # and/or RECODE_CONTROL). If not, dichotomization will be performed
      # automatically (if 'dichotomize_categorical_resp' is TRUE).
      if (!("RECODE_CASES" %in% colnames(meta_data))) {
        meta_data[[RECODE_CASES]] <- ""
      }
      if (!("RECODE_CONTROL" %in% colnames(meta_data))) {
        meta_data[[RECODE_CONTROL]] <- ""
      }
      if (util_empty(meta_data[[RECODE_CASES]][meta_data[[label_col]] == resp_vars]) &
          util_empty(meta_data[[RECODE_CONTROL]][meta_data[[label_col]] == resp_vars])) {
        # If the recoding is not defined in the metadata, dataquieR will use the
        # most frequent category as 'cases', the remaining categories as
        # 'control'. If there were too few observations in the most frequent
        # category for the 'cases', we will include further categories until we
        # reach the lower limit specified by 'min_obs_in_cat' (but we stop it
        # before all categories are being combined into one group).
        ind_cases <- 1
        n_cases <- count_nom[1, 2]
        while (n_cases < min_obs_in_cat & max(ind_cases) < nrow(count_nom) - 1) {
          ind_cases <- seq_len(max(ind_cases) + 1)
          n_cases <- sum(count_nom[ind_cases, 2])
        }
        meta_data[[RECODE_CASES]][meta_data[[label_col]] == resp_vars] <-
          paste(count_nom[ind_cases, 1], collapse = " | ")
      }
      rvs_bin <- util_dichotomize(
        study_data = ds1[, resp_vars, drop = FALSE],
        meta_data = meta_data,
        label_col = label_col)
      rvs_bin_note <- attr(rvs_bin, "Dichotomization")[[resp_vars]]
      ds1[[resp_vars]] <- unlist(rvs_bin)
      ds1 <- ds1[complete.cases(ds1), ]
      # run margins function for binary response
      mar_out <- util_margins_bin(resp_vars = resp_vars,
                                  group_vars = group_vars,
                                  co_vars = co_vars,
                                  threshold_type = threshold_type,
                                  threshold_value = threshold_value,
                                  min_obs_in_subgroup = min_obs_in_subgroup,
                                  min_obs_in_cat = min_obs_in_cat,
                                  caption = rvs_bin_note,
                                  ds1 = ds1,
                                  label_col = label_col,
                                  adjusted_hint = adjusted_hint,
                                  title = title,
                                  sort_group_var_levels = sort_group_var_levels,
                                  include_numbers_in_figures =
                                    include_numbers_in_figures,
                                  no_overall_in_bin = no_overall_in_bin,
                                  no_geom_count_in_bin = no_geom_count_in_bin)

      .plot2 <- mar_out$plot
      obj1 <- util_create_lean_ggplot(
        ggplot2::ggplot_build(.plot2),
        .plot2 = .plot2
      )
      obj1_data <- util_rbind(
        data_frames_list = util_gg_get(obj1, "data")
      )
      min_value <- min(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
      max_value <- max(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
      range_values <- max_value - min_value
      no_char_y <- nchar(range_values)
      rm(obj1, obj1_data)
      type_plot <- "count_plot"

    } else { # more than two categories to be considered in the analysis
      # Rare categories will either be collapsed or discarded.
      if (any(count_nom$below_thresh)) {
        crit_lev_ind <- which(count_nom[, 2] < min_obs_in_cat)
        critical_levels <- count_nom[crit_lev_ind, 1]
        if (sum(count_nom[crit_lev_ind, 2]) < min_obs_in_cat) {
          util_message(paste0(c(
            "The following levels:", head(critical_levels, 100),
            if (length(critical_levels) > 100)  {", ..." }, "have <",
            min_obs_in_cat, " observations and will be discarded."
          ),
          collapse = " "
          ), applicability_problem = FALSE)
          levels(ds1[[resp_vars]])[
            which(levels(ds1[[resp_vars]]) %in% critical_levels)] <- NA
          ds1 <- ds1[!is.na(ds1[[resp_vars]]), ]
        } else {
          util_message(paste0(c(
            "The following levels:", head(critical_levels, 100),
            if (length(critical_levels) > 100)  {", ..." }, "have <",
            min_obs_in_cat, " observations and will be collapsed."
          ),
          collapse = " "
          ), applicability_problem = FALSE)
          new_lev <- "other"
          # ensure that the new category is not yet present
          if (new_lev %in% levels(ds1[[resp_vars]])) {
            new_lev <- "other (collapsed)"
          }
          while (new_lev %in% levels(ds1[[resp_vars]])) {
            new_lev <- paste0(
              "other collapsed_",
              paste0(sample(c(letters, LETTERS), replace = TRUE,
                            size = length(levels(ds1[[resp_vars]])) + 1),
                     collapse = ""))
          }
          levels(ds1[[resp_vars]])[
            which(levels(ds1[[resp_vars]]) %in% critical_levels)] <- new_lev
        }
      }
      # run margins function for response with more than two categories
      mar_out <- util_margins_nom(resp_vars = resp_vars,
                                  group_vars = group_vars,
                                  co_vars = co_vars,
                                  min_obs_in_subgroup = min_obs_in_subgroup,
                                  min_obs_in_cat = min_obs_in_cat,
                                  ds1 = ds1,
                                  label_col = label_col,
                                  adjusted_hint = adjusted_hint,
                                  title = title,
                                  sort_group_var_levels = sort_group_var_levels)

      .plot3 <- mar_out$plot
      obj1 <- util_create_lean_ggplot(
        ggplot2::ggplot_build(.plot3),
        .plot3 = .plot3
      )
      obj1_data <- util_gg_get(obj1, "data")
      n_groups <- max(obj1_data[[2]]$group) * nrow(count_nom)
      min_value <- min(c(obj1_data[[1]]$xmin, obj1_data[[1]]$xmax),  na.rm = TRUE)
      max_value <- max(c(obj1_data[[1]]$xmin, obj1_data[[1]]$xmax),  na.rm = TRUE)
      range_values <- max_value - min_value
      no_char_y <- nchar(round(range_values, digits = 2))
      rm(obj1, obj1_data)
      type_plot <- "rotated_plot"
    }
    ###3rd CASE: ORDINAL (2 possible results)
  } else if (var_scale == SCALE_LEVELS$ORDINAL) {
    # Ordinal response variables will either be analyzed by a linear model
    # or by mixed effects ordered logistic models.
    count_ord <- util_table_of_vct(ds1[[resp_vars]])
    if (!is.null(cut_off_linear_model_for_ord) &&
        all(count_ord[, 2] >= cut_off_linear_model_for_ord)) {
      orig_levels <- levels(ds1[[resp_vars]])
      names(orig_levels) <- seq_along(orig_levels) - 1
      ds1[[resp_vars]] <- as.numeric(ds1[[resp_vars]]) - 1

      mar_out <- util_margins_lm(resp_vars = resp_vars,
                                 group_vars = group_vars,
                                 co_vars = co_vars,
                                 threshold_type = threshold_type,
                                 threshold_value = threshold_value,
                                 min_obs_in_subgroup = min_obs_in_subgroup,
                                 ds1 = ds1,
                                 label_col = label_col,
                                 levels = orig_levels,
                                 adjusted_hint = adjusted_hint,
                                 title = title,
                                 sort_group_var_levels = sort_group_var_levels,
                                 include_numbers_in_figures =
                                   include_numbers_in_figures,
                                 n_violin_max = n_violin_max)

      .plot4 <- mar_out$plot
      obj1 <- util_create_lean_ggplot(
        ggplot2::ggplot_build(.plot4),
        .plot4 = .plot4
      )
      obj1_data <- util_rbind(
        data_frames_list = util_gg_get(obj1, "data")
      )
      min_value <- min(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
      max_value <- max(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
      range_values <- max_value - min_value
      a <- orig_levels
      names(a) <- NULL
      no_char_y <- max(nchar(a))
      rm(a)
      rm(obj1, obj1_data)
      type_plot <- "violin_plot"

    } else {
      title <- paste("Conditional modes of",
                     prep_get_labels(original_group_vars,
                                     item_level = meta_data,
                                     label_col = label_col,
                                     label_class = "LONG",
                                     resp_vars_match_label_col_only = TRUE),
                     "levels for",
                     prep_get_labels(original_resp_vars,
                                     item_level = meta_data,
                                     label_col = label_col,
                                     label_class = "LONG",
                                     resp_vars_match_label_col_only = TRUE))

      mar_out <- util_margins_ord(resp_vars = resp_vars,
                                  group_vars = group_vars,
                                  co_vars = co_vars,
                                  min_obs_in_subgroup = min_obs_in_subgroup,
                                  min_subgroups = 4, # specific requirement from 'ordinal'
                                  ds1 = ds1,
                                  label_col = label_col,
                                  adjusted_hint = adjusted_hint,
                                  title = title,
                                  sort_group_var_levels = sort_group_var_levels)

      n_groups <- length(mar_out$plot_data$group)
      no_char_x <- max(nchar(as.character(mar_out$plot_data$group)))
      min_value <- min(mar_out$plot_data$LCL,  na.rm = TRUE)
      max_value <- max(mar_out$plot_data$UCL,  na.rm = TRUE)
      range_values <- max_value - min_value
      no_char_y <- nchar(round(range_values, digits = 3))
      type_plot <- "rotated_plot"
    }
    ###4th CASE: INTEGER categories 2 to 20 -- count data
  } else if (var_dtype == DATA_TYPES$INTEGER &
             var_prop$NCategory > 2 & var_prop$NCategory <= 20 &
             # TODO: Count data can exceed 20, of course! How do we identify count
             # data here? Maybe include a pre-test to choose between poisson and
             # linear regression? Or also consider negative binomial regression here?
             !var_prop$AnyNegative) {
    # TODO: The website states 15 as cut-off, instead of 20. Which one is the
    # preferred value??
    mar_out <- util_margins_poi(resp_vars = resp_vars,
                                group_vars = group_vars,
                                co_vars = co_vars,
                                threshold_type = threshold_type,
                                threshold_value = threshold_value,
                                min_obs_in_subgroup = min_obs_in_subgroup,
                                ds1 = ds1,
                                label_col = label_col,
                                adjusted_hint = adjusted_hint,
                                title = title,
                                sort_group_var_levels = sort_group_var_levels,
                                include_numbers_in_figures = include_numbers_in_figures)

    .plot5 <- mar_out$plot
    obj1 <- util_create_lean_ggplot(
      ggplot2::ggplot_build(.plot5),
      .plot5 = .plot5
    )
    obj1_data <- util_rbind(
      data_frames_list = util_gg_get(obj1, "data")
    )
    min_value <- min(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
    max_value <- max(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
    range_values <- max_value - min_value
    no_char_y <- nchar(range_values)
    rm(obj1, obj1_data)
    type_plot <- "count_plot"

  } else if (var_scale %in% c(SCALE_LEVELS$RATIO, SCALE_LEVELS$INTERVAL)) {
    mar_out <- util_margins_lm(resp_vars = resp_vars,
                               group_vars = group_vars,
                               co_vars = co_vars,
                               threshold_type = threshold_type,
                               threshold_value = threshold_value,
                               min_obs_in_subgroup = min_obs_in_subgroup,
                               ds1 = ds1,
                               label_col = label_col,
                               adjusted_hint = adjusted_hint,
                               title = title,
                               sort_group_var_levels = sort_group_var_levels,
                               include_numbers_in_figures =
                                 include_numbers_in_figures,
                               n_violin_max = n_violin_max)

    .plot6 <- mar_out$plot
    obj1 <- util_create_lean_ggplot(
      ggplot2::ggplot_build(.plot6),
      .plot6 = .plot6
    )
    obj1_data <- util_rbind(
      data_frames_list = util_gg_get(obj1, "data")
    )
    min_value <- min(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
    max_value <- max(c(obj1_data$x, obj1_data$xintercept),  na.rm = TRUE)
    range_values <- max_value - min_value
    no_char_y <- nchar(range_values)
    rm(obj1, obj1_data)
    type_plot <- "violin_plot"
  } else  {
    util_error("No suitable method implemented yet, sorry.")
  }


  # finalize output ------------------------------------------------------------
  res_df <- mar_out$plot_data
  res_plot <- mar_out$plot

  SummaryTable <- data.frame( # TODO: This should finally feature all data SummaryData has, but w/o any formatting and rounding.
    Variables = resp_vars#,
    # FLG_acc_ud_loc = as.numeric(any(res_df$GRADING > 0)),
    # PCT_acc_ud_loc = round(sum(res_df$GRADING == 1)/nrow(res_df)*100,
   #                        digits = 2)
  )

  SummaryData <- cbind.data.frame(
    Variables = resp_vars,
    res_df
  )

  #modify number of decimal places

  if ("sample_size" %in% colnames(SummaryData)) {
    SummaryData$sample_size <-
      util_round_to_decimal_places(SummaryData$sample_size)
  }

  SummaryData$df <- NULL
  SummaryData$threshold <- NULL
  SummaryData$overall <- NULL
  SummaryData$GRADING <- NULL

  SummaryData$margins <- util_round_to_decimal_places(SummaryData$margins)
  SummaryData$LCL <- util_round_to_decimal_places(SummaryData$LCL)
  SummaryData$UCL <- util_round_to_decimal_places(SummaryData$UCL)
  SummaryData$CL <- paste0("[", format(SummaryData$LCL), "; ",
                           format(SummaryData$UCL), "]")
  SummaryData$LCL <- NULL
  SummaryData$UCL <- NULL
  if ("SE" %in% colnames(SummaryData)) {
    SummaryData$SE <- util_round_to_decimal_places(SummaryData$SE)
  }

  colnames(SummaryData)[which(colnames(SummaryData) == "sample_size")] <- "n"

  attr(SummaryData, "description") <- character(0)
  attr(SummaryData, "description")[[group_vars]] <- "Group: Observer/Device/..."
  if (resp_vars %in% colnames(SummaryData)) { # for nominal variables
    attr(SummaryData, "description")[[resp_vars]] <- "Categories of the Outcome"
    attr(SummaryData, "description")[["margins"]] <- "Probability for the Group"
  } else {
    attr(SummaryData, "description")[["margins"]] <- "Mean for the Group"
  }
  attr(SummaryData, "description")[["SE"]] <- "Standard error for the Group"
  attr(SummaryData, "description")[["n"]] <-
    "Number of Observations of the Group"
  attr(SummaryData, "description")[["CL"]] <-
    "Confidence Interval for the Group"


  SummaryPlot <- util_set_size(res_plot, width_em = 25 +
                                 1.2 * length(unique(ds1[[group_vars]])),
                               height_em = 25)

  #Information for sizing
  #  obj1 <- ggplot2::ggplot_build(res_plot)
  #  obj1_data <- util_rbind(data_frames_list = obj1$data)

  if(!exists("n_groups")) {
    n_groups <- nrow(SummaryData)
  }

  #  min_value <- min(c(obj1_data$x,obj1_data$xintercept),  na.rm = TRUE)
  #  max_value <- max(c(obj1_data$x,obj1_data$xintercept),  na.rm = TRUE)
  #  range_values <- max_value - min_value
  #  if ((max(obj1_data$x,na.rm = TRUE) - min(obj1_data$x,na.rm = TRUE)) < 2 ) {
  #    type_plot <- "count_plot"
  #  } else {
  #    type_plot <- "violin_plot"
  #  }
  #  rm(obj1, obj1_data)

  if(!exists("no_char_x")){
    no_char_x <- max(nchar(as.character(SummaryData[[group_vars]])))
  }


  return(util_attach_attr(list(
    ResultData = SummaryData,
    SummaryTable = SummaryTable,
    SummaryPlot = SummaryPlot),
    sizing_hints = list(figure_type_id = "marg_plot",
                        n_groups = n_groups,
                        no_char_x = no_char_x,
                        no_char_y = no_char_y,
                        type_plot = type_plot
    ),
    as_plotly = "util_as_plotly_acc_margins"))
}


#' @family plotly_shims
#' @concept plotly_shims
#' @noRd
util_as_plotly_acc_margins <- function(res, ...) {
  #remove classes for plotly to work properly
  res$SummaryPlot <- util_remove_dataquieR_result_class(res$SummaryPlot)
  # use res$SummaryPlot, not res_plot to avoid depending on the enclosure
  # of the result, that may contain study data.
  util_ensure_suggested("plotly")
  if (inherits(res$SummaryPlot, "patchwork")) {
    # extract estimates on size of the plot from the patchwork object
    # obtain relative width of the single elements of the plot
    rel_w <- res$SummaryPlot$patches$layout$widths /
      sum(res$SummaryPlot$patches$layout$widths, na.rm = TRUE)
    # extract the violin plots
    py1 <- try(util_ggplotly(res$SummaryPlot[[1]],
                             ...), silent = TRUE)
    if (util_is_try_error(py1)) {
      err <- attr(py1, "condition")
      err_msg <- conditionMessage(err)
      if (suppressWarnings(util_ensure_suggested("cli", err = FALSE))) {
        err_msg <- cli::ansi_strip(err_msg)
      }
      return(util_plotly_text(err_msg))
    }
    if (identical(py1$x$data[[2]]$mode, "lines+markers")) { # no violins were created because of too many observers, see dataquieR.max_group_var_levels_with_violins
      py1$x$data[[1]]$mode <- NULL # suppress a warning on print (https://github.com/plotly/plotly.R/issues/2242)
    } else {
      py1$x$data[[2]]$mode <- NULL # suppress a warning on print (https://github.com/plotly/plotly.R/issues/2242)
    }

    if (length(res$SummaryPlot) == 1) {
      py2 <- NULL
    } else {
      # extract the overall distribution plot
      py2 <- try(util_ggplotly(res$SummaryPlot[[2]],
                               ...), silent = TRUE)
    }
    if (util_is_try_error(py2)) {
      err <- attr(py2, "condition")
      err_msg <- conditionMessage(err)
      if (suppressWarnings(util_ensure_suggested("cli", err = FALSE))) {
        err_msg <- cli::ansi_strip(err_msg)
      }
      return(util_plotly_text(err_msg))
    }
    # check if both are plotly objects
    util_stop_if_not(!inherits(py1, "try-error"))
    util_stop_if_not(!inherits(py2, "try-error"))
    # https://plotly.com/r/subplots/#subplots-with-shared-yaxes


    #  summary_ds<-as.data.frame(dplyr::summarize(dplyr::group_by_at(ds1[, c(resp_vars, group_vars), drop = FALSE],
    #                                                               group_vars), samplesize = dplyr::n()))


    # py1<- util_ggplotly(py1, tooltip = paste("Sample size:",
    #                                            as.data.frame(dplyr::summarize(dplyr::group_by_at(ds1[, c(resp_vars, group_vars), drop = FALSE],
    #                                                                                                             group_vars), samplesize = dplyr::n()))[,2]))
    #py2<- plotly::layout(py2)

    target_layers <-
      which(lapply(lapply(py1$x$data, `[[`, "marker"), `[[`, "symbol") == "diamond")

    hovertexts <- lapply(py1$x$data, `[[`, "hovertext")
    if (!all(vapply(hovertexts, is.null, logical(1)))) {
      hovertexts_matching_sample_size <- lapply(hovertexts, grepl, pattern = "sample_size: ", fixed = TRUE)

      # S7-sicher: Anzahl der Gruppen über das ggplot-Kind des patchwork bestimmen
      child1      <- res$SummaryPlot[[1]]
      child1_data <- util_gg_get(child1, "data")

      hovertexts_matching_sample_size_with_length_nr_groups <-
        vapply(
          hovertexts_matching_sample_size,
          function(x) length(x) == length(unique(child1_data[[2]])) && all(x, na.rm = TRUE),
          FUN.VALUE = logical(1)
        )

      layer_with_sample_size <- which(hovertexts_matching_sample_size_with_length_nr_groups)
      if (length(layer_with_sample_size) != 1) {
        util_warning(c("Internal error: unexpected number of",
                       "layer_with_sample_size. Sorry, please report to us"))
      } else {
        hovertexts_with_hovertexts <- py1$x$data[[layer_with_sample_size]]$hovertext

        xpositions_with_hovertexts <- py1$x$data[[layer_with_sample_size]]$x

        for (tl in target_layers) {
          # amend texts for hover-texts, matching by x position.

          py1$x$data[[tl]]$text <- paste0(
            py1$x$data[[tl]]$text,
            "<br />",
            hovertexts_with_hovertexts[match(py1$x$data[[tl]]$x,
                                             xpositions_with_hovertexts)]
          )
        }
      }
    }

    target_layer_outliers <- which(!(unlist(lapply(lapply(lapply(py1$x$data, `[[`, "marker"), `[[`, "outliercolor"), is.null))))
    if (length(target_layer_outliers) == 1) {
      # https://github.com/plotly/plotly.R/issues/1114
      py1$x$data[[target_layer_outliers]]$marker$outliercolor <-
        py1$x$data[[target_layer_outliers]]$line$color
      py1$x$data[[target_layer_outliers]]$marker$line$color <-
        py1$x$data[[target_layer_outliers]]$line$color
      py1$x$data[[target_layer_outliers]]$marker$opacity <- 0.5
    }

    note <- ""
    try(note <- res$SummaryPlot$patches$annotation$caption, silent = TRUE)
    if (length(note) != 1 || !is.character(note)) {
      note <- ""
    }

    # recombine plots
    suppressMessages(force(plotly::layout(plotly::subplot(py1,
                                                          py2,
                                                          nrows =
                                                            res$SummaryPlot$patches$layout$nrow,
                                                          shareY = TRUE, #all plots use the same y axes
                                                          widths = #define relative width
                                                            rel_w),
                                          title = list(text = #get the overall title from patch
                                                         paste0(
                                                           res$SummaryPlot$patches$annotation$title,
                                                           "<br /><sub>",
                                                           res$SummaryPlot$patches$annotation$subtitle,
                                                           "</sub>"
                                                         )),
                                          font = list(size = 12),
                                          margin = 0.01,
                                          annotations = list(
                                            list(
                                              x = 1,
                                              y = 1,
                                              yshift = 24, # px
                                              text = note,
                                              align = 'right',
                                              valign = 'top',
                                              xref = 'paper',
                                              yref = 'paper',
                                              xanchor = 'right',
                                              yanchor = 'top',
                                              showarrow = FALSE,
                                              font = list(size = 8)
                                            )
                                          ))))

  } else {
    util_plot_figure_plotly(res$SummaryPlot, attr(res, "sizing_hints"))
  }
}

#' Escape characters for HTML in a data frame
#'
#' @param x [data.frame] to be escaped
#'
#' @return [data.frame] with html escaped content
#' @noRd
util_df_escape <- function(x) {
  if (identical(attr(x, "is_html_escaped"), TRUE)) {
    return(x)
  }
  util_ensure_suggested("htmltools", goal = "for HTML escaping data frames")
  util_expect_data_frame(x)
  x[] <- lapply(x, function(y) {
    r <- htmltools::htmlEscape(y)
    attr(r, "plain_label") <- attr(y, "plain_label")
    #Add the new data type attributes for columns
    attr(r, DATA_TYPE) <- attr(y, DATA_TYPE)

    r
  })
  x
}

#' Data frame leaves `haven`
#'
#' if `df` is/contains a `haven` `labelled` or `tibble` object, convert it to
#' a base R data frame
#'
#' @param df [data.frame] may have or contain non-standard classes
#' @param symb [character] name of the data frame for error messages
#' @param .dont_cast_off_cols [logical] internal use, only.
#'
#' @return [data.frame] having all known special things removed
#'
#' @noRd
util_cast_off <- function(df, symb, .dont_cast_off_cols = FALSE) {
  if (missing(symb)) {
    symb <- as.character(substitute(df))
  }
  orig <- df
  if (requireNamespace("tibble", quietly = TRUE)) { # TODO: data.table, what else, is missing
    if (tibble::is_tibble(df)) {
      df <- as.data.frame(df)
    }
  } else { # nocov start
    if (inherits(df, "tbl_df")) {
      util_warning(
        paste(
          "%s looks like a tibble. However, the package %s seems not",
          "to be available, which is quite strange.",
          "I cannot convert the tibble to a data.frame therefore.",
          "Tibbles do not always work like base R data.frames (see %s),",
          "so this can cause errors,",
          "because %s expects %s in base R data.frames, not in tibbles."
        ),
        dQuote(symb),
        dQuote("tibble"),
        dQuote("https://r4ds.had.co.nz/tibbles.html#tibbles-vs.data.frame"),
        dQuote("dataquieR"),
        dQuote(symb),
        applicability_problem = FALSE
      )
    }
  } # nocov end
  # drop all stuff like haven labels, labelled labels
  # df[] <- lapply(df, function(cl) {
  #   mostattributes(cl) <- NULL
  #   cl
  #   # if (is.factor(cl)) {
  #   #   mostattributes(cl) <- NULL
  #   #   cl
  #   # } else if (lubridate::is.timepoint(cl)) {
  #   #   # util_parse_date(cl)
  #   #   mostattributes(cl) <- NULL
  #   #   cl
  #   # } else {
  #   #   mostattributes(cl) <- NULL
  #   #   cl
  #   # }
  # })


  if (!.dont_cast_off_cols && !!ncol(df)) {
    li <- as.list(df)

    dtypes <- prep_datatype_from_data(colnames(df), df,
                                      .dont_cast_off_cols = TRUE)

    li <- mapply(cl = li, dt = dtypes, FUN = function(cl, dt) {
      if (is.factor(cl)) {
        mostattributes(cl) <- NULL
        return(cl)
      } else {
        return(util_data_type_conversion(cl, type = dt))
      }
    }, SIMPLIFY = FALSE)

    li <-
      lapply(li, function(x) {
        if (is.list(x) &&
            all(vapply(x, inherits, "hms", FUN.VALUE = logical(1)))) {
          hms::as_hms(unlist(x, use.names = FALSE))
        } else {
          x
        }
      })

    df <- do.call(data.frame,
                  c(list(
                        check.names = FALSE,
                        stringsAsFactors = FALSE
                      ), li)
    )

  }

  known_atts <- .ds1_attribute_names

  for (att in known_atts) attr(df, att) <- attr(orig, att)

  df
}

##### Prefixes:
# cr_vv = variable 1 value <cmp> variable 2 value OR variable 1 available
#                                                   <logic> variable 2 available
# cr_lc = variable 1 in levels <logic> variable 2 <cmp> constant value
# cr_ll = variable 1 in levels <logic> variable 1 in levels
# All functions should carry an description attribute for displaying them

######################################################################
# Detect abnormalities help functions
#
# 2 variables:
#
#    if A != B
#    if A > B
#    if A >= B
#    if A & is.na(B)
#    if A & !(is.na(B))
#    if A & B %in% {set of levels}
#    if A %in% {set of levels} & B >  value
#    if A %in% {set of levels} & B == value
#    if A %in% {set of levels} & B <  value
#    if A %in% {set of levels} & B %in% {set of levels}
#    if A %in% {set of levels} & !(B %in% {set of levels})
#
#############################
# List of generic functions #
#############################
A_not_equal_B_vv <- function(study_data, A, B, A_levels, B_levels,
                             A_value, B_value) {
  X <- study_data

  grading <- ifelse(X[[A]] != X[[B]], 1, 0)
  return(grading)
}
attr(A_not_equal_B_vv, "description") <- "A \u2260 B"

A_less_than_B_vv <- function(study_data, A, B, A_levels, B_levels,
                             A_value, B_value) {
  X <- study_data
  util_warn_unordered(X[[A]], A)
  util_warn_unordered(X[[B]], B)
  if (is.factor(X[[A]])) {
    X[[A]] <- util_as_numeric(X[[A]])
  }
  if (is.factor(X[[B]])) {
    X[[B]] <- util_as_numeric(X[[B]])
  }

  grading <- ifelse(X[[A]] < X[[B]], 1, 0)
  return(grading)
}
attr(A_less_than_B_vv, "description") <- "A < B"

A_less_equal_B_vv <- function(study_data, A, B, A_levels, B_levels,
                              A_value, B_value) {
  X <- study_data
  util_warn_unordered(X[[A]], A)
  util_warn_unordered(X[[B]], B)
  if (is.factor(X[[A]])) {
    X[[A]] <- util_as_numeric(X[[A]])
  }
  if (is.factor(X[[B]])) {
    X[[B]] <- util_as_numeric(X[[B]])
  }

  grading <- ifelse(X[[A]] <= X[[B]], 1, 0)
  return(grading)
}
attr(A_less_equal_B_vv, "description") <- "A \u2264 B"

A_greater_than_B_vv <- function(study_data, A, B, A_levels, B_levels,
                                A_value, B_value) {
  X <- study_data
  util_warn_unordered(X[[A]], A)
  util_warn_unordered(X[[B]], B)
  if (is.factor(X[[A]])) {
    X[[A]] <- util_as_numeric(X[[A]])
  }
  if (is.factor(X[[B]])) {
    X[[B]] <- util_as_numeric(X[[B]])
  }

  grading <- ifelse(X[[A]] > X[[B]], 1, 0)
  return(grading)
}
attr(A_greater_than_B_vv, "description") <- "A > B"

A_greater_equal_B_vv <- function(study_data, A, B, A_levels, B_levels,
                                 A_value, B_value) {
  X <- study_data
  util_warn_unordered(X[[A]], A)
  util_warn_unordered(X[[B]], B)
  if (is.factor(X[[A]])) {
    X[[A]] <- util_as_numeric(X[[A]])
  }
  if (is.factor(X[[B]])) {
    X[[B]] <- util_as_numeric(X[[B]])
  }

  grading <- ifelse(X[[A]] >= X[[B]], 1, 0)
  return(grading)
}
attr(A_greater_equal_B_vv, "description") <- "A \u2265 B"

A_present_not_B_vv <- function(study_data, A, B, A_levels, B_levels,
                               A_value, B_value) {
  X <- study_data
  grading <- ifelse(!is.na(X[[A]]) & is.na(X[[B]]), 1, 0)
  return(grading)
}
attr(A_present_not_B_vv, "description") <- "\u2203 A \u2227 \u2204 B"

A_present_and_B_vv <- function(study_data, A, B, A_levels, B_levels,
                               A_value, B_value) {
  X <- study_data
  grading <- ifelse(!is.na(X[[A]]) & !(is.na(X[[B]])), 1, 0)
  return(grading)
}
attr(A_present_and_B_vv, "description") <- "\u2203 A \u2227 \u2203 B"

A_present_and_B_levels_vl <- function(study_data, A, B, A_levels, B_levels,
                                      A_value, B_value) {
  X <- study_data
  grading <- ifelse((!is.na(X[[A]])) & X[[B]] %in% B_levels, 1, 0)
  return(grading)
}
attr(A_present_and_B_levels_vl, "description") <-
  "\u2203 A \u2227 B \u2208 L\u2082"

A_levels_and_B_levels_ll <- function(study_data, A, B, A_levels, B_levels,
                                     A_value, B_value) {
  X <- study_data
  grading <- ifelse(X[[A]] %in% A_levels & X[[B]] %in% B_levels, 1, 0)
  return(grading)
}
attr(A_levels_and_B_levels_ll, "description") <-
  "A \u2208 L\u2081 \u2227 B \u2208 L\u2082"

A_levels_and_B_gt_value_lc <- function(study_data, A, B, A_levels, B_levels,
                                       A_value, B_value) {
  X <- study_data
  util_warn_unordered(X[[B]], B)
  if (is.factor(X[[B]])) {
    X[[B]] <- util_as_numeric(X[[B]])
  }
  grading <- ifelse(X[[A]] %in% A_levels & X[[B]] > B_value, 1, 0)
  return(grading)
}
attr(A_levels_and_B_gt_value_lc, "description") <- "A \u2208 L \u2227 B > c"

A_levels_and_B_lt_value_lc <- function(study_data, A, B, A_levels, B_levels,
                                       A_value, B_value) {
  X <- study_data
  util_warn_unordered(X[[B]], B)
  if (is.factor(X[[B]])) {
    X[[B]] <- util_as_numeric(X[[B]])
  }
  grading <- ifelse(X[[A]] %in% A_levels & X[[B]] < B_value, 1, 0)
  return(grading)
}
attr(A_levels_and_B_lt_value_lc, "description") <- "A \u2208 L \u2227 B < c"

#' contradiction_functions
#'
#' Detect abnormalities help functions
#'
#' 2 variables:
#'  - `A_not_equal_B`, if `A != B`
#'  - `A_greater_equal_B`, if `A >= B`
#'  - `A_greater_than_B`, if `A > B`
#'  - `A_less_than_B`, if `A < B`
#'  - `A_less_equal_B`, if `A <= B`
#'  - `A_present_not_B`, if `A & is.na(B)`
#'  - `A_present_and_B`, if `A & !(is.na(B))`
#'  - `A_present_and_B_levels`, if `A & B  %in% {set of levels}`
#'  - `A_levels_and_B_gt_value`, if `A %in% {set of levels} & B >  value`
#'  - `A_levels_and_B_lt_value`, if `A %in% {set of levels} & B <  value`
#'  - `A_levels_and_B_levels`, if
#'                             `A %in% {set of levels} & B %in% {set of levels}`
#'
#'
#' @export
#'
#' @keywords internal
contradiction_functions <- objects(pattern = ".*")
contradiction_functions <- mget(contradiction_functions)
contradiction_functions <- contradiction_functions[
  vapply(contradiction_functions, is.function, TRUE)]
contradiction_functions <- contradiction_functions[
  vapply(contradiction_functions, function(object) {
  !is.null(attr(object, "description"))
}, TRUE)]

#' description of the contradiction functions
#' @export
contradiction_functions_descriptions <- lapply(contradiction_functions, attr,
                                               "description")

######################################################################
#
#    TBD
#
#    > 2 Variablen
#
#    if B == level & A1 < A2
#    if A == level & sum(!(is.na(B))) > 0
#    if A == level & sum(!(is.na(B))) == 0
#    if A == level & sum(B) == 0
#    if abs(sum(sign(A-B))) == length(A)
#    if A1 %in% {set of levels} & sum(c(B) %in% {set of levels}) > 0
#

study_data <- data.frame(
  age_at_baseline = c(20, 23, 34, 36, 34, 20, 19, 18),
  age_at_followup = 2 + c(20, 23, 34, 36, 34, 20, 19, 18),
  sex_at_baseline = c("m", "f", "m", "m", "m", "f", "m", "m"),
  sex_at_followup = c("m", "f", "m", "m", "m", "m", "m", "f"),
  income_class_bl = ordered(
    c(10, 10, 10, 20, 20, 30, 30, 30),
    levels = c(10, 20, 30),
    labels = c("low", "medium", "high")
  ),
  income_class_fu = ordered(
    c(10, 10, 10, 20, 10, 20, 30, 30),
    levels = c(30, 20, 10),
    labels = c("low", "medium", "high")
  ),
  worst_grade = c("1", "2", "3", "2", "2", "2", "3", "5"),
  best_grade = c("1", "1", "1", "2", "1", "1", "2", "4"),
  piercing_date = c(NA, as.POSIXct("2020-01-01"), NA, NA, NA, NA, NA, NA),
  piercing_type = c(NA, "12", NA, NA, NA, NA, NA, NA),
  mrt_date = c(as.POSIXct("2020-01-01"), NA,
               as.POSIXct("2020-01-01"),
               as.POSIXct("2020-01-01"),
               as.POSIXct("2020-01-01"),
               as.POSIXct("2020-01-01"),
               as.POSIXct("2020-01-01"),
               as.POSIXct("2020-01-01")),
  body_size = c(180, 150, 179, 178, 150, 150, 190, 200),
  stringsAsFactors = TRUE
)

test_that("A_not_equal_B_vv works", {
  skip_on_cran()
  expect_equal(
    A_not_equal_B_vv(study_data = study_data,
                     A = "sex_at_baseline",
                     B = "sex_at_followup"
                    ),
    c(0, 0, 0, 0, 0, 1, 0, 1))
})
test_that("A_less_than_B_vv works", {
  skip_on_cran()
  expect_equal(
    A_less_than_B_vv(study_data = study_data,
                     A = "age_at_followup",
                     B = "age_at_baseline")
    ,  c(0, 0, 0, 0, 0, 0, 0, 0))
  suppressWarnings(expect_warning(A_less_than_B_vv(study_data = study_data,
                                  A = "sex_at_baseline",
                                  B = "sex_at_followup"),
                 regexp = "Could not convert .* to numeric values"))
  suppressWarnings(expect_warning(A_less_than_B_vv(study_data = study_data,
                   A = "sex_at_baseline",
                   B = "sex_at_followup"),
               regexp =
                      "Don't know, how to compare values of .+? (.+factor.+)"))
  suppressWarnings(expect_equal(A_less_than_B_vv(study_data = study_data,
                   A = "income_class_fu",
                   B = "income_class_bl"),
               c(0, 0, 0, 0, 0, 1, 1, 1)))
  suppressWarnings(expect_warning(
    expect_equal(A_less_than_B_vv(study_data = study_data,
                   A = "worst_grade",
                   B = "best_grade"),
               c(0, 0, 0, 0, 0, 0, 0, 0)),
    regexp =
      paste0("Don't know, how to compare values of .+worst_grade.+",
             "(.+factor.+, .+integer.+)", collapse = " ")
  ))
})

test_that("A_less_equal_B_vv works", {
  skip_on_cran()
  expect_equal(
    A_less_equal_B_vv(study_data = study_data,
                     A = "age_at_followup",
                     B = "age_at_baseline")
    ,  c(0, 0, 0, 0, 0, 0, 0, 0))
  suppressWarnings(expect_warning(
    A_less_equal_B_vv(study_data = study_data,
                                   A = "sex_at_baseline",
                                   B = "sex_at_followup"),
    regexp = "Could not convert .* to numeric values"))

  suppressWarnings(expect_warning(
    A_less_equal_B_vv(study_data = study_data,
                                  A = "sex_at_baseline",
                                  B = "sex_at_followup"),
    regexp = "Don't know, how to compare values of .+? (.+factor.+)"
  ))
  suppressWarnings(expect_equal(A_less_equal_B_vv(study_data = study_data,
                                A = "income_class_fu",
                                B = "income_class_bl"),
               c(0, 0, 0, 1, 0, 1, 1, 1)))
  suppressWarnings(expect_warning(
    expect_equal(A_less_equal_B_vv(study_data = study_data,
                                  A = "worst_grade",
                                  B = "best_grade"),
                 c(1, 0, 0, 1, 0, 0, 0, 0)),
    regexp =
      paste0("Don't know, how to compare values of .+worst_grade.+",
             "(.+factor.+, .+integer.+)", collapse = " ")
  ))
})


test_that("A_greater_than_B_vv works", {
  skip_on_cran()
  expect_equal(
    A_greater_than_B_vv(study_data = study_data,
                      A = "age_at_followup",
                      B = "age_at_baseline")
    ,  c(1, 1, 1, 1, 1, 1, 1, 1))
  suppressWarnings(expect_warning(
    A_greater_than_B_vv(study_data = study_data,
                                     A = "sex_at_baseline",
                                     B = "sex_at_followup"),
    regexp = "Could not convert .* to numeric values"))

  suppressWarnings(expect_warning(
    A_greater_than_B_vv(study_data = study_data,
                                   A = "sex_at_baseline",
                                   B = "sex_at_followup"),
    regexp = "Don't know, how to compare values of .+? (.+factor.+)"
  ))
  expect_equal(A_greater_than_B_vv(study_data = study_data,
                                 A = "income_class_fu",
                                 B = "income_class_bl"),
               c(1, 1, 1, 0, 1, 0, 0, 0))
  suppressWarnings(expect_warning(
    expect_equal(A_greater_than_B_vv(study_data = study_data,
                                   A = "worst_grade",
                                   B = "best_grade"),
                 c(0, 1, 1, 0, 1, 1, 1, 1)),
    regexp =
      paste("Don't know, how to compare values of .+worst_grade.+",
             "(.+factor.+, .+integer.+)", collapse = " ")
  ))
})


test_that("A_greater_equal_B_vv works", {
  skip_on_cran()
  expect_equal(
    A_greater_equal_B_vv(study_data = study_data,
                     A = "age_at_followup",
                     B = "age_at_baseline")
    ,  c(1, 1, 1, 1, 1, 1, 1, 1))
  suppressWarnings(expect_warning(
    A_greater_equal_B_vv(study_data = study_data,
                                      A = "sex_at_baseline",
                                      B = "sex_at_followup"),
    regexp = "Could not convert .* to numeric values"))

  suppressWarnings(expect_warning(
    A_greater_equal_B_vv(study_data = study_data,
                                  A = "sex_at_baseline",
                                  B = "sex_at_followup"),
    regexp = "Don't know, how to compare values of .+? (.+factor.+)"
  ))
  expect_equal(A_greater_equal_B_vv(study_data = study_data,
                                A = "income_class_fu",
                                B = "income_class_bl"),
               c(1, 1, 1, 1, 1, 0, 0, 0))
  suppressWarnings(expect_warning(
    expect_equal(A_greater_equal_B_vv(study_data = study_data,
                                  A = "worst_grade",
                                  B = "best_grade"),
                 c(1, 1, 1, 1, 1, 1, 1, 1)),
    regexp =
      paste0("Don't know, how to compare values of .+worst_grade.+",
             "(.+factor.+, .+integer.+)", collapse = " ")
  ))
})

test_that("A_present_not_B_vv works", {
  expect_equal(
    A_present_not_B_vv(study_data = study_data,
                         A = "piercing_date",
                         B = "piercing_type")
    ,  c(0, 0, 0, 0, 0, 0, 0, 0))
  expect_equal(
    A_present_not_B_vv(study_data = study_data,
                       A = "piercing_date",
                       B = "mrt_date")
    ,  c(0, 1, 0, 0, 0, 0, 0, 0))
})

test_that("A_present_and_B_vv works", {
  skip_on_cran()
  expect_equal(
    A_present_and_B_vv(study_data = study_data,
                       A = "piercing_date",
                       B = "mrt_date")
    ,  c(0, 0, 0, 0, 0, 0, 0, 0))
  expect_equal(
    A_present_and_B_vv(study_data = study_data,
                       A = "piercing_date",
                       B = "piercing_type")
    ,  c(0, 1, 0, 0, 0, 0, 0, 0))
})


test_that("A_present_and_B_levels_vl works", {
  skip_on_cran()
  expect_equal(
    A_present_and_B_levels_vl(study_data = study_data,
                     A = "piercing_date",
                     B = "piercing_type",
                     B_levels = "none")
    , c(0, 0, 0, 0, 0, 0, 0, 0))

  expect_equal(
    A_present_and_B_levels_vl(study_data = study_data,
                              A = "piercing_date",
                              B = "piercing_type",
                              B_levels = "12")
    , c(0, 1, 0, 0, 0, 0, 0, 0))


})

test_that("A_levels_and_B_levels_ll works", {
  skip_on_cran()
  expect_equal(
    A_levels_and_B_levels_ll(study_data = study_data,
                              A = "sex_at_baseline",
                              A_levels = "f",
                              B = "income_class_bl",
                              B_levels = "high")
    , c(0, 0, 0, 0, 0, 1, 0, 0))

})

test_that("A_levels_and_B_gt_value_lc works", {
  skip_on_cran()
  expect_equal(
    A_levels_and_B_gt_value_lc(study_data = study_data,
                             A = "sex_at_baseline",
                             A_levels = "f",
                             B = "body_size",
                             B_value = 185)
    , c(0, 0, 0, 0, 0, 0, 0, 0))

})

test_that("A_levels_and_B_lt_value_lc works", {
  skip_on_cran()
  expect_equal(
    A_levels_and_B_lt_value_lc(study_data = study_data,
                               A = "sex_at_baseline",
                               A_levels = "m",
                               B = "body_size",
                               B_value = 180)
    , c(0, 0, 1, 1, 1, 0, 0, 0))

})

#' Convert a [dataquieR report v2][dq_report2] to a named list of web pages
#'
#' @param report [dataquieR report v2][dq_report2].
#' @param template [character] template to use, only the name, not the path
#' @param disable_plotly [logical] do not use `plotly`, even if installed
#' @param progress [`function`] lambda for progress in percent -- 1-100
#' @param progress_msg [`function`] lambda for progress messages
#' @param block_load_factor [numeric] multiply size of parallel compute blocks
#'                                    by this factor.
#' @param dir [character] output directory for potential `iframes`.
#' @param my_dashboard [list] of class `shiny.tag.list` featuring a dashboard or
#'                            missing or `NULL`
#'
#' @return named list, each entry becomes a file with the name of the entry.
#'         the contents are `HTML` objects as used by `htmltools`.
#' @examples
#' \dontrun{
#' devtools::load_all()
#' prep_load_workbook_like_file("meta_data_v2")
#' report <- dq_report2("study_data", dimensions = NULL, label_col = "LABEL");
#' save(report, file = "report_v2.RData")
#' report <- dq_report2("study_data", label_col = "LABEL");
#' save(report, file = "report_v2_short.RData")
#' }
#'
#' @family html
#' @concept process
#' @keywords internal
util_generate_pages_from_report <- function(report, template,
                                            disable_plotly,
                                            progress = progress,
                                            progress_msg = progress_msg,
                                            block_load_factor,
                                            dir,
                                            my_dashboard) {
  util_ensure_suggested(pkg = c("htmltools"),
                                goal = "generate interactive HTML-reports.")
  have_plot_ly <- util_ensure_suggested(pkg = c("plotly"),
                                        goal = "generate interactive figures in plain HTML-reports.",
                                        err = FALSE)
  if (disable_plotly) have_plot_ly <- FALSE
  if (have_plot_ly) {
    plot_figure <- util_plot_figure_plotly
  } else {
    plot_figure <- util_plot_figure_no_plotly
  }

  vars_in_rep <- rownames(report)

  meta_data <- attr(report, "meta_data")
  label_col <- attr(report, "label_col")
  warn_pred_meta <- attr(report, "warning_pred_meta")
  title <- attr(report, "title")
  subtitle <- attr(report, "subtitle")
  label_meta_data_hints <- attr(report, "label_meta_data_hints")

  pages <- list()
  call_env <- environment()

  append_single_page <- function(drop_down_to_attach, # main menu entry (first-level)
                                 div_name, # sub-menu entry (second-level)
                                 file_name, # name where the page is stored, it is possible to add more than one page to a file
                                 ...) { # ... are the contents in htmltools compatible objects
    alternative_names <- attr(div_name, "alternative_names")
    # Setup ####
    if (file_name %in% names(call_env$pages)) {
      fil <- call_env$pages[[file_name]]
    } else {
      fil <- list()
    }
    all_ids <- unlist(lapply(call_env$pages, names))
    if (div_name %in% all_ids) {
      pgidx <- vapply(FUN = `%in%`, X = lapply(call_env$pages, names),
                      x = div_name, FUN.VALUE = logical(1))
      existing_page <- call_env$pages[pgidx]
      if (length(existing_page) == 1) {
        file_name <- names(existing_page)
        if (length(file_name) == 1 && is.character(file_name)) {
          existing_page <- existing_page[[file_name]][[div_name]]
          dd_men <- attr(existing_page, "dropdown")
          dropdown <- existing_page$dropdown
          existing_page$dropdown <- NULL
          pg <- htmltools::tagList(
            existing_page,
            ...,
            dropdown = dropdown
          )
          attr(pg, "dropdown") <- dd_men
          call_env$pages[[file_name]][[div_name]] <- pg
          return(invisible(NULL))
        }
      }
      util_error("Cannot create report, single page with ID %s already exists",
                 dQuote(div_name))
    }

    curr_url <- htmltools::tagGetAttribute(.menu_env$menu_entry(
      sprintf("%s#%s", file_name, div_name),
      alternative_names = alternative_names), "href")

    # https://matthewjamestaylor.com/custom-tags?utm_content=cmp-true
    # e.g. <r-1 curr_url=></r-1> // needs to have a hyphen, attributes in custom tags don't need the prefix data-
    curr_url <- htmltools::tag("dataquier-data", varArgs = list(`curr-url` =
                                                      gsub("\"", "\\\"",
                                                           curr_url,
                                                           fixed = TRUE)))

    sp <- util_attach_attr(htmltools::div(
      ...,
      curr_url,
      class = "singlePage",
      id = div_name
    ), dropdown = drop_down_to_attach)
    fil[[div_name]] <- sp
    call_env$pages[[file_name]] <- fil
    invisible(NULL)
  }

  progress_msg("Page generation", "Creating Page 1")
  # Note on changed labels
  if (nchar(attr(report, "label_modification_text")) > 0) {
    notes_labels <- htmltools::div(
      htmltools::h3("Label modifications"),
      util_html_table(util_df_escape(attr(report, "label_modification_table")),
                      dl_fn = "Label_modifications")
    )
  } else {
    notes_labels <- htmltools::div()
  }

  # Page 1 ####
  # Technical information about the R session and the report and first Overview
  repsum <- summary(report)
  summaryplots <- plot(repsum, dont_plot = TRUE, disable_plotly =
                         disable_plotly)
  comat <- print(repsum, dont_print = TRUE)

  integrity_issues_before_pipeline <-
    htmltools::span("")

  if (!is.null(integrity_issues_before_pipeline_conds <-
               attr(report, "integrity_issues_before_pipeline")) &&
      length(integrity_issues_before_pipeline_conds) > 0) {
    iss_df <-
      lapply(attr(report, "integrity_issues_before_pipeline"), function(cnd) {
        msg <- paste(cli::ansi_strip(conditionMessage(cnd)), collapse = "\n")
        var <- attr(cnd, "varname")
        if (length(var) == 0) {
          var <- ""
          # return(
          #   data.frame(Indicator = character(0),
          #              `Study Variable` = character(0),
          #              Issue = character(0),
          #              check.names = FALSE, stringsAsFactors = FALSE,
          #              check.rows = FALSE, fix.empty.names = FALSE,
          #              row.names = NULL)
          # )
        }
        indi <- attr(cnd, "integrity_indicator")
        if (length(indi) != 1) {
          return(
            data.frame(Indicator = character(0),
                       `Study Variable` = character(0),
                       Issue = character(0),
                       check.names = FALSE, stringsAsFactors = FALSE,
                       check.rows = FALSE, fix.empty.names = FALSE,
                       row.names = NULL)
          )
        }
        Indicator <- "Integrity issue"
        abbreviation <- NULL # to make check happy.
        try(
          Indicator <-
            head(util_get_concept_info("dqi", abbreviation == indi, "Name",
                                drop = TRUE), 1)
        , silent = TRUE)
        if (length(Indicator) != 1) {
          Indicator <- "Integrity issue"
        }

        data.frame(Indicator = Indicator, `Study Variable` = var, Issue = msg,
                   check.names = FALSE, stringsAsFactors = FALSE,
                   check.rows = FALSE, fix.empty.names = FALSE,
                   row.names = NULL)
    })
    iss_df <- util_rbind(data_frames_list = iss_df)
    # TODO: Collapse messages
    iss_df <- iss_df[!duplicated(iss_df), , FALSE]
    integrity_issues_before_pipeline <-
      htmltools::tagList(
        htmltools::h2("Integrity Issues"),
        htmltools::p(
          "There where some technical issues with the study data frame."),
        util_html_table(util_df_escape(iss_df), dl_fn = "Integrity_Issues")
      )

  }

  properties <- attr(report, "properties")
  if (!is.list(properties)) {
    properties <- list(error = "No report properties found, report file corrupted?")
  }

  if (is.list(properties)) {
    # TODO: jsTree
    p <- properties
    p[vapply(p, is.call, FUN.VALUE = logical(1))] <-
      vapply(lapply(p[vapply(p, is.call, FUN.VALUE = logical(1))], deparse),
             paste, collapse = "\n", FUN.VALUE = character(1))
    p[vapply(p, inherits, "POSIXt", FUN.VALUE = logical(1))] <-
      vapply(p[vapply(p, inherits, "POSIXt", FUN.VALUE = logical(1))],
             paste, collapse=" ", FUN.VALUE = character(1))
    p <- p[vapply(p, is.vector, FUN.VALUE = logical(1))]
    p <- p[vapply(p, length, FUN.VALUE = integer(1)) == 1]
    p <- data.frame(`  ` = names(p),
                    ` ` = unlist(unname(p)),
                    stringsAsFactors = FALSE,
                    check.names = FALSE)
  }

  # Information on dataset sizes
  vinsd <- attr(report, "study_data_dimnames")[[2]]
  vinmd <- meta_data[[VAR_NAMES]]
  with_md <- intersect(vinsd, vinmd)
  wo_md <- setdiff(vinsd, vinmd)
  if (length(wo_md) > 0) {
    mark <- function(x) {
      paste0("<strong>", x, "</strong")
    }
  } else {
    mark <- identity
  }
  meta_data_item_computation <-
    attr(report, "meta_data_item_computation")
  socs_cnt <- sum(!is.na(attr(report, "meta_data")[[
    COMPUTED_VARIABLE_ROLE]]))
  info_sd <-
    data.frame(
      "Number of ..." = c(
        "observations in study data set",
        mark("variables in study data set"),
        mark("variables with item-level metadata"),
        ifelse(socs_cnt > 0, mark("out of these, computed for social science indicators"), "")
      ),
      " " = c(
        length(attr(report, "study_data_dimnames")[[1]]),
        mark(as.character(length(vinsd))),
        mark(as.character(length(with_md))),
        ifelse(socs_cnt > 0, mark(as.character(socs_cnt)), "")),
      # TODO: Does it suffice to check for VAR_NAMES in the metadata? Should we also check if any valid metadata information apart from the name is given?
      check.names = FALSE)
  info_sd_hover <- info_sd
  info_sd_hover[] <- ""
  if (length(wo_md) > 0) {
    info_sd_hover[2, ] <- paste0(
      "<span style=\"color:#ffffbb\">Variables without metadata: ",
      util_pretty_vector_string(wo_md, n_max = 20),
      "</span>; Variables in study data: ",
      util_pretty_vector_string(vinsd, n_max = 20))
  } else {
    info_sd_hover[2, ] <- util_pretty_vector_string(vinsd, n_max = 20)
  }
  if (length(wo_md) > 0) {
    info_sd_hover[3, ] <- paste0(
      "<span style=\"color:#ffffbb\">Variables without metadata: ",
     util_pretty_vector_string(wo_md, n_max = 20),
     "</span>; Variables with metadata: ",
     util_pretty_vector_string(with_md, n_max = 20))
  } else {
    info_sd_hover[3, ] <- util_pretty_vector_string(with_md, n_max = 20)
  }
  info_md <-
    data.frame(
      "Number of ..." = "variables in item-level metadata",
      " " = nrow(meta_data),
      check.names = FALSE
    )
  if ("meta_data_segment" %in% names(attributes(report)) &&
      nrow(attr(report, "meta_data_segment")) > 0) {
    info_md <-
      rbind(info_md, list("segments", nrow(attr(
        report, "meta_data_segment"
      ))))
  }

  ## create a table with the summary of indicators and descriptors
  info_dim_dq <- util_generate_table_indicators_descriptors(report)
  # info_dim_dq <- rbind(info_dim_dq,
  #                      data.frame(Dimension = "Descriptors",
  #                                 `No. DQ indicators` = 0,
  #                                 `No. DQ descriptors` = 2,
  #                                 stringsAsFactors = FALSE,
  #                                 check.names = FALSE
  #                      ))

  # TODO: Discuss the indicator descriptor assignment problem with reference to dimensions
  info_dim_dq[info_dim_dq$Dimension == "Accuracy", "No. DQ descriptors"] <-
    info_dim_dq[info_dim_dq$Dimension == "Accuracy", "No. DQ descriptors"] + 5
  # The 5 descriptors are Location Parameters, Spread, Skewness, Kurtosis, and the graph
  info_dim_dq[info_dim_dq$Dimension == "Completeness", "No. DQ descriptors"] <-
    info_dim_dq[info_dim_dq$Dimension == "Completeness", "No. DQ descriptors"] + 1
  # The descriptor is the columns Missing/Valid

  info_dim_dq[["No. DQ descriptors"]] <- NULL

  if (!isTRUE(attr(report, "dt_adjust"))) {
    dt_adjust_info <- htmltools::tagList(
      htmltools::hr(),
      htmltools::p(htmltools::em(htmltools::strong(
        paste("Data types are assumed to be all matching, because this report",
        "was computed with the dt_adjust-option set to FALSE."))
    )))
  } else {
    dt_adjust_info <- htmltools::HTML("")
  }

  append_single_page("General",
                     util_attach_attr("Report information",
                                      alternative_names = c("index", "home")),
                     "report.html",
                     htmltools::tagList(
                       htmltools::h1(title),
                       htmltools::h2(subtitle),
                       htmltools::h3("Study data summary"),
                       htmltools::browsable(util_formattable(
                         escape_all_content = FALSE,
                         info_sd,
                         min_color = c(255, 255, 255),
                         max_color = c(255, 255, 255),
                         style_header = c(
                           "font-weight: bold;width: 15em;text-align: left;",
                           "font-weight: bold;width: 6em;"
                         ),
                         hover_texts = info_sd_hover
                       )),
                       htmltools::h3("Metadata summary"),
                       htmltools::p(style = "color:red;font-size:200%;",
                                    htmltools::strong(
                         htmltools::em(warn_pred_meta))),
                       htmltools::browsable(util_formattable(
                         info_md,
                         min_color = c(255, 255, 255),
                         max_color = c(255, 255, 255),
                         style_header = c(
                           "font-weight: bold;width: 15em;text-align: left;",
                           "font-weight: bold;width: 6em;"
                         )
                       )),
                       htmltools::h2("Scope of the data quality assessment"),
                       htmltools::browsable(util_formattable(
                         info_dim_dq,
                         min_color = c(255, 255, 255),
                         max_color = c(33, 102, 172),
                         style_header = paste0(
                           "font-weight: bold;text-align: right;",
                           c("width: 6em;", "width: 12em;", "width: 12em;")
                         )
                       )),
                       htmltools::h2("Data Quality Summary"),
                       summaryplots,
                       htmltools::a(
                         href = "#Item-level data quality summary",
                         "Display Detailed View"),
                       integrity_issues_before_pipeline,
                       htmltools::h2("Technical information"),
                       htmltools::p("The table below summarizes technical information about this report, the R session and the operating system."),
                       util_html_table(util_df_escape(p),
                                       dl_fn = "Report_Metadata",
                                       kv_table = TRUE),
                       htmltools::p(htmltools::tags$i(id = "render-time",
                                                      sprintf("Rendered using %s %s at %s",
                                                              utils::packageName(),
                                                              util_dataquieR_version(),
                                                              as.character(Sys.time()))
                       )),
                       htmltools::tags$script( # https://stackoverflow.com/a/34579496
                         '$(function() {
                              var xx = $("#render-time").html()
                              var data = window.renderingData
                              if (data.hasOwnProperty("renderingTime")) {
                                $("#render-time").html(xx + " in " + data.renderingTime)
                              } else {
                                $("#render-time").html(xx + " -- no rendering time available.")
                              }
                              // console.log(data);
                          });'),
                       if (!util_is_empty_html(notes_labels)) {
                         list(
                           htmltools::hr(),
                           htmltools::a(id = "notes_labels", notes_labels)
                         )
                       },
                       if (!util_is_empty_html(dt_adjust_info)) {
                         list(
                           dt_adjust_info,
                           htmltools::hr()
                         )
                       },
                       htmltools::h2("Related literature"),
                       do.call(htmltools::tagList, lapply(format(
                         utils::readCitationFile(system.file("CITATION",
                                                             package = utils::packageName()),
                                                 list(Encoding = "UTF-8")),
                         style="HTML"),
                         htmltools::HTML)),
                       htmltools::tags$button(class = "clipbtn",
                                              `data-clipboard-text` = paste(
                                                format(
                                                  utils::readCitationFile(
                                                    system.file("CITATION",
                                                                package =
                                                                  utils::packageName()),
                                                    list(Encoding = "UTF-8")),
                                                  style="bibtex"),
                                                collapse = "\n\n"),
                                              "\U1F4CB BibTeX",
                                              title = "copy BibTeX to clipboard")
                     ))

  progress_msg("Page generation", "Creating Page 2")

  # Page 2 ####
  # TODO apmat_not_segments_and_other_non_item_levels
  # TODO: add some hints if there is an integrity issue
  append_single_page("General",
                     util_attach_attr("Item-level data quality summary",
                                      alternative_names = c("summary table")),
                     "report.html",
                     htmltools::htmlTemplate(
                       text_ = readLines(system.file("templates", template, "overview.html",
                                                     package = utils::packageName())),
                       comat = util_apply_full_page_table_class(comat),
                       summaryplots = summaryplots,
                       # ismat = ismat,
                       # apmat = apmat,
                       # ermat = ermat,
                       # anamat = anamat,
                       # notes_labels = notes_labels,
                       util_float_index_menu = util_float_index_menu,
                       util_map_labels = util_map_labels,
                       util_get_concept_info = util_get_concept_info,
                       util_abbreviate = util_abbreviate)
  )

  if (have_plot_ly) {
    append_single_page("General",
                       util_attach_attr("Item-level data quality chart",
                                        alternative_names = c("summary chart", "sunburst")),
                       "sunburst.html",
                       plot(repsum, hierarchy = "DQ_OBS"))
  }

  if (!missing(my_dashboard) && inherits(my_dashboard, "shiny.tag.list")) {
    append_single_page("General",
                       util_attach_attr("Item-level data quality dashboard",
                                        alternative_names = c("summary dashboard")),
                       "dashboard.html",
                       util_apply_full_page_table_class(my_dashboard))
  }

  # FIXME: For multicore/fork, there are also some problems with fork on newer Macs: https://github.com/rails/rails/issues/38560
  if (!getOption("parallelMap.mode", "local") %in% c("local", "multicore", "socket")) { # TODO: Support somehow better psock, even with storr, it does not really work :()
    util_error(c("On a non-local/multicore cluster,",
                 "parallel rendering of reports is not supported.",
                 "Please call %s before",
                 "rendering the report."),
               dQuote("parallelMap::parallelStop()")
    )
    # IDEA: Addd to DESCRIPTION:
    # biocViews: Software
    # and
    # Suggests: SharedObject
    # BUT: This is not working, the package in unstable and has trouble with
    # the ulimti/stack size
    # if (util_ensure_suggested("SharedObject", "render reports in parallel",
    #                           err = FALSE)) {
    #   r <- report
    #   rm(report)
    #   gc(verbose = FALSE, full = TRUE)
    #   util_message("Sharing report with the workers")
    #   report <- SharedObject::share(r)
    # } else {
    #   util_error(c("On a non-local/multicore cluster,",
    #                "parallel rendering of reports is not supported",
    #                "without having %s installed. Please call %s before",
    #                "rendering the report, or install %s from %s."),
    #              dQuote("SharedObject"),
    #              dQuote("parallelMap::parallelStop()"),
    #              dQuote("SharedObject"),
    #              dQuote(
    # "http://www.bioconductor.org/packages/release/bioc/html/SharedObject.html")
    #              )
    # }
  }

  cores <- util_get_cores_safe()
  if (!is.numeric(cores) || !util_is_integer(cores) || !is.finite(cores)) {
    cores <- 1
  }

  if (cores > 1) {
    # rep_cls <- class(report)
    # class(report) <- NULL
    # suppressWarnings(parallelMap::parallelExport("report", "rep_cls", "have_plot_ly", "template"))
    # parallel::clusterEvalQ(cl = NULL, class(report) <- rep_cls)
    if (suppressWarnings(util_ensure_suggested("pkgload", err = FALSE,
                                               goal =
                                               "not really needed"))) {
      dev_package <- pkgload::is_dev_package(utils::packageName())
    } else {
      dev_package <- FALSE
    }

    if (dev_package && !is.null(parallel::getDefaultCluster())) {
      .d <- system.file(package = utils::packageName())
      .exp <- substitute({
        pkgload::load_all(path = .d)
        invisible(NULL)
      })
      parallel::clusterExport(envir = environment(), varlist = ".exp")
      parallel::clusterEvalQ(expr = eval(.exp))

    } else {
      parallelMap::parallelLibrary(utils::packageName(), show.info = FALSE)
    }

    ## --- export stuff ----------------------------------------------------------
    .options <- options() #options to be copied to the children (child process)
    .options <- .options[startsWith(names(.options), "dataquieR.")] #only dataquieR options selected

    progress_msg("Cluster setup: initializing parallel mode, if applicable", "exporting options")

    suppressMessages(suppressWarnings(parallelMap::parallelExport(
      objnames = ".options", show.info = FALSE)))

    progress_msg("Cluster setup: initializing parallel mode, if applicable", "exporting data frame cache")

    rule_sets <- getOption("dataquieR.grading_rulesets",
                           dataquieR.grading_rulesets_default)

    rule_formats <- getOption("dataquieR.grading_formats",
                              dataquieR.grading_formats_default)

    dataframes_list <- setNames(list(prep_get_data_frame(rule_sets),
                                     prep_get_data_frame(rule_formats)),
                                nm = c(rule_sets,
                                       rule_formats))

    suppressMessages(suppressWarnings(parallelMap::parallelExport(
      objnames = "dataframes_list", show.info = FALSE)))

    parallel::clusterEvalQ(cl = NULL, options(.options))
    parallel::clusterEvalQ(cl = NULL, prep_add_data_frames(
      data_frame_list = dataframes_list))

    parallel::clusterEvalQ(cl = NULL, {
      ..glbs <- get(".dq2_globs", envir = asNamespace("dataquieR"))
      ..glbs$.called_in_pipeline <- TRUE
    })


  }

  progress_msg("Page generation", "Creating indicator pages in memory")

# TODO: if https://community.plotly.com/t/cant-show-heatmap-inside-div-error-something-went-wrong-with-axis-scaling/30616/9 is fixed, use auto-sizing
  dim_pages <- util_html_for_dims(
    report,
    use_plot_ly = have_plot_ly,
    template = template,
    block_load_factor = block_load_factor,
    repsum = repsum,
    dir = dir
  )

  rendered_repsum <- comat

  dim_pages <- dim_pages[vapply(dim_pages, length, FUN.VALUE = integer(1))
                         != 0]

  progress_msg("Page generation", "Mounting indicator pages")
  i <- 0
  n <- length(dim_pages)
  for (args in dim_pages) {
    do.call(append_single_page, args) # creates the indicator related pages because cur_var has not yet been set
    progress(i/n * 100)
  }

  vars_to_create_pages4 <- intersect(
    rownames(report),
    util_map_labels(
      ifnotfound = NA_character_,
      attr(report, "study_data_dimnames")[[2]],
      attr(report, "meta_data"),
      to = attr(report, "label_col"))
  )

  vars_to_create_pages4 <-
    vars_to_create_pages4[vapply(vars_to_create_pages4, function(rn) {
    length(report[rn, ]) > 0
  }, FUN.VALUE = logical(1))]

  .md <- attr(report, "meta_data")
  if (!COMPUTED_VARIABLE_ROLE %in% colnames(.md)) {
    .md[[COMPUTED_VARIABLE_ROLE]] <- NA_character_
  }
  vars_to_create_pages4 <- vars_to_create_pages4[util_empty(util_map_labels(
    ifnotfound = NA_character_, # remove SSI vars from standard report.
    vars_to_create_pages4,
    .md,
    to = COMPUTED_VARIABLE_ROLE,
    from = attr(report, "label_col")))]

  i <- 0
  n <- length(vars_to_create_pages4)
  util_setup_rstudio_job("Page generation: Single Variable View", n = n)

  progress_msg("Page generation", "Generating single variable pages")

  cores <- util_get_cores_safe()
  if (!is.numeric(cores) || !util_is_integer(cores) || !is.finite(cores)) {
    cores <- 1
  }
  block_size <- block_load_factor * cores
  nblocks <- ceiling(n / block_size)
  for (cur_block in seq_len(nblocks) - 1) { # create the single variable pages
    block_indices <- seq(1 + cur_block * block_size,
                         min(cur_block * block_size + block_size,
                                              nrow(report)))
    vars_in_chunk <- vars_to_create_pages4[block_indices]
    vars_in_chunk <- intersect(
      vars_in_chunk,
      util_map_labels(
        ifnotfound = NA_character_,
        attr(report, "study_data_dimnames")[[2]],
        attr(report, "meta_data"),
        to = attr(report, "label_col"))
    )
    vars_in_chunk <- vars_in_chunk[!is.na(vars_in_chunk)]
    progress(i/n * 100)
    progress_msg("Page generation", sprintf("Single Variables %s",
                                            paste(sQuote(vars_in_chunk),
                                                  collapse = ", ")))

    # results, cll, repsum, function_alias_map,
    # meta_data, label_col, use_plot_ly, dir,
    # template, wd

    results <-
      lapply(setNames(nm = vars_in_chunk), function(nm) {
        report[nm, , as_raw = TRUE]
      })

    # find which dimensions are included in the report
    dims_in_rep <-
      unique(vapply(strsplit(colnames(report), "_"), `[[`, 1,
                    FUN.VALUE = character(1)))

    dims_in_rep <- util_sort_by_order(dims_in_rep, names(dims))

    # match the functions to the dimensions
    clls_in_rep <- lapply(setNames(nm = dims_in_rep),
                         function(prefix)
                           colnames(report)[startsWith(colnames(report),
                                                       prefix)])

    combined_list <- Map(function(results, cur_vars) list(results = results, # TODO: avoid re-compression
                                                          cur_vars = cur_vars), results = results, cur_vars = vars_in_chunk)

    use_plot_ly <- have_plot_ly
    note_meta <- warn_pred_meta
    function_alias_map <-
      attributes(attributes(report)$matrix_list)$function_alias_map

    f <-
      function(cml) {
        cml$results[vapply(cml$results, is.raw, FUN.VALUE = logical(1))] <-
          lapply(cml$results[vapply(cml$results, is.raw, FUN.VALUE = logical(1))], util_decompress)
        # rendered_repsum <- readRDS(rrs)
        util_html_for_var(results = cml$results, cur_var = cml$cur_vars,
                          dir = dir,
                          use_plot_ly = have_plot_ly,
                          template = template,
                          note_meta = warn_pred_meta,
                          rendered_repsum = rendered_repsum,
                          meta_data = meta_data,
                          label_col = label_col,
                          dims_in_rep = dims_in_rep,
                          clls_in_rep = clls_in_rep,
                          function_alias_map = function_alias_map)
      }
    use_plot_ly <- have_plot_ly
    f <- util_isolate_function(
      f,
      c(#"rrs",
        "rendered_repsum",
        "dir",
        "have_plot_ly",
        "template",
        "warn_pred_meta",
        "meta_data",
        "label_col",
        "dims_in_rep",
        "clls_in_rep",
        "function_alias_map"
      )
    )
    chunk_of_pages <-
      util_par_lapply_lb(
#        chunk.size = cores,
        X = combined_list,
        fun = f
    )

    progress_msg("Computed current chunk")
    chunk_of_pages <- do.call(`c`, chunk_of_pages)
    for (args in chunk_of_pages) {
      do.call(append_single_page, args)
    }
    i <- i + length(vars_in_chunk)
    progress(i/n * 100)
  }
  progress_msg("Computed all chunks")

  # SSI by indicator (COMPUTED_VARIABLE_ROLE) #####
  vars_to_create_pages4 <- intersect(
    rownames(report),
    util_map_labels(
      ifnotfound = NA_character_,
      vars_to_create_pages4,
      attr(report, "meta_data"),
      to = attr(report, "label_col"),
      from = attr(report, "label_col"))
  )

  vars_to_create_pages4 <-
    vars_to_create_pages4[vapply(vars_to_create_pages4, function(rn) {
      length(report[rn, ]) > 0
    }, FUN.VALUE = logical(1))]

  meta_data_cross <- util_normalize_cross_item(
    meta_data = attr(report, "meta_data"),
    meta_data_cross_item = attr(report, "meta_data_cross"),
    label_col = attr(report, "label_col")
  )

  .md <- attr(report, "meta_data")
  if (!COMPUTED_VARIABLE_ROLE %in% colnames(.md)) {
    .md[[COMPUTED_VARIABLE_ROLE]] <- NA_character_
  }

  ssi_roles <- # also used later in the other SSI output block
    util_map_labels(
      ifnotfound = NA_character_, # select SSI vars
      rownames(report),
      .md,
      to = COMPUTED_VARIABLE_ROLE,
      from = attr(report, "label_col"))

  ssi_roles_only <- unique(ssi_roles)
  ssi_roles_only <- ssi_roles_only[ssi_roles_only %in% COMPUTED_VARIABLE_ROLES]

  vars_for_each_role <- lapply(setNames(nm = ssi_roles_only),
                               function(curr_role) {
                                 names(which(ssi_roles == curr_role))
                               })

  i <- 0
  n <- length(ssi_roles_only) * 2
  util_setup_rstudio_job("Page generation: SSI Views", n = n)

  progress_msg("Page generation", "Generating SSI pages")

  fam <- attr(attr(report, "matrix_list"), "function_alias_map") # also used later in the next ssi section

  ssi_pages <- lapply(setNames(nm = ssi_roles_only), function(cvr) {
      if (!util_empty(cvr)) {
        ssi_men_lab <- util_get_concept_info("ssi",
                                             get("SSI_METRICS") == cvr,
                                             "menu_label",
                                             drop = TRUE)
        ssi_title <- util_get_concept_info("ssi",
                                           get("SSI_METRICS") == cvr,
                                           "result_caption",
                                           drop = TRUE)
        pg_list <- lapply(vars_for_each_role[[cvr]], function(variable) {
          .applicable_functions <-
            unlist(util_parse_assignments(util_get_concept_info("ssi",
                                                      get("SSI_METRICS") == cvr,
                                                                "functions",
                                                                drop = TRUE),
                                          multi_variate_text = TRUE))
          applicable_aliases_slots <- lapply(.applicable_functions,
                                       function(af) {
                                         as <- gsub("^.*\\.", "", af)
                                         af <- gsub("\\..*$", "", af)
                                         r <- subset(fam, get("name") == af,
                                                     "alias",
                                                     drop = TRUE)
                                         if (length(r) == 0) {
                                           r <- ""
                                         }
                                         res <- util_recycle(r, as)
                                         list(alias = res[[1]], slot = res[[2]])
                                       })
          rm(".applicable_functions")
          applicable_slots <- vapply(applicable_aliases_slots, `[[`, "slot", FUN.VALUE = character(1))
          applicable_aliases <- vapply(applicable_aliases_slots, `[[`, "alias", FUN.VALUE = character(1))
          # remove redundancies
          uniq_al_slo <- unique(cbind(applicable_aliases, applicable_slots))
          applicable_aliases <-
            unlist(uniq_al_slo[, "applicable_aliases", drop = TRUE])
          applicable_slots <-
            unlist(uniq_al_slo[, "applicable_slots", drop = TRUE])

          vg_title <- util_map_labels(
            util_map_labels(
              util_map_labels(
                variable,
                meta_data = .md,
                to = VAR_NAMES,
                from = attr(report, "label_col"),
                ifnotfound = NA_character_
              ),
              meta_data = meta_data_item_computation,
              from = VAR_NAMES,
              to = CHECK_ID,
              ifnotfound = NA_character_
            ),
            meta_data = meta_data_cross,
            from = CHECK_ID,
            to = CHECK_LABEL,
            ifnotfound = NA_character_
          )
          do.call(htmltools::tagList,
                  c(list(htmltools::h5(vg_title)),
                          mapply(SIMPLIFY = FALSE,
                                 alias = unique(applicable_aliases),
                                 slot = applicable_slots,
                                 FUN = function(alias, slot) {
                                   if (alias == "") {
                                     return(htmltools::HTML(""))
                                   }
                                   dqr <- lapply(report[variable, alias], `[`, slot) # TODO: Try to disentangle later, so we can combine all tables in one for a group.
                                   dqr <- util_combine_res(dqr)
                                   r <- do.call(
                                     htmltools::tagList,
                                     mapply(
                                       SIMPLIFY = FALSE,
                                       dqr = dqr,
                                       nm = names(dqr),
                                       FUN = function(dqr, nm) {
                                         util_pretty_print(
                                           dqr = dqr,
                                           nm = nm,
                                           is_single_var = !FALSE,
                                           use_plot_ly = have_plot_ly,
                                           meta_data = .md,
                                           label_col =
                                             attr(report, "label_col"),
                                           dir = dir,
                                           is_ssi = TRUE
                                         )
                                       }
                                     )
                                   )
                                   # care about memory usage
                                   rm(dqr)
                                   gc()
                                   r
                                 }))
          )
        })
        vars_to_show_in_summ <-
          util_map_labels(vars_for_each_role[[cvr]],
                          meta_data = attr(report, "meta_data"),
                          to = VAR_NAMES,
                          from = attr(report, "label_col"),
                          ifnotfound = NA_character_,
                          warn_ambiguous = FALSE)
        list( "Scales",
              prep_title_escape(ssi_men_lab, html = TRUE), # Modified title in the Menu drop-down
              paste0(prep_link_escape(cvr), ".html"),
              htmltools::tagList(
                htmltools::h2(ssi_title),
                # TODO: in the pie chart the total number of variables in not correct, it is 3 variables groups over 3 variable groups
                plot(repsum,
                     vars_to_include = "ssi",
                     filter = VAR_NAMES %in% vars_to_show_in_summ,
                     dont_plot = TRUE, disable_plotly =
                       disable_plotly),
                do.call(htmltools::tagList,
                        pg_list)
              ))
      } else {
        NULL
      }
  })

  ssi_pages <- ssi_pages[vapply(ssi_pages, length, FUN.VALUE = integer(1))
                         != 0]

  progress_msg("Page generation", "Mounting SSI pages by indicators")
  i <- 0
  n <- length(ssi_pages)
  for (args in ssi_pages) {
    do.call(append_single_page, args) # creates the indicator related pages because cur_var has not yet been set
    progress(i/n * 100)
  }

  # SSI by variable group #####
  about_scale <- c(CHECK_ID,
                   CHECK_LABEL,
                   SCALE_NAME,
                   SCALE_ACRONYM)

  missing_cols <- setdiff(about_scale, colnames(meta_data_cross))
  if (!!nrow(meta_data_cross))
    meta_data_cross[missing_cols] <- lapply(missing_cols, function(x) NA)

  groups_for_report <- meta_data_cross[
    rowSums(!util_empty(meta_data_cross[,
                                        intersect(ssi_roles_only,
                                                  colnames(meta_data_cross)),
                                        drop = FALSE])) > 0,
    intersect(c(about_scale, ssi_roles_only), colnames(meta_data_cross)),
    drop = FALSE]

  meta_data_item_computation <- attr(report, "meta_data_item_computation") # FIXME: Test w/o such sheets, that rendering wont crash, here!!

  ssi_concept_info <- util_get_concept_info("ssi")
  mapping_names_labels_ssi <- setNames(ssi_concept_info$result_caption,
                                       ssi_concept_info$SSI_METRICS)

  ssi_pages2 <- apply(meta_data_cross, 1, simplify = FALSE, function(rw) {
    if (!CHECK_ID %in% names(rw)) return(NULL);
    check_id <- rw[[CHECK_ID]]
    computed_vars_for_this_scale_vn <-
      subset(meta_data_item_computation, trimws(CHECK_ID) ==
               check_id, VAR_NAMES, drop = TRUE)
    if (length(computed_vars_for_this_scale_vn) > 0) {
      computed_vars_for_this_scale <-
        util_map_labels(computed_vars_for_this_scale_vn,
                        meta_data = meta_data,
                        to = label_col, from = VAR_NAMES,
                        ifnotfound = NA,
                        warn_ambiguous = FALSE)
      curr_roles <- ssi_roles[computed_vars_for_this_scale]

      .applicable_functions <-
        unique(unlist(util_parse_assignments(util_get_concept_info("ssi",
                                                                   get("SSI_METRICS") %in%
                                                                     curr_roles,
                                                                   "functions",
                                                                   drop = TRUE),
                      multi_variate_text = TRUE)))
      applicable_aliases_slots <- lapply(.applicable_functions,
                                         function(af) {
                                           as <- gsub("^.*\\.", "", af)
                                           af <- gsub("\\..*$", "", af)
                                           r <- subset(fam, get("name") == af,
                                                       "alias",
                                                       drop = TRUE)
                                           if (length(r) == 0) {
                                             r <- ""
                                           }
                                           res <- util_recycle(r, as)
                                           list(alias = res[[1]], slot = res[[2]])
                                         })
      rm(".applicable_functions")
      applicable_slots <- vapply(applicable_aliases_slots, `[[`, "slot", FUN.VALUE = character(1))
      applicable_aliases <- vapply(applicable_aliases_slots, `[[`, "alias", FUN.VALUE = character(1))


      # remove redundancies
      uniq_al_slo <- unique(cbind(applicable_aliases, applicable_slots))
      applicable_aliases <-
        unlist(uniq_al_slo[, "applicable_aliases", drop = TRUE])
      applicable_slots <-
        unlist(uniq_al_slo[, "applicable_slots", drop = TRUE])
      page <- do.call(htmltools::tagList, mapply(SIMPLIFY = FALSE, # parallel?
             alias = applicable_aliases,
             slot = applicable_slots,
             FUN = function(alias, slot) {
               if (alias == "") {
                 return(htmltools::HTML(""))
               }
               dqr <- report[computed_vars_for_this_scale, alias, drop = FALSE]
               dqr <- lapply(dqr, `[`, slot)
               dqr <- util_combine_res(dqr)
               if (slot %in% c("SummaryTable", "SummaryData")) {
                 old_slot <- slot
                 tb <- dqr[[alias]][[slot]]
                 vgs <- curr_roles[tb$Variables]
                 #-> correct names
                 slot <- "ResultData"
                 # avoid to auto-expand to var_names/label-column pair
                 colnames(tb)[colnames(tb) == "Variables"] <-
                   "Metrics"
                 rownames(tb) <- NULL
                 tb[["Metrics"]] <-
                   mapping_names_labels_ssi[vgs]
                 dqr[[alias]][[slot]] <- tb
                 dqr[[alias]][[old_slot]] <- NULL
               }
               r <- do.call(htmltools::tagList, mapply(SIMPLIFY = FALSE,
                                                       dqr = dqr,
                                                       nm = names(dqr),
                                                       FUN = function(dqr,
                                                                      nm) {
                                                         # variable name -> computed-varaible role -> SSI-header-title
                                                         res1 <-
                                                           util_pretty_print(
                                                             dqr = dqr,
                                                             nm = nm,
                                                             is_single_var = !FALSE,
                                                             use_plot_ly = have_plot_ly,
                                                             meta_data = .md,
                                                             label_col =
                                                               attr(report,
                                                                    "label_col"),
                                                             dir = dir,
                                                             is_ssi = TRUE)
                                                         return(htmltools::tagList(
                                                           htmltools::h5(mapping_names_labels_ssi[curr_roles[attr(attr(dqr, "call"), "entity_name")]]),
                                                           res1
                                                         ))
                                                       }
               ))
               # care about memory usage
               rm(dqr)
               gc()
               r
             }))

      # Whenever a function is assigned to the Scales menu, show it here
      fkts_in_scales <- util_get_concept_info("implementations",
                                              get("menu_location_report") ==
                                                "Scales",
                                              "function_R",
                                              drop = TRUE)
      for (fkt_in_scales in fkts_in_scales) {
        if (length(report[rw[[CHECK_LABEL]], fkt_in_scales]) == 1) {
          # add possible other SSI-only-function's results for this scale
          dqr <- report[rw[[CHECK_LABEL]], fkt_in_scales]
          dqr <- util_combine_res(dqr)
          page <- htmltools::tagList(page,
                                     htmltools::h5(util_alias2caption(fkt_in_scales, long = TRUE)),
                                     do.call(htmltools::tagList, mapply(SIMPLIFY = FALSE,
                                                                        dqr = dqr,
                                                                        nm = names(dqr),
                                                                        FUN = util_pretty_print,
                                                                        MoreArgs = list(
                                                                          is_single_var = TRUE,
                                                                          use_plot_ly = have_plot_ly,
                                                                          meta_data = .md,
                                                                          label_col =
                                                                            attr(report, "label_col"),
                                                                          dir = dir,
                                                                          is_ssi = TRUE)
                                     )))
          # care about memory usage
          rm(dqr)
          gc()
        }
      }
      short_title <- subset(meta_data_cross, trimws(CHECK_ID) ==
               check_id, CHECK_LABEL, drop = TRUE) # FIXME: Ensure, we have one
      long_title <- subset(meta_data_cross, trimws(CHECK_ID) ==
                              check_id, CHECK_LABEL, drop = TRUE) # FIXME: Add more metainfo from cil-sheet
      vars_to_show_in_summ <-
        util_map_labels(computed_vars_for_this_scale,
                        meta_data = attr(report, "meta_data"),
                        to = VAR_NAMES,
                        from = attr(report, "label_col"),
                        ifnotfound = NA_character_,
                        warn_ambiguous = FALSE)

      list("Scales",
           prep_title_escape(short_title, html = TRUE),
           paste0(prep_link_escape(short_title), ".html"),
           htmltools::tagList(
             htmltools::h2(long_title),
             plot(repsum,
                  vars_to_include = "ssi",
                  filter = VAR_NAMES %in% vars_to_show_in_summ,
                  dont_plot = TRUE,
                  disable_plotly =
                    disable_plotly),
             page))
    } else {
      NULL
    }
  })
  ssi_pages2 <- ssi_pages2[vapply(ssi_pages2, length, FUN.VALUE = integer(1))
                         != 0]

  progress_msg("Page generation", "Mounting SSI pages by variable groups")
  i <- 0
  n <- length(ssi_pages2)
  for (args in ssi_pages2) {
    do.call(append_single_page, args) # creates the indicator related pages because cur_var has not yet been set
    progress(i/n * 100)
  }


  # meta_data ----
  meta_data_frames <-
    grep("^meta_data", names(attributes(report)), value = TRUE)

  meta_data_titles <- list(
    meta_data_segment = util_attach_attr("Segment-level metadata",
                                         alternative_names = c(
                                           "meta_data_segment", "segment_level"
                                         )),
    meta_data_dataframe = util_attach_attr("Dataframe-level metadata",
                                           alternative_names = c(
                                             "meta_data_dataframe", "dataframe_level"
                                           )),
    meta_data_cross_item = util_attach_attr("Cross-item-level metadata",
                                            alternative_names = c(
                                              "meta_data_cross_item", "cross-item_level"
                                            )),
    meta_data = util_attach_attr("Item-level metadata",
                                 alternative_names = c(
                                   "meta_data", "item_level"
                                 )),
    meta_data_item_computation = util_attach_attr("Item-computation-level metadata",
                                                  alternative_names = c(
                                                    "meta_data_item_computation", "item_computation_level"
                                                  ))
  )

  for (mdn in meta_data_frames) {
    xlmd <- attr(report, mdn) # x level metadata

    if (mdn == "meta_data") {
      # these two columns should have been replaced by the v1->v2 conversion of the metadata in
      # prep_meta_data_v1_to_item_level_meta_data (.util_internal_normalize_meta_data)
      xlmd[[MISSING_LIST]] <- NULL
      xlmd[[JUMP_LIST]] <- NULL

      # there can be specific notes for item-level metadata attributes
      if (length(label_meta_data_hints)) {
        meta_data_h <- force(htmltools::tagList(
          htmltools::div( # TODO: warnings should be easy to understand if they appear in the report, and stand out while still matching the overall style of the report
            # style = htmltools::css(
            #   `background-color` = "#aaaaaa",
            #   width = "80%",
            #   left = "10%",
            #   position = "relative"
            # ),
            htmltools::tags$em(htmltools::pre(
              style = htmltools::css(
                `white-space` = "pre-wrap"
              ),
              do.call(
                paste,
                c(list(collapse = "\n"),
                  lapply(lapply(label_meta_data_hints, conditionMessage),
                         paste, collapse = "\n")))
            )))))
        label_meta_data_hints <- meta_data_h
      } else {
        label_meta_data_hints <- NULL
      }
    } else {
      label_meta_data_hints <- NULL
      # TODO: include warnings from util_normalize_cross_item on the cross-item metadata level page
    }

    xlmd[, !grepl("_TABLE$", colnames(xlmd))] <-
      util_df_escape(xlmd[, !grepl("_TABLE$", colnames(xlmd)),
                                drop = FALSE])

    for (cn in grep("_TABLE$", colnames(xlmd), value = TRUE)) {

      xlmd[!util_empty(xlmd[[cn]]), cn] <- vapply(FUN.VALUE = character(1),
                                                  xlmd[!util_empty(xlmd[[cn]]), cn],
                                                  FUN = function(tn) {
                                                    paste(as.character(htmltools::a(href = paste0(
                                                      prep_link_escape(tn, html = TRUE), ".html"), tn)),
                                                          collapse = "")
                                                  }
      )

    }

    # Hover text for headers of metadata tables
    text_to_display <- util_get_hovertext(mdn)

    # filter text_to_display only for headers actually present in the metadata
    text_to_display <- text_to_display[names(text_to_display)  %in% names(xlmd)]

    attr(xlmd, "description") <- text_to_display

    if (mdn == "meta_data") {
      hideCols <- setdiff(colnames(xlmd),
                          util_get_var_att_names_of_level(
                            VARATT_REQUIRE_LEVELS$TECHNICAL))
      col_tags <- c(list(
        "dataquieR metadata model" = intersect(colnames(xlmd),
                                               util_get_var_att_names_of_level(VARATT_REQUIRE_LEVELS$TECHNICAL)),
        "all" = colnames(xlmd)),
        lapply(lapply(setNames(nm = VARATT_REQUIRE_LEVELS_ORDER),
                      util_get_var_att_names_of_level,  cumulative = FALSE),
               function(x) {
                 union(VAR_NAMES, x)
               }
        ))
      col_tags[VARATT_REQUIRE_LEVELS_ORDER] <-
        lapply(col_tags[VARATT_REQUIRE_LEVELS_ORDER], function(x) {
            util_attach_attr(x, cssClass = "dq-hidden-col")
          })
      col_tags
    } else {
      known_cols <- colnames(xlmd)[vapply(colnames(xlmd),
                                          exists,
                                          FUN.VALUE = logical(1))]
      hideCols <- setdiff(colnames(xlmd), known_cols)
      col_tags <- list(
        "dataquieR metadata model" = known_cols,
        "all" = colnames(xlmd)
      )
    }

    meta_data_table <- util_html_table(
      xlmd, # generate links in VAR_NAMES/Variables, only possible if meta_data and label_col are available
      meta_data = meta_data,
      label_col = label_col,
      dl_fn = mdn,
      hideCols = hideCols,
      col_tags = col_tags
    )

    tmpl <- system.file("templates", template, paste0(mdn, ".html"),
                        package = utils::packageName())
    if (!file.exists(tmpl)) {
      tmpl <- system.file("templates", template,
                          paste0("generic_meta_data.html"),
                          package = utils::packageName())
    }

    meta_data_title <- meta_data_titles[[mdn]]
    if (length(meta_data_title) != 1 ||
        !is.character(meta_data_title) ||
        util_empty(meta_data_title)) {
      meta_data_title <- mdn
    }

    if (length(meta_data_table) == 0) {
      meta_data_table <- htmltools::p(paste(meta_data_title,
                                            "were not provided."))
    }

    append_single_page("Metadata",
                       meta_data_title,
                       paste0(mdn, ".html"),
                       htmltools::htmlTemplate(
                         tmpl,
                         meta_data_name = mdn,
                         meta_data_title =
                           # Use upper case for all words, not only the first one
                           gsub("(^|[[:space:]]|[[:punct:]])([[:alpha:]])",
                                "\\1\\U\\2",
                                meta_data_title,
                                perl = TRUE),
                         label_meta_data_hints = label_meta_data_hints,
                         meta_data_table = util_apply_full_page_table_class(meta_data_table)
                       )
    )
  }

  referred_tables <- attr(report, "referred_tables")
  if (length(referred_tables)) { # show all tables referred to by the report
    for (reftab in names(referred_tables)) {
      if (nrow(referred_tables[[reftab]]) > 1000) {
        append_single_page("Metadata",
                           util_attach_attr(paste("Table", sQuote(reftab)),
                                            alternative_names = c(reftab)),
                           paste0(prep_link_escape(reftab, html = TRUE), ".html"),
                           htmltools::h1(dQuote(reftab)),
                           util_html_table(
                             util_df_escape(head(referred_tables[[reftab]], 1000)), # generate links in VAR_NAMES/Variables, only possible if meta_data and label_col are available
                             meta_data = meta_data,
                             label_col = label_col,
                             dl_fn = reftab
                           ),
                           htmltools::tags$hr(),
                           htmltools::p(
                             sprintf("Showing only the first 1000 rows of a table with %d rows",
                                     nrow(referred_tables[[reftab]]))
                           )
        )
      } else {
        append_single_page("Metadata",
                           util_attach_attr(paste("Table", sQuote(reftab)),
                                            alternative_names = c(reftab)),
                           paste0(prep_link_escape(reftab, html = TRUE), ".html"),
                           htmltools::h1(dQuote(reftab)),
                           util_html_table(
                             util_df_escape(referred_tables[[reftab]]), # generate links in VAR_NAMES/Variables, only possible if meta_data and label_col are available
                             meta_data = meta_data,
                             label_col = label_col,
                             dl_fn = reftab
                           )
        )
      }
    }
  }


  rsts <- util_get_rule_sets()
  for (rstsn in names(rsts)) {
    ..tb <- util_df_escape(rsts[[rstsn]])
    ..tb[[GRADING_RULESET]] <-
      ifelse(..tb[[GRADING_RULESET]] == "0",
             "Default (0)",
             paste0("(", ..tb[[GRADING_RULESET]], ")"))
    if ("indicator_metric" %in% colnames(..tb)) {
      ..tb[["indicator_metric"]] <- lapply(..tb[["indicator_metric"]],
                                         function(im) {
                                           as.character(
                                             htmltools::span(
                                               title = im, # TODO: for downloads in Excel by datatables.js, use an alternative technical column
                                               util_translate_indicator_metrics(im)
                                               ))
                                         })
    }

    #add hover text to headers of table Grading ruleset
    text_to_display <- util_get_hovertext("grading_rulesets")
    attr(..tb, "description") <- text_to_display

    append_single_page("Metadata",
                       util_attach_attr(paste("Grading Ruleset", dQuote(rstsn)),
                                        alternative_names = c(rstsn)),
                       paste0("rulesets.html"),
                       htmltools::htmlTemplate(
                         system.file("templates", template,
                                     paste0("grading_ruleset_text.html"),
                                     package = utils::packageName()),
                         meta_data_name = paste0("grading_rulesets_", rstsn),
                         meta_data_title =
                           paste("Grading Rulesets", rstsn),
                         meta_data_table = util_apply_full_page_table_class(util_html_table(..tb,
                                                                                            dl_fn = paste0("grading_rulesets_", rstsn)))
                       )
    )
  }

  ..tb <- util_df_escape(util_get_ruleset_formats())
  if ("color" %in% colnames(..tb)) {
    ..tb[["color"]] <- lapply(..tb[["color"]],
                                         function(cl) {
                                           as.character(
                                             htmltools::span( # TODO: for downloads in Excel by datatables.js, use an alternative technical column
                                               style = htmltools::css(
                                                 background_color =
                                                   util_col2rgb(cl),
                                                 color =
                                                   util_get_fg_color(
                                                     util_col2rgb(cl))
                                               ),
                                               cl
                                             ))
                                         })
  }

  #add hover text to headers of table Ruleset formats
  text_to_display <- util_get_hovertext("grading_formats")
  attr(..tb, "description") <- text_to_display


  append_single_page("Metadata",
                     util_attach_attr(paste("Ruleset Formats"),
                                      alternative_names = c("grading_formats")), # TODO: More names
                     paste0("ruleset_formats.html"),
                     htmltools::htmlTemplate(
                       system.file("templates", template,
                                   paste0("ruleset_formats_text.html"),
                                   package = utils::packageName()),
                       meta_data_name = "grading_formats",
                       meta_data_title =
                         "Ruleset Formats",
                       meta_data_table =
                         util_apply_full_page_table_class(util_html_table(..tb,
                                                                          dl_fn = "grading_formats"))
                     )
  )


  pages
}

util_apply_full_page_table_class <- function(x) {
  util_ensure_suggested(pkg = c("htmltools"),
                        goal = "generate interactive HTML-reports.")

  htmltools::div(class = "fullpage-table", x)
}

util_is_empty_html <- function(x) {
  if (is.null(x)) return(TRUE)

  # convert to character representation
  txt <- paste0(as.character(x), collapse = "")
  !nzchar(trimws(txt))
}

#' Detects variable values exceeding limits defined in metadata
#'
#' @description
#'
#' Inadmissible numerical values can be of type integer or float. This
#' implementation requires the definition of intervals in the metadata to
#' examine the admissibility of numerical study data.
#'
#' This helps identify inadmissible measurements according to
#' hard limits (for multiple variables).
#'
#' [Indicator]
#'
#' @details
#' ### Algorithm of this implementation:
#'
#'  - Remove missing codes from the study data (if defined in the metadata)
#'  - Interpretation of variable specific intervals as supplied in the metadata.
#'  - Identification of measurements outside defined limits. Therefore two
#'    output data frames are generated:
#'    - on the level of observation to flag each deviation, and
#'    - a summary table for each variable.
#'  - A list of plots is generated for each variable examined for limit
#'    deviations. The histogram-like plots indicate respective limits as well
#'    as deviations.
#'  - Values exceeding limits are removed in a data frame of modified study data
#'
#'
#' @export
#'
#' @inheritParams .template_function_indicator
#' @param meta_data_cross_item [meta_data_cross]
#' @param cross_item_level [data.frame] alias for `meta_data_cross_item`
#' @param `cross-item_level` [data.frame] alias for `meta_data_cross_item`
#'
#' @param resp_vars [variable list] the name of the measurement variables
#' @param limits [enum] HARD_LIMITS | SOFT_LIMITS | DETECTION_LIMITS. what
#'                                             limits from metadata to check for
#' @param return_flagged_study_data [logical] return `FlaggedStudyData` in the
#'                                            result
#' @param return_limit_categorical [logical] if TRUE return limit deviations also
#'                                           for categorical variables
#' @param show_obs [logical] Should (selected) individual observations be marked
#'                           in the figure for continuous variables?
#'
#' @inheritParams acc_distributions
#'
#' @importFrom ggplot2 ggplot geom_histogram scale_fill_manual coord_flip labs theme_minimal theme geom_bar geom_vline annotate scale_linetype_manual
#' @importFrom stats setNames IQR complete.cases
#' @importFrom grDevices colorRampPalette gray.colors
#'
#' @return a list with:
#'   - `FlaggedStudyData` [data.frame] related to the study data by a 1:1
#'                                   relationship, i.e. for each observation is
#'                                   checked whether the value is below or above
#'                                   the limits. Optional, see
#'                                   `return_flagged_study_data`.
#'   - `SummaryTable` [data.frame] summarizing limit deviations for each
#'                                 variable.
#'   - `SummaryData` [data.frame] summarizing limit deviations for each
#'                                 variable for a report.
#'   - `SummaryPlotList` [list] of [ggplot2::ggplot]s The plots for each variable are
#'                            either a histogram (continuous) or a
#'                            barplot (discrete).
#'   - `ReportSummaryTable`: heatmap-like data frame about limit violations
#'
#' @seealso
#' - [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_limit_deviations.html
#' )
con_limit_deviations <- function(resp_vars = NULL,
                                 study_data,
                                 label_col,
                                 item_level = "item_level",
                                 meta_data_cross_item = "cross-item_level",
                                 limits = NULL,
                                 flip_mode = "noflip",
                                 return_flagged_study_data = FALSE,
                                 return_limit_categorical = TRUE,
                                 meta_data = item_level,
                                 cross_item_level,
                                 `cross-item_level`,
                                 meta_data_v2,
                                 show_obs = TRUE) {
  # make R CMD check happy
  Limits <- NULL
  Number <- NULL
  Section <- NULL
  all_of <- NULL
  n  <- NULL
  n_viol <- NULL
  values <- NULL
  segment_number <- NULL

  # PREP - steps -------
  # names vector containing the types of limits: e.g., HARD_LIMITS
  known_limits <- unlist(
    WELL_KNOWN_META_VARIABLE_NAMES[intersect(
      grep("LIMIT", WELL_KNOWN_META_VARIABLE_NAMES),
      which(vapply(WELL_KNOWN_META_VARIABLE_NAMES,
                   function(ll) {
                     attr(ll, "var_att_required") != "technical"
                   },
                   FUN.VALUE = logical(1)))
    )]
  )
  # map metadata to study data
  util_maybe_load_meta_data_v2()
  util_ck_arg_aliases()
  prep_prepare_dataframes()
  util_correct_variable_use(resp_vars,
                            allow_null = TRUE,
                            allow_more_than_one = TRUE,
                            allow_all_obs_na = FALSE,
                            need_type = "integer|float|datetime|time")
  try({
    meta_data_cross_item <- util_normalize_cross_item(
      meta_data = meta_data,
      meta_data_cross_item = meta_data_cross_item,
      label_col = label_col)
  }, silent = TRUE
  )
  if (!is.data.frame(meta_data_cross_item))
    meta_data_cross_item <- data.frame()
  util_expect_scalar(return_flagged_study_data, check_type = is.logical)
  util_expect_scalar(return_limit_categorical, check_type = is.logical)

  if (!DATA_PREPARATION %in% colnames(meta_data_cross_item)) {
    meta_data_cross_item[[DATA_PREPARATION]] <-
      rep("LABEL | MISSING_NA", nrow(meta_data_cross_item)) # never use LIMITS here, we are assessing the limits
  } else {
    meta_data_cross_item[[DATA_PREPARATION]] <-
      gsub(
        ignore.case = TRUE,
        perl = TRUE,
        "LIMITS",
        "",
        meta_data_cross_item[[DATA_PREPARATION]]) # never use LIMITS here, we are assessing the limits
  }

  # COMPLEX LIMITS (from cross-item_level metadata)-------
  # if *_LIMITS in cross-item-level, compute contradictions
  # with renamed column,
  # each (renamed to CONTRADICTION_TERM) and amend results
  given_cross_limits <- intersect(known_limits,
                                  colnames(meta_data_cross_item))
  no_warn_no_lim <- FALSE
  if (!all(util_empty(as.vector(meta_data_cross_item[, given_cross_limits,
                                                     FALSE])))) {
    no_warn_no_lim <- TRUE
    #ITERATE over each limit column in cross-item_level metadata
    all_viols <- lapply(given_cross_limits, function(lim) {
      #reduce the cross-item_level content to only the rows that contains limits
      cur_cross <-
        meta_data_cross_item[!util_empty(meta_data_cross_item[[lim]]), , FALSE]
      if (nrow(cur_cross) == 0) {
        return(NULL)
      } else {
        #Select only the limit column of interest in this iteration
        cur_cross <- cur_cross[, intersect(c(VARIABLE_LIST,
                                             CONTRADICTION_TERM,
                                             CONTRADICTION_TYPE,
                                             CHECK_ID,
                                             CHECK_LABEL,
                                             DATA_PREPARATION,
                                             lim),
                                           colnames(cur_cross)
        ),
        FALSE]
        #overwrite the content of the contraction_term with the rule of the limit
        cur_cross[[CONTRADICTION_TERM]][!util_empty(cur_cross[[lim]])] <-
          paste("not (", cur_cross[[lim]], ")")[!util_empty(cur_cross[[lim]])]
        # Set the contraction type to logical for hard limits and empirical for
        # all other types of limit
        if (lim == HARD_LIMITS)
          cur_cross[[CONTRADICTION_TYPE]] <- "LOGICAL"
        else
          cur_cross[[CONTRADICTION_TYPE]] <- "EMPIRICAL"
        # get names of the variables that the limits refer to
        vl <- util_parse_assignments(cur_cross[[VARIABLE_LIST]],
                                     multi_variate_text = TRUE)
        # only accept one variable per limit
        util_stop_if_not(
          "Only one variable can be the reference for complex limits" =
            all(vapply(vl, length, FUN.VALUE = integer(1)) == 1))
        # check the function argument resp_vars
        which_vars <- resp_vars
        if (length(which_vars) == 0) {
          which_vars <- unique(unlist(vl))
        }
        which_rows <- vapply(vl, function(cv) {
          length(intersect(which_vars, cv)) == 1
        }, FUN.VALUE = logical(1))
        cur_cross <- cur_cross[which_rows, , FALSE]
        if (nrow(cur_cross) == 0) {
          return(NULL)
        } else {
          # call contradiction redcap for the limit rules
          r <- suppressWarnings(con_contradictions_redcap(
            study_data = study_data,
            meta_data = meta_data,
            label_col = label_col,
            meta_data_cross_item = cur_cross)$VariableGroupTable)
          # merge by check_id
          r[[VARIABLE_LIST]] <-
            merge(cur_cross, r, by = CHECK_ID, all = FALSE,
                  suffixes = c("", ".y"))[[VARIABLE_LIST]]
          dt <- util_map_labels(
            r[[VARIABLE_LIST]],
            meta_data = meta_data,
            to = DATA_TYPE,
            from = label_col
          )
          #add temporary another column with data_type
          r$data_type <- dt[r$VARIABLE_LIST]
          if (lim == HARD_LIMITS) {
            #create a data frame for datetime and rename the columns
            r_datetime_hard <- r[r$data_type == DATA_TYPES$DATETIME, ]
            if(nrow(r_datetime_hard) > 0) {
              colnames(r_datetime_hard)[
                colnames(r_datetime_hard) == "NUM_con_con_contc"] <-
                "NUM_con_rvv_itdat"

              colnames(r_datetime_hard)[
                colnames(r_datetime_hard) == "PCT_con_con_contc"] <-
                "PCT_con_rvv_itdat"
              #Add empty rows for numeric result
              r_datetime_hard$NUM_con_rvv_inum <- NA
              r_datetime_hard$PCT_con_rvv_inum <- NA
            } else {
              r_datetime_hard$NUM_con_rvv_itdat <- numeric(0)
              r_datetime_hard$PCT_con_rvv_itdat <- numeric(0)
              #Add empty rows for numeric result
              r_datetime_hard$NUM_con_rvv_inum <- numeric(0)
              r_datetime_hard$PCT_con_rvv_inum <- numeric(0)
            }
            # remove unneeded column
            r_datetime_hard$NUM_con_con_contu <- NULL
            r_datetime_hard$PCT_con_con_contu <- NULL
            r_datetime_hard$NUM_con_con_contc <- NULL
            r_datetime_hard$PCT_con_con_contc <- NULL
            r_datetime_hard$NUM_con_con <- NULL
            r_datetime_hard$PCT_con_con <- NULL
            #create a data frame for numeric variables
            r_numeric_hard <- r[r$data_type != DATA_TYPES$DATETIME, ]
            if(nrow(r_numeric_hard) > 0) {
              colnames(r_numeric_hard)[
                colnames(r_numeric_hard) == "NUM_con_con_contc"] <-
                "NUM_con_rvv_inum"
              colnames(r_numeric_hard)[
                colnames(r_numeric_hard) == "PCT_con_con_contc"] <-
                "PCT_con_rvv_inum"
              #Add empty rows for numeric result
              r_numeric_hard$NUM_con_rvv_itdat <- NA
              r_numeric_hard$PCT_con_rvv_itdat <- NA
            } else {
              r_numeric_hard$NUM_con_rvv_inum <- numeric(0)
              r_numeric_hard$PCT_con_rvv_inum <- numeric(0)
              #Add empty rows for numeric result
              r_numeric_hard$NUM_con_rvv_itdat <- numeric(0)
              r_numeric_hard$PCT_con_rvv_itdat <- numeric(0)
            }
            # remove unneeded column
            r_numeric_hard$NUM_con_con_contu <- NULL
            r_numeric_hard$PCT_con_con_contu <- NULL
            r_datetime_hard$NUM_con_con_contc <- NULL
            r_datetime_hard$PCT_con_con_contc <- NULL
            r_numeric_hard$NUM_con_con <- NULL
            r_numeric_hard$PCT_con_con <- NULL
            #Combine the tables
            r_numeric_hard <- r_numeric_hard[, colnames(r_datetime_hard)]
            final_r_hard <- rbind(r_datetime_hard, r_numeric_hard)
            final_r_hard$limits <- lim
          } else {
            #If DETECTION OR SOFT LIMITS
            #create a data frame for datetime and rename the columns
            r_datetime <- r[r$data_type == DATA_TYPES$DATETIME, ]
            if(nrow(r_datetime) > 0) {
              colnames(r_datetime)[
                colnames(r_datetime) == "NUM_con_con_contu"] <-
                "NUM_con_rvv_utdat"
              colnames(r_datetime)[
                colnames(r_datetime) == "PCT_con_con_contu"] <-
                "PCT_con_rvv_utdat"
              #Add empty rows for numeric result
              r_datetime$NUM_con_rvv_unum <- NA
              r_datetime$PCT_con_rvv_unum <- NA
            } else {
              r_datetime$NUM_con_rvv_utdat <- numeric(0)
              r_datetime$PCT_con_rvv_utdat <- numeric(0)
              #Add empty rows for numeric result
              r_datetime$NUM_con_rvv_unum <- numeric(0)

              r_datetime$PCT_con_rvv_unum <- numeric(0)
            }
            # remove unneeded column
            r_datetime$NUM_con_con_contc <- NULL
            r_datetime$PCT_con_con_contc <- NULL
            r_datetime$NUM_con_con_contu <- NULL
            r_datetime$PCT_con_con_contu <- NULL
            r_datetime$NUM_con_con <- NULL
            r_datetime$PCT_con_con <- NULL
            #create a data frame for numeric variables
            r_numeric <- r[r$data_type != DATA_TYPES$DATETIME, ]
            if(nrow(r_numeric) > 0) {
              colnames(r_numeric)[
                colnames(r_numeric) == "NUM_con_con_contu"] <-
                "NUM_con_rvv_unum"
              colnames(r_numeric)[
                colnames(r_numeric) == "PCT_con_con_contu"] <-
                "PCT_con_rvv_unum"
              #Add empty rows for numeric result
              r_numeric$NUM_con_rvv_utdat <- NA
              r_numeric$PCT_con_rvv_utdat <- NA
            } else {
              r_numeric$NUM_con_rvv_unum <- numeric(0)
              r_numeric$PCT_con_rvv_unum <- numeric(0)
              #Add empty rows for numeric result
              r_numeric$NUM_con_rvv_utdat <- numeric(0)
              r_numeric$PCT_con_rvv_utdat <- numeric(0)
            }
            # remove unneeded column
            r_numeric$NUM_con_con_contc <- NULL
            r_numeric$PCT_con_con_contc <- NULL
            r_numeric$NUM_con_con <- NULL
            r_numeric$PCT_con_con <- NULL
            #Combine the tables
            r_numeric <- r_numeric[, colnames(r_datetime)]
            final_r <- rbind(r_datetime, r_numeric)
            final_r$limits <- lim
          }
          #merge the results for different limits
          if (exists("final_r")) {
            if (exists("final_r_hard")) {
              final_r <- merge(
                x = final_r,
                y = final_r_hard,
                by = intersect(names(final_r),
                               names(final_r_hard)),
                all = TRUE
              )
            } else {
              final_r
            }
          } else {
            final_r_hard
          }
        }
      }
    })
    all_viols <- util_rbind(data_frames_list = all_viols)
    # obtain a data frame with all the results for limits from the cross-item level
    # and a column indicating what limit is
    r_complex <-
      all_viols[, intersect(c(VARIABLE_LIST,
                              "NUM_con_rvv_unum",
                              "PCT_con_rvv_unum",
                              "NUM_con_rvv_utdat",
                              "PCT_con_rvv_utdat",
                              "NUM_con_rvv_inum",
                              "PCT_con_rvv_inum",
                              "NUM_con_rvv_itdat",
                              "PCT_con_rvv_itdat",
                              "limits"),
                            colnames(all_viols)
      ),
      FALSE]
    if(nrow(r_complex) > 0) {
      colnames(r_complex)[[1]] <- "Variables"
      r_complex$N <- nrow(study_data)
    } else {
      r_complex <- NULL
    }
  } else {
    r_complex <- NULL
  }


  #Variables with limits in cross-item level----
  if (!is.null(r_complex)) {
    vars_in_cil <- unique(r_complex$Variables)
  } else {
    vars_in_cil <- NULL
  }

  # Which limits should be assessed?
  if (!is.null(limits) && !(all(limits %in% colnames(meta_data)))) {
    limits_miss <- limits[!(limits %in% colnames(meta_data))]
    util_warning(paste0("The limit deviation check cannot be performed for ",
                        paste(dQuote(limits_miss), collapse = ", "),
                        ". There is no matching column in the metadata."),
                 applicability_problem = TRUE)
  }
  known_limits <- unlist(
    WELL_KNOWN_META_VARIABLE_NAMES[intersect(
      grep("LIMIT", WELL_KNOWN_META_VARIABLE_NAMES),
      which(vapply(WELL_KNOWN_META_VARIABLE_NAMES,
                   function(ll) {
                     attr(ll, "var_att_required") != "technical"
                   },
                   FUN.VALUE = logical(1)))
    )]
  )
  given_limits <- colnames(meta_data)[which(colnames(meta_data) %in%
                                              c(known_limits, limits))]
  LIMITS <- given_limits

  if (length(LIMITS) > 0) {
    empty_limit_column <- apply(meta_data[, LIMITS, drop = FALSE], 2,
                                function(cc) { all(util_empty(cc)) })
    if (any(empty_limit_column)) {
      LIMITS <- names(empty_limit_column)[!empty_limit_column]
    }
  }

  if (length(LIMITS) == 0) {
    util_message(paste0("Hint: If your metadata contains a column with limits ",
                        "for which the limit deviation check should be ",
                        "performed, try to pass the name of the column ",
                        "to the ", sQuote("limits"), " argument."))
    util_error(paste0("The limit deviation check cannot be performed without ",
                      "a metadata column specifying suitable limits."),
               applicability_problem = TRUE)
  }

  if (length(resp_vars) == 0) {
    any_limits <- apply(meta_data[, LIMITS, drop = FALSE], 1,
                        function(rr) { !all(util_empty(rr)) })
    if (all(!any_limits)) {
      util_error("No variables with defined limits.",
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    } else {
      util_message(paste0("All variables for which limits are specified ",
                          "in the metadata are used."),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = TRUE)
      resp_vars <- meta_data[[label_col]][any_limits &
                                            meta_data$DATA_TYPE != "string"]
      resp_vars_iteml <- meta_data[[label_col]][any_limits &
                                                  meta_data$DATA_TYPE != "string"]
      if (length(resp_vars) > 0) {
        rvs_all_na <- apply(ds1[, resp_vars, drop = FALSE], 2,
                            function(cc) { all(is.na(cc)) })
        if (any(rvs_all_na)) {
          resp_vars <- names(rvs_all_na)[!rvs_all_na]
        }
      }
    }
  } else {
    any_limits <- vapply(resp_vars, FUN.VALUE = logical(1), function(rvs) {
      !all(util_empty(meta_data[meta_data[[label_col]] == rvs,
                                LIMITS, drop = FALSE])) |
        rvs %in% r_complex$Variables
    })
    if (all(!any_limits)) {
      util_error("No limits specified.",
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    } else {
      rvs_with_lim <- names(any_limits)[any_limits]
      if (length(rvs_with_lim) < length(unique(resp_vars))) {
        util_message(paste0("No limits specified for ",
                            paste(resp_vars[!(resp_vars %in% rvs_with_lim)],
                                  collapse = ", "),
                            "."),
                     applicability_problem = TRUE,
                     intrinsic_applicability_problem = TRUE)
      }
      resp_vars <- rvs_with_lim
      any_item_lev_limits <- vapply(resp_vars,
                                    FUN.VALUE = logical(1), function(rvs) {
                                      !all(util_empty(meta_data[meta_data[[label_col]] == rvs,
                                                                LIMITS, drop = FALSE]))
                                    })
      nmrvl <- names(which(any_item_lev_limits,  useNames = TRUE))
      resp_vars_iteml <-
        nmrvl[util_map_labels(nmrvl, meta_data = meta_data, to = DATA_TYPE,
                              from = label_col, ifnotfound = "",
                              warn_ambiguous = FALSE) %in%
                unlist(setdiff(DATA_TYPES,
                               DATA_TYPES$STRING))]

      resp_vars_iteml <- intersect(resp_vars_iteml, resp_vars) # TODO: needed?
      if (length(resp_vars_iteml) == 0) {
        rm(resp_vars_iteml)
      }
    }
  }

  # Resp_vars----
  # resp_vars now contains all variables with limits in item level or cross item level
  # resp_vars_iteml contains variables with limits in item level
  # vars_in_cil contains variables with limits in cross-item level
  # ONLY cross-item_level metadata present


  # create the subset of variables in the 2 groups: the ones that uses -----
  # item level limits and the ones using complex limits

  #####################################################
  if(!is.null(vars_in_cil) & exists("resp_vars_iteml")) {
    #remove from the vector of the variables inside item_level
    # the variables that are also present in cross-item level
    resp_vars_iteml <- resp_vars_iteml[!resp_vars_iteml %in% vars_in_cil]
  }
  # Variable with only item_level metadata limits ----

  if (exists("resp_vars_iteml") && length(resp_vars_iteml) > 0 ) {
    # In both cases ('resp_vars_iteml' were specified by the user or
    # selected automatically), we have to check whether the given limits can
    # be interpreted as intervals.
    md_lim_int <- setNames(
      nm = resp_vars_iteml,
    lapply(resp_vars_iteml, function(rvs) {
      lim <- meta_data[meta_data[[label_col]] == rvs, LIMITS, drop = FALSE]
      lim_int <- lapply(lim, util_parse_interval)
      is_interval <- vapply(lim_int, FUN.VALUE = logical(1),
                            function(int) { inherits(int, "interval") })
      wrong_limits <- !is_interval != as.vector(is.na(lim))
      if (any(wrong_limits)) {
        util_warning(
          paste0("The limits given in",
                 paste(sQuote(names(lim)[wrong_limits]), collapse = ", "),
                 " for ", sQuote(rvs), " cannot be interpreted as interval."),
          applicability_problem = TRUE)
      }
      if (all(!is_interval)) {
        return(NA)
      } else {
        if(return_limit_categorical == FALSE) {
          var_scale_lev <- meta_data[meta_data[[label_col]] == rvs, SCALE_LEVEL, drop = FALSE]
          is_categorical <- var_scale_lev %in% c(SCALE_LEVELS$NOMINAL,
                                                 SCALE_LEVELS$ORDINAL)
          if (is_categorical == TRUE) {
            util_message(
              paste0("The limits for ", sQuote(rvs),
                     " cannot be interpreted as the response",
                     " variable is categorical"),
              applicability_problem = TRUE,
              intrinsic_applicability_problem = TRUE)
          }
          if (is_categorical) return(NA)
          return(lim_int[is_interval])
        } else {
          return(lim_int[is_interval])
        }
      }
    }))

  drop_rvs <- is.na(md_lim_int)
  if (all(drop_rvs)) {
    util_error("No feasible check for limit deviations.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }
  if (any(drop_rvs)) {
    resp_vars_iteml <- setdiff(resp_vars_iteml, names(md_lim_int)[drop_rvs])
  }

  # Typkennzeichnung
  is_datetime_var <- vapply(resp_vars_iteml, function(rv) {
    meta_data[["DATA_TYPE"]][meta_data[[label_col]] == rv] %in% DATA_TYPES$DATETIME
  }, FUN.VALUE = logical(1))
  is_time_var <- vapply(resp_vars_iteml, function(rv) {
    meta_data[["DATA_TYPE"]][meta_data[[label_col]] == rv] %in% DATA_TYPES$TIME
  }, FUN.VALUE = logical(1))

  # Segmente (below/within/above) je Limit
  md_lim_all_segments <- setNames(
    nm = resp_vars_iteml,
    lapply(resp_vars_iteml, function(rv) {
      lapply(md_lim_int[[rv]], function(int) {
        ll <- int
        ll$upp <- int$low
        ll$inc_u <- !int$inc_l
        ll$low <- as.numeric(-Inf)
        ll$inc_l <- FALSE
        ul <- int
        ul$upp <- as.numeric(Inf)
        ul$inc_u <- FALSE
        ul$low <- int$upp
        ul$inc_l <- !int$inc_u
        return(list(
          "below" = ll,
          "within" = int,
          "above" = ul
        ))
      })
    }))

  # Werte klassifizieren -------------------------------------------------------
  sd_lim <- setNames(
    nm = resp_vars_iteml,
    lapply(resp_vars_iteml, function(rvs) {
      col_rv <- ds1[, rvs]
      lim <- md_lim_int[[rvs]]
      col_int <- lapply(lim, function(int) {
        col_out <- rep(NA, length(col_rv))
        if (int$inc_l) below <- col_rv < int$low else below <- col_rv <= int$low
        within <- redcap_env$`in`(col_rv, int)
        if (int$inc_u) above <- col_rv > int$upp else above <- col_rv >= int$upp
        col_out[which(below)]  <- "below"
        col_out[which(within)] <- "within"
        col_out[which(above)]  <- "above"
        return(factor(col_out, levels = c("below", "within", "above")))
      })
      return(col_int)
    })
  )

  # Hard-Limits zuerst ausschließen
  sd_lim_HL_check <- setNames(
    nm = resp_vars_iteml,
    lapply(resp_vars_iteml, function(rvs) {
      sd_lim_rvs <- sd_lim[[rvs]]
      if ("HARD_LIMITS" %in% names(sd_lim_rvs) && length(sd_lim_rvs) > 1) {
        hl_viol_ind <- which(sd_lim_rvs[["HARD_LIMITS"]] %in% c("below", "above"))
        if (length(hl_viol_ind) > 0) {
          other_lim <- setdiff(names(sd_lim_rvs), "HARD_LIMITS")
          sd_lim_rvs[other_lim] <- lapply(other_lim, function(ol) {
            sd_lim_rvs[[ol]][hl_viol_ind] <- NA
            return(sd_lim_rvs[[ol]])
          })
        }
      }
      return(sd_lim_rvs)
    })
  )

  sd_n_obs <- setNames(
    nm = resp_vars_iteml,
    lapply(resp_vars_iteml, function(rvs) {
      lapply(sd_lim_HL_check[[rvs]], function(cc) {
        sum(!is.na(cc))
      })
    })
  )

  # get frequency counts of values below, within and above limits
  freq_lim_viol <- setNames(
    nm = resp_vars_iteml,
    lapply(resp_vars_iteml, function(rvs) {
      sd_lim_rvs <- sd_lim_HL_check[[rvs]]
      tab_rvs <- lapply(sd_lim_rvs, function(res_int) {
        t1 <- table(res_int)
        setNames(nm = names(t1), as.vector(t1))
      })
      tab <- as.data.frame(t(as.data.frame(tab_rvs)))
      tab$samplesize <- apply(tab,1, sum)
      tab
    }))

  # Plot-Vorbereitung ----------------------------------------------------------
  ref_env <- environment()
  base_col <- colorRampPalette(
    colors = c("#2166AC","#fdd49e","#fc8d59","#d7301f","#B2182B","#7f0000")
  )(length(given_limits) + 1)
  base_lty <- c(2, 4:6)
  spec_txt <- element_text(colour = "black", hjust = .5, vjust = .5, face = "plain")
  spec_lines <- data.frame(limits = "HARD_LIMITS", color = "#B2182B", lty = 1)
  given_lim_wo_hard <- setdiff(given_limits, "HARD_LIMITS")
  n_spec_lim <- length(given_lim_wo_hard)
  if (n_spec_lim > 0) {
    spec_lines <- rbind(
      spec_lines,
      data.frame(
        limits = given_lim_wo_hard,
        color = rep(gray.colors(n = ceiling(n_spec_lim/length(base_lty))),
                    each = length(base_lty))[1:n_spec_lim],
        lty = c(rep(base_lty, floor(n_spec_lim/length(base_lty))),
                base_lty[seq_len(n_spec_lim %% length(base_lty))])
      )
    )
  }

  # ein Plot pro Variable
  plot_list <- lapply(setNames(nm = resp_vars_iteml), function(rv) {
    # Limits einsammeln (Werte unname(), Namen separat)
    lower_limits <- do.call(c, unname(lapply(md_lim_int[[rv]], '[[', "low")))
    if (any(is.infinite(lower_limits))) lower_limits[is.infinite(lower_limits)] <- NA
    upper_limits <- do.call(c, unname(lapply(md_lim_int[[rv]], '[[', "upp")))
    if (any(is.infinite(upper_limits))) upper_limits[is.infinite(upper_limits)] <- NA
    lower_limit_names <- names(md_lim_int[[rv]])
    upper_limit_names <- names(md_lim_int[[rv]])
    rv_limit_names    <- names(md_lim_int[[rv]])

    # Daten
    rv_data <- data.frame("values" = ds1[[rv]], sd_lim_HL_check[[rv]])
    rv_data <- rv_data[complete.cases(rv_data[, "values"]), ]

    # numerisch für hist()
    if (is_datetime_var[[rv]]) {
      rv_data$values <- as.numeric(rv_data$values)                           # POSIX sec
    } else if (is_time_var[[rv]]) {
      v <- rv_data$values
      if (inherits(v, "hms")) {
        rv_data$values <- suppressWarnings(as.numeric(v))                    # Sekunden
      } else if (inherits(v, "difftime")) {
        rv_data$values <- suppressWarnings(as.numeric(v, units = "secs"))
      } else if (is.character(v)) {
        rv_data$values <- suppressWarnings(as.numeric(util_parse_time(v)))
      } else if (!is.numeric(v)) {
        rv_data$values <- suppressWarnings(as.numeric(hms::as_hms(v)))
      }
    }

    # Bereiche
    max_data <- max(rv_data$values, na.rm = TRUE)
    min_data <- min(rv_data$values, na.rm = TRUE)
    max_plot <- max(c(max_data, lower_limits, upper_limits), na.rm = TRUE)
    min_plot <- min(c(min_data, lower_limits, upper_limits), na.rm = TRUE)

    # Füllungen
    rv_data$viol_n_limits <- apply(rv_data[, rv_limit_names, drop = FALSE], 1,
                                   function(rr) length(which(rr != "within")))
    rv_data$limit_violations <- paste0("detected (", rv_data$viol_n_limits, ")")
    rv_data$limit_violations[rv_data$viol_n_limits == 0] <- "none"
    rv_data$fill <- vapply(rv_data$viol_n_limits,
                           FUN.VALUE = "character(1)",
                           FUN = function(vv) base_col[vv + 1])
    if ("HARD_LIMITS" %in% colnames(rv_data)) {
      rv_data$limit_violations[rv_data$HARD_LIMITS != "within"] <- "severe"
      rv_data$fill[rv_data$HARD_LIMITS != "within"] <- rev(base_col)[1]
    }
    spec_bars <- unique(rv_data[, c(rv_limit_names, "limit_violations", "fill")])

    plot_histogram <-
      meta_data[[SCALE_LEVEL]][which(meta_data[[label_col]] == rv)] %in%
      c(SCALE_LEVELS$INTERVAL, SCALE_LEVELS$RATIO)

    plot_histogram_integer <-
      meta_data[[SCALE_LEVEL]][which(meta_data[[label_col]] == rv)] %in%
      c(SCALE_LEVELS$INTERVAL, SCALE_LEVELS$RATIO) &&
      meta_data[[DATA_TYPE]][which(meta_data[[label_col]] == rv)] %in% c("integer")

    xlb <- prep_get_labels(resp_vars = rv,
                           resp_vars_match_label_col_only = TRUE,
                           label_class = "SHORT",
                           label_col = label_col,
                           item_level = meta_data)

    if (plot_histogram) {
      # Segmente bestimmen
      plot_segments <- rv_data %>%
        dplyr::arrange(values) %>%
        dplyr::select(all_of(rv_limit_names)) %>%
        dplyr::group_by_all(.) %>%
        dplyr::count() %>%
        as.data.frame(., drop = FALSE)

      plot_segments_intervals <-
        apply(plot_segments[, rv_limit_names, drop = FALSE], 1, function(rr) {
          rr_int <- lapply(seq_along(rr), function(ii) {
            md_lim_all_segments[[rv]][[rv_limit_names[ii]]][[rr[ii]]]
          })
          out_int <- rr_int[[1]]
          rr_low <- do.call(c, unname(lapply(rr_int, '[[', "low")))
          rr_low_max <- max(rr_low, na.rm = TRUE)
          out_int$low <- rr_low_max
          rr_low_which_max <- which(rr_low == rr_low_max)
          if (length(rr_low_which_max) > 1) {
            inc_l <- do.call(c, unname(lapply(rr_int, '[[', "inc_l")))[rr_low_which_max]
            out_int$inc_l <- if (any(!inc_l)) FALSE else TRUE
          } else {
            out_int$inc_l <- rr_int[[rr_low_which_max]][["inc_l"]]
          }
          rr_upp <- do.call(c, unname(lapply(rr_int, '[[', "upp")))
          rr_upp_min <- min(rr_upp, na.rm = TRUE)
          out_int$upp <- rr_upp_min
          rr_upp_which_min <- which(rr_upp == rr_upp_min)
          if (length(rr_upp_which_min) > 1) {
            inc_u <- do.call(c, unname(lapply(rr_int, '[[', "inc_u")))[rr_upp_which_min]
            out_int$inc_u <- if (any(!inc_u)) FALSE else TRUE
          } else {
            out_int$inc_u <- rr_int[[rr_upp_which_min]][["inc_u"]]
          }
          return(out_int)
        })

      largest_segment <- plot_segments_intervals[[which.max(plot_segments$n)]]

      bin_breaks <- util_optimize_histogram_bins(
        x = rv_data$values,
        interval_freedman_diaconis = largest_segment,
        cuts = sort(as.numeric(unique(c(
          do.call(c, unname(lapply(plot_segments_intervals, '[[', "low"))),
          do.call(c, unname(lapply(plot_segments_intervals, '[[', "upp"))))))),
        nbins_max = 100)

      if (length(plot_segments_intervals) != length(bin_breaks)) {
        bin_segments_with_data <-
          vapply(bin_breaks, FUN.VALUE = logical(1), function(bb) {
            sum(
              vapply(plot_segments_intervals, FUN.VALUE = logical(1),
                     function(p_int) {
                       bb[1] >= p_int$low & bb[1] <= p_int$upp &
                         bb[length(bb)] >= p_int$low & bb[length(bb)] <= p_int$upp
                     })
            ) == 1
          })
        bin_breaks[!bin_segments_with_data] <- NULL
      }

      # Histogramdaten
      plot_data <- lapply(seq_along(plot_segments_intervals), function(ii) {
        subd1 <- rv_data[redcap_env$`in`(rv_data$values,
                                         plot_segments_intervals[[ii]]),
                         c("values", "limit_violations")]
        h1 <- hist(subd1$values, plot = FALSE, breaks = bin_breaks[[ii]])
        return(data.frame(histogram_x = h1$mids,
                          histogram_y = h1$counts,
                          limit_violations = subd1$limit_violations[1],
                          segment_number = ii,
                          row.names = NULL))
      })
      plot_data <- do.call(rbind, plot_data)
      if (is_datetime_var[rv]) {
        plot_data[["histogram_x"]] <- util_parse_date(plot_data[["histogram_x"]])
      }

      # x-Limits
      if (!is_datetime_var[[rv]]) {
        myxlim <- c(floor(min_plot), ceiling(max_plot))
      } else {
        myxlim <- c(min_plot, max_plot)
        myxlim <- util_parse_date(myxlim)
      }

      # Limit-Linien (Werte & Namen getrennt)
      all_limits_vals  <- c(lower_limits, upper_limits)
      all_limits_names <- c(lower_limit_names, upper_limit_names)
      keep <- !is.na(all_limits_vals)
      if (!all(spec_lines$limits %in% colnames(meta_data))) {
        for (cl in setdiff(spec_lines$limits, colnames(meta_data)))
          meta_data[[cl]] <- rep("", nrow(meta_data))
      }
      all_limits_df <- data.frame(
        limits = all_limits_names[keep],
        values = all_limits_vals[keep]
      )

      all_limits_df$limits <- paste(
        all_limits_df$limits,
        meta_data[meta_data[[label_col]] == rv,
                  all_limits_df$limits, drop = FALSE])
      spec_lines$limits <- paste(
        spec_lines$limits,
        meta_data[meta_data[[label_col]] == rv,
                  spec_lines$limits, drop = FALSE])

      if (is_datetime_var[[rv]]) {
        all_limits_df$values <- as.POSIXct(all_limits_df$values,
                                           origin = min(Sys.time(), 0))
      }

      # Histogram-Plot
      p <- util_create_lean_ggplot(
        ggplot(data = plot_data,
               aes(x = .data[["histogram_x"]],
                   y = .data[["histogram_y"]],
                   fill = .data[["limit_violations"]])),
        plot_data = plot_data
      )

      for (ii in seq_along(plot_segments_intervals)) {
        dt <- subset(plot_data, segment_number == ii)
        bb_ii <- as.numeric(bin_breaks[[ii]])
        width_col <- bb_ii[2] - bb_ii[1]
        p <- p %lean+% util_create_lean_ggplot(
          geom_col(data = dt, width = width_col),
          dt = dt,
          width_col = width_col
        )
      }

      # Punkte (optional)
      bb <- do.call(c, unname(bin_breaks))
      bb <- bb[!duplicated(bb)]
      if (plot_histogram_integer) {
        if (mean(bb[-1] - bb[-length(bb)]) < 7 && length(bb) <= 6) show_obs <- FALSE
      }
      if (show_obs) {
        max_bar_height <- max(table(cut(rv_data$value, breaks = bb)))
        ypos <- -0.05 * max_bar_height
        subdata <- NULL
        for (ii in seq_len(length(bb)-1)) {
          if (ii != length(bb) - 1) {
            subd <- subset(rv_data, values >= bb[ii] & values < bb[ii+1])
          } else {
            subd <- subset(rv_data, values >= bb[ii] & values <= bb[ii+1])
          }
          if (nrow(subd) > 7) {
            sel_val <- quantile(as.numeric(subd$values),
                                probs = seq(0, 1, length.out = 7), type = 3)
            sel_ind <- vapply(sel_val, FUN.VALUE = integer(1),
                              FUN = function(x) which(as.numeric(subd$value) == x)[1])
            subd <- subd[sel_ind, ]
          }
          if (nrow(subd) > 0) subdata <- rbind(subdata, subd)
        }
        colnames(subdata)[which(colnames(subdata) == "values")] <- "histogram_x"
        if (is_datetime_var[[rv]]) {
          subdata$histogram_x <- util_parse_date(subdata$histogram_x)
        }

        p <- p %lean+% util_create_lean_ggplot(
          geom_point(aes(y = ypos),
                     data = subdata,
                     col = "darkgray",
                     position = position_jitter(height = 0.01 * max_bar_height,
                                                width = 0),
                     size = 1,
                     alpha = 0.6),
          ypos = ypos,
          subdata = subdata,
          max_bar_height = max_bar_height
        )
      }

      fli <- util_coord_flip(ref_env = ref_env, xlim = myxlim)
      p <- p %lean+% util_create_lean_ggplot(
        geom_vline(data = all_limits_df,
                   aes(xintercept = .data[["values"]],
                       linetype = .data[["limits"]],
                       color = .data[["limits"]])),
        all_limits_df = all_limits_df
      );
      p <- util_lazy_add_coord(p, fli)
      p <- p %lean+%
        scale_fill_manual(values = spec_bars$fill,
                          breaks = spec_bars$limit_violations,
                          guide = "none") %lean+%
        ggplot2::scale_linetype_manual(name = "limits",
                                       values = spec_lines$lty,
                                       breaks = spec_lines$limits) %lean+%
        scale_color_manual(name = "limits",
                           values = spec_lines$color,
                           breaks = spec_lines$limits) %lean+%
        labs(x = paste0(xlb), y = "") %lean+%
        theme_minimal() %lean+%
        theme(
          title = spec_txt,
          axis.text.x = spec_txt,
          axis.text.y = spec_txt,
          axis.title.x = spec_txt,
          axis.title.y = spec_txt
        )

      if (is_datetime_var[[rv]]) {
        p <- p %lean+% util_create_lean_ggplot(
          ggplot2::scale_x_datetime(expand = expansion(mult = 0.1))
        )
      } else if (plot_histogram_integer && !is_time_var[[rv]]) {
        pretty_x_axt <- util_int_breaks_rounded(c(rv_data$values,
                                                  all_limits_df$values,
                                                  myxlim))
        pretty_x_axt <- pretty_x_axt[
          which(pretty_x_axt >= myxlim[1] - 0.1 * abs(myxlim[2] - myxlim[1]) &
                  pretty_x_axt <= myxlim[2] + 0.1 * abs(myxlim[2] - myxlim[1]))]
        p <- p %lean+% util_create_lean_ggplot(
          scale_x_continuous(breaks = pretty_x_axt,
                             expand = expansion(mult = 0.1)),
          pretty_x_axt = pretty_x_axt
        )
      } else {
        if (is_time_var[[rv]]) {
          p <- p %lean+% util_create_lean_ggplot(
            scale_x_continuous(
              expand = expansion(mult = 0.1),
              labels = function(x) {
                util_as_character(hms::as_hms(x))
              }
            )
          )
        } else {
          p <- p %lean+% util_create_lean_ggplot(
            scale_x_continuous(expand = expansion(mult = 0.1))
          )
        }
      }

      # Sizing
      min_bin_height <- min(plot_data$histogram_y)
      max_bin_height <- max(plot_data$histogram_y)
      no_bars <- nrow(plot_data)
      obj1 <- util_create_lean_ggplot(ggplot2::ggplot_build(p), p = p)
      total_w <- c(util_rbind(data_frames_list = obj1$data)$xmin,
                   util_rbind(data_frames_list = obj1$data)$xmax,
                   util_rbind(data_frames_list = obj1$data)$xintercept)
      total_w <- total_w[!is.na(total_w)]
      min_total_w <- min(total_w); max_total_w <- max(total_w)
      total_w <- max_total_w - min_total_w
      min_x_plot <- min(util_rbind(data_frames_list = obj1$data)$xmin, na.rm = TRUE)
      max_x_plot <- max(util_rbind(data_frames_list = obj1$data)$xmax, na.rm = TRUE)
      no_char_y <- nchar(round(max_bin_height, digits = 0))
      no_char_x <- nchar(round(max_total_w, digits = 0))
      rm(obj1)

      min_limits <- min(all_limits_df$values)
      max_limits <- max(all_limits_df$values)

    } else {
      # Balkendiagramm ---------------------------------------------------------
      all_limits_vals  <- c(lower_limits - 0.5, upper_limits + 0.5)
      all_limits_names <- c(lower_limit_names, upper_limit_names)
      keep <- !is.na(all_limits_vals)
      all_limits_df <- data.frame(
        limits = all_limits_names[keep],
        values = all_limits_vals[keep]
      )

      all_limits_df$limits <- paste(
        all_limits_df$limits,
        meta_data[meta_data[[label_col]] == rv,
                  all_limits_df$limits, drop = FALSE])
      spec_lines$limits <- paste(
        spec_lines$limits,
        meta_data[meta_data[[label_col]] == rv,
                  spec_lines$limits, drop = FALSE])

      if (is_datetime_var[[rv]]) {
        all_limits_df$values <- as.POSIXct(all_limits_df$values,
                                           origin = min(Sys.time(), 0))
      }

      max_plot <- max_plot + 0.5
      min_plot <- min_plot - 0.5

      plot_data <- rv_data %>%
        dplyr::add_count(values) %>%
        dplyr::distinct()

      fli <- util_coord_flip(ref_env = ref_env, xlim = c(min_plot, max_plot))
      pretty_x_axt <- util_int_breaks_rounded(c(plot_data$values, max_plot, min_plot))
      pretty_x_axt <- pretty_x_axt[
        which(pretty_x_axt >= min_plot - 0.1 * abs(max_plot - min_plot) &
                pretty_x_axt <= max_plot + 0.1 * abs(max_plot - min_plot))]

      p <- util_create_lean_ggplot(
        ggplot(plot_data, aes(x = .data[["values"]], y = .data[["n"]],
                              fill = .data[["limit_violations"]])) %lean+%
          geom_col(width = 0.8) %lean+%
          geom_vline(data = all_limits_df,
                     aes(xintercept = .data[["values"]],
                         linetype = .data[["limits"]],
                         color = .data[["limits"]])) %lean+%
          fli %lean+%
          scale_fill_manual(values = spec_bars$fill,
                            breaks = spec_bars$limit_violations,
                            guide = "none") %lean+%
          ggplot2::scale_linetype_manual(name = "limits",
                                         values = spec_lines$lty,
                                         breaks = spec_lines$limits) %lean+%
          scale_color_manual(name = "limits",
                             values = spec_lines$color,
                             breaks = spec_lines$limits) %lean+%
          labs(x = paste0(xlb), y = "") %lean+%
          theme_minimal() %lean+%
          theme(
            title = spec_txt,
            axis.text.x = spec_txt,
            axis.text.y = spec_txt,
            axis.title.x = spec_txt,
            axis.title.y = spec_txt
          ) %lean+%
          scale_x_continuous(breaks = pretty_x_axt,
                             expand = expansion(mult = 0.1)),
        plot_data = plot_data,
        all_limits_df = all_limits_df,
        min_plot = min_plot,
        max_plot = max_plot,
        spec_bars = spec_bars,
        spec_lines = spec_lines,
        xlb = xlb,
        spec_txt = spec_txt,
        fli = fli,
        pretty_x_axt = pretty_x_axt
      )

      min_bin_height <- min(plot_data$n)
      max_bin_height <- max(plot_data$n)
      no_bars <- nrow(plot_data)
      obj1 <- util_create_lean_ggplot(ggplot2::ggplot_build(p), p = p)
      total_w <- c(util_rbind(data_frames_list = obj1$data)$xmin,
                   util_rbind(data_frames_list = obj1$data)$xmax,
                   util_rbind(data_frames_list = obj1$data)$xintercept)
      total_w <- total_w[!is.na(total_w)]
      min_total_w <- min(total_w); max_total_w <- max(total_w)
      total_w <- max_total_w - min_total_w
      min_x_plot <- min(util_rbind(data_frames_list = obj1$data)$xmin, na.rm = TRUE)
      max_x_plot <- max(util_rbind(data_frames_list = obj1$data)$xmax, na.rm = TRUE)
      no_char_y <- nchar(round(max_bin_height, digits = 0))
      no_char_x <- nchar(round(max_total_w, digits = 0))
      rm(obj1)

      min_limits <- min(all_limits_df$values)
      max_limits <- max(all_limits_df$values)
    }

    # Sizing-Hints -------------------------------------------------------------
    range_x <- max_x_plot - min_x_plot
    range_height <- max_bin_height - min_bin_height
    no_bars_in_all_w <- round((total_w * no_bars)/range_x, digits = 0)
    attr(p, "sizing_hints") <- list(figure_type_id = "bar_limit",
                                    range = range_height,
                                    no_bars_in_all_w = no_bars_in_all_w,
                                    no_char_x = no_char_x,
                                    no_char_y = no_char_y)
    return(p)
  }) # end lapply plot_list

  # FlaggedStudyData -----------------------------------------------------------
  if (return_flagged_study_data) {
    fsd <- do.call(cbind.data.frame, sd_lim)
    colnames(fsd) <- paste(
      unlist(lapply(names(sd_lim), function(rv) {
        rep(rv, length(sd_lim[[rv]]))
      })),
      unlist(lapply(sd_lim, names)),
      sep = "_")
    fsd <- cbind(ds1, fsd)
  } else {
    fsd <- NULL
  }

  # SummaryData ----------------------------------------------------------------
  sumdat <- do.call(rbind.data.frame,
                    lapply(setNames(nm = resp_vars_iteml), function(rv) {
                      divisor <- rep(unname(unlist(sd_n_obs[[rv]])), each = 4)
                      data.frame(
                        "Variables" = rv,
                        "Section" = rep(colnames(freq_lim_viol[[rv]]),
                                        nrow(freq_lim_viol[[rv]])),
                        "Limits" = rep(rownames(freq_lim_viol[[rv]]),
                                       each = 4),
                        "Number" = c(t(freq_lim_viol[[rv]])),
                        "Percentage" = ifelse(divisor == 0, 0,
                                              round(c(t(freq_lim_viol[[rv]])) /
                                                      divisor * 100, 2)),
                        check.names = FALSE, fix.empty.names = FALSE)
                    }))
  rownames(sumdat) <- NULL

  # Report-Summary-Table -------------------------------------------------------
  report_all_limits <- unique(sumdat$Limits)
  heatmap_tab <- do.call(
    rbind.data.frame,
    lapply(setNames(nm = resp_vars_iteml), function(rv) {
      n_viol <- rowSums(freq_lim_viol[[rv]][, c("below", "above"), drop = FALSE])
      divisor <- unlist(sd_n_obs[[rv]])[rownames(freq_lim_viol[[rv]])]
      n_viol_rel <- ifelse(divisor == 0, 0, n_viol / divisor)
      out_viol <- setNames(rep(NA, length(report_all_limits)), report_all_limits)
      out_viol[names(n_viol_rel)] <- n_viol_rel
      out_viol <- as.data.frame(t(out_viol))
      out_viol$Variables <- rv
      out_viol$N <- 1
      return(out_viol[, c("Variables", "N",
                          setdiff(colnames(out_viol), c("Variables", "N")))])
    }))
  rownames(heatmap_tab) <- NULL
  colnames(heatmap_tab)[3:ncol(heatmap_tab)] <-
    paste(colnames(heatmap_tab)[3:ncol(heatmap_tab)], "violations")

  heatmap_tab <- util_validate_report_summary_table(heatmap_tab,
                                                    meta_data = meta_data,
                                                    label_col = label_col)

  # SummaryTable ---------------------------------------------------------------
  sumtab_inadm_num <- vapply(
    resp_vars_iteml, FUN.VALUE = numeric(2), FUN = function(rv) {
      if (is_datetime_var[[rv]]) {
        res <- c(NA, nrow(ds1))
      } else {
        sumdat_sub <- sumdat %>%
          dplyr::filter(Variables == rv & Limits == "HARD_LIMITS")
        if (nrow(sumdat_sub) == 0) {
          res <- c(0, nrow(ds1))
        } else {
          res <- sumdat_sub %>%
            dplyr::select(Section, Number) %>%
            dplyr::mutate(n = sum(Number)) %>%
            dplyr::filter(!Section %in% c("within", "samplesize")) %>%
            dplyr::mutate(n_viol = sum(Number)) %>%
            dplyr::distinct(n_viol, n)
          res <- unlist(res)
        }
      }
      names(res) <- c("n_viol", "n")
      return(res)
    })
  sumtab_inadm_datetime <- vapply(
    resp_vars_iteml, FUN.VALUE = numeric(2), FUN = function(rv) {
      if (!is_datetime_var[[rv]]) {
        res <- c(NA, nrow(ds1))
      } else {
        sumdat_sub <- sumdat %>%
          dplyr::filter(Variables == rv & Limits == "HARD_LIMITS")
        if (nrow(sumdat_sub) == 0) {
          res <- c(0, nrow(ds1))
        } else {
          res <- sumdat_sub %>%
            dplyr::select(Section, Number) %>%
            dplyr::mutate(n = sum(Number)) %>%
            dplyr::filter(!Section %in% c("within", "samplesize")) %>%
            dplyr::mutate(n_viol = sum(Number)) %>%
            dplyr::distinct(n_viol, n)
          res <- unlist(res)
        }
      }
      names(res) <- c("n_viol", "n")
      return(res)
    })

  sumtab_unc_num <- vapply(
    resp_vars_iteml, FUN.VALUE = numeric(2), FUN = function(rv) {
      if (is_datetime_var[[rv]]) {
        res <- c(NA, nrow(ds1))
      } else {
        rv_sd_int <- sd_lim_HL_check[[rv]]
        rv_sd_int <- rv_sd_int[which(names(rv_sd_int) != "HARD_LIMITS")]
        if (length(rv_sd_int) == 0) {
          res <- c(0, nrow(ds1))
        } else {
          rv_sd_int <- lapply(rv_sd_int, function(ll) {
            levels(ll) <- c(1, 0, 1)
            return(as.numeric(as.character(ll)))
          })
          rv_sd_int <- as.data.frame(rv_sd_int)
          rv_sd_int <- rv_sd_int[complete.cases(rv_sd_int), , drop = FALSE]
          rv_sd_int$N_lim <- rowSums(rv_sd_int)
          res <- c(length(which(rv_sd_int$N_lim > 0)),
                   nrow(rv_sd_int))
        }
      }
      names(res) <- c("n_viol", "n")
      return(res)
    })
  sumtab_unc_datetime <- vapply(
    resp_vars_iteml, FUN.VALUE = numeric(2), FUN = function(rv) {
      if (!is_datetime_var[[rv]]) {
        res <- c(NA, nrow(ds1))
      } else {
        rv_sd_int <- sd_lim_HL_check[[rv]]
        rv_sd_int <- rv_sd_int[which(names(rv_sd_int) != "HARD_LIMITS")]
        if (length(rv_sd_int) == 0) {
          res <- c(0, nrow(ds1))
        } else {
          rv_sd_int <- lapply(rv_sd_int, function(ll) {
            levels(ll) <- c(1, 0, 1)
            return(as.numeric(as.character(ll)))
          })
          rv_sd_int <- as.data.frame(rv_sd_int)
          rv_sd_int <- rv_sd_int[complete.cases(rv_sd_int), , drop = FALSE]
          rv_sd_int$N_lim <- rowSums(rv_sd_int)
          res <- c(length(which(rv_sd_int$N_lim > 0)),
                   nrow(rv_sd_int))
        }
      }
      names(res) <- c("n_viol", "n")
      return(res)
    })

  sumtab <- data.frame(
    "Variables" = resp_vars_iteml,
    "NUM_con_rvv_unum" = sumtab_unc_num["n_viol", ],
    "PCT_con_rvv_unum" = round(sumtab_unc_num["n_viol", ] /
                                 sumtab_unc_num["n", ] * 100, 2),
    "FLG_con_rvv_unum" = ifelse(sumtab_unc_num["n_viol", ] > 0, TRUE, FALSE),
    "NUM_con_rvv_utdat" = sumtab_unc_datetime["n_viol", ],
    "PCT_con_rvv_utdat" = round(sumtab_unc_datetime["n_viol", ] /
                                  sumtab_unc_datetime["n", ] * 100, 2),
    "FLG_con_rvv_utdat" = ifelse(sumtab_unc_datetime["n_viol", ] > 0,
                                 TRUE, FALSE),
    "NUM_con_rvv_inum" = sumtab_inadm_num["n_viol", ],
    "PCT_con_rvv_inum" = round(sumtab_inadm_num["n_viol", ] /
                                 sumtab_inadm_num["n", ] * 100, 2),
    "FLG_con_rvv_inum" = ifelse(sumtab_inadm_num["n_viol", ] > 0, TRUE, FALSE),
    "NUM_con_rvv_itdat" = sumtab_inadm_datetime["n_viol", ],
    "PCT_con_rvv_itdat" = round(sumtab_inadm_datetime["n_viol", ] /
                                  sumtab_inadm_datetime["n", ] * 100, 2),
    "FLG_con_rvv_itdat" = ifelse(sumtab_inadm_datetime["n_viol", ] > 0,
                                 TRUE, FALSE))
  sumtab$PCT_con_rvv_unum[which(sumtab_unc_num["n", ] == 0)] <- 0
  sumtab$PCT_con_rvv_utdat[which(sumtab_unc_datetime["n", ] == 0)] <- 0
  sumtab$PCT_con_rvv_inum[which(sumtab_inadm_num["n", ] == 0)] <- 0
  sumtab$PCT_con_rvv_itdat[which(sumtab_inadm_datetime["n", ] == 0)] <- 0

  rownames(sumtab) <- NULL

  # SummaryData breit ----------------------------------------------------------
  # sumtab <- util_map_to_other_metrics(sumtab,
  #                                     meta_data,
  #                                     label_col)

  # Modify the SummaryData from long to wide format to be more concise to read
  sumdat_wide <- stats::reshape(sumdat,
                                idvar = c("Variables", "Limits"),
                                timevar = "Section",
                                direction = "wide")
  sumdat_wide$"All.outside.limitsN" <- sumdat_wide$Number.below + sumdat_wide$Number.above
  sumdat_wide$"All.outside.limits%" <- round(sumdat_wide$"All.outside.limitsN"/
                                               sumdat_wide$Number.samplesize * 100,
                                             digits = 2)
  sumdat_wide$`Below.limits-N (%)`  <- paste0(sumdat_wide$Number.below,  " (", sumdat_wide$Percentage.below,")")
  sumdat_wide$`Within.limits-N (%)` <- paste0(sumdat_wide$Number.within, " (", sumdat_wide$Percentage.within,")")
  sumdat_wide$`Above.limits-N (%)`  <- paste0(sumdat_wide$Number.above,  " (", sumdat_wide$Percentage.above,")")
  sumdat_wide$`All outside limits N (%)` <- paste0(sumdat_wide$All.outside.limitsN, " (",
                                                   sumdat_wide$`All.outside.limits%`,")")
  sumdat_wide <- sumdat_wide[, c("Variables", "Limits", "Below.limits-N (%)",
                                 "Within.limits-N (%)", "Above.limits-N (%)",
                                 "All outside limits N (%)")]
  colnames(sumdat_wide) <- c("Variables", "Limits", "Below limits N (%)",
                             "Within limits N (%)", "Above limits N (%)",
                             "All outside limits N (%)")

  text_to_display <- util_get_hovertext("[con_limit_dev_hover]")
  attr(sumdat_wide, "description") <- text_to_display

  if (length(plot_list) == 1) {
    .plot1 <- plot_list[[resp_vars]]
    obj1 <- util_create_lean_ggplot(ggplot2::ggplot_build(.plot1), .plot1 = .plot1)
    min_bin_height <- min(util_rbind(data_frames_list = obj1$data)$ymax, na.rm = TRUE)
    max_bin_height <- max(util_rbind(data_frames_list = obj1$data)$ymax, na.rm = TRUE)
    no_bars <- sum(!is.na(util_rbind(data_frames_list = obj1$data)$ymax))
    total_w <- c(util_rbind(data_frames_list = obj1$data)$xmin,
                 util_rbind(data_frames_list = obj1$data)$xmax,
                 util_rbind(data_frames_list = obj1$data)$xintercept)
    total_w <- total_w[!is.na(total_w)]
    min_total_w <- min(total_w); max_total_w <- max(total_w)
    total_w <- max_total_w - min_total_w
    min_x_plot <- min(util_rbind(data_frames_list = obj1$data)$xmin, na.rm = TRUE)
    max_x_plot <- max(util_rbind(data_frames_list = obj1$data)$xmax, na.rm = TRUE)
    range_x <- max_x_plot - min_x_plot
    no_bars_in_all_w <- round((total_w * no_bars)/range_x, digits = 0)
    range_height <- max_bin_height - min_bin_height

    if (is.infinite(min_bin_height) || is.infinite(max_bin_height)) {
      range_height <- 100
      if (is.infinite(min_bin_height)) min_bin_height <- 0
      if (is.infinite(max_bin_height)) max_bin_height <- 0
    }

    no_char_y <- nchar(round(max_bin_height, digits = 0))
    no_char_x <- nchar(round(total_w, digits = 0))
    rm(obj1)
  }

  }
  ####################################################
  #Complex limits
  # Variable with complex limits (if they also have item_level metadata limits,
  # they will be ignored) -----

  if (!is.null(vars_in_cil)) {
    # Which variables are of type 'datetime'?
    is_datetime_var_cil <- vapply(vars_in_cil, function(rv) {
      meta_data[["DATA_TYPE"]][meta_data[[label_col]] == rv] ==
        DATA_TYPES$DATETIME
    }, FUN.VALUE = logical(1))
    #create a list of variables in cross-item levels and their limits
    # with total numbers and values within
    cil_limits <- setNames(
      nm = vars_in_cil,
      lapply(vars_in_cil, function(rvs) {
        # Filter r_complex for the current variable (rvs)
        r_complex_var <- r_complex[r_complex$Variables == rvs, ]
        # Extract the limits for the current variable.
        lims <- r_complex_var$limits
        col_int <- lapply(lims, function(ll) {
          #Filter r_complex_var for the current limit (int)
          r_complex_limit <- r_complex_var[r_complex_var$limits == ll, ]
          # Find the column name that starts with "NUM_" and is NOT NA.
          num_cols <- grep("^NUM_", names(r_complex_limit), value = TRUE)
          # Select the one column in num_cols that has a non-NA value
          col_to_use <- num_cols[!is.na(r_complex_limit[, num_cols])]
          # Check if a column was found and, if so, extract the value (all_outliers)
          if (length(col_to_use) == 1) {
            all_outliers <- r_complex_limit[[col_to_use]]
          } else {
            all_outliers <- NA
          }
          n_tot <- length(ds1[, rvs]) # sample size from ds1
          within <- n_tot - all_outliers
          # Return the data frame for this limit
          return(data.frame(N = n_tot,
                            all_outliers = all_outliers,
                            within = within))
        })
        # Name the elements of the list using the limits
        names(col_int) <- lims
        return(col_int)
      })
    )
    # Flagged data and plot_list not possible
    # if limits only in cross-item level metadata -----
    if (!exists("fsd")) {
      fsd <- NULL
    }


    #sumdat from cross-item level-------
    sumdat_complex <-
      do.call(rbind.data.frame,
              lapply(setNames(nm = vars_in_cil,
                              object = vars_in_cil),
                     function(rv) {
                       # Filter r_complex for the current variable
                       r_complex_var <- r_complex[r_complex$Variables == rv, ]
                       # Extract the limits for the current variable
                       lims <- r_complex_var$limits
                       # rbind results for the current variable
                       do.call(rbind.data.frame,
                               lapply(lims,
                                      function(ll) {
                                        rv_data <- cil_limits[[rv]][[ll]]
                                        divisor <- rv_data$N
                                        data.frame(
                                          "Variables" = rv,
                                          "Section" = c("within", "all_outliers"),
                                          "Limits" = rep(ll, 2),
                                          "Number" = c(rv_data$within,
                                                       rv_data$all_outliers),
                                          "Percentage" = c(ifelse(divisor == 0, 0,
                                                                  round(rv_data$within/

                                                                          divisor * 100,
                                                                        2)),
                                                           ifelse(divisor == 0, 0,
                                                                  round(rv_data$all_outliers/
                                                                          divisor * 100,
                                                                        2))),
                                          check.names = FALSE, fix.empty.names = FALSE,
                                          stringsAsFactors = FALSE
                                        )
                                      }))
                     }))
    rownames(sumdat_complex) <- NULL

    # report summary table -----
    report_all_limits <- unique(sumdat_complex$Limits)
    dt_rcomplex <- util_map_labels(
      r_complex[["Variables"]],
      meta_data = meta_data,
      to = DATA_TYPE,
      from = label_col
    )
    # heatmap -------
    heatmap_tab_complex <- do.call(
      rbind.data.frame,
      lapply(setNames(nm = vars_in_cil),
             function(rv) {
               cur_row <- r_complex[r_complex$Variables==rv &
                                      r_complex$limits=="HARD_LIMITS", ]
               if (nrow(cur_row) > 0){
                 if (dt_rcomplex[cur_row[cur_row$Variables==rv,
                                         "Variables"]] != "datetime") {
                   n_viol <- cur_row[cur_row$Variables==rv,
                                     "NUM_con_rvv_inum"]
                 } else {
                   n_viol <- cur_row[cur_row$Variables==rv,
                                     "NUM_con_rvv_itdat"]
                 }

                 divisor <- cur_row[cur_row$Variables==rv, "N"]
                 n_viol_rel <- ifelse(divisor == 0, 0, n_viol / divisor)
                 names(n_viol_rel) <- "HARD_LIMITS"
               }
               cur_row <- r_complex[r_complex$Variables==rv &
                                      r_complex$limits=="SOFT_LIMITS", ]
               if (nrow(cur_row) > 0){
                 if (dt_rcomplex[cur_row[cur_row$Variables==rv,
                                         "Variables"]] != "datetime") {
                   n_viol <- cur_row[cur_row$Variables==rv,
                                     "NUM_con_rvv_unum"]
                 } else {
                   n_viol <- cur_row[cur_row$Variables==rv,
                                     "NUM_con_rvv_utdat"]
                 }
                 divisor <- cur_row[cur_row$Variables==rv, "N"]
                 if(exists("n_viol_rel")){
                   n_viol_rel2 <- ifelse(divisor == 0, 0, n_viol / divisor)
                   names(n_viol_rel2) <- "SOFT_LIMITS"
                   n_viol_rel <- c(n_viol_rel, n_viol_rel2)
                 } else {
                   n_viol_rel <- ifelse(divisor == 0, 0, n_viol / divisor)
                   names(n_viol_rel) <- "SOFT_LIMITS"
                 }
               }
               cur_row <- r_complex[r_complex$Variables==rv &
                                      r_complex$limits=="DETECTION_LIMITS", ]
               if (nrow(cur_row) > 0){
                 if (dt_rcomplex[cur_row[cur_row$Variables==rv,
                                         "Variables"]] != "datetime") {
                   n_viol <- cur_row[cur_row$Variables==rv,
                                     "NUM_con_rvv_unum"]
                 } else {
                   n_viol <- cur_row[cur_row$Variables==rv,
                                     "NUM_con_rvv_utdat"]
                 }
                 divisor <- cur_row[cur_row$Variables==rv, "N"]
                 if(exists("n_viol_rel")){
                   n_viol_rel2 <- ifelse(divisor == 0, 0, n_viol / divisor)
                   names(n_viol_rel2) <- "SOFT_LIMITS"
                   n_viol_rel <- c(n_viol_rel, n_viol_rel2)
                 } else {
                   n_viol_rel <- ifelse(divisor == 0, 0, n_viol / divisor)
                   names(n_viol_rel) <- "SOFT_LIMITS"
                 }
               }
               out_viol <-
                 setNames(rep(NA, length(report_all_limits)), report_all_limits)
               out_viol[names(n_viol_rel)] <- n_viol_rel
               out_viol <- as.data.frame(t(out_viol))
               out_viol$Variables <- rv
               out_viol$N <- 1
               return(out_viol[, c("Variables", "N",
                                   setdiff(colnames(out_viol),
                                           c("Variables", "N")))])
             }))

    rownames(heatmap_tab_complex) <- NULL
    colnames(heatmap_tab_complex)[3:ncol(heatmap_tab_complex)] <-
      paste(colnames(heatmap_tab_complex)[3:ncol(heatmap_tab_complex)], "violations")

    heatmap_tab_complex <- util_validate_report_summary_table(heatmap_tab_complex,
                                                              meta_data = meta_data,
                                                              label_col = label_col)

    #If sumdat does not exists, because limits are only present in the
    # cross item level create an empty one
    if (!exists("sumdat")) {
      sumdat <- data.frame(
        Variables = character(0),
        Section = character(0),
        Limits = character(0),
        Number = numeric(0),
        Percentage = numeric(0),
        stringsAsFactors = FALSE
      )
    }

    #Add datatype to r_complex
    r_complex$dt <- dt_rcomplex[match(r_complex$Variables, names(dt_rcomplex))] #TODO: is this needed?


    # summary table -----
    sumtab_inadm_num_complex <- vapply(
      vars_in_cil, FUN.VALUE = numeric(2), FUN = function(rv) {
        if (is_datetime_var_cil[[rv]]) {
          res <- c(NA, nrow(ds1))
        } else {
          sumdat_sub <- sumdat_complex[sumdat_complex$Variables == rv &
                                         sumdat_complex$Limits == "HARD_LIMITS" &
                                         sumdat_complex$Section == "all_outliers", ]
          if(length(sumdat_sub[, "Number"])>0) {
            n_viol <- sumdat_sub[, "Number"]
          } else {
            n_viol <- NA
          }
          n_tot <- nrow(ds1)
          res <- c(n_viol, n_tot)
        }
        names(res) <- c("n_viol", "n")
        return(res)
      })
    sumtab_inadm_datetime_complex <- vapply(
      vars_in_cil, FUN.VALUE = numeric(2), FUN = function(rv) {
        if (!is_datetime_var_cil[[rv]]) {
          res <- c(NA, nrow(ds1))
        } else {
          sumdat_sub <- sumdat_complex[sumdat_complex$Variables == rv &
                                         sumdat_complex$Limits == "HARD_LIMITS" &
                                         sumdat_complex$Section == "all_outliers", ]
          if(length(sumdat_sub[, "Number"])>0) {
            n_viol <- sumdat_sub[, "Number"]
          } else {
            n_viol <- NA
          }
          n_tot <- nrow(ds1)
          res <- c(n_viol, n_tot)
        }
        names(res) <- c("n_viol", "n")
        return(res)
      })
    sumtab_unc_num_soft <- vapply(
      vars_in_cil, FUN.VALUE = numeric(2), FUN = function(rv) {
        if (is_datetime_var_cil[[rv]]) {
          res <- c(NA, nrow(ds1))
        } else {
          sumdat_sub <- sumdat_complex[sumdat_complex$Variables == rv &
                                         sumdat_complex$Limits == "SOFT_LIMITS" &
                                         sumdat_complex$Section == "all_outliers", ]
          if(length(sumdat_sub[, "Number"])>0) {
            n_viol <- sumdat_sub[, "Number"]

          } else {
            n_viol <- NA
          }
          n_tot <- nrow(ds1)
          res <- c(n_viol, n_tot)
        }
        names(res) <- c("n_viol", "n")
        return(res)
      })
    sumtab_unc_num_detect <- vapply(
      vars_in_cil, FUN.VALUE = numeric(2), FUN = function(rv) {
        if (is_datetime_var_cil[[rv]]) {
          res <- c(NA, nrow(ds1))
        } else {
          sumdat_sub <- sumdat_complex[sumdat_complex$Variables == rv &
                                         sumdat_complex$Limits == "DETECTION_LIMITS" &
                                         sumdat_complex$Section == "all_outliers", ]
          if(length(sumdat_sub[, "Number"])>0) {
            n_viol <- sumdat_sub[, "Number"]
          } else {
            n_viol <- NA
          }
          n_tot <- nrow(ds1)
          res <- c(n_viol, n_tot)
        }
        names(res) <- c("n_viol", "n")
        return(res)
      })

    sumtab_unc_num_complex <- vapply(
      vars_in_cil, FUN.VALUE = numeric(2), FUN = function(rv) {
        if (is_datetime_var_cil[[rv]]) {
          res <- c(NA, nrow(ds1))
        } else {
          if(is.na(sumtab_unc_num_soft[[1]]) &&
             is.na(sumtab_unc_num_detect[[1]])){
            n_viol <- NA
          } else {
            n_viol <- max(sumtab_unc_num_soft[[1]],
                          sumtab_unc_num_detect[[1]],
                          na.rm = TRUE)
          }
          n_tot <- nrow(ds1)
          res <- c(n_viol, n_tot)
        }
        names(res) <- c("n_viol", "n")
        return(res)
      })
    sumtab_unc_datetime_soft <- vapply(
      vars_in_cil, FUN.VALUE = numeric(2), FUN = function(rv) {
        if (!is_datetime_var_cil[[rv]]) {
          res <- c(NA, nrow(ds1))
        } else {
          sumdat_sub <- sumdat_complex[sumdat_complex$Variables == rv &
                                         sumdat_complex$Limits == "SOFT_LIMITS" &
                                         sumdat_complex$Section == "all_outliers", ]
          if(length(sumdat_sub[, "Number"])>0) {
            n_viol <- sumdat_sub[, "Number"]
          } else {
            n_viol <- NA
          }
          n_tot <- nrow(ds1)
          res <- c(n_viol, n_tot)
        }
        names(res) <- c("n_viol", "n")
        return(res)
      })

    sumtab_unc_datetime_detect <- vapply(
      vars_in_cil, FUN.VALUE = numeric(2), FUN = function(rv) {
        if (!is_datetime_var_cil[[rv]]) {
          res <- c(NA, nrow(ds1))
        } else {
          sumdat_sub <- sumdat_complex[sumdat_complex$Variables == rv &
                                         sumdat_complex$Limits == "DETECTION_LIMITS" &
                                         sumdat_complex$Section == "all_outliers", ]
          if(length(sumdat_sub[, "Number"])>0) {
            n_viol <- sumdat_sub[, "Number"]
          } else {
            n_viol <- NA
          }
          n_tot <- nrow(ds1)
          res <- c(n_viol, n_tot)
        }
        names(res) <- c("n_viol", "n")
        return(res)
      })
    sumtab_unc_datetime_complex <- vapply(
      vars_in_cil, FUN.VALUE = numeric(2), FUN = function(rv) {
        if (!is_datetime_var_cil[[rv]]) {
          res <- c(NA, nrow(ds1))
        } else {
          if(is.na(sumtab_unc_datetime_soft[[1]]) &&
             is.na(sumtab_unc_datetime_detect[[1]])){
            n_viol <- NA
          } else {
            n_viol <- max(sumtab_unc_datetime_soft[[1]],
                          sumtab_unc_datetime_detect[[1]],
                          na.rm = TRUE)
          }
          n_tot <- nrow(ds1)
          res <- c(n_viol, n_tot)
        }
        names(res) <- c("n_viol", "n")
        return(res)
      })

    #Create sumtab -----
    sumtab_complex <- data.frame(
      "Variables" = vars_in_cil,
      "NUM_con_rvv_unum" = sumtab_unc_num_complex["n_viol", ],
      "PCT_con_rvv_unum" = round(sumtab_unc_num_complex["n_viol", ] /
                                   sumtab_unc_num_complex["n", ] * 100, 2),
      "FLG_con_rvv_unum" = ifelse(sumtab_unc_num_complex["n_viol", ] > 0, TRUE, FALSE),
      "NUM_con_rvv_utdat" = sumtab_unc_datetime_complex["n_viol", ],
      "PCT_con_rvv_utdat" = round(sumtab_unc_datetime_complex["n_viol", ] /
                                    sumtab_unc_datetime_complex["n", ] * 100, 2),
      "FLG_con_rvv_utdat" = ifelse(sumtab_unc_datetime_complex["n_viol", ] > 0,
                                   TRUE, FALSE),
      "NUM_con_rvv_inum" = sumtab_inadm_num_complex["n_viol", ],
      "PCT_con_rvv_inum" = round(sumtab_inadm_num_complex["n_viol", ] /
                                   sumtab_inadm_num_complex["n", ] * 100, 2),
      "FLG_con_rvv_inum" = ifelse(sumtab_inadm_num_complex["n_viol", ] > 0, TRUE, FALSE),
      "NUM_con_rvv_itdat" = sumtab_inadm_datetime_complex["n_viol", ],
      "PCT_con_rvv_itdat" = round(sumtab_inadm_datetime_complex["n_viol", ] /
                                    sumtab_inadm_datetime_complex["n", ] * 100, 2),
      "FLG_con_rvv_itdat" = ifelse(sumtab_inadm_datetime_complex["n_viol", ] > 0,
                                   TRUE, FALSE),
      check.names = FALSE, fix.empty.names = FALSE)
    # for n = 0, entries in PCT columns will be NaN
    # we can replace them by 0 (but we will not replace NA by 0, because they
    # show that an indicator did not apply)
    sumtab_complex$PCT_con_rvv_unum[which(sumtab_unc_num_complex["n", ] == 0)] <- 0
    sumtab_complex$PCT_con_rvv_utdat[which(sumtab_unc_datetime_complex["n", ] == 0)] <- 0
    sumtab_complex$PCT_con_rvv_inum[which(sumtab_inadm_num_complex["n", ] == 0)] <- 0
    sumtab_complex$PCT_con_rvv_itdat[which(sumtab_inadm_datetime_complex["n", ] == 0)] <- 0
    rownames(sumtab_complex) <- NULL

    #Create sumdat_wide -----
    #Modify the SummaryData from long to wide format to be more concise to read
    sumdat_wide_complex <- stats::reshape(sumdat_complex,
                                          idvar = c("Variables", "Limits"),
                                          timevar = "Section",
                                          direction = "wide")

    sumdat_wide_complex$`Below limits N (%)` <-
      paste0(sumdat_wide_complex$Number.below, " (",
             sumdat_wide_complex$Percentage.below,")")
    sumdat_wide_complex$`Within limits N (%)` <-
      paste0(sumdat_wide_complex$Number.within, " (",
             sumdat_wide_complex$Percentage.within,")")
    sumdat_wide_complex$`Above limits N (%)` <-
      paste0(sumdat_wide_complex$Number.above, " (",
             sumdat_wide_complex$Percentage.above,")")
    sumdat_wide_complex$`All outside limits N (%)` <-
      paste0(sumdat_wide_complex$Number.all_outliers, " (",
             sumdat_wide_complex$Percentage.all_outliers,")")
    sumdat_wide_complex <-
      sumdat_wide_complex[, c("Variables", "Limits", "Below limits N (%)",
                              "Within limits N (%)", "Above limits N (%)",
                              "All outside limits N (%)"
      )]
    text_to_display <- util_get_hovertext("[con_limit_dev_hover]")
    attr(sumdat_wide_complex, "description") <- text_to_display

    # text design
    spec_txt <- element_text(
      colour = "black",
      hjust = .5, vjust = .5, face = "plain"
    )

    #Plot complex----
    # iteration to create a plot for each 'resp_var'-----
    plot_list_complex <- lapply(setNames(nm = vars_in_cil), function(rv) {
      r_complex_rv <- r_complex[r_complex$Variables == rv, ]
      # limits for variable `rv`
      rv_limit_names <- unique(r_complex_rv$limits)


      # range of percentage
      max_data <- 100
      min_data <- 0
      # range of values to be included in the plot
      max_plot <- 100
      min_plot <- 0

      # Should the plot be a histogram? If not, it will be a bar chart.
      plot_histogram <-FALSE
      plot_histogram_integer <-FALSE

      xlb <- prep_get_labels(resp_vars = rv,
                             resp_vars_match_label_col_only = TRUE,
                             label_class = "SHORT",
                             label_col = label_col,
                             item_level = meta_data)

      # bar chart --------------------------------------------------------------
      # compute bar heights, prepare data for plotting
      plot_data <- lapply(setNames(nm = r_complex_rv$limits), function(lim) {
        r_complex_rv_lim <- r_complex_rv[r_complex_rv$limits == lim, ]
        PCT_content <-  r_complex_rv_lim[, grep("^PCT_", names(r_complex_rv_lim))]
        PCT_content <- PCT_content[, sapply(PCT_content, function(x) !all(is.na(x)))]
        if (length(PCT_content) != 1) {
          util_error(m =
                       "There should be only one value for the number of violations per limit")
        }
        PCT_content
      })

      plot_data <- data.frame(plot_data)
      plot_data <- utils::stack(plot_data)


      # Empty plot structure that uses your custom variables for consistency
      p <- util_create_lean_ggplot(
        ggplot() %lean+%
          geom_text(
            aes(x = 0.5, y = 0.5, label =
                  "No plot available for complex limits"),
            size = 5,
            color = "gray50",
            check_overlap = TRUE
          ) %lean+%
          ggplot2::xlim(0, 1) %lean+%
          ggplot2::ylim(0, 1) %lean+%
          ggplot2::labs(x = paste0(xlb), y = "") %lean+%
          ggplot2::theme_minimal() %lean+%
          ggplot2::theme(
            title = spec_txt,
            axis.text.x = element_blank(),
            axis.text.y = element_blank(),
            axis.title.x = spec_txt,
            axis.title.y = element_blank(),
            panel.grid.major = element_blank(),
            panel.grid.minor = element_blank()
          ),
        xlb = xlb,
        spec_txt = spec_txt
      )

      min_bin_height <- min(plot_data$values)
      max_bin_height  <- max(plot_data$values)
      if(max_bin_height == min_bin_height) {range_height <-max_bin_height } else{
        range_height <- max_bin_height - min_bin_height
      }

      no_bars_in_all_w <- nrow(plot_data)
      no_char_x <- 4
      no_char_y <- max(nchar(as.vector(plot_data$ind)))
      # Figure size hint for plot
      attr(p, "sizing_hints") <- list(figure_type_id = "bar_limit",
                                      range = range_height,
                                      no_bars_in_all_w = no_bars_in_all_w,
                                      no_char_x = no_char_x,
                                      no_char_y = no_char_y)
      return(p)
    })


  }

  #unify results if both itel_level limits and complex limits are present-----
  if(!exists("sumtab")) {sumtab <- NULL}
  if(!exists("sumtab_complex")) {sumtab_complex <- NULL}
  if(!exists("sumdat_wide")) {sumdat_wide <- NULL}
  if(!exists("sumdat_wide_complex")) {sumdat_wide_complex <- NULL}
  if(!exists("heatmap_tab")) {heatmap_tab <- NULL}
  if(!exists("heatmap_tab_complex")) {heatmap_tab_complex <- NULL}
  if (!exists("plot_list")) {
    plot_list <- NULL
  }
  if (!exists("plot_list_complex")) {
    plot_list_complex <- NULL
  }
  sumtab_final <- util_rbind(sumtab,sumtab_complex)
  sumdat_wide_final <- util_rbind(sumdat_wide,sumdat_wide_complex)
  sumdat_wide_final$`Below limits N (%)` <-
    gsub("\\(\\)", "0 (0)",
         sumdat_wide_final$`Below limits N (%)`)
  sumdat_wide_final$`Within limits N (%)` <-
    gsub("\\(\\)", "0 (0)",
         sumdat_wide_final$`Within limits N (%)`)
  sumdat_wide_final$`Above limits N (%)` <-
    gsub("\\(\\)", "0 (0)",
         sumdat_wide_final$`Above limits N (%)`)
  sumdat_wide_final$`All outside limits N (%)` <-
    gsub("\\(\\)", "0 (0)",
         sumdat_wide_final$`All outside limits N (%)`)

  heatmap_tab_final <- util_rbind(heatmap_tab,heatmap_tab_complex)
  plot_list_final <- c(plot_list, plot_list_complex)


  # final return statement -----------------------------------------------------
  # return(util_mark_result_layout(dqr = util_attach_attr(list( to. hard-code the layout, but it should be default, then it comes from DQ_OBS
  #   FlaggedStudyData = fsd,
  #   SummaryTable = sumtab,
  #   SummaryData = sumdat_wide,
  #   ReportSummaryTable = heatmap_tab,
  #   SummaryPlotList = plot_list),
  #   as_plotly = "util_as_plotly_con_limit_deviations"), "2-columns-fig-left")
  # )
  return(dqr = util_attach_attr(list(
    FlaggedStudyData = fsd,
    SummaryTable = sumtab_final,
    SummaryData = sumdat_wide_final,
    ReportSummaryTable = heatmap_tab_final,
    SummaryPlotList = plot_list_final),
    as_plotly = "util_as_plotly_con_limit_deviations")
  )
}



#' @family plotly_shims
#' @concept plotly_shims
#' @noRd
util_as_plotly_con_limit_deviations <- function(res, ...) {
  p <- res$SummaryPlotList
  util_stop_if_not(length(p) == 1)
  p <- p[[1]]

  gb <- ggplot2::ggplot_build(p)

  scale2 <- tryCatch(
    {
      # preferred: exactly what you had before, just factored out
      gb$plot$scales$scales[[2]]
    },
    error = function(e) {
      # fallback for exotic ggplot2 versions
      sc_obj <- util_gg_get(gb$plot$scales, "scales")
      if (is.list(sc_obj) && length(sc_obj) >= 2) {
        sc_obj[[2]]
      } else {
        stop(e)
      }
    }
  )

  lb <- scale2$get_labels(1)
  cl <- scale2$palette(1)
  br <- scale2$breaks

  py <- util_ggplotly(p, ...)
  for (i in seq_along(py$x$data)) {
    nm <- py$x$data[[i]]$name
    if (nm %in% paste0("(", br, ",1)")) {
      py$x$data[[i]]$name <- br[nm == paste0("(", br, ",1)")]
    } else {
      py$x$data[[i]]$showlegend <- FALSE
    }
  }
  py <- plotly::layout(py, legend = list(traceorder = "reversed"))
  for (l_i in which(vapply(lapply(py$x$data, `[[`, "type"), function(e1, e2) {
    if (length(e1) != length(e2)) {
      return(FALSE)
    } else {
      return(e1 == e2)
    }
  },
  "scatter",
  FUN.VALUE = logical(1)))) {
    py$x$data[[l_i]]$x <- as.numeric(py$x$data[[l_i]]$x)
  }
  py
}

#' Mark a `dataquieR_result` for a special HTML layout
#'
#' The layout information is used by `util_pretty_print()` when wrapping
#' the rendered result into the outer `div.dataquieR_result`.  Currently
#' supported layouts are `"default"` (legacy behavior) and
#' `"2-columns-fig-left"` (plot and summary table combined in a single
#' result, shown side by side).
#'
#' @param dqr [dataquieR_result] object to tag.
#' @param layout [character] one of `"default"`, `"1-column-fig-top"`
#'                           (this is the default)
#'                           or `"2-columns-fig-left"`.
#'
#' @return `dqr`, invisibly, with the attribute `dq_layout` set.
#' @noRd
util_mark_result_layout <- function(dqr,
                                    layout = c("default",
                                               "1-column-fig-top",
                                               "2-columns-fig-left")) {
  layout <- match.arg(layout)
  attr(dqr, "dq_layout") <- layout
  invisible(dqr)
}

#' Get the layout tag of a `dataquieR_result`
#'
#' @param dqr [dataquieR_result].
#'
#' @return [character] layout name, `"default"` if unset.
#' @noRd
util_get_result_layout <- function(dqr) {
  layout <- attr(dqr, "dq_layout", exact = TRUE)
  if (is.null(layout)) {
    layout <- "default"
    fn <- attr(dqr, "function_name")
    if (length(fn) == 1 && is.character(fn)) {
      lt <- suppressWarnings(try(
        unique(util_get_concept_info("implementations",
                                     get("function_R") == fn, "layout",
                                     drop = TRUE)),
        silent = TRUE))
      if (length(lt == 1) && is.character(lt)) {
        layout <- lt
      }
    }
  }
  layout
}

#' Wrap inner HTML into a `dataquieR_result` container
#'
#' This helper centralizes the construction of the outer
#' `div.dataquieR_result` element.  It keeps legacy data attributes and
#' allows attaching additional CSS classes.
#'
#' @param inner `shiny.tag|shiny.tag.list` inner HTML to wrap.
#' @param nm [character] value for the `data-nm` attribute.
#' @param dqr [dataquieR_result] the original result object (used for
#'   the stored call).
#' @param errors [character] error messages.
#' @param warnings [character] warning messages.
#' @param messages [character] additional messages.
#' @param extra_classes [character] optional vector of extra CSS
#'   classes to add to the `dataquieR_result` container.
#'
#' @return `shiny.tag` `div.dataquieR_result`.
#' @noRd
util_wrap_dqr_result <- function(inner,
                                 nm,
                                 dqr,
                                 errors,
                                 warnings,
                                 messages,
                                 extra_classes = NULL) {
  cll <- attr(dqr, "call")
  if (is.language(cll)) {
    cll <- deparse(cll)
  }

  classes <- c("dataquieR_result", extra_classes)
  classes <- classes[!is.na(classes) & nzchar(classes)]
  classes <- paste(classes, collapse = " ")

  htmltools::div(
    title = "", # hover text already suppressed in legacy code
    class = classes,
    `data-call` = paste0(cll, collapse = "\n"),
    `data-stderr` = paste(errors, warnings, messages, collapse = "\n"),
    `data-nm` = nm,
    inner
  )
}

#' Convert single `dataquieR` result to an `htmltools` compatible object
#'
#' @param dqr [dataquieR_result] an output (indicator) from `dataquieR`
#' @param nm [character] the name used in the report, the alias name of the
#'                       function call plus the variable name
#' @param is_single_var [logical] we are creating a single variable overview
#'                                page or an indicator summary page
#' @param meta_data [meta_data]  the data frame that contains metadata
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param use_plot_ly [logical] use `plotly`
#' @param dir [character] output directory for potential `iframes`.
#' @param ... further arguments passed through, if applicable
#' @param is_ssi [logical] use this is a result for the `SSI` part of the report
#'
#' @return `htmltools` compatible object with rendered `dqr`
#'
#' @noRd
util_pretty_print <- function(dqr, nm, is_single_var,
                              meta_data,
                              label_col,
                              use_plot_ly,
                              dir,
                              ...,
                              is_ssi = FALSE
                        ) { # TODO: ensure that in square2 alias names do not have points
  variable_name_context <- FALSE

  if (use_plot_ly) {
    plot_figure <- util_plot_figure_plotly
  } else {
    plot_figure <- util_plot_figure_no_plotly
  }

  fkt <- util_map_by_largest_prefix(
    nm,
    haystack = util_all_ind_functions())
  # if (dynGet("fkt") == "acc_loess_observer") browser()
  if (!inherits(dqr, "dataquieR_NULL")) { # if the function did not return NULL or trigger an error
    # is there a filter on the implementations sheet?
    outputs <- util_get_concept_info("implementations", get("function_R")
                                     == fkt, "Reportoutputs")[["Reportoutputs"]]
    if (length(outputs) == 1 && !util_empty(outputs)) {
      outputs <- names(util_parse_assignments(outputs)) ###
      if (!is_ssi) {
        if (is_single_var) {
          outputs <- grep("(I)",
                          outputs,
                          value = TRUE,
                          invert = TRUE,
                          fixed = TRUE)
        } else {
          outputs <- grep("(V)",
                          outputs,
                          value = TRUE,
                          invert = TRUE,
                          fixed = TRUE)
        }
        outputs <- gsub("(V)", "", outputs, fixed = TRUE)
        outputs <- gsub("(I)", "", outputs, fixed = TRUE)
        remaining_slots <- intersect(outputs, names(dqr))
        if (length(remaining_slots) == 0) {
          util_message(
            "Removed all outputs of %s according to %s",
            dQuote(fkt),
            sQuote("concept | implementations | Reportoutputs"),
            level = 10
          )
        }
        for (.slot in setdiff(names(dqr), remaining_slots)) {
          dqr[[.slot]] <- NULL
        }
        # dqr <- dqr[remaining_slots] loosses attrbiutes
      }
    }

    slot <- # stores the name of the result object returned by the indicator function being displayed
      head(intersect( # check if the preferred slots are in the result names
        preferred_slots, names(dqr)), 1) # take the first available result name according to the preferred slots

    if (fkt %in% c(
      "con_limit_deviations",  # check if we are working with a limits function
      "con_hard_limits",
      "con_soft_limits",
      "con_detection_limits"
    )) {
      if ("SummaryPlotList" %in% names(dqr)) { # add an exception, the histogram is preferred to the table (heat map)
        slot <- "SummaryPlotList"
      }
    }
    if (length(slot) > 0) { # if there is an output, check and collect all warnings/errors/messages
      errors <- attr(dqr, "error") # get and store the errors from the output
      warnings <- attr(dqr, "warning")
      messages <- attr(dqr, "message")

      errors <- errors[!vapply( # do not show errors about calls, that would not be possible at all (e.g., loess for only categorical scales)
        lapply(errors, attr, "intrinsic_applicability_problem"),
        identical,
        TRUE,
        FUN.VALUE = logical(1)
      )]

      warnings <- warnings[!vapply( # do not show errors about calls, that would not be possible at all (e.g., loess for only categorical scales)
        lapply(warnings, attr, "intrinsic_applicability_problem"),
        identical,
        TRUE,
        FUN.VALUE = logical(1)
      )]

      messages <- messages[!vapply( # do not show errors about calls, that would not be possible at all (e.g., loess for only categorical scales)
        lapply(messages, attr, "intrinsic_applicability_problem"),
        identical,
        TRUE,
        FUN.VALUE = logical(1)
      )]

      # extract the condition messages, and ensure that multi-line messages are in a single line
      errors <- vapply(lapply(errors, conditionMessage), paste, #. TODO: Use conditionMessage
                       FUN.VALUE = character(1))
      warnings <- vapply(lapply(warnings, conditionMessage), paste,
                         FUN.VALUE = character(1))
      messages <- vapply(lapply(messages, conditionMessage), paste,
                         FUN.VALUE = character(1))
      x <- dqr[[slot]] # extract the corresponding result according to slot

      x <- util_remove_dataquieR_result_class(x)

      if (endsWith(slot, "PlotList")) {
        if (length(x) > 1) { # check and warn if there is more than one entry in PlotList
          # util_warning(
          #   c("Internal error: %s with > 1 result should not be in a",
          #     "v2.0 report"),
          #   sQuote(slot))
          # TODO: Implement a work-around
          # x <- do.call(htmltools::div, x)
          # x <- htmltools::div(sprintf("%s should not contain > 1 result",
          #                     sQuote(slot)))
          x <- mapply(SIMPLIFY = FALSE, pl_i = x, pl_nm = names(x),
                      function(pl_i, pl_nm) {
                        if (all(grepl(".", nm, fixed = TRUE))) {
                          .nm <- util_sub_string_left_from_.(nm)
                        } else{
                          .nm <- nm
                        }
                        as_py_i <- attr(dqr, "as_plotly")
                        util_pretty_print(util_attach_attr(
                          list(SummaryPlot = pl_i), as_plotly = as_py_i),
                                          nm =
                                            paste0(.nm, ".", pl_nm),
                                          #paste0("", head(names(pl_nm), 1)), # avoid the same name, iframe FIG_.html won't work, if > 1
                                          is_single_var = TRUE,
                                          meta_data = data.frame(),
                                          label_col = VAR_NAMES,
                                          use_plot_ly = use_plot_ly,
                                          dir = dir, ...)
                      })
          x <- lapply(x, function(p) {
            htmltools::span(style = "float:left;", p)
          })
          x <- htmltools::tagList(c(x,
                                    list(htmltools::br(style =
                                                         "clear: both;"))))
        } else {
          if (length(x) == 1) {
            x <- x[[1]] # take the single plot from PLotList
          } else {
            x <- NULL
          }
        }
      }
      # check the class of x
      if (inherits(x, "ReportSummaryTable")) {
        cats <- setdiff(colnames(x), c("Variables", "N"))
        ncats <- length(cats)
        vars <- unique(x$Variables)
        nvars <- length(vars)
        if (ncats == 0 && nvars == 0) {
          x <- NULL
          slot <- "Plot" # fake, that we have a plot
        } else if (ncats == 0 || nvars == 0) {
          x <- NULL
          slot <- "Plot" # fake, that we have a plot
        } else if (ncats == 1 && nvars == 1) {
          val <- x[x$Variables == vars, cats, drop = TRUE]
          level_names <- attr(x, "level_names")
          if (!!length(level_names)) {
            val <- level_names[as.character(val)]
            x <- htmltools::p(sprintf("%s for %s is",
                                      sQuote(cats),
                                      dQuote(vars)),
                              htmltools::strong(htmltools::em(val))
            )
          } else if ("N" %in% colnames(x)) {
            x <- htmltools::p(sprintf("%s for %s is",
                                      sQuote(cats),
                                      dQuote(vars)),
                              htmltools::strong(htmltools::em(val)),
                              "(",
                              scales::percent(val/x$N),
                              ") out of",
                              x$N
            )
          } else { # TODO: May be unreachable
            x <- htmltools::p(sprintf("%s for %s is",
                                      sQuote(cats),
                                      dQuote(vars)),
                              htmltools::strong(htmltools::em(val))
            )
          }
        } else {
          x <- util_validate_report_summary_table(x,
                                                  meta_data = meta_data,
                                                  label_col = label_col)
          x <- print.ReportSummaryTable(x, view = FALSE, ...) # convert to ggplot
        }
      }
      x <- dq_lazy_unwrap(x)
      if (inherits(x, "dq_lazy_ggplot")) x <- prep_realize_ggplot(x)
      if (inherits(x, "ggplot") || inherits(x, "ggmatrix") ||
          inherits(x, "util_pairs_ggplot_panels") ||
          inherits(x, "svg_plot_proxy") ||
          "PlotlyPlot" %in% names(dqr) ||
          inherits(x, "ggplot_built")) {
        if (inherits(x, "ggplot") || inherits(x, "ggmatrix") ||
            inherits(x, "util_pairs_ggplot_panels") ||
            inherits(x, "svg_plot_proxy") ||
            inherits(x, "ggplot_built")) {
          ggthumb <- x;
        } else {
          ggthumb <- NULL
        }
        x_is_plot <- TRUE

        # ensure, sizing hint sticks at the dqr, only
        list2env(util_fix_sizing_hints(dqr = dqr, x = x), envir = environment())

        if ("PlotlyPlot" %in% names(dqr)) {
          as_plotly <- "util_as_plotly_from_res"
        } else {
          as_plotly <- attr(dqr, "as_plotly")
        }
        if (!is.null(as_plotly) && exists(as_plotly, mode = "function"))
          as_plotly <- get(as_plotly, mode = "function")
        if (!inherits(x, "ggplot_built") &&
            !util_is_svg_object(x) &&
            use_plot_ly && is.function(as_plotly)) {
          withCallingHandlers({
            x <- as_plotly(dqr)#, height = 480, width = 1040)
            if (!identical(attr(dqr, "dont_util_adjust_geom_text_for_plotly"),
                           TRUE)) {
              x <- util_adjust_geom_text_for_plotly(x)
            }
            x <- util_decorate_plotly(x)
          },
          warning = function(cond) { # suppress a waning caused by ggplotly for barplots
            if (startsWith(conditionMessage(cond),
                           "'bar' objects don't have these attributes: 'mode'") ||
                startsWith(conditionMessage(cond),
                           "'box' objects don't have these attributes: 'mode'")) {
              invokeRestart("muffleWarning")
            }
            if (any(grepl("the mode", conditionMessage(cond)))) {
              invokeRestart("muffleWarning")
            }
          },
          message = function(cond) { # suppress a waning caused by ggplotly for barplots
            if (startsWith(conditionMessage(cond),
                           "'bar' objects don't have these attributes: 'mode'") ||
                startsWith(conditionMessage(cond),
                           "'box' objects don't have these attributes: 'mode'")) {
              invokeRestart("muffleMessage")
            }
            if (any(grepl("the mode", conditionMessage(cond)))) {
              invokeRestart("muffleMessage")
            }
          })
        } else {
          x <- plot_figure(x, sizing_hints = attr(dqr, "sizing_hints"))
        }
        # convert to plotly or base 64 plot image

        # to iframe?
        x <- util_iframe_it_if_needed(x, dir = dir, nm = nm, fkt = fkt,
                                      sizing_hints = attr(dqr, "sizing_hints"),
                                      ggthumb = ggthumb)
        # NOTE: If we have two figures in the same result, nm is not unique, because the two figures may be displayed, both., but we have currently only on figure per result, the other one can olny be a table or stuff (see this function, abobve)
      } else {
        x_is_plot <- grepl("Plot", slot)
      }
      if (is.data.frame(x)) {
        # check if dataframe is empty
        if (!prod(dim(x))) {
          x <- NULL
        } else {
          rownames(x) <- NULL # TODO: Check, if this is okay, always
          x <- util_html_table(x,
                               meta_data = meta_data,
                               label_col = label_col,
                               dl_fn = nm,
                               rotate_for_one_row = TRUE,
                               df_escape = TRUE)
        }
      }
      if (length(x) > 0) {
        if (!inherits(x, "htmlwidget") && # check if output is compatible with htmltools
            !inherits(x, "shiny.tag") &&
            !inherits(x, "shiny.tag.list")) {
          x <- htmltools::p(
            sprintf(paste("Cannot display objects of class(es) %s, yet.",
                          "Please file a feature request."),
                    paste(dQuote(class(x)), collapse = ", ")))
        }
      } else {
        x <- NULL
      }
    } else {
      x <- NULL
    }
  } else {
    x <- NULL
  }

  if (!is.null(x) || (exists("x_is_plot") && x_is_plot)) { # create and add tags and links
    if (all(grepl(".", nm, fixed = TRUE))) { # get the full name, which includes a dot
      anchor <- util_generate_anchor_tag(name = nm,
                                         order_context =
                                           ifelse(
                                             is_single_var,
                                             "variable",
                                             "indicator")
      )
      # the link is most easily created here, but therefore in the wrong position, so later it must be moved
      link <-  util_generate_anchor_link(name = nm,
                                         order_context =
                                           ifelse(
                                             is_single_var,
                                             "variable",
                                             "indicator")
      )
    } else {
      anchor <- NULL
      link <- NULL
    }

    # add variable name to pages that are not single_vars and display multiple variable in same page
    if (!is_single_var && all(grepl(".", nm, fixed = TRUE))) {
      href <- htmltools::tagGetAttribute(util_generate_anchor_link(name = nm,
                                                                   order_context =
                                                                     ifelse(
                                                                       is_single_var,
                                                                       "indicator", # link to the other world
                                                                       "variable")), attr = "href")
      if (!sub("^[^\\.]*\\.", "", nm) %in% meta_data[[label_col]]) {
        href <- NULL # TODO: Better solution
      }
      var_label <- sub("^[^\\.]*\\.", "", nm)
      caption <- var_label
      if (caption != "[ALL]") {
        variable_name_context <- TRUE
        caption <- htmltools::h5(htmltools::tags$a(href = href, var_label))
        if (label_col != VAR_NAMES) {
          if (label_col != LONG_LABEL && LONG_LABEL %in% colnames(meta_data)) {
            long_label <- util_map_labels(
              var_label,
              meta_data,
              LONG_LABEL,
              label_col,
              ifnotfound = "")
          } else {
            long_label <- ""
          }
          title <- util_map_labels(
            var_label,
            meta_data,
            VAR_NAMES,
            label_col,
            ifnotfound = NA_character_)
          if (length(title) == 1 &&
              is.character(title) &&
              !is.na(title)) {
            caption <- htmltools::h5(
              title = title,
              htmltools::tags$a(href = href, var_label),
              htmltools::br(),
              htmltools::tags$small(htmltools::tags$em(title)),
              htmltools::tags$em(long_label)
            )
          }
        }
      } else {
        caption <- NULL
      }
    } else {
      caption <- NULL
    }
    slot2 <- # stores the name of the result object returned by the indicator function being displayed
      head(intersect( # check if the preferred slots are in the result names
        setdiff(preferred_summary_slots,
                slot), names(dqr)), 1) # take the first available result name according to the preferred slots
    if (length(slot2) < 1) {
      slot2 <- NULL
    } else {
      slot2 <- slot2[[1]]
    }
    if (length(slot2)) {
      y <- dqr[[slot2]]
      if (!x_is_plot)
        y <- NULL
      if (is.data.frame(y)) {
        # check if dataframe is empty
        if (!prod(dim(y))) {
          y <- NULL
        } else {
          if (endsWith(slot2, "Table")) {
            y <- util_make_data_slot_from_table_slot(y)
          }
          rownames(y) <- NULL # TODO: Check, if this is okay, always
          if (nrow(y) > 1 && (variable_name_context || is_single_var)) {
            if ("Variables" %in% colnames(y) && length(unique(y$Variables)) == 1) {
              y$Variables <- NULL
            }
          }
          y <- util_html_table(y,
                               meta_data = meta_data,
                               label_col = label_col,
                               dl_fn = nm,
                               rotate_for_one_row = TRUE,
                               df_escape = TRUE)
        }
      }
      if (length(y) > 0) {
        if (!inherits(y, "htmlwidget") && # check if output is compatible with htmltools
            !inherits(y, "html") &&
            !inherits(y, "shiny.tag") &&
            !inherits(y, "shiny.tag.list")) {
          y <- htmltools::p(
            sprintf(paste("Cannot display objects of class(es) %s, yet.",
                          "Please file a feature request."),
                    paste(dQuote(class(y)), collapse = ", ")))
        }
      } else {
        y <- NULL
      }
    } else {
      y <- NULL
    }
    if (is_ssi) {
      anchor <- NULL
      link <- NULL
    }

    ## NEW: decide how to wrap x and y into dataquieR_result containers
    layout <- util_get_result_layout(dqr)

    if (!is.null(x) && !is.null(y) &&
        exists("x_is_plot") && x_is_plot &&
        identical(layout, "2-columns-fig-left")) {

      combined_inner <- htmltools::div(
        class = "dq-plot-table-grid",
        htmltools::div(class = "dq-plot-cell", x),
        htmltools::div(class = "dq-table-cell", y)
      )

      x <- util_wrap_dqr_result(
        combined_inner,
        nm = nm,
        dqr = dqr,
        errors = errors,
        warnings = warnings,
        messages = messages,
        extra_classes = "dq-plot-table-result"
      )
      y <- NULL
    } else {
      if (!is.null(x)) {
        x <- util_wrap_dqr_result(
          x,
          nm = nm,
          dqr = dqr,
          errors = errors,
          warnings = warnings,
          messages = messages
        )
      }
      if (!is.null(y)) {
        y <- util_wrap_dqr_result(
          y,
          nm = nm,
          dqr = dqr,
          errors = errors,
          warnings = warnings,
          messages = messages
        )
      }
    }

    x <- htmltools::tagList(anchor = anchor, link = link, caption, x, y)
    # the link is most easily added here, but still in the wrong position, so later it must be moved
  }
  x
}

util_plotly_add_modebar_buttons <- function(py, buttons) {
  if (is.null(py$x$config)) py$x$config <- list()
  if (is.null(py$x$config$modeBarButtonsToAdd)) py$x$config$modeBarButtonsToAdd <- list()

  py$x$config$modeBarButtonsToAdd <- c(py$x$config$modeBarButtonsToAdd, buttons)
  py
}

#' Adds layout and buttons to a `plotly`
#'
#' meant to work inside a report
#'
#' @param x a `plotly` object
#'
#' @returns decorated `htmltools` object
#' @noRd
util_decorate_plotly <- function(x) {

  if (!.called_in_pipeline) {
    return(x)
  }

  x <- plotly::layout(x,
                      autosize = !FALSE,
                      margin = list(autoexpand = !FALSE,
                                    r = 200,
                                    b = 100)
  )
  x <- plotly::config(x, # TODO: generalize and combine with util_plot_figure_plotly
                      responsive = TRUE,
                      autosizable = TRUE,
                      fillFrame = TRUE,
                      displaylogo = FALSE,
                      modeBarButtonsToRemove = list("toImage"))
  x <- util_plotly_add_modebar_buttons(x,
                      list(
                          list(
                            name = "Save as image",
                            icon = htmlwidgets::JS("Plotly.Icons.camera"),
                            click = htmlwidgets::JS(paste(sep = "\n",
                                                          "function(gd) {",
                                                          "  plotlyDL(gd)",
                                                          "}"))
                          ),
                          list(
                            name = "Restore initial Size",
                            icon =
                              htmlwidgets::JS("PlotlyIconshomeRED()"),
                            click =
                              htmlwidgets::JS(
                                paste(sep = "\n",
                                      "function() {",
                                      "   if (togglePlotlyWindowZoom instanceof Function) {",
                                      "     togglePlotlyWindowZoom()",
                                      "   }",
                                      "}"
                                )
                              )),
                          list(
                            name = "Print plot",
                            icon = htmlwidgets::JS("PlotlyIconsprinter()"),
                            click = htmlwidgets::JS(paste(sep = "\n",
                                                          "function(gd) {",
                                                          "  plotlyPrint(gd)",
                                                          "}"))
                          ),
                          list(
                            name = "Download as PDF",
                            icon = htmlwidgets::JS("PlotlyIconspdf()"),
                            click = htmlwidgets::JS(paste(sep = "\n",
                                                          "function(gd) {",
                                                          "  downloadPlotlyAsPDF(gd, {orientation: 'landscape', dpi: 300});",
                                                          "}"))
                          ))
  )
  util_ensure_suggested("htmlwidgets",
                        goal = "Creating the Print Dialog for Plotly") # TODO: can we simply omit the print dialog, if htmlwidgets are missing?
  x <- htmlwidgets::onRender(x, "
                          function(el, x) {
                            if (typeof injectPlotlyPrintDialog === 'function') {
                              injectPlotlyPrintDialog();
                            }
                          }
                        ")
  x
}

# preferred order of the content of the report
preferred_slots <- c("ReportSummaryTable",
                     "PlotlyPlot",
                     "SummaryPlot", "SummaryPlotList", "SummaryData",
                     "OtherData", "ResultData",
                     "SummaryTable",
                     "DataframeData", "DataframeTable", "SegmentData",
                     "SegmentTable", "OtherTable",
                     "VariableGroupPlot", "VariableGroupPlotList",
                     "VariableGroupData", "VariableGroupTable")


# preferred order of the content of the report
preferred_summary_slots <- c("SummaryData",
                             "OtherData", "ResultData",
                             "SummaryTable",
                             "DataframeData", "DataframeTable", "SegmentData",
                             "SegmentTable", "OtherTable",
                             "VariableGroupData", "VariableGroupTable")

#' Pairwise numeric variable visualization using ggplot2
#'
#' This function creates a grid of plots showing pairwise relationships between
#' numeric columns in a dataset using `ggplot2`. The diagonal contains histograms
#' or density plots, the upper triangle shows correlation coefficients, and the
#' lower triangle shows scatterplots.
#'
#' @param data A data.frame to visualize.
#' @param columns Optional character vector of column names to include.
#' @param bins Number of histogram bins. Default is 30.
#' @param title Optional title string.
#' @param columnLabels Optional vector of column labels for axis titles.
#' @param diag Character string specifying the plot type on the diagonal.
#'   Either "density" (default) or "histogram".
#'
#' @return An object of class `util_pairs_ggplot_panels`, which can be printed.
#'
#' @noRd
util_pairs_ggplot <- function(data, columns = NULL, bins = 30, title = NULL,
                              columnLabels = NULL, diag = c("density", "histogram"),
                              correlation_method = c("pearson", "spearman")) {

  util_expect_scalar(correlation_method,
                     check_type = is.character,
                     allow_more_than_one = TRUE)
  correlation_method <- util_match_arg(correlation_method)

  util_expect_scalar(diag,
                     check_type = is.character,
                     allow_more_than_one = TRUE)
  diag <- util_match_arg(diag)

  cols <- if (is.null(columns)) names(data) else intersect(columns, names(data))
  df <- data[, cols, drop = FALSE]
  df <- df[, sapply(df, is.numeric), drop = FALSE]
  n <- ncol(df)
  if (n < 2) util_error("At least two numeric columns required.",
                        applicability_problem = TRUE)
  if (correlation_method == "spearman") {
    df <- as.data.frame(lapply(df, function(x) rank(x, na.last = "keep")))
  }

  palette <- c("#0072B2", "#D55E00", "#F0E442", "#009E73", "#CC79A7", "#56B4E9", "#E69F00", "#999999")
  palette <- rep(palette, length.out = n)

  var_colors <- setNames(palette[seq_len(n)], names(df))

  make_panel <- function(i, j) {
    label_x <- if (!is.null(columnLabels)) columnLabels[j] else names(df)[j]
    label_y <- if (!is.null(columnLabels)) columnLabels[i] else names(df)[i]
    x <- names(df)[j]
    y <- names(df)[i]

    if (i == j) {
      p <- ggplot(df, aes(x = !!rlang::sym(x))) +
        xlab(label_x)
      if (diag == "histogram") {
        p <- p + geom_histogram(bins = bins, fill = var_colors[[x]], color = "white")
      } else {
        p <- p + geom_density(color = var_colors[[x]], fill = var_colors[[x]], alpha = 0.7)
      }
      p + theme_minimal() +
        theme(axis.title.y = element_blank(),
              axis.text.y  = element_blank(),
              axis.ticks.y = element_blank())
    } else if (i < j) {
      corr_val <- round(cor(df[[x]], df[[y]], use = "complete.obs"), 2)
      ggplot() +
        annotate("text", x = 0.5, y = 0.5, label = corr_val, size = 6) +
        ggplot2::theme_void()
    } else {
      ggplot(df, aes(x = !!rlang::sym(x), y = !!rlang::sym(y))) +
        xlab(label_x) + ylab(label_y) +
        geom_point(alpha = 0.6, size = 1.5, color = var_colors[[x]]) +
        theme_minimal()
    }
  }

  panels <- vector("list", n * n)
  for (i in seq_len(n)) {
    for (j in seq_len(n)) {
      panels[[(i - 1) * n + j]] <- make_panel(i, j)
    }
  }

  structure(list(panels = panels, layout_dim = n, title = title),
            class = "util_pairs_ggplot_panels")
}

#' Pairwise numeric variable visualization using plotly
#'
#' This function creates an interactive grid of plots showing pairwise
#' relationships between numeric columns in a dataset using `plotly`. The
#' diagonal contains histograms or density curves, the upper triangle shows
#' correlation coefficients, and the lower triangle shows scatterplots.
#'
#' @inheritParams util_pairs_ggplot
#'
#' @return A `plotly` object representing the interactive subplot matrix.
#'
#' @noRd
util_pairs_plotly <- function(data, columns = NULL, bins = 30, title = NULL,
                              columnLabels = NULL, diag = c("density", "histogram"),
                              correlation_method = c("pearson", "spearman")) {
  # data <- cbind(data, setNames(data, nm = paste("V", seq_len(ncol(data))))); columns <- names(data)

  util_ensure_suggested("plotly", "plot interactive figures")

  used_subsampling_for_plotly <- FALSE

  util_expect_scalar(correlation_method,
                     check_type = is.character,
                     allow_more_than_one = TRUE)
  correlation_method <- util_match_arg(correlation_method)

  util_expect_scalar(diag,
                     check_type = is.character,
                     allow_more_than_one = TRUE)
  diag <- util_match_arg(diag)

  cols <- if (is.null(columns)) names(data) else intersect(columns, names(data))
  df <- data[, cols, drop = FALSE]
  df <- df[, sapply(df, is.numeric), drop = FALSE]
  n <- ncol(df)
  if (n < 2) util_error("At least two numeric columns required.",
                        applicability_problem = TRUE)

  if (correlation_method == "spearman") {
    df <- as.data.frame(lapply(df, function(x) rank(x, na.last = "keep")))
  }

  palette <- c("#0072B2", "#D55E00", "#F0E442", "#009E73", "#CC79A7", "#56B4E9", "#E69F00", "#999999")
  palette <- rep(palette, length.out = n)

  var_colors <- setNames(palette[seq_len(n)], names(df))

  panels <- vector("list", n * n)
  cnt <- 1
  for (i in seq_len(n)) {
    for (j in seq_len(n)) {
      xname <- names(df)[j]
      yname <- names(df)[i]
      xlab <- if (!is.null(columnLabels)) columnLabels[j] else xname
      ylab <- if (!is.null(columnLabels)) columnLabels[i] else yname

      if (i == j) {
        if (diag == "histogram") {
          p <- plotly::plot_ly(x = df[[xname]], type = "histogram",
                               nbinsx = bins,
                               marker = list(color = var_colors[[xname]],
                                             line = list(color = "white")))
        } else {
          dens <- stats::density(df[[xname]], na.rm = TRUE)
          p <- plotly::plot_ly(x = dens$x, y = dens$y, type = "scatter", mode = "lines",
                               fillcolor = paste0(palette[j], "B3"), # B3 = 0.7 for alpha
                               fill = 'tozeroy', line = list(color = var_colors[[xname]]))
        }
        p <- plotly::layout(p,
                            xaxis = list(visible = TRUE,
                                         showgrid = TRUE,
                                         zeroline = TRUE,
                                         showticklabels = TRUE),
                            yaxis = list(visible = TRUE,
                                         showgrid = TRUE,
                                         zeroline = TRUE,
                                         showticklabels = TRUE))
      } else if (i < j) {
        corr_val <- round(cor(df[[xname]], df[[yname]], use = "complete.obs"), 2)
        axis_id <- paste0("corr_", i, "_", j)
        p <- plotly::plot_ly(showlegend = FALSE, type = "scatter", mode = "text",
                             xaxis = paste0("x", axis_id), yaxis = paste0("y", axis_id)) %>%
          plotly::add_annotations(
            text = corr_val,
            xref = "paper", yref = "paper",
            x = 0.5, y = 0.5, showarrow = FALSE,
            font = list(size = 20)
          ) %>%
          plotly::layout(xaxis = list(visible = FALSE),
                         yaxis = list(visible = FALSE),
                         margin = list(l = 20, b = 20))

        p <- p %>% plotly::layout(
          xaxis = list(
            title = "",
            showgrid = FALSE,
            zeroline = FALSE,
            showticklabels = FALSE
          ),
          yaxis = list(
            title = "",
            showgrid = FALSE,
            zeroline = FALSE,
            showticklabels = FALSE
          ),
          margin = list(l = 20, b = 20)
        )

      } else {
        if (nrow(df) <= 3000) {
          .df <- df
        } else {
          .df <-
            df[util_subsample_cases(df,
                                    x = xname,
                                    y = yname,
                                    nmax = 3000, # TODO: Make these parameters option()able
                                    seed = 1), , drop = FALSE]
          if (nrow(.df) < nrow(df)) {
            used_subsampling_for_plotly <- TRUE
            sampling_hint <-
              "For interactive rendering performance, sampling was used"
          }
        }
        p <- plotly::plot_ly(x = .df[[xname]], y = .df[[yname]],
                             type = "scatter", mode = "markers",
                             marker = list(opacity = 0.6, size = 6, color = var_colors[[xname]]),
                             showlegend = FALSE)
        p <- plotly::layout(p, xaxis = list(visible = TRUE,
                                            showgrid = TRUE,
                                            zeroline = TRUE,
                                            showticklabels = TRUE),
                            yaxis = list(visible = TRUE,
                                         showgrid = TRUE,
                                         zeroline = TRUE,
                                         showticklabels = TRUE))
      }

      panels[[cnt]] <- p
      cnt <- cnt + 1
    }
  }

  fig <- plotly::subplot(panels,
                         nrows = n,
                         shareX = TRUE,
                         shareY = !TRUE)

  if (!is.null(title)) {
    fig <- fig %>% plotly::layout(title = list(text = title, x = 0.5))
  }

  fig <- fig %>% plotly::layout(showlegend = FALSE)

  if (is.null(columnLabels)) {
    columnLabels <- names(df)
  }

  if (!is.null(columnLabels)) {
    annotations <- list()
    for (i in seq_len(n)) {
      label <- columnLabels[i]
      annotations[[length(annotations) + 1]] <- list(
        text = label,
        x = i / n - 1 / (2 * n),
        y = 0,
        yshift = -50,
        xref = "paper", yref = "paper",
        showarrow = FALSE,
        font = list(size = 12),
        xanchor = "center",
        yanchor = "bottom",
        cliponaxis = FALSE
      )
      annotations[[length(annotations) + 1]] <- list(
        text = label,
        x = 0,
        xshift = -50,
        y = 1 - (i / n - 1 / (2 * n)),
        xref = "paper", yref = "paper",
        showarrow = FALSE,
        font = list(size = 12),
        textangle = -90,
        xanchor = "left",
        yanchor = "middle",
        cliponaxis = FALSE
      )
    }

    fig <- fig %>% plotly::layout(annotations = annotations)
  }

  if (used_subsampling_for_plotly) {
    sampling_hint <- list(
      text = sampling_hint,
      x = 1,
      y = 0,
      xref = "paper",
      yref = "paper",
      xanchor = "right",
      yanchor = "top",
      yshift = -60,
      showarrow = FALSE,
      font = list(size = 8)
    )
    ann <- fig$x$layout$annotations
    if (is.null(ann)) ann <- list()
    fig <- fig %>% plotly::layout(annotations = c(
      ann,
      list(sampling_hint)
    ))
  }

  fig <- plotly::layout(fig,
                        margin = list(l = 60, b = 80, t = 80, r = 40)
  )

  fig
}

#' Print method for `util_pairs_ggplot_panels` objects
#'
#' @param x An object of class `util_pairs_ggplot_panels`.
#' @param ... Ignored.
#' @return The input object, invisibly.
#' @export
print.util_pairs_ggplot_panels <- function(x, ...) {
  n <- x$layout_dim
  panels <- x$panels
  title <- x$title

  grid::grid.newpage()
  grid::pushViewport(grid::viewport(layout = grid::grid.layout(n, n)))
  vplayout <- function(row, col) grid::viewport(layout.pos.row = row, layout.pos.col = col)

  for (i in seq_len(n)) {
    for (j in seq_len(n)) {
      idx <- (i - 1) * n + j
      print(panels[[idx]], vp = vplayout(i, j))
    }
  }

  if (!is.null(title)) {
    grid::grid.text(title,
                    y = grid::unit(1, "npc") - grid::unit(2, "mm"),
                    gp = grid::gpar(fontsize = 15, fontface = "bold"))
  }
  invisible(x)
}

#' `grid.draw` method for `util_pairs_ggplot_panels` objects
#'
#' @param x An object of class `util_pairs_ggplot_panels`.
#' @param ... Ignored.
#' @exportS3Method grid::grid.draw
grid.draw.util_pairs_ggplot_panels <- function(x, ...) {
  print.util_pairs_ggplot_panels(x)
}

#' Create a `storr` object with a `storr_factory` attribute
#'
#' also does basic validity checks
#'
#' @param my_storr_object a `storr`-object
#' @param my_storr_factory a function creating the/a `storr_object`
#'
#' @return `storr`-object with the factory attribute and (hopefully) valid.
#' @noRd
util_storr_factory <- function(my_storr_object, my_storr_factory) {
  if (!missing(my_storr_object) && is.null(my_storr_object)) {
    return(NULL)
  }
  if (!missing(my_storr_object) && missing(my_storr_factory)) {
    my_storr_factory <- attr(my_storr_object, "storr_factory")
  }
  if ((missing(my_storr_object) || is.null(my_storr_object)) &&
      (missing(my_storr_factory) || is.null(my_storr_factory))) {
    return(NULL)
  }
  if (!is.function(my_storr_factory) ||
      length(formals(my_storr_factory)) != 0) {
    # util_error("No storr factory")
    my_storr_factory <- function() {
      return(my_storr_object)
    }
  }
  if (missing(my_storr_object)) {
    my_storr_object <- my_storr_factory()
  }
  if (!inherits(my_storr_object, "storr")) {
    util_error("storr factory should return a storr object")
  }
  is_rds <- identical(try(my_storr_object$driver$type(), silent = TRUE),
                        "rds")
  if (!is_rds) {
    rlang::warn(
      "storr classes other than RDS not yet supported, expect errors.",
      .frequency = "regularly", .frequency_id = rlang::hash(my_storr_factory))
    # for RDS, the following does not typically happen
    if (util_is_try_error(try(my_storr_object$list_hashes(), silent = TRUE))) {
      try(my_storr_object$driver$reconnect(), silent = TRUE)
    }
    if (util_is_try_error(try(my_storr_object$list_hashes(), silent = TRUE))) {
      my_storr_object <- my_storr_factory()
    }
  }
  if (!inherits(my_storr_object, "storr")) {
    util_error("storr factory should return a storr object")
  }
  if (!is_rds &&
      util_is_try_error(try(my_storr_object$list_hashes(), silent = TRUE))) {
    util_error("storr object not working")
  }
  attr(my_storr_object, "storr_factory") <- my_storr_factory
  my_storr_object
}

#' Create a `storr`-object using the factory
#'
#' also performs checks.
#'
#' @param my_storr_factory a function returning a `storr` object
#'
#' @return a `storr` object
#' @noRd
util_storr_object <- function(my_storr_factory = function() {
  storr::storr_environment()
}) {
  util_storr_factory(my_storr_factory = my_storr_factory)
}

#' Get the `storr` object backing a report
#'
#' @param r the dataquieR_resultset2 / report
#'
#' @return the `storr` object holding the results or `NULL`, if the report
#'         lives in the memory, only
#' @noRd
util_get_storr_object_from_report <- function(r) {
  my_storr_object <- attr(r, "my_storr_object")
  storr_factory <- attr(my_storr_object, "storr_factory")
  util_storr_factory(my_storr_object, storr_factory)
}

#' Fix a `storr` object, if it features the factory-attribute
#'
#' @seealso `util_storr_factory()`
#'
#' @param my_storr_object a `storr`-object
#'
#' @return a (hopefully) working `storr_object`
#' @noRd
util_fix_storr_object <- function(my_storr_object) {
  storr_factory <- attr(my_storr_object, "storr_factory")
  util_storr_factory(my_storr_object, storr_factory)
}


# Example for thor:
# unlink("/tmp/thor", recursive = TRUE)
# r_rds <- dq_report2("study_data", meta_data_v2 = "meta_data_v2", dimensions = NULL, storr_factory = function(){thor::storr_thor(thor::mdb_env("/tmp/thor", mapsize =  2 * (2 ^ 10)^4))}, amend = T)


#' Create a factory function for `storr` objects for backing
#' a [dataquieR_resultset2]
#'
#' @param db_dir [character] path to the directory for the back-end, if
#'                           one is created on the fly.
#' @param namespace [character] namespace for the report, so that one back-end
#'                              can back several reports
#'
#' the returned function will try to create a `storr` object using a temporary
#' folder or the folder in `db_dir`, if specified. The database will either
#' be the `storr_rds`.
#'
#' @return `storr` object or `NULL`, if package `storr` is not available
#' @export
prep_create_storr_factory <- function(db_dir = tempfile(),
                                      namespace = "objects") {
  if (missing(db_dir)) {
    withr::defer_parent({
      if (dir.exists(db_dir) && startsWith(db_dir, tempdir())) {
        try(unlink(db_dir, recursive = TRUE, force = TRUE), silent = TRUE)
      }
    })
  }
  force(db_dir)
  force(namespace)
  function() {
    if (util_ensure_suggested("storr", err = FALSE)) { # TODO: store somewhere the full constructor call to reproduce.
      storr_factory <- function() {
        storr::storr_rds(db_dir, default_namespace =
                           namespace)
      }
      my_storr_object <- storr_factory()
      attr(my_storr_object, "storr_factory") <- storr_factory
    } else {
      my_storr_object <- NULL
    }
    my_storr_object
  }
}

#' Utility function to put strings in quotes
#'
#' This function generates usual double-quotes for each element of the character
#' vector
#'
#' @param string Character vector
#'
#' @return quoted string
#'
#' @family string_functions
#' @concept process
#' @noRd
util_set_dQuoteString <- function(string) {
  old <- options(useFancyQuotes = FALSE)
  on.exit(options(old))
  dQuote(string)
}

#' @family reporting_functions
#' @concept process
#' @noRd
util_remove_dataquieR_result_class <- function(x) { # TODO: The [[ for dataquieR_result should not work this way, a wrapper class would be needed, instead. Then, this funcion can be abandondd
  remove_class <- function(x) { # FIXME: After the modification from this commit, is this function needed at all, any more?
    if (!is.null(x)) {
      if (is.list(x) && !inherits(x, "patchwork") && !inherits(x, "ggmatrix") &&
          !inherits(x, "svg_plot_proxy") &&
          !inherits(x, "util_pairs_ggplot_panels") &&
          !inherits(x, "dq_lazy_ggplot")) {
        x[] <- lapply(x, remove_class)
      }
      newclass <- setdiff(class(x), "dataquieR_result")
      if (length(newclass)) {
        class(x) <- newclass
      } else {
        x <- unclass(x)
      }
      attr(x, "error") <- NULL
      attr(x, "warning") <- NULL
      attr(x, "message") <- NULL
    }
    x
  }
  remove_class(x)
}

#' Generate a full DQ report, v2
#'
#' @inheritParams .template_function_indicator
#'
#' @param ... arguments to be passed to all called indicator functions if
#'            applicable.
#' @param cores [integer] number of cpu cores to use or a named list with
#'                        arguments for [parallelMap::parallelStart] or NULL,
#'                        if parallel has already been started by the caller.
#'                        Can also be a cluster.
#' @param ignore_empty_vars [enum] TRUE | FALSE | auto. See
#'                             [dataquieR.ignore_empty_vars].
#' @param specific_args [list] named list of arguments specifically for one of
#'                             the called functions, the of the list elements
#'                             correspond to the indicator functions whose calls
#'                             should be modified. The elements are lists of
#'                             arguments.
#' @param dimensions [dimensions] Vector of dimensions to address in the report.
#'                   Allowed values in the vector are Completeness, Consistency,
#'                   and Accuracy. The generated report will only cover the
#'                   listed data quality dimensions. Accuracy is computational
#'                   expensive, so this dimension is not enabled by default.
#'                   Completeness should be included, if Consistency is
#'                   included, and Consistency should be included, if Accuracy
#'                   is included to avoid misleading detections of e.g. missing
#'                   codes as outliers, please refer to the data quality concept
#'                   for more details. Integrity is always included.
#'                   If dimensions is equal to NULL or "all", all dimensions
#'                   will be covered.
#' @param author [character] author for the report documents.
#' @param debug_parallel [logical] print blocks currently evaluated in parallel
#' @param meta_data_segment [data.frame] -- optional: Segment level metadata
#' @param meta_data_dataframe [data.frame] -- optional: Data frame level
#'                                                                 metadata
#' @param meta_data_cross_item [data.frame] -- optional: Cross-item level
#'                                                                 metadata
#' @param user_info [list] additional info stored with the report, e.g.,
#'                         comments, title, ...
#' @param resp_vars [variable list] the name of the measurement variables
#'                                  for the report. If missing, all variables
#'                                  will be used. Only item level indicator
#'                                  functions are filtered, so far.
#' @param filter_indicator_functions [character] regular expressions, only
#'                                               if an indicator function's name
#'                                               matches one of these, it'll
#'                                               be used for the report. If
#'                                               of length zero, no filtering
#'                                               is performed.
#' @param exclude_indicator_functions [character] regular expressions,
#'                                               if an indicator function's name
#'                                               matches one of these, it'll
#'                                               be excluded from the report. If
#'                                               of length zero, no filtering
#'                                               is performed.
#' @param filter_result_slots [character] regular expressions, only
#'                                               if an indicator function's
#'                                               result's name
#'                                               matches one of these, it'll
#'                                               be used for the report. If
#'                                               of length zero, no filtering
#'                                               is performed.
#' @param mode [character] work mode for parallel execution. default is
#'              "default", the values mean:
#'              - default: use `queue` except `cores` has been set explicitly
#'              - futures: use the `future` package
#'              - queue: use a queue as described in the examples
#'                from the `callr` package by Csárdi and Chang and start
#'                sub-processes as workers that evaluate the queue.
#'              - parallel: use the cluster from `cores` to evaluate all
#'                         calls of indicator functions using the classic
#'                         R `parallel` back-ends
#'
#' @param mode_args [list] of arguments for the selected `mode`. As of writing
#'                         this manual, only for the mode `queue` the argument
#'                         `step` is supported, which gives the number of
#'                         function calls that are run by one worker at a time.
#'                         the default is 15, which gives on most of the tested
#'                         systems a good balance between synchronization
#'                         overhead and idling workers.
#' @param notes_from_wrapper [list] a list containing notes about changed labels
#'                                  by `dq_report_by` (otherwise NULL)
#' @param title [character] optional argument to specify the title for
#'                          the data quality report
#' @param subtitle [character] optional argument to specify a subtitle for
#'                             the data quality report
#' @param advanced_options [list] options to set during report computation,
#'                                see [options()]
#' @param meta_data_item_computation [data.frame] optional. computation rules
#'                                              for computed variables.
#' @param storr_factory [function] `NULL`, or
#'                        a function returning a `storr` object as
#'                        back-end for the report's results. If used with
#'                        `cores > 1`, the storage must be accessible from all
#'                        cores and capable of concurrent writing according
#'                        to `storr`. Hint: `dataquieR` currently only supports
#'                        `storr::storr_rds()`, officially, while other back-
#'                        ends may nevertheless work, yet, they are not tested.
#' @param amend [logical] if there is already data in.`storr_factory`,
#'                        use it anyways -- unsupported, so far!
#' @param checkpoint_resumed [logical] if using a `storr_factory` and the back-
#'                                     end there is already filled, and if
#'                                     `amend` is missing or set to `TRUE`,
#'                                     compute all missing result and add them
#'                                     to the back-end.
#' @param cross_item_level [data.frame] alias for `meta_data_cross_item`
#' @param `cross-item_level` [data.frame] alias for `meta_data_cross_item`
#' @param segment_level [data.frame] alias for `meta_data_segment`
#' @param dataframe_level [data.frame] alias for `meta_data_dataframe`
#' @param item_computation_level [data.frame] alias for
#'                               `meta_data_item_computation`
#' @param .internal [logical] internal use, only.
#' @param name_of_study_data [character] name for study data inside the report,
#'                                       internal use.
#' @param dt_adjust [logical] whether to trust data types in the study data. if
#'                            `TRUE`, data types are checked based on the
#'                            metadata and later casted to the declared type.
#'                            if your data source is already typed, this can
#'                            be turned off to speed up computations.
#'                            see [dataquieR.dt_adjust]
#' @param output_dir [character] if `output_dir` is not `NULL`, also create
#'                             `HTML` output for the report using
#'                             [print.dataquieR_resultset2()]
#'                               written to the path `output_dir`
#' @param force_overwrite [logical] force to overwrite `output_dir`, even if it
#'                                 exists
#'
#' @return a [dataquieR_resultset2] that can be
#' [printed][print.dataquieR_resultset2] creating a `HTML`-report.
#'
#' @details
#' See [dq_report_by] for a way to generate stratified or splitted reports
#' easily.
#'
#' @seealso
#' `r paste0(" * [", methods(class="dataquieR_resultset"), "]", collapse="\n")`
#' * [dq_report_by]
#' @export
#' @importFrom stats alias
#' @importFrom utils osVersion packageName packageVersion
#' @importFrom stats setNames
dq_report2 <- function(study_data, # TODO: make meta_data_segment, ... optional
                       item_level = "item_level",
                       label_col = LABEL,
                       meta_data_segment = "segment_level",
                       meta_data_dataframe = "dataframe_level",
                       meta_data_cross_item = "cross-item_level",
                       meta_data_item_computation =
                         "item_computation_level",
                       meta_data = item_level,
                       meta_data_v2,
                       ...,
                       dimensions = c("Completeness", "Consistency"),
                       cores = list(mode = "socket",
                                    logging = FALSE,
                                    cpus = util_detect_cores(),
                                    load.balancing = TRUE),
                       ignore_empty_vars =
                         getOption("dataquieR.ignore_empty_vars",
                                   dataquieR.ignore_empty_vars_default),
                       specific_args = list(), # TODO: check if list of lists, or support not only function but also call names w/o var-suffix in this list: specific_args = list(acc_margins_observer = list()) or specific_args = list(acc_margins_observer.SBP_0 = list())
                       advanced_options =  list(),
                       author = prep_get_user_name(),
                       title = "Data quality report",
                       subtitle = as.character(Sys.Date()),
                       user_info = NULL,
                       debug_parallel = FALSE,
                       resp_vars = character(0),
                       filter_indicator_functions = character(0),
                       exclude_indicator_functions = character(0),
                       filter_result_slots = c(
                         "^Summary",
                         "^Segment",
                         "^DataTypePlotList",
                         "^ReportSummaryTable",
                         "^Dataframe",
                         "^Result",
                         "^VariableGroup"
                       ),
                       mode = c("default", "futures", "queue", "parallel"),
                       mode_args = list(),
                       notes_from_wrapper = list(),
                       storr_factory = NULL,
                       amend = FALSE, # TODO: reduced output (png plus table): util_plot2svg_object would do with svg, png version was tested, but uses typically much more memory. Function with this commit but removed thereafter. use lobstr::obj_size() for testing w/o save() :)
                       cross_item_level,
                       `cross-item_level`,
                       segment_level,
                       dataframe_level,
                       item_computation_level,
                       .internal =
                         rlang::env_inherits(
                           rlang::caller_env(),
                           parent.env(environment())),
                       checkpoint_resumed =
                         getOption("dataquieR.resume_checkpoint",
                                   dataquieR.resume_checkpoint_default),
                       name_of_study_data,
                       dt_adjust = as.logical(getOption("dataquieR.dt_adjust",
                                             dataquieR.dt_adjust_default)),
                       output_dir = NULL,
                       force_overwrite = FALSE
                       ) {

  util_expect_scalar(force_overwrite, check_type = is.logical)

  rep_id <- util_make_report_id()

  util_stop_if_not(is.list(advanced_options))

  old_O <- options(
    c(
      list(
        dataquieR.CONDITIONS_WITH_STACKTRACE = FALSE,
        dataquieR.ERRORS_WITH_CALLER = FALSE,
        dataquieR.MESSAGES_WITH_CALLER = FALSE,
        dataquieR.WARNINGS_WITH_CALLER = FALSE
      ),
      advanced_options
    )
  )
  on.exit(options(old_O))

  my_storr_object <- util_storr_object(storr_factory)

  util_expect_scalar(dt_adjust, check_type = is.logical)
  util_expect_scalar(amend, check_type = is.logical)
  util_expect_scalar(checkpoint_resumed, check_type = is.logical)

  if (missing(amend) && !amend && checkpoint_resumed) {
    util_message(c("%s was %s, but %s was unset (default = %s).",
                   "I will set %s to %s to make %s work."),
                 sQuote("checkpoint_resumed"),
                 sQuote("TRUE"),
                 sQuote("amend"),
                 sQuote("FALSE"),
                 sQuote("amend"),
                 sQuote("TRUE"),
                 sQuote("checkpoint_resumed"))
    amend = TRUE
  } else if (!missing(amend) && !amend && checkpoint_resumed) {
    util_message(c("%s was %s, but %s was %s.",
                   "I will set %s to %s to avoid results from being touched",
                   "(%s = %s)"),
                 sQuote("amend"),
                 sQuote("FALSE"),
                 sQuote("checkpoint_resumed"),
                 sQuote("TRUE"),
                 sQuote("checkpoint_resumed"),
                 sQuote("FALSE"),
                 sQuote("amend"),
                 sQuote("FALSE")
                 )
    checkpoint_resumed <- FALSE
  }

  if (!is.null(my_storr_object) && (
    length(my_storr_object$list()) > 0 ||
    length(my_storr_object$list(
      util_get_storr_att_namespace(my_storr_object))) > 0 ||
    length(my_storr_object$list(
      util_get_storr_summ_namespace(my_storr_object))) > 0
  )) {
    if (amend) {
      util_message(c("Your storr-object is not empty, but %s was set %s,",
                     "so I'll amend the storage object. This is unsupported,",
                     "yet, so expect strange behavior."),
                   dQuote("amend"), sQuote(TRUE))
    } else {
      util_error(c("Your storr-object is not empty, and %s was set %s,",
                   "so I won't amend the storage object, which would",
                   "still be unsupported, so could cause strange behavior.",
                   "We strongly recommend to use clear storr objects (or",
                   "at least the default namespace (%s in your case)",
                   "and its sister namespaces (the default namespace suffixed",
                   "with %s and %s, should be empty. In case of %s, just",
                   "delete the folder that backs the storr."),
                 dQuote("amend"),
                 sQuote(FALSE),
                 sQuote(my_storr_object$default_namespace),
                 sQuote(".attributes"),
                 sQuote(".summary"),
                 sQuote("driver_rds")
                 )
    }
  }

  util_match_arg(ignore_empty_vars,
                 c("TRUE", "FALSE", "auto"))

  mode <- util_match_arg(mode)

  if (missing(title)) {
    attr(title, "default") <- TRUE
  } else {
    attr(title, "default") <- FALSE
  }

  if (missing(subtitle)) {
    attr(subtitle, "default") <- TRUE
  } else {
    attr(subtitle, "default") <- FALSE
  }

  util_expect_scalar(title, check_type = is.character,
                     error_message = sprintf("%s needs to be character(1)",
                                             sQuote("title")))

  util_expect_scalar(subtitle, check_type = is.character,
                     error_message = sprintf("%s needs to be character(1)",
                                             sQuote("subtitle")))

  .hi <- .hp <- .hm <- NULL
  content_file <- NULL

  if (!missing(output_dir)) {

    output_dir <- util_normalize_path(output_dir)

    content_file <- file.path(output_dir, "index.html")

    util_expect_scalar(output_dir, check_type = is.character)

    util_overwrite_if_requested(output_dir, force_overwrite)

    util_message("Writing to %s", dQuote(content_file))

    packageName <- utils::packageName()

    list2env(util_init_html_progress(output_dir = output_dir,
                            content_file = content_file,
                            title = title,
                            view = TRUE,
                            rep_id = rep_id), envir = environment())
  }

  if (!is.null(cores) && (missing(cores) || (
    is.vector(cores) && length(cores) == 1 && util_is_integer(cores)) &&
    cores > 1) &&
      mode == "default") {
    if (util_ensure_suggested(c("R6", "processx"),
                              goal =
     "use the queue mode, which is faster than parallel (but cannot run on distributed cluster nodes)",
                              err = FALSE)) {
      mode <- "queue"
    } else {
      mode <- "parallel"
    }
  } else if (mode == "default") {
      mode <- "parallel"
  }

  if (suppressWarnings(util_ensure_suggested("testthat", err = FALSE))) {
    if (testthat::is_testing()) {
      if (!(mode %in% c("parallel", "queue"))) {
        util_warning(
          "Internal problem: %s should be %s, %s or %s in the context of %s",
                     sQuote("mode"), dQuote("queue"), dQuote("parallel"),
          dQuote("default"),
          sQuote("testthat"))
      }
      if (!rlang::is_scalar_integerish(cores) ||
          cores > 1L) {
        if (!is.null(cores))
          util_warning(
            "Internal problem: %s should be an integer below %s in the context of %s",
            sQuote("cores"), dQuote("2"), sQuote("testthat"))
      }
    }
  }

  if (!missing(meta_data_v2)) {
    util_message("Have %s set, so I'll remove all loaded data frames",
                 sQuote("meta_data_v2"))
    prep_purge_data_frame_cache()
    prep_load_workbook_like_file(meta_data_v2)
    if (!exists("item_level", .dataframe_environment())) {
      w <- paste("Did not find any sheet named %s in %s, is this",
             "really dataquieR version 2 metadata?")
      if (requireNamespace("cli", quietly = TRUE)) {
        w <- cli::bg_red(cli::col_br_yellow(w))
      }
      util_warning(w, dQuote("item_level"), dQuote(meta_data_v2),
                   immediate = TRUE)
    }
  }

  # checks and fixes the function arguments
  util_ck_arg_aliases()

  if (missing(name_of_study_data)) {
    if (is.data.frame(study_data)) {
      name_of_study_data <- substr(head(as.character(substitute(study_data)), 1),
                                   1, 500)
    } else if (length(study_data) == 1 && is.character(study_data)) {
      name_of_study_data <- study_data
    } else {
      name_of_study_data <- "??No study data found??"
    }
  } else {
    util_expect_scalar(name_of_study_data,
                       check_type = is.character)
    substr(name_of_study_data, 1, 500)
  }
  util_expect_data_frame(study_data, keep_types = TRUE)
#  if (ncol(study_data) == 0) {
#    util_error("Study data has no column")
#  }
  prep_add_data_frames(data_frame_list = setNames(list(study_data),
                                                  nm = name_of_study_data))
  prep_add_data_frames(data_frame_list = setNames(list(study_data),
                                                  nm = "study_data"))
#TODO:  FIX with this new utility function util_fix_missing_table_references
  # to amend the metadata files, remove tables that are not existing here
  util_handle_val_tab()

  # checks the data frame names in the dataframe cache
  if (!.internal) {
    util_verify_names(name_of_study_data = name_of_study_data)
  }

  try(util_expect_data_frame(meta_data), silent = TRUE)

  case_insens <- util_is_na_0_empty_or_false(
    getOption("dataquieR.study_data_colnames_case_sensitive",
              dataquieR.study_data_colnames_case_sensitive_default))

  if (case_insens) {
    colnames(study_data) <-
      util_align_colnames_case(.colnames = colnames(study_data),
                               .var_names = meta_data[[VAR_NAMES]])
  }

  ci_in_study <- FALSE
  if (is.data.frame(meta_data) && !case_insens) {
    in_study <- meta_data[[VAR_NAMES]] %in% colnames(study_data)
    ci_in_study <- tolower(meta_data[[VAR_NAMES]]) %in%
      tolower(colnames(study_data))
  }
  if (identical(getOption("dataquieR.ELEMENT_MISSMATCH_CHECKTYPE",
                dataquieR.ELEMENT_MISSMATCH_CHECKTYPE_default), "subset_u")) {
    if (is.data.frame(meta_data)) {
      meta_data <- meta_data[in_study, , FALSE]
    }
  }

  # try, meta_data my still be missing
  # note: we had the following twice, first w/o try, then w/ try. likely a bug
  try(meta_data <- prep_meta_data_v1_to_item_level_meta_data(meta_data),
                   silent = TRUE)
  warning_pred_meta <- NULL
  if (!is.data.frame(meta_data) || !prod(dim(meta_data))) {
    try_ci <- ""
    if (!case_insens) {
      if (any(ci_in_study)) {
        try_ci <- sprintf(
          paste("But maybe, if you enable case-insensitive mapping",
                "of meta_data on study data using %s, it could work?"),
          sQuote(
            "options(dataquieR.study_data_colnames_case_sensitive = FALSE)")
        )
      }
    }
    w <- paste("No item level metadata matching study data found. Will guess",
           "some from the study data. This will not be very helpful, please",
           "consider passing an item level metadata file.",
           try_ci)
    if (requireNamespace("cli", quietly = TRUE)) {
      w <- cli::bg_red(cli::col_br_yellow(cli::ansi_toupper(w)))
      w <- gsub("%S", "%s", w)
    }
    util_warning(w,
                 immediate = TRUE)
    predicted <- prep_study2meta(study_data, convert_factors = TRUE)
    meta_data <- predicted$MetaData
    study_data <- predicted$ModifiedStudyData
    warning_pred_meta <- paste("No item-level metadata matching study data",
                               "could not be found and was guessed from the",
                               "study data.", try_ci)
  } else {
    # strip rownames from metadata to prevent confusing the html_table function
    rownames(meta_data) <- NULL
  }
  try(util_expect_data_frame(meta_data_segment), silent = TRUE)
  if (!is.data.frame(meta_data_segment)) {
    util_message("No segment level metadata %s found",
                 dQuote(meta_data_segment))
    meta_data_segment <- data.frame(STUDY_SEGMENT =
                                      unique(meta_data$STUDY_SEGMENT))
  } else {
    # strip rownames from metadata to prevent confusing the html_table function
    rownames(meta_data_segment) <- NULL
  }
  try(util_expect_data_frame(meta_data_dataframe), silent = TRUE)
  if (!is.data.frame(meta_data_dataframe)) {
    util_message("No dataframe level metadata %s found.",
                 dQuote(meta_data_dataframe))
    meta_data_dataframe <- data.frame(DF_NAME =
                                        name_of_study_data)
  } else {
    # strip rownames from metadata to prevent confusing the html_table function
    rownames(meta_data_dataframe) <- NULL
  }
  try(util_expect_data_frame(meta_data_cross_item), silent = TRUE)
  if (!is.data.frame(meta_data_cross_item)) {
    util_message("No cross-item level metadata %s found",
                   dQuote(meta_data_cross_item))
    meta_data_cross_item <- data.frame(VARIABLE_LIST = character(0),
                                      CHECK_LABEL = character(0))
  } else {
    # strip rownames from metadata to prevent confusing the html_table function
    rownames(meta_data_cross_item) <- NULL
  }

  suppressWarnings(util_ensure_in(VAR_NAMES, names(meta_data), error = TRUE,
                 err_msg =
                   sprintf("Did not find the mandatory column %%s in the %s.",
                           sQuote("meta_data"))))

  if (!(VARIABLE_ROLE %in% colnames(meta_data))) {
    util_message("No %s assigned in item level metadata. Defaulting to %s.",
                 sQuote(VARIABLE_ROLE), dQuote(VARIABLE_ROLES$PRIMARY),
                 applicability_problem = TRUE)
    meta_data$VARIABLE_ROLE <- VARIABLE_ROLES$PRIMARY
  }

  which_not <- !(meta_data[[VARIABLE_ROLE]] %in% VARIABLE_ROLES)
  if (any(which_not)) { # Also for NA cells
    util_message(
         c("The variables %s have no or an invalid %s assigned in item level",
           "metadata: %s are not in %s. Defaulting to %s."), # TODO: normalize VARIABLE_ROLES
         util_pretty_vector_string(n_max = 5, meta_data[which_not, label_col,
                                               drop = TRUE]),
         sQuote(VARIABLE_ROLE),
         dQuote(util_pretty_vector_string(sort(unique(meta_data[which_not,
                                                         VARIABLE_ROLE,
                                             drop = TRUE])))),
         util_pretty_vector_string(VARIABLE_ROLES),
         dQuote(VARIABLE_ROLES$PRIMARY),
             applicability_problem = TRUE)
    meta_data$VARIABLE_ROLE[which_not] <- VARIABLE_ROLES$PRIMARY
  }

  util_expect_scalar(label_col, check_type = is.character)
  util_ensure_in(label_col, names(meta_data), error = TRUE,
                 err_msg =
        sprintf("Did not find a label column (%s) named %%s in the %s. Did you mean %%s?",
                sQuote("label_col"),
                    sQuote("meta_data")))

  if (MISSING_CODE_RULES %in% prep_list_dataframes()) {
    # https://gitlab.com/libreumg/dataquier/-/work_items/632
    missing_code_rules <- MISSING_CODE_RULES
    util_expect_data_frame(missing_code_rules)
    pamc <- try(prep_add_missing_codes(
      use_value_labels = FALSE, # TODO
      study_data = study_data,
      meta_data = meta_data,
      rules = missing_code_rules
    ), silent = TRUE)
    if (util_is_try_error(pamc)) {
      cnd <- util_condition_from_try_error(pamc)
      util_warning("Could not amend missing codes based on the sheet %s: %s",
                   sQuote(MISSING_CODE_RULES),
                   dQuote(conditionMessage(cnd)),
                   applicability_problem = TRUE)
    } else {
      study_data <- pamc$ModifiedStudyData
      meta_data <- pamc$ModifiedMetaData
    }
  }

  try(util_expect_data_frame(meta_data_item_computation), silent = TRUE)
  if (!is.data.frame(meta_data_item_computation)) {
    meta_data_item_computation <- data.frame()
  }

  # computed variables from cross sheet ------

  uaci <-
    util_add_computed_internals(meta_data_item_computation,
                                meta_data_cross_item,
                                meta_data,
                                label_col)
  meta_data_item_computation <-
    uaci$meta_data_item_computation
  meta_data <- uaci$meta_data
  if (is.data.frame(meta_data_item_computation)) {
    if (!!prod(dim(meta_data_item_computation)))
      util_message("Computed items metadata defined. Computing them...")
    # strip rownames from metadata to prevent confusing the html_table function
    rownames(meta_data_item_computation) <- NULL
    function_e <- environment()
    status <- try(local({
      res <- prep_add_computed_variables(
        study_data = study_data,
        meta_data = meta_data,
        label_col = label_col,
        rules = meta_data_item_computation
      )
      msd <- res$ModifiedStudyData[, meta_data_item_computation$VAR_NAMES,
                                   drop = FALSE]
      mapped <- identical(attr(res$ModifiedStudyData, "MAPPED"), TRUE)
      if (mapped) {
        mapped_lc <- attr(res$ModifiedStudyData, "label_col")
        colnames(msd) <- prep_map_labels(colnames(msd),
                               meta_data = meta_data,
                               to = VAR_NAMES,
                               from = mapped_lc,
                               ifnotfound = colnames(msd),
                               warn_ambiguous = FALSE)
      }
      function_e$study_data[,
                            setdiff(colnames(msd),
                                    colnames(function_e$study_data))] <-
        msd[, setdiff(colnames(msd),
                      colnames(function_e$study_data))]
      invisible(NULL)
    }), silent = TRUE)
    if (inherits(status, "try-error")) {
      util_warning(status)
    } else {
      prep_add_data_frames(data_frame_list = setNames(list(study_data),
                                                      nm = name_of_study_data))
      prep_add_data_frames(data_frame_list = setNames(list(study_data),
                                                      nm = "study_data"))
    }
  } else {
    meta_data_item_computation <- NULL
  }

  # ensure that VAR_NAMES and labels exist, are unique and not too long
  mod_label <- util_ensure_label(meta_data = meta_data,
                                 label_col = label_col)
  if (!is.null(mod_label$label_modification_text)) {
    # There were changes in the metadata.
    meta_data <- mod_label$meta_data
    label_col <- mod_label$label_col
  }
  # Since we may also map to other metadata label columns, we have to ensure
  # that none of them contains empty fields to prevent errors.
  for (lcol in unique(c(label_col, LABEL, LONG_LABEL))) {
    if (is.data.frame(meta_data) && lcol %in% colnames(meta_data)) {
      meta_data[[lcol]][which(util_empty(meta_data[[lcol]]))] <-
        meta_data[[VAR_NAMES]][which(util_empty(meta_data[[lcol]]))]
    }
  }

  all_vars <- (length(resp_vars) == 0)

  # remove unrelated rules, if resp_vars -----
  if (!all_vars && nrow(meta_data_cross_item) > 0) {
    # TODO: consolidate with the literally copied code in dq_report_by
    # 1st: extract all rules from cross-item containing a var from resp_vars
    #combine var_names with labels
    vars <- util_map_labels(resp_vars,
                            meta_data = meta_data,
                            to = label_col,
                            from = VAR_NAMES,
                            ifnotfound = NA_character_)
    # create a list with the vars from the column
    # variable_list in cross-item metadata
    rules_vars <-
      util_parse_assignments(meta_data_cross_item$VARIABLE_LIST,
                             multi_variate_text =
                               TRUE)

    # intersect vars in the rules with vars
    # in the segment to see which rules affect the
    # current segment and discard rules not affected by the resp_vars
    rules_to_use_cil <-
      vapply(lapply(rules_vars, intersect, vars),
             length,
             FUN.VALUE = integer(1)) > 0
    meta_data_cross_item <- meta_data_cross_item[rules_to_use_cil, , FALSE]
  }
  # end remove unrelated rules -----


  meta_data_cross_item <- util_normalize_cross_item(
    meta_data = meta_data,
    meta_data_cross_item = meta_data_cross_item,
    label_col = label_col
  )

  util_expect_scalar(dimensions,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character,
                     error_message =
                       sprintf("The argument %s must be character or NULL",
                               sQuote("dimensions")))
  if (length(dimensions) == 0 || (length(dimensions) == 1 &&
                                  !is.na(dimensions) &&
                                  tolower(trimws(dimensions)) == "all")) {
    dimensions <- c("completeness", "consistency", "accuracy")
  } else {
    dimensions <- tolower(dimensions)
  }
  dimensions[dimensions %in% c("acc", "accuracy")] <- "Accuracy"
  dimensions[dimensions %in% c("con", "consistency")] <- "Consistency"
  dimensions[dimensions %in% c("com", "completeness")] <- "Completeness"
  dimensions[dimensions %in% c("int", "integrity")] <- "Integrity"
  dimensions[dimensions %in% c("des", "descriptors")] <- "Descriptors"
  .dimensions <-
    util_ensure_in(dimensions,
                   c("Completeness", "Consistency", "Accuracy", "Integrity",
                     "Descriptors"),
                   error = FALSE,
                   applicability_problem = TRUE
              )

  util_expect_scalar(resp_vars,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)

  md100 <- meta_data # metadata with study data

  miss_from_study <- (!(md100[[VAR_NAMES]] %in% (c(colnames(study_data),
                                                   meta_data_item_computation$VAR_NAMES))))

  only_nas <- vapply(setNames(md100[[VAR_NAMES]], nm = md100[[label_col]]),
                     function(vn) {
                       all(util_empty(study_data[[vn]]))
                     },
                     FUN.VALUE = logical(1))

  if (any(miss_from_study)) {
    vars_not_found <- paste0(
      dQuote(paste0(md100[miss_from_study, label_col], " (",
                    md100[miss_from_study, VAR_NAMES], ")")),
      collapse = ", ")
    util_message(
      c("Could not find the following variables in %s:",
        "%s.\nThese will be preliminarily removed from the %s."),
      sQuote("study_data"),
      vars_not_found,
      sQuote("meta_data")
    )
    md100 <- md100[!miss_from_study, , FALSE]
  }

  if (all_vars) {
    resp_vars <- md100[[label_col]] # TODO: Sort by VAR_ORDER?
  } else {
    resp_vars_m <- util_find_var_by_meta(
      resp_vars = resp_vars,
      meta_data = md100,
      label_col = label_col,
      # allowed_sources = ,
      target = label_col,
      ifnotfound = NA_character_)
    if (any(is.na(resp_vars_m))) {
      util_warning(
        c("Could not find the following variables in %s:",
          "%s.\nThese will be removed."),
        sQuote("meta_data"),
        paste0(
          dQuote(resp_vars[is.na(resp_vars_m)]),
          collapse = ", ")
      )
    }
    resp_vars <- resp_vars_m[!is.na(resp_vars_m)]
  }

  to_remove <-
    intersect(resp_vars, names(which(only_nas)))

  orig_ignore_empty_vars <- ignore_empty_vars

  if (length(resp_vars) == 0) {
    util_error("No response variables left.")
  }

  if (ignore_empty_vars == "auto") {
    ignore_empty_vars <- length(to_remove) / length(resp_vars) > .2 # 20%, see options for documentation, search for ignore_empty_vars20, if changing
  } else {
    ignore_empty_vars <- as.logical(ignore_empty_vars)
  }

  if (ignore_empty_vars) {
    util_warning(
      c("%s was %s, so removing the following variables from the report,",
        "because they only feature empty values: %s"),
      sQuote("ignore_empty_vars"),
      sQuote(orig_ignore_empty_vars),
      util_pretty_vector_string(to_remove, n_max = 6), immediate = TRUE)
    resp_vars <- setdiff(resp_vars, to_remove)
  } else if (length(to_remove) > 0) {
    util_message("Have variables only featuring empty data values: %s",
                 util_pretty_vector_string(to_remove, n_max = 6),
                 immediate = TRUE)
  }

  util_message("Pre-computing curated study data frames...")

  util_reset_cache()
  if (getOption("dataquieR.precomputeStudyData", default =
                dataquieR.precomputeStudyData_default)) {
    util_populate_study_data_cache(study_data, meta_data, label_col = LABEL)
  } else {
    util_purge_study_data_cache()
  }

  util_message("Pre-computing curated study data frames... done")

  util_expect_scalar(filter_indicator_functions,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)
  util_expect_scalar(exclude_indicator_functions,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)
  util_expect_scalar(filter_result_slots,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)

  scale_level_predicted <- FALSE

  if (!(SCALE_LEVEL %in% colnames(meta_data)) ||
      any(util_empty(meta_data[[SCALE_LEVEL]]))) {
    util_message("Estimating %s...", sQuote(SCALE_LEVEL))
    scale_level_predicted <- TRUE
    function_e <- environment()
    local(suppressWarnings(suppressMessages({
      prep_prepare_dataframes(.replace_hard_limits = FALSE,
                              .replace_missings = FALSE,
                              .adjust_data_type = TRUE,
                              .amend_scale_level = TRUE,
                              .meta_data = meta_data)
      vec_sl <- setNames(meta_data[[SCALE_LEVEL]],
                         nm = meta_data[[VAR_NAMES]])
      function_e$meta_data[[SCALE_LEVEL]] <-
        vec_sl[function_e$meta_data[[VAR_NAMES]]]
    })))
    util_message("Estimating %s... done", sQuote(SCALE_LEVEL))
  }

  all_calls <- util_generate_calls(dimensions = dimensions,
                                   meta_data = meta_data,
                                   label_col = label_col,
                                   meta_data_segment = meta_data_segment,
                                   meta_data_dataframe = meta_data_dataframe,
                                   meta_data_cross_item = meta_data_cross_item,
                                   specific_args = specific_args,
                                   arg_overrides = list(...),
                                   filter_indicator_functions =
                                     filter_indicator_functions,
                                   exclude_indicator_functions =
                                     exclude_indicator_functions,
                                   resp_vars = resp_vars)

  tm <- system.time(
    r <- util_evaluate_calls(
      cores = cores,
      all_calls = all_calls,
      study_data = study_data,
      meta_data = meta_data,
      label_col = label_col,
      meta_data_segment = meta_data_segment,
      meta_data_dataframe = meta_data_dataframe,
      meta_data_cross_item = meta_data_cross_item,
      debug_parallel = debug_parallel,
      resp_vars = resp_vars,
      filter_result_slots = filter_result_slots,
      mode = mode,
      mode_args = mode_args,
      my_storr_object = my_storr_object,
      checkpoint_resumed = checkpoint_resumed,
      dt_adjust = dt_adjust,
      content_file = content_file
    )
  )

  start_from_call <- util_find_first_externally_called_functions_in_stacktrace()
  start_from_call <- length(sys.calls()) - start_from_call # refers to reverted sys.calls, so mirror the number
  if (is.na(start_from_call))
    start_from_call <- 1
  cl <- NULL
  try({
    cl <- sys.call(start_from_call)
  }, silent = TRUE)

  # get call for dq_report_by
  get_call <- try(deparse(sys.call(-10)), silent= TRUE)
  if (!inherits(get_call , "try-error")) {
    if (startsWith(paste(deparse(sys.call(-10)), collapse = ""),
                   "dq_report_by")) {
      cl <- eval.parent(quote({call_report_by}))
    }
  }

  p <- list( # TODO: add this also in Square2
    author = author,
    date = Sys.time(),
    call = cl, # TODO: Why does this not yet work?
    version = paste(packageName(), util_dataquieR_version()),
    R = R.version.string,
    os = osVersion,
    machine = paste(Sys.info()[["nodename"]],
                    sprintf("(%s)", Sys.info()[["version"]]),
                    Sys.info()["machine"]
    ),
    runtime = paste(round(tm[["elapsed"]], 1), "secs")
  )

  if (is.list(user_info)) {
    p <- c(user_info, p)
  }

  dq_report2_env <- environment()
  meta_data_hints <- list()
  capture <- function(cnd) {
    dq_report2_env$meta_data_hints <-
      c(dq_report2_env$meta_data_hints, list(cnd))
    if (inherits(cnd, "warning")) {
      invokeRestart("muffleWarning")
    }
    if (inherits(cnd, "message")) {
      invokeRestart("muffleMessage")
    }
  }

  # TODO: Do we need this if we run prep_prepare_dataframes above?
  suppressWarnings(suppressMessages(
    try(withCallingHandlers(meta_data <- util_validate_known_meta(meta_data),
                           error = capture,
                           warning = capture,
                           message = capture), silent = TRUE)))


  if (scale_level_predicted) { # TODO: sort all the hints and have only one metadata integrity hints object?
    suppressWarnings(suppressMessages(
      try(withCallingHandlers(util_message(c(
        "Missing some or all entries in %s column in item-level %s. Predicting",
                   "it from the data -- please verify these predictions, they",
                   "may be wrong and lead to functions claiming not to be",
                   "reasonably applicable to a variable."),
                 sQuote(SCALE_LEVEL), "meta_data",
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = FALSE),
                 error = capture,
                 warning = capture,
                 message = capture), silent = TRUE)))
  }

  if (util_really_rstudio()) {
    rstudioapi::executeCommand("activateConsole")
  }

  if (!is.null(my_storr_object) &&
      inherits(my_storr_object, "storr") &&
      !util_is_try_error(try(my_storr_object$list(), silent = TRUE))) {
    # to make summary work
    atts_r <- attributes(r)
    atts_r[["my_storr_object"]] <- NULL # dont save this ever
    my_storr_object$mset(key = names(atts_r), value = atts_r, namespace =
                           util_get_storr_att_namespace(my_storr_object))
  }

  #  repsum <- summary(r) FIXME: Dead slow!!
  repsum <- NA

  p$rep_id <- rep_id

  r <- util_attach_attr(r,
                   properties = p,
                   min_render_version = as.numeric_version("1.0.0"),
                   translation_version = translation_version,
                   warning_pred_meta = warning_pred_meta,
                   label_modification_text = trimws(paste(
                     notes_from_wrapper[["label_modification_text"]],
                     mod_label$label_modification_text)),
                   label_modification_table = rbind(
                     notes_from_wrapper[["label_modification_table"]],
                     mod_label$label_modification_table),
                   label_meta_data_hints = meta_data_hints,
                   meta_data_item_computation = meta_data_item_computation,
                   repsum = repsum,
                   title = title,
                   subtitle = subtitle)

  if (!is.null(my_storr_object) &&
      inherits(my_storr_object, "storr") &&
      !util_is_try_error(try(my_storr_object$list(), silent = TRUE))) {

    # to make summary work
    atts_r <- attributes(r)
    atts_r[["my_storr_object"]] <- NULL # dont save this ever
    my_storr_object$mset(key = names(atts_r), value = atts_r, namespace =
                           util_get_storr_att_namespace(my_storr_object))

#    my_storr_object$mset(key = names(r), value = r)

    attr(r, "my_storr_object") <- my_storr_object

  }

  if (!missing(output_dir)) {
    print.dataquieR_resultset2(r, dir = output_dir,
                               view = FALSE, force_overwrite = TRUE)
    return(invisible(r))
  } else {
    return(r)
  }

}

.study_data_cache <- new.env(parent = emptyenv())
util_purge_study_data_cache <- function() {
  rm(list = ls(.study_data_cache), envir = .study_data_cache)
}

# util_populate_study_data_cache(prep_get_data_frame("study_data"), prep_get_data_frame("meta_data"), LABEL)
#
#
# options(dataquieR.study_data_cache_quick_fill = FALSE)
#
# options(dataquieR.study_data_cache_quick_fill = TRUE)
# options(dataquieR.study_data_cache_metrics = TRUE)
# metrics <- new.env(parent = emptyenv())
# options(dataquieR.study_data_cache_metrics_env = metrics)
# rx <- dq_report2("study_data", meta_data_v2 = "meta_data_v2", dimensions = NULL, cores = NULL)
# to_populate <- lapply(names(metrics$usage), function(key) attr(dataquieR:::.study_data_cache[[key]], "call"))
# cat(unlist(lapply(to_populate, deparse, width.cutoff = 300)), sep = "\n", file = "|pbcopy") # but add ".study_data = study_data, .meta_data = meta_data, .label_col = label_col," everywhere, if missing
# options(dataquieR.study_data_cache_quick_fill = TRUE)
# options(dataquieR.study_data_cache_metrics = FALSE)
# options(dataquieR.study_data_cache_metrics_env = NULL)
# rm(metrics)
# rx <- dq_report2("study_data", meta_data_v2 = "meta_data_v2", dimensions = NULL)
util_populate_study_data_cache <- function(study_data, meta_data, label_col, quick = getOption("dataquieR.study_data_cache_quick_fill", dataquieR.study_data_cache_quick_fill_default)) {
  util_purge_study_data_cache()
  invisible(lapply(study_data, util_is_na_0_empty_or_false)) # TODO: maybe RAM intensive?
  if (quick) {
    try(silent = TRUE, prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data, .label_col = label_col, .replace_hard_limits = TRUE, .replace_missings = TRUE, .adjust_data_type = TRUE, .amend_scale_level = TRUE))
    try(silent = TRUE, prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data, .label_col = label_col, .replace_missings = FALSE))
    try(silent = TRUE, prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data, .label_col = label_col, .replace_missings = FALSE))
    try(silent = TRUE, prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data, .label_col = label_col))
    try(silent = TRUE, prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data, .label_col = label_col, .allow_empty = TRUE))
    try(silent = TRUE, prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data, .label_col = label_col, .replace_hard_limits = TRUE))
    try(silent = TRUE, prep_prepare_dataframes(.study_data = study_data, .meta_data = meta_data, .label_col = label_col, .replace_missings = FALSE, .adjust_data_type = FALSE))
  } else {
    combinations <- unlist(lapply(c(0, seq_along(.to_combine)),
                                  function(x) utils::combn(.to_combine, x,simplify = FALSE)), recursive = FALSE)
    calls <- lapply(combinations, function(set_true) {
      res <- .call_template
      res[set_true] <- TRUE
      res
    })
    study_data_hash <- rlang::hash(study_data)
    meta_data_hash <- rlang::hash(meta_data)
    a <- lapply(lapply(lapply(calls, rlang::call_args), `[`, .to_combine),
                rlang::hash)
    names(calls) <-
      paste0(a, "@", study_data_hash,
             "@", meta_data_hash, "@", label_col)

    my_env <- new.env(parent = parent.env(environment()))
    my_env$study_data <- study_data
    my_env$meta_data <- meta_data
    my_env$label_col <- label_col
    # TODO: Remove unneded, rarely used, calls

    # HINT: Do not touch my_env (use new.env(parent = my_env)) to
    #       keep the original study data for all cases.

    # parallelMap::parallelLibrary("dataquieR")
    # parallelMap::parallelExport("study_data", "meta_data", "label_col")
    # for_cache <- parallelMap::parallelLapply(calls,
    #                                          function(cl, my_env) {
    #                                            try(eval(cl, envir = new.env(parent = my_env)),
    #                                                silent = TRUE)
    #                                           },
    #                                          my_env = my_env)
    for_cache <- lapply(calls,
                        function(cl, my_env) { try(util_attach_attr(eval(cl, envir = new.env(parent = my_env)), call = cl),
                                                   silent = TRUE) },
                        my_env = my_env)
    for_cache <- for_cache[vapply(for_cache, is.data.frame,
                                  FUN.VALUE = logical(1))]
    list2env(for_cache, .study_data_cache)
  }
}

.call_template <- quote(prep_prepare_dataframes(
  .study_data = study_data,
  .meta_data = meta_data,
  .label_col = label_col,
  .replace_hard_limits = FALSE,
  .replace_missings = FALSE,
  .adjust_data_type = FALSE,
  .amend_scale_level = FALSE,
  .apply_factor_metadata = FALSE,
  .apply_factor_metadata_inadm = FALSE
))
.to_combine <- setdiff(names(.call_template),
                      c("", ".study_data", ".meta_data", ".label_col"))

#' Support function to augment metadata during data quality reporting
#'
#' @description
#' adds an annotation to static metadata
#'
#' @details
#'
#' Add metadata e.g. of transformed/new variable
#' This function is not yet considered stable, but we already export it,
#' because it could help. Therefore, we have some inconsistencies in the
#' formals still.
#'
#'
#'
#' @param VAR_NAMES [character] Names of the Variables to add
#' @param DATA_TYPE [character] Data type for the added variables
#' @param LABEL [character] Labels for these variables
#' @param VALUE_LABELS [character] Value labels for the values of the variables
#'                                 as usually pipe separated and assigned with
#'                                 `=`: `1 = male | 2 = female`
#' @param ... Further defined variable attributes, see
#'            [dataquieR::prep_create_meta]
#' @param item_level [data.frame] the metadata to extend
#' @param meta_data [data.frame] old name for `item_level`
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return a data frame with amended metadata.
#'
#' @export
#'
prep_add_to_meta <- function(VAR_NAMES, DATA_TYPE, LABEL, VALUE_LABELS,
                             item_level = "item_level",
                             meta_data = item_level,
                             meta_data_v2,
                             ...) {

  util_maybe_load_meta_data_v2()

  util_expect_data_frame(meta_data)

  mini_md <- prep_create_meta(
    level = NULL,
    "VAR_NAMES" = VAR_NAMES,
    "DATA_TYPE" = DATA_TYPE,
    "LABEL" = LABEL,
    "VALUE_LABELS" = VALUE_LABELS,
    ...
  )

  mini_md <- mini_md[, names(mini_md) %in% names(meta_data)]
  new_names <- c(names(mini_md), names(meta_data)[!(names(meta_data) %in%
                                                      names(mini_md))])
  # new_names <- unique(c(names(mini_md), names(meta_data)))

  mini_md[, (dim(mini_md)[2] + 1):dim(meta_data)[2]] <- NA

  colnames(mini_md) <- new_names

  partvar_vector <-
    mini_md[[PART_VAR]]

  partvar_vector <-
    partvar_vector[!util_empty(partvar_vector)]

  if (PART_VAR %in% names(mini_md) &&
      any(is.na(util_find_var_by_meta(resp_vars = as.character(
        partvar_vector))))) {
    miss <-
      partvar_vector[
        is.na(util_find_var_by_meta(resp_vars = as.character(
          partvar_vector)))]
    util_error(c("In the existing %s, in the column %s, at least one of the",
                  "referred variables does not exist: %s."),
               sQuote("meta_data"),
               dQuote(PART_VAR),
               util_pretty_vector_string(miss))
  }

  meta_data <- dplyr::bind_rows(meta_data, mini_md)
  # Column order is auto-repaired for 2 DFs in R >= 3.3.0 by rbind too,
  # but I did not find an offical source for that. Only:
  # https://stackoverflow.com/q/16962576

  return(meta_data)
}

#' utility function to set string in backticks
#'
#' Quote a set of variable names with backticks
#'
#' @param x variable names
#'
#' @return quoted variable names
#'
#' @seealso `util_bQuote`
#' @family process_functions
#' @concept data_management
#' @noRd
util_backtickQuote <- function(x) {
  na <- is.na(x)
  res <- paste0("`", x, "`")
  res[res == "``"] <- ""
  res[na] <- NA
  res
}

test_that("util_backtickQuote works", {
  skip_on_cran()
  expect_equal(util_backtickQuote(""), "")
  expect_equal(ignore_attr = TRUE, util_backtickQuote(NA), NA_character_)
  nm <- paste(letters, LETTERS)
  t <- paste("c(", paste(util_backtickQuote(nm), collapse = ", "), ")")
  e <- parse(text = t)
  got <- eval(e, envir = as.list(setNames(LETTERS, nm = nm)))
  expect_equal(ignore_attr = TRUE, got, LETTERS)
})

#' Cache for materialized lazy `ggplot`s
#'
#' This environment stores materialized ggplot objects referenced by
#' \code{dq_lazy_ggplot} IDs. It is intentionally not exported and not
#' serialized with result objects.
#'
#' @noRd
.dq_lazy_ggplot_cache <- new.env(parent = emptyenv())
# Counter in favor of RNG no side effects on .Rand.seed; cache only in process.

#' Generate a new ID for a lazy ggplot
#'
#' @noRd
util_lazy_ggplot_next_id <- local({ # nolint
  # Counter for assigning IDs to lazy ggplots
  #
  # We do not use the global RNG state here to avoid side effects on
  # reproducible workflows and tests.
  counter <- 0L
  e <- environment()

  function() {
    e$counter <- e$counter + 1L
    sprintf("dq_lazy_%08x", counter)
  }
})

#' Construct a lazy ggplot object
#'
#' @param expr expression which, when evaluated in \code{env}, returns
#'   a \code{ggplot} object.
#' @param env environment holding all required objects for \code{expr}.
#' @param id optional character ID; if \code{NULL}, a new ID is generated.
#'
#' @return An object of class \code{c("dq_lazy_ggplot", "gg", "ggplot")}.
#'
#' @noRd
dq_lazy_ggplot <- function(expr, env, id = NULL) {

  s3 <- dq_lazy_ggplot_s3(expr = expr, env = env, id = id)

  gg_compatible <-
    as.logical(getOption("dataquieR.lazy_plots_gg_compatibility",
                         dataquieR.lazy_plots_gg_compatibility_default))

  # TODO: make all this S7 magic optional
  if (gg_compatible && dq_lazy_register_s7()) {
    obj <- .dq_lazy_state$s7_class(payload = s3)

    # add an S3 class tag for dispatch (must be a valid S3 class name)
    class(obj) <- c("dq_lazy_ggplot_s7", class(obj))

    return(obj)
  }

  s3
}

dq_lazy_ggplot_s3 <- function(expr, env, id = NULL) {
  stopifnot(is.environment(env))

  if (is.null(id)) {
    id <- util_lazy_ggplot_next_id()
  }

  .dq_lazy_ggplot_cache[[id]] <- NULL

  structure(
    list(
      id   = id,
      expr = expr,
      env  = env
    ),
    class = c("dq_lazy_ggplot")
  )
}
#' Materialize a lazy `ggplot`
#'
#' Evaluate the stored expression in its lean environment and cache
#' the resulting \code{ggplot} object in the current R session, if enabled
#' using the option [dataquieR.lazy_plots_cache].
#'
#' @param x a \code{dq_lazy_ggplot} object.
#'
#' @return A \code{ggplot} object.
#'
#' @export
prep_realize_ggplot <- function(x) {

  x <- dq_lazy_unwrap(x)

  if (!inherits(x, "dq_lazy_ggplot")) {
    return(x)
  }

  x_un <- unclass(x)

  id <- x_un$id
  cached <- .dq_lazy_ggplot_cache[[id]]
  if (!is.null(cached)) {
    return(cached)
  }

  p <- eval(x_un$expr, envir = x_un$env)

  # if (!inherits(p, "ggplot")) {
  #   util_error("dq_lazy_ggplot: expr did not evaluate to a ggplot object, but %s",
  #              util_pretty_vector_string(class(p)))
  # }

  lazy_cache <- as.logical(getOption("dataquieR.lazy_plots_cache",
                               dataquieR.lazy_plots_cache_default))
  if (length(lazy_cache) != 1 || is.na(lazy_cache)) {
    util_warning(c(
      "Cannot use option dataquieR.lazy_plots_cache %s as a logical value",
      "using %s"
    ),
    dQuote(paste(getOption("dataquieR.lazy_plots_cache",
                           dataquieR.lazy_plots_cache_default),
                 collapse = ",")),
    dQuote(dataquieR.lazy_plots_cache_default))
    lazy_cache <- as.logical(dataquieR.lazy_plots_cache_default)
  }

  if (lazy_cache) .dq_lazy_ggplot_cache[[id]] <- p
  p
}

#' Forget all materialized lazy ggplots
#'
#' This clears the in-session cache of materialized plots. It does not
#' affect any stored lazy plot objects themselves.
#'
#' @noRd
util_forget_lazy_ggplots <- function() { # nolint
  rm(list = ls(.dq_lazy_ggplot_cache), envir = .dq_lazy_ggplot_cache)
  invisible(NULL)
}

#' @rdname dq_lazy_ggplot_methods
#' @export
ggplotly.dq_lazy_ggplot_s7 <- function(p, ...) { # nolint
  util_ggplotly(p@payload, ...)
}

#' @rdname dq_lazy_ggplot_methods
#' @export
plotly_build.dq_lazy_ggplot_s7 <- function(p, ...) { # nolint
  util_plotly_build(p@payload, ...)
}

#' @rdname dq_lazy_ggplot_methods
#' @export
ggplotly.dq_lazy_ggplot <- function(p, ...) { # nolint
  util_ggplotly(p, ...)
}

#' @rdname dq_lazy_ggplot_methods
#' @export
plotly_build.dq_lazy_ggplot <- function(p, ...) { # nolint
  util_plotly_build(p, ...)
}

# patchwork ist das einzige Downstream-Paket mit relevanten gg-/ggplot-Generics; cowplot/gridExtra/egg/ggpubr hängen an ggplotGrob()
#' @export
`-.dq_lazy_ggplot` <- function(e1, e2) { # nolint
  p1 <- prep_realize_ggplot(e1)
  if (inherits(e2, "dq_lazy_ggplot")) {
    e2 <- prep_realize_ggplot(e2)
  }
  p1 - e2
}

#' @export
`/.dq_lazy_ggplot` <- function(e1, e2) { # nolint
  p1 <- prep_realize_ggplot(e1)
  if (inherits(e2, "dq_lazy_ggplot")) {
    e2 <- prep_realize_ggplot(e2)
  }
  p1 / e2
}

#' @export
`|.dq_lazy_ggplot` <- function(e1, e2) { # nolint
  p1 <- prep_realize_ggplot(e1)
  if (inherits(e2, "dq_lazy_ggplot")) {
    e2 <- prep_realize_ggplot(e2)
  }
  p1 | e2
}

#' @export
`*.dq_lazy_ggplot` <- function(e1, e2) { # nolint
  p1 <- prep_realize_ggplot(e1)
  if (inherits(e2, "dq_lazy_ggplot")) {
    e2 <- prep_realize_ggplot(e2)
  }
  p1 * e2
}

#' @export
`&.dq_lazy_ggplot` <- function(e1, e2) { # nolint
  p1 <- prep_realize_ggplot(e1)
  if (inherits(e2, "dq_lazy_ggplot")) {
    e2 <- prep_realize_ggplot(e2)
  }
  p1 & e2
}

#' Realize lazy `ggplot`s only when needed
#'
#' @noRd
util_realize_if_lazy <- function(x) { # nolint
  if (inherits(x, "dq_lazy_ggplot")) {
    prep_realize_ggplot(x)
  } else {
    x
  }
}


#' S3/S7 methods for lazy ggplot objects
#'
#' These S3/S7 methods make \code{dq_lazy_ggplot}/\code{dq_lazy_ggplot_s7}
#' objects work smoothly with
#' functions from \pkg{ggplot2} and \pkg{plotly}. They simply materialize
#' the underlying \code{ggplot} object and then delegate to the respective
#' generic.
#'
#' @param x,p A \code{dq_lazy_ggplot} object.
#' @param ... Further arguments passed on to the underlying generic.
#'
#' @return
#' The return value is the same as for the corresponding generic:
#' \itemize{
#'   \item \code{ggplotGrob()} returns a \code{gtable} object.
#'   \item \code{ggplotly()} returns a \code{plotly} object.
#'   \item \code{plotly_build()} returns a \code{plotly_proxy} or similar.
#' }
#'
#' @seealso \code{\link[ggplot2]{ggplotGrob}},
#'   \code{plotly::ggplotly}
#'   \code{plotly::plotly_build}
#'
#' @name dq_lazy_ggplot_methods
NULL

#' @export
`$.dq_lazy_ggplot_s7` <- function(x, name) { # nolint
  `$.dq_lazy_ggplot`(x@payload, name)
}

#' @export
`[[.dq_lazy_ggplot_s7` <- function(x, ...) { # nolint
  `[[.dq_lazy_ggplot`(x@payload, ...)
}

#' @export
ggplot_build.dq_lazy_ggplot_s7 <- function(plot, ...) { # nolint
  ggplot2::ggplot_build(prep_realize_ggplot(plot@payload), ...)
}

#' @export
print.dq_lazy_ggplot_s7 <- function(x, ...) { # nolint
  print(prep_realize_ggplot(x@payload), ...)
}

#' @rdname dq_lazy_ggplot_methods
#' @export
ggplotGrob.dq_lazy_ggplot_s7 <- function(x, ...) { # nolint
  ggplot2::ggplotGrob(prep_realize_ggplot(x@payload), ...)
}

#' @export
util_undisclose.dq_lazy_ggplot_s7 <- function(x, ...) { # nolint
  util_undisclose(x@payload, ...)
}

#' Create a table summarizing the number of indicators and descriptors
#' in the report
#'
#' @param report a report
#'
#' @return a table containing the number of indicators and descriptors created
#' in the report, separated by data quality dimension.
#'
#' @noRd
util_generate_table_indicators_descriptors <- function(report) {

  util_stop_if_not(inherits(report, "dataquieR_resultset2"))

  #Obtain the names of the group (e.g., observer, device...)
  tab_attrib <- attributes(attributes(report)$matrix_list)$function_alias_map

  alias2functionname <- setNames(tab_attrib$name, nm = tab_attrib$alias)

  #create the summary
  x <- summary(report)

  # this represents the current object
  this <- attr(x, "this")
  withr::with_environment(this, {

  # For robustness: in case none of the assessments work, create an empty result
  # if result is empty or contains only errors
  if (!prod(dim(this$result)) || all(
    startsWith(as.character(this$result$indicator_metric), "CAT_") |
    startsWith(as.character(this$result$indicator_metric), "MSG_"))) {
    # create a table containing only zeros
    info_dim_dq <- data.frame(Dimension = c("Accuracy", "Completeness",
                                            "Consistency", "Integrity"),
                              `No. DQ indicators` = c(0, 0, 0, 0),
                              `No. DQ descriptors` = c(0, 0, 0, 0),
                              check.names = FALSE)
  } else {

    ## import the complete list of functions run or stopped
    stopped_functions <- this$stopped_functions
    # modify from named vector to data frame
    stopped_functions <- t(data.frame(as.list(stopped_functions)))
    stopped_functions <- as.data.frame(stopped_functions)
    stopped_functions  <- cbind(rownames(stopped_functions),
                                 data.frame(stopped_functions ,
                                            row.names=NULL))
    colnames(stopped_functions) <- c("fun_names", "stopped_yn")
    # remove the suffixes(.var_names, .[ALL]) from fun names
    stopped_functions$fun_names1 <- gsub("(^[^.]*).*", "\\1",
                                          stopped_functions$fun_names)

    # translate call-names/aliases to function names,
    # i.e., remove all possible group-suffixes (_device, _observer...)
    stopped_functions$fun_names1 <-
      alias2functionname[`stopped_functions`$fun_names1]

    stopped_functions <- unique(stopped_functions[, c("fun_names1",
                                                "stopped_yn"),
                                            drop = FALSE ])
    # create list TRUE (stopped functions)
    stopped_functions_T <-
      stopped_functions[stopped_functions$stopped_yn ==TRUE, , drop=FALSE]
    # create list FALSE (run functions)
    stopped_functions_F <-
      stopped_functions[stopped_functions$stopped_yn == FALSE, , drop=FALSE]
    # create a vector with only TRUE - only fun not run
    fun_stopped <- setdiff(stopped_functions_T$fun_names1,
             stopped_functions_F$fun_names1) # VECTOR of stopped functions
    # the script works even if fun_stopped is empty
    rm(stopped_functions,
       stopped_functions_T, stopped_functions_F)


    ## import results
    result_t <- this$result


    ## Create a list of function names
    result_novars<- result_t[, c("call_names", "function_name"), drop = FALSE]
    # create the function name that corresponds to the actual call_names,
    # e.g., with con_hard_limits instead of con_limit_deviations
    result_novars$function_name2 <-
     apply(result_novars[, "call_names",
                    drop = FALSE],
           1, FUN = function(x) {
             util_map_by_largest_prefix(x,
                                        haystack = util_all_ind_functions())
           })
    # Create a unique list of function names here before to remove information
    # to use that with alias to see which function was not included in result_t
    functions_in_result_t<- unique(result_novars$function_name2)
    rm(result_novars)


    ## Define which indicator_metric is an indicator based on CAT_indicator_or_descriptor
    # select everything starting with CAT (Error category)
    result_err_cat <-
      result_t[startsWith(as.character(result_t$indicator_metric),
                          "CAT_"),
               c(VAR_NAMES, "call_names",
                 "indicator_metric",
                 "class"),
               drop = FALSE]
    # rename the class for CAT errors as "err_cat"
#    result_err_cat <- dplyr::rename(result_err_cat,
#                                    err_cat = "class")
    names(result_err_cat)[names(result_err_cat) == "class"] <- "err_cat"

    # select only the category "CAT_indicator_or_descriptor"
 #   result_err_cat <-
 #          dplyr::filter(result_err_cat,
 #                        get("indicator_metric") ==
 #                          "CAT_indicator_or_descriptor")
    result_err_cat <- result_err_cat[result_err_cat$indicator_metric ==
                                       "CAT_indicator_or_descriptor", ,
                                     drop = FALSE ]

    # no need to reshape from long to wide, only CAT_indicator_or_descriptor present
 #   result_err_cat <- suppressMessages(reshape::cast(result_err_cat,
#                                                     VAR_NAMES + call_names ~
#                                                       indicator_metric))


    # create a new column "is_indicator" logic T/F T = is an indicator
    result_err_cat$is_indicator <-
         result_err_cat$err_cat == 1
    # reduce no. columns
    result_err_cat <- result_err_cat[, c(VAR_NAMES, "call_names",
                                        "is_indicator"),drop = FALSE]
    # remove rownames
    rownames(result_err_cat) <- NULL



    ## Remove all error messages (CAT and MSG) and obtain an object with actual
    # results (metrics) only
    result_t <- result_t[!startsWith(as.character(result_t$indicator_metric),
                                    "CAT_"), , drop = FALSE]
    result_t <- result_t[!startsWith(as.character(result_t$indicator_metric),
                                    "MSG_"), , drop = FALSE]

    # merge the 2 objects: result_t and result_err_cat
    result_t <- merge(result_t,
                      result_err_cat,
                      by = c(VAR_NAMES, "call_names"),
                      all = TRUE)

    rm(result_err_cat)

    # remove empty indicator_metric, where only an error message was created
    result_t <- result_t[!is.na(result_t$indicator_metric),]

    # remove indicators that have no class (class = NA)
    # result_t <- result_t[!is.na(result_t$class),]
    # remove indicators that have no value (value = NA)
    result_t <- result_t[!is.na(result_t$value),]

    # reduce no. columns to only needed
    result_t <- result_t[, c(VAR_NAMES, "indicator_metric", "is_indicator"), drop = FALSE]


    ## calculate the no. of indicators from the result
    # keep only first row per indicator_metric+var_names,
    # in case of the same indicator metric present as both indicator and
    # descriptor, only indicator is kept
    result_out <-
      dplyr::filter(dplyr::arrange(dplyr::group_by(result_t,
                                                   result_t["indicator_metric"]),
                                   dplyr::desc(get("is_indicator"))),
                    dplyr::row_number()==1 )

    # remove the prefix to indicators
    result_out$indicator_metric <- gsub("(.*?_)(.*)","\\2",
                                        vapply(result_out$indicator_metric,
                                               as.character,
                                               FUN.VALUE = character(1)))

    # remove duplicates by indicators
    result_out <- result_out[, c("indicator_metric", "is_indicator")]
    result_out <- result_out[!duplicated(result_out$indicator_metric), ]

    # create a column with the DQ dimension of the indicator metric
    result_out$Dimension <- gsub("(.*?)_.*","\\1",
                               vapply(result_out$indicator_metric,
                                      as.character,
                                      FUN.VALUE = character(1)))

    # calculate the number of indicators and descriptors per dimension, and
    # rename the DQ dimensions
    info_dim_dq <-
      as.data.frame(dplyr::summarize(dplyr::group_by(result_out,
                                                     result_out["Dimension"],
                                                     result_out["is_indicator"]),
                                     n=dplyr::n()))
    rm(result_out)

    # remove eventual empty rows
    info_dim_dq <- info_dim_dq[!is.na(info_dim_dq$Dimension), , drop = FALSE]

    # rename the T/F with speaking labels
    info_dim_dq$is_indicator[info_dim_dq$is_indicator == "FALSE"] <-
      "descriptors"

    info_dim_dq$is_indicator[info_dim_dq$is_indicator == "TRUE"] <-
      "No_indicators"

    # change from long to wide format
 #   info_dim_dq <- suppressMessages(reshape::cast(info_dim_dq,
 #                                                 Dimension ~ is_indicator))
    info_dim_dq <- stats::reshape(info_dim_dq, idvar = "Dimension",
                                 timevar = "is_indicator",
                                 direction = "wide")

    colnames(info_dim_dq) <- gsub("^n\\.",
                                  "",
                                  colnames(info_dim_dq))



    # replace NAs with 0
    info_dim_dq[is.na(info_dim_dq)] <- 0


    # Create empty columns in case there are no descriptors or no indicators
    if(!"descriptors" %in% colnames(info_dim_dq)) {
      info_dim_dq[, "descriptors"] = 0
    } else if (!"No_indicators" %in% colnames(info_dim_dq)) {
      info_dim_dq[, "No_indicators"] = 0
    }

    ## calculate the no. of descriptors from the function_name
    alias <- unique(this$alias_names)
    # remove from alias functions that were stopped
    alias <- setdiff(alias, fun_stopped)
    # remove from alias functions that created an error comparing it with all
    # functions that created an indicator_metric
    alias <- as.data.frame(setdiff(alias, functions_in_result_t))
    colnames(alias) <- "function_name_descriptors"

    rm(functions_in_result_t, fun_stopped)

    # create a column with the DQ dimension
    alias$Dimension <- gsub("(.*?)_.*","\\1",
                               vapply(alias$function_name_descriptors,
                                      as.character,
                                      FUN.VALUE = character(1)))
    # calculate the no. descriptors per dimension
    n_descriptors <-
      as.data.frame(dplyr::summarize(dplyr::group_by(alias,
                                                     alias["Dimension"]),
                                     n_descr = dplyr::n()))

    # replace NAs with 0
    n_descriptors[is.na(n_descriptors)] <- 0

    rm(result_t, alias)

    n_descriptors <-
      n_descriptors[!is.na(n_descriptors$Dimension) &
                    n_descriptors$Dimension != "des", ]

    # merge the table with the no. indicators and the other with no. descriptors
    info_dim_dq <- merge(info_dim_dq, n_descriptors, by = "Dimension",
                         all = TRUE)
    rm(n_descriptors)

    #In case some dimensions are not present in n_descr, replace NA with 0
    info_dim_dq$n_descr[is.na(info_dim_dq$n_descr)] <- 0

    # sum the no. of descriptors of the 2 tables
    info_dim_dq <- dplyr::mutate(info_dim_dq,
                                 descr_num = get("descriptors")+
                                   get("n_descr"))

    # clean and rename table columns and content
    info_dim_dq <- info_dim_dq[, c("Dimension",
                                   "No_indicators", "descr_num"),
                               drop = FALSE]

    #re-order columns
    colnames(info_dim_dq) <- c("Dimension", "No. DQ indicators",
                               "No. DQ descriptors" )
    col_order <- c("Dimension", "No. DQ indicators",
                   "No. DQ descriptors")
    info_dim_dq <- info_dim_dq[, col_order, drop = FALSE]

    # amend missing dimensions:
    for (dim in c("int", "com", "con", "acc")) {
      if (!dim %in% info_dim_dq$Dimension) {
        info_dim_dq <- rbind(info_dim_dq,
                             data.frame(Dimension = dim,
                                        `No. DQ indicators` = 0,
                                        `No. DQ descriptors` = 0,
                                        stringsAsFactors = FALSE,
                                        check.names = FALSE
                             ))
      }
    }

    info_dim_dq$Dimension[info_dim_dq$Dimension == "acc"] <- "Accuracy"
    info_dim_dq$Dimension[info_dim_dq$Dimension == "com"] <- "Completeness"
    info_dim_dq$Dimension[info_dim_dq$Dimension == "con"] <- "Consistency"
    info_dim_dq$Dimension[info_dim_dq$Dimension == "int"] <- "Integrity"

    info_dim_dq <- info_dim_dq[match(c("Integrity", "Completeness",
                                       "Consistency", "Accuracy"),
                                     info_dim_dq$Dimension),]

    rm(col_order)
    gc()

  }
    info_dim_dq
  })
}

#' Fix encoding errors
#'
#' @param x [character] a vector, possibly with encoding errors
#'
#' @returns `x`, but with all invalidly encoded characters deleted
#' @noRd
util_fix_encoding <- function(x) {
  orig <- x
  from <- Encoding(x)
  from <- setdiff(from, "unknown")
  from <- head(names(which.max(table(from))), 1)
  if (length(from) != 1) {
    from <- "UTF-8"
  }
  x <- iconv(x, from, "UTF-8", sub = '')
  Encoding(x) <- "UTF-8"
  if (any((is.na(x) != is.na(orig)) | x != orig, na.rm = TRUE)) {
    in_pipe <- FALSE # careful, runs early
    tmp <- try(.called_in_pipeline2(), silent = TRUE)
    if (!util_is_try_error(tmp)) in_pipe <- isTRUE(tmp)
    if (!in_pipe && exists(".called_in_pipeline", .dq2_globs,
                           mode = "logical")) {
      in_pipe <- get(".called_in_pipeline", .dq2_globs, mode = "logical")
    }
    if (length(in_pipe) != 1 || is.na(in_pipe)) in_pipe <- FALSE
    if (!in_pipe)
      util_warning("Found encoding errors, deleted invalid characters")
  }
  return(x)
}

#' Fix encoding errors
#'
#' @param r [data.frame] possibly with encoding errors
#'
#' @returns `r`, but with all invalidly encoded characters deleted
#' @noRd
util_fix_encoding_cols <- function(r) {
  if (nrow(r) > 0) {
    chars <- vapply(r, is.character, FUN.VALUE = logical(1))
    if (any(chars))
      r[, chars] <-
        lapply(r[, chars, drop = FALSE], util_fix_encoding)
  }
  return(r)
}

test_that("util_fix_encoding works", {
  x <- rawToChar(as.raw(c(0xC3, 0x28)), multiple = FALSE)
  Encoding(x) <- "UTF-8"
  # x has invalid encoding
  expect_warning(xfix <- util_fix_encoding(x), regexp = "deleted invalid")
  expect_no_warning(with_pipeline(util_fix_encoding(x)))
  expect_equal(xfix, "(")
})

test_that("util_fix_encoding_cols", {
  x <- data.frame(a = letters, b = 1:26, c = LETTERS)
  expect_no_warning(expect_equal(util_fix_encoding_cols(x), x))
  x$a[c(1, 4)] <- rawToChar(as.raw(c(0xC3, 0x28)), multiple = FALSE)
  x$c[c(1, 4)] <- rawToChar(as.raw(c(0xC3, 0x28)), multiple = FALSE)
  Encoding(x$a) <- "UTF-8"
  Encoding(x$c) <- "UTF-8"
  expect_warning(expect_warning(xfix <- util_fix_encoding_cols(x),
                                regexp = "deleted invalid"),
                 regexp = "deleted invalid")
})

#' Extension of [acc_shape_or_scale] to examine uniform distributions of
#' end digits
#'
#' @description
#' This implementation contrasts the empirical distribution of a measurement
#' variables against assumed distributions. The approach is adapted from the
#' idea of rootograms (Tukey (1977)) which is also applicable for count data
#' (Kleiber and Zeileis (2016)).
#'
#' [Indicator]
#'
#' @details
#' # ALGORITHM OF THIS IMPLEMENTATION:
#' - This implementation is restricted to data of type float or integer.
#' - Missing codes are removed from resp_vars (if defined in the metadata)
#' - The user must specify the column of the metadata containing probability
#'   distribution (currently only: normal, uniform, gamma)
#' - Parameters of each distribution can be estimated from the data or are
#'   specified by the user
#' - A histogram-like plot contrasts the empirical vs. the technical
#'   distribution
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable] the names of the measurement variables,
#'                             mandatory
#'
#' @return a [list] with:
#'   - `SummaryTable`: [data.frame] with the columns `Variables` and `FLG_acc_ud_shape`
#'   - `SummaryPlot`: ggplot2 distribution plot comparing expected
#'                    with observed distribution
#'
#' @export
#' @importFrom rlang :=
#'
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_end_digits.html
#' )
acc_end_digits <- function(resp_vars = NULL,
                           study_data,
                           label_col,
                           item_level = "item_level",
                           meta_data = item_level,
                           meta_data_v2) {

  # preps ----------------------------------------------------------------------
  # map metadata to study data
  util_maybe_load_meta_data_v2()
  prep_prepare_dataframes(.replace_hard_limits = TRUE)

  # correct variable use?
  util_correct_variable_use("resp_vars",
    allow_na = FALSE,
    need_type = "integer|float",
    need_scale = "interval|ratio"
  )

  if (.called_in_pipeline && util_is_na_0_empty_or_false(
    meta_data[meta_data[[label_col]] == resp_vars, END_DIGIT_CHECK])) {
    util_error("No end digit check requested for %s",
               dQuote(resp_vars),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
    # this is not really intrinsic, but, from a user's point of view, it is.
  }

  # checks
  if (any(is.infinite(ds1[[resp_vars]]))) {
    util_error("Values in 'resp_vars' must not contain infinite data",
               applicability_problem = TRUE)
  }

  vtype <- meta_data[meta_data[[label_col]] == resp_vars, DATA_TYPE]

  decs <- meta_data[meta_data[[label_col]] == resp_vars, DECIMALS]

  if (is.null(decs)) {
    decs <- NA
  }

  decs <- as.integer(decs)

  if (vtype == DATA_TYPES$FLOAT && (is.na(decs))) {
    util_error(
      "The number of digits following the decimal point must be prespecified.",
      applicability_problem = FALSE
    )
  }

  if (vtype == DATA_TYPES$FLOAT && all(util_is_integer(ds1[[resp_vars]]))) {
    util_message("%s is of type integer.",
                 dQuote(resp_vars),
                 applicability_problem = TRUE)
    vtype <- "integer"
  }

  if (vtype == DATA_TYPES$FLOAT) {
    # use modulo ops and sprintf() to remove leading numbers
    x_lds <- sprintf(paste("%.", decs, "f", sep = ""), ds1[[resp_vars]] %% 1)
    # store last digit in `resp_vars` column
    ds1[[resp_vars]] <- util_as_numeric(
      substring(x_lds, first = decs + 2, last = decs + 2)
    )
  } else {
    ds1[[resp_vars]] <- ds1[[resp_vars]] %% 10
  }

  if (!(DISTRIBUTION %in% names(meta_data))) {
    meta_data[[DISTRIBUTION]] <- NA
  }

  mrow <- which(meta_data[[label_col]] == resp_vars)
  meta_data[[DISTRIBUTION]][mrow] <- DISTRIBUTIONS$UNIFORM
  meta_data[[DATA_TYPE]][mrow] <- DATA_TYPES$INTEGER

  res <- acc_shape_or_scale(
    resp_vars = resp_vars, guess = FALSE, par1 = 0, par2 = 9, end_digits = TRUE,
    study_data = ds1, meta_data = meta_data, dist_col = DISTRIBUTION,
    label_col = label_col
  )

  if (length(attr(res, "error")) > 0) {
    util_error(attr(res, "error")[[1]])
  }

  st <- res$SummaryTable
  st$Variables <- gsub("_x_last[\\s\\d]*$", "", st$Variables)

  return(util_attach_attr(list(SummaryTable = st,
              SummaryPlot = util_set_size(
                res$SummaryPlot,
                width_em = 15
              )), sizing_hints = attr(res, "sizing_hints")
  ))
}

#' Place all geom_texts also in `plotly` right from the x position
#'
#' @param plotly the `plotly`
#'
#' @return modified `plotly`-built object
#'
#' @family plotly
#' @concept reporting
#' @noRd
util_adjust_geom_text_for_plotly <- function(plotly) {
  util_ensure_suggested("plotly")
  util_stop_if_not(inherits(plotly, "plotly"))
  withCallingHandlers(
    pyb <- util_plotly_build(plotly),
    warning = function(cond) { # suppress a waning caused by ggplotly for barplots
      if (startsWith(conditionMessage(cond),
                     "'bar' objects don't have these attributes: 'mode'") ||
          startsWith(conditionMessage(cond),
                     "'box' objects don't have these attributes: 'mode'")) {
        invokeRestart("muffleWarning")
      }
      if (any(grepl("the mode", conditionMessage(cond)))) {
        invokeRestart("muffleWarning")
      }
    },
    message = function(cond) { # suppress a waning caused by ggplotly for barplots
      if (startsWith(conditionMessage(cond),
                     "'bar' objects don't have these attributes: 'mode'") ||
          startsWith(conditionMessage(cond),
                     "'box' objects don't have these attributes: 'mode'")) {
        invokeRestart("muffleMessage")
      }
      if (any(grepl("the mode", conditionMessage(cond)))) {
        invokeRestart("muffleMessage")
      }
    })

  no_type <-
    vapply(lapply(pyb$x$data, `[[`, "type"), is.null, FUN.VALUE = logical(1))
  type <- rep(NA_character_, length(no_type))
  type[!no_type] <-
    vapply(pyb$x$data[!no_type], `[[`, "type", FUN.VALUE = character(1))
  type_scatter <- rep(FALSE, length(no_type))
  type_scatter[!no_type] <-
    vapply(type[!no_type], `==`, "scatter", FUN.VALUE = logical(1))

  no_mode <-
    vapply(lapply(pyb$x$data, `[[`, "mode"), is.null, FUN.VALUE = logical(1))
  mode <- rep(NA_character_, length(no_mode))
  mode[!no_mode] <-
    vapply(pyb$x$data[!no_mode], `[[`, "mode", FUN.VALUE = character(1))
  mode_text <- rep(FALSE, length(no_mode))
  mode_text[!no_mode] <-
    vapply(mode[!no_mode], `==`, "text", FUN.VALUE = logical(1))

  withCallingHandlers(
    plotly::style(pyb, textposition = "right", traces = mode_text & type_scatter),
    warning = function(cond) { # suppress a waning caused by ggplotly for barplots
      if (startsWith(conditionMessage(cond),
                     "'bar' objects don't have these attributes: 'mode'") ||
          startsWith(conditionMessage(cond),
                     "'box' objects don't have these attributes: 'mode'")) {
        invokeRestart("muffleWarning")
      }
      if (any(grepl("the mode", conditionMessage(cond)))) {
        invokeRestart("muffleWarning")
      }
    },
    message = function(cond) { # suppress a waning caused by ggplotly for barplots
      if (startsWith(conditionMessage(cond),
                     "'bar' objects don't have these attributes: 'mode'") ||
          startsWith(conditionMessage(cond),
                     "'box' objects don't have these attributes: 'mode'")) {
        invokeRestart("muffleMessage")
      }
      if (any(grepl("the mode", conditionMessage(cond)))) {
        invokeRestart("muffleMessage")
      }
    })
}

#' Insert missing codes for `NA`s based on rules
#'
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables
#' @param rules [data.frame] with the columns:
#'   - `VAR_NAMES`: [VAR_NAMES] of the variable to compute
#'   - `COMPUTATION_RULE`: A rule in `REDcap` style (see, e.g.,
#'   [`REDcap` help](https://help.redcap.ualberta.ca/help-and-faq/project-best-practices/data-quality/example-data-quality-rules),
#'   [`REDcap` how-to](https://docs.google.com/document/d/1l3nGBgqqPKi5PtMe75g7q0dny8QzGMd_/edit?tab=t.0)), and
#'   [`REDcap` branching logic](https://www.iths.org/wp-content/uploads/REDCap-Branching-Logic-2017-202.pdf)
#'   that defines, how to compute the new values
#' @param use_value_labels [logical] In rules for factors, use the value labels,
#'                                   not the codes. Defaults to `TRUE`, if any
#'                                   `VALUE_LABELS` are given in the metadata.
#'
#' @return a `list` with the entry:
#'   - `ModifiedStudyData`: Study data with the new variables
#'
#' @export
#'
#' @examples
#' \dontrun{
#' study_data <- prep_get_data_frame("ship")
#' prep_load_workbook_like_file("ship_meta_v2")
#' meta_data <- prep_get_data_frame("item_level")
#' rules <- tibble::tribble(
#'   ~VAR_NAMES,  ~COMPUTATION_RULE,
#'   "BMI", '[BODY_WEIGHT_0]/(([BODY_HEIGHT_0]/100)^2)',
#'   "R", '[WAIST_CIRC_0]/2/[pi]', # in m^3
#'   "VOL_EST", '[pi]*([WAIST_CIRC_0]/2/[pi])^2*[BODY_HEIGHT_0] / 1000', # in l
#'  )
#'  r <- prep_add_computed_variables(study_data, meta_data,
#'    label_col = "LABEL", rules, use_value_labels = FALSE)
#' }
prep_add_computed_variables <- function(
                                   study_data,
                                   meta_data,
                                   label_col,
                                   rules,
                                   use_value_labels) {

  prep_prepare_dataframes(.replace_missings = FALSE,
                          .replace_hard_limits = FALSE,
                          .adjust_data_type = FALSE,
                          .amend_scale_level = FALSE)

  if (!is.data.frame(rules) || !nrow(rules)) {
    return(list(ModifiedStudyData = ds1))
  }

  util_expect_data_frame(rules, list(VAR_NAMES = function(x) {
    !any(x %in% colnames(study_data))
  },
  COMPUTATION_RULE = is.character))
  if (DATA_PREPARATION %in% colnames(rules)) {
    rules[[DATA_PREPARATION]] <-
      vapply(lapply(lapply(lapply(lapply(
        util_parse_assignments(rules[[DATA_PREPARATION]],
                               multi_variate_text = TRUE),
        toupper),
        trimws),
        sort),
        unique), FUN =
          prep_deparse_assignments, mode = "string_codes",
        FUN.VALUE = character(1))
    rules_split <- split(rules, rules[[DATA_PREPARATION]])
    res <- ds1
    for (my_rules in names(rules_split)) {
      to_apply <- util_parse_assignments(my_rules)
      curr_rules <- rules_split[[my_rules]]
      curr_rules[[DATA_PREPARATION]] <- NULL
      replace_missing_by <- ""
      if (sum("MISSING_INTERPRET" %in% to_apply,
              "MISSING_LABEL" %in% to_apply,
              "MISSING_NA" %in% to_apply) > 1) {
        util_warning("Invalid %s in computation rules. Falling back to %s",
                     sQuote(DATA_PREPARATION),
                     dQuote("MISSING_NA"))
        to_apply <- to_apply[!startsWith(to_apply, "MISSING_")]
        to_apply <- c(to_apply, "MISSING_NA")
      }
      if ("MISSING_NA" %in% to_apply) {
        replace_missing_by <- "NA"
      }
      if ("MISSING_INTERPRET" %in% to_apply) {
        replace_missing_by <- "INTERPRET"
      }
      if ("MISSING_LABEL" %in% to_apply) {
        replace_missing_by <- "LABEL"
      }
      res <- .util_add_computed_variables(ds1 = res,
                                          meta_data = meta_data,
                                          label_col = label_col,
                                          rules = curr_rules,
                                          use_value_labels =
                                            ("LABEL" %in% to_apply),
                                          replace_missing_by =
                                            replace_missing_by,
                                          replace_limits =
                                            ("LIMITS" %in% to_apply)
                                          )$ModifiedStudyData
    }
    return(list(ModifiedStudyData = res))
  } else {
    return(.util_add_computed_variables(ds1 = ds1,
                                        meta_data = meta_data,
                                        label_col = label_col,
                                        rules = rules,
                                        use_value_labels =
                                          use_value_labels))
  }

}

.util_add_computed_variables <- function(ds1,
                                         meta_data,
                                         label_col,
                                         rules,
                                         use_value_labels,
                                         replace_missing_by = "NA",
                                         replace_limits = TRUE) {

  if (missing(use_value_labels)) {
    use_value_labels <- VALUE_LABELS %in% colnames(meta_data) &&
      any(!util_empty(meta_data[[VALUE_LABELS]]))
  }

  util_expect_scalar(use_value_labels, check_type = is.logical)

  compiled_rules <- lapply(setNames(nm = rules[[VAR_NAMES]],
                                    rules[[COMPUTATION_RULE]]),
                           util_parse_redcap_rule)

  rule_res <- mapply(SIMPLIFY = FALSE,
                     rule = compiled_rules,
                     nm = names(compiled_rules),
                     function(rule, nm) {
    # util_message("Rule: %s", paste(attr(rule, "src")))
    util_message("%s", nm)
    r <- try(util_eval_rule(rule = rule,
                   ds1 = ds1,
                   meta_data = meta_data,
                   use_value_labels = use_value_labels,
                   replace_missing_by = replace_missing_by,
                   replace_limits = replace_limits))
    if (util_is_try_error(r)) {
      rl <- "?"
      try(rl <- attr(rule, "src"), silent = TRUE)
      if (length(rl) != 1)
        rl <- "?"
      util_warning("Could not evaluate rule %s: %s, results are all NA",
                   sQuote(rl),
                   dQuote(conditionMessage(attr(r, "condition"))),
                   applicability_problem =
                     TRUE)
      NA
    } else {
      r
    }
  })

  ModifiedStudyData <- ds1

  ModifiedStudyData[, names(rule_res)] <- rule_res

  return(list(ModifiedStudyData = ModifiedStudyData))
}

.redcap_cache <- new.env(parent = emptyenv())
#' Interpret a `REDcap`-style rule and create an expression, that represents this rule
#'
#' @param rule [character] `REDcap` style rule
#' @param debug [integer] debug level (0 = off, 1 = log, 2 = breakpoints)
#' @param entry_pred [character] for debugging reasons: The production
#'        rule used entry point for the parser
#' @param must_eof [logical] if `TRUE`, expect the input to be `eof`, when the
#'        parser succeeded, fail, if not.
#'
#' @return [expression] the interpreted rule
#'
#' [`REDcap` rules 1](https://help.redcap.ualberta.ca/help-and-faq/project-best-practices/data-quality/example-data-quality-rules)
#' [`REDcap` rules 2](https://docs.google.com/document/d/1l3nGBgqqPKi5PtMe75g7q0dny8QzGMd_/edit?tab=t.0)
#' [`REDcap` rules 3](https://www.iths.org/wp-content/uploads/REDCap-Branching-Logic-2017-202.pdf)
#'
#'
#' For resolving left-recursive rules,
#' [StackOverflow](https://stackoverflow.com/a/9934631)
#' helps understanding the grammar below, just in case, theoretical computer
#' science is not right in your mind currently.
#'
#' @import qmrparser
#'
#' @examples
#' \dontrun{
#' #  rules:
#' # pregnancies <- 9999 ~ SEX == 'm' |  is.na(SEX)
#' # pregnancies <- 9998 ~ AGE < 12 |  is.na(AGE)
#' # pregnancies = 9999 ~ dist > 2 |  speed == 0
#'
#' data.frame(target = "SEX_0",
#'   rule = '[speed] > 5 and [dist] > 42 or 1 = "2"',
#'   CODE = 99999, LABEL = "PREGNANCIES_NOT_ASSESSED FOR MALES",
#'   class = "JUMP")
#' ModifyiedStudyData <- replace in SEX_0 where SEX_0 is empty, if rule fits
#' ModifyedMetaData <- add missing codes with labels and class here
#'
#' subset(study_data, eval(pregnancies[[3]]))
#'
#' rule <-
#'  paste0('[con_consentdt] <> "" and [sda_osd1dt] <> "" and',
#'  ' datediff([con_consentdt],[sda_osd1dt],"d",true) < 0')
#'
#' x <- data.frame(con_consentdt = c(util_parse_date("2020-01-01"),
#'                 util_parse_date("2020-10-20")),
#'                 sda_osd1dt = c(util_parse_date("2020-01-20"),
#'                 util_parse_date("2020-10-01")))
#' eval(util_parse_redcap_rule(paste0(
#'   '[con_consentdt] <> "" and [sda_osd1dt] <> "" and ',
#'   'datediff([con_consentdt],[sda_osd1dt],"d", "Y-M-D",true) < 10')),
#'   x, util_get_redcap_rule_env())
#'
#' util_parse_redcap_rule("[a] = 12 or [b] = 13")
#' cars[eval(util_parse_redcap_rule(
#'   rule = '[speed] > 5 and [dist] > 42 or 1 = "2"'), cars,
#'   util_get_redcap_rule_env()), ]
#' cars[eval(util_parse_redcap_rule(
#'   rule = '[speed] > 5 and [dist] > 42 or 2 = "2"'), cars,
#'   util_get_redcap_rule_env()), ]
#' cars[eval(util_parse_redcap_rule(
#'   rule = '[speed] > 5 or [dist] > 42 and 1 = "2"'), cars,
#'   util_get_redcap_rule_env()), ]
#' cars[eval(util_parse_redcap_rule(
#'   rule = '[speed] > 5 or [dist] > 42 and 2 = "2"'), cars,
#'   util_get_redcap_rule_env()), ]
#' util_parse_redcap_rule(rule = '(1 = "2" or true) and (false)')
#' eval(util_parse_redcap_rule(rule =
#'   '[dist] > sum(1, +(2, [dist] + 5), [speed]) + 3 + [dist]'),
#' cars, util_get_redcap_rule_env())
#' }
#'
#' @family parser_functions
#' @concept metadata_management
#' @noRd

util_parse_redcap_rule <- function(rule, debug = 0, entry_pred = "REDcapPred",
                                   must_eof = FALSE) {
  util_expect_scalar(must_eof, check_type = is.logical)

  if (!exists("f", .redcap_cache)) {
    f <- function(rule, debug = 0, entry_pred = "REDcapPred",
                  must_eof = must_eof) {

      empty_part <- structure(list(), class = "NULL")
      empty_part_qmr <- list(type = "empty", value = "")
      separator <- structure(list(), class = "separator")

      remove_empty_part <- function(s) {
        if (is.symbol(s)) {
          return(s)
        }
        if (identical(s, empty_part)) return(NULL)
        res <- s[!vapply(s, identical, empty_part, FUN.VALUE = logical(1))]
        if (identical(res, empty_part_qmr)) return(NULL)
        res <- res[!vapply(res, identical, empty_part_qmr, FUN.VALUE = logical(1))]
        res[vapply(res, length, FUN.VALUE = integer(1)) > 1] <-
          lapply(res[vapply(res, length, FUN.VALUE = integer(1)) > 1],
                 remove_empty_part)
        res
      }

      shift <- function(x) {
        obj_name <- force(as.character(substitute(x)))
        obj <- get(obj_name, envir = parent.frame())
        if (length(obj) > 0)
          r <- obj[[1]]
        else
          r <- NULL
        obj <- tail(obj, -1)
        assign(obj_name, obj, envir = parent.frame())
        r
      }

      tag <- function(x) {
        ex <- substitute(x)
        if (debug > 0 && length(ex) == 3 && is.symbol(ex[[2]]) && ex[[1]] == as.symbol("<-")) {
          fname <- util_deparse1(ex[[2]])
          x <- eval.parent(ex)
          if (is.function(x)) {
            y <- function(...) {
              if (debug >= 2) message(sprintf("-- %s", fname))
              args <- list(...)
              if (length(body(x)) > 0 && "action" %in% names(body(x))) {
                tagged_action <- body(x)[["action"]]
                body(x)[["action"]] <- function(...) {
                  if (debug >= 1) message(sprintf(">> %s", fname))
                  if (debug >= 3) browser() # intended use of browser() -- dont modify this line
                  r <- eval(parent.env(environment())$tagged_action)(...)
                  if (debug >= 1) message(sprintf(">> %s", dQuote(util_deparse1(r))))
                  r
                }
              } else if (length(body(x)) > 0) {
                if (body(x)[[1]] == as.symbol("{")) { stop("FIXME") # FIXME: this does not work porperly
                  tagged_action <- formals(eval(body(x)[[2]][[1]]))$action
                  body(x)[[2]][["action"]] <- function(...) {
                    if (debug >= 1) message(sprintf(">> %s", fname))
                    if (debug >= 3) browser() # intended use of browser() -- dont modify this line
                    r <- eval(parent.env(environment())$tagged_action)(...)
                    if (debug >= 1) message(sprintf(">> %s", dQuote(util_deparse1(r))))
                    r
                  }
                } else {
                  tagged_action <- formals(eval(body(x)[[1]]))$action
                  body(x)[["action"]] <- function(...) {
                    if (debug >= 1) message(sprintf(">> %s", fname))
                    if (debug >= 3) browser() # intended use of browser() -- dont modify this line
                    r <- eval(parent.env(environment())$tagged_action)(...)
                    if (debug >= 1) message(sprintf(">> %s", dQuote(util_deparse1(r))))
                    r
                  }
                }
              }
              r <- do.call(x, args, quote = TRUE)
              r
            }
            assign(x = fname,
                   value = y,
                   envir = parent.frame())
          } else {
            y <- x
          }
        } else {
          y <- eval.parent(ex)
        }
        y
      }

      tag(boolean <- function()
        alternation(
          keyword("true"),
          keyword("false"),
          action = function(s) {
            if (debug >= 1) message("boolean")
            as.logical(s$value)
          }
        )
      )

      tag(dash <- function()
        keyword("-")
      )

      tag(digit <- function()
        alternation(
          keyword("1"),
          keyword("2"),
          keyword("3"),
          keyword("4"),
          keyword("5"),
          keyword("6"),
          keyword("7"),
          keyword("8"),
          keyword("9"),
          keyword("0"),
          action = function(s) {
            if (debug >= 1) message("digit")
            s$value
          })
      )

      tag(digit012345 <- function()
        alternation(
          keyword("1"),
          keyword("2"),
          keyword("3"),
          keyword("4"),
          keyword("5"),
          keyword("0"),
          action = function(s) {
            if (debug >= 1) message("digit012345")
            s$value
          })
      )

      tag(digit0123 <- function()
        alternation(
          keyword("1"),
          keyword("2"),
          keyword("3"),
          keyword("0"),
          action = function(s) {
            if (debug >= 1) message("digit0123")
            s$value
          })
      )

      tag(digit012 <- function()
        alternation(
          keyword("1"),
          keyword("2"),
          keyword("0"),
          action = function(s) {
            if (debug >= 1) message("digit012")
            s$value
          })
      )

      tag(digit01 <- function()
        alternation(
          keyword("1"),
          keyword("0"),
          action = function(s) {
            if (debug >= 1) message("digit01")
            s$value
          })

      )

      tz_parsers <- lapply(c(
        union(OlsonNames(), c("CEST", "CET"))
        ), keyword)

      tag(tz_parser <- function() do.call(alternation, c(
        tz_parsers,
        list(action = function(s) s$value))))

      # TODO: maybe, we need to downgrade this to string, since REDcap seems unaware of a datetime data type, it uses strings and converts them later, if needed
      # One way could be to write our dates w/o quotes or in single quotes, backticks or whatever, and to support then also strings in the datediff function -- maybe add
      # some converter function string to date. Then, also datediff has to be rewritten in util_get_redcap_rule_env. See also tag: 1893839
      dt_general <- function(no_brackets = FALSE) alternation(
        concatenation(
          if (!no_brackets) keyword("[", action = function(s) empty_part) else empty(action = function(s) {empty_part}),
          digit(), digit(), digit(), digit(),
          dash(),
          digit01(), digit(),
          dash(),
          digit0123(), digit(),
          alternation(concatenation(
            whitespace(),
            ignore_ws,
            digit012(),
            digit(),
            keyword(":"),
            digit012345(),
            digit(),
            option(
              concatenation(
                keyword(":"),
                digit012345(),
                digit()
              )
            ),
            option(
              concatenation(
                whitespace(),
                ignore_ws,
                tz_parser()
              )
            )
          ), empty(action = function(s) empty_part)),
          if (!no_brackets) keyword("]", action = function(s) empty_part) else empty(action = function(s) {empty_part}),
          action = function(s) {
            s <- remove_empty_part(s)

            # Flatten while keeping names to filter parser bookkeeping fields
            flat <- unlist(s, use.names = TRUE)

            # Remove internal fields like ...$type
            if (!is.null(names(flat))) {
              flat <- flat[!grepl("type$", names(flat), perl = TRUE)]
            }

            # Work with characters
            flat <- as.character(flat)

            # Drop all "option" markers (we had two: one for :SS, one for TZ)
            flat <- flat[flat != "option"]

            # Detect TZ token from tz_parser (use same source as tz_parsers)
            tz_set <- union(OlsonNames(), c("CEST", "CET"))

            tz <- ""
            tz_pos <- tail(which(flat %in% tz_set), 1)

            if (length(tz_pos)) {
              tz <- flat[tz_pos]
              # Also drop the whitespace right before the TZ (if any),
              # because we’ll rebuild the final string without it.
              drop_idx <- tz_pos
              if (tz_pos > 1 && grepl("^\\s+$", flat[tz_pos - 1])) {
                drop_idx <- c(tz_pos - 1, drop_idx)
              }
              flat <- flat[-drop_idx]
            }

            # Rebuild the literal without internal markers or TZ token
            s_str <- paste(flat, collapse = "")

            call("util_parse_date", s_str, tz = tz)
          }
        ),
        keyword("\"today\"", action = function(s) {
          call("util_parse_date", Sys.Date())
        }),
        action = function(s) {
          if (debug >= 1) message("datetime")
          remove_empty_part(s)
        }
      )

      dt_no_brackets <- dt_general

      formals(dt_no_brackets)$no_brackets <- TRUE

      tag(datetime <- dt_general)

      tag(datetime_no_brackets <- dt_no_brackets)

      # ------ 324 387
      to_general <- function(no_brackets = FALSE) alternation(
        concatenation(
          if (!no_brackets) keyword("[", action = function(s) empty_part) else empty(action = function(s) {empty_part}),
          alternation(concatenation(
            digit012(),
            digit(),
            keyword(":"),
            digit012345(),
            digit(),
            option(
              concatenation(
                keyword(":"),
                digit012345(),
                digit()
              )
            ),
            option(
              concatenation(
                whitespace(),
                ignore_ws,
                tz_parser()
              )
            )
          ), empty(action = function(s) empty_part)),
          if (!no_brackets) keyword("]", action = function(s) empty_part) else empty(action = function(s) {empty_part}),
          action = function(s) {
            s <- remove_empty_part(s)
            s_str <- unlist(s)[!grepl("type$", perl = TRUE, names(unlist(s)))]
            if (any(unlist(s) == "option")) {
              tz0 <- unname(tail(unlist(s), 1))
              if (tz0 != "option") {
                tz <- tz0
                s_str <- head(s_str, -2)
              } else {
                tz <- ""
              }
              rm("tz0")
            } else {
              tz <- ""
            }
            call("util_parse_time", paste(s_str, collapse = ""), tz = tz)
          }
        ),
        action = function(s) {
          if (debug >= 1) message("time")
          remove_empty_part(s)
        }
      )

      to_no_brackets <- to_general

      formals(to_no_brackets)$no_brackets <- TRUE

      tag(time <- to_general)

      tag(time_no_brackets <- to_no_brackets)
      # ----- 324 387

      tag(nan <- function() keyword('"NaN"', action = function(s) {
        if (debug >= 1) message("nan")
        ""
      }))

      tag(set <- function() concatenation(
                                          charParser("{"),
                                          ignore_ws,
                                          arg_part(),
                                          ignore_ws,
                                          charParser("}"),
                                          action = function(s0) {
                                            s <- s0
                                            s <- remove_empty_part(s)
                                            do.call(call, c(list(name = "set"), s[[2]]), quote = TRUE)
                                          }
      ))
      # --- strict time without brackets, no empty alternative
      tag(time_strict_no_brackets <- function()
        concatenation(
          digit012(), digit(), keyword(":"),          # HH:
          digit012345(), digit(),                     # MM
          option(concatenation(keyword(":"), digit012345(), digit())),  # :SS
          option(concatenation(whitespace(), ignore_ws, tz_parser())),  # TZ (optional)
          action = function(s0) {
            s <- remove_empty_part(s0)
            s_str <- unlist(s)[!grepl("type$", perl = TRUE, names(unlist(s)))]
            tz <- ""
            if (any(unlist(s) == "option")) {
              tz0 <- unname(tail(unlist(s), 1))
              if (tz0 != "option") {
                tz <- tz0
                s_str <- head(s_str, -2)
              }
              rm("tz0")
            }
            list(type = "time_strict", value = paste(s_str, collapse = ""), tz = tz)
          }
        )
      )

      # --- strict datetime without brackets, no empty alternative
      tag(datetime_strict_no_brackets <- function()
        concatenation(
          digit(), digit(), digit(), digit(),         # YYYY
          dash(),
          digit01(), digit(),                         # MM
          dash(),
          digit0123(), digit(),                       # DD
          alternation(
            concatenation(
              whitespace(), ignore_ws,
              digit012(), digit(), keyword(":"),      # HH:
              digit012345(), digit(), keyword(":"),   # MM:
              digit012345(), digit(),                 # SS
              option(concatenation(whitespace(), ignore_ws, tz_parser()))
            ),
            empty(action = function(s) empty_part)
          ),
          action = function(s0) {
            s <- remove_empty_part(s0)
            s_str <- unlist(s)[!grepl("type$", perl = TRUE, names(unlist(s)))]
            tz <- ""
            if (any(unlist(s) == "option")) {
              tz0 <- unname(tail(unlist(s), 1))
              if (tz0 != "option") {
                tz <- tz0
                s_str <- head(s_str, -2)
              }
              rm("tz0")
            }
            list(type = "datetime_strict", value = paste(s_str, collapse = ""), tz = tz)
          }
        )
      )

      tag(interval <- function() concatenation(
        alternation(charParser("("), charParser("["), action = function(s0) {
          s <- s0
          s <- remove_empty_part(s)
          s$value
        }),
        ignore_ws,
        alternation(
                    datetime_no_brackets(),
                    keyword("Inf", action = function(s0) {Inf}),
                    keyword("+Inf", action = function(s0) {Inf}),
                    keyword("-Inf", action = function(s0) {-Inf}),
                    term_expression(),#TODO: need alternative datetime rule here (w/o square brackets)
                    empty(action = function(s) {-Inf}),
                    time_no_brackets(),
                    action = function(s0) {
                      s <- s0
                      s <- remove_empty_part(s)
                      s
                    }),
        ignore_ws,
        alternation(charParser(";"), charParser(","),
                    action = function(s0) empty_part),
        ignore_ws,
        alternation(
                    datetime_no_brackets(),
                    keyword("Inf", action = function(s0) {Inf}),
                    keyword("+Inf", action = function(s0) {Inf}),
                    keyword("-Inf", action = function(s0) {-Inf}),
                    term_expression(),
                    empty(action = function(s) {+Inf}),
                    time_no_brackets(),
                    action = function(s0) {
                      s <- s0
                      s <- remove_empty_part(s)
                      s
                    }),
        ignore_ws,
        alternation(charParser(")"), charParser("]"), action = function(s0) {
          s <- s0
          s <- remove_empty_part(s)
          s$value
        }),
        action = function(s0) {
          s <- s0
          s <- remove_empty_part(s)
          r <- unlist(s, recursive = FALSE, use.names = FALSE)
          do.call(call, c(list(name = "interval",
                               inc_l = r[[1]] == "[",
                               low = r[[2]],
                               upp = r[[3]],
                               inc_u = r[[4]] == "]")), quote = TRUE)
        }
      ))

      tag(literal <- function()
        alternation(
          set(),
          interval(),
          nan(),
          datetime(),
          datetime_strict_no_brackets(),
          time(),
          time_strict_no_brackets(),
          string(),
          numberFloat(),
          numberInteger(),
          numberNatural(),
          numberScientific(),
          boolean(),
          action = function(s) {
            if (debug >= 1) message("literal")
            if (!is.list(s))
              return(s)

            s <- remove_empty_part(s)
            if (s$type %in% c("datetime", "datetime_strict")) {
              # util_parse_date(s$value)
              call("util_parse_date", s$value)
            } else if (s$type %in% c("time", "time_strict")) {
              if (!is.null(s$tz) && nzchar(s$tz)) {
                call("util_parse_time", s$value, tz = s$tz)
              } else {
                call("util_parse_time", s$value)
              }
            } else if (s$type == "string") {
              as.character(s$value)
            } else if (s$type %in% c("numberInteger", "numberNatural")) {
              as.integer(s$value)
            } else if (s$type %in% c("numberFloat", "numberScientific")) {
              as.numeric(s$value)
              # } else if (s$type == "boolean") {
              #   as.logical(s$value)
            }

          }
        )
      )

      tag(symbol_expression <- function()
        concatenation(keyword("["), symbolic(charFirst = function(ch) isLetter(ch) || ch == "_" || ch == ".", charRest = function(ch) isLetter(ch) ||
                                               isDigit(ch) || ch == "-" || ch == "_" || ch == "."), keyword("]"),
                      action = function(s) {
                        if (debug >= 1) message("symbol_expression")
                        s <- remove_empty_part(s)
                        as.symbol(s[[2]]$value)
                      })
      )

      tag(REDcapPred <- function() concatenation( # TODO: simplify
        alternation(
          expression()
        ),
        eofMark(),
        action = function(s) {
          if (debug >= 1) message("REDcapPred")
          s <- remove_empty_part(s)
          s[[1]]$value
        }))

      keywordfunction <- function(fname) {
        kwf <- function(...) {
          if (debug >= 2) message(sprintf("-- %s", fname))
          rf <- keyword(fname, action = function(s) {
            if (debug >= 1) message(sprintf(">> %s", fname))
            if (debug >= 3) browser() # intended use of browser() -- dont modify this line
            r <- list(type = "keyword", value = s)
            if (debug >= 1) message(sprintf(">> %s", dQuote(util_deparse1(r))))
            r
          })
          r <- rf(...)
          r
        }
        kwf
      }

      redcap_rule_env <- util_get_redcap_rule_env()

      functions <- unique(setdiff(names(redcap_rule_env), # but only, if ** is in names
                                  c("(")))

      get_internal <- function(fname) {
        r <- as.logical(attr(redcap_rule_env[[fname]], "internal"))
        if (length(r) == 0) {
          r <- FALSE
        }
        r
      }

      internal <- vapply(setNames(nm = functions), get_internal,
                          FUN.VALUE = integer(1))

      functions <- functions[!internal]

      get_order <- function(fname) {
        r <- as.integer(attr(redcap_rule_env[[fname]], "order"))
        if (length(r) == 0) {
          r <- 0L
        }
        r
      }

      get_prio <- function(fname) {
        r <- as.integer(attr(redcap_rule_env[[fname]], "prio"))
        if (length(r) == 0) {
          r <- -3L
        }
        r
      }

      fkt_order <- vapply(setNames(nm = functions), get_order,
                         FUN.VALUE = integer(1))

      functions <- functions[order(fkt_order)]

      op_prio <- vapply(setNames(nm = functions), get_prio,
                        FUN.VALUE = integer(1))
      supported_fkt <- lapply(functions, keywordfunction)
      supported_fkt <- append(supported_fkt,
                              empty(action =
                                      function(s) {
                                        list(type = "empty", value = "identity")
                                      }))
      op_prio <- c(op_prio, -3L)

      fkt <- function() do.call(
        what = alternation,
        args = append(supported_fkt,
                      c(action = function(s) {
                        f <- s$value
                        if (debug >= 1) message(sprintf("fkt: %s", dQuote(f)))
                        if (f == "identity") {
                          return(empty_part)
                        } else {
                          as.symbol(f)
                        }
                      })
        )
      )

      op_far_below_lowest <- function() do.call( # TODO: rename these with a number in the name
        what = alternation,
        args = append(supported_fkt[op_prio == -2],
                      c(action = function(s) {
                        f <- s$value
                        if (debug >= 1) message(sprintf("fkt: %s", dQuote(f)))
                        if (f == "identity") {
                          return(empty_part)
                        } else {
                          as.symbol(f)
                        }
                      })
        )
      )

      op_below_lowest <- function() do.call(
        what = alternation,
        args = append(supported_fkt[op_prio == -1],
                      c(action = function(s) {
                        f <- s$value
                        if (debug >= 1) message(sprintf("fkt: %s", dQuote(f)))
                        if (f == "identity") {
                          return(empty_part)
                        } else {
                          as.symbol(f)
                        }
                      })
        )
      )


      op_lowest <- function() do.call(
        what = alternation,
        args = append(supported_fkt[op_prio == 0],
                      c(action = function(s) {
                        f <- s$value
                        if (debug >= 1) message(sprintf("fkt: %s", dQuote(f)))
                        if (f == "identity") {
                          return(empty_part)
                        } else {
                          as.symbol(f)
                        }
                      })
        )
      )

      op_low <- function() do.call(
        what = alternation,
        args = append(supported_fkt[op_prio == 1],
                      c(action = function(s) {
                        f <- s$value
                        if (debug >= 1) message(sprintf("fkt: %s", dQuote(f)))
                        if (f == "identity") {
                          return(empty_part)
                        } else {
                          as.symbol(f)
                        }
                      })
        )
      )

      op_high <- function() do.call(
        what = alternation,
        args = append(supported_fkt[op_prio == 2],
                      c(action = function(s) {
                        f <- s$value
                        if (debug >= 1) message(sprintf("fkt: %s", dQuote(f)))
                        if (f == "identity") {
                          return(empty_part)
                        } else {
                          as.symbol(f)
                        }
                      })
        )
      )


      op_highest <- function() do.call(
        what = alternation,
        args = append(supported_fkt[op_prio == 3],
                      c(action = function(s) {
                        f <- s$value
                        if (debug >= 1) message(sprintf("fkt: %s", dQuote(f)))
                        if (f == "identity") {
                          return(empty_part)
                        } else {
                          as.symbol(f)
                        }
                      })
        )
      )

      tag(arg <- function() alternation(function_expression(),
                                        literal(),
                                        symbol_expression(),
                                        action = function(s) {
                                          if (debug >= 1) message(sprintf("arg %s", dQuote(paste0(deparse(s), collapse = " "))))
                                          if (is.list(s)) {
                                            s <- remove_empty_part(s)
                                            s$value
                                          } else {
                                            s
                                          }
                                        }))

      tag(arg_part <- function() alternation(
        concatenation(
          repetition1N(
            concatenation(ignore_ws,
                          term_expression(),
                          ignore_ws,
                          alternation(charParser(","),
                                      charParser(";"), action = function(s) {
                                        separator
                                      }),
                          ignore_ws
            )
          ),
          ignore_ws,
          term_expression(),
          ignore_ws
        ),
        concatenation(
          ignore_ws,
          term_expression(),
          ignore_ws
        ),
        ignore_ws,
        action = function(s0) {
          if (debug >= 1) message("arg_part")
          s <- remove_empty_part(s0$value)
          if (length(s) > 1 &&
              length(s[[1]]) > 1 &&
              identical(s[[1]]$type,
                        "repetition1N")) {
            rep <- s[[1]]$value
            rep <- lapply(rep,
                          function(t)
                            remove_empty_part(
                              t$value))
            rep <- lapply(
              rep,
              function(r) {
                r[!vapply(r, identical,
                          separator,
                          FUN.VALUE =
                            logical(1))]
              })
            res <- append(unlist(rep, recursive = FALSE), s[[2]])
          } else if (length(s) > 0) {
            res <- s[[1]]
          } else {
            res <- empty_part
          }
          #res <- unlist(lapply(res, deparse))
          #paste(res, collapse = ", ")
          res
        }
      ))

      tag(function_expression <- function() concatenation(fkt(), ignore_ws,
                                                          charParser("(", action = function(s) as.symbol("(")),
                                                          ignore_ws,
                                                          arg_part(),
                                                          ignore_ws,
                                                          charParser(")", action = function(s) as.symbol(")")),
                                                          action = function(s0) {
                                                            s <- s0
                                                            s <- remove_empty_part(s)

                                                            fkt <- gsub("`", "", fixed = TRUE, as.character(shift(s))) # TODO: use a better ways
                                                            if (debug >= 1) message(sprintf("function_expression: %s", sQuote(fkt)))
                                                            lng <- vapply(s, function(s_i) {
                                                              (length(s_i) == 1 && is.language(s_i))
                                                            }, FUN.VALUE = logical(1))
                                                            # val <- vapply(s, function(s_i) {
                                                            #   (length(s_i) == 2)
                                                            # }, FUN.VALUE = logical(1))
                                                            # s1 <- s
                                                            # s1[val] <- lapply(s[val], `[[`, "value")
                                                            # s[lng] <- lapply(s[lng], deparse)
                                                            r <- unlist(s, recursive = FALSE, use.names = FALSE)
                                                            # r <- r[!vapply(r, function(x) !is.language(x) && util_empty(x), FUN.VALUE = logical(1))]
                                                            # if (length(r) == 3 && is.language(r[[2]])) {
                                                            #   call(gsub("`", "", as.character(r[[2]])), r[[1]], r[[3]])
                                                            # } else {
                                                            #
                                                            # }
                                                            r <- r[!(r %in% c(as.symbol(")"), as.symbol("(")))]
                                                            do.call(call, c(list(name = fkt), r), quote = TRUE)
                                                          }
      ))

      action_subterm = function(s) {
        if (debug > 0) message(util_deparse1(sys.call()))
        s <- remove_empty_part(s)
        if (length(s)) {
          part <- function(x) {
            if (length(x) < 3) {
              if (length(x) == 1)
                x[[1]]
              else
                stop("Syntax error")
            } else {
              call(util_deparse1(x[[length(x)-1]]),
                   Recall(head(x, -2)),
                   x[[length(x)]]
              )
            }
          }
          if (length(s) > 1) {
             if (inherits(s[[1]], "POSIXct")) {
               s[[1]] <- call("util_parse_date", list(as.character(s[[1]])))
              #   s[[1]] <- as.numeric(s[[1]])
             }
            ensure_brackets(part(c(s[[1]], s[[2]])))
          } else
            ensure_brackets(part(c(s[[1]])))
        } else
          empty_part
      }

      action_op <- function(s) {
        if (debug > 0) message(util_deparse1(sys.call()))
        s <- remove_empty_part(s)
        if (inherits(s[[2]], "POSIXct")) {
          s[[2]] <- call("util_parse_date", list(as.character(s[[2]])))
        }
        c(s[[1]], s[[2]])
      }

      action_factor <- function(s) {
        if (debug > 0) message(util_deparse1(sys.call()))
        s <- remove_empty_part(s)
        if (!is.null(s))
          do.call("c", s$value)
        else
          empty_part # base::expression()
      }

      ensure_brackets <- function(x) {
        if (length(x) < 2) {
          x
        } else {
          if (!is.call(x) || x[[1]] == as.symbol("(")) {
            x
          } else {
            call("(", x)
          }
        }
      }

      tag( expression <- function() concatenation(
        alternation(op_far_below_lowest(), empty(action = function(s) {empty_part}), action = function(s) {
          s <- remove_empty_part(s)
          if (length(s) == 0)
            empty_part
          else
            s
        }), ignore_ws,
        term(), ignore_ws,
        repetition0N(concatenation(op_far_below_lowest(),  ignore_ws, term()),  ignore_ws,
                     action = function(s) {
                       s <- remove_empty_part(s)
                       if (!is.null(s))
                         do.call("c", lapply(s$value, `[[`, "value"))
                       else
                         empty_part # base::expression()
                     }),
        action = action_subterm)
      )


      tag( term <- function() concatenation(factor1(),  ignore_ws, repetition0N(concatenation(op_below_lowest(),  ignore_ws, factor1(), ignore_ws,
                                                                                              action = action_op),
                                                                                action = action_factor),
                                            action = action_subterm)
      )

      tag( factor1 <- function() concatenation(factor2(),  ignore_ws, repetition0N(concatenation(op_lowest(),  ignore_ws, factor2(),  ignore_ws,
                                                                                                 action = action_op),
                                                                                   action = action_factor),
                                               action = action_subterm)
      )

      tag( factor2 <- function() concatenation(factor3(), ignore_ws, repetition0N(concatenation(op_low(), ignore_ws, factor3(), ignore_ws,
                                                                                                action = action_op),
                                                                                  action = action_factor),
                                               action = action_subterm)
      )

      tag( factor3 <- function() concatenation(factor4(), ignore_ws, repetition0N(concatenation(op_high(), ignore_ws, factor4(), ignore_ws,
                                                                                                action = action_op),
                                                                                  action = action_factor),
                                               action = action_subterm)
      )

      tag( factor4 <- function() concatenation(factor5(), ignore_ws, repetition0N(concatenation(op_highest(), ignore_ws, factor5(), ignore_ws,
                                                                                                action = action_op),
                                                                                  action = action_factor),
                                               action = action_subterm)
      )

      tag( factor5 <- function() alternation(arg(),
                                             concatenation(
                                               charParser("(", action = function(s) empty_part),
                                               ignore_ws,
                                               expression(),
                                               ignore_ws,
                                               charParser(")", action = function(s) empty_part),
                                               action = function(s) {
                                                 do.call(call, c(
                                                   list(name = "("),
                                                   remove_empty_part(s)),
                                                   quote = TRUE)
                                               }),
                                             action = function(s) {
                                               s <- remove_empty_part(s)
                                               s
                                             })
      )

      term_expression <- expression

      # tag(term_expression <- function() alternation(
      #   term_expression_1(),
      #   arg(),
      #   action = function(s) {
      #     if (debug >= 1) message("term_expression")
      #     s
      #   }
      # ))


      ignore_ws <- whitespace(action = function(s) empty_part)

      # tag(term_expression_1 <- function() concatenation( # TODO: remove?
      #   alternation(function_expression(), arg()),
      #   ignore_ws,
      #   #      alternation(op_low()),
      #   alternation(
      #     op_highest(), # ^ * is prefix of **
      #     op_lowest(), # compare ops
      #     op_low(), # +
      #     op_high(), # *
      #     action = function(s) {
      #       s
      #     }
      #   ),
      #   ignore_ws,
      #   alternation(term_expression(), function_expression(), arg()),
      #   action = function(s) {
      #     if (debug >= 1) message("term_expression_1")
      #     s <- remove_empty_part(s)
      #     if(length(s) < 3) {
      #       stop(sprintf("two terms combined without an operator: %s",
      #                    dQuote(paste(s[[1]]$value,
      #                                 s[[2]]$value))))
      #     }
      #     call(as.character(s[[2]]),
      #          s[[1]]$value,
      #          s[[3]]$value
      #     )
      #   }
      # ))

      errorFun <- function(strmPosition,h=NULL,type="") {
        if ( is.null(h) || type != "concatenation" ) {# TODO: XXX
          warning(sprintf("Error from line %d, character %d in %s (will ignore this rule): %s",
                          strmPosition$line,
                          strmPosition$linePos,
                          rule,
                          substring(rule, strmPosition$streamPos)))
        } else {
          errorFun(h$pos,h$h,h$type)
        }
        return(list(type=type,pos=strmPosition,h=h))
      }

      # cstream <- get(entry_pred)()(streamParserFromString(sprintf("(%s)", rule))) # <ADDBR>, see below
      cstream <- get(entry_pred)()(streamParserFromString(sprintf("%s", rule)))

      if ( cstream$status == "fail" ) {
        invisible(errorFun(cstream$node$pos,cstream$node$h,cstream$node$type))
        return(util_attach_attr(list(), src = rule))
      } else if ( cstream$status != "ok" ) {
        warning(sprintf("Unknown error parsing %s: %s (will ignore this rule)", rule, cstream$status))
        return(util_attach_attr(list(), src = rule))
      } else {
        # r <- cstream$node[[2]] # 2 to remove the brackets added in <ADDBR>
        iseof <- !(cstream$stream$pos < cstream$stream$lenchar)
        if (!iseof && must_eof) {
          strmPosition <- streamParserPosition(cstream$stream)
          warning(sprintf("Found extra characters in line %d, character %d in %s (will ignore this rule): %s",
                          strmPosition$line,
                          strmPosition$linePos,
                          rule,
                          substring(rule, strmPosition$streamPos)))
          return(list())
        }
        # if (cstream$) {
        #
        # }
        r <- cstream$node
        if (debug > 0) message(r)
        if (is.symbol(r)) r <- as.expression(r)
        attr(r, "src") <- rule
        return(r)
      }
    }

    # parent.env(environment(f)) <- environment(empty)

    assign("f", f, .redcap_cache)
  }

  get("f", .redcap_cache)(rule = rule, debug = debug, entry_pred = entry_pred,
                          must_eof = must_eof)

}

#' Translate standard column names to readable ones
#'
#' TODO: Duplicate of util_make_data_slot_from_table_slot ??
#'
#' @param colnames [character] the names to translate
#' @param short [logical] include unit letter in output
#' @param long [logical] include unit description in output
#' @param ignore_unknown [logical] do not replace unknown indicator metrics
#'                                 by `NA`, keep them
#'
#' @return translated names
#'
#' @keywords internal
util_translate_indicator_metrics <- function(colnames, short = FALSE,
                                             long = TRUE,
                                             ignore_unknown = FALSE) {
  util_expect_scalar(short, check_type = is.logical)
  util_expect_scalar(long, check_type = is.logical)
  util_expect_scalar(ignore_unknown, check_type = is.logical)
  util_stop_if_not(`Cannot create short labels that are long` =
                     !(short && long))
  abbreviationMetrics <- util_get_concept_info("abbreviationMetrics")
  dqi <- util_get_concept_info("dqi")
  new_colnames <-
    vapply(colnames, FUN.VALUE = character(1), FUN = function(x) {
    if (x %in% c("Variables", VAR_NAMES, STUDY_SEGMENT)) {
      return(x)
    }
    util_stop_if_not(length(x) == 1)
    nm <- strsplit(x, "_", fixed = TRUE)[[1]]
    if (length(nm) >= 2) {
      if (long) {
        m <- head(subset(abbreviationMetrics, get("Abbreviation") == nm[[1]],
                         "Metrics", drop = TRUE), 1)
      } else {
        m <- head(subset(abbreviationMetrics, get("Abbreviation") == nm[[1]],
                         "public_name", drop = TRUE), 1)
      }
      if (short) {
        # m <- head(subset(abbreviationMetrics, get("Abbreviation") == nm[[1]],
        #                  "public_name", drop = TRUE), 1)
        d <- head(subset(dqi, get("abbreviation") == paste(tail(nm, -1),
                                                           collapse = "_"),
                         "public_name", drop = TRUE), 1)
        if (all(is.na(d))) { # FIXME: sometimes, public_name is missing, sometimes, the abbreviation is not unique any more!!
          d <- head(subset(dqi, get("abbreviation") == paste(tail(nm, -1),
                                                             collapse = "_"),
                           "Name", drop = TRUE), 1)
        }
        d <- unique(d)
        d <- paste(d, collapse = " ")

      } else {
        d <- head(subset(dqi, get("abbreviation") == paste(tail(nm, -1),
                                                           collapse = "_"),
                         "Name", drop = TRUE), 1)
      }

      if (length(m) == length(d) && length(d) == 1 &&
          !util_empty(m) && !util_empty(d)) {
       #if (short)
       #   sprintf("%s", d)
       # else
          sprintf("%s (%s)", d, m)
      } else {
        NA_character_
      }
    } else {
      NA_character_
    }
  })
  if (ignore_unknown) {
    new_colnames[is.na(new_colnames)] <-
      colnames[is.na(new_colnames)]
  }
  new_colnames
}

#' Convert two vectors from a code-value-table to a key-value list
#'
#'
#' @param codes codes, numeric or dates (as default, but string codes can be
#'              enabled using the option 'mode', see below)
#' @param labels [character] labels, same length as codes
#' @param split_char [character] split character character to split
#'                               code assignments
#' @param mode [character] one of two options to insist on numeric or datetime
#'                         codes (default) or to allow for string codes
#'
#' @return a vector with assignment strings for each row of
#'         `cbind(codes, labels)`
#' @export
prep_deparse_assignments <- function(codes, labels = codes,
                                     split_char = SPLIT_CHAR,
                                     mode = c("numeric_codes",
                                              "string_codes")) {
  mode <- util_match_arg(mode)
  no_labs <- missing(labels) || length(labels) == 0
  if (length(labels) == 0) {
    labels <- codes
  }
  if (length(codes) != length(labels)) {
    util_error("%s and %s must have the same length",
               dQuote("values"),
               dQuote("labels"), applicability_problem = TRUE)
  }
  if (is.list(codes)) codes <- unlist(codes)
  if (is.list(labels)) labels <- unlist(labels)
  if (mode == "numeric_codes" && (
    suppressWarnings(!all(
      is.na(as.numeric(codes)) == is.na(codes) |
      is.na(util_parse_date(codes)) == is.na(codes) |
      is.na(util_parse_time(codes)) == is.na(codes)
    )))) {
    util_error("All codes must be finite numeric or date/time values",
               applicability_problem = TRUE)
  } else {
    codes <- as.character(codes)
  }
  if (any(grepl(split_char, labels, fixed = TRUE))) {
    util_message("Removed seperator characters (%s) from the labels",
                 dQuote(split_char), applicability_problem = TRUE)
    labels <- gsub(split_char, "", labels, fixed = TRUE)
  }
  if (!!length(codes) && length(codes) == length(labels)) {
    if (no_labs) {
      paste(codes,
            collapse = sprintf(" %s ", split_char))
    } else {
      paste(codes, "=", labels,
            collapse = sprintf(" %s ", split_char))
    }
  } else {
    split_char
  }
}

#' Check for repetitive values using the digits 8 or 9 only
#'
#' Values not being finite (see [`is.finite`]) are also reported as missing
#' codes. Also, all missing codes must be composed out of the digits 8 and
#' 9 and they must be the largest values of a variable.
#'
#' @param x [`numeric`] vector to test
#' @param n_rules [`numeric`] Only outlying values can be missing
#'                                       codes; at least `n_rules` rules in
#'                                       [`acc_univariate_outlier`] match
#'
#' @return [`logical`] indicates for each value in `x`, if it looks like a
#'                     missing code
#'
#' @seealso [`acc_univariate_outlier`]
#'
#' @family metadata_management
#' @concept missing
#' @noRd


util_looks_like_missing <- function(x, n_rules = 1) {
  if (any(prep_dq_data_type_of(x) %in%
          tolower(c(DATA_TYPES$INTEGER, DATA_TYPES$FLOAT)))) {
    x <- as.numeric(x)
  }
  if (!is.numeric(x)) {
    util_error("%s works only on numeric vectors",
               dQuote("util_looks_like_missing"))
  }
  sysmiss <- !is.finite(x)
  if (all(sysmiss)) {
    return(!logical(length = length(x)))
  }
#  x[sysmiss] <- mean(x[!sysmiss], na.rm = TRUE)
  TYPICAL_MISSINGCODES <- c(
    99, 999, 9999, 99999, 999999, 9999999, 999999999,
    99990:99999, 999990:999999, 9999990:9999999, 999999990:999999999
  )

  .x <- abs(x)
  .x <- gsub("8", "9", .x, fixed = TRUE)
  .x <- gsub(".", "", .x, fixed = TRUE)
  .x <- gsub("^0+", "", .x)
  .x <- gsub("0+$", "", .x)
  r <- .x %in% TYPICAL_MISSINGCODES

  tuk <- util_tukey(x)
  tuk[sysmiss] <- 0
  ssig <- util_3SD(x)
  ssig[sysmiss] <- 0
  hub <- util_hubert(x)
  hub[sysmiss] <- 0
  sigg <- util_sigmagap(x)
  sigg[sysmiss] <- 0
  r <- sysmiss | (r &
                    (tuk +
                       ssig +
                       hub +
                       sigg >= n_rules)
  )
  # only, if there is no number > these numbers not being one of them
  repeat {
    done <- TRUE
    if (any(r[!is.na(r)])) {
      cur_min <- min(abs(x)[r], na.rm = TRUE)
      if (any(abs(x)[!r] > cur_min, na.rm = TRUE)) {
        r <- r & (abs(x) != cur_min)
        done <- FALSE
      }
    }
    if (done) break ;
  }
  return(r)
}

test_that("util_looks_like_missing works", {
  skip_on_cran()
  expected <- rep(FALSE, 100)
  #expected[c(88, 89, 98, 99)] <- TRUE
  expect_equal(
    util_looks_like_missing(1:100, n_rules = 0),
    expected = expected
  )
  expected <- c(TRUE, TRUE, FALSE, TRUE, TRUE, TRUE, TRUE)
  expect_equal(
    util_looks_like_missing(c(99999, 99, 8, 888, -999, 9.99, 8.9898),
      n_rules = 0),
    expected = expected)
  expected <- rep(FALSE, 100)
  expect_equal(util_looks_like_missing(1:100),
               expected = expected)
  expected <- c(TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, FALSE)
  expect_equal(
    util_looks_like_missing(c(99999, 99, 8, 888, -999, 9.99, 8.9898)),
    expected = expected
  )
  expected <- c(TRUE, FALSE)
  expect_equal(
    util_looks_like_missing(c(-Inf, 9)),
    expected = expected
  )
  expect_equal(
    util_looks_like_missing(NA_integer_),
    TRUE
  )
  expect_equal(
    util_looks_like_missing(c(Inf, -Inf, NA, NaN)),
    rep(TRUE, 4)
  )

})

#' Checks for element set
#'
#' Depends on `dataquieR.ELEMENT_MISSMATCH_CHECKTYPE` option,
#' see there -- # TODO: Rind out, how to document and link
#' it here using `Roxygen`.
#'
#' [Indicator]
#'
#' @inheritParams .template_function_indicator
#' @param study_data [data.frame] the data frame that contains the measurements, mandatory.
#' @param item_level [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param meta_data [data.frame] old name for `item_level`
#' @param meta_data_v2 [character] path to workbook like metadata file, see
#'                                 [`prep_load_workbook_like_file`] for details.
#'                                 **ALL LOADED DATAFRAMES WILL BE PURGED**,
#'                                 using [`prep_purge_data_frame_cache`],
#'                                 if you specify `meta_data_v2`.
#'
#' @return a [list] with
#'   - `SegmentData`: data frame with the unexpected elements check results.
#'                       - `Segment`: name of the corresponding segment,
#'                                    if applicable, `ALL` otherwise
#'   - `SegmentTable`: data frame with the unexpected elements check results, used for the data quality report.
#'                       - `Segment`: name of the corresponding segment,
#'                                    if applicable, `ALL` otherwise
#'
#' @export
#'
#' @examples
#' \dontrun{
#' study_data <- cars
#' meta_data <- dataquieR::prep_create_meta(VAR_NAMES = c("speedx", "distx"),
#'   DATA_TYPE = c("integer", "integer"), MISSING_LIST = "|", JUMP_LIST = "|",
#'   STUDY_SEGMENT = c("Intro", "Ex"))
#' options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "none")
#' int_sts_element_segment(study_data, meta_data)
#' options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact")
#' int_sts_element_segment(study_data, meta_data)
#' study_data <- cars
#' meta_data <- dataquieR::prep_create_meta(VAR_NAMES = c("speedx", "distx"),
#'   DATA_TYPE = c("integer", "integer"), MISSING_LIST = "|", JUMP_LIST = "|",
#'   STUDY_SEGMENT = c("Intro", "Intro"))
#' options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "none")
#' int_sts_element_segment(study_data, meta_data)
#' options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact")
#' int_sts_element_segment(study_data, meta_data)
#' study_data <- cars
#' meta_data <- dataquieR::prep_create_meta(VAR_NAMES = c("speed", "distx"),
#'   DATA_TYPE = c("integer", "integer"), MISSING_LIST = "|", JUMP_LIST = "|",
#'   STUDY_SEGMENT = c("Intro", "Intro"))
#' options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "none")
#' int_sts_element_segment(study_data, meta_data)
#' options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact")
#' int_sts_element_segment(study_data, meta_data)
#' }
int_sts_element_segment <- function(study_data,
                                    item_level = "item_level",
                                    label_col,
                                    meta_data = item_level,
                                    meta_data_v2) {
  util_maybe_load_meta_data_v2()
  # prevent error that we have item_level and meta_data
  # if prep_prepare_dataframes was not called but both objects
  # are not missing any more, when prep_prepare_dataframes is called below.
  invisible(prep_prepare_dataframes(.label_col = VAR_NAMES,
                                    .allow_empty = TRUE))
  meta_data <- prep_meta_data_v1_to_item_level_meta_data(
    meta_data = meta_data, verbose = FALSE, label_col = VAR_NAMES)
  util_expect_data_frame(meta_data, c(VAR_NAMES))
  if (!STUDY_SEGMENT %in% colnames(meta_data)) {
    meta_data[[STUDY_SEGMENT]] <- "<NO SEGMENT>"
  }
  if (any(meta_data$STUDY_SEGMENT == "ALL")) { # TODO: Use [[STUDY_SEGMENT]]
    util_message("No segment should be named %s, renaming it.",
                 dQuote("ALL"), applicability_problem = TRUE)
    meta_data$STUDY_SEGMENT[meta_data$STUDY_SEGMENT == "ALL"] <-
      "RENAMED SEGMENT: ALL"
  }
  Q <- substr(dQuote(""), 1, 1)
  E <- substr(dQuote(""), 2, 2)
  e <- new.env(parent = emptyenv())
  e$conds <- list()
  register_cond <- function(cond) {
    if (identical(
      attr(cond, "integrity_indicator"), "int_sts_element")) {
      e$conds <- c(e$conds, list(cond))
    }
  }

  util_purge_study_data_cache()

  suppressMessages(
    suppressWarnings(
      withCallingHandlers(
        without_pipeline(
          prep_prepare_dataframes(.study_data = study_data,
                                  .meta_data = meta_data,
                                  .label_col = VAR_NAMES, .allow_empty = TRUE,
                                  .internal = TRUE)
        ),
        message = register_cond,
        warning = register_cond
      )
    )
  )

  msgs <- vapply(e$conds, inherits, "message", FUN.VALUE = logical(1))
  wrns <- vapply(e$conds, inherits, "warning", FUN.VALUE = logical(1))

  pct <- data.frame(stringsAsFactors = FALSE,
             Segment = rep("ALL", sum(wrns)), # TODO: If segment-wiese, how to tell the segemnt of an uexpected study variable
             DESCRIPTION = vapply(e$conds[wrns],
                                  conditionMessage,
                                  FUN.VALUE = character(1)),
             MISSING = ifelse(grepl("of the study data", vapply(e$conds[wrns],
                                                                conditionMessage,
                                                                FUN.VALUE = character(1)), fixed = TRUE),
                              "metadata",
                              "study data"
                              ),
             PCT_int_sts_element = round(as.numeric(gsub("^.*?([0-9,.]+)%.*$", "\\1",
                                                   vapply(e$conds[wrns],
               conditionMessage,
               FUN.VALUE = character(1)))), 2))

  abs <- data.frame(stringsAsFactors = FALSE,
                    Segment = rep("ALL", sum(msgs)), # TODO: If segment-wiese, how to tell the segemnt of an uexpected study variable
                    DESCRIPTION = vapply(e$conds[msgs],
                                         conditionMessage,
                                         FUN.VALUE = character(1)),
                    MISSING = ifelse(grepl("Did not find any metadata", vapply(e$conds[msgs],
                                                                                conditionMessage,
                                                                                FUN.VALUE = character(1)), fixed = TRUE),
                                     "metadata",
                                     "study data"
                    ),
                    NUM_int_sts_element = vapply(gsub(sprintf("[^%s]", Q), "", vapply(e$conds[msgs],
                                                                                                             conditionMessage,
                                                                                                             FUN.VALUE = character(1))), nchar, FUN.VALUE = integer(1))
  )

  r <- merge(pct, abs, by = c("Segment", "MISSING"), suffixes = c("p", "a"))

  r$DESCRIPTION <- paste(r$DESCRIPTIONp, r$DESCRIPTIONa)
  r$DESCRIPTIONa <- NULL
  r$DESCRIPTIONp <- NULL

  vars <- util_extract_matches(
    r$DESCRIPTION,
    paste0(Q, ".+?", E)
  )

  vars <- lapply(
    vars,
    gsub,
    pattern = paste0("^", Q, collapse = ""),
    replacement = ""
  )

  vars <- lapply(
    vars,
    gsub,
    pattern = paste0(E, "$", collapse = ""),
    replacement = ""
  )

  rr <- lapply(seq_len(nrow(r)), function(x) r[x, , FALSE])

  segsizes <- util_table_of_vct(
    meta_data[meta_data[[VAR_NAMES]] %in% colnames(study_data),
              STUDY_SEGMENT])
  segsizes <- setNames(segsizes$Freq, nm = segsizes$Var1)
  rrr <- mapply(rr, vars, SIMPLIFY = FALSE, FUN = function(a, v) {
    x <- merge(a, v, by = NULL)
    x$Segment <- prep_map_labels(x = v,
                                  meta_data = meta_data,
                                  to = STUDY_SEGMENT,
                                  ifnotfound = NA_character_)

    x$Segment <- ifelse(is.na(x$Segment), "ALL", x$Segment)
    x$resp_vars <- x$y
    x$y <- NULL

    if (a$MISSING == "study data") {
      # prep_get_labels....... for known variables missing from the study data
      rvs <- prep_get_labels(
            resp_vars = x$resp_vars,
            item_level = meta_data,
            label_col = LABEL,
            label_class = "SHORT",
            resp_vars_are_var_names_only = TRUE
        )
    } else {
      rvs <- x$resp_vars
    }
    rvs <- sort(rvs)
    if (length(names(rvs)) == 0) {
      rvs <- prep_deparse_assignments(mode = "string_codes",
                                      rvs)
    } else {
      rvs <- prep_deparse_assignments(mode = "string_codes",
                                      names(rvs),
                                      rvs)
    }

    if (all(x$Segment != "ALL")) {
      xl <- split(x, x$Segment)
      xl <- lapply(xl, function(x) {
        x$NUM_int_sts_element <- nrow(x)
        x$PCT_int_sts_element <- nrow(x) / segsizes[unique(x$Segment)]
        x$resp_vars <- rvs
        x[!duplicated(x), , FALSE]
      })
      x <- do.call(rbind.data.frame, c(xl, list(
        make.row.names = FALSE,
        stringsAsFactors = FALSE)))
    } else {
      if (any(x$Segment != "ALL")) {
        util_error(c("Unexpected error for Segments: found a mixture of",
                   "segments. Internal error in dataquieR, please report."))
      }
      x <- data.frame(stringsAsFactors = FALSE,
                      Segment = "ALL",
                      MISSING = paste(unique(x$MISSING), collapse = ", "),
                      DESCRIPTION = paste(unique(x$DESCRIPTION),
                                          collapse = ", "),
                      NUM_int_sts_element = nrow(x),
                      PCT_int_sts_element = max(x$PCT_int_sts_element),
                      resp_vars = rvs
      )
    }
    x
  })

  SegmentTable <- do.call(rbind.data.frame,
                          c(rrr, list(
                            make.row.names = FALSE,
                            stringsAsFactors = FALSE)))

  SegmentTable$DESCRIPTION <- NULL

  # segments <- unique(names(segsizes), "ALL")

  segments <- setdiff(meta_data$STUDY_SEGMENT, SegmentTable$Segment)

  # add all segments that did not have a problem ----
  complement <- data.frame(Segment = segments,
                           MISSING = rep(NA_character_, length(segments)),
                           PCT_int_sts_element = rep(0, length(segments)),
                           NUM_int_sts_element = rep(0, length(segments)),
                           resp_vars = rep(NA_character_, length(segments)))

  SegmentTable <- rbind(complement, SegmentTable)

  SegmentData <- SegmentTable
  names(SegmentData) <- c("Segment",
                          "Missing",
                          "Percentage of unexpected elements",
                          "Number of unexpected elements",
                          "Response variables")

  list(SegmentData = SegmentData,
       SegmentTable = SegmentTable)

}

#' Get rule sets for `DQ` grading
#'
#' @return names lists, names are the ruleset names, values are [data.frame]s
#'         featuring the columns `GRADING_RULESET`, `dqi_parameterstub`,
#'         `indicator_metric`, `dqi_catnum` and `dqi_cat_1` to
#'         `dqi_cat_<dqi_catnum>`
#' @family summary_functions
#' @noRd
util_get_rule_sets <- function() {
  shipped_rulesets <- system.file("grading_rulesets.xlsx",
                                  package = "dataquieR")
  if (!nzchar(shipped_rulesets) &&
      suppressWarnings(util_ensure_suggested("pkgload", err = FALSE)) &&
      pkgload::is_dev_package("dataquieR")) {
        if (util_is_try_error(# comes from a quosure featuring an older dq version.
          try(silent = TRUE, shipped_rulesets <- pkgload::package_file("inst",
                                                  "grading_rulesets.xlsx",
                                                  path =
                                                  find.package(
                                                    "dataquieR"))))) {
          rlang::warn(sprintf(
            "Could not find package source, trying to use %s from installed package",
            sQuote("grading_rulesets.xlsx")
          ),
                      .frequency_id =
                        "pkgload_confusion",
                      .frequency = "once")

          shipped_rulesets <-
            names(head(which(vapply(
              setNames(nm = file.path(.libPaths(), "dataquieR",
                                      "grading_rulesets.xlsx")), file.exists,
              FUN.VALUE = logical(1))), 1))
          if (length(shipped_rulesets) != 1) {
            shipped_rulesets <- ""
          }
        }
  }
  if (!nzchar(shipped_rulesets)) {
    util_error(
      "Internal error with pkgload, please report, sorry: Could not find %s.",
      sQuote("shipped_rulesets"))
  }
  shipped_rulesets <- prep_get_data_frame(shipped_rulesets)

  reftab <- try(
    prep_get_data_frame(getOption("dataquieR.grading_rulesets",
                                  dataquieR.grading_rulesets_default)),
    silent = TRUE
  ) # https://gitlab.com/libreumg/dataquier/-/issues/34#note_1597720844
  if (!inherits(reftab, "data.frame")) {
    if (getOption("dataquieR.grading_rulesets",
                  dataquieR.grading_rulesets_default) !=
        dataquieR.grading_rulesets_default)
      util_message("Could not find rulesets %s, using the default rulesets.",
                   dQuote(getOption("dataquieR.grading_rulesets",
                                    dataquieR.grading_rulesets_default)),
                   applicability_problem = TRUE,
                   intrinsic_applicability_problem = FALSE)
    reftab <- shipped_rulesets
  }
  util_expect_data_frame(reftab, col_names = c("indicator_metric", GRADING_RULESET))
  reftab <- reftab[!util_empty(reftab$indicator_metric), , FALSE]
  reftab <- reftab[!util_empty(reftab[[GRADING_RULESET]]), , FALSE]
  reftab$GRADING_RULESET <- suppressWarnings(trimws(reftab[[GRADING_RULESET]]))
  GRADING_RULESET_names <- setNames(nm = setdiff(unique(reftab[[GRADING_RULESET]]), NA))

  if (!("0" %in% GRADING_RULESET_names)) {
    util_error("No default GRADING_RULESET found in grading_rule",
               applicability_problem = TRUE)
  }

  GRADING_RULESETs <- lapply(GRADING_RULESET_names, function(rs) {
    reftab[!is.na(reftab$GRADING_RULESET) & reftab$GRADING_RULESET == rs, , FALSE]
  })

  default_rs <- GRADING_RULESETs[["0"]]

  for (needed in setdiff(c("CAT_applicability", "CAT_error",
                   "CAT_anamat", "CAT_indicator_or_descriptor"),
                   unique(default_rs$indicator_metric))) {
    add_row <- shipped_rulesets[!is.na(shipped_rulesets$GRADING_RULESET) &
                       shipped_rulesets$GRADING_RULESET == "0" &
                       !is.na(shipped_rulesets$indicator_metric) &
                       shipped_rulesets$indicator_metric == needed, , FALSE]
    default_rs <- util_rbind(default_rs,
                              add_row)
  }

  GRADING_RULESETs[["0"]] <- default_rs

  to_amend <- setNames(nm = setdiff(GRADING_RULESET_names, "0"))

  GRADING_RULESETs[as.character(to_amend)] <- lapply(to_amend, function(rs) {
    rst <- GRADING_RULESETs[[as.character(rs)]]
    dst <- default_rs
    dst <- dst[!(dst$indicator_metric %in% rst$indicator_metric), , # only deafult rules, if not overwritte in current
               FALSE]
    dst$GRADING_RULESET <- rep(rs, nrow(dst))
    util_rbind(
      dst,
      rst
    )
  })

  GRADING_RULESETs
}

#' Create a Q-Q Plot for squared `Mahalanobis` distance
#'
#' This helper is meant to create \code{ggplot2} objects starting from the
#' Mahalanobis ratios.
#'
#' @param MD_ratio [numeric] the `Mahalanobis` distance values divided by the
#'                            threshold value beyond which a data point
#'                            is considered an `outlier`.
#' @param mahalanobis_threshold [numeric] if not set by the user, the confidence
#'                                        level to identify multivariate `outliers`
#'                                        is 0.975 quantile of a
#'                                        chi-square distribution
#'                                        with p degrees of freedom,
#'                                        where p = variables
#'                                        `(Mayrhofer and Filzmoser, 2023)`
#' @param df [numeric] the degree of freedom, which
#'                      corresponds to the number of variables used to
#'                      calculate the `Mahalanobis` distance
#'
#' @return a list with:
#'   - `plot_MD`: [ggplot2::ggplot2] the Q-Q plot for the squared
#'                                   `Mahalanobis` distance vs.
#'                                   Chi-squared distribution
#'   - `MD_values`: [data.frame] the squared `Mahalanobis` distance values
#'   - `theoretical_q`: values expected from a chi-squared distribution
#'                      with the degree of freedom that
#'                      corresponds to the number of variables
#'
#'
#' @family util_functions
#' @concept outlier
#' @keywords internal
#'
#' @noRd
util_create_mahalanobis_ggplot <- function(MD_ratio,
                                           mahalanobis_threshold,
                                           df) {
  #Set up of colours and labels for plot ----
  # Define colours for the plot USE SHAPES!
  #col1 <-  "#2166AC"
  #col_outliers <-  "#7f0000"
  shape_non_out <- 1  #empty circle
  shape_outliers <- 4 # plus symbol

  #Define axis labels for plot
  x_label <- paste0("Quantiles  of Chi-squared (df=", df, ")")
  y_label <- "Squared Mahalanobis distance"

  #Order the MD_ratio
  MD_ratio <- sort(MD_ratio)

  # Obtain values expected from a theoretical chi-squared distribution
  # (same no. points as the sample and degree of freedom equal to the no. variables)
  probabilities <- stats::ppoints(length(MD_ratio))
  theoretical_quantiles <- stats::qchisq(probabilities, df = df)

  # Obtain threshold to identify multivariate outliers (argument mahalanobis_threshold)
  # it is by default 0.975 quantile of a chi-square distribution
  # with p degrees of freedom, where p = variables (Mayrhofer and Filzmoser, 2023)
  MD_outliers_threshold <- stats::qchisq(mahalanobis_threshold, df = df)

  #Obtain squared Mahalanobis distance from Mahalanobis ratio----
  MD <- MD_ratio * MD_outliers_threshold

  #create the data frame
  plot_data <- data.frame(
    theoretical = theoretical_quantiles,
    MD_values = MD
  )
  #order the squared Mahal. distance, actually not needed, as ordered before----
  #plot_data <- plot_data[order(plot_data$MD_values), ]

  #create a column with identified outliers based on threshold----
  plot_data$MD_outliers <- NA
  plot_data$MD_outliers <- ifelse(plot_data$MD_values > MD_outliers_threshold, 1, 0)
  plot_data$MD_outliers <- as.factor(plot_data$MD_outliers)


  #create the plot----
  p1 <- util_create_lean_ggplot({
    ggplot(plot_data, aes(x = .data[["theoretical"]],
                          y = .data[["MD_values"]])) +
      geom_point(aes(shape=as.factor(.data[["MD_outliers"]])),
         alpha = 0.5) +
      ggplot2::scale_shape_manual(name = "Outliers",
                                  values = c(shape_non_out, shape_outliers),
                                  labels = c("0" = "No",
                                             "1" = "Yes"))+
      geom_hline(yintercept = MD_outliers_threshold,
                 linetype = "dashed",
                 color = "black",
                 linewidth = 0.6) +
      labs(x = x_label,
           y = y_label) +
      theme_minimal() +
      theme(legend.position = "right")
  },
  plot_data = plot_data,
  MD_outliers_threshold = MD_outliers_threshold,
  x_label = x_label,
  y_label = y_label,
  shape_non_out = shape_non_out,
  shape_outliers = shape_outliers
  )

  # Information for sizing----
  range_y <- max(plot_data$MD_values, na.rm = TRUE) -
    min(plot_data$MD_values, na.rm = TRUE)
  range_x <- max(plot_data$theoretical, na.rm = TRUE) -
    min(plot_data$theoretical, na.rm = TRUE)

  attr(p1, "sizing_hints") <- list(
    figure_type_id = "mahalanobis_plot",
    range_x = range_x,
    range_y = range_y
  )
  rm(range_y, range_x)


 return(list(plot_MD = p1,
             MD_values = plot_data$MD_values,
             theoretical_q = plot_data$theoretical))

}

#' Support function for a parallel `pmap`
#'
#' parallel version of `purrr::pmap`
#'
#' @param .l [data.frame] with one call per line and one function argument
#'                        per column
#' @param .f [`function`] to call with the arguments from `.l`
#' @param ... additional, static arguments for calling `.f`
#' @param cores number of cpu cores to use or a (named) list with arguments for
#'              [parallelMap::parallelStart] or NULL, if parallel has already
#'              been started by the caller. Set to 0 to run without
#'              parallelization.
#'
#' @seealso `purrr::pmap`
#' @seealso [Stack Overflow post](https://stackoverflow.com/a/47575143)
#'
#' @author [Aurèle](https://stackoverflow.com/users/6197649)
#' @author S Struckmann
#'
#' @return [list] of results of the function calls
#' @export
prep_pmap <- function(.l, .f, ..., cores = 0) {
  if (!is.function(.f)) {
    util_error("Argument %s should be a function.", dQuote(".f"))
  }

  if (!is.null(cores)) {
    if (getOption("parallelMap.status", "stopped") != "stopped") {
      util_warning("prep_pmap called encapsulated",
                   applicability_problem = FALSE)
    }
    if (identical(cores, 0)) {
      suppressMessages(parallelMap::parallelStart(mode = "local"))
    } else if (inherits(cores, "list")) { # nocov start
      # Should not test in parallel
      suppressMessages(do.call(parallelMap::parallelStart, cores))
    } else {
      suppressMessages(parallelMap::parallelStart("socket", cpus = cores,
                                                  logging = FALSE,
                                                  load.balancing = TRUE))
      # Should not test in parallel
    } # nocov end
    on.exit({Sys.sleep(2);suppressMessages(parallelMap::parallelStop());}, add = TRUE) # whyever, rstudio needs these two seconds, it hangs, otherwise.
  }
  more_args <- list(...)
  ..f <- function(...) {
    return(.f(...))
  }
  do.call(
    parallelMap::parallelMap,
    c(.l, list(
      fun = ..f, more.args = more_args, simplify = FALSE, use.names = FALSE,
      show.info = FALSE, impute.error = identity
    ))
  )
}

test_that("prep_pmap works", {
  skip_on_cran() # deprecated
  df <- data.frame(
    x = c("apple", "banana", "cherry"),
    pattern = c("p", "n", "h"),
    replacement = c("P", "N", "H"),
    stringsAsFactors = FALSE
  )
  expect_equal(
    prep_pmap(df, gsub),
    list(
      "aPPle",
      "baNaNa",
      "cHerry"
    )
  )
  expect_error(
    prep_pmap(df, 42),
    regexp = "Argument .+\\.f.+ should be a function.",
    perl = TRUE
  )
})

#' Utility function to create plots for categorical variables
#'
#' Depending on the required level of complexity, this helper function creates
#' various plots for categorical variables. Next to basic bar plots, it also
#' enables group comparisons (for example for device/examiner effects) and
#' longitudinal views.
#'
#' @param resp_vars name of the categorical variable
#' @param group_vars name of the grouping variable
#' @param time_vars name of the time variable
#' @param study_data the data frame that contains the measurements
#' @param meta_data the data frame that contains metadata
#'                  attributes of study data
#' @param n_cat_max maximum number of categories to be displayed individually
#'                  for the categorical variable (`resp_vars`)
#' @param n_group_max maximum number of categories to be displayed individually
#'                  for the grouping variable (`group_vars`, devices / examiners)
#' @param n_data_min minimum number of data points to create a time course plot
#'                   for an individual category of the `resp_vars` variable
#'
#' @return a figure
#'
#' @noRd
util_plot_categorical_vars <- function(resp_vars,
                                   group_vars = NULL,
                                   time_vars = NULL,
                                   study_data,
                                   meta_data,
                                   n_cat_max =
                                     getOption("dataquieR.max_cat_resp_var_levels_in_plot",
                                               dataquieR.max_cat_resp_var_levels_in_plot_default),
                                   n_group_max =
                                     getOption("dataquieR.max_group_var_levels_in_plot",
                                               dataquieR.max_group_var_levels_in_plot_default),
                                   n_data_min =
                                     getOption("dataquieR.min_time_points_for_cat_resp_var",
                                               dataquieR.min_time_points_for_cat_resp_var_default)) {
  # preps and checks -----------------------------------------------------------
  util_expect_data_frame(study_data)
  util_expect_data_frame(meta_data)

  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata = TRUE)

  util_correct_variable_use(resp_vars,
                            allow_all_obs_na = FALSE,
                            need_scale = "nominal | ordinal")
  util_correct_variable_use(group_vars,
                            allow_null = TRUE,
                            allow_all_obs_na = FALSE,
                            need_type = "!float",
                            need_scale = "nominal | ordinal")
  util_correct_variable_use(time_vars,
                            allow_null = TRUE,
                            allow_all_obs_na = FALSE,
                            need_type = "datetime")

  util_expect_scalar(n_cat_max,
                     check_type = util_is_numeric_in(min = 2,
                                                     whole_num = TRUE,
                                                     finite = TRUE))
  util_expect_scalar(n_group_max,
                     check_type = util_is_numeric_in(min = 2,
                                                     whole_num = TRUE,
                                                     finite = TRUE))
  util_expect_scalar(n_data_min,
                     check_type = util_is_numeric_in(min = 0,
                                                     whole_num = TRUE,
                                                     finite = TRUE))
  N <- NULL # prevent warning 'no visible binding for global variable'

  plot_data <- ds1[, c(resp_vars, group_vars, time_vars), drop = FALSE]
  plot_data <- plot_data[complete.cases(plot_data), , drop = FALSE]
  tab_cat <- table(plot_data[[resp_vars]])
  # collapse 'rare' categories to reduce the number of categories, if needed
  if (length(tab_cat) > n_cat_max) {
    tab_cat <- tab_cat[order(tab_cat, decreasing = TRUE)]
    keep_cats <- names(tab_cat)[1:n_cat_max]
    levels(plot_data[[resp_vars]])[which(!levels(plot_data[[resp_vars]]) %in%
                                           keep_cats)] <- "other"
    # new category 'other' should always be the last one
    lvl_cat <-
      c(levels(plot_data[[resp_vars]])[which(levels(plot_data[[resp_vars]])
                                              %in% keep_cats)],
        "other")
    plot_data[[resp_vars]] <- as.character(plot_data[[resp_vars]])
    plot_data[[resp_vars]] <- factor(plot_data[[resp_vars]],
                                      levels = lvl_cat)
  }
  if (!is.null(group_vars)) {
    tab_gr <- table(plot_data[[group_vars]])
    # collapse 'rare' groups to reduce the number of levels, if needed
    if (length(tab_gr) > n_group_max) {
      tab_gr <- tab_gr[order(tab_gr, decreasing = TRUE)]
      keep_gr <- names(tab_gr)[1:n_group_max]
      levels(plot_data[[group_vars]])[which(!levels(plot_data[[group_vars]])
                                            %in% keep_gr)] <- "other"
      # new category 'other' should always be the last one
      lvl_gr <-
        c(levels(plot_data[[group_vars]])[which(levels(plot_data[[group_vars]])
                                                %in% keep_gr)],
          "other")
      plot_data[[group_vars]] <- as.character(plot_data[[group_vars]])
      plot_data[[group_vars]] <- factor(plot_data[[group_vars]],
                                        levels = lvl_gr)
    }
  }
  # omit categories with too few data points from the time course plot
  if (!is.null(time_vars)) {
    tab_cat <- table(plot_data[[resp_vars]])
    if (any(tab_cat) < n_data_min) {
      keep_cats <- names(tab_cat)[which(tab_cat >= n_data_min)]
      levels(plot_data[[resp_vars]])[which(!levels(plot_data[[resp_vars]]) %in%
                                             keep_cats)] <- NA
      plot_data <- plot_data[complete.cases(plot_data), , drop = FALSE]
    }
  }

  if (nrow(plot_data) == 0) {
    util_error("Not enough data to create a plot.",
               applicability_problem = FALSE)
  }

  # create figures -------------------------------------------------------------
  if (!is.null(group_vars) & !is.null(time_vars)) {

    # colors <- c("#E69F00", "#56B4E9", "#009E73", "#F0E442", "#0072B2",
    #             "#D55E00", "#CC79A7", "#8C510A")
    # n_gr <- length(levels(plot_data[[group_vars]]))
    # if (n_gr <= 8) {
    #   colors <- colors[1:n_gr]
    # } else {
    #   colors <- colorRampPalette(colors)(n_gr)
    # }
    pp <- util_histogram(plot_data = plot_data,
                         is_datetime = TRUE,
                         num_var = time_vars,
                         facet_var = resp_vars,
                         fill_var = group_vars)#,
                         #colors = colors)
  } else if (!is.null(time_vars)) {

    pp <- util_histogram(plot_data = plot_data,
                         is_datetime = TRUE,
                         num_var = time_vars,
                         facet_var = resp_vars)
  } else if (!is.null(group_vars)) {
    data_RepSumTab <- plot_data %>%
      dplyr::mutate(N = dplyr::n()) %>%
      dplyr::group_by(.data[[resp_vars]], .data[[group_vars]], N) %>%
      dplyr::summarise(freq = dplyr::n()) %>%
      dplyr::rename("Variables" = !!resp_vars) %>%
      as.data.frame(., check.names = FALSE)
    data_RepSumTab <- stats::reshape(data_RepSumTab, direction = "wide",
                                     idvar = c("Variables", "N"),
                                     timevar = group_vars)
    data_RepSumTab <- data_RepSumTab %>% replace(is.na(.), 0)
    colnames(data_RepSumTab) <- gsub("^freq\\.", "", colnames(data_RepSumTab))
    # TODO - PROBLEM: `print.ReportSummaryTable` drops the order of the levels
    # in 'Variables'. (I tried to reverse it, so that the first category is on
    # top, but because of the 'as.character' transformation I cannot change
    # the order of levels outside of `print.ReportSummaryTable`)
    # lvl_cat <- levels(data_RepSumTab$Variables)
    # data_RepSumTab$Variables <- as.character(data_RepSumTab$Variables)
    # data_RepSumTab$Variables <- factor(data_RepSumTab$Variables,
    #                                    levels = rev(lvl_cat))
    class(data_RepSumTab) <- c("ReportSummaryTable", "data.frame")
    pp <- print(data_RepSumTab, flip_mode = "flip", view = FALSE)
    # flipped because categories (here 'Variables') should be on the y-axis
    # here, only the figure is returned, so no chance for other visualizaions,
    # yet; i.e., we need to prevent the plot from being removed in
    # util_compress_ggplots_in_res()
    attr(pp, "from_ReportSummaryTable") <- NULL
    # attr(pp, "as_plotly") <- "util_as_plotly_util_plot_categorical_vars_gv"
  } else {
    plot_data <- plot_data %>%
      dplyr::group_by(.data[[resp_vars]]) %>%
      dplyr::summarise(freq = dplyr::n()) %>%
      as.data.frame()
    pp <- util_bar_plot(plot_data = plot_data,
                        cat_var = resp_vars,
                        num_var = "freq",
                        show_numbers = FALSE) #changed to FALSE to hide numbers
    attr(pp, "as_plotly") <- "util_as_plotly_util_plot_categorical_vars"
    attr(pp, "dont_util_adjust_geom_text_for_plotly") <- TRUE

    # Information for sizing
    obj1 <- ggplot2::ggplot_build(pp)
    obj1_data <- util_gg_get(obj1, "data")[[1]]
    rotated <- unique(obj1_data$flipped_aes)
    rotated <-  rotated[!is.na(rotated)]
    number_of_bars <- nrow(obj1_data)
    # fix number in case there are categories not listed in the obj1,
    # example OBS_BP_0 in ship example data
    if (is.factor(ds1[[resp_vars]])) {
      no_levels <- length(levels(ds1[[resp_vars]]))
      number_of_bars <- max(c(number_of_bars,no_levels))
    }
    min_bar_height <- min(obj1_data$ymax, na.rm = TRUE)
    max_bar_height <- max(obj1_data$ymax, na.rm = TRUE)
    range_bar <- max_bar_height - min_bar_height
    no_char_y = nchar(max_bar_height)
    no_char_x = max(nchar(as.character(plot_data[[resp_vars]])), na.rm = TRUE)


    attr(pp, "sizing_hints") <- list(
      figure_type_id = "bar_chart",
      rotated = rotated,
      number_of_bars = number_of_bars,
      range_bar = range_bar,
      no_char_y = no_char_y,
      no_char_x = no_char_x
    )



  }
  return(pp)
}

util_as_plotly_util_plot_categorical_vars <- function(res) {
  py <- util_ggplotly(res$SummaryPlot)
  pyb <- util_plotly_build(py)
  traces <- pyb$x$data
  # the color codes are defined in util_bar_plot
  black_bar_labs <-
     vapply(lapply(lapply(traces, `[[`, "textfont"), `[[`, "color"),
            identical, "rgba(0,0,0,1)", FUN.VALUE=logical(1))
  white_bar_labs <-
    vapply(lapply(lapply(traces, `[[`, "textfont"), `[[`, "color"),
            identical, "rgba(255,255,255,1)", FUN.VALUE=logical(1))
   py <- plotly::style(py, textposition = "bottom", traces = white_bar_labs)
  py <- plotly::style(py, textposition = "top", traces = black_bar_labs)
  py <- plotly::layout(py,
                       xaxis = list(tickangle = "auto"),
                       yaxis = list(tickangle = "auto"))
  py
}

#' Register an S3 method for an external generic
#'
#' This helper allows to register methods for generics defined in
#' suggested packages (e.g., plotly, patchwork) without creating
#' hard dependencies.
#'
#' @param generic character of the form "pkg::generic".
#' @param class character class name.
#' @param method function to register.
#'
#' @noRd
util_s3_register <- function(generic, class, method) { # nolint
  parts <- strsplit(generic, "::", fixed = TRUE)[[1L]]
  if (length(parts) != 2L) {
    util_error("`generic` must be of the form \"pkg::generic\".")
  }
  pkg <- parts[[1L]]
  gen <- parts[[2L]]

  if (!requireNamespace(pkg, quietly = TRUE)) {
    return(invisible(FALSE))
  }

  ## Wichtig: explizit utils::registerS3method() und positionsbasierte Argumente
  base::registerS3method(
    gen,
    class,
    method,
    envir = asNamespace(pkg)
  )

  invisible(TRUE)
}

#' Evaluate an expression in a clean environment
#'
#' This helper is meant to create tidy \code{ggplot2} objects with fewer
#' unneeded objects included. If \code{.lazy = TRUE}, a lazy ggplot
#' object is returned which will only be materialized when actually used.
#'
#' @param expr the expression that creates a plot.
#' @param ... all objects needed by \code{expr}, must be named.
#' @param .lazy logical, if \code{TRUE}, return a \code{dq_lazy_ggplot}
#'   instead of directly evaluating the expression.
#'
#' @return A \code{ggplot} or \code{dq_lazy_ggplot} object.
#'
#' @noRd
util_create_lean_ggplot <- function(expr, ..., .lazy = FALSE) { # nolint
  expr <- substitute(expr, environment())
  ee <- new.env(parent = environment(util_create_lean_ggplot))
  list2env(list(...), ee)

  if (.lazy) {
    dq_lazy_ggplot(expr, ee)
  } else {
    eval(expr, envir = ee)
  }
}

#' Lean ggplot2 composition operator
#'
#' This operator uses \code{util_create_lean_ggplot} to combine
#' \code{ggplot2} components in a clean environment. If the option
#' \code{dataquieR.lazy_plots} is set to \code{TRUE}, a lazy
#' \code{dq_lazy_ggplot} object is returned.
#'
#' @noRd
`%lean+%` <- function(lhs, rhs) {
  lazy <- as.logical(getOption("dataquieR.lazy_plots",
                               dataquieR.lazy_plots_default))
  if (length(lazy) != 1 || is.na(lazy)) {
    util_warning(c(
      "Cannot use option dataquieR.lazy_plots %s as a logical value",
      "using %s"
    ),
    dQuote(paste(getOption("dataquieR.lazy_plots",
                           dataquieR.lazy_plots_default),
                 collapse = ",")),
    dQuote(dataquieR.lazy_plots_default))
    lazy <- as.logical(dataquieR.lazy_plots_default)
  }

  util_create_lean_ggplot(
    {
      util_realize_if_lazy(lhs) + util_realize_if_lazy(rhs)
    },
    lhs = lhs,
    rhs = rhs,
    .lazy = lazy
  )
}


#' Explicit lazy lean ggplot2 composition operator
#'
#' This variant always returns a lazy \code{dq_lazy_ggplot} object.
#'
#' @noRd
`%lean_lazy+%` <- function(lhs, rhs) {
  util_create_lean_ggplot(
    {
      util_realize_if_lazy(lhs) + util_realize_if_lazy(rhs)
    },
    lhs = lhs,
    rhs = rhs,
    .lazy = TRUE
  )
}


# global switch by option, existing  code using %lean+% remains unchanged.


#' @export
print.dq_lazy_ggplot <- function(x, ...) { # nolint
  p <- prep_realize_ggplot(x)
  print(p, ...)
  invisible(x)
}

#' @export
`$.dq_lazy_ggplot` <- function(x, name) { # nolint
  p <- prep_realize_ggplot(x)
  p[[name]]
}

#' @export
`[[.dq_lazy_ggplot` <- function(x, ...) { # nolint
  p <- prep_realize_ggplot(x)
  `[[`(p, ...)
}

# ggplot_build & ggplotGrob (important ggplot2-inteactions,
# cowplot/gridExtra/egg/ggpubr)
#' @export
ggplot_build.dq_lazy_ggplot <- function(plot, ...) { # nolint
  p <- prep_realize_ggplot(plot)
  ggplot2::ggplot_build(p, ...)
}

as_grob.dq_lazy_ggplot_s7 <- function(plot, ...) { # nolint
  ggplot2::ggplotGrob(prep_realize_ggplot(plot@payload), ...)
}

as_grob.dq_lazy_ggplot <- function(plot, ...) { # nolint
  ggplot2::ggplotGrob(prep_realize_ggplot(plot), ...)
}

#' @rdname dq_lazy_ggplot_methods
#' @export
ggplotGrob.dq_lazy_ggplot <- function(x, ...) { # nolint
  p <- prep_realize_ggplot(x)
  ggplot2::ggplotGrob(p, ...)
}

# as soon as someone build upon the plot, it's okay,
# to create a normal ggplot. Laziness will be gone
#' @export
`+.dq_lazy_ggplot` <- function(e1, e2) { # nolint
  p <- prep_realize_ggplot(e1)
  if (inherits(e2, "dq_lazy_ggplot")) {
    e2 <- prep_realize_ggplot(e2)
  }
  p + e2
}

#' Check declared data types of metadata in study data
#'
#' @description
#' Checks data types of the study data and for the data type
#' declared in the metadata
#'
#' [Indicator]
#'
#' @details
#' This is a preparatory support function that compares study data with
#' associated metadata. A prerequisite of this function is that the no. of
#' columns in the study data complies with the no. of rows in the metadata.
#'
#' For each study variable, the function searches for its data type declared in
#' static metadata and returns a heatmap like matrix indicating data type
#' mismatches in the study data.
#'
#' List function.
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable] the names of the measurement variables, if
#'                             missing or `NULL`, all variables will be checked
#' @param split_segments [logical] return one matrix per study segment
#' @param max_vars_per_plot [integer] from=0. The maximum number of variables
#'                                            per single plot.
#' @param threshold_value [numeric] from=0 to=100. percentage failing
#'                                  conversions allowed to still classify a
#'                                  study variable convertible.
#' @return a list with:
#'   - `SummaryTable`: data frame containing data quality check for
#'                    "data type mismatch" (`CLS_int_vfe_type`,
#'                    `PCT_int_vfe_type`). The following categories are possible:
#'                    categories: "Non-matching datatype",
#'                    "Non-Matching datatype, convertible",
#'                    "Matching datatype"
#'   - `SummaryData`: data frame containing data quality check for
#'                    "data type mismatch" for a report
#'   - `SummaryPlot`: [ggplot2::ggplot2] heatmap plot, graphical representation of
#'                       `SummaryTable`
#'   - `DataTypePlotList`: [list] of plots per (maybe artificial) segment
#'   - `ReportSummaryTable`: data frame underlying `SummaryPlot`
#' @export
#' @importFrom ggplot2 ggplot aes geom_tile scale_fill_manual facet_wrap theme_minimal scale_x_discrete xlab guides guide_legend theme element_text
int_datatype_matrix <- function(resp_vars = NULL,
                                study_data,
                                label_col,
                                item_level = "item_level",
                                split_segments = FALSE,
                                max_vars_per_plot = 20,
                                threshold_value = 0,
                                # , flip_mode = "noflip" # TODO: Improve style
                                meta_data = item_level,
                                meta_data_v2) {

  if (.called_in_pipeline) {
    return(list(
      SummaryData =
        "Dummy result, should not be displayed, internal error, please report."))
  }

  util_maybe_load_meta_data_v2()

  if (length(max_vars_per_plot) != 1 || !util_is_integer(max_vars_per_plot) ||
      is.na(max_vars_per_plot) || is.nan(max_vars_per_plot) ||
      max_vars_per_plot < 1) {
    util_error(c(
      "max_vars_per_plot must be one strictly positive non-complex integer",
      "value, may be Inf."
    ), applicability_problem = TRUE)
  }

  if (missing(label_col)) {
    orig_label_col <- rlang::missing_arg()
  } else {
    orig_label_col <- force(label_col)
  }

  label_col <- attr(prep_get_labels("",
                                    item_level = meta_data,
                                    label_class = "SHORT",
                                    label_col = label_col),
                    "label_col")

  prep_prepare_dataframes(.replace_missings = FALSE, .adjust_data_type = FALSE)

  # correct variable use?
  util_correct_variable_use("resp_vars",
                            allow_more_than_one = TRUE,
                            allow_null          = TRUE,
                            allow_any_obs_na    = TRUE
  )

  # if no resp_vars specified use all columns of studydata
  if (length(resp_vars) == 0) {
    resp_vars <- names(ds1)
  } else {
    ds1 <- ds1[, resp_vars, drop = FALSE]
  }

  Variables <- resp_vars
  # this matrix will be shown to users
  app_matrix <- data.frame(Variables = Variables, stringsAsFactors = FALSE)
  # this matrix can be used to trigger computations
  # tri_matrix <- data.frame(Variables = Variables)
  # defined for merge of strata (optional)
  by_y <- label_col

  # DATA_TYPE defined in metadata?
  if (!(DATA_TYPE %in% names(meta_data))) {
    util_error(
      c("The attribute DATA_TYPE is not contained in the metadata but is",
        "required for this function."), applicability_problem = TRUE)
  }

  # variables with missing DATA_TYPE?
  if (any(is.na(meta_data$DATA_TYPE))) {
    whichnot <- as.character(meta_data[[label_col]][is.na(meta_data$DATA_TYPE)])
    util_error(paste0("The DATA_TYPE for variable(s) <<", whichnot,
                      ">> is not defined in the metadata."),
               applicability_problem = TRUE)
  }

  # check whether data types adhere to conventions
  if (!all(unique(meta_data$DATA_TYPE) %in% DATA_TYPES)) {
    whichnot <- dplyr::setdiff(unique(meta_data$DATA_TYPE), DATA_TYPES)
    util_warning(paste0("The data type(s): <<", whichnot,
                        ">> is not eligible in the metadata concept."),
                 applicability_problem = TRUE)
    util_error("Please map data types to: %s.", paste0(dQuote(DATA_TYPES),
                                                       collapse = ", "),
               applicability_problem = TRUE)
  }

  # DATA TYPE CONSISTENCY ------------------------------------------------------
  # This step compares the datatype as represented in study data with expected
  # datatypes in metadata and creates a binary vector indicating whether data
  # type fits or not
  dt_appl <-
             as.numeric(util_compare_meta_with_study(sdf = ds1, mdf = meta_data,
                                                     label_col = label_col,
                                                     check_convertible = TRUE,
                                                     check_conversion_stable = TRUE,
                                                     threshold_value =
                                                       threshold_value))
  dt_appl <- recode(dt_appl,
                    `1` = 0,
                    `2` = 1,
                    `0` = 2,
                    `3` = 3) # 3, convetible with drawbacks

  # two columns in heatmap: percentage missmatch, percentage non-convertible
  # percentage convertible?

  app_matrix$MATCH = dt_appl

  # SHOULD STRATIFICATION OF SEGMENTS BE USED? ---------------------------------

  # is STUDY_SEGMENT in metadata and none of the entries has NA?
  strata_defined <- STUDY_SEGMENT %in% names(meta_data)

  if (!strata_defined && split_segments) {
    util_warning(c(
      "Stratification for STUDY_SEGMENT is not possible due to missing",
      "metadata. Will split arbitrarily avoiding too large figures"
    ), applicability_problem = TRUE)
    nvars <- nrow(meta_data)
    meta_data$STUDY_SEGMENT <- paste0("Block #", ceiling(1:nvars /
                                                             max_vars_per_plot))
  } else if (strata_defined && split_segments) {
    if (any(is.na(meta_data$STUDY_SEGMENT))) {
      util_message(c(
        "Some STUDY_SEGMENT are NA. Will assign those to an artificial",
        "segment %s"), dQuote("Other"),
        applicability_problem = TRUE
      )
      meta_data$STUDY_SEGMENT[is.na(meta_data$STUDY_SEGMENT)] <- "Other"
    }
    too_big_blocks <- table(meta_data$STUDY_SEGMENT) > max_vars_per_plot
    too_big_blocks <- names(too_big_blocks)[too_big_blocks]
    for (too_big_block in too_big_blocks) {
      util_message(
        "Will split segemnt %s arbitrarily avoiding too large figures",
        dQuote(too_big_block),
        applicability_problem = FALSE
      )
      nvars <- sum(meta_data$STUDY_SEGMENT == too_big_block, na.rm = TRUE) # TODO: Use [[STUDY_SEGMENT]]
      meta_data$STUDY_SEGMENT[
        meta_data$STUDY_SEGMENT == too_big_block] <-
        paste0(
          meta_data$STUDY_SEGMENT[meta_data$STUDY_SEGMENT ==
                                        too_big_block],
          "#",
          ceiling(1:nvars / max_vars_per_plot)
        )
    }
  }

  if (!(STUDY_SEGMENT %in% names(meta_data))) {
    meta_data[[STUDY_SEGMENT]] <- "Study"
  }

  # merge relation to segments to app_matrix
  app_matrix <- merge(app_matrix, meta_data[, intersect(c(VAR_NAMES, LABEL,
                                                          STUDY_SEGMENT,
                                                          label_col),
                                                        colnames(meta_data)),
                                            FALSE],
    by.x = "Variables",
    by.y = by_y
  )
  app_matrix <- app_matrix[, !(names(app_matrix) %in% c(VAR_NAMES, LABEL,
                                                        label_col))]

  # PREPARE PLOT ---------------------------------------------------------------
  # reorder according VARIABLE_ORDER (optional)
  if (VARIABLE_ORDER %in% colnames(meta_data)) {
    meta_data <- meta_data[order(meta_data$VARIABLE_ORDER), ]
    meta_data[[VARIABLE_ORDER]][is.na(meta_data[[VARIABLE_ORDER]])] <-
      max(meta_data[[VARIABLE_ORDER]], na.rm = TRUE)
    meta_data[[VARIABLE_ORDER]][is.na(meta_data[[VARIABLE_ORDER]])] <-
      1
    app_matrix <- app_matrix[na.omit(match(meta_data[[label_col]],
                                     app_matrix$Variables)), ]
  }

  # assign factor levels
  app_matrix$Variables <- factor(app_matrix$Variables, levels =
                                   app_matrix$Variables)


  # reshape wide to long
#  app_matrix_long <- melt(app_matrix, id.vars = c("Variables",
#                                                  STUDY_SEGMENT))
  app_matrix_long <- stats::reshape(data = app_matrix,
                                     idvar = c("Variables",
                                               STUDY_SEGMENT),
                                     varying = colnames(app_matrix)[2],
                                     v.names = "value",
                                     times = colnames(app_matrix)[2],
                                     direction = "long")
  app_matrix_long$time <- as.factor(app_matrix_long$time)

  lev <- app_matrix$Variables
  app_matrix_long$Variables <- factor(app_matrix_long$Variables,
                                      levels = lev, ordered = FALSE)



  colnames(app_matrix_long) <- c("VARIABLES", "SEGMENT", "IMPLEMENTATION",
                                 "APP_SCORE")
  rownames(app_matrix_long) <- NULL


  # assign factor labels
  app_matrix_long$APP_SCORE <- factor(app_matrix_long$APP_SCORE,
    levels = c(2:0),
    ordered = TRUE,
    labels = c(
      "Non-matching datatype",
      "Non-Matching datatype, convertible",
      "Matching datatype"
    )
  )
  # PLOT -----------------------------------------------------------------------
  # colcode <- c("#B2182B", "#ef6548", "#92C5DE", "#2166AC", "#B0B0B0")
  colcode <- c("#B2182B", "#92C5DE", "#2166AC")
  names(colcode) <- levels(app_matrix_long$APP_SCORE)

  ratio <- dim(app_matrix)[1] / dim(app_matrix)[2]

  # ref_env <- environment() TODO

  plot_me <- function(m) {
    util_create_lean_ggplot(ggplot(m, aes(
      x = IMPLEMENTATION, y = VARIABLES,
      fill = APP_SCORE
    )) +
      geom_tile(colour = "white", linewidth = 0.8) + # https://github.com/tidyverse/ggplot2/issues/5051
      scale_fill_manual(values = colcode, name = " ") +
      {
        if (split_segments) facet_wrap(~SEGMENT, scales = "free_y")# TODO: check ~
      } +
      theme_minimal() +
      scale_x_discrete(position = "top") +
      xlab("") +
      guides(fill = guide_legend(
        ncol = 1, nrow = length(colcode),
        byrow = TRUE
      )) +
      theme(
        legend.position = "bottom",
        axis.text.x = element_text(angle = 90, hjust = 0),
        axis.text.y = element_text(size = 10)#,
       # aspect.ratio = ratio
      ), # + util_coord_flip(ref_env = ref_env) # TODO: estimate w and h, since p is not using discrete axes util_lazy_add_coord(p, fli)
  m = m,
  split_segments = split_segments,
  colcode = colcode)
    }

  p <- plot_me(app_matrix_long)

  pl <- lapply(
    split(app_matrix_long, app_matrix_long$SEGMENT),
    plot_me
  )

  # Figure size hint for plot p and list of plots pl
  attr(p, "sizing_hints") <- list(
    figure_type_id = "dot_mat",
    rotated = FALSE,
    number_of_vars = nrow(app_matrix_long) ,
    number_of_cat = 3) # this is always 3 types

  attr(pl, "sizing_hints") <- list(
    figure_type_id = "dot_mat",
    rotated = FALSE,
    number_of_vars = length(lapply(
      split(app_matrix_long, app_matrix_long$SEGMENT),
      nrow)), # a list with the no. variables per segment
    number_of_cat = 3) # this is always 3 types



  cmp_res <- util_compare_meta_with_study(sdf = ds1, mdf = meta_data,
                               label_col = label_col,
                               check_convertible = TRUE,
                               check_conversion_stable = TRUE,
                               return_percentages = TRUE,
                               threshold_value =
                                 threshold_value)
  ReportSummaryTable <- as.data.frame(t(cmp_res))
  wv <- attr(cmp_res, "which_vec")
  wv <- wv[vapply(wv, length, FUN.VALUE = integer(1)) > 0] # TODO: Why is this needed, example was from NAKO with a two-row table from /Users/struckmanns/tmp/NAKO_int_prob.RData
  FlaggedStudyData <- as.data.frame(wv,
                                    stringsAsFactors = FALSE)
  # FIXME: we do not need 4 bits, 2 would suffice
  # 0 -- match
  # 12 -- convertible mismatch, stable
  # 10 -- convertible mismatch, unstable
  # 9  -- nonconvertible mismatch
  FlaggedStudyData[] <- lapply(FlaggedStudyData,
                               factor,
                               ordered = TRUE,
                               levels = c(0, 12, 10, 9),
                               labels = c(
                                 "match",
                                 "convertible mismatch, stable",
                                 "convertible mismatch, unstable",
                                 "nonconvertible mismatch"
                                 ))

  #######
  match <- ReportSummaryTable$match
  ReportSummaryTable$match <- NULL
  ReportSummaryTable$convertible_mismatch_stable <-
    ReportSummaryTable$convertible_mismatch_stable / 100
  ReportSummaryTable$convertible_mismatch_unstable <-
    ReportSummaryTable$convertible_mismatch_unstable / 100
  ReportSummaryTable$nonconvertible_mismatch <-
    ReportSummaryTable$nonconvertible_mismatch / 100
  colnames(ReportSummaryTable)[colnames(ReportSummaryTable) ==
                                 "convertible_mismatch_stable"] <-
    "convertible mismatch, stable"
  colnames(ReportSummaryTable)[colnames(ReportSummaryTable) ==
                                 "convertible_mismatch_unstable"] <-
    "convertible mismatch, unstable"
  colnames(ReportSummaryTable)[colnames(ReportSummaryTable) ==
                                 "nonconvertible_mismatch"] <-
    "nonconvertible mismatch"

  ReportSummaryTable$N <- rep(1, nrow(ReportSummaryTable))
  ReportSummaryTable$Variables <- rownames(ReportSummaryTable)

  rownames(ReportSummaryTable) <- NULL

  attr(ReportSummaryTable, "continuous") <- TRUE
  attr(ReportSummaryTable, "colscale") <- c("#B2182B", "#92C5DE", "#2166AC")

  ReportSummaryTable <- util_validate_report_summary_table(ReportSummaryTable,
                                                    meta_data = meta_data,
                                                    label_col = label_col)

  app_matrix$PCT_int_vfe_type <-
    setNames(100 * ReportSummaryTable$`nonconvertible mismatch`,
             nm = ReportSummaryTable$Variables)[app_matrix$Variables]

  app_matrix$Variables <- as.character(app_matrix$Variables)

  # create SummaryData table
  SummaryData <- app_matrix[, c("Variables", "PCT_int_vfe_type")]

  SummaryData$PCT_int_vfe_type <-
    paste0(round(SummaryData$PCT_int_vfe_type * nrow(ds1) / 100, digits = 0),
           " (",
           base::format(SummaryData$PCT_int_vfe_type, digits = 2, nsmall = 2), "%)")

  names(SummaryData)[names(SummaryData) == "PCT_int_vfe_type"] <-
    "Data type mismatch N (%)"

#  names(SummaryData)[names(SummaryData) == "PCT_int_vfe_type"] <-
#   util_translate_indicator_metrics("PCT_int_vfe_type",
#                                    short = FALSE, long = FALSE)

  SummaryData$`Convertible mismatch, stable N (%)` <-
    paste0( round(ReportSummaryTable$`convertible mismatch, stable` * nrow(ds1),
                  digits = 0), " (",
            base::format(100 * ReportSummaryTable$`convertible mismatch, stable`, digits = 2, nsmall = 2),
            "%)")
  SummaryData$`Convertible mismatch, unstable N (%)`<-
    paste0(round(ReportSummaryTable$`convertible mismatch, unstable` * nrow(ds1),
                 digits = 0), " (",
           base::format(100 * ReportSummaryTable$`convertible mismatch, unstable`, digits = 2, nsmall = 2),
           "%)")

  SummaryData$`Data type match N (%)` <-
    paste0(round(match / 100 * nrow(ds1), digits = 0), " (",
           base::format(match, digits = 2, nsmall = 2), "%)" )

  SummaryData$`Expected DATA_TYPE` <- setNames(meta_data[[DATA_TYPE]],
                        nm = meta_data[[label_col]])[SummaryData$Variables]

  SummaryData$`Observed DATA_TYPE` <-
    setNames(prep_datatype_from_data(resp_vars = # TODO: Improve
                                       SummaryData$Variables,
                                     study_data = ds1),
             nm = colnames(ds1))[SummaryData$Variables]

  SummaryData <- cbind(SummaryData,
                             app_matrix[, c("MATCH", "STUDY_SEGMENT"),
                                        drop = FALSE])

  SummaryData$MATCH <- factor(
    SummaryData$MATCH,
    levels = c(3:0),
    ordered = TRUE,
    labels = c(
      "Non-matching datatype, convertible, unstable",
      "Non-matching datatype",
      "Non-Matching datatype, convertible, stable",
      "Matching datatype"
    )
  )

  colnames(SummaryData)[colnames(SummaryData) == "MATCH"] <-
    "State, given threshold"

  #Add hover text to SummaryData table
  text_to_display <- util_get_hovertext("[int_datatype_matrix_hover]")
  attr(SummaryData, "description") <- text_to_display



  SummaryTable <- app_matrix
  SummaryTable$GRADING <- SummaryTable$MATCH == 2
  names(SummaryTable)[names(SummaryTable) == "MATCH"] <- "CLS_int_vfe_type"

  ReportSummaryTable[,
                     c("convertible mismatch, stable",
                       "convertible mismatch, unstable",
                       "nonconvertible mismatch",
                       "N")] <-
    round(ReportSummaryTable[, c("convertible mismatch, stable",
                           "convertible mismatch, unstable",
                           "nonconvertible mismatch",
                           "N")] / ReportSummaryTable$N * nrow(ds1), digits = 0)
  attr(ReportSummaryTable, "relative") <- TRUE

  if (!missing(orig_label_col)) { # map back ReportSummaryTable to requested label_col and tables to suitable table labels
    SummaryTable[["Variables"]] <-
      prep_map_labels(x = SummaryTable[["Variables"]],
                      item_level = meta_data,
                      to  = orig_label_col,
                      from = label_col,
                      ifnotfound = SummaryTable[["Variables"]],
                      warn_ambiguous = FALSE)
    ReportSummaryTable[["Variables"]] <-
      prep_map_labels(x = ReportSummaryTable[["Variables"]],
                      item_level = meta_data,
                      to  = orig_label_col,
                      from = label_col,
                      ifnotfound = SummaryTable[["Variables"]],
                      warn_ambiguous = FALSE)
    SummaryData[["Variables"]] <-
      prep_map_labels(x = SummaryData[["Variables"]],
                      item_level = meta_data,
                      to  = orig_label_col,
                      from = label_col,
                      ifnotfound = SummaryTable[["Variables"]],
                      warn_ambiguous = FALSE)
  }

  attr(ReportSummaryTable, "flip_mode") <- "noflip";
  #Fix size for data_type matrix plot
  attr(ReportSummaryTable, "size_info") <- "fix_size";

  return(list( # FIXME: Add Flagged- and Modified-StudyData
    SummaryPlot = p,
    DataTypePlotList = pl,
    SummaryTable = SummaryTable,
    SummaryData = SummaryData,
    ReportSummaryTable = ReportSummaryTable,
    FlaggedStudyData = FlaggedStudyData
  ))
}

#' print implementation for the class `ReportSummaryTable`
#'
#' @description
#' Use this function to print results objects of the class
#' `ReportSummaryTable`.
#'
#' @param relative deprecated
#' @param dt [logical] use `DT::datatables`, if installed
#' @param fillContainer [logical] if `dt` is `TRUE`, control table size,
#'        see `DT::datatables`.
#' @param displayValues [logical] if `dt` is `TRUE`, also display the actual
#'                                values
#' @param view [logical] if `view` is `FALSE`, do not print but return the
#'                       output, only
#' @param drop [logical] if `drop` is `FALSE`, keep unused levels,
#'                       see [dataquieR.droplevels_ReportSummaryTable]
#' @param x `ReportSummaryTable` objects to print
#' @inheritParams acc_distributions
#' @param ... not used, yet
#'
#' @seealso base::print
#' @importFrom grDevices colorRamp rgb col2rgb
#' @importFrom ggplot2 expansion waiver scale_color_gradientn
#' @export
#' @return the printed object
print.ReportSummaryTable <- function(x, relative = lifecycle::deprecated(),
                                     dt = FALSE,
                                     fillContainer = FALSE,
                                     displayValues = FALSE,
                                     view = TRUE,
                                     drop =
                                       getOption(
                                 "dataquieR.droplevels_ReportSummaryTable",
                                 dataquieR.droplevels_ReportSummaryTable_default
                                         ),
                                     ...,
                                     flip_mode = "auto") {

  if (lifecycle::is_present(relative)) {
    # Signal the deprecation to the user
    lifecycle::deprecate_warn(
      "2.5.0",
      "dataquieR::print.ReportSummaryTable(relative = )")
  }

  util_expect_scalar(drop,
                     check_type = is.logical,
                     error_message = "drop needs to be either TRUE or FALSE")

  if (drop) {
    x <- droplevels(x)
  }

  #definition of colscale
  colscale <- c("#B2182B", "#92C5DE", "#2166AC")


  empty <-
    (!length(setdiff(colnames(x), c("Variables", "N")))) ||
    (!c("Variables") %in% colnames(x)) ||
    (!c("N") %in% colnames(x))
  if (empty) {
    if (!dt) {
      x <- ggplot() +
        annotate("text", x = 0, y = 0, label = "Empty result.") +
        theme(
          axis.line = element_blank(),
          axis.text.x = element_blank(),
          axis.text.y = element_blank(),
          axis.ticks = element_blank(),
          axis.title.x = element_blank(),
          axis.title.y = element_blank(),
          legend.position = "none",
          panel.background = element_blank(),
          panel.border = element_blank(),
          panel.grid.major = element_blank(),
          panel.grid.minor = element_blank(),
          plot.background = element_blank()
        )
      x <- util_set_size(x)
    } else {
      x <- htmltools::HTML("")
    }
    attr(x, "warning") <- "Empty result"
    class(x) <- union("dataquieR_result", class(x))
    if (view) {
      print(x)
    }
    attr(x, "from_ReportSummaryTable") <- TRUE
    return(invisible(x))
  }

  mfm <- missing(flip_mode)
  if (!mfm) oldfm <- flip_mode

  list2env(util_init_respum_tab(x), envir = environment())

  if (!mfm) flip_mode <- oldfm

  x <- util_validate_report_summary_table(x)

  hm <- x
  if (relative) {
    hm <- cbind.data.frame(Variables = hm$Variables,
                           hm[, setdiff(colnames(hm), c("Variables", "N")),
                              drop = FALSE] /
                             hm$N)
  } else {
    hm <- cbind.data.frame(Variables = hm$Variables,
                           hm[, setdiff(colnames(hm), c("Variables", "N")),
                              drop = FALSE])
  }

  # plot usual ggplot
  # ggplot(ds2, aes(x=value)) + facet_wrap(variable ~ ., ncol = 1) +
  # geom_bar() + theme_minimal() +
  #  theme(axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5))

  #p <- util_set_size(p,
  #                   width_em  = length(unique(ctab1$Var2)) + 20,
  #                   height_em = length(unique(ctab1$Var1)) + 5)


  if (prod(dim(hm)) == 0) { # no output, # TODO: is this still possible after empty checks above?
    if (dt) {
      util_ensure_suggested("DT", "the option dt = TRUE")
      w <- DT::datatable(data.frame())
      if (view) print(w)
      attr(w, "from_ReportSummaryTable") <- TRUE
      return(invisible(x))
    } else {
      p <- ggplot()
      if (view) print(p)
      attr(p, "from_ReportSummaryTable") <- TRUE
      return(invisible(x))
    }
  }

  #tb <- reshape::melt(hm, id.vars = "Variables")
  tb <- stats::reshape(data = hm, idvar = "Variables",
                       varying = colnames(hm)[2:ncol(hm)],
                       v.names = "value",
                       times = colnames(hm)[2:ncol(hm)],
                       direction = "long")
  rownames(tb) <- NULL
  names(tb) <- c("Variables", "variable", "value")

  if (all(is.na(tb$value))) {
    tb <- tb[FALSE, , FALSE]
  } else {
    tb <- tb[!is.na(tb$value), , FALSE]
  }

  levs <- unique(tb$variable)

  if(length(levs[grep("int_", levs)]) == 0  &&
     length(levs[grep("com_", levs)]) == 0  &&
     length(levs[grep("con_", levs)]) == 0  &&
     length(levs[grep("acc_", levs)]) == 0 ) {

    tb$variable <- factor(tb$variable,
                          levels = levs)
  } else {

  #sort levels
  #levs <- sort(levs)
  levs_int <- levs[grep("int_", levs)]
  levs_com <- levs[grep("com_", levs)]
  levs_con <- levs[grep("con_", levs)]
  levs_acc <- levs[grep("acc_", levs)]

  #order the factor levels
  tb$variable <- factor(tb$variable,
                                 levels = c(levs_int,levs_com,
                                            levs_con, levs_acc))
  }


  if (dt) {
    # https://stackoverflow.com/a/50406895
    util_ensure_suggested("DT", "the option dt = TRUE")

    # if (!relative) {
    #   hm[, setdiff(colnames(hm), c("Variables", "N"))] <-
    #     hm[, setdiff(colnames(hm), c("Variables", "N")),
    #            drop = FALSE] / max(as.matrix(
    #              hm[, setdiff(colnames(hm), c("Variables", "N")),
    #                 drop = FALSE]
    #            ), na.rm = TRUE)
    # }
    mx <- max(as.matrix(
                    hm[, setdiff(colnames(hm), c("Variables", "N")),
                       drop = FALSE]
                  ), na.rm = TRUE)
    colr <- colorRamp(colors = rev(colscale))
    colors_of_hm <- lapply(hm, function(values) {
      if (!all(is.numeric(values))) { return(values) }
      if (any(is.na(values))) { return(values) }
      if (continuous) {
        if (!relative) {
          v <- colr(values / mx)
        } else {
          v <- colr(values)
        }
      } else {
        if (length(level_names) > 0) {
          cc <- setNames(colcode, nm = level_names)
          v <- t(col2rgb(cc[level_names[as.character(values)]], alpha = TRUE))
        } else {
          v <- t(col2rgb(colcode[as.character(values)], alpha = TRUE))
        }
      }
      a <- apply(v, 1, function(cl) {
        paste0(
          "<span style=\"width:100%;display:block;text-align:center;",
          "color:",
          rgb(255 - cl[[1]],
              255 - cl[[2]],
              255 - cl[[3]], maxColorValue = 255.0),
          ";",
          "overflow:hidden;background:",
          rgb(cl[[1]],
            cl[[2]],
            cl[[3]], maxColorValue = 255.0),
          "\" title=\""
        )
      })
      b <- apply(v, 1, function(cl) {
        paste0(
          "\" sort=\""
        )
      })
      cc <- apply(v, 1, function(cl) {
        paste0(
          "\">"
        )
      })
      d <- apply(v, 1, function(cl) {
        paste0(
          "</span>"
        )
      })
      if (displayValues) {
        if (relative) {
          dv <- paste0(round(100 * values, 0), "%")
        } else {
          dv <- values
        }
      } else {
        dv <- "&nbsp;"
      }
      if (relative) {
        paste0(a, round(100 * values, 1), "%", b, values, cc, dv, d)
      } else {
        if (length(level_names) > 0)
          hover <- level_names[as.character(values)]
        else
          hover <- round(values, 1)
        paste0(a, hover, b, values, cc, dv, d)
      }
    })
    x[, names(colors_of_hm)] <- colors_of_hm

    # https://www.pierrerebours.com/2017/09/custom-sorting-with-dt.html
    # https://datatables.net/manual/data/orthogonal-data
    # filter = "top" is not helpful
    w <- DT::datatable(x, # TODO: Add Buttons Extension
                       fillContainer = fillContainer,
                       rownames = FALSE,
                       options = list(
                         pageLength = nrow(x),
                         columnDefs = list(
                             list(
                               targets = seq_len(ncol(x)-2),
                               render = DT::JS("sort_heatmap_dt")
                             )
                         )
                         ),
                       class = "ReportSummaryTable",
                  colnames = paste("<div class=\"colheader\">",
                                   vapply(
                                     strsplit(
                                       colnames(x),
                                       "", fixed = TRUE),
                                     function(letters) {
                                       paste0("<span>",
                                              paste0(letters,
                                                     collapse = ""),
                                              "</span>")
                                      },
                                     FUN.VALUE = character(1)),
                                   "</div>"),
                  escape = FALSE
    )


    # https://stackoverflow.com/a/35775262
    w$dependencies <- c(
      w$dependencies,
      list(html_dependency_vert_dt())
    )

    if (view) print(w)

    attr(w, "from_ReportSummaryTable") <- TRUE

    return(invisible(w))

    # https://stackoverflow.com/a/46043032
  } else {# https://stackoverflow.com/a/64112567
    if (continuous &&
        (length(unique(tb$variable)) == 1 ||
        length(unique(tb$Variables)) == 1)) { # if only one dimension and real numbers, not categories, collapse the heatmap to a barchart
      if (length(unique(tb$Variables)) == 1) { # only one value in y direction
        x <- "variable"
        y <- "value"
        fill <- "value"
        if (relative) {
          if (suppressWarnings(max(tb$value, na.rm = TRUE)) > 0.1)
            add_amount <- 0.04
          else
            add_amount <- 0.0004
          rel_ax <- scale_y_continuous(labels = scales::percent,
                                       expand = expansion(add = c(0,
                                                                  add_amount)))
        } else {
          rel_ax <- NULL
        }
      } else {
        x <- "Variables"
        y <- "value"
        fill <- "value"
        if (relative) {
          if (suppressWarnings(max(tb$value, na.rm = TRUE)) > 0.1)
            add_amount <- 0.04
          else
            add_amount <- 0.0004
          rel_ax <- scale_y_continuous(labels = scales::percent,
                                       expand = expansion(add = c(0,
                                                                  add_amount)))
        } else {
          rel_ax <- NULL
        }
      }

      if (missing(flip_mode) && getOption("dataquieR.flip_mode",
                                          dataquieR.flip_mode_default) ==
          dataquieR.flip_mode_default) {
        # for bar charts, flip mode defaults to default (noflip)
        # fli <- coord_cartesian();
        fli <- coord_flip()
      } else {
        fli <- util_coord_flip(w = length(unique(tb[[x]])),
                               h = length(unique(tb[[y]])))
      }

      is_flipped <- inherits(fli, "CoordFlip")

      if (is_flipped) {
        hjust <- 0
        vjust <- 0
      } else {
        hjust <- 0.5
        vjust <- 0
      }
      tb_copy <- tb
      tb_copy <- tb_copy[is.finite(tb_copy$value), , drop=FALSE]
      if (nrow(tb_copy) == 0) {
        p <- ggplot() +
          annotate("text", x = 0, y = 0, label = "Empty result.") +
          theme(
            axis.line = element_blank(),
            axis.text.x = element_blank(),
            axis.text.y = element_blank(),
            axis.ticks = element_blank(),
            axis.title.x = element_blank(),
            axis.title.y = element_blank(),
            legend.position = "none",
            panel.background = element_blank(),
            panel.border = element_blank(),
            panel.grid.major = element_blank(),
            panel.grid.minor = element_blank(),
            plot.background = element_blank()
          )
        p <- util_set_size(p)
        gtlb <- character(0)
      } else {
        if (relative) {
          scale_fill <- scale_fill_gradientn(colors = rev(colscale),
                                             labels = scales::percent)
          gtlb <- paste0(" ", round(tb_copy$value * 100, digits = 2), "%")
          texts <-
            util_create_lean_ggplot(geom_text(label = gtlb,
                      hjust = hjust, vjust = vjust, size = 3.5),
                      gtlb = gtlb,
                      hjust = hjust, vjust = vjust
            )
        } else {
          scale_fill <- scale_fill_gradientn(colors = rev(colscale))
          gtlb <- paste0(" ", round(tb_copy$value, digits = 2))
          texts <-
            util_create_lean_ggplot(geom_text(label = gtlb,
                      hjust = hjust, vjust = vjust, size = 3.5),
                      gtlb = gtlb,
                      hjust = hjust, vjust = vjust
            )
        }

        p <- util_create_lean_ggplot(ggplot(tb_copy, aes(
          x = .data[[x]], y = .data[[y]],
          fill = .data[[fill]]
        )) + geom_bar(stat = "identity", na.rm = TRUE,
                      colour = "white", linewidth = 0.8) + # https://github.com/tidyverse/ggplot2/issues/5051
          theme_minimal() +
          texts +
          fli +
          scale_fill +
          xlab("") +
          guides(fill = guide_legend(
            title = ""
            #   ncol = 1, nrow = length(colcode),
            #   byrow = TRUE
          )) +
          rel_ax +
          theme(
            legend.position = "bottom",
            axis.text.x = element_text(angle = 90, hjust = 0),
            axis.text.y = element_text(size = 10)
          ) + xlab("") + ylab(""),
          tb_copy = tb_copy,
          x = x,
          y = y,
          fill = fill,
          texts = texts,
          fli = fli,
          scale_fill = scale_fill,
          rel_ax = rel_ax
        )
      }


      if (suppressWarnings(util_ensure_suggested("plotly",
                                      goal = "generate interactive plots",
                                      err = FALSE))) {
        attr(p, "py") <- local({
          tb_copy <- tb
          tb_copy <- tb_copy[is.finite(tb_copy$value), , drop=FALSE]
          py <- plotly::plot_ly(tb_copy,
                  x = tb_copy[[ifelse(is_flipped, y, x)]],
                  y = tb_copy[[ifelse(is_flipped, x, y)]],
                  type = 'bar',
                  marker = list(color = tb_copy[[fill]],
                                autocolorscale = FALSE,
                                colorscale = colscale))
          gtlb_copy <- gtlb
          gtlb_copy <- gtlb_copy[is.finite(tb$value)]
          py <- plotly::style(py,
                              text = gtlb_copy,
                              textposition = "auto")

          if (relative) {
            if (is_flipped) {
              py <- plotly::layout(py, xaxis = list(tickformat = ".2%",
                                                    rangemode="tozero"))
            } else {
              py <- plotly::layout(py, yaxis = list(tickformat = ".2%",
                                                    rangemode="tozero"))
            }
          }

          # %>%
          #   layout(xaxis = list(title = 'Parameter Name (Unit)',
          #                       range = c(0, 1), tickvals = seq(0, 1, 0.2)),
          #          yaxis = list(title = 'Sample ID', categoryorder = "total ascending"))
          #
          #
          # p <- ggplot(tb, aes(
          #   x = .data[[x]], y = .data[[y]],
          #   #fill = .data[[fill]]
          # )) + geom_bar(stat = "identity", na.rm = TRUE,
          #               colour = "white", linewidth = 0.8) + # https://github.com/tidyverse/ggplot2/issues/5051
          #   theme_minimal() +
          #   fli +
          #   #scale_fill +
          #   xlab("") +
          #   guides(fill = guide_legend(
          #     title = ""
          #   )) +
          #   rel_ax +
          #   theme(
          #     legend.position = "bottom",
          #     axis.text.x = element_text(angle = 90, hjust = 0),
          #     axis.text.y = element_text(size = 10)
          #   ) + xlab("") + ylab("")
          # py <- util_plot_figure_plotly(p)
          # for (i in seq_along(gtlb)) {
          #   py <- plotly::style(py,
          #                       traces = i,
          #                       text = gtlb[[i]],
          #                       textposition = "auto")
          # }
          py
        })
      }

      if (relative) {
        fct <- 100
      } else {
        fct <- 1
      }

      attr(p, "sizing_hints") <- list(
        figure_type_id = "bar_chart",
        rotated = is_flipped,
        number_of_bars = ifelse(is.null(nrow(util_gg_get(p, "data"))), 0,
                                nrow(util_gg_get(p, "data"))),
        range = (fct * suppressWarnings(max(util_gg_get(p, "data")[[y]],
                                            na.rm = TRUE)))-
          (fct * suppressWarnings(min(util_gg_get(p, "data")[[y]],
                                      na.rm = TRUE))),
        no_char_vars = suppressWarnings(max(nchar(
          as.character(util_gg_get(p, "data")$Variables)), na.rm = TRUE)),
        no_char_numbers = suppressWarnings(max(nchar(
          util_gg_get(p, "data")$value), na.rm = TRUE)) #max no. numbers for tick labels
      )


      if (view) print(p)
      attr(p, "from_ReportSummaryTable") <- TRUE
      return(invisible(p))

    } else {

      if (continuous) {
        xlim <- as.character(tb$Variables)
        xlim <- xlim[!duplicated(xlim)]
        ylim <- as.character(tb$variable)
      #  xsize <- util_plotly_font_size(ncol(hm) - 1)
      #  ysize <- util_plotly_font_size(nrow(hm), space = 150)

      #  if(xsize != ysize){
       #   min_size <- min(xsize, ysize)
       #   xsize <- min_size
       #   ysize <- min_size
       # }
        xsize <- 10
        ysize <- 10

        p <- util_create_lean_ggplot(ggplot(tb, aes(x = Variables, y = variable, colour = value, size = value)) +
          geom_point()  + #scale_size_continuous(range = c(-1, 10)) + scale_x_discrete() +# breaks = 50 * seq_len(length(unique(tb$Variables)))) +
          #scale_x_discrete(guide = guide_axis(n.dodge = 5)) +
          labs(title = waiver(),
               subtitle = waiver(),
               x = "",
               y = "") +
          scale_x_discrete(limits = xlim) +
          # (if (nrow(hm) < ncol(hm)) coord_flip()) + # 7x5 standard of rmarkdown, not din a4
          scale_color_gradientn(colors = rev(colscale)) +
          theme_minimal() +
          theme(#aspect.ratio=5*length(unique(tb$Variables))/7/length(unique(tb$Variables)),
                axis.text.x = element_text(angle = 35, hjust = 1,
                                           size = xsize),
                axis.text.y = element_text(size = ysize)),
          tb = tb,
          xlim = xlim,
          colscale = colscale,
          xsize = xsize,
          ysize = ysize
        )


        no_char_vars <- max(nchar(as.character(tb$Variables)))
        no_char_cat <-  max(nchar(as.character(tb$variable)))


        fli <- util_coord_flip(p = p, w = nrow(hm), h = ncol(hm))
        is_flipped <- inherits(fli, "CoordFlip")
        p <- util_lazy_add_coord(p, fli)

        p <- util_set_size(p, 500, 300)

        if (view) print(p)

        attr(p, "from_ReportSummaryTable") <- TRUE

        if (!is.null(attr(x, "size_info")) &&
            attr(x, "size_info") == "fix_size") {
          attr(p, "sizing_hints") <- list(
            figure_type_id = "dot_mat_fix",
            rotated = FALSE,
            number_of_vars = nrow(x),
            number_of_cat =  3,
            no_char_vars = no_char_vars,
            no_char_cat = no_char_cat
          )
        } else {
          attr(p, "sizing_hints") <- list(
            figure_type_id = "dot_mat",
            rotated = is_flipped,
            number_of_vars = nrow(x),
            number_of_cat =  ncol(x)-2,
            no_char_vars = no_char_vars,
            no_char_cat = no_char_cat
          )
        }



        return(invisible(p))

      } else {
        tb$value <- as.factor(tb$value)

        if (length(level_names) > 0) {
          levels(tb$value) <- level_names[levels(tb$value)]
          cc <- setNames(colcode, nm = level_names)
        } else {
          cc <- colcode
        }

        #    colcode <- c("#B2182B", "#ef6548", "#92C5DE", "#2166AC", "#B0B0B0")
        #    names(colcode) <- levels(tb$value)
        fli <- util_coord_flip(w = nrow(hm), h = ncol(hm))
        lcolcode <- length(colcode)
        p <- util_create_lean_ggplot(ggplot(tb, aes(
          x = variable, y = Variables,
          fill = value
        )) + geom_tile(colour = "white", linewidth = 0.8) + # https://github.com/tidyverse/ggplot2/issues/5051
          theme_minimal() +
          # (if (nrow(hm) > ncol(hm)) coord_flip()) +
          fli +
          scale_fill_manual(values = cc, name = " ") +
          #      scale_fill_gradientn(colors = rev(colscale)) +
          #      scale_x_discrete(position = "top") +
          xlab("") +
          ylab("") +
          guides(fill = guide_legend(
            ncol = 1, nrow = lcolcode,
            byrow = TRUE
          )) +
          theme(
            legend.position = "bottom",
            axis.text.x = element_text(angle = 90, hjust = 0, size = 10),
            axis.text.y = element_text(size = 10)
          ),
          tb = tb,
          fli = fli,
          cc = cc,
          lcolcode = lcolcode
        )
        if (view) print(p)

        attr(p, "from_ReportSummaryTable") <- TRUE

        return(invisible(x))
      }
    }
  }
}

#' @exportS3Method knitr::knit_print
knit_print.ReportSummaryTable <- print.ReportSummaryTable

#' Create an environment with several alias names for the study data variables
#'
#' generates an environment similar to `as.environment(ds1)`, but makes
#' variables available by their `VAR_NAME`, `LABEL`, and `label_col` - names.
#'
#' @param study_data [data.frame] the data frame that contains the measurements
#' @param meta_data [data.frame] the data frame that contains metadata
#'                               attributes of study data
#' @param label_col [variable attribute] the name of the column in the metadata
#'                                       with labels of variables. If
#'                                       `study_data` has already been mapped,
#'                                       i.e., `util_ds1_eval_env(ds1, ...)` is
#'                                       called, this will work too
#'
#' @family rule_functions
#' @concept process
#' @noRd
util_ds1_eval_env <- function(study_data,
                              meta_data = "item_level",
                              label_col = LABEL) {
  if (isTRUE(attr(study_data, "MAPPED"))) {
    ds1 <- study_data
  } else {
    prep_prepare_dataframes()
  }
  label_col_from <- attr(ds1, "label_col")
  label_col_to <- label_col
  res <- ds1
  lct <- setdiff(c(VAR_NAMES, LABEL, LONG_LABEL, label_col_to), label_col_from)
  lct <- intersect(lct, colnames(meta_data))
  for (cur_nm in lct) {
    res[, util_map_labels(colnames(ds1),
                          from = label_col_from,
                          to = cur_nm,
                          meta_data = meta_data)] <-
      ds1[, colnames(ds1), FALSE]
  }
  as.environment(res)
}

#' return a single page navigation menu floating on the right
#'
#' if displayed in a `dq_report2`
#'
#' @param index_menu_table [data.frame] columns: links, hovers, texts
#' @param object `htmltools` tag list, used, instead of `index_menu_table`,
#'               if passed
#'
#' @examples
#' \dontrun{
#' util_float_index_menu(tibble::tribble(
#'    ~ links, ~ hovers, ~ texts,
#'    "http://www.google.de/#xxx", "This is Google", "to Google",
#'    "http://www.uni-giessen.de/#xxx", "This is Gießen", "cruising on the A45"
#' ))
#' }
#'
#'
#' @family reporting_functions
#' @concept html
#' @noRd
util_float_index_menu <- function(index_menu_table, object) {
  if (missing(object)) {
    object <- do.call(htmltools::tagList,
                      mapply(SIMPLIFY = FALSE,
                             link = index_menu_table$links,
                             hover = index_menu_table$hovers,
                             text = index_menu_table$texts,
                             FUN = function(link, hover, text) {
                               htmltools::tags$li(
                                 htmltools::a(href = link, title = hover, text)
                               )
                             })
    )
  } else {
    util_stop_if_not(missing(index_menu_table))
  }
  if (!!length(object)) {
    htmltools::div(
      class = "floatbar",
      htmltools::tags$i(class = "fas fa-list-ul",
                        title = "Jump to section"),
      htmltools::tags$ul(
        class = "floatmenu",
        object
      )
    )
  } else {
    NULL
  }
}

#' Suppress any output to `stdout` using [sink()]
#'
#' @param expr [expression()]
#' @param warns [environment()] if given, warnings will be stored in this
#'                              environment in a newly initialized object
#'                              named `warns`.
#'
#' @return [invisible()] result of `expr`
#'
#' @family process_functions
#' @noRd
util_suppress_output <- function(expr, warns) {
  if (missing(warns)) warns <- new.env(parent = emptyenv())
  warns$warns <- list()
  sink(type = "output", file = nullfile())
  on.exit((function(){sink()})())
  withCallingHandlers(
    invisible(force(eval(expr, envir = parent.frame()))),
    warning = function(w) {
      warns$warns <- c(warns$warns, list(w))
      invokeRestart("muffleWarning")
    },
    message = function(m) {
      invokeRestart("muffleMessage")
    }
  )
}

#' Get a translation
#'
#' @param keys [character] translation keys
#' @param ns [character] translation namespace
#' @param lang [character] language to translate to
#' @param as_this_translation `dataquieR_translated` use this as template and
#'                            translate the keys as this one.
#' @param reverse [logical] translate backwards
#'
#' @return [character] translations
#'
#' @family string_functions
#' @concept reporting
#' @noRd
util_translate <- function(
    keys,
    ns = "general",
    lang = getOption("dataquieR.lang", dataquieR.lang_default),
    as_this_translation = NULL,
    reverse = FALSE) {

  if (!missing(as_this_translation)) {
    util_stop_if_not(inherits(as_this_translation, "dataquieR_translated"))
    if (!missing(ns) || !missing(lang)) {
      util_error(
        c("%s can called with %s and/or %s *or* with %s, but not both.",
        "this is an internal error, sorry. please report."),
        sQuote("util_translate"),
        sQuote("ns"), sQuote("lang"), sQuote("as_this_translation"))
    }
    ns <- attr(as_this_translation, "ns")
    lang <- attr(as_this_translation, "lang")
  }

  # Hint: the language may differ at rendering time from the language at computation time.
  translations <-
    util_get_concept_info("translations")
  values <- vapply(keys, function(key) {
    if (reverse) {
      match <-
        translations$namespace == ns &
        translations$translation == key &
        translations$lang == lang
      if (!any(match)) {
        match <-
          translations$namespace == ns &
          translations$translation == key &
          translations$lang == ""
        # if (match) {
        #   # missing translation
        # }
      }
    } else {
      match <-
        translations$namespace == ns &
        translations$key == key &
        translations$lang == lang
      if (!any(match)) {
        match <-
          translations$namespace == ns &
          translations$key == key &
          translations$lang == ""
        # if (match) {
        #   # missing translation
        # }
      }
    }
    if (sum(match) > 1) {
      util_error(
        "Internal error, sorry, please report: >1 translation for %s:%s:%s",
        sQuote(key),
        sQuote(ns),
        sQuote(lang))
    } else if (sum(match) == 1) {
      if (reverse) {
        return(translations$key[match])
      } else {
        return(translations$translation[match])
      }
    } else {
      return(key)
    }
  }, FUN.VALUE = character(1))
  structure(
    values,
    names = keys,
    ns = ns,
    lang = lang,
    class = "dataquieR_translated"
  )
}

#' `print` implementation for the class `dataquieR_translated`
#'
#' `dataquieR`'s translated texts featuring access to the language keys, still.
#'
#' @param x `dataquieR_translated` object to print
#' @param ... passed to base::print
#'
#' @seealso base::print
#'
#' @return as print
#' @export
print.dataquieR_translated <- function(x, ...) {
  y <- x
  class(y) <- "character"
  attr(y, "names") <- NULL
  attr(y, "ns") <- NULL
  attr(y, "lang") <- NULL
  print(y, ...)
  invisible(x)
}

#' `as.character` implementation for the class `dataquieR_translated`
#'
#' `dataquieR`'s translated texts featuring access to the language keys, still.
#'
#' @param x `dataquieR_translated` object to print
#' @param ... passed to base::as.character
#'
#' @seealso base::as.character
#'
#' @return character with only the translated entries
#' @export
as.character.dataquieR_translated <- function(x, ...) {
  class(x) <- "character"
  attr(x, "names") <- NULL
  attr(x, "ns") <- NULL
  attr(x, "lang") <- NULL
  x
}

#' `names` implementation for the class `dataquieR_translated`
#'
#' `dataquieR`'s translated texts featuring access to the language keys, still.
#' this function returns the language keys.
#'
#' only `setNames(nm = x)` is allowed for convenience. Any other assignment
#' would mean to change the language keys, so this is not allowed.
#'
#' @param x `dataquieR_translated` object
#' @param value the names to assign
#'
#' @seealso base::as.character
#'
#' @return names of the underlying character vector
#' @export
`names<-.dataquieR_translated` <- function(x, value) {
  if (identical(x, value)) { # enable setNames(nm = colnames(x)), if colnames(x) are translated
    x <- as.character(unclass(x))
    attr(x, "names") <- value
    return(x)
  }
  util_error("You cannot change the language keys of an %s object",
             sQuote("dataquieR_translated"))
}

#' Detect if an object is a `dataquieR_translated` object
#'
#' @param x the object to test
#'
#' @returns `TRUE`, if `x` is a `dataquieR_translated` object.
#' @export
prep_is_translated <- function(x) {
  inherits(x, "dataquieR_translated")
}

`util_translated_colnames<-` <- function(x, value) {
  util_stop_if_not(is.data.frame(x) || is.matrix(x))
  util_stop_if_not(inherits(value, "dataquieR_translated"))
  util_stop_if_not(!inherits(colnames(x), "dataquieR_translated"))
  if (is.data.frame(x)) {
    attr(x, "names") <-
      value
  } else if (is.matrix(x)) {
    attr(x, "dimnames")[[2]] <-
      value
  }
  x
}

util_untranslated_colnames <- function(x) {
  util_stop_if_not(inherits(cn <- colnames(x), "dataquieR_translated"))
  attr(cn, "names")
}

#' JSON representation for translated dataquieR objects
#'
#' @param x A `dataquieR_translated` object.
#' @param ... Passed through to `jsonlite::toJSON()` internals.
#' @return A JSON-ready representation.
#' @noRd
asJSON.dataquieR_translated <- function(x, ...) {
  if (util_ensure_suggested("jsonlite", err = FALSE)) {
   jsonlite::toJSON(unname(as.character(unclass(x))))
  } else {
    util_error(
      "Should not be reached in asJSON, internal error, sorry, please report.")
  }
}

util_with_orig_names <- function(x) {
  if (inherits(attr(x, "names"), "dataquieR_translated")) {
    names(x) <- names(attr(x, "names"))
  }
  x
}

translation_version <- as.numeric_version(attr(readRDS(system.file("translations.rds",
                                                      package = packageName())),
                                  "version"))

#' like as.character, but S3-aware for time classes (`hms`, `difftime`, `chron::times`)
#'
#' @param x  object to convert
#' @param digits.secs integer, passed via options(digits.secs) to control fractional seconds
#' @param tz optional `tz` for `POSIXt` formatting (if `NULL`, use default printing)
#' @param ... ignored
#' @return character vector (length(x)), never recurses
#' @noRd
util_as_character <- function(x, digits.secs = getOption("digits.secs", 3L), tz = NULL, ...) {
  old_dsec <- getOption("digits.secs")
  on.exit(options(digits.secs = old_dsec), add = TRUE)
  options(digits.secs = digits.secs)
  fmt_hms_vec <- function(v) {
    # v may be hms or numeric seconds
    if (inherits(v, "hms")) {
      return(.util_format_hms(v))
    }
    if (inherits(v, "difftime")) {
      return(.util_format_hms(hms::as_hms(v)))
    }
    if (inherits(v, "times")) { # chron::times -> fraction of day
      return(.util_format_hms(hms::as_hms(as.numeric(v) * 86400)))
    }
    if (is.numeric(v)) {
      return(.util_format_hms(hms::as_hms(v)))
    }
    # fallback
    return(as.character(v))
  }

  # Fast paths (vectorised) ----------------------------------------------------
  if (inherits(x, "hms")) {
    r <- .util_format_hms(x)
    r[trimws(r) == "NA"] <- NA_character_
    return(r)
  }
  if (inherits(x, "difftime")) {
    r <- .util_format_hms(hms::as_hms(x))
    r[trimws(r) == "NA"] <- NA_character_
    return(r)
  }
  if (inherits(x, "times")) {
    r <- .util_format_hms(hms::as_hms(as.numeric(x) * 86400))
    r[trimws(r) == "NA"] <- NA_character_
    return(r)
  }
  if (inherits(x, "POSIXt")) {
    if (is.null(tz)) {
      r <- format(x)
    } else {
      r <- format(x, tz = tz, usetz = TRUE)
    }
    r[trimws(r) == "NA"] <- NA_character_
    return(r)
  }
  if (is.factor(x))   return(as.character(x))
  if (is.character(x)) return(x)
  if (is.numeric(x))  return(as.character(x))
  if (is.logical(x))  return(as.character(x))
  if (length(x) == 0L) return(character(0))

  # List-handling (ohne Rekursion) --------------------------------------------
  if (is.list(x)) {
    # Check for inhomogeneous value formats (hms "vectors")
    ..is_hms <- vapply(x, inherits, "hms", FUN.VALUE = logical(1))
    ..is_na1 <- lapply(x, function(x) {
      r <- FALSE
      try(r <- length(x) == 1 && is.na(x), silent = TRUE)
      r
    })
    if (any(..is_hms) && !all(..is_hms)) {
      varname <- attr(x, "..cn")
      vn <- dQuote(varname)
      if (length(vn) == 0)
        vn <- "data"
      x[!..is_hms] <-
        lapply(x[!..is_hms], function(x) {
          if (length(x) != 1) {
            r <- hms::as_hms(NA_character_)
          } else {
            r <- try(hms::as_hms(x), silent = TRUE)
            if (util_is_try_error(r)) {
              r <- hms::as_hms(NA_character_)
            }
          }
          r
        })
      ..is_na2 <- lapply(x, function(x) {
        r <- FALSE
        try(r <- length(x) == 1 && is.na(x), silent = TRUE)
        r
      })
      lost <- sum(!vapply(mapply(..is_na1, ..is_na2, FUN = identical,
                                 SIMPLIFY = FALSE), FUN = identity,
                          FUN.VALUE = logical(1)))
      if (!.called_in_pipeline2()) {
        if (lost > 0) {
          .cnd <- util_warning
        } else {
          .cnd <- util_message
        }
        .cnd(
          # TODO: This warning is thrown several times, sometimes (if study data cache is active, it is missed)
          c("Found %d non-hms %s values in %s -- %d of which",
            "could not be casted to hms"),
          sum(!..is_hms),
          sQuote(DATA_TYPES$TIME),
          vn,
          lost,
          integrity_indicator = "int_vfe_inhom",
          varname = varname)
      } else {
        .cnd <- util_message
        .cnd(
          c("Found %d non-hms %s values in %s"),
          sum(!..is_hms),
          sQuote(DATA_TYPES$TIME),
          vn,
          integrity_indicator = "int_vfe_inhom",
          varname = varname)
      }
    }

    # Viele data.frame-Spalten mit hms sind tatsächlich Listen von Skalar-hms.
    # Wir formatieren elementweise, aber rufen NICHT wieder util_as_character() auf.
    # out <- lapply(
    #   x,
    #   function(e) {
    #     if (is.null(e) || length(e) == 0L) return(NA_character_)
    #     if (inherits(e, "hms"))       return(.util_format_hms(e))
    #     if (inherits(e, "difftime"))  return(.util_format_hms(hms::as_hms(e)))
    #     if (inherits(e, "times"))     return(.util_format_hms(hms::as_hms(as.numeric(e) * 86400)))
    #     if (inherits(e, "POSIXt")) {
    #       if (is.null(tz)) return(format(e))
    #       return(format(e, tz = tz, usetz = TRUE))
    #     }
    #     if (is.factor(e))    return(as.character(e))
    #     if (is.character(e)) return(e)
    #     if (is.numeric(e))   return(as.character(e))
    #     if (is.list(e))      return("<list>")  # Schutz vor verschachtelten Listen → keine Rekursion!
    #     return(as.character(e))
    #   }
    # )
    # if (!all(vapply(out, is.character, FUN.VALUE = logical(1)))) {
    #   save(out, file = "/tmp/xx")
    #   stop()
    # }
    out <- lapply(
      x,
      function(e) {
        if (is.null(e) || length(e) == 0L) return(NA_character_)
        if (inherits(e, "hms"))       {
          r <- .util_format_hms(e)
          r[trimws(r) == "NA"] <- NA_character_
          return(r)
        }
        if (inherits(e, "difftime"))  {
          r <- .util_format_hms(hms::as_hms(e))
          r[trimws(r) == "NA"] <- NA_character_
          return(r)
        }
        if (inherits(e, "times"))     {
          r <- .util_format_hms(hms::as_hms(as.numeric(e) * 86400))
          r[trimws(r) == "NA"] <- NA_character_
          return(r)
        }
        if (inherits(e, "POSIXt")) {
          if (is.null(tz)) r <- format(e)
          r <- format(e, tz = tz, usetz = TRUE)
          r[trimws(r) == "NA"] <- NA_character_
          return(r)
        }
        if (is.factor(e))    {
          r <- as.character(e)
          r[trimws(r) == "NA"] <- NA_character_
          return(r)
        }
        if (is.character(e)) {
          return(e)
        }
        if (is.numeric(e)) {
          r <- as.character(e)
          r[trimws(r) == "NA"] <- NA_character_
          return(r)
        }
        if (is.list(e)) {
          return(NA_character_)  # Schutz vor verschachtelten Listen → keine Rekursion!
        }
        r <- as.character(e)
        r[trimws(r) == "NA"] <- NA_character_
        return(r)
      })
    out[vapply(out, length, FUN.VALUE = integer(1)) != 1] <-
      NA_character_
    out <- vapply(
      out,
      FUN = identity,
      FUN.VALUE = character(1),
      USE.NAMES = FALSE
    )
    return(out)
  }

  # Fallback -------------------------------------------------------------------
  as.character(x)
}

.util_format_hms <- function(...) {
  r <- ..util_format_hms(...)
  if (identical(r, "hms()")) {
    r <- character(0)
  }
  r
}

..util_format_hms <- utils::getS3method("format", "hms",
                                        envir = asNamespace("hms"))

#' Determine missing and/or superfluous data elements
#'
#' Depends on [dataquieR.ELEMENT_MISSMATCH_CHECKTYPE] option,
#' see there
#'
#' [Indicator]
#' @inheritParams .template_function_indicator
#' @param check_type [enum] none | exact | subset_u | subset_m. See
#'                               [dataquieR.ELEMENT_MISSMATCH_CHECKTYPE]
#'
#' @return [list] with names lots:
#'   - `DataframeData`: data frame with the unexpected elements check results.
#'   - `DataframeTable`: [data.frame] table with all errors, used for the data quality report:
#'                       - `PCT_int_sts_element`: Percentage of element
#'                                                mismatches
#'                       - `NUM_int_sts_element`: Number of element
#'                                                mismatches
#'                       - `resp_vars`: affected element names
#'
#' @export
#' @examples
#' \dontrun{
#' prep_load_workbook_like_file("~/tmp/df_level_test.xlsx")
#' meta_data_dataframe <- "dataframe_level"
#' meta_data <- "item_level"
#' }
int_sts_element_dataframe <- function(item_level = "item_level",
                                      meta_data_dataframe = "dataframe_level",
                                      meta_data = item_level,
                                      meta_data_v2,
                                      check_type =
                                        getOption(
                                          "dataquieR.ELEMENT_MISSMATCH_CHECKTYPE",
                                          dataquieR.ELEMENT_MISSMATCH_CHECKTYPE_default),
                                      dataframe_level) { # TODO: Add MISSING-column as in segment
  # Preps and checks ----

  util_maybe_load_meta_data_v2()

  # checks and fixes the function arguments
  util_ck_arg_aliases()

  meta_data_dataframe <- prep_check_meta_data_dataframe(meta_data_dataframe)
  util_expect_data_frame(meta_data_dataframe, col_names = DF_CODE)

  util_expect_data_frame(meta_data, col_names = DATAFRAMES)
  prep_check_meta_names(meta_data = meta_data,
                        level = REQUIRED)


  util_match_arg(check_type,
                 choices = c(
                   "none",
                   "exact",
                   "subset_u",
                   "subset_m"
                 ))

  DataframeData <- data.frame(DF_NAME = meta_data_dataframe[[DF_NAME]],
                              NUM_int_sts_element = rep(NA_integer_,
                                                        nrow(meta_data_dataframe
                                                        )),
                              PCT_int_sts_element = rep(NA_real_,
                                                        nrow(meta_data_dataframe
                                                        )),
                              resp_vars = rep(NA_character_,
                                              nrow(meta_data_dataframe))
  )
  if (check_type != "none") {
    # dfr_names <- list2env(
    #   setNames(as.list(meta_data_dataframe[[DF_NAME]]),
    #            nm = meta_data_dataframe[[DF_CODE]])
    # )

    dfr_codes <- list2env(
      setNames(as.list(meta_data_dataframe[[DF_CODE]]),
               nm = meta_data_dataframe[[DF_NAME]])
    )

    meta_data_dataframe <-
      meta_data_dataframe[!util_empty(meta_data_dataframe[[DF_CODE]]), , FALSE]
    parsed_dataframes_col <- lapply(setNames(
      meta_data[[DATAFRAMES]],
      nm = meta_data[[VAR_NAMES]]
    ), util_parse_assignments,
    multi_variate_text = TRUE
    )

    # switch names and values
    # usable_df_names <- intersect(prep_list_dataframes(),
    #           meta_data_dataframe[[DF_NAME]])

    DataframeData <- data.frame(DF_NAME = meta_data_dataframe[[DF_NAME]],
                                NUM_int_sts_element = rep(NA_integer_,
                                                          nrow(meta_data_dataframe
                                                          )),
                                PCT_int_sts_element = rep(NA_real_,
                                                          nrow(meta_data_dataframe
                                                          )),
                                resp_vars = rep(NA_character_,
                                                nrow(meta_data_dataframe))
    )


    DataframeData$resp_vars <-
      vapply(meta_data_dataframe[[DF_NAME]], function(df) {
        cur_code <- dfr_codes[[df]]

        from_item_level <- meta_data[, VAR_NAMES]
        from_df <- colnames(prep_get_data_frame(df,
                                                column_names_only = TRUE))
        from_all <- union(from_item_level, from_df)

        not_in_df <- from_all[(!(from_all %in% from_df))]
        not_in_il <- from_all[(!(from_all %in% from_item_level))]

        r <- character(0)
        if (length(not_in_df) > 0) {
          r["not_in_df"] <- paste("{",
                                  util_pretty_vector_string(not_in_df, n_max = 5),
                                  "} \u2209 sd")
        }
        if (length(not_in_il) > 0) {
          r["not_in_il"] <- paste("{",
                                  util_pretty_vector_string(not_in_il, n_max = 5),
                                  "} \u2209 md")
        }

        if (check_type == "subset_u") {
          r <- r["not_in_il"]
        }
        if (check_type == "subset_m") {
          r <- r["not_in_df"]
        }
        return(paste(r, collapse = " \u2227 "))
        # optional: add the dataframes to the cache?
        # cur_df <- prep_get_data_frame(df,
        #                     keep_types = TRUE)
        # prep_add_data_frames(data_frame_list = setNames(
        #   list(cur_df),
        #   nm = get(DF_CODE, envir = dfr_names)
        # ))
      }, FUN.VALUE = character(1))

    DataframeData$NUM_int_sts_element <-
      vapply(meta_data_dataframe[[DF_NAME]], function(df) {
        cur_code <- dfr_codes[[df]]

        from_item_level <- meta_data[, VAR_NAMES]
        from_df <- colnames(prep_get_data_frame(df,
                                                column_names_only = TRUE))
        from_all <- union(from_item_level, from_df)

        not_in_df <- sum(!(from_all %in% from_df))
        not_in_il <- sum(!(from_all %in% from_item_level))

        if (check_type == "exact") {
          return(not_in_df + not_in_il)
        }
        if (check_type == "subset_u") {
          return(not_in_il)
        }
        if (check_type == "subset_m") {
          return(not_in_df)
        }
        # optional: add the dataframes to the cache?
        # cur_df <- prep_get_data_frame(df,
        #                     keep_types = TRUE)
        # prep_add_data_frames(data_frame_list = setNames(
        #   list(cur_df),
        #   nm = get(DF_CODE, envir = dfr_names)
        # ))
      }, FUN.VALUE = integer(1))

    DataframeData$PCT_int_sts_element <-
      vapply(meta_data_dataframe[[DF_NAME]], function(df) {
        cur_code <- dfr_codes[[df]]

        from_item_level <- meta_data[, VAR_NAMES]
        from_df <- colnames(prep_get_data_frame(df,
                                                column_names_only = TRUE))
        from_all <- union(from_item_level, from_df)

        not_in_df <- sum(!(from_all %in% from_df))
        not_in_il <- sum(!(from_all %in% from_item_level))

        if (check_type == "exact") {
          return(100 * (not_in_df + not_in_il) / length(from_all))
        }
        if (check_type == "subset_u") {
          return(100 * not_in_il / length(from_df))
        }
        if (check_type == "subset_m") {
          return(100 * not_in_df / length(from_item_level))
        }
      }, FUN.VALUE = numeric(1))
  }

  DataframeTable <- DataframeData[, c(
    "DF_NAME",
    "NUM_int_sts_element" ,
    "PCT_int_sts_element"
  ), FALSE]
  DataframeData$PCT_int_sts_element <-
    round(DataframeData$PCT_int_sts_element, 2)
  colnames(DataframeData) <-
    util_translate_indicator_metrics(colnames(DataframeData),
                                     ignore_unknown = TRUE)
  colnames(DataframeData)[colnames(DataframeData) == "resp_vars"] <-
    "Affected Elements"
  return(list(
    DataframeTable = DataframeTable,
    DataframeData = DataframeData
  ))

}

#' Check for duplicated IDs
#'
#' @description
#' This function tests for duplicates entries in identifiers. It is possible to
#' check duplicated identifiers by study segments or to consider only selected
#' segments.
#'
#' @inheritParams .template_function_indicator
#' @param id_vars_list [list] id variable names for each segment or data frame
#' @param level [character] a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").
#' @param repetitions [vector] an integer vector indicating the number of allowed repetitions in the `id_vars`.
#' @param identifier_name_list [vector] the segments or data frame names being assessed
#' @param ... not used.
#'
#' @return a [list] with
#'   - `DataframeData`: data frame with the results of the quality check for duplicated identifiers
#'   - `DataframeTable`: data frame with selected duplicated identifiers check results, used for the data quality report.
#'   - `Other`: named list with inner lists of unique cases containing each the
#'              row indices of duplicated identifiers separated by "|" , if any.
#'              outer names are names of the data frames
#'
#' @family integrity_indicator_functions
#' @concept integrity_indicator
#' @keywords internal
util_int_duplicate_ids_dataframe <- function(level = c("dataframe"),
                                        id_vars_list,
                                        identifier_name_list,
                                        repetitions,
                                        meta_data_dataframe = "dataframe_level",
                                        ...,
                                        dataframe_level) {

  # Checks  ----
  util_ck_arg_aliases()

  level <- util_match_arg(level)


  if (missing(id_vars_list) &&
      missing(identifier_name_list) &&
      missing(repetitions) &&
      missing(meta_data_dataframe) &&
      formals()$meta_data_dataframe %in% prep_list_dataframes()) {
    meta_data_dataframe <- force(meta_data_dataframe)
  }

  if (missing(id_vars_list) &&
      missing(identifier_name_list) &&
      missing(repetitions) &&
      !missing(meta_data_dataframe)) {
    meta_data_dataframe <- prep_check_meta_data_dataframe(meta_data_dataframe)
    meta_data_dataframe <- meta_data_dataframe[
      vapply(meta_data_dataframe[[DF_NAME]],
             function(x) { !util_is_try_error(try(prep_get_data_frame(data_frame_name = x,
                                                                      keep_types = TRUE), silent = TRUE)) },
             FUN.VALUE = logical(1))
      , , drop = FALSE]
    meta_data_dataframe <- meta_data_dataframe[
      !util_empty(meta_data_dataframe[[DF_UNIQUE_ID]]) &
        as.numeric(meta_data_dataframe[[DF_UNIQUE_ID]]) > 0 &
        !util_empty(meta_data_dataframe[[DF_ID_VARS]]), ,
      drop = FALSE
    ]

    identifier_name_list <- meta_data_dataframe[[DF_NAME]];

    id_vars_list <- lapply(setNames(meta_data_dataframe[[DF_ID_VARS]],
                                    nm = meta_data_dataframe[[DF_NAME]]),
                           util_parse_assignments,
                           multi_variate_text = TRUE)

    id_vars_list <- lapply(lapply(id_vars_list, `[[`, 1), unlist,
                           recursive = FALSE)

    repetitions <- meta_data_dataframe[[DF_UNIQUE_ID]]

  } else if (!missing(meta_data_dataframe)) {
    util_error(c("I have %s and one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "repetitions"
                 )),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "repetitions"
                 )))
  } else if (missing(meta_data_dataframe) && (
    missing(identifier_name_list) ||
    missing(id_vars_list) ||
    missing(repetitions)
  )) {
    util_error(c("I don't have %s and also miss at least",
                 "one of the following: %s.",
                 "This is not supported, please provide",
                 "either %s or all of %s."),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "repetitions"
                 )),
               sQuote("meta_data_dataframe"),
               util_pretty_vector_string(
                 c("id_vars_list",
                   "identifier_name_list",
                   "repetitions"
                 )))
  }


  util_expect_scalar(identifier_name_list,
                     allow_null = TRUE,
                     allow_more_than_one = TRUE,
                     check_type = is.character)


  util_expect_scalar(repetitions,
                     allow_more_than_one = TRUE,
                     allow_na = FALSE,
                     check_type = util_all_is_integer)

  # ID variable(s) must be specified
  util_stop_if_not(is.list(id_vars_list))
  util_stop_if_not(all(vapply(id_vars_list, is.character,
                              FUN.VALUE = logical(1))))
  # util_error(
  #   c("No ID variables specified - checks for duplicates not possible."),
  #   applicability_problem = TRUE
  # )

  util_expect_scalar(identifier_name_list,
                     allow_null = TRUE,
                     allow_more_than_one = TRUE,
                     check_type = is.character)

  util_stop_if_not(length(identifier_name_list) == length(id_vars_list))
  util_stop_if_not(length(identifier_name_list) == length(repetitions))
  util_ensure_in(names(id_vars_list), identifier_name_list)
  util_ensure_in(identifier_name_list, names(id_vars_list))

  names(repetitions) <- identifier_name_list

  level <- util_match_arg(level)


  # 1. Dataframe level check ----

  result <- lapply(setNames(nm = identifier_name_list), function(current_df) {

    # Convert data from list to data frame
    data_current_df <- util_expect_data_frame(current_df, dont_assign = TRUE)

    reps <- repetitions[[current_df]]

    id_vars <- id_vars_list[[current_df]]

    id_cols <- intersect(id_vars, colnames(data_current_df))

    if (length(id_cols) == 0) {
      util_message(
        c("None of the ID variables (%s) in the metadata are included in",
          "the study data frame %s. Ignore that data frame."),
        util_pretty_vector_string(id_vars),
        dQuote(current_df),
        applicability_problem = TRUE
      )
      return(data.frame( # dummy
        check.names = FALSE,
        "Check" = "IDs",
        "Data frame" = current_df,
        "Any duplicates" = NA,
        "Number of duplicates" = NA_real_,
        "Percentage of duplicates" =
          NA_real_,
        "GRADING" = NA_real_,
        stringsAsFactors = FALSE
      ))
    }

    dup_info <- util_find_duplicated_rows(
      study_data = data_current_df,
      id_vars = id_vars,
      repeptitions = reps)

    n_uniq <- nrow(data_current_df) - sum(dup_info$unexp_reps)

    res_tmp <- data.frame(
      check.names = FALSE,
      "Check" = "IDs",
      "Data frame" = current_df,
      "Any duplicates" = !!nrow(dup_info),
      "Number of duplicates" = nrow(data_current_df) - n_uniq,
      "Percentage of duplicates" =
        round(100 * (nrow(data_current_df) - n_uniq) /
                nrow(data_current_df), 3),
      "GRADING" = ifelse(n_uniq < nrow(data_current_df), 1, 0),
      "ID Vars" = prep_deparse_assignments(id_cols, mode = "string_codes"),
      stringsAsFactors = FALSE
    )

    if (res_tmp[[3]]) { # only if there are any duplicated observations
      vec_dup <- dup_info$which
    } else {
      vec_dup <- NULL
    }

    return(list(res_tmp, vec_dup))
  })

  res_df <- do.call(rbind.data.frame, lapply(result, `[[`, 1))
  duplicates_df <- lapply(result, `[[`, 2)

  res_pipeline <- data.frame(
    "Level" = "Dataframe",
    "DF_NAME" = res_df[, 2],
    "NUM_int_sts_dupl_ids" = res_df[, 4],
    "PCT_int_sts_dupl_ids" = res_df[, 5],
    "GRADING" = res_df$GRADING,
    stringsAsFactors = FALSE
  )

  return(list(
    DataframeData = res_df,
    DataframeTable = res_pipeline,
    Other = duplicates_df
  ))
}

util_find_duplicated_rows <- function(study_data,
                                      id_vars,
                                      repeptitions = 1) {
  util_expect_data_frame(study_data, col_names = id_vars)
  study_data <- study_data[, id_vars, FALSE]
  empty_sd <- study_data
  empty_sd[] <- lapply(study_data, util_empty)
  study_data[as.matrix(empty_sd)] <- NA
  r <- as.data.frame(table(study_data, useNA = "no"))
  r <- r[r$Freq > repeptitions, , FALSE]
  nsd <- cbind(study_data,
               ..rownr = seq_len(nrow(study_data)))
  r$which <- lapply(seq_len(nrow(r)),
                    function(my_dup_rw) {
                      my_dup <- r[my_dup_rw, , FALSE]
                      dd <- merge(nsd, my_dup, by = id_vars)
                      prep_deparse_assignments(dd$..rownr,
                                               mode = "string_codes")
                    })
  r$unexp_reps <- r$Freq - repeptitions
  r
}

#' Identify univariate outliers by four different approaches
#'
#' @description
#' A classical but still popular approach to detect univariate outlier is the
#' boxplot method introduced by Tukey 1977. The boxplot is a simple graphical
#' tool to display information about continuous univariate data (e.g., median,
#' lower and upper quartile). Outliers are defined as values deviating more
#' than \eqn{1.5 \times IQR} from the 1st (Q25) or 3rd (Q75) quartile. The
#' strength of Tukey's method is that it makes no distributional assumptions
#' and thus is also applicable to skewed or non mound-shaped data
#' Marsh and Seo, 2006. Nevertheless, this method tends to identify frequent
#' measurements which are falsely interpreted as true outliers.
#'
#' A somewhat more conservative approach in terms of symmetric and/or normal
#' distributions is the 3SD approach, i.e. any measurement not in
#' the interval of \eqn{mean(x) +/- 3 * \sigma} is considered an outlier.
#'
#' Both methods mentioned above are not ideally suited to skewed distributions.
#' As many biomarkers such as laboratory measurements represent in skewed
#' distributions the methods above may be insufficient. The approach of Hubert
#' and Vandervieren 2008 adjusts the boxplot for the skewness of the
#' distribution. This approach is implemented in several R packages such as
#' [`robustbase::mc`] which is used in this implementation of [`dataquieR`].
#'
#' Another completely heuristic approach is also included to identify outliers.
#' The approach is based on the assumption that the distances between
#' measurements of the same underlying distribution should homogeneous. For
#' comprehension of this approach:
#'  - consider an ordered sequence of all measurements.
#'  - between these measurements all distances are calculated.
#'  - the occurrence of larger distances between two neighboring measurements
#'  may
#' than indicate a distortion of the data. For the heuristic definition of a
#' large distance \eqn{1 * \sigma} has been been chosen.
#'
#' Note, that the plots are not deterministic, because they use
#' [ggplot2::geom_jitter].
#'
#' [Indicator]
#'
#' @details
#'
#' ***Hint***: *The function is designed for unimodal data only.*
#'
#' # ALGORITHM OF THIS IMPLEMENTATION:
#'
#'  - Select all variables of type float in the study data
#'  - Remove missing codes from the study data (if defined in the metadata)
#'  - Remove measurements deviating from limits defined in the metadata
#'  - Identify outliers according to the approaches of Tukey (Tukey 1977),
#'    3SD (Saleem et al. 2021), Hubert (Hubert and Vandervieren 2008),
#'    and SigmaGap (heuristic)
#'  - An output data frame is generated which indicates the no. possible
#'    outliers, the direction of deviations (Outliers, low; Outliers, high) for all methods
#'    and a summary score which sums up the deviations of the different rules
#'  - A scatter plot is generated for all examined variables, flagging
#'    observations according to the no. violated rules (step 5).
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable list] the name of the continuous measurement
#'                                  variable
#' @param exclude_roles [variable roles] a character (vector) of variable roles
#'                                       not included
#' @param n_rules [integer] from=1 to=4. the no. rules that must be violated
#'        to flag a variable as containing outliers. The default is 4, i.e. all.
#' @param max_non_outliers_plot [integer] from=0. Maximum number of non-outlier
#'                                                points to be plot. If more
#'                                                points exist, a subsample will
#'                                                be plotted only. Note, that
#'                                                sampling is not deterministic.
#' @param criteria [set] tukey | 3SD | hubert | sigmagap. a vector with
#'                       methods to be used for detecting outliers.
#'
#' @seealso
#' - [acc_robust_univariate_outlier]
#' - [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_robust_univariate_outlier.html
#' )
#'
#' @importFrom dplyr %>%
#' @importFrom ggplot2 ggplot geom_jitter position_jitter aes scale_size_continuous scale_color_manual facet_wrap vars scale_alpha theme_minimal
#' @importFrom stats median aggregate sd aggregate
#' @export
#'
#' @return a list with:
#' - `SummaryTable`: [`data.frame`] with the columns
#'        `Variables`, `Mean`, `SD`, `Median`, `Skewness`, `Tukey (N)`,
#'        `3SD (N)`, `Hubert (N)`, `Sigma-gap (N)`, `NUM_acc_ud_outlu`,
#'        `Outliers, low (N)`, `Outliers, high (N)` `Grading`
#'   - `SummaryData`: [`data.frame`] with the columns
#'        `Variables`, `Mean`, `SD`, `Median`, `Skewness`, `Tukey (N)`,
#'        `3SD (N)`, `Hubert (N)`, `Sigma-gap (N)`, `Outliers (N)`,
#'        `Outliers, low (N)`, `Outliers, high (N)`
#'   - `SummaryPlotList`: [`ggplot2::ggplot`] univariate outlier plots
#'
#'
acc_univariate_outlier <- function(resp_vars = NULL,
                                   study_data,
                                   label_col,
                                   item_level = "item_level",
                                   exclude_roles,
                                   n_rules = length(unique(criteria)),
                                   max_non_outliers_plot = 10000,
                                   criteria = c("tukey", "3sd",
                                                "hubert", "sigmagap"),
                                   meta_data = item_level,
                                   meta_data_v2) {

  # preps ----------------------------------------------------------------------
  util_maybe_load_meta_data_v2()

  #compatibility with previous name (sixsigma)
  #replace "sixsigma" (if any attributed in criteria) with 3SD
  old_name<- "sixsigma"
  if(any(criteria %in% old_name)== TRUE){
    criteria[criteria=="sixsigma"] <- "3sd"
  }

  #all lowercase
  criteria <- trimws(tolower(criteria))

  # TODO: Remove all obsoleted checks on resp_vars, the function uses util_correct_variable_use
  util_expect_scalar(criteria,
                     allow_more_than_one = TRUE,
                     allow_null = TRUE,
                     check_type = is.character)



  if (length(unique(criteria)) < 1 ||
      length(unique(criteria)) >
      length(eval(formals(acc_univariate_outlier)$criteria)) ||
      !all(criteria %in% eval(formals(acc_univariate_outlier)$criteria))) {
    if (!.called_in_pipeline)
      util_message(c("The formal criteria must have > 0 and < %d entries.",
                     "Allowed values are %s.",
                     "I was called with %s, falling back to default %s."),
                   length(eval(formals(acc_univariate_outlier)$criteria)),
                   paste(dQuote(eval(formals(acc_univariate_outlier)$criteria)),
                         collapse = ", "),
                   paste(dQuote(unique(criteria)), collapse = ", "),
                   paste(dQuote(eval(formals(acc_univariate_outlier)$criteria)),
                         collapse = ", "), applicability_problem = TRUE)
    criteria <- eval(formals(acc_univariate_outlier)$criteria)
  }

  if (length(n_rules) != 1 || !is.numeric(n_rules) ||
      !all(util_is_integer(n_rules)) ||
      !(n_rules %in% seq_len(length(unique(criteria))))) {
    if (!.called_in_pipeline)
      util_message(
        "The formal n_rules is not an integer between 1 and %d, default (%d) is used.",
        length(unique(criteria)),
        min(eval(formals(acc_univariate_outlier)$n_rules),
            length(unique(criteria))),
        applicability_problem = TRUE)
    n_rules <- min(eval(formals(acc_univariate_outlier)$n_rules),
                   length(unique(criteria)))
  }

  if (length(max_non_outliers_plot) != 1 ||
      !is.numeric(max_non_outliers_plot) ||
      !all(util_is_integer(max_non_outliers_plot)) ||
      (max_non_outliers_plot < 0)) {
    util_message(
      c("The formal max_non_outliers_plot is not an integer >= 0,",
        "default (%d) is used."),
      formals(acc_univariate_outlier)$max_non_outliers_plot,
      applicability_problem = TRUE)
    max_non_outliers_plot <-
      formals(acc_univariate_outlier)$max_non_outliers_plot
  }

  # map metadata to study data
  prep_prepare_dataframes(.replace_hard_limits = TRUE)

  util_correct_variable_use("resp_vars",
                            allow_more_than_one = TRUE,
                            allow_null = TRUE,
                            allow_any_obs_na = TRUE,
                            need_type = "integer | float",
                            need_scale = "interval | ratio",
                            do_not_stop = TRUE,
                            remove_not_found = TRUE
  )

  if (is.null(meta_data[[DATA_TYPE]]) ||
      any(is.na(meta_data[[DATA_TYPE]]))) {
    if (is.null(meta_data[[DATA_TYPE]])) {
      which_na <- rep(TRUE, nrow(meta_data))
    } else {
      which_na <- is.na(meta_data[[DATA_TYPE]])
    }
    meta_data[[DATA_TYPE]][which_na] <-
      prep_datatype_from_data(resp_vars = meta_data[[label_col]][which_na],
                              study_data = ds1)

    list_of_types <-
      paste(sQuote(meta_data[[label_col]][which_na]),
            '->',
            sQuote(meta_data[[DATA_TYPE]][which_na]))

    if (length(list_of_types) > 5) {
      dts <- "..."
    } else {
      dts <- NULL
    }
    list_of_types <- c(head(list_of_types, 5), dts)
    list_of_types <- paste0(list_of_types, collapse = ", ")
    util_warning(c(
      "No %s for all or some variables defined in the metadata.",
      "I guessed them based on data: %s"),
      dQuote(DATA_TYPE),
      list_of_types,
      applicability_problem = TRUE
    )
  }

  # no variables defined?
  if (length(resp_vars) == 0) {

    # which are float or integer?
    resp_vars <- meta_data[[label_col]][meta_data[[DATA_TYPE]] %in%
                                          c(DATA_TYPES$FLOAT,
                                            DATA_TYPES$INTEGER)]
    util_message(paste0("The following variables: ",
                        paste0(resp_vars, collapse = ", "), " were selected."),
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    if (length(resp_vars) == 0) {
      util_error(paste0("No variables with suitable data type defined."),
                 applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    }

    resp_vars <- intersect(resp_vars, colnames(ds1))
  } else {
    # defined variables are not of type float/integer?
    isfloat <- meta_data[[DATA_TYPE]] %in%
      c(DATA_TYPES$FLOAT, DATA_TYPES$INTEGER)
    isrvs <-
      meta_data[[label_col]] %in% resp_vars

    if (!all(isfloat | !isrvs)) {
      resp_vars <- meta_data[[label_col]][isfloat & isrvs]
      if (!all(!isrvs | isfloat)) util_warning(paste0("Only: ",
                                                      paste0(resp_vars,
                                                             collapse = ", "),
                                                      " are defined to be of ",
                                                      "type float or integer."),
                                               applicability_problem = TRUE,
                                               intrinsic_applicability_problem = TRUE)
    }
  }

  # should some variables not be considered?
  if (!missing(exclude_roles)) {
    if (!(all(exclude_roles %in% meta_data[[VARIABLE_ROLE]]))) {
      util_warning(
        "Specified VARIABLE_ROLE not in meta_data. No exclusion applied.",
        applicability_problem = TRUE)
    } else {
      which_vars_not <- meta_data[[label_col]][meta_data[[VARIABLE_ROLE]] %in%
                                                 exclude_roles]
      if (length(intersect(resp_vars, which_vars_not)) > 0) {
        util_message(paste0("Study variables: ",
                            paste(dQuote(intersect(resp_vars, which_vars_not)),
                                  collapse = ", "), " have been excluded."),
                     applicability_problem = TRUE,
                     intrinsic_applicability_problem = TRUE)
      }
      resp_vars <- setdiff(resp_vars, which_vars_not)
    }
  }

  # remove resp_vars with non-matching data type
  whicharenum <- vapply(FUN.VALUE = logical(1), ds1[, resp_vars, drop = FALSE],
                        function(x) is.numeric(x))

  if (!all(whicharenum)) {
    util_message(paste0(
      "Variables ", paste0(dQuote(resp_vars[!whicharenum]), collapse = ", "),
      " are not of type float or integer and will be removed",
      " from univariate outlier analysis."
    ), applicability_problem = TRUE,
    intrinsic_applicability_problem = TRUE)
    resp_vars <- resp_vars[whicharenum]
  }

  intcheck <- vapply(FUN.VALUE = logical(1), ds1[, resp_vars, drop = FALSE],
                     function(x) all(util_is_integer(x), na.rm = TRUE))

  if (any(intcheck)) {
    util_message(paste0(
      "Variables: ", paste0(dQuote(resp_vars[intcheck]), collapse = ", "),
      " show integer values only, but will be nonetheless considered."
    ), applicability_problem = TRUE, intrinsic_applicability_problem = TRUE)
  }

  if (length(resp_vars) > 0) {
    resp_vars <-
      util_no_value_labels(
        resp_vars = resp_vars,
        meta_data = meta_data,
        label_col = label_col,
        warn = TRUE,
        stop = TRUE
      )
  }

  # Label assignment -----------------------------------------------------------
  # temporary study data
  ds1_ll <- ds1

  #############
  # Results   #
  #############
  # Boxplot
  # ds2 <- melt(ds1_ll[, c(resp_vars, group_vars)], measure.vars = resp_vars)

  if (length(resp_vars) == 0) {
    util_error("No suitable response variables left.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  } else if (length(resp_vars) == 1) {
    ds2 <- ds1_ll[, resp_vars, drop = FALSE]
    ds2$variable <- resp_vars
    ds2 <- ds2[, c(2, 1)]
    names(ds2) <- c("variable", "value")
  } else {
    # ds2 <- melt(ds1_ll[, c(resp_vars)], measure.vars = resp_vars)
    ds2 <- stats::reshape(data = ds1_ll[, resp_vars],
                          varying = colnames(ds1_ll[, resp_vars]),
                          v.names = "value",
                          times = colnames(ds1_ll[, resp_vars]),
                          direction = "long")
    ds2 <- ds2[, -which(names(ds2) == "id")]
    #ds2$time <- as.factor(ds2$time)
    ds2$time <- factor(ds2$time, levels = resp_vars, ordered = FALSE)
    names(ds2)[names(ds2) == "time"] <- "variable"
    rownames(ds2) <- NULL

  }

  ds2$value <- as.numeric(ds2$value)

  # remove NAs from analysis df
  ds2plot <- ds2[!is.na(ds2$value), ]
  if (nrow(ds2plot) * ncol(ds2plot) == 0) {
    util_error("No data left, aborting.",
               applicability_problem = FALSE)
  }

  # Initialize with NA
  ds2plot$tukey <- NA
  ds2plot$threeSD <- NA
  ds2plot$hubert <- NA
  ds2plot$sigmagap <- NA

  # apply outlier functions to plot-df
  # after export/final built  correct the call of the utility functions
  ds2plotOL <- ds2plot %>%
    dplyr::group_by(variable) %>%
    dplyr::mutate(tukey = util_tukey(value)) %>%
    dplyr::mutate(threeSD = util_3SD(value)) %>%
    dplyr::mutate(hubert = util_hubert(value)) %>%
    dplyr::mutate(sigmagap = util_sigmagap(value))

  # table(util_tukey(ds2plotOL$value))

  # tibble to df
  ds2plot <- as.data.frame(ds2plotOL)


  #Fix the problem with name 3SD starting with a number replacing it with threeSD---
  orig_name<- "3sd"
  if(any(criteria %in% orig_name)== TRUE){
    criteria[criteria=="3sd"] <- "threeSD"
  }

  # calculate summary of all outlier definitions
  ds2plot$Rules <- apply(ds2plot[, criteria, drop = FALSE], 1, sum)

  # outlier to low or to high?
  ds2plot$tlta <- ifelse(ds2plot$Rules >= n_rules, 1, 0)

  # Indicate direction of deviation (low/high)
  for (i in unique(ds2plot$variable)) {
    if (length(ds2plot$tlta[ds2plot$tlta == 1 & ds2plot$variable == i]) > 0) {
      ds2plot$tlta[ds2plot$tlta == 1 & ds2plot$variable == i] <- ifelse(
        ds2plot$value[ds2plot$tlta == 1 & ds2plot$variable == i] <
          median(ds2plot$value[ds2plot$variable == i], na.rm = TRUE),
        -1, 1
      )
    }
  }

  # create summary table here --------------------------------------------------
  st1 <- aggregate(ds2plot$value, list(ds2plot$variable), mean) # TODO: here and in multi, add standard columns for indicator metrics
  colnames(st1) <- c("Variables", "Mean")
  st1$"No.records"<- aggregate(ds2plot$value, list(ds2plot$variable), length)$x
  st1$"SD" <- aggregate(ds2plot$value, list(ds2plot$variable), sd)$x
  st1$"Median" <- aggregate(ds2plot$value, list(ds2plot$variable), median)$x
  st1$"Skewness" <-
    aggregate(ds2plot$value, list(ds2plot$variable), robustbase::mc,
              doScale = FALSE)$x
  st1$"Tukey (N)" <- aggregate(ds2plot$tukey, list(ds2plot$variable), sum)$x
  st1$"3SD (N)" <-
    aggregate(ds2plot$threeSD, list(ds2plot$variable), sum)$x
  st1$"Hubert (N)" <- aggregate(ds2plot$hubert, list(ds2plot$variable), sum)$x
  st1$"Sigma-gap (N)" <-
    aggregate(ds2plot$sigmagap, list(ds2plot$variable), sum)$x
  st1$"Outliers (N)" <- aggregate(ds2plot$Rules, list(ds2plot$variable),
                                  function(x) {
                                    sum(x >= n_rules)
                                  })$x
  st1$"Outliers, low (N)" <- aggregate(ds2plot$tlta, list(ds2plot$variable),
                                       function(x) {
                                         sum(x == -1)
                                       })$x
  st1$"Outliers, high (N)" <- aggregate(ds2plot$tlta, list(ds2plot$variable),
                                        function(x) {
                                          sum(x == 1)
                                        })$x
  st1$GRADING <- ifelse(st1$"Outliers (N)" > 0, 1, 0)

  # format output
  st1$Mean <- round(st1$Mean, digits = 2)
  st1$Median <- round(st1$Median, digits = 2)
  st1$SD <- round(st1$SD, digits = 2)
  st1$Skewness <- round(st1$Skewness, digits = 2)

  SummaryTable<- st1
  names(SummaryTable)[names(SummaryTable) == "Outliers (N)"] <- "NUM_acc_ud_outlu"
  SummaryTable$PCT_acc_ud_outlu <- round(SummaryTable$NUM_acc_ud_outlu/
                                           SummaryTable$No.records*100,
                                         digits = 2)

  # create plot list here ------------------------------------------------------
  # format to factor for plot
  plot_list <- list()

  disc_cols <- c("#2166AC", "#fdd49e", "#fc8d59", "#d7301f", "#7f0000")
  names(disc_cols) <- c(0:4)

  # select as many colors as needed
  disc_cols <- disc_cols[c("0", rev(5-seq_len(length(unique(criteria)))))]
  names(disc_cols) <- c("0", seq_len(length(unique(criteria))))

  for (i in unique(ds2plot$variable)) {

    ds_i <- subset(ds2plot, variable == i)

    n_non_ol <- sum(ds_i$Rules == 0)

    if (max_non_outliers_plot < n_non_ol) {

      dsi_non_ol <- ds_i[ds_i$Rules == 0, , FALSE]
      dsi_ol <- ds_i[ds_i$Rules > 0, , FALSE]

      subsel_non_ol <- sample(seq_len(nrow(dsi_non_ol)),
                              size =
                                min(max_non_outliers_plot, nrow(dsi_non_ol)))

      ds_i <- rbind.data.frame(dsi_non_ol[subsel_non_ol, , FALSE], dsi_ol)

      util_message(
        c("For %s, %d from %d non-outlier data values were",
          "sampled to avoid large plots."),
        dQuote(i),
        max_non_outliers_plot,
        n_non_ol,
        applicability_problem = FALSE
      )
    }

    ds_i$Rules <- factor(ds_i$Rules)

    if (nrow(ds_i) > 0) {
      ds_i$variable <-
        prep_get_labels(
          ds_i$variable,
          item_level = meta_data,
          label_col = label_col,
          resp_vars_match_label_col_only = TRUE,
          label_class = "SHORT")
      if (length(unique(ds_i$variable)) == 1 && .called_in_pipeline) {
        p_i <- util_create_lean_ggplot(ggplot(ds_i, aes(x = variable, y = value)) +
          geom_jitter(data = ds_i,
                      position = position_jitter(width = 0.3, height = 0.03),
                      aes(color = Rules, alpha = 0.5, size =
                            as.numeric(Rules) / 10)) +
          scale_size_continuous(range = c(0.5, 3), guide = "none") +
          scale_color_manual(values = disc_cols) +
          scale_alpha(guide = "none") +
          xlab("") + ylab("") +
          theme_minimal() +
          theme(axis.text.x = element_blank()),
          ds_i = ds_i,
          disc_cols = disc_cols)
      } else {
        p_i <- util_create_lean_ggplot(
          ggplot(ds_i, aes(x = variable, y = value)) +
          geom_jitter(data = ds_i,
                      position = position_jitter(width = 0.3, height = 0.03),
                      aes(color = Rules, alpha = 0.5, size =
                            as.numeric(Rules) / 10)) +
          scale_size_continuous(range = c(0.5, 3), guide = "none") +
          scale_color_manual(values = disc_cols) +
          facet_wrap(vars(variable), scales = "free") +
          scale_alpha(guide = "none") +
          xlab("") + ylab("") +
          theme_minimal(),
          ds_i = ds_i,
          disc_cols = disc_cols)
      }
    } else {
      p_i <- util_create_lean_ggplot(ggplot() +
        annotate("text", x = 0, y = 0, label =
                   sprintf("No outliers detected for %s", dQuote(i))) +
        theme(
          axis.line = element_blank(),
          axis.text.x = element_blank(),
          axis.text.y = element_blank(),
          axis.ticks = element_blank(),
          axis.title.x = element_blank(),
          axis.title.y = element_blank(),
          legend.position = "none",
          panel.background = element_blank(),
          panel.border = element_blank(),
          panel.grid.major = element_blank(),
          panel.grid.minor = element_blank(),
          plot.background = element_blank()
        ),
        i = i)
    }

    # Define min and max values in plot for size hint
    if (nrow(ds_i)==0){
      #In case of an empty result in ds_i
      min_val <- 0
      max_val <- 0
    } else{
      min_val <-  min(ds_i$value)
      max_val <- max(ds_i$value)
    }

    range <- max_val - min_val
    y_variable <-  unique(ds_i$variable)
    no_char_y <- max(nchar(c(min_val, max_val)))


    # Figure size hint for plot
    attr(p_i, "sizing_hints") <- list(
      figure_type_id = "scatt_plot",
      range = range,
      number_of_vars = length(unique(ds_i$variable)),
      no_char_y = no_char_y
    )

    plot_list[[i]] <- util_set_size(p_i, width_em = 10, height_em = 25)

  }

  st1 <- st1[, !names(st1) %in% c("GRADING")]

  return(list(SummaryTable = SummaryTable,
              SummaryData = st1,
              SummaryPlotList = plot_list))
}

#' @inherit acc_univariate_outlier
#' @seealso [acc_univariate_outlier]
#' @export
acc_robust_univariate_outlier <- acc_univariate_outlier

#' Plots and checks for distributions
#'
#' @description
#' Data quality indicator checks "Unexpected location" and "Unexpected
#' proportion" with histograms.
#'
#' [Indicator]
#'
#' @details
#' # Algorithm of this implementation:
#'
#' - If no response variable is defined, select all variables of type float or
#' integer in the study data.
#' - Remove missing codes from the study data (if defined in the metadata).
#' - Remove measurements deviating from (hard) limits defined in the metadata
#' (if defined).
#' - Exclude variables containing only `NA` or only one unique value (excluding
#' `NA`s).
#' - Perform check for "Unexpected location" if defined in the metadata (needs a
#' LOCATION_METRIC (mean or median) and LOCATION_RANGE (range of expected values
#' for the mean and median, respectively)).
#' - Perform check for "Unexpected proportion" if defined in the metadata (needs
#' PROPORTION_RANGE (range of expected values for the proportions of the
#' categories)).
#' - Plot histogram(s).
#'
#' @export
#'
#' @inheritParams .template_function_indicator
#'
#' @param resp_vars [variable list] the names of the measurement variables
#' @param plot_ranges [logical] Should the plot show ranges and results from the
#'                              data quality checks? (default: TRUE)
#' @param check_param [enum] any | location | proportion. Which type of check
#'                         should be conducted (if possible): a check on the
#'                         location of the mean or median value of the study
#'                         data, a check on proportions of categories, or either
#'                         of them if the necessary metadata is available.
#' @param flip_mode [enum] default | flip | noflip | auto. Should the plot be
#'                         in default orientation, flipped, not flipped or
#'                         auto-flipped. Not all options are always supported.
#'                         In general, this con be controlled by
#'                         setting the `roptions(dataquieR.flip_mode = ...)`. If
#'                         called from `dq_report`, you can also pass
#'                         `flip_mode` to all function calls or set them
#'                         specifically using `specific_args`.
#'
#' @return A [list] with:
#'   - `SummaryTable`: [data.frame] containing data quality checks for
#'                     "Unexpected location" (`FLG_acc_ud_loc`) and "Unexpected
#'                     proportion" (`FLG_acc_ud_prop`) for each response
#'                     variable in `resp_vars`.
#'   - `SummaryData`: a [data.frame] containing data quality checks for
#'                    "Unexpected location" and / or "Unexpected proportion"
#'                    for a report
#'   - `SummaryPlotList`: [list] of [ggplot2::ggplot]s for each response variable in
#'                    `resp_vars`.
#'
#' @importFrom ggplot2 ggplot aes geom_histogram geom_bar geom_vline stat_ecdf geom_segment scale_x_continuous scale_color_manual coord_flip theme_minimal theme element_text element_blank labs ylab
#' @importFrom stats na.omit
#' @importFrom rlang .data
#' @importFrom grDevices hcl.colors
#' @seealso
#' [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_distributions.html
#' )
acc_distributions <- function(resp_vars = NULL,
                              study_data,
                              label_col,
                              item_level = "item_level",
                              check_param = c("any", "location", "proportion"),
                              plot_ranges = TRUE,
                              flip_mode = "noflip",
                              meta_data = item_level,
                              meta_data_v2) {
  # preps ----------------------------------------------------------------------
  # map metadata to study data
  util_maybe_load_meta_data_v2()
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          #.apply_factor_metadata = TRUE, # can be omitted in favor of .apply_factor_metadata_inadm
                          .apply_factor_metadata_inadm = TRUE
  )

  # If no response variable is defined, all suitable variables will be selected.
  if (length(resp_vars) == 0) {
    util_message(
      c("All variables defined to be integer or float in the metadata are used",
        "by acc_distributions."), # TODO: extend for nominal and ordinal variables
      applicability_problem = TRUE, intrinsic_applicability_problem = TRUE)
    resp_vars <- meta_data[[label_col]][meta_data$DATA_TYPE %in%
                                          c("integer", "float")]
    resp_vars <- intersect(resp_vars, colnames(ds1))
    if (length(resp_vars) == 0) {
      util_error("No suitable variables were defined for acc_distributions.",
                 applicability_problem = TRUE)
    }
  }

  util_correct_variable_use(resp_vars,
                            allow_more_than_one = TRUE,
                            allow_any_obs_na = TRUE,
                            allow_all_obs_na = FALSE,
                            min_distinct_values = 2,
                            do_not_stop = ifelse(length(resp_vars) > 1,
                                                 TRUE, FALSE),
                            need_scale = "!na"
  )

  # Which parameter of the distribution should be checked?
  dq_param <- util_match_arg(check_param)

  # DQ indicator "Unexpected location" requires metadata for the range of
  # expected values for the location parameter and the location metric.
  # DQ indicator "Unexpected proportion" requires metadata for the range of
  # expected percentages for the categories.
  if (!any(grepl("_RANGE", colnames(meta_data)))) {
    util_message(paste("Metadata does not contain columns for expected ranges",
                       "of location or proportions. So acc_distributions can",
                       "generate plots, but can not perform data quality",
                       "checks."), applicability_problem = TRUE,
                 intrinsic_applicability_problem = TRUE)
    # create empty metadata list
    rvs_meta <- list("Metric" = setNames(nm = resp_vars,
                                         rep(NA, length(resp_vars))),
                     "Range" = setNames(nm = resp_vars,
                                        rep(NA, length(resp_vars))))
  } else {
    if (dq_param == "any") {
      # split variables for location and proportion checks and call
      # acc_distributions recursively
      scale_level <- meta_data[[SCALE_LEVEL]][match(resp_vars,
                                                    meta_data[[label_col]])]
      if (any(scale_level %in% c("interval", "ratio"))) {
        rvs_loc <- resp_vars[scale_level %in%
                               c("interval", "ratio")]
        res_loc <- suppressWarnings(acc_distributions(
          resp_vars = rvs_loc, study_data = study_data,
          meta_data = meta_data, label_col = label_col,
          check_param = "location", plot_ranges = plot_ranges,
          flip_mode = flip_mode))

        res_merged <- res_loc
      }

      if (any(!(scale_level %in% c("interval", "ratio")))) {
        rvs_prop <- resp_vars[!(scale_level %in%
                                  c("interval", "ratio"))]
        res_prop <- suppressWarnings(acc_distributions(
          resp_vars = rvs_prop, study_data = study_data,
          meta_data = meta_data, label_col = label_col,
          check_param = "proportion", plot_ranges = plot_ranges,
          flip_mode = flip_mode))
        if (exists("res_merged")) {
          res_merged$SummaryTable <- merge(res_merged$SummaryTable,
                                           res_prop$SummaryTable,
                                           all = TRUE, sort = FALSE)
          res_merged$SummaryData <- merge(res_merged$SummaryData,
                                          res_prop$SummaryData,
                                          all = TRUE, sort = FALSE)
          res_merged$SummaryPlotList <- c(res_merged$SummaryPlotList,
                                          res_prop$SummaryPlotList)
        } else {
          res_merged <- res_prop
        }
      }

      # order merged results to match the order of `resp_vars`
      if (length(resp_vars) > 1) {
        res_merged$SummaryTable <- res_merged$SummaryTable[
          match(resp_vars, res_merged$SummaryTable$Variables), ]
        res_merged$SummaryData <- res_merged$SummaryData[
          match(resp_vars, res_merged$SummaryData$Variables), ]
        res_merged$SummaryPlotList <- res_merged$SummaryPlotList[resp_vars]
      }

      return(res_merged)

    } else if (dq_param == "location") {
      rvs_meta <- util_prep_location_check(resp_vars = resp_vars,
                                           meta_data = meta_data,
                                           report_problems = "warning",
                                           label_col = label_col)
    } else {
      rvs_meta <- util_prep_proportion_check(resp_vars = resp_vars,
                                             meta_data = meta_data,
                                             ds1 = ds1,
                                             report_problems = "warning",
                                             label_col = label_col)
    }
  }

  rvs_with_meta <- names(rvs_meta$Range)[which(!is.na(rvs_meta$Range))]

  # Which variables are of type 'datetime'?
  is_datetime_var <- vapply(resp_vars, function(rv) {
    meta_data[["DATA_TYPE"]][meta_data[[label_col]] == rv] %in%
      c(DATA_TYPES$DATETIME)
  }, FUN.VALUE = logical(1))

  # Which variables are of type 'time'?
  is_time_var <- vapply(resp_vars, function(rv) {
    meta_data[["DATA_TYPE"]][meta_data[[label_col]] == rv] %in%
      c(DATA_TYPES$TIME)
  }, FUN.VALUE = logical(1))

  # data quality indicator checks ----------------------------------------------
  if (dq_param == "location") {
    dq_check_list <- lapply(setNames(nm = resp_vars), function(rv) {
      rv_data <- ds1[[rv]]
      loc_func <- rvs_meta$Metric[[rv]]
      if (!is.na(loc_func)) {
        loc_val <- do.call(loc_func, args = list(x = rv_data, na.rm = TRUE))
        loc_check <- !(redcap_env$`in`(loc_val, rvs_meta$Range[[rv]]))
      } else {
        loc_val <- NA
        loc_check <- NA
      }

      loc_res <- setNames(nm = c("Variables",
                                 #                                paste(c("FLG","VAL"), "acc_ud_loc", sep = "_"),
                                 "FLG_acc_ud_loc", "values_from_data",
                                 "loc_func",
                                 "loc_range"),
                          list(rv,
                               loc_check,
                               loc_val,
                               loc_func,
                               util_find_var_by_meta(resp_vars = rv,
                                                     target = "LOCATION_RANGE",
                                                     meta_data = meta_data)))
      loc_res
    })
  } else if (dq_param == "proportion") {
    dq_check_list <- lapply(setNames(nm = resp_vars), function(rv) {
      rv_data <- ds1[[rv]]
      # compute proportions for variable `rv` in the study data
      prop_table <- table(rv_data)
      prop_table <- prop_table/length(rv_data[which(!is.na(rv_data))]) * 100
      # If some categories were expected but not observed, we have to add zeroes
      # to the table.
      if (!all(names(rvs_meta$Range[[rv]]) %in% names(prop_table))) {
        nonmiss_cat <- names(prop_table)
        miss_cat <- setdiff(names(rvs_meta$Range[[rv]]), nonmiss_cat)
        prop_table <- setNames(c(prop_table, rep(0, length(miss_cat))),
                               nm = c(nonmiss_cat, miss_cat))
      }
      # order categories as in value labels list
      prop_table <- prop_table[union(names(rvs_meta$Range[[rv]]),
                                     names(prop_table))]
      # check for each category of the variable whether the proportion lies
      # within the expected range
      prop_in_range <- mapply(as.list(prop_table),
                              rvs_meta$Range[[rv]],
                              FUN = function(pp, int) {
                                if (inherits(int, "interval")) {
                                  redcap_env$`in`(pp, int)
                                } else {
                                  NA
                                }
                              })
      # Which categories of the variable should be flagged, if any?
      cat_flg <- names(prop_in_range)[which(!prop_in_range)]
      if (length(cat_flg) == 0) cat_flg <- NA

      prop_res <-  setNames(nm = c(
        "Variables",
        #        paste(c("FLG", "VAL"), "acc_ud_prop", sep = "_"),
        "FLG_acc_ud_prop", "values_from_data",
        "prop_range",
        "flg_which"),
        list(rv,
             any(!(prop_in_range[which(!is.na(prop_in_range))])),
             paste(
               paste(names(prop_table), round(prop_table, 1), sep = " = "),
               collapse = " | "),
             util_find_var_by_meta(resp_vars = rv,
                                   target = "PROPORTION_RANGE",
                                   meta_data = meta_data),
             cat_flg))
      prop_res
    })

    # remove the list of flagged categories to arrange results as dataframe
    flg_cat_prop <- lapply(dq_check_list, function(x) { x[["flg_which"]] })
    dq_check_list <- lapply(dq_check_list, function(x) {
      x[-which(names(x) == "flg_which")]
    })
  } else {
    flg_cat_prop <- list()
    dq_check_list <- list()
  }

  res_dq_check <- do.call(rbind.data.frame, dq_check_list)
  rownames(res_dq_check) <- NULL

  # summary table --------------------------------------------------------------
  if(ncol(res_dq_check)==0) { # in case there is no check for location or range, it provides a table without columns
    res_out <- res_dq_check
    res_view <- res_dq_check
  } else{
    res_out <- res_dq_check[, 1:4]
    # GRADING for backwards compatibility
    res_out$GRADING <- as.numeric(res_out[, 2])
    res_out$GRADING[which(is.na(res_out$GRADING))] <- 0
  }

  # summary data ---------------------------------------------------------------
  if (dq_param == "location") {
    res_view <- data.frame("Variables" = res_dq_check$Variables,
                           "Measure of location" = res_dq_check$loc_func,
                           "Value" = res_dq_check$values_from_data,
                           "Range of expected values" =
                             res_dq_check$loc_range,
                           "Flag" = res_dq_check$FLG_acc_ud_loc,
                           check.names = FALSE)
  } else if (dq_param == "proportion") {
    res_view <- data.frame("Variables" = res_dq_check$Variables,

                           "Proportions" = res_dq_check$values_from_data,
                           "Range of expected values" =
                             res_dq_check$prop_range,
                           "Flag" = res_dq_check$FLG_acc_ud_prop,
                           check.names = FALSE)
  }

  # plot -----------------------------------------------------------------------
  ref_env <- environment()
  plot_list <- lapply(setNames(nm = resp_vars), function(rv) {

    # find suitable lables
    lb <- paste0(prep_get_labels(rv,
                                 item_level = meta_data,
                                 label_col = label_col,
                                 resp_vars_match_label_col_only = TRUE,
                                 label_class = "SHORT"))

    # omit NAs from data to prevent ggplot2 warning messages
    ds1 <- ds1[!(is.na(ds1[[rv]])), , drop = FALSE]

    # Should the plot be a histogram? If not, it will be a bar chart.
    plot_histogram <-
      meta_data[[SCALE_LEVEL]][which(meta_data[[label_col]] == rv)] %in%
      c(SCALE_LEVELS$INTERVAL, SCALE_LEVELS$RATIO)

    txtspec <- element_text(
      colour = "black", hjust = .5,
      vjust = .5, face = "plain", size = 10
    )

    col_bars <- "#2166AC"
    blue_red <- hcl.colors(10, "Plasma")[c(1,6)]

    if (plot_histogram) {
      # histogram --------------------------------------------------------------
      .dt <- ds1[, rv, drop = FALSE]
      rv_datetime <- is_datetime_var[rv]
      rv_time <- is_time_var[rv]
      p <- util_create_lean_ggplot(
        util_histogram(plot_data = .dt,
                       is_datetime = rv_datetime,
                       is_time = rv_time) +
          xlab(lb),
        .dt = .dt,
        rv_datetime = rv_datetime,
        lb = lb,
        rv_time = rv_time
      )

      obj1 <- ggplot2::ggplot_build(p)
      no_char_y <- max(nchar(util_gg_get(obj1, "data")[[1]]$ymax))
      no_char_x <- 4

    } else {
      # bar chart --------------------------------------------------------------
      if (!is.factor(ds1[[rv]])) {
        lvls <- union(names(rvs_meta$Range[[rv]]), unique(ds1[[rv]]))
      } else {
        lvls <- levels(ds1[[rv]])
      }
      plot_data <- data.frame("cat" = factor(lvls, levels = lvls),
                              "count" = unlist(lapply(lvls, function(ll) {
                                length(which(ds1[[rv]] == ll)) })),
                              "col" = col_bars)
      if (plot_ranges & any(!is.na(flg_cat_prop[[rv]]))) {
        # if bars outside of expected ranges shall be highlighted
        plot_data[plot_data[["cat"]] %in% flg_cat_prop[[rv]], "col"] <- blue_red[2]
      }

      p <- util_bar_plot(plot_data = plot_data,
                         cat_var = "cat",
                         num_var = "count",
                         fill_var = "col",
                         colors = unique(plot_data[, "col"]),
                         relative = FALSE, show_numbers = FALSE, flip = FALSE)
      p <- p %lean+% util_create_lean_ggplot(xlab(lb), lb = lb)

      obj1 <- ggplot2::ggplot_build(p)
      no_char_y <- max(nchar(util_gg_get(obj1, "data")[[1]]$ymax))
      no_char_x <- max(nchar(lvls))
    }

    # add ranges (if needed) ---------------------------------------------------
    if (plot_ranges) {
      # location
      if (dq_param == "location" & rv %in% rvs_with_meta) {
        # lower limit
        ll <- rvs_meta$Range[[rv]][["low"]]
        ll <- ifelse(is.infinite(ll), NA, ll)
        if (!is.na(ll)) {
          p <- p %lean+% util_create_lean_ggplot(geom_vline(xintercept = ll,
                              linetype = 2,
                              col = blue_red[1]),
                              ll = ll,
                              blue_red = blue_red)
        }
        # upper limit
        ul <- rvs_meta$Range[[rv]][["upp"]]
        ul <- ifelse(is.infinite(ul), NA, ul)
        if (!is.na(ul)) {
          p <- p %lean+% util_create_lean_ggplot(geom_vline(xintercept = ul,
                              linetype = 2,
                              col = blue_red[1]),
                              blue_red = blue_red,
                              ul = ul)
        }
        # add mean or median as line
        loc_val <- dq_check_list[[rv]][["values_from_data"]]
        if (!is.na(loc_val)) {
          loc_flg <- dq_check_list[[rv]][["FLG_acc_ud_loc"]]
          if (!is.na(loc_flg) & !loc_flg) {
            loc_col <- blue_red[1]
          } else {
            loc_col <- blue_red[2]
          }
          p <- p %lean+% util_create_lean_ggplot(geom_vline(xintercept = loc_val,
                              col = loc_col),
                              loc_col = loc_col,
                              loc_val = loc_val)
        }
      }
      # proportions
      if (dq_param == "proportion" & rv %in% rvs_with_meta) {
        # add lines for each category (if available)
        # Create a list of small data frames, one per interval category
        x1 <- NULL # to
        x2 <- NULL # make
        y1 <- NULL # R CMD check
        y2 <- NULL # happy
        segments_list <- lapply(
          seq_along(rvs_meta$Range[[rv]]),
          function(pp) {
            interval_pp <- rvs_meta$Range[[rv]][[pp]]
            if (!inherits(interval_pp, "interval")) return(NULL)

            # Determine numeric position for this category
            pp_num <- if (is.factor(ds1[[rv]])) {
              pp
            } else {
              as.numeric(names(rvs_meta$Range[[rv]])[pp])
            }

            # Collect lower and upper limit frames
            out <- list()

            ll <- interval_pp[["low"]]
            if (!is.infinite(ll) && !is.na(ll)) {
              out[[length(out) + 1]] <- data.frame(
                x1 = pp_num - 0.5,
                x2 = pp_num + 0.5,
                y1 = ll  / 100 * nrow(ds1),
                y2 = ll  / 100 * nrow(ds1)
              )
            }

            ul <- interval_pp[["upp"]]
            if (!is.infinite(ul) && !is.na(ul)) {
              out[[length(out) + 1]] <- data.frame(
                x1 = pp_num - 0.5,
                x2 = pp_num + 0.5,
                y1 = ul  / 100 * nrow(ds1),
                y2 = ul  / 100 * nrow(ds1)
              )
            }

            # Return a single data frame or NULL if nothing to draw
            if (length(out) == 0) return(NULL)
            util_rbind(data_frames_list = out)
          }
        )

        # Remove NULL entries and combine into one data frame
        segments_list <- Filter(Negate(is.null), segments_list)
        all_segments  <- if (length(segments_list) > 0) {
          util_rbind(data_frames_list = segments_list)
        } else {
          NULL
        }

        # If we have any segment data, add them in one lean call
        if (!is.null(all_segments)) {
          p <- p %lean+% util_create_lean_ggplot(
            {
              # Add all segments in one geom_segment layer
              geom_segment(
                aes(
                  x    = x1,
                  xend = x2,
                  y    = y1,
                  yend = y2
                ),
                data     = all_segments,
                linetype = 2,
                col      = blue_red[1]
              )
            },
            all_segments = all_segments,
            aes          = ggplot2::aes,
            blue_red = blue_red,
            geom_segment = ggplot2::geom_segment
          )
        }
      }
    }

    fli <- util_coord_flip(p = p, ref_env = ref_env)
    is_flipped <- inherits(fli, "CoordFlip")

    p <- util_lazy_add_coord(p, fli) %lean+%
      # TODO: estimate w and h, if p is not using discrete axes
      ylab("") %lean+%
      theme_minimal() %lean+%
      theme(
        title = txtspec,
        axis.text.x = txtspec,
        axis.text.y = txtspec,
        axis.title.x = txtspec,
        axis.title.y = txtspec
      )

    # Adust the following, if changing the bar chart plots above:
    #    if (length(P$layers) == 1 &&
    #        inherits(P$layers[[1]]$geom, "GeomBar")) {
    # bar chart sizing_hints attributes, can also be attached to the
    # figures, alone
    #        .tb <- table(P$data)
    #       attr(P, "sizing_hints") <- list(
    #         figure_type_id = "bar_chart",
    #         rotated = is_flipped,
    #         number_of_bars = dim(.tb),
    #         min_bar_height = min(.tb),
    #         max_bar_height = max(.tb)
    #       )
    #   }

    util_set_size(p)
  })
  #TODO: get here the values for the attributes from the plot
  # Add sizing hints only if resp_vars is one
  if (length(resp_vars)==1) {
    #sizing information
    if (!is.null(plot_list[[resp_vars]])) {
      obj1 <- ggplot2::ggplot_build(plot_list[[resp_vars]])
      #in case of ecdf
      if (length(util_gg_get(obj1, "layout")$facet_params$.possible_columns) == 2) {
        min_bar_height <- 0
        max_bar_height <- 0
        range_bar <- 400 #an intermediate size plot in height
      } else {
        # in case of a single ditribution plot
        min_bar_height <- min(util_rbind(data_frames_list =
                                           util_gg_get(obj1, "data"))$ymax,
                              na.rm = TRUE)
        max_bar_height <- max(util_rbind(data_frames_list =
                                           util_gg_get(obj1, "data"))$ymax,
                              na.rm = TRUE)
        range_bar <- max_bar_height - min_bar_height

        if(min_bar_height == Inf ||
           min_bar_height == -Inf ||
           max_bar_height == Inf ||
           max_bar_height == -Inf ) {
          range_bar <- 400 #an intermediate size plot in height
        }
      }

      rotated <- unique(util_rbind(data_frames_list =
                                     util_gg_get(obj1, "data"))$flipped_aes)
      rotated <-  rotated[!is.na(rotated)]
      if (inherits(ds1[[resp_vars]], "POSIXct")) {
        number_of_bars <- 10
      } else {
        #    number_of_bars <- nrow(util_rbind(data_frames_list =
        # util_gg_get(obj1, "data")$data))
        #fix number in case of acc_distrib_prop o loc
        #    if (check_param == "proportion" || check_param == "location") {
        #      number_of_bars <- nrow(obj1$data[[1]])
        #    }
        number_of_bars <- nrow(util_gg_get(obj1, "data")[[1]])

        # fix number in case there are categories not listed in the obj1,
        # example OBS_BP_0 in ship example data
        if (is.factor(ds1[[resp_vars]])) {
          no_levels <- length(levels(ds1[[resp_vars]]))
          number_of_bars <- max(c(number_of_bars,no_levels))
        }
      }

      no_char_y <- max(nchar(obj1$data[[1]]$ymax))

      if(meta_data[meta_data[[label_col]] == resp_vars,
                   SCALE_LEVEL, drop = FALSE] %in% c("ratio", "interval")) {
        no_char_x <- 4
      } else {
        exp_bars <- levels(ds1[[resp_vars]])
        no_char_x <- max(nchar(exp_bars))
      }
      rm(obj1)
    }
    return(util_attach_attr(list(SummaryTable = res_out,
                                 SummaryData = res_view,
                                 SummaryPlotList = plot_list),
                            sizing_hints = list(figure_type_id = "bar_chart",
                                                rotated = rotated,
                                                number_of_bars = number_of_bars,
                                                range = range_bar,
                                                no_char_x = no_char_x,
                                                no_char_y = no_char_y),
                            as_plotly = "util_as_plotly_acc_distributions"))
  } else {
    return(util_attach_attr(list(SummaryTable = res_out,
                                 SummaryData = res_view,
                                 SummaryPlotList = plot_list),
                            as_plotly = "util_as_plotly_acc_distributions"))
  }


}






#' Plots and checks for distributions -- Location
#' @inherit acc_distributions
#' @export
#' @seealso
#' - [acc_distributions]
#' - [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_distributions.html
#' )
acc_distributions_loc <- function(resp_vars = NULL,
                                  study_data,
                                  label_col = VAR_NAMES,
                                  item_level = "item_level",
                                  check_param = "location",
                                  plot_ranges = TRUE,
                                  flip_mode = "noflip",
                                  meta_data = item_level,
                                  meta_data_v2) {
  util_maybe_load_meta_data_v2()
  rvs_meta <- util_prep_location_check(resp_vars = resp_vars,
                                       meta_data = meta_data,
                                       report_problems = "error",
                                       label_col = label_col)
  acc_distributions(
    resp_vars = resp_vars, study_data = study_data,
    meta_data = meta_data, label_col = label_col,
    check_param = "location", plot_ranges = plot_ranges,
    flip_mode = flip_mode
  )
}

#' Plots and checks for distributions -- Proportion
#' @inherit acc_distributions
#' @export
#' @seealso
#' - [acc_distributions]
#' - [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_distributions.html
#' )
acc_distributions_prop <- function(resp_vars = NULL,
                                   study_data,
                                   label_col,
                                   item_level = "item_level",
                                   check_param = "proportion",
                                   plot_ranges = TRUE,
                                   flip_mode = "noflip",
                                   meta_data = item_level,
                                   meta_data_v2) {
  util_maybe_load_meta_data_v2()
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          #.apply_factor_metadata = TRUE, # can be omitted in favor of .apply_factor_metadata_inadm
                          .apply_factor_metadata_inadm = TRUE
                          )
  md_prep <- util_prep_proportion_check(
    resp_vars = resp_vars,
    meta_data = meta_data,
    ds1 = ds1,
    report_problems = "error",
    label_col = label_col
  )

  acc_distributions(
    resp_vars = resp_vars, study_data = study_data,
    meta_data = meta_data, label_col = label_col,
    check_param = "proportion", plot_ranges = plot_ranges,
    flip_mode = flip_mode
  )
}

#' Plots and checks for distributions -- only
#'
#' [Descriptor]
#'
#' @inherit acc_distributions
#' @export
#' @seealso
#' - [acc_distributions]
#' - [Online Documentation](
#' https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_distributions.html
#' )
acc_distributions_only <- function(resp_vars = NULL,
                                   study_data,
                                   label_col = VAR_NAMES,
                                   item_level = "item_level",
                                   flip_mode = "noflip",
                                   meta_data = item_level,
                                   meta_data_v2) {
  util_maybe_load_meta_data_v2()
  loc_avail <- !inherits(try(util_prep_location_check(resp_vars = resp_vars,
                                       meta_data = meta_data,
                                       report_problems = "error",
                                       label_col = label_col),
                   silent = TRUE), "try-error")
  if (.called_in_pipeline && loc_avail) {
    util_error("%s is already in a distribution location check plot figure",
               dQuote(resp_vars),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata = TRUE, # can be omitted in favor of .apply_factor_metadata_inadm
                          .apply_factor_metadata_inadm = TRUE)
  prop_avail <- !inherits(try(util_prep_proportion_check(
                    resp_vars = resp_vars,
                    meta_data = meta_data,
                    ds1 = ds1,
                    report_problems = "error",
                    label_col = label_col
                  ), silent = TRUE), "try-error")
  if (.called_in_pipeline && prop_avail) {
    util_error("%s is already in a distribution proportion check plot figure",
               dQuote(resp_vars),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  scl_lvl <- meta_data[[SCALE_LEVEL]][meta_data[[label_col]] == resp_vars]
  if (.called_in_pipeline && scl_lvl %in% c(SCALE_LEVELS$NOMINAL, SCALE_LEVELS$ORDINAL)) {
    util_error("%s is already in a distribution plot (cat.) figure",
               dQuote(resp_vars),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  acc_distributions(
    resp_vars = resp_vars, study_data = study_data,
    meta_data = meta_data, label_col = label_col,
    flip_mode = flip_mode
  )
}


#' @family plotly_shims
#' @concept plotly_shims
#' @noRd
util_as_plotly_acc_distributions <- function(res, ...) {
  if (length(res$SummaryPlotList) != 1) {
    return(util_ggplotly(ggplot2::ggplot() +
                              ggplot2::annotate("text", x = 0, y = 0,
                                                label =
  sprintf(paste("Internal error: I should  have exactly 1 result, if",
        "calling plotly for a dq_report2 otuput. I have %d."),
        length(res$SummaryPlotList))) +
                              theme(
                                axis.line = element_blank(),
                                axis.text.x = element_blank(),
                                axis.text.y = element_blank(),
                                axis.ticks = element_blank(),
                                axis.title.x = element_blank(),
                                axis.title.y = element_blank(),
                                legend.position = "none",
                                panel.background = element_blank(),
                                panel.border = element_blank(),
                                panel.grid.major = element_blank(),
                                panel.grid.minor = element_blank(),
                                plot.background = element_blank()
                              )))
  }
  res$SummaryPlotList <-
    util_remove_dataquieR_result_class(res$SummaryPlotList)
  # use res$SummaryPlot, not something from the enclosure
  # of the result, that may contain study data.
  util_ensure_suggested("plotly")

  res$SummaryPlot <- res$SummaryPlotList[[1]]

  withCallingHandlers(
    {
      py <- util_ggplotly(res$SummaryPlot, ...) # xxxx remove withCallingHandlers after removing xxxxx from util_histogram
    },
    warning = function(w) {
      if (grepl("position_stack", conditionMessage(w), fixed = TRUE)) {
        invokeRestart("muffleWarning")
      }
    }
  )

  plotly::layout(py, xaxis = list(tickangle = "auto"))
}

test_that("acc_distributions catches unsuitable input and works otherwise", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)

  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  # data type mismatch between study data and metadata will be converted
  # -> runs without problems
  sd0 <- study_data
  sd0$v00000 <- as.character(sd0$v00000)
  md0 <- meta_data
  expect_silent(
    res1 <- acc_distributions(
        resp_vars ="v00000",
        study_data = sd0, meta_data = md0)
  )

  # all values NA
  sd0 <- study_data
  sd0$v00000 <- NA_real_
  expect_error(
      res1 <-
        acc_distributions(
          resp_vars <- "v00000",
          study_data = sd0, meta_data = md0)
    ,
    regexp =
      "Variable .+v00000.+resp_vars.+ has only NA observations",
    perl = TRUE
  )

  # too few distinct values
  sd0 <- study_data
  sd0$v00000 <- 0
  # error if there is only one variable
  expect_error(
    res1 <-
      acc_distributions(
       resp_vars <- "v00000",
       study_data = sd0, meta_data = md0)
   ,
   regexp =
     paste("Variable .+v00000.+resp_vars.+ has fewer distinct values",
           "than required"),
   perl = TRUE
  )
  # warning if there are other variables that can be used by the function
  expect_message2(
    expect_warning(
      res1 <-
        acc_distributions(
          resp_vars <- c("v00000", "v00002"),
          study_data = sd0, meta_data = md0),
      regexp = "In .+resp_vars.+ variables .+v00000.+ were excluded.",
      perl = TRUE),
    regexp = paste("Variable .+v00000.+resp_vars.+ has fewer distinct values",
                   "than required"),
    perl = TRUE
  )

  # check that only suitable variables are selected
  expect_warning(suppressMessages(
    res1 <-
      acc_distributions(study_data = study_data, meta_data = meta_data)))
  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    39
  )

  # only two string variables in the study data, but not specified as resp_vars
  # thus, the function will only search for integer and float variables
  # (currently) and not use the two given string variables
  md0 <- subset(meta_data, VAR_NAMES %in% c("v00001", "v00103"))
  md0$KEY_STUDY_SEGMENT <- NA
  sd0 <- study_data[, c("v00001", "v00103")]
  expect_error(
    suppressWarnings(suppressMessages(
      res1 <-
        acc_distributions(
          study_data = sd0, meta_data = md0)
    )),
    regexp =
      "No suitable variables were defined for acc_distributions."
  )
})

test_that("acc_distributions works with label_col", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  expect_warning(suppressMessages(
    res1 <-
      acc_distributions(study_data = study_data, meta_data = meta_data,
                        label_col = LABEL)))
  expect_true("SummaryPlotList" %in% names(res1))
  expect_equal(
    length(res1$SummaryPlotList),
    39
  )
})


test_that("acc_distributions robust with miss-codes", {
  skip_on_cran() # slow test.
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  md0 <- meta_data
  sd0 <- study_data
  sd0$v00003[8:50] <- 8:50
  sd0$v00003[[6]] <- 9999999
  expect_silent(
    res1 <-
      acc_distributions(resp_vars = "v00003",
      study_data = sd0, meta_data = md0))
  sd0$v00003[[6]] <- 10000000
  expect_silent(
    res1 <-
      acc_distributions(resp_vars = "v00003",
                        study_data = sd0, meta_data = md0))
})

test_that("acc_distributions is robust to other issues", {
  skip_on_cran() # slow, errors obvious
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  meta_data <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = meta_data)

  # proportion check works with integers without value labels:
  expect_silent(
    res_int <- acc_distributions_prop(resp_vars = "ITEM_4_0",
                                      study_data = study_data,
                                      meta_data = meta_data,
                                      label_col = LABEL))
  # proportion check works with value labels specified:
  expect_silent(
    res_val <- acc_distributions_prop(resp_vars = "SEX_0",
                                      study_data = study_data,
                                      meta_data = meta_data,
                                      label_col = LABEL))

  md0 <- meta_data
  sd0 <- study_data
  # drop the last category, which is given as value label - this should get
  # caught by the function
  sd0$v00000[which(sd0$v00000 == 5)] <- 2
  expect_silent(
    res1 <-
      acc_distributions(resp_vars = "v00000",
                        study_data = sd0, meta_data = md0)
  )
  skip_if_not_installed("vdiffr")
  skip_on_cran()
  expect_doppelganger2("empty_cat5",
                              res1$SummaryPlotList$v00000)

  # use non-consecutive codes for categories
  sd0$v00000[which(sd0$v00000 == 3)] <- 33
  sd0$v00000[which(sd0$v00000 %in% c(2,4))] <- 111
  md0$VALUE_LABEL_TABLE[which(md0$VAR_NAMES == "v00000")] <- ""
  md0$VALUE_LABELS[which(md0$VAR_NAMES == "v00000")] <-
    "33 = Berlin | 1 = Hamburg | 111 = Cologne"

  {
    res2 <- acc_distributions_prop(resp_vars = "v00000",
                                   study_data = sd0, meta_data = md0)
  } %>%
    expect_warning(regexp = "Cannot mix")

  skip_if_not_installed("vdiffr")
  skip_on_cran()
  expect_doppelganger2("non_consecutive_codes_work",
                              res2$SummaryPlotList$v00000)

  # only few different float values, should still produce a histogram
  # (here with four bars, a bar chart would have five bars)
  sl <- meta_data$SCALE_LEVEL
  prep_purge_data_frame_cache()
  meta_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data.RData")
  meta_data[[SCALE_LEVEL]] <- sl
  study_data <- prep_get_data_frame("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/study_data.RData", keep_types = TRUE)
  md0 <- meta_data
  sd0 <- study_data
  sd0$v00009 <- rep(c(19.2, 19.9, 22.5, 25.7, 29.4), each = 600)
  expect_silent(
    res3 <- acc_distributions_loc(resp_vars = "v00009",
                                  study_data = sd0, meta_data = md0)
  )
  skip_if_not_installed("vdiffr")
  skip_on_cran()
  expect_doppelganger2("few_float_values",
                              res3$SummaryPlotList$v00009)
  skip_if_offline(host = "dataquality.qihs.uni-greifswald.de")
  prep_load_workbook_like_file("https://dataquality.qihs.uni-greifswald.de/extdata/fortests/meta_data_v2.xlsx")
  md0 <- prep_meta_data_v1_to_item_level_meta_data()
  md0 <-
    prep_scalelevel_from_data_and_metadata(study_data = study_data,
                                           meta_data = md0)

  sd0 <- study_data
  md0 <-
    prep_scalelevel_from_data_and_metadata(study_data = sd0,
                                           meta_data = md0)
  md1 <- md0
  md1[md1$LABEL == "SEX_0", VALUE_LABELS] <-
    "0 = females | 1 = males"
  suppressWarnings(sd0$v00002 <- as.character(util_study_var2factor(
    "SEX_0", study_data = sd0, code_name = c(VALUE_LABELS),
    meta_data = md1, include_sysmiss = FALSE)[["SEX_0"]]
  ))
  expect_error(
    acc_distributions(study_data = sd0, meta_data = md0, resp_vars = "SEX_0",
                      label_col = "LABEL")
  )

  md0[[DATA_TYPE]][md0[[LABEL]] == "SEX_0"] <- "string"
  expect_silent(
    acc_distributions(study_data = sd0, meta_data = md0, resp_vars = "SEX_0",
                      label_col = "LABEL")
  )
})

#' Extract report summary from reports
#'
#' @param r [dq_report2] a [dq_report2] report
#' @param ... not used
#'
#' @return [list] with two slots `Data` and `Table` with [data.frame]s
#'                featuring all metrics columns
#'                from the report `r`, the [STUDY_SEGMENT] and the [VAR_NAMES].
#'                In case of `Data`, the columns are formatted nicely but still
#'                with the standardized column names -- use
#'                `util_translate_indicator_metrics()` to rename them nicely. In
#'                case of `Table`, just as they are.
#' @family summary_functions
#' @seealso [prep_combine_report_summaries()]
#' @export
prep_extract_summary.dataquieR_resultset2 <- function(
    r,
    ...
    ) {
  te <- topenv(parent.frame(1)) # see https://stackoverflow.com/a/27870803
  if (!(isNamespace(te) && getNamespaceName(te) == "dataquieR")) {
    lifecycle::deprecate_soft("2.1.0.9007",
                              "prep_extract_summary.dataquieR_resultset2()",
                              "summary()")
  }
  util_stop_if_not(
    "Can only be called for dq_report2 objects of class dataquieR_resultset2" =
      inherits(r, "dataquieR_resultset2"))
  sts <-
    lapply(setNames(nm = rownames(r)), function(v) {
      all_cll <- lapply(setNames(nm = colnames(r)), function(cll) {
        st <- r[v, cll, res = "SummaryTable", drop = TRUE] # TODO: SegmentTable, ...
        if (is.data.frame(st)) {
          st <- util_extract_indicator_metrics(st)
          if (nrow(st) == 1 && ncol(st) > 0) {
            return(st)
          }
        }
        NULL
      })
      all_cll[vapply(all_cll, is.null, FUN.VALUE = logical(1))] <- NULL
      all_cll
    })
  sts[vapply(sts, length, FUN.VALUE = integer(1)) == 0] <- NULL
  sts <- lapply(sts, function(st) {
    st <- lapply(names(st), function(stnm) {
      colnames(st[[stnm]]) <-
        paste(stnm, colnames(st[[stnm]]), sep = ".")
      st[[stnm]]
    })
    r <- do.call(cbind, st)
    r
  })
  res <- util_rbind(data_frames_list = sts)


  if (STUDY_SEGMENT %in% colnames(attr(r, "meta_data"))) {
    res[[STUDY_SEGMENT]] <-
      prep_map_labels(
        rownames(res),
        meta_data = attr(r, "meta_data"),
        from = attr(r, "label_col"),
        to = STUDY_SEGMENT)
    if (length(res[[STUDY_SEGMENT]]) == 0)
      res[[STUDY_SEGMENT]] <- character(0)
  } else {
    res[[STUDY_SEGMENT]] <- rep("Study", nrow(res))
  }

  if (length(rownames(res)) == 0) {
    res[[VAR_NAMES]] <-
      character(0)
  } else {
    res[[VAR_NAMES]] <-
      prep_map_labels(
        rownames(res),
        meta_data = attr(r, "meta_data"),
        from = attr(r, "label_col"),
        to = VAR_NAMES)
  }

  res_raw <- res

  counts <- vapply(colnames(res), FUN.VALUE = logical(1),
                   FUN = function(x) {
                     util_stop_if_not(length(x) == 1)
                     x <- sub("^[^\\.]+\\.", "", x)
                     nm <- strsplit(x, "_", fixed = TRUE)[[1]]
                     if (length(nm) >= 2) {
                       identical(nm[[1]], "NUM")
                     } else {
                       FALSE
                     }
                   })
  res_raw[, counts] <- lapply(res[, counts, FALSE], as.numeric)
  res[, counts] <- lapply(lapply(res[, counts, FALSE], as.numeric),
                          scales::number, accuracy = 1)

  percentages <- vapply(colnames(res), FUN.VALUE = logical(1),
                        FUN = function(x) {
                          util_stop_if_not(length(x) == 1)
                          x <- sub("^[^\\.]+\\.", "", x)
                          nm <- strsplit(x, "_", fixed = TRUE)[[1]]
                          if (length(nm) >= 2) {
                            identical(nm[[1]], "PCT")
                          } else {
                            FALSE
                          }
                        })
  res_raw[, percentages] <- lapply(res_raw[, percentages, FALSE], as.numeric)
  res[, percentages] <- lapply(lapply(res[, percentages, FALSE], as.numeric),
                               scales::percent, , scale = 1, accuracy = 0.01)

  flags <- vapply(colnames(res), FUN.VALUE = logical(1),
                  FUN = function(x) {
                    util_stop_if_not(length(x) == 1)
                    x <- sub("^[^\\.]+\\.", "", x)
                    nm <- strsplit(x, "_", fixed = TRUE)[[1]]
                    if (length(nm) >= 2) {
                      identical(nm[[1]], "FLG")
                    } else {
                      FALSE
                    }
                  })
  res_raw[, flags] <- lapply(res_raw[, flags, FALSE], as.logical)
  res[, flags] <- lapply(lapply(res[, flags, FALSE], as.logical),
                         ifelse, "T", "F")

  res <- res[, sort(colnames(res))]
  res_raw <- res_raw[, sort(colnames(res_raw))]

  r <- list(Data = res, Table = res_raw, meta_data = attr(r, "meta_data"))
  class(r) <- "dq_report2_summary"
  r
}

#' Utility function to compute the rank intraclass correlation
#'
#' This implementation uses the package `rankICC` to compute the rank
#' intraclass correlation, a nonparametric version of the ICC (Tu et al., 2023).
#' In contrast to model-based ICC approaches, it is less sensitive to outliers
#' and skewed distributions. It can be applied to variables with an ordinal,
#' interval or ratio scale. However, it is not possible to adjust for
#' covariables with this approach. The calculated ICC can become negative,
#' like Fisher's ICC.
#'
#' @param resp_vars the name of the response variable
#' @param group_vars the name of the grouping variable
#' @param study_data the data frame that contains the measurements
#' @param meta_data the data frame that contains metadata attributes of
#'                  study data
#' @param min_obs_in_subgroup the minimum number of observations that is
#'                            required to include a subgroup (level) of the
#'                            grouping variable (`group_vars`) in the analysis.
#'                            Subgroups with fewer observations are excluded.
#' @param min_subgroups the minimum number of subgroups (levels) of the
#'                      grouping variable (`group_vars`). If the variable has
#'                      fewer subgroups, the analysis is not performed.
#' @param label_col the name of the column in the metadata with labels of
#'                  variables
#'
#' @return a vector from rankICC::rankICC
#'
#' @noRd
util_varcomp_robust <- function(resp_vars = NULL, group_vars = NULL,
                                study_data = study_data, meta_data = meta_data,
                                min_obs_in_subgroup = 10, min_subgroups = 5,
                                label_col = NULL) {
  # preps ----------------------------------------------------------------------
  util_ensure_suggested("rankICC")
  prep_prepare_dataframes(.replace_hard_limits = TRUE,
                          .apply_factor_metadata = TRUE)

  util_correct_variable_use("resp_vars",
                            allow_all_obs_na = FALSE,
                            need_scale = "ordinal | interval | ratio")
  util_correct_variable_use("group_vars",
                            allow_all_obs_na = FALSE,
                            need_type = "!float",
                            need_scale = "nominal | ordinal")

  util_expect_scalar(min_obs_in_subgroup,
                     check_type = util_is_numeric_in(min = 0,
                                                     whole_num = TRUE,
                                                     finite = TRUE))
  util_expect_scalar(min_subgroups,
                     check_type = util_is_numeric_in(min = 0,
                                                     whole_num = TRUE,
                                                     finite = TRUE))

  ds1 <- ds1[, c(resp_vars, group_vars), drop = FALSE]
  ds1 <- ds1[complete.cases(ds1), , drop = FALSE]

  # ensure that included levels of the grouping variable have the specified
  # minimum number of observations
  check_df <- util_table_of_vct(ds1[[group_vars]])
  critical_levels <- levels(check_df$Var1)[check_df$Freq <
                                             min_obs_in_subgroup]
  if (length(critical_levels) > 0) {
    util_message("Levels %s were excluded due to less than %d observations.",
                 paste0(c(vapply(head(critical_levels, 10), dQuote, ""),
                          if (length(critical_levels) <= 10)
                            character(0) else "..."),
                        collapse = ", "),
                 min_obs_in_subgroup,
                 applicability_problem = FALSE)
    # exclude levels with too few observations
    ds1 <- ds1[!(ds1[[group_vars]] %in% critical_levels), ]
    levels(ds1[[group_vars]])[
      which(levels(ds1[[group_vars]]) %in% critical_levels)] <- NA
  }
  ds1 <- ds1[complete.cases(ds1), , drop = FALSE]

  # ensure that the grouping variable has the specified minimum number of levels
  check_df <- util_table_of_vct(ds1[[group_vars]])
  if (length(check_df[, 1]) < min_subgroups) {
    util_error("%d < %d levels in %s. Will not compute ICCs for %s.",
               length(check_df[, 1]),
               min_subgroups,
               dQuote(group_vars),
               dQuote(resp_vars),
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  # ensure that the response variable is not constant
  if (length(unique(ds1[[resp_vars]])) < 2) {
    util_error("The response variable is constant after data preparation.",
               applicability_problem = TRUE,
               intrinsic_applicability_problem = TRUE)
  }

  res <- rankICC::rankICC(x = ds1[[resp_vars]],
                          cluster = ds1[[group_vars]],
                          weights = "clusters")

  return(res)
}

#' Distribute `CODE_LIST_TABLE` in item level metadata
#'
#' fills the columns `MISSING_LIST_TABLE` and `VALUE_LABEL_TABLE` from
#' `CODE_LIST_TABLE`, if applicable
#'
#' @inheritParams .template_function_indicator
#'
#' @return [meta_data], but `CODE_LIST_TABLE` column is distributed to the
#'         columns `VALUE_LABEL_TABLE` and `MISSING_LIST_TABLE`, respectively.
#' @noRd
util_normalize_clt <- function(meta_data) {
  if (!(CODE_LIST_TABLE %in% names(meta_data))) {
    return(meta_data)
  }
  util_handle_val_tab()
  is_mlt <- vapply(
    FUN.VALUE = logical(1),
    setNames(nm = na.omit(unique(meta_data[[CODE_LIST_TABLE]]))),
    function(tb) {
      if (util_empty(tb)) return(FALSE)
      tb <- try(prep_get_data_frame(tb), silent = TRUE)
      if (util_is_try_error(tb)) return(FALSE)
      "MISSING" %in% tb[[CODE_CLASS]] ||
        "JUMP" %in% tb[[CODE_CLASS]]
    })
  is_vlt <- vapply(
    FUN.VALUE = logical(1),
    setNames(nm = na.omit(unique(meta_data[[CODE_LIST_TABLE]]))),
    function(tb) {
      if (util_empty(tb)) return(FALSE)
      tb <- try(prep_get_data_frame(tb), silent = TRUE)
      if (util_is_try_error(tb)) return(FALSE)
      (!(CODE_CLASS %in% names(tb))) || "VALUE" %in% tb[[CODE_CLASS]]
    })
  if (any(!util_empty(meta_data[[MISSING_LIST_TABLE]][is_mlt[
    meta_data[[CODE_LIST_TABLE]]]]))) {
    util_message(c("Found %s where also %s has been assigned, discarding %s",
                   "for such items"),
                 sQuote(MISSING_LIST_TABLE),
                 sQuote(CODE_LIST_TABLE),
                 sQuote(MISSING_LIST_TABLE),
                 applicability_problem = TRUE)
  }
  if (any(!util_empty(meta_data[[VALUE_LABEL_TABLE]][is_vlt[
    meta_data[[CODE_LIST_TABLE]]]]))) {
    util_message(c("Found %s where also %s has been assigned, discarding %s",
                   "for such items"),
                 sQuote(VALUE_LABEL_TABLE),
                 sQuote(CODE_LIST_TABLE),
                 sQuote(VALUE_LABEL_TABLE),
                 applicability_problem = TRUE)
  }
  code_list_table <- meta_data[[CODE_LIST_TABLE]]
  code_list_table[util_empty(code_list_table)] <- "//**//??"
  is_mlt["//**//??"] <- FALSE
  is_vlt["//**//??"] <- FALSE
  meta_data[[MISSING_LIST_TABLE]][is_mlt[code_list_table]] <-
    meta_data[[CODE_LIST_TABLE]][is_mlt[code_list_table]]
  meta_data[[VALUE_LABEL_TABLE]][is_vlt[code_list_table]] <-
    meta_data[[CODE_LIST_TABLE]][is_vlt[code_list_table]]
  meta_data[[CODE_LIST_TABLE]] <- NULL
  meta_data
}

#' Compress an R object using available algorithms
#'
#' Serializes and compresses an R object using one of the available compression
#' algorithms supported by `memCompress()`. If no algorithm is specified, the
#' function selects the first available one from a predefined priority list.
#'
#' @param x Any R object to be serialized and compressed.
#' @param algo Character vector of candidate algorithms (e.g., `"xz"`, `"gzip"`,
#'   `"zstd"`, `"bzip2"`, `"none"`). If missing, defaults are chosen depending
#'   on the R version.
#'
#' @returns A raw vector of class `"compressed"` with attributes describing the
#'   method and algorithm used.
#'
#' @noRd
util_compress <- function(x, algo)  {
  #structure(memCompress(serialize(x, NULL, xdr = FALSE), "xz"),
  if (missing(algo)) {
    if (R.version$major > 4 || (R.version$major == 4 && R.version$minor >= 5))
      algo <- c("zstd", "gzip", "bzip2", "xz", "none")
    else
      algo <- c("gzip", "bzip2", "xz", "none")
  }
  algo <- head(intersect(
    algo,
    MEM_COMPRESS_CAPABILITIES), 1)
  structure(memCompress(serialize(x, NULL, xdr = FALSE), type = algo),
      class = "compressed",
      method = "memCompress",
      algo = algo)
  # qs::qserialize(x, preset = "high")
}

#' Decompress an object created by `util_compress()`
#'
#' Decompresses and un-serializes an object previously compressed with
#' `util_compress()`. Only objects of class `"compressed"` are supported.
#'
#' @param x A raw vector of class `"compressed"` produced by
#'   `util_compress()`.
#'
#' @returns The original R object restored after decompression.
#'
#' @noRd
util_decompress <- function(x)  {
  if (inherits(x, "compressed")) {
    algo <- attr(x, "algo")
    if (is.null(algo)) {
      algo <- head(MEM_COMPRESS_CAPABILITIES, 1)
    }
    unserialize(memDecompress(x, type = algo))
  } else {
    util_error("Decompression of objects of class %s unsupported",
               util_pretty_vector_string(class(x), quote = sQuote))
  }
  # qs::qdeserialize(r)
}

.util_mem_compress_capabilities <- function(types = c("gzip", "bzip2", "xz",
                                                      "zstd", "none")) {
  test_raw <- charToRaw("test")
  setNames(vapply(types, function(t) {
    !inherits(try(memCompress(test_raw, type = t), silent = TRUE), "try-error")
  }, logical(1)), types)
}

test_that("util_error works", {
  skip_on_cran()
  expect_error(util_error("The one and everything is %d (%s).", 42,
                          "Douglas Adams"),
               regexp = "The one and everything is 42 (Douglas Adams).",
               fixed = TRUE)
  x <- function(m) {
    stop(m)
  }
  y <- function(m) {
    x(m)
  }
  s <- function(e) {
    util_error(e)
    invokeRestart()
  }
  expect_error(
    withCallingHandlers(x(""), error = s),
    regexp = "Error",
    fixed = TRUE
  )
  expect_error(
    withCallingHandlers(x("Bye bye"), error = s),
    regexp = "Bye bye",
    fixed = TRUE
  )
  expect_error(
    do.call(y, list(m = "Hello")),
    regexp = "Hello",
    fixed = TRUE
  )
  expect_error(
    do.call(s, list(e = "Hello")),
    regexp = "Hello",
    fixed = TRUE
  )
})

---
title: "dataquieR"
output: "rmarkdown::html_vignette"
vignette: >
  %\VignetteIndexEntry{dataquieR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


[Please refer to our website: https://dataquality.qihs.uni-greifswald.de/](https://dataquality.qihs.uni-greifswald.de/).

# test-int_sts_element_dataframe works

    list(DataframeTable = structure(list(DF_NAME = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS", 
        NUM_int_sts_element = 1L, PCT_int_sts_element = 3.03030303030303), class = "data.frame", row.names = c(NA, 
    -1L)), DataframeData = structure(list(DF_NAME = "https://dataquality.qihs.uni-greifswald.de/extdata/fortests/ship.RDS", 
        "Unexpected data element set (Number)" = 1L, "Unexpected data element set (Percentage (0 to 100))" = 3.03, 
        "Affected Elements" = "{ \"sex\" } ∉ md"), row.names = c(NA, 
    -1L), class = "data.frame"))

# prep_combine_report_summaries works

    Code
      comb_sum
    Output
      $Data
         STUDY_SEGMENT VAR_NAMES acc_varcomp_observer.ICC_acc_ud_loc
      1         v10000    v00000                                  NA
      2         v10000    v00001                                  NA
      3         v10000    v00002                                  NA
      4         v10000    v00003                                  NA
      5         v20000    v00004                               0.100
      6         v20000    v00012                                  NA
      11        v20000    v00005                               0.112
         com_item_missingness.PCT_com_crm_mv
      1                                0.00%
      2                                0.00%
      3                                2.00%
      4                                2.00%
      5                               14.63%
      6                                7.50%
      11                              15.20%
      
      $Table
         STUDY_SEGMENT VAR_NAMES acc_varcomp_observer.ICC_acc_ud_loc
      1         v10000    v00000                                  NA
      2         v10000    v00001                                  NA
      3         v10000    v00002                                  NA
      4         v10000    v00003                                  NA
      5         v20000    v00004                               0.100
      6         v20000    v00012                                  NA
      11        v20000    v00005                               0.112
         com_item_missingness.PCT_com_crm_mv
      1                                 0.00
      2                                 0.00
      3                                 2.00
      4                                 2.00
      5                                14.63
      6                                 7.50
      11                               15.20
      
      $meta_data
        VAR_NAMES     LABEL DATA_TYPE
      1    v00000  CENTER_0   integer
      2    v00001 PSEUDO_ID    string
      3    v00002     SEX_0   integer
      4    v00003     AGE_0   integer
      5    v00004     SBP_0     float
      6    v00005     DBP_0     float
      7    v00012  USR_BP_0    string
                                                                                                                                                                                                                                                                                                                                                                                         MISSING_LIST
      1                                                                                                                                                                                                                                                                                                                                                                                             |
      2                                                                                                                                                                                                                                                                                                                                                                                             |
      3                                                                                                                                                                                                                                                                                                                                                                                             |
      4                                                                                                                                                                                                                                                                                                                                                                                             |
      5 99980 = MISSING 99980 | 99981 = MISSING 99981 | 99982 = MISSING 99982 | 99983 = MISSING 99983 | 99984 = MISSING 99984 | 99985 = MISSING 99985 | 99986 = MISSING 99986 | 99987 = MISSING 99987 | 99988 = MISSING 99988 | 99989 = MISSING 99989 | 99990 = MISSING 99990 | 99991 = MISSING 99991 | 99992 = MISSING 99992 | 99993 = MISSING 99993 | 99994 = MISSING 99994 | 99995 = MISSING 99995
      6 99980 = MISSING 99980 | 99981 = MISSING 99981 | 99982 = MISSING 99982 | 99983 = MISSING 99983 | 99984 = MISSING 99984 | 99985 = MISSING 99985 | 99986 = MISSING 99986 | 99987 = MISSING 99987 | 99988 = MISSING 99988 | 99989 = MISSING 99989 | 99990 = MISSING 99990 | 99991 = MISSING 99991 | 99992 = MISSING 99992 | 99993 = MISSING 99993 | 99994 = MISSING 99994 | 99995 = MISSING 99995
      7                                                                                                                                                                                                                                                                                                                                                 99981 = MISSING 99981 | 99982 = MISSING 99982
        JUMP_LIST HARD_LIMITS DETECTION_LIMITS CONTRADICTIONS SOFT_LIMITS
      1         |        <NA>             <NA>           <NA>        <NA>
      2         |        <NA>             <NA>           <NA>        <NA>
      3         |        <NA>             <NA>           1002        <NA>
      4         |    [18;Inf)             <NA>           1001        <NA>
      5         |    [80;180]          [0;265]           <NA>    (90;170)
      6         |    [50;Inf)          [0;265]           <NA>    (55;100)
      7         |        <NA>             <NA>           <NA>        <NA>
        DISTRIBUTION DECIMALS GROUP_VAR_OBSERVER GROUP_VAR_DEVICE VARIABLE_ROLE
      1         <NA>       NA               <NA>             <NA>         intro
      2         <NA>       NA               <NA>             <NA>         intro
      3         <NA>       NA               <NA>             <NA>         intro
      4         <NA>       NA               <NA>             <NA>         intro
      5       normal        0             v00012             <NA>       primary
      6       normal        0             v00012             <NA>       primary
      7         <NA>       NA               <NA>             <NA>       process
        VARIABLE_ORDER                 LONG_LABEL LOCATION_RANGE LOCATION_METRIC
      1              1                   CENTER_0           <NA>            <NA>
      2              2                  PSEUDO_ID           <NA>            <NA>
      3              3                      SEX_0           <NA>            <NA>
      4              4                      AGE_0        [45;55]            Mean
      5              9  SYSTOLIC_BLOOD_PRESSURE_0      (100;140)            Mean
      6             10 DIASTOLIC_BLOOD_PRESSURE_0       (60;100)            Mean
      7             18              EXAMINER_BP_0           <NA>            <NA>
        PROPORTION_RANGE END_DIGIT_CHECK STUDY_SEGMENT TIME_VAR
      1          [15;30]           FALSE        v10000     <NA>
      2             <NA>           FALSE        v10000     <NA>
      3     0 in [48;52]           FALSE        v10000     <NA>
      4             <NA>           FALSE        v10000     <NA>
      5             <NA>           FALSE        v20000   v00013
      6             <NA>           FALSE        v20000   v00013
      7             <NA>           FALSE        v20000     <NA>
                     VALUE_LABEL_TABLE N_RULES UNIVARIATE_OUTLIER_CHECKTYPE
      1  LABS_Berlin Hamburg...::97896      NA                         <NA>
      2                           <NA>      NA                         <NA>
      3      LABS_females males::8a1bf      NA                         <NA>
      4                           <NA>      NA                         <NA>
      5                           <NA>      NA                         <NA>
      6                           <NA>      NA                         <NA>
      7 LABS_USR_121 USR_123...::c993d      NA                         <NA>
          MISSING_LIST_TABLE SCALE_LEVEL
      1 MISSING_LIST_TABLE_0     nominal
      2 MISSING_LIST_TABLE_0          na
      3 MISSING_LIST_TABLE_0     nominal
      4 MISSING_LIST_TABLE_0       ratio
      5 MISSING_LIST_TABLE_0       ratio
      6 MISSING_LIST_TABLE_0       ratio
      7 MISSING_LIST_TABLE_0     nominal
      
      attr(,"class")
      [1] "dq_report2_summary"

# prep_compare_meta_with_study works

    list(pred = structure(list(VAR_NAMES = c("v00000", "v00001", 
    "v00002", "v00003", "v00004", "v00005", "v00006", "v00007", "v00008", 
    "v00009", "v00010", "v00011", "v00012", "v00013", "v00014", "v00015", 
    "v00016", "v00017", "v00018", "v00019", "v00020", "v00021", "v00022", 
    "v00023", "v00024", "v00025", "v00026", "v00027", "v00028", "v00029", 
    "v00030", "v00031", "v00032", "v00033", "v00034", "v00035", "v00036", 
    "v00037", "v00038", "v00039", "v00040", "v00041", "v00042", "v00103", 
    "v00109", "v01002", "v01003", "v01018", "v10000", "v20000", "v30000", 
    "v40000", "v50000"), LABEL = c("v00000", "v00001", "v00002", 
    "v00003", "v00004", "v00005", "v00006", "v00007", "v00008", "v00009", 
    "v00010", "v00011", "v00012", "v00013", "v00014", "v00015", "v00016", 
    "v00017", "v00018", "v00019", "v00020", "v00021", "v00022", "v00023", 
    "v00024", "v00025", "v00026", "v00027", "v00028", "v00029", "v00030", 
    "v00031", "v00032", "v00033", "v00034", "v00035", "v00036", "v00037", 
    "v00038", "v00039", "v00040", "v00041", "v00042", "v00103", "v00109", 
    "v01002", "v01003", "v01018", "v10000", "v20000", "v30000", "v40000", 
    "v50000"), DATA_TYPE = c("integer", "string", "integer", "integer", 
    "integer", "integer", "float", "integer", "string", "integer", 
    "integer", "string", "string", "datetime", "float", "integer", 
    "integer", "datetime", "integer", "integer", "integer", "integer", 
    "integer", "integer", "integer", "integer", "integer", "integer", 
    "integer", "integer", "integer", "integer", "string", "datetime", 
    "integer", "integer", "integer", "integer", "integer", "integer", 
    "integer", "integer", "datetime", "string", "integer", "integer", 
    "integer", "integer", "integer", "integer", "integer", "integer", 
    "integer"), SCALE_LEVEL = c("nominal", "na", "nominal", "ratio", 
    "ratio", "ratio", "ratio", "nominal", "nominal", "ratio", "nominal", 
    "nominal", "nominal", "interval", "ratio", "ratio", "nominal", 
    "interval", "nominal", "nominal", "nominal", "ordinal", "nominal", 
    "nominal", "nominal", "nominal", "ordinal", "ordinal", "nominal", 
    "nominal", "nominal", "ordinal", "nominal", "interval", "ordinal", 
    "ordinal", "ordinal", "ordinal", "ordinal", "ordinal", "ordinal", 
    "ordinal", "interval", "nominal", "nominal", "nominal", "ratio", 
    "nominal", "nominal", "nominal", "nominal", "nominal", "nominal"
    ), UNIT = c(NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_, NA_character_, 
    NA_character_, NA_character_, NA_character_, NA_character_), 
        VALUE_LABELS = c(NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_), MISSING_LIST_TABLE = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), HARD_LIMITS = c(NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_), CO_VARS = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), GROUP_VAR_OBSERVER = c(NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_), GROUP_VAR_DEVICE = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), TIME_VAR = c(NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_), PART_VAR = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), STUDY_SEGMENT = c(NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_), LONG_LABEL = c("v00000", "v00001", 
        "v00002", "v00003", "v00004", "v00005", "v00006", "v00007", 
        "v00008", "v00009", "v00010", "v00011", "v00012", "v00013", 
        "v00014", "v00015", "v00016", "v00017", "v00018", "v00019", 
        "v00020", "v00021", "v00022", "v00023", "v00024", "v00025", 
        "v00026", "v00027", "v00028", "v00029", "v00030", "v00031", 
        "v00032", "v00033", "v00034", "v00035", "v00036", "v00037", 
        "v00038", "v00039", "v00040", "v00041", "v00042", "v00103", 
        "v00109", "v01002", "v01003", "v01018", "v10000", "v20000", 
        "v30000", "v40000", "v50000"), LOCATION_RANGE = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), LOCATION_METRIC = c(NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_), PROPORTION_RANGE = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), UNIVARIATE_OUTLIER_CHECKTYPE = c(NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_), N_RULES = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), MISSING_LIST = c("|", "|", "|", "|", "99980|99981|99982|99983|99984|99985|99986|99987|99988|99989|99990|99991|99992|99993|99994|99995", 
        "99980|99981|99982|99983|99984|99985|99986|99987|99988|99989|99990|99991|99992|99993|99994|99995", 
        "99980|99983|99987|99988|99989|99990|99991|99992|99993|99994|99995", 
        "99980|99988|99989|99991|99993|99994|99995", "|", "99980|99981|99982|99983|99984|99985|99986|99988|99989|99990|99991|99992|99993|99994|99995", 
        "99980|99987", "|", "|", "|", "99980|99981|99982|99983|99984|99985|99986|99988|99989|99990|99991|99992|99994|99995", 
        "99980|99981|99982|99983|99984|99985|99986|99988|99989|99990|99991|99992|99994|99995", 
        "|", "|", "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "88880|99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "88880|99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "|", "|", "99980|99983|99988|99989|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99991|99993|99994|99995", "99980|99983|99988|99989|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99991|99993|99994|99995", "99980|99983|99988|99989|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99991|99993|99994|99995", "99980|99983|99988|99989|99991|99993|99994|99995", 
        "99980|99983|99988|99989|99991|99993|99994|99995", "|", "|", 
        "99980|99981|99982|99983|99984|99985|99986|99987|99988|99989|99990|99991|99992|99993|99994|99995", 
        "|", "|", "99980|99983|99988|99989|99990|99991|99993|99994|99995", 
        "|", "|", "|", "|", "|"), JUMP_LIST = c(NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_)), row.names = c("v00000", 
    "v00001", "v00002", "v00003", "v00004", "v00005", "v00006", "v00007", 
    "v00008", "v00009", "v00010", "v00011", "v00012", "v00013", "v00014", 
    "v00015", "v00016", "v00017", "v00018", "v00019", "v00020", "v00021", 
    "v00022", "v00023", "v00024", "v00025", "v00026", "v00027", "v00028", 
    "v00029", "v00030", "v00031", "v00032", "v00033", "v00034", "v00035", 
    "v00036", "v00037", "v00038", "v00039", "v00040", "v00041", "v00042", 
    "v00103", "v00109", "v01002", "v01003", "v01018", "v10000", "v20000", 
    "v30000", "v40000", "v50000"), class = "data.frame"), prov = structure(list(
        VAR_NAMES = c("v00000", "v00001", "v00002", "v00003", "v00004", 
        "v00005", "v00006", "v00007", "v00008", "v00009", "v00010", 
        "v00011", "v00012", "v00013", "v00014", "v00015", "v00016", 
        "v00017", "v00018", "v00019", "v00020", "v00021", "v00022", 
        "v00023", "v00024", "v00025", "v00026", "v00027", "v00028", 
        "v00029", "v00030", "v00031", "v00032", "v00033", "v00034", 
        "v00035", "v00036", "v00037", "v00038", "v00039", "v00040", 
        "v00041", "v00042", "v00103", "v00109", "v01002", "v01003", 
        "v01018", "v10000", "v20000", "v30000", "v40000", "v50000"
        ), LABEL = c("CENTER_0", "PSEUDO_ID", "SEX_0", "AGE_0", "SBP_0", 
        "DBP_0", "GLOBAL_HEALTH_VAS_0", "ASTHMA_0", "VO2_CAPCAT_0", 
        "ARM_CIRC_0", "ARM_CUFF_0", "USR_VO2_0", "USR_BP_0", "EXAM_DT_0", 
        "CRP_0", "BSG_0", "DEV_NO_0", "LAB_DT_0", "EDUCATION_0", 
        "FAM_STAT_0", "MARRIED_0", "N_CHILD_0", "EATING_PREFS_0", 
        "MEAT_CONS_0", "SMOKING_0", "SMOKE_SHOP_0", "N_INJURIES_0", 
        "N_BIRTH_0", "INCOME_GROUP_0", "PREGNANT_0", "MEDICATION_0", 
        "N_ATC_CODES_0", "USR_SOCDEM_0", "INT_DT_0", "ITEM_1_0", 
        "ITEM_2_0", "ITEM_3_0", "ITEM_4_0", "ITEM_5_0", "ITEM_6_0", 
        "ITEM_7_0", "ITEM_8_0", "QUEST_DT_0", "AGE_GROUP_0", "ARM_CIRC_DISC_0", 
        "SEX_1", "AGE_1", "EDUCATION_1", "PART_STUDY", "PART_PHYS_EXAM", 
        "PART_LAB", "PART_INTERVIEW", "PART_QUESTIONNAIRE"), DATA_TYPE = c("integer", 
        "string", "integer", "integer", "float", "float", "float", 
        "integer", "string", "float", "integer", "string", "string", 
        "datetime", "float", "float", "integer", "datetime", "integer", 
        "integer", "integer", "integer", "integer", "integer", "integer", 
        "integer", "integer", "integer", "integer", "integer", "integer", 
        "integer", "string", "datetime", "integer", "integer", "integer", 
        "integer", "integer", "integer", "integer", "integer", "datetime", 
        "string", "integer", "integer", "integer", "integer", "integer", 
        "integer", "integer", "integer", "integer"), SCALE_LEVEL = c("nominal", 
        "na", "nominal", "ratio", "ratio", "ratio", "ratio", "nominal", 
        "ordinal", "ratio", "ordinal", "nominal", "nominal", "interval", 
        "ratio", "ratio", "nominal", "interval", "ordinal", "nominal", 
        "nominal", "ratio", "nominal", "ordinal", "nominal", "ordinal", 
        "ratio", "ratio", "ordinal", "nominal", "nominal", "ratio", 
        "nominal", "interval", "ratio", "ratio", "ratio", "ratio", 
        "ratio", "ratio", "ratio", "ratio", "interval", "ordinal", 
        "ordinal", "nominal", "ratio", "ordinal", "nominal", "nominal", 
        "nominal", "nominal", "nominal"), STANDARDIZED_VOCABULARY_TABLE = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), MISSING_LIST_TABLE = c(NA, NA, NA, NA, "missing_table", 
        "missing_table", "missing_table", "missing_table", "missing_table", 
        "missing_table", "missing_table", "missing_table", "missing_table", 
        NA, "missing_table", "missing_table", NA, NA, "missing_table", 
        "missing_table", "missing_table", "missing_table", "missing_table", 
        "missing_table", "missing_table", "missing_table", "missing_table", 
        "missing_table", "missing_table", "missing_table", "missing_table", 
        "missing_table", "missing_table", NA, "missing_table", "missing_table", 
        "missing_table", "missing_table", "missing_table", "missing_table", 
        "missing_table", "missing_table", NA, NA, "missing_table", 
        NA, NA, "missing_table", NA, NA, NA, NA, NA), HARD_LIMITS = c(NA, 
        NA, NA, "[18;Inf)", "[80;180]", "[50;Inf)", "[0;10]", "[0;1]", 
        NA, "[0;Inf)", "[1;3]", NA, NA, "[2018-01-01 00:00:00 CET;)", 
        "[0;Inf)", "[0;100]", NA, "[2018-01-01 00:00:00 CET;)", "[0;6]", 
        NA, "[0;1]", NA, "[0;2]", "[0;4]", "[0;1]", "[0;4]", NA, 
        NA, NA, "[0;1]", "[0;1]", "[0;Inf]", NA, "[2018-01-01 00:00:00 CET;)", 
        "[0;10]", "[0;10]", "[0;10]", "[0;10]", "[0;10]", "[0;10]", 
        "[0;10]", "[0;10]", "[2018-01-01 00:00:00 CET;)", NA, "[1;3]", 
        NA, "[18;Inf)", "[0;6]", NA, NA, NA, NA, NA), DETECTION_LIMITS = c(NA, 
        NA, NA, NA, "[0;265]", "[0;265]", NA, NA, NA, NA, NA, NA, 
        NA, NA, "[0.16;Inf)", NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA), 
        SOFT_LIMITS = c(NA, NA, NA, NA, "(90;170)", "(55;100)", "[1;9]", 
        NA, NA, "(0;60]", NA, NA, NA, NA, "[0.2;10)", "[0.2;30)", 
        NA, NA, NA, NA, NA, "[0;5]", NA, NA, NA, NA, "[0;10]", "[0;10]", 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA), DISTRIBUTION = c(NA, 
        NA, NA, NA, "normal", "normal", "uniform", NA, NA, "normal", 
        NA, NA, NA, NA, "gamma", "normal", "uniform", NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA), DECIMALS = c(NA, NA, NA, NA, "0", "0", "1", 
        NA, NA, "0", NA, NA, NA, NA, "3", "0", NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA), GROUP_VAR_OBSERVER = c(NA, NA, NA, NA, "USR_BP_0", 
        "USR_BP_0", NA, NA, "USR_VO2_0", NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, "USR_SOCDEM_0", "USR_SOCDEM_0", "USR_SOCDEM_0", 
        "USR_SOCDEM_0", "USR_SOCDEM_0", "USR_SOCDEM_0", "USR_SOCDEM_0", 
        "USR_SOCDEM_0", "USR_SOCDEM_0", "USR_SOCDEM_0", "USR_SOCDEM_0", 
        "USR_SOCDEM_0", "USR_SOCDEM_0", "USR_SOCDEM_0", NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, "USR_SOCDEM_0", 
        NA, NA, NA, NA, NA), GROUP_VAR_DEVICE = c(NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, "ARM_CUFF_0", NA, NA, NA, NA, "DEV_NO_0", 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, "ARM_CUFF_0", 
        NA, NA, NA, NA, NA, NA, NA, NA), TIME_VAR = c(NA, NA, NA, 
        NA, "EXAM_DT_0", "EXAM_DT_0", NA, NA, "EXAM_DT_0", "EXAM_DT_0", 
        NA, NA, NA, NA, "LAB_DT_0", "LAB_DT_0", NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA), STUDY_SEGMENT = c("STUDY", "STUDY", "STUDY", "STUDY", 
        "PHYS_EXAM", "PHYS_EXAM", "PHYS_EXAM", "PHYS_EXAM", "PHYS_EXAM", 
        "PHYS_EXAM", "PHYS_EXAM", "PHYS_EXAM", "PHYS_EXAM", "PHYS_EXAM", 
        "LAB", "LAB", "LAB", "LAB", "INTERVIEW", "INTERVIEW", "INTERVIEW", 
        "INTERVIEW", "INTERVIEW", "INTERVIEW", "INTERVIEW", "INTERVIEW", 
        "INTERVIEW", "INTERVIEW", "INTERVIEW", "INTERVIEW", "INTERVIEW", 
        "INTERVIEW", "INTERVIEW", "INTERVIEW", "QUESTIONNAIRE", "QUESTIONNAIRE", 
        "QUESTIONNAIRE", "QUESTIONNAIRE", "QUESTIONNAIRE", "QUESTIONNAIRE", 
        "QUESTIONNAIRE", "QUESTIONNAIRE", "QUESTIONNAIRE", "STUDY", 
        "PHYS_EXAM", "STUDY", "STUDY", "INTERVIEW", "STUDY", "PHYS_EXAM", 
        "LAB", "INTERVIEW", "QUESTIONNAIRE"), PART_VAR = c("v10000", 
        "v10000", "v10000", "v10000", "v20000", "v20000", "v20000", 
        "v20000", "v20000", "v20000", "v20000", "v20000", "v20000", 
        "v20000", "v30000", "v30000", "v30000", "v30000", "v40000", 
        "v40000", "v40000", "v40000", "v40000", "v40000", "v40000", 
        "v40000", "v40000", "v40000", "v40000", "v40000", "v40000", 
        "v40000", "v40000", "v40000", "v50000", "v50000", "v50000", 
        "v50000", "v50000", "v50000", "v50000", "v50000", "v50000", 
        "v10000", "v20000", "v10000", "v10000", "v40000", "v10000", 
        "v10000", "v10000", "v10000", "v40000"), VARIABLE_ROLE = c("intro", 
        "intro", "intro", "intro", "primary", "primary", "primary", 
        "primary", "primary", "secondary", "process", "process", 
        "process", "process", "primary", "primary", "process", "process", 
        "primary", "primary", "primary", "primary", "primary", "primary", 
        "primary", "primary", "primary", "primary", "primary", "primary", 
        "primary", "primary", "process", "process", "primary", "primary", 
        "primary", "primary", "primary", "primary", "primary", "primary", 
        "process", "secondary", "secondary", "intro", "intro", "primary", 
        "intro", "intro", "intro", "intro", "intro"), VARIABLE_ORDER = c("1", 
        "2", "3", "4", "9", "10", "11", "12", "13", "14", "16", "17", 
        "18", "19", "21", "22", "23", "24", "26", "28", "29", "30", 
        "31", "32", "33", "34", "35", "36", "37", "38", "39", "40", 
        "41", "42", "44", "45", "46", "47", "48", "49", "50", "51", 
        "52", "5", "15", "7", "6", "27", "8", "20", "25", "43", "53"
        ), LONG_LABEL = c("Examination center", "Participant ID", 
        "Sex B/L", "Age B/L", "Systolic blood pressure", "Diastolic blood pressure", 
        "Self-reported global health", "Known asthma", "Aerobic capacity category", 
        "Upper arm circumference", "Upper arm circumference device", 
        "Aerobic capacity examiner", "Blood pressure examiner", "Examination date and time", 
        "C-reactive protein", "Erythrocyte sedimentation rate", "Device ID", 
        "Lab analysis date and time", "Highest educational level B/L", 
        "Marital status", "Currently married", "Number of children", 
        "Eating preferences", "Meat consumption", "Current smoker", 
        "Purchasing tobacco products", "Number of injuries", "Number of births", 
        "Income group", "Currently pregnant", "Medication use", "Number of ATC codes", 
        "Sociodemographics examiner", "Interview date and time", 
        "Item 1", "Item 2", "Item 3", "Item 4", "Item 5", "Item 6", 
        "Item 7", "Item 8", "Questionnaire date and time", "Age group B/L", 
        "Upper arm circumference cat", "Sex F/U", "Age F/U", "Highest educational level F/U", 
        "Study consent", "Physical exam consent", "Lab analysis consent", 
        "Interview consent", "Questionnaire consent"), ELEMENT_HOMOGENITY_CHECKTYPE = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), UNIVARIATE_OUTLIER_CHECKTYPE = c(NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, "Sigma_gap | Tukey", NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA), N_RULES = c("4", "4", "4", "4", 
        "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", 
        "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", 
        "2", "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", 
        "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", "4", 
        "4"), LOCATION_METRIC = c(NA, NA, NA, NA, "Mean", "Mean", 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA), LOCATION_RANGE = c(NA, NA, NA, NA, "(100;140)", 
        "(60;100)", NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA), PROPORTION_RANGE = c(NA, NA, "(48;52)", 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, "(48;52)", 
        NA, NA, NA, NA, NA, NA, NA), REPEATED_MEASURES_VARS = c(NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_
        ), REPEATED_MEASURES_GOLDSTANDARD = c(NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_, NA_character_, 
        NA_character_, NA_character_, NA_character_), CO_VARS = c(NA, 
        NA, NA, NA, "AGE_0 | SEX_0", "AGE_0 | SEX_0", NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA), 
        END_DIGIT_CHECK = c(FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, 
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, 
        TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, 
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, 
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, 
        FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, 
        FALSE, FALSE), VALUE_LABEL_TABLE = c("LABS_Berlin Hamburg...::97896", 
        NA, "LABS_females males::8a1bf", NA, NA, NA, NA, "LABS_no yes::cde3d", 
        "LABS_excellen good...::dfd3a", NA, "LABS_(-Inf,20 (20,30]...::8d957", 
        "LABS_USR_101 USR_103...::f7ae9", "LABS_USR_121 USR_123...::c993d", 
        NA, NA, NA, NA, NA, "LABS_pre-prim primary...::66ffb", "LABS_single married...::5eb6b", 
        "LABS_no yes::cde3d", NA, "LABS_none vegetari...::ba106", 
        "LABS_never 1-2d a w...::9f0f5", "LABS_no yes::cde3d", "LABS_never 1-2d a w...::9f0f5", 
        NA, NA, "LABS_below 10 [10-30k)...::7c9d1", "LABS_no yes::cde3d", 
        "LABS_no yes::cde3d", NA, "LABS_SR_120 USR_125...::be858", 
        NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, "LABS_(-Inf,20 (20,30]...::8d957", 
        "LABS_females males::8a1bf", NA, "LABS_pre-prim primary...::66ffb", 
        "LABS_no yes::cde3d", "LABS_no yes::cde3d", "LABS_no yes::cde3d", 
        "LABS_no yes::cde3d", "LABS_no yes::cde3d"), MISSING_LIST = c("|", 
        "|", "|", "|", "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "|", "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "|", "|", "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "|", "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "|", "|", "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "|", "|", "99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing", 
        "|", "|", "|", "|", "|"), JUMP_LIST = c("|", "|", "|", "|", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "|", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "|", "|", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "|", "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "88880 = JUMP 88880 | 88890 = JUMP 88890", "88880 = JUMP 88880 | 88890 = JUMP 88890", 
        "|", "|", "88880 = JUMP 88880 | 88890 = JUMP 88890", "|", 
        "|", "88880 = JUMP 88880 | 88890 = JUMP 88890", "|", "|", 
        "|", "|", "|")), version = 2, normalized = TRUE, row.names = c("v00000", 
    "v00001", "v00002", "v00003", "v00004", "v00005", "v00006", "v00007", 
    "v00008", "v00009", "v00010", "v00011", "v00012", "v00013", "v00014", 
    "v00015", "v00016", "v00017", "v00018", "v00019", "v00020", "v00021", 
    "v00022", "v00023", "v00024", "v00025", "v00026", "v00027", "v00028", 
    "v00029", "v00030", "v00031", "v00032", "v00033", "v00034", "v00035", 
    "v00036", "v00037", "v00038", "v00039", "v00040", "v00041", "v00042", 
    "v00103", "v00109", "v01002", "v01003", "v01018", "v10000", "v20000", 
    "v30000", "v40000", "v50000"), class = "data.frame"), ml_error = c("v00004", 
    "v00005", "v00006", "v00007", "v00008", "v00009", "v00010", "v00011", 
    "v00012", "v00014", "v00015", "v00018", "v00019", "v00020", "v00021", 
    "v00022", "v00023", "v00024", "v00025", "v00026", "v00027", "v00028", 
    "v00029", "v00030", "v00031", "v00032", "v00034", "v00035", "v00036", 
    "v00037", "v00038", "v00039", "v00040", "v00041", "v00109", "v01018"
    ), sl_error = c("v00008", "v00010", "v00018", "v00021", "v00023", 
    "v00025", "v00026", "v00027", "v00028", "v00031", "v00034", "v00035", 
    "v00036", "v00037", "v00038", "v00039", "v00040", "v00041", "v00103", 
    "v00109", "v01018"), dt_error = c("v00004", "v00005", "v00009", 
    "v00015"))

---

    c("In the item-level 'meta_data' \"v00004\", \"v00005\", \"v00009\", \"v00015\" may have assigned the wrong 'DATA_TYPE' given 'study_data'", 
    "In the item-level 'meta_data' \"v00008\", \"v00010\", \"v00018\", \"v00021\", \"v00023\", \"v00025\", \"v00026\", \"v00027\", \"v00028\", \"v00031\", \"v00034\", \"v00035\", \"v00036\", \"v00037\", \"v00038\", \"v00039\", \"v00040\", \"v00041\", \"v00103\", \"v00109\", \"v01018\" may have assigned the wrong 'SCALE_LEVEL' given 'study_data'", 
    "In the item-level 'meta_data' \"v00004\", \"v00005\", \"v00006\", \"v00007\", \"v00008\", \"v00009\", \"v00010\", \"v00011\", \"v00012\", \"v00014\", \"v00015\", \"v00018\", \"v00019\", \"v00020\", \"v00021\", \"v00022\", \"v00023\", \"v00024\", \"v00025\", \"v00026\", \"v00027\", \"v00028\", \"v00029\", \"v00030\", \"v00031\", \"v00032\", \"v00034\", \"v00035\", \"v00036\", \"v00037\", \"v00038\", \"v00039\", \"v00040\", \"v00041\", \"v00109\", \"v01018\" may have assigned the wrong 'MISSING_LIST' given 'study_data'"
    )

# complex_data_types

    Code
      attr(r, "meta_data")[attr(r, "meta_data")[[LABEL]] == "N_CHILD_0", , FALSE]
    Output
         VAR_NAMES     LABEL DATA_TYPE SCALE_LEVEL STANDARDIZED_VOCABULARY_TABLE
      30    v00021 N_CHILD_0   integer       ratio                          <NA>
         MISSING_LIST_TABLE HARD_LIMITS DETECTION_LIMITS SOFT_LIMITS DISTRIBUTION
      30      missing_table    [0; Inf)             <NA>       [0;5]         <NA>
         DECIMALS GROUP_VAR_OBSERVER GROUP_VAR_DEVICE TIME_VAR STUDY_SEGMENT PART_VAR
      30     <NA>       USR_SOCDEM_0             <NA>     <NA>     INTERVIEW   v40000
         VARIABLE_ROLE VARIABLE_ORDER         LONG_LABEL ELEMENT_HOMOGENITY_CHECKTYPE
      30       primary             30 Number of children                         <NA>
         UNIVARIATE_OUTLIER_CHECKTYPE N_RULES LOCATION_METRIC LOCATION_RANGE
      30                         <NA>       4            <NA>           <NA>
         PROPORTION_RANGE REPEATED_MEASURES_VARS REPEATED_MEASURES_GOLDSTANDARD
      30             <NA>                   <NA>                           <NA>
         CO_VARS END_DIGIT_CHECK EXTENDED_DATA_TYPE VALUE_LABEL_TABLE
      30    <NA>           FALSE              count              <NA>
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            MISSING_LIST
      30 99980 = Missing - other reason | 99981 = Missing - exclusion criteria | 99982 = Missing - refusal | 99983 = Missing - not assessable | 99984 = Missing - technical problem | 99985 = Missing - not available (material) | 99986 = Missing - not usable (material) | 99987 = Missing - reason unknown | 99988 = Missing - optional value | 99989 = Deleted - other reason | 99990 = Deleted - contradiction | 99991 = Deleted - value outside limits | 99992 = Inaccurate - other reason | 99993 = Inaccurate - value above detection limit | 99994 = Inaccurate - value below detection limit | 99995 = Data management ongoing
                                       JUMP_LIST
      30 88880 = JUMP 88880 | 88890 = JUMP 88890

---

    Code
      repsum

# multiplication works

    Code
      md0
    Output
        VAR_NAMES LABEL DATA_TYPE SCALE_LEVEL UNIT MISSING_LIST_TABLE HARD_LIMITS
      1     speed speed   integer     ordinal <NA>               <NA>        <NA>
      2      dist  dist   integer       ratio <NA>               <NA>        <NA>
        CO_VARS GROUP_VAR_OBSERVER GROUP_VAR_DEVICE TIME_VAR PART_VAR STUDY_SEGMENT
      1    <NA>               <NA>             <NA>     <NA>     <NA>          <NA>
      2    <NA>               <NA>             <NA>     <NA>     <NA>          <NA>
        LONG_LABEL LOCATION_RANGE LOCATION_METRIC PROPORTION_RANGE
      1      speed           <NA>            <NA>             <NA>
      2       dist           <NA>            <NA>             <NA>
        UNIVARIATE_OUTLIER_CHECKTYPE N_RULES END_DIGIT_CHECK VALUE_LABEL_TABLE
      1                         <NA>    <NA>           FALSE              <NA>
      2                         <NA>    <NA>           FALSE              <NA>

# prep_unsplit_val_tabs works

    Code
      clt
    Output
          CODE_VALUE                               CODE_LABEL CODE_ORDER
      1            1                                (-Inf,20]          1
      2            2                                  (20,30]          2
      3            3                                (30, Inf]          3
      4            1                                   Berlin         NA
      5            2                                  Hamburg         NA
      6            3                                  Leipzig         NA
      7            4                                  Cologne         NA
      8            5                                   Munich         NA
      9       SR_120                                   SR_120         NA
      10     USR_125                                  USR_125         NA
      11     USR_130                                  USR_130         NA
      12     USR_201                                  USR_201         NA
      13     USR_247                                  USR_247         NA
      14     USR_277                                  USR_277         NA
      15     USR_321                                  USR_321         NA
      16     USR_333                                  USR_333         NA
      17     USR_357                                  USR_357         NA
      18     USR_492                                  USR_492         NA
      19     USR_493                                  USR_493         NA
      20     USR_494                                  USR_494         NA
      21     USR_500                                  USR_500         NA
      22     USR_510                                  USR_510         NA
      23     USR_520                                  USR_520         NA
      24     USR_101                                  USR_101         NA
      25     USR_103                                  USR_103         NA
      26     USR_155                                  USR_155         NA
      27     USR_211                                  USR_211         NA
      28     USR_213                                  USR_213         NA
      29     USR_215                                  USR_215         NA
      30     USR_321                                  USR_321         NA
      31     USR_333                                  USR_333         NA
      32     USR_342                                  USR_342         NA
      33     USR_402                                  USR_402         NA
      34     USR_403                                  USR_403         NA
      35     USR_404                                  USR_404         NA
      36     USR_590                                  USR_590         NA
      37     USR_592                                  USR_592         NA
      38     USR_599                                  USR_599         NA
      39     USR_121                                  USR_121         NA
      40     USR_123                                  USR_123         NA
      41     USR_165                                  USR_165         NA
      42     USR_201                                  USR_201         NA
      43     USR_243                                  USR_243         NA
      44     USR_275                                  USR_275         NA
      45     USR_301                                  USR_301         NA
      46     USR_303                                  USR_303         NA
      47     USR_352                                  USR_352         NA
      48     USR_482                                  USR_482         NA
      49     USR_483                                  USR_483         NA
      50     USR_484                                  USR_484         NA
      51     USR_537                                  USR_537         NA
      52     USR_542                                  USR_542         NA
      53     USR_559                                  USR_559         NA
      54           0                                below 10k          1
      55           1                                 [10-30k)          2
      56           2                                 [30-50k)          3
      57           3                                 [50-70k)          4
      58           4                                 [70-90k)          5
      59           5                                above 90k          6
      60           A                                excellent          1
      61           B                                     good          2
      62           C                                 moderate          3
      63           D                               restricted          4
      64           E                             pathological          5
      65           0                                  females         NA
      66           1                                    males         NA
      67           0                                    never          1
      68           1                              1-2d a week          2
      69           2                              3-4d a week          3
      70           3                              5-6d a week          4
      71           4                                    daily          5
      72           0                                       no         NA
      73           1                                      yes         NA
      74           0                                     none         NA
      75           1                               vegetarian         NA
      76           2                                    vegan         NA
      77           0                              pre-primary          1
      78           1                                  primary          2
      79           2                                secondary          3
      80           3                              uppersecond          4
      81           4                               postsecond          5
      82           5                                 tertiary          6
      83           6                           secondtertiary          7
      84      single                                   single         NA
      85     married                                  married         NA
      86    divorced                                 divorced         NA
      87     widowed                                  widowed         NA
      88       99980                   Missing - other reason         NA
      89       99981             Missing - exclusion criteria         NA
      90       99982                        Missing - refusal         NA
      91       99983                 Missing - not assessable         NA
      92       99984              Missing - technical problem         NA
      93       99985       Missing - not available (material)         NA
      94       99986          Missing - not usable (material)         NA
      95       99987                 Missing - reason unknown         NA
      96       99988                 Missing - optional value         NA
      97       99989                   Deleted - other reason         NA
      98       99990                  Deleted - contradiction         NA
      99       99991           Deleted - value outside limits         NA
      100      99992                Inaccurate - other reason         NA
      101      99993 Inaccurate - value above detection limit         NA
      102      99994 Inaccurate - value below detection limit         NA
      103      99995                  Data management ongoing         NA
      104      88880                               JUMP 88880         NA
      105      88890                               JUMP 88890         NA
                         VALUE_LABEL_TABLE CODE_CLASS CODE_INTERPRET
      1    LABS_(-Inf,20 (20,30]...::8d957      VALUE           <NA>
      2    LABS_(-Inf,20 (20,30]...::8d957      VALUE           <NA>
      3    LABS_(-Inf,20 (20,30]...::8d957      VALUE           <NA>
      4      LABS_Berlin Hamburg...::97896      VALUE           <NA>
      5      LABS_Berlin Hamburg...::97896      VALUE           <NA>
      6      LABS_Berlin Hamburg...::97896      VALUE           <NA>
      7      LABS_Berlin Hamburg...::97896      VALUE           <NA>
      8      LABS_Berlin Hamburg...::97896      VALUE           <NA>
      9      LABS_SR_120 USR_125...::be858      VALUE           <NA>
      10     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      11     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      12     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      13     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      14     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      15     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      16     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      17     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      18     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      19     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      20     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      21     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      22     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      23     LABS_SR_120 USR_125...::be858      VALUE           <NA>
      24    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      25    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      26    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      27    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      28    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      29    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      30    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      31    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      32    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      33    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      34    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      35    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      36    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      37    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      38    LABS_USR_101 USR_103...::f7ae9      VALUE           <NA>
      39    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      40    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      41    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      42    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      43    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      44    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      45    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      46    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      47    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      48    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      49    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      50    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      51    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      52    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      53    LABS_USR_121 USR_123...::c993d      VALUE           <NA>
      54  LABS_below 10 [10-30k)...::7c9d1      VALUE           <NA>
      55  LABS_below 10 [10-30k)...::7c9d1      VALUE           <NA>
      56  LABS_below 10 [10-30k)...::7c9d1      VALUE           <NA>
      57  LABS_below 10 [10-30k)...::7c9d1      VALUE           <NA>
      58  LABS_below 10 [10-30k)...::7c9d1      VALUE           <NA>
      59  LABS_below 10 [10-30k)...::7c9d1      VALUE           <NA>
      60      LABS_excellen good...::dfd3a      VALUE           <NA>
      61      LABS_excellen good...::dfd3a      VALUE           <NA>
      62      LABS_excellen good...::dfd3a      VALUE           <NA>
      63      LABS_excellen good...::dfd3a      VALUE           <NA>
      64      LABS_excellen good...::dfd3a      VALUE           <NA>
      65         LABS_females males::8a1bf      VALUE           <NA>
      66         LABS_females males::8a1bf      VALUE           <NA>
      67     LABS_never 1-2d a w...::9f0f5      VALUE           <NA>
      68     LABS_never 1-2d a w...::9f0f5      VALUE           <NA>
      69     LABS_never 1-2d a w...::9f0f5      VALUE           <NA>
      70     LABS_never 1-2d a w...::9f0f5      VALUE           <NA>
      71     LABS_never 1-2d a w...::9f0f5      VALUE           <NA>
      72                LABS_no yes::cde3d      VALUE           <NA>
      73                LABS_no yes::cde3d      VALUE           <NA>
      74      LABS_none vegetari...::ba106      VALUE           <NA>
      75      LABS_none vegetari...::ba106      VALUE           <NA>
      76      LABS_none vegetari...::ba106      VALUE           <NA>
      77   LABS_pre-prim primary...::66ffb      VALUE           <NA>
      78   LABS_pre-prim primary...::66ffb      VALUE           <NA>
      79   LABS_pre-prim primary...::66ffb      VALUE           <NA>
      80   LABS_pre-prim primary...::66ffb      VALUE           <NA>
      81   LABS_pre-prim primary...::66ffb      VALUE           <NA>
      82   LABS_pre-prim primary...::66ffb      VALUE           <NA>
      83   LABS_pre-prim primary...::66ffb      VALUE           <NA>
      84     LABS_single married...::5eb6b      VALUE           <NA>
      85     LABS_single married...::5eb6b      VALUE           <NA>
      86     LABS_single married...::5eb6b      VALUE           <NA>
      87     LABS_single married...::5eb6b      VALUE           <NA>
      88                              <NA>    MISSING              O
      89                              <NA>    MISSING             NE
      90                              <NA>    MISSING              R
      91                              <NA>    MISSING             NC
      92                              <NA>    MISSING              O
      93                              <NA>    MISSING             NC
      94                              <NA>    MISSING             NC
      95                              <NA>    MISSING              O
      96                              <NA>    MISSING             NE
      97                              <NA>    MISSING              O
      98                              <NA>    MISSING              O
      99                              <NA>    MISSING              O
      100                             <NA>    MISSING              O
      101                             <NA>    MISSING              O
      102                             <NA>    MISSING              O
      103                             <NA>    MISSING              O
      104                             <NA>       JUMP             NE
      105                             <NA>       JUMP             NE
          MISSING_LIST_TABLE
      1                 <NA>
      2                 <NA>
      3                 <NA>
      4                 <NA>
      5                 <NA>
      6                 <NA>
      7                 <NA>
      8                 <NA>
      9                 <NA>
      10                <NA>
      11                <NA>
      12                <NA>
      13                <NA>
      14                <NA>
      15                <NA>
      16                <NA>
      17                <NA>
      18                <NA>
      19                <NA>
      20                <NA>
      21                <NA>
      22                <NA>
      23                <NA>
      24                <NA>
      25                <NA>
      26                <NA>
      27                <NA>
      28                <NA>
      29                <NA>
      30                <NA>
      31                <NA>
      32                <NA>
      33                <NA>
      34                <NA>
      35                <NA>
      36                <NA>
      37                <NA>
      38                <NA>
      39                <NA>
      40                <NA>
      41                <NA>
      42                <NA>
      43                <NA>
      44                <NA>
      45                <NA>
      46                <NA>
      47                <NA>
      48                <NA>
      49                <NA>
      50                <NA>
      51                <NA>
      52                <NA>
      53                <NA>
      54                <NA>
      55                <NA>
      56                <NA>
      57                <NA>
      58                <NA>
      59                <NA>
      60                <NA>
      61                <NA>
      62                <NA>
      63                <NA>
      64                <NA>
      65                <NA>
      66                <NA>
      67                <NA>
      68                <NA>
      69                <NA>
      70                <NA>
      71                <NA>
      72                <NA>
      73                <NA>
      74                <NA>
      75                <NA>
      76                <NA>
      77                <NA>
      78                <NA>
      79                <NA>
      80                <NA>
      81                <NA>
      82                <NA>
      83                <NA>
      84                <NA>
      85                <NA>
      86                <NA>
      87                <NA>
      88       missing_table
      89       missing_table
      90       missing_table
      91       missing_table
      92       missing_table
      93       missing_table
      94       missing_table
      95       missing_table
      96       missing_table
      97       missing_table
      98       missing_table
      99       missing_table
      100      missing_table
      101      missing_table
      102      missing_table
      103      missing_table
      104      missing_table
      105      missing_table

---

    Code
      prep_list_dataframes()
    Output
       [1] "LABS_(-Inf,20 (20,30]...::8d957"  "LABS_Berlin Hamburg...::97896"   
       [3] "LABS_SR_120 USR_125...::be858"    "LABS_USR_101 USR_103...::f7ae9"  
       [5] "LABS_USR_121 USR_123...::c993d"   "LABS_below 10 [10-30k)...::7c9d1"
       [7] "LABS_excellen good...::dfd3a"     "LABS_females males::8a1bf"       
       [9] "LABS_never 1-2d a w...::9f0f5"    "LABS_no yes::cde3d"              
      [11] "LABS_none vegetari...::ba106"     "LABS_pre-prim primary...::66ffb" 
      [13] "LABS_single married...::5eb6b"    "missing_table"                   

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_save_report.R
\name{prep_save_report}
\alias{prep_save_report}
\title{Save a \code{dq_report2}}
\usage{
prep_save_report(report, file, compression_level = 3)
}
\arguments{
\item{report}{\link{dataquieR_resultset2} the report}

\item{file}{\link{character} the file name to write to}

\item{compression_level}{\link{integer}  from=0 to=9. Compression level.
9 is very slow.}
}
\value{
\code{invisible(NULL)}
}
\description{
Save a \code{dq_report2}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_html_for_var.R
\name{util_html_for_var}
\alias{util_html_for_var}
\title{Create a dynamic single variable page for the report}
\usage{
util_html_for_var(
  results,
  cur_var,
  use_plot_ly,
  template,
  note_meta = c(),
  rendered_repsum,
  dir,
  meta_data,
  label_col,
  dims_in_rep,
  clls_in_rep,
  function_alias_map
)
}
\arguments{
\item{results}{\link{list} a list of subsets of the report matching \code{cur_var}}

\item{cur_var}{\link{character} variable name for single variable pages}

\item{use_plot_ly}{\link{logical} use \code{plotly}, if available.}

\item{template}{\link{character} template to use for the \code{dq_report2} report.}

\item{note_meta}{\link{character} notes on the metadata for a single variable
(if needed)}

\item{rendered_repsum}{the \code{dataquieR} summary, see \code{\link[=summary]{summary()}},
\code{\link[=dq_report2]{dq_report2()}} and \code{\link[=print.dataquieR_summary]{print.dataquieR_summary()}}}

\item{dir}{\link{character} output directory for potential \code{iframes}.}
}
\value{
list of arguments for \code{append_single_page()} defined locally in
\code{util_generate_pages_from_report()}.
}
\description{
Create a dynamic single variable page for the report
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_multivariate_outlier.R
\name{acc_multivariate_outlier}
\alias{acc_multivariate_outlier}
\title{Calculate and plot Mahalanobis distances}
\usage{
acc_multivariate_outlier(
  variable_group = NULL,
  id_vars = NULL,
  label_col = VAR_NAMES,
  study_data,
  item_level = "item_level",
  n_rules = 4,
  max_non_outliers_plot = 10000,
  criteria = c("tukey", "3sd", "hubert", "sigmagap"),
  meta_data = item_level,
  meta_data_v2,
  scale = getOption("dataquieR.acc_multivariate_outlier.scale",
    dataquieR.acc_multivariate_outlier.scale_default),
  multivariate_outlier_check = TRUE
)
}
\arguments{
\item{variable_group}{\link{variable list} the names of the continuous
measurement variables building
a group, for that multivariate outliers
make sense.}

\item{id_vars}{\link{variable} optional, an ID variable of
the study data. If not specified row numbers are used.}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{n_rules}{\link{numeric} from=1 to=4. the no. of rules that must be violated
to classify as outlier}

\item{max_non_outliers_plot}{\link{integer} from=0. Maximum number of non-outlier
points to be plot. If more
points exist, a subsample will
be plotted only. Note, that
sampling is not deterministic.}

\item{criteria}{\link{set} tukey | 3SD | hubert | sigmagap. a vector with
methods to be used for detecting outliers.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{scale}{\link{logical} Should min-max-scaling be applied per variable?}

\item{multivariate_outlier_check}{\link{logical} really check, pipeline use,
only.}
}
\value{
a list with:
\itemize{
\item \code{SummaryTable}: \link{data.frame} underlying the plot
\item \code{SummaryPlot}: \link[ggplot2:ggplot2]{ggplot2::ggplot2} outlier plot
\item \code{FlaggedStudyData} \link{data.frame} contains the original data frame with
the additional columns \code{tukey},
\verb{3SD},
\code{hubert}, and \code{sigmagap}. Every
observation
is coded 0 if no outlier was detected in
the respective column and 1 if an
outlier was detected. This can be used
to exclude observations with outliers.
}
}
\description{
A standard tool to detect multivariate outliers is the Mahalanobis distance.
This approach is very helpful for the interpretation of the plausibility of a
measurement given the value of another.
In this approach the Mahalanobis distance is used as a univariate measure
itself. We apply the same rules for the identification of outliers as in
univariate outliers:
\itemize{
\item the classical approach from Tukey: \eqn{1.5 * IQR} from the
1st (\eqn{Q_{25}}) or 3rd (\eqn{Q_{75}}) quartile.
\item the 3SD approach, i.e. any measurement of the Mahalanobis
distance not in the interval of \eqn{\bar{x} \pm 3*\sigma} is considered an
outlier.
\item the approach from Hubert for skewed distributions which is embedded in the
R package \pkg{robustbase}
\item a completely heuristic approach named \eqn{\sigma}-gap.
}

For further details, please see the vignette for univariate outlier.

\link{Indicator}
}
\section{ALGORITHM OF THIS IMPLEMENTATION:}{
\itemize{
\item Implementation is restricted to variables of type float
\item Remove missing codes from the study data (if defined in the metadata)
\item The covariance matrix is estimated for all variables from \code{variable_group}
\item The Mahalanobis distance of each observation is calculated
\eqn{MD^2_i  = (x_i - \mu)^T \Sigma^{-1} (x_i -  \mu)}
\item The four rules mentioned above are applied on this distance for
each observation in the study data
\item An output data frame is generated that flags each outlier
\item A parallel coordinate plot indicates respective outliers
}

List function.
}

\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_multivariate_outlier.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_sts_element_dataframe.R
\name{int_sts_element_dataframe}
\alias{int_sts_element_dataframe}
\title{Determine missing and/or superfluous data elements}
\usage{
int_sts_element_dataframe(
  item_level = "item_level",
  meta_data_dataframe = "dataframe_level",
  meta_data = item_level,
  meta_data_v2,
  check_type = getOption("dataquieR.ELEMENT_MISSMATCH_CHECKTYPE",
    dataquieR.ELEMENT_MISSMATCH_CHECKTYPE_default),
  dataframe_level
)
}
\arguments{
\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data_dataframe}{\link{data.frame} the data frame that contains the
metadata for the data frame level}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{check_type}{\link{enum} none | exact | subset_u | subset_m. See
\link{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE}}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}
}
\value{
\link{list} with names lots:
\itemize{
\item \code{DataframeData}: data frame with the unexpected elements check results.
\item \code{DataframeTable}: \link{data.frame} table with all errors, used for the data quality report:
- \code{PCT_int_sts_element}: Percentage of element
mismatches
- \code{NUM_int_sts_element}: Number of element
mismatches
- \code{resp_vars}: affected element names
}
}
\description{
Depends on \link{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE} option,
see there
}
\details{
\link{Indicator}
}
\examples{
\dontrun{
prep_load_workbook_like_file("~/tmp/df_level_test.xlsx")
meta_data_dataframe <- "dataframe_level"
meta_data <- "item_level"
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_encoding_errors.R
\name{int_encoding_errors}
\alias{int_encoding_errors}
\title{Encoding Errors}
\usage{
int_encoding_errors(
  resp_vars = NULL,
  study_data,
  label_col,
  meta_data_dataframe = "dataframe_level",
  item_level = "item_level",
  ref_encs,
  meta_data = item_level,
  meta_data_v2,
  dataframe_level
)
}
\arguments{
\item{resp_vars}{\link{variable} the names of the measurement variables, if
missing or \code{NULL}, all variables will be checked}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{meta_data_dataframe}{\link{data.frame} the data frame that contains the
metadata for the data frame level}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{ref_encs}{reference encodings (names are \code{resp_vars})}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}
}
\value{
a list with:
\itemize{
\item \code{SummaryTable}: \link{data.frame} with information on such problems
\item \code{SummaryData}: \link{data.frame} human readable version of \code{SummaryTable}
\item \code{FlaggedStudyData}: \link{data.frame} tells for each entry in study data if
its encoding is OK. has the same dimensions as
\code{study_data}
}
}
\description{
Detects errors in the character encoding of string variables

\link{Indicator}
}
\details{
Strings are stored based on
\href{https://en.wikipedia.org/wiki/Character_encoding}{code tables}, nowadays,
typically as
\href{https://en.wikipedia.org/wiki/UTF-8}{UTF-8}. However, other code systems
are still in use, so, sometimes, strings from different systems are mixed
in the data. This indicator checks for such problems and returns the count
of entries per variable, that do not match the reference coding system, which
is estimated from the study data (addition of metadata field is planned).

If not specified in the metadata (columns \code{ENCODING} in item- or data-frame-
level, the encoding is guessed from the data). Otherwise, it may be any
supported encoding as returned by \code{iconvlist()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_cat_distributions.R
\name{acc_cat_distributions}
\alias{acc_cat_distributions}
\title{Plots and checks for distributions for categorical variables}
\usage{
acc_cat_distributions(
  resp_vars = NULL,
  group_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2,
  n_cat_max = getOption("dataquieR.max_cat_resp_var_levels_in_plot",
    dataquieR.max_cat_resp_var_levels_in_plot_default),
  n_group_max = getOption("dataquieR.max_group_var_levels_in_plot",
    dataquieR.max_group_var_levels_in_plot_default),
  n_data_min = getOption("dataquieR.min_time_points_for_cat_resp_var",
    dataquieR.min_time_points_for_cat_resp_var_default)
)
}
\arguments{
\item{resp_vars}{\link{variable} the name of the measurement variable}

\item{group_vars}{\link{variable} the name of the observer, device or
reader variable}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{n_cat_max}{maximum number of categories to be displayed individually
for the categorical variable (\code{resp_vars})}

\item{n_group_max}{maximum number of categories to be displayed individually
for the grouping variable (\code{group_vars}, devices / examiners)}

\item{n_data_min}{minimum number of data points to create a time course plot
for an individual category of the \code{resp_vars} variable}
}
\value{
A \link{list} with:
\itemize{
\item \code{SummaryPlot}: \link[ggplot2:ggplot]{ggplot2::ggplot} for the response variable in
\code{resp_vars}.
}
}
\description{
This function creates distribution plots for categorical variables.

\link{Descriptor}
}
\details{
To complete
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline_vectorized.R
\name{pipeline_vectorized}
\alias{pipeline_vectorized}
\title{Call (nearly) one "Accuracy" function with many parameterizations at once
automatically}
\usage{
pipeline_vectorized(...)
}
\arguments{
\item{...}{Deprecated}
}
\value{
Deprecated
}
\description{
Deprecated
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_valuelabels_from_data.R
\name{prep_valuelabels_from_data}
\alias{prep_valuelabels_from_data}
\title{Get value labels from data}
\usage{
prep_valuelabels_from_data(resp_vars = colnames(study_data), study_data)
}
\arguments{
\item{resp_vars}{\link{variable} names of the variables to fetch the value labels
from the data}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}
}
\value{
a \link{list} with:
\itemize{
\item \code{VALUE_LABELS}: vector of value labels and modified study data
\item \code{ModifiedStudyData}: study data with factors as integers
}
}
\description{
Detects factors and converts them to compatible metadata/study data.
}
\examples{
\dontrun{
dataquieR::prep_datatype_from_data(iris)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/des_scatterplot_matrix.R
\name{des_scatterplot_matrix}
\alias{des_scatterplot_matrix}
\title{Compute Pairwise Correlations}
\usage{
des_scatterplot_matrix(
  label_col,
  study_data,
  item_level = "item_level",
  meta_data_cross_item = "cross-item_level",
  meta_data = item_level,
  meta_data_v2,
  cross_item_level,
  `cross-item_level`
)
}
\arguments{
\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data_cross_item}{\link{meta_data_cross}}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{cross_item_level}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{`cross-item_level`}{\link{data.frame} alias for \code{meta_data_cross_item}}
}
\value{
a \code{list} with the slots:
\itemize{
\item \code{SummaryPlotList}: for each variable group a \link[ggplot2:ggplot]{ggplot2::ggplot} object with
pairwise correlation plots
\item \code{SummaryData}: table with columns \code{VARIABLE_LIST}, \code{cors},
\code{max_cor}, \code{min_cor}
\item \code{SummaryTable}: like \code{SummaryData}, but machine readable and with
stable column names.
}
}
\description{
works on variable groups (\code{cross-item_level}), which are expected to show
a Pearson correlation
}
\details{
\link{Descriptor} # TODO: This can be an indicator
}
\examples{
\dontrun{
devtools::load_all()
prep_load_workbook_like_file("meta_data_v2")
des_scatterplot_matrix("study_data")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{MISSING_CODE_RULES}
\alias{MISSING_CODE_RULES}
\title{Name of the sheet with rules to introduce missing codes in the pipeline}
\usage{
MISSING_CODE_RULES
}
\description{
Name of the sheet with rules to introduce missing codes in the pipeline
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{check_table}
\alias{check_table}
\title{Data frame with contradiction rules}
\description{
Two versions exist, the newer one is used by \link{con_contradictions_redcap} and
is described \href{https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions_redcap.html#Example_output}{here}.,
the older one used by \link{con_contradictions} is described
\href{https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions.html#Example_output}{here}.
}
\seealso{
\link{meta_data_cross}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_list_voc.R
\name{prep_list_voc}
\alias{prep_list_voc}
\title{All valid \verb{voc:} vocabularies}
\usage{
prep_list_voc()
}
\value{
\code{\link[=character]{character()}} all \verb{voc:} suffixes allowed for
\code{\link[=prep_get_data_frame]{prep_get_data_frame()}}.
}
\description{
All valid \verb{voc:} vocabularies
}
\examples{
\dontrun{
prep_list_dataframes()
prep_list_voc()
prep_get_data_frame("<ICD10>")
my_voc <-
  tibble::tribble(
    ~ voc, ~ url,
    "test", "data:datasets|iris|Species+Sepal.Length")
prep_add_data_frames(`<>` = my_voc)
prep_list_dataframes()
prep_list_voc()
prep_get_data_frame("<test>")
prep_get_data_frame("<ICD10>")
my_voc <-
  tibble::tribble(
    ~ voc, ~ url,
    "ICD10", "data:datasets|iris|Species+Sepal.Length")
prep_add_data_frames(`<>` = my_voc)
prep_list_dataframes()
prep_list_voc()
prep_get_data_frame("<ICD10>")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.dataquieR_resultset2.R
\name{print.dataquieR_resultset2}
\alias{print.dataquieR_resultset2}
\title{Generate a HTML-based report from a \link{dataquieR} report}
\usage{
\method{print}{dataquieR_resultset2}(
  x,
  dir,
  view = TRUE,
  disable_plotly = FALSE,
  block_load_factor = getOption("dataquieR.print_block_load_factor",
    dataquieR.print_block_load_factor_default),
  advanced_options = list(),
  dashboard = NA,
  force_overwrite = FALSE,
  ...,
  cores = list(mode = "socket", logging = FALSE, cpus = util_detect_cores(),
    load.balancing = TRUE)
)
}
\arguments{
\item{x}{\link[=dq_report2]{dataquieR report v2}.}

\item{dir}{\link{character} directory to store the rendered report's files,
a temporary one,
if omitted. Directory will be created, if missing}

\item{view}{\link{logical} display the report}

\item{disable_plotly}{\link{logical} do not use \code{plotly}, even if installed}

\item{block_load_factor}{\link{numeric} see \link{dataquieR.print_block_load_factor}}

\item{advanced_options}{\link{list} options to set during report computation,
see \code{\link[=options]{options()}}}

\item{dashboard}{\link{logical} dashboard mode: \code{TRUE}: create a dashboard only,
\code{FALSE}: don't create a dashboard at all,
\code{NA} or missing: create a "normal" report with
a dashboard included.}

\item{force_overwrite}{\link{logical} force to overwrite \code{dir}, even if it exists}

\item{...}{additional arguments:}

\item{cores}{\link{integer} number of cpu cores to use or a named list with
arguments for \link[parallelMap:parallelStart]{parallelMap::parallelStart} or NULL,
if parallel has already been started by the caller.
Can also be a cluster.}
}
\value{
file names of the generated report's HTML files
}
\description{
Generate a HTML-based report from a \link{dataquieR} report
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_get_variant.R
\name{prep_get_variant}
\alias{prep_get_variant}
\title{Get machine variant for snapshot tests}
\usage{
prep_get_variant()
}
\value{
\link{character} the variant
}
\description{
Get machine variant for snapshot tests
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_link_escape.R
\name{prep_link_escape}
\alias{prep_link_escape}
\title{Prepare a label as part of a link for \code{RMD} files}
\usage{
prep_link_escape(s, html = FALSE)
}
\arguments{
\item{s}{the label}

\item{html}{prepare the label for direct \code{HTML} output instead of \code{RMD}}
}
\value{
the escaped label
}
\description{
Prepare a label as part of a link for \code{RMD} files
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/html_dependency_tippy.R
\name{html_dependency_tippy}
\alias{html_dependency_tippy}
\title{HTML Dependency for \code{tippy}}
\usage{
html_dependency_tippy()
}
\value{
the dependency
}
\description{
HTML Dependency for \code{tippy}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_limit_deviations.R
\name{con_limit_deviations}
\alias{con_limit_deviations}
\title{Detects variable values exceeding limits defined in metadata}
\usage{
con_limit_deviations(
  resp_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  meta_data_cross_item = "cross-item_level",
  limits = NULL,
  flip_mode = "noflip",
  return_flagged_study_data = FALSE,
  return_limit_categorical = TRUE,
  meta_data = item_level,
  cross_item_level,
  `cross-item_level`,
  meta_data_v2,
  show_obs = TRUE
)
}
\arguments{
\item{resp_vars}{\link{variable list} the name of the measurement variables}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data_cross_item}{\link{meta_data_cross}}

\item{limits}{\link{enum} HARD_LIMITS | SOFT_LIMITS | DETECTION_LIMITS. what
limits from metadata to check for}

\item{flip_mode}{\link{enum} default | flip | noflip | auto. Should the plot be
in default orientation, flipped, not flipped or
auto-flipped. Not all options are always supported.
In general, this con be controlled by
setting the \code{roptions(dataquieR.flip_mode = ...)}. If
called from \code{dq_report}, you can also pass
\code{flip_mode} to all function calls or set them
specifically using \code{specific_args}.}

\item{return_flagged_study_data}{\link{logical} return \code{FlaggedStudyData} in the
result}

\item{return_limit_categorical}{\link{logical} if TRUE return limit deviations also
for categorical variables}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{cross_item_level}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{`cross-item_level`}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{show_obs}{\link{logical} Should (selected) individual observations be marked
in the figure for continuous variables?}
}
\value{
a list with:
\itemize{
\item \code{FlaggedStudyData} \link{data.frame} related to the study data by a 1:1
relationship, i.e. for each observation is
checked whether the value is below or above
the limits. Optional, see
\code{return_flagged_study_data}.
\item \code{SummaryTable} \link{data.frame} summarizing limit deviations for each
variable.
\item \code{SummaryData} \link{data.frame} summarizing limit deviations for each
variable for a report.
\item \code{SummaryPlotList} \link{list} of \link[ggplot2:ggplot]{ggplot2::ggplot}s The plots for each variable are
either a histogram (continuous) or a
barplot (discrete).
\item \code{ReportSummaryTable}: heatmap-like data frame about limit violations
}
}
\description{
Inadmissible numerical values can be of type integer or float. This
implementation requires the definition of intervals in the metadata to
examine the admissibility of numerical study data.

This helps identify inadmissible measurements according to
hard limits (for multiple variables).

\link{Indicator}
}
\details{
\subsection{Algorithm of this implementation:}{
\itemize{
\item Remove missing codes from the study data (if defined in the metadata)
\item Interpretation of variable specific intervals as supplied in the metadata.
\item Identification of measurements outside defined limits. Therefore two
output data frames are generated:
\itemize{
\item on the level of observation to flag each deviation, and
\item a summary table for each variable.
}
\item A list of plots is generated for each variable examined for limit
deviations. The histogram-like plots indicate respective limits as well
as deviations.
\item Values exceeding limits are removed in a data frame of modified study data
}
}
}
\seealso{
\itemize{
\item \href{https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_limit_deviations.html}{Online Documentation}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_unexp_elements.R
\name{int_unexp_elements}
\alias{int_unexp_elements}
\title{Check for unexpected data element count}
\usage{
int_unexp_elements(
  identifier_name_list,
  data_element_count,
  meta_data_dataframe = "dataframe_level",
  meta_data_v2,
  dataframe_level
)
}
\arguments{
\item{identifier_name_list}{\link{character} a character vector indicating the name of each study data frame, mandatory.}

\item{data_element_count}{\link{integer}  an integer vector with the number of expected data elements, mandatory.}

\item{meta_data_dataframe}{\link{data.frame} the data frame that contains the
metadata for the data frame level}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}
}
\value{
a \link{list} with
\itemize{
\item \code{DataframeData}: data frame with the results of the quality check for unexpected data elements
\item \code{DataframeTable}: data frame with selected unexpected data elements check results, used for the data quality report.
}
}
\description{
This function contrasts the expected element number in each study in
the metadata with the actual element number in each study data frame.

\link{Indicator}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_options.R
\docType{data}
\name{dataquieR.guess_character}
\alias{dataquieR.guess_character}
\title{For metadata guessing, try to guess \link{DATA_TYPE} from the data values}
\description{
By default, the \link{DATA_TYPE} is derived from the R data type of the study
data. However, when data are imported from plain text files, it can be more
appropriate to examine the actual values and infer the data type based on
their content. This option enables that behavior: set
\code{dataquieR.guess_character} to \code{TRUE} to infer data types from the observed
values rather than relying solely on the column’s \code{class} in the data frame.
}
\seealso{
Other options:
\code{\link{dataquieR}},
\code{\link{dataquieR.CONDITIONS_LEVEL_TRHESHOLD}},
\code{\link{dataquieR.CONDITIONS_WITH_STACKTRACE}},
\code{\link{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE}},
\code{\link{dataquieR.ERRORS_WITH_CALLER}},
\code{\link{dataquieR.GAM_for_LOESS}},
\code{\link{dataquieR.MAHALANOBIS_THRESHOLD}},
\code{\link{dataquieR.MAX_LABEL_LEN}},
\code{\link{dataquieR.MAX_LONG_LABEL_LEN}},
\code{\link{dataquieR.MAX_VALUE_LABEL_LEN}},
\code{\link{dataquieR.MESSAGES_WITH_CALLER}},
\code{\link{dataquieR.MULTIVARIATE_OUTLIER_CHECK}},
\code{\link{dataquieR.VALUE_LABELS_htmlescaped}},
\code{\link{dataquieR.WARNINGS_WITH_CALLER}},
\code{\link{dataquieR.acc_loess.exclude_constant_subgroups}},
\code{\link{dataquieR.acc_loess.mark_time_points}},
\code{\link{dataquieR.acc_loess.min_bw}},
\code{\link{dataquieR.acc_loess.min_obs_in_subgroup}},
\code{\link{dataquieR.acc_loess.min_proportion}},
\code{\link{dataquieR.acc_loess.plot_format}},
\code{\link{dataquieR.acc_loess.plot_observations}},
\code{\link{dataquieR.acc_margins_num}},
\code{\link{dataquieR.acc_margins_sort}},
\code{\link{dataquieR.acc_multivariate_outlier.scale}},
\code{\link{dataquieR.acc_shape_or_scale_ci}},
\code{\link{dataquieR.col_con_con_empirical}},
\code{\link{dataquieR.col_con_con_logical}},
\code{\link{dataquieR.convert_to_list_for_lapply}},
\code{\link{dataquieR.debug}},
\code{\link{dataquieR.des_summary_hard_lim_remove}},
\code{\link{dataquieR.dontwrapresults}},
\code{\link{dataquieR.droplevels_ReportSummaryTable}},
\code{\link{dataquieR.dt_adjust}},
\code{\link{dataquieR.fix_column_type_on_read}},
\code{\link{dataquieR.flip_mode}},
\code{\link{dataquieR.force_item_specific_missing_codes}},
\code{\link{dataquieR.force_label_col}},
\code{\link{dataquieR.grading_formats}},
\code{\link{dataquieR.grading_rulesets}},
\code{\link{dataquieR.guess_missing_codes}},
\code{\link{dataquieR.ignore_empty_vars}},
\code{\link{dataquieR.lang}},
\code{\link{dataquieR.lazy_plots}},
\code{\link{dataquieR.lazy_plots_cache}},
\code{\link{dataquieR.lazy_plots_gg_compatibility}},
\code{\link{dataquieR.locale}},
\code{\link{dataquieR.max_cat_resp_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_with_violins}},
\code{\link{dataquieR.min_obs_per_group_var_in_plot}},
\code{\link{dataquieR.min_time_points_for_cat_resp_var}},
\code{\link{dataquieR.no_geom_count_in_bin}},
\code{\link{dataquieR.no_overall_in_bin}},
\code{\link{dataquieR.non_disclosure}},
\code{\link{dataquieR.old_factor_handling}},
\code{\link{dataquieR.old_type_adjust}},
\code{\link{dataquieR.precomputeStudyData}},
\code{\link{dataquieR.print_block_load_factor}},
\code{\link{dataquieR.progress_fkt_default}},
\code{\link{dataquieR.progress_msg_fkt_default}},
\code{\link{dataquieR.resume_checkpoint}},
\code{\link{dataquieR.resume_print}},
\code{\link{dataquieR.scale_level_heuristics_control_binaryrecodelimit}},
\code{\link{dataquieR.scale_level_heuristics_control_metriclevels}},
\code{\link{dataquieR.study_data_cache_max}},
\code{\link{dataquieR.study_data_cache_metrics}},
\code{\link{dataquieR.study_data_cache_metrics_env}},
\code{\link{dataquieR.study_data_cache_quick_fill}},
\code{\link{dataquieR.study_data_colnames_case_sensitive}},
\code{\link{dataquieR.testdebug}},
\code{\link{dataquieR.traceback}},
\code{\link{dataquieR.type_adjust_parallel}},
\code{\link{progress_init_fkt}}
}
\concept{options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{ASSOCIATION_RANGE}
\alias{ASSOCIATION_RANGE}
\title{Cross-item level metadata attribute name}
\usage{
ASSOCIATION_RANGE
}
\description{
Specifies the allowable range of an association. The inclusion of the
endpoints follows standard mathematical notation using round brackets
for open intervals and square brackets for closed intervals.
Values must be separated by a semicolon.
}
\seealso{
\link{meta_data_cross}

Other meta_data_cross:
\code{\link{ASSOCIATION_DIRECTION}},
\code{\link{ASSOCIATION_FORM}},
\code{\link{ASSOCIATION_METRIC}},
\code{\link{CHECK_ID}},
\code{\link{CHECK_LABEL}},
\code{\link{COMPUTED_VARIABLE_ROLES}},
\code{\link{CONTRADICTION_TERM}},
\code{\link{CONTRADICTION_TYPE}},
\code{\link{DATA_PREPARATION}},
\code{\link{GOLDSTANDARD}},
\code{\link{IRV}},
\code{\link{MAHALANOBIS_RATIO}},
\code{\link{MAHALANOBIS_THRESHOLD}},
\code{\link{MAXIMUM_LONG_STRING}},
\code{\link{MISS_RESP}},
\code{\link{MULTIVARIATE_OUTLIER_CHECK}},
\code{\link{MULTIVARIATE_OUTLIER_CHECKTYPE}},
\code{\link{RELCOMPL_SPEED}},
\code{\link{REL_VAL}},
\code{\link{RESPT_PER_ITEM}},
\code{\link{SCALE_ACRONYM}},
\code{\link{SCALE_NAME}},
\code{\link{TOTRESPT}},
\code{\link{VARIABLE_LIST}},
\code{\link{VARIABLE_LIST_ORDER}},
\code{\link{meta_data_computation}},
\code{\link{meta_data_cross}}
}
\concept{meta_data_cross}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_extract_cause_label_df.R
\name{prep_extract_cause_label_df}
\alias{prep_extract_cause_label_df}
\title{Extract all missing/jump codes from metadata and export a
cause-label-data-frame}
\usage{
prep_extract_cause_label_df(
  item_level = "item_level",
  label_col = VAR_NAMES,
  meta_data_v2,
  meta_data = item_level
)
}
\arguments{
\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}
}
\value{
\link{list} with the entries
\itemize{
\item \code{meta_data} \link{data.frame} a data frame that contains updated metadata --
you still need to add a column
\link{MISSING_LIST_TABLE} and add the
\code{cause_label_df} as such to the metadata
cache using \code{\link[=prep_add_data_frames]{prep_add_data_frames()}}, manually.
\item \code{cause_label_df} \link{data.frame} missing code table. If missing codes have
labels the respective data frame are
specified here, see \link{cause_label_df}.
}
}
\description{
Extract all missing/jump codes from metadata and export a
cause-label-data-frame
}
\seealso{
\link{prep_add_cause_label_df}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_scalelevel_from_data_and_metadata.R
\name{prep_scalelevel_from_data_and_metadata}
\alias{prep_scalelevel_from_data_and_metadata}
\title{Heuristics to amend a SCALE_LEVEL column and a UNIT column in the metadata}
\usage{
prep_scalelevel_from_data_and_metadata(
  resp_vars = lifecycle::deprecated(),
  study_data,
  item_level = "item_level",
  label_col = LABEL,
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable list} deprecated, the function always addresses
all variables.}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
\link{data.frame} modified metadata
}
\description{
...if missing
}
\examples{
\dontrun{
  prep_load_workbook_like_file("meta_data_v2")
  prep_scalelevel_from_data_and_metadata(study_data = "study_data")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_all_datastructure_dataframe.R
\name{int_all_datastructure_dataframe}
\alias{int_all_datastructure_dataframe}
\title{Wrapper function to check for studies data structure}
\usage{
int_all_datastructure_dataframe(
  meta_data_dataframe = "dataframe_level",
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2,
  dataframe_level
)
}
\arguments{
\item{meta_data_dataframe}{\link{data.frame} the data frame that contains the
metadata for the data frame level}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}
}
\value{
a \link{list} with
\itemize{
\item \code{DataframeTable}: data frame with selected check results, used for the data quality report.
}
}
\description{
This function tests for unexpected elements and records, as well as duplicated identifiers and content.
The unexpected element record check can be conducted by providing the number of expected records or
an additional table with the expected records.
It is possible to conduct the checks by study segments or to consider only selected
segments.

\link{Indicator}
}
\examples{
\dontrun{
out_dataframe <- int_all_datastructure_dataframe(
  meta_data_dataframe = "meta_data_dataframe",
  meta_data = "ship_meta"
)
md0 <- prep_get_data_frame("ship_meta")
md0
md0$VAR_NAMES
md0$VAR_NAMES[[1]] <- "Id" # is this missmatch reported -- is the data frame
                           # also reported, if nothing is wrong with it
out_dataframe <- int_all_datastructure_dataframe(
  meta_data_dataframe = "meta_data_dataframe",
  meta_data = md0
)

# This is the "normal" procedure for inside pipeline
# but outside this function  checktype is exact by default
options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "subset_u")
lapply(setNames(nm = prep_get_data_frame("meta_data_dataframe")$DF_NAME),
  int_sts_element_dataframe, meta_data = md0)
md0$VAR_NAMES[[1]] <-
  "id" # is this missmatch reported -- is the data frame also reported,
       # if nothing is wrong with it
lapply(setNames(nm = prep_get_data_frame("meta_data_dataframe")$DF_NAME),
  int_sts_element_dataframe, meta_data = md0)
options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DF_ELEMENT_COUNT}
\alias{DF_ELEMENT_COUNT}
\title{Data frame level metadata attribute name}
\usage{
DF_ELEMENT_COUNT
}
\description{
Number of expected data elements in a data frame.
\link{numeric}. Check only conducted if number entered
}
\seealso{
\link{meta_data_dataframe}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.dataquieR_result.R
\name{print.dataquieR_result}
\alias{print.dataquieR_result}
\alias{dataquieR_result}
\title{Print a \link{dataquieR} result returned by \link{dq_report2}}
\usage{
\method{print}{dataquieR_result}(x, ...)
}
\arguments{
\item{x}{\link{list} a dataquieR result from \link{dq_report2} or
\code{util_eval_to_dataquieR_result}}

\item{...}{passed to print. Additionally, the argument \code{slot} may be passed
to print only specific sub-results.}
}
\value{
see print
}
\description{
Print a \link{dataquieR} result returned by \link{dq_report2}
}
\seealso{
\code{util_pretty_print()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset2.R
\name{[<-.dataquieR_resultset2}
\alias{[<-.dataquieR_resultset2}
\title{Write to a report}
\usage{
\method{[}{dataquieR_resultset2}(x, ...) <- value
}
\arguments{
\item{x}{a `dataquieR_resultset2}

\item{...}{if this contains only one entry and this entry is not named
or its name is \code{els}, then, the report will be accessed in
list mode.}

\item{value}{new value to write}
}
\value{
nothing, stops
}
\description{
Overwriting of elements only list-wise supported
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/html_dependency_vert_dt.R
\name{html_dependency_vert_dt}
\alias{html_dependency_vert_dt}
\title{HTML Dependency for vertical headers in \code{DT::datatable}}
\usage{
html_dependency_vert_dt()
}
\value{
the dependency
}
\description{
HTML Dependency for vertical headers in \code{DT::datatable}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{study_data}
\alias{study_data}
\title{Data frame with the study data whose quality is being assessed}
\description{
Study data is expected in wide format. If should contain all variables
for all segments in one large table, even, if some variables are not
measured for all observational utils (study participants).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_fix_meta_id_dups.R
\name{prep_fix_meta_id_dups}
\alias{prep_fix_meta_id_dups}
\title{Fix metadata duplicates}
\usage{
prep_fix_meta_id_dups(
  meta_data_segment = "segment_level",
  meta_data_dataframe = "dataframe_level",
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2,
  segment_level,
  dataframe_level
)
}
\arguments{
\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{meta_data_dataframe}{\link{data.frame} the data frame that contains the
metadata for the data frame level}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}
}
\value{
\link{meta_data}
}
\description{
if \link{VAR_NAMES} have duplicates, maybe, it's because of ID-vars assigned
to different study segments multiple times (they should be in one "intro"-
segment, only), which is not the intended use of \code{STUDY_SEGMENT}.
Naturally, they will be part of more than one data-frame, so
this would also qualify for a dump duplicate, only, which can safely be
removed. Only ID-vars are by default assumed to have such duplicates in item
level metadata allowed.
}
\examples{
\dontrun{
il <- prep_get_data_frame("item_level")
il <- rbind(il, il)
il$STUDY_SEGMENT[2] <- "X"
il2 <- prep_fix_meta_id_dups(meta_data_v2 = "meta_data_v2", item_level = il)
il2$STUDY_SEGMENT
il$STUDY_SEGMENT[3] <- "X"
il3 <- prep_fix_meta_id_dups(meta_data_v2 = "meta_data_v2", item_level = il)
il3$STUDY_SEGMENT
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_translate.R
\name{print.dataquieR_translated}
\alias{print.dataquieR_translated}
\title{\code{print} implementation for the class \code{dataquieR_translated}}
\usage{
\method{print}{dataquieR_translated}(x, ...)
}
\arguments{
\item{x}{\code{dataquieR_translated} object to print}

\item{...}{passed to base::print}
}
\value{
as print
}
\description{
\code{dataquieR}'s translated texts featuring access to the language keys, still.
}
\seealso{
base::print
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.dataquieR_result.R
\name{print.Slot}
\alias{print.Slot}
\title{Print a \code{Slot} object}
\usage{
\method{print}{Slot}(x, ...)
}
\arguments{
\item{x}{the object}

\item{...}{not used}
}
\value{
calls the next print method
}
\description{
displays all warnings and stuff. then it prints \code{x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_distributions.R
\name{acc_distributions_only}
\alias{acc_distributions_only}
\title{Plots and checks for distributions -- only}
\usage{
acc_distributions_only(
  resp_vars = NULL,
  study_data,
  label_col = VAR_NAMES,
  item_level = "item_level",
  flip_mode = "noflip",
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable list} the names of the measurement variables}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{flip_mode}{\link{enum} default | flip | noflip | auto. Should the plot be
in default orientation, flipped, not flipped or
auto-flipped. Not all options are always supported.
In general, this con be controlled by
setting the \code{roptions(dataquieR.flip_mode = ...)}. If
called from \code{dq_report}, you can also pass
\code{flip_mode} to all function calls or set them
specifically using \code{specific_args}.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
A \link{list} with:
\itemize{
\item \code{SummaryTable}: \link{data.frame} containing data quality checks for
"Unexpected location" (\code{FLG_acc_ud_loc}) and "Unexpected
proportion" (\code{FLG_acc_ud_prop}) for each response
variable in \code{resp_vars}.
\item \code{SummaryData}: a \link{data.frame} containing data quality checks for
"Unexpected location" and / or "Unexpected proportion"
for a report
\item \code{SummaryPlotList}: \link{list} of \link[ggplot2:ggplot]{ggplot2::ggplot}s for each response variable in
\code{resp_vars}.
}
}
\description{
\link{Descriptor}
}
\section{Algorithm of this implementation:}{
\itemize{
\item If no response variable is defined, select all variables of type float or
integer in the study data.
\item Remove missing codes from the study data (if defined in the metadata).
\item Remove measurements deviating from (hard) limits defined in the metadata
(if defined).
\item Exclude variables containing only \code{NA} or only one unique value (excluding
\code{NA}s).
\item Perform check for "Unexpected location" if defined in the metadata (needs a
LOCATION_METRIC (mean or median) and LOCATION_RANGE (range of expected values
for the mean and median, respectively)).
\item Perform check for "Unexpected proportion" if defined in the metadata (needs
PROPORTION_RANGE (range of expected values for the proportions of the
categories)).
\item Plot histogram(s).
}
}

\seealso{
\itemize{
\item \link{acc_distributions}
\item \href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_distributions.html}{Online Documentation}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.ReportSummaryTable.R
\name{print.ReportSummaryTable}
\alias{print.ReportSummaryTable}
\title{print implementation for the class \code{ReportSummaryTable}}
\usage{
\method{print}{ReportSummaryTable}(
  x,
  relative = lifecycle::deprecated(),
  dt = FALSE,
  fillContainer = FALSE,
  displayValues = FALSE,
  view = TRUE,
  drop = getOption("dataquieR.droplevels_ReportSummaryTable",
    dataquieR.droplevels_ReportSummaryTable_default),
  ...,
  flip_mode = "auto"
)
}
\arguments{
\item{x}{\code{ReportSummaryTable} objects to print}

\item{relative}{deprecated}

\item{dt}{\link{logical} use \code{DT::datatables}, if installed}

\item{fillContainer}{\link{logical} if \code{dt} is \code{TRUE}, control table size,
see \code{DT::datatables}.}

\item{displayValues}{\link{logical} if \code{dt} is \code{TRUE}, also display the actual
values}

\item{view}{\link{logical} if \code{view} is \code{FALSE}, do not print but return the
output, only}

\item{drop}{\link{logical} if \code{drop} is \code{FALSE}, keep unused levels,
see \link{dataquieR.droplevels_ReportSummaryTable}}

\item{...}{not used, yet}

\item{flip_mode}{\link{enum} default | flip | noflip | auto. Should the plot be
in default orientation, flipped, not flipped or
auto-flipped. Not all options are always supported.
In general, this con be controlled by
setting the \code{roptions(dataquieR.flip_mode = ...)}. If
called from \code{dq_report}, you can also pass
\code{flip_mode} to all function calls or set them
specifically using \code{specific_args}.}
}
\value{
the printed object
}
\description{
Use this function to print results objects of the class
\code{ReportSummaryTable}.
}
\seealso{
base::print
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_margins.R
\name{acc_margins}
\alias{acc_margins}
\title{Estimate marginal means, see \link[emmeans:emmeans]{emmeans::emmeans}}
\usage{
acc_margins(
  resp_vars = NULL,
  group_vars = NULL,
  co_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  threshold_type = "empirical",
  threshold_value,
  min_obs_in_subgroup = 5,
  min_obs_in_cat = 5,
  dichotomize_categorical_resp = TRUE,
  cut_off_linear_model_for_ord = 10,
  meta_data = item_level,
  meta_data_v2,
  sort_group_var_levels = getOption("dataquieR.acc_margins_sort",
    dataquieR.acc_margins_sort_default),
  include_numbers_in_figures = getOption("dataquieR.acc_margins_num",
    dataquieR.acc_margins_num_default),
  n_violin_max = getOption("dataquieR.max_group_var_levels_with_violins",
    dataquieR.max_group_var_levels_with_violins_default),
  no_overall_in_bin = getOption("dataquieR.no_overall_in_bin",
    dataquieR.no_overall_in_bin_default),
  no_geom_count_in_bin = getOption("dataquieR.no_geom_count_in_bin",
    dataquieR.no_geom_count_in_bin_default)
)
}
\arguments{
\item{resp_vars}{\link{variable} the name of the measurement variable}

\item{group_vars}{\link{variable list} len=1-1. the name of the observer, device
or reader variable}

\item{co_vars}{\link{variable list} a vector of covariables, e.g. age and sex for
adjustment}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{threshold_type}{\link{enum} empirical | user | none. In case \code{empirical} is
chosen, a multiplier of the scale measure is used.
In case of \code{user}, a value of the mean or probability
(binary data) has to be defined
see \href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_margins.html#Implementation_and_use_of_thresholds}{\verb{Implementation and use of thresholds}} in the  online documentation).
In case of \code{none}, no thresholds are displayed and no
flagging of unusual group levels is applied.}

\item{threshold_value}{\link{numeric} a multiplier or absolute value (see
\href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_margins.html#Implementation_and_use_of_thresholds}{\verb{Implementation and use of thresholds}} in the
online documentation).}

\item{min_obs_in_subgroup}{\link{integer} from=0. This optional argument specifies
the minimum number of observations that is required to
include a subgroup (level) of the \code{group_var} in the
analysis. Subgroups with less observations are
excluded.}

\item{min_obs_in_cat}{\link{integer} This optional argument specifies the minimum
number of observations that is required to include
a category (level) of the outcome (\code{resp_vars}) in
the analysis. Categories with less observations are
combined into one group. If the collapsed category
contains less observations than required, it will be
excluded from the analysis.}

\item{dichotomize_categorical_resp}{\link{logical} Should nominal response
variables always be transformed to binary variables?}

\item{cut_off_linear_model_for_ord}{\link{integer} from=0. This optional argument
specifies the minimum number of observations for
individual levels of an ordinal outcome (\code{resp_var})
that is required to run a linear model instead of an
ordered regression (i.e., a cut-off value above which
linear models are considered a good approximation).
The argument can be set to \code{NULL} if ordered
regression models are preferred for ordinal data in
any case.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{sort_group_var_levels}{\link{logical} Should the levels of the grouping
variable be sorted descending by the number of
observations? Note that ordinal grouping variables
will not be reordered.}

\item{include_numbers_in_figures}{\link{logical} Should the figure report the
number of observations for each level of the grouping
variable?}

\item{n_violin_max}{\link{integer} from=0. This optional argument specifies
the maximum number of levels of the \code{group_var} for
which violin plots will be shown in the figure.}

\item{no_overall_in_bin}{\link{logical} Suppress overall distribution in 'margins'
figures for binary outcomes}

\item{no_geom_count_in_bin}{\link{logical} Suppress counts 'margins'
figures for binary outcomes, so they
.                                      are not always including 0 and 1.}
}
\value{
a list with:
\itemize{
\item \code{SummaryTable}: \link{data.frame} underlying the plot
\item \code{ResultData}: \link{data.frame}
\item \code{SummaryPlot}: \code{\link[ggplot2:ggplot]{ggplot2::ggplot()}} margins plot
}
}
\description{
This function examines the impact of so-called process variables on a
measurement variable. This implementation combines a descriptive and a
model-based approach. Process variables that can be considered in this
implementation must be categorical. It is currently not possible to
consider more than one process variable within one function call.
The measurement variable can be adjusted for (multiple) covariables, such as
age or sex, for example.

Marginal means rests on model-based results, i.e. a significantly different
marginal mean depends on sample size. Particularly in large studies, small
and irrelevant differences may become significant. The contrary holds if
sample size is low.

\link{Indicator}
}
\details{
Limitations

Selecting the appropriate distribution is complex. Dozens of continuous,
discrete or mixed distributions are conceivable in the context of
epidemiological data. Their exact exploration is beyond the scope of this
data quality approach. The present function uses the help function
\code{util_dist_selection}, the assigned \code{SCALE_LEVEL} and the \code{DATA_TYPE}
to discriminate the following cases:
\itemize{
\item continuous data
\item binary data
\item count data with <= 20 distinct values
\item count data with > 20 distinct values (treated as continuous)
\item nominal data
\item ordinal data
}
Continuous data and count data with more than 20 distinct values are analyzed
by linear models. Count data with up to 20 distinct values are modeled by a
Poisson regression. For binary data, the implementation uses logistic
regression.
Nominal response variables will either be transformed to binary variables or
analyzed by multinomial logistic regression models. The latter option is only
available if the argument \code{dichotomize_categorical_resp} is set to \code{FALSE}
and if the package \code{nnet} is installed. The transformation to a binary
variable can be user-specified using the metadata columns \code{RECODE_CASES}
and/or \code{RECODE_CONTROL}. Otherwise, the most frequent category will be
assigned to cases and the remaining categories to control.
For ordinal response variables, the argument \code{cut_off_linear_model_for_ord}
controls whether the data is analyzed in the same way as continuous data:
If every level of the variable has at least as many observations as specified
in the argument, the data will be analyzed by a linear model. Otherwise,
the data will be modeled by a ordered regression, if the package \code{ordinal}
is installed.
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_margins.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_translate_indicator_metrics.R
\name{util_translate_indicator_metrics}
\alias{util_translate_indicator_metrics}
\title{Translate standard column names to readable ones}
\usage{
util_translate_indicator_metrics(
  colnames,
  short = FALSE,
  long = TRUE,
  ignore_unknown = FALSE
)
}
\arguments{
\item{colnames}{\link{character} the names to translate}

\item{short}{\link{logical} include unit letter in output}

\item{long}{\link{logical} include unit description in output}

\item{ignore_unknown}{\link{logical} do not replace unknown indicator metrics
by \code{NA}, keep them}
}
\value{
translated names
}
\description{
TODO: Duplicate of util_make_data_slot_from_table_slot ??
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_inadmissible_categorical.R
\name{con_inadmissible_vocabulary}
\alias{con_inadmissible_vocabulary}
\title{Detects variable levels not specified in standardized vocabulary}
\usage{
con_inadmissible_vocabulary(
  resp_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  threshold_value = 0,
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable list} the name of the measurement variables}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{threshold_value}{\link{numeric} from=0 to=100. a numerical value ranging
from 0-100.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
a list with:
\itemize{
\item \code{SummaryData}: data frame summarizing inadmissible categories with the
columns:
\itemize{
\item \code{Variables}: variable name/label
\item \code{OBSERVED_CATEGORIES}: the categories observed in the study data
\item \code{DEFINED_CATEGORIES}: the categories defined in the metadata
\item \code{NON_MATCHING}: the categories observed but not defined
\item \code{NON_MATCHING_N}: the number of observations with categories not defined
\item \code{NON_MATCHING_N_PER_CATEGORY}: the number of observations for each of the
unexpected categories
\item \code{GRADING}: indicator TRUE/FALSE if inadmissible categorical values were
observed (more than indicated by the \code{threshold_value})
}
\item \code{SummaryTable}: data frame for the \code{dataquieR} pipeline reporting the number
and percentage of inadmissible categorical values
\item \code{ModifiedStudyData}: study data having inadmissible categories removed
\item \code{FlaggedStudyData}: study data having cases with inadmissible categories
flagged
}
}
\description{
For each categorical variable, value lists should be defined in the metadata.
This implementation will examine, if all observed levels in the study data
are valid.

\link{Indicator}
}
\details{
\subsection{Algorithm of this implementation:}{
\itemize{
\item Remove missing codes from the study data (if defined in the metadata)
\item Interpretation of variable specific VALUE_LABELS as supplied in the
metadata.
\item Identification of measurements not corresponding to the expected
categories. Therefore two output data frames are generated:
\itemize{
\item on the level of observation to flag each undefined category, and
\item a summary table for each variable.
}
\item Values not corresponding to defined categories are removed in a data frame
of modified study data
}
}
}
\examples{
\dontrun{
sdt <- data.frame(DIAG = c("B050", "B051", "B052", "B999"),
                  MED0 = c("S01XA28", "N07XX18", "ABC", NA), stringsAsFactors = FALSE)
mdt <- tibble::tribble(
~ VAR_NAMES, ~ DATA_TYPE, ~ STANDARDIZED_VOCABULARY_TABLE, ~ SCALE_LEVEL, ~ LABEL,
"DIAG", "string", "<ICD10>", "nominal", "Diagnosis",
"MED0", "string", "<ATC>", "nominal", "Medication"
)
con_inadmissible_vocabulary(NULL, sdt, mdt, label_col = LABEL)
prep_load_workbook_like_file("meta_data_v2")
il <- prep_get_data_frame("item_level")
il$STANDARDIZED_VOCABULARY_TABLE[[11]] <- "<ICD10GM>"
il$DATA_TYPE[[11]] <- DATA_TYPES$INTEGER
il$SCALE_LEVEL[[11]] <- SCALE_LEVELS$NOMINAL
prep_add_data_frames(item_level = il)
r <- dq_report2("study_data", dimensions = "con")
r <- dq_report2("study_data", dimensions = "con",
     advanced_options = list(dataquieR.non_disclosure = TRUE))
r
}
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_inadmissible_categorical.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_create_meta.R
\name{prep_create_meta}
\alias{prep_create_meta}
\title{Support function to create \link{data.frame}s of metadata}
\usage{
prep_create_meta(..., stringsAsFactors = FALSE, level, character.only = FALSE)
}
\arguments{
\item{...}{named column vectors, names will be mapped using
\link{WELL_KNOWN_META_VARIABLE_NAMES},
if included in \link{WELL_KNOWN_META_VARIABLE_NAMES}
can also be a data frame, then its column names will be mapped
using \link{WELL_KNOWN_META_VARIABLE_NAMES}}

\item{stringsAsFactors}{\link{logical} if the argument is a list of vectors, a
data frame will be
created. In this case, \code{stringsAsFactors} controls,
whether characters will be auto-converted to Factors,
which defaults here always to false independent from
the \link{default.stringsAsFactors}.}

\item{level}{\link{enum} level of requirement (see also \link{VARATT_REQUIRE_LEVELS})
set to \code{NULL}, if not a complete metadata frame is
created.}

\item{character.only}{\link{logical} a logical indicating whether level can be
assumed to be character strings.}
}
\value{
a data frame with:
\itemize{
\item metadata attribute names mapped and
\item metadata checked using \link{prep_check_meta_names} and do some more
verification about conventions, such as check for valid intervals
in limits)
}
}
\description{
Create a metadata data frame and map names.
Generally, this function only creates a \link{data.frame}, but using
this constructor instead of calling
\code{data.frame(..., stringsAsFactors = FALSE)}, it becomes possible, to adapt
the metadata \link{data.frame} in later developments, e.g. if we decide to use
classes for the metadata, or if certain standard names of variable attributes
change. Also, a validity check is possible to implement here.
}
\details{
For now, this calls \link{data.frame}, but it already renames variable attributes,
if they have a different name assigned in \link{WELL_KNOWN_META_VARIABLE_NAMES},
e.g. \code{WELL_KNOWN_META_VARIABLE_NAMES$RECODE} maps to \code{recode} in lower case.

NB: \code{dataquieR} exports all names from WELL_KNOWN_META_VARIABLE_NAME as
symbols, so \code{RECODE} also contains \code{"recode"}.
}
\seealso{
\link{WELL_KNOWN_META_VARIABLE_NAMES}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_options.R
\docType{data}
\name{dataquieR.study_data_cache_max}
\alias{dataquieR.study_data_cache_max}
\title{Maximum size of cache for curated study data}
\description{
\code{dataquieR} caches all used flavors of curated study data, e.g., having
missing codes replaced by \code{NA}s, having hard limits replaced by \code{NA}, ...
For larger sets of study data this can be very RAM consuming, so you can
control here the maximum size for this cache. Also, this cache is distributed
to all compute nodes in case of parallel computation, which may be very time-
consuming, and, on single-node-parallelization, also, it may be even more
RAM-consuming then.
}
\seealso{
Other options:
\code{\link{dataquieR}},
\code{\link{dataquieR.CONDITIONS_LEVEL_TRHESHOLD}},
\code{\link{dataquieR.CONDITIONS_WITH_STACKTRACE}},
\code{\link{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE}},
\code{\link{dataquieR.ERRORS_WITH_CALLER}},
\code{\link{dataquieR.GAM_for_LOESS}},
\code{\link{dataquieR.MAHALANOBIS_THRESHOLD}},
\code{\link{dataquieR.MAX_LABEL_LEN}},
\code{\link{dataquieR.MAX_LONG_LABEL_LEN}},
\code{\link{dataquieR.MAX_VALUE_LABEL_LEN}},
\code{\link{dataquieR.MESSAGES_WITH_CALLER}},
\code{\link{dataquieR.MULTIVARIATE_OUTLIER_CHECK}},
\code{\link{dataquieR.VALUE_LABELS_htmlescaped}},
\code{\link{dataquieR.WARNINGS_WITH_CALLER}},
\code{\link{dataquieR.acc_loess.exclude_constant_subgroups}},
\code{\link{dataquieR.acc_loess.mark_time_points}},
\code{\link{dataquieR.acc_loess.min_bw}},
\code{\link{dataquieR.acc_loess.min_obs_in_subgroup}},
\code{\link{dataquieR.acc_loess.min_proportion}},
\code{\link{dataquieR.acc_loess.plot_format}},
\code{\link{dataquieR.acc_loess.plot_observations}},
\code{\link{dataquieR.acc_margins_num}},
\code{\link{dataquieR.acc_margins_sort}},
\code{\link{dataquieR.acc_multivariate_outlier.scale}},
\code{\link{dataquieR.acc_shape_or_scale_ci}},
\code{\link{dataquieR.col_con_con_empirical}},
\code{\link{dataquieR.col_con_con_logical}},
\code{\link{dataquieR.convert_to_list_for_lapply}},
\code{\link{dataquieR.debug}},
\code{\link{dataquieR.des_summary_hard_lim_remove}},
\code{\link{dataquieR.dontwrapresults}},
\code{\link{dataquieR.droplevels_ReportSummaryTable}},
\code{\link{dataquieR.dt_adjust}},
\code{\link{dataquieR.fix_column_type_on_read}},
\code{\link{dataquieR.flip_mode}},
\code{\link{dataquieR.force_item_specific_missing_codes}},
\code{\link{dataquieR.force_label_col}},
\code{\link{dataquieR.grading_formats}},
\code{\link{dataquieR.grading_rulesets}},
\code{\link{dataquieR.guess_character}},
\code{\link{dataquieR.guess_missing_codes}},
\code{\link{dataquieR.ignore_empty_vars}},
\code{\link{dataquieR.lang}},
\code{\link{dataquieR.lazy_plots}},
\code{\link{dataquieR.lazy_plots_cache}},
\code{\link{dataquieR.lazy_plots_gg_compatibility}},
\code{\link{dataquieR.locale}},
\code{\link{dataquieR.max_cat_resp_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_with_violins}},
\code{\link{dataquieR.min_obs_per_group_var_in_plot}},
\code{\link{dataquieR.min_time_points_for_cat_resp_var}},
\code{\link{dataquieR.no_geom_count_in_bin}},
\code{\link{dataquieR.no_overall_in_bin}},
\code{\link{dataquieR.non_disclosure}},
\code{\link{dataquieR.old_factor_handling}},
\code{\link{dataquieR.old_type_adjust}},
\code{\link{dataquieR.precomputeStudyData}},
\code{\link{dataquieR.print_block_load_factor}},
\code{\link{dataquieR.progress_fkt_default}},
\code{\link{dataquieR.progress_msg_fkt_default}},
\code{\link{dataquieR.resume_checkpoint}},
\code{\link{dataquieR.resume_print}},
\code{\link{dataquieR.scale_level_heuristics_control_binaryrecodelimit}},
\code{\link{dataquieR.scale_level_heuristics_control_metriclevels}},
\code{\link{dataquieR.study_data_cache_metrics}},
\code{\link{dataquieR.study_data_cache_metrics_env}},
\code{\link{dataquieR.study_data_cache_quick_fill}},
\code{\link{dataquieR.study_data_colnames_case_sensitive}},
\code{\link{dataquieR.testdebug}},
\code{\link{dataquieR.traceback}},
\code{\link{dataquieR.type_adjust_parallel}},
\code{\link{progress_init_fkt}}

Other study_data_cache:
\code{\link{dataquieR.ignore_empty_vars}},
\code{\link{dataquieR.precomputeStudyData}},
\code{\link{dataquieR.print_block_load_factor}},
\code{\link{dataquieR.study_data_cache_metrics}},
\code{\link{dataquieR.study_data_cache_metrics_env}},
\code{\link{dataquieR.study_data_cache_metrics_env_default}},
\code{\link{dataquieR.study_data_cache_quick_fill}}
}
\concept{options}
\concept{study_data_cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_datatype_matrix.R
\name{int_datatype_matrix}
\alias{int_datatype_matrix}
\title{Check declared data types of metadata in study data}
\usage{
int_datatype_matrix(
  resp_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  split_segments = FALSE,
  max_vars_per_plot = 20,
  threshold_value = 0,
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable} the names of the measurement variables, if
missing or \code{NULL}, all variables will be checked}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{split_segments}{\link{logical} return one matrix per study segment}

\item{max_vars_per_plot}{\link{integer} from=0. The maximum number of variables
per single plot.}

\item{threshold_value}{\link{numeric} from=0 to=100. percentage failing
conversions allowed to still classify a
study variable convertible.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
a list with:
\itemize{
\item \code{SummaryTable}: data frame containing data quality check for
"data type mismatch" (\code{CLS_int_vfe_type},
\code{PCT_int_vfe_type}). The following categories are possible:
categories: "Non-matching datatype",
"Non-Matching datatype, convertible",
"Matching datatype"
\item \code{SummaryData}: data frame containing data quality check for
"data type mismatch" for a report
\item \code{SummaryPlot}: \link[ggplot2:ggplot2]{ggplot2::ggplot2} heatmap plot, graphical representation of
\code{SummaryTable}
\item \code{DataTypePlotList}: \link{list} of plots per (maybe artificial) segment
\item \code{ReportSummaryTable}: data frame underlying \code{SummaryPlot}
}
}
\description{
Checks data types of the study data and for the data type
declared in the metadata

\link{Indicator}
}
\details{
This is a preparatory support function that compares study data with
associated metadata. A prerequisite of this function is that the no. of
columns in the study data complies with the no. of rows in the metadata.

For each study variable, the function searches for its data type declared in
static metadata and returns a heatmap like matrix indicating data type
mismatches in the study data.

List function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_add_cause_label_df.R
\name{prep_add_cause_label_df}
\alias{prep_add_cause_label_df}
\title{Convert missing codes in metadata format v1.0 and a missing-cause-table
to v2.0 missing list / jump list assignments}
\usage{
prep_add_cause_label_df(
  item_level = "item_level",
  cause_label_df,
  label_col = VAR_NAMES,
  assume_consistent_codes = TRUE,
  replace_meta_data = ("resp_vars" \%in\% colnames(cause_label_df)),
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{cause_label_df}{\link{data.frame} missing code table. If missing codes have
labels the respective data frame can be
specified here, see \link{cause_label_df}}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{assume_consistent_codes}{\link{logical} if TRUE and no labels are given and
the same missing/jump code is used
for more than one variable, the
labels assigned for this code will
be the same for all variables.}

\item{replace_meta_data}{\link{logical} if \code{TRUE}, ignore existing missing codes
and jump codes and replace them with
data from the \code{cause_label_df}. Otherwise,
copy the labels from \code{cause_label_df} to
the existing code columns.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
\link{data.frame} updated metadata including all the code labels in
missing/jump lists
}
\description{
The function has to working modes. If \code{replace_meta_data} is \code{TRUE}, by
default, if \code{cause_label_df} contains a column
named \code{resp_vars}, then the missing/jump codes in
\code{meta_data[, c(MISSING_CODES, JUMP_CODES)]} will be overwritten, otherwise,
it will be labeled using the \code{cause_label_df}.
}
\details{
If a column \code{resp_vars} exists, then rows with a value in \code{resp_vars} will
only be used for the corresponding variable.
}
\seealso{
\link{prep_extract_cause_label_df}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_study2meta.R
\name{prep_study2meta}
\alias{prep_study2meta}
\title{Guess a metadata data frame from study data.}
\usage{
prep_study2meta(
  study_data,
  level = c(VARATT_REQUIRE_LEVELS$REQUIRED, VARATT_REQUIRE_LEVELS$RECOMMENDED),
  cumulative = TRUE,
  convert_factors = FALSE,
  guess_missing_codes = getOption("dataquieR.guess_missing_codes",
    dataquieR.guess_missing_codes_default),
  guess_character = getOption("dataquieR.guess_character", default =
    dataquieR.guess_character_default)
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{level}{\link{enum} levels to provide (see also \link{VARATT_REQUIRE_LEVELS})}

\item{cumulative}{\link{logical} include attributes of all levels up to level}

\item{convert_factors}{\link{logical} convert factor columns to coded integers.
if selected, then also the study data will
be updated and returned.}

\item{guess_missing_codes}{\link{logical} try to guess missing codes from the data}

\item{guess_character}{\link{logical} guess a data type for character columns
based on the values}
}
\value{
a meta_data data frame or a list with study data and metadata, if
\code{convert_factors == TRUE}.
}
\description{
Guess a minimum metadata data frame from study data. Minimum required variable
attributes are:
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{dataquieR:::util_get_var_att_names_of_level(VARATT_REQUIRE_LEVELS$REQUIRED)
#>            VAR_NAMES            DATA_TYPE   MISSING_LIST_TABLE 
#>          "VAR_NAMES"          "DATA_TYPE" "MISSING_LIST_TABLE"
}\if{html}{\out{</div>}}

The function also tries to detect missing codes.
}
\examples{
\dontrun{
dataquieR::prep_study2meta(Orange, convert_factors = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_get_user_name.R
\name{prep_get_user_name}
\alias{prep_get_user_name}
\title{Return the logged-in User's Full Name}
\usage{
prep_get_user_name()
}
\value{
\link{character} the user's name
}
\description{
If \code{whoami} is not installed, the user name from
\code{Sys.info()} is returned.
}
\details{
Can be overridden by options or environment:

\code{options(FULLNAME = "Stephan Struckmann")}

\code{Sys.setenv(FULLNAME = "Stephan Struckmann")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_contradiction_rules.R
\name{contradiction_functions}
\alias{contradiction_functions}
\title{contradiction_functions}
\usage{
contradiction_functions
}
\description{
Detect abnormalities help functions
}
\details{
2 variables:
\itemize{
\item \code{A_not_equal_B}, if \code{A != B}
\item \code{A_greater_equal_B}, if \code{A >= B}
\item \code{A_greater_than_B}, if \code{A > B}
\item \code{A_less_than_B}, if \code{A < B}
\item \code{A_less_equal_B}, if \code{A <= B}
\item \code{A_present_not_B}, if \code{A & is.na(B)}
\item \code{A_present_and_B}, if \code{A & !(is.na(B))}
\item \code{A_present_and_B_levels}, if \verb{A & B  \%in\% \{set of levels\}}
\item \code{A_levels_and_B_gt_value}, if \verb{A \%in\% \{set of levels\} & B >  value}
\item \code{A_levels_and_B_lt_value}, if \verb{A \%in\% \{set of levels\} & B <  value}
\item \code{A_levels_and_B_levels}, if
\verb{A \%in\% \{set of levels\} & B \%in\% \{set of levels\}}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.dataquieR_summary.R
\name{print.dataquieR_summary}
\alias{print.dataquieR_summary}
\title{Print a \code{dataquieR} summary}
\usage{
\method{print}{dataquieR_summary}(
  x,
  ...,
  grouped_by = c("call_names", "indicator_metric"),
  dont_print = FALSE,
  folder_of_report = NULL,
  vars_to_include = c("study")
)
}
\arguments{
\item{x}{the \code{dataquieR} summary, see \code{\link[=summary]{summary()}} and \code{\link[=dq_report2]{dq_report2()}}}

\item{...}{not yet used}

\item{grouped_by}{define the columns of the resulting matrix. It can be either
"call_names", one column per function,  or "indicator_metric", one column per indicator
or both c("call_names", "indicator_metric"). The last combination is the default}

\item{dont_print}{suppress the actual printing, just return a printable
object derived from \code{x}}

\item{folder_of_report}{a named vector with the location of variable and
\code{call_names}}

\item{vars_to_include}{\code{"study"}, \code{"ssi"} or \code{c("study", "ssi")}. variables
to include}
}
\value{
invisible html object
}
\description{
Print a \code{dataquieR} summary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{COMPUTATION_RULE}
\alias{COMPUTATION_RULE}
\title{Cross-item level metadata attribute name}
\usage{
COMPUTATION_RULE
}
\description{
Cross-item level metadata attribute name
}
\seealso{
\link{meta_data_computation}
}
\concept{meta_data_computation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_robust_guess_data_type.R
\name{prep_guess_encoding}
\alias{prep_guess_encoding}
\title{Guess encoding of text or text files}
\usage{
prep_guess_encoding(x, file)
}
\arguments{
\item{x}{\link{character} string to guess encoding for}

\item{file}{\link{character} file to guess encoding for}
}
\value{
encoding
}
\description{
Guess encoding of text or text files
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_html_for_dims.R
\name{util_html_for_dims}
\alias{util_html_for_dims}
\title{Create a dynamic dimension related page for the report}
\usage{
util_html_for_dims(
  report,
  use_plot_ly,
  template,
  block_load_factor,
  repsum,
  dir
)
}
\arguments{
\item{report}{\link{dataquieR_resultset2} a \code{dq_report2} report}

\item{use_plot_ly}{\link{logical} use \code{plotly}, if available.}

\item{template}{\link{character} template to use for the \code{dq_report2} report.}

\item{block_load_factor}{\link{numeric} multiply size of parallel compute blocks
by this factor.}

\item{repsum}{the \code{dataquieR} summary, see \code{\link[=summary]{summary()}} and \code{\link[=dq_report2]{dq_report2()}}}

\item{dir}{\link{character} output directory for potential \code{iframes}.}
}
\value{
list of arguments for \code{append_single_page()} defined locally in
\code{util_generate_pages_from_report()}.
}
\description{
Create a dynamic dimension related page for the report
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.list.dataquieR_resultset.R
\name{as.list.dataquieR_resultset}
\alias{as.list.dataquieR_resultset}
\title{Convert a full \code{dataquieR} report to a \code{list}}
\usage{
\method{as.list}{dataquieR_resultset}(x, ...)
}
\arguments{
\item{x}{Deprecated}

\item{...}{Deprecated}
}
\value{
Deprecated
}
\description{
Deprecated
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DISTRIBUTIONS}
\alias{DISTRIBUTIONS}
\title{All available probability distributions for
\link{acc_shape_or_scale}}
\usage{
DISTRIBUTIONS
}
\description{
\itemize{
\item \code{uniform} For uniform distribution
\item \code{normal} For Gaussian distribution
\item \code{gamma} For a gamma distribution
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_int_duplicate_content_segment.R
\name{util_int_duplicate_content_segment}
\alias{util_int_duplicate_content_segment}
\title{Check for duplicated content}
\usage{
util_int_duplicate_content_segment(
  level = c("segment"),
  identifier_name_list,
  id_vars_list,
  unique_rows,
  study_data,
  meta_data,
  meta_data_segment = "segment_level",
  segment_level
)
}
\arguments{
\item{level}{\link{character} a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").}

\item{identifier_name_list}{\link{vector} the vector that contains the name of the identifier to be used in the assessment. For the study level, corresponds to the names of the different data frames. For the segment level, indicates the name of the segments.}

\item{id_vars_list}{\link{list} the list containing the identifier variables names to be used in the assessment.}

\item{unique_rows}{\link{vector} named. for each segment, either true/false or \code{no_id} to exclude ID variables from check}

\item{study_data}{\link{data.frame} the data frame that contains the measurements, mandatory.}

\item{meta_data}{\link{data.frame} the data frame that contains metadata attributes of the study data, mandatory.}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}
}
\value{
a \link{list} with
\itemize{
\item \code{SegmentData}: data frame with the results of the quality check for duplicated entries
\item \code{SegmentTable}: data frame with selected duplicated entries check results, used for the data quality report.
\item \code{Other}: vector with row indices of duplicated entries, if any, otherwise NULL.
}
}
\description{
This function tests for duplicates entries in the data set. It is possible to
check duplicated entries by study segments or to consider only selected
segments.
}
\seealso{
Other integrity_indicator_functions:
\code{\link[=util_int_duplicate_content_dataframe]{util_int_duplicate_content_dataframe()}},
\code{\link[=util_int_duplicate_ids_dataframe]{util_int_duplicate_ids_dataframe()}},
\code{\link[=util_int_duplicate_ids_segment]{util_int_duplicate_ids_segment()}},
\code{\link[=util_int_unexp_records_set_dataframe]{util_int_unexp_records_set_dataframe()}},
\code{\link[=util_int_unexp_records_set_segment]{util_int_unexp_records_set_segment()}}
}
\concept{integrity_indicator}
\concept{integrity_indicator_functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.dataquieR_resultset.R
\name{print.dataquieR_resultset}
\alias{print.dataquieR_resultset}
\title{Generate a RMarkdown-based report from a \link{dataquieR} report}
\usage{
\method{print}{dataquieR_resultset}(...)
}
\arguments{
\item{...}{deprecated}
}
\value{
deprecated
}
\description{
Generate a RMarkdown-based report from a \link{dataquieR} report
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{SEGMENT_ID_REF_TABLE}
\alias{SEGMENT_ID_REF_TABLE}
\title{Segment level metadata attribute name}
\usage{
SEGMENT_ID_REF_TABLE
}
\description{
The name of the data frame containing the reference IDs to be compared
with the IDs in the targeted segment.
}
\seealso{
\link{meta_data_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_dq_data_type_of.R
\name{prep_dq_data_type_of}
\alias{prep_dq_data_type_of}
\title{Get the dataquieR \code{DATA_TYPE} of \code{x}}
\usage{
prep_dq_data_type_of(
  x,
  guess_character = getOption("dataquieR.guess_character", default =
    dataquieR.guess_character_default)
)
}
\arguments{
\item{x}{object to define the dataquieR data type of}

\item{guess_character}{\link{logical} guess a data type for character columns
based on the values}
}
\value{
the dataquieR data type as listed in \code{DATA_TYPES}
}
\description{
Get the dataquieR \code{DATA_TYPE} of \code{x}
}
\seealso{
\code{\link{DATA_TYPES_OF_R_TYPE}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_all_datastructure_segment.R
\name{int_all_datastructure_segment}
\alias{int_all_datastructure_segment}
\title{Wrapper function to check for segment data structure}
\usage{
int_all_datastructure_segment(
  study_data,
  label_col,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2,
  segment_level,
  meta_data_segment = "segment_level"
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}

\item{meta_data_segment}{\link{data.frame} the data frame that contains the metadata for the segment level, mandatory}
}
\value{
a \link{list} with
\itemize{
\item \code{SegmentTable}: data frame with selected check results, used for the data quality report.
}
}
\description{
This function tests for unexpected elements and records, as well as duplicated identifiers and content.
The unexpected element record check can be conducted by providing the number of expected records or
an additional table with the expected records.
It is possible to conduct the checks by study segments or to consider only selected
segments.

\link{Indicator}
}
\examples{
\dontrun{
out_segment <- int_all_datastructure_segment(
  meta_data_segment = "meta_data_segment",
  study_data = "ship",
  meta_data = "ship_meta"
)

study_data <- cars
meta_data <- dataquieR::prep_create_meta(VAR_NAMES = c("speedx", "distx"),
  DATA_TYPE = c("integer", "integer"), MISSING_LIST = "|", JUMP_LIST = "|",
  STUDY_SEGMENT = c("Intro", "Ex"))

out_segment <- int_all_datastructure_segment(
  meta_data_segment = "meta_data_segment",
  study_data = study_data,
  meta_data = meta_data
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR-package.R
\docType{package}
\name{dataquieR}
\alias{dataquieR}
\alias{dataquieR-package}
\alias{dataquieR-options}
\title{The \code{dataquieR} package about Data Quality in
Epidemiological Research}
\description{
For a quick start please read \link{dq_report2} and maybe the vignettes
or the package's \href{https://dataquality.qihs.uni-greifswald.de/}{website}.
}
\section{Options}{
This package features the following \code{options()}:
\itemize{
\item \link{dataquieR.CONDITIONS_LEVEL_TRHESHOLD}
\item \link{dataquieR.CONDITIONS_WITH_STACKTRACE}
\item \link{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE}
\item \link{dataquieR.ERRORS_WITH_CALLER}
\item \link{dataquieR.GAM_for_LOESS}
\item \link{dataquieR.MAHALANOBIS_THRESHOLD}
\item \link{dataquieR.MAX_LABEL_LEN}
\item \link{dataquieR.MAX_LONG_LABEL_LEN}
\item \link{dataquieR.MAX_VALUE_LABEL_LEN}
\item \link{dataquieR.MESSAGES_WITH_CALLER}
\item \link{dataquieR.MULTIVARIATE_OUTLIER_CHECK}
\item \link{dataquieR.VALUE_LABELS_htmlescaped}
\item \link{dataquieR.WARNINGS_WITH_CALLER}
\item \link{dataquieR.acc_loess.exclude_constant_subgroups}
\item \link{dataquieR.acc_loess.mark_time_points}
\item \link{dataquieR.acc_loess.min_bw}
\item \link{dataquieR.acc_loess.min_obs_in_subgroup}
\item \link{dataquieR.acc_loess.min_proportion}
\item \link{dataquieR.acc_loess.plot_format}
\item \link{dataquieR.acc_loess.plot_observations}
\item \link{dataquieR.acc_margins_num}
\item \link{dataquieR.acc_margins_sort}
\item \link{dataquieR.acc_multivariate_outlier.scale}
\item \link{dataquieR.acc_shape_or_scale_ci}
\item \link{dataquieR.col_con_con_empirical}
\item \link{dataquieR.col_con_con_logical}
\item \link{dataquieR.convert_to_list_for_lapply}
\item \link{dataquieR.debug}
\item \link{dataquieR.des_summary_hard_lim_remove}
\item \link{dataquieR.dontwrapresults}
\item \link{dataquieR.droplevels_ReportSummaryTable}
\item \link{dataquieR.dt_adjust}
\item \link{dataquieR.fix_column_type_on_read}
\item \link{dataquieR.flip_mode}
\item \link{dataquieR.force_item_specific_missing_codes}
\item \link{dataquieR.force_label_col}
\item \link{dataquieR.grading_formats}
\item \link{dataquieR.grading_rulesets}
\item \link{dataquieR.guess_character}
\item \link{dataquieR.guess_missing_codes}
\item \link{dataquieR.ignore_empty_vars}
\item \link{dataquieR.lang}
\item \link{dataquieR.lazy_plots}
\item \link{dataquieR.lazy_plots_cache}
\item \link{dataquieR.lazy_plots_gg_compatibility}
\item \link{dataquieR.locale}
\item \link{dataquieR.max_cat_resp_var_levels_in_plot}
\item \link{dataquieR.max_group_var_levels_in_plot}
\item \link{dataquieR.max_group_var_levels_with_violins}
\item \link{dataquieR.min_obs_per_group_var_in_plot}
\item \link{dataquieR.min_time_points_for_cat_resp_var}
\item \link{dataquieR.no_geom_count_in_bin}
\item \link{dataquieR.no_overall_in_bin}
\item \link{dataquieR.non_disclosure}
\item \link{dataquieR.old_factor_handling}
\item \link{dataquieR.old_type_adjust}
\item \link{dataquieR.precomputeStudyData}
\item \link{dataquieR.print_block_load_factor}
\item \link{dataquieR.resume_checkpoint}
\item \link{dataquieR.resume_print}
\item \link{dataquieR.scale_level_heuristics_control_binaryrecodelimit}
\item \link{dataquieR.scale_level_heuristics_control_metriclevels}
\item \link{dataquieR.study_data_cache_max}
\item \link{dataquieR.study_data_cache_metrics}
\item \link{dataquieR.study_data_cache_metrics_env}
\item \link{dataquieR.study_data_cache_quick_fill}
\item \link{dataquieR.study_data_colnames_case_sensitive}
\item \link{dataquieR.testdebug}
\item \link{dataquieR.traceback}
\item \link{dataquieR.type_adjust_parallel}
}
}

\references{
\itemize{
\item \href{https://joss.theoj.org/papers/10.21105/joss.06581}{JOSS 2024}
\item \href{https://joss.theoj.org/papers/10.21105/joss.03093}{JOSS 2021}
\item \href{https://www.researchsquare.com/article/rs-119457/v1}{DQ_OBS data quality framework}
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://dataquality.qihs.uni-greifswald.de/}
  \item Report bugs at \url{https://gitlab.com/libreumg/dataquier/-/work_items/issues}
}


Other options:
\code{\link{dataquieR.CONDITIONS_LEVEL_TRHESHOLD}},
\code{\link{dataquieR.CONDITIONS_WITH_STACKTRACE}},
\code{\link{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE}},
\code{\link{dataquieR.ERRORS_WITH_CALLER}},
\code{\link{dataquieR.GAM_for_LOESS}},
\code{\link{dataquieR.MAHALANOBIS_THRESHOLD}},
\code{\link{dataquieR.MAX_LABEL_LEN}},
\code{\link{dataquieR.MAX_LONG_LABEL_LEN}},
\code{\link{dataquieR.MAX_VALUE_LABEL_LEN}},
\code{\link{dataquieR.MESSAGES_WITH_CALLER}},
\code{\link{dataquieR.MULTIVARIATE_OUTLIER_CHECK}},
\code{\link{dataquieR.VALUE_LABELS_htmlescaped}},
\code{\link{dataquieR.WARNINGS_WITH_CALLER}},
\code{\link{dataquieR.acc_loess.exclude_constant_subgroups}},
\code{\link{dataquieR.acc_loess.mark_time_points}},
\code{\link{dataquieR.acc_loess.min_bw}},
\code{\link{dataquieR.acc_loess.min_obs_in_subgroup}},
\code{\link{dataquieR.acc_loess.min_proportion}},
\code{\link{dataquieR.acc_loess.plot_format}},
\code{\link{dataquieR.acc_loess.plot_observations}},
\code{\link{dataquieR.acc_margins_num}},
\code{\link{dataquieR.acc_margins_sort}},
\code{\link{dataquieR.acc_multivariate_outlier.scale}},
\code{\link{dataquieR.acc_shape_or_scale_ci}},
\code{\link{dataquieR.col_con_con_empirical}},
\code{\link{dataquieR.col_con_con_logical}},
\code{\link{dataquieR.convert_to_list_for_lapply}},
\code{\link{dataquieR.debug}},
\code{\link{dataquieR.des_summary_hard_lim_remove}},
\code{\link{dataquieR.dontwrapresults}},
\code{\link{dataquieR.droplevels_ReportSummaryTable}},
\code{\link{dataquieR.dt_adjust}},
\code{\link{dataquieR.fix_column_type_on_read}},
\code{\link{dataquieR.flip_mode}},
\code{\link{dataquieR.force_item_specific_missing_codes}},
\code{\link{dataquieR.force_label_col}},
\code{\link{dataquieR.grading_formats}},
\code{\link{dataquieR.grading_rulesets}},
\code{\link{dataquieR.guess_character}},
\code{\link{dataquieR.guess_missing_codes}},
\code{\link{dataquieR.ignore_empty_vars}},
\code{\link{dataquieR.lang}},
\code{\link{dataquieR.lazy_plots}},
\code{\link{dataquieR.lazy_plots_cache}},
\code{\link{dataquieR.lazy_plots_gg_compatibility}},
\code{\link{dataquieR.locale}},
\code{\link{dataquieR.max_cat_resp_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_with_violins}},
\code{\link{dataquieR.min_obs_per_group_var_in_plot}},
\code{\link{dataquieR.min_time_points_for_cat_resp_var}},
\code{\link{dataquieR.no_geom_count_in_bin}},
\code{\link{dataquieR.no_overall_in_bin}},
\code{\link{dataquieR.non_disclosure}},
\code{\link{dataquieR.old_factor_handling}},
\code{\link{dataquieR.old_type_adjust}},
\code{\link{dataquieR.precomputeStudyData}},
\code{\link{dataquieR.print_block_load_factor}},
\code{\link{dataquieR.progress_fkt_default}},
\code{\link{dataquieR.progress_msg_fkt_default}},
\code{\link{dataquieR.resume_checkpoint}},
\code{\link{dataquieR.resume_print}},
\code{\link{dataquieR.scale_level_heuristics_control_binaryrecodelimit}},
\code{\link{dataquieR.scale_level_heuristics_control_metriclevels}},
\code{\link{dataquieR.study_data_cache_max}},
\code{\link{dataquieR.study_data_cache_metrics}},
\code{\link{dataquieR.study_data_cache_metrics_env}},
\code{\link{dataquieR.study_data_cache_quick_fill}},
\code{\link{dataquieR.study_data_colnames_case_sensitive}},
\code{\link{dataquieR.testdebug}},
\code{\link{dataquieR.traceback}},
\code{\link{dataquieR.type_adjust_parallel}},
\code{\link{progress_init_fkt}}
}
\author{
\strong{Maintainer}: Stephan Struckmann \email{stephan.struckmann@uni-greifswald.de} (\href{https://orcid.org/0000-0002-8565-7962}{ORCID})

Authors:
\itemize{
  \item Stephan Struckmann \email{stephan.struckmann@uni-greifswald.de} (\href{https://orcid.org/0000-0002-8565-7962}{ORCID})
  \item Elisa Kasbohm \email{elisa.kasbohm@uni-greifswald.de} (\href{https://orcid.org/0000-0001-5261-538X}{ORCID})
  \item Elena Salogni \email{elena.salogni@uni-greifswald.de} (\href{https://orcid.org/0009-0007-3767-7145}{ORCID})
  \item Joany Marino \email{joany.marino@uni-greifswald.de} (\href{https://orcid.org/0000-0002-4657-3758}{ORCID})
  \item Adrian Richter \email{richtera@uni-greifswald.de} (\href{https://orcid.org/0000-0002-3372-2021}{ORCID})
  \item Carsten Oliver Schmidt \email{carsten.schmidt@uni-greifswald.de} (\href{https://orcid.org/0000-0001-5266-9396}{ORCID})
}

Other contributors:
\itemize{
  \item University Medicine Greifswald [copyright holder]
  \item German Research Foundation (DFG SCHM 2744/3-1, SCHM 2744/9-1, SCHM 2744/3-4) [funder]
  \item National Research Data Infrastructure for Personal Health Data: (NFDI 13/1) [funder]
  \item European Union’s Horizon 2020 programme (euCanSHare, grant agreement No. 825903) [funder]
}

}
\concept{options}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_robust_guess_data_type.R
\name{prep_robust_guess_data_type}
\alias{prep_robust_guess_data_type}
\title{Guess the data type of a vector}
\usage{
prep_robust_guess_data_type(x, k = 50, it = 200)
}
\arguments{
\item{x}{a vector with characters}

\item{k}{\link{numeric} sample size, if less than \verb{floor(length(x) / (it/20)))},
minimum sample size is 1.}

\item{it}{\link{integer} number of iterations when taking samples}
}
\value{
a guess of the data type of \code{x}. An attribute \code{orig_type} is also
attached to give the more detailed guess returned by \code{\link[readr:guess_parser]{readr::guess_parser()}}.
}
\description{
Guess the data type of a vector
}
\section{Algorithm}{
This function takes \code{x} and tries to guess the data type of random subsets of
this vector using \code{\link[readr:guess_parser]{readr::guess_parser()}}. The RNG is initialized with a
constant, so the function stays deterministic. It does such sub-sample based
checks \code{it} times, the majority of the detected datatype determines the
guessed data type.
}

\seealso{
\code{\link[=prep_datatype_from_data]{prep_datatype_from_data()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_clean_labels.R
\name{prep_clean_labels}
\alias{prep_clean_labels}
\title{Support function to scan variable labels for applicability}
\usage{
prep_clean_labels(
  label_col,
  item_level = "item_level",
  no_dups = FALSE,
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{label_col}{\link{character} label attribute to adjust or character vector to
adjust, depending on \code{meta_data} argument is
given or missing.}

\item{item_level}{\link{data.frame} metadata data frame: If \code{label_col} is a label
attribute to adjust, this is the metadata
table to process on. If missing, \code{label_col}
must be a character vector with values to
adjust.}

\item{no_dups}{\link{logical} disallow duplicates in input or output vectors of
the function, then, prep_clean_labels would call
\code{stop()} on duplicated labels.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
a data.frame with:
\itemize{
\item if \code{meta_data} is set, a list with:
\itemize{
\item modified \code{meta_data[, label_col]} column
}
\item if \code{meta_data} is not set, adjusted labels that then were directly given
in label_col
}
}
\description{
Adjust labels in meta_data to be valid variable names in formulas for
diverse r functions, such as \code{\link{glm}} or \code{\link[lme4:lmer]{lme4::lmer}}.
}
\details{
Hint: The following is still true, but the functions should be capable
of doing potentially needed fixes on-the-fly automatically, so likely
you will not need this function any more.

Currently, labels as given by \code{label_col} arguments in the most functions
are directly used in formula, so that they become natural part of the
outputs, but different models expect differently strict syntax for such
formulas, especially for valid variable names. \code{prep_clean_labels} removes
all potentially inadmissible characters from variable names (no guarantee,
that some exotic model still rejects the names, but minimizing the number
of exotic characters). However, variable names are modified, may become
unreadable or indistinguishable from other variable names. For the latter
case, a \code{stop} call is possible, controlled by the \code{no_dups} argument.

A warning is emitted, if modifications were necessary.
}
\examples{
\dontrun{
meta_data1 <- data.frame(
  LABEL =
    c(
      "syst. Blood pressure (mmHg) 1",
      "1st heart frequency in MHz",
      "body surface (\\\\u33A1)"
    )
)
print(meta_data1)
print(prep_clean_labels(meta_data1$LABEL))
meta_data1 <- prep_clean_labels("LABEL", meta_data1)
print(meta_data1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_pmap.R
\name{prep_pmap}
\alias{prep_pmap}
\title{Support function for a parallel \code{pmap}}
\usage{
prep_pmap(.l, .f, ..., cores = 0)
}
\arguments{
\item{.l}{\link{data.frame} with one call per line and one function argument
per column}

\item{.f}{\code{\link{function}} to call with the arguments from \code{.l}}

\item{...}{additional, static arguments for calling \code{.f}}

\item{cores}{number of cpu cores to use or a (named) list with arguments for
\link[parallelMap:parallelStart]{parallelMap::parallelStart} or NULL, if parallel has already
been started by the caller. Set to 0 to run without
parallelization.}
}
\value{
\link{list} of results of the function calls
}
\description{
parallel version of \code{purrr::pmap}
}
\seealso{
\code{purrr::pmap}

\href{https://stackoverflow.com/a/47575143}{Stack Overflow post}
}
\author{
\href{https://stackoverflow.com/users/6197649}{Aurèle}

S Struckmann
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_translate.R
\name{prep_is_translated}
\alias{prep_is_translated}
\title{Detect if an object is a \code{dataquieR_translated} object}
\usage{
prep_is_translated(x)
}
\arguments{
\item{x}{the object to test}
}
\value{
\code{TRUE}, if \code{x} is a \code{dataquieR_translated} object.
}
\description{
Detect if an object is a \code{dataquieR_translated} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{dims}
\alias{dims}
\title{Dimension Titles for Prefixes}
\usage{
dims
}
\description{
order does matter, because it defines the order in the \code{dq_report2}.
}
\seealso{
\code{util_html_for_var()}

\code{util_html_for_dims()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nres.R
\name{nres}
\alias{nres}
\title{return the number of result slots in a report}
\usage{
nres(x)
}
\arguments{
\item{x}{the \code{dataquieR} report (v2.0)}
}
\value{
the number of used result slots
}
\description{
return the number of result slots in a report
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_datatype_from_data.R
\name{prep_datatype_from_data}
\alias{prep_datatype_from_data}
\title{Get data types from data}
\usage{
prep_datatype_from_data(
  resp_vars = colnames(study_data),
  study_data,
  .dont_cast_off_cols = FALSE,
  guess_character = getOption("dataquieR.guess_character", default =
    dataquieR.guess_character_default)
)
}
\arguments{
\item{resp_vars}{\link{variable} names of the variables to fetch the data type
from the data}

\item{study_data}{\link{data.frame} the data frame that contains the measurements
Hint: Only data frames supported, no URL
or file names.}

\item{.dont_cast_off_cols}{\link{logical} internal use, only}

\item{guess_character}{\link{logical} guess a data type for character columns
based on the values}
}
\value{
vector of data types
}
\description{
Get data types from data
}
\examples{
\dontrun{
dataquieR::prep_datatype_from_data(cars)
}
}
\seealso{
\code{\link[=prep_robust_guess_data_type]{prep_robust_guess_data_type()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset2.R
\name{[[.dataquieR_resultset2}
\alias{[[.dataquieR_resultset2}
\title{Get a single result from a \verb{dataquieR 2} report}
\usage{
\method{[[}{dataquieR_resultset2}(x, el)
}
\arguments{
\item{x}{the report}

\item{el}{the index}
}
\value{
the \code{dataquieR} result object
}
\description{
Get a single result from a \verb{dataquieR 2} report
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_map_labels.R
\name{prep_map_labels}
\alias{prep_map_labels}
\title{Support function to allocate labels to variables}
\usage{
prep_map_labels(
  x,
  item_level = "item_level",
  to = LABEL,
  from = VAR_NAMES,
  ifnotfound,
  warn_ambiguous = FALSE,
  meta_data_v2,
  meta_data = item_level
)
}
\arguments{
\item{x}{\link{character} variable names, character vector, see parameter from}

\item{item_level}{\link{data.frame} metadata data frame, if, as a \code{dataquieR}
developer, you do not have
\strong{item-level-metadata}, you should use
\code{util_map_labels()} instead to avoid consistency
checks on for item-level \code{meta_data}.}

\item{to}{\link{character} variable attribute to map to}

\item{from}{\link{character} variable identifier to map from}

\item{ifnotfound}{\link{list} A list of values to be used if the item is not
found: it will be coerced to a list if necessary.}

\item{warn_ambiguous}{\link{logical} print a warning if mapping variables from
\code{from} to \code{to} produces ambiguous identifiers.}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}
}
\value{
a character vector with:
\itemize{
\item mapped values
}
}
\description{
Map variables to certain attributes, e.g. by default their labels.
}
\details{
This function basically calls \code{colnames(study_data) <- meta_data$LABEL},
ensuring correct merging/joining of study data columns to the corresponding
metadata rows, even if the orders differ. If a variable/study_data-column
name is not found in \code{meta_data[[from]]} (default \code{from = VAR_NAMES}),
either stop is called or, if \code{ifnotfound} has been assigned a value, that
value is returned. See \code{\link{mget}}, which is internally used by this function.

The function not only maps to the \code{LABEL} column, but \code{to} can be any
metadata variable attribute, so the function can also be used, to get, e.g.
all \code{HARD_LIMITS} from the metadata.
}
\examples{
\dontrun{
meta_data <- prep_create_meta(
  VAR_NAMES = c("ID", "SEX", "AGE", "DOE"),
  LABEL = c("Pseudo-ID", "Gender", "Age", "Examination Date"),
  DATA_TYPE = c(DATA_TYPES$INTEGER, DATA_TYPES$INTEGER, DATA_TYPES$INTEGER,
                 DATA_TYPES$DATETIME),
  MISSING_LIST = ""
)
stopifnot(all(prep_map_labels(c("AGE", "DOE"), meta_data) == c("Age",
                                                 "Examination Date")))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/menu_env.R
\name{menu_env_drop_down}
\alias{menu_env_drop_down}
\title{Creates a drop-down menu}
\arguments{
\item{title}{name of the entry in the main menu}

\item{menu_description}{description, displayed, if the main
menu entry itself is clicked}

\item{...}{the sub-menu-entries}

\item{id}{id for the entry, defaults to modified title}
}
\value{
html div object
}
\description{
Creates a drop-down menu
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_int_duplicate_ids_dataframe.R
\name{util_int_duplicate_ids_dataframe}
\alias{util_int_duplicate_ids_dataframe}
\title{Check for duplicated IDs}
\usage{
util_int_duplicate_ids_dataframe(
  level = c("dataframe"),
  id_vars_list,
  identifier_name_list,
  repetitions,
  meta_data_dataframe = "dataframe_level",
  ...,
  dataframe_level
)
}
\arguments{
\item{level}{\link{character} a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").}

\item{id_vars_list}{\link{list} id variable names for each segment or data frame}

\item{identifier_name_list}{\link{vector} the segments or data frame names being assessed}

\item{repetitions}{\link{vector} an integer vector indicating the number of allowed repetitions in the \code{id_vars}.}

\item{meta_data_dataframe}{\link{data.frame} the data frame that contains the
metadata for the data frame level}

\item{...}{not used.}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}
}
\value{
a \link{list} with
\itemize{
\item \code{DataframeData}: data frame with the results of the quality check for duplicated identifiers
\item \code{DataframeTable}: data frame with selected duplicated identifiers check results, used for the data quality report.
\item \code{Other}: named list with inner lists of unique cases containing each the
row indices of duplicated identifiers separated by "|" , if any.
outer names are names of the data frames
}
}
\description{
This function tests for duplicates entries in identifiers. It is possible to
check duplicated identifiers by study segments or to consider only selected
segments.
}
\seealso{
Other integrity_indicator_functions:
\code{\link[=util_int_duplicate_content_dataframe]{util_int_duplicate_content_dataframe()}},
\code{\link[=util_int_duplicate_content_segment]{util_int_duplicate_content_segment()}},
\code{\link[=util_int_duplicate_ids_segment]{util_int_duplicate_ids_segment()}},
\code{\link[=util_int_unexp_records_set_dataframe]{util_int_unexp_records_set_dataframe()}},
\code{\link[=util_int_unexp_records_set_segment]{util_int_unexp_records_set_segment()}}
}
\concept{integrity_indicator}
\concept{integrity_indicator_functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_filter_repsum.R
\name{util_filter_repsum}
\alias{util_filter_repsum}
\title{Delete rows from summary table for \code{SSI} or non-\code{SSI} variables}
\usage{
util_filter_repsum(
  repsumtab,
  vars_to_include,
  meta_data,
  rownames_of_report,
  label_col
)
}
\arguments{
\item{repsumtab}{\link{data.frame} the report summary table}

\item{vars_to_include}{\code{"study"}, \code{"ssi"} or \code{c("study", "ssi")}. variables to include}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}
}
\value{
\link{data.frame} the filtered \code{repsumtab} with attribute
\code{rownames_of_report}, also filtered
}
\description{
Delete rows from summary table for \code{SSI} or non-\code{SSI} variables
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_options.R
\docType{data}
\name{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE}
\alias{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE}
\title{Metadata describes more than the current study data}
\description{
\itemize{
\item \code{none}: no check will be provided about the match of variables and
records available in the study data and described in the metadata
\item \code{exact}: There must be a 1:1 match between the study data and metadata
regarding data frames and segments variables and records
\item \code{subset_u}: study data are a subset of metadata. All variables from the study
data are expected to be present in the metadata, but one or
more variables in the metadata are not expected to be
present in the study data.
In this case a variable present in
the study data but not in the metadata would produce an issue.
\item \code{subset_m}: metadata are a subset of study data. All variables in the metadata
are expected to be present in the study data, but one or more
variables in the study data are not expected to be
present in the metadata.
}
}
\seealso{
Other options:
\code{\link{dataquieR}},
\code{\link{dataquieR.CONDITIONS_LEVEL_TRHESHOLD}},
\code{\link{dataquieR.CONDITIONS_WITH_STACKTRACE}},
\code{\link{dataquieR.ERRORS_WITH_CALLER}},
\code{\link{dataquieR.GAM_for_LOESS}},
\code{\link{dataquieR.MAHALANOBIS_THRESHOLD}},
\code{\link{dataquieR.MAX_LABEL_LEN}},
\code{\link{dataquieR.MAX_LONG_LABEL_LEN}},
\code{\link{dataquieR.MAX_VALUE_LABEL_LEN}},
\code{\link{dataquieR.MESSAGES_WITH_CALLER}},
\code{\link{dataquieR.MULTIVARIATE_OUTLIER_CHECK}},
\code{\link{dataquieR.VALUE_LABELS_htmlescaped}},
\code{\link{dataquieR.WARNINGS_WITH_CALLER}},
\code{\link{dataquieR.acc_loess.exclude_constant_subgroups}},
\code{\link{dataquieR.acc_loess.mark_time_points}},
\code{\link{dataquieR.acc_loess.min_bw}},
\code{\link{dataquieR.acc_loess.min_obs_in_subgroup}},
\code{\link{dataquieR.acc_loess.min_proportion}},
\code{\link{dataquieR.acc_loess.plot_format}},
\code{\link{dataquieR.acc_loess.plot_observations}},
\code{\link{dataquieR.acc_margins_num}},
\code{\link{dataquieR.acc_margins_sort}},
\code{\link{dataquieR.acc_multivariate_outlier.scale}},
\code{\link{dataquieR.acc_shape_or_scale_ci}},
\code{\link{dataquieR.col_con_con_empirical}},
\code{\link{dataquieR.col_con_con_logical}},
\code{\link{dataquieR.convert_to_list_for_lapply}},
\code{\link{dataquieR.debug}},
\code{\link{dataquieR.des_summary_hard_lim_remove}},
\code{\link{dataquieR.dontwrapresults}},
\code{\link{dataquieR.droplevels_ReportSummaryTable}},
\code{\link{dataquieR.dt_adjust}},
\code{\link{dataquieR.fix_column_type_on_read}},
\code{\link{dataquieR.flip_mode}},
\code{\link{dataquieR.force_item_specific_missing_codes}},
\code{\link{dataquieR.force_label_col}},
\code{\link{dataquieR.grading_formats}},
\code{\link{dataquieR.grading_rulesets}},
\code{\link{dataquieR.guess_character}},
\code{\link{dataquieR.guess_missing_codes}},
\code{\link{dataquieR.ignore_empty_vars}},
\code{\link{dataquieR.lang}},
\code{\link{dataquieR.lazy_plots}},
\code{\link{dataquieR.lazy_plots_cache}},
\code{\link{dataquieR.lazy_plots_gg_compatibility}},
\code{\link{dataquieR.locale}},
\code{\link{dataquieR.max_cat_resp_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_with_violins}},
\code{\link{dataquieR.min_obs_per_group_var_in_plot}},
\code{\link{dataquieR.min_time_points_for_cat_resp_var}},
\code{\link{dataquieR.no_geom_count_in_bin}},
\code{\link{dataquieR.no_overall_in_bin}},
\code{\link{dataquieR.non_disclosure}},
\code{\link{dataquieR.old_factor_handling}},
\code{\link{dataquieR.old_type_adjust}},
\code{\link{dataquieR.precomputeStudyData}},
\code{\link{dataquieR.print_block_load_factor}},
\code{\link{dataquieR.progress_fkt_default}},
\code{\link{dataquieR.progress_msg_fkt_default}},
\code{\link{dataquieR.resume_checkpoint}},
\code{\link{dataquieR.resume_print}},
\code{\link{dataquieR.scale_level_heuristics_control_binaryrecodelimit}},
\code{\link{dataquieR.scale_level_heuristics_control_metriclevels}},
\code{\link{dataquieR.study_data_cache_max}},
\code{\link{dataquieR.study_data_cache_metrics}},
\code{\link{dataquieR.study_data_cache_metrics_env}},
\code{\link{dataquieR.study_data_cache_quick_fill}},
\code{\link{dataquieR.study_data_colnames_case_sensitive}},
\code{\link{dataquieR.testdebug}},
\code{\link{dataquieR.traceback}},
\code{\link{dataquieR.type_adjust_parallel}},
\code{\link{progress_init_fkt}}
}
\concept{options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_duplicate_content.R
\name{int_duplicate_content}
\alias{int_duplicate_content}
\title{Check for duplicated content}
\usage{
int_duplicate_content(
  level = c("dataframe", "segment"),
  study_data,
  item_level = "item_level",
  label_col,
  meta_data = item_level,
  meta_data_v2,
  ...
)
}
\arguments{
\item{level}{\link{character} a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{...}{Depending on \code{level}, passed to either
\code{util_int_duplicate_content_segment} or
\code{util_int_duplicate_content_dataframe}}
}
\value{
a \link{list}. Depending on \code{level}, see
\code{util_int_duplicate_content_segment} or
\code{util_int_duplicate_content_dataframe} for a description of the outputs.
}
\description{
This function tests for duplicates entries in the data set. It is possible to
check duplicated entries by study segments or to consider only selected
segments.

\link{Indicator}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_extract_summary.dataquieR_resultset2.R
\name{prep_extract_summary.dataquieR_resultset2}
\alias{prep_extract_summary.dataquieR_resultset2}
\title{Extract report summary from reports}
\usage{
\method{prep_extract_summary}{dataquieR_resultset2}(r, ...)
}
\arguments{
\item{r}{\link{dq_report2} a \link{dq_report2} report}

\item{...}{not used}
}
\value{
\link{list} with two slots \code{Data} and \code{Table} with \link{data.frame}s
featuring all metrics columns
from the report \code{r}, the \link{STUDY_SEGMENT} and the \link{VAR_NAMES}.
In case of \code{Data}, the columns are formatted nicely but still
with the standardized column names -- use
\code{util_translate_indicator_metrics()} to rename them nicely. In
case of \code{Table}, just as they are.
}
\description{
Extract report summary from reports
}
\seealso{
\code{\link[=prep_combine_report_summaries]{prep_combine_report_summaries()}}

Other summary_functions:
\code{\link[=prep_combine_report_summaries]{prep_combine_report_summaries()}},
\code{\link[=prep_extract_classes_by_functions]{prep_extract_classes_by_functions()}},
\code{\link[=prep_extract_summary]{prep_extract_summary()}},
\code{\link[=prep_extract_summary.dataquieR_result]{prep_extract_summary.dataquieR_result()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_ggplot2]{prep_render_pie_chart_from_summaryclasses_ggplot2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_plotly]{prep_render_pie_chart_from_summaryclasses_plotly()}},
\code{\link[=prep_summary_to_classes]{prep_summary_to_classes()}}
}
\concept{summary_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DF_UNIQUE_ID}
\alias{DF_UNIQUE_ID}
\title{Data frame level metadata attribute name}
\usage{
DF_UNIQUE_ID
}
\description{
Defines expectancies on the uniqueness of the IDs
across the rows of a data frame, or the number of times some ID can be repeated.
}
\seealso{
\link{meta_data_dataframe}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.dataquieR_summary.R
\name{plot.dataquieR_summary}
\alias{plot.dataquieR_summary}
\title{Plot a \code{dataquieR} summary}
\usage{
\method{plot}{dataquieR_summary}(
  x,
  y,
  ...,
  filter,
  dont_plot = FALSE,
  stratify_by,
  vars_to_include = "study",
  disable_plotly = FALSE,
  hierarchy,
  folder_of_report = NULL,
  var_uniquenames = NULL
)
}
\arguments{
\item{x}{the \code{dataquieR} summary, see \code{\link[=summary]{summary()}} and \code{\link[=dq_report2]{dq_report2()}}}

\item{y}{not yet used}

\item{...}{not yet used}

\item{filter}{if given, this filters the summary, e.g.,
\code{filter = call_names == "com_qualified_item_missingness"}}

\item{dont_plot}{suppress the actual plotting, just return a printable
object derived from \code{x}}

\item{stratify_by}{column to stratify the summary, may be one string.}

\item{vars_to_include}{\link{character} study | \code{ssi}. variables are study
variables or computed social science
indicator variables.}

\item{disable_plotly}{\link{logical} do not use \code{plotly}, even if installed}

\item{hierarchy}{not yet defined, but if an argument is given, a
sunburst chart is displayed, currently, only \code{DQ_OBS}
can be used a the hierarchy.}

\item{folder_of_report}{a named vector with the location of variable and
\code{call_names}}

\item{var_uniquenames}{a data frame with the original variable names and the
unique names in case of reports created with dq_report_by
containing the same variable in several reports
(e.g., creation of reports by sex)}
}
\value{
invisible html object
}
\description{
Plot a \code{dataquieR} summary
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_undisclose.R
\name{prep_undisclose}
\alias{prep_undisclose}
\title{Remove data disclosing details}
\usage{
prep_undisclose(x, cores)
}
\arguments{
\item{x}{an object to un-disclose, a}

\item{cores}{can be an integer with a number of cores to use. if not
specified, the function uses the default cluster, if available
and falls back to serial un-disclosing, otherwise.}
}
\value{
undisclosed object
}
\description{
new function: no warranty, so far.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/menu_env.R
\name{menu_env-menu}
\alias{menu_env-menu}
\title{Generate the menu for a report}
\arguments{
\item{pages}{encapsulated \code{list} with report pages as \code{tagList} objects,
its names are the desired file names}
}
\value{
the html-\code{taglist} for the menu
}
\description{
Generate the menu for a report
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{SEGMENT_PART_VARS}
\alias{SEGMENT_PART_VARS}
\title{Segment level metadata attribute name}
\usage{
SEGMENT_PART_VARS
}
\description{
The name of the segment participation status variable
}
\seealso{
\link{meta_data_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dq_report2.R
\name{dq_report2}
\alias{dq_report2}
\title{Generate a full DQ report, v2}
\usage{
dq_report2(
  study_data,
  item_level = "item_level",
  label_col = LABEL,
  meta_data_segment = "segment_level",
  meta_data_dataframe = "dataframe_level",
  meta_data_cross_item = "cross-item_level",
  meta_data_item_computation = "item_computation_level",
  meta_data = item_level,
  meta_data_v2,
  ...,
  dimensions = c("Completeness", "Consistency"),
  cores = list(mode = "socket", logging = FALSE, cpus = util_detect_cores(),
    load.balancing = TRUE),
  ignore_empty_vars = getOption("dataquieR.ignore_empty_vars",
    dataquieR.ignore_empty_vars_default),
  specific_args = list(),
  advanced_options = list(),
  author = prep_get_user_name(),
  title = "Data quality report",
  subtitle = as.character(Sys.Date()),
  user_info = NULL,
  debug_parallel = FALSE,
  resp_vars = character(0),
  filter_indicator_functions = character(0),
  exclude_indicator_functions = character(0),
  filter_result_slots = c("^Summary", "^Segment", "^DataTypePlotList",
    "^ReportSummaryTable", "^Dataframe", "^Result", "^VariableGroup"),
  mode = c("default", "futures", "queue", "parallel"),
  mode_args = list(),
  notes_from_wrapper = list(),
  storr_factory = NULL,
  amend = FALSE,
  cross_item_level,
  `cross-item_level`,
  segment_level,
  dataframe_level,
  item_computation_level,
  .internal = rlang::env_inherits(rlang::caller_env(), parent.env(environment())),
  checkpoint_resumed = getOption("dataquieR.resume_checkpoint",
    dataquieR.resume_checkpoint_default),
  name_of_study_data,
  dt_adjust = as.logical(getOption("dataquieR.dt_adjust", dataquieR.dt_adjust_default)),
  output_dir = NULL,
  force_overwrite = FALSE
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{meta_data_dataframe}{\link{data.frame} -- optional: Data frame level
metadata}

\item{meta_data_cross_item}{\link{data.frame} -- optional: Cross-item level
metadata}

\item{meta_data_item_computation}{\link{data.frame} optional. computation rules
for computed variables.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{...}{arguments to be passed to all called indicator functions if
applicable.}

\item{dimensions}{\link{dimensions} Vector of dimensions to address in the report.
Allowed values in the vector are Completeness, Consistency,
and Accuracy. The generated report will only cover the
listed data quality dimensions. Accuracy is computational
expensive, so this dimension is not enabled by default.
Completeness should be included, if Consistency is
included, and Consistency should be included, if Accuracy
is included to avoid misleading detections of e.g. missing
codes as outliers, please refer to the data quality concept
for more details. Integrity is always included.
If dimensions is equal to NULL or "all", all dimensions
will be covered.}

\item{cores}{\link{integer} number of cpu cores to use or a named list with
arguments for \link[parallelMap:parallelStart]{parallelMap::parallelStart} or NULL,
if parallel has already been started by the caller.
Can also be a cluster.}

\item{ignore_empty_vars}{\link{enum} TRUE | FALSE | auto. See
\link{dataquieR.ignore_empty_vars}.}

\item{specific_args}{\link{list} named list of arguments specifically for one of
the called functions, the of the list elements
correspond to the indicator functions whose calls
should be modified. The elements are lists of
arguments.}

\item{advanced_options}{\link{list} options to set during report computation,
see \code{\link[=options]{options()}}}

\item{author}{\link{character} author for the report documents.}

\item{title}{\link{character} optional argument to specify the title for
the data quality report}

\item{subtitle}{\link{character} optional argument to specify a subtitle for
the data quality report}

\item{user_info}{\link{list} additional info stored with the report, e.g.,
comments, title, ...}

\item{debug_parallel}{\link{logical} print blocks currently evaluated in parallel}

\item{resp_vars}{\link{variable list} the name of the measurement variables
for the report. If missing, all variables
will be used. Only item level indicator
functions are filtered, so far.}

\item{filter_indicator_functions}{\link{character} regular expressions, only
if an indicator function's name
matches one of these, it'll
be used for the report. If
of length zero, no filtering
is performed.}

\item{exclude_indicator_functions}{\link{character} regular expressions,
if an indicator function's name
matches one of these, it'll
be excluded from the report. If
of length zero, no filtering
is performed.}

\item{filter_result_slots}{\link{character} regular expressions, only
if an indicator function's
result's name
matches one of these, it'll
be used for the report. If
of length zero, no filtering
is performed.}

\item{mode}{\link{character} work mode for parallel execution. default is
"default", the values mean:
\itemize{
\item default: use \code{queue} except \code{cores} has been set explicitly
\item futures: use the \code{future} package
\item queue: use a queue as described in the examples
from the \code{callr} package by Csárdi and Chang and start
sub-processes as workers that evaluate the queue.
\item parallel: use the cluster from \code{cores} to evaluate all
calls of indicator functions using the classic
R \code{parallel} back-ends
}}

\item{mode_args}{\link{list} of arguments for the selected \code{mode}. As of writing
this manual, only for the mode \code{queue} the argument
\code{step} is supported, which gives the number of
function calls that are run by one worker at a time.
the default is 15, which gives on most of the tested
systems a good balance between synchronization
overhead and idling workers.}

\item{notes_from_wrapper}{\link{list} a list containing notes about changed labels
by \code{dq_report_by} (otherwise NULL)}

\item{storr_factory}{\link{function} \code{NULL}, or
a function returning a \code{storr} object as
back-end for the report's results. If used with
\code{cores > 1}, the storage must be accessible from all
cores and capable of concurrent writing according
to \code{storr}. Hint: \code{dataquieR} currently only supports
\code{storr::storr_rds()}, officially, while other back-
ends may nevertheless work, yet, they are not tested.}

\item{amend}{\link{logical} if there is already data in.\code{storr_factory},
use it anyways -- unsupported, so far!}

\item{cross_item_level}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{`cross-item_level`}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}

\item{item_computation_level}{\link{data.frame} alias for
\code{meta_data_item_computation}}

\item{.internal}{\link{logical} internal use, only.}

\item{checkpoint_resumed}{\link{logical} if using a \code{storr_factory} and the back-
end there is already filled, and if
\code{amend} is missing or set to \code{TRUE},
compute all missing result and add them
to the back-end.}

\item{name_of_study_data}{\link{character} name for study data inside the report,
internal use.}

\item{dt_adjust}{\link{logical} whether to trust data types in the study data. if
\code{TRUE}, data types are checked based on the
metadata and later casted to the declared type.
if your data source is already typed, this can
be turned off to speed up computations.
see \link{dataquieR.dt_adjust}}

\item{output_dir}{\link{character} if \code{output_dir} is not \code{NULL}, also create
\code{HTML} output for the report using
\code{\link[=print.dataquieR_resultset2]{print.dataquieR_resultset2()}}
written to the path \code{output_dir}}

\item{force_overwrite}{\link{logical} force to overwrite \code{output_dir}, even if it
exists}
}
\value{
a \link{dataquieR_resultset2} that can be
\link[=print.dataquieR_resultset2]{printed} creating a \code{HTML}-report.
}
\description{
Generate a full DQ report, v2
}
\details{
See \link{dq_report_by} for a way to generate stratified or splitted reports
easily.
}
\seealso{
\itemize{
\item \link{as.data.frame.dataquieR_resultset}
\item \link{as.list.dataquieR_resultset}
\item \link{print.dataquieR_resultset}
\item \link{summary.dataquieR_resultset}
\item \link{dq_report_by}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset2.R
\name{prep_load_report_from_backend}
\alias{prep_load_report_from_backend}
\title{Load a report from a back-end}
\usage{
prep_load_report_from_backend(
  namespace = "objects",
  db_dir,
  storr_factory = prep_create_storr_factory(namespace = namespace, db_dir = db_dir)
)
}
\arguments{
\item{namespace}{the namespace to read the report's results from}

\item{db_dir}{\link{character} path to the directory for the back-end, if
a \code{storr_rds} or \code{storr_torr} is used.}

\item{storr_factory}{a function returning a \code{storr} object holding the report}
}
\value{
\link{dataquieR_resultset2} the report
}
\description{
Load a report from a back-end
}
\examples{
\dontrun{
r <- dataquieR::dq_report2("study_data", meta_data_v2 = "meta_data_v2",
                           dimensions = NULL)
storr_factory <- prep_create_storr_factory()
r_storr <- prep_set_backend(r, storr_factory)
r_restorr <- prep_set_backend(r_storr, NULL)
r_loaded <- prep_load_report_from_backend(storr_factory)
}
}
\seealso{
\code{\link[=prep_create_storr_factory]{prep_create_storr_factory()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_summary_to_classes.R
\name{prep_summary_to_classes}
\alias{prep_summary_to_classes}
\title{Classify metrics from a report summary table}
\usage{
prep_summary_to_classes(report_summary)
}
\arguments{
\item{report_summary}{\code{\link[=list]{list()}} as returned by
\code{\link[=prep_extract_summary]{prep_extract_summary()}}}
}
\value{
\link{data.frame} classes for the report summary table, long format
}
\description{
Classify metrics from a report summary table
}
\seealso{
Other summary_functions:
\code{\link[=prep_combine_report_summaries]{prep_combine_report_summaries()}},
\code{\link[=prep_extract_classes_by_functions]{prep_extract_classes_by_functions()}},
\code{\link[=prep_extract_summary]{prep_extract_summary()}},
\code{\link[=prep_extract_summary.dataquieR_result]{prep_extract_summary.dataquieR_result()}},
\code{\link[=prep_extract_summary.dataquieR_resultset2]{prep_extract_summary.dataquieR_resultset2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_ggplot2]{prep_render_pie_chart_from_summaryclasses_ggplot2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_plotly]{prep_render_pie_chart_from_summaryclasses_plotly()}}
}
\concept{summary_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/com_segment_missingness.R
\name{com_segment_missingness}
\alias{com_segment_missingness}
\title{Summarizes missingness for individuals in specific segments}
\usage{
com_segment_missingness(
  study_data,
  item_level = "item_level",
  strata_vars = NULL,
  group_vars = NULL,
  label_col,
  threshold_value,
  direction,
  color_gradient_direction,
  expected_observations = c("HIERARCHY", "ALL", "SEGMENT"),
  exclude_roles = c(VARIABLE_ROLES$PROCESS),
  meta_data = item_level,
  meta_data_v2,
  segment_level,
  meta_data_segment
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{strata_vars}{\link{variable} the name of a variable used for stratification,
defaults to NULL for not grouping output}

\item{group_vars}{\link{variable} the name of a variable used for grouping,
defaults to \emph{NULL} for not grouping output}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{threshold_value}{\link{numeric} from=0 to=100. a numerical value ranging
from 0-100}

\item{direction}{\link{enum} low | high. "high" or "low", i.e. are deviations
above/below the threshold critical. This argument is deprecated and replaced by \emph{color_gradient_direction}.}

\item{color_gradient_direction}{\link{enum} above | below. "above" or "below", i.e. are deviations
above or below the threshold critical? (default: above)}

\item{expected_observations}{\link{enum} HIERARCHY | ALL | SEGMENT. If ALL, all
observations are expected to comprise
all study segments. If SEGMENT, the
\code{PART_VAR} is expected to point
to a variable with values of 0 and 1,
indicating whether the variable was
expected to be observed for each data
row. If HIERARCHY, this is also
checked recursively, so, if a variable
points to such a participation variable,
and that other variable does has also
a \code{PART_VAR} entry pointing
to a variable, the observation of the
initial variable is only
expected, if both segment variables are
1.}

\item{exclude_roles}{\link{variable roles} a character (vector) of variable roles
not included}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}

\item{meta_data_segment}{\link{data.frame} Segment level metadata. Optional.}
}
\value{
a list with:
\itemize{
\item \code{ResultData}: data frame about segment missingness
\item \code{SummaryPlot}: ggplot2 heatmap plot: a heatmap-like graphic that
highlights critical values depending on the respective
threshold_value and direction.
\item \code{ReportSummaryTable}: data frame underlying \code{SummaryPlot}
}
}
\description{
\subsection{This implementation can be applied in two use cases:}{
\enumerate{
\item participation in study segments is not recorded by respective variables,
e.g. a participant's refusal to attend a specific examination is not
recorded.
\item participation in study segments is recorded by respective
variables.
}

Use case \emph{(1)} will be common in smaller studies. For the calculation of
segment missingness it is assumed that study variables are nested in
respective segments. This structure must be specified in the static metadata.
The R-function identifies all variables within each segment and returns TRUE
if all variables within a segment are missing, otherwise FALSE.

Use case \emph{(2)} assumes a more complex structure of study data and metadata.
The study data comprise so-called intro-variables (either TRUE/FALSE or codes
for non-participation). The column \code{PART_VAR} in the metadata is
filled by variable-IDs indicating for each variable the respective
intro-variable. This structure has the benefit that subsequent calculation of
item missingness obtains correct denominators for the calculation of
missingness rates.

\link{Descriptor}
}
}
\details{
\subsection{Implementation and use of thresholds}{

This implementation uses one threshold to discriminate critical from
non-critical values. If direction is above than all values below the
threshold_value are normal (displayed in dark blue in the plot and flagged
with GRADING = 0 in the dataframe). All values above the threshold_value are
considered critical. The more they deviate from the threshold the displayed
color shifts to dark red. All critical values are highlighted with GRADING =
1 in the summary data frame. By default, highest values are always shown in
dark red irrespective of the absolute deviation.

If direction is below than all values above the threshold_value are normal
(displayed in dark blue, GRADING = 0).
}

\subsection{Hint}{

This function does not support a \code{resp_vars} argument but \code{exclude_roles} to
specify variables not relevant for detecting a missing segment.

List function.
}
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_com_impl_segment_missingness.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_geom_pointrange_robust.R
\name{to_basic.GeomPointrangeRobust}
\alias{to_basic.GeomPointrangeRobust}
\title{Internally used point-range}
\usage{
to_basic.GeomPointrangeRobust(data, prestats_data, layout, params, p, ...)
}
\arguments{
\item{data}{the data returned by \code{ggplot2::ggplot_build()}}

\item{prestats_data}{the data before statistics are computed.}

\item{layout}{the panel layout.}

\item{params}{parameters for the geom, statistic, and 'constant' aesthetics}

\item{p}{a ggplot2 object (the conversion may depend on scales, for instance).}

\item{...}{currently ignored}
}
\description{
Internally used point-range
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset2.R
\docType{class}
\name{dataquieR_resultset2-class}
\alias{dataquieR_resultset2-class}
\alias{dataquieR_resultset2}
\alias{.dataquieR_resultset2}
\title{Class \link{dataquieR_resultset2}.}
\description{
Class \link{dataquieR_resultset2}.
}
\seealso{
\link{dq_report2}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_unexp_records_segment.R
\name{int_unexp_records_segment}
\alias{int_unexp_records_segment}
\title{Check for unexpected data record count within segments}
\usage{
int_unexp_records_segment(
  study_segment,
  study_data,
  label_col,
  item_level = "item_level",
  data_record_count,
  meta_data = item_level,
  meta_data_segment = "segment_level",
  meta_data_v2,
  segment_level
)
}
\arguments{
\item{study_segment}{\link{character} a character vector indicating the name of each study data frame, mandatory.}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{data_record_count}{\link{integer}  an integer vector with the number of expected data records, mandatory.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}
}
\value{
a \link{list} with
\itemize{
\item \code{SegmentData}: data frame with the results of the quality check for unexpected data elements
\item \code{SegmentTable}: data frame with selected unexpected data elements check results, used for the data quality report.
}
}
\description{
This function contrasts the expected record number in each study segment in
the metadata with the actual record number in each segment data frame.

\link{Indicator}
}
\details{
The current implementation does not take into account jump or missing codes, the function is rather based on checking whether NAs are present in the study data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DOT.template_function_indicator.R
\name{.template_function_indicator}
\alias{.template_function_indicator}
\title{\code{Roxygen}-Template for indicator functions}
\usage{
.template_function_indicator(
  resp_vars,
  study_data,
  label_col,
  item_level,
  meta_data,
  meta_data_v2,
  meta_data_dataframe,
  meta_data_segment,
  dataframe_level,
  segment_level
)
}
\arguments{
\item{resp_vars}{\link{variable} the names of the measurement variables, if
missing or \code{NULL}, all variables will be checked}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{meta_data_dataframe}{\link{data.frame} the data frame that contains the
metadata for the data frame level}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}
}
\value{
\code{invisible(NULL)}
}
\description{
\code{Roxygen}-Template for indicator functions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dim.dataquieR_resultset2.R
\name{dim.dataquieR_resultset2}
\alias{dim.dataquieR_resultset2}
\title{Get the dimensions of a \code{dq_report2} result}
\usage{
\method{dim}{dataquieR_resultset2}(x)
}
\arguments{
\item{x}{a \code{dataquieR_resultset2} result}
}
\value{
dimensions
}
\description{
Get the dimensions of a \code{dq_report2} result
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset2.R
\name{$<-.dataquieR_resultset2}
\alias{$<-.dataquieR_resultset2}
\alias{cash-set-.dataquieR_resultset2.Rd}
\title{Write single results from a \link{dataquieR_resultset2} report}
\usage{
\method{$}{dataquieR_resultset2}(x, el) <- value
}
\arguments{
\item{x}{the report}

\item{el}{the index}

\item{value}{the single result}
}
\value{
the \code{dataquieR} result object
}
\description{
Write single results from a \link{dataquieR_resultset2} report
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_expand_codes.R
\name{prep_expand_codes}
\alias{prep_expand_codes}
\title{Expand code labels across variables}
\usage{
prep_expand_codes(
  item_level = "item_level",
  suppressWarnings = FALSE,
  mix_jumps_and_missings = FALSE,
  meta_data_v2,
  meta_data = item_level
)
}
\arguments{
\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{suppressWarnings}{\link{logical} show warnings, if labels are expanded}

\item{mix_jumps_and_missings}{\link{logical} ignore the class of the codes for
label expansion, i.e., use missing code labels
as jump code labels, if the values are the
same.}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}
}
\value{
\link{data.frame} an updated metadata data frame.
}
\description{
Code labels are copied from other
variables, if the code is the same and the
label is set only for some variables
}
\examples{
\dontrun{
meta_data <- prep_get_data_frame("meta_data")
meta_data$JUMP_LIST[meta_data$VAR_NAMES == "v00003"] <- "99980 = NOOP"
md <- prep_expand_codes(meta_data)
md$JUMP_LIST
md$MISSING_LIST
md <- prep_expand_codes(meta_data, mix_jumps_and_missings = TRUE)
md$JUMP_LIST
md$MISSING_LIST
meta_data <- prep_get_data_frame("meta_data")
meta_data$MISSING_LIST[meta_data$VAR_NAMES == "v00003"] <- "99980 = NOOP"
md <- prep_expand_codes(meta_data)
md$JUMP_LIST
md$MISSING_LIST
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_int_duplicate_ids_segment.R
\name{util_int_duplicate_ids_segment}
\alias{util_int_duplicate_ids_segment}
\title{Check for duplicated IDs}
\usage{
util_int_duplicate_ids_segment(
  level = c("segment"),
  id_vars_list,
  study_segment,
  repetitions,
  study_data,
  meta_data,
  meta_data_segment = "segment_level",
  segment_level
)
}
\arguments{
\item{level}{\link{character} a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").}

\item{id_vars_list}{\link{list} id variable names for each segment or data frame}

\item{study_segment}{\link{vector} the segments or data frame names being assessed}

\item{repetitions}{\link{vector} an integer vector indicating the number of allowed repetitions in the id_vars. Currently, no repetitions are supported.}

\item{study_data}{\link{data.frame} the data frame that contains the measurements, mandatory.}

\item{meta_data}{\link{data.frame} the data frame that contains metadata attributes of the study data, mandatory.}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}
}
\value{
a \link{list} with
\itemize{
\item \code{SegmentData}: data frame with the results of the quality check for duplicated identifiers
\item \code{SegmentTable}: data frame with selected duplicated identifiers check results, used for the data quality report.
\item \code{Other}: named list with inner lists of unique cases containing each the
row indices of duplicated identifiers separated by "|" , if any.
outer names are names of the segments.
Use \code{\link[=prep_get_study_data_segment]{prep_get_study_data_segment()}} to get the data frame the
indices refer to.
}
}
\description{
This function tests for duplicates entries in identifiers. It is possible to
check duplicated identifiers by study segments or to consider only selected
segments.
}
\seealso{
Other integrity_indicator_functions:
\code{\link[=util_int_duplicate_content_dataframe]{util_int_duplicate_content_dataframe()}},
\code{\link[=util_int_duplicate_content_segment]{util_int_duplicate_content_segment()}},
\code{\link[=util_int_duplicate_ids_dataframe]{util_int_duplicate_ids_dataframe()}},
\code{\link[=util_int_unexp_records_set_dataframe]{util_int_unexp_records_set_dataframe()}},
\code{\link[=util_int_unexp_records_set_segment]{util_int_unexp_records_set_segment()}}
}
\concept{integrity_indicator}
\concept{integrity_indicator_functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.dataquieR_resultset2.R
\name{print.list}
\alias{print.list}
\title{print a list of \code{dataquieR_result} objects}
\usage{
\method{print}{list}(x, ...)
}
\arguments{
\item{x}{\code{\link[=list]{list()}} only, if all elements inherit from \link{dataquieR_result},
this implementation runs}

\item{...}{passed to other implementations}
}
\value{
undefined
}
\description{
print a list of \code{dataquieR_result} objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_part_vars_structure.R
\name{int_part_vars_structure}
\alias{int_part_vars_structure}
\title{Detect Expected Observations}
\usage{
int_part_vars_structure(
  label_col,
  study_data,
  item_level = "item_level",
  expected_observations = c("HIERARCHY", "SEGMENT"),
  disclose_problem_paprt_var_data = FALSE,
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{label_col}{\link{character} mapping attribute \code{colnames(study_data)} vs.
\code{meta_data[label_col]}}

\item{study_data}{\link{study_data} must have all relevant \code{PART_VARS} to avoid
false-positives on \code{PART_VARS} missing from
\code{study_data}}

\item{item_level}{\link{meta_data} must be complete to avoid false positives on
non-existing \code{PART_VARS}}

\item{expected_observations}{\link{enum} HIERARCHY | SEGMENT. How should
\code{PART_VARS} be handled:
\itemize{
\item \code{SEGMENT}: if \code{PART_VAR} is 1, an
observation is expected
\item \code{HIERARCHY}: the default, if the
\code{PART_VAR} is 1 for this variable and
also for all \code{PART_VARS} of \code{PART_VARS}
up in the hierarchy, an observation is
expected.
}}

\item{disclose_problem_paprt_var_data}{\link{logical} show the problematic data
(\code{PART_VAR} only)}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
empty list, so far -- the function only warns.
}
\description{
For each participant, check, if an observation was expected, given the
\code{PART_VARS} from item-level metadata
}
\details{
\link{Descriptor}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_add_computed_variables.R
\name{prep_add_computed_variables}
\alias{prep_add_computed_variables}
\title{Insert missing codes for \code{NA}s based on rules}
\usage{
prep_add_computed_variables(
  study_data,
  meta_data,
  label_col,
  rules,
  use_value_labels
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{meta_data}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{rules}{\link{data.frame} with the columns:
\itemize{
\item \code{VAR_NAMES}: \link{VAR_NAMES} of the variable to compute
\item \code{COMPUTATION_RULE}: A rule in \code{REDcap} style (see, e.g.,
\href{https://help.redcap.ualberta.ca/help-and-faq/project-best-practices/data-quality/example-data-quality-rules}{\code{REDcap} help},
\href{https://docs.google.com/document/d/1l3nGBgqqPKi5PtMe75g7q0dny8QzGMd_/edit?tab=t.0}{\code{REDcap} how-to}), and
\href{https://www.iths.org/wp-content/uploads/REDCap-Branching-Logic-2017-202.pdf}{\code{REDcap} branching logic}
that defines, how to compute the new values
}}

\item{use_value_labels}{\link{logical} In rules for factors, use the value labels,
not the codes. Defaults to \code{TRUE}, if any
\code{VALUE_LABELS} are given in the metadata.}
}
\value{
a \code{list} with the entry:
\itemize{
\item \code{ModifiedStudyData}: Study data with the new variables
}
}
\description{
Insert missing codes for \code{NA}s based on rules
}
\examples{
\dontrun{
study_data <- prep_get_data_frame("ship")
prep_load_workbook_like_file("ship_meta_v2")
meta_data <- prep_get_data_frame("item_level")
rules <- tibble::tribble(
  ~VAR_NAMES,  ~COMPUTATION_RULE,
  "BMI", '[BODY_WEIGHT_0]/(([BODY_HEIGHT_0]/100)^2)',
  "R", '[WAIST_CIRC_0]/2/[pi]', # in m^3
  "VOL_EST", '[pi]*([WAIST_CIRC_0]/2/[pi])^2*[BODY_HEIGHT_0] / 1000', # in l
 )
 r <- prep_add_computed_variables(study_data, meta_data,
   label_col = "LABEL", rules, use_value_labels = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_open_in_excel.R
\name{prep_open_in_excel}
\alias{prep_open_in_excel}
\title{Open a data frame in Excel}
\usage{
prep_open_in_excel(dfr)
}
\arguments{
\item{dfr}{the data frame}
}
\value{
potentially modified data frame after dialog was closed
}
\description{
Open a data frame in Excel
}
\details{
if the file cannot be read on function exit, NULL will be returned
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/html_dependency_clipboard.R
\name{html_dependency_clipboard}
\alias{html_dependency_clipboard}
\title{HTML Dependency for report headers in \code{clipboard}}
\usage{
html_dependency_clipboard()
}
\value{
the dependency
}
\description{
HTML Dependency for report headers in \code{clipboard}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_lazy_ggplot.R
\name{prep_realize_ggplot}
\alias{prep_realize_ggplot}
\title{Materialize a lazy \code{ggplot}}
\usage{
prep_realize_ggplot(x)
}
\arguments{
\item{x}{a \code{dq_lazy_ggplot} object.}
}
\value{
A \code{ggplot} object.
}
\description{
Evaluate the stored expression in its lean environment and cache
the resulting \code{ggplot} object in the current R session, if enabled
using the option \link{dataquieR.lazy_plots_cache}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_add_missing_codes.R
\name{prep_add_missing_codes}
\alias{prep_add_missing_codes}
\title{Insert missing codes for \code{NA}s based on rules}
\usage{
prep_add_missing_codes(
  resp_vars,
  study_data,
  meta_data_v2,
  item_level = "item_level",
  label_col,
  rules,
  use_value_labels = NA,
  overwrite = FALSE,
  meta_data = item_level
)
}
\arguments{
\item{resp_vars}{\link{variable list} the name of the measurement variables to be
modified, all from \code{rules}, if omitted}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{rules}{\link{data.frame} with the columns:
\itemize{
\item \code{resp_vars} or \code{VAR_NAMES}:
Variable, whose \code{NA}-values should be replaced by jump codes
\item \code{CODE_CLASS}: Either \code{MISSING} or \code{JUMP}: Is the currently described case
an expected missing value (\code{JUMP}) or not (\code{MISSING})
\item \code{CODE_VALUE}: The jump code or missing code
\item \code{CODE_LABEL}: A label describing the reason for the missing value
\item \code{RULE}: A rule in \code{REDcap} style (see, e.g.,
\href{https://help.redcap.ualberta.ca/help-and-faq/project-best-practices/data-quality/example-data-quality-rules}{\code{REDcap} help},
\href{https://docs.google.com/document/d/1l3nGBgqqPKi5PtMe75g7q0dny8QzGMd_/edit?tab=t.0}{\code{REDcap} how-to}), and
\href{https://www.iths.org/wp-content/uploads/REDCap-Branching-Logic-2017-202.pdf}{\code{REDcap} branching logic}
that describes cases for the missing
\item \code{DATA_PREPARATION}: optional. if available, and \verb{use_value_labels`` is either missing or }NA`, this columns
controls the rule handling.
}}

\item{use_value_labels}{\link{logical} In rules for factors, use the value labels,
not the codes. Defaults to \code{TRUE}, if any
\code{VALUE_LABELS} are given in the metadata,
and no \code{DATA_PREPARATION} exists for the
\code{rules}. \code{NA} means to use
\code{DATA_PREPARATION}, if available.}

\item{overwrite}{\link{logical} Also insert missing codes, if the values are not
\code{NA}}

\item{meta_data}{\link{data.frame} old name for \code{item_level}
attributes of study data}
}
\value{
a \code{list} with the entries:
\itemize{
\item \code{ModifiedStudyData}: Study data with \code{NA}s replaced by the \code{CODE_VALUE}
\item \code{ModifiedMetaData}: Metadata having the new codes amended in the columns
\code{JUMP_LIST} or \code{MISSING_LIST}, respectively
}
}
\description{
Insert missing codes for \code{NA}s based on rules
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimnames_dataquieR_resultset2.R
\name{dimnames.dataquieR_resultset2}
\alias{dimnames.dataquieR_resultset2}
\title{Names of a \code{dataquieR} report object (v2.0)}
\usage{
\method{dimnames}{dataquieR_resultset2}(x)
}
\arguments{
\item{x}{the result object}
}
\value{
the names
}
\description{
Names of a \code{dataquieR} report object (v2.0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_generate_pages_from_report.R
\name{util_generate_pages_from_report}
\alias{util_generate_pages_from_report}
\title{Convert a \link[=dq_report2]{dataquieR report v2} to a named list of web pages}
\usage{
util_generate_pages_from_report(
  report,
  template,
  disable_plotly,
  progress = progress,
  progress_msg = progress_msg,
  block_load_factor,
  dir,
  my_dashboard
)
}
\arguments{
\item{report}{\link[=dq_report2]{dataquieR report v2}.}

\item{template}{\link{character} template to use, only the name, not the path}

\item{disable_plotly}{\link{logical} do not use \code{plotly}, even if installed}

\item{progress}{\code{\link{function}} lambda for progress in percent -- 1-100}

\item{progress_msg}{\code{\link{function}} lambda for progress messages}

\item{block_load_factor}{\link{numeric} multiply size of parallel compute blocks
by this factor.}

\item{dir}{\link{character} output directory for potential \code{iframes}.}

\item{my_dashboard}{\link{list} of class \code{shiny.tag.list} featuring a dashboard or
missing or \code{NULL}}
}
\value{
named list, each entry becomes a file with the name of the entry.
the contents are \code{HTML} objects as used by \code{htmltools}.
}
\description{
Convert a \link[=dq_report2]{dataquieR report v2} to a named list of web pages
}
\examples{
\dontrun{
devtools::load_all()
prep_load_workbook_like_file("meta_data_v2")
report <- dq_report2("study_data", dimensions = NULL, label_col = "LABEL");
save(report, file = "report_v2.RData")
report <- dq_report2("study_data", label_col = "LABEL");
save(report, file = "report_v2_short.RData")
}

}
\concept{html}
\concept{process}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_deparse_assignments.R
\name{prep_deparse_assignments}
\alias{prep_deparse_assignments}
\title{Convert two vectors from a code-value-table to a key-value list}
\usage{
prep_deparse_assignments(
  codes,
  labels = codes,
  split_char = SPLIT_CHAR,
  mode = c("numeric_codes", "string_codes")
)
}
\arguments{
\item{codes}{codes, numeric or dates (as default, but string codes can be
enabled using the option 'mode', see below)}

\item{labels}{\link{character} labels, same length as codes}

\item{split_char}{\link{character} split character character to split
code assignments}

\item{mode}{\link{character} one of two options to insist on numeric or datetime
codes (default) or to allow for string codes}
}
\value{
a vector with assignment strings for each row of
\code{cbind(codes, labels)}
}
\description{
Convert two vectors from a code-value-table to a key-value list
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_correct_variable_use.R
\name{.variable_arg_roles}
\alias{.variable_arg_roles}
\title{Variable-argument roles}
\usage{
.variable_arg_roles
}
\description{
A Variable-argument role is the intended use of an argument of a indicator
function -- an argument that refers variables.
In general for the table .variable_arg_roles, the suffix _var means one
variable allowed,
while _vars means more than one. The default  sets of arguments
for \code{util_correct_variable_use}/\code{util_correct_variable_use2} are defined
from the point of usage, e.g. if it could be, that NAs are in
the list of variable names, the function should be able to remove certain
response variables
from the output and not disallow them by setting \code{allow_na} to \code{FALSE}.
}
\seealso{
\code{util_correct_variable_use()}

\code{util_correct_variable_use2()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_load_workbook_like_file.R
\name{prep_load_workbook_like_file}
\alias{prep_load_workbook_like_file}
\title{Pre-load a file with named (usually more than) one table(s)}
\usage{
prep_load_workbook_like_file(file, keep_types = FALSE, append = FALSE)
}
\arguments{
\item{file}{the file name to load.}

\item{keep_types}{\link{logical} keep types as possibly defined in the file.
set \code{TRUE} for study data.}

\item{append}{\link{logical} if a data frame already exists in the cache
(by name), extend the existing one}
}
\value{
\verb{invisible(the cache environment)}
}
\description{
These can thereafter be referred to by their names only. Such files are,
e.g., spreadsheet-workbooks or \code{RData}-files.
}
\details{
Note, that this function in contrast to \link{prep_get_data_frame} does neither
support selecting specific sheets/columns from a file.
}
\seealso{
\link{prep_add_data_frames}

\link{prep_get_data_frame}

Other data-frame-cache:
\code{\link[=prep_add_data_frames]{prep_add_data_frames()}},
\code{\link[=prep_get_data_frame]{prep_get_data_frame()}},
\code{\link[=prep_list_dataframes]{prep_list_dataframes()}},
\code{\link[=prep_load_folder_with_metadata]{prep_load_folder_with_metadata()}},
\code{\link[=prep_purge_data_frame_cache]{prep_purge_data_frame_cache()}},
\code{\link[=prep_remove_from_cache]{prep_remove_from_cache()}}
}
\concept{data-frame-cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_as_numeric_with_unit.R
\name{/.numeric_with_unit}
\alias{/.numeric_with_unit}
\title{Operator caring for units}
\usage{
\method{/}{numeric_with_unit}(e1, e2)
}
\arguments{
\item{e1}{first argument}

\item{e2}{second argument}
}
\value{
result
}
\description{
Operator caring for units
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_translate.R
\name{names<-.dataquieR_translated}
\alias{names<-.dataquieR_translated}
\title{\code{names} implementation for the class \code{dataquieR_translated}}
\usage{
\method{names}{dataquieR_translated}(x) <- value
}
\arguments{
\item{x}{\code{dataquieR_translated} object}

\item{value}{the names to assign}
}
\value{
names of the underlying character vector
}
\description{
\code{dataquieR}'s translated texts featuring access to the language keys, still.
this function returns the language keys.
}
\details{
only \code{setNames(nm = x)} is allowed for convenience. Any other assignment
would mean to change the language keys, so this is not allowed.
}
\seealso{
base::as.character
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{SEGMENT_UNIQUE_ID}
\alias{SEGMENT_UNIQUE_ID}
\title{Segment level metadata attribute name}
\usage{
SEGMENT_UNIQUE_ID
}
\description{
Segment level metadata attribute name
}
\seealso{
\link{DF_UNIQUE_ID}

\link{meta_data_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dq_report_by.R
\name{dq_report_by}
\alias{dq_report_by}
\title{Generate a stratified full DQ report}
\usage{
dq_report_by(
  study_data,
  item_level = "item_level",
  meta_data_segment = "segment_level",
  meta_data_dataframe = "dataframe_level",
  meta_data_cross_item = "cross-item_level",
  meta_data_item_computation = "item_computation_level",
  missing_tables = NULL,
  label_col,
  meta_data_v2,
  segment_column = NULL,
  strata_column = NULL,
  strata_select = NULL,
  selection_type = NULL,
  segment_select = NULL,
  segment_exclude = NULL,
  strata_exclude = NULL,
  subgroup = NULL,
  resp_vars = character(0),
  id_vars = NULL,
  advanced_options = list(),
  storr_factory = NULL,
  amend = FALSE,
  checkpoint_resumed = getOption("dataquieR.resume_checkpoint",
    dataquieR.resume_checkpoint_default),
  ...,
  output_dir = NULL,
  input_dir = NULL,
  also_print = FALSE,
  force_overwrite = FALSE,
  disable_plotly = FALSE,
  view = TRUE,
  meta_data = item_level,
  cross_item_level,
  `cross-item_level`,
  segment_level,
  dataframe_level,
  item_computation_level,
  author = prep_get_user_name(),
  title = ifelse(is.null(output_dir), "Data quality report Bundle",
    paste0(basename(output_dir))),
  subtitle = as.character(Sys.Date()),
  user_info = NULL
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements:
it can be an R object (e.g., \code{bia}), a
data frame (e.g., \code{"C:/Users/data/bia.dta"}),
a vector containing data frames files (e.g.,
c(\code{"C:/Users/data/bia.dta"},
\verb{C:/Users/data/biames.dta"})), or it can be
left empty and the data frames are provided
in the data frame level metadata. If only the
file name without path is provided
(e.g., \code{"bia.dta"}), the file
name needs the extension and the path must be
provided in the argument \code{input_dir}. It can
also contain only the file name in case of
example data from the package \code{dataquieR}
(e.g., \code{"study_data"} or \code{"ship"})}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{meta_data_dataframe}{\link{data.frame} -- optional if \code{study_data} is
present: Data frame level metadata}

\item{meta_data_cross_item}{\link{data.frame} -- optional: Cross-item level
metadata}

\item{meta_data_item_computation}{\link{data.frame} -- optional: Computed items
metadata}

\item{missing_tables}{\link{character} the name of the data frame containing the
missing codes, it can be a vector if more
than one table is provided. Example:
\code{c("missing_table1", "missing_table2")}}

\item{label_col}{\link{variable attribute} the name of the column in the
metadata containing the labels of
the variables}

\item{meta_data_v2}{\link{character} path or file name of the workbook like
metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}}

\item{segment_column}{\link{variable attribute} name of a metadata attribute
usable to split the report in
sections of variables, e.g. all
blood-pressure related variables.
By default,
reports are split by
\link{STUDY_SEGMENT} if available and
no segment_column nor
strata_column or subgroup
are defined.
To create an un-split report
please write explicitly
the argument
'segment_column = NULL'}

\item{strata_column}{\link{variable} name of a study variable to stratify the
report by, e.g. the study centers.
Both labels and \code{VAR_NAMES} are accepted.
In case of NAs in the selected variable,
a separate report containing the NAs
subset will be created}

\item{strata_select}{\link{character} if given, the strata of strata_column
are limited to the content of this
vector. A character vector or a regular
expression can be provided
(e.g., "^a.*$"). This argument can not
be used if no strata_column is
provided}

\item{selection_type}{\link{character} optional, can only be specified if a
\code{strata_select} or
\code{strata_exclude} is specified.
If not present the
function try to guess what the
user typed as \code{strata_select} or
\code{strata_exclude}.
There are 3 options:
\code{value} indicating that the
stratum selected is a value and
not a value_label.
For example \code{"0"};
\code{v_label} indicating that the
stratum specified is a label.
For example \code{"male"}.
\code{regex} indicating that the user
specified strata using a regular
expression. For example \code{"^Ber"}
to select all strata starting
with that letters}

\item{segment_select}{\link{character} if given, the levels of segment_column
are limited to the content of this
vector. A character vector or a
regular expression (e.g., ".*_EXAM$")
can be provided.
This argument can not be used if no
segment_column is provided.}

\item{segment_exclude}{\link{character} optional, can only be specified if a
\code{segment_column} is specified.
The levels of \code{segment_column}
will not include the content of
this argument.
A character vector or
a regular  expression can be
provided  (e.g., "^STU").}

\item{strata_exclude}{\link{character} optional, can only be specified if a
\code{strata_column} is specified.
The strata of \code{strata_column}
will not include the content of
this argument.
A character vector or
a regular  expression can be
provided  (e.g., "^STU").}

\item{subgroup}{\link{character} optional, to define subgroups of cases. Rules are
to be written as \code{REDCap} rules.
Only VAR_NAMES are accepted in the rules.}

\item{resp_vars}{\link{variable} the names of the measurement variables, if
missing or \code{NULL}, all variables will be included}

\item{id_vars}{\link{variable} a vector containing the name/s of the variables
containing ids, to
be used to merge multiple data frames if provided
in \code{study_data} and to be add to referred vars}

\item{advanced_options}{\link{list} options to set during report computation,
see \code{\link[=options]{options()}}}

\item{storr_factory}{\link{function} \code{NULL}, or
a function returning a \code{storr} object as
back-end for the report's results. If used with
\code{cores > 1}, the storage must be accessible from all
cores and capable of concurrent writing according
to \code{storr}. Hint: \code{dataquieR} currently only supports
\code{storr::storr_rds()}, officially, while other back-
ends may nevertheless work, yet, they are not tested.}

\item{amend}{\link{logical} if there is already data in.\code{storr_factory},
use it anyways -- unsupported, so far!}

\item{checkpoint_resumed}{\link{logical} if using a \code{storr_factory} and the back-
end there is already filled, and if
\code{amend} is missing or set to \code{TRUE},
compute all missing result and add them
to the back-end.}

\item{...}{arguments to be passed through to \link{dq_report} or \link{dq_report2}}

\item{output_dir}{\link{character} if given, the output is not returned but saved
in this directory}

\item{input_dir}{\link{character} if given, the study data files that have
no path and that are not URL are searched in
this directory. Also \code{meta_data_v2} is searched
in this directory if no path is provided}

\item{also_print}{\link{logical} if \code{output_dir} is not \code{NULL}, also create
\code{HTML} output for each report using
\code{\link[=print.dataquieR_resultset2]{print.dataquieR_resultset2()}}
written to the path \code{output_dir}}

\item{force_overwrite}{\link{logical} force to overwrite \code{output_dir}, even if it
exists}

\item{disable_plotly}{\link{logical} do not use \code{plotly}, even if installed}

\item{view}{\link{logical} open the returned report}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{cross_item_level}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{`cross-item_level`}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}

\item{item_computation_level}{\link{data.frame} alias for
\code{meta_data_item_computation}}

\item{author}{\link{character} author for the report bundle's documents.}

\item{title}{\link{character} optional argument to specify the title for
the data quality report bundle}

\item{subtitle}{\link{character} optional argument to specify a subtitle for
the data quality report bundle}

\item{user_info}{\link{list} additional info stored with the report bundle,
e.g., comments, title, ...}
}
\value{
A named \link{list} of named \link{list}s of \link{dq_report2} reports, returned
invisibly unless \code{view = TRUE}. If \code{output_dir} is given, the result
is still returned (invisibly), and optionally opened in a browser
(\code{view = TRUE}, \code{also_print = TRUE}).
}
\description{
Generate a stratified full DQ report
}
\examples{
\dontrun{ # really long-running example.
prep_load_workbook_like_file("meta_data_v2")
rep <- dq_report_by("study_data", label_col =
  LABEL, strata_column = "CENTER_0")
rep <- dq_report_by("study_data",
  label_col = LABEL, strata_column = "CENTER_0",
  segment_column = NULL
)
unlink("/tmp/testRep/", force = TRUE, recursive = TRUE)
dq_report_by("study_data",
  label_col = LABEL, strata_column = "CENTER_0",
  segment_column = STUDY_SEGMENT, output_dir = "/tmp/testRep"
)
unlink("/tmp/testRep/", force = TRUE, recursive = TRUE)
dq_report_by("study_data",
  label_col = LABEL, strata_column = "CENTER_0",
  segment_column = NULL, output_dir = "/tmp/testRep"
)
dq_report_by("study_data",
  label_col = LABEL,
  segment_column = STUDY_SEGMENT, output_dir = "/tmp/testRep"
)
dq_report_by("study_data",
  label_col = LABEL,
  segment_column = STUDY_SEGMENT, output_dir = "/tmp/testRep",
  also_print = TRUE
)
dq_report_by(study_data = "study_data", meta_data_v2 = "meta_data_v2",
  advanced_options = list(dataquieR.study_data_cache_max = 0,
  dataquieR.study_data_cache_metrics = TRUE,
  dataquieR.study_data_cache_metrics_env = environment()),
  cores = NULL, dimensions = "int")
dq_report_by(study_data = "study_data", meta_data_v2 = "meta_data_v2",
  advanced_options = list(dataquieR.study_data_cache_max = 0),
  cores = NULL, dimensions = "int")
}
}
\seealso{
\link{dq_report}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset.R
\name{dataquieR_resultset}
\alias{dataquieR_resultset}
\title{Internal constructor for the internal class \link{dataquieR_resultset}.}
\usage{
dataquieR_resultset(...)
}
\arguments{
\item{...}{properties stored in the object}
}
\value{
an object of the class \link{dataquieR_resultset}.
}
\description{
creates an object of the class \link{dataquieR_resultset}.
}
\details{
The class features the following methods:
\itemize{
\item \link{as.data.frame.dataquieR_resultset},  * \link{as.list.dataquieR_resultset},  * \link{print.dataquieR_resultset},  * \link{summary.dataquieR_resultset}
}
}
\seealso{
\link{dq_report}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{SEGMENT_RECORD_CHECK}
\alias{SEGMENT_RECORD_CHECK}
\title{Segment level metadata attribute name}
\usage{
SEGMENT_RECORD_CHECK
}
\description{
The type of check to be conducted when comparing the reference ID
table with the IDs in a segment.
}
\seealso{
\link{meta_data_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{meta_data_dataframe}
\alias{meta_data_dataframe}
\title{Well known columns on the \code{meta_data_dataframe} sheet}
\description{
Metadata describing data delivered on one data frame/table sheet,
e.g., a full questionnaire, not its items.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{SEGMENT_UNIQUE_ROWS}
\alias{SEGMENT_UNIQUE_ROWS}
\title{Segment level metadata attribute name}
\usage{
SEGMENT_UNIQUE_ROWS
}
\description{
Specifies whether identical data is permitted across rows in a
segment (excluding ID variables)
}
\seealso{
\link{meta_data_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_setup_rstudio_job.R
\name{prep_deregister_progress_hook}
\alias{prep_deregister_progress_hook}
\title{De-register a hook function for progresses in computation/rendering}
\usage{
prep_deregister_progress_hook(handle, verbose = TRUE)
}
\arguments{
\item{handle}{\link{character} the handle}

\item{verbose}{\link{logical} message, if \code{handle} has currently no registration}
}
\value{
\link{logical} \code{invisible(TRUE)} on success
}
\description{
De-register a hook function for progresses in computation/rendering
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_options.R
\docType{data}
\name{dataquieR.lazy_plots_gg_compatibility}
\alias{dataquieR.lazy_plots_gg_compatibility}
\title{\link{character} be as compatible with \code{ggplot2} objects as possible}
\description{
if \code{TRUE}, plot promises are blessed in an \code{S7} class so they behave almost
like "real" \code{ggplot2} objects, so you normally do not need to call
\code{\link[=prep_realize_ggplot]{prep_realize_ggplot()}} on them. However, this comes with a small memory
overhead, so, you can disable this.
}
\seealso{
Other options:
\code{\link{dataquieR}},
\code{\link{dataquieR.CONDITIONS_LEVEL_TRHESHOLD}},
\code{\link{dataquieR.CONDITIONS_WITH_STACKTRACE}},
\code{\link{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE}},
\code{\link{dataquieR.ERRORS_WITH_CALLER}},
\code{\link{dataquieR.GAM_for_LOESS}},
\code{\link{dataquieR.MAHALANOBIS_THRESHOLD}},
\code{\link{dataquieR.MAX_LABEL_LEN}},
\code{\link{dataquieR.MAX_LONG_LABEL_LEN}},
\code{\link{dataquieR.MAX_VALUE_LABEL_LEN}},
\code{\link{dataquieR.MESSAGES_WITH_CALLER}},
\code{\link{dataquieR.MULTIVARIATE_OUTLIER_CHECK}},
\code{\link{dataquieR.VALUE_LABELS_htmlescaped}},
\code{\link{dataquieR.WARNINGS_WITH_CALLER}},
\code{\link{dataquieR.acc_loess.exclude_constant_subgroups}},
\code{\link{dataquieR.acc_loess.mark_time_points}},
\code{\link{dataquieR.acc_loess.min_bw}},
\code{\link{dataquieR.acc_loess.min_obs_in_subgroup}},
\code{\link{dataquieR.acc_loess.min_proportion}},
\code{\link{dataquieR.acc_loess.plot_format}},
\code{\link{dataquieR.acc_loess.plot_observations}},
\code{\link{dataquieR.acc_margins_num}},
\code{\link{dataquieR.acc_margins_sort}},
\code{\link{dataquieR.acc_multivariate_outlier.scale}},
\code{\link{dataquieR.acc_shape_or_scale_ci}},
\code{\link{dataquieR.col_con_con_empirical}},
\code{\link{dataquieR.col_con_con_logical}},
\code{\link{dataquieR.convert_to_list_for_lapply}},
\code{\link{dataquieR.debug}},
\code{\link{dataquieR.des_summary_hard_lim_remove}},
\code{\link{dataquieR.dontwrapresults}},
\code{\link{dataquieR.droplevels_ReportSummaryTable}},
\code{\link{dataquieR.dt_adjust}},
\code{\link{dataquieR.fix_column_type_on_read}},
\code{\link{dataquieR.flip_mode}},
\code{\link{dataquieR.force_item_specific_missing_codes}},
\code{\link{dataquieR.force_label_col}},
\code{\link{dataquieR.grading_formats}},
\code{\link{dataquieR.grading_rulesets}},
\code{\link{dataquieR.guess_character}},
\code{\link{dataquieR.guess_missing_codes}},
\code{\link{dataquieR.ignore_empty_vars}},
\code{\link{dataquieR.lang}},
\code{\link{dataquieR.lazy_plots}},
\code{\link{dataquieR.lazy_plots_cache}},
\code{\link{dataquieR.locale}},
\code{\link{dataquieR.max_cat_resp_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_in_plot}},
\code{\link{dataquieR.max_group_var_levels_with_violins}},
\code{\link{dataquieR.min_obs_per_group_var_in_plot}},
\code{\link{dataquieR.min_time_points_for_cat_resp_var}},
\code{\link{dataquieR.no_geom_count_in_bin}},
\code{\link{dataquieR.no_overall_in_bin}},
\code{\link{dataquieR.non_disclosure}},
\code{\link{dataquieR.old_factor_handling}},
\code{\link{dataquieR.old_type_adjust}},
\code{\link{dataquieR.precomputeStudyData}},
\code{\link{dataquieR.print_block_load_factor}},
\code{\link{dataquieR.progress_fkt_default}},
\code{\link{dataquieR.progress_msg_fkt_default}},
\code{\link{dataquieR.resume_checkpoint}},
\code{\link{dataquieR.resume_print}},
\code{\link{dataquieR.scale_level_heuristics_control_binaryrecodelimit}},
\code{\link{dataquieR.scale_level_heuristics_control_metriclevels}},
\code{\link{dataquieR.study_data_cache_max}},
\code{\link{dataquieR.study_data_cache_metrics}},
\code{\link{dataquieR.study_data_cache_metrics_env}},
\code{\link{dataquieR.study_data_cache_quick_fill}},
\code{\link{dataquieR.study_data_colnames_case_sensitive}},
\code{\link{dataquieR.testdebug}},
\code{\link{dataquieR.traceback}},
\code{\link{dataquieR.type_adjust_parallel}},
\code{\link{progress_init_fkt}}
}
\concept{options}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset2.R
\name{prep_set_backend}
\alias{prep_set_backend}
\title{Change the back-end of a report}
\usage{
prep_set_backend(r, storr_factory = NULL, amend = FALSE)
}
\arguments{
\item{r}{\link{dataquieR_resultset2} the report}

\item{storr_factory}{\code{storr} the \code{storr} storage or \code{NULL}, to move
the report fully back into the RAM.}

\item{amend}{\link{logical} if there is already data in.\code{storr_factory},
use it anyways -- unsupported, so far!}
}
\value{
\link{dataquieR_resultset2} but now with the desired back-end
}
\description{
with this function, you can move a report from/to a \code{storr} storage.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int_sts_element_segment.R
\name{int_sts_element_segment}
\alias{int_sts_element_segment}
\title{Checks for element set}
\usage{
int_sts_element_segment(
  study_data,
  item_level = "item_level",
  label_col,
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements, mandatory.}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
a \link{list} with
\itemize{
\item \code{SegmentData}: data frame with the unexpected elements check results.
- \code{Segment}: name of the corresponding segment,
if applicable, \code{ALL} otherwise
\item \code{SegmentTable}: data frame with the unexpected elements check results, used for the data quality report.
- \code{Segment}: name of the corresponding segment,
if applicable, \code{ALL} otherwise
}
}
\description{
Depends on \code{dataquieR.ELEMENT_MISSMATCH_CHECKTYPE} option,
see there -- # TODO: Rind out, how to document and link
it here using \code{Roxygen}.
}
\details{
\link{Indicator}
}
\examples{
\dontrun{
study_data <- cars
meta_data <- dataquieR::prep_create_meta(VAR_NAMES = c("speedx", "distx"),
  DATA_TYPE = c("integer", "integer"), MISSING_LIST = "|", JUMP_LIST = "|",
  STUDY_SEGMENT = c("Intro", "Ex"))
options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "none")
int_sts_element_segment(study_data, meta_data)
options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact")
int_sts_element_segment(study_data, meta_data)
study_data <- cars
meta_data <- dataquieR::prep_create_meta(VAR_NAMES = c("speedx", "distx"),
  DATA_TYPE = c("integer", "integer"), MISSING_LIST = "|", JUMP_LIST = "|",
  STUDY_SEGMENT = c("Intro", "Intro"))
options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "none")
int_sts_element_segment(study_data, meta_data)
options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact")
int_sts_element_segment(study_data, meta_data)
study_data <- cars
meta_data <- dataquieR::prep_create_meta(VAR_NAMES = c("speed", "distx"),
  DATA_TYPE = c("integer", "integer"), MISSING_LIST = "|", JUMP_LIST = "|",
  STUDY_SEGMENT = c("Intro", "Intro"))
options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "none")
int_sts_element_segment(study_data, meta_data)
options(dataquieR.ELEMENT_MISSMATCH_CHECKTYPE = "exact")
int_sts_element_segment(study_data, meta_data)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.interval.R
\name{print.interval}
\alias{print.interval}
\title{print implementation for the class \code{interval}}
\usage{
\method{print}{interval}(x, ...)
}
\arguments{
\item{x}{\code{interval} objects to print}

\item{...}{not used yet}
}
\value{
the printed object
}
\description{
such objects, for now, only occur in \code{RECCap} rules, so this function
is meant for internal use, mostly -- for now.
}
\seealso{
base::print
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_min_obs_level.R
\name{prep_min_obs_level}
\alias{prep_min_obs_level}
\title{Support function to identify the levels of a process variable with minimum
number of observations}
\usage{
prep_min_obs_level(study_data, group_vars, min_obs_in_subgroup)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{group_vars}{\link{variable list} the name grouping variable}

\item{min_obs_in_subgroup}{\link{integer} optional argument if a "group_var" is
used. This argument specifies the
minimum no. of observations that is
required to include a subgroup (level)
of the "group_var" in the analysis.
Subgroups with less observations are
excluded. The default is 30.}
}
\value{
a data frame with:
\itemize{
\item a subsample of original data
}
}
\description{
utility function to subset data based on minimum number of observation per
level
}
\details{
This functions removes observations having fewer than \code{min_obs_in_subgroup}
distinct values in a group variable, e.g. blood pressure measurements
performed by an examiner having fewer than e.g. 50 measurements done. It
displays a warning, if samples/rows are removed and returns the modified
study data frame.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_acc_distributions_with_ecdf.R
\name{prep_acc_distributions_with_ecdf}
\alias{prep_acc_distributions_with_ecdf}
\title{Utility function to plot a combined figure for distribution checks}
\usage{
prep_acc_distributions_with_ecdf(
  resp_vars = NULL,
  group_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2,
  n_group_max = getOption("dataquieR.max_group_var_levels_in_plot",
    dataquieR.max_group_var_levels_in_plot_default),
  n_obs_per_group_min = getOption("dataquieR.min_obs_per_group_var_in_plot",
    dataquieR.min_obs_per_group_var_in_plot_default)
)
}
\arguments{
\item{resp_vars}{\link{variable list} the name of the measurement variable}

\item{group_vars}{\link{variable list} the name of the observer, device or
reader variable}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{n_group_max}{maximum number of categories to be displayed individually
for the grouping variable (\code{group_vars}, devices / examiners)}

\item{n_obs_per_group_min}{minimum number of data points per group to create
a graph for an individual category of the \code{group_vars} variable}
}
\value{
A \code{SummaryPlot}.
}
\description{
Data quality indicator checks "Unexpected location" with histograms and
plots of empirical cumulative distributions for the subgroups.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{SPLIT_CHAR}
\alias{SPLIT_CHAR}
\title{Character used  by default as a separator in metadata such as
missing codes}
\usage{
SPLIT_CHAR
}
\description{
This 1 character is according to our metadata concept "|".
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/droplevels.ReportSummaryTable.R
\name{droplevels.ReportSummaryTable}
\alias{droplevels.ReportSummaryTable}
\title{Remove unused levels from \code{ReportSummaryTable}}
\usage{
\method{droplevels}{ReportSummaryTable}(x, ...)
}
\arguments{
\item{x}{\code{ReportSummaryTable} an object from which to drop unused factor
levels.}

\item{...}{no used.}
}
\value{
\code{ReportSummaryTable} with all (\code{NA} or \code{0})-columns removed
}
\description{
Remove unused levels from \code{ReportSummaryTable}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/com_unit_missingness.R
\name{com_unit_missingness}
\alias{com_unit_missingness}
\title{Counts all individuals with no measurements at all}
\usage{
com_unit_missingness(
  id_vars = NULL,
  strata_vars = NULL,
  label_col,
  study_data,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{id_vars}{\link{variable list} optional, a (vectorized) call of ID-variables
that should not be
considered in the calculation of unit-
missingness}

\item{strata_vars}{\link{variable} optional, a string or integer variable used for
stratification}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
A list with:
\itemize{
\item \code{FlaggedStudyData}: \link{data.frame} with id-only-rows flagged in a column
\code{Unit_missing}
\item \code{SummaryData}: \link{data.frame} with numbers and percentages of unit
missingness
}
}
\description{
This implementation examines a crude version of unit missingness or
unit-nonresponse (Kalton and Kasprzyk 1986), i.e. if all measurement
variables in the study data are missing for an observation it has unit
missingness.

The function can be applied on stratified data. In this case strata_vars must
be specified.

\link{Descriptor}
}
\details{
This implementations calculates a crude rate of unit-missingness. This type
of missingness may have several causes and is an important research outcome.
For example, unit-nonresponse may be selective regarding the targeted study
population or technical reasons such as record-linkage may cause
unit-missingness.

It has to be discriminated form segment and item missingness, since different
causes and mechanisms may be the reason for unit-missingness.
\subsection{Hint}{

This function does not support a \code{resp_vars} argument but \code{id_vars}, which
have a roughly inverse logic behind: id_vars with values do not prevent a row
from being considered missing, because an ID is the only hint for a unit that
elsewise would not occur in the data at all.

List function.
}
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_com_impl_unit_missingness.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/com_qualified_item_missingness.R
\name{com_qualified_item_missingness}
\alias{com_qualified_item_missingness}
\title{Compute Indicators for Qualified Item Missingness}
\usage{
com_qualified_item_missingness(
  resp_vars,
  study_data,
  label_col = NULL,
  item_level = "item_level",
  expected_observations = c("HIERARCHY", "ALL", "SEGMENT"),
  meta_data = item_level,
  meta_data_v2,
  meta_data_segment,
  segment_level
)
}
\arguments{
\item{resp_vars}{\link{variable list} the name of the measurement variables}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{expected_observations}{\link{enum} HIERARCHY | ALL | SEGMENT. Report the
number of observations expected using
the old \code{PART_VAR} concept. See
\link{com_item_missingness} for an
explanation.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}
}
\value{
A \link{list} with:
\itemize{
\item \code{SummaryTable}: \link{data.frame} containing data quality checks for
"Non-response rate" (\code{PCT_com_qum_nonresp}) and
"Refusal rate" (\code{PCT_com_qum_refusal}) for each response
variable in \code{resp_vars}.
\item \code{SummaryData}: a \link{data.frame} containing data quality checks for
“Non-response rate” and "Refusal rate"
for a report
}
}
\description{
\link{Indicator}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_get_labels.R
\name{prep_get_labels}
\alias{prep_get_labels}
\title{Fetch a label for a variable based on its purpose}
\usage{
prep_get_labels(
  resp_vars,
  item_level = "item_level",
  label_col,
  max_len,
  label_class = c("SHORT", "LONG"),
  label_lang = getOption("dataquieR.lang", dataquieR.lang_default),
  resp_vars_are_var_names_only = FALSE,
  resp_vars_match_label_col_only = FALSE,
  meta_data = item_level,
  meta_data_v2,
  force_label_col = getOption("dataquieR.force_label_col",
    dataquieR.force_label_col_default)
)
}
\arguments{
\item{resp_vars}{\link{variable list} the variable names to fetch for}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{max_len}{\link{integer} the maximum label length to return, if not possible
w/o causing ambiguous labels, the labels may still
be longer. For \code{label_class == "LONG"}, it defaults
to \code{200}, while for \code{label_class == "SHORT"} to \code{30}}

\item{label_class}{\link{enum} SHORT | LONG. which sort of label
according to the metadata model should be returned}

\item{label_lang}{\link{character} optional language suffix, if available in
the metadata. Can be controlled by the option
\code{dataquieR.lang}.}

\item{resp_vars_are_var_names_only}{\link{logical} If \code{TRUE}, do not use
other labels than \link{VAR_NAMES}
for finding \code{resp_vars} in
\code{meta_data}}

\item{resp_vars_match_label_col_only}{\link{logical} If \code{TRUE}, do not use
other labels than those,
referred by \code{label_col}
for finding \code{resp_vars} in
\code{meta_data}}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{force_label_col}{\link{enum} auto | FALSE | TRUE. if \code{TRUE}, always use
labels according \code{label_col}, \code{FALSE} means
use labels matching best the function's
requirements, \code{auto} means \code{FALSE}, if in a
\code{dq_report()} and \code{TRUE}, otherwise.}
}
\value{
\link{character} suitable labels for each \code{resp_vars}, names of this
vector are \link{VAR_NAMES}
}
\description{
Fetch a label for a variable based on its purpose
}
\examples{
\dontrun{
prep_load_workbook_like_file("meta_data_v2")
prep_get_labels("SEX_0", label_class = "SHORT", max_len = 2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DF_RECORD_CHECK}
\alias{DF_RECORD_CHECK}
\title{Data frame level metadata attribute name}
\usage{
DF_RECORD_CHECK
}
\description{
The type of check to be conducted when comparing the reference
ID table with the IDs delivered in the study data files.
}
\seealso{
\link{meta_data_dataframe}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_loess.R
\name{acc_loess}
\alias{acc_loess}
\title{Smoothes and plots adjusted longitudinal measurements and longitudinal trends
from logistic regression models}
\usage{
acc_loess(
  resp_vars,
  group_vars = NULL,
  time_vars,
  co_vars = NULL,
  study_data,
  label_col = VAR_NAMES,
  item_level = "item_level",
  min_obs_in_subgroup = getOption("dataquieR.acc_loess.min_obs_in_subgroup",
    dataquieR.acc_loess.min_obs_in_subgroup_default),
  resolution = 80,
  comparison_lines = list(type = c("mean/sd", "quartiles"), color = "grey30", linetype =
    2, sd_factor = 0.5),
  mark_time_points = getOption("dataquieR.acc_loess.mark_time_points",
    dataquieR.acc_loess.mark_time_points_default),
  plot_observations = getOption("dataquieR.acc_loess.plot_observations",
    dataquieR.acc_loess.plot_observations_default),
  plot_format = getOption("dataquieR.acc_loess.plot_format",
    dataquieR.acc_loess.plot_format_default),
  meta_data = item_level,
  meta_data_v2,
  n_group_max = getOption("dataquieR.max_group_var_levels_in_plot",
    dataquieR.max_group_var_levels_in_plot_default),
  enable_GAM = getOption("dataquieR.GAM_for_LOESS", dataquieR.GAM_for_LOESS_default),
  exclude_constant_subgroups =
    getOption("dataquieR.acc_loess.exclude_constant_subgroups",
    dataquieR.acc_loess.exclude_constant_subgroups_default),
  min_bandwidth = getOption("dataquieR.acc_loess.min_bw",
    dataquieR.acc_loess.min_bw_default),
  min_proportion = getOption("dataquieR.acc_loess.min_proportion",
    dataquieR.acc_loess.min_proportion_default)
)
}
\arguments{
\item{resp_vars}{\link{variable} the name of the continuous measurement variable}

\item{group_vars}{\link{variable} the name of the observer, device or reader
variable}

\item{time_vars}{\link{variable} the name of the variable giving the time
of measurement}

\item{co_vars}{\link{variable list} a vector of covariables for adjustment, for
example age and sex. Can be NULL (default) for no
adjustment.}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{min_obs_in_subgroup}{\link{integer} (optional argument) If \code{group_vars} is
specified, this argument can be used to specify
the minimum number of observations required for
each of the subgroups. Subgroups with fewer
observations are excluded. The default number
is \code{30}.}

\item{resolution}{\link{numeric} the maximum number of time points used for
plotting the trend lines}

\item{comparison_lines}{\link{list} type and style of lines with which trend
lines are to be compared. Can be mean +/- 0.5
standard deviation (the factor can be specified
differently in \code{sd_factor}) or quartiles
(Q1, Q2, and Q3). Arguments \code{color} and
\code{linetype} are passed to \code{\link[ggplot2:geom_line]{ggplot2::geom_line()}}.}

\item{mark_time_points}{\link{logical} mark time points with observations
(caution, there may be many marks)}

\item{plot_observations}{\link{logical} show observations as scatter plot in the
background. If there are \code{co_vars} specified,
the values of the observations in the plot will
also be adjusted for the specified covariables.}

\item{plot_format}{\link{enum} AUTO | COMBINED | FACETS | BOTH. Return the plot
as one combined plot for all groups or as
facet plots (one figure per group). \code{BOTH} will
return both variants, \code{AUTO} will decide based
on the number of observers.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{n_group_max}{\link{integer} maximum number of categories to be displayed
individually for the grouping variable (\code{group_vars},
devices / examiners)}

\item{enable_GAM}{\link{logical} Can LOESS computations be replaced by general
additive models to reduce memory consumption  for large
datasets?}

\item{exclude_constant_subgroups}{\link{logical} Should subgroups with constant
values be excluded?}

\item{min_bandwidth}{\link{numeric} lower limit for the LOESS bandwidth, should be
greater than 0 and less than or equal to 1. In general,
increasing the bandwidth leads to a smoother trend line.}

\item{min_proportion}{\link{numeric} lower limit for the proportion of the smaller
group (cases or controls) for creating a LOESS figure,
should be greater than 0 and less than 0.4.}
}
\value{
a \link{list} with:
\itemize{
\item \code{SummaryPlotList}: list with two plots if \code{plot_format = "BOTH"},
otherwise one of the two figures described below:
\itemize{
\item \code{Loess_fits_facets}: The plot contains LOESS-smoothed curves
for each level of the \code{group_vars} in a separate panel. Added trend
lines represent mean and standard deviation or quartiles (specified
in \code{comparison_lines}) for moving windows over the whole data.
\item \code{Loess_fits_combined}: This plot combines all curves into one
panel. Given a low number of levels in the \code{group_vars}, this plot
eases comparisons. However, if the number increases this plot may
be too crowded and unclear.
}
}
}
\description{
The following R implementation executes calculations for quality indicator
"Unexpected location" (see \href{https://dataquality.qihs.uni-greifswald.de/PDQC_DQ_3_2_1_3.html}{here}. Local regression (LOESS) is a versatile statistical method to explore an
averaged course of time series
measurements (Cleveland, Devlin, and Grosse 1988). In context of
epidemiological data, repeated measurements using the same measurement
device or by the same examiner can be considered a time series. LOESS allows
to explore changes in these measurements over time.

\link{Descriptor}
}
\details{
If \code{mark_time_points} or \code{plot_observations} is selected, but would result in
plotting more than 400 points, only a sample of the data will be displayed.

Limitations

The application of LOESS requires model fitting, i.e. the smoothness
of a model is subject to a smoothing parameter (span).
Particularly in the presence of interval-based missing data, high
variability of measurements combined with a low number of
observations in one level of the \code{group_vars} may distort the fit.
Since our approach handles data without knowledge
of such underlying characteristics, finding the best fit is complicated if
computational costs should be minimal. The default of
LOESS in R uses a span of 0.75, which provides in most cases reasonable fits.
The function \code{acc_loess} adapts the span for each level of the \code{group_vars}
(with at least as many observations as specified in \code{min_obs_in_subgroup}
and with at least three time points) based on the respective
number of observations.
LOESS consumes a lot of memory for larger datasets. That is why \code{acc_loess}
switches to a generalized additive model with integrated smoothness
estimation (\code{gam} by \code{mgcv}) if there are 1000 observations or more for
at least one level of the \code{group_vars} (similar to \code{geom_smooth}
from \code{ggplot2}).
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_loess.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{meta_data}
\alias{meta_data}
\title{Data frame with metadata about the study data on variable level}
\description{
Variable level metadata.
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_Annotation_of_Metadata.html}{further details on variable level metadata.}

\link{meta_data_segment}

\link{meta_data_dataframe}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{WELL_KNOWN_META_VARIABLE_NAMES}
\alias{WELL_KNOWN_META_VARIABLE_NAMES}
\alias{VAR_NAMES}
\alias{LABEL}
\alias{DATA_TYPE}
\alias{SCALE_LEVEL}
\alias{UNIT}
\alias{VALUE_LABELS}
\alias{VALUE_LABEL_TABLE}
\alias{MISSING_LIST}
\alias{JUMP_LIST}
\alias{MISSING_LIST_TABLE}
\alias{HARD_LIMITS}
\alias{DETECTION_LIMITS}
\alias{SOFT_LIMITS}
\alias{CONTRADICTIONS}
\alias{DISTRIBUTION}
\alias{DECIMALS}
\alias{DATA_ENTRY_TYPE}
\alias{END_DIGIT_CHECK}
\alias{CO_VARS}
\alias{GROUP_VAR_OBSERVER}
\alias{GROUP_VAR_DEVICE}
\alias{KEY_OBSERVER}
\alias{KEY_DEVICE}
\alias{TIME_VAR}
\alias{TIME_VAR_END}
\alias{KEY_DATETIME}
\alias{PART_VAR}
\alias{STUDY_SEGMENT}
\alias{KEY_STUDY_SEGMENT}
\alias{VARIABLE_ROLE}
\alias{VARIABLE_ORDER}
\alias{LONG_LABEL}
\alias{SOFT_LIMIT_LOW}
\alias{SOFT_LIMIT_UP}
\alias{HARD_LIMIT_LOW}
\alias{HARD_LIMIT_UP}
\alias{DETECTION_LIMIT_LOW}
\alias{DETECTION_LIMIT_UP}
\alias{INCL_SOFT_LIMIT_LOW}
\alias{INCL_SOFT_LIMIT_UP}
\alias{INCL_HARD_LIMIT_LOW}
\alias{INCL_HARD_LIMIT_UP}
\alias{LOCATION_RANGE}
\alias{LOCATION_METRIC}
\alias{PROPORTION_RANGE}
\alias{LOCATION_LIMIT_LOW}
\alias{LOCATION_LIMIT_UP}
\alias{INCL_LOCATION_LIMIT_LOW}
\alias{INCL_LOCATION_LIMIT_UP}
\alias{PROPORTION_LIMIT_LOW}
\alias{PROPORTION_LIMIT_UP}
\alias{INCL_PROPORTION_LIMIT_LOW}
\alias{INCL_PROPORTION_LIMIT_UP}
\alias{RECODE_CASES}
\alias{RECODE_CONTROL}
\alias{GRADING_RULESET}
\alias{STANDARDIZED_VOCABULARY_TABLE}
\alias{DATAFRAMES}
\alias{ENCODING}
\alias{UNIVARIATE_OUTLIER_CHECKTYPE}
\alias{N_RULES}
\alias{EXTENDED_DATA_TYPE}
\alias{COMPUTED_VARIABLE_ROLE}
\title{Well-known metadata column names, names of metadata columns}
\usage{
WELL_KNOWN_META_VARIABLE_NAMES
}
\description{
names of the variable attributes in the metadata frame holding
the names of the respective observers, devices, lower limits for plausible
values, upper limits for plausible values, lower limits for allowed values,
upper limits for allowed values, the variable name (column name, e.g.
v0020349) used in the study data,  the variable name used for processing
(readable name, e.g. RR_DIAST_1) and in parameters of the QA-Functions, the
variable label, variable long label, variable short label, variable data
type (see also \link{DATA_TYPES}), re-code for definition of lists of event
categories, missing lists and jump lists as CSV strings. For valid units
see \link{UNITS}.
}
\details{
all entries of this list will be mapped to the package's exported NAMESPACE
environment directly, i.e. they are available directly by their names too:
\itemize{
\item \link{VAR_NAMES}
\item \link{LABEL}
\item \link{DATA_TYPE}
\item \link{SCALE_LEVEL}
\item \link{UNIT}
\item \link{VALUE_LABELS}
\item \link{VALUE_LABEL_TABLE}
\item \link{MISSING_LIST}
\item \link{JUMP_LIST}
\item \link{MISSING_LIST_TABLE}
\item \link{HARD_LIMITS}
\item \link{DETECTION_LIMITS}
\item \link{SOFT_LIMITS}
\item \link{CONTRADICTIONS}
\item \link{DISTRIBUTION}
\item \link{DECIMALS}
\item \link{DATA_ENTRY_TYPE}
\item \link{END_DIGIT_CHECK}
\item \link{CO_VARS}
\item \link{GROUP_VAR_OBSERVER}
\item \link{GROUP_VAR_DEVICE}
\item \link{KEY_OBSERVER}
\item \link{KEY_DEVICE}
\item \link{TIME_VAR}
\item \link{TIME_VAR_END}
\item \link{KEY_DATETIME}
\item \link{PART_VAR}
\item \link{STUDY_SEGMENT}
\item \link{KEY_STUDY_SEGMENT}
\item \link{VARIABLE_ROLE}
\item \link{VARIABLE_ORDER}
\item \link{LONG_LABEL}
\item \link{SOFT_LIMIT_LOW}
\item \link{SOFT_LIMIT_UP}
\item \link{HARD_LIMIT_LOW}
\item \link{HARD_LIMIT_UP}
\item \link{DETECTION_LIMIT_LOW}
\item \link{DETECTION_LIMIT_UP}
\item \link{INCL_SOFT_LIMIT_LOW}
\item \link{INCL_SOFT_LIMIT_UP}
\item \link{INCL_HARD_LIMIT_LOW}
\item \link{INCL_HARD_LIMIT_UP}
\item \link{LOCATION_RANGE}
\item \link{LOCATION_METRIC}
\item \link{PROPORTION_RANGE}
\item \link{LOCATION_LIMIT_LOW}
\item \link{LOCATION_LIMIT_UP}
\item \link{INCL_LOCATION_LIMIT_LOW}
\item \link{INCL_LOCATION_LIMIT_UP}
\item \link{PROPORTION_LIMIT_LOW}
\item \link{PROPORTION_LIMIT_UP}
\item \link{INCL_PROPORTION_LIMIT_LOW}
\item \link{INCL_PROPORTION_LIMIT_UP}
\item \link{RECODE_CASES}
\item \link{RECODE_CONTROL}
\item \link{GRADING_RULESET}
\item \link{STANDARDIZED_VOCABULARY_TABLE}
\item \link{DATAFRAMES}
\item \link{ENCODING}
\item \link{UNIVARIATE_OUTLIER_CHECKTYPE}
\item \link{N_RULES}
\item \link{EXTENDED_DATA_TYPE}
\item \link{COMPUTED_VARIABLE_ROLE}
}
}
\examples{
print(WELL_KNOWN_META_VARIABLE_NAMES$VAR_NAMES)
# print(VAR_NAMES) # should usually also work
}
\seealso{
\link{meta_data_segment} for \code{STUDY_SEGMENT}

\href{https://dataquality.qihs.uni-greifswald.de/VIN_Item_Level_Metadata.html}{online}

Other UNITS:
\code{\link{UNITS}},
\code{\link{UNIT_IS_COUNT}},
\code{\link{UNIT_PREFIXES}},
\code{\link{UNIT_PREFIX_FACTORS}},
\code{\link{UNIT_SOURCES}}
}
\concept{UNITS}
\alias{variable attribute}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_options.R
\name{dataquieR.study_data_cache_metrics_env_default}
\alias{dataquieR.study_data_cache_metrics_env_default}
\title{Default space for some metrics during report computation}
\usage{
dataquieR.study_data_cache_metrics_env_default
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\seealso{
Other study_data_cache:
\code{\link{dataquieR.ignore_empty_vars}},
\code{\link{dataquieR.precomputeStudyData}},
\code{\link{dataquieR.print_block_load_factor}},
\code{\link{dataquieR.study_data_cache_max}},
\code{\link{dataquieR.study_data_cache_metrics}},
\code{\link{dataquieR.study_data_cache_metrics_env}},
\code{\link{dataquieR.study_data_cache_quick_fill}}
}
\concept{study_data_cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.dataquieR_resultset.R
\name{summary.dataquieR_resultset}
\alias{summary.dataquieR_resultset}
\title{Summarize a \link{dataquieR} report}
\usage{
\method{summary}{dataquieR_resultset}(...)
}
\arguments{
\item{...}{Deprecated}
}
\value{
Deprecated
}
\description{
Deprecated
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_contradictions.R
\name{con_contradictions}
\alias{con_contradictions}
\title{Checks user-defined contradictions in study data}
\usage{
con_contradictions(
  resp_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  threshold_value,
  check_table,
  summarize_categories = FALSE,
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable list} the name of the measurement variables}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{threshold_value}{\link{numeric} from=0 to=100. a numerical value
ranging from 0-100}

\item{check_table}{\link{data.frame} contradiction rules table.  Table defining
contradictions. See details for
its required structure.}

\item{summarize_categories}{\link{logical} Needs a column 'tag' in the
\code{check_table}.
If set, a summary output is generated for the
defined categories plus one plot per
category.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
If \code{summarize_categories} is \code{FALSE}:
A \link{list} with:
\itemize{
\item \code{FlaggedStudyData}: The first output of the contradiction function is a
data frame of similar dimension regarding the number
of observations in the study data. In addition, for
each applied check on the variables an additional
column is added which flags observations with a
contradiction given the applied check.
\item \code{SummaryTable}: The second output summarizes this information into one
data frame. This output can be used to provide an
executive overview on the amount of contradictions. This
output is meant for automatic digestion within pipelines.
\item \code{SummaryData}: The third output is the same as \code{SummaryTable} but for
human readers.
\item \code{SummaryPlot}: The fourth output visualizes summarized information
of \code{SummaryData}.
}

if \code{summarize_categories} is \code{TRUE}, other objects are returned:
one per category named by that category (e.g. "Empirical") containing a
result for contradictions within that category only. Additionally, in the
slot \code{all_checks} a result as it would have been returned with
\code{summarize_categories} set to \code{FALSE}. Finally, a slot \code{SummaryData} is
returned containing sums per Category and an according \link[ggplot2:ggplot]{ggplot2::ggplot} in
\code{SummaryPlot}.
}
\description{
This approach considers a contradiction if impossible combinations of data
are observed in one participant. For example, if age of a participant is
recorded repeatedly the value of age is (unfortunately) not able to decline.
Most cases of contradictions rest on comparison of two variables.

Important to note, each value that is used for comparison may represent a
possible characteristic but the combination of these two values is considered
to be impossible. The approach does not consider implausible or inadmissible
values.

\link{Descriptor}
}
\details{
\subsection{Algorithm of this implementation:}{
\itemize{
\item Select all variables in the data with defined contradiction rules (static
metadata column CONTRADICTIONS)
\item Remove missing codes from the study data (if defined in the metadata)
\item Remove measurements deviating from limits defined in the metadata
\item Assign label to levels of categorical variables (if applicable)
\item Apply contradiction checks on predefined sets of variables
\item Identification of measurements fulfilling contradiction rules. Therefore
two output data frames are generated:
\itemize{
\item on the level of observation to flag each contradictory value
combination, and
\item a summary table for each contradiction check.
}
\item A summary plot illustrating the number of contradictions is generated.
}

List function.
}
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{CODE_ORDER}
\alias{CODE_ORDER}
\title{Only existence is checked, order not yet used}
\usage{
CODE_ORDER
}
\description{
Only existence is checked, order not yet used
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_get_study_data_segment.R
\name{prep_get_study_data_segment}
\alias{prep_get_study_data_segment}
\title{Get data frame for a given segment}
\usage{
prep_get_study_data_segment(
  segment,
  study_data,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2,
  segment_level,
  meta_data_segment = "segment_level"
)
}
\arguments{
\item{segment}{\link{character} name of the segment to return data for}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}
}
\value{
\link{data.frame} the data for the segment
}
\description{
Get data frame for a given segment
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_internal_api.R
\name{.get_internal_api}
\alias{.get_internal_api}
\title{Get Access to Utility Functions}
\usage{
.get_internal_api(fkt, version = API_VERSION, or_newer = TRUE)
}
\arguments{
\item{fkt}{function name}

\item{version}{version number to get}
}
\value{
an API object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DATA_TYPES}
\alias{DATA_TYPES}
\alias{integer}
\alias{string}
\alias{float}
\alias{datetime}
\alias{time}
\alias{numeric}
\alias{enum}
\alias{FLOAT}
\alias{INTEGER}
\alias{STRING}
\alias{DATETIME}
\alias{TIME}
\alias{variable}
\alias{set}
\title{Data Types}
\usage{
DATA_TYPES
}
\description{
\subsection{Data Types of Study Data}{

In the metadata, the following entries are allowed for the
\link{variable attribute} \link{DATA_TYPE}:
}
}
\details{
\itemize{
\item \code{integer} for integer numbers
\item \code{string} for text/string/character data
\item \code{float} for decimal/floating point numbers
\item \code{datetime} for timepoints
\item \code{time} for time of day
}
\subsection{Data Types of Function Arguments}{

As function arguments, \link{dataquieR} uses additional type specifications:
\itemize{
\item \code{numeric} is a numerical value (\link{float} or \link{integer}), but it is not an
allowed \code{DATA_TYPE} in the metadata. However, some functions may accept
\code{float} or \code{integer} for specific function arguments. This is, where we
use the term \code{numeric}.
\item \code{enum} allows one element out of a set of allowed options similar to
\link{match.arg}
\item \code{set} allows a subset out of a set of allowed options similar to
\link{match.arg} with \code{several.ok = TRUE}.
\item \code{variable} Function arguments of this type expect a character scalar that
specifies one variable using the variable identifier given in
the metadata attribute \code{VAR_NAMES} or, if \code{label_col} is set,
given in the metadata attribute given in that argument.
Labels can easily be translated using \link{prep_map_labels}
\item \verb{variable list} Function arguments of this type expect a character vector
that specifies variables using the variable identifiers
given in the metadata attribute \code{VAR_NAMES} or,
if \code{label_col} is set, given in the metadata attribute
given in that argument. Labels can easily be translated
using \link{prep_map_labels}
}
}
}
\seealso{
\link{integer} \link{string}
}
\alias{variable list}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{SCALE_LEVELS}
\alias{SCALE_LEVELS}
\title{Scale Levels}
\usage{
SCALE_LEVELS
}
\description{
\subsection{Scale Levels of Study Data according to \verb{Stevens's} Typology}{

In the metadata, the following entries are allowed for the
\link{variable attribute} \link{SCALE_LEVEL}:
}
}
\details{
\itemize{
\item \code{nominal} for categorical variables
\item \code{ordinal} for ordinal variables (i.e., comparison of values is possible)
\item \code{interval} for interval scales, i.e., distances are meaningful
\item \code{ratio} for ratio scales, i.e., ratios are meaningful
\item \code{na} for variables, that contain e.g. unstructured texts, \code{json},
\code{xml}, ... to distinguish them from variables, that still need to
have the \code{SCALE_LEVEL} estimated by
\code{prep_scalelevel_from_data_and_metadata()}
}
\subsection{Examples}{
\itemize{
\item sex, eye color -- \code{nominal}
\item income group, education level -- \code{ordinal}
\item temperature in degree Celsius -- \code{interval}
\item body weight, temperature in Kelvin -- \code{ratio}
}
}
}
\seealso{
\href{https://en.wikipedia.org/wiki/Level_of_measurement}{Wikipedia}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_pairs_matrix.R
\name{grid.draw.util_pairs_ggplot_panels}
\alias{grid.draw.util_pairs_ggplot_panels}
\title{\code{grid.draw} method for \code{util_pairs_ggplot_panels} objects}
\usage{
\method{grid.draw}{util_pairs_ggplot_panels}(x, ...)
}
\arguments{
\item{x}{An object of class \code{util_pairs_ggplot_panels}.}

\item{...}{Ignored.}
}
\description{
\code{grid.draw} method for \code{util_pairs_ggplot_panels} objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_apply_coding.R
\name{prep_apply_coding}
\alias{prep_apply_coding}
\title{Re-Code labels with their respective codes according to the \code{meta_data}}
\usage{
prep_apply_coding(
  study_data,
  meta_data_v2,
  item_level = "item_level",
  meta_data = item_level
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}
}
\value{
\link{data.frame} modified study data with labels replaced by the codes
}
\description{
Re-Code labels with their respective codes according to the \code{meta_data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reflection.R
\docType{data}
\name{Descriptor}
\alias{Descriptor}
\title{Descriptor Function}
\usage{
Descriptor
}
\description{
A function that returns some figure or table to assess data quality, but it
does not return a value correlating with the magnitude of a data quality
problem. It's the opposite of an \link{Indicator}.

The object \code{Descriptor} only contains the name used internally to tag
such functions.
}
\seealso{
\link{Indicator}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resnames.dataquieR_resultset2.R
\name{resnames.dataquieR_resultset2}
\alias{resnames.dataquieR_resultset2}
\title{Return names of result slots (e.g., 3rd dimension of dataquieR results)}
\usage{
\method{resnames}{dataquieR_resultset2}(x)
}
\arguments{
\item{x}{the objects}
}
\value{
character vector with names
}
\description{
Return names of result slots (e.g., 3rd dimension of dataquieR results)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/con_contradictions_redcap.R
\name{con_contradictions_redcap}
\alias{con_contradictions_redcap}
\title{Checks user-defined contradictions in study data}
\usage{
con_contradictions_redcap(
  study_data,
  item_level = "item_level",
  label_col,
  threshold_value,
  meta_data_cross_item = "cross-item_level",
  use_value_labels,
  summarize_categories = FALSE,
  meta_data = item_level,
  cross_item_level,
  `cross-item_level`,
  meta_data_v2
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{threshold_value}{\link{numeric} from=0 to=100. a numerical value
ranging from 0-100}

\item{meta_data_cross_item}{\link{data.frame} contradiction rules table.  Table
defining contradictions. See
\href{https://dataquality.qihs.uni-greifswald.de/VIN_Cross_Item_Level_Metadata.html}{online documentation}
for its required structure.}

\item{use_value_labels}{\link{logical} Deprecated in favor of \link{DATA_PREPARATION}.
If set to \code{TRUE}, labels can be used in the
\code{REDCap} syntax to specify contraction checks for
categorical variables. If set to \code{FALSE},
contractions have to be specified using the coded
values. In case that this argument is not set in
the function call, it will be set to \code{TRUE} if
the metadata contains a column \code{VALUE_LABELS}
which is not empty.}

\item{summarize_categories}{\link{logical} Needs a column \code{CONTRADICTION_TYPE} in
the \code{meta_data_cross_item}.
If set, a summary output is generated for the
defined categories plus one plot per
category. TODO: Not yet controllable by metadata.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{cross_item_level}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{`cross-item_level`}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
If \code{summarize_categories} is \code{FALSE}:
A \link{list} with:
\itemize{
\item \code{FlaggedStudyData}: The first output of the contradiction function is a
data frame of similar dimension regarding the number
of observations in the study data. In addition, for
each applied check on the variables an additional
column is added which flags observations with a
contradiction given the applied check.
\item \code{VariableGroupData}: The second output summarizes this information
into one
data frame. This output can be used to provide an
executive overview on the amount of contradictions.
\item \code{VariableGroupTable}: A subset of \code{VariableGroupData} used within the
pipeline.
\item \code{SummaryPlot}: The third output visualizes summarized information
of \code{SummaryData}.
}

If \code{summarize_categories} is \code{TRUE}, other objects are returned:
A list with one element \code{Other}, a list with the following entries:
One per category named by that category (e.g. "Empirical") containing a
result for contradiction checks within that category only. Additionally, in the
slot \code{all_checks}, a result as it would have been returned with
\code{summarize_categories} set to \code{FALSE}. Finally, in
the top-level list, a slot \code{SummaryData} is
returned containing sums per Category and an according \link[ggplot2:ggplot]{ggplot2::ggplot} in
\code{SummaryPlot}.
}
\description{
This approach considers a contradiction if impossible combinations of data
are observed in one participant. For example, if age of a participant is
recorded repeatedly the value of age is (unfortunately) not able to decline.
Most cases of contradictions rest on comparison of two variables.

Important to note, each value that is used for comparison may represent a
possible characteristic but the combination of these two values is considered
to be impossible. The approach does not consider implausible or inadmissible
values.

\link{Indicator}
}
\details{
\subsection{Algorithm of this implementation:}{
\itemize{
\item Remove missing codes from the study data (if defined in the metadata)
\item Remove measurements deviating from limits defined in the metadata
\item Assign label to levels of categorical variables (if applicable)
\item Apply contradiction checks (given as \code{REDCap}-like rules in a separate
metadata table)
\item Identification of measurements fulfilling contradiction rules. Therefore
two output data frames are generated:
\itemize{
\item on the level of observation to flag each contradictory value
combination, and
\item a summary table for each contradiction check.
}
\item A summary plot illustrating the number of contradictions is generated.
}

List function.
}
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_con_impl_contradictions_redcap.html}{Online Documentation for the function}
\link{meta_data_cross}
\href{https://dataquality.qihs.uni-greifswald.de/Cross_Item_Level_Metadata.html}{Online Documentation for the required cross-item-level metadata}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{value/missing-lists}
\alias{value/missing-lists}
\alias{cause_label_df}
\alias{missing_matchtable}
\alias{CODE_VALUE}
\alias{CODE_LABEL}
\alias{CODE_CLASS}
\alias{CODE_INTERPRET}
\alias{value_label_table}
\title{Data frame with labels for missing- and jump-codes
#' Metadata about value and missing codes}
\description{
\link{data.frame} with the following columns:
\itemize{
\item \code{CODE_VALUE}: \link{numeric} | \link{DATETIME} Missing or categorical code
(the number or date representing a
missing/category)
\item \code{CODE_LABEL}: \link{character} a label for the missing code or category
\item \code{CODE_CLASS}: \link{enum} JUMP | MISSING. For missing lists: Class of the
missing code.
\item \code{CODE_INTERPRET} \link{enum} I | P | PL | R | BO | NC | O | UH | UO | NE.
For missing lists: Class of the missing code
according to
\href{https://aapor.org/standards-and-ethics/standard-definitions/}{\code{AAPOR}}.
\item \code{resp_vars}: \link{character} For missing lists: optional, if a missing code
is specific for some
variables, it is listed for each such variable
with one entry in \code{resp_vars}, If \code{NA}, the
code is assumed shared among all variables.
For v1.0 metadata, you need to refer to
\code{VAR_NAMES} here.
}
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_Item_Level_Metadata.html#MISSING_LIST_TABLE}{Online}

\code{\link[=com_item_missingness]{com_item_missingness()}}

\code{\link[=com_segment_missingness]{com_segment_missingness()}}

\code{\link[=com_qualified_item_missingness]{com_qualified_item_missingness()}}

\code{\link[=com_qualified_segment_missingness]{com_qualified_segment_missingness()}}

\code{\link[=con_inadmissible_categorical]{con_inadmissible_categorical()}}

\code{\link[=con_inadmissible_vocabulary]{con_inadmissible_vocabulary()}}

\link{MISSING_LIST_TABLE}

\link{VALUE_LABEL_TABLE}

\link{STANDARDIZED_VOCABULARY_TABLE}

\link{cause_label_df}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{meta_data_segment}
\alias{meta_data_segment}
\title{Well known columns on the \code{meta_data_segment} sheet}
\description{
Metadata describing study segments, e.g., a full questionnaire, not its
items.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/html_dependency_jspdf.R
\name{html_dependency_jspdf}
\alias{html_dependency_jspdf}
\title{HTML dependency for \code{jsPDF}}
\usage{
html_dependency_jspdf()
}
\value{
An \code{\link[htmltools:htmlDependency]{htmltools::htmlDependency()}} object
}
\description{
Provides \code{jsPDF} for use in \code{Shiny} or \code{RMarkdown} via \code{htmltools}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset2.R
\name{as.list.dataquieR_resultset2}
\alias{as.list.dataquieR_resultset2}
\title{inefficient way to convert a report to a list. try \code{\link[=prep_set_backend]{prep_set_backend()}}}
\usage{
\method{as.list}{dataquieR_resultset2}(x, ...)
}
\arguments{
\item{x}{\link{dataquieR_resultset2}}

\item{...}{no used}
}
\value{
\link{list}
}
\description{
inefficient way to convert a report to a list. try \code{\link[=prep_set_backend]{prep_set_backend()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_mahalanobis.R
\name{acc_mahalanobis}
\alias{acc_mahalanobis}
\title{Calculate and plot \code{Mahalanobis} distances}
\usage{
acc_mahalanobis(
  variable_group = NULL,
  study_data,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_cross_item = "cross-item_level",
  label_col = VAR_NAMES,
  meta_data_v2,
  cross_item_level,
  `cross-item_level`,
  mahalanobis_threshold =
    suppressWarnings(as.numeric(getOption("dataquieR.MAHALANOBIS_THRESHOLD",
    dataquieR.MAHALANOBIS_THRESHOLD_default)))
)
}
\arguments{
\item{variable_group}{\link{variable list} the names of the variables used to
calculate the \code{Mahalanobis} distance}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_cross_item}{\link{data.frame} -- Cross-item level metadata}

\item{label_col}{\link{variable attribute} the name of the column in the
metadata containing the labels of
the variables}

\item{meta_data_v2}{\link{character} path or file name of the workbook like
metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}}

\item{cross_item_level}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{`cross-item_level`}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{mahalanobis_threshold}{\link{numeric} the confidence level to use to define
\code{outliers}, if not stated it is by default
0.975.}
}
\value{
a list with:
\itemize{
\item \code{SummaryTable}: \link{data.frame} underlying the plot
\item \code{SummaryData}: \link{data.frame} underlying the plot with speaking column labels
\item \code{SummaryPlot}: \link[ggplot2:ggplot2]{ggplot2::ggplot2} Q-Q plot of squared \code{Mahalanobis}
distances vs. a theoretical
chi-squared distribution showing \code{outliers}.
\item \code{FlaggedStudyData}: \link{data.frame} contains the original data frame of the
variables used to calculate
the squared \code{Mahalanobis} distances
with the additional column,
containing the squared
\code{Mahalanobis} distance, and a column
called \code{MD_outliers}, that contains
1 if the observational unit is considered
a multivariate \code{outlier}.
}
}
\description{
A standard tool to calculate \code{Mahalanobis} distance.
In this approach the squared \code{Mahalanobis} distance is calculated for ordinal
variables (treated as continuous) to identify inattentive responses.
It calculates the distance for each observational unit from the sample mean.
The greater the distance, the atypical the responses.

\link{Indicator}
}
\section{ALGORITHM OF THIS IMPLEMENTATION:}{
\itemize{
\item Implementation is restricted to variables of type integer
\item Remove missing codes from the study data (if defined in the metadata)
\item The covariance matrix is estimated for all variables from \code{variable_group}
\item The \code{Mahalanobis} distance of each observation is calculated
\eqn{MD^2_i  = (x_i - \mu)^T \Sigma^{-1} (x_i -  \mu)}
\item The default to consider a value an \code{outlier} is to use the 0.975 quantile
of a theoretical chi-square distribution with degrees of freedom
equals to the number of variables used to calculate the
\code{Mahalanobis} distance (\verb{Mayrhofer and Filzmoser}, 2023)
}
}

\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_multivariate_outlier.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{VARATT_REQUIRE_LEVELS}
\alias{VARATT_REQUIRE_LEVELS}
\alias{COMPATIBILITY}
\alias{REQUIRED}
\alias{RECOMMENDED}
\alias{OPTIONAL}
\alias{TECHNICAL}
\alias{UNKNOWN}
\title{Requirement levels of certain metadata columns}
\usage{
VARATT_REQUIRE_LEVELS
}
\description{
These levels are cumulatively used by the function \link{prep_create_meta} and
related in the argument \code{level} therein.
}
\details{
currently available:
\itemize{
\item 'COMPATIBILITY' = "compatibility"
\item 'REQUIRED' = "required"
\item 'RECOMMENDED' = "recommended"
\item 'OPTIONAL' = "optional"
\item 'TECHNICAL' = "technical"
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_get_data_frame.R
\name{prep_get_data_frame}
\alias{prep_get_data_frame}
\title{Read data from files/URLs}
\usage{
prep_get_data_frame(
  data_frame_name,
  .data_frame_list = .dataframe_environment(),
  keep_types = FALSE,
  column_names_only = FALSE
)
}
\arguments{
\item{data_frame_name}{\link{character} name of the data frame to read, see
details}

\item{.data_frame_list}{\link{environment} cache for loaded data frames}

\item{keep_types}{\link{logical} keep types as possibly defined in a file, if the
data frame is loaded from one. set \code{TRUE} for
study data.}

\item{column_names_only}{\link{logical} if TRUE imports only headers (column names)
of the data frame and no content
(an empty data frame)}
}
\value{
\link{data.frame} a data frame
}
\description{
data_frame_name can be a file path or an URL you can append a pipe and a
sheet name for Excel files or object name e.g. for \code{RData} files. Numbers
may also work. All file formats supported by your \code{rio} installation will
work.
}
\details{
The data frames will be cached automatically, you can define an alternative
environment for this using the argument \code{.data_frame_list}, and you can purge
the cache using \link{prep_purge_data_frame_cache}.

Use \link{prep_add_data_frames} to manually add data frames to the
cache, e.g., if you have loaded them from more complex sources, before.
}
\examples{
\dontrun{
bl <- as.factor(prep_get_data_frame(
  paste0("https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus",
    "/Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=",
    "publicationFile|COVID_Todesfälle_BL|Bundesland"))[[1]])

n <- as.numeric(prep_get_data_frame(paste0(
  "https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus/",
  "Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=",
  "publicationFile|COVID_Todesfälle_BL|Anzahl verstorbene",
  " COVID-19 Fälle"))[[1]])
plot(bl, n)
# Working names would be to date (2022-10-21), e.g.:
#
# https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus/ \\
#    Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=publicationFile
# https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus/  \\
#    Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=publicationFile|2
# https://www.rki.de/DE/Content/InfAZ/N/Neuartiges_Coronavirus/ \\
#    Projekte_RKI/COVID-19_Todesfaelle.xlsx?__blob=publicationFile|name
# study_data
# ship
# meta_data
# ship_meta
#
prep_get_data_frame("meta_data | meta_data")
}
}
\seealso{
\link{prep_add_data_frames}

\link{prep_load_workbook_like_file}

Other data-frame-cache:
\code{\link[=prep_add_data_frames]{prep_add_data_frames()}},
\code{\link[=prep_list_dataframes]{prep_list_dataframes()}},
\code{\link[=prep_load_folder_with_metadata]{prep_load_folder_with_metadata()}},
\code{\link[=prep_load_workbook_like_file]{prep_load_workbook_like_file()}},
\code{\link[=prep_purge_data_frame_cache]{prep_purge_data_frame_cache()}},
\code{\link[=prep_remove_from_cache]{prep_remove_from_cache()}}
}
\concept{data-frame-cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_int_unexp_records_set_segment.R
\name{util_int_unexp_records_set_segment}
\alias{util_int_unexp_records_set_segment}
\title{Check for unexpected data record set}
\usage{
util_int_unexp_records_set_segment(
  level = c("segment"),
  id_vars_list,
  identifier_name_list,
  valid_id_table_list,
  meta_data_record_check_list,
  study_data,
  label_col,
  meta_data,
  item_level,
  meta_data_segment = "segment_level",
  segment_level
)
}
\arguments{
\item{level}{\link{character} a character vector indicating whether the assessment should be conducted at the study level (level = "dataframe") or at the segment level (level = "segment").}

\item{id_vars_list}{\link{list} the list containing the identifier variables names to be used in the assessment.}

\item{identifier_name_list}{\link{list} the list that contains the name of the identifier to be used in the assessment. For the study level, corresponds to the names of the different data frames. For the segment level, indicates the name of the segments.}

\item{valid_id_table_list}{\link{list} the reference list with the identifier variable values.}

\item{meta_data_record_check_list}{\link{character} a character vector indicating the type of check to conduct, either "subset" or "exact".}

\item{study_data}{\link{data.frame} the data frame that contains the measurements, mandatory.}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{meta_data}{\link{data.frame} the data frame that contains metadata attributes of the study data, mandatory.}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}
}
\value{
a \link{list} with
\itemize{
\item \code{SegmentData}: data frame with the results of the quality check for unexpected data elements
\item \code{SegmentTable}: data frame with selected unexpected data elements check results, used for the data quality report.
\item \code{UnexpectedRecords}: vector with row indices of duplicated records, if any, otherwise NULL.
}
}
\description{
This function tests that the identifiers match a provided record set. It is possible to
check for unexpected data record sets by study segments or to consider only selected
segments.
}
\seealso{
Other integrity_indicator_functions:
\code{\link[=util_int_duplicate_content_dataframe]{util_int_duplicate_content_dataframe()}},
\code{\link[=util_int_duplicate_content_segment]{util_int_duplicate_content_segment()}},
\code{\link[=util_int_duplicate_ids_dataframe]{util_int_duplicate_ids_dataframe()}},
\code{\link[=util_int_duplicate_ids_segment]{util_int_duplicate_ids_segment()}},
\code{\link[=util_int_unexp_records_set_dataframe]{util_int_unexp_records_set_dataframe()}}
}
\concept{integrity_indicator}
\concept{integrity_indicator_functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_setup_rstudio_job.R
\name{prep_register_progress_hook}
\alias{prep_register_progress_hook}
\title{Register a hook function for progresses in computation/rendering}
\usage{
prep_register_progress_hook(type = c("progress", "init", "msg"), hook)
}
\arguments{
\item{type}{\link{character} what event}

\item{hook}{\link{function} hook function}
}
\value{
\link{character} a handle for de-registering, \code{invisible}
}
\description{
The order hooks are called is not defined.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/html_dependency_dataquieR.R
\name{html_dependency_dataquieR}
\alias{html_dependency_dataquieR}
\title{HTML Dependency for \code{dataquieR}}
\usage{
html_dependency_dataquieR(iframe = FALSE)
}
\arguments{
\item{iframe}{\link{logical}\code{(1)} if \code{TRUE}, create the dependency used in figure
\code{iframes}.}
}
\value{
the dependency
}
\description{
generate all dependencies used in static \code{dataquieR} reports
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline_recursive_result.R
\name{pipeline_recursive_result}
\alias{pipeline_recursive_result}
\title{Convert a pipeline result data frame to named encapsulated lists}
\usage{
pipeline_recursive_result(...)
}
\arguments{
\item{...}{Deprecated}
}
\value{
Deprecated
}
\description{
Deprecated
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind.R
\name{rbind.ReportSummaryTable}
\alias{rbind.ReportSummaryTable}
\title{Combine \code{ReportSummaryTable} outputs}
\usage{
\method{rbind}{ReportSummaryTable}(...)
}
\arguments{
\item{...}{\code{ReportSummaryTable} objects to combine.}
}
\description{
Using this \code{rbind} implementation, you can combine different
heatmap-like results of the class \code{ReportSummaryTable}.
}
\seealso{
\link[base:rbind.data.frame]{base::rbind.data.frame}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_check_meta_data_dataframe.R
\name{prep_check_meta_data_dataframe}
\alias{prep_check_meta_data_dataframe}
\title{Verify and normalize metadata on data frame level}
\usage{
prep_check_meta_data_dataframe(
  meta_data_dataframe = "dataframe_level",
  meta_data_v2,
  dataframe_level
)
}
\arguments{
\item{meta_data_dataframe}{\link{data.frame} data frame or path/url of a metadata
sheet for the data frame level}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}
}
\value{
standardized metadata sheet as data frame
}
\description{
if possible, mismatching data types are converted (\code{"true"} becomes \code{TRUE})
}
\details{
missing columns are added, filled with \code{NA}, if this is valid, i.e., n.a.
for \code{DF_NAME} as the key column
}
\examples{
\dontrun{
mds <- prep_check_meta_data_dataframe("ship_meta_dataframe|dataframe_level") # also converts
print(mds)
prep_check_meta_data_dataframe(mds)
mds1 <- mds
mds1$DF_RECORD_COUNT <- NULL
print(prep_check_meta_data_dataframe(mds1)) # fixes the missing column by NAs
mds1 <- mds
mds1$DF_UNIQUE_ROWS[[2]] <- "xxx" # not convertible
# print(prep_check_meta_data_dataframe(mds1)) # fail
mds1 <- mds
mds1$DF_UNIQUE_ID[[2]] <- 12
# print(prep_check_meta_data_dataframe(mds1)) # fail
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DF_ID_REF_TABLE}
\alias{DF_ID_REF_TABLE}
\title{Data frame level metadata attribute name}
\usage{
DF_ID_REF_TABLE
}
\description{
The name of the data frame containing the reference IDs to be
compared with the IDs in the study data set.
}
\seealso{
\link{meta_data_dataframe}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_exception_classes.R
\name{dataquieR.intrinsic_applicability_problem}
\alias{dataquieR.intrinsic_applicability_problem}
\title{An exception class assigned for exceptions caused by trying to apply
a non-applicable indicator function, which is not caused by deficient
metadata}
\usage{
dataquieR.intrinsic_applicability_problem
}
\description{
Also amending meta data could not make the function running, e.g., a test
for numbers applied to a character.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset2.R
\name{[.dataquieR_resultset2}
\alias{[.dataquieR_resultset2}
\title{Get a subset of a \code{dataquieR} \code{dq_report2} report}
\usage{
\method{[}{dataquieR_resultset2}(x, row, col, res, drop = FALSE, els = row, as_raw = FALSE)
}
\arguments{
\item{x}{the report}

\item{row}{the variable names, must be unique}

\item{col}{the function-call-names, must be unique}

\item{res}{the result slot, must be unique}

\item{drop}{drop, if length is 1}

\item{els}{used, if in list-mode with named argument}

\item{as_raw}{retrieve the result maybe as compressed \code{raw} \code{util_compress()}
serialized object}
}
\value{
a list with results, depending on \code{drop} and the number of results,
the list may contain all requested results in sub-lists. The order
of the results follows the order of the row/column/result-names given
}
\description{
Get a subset of a \code{dataquieR} \code{dq_report2} report
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_remove_from_cache.R
\name{prep_remove_from_cache}
\alias{prep_remove_from_cache}
\title{Remove a specified element from the data frame cache}
\usage{
prep_remove_from_cache(object_to_remove)
}
\arguments{
\item{object_to_remove}{\link{character} name of the object to be removed
as character string (quoted), or character vector containing the names of the
objects to remove from the cache}
}
\value{
nothing
}
\description{
Remove a specified element from the data frame cache
}
\examples{
\dontrun{
prep_load_workbook_like_file("meta_data_v2") #load metadata in the cache
ls(.dataframe_environment()) #get the list of dataframes in the cache

#remove cross-item_level from the cache
prep_remove_from_cache("cross-item_level")

#remove dataframe_level and expected_id from the cache
prep_remove_from_cache(c("dataframe_level", "expected_id"))

#remove missing_table and segment_level from the cache
x<- c("missing_table", "segment_level")
prep_remove_from_cache(x)
}

}
\seealso{
Other data-frame-cache:
\code{\link[=prep_add_data_frames]{prep_add_data_frames()}},
\code{\link[=prep_get_data_frame]{prep_get_data_frame()}},
\code{\link[=prep_list_dataframes]{prep_list_dataframes()}},
\code{\link[=prep_load_folder_with_metadata]{prep_load_folder_with_metadata()}},
\code{\link[=prep_load_workbook_like_file]{prep_load_workbook_like_file()}},
\code{\link[=prep_purge_data_frame_cache]{prep_purge_data_frame_cache()}}
}
\concept{data-frame-cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{dimensions}
\alias{dimensions}
\title{Names of DQ dimensions}
\usage{
dimensions
}
\value{
Only a definition, not a function, so no return value
}
\description{
a vector of data quality dimensions. The supported
dimensions are Completeness, Consistency and Accuracy.
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/DQconceptNew.html}{Data Quality Concept}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DF_ID_VARS}
\alias{DF_ID_VARS}
\title{Data frame level metadata attribute name}
\usage{
DF_ID_VARS
}
\description{
All variables that are to be used as one single ID
variable (combined key) in a data frame.
}
\seealso{
\link{meta_data_dataframe}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.dataquieR_result.R
\name{print.DataSlot}
\alias{print.DataSlot}
\title{Print a \code{DataSlot} object}
\usage{
\method{print}{DataSlot}(x, ...)
}
\arguments{
\item{x}{the object}

\item{...}{not used}
}
\value{
see print
}
\description{
Print a \code{DataSlot} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{UNIT_PREFIXES}
\alias{UNIT_PREFIXES}
\title{Valid unit prefixes according to \code{\link[units:valid_udunits_prefixes]{units::valid_udunits_prefixes()}}}
\description{
like k, m, M, c, ...
}
\seealso{
Other UNITS:
\code{\link{UNITS}},
\code{\link{UNIT_IS_COUNT}},
\code{\link{UNIT_PREFIX_FACTORS}},
\code{\link{UNIT_SOURCES}},
\code{\link{WELL_KNOWN_META_VARIABLE_NAMES}}
}
\concept{UNITS}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{SEGMENT_RECORD_COUNT}
\alias{SEGMENT_RECORD_COUNT}
\title{Segment level metadata attribute name}
\usage{
SEGMENT_RECORD_COUNT
}
\description{
Number of expected data records in each segment.
\link{numeric}. Check only conducted if number entered
}
\seealso{
\link{meta_data_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pro_applicability_matrix.R
\name{pro_applicability_matrix}
\alias{pro_applicability_matrix}
\title{Check applicability of DQ functions on study data}
\usage{
pro_applicability_matrix(
  study_data,
  item_level = "item_level",
  split_segments = FALSE,
  label_col,
  max_vars_per_plot = 20,
  meta_data_segment,
  meta_data_dataframe,
  flip_mode = "noflip",
  meta_data_v2,
  meta_data = item_level,
  segment_level,
  dataframe_level
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{split_segments}{\link{logical} return one matrix per study segment}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{max_vars_per_plot}{\link{integer} from=0. The maximum number of variables
per single plot.}

\item{meta_data_segment}{\link{data.frame} -- optional: Segment level metadata}

\item{meta_data_dataframe}{\link{data.frame} -- optional: Data frame level
metadata}

\item{flip_mode}{\link{enum} default | flip | noflip | auto. Should the plot be
in default orientation, flipped, not flipped or
auto-flipped. Not all options are always supported.
In general, this con be controlled by
setting the \code{roptions(dataquieR.flip_mode = ...)}. If
called from \code{dq_report}, you can also pass
\code{flip_mode} to all function calls or set them
specifically using \code{specific_args}.}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{segment_level}{\link{data.frame} alias for \code{meta_data_segment}}

\item{dataframe_level}{\link{data.frame} alias for \code{meta_data_dataframe}}
}
\value{
a list with:
\itemize{
\item \code{SummaryTable}: data frame about the applicability of each indicator
function (each function in a column).
its \link{integer} values can be one of the following four
categories:
0. Non-matching datatype + Incomplete metadata,
1. Non-matching datatype + complete metadata,
2. Matching datatype + Incomplete metadata,
3. Matching datatype + complete metadata,
4. Not applicable according to data type
\item \code{ApplicabilityPlot}: \link[ggplot2:ggplot2]{ggplot2::ggplot2} heatmap plot, graphical representation of
\code{SummaryTable}
\item \code{ApplicabilityPlotList}: \link{list} of plots per (maybe artificial) segment
\item \code{ReportSummaryTable}: data frame underlying \code{ApplicabilityPlot}
}
}
\description{
Checks applicability of DQ functions based on study data and metadata
characteristics
}
\details{
This is a preparatory support function that compares study data with
associated metadata. A prerequisite of this function is that the no. of
columns in the study data complies with the no. of rows in the metadata.

For each existing R-implementation, the function searches for necessary
static metadata and returns a heatmap like matrix indicating the
applicability of each data quality implementation.

In addition, the data type defined in the metadata is compared with the
observed data type in the study data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_internal_api.R
\name{API_VERSION}
\alias{API_VERSION}
\title{Version of the API}
\usage{
API_VERSION
}
\description{
Version of the API
}
\seealso{
\code{.get_internal_api()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_check_meta_names.R
\name{prep_check_meta_names}
\alias{prep_check_meta_names}
\title{Checks the validity of metadata w.r.t. the provided column names}
\usage{
prep_check_meta_names(
  item_level = "item_level",
  level,
  character.only = FALSE,
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{level}{\link{enum} level of requirement (see also \link{VARATT_REQUIRE_LEVELS}).
set to \code{NULL} to deactivate the check of richness.}

\item{character.only}{\link{logical} a logical indicating whether level can be
assumed to be character strings.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
a logical with:
\itemize{
\item invisible(TRUE). In case of problems with the metadata, a condition is
raised (\code{stop()}).
}
}
\description{
This function verifies, if a data frame complies to metadata conventions and
provides a given richness of meta information as specified by \code{level}.
}
\details{
Note, that only the given level is checked despite, levels are somehow
hierarchical.
}
\examples{
\dontrun{
prep_check_meta_names(data.frame(VAR_NAMES = 1, DATA_TYPE = 2,
                      MISSING_LIST = 3))

prep_check_meta_names(
  data.frame(
    VAR_NAMES = 1, DATA_TYPE = 2, MISSING_LIST = 3,
    LABEL = "LABEL", VALUE_LABELS = "VALUE_LABELS",
    JUMP_LIST = "JUMP_LIST", HARD_LIMITS = "HARD_LIMITS",
    GROUP_VAR_OBSERVER = "GROUP_VAR_OBSERVER",
    GROUP_VAR_DEVICE = "GROUP_VAR_DEVICE",
    TIME_VAR = "TIME_VAR",
    PART_VAR = "PART_VAR",
    STUDY_SEGMENT = "STUDY_SEGMENT",
    LOCATION_RANGE = "LOCATION_RANGE",
    LOCATION_METRIC = "LOCATION_METRIC",
    PROPORTION_RANGE = "PROPORTION_RANGE",
    MISSING_LIST_TABLE = "MISSING_LIST_TABLE",
    CO_VARS = "CO_VARS",
    LONG_LABEL = "LONG_LABEL"
  ),
  RECOMMENDED
)

prep_check_meta_names(
  data.frame(
    VAR_NAMES = 1, DATA_TYPE = 2, MISSING_LIST = 3,
    LABEL = "LABEL", VALUE_LABELS = "VALUE_LABELS",
    JUMP_LIST = "JUMP_LIST", HARD_LIMITS = "HARD_LIMITS",
    GROUP_VAR_OBSERVER = "GROUP_VAR_OBSERVER",
    GROUP_VAR_DEVICE = "GROUP_VAR_DEVICE",
    TIME_VAR = "TIME_VAR",
    PART_VAR = "PART_VAR",
    STUDY_SEGMENT = "STUDY_SEGMENT",
    LOCATION_RANGE = "LOCATION_RANGE",
    LOCATION_METRIC = "LOCATION_METRIC",
    PROPORTION_RANGE = "PROPORTION_RANGE",
    DETECTION_LIMITS = "DETECTION_LIMITS", SOFT_LIMITS = "SOFT_LIMITS",
    CONTRADICTIONS = "CONTRADICTIONS", DISTRIBUTION = "DISTRIBUTION",
    DECIMALS = "DECIMALS", VARIABLE_ROLE = "VARIABLE_ROLE",
    DATA_ENTRY_TYPE = "DATA_ENTRY_TYPE",
    CO_VARS = "CO_VARS",
    END_DIGIT_CHECK = "END_DIGIT_CHECK",
    VARIABLE_ORDER = "VARIABLE_ORDER", LONG_LABEL =
      "LONG_LABEL", recode = "recode",
      MISSING_LIST_TABLE = "MISSING_LIST_TABLE"
  ),
  OPTIONAL
)

# Next one will fail
try(
  prep_check_meta_names(data.frame(VAR_NAMES = 1, DATA_TYPE = 2,
    MISSING_LIST = 3), TECHNICAL)
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_storr_factory.R
\name{prep_create_storr_factory}
\alias{prep_create_storr_factory}
\title{Create a factory function for \code{storr} objects for backing
a \link{dataquieR_resultset2}}
\usage{
prep_create_storr_factory(db_dir = tempfile(), namespace = "objects")
}
\arguments{
\item{db_dir}{\link{character} path to the directory for the back-end, if
one is created on the fly.}

\item{namespace}{\link{character} namespace for the report, so that one back-end
can back several reports

the returned function will try to create a \code{storr} object using a temporary
folder or the folder in \code{db_dir}, if specified. The database will either
be the \code{storr_rds}.}
}
\value{
\code{storr} object or \code{NULL}, if package \code{storr} is not available
}
\description{
Create a factory function for \code{storr} objects for backing
a \link{dataquieR_resultset2}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DF_NAME}
\alias{DF_NAME}
\title{Data frame level metadata attribute name}
\usage{
DF_NAME
}
\description{
Name of the data frame
}
\seealso{
\link{meta_data_dataframe}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/des_summary_continuous.R
\name{des_summary_continuous}
\alias{des_summary_continuous}
\title{Compute Descriptive Statistics - continuous variables}
\usage{
des_summary_continuous(
  resp_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2,
  hard_limits_removal = getOption("dataquieR.des_summary_hard_lim_remove",
    dataquieR.des_summary_hard_lim_remove_default),
  ...
)
}
\arguments{
\item{resp_vars}{\link{variable} the name of the continuous measurement variable}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{hard_limits_removal}{\link{logical} if TRUE values outside hard limits are
removed from the data before calculating
descriptive statistics.
The default is FALSE}

\item{...}{arguments to be passed to all called indicator functions if
applicable.}
}
\value{
a \link{list} with:
\itemize{
\item \code{SummaryTable}: \link{data.frame}
\item \code{SummaryData}: \link{data.frame}
}
}
\description{
generates a descriptive overview of continuous variables (ratio and interval) in \code{resp_vars}.

\link{Descriptor}
}
\details{
TODO
}
\examples{
\dontrun{
prep_load_workbook_like_file("meta_data_v2")
xx <- des_summary_continuous(study_data = "study_data", meta_data =
                              prep_get_data_frame("item_level"))
xx$SummaryData
}
}
\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_des_impl_summary.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_prepare_dataframes.R
\name{prep_prepare_dataframes}
\alias{prep_prepare_dataframes}
\title{Prepare and verify study data with metadata}
\usage{
prep_prepare_dataframes(
  .study_data,
  .meta_data,
  .label_col,
  .replace_hard_limits,
  .replace_missings,
  .sm_code = NULL,
  .allow_empty = FALSE,
  .adjust_data_type = TRUE,
  .amend_scale_level = TRUE,
  .apply_factor_metadata = FALSE,
  .apply_factor_metadata_inadm = FALSE,
  .internal = rlang::env_inherits(rlang::caller_env(), parent.env(environment()))
)
}
\arguments{
\item{.study_data}{if provided, use this data set as study_data}

\item{.meta_data}{if provided, use this data set as meta_data}

\item{.label_col}{if provided, use this as label_col}

\item{.replace_hard_limits}{replace \code{HARD_LIMIT} violations by \code{NA},
defaults to \code{FALSE}.}

\item{.replace_missings}{replace missing codes, defaults to \code{TRUE}}

\item{.sm_code}{missing code for \code{NAs}, if they have been
re-coded by \code{util_combine_missing_lists}}

\item{.allow_empty}{allow \code{ds1} to be empty, i.e., 0 rows and/or 0 columns}

\item{.adjust_data_type}{ensure that the data type of variables in the study
data corresponds to their data type specified in the metadata}

\item{.amend_scale_level}{ensure that \code{SCALE_LEVEL} is available in the
item-level \code{meta_data}. internally used to prevent
recursion, if called from
\code{\link[=prep_scalelevel_from_data_and_metadata]{prep_scalelevel_from_data_and_metadata()}}.}

\item{.apply_factor_metadata}{\link{logical} convert categorical variables to
labeled factors.}

\item{.apply_factor_metadata_inadm}{\link{logical} convert categorical variables
to labeled factors keeping
inadmissible values. Implies, that
.apply_factor_metadata will be set
to \code{TRUE}, too.}

\item{.internal}{\link{logical} internally called, modify caller's environment.}
}
\value{
\code{ds1} the study data with mapped column names, \code{invisible()}, if
not \code{.internal}
}
\description{
This function ensures, that a data frame \code{ds1} with suitable variable
names study_data and meta_data exist as base \link{data.frame}s.
}
\details{
This function defines \code{ds1} and modifies \code{study_data} and \code{meta_data} in the
environment of its caller (see \link{eval.parent}). It also defines or modifies
the object \code{label_col} in the calling environment. Almost all functions
exported by \code{dataquieR} call this function initially, so that aspects common
to all functions live here, e.g. testing, if an argument \code{meta_data} has been
given and features really a \link{data.frame}. It verifies the existence of
required metadata attributes (\link{VARATT_REQUIRE_LEVELS}). It can also replace
missing codes by \code{NA}s, and calls \link{prep_study2meta} to generate a minimum
set of metadata from the study data on the fly (should be amended, so
on-the-fly-calling is not recommended for an instructive use of \code{dataquieR}).

The function also detects \code{tibbles}, which are then converted to base-R
\link{data.frame}s, which are expected by \code{dataquieR}.

If \code{.internal} is \code{TRUE}, differently from the other utility function that
work in their caller's environment, this function modifies objects in the
calling function's environment. It defines a new object \code{ds1},
it modifies \code{study_data} and/or \code{meta_data}
and \code{label_col}.
}
\examples{
\dontrun{
acc_test1 <- function(resp_variable, aux_variable,
                      time_variable, co_variables,
                      group_vars, study_data, meta_data) {
  prep_prepare_dataframes()
  invisible(ds1)
}
acc_test2 <- function(resp_variable, aux_variable,
                      time_variable, co_variables,
                      group_vars, study_data, meta_data, label_col) {
  ds1 <- prep_prepare_dataframes(study_data, meta_data)
  invisible(ds1)
}
environment(acc_test1) <- asNamespace("dataquieR")
# perform this inside the package (not needed for functions that have been
# integrated with the package already)

environment(acc_test2) <- asNamespace("dataquieR")
# perform this inside the package (not needed for functions that have been
# integrated with the package already)
acc_test3 <- function(resp_variable, aux_variable, time_variable,
                      co_variables, group_vars, study_data, meta_data,
                      label_col) {
  prep_prepare_dataframes()
  invisible(ds1)
}
acc_test4 <- function(resp_variable, aux_variable, time_variable,
                      co_variables, group_vars, study_data, meta_data,
                      label_col) {
  ds1 <- prep_prepare_dataframes(study_data, meta_data)
  invisible(ds1)
}
environment(acc_test3) <- asNamespace("dataquieR")
# perform this inside the package (not needed for functions that have been
# integrated with the package already)

environment(acc_test4) <- asNamespace("dataquieR")
# perform this inside the package (not needed for functions that have been
# integrated with the package already)
meta_data <- prep_get_data_frame("meta_data")
study_data <- prep_get_data_frame("study_data")
try(acc_test1())
try(acc_test2())
acc_test1(study_data = study_data)
try(acc_test1(meta_data = meta_data))
try(acc_test2(study_data = 12, meta_data = meta_data))
print(head(acc_test1(study_data = study_data, meta_data = meta_data)))
print(head(acc_test2(study_data = study_data, meta_data = meta_data)))
print(head(acc_test3(study_data = study_data, meta_data = meta_data)))
print(head(acc_test3(study_data = study_data, meta_data = meta_data,
  label_col = LABEL)))
print(head(acc_test4(study_data = study_data, meta_data = meta_data)))
print(head(acc_test4(study_data = study_data, meta_data = meta_data,
  label_col = LABEL)))
try(acc_test2(study_data = NULL, meta_data = meta_data))
}

}
\seealso{
acc_margins
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/com_item_missingness.R
\name{com_item_missingness}
\alias{com_item_missingness}
\title{Summarize missingness columnwise (in variable)}
\usage{
com_item_missingness(
  resp_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  show_causes = TRUE,
  cause_label_df,
  include_sysmiss = TRUE,
  threshold_value,
  suppressWarnings = FALSE,
  assume_consistent_codes = TRUE,
  expand_codes = assume_consistent_codes,
  drop_levels = FALSE,
  expected_observations = c("HIERARCHY", "ALL", "SEGMENT"),
  pretty_print = lifecycle::deprecated(),
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable list} the name of the measurement variables}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{show_causes}{\link{logical} if TRUE, then the distribution of missing codes
is shown}

\item{cause_label_df}{\link{data.frame} missing code table. If missing codes have
labels the respective data frame can be
specified here or in the metadata as
assignments, see \link{cause_label_df}}

\item{include_sysmiss}{\link{logical} Optional, if TRUE system missingness (NAs)
is evaluated in the summary plot}

\item{threshold_value}{\link{numeric} from=0 to=100. a numerical value ranging
from 0-100}

\item{suppressWarnings}{\link{logical} warn about consistency issues with missing
and jump lists}

\item{assume_consistent_codes}{\link{logical} if TRUE and no labels are given and
the same missing/jump code is used
for more than one variable, the
labels assigned for this code are
treated as being be the same for
all variables.}

\item{expand_codes}{\link{logical} if TRUE, code labels are copied from other
variables, if the code is the same and the
label is set somewhere}

\item{drop_levels}{\link{logical} if TRUE, do not display unused missing codes in
the figure legend.}

\item{expected_observations}{\link{enum} HIERARCHY | ALL | SEGMENT. If ALL, all
observations are expected to comprise
all study segments. If SEGMENT, the
\code{PART_VAR} is expected to point
to a variable with values of 0 and 1,
indicating whether the variable was
expected to be observed for each data
row. If HIERARCHY, this is also
checked recursively, so, if a variable
points to such a participation variable,
and that other variable does has also
a \code{PART_VAR} entry pointing
to a variable, the observation of the
initial variable is only
expected, if both segment variables are
1.}

\item{pretty_print}{\link{logical} deprecated. If you want to have a human
readable output, use \code{SummaryData} instead
of \code{SummaryTable}}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
a list with:
\itemize{
\item \code{SummaryTable}: data frame about item missingness per response variable
\item \code{SummaryData}: data frame about item missingness per response variable
formatted for user
\item \code{SummaryPlot}: ggplot2 heatmap plot, if show_causes was TRUE
\item \code{ReportSummaryTable}: data frame underlying \code{SummaryPlot}
}
}
\description{
Item-Missingness (also referred to as item nonresponse (De Leeuw et al.
2003)) describes the missingness of single values, e.g. blanks or empty data
cells in a data set. Item-Missingness occurs for example in case a respondent
does not provide information for a certain question, a question is overlooked
by accident, a programming failure occurs or a provided answer were missed
while entering the data.

\link{Indicator}
}
\section{ALGORITHM OF THIS IMPLEMENTATION:}{
\itemize{
\item Lists of missing codes and, if applicable, jump codes are selected from
the metadata
\item The no. of system missings (NA) in each variable is calculated
\item The no. of used missing codes is calculated for each variable
\item The no. of used jump codes is calculated for each variable
\item Two result dataframes (1: on the level of observations, 2: a summary for
each variable) are generated
\item \emph{OPTIONAL:} if \code{show_causes} is selected, one summary plot for all
\code{resp_vars} is provided
}
}

\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_com_impl_item_missingness.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_create_meta_data_file.R
\name{prep_create_meta_data_file}
\alias{prep_create_meta_data_file}
\title{Instantiate a new metadata file}
\usage{
prep_create_meta_data_file(
  file_name,
  study_data,
  open = TRUE,
  overwrite = FALSE
)
}
\arguments{
\item{file_name}{\link{character} file path to write to}

\item{study_data}{\link{data.frame} optional, study data to guess metadata from}

\item{open}{\link{logical} open the file after creation}

\item{overwrite}{\link{logical} overwrite \code{file}, if exists}
}
\value{
\code{invisible(NULL)}
}
\description{
Instantiate a new metadata file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_create_lean_ggplot.R,
%   R/util_lazy_ggplot.R
\name{ggplotGrob.dq_lazy_ggplot}
\alias{ggplotGrob.dq_lazy_ggplot}
\alias{ggplotly.dq_lazy_ggplot_s7}
\alias{plotly_build.dq_lazy_ggplot_s7}
\alias{ggplotly.dq_lazy_ggplot}
\alias{plotly_build.dq_lazy_ggplot}
\alias{dq_lazy_ggplot_methods}
\alias{ggplotGrob.dq_lazy_ggplot_s7}
\title{S3/S7 methods for lazy ggplot objects}
\usage{
ggplotGrob.dq_lazy_ggplot(x, ...)

ggplotly.dq_lazy_ggplot_s7(p, ...)

plotly_build.dq_lazy_ggplot_s7(p, ...)

ggplotly.dq_lazy_ggplot(p, ...)

plotly_build.dq_lazy_ggplot(p, ...)

ggplotGrob.dq_lazy_ggplot_s7(x, ...)
}
\arguments{
\item{x, p}{A \code{dq_lazy_ggplot} object.}

\item{...}{Further arguments passed on to the underlying generic.}
}
\value{
The return value is the same as for the corresponding generic:
\itemize{
\item \code{ggplotGrob()} returns a \code{gtable} object.
\item \code{ggplotly()} returns a \code{plotly} object.
\item \code{plotly_build()} returns a \code{plotly_proxy} or similar.
}
}
\description{
These S3/S7 methods make \code{dq_lazy_ggplot}/\code{dq_lazy_ggplot_s7}
objects work smoothly with
functions from \pkg{ggplot2} and \pkg{plotly}. They simply materialize
the underlying \code{ggplot} object and then delegate to the respective
generic.
}
\seealso{
\code{\link[ggplot2]{ggplotGrob}},
\code{plotly::ggplotly}
\code{plotly::plotly_build}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_exception_classes.R
\name{dataquieR.applicability_problem}
\alias{dataquieR.applicability_problem}
\title{An exception class assigned for exceptions caused by trying to apply
a non-applicable indicator function}
\usage{
dataquieR.applicability_problem
}
\description{
Amending metadata could make the function running, e.g., a test
for missingness without any declared missing codes
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in
%   R/prep_render_pie_chart_from_summaryclasses_plotly.R
\name{prep_render_pie_chart_from_summaryclasses_plotly}
\alias{prep_render_pie_chart_from_summaryclasses_plotly}
\title{Create a \code{plotly} pie chart}
\usage{
prep_render_pie_chart_from_summaryclasses_plotly(
  data,
  meta_data = "item_level"
)
}
\arguments{
\item{data}{data as returned by \code{prep_summary_to_classes} but
summarized by one column (currently, we support
\code{indicator_metric}, \code{call_names},
\code{STUDY_SEGMENT}, and \code{VAR_NAMES})}

\item{meta_data}{\link{meta_data}}
}
\value{
a \code{htmltools} compatible object
}
\description{
Create a \code{plotly} pie chart
}
\seealso{
Other summary_functions:
\code{\link[=prep_combine_report_summaries]{prep_combine_report_summaries()}},
\code{\link[=prep_extract_classes_by_functions]{prep_extract_classes_by_functions()}},
\code{\link[=prep_extract_summary]{prep_extract_summary()}},
\code{\link[=prep_extract_summary.dataquieR_result]{prep_extract_summary.dataquieR_result()}},
\code{\link[=prep_extract_summary.dataquieR_resultset2]{prep_extract_summary.dataquieR_resultset2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_ggplot2]{prep_render_pie_chart_from_summaryclasses_ggplot2()}},
\code{\link[=prep_summary_to_classes]{prep_summary_to_classes()}}
}
\concept{summary_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_list_dataframes.R
\name{prep_list_dataframes}
\alias{prep_list_dataframes}
\title{List Loaded Data Frames}
\usage{
prep_list_dataframes()
}
\value{
names of all loaded data frames
}
\description{
List Loaded Data Frames
}
\seealso{
Other data-frame-cache:
\code{\link[=prep_add_data_frames]{prep_add_data_frames()}},
\code{\link[=prep_get_data_frame]{prep_get_data_frame()}},
\code{\link[=prep_load_folder_with_metadata]{prep_load_folder_with_metadata()}},
\code{\link[=prep_load_workbook_like_file]{prep_load_workbook_like_file()}},
\code{\link[=prep_purge_data_frame_cache]{prep_purge_data_frame_cache()}},
\code{\link[=prep_remove_from_cache]{prep_remove_from_cache()}}
}
\concept{data-frame-cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_compare_meta_with_study.R
\name{prep_compare_meta_with_study}
\alias{prep_compare_meta_with_study}
\title{Verify item-level metadata}
\usage{
prep_compare_meta_with_study(
  study_data,
  label_col,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
an \code{\link[=invisible]{invisible()}} list with the entries.
\itemize{
\item \code{pred} \link{data.frame} metadata predicted from \code{study_data},
reduced to such metadata also available in the provided metadata
\item \code{prov} \link{data.frame} provided metadata,
reduced to such metadata also available in the provided \code{study_data}
\item \code{ml_error} \link{character} \code{VAR_NAMES} of variables with potentially wrong
\code{MISSING_LIST}
\item \code{sl_error} \link{character} \code{VAR_NAMES} of variables with potentially wrong
\code{SCALE_LEVEL}
\item \code{dt_error} \link{character} \code{VAR_NAMES} of variables with potentially wrong
\code{DATA_TYPE}
}
}
\description{
are the provided item-level \code{meta_data} plausible given \code{study_data}?
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DF_UNIQUE_ROWS}
\alias{DF_UNIQUE_ROWS}
\title{Data frame level metadata attribute name}
\usage{
DF_UNIQUE_ROWS
}
\description{
Specifies whether identical data is permitted across rows in a
data frame (excluding ID variables)
}
\seealso{
\link{meta_data_dataframe}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_combine_report_summaries.R
\name{prep_combine_report_summaries}
\alias{prep_combine_report_summaries}
\title{Combine two report summaries}
\usage{
prep_combine_report_summaries(..., summaries_list, amend_segment_names = FALSE)
}
\arguments{
\item{...}{objects returned by \link{prep_extract_summary}}

\item{summaries_list}{if given, \link{list} of objects returned by
\link{prep_extract_summary}}

\item{amend_segment_names}{\link{logical} use names of the \code{summaries_list} and
argument names as segment prefixes}
}
\value{
combined summaries
}
\description{
Combine two report summaries
}
\seealso{
Other summary_functions:
\code{\link[=prep_extract_classes_by_functions]{prep_extract_classes_by_functions()}},
\code{\link[=prep_extract_summary]{prep_extract_summary()}},
\code{\link[=prep_extract_summary.dataquieR_result]{prep_extract_summary.dataquieR_result()}},
\code{\link[=prep_extract_summary.dataquieR_resultset2]{prep_extract_summary.dataquieR_resultset2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_ggplot2]{prep_render_pie_chart_from_summaryclasses_ggplot2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_plotly]{prep_render_pie_chart_from_summaryclasses_plotly()}},
\code{\link[=prep_summary_to_classes]{prep_summary_to_classes()}}
}
\concept{summary_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_extract_classes_by_functions.R
\name{prep_extract_classes_by_functions}
\alias{prep_extract_classes_by_functions}
\title{Extract old function based summary from data quality results}
\usage{
prep_extract_classes_by_functions(r)
}
\arguments{
\item{r}{\link{dq_report2}}
}
\value{
\link{data.frame} long format, compatible with \code{\link[=prep_summary_to_classes]{prep_summary_to_classes()}}
}
\description{
Extract old function based summary from data quality results
}
\seealso{
Other summary_functions:
\code{\link[=prep_combine_report_summaries]{prep_combine_report_summaries()}},
\code{\link[=prep_extract_summary]{prep_extract_summary()}},
\code{\link[=prep_extract_summary.dataquieR_result]{prep_extract_summary.dataquieR_result()}},
\code{\link[=prep_extract_summary.dataquieR_resultset2]{prep_extract_summary.dataquieR_resultset2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_ggplot2]{prep_render_pie_chart_from_summaryclasses_ggplot2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_plotly]{prep_render_pie_chart_from_summaryclasses_plotly()}},
\code{\link[=prep_summary_to_classes]{prep_summary_to_classes()}}
}
\concept{summary_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_univariate_outlier.R
\name{acc_univariate_outlier}
\alias{acc_univariate_outlier}
\title{Identify univariate outliers by four different approaches}
\usage{
acc_univariate_outlier(
  resp_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  exclude_roles,
  n_rules = length(unique(criteria)),
  max_non_outliers_plot = 10000,
  criteria = c("tukey", "3sd", "hubert", "sigmagap"),
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable list} the name of the continuous measurement
variable}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{exclude_roles}{\link{variable roles} a character (vector) of variable roles
not included}

\item{n_rules}{\link{integer} from=1 to=4. the no. rules that must be violated
to flag a variable as containing outliers. The default is 4, i.e. all.}

\item{max_non_outliers_plot}{\link{integer} from=0. Maximum number of non-outlier
points to be plot. If more
points exist, a subsample will
be plotted only. Note, that
sampling is not deterministic.}

\item{criteria}{\link{set} tukey | 3SD | hubert | sigmagap. a vector with
methods to be used for detecting outliers.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
a list with:
\itemize{
\item \code{SummaryTable}: \code{\link{data.frame}} with the columns
\code{Variables}, \code{Mean}, \code{SD}, \code{Median}, \code{Skewness}, \code{Tukey (N)},
\verb{3SD (N)}, \code{Hubert (N)}, \code{Sigma-gap (N)}, \code{NUM_acc_ud_outlu},
\verb{Outliers, low (N)}, \verb{Outliers, high (N)} \code{Grading}
\itemize{
\item \code{SummaryData}: \code{\link{data.frame}} with the columns
\code{Variables}, \code{Mean}, \code{SD}, \code{Median}, \code{Skewness}, \code{Tukey (N)},
\verb{3SD (N)}, \code{Hubert (N)}, \code{Sigma-gap (N)}, \code{Outliers (N)},
\verb{Outliers, low (N)}, \verb{Outliers, high (N)}
\item \code{SummaryPlotList}: \code{\link[ggplot2:ggplot]{ggplot2::ggplot}} univariate outlier plots
}
}
}
\description{
A classical but still popular approach to detect univariate outlier is the
boxplot method introduced by Tukey 1977. The boxplot is a simple graphical
tool to display information about continuous univariate data (e.g., median,
lower and upper quartile). Outliers are defined as values deviating more
than \eqn{1.5 \times IQR} from the 1st (Q25) or 3rd (Q75) quartile. The
strength of Tukey's method is that it makes no distributional assumptions
and thus is also applicable to skewed or non mound-shaped data
Marsh and Seo, 2006. Nevertheless, this method tends to identify frequent
measurements which are falsely interpreted as true outliers.

A somewhat more conservative approach in terms of symmetric and/or normal
distributions is the 3SD approach, i.e. any measurement not in
the interval of \eqn{mean(x) +/- 3 * \sigma} is considered an outlier.

Both methods mentioned above are not ideally suited to skewed distributions.
As many biomarkers such as laboratory measurements represent in skewed
distributions the methods above may be insufficient. The approach of Hubert
and Vandervieren 2008 adjusts the boxplot for the skewness of the
distribution. This approach is implemented in several R packages such as
\code{\link[robustbase:mc]{robustbase::mc}} which is used in this implementation of \code{\link{dataquieR}}.

Another completely heuristic approach is also included to identify outliers.
The approach is based on the assumption that the distances between
measurements of the same underlying distribution should homogeneous. For
comprehension of this approach:
\itemize{
\item consider an ordered sequence of all measurements.
\item between these measurements all distances are calculated.
\item the occurrence of larger distances between two neighboring measurements
may
than indicate a distortion of the data. For the heuristic definition of a
large distance \eqn{1 * \sigma} has been been chosen.
}

Note, that the plots are not deterministic, because they use
\link[ggplot2:geom_jitter]{ggplot2::geom_jitter}.

\link{Indicator}
}
\details{
\emph{\strong{Hint}}: \emph{The function is designed for unimodal data only.}
}
\section{ALGORITHM OF THIS IMPLEMENTATION:}{
\itemize{
\item Select all variables of type float in the study data
\item Remove missing codes from the study data (if defined in the metadata)
\item Remove measurements deviating from limits defined in the metadata
\item Identify outliers according to the approaches of Tukey (Tukey 1977),
3SD (Saleem et al. 2021), Hubert (Hubert and Vandervieren 2008),
and SigmaGap (heuristic)
\item An output data frame is generated which indicates the no. possible
outliers, the direction of deviations (Outliers, low; Outliers, high) for all methods
and a summary score which sums up the deviations of the different rules
\item A scatter plot is generated for all examined variables, flagging
observations according to the no. violated rules (step 5).
}
}

\seealso{
\itemize{
\item \link{acc_robust_univariate_outlier}
\item \href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_robust_univariate_outlier.html}{Online Documentation}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.dataquieR_resultset2.R
\name{summary.dataquieR_resultset2}
\alias{summary.dataquieR_resultset2}
\title{Generate a report summary table}
\usage{
\method{summary}{dataquieR_resultset2}(
  object,
  aspect = c("applicability", "error", "anamat", "indicator_or_descriptor"),
  FUN,
  collapse = "\\n<br />\\n",
  ...
)
}
\arguments{
\item{object}{a square result set}

\item{aspect}{an aspect/problem category of results}

\item{FUN}{function to apply to the cells of the result table}

\item{collapse}{passed to \code{FUN}}

\item{...}{not used}
}
\value{
a summary of a \code{dataquieR} report
}
\description{
Generate a report summary table
}
\examples{
\dontrun{
  util_html_table(summary(report),
       filter = "top", options = list(scrollCollapse = TRUE, scrollY = "75vh"),
       is_matrix_table = TRUE, rotate_headers = TRUE
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_end_digits.R
\name{acc_end_digits}
\alias{acc_end_digits}
\title{Extension of \link{acc_shape_or_scale} to examine uniform distributions of
end digits}
\usage{
acc_end_digits(
  resp_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable} the names of the measurement variables,
mandatory}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
a \link{list} with:
\itemize{
\item \code{SummaryTable}: \link{data.frame} with the columns \code{Variables} and \code{FLG_acc_ud_shape}
\item \code{SummaryPlot}: ggplot2 distribution plot comparing expected
with observed distribution
}
}
\description{
This implementation contrasts the empirical distribution of a measurement
variables against assumed distributions. The approach is adapted from the
idea of rootograms (Tukey (1977)) which is also applicable for count data
(Kleiber and Zeileis (2016)).

\link{Indicator}
}
\section{ALGORITHM OF THIS IMPLEMENTATION:}{
\itemize{
\item This implementation is restricted to data of type float or integer.
\item Missing codes are removed from resp_vars (if defined in the metadata)
\item The user must specify the column of the metadata containing probability
distribution (currently only: normal, uniform, gamma)
\item Parameters of each distribution can be estimated from the data or are
specified by the user
\item A histogram-like plot contrasts the empirical vs. the technical
distribution
}
}

\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_end_digits.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{CODE_CLASSES}
\alias{CODE_CLASSES}
\title{types of value codes}
\usage{
CODE_CLASSES
}
\description{
types of value codes
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_split_val_tab.R
\name{prep_unsplit_val_tabs}
\alias{prep_unsplit_val_tabs}
\title{Combine all missing and value lists to one big table}
\usage{
prep_unsplit_val_tabs(meta_data = "item_level", val_tab = NULL)
}
\arguments{
\item{meta_data}{\link{data.frame} item level meta data to be used, defaults to
\code{"item_level"}}

\item{val_tab}{\link{character} name of the table being created: This table will
be added to the data frame cache (or overwritten). If \code{NULL}, the table
will only be returned}
}
\value{
\link{data.frame} the combined table
}
\description{
Combine all missing and value lists to one big table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_check_for_dataquieR_updates.R
\name{prep_check_for_dataquieR_updates}
\alias{prep_check_for_dataquieR_updates}
\title{Check for package updates}
\usage{
prep_check_for_dataquieR_updates(
  beta = FALSE,
  deps = TRUE,
  ask = interactive()
)
}
\arguments{
\item{beta}{\link{logical} check for beta version too}

\item{deps}{\link{logical} check for missing (optional) dependencies}

\item{ask}{\link{logical} ask for updates}
}
\value{
\code{invisible(NULL)}
}
\description{
Check for package updates
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/meta_data_env.R
\name{meta_data_env_co_vars}
\alias{meta_data_env_co_vars}
\title{Extract co-variables for a given item}
\arguments{
\item{entity}{vector of item-identifiers}
}
\value{
a vector with co-variables for each entity-entry, having the
\code{explode} attribute set to \code{FALSE}
}
\description{
Extract co-variables for a given item
}
\seealso{
\link{meta_data_env}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dq_report.R
\name{dq_report}
\alias{dq_report}
\title{Generate a full DQ report}
\usage{
dq_report(...)
}
\arguments{
\item{...}{Deprecated}
}
\value{
Deprecated
}
\description{
Deprecated
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_add_to_meta.R
\name{prep_add_to_meta}
\alias{prep_add_to_meta}
\title{Support function to augment metadata during data quality reporting}
\usage{
prep_add_to_meta(
  VAR_NAMES,
  DATA_TYPE,
  LABEL,
  VALUE_LABELS,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2,
  ...
)
}
\arguments{
\item{VAR_NAMES}{\link{character} Names of the Variables to add}

\item{DATA_TYPE}{\link{character} Data type for the added variables}

\item{LABEL}{\link{character} Labels for these variables}

\item{VALUE_LABELS}{\link{character} Value labels for the values of the variables
as usually pipe separated and assigned with
\code{=}: \code{1 = male | 2 = female}}

\item{item_level}{\link{data.frame} the metadata to extend}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}

\item{...}{Further defined variable attributes, see
\link{prep_create_meta}}
}
\value{
a data frame with amended metadata.
}
\description{
adds an annotation to static metadata
}
\details{
Add metadata e.g. of transformed/new variable
This function is not yet considered stable, but we already export it,
because it could help. Therefore, we have some inconsistencies in the
formals still.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_save_report.R
\name{prep_load_report}
\alias{prep_load_report}
\title{Load a \code{dq_report2}}
\usage{
prep_load_report(file)
}
\arguments{
\item{file}{\link{character} the file name to load from}
}
\value{
\link{dataquieR_resultset2} the report
}
\description{
Load a \code{dq_report2}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_meta_data_v1_to_item_level_meta_data.R
\name{prep_meta_data_v1_to_item_level_meta_data}
\alias{prep_meta_data_v1_to_item_level_meta_data}
\title{Convert item-level metadata from v1.0 to v2.0}
\usage{
prep_meta_data_v1_to_item_level_meta_data(
  item_level = "item_level",
  verbose = TRUE,
  label_col = LABEL,
  cause_label_df,
  meta_data = item_level
)
}
\arguments{
\item{item_level}{\link{data.frame} the old item-level-metadata}

\item{verbose}{\link{logical} display all estimated decisions, defaults to \code{TRUE},
except if called in a \link{dq_report2} pipeline.}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{cause_label_df}{\link{data.frame} missing code table, see \link{cause_label_df}.
Optional. If this argument is given,
you can add missing code tables.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}
}
\value{
\link{data.frame} the updated metadata
}
\description{
This function is idempotent..
}
\details{
The \code{options("dataquieR.force_item_specific_missing_codes")} (default
\code{FALSE}) tells the system, to always fill in \code{res_vars} columns to the
\code{MISSING_LIST_TABLE}, even, if the column already exists, but is empty.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_load_folder_with_metadata.R
\name{prep_load_folder_with_metadata}
\alias{prep_load_folder_with_metadata}
\title{Pre-load a folder with named (usually more than) one table(s)}
\usage{
prep_load_folder_with_metadata(folder, keep_types = FALSE, append = FALSE, ...)
}
\arguments{
\item{folder}{the folder name to load.}

\item{keep_types}{\link{logical} keep types as possibly defined in the file.
set \code{TRUE} for study data.}

\item{append}{\link{logical} if a data frame already exists in the cache
(by name), extend the existing one}

\item{...}{arguments passed to \code{\link[=list.files]{list.files()}}}
}
\value{
\verb{invisible(the cache environment)}
}
\description{
The original purpose of this function is to load metadata, not study data.
If you want to load study data, you should keep them in a different folder,
then you can call this function once for the metadata and once for the study
data but this time setting \code{keep_types = TRUE} to avoid all data being read
as \code{\link[=character]{character()}}.
}
\details{
Note, that once loaded to the data frame cache, a file won't be read again,
except you call \code{\link[=prep_purge_data_frame_cache]{prep_purge_data_frame_cache()}} or
\code{\link[=prep_remove_from_cache]{prep_remove_from_cache()}}. That is, if you call this function first, and
\code{\link[=prep_get_data_frame]{prep_get_data_frame()}} later, of if \code{dataquieR} wants to read a file, e.g.,
for \code{\link[=dq_report2]{dq_report2()}}, the file will come from the cache in the way it was
initially read in (\code{keep_types} may thus be used inadequately).

By default, this function works not recursively, but you can tweak that by
passing \code{...}-arguments passed through to the initially running
\code{\link[=list.files]{list.files()}} function.

These can thereafter be referred to by their names only. Such files are,
e.g., spreadsheet-workbooks or \code{RData}-files.

Note, that this function in contrast to \link{prep_get_data_frame} does neither
support selecting specific sheets/columns from a file.
}
\seealso{
\link{prep_add_data_frames}

\link{prep_get_data_frame}

Other data-frame-cache:
\code{\link[=prep_add_data_frames]{prep_add_data_frames()}},
\code{\link[=prep_get_data_frame]{prep_get_data_frame()}},
\code{\link[=prep_list_dataframes]{prep_list_dataframes()}},
\code{\link[=prep_load_workbook_like_file]{prep_load_workbook_like_file()}},
\code{\link[=prep_purge_data_frame_cache]{prep_purge_data_frame_cache()}},
\code{\link[=prep_remove_from_cache]{prep_remove_from_cache()}}
}
\concept{data-frame-cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{VARIABLE_ROLES}
\alias{VARIABLE_ROLES}
\title{Variable roles can be one of the following:}
\usage{
VARIABLE_ROLES
}
\description{
\itemize{
\item \code{intro} a variable holding consent-data
\item \code{primary} a primary outcome variable
\item \code{secondary} a secondary outcome variable
\item \code{process} a variable describing the measurement process
\item \code{suppress} a variable added on the fly computing sub-reports, i.e., by
\link{dq_report_by} to have all referred variables available,
even if they are not part of the currently processed segment.
But they will only be fully assessed in their real segment's
report.
}
}
\alias{variable roles}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.dataquieR_result.R
\name{print.master_result}
\alias{print.master_result}
\title{Print a \code{master_result} object}
\usage{
\method{print}{master_result}(x, template = "default", ...)
}
\arguments{
\item{x}{the object}

\item{template}{the template for the \code{iframes}, not used, so far.}

\item{...}{not used}
}
\value{
\code{invisible(NULL)}
}
\description{
Print a \code{master_result} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{DATA_TYPES_OF_R_TYPE}
\alias{DATA_TYPES_OF_R_TYPE}
\title{All available data types, mapped from their respective
R types}
\usage{
DATA_TYPES_OF_R_TYPE
}
\description{
All available data types, mapped from their respective
R types
}
\seealso{
\code{\link{prep_dq_data_type_of}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_contradiction_rules.R
\name{contradiction_functions_descriptions}
\alias{contradiction_functions_descriptions}
\title{description of the contradiction functions}
\usage{
contradiction_functions_descriptions
}
\description{
description of the contradiction functions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_merge_study_data.R
\name{prep_merge_study_data}
\alias{prep_merge_study_data}
\title{Merge a list of study data frames to one (sparse) study data frame}
\usage{
prep_merge_study_data(study_data_list)
}
\arguments{
\item{study_data_list}{\link{list} the list}
}
\value{
\link{data.frame} \link{study_data}
}
\description{
Merge a list of study data frames to one (sparse) study data frame
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_shape_or_scale.R
\name{acc_shape_or_scale}
\alias{acc_shape_or_scale}
\title{Compare observed versus expected distributions}
\usage{
acc_shape_or_scale(
  resp_vars,
  study_data,
  label_col,
  item_level = "item_level",
  dist_col,
  guess,
  par1,
  par2,
  end_digits,
  flip_mode = "noflip",
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable} the name of the continuous measurement variable}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{dist_col}{\link{variable attribute} the name of the variable attribute in
meta_data that provides the expected
distribution of a study variable}

\item{guess}{\link{logical} estimate parameters}

\item{par1}{\link{numeric} first parameter of the distribution if applicable}

\item{par2}{\link{numeric} second parameter of the distribution if applicable}

\item{end_digits}{\link{logical} internal use. check for end digits preferences}

\item{flip_mode}{\link{enum} default | flip | noflip | auto. Should the plot be
in default orientation, flipped, not flipped or
auto-flipped. Not all options are always supported.
In general, this con be controlled by
setting the \code{roptions(dataquieR.flip_mode = ...)}. If
called from \code{dq_report}, you can also pass
\code{flip_mode} to all function calls or set them
specifically using \code{specific_args}.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
a list with:
\itemize{
\item \code{ResultData}: \link{data.frame} underlying the plot
\item \code{SummaryPlot}: \link[ggplot2:ggplot2]{ggplot2::ggplot2} probability distribution plot
\item \code{SummaryTable}: \link{data.frame} with the columns \code{Variables} and \code{FLG_acc_ud_shape}
}
}
\description{
This implementation contrasts the empirical distribution of a measurement
variables against assumed distributions. The approach is adapted from the
idea of rootograms (Tukey 1977) which is also applicable for count data
(Kleiber and Zeileis 2016).

\link{Indicator}
}
\section{ALGORITHM OF THIS IMPLEMENTATION:}{
\itemize{
\item This implementation is restricted to data of type float or integer.
\item Missing codes are removed from resp_vars (if defined in the metadata)
\item The user must specify the column of the metadata containing probability
distribution (currently only: normal, uniform, gamma)
\item Parameters of each distribution can be estimated from the data or are
specified by the user
\item A histogram-like plot contrasts the empirical vs. the technical
distribution
}
}

\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_shape_or_scale.html}{Online Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset2.R
\name{[[<-.dataquieR_resultset2}
\alias{[[<-.dataquieR_resultset2}
\title{Set a single result from a \verb{dataquieR 2} report}
\usage{
\method{[[}{dataquieR_resultset2}(x, el) <- value
}
\arguments{
\item{x}{the report}

\item{el}{the index}

\item{value}{the single result}
}
\value{
the \code{dataquieR} result object
}
\description{
Set a single result from a \verb{dataquieR 2} report
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_pairs_matrix.R
\name{print.util_pairs_ggplot_panels}
\alias{print.util_pairs_ggplot_panels}
\title{Print method for \code{util_pairs_ggplot_panels} objects}
\usage{
\method{print}{util_pairs_ggplot_panels}(x, ...)
}
\arguments{
\item{x}{An object of class \code{util_pairs_ggplot_panels}.}

\item{...}{Ignored.}
}
\value{
The input object, invisibly.
}
\description{
Print method for \code{util_pairs_ggplot_panels} objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reflection.R
\docType{data}
\name{Indicator}
\alias{Indicator}
\title{Indicator Function}
\usage{
Indicator
}
\description{
A function that returns some value that correlates with the magnitude of
a certain class of data quality problems. Typically, in \code{dataquieR}, such
functions return a \code{SummaryTable} that features columns with names, that
start with a short abbreviation that describes the specific semantics of
the value (e.g., \code{PCT} for a percentage or \code{COR} for a correlation) and
the public name of the indicator according to the data quality concept
\code{DQ_OBS}, e.g., \code{com_qum_nonresp} for item-non-response-rate. A name could
therefore be \code{PCT_com_qum_nonresp}.

The object \code{Indicator} only contains the name used internally to tag
such functions.
}
\seealso{
\link{Descriptor}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataquieR_resultset_verify.R
\name{dataquieR_resultset_verify}
\alias{dataquieR_resultset_verify}
\title{Verify an object of class dataquieR_resultset}
\usage{
dataquieR_resultset_verify(...)
}
\arguments{
\item{...}{Deprecated}
}
\value{
Deprecated
}
\description{
Deprecated
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/menu_env.R
\name{menu_env_menu_entry}
\alias{menu_env_menu_entry}
\title{Create a single menu entry}
\arguments{
\item{title}{of the entry}

\item{id}{linked \code{href}, defaults to modified title. can be a word, then
a single-page-link with an anchor tag is created.}

\item{...}{additional arguments for the menu link}
}
\value{
html-a-tag object
}
\description{
Create a single menu entry
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{CODE_LIST_TABLE}
\alias{CODE_LIST_TABLE}
\title{Default Name of the Table featuring Code Lists}
\usage{
CODE_LIST_TABLE

CODE_LIST_TABLE
}
\description{
Default Name of the Table featuring Code Lists

Metadata sheet name containing VALUE_LABEL_TABLES
This metadata sheet can contain both value labels of several
VALUE_LABEL_TABLE and also Missing and JUMP tables
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_add_data_frames.R
\name{prep_add_data_frames}
\alias{prep_add_data_frames}
\title{Add data frames to the pre-loaded / cache data frame environment}
\usage{
prep_add_data_frames(..., data_frame_list = list(), append = FALSE)
}
\arguments{
\item{...}{data frames, if passed with names, these will be the names
of these tables in the data frame environment. If not, then the
names in the calling environment will be used.}

\item{data_frame_list}{a named list with data frames. Also these will be
added and names will be handled as for the \code{...}
argument.}

\item{append}{\link{logical} if a data frame already exists in the cache
(by name), extend the existing one}
}
\value{
\link{data.frame} \verb{invisible(the cache environment)}
}
\description{
These can be referred to by their names, then, wherever \code{dataquieR} expects
a \link{data.frame} -- just pass a character instead. If this character is not
found, \code{dataquieR} would additionally look for files with the name and for
\code{URLs}. You can also refer to specific sheets of a workbook or specific
object from an \code{RData} by appending a pipe symbol and its name. A second
pipe symbol allows to extract certain columns from such sheets (but
they will remain data frames).
}
\seealso{
\link{prep_load_workbook_like_file}

\link{prep_get_data_frame}

Other data-frame-cache:
\code{\link[=prep_get_data_frame]{prep_get_data_frame()}},
\code{\link[=prep_list_dataframes]{prep_list_dataframes()}},
\code{\link[=prep_load_folder_with_metadata]{prep_load_folder_with_metadata()}},
\code{\link[=prep_load_workbook_like_file]{prep_load_workbook_like_file()}},
\code{\link[=prep_purge_data_frame_cache]{prep_purge_data_frame_cache()}},
\code{\link[=prep_remove_from_cache]{prep_remove_from_cache()}}
}
\concept{data-frame-cache}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{UNIT_IS_COUNT}
\alias{UNIT_IS_COUNT}
\title{Is a unit a count according to \code{\link[units:valid_udunits]{units::valid_udunits()}}}
\description{
see column \code{def}, therein
}
\details{
like \verb{\%}, \code{ppt}, \code{ppm}
}
\seealso{
Other UNITS:
\code{\link{UNITS}},
\code{\link{UNIT_PREFIXES}},
\code{\link{UNIT_PREFIX_FACTORS}},
\code{\link{UNIT_SOURCES}},
\code{\link{WELL_KNOWN_META_VARIABLE_NAMES}}
}
\concept{UNITS}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{MAHALANOBIS_RATIO}
\alias{MAHALANOBIS_RATIO}
\title{Cross-item level metadata attribute name}
\description{
TODO
}
\seealso{
\link{meta_data_cross}

Other meta_data_cross:
\code{\link{ASSOCIATION_DIRECTION}},
\code{\link{ASSOCIATION_FORM}},
\code{\link{ASSOCIATION_METRIC}},
\code{\link{ASSOCIATION_RANGE}},
\code{\link{CHECK_ID}},
\code{\link{CHECK_LABEL}},
\code{\link{COMPUTED_VARIABLE_ROLES}},
\code{\link{CONTRADICTION_TERM}},
\code{\link{CONTRADICTION_TYPE}},
\code{\link{DATA_PREPARATION}},
\code{\link{GOLDSTANDARD}},
\code{\link{IRV}},
\code{\link{MAHALANOBIS_THRESHOLD}},
\code{\link{MAXIMUM_LONG_STRING}},
\code{\link{MISS_RESP}},
\code{\link{MULTIVARIATE_OUTLIER_CHECK}},
\code{\link{MULTIVARIATE_OUTLIER_CHECKTYPE}},
\code{\link{RELCOMPL_SPEED}},
\code{\link{REL_VAL}},
\code{\link{RESPT_PER_ITEM}},
\code{\link{SCALE_ACRONYM}},
\code{\link{SCALE_NAME}},
\code{\link{TOTRESPT}},
\code{\link{VARIABLE_LIST}},
\code{\link{VARIABLE_LIST_ORDER}},
\code{\link{meta_data_computation}},
\code{\link{meta_data_cross}}

Other SSI:
\code{\link{COMPUTED_VARIABLE_ROLES}},
\code{\link{IRV}},
\code{\link{MAXIMUM_LONG_STRING}},
\code{\link{MISS_RESP}},
\code{\link{RELCOMPL_SPEED}},
\code{\link{RESPT_PER_ITEM}},
\code{\link{TOTRESPT}}
}
\concept{SSI}
\concept{meta_data_cross}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/meta_data_env.R
\name{meta_data_env}
\alias{meta_data_env}
\alias{.meta_data_env}
\title{\code{.meta_data_env} -- an environment for easy metadata access}
\usage{
.meta_data_env
}
\description{
used by the dq_report2-pipeline
}
\seealso{
\code{meta_data_env_id_vars()} \code{meta_data_env_co_vars()}
\code{meta_data_env_time_vars()} \code{meta_data_env_group_vars()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/000_globs.R
\name{SEGMENT_MISS}
\alias{SEGMENT_MISS}
\title{Segment level metadata attribute name}
\usage{
SEGMENT_MISS
}
\description{
\code{true} or \code{false} to suppress crude segment missingness output
(\verb{Completeness/Misg. Segments} in the report). Defaults to compute
the output, if more than one segment is available in the item-level
metadata.
}
\seealso{
\link{meta_data_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prep_extract_summary.R
\name{prep_extract_summary}
\alias{prep_extract_summary}
\title{Extract summary from data quality results}
\usage{
prep_extract_summary(r, ...)
}
\arguments{
\item{r}{\link{dq_report2} or \link{dataquieR_result} object}

\item{...}{further arguments, maybe needed for some implementations}
}
\value{
\link{list} with two slots \code{Data} and \code{Table} with \link{data.frame}s
featuring all metrics columns
from the report or result in \code{x},
the \link{STUDY_SEGMENT} and the \link{VAR_NAMES}.
In case of \code{Data}, the columns are formatted nicely but still
with the standardized column names -- use
\code{util_translate_indicator_metrics()} to rename them nicely. In
case of \code{Table}, just as they are.
}
\description{
Generic function, currently supports \link{dq_report2} and \link{dataquieR_result}
}
\seealso{
Other summary_functions:
\code{\link[=prep_combine_report_summaries]{prep_combine_report_summaries()}},
\code{\link[=prep_extract_classes_by_functions]{prep_extract_classes_by_functions()}},
\code{\link[=prep_extract_summary.dataquieR_result]{prep_extract_summary.dataquieR_result()}},
\code{\link[=prep_extract_summary.dataquieR_resultset2]{prep_extract_summary.dataquieR_resultset2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_ggplot2]{prep_render_pie_chart_from_summaryclasses_ggplot2()}},
\code{\link[=prep_render_pie_chart_from_summaryclasses_plotly]{prep_render_pie_chart_from_summaryclasses_plotly()}},
\code{\link[=prep_summary_to_classes]{prep_summary_to_classes()}}
}
\concept{summary_functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_varcomp.R
\name{acc_varcomp}
\alias{acc_varcomp}
\title{Utility function to compute model-based ICC depending on the (statistical) data type}
\usage{
acc_varcomp(
  resp_vars = NULL,
  group_vars = NULL,
  co_vars = NULL,
  study_data,
  label_col,
  item_level = "item_level",
  min_obs_in_subgroup = 10,
  min_subgroups = 5,
  cut_off_linear_model_for_ord = 10,
  threshold_value = lifecycle::deprecated(),
  meta_data = item_level,
  meta_data_v2
)
}
\arguments{
\item{resp_vars}{\link{variable} the name of the measurement variable}

\item{group_vars}{\link{variable} the name of the examiner, device or
reader variable}

\item{co_vars}{\link{variable list} a vector of covariables, e.g. age and sex,
for adjustment}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the metadata
with labels of variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{min_obs_in_subgroup}{\link{integer} from=0. This optional argument specifies
the minimum number of observations that is
required to include a subgroup (level) of the
\code{group_var} in the analysis. Subgroups with less
observations are excluded.}

\item{min_subgroups}{\link{integer} from=0. This optional argument specifies
the minimum number of subgroups (level) of the
\code{group_var} that is required to run the
analysis. If there are less subgroups, the
analysis is not conducted.}

\item{cut_off_linear_model_for_ord}{\link{integer} from=0. This optional argument
specifies the minimum number of observations for
individual levels of an ordinal outcome
(\code{resp_var}) that is required to run a linear
mixed effects model instead of a mixed effects
ordered regression (i.e., a cut-off value above
which linear models are considered a good
approximation). The argument can be set to
\code{NULL} if ordered regression models are
preferred for ordinal data in any case.}

\item{threshold_value}{Deprecated.}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path to workbook like metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}.}
}
\value{
The function returns two data frames, 'SummaryTable' and
'SummaryData', that differ only in the names of the columns.
}
\description{
This function is still under construction. It is designed to run for any
statistical data type as follows:
\itemize{
\item Variables with only two distinct values will be modeled by mixed effects
logistic regression.
\item Nominal variables will be transformed to binary variables. This can be
user-specified using the metadata columns \code{RECODE_CASES} and/or
\code{RECODE_CONTROL}. Otherwise, the most frequent category will be assigned
to cases and the remaining categories to control. As for other binary
variables, the ICC will be computed using a mixed effects logistic
regression.
\item Ordinal variables will be analyzed by linear mixed effects models, if
every level of the variable has at least as many observations as
specified in the argument \code{cut_off_linear_model_for_ord}. Otherwise, the
data will be modeled by a mixed effects ordered regression, if the
package \code{ordinal} is available.
\item Metric variables with integer values are analyzed by linear mixed
effects models.
\item For variables with data type \code{float}, the existing implementation
\code{acc_varcomp} is called, which also uses linear mixed effects models.
}

\link{Indicator}
}
\details{
Not yet described
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util_as_numeric_with_unit.R
\name{print.numeric_with_unit}
\alias{print.numeric_with_unit}
\title{Print a number with unit}
\usage{
\method{print}{numeric_with_unit}(x, ...)
}
\arguments{
\item{x}{number with unit}

\item{...}{not used}
}
\value{
\code{invisible(x)}
}
\description{
Print a number with unit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acc_mahalanobis_ratio.R
\name{acc_mahalanobis_ratio}
\alias{acc_mahalanobis_ratio}
\title{Internal function only existing for technical reasons, planned to be removed in
future releases}
\usage{
acc_mahalanobis_ratio(
  resp_vars = NULL,
  study_data,
  label_col = VAR_NAMES,
  item_level = "item_level",
  meta_data = item_level,
  meta_data_v2,
  meta_data_cross_item = "cross-item_level",
  cross_item_level,
  `cross-item_level`
)
}
\arguments{
\item{resp_vars}{\link{variable} the names of the computed variable
containing \code{Mahalanobis} distance ratio}

\item{study_data}{\link{data.frame} the data frame that contains the measurements}

\item{label_col}{\link{variable attribute} the name of the column in the
metadata containing the labels of
the variables}

\item{item_level}{\link{data.frame} the data frame that contains metadata
attributes of study data}

\item{meta_data}{\link{data.frame} old name for \code{item_level}}

\item{meta_data_v2}{\link{character} path or file name of the workbook like
metadata file, see
\code{\link{prep_load_workbook_like_file}} for details.
\strong{ALL LOADED DATAFRAMES WILL BE PURGED},
using \code{\link{prep_purge_data_frame_cache}},
if you specify \code{meta_data_v2}}

\item{meta_data_cross_item}{\link{data.frame} -- Cross-item level metadata}

\item{cross_item_level}{\link{data.frame} alias for \code{meta_data_cross_item}}

\item{`cross-item_level`}{\link{data.frame} alias for \code{meta_data_cross_item}}
}
\value{
a list with:
\itemize{
\item \code{SummaryData}: \link{data.frame} underlying the plot with user friendly caption
\item \code{SummaryTable}: \link{data.frame} underlying the plot
\item \code{SummaryPlot}: \link[ggplot2:ggplot2]{ggplot2::ggplot2} Q-Q plot of squared \code{Mahalanobis}
distances vs. a theoretical
chi-squared distribution showing \code{outliers}.
\item \code{FlaggedStudyData} \link{data.frame} contains the original data frame of the
variables used to calculate
the squared \code{Mahalanobis} distances
with an additional column indicating if
for a group of variables if the
observational unit is a
multivariate \code{outlier}.
}
}
\description{
Please use instead the function \code{\link[=acc_mahalanobis]{acc_mahalanobis()}}

\link{Indicator}
}
\details{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\section{ALGORITHM OF THIS IMPLEMENTATION:}{
\itemize{
\item Implementation is restricted to variables of type integer
\item Remove missing codes from the study data (if defined in the metadata)
\item The covariance matrix is estimated for all variables of \code{resp_vars}
\item The \code{Mahalanobis} distance of each observation is calculated
\eqn{MD^2_i  = (x_i - \mu)^T \Sigma^{-1} (x_i -  \mu)}
\item The default to consider a value an \code{outlier} is to use the 0.975 quantile
of a theoretical chi-square distribution with degrees of freedom
equals to the number of variables used to calculate the
\code{Mahalanobis} distance (\verb{Mayrhofer and Filzmoser}, 2023)
}
}

\seealso{
\href{https://dataquality.qihs.uni-greifswald.de/VIN_acc_impl_multivariate_outlier.html}{Online Documentation}
}
